/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 22831
#define LARGE_STATE_COUNT 8933
#define SYMBOL_COUNT 514
#define ALIAS_COUNT 4
#define TOKEN_COUNT 221
#define EXTERNAL_TOKEN_COUNT 34
#define FIELD_COUNT 28
#define MAX_ALIAS_SEQUENCE_LENGTH 11
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 102
#define SUPERTYPE_COUNT 6

enum ts_symbol_identifiers {
  sym_identifier = 1,
  anon_sym_namespace = 2,
  anon_sym_global = 3,
  anon_sym_rec = 4,
  anon_sym_module = 5,
  anon_sym_null = 6,
  anon_sym_EQ = 7,
  anon_sym_begin = 8,
  anon_sym_end = 9,
  anon_sym_open = 10,
  anon_sym_type = 11,
  anon_sym_LBRACK_LT = 12,
  anon_sym_GT_RBRACK = 13,
  anon_sym_COLON = 14,
  anon_sym_do = 15,
  anon_sym_and = 16,
  anon_sym_let = 17,
  anon_sym_let_BANG = 18,
  anon_sym_inline = 19,
  anon_sym_mutable = 20,
  aux_sym_access_modifier_token1 = 21,
  anon_sym_as = 22,
  anon_sym_LPAREN = 23,
  anon_sym_RPAREN = 24,
  anon_sym_COMMA = 25,
  anon_sym__ = 26,
  anon_sym_QMARK = 27,
  anon_sym_COLON_QMARK = 28,
  anon_sym_COLON_COLON = 29,
  anon_sym_PIPE = 30,
  anon_sym_AMP = 31,
  anon_sym_GT = 32,
  anon_sym_LBRACK = 33,
  anon_sym_RBRACK = 34,
  anon_sym_LBRACK_PIPE = 35,
  anon_sym_PIPE_RBRACK = 36,
  anon_sym_LBRACE = 37,
  anon_sym_RBRACE = 38,
  anon_sym_LT_AT = 39,
  anon_sym_LT_AT_AT = 40,
  anon_sym_CARET = 41,
  anon_sym_or = 42,
  anon_sym_DOT = 43,
  anon_sym_LBRACE_PIPE = 44,
  anon_sym_PIPE_RBRACE = 45,
  anon_sym_new = 46,
  anon_sym_with = 47,
  anon_sym_return = 48,
  anon_sym_return_BANG = 49,
  anon_sym_yield = 50,
  anon_sym_yield_BANG = 51,
  anon_sym_lazy = 52,
  anon_sym_assert = 53,
  anon_sym_upcast = 54,
  anon_sym_downcast = 55,
  anon_sym_fixed = 56,
  anon_sym_COLON_GT = 57,
  anon_sym_COLON_QMARK_GT = 58,
  anon_sym_for = 59,
  anon_sym_in = 60,
  anon_sym_to = 61,
  anon_sym_downto = 62,
  anon_sym_done = 63,
  anon_sym_while = 64,
  anon_sym_while_BANG = 65,
  anon_sym_else = 66,
  anon_sym_then = 67,
  anon_sym_elif = 68,
  anon_sym_if = 69,
  anon_sym_fun = 70,
  anon_sym_DASH_GT = 71,
  anon_sym_try = 72,
  anon_sym_finally = 73,
  anon_sym_match = 74,
  anon_sym_match_BANG = 75,
  anon_sym_function = 76,
  anon_sym_LT_DASH = 77,
  anon_sym_DOT_LBRACK = 78,
  anon_sym_use = 79,
  anon_sym_use_BANG = 80,
  anon_sym_and_BANG = 81,
  anon_sym_do_BANG = 82,
  anon_sym_DOT_DOT = 83,
  anon_sym_when = 84,
  anon_sym_LPAREN2 = 85,
  anon_sym_PIPE_PIPE = 86,
  anon_sym_AMP_AMP = 87,
  anon_sym_DOT_DOT2 = 88,
  anon_sym_DOT_DOT3 = 89,
  anon_sym_STAR = 90,
  anon_sym_1 = 91,
  anon_sym_SLASH = 92,
  anon_sym_LT = 93,
  anon_sym_struct = 94,
  anon_sym_LBRACK_RBRACK = 95,
  anon_sym_POUND = 96,
  anon_sym_SQUOTE = 97,
  anon_sym_unit = 98,
  anon_sym_not = 99,
  anon_sym_enum = 100,
  anon_sym_unmanaged = 101,
  anon_sym_equality = 102,
  anon_sym_comparison = 103,
  anon_sym_delegate = 104,
  anon_sym_default = 105,
  anon_sym_static = 106,
  anon_sym_member = 107,
  anon_sym_get = 108,
  anon_sym_set = 109,
  anon_sym_exception = 110,
  anon_sym_of = 111,
  anon_sym_class = 112,
  anon_sym_interface = 113,
  anon_sym_abstract = 114,
  anon_sym_val = 115,
  anon_sym_override = 116,
  anon_sym_extern = 117,
  anon_sym_inherit = 118,
  anon_sym_EQ2 = 119,
  sym__escape_char = 120,
  sym__non_escape_char = 121,
  sym__unicodegraph_short = 122,
  sym__unicodegraph_long = 123,
  sym__trigraph = 124,
  sym__hexgraph_short = 125,
  aux_sym__simple_string_char_token1 = 126,
  aux_sym_char_token1 = 127,
  anon_sym_LBRACE2 = 128,
  anon_sym_DOLLAR_DQUOTE = 129,
  anon_sym_DQUOTE = 130,
  anon_sym_BSLASH = 131,
  aux_sym__verbatim_string_char_token1 = 132,
  anon_sym_AT_DQUOTE = 133,
  anon_sym_DQUOTE2 = 134,
  anon_sym_DQUOTEB = 135,
  anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE = 136,
  anon_sym_DQUOTE_DQUOTE_DQUOTE = 137,
  sym_bool = 138,
  sym_unit = 139,
  anon_sym_LPAREN_PIPE = 140,
  anon_sym_PIPE_RPAREN = 141,
  sym_op_identifier = 142,
  anon_sym_PLUS = 143,
  anon_sym_DASH = 144,
  anon_sym_PLUS_DOT = 145,
  anon_sym_DASH_DOT = 146,
  anon_sym_PERCENT = 147,
  anon_sym_PERCENT_PERCENT = 148,
  anon_sym_TILDE = 149,
  aux_sym_prefix_op_token1 = 150,
  aux_sym_infix_op_token1 = 151,
  aux_sym_infix_op_token2 = 152,
  aux_sym_infix_op_token3 = 153,
  anon_sym_BANG_EQ = 154,
  anon_sym_COLON_EQ = 155,
  anon_sym_DOLLAR = 156,
  anon_sym_QMARK_LT_DASH = 157,
  anon_sym_QMARK_DASH_GT = 158,
  sym_int = 159,
  sym_xint = 160,
  anon_sym_y = 161,
  anon_sym_uy = 162,
  anon_sym_s = 163,
  anon_sym_us = 164,
  anon_sym_l = 165,
  aux_sym_uint32_token1 = 166,
  anon_sym_n = 167,
  anon_sym_un = 168,
  anon_sym_L = 169,
  aux_sym_uint64_token1 = 170,
  anon_sym_f = 171,
  anon_sym_lf = 172,
  anon_sym_LF = 173,
  aux_sym_bignum_token1 = 174,
  aux_sym_decimal_token1 = 175,
  anon_sym_DOT2 = 176,
  aux_sym_float_token1 = 177,
  anon_sym_LPAREN_STAR = 178,
  anon_sym_STAR_RPAREN = 179,
  sym_inline_line_comment = 180,
  sym_line_comment = 181,
  sym_xml_doc = 182,
  anon_sym_POUNDnowarn = 183,
  anon_sym_POUNDwarnon = 184,
  anon_sym_POUNDlight = 185,
  anon_sym_POUNDr = 186,
  anon_sym_POUNDload = 187,
  anon_sym_POUNDtime = 188,
  anon_sym_POUNDI = 189,
  anon_sym_POUNDhelp = 190,
  anon_sym_POUNDquit = 191,
  aux_sym_fsi_directive_decl_token1 = 192,
  aux_sym_preproc_line_token1 = 193,
  anon_sym_BANG = 194,
  anon_sym_POUNDif = 195,
  anon_sym_POUNDendif = 196,
  anon_sym_POUNDelse = 197,
  sym__newline = 198,
  sym__indent = 199,
  sym__dedent = 200,
  sym__struct_begin = 201,
  sym__interface_begin = 202,
  sym__triple_quoted_content = 203,
  sym__format_triple_quoted_content = 204,
  sym_block_comment_content = 205,
  sym__inside_string_marker = 206,
  sym__newline_not_aligned = 207,
  sym__tuple_marker = 208,
  sym__quoted_close = 209,
  sym__untyped_quoted_close = 210,
  sym__multi_dollar_triple_quote_start = 211,
  sym__multi_dollar_triple_quoted_content = 212,
  sym__multi_dollar_interp_start = 213,
  sym__multi_dollar_interp_end = 214,
  sym__multi_dollar_triple_quote_end = 215,
  sym__tyapp_open = 216,
  sym__paren_indent = 217,
  sym__type_decl_newline = 218,
  sym__in = 219,
  sym__error_sentinel = 220,
  sym_file = 221,
  sym_namespace = 222,
  sym_named_module = 223,
  sym__preproc_toplevel_module = 224,
  sym__module_body_elem = 225,
  sym__module_expression = 226,
  sym__module_elem = 227,
  sym__attribute_expression = 228,
  sym_module_abbrev = 229,
  sym_module_defn = 230,
  sym__module_body = 231,
  sym_import_decl = 232,
  sym_attributes = 233,
  sym__attribute_set = 234,
  sym_attribute = 235,
  sym__object_construction = 236,
  sym_value_declaration = 237,
  sym_do = 238,
  sym__function_or_value_defns = 239,
  sym_function_or_value_defn = 240,
  sym__function_or_value_defn_body = 241,
  sym_function_declaration_left = 242,
  sym_value_declaration_left = 243,
  sym_access_modifier = 244,
  sym_class_as_reference = 245,
  sym_primary_constr_args = 246,
  sym_repeat_pattern = 247,
  sym__pattern = 248,
  sym_optional_pattern = 249,
  sym_type_check_pattern = 250,
  sym_attribute_pattern = 251,
  sym_paren_pattern = 252,
  sym_as_pattern = 253,
  sym_cons_pattern = 254,
  sym_disjunct_pattern = 255,
  sym_conjunct_pattern = 256,
  sym_typed_pattern = 257,
  sym_typed_const_pattern = 258,
  sym_argument_patterns = 259,
  sym_field_pattern = 260,
  sym__atomic_pattern = 261,
  sym__list_pattern_content = 262,
  sym_list_pattern = 263,
  sym_array_pattern = 264,
  sym_record_pattern = 265,
  sym_named_field = 266,
  sym_named_field_pattern = 267,
  sym_identifier_pattern = 268,
  sym__expression_block = 269,
  sym__expression_block_for_let = 270,
  sym__paren_expression_block = 271,
  sym__mutable_expression = 272,
  sym__expression = 273,
  sym_literal_expression = 274,
  sym_srtp_call_expression = 275,
  sym__srtp_type_argument = 276,
  sym_long_identifier_or_op = 277,
  sym_tuple_expression = 278,
  sym_brace_expression = 279,
  sym_anon_record_expression = 280,
  sym_object_expression = 281,
  sym_with_field_expression = 282,
  sym_prefixed_expression = 283,
  sym_typecast_expression = 284,
  sym_for_expression = 285,
  sym_while_expression = 286,
  sym__else_expression = 287,
  sym_elif_expression = 288,
  sym__if_branch = 289,
  sym_if_expression = 290,
  sym_fun_expression = 291,
  sym_try_expression = 292,
  sym_match_expression = 293,
  sym_function_expression = 294,
  sym_mutate_expression = 295,
  sym_index_expression = 296,
  sym_typed_expression = 297,
  sym_declaration_expression = 298,
  sym_and_bang = 299,
  sym_do_expression = 300,
  sym__list_elements = 301,
  sym__list_element = 302,
  sym_list_expression = 303,
  sym_array_expression = 304,
  sym_range_expression = 305,
  sym_rule = 306,
  sym_rules = 307,
  sym_begin_end_expression = 308,
  sym_paren_expression = 309,
  sym__high_prec_app = 310,
  sym__low_prec_app = 311,
  sym_application_expression = 312,
  sym_dot_expression = 313,
  sym_infix_expression = 314,
  sym_ce_expression = 315,
  sym__comp_expression_block = 316,
  sym_sequential_expression = 317,
  sym__comp_or_range_expression = 318,
  sym_comp_declaration_expression = 319,
  sym_short_comp_expression = 320,
  sym_slice_ranges = 321,
  sym__slice_range_special = 322,
  sym_slice_range = 323,
  sym__type = 324,
  sym__argument_type = 325,
  sym__curried_return_type = 326,
  sym_measure_atom = 327,
  sym_measure_power = 328,
  sym__measure_operand = 329,
  sym_measure_quotient = 330,
  sym_measure = 331,
  sym_simple_type = 332,
  sym_generic_type = 333,
  sym_paren_type = 334,
  sym_function_type = 335,
  sym_compound_type = 336,
  sym_struct_type = 337,
  sym_postfix_type = 338,
  sym_list_type = 339,
  sym_static_type = 340,
  sym_constrained_type = 341,
  sym_flexible_type = 342,
  sym_anon_record_type = 343,
  sym_types = 344,
  sym__static_type_identifier = 345,
  sym__static_parameter = 346,
  sym_named_static_parameter = 347,
  sym_type_attribute = 348,
  sym_type_attributes = 349,
  sym__multiline_generic_type = 350,
  sym__multiline_generic_type_head = 351,
  sym__multiline_generic_function_type = 352,
  sym_atomic_type = 353,
  sym_constraint = 354,
  sym_type_argument_constraints = 355,
  sym_type_argument = 356,
  sym_type_argument_defn = 357,
  sym_type_arguments = 358,
  sym_trait_member_constraint = 359,
  sym_member_signature = 360,
  sym_curried_spec = 361,
  sym_argument_spec = 362,
  sym_arguments_spec = 363,
  sym_argument_name_spec = 364,
  sym_static_parameter_value = 365,
  sym_exception_definition = 366,
  sym_type_definition = 367,
  sym__type_defn_body = 368,
  sym_type_declaration = 369,
  sym_type_name = 370,
  sym_type_extension = 371,
  sym__type_extension_with = 372,
  sym_delegate_type_defn = 373,
  sym_delegate_signature = 374,
  sym_type_abbrev_defn = 375,
  sym__class_type_body_inner = 376,
  sym__class_type_body = 377,
  sym_record_type_defn = 378,
  sym_record_fields = 379,
  sym_record_field = 380,
  sym_enum_type_defn = 381,
  sym_enum_type_cases = 382,
  sym_enum_type_case = 383,
  sym_union_type_defn = 384,
  sym_union_type_cases = 385,
  sym_union_type_case = 386,
  sym_union_type_fields = 387,
  sym_union_type_field = 388,
  sym_interface_type_defn = 389,
  sym_anon_type_defn = 390,
  sym__class_function_or_value_defn = 391,
  aux_sym__type_extension_inner = 392,
  sym_type_extension_elements = 393,
  sym__type_defn_elements = 394,
  sym_interface_implementation = 395,
  sym__member_defns = 396,
  sym__object_members = 397,
  sym_member_defn = 398,
  sym_property_or_ident = 399,
  sym__method_defn = 400,
  sym__property_accessor_body = 401,
  sym_property_accessor = 402,
  sym__property_defn = 403,
  sym__val_property_defn = 404,
  sym_method_or_prop_defn = 405,
  sym_additional_constr_defn = 406,
  sym_extern_binding = 407,
  sym_extern_param = 408,
  sym_class_inherits_decl = 409,
  sym_field_initializer = 410,
  sym_field_initializers = 411,
  sym__simple_string_char = 412,
  sym__string_char = 413,
  sym_char = 414,
  sym_format_string_eval = 415,
  sym_format_string = 416,
  sym__string_literal = 417,
  sym_string = 418,
  sym__verbatim_string_char = 419,
  sym_verbatim_string = 420,
  sym_bytearray = 421,
  sym_verbatim_bytearray = 422,
  sym_format_triple_quoted_string = 423,
  sym_triple_quoted_string = 424,
  sym_const = 425,
  sym_long_identifier = 426,
  sym_active_pattern = 427,
  sym__identifier_or_op = 428,
  sym__infix_or_prefix_op = 429,
  sym_prefix_op = 430,
  sym_infix_op = 431,
  sym_sbyte = 432,
  sym_byte = 433,
  sym_int16 = 434,
  sym_uint16 = 435,
  sym_int32 = 436,
  sym_uint32 = 437,
  sym_nativeint = 438,
  sym_unativeint = 439,
  sym_int64 = 440,
  sym_uint64 = 441,
  sym_ieee32 = 442,
  sym_ieee64 = 443,
  sym_bignum = 444,
  sym_decimal = 445,
  sym_float = 446,
  sym_block_comment = 447,
  sym_compiler_directive_decl = 448,
  sym_fsi_directive_decl = 449,
  sym_preproc_line = 450,
  sym__preproc_expression = 451,
  sym_preproc_if_not_expression = 452,
  sym_preproc_if_and_expression = 453,
  sym_preproc_if_or_expression = 454,
  sym_preproc_if = 455,
  sym_preproc_else = 456,
  sym_preproc_if_in_expression = 457,
  sym_preproc_else_in_expression = 458,
  sym_preproc_if_in_module_body = 459,
  sym_preproc_else_in_module_body = 460,
  sym_preproc_if_in_class_definition = 461,
  sym_preproc_else_in_class_definition = 462,
  aux_sym_file_repeat1 = 463,
  aux_sym_file_repeat2 = 464,
  aux_sym__module_body_repeat1 = 465,
  aux_sym_attributes_repeat1 = 466,
  aux_sym__attribute_set_repeat1 = 467,
  aux_sym__function_or_value_defns_repeat1 = 468,
  aux_sym_repeat_pattern_repeat1 = 469,
  aux_sym_argument_patterns_repeat1 = 470,
  aux_sym__list_pattern_content_repeat1 = 471,
  aux_sym_record_pattern_repeat1 = 472,
  aux_sym_named_field_pattern_repeat1 = 473,
  aux_sym__srtp_type_argument_repeat1 = 474,
  aux_sym__object_expression_inner_repeat1 = 475,
  aux_sym_if_expression_repeat1 = 476,
  aux_sym_declaration_expression_repeat1 = 477,
  aux_sym__list_elements_repeat1 = 478,
  aux_sym_rules_repeat1 = 479,
  aux_sym__comp_expression_block_repeat1 = 480,
  aux_sym_sequential_expression_repeat1 = 481,
  aux_sym_slice_ranges_repeat1 = 482,
  aux_sym_compound_type_repeat1 = 483,
  aux_sym_types_repeat1 = 484,
  aux_sym_type_attributes_repeat1 = 485,
  aux_sym_type_argument_constraints_repeat1 = 486,
  aux_sym_type_argument_repeat1 = 487,
  aux_sym_type_arguments_repeat1 = 488,
  aux_sym_curried_spec_repeat1 = 489,
  aux_sym_arguments_spec_repeat1 = 490,
  aux_sym_type_definition_repeat1 = 491,
  aux_sym__class_type_body_repeat1 = 492,
  aux_sym_record_fields_repeat1 = 493,
  aux_sym_enum_type_cases_repeat1 = 494,
  aux_sym_union_type_cases_repeat1 = 495,
  aux_sym_union_type_fields_repeat1 = 496,
  aux_sym_interface_type_defn_repeat1 = 497,
  aux_sym__member_defns_repeat1 = 498,
  aux_sym__method_defn_repeat1 = 499,
  aux_sym__property_defn_repeat1 = 500,
  aux_sym_extern_binding_repeat1 = 501,
  aux_sym_field_initializers_repeat1 = 502,
  aux_sym_format_string_repeat1 = 503,
  aux_sym__string_literal_repeat1 = 504,
  aux_sym_verbatim_string_repeat1 = 505,
  aux_sym_format_triple_quoted_string_repeat1 = 506,
  aux_sym_format_triple_quoted_string_repeat2 = 507,
  aux_sym_long_identifier_repeat1 = 508,
  aux_sym_active_pattern_repeat1 = 509,
  aux_sym_prefix_op_repeat1 = 510,
  aux_sym_preproc_if_in_expression_repeat1 = 511,
  aux_sym_preproc_if_in_module_body_repeat1 = 512,
  aux_sym_preproc_if_in_class_definition_repeat1 = 513,
  anon_alias_sym_SEMI = 514,
  alias_sym_active_pattern_op_name = 515,
  alias_sym_wildcard_active_pattern_op = 516,
  alias_sym_wildcard_pattern = 517,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [anon_sym_namespace] = "namespace",
  [anon_sym_global] = "global",
  [anon_sym_rec] = "rec",
  [anon_sym_module] = "module",
  [anon_sym_null] = "null",
  [anon_sym_EQ] = "=",
  [anon_sym_begin] = "begin",
  [anon_sym_end] = "end",
  [anon_sym_open] = "open",
  [anon_sym_type] = "type",
  [anon_sym_LBRACK_LT] = "[<",
  [anon_sym_GT_RBRACK] = ">]",
  [anon_sym_COLON] = ":",
  [anon_sym_do] = "do",
  [anon_sym_and] = "and",
  [anon_sym_let] = "let",
  [anon_sym_let_BANG] = "let!",
  [anon_sym_inline] = "inline",
  [anon_sym_mutable] = "mutable",
  [aux_sym_access_modifier_token1] = "access_modifier_token1",
  [anon_sym_as] = "as",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_COMMA] = ",",
  [anon_sym__] = "_",
  [anon_sym_QMARK] = "\?",
  [anon_sym_COLON_QMARK] = ":\?",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_PIPE] = "|",
  [anon_sym_AMP] = "&",
  [anon_sym_GT] = ">",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_LBRACK_PIPE] = "[|",
  [anon_sym_PIPE_RBRACK] = "|]",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_LT_AT] = "<@",
  [anon_sym_LT_AT_AT] = "<@@",
  [anon_sym_CARET] = "^",
  [anon_sym_or] = "or",
  [anon_sym_DOT] = ".",
  [anon_sym_LBRACE_PIPE] = "{|",
  [anon_sym_PIPE_RBRACE] = "|}",
  [anon_sym_new] = "new",
  [anon_sym_with] = "with",
  [anon_sym_return] = "return",
  [anon_sym_return_BANG] = "return!",
  [anon_sym_yield] = "yield",
  [anon_sym_yield_BANG] = "yield!",
  [anon_sym_lazy] = "lazy",
  [anon_sym_assert] = "assert",
  [anon_sym_upcast] = "upcast",
  [anon_sym_downcast] = "downcast",
  [anon_sym_fixed] = "fixed",
  [anon_sym_COLON_GT] = ":>",
  [anon_sym_COLON_QMARK_GT] = ":\?>",
  [anon_sym_for] = "for",
  [anon_sym_in] = "in",
  [anon_sym_to] = "to",
  [anon_sym_downto] = "downto",
  [anon_sym_done] = "done",
  [anon_sym_while] = "while",
  [anon_sym_while_BANG] = "while!",
  [anon_sym_else] = "else",
  [anon_sym_then] = "then",
  [anon_sym_elif] = "elif",
  [anon_sym_if] = "if",
  [anon_sym_fun] = "fun",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_try] = "try",
  [anon_sym_finally] = "finally",
  [anon_sym_match] = "match",
  [anon_sym_match_BANG] = "match!",
  [anon_sym_function] = "function",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_DOT_LBRACK] = ".[",
  [anon_sym_use] = "use",
  [anon_sym_use_BANG] = "use!",
  [anon_sym_and_BANG] = "and!",
  [anon_sym_do_BANG] = "do!",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_when] = "when",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_DOT_DOT2] = "..",
  [anon_sym_DOT_DOT3] = "..",
  [anon_sym_STAR] = "*",
  [anon_sym_1] = "1",
  [anon_sym_SLASH] = "/",
  [anon_sym_LT] = "<",
  [anon_sym_struct] = "struct",
  [anon_sym_LBRACK_RBRACK] = "[]",
  [anon_sym_POUND] = "#",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_unit] = "unit",
  [anon_sym_not] = "not",
  [anon_sym_enum] = "enum",
  [anon_sym_unmanaged] = "unmanaged",
  [anon_sym_equality] = "equality",
  [anon_sym_comparison] = "comparison",
  [anon_sym_delegate] = "delegate",
  [anon_sym_default] = "default",
  [anon_sym_static] = "static",
  [anon_sym_member] = "member",
  [anon_sym_get] = "get",
  [anon_sym_set] = "set",
  [anon_sym_exception] = "exception",
  [anon_sym_of] = "of",
  [anon_sym_class] = "class",
  [anon_sym_interface] = "interface",
  [anon_sym_abstract] = "abstract",
  [anon_sym_val] = "val",
  [anon_sym_override] = "override",
  [anon_sym_extern] = "extern",
  [anon_sym_inherit] = "inherit",
  [anon_sym_EQ2] = "=",
  [sym__escape_char] = "_escape_char",
  [sym__non_escape_char] = "_non_escape_char",
  [sym__unicodegraph_short] = "_unicodegraph_short",
  [sym__unicodegraph_long] = "_unicodegraph_long",
  [sym__trigraph] = "_trigraph",
  [sym__hexgraph_short] = "_hexgraph_short",
  [aux_sym__simple_string_char_token1] = "_simple_string_char_token1",
  [aux_sym_char_token1] = "char_token1",
  [anon_sym_LBRACE2] = "{",
  [anon_sym_DOLLAR_DQUOTE] = "$\"",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_BSLASH] = "\\",
  [aux_sym__verbatim_string_char_token1] = "_verbatim_string_char_token1",
  [anon_sym_AT_DQUOTE] = "@\"",
  [anon_sym_DQUOTE2] = "\"",
  [anon_sym_DQUOTEB] = "\"B",
  [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = "$\"\"\"",
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = "\"\"\"",
  [sym_bool] = "bool",
  [sym_unit] = "unit",
  [anon_sym_LPAREN_PIPE] = "(|",
  [anon_sym_PIPE_RPAREN] = "|)",
  [sym_op_identifier] = "op_identifier",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_PLUS_DOT] = "+.",
  [anon_sym_DASH_DOT] = "-.",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PERCENT_PERCENT] = "%%",
  [anon_sym_TILDE] = "~",
  [aux_sym_prefix_op_token1] = "prefix_op_token1",
  [aux_sym_infix_op_token1] = "infix_op_token1",
  [aux_sym_infix_op_token2] = "infix_op_token2",
  [aux_sym_infix_op_token3] = "infix_op_token3",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_DOLLAR] = "$",
  [anon_sym_QMARK_LT_DASH] = "\?<-",
  [anon_sym_QMARK_DASH_GT] = "\?->",
  [sym_int] = "int",
  [sym_xint] = "xint",
  [anon_sym_y] = "y",
  [anon_sym_uy] = "uy",
  [anon_sym_s] = "s",
  [anon_sym_us] = "us",
  [anon_sym_l] = "l",
  [aux_sym_uint32_token1] = "uint32_token1",
  [anon_sym_n] = "n",
  [anon_sym_un] = "un",
  [anon_sym_L] = "L",
  [aux_sym_uint64_token1] = "uint64_token1",
  [anon_sym_f] = "f",
  [anon_sym_lf] = "lf",
  [anon_sym_LF] = "LF",
  [aux_sym_bignum_token1] = "bignum_token1",
  [aux_sym_decimal_token1] = "decimal_token1",
  [anon_sym_DOT2] = "float",
  [aux_sym_float_token1] = "float",
  [anon_sym_LPAREN_STAR] = "(*",
  [anon_sym_STAR_RPAREN] = "*)",
  [sym_inline_line_comment] = "line_comment",
  [sym_line_comment] = "line_comment",
  [sym_xml_doc] = "xml_doc",
  [anon_sym_POUNDnowarn] = "#nowarn",
  [anon_sym_POUNDwarnon] = "#warnon",
  [anon_sym_POUNDlight] = "#light",
  [anon_sym_POUNDr] = "#r",
  [anon_sym_POUNDload] = "#load",
  [anon_sym_POUNDtime] = "#time",
  [anon_sym_POUNDI] = "#I",
  [anon_sym_POUNDhelp] = "#help",
  [anon_sym_POUNDquit] = "#quit",
  [aux_sym_fsi_directive_decl_token1] = "fsi_directive_decl_token1",
  [aux_sym_preproc_line_token1] = "#line",
  [anon_sym_BANG] = "!",
  [anon_sym_POUNDif] = "#if",
  [anon_sym_POUNDendif] = "#endif",
  [anon_sym_POUNDelse] = "#else",
  [sym__newline] = "_newline",
  [sym__indent] = "_indent",
  [sym__dedent] = "_dedent",
  [sym__struct_begin] = "struct",
  [sym__interface_begin] = "interface",
  [sym__triple_quoted_content] = "_triple_quoted_content",
  [sym__format_triple_quoted_content] = "_format_triple_quoted_content",
  [sym_block_comment_content] = "block_comment_content",
  [sym__inside_string_marker] = "_inside_string_marker",
  [sym__newline_not_aligned] = "_newline_not_aligned",
  [sym__tuple_marker] = "_tuple_marker",
  [sym__quoted_close] = "_quoted_close",
  [sym__untyped_quoted_close] = "_untyped_quoted_close",
  [sym__multi_dollar_triple_quote_start] = "_multi_dollar_triple_quote_start",
  [sym__multi_dollar_triple_quoted_content] = "_multi_dollar_triple_quoted_content",
  [sym__multi_dollar_interp_start] = "_multi_dollar_interp_start",
  [sym__multi_dollar_interp_end] = "_multi_dollar_interp_end",
  [sym__multi_dollar_triple_quote_end] = "_multi_dollar_triple_quote_end",
  [sym__tyapp_open] = "_tyapp_open",
  [sym__paren_indent] = "_paren_indent",
  [sym__type_decl_newline] = "_type_decl_newline",
  [sym__in] = "_in",
  [sym__error_sentinel] = "_error_sentinel",
  [sym_file] = "file",
  [sym_namespace] = "namespace",
  [sym_named_module] = "named_module",
  [sym__preproc_toplevel_module] = "named_module",
  [sym__module_body_elem] = "_module_body_elem",
  [sym__module_expression] = "_module_expression",
  [sym__module_elem] = "_module_elem",
  [sym__attribute_expression] = "declaration_expression",
  [sym_module_abbrev] = "module_abbrev",
  [sym_module_defn] = "module_defn",
  [sym__module_body] = "_module_body",
  [sym_import_decl] = "import_decl",
  [sym_attributes] = "attributes",
  [sym__attribute_set] = "_attribute_set",
  [sym_attribute] = "attribute",
  [sym__object_construction] = "_object_construction",
  [sym_value_declaration] = "value_declaration",
  [sym_do] = "do",
  [sym__function_or_value_defns] = "_function_or_value_defns",
  [sym_function_or_value_defn] = "function_or_value_defn",
  [sym__function_or_value_defn_body] = "_function_or_value_defn_body",
  [sym_function_declaration_left] = "function_declaration_left",
  [sym_value_declaration_left] = "value_declaration_left",
  [sym_access_modifier] = "access_modifier",
  [sym_class_as_reference] = "class_as_reference",
  [sym_primary_constr_args] = "primary_constr_args",
  [sym_repeat_pattern] = "repeat_pattern",
  [sym__pattern] = "_pattern",
  [sym_optional_pattern] = "optional_pattern",
  [sym_type_check_pattern] = "type_check_pattern",
  [sym_attribute_pattern] = "attribute_pattern",
  [sym_paren_pattern] = "paren_pattern",
  [sym_as_pattern] = "as_pattern",
  [sym_cons_pattern] = "cons_pattern",
  [sym_disjunct_pattern] = "disjunct_pattern",
  [sym_conjunct_pattern] = "conjunct_pattern",
  [sym_typed_pattern] = "typed_pattern",
  [sym_typed_const_pattern] = "typed_const_pattern",
  [sym_argument_patterns] = "argument_patterns",
  [sym_field_pattern] = "field_pattern",
  [sym__atomic_pattern] = "_atomic_pattern",
  [sym__list_pattern_content] = "_list_pattern_content",
  [sym_list_pattern] = "list_pattern",
  [sym_array_pattern] = "array_pattern",
  [sym_record_pattern] = "record_pattern",
  [sym_named_field] = "named_field",
  [sym_named_field_pattern] = "named_field_pattern",
  [sym_identifier_pattern] = "identifier_pattern",
  [sym__expression_block] = "_expression_block",
  [sym__expression_block_for_let] = "_expression_block_for_let",
  [sym__paren_expression_block] = "_paren_expression_block",
  [sym__mutable_expression] = "_mutable_expression",
  [sym__expression] = "_expression",
  [sym_literal_expression] = "literal_expression",
  [sym_srtp_call_expression] = "srtp_call_expression",
  [sym__srtp_type_argument] = "type_argument",
  [sym_long_identifier_or_op] = "long_identifier_or_op",
  [sym_tuple_expression] = "tuple_expression",
  [sym_brace_expression] = "brace_expression",
  [sym_anon_record_expression] = "anon_record_expression",
  [sym_object_expression] = "object_expression",
  [sym_with_field_expression] = "with_field_expression",
  [sym_prefixed_expression] = "prefixed_expression",
  [sym_typecast_expression] = "typecast_expression",
  [sym_for_expression] = "for_expression",
  [sym_while_expression] = "while_expression",
  [sym__else_expression] = "_else_expression",
  [sym_elif_expression] = "elif_expression",
  [sym__if_branch] = "_if_branch",
  [sym_if_expression] = "if_expression",
  [sym_fun_expression] = "fun_expression",
  [sym_try_expression] = "try_expression",
  [sym_match_expression] = "match_expression",
  [sym_function_expression] = "function_expression",
  [sym_mutate_expression] = "mutate_expression",
  [sym_index_expression] = "index_expression",
  [sym_typed_expression] = "typed_expression",
  [sym_declaration_expression] = "declaration_expression",
  [sym_and_bang] = "and_bang",
  [sym_do_expression] = "do_expression",
  [sym__list_elements] = "_list_elements",
  [sym__list_element] = "_list_element",
  [sym_list_expression] = "list_expression",
  [sym_array_expression] = "array_expression",
  [sym_range_expression] = "range_expression",
  [sym_rule] = "rule",
  [sym_rules] = "rules",
  [sym_begin_end_expression] = "begin_end_expression",
  [sym_paren_expression] = "paren_expression",
  [sym__high_prec_app] = "_high_prec_app",
  [sym__low_prec_app] = "_low_prec_app",
  [sym_application_expression] = "application_expression",
  [sym_dot_expression] = "dot_expression",
  [sym_infix_expression] = "infix_expression",
  [sym_ce_expression] = "ce_expression",
  [sym__comp_expression_block] = "_comp_expression_block",
  [sym_sequential_expression] = "sequential_expression",
  [sym__comp_or_range_expression] = "_comp_or_range_expression",
  [sym_comp_declaration_expression] = "declaration_expression",
  [sym_short_comp_expression] = "short_comp_expression",
  [sym_slice_ranges] = "slice_ranges",
  [sym__slice_range_special] = "_slice_range_special",
  [sym_slice_range] = "slice_range",
  [sym__type] = "_type",
  [sym__argument_type] = "_argument_type",
  [sym__curried_return_type] = "_curried_return_type",
  [sym_measure_atom] = "measure_atom",
  [sym_measure_power] = "measure_power",
  [sym__measure_operand] = "_measure_operand",
  [sym_measure_quotient] = "measure_quotient",
  [sym_measure] = "measure",
  [sym_simple_type] = "simple_type",
  [sym_generic_type] = "generic_type",
  [sym_paren_type] = "paren_type",
  [sym_function_type] = "function_type",
  [sym_compound_type] = "compound_type",
  [sym_struct_type] = "struct_type",
  [sym_postfix_type] = "postfix_type",
  [sym_list_type] = "list_type",
  [sym_static_type] = "static_type",
  [sym_constrained_type] = "constrained_type",
  [sym_flexible_type] = "flexible_type",
  [sym_anon_record_type] = "anon_record_type",
  [sym_types] = "types",
  [sym__static_type_identifier] = "_static_type_identifier",
  [sym__static_parameter] = "_static_parameter",
  [sym_named_static_parameter] = "named_static_parameter",
  [sym_type_attribute] = "type_attribute",
  [sym_type_attributes] = "type_attributes",
  [sym__multiline_generic_type] = "generic_type",
  [sym__multiline_generic_type_head] = "generic_type",
  [sym__multiline_generic_function_type] = "function_type",
  [sym_atomic_type] = "atomic_type",
  [sym_constraint] = "constraint",
  [sym_type_argument_constraints] = "type_argument_constraints",
  [sym_type_argument] = "type_argument",
  [sym_type_argument_defn] = "type_argument_defn",
  [sym_type_arguments] = "type_arguments",
  [sym_trait_member_constraint] = "trait_member_constraint",
  [sym_member_signature] = "member_signature",
  [sym_curried_spec] = "curried_spec",
  [sym_argument_spec] = "argument_spec",
  [sym_arguments_spec] = "arguments_spec",
  [sym_argument_name_spec] = "argument_name_spec",
  [sym_static_parameter_value] = "static_parameter_value",
  [sym_exception_definition] = "exception_definition",
  [sym_type_definition] = "type_definition",
  [sym__type_defn_body] = "_type_defn_body",
  [sym_type_declaration] = "type_declaration",
  [sym_type_name] = "type_name",
  [sym_type_extension] = "type_extension",
  [sym__type_extension_with] = "type_extension_elements",
  [sym_delegate_type_defn] = "delegate_type_defn",
  [sym_delegate_signature] = "delegate_signature",
  [sym_type_abbrev_defn] = "type_abbrev_defn",
  [sym__class_type_body_inner] = "_class_type_body_inner",
  [sym__class_type_body] = "_class_type_body",
  [sym_record_type_defn] = "record_type_defn",
  [sym_record_fields] = "record_fields",
  [sym_record_field] = "record_field",
  [sym_enum_type_defn] = "enum_type_defn",
  [sym_enum_type_cases] = "enum_type_cases",
  [sym_enum_type_case] = "enum_type_case",
  [sym_union_type_defn] = "union_type_defn",
  [sym_union_type_cases] = "union_type_cases",
  [sym_union_type_case] = "union_type_case",
  [sym_union_type_fields] = "union_type_fields",
  [sym_union_type_field] = "union_type_field",
  [sym_interface_type_defn] = "interface_type_defn",
  [sym_anon_type_defn] = "anon_type_defn",
  [sym__class_function_or_value_defn] = "_class_function_or_value_defn",
  [aux_sym__type_extension_inner] = "_type_extension_inner",
  [sym_type_extension_elements] = "type_extension_elements",
  [sym__type_defn_elements] = "_type_defn_elements",
  [sym_interface_implementation] = "interface_implementation",
  [sym__member_defns] = "_member_defns",
  [sym__object_members] = "_object_members",
  [sym_member_defn] = "member_defn",
  [sym_property_or_ident] = "property_or_ident",
  [sym__method_defn] = "_method_defn",
  [sym__property_accessor_body] = "_property_accessor_body",
  [sym_property_accessor] = "property_accessor",
  [sym__property_defn] = "_property_defn",
  [sym__val_property_defn] = "_val_property_defn",
  [sym_method_or_prop_defn] = "method_or_prop_defn",
  [sym_additional_constr_defn] = "additional_constr_defn",
  [sym_extern_binding] = "extern_binding",
  [sym_extern_param] = "extern_param",
  [sym_class_inherits_decl] = "class_inherits_decl",
  [sym_field_initializer] = "field_initializer",
  [sym_field_initializers] = "field_initializers",
  [sym__simple_string_char] = "_simple_string_char",
  [sym__string_char] = "_string_char",
  [sym_char] = "char",
  [sym_format_string_eval] = "format_string_eval",
  [sym_format_string] = "format_string",
  [sym__string_literal] = "_string_literal",
  [sym_string] = "string",
  [sym__verbatim_string_char] = "_verbatim_string_char",
  [sym_verbatim_string] = "verbatim_string",
  [sym_bytearray] = "bytearray",
  [sym_verbatim_bytearray] = "verbatim_bytearray",
  [sym_format_triple_quoted_string] = "format_triple_quoted_string",
  [sym_triple_quoted_string] = "triple_quoted_string",
  [sym_const] = "const",
  [sym_long_identifier] = "long_identifier",
  [sym_active_pattern] = "active_pattern",
  [sym__identifier_or_op] = "_identifier_or_op",
  [sym__infix_or_prefix_op] = "_infix_or_prefix_op",
  [sym_prefix_op] = "prefix_op",
  [sym_infix_op] = "infix_op",
  [sym_sbyte] = "sbyte",
  [sym_byte] = "byte",
  [sym_int16] = "int16",
  [sym_uint16] = "uint16",
  [sym_int32] = "int32",
  [sym_uint32] = "uint32",
  [sym_nativeint] = "nativeint",
  [sym_unativeint] = "unativeint",
  [sym_int64] = "int64",
  [sym_uint64] = "uint64",
  [sym_ieee32] = "ieee32",
  [sym_ieee64] = "ieee64",
  [sym_bignum] = "bignum",
  [sym_decimal] = "decimal",
  [sym_float] = "float",
  [sym_block_comment] = "block_comment",
  [sym_compiler_directive_decl] = "compiler_directive_decl",
  [sym_fsi_directive_decl] = "fsi_directive_decl",
  [sym_preproc_line] = "preproc_line",
  [sym__preproc_expression] = "_preproc_expression",
  [sym_preproc_if_not_expression] = "preproc_if_not_expression",
  [sym_preproc_if_and_expression] = "preproc_if_and_expression",
  [sym_preproc_if_or_expression] = "preproc_if_or_expression",
  [sym_preproc_if] = "preproc_if",
  [sym_preproc_else] = "preproc_else",
  [sym_preproc_if_in_expression] = "preproc_if",
  [sym_preproc_else_in_expression] = "preproc_else",
  [sym_preproc_if_in_module_body] = "preproc_if",
  [sym_preproc_else_in_module_body] = "preproc_else",
  [sym_preproc_if_in_class_definition] = "preproc_if",
  [sym_preproc_else_in_class_definition] = "preproc_else",
  [aux_sym_file_repeat1] = "file_repeat1",
  [aux_sym_file_repeat2] = "file_repeat2",
  [aux_sym__module_body_repeat1] = "_module_body_repeat1",
  [aux_sym_attributes_repeat1] = "attributes_repeat1",
  [aux_sym__attribute_set_repeat1] = "_attribute_set_repeat1",
  [aux_sym__function_or_value_defns_repeat1] = "_function_or_value_defns_repeat1",
  [aux_sym_repeat_pattern_repeat1] = "repeat_pattern_repeat1",
  [aux_sym_argument_patterns_repeat1] = "argument_patterns_repeat1",
  [aux_sym__list_pattern_content_repeat1] = "_list_pattern_content_repeat1",
  [aux_sym_record_pattern_repeat1] = "record_pattern_repeat1",
  [aux_sym_named_field_pattern_repeat1] = "named_field_pattern_repeat1",
  [aux_sym__srtp_type_argument_repeat1] = "_srtp_type_argument_repeat1",
  [aux_sym__object_expression_inner_repeat1] = "_object_expression_inner_repeat1",
  [aux_sym_if_expression_repeat1] = "if_expression_repeat1",
  [aux_sym_declaration_expression_repeat1] = "declaration_expression_repeat1",
  [aux_sym__list_elements_repeat1] = "_list_elements_repeat1",
  [aux_sym_rules_repeat1] = "rules_repeat1",
  [aux_sym__comp_expression_block_repeat1] = "_comp_expression_block_repeat1",
  [aux_sym_sequential_expression_repeat1] = "sequential_expression_repeat1",
  [aux_sym_slice_ranges_repeat1] = "slice_ranges_repeat1",
  [aux_sym_compound_type_repeat1] = "compound_type_repeat1",
  [aux_sym_types_repeat1] = "types_repeat1",
  [aux_sym_type_attributes_repeat1] = "type_attributes_repeat1",
  [aux_sym_type_argument_constraints_repeat1] = "type_argument_constraints_repeat1",
  [aux_sym_type_argument_repeat1] = "type_argument_repeat1",
  [aux_sym_type_arguments_repeat1] = "type_arguments_repeat1",
  [aux_sym_curried_spec_repeat1] = "curried_spec_repeat1",
  [aux_sym_arguments_spec_repeat1] = "arguments_spec_repeat1",
  [aux_sym_type_definition_repeat1] = "type_definition_repeat1",
  [aux_sym__class_type_body_repeat1] = "_class_type_body_repeat1",
  [aux_sym_record_fields_repeat1] = "record_fields_repeat1",
  [aux_sym_enum_type_cases_repeat1] = "enum_type_cases_repeat1",
  [aux_sym_union_type_cases_repeat1] = "union_type_cases_repeat1",
  [aux_sym_union_type_fields_repeat1] = "union_type_fields_repeat1",
  [aux_sym_interface_type_defn_repeat1] = "interface_type_defn_repeat1",
  [aux_sym__member_defns_repeat1] = "_member_defns_repeat1",
  [aux_sym__method_defn_repeat1] = "_method_defn_repeat1",
  [aux_sym__property_defn_repeat1] = "_property_defn_repeat1",
  [aux_sym_extern_binding_repeat1] = "extern_binding_repeat1",
  [aux_sym_field_initializers_repeat1] = "field_initializers_repeat1",
  [aux_sym_format_string_repeat1] = "format_string_repeat1",
  [aux_sym__string_literal_repeat1] = "_string_literal_repeat1",
  [aux_sym_verbatim_string_repeat1] = "verbatim_string_repeat1",
  [aux_sym_format_triple_quoted_string_repeat1] = "format_triple_quoted_string_repeat1",
  [aux_sym_format_triple_quoted_string_repeat2] = "format_triple_quoted_string_repeat2",
  [aux_sym_long_identifier_repeat1] = "long_identifier_repeat1",
  [aux_sym_active_pattern_repeat1] = "active_pattern_repeat1",
  [aux_sym_prefix_op_repeat1] = "prefix_op_repeat1",
  [aux_sym_preproc_if_in_expression_repeat1] = "preproc_if_in_expression_repeat1",
  [aux_sym_preproc_if_in_module_body_repeat1] = "preproc_if_in_module_body_repeat1",
  [aux_sym_preproc_if_in_class_definition_repeat1] = "preproc_if_in_class_definition_repeat1",
  [anon_alias_sym_SEMI] = ";",
  [alias_sym_active_pattern_op_name] = "active_pattern_op_name",
  [alias_sym_wildcard_active_pattern_op] = "wildcard_active_pattern_op",
  [alias_sym_wildcard_pattern] = "wildcard_pattern",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [anon_sym_namespace] = anon_sym_namespace,
  [anon_sym_global] = anon_sym_global,
  [anon_sym_rec] = anon_sym_rec,
  [anon_sym_module] = anon_sym_module,
  [anon_sym_null] = anon_sym_null,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_begin] = anon_sym_begin,
  [anon_sym_end] = anon_sym_end,
  [anon_sym_open] = anon_sym_open,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_LBRACK_LT] = anon_sym_LBRACK_LT,
  [anon_sym_GT_RBRACK] = anon_sym_GT_RBRACK,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_let_BANG] = anon_sym_let_BANG,
  [anon_sym_inline] = anon_sym_inline,
  [anon_sym_mutable] = anon_sym_mutable,
  [aux_sym_access_modifier_token1] = aux_sym_access_modifier_token1,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym__] = anon_sym__,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_COLON_QMARK] = anon_sym_COLON_QMARK,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_LBRACK_PIPE] = anon_sym_LBRACK_PIPE,
  [anon_sym_PIPE_RBRACK] = anon_sym_PIPE_RBRACK,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_LT_AT] = anon_sym_LT_AT,
  [anon_sym_LT_AT_AT] = anon_sym_LT_AT_AT,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_LBRACE_PIPE] = anon_sym_LBRACE_PIPE,
  [anon_sym_PIPE_RBRACE] = anon_sym_PIPE_RBRACE,
  [anon_sym_new] = anon_sym_new,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_return_BANG] = anon_sym_return_BANG,
  [anon_sym_yield] = anon_sym_yield,
  [anon_sym_yield_BANG] = anon_sym_yield_BANG,
  [anon_sym_lazy] = anon_sym_lazy,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_upcast] = anon_sym_upcast,
  [anon_sym_downcast] = anon_sym_downcast,
  [anon_sym_fixed] = anon_sym_fixed,
  [anon_sym_COLON_GT] = anon_sym_COLON_GT,
  [anon_sym_COLON_QMARK_GT] = anon_sym_COLON_QMARK_GT,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_to] = anon_sym_to,
  [anon_sym_downto] = anon_sym_downto,
  [anon_sym_done] = anon_sym_done,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_while_BANG] = anon_sym_while_BANG,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_then] = anon_sym_then,
  [anon_sym_elif] = anon_sym_elif,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_fun] = anon_sym_fun,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_finally] = anon_sym_finally,
  [anon_sym_match] = anon_sym_match,
  [anon_sym_match_BANG] = anon_sym_match_BANG,
  [anon_sym_function] = anon_sym_function,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_DOT_LBRACK] = anon_sym_DOT_LBRACK,
  [anon_sym_use] = anon_sym_use,
  [anon_sym_use_BANG] = anon_sym_use_BANG,
  [anon_sym_and_BANG] = anon_sym_and_BANG,
  [anon_sym_do_BANG] = anon_sym_do_BANG,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_when] = anon_sym_when,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_DOT_DOT2] = anon_sym_DOT_DOT,
  [anon_sym_DOT_DOT3] = anon_sym_DOT_DOT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_1] = anon_sym_1,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_LBRACK_RBRACK] = anon_sym_LBRACK_RBRACK,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_unit] = anon_sym_unit,
  [anon_sym_not] = anon_sym_not,
  [anon_sym_enum] = anon_sym_enum,
  [anon_sym_unmanaged] = anon_sym_unmanaged,
  [anon_sym_equality] = anon_sym_equality,
  [anon_sym_comparison] = anon_sym_comparison,
  [anon_sym_delegate] = anon_sym_delegate,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_static] = anon_sym_static,
  [anon_sym_member] = anon_sym_member,
  [anon_sym_get] = anon_sym_get,
  [anon_sym_set] = anon_sym_set,
  [anon_sym_exception] = anon_sym_exception,
  [anon_sym_of] = anon_sym_of,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_interface] = anon_sym_interface,
  [anon_sym_abstract] = anon_sym_abstract,
  [anon_sym_val] = anon_sym_val,
  [anon_sym_override] = anon_sym_override,
  [anon_sym_extern] = anon_sym_extern,
  [anon_sym_inherit] = anon_sym_inherit,
  [anon_sym_EQ2] = anon_sym_EQ,
  [sym__escape_char] = sym__escape_char,
  [sym__non_escape_char] = sym__non_escape_char,
  [sym__unicodegraph_short] = sym__unicodegraph_short,
  [sym__unicodegraph_long] = sym__unicodegraph_long,
  [sym__trigraph] = sym__trigraph,
  [sym__hexgraph_short] = sym__hexgraph_short,
  [aux_sym__simple_string_char_token1] = aux_sym__simple_string_char_token1,
  [aux_sym_char_token1] = aux_sym_char_token1,
  [anon_sym_LBRACE2] = anon_sym_LBRACE,
  [anon_sym_DOLLAR_DQUOTE] = anon_sym_DOLLAR_DQUOTE,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [aux_sym__verbatim_string_char_token1] = aux_sym__verbatim_string_char_token1,
  [anon_sym_AT_DQUOTE] = anon_sym_AT_DQUOTE,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [anon_sym_DQUOTEB] = anon_sym_DQUOTEB,
  [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE,
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = anon_sym_DQUOTE_DQUOTE_DQUOTE,
  [sym_bool] = sym_bool,
  [sym_unit] = sym_unit,
  [anon_sym_LPAREN_PIPE] = anon_sym_LPAREN_PIPE,
  [anon_sym_PIPE_RPAREN] = anon_sym_PIPE_RPAREN,
  [sym_op_identifier] = sym_op_identifier,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_PLUS_DOT] = anon_sym_PLUS_DOT,
  [anon_sym_DASH_DOT] = anon_sym_DASH_DOT,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PERCENT_PERCENT] = anon_sym_PERCENT_PERCENT,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [aux_sym_prefix_op_token1] = aux_sym_prefix_op_token1,
  [aux_sym_infix_op_token1] = aux_sym_infix_op_token1,
  [aux_sym_infix_op_token2] = aux_sym_infix_op_token2,
  [aux_sym_infix_op_token3] = aux_sym_infix_op_token3,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [anon_sym_QMARK_LT_DASH] = anon_sym_QMARK_LT_DASH,
  [anon_sym_QMARK_DASH_GT] = anon_sym_QMARK_DASH_GT,
  [sym_int] = sym_int,
  [sym_xint] = sym_xint,
  [anon_sym_y] = anon_sym_y,
  [anon_sym_uy] = anon_sym_uy,
  [anon_sym_s] = anon_sym_s,
  [anon_sym_us] = anon_sym_us,
  [anon_sym_l] = anon_sym_l,
  [aux_sym_uint32_token1] = aux_sym_uint32_token1,
  [anon_sym_n] = anon_sym_n,
  [anon_sym_un] = anon_sym_un,
  [anon_sym_L] = anon_sym_L,
  [aux_sym_uint64_token1] = aux_sym_uint64_token1,
  [anon_sym_f] = anon_sym_f,
  [anon_sym_lf] = anon_sym_lf,
  [anon_sym_LF] = anon_sym_LF,
  [aux_sym_bignum_token1] = aux_sym_bignum_token1,
  [aux_sym_decimal_token1] = aux_sym_decimal_token1,
  [anon_sym_DOT2] = anon_sym_DOT2,
  [aux_sym_float_token1] = anon_sym_DOT2,
  [anon_sym_LPAREN_STAR] = anon_sym_LPAREN_STAR,
  [anon_sym_STAR_RPAREN] = anon_sym_STAR_RPAREN,
  [sym_inline_line_comment] = sym_line_comment,
  [sym_line_comment] = sym_line_comment,
  [sym_xml_doc] = sym_xml_doc,
  [anon_sym_POUNDnowarn] = anon_sym_POUNDnowarn,
  [anon_sym_POUNDwarnon] = anon_sym_POUNDwarnon,
  [anon_sym_POUNDlight] = anon_sym_POUNDlight,
  [anon_sym_POUNDr] = anon_sym_POUNDr,
  [anon_sym_POUNDload] = anon_sym_POUNDload,
  [anon_sym_POUNDtime] = anon_sym_POUNDtime,
  [anon_sym_POUNDI] = anon_sym_POUNDI,
  [anon_sym_POUNDhelp] = anon_sym_POUNDhelp,
  [anon_sym_POUNDquit] = anon_sym_POUNDquit,
  [aux_sym_fsi_directive_decl_token1] = aux_sym_fsi_directive_decl_token1,
  [aux_sym_preproc_line_token1] = aux_sym_preproc_line_token1,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_POUNDif] = anon_sym_POUNDif,
  [anon_sym_POUNDendif] = anon_sym_POUNDendif,
  [anon_sym_POUNDelse] = anon_sym_POUNDelse,
  [sym__newline] = sym__newline,
  [sym__indent] = sym__indent,
  [sym__dedent] = sym__dedent,
  [sym__struct_begin] = anon_sym_struct,
  [sym__interface_begin] = anon_sym_interface,
  [sym__triple_quoted_content] = sym__triple_quoted_content,
  [sym__format_triple_quoted_content] = sym__format_triple_quoted_content,
  [sym_block_comment_content] = sym_block_comment_content,
  [sym__inside_string_marker] = sym__inside_string_marker,
  [sym__newline_not_aligned] = sym__newline_not_aligned,
  [sym__tuple_marker] = sym__tuple_marker,
  [sym__quoted_close] = sym__quoted_close,
  [sym__untyped_quoted_close] = sym__untyped_quoted_close,
  [sym__multi_dollar_triple_quote_start] = sym__multi_dollar_triple_quote_start,
  [sym__multi_dollar_triple_quoted_content] = sym__multi_dollar_triple_quoted_content,
  [sym__multi_dollar_interp_start] = sym__multi_dollar_interp_start,
  [sym__multi_dollar_interp_end] = sym__multi_dollar_interp_end,
  [sym__multi_dollar_triple_quote_end] = sym__multi_dollar_triple_quote_end,
  [sym__tyapp_open] = sym__tyapp_open,
  [sym__paren_indent] = sym__paren_indent,
  [sym__type_decl_newline] = sym__type_decl_newline,
  [sym__in] = sym__in,
  [sym__error_sentinel] = sym__error_sentinel,
  [sym_file] = sym_file,
  [sym_namespace] = sym_namespace,
  [sym_named_module] = sym_named_module,
  [sym__preproc_toplevel_module] = sym_named_module,
  [sym__module_body_elem] = sym__module_body_elem,
  [sym__module_expression] = sym__module_expression,
  [sym__module_elem] = sym__module_elem,
  [sym__attribute_expression] = sym_declaration_expression,
  [sym_module_abbrev] = sym_module_abbrev,
  [sym_module_defn] = sym_module_defn,
  [sym__module_body] = sym__module_body,
  [sym_import_decl] = sym_import_decl,
  [sym_attributes] = sym_attributes,
  [sym__attribute_set] = sym__attribute_set,
  [sym_attribute] = sym_attribute,
  [sym__object_construction] = sym__object_construction,
  [sym_value_declaration] = sym_value_declaration,
  [sym_do] = sym_do,
  [sym__function_or_value_defns] = sym__function_or_value_defns,
  [sym_function_or_value_defn] = sym_function_or_value_defn,
  [sym__function_or_value_defn_body] = sym__function_or_value_defn_body,
  [sym_function_declaration_left] = sym_function_declaration_left,
  [sym_value_declaration_left] = sym_value_declaration_left,
  [sym_access_modifier] = sym_access_modifier,
  [sym_class_as_reference] = sym_class_as_reference,
  [sym_primary_constr_args] = sym_primary_constr_args,
  [sym_repeat_pattern] = sym_repeat_pattern,
  [sym__pattern] = sym__pattern,
  [sym_optional_pattern] = sym_optional_pattern,
  [sym_type_check_pattern] = sym_type_check_pattern,
  [sym_attribute_pattern] = sym_attribute_pattern,
  [sym_paren_pattern] = sym_paren_pattern,
  [sym_as_pattern] = sym_as_pattern,
  [sym_cons_pattern] = sym_cons_pattern,
  [sym_disjunct_pattern] = sym_disjunct_pattern,
  [sym_conjunct_pattern] = sym_conjunct_pattern,
  [sym_typed_pattern] = sym_typed_pattern,
  [sym_typed_const_pattern] = sym_typed_const_pattern,
  [sym_argument_patterns] = sym_argument_patterns,
  [sym_field_pattern] = sym_field_pattern,
  [sym__atomic_pattern] = sym__atomic_pattern,
  [sym__list_pattern_content] = sym__list_pattern_content,
  [sym_list_pattern] = sym_list_pattern,
  [sym_array_pattern] = sym_array_pattern,
  [sym_record_pattern] = sym_record_pattern,
  [sym_named_field] = sym_named_field,
  [sym_named_field_pattern] = sym_named_field_pattern,
  [sym_identifier_pattern] = sym_identifier_pattern,
  [sym__expression_block] = sym__expression_block,
  [sym__expression_block_for_let] = sym__expression_block_for_let,
  [sym__paren_expression_block] = sym__paren_expression_block,
  [sym__mutable_expression] = sym__mutable_expression,
  [sym__expression] = sym__expression,
  [sym_literal_expression] = sym_literal_expression,
  [sym_srtp_call_expression] = sym_srtp_call_expression,
  [sym__srtp_type_argument] = sym_type_argument,
  [sym_long_identifier_or_op] = sym_long_identifier_or_op,
  [sym_tuple_expression] = sym_tuple_expression,
  [sym_brace_expression] = sym_brace_expression,
  [sym_anon_record_expression] = sym_anon_record_expression,
  [sym_object_expression] = sym_object_expression,
  [sym_with_field_expression] = sym_with_field_expression,
  [sym_prefixed_expression] = sym_prefixed_expression,
  [sym_typecast_expression] = sym_typecast_expression,
  [sym_for_expression] = sym_for_expression,
  [sym_while_expression] = sym_while_expression,
  [sym__else_expression] = sym__else_expression,
  [sym_elif_expression] = sym_elif_expression,
  [sym__if_branch] = sym__if_branch,
  [sym_if_expression] = sym_if_expression,
  [sym_fun_expression] = sym_fun_expression,
  [sym_try_expression] = sym_try_expression,
  [sym_match_expression] = sym_match_expression,
  [sym_function_expression] = sym_function_expression,
  [sym_mutate_expression] = sym_mutate_expression,
  [sym_index_expression] = sym_index_expression,
  [sym_typed_expression] = sym_typed_expression,
  [sym_declaration_expression] = sym_declaration_expression,
  [sym_and_bang] = sym_and_bang,
  [sym_do_expression] = sym_do_expression,
  [sym__list_elements] = sym__list_elements,
  [sym__list_element] = sym__list_element,
  [sym_list_expression] = sym_list_expression,
  [sym_array_expression] = sym_array_expression,
  [sym_range_expression] = sym_range_expression,
  [sym_rule] = sym_rule,
  [sym_rules] = sym_rules,
  [sym_begin_end_expression] = sym_begin_end_expression,
  [sym_paren_expression] = sym_paren_expression,
  [sym__high_prec_app] = sym__high_prec_app,
  [sym__low_prec_app] = sym__low_prec_app,
  [sym_application_expression] = sym_application_expression,
  [sym_dot_expression] = sym_dot_expression,
  [sym_infix_expression] = sym_infix_expression,
  [sym_ce_expression] = sym_ce_expression,
  [sym__comp_expression_block] = sym__comp_expression_block,
  [sym_sequential_expression] = sym_sequential_expression,
  [sym__comp_or_range_expression] = sym__comp_or_range_expression,
  [sym_comp_declaration_expression] = sym_declaration_expression,
  [sym_short_comp_expression] = sym_short_comp_expression,
  [sym_slice_ranges] = sym_slice_ranges,
  [sym__slice_range_special] = sym__slice_range_special,
  [sym_slice_range] = sym_slice_range,
  [sym__type] = sym__type,
  [sym__argument_type] = sym__argument_type,
  [sym__curried_return_type] = sym__curried_return_type,
  [sym_measure_atom] = sym_measure_atom,
  [sym_measure_power] = sym_measure_power,
  [sym__measure_operand] = sym__measure_operand,
  [sym_measure_quotient] = sym_measure_quotient,
  [sym_measure] = sym_measure,
  [sym_simple_type] = sym_simple_type,
  [sym_generic_type] = sym_generic_type,
  [sym_paren_type] = sym_paren_type,
  [sym_function_type] = sym_function_type,
  [sym_compound_type] = sym_compound_type,
  [sym_struct_type] = sym_struct_type,
  [sym_postfix_type] = sym_postfix_type,
  [sym_list_type] = sym_list_type,
  [sym_static_type] = sym_static_type,
  [sym_constrained_type] = sym_constrained_type,
  [sym_flexible_type] = sym_flexible_type,
  [sym_anon_record_type] = sym_anon_record_type,
  [sym_types] = sym_types,
  [sym__static_type_identifier] = sym__static_type_identifier,
  [sym__static_parameter] = sym__static_parameter,
  [sym_named_static_parameter] = sym_named_static_parameter,
  [sym_type_attribute] = sym_type_attribute,
  [sym_type_attributes] = sym_type_attributes,
  [sym__multiline_generic_type] = sym_generic_type,
  [sym__multiline_generic_type_head] = sym_generic_type,
  [sym__multiline_generic_function_type] = sym_function_type,
  [sym_atomic_type] = sym_atomic_type,
  [sym_constraint] = sym_constraint,
  [sym_type_argument_constraints] = sym_type_argument_constraints,
  [sym_type_argument] = sym_type_argument,
  [sym_type_argument_defn] = sym_type_argument_defn,
  [sym_type_arguments] = sym_type_arguments,
  [sym_trait_member_constraint] = sym_trait_member_constraint,
  [sym_member_signature] = sym_member_signature,
  [sym_curried_spec] = sym_curried_spec,
  [sym_argument_spec] = sym_argument_spec,
  [sym_arguments_spec] = sym_arguments_spec,
  [sym_argument_name_spec] = sym_argument_name_spec,
  [sym_static_parameter_value] = sym_static_parameter_value,
  [sym_exception_definition] = sym_exception_definition,
  [sym_type_definition] = sym_type_definition,
  [sym__type_defn_body] = sym__type_defn_body,
  [sym_type_declaration] = sym_type_declaration,
  [sym_type_name] = sym_type_name,
  [sym_type_extension] = sym_type_extension,
  [sym__type_extension_with] = sym_type_extension_elements,
  [sym_delegate_type_defn] = sym_delegate_type_defn,
  [sym_delegate_signature] = sym_delegate_signature,
  [sym_type_abbrev_defn] = sym_type_abbrev_defn,
  [sym__class_type_body_inner] = sym__class_type_body_inner,
  [sym__class_type_body] = sym__class_type_body,
  [sym_record_type_defn] = sym_record_type_defn,
  [sym_record_fields] = sym_record_fields,
  [sym_record_field] = sym_record_field,
  [sym_enum_type_defn] = sym_enum_type_defn,
  [sym_enum_type_cases] = sym_enum_type_cases,
  [sym_enum_type_case] = sym_enum_type_case,
  [sym_union_type_defn] = sym_union_type_defn,
  [sym_union_type_cases] = sym_union_type_cases,
  [sym_union_type_case] = sym_union_type_case,
  [sym_union_type_fields] = sym_union_type_fields,
  [sym_union_type_field] = sym_union_type_field,
  [sym_interface_type_defn] = sym_interface_type_defn,
  [sym_anon_type_defn] = sym_anon_type_defn,
  [sym__class_function_or_value_defn] = sym__class_function_or_value_defn,
  [aux_sym__type_extension_inner] = aux_sym__type_extension_inner,
  [sym_type_extension_elements] = sym_type_extension_elements,
  [sym__type_defn_elements] = sym__type_defn_elements,
  [sym_interface_implementation] = sym_interface_implementation,
  [sym__member_defns] = sym__member_defns,
  [sym__object_members] = sym__object_members,
  [sym_member_defn] = sym_member_defn,
  [sym_property_or_ident] = sym_property_or_ident,
  [sym__method_defn] = sym__method_defn,
  [sym__property_accessor_body] = sym__property_accessor_body,
  [sym_property_accessor] = sym_property_accessor,
  [sym__property_defn] = sym__property_defn,
  [sym__val_property_defn] = sym__val_property_defn,
  [sym_method_or_prop_defn] = sym_method_or_prop_defn,
  [sym_additional_constr_defn] = sym_additional_constr_defn,
  [sym_extern_binding] = sym_extern_binding,
  [sym_extern_param] = sym_extern_param,
  [sym_class_inherits_decl] = sym_class_inherits_decl,
  [sym_field_initializer] = sym_field_initializer,
  [sym_field_initializers] = sym_field_initializers,
  [sym__simple_string_char] = sym__simple_string_char,
  [sym__string_char] = sym__string_char,
  [sym_char] = sym_char,
  [sym_format_string_eval] = sym_format_string_eval,
  [sym_format_string] = sym_format_string,
  [sym__string_literal] = sym__string_literal,
  [sym_string] = sym_string,
  [sym__verbatim_string_char] = sym__verbatim_string_char,
  [sym_verbatim_string] = sym_verbatim_string,
  [sym_bytearray] = sym_bytearray,
  [sym_verbatim_bytearray] = sym_verbatim_bytearray,
  [sym_format_triple_quoted_string] = sym_format_triple_quoted_string,
  [sym_triple_quoted_string] = sym_triple_quoted_string,
  [sym_const] = sym_const,
  [sym_long_identifier] = sym_long_identifier,
  [sym_active_pattern] = sym_active_pattern,
  [sym__identifier_or_op] = sym__identifier_or_op,
  [sym__infix_or_prefix_op] = sym__infix_or_prefix_op,
  [sym_prefix_op] = sym_prefix_op,
  [sym_infix_op] = sym_infix_op,
  [sym_sbyte] = sym_sbyte,
  [sym_byte] = sym_byte,
  [sym_int16] = sym_int16,
  [sym_uint16] = sym_uint16,
  [sym_int32] = sym_int32,
  [sym_uint32] = sym_uint32,
  [sym_nativeint] = sym_nativeint,
  [sym_unativeint] = sym_unativeint,
  [sym_int64] = sym_int64,
  [sym_uint64] = sym_uint64,
  [sym_ieee32] = sym_ieee32,
  [sym_ieee64] = sym_ieee64,
  [sym_bignum] = sym_bignum,
  [sym_decimal] = sym_decimal,
  [sym_float] = sym_float,
  [sym_block_comment] = sym_block_comment,
  [sym_compiler_directive_decl] = sym_compiler_directive_decl,
  [sym_fsi_directive_decl] = sym_fsi_directive_decl,
  [sym_preproc_line] = sym_preproc_line,
  [sym__preproc_expression] = sym__preproc_expression,
  [sym_preproc_if_not_expression] = sym_preproc_if_not_expression,
  [sym_preproc_if_and_expression] = sym_preproc_if_and_expression,
  [sym_preproc_if_or_expression] = sym_preproc_if_or_expression,
  [sym_preproc_if] = sym_preproc_if,
  [sym_preproc_else] = sym_preproc_else,
  [sym_preproc_if_in_expression] = sym_preproc_if,
  [sym_preproc_else_in_expression] = sym_preproc_else,
  [sym_preproc_if_in_module_body] = sym_preproc_if,
  [sym_preproc_else_in_module_body] = sym_preproc_else,
  [sym_preproc_if_in_class_definition] = sym_preproc_if,
  [sym_preproc_else_in_class_definition] = sym_preproc_else,
  [aux_sym_file_repeat1] = aux_sym_file_repeat1,
  [aux_sym_file_repeat2] = aux_sym_file_repeat2,
  [aux_sym__module_body_repeat1] = aux_sym__module_body_repeat1,
  [aux_sym_attributes_repeat1] = aux_sym_attributes_repeat1,
  [aux_sym__attribute_set_repeat1] = aux_sym__attribute_set_repeat1,
  [aux_sym__function_or_value_defns_repeat1] = aux_sym__function_or_value_defns_repeat1,
  [aux_sym_repeat_pattern_repeat1] = aux_sym_repeat_pattern_repeat1,
  [aux_sym_argument_patterns_repeat1] = aux_sym_argument_patterns_repeat1,
  [aux_sym__list_pattern_content_repeat1] = aux_sym__list_pattern_content_repeat1,
  [aux_sym_record_pattern_repeat1] = aux_sym_record_pattern_repeat1,
  [aux_sym_named_field_pattern_repeat1] = aux_sym_named_field_pattern_repeat1,
  [aux_sym__srtp_type_argument_repeat1] = aux_sym__srtp_type_argument_repeat1,
  [aux_sym__object_expression_inner_repeat1] = aux_sym__object_expression_inner_repeat1,
  [aux_sym_if_expression_repeat1] = aux_sym_if_expression_repeat1,
  [aux_sym_declaration_expression_repeat1] = aux_sym_declaration_expression_repeat1,
  [aux_sym__list_elements_repeat1] = aux_sym__list_elements_repeat1,
  [aux_sym_rules_repeat1] = aux_sym_rules_repeat1,
  [aux_sym__comp_expression_block_repeat1] = aux_sym__comp_expression_block_repeat1,
  [aux_sym_sequential_expression_repeat1] = aux_sym_sequential_expression_repeat1,
  [aux_sym_slice_ranges_repeat1] = aux_sym_slice_ranges_repeat1,
  [aux_sym_compound_type_repeat1] = aux_sym_compound_type_repeat1,
  [aux_sym_types_repeat1] = aux_sym_types_repeat1,
  [aux_sym_type_attributes_repeat1] = aux_sym_type_attributes_repeat1,
  [aux_sym_type_argument_constraints_repeat1] = aux_sym_type_argument_constraints_repeat1,
  [aux_sym_type_argument_repeat1] = aux_sym_type_argument_repeat1,
  [aux_sym_type_arguments_repeat1] = aux_sym_type_arguments_repeat1,
  [aux_sym_curried_spec_repeat1] = aux_sym_curried_spec_repeat1,
  [aux_sym_arguments_spec_repeat1] = aux_sym_arguments_spec_repeat1,
  [aux_sym_type_definition_repeat1] = aux_sym_type_definition_repeat1,
  [aux_sym__class_type_body_repeat1] = aux_sym__class_type_body_repeat1,
  [aux_sym_record_fields_repeat1] = aux_sym_record_fields_repeat1,
  [aux_sym_enum_type_cases_repeat1] = aux_sym_enum_type_cases_repeat1,
  [aux_sym_union_type_cases_repeat1] = aux_sym_union_type_cases_repeat1,
  [aux_sym_union_type_fields_repeat1] = aux_sym_union_type_fields_repeat1,
  [aux_sym_interface_type_defn_repeat1] = aux_sym_interface_type_defn_repeat1,
  [aux_sym__member_defns_repeat1] = aux_sym__member_defns_repeat1,
  [aux_sym__method_defn_repeat1] = aux_sym__method_defn_repeat1,
  [aux_sym__property_defn_repeat1] = aux_sym__property_defn_repeat1,
  [aux_sym_extern_binding_repeat1] = aux_sym_extern_binding_repeat1,
  [aux_sym_field_initializers_repeat1] = aux_sym_field_initializers_repeat1,
  [aux_sym_format_string_repeat1] = aux_sym_format_string_repeat1,
  [aux_sym__string_literal_repeat1] = aux_sym__string_literal_repeat1,
  [aux_sym_verbatim_string_repeat1] = aux_sym_verbatim_string_repeat1,
  [aux_sym_format_triple_quoted_string_repeat1] = aux_sym_format_triple_quoted_string_repeat1,
  [aux_sym_format_triple_quoted_string_repeat2] = aux_sym_format_triple_quoted_string_repeat2,
  [aux_sym_long_identifier_repeat1] = aux_sym_long_identifier_repeat1,
  [aux_sym_active_pattern_repeat1] = aux_sym_active_pattern_repeat1,
  [aux_sym_prefix_op_repeat1] = aux_sym_prefix_op_repeat1,
  [aux_sym_preproc_if_in_expression_repeat1] = aux_sym_preproc_if_in_expression_repeat1,
  [aux_sym_preproc_if_in_module_body_repeat1] = aux_sym_preproc_if_in_module_body_repeat1,
  [aux_sym_preproc_if_in_class_definition_repeat1] = aux_sym_preproc_if_in_class_definition_repeat1,
  [anon_alias_sym_SEMI] = anon_alias_sym_SEMI,
  [alias_sym_active_pattern_op_name] = alias_sym_active_pattern_op_name,
  [alias_sym_wildcard_active_pattern_op] = alias_sym_wildcard_active_pattern_op,
  [alias_sym_wildcard_pattern] = alias_sym_wildcard_pattern,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_namespace] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_global] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_module] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_null] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_begin] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_end] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_open] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mutable] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_access_modifier_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_yield] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_yield_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lazy] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_upcast] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_downcast] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fixed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_QMARK_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_to] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_downto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_done] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_then] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_elif] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fun] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_match] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_match_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_function] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_use] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_use_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_when] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unit] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_not] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unmanaged] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_equality] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_comparison] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_delegate] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_static] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_member] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_get] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_set] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exception] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_interface] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_abstract] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_val] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_override] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extern] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inherit] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ2] = {
    .visible = true,
    .named = false,
  },
  [sym__escape_char] = {
    .visible = false,
    .named = true,
  },
  [sym__non_escape_char] = {
    .visible = false,
    .named = true,
  },
  [sym__unicodegraph_short] = {
    .visible = false,
    .named = true,
  },
  [sym__unicodegraph_long] = {
    .visible = false,
    .named = true,
  },
  [sym__trigraph] = {
    .visible = false,
    .named = true,
  },
  [sym__hexgraph_short] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__simple_string_char_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_char_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LBRACE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__verbatim_string_char_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_AT_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTEB] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_bool] = {
    .visible = true,
    .named = true,
  },
  [sym_unit] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LPAREN_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [sym_op_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_prefix_op_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_infix_op_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_infix_op_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_infix_op_token3] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_int] = {
    .visible = true,
    .named = true,
  },
  [sym_xint] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_y] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_uy] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_s] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_us] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_l] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_uint32_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_n] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_un] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_L] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_uint64_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_f] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lf] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LF] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_bignum_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_decimal_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DOT2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_float_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [sym_inline_line_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_line_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_xml_doc] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_POUNDnowarn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDwarnon] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDlight] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDload] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDtime] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDhelp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDquit] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_fsi_directive_decl_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_line_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDif] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDendif] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUNDelse] = {
    .visible = true,
    .named = false,
  },
  [sym__newline] = {
    .visible = false,
    .named = true,
  },
  [sym__indent] = {
    .visible = false,
    .named = true,
  },
  [sym__dedent] = {
    .visible = false,
    .named = true,
  },
  [sym__struct_begin] = {
    .visible = true,
    .named = false,
  },
  [sym__interface_begin] = {
    .visible = true,
    .named = false,
  },
  [sym__triple_quoted_content] = {
    .visible = false,
    .named = true,
  },
  [sym__format_triple_quoted_content] = {
    .visible = false,
    .named = true,
  },
  [sym_block_comment_content] = {
    .visible = true,
    .named = true,
  },
  [sym__inside_string_marker] = {
    .visible = false,
    .named = true,
  },
  [sym__newline_not_aligned] = {
    .visible = false,
    .named = true,
  },
  [sym__tuple_marker] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_close] = {
    .visible = false,
    .named = true,
  },
  [sym__untyped_quoted_close] = {
    .visible = false,
    .named = true,
  },
  [sym__multi_dollar_triple_quote_start] = {
    .visible = false,
    .named = true,
  },
  [sym__multi_dollar_triple_quoted_content] = {
    .visible = false,
    .named = true,
  },
  [sym__multi_dollar_interp_start] = {
    .visible = false,
    .named = true,
  },
  [sym__multi_dollar_interp_end] = {
    .visible = false,
    .named = true,
  },
  [sym__multi_dollar_triple_quote_end] = {
    .visible = false,
    .named = true,
  },
  [sym__tyapp_open] = {
    .visible = false,
    .named = true,
  },
  [sym__paren_indent] = {
    .visible = false,
    .named = true,
  },
  [sym__type_decl_newline] = {
    .visible = false,
    .named = true,
  },
  [sym__in] = {
    .visible = false,
    .named = true,
  },
  [sym__error_sentinel] = {
    .visible = false,
    .named = true,
  },
  [sym_file] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace] = {
    .visible = true,
    .named = true,
  },
  [sym_named_module] = {
    .visible = true,
    .named = true,
  },
  [sym__preproc_toplevel_module] = {
    .visible = true,
    .named = true,
  },
  [sym__module_body_elem] = {
    .visible = false,
    .named = true,
  },
  [sym__module_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__module_elem] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__attribute_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_module_abbrev] = {
    .visible = true,
    .named = true,
  },
  [sym_module_defn] = {
    .visible = true,
    .named = true,
  },
  [sym__module_body] = {
    .visible = false,
    .named = true,
  },
  [sym_import_decl] = {
    .visible = true,
    .named = true,
  },
  [sym_attributes] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute_set] = {
    .visible = false,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym__object_construction] = {
    .visible = false,
    .named = true,
  },
  [sym_value_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_do] = {
    .visible = true,
    .named = true,
  },
  [sym__function_or_value_defns] = {
    .visible = false,
    .named = true,
  },
  [sym_function_or_value_defn] = {
    .visible = true,
    .named = true,
  },
  [sym__function_or_value_defn_body] = {
    .visible = false,
    .named = true,
  },
  [sym_function_declaration_left] = {
    .visible = true,
    .named = true,
  },
  [sym_value_declaration_left] = {
    .visible = true,
    .named = true,
  },
  [sym_access_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_class_as_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_primary_constr_args] = {
    .visible = true,
    .named = true,
  },
  [sym_repeat_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_optional_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_type_check_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_paren_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_as_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_disjunct_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_conjunct_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_const_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_patterns] = {
    .visible = true,
    .named = true,
  },
  [sym_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__atomic_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym__list_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym_list_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_record_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_named_field] = {
    .visible = true,
    .named = true,
  },
  [sym_named_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_identifier_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_block] = {
    .visible = false,
    .named = true,
  },
  [sym__expression_block_for_let] = {
    .visible = false,
    .named = true,
  },
  [sym__paren_expression_block] = {
    .visible = false,
    .named = true,
  },
  [sym__mutable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_literal_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_srtp_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__srtp_type_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_long_identifier_or_op] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_brace_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_anon_record_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_object_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_with_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_prefixed_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_typecast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_for_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_while_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__else_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_elif_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__if_branch] = {
    .visible = false,
    .named = true,
  },
  [sym_if_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_fun_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_try_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_function_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_mutate_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_index_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_and_bang] = {
    .visible = true,
    .named = true,
  },
  [sym_do_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__list_elements] = {
    .visible = false,
    .named = true,
  },
  [sym__list_element] = {
    .visible = false,
    .named = true,
  },
  [sym_list_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_range_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_rule] = {
    .visible = true,
    .named = true,
  },
  [sym_rules] = {
    .visible = true,
    .named = true,
  },
  [sym_begin_end_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_paren_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__high_prec_app] = {
    .visible = false,
    .named = true,
  },
  [sym__low_prec_app] = {
    .visible = false,
    .named = true,
  },
  [sym_application_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_dot_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_infix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_ce_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__comp_expression_block] = {
    .visible = false,
    .named = true,
  },
  [sym_sequential_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__comp_or_range_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_comp_declaration_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_short_comp_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_slice_ranges] = {
    .visible = true,
    .named = true,
  },
  [sym__slice_range_special] = {
    .visible = false,
    .named = true,
  },
  [sym_slice_range] = {
    .visible = true,
    .named = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__argument_type] = {
    .visible = false,
    .named = true,
  },
  [sym__curried_return_type] = {
    .visible = false,
    .named = true,
  },
  [sym_measure_atom] = {
    .visible = true,
    .named = true,
  },
  [sym_measure_power] = {
    .visible = true,
    .named = true,
  },
  [sym__measure_operand] = {
    .visible = false,
    .named = true,
  },
  [sym_measure_quotient] = {
    .visible = true,
    .named = true,
  },
  [sym_measure] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_type] = {
    .visible = true,
    .named = true,
  },
  [sym_paren_type] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_type] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_type] = {
    .visible = true,
    .named = true,
  },
  [sym_postfix_type] = {
    .visible = true,
    .named = true,
  },
  [sym_list_type] = {
    .visible = true,
    .named = true,
  },
  [sym_static_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constrained_type] = {
    .visible = true,
    .named = true,
  },
  [sym_flexible_type] = {
    .visible = true,
    .named = true,
  },
  [sym_anon_record_type] = {
    .visible = true,
    .named = true,
  },
  [sym_types] = {
    .visible = true,
    .named = true,
  },
  [sym__static_type_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__static_parameter] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_named_static_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_type_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_type_attributes] = {
    .visible = true,
    .named = true,
  },
  [sym__multiline_generic_type] = {
    .visible = true,
    .named = true,
  },
  [sym__multiline_generic_type_head] = {
    .visible = true,
    .named = true,
  },
  [sym__multiline_generic_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_atomic_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_type_argument_constraints] = {
    .visible = true,
    .named = true,
  },
  [sym_type_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_type_argument_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_type_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_trait_member_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_member_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_curried_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_arguments_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_name_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_static_parameter_value] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__type_defn_body] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_type_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_name] = {
    .visible = true,
    .named = true,
  },
  [sym_type_extension] = {
    .visible = true,
    .named = true,
  },
  [sym__type_extension_with] = {
    .visible = true,
    .named = true,
  },
  [sym_delegate_type_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_delegate_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_type_abbrev_defn] = {
    .visible = true,
    .named = true,
  },
  [sym__class_type_body_inner] = {
    .visible = false,
    .named = true,
  },
  [sym__class_type_body] = {
    .visible = false,
    .named = true,
  },
  [sym_record_type_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_record_fields] = {
    .visible = true,
    .named = true,
  },
  [sym_record_field] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_type_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_type_cases] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_type_case] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type_cases] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type_case] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type_fields] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type_field] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_type_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_anon_type_defn] = {
    .visible = true,
    .named = true,
  },
  [sym__class_function_or_value_defn] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__type_extension_inner] = {
    .visible = false,
    .named = false,
  },
  [sym_type_extension_elements] = {
    .visible = true,
    .named = true,
  },
  [sym__type_defn_elements] = {
    .visible = false,
    .named = true,
  },
  [sym_interface_implementation] = {
    .visible = true,
    .named = true,
  },
  [sym__member_defns] = {
    .visible = false,
    .named = true,
  },
  [sym__object_members] = {
    .visible = false,
    .named = true,
  },
  [sym_member_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_property_or_ident] = {
    .visible = true,
    .named = true,
  },
  [sym__method_defn] = {
    .visible = false,
    .named = true,
  },
  [sym__property_accessor_body] = {
    .visible = false,
    .named = true,
  },
  [sym_property_accessor] = {
    .visible = true,
    .named = true,
  },
  [sym__property_defn] = {
    .visible = false,
    .named = true,
  },
  [sym__val_property_defn] = {
    .visible = false,
    .named = true,
  },
  [sym_method_or_prop_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_additional_constr_defn] = {
    .visible = true,
    .named = true,
  },
  [sym_extern_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_extern_param] = {
    .visible = true,
    .named = true,
  },
  [sym_class_inherits_decl] = {
    .visible = true,
    .named = true,
  },
  [sym_field_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_field_initializers] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_char] = {
    .visible = false,
    .named = true,
  },
  [sym__string_char] = {
    .visible = false,
    .named = true,
  },
  [sym_char] = {
    .visible = true,
    .named = true,
  },
  [sym_format_string_eval] = {
    .visible = true,
    .named = true,
  },
  [sym_format_string] = {
    .visible = true,
    .named = true,
  },
  [sym__string_literal] = {
    .visible = false,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym__verbatim_string_char] = {
    .visible = false,
    .named = true,
  },
  [sym_verbatim_string] = {
    .visible = true,
    .named = true,
  },
  [sym_bytearray] = {
    .visible = true,
    .named = true,
  },
  [sym_verbatim_bytearray] = {
    .visible = true,
    .named = true,
  },
  [sym_format_triple_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym_triple_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym_const] = {
    .visible = true,
    .named = true,
  },
  [sym_long_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_active_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__identifier_or_op] = {
    .visible = false,
    .named = true,
  },
  [sym__infix_or_prefix_op] = {
    .visible = false,
    .named = true,
  },
  [sym_prefix_op] = {
    .visible = true,
    .named = true,
  },
  [sym_infix_op] = {
    .visible = true,
    .named = true,
  },
  [sym_sbyte] = {
    .visible = true,
    .named = true,
  },
  [sym_byte] = {
    .visible = true,
    .named = true,
  },
  [sym_int16] = {
    .visible = true,
    .named = true,
  },
  [sym_uint16] = {
    .visible = true,
    .named = true,
  },
  [sym_int32] = {
    .visible = true,
    .named = true,
  },
  [sym_uint32] = {
    .visible = true,
    .named = true,
  },
  [sym_nativeint] = {
    .visible = true,
    .named = true,
  },
  [sym_unativeint] = {
    .visible = true,
    .named = true,
  },
  [sym_int64] = {
    .visible = true,
    .named = true,
  },
  [sym_uint64] = {
    .visible = true,
    .named = true,
  },
  [sym_ieee32] = {
    .visible = true,
    .named = true,
  },
  [sym_ieee64] = {
    .visible = true,
    .named = true,
  },
  [sym_bignum] = {
    .visible = true,
    .named = true,
  },
  [sym_decimal] = {
    .visible = true,
    .named = true,
  },
  [sym_float] = {
    .visible = true,
    .named = true,
  },
  [sym_block_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_compiler_directive_decl] = {
    .visible = true,
    .named = true,
  },
  [sym_fsi_directive_decl] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_line] = {
    .visible = true,
    .named = true,
  },
  [sym__preproc_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_preproc_if_not_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_and_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_or_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_module_body] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_module_body] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_class_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_class_definition] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_file_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__module_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attributes_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__attribute_set_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__function_or_value_defns_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_repeat_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_patterns_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_named_field_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__srtp_type_argument_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__object_expression_inner_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_elements_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_rules_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__comp_expression_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequential_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_slice_ranges_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_compound_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_types_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_attributes_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_argument_constraints_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_argument_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_curried_spec_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arguments_spec_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__class_type_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_fields_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_type_cases_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_union_type_cases_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_union_type_fields_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_interface_type_defn_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__member_defns_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__method_defn_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__property_defn_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_extern_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_field_initializers_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_format_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_verbatim_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_format_triple_quoted_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_format_triple_quoted_string_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_long_identifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_active_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_prefix_op_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_module_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_class_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [anon_alias_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [alias_sym_active_pattern_op_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_wildcard_active_pattern_op] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_wildcard_pattern] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alternative = 1,
  field_args = 2,
  field_assignee = 3,
  field_base = 4,
  field_block = 5,
  field_body = 6,
  field_condition = 7,
  field_constraints = 8,
  field_else = 9,
  field_exception_name = 10,
  field_field = 11,
  field_from = 12,
  field_guard = 13,
  field_in = 14,
  field_index = 15,
  field_instance = 16,
  field_method = 17,
  field_name = 18,
  field_parameters = 19,
  field_pattern = 20,
  field_return_type = 21,
  field_step = 22,
  field_target = 23,
  field_then = 24,
  field_to = 25,
  field_type = 26,
  field_type_name = 27,
  field_value = 28,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alternative] = "alternative",
  [field_args] = "args",
  [field_assignee] = "assignee",
  [field_base] = "base",
  [field_block] = "block",
  [field_body] = "body",
  [field_condition] = "condition",
  [field_constraints] = "constraints",
  [field_else] = "else",
  [field_exception_name] = "exception_name",
  [field_field] = "field",
  [field_from] = "from",
  [field_guard] = "guard",
  [field_in] = "in",
  [field_index] = "index",
  [field_instance] = "instance",
  [field_method] = "method",
  [field_name] = "name",
  [field_parameters] = "parameters",
  [field_pattern] = "pattern",
  [field_return_type] = "return_type",
  [field_step] = "step",
  [field_target] = "target",
  [field_then] = "then",
  [field_to] = "to",
  [field_type] = "type",
  [field_type_name] = "type_name",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 1},
  [5] = {.index = 2, .length = 1},
  [6] = {.index = 3, .length = 1},
  [7] = {.index = 4, .length = 1},
  [9] = {.index = 5, .length = 1},
  [11] = {.index = 6, .length = 2},
  [12] = {.index = 8, .length = 1},
  [13] = {.index = 9, .length = 1},
  [14] = {.index = 10, .length = 1},
  [15] = {.index = 11, .length = 1},
  [16] = {.index = 12, .length = 1},
  [17] = {.index = 13, .length = 3},
  [18] = {.index = 16, .length = 1},
  [21] = {.index = 17, .length = 1},
  [22] = {.index = 18, .length = 2},
  [25] = {.index = 20, .length = 2},
  [26] = {.index = 22, .length = 2},
  [28] = {.index = 24, .length = 1},
  [29] = {.index = 25, .length = 1},
  [30] = {.index = 26, .length = 1},
  [31] = {.index = 27, .length = 2},
  [32] = {.index = 29, .length = 1},
  [33] = {.index = 30, .length = 1},
  [34] = {.index = 31, .length = 1},
  [35] = {.index = 32, .length = 3},
  [36] = {.index = 35, .length = 1},
  [37] = {.index = 36, .length = 1},
  [38] = {.index = 37, .length = 1},
  [39] = {.index = 38, .length = 3},
  [40] = {.index = 41, .length = 1},
  [41] = {.index = 42, .length = 3},
  [42] = {.index = 45, .length = 1},
  [43] = {.index = 46, .length = 1},
  [44] = {.index = 47, .length = 2},
  [45] = {.index = 49, .length = 1},
  [46] = {.index = 50, .length = 2},
  [47] = {.index = 52, .length = 2},
  [48] = {.index = 54, .length = 1},
  [49] = {.index = 55, .length = 2},
  [52] = {.index = 57, .length = 2},
  [53] = {.index = 59, .length = 2},
  [54] = {.index = 61, .length = 1},
  [55] = {.index = 62, .length = 1},
  [56] = {.index = 63, .length = 3},
  [57] = {.index = 66, .length = 3},
  [58] = {.index = 69, .length = 3},
  [59] = {.index = 72, .length = 4},
  [60] = {.index = 76, .length = 1},
  [61] = {.index = 77, .length = 2},
  [62] = {.index = 79, .length = 3},
  [64] = {.index = 82, .length = 1},
  [65] = {.index = 83, .length = 3},
  [66] = {.index = 86, .length = 2},
  [67] = {.index = 88, .length = 2},
  [68] = {.index = 90, .length = 3},
  [69] = {.index = 93, .length = 2},
  [70] = {.index = 95, .length = 2},
  [71] = {.index = 97, .length = 2},
  [72] = {.index = 99, .length = 4},
  [73] = {.index = 103, .length = 2},
  [74] = {.index = 105, .length = 5},
  [75] = {.index = 110, .length = 1},
  [76] = {.index = 111, .length = 3},
  [77] = {.index = 114, .length = 1},
  [78] = {.index = 115, .length = 3},
  [79] = {.index = 118, .length = 2},
  [80] = {.index = 120, .length = 4},
  [81] = {.index = 124, .length = 3},
  [82] = {.index = 127, .length = 2},
  [83] = {.index = 129, .length = 2},
  [84] = {.index = 131, .length = 3},
  [85] = {.index = 134, .length = 2},
  [86] = {.index = 136, .length = 5},
  [87] = {.index = 141, .length = 4},
  [88] = {.index = 145, .length = 2},
  [89] = {.index = 147, .length = 1},
  [90] = {.index = 148, .length = 1},
  [91] = {.index = 149, .length = 4},
  [92] = {.index = 153, .length = 4},
  [93] = {.index = 157, .length = 3},
  [94] = {.index = 160, .length = 7},
  [95] = {.index = 167, .length = 5},
  [96] = {.index = 172, .length = 1},
  [97] = {.index = 173, .length = 4},
  [98] = {.index = 177, .length = 2},
  [99] = {.index = 179, .length = 8},
  [100] = {.index = 187, .length = 3},
  [101] = {.index = 190, .length = 9},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_name, 1},
  [1] =
    {field_type_name, 0},
  [2] =
    {field_body, 1, .inherited = true},
  [3] =
    {field_guard, 1},
  [4] =
    {field_exception_name, 1},
  [5] =
    {field_in, 1},
  [6] =
    {field_name, 1},
    {field_name, 2},
  [8] =
    {field_name, 2},
  [9] =
    {field_type_name, 1},
  [10] =
    {field_block, 1, .inherited = true},
  [11] =
    {field_body, 2, .inherited = true},
  [12] =
    {field_body, 0, .inherited = true},
  [13] =
    {field_from, 0, .inherited = true},
    {field_step, 0, .inherited = true},
    {field_to, 0, .inherited = true},
  [16] =
    {field_exception_name, 2},
  [17] =
    {field_in, 2},
  [18] =
    {field_assignee, 0},
    {field_value, 2},
  [20] =
    {field_base, 0},
    {field_field, 2},
  [22] =
    {field_guard, 0, .inherited = true},
    {field_then, 2},
  [24] =
    {field_name, 3},
  [25] =
    {field_type_name, 2},
  [26] =
    {field_target, 0},
  [27] =
    {field_body, 0, .inherited = true},
    {field_body, 1, .inherited = true},
  [29] =
    {field_body, 2},
  [30] =
    {field_to, 1},
  [31] =
    {field_from, 0},
  [32] =
    {field_block, 1},
    {field_block, 2},
    {field_block, 3},
  [35] =
    {field_condition, 1},
  [36] =
    {field_exception_name, 3},
  [37] =
    {field_index, 2},
  [38] =
    {field_else, 3, .inherited = true},
    {field_guard, 0, .inherited = true},
    {field_then, 2},
  [41] =
    {field_type_name, 3},
  [42] =
    {field_block, 0},
    {field_block, 1},
    {field_block, 2},
  [45] =
    {field_body, 3},
  [46] =
    {field_constraints, 3},
  [47] =
    {field_from, 0},
    {field_to, 2},
  [49] =
    {field_block, 2, .inherited = true},
  [50] =
    {field_field, 0},
    {field_value, 2},
  [52] =
    {field_block, 2},
    {field_pattern, 0},
  [54] =
    {field_in, 4},
  [55] =
    {field_name, 2},
    {field_return_type, 1},
  [57] =
    {field_condition, 1},
    {field_name, 3, .inherited = true},
  [59] =
    {field_alternative, 3},
    {field_condition, 1},
  [61] =
    {field_name, 4},
  [62] =
    {field_else, 1},
  [63] =
    {field_else, 4, .inherited = true},
    {field_guard, 0, .inherited = true},
    {field_then, 2},
  [66] =
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
  [69] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
  [72] =
    {field_block, 0},
    {field_block, 1},
    {field_block, 2},
    {field_block, 3},
  [76] =
    {field_body, 4},
  [77] =
    {field_name, 1},
    {field_type, 0},
  [79] =
    {field_name, 2},
    {field_parameters, 4},
    {field_return_type, 1},
  [82] =
    {field_name, 1, .inherited = true},
  [83] =
    {field_alternative, 4},
    {field_condition, 1},
    {field_name, 3, .inherited = true},
  [86] =
    {field_alternative, 4},
    {field_condition, 1},
  [88] =
    {field_name, 3},
    {field_return_type, 2},
  [90] =
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
  [93] =
    {field_block, 4},
    {field_block, 5},
  [95] =
    {field_args, 1, .inherited = true},
    {field_name, 0},
  [97] =
    {field_block, 1, .inherited = true},
    {field_name, 0},
  [99] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
  [103] =
    {field_block, 3},
    {field_block, 4},
  [105] =
    {field_block, 0},
    {field_block, 1},
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
  [110] =
    {field_body, 5},
  [111] =
    {field_from, 0},
    {field_step, 2},
    {field_to, 4},
  [114] =
    {field_block, 4, .inherited = true},
  [115] =
    {field_block, 4},
    {field_guard, 2},
    {field_pattern, 0},
  [118] =
    {field_name, 2},
    {field_type, 1},
  [120] =
    {field_name, 2},
    {field_parameters, 4},
    {field_parameters, 5},
    {field_return_type, 1},
  [124] =
    {field_name, 3},
    {field_parameters, 5},
    {field_return_type, 2},
  [127] =
    {field_guard, 1},
    {field_then, 3},
  [129] =
    {field_block, 5},
    {field_block, 6},
  [131] =
    {field_block, 5},
    {field_block, 6},
    {field_block, 7},
  [134] =
    {field_instance, 0},
    {field_method, 2},
  [136] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
  [141] =
    {field_name, 3},
    {field_parameters, 5},
    {field_parameters, 6},
    {field_return_type, 2},
  [145] =
    {field_block, 6},
    {field_block, 7},
  [147] =
    {field_args, 0},
  [148] =
    {field_name, 0},
  [149] =
    {field_block, 1},
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
  [153] =
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
  [157] =
    {field_block, 6},
    {field_block, 7},
    {field_block, 8},
  [160] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
    {field_block, 7},
    {field_block, 8},
  [167] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
    {field_body, 3, .inherited = true},
    {field_name, 0},
  [172] =
    {field_args, 1},
  [173] =
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
    {field_block, 7},
  [177] =
    {field_block, 7},
    {field_block, 8},
  [179] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
    {field_block, 7},
    {field_block, 8},
    {field_block, 9},
  [187] =
    {field_block, 7},
    {field_block, 8},
    {field_block, 9},
  [190] =
    {field_block, 2},
    {field_block, 3},
    {field_block, 4},
    {field_block, 5},
    {field_block, 6},
    {field_block, 7},
    {field_block, 8},
    {field_block, 9},
    {field_block, 10},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_declaration_expression,
  },
  [4] = {
    [0] = alias_sym_wildcard_pattern,
  },
  [8] = {
    [0] = anon_sym_DOT2,
  },
  [10] = {
    [1] = sym_string,
  },
  [19] = {
    [1] = alias_sym_active_pattern_op_name,
  },
  [20] = {
    [0] = anon_sym_DOT2,
    [2] = anon_sym_DOT2,
  },
  [23] = {
    [0] = anon_alias_sym_SEMI,
  },
  [24] = {
    [1] = sym_infix_op,
  },
  [27] = {
    [2] = sym_string,
  },
  [50] = {
    [1] = alias_sym_active_pattern_op_name,
    [3] = alias_sym_wildcard_active_pattern_op,
  },
  [51] = {
    [0] = anon_sym_DOT2,
    [2] = anon_sym_DOT2,
    [4] = anon_sym_DOT2,
  },
  [63] = {
    [1] = alias_sym_active_pattern_op_name,
    [4] = alias_sym_wildcard_active_pattern_op,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_value_declaration, 2,
    sym_value_declaration,
    sym_declaration_expression,
  sym__string_literal, 2,
    sym__string_literal,
    sym_string,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 2,
  [6] = 4,
  [7] = 7,
  [8] = 8,
  [9] = 3,
  [10] = 8,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 11,
  [22] = 12,
  [23] = 13,
  [24] = 14,
  [25] = 15,
  [26] = 16,
  [27] = 17,
  [28] = 18,
  [29] = 19,
  [30] = 20,
  [31] = 2,
  [32] = 4,
  [33] = 7,
  [34] = 4,
  [35] = 20,
  [36] = 2,
  [37] = 2,
  [38] = 7,
  [39] = 4,
  [40] = 2,
  [41] = 4,
  [42] = 7,
  [43] = 3,
  [44] = 8,
  [45] = 11,
  [46] = 12,
  [47] = 3,
  [48] = 8,
  [49] = 11,
  [50] = 12,
  [51] = 13,
  [52] = 14,
  [53] = 15,
  [54] = 16,
  [55] = 17,
  [56] = 18,
  [57] = 19,
  [58] = 20,
  [59] = 13,
  [60] = 14,
  [61] = 15,
  [62] = 16,
  [63] = 17,
  [64] = 18,
  [65] = 19,
  [66] = 7,
  [67] = 14,
  [68] = 14,
  [69] = 15,
  [70] = 16,
  [71] = 17,
  [72] = 18,
  [73] = 19,
  [74] = 20,
  [75] = 2,
  [76] = 4,
  [77] = 15,
  [78] = 16,
  [79] = 17,
  [80] = 18,
  [81] = 2,
  [82] = 3,
  [83] = 8,
  [84] = 19,
  [85] = 20,
  [86] = 4,
  [87] = 11,
  [88] = 12,
  [89] = 8,
  [90] = 7,
  [91] = 13,
  [92] = 11,
  [93] = 12,
  [94] = 3,
  [95] = 13,
  [96] = 19,
  [97] = 3,
  [98] = 11,
  [99] = 12,
  [100] = 8,
  [101] = 3,
  [102] = 13,
  [103] = 11,
  [104] = 12,
  [105] = 14,
  [106] = 13,
  [107] = 14,
  [108] = 15,
  [109] = 16,
  [110] = 17,
  [111] = 18,
  [112] = 15,
  [113] = 19,
  [114] = 20,
  [115] = 16,
  [116] = 7,
  [117] = 18,
  [118] = 8,
  [119] = 4,
  [120] = 20,
  [121] = 7,
  [122] = 11,
  [123] = 12,
  [124] = 2,
  [125] = 13,
  [126] = 14,
  [127] = 15,
  [128] = 16,
  [129] = 17,
  [130] = 18,
  [131] = 4,
  [132] = 19,
  [133] = 20,
  [134] = 2,
  [135] = 4,
  [136] = 7,
  [137] = 3,
  [138] = 8,
  [139] = 2,
  [140] = 17,
  [141] = 11,
  [142] = 4,
  [143] = 2,
  [144] = 4,
  [145] = 2,
  [146] = 4,
  [147] = 2,
  [148] = 7,
  [149] = 3,
  [150] = 8,
  [151] = 14,
  [152] = 15,
  [153] = 16,
  [154] = 17,
  [155] = 18,
  [156] = 19,
  [157] = 12,
  [158] = 13,
  [159] = 20,
  [160] = 2,
  [161] = 4,
  [162] = 12,
  [163] = 163,
  [164] = 164,
  [165] = 11,
  [166] = 164,
  [167] = 13,
  [168] = 14,
  [169] = 15,
  [170] = 16,
  [171] = 17,
  [172] = 18,
  [173] = 19,
  [174] = 20,
  [175] = 3,
  [176] = 8,
  [177] = 7,
  [178] = 11,
  [179] = 12,
  [180] = 13,
  [181] = 14,
  [182] = 15,
  [183] = 16,
  [184] = 17,
  [185] = 18,
  [186] = 19,
  [187] = 20,
  [188] = 7,
  [189] = 3,
  [190] = 8,
  [191] = 163,
  [192] = 11,
  [193] = 19,
  [194] = 20,
  [195] = 7,
  [196] = 7,
  [197] = 163,
  [198] = 3,
  [199] = 12,
  [200] = 164,
  [201] = 13,
  [202] = 3,
  [203] = 8,
  [204] = 163,
  [205] = 14,
  [206] = 15,
  [207] = 16,
  [208] = 17,
  [209] = 18,
  [210] = 8,
  [211] = 164,
  [212] = 11,
  [213] = 12,
  [214] = 13,
  [215] = 14,
  [216] = 15,
  [217] = 16,
  [218] = 17,
  [219] = 18,
  [220] = 19,
  [221] = 20,
  [222] = 12,
  [223] = 3,
  [224] = 13,
  [225] = 225,
  [226] = 14,
  [227] = 15,
  [228] = 16,
  [229] = 17,
  [230] = 18,
  [231] = 19,
  [232] = 8,
  [233] = 20,
  [234] = 7,
  [235] = 11,
  [236] = 236,
  [237] = 237,
  [238] = 7,
  [239] = 12,
  [240] = 3,
  [241] = 13,
  [242] = 8,
  [243] = 20,
  [244] = 11,
  [245] = 14,
  [246] = 15,
  [247] = 16,
  [248] = 17,
  [249] = 18,
  [250] = 19,
  [251] = 12,
  [252] = 16,
  [253] = 17,
  [254] = 18,
  [255] = 19,
  [256] = 20,
  [257] = 15,
  [258] = 16,
  [259] = 3,
  [260] = 225,
  [261] = 17,
  [262] = 13,
  [263] = 8,
  [264] = 7,
  [265] = 11,
  [266] = 12,
  [267] = 267,
  [268] = 7,
  [269] = 18,
  [270] = 14,
  [271] = 20,
  [272] = 272,
  [273] = 273,
  [274] = 267,
  [275] = 19,
  [276] = 276,
  [277] = 272,
  [278] = 278,
  [279] = 11,
  [280] = 3,
  [281] = 15,
  [282] = 14,
  [283] = 8,
  [284] = 273,
  [285] = 13,
  [286] = 286,
  [287] = 11,
  [288] = 12,
  [289] = 3,
  [290] = 13,
  [291] = 14,
  [292] = 15,
  [293] = 16,
  [294] = 17,
  [295] = 18,
  [296] = 8,
  [297] = 19,
  [298] = 20,
  [299] = 7,
  [300] = 272,
  [301] = 11,
  [302] = 7,
  [303] = 14,
  [304] = 15,
  [305] = 16,
  [306] = 17,
  [307] = 18,
  [308] = 19,
  [309] = 309,
  [310] = 11,
  [311] = 12,
  [312] = 13,
  [313] = 14,
  [314] = 15,
  [315] = 315,
  [316] = 16,
  [317] = 286,
  [318] = 318,
  [319] = 319,
  [320] = 17,
  [321] = 321,
  [322] = 11,
  [323] = 12,
  [324] = 3,
  [325] = 13,
  [326] = 14,
  [327] = 15,
  [328] = 16,
  [329] = 17,
  [330] = 18,
  [331] = 8,
  [332] = 19,
  [333] = 20,
  [334] = 7,
  [335] = 18,
  [336] = 19,
  [337] = 20,
  [338] = 7,
  [339] = 267,
  [340] = 315,
  [341] = 3,
  [342] = 8,
  [343] = 309,
  [344] = 344,
  [345] = 345,
  [346] = 11,
  [347] = 12,
  [348] = 318,
  [349] = 13,
  [350] = 14,
  [351] = 15,
  [352] = 16,
  [353] = 17,
  [354] = 18,
  [355] = 355,
  [356] = 19,
  [357] = 20,
  [358] = 12,
  [359] = 13,
  [360] = 20,
  [361] = 11,
  [362] = 14,
  [363] = 15,
  [364] = 16,
  [365] = 17,
  [366] = 18,
  [367] = 19,
  [368] = 319,
  [369] = 369,
  [370] = 344,
  [371] = 345,
  [372] = 372,
  [373] = 286,
  [374] = 321,
  [375] = 375,
  [376] = 309,
  [377] = 7,
  [378] = 164,
  [379] = 273,
  [380] = 12,
  [381] = 315,
  [382] = 286,
  [383] = 321,
  [384] = 3,
  [385] = 8,
  [386] = 7,
  [387] = 11,
  [388] = 12,
  [389] = 13,
  [390] = 14,
  [391] = 15,
  [392] = 16,
  [393] = 17,
  [394] = 18,
  [395] = 19,
  [396] = 20,
  [397] = 315,
  [398] = 286,
  [399] = 321,
  [400] = 315,
  [401] = 286,
  [402] = 8,
  [403] = 315,
  [404] = 286,
  [405] = 315,
  [406] = 286,
  [407] = 315,
  [408] = 286,
  [409] = 315,
  [410] = 286,
  [411] = 315,
  [412] = 286,
  [413] = 315,
  [414] = 286,
  [415] = 315,
  [416] = 286,
  [417] = 315,
  [418] = 286,
  [419] = 315,
  [420] = 286,
  [421] = 315,
  [422] = 286,
  [423] = 355,
  [424] = 315,
  [425] = 286,
  [426] = 315,
  [427] = 286,
  [428] = 315,
  [429] = 286,
  [430] = 315,
  [431] = 286,
  [432] = 315,
  [433] = 286,
  [434] = 315,
  [435] = 286,
  [436] = 315,
  [437] = 286,
  [438] = 315,
  [439] = 286,
  [440] = 315,
  [441] = 286,
  [442] = 315,
  [443] = 286,
  [444] = 315,
  [445] = 286,
  [446] = 315,
  [447] = 286,
  [448] = 286,
  [449] = 286,
  [450] = 286,
  [451] = 286,
  [452] = 286,
  [453] = 286,
  [454] = 355,
  [455] = 286,
  [456] = 286,
  [457] = 286,
  [458] = 286,
  [459] = 286,
  [460] = 369,
  [461] = 3,
  [462] = 13,
  [463] = 7,
  [464] = 3,
  [465] = 8,
  [466] = 8,
  [467] = 20,
  [468] = 3,
  [469] = 369,
  [470] = 355,
  [471] = 355,
  [472] = 355,
  [473] = 355,
  [474] = 355,
  [475] = 355,
  [476] = 355,
  [477] = 355,
  [478] = 355,
  [479] = 355,
  [480] = 355,
  [481] = 355,
  [482] = 355,
  [483] = 355,
  [484] = 355,
  [485] = 355,
  [486] = 355,
  [487] = 355,
  [488] = 355,
  [489] = 355,
  [490] = 355,
  [491] = 355,
  [492] = 355,
  [493] = 355,
  [494] = 321,
  [495] = 495,
  [496] = 496,
  [497] = 11,
  [498] = 12,
  [499] = 13,
  [500] = 14,
  [501] = 15,
  [502] = 16,
  [503] = 17,
  [504] = 18,
  [505] = 19,
  [506] = 20,
  [507] = 7,
  [508] = 11,
  [509] = 509,
  [510] = 12,
  [511] = 3,
  [512] = 13,
  [513] = 14,
  [514] = 15,
  [515] = 16,
  [516] = 17,
  [517] = 18,
  [518] = 8,
  [519] = 519,
  [520] = 7,
  [521] = 19,
  [522] = 12,
  [523] = 3,
  [524] = 13,
  [525] = 309,
  [526] = 496,
  [527] = 527,
  [528] = 528,
  [529] = 529,
  [530] = 530,
  [531] = 318,
  [532] = 8,
  [533] = 20,
  [534] = 20,
  [535] = 319,
  [536] = 536,
  [537] = 318,
  [538] = 319,
  [539] = 11,
  [540] = 7,
  [541] = 14,
  [542] = 15,
  [543] = 16,
  [544] = 17,
  [545] = 18,
  [546] = 19,
  [547] = 547,
  [548] = 7,
  [549] = 519,
  [550] = 11,
  [551] = 12,
  [552] = 3,
  [553] = 7,
  [554] = 13,
  [555] = 14,
  [556] = 12,
  [557] = 3,
  [558] = 13,
  [559] = 15,
  [560] = 16,
  [561] = 17,
  [562] = 18,
  [563] = 8,
  [564] = 8,
  [565] = 20,
  [566] = 19,
  [567] = 11,
  [568] = 12,
  [569] = 3,
  [570] = 13,
  [571] = 14,
  [572] = 15,
  [573] = 16,
  [574] = 17,
  [575] = 18,
  [576] = 8,
  [577] = 19,
  [578] = 20,
  [579] = 527,
  [580] = 11,
  [581] = 581,
  [582] = 495,
  [583] = 496,
  [584] = 527,
  [585] = 528,
  [586] = 529,
  [587] = 530,
  [588] = 20,
  [589] = 7,
  [590] = 14,
  [591] = 536,
  [592] = 15,
  [593] = 16,
  [594] = 17,
  [595] = 18,
  [596] = 528,
  [597] = 19,
  [598] = 598,
  [599] = 163,
  [600] = 495,
  [601] = 496,
  [602] = 529,
  [603] = 530,
  [604] = 495,
  [605] = 496,
  [606] = 529,
  [607] = 530,
  [608] = 608,
  [609] = 495,
  [610] = 496,
  [611] = 529,
  [612] = 612,
  [613] = 495,
  [614] = 496,
  [615] = 529,
  [616] = 616,
  [617] = 529,
  [618] = 495,
  [619] = 496,
  [620] = 529,
  [621] = 621,
  [622] = 622,
  [623] = 495,
  [624] = 496,
  [625] = 529,
  [626] = 626,
  [627] = 495,
  [628] = 496,
  [629] = 529,
  [630] = 598,
  [631] = 495,
  [632] = 496,
  [633] = 529,
  [634] = 495,
  [635] = 496,
  [636] = 529,
  [637] = 11,
  [638] = 495,
  [639] = 496,
  [640] = 529,
  [641] = 164,
  [642] = 12,
  [643] = 495,
  [644] = 496,
  [645] = 529,
  [646] = 3,
  [647] = 13,
  [648] = 495,
  [649] = 496,
  [650] = 529,
  [651] = 14,
  [652] = 15,
  [653] = 16,
  [654] = 496,
  [655] = 529,
  [656] = 17,
  [657] = 18,
  [658] = 495,
  [659] = 496,
  [660] = 529,
  [661] = 8,
  [662] = 19,
  [663] = 495,
  [664] = 496,
  [665] = 529,
  [666] = 20,
  [667] = 7,
  [668] = 495,
  [669] = 496,
  [670] = 529,
  [671] = 495,
  [672] = 496,
  [673] = 529,
  [674] = 11,
  [675] = 495,
  [676] = 496,
  [677] = 529,
  [678] = 12,
  [679] = 495,
  [680] = 496,
  [681] = 529,
  [682] = 3,
  [683] = 13,
  [684] = 495,
  [685] = 496,
  [686] = 529,
  [687] = 14,
  [688] = 15,
  [689] = 495,
  [690] = 496,
  [691] = 529,
  [692] = 16,
  [693] = 17,
  [694] = 495,
  [695] = 496,
  [696] = 529,
  [697] = 18,
  [698] = 8,
  [699] = 495,
  [700] = 496,
  [701] = 529,
  [702] = 19,
  [703] = 20,
  [704] = 495,
  [705] = 496,
  [706] = 529,
  [707] = 7,
  [708] = 529,
  [709] = 529,
  [710] = 529,
  [711] = 529,
  [712] = 712,
  [713] = 529,
  [714] = 3,
  [715] = 529,
  [716] = 529,
  [717] = 529,
  [718] = 529,
  [719] = 529,
  [720] = 720,
  [721] = 529,
  [722] = 530,
  [723] = 529,
  [724] = 8,
  [725] = 529,
  [726] = 529,
  [727] = 529,
  [728] = 529,
  [729] = 529,
  [730] = 529,
  [731] = 529,
  [732] = 529,
  [733] = 529,
  [734] = 529,
  [735] = 529,
  [736] = 529,
  [737] = 529,
  [738] = 529,
  [739] = 529,
  [740] = 529,
  [741] = 529,
  [742] = 529,
  [743] = 529,
  [744] = 529,
  [745] = 529,
  [746] = 529,
  [747] = 529,
  [748] = 529,
  [749] = 529,
  [750] = 529,
  [751] = 529,
  [752] = 529,
  [753] = 529,
  [754] = 529,
  [755] = 529,
  [756] = 529,
  [757] = 529,
  [758] = 529,
  [759] = 529,
  [760] = 529,
  [761] = 529,
  [762] = 529,
  [763] = 529,
  [764] = 529,
  [765] = 529,
  [766] = 529,
  [767] = 529,
  [768] = 529,
  [769] = 529,
  [770] = 529,
  [771] = 529,
  [772] = 529,
  [773] = 529,
  [774] = 529,
  [775] = 529,
  [776] = 529,
  [777] = 529,
  [778] = 529,
  [779] = 529,
  [780] = 529,
  [781] = 529,
  [782] = 529,
  [783] = 529,
  [784] = 529,
  [785] = 529,
  [786] = 529,
  [787] = 529,
  [788] = 529,
  [789] = 529,
  [790] = 529,
  [791] = 529,
  [792] = 529,
  [793] = 529,
  [794] = 529,
  [795] = 529,
  [796] = 529,
  [797] = 529,
  [798] = 529,
  [799] = 529,
  [800] = 529,
  [801] = 612,
  [802] = 581,
  [803] = 626,
  [804] = 7,
  [805] = 495,
  [806] = 11,
  [807] = 12,
  [808] = 598,
  [809] = 612,
  [810] = 581,
  [811] = 626,
  [812] = 13,
  [813] = 612,
  [814] = 581,
  [815] = 626,
  [816] = 612,
  [817] = 581,
  [818] = 626,
  [819] = 612,
  [820] = 581,
  [821] = 626,
  [822] = 612,
  [823] = 581,
  [824] = 626,
  [825] = 612,
  [826] = 581,
  [827] = 626,
  [828] = 612,
  [829] = 581,
  [830] = 626,
  [831] = 612,
  [832] = 581,
  [833] = 626,
  [834] = 612,
  [835] = 581,
  [836] = 626,
  [837] = 612,
  [838] = 581,
  [839] = 626,
  [840] = 612,
  [841] = 581,
  [842] = 626,
  [843] = 612,
  [844] = 581,
  [845] = 626,
  [846] = 612,
  [847] = 581,
  [848] = 626,
  [849] = 612,
  [850] = 581,
  [851] = 626,
  [852] = 612,
  [853] = 581,
  [854] = 626,
  [855] = 612,
  [856] = 581,
  [857] = 626,
  [858] = 612,
  [859] = 581,
  [860] = 626,
  [861] = 612,
  [862] = 581,
  [863] = 626,
  [864] = 612,
  [865] = 581,
  [866] = 626,
  [867] = 612,
  [868] = 581,
  [869] = 626,
  [870] = 612,
  [871] = 581,
  [872] = 626,
  [873] = 612,
  [874] = 581,
  [875] = 626,
  [876] = 612,
  [877] = 581,
  [878] = 626,
  [879] = 612,
  [880] = 581,
  [881] = 626,
  [882] = 612,
  [883] = 581,
  [884] = 626,
  [885] = 581,
  [886] = 626,
  [887] = 581,
  [888] = 581,
  [889] = 581,
  [890] = 581,
  [891] = 509,
  [892] = 509,
  [893] = 509,
  [894] = 509,
  [895] = 509,
  [896] = 509,
  [897] = 509,
  [898] = 509,
  [899] = 509,
  [900] = 509,
  [901] = 509,
  [902] = 509,
  [903] = 509,
  [904] = 509,
  [905] = 509,
  [906] = 509,
  [907] = 509,
  [908] = 509,
  [909] = 509,
  [910] = 509,
  [911] = 509,
  [912] = 509,
  [913] = 509,
  [914] = 509,
  [915] = 509,
  [916] = 509,
  [917] = 14,
  [918] = 15,
  [919] = 16,
  [920] = 17,
  [921] = 18,
  [922] = 547,
  [923] = 19,
  [924] = 20,
  [925] = 3,
  [926] = 536,
  [927] = 927,
  [928] = 8,
  [929] = 495,
  [930] = 930,
  [931] = 930,
  [932] = 932,
  [933] = 930,
  [934] = 934,
  [935] = 932,
  [936] = 930,
  [937] = 932,
  [938] = 930,
  [939] = 939,
  [940] = 932,
  [941] = 930,
  [942] = 939,
  [943] = 932,
  [944] = 930,
  [945] = 932,
  [946] = 930,
  [947] = 932,
  [948] = 930,
  [949] = 932,
  [950] = 930,
  [951] = 934,
  [952] = 932,
  [953] = 930,
  [954] = 932,
  [955] = 930,
  [956] = 932,
  [957] = 930,
  [958] = 932,
  [959] = 930,
  [960] = 932,
  [961] = 932,
  [962] = 930,
  [963] = 932,
  [964] = 930,
  [965] = 932,
  [966] = 930,
  [967] = 932,
  [968] = 930,
  [969] = 598,
  [970] = 934,
  [971] = 932,
  [972] = 930,
  [973] = 932,
  [974] = 930,
  [975] = 930,
  [976] = 932,
  [977] = 930,
  [978] = 932,
  [979] = 930,
  [980] = 932,
  [981] = 930,
  [982] = 932,
  [983] = 932,
  [984] = 930,
  [985] = 930,
  [986] = 932,
  [987] = 930,
  [988] = 932,
  [989] = 939,
  [990] = 932,
  [991] = 991,
  [992] = 992,
  [993] = 993,
  [994] = 994,
  [995] = 995,
  [996] = 996,
  [997] = 997,
  [998] = 998,
  [999] = 999,
  [1000] = 1000,
  [1001] = 1001,
  [1002] = 1002,
  [1003] = 992,
  [1004] = 1004,
  [1005] = 997,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 1011,
  [1012] = 1012,
  [1013] = 1013,
  [1014] = 1004,
  [1015] = 1012,
  [1016] = 1016,
  [1017] = 1012,
  [1018] = 999,
  [1019] = 1019,
  [1020] = 1020,
  [1021] = 1021,
  [1022] = 1012,
  [1023] = 1023,
  [1024] = 1020,
  [1025] = 1007,
  [1026] = 1008,
  [1027] = 1013,
  [1028] = 1013,
  [1029] = 1013,
  [1030] = 1012,
  [1031] = 1012,
  [1032] = 1020,
  [1033] = 1021,
  [1034] = 1020,
  [1035] = 1035,
  [1036] = 1021,
  [1037] = 1010,
  [1038] = 1020,
  [1039] = 1039,
  [1040] = 1021,
  [1041] = 1013,
  [1042] = 1000,
  [1043] = 1012,
  [1044] = 1021,
  [1045] = 1020,
  [1046] = 1021,
  [1047] = 1012,
  [1048] = 1048,
  [1049] = 1013,
  [1050] = 1013,
  [1051] = 997,
  [1052] = 1013,
  [1053] = 1053,
  [1054] = 1020,
  [1055] = 1021,
  [1056] = 997,
  [1057] = 1057,
  [1058] = 1021,
  [1059] = 1020,
  [1060] = 998,
  [1061] = 1061,
  [1062] = 1009,
  [1063] = 1063,
  [1064] = 1064,
  [1065] = 1065,
  [1066] = 1063,
  [1067] = 1065,
  [1068] = 1068,
  [1069] = 1069,
  [1070] = 1069,
  [1071] = 1064,
  [1072] = 1064,
  [1073] = 1063,
  [1074] = 1069,
  [1075] = 1065,
  [1076] = 1068,
  [1077] = 1068,
  [1078] = 1064,
  [1079] = 1068,
  [1080] = 1069,
  [1081] = 1068,
  [1082] = 1069,
  [1083] = 1069,
  [1084] = 1064,
  [1085] = 1064,
  [1086] = 1063,
  [1087] = 1064,
  [1088] = 1065,
  [1089] = 1089,
  [1090] = 1063,
  [1091] = 1063,
  [1092] = 1068,
  [1093] = 1063,
  [1094] = 1068,
  [1095] = 1069,
  [1096] = 1063,
  [1097] = 1065,
  [1098] = 1065,
  [1099] = 1065,
  [1100] = 1064,
  [1101] = 1068,
  [1102] = 1065,
  [1103] = 1069,
  [1104] = 1104,
  [1105] = 1104,
  [1106] = 1106,
  [1107] = 1107,
  [1108] = 1106,
  [1109] = 1109,
  [1110] = 1109,
  [1111] = 1109,
  [1112] = 1109,
  [1113] = 1113,
  [1114] = 1109,
  [1115] = 1115,
  [1116] = 1115,
  [1117] = 1109,
  [1118] = 1109,
  [1119] = 1115,
  [1120] = 1115,
  [1121] = 1109,
  [1122] = 1122,
  [1123] = 1122,
  [1124] = 1122,
  [1125] = 1122,
  [1126] = 1122,
  [1127] = 1122,
  [1128] = 1122,
  [1129] = 1122,
  [1130] = 1122,
  [1131] = 1122,
  [1132] = 1122,
  [1133] = 1122,
  [1134] = 1122,
  [1135] = 1122,
  [1136] = 1122,
  [1137] = 1122,
  [1138] = 1122,
  [1139] = 1139,
  [1140] = 1122,
  [1141] = 1122,
  [1142] = 1122,
  [1143] = 1122,
  [1144] = 1122,
  [1145] = 1122,
  [1146] = 1122,
  [1147] = 1122,
  [1148] = 1122,
  [1149] = 1122,
  [1150] = 1150,
  [1151] = 1151,
  [1152] = 1152,
  [1153] = 1151,
  [1154] = 1154,
  [1155] = 1154,
  [1156] = 1151,
  [1157] = 1152,
  [1158] = 1154,
  [1159] = 1152,
  [1160] = 1160,
  [1161] = 1160,
  [1162] = 1162,
  [1163] = 1163,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 1162,
  [1167] = 1163,
  [1168] = 1165,
  [1169] = 1109,
  [1170] = 1163,
  [1171] = 1165,
  [1172] = 1115,
  [1173] = 1163,
  [1174] = 1165,
  [1175] = 1163,
  [1176] = 1165,
  [1177] = 1163,
  [1178] = 1165,
  [1179] = 1163,
  [1180] = 1165,
  [1181] = 1163,
  [1182] = 1165,
  [1183] = 1163,
  [1184] = 1165,
  [1185] = 1163,
  [1186] = 1165,
  [1187] = 1163,
  [1188] = 1165,
  [1189] = 1163,
  [1190] = 1165,
  [1191] = 1163,
  [1192] = 1165,
  [1193] = 1163,
  [1194] = 1165,
  [1195] = 1162,
  [1196] = 1163,
  [1197] = 1165,
  [1198] = 1163,
  [1199] = 1165,
  [1200] = 1163,
  [1201] = 1165,
  [1202] = 1163,
  [1203] = 1165,
  [1204] = 1163,
  [1205] = 1165,
  [1206] = 1163,
  [1207] = 1165,
  [1208] = 1160,
  [1209] = 1165,
  [1210] = 1163,
  [1211] = 1165,
  [1212] = 1212,
  [1213] = 1163,
  [1214] = 1165,
  [1215] = 1163,
  [1216] = 1165,
  [1217] = 1163,
  [1218] = 1165,
  [1219] = 1163,
  [1220] = 1165,
  [1221] = 1109,
  [1222] = 1163,
  [1223] = 1162,
  [1224] = 1160,
  [1225] = 1160,
  [1226] = 1160,
  [1227] = 1160,
  [1228] = 1164,
  [1229] = 1160,
  [1230] = 1160,
  [1231] = 1165,
  [1232] = 1160,
  [1233] = 1160,
  [1234] = 1160,
  [1235] = 1160,
  [1236] = 1160,
  [1237] = 1160,
  [1238] = 1160,
  [1239] = 1160,
  [1240] = 1160,
  [1241] = 1160,
  [1242] = 1160,
  [1243] = 1160,
  [1244] = 1160,
  [1245] = 1160,
  [1246] = 1160,
  [1247] = 1160,
  [1248] = 1160,
  [1249] = 1160,
  [1250] = 1163,
  [1251] = 1251,
  [1252] = 1252,
  [1253] = 1253,
  [1254] = 1254,
  [1255] = 1109,
  [1256] = 1254,
  [1257] = 1257,
  [1258] = 1257,
  [1259] = 1254,
  [1260] = 1251,
  [1261] = 1115,
  [1262] = 1251,
  [1263] = 1251,
  [1264] = 1251,
  [1265] = 1251,
  [1266] = 1251,
  [1267] = 1251,
  [1268] = 1251,
  [1269] = 1251,
  [1270] = 1251,
  [1271] = 1251,
  [1272] = 1251,
  [1273] = 1251,
  [1274] = 1251,
  [1275] = 1251,
  [1276] = 1251,
  [1277] = 1251,
  [1278] = 1251,
  [1279] = 1251,
  [1280] = 1251,
  [1281] = 1251,
  [1282] = 1251,
  [1283] = 1251,
  [1284] = 1251,
  [1285] = 1251,
  [1286] = 1251,
  [1287] = 1287,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1289,
  [1291] = 1288,
  [1292] = 1289,
  [1293] = 1288,
  [1294] = 1288,
  [1295] = 1289,
  [1296] = 1289,
  [1297] = 1289,
  [1298] = 1288,
  [1299] = 1289,
  [1300] = 1289,
  [1301] = 1301,
  [1302] = 1288,
  [1303] = 1288,
  [1304] = 1289,
  [1305] = 1289,
  [1306] = 1289,
  [1307] = 1288,
  [1308] = 1289,
  [1309] = 1288,
  [1310] = 1289,
  [1311] = 1288,
  [1312] = 1289,
  [1313] = 1289,
  [1314] = 1289,
  [1315] = 1289,
  [1316] = 1252,
  [1317] = 1289,
  [1318] = 1288,
  [1319] = 1289,
  [1320] = 1288,
  [1321] = 1288,
  [1322] = 1289,
  [1323] = 1288,
  [1324] = 1288,
  [1325] = 1289,
  [1326] = 1326,
  [1327] = 1288,
  [1328] = 1289,
  [1329] = 1288,
  [1330] = 1288,
  [1331] = 1289,
  [1332] = 1288,
  [1333] = 1288,
  [1334] = 1288,
  [1335] = 1289,
  [1336] = 1288,
  [1337] = 1288,
  [1338] = 1288,
  [1339] = 1289,
  [1340] = 1288,
  [1341] = 1288,
  [1342] = 1301,
  [1343] = 1301,
  [1344] = 1289,
  [1345] = 1301,
  [1346] = 1289,
  [1347] = 1301,
  [1348] = 1288,
  [1349] = 1301,
  [1350] = 1288,
  [1351] = 1301,
  [1352] = 1301,
  [1353] = 1289,
  [1354] = 1301,
  [1355] = 1301,
  [1356] = 1289,
  [1357] = 1301,
  [1358] = 1288,
  [1359] = 1301,
  [1360] = 1288,
  [1361] = 1301,
  [1362] = 1301,
  [1363] = 1289,
  [1364] = 1301,
  [1365] = 1289,
  [1366] = 1301,
  [1367] = 1301,
  [1368] = 1301,
  [1369] = 1288,
  [1370] = 1301,
  [1371] = 1288,
  [1372] = 1301,
  [1373] = 1301,
  [1374] = 1289,
  [1375] = 1301,
  [1376] = 1301,
  [1377] = 1288,
  [1378] = 1301,
  [1379] = 1288,
  [1380] = 1301,
  [1381] = 1301,
  [1382] = 1289,
  [1383] = 1301,
  [1384] = 1289,
  [1385] = 1385,
  [1386] = 1385,
  [1387] = 1385,
  [1388] = 1385,
  [1389] = 1385,
  [1390] = 1385,
  [1391] = 1385,
  [1392] = 1385,
  [1393] = 1385,
  [1394] = 1385,
  [1395] = 1385,
  [1396] = 1385,
  [1397] = 1385,
  [1398] = 1385,
  [1399] = 1385,
  [1400] = 1385,
  [1401] = 1385,
  [1402] = 1385,
  [1403] = 1385,
  [1404] = 1385,
  [1405] = 1385,
  [1406] = 1385,
  [1407] = 1385,
  [1408] = 1385,
  [1409] = 1385,
  [1410] = 1385,
  [1411] = 1385,
  [1412] = 1385,
  [1413] = 1413,
  [1414] = 1385,
  [1415] = 1415,
  [1416] = 1385,
  [1417] = 1385,
  [1418] = 1385,
  [1419] = 1385,
  [1420] = 1385,
  [1421] = 1415,
  [1422] = 1422,
  [1423] = 1423,
  [1424] = 1424,
  [1425] = 1424,
  [1426] = 1422,
  [1427] = 1423,
  [1428] = 1424,
  [1429] = 1424,
  [1430] = 1422,
  [1431] = 1423,
  [1432] = 1424,
  [1433] = 1424,
  [1434] = 1422,
  [1435] = 1423,
  [1436] = 1424,
  [1437] = 1424,
  [1438] = 1422,
  [1439] = 1423,
  [1440] = 1424,
  [1441] = 1424,
  [1442] = 1422,
  [1443] = 1423,
  [1444] = 1444,
  [1445] = 1424,
  [1446] = 1422,
  [1447] = 1423,
  [1448] = 1424,
  [1449] = 1424,
  [1450] = 1424,
  [1451] = 1422,
  [1452] = 1423,
  [1453] = 1424,
  [1454] = 1424,
  [1455] = 1422,
  [1456] = 1423,
  [1457] = 1424,
  [1458] = 1424,
  [1459] = 1422,
  [1460] = 1423,
  [1461] = 1424,
  [1462] = 1462,
  [1463] = 1422,
  [1464] = 1423,
  [1465] = 1424,
  [1466] = 1424,
  [1467] = 1422,
  [1468] = 1423,
  [1469] = 1424,
  [1470] = 1424,
  [1471] = 1422,
  [1472] = 1423,
  [1473] = 1424,
  [1474] = 1474,
  [1475] = 1422,
  [1476] = 1423,
  [1477] = 1424,
  [1478] = 1424,
  [1479] = 1422,
  [1480] = 1423,
  [1481] = 1424,
  [1482] = 1424,
  [1483] = 1422,
  [1484] = 1423,
  [1485] = 1424,
  [1486] = 1486,
  [1487] = 1422,
  [1488] = 1423,
  [1489] = 1422,
  [1490] = 1424,
  [1491] = 1422,
  [1492] = 1423,
  [1493] = 1424,
  [1494] = 1422,
  [1495] = 1422,
  [1496] = 1423,
  [1497] = 1423,
  [1498] = 1462,
  [1499] = 1422,
  [1500] = 1423,
  [1501] = 1423,
  [1502] = 1502,
  [1503] = 1422,
  [1504] = 1423,
  [1505] = 1422,
  [1506] = 1423,
  [1507] = 1422,
  [1508] = 1423,
  [1509] = 1462,
  [1510] = 1424,
  [1511] = 1422,
  [1512] = 1423,
  [1513] = 1424,
  [1514] = 1422,
  [1515] = 1422,
  [1516] = 1423,
  [1517] = 1422,
  [1518] = 1422,
  [1519] = 1422,
  [1520] = 1422,
  [1521] = 1424,
  [1522] = 1522,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1526,
  [1529] = 1529,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1527,
  [1534] = 1524,
  [1535] = 1524,
  [1536] = 1536,
  [1537] = 1537,
  [1538] = 1525,
  [1539] = 1539,
  [1540] = 1529,
  [1541] = 1541,
  [1542] = 1526,
  [1543] = 1543,
  [1544] = 1529,
  [1545] = 1545,
  [1546] = 1546,
  [1547] = 1547,
  [1548] = 1548,
  [1549] = 1530,
  [1550] = 1536,
  [1551] = 1537,
  [1552] = 1531,
  [1553] = 1532,
  [1554] = 1527,
  [1555] = 1555,
  [1556] = 1556,
  [1557] = 1557,
  [1558] = 1529,
  [1559] = 1524,
  [1560] = 1530,
  [1561] = 1531,
  [1562] = 1532,
  [1563] = 1527,
  [1564] = 1564,
  [1565] = 1524,
  [1566] = 1525,
  [1567] = 1525,
  [1568] = 1526,
  [1569] = 1530,
  [1570] = 1529,
  [1571] = 1571,
  [1572] = 1526,
  [1573] = 1522,
  [1574] = 1529,
  [1575] = 1531,
  [1576] = 1532,
  [1577] = 1527,
  [1578] = 1524,
  [1579] = 1543,
  [1580] = 1525,
  [1581] = 1526,
  [1582] = 1529,
  [1583] = 1531,
  [1584] = 1532,
  [1585] = 1527,
  [1586] = 1524,
  [1587] = 1525,
  [1588] = 1526,
  [1589] = 1531,
  [1590] = 1532,
  [1591] = 1591,
  [1592] = 1531,
  [1593] = 1532,
  [1594] = 1527,
  [1595] = 1527,
  [1596] = 1524,
  [1597] = 1525,
  [1598] = 1524,
  [1599] = 1526,
  [1600] = 1545,
  [1601] = 1555,
  [1602] = 1525,
  [1603] = 1526,
  [1604] = 1523,
  [1605] = 1529,
  [1606] = 1531,
  [1607] = 1532,
  [1608] = 1527,
  [1609] = 1546,
  [1610] = 1524,
  [1611] = 1536,
  [1612] = 1556,
  [1613] = 1555,
  [1614] = 1525,
  [1615] = 1556,
  [1616] = 1530,
  [1617] = 1529,
  [1618] = 1529,
  [1619] = 1526,
  [1620] = 1531,
  [1621] = 1532,
  [1622] = 1527,
  [1623] = 1529,
  [1624] = 1524,
  [1625] = 1525,
  [1626] = 1526,
  [1627] = 1529,
  [1628] = 1531,
  [1629] = 1532,
  [1630] = 1527,
  [1631] = 1524,
  [1632] = 1525,
  [1633] = 1526,
  [1634] = 1546,
  [1635] = 1536,
  [1636] = 1555,
  [1637] = 1556,
  [1638] = 1530,
  [1639] = 1529,
  [1640] = 1531,
  [1641] = 1532,
  [1642] = 1527,
  [1643] = 1524,
  [1644] = 1525,
  [1645] = 1526,
  [1646] = 1531,
  [1647] = 1532,
  [1648] = 1555,
  [1649] = 1556,
  [1650] = 1530,
  [1651] = 1529,
  [1652] = 1531,
  [1653] = 1532,
  [1654] = 1527,
  [1655] = 1524,
  [1656] = 1525,
  [1657] = 1526,
  [1658] = 1527,
  [1659] = 1555,
  [1660] = 1556,
  [1661] = 1530,
  [1662] = 1529,
  [1663] = 1531,
  [1664] = 1532,
  [1665] = 1527,
  [1666] = 1524,
  [1667] = 1525,
  [1668] = 1526,
  [1669] = 1555,
  [1670] = 1556,
  [1671] = 1530,
  [1672] = 1529,
  [1673] = 1525,
  [1674] = 1531,
  [1675] = 1532,
  [1676] = 1527,
  [1677] = 1524,
  [1678] = 1525,
  [1679] = 1526,
  [1680] = 1555,
  [1681] = 1556,
  [1682] = 1530,
  [1683] = 1529,
  [1684] = 1531,
  [1685] = 1532,
  [1686] = 1527,
  [1687] = 1524,
  [1688] = 1525,
  [1689] = 1526,
  [1690] = 1530,
  [1691] = 1524,
  [1692] = 1555,
  [1693] = 1556,
  [1694] = 1530,
  [1695] = 1529,
  [1696] = 1531,
  [1697] = 1532,
  [1698] = 1527,
  [1699] = 1524,
  [1700] = 1525,
  [1701] = 1526,
  [1702] = 1531,
  [1703] = 1532,
  [1704] = 1555,
  [1705] = 1556,
  [1706] = 1530,
  [1707] = 1529,
  [1708] = 1531,
  [1709] = 1532,
  [1710] = 1527,
  [1711] = 1524,
  [1712] = 1525,
  [1713] = 1526,
  [1714] = 1555,
  [1715] = 1556,
  [1716] = 1530,
  [1717] = 1529,
  [1718] = 1531,
  [1719] = 1532,
  [1720] = 1527,
  [1721] = 1524,
  [1722] = 1525,
  [1723] = 1526,
  [1724] = 1555,
  [1725] = 1556,
  [1726] = 1530,
  [1727] = 1529,
  [1728] = 1531,
  [1729] = 1532,
  [1730] = 1527,
  [1731] = 1524,
  [1732] = 1525,
  [1733] = 1526,
  [1734] = 1525,
  [1735] = 1527,
  [1736] = 1555,
  [1737] = 1556,
  [1738] = 1530,
  [1739] = 1529,
  [1740] = 1531,
  [1741] = 1532,
  [1742] = 1523,
  [1743] = 1524,
  [1744] = 1525,
  [1745] = 1526,
  [1746] = 1555,
  [1747] = 1556,
  [1748] = 1530,
  [1749] = 1529,
  [1750] = 1531,
  [1751] = 1532,
  [1752] = 1527,
  [1753] = 1524,
  [1754] = 1525,
  [1755] = 1526,
  [1756] = 1555,
  [1757] = 1556,
  [1758] = 1530,
  [1759] = 1529,
  [1760] = 1526,
  [1761] = 1531,
  [1762] = 1532,
  [1763] = 1527,
  [1764] = 1524,
  [1765] = 1525,
  [1766] = 1526,
  [1767] = 1530,
  [1768] = 1529,
  [1769] = 1531,
  [1770] = 1532,
  [1771] = 1527,
  [1772] = 1524,
  [1773] = 1525,
  [1774] = 1526,
  [1775] = 1526,
  [1776] = 1530,
  [1777] = 1529,
  [1778] = 1531,
  [1779] = 1532,
  [1780] = 1527,
  [1781] = 1524,
  [1782] = 1525,
  [1783] = 1526,
  [1784] = 1530,
  [1785] = 1529,
  [1786] = 1531,
  [1787] = 1532,
  [1788] = 1527,
  [1789] = 1524,
  [1790] = 1525,
  [1791] = 1526,
  [1792] = 1530,
  [1793] = 1530,
  [1794] = 1529,
  [1795] = 1531,
  [1796] = 1532,
  [1797] = 1527,
  [1798] = 1524,
  [1799] = 1525,
  [1800] = 1526,
  [1801] = 1530,
  [1802] = 1529,
  [1803] = 1531,
  [1804] = 1530,
  [1805] = 1529,
  [1806] = 1532,
  [1807] = 1527,
  [1808] = 1808,
  [1809] = 1529,
  [1810] = 1524,
  [1811] = 1530,
  [1812] = 1524,
  [1813] = 1524,
  [1814] = 1530,
  [1815] = 1529,
  [1816] = 1530,
  [1817] = 1525,
  [1818] = 1530,
  [1819] = 1526,
  [1820] = 1820,
  [1821] = 1530,
  [1822] = 1525,
  [1823] = 1530,
  [1824] = 1531,
  [1825] = 1532,
  [1826] = 1527,
  [1827] = 1524,
  [1828] = 1525,
  [1829] = 1526,
  [1830] = 1526,
  [1831] = 1831,
  [1832] = 1546,
  [1833] = 1833,
  [1834] = 1529,
  [1835] = 1530,
  [1836] = 1531,
  [1837] = 1532,
  [1838] = 1527,
  [1839] = 1839,
  [1840] = 1840,
  [1841] = 1548,
  [1842] = 1557,
  [1843] = 1522,
  [1844] = 1844,
  [1845] = 1831,
  [1846] = 1525,
  [1847] = 1839,
  [1848] = 1840,
  [1849] = 1548,
  [1850] = 1557,
  [1851] = 1522,
  [1852] = 1831,
  [1853] = 1839,
  [1854] = 1840,
  [1855] = 1522,
  [1856] = 1831,
  [1857] = 1524,
  [1858] = 1839,
  [1859] = 1840,
  [1860] = 1522,
  [1861] = 1831,
  [1862] = 1547,
  [1863] = 1839,
  [1864] = 1840,
  [1865] = 1522,
  [1866] = 1831,
  [1867] = 1839,
  [1868] = 1840,
  [1869] = 1522,
  [1870] = 1525,
  [1871] = 1831,
  [1872] = 1839,
  [1873] = 1840,
  [1874] = 1522,
  [1875] = 1831,
  [1876] = 1839,
  [1877] = 1840,
  [1878] = 1522,
  [1879] = 1831,
  [1880] = 1839,
  [1881] = 1840,
  [1882] = 1522,
  [1883] = 1526,
  [1884] = 1831,
  [1885] = 1839,
  [1886] = 1840,
  [1887] = 1522,
  [1888] = 1831,
  [1889] = 1839,
  [1890] = 1840,
  [1891] = 1522,
  [1892] = 1831,
  [1893] = 1839,
  [1894] = 1840,
  [1895] = 1522,
  [1896] = 1831,
  [1897] = 1839,
  [1898] = 1840,
  [1899] = 1522,
  [1900] = 1831,
  [1901] = 1839,
  [1902] = 1840,
  [1903] = 1522,
  [1904] = 1831,
  [1905] = 1839,
  [1906] = 1840,
  [1907] = 1522,
  [1908] = 1831,
  [1909] = 1839,
  [1910] = 1840,
  [1911] = 1522,
  [1912] = 1831,
  [1913] = 1839,
  [1914] = 1840,
  [1915] = 1522,
  [1916] = 1529,
  [1917] = 1831,
  [1918] = 1839,
  [1919] = 1840,
  [1920] = 1522,
  [1921] = 1831,
  [1922] = 1839,
  [1923] = 1840,
  [1924] = 1522,
  [1925] = 1839,
  [1926] = 1831,
  [1927] = 1839,
  [1928] = 1840,
  [1929] = 1522,
  [1930] = 1831,
  [1931] = 1839,
  [1932] = 1840,
  [1933] = 1522,
  [1934] = 1831,
  [1935] = 1839,
  [1936] = 1840,
  [1937] = 1522,
  [1938] = 1831,
  [1939] = 1839,
  [1940] = 1840,
  [1941] = 1522,
  [1942] = 1831,
  [1943] = 1839,
  [1944] = 1840,
  [1945] = 1522,
  [1946] = 1831,
  [1947] = 1839,
  [1948] = 1840,
  [1949] = 1522,
  [1950] = 1831,
  [1951] = 1839,
  [1952] = 1840,
  [1953] = 1522,
  [1954] = 1831,
  [1955] = 1522,
  [1956] = 1831,
  [1957] = 1522,
  [1958] = 1831,
  [1959] = 1522,
  [1960] = 1831,
  [1961] = 1530,
  [1962] = 1522,
  [1963] = 1831,
  [1964] = 1522,
  [1965] = 1831,
  [1966] = 1522,
  [1967] = 1831,
  [1968] = 1522,
  [1969] = 1831,
  [1970] = 1522,
  [1971] = 1831,
  [1972] = 1522,
  [1973] = 1831,
  [1974] = 1522,
  [1975] = 1831,
  [1976] = 1522,
  [1977] = 1831,
  [1978] = 1522,
  [1979] = 1831,
  [1980] = 1522,
  [1981] = 1522,
  [1982] = 1522,
  [1983] = 1522,
  [1984] = 1522,
  [1985] = 1522,
  [1986] = 1522,
  [1987] = 1522,
  [1988] = 1522,
  [1989] = 1522,
  [1990] = 1522,
  [1991] = 1522,
  [1992] = 1522,
  [1993] = 1522,
  [1994] = 1522,
  [1995] = 1522,
  [1996] = 1522,
  [1997] = 1522,
  [1998] = 1522,
  [1999] = 1522,
  [2000] = 1522,
  [2001] = 1522,
  [2002] = 1522,
  [2003] = 1522,
  [2004] = 1522,
  [2005] = 1522,
  [2006] = 1522,
  [2007] = 1522,
  [2008] = 1522,
  [2009] = 1522,
  [2010] = 1522,
  [2011] = 1522,
  [2012] = 1522,
  [2013] = 1522,
  [2014] = 1522,
  [2015] = 1522,
  [2016] = 1522,
  [2017] = 1522,
  [2018] = 1522,
  [2019] = 1522,
  [2020] = 1522,
  [2021] = 1522,
  [2022] = 1522,
  [2023] = 1522,
  [2024] = 1522,
  [2025] = 1840,
  [2026] = 1522,
  [2027] = 1522,
  [2028] = 1522,
  [2029] = 1522,
  [2030] = 1522,
  [2031] = 1522,
  [2032] = 1522,
  [2033] = 1522,
  [2034] = 1522,
  [2035] = 1522,
  [2036] = 1522,
  [2037] = 1522,
  [2038] = 1522,
  [2039] = 1522,
  [2040] = 1522,
  [2041] = 1522,
  [2042] = 1522,
  [2043] = 1522,
  [2044] = 1522,
  [2045] = 1522,
  [2046] = 1522,
  [2047] = 1522,
  [2048] = 1522,
  [2049] = 1522,
  [2050] = 1522,
  [2051] = 1522,
  [2052] = 1522,
  [2053] = 1522,
  [2054] = 1522,
  [2055] = 1522,
  [2056] = 1820,
  [2057] = 1564,
  [2058] = 1844,
  [2059] = 1808,
  [2060] = 1820,
  [2061] = 1564,
  [2062] = 1531,
  [2063] = 1844,
  [2064] = 1527,
  [2065] = 1808,
  [2066] = 1820,
  [2067] = 1564,
  [2068] = 1530,
  [2069] = 1844,
  [2070] = 1808,
  [2071] = 1820,
  [2072] = 1564,
  [2073] = 1844,
  [2074] = 1808,
  [2075] = 1820,
  [2076] = 1564,
  [2077] = 1844,
  [2078] = 1808,
  [2079] = 1820,
  [2080] = 1564,
  [2081] = 1844,
  [2082] = 1808,
  [2083] = 1820,
  [2084] = 1564,
  [2085] = 1844,
  [2086] = 1808,
  [2087] = 1820,
  [2088] = 1564,
  [2089] = 1844,
  [2090] = 1808,
  [2091] = 1820,
  [2092] = 1564,
  [2093] = 1844,
  [2094] = 1808,
  [2095] = 1820,
  [2096] = 1564,
  [2097] = 1531,
  [2098] = 1844,
  [2099] = 1808,
  [2100] = 1820,
  [2101] = 1564,
  [2102] = 1844,
  [2103] = 1808,
  [2104] = 1820,
  [2105] = 1564,
  [2106] = 1844,
  [2107] = 1808,
  [2108] = 1820,
  [2109] = 1564,
  [2110] = 1532,
  [2111] = 1844,
  [2112] = 1808,
  [2113] = 1820,
  [2114] = 1564,
  [2115] = 1844,
  [2116] = 1808,
  [2117] = 1820,
  [2118] = 1564,
  [2119] = 1844,
  [2120] = 1808,
  [2121] = 1820,
  [2122] = 1564,
  [2123] = 1844,
  [2124] = 1808,
  [2125] = 1820,
  [2126] = 1564,
  [2127] = 1527,
  [2128] = 1844,
  [2129] = 1808,
  [2130] = 1820,
  [2131] = 1564,
  [2132] = 1844,
  [2133] = 1808,
  [2134] = 1820,
  [2135] = 1564,
  [2136] = 1844,
  [2137] = 1808,
  [2138] = 1820,
  [2139] = 1564,
  [2140] = 1844,
  [2141] = 1808,
  [2142] = 1820,
  [2143] = 1564,
  [2144] = 1844,
  [2145] = 1808,
  [2146] = 1820,
  [2147] = 1564,
  [2148] = 1532,
  [2149] = 1844,
  [2150] = 1808,
  [2151] = 1820,
  [2152] = 1564,
  [2153] = 1844,
  [2154] = 1808,
  [2155] = 1820,
  [2156] = 1564,
  [2157] = 1844,
  [2158] = 1808,
  [2159] = 1820,
  [2160] = 1564,
  [2161] = 1844,
  [2162] = 1808,
  [2163] = 1820,
  [2164] = 1564,
  [2165] = 1844,
  [2166] = 1808,
  [2167] = 1844,
  [2168] = 1541,
  [2169] = 1541,
  [2170] = 1541,
  [2171] = 1531,
  [2172] = 1541,
  [2173] = 1541,
  [2174] = 1541,
  [2175] = 1541,
  [2176] = 1541,
  [2177] = 1541,
  [2178] = 1541,
  [2179] = 1541,
  [2180] = 1541,
  [2181] = 1541,
  [2182] = 1541,
  [2183] = 1541,
  [2184] = 1541,
  [2185] = 1541,
  [2186] = 1541,
  [2187] = 1541,
  [2188] = 1541,
  [2189] = 1541,
  [2190] = 1541,
  [2191] = 1532,
  [2192] = 1541,
  [2193] = 1541,
  [2194] = 1541,
  [2195] = 1541,
  [2196] = 1523,
  [2197] = 1523,
  [2198] = 1523,
  [2199] = 1523,
  [2200] = 1523,
  [2201] = 1523,
  [2202] = 1523,
  [2203] = 1523,
  [2204] = 1523,
  [2205] = 1523,
  [2206] = 1523,
  [2207] = 1523,
  [2208] = 1523,
  [2209] = 1523,
  [2210] = 1523,
  [2211] = 1523,
  [2212] = 1523,
  [2213] = 1523,
  [2214] = 1523,
  [2215] = 1523,
  [2216] = 1523,
  [2217] = 1523,
  [2218] = 1523,
  [2219] = 1523,
  [2220] = 1527,
  [2221] = 2221,
  [2222] = 2221,
  [2223] = 2221,
  [2224] = 2221,
  [2225] = 2225,
  [2226] = 2225,
  [2227] = 2225,
  [2228] = 2225,
  [2229] = 2221,
  [2230] = 2221,
  [2231] = 2221,
  [2232] = 2221,
  [2233] = 2221,
  [2234] = 2234,
  [2235] = 2225,
  [2236] = 2236,
  [2237] = 2225,
  [2238] = 2234,
  [2239] = 2221,
  [2240] = 2240,
  [2241] = 2241,
  [2242] = 2242,
  [2243] = 2240,
  [2244] = 2241,
  [2245] = 2236,
  [2246] = 2242,
  [2247] = 2225,
  [2248] = 2225,
  [2249] = 2225,
  [2250] = 2242,
  [2251] = 2240,
  [2252] = 2236,
  [2253] = 2234,
  [2254] = 2241,
  [2255] = 2255,
  [2256] = 2256,
  [2257] = 2225,
  [2258] = 2258,
  [2259] = 2259,
  [2260] = 2260,
  [2261] = 2255,
  [2262] = 2262,
  [2263] = 2263,
  [2264] = 2264,
  [2265] = 2265,
  [2266] = 2263,
  [2267] = 2265,
  [2268] = 2264,
  [2269] = 2258,
  [2270] = 2259,
  [2271] = 2260,
  [2272] = 2272,
  [2273] = 2256,
  [2274] = 2262,
  [2275] = 2272,
  [2276] = 2276,
  [2277] = 2277,
  [2278] = 2278,
  [2279] = 2276,
  [2280] = 2277,
  [2281] = 2281,
  [2282] = 2282,
  [2283] = 2283,
  [2284] = 2284,
  [2285] = 2285,
  [2286] = 2286,
  [2287] = 2287,
  [2288] = 2288,
  [2289] = 2289,
  [2290] = 2290,
  [2291] = 2263,
  [2292] = 2292,
  [2293] = 2264,
  [2294] = 2278,
  [2295] = 2289,
  [2296] = 2265,
  [2297] = 2281,
  [2298] = 2282,
  [2299] = 2299,
  [2300] = 2287,
  [2301] = 2301,
  [2302] = 2283,
  [2303] = 2286,
  [2304] = 2292,
  [2305] = 2290,
  [2306] = 2299,
  [2307] = 2288,
  [2308] = 2258,
  [2309] = 2259,
  [2310] = 2221,
  [2311] = 2221,
  [2312] = 2312,
  [2313] = 2313,
  [2314] = 2314,
  [2315] = 2255,
  [2316] = 2316,
  [2317] = 2316,
  [2318] = 2313,
  [2319] = 2301,
  [2320] = 2314,
  [2321] = 2272,
  [2322] = 2256,
  [2323] = 2260,
  [2324] = 2262,
  [2325] = 2284,
  [2326] = 2285,
  [2327] = 2312,
  [2328] = 2328,
  [2329] = 2288,
  [2330] = 2299,
  [2331] = 2289,
  [2332] = 2290,
  [2333] = 2316,
  [2334] = 2281,
  [2335] = 2282,
  [2336] = 2221,
  [2337] = 2284,
  [2338] = 2285,
  [2339] = 2312,
  [2340] = 2301,
  [2341] = 2283,
  [2342] = 2277,
  [2343] = 2278,
  [2344] = 2313,
  [2345] = 2276,
  [2346] = 2328,
  [2347] = 2287,
  [2348] = 2286,
  [2349] = 2314,
  [2350] = 2292,
  [2351] = 2221,
  [2352] = 2352,
  [2353] = 2353,
  [2354] = 2354,
  [2355] = 2242,
  [2356] = 2236,
  [2357] = 2328,
  [2358] = 2234,
  [2359] = 2241,
  [2360] = 2352,
  [2361] = 2236,
  [2362] = 2234,
  [2363] = 2241,
  [2364] = 2364,
  [2365] = 2365,
  [2366] = 2366,
  [2367] = 2367,
  [2368] = 2368,
  [2369] = 2369,
  [2370] = 2370,
  [2371] = 2272,
  [2372] = 2372,
  [2373] = 2373,
  [2374] = 2256,
  [2375] = 2370,
  [2376] = 2367,
  [2377] = 2369,
  [2378] = 2373,
  [2379] = 2379,
  [2380] = 2272,
  [2381] = 2256,
  [2382] = 2225,
  [2383] = 2383,
  [2384] = 2383,
  [2385] = 2364,
  [2386] = 2353,
  [2387] = 2367,
  [2388] = 2366,
  [2389] = 2368,
  [2390] = 2379,
  [2391] = 2242,
  [2392] = 2240,
  [2393] = 2240,
  [2394] = 2367,
  [2395] = 2225,
  [2396] = 2372,
  [2397] = 2365,
  [2398] = 2354,
  [2399] = 2399,
  [2400] = 2400,
  [2401] = 2221,
  [2402] = 2242,
  [2403] = 2240,
  [2404] = 2225,
  [2405] = 2405,
  [2406] = 2242,
  [2407] = 2407,
  [2408] = 2240,
  [2409] = 2365,
  [2410] = 2410,
  [2411] = 2236,
  [2412] = 2240,
  [2413] = 2236,
  [2414] = 2225,
  [2415] = 2234,
  [2416] = 2416,
  [2417] = 2241,
  [2418] = 2242,
  [2419] = 2405,
  [2420] = 2420,
  [2421] = 2421,
  [2422] = 2422,
  [2423] = 2313,
  [2424] = 2424,
  [2425] = 2425,
  [2426] = 2367,
  [2427] = 2236,
  [2428] = 2234,
  [2429] = 2241,
  [2430] = 2234,
  [2431] = 2272,
  [2432] = 2256,
  [2433] = 2407,
  [2434] = 2400,
  [2435] = 2369,
  [2436] = 2416,
  [2437] = 2236,
  [2438] = 2438,
  [2439] = 2234,
  [2440] = 2410,
  [2441] = 2242,
  [2442] = 2442,
  [2443] = 2443,
  [2444] = 2241,
  [2445] = 2445,
  [2446] = 2446,
  [2447] = 2354,
  [2448] = 2448,
  [2449] = 2313,
  [2450] = 2442,
  [2451] = 2424,
  [2452] = 2399,
  [2453] = 2352,
  [2454] = 2420,
  [2455] = 2421,
  [2456] = 2373,
  [2457] = 2448,
  [2458] = 2364,
  [2459] = 2241,
  [2460] = 2425,
  [2461] = 2383,
  [2462] = 2366,
  [2463] = 2368,
  [2464] = 2370,
  [2465] = 2367,
  [2466] = 2379,
  [2467] = 2467,
  [2468] = 2438,
  [2469] = 2469,
  [2470] = 2240,
  [2471] = 2353,
  [2472] = 2472,
  [2473] = 2473,
  [2474] = 2443,
  [2475] = 2372,
  [2476] = 2367,
  [2477] = 2445,
  [2478] = 2446,
  [2479] = 2473,
  [2480] = 2467,
  [2481] = 2367,
  [2482] = 2469,
  [2483] = 2472,
  [2484] = 2422,
  [2485] = 2485,
  [2486] = 2264,
  [2487] = 2487,
  [2488] = 2488,
  [2489] = 2489,
  [2490] = 2490,
  [2491] = 2491,
  [2492] = 2492,
  [2493] = 2405,
  [2494] = 2265,
  [2495] = 2495,
  [2496] = 2496,
  [2497] = 2497,
  [2498] = 2498,
  [2499] = 2499,
  [2500] = 2487,
  [2501] = 2263,
  [2502] = 2502,
  [2503] = 2503,
  [2504] = 2504,
  [2505] = 2264,
  [2506] = 2506,
  [2507] = 2507,
  [2508] = 2508,
  [2509] = 2265,
  [2510] = 2510,
  [2511] = 2488,
  [2512] = 2512,
  [2513] = 2513,
  [2514] = 2514,
  [2515] = 2515,
  [2516] = 2516,
  [2517] = 2517,
  [2518] = 2518,
  [2519] = 2519,
  [2520] = 2520,
  [2521] = 2521,
  [2522] = 2522,
  [2523] = 2523,
  [2524] = 2524,
  [2525] = 2525,
  [2526] = 2489,
  [2527] = 2527,
  [2528] = 2528,
  [2529] = 2529,
  [2530] = 2530,
  [2531] = 2531,
  [2532] = 2490,
  [2533] = 2533,
  [2534] = 2534,
  [2535] = 2535,
  [2536] = 2536,
  [2537] = 2537,
  [2538] = 2538,
  [2539] = 2539,
  [2540] = 2540,
  [2541] = 2491,
  [2542] = 2542,
  [2543] = 2485,
  [2544] = 2544,
  [2545] = 2545,
  [2546] = 2492,
  [2547] = 2547,
  [2548] = 2495,
  [2549] = 2260,
  [2550] = 2550,
  [2551] = 2551,
  [2552] = 2552,
  [2553] = 2496,
  [2554] = 2497,
  [2555] = 2555,
  [2556] = 2556,
  [2557] = 2498,
  [2558] = 2558,
  [2559] = 2262,
  [2560] = 2560,
  [2561] = 2561,
  [2562] = 2562,
  [2563] = 2563,
  [2564] = 2503,
  [2565] = 2504,
  [2566] = 2525,
  [2567] = 2530,
  [2568] = 2535,
  [2569] = 2536,
  [2570] = 2540,
  [2571] = 2542,
  [2572] = 2572,
  [2573] = 2573,
  [2574] = 2574,
  [2575] = 2575,
  [2576] = 2572,
  [2577] = 2573,
  [2578] = 2574,
  [2579] = 2575,
  [2580] = 2537,
  [2581] = 2399,
  [2582] = 2582,
  [2583] = 2583,
  [2584] = 2258,
  [2585] = 2259,
  [2586] = 2258,
  [2587] = 2259,
  [2588] = 2588,
  [2589] = 2589,
  [2590] = 2221,
  [2591] = 2255,
  [2592] = 2367,
  [2593] = 2255,
  [2594] = 2499,
  [2595] = 2502,
  [2596] = 2539,
  [2597] = 2469,
  [2598] = 2544,
  [2599] = 2472,
  [2600] = 2473,
  [2601] = 2545,
  [2602] = 2506,
  [2603] = 2507,
  [2604] = 2508,
  [2605] = 2510,
  [2606] = 2512,
  [2607] = 2272,
  [2608] = 2443,
  [2609] = 2547,
  [2610] = 2513,
  [2611] = 2256,
  [2612] = 2550,
  [2613] = 2225,
  [2614] = 2514,
  [2615] = 2582,
  [2616] = 2583,
  [2617] = 2515,
  [2618] = 2588,
  [2619] = 2272,
  [2620] = 2256,
  [2621] = 2516,
  [2622] = 2517,
  [2623] = 2518,
  [2624] = 2519,
  [2625] = 2520,
  [2626] = 2589,
  [2627] = 2551,
  [2628] = 2521,
  [2629] = 2522,
  [2630] = 2523,
  [2631] = 2313,
  [2632] = 2524,
  [2633] = 2527,
  [2634] = 2552,
  [2635] = 2528,
  [2636] = 2529,
  [2637] = 2531,
  [2638] = 2260,
  [2639] = 2533,
  [2640] = 2534,
  [2641] = 2538,
  [2642] = 2407,
  [2643] = 2555,
  [2644] = 2416,
  [2645] = 2420,
  [2646] = 2421,
  [2647] = 2556,
  [2648] = 2400,
  [2649] = 2410,
  [2650] = 2262,
  [2651] = 2448,
  [2652] = 2442,
  [2653] = 2558,
  [2654] = 2425,
  [2655] = 2438,
  [2656] = 2560,
  [2657] = 2424,
  [2658] = 2561,
  [2659] = 2422,
  [2660] = 2562,
  [2661] = 2563,
  [2662] = 2422,
  [2663] = 2405,
  [2664] = 2263,
  [2665] = 2445,
  [2666] = 2446,
  [2667] = 2467,
  [2668] = 2276,
  [2669] = 2278,
  [2670] = 2265,
  [2671] = 2316,
  [2672] = 2287,
  [2673] = 2537,
  [2674] = 2539,
  [2675] = 2544,
  [2676] = 2545,
  [2677] = 2547,
  [2678] = 2550,
  [2679] = 2551,
  [2680] = 2552,
  [2681] = 2555,
  [2682] = 2556,
  [2683] = 2558,
  [2684] = 2560,
  [2685] = 2561,
  [2686] = 2562,
  [2687] = 2563,
  [2688] = 2503,
  [2689] = 2504,
  [2690] = 2525,
  [2691] = 2530,
  [2692] = 2535,
  [2693] = 2536,
  [2694] = 2540,
  [2695] = 2258,
  [2696] = 2286,
  [2697] = 2292,
  [2698] = 2284,
  [2699] = 2299,
  [2700] = 2287,
  [2701] = 2286,
  [2702] = 2292,
  [2703] = 2299,
  [2704] = 2542,
  [2705] = 2572,
  [2706] = 2288,
  [2707] = 2258,
  [2708] = 2259,
  [2709] = 2258,
  [2710] = 2259,
  [2711] = 2573,
  [2712] = 2574,
  [2713] = 2575,
  [2714] = 2259,
  [2715] = 2263,
  [2716] = 2258,
  [2717] = 2259,
  [2718] = 2255,
  [2719] = 2277,
  [2720] = 2314,
  [2721] = 2255,
  [2722] = 2582,
  [2723] = 2255,
  [2724] = 2289,
  [2725] = 2583,
  [2726] = 2264,
  [2727] = 2260,
  [2728] = 2588,
  [2729] = 2589,
  [2730] = 2290,
  [2731] = 2265,
  [2732] = 2282,
  [2733] = 2313,
  [2734] = 2312,
  [2735] = 2255,
  [2736] = 2301,
  [2737] = 2278,
  [2738] = 2276,
  [2739] = 2313,
  [2740] = 2283,
  [2741] = 2263,
  [2742] = 2312,
  [2743] = 2272,
  [2744] = 2256,
  [2745] = 2272,
  [2746] = 2256,
  [2747] = 2262,
  [2748] = 2301,
  [2749] = 2283,
  [2750] = 2285,
  [2751] = 2281,
  [2752] = 2272,
  [2753] = 2256,
  [2754] = 2221,
  [2755] = 2495,
  [2756] = 2496,
  [2757] = 2497,
  [2758] = 2272,
  [2759] = 2262,
  [2760] = 2498,
  [2761] = 2499,
  [2762] = 2502,
  [2763] = 2256,
  [2764] = 2221,
  [2765] = 2264,
  [2766] = 2282,
  [2767] = 2263,
  [2768] = 2260,
  [2769] = 2506,
  [2770] = 2507,
  [2771] = 2508,
  [2772] = 2510,
  [2773] = 2512,
  [2774] = 2288,
  [2775] = 2284,
  [2776] = 2513,
  [2777] = 2514,
  [2778] = 2515,
  [2779] = 2516,
  [2780] = 2517,
  [2781] = 2518,
  [2782] = 2519,
  [2783] = 2520,
  [2784] = 2521,
  [2785] = 2522,
  [2786] = 2523,
  [2787] = 2524,
  [2788] = 2527,
  [2789] = 2528,
  [2790] = 2529,
  [2791] = 2260,
  [2792] = 2531,
  [2793] = 2533,
  [2794] = 2534,
  [2795] = 2262,
  [2796] = 2264,
  [2797] = 2538,
  [2798] = 2265,
  [2799] = 2262,
  [2800] = 2260,
  [2801] = 2281,
  [2802] = 2289,
  [2803] = 2285,
  [2804] = 2487,
  [2805] = 2265,
  [2806] = 2488,
  [2807] = 2290,
  [2808] = 2263,
  [2809] = 2489,
  [2810] = 2490,
  [2811] = 2491,
  [2812] = 2492,
  [2813] = 2316,
  [2814] = 2314,
  [2815] = 2264,
  [2816] = 2277,
  [2817] = 2485,
  [2818] = 2288,
  [2819] = 2301,
  [2820] = 2313,
  [2821] = 2312,
  [2822] = 2221,
  [2823] = 2281,
  [2824] = 2312,
  [2825] = 2301,
  [2826] = 2299,
  [2827] = 2283,
  [2828] = 2301,
  [2829] = 2276,
  [2830] = 2314,
  [2831] = 2283,
  [2832] = 2286,
  [2833] = 2281,
  [2834] = 2282,
  [2835] = 2313,
  [2836] = 2282,
  [2837] = 2314,
  [2838] = 2328,
  [2839] = 2316,
  [2840] = 2283,
  [2841] = 2290,
  [2842] = 2314,
  [2843] = 2290,
  [2844] = 2284,
  [2845] = 2289,
  [2846] = 2286,
  [2847] = 2292,
  [2848] = 2285,
  [2849] = 2221,
  [2850] = 2277,
  [2851] = 2285,
  [2852] = 2278,
  [2853] = 2221,
  [2854] = 2299,
  [2855] = 2287,
  [2856] = 2287,
  [2857] = 2286,
  [2858] = 2276,
  [2859] = 2284,
  [2860] = 2285,
  [2861] = 2312,
  [2862] = 2292,
  [2863] = 2316,
  [2864] = 2292,
  [2865] = 2292,
  [2866] = 2276,
  [2867] = 2281,
  [2868] = 2282,
  [2869] = 2283,
  [2870] = 2289,
  [2871] = 2312,
  [2872] = 2288,
  [2873] = 2284,
  [2874] = 2285,
  [2875] = 2314,
  [2876] = 2287,
  [2877] = 2288,
  [2878] = 2328,
  [2879] = 2289,
  [2880] = 2290,
  [2881] = 2290,
  [2882] = 2288,
  [2883] = 2313,
  [2884] = 2299,
  [2885] = 2301,
  [2886] = 2316,
  [2887] = 2277,
  [2888] = 2313,
  [2889] = 2277,
  [2890] = 2278,
  [2891] = 2221,
  [2892] = 2316,
  [2893] = 2289,
  [2894] = 2278,
  [2895] = 2276,
  [2896] = 2277,
  [2897] = 2221,
  [2898] = 2221,
  [2899] = 2225,
  [2900] = 2287,
  [2901] = 2284,
  [2902] = 2221,
  [2903] = 2281,
  [2904] = 2282,
  [2905] = 2286,
  [2906] = 2278,
  [2907] = 2299,
  [2908] = 2365,
  [2909] = 2368,
  [2910] = 2353,
  [2911] = 2367,
  [2912] = 2369,
  [2913] = 2328,
  [2914] = 2328,
  [2915] = 2221,
  [2916] = 2370,
  [2917] = 2383,
  [2918] = 2328,
  [2919] = 2367,
  [2920] = 2373,
  [2921] = 2367,
  [2922] = 2221,
  [2923] = 2236,
  [2924] = 2370,
  [2925] = 2242,
  [2926] = 2364,
  [2927] = 2353,
  [2928] = 2379,
  [2929] = 2352,
  [2930] = 2234,
  [2931] = 2221,
  [2932] = 2221,
  [2933] = 2369,
  [2934] = 2272,
  [2935] = 2367,
  [2936] = 2368,
  [2937] = 2221,
  [2938] = 2354,
  [2939] = 2221,
  [2940] = 2354,
  [2941] = 2379,
  [2942] = 2366,
  [2943] = 2372,
  [2944] = 2240,
  [2945] = 2256,
  [2946] = 2272,
  [2947] = 2236,
  [2948] = 2234,
  [2949] = 2221,
  [2950] = 2225,
  [2951] = 2256,
  [2952] = 2372,
  [2953] = 2373,
  [2954] = 2365,
  [2955] = 2221,
  [2956] = 2366,
  [2957] = 2383,
  [2958] = 2225,
  [2959] = 2242,
  [2960] = 2221,
  [2961] = 2352,
  [2962] = 2241,
  [2963] = 2364,
  [2964] = 2240,
  [2965] = 2241,
  [2966] = 2442,
  [2967] = 2234,
  [2968] = 2241,
  [2969] = 2354,
  [2970] = 2372,
  [2971] = 2407,
  [2972] = 2448,
  [2973] = 2367,
  [2974] = 2365,
  [2975] = 2364,
  [2976] = 2367,
  [2977] = 2354,
  [2978] = 2370,
  [2979] = 2379,
  [2980] = 2425,
  [2981] = 2467,
  [2982] = 2352,
  [2983] = 2469,
  [2984] = 2438,
  [2985] = 2352,
  [2986] = 2472,
  [2987] = 2352,
  [2988] = 2367,
  [2989] = 2364,
  [2990] = 2364,
  [2991] = 2367,
  [2992] = 2422,
  [2993] = 2366,
  [2994] = 2368,
  [2995] = 2366,
  [2996] = 2370,
  [2997] = 2368,
  [2998] = 2379,
  [2999] = 2354,
  [3000] = 2473,
  [3001] = 2370,
  [3002] = 2379,
  [3003] = 2443,
  [3004] = 2467,
  [3005] = 2367,
  [3006] = 2469,
  [3007] = 2352,
  [3008] = 2472,
  [3009] = 2225,
  [3010] = 2473,
  [3011] = 2443,
  [3012] = 2369,
  [3013] = 2367,
  [3014] = 2373,
  [3015] = 2367,
  [3016] = 2383,
  [3017] = 2353,
  [3018] = 2242,
  [3019] = 2367,
  [3020] = 2369,
  [3021] = 2373,
  [3022] = 2383,
  [3023] = 2353,
  [3024] = 2313,
  [3025] = 2313,
  [3026] = 2328,
  [3027] = 2256,
  [3028] = 2272,
  [3029] = 2256,
  [3030] = 2424,
  [3031] = 2272,
  [3032] = 2256,
  [3033] = 2405,
  [3034] = 2407,
  [3035] = 2416,
  [3036] = 2420,
  [3037] = 2421,
  [3038] = 2416,
  [3039] = 2364,
  [3040] = 2372,
  [3041] = 2369,
  [3042] = 2399,
  [3043] = 2225,
  [3044] = 2420,
  [3045] = 2421,
  [3046] = 2365,
  [3047] = 2372,
  [3048] = 2445,
  [3049] = 2446,
  [3050] = 2373,
  [3051] = 2225,
  [3052] = 2367,
  [3053] = 2365,
  [3054] = 2366,
  [3055] = 2368,
  [3056] = 2383,
  [3057] = 2445,
  [3058] = 2353,
  [3059] = 2448,
  [3060] = 2442,
  [3061] = 2400,
  [3062] = 2366,
  [3063] = 2368,
  [3064] = 2400,
  [3065] = 2425,
  [3066] = 2399,
  [3067] = 2446,
  [3068] = 2370,
  [3069] = 2379,
  [3070] = 2438,
  [3071] = 2410,
  [3072] = 2372,
  [3073] = 2410,
  [3074] = 2272,
  [3075] = 2256,
  [3076] = 2225,
  [3077] = 2424,
  [3078] = 2354,
  [3079] = 2422,
  [3080] = 2240,
  [3081] = 2405,
  [3082] = 2225,
  [3083] = 2236,
  [3084] = 2272,
  [3085] = 2507,
  [3086] = 2272,
  [3087] = 2445,
  [3088] = 2256,
  [3089] = 2446,
  [3090] = 2535,
  [3091] = 2536,
  [3092] = 2540,
  [3093] = 2260,
  [3094] = 2542,
  [3095] = 2420,
  [3096] = 2572,
  [3097] = 2448,
  [3098] = 2573,
  [3099] = 2442,
  [3100] = 2574,
  [3101] = 2259,
  [3102] = 2421,
  [3103] = 2575,
  [3104] = 2225,
  [3105] = 2263,
  [3106] = 2446,
  [3107] = 2400,
  [3108] = 2443,
  [3109] = 2399,
  [3110] = 2467,
  [3111] = 2425,
  [3112] = 2236,
  [3113] = 2582,
  [3114] = 2469,
  [3115] = 2264,
  [3116] = 2583,
  [3117] = 2425,
  [3118] = 2260,
  [3119] = 2472,
  [3120] = 2262,
  [3121] = 2473,
  [3122] = 2443,
  [3123] = 2588,
  [3124] = 2589,
  [3125] = 2265,
  [3126] = 2410,
  [3127] = 2438,
  [3128] = 2225,
  [3129] = 2410,
  [3130] = 2424,
  [3131] = 2399,
  [3132] = 2234,
  [3133] = 2502,
  [3134] = 2537,
  [3135] = 2539,
  [3136] = 2485,
  [3137] = 2582,
  [3138] = 2583,
  [3139] = 2544,
  [3140] = 2588,
  [3141] = 2589,
  [3142] = 2545,
  [3143] = 2547,
  [3144] = 2550,
  [3145] = 2400,
  [3146] = 2551,
  [3147] = 2367,
  [3148] = 2367,
  [3149] = 2367,
  [3150] = 2552,
  [3151] = 2272,
  [3152] = 2467,
  [3153] = 2555,
  [3154] = 2556,
  [3155] = 2438,
  [3156] = 2445,
  [3157] = 2446,
  [3158] = 2558,
  [3159] = 2560,
  [3160] = 2561,
  [3161] = 2313,
  [3162] = 2562,
  [3163] = 2563,
  [3164] = 2313,
  [3165] = 2405,
  [3166] = 2487,
  [3167] = 2421,
  [3168] = 2488,
  [3169] = 2503,
  [3170] = 2313,
  [3171] = 2489,
  [3172] = 2241,
  [3173] = 2490,
  [3174] = 2491,
  [3175] = 2225,
  [3176] = 2492,
  [3177] = 2262,
  [3178] = 2469,
  [3179] = 2504,
  [3180] = 2495,
  [3181] = 2496,
  [3182] = 2497,
  [3183] = 2498,
  [3184] = 2499,
  [3185] = 2502,
  [3186] = 2525,
  [3187] = 2530,
  [3188] = 2506,
  [3189] = 2472,
  [3190] = 2508,
  [3191] = 2510,
  [3192] = 2512,
  [3193] = 2535,
  [3194] = 2225,
  [3195] = 2536,
  [3196] = 2513,
  [3197] = 2514,
  [3198] = 2367,
  [3199] = 2515,
  [3200] = 2540,
  [3201] = 2506,
  [3202] = 2516,
  [3203] = 2416,
  [3204] = 2517,
  [3205] = 2518,
  [3206] = 2507,
  [3207] = 2519,
  [3208] = 2520,
  [3209] = 2399,
  [3210] = 2521,
  [3211] = 2522,
  [3212] = 2508,
  [3213] = 2510,
  [3214] = 2512,
  [3215] = 2472,
  [3216] = 2473,
  [3217] = 2443,
  [3218] = 2410,
  [3219] = 2542,
  [3220] = 2467,
  [3221] = 2572,
  [3222] = 2469,
  [3223] = 2573,
  [3224] = 2256,
  [3225] = 2574,
  [3226] = 2513,
  [3227] = 2472,
  [3228] = 2473,
  [3229] = 2473,
  [3230] = 2443,
  [3231] = 2225,
  [3232] = 2537,
  [3233] = 2255,
  [3234] = 2539,
  [3235] = 2485,
  [3236] = 2514,
  [3237] = 2515,
  [3238] = 2516,
  [3239] = 2544,
  [3240] = 2399,
  [3241] = 2517,
  [3242] = 2410,
  [3243] = 2518,
  [3244] = 2519,
  [3245] = 2520,
  [3246] = 2400,
  [3247] = 2545,
  [3248] = 2547,
  [3249] = 2550,
  [3250] = 2407,
  [3251] = 2551,
  [3252] = 2467,
  [3253] = 2552,
  [3254] = 2523,
  [3255] = 2555,
  [3256] = 2448,
  [3257] = 2524,
  [3258] = 2521,
  [3259] = 2369,
  [3260] = 2422,
  [3261] = 2522,
  [3262] = 2523,
  [3263] = 2400,
  [3264] = 2556,
  [3265] = 2524,
  [3266] = 2558,
  [3267] = 2527,
  [3268] = 2499,
  [3269] = 2528,
  [3270] = 2560,
  [3271] = 2527,
  [3272] = 2528,
  [3273] = 2442,
  [3274] = 2422,
  [3275] = 2416,
  [3276] = 2242,
  [3277] = 2424,
  [3278] = 2529,
  [3279] = 2529,
  [3280] = 2531,
  [3281] = 2561,
  [3282] = 2531,
  [3283] = 2448,
  [3284] = 2442,
  [3285] = 2562,
  [3286] = 2533,
  [3287] = 2533,
  [3288] = 2534,
  [3289] = 2365,
  [3290] = 2425,
  [3291] = 2534,
  [3292] = 2496,
  [3293] = 2353,
  [3294] = 2538,
  [3295] = 2225,
  [3296] = 2438,
  [3297] = 2367,
  [3298] = 2424,
  [3299] = 2445,
  [3300] = 2263,
  [3301] = 2448,
  [3302] = 2495,
  [3303] = 2424,
  [3304] = 2373,
  [3305] = 2538,
  [3306] = 2264,
  [3307] = 2487,
  [3308] = 2265,
  [3309] = 2405,
  [3310] = 2240,
  [3311] = 2258,
  [3312] = 2442,
  [3313] = 2405,
  [3314] = 2313,
  [3315] = 2563,
  [3316] = 2497,
  [3317] = 2225,
  [3318] = 2407,
  [3319] = 2503,
  [3320] = 2422,
  [3321] = 2416,
  [3322] = 2488,
  [3323] = 2420,
  [3324] = 2421,
  [3325] = 2498,
  [3326] = 2425,
  [3327] = 2258,
  [3328] = 2489,
  [3329] = 2490,
  [3330] = 2259,
  [3331] = 2491,
  [3332] = 2407,
  [3333] = 2504,
  [3334] = 2420,
  [3335] = 2492,
  [3336] = 2525,
  [3337] = 2416,
  [3338] = 2530,
  [3339] = 2445,
  [3340] = 2407,
  [3341] = 2255,
  [3342] = 2421,
  [3343] = 2469,
  [3344] = 2438,
  [3345] = 2383,
  [3346] = 2446,
  [3347] = 2420,
  [3348] = 2575,
  [3349] = 2289,
  [3350] = 2281,
  [3351] = 2282,
  [3352] = 2528,
  [3353] = 2422,
  [3354] = 2284,
  [3355] = 2550,
  [3356] = 2529,
  [3357] = 2551,
  [3358] = 2552,
  [3359] = 2555,
  [3360] = 2556,
  [3361] = 2582,
  [3362] = 2583,
  [3363] = 2558,
  [3364] = 2531,
  [3365] = 2560,
  [3366] = 2561,
  [3367] = 2562,
  [3368] = 2529,
  [3369] = 2513,
  [3370] = 2563,
  [3371] = 2503,
  [3372] = 2314,
  [3373] = 2504,
  [3374] = 2525,
  [3375] = 2530,
  [3376] = 2531,
  [3377] = 2524,
  [3378] = 2535,
  [3379] = 2536,
  [3380] = 2540,
  [3381] = 2542,
  [3382] = 2533,
  [3383] = 2534,
  [3384] = 2285,
  [3385] = 2542,
  [3386] = 2572,
  [3387] = 2492,
  [3388] = 2573,
  [3389] = 2574,
  [3390] = 2287,
  [3391] = 2260,
  [3392] = 2521,
  [3393] = 2522,
  [3394] = 2575,
  [3395] = 2262,
  [3396] = 2533,
  [3397] = 2288,
  [3398] = 2534,
  [3399] = 2533,
  [3400] = 2495,
  [3401] = 2523,
  [3402] = 2538,
  [3403] = 2496,
  [3404] = 2497,
  [3405] = 2498,
  [3406] = 2513,
  [3407] = 2537,
  [3408] = 2539,
  [3409] = 2485,
  [3410] = 2289,
  [3411] = 2544,
  [3412] = 2545,
  [3413] = 2514,
  [3414] = 2547,
  [3415] = 2582,
  [3416] = 2290,
  [3417] = 2550,
  [3418] = 2583,
  [3419] = 2551,
  [3420] = 2552,
  [3421] = 2555,
  [3422] = 2556,
  [3423] = 2558,
  [3424] = 2560,
  [3425] = 2561,
  [3426] = 2562,
  [3427] = 2563,
  [3428] = 2503,
  [3429] = 2504,
  [3430] = 2525,
  [3431] = 2530,
  [3432] = 2524,
  [3433] = 2516,
  [3434] = 2535,
  [3435] = 2536,
  [3436] = 2540,
  [3437] = 2542,
  [3438] = 2588,
  [3439] = 2572,
  [3440] = 2589,
  [3441] = 2573,
  [3442] = 2574,
  [3443] = 2575,
  [3444] = 2538,
  [3445] = 2588,
  [3446] = 2589,
  [3447] = 2582,
  [3448] = 2583,
  [3449] = 2588,
  [3450] = 2589,
  [3451] = 2313,
  [3452] = 2531,
  [3453] = 2272,
  [3454] = 2256,
  [3455] = 2506,
  [3456] = 2527,
  [3457] = 2499,
  [3458] = 2502,
  [3459] = 2528,
  [3460] = 2507,
  [3461] = 2517,
  [3462] = 2518,
  [3463] = 2508,
  [3464] = 2515,
  [3465] = 2516,
  [3466] = 2517,
  [3467] = 2529,
  [3468] = 2531,
  [3469] = 2518,
  [3470] = 2537,
  [3471] = 2539,
  [3472] = 2286,
  [3473] = 2485,
  [3474] = 2544,
  [3475] = 2545,
  [3476] = 2547,
  [3477] = 2550,
  [3478] = 2551,
  [3479] = 2299,
  [3480] = 2552,
  [3481] = 2312,
  [3482] = 2555,
  [3483] = 2556,
  [3484] = 2558,
  [3485] = 2301,
  [3486] = 2534,
  [3487] = 2514,
  [3488] = 2283,
  [3489] = 2560,
  [3490] = 2561,
  [3491] = 2255,
  [3492] = 2562,
  [3493] = 2533,
  [3494] = 2563,
  [3495] = 2534,
  [3496] = 2506,
  [3497] = 2523,
  [3498] = 2503,
  [3499] = 2519,
  [3500] = 2507,
  [3501] = 2575,
  [3502] = 2504,
  [3503] = 2520,
  [3504] = 2525,
  [3505] = 2508,
  [3506] = 2519,
  [3507] = 2520,
  [3508] = 2510,
  [3509] = 2512,
  [3510] = 2530,
  [3511] = 2535,
  [3512] = 2536,
  [3513] = 2540,
  [3514] = 2542,
  [3515] = 2572,
  [3516] = 2573,
  [3517] = 2574,
  [3518] = 2521,
  [3519] = 2538,
  [3520] = 2522,
  [3521] = 2575,
  [3522] = 2487,
  [3523] = 2258,
  [3524] = 2513,
  [3525] = 2521,
  [3526] = 2514,
  [3527] = 2313,
  [3528] = 2515,
  [3529] = 2516,
  [3530] = 2517,
  [3531] = 2518,
  [3532] = 2519,
  [3533] = 2520,
  [3534] = 2488,
  [3535] = 2573,
  [3536] = 2405,
  [3537] = 2523,
  [3538] = 2574,
  [3539] = 2527,
  [3540] = 2582,
  [3541] = 2583,
  [3542] = 2588,
  [3543] = 2589,
  [3544] = 2314,
  [3545] = 2487,
  [3546] = 2538,
  [3547] = 2572,
  [3548] = 2527,
  [3549] = 2488,
  [3550] = 2259,
  [3551] = 2489,
  [3552] = 2490,
  [3553] = 2521,
  [3554] = 2489,
  [3555] = 2490,
  [3556] = 2522,
  [3557] = 2528,
  [3558] = 2491,
  [3559] = 2522,
  [3560] = 2316,
  [3561] = 2277,
  [3562] = 2278,
  [3563] = 2487,
  [3564] = 2523,
  [3565] = 2510,
  [3566] = 2488,
  [3567] = 2316,
  [3568] = 2487,
  [3569] = 2489,
  [3570] = 2492,
  [3571] = 2277,
  [3572] = 2278,
  [3573] = 2488,
  [3574] = 2276,
  [3575] = 2524,
  [3576] = 2490,
  [3577] = 2491,
  [3578] = 2281,
  [3579] = 2489,
  [3580] = 2490,
  [3581] = 2282,
  [3582] = 2284,
  [3583] = 2285,
  [3584] = 2491,
  [3585] = 2537,
  [3586] = 2512,
  [3587] = 2492,
  [3588] = 2287,
  [3589] = 2292,
  [3590] = 2539,
  [3591] = 2537,
  [3592] = 2495,
  [3593] = 2496,
  [3594] = 2492,
  [3595] = 2497,
  [3596] = 2498,
  [3597] = 2539,
  [3598] = 2288,
  [3599] = 2520,
  [3600] = 2290,
  [3601] = 2499,
  [3602] = 2502,
  [3603] = 2495,
  [3604] = 2529,
  [3605] = 2496,
  [3606] = 2495,
  [3607] = 2496,
  [3608] = 2497,
  [3609] = 2498,
  [3610] = 2497,
  [3611] = 2527,
  [3612] = 2506,
  [3613] = 2507,
  [3614] = 2312,
  [3615] = 2499,
  [3616] = 2485,
  [3617] = 2508,
  [3618] = 2510,
  [3619] = 2301,
  [3620] = 2283,
  [3621] = 2512,
  [3622] = 2544,
  [3623] = 2286,
  [3624] = 2292,
  [3625] = 2498,
  [3626] = 2513,
  [3627] = 2514,
  [3628] = 2545,
  [3629] = 2499,
  [3630] = 2515,
  [3631] = 2502,
  [3632] = 2516,
  [3633] = 2517,
  [3634] = 2518,
  [3635] = 2519,
  [3636] = 2515,
  [3637] = 2299,
  [3638] = 2485,
  [3639] = 2544,
  [3640] = 2547,
  [3641] = 2550,
  [3642] = 2491,
  [3643] = 2551,
  [3644] = 2552,
  [3645] = 2555,
  [3646] = 2556,
  [3647] = 2558,
  [3648] = 2560,
  [3649] = 2561,
  [3650] = 2263,
  [3651] = 2562,
  [3652] = 2563,
  [3653] = 2503,
  [3654] = 2502,
  [3655] = 2504,
  [3656] = 2525,
  [3657] = 2530,
  [3658] = 2535,
  [3659] = 2536,
  [3660] = 2528,
  [3661] = 2264,
  [3662] = 2545,
  [3663] = 2540,
  [3664] = 2506,
  [3665] = 2507,
  [3666] = 2276,
  [3667] = 2547,
  [3668] = 2265,
  [3669] = 2508,
  [3670] = 2510,
  [3671] = 2512,
  [3672] = 2524,
  [3673] = 2242,
  [3674] = 2313,
  [3675] = 2236,
  [3676] = 2292,
  [3677] = 2263,
  [3678] = 2234,
  [3679] = 2241,
  [3680] = 2283,
  [3681] = 2328,
  [3682] = 2301,
  [3683] = 2255,
  [3684] = 2262,
  [3685] = 2299,
  [3686] = 2259,
  [3687] = 2312,
  [3688] = 2258,
  [3689] = 2264,
  [3690] = 2277,
  [3691] = 2240,
  [3692] = 2278,
  [3693] = 2260,
  [3694] = 2276,
  [3695] = 2314,
  [3696] = 2281,
  [3697] = 2282,
  [3698] = 2284,
  [3699] = 2328,
  [3700] = 2272,
  [3701] = 2286,
  [3702] = 2285,
  [3703] = 2287,
  [3704] = 2288,
  [3705] = 2289,
  [3706] = 2290,
  [3707] = 2265,
  [3708] = 2256,
  [3709] = 2316,
  [3710] = 2299,
  [3711] = 2370,
  [3712] = 2240,
  [3713] = 2286,
  [3714] = 2242,
  [3715] = 2365,
  [3716] = 2352,
  [3717] = 2369,
  [3718] = 2367,
  [3719] = 2272,
  [3720] = 2379,
  [3721] = 2277,
  [3722] = 2278,
  [3723] = 2383,
  [3724] = 2352,
  [3725] = 2365,
  [3726] = 2256,
  [3727] = 2372,
  [3728] = 2353,
  [3729] = 2288,
  [3730] = 2282,
  [3731] = 2370,
  [3732] = 2285,
  [3733] = 2367,
  [3734] = 2366,
  [3735] = 2272,
  [3736] = 2373,
  [3737] = 2368,
  [3738] = 2316,
  [3739] = 2276,
  [3740] = 2379,
  [3741] = 2234,
  [3742] = 2366,
  [3743] = 2367,
  [3744] = 2312,
  [3745] = 2367,
  [3746] = 2314,
  [3747] = 2301,
  [3748] = 2283,
  [3749] = 2372,
  [3750] = 2281,
  [3751] = 2287,
  [3752] = 2354,
  [3753] = 2289,
  [3754] = 2354,
  [3755] = 2373,
  [3756] = 2368,
  [3757] = 2290,
  [3758] = 2383,
  [3759] = 2284,
  [3760] = 2241,
  [3761] = 2364,
  [3762] = 2313,
  [3763] = 2353,
  [3764] = 2236,
  [3765] = 2328,
  [3766] = 2292,
  [3767] = 2364,
  [3768] = 2369,
  [3769] = 2256,
  [3770] = 2442,
  [3771] = 2383,
  [3772] = 2445,
  [3773] = 2399,
  [3774] = 2353,
  [3775] = 2242,
  [3776] = 2448,
  [3777] = 2472,
  [3778] = 2410,
  [3779] = 2263,
  [3780] = 2425,
  [3781] = 2242,
  [3782] = 2445,
  [3783] = 2400,
  [3784] = 2313,
  [3785] = 2438,
  [3786] = 2264,
  [3787] = 2265,
  [3788] = 2469,
  [3789] = 2446,
  [3790] = 3790,
  [3791] = 2272,
  [3792] = 2473,
  [3793] = 2469,
  [3794] = 2240,
  [3795] = 2420,
  [3796] = 2236,
  [3797] = 2367,
  [3798] = 2365,
  [3799] = 2234,
  [3800] = 2241,
  [3801] = 2364,
  [3802] = 2416,
  [3803] = 2421,
  [3804] = 2410,
  [3805] = 2473,
  [3806] = 2472,
  [3807] = 2420,
  [3808] = 2407,
  [3809] = 2421,
  [3810] = 2399,
  [3811] = 2405,
  [3812] = 2272,
  [3813] = 2400,
  [3814] = 2422,
  [3815] = 2260,
  [3816] = 2407,
  [3817] = 2416,
  [3818] = 2328,
  [3819] = 2443,
  [3820] = 2262,
  [3821] = 2366,
  [3822] = 2241,
  [3823] = 2236,
  [3824] = 2234,
  [3825] = 2258,
  [3826] = 2259,
  [3827] = 2240,
  [3828] = 2354,
  [3829] = 2255,
  [3830] = 2367,
  [3831] = 2256,
  [3832] = 2448,
  [3833] = 2424,
  [3834] = 2442,
  [3835] = 2367,
  [3836] = 2240,
  [3837] = 2236,
  [3838] = 2241,
  [3839] = 2234,
  [3840] = 2372,
  [3841] = 2241,
  [3842] = 2425,
  [3843] = 3843,
  [3844] = 2313,
  [3845] = 2467,
  [3846] = 2422,
  [3847] = 2240,
  [3848] = 2367,
  [3849] = 2467,
  [3850] = 2370,
  [3851] = 2424,
  [3852] = 2443,
  [3853] = 3853,
  [3854] = 2438,
  [3855] = 2379,
  [3856] = 2242,
  [3857] = 2242,
  [3858] = 2446,
  [3859] = 2352,
  [3860] = 2369,
  [3861] = 2256,
  [3862] = 3843,
  [3863] = 2373,
  [3864] = 2405,
  [3865] = 2236,
  [3866] = 2234,
  [3867] = 2368,
  [3868] = 2234,
  [3869] = 2530,
  [3870] = 2369,
  [3871] = 2366,
  [3872] = 2512,
  [3873] = 2527,
  [3874] = 2535,
  [3875] = 2497,
  [3876] = 2472,
  [3877] = 2236,
  [3878] = 2240,
  [3879] = 2556,
  [3880] = 2498,
  [3881] = 2512,
  [3882] = 2242,
  [3883] = 2528,
  [3884] = 2485,
  [3885] = 2234,
  [3886] = 2241,
  [3887] = 2536,
  [3888] = 2562,
  [3889] = 2405,
  [3890] = 2370,
  [3891] = 2242,
  [3892] = 2379,
  [3893] = 2242,
  [3894] = 2547,
  [3895] = 2540,
  [3896] = 2542,
  [3897] = 2489,
  [3898] = 2242,
  [3899] = 3853,
  [3900] = 2236,
  [3901] = 2234,
  [3902] = 2572,
  [3903] = 2573,
  [3904] = 2574,
  [3905] = 2492,
  [3906] = 2540,
  [3907] = 2241,
  [3908] = 2529,
  [3909] = 2531,
  [3910] = 2312,
  [3911] = 2550,
  [3912] = 2535,
  [3913] = 2533,
  [3914] = 2301,
  [3915] = 2283,
  [3916] = 2242,
  [3917] = 2498,
  [3918] = 2473,
  [3919] = 2537,
  [3920] = 3790,
  [3921] = 2263,
  [3922] = 2284,
  [3923] = 2521,
  [3924] = 2558,
  [3925] = 2551,
  [3926] = 2534,
  [3927] = 2560,
  [3928] = 2523,
  [3929] = 2563,
  [3930] = 2373,
  [3931] = 2503,
  [3932] = 2372,
  [3933] = 3933,
  [3934] = 3933,
  [3935] = 2264,
  [3936] = 2506,
  [3937] = 2504,
  [3938] = 2547,
  [3939] = 2538,
  [3940] = 2424,
  [3941] = 2544,
  [3942] = 2539,
  [3943] = 2510,
  [3944] = 2368,
  [3945] = 2518,
  [3946] = 2519,
  [3947] = 2525,
  [3948] = 2513,
  [3949] = 2514,
  [3950] = 2530,
  [3951] = 2520,
  [3952] = 2552,
  [3953] = 2582,
  [3954] = 2242,
  [3955] = 2499,
  [3956] = 2422,
  [3957] = 2490,
  [3958] = 2405,
  [3959] = 2272,
  [3960] = 2508,
  [3961] = 2316,
  [3962] = 2487,
  [3963] = 2555,
  [3964] = 2277,
  [3965] = 2542,
  [3966] = 2515,
  [3967] = 3853,
  [3968] = 2272,
  [3969] = 2527,
  [3970] = 2513,
  [3971] = 2588,
  [3972] = 2516,
  [3973] = 2550,
  [3974] = 2556,
  [3975] = 2242,
  [3976] = 2516,
  [3977] = 2517,
  [3978] = 2407,
  [3979] = 2364,
  [3980] = 2258,
  [3981] = 2278,
  [3982] = 2518,
  [3983] = 2528,
  [3984] = 2488,
  [3985] = 2265,
  [3986] = 2443,
  [3987] = 2561,
  [3988] = 2383,
  [3989] = 2507,
  [3990] = 2519,
  [3991] = 2558,
  [3992] = 2416,
  [3993] = 2286,
  [3994] = 2425,
  [3995] = 2560,
  [3996] = 2520,
  [3997] = 2256,
  [3998] = 2313,
  [3999] = 2314,
  [4000] = 2236,
  [4001] = 2572,
  [4002] = 2506,
  [4003] = 2236,
  [4004] = 2234,
  [4005] = 2292,
  [4006] = 2241,
  [4007] = 2487,
  [4008] = 2521,
  [4009] = 2522,
  [4010] = 2583,
  [4011] = 2299,
  [4012] = 2410,
  [4013] = 2589,
  [4014] = 2523,
  [4015] = 2367,
  [4016] = 2514,
  [4017] = 2420,
  [4018] = 2421,
  [4019] = 2236,
  [4020] = 2399,
  [4021] = 2573,
  [4022] = 2538,
  [4023] = 2524,
  [4024] = 2242,
  [4025] = 2507,
  [4026] = 2313,
  [4027] = 2240,
  [4028] = 2276,
  [4029] = 2545,
  [4030] = 2499,
  [4031] = 2561,
  [4032] = 2445,
  [4033] = 2240,
  [4034] = 2529,
  [4035] = 2488,
  [4036] = 2508,
  [4037] = 2354,
  [4038] = 3843,
  [4039] = 2575,
  [4040] = 2531,
  [4041] = 2234,
  [4042] = 2469,
  [4043] = 2240,
  [4044] = 2502,
  [4045] = 2551,
  [4046] = 3853,
  [4047] = 2489,
  [4048] = 3790,
  [4049] = 2400,
  [4050] = 2574,
  [4051] = 2502,
  [4052] = 2536,
  [4053] = 2281,
  [4054] = 2282,
  [4055] = 2367,
  [4056] = 2537,
  [4057] = 4057,
  [4058] = 3843,
  [4059] = 2539,
  [4060] = 2234,
  [4061] = 2545,
  [4062] = 2446,
  [4063] = 2496,
  [4064] = 2352,
  [4065] = 2367,
  [4066] = 2260,
  [4067] = 2236,
  [4068] = 2234,
  [4069] = 2240,
  [4070] = 2241,
  [4071] = 4071,
  [4072] = 2241,
  [4073] = 2510,
  [4074] = 2524,
  [4075] = 3790,
  [4076] = 2517,
  [4077] = 2262,
  [4078] = 2448,
  [4079] = 2442,
  [4080] = 2467,
  [4081] = 2522,
  [4082] = 2562,
  [4083] = 2495,
  [4084] = 3933,
  [4085] = 2285,
  [4086] = 2485,
  [4087] = 2490,
  [4088] = 2552,
  [4089] = 2259,
  [4090] = 2236,
  [4091] = 2544,
  [4092] = 2241,
  [4093] = 2287,
  [4094] = 2515,
  [4095] = 2240,
  [4096] = 2255,
  [4097] = 2491,
  [4098] = 2438,
  [4099] = 2533,
  [4100] = 4057,
  [4101] = 2555,
  [4102] = 2534,
  [4103] = 2241,
  [4104] = 2240,
  [4105] = 2563,
  [4106] = 2353,
  [4107] = 4071,
  [4108] = 2582,
  [4109] = 2492,
  [4110] = 2288,
  [4111] = 2583,
  [4112] = 2575,
  [4113] = 2491,
  [4114] = 2289,
  [4115] = 2503,
  [4116] = 2588,
  [4117] = 2422,
  [4118] = 2495,
  [4119] = 2589,
  [4120] = 2496,
  [4121] = 2497,
  [4122] = 2504,
  [4123] = 2290,
  [4124] = 2525,
  [4125] = 2256,
  [4126] = 2582,
  [4127] = 2438,
  [4128] = 4071,
  [4129] = 2467,
  [4130] = 2445,
  [4131] = 2255,
  [4132] = 2469,
  [4133] = 2255,
  [4134] = 2487,
  [4135] = 2258,
  [4136] = 2259,
  [4137] = 2488,
  [4138] = 2489,
  [4139] = 2490,
  [4140] = 2491,
  [4141] = 2255,
  [4142] = 2492,
  [4143] = 2495,
  [4144] = 2496,
  [4145] = 2272,
  [4146] = 2497,
  [4147] = 2498,
  [4148] = 2256,
  [4149] = 2399,
  [4150] = 2272,
  [4151] = 2256,
  [4152] = 2499,
  [4153] = 2502,
  [4154] = 2506,
  [4155] = 2507,
  [4156] = 2508,
  [4157] = 2510,
  [4158] = 2512,
  [4159] = 2472,
  [4160] = 2513,
  [4161] = 2514,
  [4162] = 2515,
  [4163] = 2516,
  [4164] = 2517,
  [4165] = 2518,
  [4166] = 2519,
  [4167] = 2520,
  [4168] = 2521,
  [4169] = 2272,
  [4170] = 2256,
  [4171] = 2522,
  [4172] = 2523,
  [4173] = 2524,
  [4174] = 2527,
  [4175] = 2528,
  [4176] = 2529,
  [4177] = 2531,
  [4178] = 2533,
  [4179] = 2272,
  [4180] = 2256,
  [4181] = 2534,
  [4182] = 2538,
  [4183] = 2473,
  [4184] = 2443,
  [4185] = 4057,
  [4186] = 2312,
  [4187] = 3933,
  [4188] = 2301,
  [4189] = 2283,
  [4190] = 2263,
  [4191] = 2264,
  [4192] = 2265,
  [4193] = 3853,
  [4194] = 2367,
  [4195] = 2258,
  [4196] = 2259,
  [4197] = 2255,
  [4198] = 3790,
  [4199] = 2314,
  [4200] = 2272,
  [4201] = 2256,
  [4202] = 3933,
  [4203] = 4071,
  [4204] = 2286,
  [4205] = 2292,
  [4206] = 2424,
  [4207] = 2260,
  [4208] = 2299,
  [4209] = 2262,
  [4210] = 2372,
  [4211] = 2313,
  [4212] = 2260,
  [4213] = 2262,
  [4214] = 3790,
  [4215] = 2407,
  [4216] = 2416,
  [4217] = 2420,
  [4218] = 2259,
  [4219] = 4219,
  [4220] = 2365,
  [4221] = 2400,
  [4222] = 4222,
  [4223] = 2410,
  [4224] = 2446,
  [4225] = 2372,
  [4226] = 2328,
  [4227] = 2260,
  [4228] = 2262,
  [4229] = 2260,
  [4230] = 2262,
  [4231] = 2537,
  [4232] = 2539,
  [4233] = 2485,
  [4234] = 2544,
  [4235] = 2545,
  [4236] = 2547,
  [4237] = 2550,
  [4238] = 2551,
  [4239] = 2552,
  [4240] = 2555,
  [4241] = 2556,
  [4242] = 2558,
  [4243] = 2560,
  [4244] = 2561,
  [4245] = 2562,
  [4246] = 2563,
  [4247] = 2503,
  [4248] = 2504,
  [4249] = 2525,
  [4250] = 2530,
  [4251] = 2535,
  [4252] = 2536,
  [4253] = 2540,
  [4254] = 2542,
  [4255] = 2572,
  [4256] = 2573,
  [4257] = 2574,
  [4258] = 2575,
  [4259] = 2263,
  [4260] = 3853,
  [4261] = 2264,
  [4262] = 2242,
  [4263] = 2260,
  [4264] = 2265,
  [4265] = 2583,
  [4266] = 2588,
  [4267] = 2589,
  [4268] = 2262,
  [4269] = 2263,
  [4270] = 2264,
  [4271] = 2265,
  [4272] = 2367,
  [4273] = 2313,
  [4274] = 2316,
  [4275] = 2277,
  [4276] = 2278,
  [4277] = 2272,
  [4278] = 4057,
  [4279] = 2276,
  [4280] = 2281,
  [4281] = 2282,
  [4282] = 2284,
  [4283] = 2256,
  [4284] = 2263,
  [4285] = 2448,
  [4286] = 2258,
  [4287] = 2442,
  [4288] = 2264,
  [4289] = 2259,
  [4290] = 2265,
  [4291] = 2285,
  [4292] = 2287,
  [4293] = 2288,
  [4294] = 2289,
  [4295] = 2425,
  [4296] = 2258,
  [4297] = 2290,
  [4298] = 2421,
  [4299] = 2301,
  [4300] = 2299,
  [4301] = 2264,
  [4302] = 2265,
  [4303] = 2264,
  [4304] = 2316,
  [4305] = 2299,
  [4306] = 2277,
  [4307] = 2278,
  [4308] = 2276,
  [4309] = 2281,
  [4310] = 2282,
  [4311] = 2263,
  [4312] = 2284,
  [4313] = 2285,
  [4314] = 2287,
  [4315] = 2288,
  [4316] = 2289,
  [4317] = 2588,
  [4318] = 2290,
  [4319] = 2312,
  [4320] = 2301,
  [4321] = 2283,
  [4322] = 2272,
  [4323] = 2265,
  [4324] = 2258,
  [4325] = 2259,
  [4326] = 2286,
  [4327] = 2292,
  [4328] = 2299,
  [4329] = 2255,
  [4330] = 2256,
  [4331] = 2258,
  [4332] = 2328,
  [4333] = 2272,
  [4334] = 2256,
  [4335] = 2313,
  [4336] = 2259,
  [4337] = 2258,
  [4338] = 2487,
  [4339] = 2255,
  [4340] = 2259,
  [4341] = 2260,
  [4342] = 4342,
  [4343] = 2264,
  [4344] = 2260,
  [4345] = 2488,
  [4346] = 2255,
  [4347] = 4219,
  [4348] = 2314,
  [4349] = 4222,
  [4350] = 2489,
  [4351] = 2255,
  [4352] = 2490,
  [4353] = 2491,
  [4354] = 2328,
  [4355] = 2372,
  [4356] = 2583,
  [4357] = 2366,
  [4358] = 2314,
  [4359] = 2492,
  [4360] = 2405,
  [4361] = 2495,
  [4362] = 2496,
  [4363] = 2497,
  [4364] = 2498,
  [4365] = 2499,
  [4366] = 2368,
  [4367] = 2502,
  [4368] = 2313,
  [4369] = 2422,
  [4370] = 2506,
  [4371] = 2507,
  [4372] = 2508,
  [4373] = 2510,
  [4374] = 2512,
  [4375] = 2265,
  [4376] = 2513,
  [4377] = 2514,
  [4378] = 2515,
  [4379] = 2516,
  [4380] = 2517,
  [4381] = 2518,
  [4382] = 2519,
  [4383] = 2520,
  [4384] = 2521,
  [4385] = 2522,
  [4386] = 2523,
  [4387] = 2524,
  [4388] = 2313,
  [4389] = 2527,
  [4390] = 2528,
  [4391] = 2529,
  [4392] = 2531,
  [4393] = 2533,
  [4394] = 2534,
  [4395] = 2538,
  [4396] = 2537,
  [4397] = 2539,
  [4398] = 2258,
  [4399] = 2485,
  [4400] = 2544,
  [4401] = 2545,
  [4402] = 2260,
  [4403] = 2259,
  [4404] = 2260,
  [4405] = 2262,
  [4406] = 2262,
  [4407] = 2547,
  [4408] = 2550,
  [4409] = 2551,
  [4410] = 2552,
  [4411] = 2369,
  [4412] = 2373,
  [4413] = 2555,
  [4414] = 2265,
  [4415] = 2383,
  [4416] = 2556,
  [4417] = 2558,
  [4418] = 2560,
  [4419] = 2561,
  [4420] = 2353,
  [4421] = 2314,
  [4422] = 2263,
  [4423] = 2562,
  [4424] = 2563,
  [4425] = 2503,
  [4426] = 2504,
  [4427] = 2525,
  [4428] = 2530,
  [4429] = 2536,
  [4430] = 2258,
  [4431] = 2264,
  [4432] = 2272,
  [4433] = 2256,
  [4434] = 2259,
  [4435] = 2272,
  [4436] = 2265,
  [4437] = 2540,
  [4438] = 2256,
  [4439] = 4219,
  [4440] = 2262,
  [4441] = 2314,
  [4442] = 4222,
  [4443] = 2264,
  [4444] = 2370,
  [4445] = 2312,
  [4446] = 2379,
  [4447] = 2272,
  [4448] = 2256,
  [4449] = 2283,
  [4450] = 2589,
  [4451] = 2313,
  [4452] = 2364,
  [4453] = 2272,
  [4454] = 2256,
  [4455] = 2262,
  [4456] = 2316,
  [4457] = 2258,
  [4458] = 2277,
  [4459] = 2278,
  [4460] = 2542,
  [4461] = 2572,
  [4462] = 2276,
  [4463] = 2281,
  [4464] = 2282,
  [4465] = 2284,
  [4466] = 2573,
  [4467] = 2259,
  [4468] = 2574,
  [4469] = 2285,
  [4470] = 2260,
  [4471] = 2287,
  [4472] = 2288,
  [4473] = 2289,
  [4474] = 2290,
  [4475] = 2313,
  [4476] = 2316,
  [4477] = 2277,
  [4478] = 2278,
  [4479] = 2272,
  [4480] = 2276,
  [4481] = 2281,
  [4482] = 2282,
  [4483] = 2284,
  [4484] = 2256,
  [4485] = 2285,
  [4486] = 2287,
  [4487] = 2288,
  [4488] = 2289,
  [4489] = 2290,
  [4490] = 2314,
  [4491] = 4342,
  [4492] = 2263,
  [4493] = 2312,
  [4494] = 2301,
  [4495] = 2283,
  [4496] = 2272,
  [4497] = 2255,
  [4498] = 2263,
  [4499] = 2260,
  [4500] = 2575,
  [4501] = 2256,
  [4502] = 2264,
  [4503] = 2263,
  [4504] = 2272,
  [4505] = 2260,
  [4506] = 2264,
  [4507] = 2265,
  [4508] = 2256,
  [4509] = 2265,
  [4510] = 2312,
  [4511] = 2301,
  [4512] = 2283,
  [4513] = 2262,
  [4514] = 2255,
  [4515] = 2258,
  [4516] = 2262,
  [4517] = 2255,
  [4518] = 2259,
  [4519] = 2262,
  [4520] = 2372,
  [4521] = 2313,
  [4522] = 2367,
  [4523] = 2284,
  [4524] = 2286,
  [4525] = 2316,
  [4526] = 2277,
  [4527] = 2278,
  [4528] = 2263,
  [4529] = 2292,
  [4530] = 2276,
  [4531] = 2281,
  [4532] = 2282,
  [4533] = 2286,
  [4534] = 2299,
  [4535] = 2292,
  [4536] = 2285,
  [4537] = 2582,
  [4538] = 2287,
  [4539] = 2288,
  [4540] = 2289,
  [4541] = 2290,
  [4542] = 4219,
  [4543] = 4222,
  [4544] = 2354,
  [4545] = 2286,
  [4546] = 2263,
  [4547] = 2292,
  [4548] = 2352,
  [4549] = 2535,
  [4550] = 2285,
  [4551] = 2290,
  [4552] = 2281,
  [4553] = 2282,
  [4554] = 2289,
  [4555] = 2292,
  [4556] = 2352,
  [4557] = 2448,
  [4558] = 2442,
  [4559] = 2367,
  [4560] = 2312,
  [4561] = 2372,
  [4562] = 2328,
  [4563] = 2410,
  [4564] = 2299,
  [4565] = 2313,
  [4566] = 2285,
  [4567] = 2285,
  [4568] = 2284,
  [4569] = 2353,
  [4570] = 2278,
  [4571] = 2299,
  [4572] = 2425,
  [4573] = 2314,
  [4574] = 2364,
  [4575] = 2313,
  [4576] = 2446,
  [4577] = 2286,
  [4578] = 2290,
  [4579] = 2292,
  [4580] = 2301,
  [4581] = 2287,
  [4582] = 2312,
  [4583] = 2277,
  [4584] = 2287,
  [4585] = 2282,
  [4586] = 2328,
  [4587] = 2278,
  [4588] = 2286,
  [4589] = 2288,
  [4590] = 2292,
  [4591] = 4222,
  [4592] = 2299,
  [4593] = 2301,
  [4594] = 2281,
  [4595] = 2299,
  [4596] = 2399,
  [4597] = 4219,
  [4598] = 2283,
  [4599] = 2283,
  [4600] = 2284,
  [4601] = 2366,
  [4602] = 2328,
  [4603] = 2289,
  [4604] = 2368,
  [4605] = 2312,
  [4606] = 2286,
  [4607] = 2301,
  [4608] = 2283,
  [4609] = 4609,
  [4610] = 2288,
  [4611] = 2282,
  [4612] = 2292,
  [4613] = 2370,
  [4614] = 2379,
  [4615] = 2365,
  [4616] = 2285,
  [4617] = 2438,
  [4618] = 2282,
  [4619] = 2313,
  [4620] = 2276,
  [4621] = 2299,
  [4622] = 2286,
  [4623] = 2365,
  [4624] = 2290,
  [4625] = 2467,
  [4626] = 2287,
  [4627] = 4219,
  [4628] = 2314,
  [4629] = 2354,
  [4630] = 2445,
  [4631] = 2314,
  [4632] = 2367,
  [4633] = 2369,
  [4634] = 2314,
  [4635] = 2316,
  [4636] = 4222,
  [4637] = 2277,
  [4638] = 2278,
  [4639] = 2272,
  [4640] = 2469,
  [4641] = 2284,
  [4642] = 2276,
  [4643] = 2472,
  [4644] = 2281,
  [4645] = 2282,
  [4646] = 2313,
  [4647] = 2373,
  [4648] = 2473,
  [4649] = 2284,
  [4650] = 2316,
  [4651] = 2367,
  [4652] = 2277,
  [4653] = 2278,
  [4654] = 2256,
  [4655] = 2443,
  [4656] = 2369,
  [4657] = 2281,
  [4658] = 2286,
  [4659] = 2281,
  [4660] = 2282,
  [4661] = 2287,
  [4662] = 2383,
  [4663] = 2287,
  [4664] = 2313,
  [4665] = 2286,
  [4666] = 2285,
  [4667] = 2422,
  [4668] = 2287,
  [4669] = 2284,
  [4670] = 2288,
  [4671] = 2281,
  [4672] = 2289,
  [4673] = 2286,
  [4674] = 2290,
  [4675] = 2316,
  [4676] = 2288,
  [4677] = 2314,
  [4678] = 2289,
  [4679] = 2365,
  [4680] = 2292,
  [4681] = 2290,
  [4682] = 2424,
  [4683] = 2312,
  [4684] = 2277,
  [4685] = 2314,
  [4686] = 2289,
  [4687] = 2278,
  [4688] = 2353,
  [4689] = 2290,
  [4690] = 2405,
  [4691] = 2299,
  [4692] = 2313,
  [4693] = 2301,
  [4694] = 2283,
  [4695] = 2288,
  [4696] = 2367,
  [4697] = 2373,
  [4698] = 2316,
  [4699] = 2312,
  [4700] = 2301,
  [4701] = 2283,
  [4702] = 2312,
  [4703] = 2292,
  [4704] = 2316,
  [4705] = 2289,
  [4706] = 2276,
  [4707] = 2281,
  [4708] = 2407,
  [4709] = 2282,
  [4710] = 2277,
  [4711] = 2292,
  [4712] = 2278,
  [4713] = 2301,
  [4714] = 2313,
  [4715] = 2290,
  [4716] = 2284,
  [4717] = 2277,
  [4718] = 2299,
  [4719] = 2283,
  [4720] = 2285,
  [4721] = 2383,
  [4722] = 2328,
  [4723] = 2287,
  [4724] = 2316,
  [4725] = 2288,
  [4726] = 2276,
  [4727] = 2328,
  [4728] = 2416,
  [4729] = 2420,
  [4730] = 2421,
  [4731] = 2313,
  [4732] = 2314,
  [4733] = 2312,
  [4734] = 2316,
  [4735] = 2313,
  [4736] = 2301,
  [4737] = 2285,
  [4738] = 2400,
  [4739] = 2277,
  [4740] = 2288,
  [4741] = 2278,
  [4742] = 2284,
  [4743] = 2283,
  [4744] = 2276,
  [4745] = 2289,
  [4746] = 2367,
  [4747] = 2276,
  [4748] = 2276,
  [4749] = 2531,
  [4750] = 2353,
  [4751] = 2370,
  [4752] = 2379,
  [4753] = 2369,
  [4754] = 2365,
  [4755] = 2373,
  [4756] = 2416,
  [4757] = 2383,
  [4758] = 2367,
  [4759] = 2372,
  [4760] = 2420,
  [4761] = 2353,
  [4762] = 2365,
  [4763] = 2272,
  [4764] = 2373,
  [4765] = 2421,
  [4766] = 2328,
  [4767] = 2367,
  [4768] = 2256,
  [4769] = 2367,
  [4770] = 2367,
  [4771] = 2422,
  [4772] = 2537,
  [4773] = 2539,
  [4774] = 2485,
  [4775] = 2544,
  [4776] = 2365,
  [4777] = 2545,
  [4778] = 2547,
  [4779] = 2550,
  [4780] = 2551,
  [4781] = 2552,
  [4782] = 2555,
  [4783] = 2556,
  [4784] = 2558,
  [4785] = 2560,
  [4786] = 2561,
  [4787] = 2562,
  [4788] = 2563,
  [4789] = 2503,
  [4790] = 2504,
  [4791] = 2525,
  [4792] = 2530,
  [4793] = 2535,
  [4794] = 2536,
  [4795] = 2540,
  [4796] = 2542,
  [4797] = 2572,
  [4798] = 2573,
  [4799] = 2574,
  [4800] = 2575,
  [4801] = 2354,
  [4802] = 2352,
  [4803] = 2272,
  [4804] = 2424,
  [4805] = 2364,
  [4806] = 2256,
  [4807] = 4807,
  [4808] = 2400,
  [4809] = 2366,
  [4810] = 2368,
  [4811] = 2370,
  [4812] = 2379,
  [4813] = 2328,
  [4814] = 2582,
  [4815] = 2583,
  [4816] = 4816,
  [4817] = 2588,
  [4818] = 2589,
  [4819] = 2422,
  [4820] = 2369,
  [4821] = 2272,
  [4822] = 2256,
  [4823] = 2354,
  [4824] = 2365,
  [4825] = 2367,
  [4826] = 2410,
  [4827] = 4827,
  [4828] = 2446,
  [4829] = 2258,
  [4830] = 2369,
  [4831] = 2372,
  [4832] = 2259,
  [4833] = 2272,
  [4834] = 2373,
  [4835] = 2328,
  [4836] = 2352,
  [4837] = 2328,
  [4838] = 2364,
  [4839] = 2256,
  [4840] = 2353,
  [4841] = 2383,
  [4842] = 2354,
  [4843] = 2448,
  [4844] = 2366,
  [4845] = 2368,
  [4846] = 2442,
  [4847] = 2373,
  [4848] = 2354,
  [4849] = 2367,
  [4850] = 2370,
  [4851] = 2383,
  [4852] = 2425,
  [4853] = 2379,
  [4854] = 2369,
  [4855] = 2438,
  [4856] = 2352,
  [4857] = 2263,
  [4858] = 2383,
  [4859] = 2467,
  [4860] = 2365,
  [4861] = 2372,
  [4862] = 2372,
  [4863] = 2272,
  [4864] = 2405,
  [4865] = 2256,
  [4866] = 2445,
  [4867] = 2240,
  [4868] = 2364,
  [4869] = 2313,
  [4870] = 2256,
  [4871] = 2364,
  [4872] = 2365,
  [4873] = 2469,
  [4874] = 2353,
  [4875] = 2367,
  [4876] = 2366,
  [4877] = 2399,
  [4878] = 2368,
  [4879] = 2366,
  [4880] = 2370,
  [4881] = 2379,
  [4882] = 2368,
  [4883] = 2472,
  [4884] = 2487,
  [4885] = 2372,
  [4886] = 2354,
  [4887] = 2488,
  [4888] = 2236,
  [4889] = 2370,
  [4890] = 2489,
  [4891] = 2490,
  [4892] = 2491,
  [4893] = 2379,
  [4894] = 2473,
  [4895] = 2492,
  [4896] = 2234,
  [4897] = 2241,
  [4898] = 2495,
  [4899] = 2496,
  [4900] = 2497,
  [4901] = 2498,
  [4902] = 2365,
  [4903] = 2443,
  [4904] = 2499,
  [4905] = 2502,
  [4906] = 2352,
  [4907] = 2372,
  [4908] = 2367,
  [4909] = 2367,
  [4910] = 2272,
  [4911] = 2369,
  [4912] = 2328,
  [4913] = 2373,
  [4914] = 2405,
  [4915] = 2365,
  [4916] = 2367,
  [4917] = 2383,
  [4918] = 2328,
  [4919] = 2364,
  [4920] = 2506,
  [4921] = 2507,
  [4922] = 2508,
  [4923] = 4923,
  [4924] = 4924,
  [4925] = 2510,
  [4926] = 2512,
  [4927] = 2513,
  [4928] = 2514,
  [4929] = 2328,
  [4930] = 2515,
  [4931] = 2516,
  [4932] = 2517,
  [4933] = 2518,
  [4934] = 2519,
  [4935] = 2520,
  [4936] = 2521,
  [4937] = 2522,
  [4938] = 2523,
  [4939] = 2524,
  [4940] = 2527,
  [4941] = 2528,
  [4942] = 4942,
  [4943] = 2529,
  [4944] = 2367,
  [4945] = 2366,
  [4946] = 2368,
  [4947] = 2407,
  [4948] = 2533,
  [4949] = 2534,
  [4950] = 2353,
  [4951] = 2538,
  [4952] = 2352,
  [4953] = 2372,
  [4954] = 2256,
  [4955] = 4924,
  [4956] = 2313,
  [4957] = 2367,
  [4958] = 2416,
  [4959] = 2366,
  [4960] = 2368,
  [4961] = 2354,
  [4962] = 2354,
  [4963] = 2420,
  [4964] = 2421,
  [4965] = 4827,
  [4966] = 2489,
  [4967] = 2490,
  [4968] = 2368,
  [4969] = 2313,
  [4970] = 2491,
  [4971] = 2383,
  [4972] = 2370,
  [4973] = 2379,
  [4974] = 2263,
  [4975] = 2272,
  [4976] = 2399,
  [4977] = 2354,
  [4978] = 2538,
  [4979] = 2240,
  [4980] = 2467,
  [4981] = 2292,
  [4982] = 2400,
  [4983] = 4609,
  [4984] = 2424,
  [4985] = 2472,
  [4986] = 2365,
  [4987] = 2369,
  [4988] = 4827,
  [4989] = 2473,
  [4990] = 2446,
  [4991] = 2365,
  [4992] = 2373,
  [4993] = 2352,
  [4994] = 2443,
  [4995] = 2513,
  [4996] = 2422,
  [4997] = 2373,
  [4998] = 2405,
  [4999] = 2353,
  [5000] = 2299,
  [5001] = 2514,
  [5002] = 2369,
  [5003] = 2445,
  [5004] = 2367,
  [5005] = 2373,
  [5006] = 2405,
  [5007] = 2383,
  [5008] = 4942,
  [5009] = 2372,
  [5010] = 2263,
  [5011] = 2410,
  [5012] = 2445,
  [5013] = 2446,
  [5014] = 2424,
  [5015] = 2352,
  [5016] = 2383,
  [5017] = 2492,
  [5018] = 2313,
  [5019] = 2445,
  [5020] = 2446,
  [5021] = 2352,
  [5022] = 2272,
  [5023] = 2407,
  [5024] = 2422,
  [5025] = 2256,
  [5026] = 2443,
  [5027] = 2469,
  [5028] = 2256,
  [5029] = 2416,
  [5030] = 2352,
  [5031] = 2236,
  [5032] = 2234,
  [5033] = 2241,
  [5034] = 2420,
  [5035] = 2407,
  [5036] = 2421,
  [5037] = 2399,
  [5038] = 2424,
  [5039] = 2372,
  [5040] = 2399,
  [5041] = 2383,
  [5042] = 2369,
  [5043] = 2367,
  [5044] = 2416,
  [5045] = 2256,
  [5046] = 2367,
  [5047] = 2259,
  [5048] = 2487,
  [5049] = 2379,
  [5050] = 2353,
  [5051] = 2420,
  [5052] = 2258,
  [5053] = 2372,
  [5054] = 2353,
  [5055] = 2421,
  [5056] = 2259,
  [5057] = 4924,
  [5058] = 2352,
  [5059] = 2425,
  [5060] = 2240,
  [5061] = 2506,
  [5062] = 2313,
  [5063] = 2445,
  [5064] = 2400,
  [5065] = 2448,
  [5066] = 2442,
  [5067] = 2507,
  [5068] = 4816,
  [5069] = 2366,
  [5070] = 2370,
  [5071] = 2368,
  [5072] = 2364,
  [5073] = 2508,
  [5074] = 2353,
  [5075] = 2410,
  [5076] = 2422,
  [5077] = 2354,
  [5078] = 2537,
  [5079] = 2539,
  [5080] = 2438,
  [5081] = 2366,
  [5082] = 2372,
  [5083] = 2467,
  [5084] = 2424,
  [5085] = 2510,
  [5086] = 2400,
  [5087] = 2425,
  [5088] = 2469,
  [5089] = 2485,
  [5090] = 2438,
  [5091] = 2236,
  [5092] = 2472,
  [5093] = 5093,
  [5094] = 2473,
  [5095] = 2443,
  [5096] = 2410,
  [5097] = 2234,
  [5098] = 2241,
  [5099] = 2446,
  [5100] = 2313,
  [5101] = 2405,
  [5102] = 2515,
  [5103] = 2544,
  [5104] = 2512,
  [5105] = 2582,
  [5106] = 2365,
  [5107] = 2405,
  [5108] = 2467,
  [5109] = 2445,
  [5110] = 2545,
  [5111] = 2446,
  [5112] = 2547,
  [5113] = 2550,
  [5114] = 2551,
  [5115] = 2552,
  [5116] = 2555,
  [5117] = 2556,
  [5118] = 2558,
  [5119] = 2560,
  [5120] = 2561,
  [5121] = 2562,
  [5122] = 2563,
  [5123] = 2503,
  [5124] = 2504,
  [5125] = 2369,
  [5126] = 2525,
  [5127] = 2405,
  [5128] = 2530,
  [5129] = 2365,
  [5130] = 2535,
  [5131] = 2536,
  [5132] = 2405,
  [5133] = 2583,
  [5134] = 2540,
  [5135] = 2469,
  [5136] = 2364,
  [5137] = 2367,
  [5138] = 2422,
  [5139] = 2272,
  [5140] = 2399,
  [5141] = 2542,
  [5142] = 2572,
  [5143] = 2573,
  [5144] = 4924,
  [5145] = 2574,
  [5146] = 4942,
  [5147] = 2467,
  [5148] = 2367,
  [5149] = 2258,
  [5150] = 2495,
  [5151] = 2258,
  [5152] = 2373,
  [5153] = 2372,
  [5154] = 2354,
  [5155] = 2407,
  [5156] = 2422,
  [5157] = 2367,
  [5158] = 2496,
  [5159] = 2367,
  [5160] = 2575,
  [5161] = 2367,
  [5162] = 4942,
  [5163] = 2472,
  [5164] = 2354,
  [5165] = 2416,
  [5166] = 2367,
  [5167] = 2407,
  [5168] = 2448,
  [5169] = 2442,
  [5170] = 2516,
  [5171] = 2473,
  [5172] = 2352,
  [5173] = 4827,
  [5174] = 2517,
  [5175] = 4807,
  [5176] = 2420,
  [5177] = 2425,
  [5178] = 2443,
  [5179] = 2421,
  [5180] = 2438,
  [5181] = 2352,
  [5182] = 2367,
  [5183] = 2367,
  [5184] = 2518,
  [5185] = 2467,
  [5186] = 2416,
  [5187] = 2366,
  [5188] = 2367,
  [5189] = 2368,
  [5190] = 2354,
  [5191] = 2519,
  [5192] = 2448,
  [5193] = 2420,
  [5194] = 2272,
  [5195] = 2421,
  [5196] = 2442,
  [5197] = 2588,
  [5198] = 2520,
  [5199] = 2370,
  [5200] = 2379,
  [5201] = 2424,
  [5202] = 2370,
  [5203] = 4807,
  [5204] = 2497,
  [5205] = 2272,
  [5206] = 2400,
  [5207] = 2256,
  [5208] = 2379,
  [5209] = 2364,
  [5210] = 2448,
  [5211] = 2400,
  [5212] = 2425,
  [5213] = 2256,
  [5214] = 2469,
  [5215] = 2410,
  [5216] = 2353,
  [5217] = 4816,
  [5218] = 2521,
  [5219] = 2498,
  [5220] = 2438,
  [5221] = 2522,
  [5222] = 2410,
  [5223] = 2365,
  [5224] = 2259,
  [5225] = 2523,
  [5226] = 2589,
  [5227] = 2399,
  [5228] = 2263,
  [5229] = 2256,
  [5230] = 2442,
  [5231] = 2364,
  [5232] = 2313,
  [5233] = 2370,
  [5234] = 2524,
  [5235] = 2383,
  [5236] = 2364,
  [5237] = 2240,
  [5238] = 2488,
  [5239] = 2370,
  [5240] = 2366,
  [5241] = 2369,
  [5242] = 2365,
  [5243] = 2372,
  [5244] = 2368,
  [5245] = 2364,
  [5246] = 2365,
  [5247] = 2422,
  [5248] = 4923,
  [5249] = 2367,
  [5250] = 2367,
  [5251] = 2367,
  [5252] = 2405,
  [5253] = 2367,
  [5254] = 2366,
  [5255] = 2368,
  [5256] = 2236,
  [5257] = 2422,
  [5258] = 2527,
  [5259] = 2472,
  [5260] = 2272,
  [5261] = 2234,
  [5262] = 2472,
  [5263] = 2256,
  [5264] = 2448,
  [5265] = 2442,
  [5266] = 2367,
  [5267] = 2366,
  [5268] = 2365,
  [5269] = 2368,
  [5270] = 2383,
  [5271] = 2528,
  [5272] = 2425,
  [5273] = 2529,
  [5274] = 2438,
  [5275] = 2531,
  [5276] = 2370,
  [5277] = 2379,
  [5278] = 2255,
  [5279] = 4807,
  [5280] = 2473,
  [5281] = 2364,
  [5282] = 2241,
  [5283] = 2367,
  [5284] = 2369,
  [5285] = 2365,
  [5286] = 2499,
  [5287] = 2379,
  [5288] = 2365,
  [5289] = 2502,
  [5290] = 2373,
  [5291] = 4923,
  [5292] = 2353,
  [5293] = 2443,
  [5294] = 2533,
  [5295] = 4816,
  [5296] = 2353,
  [5297] = 2469,
  [5298] = 2367,
  [5299] = 2373,
  [5300] = 2407,
  [5301] = 2379,
  [5302] = 2383,
  [5303] = 2369,
  [5304] = 2272,
  [5305] = 2473,
  [5306] = 4923,
  [5307] = 2534,
  [5308] = 2272,
  [5309] = 2373,
  [5310] = 2516,
  [5311] = 5311,
  [5312] = 5312,
  [5313] = 2506,
  [5314] = 2507,
  [5315] = 2503,
  [5316] = 2508,
  [5317] = 2504,
  [5318] = 2510,
  [5319] = 2512,
  [5320] = 2255,
  [5321] = 2513,
  [5322] = 2514,
  [5323] = 2515,
  [5324] = 2517,
  [5325] = 2518,
  [5326] = 2519,
  [5327] = 2520,
  [5328] = 2537,
  [5329] = 2539,
  [5330] = 2521,
  [5331] = 2522,
  [5332] = 2416,
  [5333] = 2523,
  [5334] = 2524,
  [5335] = 2286,
  [5336] = 2485,
  [5337] = 2527,
  [5338] = 2528,
  [5339] = 2544,
  [5340] = 2445,
  [5341] = 2529,
  [5342] = 2531,
  [5343] = 2525,
  [5344] = 2272,
  [5345] = 2533,
  [5346] = 2534,
  [5347] = 2530,
  [5348] = 2545,
  [5349] = 2538,
  [5350] = 2256,
  [5351] = 2535,
  [5352] = 2547,
  [5353] = 2537,
  [5354] = 2446,
  [5355] = 2485,
  [5356] = 2544,
  [5357] = 2545,
  [5358] = 2469,
  [5359] = 2582,
  [5360] = 2547,
  [5361] = 2550,
  [5362] = 2551,
  [5363] = 2552,
  [5364] = 2555,
  [5365] = 2556,
  [5366] = 2558,
  [5367] = 2560,
  [5368] = 2561,
  [5369] = 2562,
  [5370] = 2563,
  [5371] = 2503,
  [5372] = 2504,
  [5373] = 2525,
  [5374] = 2530,
  [5375] = 2535,
  [5376] = 2536,
  [5377] = 2550,
  [5378] = 2583,
  [5379] = 2540,
  [5380] = 2551,
  [5381] = 2536,
  [5382] = 2552,
  [5383] = 2542,
  [5384] = 2572,
  [5385] = 2573,
  [5386] = 2574,
  [5387] = 2555,
  [5388] = 2556,
  [5389] = 2558,
  [5390] = 2467,
  [5391] = 2560,
  [5392] = 2399,
  [5393] = 2561,
  [5394] = 2562,
  [5395] = 2575,
  [5396] = 2563,
  [5397] = 2503,
  [5398] = 2504,
  [5399] = 2525,
  [5400] = 2530,
  [5401] = 2424,
  [5402] = 2472,
  [5403] = 2422,
  [5404] = 2469,
  [5405] = 2535,
  [5406] = 2536,
  [5407] = 2445,
  [5408] = 2446,
  [5409] = 2367,
  [5410] = 2540,
  [5411] = 5311,
  [5412] = 2258,
  [5413] = 2448,
  [5414] = 2442,
  [5415] = 2542,
  [5416] = 2588,
  [5417] = 2572,
  [5418] = 2473,
  [5419] = 2589,
  [5420] = 2573,
  [5421] = 2259,
  [5422] = 2574,
  [5423] = 2422,
  [5424] = 2537,
  [5425] = 2539,
  [5426] = 2537,
  [5427] = 2485,
  [5428] = 2544,
  [5429] = 2405,
  [5430] = 2545,
  [5431] = 2539,
  [5432] = 2540,
  [5433] = 2547,
  [5434] = 2550,
  [5435] = 2551,
  [5436] = 2552,
  [5437] = 2555,
  [5438] = 2556,
  [5439] = 2558,
  [5440] = 2560,
  [5441] = 2561,
  [5442] = 2562,
  [5443] = 2563,
  [5444] = 2503,
  [5445] = 2504,
  [5446] = 2525,
  [5447] = 2530,
  [5448] = 2367,
  [5449] = 2535,
  [5450] = 2536,
  [5451] = 2425,
  [5452] = 2443,
  [5453] = 2485,
  [5454] = 2540,
  [5455] = 2544,
  [5456] = 2545,
  [5457] = 2542,
  [5458] = 2572,
  [5459] = 2573,
  [5460] = 2574,
  [5461] = 2547,
  [5462] = 2410,
  [5463] = 2550,
  [5464] = 2438,
  [5465] = 2542,
  [5466] = 2420,
  [5467] = 2572,
  [5468] = 2575,
  [5469] = 2573,
  [5470] = 2574,
  [5471] = 2469,
  [5472] = 2424,
  [5473] = 2551,
  [5474] = 2421,
  [5475] = 2552,
  [5476] = 2555,
  [5477] = 2556,
  [5478] = 2292,
  [5479] = 2558,
  [5480] = 2405,
  [5481] = 2560,
  [5482] = 2561,
  [5483] = 2562,
  [5484] = 2563,
  [5485] = 2503,
  [5486] = 2313,
  [5487] = 2416,
  [5488] = 2424,
  [5489] = 2504,
  [5490] = 2299,
  [5491] = 2525,
  [5492] = 2407,
  [5493] = 2416,
  [5494] = 2575,
  [5495] = 2530,
  [5496] = 2575,
  [5497] = 2416,
  [5498] = 2424,
  [5499] = 2582,
  [5500] = 2583,
  [5501] = 2535,
  [5502] = 2405,
  [5503] = 2588,
  [5504] = 2589,
  [5505] = 2536,
  [5506] = 2540,
  [5507] = 2542,
  [5508] = 2420,
  [5509] = 2421,
  [5510] = 2407,
  [5511] = 2416,
  [5512] = 2313,
  [5513] = 2420,
  [5514] = 2421,
  [5515] = 2572,
  [5516] = 2469,
  [5517] = 2424,
  [5518] = 2472,
  [5519] = 2573,
  [5520] = 2448,
  [5521] = 2400,
  [5522] = 2574,
  [5523] = 2442,
  [5524] = 2410,
  [5525] = 2487,
  [5526] = 2473,
  [5527] = 2575,
  [5528] = 2313,
  [5529] = 2488,
  [5530] = 2489,
  [5531] = 2490,
  [5532] = 2400,
  [5533] = 2491,
  [5534] = 4923,
  [5535] = 2582,
  [5536] = 2583,
  [5537] = 2258,
  [5538] = 2259,
  [5539] = 4942,
  [5540] = 2422,
  [5541] = 2399,
  [5542] = 4924,
  [5543] = 2422,
  [5544] = 2492,
  [5545] = 2495,
  [5546] = 2496,
  [5547] = 2497,
  [5548] = 2498,
  [5549] = 2255,
  [5550] = 2588,
  [5551] = 2589,
  [5552] = 2420,
  [5553] = 2499,
  [5554] = 2502,
  [5555] = 2421,
  [5556] = 5093,
  [5557] = 2410,
  [5558] = 2446,
  [5559] = 2506,
  [5560] = 2507,
  [5561] = 2400,
  [5562] = 2472,
  [5563] = 2508,
  [5564] = 2510,
  [5565] = 2512,
  [5566] = 2513,
  [5567] = 2514,
  [5568] = 2515,
  [5569] = 2516,
  [5570] = 2517,
  [5571] = 2518,
  [5572] = 2519,
  [5573] = 2520,
  [5574] = 2448,
  [5575] = 2442,
  [5576] = 2487,
  [5577] = 2521,
  [5578] = 2522,
  [5579] = 2523,
  [5580] = 2524,
  [5581] = 2488,
  [5582] = 2527,
  [5583] = 2528,
  [5584] = 2425,
  [5585] = 2529,
  [5586] = 2531,
  [5587] = 2533,
  [5588] = 2534,
  [5589] = 2438,
  [5590] = 2473,
  [5591] = 2538,
  [5592] = 2489,
  [5593] = 2490,
  [5594] = 3933,
  [5595] = 2491,
  [5596] = 2367,
  [5597] = 2399,
  [5598] = 2492,
  [5599] = 2446,
  [5600] = 2221,
  [5601] = 2272,
  [5602] = 2256,
  [5603] = 2582,
  [5604] = 2277,
  [5605] = 2583,
  [5606] = 2588,
  [5607] = 2495,
  [5608] = 2496,
  [5609] = 2497,
  [5610] = 2498,
  [5611] = 2448,
  [5612] = 2313,
  [5613] = 2264,
  [5614] = 2442,
  [5615] = 2425,
  [5616] = 2499,
  [5617] = 2502,
  [5618] = 2272,
  [5619] = 2256,
  [5620] = 2473,
  [5621] = 2589,
  [5622] = 2405,
  [5623] = 2443,
  [5624] = 2265,
  [5625] = 2506,
  [5626] = 2507,
  [5627] = 2508,
  [5628] = 2510,
  [5629] = 2512,
  [5630] = 2513,
  [5631] = 2514,
  [5632] = 2515,
  [5633] = 2516,
  [5634] = 2517,
  [5635] = 2443,
  [5636] = 2518,
  [5637] = 2519,
  [5638] = 2520,
  [5639] = 2425,
  [5640] = 2448,
  [5641] = 2442,
  [5642] = 2521,
  [5643] = 2522,
  [5644] = 2523,
  [5645] = 2443,
  [5646] = 2524,
  [5647] = 2263,
  [5648] = 2240,
  [5649] = 2365,
  [5650] = 2527,
  [5651] = 2445,
  [5652] = 2446,
  [5653] = 2528,
  [5654] = 2438,
  [5655] = 2425,
  [5656] = 2529,
  [5657] = 2313,
  [5658] = 2537,
  [5659] = 2531,
  [5660] = 2487,
  [5661] = 2407,
  [5662] = 2533,
  [5663] = 2534,
  [5664] = 2438,
  [5665] = 2399,
  [5666] = 2467,
  [5667] = 2399,
  [5668] = 5668,
  [5669] = 5669,
  [5670] = 2538,
  [5671] = 5671,
  [5672] = 5672,
  [5673] = 5673,
  [5674] = 2487,
  [5675] = 2488,
  [5676] = 2488,
  [5677] = 2445,
  [5678] = 2489,
  [5679] = 2490,
  [5680] = 2491,
  [5681] = 5311,
  [5682] = 2492,
  [5683] = 2405,
  [5684] = 2495,
  [5685] = 2496,
  [5686] = 2497,
  [5687] = 2498,
  [5688] = 2236,
  [5689] = 2313,
  [5690] = 2234,
  [5691] = 2499,
  [5692] = 2489,
  [5693] = 2490,
  [5694] = 5312,
  [5695] = 2438,
  [5696] = 2491,
  [5697] = 2502,
  [5698] = 2405,
  [5699] = 2422,
  [5700] = 2241,
  [5701] = 2506,
  [5702] = 2507,
  [5703] = 2508,
  [5704] = 2510,
  [5705] = 2512,
  [5706] = 3843,
  [5707] = 2292,
  [5708] = 2513,
  [5709] = 2514,
  [5710] = 2299,
  [5711] = 2410,
  [5712] = 2515,
  [5713] = 2516,
  [5714] = 2517,
  [5715] = 2518,
  [5716] = 2519,
  [5717] = 2520,
  [5718] = 2467,
  [5719] = 2492,
  [5720] = 2422,
  [5721] = 2467,
  [5722] = 2495,
  [5723] = 2496,
  [5724] = 2521,
  [5725] = 2497,
  [5726] = 2498,
  [5727] = 2469,
  [5728] = 2400,
  [5729] = 2367,
  [5730] = 2472,
  [5731] = 2473,
  [5732] = 2522,
  [5733] = 2443,
  [5734] = 2523,
  [5735] = 2524,
  [5736] = 4827,
  [5737] = 2424,
  [5738] = 2527,
  [5739] = 2528,
  [5740] = 2529,
  [5741] = 2531,
  [5742] = 2582,
  [5743] = 2405,
  [5744] = 2533,
  [5745] = 2534,
  [5746] = 2583,
  [5747] = 2367,
  [5748] = 2538,
  [5749] = 2292,
  [5750] = 2407,
  [5751] = 2255,
  [5752] = 2424,
  [5753] = 2469,
  [5754] = 2539,
  [5755] = 2499,
  [5756] = 2422,
  [5757] = 2502,
  [5758] = 2405,
  [5759] = 2467,
  [5760] = 2588,
  [5761] = 2589,
  [5762] = 2469,
  [5763] = 2367,
  [5764] = 2299,
  [5765] = 2467,
  [5766] = 2422,
  [5767] = 2399,
  [5768] = 2416,
  [5769] = 2407,
  [5770] = 2448,
  [5771] = 2442,
  [5772] = 2506,
  [5773] = 2507,
  [5774] = 2472,
  [5775] = 2407,
  [5776] = 2508,
  [5777] = 2416,
  [5778] = 2472,
  [5779] = 2472,
  [5780] = 2510,
  [5781] = 2512,
  [5782] = 2445,
  [5783] = 2513,
  [5784] = 2514,
  [5785] = 5311,
  [5786] = 2313,
  [5787] = 2515,
  [5788] = 2420,
  [5789] = 2421,
  [5790] = 5312,
  [5791] = 2516,
  [5792] = 2517,
  [5793] = 2518,
  [5794] = 2519,
  [5795] = 2520,
  [5796] = 2405,
  [5797] = 2473,
  [5798] = 2400,
  [5799] = 2407,
  [5800] = 2420,
  [5801] = 2473,
  [5802] = 2521,
  [5803] = 2410,
  [5804] = 2421,
  [5805] = 2485,
  [5806] = 2522,
  [5807] = 5093,
  [5808] = 2523,
  [5809] = 2524,
  [5810] = 2443,
  [5811] = 2445,
  [5812] = 2400,
  [5813] = 2544,
  [5814] = 4807,
  [5815] = 2527,
  [5816] = 2528,
  [5817] = 2425,
  [5818] = 2443,
  [5819] = 2529,
  [5820] = 2531,
  [5821] = 4816,
  [5822] = 2533,
  [5823] = 2534,
  [5824] = 2420,
  [5825] = 2399,
  [5826] = 2538,
  [5827] = 2487,
  [5828] = 2313,
  [5829] = 2545,
  [5830] = 5311,
  [5831] = 5312,
  [5832] = 2438,
  [5833] = 2367,
  [5834] = 2410,
  [5835] = 2488,
  [5836] = 2421,
  [5837] = 2489,
  [5838] = 2490,
  [5839] = 2491,
  [5840] = 2422,
  [5841] = 2467,
  [5842] = 2263,
  [5843] = 5311,
  [5844] = 5312,
  [5845] = 2240,
  [5846] = 2547,
  [5847] = 2550,
  [5848] = 2551,
  [5849] = 2552,
  [5850] = 2492,
  [5851] = 5311,
  [5852] = 5312,
  [5853] = 2445,
  [5854] = 2446,
  [5855] = 2495,
  [5856] = 2496,
  [5857] = 2497,
  [5858] = 2498,
  [5859] = 2555,
  [5860] = 2556,
  [5861] = 2448,
  [5862] = 2442,
  [5863] = 2558,
  [5864] = 2560,
  [5865] = 2561,
  [5866] = 2562,
  [5867] = 2446,
  [5868] = 5311,
  [5869] = 5312,
  [5870] = 2499,
  [5871] = 2425,
  [5872] = 2502,
  [5873] = 2400,
  [5874] = 2563,
  [5875] = 2410,
  [5876] = 5312,
  [5877] = 2438,
  [5878] = 5093,
  [5879] = 2236,
  [5880] = 2234,
  [5881] = 2241,
  [5882] = 2539,
  [5883] = 2489,
  [5884] = 2496,
  [5885] = 2491,
  [5886] = 2588,
  [5887] = 2588,
  [5888] = 2545,
  [5889] = 2589,
  [5890] = 2499,
  [5891] = 2547,
  [5892] = 2487,
  [5893] = 2491,
  [5894] = 2504,
  [5895] = 2582,
  [5896] = 2583,
  [5897] = 2550,
  [5898] = 2499,
  [5899] = 2521,
  [5900] = 2588,
  [5901] = 2589,
  [5902] = 2516,
  [5903] = 2547,
  [5904] = 2561,
  [5905] = 2550,
  [5906] = 2551,
  [5907] = 2492,
  [5908] = 2528,
  [5909] = 2502,
  [5910] = 2575,
  [5911] = 2263,
  [5912] = 2316,
  [5913] = 2551,
  [5914] = 2313,
  [5915] = 2552,
  [5916] = 2517,
  [5917] = 2518,
  [5918] = 2278,
  [5919] = 2562,
  [5920] = 2292,
  [5921] = 2240,
  [5922] = 2276,
  [5923] = 2495,
  [5924] = 2281,
  [5925] = 2496,
  [5926] = 2497,
  [5927] = 2498,
  [5928] = 2555,
  [5929] = 2506,
  [5930] = 2282,
  [5931] = 2556,
  [5932] = 2525,
  [5933] = 2497,
  [5934] = 2284,
  [5935] = 5671,
  [5936] = 2563,
  [5937] = 2272,
  [5938] = 2503,
  [5939] = 2504,
  [5940] = 2499,
  [5941] = 2525,
  [5942] = 2530,
  [5943] = 2535,
  [5944] = 2256,
  [5945] = 2536,
  [5946] = 2492,
  [5947] = 2313,
  [5948] = 5672,
  [5949] = 2488,
  [5950] = 2506,
  [5951] = 2513,
  [5952] = 2514,
  [5953] = 2552,
  [5954] = 2285,
  [5955] = 2502,
  [5956] = 2530,
  [5957] = 2313,
  [5958] = 2287,
  [5959] = 2556,
  [5960] = 2507,
  [5961] = 2288,
  [5962] = 2289,
  [5963] = 5673,
  [5964] = 2290,
  [5965] = 2558,
  [5966] = 2263,
  [5967] = 2555,
  [5968] = 2240,
  [5969] = 3933,
  [5970] = 2299,
  [5971] = 2556,
  [5972] = 2515,
  [5973] = 2272,
  [5974] = 2507,
  [5975] = 2256,
  [5976] = 2558,
  [5977] = 2487,
  [5978] = 2506,
  [5979] = 2487,
  [5980] = 2537,
  [5981] = 2539,
  [5982] = 2540,
  [5983] = 2572,
  [5984] = 2533,
  [5985] = 2488,
  [5986] = 2485,
  [5987] = 2495,
  [5988] = 2544,
  [5989] = 2506,
  [5990] = 2507,
  [5991] = 2507,
  [5992] = 2529,
  [5993] = 2560,
  [5994] = 2502,
  [5995] = 2545,
  [5996] = 2489,
  [5997] = 2490,
  [5998] = 2508,
  [5999] = 2523,
  [6000] = 2489,
  [6001] = 2491,
  [6002] = 2547,
  [6003] = 2524,
  [6004] = 2535,
  [6005] = 2550,
  [6006] = 2561,
  [6007] = 2510,
  [6008] = 2551,
  [6009] = 2236,
  [6010] = 2234,
  [6011] = 2241,
  [6012] = 2552,
  [6013] = 2562,
  [6014] = 2277,
  [6015] = 2563,
  [6016] = 2555,
  [6017] = 2531,
  [6018] = 2503,
  [6019] = 2491,
  [6020] = 2556,
  [6021] = 2496,
  [6022] = 2558,
  [6023] = 2537,
  [6024] = 2539,
  [6025] = 2560,
  [6026] = 2561,
  [6027] = 2562,
  [6028] = 2563,
  [6029] = 2503,
  [6030] = 2504,
  [6031] = 6031,
  [6032] = 2525,
  [6033] = 4071,
  [6034] = 2530,
  [6035] = 2497,
  [6036] = 2498,
  [6037] = 2535,
  [6038] = 2536,
  [6039] = 2498,
  [6040] = 2490,
  [6041] = 2264,
  [6042] = 2492,
  [6043] = 2533,
  [6044] = 2236,
  [6045] = 2508,
  [6046] = 2485,
  [6047] = 2531,
  [6048] = 2544,
  [6049] = 2540,
  [6050] = 2495,
  [6051] = 2496,
  [6052] = 2234,
  [6053] = 2497,
  [6054] = 2498,
  [6055] = 2560,
  [6056] = 2588,
  [6057] = 2542,
  [6058] = 2512,
  [6059] = 2572,
  [6060] = 2573,
  [6061] = 2574,
  [6062] = 2542,
  [6063] = 2545,
  [6064] = 2534,
  [6065] = 2519,
  [6066] = 2537,
  [6067] = 2539,
  [6068] = 2499,
  [6069] = 2510,
  [6070] = 2512,
  [6071] = 2589,
  [6072] = 2502,
  [6073] = 2491,
  [6074] = 2547,
  [6075] = 2550,
  [6076] = 2520,
  [6077] = 2583,
  [6078] = 2515,
  [6079] = 2551,
  [6080] = 2552,
  [6081] = 2499,
  [6082] = 2555,
  [6083] = 2556,
  [6084] = 2558,
  [6085] = 2560,
  [6086] = 2499,
  [6087] = 2561,
  [6088] = 2562,
  [6089] = 2265,
  [6090] = 2540,
  [6091] = 2513,
  [6092] = 2514,
  [6093] = 2563,
  [6094] = 2503,
  [6095] = 2504,
  [6096] = 2525,
  [6097] = 2572,
  [6098] = 2573,
  [6099] = 2498,
  [6100] = 2286,
  [6101] = 2575,
  [6102] = 2530,
  [6103] = 2574,
  [6104] = 2241,
  [6105] = 2535,
  [6106] = 2536,
  [6107] = 2312,
  [6108] = 2589,
  [6109] = 2506,
  [6110] = 2507,
  [6111] = 2301,
  [6112] = 2515,
  [6113] = 2533,
  [6114] = 2492,
  [6115] = 2508,
  [6116] = 2516,
  [6117] = 2510,
  [6118] = 2512,
  [6119] = 2517,
  [6120] = 2258,
  [6121] = 2538,
  [6122] = 2534,
  [6123] = 2277,
  [6124] = 2283,
  [6125] = 2513,
  [6126] = 2540,
  [6127] = 2537,
  [6128] = 2539,
  [6129] = 2286,
  [6130] = 2514,
  [6131] = 2502,
  [6132] = 2515,
  [6133] = 2516,
  [6134] = 2517,
  [6135] = 2518,
  [6136] = 2485,
  [6137] = 2544,
  [6138] = 2519,
  [6139] = 2520,
  [6140] = 2518,
  [6141] = 2504,
  [6142] = 2519,
  [6143] = 2516,
  [6144] = 2520,
  [6145] = 2259,
  [6146] = 2545,
  [6147] = 2221,
  [6148] = 2582,
  [6149] = 2563,
  [6150] = 2542,
  [6151] = 2503,
  [6152] = 2521,
  [6153] = 2572,
  [6154] = 2589,
  [6155] = 2522,
  [6156] = 2523,
  [6157] = 2573,
  [6158] = 2524,
  [6159] = 2547,
  [6160] = 2583,
  [6161] = 2542,
  [6162] = 2527,
  [6163] = 2528,
  [6164] = 2550,
  [6165] = 2572,
  [6166] = 2521,
  [6167] = 2551,
  [6168] = 2552,
  [6169] = 2555,
  [6170] = 2556,
  [6171] = 2558,
  [6172] = 2560,
  [6173] = 2495,
  [6174] = 2561,
  [6175] = 2562,
  [6176] = 2563,
  [6177] = 2503,
  [6178] = 2504,
  [6179] = 2529,
  [6180] = 2525,
  [6181] = 2530,
  [6182] = 2531,
  [6183] = 2535,
  [6184] = 2522,
  [6185] = 2487,
  [6186] = 2536,
  [6187] = 2574,
  [6188] = 2516,
  [6189] = 2533,
  [6190] = 2488,
  [6191] = 2534,
  [6192] = 2525,
  [6193] = 2517,
  [6194] = 2540,
  [6195] = 2589,
  [6196] = 2537,
  [6197] = 2508,
  [6198] = 2539,
  [6199] = 2514,
  [6200] = 2490,
  [6201] = 2540,
  [6202] = 2491,
  [6203] = 2518,
  [6204] = 2490,
  [6205] = 2492,
  [6206] = 2537,
  [6207] = 2552,
  [6208] = 2538,
  [6209] = 2542,
  [6210] = 2487,
  [6211] = 2572,
  [6212] = 2508,
  [6213] = 2573,
  [6214] = 2574,
  [6215] = 2521,
  [6216] = 2495,
  [6217] = 2496,
  [6218] = 2221,
  [6219] = 2497,
  [6220] = 2498,
  [6221] = 2519,
  [6222] = 2520,
  [6223] = 2530,
  [6224] = 2523,
  [6225] = 2492,
  [6226] = 2524,
  [6227] = 2258,
  [6228] = 2535,
  [6229] = 2496,
  [6230] = 2499,
  [6231] = 2517,
  [6232] = 2522,
  [6233] = 2502,
  [6234] = 2492,
  [6235] = 2506,
  [6236] = 2507,
  [6237] = 2529,
  [6238] = 2518,
  [6239] = 2519,
  [6240] = 2575,
  [6241] = 2536,
  [6242] = 2485,
  [6243] = 2523,
  [6244] = 2508,
  [6245] = 2524,
  [6246] = 2520,
  [6247] = 2575,
  [6248] = 2540,
  [6249] = 2510,
  [6250] = 2512,
  [6251] = 2510,
  [6252] = 2539,
  [6253] = 2588,
  [6254] = 2589,
  [6255] = 2544,
  [6256] = 2583,
  [6257] = 2506,
  [6258] = 2507,
  [6259] = 2575,
  [6260] = 2582,
  [6261] = 2583,
  [6262] = 2521,
  [6263] = 2522,
  [6264] = 2292,
  [6265] = 2263,
  [6266] = 2299,
  [6267] = 2522,
  [6268] = 2508,
  [6269] = 2573,
  [6270] = 2510,
  [6271] = 2512,
  [6272] = 2512,
  [6273] = 2485,
  [6274] = 2513,
  [6275] = 2514,
  [6276] = 2544,
  [6277] = 2510,
  [6278] = 2515,
  [6279] = 2523,
  [6280] = 2516,
  [6281] = 2517,
  [6282] = 2518,
  [6283] = 2519,
  [6284] = 2520,
  [6285] = 2574,
  [6286] = 2534,
  [6287] = 2255,
  [6288] = 2558,
  [6289] = 2495,
  [6290] = 2575,
  [6291] = 2521,
  [6292] = 2522,
  [6293] = 2523,
  [6294] = 2524,
  [6295] = 2524,
  [6296] = 2512,
  [6297] = 2545,
  [6298] = 2527,
  [6299] = 2528,
  [6300] = 2488,
  [6301] = 2545,
  [6302] = 2529,
  [6303] = 2531,
  [6304] = 2527,
  [6305] = 2533,
  [6306] = 2534,
  [6307] = 2489,
  [6308] = 5093,
  [6309] = 5668,
  [6310] = 2528,
  [6311] = 5669,
  [6312] = 2538,
  [6313] = 5671,
  [6314] = 5672,
  [6315] = 2513,
  [6316] = 5673,
  [6317] = 2582,
  [6318] = 2264,
  [6319] = 2259,
  [6320] = 4057,
  [6321] = 2514,
  [6322] = 3933,
  [6323] = 2258,
  [6324] = 2527,
  [6325] = 2528,
  [6326] = 2489,
  [6327] = 2259,
  [6328] = 2531,
  [6329] = 2583,
  [6330] = 2542,
  [6331] = 2529,
  [6332] = 2488,
  [6333] = 2496,
  [6334] = 2531,
  [6335] = 2560,
  [6336] = 2515,
  [6337] = 2529,
  [6338] = 2516,
  [6339] = 2527,
  [6340] = 2531,
  [6341] = 2561,
  [6342] = 2533,
  [6343] = 2572,
  [6344] = 2562,
  [6345] = 2533,
  [6346] = 2534,
  [6347] = 2513,
  [6348] = 2491,
  [6349] = 2517,
  [6350] = 2534,
  [6351] = 2518,
  [6352] = 2519,
  [6353] = 2502,
  [6354] = 2538,
  [6355] = 2573,
  [6356] = 2520,
  [6357] = 2547,
  [6358] = 2550,
  [6359] = 2497,
  [6360] = 2277,
  [6361] = 2236,
  [6362] = 2538,
  [6363] = 2513,
  [6364] = 2574,
  [6365] = 2514,
  [6366] = 2537,
  [6367] = 2234,
  [6368] = 2583,
  [6369] = 2488,
  [6370] = 2539,
  [6371] = 2582,
  [6372] = 2498,
  [6373] = 5669,
  [6374] = 2582,
  [6375] = 2241,
  [6376] = 2551,
  [6377] = 2588,
  [6378] = 2489,
  [6379] = 2538,
  [6380] = 5668,
  [6381] = 2497,
  [6382] = 2495,
  [6383] = 2521,
  [6384] = 2575,
  [6385] = 2522,
  [6386] = 2542,
  [6387] = 2523,
  [6388] = 2264,
  [6389] = 2485,
  [6390] = 2544,
  [6391] = 2552,
  [6392] = 2555,
  [6393] = 2556,
  [6394] = 2558,
  [6395] = 2560,
  [6396] = 2524,
  [6397] = 2561,
  [6398] = 2490,
  [6399] = 5669,
  [6400] = 2562,
  [6401] = 2563,
  [6402] = 2551,
  [6403] = 5671,
  [6404] = 5672,
  [6405] = 2536,
  [6406] = 2545,
  [6407] = 2221,
  [6408] = 5673,
  [6409] = 2527,
  [6410] = 2487,
  [6411] = 2588,
  [6412] = 2485,
  [6413] = 2265,
  [6414] = 2573,
  [6415] = 2503,
  [6416] = 2255,
  [6417] = 2538,
  [6418] = 2574,
  [6419] = 2544,
  [6420] = 2528,
  [6421] = 2555,
  [6422] = 2504,
  [6423] = 2265,
  [6424] = 2286,
  [6425] = 2525,
  [6426] = 2530,
  [6427] = 2487,
  [6428] = 2535,
  [6429] = 2536,
  [6430] = 2547,
  [6431] = 5668,
  [6432] = 2515,
  [6433] = 2490,
  [6434] = 2550,
  [6435] = 2488,
  [6436] = 2240,
  [6437] = 2527,
  [6438] = 2489,
  [6439] = 2528,
  [6440] = 2582,
  [6441] = 2529,
  [6442] = 2490,
  [6443] = 2256,
  [6444] = 2276,
  [6445] = 6445,
  [6446] = 2276,
  [6447] = 2313,
  [6448] = 2287,
  [6449] = 2258,
  [6450] = 6450,
  [6451] = 6451,
  [6452] = 6445,
  [6453] = 6445,
  [6454] = 2255,
  [6455] = 2272,
  [6456] = 6031,
  [6457] = 2288,
  [6458] = 6458,
  [6459] = 2277,
  [6460] = 2272,
  [6461] = 6031,
  [6462] = 2264,
  [6463] = 2259,
  [6464] = 2265,
  [6465] = 2288,
  [6466] = 2221,
  [6467] = 2265,
  [6468] = 2256,
  [6469] = 2316,
  [6470] = 2290,
  [6471] = 2281,
  [6472] = 2312,
  [6473] = 2301,
  [6474] = 2283,
  [6475] = 2281,
  [6476] = 2282,
  [6477] = 2282,
  [6478] = 2312,
  [6479] = 2282,
  [6480] = 6445,
  [6481] = 2255,
  [6482] = 2234,
  [6483] = 2292,
  [6484] = 2236,
  [6485] = 2289,
  [6486] = 2286,
  [6487] = 2316,
  [6488] = 2284,
  [6489] = 2299,
  [6490] = 6445,
  [6491] = 2272,
  [6492] = 2284,
  [6493] = 2285,
  [6494] = 6445,
  [6495] = 2286,
  [6496] = 2278,
  [6497] = 2301,
  [6498] = 2292,
  [6499] = 2289,
  [6500] = 2276,
  [6501] = 2225,
  [6502] = 6445,
  [6503] = 2283,
  [6504] = 2277,
  [6505] = 2284,
  [6506] = 5668,
  [6507] = 2290,
  [6508] = 2285,
  [6509] = 2287,
  [6510] = 2241,
  [6511] = 2312,
  [6512] = 5669,
  [6513] = 2256,
  [6514] = 6445,
  [6515] = 2313,
  [6516] = 6031,
  [6517] = 6445,
  [6518] = 2285,
  [6519] = 2278,
  [6520] = 2316,
  [6521] = 2299,
  [6522] = 2301,
  [6523] = 2288,
  [6524] = 2264,
  [6525] = 6445,
  [6526] = 5671,
  [6527] = 6445,
  [6528] = 5672,
  [6529] = 6445,
  [6530] = 2287,
  [6531] = 2289,
  [6532] = 2281,
  [6533] = 6445,
  [6534] = 2255,
  [6535] = 2292,
  [6536] = 2263,
  [6537] = 2240,
  [6538] = 2290,
  [6539] = 6539,
  [6540] = 5673,
  [6541] = 2299,
  [6542] = 2221,
  [6543] = 2278,
  [6544] = 2283,
  [6545] = 6458,
  [6546] = 2225,
  [6547] = 2225,
  [6548] = 6450,
  [6549] = 6451,
  [6550] = 6031,
  [6551] = 6450,
  [6552] = 2225,
  [6553] = 6451,
  [6554] = 2277,
  [6555] = 2264,
  [6556] = 2265,
  [6557] = 6557,
  [6558] = 2286,
  [6559] = 6559,
  [6560] = 6560,
  [6561] = 6458,
  [6562] = 2277,
  [6563] = 2264,
  [6564] = 2265,
  [6565] = 2316,
  [6566] = 2278,
  [6567] = 2276,
  [6568] = 2281,
  [6569] = 2282,
  [6570] = 2284,
  [6571] = 2285,
  [6572] = 2287,
  [6573] = 2288,
  [6574] = 2289,
  [6575] = 2290,
  [6576] = 2313,
  [6577] = 6458,
  [6578] = 2312,
  [6579] = 2301,
  [6580] = 2283,
  [6581] = 2316,
  [6582] = 6458,
  [6583] = 2276,
  [6584] = 2281,
  [6585] = 2282,
  [6586] = 2284,
  [6587] = 2285,
  [6588] = 2287,
  [6589] = 2288,
  [6590] = 2289,
  [6591] = 2290,
  [6592] = 2277,
  [6593] = 2264,
  [6594] = 2265,
  [6595] = 2312,
  [6596] = 2301,
  [6597] = 2283,
  [6598] = 2286,
  [6599] = 6599,
  [6600] = 6600,
  [6601] = 2292,
  [6602] = 2299,
  [6603] = 2286,
  [6604] = 2255,
  [6605] = 2313,
  [6606] = 2272,
  [6607] = 2256,
  [6608] = 2313,
  [6609] = 6609,
  [6610] = 6539,
  [6611] = 6539,
  [6612] = 6450,
  [6613] = 6451,
  [6614] = 6458,
  [6615] = 6539,
  [6616] = 6609,
  [6617] = 2278,
  [6618] = 6599,
  [6619] = 2301,
  [6620] = 2283,
  [6621] = 2301,
  [6622] = 2283,
  [6623] = 6609,
  [6624] = 6450,
  [6625] = 6625,
  [6626] = 6451,
  [6627] = 6539,
  [6628] = 6609,
  [6629] = 6557,
  [6630] = 2316,
  [6631] = 6631,
  [6632] = 2278,
  [6633] = 6559,
  [6634] = 2277,
  [6635] = 2264,
  [6636] = 2265,
  [6637] = 2316,
  [6638] = 2276,
  [6639] = 2278,
  [6640] = 2276,
  [6641] = 2281,
  [6642] = 2282,
  [6643] = 2284,
  [6644] = 2285,
  [6645] = 2281,
  [6646] = 6560,
  [6647] = 2287,
  [6648] = 2288,
  [6649] = 2289,
  [6650] = 2290,
  [6651] = 2282,
  [6652] = 6652,
  [6653] = 2316,
  [6654] = 2284,
  [6655] = 2285,
  [6656] = 6458,
  [6657] = 2287,
  [6658] = 6658,
  [6659] = 2312,
  [6660] = 2301,
  [6661] = 2283,
  [6662] = 2312,
  [6663] = 6663,
  [6664] = 2286,
  [6665] = 2288,
  [6666] = 2289,
  [6667] = 2290,
  [6668] = 2278,
  [6669] = 2276,
  [6670] = 2281,
  [6671] = 2313,
  [6672] = 2282,
  [6673] = 6458,
  [6674] = 2284,
  [6675] = 6539,
  [6676] = 6458,
  [6677] = 2285,
  [6678] = 6600,
  [6679] = 2225,
  [6680] = 6458,
  [6681] = 6557,
  [6682] = 6599,
  [6683] = 6600,
  [6684] = 2287,
  [6685] = 2288,
  [6686] = 6609,
  [6687] = 2289,
  [6688] = 6559,
  [6689] = 6557,
  [6690] = 6559,
  [6691] = 6560,
  [6692] = 2225,
  [6693] = 2290,
  [6694] = 6560,
  [6695] = 6609,
  [6696] = 6599,
  [6697] = 6600,
  [6698] = 2312,
  [6699] = 6699,
  [6700] = 6700,
  [6701] = 6701,
  [6702] = 6702,
  [6703] = 6609,
  [6704] = 6699,
  [6705] = 2278,
  [6706] = 6700,
  [6707] = 6701,
  [6708] = 6702,
  [6709] = 6599,
  [6710] = 6600,
  [6711] = 6700,
  [6712] = 6701,
  [6713] = 6702,
  [6714] = 6700,
  [6715] = 6700,
  [6716] = 6701,
  [6717] = 6702,
  [6718] = 6663,
  [6719] = 6701,
  [6720] = 6702,
  [6721] = 6700,
  [6722] = 6701,
  [6723] = 6702,
  [6724] = 6700,
  [6725] = 6701,
  [6726] = 6702,
  [6727] = 6700,
  [6728] = 6701,
  [6729] = 6702,
  [6730] = 6700,
  [6731] = 6701,
  [6732] = 6702,
  [6733] = 6700,
  [6734] = 6701,
  [6735] = 6702,
  [6736] = 6700,
  [6737] = 6701,
  [6738] = 6702,
  [6739] = 6700,
  [6740] = 6701,
  [6741] = 6702,
  [6742] = 6700,
  [6743] = 6701,
  [6744] = 6609,
  [6745] = 6700,
  [6746] = 6701,
  [6747] = 6700,
  [6748] = 6701,
  [6749] = 6700,
  [6750] = 6701,
  [6751] = 6700,
  [6752] = 6701,
  [6753] = 2276,
  [6754] = 6458,
  [6755] = 6700,
  [6756] = 6609,
  [6757] = 2312,
  [6758] = 6700,
  [6759] = 6701,
  [6760] = 6631,
  [6761] = 6700,
  [6762] = 6701,
  [6763] = 6700,
  [6764] = 6701,
  [6765] = 6700,
  [6766] = 6701,
  [6767] = 6700,
  [6768] = 6701,
  [6769] = 6652,
  [6770] = 6699,
  [6771] = 2301,
  [6772] = 6658,
  [6773] = 6699,
  [6774] = 6663,
  [6775] = 2283,
  [6776] = 6609,
  [6777] = 6559,
  [6778] = 6652,
  [6779] = 2281,
  [6780] = 2282,
  [6781] = 6625,
  [6782] = 6658,
  [6783] = 6783,
  [6784] = 6784,
  [6785] = 6557,
  [6786] = 2284,
  [6787] = 6458,
  [6788] = 6701,
  [6789] = 2285,
  [6790] = 6663,
  [6791] = 6560,
  [6792] = 6631,
  [6793] = 2287,
  [6794] = 6625,
  [6795] = 6795,
  [6796] = 2288,
  [6797] = 2289,
  [6798] = 6625,
  [6799] = 2290,
  [6800] = 6631,
  [6801] = 6652,
  [6802] = 2316,
  [6803] = 6702,
  [6804] = 6658,
  [6805] = 6700,
  [6806] = 6701,
  [6807] = 6807,
  [6808] = 6808,
  [6809] = 6808,
  [6810] = 6810,
  [6811] = 6808,
  [6812] = 6808,
  [6813] = 6808,
  [6814] = 6808,
  [6815] = 2367,
  [6816] = 6808,
  [6817] = 6808,
  [6818] = 6808,
  [6819] = 6819,
  [6820] = 6808,
  [6821] = 6808,
  [6822] = 6808,
  [6823] = 6783,
  [6824] = 6808,
  [6825] = 6795,
  [6826] = 6808,
  [6827] = 6808,
  [6828] = 6808,
  [6829] = 6783,
  [6830] = 6631,
  [6831] = 6784,
  [6832] = 6808,
  [6833] = 6783,
  [6834] = 6784,
  [6835] = 6835,
  [6836] = 6652,
  [6837] = 6658,
  [6838] = 6795,
  [6839] = 6808,
  [6840] = 6609,
  [6841] = 6625,
  [6842] = 6609,
  [6843] = 6699,
  [6844] = 6784,
  [6845] = 6808,
  [6846] = 6663,
  [6847] = 6808,
  [6848] = 6848,
  [6849] = 6849,
  [6850] = 6850,
  [6851] = 6851,
  [6852] = 6852,
  [6853] = 6853,
  [6854] = 6854,
  [6855] = 6855,
  [6856] = 6856,
  [6857] = 6857,
  [6858] = 6858,
  [6859] = 6859,
  [6860] = 6860,
  [6861] = 6808,
  [6862] = 6808,
  [6863] = 6795,
  [6864] = 6864,
  [6865] = 6859,
  [6866] = 6819,
  [6867] = 6835,
  [6868] = 6852,
  [6869] = 2367,
  [6870] = 6848,
  [6871] = 6871,
  [6872] = 6872,
  [6873] = 6858,
  [6874] = 6848,
  [6875] = 6849,
  [6876] = 6850,
  [6877] = 6851,
  [6878] = 6852,
  [6879] = 6853,
  [6880] = 6854,
  [6881] = 6855,
  [6882] = 6856,
  [6883] = 6857,
  [6884] = 6858,
  [6885] = 6859,
  [6886] = 6886,
  [6887] = 6807,
  [6888] = 6860,
  [6889] = 6853,
  [6890] = 6890,
  [6891] = 6891,
  [6892] = 6892,
  [6893] = 6893,
  [6894] = 6807,
  [6895] = 6895,
  [6896] = 6896,
  [6897] = 6897,
  [6898] = 6898,
  [6899] = 6859,
  [6900] = 6860,
  [6901] = 6856,
  [6902] = 6902,
  [6903] = 6903,
  [6904] = 6835,
  [6905] = 6807,
  [6906] = 6810,
  [6907] = 6907,
  [6908] = 6908,
  [6909] = 6909,
  [6910] = 6910,
  [6911] = 6911,
  [6912] = 6848,
  [6913] = 6783,
  [6914] = 6849,
  [6915] = 6854,
  [6916] = 6860,
  [6917] = 6784,
  [6918] = 6918,
  [6919] = 6919,
  [6920] = 6857,
  [6921] = 6921,
  [6922] = 6922,
  [6923] = 6850,
  [6924] = 6849,
  [6925] = 6925,
  [6926] = 6926,
  [6927] = 6927,
  [6928] = 6928,
  [6929] = 6858,
  [6930] = 6851,
  [6931] = 6931,
  [6932] = 6932,
  [6933] = 6810,
  [6934] = 6819,
  [6935] = 6935,
  [6936] = 6936,
  [6937] = 6937,
  [6938] = 6938,
  [6939] = 6855,
  [6940] = 2367,
  [6941] = 6941,
  [6942] = 6942,
  [6943] = 6943,
  [6944] = 6944,
  [6945] = 6945,
  [6946] = 6856,
  [6947] = 2367,
  [6948] = 6948,
  [6949] = 6949,
  [6950] = 2367,
  [6951] = 6850,
  [6952] = 6851,
  [6953] = 6953,
  [6954] = 6954,
  [6955] = 6955,
  [6956] = 6956,
  [6957] = 6957,
  [6958] = 6958,
  [6959] = 6795,
  [6960] = 6835,
  [6961] = 6961,
  [6962] = 6962,
  [6963] = 6852,
  [6964] = 6853,
  [6965] = 6857,
  [6966] = 6854,
  [6967] = 6967,
  [6968] = 6855,
  [6969] = 6819,
  [6970] = 6810,
  [6971] = 6971,
  [6972] = 6936,
  [6973] = 6893,
  [6974] = 6921,
  [6975] = 6896,
  [6976] = 6897,
  [6977] = 6908,
  [6978] = 6938,
  [6979] = 6864,
  [6980] = 6902,
  [6981] = 6903,
  [6982] = 6903,
  [6983] = 6937,
  [6984] = 6872,
  [6985] = 6918,
  [6986] = 6932,
  [6987] = 6902,
  [6988] = 6871,
  [6989] = 6907,
  [6990] = 6908,
  [6991] = 6891,
  [6992] = 6909,
  [6993] = 6909,
  [6994] = 6910,
  [6995] = 6910,
  [6996] = 6892,
  [6997] = 6931,
  [6998] = 6922,
  [6999] = 6893,
  [7000] = 6925,
  [7001] = 6926,
  [7002] = 6911,
  [7003] = 6927,
  [7004] = 6928,
  [7005] = 6896,
  [7006] = 6931,
  [7007] = 6932,
  [7008] = 6835,
  [7009] = 6971,
  [7010] = 2367,
  [7011] = 6897,
  [7012] = 6921,
  [7013] = 6935,
  [7014] = 2367,
  [7015] = 6810,
  [7016] = 6936,
  [7017] = 6937,
  [7018] = 6938,
  [7019] = 6819,
  [7020] = 6872,
  [7021] = 6941,
  [7022] = 6941,
  [7023] = 6942,
  [7024] = 6943,
  [7025] = 6944,
  [7026] = 6945,
  [7027] = 6942,
  [7028] = 6948,
  [7029] = 6949,
  [7030] = 6943,
  [7031] = 6864,
  [7032] = 6944,
  [7033] = 6902,
  [7034] = 6903,
  [7035] = 6954,
  [7036] = 6955,
  [7037] = 6956,
  [7038] = 6945,
  [7039] = 6957,
  [7040] = 6958,
  [7041] = 6919,
  [7042] = 6848,
  [7043] = 6961,
  [7044] = 6962,
  [7045] = 6918,
  [7046] = 6911,
  [7047] = 6948,
  [7048] = 6949,
  [7049] = 6849,
  [7050] = 6871,
  [7051] = 6895,
  [7052] = 6922,
  [7053] = 6907,
  [7054] = 6895,
  [7055] = 6850,
  [7056] = 6971,
  [7057] = 6919,
  [7058] = 6925,
  [7059] = 6926,
  [7060] = 6927,
  [7061] = 6928,
  [7062] = 6907,
  [7063] = 6921,
  [7064] = 6953,
  [7065] = 6932,
  [7066] = 6908,
  [7067] = 6954,
  [7068] = 6935,
  [7069] = 6936,
  [7070] = 6937,
  [7071] = 6955,
  [7072] = 6956,
  [7073] = 6957,
  [7074] = 6941,
  [7075] = 6942,
  [7076] = 6943,
  [7077] = 6944,
  [7078] = 6945,
  [7079] = 6958,
  [7080] = 6938,
  [7081] = 6948,
  [7082] = 6949,
  [7083] = 6872,
  [7084] = 6961,
  [7085] = 6962,
  [7086] = 6954,
  [7087] = 6955,
  [7088] = 6956,
  [7089] = 6957,
  [7090] = 6958,
  [7091] = 2367,
  [7092] = 6961,
  [7093] = 6962,
  [7094] = 6898,
  [7095] = 6967,
  [7096] = 6851,
  [7097] = 6852,
  [7098] = 6909,
  [7099] = 6910,
  [7100] = 6853,
  [7101] = 6886,
  [7102] = 6890,
  [7103] = 6854,
  [7104] = 6855,
  [7105] = 6911,
  [7106] = 6891,
  [7107] = 6892,
  [7108] = 6893,
  [7109] = 6918,
  [7110] = 6953,
  [7111] = 6856,
  [7112] = 6857,
  [7113] = 6858,
  [7114] = 6896,
  [7115] = 6922,
  [7116] = 6897,
  [7117] = 6859,
  [7118] = 6890,
  [7119] = 6898,
  [7120] = 6928,
  [7121] = 6919,
  [7122] = 6886,
  [7123] = 6935,
  [7124] = 6967,
  [7125] = 6871,
  [7126] = 6953,
  [7127] = 6895,
  [7128] = 6890,
  [7129] = 2367,
  [7130] = 6807,
  [7131] = 6860,
  [7132] = 6971,
  [7133] = 6886,
  [7134] = 6898,
  [7135] = 6967,
  [7136] = 6925,
  [7137] = 6926,
  [7138] = 6927,
  [7139] = 6864,
  [7140] = 6891,
  [7141] = 6892,
  [7142] = 6931,
  [7143] = 6967,
  [7144] = 6932,
  [7145] = 6922,
  [7146] = 6918,
  [7147] = 6961,
  [7148] = 2221,
  [7149] = 2258,
  [7150] = 6935,
  [7151] = 2234,
  [7152] = 6886,
  [7153] = 6962,
  [7154] = 6936,
  [7155] = 6937,
  [7156] = 4923,
  [7157] = 4924,
  [7158] = 6890,
  [7159] = 6925,
  [7160] = 6898,
  [7161] = 4942,
  [7162] = 6941,
  [7163] = 6942,
  [7164] = 6943,
  [7165] = 6944,
  [7166] = 6945,
  [7167] = 2241,
  [7168] = 6948,
  [7169] = 6949,
  [7170] = 2236,
  [7171] = 6931,
  [7172] = 6926,
  [7173] = 6891,
  [7174] = 6892,
  [7175] = 6893,
  [7176] = 6895,
  [7177] = 6954,
  [7178] = 6953,
  [7179] = 6927,
  [7180] = 6896,
  [7181] = 6955,
  [7182] = 6897,
  [7183] = 6956,
  [7184] = 6919,
  [7185] = 6971,
  [7186] = 6864,
  [7187] = 6921,
  [7188] = 6902,
  [7189] = 6903,
  [7190] = 6938,
  [7191] = 6872,
  [7192] = 6907,
  [7193] = 6908,
  [7194] = 6957,
  [7195] = 6958,
  [7196] = 6909,
  [7197] = 6910,
  [7198] = 6911,
  [7199] = 6928,
  [7200] = 2259,
  [7201] = 2263,
  [7202] = 2240,
  [7203] = 2367,
  [7204] = 6871,
  [7205] = 7205,
  [7206] = 7206,
  [7207] = 7207,
  [7208] = 7208,
  [7209] = 7209,
  [7210] = 7209,
  [7211] = 2292,
  [7212] = 7207,
  [7213] = 7206,
  [7214] = 7208,
  [7215] = 7207,
  [7216] = 7208,
  [7217] = 7209,
  [7218] = 7209,
  [7219] = 7207,
  [7220] = 7220,
  [7221] = 7208,
  [7222] = 7208,
  [7223] = 7209,
  [7224] = 4923,
  [7225] = 7206,
  [7226] = 7206,
  [7227] = 7227,
  [7228] = 7209,
  [7229] = 7205,
  [7230] = 7207,
  [7231] = 7208,
  [7232] = 7209,
  [7233] = 4924,
  [7234] = 7207,
  [7235] = 7206,
  [7236] = 7207,
  [7237] = 7220,
  [7238] = 7208,
  [7239] = 7227,
  [7240] = 7205,
  [7241] = 7220,
  [7242] = 7227,
  [7243] = 7205,
  [7244] = 7209,
  [7245] = 7220,
  [7246] = 7227,
  [7247] = 7205,
  [7248] = 7220,
  [7249] = 7227,
  [7250] = 7205,
  [7251] = 7220,
  [7252] = 7227,
  [7253] = 7205,
  [7254] = 7220,
  [7255] = 2258,
  [7256] = 7227,
  [7257] = 2236,
  [7258] = 7205,
  [7259] = 7206,
  [7260] = 7220,
  [7261] = 7227,
  [7262] = 7205,
  [7263] = 4923,
  [7264] = 4924,
  [7265] = 7220,
  [7266] = 7227,
  [7267] = 7205,
  [7268] = 7220,
  [7269] = 7207,
  [7270] = 7227,
  [7271] = 7205,
  [7272] = 4942,
  [7273] = 7220,
  [7274] = 7227,
  [7275] = 7205,
  [7276] = 7208,
  [7277] = 7207,
  [7278] = 7220,
  [7279] = 7206,
  [7280] = 7208,
  [7281] = 7227,
  [7282] = 7205,
  [7283] = 7207,
  [7284] = 7220,
  [7285] = 4923,
  [7286] = 7227,
  [7287] = 7205,
  [7288] = 7220,
  [7289] = 7227,
  [7290] = 7205,
  [7291] = 7208,
  [7292] = 7209,
  [7293] = 7209,
  [7294] = 7206,
  [7295] = 7220,
  [7296] = 7227,
  [7297] = 7205,
  [7298] = 7208,
  [7299] = 7220,
  [7300] = 2255,
  [7301] = 7209,
  [7302] = 7227,
  [7303] = 7205,
  [7304] = 2234,
  [7305] = 7209,
  [7306] = 7220,
  [7307] = 2259,
  [7308] = 7206,
  [7309] = 7227,
  [7310] = 2258,
  [7311] = 2259,
  [7312] = 7207,
  [7313] = 7220,
  [7314] = 7208,
  [7315] = 7227,
  [7316] = 7205,
  [7317] = 7220,
  [7318] = 7227,
  [7319] = 7205,
  [7320] = 7209,
  [7321] = 2263,
  [7322] = 7220,
  [7323] = 7227,
  [7324] = 7205,
  [7325] = 7206,
  [7326] = 7220,
  [7327] = 7227,
  [7328] = 7205,
  [7329] = 7220,
  [7330] = 2263,
  [7331] = 7227,
  [7332] = 7205,
  [7333] = 7220,
  [7334] = 7227,
  [7335] = 7205,
  [7336] = 7220,
  [7337] = 7227,
  [7338] = 7205,
  [7339] = 7220,
  [7340] = 7227,
  [7341] = 7205,
  [7342] = 7206,
  [7343] = 7220,
  [7344] = 7227,
  [7345] = 7205,
  [7346] = 7207,
  [7347] = 7208,
  [7348] = 7206,
  [7349] = 7206,
  [7350] = 7209,
  [7351] = 7207,
  [7352] = 7208,
  [7353] = 7209,
  [7354] = 7209,
  [7355] = 7205,
  [7356] = 2241,
  [7357] = 7206,
  [7358] = 7207,
  [7359] = 7208,
  [7360] = 7209,
  [7361] = 7206,
  [7362] = 2258,
  [7363] = 7207,
  [7364] = 7208,
  [7365] = 7209,
  [7366] = 7206,
  [7367] = 7207,
  [7368] = 2272,
  [7369] = 7208,
  [7370] = 2256,
  [7371] = 7209,
  [7372] = 2259,
  [7373] = 7207,
  [7374] = 7208,
  [7375] = 7209,
  [7376] = 7220,
  [7377] = 7206,
  [7378] = 7206,
  [7379] = 7208,
  [7380] = 4942,
  [7381] = 7227,
  [7382] = 2299,
  [7383] = 2236,
  [7384] = 7206,
  [7385] = 4942,
  [7386] = 7207,
  [7387] = 2234,
  [7388] = 7207,
  [7389] = 2263,
  [7390] = 7208,
  [7391] = 7209,
  [7392] = 7206,
  [7393] = 2240,
  [7394] = 7207,
  [7395] = 7208,
  [7396] = 7208,
  [7397] = 7209,
  [7398] = 2236,
  [7399] = 7209,
  [7400] = 7206,
  [7401] = 2234,
  [7402] = 7207,
  [7403] = 2241,
  [7404] = 2241,
  [7405] = 7208,
  [7406] = 7206,
  [7407] = 7209,
  [7408] = 7207,
  [7409] = 7208,
  [7410] = 7209,
  [7411] = 7207,
  [7412] = 2240,
  [7413] = 7207,
  [7414] = 7206,
  [7415] = 7208,
  [7416] = 7207,
  [7417] = 7208,
  [7418] = 7209,
  [7419] = 4924,
  [7420] = 7206,
  [7421] = 7206,
  [7422] = 2240,
  [7423] = 7206,
  [7424] = 7207,
  [7425] = 7207,
  [7426] = 7208,
  [7427] = 7209,
  [7428] = 7206,
  [7429] = 4923,
  [7430] = 7430,
  [7431] = 2225,
  [7432] = 7432,
  [7433] = 2328,
  [7434] = 2277,
  [7435] = 2264,
  [7436] = 2265,
  [7437] = 2292,
  [7438] = 2299,
  [7439] = 2263,
  [7440] = 2240,
  [7441] = 2236,
  [7442] = 2234,
  [7443] = 2241,
  [7444] = 7444,
  [7445] = 2286,
  [7446] = 7444,
  [7447] = 7447,
  [7448] = 7444,
  [7449] = 7447,
  [7450] = 7444,
  [7451] = 7447,
  [7452] = 7444,
  [7453] = 7447,
  [7454] = 7444,
  [7455] = 7447,
  [7456] = 7444,
  [7457] = 7447,
  [7458] = 2255,
  [7459] = 7444,
  [7460] = 7447,
  [7461] = 7444,
  [7462] = 7447,
  [7463] = 7444,
  [7464] = 7447,
  [7465] = 7465,
  [7466] = 7447,
  [7467] = 7444,
  [7468] = 7447,
  [7469] = 2258,
  [7470] = 7444,
  [7471] = 7447,
  [7472] = 2259,
  [7473] = 4942,
  [7474] = 2272,
  [7475] = 2292,
  [7476] = 2299,
  [7477] = 7447,
  [7478] = 7478,
  [7479] = 2256,
  [7480] = 7480,
  [7481] = 2255,
  [7482] = 7482,
  [7483] = 4924,
  [7484] = 2255,
  [7485] = 2272,
  [7486] = 2256,
  [7487] = 2292,
  [7488] = 2313,
  [7489] = 7489,
  [7490] = 2272,
  [7491] = 2256,
  [7492] = 2299,
  [7493] = 7444,
  [7494] = 7494,
  [7495] = 2278,
  [7496] = 2276,
  [7497] = 7465,
  [7498] = 2281,
  [7499] = 2282,
  [7500] = 7500,
  [7501] = 7494,
  [7502] = 2284,
  [7503] = 2285,
  [7504] = 7482,
  [7505] = 7505,
  [7506] = 2287,
  [7507] = 7507,
  [7508] = 2288,
  [7509] = 2289,
  [7510] = 2290,
  [7511] = 7494,
  [7512] = 7505,
  [7513] = 7507,
  [7514] = 7494,
  [7515] = 7430,
  [7516] = 2286,
  [7517] = 7505,
  [7518] = 7507,
  [7519] = 7432,
  [7520] = 7489,
  [7521] = 7494,
  [7522] = 7505,
  [7523] = 7507,
  [7524] = 2312,
  [7525] = 7494,
  [7526] = 2301,
  [7527] = 2283,
  [7528] = 7505,
  [7529] = 7507,
  [7530] = 7465,
  [7531] = 7494,
  [7532] = 7432,
  [7533] = 7505,
  [7534] = 7507,
  [7535] = 7494,
  [7536] = 7505,
  [7537] = 7537,
  [7538] = 7507,
  [7539] = 7494,
  [7540] = 7505,
  [7541] = 7541,
  [7542] = 7430,
  [7543] = 7507,
  [7544] = 7494,
  [7545] = 7505,
  [7546] = 7494,
  [7547] = 7505,
  [7548] = 2277,
  [7549] = 7507,
  [7550] = 7494,
  [7551] = 2264,
  [7552] = 7505,
  [7553] = 2265,
  [7554] = 7507,
  [7555] = 7494,
  [7556] = 7505,
  [7557] = 7507,
  [7558] = 7494,
  [7559] = 7505,
  [7560] = 7507,
  [7561] = 7432,
  [7562] = 7505,
  [7563] = 7507,
  [7564] = 7494,
  [7565] = 7505,
  [7566] = 7507,
  [7567] = 7567,
  [7568] = 7507,
  [7569] = 7494,
  [7570] = 7570,
  [7571] = 7537,
  [7572] = 7482,
  [7573] = 7494,
  [7574] = 7570,
  [7575] = 7537,
  [7576] = 7505,
  [7577] = 7507,
  [7578] = 7570,
  [7579] = 7537,
  [7580] = 7505,
  [7581] = 2369,
  [7582] = 7570,
  [7583] = 7537,
  [7584] = 7494,
  [7585] = 7505,
  [7586] = 7570,
  [7587] = 7537,
  [7588] = 7507,
  [7589] = 7494,
  [7590] = 7507,
  [7591] = 7537,
  [7592] = 2328,
  [7593] = 7505,
  [7594] = 7570,
  [7595] = 7537,
  [7596] = 7507,
  [7597] = 7505,
  [7598] = 7494,
  [7599] = 7570,
  [7600] = 7505,
  [7601] = 7507,
  [7602] = 7570,
  [7603] = 7603,
  [7604] = 7489,
  [7605] = 7494,
  [7606] = 7505,
  [7607] = 7507,
  [7608] = 7494,
  [7609] = 2373,
  [7610] = 2292,
  [7611] = 7505,
  [7612] = 2299,
  [7613] = 7507,
  [7614] = 2277,
  [7615] = 7494,
  [7616] = 2264,
  [7617] = 7505,
  [7618] = 7618,
  [7619] = 2286,
  [7620] = 2265,
  [7621] = 7507,
  [7622] = 7494,
  [7623] = 7505,
  [7624] = 2313,
  [7625] = 7625,
  [7626] = 2328,
  [7627] = 2277,
  [7628] = 7507,
  [7629] = 7570,
  [7630] = 2383,
  [7631] = 7494,
  [7632] = 7494,
  [7633] = 7567,
  [7634] = 2328,
  [7635] = 7567,
  [7636] = 2353,
  [7637] = 7567,
  [7638] = 7567,
  [7639] = 2264,
  [7640] = 7567,
  [7641] = 7567,
  [7642] = 7567,
  [7643] = 7567,
  [7644] = 7567,
  [7645] = 7567,
  [7646] = 7567,
  [7647] = 7567,
  [7648] = 2255,
  [7649] = 7507,
  [7650] = 7505,
  [7651] = 7507,
  [7652] = 2313,
  [7653] = 7494,
  [7654] = 2286,
  [7655] = 2265,
  [7656] = 7656,
  [7657] = 2316,
  [7658] = 7507,
  [7659] = 7505,
  [7660] = 2272,
  [7661] = 7482,
  [7662] = 7489,
  [7663] = 7465,
  [7664] = 2256,
  [7665] = 7430,
  [7666] = 7507,
  [7667] = 2313,
  [7668] = 7494,
  [7669] = 7505,
  [7670] = 7570,
  [7671] = 7671,
  [7672] = 2282,
  [7673] = 7673,
  [7674] = 2284,
  [7675] = 2285,
  [7676] = 7673,
  [7677] = 2287,
  [7678] = 2288,
  [7679] = 2289,
  [7680] = 7671,
  [7681] = 2290,
  [7682] = 2289,
  [7683] = 7683,
  [7684] = 2312,
  [7685] = 2301,
  [7686] = 2283,
  [7687] = 7671,
  [7688] = 7671,
  [7689] = 7683,
  [7690] = 7671,
  [7691] = 7432,
  [7692] = 7692,
  [7693] = 7671,
  [7694] = 7673,
  [7695] = 7683,
  [7696] = 7673,
  [7697] = 7671,
  [7698] = 7683,
  [7699] = 7673,
  [7700] = 7683,
  [7701] = 7671,
  [7702] = 7702,
  [7703] = 7703,
  [7704] = 7671,
  [7705] = 7705,
  [7706] = 7706,
  [7707] = 2255,
  [7708] = 7430,
  [7709] = 7671,
  [7710] = 7673,
  [7711] = 7711,
  [7712] = 7702,
  [7713] = 7703,
  [7714] = 7705,
  [7715] = 7706,
  [7716] = 7683,
  [7717] = 7671,
  [7718] = 7702,
  [7719] = 7703,
  [7720] = 7482,
  [7721] = 7705,
  [7722] = 7706,
  [7723] = 7702,
  [7724] = 7671,
  [7725] = 7673,
  [7726] = 2277,
  [7727] = 7671,
  [7728] = 2264,
  [7729] = 7671,
  [7730] = 7465,
  [7731] = 2265,
  [7732] = 7671,
  [7733] = 7671,
  [7734] = 7671,
  [7735] = 7432,
  [7736] = 2316,
  [7737] = 7671,
  [7738] = 7702,
  [7739] = 2278,
  [7740] = 2276,
  [7741] = 2281,
  [7742] = 7671,
  [7743] = 2282,
  [7744] = 7702,
  [7745] = 2284,
  [7746] = 2285,
  [7747] = 7683,
  [7748] = 2287,
  [7749] = 7702,
  [7750] = 7671,
  [7751] = 7703,
  [7752] = 2288,
  [7753] = 2289,
  [7754] = 2290,
  [7755] = 7703,
  [7756] = 7671,
  [7757] = 7705,
  [7758] = 7705,
  [7759] = 7706,
  [7760] = 7671,
  [7761] = 7671,
  [7762] = 7671,
  [7763] = 7671,
  [7764] = 7671,
  [7765] = 7673,
  [7766] = 7671,
  [7767] = 7671,
  [7768] = 7683,
  [7769] = 7671,
  [7770] = 7671,
  [7771] = 7671,
  [7772] = 7671,
  [7773] = 7671,
  [7774] = 2312,
  [7775] = 7671,
  [7776] = 7671,
  [7777] = 7671,
  [7778] = 7671,
  [7779] = 7671,
  [7780] = 7671,
  [7781] = 7671,
  [7782] = 7671,
  [7783] = 7671,
  [7784] = 7671,
  [7785] = 7671,
  [7786] = 7671,
  [7787] = 7671,
  [7788] = 7671,
  [7789] = 7671,
  [7790] = 7671,
  [7791] = 7671,
  [7792] = 7671,
  [7793] = 7671,
  [7794] = 7671,
  [7795] = 7671,
  [7796] = 7671,
  [7797] = 7671,
  [7798] = 7671,
  [7799] = 7671,
  [7800] = 7671,
  [7801] = 7671,
  [7802] = 7671,
  [7803] = 7671,
  [7804] = 7671,
  [7805] = 7671,
  [7806] = 7671,
  [7807] = 7671,
  [7808] = 7671,
  [7809] = 7671,
  [7810] = 7671,
  [7811] = 7671,
  [7812] = 7671,
  [7813] = 7706,
  [7814] = 7671,
  [7815] = 7671,
  [7816] = 7671,
  [7817] = 7671,
  [7818] = 7671,
  [7819] = 7671,
  [7820] = 7671,
  [7821] = 7671,
  [7822] = 7671,
  [7823] = 7671,
  [7824] = 7671,
  [7825] = 7671,
  [7826] = 7671,
  [7827] = 7671,
  [7828] = 7671,
  [7829] = 7671,
  [7830] = 7671,
  [7831] = 7671,
  [7832] = 7671,
  [7833] = 7671,
  [7834] = 7671,
  [7835] = 7671,
  [7836] = 7671,
  [7837] = 7671,
  [7838] = 2353,
  [7839] = 6810,
  [7840] = 7840,
  [7841] = 7841,
  [7842] = 2290,
  [7843] = 2301,
  [7844] = 7673,
  [7845] = 7683,
  [7846] = 7671,
  [7847] = 2369,
  [7848] = 2312,
  [7849] = 2301,
  [7850] = 2288,
  [7851] = 7702,
  [7852] = 2373,
  [7853] = 2283,
  [7854] = 2283,
  [7855] = 7711,
  [7856] = 7482,
  [7857] = 7857,
  [7858] = 7671,
  [7859] = 7702,
  [7860] = 7703,
  [7861] = 7673,
  [7862] = 7705,
  [7863] = 7706,
  [7864] = 7489,
  [7865] = 7683,
  [7866] = 2286,
  [7867] = 7489,
  [7868] = 7868,
  [7869] = 2328,
  [7870] = 7673,
  [7871] = 7683,
  [7872] = 7465,
  [7873] = 7840,
  [7874] = 7874,
  [7875] = 7671,
  [7876] = 2369,
  [7877] = 2383,
  [7878] = 2373,
  [7879] = 7879,
  [7880] = 7880,
  [7881] = 2383,
  [7882] = 7882,
  [7883] = 6819,
  [7884] = 7430,
  [7885] = 2316,
  [7886] = 2353,
  [7887] = 7702,
  [7888] = 7888,
  [7889] = 2278,
  [7890] = 7673,
  [7891] = 7683,
  [7892] = 2272,
  [7893] = 7893,
  [7894] = 2256,
  [7895] = 2369,
  [7896] = 2373,
  [7897] = 7857,
  [7898] = 7898,
  [7899] = 2276,
  [7900] = 7857,
  [7901] = 7671,
  [7902] = 2383,
  [7903] = 7898,
  [7904] = 2281,
  [7905] = 7857,
  [7906] = 7898,
  [7907] = 7857,
  [7908] = 7898,
  [7909] = 2282,
  [7910] = 7857,
  [7911] = 7898,
  [7912] = 7912,
  [7913] = 2353,
  [7914] = 7857,
  [7915] = 7898,
  [7916] = 7857,
  [7917] = 7898,
  [7918] = 7857,
  [7919] = 7898,
  [7920] = 7857,
  [7921] = 7898,
  [7922] = 7857,
  [7923] = 7898,
  [7924] = 7857,
  [7925] = 7898,
  [7926] = 7857,
  [7927] = 7898,
  [7928] = 2284,
  [7929] = 7702,
  [7930] = 7930,
  [7931] = 7703,
  [7932] = 7705,
  [7933] = 7933,
  [7934] = 7706,
  [7935] = 2285,
  [7936] = 7671,
  [7937] = 2287,
  [7938] = 7938,
  [7939] = 2316,
  [7940] = 2313,
  [7941] = 7702,
  [7942] = 2278,
  [7943] = 2276,
  [7944] = 7944,
  [7945] = 7898,
  [7946] = 7880,
  [7947] = 2281,
  [7948] = 7868,
  [7949] = 7711,
  [7950] = 7880,
  [7951] = 7868,
  [7952] = 7880,
  [7953] = 7880,
  [7954] = 7880,
  [7955] = 7880,
  [7956] = 7880,
  [7957] = 7880,
  [7958] = 7880,
  [7959] = 7880,
  [7960] = 7880,
  [7961] = 7880,
  [7962] = 7880,
  [7963] = 7880,
  [7964] = 7880,
  [7965] = 7880,
  [7966] = 7880,
  [7967] = 7880,
  [7968] = 6835,
  [7969] = 7880,
  [7970] = 7880,
  [7971] = 7880,
  [7972] = 7880,
  [7973] = 7880,
  [7974] = 7880,
  [7975] = 7880,
  [7976] = 7880,
  [7977] = 7880,
  [7978] = 7880,
  [7979] = 7880,
  [7980] = 7880,
  [7981] = 7880,
  [7982] = 7671,
  [7983] = 2539,
  [7984] = 2276,
  [7985] = 2369,
  [7986] = 2313,
  [7987] = 2535,
  [7988] = 2550,
  [7989] = 2281,
  [7990] = 7874,
  [7991] = 2282,
  [7992] = 7888,
  [7993] = 7993,
  [7994] = 7930,
  [7995] = 6819,
  [7996] = 7893,
  [7997] = 7930,
  [7998] = 7933,
  [7999] = 6835,
  [8000] = 7874,
  [8001] = 2284,
  [8002] = 2574,
  [8003] = 7993,
  [8004] = 2285,
  [8005] = 2551,
  [8006] = 2561,
  [8007] = 7993,
  [8008] = 2582,
  [8009] = 2536,
  [8010] = 6810,
  [8011] = 2560,
  [8012] = 2575,
  [8013] = 7993,
  [8014] = 6971,
  [8015] = 2552,
  [8016] = 2287,
  [8017] = 7993,
  [8018] = 2373,
  [8019] = 6921,
  [8020] = 2312,
  [8021] = 2272,
  [8022] = 7993,
  [8023] = 2301,
  [8024] = 6938,
  [8025] = 2283,
  [8026] = 2256,
  [8027] = 6872,
  [8028] = 7993,
  [8029] = 8029,
  [8030] = 2255,
  [8031] = 8031,
  [8032] = 2288,
  [8033] = 6557,
  [8034] = 2289,
  [8035] = 2562,
  [8036] = 2256,
  [8037] = 2255,
  [8038] = 2272,
  [8039] = 8039,
  [8040] = 2256,
  [8041] = 2537,
  [8042] = 8042,
  [8043] = 2485,
  [8044] = 2290,
  [8045] = 7993,
  [8046] = 8046,
  [8047] = 7930,
  [8048] = 2558,
  [8049] = 7993,
  [8050] = 7993,
  [8051] = 2544,
  [8052] = 6559,
  [8053] = 2545,
  [8054] = 7888,
  [8055] = 7993,
  [8056] = 2503,
  [8057] = 2540,
  [8058] = 2255,
  [8059] = 7933,
  [8060] = 6810,
  [8061] = 2504,
  [8062] = 2547,
  [8063] = 2383,
  [8064] = 6560,
  [8065] = 2542,
  [8066] = 6835,
  [8067] = 7893,
  [8068] = 2588,
  [8069] = 7933,
  [8070] = 2525,
  [8071] = 2353,
  [8072] = 6835,
  [8073] = 7888,
  [8074] = 8074,
  [8075] = 6919,
  [8076] = 2272,
  [8077] = 6886,
  [8078] = 2530,
  [8079] = 6819,
  [8080] = 2316,
  [8081] = 6819,
  [8082] = 2589,
  [8083] = 7874,
  [8084] = 7893,
  [8085] = 8085,
  [8086] = 2563,
  [8087] = 2278,
  [8088] = 6810,
  [8089] = 2583,
  [8090] = 2555,
  [8091] = 2572,
  [8092] = 2556,
  [8093] = 2573,
  [8094] = 7993,
  [8095] = 8095,
  [8096] = 8096,
  [8097] = 7933,
  [8098] = 2503,
  [8099] = 8042,
  [8100] = 8085,
  [8101] = 2504,
  [8102] = 6971,
  [8103] = 6835,
  [8104] = 2272,
  [8105] = 2525,
  [8106] = 2530,
  [8107] = 2256,
  [8108] = 2313,
  [8109] = 6938,
  [8110] = 8110,
  [8111] = 2535,
  [8112] = 8112,
  [8113] = 6559,
  [8114] = 2536,
  [8115] = 2540,
  [8116] = 2485,
  [8117] = 8117,
  [8118] = 2544,
  [8119] = 8119,
  [8120] = 8120,
  [8121] = 8074,
  [8122] = 2542,
  [8123] = 8029,
  [8124] = 2572,
  [8125] = 2582,
  [8126] = 7888,
  [8127] = 8127,
  [8128] = 8128,
  [8129] = 2573,
  [8130] = 8130,
  [8131] = 6557,
  [8132] = 8074,
  [8133] = 2574,
  [8134] = 6810,
  [8135] = 8039,
  [8136] = 8029,
  [8137] = 6625,
  [8138] = 2575,
  [8139] = 2545,
  [8140] = 8031,
  [8141] = 2583,
  [8142] = 2547,
  [8143] = 2550,
  [8144] = 6560,
  [8145] = 7893,
  [8146] = 8146,
  [8147] = 2551,
  [8148] = 8148,
  [8149] = 8149,
  [8150] = 2552,
  [8151] = 2555,
  [8152] = 2556,
  [8153] = 2558,
  [8154] = 6886,
  [8155] = 8155,
  [8156] = 2560,
  [8157] = 2561,
  [8158] = 2562,
  [8159] = 8159,
  [8160] = 2563,
  [8161] = 2503,
  [8162] = 2504,
  [8163] = 8163,
  [8164] = 2588,
  [8165] = 2589,
  [8166] = 8166,
  [8167] = 2525,
  [8168] = 2530,
  [8169] = 2535,
  [8170] = 2536,
  [8171] = 2540,
  [8172] = 6871,
  [8173] = 6560,
  [8174] = 2542,
  [8175] = 2572,
  [8176] = 2573,
  [8177] = 2574,
  [8178] = 8178,
  [8179] = 8179,
  [8180] = 6921,
  [8181] = 2575,
  [8182] = 6886,
  [8183] = 8046,
  [8184] = 6895,
  [8185] = 6971,
  [8186] = 8029,
  [8187] = 6557,
  [8188] = 6872,
  [8189] = 2582,
  [8190] = 2583,
  [8191] = 2537,
  [8192] = 2539,
  [8193] = 7874,
  [8194] = 7888,
  [8195] = 8195,
  [8196] = 2485,
  [8197] = 2544,
  [8198] = 2545,
  [8199] = 8199,
  [8200] = 2547,
  [8201] = 8201,
  [8202] = 2550,
  [8203] = 2551,
  [8204] = 2552,
  [8205] = 2555,
  [8206] = 2556,
  [8207] = 2562,
  [8208] = 2560,
  [8209] = 2561,
  [8210] = 2562,
  [8211] = 2563,
  [8212] = 2503,
  [8213] = 2504,
  [8214] = 2525,
  [8215] = 2530,
  [8216] = 2535,
  [8217] = 2536,
  [8218] = 2540,
  [8219] = 8046,
  [8220] = 2563,
  [8221] = 2542,
  [8222] = 2572,
  [8223] = 2573,
  [8224] = 2574,
  [8225] = 7930,
  [8226] = 6921,
  [8227] = 7893,
  [8228] = 2575,
  [8229] = 7930,
  [8230] = 2313,
  [8231] = 8042,
  [8232] = 8031,
  [8233] = 8233,
  [8234] = 8234,
  [8235] = 2588,
  [8236] = 7933,
  [8237] = 2589,
  [8238] = 8238,
  [8239] = 8239,
  [8240] = 8240,
  [8241] = 6971,
  [8242] = 6919,
  [8243] = 6921,
  [8244] = 2255,
  [8245] = 6938,
  [8246] = 2582,
  [8247] = 2583,
  [8248] = 6938,
  [8249] = 6919,
  [8250] = 6819,
  [8251] = 6872,
  [8252] = 8252,
  [8253] = 6560,
  [8254] = 2588,
  [8255] = 2589,
  [8256] = 8256,
  [8257] = 2313,
  [8258] = 6919,
  [8259] = 2537,
  [8260] = 8260,
  [8261] = 2539,
  [8262] = 8262,
  [8263] = 8263,
  [8264] = 6971,
  [8265] = 6921,
  [8266] = 6938,
  [8267] = 6872,
  [8268] = 8031,
  [8269] = 2485,
  [8270] = 7874,
  [8271] = 2544,
  [8272] = 6559,
  [8273] = 2545,
  [8274] = 6557,
  [8275] = 6872,
  [8276] = 6559,
  [8277] = 6919,
  [8278] = 2537,
  [8279] = 2539,
  [8280] = 2547,
  [8281] = 8281,
  [8282] = 2550,
  [8283] = 2551,
  [8284] = 2552,
  [8285] = 2555,
  [8286] = 2556,
  [8287] = 8085,
  [8288] = 8288,
  [8289] = 2558,
  [8290] = 2560,
  [8291] = 8291,
  [8292] = 6886,
  [8293] = 8039,
  [8294] = 2561,
  [8295] = 2558,
  [8296] = 2536,
  [8297] = 6921,
  [8298] = 8159,
  [8299] = 2367,
  [8300] = 8300,
  [8301] = 8159,
  [8302] = 6919,
  [8303] = 6938,
  [8304] = 8149,
  [8305] = 6871,
  [8306] = 6967,
  [8307] = 6872,
  [8308] = 2547,
  [8309] = 8309,
  [8310] = 8256,
  [8311] = 8117,
  [8312] = 8312,
  [8313] = 2241,
  [8314] = 8146,
  [8315] = 8315,
  [8316] = 8233,
  [8317] = 8234,
  [8318] = 8238,
  [8319] = 8319,
  [8320] = 8260,
  [8321] = 8260,
  [8322] = 8256,
  [8323] = 8155,
  [8324] = 8288,
  [8325] = 8325,
  [8326] = 6971,
  [8327] = 8252,
  [8328] = 8328,
  [8329] = 8329,
  [8330] = 8330,
  [8331] = 2556,
  [8332] = 8332,
  [8333] = 8195,
  [8334] = 8195,
  [8335] = 8148,
  [8336] = 2530,
  [8337] = 8337,
  [8338] = 8338,
  [8339] = 8110,
  [8340] = 8340,
  [8341] = 8117,
  [8342] = 6971,
  [8343] = 8178,
  [8344] = 8256,
  [8345] = 2551,
  [8346] = 8159,
  [8347] = 2535,
  [8348] = 8348,
  [8349] = 8031,
  [8350] = 8112,
  [8351] = 6871,
  [8352] = 8263,
  [8353] = 8166,
  [8354] = 8354,
  [8355] = 8355,
  [8356] = 8263,
  [8357] = 8110,
  [8358] = 8199,
  [8359] = 6625,
  [8360] = 2540,
  [8361] = 8361,
  [8362] = 8128,
  [8363] = 8110,
  [8364] = 8364,
  [8365] = 8166,
  [8366] = 8146,
  [8367] = 8367,
  [8368] = 8112,
  [8369] = 8239,
  [8370] = 6872,
  [8371] = 6938,
  [8372] = 8372,
  [8373] = 8201,
  [8374] = 8240,
  [8375] = 6921,
  [8376] = 8260,
  [8377] = 2552,
  [8378] = 8378,
  [8379] = 6919,
  [8380] = 2537,
  [8381] = 8119,
  [8382] = 2561,
  [8383] = 8029,
  [8384] = 8179,
  [8385] = 2562,
  [8386] = 8119,
  [8387] = 2542,
  [8388] = 8155,
  [8389] = 8179,
  [8390] = 8281,
  [8391] = 8262,
  [8392] = 8392,
  [8393] = 8281,
  [8394] = 8199,
  [8395] = 8096,
  [8396] = 8201,
  [8397] = 2539,
  [8398] = 8288,
  [8399] = 8130,
  [8400] = 2572,
  [8401] = 8291,
  [8402] = 8402,
  [8403] = 8178,
  [8404] = 2582,
  [8405] = 6919,
  [8406] = 8130,
  [8407] = 2583,
  [8408] = 8408,
  [8409] = 8409,
  [8410] = 8120,
  [8411] = 2563,
  [8412] = 8412,
  [8413] = 8095,
  [8414] = 6886,
  [8415] = 6871,
  [8416] = 2573,
  [8417] = 2555,
  [8418] = 8119,
  [8419] = 8166,
  [8420] = 8252,
  [8421] = 8239,
  [8422] = 8240,
  [8423] = 8262,
  [8424] = 8096,
  [8425] = 8127,
  [8426] = 8120,
  [8427] = 6895,
  [8428] = 6898,
  [8429] = 6921,
  [8430] = 6625,
  [8431] = 8127,
  [8432] = 2574,
  [8433] = 6559,
  [8434] = 8146,
  [8435] = 6872,
  [8436] = 8149,
  [8437] = 8437,
  [8438] = 6871,
  [8439] = 8127,
  [8440] = 8291,
  [8441] = 2503,
  [8442] = 6953,
  [8443] = 6895,
  [8444] = 2504,
  [8445] = 6971,
  [8446] = 8446,
  [8447] = 6921,
  [8448] = 6890,
  [8449] = 2258,
  [8450] = 6938,
  [8451] = 2575,
  [8452] = 8452,
  [8453] = 6872,
  [8454] = 8239,
  [8455] = 8199,
  [8456] = 6625,
  [8457] = 6560,
  [8458] = 8458,
  [8459] = 8240,
  [8460] = 8148,
  [8461] = 6895,
  [8462] = 6919,
  [8463] = 2558,
  [8464] = 8262,
  [8465] = 8201,
  [8466] = 8155,
  [8467] = 8095,
  [8468] = 8096,
  [8469] = 8469,
  [8470] = 2588,
  [8471] = 8291,
  [8472] = 8263,
  [8473] = 8288,
  [8474] = 2589,
  [8475] = 2236,
  [8476] = 6938,
  [8477] = 8233,
  [8478] = 2313,
  [8479] = 8149,
  [8480] = 8128,
  [8481] = 8481,
  [8482] = 6872,
  [8483] = 8179,
  [8484] = 2550,
  [8485] = 2560,
  [8486] = 8486,
  [8487] = 8234,
  [8488] = 2240,
  [8489] = 2544,
  [8490] = 8490,
  [8491] = 6919,
  [8492] = 8095,
  [8493] = 8195,
  [8494] = 8494,
  [8495] = 2234,
  [8496] = 8281,
  [8497] = 8238,
  [8498] = 8112,
  [8499] = 2545,
  [8500] = 6557,
  [8501] = 2485,
  [8502] = 8502,
  [8503] = 6921,
  [8504] = 8128,
  [8505] = 2259,
  [8506] = 2263,
  [8507] = 8252,
  [8508] = 8130,
  [8509] = 6971,
  [8510] = 8120,
  [8511] = 8178,
  [8512] = 6971,
  [8513] = 8148,
  [8514] = 8233,
  [8515] = 2525,
  [8516] = 6895,
  [8517] = 8234,
  [8518] = 8238,
  [8519] = 8117,
  [8520] = 6938,
  [8521] = 8340,
  [8522] = 8159,
  [8523] = 8354,
  [8524] = 8330,
  [8525] = 8469,
  [8526] = 8337,
  [8527] = 8348,
  [8528] = 8367,
  [8529] = 8117,
  [8530] = 8530,
  [8531] = 6895,
  [8532] = 8195,
  [8533] = 8533,
  [8534] = 8148,
  [8535] = 8535,
  [8536] = 6625,
  [8537] = 8166,
  [8538] = 8179,
  [8539] = 2367,
  [8540] = 8481,
  [8541] = 8486,
  [8542] = 8319,
  [8543] = 8110,
  [8544] = 8112,
  [8545] = 8119,
  [8546] = 8120,
  [8547] = 8127,
  [8548] = 8128,
  [8549] = 8095,
  [8550] = 8319,
  [8551] = 8325,
  [8552] = 8233,
  [8553] = 8234,
  [8554] = 8238,
  [8555] = 8555,
  [8556] = 6871,
  [8557] = 8252,
  [8558] = 8558,
  [8559] = 8533,
  [8560] = 8239,
  [8561] = 8240,
  [8562] = 8262,
  [8563] = 8096,
  [8564] = 6895,
  [8565] = 6971,
  [8566] = 8146,
  [8567] = 8155,
  [8568] = 8325,
  [8569] = 8263,
  [8570] = 8281,
  [8571] = 8288,
  [8572] = 6921,
  [8573] = 8130,
  [8574] = 8338,
  [8575] = 8312,
  [8576] = 8576,
  [8577] = 8577,
  [8578] = 8437,
  [8579] = 8446,
  [8580] = 6938,
  [8581] = 2263,
  [8582] = 6872,
  [8583] = 8555,
  [8584] = 8452,
  [8585] = 8469,
  [8586] = 8502,
  [8587] = 8364,
  [8588] = 2299,
  [8589] = 8309,
  [8590] = 8315,
  [8591] = 8361,
  [8592] = 8372,
  [8593] = 8408,
  [8594] = 2255,
  [8595] = 8437,
  [8596] = 8312,
  [8597] = 2258,
  [8598] = 8409,
  [8599] = 8355,
  [8600] = 8300,
  [8601] = 8332,
  [8602] = 2236,
  [8603] = 8329,
  [8604] = 2234,
  [8605] = 2241,
  [8606] = 8348,
  [8607] = 8490,
  [8608] = 2272,
  [8609] = 8458,
  [8610] = 2259,
  [8611] = 8340,
  [8612] = 8338,
  [8613] = 8502,
  [8614] = 8530,
  [8615] = 8354,
  [8616] = 6919,
  [8617] = 8408,
  [8618] = 8409,
  [8619] = 8558,
  [8620] = 2240,
  [8621] = 8378,
  [8622] = 8490,
  [8623] = 8458,
  [8624] = 6967,
  [8625] = 8309,
  [8626] = 8315,
  [8627] = 8312,
  [8628] = 8577,
  [8629] = 8481,
  [8630] = 8486,
  [8631] = 8378,
  [8632] = 8319,
  [8633] = 8325,
  [8634] = 8291,
  [8635] = 8361,
  [8636] = 8494,
  [8637] = 6967,
  [8638] = 8469,
  [8639] = 8309,
  [8640] = 8315,
  [8641] = 8361,
  [8642] = 8372,
  [8643] = 6953,
  [8644] = 8330,
  [8645] = 8337,
  [8646] = 2256,
  [8647] = 8372,
  [8648] = 8437,
  [8649] = 8355,
  [8650] = 8300,
  [8651] = 8332,
  [8652] = 8329,
  [8653] = 8348,
  [8654] = 8367,
  [8655] = 8260,
  [8656] = 6890,
  [8657] = 8576,
  [8658] = 8340,
  [8659] = 8338,
  [8660] = 8502,
  [8661] = 8408,
  [8662] = 8409,
  [8663] = 8490,
  [8664] = 8149,
  [8665] = 6953,
  [8666] = 6967,
  [8667] = 8458,
  [8668] = 8354,
  [8669] = 8199,
  [8670] = 8201,
  [8671] = 8535,
  [8672] = 8481,
  [8673] = 6919,
  [8674] = 8256,
  [8675] = 6871,
  [8676] = 8355,
  [8677] = 8330,
  [8678] = 8337,
  [8679] = 8486,
  [8680] = 6953,
  [8681] = 8300,
  [8682] = 6971,
  [8683] = 6921,
  [8684] = 8378,
  [8685] = 6938,
  [8686] = 8332,
  [8687] = 6898,
  [8688] = 8329,
  [8689] = 6872,
  [8690] = 8392,
  [8691] = 8328,
  [8692] = 8402,
  [8693] = 2292,
  [8694] = 8178,
  [8695] = 8367,
  [8696] = 8325,
  [8697] = 8555,
  [8698] = 8698,
  [8699] = 2286,
  [8700] = 8700,
  [8701] = 8576,
  [8702] = 8577,
  [8703] = 8533,
  [8704] = 8481,
  [8705] = 8486,
  [8706] = 8437,
  [8707] = 2272,
  [8708] = 8355,
  [8709] = 8300,
  [8710] = 2256,
  [8711] = 2313,
  [8712] = 8332,
  [8713] = 8329,
  [8714] = 8348,
  [8715] = 6967,
  [8716] = 2277,
  [8717] = 2255,
  [8718] = 8367,
  [8719] = 8319,
  [8720] = 8378,
  [8721] = 8337,
  [8722] = 2292,
  [8723] = 2264,
  [8724] = 8340,
  [8725] = 6953,
  [8726] = 8338,
  [8727] = 8502,
  [8728] = 8535,
  [8729] = 8312,
  [8730] = 8408,
  [8731] = 8409,
  [8732] = 8354,
  [8733] = 8733,
  [8734] = 8490,
  [8735] = 8469,
  [8736] = 8558,
  [8737] = 2299,
  [8738] = 8309,
  [8739] = 2265,
  [8740] = 8740,
  [8741] = 8458,
  [8742] = 8315,
  [8743] = 8530,
  [8744] = 8361,
  [8745] = 8372,
  [8746] = 8330,
  [8747] = 2283,
  [8748] = 2289,
  [8749] = 2288,
  [8750] = 2285,
  [8751] = 2316,
  [8752] = 2276,
  [8753] = 2264,
  [8754] = 2282,
  [8755] = 2277,
  [8756] = 2313,
  [8757] = 2301,
  [8758] = 2278,
  [8759] = 2284,
  [8760] = 2290,
  [8761] = 2265,
  [8762] = 2287,
  [8763] = 2286,
  [8764] = 2312,
  [8765] = 2281,
  [8766] = 6819,
  [8767] = 6810,
  [8768] = 2301,
  [8769] = 2316,
  [8770] = 2328,
  [8771] = 2283,
  [8772] = 2281,
  [8773] = 6835,
  [8774] = 6819,
  [8775] = 2285,
  [8776] = 6810,
  [8777] = 2289,
  [8778] = 6835,
  [8779] = 2282,
  [8780] = 2312,
  [8781] = 8781,
  [8782] = 2290,
  [8783] = 2278,
  [8784] = 2287,
  [8785] = 2284,
  [8786] = 8786,
  [8787] = 8787,
  [8788] = 2276,
  [8789] = 2288,
  [8790] = 6886,
  [8791] = 2383,
  [8792] = 6835,
  [8793] = 6921,
  [8794] = 2369,
  [8795] = 6919,
  [8796] = 6938,
  [8797] = 6938,
  [8798] = 6971,
  [8799] = 6810,
  [8800] = 6819,
  [8801] = 6971,
  [8802] = 2353,
  [8803] = 6872,
  [8804] = 6872,
  [8805] = 6919,
  [8806] = 2373,
  [8807] = 6921,
  [8808] = 6886,
  [8809] = 6919,
  [8810] = 6921,
  [8811] = 6886,
  [8812] = 6871,
  [8813] = 6919,
  [8814] = 6971,
  [8815] = 6971,
  [8816] = 6872,
  [8817] = 8817,
  [8818] = 6938,
  [8819] = 6871,
  [8820] = 6895,
  [8821] = 6810,
  [8822] = 6895,
  [8823] = 6938,
  [8824] = 6921,
  [8825] = 6919,
  [8826] = 6819,
  [8827] = 6938,
  [8828] = 6872,
  [8829] = 6921,
  [8830] = 6971,
  [8831] = 6872,
  [8832] = 6835,
  [8833] = 2530,
  [8834] = 2589,
  [8835] = 6919,
  [8836] = 2582,
  [8837] = 2583,
  [8838] = 6938,
  [8839] = 2537,
  [8840] = 2539,
  [8841] = 6872,
  [8842] = 2485,
  [8843] = 2574,
  [8844] = 6938,
  [8845] = 6872,
  [8846] = 2563,
  [8847] = 2544,
  [8848] = 2503,
  [8849] = 2504,
  [8850] = 2525,
  [8851] = 6895,
  [8852] = 6886,
  [8853] = 2588,
  [8854] = 2545,
  [8855] = 2558,
  [8856] = 6919,
  [8857] = 6921,
  [8858] = 2560,
  [8859] = 2561,
  [8860] = 2535,
  [8861] = 2536,
  [8862] = 2540,
  [8863] = 2547,
  [8864] = 2542,
  [8865] = 2550,
  [8866] = 6871,
  [8867] = 2551,
  [8868] = 6971,
  [8869] = 2572,
  [8870] = 2552,
  [8871] = 6921,
  [8872] = 2555,
  [8873] = 2562,
  [8874] = 2556,
  [8875] = 6971,
  [8876] = 2573,
  [8877] = 2575,
  [8878] = 8878,
  [8879] = 6938,
  [8880] = 6921,
  [8881] = 8881,
  [8882] = 8882,
  [8883] = 8883,
  [8884] = 6895,
  [8885] = 6919,
  [8886] = 6971,
  [8887] = 6871,
  [8888] = 6872,
  [8889] = 6919,
  [8890] = 8890,
  [8891] = 8891,
  [8892] = 8892,
  [8893] = 8893,
  [8894] = 8894,
  [8895] = 8895,
  [8896] = 8896,
  [8897] = 8897,
  [8898] = 8897,
  [8899] = 8895,
  [8900] = 8900,
  [8901] = 8901,
  [8902] = 8902,
  [8903] = 8902,
  [8904] = 8902,
  [8905] = 8902,
  [8906] = 8902,
  [8907] = 8902,
  [8908] = 8902,
  [8909] = 8902,
  [8910] = 8902,
  [8911] = 8902,
  [8912] = 8902,
  [8913] = 8913,
  [8914] = 8902,
  [8915] = 8902,
  [8916] = 8902,
  [8917] = 8902,
  [8918] = 8902,
  [8919] = 8919,
  [8920] = 8902,
  [8921] = 8902,
  [8922] = 8902,
  [8923] = 8902,
  [8924] = 8902,
  [8925] = 8902,
  [8926] = 8902,
  [8927] = 8902,
  [8928] = 8928,
  [8929] = 8902,
  [8930] = 8902,
  [8931] = 8902,
  [8932] = 2221,
  [8933] = 2221,
  [8934] = 2221,
  [8935] = 2221,
  [8936] = 2221,
  [8937] = 2225,
  [8938] = 2221,
  [8939] = 2225,
  [8940] = 2225,
  [8941] = 2225,
  [8942] = 2225,
  [8943] = 2225,
  [8944] = 8944,
  [8945] = 8944,
  [8946] = 2234,
  [8947] = 2258,
  [8948] = 2259,
  [8949] = 8949,
  [8950] = 2263,
  [8951] = 2240,
  [8952] = 2240,
  [8953] = 2258,
  [8954] = 8954,
  [8955] = 2263,
  [8956] = 2236,
  [8957] = 8954,
  [8958] = 2241,
  [8959] = 2240,
  [8960] = 2263,
  [8961] = 8944,
  [8962] = 2258,
  [8963] = 2240,
  [8964] = 8949,
  [8965] = 2263,
  [8966] = 2236,
  [8967] = 2236,
  [8968] = 2234,
  [8969] = 2241,
  [8970] = 2234,
  [8971] = 2241,
  [8972] = 2236,
  [8973] = 2234,
  [8974] = 2241,
  [8975] = 2259,
  [8976] = 2258,
  [8977] = 2259,
  [8978] = 8944,
  [8979] = 2259,
  [8980] = 2240,
  [8981] = 2299,
  [8982] = 2234,
  [8983] = 2241,
  [8984] = 2258,
  [8985] = 2234,
  [8986] = 8986,
  [8987] = 2241,
  [8988] = 2236,
  [8989] = 8986,
  [8990] = 2259,
  [8991] = 2258,
  [8992] = 2292,
  [8993] = 2234,
  [8994] = 2258,
  [8995] = 2236,
  [8996] = 2259,
  [8997] = 2259,
  [8998] = 2241,
  [8999] = 2258,
  [9000] = 2259,
  [9001] = 2299,
  [9002] = 2263,
  [9003] = 2263,
  [9004] = 2240,
  [9005] = 2240,
  [9006] = 2256,
  [9007] = 2255,
  [9008] = 2236,
  [9009] = 2234,
  [9010] = 2241,
  [9011] = 2234,
  [9012] = 2241,
  [9013] = 8954,
  [9014] = 8944,
  [9015] = 2259,
  [9016] = 2236,
  [9017] = 2234,
  [9018] = 2241,
  [9019] = 8949,
  [9020] = 2236,
  [9021] = 8954,
  [9022] = 2256,
  [9023] = 2299,
  [9024] = 8954,
  [9025] = 2258,
  [9026] = 2258,
  [9027] = 8949,
  [9028] = 2259,
  [9029] = 2259,
  [9030] = 2236,
  [9031] = 2255,
  [9032] = 2234,
  [9033] = 2241,
  [9034] = 2292,
  [9035] = 2263,
  [9036] = 2272,
  [9037] = 8949,
  [9038] = 2263,
  [9039] = 2263,
  [9040] = 2255,
  [9041] = 2240,
  [9042] = 2292,
  [9043] = 2240,
  [9044] = 2255,
  [9045] = 2299,
  [9046] = 2263,
  [9047] = 2272,
  [9048] = 2256,
  [9049] = 2292,
  [9050] = 2240,
  [9051] = 2263,
  [9052] = 2236,
  [9053] = 8944,
  [9054] = 2272,
  [9055] = 2240,
  [9056] = 2258,
  [9057] = 2256,
  [9058] = 2272,
  [9059] = 2313,
  [9060] = 2272,
  [9061] = 2264,
  [9062] = 2263,
  [9063] = 2240,
  [9064] = 2265,
  [9065] = 2241,
  [9066] = 2256,
  [9067] = 2299,
  [9068] = 2277,
  [9069] = 2272,
  [9070] = 2234,
  [9071] = 2265,
  [9072] = 2299,
  [9073] = 2263,
  [9074] = 8954,
  [9075] = 2236,
  [9076] = 2255,
  [9077] = 2264,
  [9078] = 2272,
  [9079] = 2259,
  [9080] = 2263,
  [9081] = 2292,
  [9082] = 2258,
  [9083] = 2264,
  [9084] = 2256,
  [9085] = 2265,
  [9086] = 8986,
  [9087] = 2292,
  [9088] = 2292,
  [9089] = 2256,
  [9090] = 2299,
  [9091] = 2263,
  [9092] = 2255,
  [9093] = 2299,
  [9094] = 2240,
  [9095] = 2299,
  [9096] = 2241,
  [9097] = 2255,
  [9098] = 8949,
  [9099] = 2265,
  [9100] = 8949,
  [9101] = 2272,
  [9102] = 2258,
  [9103] = 2258,
  [9104] = 2258,
  [9105] = 2277,
  [9106] = 2277,
  [9107] = 2264,
  [9108] = 8954,
  [9109] = 2286,
  [9110] = 2255,
  [9111] = 2234,
  [9112] = 2299,
  [9113] = 2259,
  [9114] = 2256,
  [9115] = 2256,
  [9116] = 2236,
  [9117] = 2234,
  [9118] = 2241,
  [9119] = 2240,
  [9120] = 2236,
  [9121] = 2286,
  [9122] = 2255,
  [9123] = 2234,
  [9124] = 2240,
  [9125] = 2272,
  [9126] = 2277,
  [9127] = 2313,
  [9128] = 2292,
  [9129] = 2272,
  [9130] = 2313,
  [9131] = 2259,
  [9132] = 2286,
  [9133] = 2286,
  [9134] = 2292,
  [9135] = 2272,
  [9136] = 2256,
  [9137] = 2299,
  [9138] = 2255,
  [9139] = 2255,
  [9140] = 2292,
  [9141] = 2241,
  [9142] = 2259,
  [9143] = 2256,
  [9144] = 2292,
  [9145] = 2236,
  [9146] = 2313,
  [9147] = 8986,
  [9148] = 8986,
  [9149] = 2264,
  [9150] = 2313,
  [9151] = 2292,
  [9152] = 2265,
  [9153] = 2278,
  [9154] = 2265,
  [9155] = 2277,
  [9156] = 2277,
  [9157] = 2272,
  [9158] = 2292,
  [9159] = 2264,
  [9160] = 2265,
  [9161] = 2287,
  [9162] = 2288,
  [9163] = 2255,
  [9164] = 2240,
  [9165] = 2286,
  [9166] = 2265,
  [9167] = 2289,
  [9168] = 2264,
  [9169] = 2290,
  [9170] = 2299,
  [9171] = 2316,
  [9172] = 2292,
  [9173] = 2221,
  [9174] = 2278,
  [9175] = 2276,
  [9176] = 2256,
  [9177] = 2281,
  [9178] = 2282,
  [9179] = 2299,
  [9180] = 2265,
  [9181] = 9181,
  [9182] = 2286,
  [9183] = 2284,
  [9184] = 2285,
  [9185] = 2287,
  [9186] = 2272,
  [9187] = 2272,
  [9188] = 2288,
  [9189] = 2289,
  [9190] = 2290,
  [9191] = 2255,
  [9192] = 2256,
  [9193] = 2316,
  [9194] = 2286,
  [9195] = 2263,
  [9196] = 2278,
  [9197] = 2313,
  [9198] = 2286,
  [9199] = 2316,
  [9200] = 2258,
  [9201] = 2276,
  [9202] = 2281,
  [9203] = 2221,
  [9204] = 2282,
  [9205] = 2283,
  [9206] = 2265,
  [9207] = 2284,
  [9208] = 2264,
  [9209] = 2278,
  [9210] = 2312,
  [9211] = 2259,
  [9212] = 2285,
  [9213] = 2264,
  [9214] = 2287,
  [9215] = 2276,
  [9216] = 2286,
  [9217] = 2313,
  [9218] = 2288,
  [9219] = 2256,
  [9220] = 2289,
  [9221] = 2281,
  [9222] = 2290,
  [9223] = 2316,
  [9224] = 2277,
  [9225] = 2282,
  [9226] = 2313,
  [9227] = 2312,
  [9228] = 2301,
  [9229] = 2283,
  [9230] = 2284,
  [9231] = 2276,
  [9232] = 2277,
  [9233] = 2292,
  [9234] = 2299,
  [9235] = 2236,
  [9236] = 2234,
  [9237] = 2241,
  [9238] = 2277,
  [9239] = 2299,
  [9240] = 2312,
  [9241] = 2313,
  [9242] = 2286,
  [9243] = 2285,
  [9244] = 2281,
  [9245] = 2287,
  [9246] = 2255,
  [9247] = 2328,
  [9248] = 2288,
  [9249] = 2282,
  [9250] = 2312,
  [9251] = 2264,
  [9252] = 2301,
  [9253] = 2283,
  [9254] = 2301,
  [9255] = 2277,
  [9256] = 2283,
  [9257] = 2289,
  [9258] = 2286,
  [9259] = 2265,
  [9260] = 2284,
  [9261] = 2290,
  [9262] = 2313,
  [9263] = 2264,
  [9264] = 2272,
  [9265] = 2255,
  [9266] = 2256,
  [9267] = 2277,
  [9268] = 2285,
  [9269] = 8986,
  [9270] = 2301,
  [9271] = 2313,
  [9272] = 2289,
  [9273] = 9273,
  [9274] = 9274,
  [9275] = 2288,
  [9276] = 9276,
  [9277] = 9277,
  [9278] = 2278,
  [9279] = 2276,
  [9280] = 2290,
  [9281] = 2284,
  [9282] = 2264,
  [9283] = 2288,
  [9284] = 2292,
  [9285] = 2290,
  [9286] = 2285,
  [9287] = 2369,
  [9288] = 2328,
  [9289] = 2286,
  [9290] = 9274,
  [9291] = 9274,
  [9292] = 9274,
  [9293] = 2277,
  [9294] = 2277,
  [9295] = 9273,
  [9296] = 2287,
  [9297] = 2281,
  [9298] = 2284,
  [9299] = 2288,
  [9300] = 2328,
  [9301] = 2282,
  [9302] = 2299,
  [9303] = 2284,
  [9304] = 2312,
  [9305] = 2286,
  [9306] = 2313,
  [9307] = 2264,
  [9308] = 9276,
  [9309] = 9277,
  [9310] = 2316,
  [9311] = 2313,
  [9312] = 2353,
  [9313] = 2284,
  [9314] = 9314,
  [9315] = 2285,
  [9316] = 2373,
  [9317] = 2265,
  [9318] = 9314,
  [9319] = 2287,
  [9320] = 2316,
  [9321] = 2287,
  [9322] = 9273,
  [9323] = 9274,
  [9324] = 2285,
  [9325] = 9277,
  [9326] = 9276,
  [9327] = 9276,
  [9328] = 9277,
  [9329] = 9314,
  [9330] = 2264,
  [9331] = 2255,
  [9332] = 9314,
  [9333] = 2281,
  [9334] = 2312,
  [9335] = 2287,
  [9336] = 2287,
  [9337] = 9273,
  [9338] = 9277,
  [9339] = 9274,
  [9340] = 9273,
  [9341] = 2316,
  [9342] = 9314,
  [9343] = 2312,
  [9344] = 2284,
  [9345] = 9274,
  [9346] = 2301,
  [9347] = 2316,
  [9348] = 9277,
  [9349] = 9276,
  [9350] = 9273,
  [9351] = 9274,
  [9352] = 2284,
  [9353] = 2285,
  [9354] = 9276,
  [9355] = 9276,
  [9356] = 2290,
  [9357] = 9277,
  [9358] = 2282,
  [9359] = 2288,
  [9360] = 9276,
  [9361] = 9277,
  [9362] = 9314,
  [9363] = 9314,
  [9364] = 2301,
  [9365] = 2278,
  [9366] = 9273,
  [9367] = 2285,
  [9368] = 2288,
  [9369] = 2277,
  [9370] = 2278,
  [9371] = 2287,
  [9372] = 2289,
  [9373] = 9274,
  [9374] = 2301,
  [9375] = 2278,
  [9376] = 9276,
  [9377] = 9273,
  [9378] = 2289,
  [9379] = 2283,
  [9380] = 2265,
  [9381] = 2301,
  [9382] = 2276,
  [9383] = 2272,
  [9384] = 2283,
  [9385] = 9314,
  [9386] = 2289,
  [9387] = 2283,
  [9388] = 2282,
  [9389] = 9277,
  [9390] = 2383,
  [9391] = 2276,
  [9392] = 2316,
  [9393] = 2289,
  [9394] = 9273,
  [9395] = 9277,
  [9396] = 9273,
  [9397] = 2281,
  [9398] = 9274,
  [9399] = 2282,
  [9400] = 2316,
  [9401] = 2289,
  [9402] = 2290,
  [9403] = 2284,
  [9404] = 9274,
  [9405] = 9273,
  [9406] = 9274,
  [9407] = 9276,
  [9408] = 2276,
  [9409] = 2313,
  [9410] = 2285,
  [9411] = 2281,
  [9412] = 2312,
  [9413] = 9276,
  [9414] = 2278,
  [9415] = 9273,
  [9416] = 2301,
  [9417] = 2281,
  [9418] = 9273,
  [9419] = 9277,
  [9420] = 2287,
  [9421] = 2312,
  [9422] = 2283,
  [9423] = 2312,
  [9424] = 2288,
  [9425] = 2301,
  [9426] = 2289,
  [9427] = 2283,
  [9428] = 2283,
  [9429] = 2264,
  [9430] = 2290,
  [9431] = 2265,
  [9432] = 2282,
  [9433] = 9314,
  [9434] = 9276,
  [9435] = 2290,
  [9436] = 2276,
  [9437] = 9273,
  [9438] = 2286,
  [9439] = 2286,
  [9440] = 2265,
  [9441] = 2316,
  [9442] = 9277,
  [9443] = 2313,
  [9444] = 2256,
  [9445] = 2278,
  [9446] = 2301,
  [9447] = 2276,
  [9448] = 2288,
  [9449] = 2281,
  [9450] = 2282,
  [9451] = 2278,
  [9452] = 2276,
  [9453] = 9277,
  [9454] = 2283,
  [9455] = 9314,
  [9456] = 2281,
  [9457] = 2282,
  [9458] = 2285,
  [9459] = 9277,
  [9460] = 2277,
  [9461] = 2290,
  [9462] = 2312,
  [9463] = 2284,
  [9464] = 2278,
  [9465] = 2256,
  [9466] = 9466,
  [9467] = 2264,
  [9468] = 9466,
  [9469] = 2313,
  [9470] = 2373,
  [9471] = 9466,
  [9472] = 2265,
  [9473] = 9466,
  [9474] = 9466,
  [9475] = 2284,
  [9476] = 9466,
  [9477] = 2316,
  [9478] = 9466,
  [9479] = 9466,
  [9480] = 2276,
  [9481] = 2281,
  [9482] = 2282,
  [9483] = 2284,
  [9484] = 2276,
  [9485] = 2328,
  [9486] = 2316,
  [9487] = 2278,
  [9488] = 2285,
  [9489] = 2312,
  [9490] = 2287,
  [9491] = 2383,
  [9492] = 2301,
  [9493] = 2283,
  [9494] = 2281,
  [9495] = 2288,
  [9496] = 2289,
  [9497] = 2290,
  [9498] = 2258,
  [9499] = 2276,
  [9500] = 2282,
  [9501] = 2281,
  [9502] = 2282,
  [9503] = 2259,
  [9504] = 2285,
  [9505] = 2255,
  [9506] = 2287,
  [9507] = 2288,
  [9508] = 2225,
  [9509] = 2289,
  [9510] = 2290,
  [9511] = 9466,
  [9512] = 2285,
  [9513] = 2272,
  [9514] = 2316,
  [9515] = 2369,
  [9516] = 2278,
  [9517] = 2276,
  [9518] = 2281,
  [9519] = 2282,
  [9520] = 2284,
  [9521] = 2285,
  [9522] = 2287,
  [9523] = 2288,
  [9524] = 2289,
  [9525] = 2290,
  [9526] = 2288,
  [9527] = 2256,
  [9528] = 9466,
  [9529] = 2328,
  [9530] = 2289,
  [9531] = 2353,
  [9532] = 2290,
  [9533] = 9466,
  [9534] = 2312,
  [9535] = 2301,
  [9536] = 2283,
  [9537] = 2312,
  [9538] = 2301,
  [9539] = 2283,
  [9540] = 2255,
  [9541] = 9541,
  [9542] = 2272,
  [9543] = 2316,
  [9544] = 2256,
  [9545] = 9466,
  [9546] = 2278,
  [9547] = 2286,
  [9548] = 9548,
  [9549] = 2277,
  [9550] = 9466,
  [9551] = 2369,
  [9552] = 2373,
  [9553] = 2312,
  [9554] = 2301,
  [9555] = 9466,
  [9556] = 2383,
  [9557] = 2283,
  [9558] = 2353,
  [9559] = 2372,
  [9560] = 2272,
  [9561] = 2255,
  [9562] = 2225,
  [9563] = 2287,
  [9564] = 9564,
  [9565] = 9564,
  [9566] = 2312,
  [9567] = 2292,
  [9568] = 2369,
  [9569] = 2313,
  [9570] = 2301,
  [9571] = 2313,
  [9572] = 2373,
  [9573] = 2283,
  [9574] = 9574,
  [9575] = 9575,
  [9576] = 9576,
  [9577] = 9577,
  [9578] = 9578,
  [9579] = 2383,
  [9580] = 2272,
  [9581] = 9581,
  [9582] = 9575,
  [9583] = 2258,
  [9584] = 9576,
  [9585] = 2256,
  [9586] = 9575,
  [9587] = 9576,
  [9588] = 9588,
  [9589] = 9589,
  [9590] = 2259,
  [9591] = 9576,
  [9592] = 9575,
  [9593] = 2276,
  [9594] = 9576,
  [9595] = 9595,
  [9596] = 2285,
  [9597] = 9575,
  [9598] = 2353,
  [9599] = 9576,
  [9600] = 9600,
  [9601] = 9564,
  [9602] = 2281,
  [9603] = 9575,
  [9604] = 9576,
  [9605] = 9605,
  [9606] = 9606,
  [9607] = 9575,
  [9608] = 2372,
  [9609] = 9609,
  [9610] = 9576,
  [9611] = 2299,
  [9612] = 2287,
  [9613] = 9613,
  [9614] = 9575,
  [9615] = 9595,
  [9616] = 2282,
  [9617] = 2383,
  [9618] = 2537,
  [9619] = 2539,
  [9620] = 2485,
  [9621] = 2544,
  [9622] = 2545,
  [9623] = 2547,
  [9624] = 2550,
  [9625] = 2551,
  [9626] = 2552,
  [9627] = 2555,
  [9628] = 2316,
  [9629] = 2558,
  [9630] = 2560,
  [9631] = 2561,
  [9632] = 2562,
  [9633] = 2563,
  [9634] = 2503,
  [9635] = 2504,
  [9636] = 2525,
  [9637] = 2530,
  [9638] = 2535,
  [9639] = 2536,
  [9640] = 2540,
  [9641] = 9641,
  [9642] = 2542,
  [9643] = 2572,
  [9644] = 2573,
  [9645] = 2574,
  [9646] = 2575,
  [9647] = 2288,
  [9648] = 2313,
  [9649] = 2289,
  [9650] = 2582,
  [9651] = 9651,
  [9652] = 9541,
  [9653] = 2583,
  [9654] = 9595,
  [9655] = 9655,
  [9656] = 9548,
  [9657] = 9657,
  [9658] = 9658,
  [9659] = 9659,
  [9660] = 9660,
  [9661] = 9661,
  [9662] = 9613,
  [9663] = 2278,
  [9664] = 2284,
  [9665] = 2290,
  [9666] = 2272,
  [9667] = 9667,
  [9668] = 2369,
  [9669] = 2258,
  [9670] = 2259,
  [9671] = 2588,
  [9672] = 2589,
  [9673] = 9673,
  [9674] = 2256,
  [9675] = 9675,
  [9676] = 2255,
  [9677] = 2328,
  [9678] = 2372,
  [9679] = 2353,
  [9680] = 2445,
  [9681] = 9541,
  [9682] = 9613,
  [9683] = 2255,
  [9684] = 2373,
  [9685] = 2556,
  [9686] = 9686,
  [9687] = 2589,
  [9688] = 9641,
  [9689] = 9689,
  [9690] = 9690,
  [9691] = 9588,
  [9692] = 2547,
  [9693] = 9693,
  [9694] = 9694,
  [9695] = 2259,
  [9696] = 9595,
  [9697] = 9651,
  [9698] = 2550,
  [9699] = 2469,
  [9700] = 9686,
  [9701] = 9574,
  [9702] = 9577,
  [9703] = 9690,
  [9704] = 9657,
  [9705] = 9578,
  [9706] = 9589,
  [9707] = 9693,
  [9708] = 9605,
  [9709] = 9606,
  [9710] = 2560,
  [9711] = 2561,
  [9712] = 2292,
  [9713] = 2551,
  [9714] = 2299,
  [9715] = 9690,
  [9716] = 9600,
  [9717] = 9717,
  [9718] = 9693,
  [9719] = 9595,
  [9720] = 2372,
  [9721] = 9564,
  [9722] = 2472,
  [9723] = 9588,
  [9724] = 2446,
  [9725] = 9600,
  [9726] = 9726,
  [9727] = 9675,
  [9728] = 9641,
  [9729] = 9651,
  [9730] = 2562,
  [9731] = 9655,
  [9732] = 9657,
  [9733] = 9658,
  [9734] = 9734,
  [9735] = 2286,
  [9736] = 2563,
  [9737] = 2552,
  [9738] = 9738,
  [9739] = 9541,
  [9740] = 2503,
  [9741] = 9741,
  [9742] = 9742,
  [9743] = 9743,
  [9744] = 9744,
  [9745] = 9745,
  [9746] = 2542,
  [9747] = 9693,
  [9748] = 2383,
  [9749] = 2373,
  [9750] = 9658,
  [9751] = 9751,
  [9752] = 9752,
  [9753] = 2255,
  [9754] = 9754,
  [9755] = 2399,
  [9756] = 2504,
  [9757] = 2575,
  [9758] = 2525,
  [9759] = 2530,
  [9760] = 2258,
  [9761] = 9686,
  [9762] = 9613,
  [9763] = 9661,
  [9764] = 9667,
  [9765] = 9673,
  [9766] = 9766,
  [9767] = 2473,
  [9768] = 9690,
  [9769] = 2555,
  [9770] = 2572,
  [9771] = 2582,
  [9772] = 2535,
  [9773] = 9693,
  [9774] = 2583,
  [9775] = 2445,
  [9776] = 9541,
  [9777] = 2353,
  [9778] = 2573,
  [9779] = 2272,
  [9780] = 9675,
  [9781] = 2292,
  [9782] = 9782,
  [9783] = 2574,
  [9784] = 9690,
  [9785] = 2536,
  [9786] = 9786,
  [9787] = 2537,
  [9788] = 2539,
  [9789] = 9693,
  [9790] = 2485,
  [9791] = 2544,
  [9792] = 9574,
  [9793] = 2313,
  [9794] = 2545,
  [9795] = 2547,
  [9796] = 9782,
  [9797] = 9655,
  [9798] = 9613,
  [9799] = 2551,
  [9800] = 9577,
  [9801] = 2552,
  [9802] = 2555,
  [9803] = 2556,
  [9804] = 2558,
  [9805] = 2560,
  [9806] = 2561,
  [9807] = 2562,
  [9808] = 2563,
  [9809] = 2503,
  [9810] = 2504,
  [9811] = 2525,
  [9812] = 2530,
  [9813] = 2556,
  [9814] = 9661,
  [9815] = 2535,
  [9816] = 2536,
  [9817] = 2540,
  [9818] = 9690,
  [9819] = 9690,
  [9820] = 9820,
  [9821] = 2558,
  [9822] = 2542,
  [9823] = 9578,
  [9824] = 2572,
  [9825] = 9589,
  [9826] = 2573,
  [9827] = 2574,
  [9828] = 9693,
  [9829] = 2575,
  [9830] = 2369,
  [9831] = 2256,
  [9832] = 9667,
  [9833] = 2313,
  [9834] = 9673,
  [9835] = 9835,
  [9836] = 9605,
  [9837] = 9606,
  [9838] = 2445,
  [9839] = 9839,
  [9840] = 9564,
  [9841] = 2299,
  [9842] = 2467,
  [9843] = 9843,
  [9844] = 2582,
  [9845] = 2583,
  [9846] = 2588,
  [9847] = 2589,
  [9848] = 9848,
  [9849] = 2372,
  [9850] = 9564,
  [9851] = 9613,
  [9852] = 9595,
  [9853] = 9853,
  [9854] = 2443,
  [9855] = 2540,
  [9856] = 2258,
  [9857] = 2537,
  [9858] = 2539,
  [9859] = 2259,
  [9860] = 9690,
  [9861] = 2485,
  [9862] = 2544,
  [9863] = 9863,
  [9864] = 9693,
  [9865] = 2588,
  [9866] = 9782,
  [9867] = 2545,
  [9868] = 9868,
  [9869] = 9686,
  [9870] = 9686,
  [9871] = 9686,
  [9872] = 9686,
  [9873] = 9686,
  [9874] = 9686,
  [9875] = 9686,
  [9876] = 9686,
  [9877] = 9686,
  [9878] = 9686,
  [9879] = 9686,
  [9880] = 9686,
  [9881] = 9686,
  [9882] = 9686,
  [9883] = 9686,
  [9884] = 9686,
  [9885] = 9686,
  [9886] = 9686,
  [9887] = 9686,
  [9888] = 9686,
  [9889] = 9686,
  [9890] = 9686,
  [9891] = 9686,
  [9892] = 9686,
  [9893] = 2550,
  [9894] = 9894,
  [9895] = 9895,
  [9896] = 9895,
  [9897] = 9897,
  [9898] = 2399,
  [9899] = 9895,
  [9900] = 9743,
  [9901] = 9745,
  [9902] = 9895,
  [9903] = 2286,
  [9904] = 9895,
  [9905] = 2582,
  [9906] = 2583,
  [9907] = 9895,
  [9908] = 9751,
  [9909] = 9741,
  [9910] = 9897,
  [9911] = 9742,
  [9912] = 9744,
  [9913] = 9895,
  [9914] = 9574,
  [9915] = 9577,
  [9916] = 9578,
  [9917] = 9589,
  [9918] = 9605,
  [9919] = 9606,
  [9920] = 9751,
  [9921] = 2588,
  [9922] = 2589,
  [9923] = 9868,
  [9924] = 9541,
  [9925] = 9895,
  [9926] = 9897,
  [9927] = 2258,
  [9928] = 2537,
  [9929] = 9595,
  [9930] = 2539,
  [9931] = 2467,
  [9932] = 2485,
  [9933] = 2544,
  [9934] = 2545,
  [9935] = 2547,
  [9936] = 2550,
  [9937] = 2551,
  [9938] = 2552,
  [9939] = 2555,
  [9940] = 2556,
  [9941] = 2558,
  [9942] = 2560,
  [9943] = 2561,
  [9944] = 2562,
  [9945] = 2563,
  [9946] = 2503,
  [9947] = 2504,
  [9948] = 2525,
  [9949] = 2530,
  [9950] = 2535,
  [9951] = 2536,
  [9952] = 2292,
  [9953] = 2540,
  [9954] = 9843,
  [9955] = 9848,
  [9956] = 9853,
  [9957] = 9689,
  [9958] = 2542,
  [9959] = 9894,
  [9960] = 2572,
  [9961] = 2573,
  [9962] = 2574,
  [9963] = 9651,
  [9964] = 2469,
  [9965] = 9843,
  [9966] = 2446,
  [9967] = 9820,
  [9968] = 9835,
  [9969] = 9894,
  [9970] = 9895,
  [9971] = 9897,
  [9972] = 9694,
  [9973] = 2537,
  [9974] = 2575,
  [9975] = 2539,
  [9976] = 9894,
  [9977] = 9782,
  [9978] = 9655,
  [9979] = 2472,
  [9980] = 2485,
  [9981] = 2544,
  [9982] = 9894,
  [9983] = 2545,
  [9984] = 9848,
  [9985] = 9895,
  [9986] = 9853,
  [9987] = 2299,
  [9988] = 2473,
  [9989] = 9897,
  [9990] = 2399,
  [9991] = 9894,
  [9992] = 2547,
  [9993] = 2550,
  [9994] = 9743,
  [9995] = 9745,
  [9996] = 9752,
  [9997] = 9588,
  [9998] = 9754,
  [9999] = 2551,
  [10000] = 9600,
  [10001] = 2552,
  [10002] = 2555,
  [10003] = 2443,
  [10004] = 9766,
  [10005] = 2556,
  [10006] = 2558,
  [10007] = 2560,
  [10008] = 2561,
  [10009] = 2562,
  [10010] = 2563,
  [10011] = 9689,
  [10012] = 2503,
  [10013] = 2504,
  [10014] = 2525,
  [10015] = 2530,
  [10016] = 9894,
  [10017] = 9895,
  [10018] = 9895,
  [10019] = 9897,
  [10020] = 2536,
  [10021] = 9717,
  [10022] = 9734,
  [10023] = 9786,
  [10024] = 9641,
  [10025] = 9675,
  [10026] = 2540,
  [10027] = 9820,
  [10028] = 9835,
  [10029] = 9895,
  [10030] = 9894,
  [10031] = 9868,
  [10032] = 9651,
  [10033] = 9895,
  [10034] = 9895,
  [10035] = 9897,
  [10036] = 2372,
  [10037] = 2467,
  [10038] = 2542,
  [10039] = 9895,
  [10040] = 2445,
  [10041] = 9897,
  [10042] = 9694,
  [10043] = 2572,
  [10044] = 9655,
  [10045] = 2573,
  [10046] = 2574,
  [10047] = 9657,
  [10048] = 9897,
  [10049] = 9658,
  [10050] = 9895,
  [10051] = 9744,
  [10052] = 9657,
  [10053] = 2575,
  [10054] = 2469,
  [10055] = 9658,
  [10056] = 9588,
  [10057] = 9741,
  [10058] = 9895,
  [10059] = 2472,
  [10060] = 2286,
  [10061] = 9600,
  [10062] = 9675,
  [10063] = 9895,
  [10064] = 9897,
  [10065] = 2473,
  [10066] = 9752,
  [10067] = 10067,
  [10068] = 9754,
  [10069] = 9895,
  [10070] = 9766,
  [10071] = 2582,
  [10072] = 2583,
  [10073] = 2443,
  [10074] = 9564,
  [10075] = 9895,
  [10076] = 9895,
  [10077] = 9661,
  [10078] = 9897,
  [10079] = 9667,
  [10080] = 9673,
  [10081] = 9895,
  [10082] = 2446,
  [10083] = 2259,
  [10084] = 9742,
  [10085] = 2313,
  [10086] = 2588,
  [10087] = 2589,
  [10088] = 9782,
  [10089] = 9786,
  [10090] = 9574,
  [10091] = 9577,
  [10092] = 9578,
  [10093] = 9589,
  [10094] = 9605,
  [10095] = 9641,
  [10096] = 9606,
  [10097] = 9661,
  [10098] = 9726,
  [10099] = 9738,
  [10100] = 9895,
  [10101] = 9895,
  [10102] = 9839,
  [10103] = 9863,
  [10104] = 9667,
  [10105] = 9897,
  [10106] = 2445,
  [10107] = 9673,
  [10108] = 2292,
  [10109] = 2299,
  [10110] = 9895,
  [10111] = 9717,
  [10112] = 9726,
  [10113] = 9738,
  [10114] = 9839,
  [10115] = 9863,
  [10116] = 9734,
  [10117] = 9613,
  [10118] = 9895,
  [10119] = 2535,
  [10120] = 2537,
  [10121] = 9863,
  [10122] = 9868,
  [10123] = 2562,
  [10124] = 9754,
  [10125] = 9694,
  [10126] = 2545,
  [10127] = 2589,
  [10128] = 2563,
  [10129] = 2503,
  [10130] = 3790,
  [10131] = 2399,
  [10132] = 9786,
  [10133] = 2446,
  [10134] = 2550,
  [10135] = 3790,
  [10136] = 2286,
  [10137] = 9717,
  [10138] = 2540,
  [10139] = 9651,
  [10140] = 9600,
  [10141] = 4222,
  [10142] = 9734,
  [10143] = 9853,
  [10144] = 2551,
  [10145] = 2504,
  [10146] = 2292,
  [10147] = 2399,
  [10148] = 9868,
  [10149] = 2525,
  [10150] = 9661,
  [10151] = 9820,
  [10152] = 9835,
  [10153] = 9689,
  [10154] = 2473,
  [10155] = 9741,
  [10156] = 10156,
  [10157] = 9742,
  [10158] = 2542,
  [10159] = 9744,
  [10160] = 9673,
  [10161] = 2572,
  [10162] = 9820,
  [10163] = 3853,
  [10164] = 2573,
  [10165] = 2574,
  [10166] = 2582,
  [10167] = 10167,
  [10168] = 9734,
  [10169] = 2469,
  [10170] = 10167,
  [10171] = 3853,
  [10172] = 2547,
  [10173] = 9743,
  [10174] = 2583,
  [10175] = 2552,
  [10176] = 9574,
  [10177] = 9577,
  [10178] = 9578,
  [10179] = 9589,
  [10180] = 9745,
  [10181] = 9605,
  [10182] = 9606,
  [10183] = 9717,
  [10184] = 9843,
  [10185] = 2575,
  [10186] = 9766,
  [10187] = 9835,
  [10188] = 9751,
  [10189] = 9655,
  [10190] = 2299,
  [10191] = 2530,
  [10192] = 2469,
  [10193] = 2473,
  [10194] = 9726,
  [10195] = 9675,
  [10196] = 9848,
  [10197] = 2556,
  [10198] = 9738,
  [10199] = 9641,
  [10200] = 2472,
  [10201] = 9839,
  [10202] = 9863,
  [10203] = 2443,
  [10204] = 9752,
  [10205] = 9754,
  [10206] = 9766,
  [10207] = 9657,
  [10208] = 9658,
  [10209] = 9782,
  [10210] = 9843,
  [10211] = 9848,
  [10212] = 9853,
  [10213] = 2535,
  [10214] = 2558,
  [10215] = 9689,
  [10216] = 2539,
  [10217] = 2286,
  [10218] = 2536,
  [10219] = 9752,
  [10220] = 2467,
  [10221] = 2445,
  [10222] = 2560,
  [10223] = 2472,
  [10224] = 2446,
  [10225] = 9743,
  [10226] = 2467,
  [10227] = 9745,
  [10228] = 4219,
  [10229] = 2588,
  [10230] = 10230,
  [10231] = 2485,
  [10232] = 9726,
  [10233] = 2561,
  [10234] = 9738,
  [10235] = 2443,
  [10236] = 2544,
  [10237] = 9667,
  [10238] = 9786,
  [10239] = 9694,
  [10240] = 9588,
  [10241] = 9741,
  [10242] = 9742,
  [10243] = 9744,
  [10244] = 9751,
  [10245] = 9839,
  [10246] = 2555,
  [10247] = 10247,
  [10248] = 2234,
  [10249] = 2241,
  [10250] = 2446,
  [10251] = 2443,
  [10252] = 2473,
  [10253] = 9726,
  [10254] = 8986,
  [10255] = 10247,
  [10256] = 9738,
  [10257] = 9766,
  [10258] = 9741,
  [10259] = 9742,
  [10260] = 2399,
  [10261] = 9868,
  [10262] = 9744,
  [10263] = 10263,
  [10264] = 10264,
  [10265] = 9751,
  [10266] = 2467,
  [10267] = 3853,
  [10268] = 9745,
  [10269] = 10264,
  [10270] = 9839,
  [10271] = 2469,
  [10272] = 2472,
  [10273] = 9786,
  [10274] = 10247,
  [10275] = 2236,
  [10276] = 2259,
  [10277] = 10264,
  [10278] = 10278,
  [10279] = 9820,
  [10280] = 9835,
  [10281] = 9863,
  [10282] = 2258,
  [10283] = 10247,
  [10284] = 10264,
  [10285] = 10264,
  [10286] = 9843,
  [10287] = 3790,
  [10288] = 10247,
  [10289] = 9848,
  [10290] = 10263,
  [10291] = 9853,
  [10292] = 10247,
  [10293] = 2240,
  [10294] = 9743,
  [10295] = 10278,
  [10296] = 2263,
  [10297] = 10264,
  [10298] = 2286,
  [10299] = 10264,
  [10300] = 9717,
  [10301] = 9689,
  [10302] = 10264,
  [10303] = 4219,
  [10304] = 10247,
  [10305] = 4222,
  [10306] = 9752,
  [10307] = 9754,
  [10308] = 9734,
  [10309] = 10247,
  [10310] = 9694,
  [10311] = 10311,
  [10312] = 10312,
  [10313] = 10312,
  [10314] = 2263,
  [10315] = 10315,
  [10316] = 10311,
  [10317] = 2240,
  [10318] = 10311,
  [10319] = 10312,
  [10320] = 2236,
  [10321] = 2234,
  [10322] = 2241,
  [10323] = 10323,
  [10324] = 10312,
  [10325] = 10325,
  [10326] = 2292,
  [10327] = 10323,
  [10328] = 10311,
  [10329] = 2299,
  [10330] = 10311,
  [10331] = 10312,
  [10332] = 10323,
  [10333] = 10325,
  [10334] = 2255,
  [10335] = 10323,
  [10336] = 4222,
  [10337] = 10323,
  [10338] = 2259,
  [10339] = 10312,
  [10340] = 10311,
  [10341] = 8986,
  [10342] = 10325,
  [10343] = 10311,
  [10344] = 10325,
  [10345] = 4219,
  [10346] = 10312,
  [10347] = 10347,
  [10348] = 10311,
  [10349] = 10323,
  [10350] = 10325,
  [10351] = 10323,
  [10352] = 10325,
  [10353] = 2272,
  [10354] = 10325,
  [10355] = 4222,
  [10356] = 4219,
  [10357] = 10312,
  [10358] = 2258,
  [10359] = 10323,
  [10360] = 10325,
  [10361] = 2256,
  [10362] = 10362,
  [10363] = 2313,
  [10364] = 2255,
  [10365] = 4219,
  [10366] = 2277,
  [10367] = 2264,
  [10368] = 4222,
  [10369] = 10369,
  [10370] = 2265,
  [10371] = 4827,
  [10372] = 2286,
  [10373] = 2292,
  [10374] = 4222,
  [10375] = 10369,
  [10376] = 2256,
  [10377] = 4219,
  [10378] = 2299,
  [10379] = 2272,
  [10380] = 2286,
  [10381] = 2236,
  [10382] = 4816,
  [10383] = 2258,
  [10384] = 2259,
  [10385] = 2265,
  [10386] = 2443,
  [10387] = 2277,
  [10388] = 10388,
  [10389] = 10388,
  [10390] = 4807,
  [10391] = 2258,
  [10392] = 10388,
  [10393] = 10393,
  [10394] = 5668,
  [10395] = 2234,
  [10396] = 10396,
  [10397] = 10388,
  [10398] = 2469,
  [10399] = 4827,
  [10400] = 2241,
  [10401] = 2301,
  [10402] = 10388,
  [10403] = 2467,
  [10404] = 2276,
  [10405] = 2281,
  [10406] = 2282,
  [10407] = 2236,
  [10408] = 2263,
  [10409] = 2313,
  [10410] = 4942,
  [10411] = 2284,
  [10412] = 4923,
  [10413] = 10413,
  [10414] = 10413,
  [10415] = 2473,
  [10416] = 2285,
  [10417] = 5673,
  [10418] = 4924,
  [10419] = 10419,
  [10420] = 2287,
  [10421] = 10388,
  [10422] = 5671,
  [10423] = 2234,
  [10424] = 4942,
  [10425] = 2263,
  [10426] = 2288,
  [10427] = 2240,
  [10428] = 4222,
  [10429] = 4219,
  [10430] = 2221,
  [10431] = 2472,
  [10432] = 2236,
  [10433] = 4923,
  [10434] = 2283,
  [10435] = 5672,
  [10436] = 2234,
  [10437] = 10388,
  [10438] = 10413,
  [10439] = 5669,
  [10440] = 4924,
  [10441] = 10388,
  [10442] = 2258,
  [10443] = 2278,
  [10444] = 2259,
  [10445] = 2316,
  [10446] = 2241,
  [10447] = 10388,
  [10448] = 2289,
  [10449] = 10413,
  [10450] = 2241,
  [10451] = 10388,
  [10452] = 2264,
  [10453] = 10413,
  [10454] = 4816,
  [10455] = 4827,
  [10456] = 10413,
  [10457] = 2290,
  [10458] = 10388,
  [10459] = 2259,
  [10460] = 10413,
  [10461] = 4827,
  [10462] = 2312,
  [10463] = 10413,
  [10464] = 5093,
  [10465] = 4807,
  [10466] = 2263,
  [10467] = 10388,
  [10468] = 2240,
  [10469] = 2240,
  [10470] = 4923,
  [10471] = 2255,
  [10472] = 10472,
  [10473] = 5669,
  [10474] = 10474,
  [10475] = 2258,
  [10476] = 10476,
  [10477] = 10477,
  [10478] = 2272,
  [10479] = 10474,
  [10480] = 4816,
  [10481] = 10472,
  [10482] = 4827,
  [10483] = 10483,
  [10484] = 4827,
  [10485] = 2289,
  [10486] = 10477,
  [10487] = 2258,
  [10488] = 10476,
  [10489] = 10472,
  [10490] = 2256,
  [10491] = 10474,
  [10492] = 2292,
  [10493] = 10476,
  [10494] = 10494,
  [10495] = 10495,
  [10496] = 2292,
  [10497] = 2299,
  [10498] = 10472,
  [10499] = 5093,
  [10500] = 2236,
  [10501] = 2259,
  [10502] = 4807,
  [10503] = 10483,
  [10504] = 2234,
  [10505] = 2241,
  [10506] = 10474,
  [10507] = 2259,
  [10508] = 2276,
  [10509] = 5671,
  [10510] = 2301,
  [10511] = 10495,
  [10512] = 2256,
  [10513] = 5672,
  [10514] = 2281,
  [10515] = 2282,
  [10516] = 10476,
  [10517] = 2272,
  [10518] = 10476,
  [10519] = 10519,
  [10520] = 2284,
  [10521] = 10483,
  [10522] = 10477,
  [10523] = 10472,
  [10524] = 10477,
  [10525] = 10474,
  [10526] = 10477,
  [10527] = 2256,
  [10528] = 10477,
  [10529] = 10519,
  [10530] = 10474,
  [10531] = 10494,
  [10532] = 5673,
  [10533] = 2290,
  [10534] = 10477,
  [10535] = 10495,
  [10536] = 5093,
  [10537] = 2255,
  [10538] = 2283,
  [10539] = 10476,
  [10540] = 2263,
  [10541] = 2285,
  [10542] = 2328,
  [10543] = 2255,
  [10544] = 10495,
  [10545] = 10545,
  [10546] = 2299,
  [10547] = 2287,
  [10548] = 10483,
  [10549] = 10476,
  [10550] = 10483,
  [10551] = 10476,
  [10552] = 10495,
  [10553] = 10519,
  [10554] = 10494,
  [10555] = 2272,
  [10556] = 5093,
  [10557] = 5668,
  [10558] = 10483,
  [10559] = 10519,
  [10560] = 2256,
  [10561] = 4924,
  [10562] = 2328,
  [10563] = 10476,
  [10564] = 2255,
  [10565] = 2255,
  [10566] = 10483,
  [10567] = 10519,
  [10568] = 10494,
  [10569] = 2316,
  [10570] = 2236,
  [10571] = 10519,
  [10572] = 2288,
  [10573] = 10494,
  [10574] = 10474,
  [10575] = 2272,
  [10576] = 10472,
  [10577] = 2256,
  [10578] = 2278,
  [10579] = 10476,
  [10580] = 10477,
  [10581] = 2312,
  [10582] = 2240,
  [10583] = 10474,
  [10584] = 4942,
  [10585] = 10483,
  [10586] = 2263,
  [10587] = 10477,
  [10588] = 10494,
  [10589] = 2241,
  [10590] = 10476,
  [10591] = 2234,
  [10592] = 2240,
  [10593] = 10494,
  [10594] = 10476,
  [10595] = 10494,
  [10596] = 10472,
  [10597] = 2292,
  [10598] = 10483,
  [10599] = 10519,
  [10600] = 10495,
  [10601] = 10519,
  [10602] = 10495,
  [10603] = 2272,
  [10604] = 10472,
  [10605] = 2299,
  [10606] = 2259,
  [10607] = 10607,
  [10608] = 2277,
  [10609] = 2263,
  [10610] = 10607,
  [10611] = 2299,
  [10612] = 2264,
  [10613] = 10607,
  [10614] = 2264,
  [10615] = 2272,
  [10616] = 2255,
  [10617] = 2225,
  [10618] = 5093,
  [10619] = 2241,
  [10620] = 10607,
  [10621] = 2240,
  [10622] = 10607,
  [10623] = 5671,
  [10624] = 2259,
  [10625] = 4827,
  [10626] = 2313,
  [10627] = 5672,
  [10628] = 5672,
  [10629] = 10607,
  [10630] = 2258,
  [10631] = 2373,
  [10632] = 2258,
  [10633] = 2265,
  [10634] = 2313,
  [10635] = 9541,
  [10636] = 2286,
  [10637] = 5093,
  [10638] = 2369,
  [10639] = 2383,
  [10640] = 2256,
  [10641] = 5673,
  [10642] = 10607,
  [10643] = 2241,
  [10644] = 10607,
  [10645] = 4816,
  [10646] = 10607,
  [10647] = 10607,
  [10648] = 2373,
  [10649] = 10649,
  [10650] = 2369,
  [10651] = 5671,
  [10652] = 10607,
  [10653] = 2383,
  [10654] = 2313,
  [10655] = 2259,
  [10656] = 4942,
  [10657] = 2313,
  [10658] = 10607,
  [10659] = 5668,
  [10660] = 2286,
  [10661] = 2234,
  [10662] = 5673,
  [10663] = 10607,
  [10664] = 10607,
  [10665] = 2240,
  [10666] = 2236,
  [10667] = 2240,
  [10668] = 2240,
  [10669] = 6031,
  [10670] = 2299,
  [10671] = 5669,
  [10672] = 9541,
  [10673] = 2265,
  [10674] = 2353,
  [10675] = 2264,
  [10676] = 2258,
  [10677] = 2353,
  [10678] = 4807,
  [10679] = 2258,
  [10680] = 5669,
  [10681] = 2277,
  [10682] = 2272,
  [10683] = 2263,
  [10684] = 8740,
  [10685] = 10607,
  [10686] = 2256,
  [10687] = 10687,
  [10688] = 2292,
  [10689] = 2234,
  [10690] = 2258,
  [10691] = 2255,
  [10692] = 2313,
  [10693] = 2236,
  [10694] = 10694,
  [10695] = 2292,
  [10696] = 10607,
  [10697] = 2259,
  [10698] = 2263,
  [10699] = 2259,
  [10700] = 2286,
  [10701] = 2234,
  [10702] = 4923,
  [10703] = 2241,
  [10704] = 2259,
  [10705] = 4924,
  [10706] = 2236,
  [10707] = 2241,
  [10708] = 2277,
  [10709] = 2258,
  [10710] = 2234,
  [10711] = 2265,
  [10712] = 2263,
  [10713] = 10607,
  [10714] = 5668,
  [10715] = 2236,
  [10716] = 8733,
  [10717] = 2292,
  [10718] = 10718,
  [10719] = 10719,
  [10720] = 10720,
  [10721] = 10718,
  [10722] = 2258,
  [10723] = 10719,
  [10724] = 10720,
  [10725] = 10718,
  [10726] = 10726,
  [10727] = 10727,
  [10728] = 10719,
  [10729] = 10719,
  [10730] = 10720,
  [10731] = 10718,
  [10732] = 10726,
  [10733] = 10719,
  [10734] = 10720,
  [10735] = 10718,
  [10736] = 10736,
  [10737] = 10719,
  [10738] = 10720,
  [10739] = 10718,
  [10740] = 10719,
  [10741] = 10720,
  [10742] = 10718,
  [10743] = 10726,
  [10744] = 10719,
  [10745] = 10720,
  [10746] = 10718,
  [10747] = 10726,
  [10748] = 10748,
  [10749] = 10719,
  [10750] = 10720,
  [10751] = 10718,
  [10752] = 10726,
  [10753] = 10719,
  [10754] = 10720,
  [10755] = 10718,
  [10756] = 10726,
  [10757] = 10719,
  [10758] = 10720,
  [10759] = 10718,
  [10760] = 10726,
  [10761] = 10719,
  [10762] = 10720,
  [10763] = 10718,
  [10764] = 10726,
  [10765] = 2259,
  [10766] = 10766,
  [10767] = 10719,
  [10768] = 10720,
  [10769] = 10718,
  [10770] = 10726,
  [10771] = 10719,
  [10772] = 10720,
  [10773] = 10718,
  [10774] = 10726,
  [10775] = 10775,
  [10776] = 10719,
  [10777] = 10720,
  [10778] = 10718,
  [10779] = 10719,
  [10780] = 10720,
  [10781] = 10718,
  [10782] = 10726,
  [10783] = 2277,
  [10784] = 10719,
  [10785] = 10720,
  [10786] = 10718,
  [10787] = 10726,
  [10788] = 10788,
  [10789] = 10719,
  [10790] = 10720,
  [10791] = 10718,
  [10792] = 10726,
  [10793] = 10793,
  [10794] = 10719,
  [10795] = 10720,
  [10796] = 10718,
  [10797] = 10726,
  [10798] = 2312,
  [10799] = 10719,
  [10800] = 10719,
  [10801] = 10718,
  [10802] = 10726,
  [10803] = 2301,
  [10804] = 10719,
  [10805] = 10720,
  [10806] = 10718,
  [10807] = 10726,
  [10808] = 2255,
  [10809] = 10809,
  [10810] = 10719,
  [10811] = 10720,
  [10812] = 10718,
  [10813] = 10726,
  [10814] = 10814,
  [10815] = 10719,
  [10816] = 10720,
  [10817] = 10718,
  [10818] = 10726,
  [10819] = 10819,
  [10820] = 10719,
  [10821] = 10720,
  [10822] = 10718,
  [10823] = 10726,
  [10824] = 10775,
  [10825] = 10719,
  [10826] = 10720,
  [10827] = 10718,
  [10828] = 10726,
  [10829] = 10829,
  [10830] = 10719,
  [10831] = 10720,
  [10832] = 10718,
  [10833] = 10726,
  [10834] = 10720,
  [10835] = 2283,
  [10836] = 10719,
  [10837] = 10720,
  [10838] = 10718,
  [10839] = 10726,
  [10840] = 10718,
  [10841] = 10841,
  [10842] = 10719,
  [10843] = 10720,
  [10844] = 10718,
  [10845] = 10726,
  [10846] = 10846,
  [10847] = 10847,
  [10848] = 10719,
  [10849] = 10720,
  [10850] = 10718,
  [10851] = 10726,
  [10852] = 10852,
  [10853] = 10829,
  [10854] = 10719,
  [10855] = 10720,
  [10856] = 10718,
  [10857] = 10726,
  [10858] = 10858,
  [10859] = 10719,
  [10860] = 10720,
  [10861] = 10718,
  [10862] = 10719,
  [10863] = 10720,
  [10864] = 10718,
  [10865] = 10726,
  [10866] = 10866,
  [10867] = 10719,
  [10868] = 10720,
  [10869] = 10718,
  [10870] = 10719,
  [10871] = 10719,
  [10872] = 10720,
  [10873] = 10718,
  [10874] = 10726,
  [10875] = 10875,
  [10876] = 10719,
  [10877] = 10720,
  [10878] = 10718,
  [10879] = 10726,
  [10880] = 10720,
  [10881] = 10719,
  [10882] = 10720,
  [10883] = 10718,
  [10884] = 10726,
  [10885] = 10719,
  [10886] = 10720,
  [10887] = 10718,
  [10888] = 10726,
  [10889] = 10719,
  [10890] = 10720,
  [10891] = 10718,
  [10892] = 10726,
  [10893] = 10893,
  [10894] = 10894,
  [10895] = 10719,
  [10896] = 10720,
  [10897] = 10718,
  [10898] = 10726,
  [10899] = 10718,
  [10900] = 10719,
  [10901] = 10720,
  [10902] = 10718,
  [10903] = 10726,
  [10904] = 10727,
  [10905] = 10719,
  [10906] = 10720,
  [10907] = 10718,
  [10908] = 10726,
  [10909] = 10719,
  [10910] = 10720,
  [10911] = 10718,
  [10912] = 10726,
  [10913] = 10894,
  [10914] = 10719,
  [10915] = 10720,
  [10916] = 10718,
  [10917] = 10726,
  [10918] = 10719,
  [10919] = 10720,
  [10920] = 10718,
  [10921] = 10726,
  [10922] = 6031,
  [10923] = 10719,
  [10924] = 10720,
  [10925] = 10718,
  [10926] = 10726,
  [10927] = 10719,
  [10928] = 10720,
  [10929] = 10718,
  [10930] = 10726,
  [10931] = 10736,
  [10932] = 10719,
  [10933] = 10720,
  [10934] = 10718,
  [10935] = 10726,
  [10936] = 2255,
  [10937] = 10766,
  [10938] = 10719,
  [10939] = 10720,
  [10940] = 10718,
  [10941] = 10719,
  [10942] = 10720,
  [10943] = 10718,
  [10944] = 10719,
  [10945] = 10720,
  [10946] = 10718,
  [10947] = 10788,
  [10948] = 10719,
  [10949] = 10720,
  [10950] = 10718,
  [10951] = 10726,
  [10952] = 2272,
  [10953] = 10719,
  [10954] = 10720,
  [10955] = 10718,
  [10956] = 10809,
  [10957] = 10719,
  [10958] = 10720,
  [10959] = 10718,
  [10960] = 10726,
  [10961] = 2299,
  [10962] = 10719,
  [10963] = 10720,
  [10964] = 10718,
  [10965] = 10726,
  [10966] = 10841,
  [10967] = 10846,
  [10968] = 10719,
  [10969] = 10720,
  [10970] = 10718,
  [10971] = 10875,
  [10972] = 10719,
  [10973] = 10720,
  [10974] = 10718,
  [10975] = 2292,
  [10976] = 10719,
  [10977] = 10720,
  [10978] = 10718,
  [10979] = 2264,
  [10980] = 10847,
  [10981] = 10829,
  [10982] = 10788,
  [10983] = 2299,
  [10984] = 10719,
  [10985] = 10985,
  [10986] = 2255,
  [10987] = 10719,
  [10988] = 10866,
  [10989] = 10829,
  [10990] = 10720,
  [10991] = 10718,
  [10992] = 10726,
  [10993] = 10993,
  [10994] = 10847,
  [10995] = 10995,
  [10996] = 10775,
  [10997] = 10997,
  [10998] = 10829,
  [10999] = 10858,
  [11000] = 10894,
  [11001] = 10727,
  [11002] = 10720,
  [11003] = 10736,
  [11004] = 2256,
  [11005] = 11005,
  [11006] = 10766,
  [11007] = 10718,
  [11008] = 10726,
  [11009] = 10788,
  [11010] = 2265,
  [11011] = 10809,
  [11012] = 11012,
  [11013] = 10841,
  [11014] = 11014,
  [11015] = 11015,
  [11016] = 10846,
  [11017] = 5668,
  [11018] = 10847,
  [11019] = 10875,
  [11020] = 10858,
  [11021] = 10866,
  [11022] = 2258,
  [11023] = 10997,
  [11024] = 10993,
  [11025] = 2259,
  [11026] = 10719,
  [11027] = 10997,
  [11028] = 2272,
  [11029] = 10829,
  [11030] = 11005,
  [11031] = 10720,
  [11032] = 10718,
  [11033] = 10847,
  [11034] = 11012,
  [11035] = 11014,
  [11036] = 11015,
  [11037] = 10866,
  [11038] = 10993,
  [11039] = 10894,
  [11040] = 10727,
  [11041] = 10736,
  [11042] = 10766,
  [11043] = 10788,
  [11044] = 10809,
  [11045] = 10841,
  [11046] = 10846,
  [11047] = 10727,
  [11048] = 11012,
  [11049] = 10875,
  [11050] = 11014,
  [11051] = 11015,
  [11052] = 10866,
  [11053] = 10726,
  [11054] = 10993,
  [11055] = 10894,
  [11056] = 10719,
  [11057] = 2316,
  [11058] = 10995,
  [11059] = 2278,
  [11060] = 2256,
  [11061] = 10829,
  [11062] = 10720,
  [11063] = 10718,
  [11064] = 10894,
  [11065] = 11012,
  [11066] = 11014,
  [11067] = 11015,
  [11068] = 10829,
  [11069] = 10847,
  [11070] = 10866,
  [11071] = 2276,
  [11072] = 2263,
  [11073] = 2281,
  [11074] = 10727,
  [11075] = 2299,
  [11076] = 10993,
  [11077] = 2282,
  [11078] = 2284,
  [11079] = 10736,
  [11080] = 10995,
  [11081] = 10894,
  [11082] = 10727,
  [11083] = 10748,
  [11084] = 10736,
  [11085] = 2285,
  [11086] = 2299,
  [11087] = 10766,
  [11088] = 10736,
  [11089] = 10788,
  [11090] = 10809,
  [11091] = 10766,
  [11092] = 10841,
  [11093] = 10846,
  [11094] = 11012,
  [11095] = 10875,
  [11096] = 2287,
  [11097] = 11014,
  [11098] = 11015,
  [11099] = 10720,
  [11100] = 10866,
  [11101] = 2263,
  [11102] = 10993,
  [11103] = 10788,
  [11104] = 2240,
  [11105] = 2288,
  [11106] = 10719,
  [11107] = 2289,
  [11108] = 2290,
  [11109] = 10727,
  [11110] = 10829,
  [11111] = 10720,
  [11112] = 10718,
  [11113] = 10726,
  [11114] = 11012,
  [11115] = 10809,
  [11116] = 11014,
  [11117] = 11015,
  [11118] = 10866,
  [11119] = 2240,
  [11120] = 10993,
  [11121] = 10727,
  [11122] = 10736,
  [11123] = 10766,
  [11124] = 10841,
  [11125] = 11125,
  [11126] = 10846,
  [11127] = 10736,
  [11128] = 10875,
  [11129] = 2286,
  [11130] = 11012,
  [11131] = 10894,
  [11132] = 11014,
  [11133] = 11015,
  [11134] = 2316,
  [11135] = 10866,
  [11136] = 10718,
  [11137] = 10993,
  [11138] = 10748,
  [11139] = 2278,
  [11140] = 2276,
  [11141] = 10719,
  [11142] = 2281,
  [11143] = 11012,
  [11144] = 2282,
  [11145] = 11014,
  [11146] = 11015,
  [11147] = 10866,
  [11148] = 10748,
  [11149] = 2284,
  [11150] = 10829,
  [11151] = 2285,
  [11152] = 10720,
  [11153] = 10993,
  [11154] = 10718,
  [11155] = 10726,
  [11156] = 6031,
  [11157] = 11012,
  [11158] = 11014,
  [11159] = 11015,
  [11160] = 10866,
  [11161] = 10985,
  [11162] = 11162,
  [11163] = 10993,
  [11164] = 10736,
  [11165] = 10719,
  [11166] = 2287,
  [11167] = 11167,
  [11168] = 11012,
  [11169] = 2288,
  [11170] = 10720,
  [11171] = 2289,
  [11172] = 10718,
  [11173] = 11173,
  [11174] = 10726,
  [11175] = 2290,
  [11176] = 10866,
  [11177] = 10993,
  [11178] = 10995,
  [11179] = 10766,
  [11180] = 2277,
  [11181] = 10719,
  [11182] = 10866,
  [11183] = 2264,
  [11184] = 2313,
  [11185] = 10829,
  [11186] = 10720,
  [11187] = 10718,
  [11188] = 10726,
  [11189] = 10866,
  [11190] = 10775,
  [11191] = 6031,
  [11192] = 2265,
  [11193] = 10726,
  [11194] = 10866,
  [11195] = 10829,
  [11196] = 10858,
  [11197] = 10866,
  [11198] = 10720,
  [11199] = 10718,
  [11200] = 10726,
  [11201] = 2255,
  [11202] = 10866,
  [11203] = 10736,
  [11204] = 10847,
  [11205] = 10852,
  [11206] = 10766,
  [11207] = 10866,
  [11208] = 10788,
  [11209] = 10997,
  [11210] = 10719,
  [11211] = 11211,
  [11212] = 10866,
  [11213] = 10809,
  [11214] = 10829,
  [11215] = 10720,
  [11216] = 10718,
  [11217] = 10866,
  [11218] = 10726,
  [11219] = 11219,
  [11220] = 2272,
  [11221] = 2221,
  [11222] = 2256,
  [11223] = 9782,
  [11224] = 10894,
  [11225] = 10866,
  [11226] = 10995,
  [11227] = 2313,
  [11228] = 2221,
  [11229] = 2236,
  [11230] = 2272,
  [11231] = 2234,
  [11232] = 10841,
  [11233] = 2241,
  [11234] = 10846,
  [11235] = 10866,
  [11236] = 2256,
  [11237] = 2312,
  [11238] = 10727,
  [11239] = 2301,
  [11240] = 11240,
  [11241] = 10736,
  [11242] = 10866,
  [11243] = 2283,
  [11244] = 10847,
  [11245] = 10875,
  [11246] = 10719,
  [11247] = 10788,
  [11248] = 10829,
  [11249] = 10866,
  [11250] = 10720,
  [11251] = 10718,
  [11252] = 10726,
  [11253] = 10748,
  [11254] = 10766,
  [11255] = 10866,
  [11256] = 10809,
  [11257] = 2292,
  [11258] = 10866,
  [11259] = 10788,
  [11260] = 10809,
  [11261] = 10841,
  [11262] = 10846,
  [11263] = 10841,
  [11264] = 10846,
  [11265] = 6450,
  [11266] = 10866,
  [11267] = 10727,
  [11268] = 11005,
  [11269] = 10866,
  [11270] = 10719,
  [11271] = 10866,
  [11272] = 6451,
  [11273] = 2299,
  [11274] = 10866,
  [11275] = 10866,
  [11276] = 10875,
  [11277] = 10829,
  [11278] = 10866,
  [11279] = 10720,
  [11280] = 10866,
  [11281] = 9782,
  [11282] = 10866,
  [11283] = 10866,
  [11284] = 10718,
  [11285] = 10866,
  [11286] = 10726,
  [11287] = 10866,
  [11288] = 10866,
  [11289] = 2316,
  [11290] = 10866,
  [11291] = 10866,
  [11292] = 10866,
  [11293] = 10866,
  [11294] = 10866,
  [11295] = 10866,
  [11296] = 5093,
  [11297] = 10866,
  [11298] = 10866,
  [11299] = 10866,
  [11300] = 10866,
  [11301] = 10866,
  [11302] = 10866,
  [11303] = 10875,
  [11304] = 10866,
  [11305] = 10985,
  [11306] = 10866,
  [11307] = 10866,
  [11308] = 10866,
  [11309] = 10866,
  [11310] = 10866,
  [11311] = 10866,
  [11312] = 10866,
  [11313] = 10866,
  [11314] = 2278,
  [11315] = 10866,
  [11316] = 2312,
  [11317] = 10866,
  [11318] = 10866,
  [11319] = 10866,
  [11320] = 2276,
  [11321] = 10866,
  [11322] = 2301,
  [11323] = 10866,
  [11324] = 10866,
  [11325] = 10866,
  [11326] = 10866,
  [11327] = 10719,
  [11328] = 10866,
  [11329] = 10866,
  [11330] = 10866,
  [11331] = 10719,
  [11332] = 10866,
  [11333] = 2283,
  [11334] = 10866,
  [11335] = 10866,
  [11336] = 10866,
  [11337] = 10719,
  [11338] = 10866,
  [11339] = 10866,
  [11340] = 10997,
  [11341] = 10866,
  [11342] = 2292,
  [11343] = 10866,
  [11344] = 10858,
  [11345] = 10866,
  [11346] = 10829,
  [11347] = 10866,
  [11348] = 10720,
  [11349] = 10866,
  [11350] = 10866,
  [11351] = 10718,
  [11352] = 10866,
  [11353] = 10866,
  [11354] = 10726,
  [11355] = 10866,
  [11356] = 10866,
  [11357] = 10866,
  [11358] = 2281,
  [11359] = 10866,
  [11360] = 10866,
  [11361] = 10985,
  [11362] = 10866,
  [11363] = 2282,
  [11364] = 10866,
  [11365] = 10866,
  [11366] = 10985,
  [11367] = 10866,
  [11368] = 10866,
  [11369] = 10993,
  [11370] = 10866,
  [11371] = 10866,
  [11372] = 10719,
  [11373] = 10866,
  [11374] = 11374,
  [11375] = 10866,
  [11376] = 10866,
  [11377] = 2284,
  [11378] = 10866,
  [11379] = 11379,
  [11380] = 10775,
  [11381] = 2285,
  [11382] = 10847,
  [11383] = 10829,
  [11384] = 10858,
  [11385] = 10720,
  [11386] = 10718,
  [11387] = 10829,
  [11388] = 2287,
  [11389] = 10847,
  [11390] = 11390,
  [11391] = 10720,
  [11392] = 10852,
  [11393] = 10718,
  [11394] = 10719,
  [11395] = 2288,
  [11396] = 10893,
  [11397] = 10793,
  [11398] = 10847,
  [11399] = 10829,
  [11400] = 10720,
  [11401] = 10997,
  [11402] = 10718,
  [11403] = 11390,
  [11404] = 10726,
  [11405] = 2289,
  [11406] = 10852,
  [11407] = 10893,
  [11408] = 10793,
  [11409] = 2290,
  [11410] = 10775,
  [11411] = 2272,
  [11412] = 11390,
  [11413] = 10841,
  [11414] = 5669,
  [11415] = 10894,
  [11416] = 10893,
  [11417] = 10793,
  [11418] = 11390,
  [11419] = 10995,
  [11420] = 10829,
  [11421] = 11390,
  [11422] = 10727,
  [11423] = 10858,
  [11424] = 10719,
  [11425] = 10893,
  [11426] = 10736,
  [11427] = 10793,
  [11428] = 10720,
  [11429] = 11390,
  [11430] = 10829,
  [11431] = 10748,
  [11432] = 10720,
  [11433] = 10766,
  [11434] = 10893,
  [11435] = 10718,
  [11436] = 10793,
  [11437] = 10726,
  [11438] = 10718,
  [11439] = 10766,
  [11440] = 11390,
  [11441] = 2286,
  [11442] = 10852,
  [11443] = 10788,
  [11444] = 10893,
  [11445] = 10793,
  [11446] = 10894,
  [11447] = 10809,
  [11448] = 11390,
  [11449] = 2256,
  [11450] = 10893,
  [11451] = 10793,
  [11452] = 10995,
  [11453] = 10841,
  [11454] = 10719,
  [11455] = 11390,
  [11456] = 10847,
  [11457] = 10727,
  [11458] = 10852,
  [11459] = 10846,
  [11460] = 11014,
  [11461] = 10893,
  [11462] = 10875,
  [11463] = 10793,
  [11464] = 10829,
  [11465] = 10720,
  [11466] = 10718,
  [11467] = 11390,
  [11468] = 10726,
  [11469] = 10736,
  [11470] = 10893,
  [11471] = 10793,
  [11472] = 10748,
  [11473] = 10719,
  [11474] = 10766,
  [11475] = 11390,
  [11476] = 10788,
  [11477] = 10893,
  [11478] = 10793,
  [11479] = 10719,
  [11480] = 10809,
  [11481] = 11390,
  [11482] = 10829,
  [11483] = 10720,
  [11484] = 10718,
  [11485] = 10726,
  [11486] = 10893,
  [11487] = 10793,
  [11488] = 10841,
  [11489] = 10846,
  [11490] = 11390,
  [11491] = 10875,
  [11492] = 10893,
  [11493] = 10793,
  [11494] = 10719,
  [11495] = 10997,
  [11496] = 11390,
  [11497] = 10829,
  [11498] = 10720,
  [11499] = 10718,
  [11500] = 10793,
  [11501] = 10726,
  [11502] = 2292,
  [11503] = 11390,
  [11504] = 10793,
  [11505] = 10719,
  [11506] = 2299,
  [11507] = 11390,
  [11508] = 5671,
  [11509] = 10719,
  [11510] = 11390,
  [11511] = 10846,
  [11512] = 10829,
  [11513] = 10829,
  [11514] = 10720,
  [11515] = 11390,
  [11516] = 10718,
  [11517] = 10726,
  [11518] = 11015,
  [11519] = 10720,
  [11520] = 10718,
  [11521] = 11521,
  [11522] = 5672,
  [11523] = 11390,
  [11524] = 10726,
  [11525] = 10875,
  [11526] = 5673,
  [11527] = 5668,
  [11528] = 11390,
  [11529] = 10775,
  [11530] = 5669,
  [11531] = 10847,
  [11532] = 11390,
  [11533] = 11533,
  [11534] = 2255,
  [11535] = 10894,
  [11536] = 2272,
  [11537] = 11390,
  [11538] = 10719,
  [11539] = 10829,
  [11540] = 5671,
  [11541] = 10720,
  [11542] = 10718,
  [11543] = 10726,
  [11544] = 11390,
  [11545] = 2236,
  [11546] = 5672,
  [11547] = 2234,
  [11548] = 10719,
  [11549] = 11390,
  [11550] = 10720,
  [11551] = 10718,
  [11552] = 10985,
  [11553] = 10719,
  [11554] = 11390,
  [11555] = 10726,
  [11556] = 5673,
  [11557] = 10720,
  [11558] = 10775,
  [11559] = 10829,
  [11560] = 11390,
  [11561] = 10719,
  [11562] = 2241,
  [11563] = 10720,
  [11564] = 10718,
  [11565] = 10718,
  [11566] = 10858,
  [11567] = 10736,
  [11568] = 10720,
  [11569] = 11390,
  [11570] = 10718,
  [11571] = 10726,
  [11572] = 11390,
  [11573] = 10726,
  [11574] = 11390,
  [11575] = 11390,
  [11576] = 11390,
  [11577] = 10847,
  [11578] = 11390,
  [11579] = 10852,
  [11580] = 11390,
  [11581] = 11390,
  [11582] = 11390,
  [11583] = 11390,
  [11584] = 11390,
  [11585] = 11390,
  [11586] = 11167,
  [11587] = 10775,
  [11588] = 11533,
  [11589] = 10858,
  [11590] = 10997,
  [11591] = 10719,
  [11592] = 10720,
  [11593] = 11173,
  [11594] = 10718,
  [11595] = 10726,
  [11596] = 2256,
  [11597] = 11167,
  [11598] = 11533,
  [11599] = 10719,
  [11600] = 11173,
  [11601] = 11167,
  [11602] = 11533,
  [11603] = 10894,
  [11604] = 11173,
  [11605] = 10720,
  [11606] = 11167,
  [11607] = 11533,
  [11608] = 10718,
  [11609] = 11173,
  [11610] = 11167,
  [11611] = 11533,
  [11612] = 11173,
  [11613] = 11167,
  [11614] = 11533,
  [11615] = 11173,
  [11616] = 11167,
  [11617] = 11533,
  [11618] = 11618,
  [11619] = 11173,
  [11620] = 10719,
  [11621] = 11173,
  [11622] = 10997,
  [11623] = 11173,
  [11624] = 10720,
  [11625] = 10718,
  [11626] = 10719,
  [11627] = 10720,
  [11628] = 10718,
  [11629] = 10748,
  [11630] = 10726,
  [11631] = 2255,
  [11632] = 10727,
  [11633] = 10995,
  [11634] = 10985,
  [11635] = 10719,
  [11636] = 10720,
  [11637] = 10894,
  [11638] = 10718,
  [11639] = 10995,
  [11640] = 2292,
  [11641] = 10736,
  [11642] = 10727,
  [11643] = 10719,
  [11644] = 10720,
  [11645] = 10736,
  [11646] = 10718,
  [11647] = 10748,
  [11648] = 10766,
  [11649] = 10719,
  [11650] = 10766,
  [11651] = 10720,
  [11652] = 10719,
  [11653] = 10718,
  [11654] = 10726,
  [11655] = 10985,
  [11656] = 10788,
  [11657] = 10719,
  [11658] = 10809,
  [11659] = 10720,
  [11660] = 10718,
  [11661] = 10726,
  [11662] = 10788,
  [11663] = 10985,
  [11664] = 10719,
  [11665] = 10841,
  [11666] = 10846,
  [11667] = 10720,
  [11668] = 10718,
  [11669] = 10875,
  [11670] = 10809,
  [11671] = 10719,
  [11672] = 10720,
  [11673] = 10718,
  [11674] = 10726,
  [11675] = 10719,
  [11676] = 10720,
  [11677] = 10718,
  [11678] = 10726,
  [11679] = 10841,
  [11680] = 10846,
  [11681] = 10719,
  [11682] = 10720,
  [11683] = 10718,
  [11684] = 10726,
  [11685] = 10809,
  [11686] = 10719,
  [11687] = 10720,
  [11688] = 10718,
  [11689] = 10726,
  [11690] = 10875,
  [11691] = 10719,
  [11692] = 10720,
  [11693] = 10720,
  [11694] = 2276,
  [11695] = 2530,
  [11696] = 2558,
  [11697] = 2272,
  [11698] = 2537,
  [11699] = 2313,
  [11700] = 2256,
  [11701] = 2258,
  [11702] = 8787,
  [11703] = 2539,
  [11704] = 2285,
  [11705] = 2542,
  [11706] = 2292,
  [11707] = 9726,
  [11708] = 2583,
  [11709] = 9738,
  [11710] = 2583,
  [11711] = 2281,
  [11712] = 2582,
  [11713] = 2312,
  [11714] = 2589,
  [11715] = 2265,
  [11716] = 2265,
  [11717] = 11717,
  [11718] = 2264,
  [11719] = 2575,
  [11720] = 2301,
  [11721] = 9743,
  [11722] = 2313,
  [11723] = 2283,
  [11724] = 5668,
  [11725] = 2272,
  [11726] = 2503,
  [11727] = 2574,
  [11728] = 2573,
  [11729] = 2299,
  [11730] = 2256,
  [11731] = 2286,
  [11732] = 2504,
  [11733] = 5669,
  [11734] = 2265,
  [11735] = 2588,
  [11736] = 2312,
  [11737] = 2589,
  [11738] = 2301,
  [11739] = 6450,
  [11740] = 2283,
  [11741] = 2563,
  [11742] = 2582,
  [11743] = 5671,
  [11744] = 5672,
  [11745] = 2286,
  [11746] = 9743,
  [11747] = 2287,
  [11748] = 5673,
  [11749] = 2547,
  [11750] = 2221,
  [11751] = 2277,
  [11752] = 2550,
  [11753] = 2255,
  [11754] = 2292,
  [11755] = 2312,
  [11756] = 2277,
  [11757] = 2259,
  [11758] = 11717,
  [11759] = 2313,
  [11760] = 2255,
  [11761] = 2299,
  [11762] = 6451,
  [11763] = 2288,
  [11764] = 6450,
  [11765] = 6451,
  [11766] = 8781,
  [11767] = 2263,
  [11768] = 2551,
  [11769] = 2286,
  [11770] = 2277,
  [11771] = 2282,
  [11772] = 2289,
  [11773] = 2286,
  [11774] = 2240,
  [11775] = 2290,
  [11776] = 2545,
  [11777] = 2316,
  [11778] = 2588,
  [11779] = 2264,
  [11780] = 2278,
  [11781] = 2286,
  [11782] = 9843,
  [11783] = 2316,
  [11784] = 9726,
  [11785] = 2281,
  [11786] = 9738,
  [11787] = 2282,
  [11788] = 2555,
  [11789] = 6031,
  [11790] = 6450,
  [11791] = 2485,
  [11792] = 2284,
  [11793] = 6451,
  [11794] = 2285,
  [11795] = 2540,
  [11796] = 2287,
  [11797] = 2265,
  [11798] = 2313,
  [11799] = 2542,
  [11800] = 2288,
  [11801] = 2289,
  [11802] = 2290,
  [11803] = 2537,
  [11804] = 2539,
  [11805] = 2572,
  [11806] = 2573,
  [11807] = 2574,
  [11808] = 2264,
  [11809] = 2485,
  [11810] = 2544,
  [11811] = 2525,
  [11812] = 2575,
  [11813] = 2545,
  [11814] = 2560,
  [11815] = 2561,
  [11816] = 2562,
  [11817] = 2535,
  [11818] = 2572,
  [11819] = 2301,
  [11820] = 2547,
  [11821] = 2550,
  [11822] = 2551,
  [11823] = 2552,
  [11824] = 2555,
  [11825] = 2556,
  [11826] = 2558,
  [11827] = 2560,
  [11828] = 2561,
  [11829] = 2562,
  [11830] = 2276,
  [11831] = 2563,
  [11832] = 2503,
  [11833] = 2504,
  [11834] = 2525,
  [11835] = 2530,
  [11836] = 2535,
  [11837] = 2536,
  [11838] = 2284,
  [11839] = 6031,
  [11840] = 2313,
  [11841] = 2283,
  [11842] = 8786,
  [11843] = 11843,
  [11844] = 2236,
  [11845] = 2540,
  [11846] = 2544,
  [11847] = 2277,
  [11848] = 2278,
  [11849] = 9843,
  [11850] = 2234,
  [11851] = 2536,
  [11852] = 2286,
  [11853] = 2313,
  [11854] = 2241,
  [11855] = 2264,
  [11856] = 2556,
  [11857] = 2552,
  [11858] = 11858,
  [11859] = 11859,
  [11860] = 11860,
  [11861] = 11861,
  [11862] = 11862,
  [11863] = 11863,
  [11864] = 6539,
  [11865] = 2264,
  [11866] = 11858,
  [11867] = 11860,
  [11868] = 11861,
  [11869] = 11859,
  [11870] = 11859,
  [11871] = 2313,
  [11872] = 2312,
  [11873] = 11858,
  [11874] = 2286,
  [11875] = 11863,
  [11876] = 2290,
  [11877] = 6539,
  [11878] = 2265,
  [11879] = 6451,
  [11880] = 2316,
  [11881] = 2299,
  [11882] = 11862,
  [11883] = 11863,
  [11884] = 2301,
  [11885] = 2287,
  [11886] = 2256,
  [11887] = 9745,
  [11888] = 11860,
  [11889] = 11861,
  [11890] = 11859,
  [11891] = 2277,
  [11892] = 11892,
  [11893] = 2283,
  [11894] = 11858,
  [11895] = 9868,
  [11896] = 2287,
  [11897] = 9741,
  [11898] = 2312,
  [11899] = 2316,
  [11900] = 9742,
  [11901] = 2278,
  [11902] = 2301,
  [11903] = 2225,
  [11904] = 2283,
  [11905] = 9694,
  [11906] = 9744,
  [11907] = 2285,
  [11908] = 2278,
  [11909] = 9689,
  [11910] = 2287,
  [11911] = 11862,
  [11912] = 11863,
  [11913] = 11858,
  [11914] = 9839,
  [11915] = 9863,
  [11916] = 11916,
  [11917] = 2316,
  [11918] = 11860,
  [11919] = 11862,
  [11920] = 11861,
  [11921] = 11863,
  [11922] = 2278,
  [11923] = 2288,
  [11924] = 2265,
  [11925] = 11860,
  [11926] = 11861,
  [11927] = 11859,
  [11928] = 2276,
  [11929] = 11859,
  [11930] = 2288,
  [11931] = 11858,
  [11932] = 2272,
  [11933] = 11933,
  [11934] = 2276,
  [11935] = 2281,
  [11936] = 2288,
  [11937] = 2282,
  [11938] = 9786,
  [11939] = 2281,
  [11940] = 2276,
  [11941] = 9751,
  [11942] = 2284,
  [11943] = 11862,
  [11944] = 6450,
  [11945] = 6451,
  [11946] = 2281,
  [11947] = 2282,
  [11948] = 2289,
  [11949] = 9741,
  [11950] = 4827,
  [11951] = 2289,
  [11952] = 9868,
  [11953] = 9742,
  [11954] = 2316,
  [11955] = 2312,
  [11956] = 2282,
  [11957] = 11892,
  [11958] = 2301,
  [11959] = 2312,
  [11960] = 2285,
  [11961] = 2301,
  [11962] = 2283,
  [11963] = 2287,
  [11964] = 2276,
  [11965] = 11862,
  [11966] = 11863,
  [11967] = 2255,
  [11968] = 2281,
  [11969] = 11916,
  [11970] = 9744,
  [11971] = 9786,
  [11972] = 2290,
  [11973] = 2288,
  [11974] = 2282,
  [11975] = 11862,
  [11976] = 11863,
  [11977] = 11862,
  [11978] = 11863,
  [11979] = 2283,
  [11980] = 2264,
  [11981] = 2289,
  [11982] = 11860,
  [11983] = 11861,
  [11984] = 11860,
  [11985] = 2289,
  [11986] = 11859,
  [11987] = 2284,
  [11988] = 11861,
  [11989] = 11858,
  [11990] = 2284,
  [11991] = 2284,
  [11992] = 11859,
  [11993] = 2290,
  [11994] = 9745,
  [11995] = 6031,
  [11996] = 2225,
  [11997] = 11858,
  [11998] = 11860,
  [11999] = 11861,
  [12000] = 9839,
  [12001] = 9863,
  [12002] = 12002,
  [12003] = 2313,
  [12004] = 6450,
  [12005] = 2285,
  [12006] = 9848,
  [12007] = 9853,
  [12008] = 2277,
  [12009] = 9848,
  [12010] = 2285,
  [12011] = 9853,
  [12012] = 9694,
  [12013] = 2286,
  [12014] = 2278,
  [12015] = 2290,
  [12016] = 11933,
  [12017] = 9689,
  [12018] = 2292,
  [12019] = 9751,
  [12020] = 2287,
  [12021] = 6557,
  [12022] = 5093,
  [12023] = 2259,
  [12024] = 6559,
  [12025] = 2225,
  [12026] = 6560,
  [12027] = 2287,
  [12028] = 2285,
  [12029] = 6559,
  [12030] = 6599,
  [12031] = 2282,
  [12032] = 6557,
  [12033] = 2276,
  [12034] = 6450,
  [12035] = 6451,
  [12036] = 2289,
  [12037] = 2281,
  [12038] = 2285,
  [12039] = 2286,
  [12040] = 2265,
  [12041] = 2276,
  [12042] = 2288,
  [12043] = 2312,
  [12044] = 6560,
  [12045] = 2278,
  [12046] = 6835,
  [12047] = 2289,
  [12048] = 6819,
  [12049] = 2264,
  [12050] = 2241,
  [12051] = 2316,
  [12052] = 2301,
  [12053] = 2290,
  [12054] = 2316,
  [12055] = 6600,
  [12056] = 2313,
  [12057] = 2240,
  [12058] = 2283,
  [12059] = 2288,
  [12060] = 2258,
  [12061] = 2278,
  [12062] = 2284,
  [12063] = 2236,
  [12064] = 2234,
  [12065] = 6539,
  [12066] = 6810,
  [12067] = 2284,
  [12068] = 2281,
  [12069] = 2283,
  [12070] = 2312,
  [12071] = 2277,
  [12072] = 6599,
  [12073] = 6600,
  [12074] = 2282,
  [12075] = 2290,
  [12076] = 2301,
  [12077] = 2263,
  [12078] = 6559,
  [12079] = 6631,
  [12080] = 6819,
  [12081] = 2467,
  [12082] = 2278,
  [12083] = 2472,
  [12084] = 6658,
  [12085] = 2276,
  [12086] = 2258,
  [12087] = 2281,
  [12088] = 2282,
  [12089] = 2473,
  [12090] = 2263,
  [12091] = 2284,
  [12092] = 5672,
  [12093] = 2240,
  [12094] = 6919,
  [12095] = 6835,
  [12096] = 2312,
  [12097] = 2299,
  [12098] = 5668,
  [12099] = 6658,
  [12100] = 2259,
  [12101] = 2285,
  [12102] = 5673,
  [12103] = 2272,
  [12104] = 2301,
  [12105] = 2283,
  [12106] = 2236,
  [12107] = 6858,
  [12108] = 2263,
  [12109] = 2443,
  [12110] = 6699,
  [12111] = 6971,
  [12112] = 6652,
  [12113] = 6625,
  [12114] = 2256,
  [12115] = 6807,
  [12116] = 6599,
  [12117] = 6600,
  [12118] = 6857,
  [12119] = 6921,
  [12120] = 6810,
  [12121] = 6886,
  [12122] = 6652,
  [12123] = 2255,
  [12124] = 6539,
  [12125] = 2259,
  [12126] = 2236,
  [12127] = 6663,
  [12128] = 5669,
  [12129] = 2240,
  [12130] = 5093,
  [12131] = 2234,
  [12132] = 2287,
  [12133] = 6859,
  [12134] = 2258,
  [12135] = 2469,
  [12136] = 6938,
  [12137] = 6699,
  [12138] = 6557,
  [12139] = 6856,
  [12140] = 5671,
  [12141] = 2241,
  [12142] = 6872,
  [12143] = 2234,
  [12144] = 6854,
  [12145] = 2288,
  [12146] = 2367,
  [12147] = 2241,
  [12148] = 6631,
  [12149] = 6860,
  [12150] = 6663,
  [12151] = 6853,
  [12152] = 2292,
  [12153] = 2316,
  [12154] = 2289,
  [12155] = 6848,
  [12156] = 6855,
  [12157] = 2290,
  [12158] = 6625,
  [12159] = 6560,
  [12160] = 6955,
  [12161] = 2221,
  [12162] = 6896,
  [12163] = 6560,
  [12164] = 6897,
  [12165] = 2272,
  [12166] = 6853,
  [12167] = 2312,
  [12168] = 6795,
  [12169] = 6938,
  [12170] = 6849,
  [12171] = 6864,
  [12172] = 6631,
  [12173] = 2255,
  [12174] = 2299,
  [12175] = 2236,
  [12176] = 6921,
  [12177] = 2234,
  [12178] = 6898,
  [12179] = 6850,
  [12180] = 6902,
  [12181] = 6925,
  [12182] = 6938,
  [12183] = 5668,
  [12184] = 6927,
  [12185] = 6922,
  [12186] = 2258,
  [12187] = 2255,
  [12188] = 2241,
  [12189] = 6944,
  [12190] = 6957,
  [12191] = 6958,
  [12192] = 8944,
  [12193] = 6936,
  [12194] = 6625,
  [12195] = 2313,
  [12196] = 6935,
  [12197] = 2283,
  [12198] = 6859,
  [12199] = 2286,
  [12200] = 6652,
  [12201] = 6956,
  [12202] = 6948,
  [12203] = 6931,
  [12204] = 6699,
  [12205] = 2272,
  [12206] = 6907,
  [12207] = 6854,
  [12208] = 2301,
  [12209] = 2256,
  [12210] = 2367,
  [12211] = 6919,
  [12212] = 2277,
  [12213] = 6872,
  [12214] = 12214,
  [12215] = 6967,
  [12216] = 2292,
  [12217] = 6784,
  [12218] = 6908,
  [12219] = 6926,
  [12220] = 6795,
  [12221] = 6848,
  [12222] = 6557,
  [12223] = 6807,
  [12224] = 2264,
  [12225] = 6932,
  [12226] = 6971,
  [12227] = 8944,
  [12228] = 6962,
  [12229] = 12229,
  [12230] = 5671,
  [12231] = 6949,
  [12232] = 6658,
  [12233] = 5672,
  [12234] = 2292,
  [12235] = 6871,
  [12236] = 6945,
  [12237] = 6928,
  [12238] = 12238,
  [12239] = 12239,
  [12240] = 6663,
  [12241] = 6903,
  [12242] = 6909,
  [12243] = 5669,
  [12244] = 6852,
  [12245] = 6858,
  [12246] = 2299,
  [12247] = 6910,
  [12248] = 2367,
  [12249] = 6953,
  [12250] = 6918,
  [12251] = 6941,
  [12252] = 6942,
  [12253] = 6784,
  [12254] = 6851,
  [12255] = 2256,
  [12256] = 2221,
  [12257] = 6911,
  [12258] = 12258,
  [12259] = 6943,
  [12260] = 2240,
  [12261] = 6856,
  [12262] = 6895,
  [12263] = 6971,
  [12264] = 6857,
  [12265] = 6783,
  [12266] = 6890,
  [12267] = 6599,
  [12268] = 6921,
  [12269] = 2263,
  [12270] = 6600,
  [12271] = 6860,
  [12272] = 6954,
  [12273] = 6783,
  [12274] = 5673,
  [12275] = 6891,
  [12276] = 6892,
  [12277] = 6855,
  [12278] = 6893,
  [12279] = 6937,
  [12280] = 2265,
  [12281] = 6886,
  [12282] = 8149,
  [12283] = 2259,
  [12284] = 6031,
  [12285] = 6559,
  [12286] = 6961,
  [12287] = 6919,
  [12288] = 6872,
  [12289] = 2240,
  [12290] = 2259,
  [12291] = 6911,
  [12292] = 2241,
  [12293] = 6850,
  [12294] = 6819,
  [12295] = 6859,
  [12296] = 2281,
  [12297] = 8949,
  [12298] = 6860,
  [12299] = 2240,
  [12300] = 6848,
  [12301] = 6927,
  [12302] = 2282,
  [12303] = 2234,
  [12304] = 6856,
  [12305] = 6957,
  [12306] = 6857,
  [12307] = 8954,
  [12308] = 12308,
  [12309] = 2287,
  [12310] = 8944,
  [12311] = 6958,
  [12312] = 6849,
  [12313] = 2286,
  [12314] = 2236,
  [12315] = 8740,
  [12316] = 2278,
  [12317] = 2289,
  [12318] = 6956,
  [12319] = 6854,
  [12320] = 2292,
  [12321] = 6854,
  [12322] = 8944,
  [12323] = 6955,
  [12324] = 12324,
  [12325] = 6872,
  [12326] = 2259,
  [12327] = 6921,
  [12328] = 2241,
  [12329] = 2265,
  [12330] = 2259,
  [12331] = 6851,
  [12332] = 6941,
  [12333] = 6631,
  [12334] = 6852,
  [12335] = 2234,
  [12336] = 8954,
  [12337] = 2240,
  [12338] = 6849,
  [12339] = 2286,
  [12340] = 6918,
  [12341] = 2285,
  [12342] = 6850,
  [12343] = 6909,
  [12344] = 6922,
  [12345] = 6910,
  [12346] = 6853,
  [12347] = 6855,
  [12348] = 12348,
  [12349] = 12239,
  [12350] = 2236,
  [12351] = 6864,
  [12352] = 2234,
  [12353] = 2312,
  [12354] = 12308,
  [12355] = 6851,
  [12356] = 2264,
  [12357] = 6891,
  [12358] = 6893,
  [12359] = 2277,
  [12360] = 6807,
  [12361] = 6851,
  [12362] = 6954,
  [12363] = 2316,
  [12364] = 6925,
  [12365] = 6907,
  [12366] = 12308,
  [12367] = 6795,
  [12368] = 6931,
  [12369] = 6942,
  [12370] = 6892,
  [12371] = 12214,
  [12372] = 6938,
  [12373] = 2277,
  [12374] = 6858,
  [12375] = 2256,
  [12376] = 6699,
  [12377] = 2259,
  [12378] = 2284,
  [12379] = 6919,
  [12380] = 6932,
  [12381] = 2255,
  [12382] = 2263,
  [12383] = 2236,
  [12384] = 6967,
  [12385] = 6852,
  [12386] = 6031,
  [12387] = 12308,
  [12388] = 6890,
  [12389] = 8733,
  [12390] = 12308,
  [12391] = 2301,
  [12392] = 2241,
  [12393] = 6928,
  [12394] = 12229,
  [12395] = 12238,
  [12396] = 12308,
  [12397] = 6852,
  [12398] = 6807,
  [12399] = 2240,
  [12400] = 6949,
  [12401] = 6895,
  [12402] = 2283,
  [12403] = 6860,
  [12404] = 6625,
  [12405] = 2288,
  [12406] = 6835,
  [12407] = 6652,
  [12408] = 6819,
  [12409] = 6850,
  [12410] = 6871,
  [12411] = 2264,
  [12412] = 6859,
  [12413] = 6858,
  [12414] = 2290,
  [12415] = 6902,
  [12416] = 6849,
  [12417] = 6783,
  [12418] = 6856,
  [12419] = 8949,
  [12420] = 6855,
  [12421] = 6908,
  [12422] = 2258,
  [12423] = 2258,
  [12424] = 12424,
  [12425] = 12258,
  [12426] = 6857,
  [12427] = 2367,
  [12428] = 12308,
  [12429] = 6953,
  [12430] = 6784,
  [12431] = 6903,
  [12432] = 2263,
  [12433] = 2258,
  [12434] = 6898,
  [12435] = 2263,
  [12436] = 6971,
  [12437] = 12308,
  [12438] = 2241,
  [12439] = 6896,
  [12440] = 6658,
  [12441] = 2263,
  [12442] = 6936,
  [12443] = 6935,
  [12444] = 6961,
  [12445] = 12308,
  [12446] = 2265,
  [12447] = 2276,
  [12448] = 2299,
  [12449] = 6810,
  [12450] = 6897,
  [12451] = 6937,
  [12452] = 6943,
  [12453] = 6853,
  [12454] = 6945,
  [12455] = 6835,
  [12456] = 2221,
  [12457] = 2367,
  [12458] = 6944,
  [12459] = 6848,
  [12460] = 2258,
  [12461] = 2367,
  [12462] = 6663,
  [12463] = 6450,
  [12464] = 6962,
  [12465] = 8149,
  [12466] = 6451,
  [12467] = 6948,
  [12468] = 6926,
  [12469] = 2272,
  [12470] = 2236,
  [12471] = 2313,
  [12472] = 6810,
  [12473] = 2234,
  [12474] = 2313,
  [12475] = 6857,
  [12476] = 6911,
  [12477] = 6783,
  [12478] = 6784,
  [12479] = 8954,
  [12480] = 8954,
  [12481] = 2263,
  [12482] = 8949,
  [12483] = 6848,
  [12484] = 6849,
  [12485] = 2240,
  [12486] = 6850,
  [12487] = 6851,
  [12488] = 6852,
  [12489] = 6795,
  [12490] = 6853,
  [12491] = 6854,
  [12492] = 6855,
  [12493] = 8949,
  [12494] = 6856,
  [12495] = 6857,
  [12496] = 6858,
  [12497] = 6859,
  [12498] = 6807,
  [12499] = 6860,
  [12500] = 2236,
  [12501] = 2234,
  [12502] = 2241,
  [12503] = 2272,
  [12504] = 2256,
  [12505] = 6835,
  [12506] = 6810,
  [12507] = 2367,
  [12508] = 6819,
  [12509] = 6891,
  [12510] = 6892,
  [12511] = 6893,
  [12512] = 6918,
  [12513] = 6896,
  [12514] = 6922,
  [12515] = 6897,
  [12516] = 6925,
  [12517] = 6926,
  [12518] = 6927,
  [12519] = 6864,
  [12520] = 6928,
  [12521] = 6902,
  [12522] = 6903,
  [12523] = 6931,
  [12524] = 6932,
  [12525] = 6907,
  [12526] = 6908,
  [12527] = 6935,
  [12528] = 6936,
  [12529] = 6937,
  [12530] = 6909,
  [12531] = 6910,
  [12532] = 2258,
  [12533] = 6941,
  [12534] = 6942,
  [12535] = 6943,
  [12536] = 6944,
  [12537] = 6945,
  [12538] = 6911,
  [12539] = 6948,
  [12540] = 2259,
  [12541] = 6971,
  [12542] = 6949,
  [12543] = 6954,
  [12544] = 6955,
  [12545] = 6956,
  [12546] = 2367,
  [12547] = 6921,
  [12548] = 6957,
  [12549] = 6958,
  [12550] = 6961,
  [12551] = 6962,
  [12552] = 2258,
  [12553] = 2259,
  [12554] = 6938,
  [12555] = 6872,
  [12556] = 2292,
  [12557] = 2299,
  [12558] = 2263,
  [12559] = 2240,
  [12560] = 2292,
  [12561] = 2236,
  [12562] = 2234,
  [12563] = 2241,
  [12564] = 2367,
  [12565] = 2258,
  [12566] = 2255,
  [12567] = 6835,
  [12568] = 2299,
  [12569] = 2316,
  [12570] = 2278,
  [12571] = 2276,
  [12572] = 2281,
  [12573] = 2282,
  [12574] = 2284,
  [12575] = 2285,
  [12576] = 2287,
  [12577] = 2288,
  [12578] = 2289,
  [12579] = 2290,
  [12580] = 2259,
  [12581] = 6919,
  [12582] = 6810,
  [12583] = 6967,
  [12584] = 2313,
  [12585] = 6871,
  [12586] = 2255,
  [12587] = 2263,
  [12588] = 2240,
  [12589] = 2312,
  [12590] = 2301,
  [12591] = 2283,
  [12592] = 6953,
  [12593] = 6918,
  [12594] = 6895,
  [12595] = 6819,
  [12596] = 2236,
  [12597] = 2234,
  [12598] = 2241,
  [12599] = 8986,
  [12600] = 6971,
  [12601] = 6922,
  [12602] = 2258,
  [12603] = 2259,
  [12604] = 6921,
  [12605] = 8954,
  [12606] = 6938,
  [12607] = 6925,
  [12608] = 6926,
  [12609] = 6927,
  [12610] = 6928,
  [12611] = 6872,
  [12612] = 8949,
  [12613] = 2263,
  [12614] = 2240,
  [12615] = 6931,
  [12616] = 6932,
  [12617] = 2278,
  [12618] = 2276,
  [12619] = 6910,
  [12620] = 2282,
  [12621] = 2284,
  [12622] = 6848,
  [12623] = 2263,
  [12624] = 2285,
  [12625] = 6935,
  [12626] = 2287,
  [12627] = 2288,
  [12628] = 2289,
  [12629] = 2290,
  [12630] = 6936,
  [12631] = 6937,
  [12632] = 2292,
  [12633] = 2299,
  [12634] = 2236,
  [12635] = 2234,
  [12636] = 2241,
  [12637] = 6941,
  [12638] = 6942,
  [12639] = 6943,
  [12640] = 6944,
  [12641] = 6945,
  [12642] = 6967,
  [12643] = 6849,
  [12644] = 6948,
  [12645] = 6850,
  [12646] = 6949,
  [12647] = 6954,
  [12648] = 6955,
  [12649] = 6956,
  [12650] = 6957,
  [12651] = 6958,
  [12652] = 6961,
  [12653] = 6962,
  [12654] = 6953,
  [12655] = 2258,
  [12656] = 2240,
  [12657] = 2259,
  [12658] = 2312,
  [12659] = 2301,
  [12660] = 2283,
  [12661] = 8944,
  [12662] = 2277,
  [12663] = 2316,
  [12664] = 2255,
  [12665] = 6886,
  [12666] = 2264,
  [12667] = 6919,
  [12668] = 2236,
  [12669] = 2234,
  [12670] = 2241,
  [12671] = 2265,
  [12672] = 8986,
  [12673] = 6851,
  [12674] = 6852,
  [12675] = 2272,
  [12676] = 6853,
  [12677] = 2256,
  [12678] = 6871,
  [12679] = 2258,
  [12680] = 2259,
  [12681] = 6854,
  [12682] = 6855,
  [12683] = 6895,
  [12684] = 6856,
  [12685] = 6858,
  [12686] = 6859,
  [12687] = 6890,
  [12688] = 6807,
  [12689] = 6860,
  [12690] = 2367,
  [12691] = 6898,
  [12692] = 8740,
  [12693] = 2292,
  [12694] = 2299,
  [12695] = 2263,
  [12696] = 2240,
  [12697] = 2236,
  [12698] = 2234,
  [12699] = 2241,
  [12700] = 2255,
  [12701] = 2263,
  [12702] = 8733,
  [12703] = 2240,
  [12704] = 2236,
  [12705] = 2234,
  [12706] = 2241,
  [12707] = 2272,
  [12708] = 2256,
  [12709] = 6890,
  [12710] = 2286,
  [12711] = 2272,
  [12712] = 2256,
  [12713] = 6898,
  [12714] = 8944,
  [12715] = 6909,
  [12716] = 6891,
  [12717] = 6892,
  [12718] = 6893,
  [12719] = 2225,
  [12720] = 6886,
  [12721] = 6896,
  [12722] = 6897,
  [12723] = 2225,
  [12724] = 6864,
  [12725] = 6902,
  [12726] = 6903,
  [12727] = 2258,
  [12728] = 6450,
  [12729] = 6451,
  [12730] = 6907,
  [12731] = 6908,
  [12732] = 2259,
  [12733] = 2281,
  [12734] = 6872,
  [12735] = 2264,
  [12736] = 2265,
  [12737] = 2299,
  [12738] = 6922,
  [12739] = 6897,
  [12740] = 8954,
  [12741] = 6971,
  [12742] = 6927,
  [12743] = 2367,
  [12744] = 8949,
  [12745] = 6886,
  [12746] = 6941,
  [12747] = 6848,
  [12748] = 6921,
  [12749] = 6849,
  [12750] = 6850,
  [12751] = 6942,
  [12752] = 6943,
  [12753] = 2272,
  [12754] = 2256,
  [12755] = 6944,
  [12756] = 6851,
  [12757] = 6852,
  [12758] = 2292,
  [12759] = 6945,
  [12760] = 6853,
  [12761] = 2299,
  [12762] = 2258,
  [12763] = 6854,
  [12764] = 6855,
  [12765] = 6938,
  [12766] = 2259,
  [12767] = 6856,
  [12768] = 6857,
  [12769] = 6858,
  [12770] = 6859,
  [12771] = 2277,
  [12772] = 6928,
  [12773] = 6807,
  [12774] = 6860,
  [12775] = 6872,
  [12776] = 6948,
  [12777] = 6949,
  [12778] = 6918,
  [12779] = 6967,
  [12780] = 2272,
  [12781] = 2286,
  [12782] = 2263,
  [12783] = 6898,
  [12784] = 6922,
  [12785] = 6931,
  [12786] = 6932,
  [12787] = 2292,
  [12788] = 6954,
  [12789] = 2240,
  [12790] = 6931,
  [12791] = 6955,
  [12792] = 6956,
  [12793] = 2255,
  [12794] = 2286,
  [12795] = 6957,
  [12796] = 6958,
  [12797] = 6932,
  [12798] = 2225,
  [12799] = 2299,
  [12800] = 8986,
  [12801] = 6961,
  [12802] = 2236,
  [12803] = 2234,
  [12804] = 2241,
  [12805] = 12805,
  [12806] = 2264,
  [12807] = 6907,
  [12808] = 6953,
  [12809] = 6962,
  [12810] = 2256,
  [12811] = 6908,
  [12812] = 2258,
  [12813] = 2367,
  [12814] = 2292,
  [12815] = 2299,
  [12816] = 6890,
  [12817] = 2292,
  [12818] = 2265,
  [12819] = 2299,
  [12820] = 2277,
  [12821] = 2259,
  [12822] = 2316,
  [12823] = 2278,
  [12824] = 6971,
  [12825] = 2264,
  [12826] = 2265,
  [12827] = 2286,
  [12828] = 6921,
  [12829] = 2276,
  [12830] = 2281,
  [12831] = 2282,
  [12832] = 2284,
  [12833] = 6936,
  [12834] = 6937,
  [12835] = 2272,
  [12836] = 6938,
  [12837] = 2256,
  [12838] = 2285,
  [12839] = 2287,
  [12840] = 2255,
  [12841] = 6872,
  [12842] = 2277,
  [12843] = 6909,
  [12844] = 6910,
  [12845] = 2255,
  [12846] = 2313,
  [12847] = 2288,
  [12848] = 2255,
  [12849] = 8986,
  [12850] = 2289,
  [12851] = 2290,
  [12852] = 6941,
  [12853] = 6942,
  [12854] = 6943,
  [12855] = 2292,
  [12856] = 6944,
  [12857] = 6945,
  [12858] = 2299,
  [12859] = 2272,
  [12860] = 2263,
  [12861] = 2240,
  [12862] = 12805,
  [12863] = 6911,
  [12864] = 6948,
  [12865] = 6949,
  [12866] = 6903,
  [12867] = 6955,
  [12868] = 6956,
  [12869] = 12805,
  [12870] = 6919,
  [12871] = 6957,
  [12872] = 8954,
  [12873] = 6958,
  [12874] = 6890,
  [12875] = 6961,
  [12876] = 6962,
  [12877] = 6919,
  [12878] = 2255,
  [12879] = 12805,
  [12880] = 8949,
  [12881] = 2255,
  [12882] = 12805,
  [12883] = 2263,
  [12884] = 2286,
  [12885] = 6925,
  [12886] = 6871,
  [12887] = 12805,
  [12888] = 2240,
  [12889] = 2236,
  [12890] = 2234,
  [12891] = 2241,
  [12892] = 12805,
  [12893] = 6935,
  [12894] = 2258,
  [12895] = 12805,
  [12896] = 2259,
  [12897] = 2313,
  [12898] = 2292,
  [12899] = 2312,
  [12900] = 12805,
  [12901] = 2301,
  [12902] = 2283,
  [12903] = 2272,
  [12904] = 2373,
  [12905] = 12805,
  [12906] = 2256,
  [12907] = 2263,
  [12908] = 12805,
  [12909] = 12805,
  [12910] = 8787,
  [12911] = 6967,
  [12912] = 12805,
  [12913] = 6925,
  [12914] = 2277,
  [12915] = 6871,
  [12916] = 2240,
  [12917] = 6953,
  [12918] = 2256,
  [12919] = 6898,
  [12920] = 2236,
  [12921] = 2234,
  [12922] = 2241,
  [12923] = 8786,
  [12924] = 6895,
  [12925] = 2328,
  [12926] = 2313,
  [12927] = 2236,
  [12928] = 6891,
  [12929] = 6892,
  [12930] = 6893,
  [12931] = 2272,
  [12932] = 2256,
  [12933] = 6891,
  [12934] = 2234,
  [12935] = 2241,
  [12936] = 6896,
  [12937] = 6897,
  [12938] = 2272,
  [12939] = 8781,
  [12940] = 2256,
  [12941] = 2264,
  [12942] = 6926,
  [12943] = 2313,
  [12944] = 6927,
  [12945] = 6892,
  [12946] = 6864,
  [12947] = 6971,
  [12948] = 6902,
  [12949] = 6921,
  [12950] = 6903,
  [12951] = 6938,
  [12952] = 6864,
  [12953] = 2265,
  [12954] = 6893,
  [12955] = 6918,
  [12956] = 6907,
  [12957] = 6908,
  [12958] = 6895,
  [12959] = 6935,
  [12960] = 6909,
  [12961] = 6910,
  [12962] = 6919,
  [12963] = 2221,
  [12964] = 2367,
  [12965] = 6911,
  [12966] = 6928,
  [12967] = 2255,
  [12968] = 6936,
  [12969] = 6937,
  [12970] = 2258,
  [12971] = 2259,
  [12972] = 6902,
  [12973] = 8986,
  [12974] = 2292,
  [12975] = 6886,
  [12976] = 6926,
  [12977] = 6896,
  [12978] = 2299,
  [12979] = 6954,
  [12980] = 2290,
  [12981] = 2264,
  [12982] = 2313,
  [12983] = 2284,
  [12984] = 12984,
  [12985] = 2301,
  [12986] = 12986,
  [12987] = 12984,
  [12988] = 12984,
  [12989] = 2316,
  [12990] = 2258,
  [12991] = 12991,
  [12992] = 2299,
  [12993] = 6936,
  [12994] = 6937,
  [12995] = 12986,
  [12996] = 2286,
  [12997] = 2283,
  [12998] = 2287,
  [12999] = 2278,
  [13000] = 2276,
  [13001] = 2281,
  [13002] = 2282,
  [13003] = 12986,
  [13004] = 2284,
  [13005] = 2285,
  [13006] = 12984,
  [13007] = 2287,
  [13008] = 2288,
  [13009] = 2289,
  [13010] = 2290,
  [13011] = 12986,
  [13012] = 2284,
  [13013] = 2240,
  [13014] = 12984,
  [13015] = 6909,
  [13016] = 12986,
  [13017] = 6910,
  [13018] = 12984,
  [13019] = 6896,
  [13020] = 12984,
  [13021] = 2265,
  [13022] = 2288,
  [13023] = 2255,
  [13024] = 12986,
  [13025] = 6953,
  [13026] = 2286,
  [13027] = 2313,
  [13028] = 12984,
  [13029] = 2265,
  [13030] = 2289,
  [13031] = 6941,
  [13032] = 6871,
  [13033] = 6942,
  [13034] = 12986,
  [13035] = 2259,
  [13036] = 6943,
  [13037] = 6944,
  [13038] = 12984,
  [13039] = 2316,
  [13040] = 2277,
  [13041] = 6945,
  [13042] = 12986,
  [13043] = 12986,
  [13044] = 2292,
  [13045] = 2236,
  [13046] = 2234,
  [13047] = 2241,
  [13048] = 6922,
  [13049] = 12984,
  [13050] = 8740,
  [13051] = 2264,
  [13052] = 12986,
  [13053] = 6898,
  [13054] = 2312,
  [13055] = 6911,
  [13056] = 2301,
  [13057] = 2283,
  [13058] = 2286,
  [13059] = 6948,
  [13060] = 12986,
  [13061] = 12984,
  [13062] = 12984,
  [13063] = 2299,
  [13064] = 6925,
  [13065] = 2265,
  [13066] = 2258,
  [13067] = 2277,
  [13068] = 12986,
  [13069] = 12984,
  [13070] = 6949,
  [13071] = 6926,
  [13072] = 12986,
  [13073] = 6927,
  [13074] = 12986,
  [13075] = 6891,
  [13076] = 6892,
  [13077] = 6890,
  [13078] = 12984,
  [13079] = 12984,
  [13080] = 12984,
  [13081] = 6864,
  [13082] = 6954,
  [13083] = 12986,
  [13084] = 6955,
  [13085] = 12984,
  [13086] = 6956,
  [13087] = 6928,
  [13088] = 6893,
  [13089] = 12986,
  [13090] = 6918,
  [13091] = 12984,
  [13092] = 6957,
  [13093] = 2278,
  [13094] = 8787,
  [13095] = 12986,
  [13096] = 12984,
  [13097] = 2369,
  [13098] = 8733,
  [13099] = 2255,
  [13100] = 2277,
  [13101] = 9541,
  [13102] = 6958,
  [13103] = 12986,
  [13104] = 12986,
  [13105] = 12984,
  [13106] = 2277,
  [13107] = 6902,
  [13108] = 6961,
  [13109] = 6962,
  [13110] = 6935,
  [13111] = 12986,
  [13112] = 12984,
  [13113] = 6919,
  [13114] = 2264,
  [13115] = 12986,
  [13116] = 2277,
  [13117] = 2272,
  [13118] = 2292,
  [13119] = 6903,
  [13120] = 12984,
  [13121] = 12986,
  [13122] = 2264,
  [13123] = 2256,
  [13124] = 2299,
  [13125] = 12986,
  [13126] = 12986,
  [13127] = 12984,
  [13128] = 2265,
  [13129] = 12984,
  [13130] = 2277,
  [13131] = 2265,
  [13132] = 12986,
  [13133] = 12984,
  [13134] = 12984,
  [13135] = 12984,
  [13136] = 2278,
  [13137] = 12986,
  [13138] = 2255,
  [13139] = 2313,
  [13140] = 12984,
  [13141] = 12986,
  [13142] = 2272,
  [13143] = 12984,
  [13144] = 2272,
  [13145] = 2256,
  [13146] = 2265,
  [13147] = 12986,
  [13148] = 2256,
  [13149] = 12984,
  [13150] = 12986,
  [13151] = 2255,
  [13152] = 12984,
  [13153] = 12986,
  [13154] = 2278,
  [13155] = 13155,
  [13156] = 12984,
  [13157] = 12986,
  [13158] = 12984,
  [13159] = 8786,
  [13160] = 2313,
  [13161] = 2264,
  [13162] = 12984,
  [13163] = 2286,
  [13164] = 2265,
  [13165] = 6895,
  [13166] = 12986,
  [13167] = 12986,
  [13168] = 6897,
  [13169] = 12984,
  [13170] = 6971,
  [13171] = 2312,
  [13172] = 2285,
  [13173] = 2258,
  [13174] = 2383,
  [13175] = 12986,
  [13176] = 12986,
  [13177] = 12984,
  [13178] = 2301,
  [13179] = 12986,
  [13180] = 2277,
  [13181] = 12986,
  [13182] = 2313,
  [13183] = 2283,
  [13184] = 6921,
  [13185] = 2316,
  [13186] = 13186,
  [13187] = 12984,
  [13188] = 12986,
  [13189] = 12984,
  [13190] = 12984,
  [13191] = 2312,
  [13192] = 2301,
  [13193] = 2276,
  [13194] = 2313,
  [13195] = 2283,
  [13196] = 8781,
  [13197] = 2276,
  [13198] = 2281,
  [13199] = 2282,
  [13200] = 2287,
  [13201] = 6938,
  [13202] = 2284,
  [13203] = 8986,
  [13204] = 2288,
  [13205] = 6967,
  [13206] = 6931,
  [13207] = 6932,
  [13208] = 12986,
  [13209] = 12984,
  [13210] = 2276,
  [13211] = 12986,
  [13212] = 2272,
  [13213] = 2289,
  [13214] = 2290,
  [13215] = 12984,
  [13216] = 12986,
  [13217] = 6872,
  [13218] = 2285,
  [13219] = 2367,
  [13220] = 2287,
  [13221] = 2272,
  [13222] = 2255,
  [13223] = 2281,
  [13224] = 2282,
  [13225] = 2288,
  [13226] = 12984,
  [13227] = 2256,
  [13228] = 12986,
  [13229] = 2256,
  [13230] = 12984,
  [13231] = 2281,
  [13232] = 2259,
  [13233] = 12986,
  [13234] = 12984,
  [13235] = 6907,
  [13236] = 12984,
  [13237] = 2289,
  [13238] = 12986,
  [13239] = 2290,
  [13240] = 12984,
  [13241] = 2313,
  [13242] = 6908,
  [13243] = 12986,
  [13244] = 12986,
  [13245] = 2292,
  [13246] = 12986,
  [13247] = 2299,
  [13248] = 2285,
  [13249] = 12984,
  [13250] = 2286,
  [13251] = 2292,
  [13252] = 2312,
  [13253] = 2353,
  [13254] = 2259,
  [13255] = 12984,
  [13256] = 2286,
  [13257] = 2263,
  [13258] = 2282,
  [13259] = 12986,
  [13260] = 2316,
  [13261] = 2264,
  [13262] = 2286,
  [13263] = 12986,
  [13264] = 12984,
  [13265] = 2264,
  [13266] = 9651,
  [13267] = 2278,
  [13268] = 13268,
  [13269] = 2292,
  [13270] = 13268,
  [13271] = 2312,
  [13272] = 2301,
  [13273] = 2283,
  [13274] = 13268,
  [13275] = 2286,
  [13276] = 2285,
  [13277] = 2284,
  [13278] = 13268,
  [13279] = 9588,
  [13280] = 2312,
  [13281] = 2301,
  [13282] = 2283,
  [13283] = 2276,
  [13284] = 2287,
  [13285] = 2281,
  [13286] = 2282,
  [13287] = 9600,
  [13288] = 2284,
  [13289] = 13268,
  [13290] = 2288,
  [13291] = 2264,
  [13292] = 2316,
  [13293] = 2278,
  [13294] = 2276,
  [13295] = 2281,
  [13296] = 2282,
  [13297] = 2284,
  [13298] = 2285,
  [13299] = 13268,
  [13300] = 2285,
  [13301] = 2287,
  [13302] = 2287,
  [13303] = 2292,
  [13304] = 2288,
  [13305] = 2289,
  [13306] = 2290,
  [13307] = 2292,
  [13308] = 13268,
  [13309] = 9655,
  [13310] = 2312,
  [13311] = 2301,
  [13312] = 2289,
  [13313] = 2290,
  [13314] = 2265,
  [13315] = 9641,
  [13316] = 2288,
  [13317] = 2276,
  [13318] = 2313,
  [13319] = 2313,
  [13320] = 2281,
  [13321] = 2299,
  [13322] = 9657,
  [13323] = 2299,
  [13324] = 2282,
  [13325] = 9658,
  [13326] = 2289,
  [13327] = 13268,
  [13328] = 2312,
  [13329] = 2290,
  [13330] = 2278,
  [13331] = 2312,
  [13332] = 2263,
  [13333] = 2286,
  [13334] = 2277,
  [13335] = 2301,
  [13336] = 2283,
  [13337] = 2264,
  [13338] = 2265,
  [13339] = 2316,
  [13340] = 2312,
  [13341] = 2278,
  [13342] = 2301,
  [13343] = 2316,
  [13344] = 2276,
  [13345] = 2283,
  [13346] = 2316,
  [13347] = 2282,
  [13348] = 2284,
  [13349] = 2313,
  [13350] = 2285,
  [13351] = 2286,
  [13352] = 2301,
  [13353] = 2287,
  [13354] = 13268,
  [13355] = 2288,
  [13356] = 2289,
  [13357] = 2290,
  [13358] = 2299,
  [13359] = 2316,
  [13360] = 2277,
  [13361] = 2277,
  [13362] = 2313,
  [13363] = 2278,
  [13364] = 2276,
  [13365] = 2313,
  [13366] = 2281,
  [13367] = 2255,
  [13368] = 2316,
  [13369] = 2282,
  [13370] = 2225,
  [13371] = 2272,
  [13372] = 2284,
  [13373] = 2264,
  [13374] = 2276,
  [13375] = 2281,
  [13376] = 2282,
  [13377] = 2284,
  [13378] = 2256,
  [13379] = 2285,
  [13380] = 2285,
  [13381] = 2287,
  [13382] = 2287,
  [13383] = 2288,
  [13384] = 2265,
  [13385] = 2289,
  [13386] = 2288,
  [13387] = 2290,
  [13388] = 2289,
  [13389] = 13268,
  [13390] = 2277,
  [13391] = 2290,
  [13392] = 2264,
  [13393] = 2312,
  [13394] = 2301,
  [13395] = 2283,
  [13396] = 2265,
  [13397] = 2283,
  [13398] = 2316,
  [13399] = 2278,
  [13400] = 2276,
  [13401] = 2281,
  [13402] = 2282,
  [13403] = 2284,
  [13404] = 2285,
  [13405] = 2287,
  [13406] = 2288,
  [13407] = 13268,
  [13408] = 2289,
  [13409] = 2290,
  [13410] = 2278,
  [13411] = 13268,
  [13412] = 2283,
  [13413] = 2286,
  [13414] = 13268,
  [13415] = 2281,
  [13416] = 13416,
  [13417] = 2272,
  [13418] = 2328,
  [13419] = 13419,
  [13420] = 13419,
  [13421] = 13416,
  [13422] = 8781,
  [13423] = 2265,
  [13424] = 13416,
  [13425] = 6810,
  [13426] = 13419,
  [13427] = 13419,
  [13428] = 2285,
  [13429] = 13416,
  [13430] = 13416,
  [13431] = 2287,
  [13432] = 13419,
  [13433] = 2589,
  [13434] = 2256,
  [13435] = 13419,
  [13436] = 6819,
  [13437] = 13416,
  [13438] = 2312,
  [13439] = 13416,
  [13440] = 13440,
  [13441] = 13416,
  [13442] = 13419,
  [13443] = 13419,
  [13444] = 8787,
  [13445] = 13419,
  [13446] = 13416,
  [13447] = 13419,
  [13448] = 2288,
  [13449] = 2537,
  [13450] = 2539,
  [13451] = 2485,
  [13452] = 2544,
  [13453] = 2312,
  [13454] = 2545,
  [13455] = 13416,
  [13456] = 2301,
  [13457] = 13416,
  [13458] = 2316,
  [13459] = 2301,
  [13460] = 2278,
  [13461] = 13419,
  [13462] = 2277,
  [13463] = 13419,
  [13464] = 2276,
  [13465] = 2281,
  [13466] = 2282,
  [13467] = 2284,
  [13468] = 2289,
  [13469] = 2285,
  [13470] = 2283,
  [13471] = 13416,
  [13472] = 2287,
  [13473] = 2288,
  [13474] = 2547,
  [13475] = 2289,
  [13476] = 2550,
  [13477] = 2290,
  [13478] = 2551,
  [13479] = 2552,
  [13480] = 13419,
  [13481] = 13419,
  [13482] = 2555,
  [13483] = 13416,
  [13484] = 2316,
  [13485] = 13416,
  [13486] = 2283,
  [13487] = 13416,
  [13488] = 2556,
  [13489] = 13419,
  [13490] = 13419,
  [13491] = 2290,
  [13492] = 13419,
  [13493] = 13419,
  [13494] = 13416,
  [13495] = 2558,
  [13496] = 2560,
  [13497] = 2561,
  [13498] = 2562,
  [13499] = 13416,
  [13500] = 2563,
  [13501] = 2503,
  [13502] = 13416,
  [13503] = 2504,
  [13504] = 2525,
  [13505] = 13419,
  [13506] = 2530,
  [13507] = 13419,
  [13508] = 2255,
  [13509] = 2312,
  [13510] = 13419,
  [13511] = 2316,
  [13512] = 8786,
  [13513] = 2278,
  [13514] = 2535,
  [13515] = 13416,
  [13516] = 2536,
  [13517] = 2278,
  [13518] = 13416,
  [13519] = 13419,
  [13520] = 2588,
  [13521] = 2286,
  [13522] = 2582,
  [13523] = 13416,
  [13524] = 13419,
  [13525] = 13416,
  [13526] = 2540,
  [13527] = 13416,
  [13528] = 2276,
  [13529] = 13419,
  [13530] = 2281,
  [13531] = 2282,
  [13532] = 13419,
  [13533] = 13419,
  [13534] = 2286,
  [13535] = 13419,
  [13536] = 2373,
  [13537] = 2284,
  [13538] = 2313,
  [13539] = 13416,
  [13540] = 13416,
  [13541] = 2276,
  [13542] = 2542,
  [13543] = 2572,
  [13544] = 2285,
  [13545] = 13419,
  [13546] = 2281,
  [13547] = 2287,
  [13548] = 2282,
  [13549] = 13419,
  [13550] = 2573,
  [13551] = 2574,
  [13552] = 2312,
  [13553] = 2301,
  [13554] = 13416,
  [13555] = 2316,
  [13556] = 2283,
  [13557] = 13419,
  [13558] = 13416,
  [13559] = 2278,
  [13560] = 2276,
  [13561] = 2281,
  [13562] = 13416,
  [13563] = 2282,
  [13564] = 2288,
  [13565] = 13416,
  [13566] = 13416,
  [13567] = 13419,
  [13568] = 2284,
  [13569] = 2285,
  [13570] = 2575,
  [13571] = 13416,
  [13572] = 13416,
  [13573] = 13419,
  [13574] = 2264,
  [13575] = 2289,
  [13576] = 2287,
  [13577] = 2236,
  [13578] = 13416,
  [13579] = 13419,
  [13580] = 2234,
  [13581] = 13419,
  [13582] = 2241,
  [13583] = 13416,
  [13584] = 2288,
  [13585] = 2289,
  [13586] = 13419,
  [13587] = 6835,
  [13588] = 2290,
  [13589] = 13416,
  [13590] = 13416,
  [13591] = 13419,
  [13592] = 2290,
  [13593] = 13419,
  [13594] = 2301,
  [13595] = 9675,
  [13596] = 13416,
  [13597] = 13419,
  [13598] = 13416,
  [13599] = 13419,
  [13600] = 2583,
  [13601] = 13416,
  [13602] = 13416,
  [13603] = 2284,
  [13604] = 13419,
  [13605] = 13416,
  [13606] = 13419,
  [13607] = 13419,
  [13608] = 13416,
  [13609] = 13416,
  [13610] = 13419,
  [13611] = 2373,
  [13612] = 13416,
  [13613] = 13419,
  [13614] = 13416,
  [13615] = 13419,
  [13616] = 13419,
  [13617] = 2328,
  [13618] = 13419,
  [13619] = 13416,
  [13620] = 13416,
  [13621] = 13419,
  [13622] = 13419,
  [13623] = 2283,
  [13624] = 13416,
  [13625] = 13416,
  [13626] = 2240,
  [13627] = 2255,
  [13628] = 6859,
  [13629] = 13629,
  [13630] = 9786,
  [13631] = 2290,
  [13632] = 2312,
  [13633] = 2255,
  [13634] = 6807,
  [13635] = 6860,
  [13636] = 2272,
  [13637] = 9745,
  [13638] = 9742,
  [13639] = 13639,
  [13640] = 2301,
  [13641] = 2272,
  [13642] = 13629,
  [13643] = 13629,
  [13644] = 9751,
  [13645] = 2373,
  [13646] = 2316,
  [13647] = 6848,
  [13648] = 9752,
  [13649] = 2285,
  [13650] = 2353,
  [13651] = 9835,
  [13652] = 13652,
  [13653] = 13653,
  [13654] = 13654,
  [13655] = 13639,
  [13656] = 6849,
  [13657] = 2272,
  [13658] = 6971,
  [13659] = 9726,
  [13660] = 6850,
  [13661] = 13639,
  [13662] = 2328,
  [13663] = 2276,
  [13664] = 13440,
  [13665] = 2256,
  [13666] = 6921,
  [13667] = 2313,
  [13668] = 2278,
  [13669] = 9738,
  [13670] = 13629,
  [13671] = 13671,
  [13672] = 6938,
  [13673] = 2383,
  [13674] = 9782,
  [13675] = 13653,
  [13676] = 2281,
  [13677] = 6872,
  [13678] = 13653,
  [13679] = 2288,
  [13680] = 6886,
  [13681] = 2287,
  [13682] = 13639,
  [13683] = 6851,
  [13684] = 6852,
  [13685] = 2255,
  [13686] = 2282,
  [13687] = 2367,
  [13688] = 13629,
  [13689] = 6853,
  [13690] = 9694,
  [13691] = 13629,
  [13692] = 2286,
  [13693] = 9741,
  [13694] = 2283,
  [13695] = 6854,
  [13696] = 2369,
  [13697] = 9863,
  [13698] = 13698,
  [13699] = 2256,
  [13700] = 9744,
  [13701] = 9868,
  [13702] = 13653,
  [13703] = 2289,
  [13704] = 2367,
  [13705] = 9689,
  [13706] = 6855,
  [13707] = 2284,
  [13708] = 8733,
  [13709] = 13629,
  [13710] = 13653,
  [13711] = 2353,
  [13712] = 13639,
  [13713] = 13713,
  [13714] = 13671,
  [13715] = 13653,
  [13716] = 8740,
  [13717] = 13639,
  [13718] = 6919,
  [13719] = 9754,
  [13720] = 13720,
  [13721] = 13652,
  [13722] = 13629,
  [13723] = 9743,
  [13724] = 2369,
  [13725] = 13639,
  [13726] = 13639,
  [13727] = 13629,
  [13728] = 4827,
  [13729] = 13629,
  [13730] = 13698,
  [13731] = 13731,
  [13732] = 2256,
  [13733] = 13731,
  [13734] = 13639,
  [13735] = 13653,
  [13736] = 13639,
  [13737] = 13653,
  [13738] = 13639,
  [13739] = 9734,
  [13740] = 9717,
  [13741] = 2383,
  [13742] = 9843,
  [13743] = 9848,
  [13744] = 13639,
  [13745] = 9853,
  [13746] = 13629,
  [13747] = 13629,
  [13748] = 9766,
  [13749] = 6856,
  [13750] = 6857,
  [13751] = 6858,
  [13752] = 9820,
  [13753] = 13653,
  [13754] = 9839,
  [13755] = 9667,
  [13756] = 2236,
  [13757] = 13757,
  [13758] = 2234,
  [13759] = 2241,
  [13760] = 6971,
  [13761] = 13720,
  [13762] = 6926,
  [13763] = 13757,
  [13764] = 9605,
  [13765] = 13757,
  [13766] = 13757,
  [13767] = 13767,
  [13768] = 13757,
  [13769] = 6957,
  [13770] = 13767,
  [13771] = 13757,
  [13772] = 13757,
  [13773] = 13757,
  [13774] = 6918,
  [13775] = 13757,
  [13776] = 13757,
  [13777] = 13757,
  [13778] = 13757,
  [13779] = 13757,
  [13780] = 13757,
  [13781] = 9613,
  [13782] = 13757,
  [13783] = 13783,
  [13784] = 6958,
  [13785] = 9606,
  [13786] = 6955,
  [13787] = 13783,
  [13788] = 13757,
  [13789] = 8954,
  [13790] = 9577,
  [13791] = 2277,
  [13792] = 13792,
  [13793] = 9541,
  [13794] = 13794,
  [13795] = 2259,
  [13796] = 2278,
  [13797] = 6927,
  [13798] = 2259,
  [13799] = 2263,
  [13800] = 8949,
  [13801] = 13767,
  [13802] = 6941,
  [13803] = 6942,
  [13804] = 13767,
  [13805] = 6943,
  [13806] = 6944,
  [13807] = 6945,
  [13808] = 6898,
  [13809] = 13783,
  [13810] = 13757,
  [13811] = 9595,
  [13812] = 2265,
  [13813] = 2258,
  [13814] = 6895,
  [13815] = 2258,
  [13816] = 13757,
  [13817] = 6922,
  [13818] = 13767,
  [13819] = 13757,
  [13820] = 6956,
  [13821] = 9564,
  [13822] = 2258,
  [13823] = 13757,
  [13824] = 6871,
  [13825] = 2236,
  [13826] = 2313,
  [13827] = 9578,
  [13828] = 2234,
  [13829] = 9613,
  [13830] = 9595,
  [13831] = 2241,
  [13832] = 13767,
  [13833] = 13767,
  [13834] = 13757,
  [13835] = 6921,
  [13836] = 13757,
  [13837] = 2258,
  [13838] = 9564,
  [13839] = 2264,
  [13840] = 13783,
  [13841] = 13767,
  [13842] = 13783,
  [13843] = 13783,
  [13844] = 6938,
  [13845] = 2259,
  [13846] = 13757,
  [13847] = 13783,
  [13848] = 6931,
  [13849] = 13849,
  [13850] = 2259,
  [13851] = 2258,
  [13852] = 13757,
  [13853] = 13853,
  [13854] = 13854,
  [13855] = 13757,
  [13856] = 13783,
  [13857] = 6948,
  [13858] = 9661,
  [13859] = 2263,
  [13860] = 6932,
  [13861] = 13783,
  [13862] = 13767,
  [13863] = 13757,
  [13864] = 13783,
  [13865] = 2353,
  [13866] = 6961,
  [13867] = 6890,
  [13868] = 6962,
  [13869] = 13869,
  [13870] = 13757,
  [13871] = 13871,
  [13872] = 9673,
  [13873] = 13757,
  [13874] = 13767,
  [13875] = 13875,
  [13876] = 2313,
  [13877] = 13783,
  [13878] = 2313,
  [13879] = 6925,
  [13880] = 13757,
  [13881] = 2369,
  [13882] = 13757,
  [13883] = 13767,
  [13884] = 2383,
  [13885] = 2240,
  [13886] = 2259,
  [13887] = 13757,
  [13888] = 13757,
  [13889] = 2263,
  [13890] = 13767,
  [13891] = 13757,
  [13892] = 6919,
  [13893] = 9613,
  [13894] = 13757,
  [13895] = 9564,
  [13896] = 13896,
  [13897] = 13757,
  [13898] = 6949,
  [13899] = 13783,
  [13900] = 13783,
  [13901] = 13757,
  [13902] = 2240,
  [13903] = 13903,
  [13904] = 6953,
  [13905] = 13905,
  [13906] = 9595,
  [13907] = 9589,
  [13908] = 6967,
  [13909] = 6954,
  [13910] = 13757,
  [13911] = 13767,
  [13912] = 9541,
  [13913] = 13767,
  [13914] = 13914,
  [13915] = 13783,
  [13916] = 13757,
  [13917] = 13783,
  [13918] = 6936,
  [13919] = 6937,
  [13920] = 6872,
  [13921] = 9574,
  [13922] = 13767,
  [13923] = 13923,
  [13924] = 13924,
  [13925] = 13925,
  [13926] = 2551,
  [13927] = 2292,
  [13928] = 13925,
  [13929] = 2552,
  [13930] = 2555,
  [13931] = 2556,
  [13932] = 2558,
  [13933] = 13933,
  [13934] = 13924,
  [13935] = 13925,
  [13936] = 2560,
  [13937] = 2561,
  [13938] = 13924,
  [13939] = 2562,
  [13940] = 2563,
  [13941] = 2503,
  [13942] = 13925,
  [13943] = 13943,
  [13944] = 2504,
  [13945] = 13924,
  [13946] = 2525,
  [13947] = 2530,
  [13948] = 13925,
  [13949] = 13949,
  [13950] = 2535,
  [13951] = 2536,
  [13952] = 13924,
  [13953] = 13925,
  [13954] = 2540,
  [13955] = 9655,
  [13956] = 13924,
  [13957] = 2256,
  [13958] = 13925,
  [13959] = 13924,
  [13960] = 13925,
  [13961] = 13924,
  [13962] = 13962,
  [13963] = 13924,
  [13964] = 13914,
  [13965] = 13933,
  [13966] = 13925,
  [13967] = 13925,
  [13968] = 13924,
  [13969] = 2542,
  [13970] = 2572,
  [13971] = 13925,
  [13972] = 13933,
  [13973] = 2573,
  [13974] = 13924,
  [13975] = 2574,
  [13976] = 13925,
  [13977] = 2299,
  [13978] = 13924,
  [13979] = 13925,
  [13980] = 13933,
  [13981] = 2236,
  [13982] = 13924,
  [13983] = 2234,
  [13984] = 2241,
  [13985] = 13925,
  [13986] = 13933,
  [13987] = 9641,
  [13988] = 13924,
  [13989] = 2588,
  [13990] = 13925,
  [13991] = 2589,
  [13992] = 13943,
  [13993] = 13993,
  [13994] = 2575,
  [13995] = 9588,
  [13996] = 2258,
  [13997] = 13943,
  [13998] = 9600,
  [13999] = 13924,
  [14000] = 14000,
  [14001] = 2259,
  [14002] = 13925,
  [14003] = 9657,
  [14004] = 13924,
  [14005] = 13943,
  [14006] = 13924,
  [14007] = 13933,
  [14008] = 13943,
  [14009] = 8787,
  [14010] = 13962,
  [14011] = 14011,
  [14012] = 13925,
  [14013] = 13924,
  [14014] = 9564,
  [14015] = 13925,
  [14016] = 2263,
  [14017] = 2258,
  [14018] = 13962,
  [14019] = 2272,
  [14020] = 9658,
  [14021] = 13924,
  [14022] = 14022,
  [14023] = 2256,
  [14024] = 2292,
  [14025] = 13925,
  [14026] = 14026,
  [14027] = 2240,
  [14028] = 2259,
  [14029] = 13933,
  [14030] = 14030,
  [14031] = 13925,
  [14032] = 13924,
  [14033] = 13933,
  [14034] = 4923,
  [14035] = 4924,
  [14036] = 13925,
  [14037] = 13943,
  [14038] = 13943,
  [14039] = 13962,
  [14040] = 14040,
  [14041] = 2272,
  [14042] = 2292,
  [14043] = 14043,
  [14044] = 2256,
  [14045] = 2299,
  [14046] = 13993,
  [14047] = 13924,
  [14048] = 13962,
  [14049] = 14040,
  [14050] = 13925,
  [14051] = 2299,
  [14052] = 14043,
  [14053] = 8786,
  [14054] = 2258,
  [14055] = 13993,
  [14056] = 13924,
  [14057] = 13924,
  [14058] = 14040,
  [14059] = 2259,
  [14060] = 4942,
  [14061] = 13962,
  [14062] = 14043,
  [14063] = 9613,
  [14064] = 14040,
  [14065] = 2255,
  [14066] = 9595,
  [14067] = 2299,
  [14068] = 13925,
  [14069] = 13925,
  [14070] = 13993,
  [14071] = 13924,
  [14072] = 13962,
  [14073] = 14040,
  [14074] = 13925,
  [14075] = 14043,
  [14076] = 13993,
  [14077] = 2316,
  [14078] = 13925,
  [14079] = 13962,
  [14080] = 2240,
  [14081] = 14040,
  [14082] = 14043,
  [14083] = 9613,
  [14084] = 13943,
  [14085] = 13924,
  [14086] = 9675,
  [14087] = 13993,
  [14088] = 13925,
  [14089] = 14040,
  [14090] = 2236,
  [14091] = 2234,
  [14092] = 14043,
  [14093] = 2241,
  [14094] = 13854,
  [14095] = 13869,
  [14096] = 13933,
  [14097] = 14097,
  [14098] = 2292,
  [14099] = 13993,
  [14100] = 2255,
  [14101] = 14040,
  [14102] = 13924,
  [14103] = 2258,
  [14104] = 2290,
  [14105] = 14043,
  [14106] = 11843,
  [14107] = 13925,
  [14108] = 13943,
  [14109] = 2259,
  [14110] = 13993,
  [14111] = 8781,
  [14112] = 14112,
  [14113] = 14040,
  [14114] = 13924,
  [14115] = 13925,
  [14116] = 13993,
  [14117] = 14117,
  [14118] = 2272,
  [14119] = 13924,
  [14120] = 14040,
  [14121] = 2292,
  [14122] = 13925,
  [14123] = 13925,
  [14124] = 13924,
  [14125] = 13962,
  [14126] = 14126,
  [14127] = 13993,
  [14128] = 2582,
  [14129] = 2583,
  [14130] = 9541,
  [14131] = 13925,
  [14132] = 14040,
  [14133] = 9675,
  [14134] = 13924,
  [14135] = 13993,
  [14136] = 14040,
  [14137] = 2299,
  [14138] = 9595,
  [14139] = 13925,
  [14140] = 13993,
  [14141] = 14040,
  [14142] = 13993,
  [14143] = 14040,
  [14144] = 2263,
  [14145] = 13933,
  [14146] = 14040,
  [14147] = 13933,
  [14148] = 13924,
  [14149] = 14149,
  [14150] = 14040,
  [14151] = 9661,
  [14152] = 13925,
  [14153] = 2276,
  [14154] = 9667,
  [14155] = 13943,
  [14156] = 2289,
  [14157] = 14040,
  [14158] = 2281,
  [14159] = 9673,
  [14160] = 2282,
  [14161] = 2588,
  [14162] = 2589,
  [14163] = 9661,
  [14164] = 14040,
  [14165] = 9667,
  [14166] = 9673,
  [14167] = 14040,
  [14168] = 13924,
  [14169] = 2236,
  [14170] = 2234,
  [14171] = 14040,
  [14172] = 13925,
  [14173] = 14173,
  [14174] = 13943,
  [14175] = 2241,
  [14176] = 14040,
  [14177] = 9651,
  [14178] = 14178,
  [14179] = 14040,
  [14180] = 14040,
  [14181] = 13924,
  [14182] = 13924,
  [14183] = 2240,
  [14184] = 9574,
  [14185] = 9577,
  [14186] = 9578,
  [14187] = 14040,
  [14188] = 9589,
  [14189] = 9605,
  [14190] = 13962,
  [14191] = 9606,
  [14192] = 13925,
  [14193] = 14040,
  [14194] = 14040,
  [14195] = 13925,
  [14196] = 2255,
  [14197] = 14040,
  [14198] = 13924,
  [14199] = 14040,
  [14200] = 14040,
  [14201] = 14040,
  [14202] = 9782,
  [14203] = 13925,
  [14204] = 14040,
  [14205] = 14040,
  [14206] = 14040,
  [14207] = 2263,
  [14208] = 14040,
  [14209] = 14040,
  [14210] = 14040,
  [14211] = 14040,
  [14212] = 14040,
  [14213] = 14040,
  [14214] = 14040,
  [14215] = 14040,
  [14216] = 14040,
  [14217] = 2582,
  [14218] = 14040,
  [14219] = 14040,
  [14220] = 2583,
  [14221] = 14040,
  [14222] = 13924,
  [14223] = 14040,
  [14224] = 14043,
  [14225] = 14040,
  [14226] = 14040,
  [14227] = 14040,
  [14228] = 14040,
  [14229] = 14040,
  [14230] = 14040,
  [14231] = 14040,
  [14232] = 14040,
  [14233] = 2255,
  [14234] = 14040,
  [14235] = 14040,
  [14236] = 13925,
  [14237] = 14040,
  [14238] = 14040,
  [14239] = 14040,
  [14240] = 14040,
  [14241] = 2236,
  [14242] = 14040,
  [14243] = 14040,
  [14244] = 2234,
  [14245] = 14040,
  [14246] = 2241,
  [14247] = 14040,
  [14248] = 2285,
  [14249] = 14040,
  [14250] = 14040,
  [14251] = 14040,
  [14252] = 14040,
  [14253] = 13933,
  [14254] = 13924,
  [14255] = 14040,
  [14256] = 14040,
  [14257] = 9588,
  [14258] = 14040,
  [14259] = 9600,
  [14260] = 13925,
  [14261] = 14040,
  [14262] = 2284,
  [14263] = 14040,
  [14264] = 2287,
  [14265] = 14040,
  [14266] = 14040,
  [14267] = 14040,
  [14268] = 14040,
  [14269] = 14040,
  [14270] = 14040,
  [14271] = 14040,
  [14272] = 14040,
  [14273] = 14040,
  [14274] = 13943,
  [14275] = 14040,
  [14276] = 13924,
  [14277] = 14040,
  [14278] = 9574,
  [14279] = 14040,
  [14280] = 14040,
  [14281] = 9577,
  [14282] = 14040,
  [14283] = 2240,
  [14284] = 14040,
  [14285] = 13925,
  [14286] = 14040,
  [14287] = 9578,
  [14288] = 14040,
  [14289] = 9589,
  [14290] = 14040,
  [14291] = 9605,
  [14292] = 14040,
  [14293] = 14040,
  [14294] = 14040,
  [14295] = 9606,
  [14296] = 14040,
  [14297] = 4923,
  [14298] = 14040,
  [14299] = 14040,
  [14300] = 9641,
  [14301] = 14040,
  [14302] = 4924,
  [14303] = 2272,
  [14304] = 14040,
  [14305] = 14040,
  [14306] = 14040,
  [14307] = 13962,
  [14308] = 14040,
  [14309] = 14040,
  [14310] = 14040,
  [14311] = 13923,
  [14312] = 14173,
  [14313] = 13924,
  [14314] = 9613,
  [14315] = 9595,
  [14316] = 13925,
  [14317] = 9651,
  [14318] = 13923,
  [14319] = 2537,
  [14320] = 14173,
  [14321] = 2539,
  [14322] = 13924,
  [14323] = 2485,
  [14324] = 9655,
  [14325] = 13923,
  [14326] = 14173,
  [14327] = 9564,
  [14328] = 13924,
  [14329] = 2256,
  [14330] = 13924,
  [14331] = 2544,
  [14332] = 13923,
  [14333] = 14173,
  [14334] = 9564,
  [14335] = 13925,
  [14336] = 13962,
  [14337] = 13923,
  [14338] = 14173,
  [14339] = 2312,
  [14340] = 13925,
  [14341] = 13923,
  [14342] = 14173,
  [14343] = 14343,
  [14344] = 13923,
  [14345] = 9657,
  [14346] = 14173,
  [14347] = 13924,
  [14348] = 9658,
  [14349] = 13923,
  [14350] = 14173,
  [14351] = 13924,
  [14352] = 2537,
  [14353] = 13923,
  [14354] = 2539,
  [14355] = 14173,
  [14356] = 13923,
  [14357] = 14173,
  [14358] = 13923,
  [14359] = 14173,
  [14360] = 2485,
  [14361] = 2544,
  [14362] = 14362,
  [14363] = 13925,
  [14364] = 2545,
  [14365] = 13923,
  [14366] = 14173,
  [14367] = 13923,
  [14368] = 2545,
  [14369] = 14173,
  [14370] = 13923,
  [14371] = 14173,
  [14372] = 13923,
  [14373] = 14173,
  [14374] = 13962,
  [14375] = 13923,
  [14376] = 14173,
  [14377] = 13923,
  [14378] = 14173,
  [14379] = 13933,
  [14380] = 13923,
  [14381] = 13924,
  [14382] = 14173,
  [14383] = 2547,
  [14384] = 14112,
  [14385] = 13923,
  [14386] = 2550,
  [14387] = 14173,
  [14388] = 2551,
  [14389] = 13923,
  [14390] = 2552,
  [14391] = 14173,
  [14392] = 13924,
  [14393] = 2555,
  [14394] = 13923,
  [14395] = 2556,
  [14396] = 14173,
  [14397] = 2558,
  [14398] = 13923,
  [14399] = 2560,
  [14400] = 14173,
  [14401] = 2561,
  [14402] = 2562,
  [14403] = 13943,
  [14404] = 13923,
  [14405] = 13925,
  [14406] = 14173,
  [14407] = 13923,
  [14408] = 2563,
  [14409] = 2503,
  [14410] = 14173,
  [14411] = 2504,
  [14412] = 2525,
  [14413] = 2530,
  [14414] = 13923,
  [14415] = 14173,
  [14416] = 13923,
  [14417] = 2535,
  [14418] = 14173,
  [14419] = 13923,
  [14420] = 14173,
  [14421] = 13923,
  [14422] = 2536,
  [14423] = 14173,
  [14424] = 13923,
  [14425] = 14173,
  [14426] = 13923,
  [14427] = 14173,
  [14428] = 13923,
  [14429] = 14173,
  [14430] = 13923,
  [14431] = 14173,
  [14432] = 13923,
  [14433] = 14173,
  [14434] = 13923,
  [14435] = 14435,
  [14436] = 14173,
  [14437] = 13923,
  [14438] = 14173,
  [14439] = 13923,
  [14440] = 14173,
  [14441] = 13923,
  [14442] = 14173,
  [14443] = 13923,
  [14444] = 14173,
  [14445] = 13923,
  [14446] = 14173,
  [14447] = 13923,
  [14448] = 14173,
  [14449] = 13923,
  [14450] = 14173,
  [14451] = 13923,
  [14452] = 14173,
  [14453] = 13923,
  [14454] = 14173,
  [14455] = 13923,
  [14456] = 14173,
  [14457] = 13923,
  [14458] = 14173,
  [14459] = 13923,
  [14460] = 14173,
  [14461] = 13923,
  [14462] = 14173,
  [14463] = 13923,
  [14464] = 14173,
  [14465] = 13923,
  [14466] = 14173,
  [14467] = 13923,
  [14468] = 14173,
  [14469] = 13923,
  [14470] = 14173,
  [14471] = 13923,
  [14472] = 14173,
  [14473] = 13923,
  [14474] = 14173,
  [14475] = 13923,
  [14476] = 14173,
  [14477] = 13923,
  [14478] = 14173,
  [14479] = 13923,
  [14480] = 14173,
  [14481] = 13923,
  [14482] = 14173,
  [14483] = 13923,
  [14484] = 14173,
  [14485] = 13923,
  [14486] = 14173,
  [14487] = 13923,
  [14488] = 14173,
  [14489] = 13923,
  [14490] = 14173,
  [14491] = 13923,
  [14492] = 14173,
  [14493] = 13923,
  [14494] = 14173,
  [14495] = 13923,
  [14496] = 14173,
  [14497] = 13923,
  [14498] = 14173,
  [14499] = 13923,
  [14500] = 14173,
  [14501] = 13923,
  [14502] = 14173,
  [14503] = 13923,
  [14504] = 14173,
  [14505] = 13923,
  [14506] = 14173,
  [14507] = 13923,
  [14508] = 14173,
  [14509] = 13923,
  [14510] = 14173,
  [14511] = 13923,
  [14512] = 14173,
  [14513] = 13923,
  [14514] = 14173,
  [14515] = 13923,
  [14516] = 14173,
  [14517] = 13923,
  [14518] = 14173,
  [14519] = 13923,
  [14520] = 14173,
  [14521] = 13923,
  [14522] = 14173,
  [14523] = 13923,
  [14524] = 14173,
  [14525] = 13923,
  [14526] = 14173,
  [14527] = 13923,
  [14528] = 14173,
  [14529] = 13923,
  [14530] = 14173,
  [14531] = 13923,
  [14532] = 14173,
  [14533] = 13923,
  [14534] = 14173,
  [14535] = 13923,
  [14536] = 14173,
  [14537] = 13923,
  [14538] = 14173,
  [14539] = 13923,
  [14540] = 14173,
  [14541] = 13923,
  [14542] = 14173,
  [14543] = 13923,
  [14544] = 14173,
  [14545] = 13923,
  [14546] = 14173,
  [14547] = 13923,
  [14548] = 14173,
  [14549] = 13923,
  [14550] = 14173,
  [14551] = 13923,
  [14552] = 14173,
  [14553] = 13923,
  [14554] = 14173,
  [14555] = 13923,
  [14556] = 14173,
  [14557] = 13923,
  [14558] = 14173,
  [14559] = 13923,
  [14560] = 14173,
  [14561] = 13923,
  [14562] = 14173,
  [14563] = 13923,
  [14564] = 14173,
  [14565] = 13923,
  [14566] = 14173,
  [14567] = 13923,
  [14568] = 14173,
  [14569] = 13925,
  [14570] = 14022,
  [14571] = 14097,
  [14572] = 2288,
  [14573] = 14022,
  [14574] = 14097,
  [14575] = 14022,
  [14576] = 14097,
  [14577] = 2540,
  [14578] = 2301,
  [14579] = 14022,
  [14580] = 14097,
  [14581] = 14581,
  [14582] = 14022,
  [14583] = 14097,
  [14584] = 14022,
  [14585] = 14097,
  [14586] = 14022,
  [14587] = 14097,
  [14588] = 4942,
  [14589] = 13924,
  [14590] = 2542,
  [14591] = 2572,
  [14592] = 2573,
  [14593] = 2574,
  [14594] = 13925,
  [14595] = 13962,
  [14596] = 14596,
  [14597] = 2283,
  [14598] = 2575,
  [14599] = 13924,
  [14600] = 14600,
  [14601] = 14601,
  [14602] = 9782,
  [14603] = 13925,
  [14604] = 2547,
  [14605] = 13924,
  [14606] = 2550,
  [14607] = 14040,
  [14608] = 14608,
  [14609] = 14609,
  [14610] = 14610,
  [14611] = 14611,
  [14612] = 2575,
  [14613] = 14613,
  [14614] = 14611,
  [14615] = 14611,
  [14616] = 2292,
  [14617] = 14611,
  [14618] = 14618,
  [14619] = 14611,
  [14620] = 14620,
  [14621] = 14621,
  [14622] = 14611,
  [14623] = 2277,
  [14624] = 14608,
  [14625] = 14611,
  [14626] = 14626,
  [14627] = 14627,
  [14628] = 14581,
  [14629] = 14629,
  [14630] = 9786,
  [14631] = 13869,
  [14632] = 14609,
  [14633] = 14629,
  [14634] = 14611,
  [14635] = 14613,
  [14636] = 2292,
  [14637] = 2264,
  [14638] = 14611,
  [14639] = 13854,
  [14640] = 2255,
  [14641] = 14620,
  [14642] = 14611,
  [14643] = 14643,
  [14644] = 2299,
  [14645] = 14611,
  [14646] = 14609,
  [14647] = 9743,
  [14648] = 9745,
  [14649] = 14621,
  [14650] = 14621,
  [14651] = 14613,
  [14652] = 14652,
  [14653] = 14652,
  [14654] = 14620,
  [14655] = 9673,
  [14656] = 2286,
  [14657] = 14626,
  [14658] = 14611,
  [14659] = 14659,
  [14660] = 2272,
  [14661] = 14611,
  [14662] = 14662,
  [14663] = 14609,
  [14664] = 2256,
  [14665] = 14610,
  [14666] = 9577,
  [14667] = 14610,
  [14668] = 14613,
  [14669] = 14629,
  [14670] = 9868,
  [14671] = 14611,
  [14672] = 14659,
  [14673] = 14611,
  [14674] = 14674,
  [14675] = 14611,
  [14676] = 14609,
  [14677] = 14626,
  [14678] = 2277,
  [14679] = 14662,
  [14680] = 14611,
  [14681] = 14613,
  [14682] = 14618,
  [14683] = 14627,
  [14684] = 9820,
  [14685] = 9694,
  [14686] = 9835,
  [14687] = 14620,
  [14688] = 14611,
  [14689] = 2299,
  [14690] = 14609,
  [14691] = 14613,
  [14692] = 14613,
  [14693] = 9588,
  [14694] = 14611,
  [14695] = 2264,
  [14696] = 9600,
  [14697] = 14609,
  [14698] = 14626,
  [14699] = 14611,
  [14700] = 14700,
  [14701] = 14611,
  [14702] = 14613,
  [14703] = 2265,
  [14704] = 2286,
  [14705] = 2255,
  [14706] = 9657,
  [14707] = 14609,
  [14708] = 14708,
  [14709] = 9786,
  [14710] = 14618,
  [14711] = 14611,
  [14712] = 14613,
  [14713] = 14609,
  [14714] = 2265,
  [14715] = 14608,
  [14716] = 14613,
  [14717] = 14611,
  [14718] = 9578,
  [14719] = 14609,
  [14720] = 14608,
  [14721] = 14629,
  [14722] = 9658,
  [14723] = 14613,
  [14724] = 14611,
  [14725] = 14611,
  [14726] = 9741,
  [14727] = 14621,
  [14728] = 14611,
  [14729] = 14610,
  [14730] = 9742,
  [14731] = 14621,
  [14732] = 14613,
  [14733] = 14652,
  [14734] = 14611,
  [14735] = 14626,
  [14736] = 9744,
  [14737] = 14611,
  [14738] = 9661,
  [14739] = 14611,
  [14740] = 14613,
  [14741] = 14629,
  [14742] = 14611,
  [14743] = 14743,
  [14744] = 9863,
  [14745] = 14613,
  [14746] = 14627,
  [14747] = 14659,
  [14748] = 14611,
  [14749] = 14662,
  [14750] = 14613,
  [14751] = 14608,
  [14752] = 14611,
  [14753] = 14611,
  [14754] = 14629,
  [14755] = 14613,
  [14756] = 9667,
  [14757] = 14621,
  [14758] = 14620,
  [14759] = 14613,
  [14760] = 14652,
  [14761] = 9752,
  [14762] = 9754,
  [14763] = 14613,
  [14764] = 14611,
  [14765] = 14621,
  [14766] = 9868,
  [14767] = 9766,
  [14768] = 14126,
  [14769] = 14613,
  [14770] = 14611,
  [14771] = 2255,
  [14772] = 2286,
  [14773] = 9641,
  [14774] = 14343,
  [14775] = 14611,
  [14776] = 14611,
  [14777] = 9751,
  [14778] = 14613,
  [14779] = 2263,
  [14780] = 2240,
  [14781] = 9589,
  [14782] = 14613,
  [14783] = 14652,
  [14784] = 14613,
  [14785] = 9726,
  [14786] = 9738,
  [14787] = 14611,
  [14788] = 14788,
  [14789] = 14613,
  [14790] = 14611,
  [14791] = 14621,
  [14792] = 14611,
  [14793] = 14611,
  [14794] = 9655,
  [14795] = 13794,
  [14796] = 14626,
  [14797] = 14611,
  [14798] = 14611,
  [14799] = 9717,
  [14800] = 13914,
  [14801] = 9574,
  [14802] = 14627,
  [14803] = 14803,
  [14804] = 9839,
  [14805] = 9741,
  [14806] = 9734,
  [14807] = 9863,
  [14808] = 14652,
  [14809] = 14611,
  [14810] = 14620,
  [14811] = 14611,
  [14812] = 14812,
  [14813] = 14611,
  [14814] = 14629,
  [14815] = 9742,
  [14816] = 9744,
  [14817] = 14626,
  [14818] = 14611,
  [14819] = 14819,
  [14820] = 14820,
  [14821] = 14610,
  [14822] = 9673,
  [14823] = 14611,
  [14824] = 2582,
  [14825] = 2583,
  [14826] = 14611,
  [14827] = 14659,
  [14828] = 14611,
  [14829] = 14662,
  [14830] = 14830,
  [14831] = 9751,
  [14832] = 14611,
  [14833] = 14627,
  [14834] = 14611,
  [14835] = 14611,
  [14836] = 14618,
  [14837] = 2292,
  [14838] = 9577,
  [14839] = 14611,
  [14840] = 9694,
  [14841] = 14841,
  [14842] = 14611,
  [14843] = 9578,
  [14844] = 14608,
  [14845] = 9589,
  [14846] = 9605,
  [14847] = 14611,
  [14848] = 14621,
  [14849] = 14611,
  [14850] = 2258,
  [14851] = 9606,
  [14852] = 2588,
  [14853] = 14611,
  [14854] = 2589,
  [14855] = 14620,
  [14856] = 9752,
  [14857] = 14611,
  [14858] = 9754,
  [14859] = 14618,
  [14860] = 9675,
  [14861] = 14629,
  [14862] = 14611,
  [14863] = 2313,
  [14864] = 14626,
  [14865] = 14626,
  [14866] = 9766,
  [14867] = 9782,
  [14868] = 14611,
  [14869] = 14626,
  [14870] = 2313,
  [14871] = 14611,
  [14872] = 14609,
  [14873] = 14611,
  [14874] = 14611,
  [14875] = 14629,
  [14876] = 14621,
  [14877] = 14652,
  [14878] = 14611,
  [14879] = 14610,
  [14880] = 14611,
  [14881] = 14611,
  [14882] = 14659,
  [14883] = 14611,
  [14884] = 9605,
  [14885] = 14662,
  [14886] = 14611,
  [14887] = 14608,
  [14888] = 14618,
  [14889] = 14788,
  [14890] = 14659,
  [14891] = 14891,
  [14892] = 14662,
  [14893] = 14611,
  [14894] = 14894,
  [14895] = 14611,
  [14896] = 14611,
  [14897] = 14611,
  [14898] = 9564,
  [14899] = 9651,
  [14900] = 14788,
  [14901] = 14618,
  [14902] = 14611,
  [14903] = 2299,
  [14904] = 14611,
  [14905] = 14788,
  [14906] = 2313,
  [14907] = 14611,
  [14908] = 14608,
  [14909] = 2256,
  [14910] = 14700,
  [14911] = 14611,
  [14912] = 2272,
  [14913] = 14788,
  [14914] = 14611,
  [14915] = 14611,
  [14916] = 14788,
  [14917] = 14917,
  [14918] = 2236,
  [14919] = 14629,
  [14920] = 2234,
  [14921] = 14803,
  [14922] = 14788,
  [14923] = 14611,
  [14924] = 14611,
  [14925] = 14611,
  [14926] = 9843,
  [14927] = 14627,
  [14928] = 14788,
  [14929] = 9848,
  [14930] = 14611,
  [14931] = 14788,
  [14932] = 9853,
  [14933] = 14788,
  [14934] = 14611,
  [14935] = 14620,
  [14936] = 14611,
  [14937] = 9689,
  [14938] = 14621,
  [14939] = 14611,
  [14940] = 4923,
  [14941] = 14743,
  [14942] = 14652,
  [14943] = 14626,
  [14944] = 2241,
  [14945] = 4924,
  [14946] = 14610,
  [14947] = 14613,
  [14948] = 14618,
  [14949] = 9606,
  [14950] = 14618,
  [14951] = 2286,
  [14952] = 14610,
  [14953] = 14608,
  [14954] = 2313,
  [14955] = 14611,
  [14956] = 14608,
  [14957] = 14659,
  [14958] = 14629,
  [14959] = 14662,
  [14960] = 14611,
  [14961] = 14621,
  [14962] = 14609,
  [14963] = 14652,
  [14964] = 14611,
  [14965] = 14659,
  [14966] = 14620,
  [14967] = 9661,
  [14968] = 9667,
  [14969] = 14613,
  [14970] = 14627,
  [14971] = 14971,
  [14972] = 14611,
  [14973] = 14662,
  [14974] = 14620,
  [14975] = 14975,
  [14976] = 9613,
  [14977] = 9595,
  [14978] = 14621,
  [14979] = 9843,
  [14980] = 2272,
  [14981] = 2292,
  [14982] = 14618,
  [14983] = 14627,
  [14984] = 9848,
  [14985] = 9853,
  [14986] = 14609,
  [14987] = 14600,
  [14988] = 14652,
  [14989] = 2256,
  [14990] = 14611,
  [14991] = 14626,
  [14992] = 4942,
  [14993] = 2299,
  [14994] = 14613,
  [14995] = 14608,
  [14996] = 14627,
  [14997] = 14608,
  [14998] = 14611,
  [14999] = 9689,
  [15000] = 9574,
  [15001] = 14611,
  [15002] = 14611,
  [15003] = 2537,
  [15004] = 2539,
  [15005] = 14609,
  [15006] = 14611,
  [15007] = 2485,
  [15008] = 2544,
  [15009] = 2272,
  [15010] = 14629,
  [15011] = 2545,
  [15012] = 14613,
  [15013] = 2256,
  [15014] = 14608,
  [15015] = 14627,
  [15016] = 14611,
  [15017] = 2547,
  [15018] = 14659,
  [15019] = 2550,
  [15020] = 2551,
  [15021] = 2552,
  [15022] = 2555,
  [15023] = 2556,
  [15024] = 2558,
  [15025] = 2560,
  [15026] = 14662,
  [15027] = 2561,
  [15028] = 2562,
  [15029] = 2563,
  [15030] = 2503,
  [15031] = 2504,
  [15032] = 2525,
  [15033] = 2530,
  [15034] = 2535,
  [15035] = 2536,
  [15036] = 14609,
  [15037] = 9743,
  [15038] = 2540,
  [15039] = 14659,
  [15040] = 14662,
  [15041] = 9745,
  [15042] = 14629,
  [15043] = 9717,
  [15044] = 14613,
  [15045] = 14600,
  [15046] = 14611,
  [15047] = 14620,
  [15048] = 9734,
  [15049] = 14743,
  [15050] = 9726,
  [15051] = 15051,
  [15052] = 14611,
  [15053] = 15053,
  [15054] = 9820,
  [15055] = 9835,
  [15056] = 14611,
  [15057] = 2542,
  [15058] = 2572,
  [15059] = 9738,
  [15060] = 2573,
  [15061] = 2574,
  [15062] = 14611,
  [15063] = 14620,
  [15064] = 14626,
  [15065] = 14611,
  [15066] = 9839,
  [15067] = 15067,
  [15068] = 2259,
  [15069] = 5093,
  [15070] = 14674,
  [15071] = 2277,
  [15072] = 2285,
  [15073] = 7465,
  [15074] = 15074,
  [15075] = 15075,
  [15076] = 15076,
  [15077] = 15077,
  [15078] = 2287,
  [15079] = 15079,
  [15080] = 13914,
  [15081] = 15081,
  [15082] = 2265,
  [15083] = 2288,
  [15084] = 15084,
  [15085] = 15085,
  [15086] = 2373,
  [15087] = 15079,
  [15088] = 2289,
  [15089] = 15085,
  [15090] = 2292,
  [15091] = 2290,
  [15092] = 15085,
  [15093] = 2277,
  [15094] = 2328,
  [15095] = 15085,
  [15096] = 2299,
  [15097] = 2287,
  [15098] = 15076,
  [15099] = 9726,
  [15100] = 2313,
  [15101] = 9738,
  [15102] = 9839,
  [15103] = 9863,
  [15104] = 2312,
  [15105] = 15076,
  [15106] = 15081,
  [15107] = 2301,
  [15108] = 15108,
  [15109] = 7465,
  [15110] = 7482,
  [15111] = 2283,
  [15112] = 15081,
  [15113] = 2288,
  [15114] = 2265,
  [15115] = 15115,
  [15116] = 7430,
  [15117] = 15117,
  [15118] = 2264,
  [15119] = 15085,
  [15120] = 15120,
  [15121] = 15121,
  [15122] = 15122,
  [15123] = 15123,
  [15124] = 15124,
  [15125] = 15079,
  [15126] = 15126,
  [15127] = 15127,
  [15128] = 15128,
  [15129] = 7482,
  [15130] = 2286,
  [15131] = 2289,
  [15132] = 15121,
  [15133] = 15079,
  [15134] = 15122,
  [15135] = 15123,
  [15136] = 2286,
  [15137] = 15124,
  [15138] = 15081,
  [15139] = 15126,
  [15140] = 15127,
  [15141] = 2313,
  [15142] = 15121,
  [15143] = 15123,
  [15144] = 15122,
  [15145] = 15123,
  [15146] = 15124,
  [15147] = 15126,
  [15148] = 2264,
  [15149] = 15127,
  [15150] = 15150,
  [15151] = 15121,
  [15152] = 15122,
  [15153] = 15123,
  [15154] = 15126,
  [15155] = 15155,
  [15156] = 15124,
  [15157] = 15126,
  [15158] = 15127,
  [15159] = 15122,
  [15160] = 15123,
  [15161] = 15124,
  [15162] = 15127,
  [15163] = 2286,
  [15164] = 15122,
  [15165] = 15123,
  [15166] = 15079,
  [15167] = 15167,
  [15168] = 15124,
  [15169] = 15127,
  [15170] = 15079,
  [15171] = 7430,
  [15172] = 15122,
  [15173] = 15123,
  [15174] = 15076,
  [15175] = 15124,
  [15176] = 15121,
  [15177] = 15127,
  [15178] = 15122,
  [15179] = 9741,
  [15180] = 9673,
  [15181] = 15181,
  [15182] = 15127,
  [15183] = 9742,
  [15184] = 15184,
  [15185] = 9744,
  [15186] = 15122,
  [15187] = 2278,
  [15188] = 15127,
  [15189] = 15122,
  [15190] = 15085,
  [15191] = 15191,
  [15192] = 15127,
  [15193] = 2290,
  [15194] = 9574,
  [15195] = 9577,
  [15196] = 9578,
  [15197] = 15122,
  [15198] = 9589,
  [15199] = 9605,
  [15200] = 9751,
  [15201] = 9606,
  [15202] = 15127,
  [15203] = 15081,
  [15204] = 15122,
  [15205] = 15205,
  [15206] = 15206,
  [15207] = 15122,
  [15208] = 2286,
  [15209] = 2316,
  [15210] = 15122,
  [15211] = 9661,
  [15212] = 9667,
  [15213] = 2265,
  [15214] = 15122,
  [15215] = 15205,
  [15216] = 15122,
  [15217] = 15122,
  [15218] = 15122,
  [15219] = 15155,
  [15220] = 15122,
  [15221] = 2278,
  [15222] = 15122,
  [15223] = 15122,
  [15224] = 15076,
  [15225] = 15225,
  [15226] = 15122,
  [15227] = 15081,
  [15228] = 2265,
  [15229] = 15079,
  [15230] = 15122,
  [15231] = 15079,
  [15232] = 15122,
  [15233] = 15081,
  [15234] = 9820,
  [15235] = 15122,
  [15236] = 15155,
  [15237] = 15122,
  [15238] = 15206,
  [15239] = 9843,
  [15240] = 9848,
  [15241] = 9853,
  [15242] = 9689,
  [15243] = 15184,
  [15244] = 2264,
  [15245] = 15205,
  [15246] = 15079,
  [15247] = 15079,
  [15248] = 15076,
  [15249] = 15155,
  [15250] = 2286,
  [15251] = 15076,
  [15252] = 2281,
  [15253] = 9743,
  [15254] = 9745,
  [15255] = 15081,
  [15256] = 15079,
  [15257] = 15117,
  [15258] = 15120,
  [15259] = 2313,
  [15260] = 7489,
  [15261] = 2284,
  [15262] = 15205,
  [15263] = 15155,
  [15264] = 15076,
  [15265] = 15085,
  [15266] = 9786,
  [15267] = 15079,
  [15268] = 15081,
  [15269] = 15076,
  [15270] = 9694,
  [15271] = 15205,
  [15272] = 9752,
  [15273] = 15205,
  [15274] = 2313,
  [15275] = 15081,
  [15276] = 2373,
  [15277] = 15155,
  [15278] = 15122,
  [15279] = 7432,
  [15280] = 15076,
  [15281] = 15115,
  [15282] = 15155,
  [15283] = 15283,
  [15284] = 2277,
  [15285] = 15081,
  [15286] = 14126,
  [15287] = 9766,
  [15288] = 2276,
  [15289] = 2276,
  [15290] = 2316,
  [15291] = 2285,
  [15292] = 9868,
  [15293] = 2328,
  [15294] = 15081,
  [15295] = 15295,
  [15296] = 15074,
  [15297] = 15295,
  [15298] = 15074,
  [15299] = 15079,
  [15300] = 15295,
  [15301] = 2264,
  [15302] = 2282,
  [15303] = 15205,
  [15304] = 15295,
  [15305] = 15305,
  [15306] = 9717,
  [15307] = 2312,
  [15308] = 15295,
  [15309] = 15079,
  [15310] = 15085,
  [15311] = 15295,
  [15312] = 2281,
  [15313] = 7432,
  [15314] = 15295,
  [15315] = 15155,
  [15316] = 2282,
  [15317] = 15295,
  [15318] = 9835,
  [15319] = 15295,
  [15320] = 15295,
  [15321] = 15295,
  [15322] = 15205,
  [15323] = 15295,
  [15324] = 15295,
  [15325] = 2284,
  [15326] = 15076,
  [15327] = 15295,
  [15328] = 15295,
  [15329] = 9734,
  [15330] = 15295,
  [15331] = 15295,
  [15332] = 2301,
  [15333] = 15295,
  [15334] = 15295,
  [15335] = 2283,
  [15336] = 15295,
  [15337] = 15295,
  [15338] = 15295,
  [15339] = 15295,
  [15340] = 2277,
  [15341] = 15295,
  [15342] = 15295,
  [15343] = 15295,
  [15344] = 15295,
  [15345] = 15295,
  [15346] = 15295,
  [15347] = 15295,
  [15348] = 15295,
  [15349] = 15295,
  [15350] = 15076,
  [15351] = 15085,
  [15352] = 15127,
  [15353] = 15081,
  [15354] = 9754,
  [15355] = 15076,
  [15356] = 15124,
  [15357] = 15079,
  [15358] = 7489,
  [15359] = 15079,
  [15360] = 15360,
  [15361] = 15361,
  [15362] = 2282,
  [15363] = 15363,
  [15364] = 15363,
  [15365] = 15365,
  [15366] = 15366,
  [15367] = 15360,
  [15368] = 15368,
  [15369] = 15369,
  [15370] = 15360,
  [15371] = 15366,
  [15372] = 2328,
  [15373] = 14126,
  [15374] = 15360,
  [15375] = 15365,
  [15376] = 15368,
  [15377] = 15368,
  [15378] = 7482,
  [15379] = 15368,
  [15380] = 15365,
  [15381] = 15368,
  [15382] = 13854,
  [15383] = 2286,
  [15384] = 15365,
  [15385] = 2276,
  [15386] = 7430,
  [15387] = 15360,
  [15388] = 15363,
  [15389] = 15366,
  [15390] = 2285,
  [15391] = 15369,
  [15392] = 15392,
  [15393] = 15368,
  [15394] = 2287,
  [15395] = 2276,
  [15396] = 15363,
  [15397] = 15360,
  [15398] = 15365,
  [15399] = 2289,
  [15400] = 15369,
  [15401] = 15360,
  [15402] = 15360,
  [15403] = 15361,
  [15404] = 15365,
  [15405] = 15360,
  [15406] = 15360,
  [15407] = 15365,
  [15408] = 2287,
  [15409] = 15360,
  [15410] = 2284,
  [15411] = 2312,
  [15412] = 15368,
  [15413] = 15361,
  [15414] = 15360,
  [15415] = 15366,
  [15416] = 2285,
  [15417] = 15363,
  [15418] = 15360,
  [15419] = 2281,
  [15420] = 15360,
  [15421] = 15421,
  [15422] = 2278,
  [15423] = 2288,
  [15424] = 15421,
  [15425] = 2281,
  [15426] = 15363,
  [15427] = 2282,
  [15428] = 2289,
  [15429] = 2288,
  [15430] = 15365,
  [15431] = 7432,
  [15432] = 2282,
  [15433] = 15366,
  [15434] = 2290,
  [15435] = 15366,
  [15436] = 15365,
  [15437] = 15363,
  [15438] = 2301,
  [15439] = 2287,
  [15440] = 15369,
  [15441] = 15441,
  [15442] = 2277,
  [15443] = 15365,
  [15444] = 15363,
  [15445] = 15368,
  [15446] = 15363,
  [15447] = 2264,
  [15448] = 15360,
  [15449] = 15421,
  [15450] = 15360,
  [15451] = 15366,
  [15452] = 2316,
  [15453] = 15366,
  [15454] = 2312,
  [15455] = 15360,
  [15456] = 2369,
  [15457] = 2278,
  [15458] = 15365,
  [15459] = 15360,
  [15460] = 15366,
  [15461] = 2283,
  [15462] = 15421,
  [15463] = 15421,
  [15464] = 2284,
  [15465] = 15363,
  [15466] = 15360,
  [15467] = 2312,
  [15468] = 2301,
  [15469] = 15363,
  [15470] = 15369,
  [15471] = 15366,
  [15472] = 15360,
  [15473] = 2383,
  [15474] = 15361,
  [15475] = 2265,
  [15476] = 15421,
  [15477] = 2316,
  [15478] = 15478,
  [15479] = 2290,
  [15480] = 15363,
  [15481] = 2316,
  [15482] = 15361,
  [15483] = 15421,
  [15484] = 2283,
  [15485] = 2285,
  [15486] = 15369,
  [15487] = 14674,
  [15488] = 15360,
  [15489] = 15361,
  [15490] = 2278,
  [15491] = 15491,
  [15492] = 15360,
  [15493] = 15361,
  [15494] = 15366,
  [15495] = 15361,
  [15496] = 2383,
  [15497] = 7465,
  [15498] = 2276,
  [15499] = 2301,
  [15500] = 15360,
  [15501] = 2281,
  [15502] = 15363,
  [15503] = 2282,
  [15504] = 7489,
  [15505] = 15365,
  [15506] = 2285,
  [15507] = 15360,
  [15508] = 2289,
  [15509] = 15366,
  [15510] = 2288,
  [15511] = 15369,
  [15512] = 15368,
  [15513] = 2284,
  [15514] = 2276,
  [15515] = 2288,
  [15516] = 15421,
  [15517] = 15368,
  [15518] = 15518,
  [15519] = 15360,
  [15520] = 15368,
  [15521] = 15360,
  [15522] = 15361,
  [15523] = 15361,
  [15524] = 15366,
  [15525] = 2278,
  [15526] = 2281,
  [15527] = 15365,
  [15528] = 15421,
  [15529] = 2353,
  [15530] = 2289,
  [15531] = 15368,
  [15532] = 2284,
  [15533] = 15421,
  [15534] = 15369,
  [15535] = 2290,
  [15536] = 2283,
  [15537] = 15368,
  [15538] = 2287,
  [15539] = 15361,
  [15540] = 15421,
  [15541] = 2373,
  [15542] = 2369,
  [15543] = 15361,
  [15544] = 15361,
  [15545] = 2290,
  [15546] = 2312,
  [15547] = 2278,
  [15548] = 2301,
  [15549] = 2353,
  [15550] = 2316,
  [15551] = 15365,
  [15552] = 2283,
  [15553] = 15369,
  [15554] = 15369,
  [15555] = 15365,
  [15556] = 15360,
  [15557] = 15557,
  [15558] = 15558,
  [15559] = 15559,
  [15560] = 15560,
  [15561] = 15561,
  [15562] = 15558,
  [15563] = 15558,
  [15564] = 2312,
  [15565] = 15558,
  [15566] = 15561,
  [15567] = 15558,
  [15568] = 15557,
  [15569] = 15569,
  [15570] = 15570,
  [15571] = 15569,
  [15572] = 15558,
  [15573] = 15569,
  [15574] = 2301,
  [15575] = 15569,
  [15576] = 15558,
  [15577] = 2312,
  [15578] = 15558,
  [15579] = 15569,
  [15580] = 15560,
  [15581] = 15569,
  [15582] = 15569,
  [15583] = 14674,
  [15584] = 7888,
  [15585] = 15558,
  [15586] = 7930,
  [15587] = 15569,
  [15588] = 7933,
  [15589] = 15558,
  [15590] = 15569,
  [15591] = 15569,
  [15592] = 15558,
  [15593] = 2287,
  [15594] = 15560,
  [15595] = 15561,
  [15596] = 15569,
  [15597] = 15569,
  [15598] = 15561,
  [15599] = 15569,
  [15600] = 15557,
  [15601] = 15558,
  [15602] = 7893,
  [15603] = 15560,
  [15604] = 15570,
  [15605] = 15558,
  [15606] = 15569,
  [15607] = 15569,
  [15608] = 2288,
  [15609] = 7874,
  [15610] = 15560,
  [15611] = 15558,
  [15612] = 2301,
  [15613] = 15558,
  [15614] = 2289,
  [15615] = 15561,
  [15616] = 15558,
  [15617] = 15570,
  [15618] = 13854,
  [15619] = 15569,
  [15620] = 15558,
  [15621] = 15621,
  [15622] = 15622,
  [15623] = 15557,
  [15624] = 15569,
  [15625] = 15570,
  [15626] = 15569,
  [15627] = 2369,
  [15628] = 15569,
  [15629] = 2316,
  [15630] = 7893,
  [15631] = 15560,
  [15632] = 2290,
  [15633] = 15633,
  [15634] = 15558,
  [15635] = 15635,
  [15636] = 15560,
  [15637] = 15569,
  [15638] = 15557,
  [15639] = 7930,
  [15640] = 15570,
  [15641] = 15569,
  [15642] = 15558,
  [15643] = 15569,
  [15644] = 15569,
  [15645] = 15558,
  [15646] = 15557,
  [15647] = 15569,
  [15648] = 15557,
  [15649] = 7933,
  [15650] = 15569,
  [15651] = 15558,
  [15652] = 15561,
  [15653] = 15569,
  [15654] = 15570,
  [15655] = 15557,
  [15656] = 15558,
  [15657] = 15569,
  [15658] = 15558,
  [15659] = 15558,
  [15660] = 15558,
  [15661] = 15560,
  [15662] = 15561,
  [15663] = 2283,
  [15664] = 15569,
  [15665] = 15570,
  [15666] = 2278,
  [15667] = 15569,
  [15668] = 2276,
  [15669] = 15558,
  [15670] = 15558,
  [15671] = 15671,
  [15672] = 15560,
  [15673] = 15561,
  [15674] = 2281,
  [15675] = 15558,
  [15676] = 2282,
  [15677] = 15558,
  [15678] = 15560,
  [15679] = 15569,
  [15680] = 15518,
  [15681] = 15569,
  [15682] = 15569,
  [15683] = 15557,
  [15684] = 15558,
  [15685] = 2284,
  [15686] = 15558,
  [15687] = 15558,
  [15688] = 15569,
  [15689] = 15570,
  [15690] = 15558,
  [15691] = 15478,
  [15692] = 15569,
  [15693] = 15557,
  [15694] = 15558,
  [15695] = 15557,
  [15696] = 15561,
  [15697] = 2383,
  [15698] = 15569,
  [15699] = 15557,
  [15700] = 15558,
  [15701] = 15560,
  [15702] = 15569,
  [15703] = 15569,
  [15704] = 15569,
  [15705] = 15560,
  [15706] = 2283,
  [15707] = 2353,
  [15708] = 15558,
  [15709] = 15569,
  [15710] = 15558,
  [15711] = 15558,
  [15712] = 15570,
  [15713] = 15558,
  [15714] = 15714,
  [15715] = 15558,
  [15716] = 15569,
  [15717] = 15558,
  [15718] = 15558,
  [15719] = 2285,
  [15720] = 15560,
  [15721] = 15558,
  [15722] = 15569,
  [15723] = 15557,
  [15724] = 15558,
  [15725] = 15569,
  [15726] = 15557,
  [15727] = 15558,
  [15728] = 15569,
  [15729] = 15569,
  [15730] = 15560,
  [15731] = 15569,
  [15732] = 15558,
  [15733] = 7888,
  [15734] = 7874,
  [15735] = 15557,
  [15736] = 15569,
  [15737] = 15569,
  [15738] = 15558,
  [15739] = 15560,
  [15740] = 15558,
  [15741] = 15569,
  [15742] = 15558,
  [15743] = 15569,
  [15744] = 15744,
  [15745] = 15745,
  [15746] = 15746,
  [15747] = 15747,
  [15748] = 15748,
  [15749] = 15749,
  [15750] = 15749,
  [15751] = 15751,
  [15752] = 15752,
  [15753] = 15744,
  [15754] = 15754,
  [15755] = 15745,
  [15756] = 15746,
  [15757] = 15747,
  [15758] = 15748,
  [15759] = 15749,
  [15760] = 15751,
  [15761] = 15752,
  [15762] = 15744,
  [15763] = 15751,
  [15764] = 15745,
  [15765] = 15746,
  [15766] = 15747,
  [15767] = 15748,
  [15768] = 15749,
  [15769] = 15751,
  [15770] = 15752,
  [15771] = 15744,
  [15772] = 15745,
  [15773] = 15746,
  [15774] = 15747,
  [15775] = 15748,
  [15776] = 15749,
  [15777] = 15751,
  [15778] = 15752,
  [15779] = 15744,
  [15780] = 15780,
  [15781] = 15751,
  [15782] = 15745,
  [15783] = 15746,
  [15784] = 15747,
  [15785] = 15748,
  [15786] = 15749,
  [15787] = 15749,
  [15788] = 15751,
  [15789] = 15752,
  [15790] = 15744,
  [15791] = 15745,
  [15792] = 15746,
  [15793] = 15745,
  [15794] = 15746,
  [15795] = 15795,
  [15796] = 15752,
  [15797] = 15747,
  [15798] = 15748,
  [15799] = 15749,
  [15800] = 15751,
  [15801] = 15752,
  [15802] = 15744,
  [15803] = 15747,
  [15804] = 15748,
  [15805] = 15745,
  [15806] = 15746,
  [15807] = 15749,
  [15808] = 15747,
  [15809] = 15748,
  [15810] = 15749,
  [15811] = 15751,
  [15812] = 15752,
  [15813] = 15744,
  [15814] = 15814,
  [15815] = 15745,
  [15816] = 15746,
  [15817] = 15817,
  [15818] = 15747,
  [15819] = 15748,
  [15820] = 15749,
  [15821] = 15751,
  [15822] = 15752,
  [15823] = 15744,
  [15824] = 15751,
  [15825] = 15745,
  [15826] = 15746,
  [15827] = 15752,
  [15828] = 15817,
  [15829] = 15747,
  [15830] = 15748,
  [15831] = 15749,
  [15832] = 15751,
  [15833] = 15752,
  [15834] = 15744,
  [15835] = 15835,
  [15836] = 15745,
  [15837] = 15746,
  [15838] = 15747,
  [15839] = 15748,
  [15840] = 15749,
  [15841] = 15817,
  [15842] = 15744,
  [15843] = 15751,
  [15844] = 15752,
  [15845] = 15744,
  [15846] = 15744,
  [15847] = 15745,
  [15848] = 15746,
  [15849] = 15744,
  [15850] = 15747,
  [15851] = 15748,
  [15852] = 15749,
  [15853] = 9541,
  [15854] = 15751,
  [15855] = 15752,
  [15856] = 15744,
  [15857] = 15857,
  [15858] = 15745,
  [15859] = 15746,
  [15860] = 7874,
  [15861] = 15861,
  [15862] = 15747,
  [15863] = 15748,
  [15864] = 15749,
  [15865] = 15751,
  [15866] = 15752,
  [15867] = 15744,
  [15868] = 15752,
  [15869] = 15744,
  [15870] = 15857,
  [15871] = 15752,
  [15872] = 15744,
  [15873] = 15873,
  [15874] = 15752,
  [15875] = 15744,
  [15876] = 15752,
  [15877] = 15744,
  [15878] = 15861,
  [15879] = 15752,
  [15880] = 15744,
  [15881] = 15752,
  [15882] = 15744,
  [15883] = 15752,
  [15884] = 15744,
  [15885] = 15752,
  [15886] = 15744,
  [15887] = 15752,
  [15888] = 15744,
  [15889] = 15780,
  [15890] = 15752,
  [15891] = 15744,
  [15892] = 15752,
  [15893] = 15744,
  [15894] = 15745,
  [15895] = 15746,
  [15896] = 15752,
  [15897] = 15744,
  [15898] = 15752,
  [15899] = 15744,
  [15900] = 15752,
  [15901] = 15744,
  [15902] = 15780,
  [15903] = 15752,
  [15904] = 15744,
  [15905] = 15905,
  [15906] = 15857,
  [15907] = 15752,
  [15908] = 15744,
  [15909] = 15747,
  [15910] = 15748,
  [15911] = 15752,
  [15912] = 15744,
  [15913] = 15749,
  [15914] = 15752,
  [15915] = 15744,
  [15916] = 15752,
  [15917] = 15744,
  [15918] = 15752,
  [15919] = 15744,
  [15920] = 15751,
  [15921] = 15752,
  [15922] = 15752,
  [15923] = 15744,
  [15924] = 15752,
  [15925] = 15744,
  [15926] = 15752,
  [15927] = 15744,
  [15928] = 15861,
  [15929] = 15752,
  [15930] = 15744,
  [15931] = 15817,
  [15932] = 15744,
  [15933] = 15752,
  [15934] = 15744,
  [15935] = 15752,
  [15936] = 15744,
  [15937] = 15857,
  [15938] = 15752,
  [15939] = 15744,
  [15940] = 15752,
  [15941] = 15744,
  [15942] = 15752,
  [15943] = 15744,
  [15944] = 15861,
  [15945] = 15780,
  [15946] = 15752,
  [15947] = 15744,
  [15948] = 15752,
  [15949] = 15744,
  [15950] = 15752,
  [15951] = 15744,
  [15952] = 15752,
  [15953] = 15744,
  [15954] = 15780,
  [15955] = 15752,
  [15956] = 15744,
  [15957] = 15752,
  [15958] = 15744,
  [15959] = 15959,
  [15960] = 15752,
  [15961] = 15744,
  [15962] = 15752,
  [15963] = 15744,
  [15964] = 15752,
  [15965] = 15744,
  [15966] = 15780,
  [15967] = 15752,
  [15968] = 15744,
  [15969] = 15745,
  [15970] = 15970,
  [15971] = 15746,
  [15972] = 15752,
  [15973] = 15744,
  [15974] = 15752,
  [15975] = 15744,
  [15976] = 15745,
  [15977] = 15746,
  [15978] = 15752,
  [15979] = 15744,
  [15980] = 15752,
  [15981] = 15744,
  [15982] = 15747,
  [15983] = 15748,
  [15984] = 15752,
  [15985] = 15744,
  [15986] = 15749,
  [15987] = 15752,
  [15988] = 15744,
  [15989] = 15747,
  [15990] = 15748,
  [15991] = 15752,
  [15992] = 15752,
  [15993] = 15744,
  [15994] = 15751,
  [15995] = 15752,
  [15996] = 15744,
  [15997] = 15752,
  [15998] = 15749,
  [15999] = 15752,
  [16000] = 15744,
  [16001] = 15752,
  [16002] = 15744,
  [16003] = 15817,
  [16004] = 15744,
  [16005] = 15752,
  [16006] = 15744,
  [16007] = 15752,
  [16008] = 15744,
  [16009] = 15752,
  [16010] = 15744,
  [16011] = 15752,
  [16012] = 15744,
  [16013] = 15752,
  [16014] = 15744,
  [16015] = 15752,
  [16016] = 15744,
  [16017] = 15857,
  [16018] = 15861,
  [16019] = 15751,
  [16020] = 15780,
  [16021] = 15795,
  [16022] = 15752,
  [16023] = 16023,
  [16024] = 15745,
  [16025] = 15746,
  [16026] = 15747,
  [16027] = 15748,
  [16028] = 15749,
  [16029] = 15749,
  [16030] = 8029,
  [16031] = 8031,
  [16032] = 15751,
  [16033] = 15752,
  [16034] = 15817,
  [16035] = 15744,
  [16036] = 15744,
  [16037] = 15817,
  [16038] = 15780,
  [16039] = 15745,
  [16040] = 15746,
  [16041] = 15780,
  [16042] = 15857,
  [16043] = 15747,
  [16044] = 15748,
  [16045] = 15861,
  [16046] = 15749,
  [16047] = 15959,
  [16048] = 15751,
  [16049] = 15752,
  [16050] = 15744,
  [16051] = 15780,
  [16052] = 15745,
  [16053] = 15746,
  [16054] = 16054,
  [16055] = 15857,
  [16056] = 15747,
  [16057] = 15748,
  [16058] = 15749,
  [16059] = 15745,
  [16060] = 15751,
  [16061] = 15752,
  [16062] = 15746,
  [16063] = 15744,
  [16064] = 15780,
  [16065] = 15780,
  [16066] = 15780,
  [16067] = 15745,
  [16068] = 15746,
  [16069] = 15744,
  [16070] = 15751,
  [16071] = 15780,
  [16072] = 15747,
  [16073] = 15748,
  [16074] = 2373,
  [16075] = 15749,
  [16076] = 15747,
  [16077] = 15780,
  [16078] = 15748,
  [16079] = 15959,
  [16080] = 15751,
  [16081] = 2328,
  [16082] = 15752,
  [16083] = 15744,
  [16084] = 15959,
  [16085] = 15749,
  [16086] = 15745,
  [16087] = 15746,
  [16088] = 15780,
  [16089] = 7888,
  [16090] = 15745,
  [16091] = 15746,
  [16092] = 15747,
  [16093] = 15748,
  [16094] = 7893,
  [16095] = 15749,
  [16096] = 15747,
  [16097] = 15748,
  [16098] = 15749,
  [16099] = 15751,
  [16100] = 15752,
  [16101] = 15744,
  [16102] = 15780,
  [16103] = 15751,
  [16104] = 15745,
  [16105] = 15746,
  [16106] = 15751,
  [16107] = 15795,
  [16108] = 15752,
  [16109] = 15747,
  [16110] = 15748,
  [16111] = 15749,
  [16112] = 15780,
  [16113] = 15751,
  [16114] = 15752,
  [16115] = 15861,
  [16116] = 16116,
  [16117] = 15780,
  [16118] = 16118,
  [16119] = 15744,
  [16120] = 15795,
  [16121] = 15752,
  [16122] = 15817,
  [16123] = 15780,
  [16124] = 16124,
  [16125] = 16125,
  [16126] = 15744,
  [16127] = 15745,
  [16128] = 15746,
  [16129] = 8029,
  [16130] = 15745,
  [16131] = 15745,
  [16132] = 15747,
  [16133] = 15748,
  [16134] = 15749,
  [16135] = 8031,
  [16136] = 2537,
  [16137] = 2539,
  [16138] = 15780,
  [16139] = 2485,
  [16140] = 2544,
  [16141] = 2545,
  [16142] = 15751,
  [16143] = 15752,
  [16144] = 2547,
  [16145] = 2550,
  [16146] = 2551,
  [16147] = 2552,
  [16148] = 2555,
  [16149] = 2556,
  [16150] = 2558,
  [16151] = 2560,
  [16152] = 2561,
  [16153] = 2562,
  [16154] = 2563,
  [16155] = 2503,
  [16156] = 2504,
  [16157] = 2525,
  [16158] = 2530,
  [16159] = 2535,
  [16160] = 2536,
  [16161] = 2540,
  [16162] = 15744,
  [16163] = 15857,
  [16164] = 15780,
  [16165] = 2542,
  [16166] = 2572,
  [16167] = 2573,
  [16168] = 2574,
  [16169] = 15780,
  [16170] = 2575,
  [16171] = 15861,
  [16172] = 15746,
  [16173] = 15745,
  [16174] = 15746,
  [16175] = 15795,
  [16176] = 16176,
  [16177] = 15747,
  [16178] = 15748,
  [16179] = 15749,
  [16180] = 15752,
  [16181] = 2537,
  [16182] = 2539,
  [16183] = 15751,
  [16184] = 2485,
  [16185] = 2544,
  [16186] = 15752,
  [16187] = 2545,
  [16188] = 2547,
  [16189] = 2550,
  [16190] = 15744,
  [16191] = 2551,
  [16192] = 2552,
  [16193] = 2555,
  [16194] = 2556,
  [16195] = 2558,
  [16196] = 2560,
  [16197] = 2561,
  [16198] = 2562,
  [16199] = 2563,
  [16200] = 2503,
  [16201] = 15817,
  [16202] = 2504,
  [16203] = 2525,
  [16204] = 2530,
  [16205] = 2535,
  [16206] = 2536,
  [16207] = 15780,
  [16208] = 2540,
  [16209] = 15795,
  [16210] = 2542,
  [16211] = 15745,
  [16212] = 2572,
  [16213] = 15746,
  [16214] = 2573,
  [16215] = 2574,
  [16216] = 15747,
  [16217] = 15748,
  [16218] = 2575,
  [16219] = 15749,
  [16220] = 15744,
  [16221] = 15817,
  [16222] = 15751,
  [16223] = 15752,
  [16224] = 15752,
  [16225] = 15857,
  [16226] = 15744,
  [16227] = 15780,
  [16228] = 2582,
  [16229] = 2583,
  [16230] = 15747,
  [16231] = 15748,
  [16232] = 15745,
  [16233] = 15746,
  [16234] = 16234,
  [16235] = 15857,
  [16236] = 16236,
  [16237] = 15747,
  [16238] = 15748,
  [16239] = 15780,
  [16240] = 15749,
  [16241] = 15959,
  [16242] = 15751,
  [16243] = 15752,
  [16244] = 2588,
  [16245] = 2589,
  [16246] = 15744,
  [16247] = 15745,
  [16248] = 15746,
  [16249] = 15780,
  [16250] = 7930,
  [16251] = 15861,
  [16252] = 15745,
  [16253] = 15746,
  [16254] = 15747,
  [16255] = 15747,
  [16256] = 15748,
  [16257] = 15749,
  [16258] = 15748,
  [16259] = 15751,
  [16260] = 15752,
  [16261] = 15749,
  [16262] = 15959,
  [16263] = 7933,
  [16264] = 15744,
  [16265] = 2582,
  [16266] = 2583,
  [16267] = 15749,
  [16268] = 15780,
  [16269] = 15905,
  [16270] = 2588,
  [16271] = 2589,
  [16272] = 15745,
  [16273] = 15746,
  [16274] = 15817,
  [16275] = 15861,
  [16276] = 15751,
  [16277] = 15747,
  [16278] = 15748,
  [16279] = 15752,
  [16280] = 15749,
  [16281] = 15905,
  [16282] = 15751,
  [16283] = 15752,
  [16284] = 15959,
  [16285] = 15744,
  [16286] = 15905,
  [16287] = 15905,
  [16288] = 15745,
  [16289] = 15745,
  [16290] = 15746,
  [16291] = 15746,
  [16292] = 15817,
  [16293] = 15747,
  [16294] = 15748,
  [16295] = 15905,
  [16296] = 15749,
  [16297] = 15747,
  [16298] = 15748,
  [16299] = 15744,
  [16300] = 15905,
  [16301] = 15751,
  [16302] = 15749,
  [16303] = 15752,
  [16304] = 15780,
  [16305] = 15905,
  [16306] = 15744,
  [16307] = 15905,
  [16308] = 15857,
  [16309] = 15861,
  [16310] = 15745,
  [16311] = 15751,
  [16312] = 15746,
  [16313] = 15905,
  [16314] = 15861,
  [16315] = 15795,
  [16316] = 15752,
  [16317] = 15747,
  [16318] = 15905,
  [16319] = 15748,
  [16320] = 15749,
  [16321] = 15905,
  [16322] = 15751,
  [16323] = 16054,
  [16324] = 15747,
  [16325] = 15905,
  [16326] = 15752,
  [16327] = 15744,
  [16328] = 15905,
  [16329] = 15817,
  [16330] = 15746,
  [16331] = 15744,
  [16332] = 15905,
  [16333] = 15744,
  [16334] = 16334,
  [16335] = 15745,
  [16336] = 15746,
  [16337] = 15857,
  [16338] = 15747,
  [16339] = 15748,
  [16340] = 15861,
  [16341] = 15749,
  [16342] = 15751,
  [16343] = 15752,
  [16344] = 15744,
  [16345] = 15745,
  [16346] = 15746,
  [16347] = 15747,
  [16348] = 15748,
  [16349] = 15749,
  [16350] = 15751,
  [16351] = 15752,
  [16352] = 15744,
  [16353] = 15745,
  [16354] = 15746,
  [16355] = 15780,
  [16356] = 15747,
  [16357] = 15748,
  [16358] = 15749,
  [16359] = 16334,
  [16360] = 15751,
  [16361] = 15752,
  [16362] = 15744,
  [16363] = 15745,
  [16364] = 15745,
  [16365] = 15746,
  [16366] = 15746,
  [16367] = 15747,
  [16368] = 15748,
  [16369] = 15749,
  [16370] = 15780,
  [16371] = 15751,
  [16372] = 15752,
  [16373] = 15744,
  [16374] = 15959,
  [16375] = 15747,
  [16376] = 15748,
  [16377] = 15745,
  [16378] = 15746,
  [16379] = 15751,
  [16380] = 15749,
  [16381] = 15747,
  [16382] = 15748,
  [16383] = 15749,
  [16384] = 15751,
  [16385] = 15752,
  [16386] = 15744,
  [16387] = 15745,
  [16388] = 15746,
  [16389] = 15747,
  [16390] = 15748,
  [16391] = 15749,
  [16392] = 15751,
  [16393] = 15752,
  [16394] = 15744,
  [16395] = 15748,
  [16396] = 15751,
  [16397] = 15752,
  [16398] = 15745,
  [16399] = 15746,
  [16400] = 15745,
  [16401] = 15747,
  [16402] = 15748,
  [16403] = 15749,
  [16404] = 15746,
  [16405] = 15751,
  [16406] = 15752,
  [16407] = 15780,
  [16408] = 15744,
  [16409] = 15795,
  [16410] = 15745,
  [16411] = 15746,
  [16412] = 15752,
  [16413] = 15747,
  [16414] = 15748,
  [16415] = 15749,
  [16416] = 15817,
  [16417] = 15751,
  [16418] = 15752,
  [16419] = 15744,
  [16420] = 15744,
  [16421] = 15745,
  [16422] = 15746,
  [16423] = 15747,
  [16424] = 15748,
  [16425] = 15749,
  [16426] = 15780,
  [16427] = 15751,
  [16428] = 15752,
  [16429] = 15744,
  [16430] = 15857,
  [16431] = 15745,
  [16432] = 15746,
  [16433] = 15747,
  [16434] = 15748,
  [16435] = 15861,
  [16436] = 15747,
  [16437] = 15748,
  [16438] = 15749,
  [16439] = 15959,
  [16440] = 15745,
  [16441] = 15746,
  [16442] = 15751,
  [16443] = 15752,
  [16444] = 15744,
  [16445] = 15747,
  [16446] = 15748,
  [16447] = 16447,
  [16448] = 16448,
  [16449] = 8029,
  [16450] = 16450,
  [16451] = 16451,
  [16452] = 8291,
  [16453] = 16453,
  [16454] = 16448,
  [16455] = 2370,
  [16456] = 16451,
  [16457] = 8031,
  [16458] = 2537,
  [16459] = 2539,
  [16460] = 16460,
  [16461] = 2485,
  [16462] = 8110,
  [16463] = 2544,
  [16464] = 8112,
  [16465] = 2545,
  [16466] = 16466,
  [16467] = 16467,
  [16468] = 8740,
  [16469] = 2547,
  [16470] = 8119,
  [16471] = 2550,
  [16472] = 16451,
  [16473] = 2551,
  [16474] = 6557,
  [16475] = 2552,
  [16476] = 2555,
  [16477] = 2556,
  [16478] = 2558,
  [16479] = 2560,
  [16480] = 2561,
  [16481] = 2562,
  [16482] = 2563,
  [16483] = 2503,
  [16484] = 2504,
  [16485] = 2525,
  [16486] = 2530,
  [16487] = 2535,
  [16488] = 2536,
  [16489] = 2540,
  [16490] = 8199,
  [16491] = 8120,
  [16492] = 16492,
  [16493] = 16450,
  [16494] = 16494,
  [16495] = 2542,
  [16496] = 16496,
  [16497] = 8201,
  [16498] = 2572,
  [16499] = 16499,
  [16500] = 16500,
  [16501] = 2573,
  [16502] = 2574,
  [16503] = 16492,
  [16504] = 16451,
  [16505] = 16505,
  [16506] = 16451,
  [16507] = 16447,
  [16508] = 8127,
  [16509] = 16450,
  [16510] = 16510,
  [16511] = 2575,
  [16512] = 8239,
  [16513] = 16448,
  [16514] = 16447,
  [16515] = 8095,
  [16516] = 16448,
  [16517] = 16505,
  [16518] = 16448,
  [16519] = 16494,
  [16520] = 16496,
  [16521] = 16500,
  [16522] = 16451,
  [16523] = 16505,
  [16524] = 8733,
  [16525] = 8117,
  [16526] = 16510,
  [16527] = 8239,
  [16528] = 16494,
  [16529] = 16496,
  [16530] = 8240,
  [16531] = 16500,
  [16532] = 16467,
  [16533] = 8262,
  [16534] = 8096,
  [16535] = 8240,
  [16536] = 8291,
  [16537] = 16451,
  [16538] = 16496,
  [16539] = 16494,
  [16540] = 16492,
  [16541] = 16467,
  [16542] = 16496,
  [16543] = 16500,
  [16544] = 8262,
  [16545] = 8096,
  [16546] = 8128,
  [16547] = 16450,
  [16548] = 16451,
  [16549] = 16549,
  [16550] = 16453,
  [16551] = 16494,
  [16552] = 16552,
  [16553] = 16496,
  [16554] = 16460,
  [16555] = 16500,
  [16556] = 9595,
  [16557] = 16451,
  [16558] = 16453,
  [16559] = 16505,
  [16560] = 16451,
  [16561] = 16494,
  [16562] = 16450,
  [16563] = 16496,
  [16564] = 16500,
  [16565] = 16505,
  [16566] = 9564,
  [16567] = 8095,
  [16568] = 8148,
  [16569] = 8159,
  [16570] = 16494,
  [16571] = 16496,
  [16572] = 16572,
  [16573] = 16500,
  [16574] = 16492,
  [16575] = 8155,
  [16576] = 16494,
  [16577] = 16496,
  [16578] = 16500,
  [16579] = 16467,
  [16580] = 16510,
  [16581] = 16447,
  [16582] = 16451,
  [16583] = 16549,
  [16584] = 16453,
  [16585] = 16451,
  [16586] = 16494,
  [16587] = 16496,
  [16588] = 16588,
  [16589] = 16500,
  [16590] = 8148,
  [16591] = 16492,
  [16592] = 16460,
  [16593] = 8119,
  [16594] = 16451,
  [16595] = 16505,
  [16596] = 6559,
  [16597] = 16494,
  [16598] = 16598,
  [16599] = 16496,
  [16600] = 16448,
  [16601] = 16500,
  [16602] = 16602,
  [16603] = 8128,
  [16604] = 8288,
  [16605] = 16494,
  [16606] = 16496,
  [16607] = 16500,
  [16608] = 6557,
  [16609] = 16447,
  [16610] = 8166,
  [16611] = 16451,
  [16612] = 16612,
  [16613] = 16494,
  [16614] = 16496,
  [16615] = 16500,
  [16616] = 8120,
  [16617] = 16448,
  [16618] = 16549,
  [16619] = 16494,
  [16620] = 16453,
  [16621] = 16496,
  [16622] = 16500,
  [16623] = 16447,
  [16624] = 16624,
  [16625] = 8179,
  [16626] = 16451,
  [16627] = 16505,
  [16628] = 16450,
  [16629] = 8238,
  [16630] = 16460,
  [16631] = 16510,
  [16632] = 2369,
  [16633] = 16451,
  [16634] = 16447,
  [16635] = 16549,
  [16636] = 16451,
  [16637] = 16447,
  [16638] = 8146,
  [16639] = 16505,
  [16640] = 8155,
  [16641] = 8166,
  [16642] = 16451,
  [16643] = 16643,
  [16644] = 16447,
  [16645] = 16492,
  [16646] = 16450,
  [16647] = 16492,
  [16648] = 16448,
  [16649] = 16451,
  [16650] = 16450,
  [16651] = 2582,
  [16652] = 16652,
  [16653] = 2583,
  [16654] = 8263,
  [16655] = 13792,
  [16656] = 16549,
  [16657] = 16549,
  [16658] = 16510,
  [16659] = 16453,
  [16660] = 16453,
  [16661] = 8199,
  [16662] = 8201,
  [16663] = 16451,
  [16664] = 8263,
  [16665] = 16460,
  [16666] = 8179,
  [16667] = 16460,
  [16668] = 13792,
  [16669] = 16669,
  [16670] = 16492,
  [16671] = 2383,
  [16672] = 16492,
  [16673] = 16451,
  [16674] = 16451,
  [16675] = 16675,
  [16676] = 8127,
  [16677] = 6559,
  [16678] = 16467,
  [16679] = 8110,
  [16680] = 16451,
  [16681] = 16549,
  [16682] = 16453,
  [16683] = 16460,
  [16684] = 8281,
  [16685] = 16451,
  [16686] = 16451,
  [16687] = 2353,
  [16688] = 16451,
  [16689] = 16448,
  [16690] = 16450,
  [16691] = 16450,
  [16692] = 16451,
  [16693] = 16451,
  [16694] = 16694,
  [16695] = 16695,
  [16696] = 16451,
  [16697] = 16450,
  [16698] = 16467,
  [16699] = 16699,
  [16700] = 16492,
  [16701] = 16451,
  [16702] = 2379,
  [16703] = 16448,
  [16704] = 16451,
  [16705] = 16450,
  [16706] = 8288,
  [16707] = 16450,
  [16708] = 16451,
  [16709] = 16451,
  [16710] = 16710,
  [16711] = 16549,
  [16712] = 16453,
  [16713] = 16451,
  [16714] = 16714,
  [16715] = 16447,
  [16716] = 16695,
  [16717] = 16710,
  [16718] = 16448,
  [16719] = 16505,
  [16720] = 2366,
  [16721] = 16494,
  [16722] = 16510,
  [16723] = 16695,
  [16724] = 16710,
  [16725] = 16549,
  [16726] = 16460,
  [16727] = 16695,
  [16728] = 16710,
  [16729] = 16453,
  [16730] = 16451,
  [16731] = 16450,
  [16732] = 16695,
  [16733] = 16710,
  [16734] = 8233,
  [16735] = 16695,
  [16736] = 16710,
  [16737] = 8178,
  [16738] = 8234,
  [16739] = 2588,
  [16740] = 8238,
  [16741] = 8117,
  [16742] = 2364,
  [16743] = 2589,
  [16744] = 16460,
  [16745] = 8149,
  [16746] = 16746,
  [16747] = 8252,
  [16748] = 6560,
  [16749] = 8159,
  [16750] = 16549,
  [16751] = 16447,
  [16752] = 8130,
  [16753] = 16492,
  [16754] = 16505,
  [16755] = 16460,
  [16756] = 16510,
  [16757] = 16448,
  [16758] = 16451,
  [16759] = 8252,
  [16760] = 8233,
  [16761] = 16549,
  [16762] = 16492,
  [16763] = 8281,
  [16764] = 16549,
  [16765] = 16460,
  [16766] = 8178,
  [16767] = 16447,
  [16768] = 16467,
  [16769] = 16505,
  [16770] = 16770,
  [16771] = 8149,
  [16772] = 16451,
  [16773] = 16451,
  [16774] = 16447,
  [16775] = 16447,
  [16776] = 16510,
  [16777] = 16500,
  [16778] = 8234,
  [16779] = 16467,
  [16780] = 6560,
  [16781] = 8112,
  [16782] = 2368,
  [16783] = 8130,
  [16784] = 9613,
  [16785] = 16505,
  [16786] = 16460,
  [16787] = 8146,
  [16788] = 16788,
  [16789] = 16789,
  [16790] = 16790,
  [16791] = 6560,
  [16792] = 16792,
  [16793] = 16793,
  [16794] = 16794,
  [16795] = 16795,
  [16796] = 16788,
  [16797] = 16797,
  [16798] = 6625,
  [16799] = 16799,
  [16800] = 16800,
  [16801] = 16800,
  [16802] = 16802,
  [16803] = 16803,
  [16804] = 16804,
  [16805] = 16805,
  [16806] = 16806,
  [16807] = 16807,
  [16808] = 16808,
  [16809] = 16808,
  [16810] = 16810,
  [16811] = 16811,
  [16812] = 16800,
  [16813] = 2448,
  [16814] = 16814,
  [16815] = 16815,
  [16816] = 16816,
  [16817] = 16799,
  [16818] = 16818,
  [16819] = 16800,
  [16820] = 16799,
  [16821] = 16789,
  [16822] = 16814,
  [16823] = 16792,
  [16824] = 16793,
  [16825] = 16800,
  [16826] = 16794,
  [16827] = 16827,
  [16828] = 2442,
  [16829] = 16799,
  [16830] = 16788,
  [16831] = 16831,
  [16832] = 16800,
  [16833] = 16833,
  [16834] = 16800,
  [16835] = 16835,
  [16836] = 16802,
  [16837] = 8195,
  [16838] = 16804,
  [16839] = 16839,
  [16840] = 16806,
  [16841] = 16807,
  [16842] = 16842,
  [16843] = 16843,
  [16844] = 16844,
  [16845] = 16810,
  [16846] = 16846,
  [16847] = 16805,
  [16848] = 16808,
  [16849] = 16799,
  [16850] = 16850,
  [16851] = 16851,
  [16852] = 16818,
  [16853] = 6819,
  [16854] = 16789,
  [16855] = 16800,
  [16856] = 16792,
  [16857] = 16803,
  [16858] = 16858,
  [16859] = 16833,
  [16860] = 16794,
  [16861] = 16788,
  [16862] = 16815,
  [16863] = 16863,
  [16864] = 16816,
  [16865] = 16815,
  [16866] = 16816,
  [16867] = 16867,
  [16868] = 16815,
  [16869] = 16816,
  [16870] = 16799,
  [16871] = 16802,
  [16872] = 16872,
  [16873] = 16803,
  [16874] = 16806,
  [16875] = 16807,
  [16876] = 16799,
  [16877] = 16877,
  [16878] = 16878,
  [16879] = 16810,
  [16880] = 16800,
  [16881] = 16831,
  [16882] = 16882,
  [16883] = 16883,
  [16884] = 16884,
  [16885] = 16867,
  [16886] = 16800,
  [16887] = 16887,
  [16888] = 16878,
  [16889] = 16818,
  [16890] = 16890,
  [16891] = 16891,
  [16892] = 16789,
  [16893] = 8263,
  [16894] = 16792,
  [16895] = 8260,
  [16896] = 16833,
  [16897] = 16794,
  [16898] = 16799,
  [16899] = 16842,
  [16900] = 16788,
  [16901] = 16901,
  [16902] = 16799,
  [16903] = 16858,
  [16904] = 16800,
  [16905] = 16799,
  [16906] = 16799,
  [16907] = 16802,
  [16908] = 16790,
  [16909] = 16806,
  [16910] = 16807,
  [16911] = 16911,
  [16912] = 16912,
  [16913] = 16844,
  [16914] = 16803,
  [16915] = 16810,
  [16916] = 16799,
  [16917] = 16800,
  [16918] = 8130,
  [16919] = 16815,
  [16920] = 16816,
  [16921] = 16818,
  [16922] = 16922,
  [16923] = 16923,
  [16924] = 16789,
  [16925] = 16922,
  [16926] = 16792,
  [16927] = 16923,
  [16928] = 16799,
  [16929] = 2312,
  [16930] = 16793,
  [16931] = 16794,
  [16932] = 16804,
  [16933] = 16788,
  [16934] = 16799,
  [16935] = 16799,
  [16936] = 16800,
  [16937] = 16802,
  [16938] = 16877,
  [16939] = 16831,
  [16940] = 16806,
  [16941] = 16807,
  [16942] = 16800,
  [16943] = 16803,
  [16944] = 16810,
  [16945] = 16883,
  [16946] = 16800,
  [16947] = 16806,
  [16948] = 16799,
  [16949] = 16831,
  [16950] = 16800,
  [16951] = 16842,
  [16952] = 16818,
  [16953] = 16842,
  [16954] = 16954,
  [16955] = 16858,
  [16956] = 16789,
  [16957] = 16792,
  [16958] = 16958,
  [16959] = 16794,
  [16960] = 16815,
  [16961] = 16788,
  [16962] = 16816,
  [16963] = 16799,
  [16964] = 16802,
  [16965] = 16799,
  [16966] = 16800,
  [16967] = 16806,
  [16968] = 16807,
  [16969] = 16884,
  [16970] = 16810,
  [16971] = 16831,
  [16972] = 16788,
  [16973] = 16973,
  [16974] = 16800,
  [16975] = 16800,
  [16976] = 16799,
  [16977] = 16818,
  [16978] = 16800,
  [16979] = 16799,
  [16980] = 16800,
  [16981] = 16789,
  [16982] = 16858,
  [16983] = 16792,
  [16984] = 16922,
  [16985] = 16923,
  [16986] = 16815,
  [16987] = 16816,
  [16988] = 16794,
  [16989] = 16884,
  [16990] = 16788,
  [16991] = 16800,
  [16992] = 16799,
  [16993] = 16802,
  [16994] = 16842,
  [16995] = 16831,
  [16996] = 16996,
  [16997] = 16806,
  [16998] = 16807,
  [16999] = 16799,
  [17000] = 16810,
  [17001] = 16878,
  [17002] = 16800,
  [17003] = 16842,
  [17004] = 16831,
  [17005] = 16800,
  [17006] = 17006,
  [17007] = 16800,
  [17008] = 16818,
  [17009] = 16807,
  [17010] = 16833,
  [17011] = 16789,
  [17012] = 16792,
  [17013] = 16815,
  [17014] = 17014,
  [17015] = 16794,
  [17016] = 16799,
  [17017] = 16842,
  [17018] = 16788,
  [17019] = 17019,
  [17020] = 11843,
  [17021] = 16802,
  [17022] = 16806,
  [17023] = 16807,
  [17024] = 16799,
  [17025] = 16810,
  [17026] = 16800,
  [17027] = 17027,
  [17028] = 17028,
  [17029] = 16818,
  [17030] = 16789,
  [17031] = 16901,
  [17032] = 16792,
  [17033] = 17014,
  [17034] = 16816,
  [17035] = 16794,
  [17036] = 16788,
  [17037] = 16802,
  [17038] = 16806,
  [17039] = 16807,
  [17040] = 16810,
  [17041] = 16815,
  [17042] = 16816,
  [17043] = 17043,
  [17044] = 17044,
  [17045] = 16818,
  [17046] = 16842,
  [17047] = 16799,
  [17048] = 16800,
  [17049] = 16842,
  [17050] = 16792,
  [17051] = 16799,
  [17052] = 16803,
  [17053] = 16794,
  [17054] = 16788,
  [17055] = 16842,
  [17056] = 17056,
  [17057] = 16800,
  [17058] = 16806,
  [17059] = 16831,
  [17060] = 17060,
  [17061] = 16799,
  [17062] = 16800,
  [17063] = 16800,
  [17064] = 16818,
  [17065] = 16850,
  [17066] = 16850,
  [17067] = 16850,
  [17068] = 16792,
  [17069] = 16794,
  [17070] = 16788,
  [17071] = 16842,
  [17072] = 16806,
  [17073] = 8291,
  [17074] = 16867,
  [17075] = 17075,
  [17076] = 16799,
  [17077] = 16792,
  [17078] = 16800,
  [17079] = 16794,
  [17080] = 16788,
  [17081] = 16800,
  [17082] = 16792,
  [17083] = 16794,
  [17084] = 16788,
  [17085] = 16814,
  [17086] = 16799,
  [17087] = 16800,
  [17088] = 17056,
  [17089] = 10362,
  [17090] = 16816,
  [17091] = 16790,
  [17092] = 16792,
  [17093] = 17093,
  [17094] = 16844,
  [17095] = 16794,
  [17096] = 16815,
  [17097] = 16788,
  [17098] = 16816,
  [17099] = 16850,
  [17100] = 8288,
  [17101] = 16799,
  [17102] = 16792,
  [17103] = 16800,
  [17104] = 16792,
  [17105] = 16842,
  [17106] = 16799,
  [17107] = 16794,
  [17108] = 16922,
  [17109] = 16867,
  [17110] = 16788,
  [17111] = 16800,
  [17112] = 17112,
  [17113] = 16799,
  [17114] = 16792,
  [17115] = 16799,
  [17116] = 16794,
  [17117] = 16831,
  [17118] = 16788,
  [17119] = 17075,
  [17120] = 17014,
  [17121] = 16800,
  [17122] = 16800,
  [17123] = 16792,
  [17124] = 16814,
  [17125] = 16794,
  [17126] = 16901,
  [17127] = 8256,
  [17128] = 17128,
  [17129] = 16799,
  [17130] = 16792,
  [17131] = 16800,
  [17132] = 16842,
  [17133] = 16794,
  [17134] = 8260,
  [17135] = 16788,
  [17136] = 17136,
  [17137] = 2301,
  [17138] = 7930,
  [17139] = 16799,
  [17140] = 16815,
  [17141] = 16799,
  [17142] = 16800,
  [17143] = 16792,
  [17144] = 16816,
  [17145] = 16794,
  [17146] = 16788,
  [17147] = 16800,
  [17148] = 16799,
  [17149] = 16800,
  [17150] = 16799,
  [17151] = 16792,
  [17152] = 8199,
  [17153] = 8201,
  [17154] = 16794,
  [17155] = 16800,
  [17156] = 16788,
  [17157] = 17157,
  [17158] = 17158,
  [17159] = 16792,
  [17160] = 16877,
  [17161] = 16815,
  [17162] = 16816,
  [17163] = 16794,
  [17164] = 16883,
  [17165] = 16788,
  [17166] = 16799,
  [17167] = 16800,
  [17168] = 16799,
  [17169] = 16833,
  [17170] = 16792,
  [17171] = 16805,
  [17172] = 16831,
  [17173] = 16794,
  [17174] = 16799,
  [17175] = 16788,
  [17176] = 16800,
  [17177] = 16808,
  [17178] = 17178,
  [17179] = 16800,
  [17180] = 17180,
  [17181] = 17181,
  [17182] = 16799,
  [17183] = 16792,
  [17184] = 16800,
  [17185] = 16794,
  [17186] = 16788,
  [17187] = 6810,
  [17188] = 16842,
  [17189] = 16923,
  [17190] = 17190,
  [17191] = 16792,
  [17192] = 16794,
  [17193] = 16788,
  [17194] = 16799,
  [17195] = 16805,
  [17196] = 16800,
  [17197] = 16799,
  [17198] = 16878,
  [17199] = 16788,
  [17200] = 16799,
  [17201] = 16800,
  [17202] = 16789,
  [17203] = 16788,
  [17204] = 16808,
  [17205] = 16911,
  [17206] = 16803,
  [17207] = 16788,
  [17208] = 16799,
  [17209] = 16833,
  [17210] = 16794,
  [17211] = 16788,
  [17212] = 16799,
  [17213] = 16800,
  [17214] = 16814,
  [17215] = 16788,
  [17216] = 16815,
  [17217] = 16816,
  [17218] = 16878,
  [17219] = 16788,
  [17220] = 16884,
  [17221] = 16922,
  [17222] = 16788,
  [17223] = 16799,
  [17224] = 17224,
  [17225] = 16799,
  [17226] = 16800,
  [17227] = 16901,
  [17228] = 16867,
  [17229] = 16923,
  [17230] = 16858,
  [17231] = 16831,
  [17232] = 16803,
  [17233] = 16878,
  [17234] = 16800,
  [17235] = 16799,
  [17236] = 16800,
  [17237] = 16790,
  [17238] = 16878,
  [17239] = 16831,
  [17240] = 16842,
  [17241] = 16844,
  [17242] = 16901,
  [17243] = 8281,
  [17244] = 17244,
  [17245] = 16800,
  [17246] = 8178,
  [17247] = 16799,
  [17248] = 16815,
  [17249] = 16800,
  [17250] = 8117,
  [17251] = 16790,
  [17252] = 16816,
  [17253] = 16814,
  [17254] = 8159,
  [17255] = 16815,
  [17256] = 16790,
  [17257] = 16816,
  [17258] = 7874,
  [17259] = 16833,
  [17260] = 16844,
  [17261] = 16799,
  [17262] = 16800,
  [17263] = 16799,
  [17264] = 16831,
  [17265] = 16799,
  [17266] = 17266,
  [17267] = 16831,
  [17268] = 16800,
  [17269] = 16800,
  [17270] = 17270,
  [17271] = 16799,
  [17272] = 16815,
  [17273] = 16816,
  [17274] = 2283,
  [17275] = 16815,
  [17276] = 16842,
  [17277] = 16799,
  [17278] = 16799,
  [17279] = 16816,
  [17280] = 16800,
  [17281] = 16799,
  [17282] = 16799,
  [17283] = 16844,
  [17284] = 16815,
  [17285] = 16816,
  [17286] = 16922,
  [17287] = 16923,
  [17288] = 16799,
  [17289] = 16911,
  [17290] = 16799,
  [17291] = 17291,
  [17292] = 16800,
  [17293] = 16799,
  [17294] = 16831,
  [17295] = 16799,
  [17296] = 16800,
  [17297] = 16831,
  [17298] = 16805,
  [17299] = 17299,
  [17300] = 16814,
  [17301] = 16799,
  [17302] = 16842,
  [17303] = 16799,
  [17304] = 16831,
  [17305] = 16922,
  [17306] = 16800,
  [17307] = 16923,
  [17308] = 16799,
  [17309] = 16901,
  [17310] = 16799,
  [17311] = 17311,
  [17312] = 16800,
  [17313] = 16808,
  [17314] = 16800,
  [17315] = 16800,
  [17316] = 16831,
  [17317] = 17317,
  [17318] = 16799,
  [17319] = 16867,
  [17320] = 16800,
  [17321] = 16800,
  [17322] = 17322,
  [17323] = 17014,
  [17324] = 8149,
  [17325] = 16922,
  [17326] = 16799,
  [17327] = 16815,
  [17328] = 16839,
  [17329] = 16800,
  [17330] = 16816,
  [17331] = 16923,
  [17332] = 17056,
  [17333] = 16831,
  [17334] = 16958,
  [17335] = 16799,
  [17336] = 17006,
  [17337] = 16799,
  [17338] = 16912,
  [17339] = 16800,
  [17340] = 17340,
  [17341] = 16800,
  [17342] = 16799,
  [17343] = 16911,
  [17344] = 17344,
  [17345] = 16799,
  [17346] = 16800,
  [17347] = 16842,
  [17348] = 16842,
  [17349] = 16831,
  [17350] = 17350,
  [17351] = 17340,
  [17352] = 16800,
  [17353] = 17056,
  [17354] = 16839,
  [17355] = 17355,
  [17356] = 8148,
  [17357] = 16833,
  [17358] = 16799,
  [17359] = 16800,
  [17360] = 16958,
  [17361] = 17006,
  [17362] = 16912,
  [17363] = 16842,
  [17364] = 17364,
  [17365] = 16850,
  [17366] = 16850,
  [17367] = 16800,
  [17368] = 16858,
  [17369] = 16833,
  [17370] = 16839,
  [17371] = 16799,
  [17372] = 16799,
  [17373] = 16800,
  [17374] = 16815,
  [17375] = 16800,
  [17376] = 16958,
  [17377] = 16816,
  [17378] = 17006,
  [17379] = 16912,
  [17380] = 16833,
  [17381] = 16800,
  [17382] = 16815,
  [17383] = 16816,
  [17384] = 16839,
  [17385] = 16799,
  [17386] = 16800,
  [17387] = 16884,
  [17388] = 16803,
  [17389] = 16958,
  [17390] = 16799,
  [17391] = 16799,
  [17392] = 17006,
  [17393] = 8166,
  [17394] = 16912,
  [17395] = 16911,
  [17396] = 17396,
  [17397] = 16922,
  [17398] = 16923,
  [17399] = 16799,
  [17400] = 16851,
  [17401] = 16800,
  [17402] = 16839,
  [17403] = 8179,
  [17404] = 16831,
  [17405] = 16799,
  [17406] = 16958,
  [17407] = 16800,
  [17408] = 17006,
  [17409] = 16800,
  [17410] = 16912,
  [17411] = 16800,
  [17412] = 16800,
  [17413] = 16831,
  [17414] = 16867,
  [17415] = 16839,
  [17416] = 16800,
  [17417] = 16890,
  [17418] = 16800,
  [17419] = 16958,
  [17420] = 16842,
  [17421] = 17421,
  [17422] = 17006,
  [17423] = 16800,
  [17424] = 16912,
  [17425] = 16831,
  [17426] = 16891,
  [17427] = 16800,
  [17428] = 8195,
  [17429] = 17056,
  [17430] = 16884,
  [17431] = 16839,
  [17432] = 16799,
  [17433] = 16800,
  [17434] = 16800,
  [17435] = 16958,
  [17436] = 17014,
  [17437] = 16800,
  [17438] = 17006,
  [17439] = 17439,
  [17440] = 16912,
  [17441] = 16800,
  [17442] = 17442,
  [17443] = 16839,
  [17444] = 16884,
  [17445] = 16800,
  [17446] = 8146,
  [17447] = 16815,
  [17448] = 16800,
  [17449] = 16958,
  [17450] = 17450,
  [17451] = 16816,
  [17452] = 17006,
  [17453] = 16912,
  [17454] = 16800,
  [17455] = 17455,
  [17456] = 16800,
  [17457] = 16877,
  [17458] = 16842,
  [17459] = 16839,
  [17460] = 17460,
  [17461] = 16799,
  [17462] = 16800,
  [17463] = 16958,
  [17464] = 17014,
  [17465] = 17006,
  [17466] = 16800,
  [17467] = 16912,
  [17468] = 16803,
  [17469] = 16789,
  [17470] = 16803,
  [17471] = 17056,
  [17472] = 16839,
  [17473] = 16792,
  [17474] = 17056,
  [17475] = 16803,
  [17476] = 16793,
  [17477] = 16831,
  [17478] = 16958,
  [17479] = 16850,
  [17480] = 17006,
  [17481] = 16912,
  [17482] = 16800,
  [17483] = 17483,
  [17484] = 16839,
  [17485] = 16818,
  [17486] = 16794,
  [17487] = 16958,
  [17488] = 17006,
  [17489] = 16912,
  [17490] = 16842,
  [17491] = 16877,
  [17492] = 16878,
  [17493] = 16883,
  [17494] = 16833,
  [17495] = 6557,
  [17496] = 16958,
  [17497] = 17006,
  [17498] = 16788,
  [17499] = 16912,
  [17500] = 16833,
  [17501] = 7888,
  [17502] = 16958,
  [17503] = 17006,
  [17504] = 16912,
  [17505] = 17505,
  [17506] = 7893,
  [17507] = 16878,
  [17508] = 16814,
  [17509] = 8110,
  [17510] = 17291,
  [17511] = 17511,
  [17512] = 16883,
  [17513] = 8112,
  [17514] = 16901,
  [17515] = 16799,
  [17516] = 16901,
  [17517] = 8119,
  [17518] = 16803,
  [17519] = 16803,
  [17520] = 16850,
  [17521] = 16831,
  [17522] = 16802,
  [17523] = 16800,
  [17524] = 8120,
  [17525] = 16878,
  [17526] = 16878,
  [17527] = 17355,
  [17528] = 8127,
  [17529] = 17014,
  [17530] = 16805,
  [17531] = 16797,
  [17532] = 16808,
  [17533] = 16842,
  [17534] = 17534,
  [17535] = 2438,
  [17536] = 16833,
  [17537] = 16901,
  [17538] = 16804,
  [17539] = 16799,
  [17540] = 16867,
  [17541] = 16790,
  [17542] = 16790,
  [17543] = 16901,
  [17544] = 16831,
  [17545] = 16806,
  [17546] = 16800,
  [17547] = 16807,
  [17548] = 16803,
  [17549] = 16844,
  [17550] = 16844,
  [17551] = 16842,
  [17552] = 16878,
  [17553] = 16790,
  [17554] = 16901,
  [17555] = 17014,
  [17556] = 17556,
  [17557] = 16799,
  [17558] = 16799,
  [17559] = 16810,
  [17560] = 16844,
  [17561] = 16799,
  [17562] = 16831,
  [17563] = 16800,
  [17564] = 17564,
  [17565] = 16815,
  [17566] = 16816,
  [17567] = 16800,
  [17568] = 16842,
  [17569] = 16842,
  [17570] = 8155,
  [17571] = 16815,
  [17572] = 16799,
  [17573] = 16901,
  [17574] = 16799,
  [17575] = 2425,
  [17576] = 16831,
  [17577] = 2367,
  [17578] = 16800,
  [17579] = 16922,
  [17580] = 16923,
  [17581] = 17581,
  [17582] = 16842,
  [17583] = 16818,
  [17584] = 16878,
  [17585] = 16815,
  [17586] = 16831,
  [17587] = 16790,
  [17588] = 16799,
  [17589] = 16858,
  [17590] = 17590,
  [17591] = 16800,
  [17592] = 16816,
  [17593] = 16831,
  [17594] = 16833,
  [17595] = 16800,
  [17596] = 16842,
  [17597] = 16805,
  [17598] = 16815,
  [17599] = 16816,
  [17600] = 16833,
  [17601] = 16842,
  [17602] = 17014,
  [17603] = 17603,
  [17604] = 6559,
  [17605] = 16799,
  [17606] = 16808,
  [17607] = 17056,
  [17608] = 16789,
  [17609] = 16831,
  [17610] = 16799,
  [17611] = 16800,
  [17612] = 17612,
  [17613] = 16792,
  [17614] = 17614,
  [17615] = 16842,
  [17616] = 16793,
  [17617] = 16842,
  [17618] = 16800,
  [17619] = 16794,
  [17620] = 16799,
  [17621] = 17056,
  [17622] = 16790,
  [17623] = 17623,
  [17624] = 16799,
  [17625] = 16850,
  [17626] = 16844,
  [17627] = 16831,
  [17628] = 16788,
  [17629] = 16901,
  [17630] = 16800,
  [17631] = 16850,
  [17632] = 7933,
  [17633] = 17633,
  [17634] = 8128,
  [17635] = 17635,
  [17636] = 17027,
  [17637] = 16842,
  [17638] = 8095,
  [17639] = 16831,
  [17640] = 17640,
  [17641] = 17641,
  [17642] = 16799,
  [17643] = 16844,
  [17644] = 17644,
  [17645] = 16799,
  [17646] = 6835,
  [17647] = 16911,
  [17648] = 16802,
  [17649] = 16800,
  [17650] = 16790,
  [17651] = 16831,
  [17652] = 17014,
  [17653] = 16800,
  [17654] = 16804,
  [17655] = 17655,
  [17656] = 6625,
  [17657] = 16922,
  [17658] = 17460,
  [17659] = 17659,
  [17660] = 16842,
  [17661] = 17244,
  [17662] = 16806,
  [17663] = 16807,
  [17664] = 16799,
  [17665] = 16923,
  [17666] = 17056,
  [17667] = 16799,
  [17668] = 17668,
  [17669] = 16810,
  [17670] = 16800,
  [17671] = 16831,
  [17672] = 16800,
  [17673] = 16810,
  [17674] = 8233,
  [17675] = 8234,
  [17676] = 8238,
  [17677] = 16850,
  [17678] = 8256,
  [17679] = 8252,
  [17680] = 16842,
  [17681] = 16844,
  [17682] = 17682,
  [17683] = 17683,
  [17684] = 16790,
  [17685] = 16802,
  [17686] = 16842,
  [17687] = 17180,
  [17688] = 16799,
  [17689] = 16844,
  [17690] = 17644,
  [17691] = 16805,
  [17692] = 15754,
  [17693] = 8239,
  [17694] = 16800,
  [17695] = 8240,
  [17696] = 8262,
  [17697] = 8096,
  [17698] = 16831,
  [17699] = 16858,
  [17700] = 16815,
  [17701] = 16816,
  [17702] = 16818,
  [17703] = 16799,
  [17704] = 16911,
  [17705] = 16922,
  [17706] = 16923,
  [17707] = 16800,
  [17708] = 17708,
  [17709] = 17709,
  [17710] = 17710,
  [17711] = 17711,
  [17712] = 17712,
  [17713] = 17713,
  [17714] = 17714,
  [17715] = 17715,
  [17716] = 17716,
  [17717] = 17717,
  [17718] = 17718,
  [17719] = 8149,
  [17720] = 17713,
  [17721] = 17721,
  [17722] = 17711,
  [17723] = 17723,
  [17724] = 17724,
  [17725] = 17725,
  [17726] = 17714,
  [17727] = 17727,
  [17728] = 17728,
  [17729] = 17729,
  [17730] = 17730,
  [17731] = 17718,
  [17732] = 17716,
  [17733] = 17733,
  [17734] = 17734,
  [17735] = 17735,
  [17736] = 17736,
  [17737] = 17729,
  [17738] = 17709,
  [17739] = 17725,
  [17740] = 17740,
  [17741] = 17741,
  [17742] = 17713,
  [17743] = 17743,
  [17744] = 17711,
  [17745] = 17714,
  [17746] = 17721,
  [17747] = 17747,
  [17748] = 17714,
  [17749] = 17721,
  [17750] = 2542,
  [17751] = 17724,
  [17752] = 17723,
  [17753] = 17753,
  [17754] = 17754,
  [17755] = 17755,
  [17756] = 17727,
  [17757] = 17728,
  [17758] = 2572,
  [17759] = 17730,
  [17760] = 17760,
  [17761] = 17761,
  [17762] = 17725,
  [17763] = 2573,
  [17764] = 17710,
  [17765] = 17736,
  [17766] = 2574,
  [17767] = 17709,
  [17768] = 8355,
  [17769] = 17769,
  [17770] = 17718,
  [17771] = 17713,
  [17772] = 17713,
  [17773] = 17740,
  [17774] = 17747,
  [17775] = 17741,
  [17776] = 8337,
  [17777] = 17754,
  [17778] = 17710,
  [17779] = 17734,
  [17780] = 17761,
  [17781] = 17723,
  [17782] = 17782,
  [17783] = 17761,
  [17784] = 17728,
  [17785] = 17727,
  [17786] = 17728,
  [17787] = 8469,
  [17788] = 17730,
  [17789] = 17769,
  [17790] = 17723,
  [17791] = 17791,
  [17792] = 8300,
  [17793] = 17736,
  [17794] = 2467,
  [17795] = 17709,
  [17796] = 17796,
  [17797] = 17797,
  [17798] = 17798,
  [17799] = 17713,
  [17800] = 17800,
  [17801] = 8325,
  [17802] = 8256,
  [17803] = 17803,
  [17804] = 2575,
  [17805] = 17734,
  [17806] = 8378,
  [17807] = 17725,
  [17808] = 17808,
  [17809] = 17723,
  [17810] = 17810,
  [17811] = 17743,
  [17812] = 17782,
  [17813] = 17727,
  [17814] = 17728,
  [17815] = 8309,
  [17816] = 17730,
  [17817] = 17735,
  [17818] = 17723,
  [17819] = 17803,
  [17820] = 17820,
  [17821] = 8315,
  [17822] = 17736,
  [17823] = 17725,
  [17824] = 17709,
  [17825] = 17713,
  [17826] = 2367,
  [17827] = 17710,
  [17828] = 17727,
  [17829] = 17829,
  [17830] = 17810,
  [17831] = 17831,
  [17832] = 17730,
  [17833] = 17728,
  [17834] = 17831,
  [17835] = 17835,
  [17836] = 17741,
  [17837] = 17709,
  [17838] = 17808,
  [17839] = 17839,
  [17840] = 17730,
  [17841] = 8458,
  [17842] = 17727,
  [17843] = 17730,
  [17844] = 17728,
  [17845] = 17743,
  [17846] = 17820,
  [17847] = 17718,
  [17848] = 17709,
  [17849] = 17729,
  [17850] = 17711,
  [17851] = 17733,
  [17852] = 17714,
  [17853] = 17730,
  [17854] = 17854,
  [17855] = 17718,
  [17856] = 17710,
  [17857] = 17735,
  [17858] = 17709,
  [17859] = 17736,
  [17860] = 17782,
  [17861] = 17835,
  [17862] = 8332,
  [17863] = 17730,
  [17864] = 17709,
  [17865] = 17730,
  [17866] = 17755,
  [17867] = 17867,
  [17868] = 17709,
  [17869] = 17810,
  [17870] = 17724,
  [17871] = 8361,
  [17872] = 17730,
  [17873] = 17831,
  [17874] = 17835,
  [17875] = 17754,
  [17876] = 16234,
  [17877] = 17709,
  [17878] = 16236,
  [17879] = 17729,
  [17880] = 17753,
  [17881] = 17753,
  [17882] = 17730,
  [17883] = 17883,
  [17884] = 16770,
  [17885] = 17798,
  [17886] = 17709,
  [17887] = 17887,
  [17888] = 17829,
  [17889] = 17798,
  [17890] = 17730,
  [17891] = 17721,
  [17892] = 17892,
  [17893] = 17810,
  [17894] = 17709,
  [17895] = 8469,
  [17896] = 17725,
  [17897] = 17711,
  [17898] = 17714,
  [17899] = 17730,
  [17900] = 17718,
  [17901] = 17710,
  [17902] = 17709,
  [17903] = 17710,
  [17904] = 8372,
  [17905] = 17734,
  [17906] = 17854,
  [17907] = 17730,
  [17908] = 17829,
  [17909] = 17760,
  [17910] = 17709,
  [17911] = 17715,
  [17912] = 17831,
  [17913] = 17708,
  [17914] = 17867,
  [17915] = 17725,
  [17916] = 17712,
  [17917] = 17839,
  [17918] = 17736,
  [17919] = 17753,
  [17920] = 17920,
  [17921] = 17733,
  [17922] = 17760,
  [17923] = 17769,
  [17924] = 17709,
  [17925] = 17755,
  [17926] = 17729,
  [17927] = 17887,
  [17928] = 17724,
  [17929] = 17753,
  [17930] = 17798,
  [17931] = 17713,
  [17932] = 17710,
  [17933] = 17740,
  [17934] = 17747,
  [17935] = 17754,
  [17936] = 17887,
  [17937] = 17761,
  [17938] = 17810,
  [17939] = 17769,
  [17940] = 16714,
  [17941] = 17854,
  [17942] = 17730,
  [17943] = 17943,
  [17944] = 17760,
  [17945] = 17803,
  [17946] = 17820,
  [17947] = 17831,
  [17948] = 17808,
  [17949] = 17835,
  [17950] = 17743,
  [17951] = 17839,
  [17952] = 17829,
  [17953] = 17735,
  [17954] = 17725,
  [17955] = 17711,
  [17956] = 17724,
  [17957] = 17810,
  [17958] = 17831,
  [17959] = 17835,
  [17960] = 17960,
  [17961] = 17961,
  [17962] = 8576,
  [17963] = 17723,
  [17964] = 17729,
  [17965] = 17711,
  [17966] = 17714,
  [17967] = 17718,
  [17968] = 17710,
  [17969] = 17729,
  [17970] = 17729,
  [17971] = 17729,
  [17972] = 17972,
  [17973] = 17727,
  [17974] = 17721,
  [17975] = 17728,
  [17976] = 17735,
  [17977] = 17730,
  [17978] = 17831,
  [17979] = 8558,
  [17980] = 17729,
  [17981] = 17943,
  [17982] = 17820,
  [17983] = 17798,
  [17984] = 17721,
  [17985] = 17831,
  [17986] = 8361,
  [17987] = 17724,
  [17988] = 17854,
  [17989] = 17753,
  [17990] = 17782,
  [17991] = 17867,
  [17992] = 17867,
  [17993] = 17993,
  [17994] = 17835,
  [17995] = 17887,
  [17996] = 17736,
  [17997] = 17997,
  [17998] = 17998,
  [17999] = 17709,
  [18000] = 17723,
  [18001] = 18001,
  [18002] = 17760,
  [18003] = 18003,
  [18004] = 8577,
  [18005] = 17710,
  [18006] = 17716,
  [18007] = 17892,
  [18008] = 2472,
  [18009] = 17733,
  [18010] = 8378,
  [18011] = 17740,
  [18012] = 17747,
  [18013] = 17754,
  [18014] = 17820,
  [18015] = 17761,
  [18016] = 17729,
  [18017] = 17769,
  [18018] = 17717,
  [18019] = 17831,
  [18020] = 17716,
  [18021] = 17740,
  [18022] = 17803,
  [18023] = 17717,
  [18024] = 17808,
  [18025] = 17727,
  [18026] = 17743,
  [18027] = 18027,
  [18028] = 17728,
  [18029] = 17735,
  [18030] = 17714,
  [18031] = 17760,
  [18032] = 17760,
  [18033] = 17810,
  [18034] = 17831,
  [18035] = 17835,
  [18036] = 17729,
  [18037] = 17716,
  [18038] = 17711,
  [18039] = 17714,
  [18040] = 17718,
  [18041] = 17710,
  [18042] = 17887,
  [18043] = 17740,
  [18044] = 8781,
  [18045] = 17892,
  [18046] = 17717,
  [18047] = 17747,
  [18048] = 17835,
  [18049] = 17729,
  [18050] = 17755,
  [18051] = 17730,
  [18052] = 17782,
  [18053] = 18053,
  [18054] = 17760,
  [18055] = 17740,
  [18056] = 17760,
  [18057] = 17710,
  [18058] = 8355,
  [18059] = 17711,
  [18060] = 8300,
  [18061] = 17769,
  [18062] = 17714,
  [18063] = 6967,
  [18064] = 17740,
  [18065] = 17713,
  [18066] = 17835,
  [18067] = 8332,
  [18068] = 17760,
  [18069] = 8309,
  [18070] = 8481,
  [18071] = 18071,
  [18072] = 8329,
  [18073] = 17740,
  [18074] = 8348,
  [18075] = 17709,
  [18076] = 17740,
  [18077] = 17747,
  [18078] = 17754,
  [18079] = 17798,
  [18080] = 17761,
  [18081] = 17754,
  [18082] = 17747,
  [18083] = 8367,
  [18084] = 17769,
  [18085] = 17808,
  [18086] = 17754,
  [18087] = 17760,
  [18088] = 17711,
  [18089] = 17723,
  [18090] = 17711,
  [18091] = 17803,
  [18092] = 17761,
  [18093] = 17892,
  [18094] = 6872,
  [18095] = 17808,
  [18096] = 17769,
  [18097] = 17743,
  [18098] = 17943,
  [18099] = 17724,
  [18100] = 17735,
  [18101] = 17727,
  [18102] = 17803,
  [18103] = 17810,
  [18104] = 17728,
  [18105] = 17831,
  [18106] = 17753,
  [18107] = 17835,
  [18108] = 17730,
  [18109] = 17711,
  [18110] = 17887,
  [18111] = 17718,
  [18112] = 8329,
  [18113] = 8490,
  [18114] = 18114,
  [18115] = 17710,
  [18116] = 17740,
  [18117] = 17714,
  [18118] = 17854,
  [18119] = 17803,
  [18120] = 17736,
  [18121] = 17760,
  [18122] = 17718,
  [18123] = 17782,
  [18124] = 17972,
  [18125] = 17711,
  [18126] = 12238,
  [18127] = 12239,
  [18128] = 17714,
  [18129] = 17709,
  [18130] = 18130,
  [18131] = 17718,
  [18132] = 18132,
  [18133] = 17734,
  [18134] = 17710,
  [18135] = 17740,
  [18136] = 17717,
  [18137] = 17808,
  [18138] = 8486,
  [18139] = 18139,
  [18140] = 17829,
  [18141] = 8195,
  [18142] = 17743,
  [18143] = 17735,
  [18144] = 17713,
  [18145] = 17747,
  [18146] = 17729,
  [18147] = 18147,
  [18148] = 18148,
  [18149] = 17740,
  [18150] = 18150,
  [18151] = 18151,
  [18152] = 17754,
  [18153] = 17734,
  [18154] = 18154,
  [18155] = 17760,
  [18156] = 8786,
  [18157] = 17854,
  [18158] = 18158,
  [18159] = 17920,
  [18160] = 17810,
  [18161] = 17721,
  [18162] = 17831,
  [18163] = 17835,
  [18164] = 17740,
  [18165] = 17747,
  [18166] = 17747,
  [18167] = 17740,
  [18168] = 17892,
  [18169] = 17747,
  [18170] = 17754,
  [18171] = 17810,
  [18172] = 17711,
  [18173] = 17761,
  [18174] = 17754,
  [18175] = 17725,
  [18176] = 17716,
  [18177] = 17736,
  [18178] = 17713,
  [18179] = 17782,
  [18180] = 17714,
  [18181] = 17733,
  [18182] = 17761,
  [18183] = 17769,
  [18184] = 17718,
  [18185] = 17754,
  [18186] = 18186,
  [18187] = 17972,
  [18188] = 17769,
  [18189] = 17760,
  [18190] = 6953,
  [18191] = 18147,
  [18192] = 8315,
  [18193] = 17740,
  [18194] = 18194,
  [18195] = 18132,
  [18196] = 17760,
  [18197] = 18197,
  [18198] = 18198,
  [18199] = 17710,
  [18200] = 18200,
  [18201] = 18201,
  [18202] = 18139,
  [18203] = 17723,
  [18204] = 18197,
  [18205] = 17723,
  [18206] = 18206,
  [18207] = 17761,
  [18208] = 18147,
  [18209] = 18148,
  [18210] = 17725,
  [18211] = 18150,
  [18212] = 17709,
  [18213] = 17727,
  [18214] = 17740,
  [18215] = 17747,
  [18216] = 17728,
  [18217] = 17798,
  [18218] = 17754,
  [18219] = 17920,
  [18220] = 17761,
  [18221] = 17769,
  [18222] = 17730,
  [18223] = 17760,
  [18224] = 17782,
  [18225] = 17740,
  [18226] = 17761,
  [18227] = 17729,
  [18228] = 8437,
  [18229] = 17803,
  [18230] = 17972,
  [18231] = 17747,
  [18232] = 17803,
  [18233] = 17808,
  [18234] = 8354,
  [18235] = 17803,
  [18236] = 17854,
  [18237] = 18132,
  [18238] = 17743,
  [18239] = 17716,
  [18240] = 17808,
  [18241] = 17747,
  [18242] = 17715,
  [18243] = 17735,
  [18244] = 18139,
  [18245] = 17736,
  [18246] = 2469,
  [18247] = 17709,
  [18248] = 6938,
  [18249] = 17820,
  [18250] = 18147,
  [18251] = 18148,
  [18252] = 17810,
  [18253] = 18150,
  [18254] = 17831,
  [18255] = 17754,
  [18256] = 17835,
  [18257] = 18257,
  [18258] = 17754,
  [18259] = 17761,
  [18260] = 17743,
  [18261] = 17920,
  [18262] = 18197,
  [18263] = 17729,
  [18264] = 17727,
  [18265] = 17808,
  [18266] = 17743,
  [18267] = 17769,
  [18268] = 17867,
  [18269] = 17711,
  [18270] = 17714,
  [18271] = 17972,
  [18272] = 17718,
  [18273] = 17710,
  [18274] = 17728,
  [18275] = 17754,
  [18276] = 17743,
  [18277] = 17769,
  [18278] = 18132,
  [18279] = 17769,
  [18280] = 17729,
  [18281] = 17782,
  [18282] = 8029,
  [18283] = 18139,
  [18284] = 17735,
  [18285] = 17714,
  [18286] = 17760,
  [18287] = 17741,
  [18288] = 18147,
  [18289] = 18148,
  [18290] = 17729,
  [18291] = 18150,
  [18292] = 17735,
  [18293] = 17730,
  [18294] = 8031,
  [18295] = 17761,
  [18296] = 8372,
  [18297] = 17761,
  [18298] = 17920,
  [18299] = 18158,
  [18300] = 17713,
  [18301] = 8437,
  [18302] = 17798,
  [18303] = 17711,
  [18304] = 10419,
  [18305] = 17810,
  [18306] = 17808,
  [18307] = 17710,
  [18308] = 17972,
  [18309] = 17798,
  [18310] = 17803,
  [18311] = 17829,
  [18312] = 17708,
  [18313] = 17808,
  [18314] = 17717,
  [18315] = 18132,
  [18316] = 17723,
  [18317] = 17854,
  [18318] = 17769,
  [18319] = 18139,
  [18320] = 17733,
  [18321] = 17727,
  [18322] = 17810,
  [18323] = 18147,
  [18324] = 18148,
  [18325] = 18150,
  [18326] = 17728,
  [18327] = 17803,
  [18328] = 17769,
  [18329] = 8408,
  [18330] = 17730,
  [18331] = 17736,
  [18332] = 17715,
  [18333] = 17920,
  [18334] = 17760,
  [18335] = 8348,
  [18336] = 17733,
  [18337] = 17831,
  [18338] = 17854,
  [18339] = 17803,
  [18340] = 17710,
  [18341] = 17710,
  [18342] = 17736,
  [18343] = 17972,
  [18344] = 17708,
  [18345] = 17709,
  [18346] = 17835,
  [18347] = 18132,
  [18348] = 17810,
  [18349] = 17743,
  [18350] = 17740,
  [18351] = 17747,
  [18352] = 17754,
  [18353] = 18139,
  [18354] = 17761,
  [18355] = 17709,
  [18356] = 17829,
  [18357] = 17829,
  [18358] = 18147,
  [18359] = 18148,
  [18360] = 18150,
  [18361] = 17769,
  [18362] = 17810,
  [18363] = 17755,
  [18364] = 17712,
  [18365] = 17735,
  [18366] = 17867,
  [18367] = 17760,
  [18368] = 17920,
  [18369] = 18369,
  [18370] = 17740,
  [18371] = 8319,
  [18372] = 17716,
  [18373] = 17760,
  [18374] = 17810,
  [18375] = 17831,
  [18376] = 17835,
  [18377] = 17803,
  [18378] = 17972,
  [18379] = 17747,
  [18380] = 17713,
  [18381] = 17831,
  [18382] = 17835,
  [18383] = 17808,
  [18384] = 18132,
  [18385] = 17716,
  [18386] = 17754,
  [18387] = 17743,
  [18388] = 17712,
  [18389] = 18158,
  [18390] = 18139,
  [18391] = 17735,
  [18392] = 17839,
  [18393] = 17711,
  [18394] = 8325,
  [18395] = 18147,
  [18396] = 18148,
  [18397] = 18150,
  [18398] = 17721,
  [18399] = 17755,
  [18400] = 17810,
  [18401] = 17831,
  [18402] = 17714,
  [18403] = 17803,
  [18404] = 17803,
  [18405] = 17920,
  [18406] = 17723,
  [18407] = 17740,
  [18408] = 17747,
  [18409] = 17718,
  [18410] = 17835,
  [18411] = 17727,
  [18412] = 17740,
  [18413] = 17972,
  [18414] = 8340,
  [18415] = 8338,
  [18416] = 17711,
  [18417] = 17714,
  [18418] = 17728,
  [18419] = 17718,
  [18420] = 17710,
  [18421] = 17754,
  [18422] = 17791,
  [18423] = 17730,
  [18424] = 18148,
  [18425] = 17867,
  [18426] = 8367,
  [18427] = 17710,
  [18428] = 17761,
  [18429] = 17854,
  [18430] = 17808,
  [18431] = 17736,
  [18432] = 17709,
  [18433] = 17747,
  [18434] = 17769,
  [18435] = 17717,
  [18436] = 17729,
  [18437] = 17712,
  [18438] = 17715,
  [18439] = 17754,
  [18440] = 17972,
  [18441] = 17761,
  [18442] = 17723,
  [18443] = 17829,
  [18444] = 17769,
  [18445] = 17716,
  [18446] = 17721,
  [18447] = 17727,
  [18448] = 17803,
  [18449] = 17724,
  [18450] = 17713,
  [18451] = 17713,
  [18452] = 17743,
  [18453] = 18148,
  [18454] = 8364,
  [18455] = 17803,
  [18456] = 17718,
  [18457] = 12258,
  [18458] = 17867,
  [18459] = 17808,
  [18460] = 17761,
  [18461] = 17717,
  [18462] = 17717,
  [18463] = 17972,
  [18464] = 17753,
  [18465] = 8502,
  [18466] = 17723,
  [18467] = 17733,
  [18468] = 17743,
  [18469] = 18469,
  [18470] = 17831,
  [18471] = 17716,
  [18472] = 17727,
  [18473] = 17728,
  [18474] = 17733,
  [18475] = 18148,
  [18476] = 17808,
  [18477] = 17730,
  [18478] = 17839,
  [18479] = 17854,
  [18480] = 17711,
  [18481] = 17743,
  [18482] = 17723,
  [18483] = 17735,
  [18484] = 6625,
  [18485] = 17714,
  [18486] = 8312,
  [18487] = 17803,
  [18488] = 17718,
  [18489] = 17972,
  [18490] = 17711,
  [18491] = 17760,
  [18492] = 17734,
  [18493] = 17736,
  [18494] = 8330,
  [18495] = 17714,
  [18496] = 17718,
  [18497] = 17709,
  [18498] = 17734,
  [18499] = 17769,
  [18500] = 17735,
  [18501] = 8337,
  [18502] = 18148,
  [18503] = 17725,
  [18504] = 17725,
  [18505] = 17740,
  [18506] = 17735,
  [18507] = 17810,
  [18508] = 17747,
  [18509] = 17808,
  [18510] = 17831,
  [18511] = 17727,
  [18512] = 17754,
  [18513] = 17835,
  [18514] = 17743,
  [18515] = 17755,
  [18516] = 17972,
  [18517] = 17761,
  [18518] = 17710,
  [18519] = 17769,
  [18520] = 17728,
  [18521] = 17713,
  [18522] = 17740,
  [18523] = 17892,
  [18524] = 17791,
  [18525] = 17730,
  [18526] = 17711,
  [18527] = 18148,
  [18528] = 2582,
  [18529] = 16236,
  [18530] = 6935,
  [18531] = 17714,
  [18532] = 17733,
  [18533] = 2583,
  [18534] = 17718,
  [18535] = 17735,
  [18536] = 17972,
  [18537] = 17803,
  [18538] = 18538,
  [18539] = 17723,
  [18540] = 17741,
  [18541] = 17782,
  [18542] = 17808,
  [18543] = 18543,
  [18544] = 17743,
  [18545] = 8312,
  [18546] = 17782,
  [18547] = 17769,
  [18548] = 18148,
  [18549] = 2473,
  [18550] = 17887,
  [18551] = 17747,
  [18552] = 17710,
  [18553] = 17943,
  [18554] = 17733,
  [18555] = 17798,
  [18556] = 17727,
  [18557] = 17972,
  [18558] = 17735,
  [18559] = 17854,
  [18560] = 17710,
  [18561] = 17728,
  [18562] = 17724,
  [18563] = 17810,
  [18564] = 17736,
  [18565] = 17715,
  [18566] = 18148,
  [18567] = 17831,
  [18568] = 17829,
  [18569] = 17972,
  [18570] = 17835,
  [18571] = 17709,
  [18572] = 17754,
  [18573] = 17810,
  [18574] = 17711,
  [18575] = 17714,
  [18576] = 17718,
  [18577] = 17710,
  [18578] = 18148,
  [18579] = 17741,
  [18580] = 17831,
  [18581] = 17835,
  [18582] = 17972,
  [18583] = 17710,
  [18584] = 17829,
  [18585] = 17753,
  [18586] = 17730,
  [18587] = 17887,
  [18588] = 17735,
  [18589] = 17734,
  [18590] = 17755,
  [18591] = 18053,
  [18592] = 18148,
  [18593] = 17741,
  [18594] = 17972,
  [18595] = 18158,
  [18596] = 18596,
  [18597] = 17736,
  [18598] = 18197,
  [18599] = 17721,
  [18600] = 17709,
  [18601] = 17713,
  [18602] = 18148,
  [18603] = 8260,
  [18604] = 17867,
  [18605] = 17735,
  [18606] = 17972,
  [18607] = 17713,
  [18608] = 17733,
  [18609] = 17755,
  [18610] = 17733,
  [18611] = 17716,
  [18612] = 17747,
  [18613] = 17717,
  [18614] = 17715,
  [18615] = 17755,
  [18616] = 2537,
  [18617] = 17724,
  [18618] = 18148,
  [18619] = 17741,
  [18620] = 17972,
  [18621] = 17753,
  [18622] = 17782,
  [18623] = 17724,
  [18624] = 2539,
  [18625] = 18625,
  [18626] = 17753,
  [18627] = 17835,
  [18628] = 18148,
  [18629] = 17798,
  [18630] = 17972,
  [18631] = 17711,
  [18632] = 17761,
  [18633] = 17887,
  [18634] = 17723,
  [18635] = 17740,
  [18636] = 17998,
  [18637] = 17892,
  [18638] = 17769,
  [18639] = 17714,
  [18640] = 18148,
  [18641] = 18148,
  [18642] = 18642,
  [18643] = 17721,
  [18644] = 17972,
  [18645] = 17887,
  [18646] = 8354,
  [18647] = 17727,
  [18648] = 17728,
  [18649] = 17803,
  [18650] = 17718,
  [18651] = 17743,
  [18652] = 17729,
  [18653] = 18148,
  [18654] = 17734,
  [18655] = 17839,
  [18656] = 17972,
  [18657] = 17725,
  [18658] = 17711,
  [18659] = 17810,
  [18660] = 17831,
  [18661] = 8408,
  [18662] = 17730,
  [18663] = 17782,
  [18664] = 17711,
  [18665] = 17714,
  [18666] = 18148,
  [18667] = 17808,
  [18668] = 17839,
  [18669] = 17708,
  [18670] = 17972,
  [18671] = 17798,
  [18672] = 17892,
  [18673] = 17712,
  [18674] = 2485,
  [18675] = 17708,
  [18676] = 17761,
  [18677] = 17718,
  [18678] = 18148,
  [18679] = 17972,
  [18680] = 17710,
  [18681] = 18681,
  [18682] = 18682,
  [18683] = 17729,
  [18684] = 8409,
  [18685] = 17839,
  [18686] = 17715,
  [18687] = 17736,
  [18688] = 18148,
  [18689] = 17829,
  [18690] = 2544,
  [18691] = 17708,
  [18692] = 17972,
  [18693] = 17709,
  [18694] = 17713,
  [18695] = 6921,
  [18696] = 17791,
  [18697] = 17740,
  [18698] = 17743,
  [18699] = 17712,
  [18700] = 17713,
  [18701] = 17791,
  [18702] = 17733,
  [18703] = 18148,
  [18704] = 17755,
  [18705] = 17972,
  [18706] = 2588,
  [18707] = 17943,
  [18708] = 18158,
  [18709] = 17740,
  [18710] = 17769,
  [18711] = 18711,
  [18712] = 17724,
  [18713] = 18148,
  [18714] = 17753,
  [18715] = 17730,
  [18716] = 17943,
  [18717] = 17839,
  [18718] = 17887,
  [18719] = 17803,
  [18720] = 2545,
  [18721] = 17892,
  [18722] = 18197,
  [18723] = 17810,
  [18724] = 17831,
  [18725] = 17711,
  [18726] = 18726,
  [18727] = 17803,
  [18728] = 17791,
  [18729] = 2589,
  [18730] = 8340,
  [18731] = 17714,
  [18732] = 17769,
  [18733] = 17723,
  [18734] = 17803,
  [18735] = 17708,
  [18736] = 8338,
  [18737] = 17736,
  [18738] = 17727,
  [18739] = 17728,
  [18740] = 18740,
  [18741] = 17712,
  [18742] = 18742,
  [18743] = 17712,
  [18744] = 18596,
  [18745] = 17730,
  [18746] = 17943,
  [18747] = 17839,
  [18748] = 17808,
  [18749] = 17892,
  [18750] = 17755,
  [18751] = 17747,
  [18752] = 17803,
  [18753] = 17810,
  [18754] = 17740,
  [18755] = 17769,
  [18756] = 18756,
  [18757] = 17736,
  [18758] = 17709,
  [18759] = 17803,
  [18760] = 17723,
  [18761] = 17729,
  [18762] = 17810,
  [18763] = 17831,
  [18764] = 17711,
  [18765] = 17754,
  [18766] = 10393,
  [18767] = 17810,
  [18768] = 8502,
  [18769] = 17714,
  [18770] = 17724,
  [18771] = 17839,
  [18772] = 17713,
  [18773] = 6886,
  [18774] = 17808,
  [18775] = 18775,
  [18776] = 18776,
  [18777] = 17761,
  [18778] = 17723,
  [18779] = 17735,
  [18780] = 8392,
  [18781] = 17743,
  [18782] = 6928,
  [18783] = 17831,
  [18784] = 17760,
  [18785] = 17723,
  [18786] = 17714,
  [18787] = 17753,
  [18788] = 17769,
  [18789] = 17740,
  [18790] = 17769,
  [18791] = 17708,
  [18792] = 8402,
  [18793] = 18158,
  [18794] = 17803,
  [18795] = 18139,
  [18796] = 17727,
  [18797] = 17810,
  [18798] = 17831,
  [18799] = 17711,
  [18800] = 17835,
  [18801] = 17887,
  [18802] = 17808,
  [18803] = 17727,
  [18804] = 17728,
  [18805] = 17729,
  [18806] = 17727,
  [18807] = 17730,
  [18808] = 18808,
  [18809] = 18809,
  [18810] = 17760,
  [18811] = 17728,
  [18812] = 18812,
  [18813] = 18813,
  [18814] = 17835,
  [18815] = 18369,
  [18816] = 17736,
  [18817] = 17735,
  [18818] = 17709,
  [18819] = 17820,
  [18820] = 17760,
  [18821] = 17754,
  [18822] = 17730,
  [18823] = 17740,
  [18824] = 17769,
  [18825] = 17713,
  [18826] = 2547,
  [18827] = 17729,
  [18828] = 17803,
  [18829] = 17713,
  [18830] = 17729,
  [18831] = 17810,
  [18832] = 17831,
  [18833] = 17711,
  [18834] = 8409,
  [18835] = 17761,
  [18836] = 17743,
  [18837] = 17808,
  [18838] = 17740,
  [18839] = 17760,
  [18840] = 17831,
  [18841] = 17835,
  [18842] = 8490,
  [18843] = 18197,
  [18844] = 17718,
  [18845] = 17747,
  [18846] = 17743,
  [18847] = 17810,
  [18848] = 18848,
  [18849] = 17723,
  [18850] = 18150,
  [18851] = 17754,
  [18852] = 17740,
  [18853] = 2550,
  [18854] = 17761,
  [18855] = 17727,
  [18856] = 17740,
  [18857] = 17769,
  [18858] = 17736,
  [18859] = 17728,
  [18860] = 17747,
  [18861] = 17803,
  [18862] = 17723,
  [18863] = 17712,
  [18864] = 17810,
  [18865] = 17831,
  [18866] = 17711,
  [18867] = 17769,
  [18868] = 17892,
  [18869] = 17730,
  [18870] = 18870,
  [18871] = 17754,
  [18872] = 17755,
  [18873] = 8319,
  [18874] = 17721,
  [18875] = 6919,
  [18876] = 17829,
  [18877] = 17727,
  [18878] = 17736,
  [18879] = 17728,
  [18880] = 17718,
  [18881] = 17796,
  [18882] = 17709,
  [18883] = 2551,
  [18884] = 17721,
  [18885] = 17730,
  [18886] = 17943,
  [18887] = 18756,
  [18888] = 2443,
  [18889] = 17761,
  [18890] = 17713,
  [18891] = 18158,
  [18892] = 18197,
  [18893] = 17831,
  [18894] = 17735,
  [18895] = 17791,
  [18896] = 17741,
  [18897] = 17803,
  [18898] = 17803,
  [18899] = 17796,
  [18900] = 17887,
  [18901] = 17791,
  [18902] = 17854,
  [18903] = 18756,
  [18904] = 17808,
  [18905] = 17808,
  [18906] = 17769,
  [18907] = 17740,
  [18908] = 17769,
  [18909] = 17723,
  [18910] = 17854,
  [18911] = 17796,
  [18912] = 17743,
  [18913] = 2552,
  [18914] = 18756,
  [18915] = 17736,
  [18916] = 17803,
  [18917] = 17736,
  [18918] = 17831,
  [18919] = 17709,
  [18920] = 8458,
  [18921] = 17810,
  [18922] = 17708,
  [18923] = 17796,
  [18924] = 2555,
  [18925] = 17831,
  [18926] = 18756,
  [18927] = 17711,
  [18928] = 17839,
  [18929] = 17743,
  [18930] = 17727,
  [18931] = 17728,
  [18932] = 17709,
  [18933] = 17735,
  [18934] = 17711,
  [18935] = 17796,
  [18936] = 17730,
  [18937] = 18937,
  [18938] = 18756,
  [18939] = 18939,
  [18940] = 17760,
  [18941] = 6971,
  [18942] = 17782,
  [18943] = 17741,
  [18944] = 2556,
  [18945] = 17803,
  [18946] = 17867,
  [18947] = 17712,
  [18948] = 17736,
  [18949] = 17998,
  [18950] = 17708,
  [18951] = 17709,
  [18952] = 2558,
  [18953] = 18953,
  [18954] = 18954,
  [18955] = 18955,
  [18956] = 17808,
  [18957] = 17713,
  [18958] = 17943,
  [18959] = 18681,
  [18960] = 2560,
  [18961] = 17735,
  [18962] = 17867,
  [18963] = 17740,
  [18964] = 2561,
  [18965] = 17741,
  [18966] = 2562,
  [18967] = 17797,
  [18968] = 8446,
  [18969] = 2563,
  [18970] = 17835,
  [18971] = 17803,
  [18972] = 17721,
  [18973] = 8787,
  [18974] = 17810,
  [18975] = 17831,
  [18976] = 17711,
  [18977] = 18977,
  [18978] = 17741,
  [18979] = 2503,
  [18980] = 17723,
  [18981] = 17810,
  [18982] = 8452,
  [18983] = 17743,
  [18984] = 17839,
  [18985] = 17743,
  [18986] = 17734,
  [18987] = 2504,
  [18988] = 17727,
  [18989] = 17831,
  [18990] = 17835,
  [18991] = 17747,
  [18992] = 17728,
  [18993] = 17728,
  [18994] = 17740,
  [18995] = 17730,
  [18996] = 18596,
  [18997] = 18132,
  [18998] = 2525,
  [18999] = 17721,
  [19000] = 2530,
  [19001] = 17735,
  [19002] = 17715,
  [19003] = 17736,
  [19004] = 17740,
  [19005] = 17747,
  [19006] = 17718,
  [19007] = 17754,
  [19008] = 17709,
  [19009] = 17747,
  [19010] = 17803,
  [19011] = 17761,
  [19012] = 17867,
  [19013] = 17810,
  [19014] = 17711,
  [19015] = 17803,
  [19016] = 19016,
  [19017] = 17769,
  [19018] = 8481,
  [19019] = 17716,
  [19020] = 17713,
  [19021] = 17717,
  [19022] = 8486,
  [19023] = 18937,
  [19024] = 17716,
  [19025] = 8555,
  [19026] = 17711,
  [19027] = 17715,
  [19028] = 17835,
  [19029] = 17729,
  [19030] = 8533,
  [19031] = 2535,
  [19032] = 8535,
  [19033] = 17810,
  [19034] = 17717,
  [19035] = 18809,
  [19036] = 19036,
  [19037] = 17803,
  [19038] = 17810,
  [19039] = 17810,
  [19040] = 17709,
  [19041] = 19041,
  [19042] = 17943,
  [19043] = 17723,
  [19044] = 17808,
  [19045] = 17734,
  [19046] = 17831,
  [19047] = 2536,
  [19048] = 17727,
  [19049] = 16612,
  [19050] = 17711,
  [19051] = 17743,
  [19052] = 17728,
  [19053] = 17835,
  [19054] = 17714,
  [19055] = 17820,
  [19056] = 17730,
  [19057] = 2540,
  [19058] = 17729,
  [19059] = 17760,
  [19060] = 17718,
  [19061] = 17734,
  [19062] = 19062,
  [19063] = 17740,
  [19064] = 18642,
  [19065] = 18158,
  [19066] = 17735,
  [19067] = 17711,
  [19068] = 17714,
  [19069] = 17713,
  [19070] = 17718,
  [19071] = 17736,
  [19072] = 17803,
  [19073] = 18197,
  [19074] = 17943,
  [19075] = 17810,
  [19076] = 17711,
  [19077] = 19077,
  [19078] = 17810,
  [19079] = 8330,
  [19080] = 17854,
  [19081] = 19081,
  [19082] = 19082,
  [19083] = 19083,
  [19084] = 19084,
  [19085] = 19085,
  [19086] = 19086,
  [19087] = 19087,
  [19088] = 19088,
  [19089] = 19089,
  [19090] = 19090,
  [19091] = 19091,
  [19092] = 19092,
  [19093] = 19093,
  [19094] = 19094,
  [19095] = 19095,
  [19096] = 19096,
  [19097] = 19097,
  [19098] = 19098,
  [19099] = 19099,
  [19100] = 19100,
  [19101] = 19101,
  [19102] = 19102,
  [19103] = 19103,
  [19104] = 19104,
  [19105] = 19105,
  [19106] = 19106,
  [19107] = 19107,
  [19108] = 19108,
  [19109] = 19109,
  [19110] = 19110,
  [19111] = 19111,
  [19112] = 19112,
  [19113] = 19113,
  [19114] = 19089,
  [19115] = 19115,
  [19116] = 19116,
  [19117] = 19117,
  [19118] = 19118,
  [19119] = 19119,
  [19120] = 19120,
  [19121] = 19121,
  [19122] = 19122,
  [19123] = 19123,
  [19124] = 19124,
  [19125] = 19125,
  [19126] = 19126,
  [19127] = 19097,
  [19128] = 19128,
  [19129] = 19129,
  [19130] = 19130,
  [19131] = 19131,
  [19132] = 19113,
  [19133] = 19133,
  [19134] = 19134,
  [19135] = 19135,
  [19136] = 19136,
  [19137] = 19096,
  [19138] = 19138,
  [19139] = 19139,
  [19140] = 19140,
  [19141] = 19141,
  [19142] = 19113,
  [19143] = 19143,
  [19144] = 19089,
  [19145] = 19145,
  [19146] = 19121,
  [19147] = 19147,
  [19148] = 19124,
  [19149] = 19089,
  [19150] = 19097,
  [19151] = 19086,
  [19152] = 19152,
  [19153] = 19153,
  [19154] = 19154,
  [19155] = 19155,
  [19156] = 19156,
  [19157] = 19086,
  [19158] = 19087,
  [19159] = 19088,
  [19160] = 19087,
  [19161] = 19088,
  [19162] = 19162,
  [19163] = 19103,
  [19164] = 19164,
  [19165] = 19165,
  [19166] = 19100,
  [19167] = 19167,
  [19168] = 19100,
  [19169] = 19169,
  [19170] = 19170,
  [19171] = 19171,
  [19172] = 19098,
  [19173] = 19173,
  [19174] = 19174,
  [19175] = 19104,
  [19176] = 19176,
  [19177] = 19124,
  [19178] = 19104,
  [19179] = 19111,
  [19180] = 19128,
  [19181] = 19131,
  [19182] = 19182,
  [19183] = 19097,
  [19184] = 19184,
  [19185] = 19115,
  [19186] = 19116,
  [19187] = 19187,
  [19188] = 19117,
  [19189] = 19118,
  [19190] = 19190,
  [19191] = 19191,
  [19192] = 19118,
  [19193] = 19193,
  [19194] = 19194,
  [19195] = 19195,
  [19196] = 19196,
  [19197] = 19197,
  [19198] = 19198,
  [19199] = 19083,
  [19200] = 19184,
  [19201] = 19121,
  [19202] = 19202,
  [19203] = 19203,
  [19204] = 19204,
  [19205] = 19096,
  [19206] = 19206,
  [19207] = 19091,
  [19208] = 19113,
  [19209] = 19093,
  [19210] = 19094,
  [19211] = 19089,
  [19212] = 19121,
  [19213] = 19124,
  [19214] = 19098,
  [19215] = 19100,
  [19216] = 19101,
  [19217] = 19097,
  [19218] = 19218,
  [19219] = 19107,
  [19220] = 19104,
  [19221] = 19086,
  [19222] = 19107,
  [19223] = 19087,
  [19224] = 19088,
  [19225] = 19225,
  [19226] = 19226,
  [19227] = 19112,
  [19228] = 19228,
  [19229] = 8330,
  [19230] = 19202,
  [19231] = 19203,
  [19232] = 19206,
  [19233] = 19100,
  [19234] = 19117,
  [19235] = 19120,
  [19236] = 19236,
  [19237] = 19104,
  [19238] = 19123,
  [19239] = 19136,
  [19240] = 19240,
  [19241] = 19126,
  [19242] = 19115,
  [19243] = 19116,
  [19244] = 19129,
  [19245] = 19164,
  [19246] = 19118,
  [19247] = 19115,
  [19248] = 19133,
  [19249] = 19134,
  [19250] = 19115,
  [19251] = 19117,
  [19252] = 19116,
  [19253] = 19138,
  [19254] = 19139,
  [19255] = 19116,
  [19256] = 19141,
  [19257] = 19118,
  [19258] = 19143,
  [19259] = 19190,
  [19260] = 19145,
  [19261] = 19118,
  [19262] = 19147,
  [19263] = 19191,
  [19264] = 19264,
  [19265] = 19086,
  [19266] = 19152,
  [19267] = 19153,
  [19268] = 19154,
  [19269] = 19155,
  [19270] = 8337,
  [19271] = 19156,
  [19272] = 19100,
  [19273] = 19273,
  [19274] = 19274,
  [19275] = 19104,
  [19276] = 19162,
  [19277] = 19096,
  [19278] = 19278,
  [19279] = 19165,
  [19280] = 19113,
  [19281] = 19089,
  [19282] = 19121,
  [19283] = 19169,
  [19284] = 19170,
  [19285] = 19124,
  [19286] = 19194,
  [19287] = 19173,
  [19288] = 19174,
  [19289] = 19097,
  [19290] = 19176,
  [19291] = 19106,
  [19292] = 19108,
  [19293] = 19110,
  [19294] = 19086,
  [19295] = 19087,
  [19296] = 19088,
  [19297] = 19119,
  [19298] = 19202,
  [19299] = 19274,
  [19300] = 19203,
  [19301] = 19187,
  [19302] = 19206,
  [19303] = 19303,
  [19304] = 19082,
  [19305] = 19195,
  [19306] = 19104,
  [19307] = 19193,
  [19308] = 19085,
  [19309] = 19090,
  [19310] = 19196,
  [19311] = 19092,
  [19312] = 19197,
  [19313] = 19083,
  [19314] = 19096,
  [19315] = 19315,
  [19316] = 19316,
  [19317] = 19115,
  [19318] = 19116,
  [19319] = 19102,
  [19320] = 19118,
  [19321] = 19091,
  [19322] = 19198,
  [19323] = 19093,
  [19324] = 19094,
  [19325] = 19105,
  [19326] = 19121,
  [19327] = 19315,
  [19328] = 19098,
  [19329] = 19115,
  [19330] = 19101,
  [19331] = 19116,
  [19332] = 19332,
  [19333] = 19113,
  [19334] = 19089,
  [19335] = 19117,
  [19336] = 19107,
  [19337] = 19118,
  [19338] = 2582,
  [19339] = 19096,
  [19340] = 19121,
  [19341] = 19112,
  [19342] = 19113,
  [19343] = 19089,
  [19344] = 19124,
  [19345] = 19345,
  [19346] = 19097,
  [19347] = 19124,
  [19348] = 19348,
  [19349] = 19120,
  [19350] = 2583,
  [19351] = 19097,
  [19352] = 19123,
  [19353] = 19086,
  [19354] = 19087,
  [19355] = 19126,
  [19356] = 19088,
  [19357] = 19357,
  [19358] = 19129,
  [19359] = 19359,
  [19360] = 19360,
  [19361] = 19124,
  [19362] = 19133,
  [19363] = 19134,
  [19364] = 19274,
  [19365] = 19365,
  [19366] = 19136,
  [19367] = 19138,
  [19368] = 19139,
  [19369] = 19369,
  [19370] = 19141,
  [19371] = 19371,
  [19372] = 19143,
  [19373] = 19104,
  [19374] = 19145,
  [19375] = 19086,
  [19376] = 19147,
  [19377] = 19152,
  [19378] = 19156,
  [19379] = 19087,
  [19380] = 19088,
  [19381] = 19153,
  [19382] = 19154,
  [19383] = 19155,
  [19384] = 19384,
  [19385] = 19164,
  [19386] = 19115,
  [19387] = 19116,
  [19388] = 19087,
  [19389] = 19118,
  [19390] = 19162,
  [19391] = 19088,
  [19392] = 19315,
  [19393] = 19165,
  [19394] = 19316,
  [19395] = 19082,
  [19396] = 19396,
  [19397] = 19169,
  [19398] = 19170,
  [19399] = 19085,
  [19400] = 19090,
  [19401] = 19173,
  [19402] = 19174,
  [19403] = 19092,
  [19404] = 19176,
  [19405] = 19405,
  [19406] = 19406,
  [19407] = 19097,
  [19408] = 19096,
  [19409] = 19371,
  [19410] = 19113,
  [19411] = 19089,
  [19412] = 19124,
  [19413] = 19122,
  [19414] = 19097,
  [19415] = 19187,
  [19416] = 19274,
  [19417] = 19164,
  [19418] = 19086,
  [19419] = 19087,
  [19420] = 19088,
  [19421] = 19193,
  [19422] = 19184,
  [19423] = 19096,
  [19424] = 19196,
  [19425] = 19082,
  [19426] = 19426,
  [19427] = 19083,
  [19428] = 19085,
  [19429] = 19090,
  [19430] = 19092,
  [19431] = 19154,
  [19432] = 19104,
  [19433] = 19096,
  [19434] = 19100,
  [19435] = 19091,
  [19436] = 19436,
  [19437] = 19093,
  [19438] = 19094,
  [19439] = 19439,
  [19440] = 19104,
  [19441] = 19439,
  [19442] = 19098,
  [19443] = 19443,
  [19444] = 19101,
  [19445] = 19115,
  [19446] = 19116,
  [19447] = 19118,
  [19448] = 19130,
  [19449] = 19202,
  [19450] = 19107,
  [19451] = 19203,
  [19452] = 19206,
  [19453] = 19103,
  [19454] = 19454,
  [19455] = 19112,
  [19456] = 19102,
  [19457] = 19454,
  [19458] = 19105,
  [19459] = 19102,
  [19460] = 19111,
  [19461] = 19096,
  [19462] = 19105,
  [19463] = 19120,
  [19464] = 19089,
  [19465] = 19124,
  [19466] = 19123,
  [19467] = 19115,
  [19468] = 19097,
  [19469] = 19126,
  [19470] = 19470,
  [19471] = 19116,
  [19472] = 19129,
  [19473] = 19087,
  [19474] = 19088,
  [19475] = 19475,
  [19476] = 19133,
  [19477] = 19134,
  [19478] = 19118,
  [19479] = 19086,
  [19480] = 19113,
  [19481] = 19138,
  [19482] = 19139,
  [19483] = 19089,
  [19484] = 19141,
  [19485] = 19117,
  [19486] = 19143,
  [19487] = 19316,
  [19488] = 19145,
  [19489] = 19121,
  [19490] = 19147,
  [19491] = 19104,
  [19492] = 19123,
  [19493] = 19096,
  [19494] = 19494,
  [19495] = 19153,
  [19496] = 19154,
  [19497] = 19155,
  [19498] = 19498,
  [19499] = 19136,
  [19500] = 19081,
  [19501] = 19113,
  [19502] = 19115,
  [19503] = 19116,
  [19504] = 19162,
  [19505] = 19118,
  [19506] = 19122,
  [19507] = 19165,
  [19508] = 19089,
  [19509] = 19117,
  [19510] = 19124,
  [19511] = 19274,
  [19512] = 19512,
  [19513] = 19173,
  [19514] = 19174,
  [19515] = 19097,
  [19516] = 19176,
  [19517] = 19184,
  [19518] = 19082,
  [19519] = 19121,
  [19520] = 19085,
  [19521] = 19096,
  [19522] = 19090,
  [19523] = 19187,
  [19524] = 19089,
  [19525] = 19124,
  [19526] = 19092,
  [19527] = 19097,
  [19528] = 19528,
  [19529] = 19193,
  [19530] = 19096,
  [19531] = 19087,
  [19532] = 19196,
  [19533] = 19088,
  [19534] = 19083,
  [19535] = 19091,
  [19536] = 19226,
  [19537] = 19093,
  [19538] = 19094,
  [19539] = 19303,
  [19540] = 19102,
  [19541] = 19098,
  [19542] = 19226,
  [19543] = 19105,
  [19544] = 19273,
  [19545] = 19086,
  [19546] = 19107,
  [19547] = 19136,
  [19548] = 19104,
  [19549] = 19549,
  [19550] = 19550,
  [19551] = 19113,
  [19552] = 19089,
  [19553] = 19117,
  [19554] = 19123,
  [19555] = 19303,
  [19556] = 19121,
  [19557] = 19115,
  [19558] = 19116,
  [19559] = 19118,
  [19560] = 19560,
  [19561] = 19124,
  [19562] = 19138,
  [19563] = 19141,
  [19564] = 19086,
  [19565] = 19097,
  [19566] = 19096,
  [19567] = 19147,
  [19568] = 8576,
  [19569] = 8577,
  [19570] = 19152,
  [19571] = 19156,
  [19572] = 19153,
  [19573] = 19154,
  [19574] = 19155,
  [19575] = 19096,
  [19576] = 19089,
  [19577] = 19124,
  [19578] = 19332,
  [19579] = 19097,
  [19580] = 19360,
  [19581] = 19162,
  [19582] = 19087,
  [19583] = 19088,
  [19584] = 19165,
  [19585] = 19106,
  [19586] = 19136,
  [19587] = 19108,
  [19588] = 19086,
  [19589] = 19152,
  [19590] = 19173,
  [19591] = 19174,
  [19592] = 19156,
  [19593] = 19176,
  [19594] = 19087,
  [19595] = 19104,
  [19596] = 19088,
  [19597] = 19110,
  [19598] = 19164,
  [19599] = 19119,
  [19600] = 19187,
  [19601] = 19406,
  [19602] = 19426,
  [19603] = 19436,
  [19604] = 19115,
  [19605] = 19116,
  [19606] = 19193,
  [19607] = 19118,
  [19608] = 19443,
  [19609] = 19196,
  [19610] = 19475,
  [19611] = 19083,
  [19612] = 19091,
  [19613] = 19152,
  [19614] = 19093,
  [19615] = 19094,
  [19616] = 19454,
  [19617] = 19087,
  [19618] = 19098,
  [19619] = 19088,
  [19620] = 19396,
  [19621] = 19621,
  [19622] = 19622,
  [19623] = 19107,
  [19624] = 19124,
  [19625] = 19096,
  [19626] = 19089,
  [19627] = 19124,
  [19628] = 19164,
  [19629] = 19097,
  [19630] = 19087,
  [19631] = 19123,
  [19632] = 19088,
  [19633] = 19512,
  [19634] = 19097,
  [19635] = 19184,
  [19636] = 19528,
  [19637] = 19156,
  [19638] = 19273,
  [19639] = 19138,
  [19640] = 19141,
  [19641] = 19130,
  [19642] = 19104,
  [19643] = 19278,
  [19644] = 19147,
  [19645] = 19084,
  [19646] = 19646,
  [19647] = 19303,
  [19648] = 19100,
  [19649] = 19153,
  [19650] = 19154,
  [19651] = 19155,
  [19652] = 19082,
  [19653] = 19348,
  [19654] = 19115,
  [19655] = 19116,
  [19656] = 19118,
  [19657] = 19162,
  [19658] = 19104,
  [19659] = 19136,
  [19660] = 19165,
  [19661] = 19084,
  [19662] = 19103,
  [19663] = 19202,
  [19664] = 19203,
  [19665] = 19206,
  [19666] = 19173,
  [19667] = 19174,
  [19668] = 19436,
  [19669] = 19176,
  [19670] = 19475,
  [19671] = 19111,
  [19672] = 19096,
  [19673] = 19089,
  [19674] = 19124,
  [19675] = 19086,
  [19676] = 19187,
  [19677] = 19097,
  [19678] = 19087,
  [19679] = 19088,
  [19680] = 19152,
  [19681] = 19357,
  [19682] = 19193,
  [19683] = 19156,
  [19684] = 19084,
  [19685] = 19196,
  [19686] = 19115,
  [19687] = 19083,
  [19688] = 19091,
  [19689] = 19116,
  [19690] = 19093,
  [19691] = 19094,
  [19692] = 19273,
  [19693] = 19104,
  [19694] = 19098,
  [19695] = 19118,
  [19696] = 19512,
  [19697] = 19697,
  [19698] = 19278,
  [19699] = 19107,
  [19700] = 19700,
  [19701] = 19109,
  [19702] = 19109,
  [19703] = 19115,
  [19704] = 19116,
  [19705] = 19118,
  [19706] = 19123,
  [19707] = 19316,
  [19708] = 19708,
  [19709] = 19125,
  [19710] = 19710,
  [19711] = 19711,
  [19712] = 19102,
  [19713] = 19138,
  [19714] = 19141,
  [19715] = 19274,
  [19716] = 19697,
  [19717] = 19147,
  [19718] = 19135,
  [19719] = 19125,
  [19720] = 19096,
  [19721] = 19153,
  [19722] = 19154,
  [19723] = 19155,
  [19724] = 19089,
  [19725] = 19124,
  [19726] = 19700,
  [19727] = 19097,
  [19728] = 19728,
  [19729] = 19162,
  [19730] = 19087,
  [19731] = 19088,
  [19732] = 19165,
  [19733] = 2583,
  [19734] = 19109,
  [19735] = 19136,
  [19736] = 19274,
  [19737] = 19708,
  [19738] = 19173,
  [19739] = 19174,
  [19740] = 19348,
  [19741] = 19176,
  [19742] = 19357,
  [19743] = 19082,
  [19744] = 19104,
  [19745] = 19359,
  [19746] = 19746,
  [19747] = 19085,
  [19748] = 19187,
  [19749] = 19090,
  [19750] = 19092,
  [19751] = 19128,
  [19752] = 19096,
  [19753] = 19115,
  [19754] = 19193,
  [19755] = 19116,
  [19756] = 19118,
  [19757] = 19196,
  [19758] = 19131,
  [19759] = 19083,
  [19760] = 19091,
  [19761] = 19135,
  [19762] = 19093,
  [19763] = 19094,
  [19764] = 19102,
  [19765] = 19711,
  [19766] = 19098,
  [19767] = 19105,
  [19768] = 19768,
  [19769] = 19190,
  [19770] = 19191,
  [19771] = 19107,
  [19772] = 19369,
  [19773] = 19194,
  [19774] = 19195,
  [19775] = 19096,
  [19776] = 19089,
  [19777] = 19124,
  [19778] = 19123,
  [19779] = 19113,
  [19780] = 19097,
  [19781] = 19087,
  [19782] = 19088,
  [19783] = 19089,
  [19784] = 19784,
  [19785] = 19138,
  [19786] = 19141,
  [19787] = 19117,
  [19788] = 19197,
  [19789] = 19147,
  [19790] = 19121,
  [19791] = 19198,
  [19792] = 19124,
  [19793] = 19153,
  [19794] = 19154,
  [19795] = 19155,
  [19796] = 19384,
  [19797] = 19104,
  [19798] = 19097,
  [19799] = 19162,
  [19800] = 19087,
  [19801] = 19801,
  [19802] = 19165,
  [19803] = 19088,
  [19804] = 19804,
  [19805] = 19805,
  [19806] = 19173,
  [19807] = 19174,
  [19808] = 19396,
  [19809] = 19176,
  [19810] = 19136,
  [19811] = 19115,
  [19812] = 19116,
  [19813] = 19118,
  [19814] = 19083,
  [19815] = 19815,
  [19816] = 19187,
  [19817] = 19086,
  [19818] = 19152,
  [19819] = 19156,
  [19820] = 19087,
  [19821] = 19193,
  [19822] = 19088,
  [19823] = 19823,
  [19824] = 19196,
  [19825] = 19164,
  [19826] = 19083,
  [19827] = 19091,
  [19828] = 19093,
  [19829] = 19094,
  [19830] = 19122,
  [19831] = 19096,
  [19832] = 19098,
  [19833] = 19089,
  [19834] = 19124,
  [19835] = 19164,
  [19836] = 19107,
  [19837] = 19097,
  [19838] = 19087,
  [19839] = 19088,
  [19840] = 19123,
  [19841] = 19841,
  [19842] = 19348,
  [19843] = 19357,
  [19844] = 19359,
  [19845] = 19162,
  [19846] = 19138,
  [19847] = 19141,
  [19848] = 18954,
  [19849] = 19130,
  [19850] = 19147,
  [19851] = 19103,
  [19852] = 19167,
  [19853] = 19104,
  [19854] = 19153,
  [19855] = 19154,
  [19856] = 19155,
  [19857] = 19111,
  [19858] = 19274,
  [19859] = 19369,
  [19860] = 19384,
  [19861] = 19162,
  [19862] = 19087,
  [19863] = 19099,
  [19864] = 19165,
  [19865] = 19088,
  [19866] = 19115,
  [19867] = 19116,
  [19868] = 19118,
  [19869] = 19173,
  [19870] = 19174,
  [19871] = 19128,
  [19872] = 19176,
  [19873] = 19100,
  [19874] = 19131,
  [19875] = 19303,
  [19876] = 19876,
  [19877] = 19104,
  [19878] = 19396,
  [19879] = 19187,
  [19880] = 8437,
  [19881] = 19218,
  [19882] = 19085,
  [19883] = 19096,
  [19884] = 19193,
  [19885] = 19089,
  [19886] = 19171,
  [19887] = 19196,
  [19888] = 19124,
  [19889] = 19083,
  [19890] = 19091,
  [19891] = 19093,
  [19892] = 19094,
  [19893] = 19105,
  [19894] = 19097,
  [19895] = 19098,
  [19896] = 19087,
  [19897] = 19088,
  [19898] = 19107,
  [19899] = 19100,
  [19900] = 19123,
  [19901] = 19190,
  [19902] = 19138,
  [19903] = 19141,
  [19904] = 19332,
  [19905] = 19147,
  [19906] = 19104,
  [19907] = 19202,
  [19908] = 19203,
  [19909] = 19153,
  [19910] = 19154,
  [19911] = 19155,
  [19912] = 19206,
  [19913] = 19104,
  [19914] = 19202,
  [19915] = 19203,
  [19916] = 19162,
  [19917] = 19206,
  [19918] = 19918,
  [19919] = 19165,
  [19920] = 19360,
  [19921] = 19406,
  [19922] = 19426,
  [19923] = 19436,
  [19924] = 19173,
  [19925] = 19174,
  [19926] = 19115,
  [19927] = 19176,
  [19928] = 19116,
  [19929] = 19118,
  [19930] = 19443,
  [19931] = 19475,
  [19932] = 19191,
  [19933] = 19122,
  [19934] = 19187,
  [19935] = 19528,
  [19936] = 19133,
  [19937] = 19115,
  [19938] = 19134,
  [19939] = 19193,
  [19940] = 19116,
  [19941] = 19121,
  [19942] = 19196,
  [19943] = 19115,
  [19944] = 19083,
  [19945] = 19091,
  [19946] = 19093,
  [19947] = 19094,
  [19948] = 19096,
  [19949] = 19089,
  [19950] = 19098,
  [19951] = 19124,
  [19952] = 19118,
  [19953] = 19097,
  [19954] = 19087,
  [19955] = 19088,
  [19956] = 19116,
  [19957] = 19332,
  [19958] = 19153,
  [19959] = 19154,
  [19960] = 19155,
  [19961] = 19512,
  [19962] = 19118,
  [19963] = 19528,
  [19964] = 19194,
  [19965] = 19162,
  [19966] = 19195,
  [19967] = 19104,
  [19968] = 19165,
  [19969] = 2582,
  [19970] = 19197,
  [19971] = 19084,
  [19972] = 19198,
  [19973] = 19173,
  [19974] = 19174,
  [19975] = 19975,
  [19976] = 19176,
  [19977] = 19164,
  [19978] = 19125,
  [19979] = 19115,
  [19980] = 19116,
  [19981] = 19118,
  [19982] = 19086,
  [19983] = 19187,
  [19984] = 19082,
  [19985] = 19315,
  [19986] = 19274,
  [19987] = 19193,
  [19988] = 19316,
  [19989] = 19697,
  [19990] = 19196,
  [19991] = 19700,
  [19992] = 19083,
  [19993] = 19091,
  [19994] = 19093,
  [19995] = 19082,
  [19996] = 19109,
  [19997] = 19098,
  [19998] = 19085,
  [19999] = 19096,
  [20000] = 19089,
  [20001] = 19124,
  [20002] = 19153,
  [20003] = 19154,
  [20004] = 19155,
  [20005] = 19090,
  [20006] = 19097,
  [20007] = 19087,
  [20008] = 19088,
  [20009] = 19162,
  [20010] = 19092,
  [20011] = 19549,
  [20012] = 19165,
  [20013] = 19096,
  [20014] = 19708,
  [20015] = 19711,
  [20016] = 19102,
  [20017] = 19173,
  [20018] = 19174,
  [20019] = 19085,
  [20020] = 19176,
  [20021] = 19104,
  [20022] = 19105,
  [20023] = 19130,
  [20024] = 19274,
  [20025] = 19135,
  [20026] = 19125,
  [20027] = 19187,
  [20028] = 19113,
  [20029] = 19115,
  [20030] = 19193,
  [20031] = 19116,
  [20032] = 19118,
  [20033] = 19196,
  [20034] = 19089,
  [20035] = 19083,
  [20036] = 19091,
  [20037] = 19093,
  [20038] = 19117,
  [20039] = 19103,
  [20040] = 19098,
  [20041] = 19121,
  [20042] = 19082,
  [20043] = 19124,
  [20044] = 20044,
  [20045] = 19153,
  [20046] = 19154,
  [20047] = 19155,
  [20048] = 20048,
  [20049] = 19111,
  [20050] = 19097,
  [20051] = 19085,
  [20052] = 19162,
  [20053] = 19096,
  [20054] = 19089,
  [20055] = 19165,
  [20056] = 19124,
  [20057] = 19090,
  [20058] = 20058,
  [20059] = 19097,
  [20060] = 19173,
  [20061] = 19174,
  [20062] = 19087,
  [20063] = 19176,
  [20064] = 19088,
  [20065] = 19092,
  [20066] = 19090,
  [20067] = 20067,
  [20068] = 19136,
  [20069] = 19096,
  [20070] = 19187,
  [20071] = 19086,
  [20072] = 19152,
  [20073] = 19104,
  [20074] = 19193,
  [20075] = 19156,
  [20076] = 2589,
  [20077] = 19196,
  [20078] = 19087,
  [20079] = 19083,
  [20080] = 19091,
  [20081] = 19093,
  [20082] = 19088,
  [20083] = 20083,
  [20084] = 19098,
  [20085] = 19135,
  [20086] = 19164,
  [20087] = 19815,
  [20088] = 19439,
  [20089] = 19153,
  [20090] = 19154,
  [20091] = 19155,
  [20092] = 19115,
  [20093] = 19116,
  [20094] = 19118,
  [20095] = 19226,
  [20096] = 19162,
  [20097] = 19815,
  [20098] = 19823,
  [20099] = 19165,
  [20100] = 19841,
  [20101] = 19454,
  [20102] = 19102,
  [20103] = 19823,
  [20104] = 19173,
  [20105] = 19174,
  [20106] = 19360,
  [20107] = 19176,
  [20108] = 19105,
  [20109] = 19841,
  [20110] = 19096,
  [20111] = 19089,
  [20112] = 19124,
  [20113] = 19092,
  [20114] = 19187,
  [20115] = 19097,
  [20116] = 19087,
  [20117] = 19088,
  [20118] = 19193,
  [20119] = 19104,
  [20120] = 8378,
  [20121] = 19196,
  [20122] = 19303,
  [20123] = 19083,
  [20124] = 19091,
  [20125] = 19093,
  [20126] = 19116,
  [20127] = 19118,
  [20128] = 19098,
  [20129] = 19371,
  [20130] = 19113,
  [20131] = 19100,
  [20132] = 19089,
  [20133] = 19153,
  [20134] = 19154,
  [20135] = 19155,
  [20136] = 19117,
  [20137] = 19104,
  [20138] = 19099,
  [20139] = 19121,
  [20140] = 19162,
  [20141] = 19096,
  [20142] = 19089,
  [20143] = 19165,
  [20144] = 19124,
  [20145] = 19303,
  [20146] = 19097,
  [20147] = 19087,
  [20148] = 19173,
  [20149] = 19174,
  [20150] = 19088,
  [20151] = 19176,
  [20152] = 19104,
  [20153] = 19202,
  [20154] = 19203,
  [20155] = 19116,
  [20156] = 19206,
  [20157] = 19128,
  [20158] = 19187,
  [20159] = 19165,
  [20160] = 19131,
  [20161] = 19406,
  [20162] = 19193,
  [20163] = 19115,
  [20164] = 19218,
  [20165] = 19196,
  [20166] = 19560,
  [20167] = 19083,
  [20168] = 19091,
  [20169] = 19093,
  [20170] = 19152,
  [20171] = 19096,
  [20172] = 19098,
  [20173] = 19124,
  [20174] = 19190,
  [20175] = 19097,
  [20176] = 19087,
  [20177] = 19153,
  [20178] = 19154,
  [20179] = 19155,
  [20180] = 19088,
  [20181] = 2588,
  [20182] = 19104,
  [20183] = 19115,
  [20184] = 19162,
  [20185] = 19116,
  [20186] = 19116,
  [20187] = 19165,
  [20188] = 8312,
  [20189] = 19124,
  [20190] = 19118,
  [20191] = 19126,
  [20192] = 19173,
  [20193] = 19174,
  [20194] = 19191,
  [20195] = 19176,
  [20196] = 19097,
  [20197] = 19138,
  [20198] = 20198,
  [20199] = 19184,
  [20200] = 19096,
  [20201] = 19124,
  [20202] = 19187,
  [20203] = 19194,
  [20204] = 19097,
  [20205] = 19087,
  [20206] = 19193,
  [20207] = 19088,
  [20208] = 19104,
  [20209] = 19196,
  [20210] = 20210,
  [20211] = 19083,
  [20212] = 19091,
  [20213] = 19093,
  [20214] = 19195,
  [20215] = 19116,
  [20216] = 19098,
  [20217] = 19197,
  [20218] = 20218,
  [20219] = 19198,
  [20220] = 19096,
  [20221] = 19153,
  [20222] = 19154,
  [20223] = 19155,
  [20224] = 19099,
  [20225] = 19136,
  [20226] = 19332,
  [20227] = 20227,
  [20228] = 19162,
  [20229] = 19096,
  [20230] = 19124,
  [20231] = 19165,
  [20232] = 19086,
  [20233] = 19097,
  [20234] = 19087,
  [20235] = 19088,
  [20236] = 19173,
  [20237] = 19174,
  [20238] = 19104,
  [20239] = 19176,
  [20240] = 19152,
  [20241] = 20241,
  [20242] = 19116,
  [20243] = 19274,
  [20244] = 19156,
  [20245] = 19359,
  [20246] = 19187,
  [20247] = 19360,
  [20248] = 19082,
  [20249] = 19426,
  [20250] = 19193,
  [20251] = 19439,
  [20252] = 19085,
  [20253] = 19196,
  [20254] = 19096,
  [20255] = 19083,
  [20256] = 19091,
  [20257] = 19093,
  [20258] = 19124,
  [20259] = 19090,
  [20260] = 19098,
  [20261] = 2589,
  [20262] = 19097,
  [20263] = 19087,
  [20264] = 19088,
  [20265] = 19153,
  [20266] = 19154,
  [20267] = 19155,
  [20268] = 19104,
  [20269] = 19092,
  [20270] = 19406,
  [20271] = 19116,
  [20272] = 19162,
  [20273] = 19096,
  [20274] = 19426,
  [20275] = 19165,
  [20276] = 19436,
  [20277] = 19102,
  [20278] = 19443,
  [20279] = 19105,
  [20280] = 19173,
  [20281] = 19174,
  [20282] = 19112,
  [20283] = 19176,
  [20284] = 19475,
  [20285] = 19096,
  [20286] = 19124,
  [20287] = 20287,
  [20288] = 19097,
  [20289] = 19087,
  [20290] = 19187,
  [20291] = 19088,
  [20292] = 19104,
  [20293] = 20293,
  [20294] = 19193,
  [20295] = 19129,
  [20296] = 19116,
  [20297] = 19196,
  [20298] = 19113,
  [20299] = 19083,
  [20300] = 19091,
  [20301] = 19093,
  [20302] = 19089,
  [20303] = 19117,
  [20304] = 19098,
  [20305] = 19128,
  [20306] = 19121,
  [20307] = 20307,
  [20308] = 19155,
  [20309] = 19153,
  [20310] = 19154,
  [20311] = 19155,
  [20312] = 19124,
  [20313] = 19096,
  [20314] = 19124,
  [20315] = 19512,
  [20316] = 19162,
  [20317] = 19097,
  [20318] = 19087,
  [20319] = 19165,
  [20320] = 19088,
  [20321] = 19104,
  [20322] = 19097,
  [20323] = 19528,
  [20324] = 19173,
  [20325] = 19174,
  [20326] = 19116,
  [20327] = 19176,
  [20328] = 19087,
  [20329] = 19088,
  [20330] = 19396,
  [20331] = 19100,
  [20332] = 19136,
  [20333] = 19167,
  [20334] = 19187,
  [20335] = 19086,
  [20336] = 19152,
  [20337] = 19096,
  [20338] = 19193,
  [20339] = 19124,
  [20340] = 19156,
  [20341] = 19196,
  [20342] = 19097,
  [20343] = 19083,
  [20344] = 19091,
  [20345] = 19093,
  [20346] = 19087,
  [20347] = 19088,
  [20348] = 19098,
  [20349] = 19104,
  [20350] = 19087,
  [20351] = 19088,
  [20352] = 19116,
  [20353] = 19153,
  [20354] = 19154,
  [20355] = 19155,
  [20356] = 19164,
  [20357] = 19164,
  [20358] = 20358,
  [20359] = 19171,
  [20360] = 19162,
  [20361] = 19131,
  [20362] = 19439,
  [20363] = 19165,
  [20364] = 20364,
  [20365] = 19096,
  [20366] = 19124,
  [20367] = 19104,
  [20368] = 19173,
  [20369] = 19174,
  [20370] = 19097,
  [20371] = 19176,
  [20372] = 19087,
  [20373] = 19088,
  [20374] = 19104,
  [20375] = 19218,
  [20376] = 19156,
  [20377] = 19116,
  [20378] = 19187,
  [20379] = 19084,
  [20380] = 19100,
  [20381] = 19167,
  [20382] = 19193,
  [20383] = 19303,
  [20384] = 19171,
  [20385] = 19196,
  [20386] = 19439,
  [20387] = 19083,
  [20388] = 19091,
  [20389] = 19093,
  [20390] = 19190,
  [20391] = 19104,
  [20392] = 19098,
  [20393] = 19096,
  [20394] = 19124,
  [20395] = 19191,
  [20396] = 19097,
  [20397] = 19153,
  [20398] = 19154,
  [20399] = 19155,
  [20400] = 19087,
  [20401] = 19088,
  [20402] = 19104,
  [20403] = 19100,
  [20404] = 19162,
  [20405] = 19436,
  [20406] = 19116,
  [20407] = 19165,
  [20408] = 19475,
  [20409] = 19104,
  [20410] = 19273,
  [20411] = 19082,
  [20412] = 19173,
  [20413] = 19174,
  [20414] = 19194,
  [20415] = 19176,
  [20416] = 19202,
  [20417] = 19203,
  [20418] = 19206,
  [20419] = 19096,
  [20420] = 19124,
  [20421] = 19202,
  [20422] = 19187,
  [20423] = 19097,
  [20424] = 19087,
  [20425] = 20425,
  [20426] = 19193,
  [20427] = 19088,
  [20428] = 19104,
  [20429] = 19196,
  [20430] = 19203,
  [20431] = 19083,
  [20432] = 19091,
  [20433] = 19093,
  [20434] = 19084,
  [20435] = 19116,
  [20436] = 19098,
  [20437] = 19202,
  [20438] = 19203,
  [20439] = 19206,
  [20440] = 19206,
  [20441] = 19153,
  [20442] = 19154,
  [20443] = 19155,
  [20444] = 19084,
  [20445] = 19697,
  [20446] = 19115,
  [20447] = 19116,
  [20448] = 19162,
  [20449] = 19096,
  [20450] = 19124,
  [20451] = 19165,
  [20452] = 19109,
  [20453] = 19097,
  [20454] = 19087,
  [20455] = 19088,
  [20456] = 19173,
  [20457] = 19174,
  [20458] = 19104,
  [20459] = 19176,
  [20460] = 19118,
  [20461] = 19700,
  [20462] = 19116,
  [20463] = 19278,
  [20464] = 19125,
  [20465] = 19109,
  [20466] = 19187,
  [20467] = 19195,
  [20468] = 19708,
  [20469] = 19711,
  [20470] = 19193,
  [20471] = 19135,
  [20472] = 19115,
  [20473] = 19196,
  [20474] = 19096,
  [20475] = 19083,
  [20476] = 19091,
  [20477] = 19093,
  [20478] = 20478,
  [20479] = 19124,
  [20480] = 19098,
  [20481] = 19116,
  [20482] = 19153,
  [20483] = 19154,
  [20484] = 19097,
  [20485] = 19087,
  [20486] = 19088,
  [20487] = 19165,
  [20488] = 19104,
  [20489] = 19173,
  [20490] = 19197,
  [20491] = 20491,
  [20492] = 19116,
  [20493] = 19187,
  [20494] = 19118,
  [20495] = 19115,
  [20496] = 19116,
  [20497] = 19196,
  [20498] = 19198,
  [20499] = 19091,
  [20500] = 19093,
  [20501] = 6967,
  [20502] = 19153,
  [20503] = 19154,
  [20504] = 19118,
  [20505] = 19128,
  [20506] = 19274,
  [20507] = 19165,
  [20508] = 19096,
  [20509] = 19173,
  [20510] = 19124,
  [20511] = 19131,
  [20512] = 19187,
  [20513] = 19097,
  [20514] = 19087,
  [20515] = 19088,
  [20516] = 19104,
  [20517] = 19091,
  [20518] = 19093,
  [20519] = 19153,
  [20520] = 19154,
  [20521] = 19332,
  [20522] = 20522,
  [20523] = 19116,
  [20524] = 19165,
  [20525] = 19125,
  [20526] = 19173,
  [20527] = 19082,
  [20528] = 19190,
  [20529] = 19187,
  [20530] = 19085,
  [20531] = 19090,
  [20532] = 19092,
  [20533] = 19191,
  [20534] = 19091,
  [20535] = 19093,
  [20536] = 19153,
  [20537] = 19154,
  [20538] = 19096,
  [20539] = 19096,
  [20540] = 19124,
  [20541] = 19165,
  [20542] = 19360,
  [20543] = 19173,
  [20544] = 19097,
  [20545] = 19087,
  [20546] = 19187,
  [20547] = 19088,
  [20548] = 19104,
  [20549] = 19194,
  [20550] = 19102,
  [20551] = 19091,
  [20552] = 19093,
  [20553] = 19153,
  [20554] = 19154,
  [20555] = 20555,
  [20556] = 19116,
  [20557] = 19195,
  [20558] = 19165,
  [20559] = 19105,
  [20560] = 19173,
  [20561] = 19197,
  [20562] = 20562,
  [20563] = 19187,
  [20564] = 19198,
  [20565] = 19369,
  [20566] = 19454,
  [20567] = 19549,
  [20568] = 19091,
  [20569] = 19093,
  [20570] = 19153,
  [20571] = 19154,
  [20572] = 19096,
  [20573] = 19124,
  [20574] = 19113,
  [20575] = 19165,
  [20576] = 19097,
  [20577] = 19173,
  [20578] = 19087,
  [20579] = 19088,
  [20580] = 19187,
  [20581] = 19104,
  [20582] = 19089,
  [20583] = 19117,
  [20584] = 19091,
  [20585] = 19093,
  [20586] = 19153,
  [20587] = 19154,
  [20588] = 17270,
  [20589] = 19116,
  [20590] = 19135,
  [20591] = 19165,
  [20592] = 19121,
  [20593] = 19173,
  [20594] = 19697,
  [20595] = 19406,
  [20596] = 19187,
  [20597] = 20597,
  [20598] = 19124,
  [20599] = 19815,
  [20600] = 19091,
  [20601] = 19093,
  [20602] = 19153,
  [20603] = 19154,
  [20604] = 19097,
  [20605] = 19096,
  [20606] = 19124,
  [20607] = 19315,
  [20608] = 19173,
  [20609] = 19097,
  [20610] = 19187,
  [20611] = 19087,
  [20612] = 19088,
  [20613] = 19104,
  [20614] = 19091,
  [20615] = 19093,
  [20616] = 19153,
  [20617] = 19154,
  [20618] = 19316,
  [20619] = 19823,
  [20620] = 19116,
  [20621] = 19841,
  [20622] = 19173,
  [20623] = 19102,
  [20624] = 19187,
  [20625] = 19136,
  [20626] = 20626,
  [20627] = 19700,
  [20628] = 19091,
  [20629] = 19093,
  [20630] = 19153,
  [20631] = 19086,
  [20632] = 19152,
  [20633] = 19156,
  [20634] = 19096,
  [20635] = 19173,
  [20636] = 19187,
  [20637] = 19124,
  [20638] = 19087,
  [20639] = 19091,
  [20640] = 19093,
  [20641] = 19153,
  [20642] = 19097,
  [20643] = 19087,
  [20644] = 19088,
  [20645] = 19104,
  [20646] = 19173,
  [20647] = 19187,
  [20648] = 19088,
  [20649] = 20649,
  [20650] = 19091,
  [20651] = 19093,
  [20652] = 19153,
  [20653] = 19116,
  [20654] = 19164,
  [20655] = 20655,
  [20656] = 19187,
  [20657] = 19105,
  [20658] = 19091,
  [20659] = 19153,
  [20660] = 19426,
  [20661] = 19436,
  [20662] = 19443,
  [20663] = 19187,
  [20664] = 19091,
  [20665] = 19153,
  [20666] = 20666,
  [20667] = 19274,
  [20668] = 19096,
  [20669] = 19187,
  [20670] = 19091,
  [20671] = 19153,
  [20672] = 19124,
  [20673] = 20673,
  [20674] = 19099,
  [20675] = 19187,
  [20676] = 19091,
  [20677] = 19153,
  [20678] = 19097,
  [20679] = 19087,
  [20680] = 19088,
  [20681] = 19187,
  [20682] = 19091,
  [20683] = 19153,
  [20684] = 19104,
  [20685] = 19475,
  [20686] = 19315,
  [20687] = 19187,
  [20688] = 19091,
  [20689] = 19153,
  [20690] = 19116,
  [20691] = 19100,
  [20692] = 19082,
  [20693] = 19187,
  [20694] = 19091,
  [20695] = 19153,
  [20696] = 20696,
  [20697] = 19104,
  [20698] = 19085,
  [20699] = 19187,
  [20700] = 19091,
  [20701] = 19153,
  [20702] = 19100,
  [20703] = 19090,
  [20704] = 19092,
  [20705] = 19187,
  [20706] = 19091,
  [20707] = 19153,
  [20708] = 19096,
  [20709] = 19124,
  [20710] = 19104,
  [20711] = 19187,
  [20712] = 19091,
  [20713] = 19153,
  [20714] = 19097,
  [20715] = 19106,
  [20716] = 19087,
  [20717] = 19187,
  [20718] = 19091,
  [20719] = 19088,
  [20720] = 19104,
  [20721] = 19202,
  [20722] = 19187,
  [20723] = 19091,
  [20724] = 20724,
  [20725] = 20725,
  [20726] = 19116,
  [20727] = 19187,
  [20728] = 19091,
  [20729] = 19203,
  [20730] = 19202,
  [20731] = 19203,
  [20732] = 19187,
  [20733] = 19091,
  [20734] = 19139,
  [20735] = 19206,
  [20736] = 19206,
  [20737] = 19187,
  [20738] = 19091,
  [20739] = 19316,
  [20740] = 19371,
  [20741] = 19096,
  [20742] = 19187,
  [20743] = 19091,
  [20744] = 19124,
  [20745] = 19128,
  [20746] = 19097,
  [20747] = 19187,
  [20748] = 19091,
  [20749] = 19087,
  [20750] = 19088,
  [20751] = 19104,
  [20752] = 19187,
  [20753] = 19091,
  [20754] = 19096,
  [20755] = 19109,
  [20756] = 19116,
  [20757] = 19187,
  [20758] = 19091,
  [20759] = 19131,
  [20760] = 19560,
  [20761] = 20761,
  [20762] = 19187,
  [20763] = 19091,
  [20764] = 19108,
  [20765] = 20649,
  [20766] = 19115,
  [20767] = 19187,
  [20768] = 19091,
  [20769] = 19116,
  [20770] = 19439,
  [20771] = 19118,
  [20772] = 19187,
  [20773] = 19091,
  [20774] = 19096,
  [20775] = 19124,
  [20776] = 19115,
  [20777] = 19187,
  [20778] = 19091,
  [20779] = 19097,
  [20780] = 19087,
  [20781] = 19088,
  [20782] = 19187,
  [20783] = 19091,
  [20784] = 19104,
  [20785] = 19116,
  [20786] = 19218,
  [20787] = 19187,
  [20788] = 19091,
  [20789] = 19116,
  [20790] = 19118,
  [20791] = 19136,
  [20792] = 19187,
  [20793] = 19091,
  [20794] = 19116,
  [20795] = 19454,
  [20796] = 19102,
  [20797] = 19187,
  [20798] = 19091,
  [20799] = 19190,
  [20800] = 19105,
  [20801] = 19191,
  [20802] = 19187,
  [20803] = 19091,
  [20804] = 19096,
  [20805] = 19124,
  [20806] = 19194,
  [20807] = 19187,
  [20808] = 19091,
  [20809] = 19097,
  [20810] = 19087,
  [20811] = 19088,
  [20812] = 19187,
  [20813] = 19091,
  [20814] = 19104,
  [20815] = 19274,
  [20816] = 19195,
  [20817] = 19187,
  [20818] = 19091,
  [20819] = 19116,
  [20820] = 19197,
  [20821] = 19113,
  [20822] = 19187,
  [20823] = 19091,
  [20824] = 19089,
  [20825] = 19274,
  [20826] = 19117,
  [20827] = 19187,
  [20828] = 19091,
  [20829] = 19198,
  [20830] = 19121,
  [20831] = 19082,
  [20832] = 19187,
  [20833] = 19091,
  [20834] = 19096,
  [20835] = 19124,
  [20836] = 8458,
  [20837] = 19187,
  [20838] = 19091,
  [20839] = 19097,
  [20840] = 19087,
  [20841] = 19088,
  [20842] = 19187,
  [20843] = 19091,
  [20844] = 19104,
  [20845] = 2367,
  [20846] = 19085,
  [20847] = 19187,
  [20848] = 19091,
  [20849] = 19090,
  [20850] = 19116,
  [20851] = 19092,
  [20852] = 19187,
  [20853] = 19091,
  [20854] = 19110,
  [20855] = 19274,
  [20856] = 19096,
  [20857] = 19187,
  [20858] = 19091,
  [20859] = 19082,
  [20860] = 19102,
  [20861] = 19708,
  [20862] = 19187,
  [20863] = 19091,
  [20864] = 19278,
  [20865] = 19096,
  [20866] = 19124,
  [20867] = 19187,
  [20868] = 19091,
  [20869] = 19119,
  [20870] = 19315,
  [20871] = 19105,
  [20872] = 19187,
  [20873] = 19091,
  [20874] = 19097,
  [20875] = 19087,
  [20876] = 19088,
  [20877] = 19187,
  [20878] = 19091,
  [20879] = 19104,
  [20880] = 20880,
  [20881] = 19085,
  [20882] = 19187,
  [20883] = 19091,
  [20884] = 19116,
  [20885] = 19082,
  [20886] = 19090,
  [20887] = 19187,
  [20888] = 19091,
  [20889] = 19085,
  [20890] = 19113,
  [20891] = 19089,
  [20892] = 19187,
  [20893] = 19091,
  [20894] = 19117,
  [20895] = 19436,
  [20896] = 19121,
  [20897] = 19187,
  [20898] = 19091,
  [20899] = 19096,
  [20900] = 19124,
  [20901] = 19090,
  [20902] = 19187,
  [20903] = 19091,
  [20904] = 19097,
  [20905] = 19087,
  [20906] = 19088,
  [20907] = 19187,
  [20908] = 19091,
  [20909] = 19104,
  [20910] = 19124,
  [20911] = 19092,
  [20912] = 19187,
  [20913] = 19091,
  [20914] = 19116,
  [20915] = 19097,
  [20916] = 19092,
  [20917] = 19187,
  [20918] = 19091,
  [20919] = 19096,
  [20920] = 19113,
  [20921] = 19089,
  [20922] = 19187,
  [20923] = 19091,
  [20924] = 19454,
  [20925] = 19102,
  [20926] = 19096,
  [20927] = 19187,
  [20928] = 19091,
  [20929] = 19124,
  [20930] = 19136,
  [20931] = 19097,
  [20932] = 19187,
  [20933] = 19091,
  [20934] = 19087,
  [20935] = 19088,
  [20936] = 19104,
  [20937] = 19187,
  [20938] = 19091,
  [20939] = 19384,
  [20940] = 19086,
  [20941] = 19116,
  [20942] = 19187,
  [20943] = 19091,
  [20944] = 19152,
  [20945] = 19156,
  [20946] = 19087,
  [20947] = 19187,
  [20948] = 19091,
  [20949] = 6953,
  [20950] = 20950,
  [20951] = 19088,
  [20952] = 19187,
  [20953] = 19091,
  [20954] = 19105,
  [20955] = 19164,
  [20956] = 19096,
  [20957] = 19187,
  [20958] = 19091,
  [20959] = 19124,
  [20960] = 20960,
  [20961] = 19097,
  [20962] = 19187,
  [20963] = 19091,
  [20964] = 19087,
  [20965] = 19088,
  [20966] = 19104,
  [20967] = 19187,
  [20968] = 19091,
  [20969] = 8354,
  [20970] = 19371,
  [20971] = 20971,
  [20972] = 19187,
  [20973] = 19091,
  [20974] = 19116,
  [20975] = 20975,
  [20976] = 19512,
  [20977] = 19187,
  [20978] = 19091,
  [20979] = 20979,
  [20980] = 19746,
  [20981] = 19113,
  [20982] = 19711,
  [20983] = 19089,
  [20984] = 19117,
  [20985] = 19096,
  [20986] = 19124,
  [20987] = 19124,
  [20988] = 19097,
  [20989] = 19087,
  [20990] = 19088,
  [20991] = 19104,
  [20992] = 19121,
  [20993] = 20993,
  [20994] = 19116,
  [20995] = 19117,
  [20996] = 19097,
  [20997] = 19184,
  [20998] = 19096,
  [20999] = 20999,
  [21000] = 21000,
  [21001] = 21001,
  [21002] = 19100,
  [21003] = 19124,
  [21004] = 19090,
  [21005] = 19096,
  [21006] = 19124,
  [21007] = 19104,
  [21008] = 19097,
  [21009] = 19240,
  [21010] = 21010,
  [21011] = 19087,
  [21012] = 19088,
  [21013] = 21013,
  [21014] = 19104,
  [21015] = 19097,
  [21016] = 19187,
  [21017] = 19116,
  [21018] = 19121,
  [21019] = 20364,
  [21020] = 19202,
  [21021] = 21021,
  [21022] = 21022,
  [21023] = 21023,
  [21024] = 21024,
  [21025] = 19203,
  [21026] = 19206,
  [21027] = 19528,
  [21028] = 19439,
  [21029] = 19226,
  [21030] = 19092,
  [21031] = 20993,
  [21032] = 19096,
  [21033] = 19124,
  [21034] = 19106,
  [21035] = 20044,
  [21036] = 20058,
  [21037] = 20083,
  [21038] = 19097,
  [21039] = 19087,
  [21040] = 19088,
  [21041] = 20478,
  [21042] = 19104,
  [21043] = 20673,
  [21044] = 19136,
  [21045] = 21045,
  [21046] = 21046,
  [21047] = 21047,
  [21048] = 19316,
  [21049] = 21049,
  [21050] = 19116,
  [21051] = 19236,
  [21052] = 21052,
  [21053] = 19136,
  [21054] = 19454,
  [21055] = 19225,
  [21056] = 19115,
  [21057] = 19804,
  [21058] = 19116,
  [21059] = 21059,
  [21060] = 19140,
  [21061] = 19086,
  [21062] = 19405,
  [21063] = 19118,
  [21064] = 19152,
  [21065] = 19494,
  [21066] = 21066,
  [21067] = 19470,
  [21068] = 20198,
  [21069] = 19096,
  [21070] = 19124,
  [21071] = 20491,
  [21072] = 19156,
  [21073] = 21045,
  [21074] = 21074,
  [21075] = 21075,
  [21076] = 19097,
  [21077] = 19087,
  [21078] = 19088,
  [21079] = 21079,
  [21080] = 19104,
  [21081] = 19102,
  [21082] = 19086,
  [21083] = 19116,
  [21084] = 19152,
  [21085] = 19156,
  [21086] = 19152,
  [21087] = 19125,
  [21088] = 19443,
  [21089] = 19105,
  [21090] = 19087,
  [21091] = 19101,
  [21092] = 19096,
  [21093] = 19124,
  [21094] = 19108,
  [21095] = 19097,
  [21096] = 19087,
  [21097] = 19088,
  [21098] = 21000,
  [21099] = 21001,
  [21100] = 19104,
  [21101] = 21101,
  [21102] = 19087,
  [21103] = 19116,
  [21104] = 19088,
  [21105] = 21049,
  [21106] = 19274,
  [21107] = 19240,
  [21108] = 21010,
  [21109] = 21013,
  [21110] = 19164,
  [21111] = 21052,
  [21112] = 19082,
  [21113] = 19084,
  [21114] = 19096,
  [21115] = 19124,
  [21116] = 19085,
  [21117] = 21021,
  [21118] = 21022,
  [21119] = 21023,
  [21120] = 21024,
  [21121] = 19097,
  [21122] = 19087,
  [21123] = 19088,
  [21124] = 19104,
  [21125] = 20993,
  [21126] = 19090,
  [21127] = 19092,
  [21128] = 19116,
  [21129] = 20083,
  [21130] = 19104,
  [21131] = 19096,
  [21132] = 19113,
  [21133] = 20478,
  [21134] = 19102,
  [21135] = 20673,
  [21136] = 19089,
  [21137] = 21046,
  [21138] = 21047,
  [21139] = 19105,
  [21140] = 21049,
  [21141] = 19173,
  [21142] = 21052,
  [21143] = 19117,
  [21144] = 19110,
  [21145] = 19225,
  [21146] = 19096,
  [21147] = 19804,
  [21148] = 19124,
  [21149] = 19140,
  [21150] = 19085,
  [21151] = 19405,
  [21152] = 19097,
  [21153] = 19087,
  [21154] = 19494,
  [21155] = 19088,
  [21156] = 19470,
  [21157] = 20198,
  [21158] = 19104,
  [21159] = 19369,
  [21160] = 20491,
  [21161] = 19303,
  [21162] = 21045,
  [21163] = 21074,
  [21164] = 21075,
  [21165] = 19116,
  [21166] = 19113,
  [21167] = 19089,
  [21168] = 19117,
  [21169] = 21000,
  [21170] = 21001,
  [21171] = 19087,
  [21172] = 19121,
  [21173] = 20649,
  [21174] = 19088,
  [21175] = 19124,
  [21176] = 19396,
  [21177] = 19096,
  [21178] = 19240,
  [21179] = 21010,
  [21180] = 19124,
  [21181] = 21013,
  [21182] = 19097,
  [21183] = 19097,
  [21184] = 19087,
  [21185] = 19088,
  [21186] = 19104,
  [21187] = 8149,
  [21188] = 19475,
  [21189] = 21021,
  [21190] = 21022,
  [21191] = 21023,
  [21192] = 21024,
  [21193] = 19116,
  [21194] = 19121,
  [21195] = 19119,
  [21196] = 19122,
  [21197] = 19136,
  [21198] = 19086,
  [21199] = 20083,
  [21200] = 19152,
  [21201] = 19156,
  [21202] = 19096,
  [21203] = 20478,
  [21204] = 19124,
  [21205] = 19087,
  [21206] = 21046,
  [21207] = 19097,
  [21208] = 21049,
  [21209] = 19097,
  [21210] = 21052,
  [21211] = 19087,
  [21212] = 19088,
  [21213] = 19225,
  [21214] = 19104,
  [21215] = 19804,
  [21216] = 19088,
  [21217] = 19140,
  [21218] = 19164,
  [21219] = 19405,
  [21220] = 19116,
  [21221] = 19164,
  [21222] = 19494,
  [21223] = 19226,
  [21224] = 19470,
  [21225] = 20198,
  [21226] = 19106,
  [21227] = 19108,
  [21228] = 20491,
  [21229] = 19110,
  [21230] = 21045,
  [21231] = 21074,
  [21232] = 21075,
  [21233] = 19100,
  [21234] = 19167,
  [21235] = 19125,
  [21236] = 21000,
  [21237] = 21001,
  [21238] = 19096,
  [21239] = 19124,
  [21240] = 19171,
  [21241] = 19097,
  [21242] = 19087,
  [21243] = 19100,
  [21244] = 19088,
  [21245] = 19240,
  [21246] = 21010,
  [21247] = 19104,
  [21248] = 21013,
  [21249] = 19119,
  [21250] = 19141,
  [21251] = 19116,
  [21252] = 19167,
  [21253] = 18812,
  [21254] = 21254,
  [21255] = 21021,
  [21256] = 21022,
  [21257] = 21023,
  [21258] = 21024,
  [21259] = 18151,
  [21260] = 21000,
  [21261] = 19303,
  [21262] = 21262,
  [21263] = 19171,
  [21264] = 19106,
  [21265] = 19164,
  [21266] = 20083,
  [21267] = 19096,
  [21268] = 19124,
  [21269] = 19100,
  [21270] = 20478,
  [21271] = 19097,
  [21272] = 19087,
  [21273] = 21046,
  [21274] = 21047,
  [21275] = 19088,
  [21276] = 21052,
  [21277] = 19104,
  [21278] = 19108,
  [21279] = 19225,
  [21280] = 19104,
  [21281] = 19804,
  [21282] = 19110,
  [21283] = 19140,
  [21284] = 19116,
  [21285] = 19405,
  [21286] = 19104,
  [21287] = 19119,
  [21288] = 19494,
  [21289] = 19104,
  [21290] = 19470,
  [21291] = 20198,
  [21292] = 19202,
  [21293] = 19203,
  [21294] = 20491,
  [21295] = 19206,
  [21296] = 21045,
  [21297] = 21074,
  [21298] = 21075,
  [21299] = 19436,
  [21300] = 19128,
  [21301] = 19096,
  [21302] = 21000,
  [21303] = 21001,
  [21304] = 19124,
  [21305] = 19131,
  [21306] = 19097,
  [21307] = 19087,
  [21308] = 19088,
  [21309] = 19104,
  [21310] = 19240,
  [21311] = 21010,
  [21312] = 19443,
  [21313] = 21013,
  [21314] = 19475,
  [21315] = 19116,
  [21316] = 19202,
  [21317] = 19203,
  [21318] = 19190,
  [21319] = 19191,
  [21320] = 21021,
  [21321] = 21022,
  [21322] = 21023,
  [21323] = 21024,
  [21324] = 8490,
  [21325] = 21001,
  [21326] = 19115,
  [21327] = 19206,
  [21328] = 19116,
  [21329] = 19194,
  [21330] = 20083,
  [21331] = 19118,
  [21332] = 19096,
  [21333] = 20478,
  [21334] = 19124,
  [21335] = 19195,
  [21336] = 21046,
  [21337] = 21047,
  [21338] = 21049,
  [21339] = 19097,
  [21340] = 21052,
  [21341] = 19087,
  [21342] = 19088,
  [21343] = 19225,
  [21344] = 19104,
  [21345] = 19804,
  [21346] = 19197,
  [21347] = 19140,
  [21348] = 19198,
  [21349] = 19405,
  [21350] = 19116,
  [21351] = 21351,
  [21352] = 19494,
  [21353] = 19697,
  [21354] = 19470,
  [21355] = 20198,
  [21356] = 19700,
  [21357] = 19109,
  [21358] = 20491,
  [21359] = 19084,
  [21360] = 21045,
  [21361] = 21074,
  [21362] = 21075,
  [21363] = 19100,
  [21364] = 19167,
  [21365] = 19096,
  [21366] = 21000,
  [21367] = 21001,
  [21368] = 19115,
  [21369] = 19116,
  [21370] = 19124,
  [21371] = 19124,
  [21372] = 19118,
  [21373] = 19097,
  [21374] = 19240,
  [21375] = 21010,
  [21376] = 19087,
  [21377] = 21013,
  [21378] = 19088,
  [21379] = 19104,
  [21380] = 19171,
  [21381] = 19097,
  [21382] = 19116,
  [21383] = 19109,
  [21384] = 21021,
  [21385] = 21022,
  [21386] = 21023,
  [21387] = 21024,
  [21388] = 18071,
  [21389] = 19184,
  [21390] = 19274,
  [21391] = 8408,
  [21392] = 19708,
  [21393] = 20083,
  [21394] = 19096,
  [21395] = 19124,
  [21396] = 20478,
  [21397] = 19125,
  [21398] = 19097,
  [21399] = 21046,
  [21400] = 21047,
  [21401] = 21049,
  [21402] = 19087,
  [21403] = 21052,
  [21404] = 19088,
  [21405] = 19104,
  [21406] = 19225,
  [21407] = 19082,
  [21408] = 19804,
  [21409] = 19104,
  [21410] = 19140,
  [21411] = 19116,
  [21412] = 19405,
  [21413] = 19085,
  [21414] = 19090,
  [21415] = 19494,
  [21416] = 19092,
  [21417] = 19470,
  [21418] = 20198,
  [21419] = 19124,
  [21420] = 19096,
  [21421] = 20491,
  [21422] = 19202,
  [21423] = 21045,
  [21424] = 21074,
  [21425] = 21075,
  [21426] = 19102,
  [21427] = 19096,
  [21428] = 19124,
  [21429] = 21000,
  [21430] = 21001,
  [21431] = 19203,
  [21432] = 19097,
  [21433] = 19087,
  [21434] = 21434,
  [21435] = 19088,
  [21436] = 19104,
  [21437] = 19240,
  [21438] = 21010,
  [21439] = 19105,
  [21440] = 21013,
  [21441] = 19135,
  [21442] = 19116,
  [21443] = 19206,
  [21444] = 20971,
  [21445] = 19097,
  [21446] = 19100,
  [21447] = 21021,
  [21448] = 21022,
  [21449] = 21023,
  [21450] = 21024,
  [21451] = 19113,
  [21452] = 19089,
  [21453] = 19117,
  [21454] = 19136,
  [21455] = 19096,
  [21456] = 19124,
  [21457] = 20083,
  [21458] = 19121,
  [21459] = 19097,
  [21460] = 20478,
  [21461] = 19087,
  [21462] = 19088,
  [21463] = 21046,
  [21464] = 21047,
  [21465] = 19104,
  [21466] = 21052,
  [21467] = 19184,
  [21468] = 19124,
  [21469] = 19225,
  [21470] = 19116,
  [21471] = 19804,
  [21472] = 19104,
  [21473] = 19140,
  [21474] = 19097,
  [21475] = 19405,
  [21476] = 19315,
  [21477] = 19711,
  [21478] = 19494,
  [21479] = 19135,
  [21480] = 19470,
  [21481] = 20198,
  [21482] = 19316,
  [21483] = 19202,
  [21484] = 20491,
  [21485] = 19086,
  [21486] = 21045,
  [21487] = 21074,
  [21488] = 21075,
  [21489] = 19202,
  [21490] = 19096,
  [21491] = 19124,
  [21492] = 21000,
  [21493] = 21001,
  [21494] = 19203,
  [21495] = 19097,
  [21496] = 19087,
  [21497] = 20649,
  [21498] = 21498,
  [21499] = 19240,
  [21500] = 21010,
  [21501] = 19088,
  [21502] = 21502,
  [21503] = 19104,
  [21504] = 19136,
  [21505] = 19206,
  [21506] = 19116,
  [21507] = 19086,
  [21508] = 19152,
  [21509] = 21074,
  [21510] = 19156,
  [21511] = 20478,
  [21512] = 19087,
  [21513] = 19143,
  [21514] = 21049,
  [21515] = 19088,
  [21516] = 21516,
  [21517] = 19096,
  [21518] = 19225,
  [21519] = 19124,
  [21520] = 19804,
  [21521] = 19164,
  [21522] = 19405,
  [21523] = 19097,
  [21524] = 19087,
  [21525] = 19494,
  [21526] = 19088,
  [21527] = 20198,
  [21528] = 19104,
  [21529] = 19152,
  [21530] = 21530,
  [21531] = 21045,
  [21532] = 21074,
  [21533] = 21075,
  [21534] = 19116,
  [21535] = 21046,
  [21536] = 19128,
  [21537] = 21000,
  [21538] = 21001,
  [21539] = 19131,
  [21540] = 19203,
  [21541] = 19156,
  [21542] = 19274,
  [21543] = 19273,
  [21544] = 19240,
  [21545] = 21010,
  [21546] = 19190,
  [21547] = 19115,
  [21548] = 19116,
  [21549] = 19096,
  [21550] = 19124,
  [21551] = 19191,
  [21552] = 19097,
  [21553] = 19087,
  [21554] = 19088,
  [21555] = 21047,
  [21556] = 20478,
  [21557] = 19104,
  [21558] = 19118,
  [21559] = 21049,
  [21560] = 20307,
  [21561] = 21561,
  [21562] = 19116,
  [21563] = 19225,
  [21564] = 19082,
  [21565] = 19804,
  [21566] = 19206,
  [21567] = 19405,
  [21568] = 19085,
  [21569] = 19194,
  [21570] = 19494,
  [21571] = 19815,
  [21572] = 20198,
  [21573] = 19090,
  [21574] = 19100,
  [21575] = 19195,
  [21576] = 21045,
  [21577] = 21074,
  [21578] = 21075,
  [21579] = 8481,
  [21580] = 19092,
  [21581] = 19197,
  [21582] = 21000,
  [21583] = 21001,
  [21584] = 19104,
  [21585] = 19096,
  [21586] = 19124,
  [21587] = 19371,
  [21588] = 19198,
  [21589] = 19240,
  [21590] = 21010,
  [21591] = 19097,
  [21592] = 19087,
  [21593] = 19088,
  [21594] = 19096,
  [21595] = 19104,
  [21596] = 19278,
  [21597] = 19202,
  [21598] = 19116,
  [21599] = 19203,
  [21600] = 19206,
  [21601] = 20478,
  [21602] = 19371,
  [21603] = 19100,
  [21604] = 21049,
  [21605] = 19167,
  [21606] = 19171,
  [21607] = 19115,
  [21608] = 19225,
  [21609] = 19439,
  [21610] = 19096,
  [21611] = 19405,
  [21612] = 19124,
  [21613] = 19116,
  [21614] = 19494,
  [21615] = 19097,
  [21616] = 20198,
  [21617] = 8486,
  [21618] = 19087,
  [21619] = 21045,
  [21620] = 21074,
  [21621] = 19088,
  [21622] = 19104,
  [21623] = 19113,
  [21624] = 21000,
  [21625] = 21001,
  [21626] = 19118,
  [21627] = 19116,
  [21628] = 19115,
  [21629] = 19116,
  [21630] = 19115,
  [21631] = 19240,
  [21632] = 21010,
  [21633] = 19118,
  [21634] = 21634,
  [21635] = 19454,
  [21636] = 21021,
  [21637] = 19274,
  [21638] = 19096,
  [21639] = 19124,
  [21640] = 19090,
  [21641] = 19097,
  [21642] = 19102,
  [21643] = 20478,
  [21644] = 19087,
  [21645] = 19088,
  [21646] = 21049,
  [21647] = 19104,
  [21648] = 19116,
  [21649] = 19116,
  [21650] = 19225,
  [21651] = 19105,
  [21652] = 19089,
  [21653] = 19405,
  [21654] = 19082,
  [21655] = 19118,
  [21656] = 19494,
  [21657] = 19085,
  [21658] = 20198,
  [21659] = 19090,
  [21660] = 19096,
  [21661] = 21045,
  [21662] = 21074,
  [21663] = 19124,
  [21664] = 19475,
  [21665] = 19097,
  [21666] = 21000,
  [21667] = 21001,
  [21668] = 19087,
  [21669] = 19088,
  [21670] = 19104,
  [21671] = 19113,
  [21672] = 19089,
  [21673] = 19240,
  [21674] = 21010,
  [21675] = 19116,
  [21676] = 19092,
  [21677] = 19104,
  [21678] = 19096,
  [21679] = 19124,
  [21680] = 19087,
  [21681] = 19104,
  [21682] = 19116,
  [21683] = 19348,
  [21684] = 21049,
  [21685] = 19117,
  [21686] = 19088,
  [21687] = 19274,
  [21688] = 19225,
  [21689] = 19124,
  [21690] = 19768,
  [21691] = 19405,
  [21692] = 19784,
  [21693] = 19124,
  [21694] = 19494,
  [21695] = 19102,
  [21696] = 20198,
  [21697] = 19121,
  [21698] = 19087,
  [21699] = 21045,
  [21700] = 21074,
  [21701] = 19104,
  [21702] = 19226,
  [21703] = 19116,
  [21704] = 21000,
  [21705] = 21001,
  [21706] = 19357,
  [21707] = 21075,
  [21708] = 19105,
  [21709] = 21709,
  [21710] = 19124,
  [21711] = 19240,
  [21712] = 21010,
  [21713] = 19082,
  [21714] = 19087,
  [21715] = 19104,
  [21716] = 19116,
  [21717] = 19359,
  [21718] = 19085,
  [21719] = 19090,
  [21720] = 19124,
  [21721] = 19092,
  [21722] = 19087,
  [21723] = 19104,
  [21724] = 19116,
  [21725] = 19225,
  [21726] = 19125,
  [21727] = 19494,
  [21728] = 19096,
  [21729] = 20198,
  [21730] = 19104,
  [21731] = 19116,
  [21732] = 21074,
  [21733] = 19106,
  [21734] = 19108,
  [21735] = 19092,
  [21736] = 21000,
  [21737] = 21001,
  [21738] = 19110,
  [21739] = 19119,
  [21740] = 21740,
  [21741] = 21741,
  [21742] = 19104,
  [21743] = 19240,
  [21744] = 21010,
  [21745] = 19116,
  [21746] = 19102,
  [21747] = 8319,
  [21748] = 19136,
  [21749] = 19104,
  [21750] = 19116,
  [21751] = 19105,
  [21752] = 19369,
  [21753] = 19104,
  [21754] = 19097,
  [21755] = 19124,
  [21756] = 19116,
  [21757] = 19113,
  [21758] = 21000,
  [21759] = 21001,
  [21760] = 19097,
  [21761] = 19184,
  [21762] = 19102,
  [21763] = 19104,
  [21764] = 8325,
  [21765] = 19240,
  [21766] = 21010,
  [21767] = 19116,
  [21768] = 19089,
  [21769] = 19117,
  [21770] = 19104,
  [21771] = 19116,
  [21772] = 19384,
  [21773] = 19113,
  [21774] = 19104,
  [21775] = 19116,
  [21776] = 19089,
  [21777] = 19117,
  [21778] = 21000,
  [21779] = 21001,
  [21780] = 19116,
  [21781] = 19136,
  [21782] = 19823,
  [21783] = 19116,
  [21784] = 19121,
  [21785] = 19240,
  [21786] = 21010,
  [21787] = 19116,
  [21788] = 19116,
  [21789] = 19086,
  [21790] = 19116,
  [21791] = 19116,
  [21792] = 19116,
  [21793] = 19116,
  [21794] = 19116,
  [21795] = 19549,
  [21796] = 19560,
  [21797] = 19116,
  [21798] = 21000,
  [21799] = 21001,
  [21800] = 19116,
  [21801] = 19152,
  [21802] = 19116,
  [21803] = 19116,
  [21804] = 19156,
  [21805] = 19240,
  [21806] = 21010,
  [21807] = 19116,
  [21808] = 19273,
  [21809] = 19116,
  [21810] = 19278,
  [21811] = 19116,
  [21812] = 19116,
  [21813] = 19116,
  [21814] = 19116,
  [21815] = 19116,
  [21816] = 19116,
  [21817] = 19116,
  [21818] = 21000,
  [21819] = 21001,
  [21820] = 19116,
  [21821] = 19116,
  [21822] = 19116,
  [21823] = 19116,
  [21824] = 19116,
  [21825] = 19240,
  [21826] = 21010,
  [21827] = 19116,
  [21828] = 19116,
  [21829] = 19116,
  [21830] = 19116,
  [21831] = 19116,
  [21832] = 19116,
  [21833] = 19116,
  [21834] = 19116,
  [21835] = 19348,
  [21836] = 19357,
  [21837] = 19116,
  [21838] = 21000,
  [21839] = 21001,
  [21840] = 19116,
  [21841] = 19116,
  [21842] = 19116,
  [21843] = 19359,
  [21844] = 19121,
  [21845] = 19240,
  [21846] = 21010,
  [21847] = 19087,
  [21848] = 19145,
  [21849] = 19369,
  [21850] = 19384,
  [21851] = 19124,
  [21852] = 19087,
  [21853] = 19153,
  [21854] = 19088,
  [21855] = 19396,
  [21856] = 19154,
  [21857] = 19155,
  [21858] = 21000,
  [21859] = 21001,
  [21860] = 19100,
  [21861] = 21861,
  [21862] = 19841,
  [21863] = 19122,
  [21864] = 19097,
  [21865] = 19240,
  [21866] = 21010,
  [21867] = 19124,
  [21868] = 19088,
  [21869] = 19097,
  [21870] = 19164,
  [21871] = 21871,
  [21872] = 19396,
  [21873] = 19136,
  [21874] = 19162,
  [21875] = 21875,
  [21876] = 19086,
  [21877] = 19165,
  [21878] = 21000,
  [21879] = 21001,
  [21880] = 19152,
  [21881] = 19156,
  [21882] = 19087,
  [21883] = 21883,
  [21884] = 19088,
  [21885] = 19240,
  [21886] = 21010,
  [21887] = 19104,
  [21888] = 19169,
  [21889] = 19130,
  [21890] = 19170,
  [21891] = 19164,
  [21892] = 19103,
  [21893] = 19176,
  [21894] = 19111,
  [21895] = 19348,
  [21896] = 21079,
  [21897] = 19136,
  [21898] = 21000,
  [21899] = 21001,
  [21900] = 19264,
  [21901] = 19173,
  [21902] = 19174,
  [21903] = 21903,
  [21904] = 19176,
  [21905] = 19240,
  [21906] = 21010,
  [21907] = 19122,
  [21908] = 19086,
  [21909] = 19152,
  [21910] = 19156,
  [21911] = 19164,
  [21912] = 19303,
  [21913] = 19087,
  [21914] = 19088,
  [21915] = 19357,
  [21916] = 19086,
  [21917] = 19100,
  [21918] = 21000,
  [21919] = 21001,
  [21920] = 19164,
  [21921] = 19202,
  [21922] = 19104,
  [21923] = 19187,
  [21924] = 19332,
  [21925] = 19240,
  [21926] = 21010,
  [21927] = 19203,
  [21928] = 19206,
  [21929] = 19360,
  [21930] = 19406,
  [21931] = 19202,
  [21932] = 19426,
  [21933] = 19203,
  [21934] = 19436,
  [21935] = 19443,
  [21936] = 19475,
  [21937] = 19206,
  [21938] = 21000,
  [21939] = 21001,
  [21940] = 19193,
  [21941] = 21941,
  [21942] = 19303,
  [21943] = 19196,
  [21944] = 19512,
  [21945] = 19240,
  [21946] = 21010,
  [21947] = 19549,
  [21948] = 19560,
  [21949] = 19746,
  [21950] = 19083,
  [21951] = 19152,
  [21952] = 19528,
  [21953] = 19130,
  [21954] = 19103,
  [21955] = 19111,
  [21956] = 19115,
  [21957] = 19116,
  [21958] = 21000,
  [21959] = 21001,
  [21960] = 19359,
  [21961] = 19118,
  [21962] = 19156,
  [21963] = 19091,
  [21964] = 19084,
  [21965] = 19240,
  [21966] = 21010,
  [21967] = 19115,
  [21968] = 19093,
  [21969] = 19094,
  [21970] = 19116,
  [21971] = 19273,
  [21972] = 19100,
  [21973] = 19118,
  [21974] = 19315,
  [21975] = 19697,
  [21976] = 19700,
  [21977] = 19278,
  [21978] = 21000,
  [21979] = 21001,
  [21980] = 19109,
  [21981] = 19105,
  [21982] = 19098,
  [21983] = 19708,
  [21984] = 19135,
  [21985] = 19240,
  [21986] = 21010,
  [21987] = 19315,
  [21988] = 20971,
  [21989] = 19711,
  [21990] = 19101,
  [21991] = 19315,
  [21992] = 19303,
  [21993] = 19316,
  [21994] = 21883,
  [21995] = 21001,
  [21996] = 21010,
  [21997] = 19316,
  [21998] = 19125,
  [21999] = 19107,
  [22000] = 19124,
  [22001] = 19815,
  [22002] = 21001,
  [22003] = 21010,
  [22004] = 19274,
  [22005] = 19202,
  [22006] = 19332,
  [22007] = 19112,
  [22008] = 21001,
  [22009] = 21010,
  [22010] = 19360,
  [22011] = 19135,
  [22012] = 19082,
  [22013] = 19406,
  [22014] = 21001,
  [22015] = 21010,
  [22016] = 19085,
  [22017] = 19815,
  [22018] = 19823,
  [22019] = 19090,
  [22020] = 21001,
  [22021] = 21010,
  [22022] = 19092,
  [22023] = 19841,
  [22024] = 19315,
  [22025] = 19096,
  [22026] = 21001,
  [22027] = 21010,
  [22028] = 19316,
  [22029] = 19102,
  [22030] = 19426,
  [22031] = 21001,
  [22032] = 21010,
  [22033] = 19105,
  [22034] = 19120,
  [22035] = 19436,
  [22036] = 19147,
  [22037] = 19123,
  [22038] = 19100,
  [22039] = 19443,
  [22040] = 19126,
  [22041] = 19475,
  [22042] = 19104,
  [22043] = 19129,
  [22044] = 19113,
  [22045] = 19089,
  [22046] = 19117,
  [22047] = 19133,
  [22048] = 19134,
  [22049] = 20649,
  [22050] = 19121,
  [22051] = 19203,
  [22052] = 19138,
  [22053] = 19139,
  [22054] = 19124,
  [22055] = 19141,
  [22056] = 19512,
  [22057] = 19143,
  [22058] = 22058,
  [22059] = 19145,
  [22060] = 19097,
  [22061] = 19147,
  [22062] = 19202,
  [22063] = 19203,
  [22064] = 19206,
  [22065] = 19193,
  [22066] = 19202,
  [22067] = 19528,
  [22068] = 19206,
  [22069] = 19169,
  [22070] = 19099,
  [22071] = 19136,
  [22072] = 19086,
  [22073] = 19152,
  [22074] = 19156,
  [22075] = 19087,
  [22076] = 19088,
  [22077] = 19274,
  [22078] = 19164,
  [22079] = 22079,
  [22080] = 19823,
  [22081] = 19841,
  [22082] = 19153,
  [22083] = 19154,
  [22084] = 19155,
  [22085] = 19082,
  [22086] = 19084,
  [22087] = 19115,
  [22088] = 8533,
  [22089] = 8535,
  [22090] = 19116,
  [22091] = 19085,
  [22092] = 19118,
  [22093] = 19203,
  [22094] = 19162,
  [22095] = 19100,
  [22096] = 19090,
  [22097] = 19165,
  [22098] = 19092,
  [22099] = 19128,
  [22100] = 19104,
  [22101] = 19131,
  [22102] = 19169,
  [22103] = 19170,
  [22104] = 19369,
  [22105] = 19096,
  [22106] = 19274,
  [22107] = 21079,
  [22108] = 19274,
  [22109] = 19173,
  [22110] = 19174,
  [22111] = 19174,
  [22112] = 19176,
  [22113] = 21010,
  [22114] = 19439,
  [22115] = 19384,
  [22116] = 19115,
  [22117] = 19116,
  [22118] = 19697,
  [22119] = 19118,
  [22120] = 19700,
  [22121] = 19454,
  [22122] = 19102,
  [22123] = 19187,
  [22124] = 19218,
  [22125] = 19109,
  [22126] = 19105,
  [22127] = 19087,
  [22128] = 19274,
  [22129] = 19193,
  [22130] = 19708,
  [22131] = 19711,
  [22132] = 19196,
  [22133] = 19082,
  [22134] = 19082,
  [22135] = 19083,
  [22136] = 19348,
  [22137] = 21013,
  [22138] = 19085,
  [22139] = 19090,
  [22140] = 19092,
  [22141] = 19096,
  [22142] = 19113,
  [22143] = 19091,
  [22144] = 19096,
  [22145] = 19093,
  [22146] = 19094,
  [22147] = 19089,
  [22148] = 19113,
  [22149] = 19089,
  [22150] = 19190,
  [22151] = 19121,
  [22152] = 19191,
  [22153] = 19098,
  [22154] = 19124,
  [22155] = 19117,
  [22156] = 19101,
  [22157] = 19097,
  [22158] = 19102,
  [22159] = 22159,
  [22160] = 19105,
  [22161] = 19085,
  [22162] = 19107,
  [22163] = 19194,
  [22164] = 19086,
  [22165] = 19087,
  [22166] = 19088,
  [22167] = 19195,
  [22168] = 19112,
  [22169] = 19121,
  [22170] = 19125,
  [22171] = 19090,
  [22172] = 19092,
  [22173] = 19876,
  [22174] = 19622,
  [22175] = 19975,
  [22176] = 19113,
  [22177] = 19099,
  [22178] = 19550,
  [22179] = 19089,
  [22180] = 20227,
  [22181] = 19117,
  [22182] = 19100,
  [22183] = 19128,
  [22184] = 19357,
  [22185] = 19131,
  [22186] = 19121,
  [22187] = 22187,
  [22188] = 22188,
  [22189] = 19218,
  [22190] = 20210,
  [22191] = 19120,
  [22192] = 19104,
  [22193] = 19082,
  [22194] = 19123,
  [22195] = 19190,
  [22196] = 19124,
  [22197] = 19191,
  [22198] = 19135,
  [22199] = 20048,
  [22200] = 19126,
  [22201] = 19194,
  [22202] = 19097,
  [22203] = 21434,
  [22204] = 19345,
  [22205] = 21022,
  [22206] = 19129,
  [22207] = 19195,
  [22208] = 20761,
  [22209] = 19197,
  [22210] = 19198,
  [22211] = 19371,
  [22212] = 19815,
  [22213] = 19823,
  [22214] = 20364,
  [22215] = 20044,
  [22216] = 20058,
  [22217] = 19236,
  [22218] = 21059,
  [22219] = 19876,
  [22220] = 19622,
  [22221] = 19975,
  [22222] = 19133,
  [22223] = 19134,
  [22224] = 19550,
  [22225] = 20227,
  [22226] = 19115,
  [22227] = 19116,
  [22228] = 19124,
  [22229] = 22187,
  [22230] = 22188,
  [22231] = 20210,
  [22232] = 19138,
  [22233] = 19139,
  [22234] = 19118,
  [22235] = 19141,
  [22236] = 20048,
  [22237] = 19841,
  [22238] = 21434,
  [22239] = 19345,
  [22240] = 19143,
  [22241] = 19136,
  [22242] = 19091,
  [22243] = 20761,
  [22244] = 20364,
  [22245] = 20044,
  [22246] = 20058,
  [22247] = 19236,
  [22248] = 21059,
  [22249] = 19876,
  [22250] = 19622,
  [22251] = 19975,
  [22252] = 19145,
  [22253] = 19097,
  [22254] = 19550,
  [22255] = 20227,
  [22256] = 19147,
  [22257] = 19184,
  [22258] = 19086,
  [22259] = 22187,
  [22260] = 20210,
  [22261] = 19152,
  [22262] = 19156,
  [22263] = 19359,
  [22264] = 20048,
  [22265] = 19087,
  [22266] = 21434,
  [22267] = 19345,
  [22268] = 19153,
  [22269] = 19154,
  [22270] = 19155,
  [22271] = 20761,
  [22272] = 20364,
  [22273] = 20044,
  [22274] = 20058,
  [22275] = 19236,
  [22276] = 21059,
  [22277] = 19876,
  [22278] = 19622,
  [22279] = 19975,
  [22280] = 19088,
  [22281] = 19085,
  [22282] = 19550,
  [22283] = 20227,
  [22284] = 19090,
  [22285] = 19164,
  [22286] = 19096,
  [22287] = 22187,
  [22288] = 20210,
  [22289] = 19096,
  [22290] = 19162,
  [22291] = 19113,
  [22292] = 20048,
  [22293] = 19197,
  [22294] = 21434,
  [22295] = 19345,
  [22296] = 19165,
  [22297] = 19089,
  [22298] = 19198,
  [22299] = 20761,
  [22300] = 20364,
  [22301] = 20044,
  [22302] = 20058,
  [22303] = 19236,
  [22304] = 21059,
  [22305] = 19876,
  [22306] = 19622,
  [22307] = 19975,
  [22308] = 19121,
  [22309] = 19124,
  [22310] = 19550,
  [22311] = 20227,
  [22312] = 19169,
  [22313] = 19170,
  [22314] = 19092,
  [22315] = 22187,
  [22316] = 20210,
  [22317] = 19097,
  [22318] = 19173,
  [22319] = 19174,
  [22320] = 20048,
  [22321] = 19088,
  [22322] = 21434,
  [22323] = 19345,
  [22324] = 19176,
  [22325] = 21059,
  [22326] = 19115,
  [22327] = 20761,
  [22328] = 19876,
  [22329] = 19622,
  [22330] = 19975,
  [22331] = 19086,
  [22332] = 19087,
  [22333] = 19550,
  [22334] = 20227,
  [22335] = 19088,
  [22336] = 21023,
  [22337] = 22187,
  [22338] = 19439,
  [22339] = 21434,
  [22340] = 19196,
  [22341] = 19130,
  [22342] = 19099,
  [22343] = 20761,
  [22344] = 19876,
  [22345] = 19622,
  [22346] = 19975,
  [22347] = 19187,
  [22348] = 19136,
  [22349] = 19550,
  [22350] = 20227,
  [22351] = 19396,
  [22352] = 19100,
  [22353] = 22187,
  [22354] = 19371,
  [22355] = 21434,
  [22356] = 19086,
  [22357] = 19193,
  [22358] = 19104,
  [22359] = 20761,
  [22360] = 19876,
  [22361] = 19622,
  [22362] = 19975,
  [22363] = 22188,
  [22364] = 19550,
  [22365] = 20227,
  [22366] = 19152,
  [22367] = 22187,
  [22368] = 19196,
  [22369] = 21434,
  [22370] = 19120,
  [22371] = 20761,
  [22372] = 19876,
  [22373] = 19622,
  [22374] = 19975,
  [22375] = 19156,
  [22376] = 19550,
  [22377] = 20227,
  [22378] = 19083,
  [22379] = 22187,
  [22380] = 21024,
  [22381] = 21434,
  [22382] = 22382,
  [22383] = 20761,
  [22384] = 19876,
  [22385] = 19622,
  [22386] = 19975,
  [22387] = 8469,
  [22388] = 19550,
  [22389] = 20227,
  [22390] = 19128,
  [22391] = 22187,
  [22392] = 8555,
  [22393] = 21434,
  [22394] = 19876,
  [22395] = 19622,
  [22396] = 19975,
  [22397] = 8558,
  [22398] = 19550,
  [22399] = 20227,
  [22400] = 19131,
  [22401] = 22187,
  [22402] = 19170,
  [22403] = 21434,
  [22404] = 19876,
  [22405] = 19622,
  [22406] = 19975,
  [22407] = 19091,
  [22408] = 19550,
  [22409] = 20227,
  [22410] = 19369,
  [22411] = 22187,
  [22412] = 8309,
  [22413] = 21434,
  [22414] = 19876,
  [22415] = 19622,
  [22416] = 19975,
  [22417] = 8315,
  [22418] = 19550,
  [22419] = 20227,
  [22420] = 19093,
  [22421] = 22187,
  [22422] = 19094,
  [22423] = 19876,
  [22424] = 19622,
  [22425] = 19975,
  [22426] = 19218,
  [22427] = 19550,
  [22428] = 20227,
  [22429] = 19115,
  [22430] = 22187,
  [22431] = 19206,
  [22432] = 19876,
  [22433] = 19622,
  [22434] = 19975,
  [22435] = 19116,
  [22436] = 19550,
  [22437] = 20227,
  [22438] = 19384,
  [22439] = 22187,
  [22440] = 19098,
  [22441] = 19876,
  [22442] = 19622,
  [22443] = 19975,
  [22444] = 19118,
  [22445] = 19550,
  [22446] = 20227,
  [22447] = 19454,
  [22448] = 22187,
  [22449] = 8361,
  [22450] = 19876,
  [22451] = 19622,
  [22452] = 19975,
  [22453] = 19101,
  [22454] = 19550,
  [22455] = 20227,
  [22456] = 19190,
  [22457] = 22187,
  [22458] = 8372,
  [22459] = 19876,
  [22460] = 19622,
  [22461] = 19975,
  [22462] = 19191,
  [22463] = 19550,
  [22464] = 20227,
  [22465] = 19102,
  [22466] = 22187,
  [22467] = 19194,
  [22468] = 19876,
  [22469] = 19622,
  [22470] = 19975,
  [22471] = 19195,
  [22472] = 19550,
  [22473] = 20227,
  [22474] = 19107,
  [22475] = 22187,
  [22476] = 19100,
  [22477] = 19876,
  [22478] = 19622,
  [22479] = 19975,
  [22480] = 19550,
  [22481] = 20227,
  [22482] = 19197,
  [22483] = 22187,
  [22484] = 19198,
  [22485] = 19876,
  [22486] = 19622,
  [22487] = 19975,
  [22488] = 19184,
  [22489] = 19550,
  [22490] = 20227,
  [22491] = 19112,
  [22492] = 22187,
  [22493] = 19104,
  [22494] = 19876,
  [22495] = 19622,
  [22496] = 19975,
  [22497] = 19369,
  [22498] = 19550,
  [22499] = 20227,
  [22500] = 19087,
  [22501] = 22187,
  [22502] = 19096,
  [22503] = 19876,
  [22504] = 19622,
  [22505] = 19975,
  [22506] = 19088,
  [22507] = 19550,
  [22508] = 20227,
  [22509] = 19113,
  [22510] = 22187,
  [22511] = 19089,
  [22512] = 19876,
  [22513] = 19622,
  [22514] = 19975,
  [22515] = 19121,
  [22516] = 19550,
  [22517] = 20227,
  [22518] = 19120,
  [22519] = 22187,
  [22520] = 19124,
  [22521] = 19876,
  [22522] = 19622,
  [22523] = 19975,
  [22524] = 19153,
  [22525] = 19550,
  [22526] = 20227,
  [22527] = 19202,
  [22528] = 22187,
  [22529] = 19123,
  [22530] = 19876,
  [22531] = 19622,
  [22532] = 19975,
  [22533] = 19097,
  [22534] = 19550,
  [22535] = 20227,
  [22536] = 19203,
  [22537] = 22187,
  [22538] = 19126,
  [22539] = 19550,
  [22540] = 20227,
  [22541] = 19206,
  [22542] = 22187,
  [22543] = 19550,
  [22544] = 20227,
  [22545] = 19396,
  [22546] = 22187,
  [22547] = 19550,
  [22548] = 20227,
  [22549] = 19129,
  [22550] = 22187,
  [22551] = 19550,
  [22552] = 20227,
  [22553] = 19086,
  [22554] = 22187,
  [22555] = 19550,
  [22556] = 20227,
  [22557] = 19087,
  [22558] = 22187,
  [22559] = 19550,
  [22560] = 20227,
  [22561] = 22187,
  [22562] = 19550,
  [22563] = 20227,
  [22564] = 22187,
  [22565] = 20227,
  [22566] = 22187,
  [22567] = 20227,
  [22568] = 22187,
  [22569] = 20227,
  [22570] = 22187,
  [22571] = 20227,
  [22572] = 22187,
  [22573] = 20227,
  [22574] = 22187,
  [22575] = 20227,
  [22576] = 22187,
  [22577] = 20227,
  [22578] = 22187,
  [22579] = 20227,
  [22580] = 22187,
  [22581] = 20227,
  [22582] = 22187,
  [22583] = 20227,
  [22584] = 22187,
  [22585] = 20227,
  [22586] = 22187,
  [22587] = 20227,
  [22588] = 22187,
  [22589] = 20227,
  [22590] = 22187,
  [22591] = 20227,
  [22592] = 22187,
  [22593] = 20227,
  [22594] = 22187,
  [22595] = 20227,
  [22596] = 22187,
  [22597] = 20227,
  [22598] = 22187,
  [22599] = 20227,
  [22600] = 22187,
  [22601] = 20227,
  [22602] = 22187,
  [22603] = 20227,
  [22604] = 22187,
  [22605] = 20227,
  [22606] = 22187,
  [22607] = 20227,
  [22608] = 22187,
  [22609] = 20227,
  [22610] = 22187,
  [22611] = 20227,
  [22612] = 22187,
  [22613] = 20227,
  [22614] = 22187,
  [22615] = 20227,
  [22616] = 22187,
  [22617] = 20227,
  [22618] = 22187,
  [22619] = 20227,
  [22620] = 22187,
  [22621] = 20227,
  [22622] = 22187,
  [22623] = 20227,
  [22624] = 22187,
  [22625] = 20227,
  [22626] = 22187,
  [22627] = 20227,
  [22628] = 22187,
  [22629] = 20227,
  [22630] = 22187,
  [22631] = 20227,
  [22632] = 22187,
  [22633] = 20227,
  [22634] = 22187,
  [22635] = 20227,
  [22636] = 22187,
  [22637] = 20227,
  [22638] = 22187,
  [22639] = 20227,
  [22640] = 22187,
  [22641] = 20227,
  [22642] = 22187,
  [22643] = 20227,
  [22644] = 22187,
  [22645] = 20227,
  [22646] = 22187,
  [22647] = 20227,
  [22648] = 22187,
  [22649] = 20227,
  [22650] = 22187,
  [22651] = 20227,
  [22652] = 22187,
  [22653] = 20227,
  [22654] = 22187,
  [22655] = 20227,
  [22656] = 22187,
  [22657] = 20227,
  [22658] = 22187,
  [22659] = 20227,
  [22660] = 22187,
  [22661] = 20227,
  [22662] = 22187,
  [22663] = 20227,
  [22664] = 22187,
  [22665] = 20227,
  [22666] = 22187,
  [22667] = 20227,
  [22668] = 22187,
  [22669] = 20227,
  [22670] = 22187,
  [22671] = 20227,
  [22672] = 22187,
  [22673] = 20227,
  [22674] = 22187,
  [22675] = 20227,
  [22676] = 22187,
  [22677] = 20227,
  [22678] = 22187,
  [22679] = 20227,
  [22680] = 22187,
  [22681] = 20227,
  [22682] = 22187,
  [22683] = 20227,
  [22684] = 22187,
  [22685] = 20227,
  [22686] = 22187,
  [22687] = 20227,
  [22688] = 22187,
  [22689] = 20227,
  [22690] = 22187,
  [22691] = 20227,
  [22692] = 22187,
  [22693] = 20227,
  [22694] = 22187,
  [22695] = 20227,
  [22696] = 22187,
  [22697] = 20227,
  [22698] = 22187,
  [22699] = 22187,
  [22700] = 19133,
  [22701] = 19134,
  [22702] = 19169,
  [22703] = 19088,
  [22704] = 19087,
  [22705] = 19105,
  [22706] = 20425,
  [22707] = 19081,
  [22708] = 19805,
  [22709] = 21066,
  [22710] = 19138,
  [22711] = 19139,
  [22712] = 19122,
  [22713] = 20425,
  [22714] = 19081,
  [22715] = 19805,
  [22716] = 21066,
  [22717] = 19141,
  [22718] = 19164,
  [22719] = 20425,
  [22720] = 19081,
  [22721] = 19805,
  [22722] = 21066,
  [22723] = 8355,
  [22724] = 8300,
  [22725] = 20425,
  [22726] = 19081,
  [22727] = 19805,
  [22728] = 21066,
  [22729] = 8332,
  [22730] = 19143,
  [22731] = 20425,
  [22732] = 19081,
  [22733] = 19805,
  [22734] = 21066,
  [22735] = 8329,
  [22736] = 19156,
  [22737] = 20425,
  [22738] = 19081,
  [22739] = 19805,
  [22740] = 21066,
  [22741] = 19145,
  [22742] = 19096,
  [22743] = 20425,
  [22744] = 19118,
  [22745] = 19805,
  [22746] = 21066,
  [22747] = 19147,
  [22748] = 8348,
  [22749] = 19805,
  [22750] = 21066,
  [22751] = 8367,
  [22752] = 19100,
  [22753] = 19805,
  [22754] = 21066,
  [22755] = 19113,
  [22756] = 19088,
  [22757] = 19115,
  [22758] = 19104,
  [22759] = 19153,
  [22760] = 19154,
  [22761] = 19155,
  [22762] = 19116,
  [22763] = 19396,
  [22764] = 19093,
  [22765] = 19118,
  [22766] = 19116,
  [22767] = 19439,
  [22768] = 19396,
  [22769] = 19162,
  [22770] = 19113,
  [22771] = 19089,
  [22772] = 19165,
  [22773] = 19115,
  [22774] = 19116,
  [22775] = 2588,
  [22776] = 19118,
  [22777] = 19169,
  [22778] = 19170,
  [22779] = 19303,
  [22780] = 19117,
  [22781] = 17019,
  [22782] = 19173,
  [22783] = 19174,
  [22784] = 19122,
  [22785] = 19176,
  [22786] = 19121,
  [22787] = 19118,
  [22788] = 19226,
  [22789] = 19436,
  [22790] = 19100,
  [22791] = 19167,
  [22792] = 19171,
  [22793] = 19443,
  [22794] = 19475,
  [22795] = 19104,
  [22796] = 19454,
  [22797] = 19102,
  [22798] = 17299,
  [22799] = 19274,
  [22800] = 8340,
  [22801] = 8338,
  [22802] = 19164,
  [22803] = 19187,
  [22804] = 19096,
  [22805] = 8502,
  [22806] = 19202,
  [22807] = 19203,
  [22808] = 19206,
  [22809] = 19094,
  [22810] = 19105,
  [22811] = 19113,
  [22812] = 19089,
  [22813] = 19121,
  [22814] = 19193,
  [22815] = 19124,
  [22816] = 8409,
  [22817] = 19115,
  [22818] = 19116,
  [22819] = 19196,
  [22820] = 21047,
  [22821] = 22821,
  [22822] = 22822,
  [22823] = 8378,
  [22824] = 22824,
  [22825] = 22825,
  [22826] = 8330,
  [22827] = 22827,
  [22828] = 22828,
  [22829] = 22829,
  [22830] = 8337,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__expression,
  sym__module_elem,
  sym__pattern,
  sym__static_parameter,
  sym__type,
  sym__type_defn_body,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__expression] = {.index = 0, .length = 35},
  [sym__module_elem] = {.index = 35, .length = 10},
  [sym__pattern] = {.index = 45, .length = 19},
  [sym__static_parameter] = {.index = 64, .length = 2},
  [sym__type] = {.index = 66, .length = 13},
  [sym__type_defn_body] = {.index = 79, .length = 9},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    anon_sym_null,
    sym_anon_record_expression,
    sym_application_expression,
    sym_array_expression,
    sym_begin_end_expression,
    sym_brace_expression,
    sym_ce_expression,
    sym_const,
    sym_declaration_expression,
    sym_do_expression,
    sym_dot_expression,
    sym_for_expression,
    sym_fun_expression,
    sym_function_expression,
    sym_if_expression,
    sym_index_expression,
    sym_infix_expression,
    sym_list_expression,
    sym_literal_expression,
    sym_long_identifier_or_op,
    sym_match_expression,
    sym_mutate_expression,
    sym_paren_expression,
    sym_prefixed_expression,
    sym_preproc_if,
    sym_preproc_if_in_class_definition,
    sym_preproc_if_in_expression,
    sym_preproc_if_in_module_body,
    sym_sequential_expression,
    sym_srtp_call_expression,
    sym_try_expression,
    sym_tuple_expression,
    sym_typecast_expression,
    sym_typed_expression,
    sym_while_expression,
  [35] =
    sym__expression,
    sym_declaration_expression,
    sym_exception_definition,
    sym_extern_binding,
    sym_fsi_directive_decl,
    sym_import_decl,
    sym_module_abbrev,
    sym_module_defn,
    sym_preproc_if,
    sym_type_definition,
  [45] =
    alias_sym_wildcard_pattern,
    anon_sym_null,
    sym_array_pattern,
    sym_as_pattern,
    sym_attribute_pattern,
    sym_conjunct_pattern,
    sym_cons_pattern,
    sym_const,
    sym_disjunct_pattern,
    sym_identifier_pattern,
    sym_list_pattern,
    sym_named_field_pattern,
    sym_optional_pattern,
    sym_paren_pattern,
    sym_record_pattern,
    sym_repeat_pattern,
    sym_type_check_pattern,
    sym_typed_const_pattern,
    sym_typed_pattern,
  [64] =
    sym_named_static_parameter,
    sym_static_parameter_value,
  [66] =
    sym_anon_record_type,
    sym_compound_type,
    sym_constrained_type,
    sym_flexible_type,
    sym_function_type,
    sym_generic_type,
    sym_list_type,
    sym_paren_type,
    sym_postfix_type,
    sym_simple_type,
    sym_static_type,
    sym_struct_type,
    sym_type_argument,
  [79] =
    sym_anon_type_defn,
    sym_delegate_type_defn,
    sym_enum_type_defn,
    sym_interface_type_defn,
    sym_record_type_defn,
    sym_type_abbrev_defn,
    sym_type_declaration,
    sym_type_extension,
    sym_union_type_defn,
};

static const TSCharacterRange sym__non_escape_char_character_set_1[] = {
  {0, '!'}, {'#', '&'}, {'(', '`'}, {'c', 'e'}, {'g', 'm'}, {'o', 'q'}, {'s', 's'}, {'u', 'u'},
  {'w', 0x10ffff},
};

static const TSCharacterRange sym_identifier_character_set_1[] = {
  {'A', 'Z'}, {'_', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1},
  {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x370, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f},
  {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x48a, 0x52f}, {0x531, 0x556},
  {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a}, {0x66e, 0x66f}, {0x671, 0x6d3}, {0x6d5, 0x6d5},
  {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710}, {0x712, 0x72f}, {0x74d, 0x7a5}, {0x7b1, 0x7b1},
  {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a}, {0x824, 0x824}, {0x828, 0x828}, {0x840, 0x858},
  {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939}, {0x93d, 0x93d}, {0x950, 0x950}, {0x958, 0x961},
  {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bd, 0x9bd},
  {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28},
  {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa72, 0xa74}, {0xa85, 0xa8d},
  {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabd, 0xabd}, {0xad0, 0xad0}, {0xae0, 0xae1},
  {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3d, 0xb3d},
  {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a},
  {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbd0, 0xbd0}, {0xc05, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc61}, {0xc80, 0xc80}, {0xc85, 0xc8c},
  {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd}, {0xcdd, 0xcde}, {0xce0, 0xce1}, {0xcf1, 0xcf2},
  {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e}, {0xd54, 0xd56}, {0xd5f, 0xd61}, {0xd7a, 0xd7f},
  {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xe01, 0xe30}, {0xe32, 0xe32}, {0xe40, 0xe46},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xeb0}, {0xeb2, 0xeb2}, {0xebd, 0xebd},
  {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47}, {0xf49, 0xf6c}, {0xf88, 0xf8c}, {0x1000, 0x102a},
  {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066}, {0x106e, 0x1070}, {0x1075, 0x1081}, {0x108e, 0x108e},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c},
  {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711}, {0x171f, 0x1731}, {0x1740, 0x1751}, {0x1760, 0x176c},
  {0x176e, 0x1770}, {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878}, {0x1880, 0x18a8}, {0x18aa, 0x18aa}, {0x18b0, 0x18f5},
  {0x1900, 0x191e}, {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x1a00, 0x1a16}, {0x1a20, 0x1a54}, {0x1aa7, 0x1aa7},
  {0x1b05, 0x1b33}, {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5}, {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f}, {0x1c5a, 0x1c7d},
  {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3}, {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa}, {0x1d00, 0x1dbf},
  {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d},
  {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb},
  {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cee}, {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe},
  {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x3005, 0x3007}, {0x3021, 0x3029}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa61f}, {0xa62a, 0xa62b}, {0xa640, 0xa66e}, {0xa67f, 0xa69d},
  {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa801},
  {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3}, {0xa8f2, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa8fe},
  {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf}, {0xa9e0, 0xa9e4}, {0xa9e6, 0xa9ef}, {0xa9fa, 0xa9fe},
  {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a}, {0xaa7e, 0xaaaf}, {0xaab1, 0xaab1}, {0xaab5, 0xaab6},
  {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea}, {0xaaf2, 0xaaf4}, {0xab01, 0xab06}, {0xab09, 0xab0e},
  {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabe2}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6},
  {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb1d}, {0xfb1f, 0xfb28}, {0xfb2a, 0xfb36},
  {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f},
  {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d},
  {0xfe7f, 0xfefc}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xff9d}, {0xffa0, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7},
  {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa},
  {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x10375}, {0x10380, 0x1039d}, {0x103a0, 0x103c3},
  {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a},
  {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3},
  {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808},
  {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5},
  {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a00}, {0x10a10, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35},
  {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91},
  {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23}, {0x10d4a, 0x10d65}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eb0, 0x10eb1},
  {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45}, {0x10f70, 0x10f81}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11003, 0x11037},
  {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8}, {0x11103, 0x11126}, {0x11144, 0x11144}, {0x11147, 0x11147}, {0x11150, 0x11172},
  {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x1122b}, {0x1123f, 0x11240},
  {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112de}, {0x11305, 0x1130c}, {0x1130f, 0x11310},
  {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133d, 0x1133d}, {0x11350, 0x11350}, {0x1135d, 0x11361}, {0x11380, 0x11389},
  {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7}, {0x113d1, 0x113d1}, {0x113d3, 0x113d3}, {0x11400, 0x11434}, {0x11447, 0x1144a},
  {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7}, {0x11580, 0x115ae}, {0x115d8, 0x115db}, {0x11600, 0x1162f}, {0x11644, 0x11644},
  {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746}, {0x11800, 0x1182b}, {0x118a0, 0x118df}, {0x118ff, 0x11906}, {0x11909, 0x11909},
  {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f}, {0x11941, 0x11941}, {0x119a0, 0x119a7}, {0x119aa, 0x119d0}, {0x119e1, 0x119e1},
  {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a}, {0x11a50, 0x11a50}, {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8},
  {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40}, {0x11c72, 0x11c8f}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d30},
  {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89}, {0x11d98, 0x11d98}, {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02}, {0x11f04, 0x11f10},
  {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13441, 0x13446},
  {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a70, 0x16abe}, {0x16ad0, 0x16aed}, {0x16b00, 0x16b2f},
  {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f50, 0x16f50}, {0x16f93, 0x16f9f},
  {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe},
  {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c},
  {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac},
  {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539},
  {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa},
  {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb},
  {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ad}, {0x1e2c0, 0x1e2eb},
  {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4},
  {0x1e900, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32},
  {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f},
  {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62},
  {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b},
  {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0},
  {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange sym_identifier_character_set_2[] = {
  {'\'', '\''}, {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7},
  {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x300, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf},
  {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3},
  {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa},
  {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963},
  {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054},
  {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c},
  {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2},
  {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e},
  {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb},
  {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e},
  {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9},
  {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79},
  {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a}, {0xff65, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102e0},
  {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d},
  {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592},
  {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755},
  {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838},
  {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939},
  {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a},
  {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d},
  {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f50},
  {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8},
  {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc},
  {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d},
  {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328},
  {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357},
  {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0},
  {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461},
  {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644},
  {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739}, {0x11740, 0x11746},
  {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938},
  {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47},
  {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40},
  {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a},
  {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98},
  {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399},
  {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139},
  {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36},
  {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a},
  {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d},
  {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d400, 0x1d454},
  {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3},
  {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546},
  {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e},
  {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c},
  {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018},
  {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149},
  {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee},
  {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df},
  {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
  {0xe0100, 0xe01ef},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 775,
        '"', 604,
        '#', 579,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 581,
        '(', 561,
        ')', 503,
        '*', 569,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 743,
        '/', 573,
        '0', 700,
        '1', 572,
        ':', 486,
        '<', 575,
        '=', 582,
        '>', 519,
        '?', 507,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 520,
        '\\', 2,
        ']', 524,
        '^', 535,
        '`', 253,
        'a', 639,
        'd', 647,
        'f', 736,
        'i', 641,
        'l', 720,
        'm', 740,
        'n', 728,
        'r', 629,
        's', 715,
        'u', 723,
        'w', 631,
        'y', 710,
        '{', 593,
        '|', 512,
        '}', 529,
        '~', 673,
        'E', 748,
        'e', 748,
        'G', 738,
        'I', 738,
        'N', 738,
        'Q', 738,
        'R', 738,
        'Z', 738,
      );
      if (('2' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          (' ' <= lookahead && lookahead <= ';') ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(414);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(768);
      if (lookahead == '"') ADVANCE(595);
      if (lookahead == '#') ADVANCE(773);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == '@') ADVANCE(140);
      if (lookahead == '\\') SKIP(72);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(1);
      END_STATE();
    case 2:
      if (lookahead == '\r') SKIP(279);
      if (lookahead == 'n') SKIP(77);
      if (lookahead == 'u') ADVANCE(390);
      if (lookahead == 'x') ADVANCE(383);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(376);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\r', 584,
        'n', 583,
        'u', 584,
        'x', 584,
        '"', 583,
        '\'', 583,
        'a', 583,
        'b', 583,
        'f', 583,
        'r', 583,
        't', 583,
        'v', 583,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(584);
      if (lookahead != 0) ADVANCE(584);
      END_STATE();
    case 4:
      if (lookahead == '\r') SKIP(329);
      if (lookahead == 'n') SKIP(133);
      END_STATE();
    case 5:
      if (lookahead == '\r') SKIP(330);
      if (lookahead == 'n') SKIP(127);
      END_STATE();
    case 6:
      if (lookahead == '\r') SKIP(331);
      if (lookahead == 'n') SKIP(120);
      END_STATE();
    case 7:
      if (lookahead == '\r') SKIP(342);
      if (lookahead == 'n') SKIP(131);
      END_STATE();
    case 8:
      if (lookahead == '\r') SKIP(344);
      if (lookahead == 'n') SKIP(134);
      END_STATE();
    case 9:
      if (lookahead == '\r') SKIP(282);
      if (lookahead == 'n') SKIP(130);
      END_STATE();
    case 10:
      if (lookahead == '\r') SKIP(343);
      if (lookahead == 'n') SKIP(117);
      END_STATE();
    case 11:
      if (lookahead == '\r') SKIP(347);
      if (lookahead == 'n') SKIP(126);
      END_STATE();
    case 12:
      if (lookahead == '\r') SKIP(332);
      if (lookahead == 'n') SKIP(121);
      END_STATE();
    case 13:
      if (lookahead == '\r') SKIP(333);
      if (lookahead == 'n') SKIP(124);
      END_STATE();
    case 14:
      if (lookahead == '\r') SKIP(345);
      if (lookahead == 'n') SKIP(115);
      END_STATE();
    case 15:
      if (lookahead == '\r') SKIP(348);
      if (lookahead == 'n') SKIP(123);
      END_STATE();
    case 16:
      if (lookahead == '\r') SKIP(349);
      if (lookahead == 'n') SKIP(116);
      END_STATE();
    case 17:
      if (lookahead == '\r') SKIP(352);
      if (lookahead == 'n') SKIP(122);
      END_STATE();
    case 18:
      if (lookahead == '\r') SKIP(351);
      if (lookahead == 'n') SKIP(136);
      END_STATE();
    case 19:
      if (lookahead == '\r') SKIP(353);
      if (lookahead == 'n') SKIP(125);
      END_STATE();
    case 20:
      if (lookahead == '\r') SKIP(355);
      if (lookahead == 'n') SKIP(129);
      END_STATE();
    case 21:
      if (lookahead == '\r') SKIP(357);
      if (lookahead == 'n') SKIP(128);
      END_STATE();
    case 22:
      if (lookahead == '\r') SKIP(354);
      if (lookahead == 'n') SKIP(118);
      END_STATE();
    case 23:
      if (lookahead == '\r') SKIP(356);
      if (lookahead == 'n') SKIP(119);
      END_STATE();
    case 24:
      if (lookahead == '\r') SKIP(283);
      if (lookahead == 'n') SKIP(179);
      END_STATE();
    case 25:
      if (lookahead == '\r') SKIP(284);
      if (lookahead == 'n') SKIP(165);
      END_STATE();
    case 26:
      if (lookahead == '\r') SKIP(285);
      if (lookahead == 'n') SKIP(154);
      END_STATE();
    case 27:
      if (lookahead == '\r') SKIP(286);
      if (lookahead == 'n') SKIP(172);
      END_STATE();
    case 28:
      if (lookahead == '\r') SKIP(287);
      if (lookahead == 'n') SKIP(150);
      END_STATE();
    case 29:
      if (lookahead == '\r') SKIP(288);
      if (lookahead == 'n') SKIP(206);
      END_STATE();
    case 30:
      if (lookahead == '\r') SKIP(289);
      if (lookahead == 'n') SKIP(196);
      END_STATE();
    case 31:
      if (lookahead == '\r') SKIP(290);
      if (lookahead == 'n') SKIP(170);
      END_STATE();
    case 32:
      if (lookahead == '\r') SKIP(291);
      if (lookahead == 'n') SKIP(159);
      END_STATE();
    case 33:
      if (lookahead == '\r') SKIP(292);
      if (lookahead == 'n') SKIP(177);
      END_STATE();
    case 34:
      if (lookahead == '\r') SKIP(293);
      if (lookahead == 'n') SKIP(151);
      END_STATE();
    case 35:
      if (lookahead == '\r') SKIP(294);
      if (lookahead == 'n') SKIP(152);
      END_STATE();
    case 36:
      if (lookahead == '\r') SKIP(295);
      if (lookahead == 'n') SKIP(161);
      END_STATE();
    case 37:
      if (lookahead == '\r') SKIP(337);
      if (lookahead == 'n') SKIP(153);
      END_STATE();
    case 38:
      if (lookahead == '\r') SKIP(296);
      if (lookahead == 'n') SKIP(162);
      END_STATE();
    case 39:
      if (lookahead == '\r') SKIP(297);
      if (lookahead == 'n') SKIP(163);
      END_STATE();
    case 40:
      if (lookahead == '\r') SKIP(338);
      if (lookahead == 'n') SKIP(164);
      END_STATE();
    case 41:
      if (lookahead == '\r') SKIP(298);
      if (lookahead == 'n') SKIP(192);
      END_STATE();
    case 42:
      if (lookahead == '\r') SKIP(299);
      if (lookahead == 'n') SKIP(187);
      END_STATE();
    case 43:
      if (lookahead == '\r') SKIP(300);
      if (lookahead == 'n') SKIP(223);
      END_STATE();
    case 44:
      if (lookahead == '\r') SKIP(301);
      if (lookahead == 'n') SKIP(204);
      END_STATE();
    case 45:
      if (lookahead == '\r') SKIP(302);
      if (lookahead == 'n') SKIP(188);
      END_STATE();
    case 46:
      if (lookahead == '\r') SKIP(303);
      if (lookahead == 'n') SKIP(195);
      END_STATE();
    case 47:
      if (lookahead == '\r') SKIP(335);
      if (lookahead == 'n') SKIP(160);
      END_STATE();
    case 48:
      if (lookahead == '\r') SKIP(334);
      if (lookahead == 'n') SKIP(171);
      END_STATE();
    case 49:
      if (lookahead == '\r') SKIP(336);
      if (lookahead == 'n') SKIP(178);
      END_STATE();
    case 50:
      if (lookahead == '\r') SKIP(304);
      if (lookahead == 'n') SKIP(141);
      END_STATE();
    case 51:
      if (lookahead == '\r') SKIP(305);
      if (lookahead == 'n') SKIP(180);
      END_STATE();
    case 52:
      if (lookahead == '\r') SKIP(306);
      if (lookahead == 'n') SKIP(181);
      END_STATE();
    case 53:
      if (lookahead == '\r') SKIP(307);
      if (lookahead == 'n') SKIP(225);
      END_STATE();
    case 54:
      if (lookahead == '\r') SKIP(310);
      if (lookahead == 'n') SKIP(189);
      END_STATE();
    case 55:
      if (lookahead == '\r') SKIP(311);
      if (lookahead == 'n') SKIP(182);
      END_STATE();
    case 56:
      if (lookahead == '\r') SKIP(312);
      if (lookahead == 'n') SKIP(213);
      END_STATE();
    case 57:
      if (lookahead == '\r') SKIP(313);
      if (lookahead == 'n') SKIP(138);
      END_STATE();
    case 58:
      if (lookahead == '\r') SKIP(314);
      if (lookahead == 'n') SKIP(215);
      END_STATE();
    case 59:
      if (lookahead == '\r') SKIP(315);
      if (lookahead == 'n') SKIP(208);
      END_STATE();
    case 60:
      if (lookahead == '\r') SKIP(316);
      if (lookahead == 'n') SKIP(209);
      END_STATE();
    case 61:
      if (lookahead == '\r') SKIP(317);
      if (lookahead == 'n') SKIP(221);
      END_STATE();
    case 62:
      if (lookahead == '\r') SKIP(318);
      if (lookahead == 'n') SKIP(217);
      END_STATE();
    case 63:
      if (lookahead == '\r') SKIP(339);
      if (lookahead == 'n') SKIP(210);
      END_STATE();
    case 64:
      if (lookahead == '\r') SKIP(319);
      if (lookahead == 'n') SKIP(207);
      END_STATE();
    case 65:
      if (lookahead == '\r') SKIP(320);
      if (lookahead == 'n') SKIP(227);
      END_STATE();
    case 66:
      if (lookahead == '\r') SKIP(321);
      if (lookahead == 'n') SKIP(218);
      END_STATE();
    case 67:
      if (lookahead == '\r') SKIP(322);
      if (lookahead == 'n') SKIP(142);
      if (lookahead == 'u') ADVANCE(390);
      if (lookahead == 'x') ADVANCE(383);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(376);
      END_STATE();
    case 68:
      if (lookahead == '\r') SKIP(323);
      if (lookahead == 'n') SKIP(216);
      END_STATE();
    case 69:
      if (lookahead == '\r') SKIP(324);
      if (lookahead == 'n') SKIP(211);
      END_STATE();
    case 70:
      if (lookahead == '\r') SKIP(325);
      if (lookahead == 'n') SKIP(219);
      END_STATE();
    case 71:
      if (lookahead == '\r') SKIP(326);
      if (lookahead == 'n') SKIP(220);
      END_STATE();
    case 72:
      if (lookahead == '\r') SKIP(327);
      if (lookahead == 'n') SKIP(1);
      END_STATE();
    case 73:
      if (lookahead == '\r') SKIP(341);
      if (lookahead == 'n') SKIP(132);
      END_STATE();
    case 74:
      if (lookahead == '\r') SKIP(346);
      if (lookahead == 'n') SKIP(137);
      END_STATE();
    case 75:
      if (lookahead == '\r') SKIP(350);
      if (lookahead == 'n') SKIP(135);
      END_STATE();
    case 76:
      if (lookahead == ' ') ADVANCE(233);
      if (lookahead == ')') ADVANCE(658);
      if (('\t' <= lookahead && lookahead <= '\r')) ADVANCE(234);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 77:
      ADVANCE_MAP(
        '!', 775,
        '"', 595,
        '#', 579,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 581,
        '(', 500,
        ')', 503,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 537,
        '/', 573,
        '0', 700,
        '1', 572,
        ':', 486,
        '<', 575,
        '=', 582,
        '>', 519,
        '?', 507,
        '@', 680,
        '[', 520,
        '\\', 2,
        ']', 524,
        '^', 535,
        '`', 253,
        'a', 639,
        'd', 647,
        'i', 641,
        'l', 624,
        'm', 618,
        'r', 629,
        'u', 650,
        'w', 631,
        'y', 636,
        '{', 528,
        '|', 512,
        '}', 529,
        '~', 673,
      );
      if (('2' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          (' ' <= lookahead && lookahead <= ';') ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(77);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 78:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(14);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(115);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 79:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(16);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(116);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 80:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 745,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(10);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(117);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 81:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(10);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(117);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 82:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(10);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(117);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 83:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(10);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(117);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 84:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(22);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(118);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 85:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(23);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(119);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 86:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(6);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(120);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 87:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 570,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(12);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(121);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 88:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 570,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 575,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(17);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(122);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 89:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 570,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(15);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(123);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 90:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 570,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 689,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(13);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(124);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 91:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 743,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(9);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(130);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 92:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 743,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(7);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(131);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 93:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(19);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(125);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 94:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 745,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(73);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(132);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 95:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 745,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(11);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(126);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 96:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 744,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 97:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 744,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(5);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 98:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 744,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(8);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(134);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 99:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 100:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(5);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 101:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(8);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(134);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 102:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 103:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 104:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(5);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 105:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(5);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 106:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(8);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(134);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 107:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(8);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(134);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 108:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 109:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(5);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 110:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(75);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(135);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 111:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(21);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(128);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 112:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 582,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(18);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(136);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 113:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(74);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(137);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 114:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(20);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(129);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 115:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(14);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(115);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 116:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(16);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(116);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 117:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(10);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(117);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 118:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(22);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(118);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 119:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(23);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(119);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 120:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(6);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(120);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 121:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(12);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(121);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 122:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 575,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(17);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(122);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 123:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 519,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(15);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(123);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 124:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 689,
        '?', 506,
        '@', 680,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(13);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(124);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 125:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(19);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(125);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 126:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(11);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(126);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 127:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(5);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(127);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 128:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(21);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(128);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 129:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(20);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(129);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 130:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(9);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(130);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 131:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(7);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(131);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 132:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(73);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(132);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 133:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(133);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 134:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(8);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(134);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 135:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(75);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(135);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 136:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 582,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(18);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(136);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 137:
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 772,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(74);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(137);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 138:
      ADVANCE_MAP(
        '!', 774,
        '#', 770,
        '\'', 581,
        '(', 502,
        ')', 503,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 573,
        ':', 490,
        '<', 574,
        '=', 480,
        '>', 518,
        '[', 244,
      );
      if (lookahead == '\\') SKIP(57);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(138);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 139:
      if (lookahead == '"') ADVANCE(594);
      END_STATE();
    case 140:
      if (lookahead == '"') ADVANCE(601);
      END_STATE();
    case 141:
      ADVANCE_MAP(
        '"', 595,
        '#', 770,
        '(', 502,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        ':', 490,
        '<', 574,
        '=', 480,
        '@', 140,
        '[', 244,
      );
      if (lookahead == '\\') SKIP(50);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == '{') ADVANCE(370);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(141);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 142:
      if (lookahead == '"') ADVANCE(595);
      if (lookahead == '#') ADVANCE(773);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == '\\') ADVANCE(67);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(142);
      END_STATE();
    case 143:
      if (lookahead == '"') ADVANCE(595);
      if (lookahead == '#') ADVANCE(589);
      if (lookahead == '(') ADVANCE(589);
      if (lookahead == '/') ADVANCE(589);
      if (lookahead == '\\') ADVANCE(3);
      if (lookahead == '{') ADVANCE(593);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r')) SKIP(142);
      if (lookahead == '\n' ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(589);
      END_STATE();
    case 144:
      if (lookahead == '"') ADVANCE(595);
      if (lookahead == '#') ADVANCE(589);
      if (lookahead == '(') ADVANCE(589);
      if (lookahead == '/') ADVANCE(589);
      if (lookahead == '\\') ADVANCE(3);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r')) SKIP(142);
      if (lookahead == '\n' ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(589);
      END_STATE();
    case 145:
      if (lookahead == '"') ADVANCE(606);
      END_STATE();
    case 146:
      ADVANCE_MAP(
        '"', 596,
        '#', 772,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        ',', 504,
        '-', 664,
        '.', 742,
        '/', 241,
        '0', 700,
        '<', 248,
        '>', 518,
        '@', 140,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(28);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(150);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 147:
      ADVANCE_MAP(
        '"', 596,
        '#', 772,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        ',', 504,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '>', 518,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(28);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(150);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 148:
      ADVANCE_MAP(
        '"', 596,
        '#', 772,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        ',', 504,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '>', 518,
        '@', 140,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(28);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(150);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 149:
      ADVANCE_MAP(
        '"', 596,
        '#', 772,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        ',', 504,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '>', 518,
        '@', 140,
        'M', 741,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(28);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(150);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 150:
      ADVANCE_MAP(
        '"', 596,
        '#', 772,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        ',', 504,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '>', 518,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(28);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(150);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 151:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(34);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(151);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 152:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(35);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(152);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 153:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(37);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(153);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 154:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(26);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(154);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 155:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '.', 742,
        '/', 241,
        '0', 700,
        ':', 489,
        '=', 480,
        '?', 505,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(32);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 156:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '=', 480,
        '?', 505,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(32);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 157:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(32);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 158:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(32);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 159:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(32);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(159);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 160:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ')', 503,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(47);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(160);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 161:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(36);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(161);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 162:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(38);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(162);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 163:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(39);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(163);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 164:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 488,
        '<', 574,
        '?', 505,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(40);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(164);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 165:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(25);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(165);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 166:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 742,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(170);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 167:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(170);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 168:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(170);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 169:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(170);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 170:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(170);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 171:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        '+', 373,
        ',', 504,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(48);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(171);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 172:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 489,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(27);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(172);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 173:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '.', 742,
        '/', 241,
        '0', 700,
        ':', 489,
        '?', 505,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(177);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 174:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '?', 505,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(177);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 175:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(177);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 176:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(177);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 177:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(177);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 178:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '&', 515,
        '\'', 229,
        '(', 500,
        ',', 504,
        '/', 241,
        '0', 700,
        ':', 489,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(49);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('+' <= lookahead && lookahead <= '-')) ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(178);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 179:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 500,
        ')', 503,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 492,
        '<', 574,
        '=', 480,
        '?', 505,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(24);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(179);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 180:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '*', 568,
        '+', 373,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 491,
        '<', 574,
        '=', 480,
        '@', 140,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(51);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(180);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 181:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '*', 568,
        '+', 373,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 491,
        '<', 574,
        '=', 480,
        '@', 140,
        '[', 522,
      );
      if (lookahead == '\\') SKIP(52);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '{') ADVANCE(527);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(181);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 182:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 492,
        '<', 574,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(55);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '{') ADVANCE(527);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(182);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 183:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '.', 742,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(42);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(187);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 184:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(45);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(188);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 185:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(42);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(187);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 186:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(42);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(187);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 187:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(42);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(527);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(187);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 188:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        '-', 247,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(45);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(188);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 189:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 492,
        '<', 574,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(54);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(189);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 190:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '.', 742,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(41);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(251);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(192);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 191:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(46);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(195);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 192:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(41);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(251);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(192);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 193:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(46);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(195);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 194:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(46);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(741);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(195);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 195:
      ADVANCE_MAP(
        '"', 596,
        '#', 773,
        '$', 139,
        '\'', 229,
        '(', 501,
        '/', 241,
        '0', 700,
        ':', 492,
        '=', 480,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(46);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(195);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 196:
      ADVANCE_MAP(
        '"', 596,
        '#', 580,
        '$', 139,
        '\'', 581,
        '(', 501,
        '/', 241,
        '0', 700,
        '1', 572,
        '>', 518,
        '@', 140,
      );
      if (lookahead == '\\') SKIP(30);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(370);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(373);
      if (('2' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(196);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 197:
      if (lookahead == '"') ADVANCE(607);
      END_STATE();
    case 198:
      if (lookahead == '"') ADVANCE(597);
      if (lookahead == '#') ADVANCE(589);
      if (lookahead == '(') ADVANCE(589);
      if (lookahead == '/') ADVANCE(589);
      if (lookahead == '\\') ADVANCE(3);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r')) SKIP(142);
      if (lookahead == '\n' ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(589);
      END_STATE();
    case 199:
      if (lookahead == '"') ADVANCE(603);
      if (lookahead == '#') ADVANCE(589);
      if (lookahead == '(') ADVANCE(589);
      if (lookahead == '/') ADVANCE(589);
      if (lookahead == '\\') ADVANCE(599);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r')) SKIP(205);
      if (lookahead == '\n' ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(589);
      END_STATE();
    case 200:
      if (lookahead == '"') ADVANCE(600);
      END_STATE();
    case 201:
      if (lookahead == '"') ADVANCE(602);
      if (lookahead == '#') ADVANCE(589);
      if (lookahead == '(') ADVANCE(589);
      if (lookahead == '/') ADVANCE(589);
      if (lookahead == '\\') ADVANCE(599);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r')) SKIP(205);
      if (lookahead == '\n' ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(589);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(589);
      END_STATE();
    case 202:
      if (lookahead == '"') ADVANCE(197);
      END_STATE();
    case 203:
      ADVANCE_MAP(
        '"', 202,
        '#', 770,
        '\'', 581,
        '(', 502,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 573,
        ':', 490,
        '<', 574,
        '=', 480,
        '[', 244,
      );
      if (lookahead == '\\') SKIP(44);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == '{') ADVANCE(593);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(204);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 204:
      ADVANCE_MAP(
        '"', 202,
        '#', 770,
        '\'', 581,
        '(', 502,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 573,
        ':', 490,
        '<', 574,
        '=', 480,
        '[', 244,
      );
      if (lookahead == '\\') SKIP(44);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(204);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 205:
      if (lookahead == '"') ADVANCE(200);
      if (lookahead == '#') ADVANCE(773);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == '\\') ADVANCE(598);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(205);
      END_STATE();
    case 206:
      if (lookahead == '#') ADVANCE(579);
      if (lookahead == '\'') ADVANCE(581);
      if (lookahead == '(') ADVANCE(502);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == '1') ADVANCE(571);
      if (lookahead == '[') ADVANCE(243);
      if (lookahead == '\\') SKIP(29);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(206);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 207:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '=', 480,
        '>', 518,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(64);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(207);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 208:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '=', 480,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(59);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(208);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 209:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '=', 480,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(60);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(209);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 210:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '=', 480,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(63);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(210);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 211:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 485,
        '<', 574,
        '=', 480,
      );
      if (lookahead == '\\') SKIP(69);
      if (lookahead == 'a') ADVANCE(363);
      if (lookahead == 'i') ADVANCE(308);
      if (lookahead == 'w') ADVANCE(271);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(211);
      END_STATE();
    case 212:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        ',', 504,
        '-', 246,
        '/', 241,
        ':', 485,
        '<', 574,
        '=', 480,
        'L', 732,
        'U', 617,
      );
      if (lookahead == '\\') SKIP(56);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(213);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 213:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        ',', 504,
        '-', 246,
        '/', 241,
        ':', 485,
        '<', 574,
        '=', 480,
      );
      if (lookahead == '\\') SKIP(56);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(213);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 214:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        ',', 504,
        '/', 241,
        ':', 485,
        '=', 582,
        'L', 732,
        'U', 617,
      );
      if (lookahead == '\\') SKIP(58);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(215);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 215:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        ')', 503,
        ',', 504,
        '/', 241,
        ':', 485,
        '=', 582,
      );
      if (lookahead == '\\') SKIP(58);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(215);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 216:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '=', 582,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(68);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(649);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(216);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 217:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '=', 480,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(62);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(217);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 218:
      ADVANCE_MAP(
        '#', 773,
        '&', 515,
        '(', 236,
        '*', 568,
        ',', 504,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 487,
        '<', 574,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(66);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(643);
      if (lookahead == 'i') ADVANCE(641);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(218);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 219:
      ADVANCE_MAP(
        '#', 773,
        '(', 236,
        '*', 568,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 245,
        '<', 574,
        '=', 480,
        '>', 518,
        '[', 244,
      );
      if (lookahead == '\\') SKIP(70);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(219);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 220:
      ADVANCE_MAP(
        '#', 773,
        '(', 236,
        '*', 568,
        '-', 246,
        '.', 536,
        '/', 241,
        ':', 245,
        '<', 574,
        '=', 480,
        '[', 250,
      );
      if (lookahead == '\\') SKIP(71);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'w') ADVANCE(634);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(220);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 221:
      if (lookahead == '#') ADVANCE(773);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == '[') ADVANCE(243);
      if (lookahead == '\\') SKIP(61);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(221);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 222:
      if (lookahead == '#') ADVANCE(773);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(242);
      if (lookahead == '[') ADVANCE(243);
      if (lookahead == '\\') SKIP(61);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(222);
      if (('\n' <= lookahead && lookahead <= '\r') ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(221);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 223:
      ADVANCE_MAP(
        '#', 580,
        '\'', 581,
        '(', 502,
        ')', 503,
        '/', 241,
        '1', 571,
        '>', 518,
        '?', 505,
        '[', 243,
      );
      if (lookahead == '\\') SKIP(43);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(370);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(223);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 224:
      ADVANCE_MAP(
        '#', 771,
        '&', 515,
        '(', 236,
        ')', 503,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 742,
        '/', 241,
        ':', 485,
        '<', 574,
        '=', 480,
        '>', 518,
        'L', 731,
        'U', 249,
        '[', 243,
      );
      if (lookahead == '\\') SKIP(53);
      if (lookahead == 'a') ADVANCE(340);
      if (lookahead == 'f') ADVANCE(735);
      if (lookahead == 'i') ADVANCE(308);
      if (lookahead == 'l') ADVANCE(719);
      if (lookahead == 'n') ADVANCE(727);
      if (lookahead == 's') ADVANCE(714);
      if (lookahead == 'u') ADVANCE(725);
      if (lookahead == 'w') ADVANCE(271);
      if (lookahead == 'y') ADVANCE(709);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(747);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(739);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(737);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(225);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 225:
      ADVANCE_MAP(
        '#', 771,
        '&', 515,
        '(', 236,
        ')', 503,
        '+', 373,
        ',', 504,
        '-', 247,
        '/', 241,
        ':', 485,
        '<', 574,
        '=', 480,
        '>', 518,
        '[', 243,
      );
      if (lookahead == '\\') SKIP(53);
      if (lookahead == 'a') ADVANCE(340);
      if (lookahead == 'i') ADVANCE(308);
      if (lookahead == 'w') ADVANCE(271);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(225);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 226:
      if (lookahead == '#') ADVANCE(771);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == 'L') ADVANCE(732);
      if (lookahead == 'U') ADVANCE(617);
      if (lookahead == '[') ADVANCE(243);
      if (lookahead == '\\') SKIP(65);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'l') ADVANCE(721);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(724);
      if (lookahead == 'y') ADVANCE(711);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(227);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 227:
      if (lookahead == '#') ADVANCE(771);
      if (lookahead == '(') ADVANCE(236);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == '[') ADVANCE(243);
      if (lookahead == '\\') SKIP(65);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == '|') ADVANCE(511);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(227);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 228:
      if (lookahead == '&') ADVANCE(564);
      END_STATE();
    case 229:
      if (lookahead == '\'') ADVANCE(591);
      if (lookahead == '\\') ADVANCE(368);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(230);
      END_STATE();
    case 230:
      if (lookahead == '\'') ADVANCE(592);
      END_STATE();
    case 231:
      if (lookahead == ')') ADVANCE(751);
      END_STATE();
    case 232:
      if (lookahead == ')') ADVANCE(658);
      if (lookahead == '.') ADVANCE(76);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 233:
      if (lookahead == ')') ADVANCE(658);
      if (lookahead == '.') ADVANCE(239);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      END_STATE();
    case 234:
      if (lookahead == ')') ADVANCE(658);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      END_STATE();
    case 235:
      if (lookahead == ')') ADVANCE(658);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 236:
      if (lookahead == '*') ADVANCE(749);
      END_STATE();
    case 237:
      if (lookahead == '-') ADVANCE(552);
      END_STATE();
    case 238:
      if (lookahead == '.') ADVANCE(232);
      if (lookahead == '?') ADVANCE(235);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(238);
      if (lookahead == '!' ||
          ('$' <= lookahead && lookahead <= '&') ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 239:
      if (lookahead == '.') ADVANCE(234);
      END_STATE();
    case 240:
      if (lookahead == '.') ADVANCE(567);
      END_STATE();
    case 241:
      if (lookahead == '/') ADVANCE(754);
      END_STATE();
    case 242:
      if (lookahead == '/') ADVANCE(752);
      END_STATE();
    case 243:
      if (lookahead == '<') ADVANCE(482);
      END_STATE();
    case 244:
      if (lookahead == '<') ADVANCE(482);
      if (lookahead == ']') ADVANCE(578);
      END_STATE();
    case 245:
      if (lookahead == '>') ADVANCE(544);
      END_STATE();
    case 246:
      if (lookahead == '>') ADVANCE(549);
      END_STATE();
    case 247:
      if (lookahead == '>') ADVANCE(549);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 248:
      if (lookahead == '@') ADVANCE(531);
      END_STATE();
    case 249:
      if (lookahead == 'L') ADVANCE(733);
      END_STATE();
    case 250:
      if (lookahead == ']') ADVANCE(578);
      END_STATE();
    case 251:
      if (lookahead == ']') ADVANCE(526);
      if (lookahead == '}') ADVANCE(541);
      END_STATE();
    case 252:
      if (lookahead == ']') ADVANCE(483);
      END_STATE();
    case 253:
      if (lookahead == '`') ADVANCE(392);
      END_STATE();
    case 254:
      if (lookahead == '`') ADVANCE(609);
      END_STATE();
    case 255:
      if (lookahead == '`') ADVANCE(254);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(255);
      END_STATE();
    case 256:
      if (lookahead == 'a') ADVANCE(361);
      END_STATE();
    case 257:
      if (lookahead == 'a') ADVANCE(259);
      END_STATE();
    case 258:
      if (lookahead == 'a') ADVANCE(362);
      END_STATE();
    case 259:
      if (lookahead == 'd') ADVANCE(763);
      END_STATE();
    case 260:
      if (lookahead == 'd') ADVANCE(493);
      END_STATE();
    case 261:
      if (lookahead == 'd') ADVANCE(275);
      END_STATE();
    case 262:
      if (lookahead == 'e') ADVANCE(276);
      END_STATE();
    case 263:
      if (lookahead == 'e') ADVANCE(769);
      END_STATE();
    case 264:
      if (lookahead == 'e') ADVANCE(764);
      END_STATE();
    case 265:
      if (lookahead == 'e') ADVANCE(778);
      END_STATE();
    case 266:
      if (lookahead == 'e') ADVANCE(309);
      END_STATE();
    case 267:
      if (lookahead == 'f') ADVANCE(776);
      END_STATE();
    case 268:
      if (lookahead == 'f') ADVANCE(777);
      END_STATE();
    case 269:
      if (lookahead == 'g') ADVANCE(270);
      if (lookahead == 'n') ADVANCE(263);
      END_STATE();
    case 270:
      if (lookahead == 'h') ADVANCE(366);
      END_STATE();
    case 271:
      if (lookahead == 'h') ADVANCE(266);
      END_STATE();
    case 272:
      if (lookahead == 'i') ADVANCE(269);
      if (lookahead == 'o') ADVANCE(257);
      END_STATE();
    case 273:
      if (lookahead == 'i') ADVANCE(278);
      END_STATE();
    case 274:
      if (lookahead == 'i') ADVANCE(365);
      END_STATE();
    case 275:
      if (lookahead == 'i') ADVANCE(268);
      END_STATE();
    case 276:
      if (lookahead == 'l') ADVANCE(360);
      END_STATE();
    case 277:
      if (lookahead == 'l') ADVANCE(364);
      if (lookahead == 'n') ADVANCE(261);
      END_STATE();
    case 278:
      if (lookahead == 'm') ADVANCE(264);
      END_STATE();
    case 279:
      if (lookahead == 'n') SKIP(77);
      END_STATE();
    case 280:
      if (lookahead == 'n') ADVANCE(759);
      END_STATE();
    case 281:
      if (lookahead == 'n') ADVANCE(760);
      END_STATE();
    case 282:
      if (lookahead == 'n') SKIP(130);
      END_STATE();
    case 283:
      if (lookahead == 'n') SKIP(179);
      END_STATE();
    case 284:
      if (lookahead == 'n') SKIP(165);
      END_STATE();
    case 285:
      if (lookahead == 'n') SKIP(154);
      END_STATE();
    case 286:
      if (lookahead == 'n') SKIP(172);
      END_STATE();
    case 287:
      if (lookahead == 'n') SKIP(150);
      END_STATE();
    case 288:
      if (lookahead == 'n') SKIP(206);
      END_STATE();
    case 289:
      if (lookahead == 'n') SKIP(196);
      END_STATE();
    case 290:
      if (lookahead == 'n') SKIP(170);
      END_STATE();
    case 291:
      if (lookahead == 'n') SKIP(159);
      END_STATE();
    case 292:
      if (lookahead == 'n') SKIP(177);
      END_STATE();
    case 293:
      if (lookahead == 'n') SKIP(151);
      END_STATE();
    case 294:
      if (lookahead == 'n') SKIP(152);
      END_STATE();
    case 295:
      if (lookahead == 'n') SKIP(161);
      END_STATE();
    case 296:
      if (lookahead == 'n') SKIP(162);
      END_STATE();
    case 297:
      if (lookahead == 'n') SKIP(163);
      END_STATE();
    case 298:
      if (lookahead == 'n') SKIP(192);
      END_STATE();
    case 299:
      if (lookahead == 'n') SKIP(187);
      END_STATE();
    case 300:
      if (lookahead == 'n') SKIP(223);
      END_STATE();
    case 301:
      if (lookahead == 'n') SKIP(204);
      END_STATE();
    case 302:
      if (lookahead == 'n') SKIP(188);
      END_STATE();
    case 303:
      if (lookahead == 'n') SKIP(195);
      END_STATE();
    case 304:
      if (lookahead == 'n') SKIP(141);
      END_STATE();
    case 305:
      if (lookahead == 'n') SKIP(180);
      END_STATE();
    case 306:
      if (lookahead == 'n') SKIP(181);
      END_STATE();
    case 307:
      if (lookahead == 'n') SKIP(225);
      END_STATE();
    case 308:
      if (lookahead == 'n') ADVANCE(546);
      END_STATE();
    case 309:
      if (lookahead == 'n') ADVANCE(559);
      END_STATE();
    case 310:
      if (lookahead == 'n') SKIP(189);
      END_STATE();
    case 311:
      if (lookahead == 'n') SKIP(182);
      END_STATE();
    case 312:
      if (lookahead == 'n') SKIP(213);
      END_STATE();
    case 313:
      if (lookahead == 'n') SKIP(138);
      END_STATE();
    case 314:
      if (lookahead == 'n') SKIP(215);
      END_STATE();
    case 315:
      if (lookahead == 'n') SKIP(208);
      END_STATE();
    case 316:
      if (lookahead == 'n') SKIP(209);
      END_STATE();
    case 317:
      if (lookahead == 'n') SKIP(221);
      END_STATE();
    case 318:
      if (lookahead == 'n') SKIP(217);
      END_STATE();
    case 319:
      if (lookahead == 'n') SKIP(207);
      END_STATE();
    case 320:
      if (lookahead == 'n') SKIP(227);
      END_STATE();
    case 321:
      if (lookahead == 'n') SKIP(218);
      END_STATE();
    case 322:
      if (lookahead == 'n') SKIP(142);
      END_STATE();
    case 323:
      if (lookahead == 'n') SKIP(216);
      END_STATE();
    case 324:
      if (lookahead == 'n') SKIP(211);
      END_STATE();
    case 325:
      if (lookahead == 'n') SKIP(219);
      END_STATE();
    case 326:
      if (lookahead == 'n') SKIP(220);
      END_STATE();
    case 327:
      if (lookahead == 'n') SKIP(1);
      END_STATE();
    case 328:
      if (lookahead == 'n') ADVANCE(359);
      END_STATE();
    case 329:
      if (lookahead == 'n') SKIP(133);
      END_STATE();
    case 330:
      if (lookahead == 'n') SKIP(127);
      END_STATE();
    case 331:
      if (lookahead == 'n') SKIP(120);
      END_STATE();
    case 332:
      if (lookahead == 'n') SKIP(121);
      END_STATE();
    case 333:
      if (lookahead == 'n') SKIP(124);
      END_STATE();
    case 334:
      if (lookahead == 'n') SKIP(171);
      END_STATE();
    case 335:
      if (lookahead == 'n') SKIP(160);
      END_STATE();
    case 336:
      if (lookahead == 'n') SKIP(178);
      END_STATE();
    case 337:
      if (lookahead == 'n') SKIP(153);
      END_STATE();
    case 338:
      if (lookahead == 'n') SKIP(164);
      END_STATE();
    case 339:
      if (lookahead == 'n') SKIP(210);
      END_STATE();
    case 340:
      if (lookahead == 'n') ADVANCE(260);
      if (lookahead == 's') ADVANCE(498);
      END_STATE();
    case 341:
      if (lookahead == 'n') SKIP(132);
      END_STATE();
    case 342:
      if (lookahead == 'n') SKIP(131);
      END_STATE();
    case 343:
      if (lookahead == 'n') SKIP(117);
      END_STATE();
    case 344:
      if (lookahead == 'n') SKIP(134);
      END_STATE();
    case 345:
      if (lookahead == 'n') SKIP(115);
      END_STATE();
    case 346:
      if (lookahead == 'n') SKIP(137);
      END_STATE();
    case 347:
      if (lookahead == 'n') SKIP(126);
      END_STATE();
    case 348:
      if (lookahead == 'n') SKIP(123);
      END_STATE();
    case 349:
      if (lookahead == 'n') SKIP(116);
      END_STATE();
    case 350:
      if (lookahead == 'n') SKIP(135);
      END_STATE();
    case 351:
      if (lookahead == 'n') SKIP(136);
      END_STATE();
    case 352:
      if (lookahead == 'n') SKIP(122);
      END_STATE();
    case 353:
      if (lookahead == 'n') SKIP(125);
      END_STATE();
    case 354:
      if (lookahead == 'n') SKIP(118);
      END_STATE();
    case 355:
      if (lookahead == 'n') SKIP(129);
      END_STATE();
    case 356:
      if (lookahead == 'n') SKIP(119);
      END_STATE();
    case 357:
      if (lookahead == 'n') SKIP(128);
      END_STATE();
    case 358:
      if (lookahead == 'o') ADVANCE(369);
      END_STATE();
    case 359:
      if (lookahead == 'o') ADVANCE(281);
      END_STATE();
    case 360:
      if (lookahead == 'p') ADVANCE(766);
      END_STATE();
    case 361:
      if (lookahead == 'r') ADVANCE(328);
      END_STATE();
    case 362:
      if (lookahead == 'r') ADVANCE(280);
      END_STATE();
    case 363:
      if (lookahead == 's') ADVANCE(498);
      END_STATE();
    case 364:
      if (lookahead == 's') ADVANCE(265);
      END_STATE();
    case 365:
      if (lookahead == 't') ADVANCE(767);
      END_STATE();
    case 366:
      if (lookahead == 't') ADVANCE(761);
      END_STATE();
    case 367:
      if (lookahead == 'u') ADVANCE(274);
      END_STATE();
    case 368:
      ADVANCE_MAP(
        'u', 391,
        'x', 386,
        '"', 230,
        '\'', 230,
        '\\', 230,
        'a', 230,
        'b', 230,
        'f', 230,
        'n', 230,
        'r', 230,
        't', 230,
        'v', 230,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(377);
      END_STATE();
    case 369:
      if (lookahead == 'w') ADVANCE(258);
      END_STATE();
    case 370:
      if (lookahead == '|') ADVANCE(540);
      END_STATE();
    case 371:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(703);
      END_STATE();
    case 372:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(704);
      END_STATE();
    case 373:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 374:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(587);
      END_STATE();
    case 375:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(230);
      END_STATE();
    case 376:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(374);
      END_STATE();
    case 377:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(375);
      END_STATE();
    case 378:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(705);
      END_STATE();
    case 379:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(588);
      END_STATE();
    case 380:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(585);
      END_STATE();
    case 381:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(586);
      END_STATE();
    case 382:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(230);
      END_STATE();
    case 383:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(379);
      END_STATE();
    case 384:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(380);
      END_STATE();
    case 385:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(381);
      END_STATE();
    case 386:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(382);
      END_STATE();
    case 387:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(384);
      END_STATE();
    case 388:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(385);
      END_STATE();
    case 389:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(386);
      END_STATE();
    case 390:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(387);
      END_STATE();
    case 391:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(389);
      END_STATE();
    case 392:
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '`') ADVANCE(255);
      END_STATE();
    case 393:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(458);
      if (lookahead == 'n') SKIP(443);
      END_STATE();
    case 394:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(467);
      if (lookahead == 'n') SKIP(440);
      END_STATE();
    case 395:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(468);
      if (lookahead == 'n') SKIP(432);
      END_STATE();
    case 396:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(473);
      if (lookahead == 'n') SKIP(434);
      END_STATE();
    case 397:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(459);
      if (lookahead == 'n') SKIP(452);
      END_STATE();
    case 398:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(460);
      if (lookahead == 'n') SKIP(445);
      END_STATE();
    case 399:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(469);
      if (lookahead == 'n') SKIP(435);
      END_STATE();
    case 400:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(470);
      if (lookahead == 'n') SKIP(436);
      END_STATE();
    case 401:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(471);
      if (lookahead == 'n') SKIP(441);
      END_STATE();
    case 402:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(474);
      if (lookahead == 'n') SKIP(442);
      END_STATE();
    case 403:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(461);
      if (lookahead == 'n') SKIP(444);
      END_STATE();
    case 404:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(462);
      if (lookahead == 'n') SKIP(446);
      END_STATE();
    case 405:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(463);
      if (lookahead == 'n') SKIP(453);
      END_STATE();
    case 406:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(464);
      if (lookahead == 'n') SKIP(447);
      END_STATE();
    case 407:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(465);
      if (lookahead == 'n') SKIP(457);
      END_STATE();
    case 408:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(466);
      if (lookahead == 'n') SKIP(456);
      END_STATE();
    case 409:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(472);
      if (lookahead == 'n') SKIP(437);
      END_STATE();
    case 410:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(476);
      if (lookahead == 'n') SKIP(433);
      END_STATE();
    case 411:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(478);
      if (lookahead == 'n') SKIP(454);
      END_STATE();
    case 412:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(475);
      if (lookahead == 'n') SKIP(438);
      END_STATE();
    case 413:
      if (eof) ADVANCE(479);
      if (lookahead == '\r') SKIP(477);
      if (lookahead == 'n') SKIP(439);
      END_STATE();
    case 414:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 775,
        '"', 595,
        '#', 579,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 581,
        '(', 500,
        ')', 503,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 537,
        '/', 573,
        '0', 700,
        '1', 572,
        ':', 486,
        '<', 575,
        '=', 582,
        '>', 519,
        '?', 507,
        '@', 680,
        '[', 520,
        '\\', 2,
        ']', 524,
        '^', 535,
        '`', 253,
        'a', 639,
        'd', 647,
        'i', 641,
        'l', 624,
        'm', 618,
        'r', 629,
        'u', 650,
        'w', 631,
        'y', 636,
        '{', 528,
        '|', 512,
        '}', 529,
        '~', 673,
      );
      if (('2' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          (' ' <= lookahead && lookahead <= ';') ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(414);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 415:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(395);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(432);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 416:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(395);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(432);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 417:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(395);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(432);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 418:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(395);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(432);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 419:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 745,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(410);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(433);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 420:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 689,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(396);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(434);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 421:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 570,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 689,
        '?', 506,
        '@', 680,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(399);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(435);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 422:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '*', 570,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 689,
        '?', 506,
        '@', 680,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(400);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(436);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 423:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 745,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(409);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(437);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 424:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(409);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(437);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 425:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(409);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(437);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 426:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        'M', 741,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(409);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(437);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 427:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(412);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(438);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 428:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(413);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(439);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 429:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(394);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(440);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 430:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(401);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(441);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 431:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 561,
        '+', 679,
        ',', 504,
        '-', 679,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(402);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(442);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 432:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 537,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(395);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(432);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 433:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(410);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(433);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 434:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 689,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '>', 687,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(396);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '}') ADVANCE(529);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(434);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 435:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 689,
        '?', 506,
        '@', 680,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(399);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(435);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 436:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '*', 570,
        '+', 659,
        ',', 504,
        '-', 661,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 576,
        '=', 481,
        '>', 689,
        '?', 506,
        '@', 680,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(400);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(436);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 437:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(409);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(437);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 438:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(412);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(438);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 439:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 539,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(413);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(439);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 440:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(394);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(440);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 441:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 686,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(401);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(514);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(441);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 442:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '!', 677,
        '"', 596,
        '#', 770,
        '$', 695,
        '%', 669,
        '&', 516,
        '\'', 681,
        '(', 500,
        '+', 659,
        ',', 504,
        '-', 662,
        '.', 538,
        '/', 690,
        '0', 700,
        ':', 486,
        '<', 683,
        '=', 481,
        '?', 506,
        '@', 680,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(402);
      if (lookahead == '^') ADVANCE(535);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(688);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '*' ||
          lookahead == '>') ADVANCE(689);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(442);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 443:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 660,
        ',', 504,
        '-', 664,
        '.', 240,
        '/', 241,
        '0', 700,
        '<', 248,
        '>', 518,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(393);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(443);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 444:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 660,
        ',', 504,
        '-', 664,
        '/', 241,
        '0', 700,
        ':', 484,
        '<', 248,
        '=', 480,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(403);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(444);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 445:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 660,
        '-', 663,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 490,
        '<', 577,
        '>', 252,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(398);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(445);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 446:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '*', 568,
        '+', 660,
        '-', 663,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 490,
        '<', 577,
        '@', 140,
        '[', 520,
      );
      if (lookahead == '\\') SKIP(404);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(446);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 447:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '.', 536,
        '/', 241,
        '0', 700,
        '<', 248,
        '=', 480,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(406);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(447);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 448:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '.', 742,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        'L', 732,
        'M', 741,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(405);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (lookahead == 'G' ||
          lookahead == 'I' ||
          lookahead == 'N' ||
          lookahead == 'Q' ||
          lookahead == 'R' ||
          lookahead == 'Z') ADVANCE(738);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(453);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 449:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        'L', 732,
        'U', 617,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(405);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(720);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'n') ADVANCE(728);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 's') ADVANCE(715);
      if (lookahead == 'u') ADVANCE(723);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(710);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(453);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 450:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        'M', 741,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(405);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(748);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(453);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 451:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        'M', 741,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(405);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'f') ADVANCE(736);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(740);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(453);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 452:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(397);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(646);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(452);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 453:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(405);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(644);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(453);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 454:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '%', 670,
        '&', 517,
        '\'', 229,
        '(', 500,
        '+', 660,
        '-', 664,
        '/', 241,
        '0', 700,
        '<', 248,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(411);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'a') ADVANCE(640);
      if (lookahead == 'd') ADVANCE(647);
      if (lookahead == 'l') ADVANCE(624);
      if (lookahead == 'm') ADVANCE(618);
      if (lookahead == 'r') ADVANCE(629);
      if (lookahead == 'u') ADVANCE(650);
      if (lookahead == 'w') ADVANCE(632);
      if (lookahead == 'y') ADVANCE(636);
      if (lookahead == '{') ADVANCE(528);
      if (lookahead == '~') ADVANCE(673);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(678);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(454);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 455:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '\'', 229,
        '(', 501,
        '*', 231,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 492,
        '<', 574,
        '=', 480,
        '>', 252,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(408);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(456);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 456:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 770,
        '$', 139,
        '\'', 229,
        '(', 501,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 241,
        '0', 700,
        ':', 492,
        '<', 574,
        '=', 480,
        '>', 252,
        '@', 140,
        '[', 521,
      );
      if (lookahead == '\\') SKIP(408);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(511);
      if (lookahead == '}') ADVANCE(529);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(456);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 457:
      if (eof) ADVANCE(479);
      ADVANCE_MAP(
        '"', 596,
        '#', 771,
        '$', 139,
        '&', 228,
        '\'', 229,
        '(', 501,
        ')', 503,
        '*', 568,
        '+', 373,
        ',', 504,
        '-', 247,
        '.', 536,
        '/', 573,
        '0', 700,
        ':', 492,
        '<', 237,
        '=', 480,
        '>', 518,
        '@', 140,
        '[', 523,
      );
      if (lookahead == '\\') SKIP(407);
      if (lookahead == ']') ADVANCE(524);
      if (lookahead == '^') ADVANCE(534);
      if (lookahead == '`') ADVANCE(253);
      if (lookahead == 'w') ADVANCE(634);
      if (lookahead == '{') ADVANCE(527);
      if (lookahead == '|') ADVANCE(513);
      if (lookahead == '}') ADVANCE(529);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(457);
      if (set_contains(sym_identifier_character_set_1, 684, lookahead)) ADVANCE(655);
      END_STATE();
    case 458:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(443);
      END_STATE();
    case 459:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(452);
      END_STATE();
    case 460:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(445);
      END_STATE();
    case 461:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(444);
      END_STATE();
    case 462:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(446);
      END_STATE();
    case 463:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(453);
      END_STATE();
    case 464:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(447);
      END_STATE();
    case 465:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(457);
      END_STATE();
    case 466:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(456);
      END_STATE();
    case 467:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(440);
      END_STATE();
    case 468:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(432);
      END_STATE();
    case 469:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(435);
      END_STATE();
    case 470:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(436);
      END_STATE();
    case 471:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(441);
      END_STATE();
    case 472:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(437);
      END_STATE();
    case 473:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(434);
      END_STATE();
    case 474:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(442);
      END_STATE();
    case 475:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(438);
      END_STATE();
    case 476:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(433);
      END_STATE();
    case 477:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(439);
      END_STATE();
    case 478:
      if (eof) ADVANCE(479);
      if (lookahead == 'n') SKIP(454);
      END_STATE();
    case 479:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 480:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 481:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 482:
      ACCEPT_TOKEN(anon_sym_LBRACK_LT);
      END_STATE();
    case 483:
      ACCEPT_TOKEN(anon_sym_GT_RBRACK);
      END_STATE();
    case 484:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 485:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(510);
      END_STATE();
    case 486:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(510);
      if (lookahead == '=') ADVANCE(694);
      if (lookahead == '>') ADVANCE(544);
      if (lookahead == '?') ADVANCE(509);
      END_STATE();
    case 487:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(510);
      if (lookahead == '>') ADVANCE(544);
      END_STATE();
    case 488:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(510);
      if (lookahead == '>') ADVANCE(544);
      if (lookahead == '?') ADVANCE(508);
      END_STATE();
    case 489:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(510);
      if (lookahead == '?') ADVANCE(508);
      END_STATE();
    case 490:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '>') ADVANCE(544);
      END_STATE();
    case 491:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '>') ADVANCE(544);
      if (lookahead == '?') ADVANCE(508);
      END_STATE();
    case 492:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '?') ADVANCE(508);
      END_STATE();
    case 493:
      ACCEPT_TOKEN(anon_sym_and);
      END_STATE();
    case 494:
      ACCEPT_TOKEN(anon_sym_and);
      if (lookahead == '!') ADVANCE(556);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 495:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 496:
      ACCEPT_TOKEN(anon_sym_let);
      if (lookahead == '!') ADVANCE(497);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 497:
      ACCEPT_TOKEN(anon_sym_let_BANG);
      END_STATE();
    case 498:
      ACCEPT_TOKEN(anon_sym_as);
      END_STATE();
    case 499:
      ACCEPT_TOKEN(anon_sym_as);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 500:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      if (lookahead == ')') ADVANCE(608);
      if (lookahead == '*') ADVANCE(750);
      if (lookahead == '.') ADVANCE(232);
      if (lookahead == '?') ADVANCE(235);
      if (lookahead == '|') ADVANCE(656);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(238);
      if (lookahead == '!' ||
          ('$' <= lookahead && lookahead <= '&') ||
          ('+' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 501:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      if (lookahead == ')') ADVANCE(608);
      if (lookahead == '*') ADVANCE(749);
      END_STATE();
    case 502:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      if (lookahead == '*') ADVANCE(749);
      END_STATE();
    case 503:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 504:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 505:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 506:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == ',') ADVANCE(678);
      if (lookahead == '-') ADVANCE(675);
      if (lookahead == '<') ADVANCE(674);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('=' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(676);
      END_STATE();
    case 507:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(685);
      if (lookahead == '<') ADVANCE(684);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '.' ||
          lookahead == '/' ||
          ('=' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 508:
      ACCEPT_TOKEN(anon_sym_COLON_QMARK);
      END_STATE();
    case 509:
      ACCEPT_TOKEN(anon_sym_COLON_QMARK);
      if (lookahead == '>') ADVANCE(545);
      END_STATE();
    case 510:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 511:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 512:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ')') ADVANCE(657);
      if (lookahead == ']') ADVANCE(526);
      if (lookahead == '|') ADVANCE(563);
      if (lookahead == '}') ADVANCE(541);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 513:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ')') ADVANCE(657);
      if (lookahead == '|') ADVANCE(562);
      END_STATE();
    case 514:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(563);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 515:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 516:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(565);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 517:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(564);
      END_STATE();
    case 518:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 519:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 520:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '<') ADVANCE(482);
      if (lookahead == ']') ADVANCE(578);
      if (lookahead == '|') ADVANCE(525);
      END_STATE();
    case 521:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '<') ADVANCE(482);
      if (lookahead == '|') ADVANCE(525);
      END_STATE();
    case 522:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == ']') ADVANCE(578);
      if (lookahead == '|') ADVANCE(525);
      END_STATE();
    case 523:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '|') ADVANCE(525);
      END_STATE();
    case 524:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 525:
      ACCEPT_TOKEN(anon_sym_LBRACK_PIPE);
      END_STATE();
    case 526:
      ACCEPT_TOKEN(anon_sym_PIPE_RBRACK);
      END_STATE();
    case 527:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 528:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '|') ADVANCE(540);
      END_STATE();
    case 529:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 530:
      ACCEPT_TOKEN(anon_sym_LT_AT);
      if (lookahead == '@') ADVANCE(533);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '?') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 531:
      ACCEPT_TOKEN(anon_sym_LT_AT);
      if (lookahead == '@') ADVANCE(532);
      END_STATE();
    case 532:
      ACCEPT_TOKEN(anon_sym_LT_AT_AT);
      END_STATE();
    case 533:
      ACCEPT_TOKEN(anon_sym_LT_AT_AT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 534:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 535:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 536:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 537:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(566);
      if (lookahead == '[') ADVANCE(554);
      END_STATE();
    case 538:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(558);
      if (lookahead == '[') ADVANCE(554);
      END_STATE();
    case 539:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '[') ADVANCE(554);
      END_STATE();
    case 540:
      ACCEPT_TOKEN(anon_sym_LBRACE_PIPE);
      END_STATE();
    case 541:
      ACCEPT_TOKEN(anon_sym_PIPE_RBRACE);
      END_STATE();
    case 542:
      ACCEPT_TOKEN(anon_sym_return_BANG);
      END_STATE();
    case 543:
      ACCEPT_TOKEN(anon_sym_yield_BANG);
      END_STATE();
    case 544:
      ACCEPT_TOKEN(anon_sym_COLON_GT);
      END_STATE();
    case 545:
      ACCEPT_TOKEN(anon_sym_COLON_QMARK_GT);
      END_STATE();
    case 546:
      ACCEPT_TOKEN(anon_sym_in);
      END_STATE();
    case 547:
      ACCEPT_TOKEN(anon_sym_in);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 548:
      ACCEPT_TOKEN(anon_sym_while_BANG);
      END_STATE();
    case 549:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 550:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 551:
      ACCEPT_TOKEN(anon_sym_match_BANG);
      END_STATE();
    case 552:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 553:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 554:
      ACCEPT_TOKEN(anon_sym_DOT_LBRACK);
      END_STATE();
    case 555:
      ACCEPT_TOKEN(anon_sym_use_BANG);
      END_STATE();
    case 556:
      ACCEPT_TOKEN(anon_sym_and_BANG);
      END_STATE();
    case 557:
      ACCEPT_TOKEN(anon_sym_do_BANG);
      END_STATE();
    case 558:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 559:
      ACCEPT_TOKEN(anon_sym_when);
      END_STATE();
    case 560:
      ACCEPT_TOKEN(anon_sym_when);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 561:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      if (lookahead == ')') ADVANCE(608);
      END_STATE();
    case 562:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 563:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 564:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 565:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 566:
      ACCEPT_TOKEN(anon_sym_DOT_DOT2);
      END_STATE();
    case 567:
      ACCEPT_TOKEN(anon_sym_DOT_DOT3);
      END_STATE();
    case 568:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 569:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == ')') ADVANCE(751);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 570:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 571:
      ACCEPT_TOKEN(anon_sym_1);
      END_STATE();
    case 572:
      ACCEPT_TOKEN(anon_sym_1);
      if (lookahead == '_') ADVANCE(702);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 573:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(754);
      END_STATE();
    case 574:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 575:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(553);
      if (lookahead == '@') ADVANCE(530);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '.' ||
          lookahead == '/' ||
          ('<' <= lookahead && lookahead <= '?') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 576:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '@') ADVANCE(530);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '?') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 577:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '@') ADVANCE(531);
      END_STATE();
    case 578:
      ACCEPT_TOKEN(anon_sym_LBRACK_RBRACK);
      END_STATE();
    case 579:
      ACCEPT_TOKEN(anon_sym_POUND);
      ADVANCE_MAP(
        'I', 765,
        'h', 262,
        'i', 267,
        'l', 272,
        'n', 358,
        'q', 367,
        'r', 762,
        't', 273,
        'w', 256,
      );
      END_STATE();
    case 580:
      ACCEPT_TOKEN(anon_sym_POUND);
      ADVANCE_MAP(
        'I', 765,
        'h', 262,
        'l', 272,
        'n', 358,
        'q', 367,
        'r', 762,
        't', 273,
        'w', 256,
      );
      END_STATE();
    case 581:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 582:
      ACCEPT_TOKEN(anon_sym_EQ2);
      END_STATE();
    case 583:
      ACCEPT_TOKEN(sym__escape_char);
      END_STATE();
    case 584:
      ACCEPT_TOKEN(sym__non_escape_char);
      END_STATE();
    case 585:
      ACCEPT_TOKEN(sym__unicodegraph_short);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(388);
      END_STATE();
    case 586:
      ACCEPT_TOKEN(sym__unicodegraph_long);
      END_STATE();
    case 587:
      ACCEPT_TOKEN(sym__trigraph);
      END_STATE();
    case 588:
      ACCEPT_TOKEN(sym__hexgraph_short);
      END_STATE();
    case 589:
      ACCEPT_TOKEN(aux_sym__simple_string_char_token1);
      END_STATE();
    case 590:
      ACCEPT_TOKEN(aux_sym_char_token1);
      END_STATE();
    case 591:
      ACCEPT_TOKEN(aux_sym_char_token1);
      if (lookahead == '\'') ADVANCE(592);
      if (lookahead == 'B') ADVANCE(590);
      END_STATE();
    case 592:
      ACCEPT_TOKEN(aux_sym_char_token1);
      if (lookahead == 'B') ADVANCE(590);
      END_STATE();
    case 593:
      ACCEPT_TOKEN(anon_sym_LBRACE2);
      END_STATE();
    case 594:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DQUOTE);
      if (lookahead == '"') ADVANCE(145);
      END_STATE();
    case 595:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 596:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(197);
      END_STATE();
    case 597:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == 'B') ADVANCE(605);
      END_STATE();
    case 598:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 599:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '\r') ADVANCE(584);
      if ((!eof && set_contains(sym__non_escape_char_character_set_1, 9, lookahead))) ADVANCE(584);
      END_STATE();
    case 600:
      ACCEPT_TOKEN(aux_sym__verbatim_string_char_token1);
      END_STATE();
    case 601:
      ACCEPT_TOKEN(anon_sym_AT_DQUOTE);
      END_STATE();
    case 602:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      if (lookahead == '"') ADVANCE(600);
      END_STATE();
    case 603:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      if (lookahead == '"') ADVANCE(600);
      if (lookahead == 'B') ADVANCE(605);
      END_STATE();
    case 604:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      if (lookahead == 'B') ADVANCE(605);
      END_STATE();
    case 605:
      ACCEPT_TOKEN(anon_sym_DQUOTEB);
      END_STATE();
    case 606:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 607:
      ACCEPT_TOKEN(anon_sym_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 608:
      ACCEPT_TOKEN(sym_unit);
      END_STATE();
    case 609:
      ACCEPT_TOKEN(sym_identifier);
      END_STATE();
    case 610:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(557);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 611:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(556);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 612:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(555);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 613:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(551);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 614:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(548);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 615:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(543);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 616:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(542);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 617:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'L') ADVANCE(734);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 618:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(652);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 619:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(633);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 620:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(494);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 621:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(495);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 622:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(615);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 623:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(611);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 624:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(651);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 625:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(642);
      if (lookahead == 'i') ADVANCE(638);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 626:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(642);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 627:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(612);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 628:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(637);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 629:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(653);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 630:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(614);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 631:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(625);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 632:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(635);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 633:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(613);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 634:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(626);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 635:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(638);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 636:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(628);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 637:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(622);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 638:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(630);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 639:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(620);
      if (lookahead == 's') ADVANCE(499);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 640:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(620);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 641:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(547);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 642:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(560);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 643:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(621);
      if (lookahead == 's') ADVANCE(499);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 644:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(621);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 645:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(616);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 646:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(623);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 647:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(610);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 648:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(645);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 649:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(499);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 650:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(627);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 651:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(496);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 652:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(619);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 653:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(654);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 654:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'u') ADVANCE(648);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 655:
      ACCEPT_TOKEN(sym_identifier);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 656:
      ACCEPT_TOKEN(anon_sym_LPAREN_PIPE);
      if (lookahead == ')') ADVANCE(658);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 657:
      ACCEPT_TOKEN(anon_sym_PIPE_RPAREN);
      END_STATE();
    case 658:
      ACCEPT_TOKEN(sym_op_identifier);
      END_STATE();
    case 659:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(666);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 660:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(665);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 661:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(668);
      if (lookahead == '>') ADVANCE(550);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 662:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(668);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 663:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(667);
      if (lookahead == '>') ADVANCE(549);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 664:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(667);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 665:
      ACCEPT_TOKEN(anon_sym_PLUS_DOT);
      END_STATE();
    case 666:
      ACCEPT_TOKEN(anon_sym_PLUS_DOT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 667:
      ACCEPT_TOKEN(anon_sym_DASH_DOT);
      END_STATE();
    case 668:
      ACCEPT_TOKEN(anon_sym_DASH_DOT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 669:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '%') ADVANCE(672);
      if (lookahead == '!' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 670:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '%') ADVANCE(671);
      END_STATE();
    case 671:
      ACCEPT_TOKEN(anon_sym_PERCENT_PERCENT);
      END_STATE();
    case 672:
      ACCEPT_TOKEN(anon_sym_PERCENT_PERCENT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 673:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 674:
      ACCEPT_TOKEN(aux_sym_prefix_op_token1);
      if (lookahead == ',') ADVANCE(678);
      if (lookahead == '-') ADVANCE(696);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(676);
      END_STATE();
    case 675:
      ACCEPT_TOKEN(aux_sym_prefix_op_token1);
      if (lookahead == ',') ADVANCE(678);
      if (lookahead == '>') ADVANCE(698);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(676);
      END_STATE();
    case 676:
      ACCEPT_TOKEN(aux_sym_prefix_op_token1);
      if (lookahead == ',') ADVANCE(678);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(676);
      END_STATE();
    case 677:
      ACCEPT_TOKEN(aux_sym_prefix_op_token1);
      if (lookahead == '=') ADVANCE(693);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(678);
      END_STATE();
    case 678:
      ACCEPT_TOKEN(aux_sym_prefix_op_token1);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(678);
      END_STATE();
    case 679:
      ACCEPT_TOKEN(aux_sym_infix_op_token1);
      END_STATE();
    case 680:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '"') ADVANCE(601);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 681:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '\'') ADVANCE(591);
      if (lookahead == '\\') ADVANCE(368);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(682);
      if (lookahead != 0 &&
          (lookahead < 0x07 || '\r' < lookahead)) ADVANCE(230);
      END_STATE();
    case 682:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '\'') ADVANCE(592);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 683:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '-') ADVANCE(553);
      if (lookahead == '@') ADVANCE(530);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '.' ||
          lookahead == '/' ||
          ('<' <= lookahead && lookahead <= '?') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 684:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '-') ADVANCE(697);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '.' ||
          lookahead == '/' ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 685:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '>') ADVANCE(699);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 686:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '@') ADVANCE(530);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '?') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 687:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == ']') ADVANCE(483);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 688:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '|') ADVANCE(563);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 689:
      ACCEPT_TOKEN(aux_sym_infix_op_token2);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 690:
      ACCEPT_TOKEN(aux_sym_infix_op_token3);
      if (lookahead == '/') ADVANCE(754);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '-' ||
          lookahead == '.' ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(691);
      END_STATE();
    case 691:
      ACCEPT_TOKEN(aux_sym_infix_op_token3);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '-' ||
          lookahead == '.' ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(691);
      END_STATE();
    case 692:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 693:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(678);
      END_STATE();
    case 694:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 695:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '"') ADVANCE(594);
      END_STATE();
    case 696:
      ACCEPT_TOKEN(anon_sym_QMARK_LT_DASH);
      if (lookahead == ',') ADVANCE(678);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(676);
      END_STATE();
    case 697:
      ACCEPT_TOKEN(anon_sym_QMARK_LT_DASH);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 698:
      ACCEPT_TOKEN(anon_sym_QMARK_DASH_GT);
      if (lookahead == ',') ADVANCE(678);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(676);
      END_STATE();
    case 699:
      ACCEPT_TOKEN(anon_sym_QMARK_DASH_GT);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(689);
      END_STATE();
    case 700:
      ACCEPT_TOKEN(sym_int);
      if (lookahead == '_') ADVANCE(702);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(371);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(372);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(378);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 701:
      ACCEPT_TOKEN(sym_int);
      if (lookahead == '_') ADVANCE(702);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 702:
      ACCEPT_TOKEN(sym_int);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(701);
      END_STATE();
    case 703:
      ACCEPT_TOKEN(sym_xint);
      if (lookahead == '_') ADVANCE(706);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(703);
      END_STATE();
    case 704:
      ACCEPT_TOKEN(sym_xint);
      if (lookahead == '_') ADVANCE(707);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(704);
      END_STATE();
    case 705:
      ACCEPT_TOKEN(sym_xint);
      if (lookahead == '_') ADVANCE(708);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(705);
      END_STATE();
    case 706:
      ACCEPT_TOKEN(sym_xint);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(703);
      END_STATE();
    case 707:
      ACCEPT_TOKEN(sym_xint);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(704);
      END_STATE();
    case 708:
      ACCEPT_TOKEN(sym_xint);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(705);
      END_STATE();
    case 709:
      ACCEPT_TOKEN(anon_sym_y);
      END_STATE();
    case 710:
      ACCEPT_TOKEN(anon_sym_y);
      if (lookahead == 'i') ADVANCE(628);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 711:
      ACCEPT_TOKEN(anon_sym_y);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 712:
      ACCEPT_TOKEN(anon_sym_uy);
      END_STATE();
    case 713:
      ACCEPT_TOKEN(anon_sym_uy);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 714:
      ACCEPT_TOKEN(anon_sym_s);
      END_STATE();
    case 715:
      ACCEPT_TOKEN(anon_sym_s);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 716:
      ACCEPT_TOKEN(anon_sym_us);
      END_STATE();
    case 717:
      ACCEPT_TOKEN(anon_sym_us);
      if (lookahead == 'e') ADVANCE(612);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 718:
      ACCEPT_TOKEN(anon_sym_us);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 719:
      ACCEPT_TOKEN(anon_sym_l);
      END_STATE();
    case 720:
      ACCEPT_TOKEN(anon_sym_l);
      if (lookahead == 'e') ADVANCE(651);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 721:
      ACCEPT_TOKEN(anon_sym_l);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 722:
      ACCEPT_TOKEN(aux_sym_uint32_token1);
      END_STATE();
    case 723:
      ACCEPT_TOKEN(aux_sym_uint32_token1);
      if (lookahead == 'L') ADVANCE(734);
      if (lookahead == 'l') ADVANCE(726);
      if (lookahead == 'n') ADVANCE(730);
      if (lookahead == 's') ADVANCE(717);
      if (lookahead == 'y') ADVANCE(713);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 724:
      ACCEPT_TOKEN(aux_sym_uint32_token1);
      if (lookahead == 'L') ADVANCE(734);
      if (lookahead == 'l') ADVANCE(726);
      if (lookahead == 'n') ADVANCE(730);
      if (lookahead == 's') ADVANCE(718);
      if (lookahead == 'y') ADVANCE(713);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 725:
      ACCEPT_TOKEN(aux_sym_uint32_token1);
      if (lookahead == 'L') ADVANCE(733);
      if (lookahead == 'l') ADVANCE(722);
      if (lookahead == 'n') ADVANCE(729);
      if (lookahead == 's') ADVANCE(716);
      if (lookahead == 'y') ADVANCE(712);
      END_STATE();
    case 726:
      ACCEPT_TOKEN(aux_sym_uint32_token1);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 727:
      ACCEPT_TOKEN(anon_sym_n);
      END_STATE();
    case 728:
      ACCEPT_TOKEN(anon_sym_n);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 729:
      ACCEPT_TOKEN(anon_sym_un);
      END_STATE();
    case 730:
      ACCEPT_TOKEN(anon_sym_un);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 731:
      ACCEPT_TOKEN(anon_sym_L);
      END_STATE();
    case 732:
      ACCEPT_TOKEN(anon_sym_L);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 733:
      ACCEPT_TOKEN(aux_sym_uint64_token1);
      END_STATE();
    case 734:
      ACCEPT_TOKEN(aux_sym_uint64_token1);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 735:
      ACCEPT_TOKEN(anon_sym_f);
      END_STATE();
    case 736:
      ACCEPT_TOKEN(anon_sym_f);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 737:
      ACCEPT_TOKEN(aux_sym_bignum_token1);
      END_STATE();
    case 738:
      ACCEPT_TOKEN(aux_sym_bignum_token1);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 739:
      ACCEPT_TOKEN(aux_sym_decimal_token1);
      END_STATE();
    case 740:
      ACCEPT_TOKEN(aux_sym_decimal_token1);
      if (lookahead == 'a') ADVANCE(652);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 741:
      ACCEPT_TOKEN(aux_sym_decimal_token1);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 742:
      ACCEPT_TOKEN(anon_sym_DOT2);
      END_STATE();
    case 743:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == '.') ADVANCE(566);
      if (lookahead == '[') ADVANCE(554);
      END_STATE();
    case 744:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == '.') ADVANCE(558);
      if (lookahead == '[') ADVANCE(554);
      END_STATE();
    case 745:
      ACCEPT_TOKEN(anon_sym_DOT2);
      if (lookahead == '[') ADVANCE(554);
      END_STATE();
    case 746:
      ACCEPT_TOKEN(aux_sym_float_token1);
      END_STATE();
    case 747:
      ACCEPT_TOKEN(aux_sym_float_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(746);
      END_STATE();
    case 748:
      ACCEPT_TOKEN(aux_sym_float_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(746);
      if (set_contains(sym_identifier_character_set_2, 801, lookahead)) ADVANCE(655);
      END_STATE();
    case 749:
      ACCEPT_TOKEN(anon_sym_LPAREN_STAR);
      END_STATE();
    case 750:
      ACCEPT_TOKEN(anon_sym_LPAREN_STAR);
      if (lookahead == ')') ADVANCE(658);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(234);
      if (lookahead == '!' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          ('<' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(235);
      END_STATE();
    case 751:
      ACCEPT_TOKEN(anon_sym_STAR_RPAREN);
      END_STATE();
    case 752:
      ACCEPT_TOKEN(sym_inline_line_comment);
      if (lookahead == '/') ADVANCE(757);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(753);
      END_STATE();
    case 753:
      ACCEPT_TOKEN(sym_inline_line_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(753);
      END_STATE();
    case 754:
      ACCEPT_TOKEN(sym_line_comment);
      if (lookahead == '/') ADVANCE(756);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(755);
      END_STATE();
    case 755:
      ACCEPT_TOKEN(sym_line_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(755);
      END_STATE();
    case 756:
      ACCEPT_TOKEN(sym_xml_doc);
      if (lookahead == '/') ADVANCE(755);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(758);
      END_STATE();
    case 757:
      ACCEPT_TOKEN(sym_xml_doc);
      if (lookahead == '/') ADVANCE(753);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(758);
      END_STATE();
    case 758:
      ACCEPT_TOKEN(sym_xml_doc);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(758);
      END_STATE();
    case 759:
      ACCEPT_TOKEN(anon_sym_POUNDnowarn);
      END_STATE();
    case 760:
      ACCEPT_TOKEN(anon_sym_POUNDwarnon);
      END_STATE();
    case 761:
      ACCEPT_TOKEN(anon_sym_POUNDlight);
      END_STATE();
    case 762:
      ACCEPT_TOKEN(anon_sym_POUNDr);
      END_STATE();
    case 763:
      ACCEPT_TOKEN(anon_sym_POUNDload);
      END_STATE();
    case 764:
      ACCEPT_TOKEN(anon_sym_POUNDtime);
      END_STATE();
    case 765:
      ACCEPT_TOKEN(anon_sym_POUNDI);
      END_STATE();
    case 766:
      ACCEPT_TOKEN(anon_sym_POUNDhelp);
      END_STATE();
    case 767:
      ACCEPT_TOKEN(anon_sym_POUNDquit);
      END_STATE();
    case 768:
      ACCEPT_TOKEN(aux_sym_fsi_directive_decl_token1);
      if (lookahead == '\n') ADVANCE(768);
      END_STATE();
    case 769:
      ACCEPT_TOKEN(aux_sym_preproc_line_token1);
      END_STATE();
    case 770:
      ACCEPT_TOKEN(aux_sym_preproc_line_token1);
      ADVANCE_MAP(
        'I', 765,
        'e', 277,
        'h', 262,
        'i', 267,
        'l', 272,
        'n', 358,
        'q', 367,
        'r', 762,
        't', 273,
        'w', 256,
      );
      END_STATE();
    case 771:
      ACCEPT_TOKEN(aux_sym_preproc_line_token1);
      ADVANCE_MAP(
        'I', 765,
        'e', 277,
        'h', 262,
        'l', 272,
        'n', 358,
        'q', 367,
        'r', 762,
        't', 273,
        'w', 256,
      );
      END_STATE();
    case 772:
      ACCEPT_TOKEN(aux_sym_preproc_line_token1);
      ADVANCE_MAP(
        'I', 765,
        'h', 262,
        'i', 267,
        'l', 272,
        'n', 358,
        'q', 367,
        'r', 762,
        't', 273,
        'w', 256,
      );
      END_STATE();
    case 773:
      ACCEPT_TOKEN(aux_sym_preproc_line_token1);
      ADVANCE_MAP(
        'I', 765,
        'h', 262,
        'l', 272,
        'n', 358,
        'q', 367,
        'r', 762,
        't', 273,
        'w', 256,
      );
      END_STATE();
    case 774:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 775:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(692);
      END_STATE();
    case 776:
      ACCEPT_TOKEN(anon_sym_POUNDif);
      END_STATE();
    case 777:
      ACCEPT_TOKEN(anon_sym_POUNDendif);
      END_STATE();
    case 778:
      ACCEPT_TOKEN(anon_sym_POUNDelse);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == 'L') ADVANCE(1);
      if (lookahead == '\\') SKIP(2);
      if (lookahead == '_') ADVANCE(3);
      if (lookahead == 'a') ADVANCE(4);
      if (lookahead == 'b') ADVANCE(5);
      if (lookahead == 'c') ADVANCE(6);
      if (lookahead == 'd') ADVANCE(7);
      if (lookahead == 'e') ADVANCE(8);
      if (lookahead == 'f') ADVANCE(9);
      if (lookahead == 'g') ADVANCE(10);
      if (lookahead == 'i') ADVANCE(11);
      if (lookahead == 'l') ADVANCE(12);
      if (lookahead == 'm') ADVANCE(13);
      if (lookahead == 'n') ADVANCE(14);
      if (lookahead == 'o') ADVANCE(15);
      if (lookahead == 'p') ADVANCE(16);
      if (lookahead == 'r') ADVANCE(17);
      if (lookahead == 's') ADVANCE(18);
      if (lookahead == 't') ADVANCE(19);
      if (lookahead == 'u') ADVANCE(20);
      if (lookahead == 'v') ADVANCE(21);
      if (lookahead == 'w') ADVANCE(22);
      if (lookahead == 'y') ADVANCE(23);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(24);
      END_STATE();
    case 1:
      if (lookahead == 'F') ADVANCE(25);
      END_STATE();
    case 2:
      if (lookahead == '\r') SKIP(26);
      if (lookahead == 'n') SKIP(24);
      END_STATE();
    case 3:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 4:
      if (lookahead == 'b') ADVANCE(27);
      if (lookahead == 's') ADVANCE(28);
      END_STATE();
    case 5:
      if (lookahead == 'e') ADVANCE(29);
      END_STATE();
    case 6:
      if (lookahead == 'l') ADVANCE(30);
      if (lookahead == 'o') ADVANCE(31);
      END_STATE();
    case 7:
      if (lookahead == 'e') ADVANCE(32);
      if (lookahead == 'o') ADVANCE(33);
      END_STATE();
    case 8:
      if (lookahead == 'l') ADVANCE(34);
      if (lookahead == 'n') ADVANCE(35);
      if (lookahead == 'q') ADVANCE(36);
      if (lookahead == 'x') ADVANCE(37);
      END_STATE();
    case 9:
      if (lookahead == 'a') ADVANCE(38);
      if (lookahead == 'i') ADVANCE(39);
      if (lookahead == 'o') ADVANCE(40);
      if (lookahead == 'u') ADVANCE(41);
      END_STATE();
    case 10:
      if (lookahead == 'e') ADVANCE(42);
      if (lookahead == 'l') ADVANCE(43);
      END_STATE();
    case 11:
      if (lookahead == 'f') ADVANCE(44);
      if (lookahead == 'n') ADVANCE(45);
      END_STATE();
    case 12:
      if (lookahead == 'a') ADVANCE(46);
      if (lookahead == 'f') ADVANCE(47);
      END_STATE();
    case 13:
      if (lookahead == 'a') ADVANCE(48);
      if (lookahead == 'e') ADVANCE(49);
      if (lookahead == 'o') ADVANCE(50);
      if (lookahead == 'u') ADVANCE(51);
      END_STATE();
    case 14:
      if (lookahead == 'a') ADVANCE(52);
      if (lookahead == 'e') ADVANCE(53);
      if (lookahead == 'o') ADVANCE(54);
      if (lookahead == 'u') ADVANCE(55);
      END_STATE();
    case 15:
      if (lookahead == 'f') ADVANCE(56);
      if (lookahead == 'p') ADVANCE(57);
      if (lookahead == 'r') ADVANCE(58);
      if (lookahead == 'v') ADVANCE(59);
      END_STATE();
    case 16:
      if (lookahead == 'r') ADVANCE(60);
      if (lookahead == 'u') ADVANCE(61);
      END_STATE();
    case 17:
      if (lookahead == 'e') ADVANCE(62);
      END_STATE();
    case 18:
      if (lookahead == 'e') ADVANCE(63);
      if (lookahead == 't') ADVANCE(64);
      END_STATE();
    case 19:
      if (lookahead == 'h') ADVANCE(65);
      if (lookahead == 'o') ADVANCE(66);
      if (lookahead == 'r') ADVANCE(67);
      if (lookahead == 'y') ADVANCE(68);
      END_STATE();
    case 20:
      if (lookahead == 'n') ADVANCE(69);
      if (lookahead == 'p') ADVANCE(70);
      if (lookahead == 's') ADVANCE(71);
      END_STATE();
    case 21:
      if (lookahead == 'a') ADVANCE(72);
      END_STATE();
    case 22:
      if (lookahead == 'h') ADVANCE(73);
      if (lookahead == 'i') ADVANCE(74);
      END_STATE();
    case 23:
      if (lookahead == 'i') ADVANCE(75);
      END_STATE();
    case 24:
      if (lookahead == '\\') SKIP(2);
      if (lookahead == '_') ADVANCE(3);
      if (lookahead == 'a') ADVANCE(4);
      if (lookahead == 'b') ADVANCE(5);
      if (lookahead == 'c') ADVANCE(6);
      if (lookahead == 'd') ADVANCE(7);
      if (lookahead == 'e') ADVANCE(8);
      if (lookahead == 'f') ADVANCE(9);
      if (lookahead == 'g') ADVANCE(10);
      if (lookahead == 'i') ADVANCE(11);
      if (lookahead == 'l') ADVANCE(76);
      if (lookahead == 'm') ADVANCE(13);
      if (lookahead == 'n') ADVANCE(14);
      if (lookahead == 'o') ADVANCE(15);
      if (lookahead == 'p') ADVANCE(16);
      if (lookahead == 'r') ADVANCE(17);
      if (lookahead == 's') ADVANCE(18);
      if (lookahead == 't') ADVANCE(19);
      if (lookahead == 'u') ADVANCE(20);
      if (lookahead == 'v') ADVANCE(21);
      if (lookahead == 'w') ADVANCE(22);
      if (lookahead == 'y') ADVANCE(23);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == ';' ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(24);
      END_STATE();
    case 25:
      ACCEPT_TOKEN(anon_sym_LF);
      END_STATE();
    case 26:
      if (lookahead == 'n') SKIP(24);
      END_STATE();
    case 27:
      if (lookahead == 's') ADVANCE(77);
      END_STATE();
    case 28:
      if (lookahead == 's') ADVANCE(78);
      END_STATE();
    case 29:
      if (lookahead == 'g') ADVANCE(79);
      END_STATE();
    case 30:
      if (lookahead == 'a') ADVANCE(80);
      END_STATE();
    case 31:
      if (lookahead == 'm') ADVANCE(81);
      END_STATE();
    case 32:
      if (lookahead == 'f') ADVANCE(82);
      if (lookahead == 'l') ADVANCE(83);
      END_STATE();
    case 33:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'n') ADVANCE(84);
      if (lookahead == 'w') ADVANCE(85);
      END_STATE();
    case 34:
      if (lookahead == 'i') ADVANCE(86);
      if (lookahead == 's') ADVANCE(87);
      END_STATE();
    case 35:
      if (lookahead == 'd') ADVANCE(88);
      if (lookahead == 'u') ADVANCE(89);
      END_STATE();
    case 36:
      if (lookahead == 'u') ADVANCE(90);
      END_STATE();
    case 37:
      if (lookahead == 'c') ADVANCE(91);
      if (lookahead == 't') ADVANCE(92);
      END_STATE();
    case 38:
      if (lookahead == 'l') ADVANCE(93);
      END_STATE();
    case 39:
      if (lookahead == 'n') ADVANCE(94);
      if (lookahead == 'x') ADVANCE(95);
      END_STATE();
    case 40:
      if (lookahead == 'r') ADVANCE(96);
      END_STATE();
    case 41:
      if (lookahead == 'n') ADVANCE(97);
      END_STATE();
    case 42:
      if (lookahead == 't') ADVANCE(98);
      END_STATE();
    case 43:
      if (lookahead == 'o') ADVANCE(99);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 45:
      if (lookahead == 'h') ADVANCE(100);
      if (lookahead == 'l') ADVANCE(101);
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 46:
      if (lookahead == 'z') ADVANCE(103);
      END_STATE();
    case 47:
      ACCEPT_TOKEN(anon_sym_lf);
      END_STATE();
    case 48:
      if (lookahead == 't') ADVANCE(104);
      END_STATE();
    case 49:
      if (lookahead == 'm') ADVANCE(105);
      END_STATE();
    case 50:
      if (lookahead == 'd') ADVANCE(106);
      END_STATE();
    case 51:
      if (lookahead == 't') ADVANCE(107);
      END_STATE();
    case 52:
      if (lookahead == 'm') ADVANCE(108);
      END_STATE();
    case 53:
      if (lookahead == 'w') ADVANCE(109);
      END_STATE();
    case 54:
      if (lookahead == 't') ADVANCE(110);
      END_STATE();
    case 55:
      if (lookahead == 'l') ADVANCE(111);
      END_STATE();
    case 56:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 57:
      if (lookahead == 'e') ADVANCE(112);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 59:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 60:
      if (lookahead == 'i') ADVANCE(114);
      END_STATE();
    case 61:
      if (lookahead == 'b') ADVANCE(115);
      END_STATE();
    case 62:
      if (lookahead == 'c') ADVANCE(116);
      if (lookahead == 't') ADVANCE(117);
      END_STATE();
    case 63:
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 64:
      if (lookahead == 'a') ADVANCE(119);
      if (lookahead == 'r') ADVANCE(120);
      END_STATE();
    case 65:
      if (lookahead == 'e') ADVANCE(121);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_to);
      END_STATE();
    case 67:
      if (lookahead == 'u') ADVANCE(122);
      if (lookahead == 'y') ADVANCE(123);
      END_STATE();
    case 68:
      if (lookahead == 'p') ADVANCE(124);
      END_STATE();
    case 69:
      if (lookahead == 'i') ADVANCE(125);
      if (lookahead == 'm') ADVANCE(126);
      END_STATE();
    case 70:
      if (lookahead == 'c') ADVANCE(127);
      END_STATE();
    case 71:
      if (lookahead == 'e') ADVANCE(128);
      END_STATE();
    case 72:
      if (lookahead == 'l') ADVANCE(129);
      END_STATE();
    case 73:
      if (lookahead == 'i') ADVANCE(130);
      END_STATE();
    case 74:
      if (lookahead == 't') ADVANCE(131);
      END_STATE();
    case 75:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 76:
      if (lookahead == 'a') ADVANCE(46);
      END_STATE();
    case 77:
      if (lookahead == 't') ADVANCE(133);
      END_STATE();
    case 78:
      if (lookahead == 'e') ADVANCE(134);
      END_STATE();
    case 79:
      if (lookahead == 'i') ADVANCE(135);
      END_STATE();
    case 80:
      if (lookahead == 's') ADVANCE(136);
      END_STATE();
    case 81:
      if (lookahead == 'p') ADVANCE(137);
      END_STATE();
    case 82:
      if (lookahead == 'a') ADVANCE(138);
      END_STATE();
    case 83:
      if (lookahead == 'e') ADVANCE(139);
      END_STATE();
    case 84:
      if (lookahead == 'e') ADVANCE(140);
      END_STATE();
    case 85:
      if (lookahead == 'n') ADVANCE(141);
      END_STATE();
    case 86:
      if (lookahead == 'f') ADVANCE(142);
      END_STATE();
    case 87:
      if (lookahead == 'e') ADVANCE(143);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_end);
      END_STATE();
    case 89:
      if (lookahead == 'm') ADVANCE(144);
      END_STATE();
    case 90:
      if (lookahead == 'a') ADVANCE(145);
      END_STATE();
    case 91:
      if (lookahead == 'e') ADVANCE(146);
      END_STATE();
    case 92:
      if (lookahead == 'e') ADVANCE(147);
      END_STATE();
    case 93:
      if (lookahead == 's') ADVANCE(148);
      END_STATE();
    case 94:
      if (lookahead == 'a') ADVANCE(149);
      END_STATE();
    case 95:
      if (lookahead == 'e') ADVANCE(150);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_fun);
      if (lookahead == 'c') ADVANCE(151);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_get);
      END_STATE();
    case 99:
      if (lookahead == 'b') ADVANCE(152);
      END_STATE();
    case 100:
      if (lookahead == 'e') ADVANCE(153);
      END_STATE();
    case 101:
      if (lookahead == 'i') ADVANCE(154);
      END_STATE();
    case 102:
      if (lookahead == 'e') ADVANCE(155);
      END_STATE();
    case 103:
      if (lookahead == 'y') ADVANCE(156);
      END_STATE();
    case 104:
      if (lookahead == 'c') ADVANCE(157);
      END_STATE();
    case 105:
      if (lookahead == 'b') ADVANCE(158);
      END_STATE();
    case 106:
      if (lookahead == 'u') ADVANCE(159);
      END_STATE();
    case 107:
      if (lookahead == 'a') ADVANCE(160);
      END_STATE();
    case 108:
      if (lookahead == 'e') ADVANCE(161);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(anon_sym_not);
      END_STATE();
    case 111:
      if (lookahead == 'l') ADVANCE(162);
      END_STATE();
    case 112:
      if (lookahead == 'n') ADVANCE(163);
      END_STATE();
    case 113:
      if (lookahead == 'r') ADVANCE(164);
      END_STATE();
    case 114:
      if (lookahead == 'v') ADVANCE(165);
      END_STATE();
    case 115:
      if (lookahead == 'l') ADVANCE(166);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(anon_sym_rec);
      END_STATE();
    case 117:
      if (lookahead == 'u') ADVANCE(167);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_set);
      END_STATE();
    case 119:
      if (lookahead == 't') ADVANCE(168);
      END_STATE();
    case 120:
      if (lookahead == 'u') ADVANCE(169);
      END_STATE();
    case 121:
      if (lookahead == 'n') ADVANCE(170);
      END_STATE();
    case 122:
      if (lookahead == 'e') ADVANCE(171);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 124:
      if (lookahead == 'e') ADVANCE(172);
      END_STATE();
    case 125:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 126:
      if (lookahead == 'a') ADVANCE(174);
      END_STATE();
    case 127:
      if (lookahead == 'a') ADVANCE(175);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_use);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_val);
      END_STATE();
    case 130:
      if (lookahead == 'l') ADVANCE(176);
      END_STATE();
    case 131:
      if (lookahead == 'h') ADVANCE(177);
      END_STATE();
    case 132:
      if (lookahead == 'l') ADVANCE(178);
      END_STATE();
    case 133:
      if (lookahead == 'r') ADVANCE(179);
      END_STATE();
    case 134:
      if (lookahead == 'r') ADVANCE(180);
      END_STATE();
    case 135:
      if (lookahead == 'n') ADVANCE(181);
      END_STATE();
    case 136:
      if (lookahead == 's') ADVANCE(182);
      END_STATE();
    case 137:
      if (lookahead == 'a') ADVANCE(183);
      END_STATE();
    case 138:
      if (lookahead == 'u') ADVANCE(184);
      END_STATE();
    case 139:
      if (lookahead == 'g') ADVANCE(185);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_done);
      END_STATE();
    case 141:
      if (lookahead == 'c') ADVANCE(186);
      if (lookahead == 't') ADVANCE(187);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_elif);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_enum);
      END_STATE();
    case 145:
      if (lookahead == 'l') ADVANCE(188);
      END_STATE();
    case 146:
      if (lookahead == 'p') ADVANCE(189);
      END_STATE();
    case 147:
      if (lookahead == 'r') ADVANCE(190);
      END_STATE();
    case 148:
      if (lookahead == 'e') ADVANCE(171);
      END_STATE();
    case 149:
      if (lookahead == 'l') ADVANCE(191);
      END_STATE();
    case 150:
      if (lookahead == 'd') ADVANCE(192);
      END_STATE();
    case 151:
      if (lookahead == 't') ADVANCE(193);
      END_STATE();
    case 152:
      if (lookahead == 'a') ADVANCE(194);
      END_STATE();
    case 153:
      if (lookahead == 'r') ADVANCE(195);
      END_STATE();
    case 154:
      if (lookahead == 'n') ADVANCE(196);
      END_STATE();
    case 155:
      if (lookahead == 'r') ADVANCE(197);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_lazy);
      END_STATE();
    case 157:
      if (lookahead == 'h') ADVANCE(198);
      END_STATE();
    case 158:
      if (lookahead == 'e') ADVANCE(199);
      END_STATE();
    case 159:
      if (lookahead == 'l') ADVANCE(200);
      END_STATE();
    case 160:
      if (lookahead == 'b') ADVANCE(201);
      END_STATE();
    case 161:
      if (lookahead == 's') ADVANCE(202);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_null);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_open);
      END_STATE();
    case 164:
      if (lookahead == 'r') ADVANCE(203);
      END_STATE();
    case 165:
      if (lookahead == 'a') ADVANCE(204);
      END_STATE();
    case 166:
      if (lookahead == 'i') ADVANCE(205);
      END_STATE();
    case 167:
      if (lookahead == 'r') ADVANCE(206);
      END_STATE();
    case 168:
      if (lookahead == 'i') ADVANCE(207);
      END_STATE();
    case 169:
      if (lookahead == 'c') ADVANCE(208);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_then);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(sym_bool);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_unit);
      END_STATE();
    case 174:
      if (lookahead == 'n') ADVANCE(209);
      END_STATE();
    case 175:
      if (lookahead == 's') ADVANCE(210);
      END_STATE();
    case 176:
      if (lookahead == 'e') ADVANCE(211);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_with);
      END_STATE();
    case 178:
      if (lookahead == 'd') ADVANCE(212);
      END_STATE();
    case 179:
      if (lookahead == 'a') ADVANCE(213);
      END_STATE();
    case 180:
      if (lookahead == 't') ADVANCE(214);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_begin);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 183:
      if (lookahead == 'r') ADVANCE(215);
      END_STATE();
    case 184:
      if (lookahead == 'l') ADVANCE(216);
      END_STATE();
    case 185:
      if (lookahead == 'a') ADVANCE(217);
      END_STATE();
    case 186:
      if (lookahead == 'a') ADVANCE(218);
      END_STATE();
    case 187:
      if (lookahead == 'o') ADVANCE(219);
      END_STATE();
    case 188:
      if (lookahead == 'i') ADVANCE(220);
      END_STATE();
    case 189:
      if (lookahead == 't') ADVANCE(221);
      END_STATE();
    case 190:
      if (lookahead == 'n') ADVANCE(222);
      END_STATE();
    case 191:
      if (lookahead == 'l') ADVANCE(223);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_fixed);
      END_STATE();
    case 193:
      if (lookahead == 'i') ADVANCE(224);
      END_STATE();
    case 194:
      if (lookahead == 'l') ADVANCE(225);
      END_STATE();
    case 195:
      if (lookahead == 'i') ADVANCE(226);
      END_STATE();
    case 196:
      if (lookahead == 'e') ADVANCE(227);
      END_STATE();
    case 197:
      if (lookahead == 'f') ADVANCE(228);
      if (lookahead == 'n') ADVANCE(229);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_match);
      END_STATE();
    case 199:
      if (lookahead == 'r') ADVANCE(230);
      END_STATE();
    case 200:
      if (lookahead == 'e') ADVANCE(231);
      END_STATE();
    case 201:
      if (lookahead == 'l') ADVANCE(232);
      END_STATE();
    case 202:
      if (lookahead == 'p') ADVANCE(233);
      END_STATE();
    case 203:
      if (lookahead == 'i') ADVANCE(234);
      END_STATE();
    case 204:
      if (lookahead == 't') ADVANCE(235);
      END_STATE();
    case 205:
      if (lookahead == 'c') ADVANCE(236);
      END_STATE();
    case 206:
      if (lookahead == 'n') ADVANCE(237);
      END_STATE();
    case 207:
      if (lookahead == 'c') ADVANCE(238);
      END_STATE();
    case 208:
      if (lookahead == 't') ADVANCE(239);
      END_STATE();
    case 209:
      if (lookahead == 'a') ADVANCE(240);
      END_STATE();
    case 210:
      if (lookahead == 't') ADVANCE(241);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_yield);
      END_STATE();
    case 213:
      if (lookahead == 'c') ADVANCE(242);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 215:
      if (lookahead == 'i') ADVANCE(243);
      END_STATE();
    case 216:
      if (lookahead == 't') ADVANCE(244);
      END_STATE();
    case 217:
      if (lookahead == 't') ADVANCE(245);
      END_STATE();
    case 218:
      if (lookahead == 's') ADVANCE(246);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_downto);
      END_STATE();
    case 220:
      if (lookahead == 't') ADVANCE(247);
      END_STATE();
    case 221:
      if (lookahead == 'i') ADVANCE(248);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_extern);
      END_STATE();
    case 223:
      if (lookahead == 'y') ADVANCE(249);
      END_STATE();
    case 224:
      if (lookahead == 'o') ADVANCE(250);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_global);
      END_STATE();
    case 226:
      if (lookahead == 't') ADVANCE(251);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_inline);
      END_STATE();
    case 228:
      if (lookahead == 'a') ADVANCE(252);
      END_STATE();
    case 229:
      if (lookahead == 'a') ADVANCE(253);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_member);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_module);
      END_STATE();
    case 232:
      if (lookahead == 'e') ADVANCE(254);
      END_STATE();
    case 233:
      if (lookahead == 'a') ADVANCE(255);
      END_STATE();
    case 234:
      if (lookahead == 'd') ADVANCE(256);
      END_STATE();
    case 235:
      if (lookahead == 'e') ADVANCE(236);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(aux_sym_access_modifier_token1);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_static);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 240:
      if (lookahead == 'g') ADVANCE(257);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_upcast);
      END_STATE();
    case 242:
      if (lookahead == 't') ADVANCE(258);
      END_STATE();
    case 243:
      if (lookahead == 's') ADVANCE(259);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 245:
      if (lookahead == 'e') ADVANCE(260);
      END_STATE();
    case 246:
      if (lookahead == 't') ADVANCE(261);
      END_STATE();
    case 247:
      if (lookahead == 'y') ADVANCE(262);
      END_STATE();
    case 248:
      if (lookahead == 'o') ADVANCE(263);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_finally);
      END_STATE();
    case 250:
      if (lookahead == 'n') ADVANCE(264);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_inherit);
      END_STATE();
    case 252:
      if (lookahead == 'c') ADVANCE(265);
      END_STATE();
    case 253:
      if (lookahead == 'l') ADVANCE(236);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_mutable);
      END_STATE();
    case 255:
      if (lookahead == 'c') ADVANCE(266);
      END_STATE();
    case 256:
      if (lookahead == 'e') ADVANCE(267);
      END_STATE();
    case 257:
      if (lookahead == 'e') ADVANCE(268);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_abstract);
      END_STATE();
    case 259:
      if (lookahead == 'o') ADVANCE(269);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_delegate);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(anon_sym_downcast);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(anon_sym_equality);
      END_STATE();
    case 263:
      if (lookahead == 'n') ADVANCE(270);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(anon_sym_function);
      END_STATE();
    case 265:
      if (lookahead == 'e') ADVANCE(271);
      END_STATE();
    case 266:
      if (lookahead == 'e') ADVANCE(272);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(anon_sym_override);
      END_STATE();
    case 268:
      if (lookahead == 'd') ADVANCE(273);
      END_STATE();
    case 269:
      if (lookahead == 'n') ADVANCE(274);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(anon_sym_exception);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(anon_sym_interface);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(anon_sym_namespace);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(anon_sym_unmanaged);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(anon_sym_comparison);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 443, .external_lex_state = 2},
  [2] = {.lex_state = 429, .external_lex_state = 3},
  [3] = {.lex_state = 429, .external_lex_state = 4},
  [4] = {.lex_state = 429, .external_lex_state = 3},
  [5] = {.lex_state = 429, .external_lex_state = 3},
  [6] = {.lex_state = 429, .external_lex_state = 3},
  [7] = {.lex_state = 429, .external_lex_state = 3},
  [8] = {.lex_state = 429, .external_lex_state = 4},
  [9] = {.lex_state = 429, .external_lex_state = 3},
  [10] = {.lex_state = 429, .external_lex_state = 3},
  [11] = {.lex_state = 429, .external_lex_state = 3},
  [12] = {.lex_state = 429, .external_lex_state = 3},
  [13] = {.lex_state = 429, .external_lex_state = 3},
  [14] = {.lex_state = 429, .external_lex_state = 3},
  [15] = {.lex_state = 429, .external_lex_state = 3},
  [16] = {.lex_state = 429, .external_lex_state = 3},
  [17] = {.lex_state = 429, .external_lex_state = 3},
  [18] = {.lex_state = 429, .external_lex_state = 3},
  [19] = {.lex_state = 429, .external_lex_state = 3},
  [20] = {.lex_state = 429, .external_lex_state = 3},
  [21] = {.lex_state = 429, .external_lex_state = 4},
  [22] = {.lex_state = 429, .external_lex_state = 4},
  [23] = {.lex_state = 429, .external_lex_state = 4},
  [24] = {.lex_state = 429, .external_lex_state = 4},
  [25] = {.lex_state = 429, .external_lex_state = 4},
  [26] = {.lex_state = 429, .external_lex_state = 4},
  [27] = {.lex_state = 429, .external_lex_state = 4},
  [28] = {.lex_state = 429, .external_lex_state = 4},
  [29] = {.lex_state = 429, .external_lex_state = 4},
  [30] = {.lex_state = 429, .external_lex_state = 4},
  [31] = {.lex_state = 429, .external_lex_state = 4},
  [32] = {.lex_state = 429, .external_lex_state = 4},
  [33] = {.lex_state = 429, .external_lex_state = 4},
  [34] = {.lex_state = 429, .external_lex_state = 4},
  [35] = {.lex_state = 429, .external_lex_state = 4},
  [36] = {.lex_state = 429, .external_lex_state = 4},
  [37] = {.lex_state = 429, .external_lex_state = 5},
  [38] = {.lex_state = 429, .external_lex_state = 5},
  [39] = {.lex_state = 429, .external_lex_state = 5},
  [40] = {.lex_state = 429, .external_lex_state = 5},
  [41] = {.lex_state = 429, .external_lex_state = 5},
  [42] = {.lex_state = 429, .external_lex_state = 4},
  [43] = {.lex_state = 429, .external_lex_state = 4},
  [44] = {.lex_state = 429, .external_lex_state = 4},
  [45] = {.lex_state = 429, .external_lex_state = 4},
  [46] = {.lex_state = 429, .external_lex_state = 4},
  [47] = {.lex_state = 429, .external_lex_state = 5},
  [48] = {.lex_state = 429, .external_lex_state = 5},
  [49] = {.lex_state = 429, .external_lex_state = 5},
  [50] = {.lex_state = 429, .external_lex_state = 5},
  [51] = {.lex_state = 429, .external_lex_state = 5},
  [52] = {.lex_state = 429, .external_lex_state = 5},
  [53] = {.lex_state = 429, .external_lex_state = 5},
  [54] = {.lex_state = 429, .external_lex_state = 5},
  [55] = {.lex_state = 429, .external_lex_state = 5},
  [56] = {.lex_state = 429, .external_lex_state = 5},
  [57] = {.lex_state = 429, .external_lex_state = 5},
  [58] = {.lex_state = 429, .external_lex_state = 5},
  [59] = {.lex_state = 429, .external_lex_state = 4},
  [60] = {.lex_state = 429, .external_lex_state = 4},
  [61] = {.lex_state = 429, .external_lex_state = 4},
  [62] = {.lex_state = 429, .external_lex_state = 4},
  [63] = {.lex_state = 429, .external_lex_state = 4},
  [64] = {.lex_state = 429, .external_lex_state = 4},
  [65] = {.lex_state = 429, .external_lex_state = 4},
  [66] = {.lex_state = 418, .external_lex_state = 6},
  [67] = {.lex_state = 418, .external_lex_state = 6},
  [68] = {.lex_state = 429, .external_lex_state = 3},
  [69] = {.lex_state = 429, .external_lex_state = 3},
  [70] = {.lex_state = 429, .external_lex_state = 3},
  [71] = {.lex_state = 429, .external_lex_state = 3},
  [72] = {.lex_state = 429, .external_lex_state = 3},
  [73] = {.lex_state = 429, .external_lex_state = 3},
  [74] = {.lex_state = 429, .external_lex_state = 3},
  [75] = {.lex_state = 429, .external_lex_state = 3},
  [76] = {.lex_state = 429, .external_lex_state = 3},
  [77] = {.lex_state = 418, .external_lex_state = 6},
  [78] = {.lex_state = 418, .external_lex_state = 6},
  [79] = {.lex_state = 418, .external_lex_state = 6},
  [80] = {.lex_state = 418, .external_lex_state = 6},
  [81] = {.lex_state = 418, .external_lex_state = 6},
  [82] = {.lex_state = 418, .external_lex_state = 6},
  [83] = {.lex_state = 429, .external_lex_state = 3},
  [84] = {.lex_state = 418, .external_lex_state = 6},
  [85] = {.lex_state = 418, .external_lex_state = 6},
  [86] = {.lex_state = 418, .external_lex_state = 6},
  [87] = {.lex_state = 429, .external_lex_state = 3},
  [88] = {.lex_state = 429, .external_lex_state = 3},
  [89] = {.lex_state = 418, .external_lex_state = 6},
  [90] = {.lex_state = 429, .external_lex_state = 3},
  [91] = {.lex_state = 429, .external_lex_state = 3},
  [92] = {.lex_state = 418, .external_lex_state = 6},
  [93] = {.lex_state = 418, .external_lex_state = 6},
  [94] = {.lex_state = 429, .external_lex_state = 3},
  [95] = {.lex_state = 418, .external_lex_state = 6},
  [96] = {.lex_state = 429, .external_lex_state = 7},
  [97] = {.lex_state = 429, .external_lex_state = 7},
  [98] = {.lex_state = 429, .external_lex_state = 7},
  [99] = {.lex_state = 429, .external_lex_state = 7},
  [100] = {.lex_state = 429, .external_lex_state = 5},
  [101] = {.lex_state = 418, .external_lex_state = 8},
  [102] = {.lex_state = 429, .external_lex_state = 7},
  [103] = {.lex_state = 429, .external_lex_state = 5},
  [104] = {.lex_state = 429, .external_lex_state = 5},
  [105] = {.lex_state = 429, .external_lex_state = 7},
  [106] = {.lex_state = 429, .external_lex_state = 5},
  [107] = {.lex_state = 429, .external_lex_state = 5},
  [108] = {.lex_state = 429, .external_lex_state = 5},
  [109] = {.lex_state = 429, .external_lex_state = 5},
  [110] = {.lex_state = 429, .external_lex_state = 5},
  [111] = {.lex_state = 429, .external_lex_state = 5},
  [112] = {.lex_state = 429, .external_lex_state = 7},
  [113] = {.lex_state = 429, .external_lex_state = 5},
  [114] = {.lex_state = 429, .external_lex_state = 5},
  [115] = {.lex_state = 429, .external_lex_state = 7},
  [116] = {.lex_state = 429, .external_lex_state = 7},
  [117] = {.lex_state = 429, .external_lex_state = 7},
  [118] = {.lex_state = 418, .external_lex_state = 8},
  [119] = {.lex_state = 429, .external_lex_state = 5},
  [120] = {.lex_state = 429, .external_lex_state = 7},
  [121] = {.lex_state = 429, .external_lex_state = 5},
  [122] = {.lex_state = 418, .external_lex_state = 8},
  [123] = {.lex_state = 418, .external_lex_state = 8},
  [124] = {.lex_state = 418, .external_lex_state = 8},
  [125] = {.lex_state = 418, .external_lex_state = 8},
  [126] = {.lex_state = 418, .external_lex_state = 8},
  [127] = {.lex_state = 418, .external_lex_state = 8},
  [128] = {.lex_state = 418, .external_lex_state = 8},
  [129] = {.lex_state = 418, .external_lex_state = 8},
  [130] = {.lex_state = 418, .external_lex_state = 8},
  [131] = {.lex_state = 418, .external_lex_state = 8},
  [132] = {.lex_state = 418, .external_lex_state = 8},
  [133] = {.lex_state = 418, .external_lex_state = 8},
  [134] = {.lex_state = 429, .external_lex_state = 7},
  [135] = {.lex_state = 429, .external_lex_state = 7},
  [136] = {.lex_state = 418, .external_lex_state = 8},
  [137] = {.lex_state = 429, .external_lex_state = 5},
  [138] = {.lex_state = 429, .external_lex_state = 7},
  [139] = {.lex_state = 429, .external_lex_state = 5},
  [140] = {.lex_state = 429, .external_lex_state = 7},
  [141] = {.lex_state = 418, .external_lex_state = 9},
  [142] = {.lex_state = 418, .external_lex_state = 9},
  [143] = {.lex_state = 418, .external_lex_state = 9},
  [144] = {.lex_state = 418, .external_lex_state = 9},
  [145] = {.lex_state = 418, .external_lex_state = 9},
  [146] = {.lex_state = 418, .external_lex_state = 9},
  [147] = {.lex_state = 418, .external_lex_state = 9},
  [148] = {.lex_state = 418, .external_lex_state = 9},
  [149] = {.lex_state = 418, .external_lex_state = 9},
  [150] = {.lex_state = 418, .external_lex_state = 9},
  [151] = {.lex_state = 418, .external_lex_state = 9},
  [152] = {.lex_state = 418, .external_lex_state = 9},
  [153] = {.lex_state = 418, .external_lex_state = 9},
  [154] = {.lex_state = 418, .external_lex_state = 9},
  [155] = {.lex_state = 418, .external_lex_state = 9},
  [156] = {.lex_state = 418, .external_lex_state = 9},
  [157] = {.lex_state = 418, .external_lex_state = 9},
  [158] = {.lex_state = 418, .external_lex_state = 9},
  [159] = {.lex_state = 418, .external_lex_state = 9},
  [160] = {.lex_state = 418, .external_lex_state = 9},
  [161] = {.lex_state = 418, .external_lex_state = 9},
  [162] = {.lex_state = 418, .external_lex_state = 10},
  [163] = {.lex_state = 418, .external_lex_state = 10},
  [164] = {.lex_state = 418, .external_lex_state = 10},
  [165] = {.lex_state = 418, .external_lex_state = 10},
  [166] = {.lex_state = 418, .external_lex_state = 11},
  [167] = {.lex_state = 418, .external_lex_state = 10},
  [168] = {.lex_state = 418, .external_lex_state = 10},
  [169] = {.lex_state = 418, .external_lex_state = 10},
  [170] = {.lex_state = 418, .external_lex_state = 10},
  [171] = {.lex_state = 418, .external_lex_state = 10},
  [172] = {.lex_state = 418, .external_lex_state = 10},
  [173] = {.lex_state = 418, .external_lex_state = 10},
  [174] = {.lex_state = 418, .external_lex_state = 10},
  [175] = {.lex_state = 418, .external_lex_state = 11},
  [176] = {.lex_state = 418, .external_lex_state = 11},
  [177] = {.lex_state = 418, .external_lex_state = 11},
  [178] = {.lex_state = 418, .external_lex_state = 11},
  [179] = {.lex_state = 418, .external_lex_state = 11},
  [180] = {.lex_state = 418, .external_lex_state = 11},
  [181] = {.lex_state = 418, .external_lex_state = 11},
  [182] = {.lex_state = 418, .external_lex_state = 11},
  [183] = {.lex_state = 418, .external_lex_state = 11},
  [184] = {.lex_state = 418, .external_lex_state = 11},
  [185] = {.lex_state = 418, .external_lex_state = 11},
  [186] = {.lex_state = 418, .external_lex_state = 11},
  [187] = {.lex_state = 418, .external_lex_state = 11},
  [188] = {.lex_state = 418, .external_lex_state = 10},
  [189] = {.lex_state = 418, .external_lex_state = 10},
  [190] = {.lex_state = 418, .external_lex_state = 10},
  [191] = {.lex_state = 418, .external_lex_state = 11},
  [192] = {.lex_state = 418, .external_lex_state = 11},
  [193] = {.lex_state = 418, .external_lex_state = 11},
  [194] = {.lex_state = 418, .external_lex_state = 11},
  [195] = {.lex_state = 418, .external_lex_state = 11},
  [196] = {.lex_state = 418, .external_lex_state = 12},
  [197] = {.lex_state = 418, .external_lex_state = 12},
  [198] = {.lex_state = 418, .external_lex_state = 12},
  [199] = {.lex_state = 418, .external_lex_state = 11},
  [200] = {.lex_state = 418, .external_lex_state = 12},
  [201] = {.lex_state = 418, .external_lex_state = 11},
  [202] = {.lex_state = 418, .external_lex_state = 11},
  [203] = {.lex_state = 418, .external_lex_state = 12},
  [204] = {.lex_state = 418, .external_lex_state = 11},
  [205] = {.lex_state = 418, .external_lex_state = 11},
  [206] = {.lex_state = 418, .external_lex_state = 11},
  [207] = {.lex_state = 418, .external_lex_state = 11},
  [208] = {.lex_state = 418, .external_lex_state = 11},
  [209] = {.lex_state = 418, .external_lex_state = 11},
  [210] = {.lex_state = 418, .external_lex_state = 11},
  [211] = {.lex_state = 418, .external_lex_state = 11},
  [212] = {.lex_state = 418, .external_lex_state = 12},
  [213] = {.lex_state = 418, .external_lex_state = 12},
  [214] = {.lex_state = 418, .external_lex_state = 12},
  [215] = {.lex_state = 418, .external_lex_state = 12},
  [216] = {.lex_state = 418, .external_lex_state = 12},
  [217] = {.lex_state = 418, .external_lex_state = 12},
  [218] = {.lex_state = 418, .external_lex_state = 12},
  [219] = {.lex_state = 418, .external_lex_state = 12},
  [220] = {.lex_state = 418, .external_lex_state = 12},
  [221] = {.lex_state = 418, .external_lex_state = 12},
  [222] = {.lex_state = 418, .external_lex_state = 11},
  [223] = {.lex_state = 418, .external_lex_state = 11},
  [224] = {.lex_state = 418, .external_lex_state = 11},
  [225] = {.lex_state = 108, .external_lex_state = 9},
  [226] = {.lex_state = 418, .external_lex_state = 11},
  [227] = {.lex_state = 418, .external_lex_state = 11},
  [228] = {.lex_state = 418, .external_lex_state = 11},
  [229] = {.lex_state = 418, .external_lex_state = 11},
  [230] = {.lex_state = 418, .external_lex_state = 11},
  [231] = {.lex_state = 418, .external_lex_state = 11},
  [232] = {.lex_state = 418, .external_lex_state = 11},
  [233] = {.lex_state = 418, .external_lex_state = 11},
  [234] = {.lex_state = 418, .external_lex_state = 11},
  [235] = {.lex_state = 418, .external_lex_state = 11},
  [236] = {.lex_state = 418, .external_lex_state = 13},
  [237] = {.lex_state = 418, .external_lex_state = 13},
  [238] = {.lex_state = 108, .external_lex_state = 9},
  [239] = {.lex_state = 108, .external_lex_state = 9},
  [240] = {.lex_state = 108, .external_lex_state = 9},
  [241] = {.lex_state = 108, .external_lex_state = 9},
  [242] = {.lex_state = 108, .external_lex_state = 9},
  [243] = {.lex_state = 108, .external_lex_state = 9},
  [244] = {.lex_state = 108, .external_lex_state = 9},
  [245] = {.lex_state = 108, .external_lex_state = 9},
  [246] = {.lex_state = 108, .external_lex_state = 9},
  [247] = {.lex_state = 108, .external_lex_state = 9},
  [248] = {.lex_state = 108, .external_lex_state = 9},
  [249] = {.lex_state = 108, .external_lex_state = 9},
  [250] = {.lex_state = 108, .external_lex_state = 9},
  [251] = {.lex_state = 420, .external_lex_state = 9},
  [252] = {.lex_state = 418, .external_lex_state = 13},
  [253] = {.lex_state = 418, .external_lex_state = 13},
  [254] = {.lex_state = 418, .external_lex_state = 13},
  [255] = {.lex_state = 418, .external_lex_state = 13},
  [256] = {.lex_state = 418, .external_lex_state = 13},
  [257] = {.lex_state = 420, .external_lex_state = 9},
  [258] = {.lex_state = 420, .external_lex_state = 9},
  [259] = {.lex_state = 418, .external_lex_state = 13},
  [260] = {.lex_state = 108, .external_lex_state = 14},
  [261] = {.lex_state = 420, .external_lex_state = 9},
  [262] = {.lex_state = 420, .external_lex_state = 9},
  [263] = {.lex_state = 418, .external_lex_state = 13},
  [264] = {.lex_state = 418, .external_lex_state = 13},
  [265] = {.lex_state = 418, .external_lex_state = 13},
  [266] = {.lex_state = 418, .external_lex_state = 13},
  [267] = {.lex_state = 418, .external_lex_state = 13},
  [268] = {.lex_state = 420, .external_lex_state = 9},
  [269] = {.lex_state = 420, .external_lex_state = 9},
  [270] = {.lex_state = 418, .external_lex_state = 13},
  [271] = {.lex_state = 420, .external_lex_state = 9},
  [272] = {.lex_state = 420, .external_lex_state = 9},
  [273] = {.lex_state = 418, .external_lex_state = 13},
  [274] = {.lex_state = 420, .external_lex_state = 9},
  [275] = {.lex_state = 420, .external_lex_state = 9},
  [276] = {.lex_state = 420, .external_lex_state = 9},
  [277] = {.lex_state = 418, .external_lex_state = 13},
  [278] = {.lex_state = 418, .external_lex_state = 13},
  [279] = {.lex_state = 420, .external_lex_state = 9},
  [280] = {.lex_state = 420, .external_lex_state = 9},
  [281] = {.lex_state = 418, .external_lex_state = 13},
  [282] = {.lex_state = 420, .external_lex_state = 9},
  [283] = {.lex_state = 420, .external_lex_state = 9},
  [284] = {.lex_state = 420, .external_lex_state = 9},
  [285] = {.lex_state = 418, .external_lex_state = 13},
  [286] = {.lex_state = 418, .external_lex_state = 15},
  [287] = {.lex_state = 418, .external_lex_state = 15},
  [288] = {.lex_state = 420, .external_lex_state = 13},
  [289] = {.lex_state = 420, .external_lex_state = 13},
  [290] = {.lex_state = 420, .external_lex_state = 13},
  [291] = {.lex_state = 418, .external_lex_state = 15},
  [292] = {.lex_state = 418, .external_lex_state = 15},
  [293] = {.lex_state = 418, .external_lex_state = 15},
  [294] = {.lex_state = 418, .external_lex_state = 15},
  [295] = {.lex_state = 418, .external_lex_state = 15},
  [296] = {.lex_state = 420, .external_lex_state = 13},
  [297] = {.lex_state = 418, .external_lex_state = 15},
  [298] = {.lex_state = 420, .external_lex_state = 13},
  [299] = {.lex_state = 108, .external_lex_state = 14},
  [300] = {.lex_state = 420, .external_lex_state = 13},
  [301] = {.lex_state = 420, .external_lex_state = 13},
  [302] = {.lex_state = 420, .external_lex_state = 13},
  [303] = {.lex_state = 420, .external_lex_state = 13},
  [304] = {.lex_state = 420, .external_lex_state = 13},
  [305] = {.lex_state = 420, .external_lex_state = 13},
  [306] = {.lex_state = 420, .external_lex_state = 13},
  [307] = {.lex_state = 420, .external_lex_state = 13},
  [308] = {.lex_state = 420, .external_lex_state = 13},
  [309] = {.lex_state = 420, .external_lex_state = 13},
  [310] = {.lex_state = 108, .external_lex_state = 14},
  [311] = {.lex_state = 108, .external_lex_state = 14},
  [312] = {.lex_state = 108, .external_lex_state = 14},
  [313] = {.lex_state = 108, .external_lex_state = 14},
  [314] = {.lex_state = 108, .external_lex_state = 14},
  [315] = {.lex_state = 418, .external_lex_state = 11},
  [316] = {.lex_state = 108, .external_lex_state = 14},
  [317] = {.lex_state = 418, .external_lex_state = 15},
  [318] = {.lex_state = 418, .external_lex_state = 10},
  [319] = {.lex_state = 418, .external_lex_state = 10},
  [320] = {.lex_state = 108, .external_lex_state = 14},
  [321] = {.lex_state = 443, .external_lex_state = 16},
  [322] = {.lex_state = 418, .external_lex_state = 11},
  [323] = {.lex_state = 418, .external_lex_state = 11},
  [324] = {.lex_state = 418, .external_lex_state = 11},
  [325] = {.lex_state = 418, .external_lex_state = 11},
  [326] = {.lex_state = 418, .external_lex_state = 11},
  [327] = {.lex_state = 418, .external_lex_state = 11},
  [328] = {.lex_state = 418, .external_lex_state = 11},
  [329] = {.lex_state = 418, .external_lex_state = 11},
  [330] = {.lex_state = 418, .external_lex_state = 11},
  [331] = {.lex_state = 418, .external_lex_state = 11},
  [332] = {.lex_state = 418, .external_lex_state = 11},
  [333] = {.lex_state = 418, .external_lex_state = 11},
  [334] = {.lex_state = 418, .external_lex_state = 11},
  [335] = {.lex_state = 108, .external_lex_state = 14},
  [336] = {.lex_state = 108, .external_lex_state = 14},
  [337] = {.lex_state = 108, .external_lex_state = 14},
  [338] = {.lex_state = 418, .external_lex_state = 10},
  [339] = {.lex_state = 420, .external_lex_state = 13},
  [340] = {.lex_state = 418, .external_lex_state = 11},
  [341] = {.lex_state = 418, .external_lex_state = 10},
  [342] = {.lex_state = 418, .external_lex_state = 10},
  [343] = {.lex_state = 420, .external_lex_state = 13},
  [344] = {.lex_state = 418, .external_lex_state = 13},
  [345] = {.lex_state = 418, .external_lex_state = 13},
  [346] = {.lex_state = 418, .external_lex_state = 10},
  [347] = {.lex_state = 418, .external_lex_state = 10},
  [348] = {.lex_state = 418, .external_lex_state = 10},
  [349] = {.lex_state = 418, .external_lex_state = 10},
  [350] = {.lex_state = 418, .external_lex_state = 10},
  [351] = {.lex_state = 418, .external_lex_state = 10},
  [352] = {.lex_state = 418, .external_lex_state = 10},
  [353] = {.lex_state = 418, .external_lex_state = 10},
  [354] = {.lex_state = 418, .external_lex_state = 10},
  [355] = {.lex_state = 418, .external_lex_state = 11},
  [356] = {.lex_state = 418, .external_lex_state = 10},
  [357] = {.lex_state = 418, .external_lex_state = 10},
  [358] = {.lex_state = 109, .external_lex_state = 11},
  [359] = {.lex_state = 109, .external_lex_state = 11},
  [360] = {.lex_state = 109, .external_lex_state = 11},
  [361] = {.lex_state = 109, .external_lex_state = 11},
  [362] = {.lex_state = 109, .external_lex_state = 11},
  [363] = {.lex_state = 109, .external_lex_state = 11},
  [364] = {.lex_state = 109, .external_lex_state = 11},
  [365] = {.lex_state = 109, .external_lex_state = 11},
  [366] = {.lex_state = 109, .external_lex_state = 11},
  [367] = {.lex_state = 109, .external_lex_state = 11},
  [368] = {.lex_state = 418, .external_lex_state = 10},
  [369] = {.lex_state = 109, .external_lex_state = 11},
  [370] = {.lex_state = 418, .external_lex_state = 11},
  [371] = {.lex_state = 418, .external_lex_state = 11},
  [372] = {.lex_state = 420, .external_lex_state = 13},
  [373] = {.lex_state = 418, .external_lex_state = 15},
  [374] = {.lex_state = 443, .external_lex_state = 16},
  [375] = {.lex_state = 418, .external_lex_state = 10},
  [376] = {.lex_state = 109, .external_lex_state = 11},
  [377] = {.lex_state = 418, .external_lex_state = 15},
  [378] = {.lex_state = 418, .external_lex_state = 10},
  [379] = {.lex_state = 420, .external_lex_state = 13},
  [380] = {.lex_state = 418, .external_lex_state = 15},
  [381] = {.lex_state = 418, .external_lex_state = 11},
  [382] = {.lex_state = 418, .external_lex_state = 15},
  [383] = {.lex_state = 443, .external_lex_state = 16},
  [384] = {.lex_state = 418, .external_lex_state = 13},
  [385] = {.lex_state = 418, .external_lex_state = 13},
  [386] = {.lex_state = 418, .external_lex_state = 13},
  [387] = {.lex_state = 418, .external_lex_state = 13},
  [388] = {.lex_state = 418, .external_lex_state = 13},
  [389] = {.lex_state = 418, .external_lex_state = 13},
  [390] = {.lex_state = 418, .external_lex_state = 13},
  [391] = {.lex_state = 418, .external_lex_state = 13},
  [392] = {.lex_state = 418, .external_lex_state = 13},
  [393] = {.lex_state = 418, .external_lex_state = 13},
  [394] = {.lex_state = 418, .external_lex_state = 13},
  [395] = {.lex_state = 418, .external_lex_state = 13},
  [396] = {.lex_state = 418, .external_lex_state = 13},
  [397] = {.lex_state = 418, .external_lex_state = 11},
  [398] = {.lex_state = 418, .external_lex_state = 15},
  [399] = {.lex_state = 443, .external_lex_state = 16},
  [400] = {.lex_state = 418, .external_lex_state = 11},
  [401] = {.lex_state = 418, .external_lex_state = 15},
  [402] = {.lex_state = 108, .external_lex_state = 14},
  [403] = {.lex_state = 418, .external_lex_state = 11},
  [404] = {.lex_state = 418, .external_lex_state = 15},
  [405] = {.lex_state = 418, .external_lex_state = 11},
  [406] = {.lex_state = 418, .external_lex_state = 15},
  [407] = {.lex_state = 418, .external_lex_state = 11},
  [408] = {.lex_state = 418, .external_lex_state = 15},
  [409] = {.lex_state = 418, .external_lex_state = 11},
  [410] = {.lex_state = 418, .external_lex_state = 15},
  [411] = {.lex_state = 418, .external_lex_state = 11},
  [412] = {.lex_state = 418, .external_lex_state = 15},
  [413] = {.lex_state = 418, .external_lex_state = 11},
  [414] = {.lex_state = 418, .external_lex_state = 15},
  [415] = {.lex_state = 418, .external_lex_state = 11},
  [416] = {.lex_state = 418, .external_lex_state = 15},
  [417] = {.lex_state = 418, .external_lex_state = 11},
  [418] = {.lex_state = 418, .external_lex_state = 15},
  [419] = {.lex_state = 418, .external_lex_state = 11},
  [420] = {.lex_state = 418, .external_lex_state = 15},
  [421] = {.lex_state = 418, .external_lex_state = 11},
  [422] = {.lex_state = 418, .external_lex_state = 15},
  [423] = {.lex_state = 418, .external_lex_state = 11},
  [424] = {.lex_state = 418, .external_lex_state = 11},
  [425] = {.lex_state = 418, .external_lex_state = 15},
  [426] = {.lex_state = 418, .external_lex_state = 11},
  [427] = {.lex_state = 418, .external_lex_state = 15},
  [428] = {.lex_state = 418, .external_lex_state = 11},
  [429] = {.lex_state = 418, .external_lex_state = 15},
  [430] = {.lex_state = 418, .external_lex_state = 11},
  [431] = {.lex_state = 418, .external_lex_state = 15},
  [432] = {.lex_state = 418, .external_lex_state = 11},
  [433] = {.lex_state = 418, .external_lex_state = 15},
  [434] = {.lex_state = 418, .external_lex_state = 11},
  [435] = {.lex_state = 418, .external_lex_state = 15},
  [436] = {.lex_state = 418, .external_lex_state = 11},
  [437] = {.lex_state = 418, .external_lex_state = 15},
  [438] = {.lex_state = 418, .external_lex_state = 11},
  [439] = {.lex_state = 418, .external_lex_state = 15},
  [440] = {.lex_state = 418, .external_lex_state = 11},
  [441] = {.lex_state = 418, .external_lex_state = 15},
  [442] = {.lex_state = 418, .external_lex_state = 11},
  [443] = {.lex_state = 418, .external_lex_state = 15},
  [444] = {.lex_state = 418, .external_lex_state = 11},
  [445] = {.lex_state = 418, .external_lex_state = 15},
  [446] = {.lex_state = 418, .external_lex_state = 11},
  [447] = {.lex_state = 418, .external_lex_state = 15},
  [448] = {.lex_state = 418, .external_lex_state = 15},
  [449] = {.lex_state = 418, .external_lex_state = 15},
  [450] = {.lex_state = 418, .external_lex_state = 15},
  [451] = {.lex_state = 418, .external_lex_state = 15},
  [452] = {.lex_state = 418, .external_lex_state = 15},
  [453] = {.lex_state = 418, .external_lex_state = 15},
  [454] = {.lex_state = 418, .external_lex_state = 11},
  [455] = {.lex_state = 418, .external_lex_state = 15},
  [456] = {.lex_state = 418, .external_lex_state = 15},
  [457] = {.lex_state = 418, .external_lex_state = 15},
  [458] = {.lex_state = 418, .external_lex_state = 15},
  [459] = {.lex_state = 418, .external_lex_state = 15},
  [460] = {.lex_state = 109, .external_lex_state = 11},
  [461] = {.lex_state = 418, .external_lex_state = 15},
  [462] = {.lex_state = 418, .external_lex_state = 15},
  [463] = {.lex_state = 109, .external_lex_state = 11},
  [464] = {.lex_state = 109, .external_lex_state = 11},
  [465] = {.lex_state = 109, .external_lex_state = 11},
  [466] = {.lex_state = 418, .external_lex_state = 15},
  [467] = {.lex_state = 418, .external_lex_state = 15},
  [468] = {.lex_state = 108, .external_lex_state = 14},
  [469] = {.lex_state = 109, .external_lex_state = 11},
  [470] = {.lex_state = 418, .external_lex_state = 11},
  [471] = {.lex_state = 418, .external_lex_state = 11},
  [472] = {.lex_state = 418, .external_lex_state = 11},
  [473] = {.lex_state = 418, .external_lex_state = 11},
  [474] = {.lex_state = 418, .external_lex_state = 11},
  [475] = {.lex_state = 418, .external_lex_state = 11},
  [476] = {.lex_state = 418, .external_lex_state = 11},
  [477] = {.lex_state = 418, .external_lex_state = 11},
  [478] = {.lex_state = 418, .external_lex_state = 11},
  [479] = {.lex_state = 418, .external_lex_state = 11},
  [480] = {.lex_state = 418, .external_lex_state = 11},
  [481] = {.lex_state = 418, .external_lex_state = 11},
  [482] = {.lex_state = 418, .external_lex_state = 11},
  [483] = {.lex_state = 418, .external_lex_state = 11},
  [484] = {.lex_state = 418, .external_lex_state = 11},
  [485] = {.lex_state = 418, .external_lex_state = 11},
  [486] = {.lex_state = 418, .external_lex_state = 11},
  [487] = {.lex_state = 418, .external_lex_state = 11},
  [488] = {.lex_state = 418, .external_lex_state = 11},
  [489] = {.lex_state = 418, .external_lex_state = 11},
  [490] = {.lex_state = 418, .external_lex_state = 11},
  [491] = {.lex_state = 418, .external_lex_state = 11},
  [492] = {.lex_state = 418, .external_lex_state = 11},
  [493] = {.lex_state = 418, .external_lex_state = 11},
  [494] = {.lex_state = 443, .external_lex_state = 16},
  [495] = {.lex_state = 418, .external_lex_state = 17},
  [496] = {.lex_state = 418, .external_lex_state = 18},
  [497] = {.lex_state = 420, .external_lex_state = 11},
  [498] = {.lex_state = 420, .external_lex_state = 11},
  [499] = {.lex_state = 420, .external_lex_state = 11},
  [500] = {.lex_state = 420, .external_lex_state = 11},
  [501] = {.lex_state = 420, .external_lex_state = 11},
  [502] = {.lex_state = 420, .external_lex_state = 11},
  [503] = {.lex_state = 420, .external_lex_state = 11},
  [504] = {.lex_state = 420, .external_lex_state = 11},
  [505] = {.lex_state = 420, .external_lex_state = 11},
  [506] = {.lex_state = 420, .external_lex_state = 11},
  [507] = {.lex_state = 420, .external_lex_state = 11},
  [508] = {.lex_state = 418, .external_lex_state = 17},
  [509] = {.lex_state = 418, .external_lex_state = 14},
  [510] = {.lex_state = 418, .external_lex_state = 17},
  [511] = {.lex_state = 418, .external_lex_state = 17},
  [512] = {.lex_state = 418, .external_lex_state = 17},
  [513] = {.lex_state = 418, .external_lex_state = 17},
  [514] = {.lex_state = 418, .external_lex_state = 17},
  [515] = {.lex_state = 418, .external_lex_state = 17},
  [516] = {.lex_state = 418, .external_lex_state = 17},
  [517] = {.lex_state = 418, .external_lex_state = 17},
  [518] = {.lex_state = 418, .external_lex_state = 17},
  [519] = {.lex_state = 418, .external_lex_state = 11},
  [520] = {.lex_state = 86, .external_lex_state = 11},
  [521] = {.lex_state = 418, .external_lex_state = 17},
  [522] = {.lex_state = 86, .external_lex_state = 11},
  [523] = {.lex_state = 86, .external_lex_state = 11},
  [524] = {.lex_state = 86, .external_lex_state = 11},
  [525] = {.lex_state = 420, .external_lex_state = 11},
  [526] = {.lex_state = 418, .external_lex_state = 18},
  [527] = {.lex_state = 418, .external_lex_state = 19},
  [528] = {.lex_state = 418, .external_lex_state = 11},
  [529] = {.lex_state = 418, .external_lex_state = 13},
  [530] = {.lex_state = 443, .external_lex_state = 16},
  [531] = {.lex_state = 418, .external_lex_state = 12},
  [532] = {.lex_state = 86, .external_lex_state = 11},
  [533] = {.lex_state = 418, .external_lex_state = 17},
  [534] = {.lex_state = 86, .external_lex_state = 11},
  [535] = {.lex_state = 418, .external_lex_state = 12},
  [536] = {.lex_state = 418, .external_lex_state = 13},
  [537] = {.lex_state = 418, .external_lex_state = 12},
  [538] = {.lex_state = 418, .external_lex_state = 12},
  [539] = {.lex_state = 86, .external_lex_state = 11},
  [540] = {.lex_state = 418, .external_lex_state = 17},
  [541] = {.lex_state = 86, .external_lex_state = 11},
  [542] = {.lex_state = 86, .external_lex_state = 11},
  [543] = {.lex_state = 86, .external_lex_state = 11},
  [544] = {.lex_state = 86, .external_lex_state = 11},
  [545] = {.lex_state = 86, .external_lex_state = 11},
  [546] = {.lex_state = 86, .external_lex_state = 11},
  [547] = {.lex_state = 418, .external_lex_state = 11},
  [548] = {.lex_state = 109, .external_lex_state = 11},
  [549] = {.lex_state = 418, .external_lex_state = 13},
  [550] = {.lex_state = 418, .external_lex_state = 18},
  [551] = {.lex_state = 418, .external_lex_state = 18},
  [552] = {.lex_state = 418, .external_lex_state = 18},
  [553] = {.lex_state = 418, .external_lex_state = 12},
  [554] = {.lex_state = 418, .external_lex_state = 18},
  [555] = {.lex_state = 418, .external_lex_state = 18},
  [556] = {.lex_state = 109, .external_lex_state = 11},
  [557] = {.lex_state = 109, .external_lex_state = 11},
  [558] = {.lex_state = 109, .external_lex_state = 11},
  [559] = {.lex_state = 418, .external_lex_state = 18},
  [560] = {.lex_state = 418, .external_lex_state = 18},
  [561] = {.lex_state = 418, .external_lex_state = 18},
  [562] = {.lex_state = 418, .external_lex_state = 18},
  [563] = {.lex_state = 418, .external_lex_state = 18},
  [564] = {.lex_state = 109, .external_lex_state = 11},
  [565] = {.lex_state = 109, .external_lex_state = 11},
  [566] = {.lex_state = 418, .external_lex_state = 18},
  [567] = {.lex_state = 109, .external_lex_state = 11},
  [568] = {.lex_state = 418, .external_lex_state = 12},
  [569] = {.lex_state = 418, .external_lex_state = 12},
  [570] = {.lex_state = 418, .external_lex_state = 12},
  [571] = {.lex_state = 109, .external_lex_state = 11},
  [572] = {.lex_state = 109, .external_lex_state = 11},
  [573] = {.lex_state = 109, .external_lex_state = 11},
  [574] = {.lex_state = 109, .external_lex_state = 11},
  [575] = {.lex_state = 109, .external_lex_state = 11},
  [576] = {.lex_state = 418, .external_lex_state = 12},
  [577] = {.lex_state = 109, .external_lex_state = 11},
  [578] = {.lex_state = 418, .external_lex_state = 12},
  [579] = {.lex_state = 418, .external_lex_state = 19},
  [580] = {.lex_state = 418, .external_lex_state = 12},
  [581] = {.lex_state = 420, .external_lex_state = 11},
  [582] = {.lex_state = 418, .external_lex_state = 17},
  [583] = {.lex_state = 418, .external_lex_state = 18},
  [584] = {.lex_state = 418, .external_lex_state = 19},
  [585] = {.lex_state = 418, .external_lex_state = 11},
  [586] = {.lex_state = 418, .external_lex_state = 13},
  [587] = {.lex_state = 443, .external_lex_state = 16},
  [588] = {.lex_state = 418, .external_lex_state = 18},
  [589] = {.lex_state = 418, .external_lex_state = 18},
  [590] = {.lex_state = 418, .external_lex_state = 12},
  [591] = {.lex_state = 418, .external_lex_state = 13},
  [592] = {.lex_state = 418, .external_lex_state = 12},
  [593] = {.lex_state = 418, .external_lex_state = 12},
  [594] = {.lex_state = 418, .external_lex_state = 12},
  [595] = {.lex_state = 418, .external_lex_state = 12},
  [596] = {.lex_state = 418, .external_lex_state = 11},
  [597] = {.lex_state = 418, .external_lex_state = 12},
  [598] = {.lex_state = 418, .external_lex_state = 13},
  [599] = {.lex_state = 418, .external_lex_state = 12},
  [600] = {.lex_state = 418, .external_lex_state = 17},
  [601] = {.lex_state = 418, .external_lex_state = 18},
  [602] = {.lex_state = 418, .external_lex_state = 13},
  [603] = {.lex_state = 443, .external_lex_state = 16},
  [604] = {.lex_state = 418, .external_lex_state = 17},
  [605] = {.lex_state = 418, .external_lex_state = 18},
  [606] = {.lex_state = 418, .external_lex_state = 13},
  [607] = {.lex_state = 443, .external_lex_state = 16},
  [608] = {.lex_state = 86, .external_lex_state = 11},
  [609] = {.lex_state = 418, .external_lex_state = 17},
  [610] = {.lex_state = 418, .external_lex_state = 18},
  [611] = {.lex_state = 418, .external_lex_state = 13},
  [612] = {.lex_state = 418, .external_lex_state = 13},
  [613] = {.lex_state = 418, .external_lex_state = 17},
  [614] = {.lex_state = 418, .external_lex_state = 18},
  [615] = {.lex_state = 418, .external_lex_state = 13},
  [616] = {.lex_state = 418, .external_lex_state = 11},
  [617] = {.lex_state = 418, .external_lex_state = 13},
  [618] = {.lex_state = 418, .external_lex_state = 17},
  [619] = {.lex_state = 418, .external_lex_state = 18},
  [620] = {.lex_state = 418, .external_lex_state = 13},
  [621] = {.lex_state = 418, .external_lex_state = 13},
  [622] = {.lex_state = 418, .external_lex_state = 13},
  [623] = {.lex_state = 418, .external_lex_state = 17},
  [624] = {.lex_state = 418, .external_lex_state = 18},
  [625] = {.lex_state = 418, .external_lex_state = 13},
  [626] = {.lex_state = 109, .external_lex_state = 11},
  [627] = {.lex_state = 418, .external_lex_state = 17},
  [628] = {.lex_state = 418, .external_lex_state = 18},
  [629] = {.lex_state = 418, .external_lex_state = 13},
  [630] = {.lex_state = 109, .external_lex_state = 11},
  [631] = {.lex_state = 418, .external_lex_state = 17},
  [632] = {.lex_state = 418, .external_lex_state = 18},
  [633] = {.lex_state = 418, .external_lex_state = 13},
  [634] = {.lex_state = 418, .external_lex_state = 17},
  [635] = {.lex_state = 418, .external_lex_state = 18},
  [636] = {.lex_state = 418, .external_lex_state = 13},
  [637] = {.lex_state = 418, .external_lex_state = 14},
  [638] = {.lex_state = 418, .external_lex_state = 17},
  [639] = {.lex_state = 418, .external_lex_state = 18},
  [640] = {.lex_state = 418, .external_lex_state = 13},
  [641] = {.lex_state = 418, .external_lex_state = 12},
  [642] = {.lex_state = 418, .external_lex_state = 14},
  [643] = {.lex_state = 418, .external_lex_state = 17},
  [644] = {.lex_state = 418, .external_lex_state = 18},
  [645] = {.lex_state = 418, .external_lex_state = 13},
  [646] = {.lex_state = 418, .external_lex_state = 14},
  [647] = {.lex_state = 418, .external_lex_state = 14},
  [648] = {.lex_state = 418, .external_lex_state = 17},
  [649] = {.lex_state = 418, .external_lex_state = 18},
  [650] = {.lex_state = 418, .external_lex_state = 13},
  [651] = {.lex_state = 418, .external_lex_state = 14},
  [652] = {.lex_state = 418, .external_lex_state = 14},
  [653] = {.lex_state = 418, .external_lex_state = 14},
  [654] = {.lex_state = 418, .external_lex_state = 18},
  [655] = {.lex_state = 418, .external_lex_state = 13},
  [656] = {.lex_state = 418, .external_lex_state = 14},
  [657] = {.lex_state = 418, .external_lex_state = 14},
  [658] = {.lex_state = 418, .external_lex_state = 17},
  [659] = {.lex_state = 418, .external_lex_state = 18},
  [660] = {.lex_state = 418, .external_lex_state = 13},
  [661] = {.lex_state = 418, .external_lex_state = 14},
  [662] = {.lex_state = 418, .external_lex_state = 14},
  [663] = {.lex_state = 418, .external_lex_state = 17},
  [664] = {.lex_state = 418, .external_lex_state = 18},
  [665] = {.lex_state = 418, .external_lex_state = 13},
  [666] = {.lex_state = 418, .external_lex_state = 14},
  [667] = {.lex_state = 418, .external_lex_state = 14},
  [668] = {.lex_state = 418, .external_lex_state = 17},
  [669] = {.lex_state = 418, .external_lex_state = 18},
  [670] = {.lex_state = 418, .external_lex_state = 13},
  [671] = {.lex_state = 418, .external_lex_state = 17},
  [672] = {.lex_state = 418, .external_lex_state = 18},
  [673] = {.lex_state = 418, .external_lex_state = 13},
  [674] = {.lex_state = 418, .external_lex_state = 19},
  [675] = {.lex_state = 418, .external_lex_state = 17},
  [676] = {.lex_state = 418, .external_lex_state = 18},
  [677] = {.lex_state = 418, .external_lex_state = 13},
  [678] = {.lex_state = 418, .external_lex_state = 19},
  [679] = {.lex_state = 418, .external_lex_state = 17},
  [680] = {.lex_state = 418, .external_lex_state = 18},
  [681] = {.lex_state = 418, .external_lex_state = 13},
  [682] = {.lex_state = 418, .external_lex_state = 19},
  [683] = {.lex_state = 418, .external_lex_state = 19},
  [684] = {.lex_state = 418, .external_lex_state = 17},
  [685] = {.lex_state = 418, .external_lex_state = 18},
  [686] = {.lex_state = 418, .external_lex_state = 13},
  [687] = {.lex_state = 418, .external_lex_state = 19},
  [688] = {.lex_state = 418, .external_lex_state = 19},
  [689] = {.lex_state = 418, .external_lex_state = 17},
  [690] = {.lex_state = 418, .external_lex_state = 18},
  [691] = {.lex_state = 418, .external_lex_state = 13},
  [692] = {.lex_state = 418, .external_lex_state = 19},
  [693] = {.lex_state = 418, .external_lex_state = 19},
  [694] = {.lex_state = 418, .external_lex_state = 17},
  [695] = {.lex_state = 418, .external_lex_state = 18},
  [696] = {.lex_state = 418, .external_lex_state = 13},
  [697] = {.lex_state = 418, .external_lex_state = 19},
  [698] = {.lex_state = 418, .external_lex_state = 19},
  [699] = {.lex_state = 418, .external_lex_state = 17},
  [700] = {.lex_state = 418, .external_lex_state = 18},
  [701] = {.lex_state = 418, .external_lex_state = 13},
  [702] = {.lex_state = 418, .external_lex_state = 19},
  [703] = {.lex_state = 418, .external_lex_state = 19},
  [704] = {.lex_state = 418, .external_lex_state = 17},
  [705] = {.lex_state = 418, .external_lex_state = 18},
  [706] = {.lex_state = 418, .external_lex_state = 13},
  [707] = {.lex_state = 418, .external_lex_state = 19},
  [708] = {.lex_state = 418, .external_lex_state = 13},
  [709] = {.lex_state = 418, .external_lex_state = 13},
  [710] = {.lex_state = 418, .external_lex_state = 13},
  [711] = {.lex_state = 418, .external_lex_state = 13},
  [712] = {.lex_state = 418, .external_lex_state = 13},
  [713] = {.lex_state = 418, .external_lex_state = 13},
  [714] = {.lex_state = 418, .external_lex_state = 13},
  [715] = {.lex_state = 418, .external_lex_state = 13},
  [716] = {.lex_state = 418, .external_lex_state = 13},
  [717] = {.lex_state = 418, .external_lex_state = 13},
  [718] = {.lex_state = 418, .external_lex_state = 13},
  [719] = {.lex_state = 418, .external_lex_state = 13},
  [720] = {.lex_state = 418, .external_lex_state = 14},
  [721] = {.lex_state = 418, .external_lex_state = 13},
  [722] = {.lex_state = 443, .external_lex_state = 16},
  [723] = {.lex_state = 418, .external_lex_state = 13},
  [724] = {.lex_state = 418, .external_lex_state = 13},
  [725] = {.lex_state = 418, .external_lex_state = 13},
  [726] = {.lex_state = 418, .external_lex_state = 13},
  [727] = {.lex_state = 418, .external_lex_state = 13},
  [728] = {.lex_state = 418, .external_lex_state = 13},
  [729] = {.lex_state = 418, .external_lex_state = 13},
  [730] = {.lex_state = 418, .external_lex_state = 13},
  [731] = {.lex_state = 418, .external_lex_state = 13},
  [732] = {.lex_state = 418, .external_lex_state = 13},
  [733] = {.lex_state = 418, .external_lex_state = 13},
  [734] = {.lex_state = 418, .external_lex_state = 13},
  [735] = {.lex_state = 418, .external_lex_state = 13},
  [736] = {.lex_state = 418, .external_lex_state = 13},
  [737] = {.lex_state = 418, .external_lex_state = 13},
  [738] = {.lex_state = 418, .external_lex_state = 13},
  [739] = {.lex_state = 418, .external_lex_state = 13},
  [740] = {.lex_state = 418, .external_lex_state = 13},
  [741] = {.lex_state = 418, .external_lex_state = 13},
  [742] = {.lex_state = 418, .external_lex_state = 13},
  [743] = {.lex_state = 418, .external_lex_state = 13},
  [744] = {.lex_state = 418, .external_lex_state = 13},
  [745] = {.lex_state = 418, .external_lex_state = 13},
  [746] = {.lex_state = 418, .external_lex_state = 13},
  [747] = {.lex_state = 418, .external_lex_state = 13},
  [748] = {.lex_state = 418, .external_lex_state = 13},
  [749] = {.lex_state = 418, .external_lex_state = 13},
  [750] = {.lex_state = 418, .external_lex_state = 13},
  [751] = {.lex_state = 418, .external_lex_state = 13},
  [752] = {.lex_state = 418, .external_lex_state = 13},
  [753] = {.lex_state = 418, .external_lex_state = 13},
  [754] = {.lex_state = 418, .external_lex_state = 13},
  [755] = {.lex_state = 418, .external_lex_state = 13},
  [756] = {.lex_state = 418, .external_lex_state = 13},
  [757] = {.lex_state = 418, .external_lex_state = 13},
  [758] = {.lex_state = 418, .external_lex_state = 13},
  [759] = {.lex_state = 418, .external_lex_state = 13},
  [760] = {.lex_state = 418, .external_lex_state = 13},
  [761] = {.lex_state = 418, .external_lex_state = 13},
  [762] = {.lex_state = 418, .external_lex_state = 13},
  [763] = {.lex_state = 418, .external_lex_state = 13},
  [764] = {.lex_state = 418, .external_lex_state = 13},
  [765] = {.lex_state = 418, .external_lex_state = 13},
  [766] = {.lex_state = 418, .external_lex_state = 13},
  [767] = {.lex_state = 418, .external_lex_state = 13},
  [768] = {.lex_state = 418, .external_lex_state = 13},
  [769] = {.lex_state = 418, .external_lex_state = 13},
  [770] = {.lex_state = 418, .external_lex_state = 13},
  [771] = {.lex_state = 418, .external_lex_state = 13},
  [772] = {.lex_state = 418, .external_lex_state = 13},
  [773] = {.lex_state = 418, .external_lex_state = 13},
  [774] = {.lex_state = 418, .external_lex_state = 13},
  [775] = {.lex_state = 418, .external_lex_state = 13},
  [776] = {.lex_state = 418, .external_lex_state = 13},
  [777] = {.lex_state = 418, .external_lex_state = 13},
  [778] = {.lex_state = 418, .external_lex_state = 13},
  [779] = {.lex_state = 418, .external_lex_state = 13},
  [780] = {.lex_state = 418, .external_lex_state = 13},
  [781] = {.lex_state = 418, .external_lex_state = 13},
  [782] = {.lex_state = 418, .external_lex_state = 13},
  [783] = {.lex_state = 418, .external_lex_state = 13},
  [784] = {.lex_state = 418, .external_lex_state = 13},
  [785] = {.lex_state = 418, .external_lex_state = 13},
  [786] = {.lex_state = 418, .external_lex_state = 13},
  [787] = {.lex_state = 418, .external_lex_state = 13},
  [788] = {.lex_state = 418, .external_lex_state = 13},
  [789] = {.lex_state = 418, .external_lex_state = 13},
  [790] = {.lex_state = 418, .external_lex_state = 13},
  [791] = {.lex_state = 418, .external_lex_state = 13},
  [792] = {.lex_state = 418, .external_lex_state = 13},
  [793] = {.lex_state = 418, .external_lex_state = 13},
  [794] = {.lex_state = 418, .external_lex_state = 13},
  [795] = {.lex_state = 418, .external_lex_state = 13},
  [796] = {.lex_state = 418, .external_lex_state = 13},
  [797] = {.lex_state = 418, .external_lex_state = 13},
  [798] = {.lex_state = 418, .external_lex_state = 13},
  [799] = {.lex_state = 418, .external_lex_state = 13},
  [800] = {.lex_state = 418, .external_lex_state = 13},
  [801] = {.lex_state = 418, .external_lex_state = 13},
  [802] = {.lex_state = 420, .external_lex_state = 11},
  [803] = {.lex_state = 109, .external_lex_state = 11},
  [804] = {.lex_state = 418, .external_lex_state = 13},
  [805] = {.lex_state = 418, .external_lex_state = 17},
  [806] = {.lex_state = 418, .external_lex_state = 13},
  [807] = {.lex_state = 418, .external_lex_state = 13},
  [808] = {.lex_state = 418, .external_lex_state = 13},
  [809] = {.lex_state = 418, .external_lex_state = 13},
  [810] = {.lex_state = 420, .external_lex_state = 11},
  [811] = {.lex_state = 109, .external_lex_state = 11},
  [812] = {.lex_state = 418, .external_lex_state = 13},
  [813] = {.lex_state = 418, .external_lex_state = 13},
  [814] = {.lex_state = 420, .external_lex_state = 11},
  [815] = {.lex_state = 109, .external_lex_state = 11},
  [816] = {.lex_state = 418, .external_lex_state = 13},
  [817] = {.lex_state = 420, .external_lex_state = 11},
  [818] = {.lex_state = 109, .external_lex_state = 11},
  [819] = {.lex_state = 418, .external_lex_state = 13},
  [820] = {.lex_state = 420, .external_lex_state = 11},
  [821] = {.lex_state = 109, .external_lex_state = 11},
  [822] = {.lex_state = 418, .external_lex_state = 13},
  [823] = {.lex_state = 420, .external_lex_state = 11},
  [824] = {.lex_state = 109, .external_lex_state = 11},
  [825] = {.lex_state = 418, .external_lex_state = 13},
  [826] = {.lex_state = 420, .external_lex_state = 11},
  [827] = {.lex_state = 109, .external_lex_state = 11},
  [828] = {.lex_state = 418, .external_lex_state = 13},
  [829] = {.lex_state = 420, .external_lex_state = 11},
  [830] = {.lex_state = 109, .external_lex_state = 11},
  [831] = {.lex_state = 418, .external_lex_state = 13},
  [832] = {.lex_state = 420, .external_lex_state = 11},
  [833] = {.lex_state = 109, .external_lex_state = 11},
  [834] = {.lex_state = 418, .external_lex_state = 13},
  [835] = {.lex_state = 420, .external_lex_state = 11},
  [836] = {.lex_state = 109, .external_lex_state = 11},
  [837] = {.lex_state = 418, .external_lex_state = 13},
  [838] = {.lex_state = 420, .external_lex_state = 11},
  [839] = {.lex_state = 109, .external_lex_state = 11},
  [840] = {.lex_state = 418, .external_lex_state = 13},
  [841] = {.lex_state = 420, .external_lex_state = 11},
  [842] = {.lex_state = 109, .external_lex_state = 11},
  [843] = {.lex_state = 418, .external_lex_state = 13},
  [844] = {.lex_state = 420, .external_lex_state = 11},
  [845] = {.lex_state = 109, .external_lex_state = 11},
  [846] = {.lex_state = 418, .external_lex_state = 13},
  [847] = {.lex_state = 420, .external_lex_state = 11},
  [848] = {.lex_state = 109, .external_lex_state = 11},
  [849] = {.lex_state = 418, .external_lex_state = 13},
  [850] = {.lex_state = 420, .external_lex_state = 11},
  [851] = {.lex_state = 109, .external_lex_state = 11},
  [852] = {.lex_state = 418, .external_lex_state = 13},
  [853] = {.lex_state = 420, .external_lex_state = 11},
  [854] = {.lex_state = 109, .external_lex_state = 11},
  [855] = {.lex_state = 418, .external_lex_state = 13},
  [856] = {.lex_state = 420, .external_lex_state = 11},
  [857] = {.lex_state = 109, .external_lex_state = 11},
  [858] = {.lex_state = 418, .external_lex_state = 13},
  [859] = {.lex_state = 420, .external_lex_state = 11},
  [860] = {.lex_state = 109, .external_lex_state = 11},
  [861] = {.lex_state = 418, .external_lex_state = 13},
  [862] = {.lex_state = 420, .external_lex_state = 11},
  [863] = {.lex_state = 109, .external_lex_state = 11},
  [864] = {.lex_state = 418, .external_lex_state = 13},
  [865] = {.lex_state = 420, .external_lex_state = 11},
  [866] = {.lex_state = 109, .external_lex_state = 11},
  [867] = {.lex_state = 418, .external_lex_state = 13},
  [868] = {.lex_state = 420, .external_lex_state = 11},
  [869] = {.lex_state = 109, .external_lex_state = 11},
  [870] = {.lex_state = 418, .external_lex_state = 13},
  [871] = {.lex_state = 420, .external_lex_state = 11},
  [872] = {.lex_state = 109, .external_lex_state = 11},
  [873] = {.lex_state = 418, .external_lex_state = 13},
  [874] = {.lex_state = 420, .external_lex_state = 11},
  [875] = {.lex_state = 109, .external_lex_state = 11},
  [876] = {.lex_state = 418, .external_lex_state = 13},
  [877] = {.lex_state = 420, .external_lex_state = 11},
  [878] = {.lex_state = 109, .external_lex_state = 11},
  [879] = {.lex_state = 418, .external_lex_state = 13},
  [880] = {.lex_state = 420, .external_lex_state = 11},
  [881] = {.lex_state = 109, .external_lex_state = 11},
  [882] = {.lex_state = 418, .external_lex_state = 13},
  [883] = {.lex_state = 420, .external_lex_state = 11},
  [884] = {.lex_state = 109, .external_lex_state = 11},
  [885] = {.lex_state = 420, .external_lex_state = 11},
  [886] = {.lex_state = 109, .external_lex_state = 11},
  [887] = {.lex_state = 420, .external_lex_state = 11},
  [888] = {.lex_state = 420, .external_lex_state = 11},
  [889] = {.lex_state = 420, .external_lex_state = 11},
  [890] = {.lex_state = 420, .external_lex_state = 11},
  [891] = {.lex_state = 418, .external_lex_state = 14},
  [892] = {.lex_state = 418, .external_lex_state = 14},
  [893] = {.lex_state = 418, .external_lex_state = 14},
  [894] = {.lex_state = 418, .external_lex_state = 14},
  [895] = {.lex_state = 418, .external_lex_state = 14},
  [896] = {.lex_state = 418, .external_lex_state = 14},
  [897] = {.lex_state = 418, .external_lex_state = 14},
  [898] = {.lex_state = 418, .external_lex_state = 14},
  [899] = {.lex_state = 418, .external_lex_state = 14},
  [900] = {.lex_state = 418, .external_lex_state = 14},
  [901] = {.lex_state = 418, .external_lex_state = 14},
  [902] = {.lex_state = 418, .external_lex_state = 14},
  [903] = {.lex_state = 418, .external_lex_state = 14},
  [904] = {.lex_state = 418, .external_lex_state = 14},
  [905] = {.lex_state = 418, .external_lex_state = 14},
  [906] = {.lex_state = 418, .external_lex_state = 14},
  [907] = {.lex_state = 418, .external_lex_state = 14},
  [908] = {.lex_state = 418, .external_lex_state = 14},
  [909] = {.lex_state = 418, .external_lex_state = 14},
  [910] = {.lex_state = 418, .external_lex_state = 14},
  [911] = {.lex_state = 418, .external_lex_state = 14},
  [912] = {.lex_state = 418, .external_lex_state = 14},
  [913] = {.lex_state = 418, .external_lex_state = 14},
  [914] = {.lex_state = 418, .external_lex_state = 14},
  [915] = {.lex_state = 418, .external_lex_state = 14},
  [916] = {.lex_state = 418, .external_lex_state = 14},
  [917] = {.lex_state = 418, .external_lex_state = 13},
  [918] = {.lex_state = 418, .external_lex_state = 13},
  [919] = {.lex_state = 418, .external_lex_state = 13},
  [920] = {.lex_state = 418, .external_lex_state = 13},
  [921] = {.lex_state = 418, .external_lex_state = 13},
  [922] = {.lex_state = 418, .external_lex_state = 13},
  [923] = {.lex_state = 418, .external_lex_state = 13},
  [924] = {.lex_state = 418, .external_lex_state = 13},
  [925] = {.lex_state = 420, .external_lex_state = 11},
  [926] = {.lex_state = 418, .external_lex_state = 13},
  [927] = {.lex_state = 418, .external_lex_state = 13},
  [928] = {.lex_state = 420, .external_lex_state = 11},
  [929] = {.lex_state = 418, .external_lex_state = 17},
  [930] = {.lex_state = 418, .external_lex_state = 11},
  [931] = {.lex_state = 418, .external_lex_state = 11},
  [932] = {.lex_state = 418, .external_lex_state = 11},
  [933] = {.lex_state = 418, .external_lex_state = 11},
  [934] = {.lex_state = 418, .external_lex_state = 11},
  [935] = {.lex_state = 418, .external_lex_state = 11},
  [936] = {.lex_state = 418, .external_lex_state = 11},
  [937] = {.lex_state = 418, .external_lex_state = 11},
  [938] = {.lex_state = 418, .external_lex_state = 11},
  [939] = {.lex_state = 443, .external_lex_state = 16},
  [940] = {.lex_state = 418, .external_lex_state = 11},
  [941] = {.lex_state = 418, .external_lex_state = 11},
  [942] = {.lex_state = 443, .external_lex_state = 16},
  [943] = {.lex_state = 418, .external_lex_state = 11},
  [944] = {.lex_state = 418, .external_lex_state = 11},
  [945] = {.lex_state = 418, .external_lex_state = 11},
  [946] = {.lex_state = 418, .external_lex_state = 11},
  [947] = {.lex_state = 418, .external_lex_state = 11},
  [948] = {.lex_state = 418, .external_lex_state = 11},
  [949] = {.lex_state = 418, .external_lex_state = 11},
  [950] = {.lex_state = 418, .external_lex_state = 11},
  [951] = {.lex_state = 418, .external_lex_state = 11},
  [952] = {.lex_state = 418, .external_lex_state = 11},
  [953] = {.lex_state = 418, .external_lex_state = 11},
  [954] = {.lex_state = 418, .external_lex_state = 11},
  [955] = {.lex_state = 418, .external_lex_state = 11},
  [956] = {.lex_state = 418, .external_lex_state = 11},
  [957] = {.lex_state = 418, .external_lex_state = 11},
  [958] = {.lex_state = 418, .external_lex_state = 11},
  [959] = {.lex_state = 418, .external_lex_state = 11},
  [960] = {.lex_state = 418, .external_lex_state = 11},
  [961] = {.lex_state = 418, .external_lex_state = 11},
  [962] = {.lex_state = 418, .external_lex_state = 11},
  [963] = {.lex_state = 418, .external_lex_state = 11},
  [964] = {.lex_state = 418, .external_lex_state = 11},
  [965] = {.lex_state = 418, .external_lex_state = 11},
  [966] = {.lex_state = 418, .external_lex_state = 11},
  [967] = {.lex_state = 418, .external_lex_state = 11},
  [968] = {.lex_state = 418, .external_lex_state = 11},
  [969] = {.lex_state = 418, .external_lex_state = 11},
  [970] = {.lex_state = 418, .external_lex_state = 11},
  [971] = {.lex_state = 418, .external_lex_state = 11},
  [972] = {.lex_state = 418, .external_lex_state = 11},
  [973] = {.lex_state = 418, .external_lex_state = 11},
  [974] = {.lex_state = 418, .external_lex_state = 11},
  [975] = {.lex_state = 418, .external_lex_state = 11},
  [976] = {.lex_state = 418, .external_lex_state = 11},
  [977] = {.lex_state = 418, .external_lex_state = 11},
  [978] = {.lex_state = 418, .external_lex_state = 11},
  [979] = {.lex_state = 418, .external_lex_state = 11},
  [980] = {.lex_state = 418, .external_lex_state = 11},
  [981] = {.lex_state = 418, .external_lex_state = 11},
  [982] = {.lex_state = 418, .external_lex_state = 11},
  [983] = {.lex_state = 418, .external_lex_state = 11},
  [984] = {.lex_state = 418, .external_lex_state = 11},
  [985] = {.lex_state = 418, .external_lex_state = 11},
  [986] = {.lex_state = 418, .external_lex_state = 11},
  [987] = {.lex_state = 418, .external_lex_state = 11},
  [988] = {.lex_state = 418, .external_lex_state = 11},
  [989] = {.lex_state = 443, .external_lex_state = 16},
  [990] = {.lex_state = 418, .external_lex_state = 11},
  [991] = {.lex_state = 443, .external_lex_state = 20},
  [992] = {.lex_state = 443, .external_lex_state = 16},
  [993] = {.lex_state = 443, .external_lex_state = 20},
  [994] = {.lex_state = 443, .external_lex_state = 2},
  [995] = {.lex_state = 443, .external_lex_state = 2},
  [996] = {.lex_state = 443, .external_lex_state = 2},
  [997] = {.lex_state = 443, .external_lex_state = 2},
  [998] = {.lex_state = 443, .external_lex_state = 21},
  [999] = {.lex_state = 443, .external_lex_state = 21},
  [1000] = {.lex_state = 443, .external_lex_state = 21},
  [1001] = {.lex_state = 443, .external_lex_state = 2},
  [1002] = {.lex_state = 443, .external_lex_state = 2},
  [1003] = {.lex_state = 443, .external_lex_state = 20},
  [1004] = {.lex_state = 443, .external_lex_state = 21},
  [1005] = {.lex_state = 443, .external_lex_state = 21},
  [1006] = {.lex_state = 443, .external_lex_state = 2},
  [1007] = {.lex_state = 443, .external_lex_state = 21},
  [1008] = {.lex_state = 443, .external_lex_state = 21},
  [1009] = {.lex_state = 443, .external_lex_state = 21},
  [1010] = {.lex_state = 443, .external_lex_state = 21},
  [1011] = {.lex_state = 443, .external_lex_state = 20},
  [1012] = {.lex_state = 443, .external_lex_state = 22},
  [1013] = {.lex_state = 443, .external_lex_state = 22},
  [1014] = {.lex_state = 443, .external_lex_state = 23},
  [1015] = {.lex_state = 443, .external_lex_state = 22},
  [1016] = {.lex_state = 443, .external_lex_state = 2},
  [1017] = {.lex_state = 443, .external_lex_state = 22},
  [1018] = {.lex_state = 443, .external_lex_state = 23},
  [1019] = {.lex_state = 443, .external_lex_state = 2},
  [1020] = {.lex_state = 443, .external_lex_state = 22},
  [1021] = {.lex_state = 443, .external_lex_state = 22},
  [1022] = {.lex_state = 443, .external_lex_state = 22},
  [1023] = {.lex_state = 443, .external_lex_state = 2},
  [1024] = {.lex_state = 443, .external_lex_state = 22},
  [1025] = {.lex_state = 443, .external_lex_state = 23},
  [1026] = {.lex_state = 443, .external_lex_state = 23},
  [1027] = {.lex_state = 443, .external_lex_state = 22},
  [1028] = {.lex_state = 443, .external_lex_state = 22},
  [1029] = {.lex_state = 443, .external_lex_state = 22},
  [1030] = {.lex_state = 443, .external_lex_state = 22},
  [1031] = {.lex_state = 443, .external_lex_state = 22},
  [1032] = {.lex_state = 443, .external_lex_state = 22},
  [1033] = {.lex_state = 443, .external_lex_state = 22},
  [1034] = {.lex_state = 443, .external_lex_state = 22},
  [1035] = {.lex_state = 443, .external_lex_state = 2},
  [1036] = {.lex_state = 443, .external_lex_state = 22},
  [1037] = {.lex_state = 443, .external_lex_state = 23},
  [1038] = {.lex_state = 443, .external_lex_state = 22},
  [1039] = {.lex_state = 443, .external_lex_state = 2},
  [1040] = {.lex_state = 443, .external_lex_state = 22},
  [1041] = {.lex_state = 443, .external_lex_state = 22},
  [1042] = {.lex_state = 443, .external_lex_state = 23},
  [1043] = {.lex_state = 443, .external_lex_state = 22},
  [1044] = {.lex_state = 443, .external_lex_state = 22},
  [1045] = {.lex_state = 443, .external_lex_state = 22},
  [1046] = {.lex_state = 443, .external_lex_state = 22},
  [1047] = {.lex_state = 443, .external_lex_state = 22},
  [1048] = {.lex_state = 443, .external_lex_state = 2},
  [1049] = {.lex_state = 443, .external_lex_state = 22},
  [1050] = {.lex_state = 443, .external_lex_state = 22},
  [1051] = {.lex_state = 443, .external_lex_state = 2},
  [1052] = {.lex_state = 443, .external_lex_state = 22},
  [1053] = {.lex_state = 443, .external_lex_state = 2},
  [1054] = {.lex_state = 443, .external_lex_state = 22},
  [1055] = {.lex_state = 443, .external_lex_state = 22},
  [1056] = {.lex_state = 443, .external_lex_state = 23},
  [1057] = {.lex_state = 443, .external_lex_state = 2},
  [1058] = {.lex_state = 443, .external_lex_state = 22},
  [1059] = {.lex_state = 443, .external_lex_state = 22},
  [1060] = {.lex_state = 443, .external_lex_state = 23},
  [1061] = {.lex_state = 443, .external_lex_state = 2},
  [1062] = {.lex_state = 443, .external_lex_state = 23},
  [1063] = {.lex_state = 443, .external_lex_state = 2},
  [1064] = {.lex_state = 443, .external_lex_state = 2},
  [1065] = {.lex_state = 443, .external_lex_state = 2},
  [1066] = {.lex_state = 443, .external_lex_state = 2},
  [1067] = {.lex_state = 443, .external_lex_state = 2},
  [1068] = {.lex_state = 443, .external_lex_state = 2},
  [1069] = {.lex_state = 443, .external_lex_state = 2},
  [1070] = {.lex_state = 443, .external_lex_state = 2},
  [1071] = {.lex_state = 443, .external_lex_state = 2},
  [1072] = {.lex_state = 443, .external_lex_state = 2},
  [1073] = {.lex_state = 443, .external_lex_state = 2},
  [1074] = {.lex_state = 443, .external_lex_state = 2},
  [1075] = {.lex_state = 443, .external_lex_state = 2},
  [1076] = {.lex_state = 443, .external_lex_state = 2},
  [1077] = {.lex_state = 443, .external_lex_state = 2},
  [1078] = {.lex_state = 443, .external_lex_state = 2},
  [1079] = {.lex_state = 443, .external_lex_state = 2},
  [1080] = {.lex_state = 443, .external_lex_state = 2},
  [1081] = {.lex_state = 443, .external_lex_state = 2},
  [1082] = {.lex_state = 443, .external_lex_state = 2},
  [1083] = {.lex_state = 443, .external_lex_state = 2},
  [1084] = {.lex_state = 443, .external_lex_state = 2},
  [1085] = {.lex_state = 443, .external_lex_state = 2},
  [1086] = {.lex_state = 443, .external_lex_state = 2},
  [1087] = {.lex_state = 443, .external_lex_state = 2},
  [1088] = {.lex_state = 443, .external_lex_state = 2},
  [1089] = {.lex_state = 443, .external_lex_state = 2},
  [1090] = {.lex_state = 443, .external_lex_state = 2},
  [1091] = {.lex_state = 443, .external_lex_state = 2},
  [1092] = {.lex_state = 443, .external_lex_state = 2},
  [1093] = {.lex_state = 443, .external_lex_state = 2},
  [1094] = {.lex_state = 443, .external_lex_state = 2},
  [1095] = {.lex_state = 443, .external_lex_state = 2},
  [1096] = {.lex_state = 443, .external_lex_state = 2},
  [1097] = {.lex_state = 443, .external_lex_state = 2},
  [1098] = {.lex_state = 443, .external_lex_state = 2},
  [1099] = {.lex_state = 443, .external_lex_state = 2},
  [1100] = {.lex_state = 443, .external_lex_state = 2},
  [1101] = {.lex_state = 443, .external_lex_state = 2},
  [1102] = {.lex_state = 443, .external_lex_state = 2},
  [1103] = {.lex_state = 443, .external_lex_state = 2},
  [1104] = {.lex_state = 443, .external_lex_state = 2},
  [1105] = {.lex_state = 443, .external_lex_state = 2},
  [1106] = {.lex_state = 443, .external_lex_state = 2},
  [1107] = {.lex_state = 443, .external_lex_state = 2},
  [1108] = {.lex_state = 443, .external_lex_state = 2},
  [1109] = {.lex_state = 452, .external_lex_state = 16},
  [1110] = {.lex_state = 452, .external_lex_state = 16},
  [1111] = {.lex_state = 452, .external_lex_state = 20},
  [1112] = {.lex_state = 452, .external_lex_state = 2},
  [1113] = {.lex_state = 443, .external_lex_state = 24},
  [1114] = {.lex_state = 452, .external_lex_state = 21},
  [1115] = {.lex_state = 452, .external_lex_state = 21},
  [1116] = {.lex_state = 452, .external_lex_state = 2},
  [1117] = {.lex_state = 452, .external_lex_state = 20},
  [1118] = {.lex_state = 452, .external_lex_state = 2},
  [1119] = {.lex_state = 452, .external_lex_state = 2},
  [1120] = {.lex_state = 452, .external_lex_state = 23},
  [1121] = {.lex_state = 452, .external_lex_state = 23},
  [1122] = {.lex_state = 443, .external_lex_state = 2},
  [1123] = {.lex_state = 443, .external_lex_state = 2},
  [1124] = {.lex_state = 443, .external_lex_state = 2},
  [1125] = {.lex_state = 443, .external_lex_state = 2},
  [1126] = {.lex_state = 443, .external_lex_state = 2},
  [1127] = {.lex_state = 443, .external_lex_state = 2},
  [1128] = {.lex_state = 443, .external_lex_state = 2},
  [1129] = {.lex_state = 443, .external_lex_state = 2},
  [1130] = {.lex_state = 443, .external_lex_state = 2},
  [1131] = {.lex_state = 443, .external_lex_state = 2},
  [1132] = {.lex_state = 443, .external_lex_state = 2},
  [1133] = {.lex_state = 443, .external_lex_state = 2},
  [1134] = {.lex_state = 443, .external_lex_state = 2},
  [1135] = {.lex_state = 443, .external_lex_state = 2},
  [1136] = {.lex_state = 443, .external_lex_state = 2},
  [1137] = {.lex_state = 443, .external_lex_state = 2},
  [1138] = {.lex_state = 443, .external_lex_state = 2},
  [1139] = {.lex_state = 443, .external_lex_state = 2},
  [1140] = {.lex_state = 443, .external_lex_state = 2},
  [1141] = {.lex_state = 443, .external_lex_state = 2},
  [1142] = {.lex_state = 443, .external_lex_state = 2},
  [1143] = {.lex_state = 443, .external_lex_state = 2},
  [1144] = {.lex_state = 443, .external_lex_state = 2},
  [1145] = {.lex_state = 443, .external_lex_state = 2},
  [1146] = {.lex_state = 443, .external_lex_state = 2},
  [1147] = {.lex_state = 443, .external_lex_state = 2},
  [1148] = {.lex_state = 443, .external_lex_state = 2},
  [1149] = {.lex_state = 443, .external_lex_state = 2},
  [1150] = {.lex_state = 445, .external_lex_state = 24},
  [1151] = {.lex_state = 452, .external_lex_state = 2},
  [1152] = {.lex_state = 445, .external_lex_state = 22},
  [1153] = {.lex_state = 452, .external_lex_state = 2},
  [1154] = {.lex_state = 452, .external_lex_state = 2},
  [1155] = {.lex_state = 452, .external_lex_state = 2},
  [1156] = {.lex_state = 452, .external_lex_state = 2},
  [1157] = {.lex_state = 445, .external_lex_state = 22},
  [1158] = {.lex_state = 452, .external_lex_state = 2},
  [1159] = {.lex_state = 445, .external_lex_state = 22},
  [1160] = {.lex_state = 443, .external_lex_state = 2},
  [1161] = {.lex_state = 443, .external_lex_state = 2},
  [1162] = {.lex_state = 443, .external_lex_state = 2},
  [1163] = {.lex_state = 443, .external_lex_state = 16},
  [1164] = {.lex_state = 443, .external_lex_state = 2},
  [1165] = {.lex_state = 443, .external_lex_state = 16},
  [1166] = {.lex_state = 443, .external_lex_state = 2},
  [1167] = {.lex_state = 443, .external_lex_state = 16},
  [1168] = {.lex_state = 443, .external_lex_state = 16},
  [1169] = {.lex_state = 452, .external_lex_state = 21},
  [1170] = {.lex_state = 443, .external_lex_state = 16},
  [1171] = {.lex_state = 443, .external_lex_state = 16},
  [1172] = {.lex_state = 452, .external_lex_state = 21},
  [1173] = {.lex_state = 443, .external_lex_state = 16},
  [1174] = {.lex_state = 443, .external_lex_state = 16},
  [1175] = {.lex_state = 443, .external_lex_state = 16},
  [1176] = {.lex_state = 443, .external_lex_state = 16},
  [1177] = {.lex_state = 443, .external_lex_state = 16},
  [1178] = {.lex_state = 443, .external_lex_state = 16},
  [1179] = {.lex_state = 443, .external_lex_state = 16},
  [1180] = {.lex_state = 443, .external_lex_state = 16},
  [1181] = {.lex_state = 443, .external_lex_state = 16},
  [1182] = {.lex_state = 443, .external_lex_state = 16},
  [1183] = {.lex_state = 443, .external_lex_state = 16},
  [1184] = {.lex_state = 443, .external_lex_state = 16},
  [1185] = {.lex_state = 443, .external_lex_state = 16},
  [1186] = {.lex_state = 443, .external_lex_state = 16},
  [1187] = {.lex_state = 443, .external_lex_state = 16},
  [1188] = {.lex_state = 443, .external_lex_state = 16},
  [1189] = {.lex_state = 443, .external_lex_state = 16},
  [1190] = {.lex_state = 443, .external_lex_state = 16},
  [1191] = {.lex_state = 443, .external_lex_state = 16},
  [1192] = {.lex_state = 443, .external_lex_state = 16},
  [1193] = {.lex_state = 443, .external_lex_state = 16},
  [1194] = {.lex_state = 443, .external_lex_state = 16},
  [1195] = {.lex_state = 443, .external_lex_state = 2},
  [1196] = {.lex_state = 443, .external_lex_state = 16},
  [1197] = {.lex_state = 443, .external_lex_state = 16},
  [1198] = {.lex_state = 443, .external_lex_state = 16},
  [1199] = {.lex_state = 443, .external_lex_state = 16},
  [1200] = {.lex_state = 443, .external_lex_state = 16},
  [1201] = {.lex_state = 443, .external_lex_state = 16},
  [1202] = {.lex_state = 443, .external_lex_state = 16},
  [1203] = {.lex_state = 443, .external_lex_state = 16},
  [1204] = {.lex_state = 443, .external_lex_state = 16},
  [1205] = {.lex_state = 443, .external_lex_state = 16},
  [1206] = {.lex_state = 443, .external_lex_state = 16},
  [1207] = {.lex_state = 443, .external_lex_state = 16},
  [1208] = {.lex_state = 443, .external_lex_state = 2},
  [1209] = {.lex_state = 443, .external_lex_state = 16},
  [1210] = {.lex_state = 443, .external_lex_state = 16},
  [1211] = {.lex_state = 443, .external_lex_state = 16},
  [1212] = {.lex_state = 443, .external_lex_state = 2},
  [1213] = {.lex_state = 443, .external_lex_state = 16},
  [1214] = {.lex_state = 443, .external_lex_state = 16},
  [1215] = {.lex_state = 443, .external_lex_state = 16},
  [1216] = {.lex_state = 443, .external_lex_state = 16},
  [1217] = {.lex_state = 443, .external_lex_state = 16},
  [1218] = {.lex_state = 443, .external_lex_state = 16},
  [1219] = {.lex_state = 443, .external_lex_state = 16},
  [1220] = {.lex_state = 443, .external_lex_state = 16},
  [1221] = {.lex_state = 452, .external_lex_state = 25},
  [1222] = {.lex_state = 443, .external_lex_state = 16},
  [1223] = {.lex_state = 443, .external_lex_state = 2},
  [1224] = {.lex_state = 443, .external_lex_state = 2},
  [1225] = {.lex_state = 443, .external_lex_state = 2},
  [1226] = {.lex_state = 443, .external_lex_state = 2},
  [1227] = {.lex_state = 443, .external_lex_state = 2},
  [1228] = {.lex_state = 443, .external_lex_state = 2},
  [1229] = {.lex_state = 443, .external_lex_state = 2},
  [1230] = {.lex_state = 443, .external_lex_state = 2},
  [1231] = {.lex_state = 443, .external_lex_state = 16},
  [1232] = {.lex_state = 443, .external_lex_state = 2},
  [1233] = {.lex_state = 443, .external_lex_state = 2},
  [1234] = {.lex_state = 443, .external_lex_state = 2},
  [1235] = {.lex_state = 443, .external_lex_state = 2},
  [1236] = {.lex_state = 443, .external_lex_state = 2},
  [1237] = {.lex_state = 443, .external_lex_state = 2},
  [1238] = {.lex_state = 443, .external_lex_state = 2},
  [1239] = {.lex_state = 443, .external_lex_state = 2},
  [1240] = {.lex_state = 443, .external_lex_state = 2},
  [1241] = {.lex_state = 443, .external_lex_state = 2},
  [1242] = {.lex_state = 443, .external_lex_state = 2},
  [1243] = {.lex_state = 443, .external_lex_state = 2},
  [1244] = {.lex_state = 443, .external_lex_state = 2},
  [1245] = {.lex_state = 443, .external_lex_state = 2},
  [1246] = {.lex_state = 443, .external_lex_state = 2},
  [1247] = {.lex_state = 443, .external_lex_state = 2},
  [1248] = {.lex_state = 443, .external_lex_state = 2},
  [1249] = {.lex_state = 443, .external_lex_state = 2},
  [1250] = {.lex_state = 443, .external_lex_state = 16},
  [1251] = {.lex_state = 443, .external_lex_state = 2},
  [1252] = {.lex_state = 443, .external_lex_state = 16},
  [1253] = {.lex_state = 443, .external_lex_state = 2},
  [1254] = {.lex_state = 443, .external_lex_state = 2},
  [1255] = {.lex_state = 452, .external_lex_state = 23},
  [1256] = {.lex_state = 443, .external_lex_state = 2},
  [1257] = {.lex_state = 443, .external_lex_state = 2},
  [1258] = {.lex_state = 443, .external_lex_state = 2},
  [1259] = {.lex_state = 443, .external_lex_state = 2},
  [1260] = {.lex_state = 443, .external_lex_state = 2},
  [1261] = {.lex_state = 452, .external_lex_state = 23},
  [1262] = {.lex_state = 443, .external_lex_state = 2},
  [1263] = {.lex_state = 443, .external_lex_state = 2},
  [1264] = {.lex_state = 443, .external_lex_state = 2},
  [1265] = {.lex_state = 443, .external_lex_state = 2},
  [1266] = {.lex_state = 443, .external_lex_state = 2},
  [1267] = {.lex_state = 443, .external_lex_state = 2},
  [1268] = {.lex_state = 443, .external_lex_state = 2},
  [1269] = {.lex_state = 443, .external_lex_state = 2},
  [1270] = {.lex_state = 443, .external_lex_state = 2},
  [1271] = {.lex_state = 443, .external_lex_state = 2},
  [1272] = {.lex_state = 443, .external_lex_state = 2},
  [1273] = {.lex_state = 443, .external_lex_state = 2},
  [1274] = {.lex_state = 443, .external_lex_state = 2},
  [1275] = {.lex_state = 443, .external_lex_state = 2},
  [1276] = {.lex_state = 443, .external_lex_state = 2},
  [1277] = {.lex_state = 443, .external_lex_state = 2},
  [1278] = {.lex_state = 443, .external_lex_state = 2},
  [1279] = {.lex_state = 443, .external_lex_state = 2},
  [1280] = {.lex_state = 443, .external_lex_state = 2},
  [1281] = {.lex_state = 443, .external_lex_state = 2},
  [1282] = {.lex_state = 443, .external_lex_state = 2},
  [1283] = {.lex_state = 443, .external_lex_state = 2},
  [1284] = {.lex_state = 443, .external_lex_state = 2},
  [1285] = {.lex_state = 443, .external_lex_state = 2},
  [1286] = {.lex_state = 443, .external_lex_state = 2},
  [1287] = {.lex_state = 443, .external_lex_state = 20},
  [1288] = {.lex_state = 452, .external_lex_state = 2},
  [1289] = {.lex_state = 452, .external_lex_state = 2},
  [1290] = {.lex_state = 452, .external_lex_state = 2},
  [1291] = {.lex_state = 452, .external_lex_state = 2},
  [1292] = {.lex_state = 452, .external_lex_state = 2},
  [1293] = {.lex_state = 452, .external_lex_state = 2},
  [1294] = {.lex_state = 452, .external_lex_state = 2},
  [1295] = {.lex_state = 452, .external_lex_state = 2},
  [1296] = {.lex_state = 452, .external_lex_state = 2},
  [1297] = {.lex_state = 452, .external_lex_state = 2},
  [1298] = {.lex_state = 452, .external_lex_state = 2},
  [1299] = {.lex_state = 452, .external_lex_state = 2},
  [1300] = {.lex_state = 452, .external_lex_state = 2},
  [1301] = {.lex_state = 443, .external_lex_state = 2},
  [1302] = {.lex_state = 452, .external_lex_state = 2},
  [1303] = {.lex_state = 452, .external_lex_state = 2},
  [1304] = {.lex_state = 452, .external_lex_state = 2},
  [1305] = {.lex_state = 452, .external_lex_state = 2},
  [1306] = {.lex_state = 452, .external_lex_state = 2},
  [1307] = {.lex_state = 452, .external_lex_state = 2},
  [1308] = {.lex_state = 452, .external_lex_state = 2},
  [1309] = {.lex_state = 452, .external_lex_state = 2},
  [1310] = {.lex_state = 452, .external_lex_state = 2},
  [1311] = {.lex_state = 452, .external_lex_state = 2},
  [1312] = {.lex_state = 452, .external_lex_state = 2},
  [1313] = {.lex_state = 452, .external_lex_state = 2},
  [1314] = {.lex_state = 452, .external_lex_state = 2},
  [1315] = {.lex_state = 452, .external_lex_state = 2},
  [1316] = {.lex_state = 443, .external_lex_state = 20},
  [1317] = {.lex_state = 452, .external_lex_state = 2},
  [1318] = {.lex_state = 452, .external_lex_state = 2},
  [1319] = {.lex_state = 452, .external_lex_state = 2},
  [1320] = {.lex_state = 452, .external_lex_state = 2},
  [1321] = {.lex_state = 452, .external_lex_state = 2},
  [1322] = {.lex_state = 452, .external_lex_state = 2},
  [1323] = {.lex_state = 452, .external_lex_state = 2},
  [1324] = {.lex_state = 452, .external_lex_state = 2},
  [1325] = {.lex_state = 452, .external_lex_state = 2},
  [1326] = {.lex_state = 443, .external_lex_state = 20},
  [1327] = {.lex_state = 452, .external_lex_state = 2},
  [1328] = {.lex_state = 452, .external_lex_state = 2},
  [1329] = {.lex_state = 452, .external_lex_state = 2},
  [1330] = {.lex_state = 452, .external_lex_state = 2},
  [1331] = {.lex_state = 452, .external_lex_state = 2},
  [1332] = {.lex_state = 452, .external_lex_state = 2},
  [1333] = {.lex_state = 452, .external_lex_state = 2},
  [1334] = {.lex_state = 452, .external_lex_state = 2},
  [1335] = {.lex_state = 452, .external_lex_state = 2},
  [1336] = {.lex_state = 452, .external_lex_state = 2},
  [1337] = {.lex_state = 452, .external_lex_state = 2},
  [1338] = {.lex_state = 452, .external_lex_state = 2},
  [1339] = {.lex_state = 452, .external_lex_state = 2},
  [1340] = {.lex_state = 452, .external_lex_state = 2},
  [1341] = {.lex_state = 452, .external_lex_state = 2},
  [1342] = {.lex_state = 443, .external_lex_state = 2},
  [1343] = {.lex_state = 443, .external_lex_state = 2},
  [1344] = {.lex_state = 452, .external_lex_state = 2},
  [1345] = {.lex_state = 443, .external_lex_state = 2},
  [1346] = {.lex_state = 452, .external_lex_state = 2},
  [1347] = {.lex_state = 443, .external_lex_state = 2},
  [1348] = {.lex_state = 452, .external_lex_state = 2},
  [1349] = {.lex_state = 443, .external_lex_state = 2},
  [1350] = {.lex_state = 452, .external_lex_state = 2},
  [1351] = {.lex_state = 443, .external_lex_state = 2},
  [1352] = {.lex_state = 443, .external_lex_state = 2},
  [1353] = {.lex_state = 452, .external_lex_state = 2},
  [1354] = {.lex_state = 443, .external_lex_state = 2},
  [1355] = {.lex_state = 443, .external_lex_state = 2},
  [1356] = {.lex_state = 452, .external_lex_state = 2},
  [1357] = {.lex_state = 443, .external_lex_state = 2},
  [1358] = {.lex_state = 452, .external_lex_state = 2},
  [1359] = {.lex_state = 443, .external_lex_state = 2},
  [1360] = {.lex_state = 452, .external_lex_state = 2},
  [1361] = {.lex_state = 443, .external_lex_state = 2},
  [1362] = {.lex_state = 443, .external_lex_state = 2},
  [1363] = {.lex_state = 452, .external_lex_state = 2},
  [1364] = {.lex_state = 443, .external_lex_state = 2},
  [1365] = {.lex_state = 452, .external_lex_state = 2},
  [1366] = {.lex_state = 443, .external_lex_state = 2},
  [1367] = {.lex_state = 443, .external_lex_state = 2},
  [1368] = {.lex_state = 443, .external_lex_state = 2},
  [1369] = {.lex_state = 452, .external_lex_state = 2},
  [1370] = {.lex_state = 443, .external_lex_state = 2},
  [1371] = {.lex_state = 452, .external_lex_state = 2},
  [1372] = {.lex_state = 443, .external_lex_state = 2},
  [1373] = {.lex_state = 443, .external_lex_state = 2},
  [1374] = {.lex_state = 452, .external_lex_state = 2},
  [1375] = {.lex_state = 443, .external_lex_state = 2},
  [1376] = {.lex_state = 443, .external_lex_state = 2},
  [1377] = {.lex_state = 452, .external_lex_state = 2},
  [1378] = {.lex_state = 443, .external_lex_state = 2},
  [1379] = {.lex_state = 452, .external_lex_state = 2},
  [1380] = {.lex_state = 443, .external_lex_state = 2},
  [1381] = {.lex_state = 443, .external_lex_state = 2},
  [1382] = {.lex_state = 452, .external_lex_state = 2},
  [1383] = {.lex_state = 443, .external_lex_state = 2},
  [1384] = {.lex_state = 452, .external_lex_state = 2},
  [1385] = {.lex_state = 443, .external_lex_state = 26},
  [1386] = {.lex_state = 443, .external_lex_state = 26},
  [1387] = {.lex_state = 443, .external_lex_state = 26},
  [1388] = {.lex_state = 443, .external_lex_state = 26},
  [1389] = {.lex_state = 443, .external_lex_state = 26},
  [1390] = {.lex_state = 443, .external_lex_state = 26},
  [1391] = {.lex_state = 443, .external_lex_state = 26},
  [1392] = {.lex_state = 443, .external_lex_state = 26},
  [1393] = {.lex_state = 443, .external_lex_state = 26},
  [1394] = {.lex_state = 443, .external_lex_state = 26},
  [1395] = {.lex_state = 443, .external_lex_state = 26},
  [1396] = {.lex_state = 443, .external_lex_state = 26},
  [1397] = {.lex_state = 443, .external_lex_state = 26},
  [1398] = {.lex_state = 443, .external_lex_state = 26},
  [1399] = {.lex_state = 443, .external_lex_state = 26},
  [1400] = {.lex_state = 443, .external_lex_state = 26},
  [1401] = {.lex_state = 443, .external_lex_state = 26},
  [1402] = {.lex_state = 443, .external_lex_state = 26},
  [1403] = {.lex_state = 443, .external_lex_state = 26},
  [1404] = {.lex_state = 443, .external_lex_state = 26},
  [1405] = {.lex_state = 443, .external_lex_state = 26},
  [1406] = {.lex_state = 443, .external_lex_state = 26},
  [1407] = {.lex_state = 443, .external_lex_state = 26},
  [1408] = {.lex_state = 443, .external_lex_state = 26},
  [1409] = {.lex_state = 443, .external_lex_state = 26},
  [1410] = {.lex_state = 443, .external_lex_state = 26},
  [1411] = {.lex_state = 443, .external_lex_state = 26},
  [1412] = {.lex_state = 443, .external_lex_state = 26},
  [1413] = {.lex_state = 443, .external_lex_state = 2},
  [1414] = {.lex_state = 443, .external_lex_state = 26},
  [1415] = {.lex_state = 443, .external_lex_state = 22},
  [1416] = {.lex_state = 443, .external_lex_state = 26},
  [1417] = {.lex_state = 443, .external_lex_state = 26},
  [1418] = {.lex_state = 443, .external_lex_state = 26},
  [1419] = {.lex_state = 443, .external_lex_state = 26},
  [1420] = {.lex_state = 443, .external_lex_state = 26},
  [1421] = {.lex_state = 443, .external_lex_state = 2},
  [1422] = {.lex_state = 443, .external_lex_state = 2},
  [1423] = {.lex_state = 443, .external_lex_state = 27},
  [1424] = {.lex_state = 443, .external_lex_state = 28},
  [1425] = {.lex_state = 443, .external_lex_state = 28},
  [1426] = {.lex_state = 443, .external_lex_state = 2},
  [1427] = {.lex_state = 443, .external_lex_state = 27},
  [1428] = {.lex_state = 443, .external_lex_state = 28},
  [1429] = {.lex_state = 443, .external_lex_state = 28},
  [1430] = {.lex_state = 443, .external_lex_state = 2},
  [1431] = {.lex_state = 443, .external_lex_state = 27},
  [1432] = {.lex_state = 443, .external_lex_state = 28},
  [1433] = {.lex_state = 443, .external_lex_state = 28},
  [1434] = {.lex_state = 443, .external_lex_state = 2},
  [1435] = {.lex_state = 443, .external_lex_state = 27},
  [1436] = {.lex_state = 443, .external_lex_state = 28},
  [1437] = {.lex_state = 443, .external_lex_state = 28},
  [1438] = {.lex_state = 443, .external_lex_state = 2},
  [1439] = {.lex_state = 443, .external_lex_state = 27},
  [1440] = {.lex_state = 443, .external_lex_state = 28},
  [1441] = {.lex_state = 443, .external_lex_state = 28},
  [1442] = {.lex_state = 443, .external_lex_state = 2},
  [1443] = {.lex_state = 443, .external_lex_state = 27},
  [1444] = {.lex_state = 443, .external_lex_state = 22},
  [1445] = {.lex_state = 443, .external_lex_state = 28},
  [1446] = {.lex_state = 443, .external_lex_state = 2},
  [1447] = {.lex_state = 443, .external_lex_state = 27},
  [1448] = {.lex_state = 443, .external_lex_state = 28},
  [1449] = {.lex_state = 443, .external_lex_state = 28},
  [1450] = {.lex_state = 443, .external_lex_state = 28},
  [1451] = {.lex_state = 443, .external_lex_state = 2},
  [1452] = {.lex_state = 443, .external_lex_state = 27},
  [1453] = {.lex_state = 443, .external_lex_state = 28},
  [1454] = {.lex_state = 443, .external_lex_state = 28},
  [1455] = {.lex_state = 443, .external_lex_state = 2},
  [1456] = {.lex_state = 443, .external_lex_state = 27},
  [1457] = {.lex_state = 443, .external_lex_state = 28},
  [1458] = {.lex_state = 443, .external_lex_state = 28},
  [1459] = {.lex_state = 443, .external_lex_state = 2},
  [1460] = {.lex_state = 443, .external_lex_state = 27},
  [1461] = {.lex_state = 443, .external_lex_state = 28},
  [1462] = {.lex_state = 443, .external_lex_state = 2},
  [1463] = {.lex_state = 443, .external_lex_state = 2},
  [1464] = {.lex_state = 443, .external_lex_state = 27},
  [1465] = {.lex_state = 443, .external_lex_state = 28},
  [1466] = {.lex_state = 443, .external_lex_state = 28},
  [1467] = {.lex_state = 443, .external_lex_state = 2},
  [1468] = {.lex_state = 443, .external_lex_state = 27},
  [1469] = {.lex_state = 443, .external_lex_state = 28},
  [1470] = {.lex_state = 443, .external_lex_state = 28},
  [1471] = {.lex_state = 443, .external_lex_state = 2},
  [1472] = {.lex_state = 443, .external_lex_state = 27},
  [1473] = {.lex_state = 443, .external_lex_state = 28},
  [1474] = {.lex_state = 443, .external_lex_state = 22},
  [1475] = {.lex_state = 443, .external_lex_state = 2},
  [1476] = {.lex_state = 443, .external_lex_state = 27},
  [1477] = {.lex_state = 443, .external_lex_state = 28},
  [1478] = {.lex_state = 443, .external_lex_state = 28},
  [1479] = {.lex_state = 443, .external_lex_state = 2},
  [1480] = {.lex_state = 443, .external_lex_state = 27},
  [1481] = {.lex_state = 443, .external_lex_state = 28},
  [1482] = {.lex_state = 443, .external_lex_state = 28},
  [1483] = {.lex_state = 443, .external_lex_state = 2},
  [1484] = {.lex_state = 443, .external_lex_state = 27},
  [1485] = {.lex_state = 443, .external_lex_state = 28},
  [1486] = {.lex_state = 443, .external_lex_state = 22},
  [1487] = {.lex_state = 443, .external_lex_state = 2},
  [1488] = {.lex_state = 443, .external_lex_state = 27},
  [1489] = {.lex_state = 443, .external_lex_state = 2},
  [1490] = {.lex_state = 443, .external_lex_state = 28},
  [1491] = {.lex_state = 443, .external_lex_state = 2},
  [1492] = {.lex_state = 443, .external_lex_state = 27},
  [1493] = {.lex_state = 443, .external_lex_state = 28},
  [1494] = {.lex_state = 443, .external_lex_state = 2},
  [1495] = {.lex_state = 443, .external_lex_state = 2},
  [1496] = {.lex_state = 443, .external_lex_state = 27},
  [1497] = {.lex_state = 443, .external_lex_state = 27},
  [1498] = {.lex_state = 443, .external_lex_state = 2},
  [1499] = {.lex_state = 443, .external_lex_state = 2},
  [1500] = {.lex_state = 443, .external_lex_state = 27},
  [1501] = {.lex_state = 443, .external_lex_state = 27},
  [1502] = {.lex_state = 443, .external_lex_state = 22},
  [1503] = {.lex_state = 443, .external_lex_state = 2},
  [1504] = {.lex_state = 443, .external_lex_state = 27},
  [1505] = {.lex_state = 443, .external_lex_state = 2},
  [1506] = {.lex_state = 443, .external_lex_state = 27},
  [1507] = {.lex_state = 443, .external_lex_state = 2},
  [1508] = {.lex_state = 443, .external_lex_state = 27},
  [1509] = {.lex_state = 443, .external_lex_state = 2},
  [1510] = {.lex_state = 443, .external_lex_state = 28},
  [1511] = {.lex_state = 443, .external_lex_state = 2},
  [1512] = {.lex_state = 443, .external_lex_state = 27},
  [1513] = {.lex_state = 443, .external_lex_state = 28},
  [1514] = {.lex_state = 443, .external_lex_state = 2},
  [1515] = {.lex_state = 443, .external_lex_state = 2},
  [1516] = {.lex_state = 443, .external_lex_state = 27},
  [1517] = {.lex_state = 443, .external_lex_state = 2},
  [1518] = {.lex_state = 443, .external_lex_state = 2},
  [1519] = {.lex_state = 443, .external_lex_state = 2},
  [1520] = {.lex_state = 443, .external_lex_state = 2},
  [1521] = {.lex_state = 443, .external_lex_state = 28},
  [1522] = {.lex_state = 443, .external_lex_state = 2},
  [1523] = {.lex_state = 443, .external_lex_state = 2},
  [1524] = {.lex_state = 443, .external_lex_state = 2},
  [1525] = {.lex_state = 443, .external_lex_state = 2},
  [1526] = {.lex_state = 443, .external_lex_state = 2},
  [1527] = {.lex_state = 443, .external_lex_state = 2},
  [1528] = {.lex_state = 443, .external_lex_state = 2},
  [1529] = {.lex_state = 443, .external_lex_state = 2},
  [1530] = {.lex_state = 443, .external_lex_state = 2},
  [1531] = {.lex_state = 443, .external_lex_state = 2},
  [1532] = {.lex_state = 443, .external_lex_state = 2},
  [1533] = {.lex_state = 443, .external_lex_state = 2},
  [1534] = {.lex_state = 443, .external_lex_state = 2},
  [1535] = {.lex_state = 443, .external_lex_state = 2},
  [1536] = {.lex_state = 443, .external_lex_state = 2},
  [1537] = {.lex_state = 443, .external_lex_state = 2},
  [1538] = {.lex_state = 443, .external_lex_state = 2},
  [1539] = {.lex_state = 443, .external_lex_state = 2},
  [1540] = {.lex_state = 443, .external_lex_state = 2},
  [1541] = {.lex_state = 443, .external_lex_state = 2},
  [1542] = {.lex_state = 443, .external_lex_state = 2},
  [1543] = {.lex_state = 443, .external_lex_state = 2},
  [1544] = {.lex_state = 443, .external_lex_state = 2},
  [1545] = {.lex_state = 443, .external_lex_state = 2},
  [1546] = {.lex_state = 443, .external_lex_state = 2},
  [1547] = {.lex_state = 443, .external_lex_state = 2},
  [1548] = {.lex_state = 443, .external_lex_state = 2},
  [1549] = {.lex_state = 443, .external_lex_state = 2},
  [1550] = {.lex_state = 443, .external_lex_state = 2},
  [1551] = {.lex_state = 443, .external_lex_state = 2},
  [1552] = {.lex_state = 443, .external_lex_state = 2},
  [1553] = {.lex_state = 443, .external_lex_state = 2},
  [1554] = {.lex_state = 443, .external_lex_state = 2},
  [1555] = {.lex_state = 443, .external_lex_state = 2},
  [1556] = {.lex_state = 443, .external_lex_state = 2},
  [1557] = {.lex_state = 443, .external_lex_state = 2},
  [1558] = {.lex_state = 443, .external_lex_state = 2},
  [1559] = {.lex_state = 443, .external_lex_state = 2},
  [1560] = {.lex_state = 443, .external_lex_state = 2},
  [1561] = {.lex_state = 443, .external_lex_state = 2},
  [1562] = {.lex_state = 443, .external_lex_state = 2},
  [1563] = {.lex_state = 443, .external_lex_state = 2},
  [1564] = {.lex_state = 443, .external_lex_state = 2},
  [1565] = {.lex_state = 443, .external_lex_state = 2},
  [1566] = {.lex_state = 443, .external_lex_state = 2},
  [1567] = {.lex_state = 443, .external_lex_state = 2},
  [1568] = {.lex_state = 443, .external_lex_state = 2},
  [1569] = {.lex_state = 443, .external_lex_state = 2},
  [1570] = {.lex_state = 443, .external_lex_state = 2},
  [1571] = {.lex_state = 443, .external_lex_state = 2},
  [1572] = {.lex_state = 443, .external_lex_state = 2},
  [1573] = {.lex_state = 443, .external_lex_state = 2},
  [1574] = {.lex_state = 443, .external_lex_state = 2},
  [1575] = {.lex_state = 443, .external_lex_state = 2},
  [1576] = {.lex_state = 443, .external_lex_state = 2},
  [1577] = {.lex_state = 443, .external_lex_state = 2},
  [1578] = {.lex_state = 443, .external_lex_state = 2},
  [1579] = {.lex_state = 443, .external_lex_state = 2},
  [1580] = {.lex_state = 443, .external_lex_state = 2},
  [1581] = {.lex_state = 443, .external_lex_state = 2},
  [1582] = {.lex_state = 443, .external_lex_state = 2},
  [1583] = {.lex_state = 443, .external_lex_state = 2},
  [1584] = {.lex_state = 443, .external_lex_state = 2},
  [1585] = {.lex_state = 443, .external_lex_state = 2},
  [1586] = {.lex_state = 443, .external_lex_state = 2},
  [1587] = {.lex_state = 443, .external_lex_state = 2},
  [1588] = {.lex_state = 443, .external_lex_state = 2},
  [1589] = {.lex_state = 443, .external_lex_state = 2},
  [1590] = {.lex_state = 443, .external_lex_state = 2},
  [1591] = {.lex_state = 443, .external_lex_state = 2},
  [1592] = {.lex_state = 443, .external_lex_state = 2},
  [1593] = {.lex_state = 443, .external_lex_state = 2},
  [1594] = {.lex_state = 443, .external_lex_state = 2},
  [1595] = {.lex_state = 443, .external_lex_state = 2},
  [1596] = {.lex_state = 443, .external_lex_state = 2},
  [1597] = {.lex_state = 443, .external_lex_state = 2},
  [1598] = {.lex_state = 443, .external_lex_state = 2},
  [1599] = {.lex_state = 443, .external_lex_state = 2},
  [1600] = {.lex_state = 443, .external_lex_state = 2},
  [1601] = {.lex_state = 443, .external_lex_state = 2},
  [1602] = {.lex_state = 443, .external_lex_state = 2},
  [1603] = {.lex_state = 443, .external_lex_state = 2},
  [1604] = {.lex_state = 443, .external_lex_state = 2},
  [1605] = {.lex_state = 443, .external_lex_state = 2},
  [1606] = {.lex_state = 443, .external_lex_state = 2},
  [1607] = {.lex_state = 443, .external_lex_state = 2},
  [1608] = {.lex_state = 443, .external_lex_state = 2},
  [1609] = {.lex_state = 443, .external_lex_state = 2},
  [1610] = {.lex_state = 443, .external_lex_state = 2},
  [1611] = {.lex_state = 443, .external_lex_state = 2},
  [1612] = {.lex_state = 443, .external_lex_state = 2},
  [1613] = {.lex_state = 443, .external_lex_state = 2},
  [1614] = {.lex_state = 443, .external_lex_state = 2},
  [1615] = {.lex_state = 443, .external_lex_state = 2},
  [1616] = {.lex_state = 443, .external_lex_state = 2},
  [1617] = {.lex_state = 443, .external_lex_state = 2},
  [1618] = {.lex_state = 443, .external_lex_state = 2},
  [1619] = {.lex_state = 443, .external_lex_state = 2},
  [1620] = {.lex_state = 443, .external_lex_state = 2},
  [1621] = {.lex_state = 443, .external_lex_state = 2},
  [1622] = {.lex_state = 443, .external_lex_state = 2},
  [1623] = {.lex_state = 443, .external_lex_state = 2},
  [1624] = {.lex_state = 443, .external_lex_state = 2},
  [1625] = {.lex_state = 443, .external_lex_state = 2},
  [1626] = {.lex_state = 443, .external_lex_state = 2},
  [1627] = {.lex_state = 443, .external_lex_state = 2},
  [1628] = {.lex_state = 443, .external_lex_state = 2},
  [1629] = {.lex_state = 443, .external_lex_state = 2},
  [1630] = {.lex_state = 443, .external_lex_state = 2},
  [1631] = {.lex_state = 443, .external_lex_state = 2},
  [1632] = {.lex_state = 443, .external_lex_state = 2},
  [1633] = {.lex_state = 443, .external_lex_state = 2},
  [1634] = {.lex_state = 443, .external_lex_state = 2},
  [1635] = {.lex_state = 443, .external_lex_state = 2},
  [1636] = {.lex_state = 443, .external_lex_state = 2},
  [1637] = {.lex_state = 443, .external_lex_state = 2},
  [1638] = {.lex_state = 443, .external_lex_state = 2},
  [1639] = {.lex_state = 443, .external_lex_state = 2},
  [1640] = {.lex_state = 443, .external_lex_state = 2},
  [1641] = {.lex_state = 443, .external_lex_state = 2},
  [1642] = {.lex_state = 443, .external_lex_state = 2},
  [1643] = {.lex_state = 443, .external_lex_state = 2},
  [1644] = {.lex_state = 443, .external_lex_state = 2},
  [1645] = {.lex_state = 443, .external_lex_state = 2},
  [1646] = {.lex_state = 443, .external_lex_state = 2},
  [1647] = {.lex_state = 443, .external_lex_state = 2},
  [1648] = {.lex_state = 443, .external_lex_state = 2},
  [1649] = {.lex_state = 443, .external_lex_state = 2},
  [1650] = {.lex_state = 443, .external_lex_state = 2},
  [1651] = {.lex_state = 443, .external_lex_state = 2},
  [1652] = {.lex_state = 443, .external_lex_state = 2},
  [1653] = {.lex_state = 443, .external_lex_state = 2},
  [1654] = {.lex_state = 443, .external_lex_state = 2},
  [1655] = {.lex_state = 443, .external_lex_state = 2},
  [1656] = {.lex_state = 443, .external_lex_state = 2},
  [1657] = {.lex_state = 443, .external_lex_state = 2},
  [1658] = {.lex_state = 443, .external_lex_state = 2},
  [1659] = {.lex_state = 443, .external_lex_state = 2},
  [1660] = {.lex_state = 443, .external_lex_state = 2},
  [1661] = {.lex_state = 443, .external_lex_state = 2},
  [1662] = {.lex_state = 443, .external_lex_state = 2},
  [1663] = {.lex_state = 443, .external_lex_state = 2},
  [1664] = {.lex_state = 443, .external_lex_state = 2},
  [1665] = {.lex_state = 443, .external_lex_state = 2},
  [1666] = {.lex_state = 443, .external_lex_state = 2},
  [1667] = {.lex_state = 443, .external_lex_state = 2},
  [1668] = {.lex_state = 443, .external_lex_state = 2},
  [1669] = {.lex_state = 443, .external_lex_state = 2},
  [1670] = {.lex_state = 443, .external_lex_state = 2},
  [1671] = {.lex_state = 443, .external_lex_state = 2},
  [1672] = {.lex_state = 443, .external_lex_state = 2},
  [1673] = {.lex_state = 443, .external_lex_state = 2},
  [1674] = {.lex_state = 443, .external_lex_state = 2},
  [1675] = {.lex_state = 443, .external_lex_state = 2},
  [1676] = {.lex_state = 443, .external_lex_state = 2},
  [1677] = {.lex_state = 443, .external_lex_state = 2},
  [1678] = {.lex_state = 443, .external_lex_state = 2},
  [1679] = {.lex_state = 443, .external_lex_state = 2},
  [1680] = {.lex_state = 443, .external_lex_state = 2},
  [1681] = {.lex_state = 443, .external_lex_state = 2},
  [1682] = {.lex_state = 443, .external_lex_state = 2},
  [1683] = {.lex_state = 443, .external_lex_state = 2},
  [1684] = {.lex_state = 443, .external_lex_state = 2},
  [1685] = {.lex_state = 443, .external_lex_state = 2},
  [1686] = {.lex_state = 443, .external_lex_state = 2},
  [1687] = {.lex_state = 443, .external_lex_state = 2},
  [1688] = {.lex_state = 443, .external_lex_state = 2},
  [1689] = {.lex_state = 443, .external_lex_state = 2},
  [1690] = {.lex_state = 443, .external_lex_state = 2},
  [1691] = {.lex_state = 443, .external_lex_state = 2},
  [1692] = {.lex_state = 443, .external_lex_state = 2},
  [1693] = {.lex_state = 443, .external_lex_state = 2},
  [1694] = {.lex_state = 443, .external_lex_state = 2},
  [1695] = {.lex_state = 443, .external_lex_state = 2},
  [1696] = {.lex_state = 443, .external_lex_state = 2},
  [1697] = {.lex_state = 443, .external_lex_state = 2},
  [1698] = {.lex_state = 443, .external_lex_state = 2},
  [1699] = {.lex_state = 443, .external_lex_state = 2},
  [1700] = {.lex_state = 443, .external_lex_state = 2},
  [1701] = {.lex_state = 443, .external_lex_state = 2},
  [1702] = {.lex_state = 443, .external_lex_state = 2},
  [1703] = {.lex_state = 443, .external_lex_state = 2},
  [1704] = {.lex_state = 443, .external_lex_state = 2},
  [1705] = {.lex_state = 443, .external_lex_state = 2},
  [1706] = {.lex_state = 443, .external_lex_state = 2},
  [1707] = {.lex_state = 443, .external_lex_state = 2},
  [1708] = {.lex_state = 443, .external_lex_state = 2},
  [1709] = {.lex_state = 443, .external_lex_state = 2},
  [1710] = {.lex_state = 443, .external_lex_state = 2},
  [1711] = {.lex_state = 443, .external_lex_state = 2},
  [1712] = {.lex_state = 443, .external_lex_state = 2},
  [1713] = {.lex_state = 443, .external_lex_state = 2},
  [1714] = {.lex_state = 443, .external_lex_state = 2},
  [1715] = {.lex_state = 443, .external_lex_state = 2},
  [1716] = {.lex_state = 443, .external_lex_state = 2},
  [1717] = {.lex_state = 443, .external_lex_state = 2},
  [1718] = {.lex_state = 443, .external_lex_state = 2},
  [1719] = {.lex_state = 443, .external_lex_state = 2},
  [1720] = {.lex_state = 443, .external_lex_state = 2},
  [1721] = {.lex_state = 443, .external_lex_state = 2},
  [1722] = {.lex_state = 443, .external_lex_state = 2},
  [1723] = {.lex_state = 443, .external_lex_state = 2},
  [1724] = {.lex_state = 443, .external_lex_state = 2},
  [1725] = {.lex_state = 443, .external_lex_state = 2},
  [1726] = {.lex_state = 443, .external_lex_state = 2},
  [1727] = {.lex_state = 443, .external_lex_state = 2},
  [1728] = {.lex_state = 443, .external_lex_state = 2},
  [1729] = {.lex_state = 443, .external_lex_state = 2},
  [1730] = {.lex_state = 443, .external_lex_state = 2},
  [1731] = {.lex_state = 443, .external_lex_state = 2},
  [1732] = {.lex_state = 443, .external_lex_state = 2},
  [1733] = {.lex_state = 443, .external_lex_state = 2},
  [1734] = {.lex_state = 443, .external_lex_state = 2},
  [1735] = {.lex_state = 443, .external_lex_state = 2},
  [1736] = {.lex_state = 443, .external_lex_state = 2},
  [1737] = {.lex_state = 443, .external_lex_state = 2},
  [1738] = {.lex_state = 443, .external_lex_state = 2},
  [1739] = {.lex_state = 443, .external_lex_state = 2},
  [1740] = {.lex_state = 443, .external_lex_state = 2},
  [1741] = {.lex_state = 443, .external_lex_state = 2},
  [1742] = {.lex_state = 443, .external_lex_state = 2},
  [1743] = {.lex_state = 443, .external_lex_state = 2},
  [1744] = {.lex_state = 443, .external_lex_state = 2},
  [1745] = {.lex_state = 443, .external_lex_state = 2},
  [1746] = {.lex_state = 443, .external_lex_state = 2},
  [1747] = {.lex_state = 443, .external_lex_state = 2},
  [1748] = {.lex_state = 443, .external_lex_state = 2},
  [1749] = {.lex_state = 443, .external_lex_state = 2},
  [1750] = {.lex_state = 443, .external_lex_state = 2},
  [1751] = {.lex_state = 443, .external_lex_state = 2},
  [1752] = {.lex_state = 443, .external_lex_state = 2},
  [1753] = {.lex_state = 443, .external_lex_state = 2},
  [1754] = {.lex_state = 443, .external_lex_state = 2},
  [1755] = {.lex_state = 443, .external_lex_state = 2},
  [1756] = {.lex_state = 443, .external_lex_state = 2},
  [1757] = {.lex_state = 443, .external_lex_state = 2},
  [1758] = {.lex_state = 443, .external_lex_state = 2},
  [1759] = {.lex_state = 443, .external_lex_state = 2},
  [1760] = {.lex_state = 443, .external_lex_state = 2},
  [1761] = {.lex_state = 443, .external_lex_state = 2},
  [1762] = {.lex_state = 443, .external_lex_state = 2},
  [1763] = {.lex_state = 443, .external_lex_state = 2},
  [1764] = {.lex_state = 443, .external_lex_state = 2},
  [1765] = {.lex_state = 443, .external_lex_state = 2},
  [1766] = {.lex_state = 443, .external_lex_state = 2},
  [1767] = {.lex_state = 443, .external_lex_state = 2},
  [1768] = {.lex_state = 443, .external_lex_state = 2},
  [1769] = {.lex_state = 443, .external_lex_state = 2},
  [1770] = {.lex_state = 443, .external_lex_state = 2},
  [1771] = {.lex_state = 443, .external_lex_state = 2},
  [1772] = {.lex_state = 443, .external_lex_state = 2},
  [1773] = {.lex_state = 443, .external_lex_state = 2},
  [1774] = {.lex_state = 443, .external_lex_state = 2},
  [1775] = {.lex_state = 443, .external_lex_state = 2},
  [1776] = {.lex_state = 443, .external_lex_state = 2},
  [1777] = {.lex_state = 443, .external_lex_state = 2},
  [1778] = {.lex_state = 443, .external_lex_state = 2},
  [1779] = {.lex_state = 443, .external_lex_state = 2},
  [1780] = {.lex_state = 443, .external_lex_state = 2},
  [1781] = {.lex_state = 443, .external_lex_state = 2},
  [1782] = {.lex_state = 443, .external_lex_state = 2},
  [1783] = {.lex_state = 443, .external_lex_state = 2},
  [1784] = {.lex_state = 443, .external_lex_state = 2},
  [1785] = {.lex_state = 443, .external_lex_state = 2},
  [1786] = {.lex_state = 443, .external_lex_state = 2},
  [1787] = {.lex_state = 443, .external_lex_state = 2},
  [1788] = {.lex_state = 443, .external_lex_state = 2},
  [1789] = {.lex_state = 443, .external_lex_state = 2},
  [1790] = {.lex_state = 443, .external_lex_state = 2},
  [1791] = {.lex_state = 443, .external_lex_state = 2},
  [1792] = {.lex_state = 443, .external_lex_state = 2},
  [1793] = {.lex_state = 443, .external_lex_state = 2},
  [1794] = {.lex_state = 443, .external_lex_state = 2},
  [1795] = {.lex_state = 443, .external_lex_state = 2},
  [1796] = {.lex_state = 443, .external_lex_state = 2},
  [1797] = {.lex_state = 443, .external_lex_state = 2},
  [1798] = {.lex_state = 443, .external_lex_state = 2},
  [1799] = {.lex_state = 443, .external_lex_state = 2},
  [1800] = {.lex_state = 443, .external_lex_state = 2},
  [1801] = {.lex_state = 443, .external_lex_state = 2},
  [1802] = {.lex_state = 443, .external_lex_state = 2},
  [1803] = {.lex_state = 443, .external_lex_state = 2},
  [1804] = {.lex_state = 443, .external_lex_state = 2},
  [1805] = {.lex_state = 443, .external_lex_state = 2},
  [1806] = {.lex_state = 443, .external_lex_state = 2},
  [1807] = {.lex_state = 443, .external_lex_state = 2},
  [1808] = {.lex_state = 443, .external_lex_state = 2},
  [1809] = {.lex_state = 443, .external_lex_state = 2},
  [1810] = {.lex_state = 443, .external_lex_state = 2},
  [1811] = {.lex_state = 443, .external_lex_state = 2},
  [1812] = {.lex_state = 443, .external_lex_state = 2},
  [1813] = {.lex_state = 443, .external_lex_state = 2},
  [1814] = {.lex_state = 443, .external_lex_state = 2},
  [1815] = {.lex_state = 443, .external_lex_state = 2},
  [1816] = {.lex_state = 443, .external_lex_state = 2},
  [1817] = {.lex_state = 443, .external_lex_state = 2},
  [1818] = {.lex_state = 443, .external_lex_state = 2},
  [1819] = {.lex_state = 443, .external_lex_state = 2},
  [1820] = {.lex_state = 443, .external_lex_state = 2},
  [1821] = {.lex_state = 443, .external_lex_state = 2},
  [1822] = {.lex_state = 443, .external_lex_state = 2},
  [1823] = {.lex_state = 443, .external_lex_state = 2},
  [1824] = {.lex_state = 443, .external_lex_state = 2},
  [1825] = {.lex_state = 443, .external_lex_state = 2},
  [1826] = {.lex_state = 443, .external_lex_state = 2},
  [1827] = {.lex_state = 443, .external_lex_state = 2},
  [1828] = {.lex_state = 443, .external_lex_state = 2},
  [1829] = {.lex_state = 443, .external_lex_state = 2},
  [1830] = {.lex_state = 443, .external_lex_state = 2},
  [1831] = {.lex_state = 443, .external_lex_state = 2},
  [1832] = {.lex_state = 443, .external_lex_state = 2},
  [1833] = {.lex_state = 443, .external_lex_state = 2},
  [1834] = {.lex_state = 443, .external_lex_state = 2},
  [1835] = {.lex_state = 443, .external_lex_state = 2},
  [1836] = {.lex_state = 443, .external_lex_state = 2},
  [1837] = {.lex_state = 443, .external_lex_state = 2},
  [1838] = {.lex_state = 443, .external_lex_state = 2},
  [1839] = {.lex_state = 443, .external_lex_state = 2},
  [1840] = {.lex_state = 443, .external_lex_state = 2},
  [1841] = {.lex_state = 443, .external_lex_state = 2},
  [1842] = {.lex_state = 443, .external_lex_state = 2},
  [1843] = {.lex_state = 443, .external_lex_state = 2},
  [1844] = {.lex_state = 443, .external_lex_state = 2},
  [1845] = {.lex_state = 443, .external_lex_state = 2},
  [1846] = {.lex_state = 443, .external_lex_state = 2},
  [1847] = {.lex_state = 443, .external_lex_state = 2},
  [1848] = {.lex_state = 443, .external_lex_state = 2},
  [1849] = {.lex_state = 443, .external_lex_state = 2},
  [1850] = {.lex_state = 443, .external_lex_state = 2},
  [1851] = {.lex_state = 443, .external_lex_state = 2},
  [1852] = {.lex_state = 443, .external_lex_state = 2},
  [1853] = {.lex_state = 443, .external_lex_state = 2},
  [1854] = {.lex_state = 443, .external_lex_state = 2},
  [1855] = {.lex_state = 443, .external_lex_state = 2},
  [1856] = {.lex_state = 443, .external_lex_state = 2},
  [1857] = {.lex_state = 443, .external_lex_state = 2},
  [1858] = {.lex_state = 443, .external_lex_state = 2},
  [1859] = {.lex_state = 443, .external_lex_state = 2},
  [1860] = {.lex_state = 443, .external_lex_state = 2},
  [1861] = {.lex_state = 443, .external_lex_state = 2},
  [1862] = {.lex_state = 443, .external_lex_state = 2},
  [1863] = {.lex_state = 443, .external_lex_state = 2},
  [1864] = {.lex_state = 443, .external_lex_state = 2},
  [1865] = {.lex_state = 443, .external_lex_state = 2},
  [1866] = {.lex_state = 443, .external_lex_state = 2},
  [1867] = {.lex_state = 443, .external_lex_state = 2},
  [1868] = {.lex_state = 443, .external_lex_state = 2},
  [1869] = {.lex_state = 443, .external_lex_state = 2},
  [1870] = {.lex_state = 443, .external_lex_state = 2},
  [1871] = {.lex_state = 443, .external_lex_state = 2},
  [1872] = {.lex_state = 443, .external_lex_state = 2},
  [1873] = {.lex_state = 443, .external_lex_state = 2},
  [1874] = {.lex_state = 443, .external_lex_state = 2},
  [1875] = {.lex_state = 443, .external_lex_state = 2},
  [1876] = {.lex_state = 443, .external_lex_state = 2},
  [1877] = {.lex_state = 443, .external_lex_state = 2},
  [1878] = {.lex_state = 443, .external_lex_state = 2},
  [1879] = {.lex_state = 443, .external_lex_state = 2},
  [1880] = {.lex_state = 443, .external_lex_state = 2},
  [1881] = {.lex_state = 443, .external_lex_state = 2},
  [1882] = {.lex_state = 443, .external_lex_state = 2},
  [1883] = {.lex_state = 443, .external_lex_state = 2},
  [1884] = {.lex_state = 443, .external_lex_state = 2},
  [1885] = {.lex_state = 443, .external_lex_state = 2},
  [1886] = {.lex_state = 443, .external_lex_state = 2},
  [1887] = {.lex_state = 443, .external_lex_state = 2},
  [1888] = {.lex_state = 443, .external_lex_state = 2},
  [1889] = {.lex_state = 443, .external_lex_state = 2},
  [1890] = {.lex_state = 443, .external_lex_state = 2},
  [1891] = {.lex_state = 443, .external_lex_state = 2},
  [1892] = {.lex_state = 443, .external_lex_state = 2},
  [1893] = {.lex_state = 443, .external_lex_state = 2},
  [1894] = {.lex_state = 443, .external_lex_state = 2},
  [1895] = {.lex_state = 443, .external_lex_state = 2},
  [1896] = {.lex_state = 443, .external_lex_state = 2},
  [1897] = {.lex_state = 443, .external_lex_state = 2},
  [1898] = {.lex_state = 443, .external_lex_state = 2},
  [1899] = {.lex_state = 443, .external_lex_state = 2},
  [1900] = {.lex_state = 443, .external_lex_state = 2},
  [1901] = {.lex_state = 443, .external_lex_state = 2},
  [1902] = {.lex_state = 443, .external_lex_state = 2},
  [1903] = {.lex_state = 443, .external_lex_state = 2},
  [1904] = {.lex_state = 443, .external_lex_state = 2},
  [1905] = {.lex_state = 443, .external_lex_state = 2},
  [1906] = {.lex_state = 443, .external_lex_state = 2},
  [1907] = {.lex_state = 443, .external_lex_state = 2},
  [1908] = {.lex_state = 443, .external_lex_state = 2},
  [1909] = {.lex_state = 443, .external_lex_state = 2},
  [1910] = {.lex_state = 443, .external_lex_state = 2},
  [1911] = {.lex_state = 443, .external_lex_state = 2},
  [1912] = {.lex_state = 443, .external_lex_state = 2},
  [1913] = {.lex_state = 443, .external_lex_state = 2},
  [1914] = {.lex_state = 443, .external_lex_state = 2},
  [1915] = {.lex_state = 443, .external_lex_state = 2},
  [1916] = {.lex_state = 443, .external_lex_state = 2},
  [1917] = {.lex_state = 443, .external_lex_state = 2},
  [1918] = {.lex_state = 443, .external_lex_state = 2},
  [1919] = {.lex_state = 443, .external_lex_state = 2},
  [1920] = {.lex_state = 443, .external_lex_state = 2},
  [1921] = {.lex_state = 443, .external_lex_state = 2},
  [1922] = {.lex_state = 443, .external_lex_state = 2},
  [1923] = {.lex_state = 443, .external_lex_state = 2},
  [1924] = {.lex_state = 443, .external_lex_state = 2},
  [1925] = {.lex_state = 443, .external_lex_state = 2},
  [1926] = {.lex_state = 443, .external_lex_state = 2},
  [1927] = {.lex_state = 443, .external_lex_state = 2},
  [1928] = {.lex_state = 443, .external_lex_state = 2},
  [1929] = {.lex_state = 443, .external_lex_state = 2},
  [1930] = {.lex_state = 443, .external_lex_state = 2},
  [1931] = {.lex_state = 443, .external_lex_state = 2},
  [1932] = {.lex_state = 443, .external_lex_state = 2},
  [1933] = {.lex_state = 443, .external_lex_state = 2},
  [1934] = {.lex_state = 443, .external_lex_state = 2},
  [1935] = {.lex_state = 443, .external_lex_state = 2},
  [1936] = {.lex_state = 443, .external_lex_state = 2},
  [1937] = {.lex_state = 443, .external_lex_state = 2},
  [1938] = {.lex_state = 443, .external_lex_state = 2},
  [1939] = {.lex_state = 443, .external_lex_state = 2},
  [1940] = {.lex_state = 443, .external_lex_state = 2},
  [1941] = {.lex_state = 443, .external_lex_state = 2},
  [1942] = {.lex_state = 443, .external_lex_state = 2},
  [1943] = {.lex_state = 443, .external_lex_state = 2},
  [1944] = {.lex_state = 443, .external_lex_state = 2},
  [1945] = {.lex_state = 443, .external_lex_state = 2},
  [1946] = {.lex_state = 443, .external_lex_state = 2},
  [1947] = {.lex_state = 443, .external_lex_state = 2},
  [1948] = {.lex_state = 443, .external_lex_state = 2},
  [1949] = {.lex_state = 443, .external_lex_state = 2},
  [1950] = {.lex_state = 443, .external_lex_state = 2},
  [1951] = {.lex_state = 443, .external_lex_state = 2},
  [1952] = {.lex_state = 443, .external_lex_state = 2},
  [1953] = {.lex_state = 443, .external_lex_state = 2},
  [1954] = {.lex_state = 443, .external_lex_state = 2},
  [1955] = {.lex_state = 443, .external_lex_state = 2},
  [1956] = {.lex_state = 443, .external_lex_state = 2},
  [1957] = {.lex_state = 443, .external_lex_state = 2},
  [1958] = {.lex_state = 443, .external_lex_state = 2},
  [1959] = {.lex_state = 443, .external_lex_state = 2},
  [1960] = {.lex_state = 443, .external_lex_state = 2},
  [1961] = {.lex_state = 443, .external_lex_state = 2},
  [1962] = {.lex_state = 443, .external_lex_state = 2},
  [1963] = {.lex_state = 443, .external_lex_state = 2},
  [1964] = {.lex_state = 443, .external_lex_state = 2},
  [1965] = {.lex_state = 443, .external_lex_state = 2},
  [1966] = {.lex_state = 443, .external_lex_state = 2},
  [1967] = {.lex_state = 443, .external_lex_state = 2},
  [1968] = {.lex_state = 443, .external_lex_state = 2},
  [1969] = {.lex_state = 443, .external_lex_state = 2},
  [1970] = {.lex_state = 443, .external_lex_state = 2},
  [1971] = {.lex_state = 443, .external_lex_state = 2},
  [1972] = {.lex_state = 443, .external_lex_state = 2},
  [1973] = {.lex_state = 443, .external_lex_state = 2},
  [1974] = {.lex_state = 443, .external_lex_state = 2},
  [1975] = {.lex_state = 443, .external_lex_state = 2},
  [1976] = {.lex_state = 443, .external_lex_state = 2},
  [1977] = {.lex_state = 443, .external_lex_state = 2},
  [1978] = {.lex_state = 443, .external_lex_state = 2},
  [1979] = {.lex_state = 443, .external_lex_state = 2},
  [1980] = {.lex_state = 443, .external_lex_state = 2},
  [1981] = {.lex_state = 443, .external_lex_state = 2},
  [1982] = {.lex_state = 443, .external_lex_state = 2},
  [1983] = {.lex_state = 443, .external_lex_state = 2},
  [1984] = {.lex_state = 443, .external_lex_state = 2},
  [1985] = {.lex_state = 443, .external_lex_state = 2},
  [1986] = {.lex_state = 443, .external_lex_state = 2},
  [1987] = {.lex_state = 443, .external_lex_state = 2},
  [1988] = {.lex_state = 443, .external_lex_state = 2},
  [1989] = {.lex_state = 443, .external_lex_state = 2},
  [1990] = {.lex_state = 443, .external_lex_state = 2},
  [1991] = {.lex_state = 443, .external_lex_state = 2},
  [1992] = {.lex_state = 443, .external_lex_state = 2},
  [1993] = {.lex_state = 443, .external_lex_state = 2},
  [1994] = {.lex_state = 443, .external_lex_state = 2},
  [1995] = {.lex_state = 443, .external_lex_state = 2},
  [1996] = {.lex_state = 443, .external_lex_state = 2},
  [1997] = {.lex_state = 443, .external_lex_state = 2},
  [1998] = {.lex_state = 443, .external_lex_state = 2},
  [1999] = {.lex_state = 443, .external_lex_state = 2},
  [2000] = {.lex_state = 443, .external_lex_state = 2},
  [2001] = {.lex_state = 443, .external_lex_state = 2},
  [2002] = {.lex_state = 443, .external_lex_state = 2},
  [2003] = {.lex_state = 443, .external_lex_state = 2},
  [2004] = {.lex_state = 443, .external_lex_state = 2},
  [2005] = {.lex_state = 443, .external_lex_state = 2},
  [2006] = {.lex_state = 443, .external_lex_state = 2},
  [2007] = {.lex_state = 443, .external_lex_state = 2},
  [2008] = {.lex_state = 443, .external_lex_state = 2},
  [2009] = {.lex_state = 443, .external_lex_state = 2},
  [2010] = {.lex_state = 443, .external_lex_state = 2},
  [2011] = {.lex_state = 443, .external_lex_state = 2},
  [2012] = {.lex_state = 443, .external_lex_state = 2},
  [2013] = {.lex_state = 443, .external_lex_state = 2},
  [2014] = {.lex_state = 443, .external_lex_state = 2},
  [2015] = {.lex_state = 443, .external_lex_state = 2},
  [2016] = {.lex_state = 443, .external_lex_state = 2},
  [2017] = {.lex_state = 443, .external_lex_state = 2},
  [2018] = {.lex_state = 443, .external_lex_state = 2},
  [2019] = {.lex_state = 443, .external_lex_state = 2},
  [2020] = {.lex_state = 443, .external_lex_state = 2},
  [2021] = {.lex_state = 443, .external_lex_state = 2},
  [2022] = {.lex_state = 443, .external_lex_state = 2},
  [2023] = {.lex_state = 443, .external_lex_state = 2},
  [2024] = {.lex_state = 443, .external_lex_state = 2},
  [2025] = {.lex_state = 443, .external_lex_state = 2},
  [2026] = {.lex_state = 443, .external_lex_state = 2},
  [2027] = {.lex_state = 443, .external_lex_state = 2},
  [2028] = {.lex_state = 443, .external_lex_state = 2},
  [2029] = {.lex_state = 443, .external_lex_state = 2},
  [2030] = {.lex_state = 443, .external_lex_state = 2},
  [2031] = {.lex_state = 443, .external_lex_state = 2},
  [2032] = {.lex_state = 443, .external_lex_state = 2},
  [2033] = {.lex_state = 443, .external_lex_state = 2},
  [2034] = {.lex_state = 443, .external_lex_state = 2},
  [2035] = {.lex_state = 443, .external_lex_state = 2},
  [2036] = {.lex_state = 443, .external_lex_state = 2},
  [2037] = {.lex_state = 443, .external_lex_state = 2},
  [2038] = {.lex_state = 443, .external_lex_state = 2},
  [2039] = {.lex_state = 443, .external_lex_state = 2},
  [2040] = {.lex_state = 443, .external_lex_state = 2},
  [2041] = {.lex_state = 443, .external_lex_state = 2},
  [2042] = {.lex_state = 443, .external_lex_state = 2},
  [2043] = {.lex_state = 443, .external_lex_state = 2},
  [2044] = {.lex_state = 443, .external_lex_state = 2},
  [2045] = {.lex_state = 443, .external_lex_state = 2},
  [2046] = {.lex_state = 443, .external_lex_state = 2},
  [2047] = {.lex_state = 443, .external_lex_state = 2},
  [2048] = {.lex_state = 443, .external_lex_state = 2},
  [2049] = {.lex_state = 443, .external_lex_state = 2},
  [2050] = {.lex_state = 443, .external_lex_state = 2},
  [2051] = {.lex_state = 443, .external_lex_state = 2},
  [2052] = {.lex_state = 443, .external_lex_state = 2},
  [2053] = {.lex_state = 443, .external_lex_state = 2},
  [2054] = {.lex_state = 443, .external_lex_state = 2},
  [2055] = {.lex_state = 443, .external_lex_state = 2},
  [2056] = {.lex_state = 443, .external_lex_state = 2},
  [2057] = {.lex_state = 443, .external_lex_state = 2},
  [2058] = {.lex_state = 443, .external_lex_state = 2},
  [2059] = {.lex_state = 443, .external_lex_state = 2},
  [2060] = {.lex_state = 443, .external_lex_state = 2},
  [2061] = {.lex_state = 443, .external_lex_state = 2},
  [2062] = {.lex_state = 443, .external_lex_state = 2},
  [2063] = {.lex_state = 443, .external_lex_state = 2},
  [2064] = {.lex_state = 443, .external_lex_state = 2},
  [2065] = {.lex_state = 443, .external_lex_state = 2},
  [2066] = {.lex_state = 443, .external_lex_state = 2},
  [2067] = {.lex_state = 443, .external_lex_state = 2},
  [2068] = {.lex_state = 443, .external_lex_state = 2},
  [2069] = {.lex_state = 443, .external_lex_state = 2},
  [2070] = {.lex_state = 443, .external_lex_state = 2},
  [2071] = {.lex_state = 443, .external_lex_state = 2},
  [2072] = {.lex_state = 443, .external_lex_state = 2},
  [2073] = {.lex_state = 443, .external_lex_state = 2},
  [2074] = {.lex_state = 443, .external_lex_state = 2},
  [2075] = {.lex_state = 443, .external_lex_state = 2},
  [2076] = {.lex_state = 443, .external_lex_state = 2},
  [2077] = {.lex_state = 443, .external_lex_state = 2},
  [2078] = {.lex_state = 443, .external_lex_state = 2},
  [2079] = {.lex_state = 443, .external_lex_state = 2},
  [2080] = {.lex_state = 443, .external_lex_state = 2},
  [2081] = {.lex_state = 443, .external_lex_state = 2},
  [2082] = {.lex_state = 443, .external_lex_state = 2},
  [2083] = {.lex_state = 443, .external_lex_state = 2},
  [2084] = {.lex_state = 443, .external_lex_state = 2},
  [2085] = {.lex_state = 443, .external_lex_state = 2},
  [2086] = {.lex_state = 443, .external_lex_state = 2},
  [2087] = {.lex_state = 443, .external_lex_state = 2},
  [2088] = {.lex_state = 443, .external_lex_state = 2},
  [2089] = {.lex_state = 443, .external_lex_state = 2},
  [2090] = {.lex_state = 443, .external_lex_state = 2},
  [2091] = {.lex_state = 443, .external_lex_state = 2},
  [2092] = {.lex_state = 443, .external_lex_state = 2},
  [2093] = {.lex_state = 443, .external_lex_state = 2},
  [2094] = {.lex_state = 443, .external_lex_state = 2},
  [2095] = {.lex_state = 443, .external_lex_state = 2},
  [2096] = {.lex_state = 443, .external_lex_state = 2},
  [2097] = {.lex_state = 443, .external_lex_state = 2},
  [2098] = {.lex_state = 443, .external_lex_state = 2},
  [2099] = {.lex_state = 443, .external_lex_state = 2},
  [2100] = {.lex_state = 443, .external_lex_state = 2},
  [2101] = {.lex_state = 443, .external_lex_state = 2},
  [2102] = {.lex_state = 443, .external_lex_state = 2},
  [2103] = {.lex_state = 443, .external_lex_state = 2},
  [2104] = {.lex_state = 443, .external_lex_state = 2},
  [2105] = {.lex_state = 443, .external_lex_state = 2},
  [2106] = {.lex_state = 443, .external_lex_state = 2},
  [2107] = {.lex_state = 443, .external_lex_state = 2},
  [2108] = {.lex_state = 443, .external_lex_state = 2},
  [2109] = {.lex_state = 443, .external_lex_state = 2},
  [2110] = {.lex_state = 443, .external_lex_state = 2},
  [2111] = {.lex_state = 443, .external_lex_state = 2},
  [2112] = {.lex_state = 443, .external_lex_state = 2},
  [2113] = {.lex_state = 443, .external_lex_state = 2},
  [2114] = {.lex_state = 443, .external_lex_state = 2},
  [2115] = {.lex_state = 443, .external_lex_state = 2},
  [2116] = {.lex_state = 443, .external_lex_state = 2},
  [2117] = {.lex_state = 443, .external_lex_state = 2},
  [2118] = {.lex_state = 443, .external_lex_state = 2},
  [2119] = {.lex_state = 443, .external_lex_state = 2},
  [2120] = {.lex_state = 443, .external_lex_state = 2},
  [2121] = {.lex_state = 443, .external_lex_state = 2},
  [2122] = {.lex_state = 443, .external_lex_state = 2},
  [2123] = {.lex_state = 443, .external_lex_state = 2},
  [2124] = {.lex_state = 443, .external_lex_state = 2},
  [2125] = {.lex_state = 443, .external_lex_state = 2},
  [2126] = {.lex_state = 443, .external_lex_state = 2},
  [2127] = {.lex_state = 443, .external_lex_state = 2},
  [2128] = {.lex_state = 443, .external_lex_state = 2},
  [2129] = {.lex_state = 443, .external_lex_state = 2},
  [2130] = {.lex_state = 443, .external_lex_state = 2},
  [2131] = {.lex_state = 443, .external_lex_state = 2},
  [2132] = {.lex_state = 443, .external_lex_state = 2},
  [2133] = {.lex_state = 443, .external_lex_state = 2},
  [2134] = {.lex_state = 443, .external_lex_state = 2},
  [2135] = {.lex_state = 443, .external_lex_state = 2},
  [2136] = {.lex_state = 443, .external_lex_state = 2},
  [2137] = {.lex_state = 443, .external_lex_state = 2},
  [2138] = {.lex_state = 443, .external_lex_state = 2},
  [2139] = {.lex_state = 443, .external_lex_state = 2},
  [2140] = {.lex_state = 443, .external_lex_state = 2},
  [2141] = {.lex_state = 443, .external_lex_state = 2},
  [2142] = {.lex_state = 443, .external_lex_state = 2},
  [2143] = {.lex_state = 443, .external_lex_state = 2},
  [2144] = {.lex_state = 443, .external_lex_state = 2},
  [2145] = {.lex_state = 443, .external_lex_state = 2},
  [2146] = {.lex_state = 443, .external_lex_state = 2},
  [2147] = {.lex_state = 443, .external_lex_state = 2},
  [2148] = {.lex_state = 443, .external_lex_state = 2},
  [2149] = {.lex_state = 443, .external_lex_state = 2},
  [2150] = {.lex_state = 443, .external_lex_state = 2},
  [2151] = {.lex_state = 443, .external_lex_state = 2},
  [2152] = {.lex_state = 443, .external_lex_state = 2},
  [2153] = {.lex_state = 443, .external_lex_state = 2},
  [2154] = {.lex_state = 443, .external_lex_state = 2},
  [2155] = {.lex_state = 443, .external_lex_state = 2},
  [2156] = {.lex_state = 443, .external_lex_state = 2},
  [2157] = {.lex_state = 443, .external_lex_state = 2},
  [2158] = {.lex_state = 443, .external_lex_state = 2},
  [2159] = {.lex_state = 443, .external_lex_state = 2},
  [2160] = {.lex_state = 443, .external_lex_state = 2},
  [2161] = {.lex_state = 443, .external_lex_state = 2},
  [2162] = {.lex_state = 443, .external_lex_state = 2},
  [2163] = {.lex_state = 443, .external_lex_state = 2},
  [2164] = {.lex_state = 443, .external_lex_state = 2},
  [2165] = {.lex_state = 443, .external_lex_state = 2},
  [2166] = {.lex_state = 443, .external_lex_state = 2},
  [2167] = {.lex_state = 443, .external_lex_state = 2},
  [2168] = {.lex_state = 443, .external_lex_state = 2},
  [2169] = {.lex_state = 443, .external_lex_state = 2},
  [2170] = {.lex_state = 443, .external_lex_state = 2},
  [2171] = {.lex_state = 443, .external_lex_state = 2},
  [2172] = {.lex_state = 443, .external_lex_state = 2},
  [2173] = {.lex_state = 443, .external_lex_state = 2},
  [2174] = {.lex_state = 443, .external_lex_state = 2},
  [2175] = {.lex_state = 443, .external_lex_state = 2},
  [2176] = {.lex_state = 443, .external_lex_state = 2},
  [2177] = {.lex_state = 443, .external_lex_state = 2},
  [2178] = {.lex_state = 443, .external_lex_state = 2},
  [2179] = {.lex_state = 443, .external_lex_state = 2},
  [2180] = {.lex_state = 443, .external_lex_state = 2},
  [2181] = {.lex_state = 443, .external_lex_state = 2},
  [2182] = {.lex_state = 443, .external_lex_state = 2},
  [2183] = {.lex_state = 443, .external_lex_state = 2},
  [2184] = {.lex_state = 443, .external_lex_state = 2},
  [2185] = {.lex_state = 443, .external_lex_state = 2},
  [2186] = {.lex_state = 443, .external_lex_state = 2},
  [2187] = {.lex_state = 443, .external_lex_state = 2},
  [2188] = {.lex_state = 443, .external_lex_state = 2},
  [2189] = {.lex_state = 443, .external_lex_state = 2},
  [2190] = {.lex_state = 443, .external_lex_state = 2},
  [2191] = {.lex_state = 443, .external_lex_state = 2},
  [2192] = {.lex_state = 443, .external_lex_state = 2},
  [2193] = {.lex_state = 443, .external_lex_state = 2},
  [2194] = {.lex_state = 443, .external_lex_state = 2},
  [2195] = {.lex_state = 443, .external_lex_state = 2},
  [2196] = {.lex_state = 443, .external_lex_state = 2},
  [2197] = {.lex_state = 443, .external_lex_state = 2},
  [2198] = {.lex_state = 443, .external_lex_state = 2},
  [2199] = {.lex_state = 443, .external_lex_state = 2},
  [2200] = {.lex_state = 443, .external_lex_state = 2},
  [2201] = {.lex_state = 443, .external_lex_state = 2},
  [2202] = {.lex_state = 443, .external_lex_state = 2},
  [2203] = {.lex_state = 443, .external_lex_state = 2},
  [2204] = {.lex_state = 443, .external_lex_state = 2},
  [2205] = {.lex_state = 443, .external_lex_state = 2},
  [2206] = {.lex_state = 443, .external_lex_state = 2},
  [2207] = {.lex_state = 443, .external_lex_state = 2},
  [2208] = {.lex_state = 443, .external_lex_state = 2},
  [2209] = {.lex_state = 443, .external_lex_state = 2},
  [2210] = {.lex_state = 443, .external_lex_state = 2},
  [2211] = {.lex_state = 443, .external_lex_state = 2},
  [2212] = {.lex_state = 443, .external_lex_state = 2},
  [2213] = {.lex_state = 443, .external_lex_state = 2},
  [2214] = {.lex_state = 443, .external_lex_state = 2},
  [2215] = {.lex_state = 443, .external_lex_state = 2},
  [2216] = {.lex_state = 443, .external_lex_state = 2},
  [2217] = {.lex_state = 443, .external_lex_state = 2},
  [2218] = {.lex_state = 443, .external_lex_state = 2},
  [2219] = {.lex_state = 443, .external_lex_state = 2},
  [2220] = {.lex_state = 443, .external_lex_state = 2},
  [2221] = {.lex_state = 423, .external_lex_state = 3},
  [2222] = {.lex_state = 423, .external_lex_state = 4},
  [2223] = {.lex_state = 423, .external_lex_state = 5},
  [2224] = {.lex_state = 423, .external_lex_state = 4},
  [2225] = {.lex_state = 424, .external_lex_state = 3},
  [2226] = {.lex_state = 424, .external_lex_state = 4},
  [2227] = {.lex_state = 424, .external_lex_state = 4},
  [2228] = {.lex_state = 424, .external_lex_state = 5},
  [2229] = {.lex_state = 419, .external_lex_state = 6},
  [2230] = {.lex_state = 423, .external_lex_state = 3},
  [2231] = {.lex_state = 419, .external_lex_state = 8},
  [2232] = {.lex_state = 423, .external_lex_state = 7},
  [2233] = {.lex_state = 423, .external_lex_state = 5},
  [2234] = {.lex_state = 421, .external_lex_state = 3},
  [2235] = {.lex_state = 424, .external_lex_state = 3},
  [2236] = {.lex_state = 421, .external_lex_state = 4},
  [2237] = {.lex_state = 415, .external_lex_state = 6},
  [2238] = {.lex_state = 421, .external_lex_state = 4},
  [2239] = {.lex_state = 419, .external_lex_state = 9},
  [2240] = {.lex_state = 421, .external_lex_state = 4},
  [2241] = {.lex_state = 421, .external_lex_state = 4},
  [2242] = {.lex_state = 421, .external_lex_state = 4},
  [2243] = {.lex_state = 421, .external_lex_state = 3},
  [2244] = {.lex_state = 421, .external_lex_state = 3},
  [2245] = {.lex_state = 421, .external_lex_state = 3},
  [2246] = {.lex_state = 421, .external_lex_state = 3},
  [2247] = {.lex_state = 424, .external_lex_state = 7},
  [2248] = {.lex_state = 415, .external_lex_state = 8},
  [2249] = {.lex_state = 424, .external_lex_state = 5},
  [2250] = {.lex_state = 421, .external_lex_state = 5},
  [2251] = {.lex_state = 421, .external_lex_state = 5},
  [2252] = {.lex_state = 421, .external_lex_state = 5},
  [2253] = {.lex_state = 421, .external_lex_state = 5},
  [2254] = {.lex_state = 421, .external_lex_state = 5},
  [2255] = {.lex_state = 421, .external_lex_state = 3},
  [2256] = {.lex_state = 421, .external_lex_state = 3},
  [2257] = {.lex_state = 415, .external_lex_state = 9},
  [2258] = {.lex_state = 421, .external_lex_state = 4},
  [2259] = {.lex_state = 421, .external_lex_state = 4},
  [2260] = {.lex_state = 429, .external_lex_state = 29},
  [2261] = {.lex_state = 421, .external_lex_state = 4},
  [2262] = {.lex_state = 429, .external_lex_state = 29},
  [2263] = {.lex_state = 421, .external_lex_state = 3},
  [2264] = {.lex_state = 421, .external_lex_state = 3},
  [2265] = {.lex_state = 421, .external_lex_state = 3},
  [2266] = {.lex_state = 421, .external_lex_state = 4},
  [2267] = {.lex_state = 421, .external_lex_state = 4},
  [2268] = {.lex_state = 421, .external_lex_state = 4},
  [2269] = {.lex_state = 421, .external_lex_state = 3},
  [2270] = {.lex_state = 421, .external_lex_state = 3},
  [2271] = {.lex_state = 429, .external_lex_state = 30},
  [2272] = {.lex_state = 421, .external_lex_state = 4},
  [2273] = {.lex_state = 421, .external_lex_state = 4},
  [2274] = {.lex_state = 429, .external_lex_state = 30},
  [2275] = {.lex_state = 421, .external_lex_state = 3},
  [2276] = {.lex_state = 421, .external_lex_state = 4},
  [2277] = {.lex_state = 421, .external_lex_state = 3},
  [2278] = {.lex_state = 421, .external_lex_state = 3},
  [2279] = {.lex_state = 421, .external_lex_state = 3},
  [2280] = {.lex_state = 421, .external_lex_state = 4},
  [2281] = {.lex_state = 421, .external_lex_state = 3},
  [2282] = {.lex_state = 421, .external_lex_state = 3},
  [2283] = {.lex_state = 421, .external_lex_state = 4},
  [2284] = {.lex_state = 421, .external_lex_state = 3},
  [2285] = {.lex_state = 421, .external_lex_state = 3},
  [2286] = {.lex_state = 421, .external_lex_state = 4},
  [2287] = {.lex_state = 421, .external_lex_state = 3},
  [2288] = {.lex_state = 421, .external_lex_state = 3},
  [2289] = {.lex_state = 421, .external_lex_state = 3},
  [2290] = {.lex_state = 421, .external_lex_state = 3},
  [2291] = {.lex_state = 421, .external_lex_state = 5},
  [2292] = {.lex_state = 421, .external_lex_state = 4},
  [2293] = {.lex_state = 421, .external_lex_state = 5},
  [2294] = {.lex_state = 421, .external_lex_state = 4},
  [2295] = {.lex_state = 421, .external_lex_state = 4},
  [2296] = {.lex_state = 421, .external_lex_state = 5},
  [2297] = {.lex_state = 421, .external_lex_state = 4},
  [2298] = {.lex_state = 421, .external_lex_state = 4},
  [2299] = {.lex_state = 421, .external_lex_state = 4},
  [2300] = {.lex_state = 421, .external_lex_state = 4},
  [2301] = {.lex_state = 421, .external_lex_state = 3},
  [2302] = {.lex_state = 421, .external_lex_state = 3},
  [2303] = {.lex_state = 421, .external_lex_state = 3},
  [2304] = {.lex_state = 421, .external_lex_state = 3},
  [2305] = {.lex_state = 421, .external_lex_state = 4},
  [2306] = {.lex_state = 421, .external_lex_state = 3},
  [2307] = {.lex_state = 421, .external_lex_state = 4},
  [2308] = {.lex_state = 421, .external_lex_state = 5},
  [2309] = {.lex_state = 421, .external_lex_state = 5},
  [2310] = {.lex_state = 419, .external_lex_state = 10},
  [2311] = {.lex_state = 419, .external_lex_state = 11},
  [2312] = {.lex_state = 421, .external_lex_state = 4},
  [2313] = {.lex_state = 421, .external_lex_state = 4},
  [2314] = {.lex_state = 429, .external_lex_state = 29},
  [2315] = {.lex_state = 421, .external_lex_state = 5},
  [2316] = {.lex_state = 421, .external_lex_state = 4},
  [2317] = {.lex_state = 421, .external_lex_state = 3},
  [2318] = {.lex_state = 421, .external_lex_state = 3},
  [2319] = {.lex_state = 421, .external_lex_state = 4},
  [2320] = {.lex_state = 429, .external_lex_state = 30},
  [2321] = {.lex_state = 421, .external_lex_state = 5},
  [2322] = {.lex_state = 421, .external_lex_state = 5},
  [2323] = {.lex_state = 429, .external_lex_state = 31},
  [2324] = {.lex_state = 429, .external_lex_state = 31},
  [2325] = {.lex_state = 421, .external_lex_state = 4},
  [2326] = {.lex_state = 421, .external_lex_state = 4},
  [2327] = {.lex_state = 421, .external_lex_state = 3},
  [2328] = {.lex_state = 425, .external_lex_state = 3},
  [2329] = {.lex_state = 421, .external_lex_state = 5},
  [2330] = {.lex_state = 421, .external_lex_state = 5},
  [2331] = {.lex_state = 421, .external_lex_state = 5},
  [2332] = {.lex_state = 421, .external_lex_state = 5},
  [2333] = {.lex_state = 421, .external_lex_state = 5},
  [2334] = {.lex_state = 421, .external_lex_state = 5},
  [2335] = {.lex_state = 421, .external_lex_state = 5},
  [2336] = {.lex_state = 419, .external_lex_state = 12},
  [2337] = {.lex_state = 421, .external_lex_state = 5},
  [2338] = {.lex_state = 421, .external_lex_state = 5},
  [2339] = {.lex_state = 421, .external_lex_state = 5},
  [2340] = {.lex_state = 421, .external_lex_state = 5},
  [2341] = {.lex_state = 421, .external_lex_state = 5},
  [2342] = {.lex_state = 421, .external_lex_state = 5},
  [2343] = {.lex_state = 421, .external_lex_state = 5},
  [2344] = {.lex_state = 421, .external_lex_state = 5},
  [2345] = {.lex_state = 421, .external_lex_state = 5},
  [2346] = {.lex_state = 425, .external_lex_state = 4},
  [2347] = {.lex_state = 421, .external_lex_state = 5},
  [2348] = {.lex_state = 421, .external_lex_state = 5},
  [2349] = {.lex_state = 429, .external_lex_state = 31},
  [2350] = {.lex_state = 421, .external_lex_state = 5},
  [2351] = {.lex_state = 419, .external_lex_state = 11},
  [2352] = {.lex_state = 429, .external_lex_state = 30},
  [2353] = {.lex_state = 426, .external_lex_state = 4},
  [2354] = {.lex_state = 429, .external_lex_state = 29},
  [2355] = {.lex_state = 421, .external_lex_state = 3},
  [2356] = {.lex_state = 422, .external_lex_state = 6},
  [2357] = {.lex_state = 425, .external_lex_state = 5},
  [2358] = {.lex_state = 422, .external_lex_state = 6},
  [2359] = {.lex_state = 422, .external_lex_state = 6},
  [2360] = {.lex_state = 429, .external_lex_state = 29},
  [2361] = {.lex_state = 421, .external_lex_state = 3},
  [2362] = {.lex_state = 421, .external_lex_state = 3},
  [2363] = {.lex_state = 421, .external_lex_state = 3},
  [2364] = {.lex_state = 427, .external_lex_state = 4},
  [2365] = {.lex_state = 429, .external_lex_state = 3},
  [2366] = {.lex_state = 427, .external_lex_state = 4},
  [2367] = {.lex_state = 429, .external_lex_state = 3},
  [2368] = {.lex_state = 427, .external_lex_state = 4},
  [2369] = {.lex_state = 426, .external_lex_state = 4},
  [2370] = {.lex_state = 427, .external_lex_state = 4},
  [2371] = {.lex_state = 428, .external_lex_state = 3},
  [2372] = {.lex_state = 428, .external_lex_state = 3},
  [2373] = {.lex_state = 426, .external_lex_state = 4},
  [2374] = {.lex_state = 428, .external_lex_state = 3},
  [2375] = {.lex_state = 427, .external_lex_state = 3},
  [2376] = {.lex_state = 429, .external_lex_state = 30},
  [2377] = {.lex_state = 426, .external_lex_state = 3},
  [2378] = {.lex_state = 426, .external_lex_state = 3},
  [2379] = {.lex_state = 427, .external_lex_state = 4},
  [2380] = {.lex_state = 428, .external_lex_state = 4},
  [2381] = {.lex_state = 428, .external_lex_state = 4},
  [2382] = {.lex_state = 415, .external_lex_state = 10},
  [2383] = {.lex_state = 426, .external_lex_state = 4},
  [2384] = {.lex_state = 426, .external_lex_state = 3},
  [2385] = {.lex_state = 427, .external_lex_state = 3},
  [2386] = {.lex_state = 426, .external_lex_state = 3},
  [2387] = {.lex_state = 429, .external_lex_state = 4},
  [2388] = {.lex_state = 427, .external_lex_state = 3},
  [2389] = {.lex_state = 427, .external_lex_state = 3},
  [2390] = {.lex_state = 427, .external_lex_state = 3},
  [2391] = {.lex_state = 422, .external_lex_state = 6},
  [2392] = {.lex_state = 422, .external_lex_state = 6},
  [2393] = {.lex_state = 421, .external_lex_state = 3},
  [2394] = {.lex_state = 429, .external_lex_state = 29},
  [2395] = {.lex_state = 415, .external_lex_state = 11},
  [2396] = {.lex_state = 428, .external_lex_state = 4},
  [2397] = {.lex_state = 429, .external_lex_state = 4},
  [2398] = {.lex_state = 429, .external_lex_state = 30},
  [2399] = {.lex_state = 428, .external_lex_state = 3},
  [2400] = {.lex_state = 429, .external_lex_state = 4},
  [2401] = {.lex_state = 419, .external_lex_state = 11},
  [2402] = {.lex_state = 422, .external_lex_state = 9},
  [2403] = {.lex_state = 422, .external_lex_state = 9},
  [2404] = {.lex_state = 415, .external_lex_state = 12},
  [2405] = {.lex_state = 429, .external_lex_state = 4},
  [2406] = {.lex_state = 422, .external_lex_state = 8},
  [2407] = {.lex_state = 428, .external_lex_state = 3},
  [2408] = {.lex_state = 421, .external_lex_state = 5},
  [2409] = {.lex_state = 429, .external_lex_state = 5},
  [2410] = {.lex_state = 429, .external_lex_state = 4},
  [2411] = {.lex_state = 422, .external_lex_state = 9},
  [2412] = {.lex_state = 422, .external_lex_state = 8},
  [2413] = {.lex_state = 421, .external_lex_state = 7},
  [2414] = {.lex_state = 415, .external_lex_state = 11},
  [2415] = {.lex_state = 421, .external_lex_state = 7},
  [2416] = {.lex_state = 429, .external_lex_state = 3},
  [2417] = {.lex_state = 421, .external_lex_state = 7},
  [2418] = {.lex_state = 421, .external_lex_state = 7},
  [2419] = {.lex_state = 429, .external_lex_state = 3},
  [2420] = {.lex_state = 428, .external_lex_state = 3},
  [2421] = {.lex_state = 428, .external_lex_state = 3},
  [2422] = {.lex_state = 429, .external_lex_state = 4},
  [2423] = {.lex_state = 428, .external_lex_state = 3},
  [2424] = {.lex_state = 428, .external_lex_state = 4},
  [2425] = {.lex_state = 427, .external_lex_state = 4},
  [2426] = {.lex_state = 427, .external_lex_state = 4},
  [2427] = {.lex_state = 422, .external_lex_state = 8},
  [2428] = {.lex_state = 422, .external_lex_state = 8},
  [2429] = {.lex_state = 422, .external_lex_state = 8},
  [2430] = {.lex_state = 422, .external_lex_state = 9},
  [2431] = {.lex_state = 428, .external_lex_state = 5},
  [2432] = {.lex_state = 428, .external_lex_state = 5},
  [2433] = {.lex_state = 428, .external_lex_state = 4},
  [2434] = {.lex_state = 429, .external_lex_state = 3},
  [2435] = {.lex_state = 426, .external_lex_state = 5},
  [2436] = {.lex_state = 429, .external_lex_state = 4},
  [2437] = {.lex_state = 421, .external_lex_state = 5},
  [2438] = {.lex_state = 427, .external_lex_state = 4},
  [2439] = {.lex_state = 421, .external_lex_state = 5},
  [2440] = {.lex_state = 429, .external_lex_state = 3},
  [2441] = {.lex_state = 421, .external_lex_state = 5},
  [2442] = {.lex_state = 427, .external_lex_state = 3},
  [2443] = {.lex_state = 428, .external_lex_state = 3},
  [2444] = {.lex_state = 421, .external_lex_state = 5},
  [2445] = {.lex_state = 428, .external_lex_state = 4},
  [2446] = {.lex_state = 428, .external_lex_state = 4},
  [2447] = {.lex_state = 429, .external_lex_state = 31},
  [2448] = {.lex_state = 427, .external_lex_state = 4},
  [2449] = {.lex_state = 428, .external_lex_state = 4},
  [2450] = {.lex_state = 427, .external_lex_state = 4},
  [2451] = {.lex_state = 428, .external_lex_state = 3},
  [2452] = {.lex_state = 428, .external_lex_state = 4},
  [2453] = {.lex_state = 429, .external_lex_state = 31},
  [2454] = {.lex_state = 428, .external_lex_state = 4},
  [2455] = {.lex_state = 428, .external_lex_state = 4},
  [2456] = {.lex_state = 426, .external_lex_state = 5},
  [2457] = {.lex_state = 427, .external_lex_state = 3},
  [2458] = {.lex_state = 427, .external_lex_state = 5},
  [2459] = {.lex_state = 422, .external_lex_state = 9},
  [2460] = {.lex_state = 427, .external_lex_state = 3},
  [2461] = {.lex_state = 426, .external_lex_state = 5},
  [2462] = {.lex_state = 427, .external_lex_state = 5},
  [2463] = {.lex_state = 427, .external_lex_state = 5},
  [2464] = {.lex_state = 427, .external_lex_state = 5},
  [2465] = {.lex_state = 429, .external_lex_state = 5},
  [2466] = {.lex_state = 427, .external_lex_state = 5},
  [2467] = {.lex_state = 428, .external_lex_state = 4},
  [2468] = {.lex_state = 427, .external_lex_state = 3},
  [2469] = {.lex_state = 428, .external_lex_state = 4},
  [2470] = {.lex_state = 421, .external_lex_state = 7},
  [2471] = {.lex_state = 426, .external_lex_state = 5},
  [2472] = {.lex_state = 428, .external_lex_state = 4},
  [2473] = {.lex_state = 428, .external_lex_state = 4},
  [2474] = {.lex_state = 428, .external_lex_state = 4},
  [2475] = {.lex_state = 428, .external_lex_state = 5},
  [2476] = {.lex_state = 429, .external_lex_state = 31},
  [2477] = {.lex_state = 428, .external_lex_state = 3},
  [2478] = {.lex_state = 428, .external_lex_state = 3},
  [2479] = {.lex_state = 428, .external_lex_state = 3},
  [2480] = {.lex_state = 428, .external_lex_state = 3},
  [2481] = {.lex_state = 427, .external_lex_state = 3},
  [2482] = {.lex_state = 428, .external_lex_state = 3},
  [2483] = {.lex_state = 428, .external_lex_state = 3},
  [2484] = {.lex_state = 429, .external_lex_state = 3},
  [2485] = {.lex_state = 429, .external_lex_state = 4},
  [2486] = {.lex_state = 422, .external_lex_state = 6},
  [2487] = {.lex_state = 429, .external_lex_state = 4},
  [2488] = {.lex_state = 429, .external_lex_state = 4},
  [2489] = {.lex_state = 429, .external_lex_state = 4},
  [2490] = {.lex_state = 429, .external_lex_state = 4},
  [2491] = {.lex_state = 429, .external_lex_state = 4},
  [2492] = {.lex_state = 429, .external_lex_state = 4},
  [2493] = {.lex_state = 429, .external_lex_state = 4},
  [2494] = {.lex_state = 422, .external_lex_state = 6},
  [2495] = {.lex_state = 429, .external_lex_state = 4},
  [2496] = {.lex_state = 429, .external_lex_state = 4},
  [2497] = {.lex_state = 429, .external_lex_state = 4},
  [2498] = {.lex_state = 429, .external_lex_state = 4},
  [2499] = {.lex_state = 429, .external_lex_state = 4},
  [2500] = {.lex_state = 429, .external_lex_state = 3},
  [2501] = {.lex_state = 421, .external_lex_state = 3},
  [2502] = {.lex_state = 429, .external_lex_state = 4},
  [2503] = {.lex_state = 429, .external_lex_state = 4},
  [2504] = {.lex_state = 429, .external_lex_state = 4},
  [2505] = {.lex_state = 421, .external_lex_state = 3},
  [2506] = {.lex_state = 429, .external_lex_state = 4},
  [2507] = {.lex_state = 429, .external_lex_state = 4},
  [2508] = {.lex_state = 429, .external_lex_state = 4},
  [2509] = {.lex_state = 421, .external_lex_state = 3},
  [2510] = {.lex_state = 429, .external_lex_state = 4},
  [2511] = {.lex_state = 429, .external_lex_state = 3},
  [2512] = {.lex_state = 429, .external_lex_state = 4},
  [2513] = {.lex_state = 429, .external_lex_state = 4},
  [2514] = {.lex_state = 429, .external_lex_state = 4},
  [2515] = {.lex_state = 429, .external_lex_state = 4},
  [2516] = {.lex_state = 429, .external_lex_state = 4},
  [2517] = {.lex_state = 429, .external_lex_state = 4},
  [2518] = {.lex_state = 429, .external_lex_state = 4},
  [2519] = {.lex_state = 429, .external_lex_state = 4},
  [2520] = {.lex_state = 429, .external_lex_state = 4},
  [2521] = {.lex_state = 429, .external_lex_state = 4},
  [2522] = {.lex_state = 429, .external_lex_state = 4},
  [2523] = {.lex_state = 429, .external_lex_state = 4},
  [2524] = {.lex_state = 429, .external_lex_state = 4},
  [2525] = {.lex_state = 429, .external_lex_state = 4},
  [2526] = {.lex_state = 429, .external_lex_state = 3},
  [2527] = {.lex_state = 429, .external_lex_state = 4},
  [2528] = {.lex_state = 429, .external_lex_state = 4},
  [2529] = {.lex_state = 429, .external_lex_state = 4},
  [2530] = {.lex_state = 429, .external_lex_state = 4},
  [2531] = {.lex_state = 429, .external_lex_state = 4},
  [2532] = {.lex_state = 429, .external_lex_state = 3},
  [2533] = {.lex_state = 429, .external_lex_state = 4},
  [2534] = {.lex_state = 429, .external_lex_state = 4},
  [2535] = {.lex_state = 429, .external_lex_state = 4},
  [2536] = {.lex_state = 429, .external_lex_state = 4},
  [2537] = {.lex_state = 429, .external_lex_state = 3},
  [2538] = {.lex_state = 429, .external_lex_state = 4},
  [2539] = {.lex_state = 429, .external_lex_state = 3},
  [2540] = {.lex_state = 429, .external_lex_state = 4},
  [2541] = {.lex_state = 429, .external_lex_state = 3},
  [2542] = {.lex_state = 429, .external_lex_state = 4},
  [2543] = {.lex_state = 429, .external_lex_state = 3},
  [2544] = {.lex_state = 429, .external_lex_state = 3},
  [2545] = {.lex_state = 429, .external_lex_state = 3},
  [2546] = {.lex_state = 429, .external_lex_state = 3},
  [2547] = {.lex_state = 429, .external_lex_state = 3},
  [2548] = {.lex_state = 429, .external_lex_state = 3},
  [2549] = {.lex_state = 418, .external_lex_state = 32},
  [2550] = {.lex_state = 429, .external_lex_state = 3},
  [2551] = {.lex_state = 429, .external_lex_state = 3},
  [2552] = {.lex_state = 429, .external_lex_state = 3},
  [2553] = {.lex_state = 429, .external_lex_state = 3},
  [2554] = {.lex_state = 429, .external_lex_state = 3},
  [2555] = {.lex_state = 429, .external_lex_state = 3},
  [2556] = {.lex_state = 429, .external_lex_state = 3},
  [2557] = {.lex_state = 429, .external_lex_state = 3},
  [2558] = {.lex_state = 429, .external_lex_state = 3},
  [2559] = {.lex_state = 418, .external_lex_state = 32},
  [2560] = {.lex_state = 429, .external_lex_state = 3},
  [2561] = {.lex_state = 429, .external_lex_state = 3},
  [2562] = {.lex_state = 429, .external_lex_state = 3},
  [2563] = {.lex_state = 429, .external_lex_state = 3},
  [2564] = {.lex_state = 429, .external_lex_state = 3},
  [2565] = {.lex_state = 429, .external_lex_state = 3},
  [2566] = {.lex_state = 429, .external_lex_state = 3},
  [2567] = {.lex_state = 429, .external_lex_state = 3},
  [2568] = {.lex_state = 429, .external_lex_state = 3},
  [2569] = {.lex_state = 429, .external_lex_state = 3},
  [2570] = {.lex_state = 429, .external_lex_state = 3},
  [2571] = {.lex_state = 429, .external_lex_state = 3},
  [2572] = {.lex_state = 429, .external_lex_state = 3},
  [2573] = {.lex_state = 429, .external_lex_state = 3},
  [2574] = {.lex_state = 429, .external_lex_state = 3},
  [2575] = {.lex_state = 429, .external_lex_state = 3},
  [2576] = {.lex_state = 429, .external_lex_state = 4},
  [2577] = {.lex_state = 429, .external_lex_state = 4},
  [2578] = {.lex_state = 429, .external_lex_state = 4},
  [2579] = {.lex_state = 429, .external_lex_state = 4},
  [2580] = {.lex_state = 429, .external_lex_state = 4},
  [2581] = {.lex_state = 428, .external_lex_state = 5},
  [2582] = {.lex_state = 429, .external_lex_state = 3},
  [2583] = {.lex_state = 429, .external_lex_state = 3},
  [2584] = {.lex_state = 422, .external_lex_state = 6},
  [2585] = {.lex_state = 422, .external_lex_state = 6},
  [2586] = {.lex_state = 421, .external_lex_state = 3},
  [2587] = {.lex_state = 421, .external_lex_state = 3},
  [2588] = {.lex_state = 429, .external_lex_state = 3},
  [2589] = {.lex_state = 429, .external_lex_state = 3},
  [2590] = {.lex_state = 96, .external_lex_state = 9},
  [2591] = {.lex_state = 422, .external_lex_state = 6},
  [2592] = {.lex_state = 427, .external_lex_state = 5},
  [2593] = {.lex_state = 421, .external_lex_state = 3},
  [2594] = {.lex_state = 429, .external_lex_state = 3},
  [2595] = {.lex_state = 429, .external_lex_state = 3},
  [2596] = {.lex_state = 429, .external_lex_state = 4},
  [2597] = {.lex_state = 428, .external_lex_state = 5},
  [2598] = {.lex_state = 429, .external_lex_state = 4},
  [2599] = {.lex_state = 428, .external_lex_state = 5},
  [2600] = {.lex_state = 428, .external_lex_state = 5},
  [2601] = {.lex_state = 429, .external_lex_state = 4},
  [2602] = {.lex_state = 429, .external_lex_state = 3},
  [2603] = {.lex_state = 429, .external_lex_state = 3},
  [2604] = {.lex_state = 429, .external_lex_state = 3},
  [2605] = {.lex_state = 429, .external_lex_state = 3},
  [2606] = {.lex_state = 429, .external_lex_state = 3},
  [2607] = {.lex_state = 422, .external_lex_state = 6},
  [2608] = {.lex_state = 428, .external_lex_state = 5},
  [2609] = {.lex_state = 429, .external_lex_state = 4},
  [2610] = {.lex_state = 429, .external_lex_state = 3},
  [2611] = {.lex_state = 422, .external_lex_state = 6},
  [2612] = {.lex_state = 429, .external_lex_state = 4},
  [2613] = {.lex_state = 415, .external_lex_state = 11},
  [2614] = {.lex_state = 429, .external_lex_state = 3},
  [2615] = {.lex_state = 429, .external_lex_state = 4},
  [2616] = {.lex_state = 429, .external_lex_state = 4},
  [2617] = {.lex_state = 429, .external_lex_state = 3},
  [2618] = {.lex_state = 429, .external_lex_state = 4},
  [2619] = {.lex_state = 421, .external_lex_state = 3},
  [2620] = {.lex_state = 421, .external_lex_state = 3},
  [2621] = {.lex_state = 429, .external_lex_state = 3},
  [2622] = {.lex_state = 429, .external_lex_state = 3},
  [2623] = {.lex_state = 429, .external_lex_state = 3},
  [2624] = {.lex_state = 429, .external_lex_state = 3},
  [2625] = {.lex_state = 429, .external_lex_state = 3},
  [2626] = {.lex_state = 429, .external_lex_state = 4},
  [2627] = {.lex_state = 429, .external_lex_state = 4},
  [2628] = {.lex_state = 429, .external_lex_state = 3},
  [2629] = {.lex_state = 429, .external_lex_state = 3},
  [2630] = {.lex_state = 429, .external_lex_state = 3},
  [2631] = {.lex_state = 428, .external_lex_state = 5},
  [2632] = {.lex_state = 429, .external_lex_state = 3},
  [2633] = {.lex_state = 429, .external_lex_state = 3},
  [2634] = {.lex_state = 429, .external_lex_state = 4},
  [2635] = {.lex_state = 429, .external_lex_state = 3},
  [2636] = {.lex_state = 429, .external_lex_state = 3},
  [2637] = {.lex_state = 429, .external_lex_state = 3},
  [2638] = {.lex_state = 429, .external_lex_state = 30},
  [2639] = {.lex_state = 429, .external_lex_state = 3},
  [2640] = {.lex_state = 429, .external_lex_state = 3},
  [2641] = {.lex_state = 429, .external_lex_state = 3},
  [2642] = {.lex_state = 428, .external_lex_state = 5},
  [2643] = {.lex_state = 429, .external_lex_state = 4},
  [2644] = {.lex_state = 429, .external_lex_state = 5},
  [2645] = {.lex_state = 428, .external_lex_state = 5},
  [2646] = {.lex_state = 428, .external_lex_state = 5},
  [2647] = {.lex_state = 429, .external_lex_state = 4},
  [2648] = {.lex_state = 429, .external_lex_state = 5},
  [2649] = {.lex_state = 429, .external_lex_state = 5},
  [2650] = {.lex_state = 429, .external_lex_state = 30},
  [2651] = {.lex_state = 427, .external_lex_state = 5},
  [2652] = {.lex_state = 427, .external_lex_state = 5},
  [2653] = {.lex_state = 429, .external_lex_state = 4},
  [2654] = {.lex_state = 427, .external_lex_state = 5},
  [2655] = {.lex_state = 427, .external_lex_state = 5},
  [2656] = {.lex_state = 429, .external_lex_state = 4},
  [2657] = {.lex_state = 428, .external_lex_state = 5},
  [2658] = {.lex_state = 429, .external_lex_state = 4},
  [2659] = {.lex_state = 429, .external_lex_state = 5},
  [2660] = {.lex_state = 429, .external_lex_state = 4},
  [2661] = {.lex_state = 429, .external_lex_state = 4},
  [2662] = {.lex_state = 429, .external_lex_state = 4},
  [2663] = {.lex_state = 429, .external_lex_state = 5},
  [2664] = {.lex_state = 422, .external_lex_state = 6},
  [2665] = {.lex_state = 428, .external_lex_state = 5},
  [2666] = {.lex_state = 428, .external_lex_state = 5},
  [2667] = {.lex_state = 428, .external_lex_state = 5},
  [2668] = {.lex_state = 422, .external_lex_state = 6},
  [2669] = {.lex_state = 422, .external_lex_state = 6},
  [2670] = {.lex_state = 421, .external_lex_state = 5},
  [2671] = {.lex_state = 421, .external_lex_state = 3},
  [2672] = {.lex_state = 421, .external_lex_state = 3},
  [2673] = {.lex_state = 429, .external_lex_state = 5},
  [2674] = {.lex_state = 429, .external_lex_state = 5},
  [2675] = {.lex_state = 429, .external_lex_state = 5},
  [2676] = {.lex_state = 429, .external_lex_state = 5},
  [2677] = {.lex_state = 429, .external_lex_state = 5},
  [2678] = {.lex_state = 429, .external_lex_state = 5},
  [2679] = {.lex_state = 429, .external_lex_state = 5},
  [2680] = {.lex_state = 429, .external_lex_state = 5},
  [2681] = {.lex_state = 429, .external_lex_state = 5},
  [2682] = {.lex_state = 429, .external_lex_state = 5},
  [2683] = {.lex_state = 429, .external_lex_state = 5},
  [2684] = {.lex_state = 429, .external_lex_state = 5},
  [2685] = {.lex_state = 429, .external_lex_state = 5},
  [2686] = {.lex_state = 429, .external_lex_state = 5},
  [2687] = {.lex_state = 429, .external_lex_state = 5},
  [2688] = {.lex_state = 429, .external_lex_state = 5},
  [2689] = {.lex_state = 429, .external_lex_state = 5},
  [2690] = {.lex_state = 429, .external_lex_state = 5},
  [2691] = {.lex_state = 429, .external_lex_state = 5},
  [2692] = {.lex_state = 429, .external_lex_state = 5},
  [2693] = {.lex_state = 429, .external_lex_state = 5},
  [2694] = {.lex_state = 429, .external_lex_state = 5},
  [2695] = {.lex_state = 422, .external_lex_state = 9},
  [2696] = {.lex_state = 422, .external_lex_state = 6},
  [2697] = {.lex_state = 422, .external_lex_state = 6},
  [2698] = {.lex_state = 422, .external_lex_state = 6},
  [2699] = {.lex_state = 422, .external_lex_state = 6},
  [2700] = {.lex_state = 422, .external_lex_state = 6},
  [2701] = {.lex_state = 421, .external_lex_state = 3},
  [2702] = {.lex_state = 421, .external_lex_state = 3},
  [2703] = {.lex_state = 421, .external_lex_state = 3},
  [2704] = {.lex_state = 429, .external_lex_state = 5},
  [2705] = {.lex_state = 429, .external_lex_state = 5},
  [2706] = {.lex_state = 421, .external_lex_state = 3},
  [2707] = {.lex_state = 421, .external_lex_state = 7},
  [2708] = {.lex_state = 421, .external_lex_state = 7},
  [2709] = {.lex_state = 422, .external_lex_state = 8},
  [2710] = {.lex_state = 422, .external_lex_state = 8},
  [2711] = {.lex_state = 429, .external_lex_state = 5},
  [2712] = {.lex_state = 429, .external_lex_state = 5},
  [2713] = {.lex_state = 429, .external_lex_state = 5},
  [2714] = {.lex_state = 422, .external_lex_state = 9},
  [2715] = {.lex_state = 422, .external_lex_state = 8},
  [2716] = {.lex_state = 421, .external_lex_state = 5},
  [2717] = {.lex_state = 421, .external_lex_state = 5},
  [2718] = {.lex_state = 421, .external_lex_state = 7},
  [2719] = {.lex_state = 421, .external_lex_state = 3},
  [2720] = {.lex_state = 418, .external_lex_state = 32},
  [2721] = {.lex_state = 422, .external_lex_state = 8},
  [2722] = {.lex_state = 429, .external_lex_state = 5},
  [2723] = {.lex_state = 421, .external_lex_state = 5},
  [2724] = {.lex_state = 421, .external_lex_state = 3},
  [2725] = {.lex_state = 429, .external_lex_state = 5},
  [2726] = {.lex_state = 422, .external_lex_state = 8},
  [2727] = {.lex_state = 418, .external_lex_state = 33},
  [2728] = {.lex_state = 429, .external_lex_state = 5},
  [2729] = {.lex_state = 429, .external_lex_state = 5},
  [2730] = {.lex_state = 421, .external_lex_state = 3},
  [2731] = {.lex_state = 422, .external_lex_state = 8},
  [2732] = {.lex_state = 422, .external_lex_state = 6},
  [2733] = {.lex_state = 422, .external_lex_state = 6},
  [2734] = {.lex_state = 422, .external_lex_state = 6},
  [2735] = {.lex_state = 422, .external_lex_state = 9},
  [2736] = {.lex_state = 422, .external_lex_state = 6},
  [2737] = {.lex_state = 421, .external_lex_state = 3},
  [2738] = {.lex_state = 421, .external_lex_state = 3},
  [2739] = {.lex_state = 421, .external_lex_state = 3},
  [2740] = {.lex_state = 422, .external_lex_state = 6},
  [2741] = {.lex_state = 421, .external_lex_state = 7},
  [2742] = {.lex_state = 421, .external_lex_state = 3},
  [2743] = {.lex_state = 421, .external_lex_state = 7},
  [2744] = {.lex_state = 421, .external_lex_state = 7},
  [2745] = {.lex_state = 422, .external_lex_state = 8},
  [2746] = {.lex_state = 422, .external_lex_state = 8},
  [2747] = {.lex_state = 418, .external_lex_state = 33},
  [2748] = {.lex_state = 421, .external_lex_state = 3},
  [2749] = {.lex_state = 421, .external_lex_state = 3},
  [2750] = {.lex_state = 422, .external_lex_state = 6},
  [2751] = {.lex_state = 421, .external_lex_state = 3},
  [2752] = {.lex_state = 421, .external_lex_state = 5},
  [2753] = {.lex_state = 421, .external_lex_state = 5},
  [2754] = {.lex_state = 92, .external_lex_state = 13},
  [2755] = {.lex_state = 429, .external_lex_state = 5},
  [2756] = {.lex_state = 429, .external_lex_state = 5},
  [2757] = {.lex_state = 429, .external_lex_state = 5},
  [2758] = {.lex_state = 422, .external_lex_state = 9},
  [2759] = {.lex_state = 420, .external_lex_state = 34},
  [2760] = {.lex_state = 429, .external_lex_state = 5},
  [2761] = {.lex_state = 429, .external_lex_state = 5},
  [2762] = {.lex_state = 429, .external_lex_state = 5},
  [2763] = {.lex_state = 422, .external_lex_state = 9},
  [2764] = {.lex_state = 98, .external_lex_state = 9},
  [2765] = {.lex_state = 421, .external_lex_state = 7},
  [2766] = {.lex_state = 421, .external_lex_state = 3},
  [2767] = {.lex_state = 422, .external_lex_state = 9},
  [2768] = {.lex_state = 429, .external_lex_state = 35},
  [2769] = {.lex_state = 429, .external_lex_state = 5},
  [2770] = {.lex_state = 429, .external_lex_state = 5},
  [2771] = {.lex_state = 429, .external_lex_state = 5},
  [2772] = {.lex_state = 429, .external_lex_state = 5},
  [2773] = {.lex_state = 429, .external_lex_state = 5},
  [2774] = {.lex_state = 422, .external_lex_state = 6},
  [2775] = {.lex_state = 421, .external_lex_state = 3},
  [2776] = {.lex_state = 429, .external_lex_state = 5},
  [2777] = {.lex_state = 429, .external_lex_state = 5},
  [2778] = {.lex_state = 429, .external_lex_state = 5},
  [2779] = {.lex_state = 429, .external_lex_state = 5},
  [2780] = {.lex_state = 429, .external_lex_state = 5},
  [2781] = {.lex_state = 429, .external_lex_state = 5},
  [2782] = {.lex_state = 429, .external_lex_state = 5},
  [2783] = {.lex_state = 429, .external_lex_state = 5},
  [2784] = {.lex_state = 429, .external_lex_state = 5},
  [2785] = {.lex_state = 429, .external_lex_state = 5},
  [2786] = {.lex_state = 429, .external_lex_state = 5},
  [2787] = {.lex_state = 429, .external_lex_state = 5},
  [2788] = {.lex_state = 429, .external_lex_state = 5},
  [2789] = {.lex_state = 429, .external_lex_state = 5},
  [2790] = {.lex_state = 429, .external_lex_state = 5},
  [2791] = {.lex_state = 429, .external_lex_state = 31},
  [2792] = {.lex_state = 429, .external_lex_state = 5},
  [2793] = {.lex_state = 429, .external_lex_state = 5},
  [2794] = {.lex_state = 429, .external_lex_state = 5},
  [2795] = {.lex_state = 429, .external_lex_state = 35},
  [2796] = {.lex_state = 422, .external_lex_state = 9},
  [2797] = {.lex_state = 429, .external_lex_state = 5},
  [2798] = {.lex_state = 422, .external_lex_state = 9},
  [2799] = {.lex_state = 429, .external_lex_state = 31},
  [2800] = {.lex_state = 420, .external_lex_state = 34},
  [2801] = {.lex_state = 422, .external_lex_state = 6},
  [2802] = {.lex_state = 422, .external_lex_state = 6},
  [2803] = {.lex_state = 421, .external_lex_state = 3},
  [2804] = {.lex_state = 429, .external_lex_state = 5},
  [2805] = {.lex_state = 421, .external_lex_state = 7},
  [2806] = {.lex_state = 429, .external_lex_state = 5},
  [2807] = {.lex_state = 422, .external_lex_state = 6},
  [2808] = {.lex_state = 421, .external_lex_state = 5},
  [2809] = {.lex_state = 429, .external_lex_state = 5},
  [2810] = {.lex_state = 429, .external_lex_state = 5},
  [2811] = {.lex_state = 429, .external_lex_state = 5},
  [2812] = {.lex_state = 429, .external_lex_state = 5},
  [2813] = {.lex_state = 422, .external_lex_state = 6},
  [2814] = {.lex_state = 429, .external_lex_state = 30},
  [2815] = {.lex_state = 421, .external_lex_state = 5},
  [2816] = {.lex_state = 422, .external_lex_state = 6},
  [2817] = {.lex_state = 429, .external_lex_state = 5},
  [2818] = {.lex_state = 422, .external_lex_state = 8},
  [2819] = {.lex_state = 422, .external_lex_state = 9},
  [2820] = {.lex_state = 421, .external_lex_state = 5},
  [2821] = {.lex_state = 421, .external_lex_state = 5},
  [2822] = {.lex_state = 97, .external_lex_state = 11},
  [2823] = {.lex_state = 421, .external_lex_state = 7},
  [2824] = {.lex_state = 421, .external_lex_state = 7},
  [2825] = {.lex_state = 421, .external_lex_state = 7},
  [2826] = {.lex_state = 421, .external_lex_state = 5},
  [2827] = {.lex_state = 421, .external_lex_state = 7},
  [2828] = {.lex_state = 421, .external_lex_state = 5},
  [2829] = {.lex_state = 422, .external_lex_state = 8},
  [2830] = {.lex_state = 429, .external_lex_state = 35},
  [2831] = {.lex_state = 421, .external_lex_state = 5},
  [2832] = {.lex_state = 422, .external_lex_state = 9},
  [2833] = {.lex_state = 422, .external_lex_state = 8},
  [2834] = {.lex_state = 422, .external_lex_state = 8},
  [2835] = {.lex_state = 422, .external_lex_state = 9},
  [2836] = {.lex_state = 421, .external_lex_state = 7},
  [2837] = {.lex_state = 418, .external_lex_state = 33},
  [2838] = {.lex_state = 416, .external_lex_state = 6},
  [2839] = {.lex_state = 422, .external_lex_state = 9},
  [2840] = {.lex_state = 422, .external_lex_state = 8},
  [2841] = {.lex_state = 421, .external_lex_state = 7},
  [2842] = {.lex_state = 420, .external_lex_state = 34},
  [2843] = {.lex_state = 421, .external_lex_state = 5},
  [2844] = {.lex_state = 422, .external_lex_state = 8},
  [2845] = {.lex_state = 421, .external_lex_state = 7},
  [2846] = {.lex_state = 421, .external_lex_state = 7},
  [2847] = {.lex_state = 421, .external_lex_state = 7},
  [2848] = {.lex_state = 422, .external_lex_state = 8},
  [2849] = {.lex_state = 91, .external_lex_state = 11},
  [2850] = {.lex_state = 422, .external_lex_state = 9},
  [2851] = {.lex_state = 421, .external_lex_state = 5},
  [2852] = {.lex_state = 422, .external_lex_state = 9},
  [2853] = {.lex_state = 94, .external_lex_state = 14},
  [2854] = {.lex_state = 421, .external_lex_state = 7},
  [2855] = {.lex_state = 421, .external_lex_state = 5},
  [2856] = {.lex_state = 422, .external_lex_state = 8},
  [2857] = {.lex_state = 422, .external_lex_state = 8},
  [2858] = {.lex_state = 422, .external_lex_state = 9},
  [2859] = {.lex_state = 421, .external_lex_state = 7},
  [2860] = {.lex_state = 421, .external_lex_state = 7},
  [2861] = {.lex_state = 422, .external_lex_state = 9},
  [2862] = {.lex_state = 422, .external_lex_state = 8},
  [2863] = {.lex_state = 421, .external_lex_state = 7},
  [2864] = {.lex_state = 422, .external_lex_state = 9},
  [2865] = {.lex_state = 421, .external_lex_state = 5},
  [2866] = {.lex_state = 421, .external_lex_state = 7},
  [2867] = {.lex_state = 422, .external_lex_state = 9},
  [2868] = {.lex_state = 422, .external_lex_state = 9},
  [2869] = {.lex_state = 422, .external_lex_state = 9},
  [2870] = {.lex_state = 422, .external_lex_state = 8},
  [2871] = {.lex_state = 422, .external_lex_state = 8},
  [2872] = {.lex_state = 421, .external_lex_state = 5},
  [2873] = {.lex_state = 422, .external_lex_state = 9},
  [2874] = {.lex_state = 422, .external_lex_state = 9},
  [2875] = {.lex_state = 429, .external_lex_state = 31},
  [2876] = {.lex_state = 422, .external_lex_state = 9},
  [2877] = {.lex_state = 422, .external_lex_state = 9},
  [2878] = {.lex_state = 425, .external_lex_state = 3},
  [2879] = {.lex_state = 422, .external_lex_state = 9},
  [2880] = {.lex_state = 422, .external_lex_state = 9},
  [2881] = {.lex_state = 422, .external_lex_state = 8},
  [2882] = {.lex_state = 421, .external_lex_state = 7},
  [2883] = {.lex_state = 421, .external_lex_state = 7},
  [2884] = {.lex_state = 422, .external_lex_state = 9},
  [2885] = {.lex_state = 422, .external_lex_state = 8},
  [2886] = {.lex_state = 421, .external_lex_state = 5},
  [2887] = {.lex_state = 421, .external_lex_state = 7},
  [2888] = {.lex_state = 422, .external_lex_state = 8},
  [2889] = {.lex_state = 421, .external_lex_state = 5},
  [2890] = {.lex_state = 421, .external_lex_state = 5},
  [2891] = {.lex_state = 419, .external_lex_state = 15},
  [2892] = {.lex_state = 422, .external_lex_state = 8},
  [2893] = {.lex_state = 421, .external_lex_state = 5},
  [2894] = {.lex_state = 421, .external_lex_state = 7},
  [2895] = {.lex_state = 421, .external_lex_state = 5},
  [2896] = {.lex_state = 422, .external_lex_state = 8},
  [2897] = {.lex_state = 419, .external_lex_state = 10},
  [2898] = {.lex_state = 98, .external_lex_state = 13},
  [2899] = {.lex_state = 99, .external_lex_state = 9},
  [2900] = {.lex_state = 421, .external_lex_state = 7},
  [2901] = {.lex_state = 421, .external_lex_state = 5},
  [2902] = {.lex_state = 91, .external_lex_state = 13},
  [2903] = {.lex_state = 421, .external_lex_state = 5},
  [2904] = {.lex_state = 421, .external_lex_state = 5},
  [2905] = {.lex_state = 421, .external_lex_state = 5},
  [2906] = {.lex_state = 422, .external_lex_state = 8},
  [2907] = {.lex_state = 422, .external_lex_state = 8},
  [2908] = {.lex_state = 420, .external_lex_state = 6},
  [2909] = {.lex_state = 427, .external_lex_state = 3},
  [2910] = {.lex_state = 417, .external_lex_state = 6},
  [2911] = {.lex_state = 429, .external_lex_state = 30},
  [2912] = {.lex_state = 426, .external_lex_state = 3},
  [2913] = {.lex_state = 425, .external_lex_state = 7},
  [2914] = {.lex_state = 416, .external_lex_state = 8},
  [2915] = {.lex_state = 419, .external_lex_state = 14},
  [2916] = {.lex_state = 427, .external_lex_state = 3},
  [2917] = {.lex_state = 417, .external_lex_state = 6},
  [2918] = {.lex_state = 425, .external_lex_state = 5},
  [2919] = {.lex_state = 420, .external_lex_state = 6},
  [2920] = {.lex_state = 417, .external_lex_state = 6},
  [2921] = {.lex_state = 429, .external_lex_state = 3},
  [2922] = {.lex_state = 80, .external_lex_state = 11},
  [2923] = {.lex_state = 422, .external_lex_state = 10},
  [2924] = {.lex_state = 430, .external_lex_state = 6},
  [2925] = {.lex_state = 422, .external_lex_state = 11},
  [2926] = {.lex_state = 430, .external_lex_state = 6},
  [2927] = {.lex_state = 426, .external_lex_state = 3},
  [2928] = {.lex_state = 430, .external_lex_state = 6},
  [2929] = {.lex_state = 418, .external_lex_state = 32},
  [2930] = {.lex_state = 422, .external_lex_state = 10},
  [2931] = {.lex_state = 419, .external_lex_state = 19},
  [2932] = {.lex_state = 419, .external_lex_state = 17},
  [2933] = {.lex_state = 417, .external_lex_state = 6},
  [2934] = {.lex_state = 431, .external_lex_state = 6},
  [2935] = {.lex_state = 418, .external_lex_state = 32},
  [2936] = {.lex_state = 430, .external_lex_state = 6},
  [2937] = {.lex_state = 419, .external_lex_state = 18},
  [2938] = {.lex_state = 429, .external_lex_state = 30},
  [2939] = {.lex_state = 98, .external_lex_state = 11},
  [2940] = {.lex_state = 418, .external_lex_state = 32},
  [2941] = {.lex_state = 427, .external_lex_state = 3},
  [2942] = {.lex_state = 430, .external_lex_state = 6},
  [2943] = {.lex_state = 431, .external_lex_state = 6},
  [2944] = {.lex_state = 422, .external_lex_state = 11},
  [2945] = {.lex_state = 431, .external_lex_state = 6},
  [2946] = {.lex_state = 428, .external_lex_state = 3},
  [2947] = {.lex_state = 422, .external_lex_state = 11},
  [2948] = {.lex_state = 422, .external_lex_state = 11},
  [2949] = {.lex_state = 419, .external_lex_state = 13},
  [2950] = {.lex_state = 415, .external_lex_state = 13},
  [2951] = {.lex_state = 428, .external_lex_state = 3},
  [2952] = {.lex_state = 428, .external_lex_state = 3},
  [2953] = {.lex_state = 426, .external_lex_state = 3},
  [2954] = {.lex_state = 429, .external_lex_state = 3},
  [2955] = {.lex_state = 95, .external_lex_state = 11},
  [2956] = {.lex_state = 427, .external_lex_state = 3},
  [2957] = {.lex_state = 426, .external_lex_state = 3},
  [2958] = {.lex_state = 101, .external_lex_state = 9},
  [2959] = {.lex_state = 422, .external_lex_state = 10},
  [2960] = {.lex_state = 419, .external_lex_state = 12},
  [2961] = {.lex_state = 429, .external_lex_state = 30},
  [2962] = {.lex_state = 422, .external_lex_state = 11},
  [2963] = {.lex_state = 427, .external_lex_state = 3},
  [2964] = {.lex_state = 422, .external_lex_state = 10},
  [2965] = {.lex_state = 422, .external_lex_state = 10},
  [2966] = {.lex_state = 427, .external_lex_state = 3},
  [2967] = {.lex_state = 422, .external_lex_state = 12},
  [2968] = {.lex_state = 422, .external_lex_state = 12},
  [2969] = {.lex_state = 429, .external_lex_state = 35},
  [2970] = {.lex_state = 428, .external_lex_state = 5},
  [2971] = {.lex_state = 428, .external_lex_state = 3},
  [2972] = {.lex_state = 427, .external_lex_state = 3},
  [2973] = {.lex_state = 429, .external_lex_state = 7},
  [2974] = {.lex_state = 429, .external_lex_state = 5},
  [2975] = {.lex_state = 427, .external_lex_state = 5},
  [2976] = {.lex_state = 420, .external_lex_state = 8},
  [2977] = {.lex_state = 420, .external_lex_state = 34},
  [2978] = {.lex_state = 427, .external_lex_state = 5},
  [2979] = {.lex_state = 427, .external_lex_state = 5},
  [2980] = {.lex_state = 427, .external_lex_state = 3},
  [2981] = {.lex_state = 431, .external_lex_state = 6},
  [2982] = {.lex_state = 429, .external_lex_state = 31},
  [2983] = {.lex_state = 431, .external_lex_state = 6},
  [2984] = {.lex_state = 427, .external_lex_state = 3},
  [2985] = {.lex_state = 429, .external_lex_state = 35},
  [2986] = {.lex_state = 431, .external_lex_state = 6},
  [2987] = {.lex_state = 420, .external_lex_state = 34},
  [2988] = {.lex_state = 429, .external_lex_state = 35},
  [2989] = {.lex_state = 427, .external_lex_state = 7},
  [2990] = {.lex_state = 430, .external_lex_state = 9},
  [2991] = {.lex_state = 418, .external_lex_state = 33},
  [2992] = {.lex_state = 429, .external_lex_state = 3},
  [2993] = {.lex_state = 430, .external_lex_state = 9},
  [2994] = {.lex_state = 430, .external_lex_state = 9},
  [2995] = {.lex_state = 427, .external_lex_state = 7},
  [2996] = {.lex_state = 430, .external_lex_state = 9},
  [2997] = {.lex_state = 427, .external_lex_state = 7},
  [2998] = {.lex_state = 430, .external_lex_state = 9},
  [2999] = {.lex_state = 418, .external_lex_state = 33},
  [3000] = {.lex_state = 431, .external_lex_state = 6},
  [3001] = {.lex_state = 427, .external_lex_state = 7},
  [3002] = {.lex_state = 427, .external_lex_state = 7},
  [3003] = {.lex_state = 431, .external_lex_state = 6},
  [3004] = {.lex_state = 428, .external_lex_state = 3},
  [3005] = {.lex_state = 429, .external_lex_state = 5},
  [3006] = {.lex_state = 428, .external_lex_state = 3},
  [3007] = {.lex_state = 418, .external_lex_state = 33},
  [3008] = {.lex_state = 428, .external_lex_state = 3},
  [3009] = {.lex_state = 99, .external_lex_state = 14},
  [3010] = {.lex_state = 428, .external_lex_state = 3},
  [3011] = {.lex_state = 428, .external_lex_state = 3},
  [3012] = {.lex_state = 426, .external_lex_state = 7},
  [3013] = {.lex_state = 429, .external_lex_state = 31},
  [3014] = {.lex_state = 426, .external_lex_state = 7},
  [3015] = {.lex_state = 430, .external_lex_state = 6},
  [3016] = {.lex_state = 426, .external_lex_state = 7},
  [3017] = {.lex_state = 426, .external_lex_state = 7},
  [3018] = {.lex_state = 422, .external_lex_state = 12},
  [3019] = {.lex_state = 427, .external_lex_state = 3},
  [3020] = {.lex_state = 417, .external_lex_state = 8},
  [3021] = {.lex_state = 417, .external_lex_state = 8},
  [3022] = {.lex_state = 417, .external_lex_state = 8},
  [3023] = {.lex_state = 417, .external_lex_state = 8},
  [3024] = {.lex_state = 431, .external_lex_state = 6},
  [3025] = {.lex_state = 428, .external_lex_state = 3},
  [3026] = {.lex_state = 416, .external_lex_state = 9},
  [3027] = {.lex_state = 428, .external_lex_state = 7},
  [3028] = {.lex_state = 431, .external_lex_state = 8},
  [3029] = {.lex_state = 431, .external_lex_state = 8},
  [3030] = {.lex_state = 431, .external_lex_state = 6},
  [3031] = {.lex_state = 428, .external_lex_state = 5},
  [3032] = {.lex_state = 428, .external_lex_state = 5},
  [3033] = {.lex_state = 418, .external_lex_state = 6},
  [3034] = {.lex_state = 431, .external_lex_state = 6},
  [3035] = {.lex_state = 418, .external_lex_state = 6},
  [3036] = {.lex_state = 431, .external_lex_state = 6},
  [3037] = {.lex_state = 431, .external_lex_state = 6},
  [3038] = {.lex_state = 429, .external_lex_state = 3},
  [3039] = {.lex_state = 430, .external_lex_state = 8},
  [3040] = {.lex_state = 428, .external_lex_state = 7},
  [3041] = {.lex_state = 426, .external_lex_state = 5},
  [3042] = {.lex_state = 431, .external_lex_state = 6},
  [3043] = {.lex_state = 101, .external_lex_state = 13},
  [3044] = {.lex_state = 428, .external_lex_state = 3},
  [3045] = {.lex_state = 428, .external_lex_state = 3},
  [3046] = {.lex_state = 429, .external_lex_state = 7},
  [3047] = {.lex_state = 431, .external_lex_state = 8},
  [3048] = {.lex_state = 428, .external_lex_state = 3},
  [3049] = {.lex_state = 428, .external_lex_state = 3},
  [3050] = {.lex_state = 426, .external_lex_state = 5},
  [3051] = {.lex_state = 415, .external_lex_state = 15},
  [3052] = {.lex_state = 420, .external_lex_state = 34},
  [3053] = {.lex_state = 420, .external_lex_state = 8},
  [3054] = {.lex_state = 427, .external_lex_state = 5},
  [3055] = {.lex_state = 427, .external_lex_state = 5},
  [3056] = {.lex_state = 426, .external_lex_state = 5},
  [3057] = {.lex_state = 431, .external_lex_state = 6},
  [3058] = {.lex_state = 426, .external_lex_state = 5},
  [3059] = {.lex_state = 430, .external_lex_state = 6},
  [3060] = {.lex_state = 430, .external_lex_state = 6},
  [3061] = {.lex_state = 418, .external_lex_state = 6},
  [3062] = {.lex_state = 430, .external_lex_state = 8},
  [3063] = {.lex_state = 430, .external_lex_state = 8},
  [3064] = {.lex_state = 429, .external_lex_state = 3},
  [3065] = {.lex_state = 430, .external_lex_state = 6},
  [3066] = {.lex_state = 428, .external_lex_state = 3},
  [3067] = {.lex_state = 431, .external_lex_state = 6},
  [3068] = {.lex_state = 430, .external_lex_state = 8},
  [3069] = {.lex_state = 430, .external_lex_state = 8},
  [3070] = {.lex_state = 430, .external_lex_state = 6},
  [3071] = {.lex_state = 429, .external_lex_state = 3},
  [3072] = {.lex_state = 431, .external_lex_state = 9},
  [3073] = {.lex_state = 418, .external_lex_state = 6},
  [3074] = {.lex_state = 431, .external_lex_state = 9},
  [3075] = {.lex_state = 431, .external_lex_state = 9},
  [3076] = {.lex_state = 415, .external_lex_state = 10},
  [3077] = {.lex_state = 428, .external_lex_state = 3},
  [3078] = {.lex_state = 429, .external_lex_state = 31},
  [3079] = {.lex_state = 418, .external_lex_state = 6},
  [3080] = {.lex_state = 422, .external_lex_state = 12},
  [3081] = {.lex_state = 429, .external_lex_state = 3},
  [3082] = {.lex_state = 100, .external_lex_state = 11},
  [3083] = {.lex_state = 422, .external_lex_state = 12},
  [3084] = {.lex_state = 428, .external_lex_state = 7},
  [3085] = {.lex_state = 418, .external_lex_state = 6},
  [3086] = {.lex_state = 422, .external_lex_state = 10},
  [3087] = {.lex_state = 431, .external_lex_state = 8},
  [3088] = {.lex_state = 422, .external_lex_state = 10},
  [3089] = {.lex_state = 431, .external_lex_state = 8},
  [3090] = {.lex_state = 429, .external_lex_state = 3},
  [3091] = {.lex_state = 429, .external_lex_state = 3},
  [3092] = {.lex_state = 429, .external_lex_state = 3},
  [3093] = {.lex_state = 418, .external_lex_state = 36},
  [3094] = {.lex_state = 429, .external_lex_state = 3},
  [3095] = {.lex_state = 431, .external_lex_state = 9},
  [3096] = {.lex_state = 429, .external_lex_state = 3},
  [3097] = {.lex_state = 427, .external_lex_state = 7},
  [3098] = {.lex_state = 429, .external_lex_state = 3},
  [3099] = {.lex_state = 427, .external_lex_state = 7},
  [3100] = {.lex_state = 429, .external_lex_state = 3},
  [3101] = {.lex_state = 422, .external_lex_state = 11},
  [3102] = {.lex_state = 428, .external_lex_state = 7},
  [3103] = {.lex_state = 429, .external_lex_state = 3},
  [3104] = {.lex_state = 415, .external_lex_state = 17},
  [3105] = {.lex_state = 422, .external_lex_state = 11},
  [3106] = {.lex_state = 431, .external_lex_state = 9},
  [3107] = {.lex_state = 418, .external_lex_state = 8},
  [3108] = {.lex_state = 428, .external_lex_state = 5},
  [3109] = {.lex_state = 431, .external_lex_state = 9},
  [3110] = {.lex_state = 431, .external_lex_state = 9},
  [3111] = {.lex_state = 427, .external_lex_state = 5},
  [3112] = {.lex_state = 87, .external_lex_state = 11},
  [3113] = {.lex_state = 418, .external_lex_state = 6},
  [3114] = {.lex_state = 431, .external_lex_state = 9},
  [3115] = {.lex_state = 422, .external_lex_state = 11},
  [3116] = {.lex_state = 418, .external_lex_state = 6},
  [3117] = {.lex_state = 427, .external_lex_state = 7},
  [3118] = {.lex_state = 418, .external_lex_state = 37},
  [3119] = {.lex_state = 431, .external_lex_state = 9},
  [3120] = {.lex_state = 418, .external_lex_state = 37},
  [3121] = {.lex_state = 431, .external_lex_state = 9},
  [3122] = {.lex_state = 431, .external_lex_state = 9},
  [3123] = {.lex_state = 418, .external_lex_state = 6},
  [3124] = {.lex_state = 418, .external_lex_state = 6},
  [3125] = {.lex_state = 422, .external_lex_state = 11},
  [3126] = {.lex_state = 429, .external_lex_state = 7},
  [3127] = {.lex_state = 427, .external_lex_state = 7},
  [3128] = {.lex_state = 81, .external_lex_state = 11},
  [3129] = {.lex_state = 418, .external_lex_state = 8},
  [3130] = {.lex_state = 431, .external_lex_state = 8},
  [3131] = {.lex_state = 431, .external_lex_state = 8},
  [3132] = {.lex_state = 87, .external_lex_state = 11},
  [3133] = {.lex_state = 429, .external_lex_state = 3},
  [3134] = {.lex_state = 418, .external_lex_state = 6},
  [3135] = {.lex_state = 418, .external_lex_state = 6},
  [3136] = {.lex_state = 418, .external_lex_state = 6},
  [3137] = {.lex_state = 429, .external_lex_state = 3},
  [3138] = {.lex_state = 429, .external_lex_state = 3},
  [3139] = {.lex_state = 418, .external_lex_state = 6},
  [3140] = {.lex_state = 429, .external_lex_state = 3},
  [3141] = {.lex_state = 429, .external_lex_state = 3},
  [3142] = {.lex_state = 418, .external_lex_state = 6},
  [3143] = {.lex_state = 418, .external_lex_state = 6},
  [3144] = {.lex_state = 418, .external_lex_state = 6},
  [3145] = {.lex_state = 420, .external_lex_state = 9},
  [3146] = {.lex_state = 418, .external_lex_state = 6},
  [3147] = {.lex_state = 427, .external_lex_state = 7},
  [3148] = {.lex_state = 430, .external_lex_state = 8},
  [3149] = {.lex_state = 427, .external_lex_state = 5},
  [3150] = {.lex_state = 418, .external_lex_state = 6},
  [3151] = {.lex_state = 422, .external_lex_state = 11},
  [3152] = {.lex_state = 428, .external_lex_state = 7},
  [3153] = {.lex_state = 418, .external_lex_state = 6},
  [3154] = {.lex_state = 418, .external_lex_state = 6},
  [3155] = {.lex_state = 427, .external_lex_state = 5},
  [3156] = {.lex_state = 428, .external_lex_state = 5},
  [3157] = {.lex_state = 428, .external_lex_state = 5},
  [3158] = {.lex_state = 418, .external_lex_state = 6},
  [3159] = {.lex_state = 418, .external_lex_state = 6},
  [3160] = {.lex_state = 418, .external_lex_state = 6},
  [3161] = {.lex_state = 428, .external_lex_state = 7},
  [3162] = {.lex_state = 418, .external_lex_state = 6},
  [3163] = {.lex_state = 418, .external_lex_state = 6},
  [3164] = {.lex_state = 431, .external_lex_state = 8},
  [3165] = {.lex_state = 418, .external_lex_state = 8},
  [3166] = {.lex_state = 418, .external_lex_state = 6},
  [3167] = {.lex_state = 431, .external_lex_state = 9},
  [3168] = {.lex_state = 418, .external_lex_state = 6},
  [3169] = {.lex_state = 418, .external_lex_state = 6},
  [3170] = {.lex_state = 428, .external_lex_state = 5},
  [3171] = {.lex_state = 418, .external_lex_state = 6},
  [3172] = {.lex_state = 87, .external_lex_state = 11},
  [3173] = {.lex_state = 418, .external_lex_state = 6},
  [3174] = {.lex_state = 418, .external_lex_state = 6},
  [3175] = {.lex_state = 415, .external_lex_state = 18},
  [3176] = {.lex_state = 418, .external_lex_state = 6},
  [3177] = {.lex_state = 418, .external_lex_state = 36},
  [3178] = {.lex_state = 428, .external_lex_state = 7},
  [3179] = {.lex_state = 418, .external_lex_state = 6},
  [3180] = {.lex_state = 418, .external_lex_state = 6},
  [3181] = {.lex_state = 418, .external_lex_state = 6},
  [3182] = {.lex_state = 418, .external_lex_state = 6},
  [3183] = {.lex_state = 418, .external_lex_state = 6},
  [3184] = {.lex_state = 418, .external_lex_state = 6},
  [3185] = {.lex_state = 418, .external_lex_state = 6},
  [3186] = {.lex_state = 418, .external_lex_state = 6},
  [3187] = {.lex_state = 418, .external_lex_state = 6},
  [3188] = {.lex_state = 418, .external_lex_state = 6},
  [3189] = {.lex_state = 428, .external_lex_state = 5},
  [3190] = {.lex_state = 418, .external_lex_state = 6},
  [3191] = {.lex_state = 418, .external_lex_state = 6},
  [3192] = {.lex_state = 418, .external_lex_state = 6},
  [3193] = {.lex_state = 418, .external_lex_state = 6},
  [3194] = {.lex_state = 415, .external_lex_state = 14},
  [3195] = {.lex_state = 418, .external_lex_state = 6},
  [3196] = {.lex_state = 418, .external_lex_state = 6},
  [3197] = {.lex_state = 418, .external_lex_state = 6},
  [3198] = {.lex_state = 420, .external_lex_state = 9},
  [3199] = {.lex_state = 418, .external_lex_state = 6},
  [3200] = {.lex_state = 418, .external_lex_state = 6},
  [3201] = {.lex_state = 429, .external_lex_state = 3},
  [3202] = {.lex_state = 418, .external_lex_state = 6},
  [3203] = {.lex_state = 420, .external_lex_state = 9},
  [3204] = {.lex_state = 418, .external_lex_state = 6},
  [3205] = {.lex_state = 418, .external_lex_state = 6},
  [3206] = {.lex_state = 429, .external_lex_state = 3},
  [3207] = {.lex_state = 418, .external_lex_state = 6},
  [3208] = {.lex_state = 418, .external_lex_state = 6},
  [3209] = {.lex_state = 428, .external_lex_state = 5},
  [3210] = {.lex_state = 418, .external_lex_state = 6},
  [3211] = {.lex_state = 418, .external_lex_state = 6},
  [3212] = {.lex_state = 429, .external_lex_state = 3},
  [3213] = {.lex_state = 429, .external_lex_state = 3},
  [3214] = {.lex_state = 429, .external_lex_state = 3},
  [3215] = {.lex_state = 428, .external_lex_state = 7},
  [3216] = {.lex_state = 428, .external_lex_state = 7},
  [3217] = {.lex_state = 428, .external_lex_state = 7},
  [3218] = {.lex_state = 420, .external_lex_state = 9},
  [3219] = {.lex_state = 418, .external_lex_state = 6},
  [3220] = {.lex_state = 431, .external_lex_state = 8},
  [3221] = {.lex_state = 418, .external_lex_state = 6},
  [3222] = {.lex_state = 431, .external_lex_state = 8},
  [3223] = {.lex_state = 418, .external_lex_state = 6},
  [3224] = {.lex_state = 422, .external_lex_state = 11},
  [3225] = {.lex_state = 418, .external_lex_state = 6},
  [3226] = {.lex_state = 429, .external_lex_state = 3},
  [3227] = {.lex_state = 431, .external_lex_state = 8},
  [3228] = {.lex_state = 431, .external_lex_state = 8},
  [3229] = {.lex_state = 428, .external_lex_state = 5},
  [3230] = {.lex_state = 431, .external_lex_state = 8},
  [3231] = {.lex_state = 415, .external_lex_state = 19},
  [3232] = {.lex_state = 429, .external_lex_state = 3},
  [3233] = {.lex_state = 422, .external_lex_state = 11},
  [3234] = {.lex_state = 429, .external_lex_state = 3},
  [3235] = {.lex_state = 429, .external_lex_state = 3},
  [3236] = {.lex_state = 429, .external_lex_state = 3},
  [3237] = {.lex_state = 429, .external_lex_state = 3},
  [3238] = {.lex_state = 429, .external_lex_state = 3},
  [3239] = {.lex_state = 429, .external_lex_state = 3},
  [3240] = {.lex_state = 428, .external_lex_state = 7},
  [3241] = {.lex_state = 429, .external_lex_state = 3},
  [3242] = {.lex_state = 429, .external_lex_state = 5},
  [3243] = {.lex_state = 429, .external_lex_state = 3},
  [3244] = {.lex_state = 429, .external_lex_state = 3},
  [3245] = {.lex_state = 429, .external_lex_state = 3},
  [3246] = {.lex_state = 429, .external_lex_state = 5},
  [3247] = {.lex_state = 429, .external_lex_state = 3},
  [3248] = {.lex_state = 429, .external_lex_state = 3},
  [3249] = {.lex_state = 429, .external_lex_state = 3},
  [3250] = {.lex_state = 431, .external_lex_state = 8},
  [3251] = {.lex_state = 429, .external_lex_state = 3},
  [3252] = {.lex_state = 428, .external_lex_state = 5},
  [3253] = {.lex_state = 429, .external_lex_state = 3},
  [3254] = {.lex_state = 418, .external_lex_state = 6},
  [3255] = {.lex_state = 429, .external_lex_state = 3},
  [3256] = {.lex_state = 427, .external_lex_state = 5},
  [3257] = {.lex_state = 418, .external_lex_state = 6},
  [3258] = {.lex_state = 429, .external_lex_state = 3},
  [3259] = {.lex_state = 417, .external_lex_state = 9},
  [3260] = {.lex_state = 429, .external_lex_state = 7},
  [3261] = {.lex_state = 429, .external_lex_state = 3},
  [3262] = {.lex_state = 429, .external_lex_state = 3},
  [3263] = {.lex_state = 429, .external_lex_state = 7},
  [3264] = {.lex_state = 429, .external_lex_state = 3},
  [3265] = {.lex_state = 429, .external_lex_state = 3},
  [3266] = {.lex_state = 429, .external_lex_state = 3},
  [3267] = {.lex_state = 418, .external_lex_state = 6},
  [3268] = {.lex_state = 429, .external_lex_state = 3},
  [3269] = {.lex_state = 418, .external_lex_state = 6},
  [3270] = {.lex_state = 429, .external_lex_state = 3},
  [3271] = {.lex_state = 429, .external_lex_state = 3},
  [3272] = {.lex_state = 429, .external_lex_state = 3},
  [3273] = {.lex_state = 427, .external_lex_state = 5},
  [3274] = {.lex_state = 429, .external_lex_state = 5},
  [3275] = {.lex_state = 418, .external_lex_state = 8},
  [3276] = {.lex_state = 87, .external_lex_state = 11},
  [3277] = {.lex_state = 431, .external_lex_state = 9},
  [3278] = {.lex_state = 418, .external_lex_state = 6},
  [3279] = {.lex_state = 429, .external_lex_state = 3},
  [3280] = {.lex_state = 429, .external_lex_state = 3},
  [3281] = {.lex_state = 429, .external_lex_state = 3},
  [3282] = {.lex_state = 418, .external_lex_state = 6},
  [3283] = {.lex_state = 430, .external_lex_state = 8},
  [3284] = {.lex_state = 430, .external_lex_state = 8},
  [3285] = {.lex_state = 429, .external_lex_state = 3},
  [3286] = {.lex_state = 429, .external_lex_state = 3},
  [3287] = {.lex_state = 418, .external_lex_state = 6},
  [3288] = {.lex_state = 429, .external_lex_state = 3},
  [3289] = {.lex_state = 420, .external_lex_state = 9},
  [3290] = {.lex_state = 430, .external_lex_state = 8},
  [3291] = {.lex_state = 418, .external_lex_state = 6},
  [3292] = {.lex_state = 429, .external_lex_state = 3},
  [3293] = {.lex_state = 417, .external_lex_state = 9},
  [3294] = {.lex_state = 429, .external_lex_state = 3},
  [3295] = {.lex_state = 101, .external_lex_state = 11},
  [3296] = {.lex_state = 430, .external_lex_state = 8},
  [3297] = {.lex_state = 430, .external_lex_state = 9},
  [3298] = {.lex_state = 428, .external_lex_state = 7},
  [3299] = {.lex_state = 431, .external_lex_state = 9},
  [3300] = {.lex_state = 422, .external_lex_state = 10},
  [3301] = {.lex_state = 430, .external_lex_state = 9},
  [3302] = {.lex_state = 429, .external_lex_state = 3},
  [3303] = {.lex_state = 428, .external_lex_state = 5},
  [3304] = {.lex_state = 417, .external_lex_state = 9},
  [3305] = {.lex_state = 418, .external_lex_state = 6},
  [3306] = {.lex_state = 422, .external_lex_state = 10},
  [3307] = {.lex_state = 429, .external_lex_state = 3},
  [3308] = {.lex_state = 422, .external_lex_state = 10},
  [3309] = {.lex_state = 429, .external_lex_state = 5},
  [3310] = {.lex_state = 87, .external_lex_state = 11},
  [3311] = {.lex_state = 422, .external_lex_state = 11},
  [3312] = {.lex_state = 430, .external_lex_state = 9},
  [3313] = {.lex_state = 429, .external_lex_state = 7},
  [3314] = {.lex_state = 431, .external_lex_state = 9},
  [3315] = {.lex_state = 429, .external_lex_state = 3},
  [3316] = {.lex_state = 429, .external_lex_state = 3},
  [3317] = {.lex_state = 415, .external_lex_state = 12},
  [3318] = {.lex_state = 428, .external_lex_state = 5},
  [3319] = {.lex_state = 429, .external_lex_state = 3},
  [3320] = {.lex_state = 418, .external_lex_state = 8},
  [3321] = {.lex_state = 429, .external_lex_state = 5},
  [3322] = {.lex_state = 429, .external_lex_state = 3},
  [3323] = {.lex_state = 428, .external_lex_state = 5},
  [3324] = {.lex_state = 428, .external_lex_state = 5},
  [3325] = {.lex_state = 429, .external_lex_state = 3},
  [3326] = {.lex_state = 430, .external_lex_state = 9},
  [3327] = {.lex_state = 422, .external_lex_state = 10},
  [3328] = {.lex_state = 429, .external_lex_state = 3},
  [3329] = {.lex_state = 429, .external_lex_state = 3},
  [3330] = {.lex_state = 422, .external_lex_state = 10},
  [3331] = {.lex_state = 429, .external_lex_state = 3},
  [3332] = {.lex_state = 428, .external_lex_state = 7},
  [3333] = {.lex_state = 429, .external_lex_state = 3},
  [3334] = {.lex_state = 431, .external_lex_state = 8},
  [3335] = {.lex_state = 429, .external_lex_state = 3},
  [3336] = {.lex_state = 429, .external_lex_state = 3},
  [3337] = {.lex_state = 429, .external_lex_state = 7},
  [3338] = {.lex_state = 429, .external_lex_state = 3},
  [3339] = {.lex_state = 428, .external_lex_state = 7},
  [3340] = {.lex_state = 431, .external_lex_state = 9},
  [3341] = {.lex_state = 422, .external_lex_state = 10},
  [3342] = {.lex_state = 431, .external_lex_state = 8},
  [3343] = {.lex_state = 428, .external_lex_state = 5},
  [3344] = {.lex_state = 430, .external_lex_state = 9},
  [3345] = {.lex_state = 417, .external_lex_state = 9},
  [3346] = {.lex_state = 428, .external_lex_state = 7},
  [3347] = {.lex_state = 428, .external_lex_state = 7},
  [3348] = {.lex_state = 418, .external_lex_state = 6},
  [3349] = {.lex_state = 422, .external_lex_state = 10},
  [3350] = {.lex_state = 422, .external_lex_state = 11},
  [3351] = {.lex_state = 422, .external_lex_state = 11},
  [3352] = {.lex_state = 420, .external_lex_state = 9},
  [3353] = {.lex_state = 418, .external_lex_state = 9},
  [3354] = {.lex_state = 422, .external_lex_state = 11},
  [3355] = {.lex_state = 429, .external_lex_state = 7},
  [3356] = {.lex_state = 420, .external_lex_state = 9},
  [3357] = {.lex_state = 429, .external_lex_state = 7},
  [3358] = {.lex_state = 429, .external_lex_state = 7},
  [3359] = {.lex_state = 429, .external_lex_state = 7},
  [3360] = {.lex_state = 429, .external_lex_state = 7},
  [3361] = {.lex_state = 420, .external_lex_state = 9},
  [3362] = {.lex_state = 420, .external_lex_state = 9},
  [3363] = {.lex_state = 429, .external_lex_state = 7},
  [3364] = {.lex_state = 420, .external_lex_state = 9},
  [3365] = {.lex_state = 429, .external_lex_state = 7},
  [3366] = {.lex_state = 429, .external_lex_state = 7},
  [3367] = {.lex_state = 429, .external_lex_state = 7},
  [3368] = {.lex_state = 418, .external_lex_state = 8},
  [3369] = {.lex_state = 429, .external_lex_state = 7},
  [3370] = {.lex_state = 429, .external_lex_state = 7},
  [3371] = {.lex_state = 429, .external_lex_state = 7},
  [3372] = {.lex_state = 418, .external_lex_state = 37},
  [3373] = {.lex_state = 429, .external_lex_state = 7},
  [3374] = {.lex_state = 429, .external_lex_state = 7},
  [3375] = {.lex_state = 429, .external_lex_state = 7},
  [3376] = {.lex_state = 418, .external_lex_state = 8},
  [3377] = {.lex_state = 429, .external_lex_state = 5},
  [3378] = {.lex_state = 429, .external_lex_state = 7},
  [3379] = {.lex_state = 429, .external_lex_state = 7},
  [3380] = {.lex_state = 429, .external_lex_state = 7},
  [3381] = {.lex_state = 429, .external_lex_state = 7},
  [3382] = {.lex_state = 418, .external_lex_state = 8},
  [3383] = {.lex_state = 418, .external_lex_state = 8},
  [3384] = {.lex_state = 422, .external_lex_state = 11},
  [3385] = {.lex_state = 420, .external_lex_state = 9},
  [3386] = {.lex_state = 429, .external_lex_state = 7},
  [3387] = {.lex_state = 418, .external_lex_state = 8},
  [3388] = {.lex_state = 429, .external_lex_state = 7},
  [3389] = {.lex_state = 429, .external_lex_state = 7},
  [3390] = {.lex_state = 422, .external_lex_state = 11},
  [3391] = {.lex_state = 418, .external_lex_state = 38},
  [3392] = {.lex_state = 429, .external_lex_state = 7},
  [3393] = {.lex_state = 429, .external_lex_state = 7},
  [3394] = {.lex_state = 429, .external_lex_state = 7},
  [3395] = {.lex_state = 418, .external_lex_state = 38},
  [3396] = {.lex_state = 420, .external_lex_state = 9},
  [3397] = {.lex_state = 422, .external_lex_state = 11},
  [3398] = {.lex_state = 420, .external_lex_state = 9},
  [3399] = {.lex_state = 429, .external_lex_state = 5},
  [3400] = {.lex_state = 418, .external_lex_state = 8},
  [3401] = {.lex_state = 429, .external_lex_state = 7},
  [3402] = {.lex_state = 418, .external_lex_state = 8},
  [3403] = {.lex_state = 418, .external_lex_state = 8},
  [3404] = {.lex_state = 418, .external_lex_state = 8},
  [3405] = {.lex_state = 418, .external_lex_state = 8},
  [3406] = {.lex_state = 420, .external_lex_state = 9},
  [3407] = {.lex_state = 418, .external_lex_state = 8},
  [3408] = {.lex_state = 418, .external_lex_state = 8},
  [3409] = {.lex_state = 418, .external_lex_state = 8},
  [3410] = {.lex_state = 422, .external_lex_state = 11},
  [3411] = {.lex_state = 418, .external_lex_state = 8},
  [3412] = {.lex_state = 418, .external_lex_state = 8},
  [3413] = {.lex_state = 420, .external_lex_state = 9},
  [3414] = {.lex_state = 418, .external_lex_state = 8},
  [3415] = {.lex_state = 429, .external_lex_state = 7},
  [3416] = {.lex_state = 422, .external_lex_state = 11},
  [3417] = {.lex_state = 418, .external_lex_state = 8},
  [3418] = {.lex_state = 429, .external_lex_state = 7},
  [3419] = {.lex_state = 418, .external_lex_state = 8},
  [3420] = {.lex_state = 418, .external_lex_state = 8},
  [3421] = {.lex_state = 418, .external_lex_state = 8},
  [3422] = {.lex_state = 418, .external_lex_state = 8},
  [3423] = {.lex_state = 418, .external_lex_state = 8},
  [3424] = {.lex_state = 418, .external_lex_state = 8},
  [3425] = {.lex_state = 418, .external_lex_state = 8},
  [3426] = {.lex_state = 418, .external_lex_state = 8},
  [3427] = {.lex_state = 418, .external_lex_state = 8},
  [3428] = {.lex_state = 418, .external_lex_state = 8},
  [3429] = {.lex_state = 418, .external_lex_state = 8},
  [3430] = {.lex_state = 418, .external_lex_state = 8},
  [3431] = {.lex_state = 418, .external_lex_state = 8},
  [3432] = {.lex_state = 429, .external_lex_state = 7},
  [3433] = {.lex_state = 429, .external_lex_state = 7},
  [3434] = {.lex_state = 418, .external_lex_state = 8},
  [3435] = {.lex_state = 418, .external_lex_state = 8},
  [3436] = {.lex_state = 418, .external_lex_state = 8},
  [3437] = {.lex_state = 418, .external_lex_state = 8},
  [3438] = {.lex_state = 429, .external_lex_state = 7},
  [3439] = {.lex_state = 418, .external_lex_state = 8},
  [3440] = {.lex_state = 429, .external_lex_state = 7},
  [3441] = {.lex_state = 418, .external_lex_state = 8},
  [3442] = {.lex_state = 418, .external_lex_state = 8},
  [3443] = {.lex_state = 418, .external_lex_state = 8},
  [3444] = {.lex_state = 420, .external_lex_state = 9},
  [3445] = {.lex_state = 420, .external_lex_state = 9},
  [3446] = {.lex_state = 420, .external_lex_state = 9},
  [3447] = {.lex_state = 418, .external_lex_state = 8},
  [3448] = {.lex_state = 418, .external_lex_state = 8},
  [3449] = {.lex_state = 418, .external_lex_state = 8},
  [3450] = {.lex_state = 418, .external_lex_state = 8},
  [3451] = {.lex_state = 422, .external_lex_state = 11},
  [3452] = {.lex_state = 429, .external_lex_state = 5},
  [3453] = {.lex_state = 422, .external_lex_state = 12},
  [3454] = {.lex_state = 422, .external_lex_state = 12},
  [3455] = {.lex_state = 420, .external_lex_state = 9},
  [3456] = {.lex_state = 429, .external_lex_state = 7},
  [3457] = {.lex_state = 418, .external_lex_state = 8},
  [3458] = {.lex_state = 418, .external_lex_state = 8},
  [3459] = {.lex_state = 429, .external_lex_state = 7},
  [3460] = {.lex_state = 420, .external_lex_state = 9},
  [3461] = {.lex_state = 429, .external_lex_state = 7},
  [3462] = {.lex_state = 429, .external_lex_state = 7},
  [3463] = {.lex_state = 420, .external_lex_state = 9},
  [3464] = {.lex_state = 420, .external_lex_state = 9},
  [3465] = {.lex_state = 420, .external_lex_state = 9},
  [3466] = {.lex_state = 420, .external_lex_state = 9},
  [3467] = {.lex_state = 429, .external_lex_state = 7},
  [3468] = {.lex_state = 429, .external_lex_state = 7},
  [3469] = {.lex_state = 420, .external_lex_state = 9},
  [3470] = {.lex_state = 429, .external_lex_state = 5},
  [3471] = {.lex_state = 429, .external_lex_state = 5},
  [3472] = {.lex_state = 422, .external_lex_state = 11},
  [3473] = {.lex_state = 429, .external_lex_state = 5},
  [3474] = {.lex_state = 429, .external_lex_state = 5},
  [3475] = {.lex_state = 429, .external_lex_state = 5},
  [3476] = {.lex_state = 429, .external_lex_state = 5},
  [3477] = {.lex_state = 429, .external_lex_state = 5},
  [3478] = {.lex_state = 429, .external_lex_state = 5},
  [3479] = {.lex_state = 422, .external_lex_state = 11},
  [3480] = {.lex_state = 429, .external_lex_state = 5},
  [3481] = {.lex_state = 422, .external_lex_state = 11},
  [3482] = {.lex_state = 429, .external_lex_state = 5},
  [3483] = {.lex_state = 429, .external_lex_state = 5},
  [3484] = {.lex_state = 429, .external_lex_state = 5},
  [3485] = {.lex_state = 422, .external_lex_state = 11},
  [3486] = {.lex_state = 429, .external_lex_state = 5},
  [3487] = {.lex_state = 429, .external_lex_state = 7},
  [3488] = {.lex_state = 422, .external_lex_state = 11},
  [3489] = {.lex_state = 429, .external_lex_state = 5},
  [3490] = {.lex_state = 429, .external_lex_state = 5},
  [3491] = {.lex_state = 422, .external_lex_state = 12},
  [3492] = {.lex_state = 429, .external_lex_state = 5},
  [3493] = {.lex_state = 429, .external_lex_state = 7},
  [3494] = {.lex_state = 429, .external_lex_state = 5},
  [3495] = {.lex_state = 429, .external_lex_state = 7},
  [3496] = {.lex_state = 418, .external_lex_state = 8},
  [3497] = {.lex_state = 429, .external_lex_state = 5},
  [3498] = {.lex_state = 429, .external_lex_state = 5},
  [3499] = {.lex_state = 429, .external_lex_state = 7},
  [3500] = {.lex_state = 418, .external_lex_state = 8},
  [3501] = {.lex_state = 420, .external_lex_state = 9},
  [3502] = {.lex_state = 429, .external_lex_state = 5},
  [3503] = {.lex_state = 429, .external_lex_state = 7},
  [3504] = {.lex_state = 429, .external_lex_state = 5},
  [3505] = {.lex_state = 418, .external_lex_state = 8},
  [3506] = {.lex_state = 420, .external_lex_state = 9},
  [3507] = {.lex_state = 420, .external_lex_state = 9},
  [3508] = {.lex_state = 418, .external_lex_state = 8},
  [3509] = {.lex_state = 418, .external_lex_state = 8},
  [3510] = {.lex_state = 429, .external_lex_state = 5},
  [3511] = {.lex_state = 429, .external_lex_state = 5},
  [3512] = {.lex_state = 429, .external_lex_state = 5},
  [3513] = {.lex_state = 429, .external_lex_state = 5},
  [3514] = {.lex_state = 429, .external_lex_state = 5},
  [3515] = {.lex_state = 429, .external_lex_state = 5},
  [3516] = {.lex_state = 429, .external_lex_state = 5},
  [3517] = {.lex_state = 429, .external_lex_state = 5},
  [3518] = {.lex_state = 420, .external_lex_state = 9},
  [3519] = {.lex_state = 429, .external_lex_state = 7},
  [3520] = {.lex_state = 420, .external_lex_state = 9},
  [3521] = {.lex_state = 429, .external_lex_state = 5},
  [3522] = {.lex_state = 418, .external_lex_state = 8},
  [3523] = {.lex_state = 422, .external_lex_state = 12},
  [3524] = {.lex_state = 418, .external_lex_state = 8},
  [3525] = {.lex_state = 429, .external_lex_state = 5},
  [3526] = {.lex_state = 418, .external_lex_state = 8},
  [3527] = {.lex_state = 422, .external_lex_state = 10},
  [3528] = {.lex_state = 418, .external_lex_state = 8},
  [3529] = {.lex_state = 418, .external_lex_state = 8},
  [3530] = {.lex_state = 418, .external_lex_state = 8},
  [3531] = {.lex_state = 418, .external_lex_state = 8},
  [3532] = {.lex_state = 418, .external_lex_state = 8},
  [3533] = {.lex_state = 418, .external_lex_state = 8},
  [3534] = {.lex_state = 418, .external_lex_state = 8},
  [3535] = {.lex_state = 420, .external_lex_state = 9},
  [3536] = {.lex_state = 418, .external_lex_state = 9},
  [3537] = {.lex_state = 420, .external_lex_state = 9},
  [3538] = {.lex_state = 420, .external_lex_state = 9},
  [3539] = {.lex_state = 420, .external_lex_state = 9},
  [3540] = {.lex_state = 429, .external_lex_state = 5},
  [3541] = {.lex_state = 429, .external_lex_state = 5},
  [3542] = {.lex_state = 429, .external_lex_state = 5},
  [3543] = {.lex_state = 429, .external_lex_state = 5},
  [3544] = {.lex_state = 418, .external_lex_state = 36},
  [3545] = {.lex_state = 420, .external_lex_state = 9},
  [3546] = {.lex_state = 429, .external_lex_state = 5},
  [3547] = {.lex_state = 420, .external_lex_state = 9},
  [3548] = {.lex_state = 429, .external_lex_state = 5},
  [3549] = {.lex_state = 420, .external_lex_state = 9},
  [3550] = {.lex_state = 422, .external_lex_state = 12},
  [3551] = {.lex_state = 418, .external_lex_state = 8},
  [3552] = {.lex_state = 418, .external_lex_state = 8},
  [3553] = {.lex_state = 418, .external_lex_state = 8},
  [3554] = {.lex_state = 420, .external_lex_state = 9},
  [3555] = {.lex_state = 420, .external_lex_state = 9},
  [3556] = {.lex_state = 418, .external_lex_state = 8},
  [3557] = {.lex_state = 429, .external_lex_state = 5},
  [3558] = {.lex_state = 420, .external_lex_state = 9},
  [3559] = {.lex_state = 429, .external_lex_state = 5},
  [3560] = {.lex_state = 422, .external_lex_state = 11},
  [3561] = {.lex_state = 422, .external_lex_state = 11},
  [3562] = {.lex_state = 422, .external_lex_state = 11},
  [3563] = {.lex_state = 429, .external_lex_state = 7},
  [3564] = {.lex_state = 418, .external_lex_state = 8},
  [3565] = {.lex_state = 420, .external_lex_state = 9},
  [3566] = {.lex_state = 429, .external_lex_state = 7},
  [3567] = {.lex_state = 422, .external_lex_state = 10},
  [3568] = {.lex_state = 429, .external_lex_state = 5},
  [3569] = {.lex_state = 429, .external_lex_state = 7},
  [3570] = {.lex_state = 420, .external_lex_state = 9},
  [3571] = {.lex_state = 422, .external_lex_state = 10},
  [3572] = {.lex_state = 422, .external_lex_state = 10},
  [3573] = {.lex_state = 429, .external_lex_state = 5},
  [3574] = {.lex_state = 422, .external_lex_state = 10},
  [3575] = {.lex_state = 418, .external_lex_state = 8},
  [3576] = {.lex_state = 429, .external_lex_state = 7},
  [3577] = {.lex_state = 429, .external_lex_state = 7},
  [3578] = {.lex_state = 422, .external_lex_state = 10},
  [3579] = {.lex_state = 429, .external_lex_state = 5},
  [3580] = {.lex_state = 429, .external_lex_state = 5},
  [3581] = {.lex_state = 422, .external_lex_state = 10},
  [3582] = {.lex_state = 422, .external_lex_state = 10},
  [3583] = {.lex_state = 422, .external_lex_state = 10},
  [3584] = {.lex_state = 429, .external_lex_state = 5},
  [3585] = {.lex_state = 429, .external_lex_state = 7},
  [3586] = {.lex_state = 420, .external_lex_state = 9},
  [3587] = {.lex_state = 429, .external_lex_state = 5},
  [3588] = {.lex_state = 422, .external_lex_state = 10},
  [3589] = {.lex_state = 422, .external_lex_state = 11},
  [3590] = {.lex_state = 429, .external_lex_state = 7},
  [3591] = {.lex_state = 420, .external_lex_state = 9},
  [3592] = {.lex_state = 429, .external_lex_state = 5},
  [3593] = {.lex_state = 429, .external_lex_state = 5},
  [3594] = {.lex_state = 429, .external_lex_state = 7},
  [3595] = {.lex_state = 429, .external_lex_state = 5},
  [3596] = {.lex_state = 429, .external_lex_state = 5},
  [3597] = {.lex_state = 420, .external_lex_state = 9},
  [3598] = {.lex_state = 422, .external_lex_state = 10},
  [3599] = {.lex_state = 429, .external_lex_state = 5},
  [3600] = {.lex_state = 422, .external_lex_state = 10},
  [3601] = {.lex_state = 429, .external_lex_state = 5},
  [3602] = {.lex_state = 429, .external_lex_state = 5},
  [3603] = {.lex_state = 429, .external_lex_state = 7},
  [3604] = {.lex_state = 429, .external_lex_state = 5},
  [3605] = {.lex_state = 429, .external_lex_state = 7},
  [3606] = {.lex_state = 420, .external_lex_state = 9},
  [3607] = {.lex_state = 420, .external_lex_state = 9},
  [3608] = {.lex_state = 429, .external_lex_state = 7},
  [3609] = {.lex_state = 429, .external_lex_state = 7},
  [3610] = {.lex_state = 420, .external_lex_state = 9},
  [3611] = {.lex_state = 418, .external_lex_state = 8},
  [3612] = {.lex_state = 429, .external_lex_state = 5},
  [3613] = {.lex_state = 429, .external_lex_state = 5},
  [3614] = {.lex_state = 422, .external_lex_state = 10},
  [3615] = {.lex_state = 429, .external_lex_state = 7},
  [3616] = {.lex_state = 420, .external_lex_state = 9},
  [3617] = {.lex_state = 429, .external_lex_state = 5},
  [3618] = {.lex_state = 429, .external_lex_state = 5},
  [3619] = {.lex_state = 422, .external_lex_state = 10},
  [3620] = {.lex_state = 422, .external_lex_state = 10},
  [3621] = {.lex_state = 429, .external_lex_state = 5},
  [3622] = {.lex_state = 420, .external_lex_state = 9},
  [3623] = {.lex_state = 422, .external_lex_state = 10},
  [3624] = {.lex_state = 422, .external_lex_state = 10},
  [3625] = {.lex_state = 420, .external_lex_state = 9},
  [3626] = {.lex_state = 429, .external_lex_state = 5},
  [3627] = {.lex_state = 429, .external_lex_state = 5},
  [3628] = {.lex_state = 420, .external_lex_state = 9},
  [3629] = {.lex_state = 420, .external_lex_state = 9},
  [3630] = {.lex_state = 429, .external_lex_state = 5},
  [3631] = {.lex_state = 429, .external_lex_state = 7},
  [3632] = {.lex_state = 429, .external_lex_state = 5},
  [3633] = {.lex_state = 429, .external_lex_state = 5},
  [3634] = {.lex_state = 429, .external_lex_state = 5},
  [3635] = {.lex_state = 429, .external_lex_state = 5},
  [3636] = {.lex_state = 429, .external_lex_state = 7},
  [3637] = {.lex_state = 422, .external_lex_state = 10},
  [3638] = {.lex_state = 429, .external_lex_state = 7},
  [3639] = {.lex_state = 429, .external_lex_state = 7},
  [3640] = {.lex_state = 420, .external_lex_state = 9},
  [3641] = {.lex_state = 420, .external_lex_state = 9},
  [3642] = {.lex_state = 418, .external_lex_state = 8},
  [3643] = {.lex_state = 420, .external_lex_state = 9},
  [3644] = {.lex_state = 420, .external_lex_state = 9},
  [3645] = {.lex_state = 420, .external_lex_state = 9},
  [3646] = {.lex_state = 420, .external_lex_state = 9},
  [3647] = {.lex_state = 420, .external_lex_state = 9},
  [3648] = {.lex_state = 420, .external_lex_state = 9},
  [3649] = {.lex_state = 420, .external_lex_state = 9},
  [3650] = {.lex_state = 422, .external_lex_state = 12},
  [3651] = {.lex_state = 420, .external_lex_state = 9},
  [3652] = {.lex_state = 420, .external_lex_state = 9},
  [3653] = {.lex_state = 420, .external_lex_state = 9},
  [3654] = {.lex_state = 420, .external_lex_state = 9},
  [3655] = {.lex_state = 420, .external_lex_state = 9},
  [3656] = {.lex_state = 420, .external_lex_state = 9},
  [3657] = {.lex_state = 420, .external_lex_state = 9},
  [3658] = {.lex_state = 420, .external_lex_state = 9},
  [3659] = {.lex_state = 420, .external_lex_state = 9},
  [3660] = {.lex_state = 418, .external_lex_state = 8},
  [3661] = {.lex_state = 422, .external_lex_state = 12},
  [3662] = {.lex_state = 429, .external_lex_state = 7},
  [3663] = {.lex_state = 420, .external_lex_state = 9},
  [3664] = {.lex_state = 429, .external_lex_state = 7},
  [3665] = {.lex_state = 429, .external_lex_state = 7},
  [3666] = {.lex_state = 422, .external_lex_state = 11},
  [3667] = {.lex_state = 429, .external_lex_state = 7},
  [3668] = {.lex_state = 422, .external_lex_state = 12},
  [3669] = {.lex_state = 429, .external_lex_state = 7},
  [3670] = {.lex_state = 429, .external_lex_state = 7},
  [3671] = {.lex_state = 429, .external_lex_state = 7},
  [3672] = {.lex_state = 420, .external_lex_state = 9},
  [3673] = {.lex_state = 90, .external_lex_state = 9},
  [3674] = {.lex_state = 422, .external_lex_state = 12},
  [3675] = {.lex_state = 90, .external_lex_state = 9},
  [3676] = {.lex_state = 422, .external_lex_state = 12},
  [3677] = {.lex_state = 87, .external_lex_state = 11},
  [3678] = {.lex_state = 90, .external_lex_state = 9},
  [3679] = {.lex_state = 90, .external_lex_state = 9},
  [3680] = {.lex_state = 422, .external_lex_state = 12},
  [3681] = {.lex_state = 416, .external_lex_state = 10},
  [3682] = {.lex_state = 422, .external_lex_state = 12},
  [3683] = {.lex_state = 87, .external_lex_state = 11},
  [3684] = {.lex_state = 418, .external_lex_state = 36},
  [3685] = {.lex_state = 422, .external_lex_state = 12},
  [3686] = {.lex_state = 87, .external_lex_state = 11},
  [3687] = {.lex_state = 422, .external_lex_state = 12},
  [3688] = {.lex_state = 87, .external_lex_state = 11},
  [3689] = {.lex_state = 87, .external_lex_state = 11},
  [3690] = {.lex_state = 422, .external_lex_state = 12},
  [3691] = {.lex_state = 90, .external_lex_state = 9},
  [3692] = {.lex_state = 422, .external_lex_state = 12},
  [3693] = {.lex_state = 418, .external_lex_state = 36},
  [3694] = {.lex_state = 422, .external_lex_state = 12},
  [3695] = {.lex_state = 418, .external_lex_state = 38},
  [3696] = {.lex_state = 422, .external_lex_state = 12},
  [3697] = {.lex_state = 422, .external_lex_state = 12},
  [3698] = {.lex_state = 422, .external_lex_state = 12},
  [3699] = {.lex_state = 416, .external_lex_state = 11},
  [3700] = {.lex_state = 87, .external_lex_state = 11},
  [3701] = {.lex_state = 422, .external_lex_state = 12},
  [3702] = {.lex_state = 422, .external_lex_state = 12},
  [3703] = {.lex_state = 422, .external_lex_state = 12},
  [3704] = {.lex_state = 422, .external_lex_state = 12},
  [3705] = {.lex_state = 422, .external_lex_state = 12},
  [3706] = {.lex_state = 422, .external_lex_state = 12},
  [3707] = {.lex_state = 87, .external_lex_state = 11},
  [3708] = {.lex_state = 87, .external_lex_state = 11},
  [3709] = {.lex_state = 422, .external_lex_state = 12},
  [3710] = {.lex_state = 87, .external_lex_state = 11},
  [3711] = {.lex_state = 430, .external_lex_state = 11},
  [3712] = {.lex_state = 87, .external_lex_state = 13},
  [3713] = {.lex_state = 87, .external_lex_state = 11},
  [3714] = {.lex_state = 87, .external_lex_state = 13},
  [3715] = {.lex_state = 420, .external_lex_state = 10},
  [3716] = {.lex_state = 418, .external_lex_state = 36},
  [3717] = {.lex_state = 417, .external_lex_state = 11},
  [3718] = {.lex_state = 418, .external_lex_state = 36},
  [3719] = {.lex_state = 431, .external_lex_state = 10},
  [3720] = {.lex_state = 430, .external_lex_state = 11},
  [3721] = {.lex_state = 87, .external_lex_state = 11},
  [3722] = {.lex_state = 87, .external_lex_state = 11},
  [3723] = {.lex_state = 417, .external_lex_state = 10},
  [3724] = {.lex_state = 418, .external_lex_state = 37},
  [3725] = {.lex_state = 420, .external_lex_state = 11},
  [3726] = {.lex_state = 431, .external_lex_state = 10},
  [3727] = {.lex_state = 431, .external_lex_state = 11},
  [3728] = {.lex_state = 417, .external_lex_state = 10},
  [3729] = {.lex_state = 87, .external_lex_state = 11},
  [3730] = {.lex_state = 87, .external_lex_state = 11},
  [3731] = {.lex_state = 430, .external_lex_state = 10},
  [3732] = {.lex_state = 87, .external_lex_state = 11},
  [3733] = {.lex_state = 420, .external_lex_state = 11},
  [3734] = {.lex_state = 430, .external_lex_state = 11},
  [3735] = {.lex_state = 431, .external_lex_state = 11},
  [3736] = {.lex_state = 417, .external_lex_state = 11},
  [3737] = {.lex_state = 430, .external_lex_state = 11},
  [3738] = {.lex_state = 87, .external_lex_state = 11},
  [3739] = {.lex_state = 87, .external_lex_state = 11},
  [3740] = {.lex_state = 430, .external_lex_state = 10},
  [3741] = {.lex_state = 87, .external_lex_state = 13},
  [3742] = {.lex_state = 430, .external_lex_state = 10},
  [3743] = {.lex_state = 420, .external_lex_state = 10},
  [3744] = {.lex_state = 87, .external_lex_state = 11},
  [3745] = {.lex_state = 418, .external_lex_state = 37},
  [3746] = {.lex_state = 418, .external_lex_state = 36},
  [3747] = {.lex_state = 87, .external_lex_state = 11},
  [3748] = {.lex_state = 87, .external_lex_state = 11},
  [3749] = {.lex_state = 431, .external_lex_state = 10},
  [3750] = {.lex_state = 87, .external_lex_state = 11},
  [3751] = {.lex_state = 87, .external_lex_state = 11},
  [3752] = {.lex_state = 418, .external_lex_state = 36},
  [3753] = {.lex_state = 87, .external_lex_state = 11},
  [3754] = {.lex_state = 418, .external_lex_state = 37},
  [3755] = {.lex_state = 417, .external_lex_state = 10},
  [3756] = {.lex_state = 430, .external_lex_state = 10},
  [3757] = {.lex_state = 87, .external_lex_state = 11},
  [3758] = {.lex_state = 417, .external_lex_state = 11},
  [3759] = {.lex_state = 87, .external_lex_state = 11},
  [3760] = {.lex_state = 87, .external_lex_state = 13},
  [3761] = {.lex_state = 430, .external_lex_state = 11},
  [3762] = {.lex_state = 87, .external_lex_state = 11},
  [3763] = {.lex_state = 417, .external_lex_state = 11},
  [3764] = {.lex_state = 87, .external_lex_state = 13},
  [3765] = {.lex_state = 416, .external_lex_state = 12},
  [3766] = {.lex_state = 87, .external_lex_state = 11},
  [3767] = {.lex_state = 430, .external_lex_state = 10},
  [3768] = {.lex_state = 417, .external_lex_state = 10},
  [3769] = {.lex_state = 431, .external_lex_state = 11},
  [3770] = {.lex_state = 430, .external_lex_state = 11},
  [3771] = {.lex_state = 417, .external_lex_state = 12},
  [3772] = {.lex_state = 431, .external_lex_state = 11},
  [3773] = {.lex_state = 431, .external_lex_state = 11},
  [3774] = {.lex_state = 417, .external_lex_state = 12},
  [3775] = {.lex_state = 90, .external_lex_state = 14},
  [3776] = {.lex_state = 430, .external_lex_state = 11},
  [3777] = {.lex_state = 431, .external_lex_state = 10},
  [3778] = {.lex_state = 418, .external_lex_state = 11},
  [3779] = {.lex_state = 90, .external_lex_state = 9},
  [3780] = {.lex_state = 430, .external_lex_state = 11},
  [3781] = {.lex_state = 422, .external_lex_state = 10},
  [3782] = {.lex_state = 431, .external_lex_state = 10},
  [3783] = {.lex_state = 418, .external_lex_state = 10},
  [3784] = {.lex_state = 431, .external_lex_state = 11},
  [3785] = {.lex_state = 430, .external_lex_state = 11},
  [3786] = {.lex_state = 90, .external_lex_state = 9},
  [3787] = {.lex_state = 90, .external_lex_state = 9},
  [3788] = {.lex_state = 431, .external_lex_state = 11},
  [3789] = {.lex_state = 431, .external_lex_state = 10},
  [3790] = {.lex_state = 444, .external_lex_state = 39},
  [3791] = {.lex_state = 431, .external_lex_state = 12},
  [3792] = {.lex_state = 431, .external_lex_state = 10},
  [3793] = {.lex_state = 431, .external_lex_state = 10},
  [3794] = {.lex_state = 422, .external_lex_state = 13},
  [3795] = {.lex_state = 431, .external_lex_state = 10},
  [3796] = {.lex_state = 422, .external_lex_state = 15},
  [3797] = {.lex_state = 420, .external_lex_state = 12},
  [3798] = {.lex_state = 420, .external_lex_state = 12},
  [3799] = {.lex_state = 422, .external_lex_state = 15},
  [3800] = {.lex_state = 422, .external_lex_state = 15},
  [3801] = {.lex_state = 430, .external_lex_state = 12},
  [3802] = {.lex_state = 418, .external_lex_state = 11},
  [3803] = {.lex_state = 431, .external_lex_state = 10},
  [3804] = {.lex_state = 418, .external_lex_state = 10},
  [3805] = {.lex_state = 431, .external_lex_state = 11},
  [3806] = {.lex_state = 431, .external_lex_state = 11},
  [3807] = {.lex_state = 431, .external_lex_state = 11},
  [3808] = {.lex_state = 431, .external_lex_state = 10},
  [3809] = {.lex_state = 431, .external_lex_state = 11},
  [3810] = {.lex_state = 431, .external_lex_state = 10},
  [3811] = {.lex_state = 418, .external_lex_state = 11},
  [3812] = {.lex_state = 90, .external_lex_state = 9},
  [3813] = {.lex_state = 418, .external_lex_state = 11},
  [3814] = {.lex_state = 418, .external_lex_state = 10},
  [3815] = {.lex_state = 108, .external_lex_state = 34},
  [3816] = {.lex_state = 431, .external_lex_state = 11},
  [3817] = {.lex_state = 418, .external_lex_state = 10},
  [3818] = {.lex_state = 416, .external_lex_state = 11},
  [3819] = {.lex_state = 431, .external_lex_state = 11},
  [3820] = {.lex_state = 108, .external_lex_state = 34},
  [3821] = {.lex_state = 430, .external_lex_state = 12},
  [3822] = {.lex_state = 90, .external_lex_state = 14},
  [3823] = {.lex_state = 422, .external_lex_state = 10},
  [3824] = {.lex_state = 422, .external_lex_state = 10},
  [3825] = {.lex_state = 90, .external_lex_state = 9},
  [3826] = {.lex_state = 90, .external_lex_state = 9},
  [3827] = {.lex_state = 90, .external_lex_state = 14},
  [3828] = {.lex_state = 418, .external_lex_state = 38},
  [3829] = {.lex_state = 90, .external_lex_state = 9},
  [3830] = {.lex_state = 418, .external_lex_state = 38},
  [3831] = {.lex_state = 431, .external_lex_state = 12},
  [3832] = {.lex_state = 430, .external_lex_state = 10},
  [3833] = {.lex_state = 431, .external_lex_state = 10},
  [3834] = {.lex_state = 430, .external_lex_state = 10},
  [3835] = {.lex_state = 430, .external_lex_state = 11},
  [3836] = {.lex_state = 422, .external_lex_state = 10},
  [3837] = {.lex_state = 422, .external_lex_state = 13},
  [3838] = {.lex_state = 422, .external_lex_state = 10},
  [3839] = {.lex_state = 422, .external_lex_state = 13},
  [3840] = {.lex_state = 431, .external_lex_state = 12},
  [3841] = {.lex_state = 422, .external_lex_state = 13},
  [3842] = {.lex_state = 430, .external_lex_state = 10},
  [3843] = {.lex_state = 431, .external_lex_state = 10},
  [3844] = {.lex_state = 431, .external_lex_state = 10},
  [3845] = {.lex_state = 431, .external_lex_state = 10},
  [3846] = {.lex_state = 418, .external_lex_state = 11},
  [3847] = {.lex_state = 422, .external_lex_state = 15},
  [3848] = {.lex_state = 430, .external_lex_state = 10},
  [3849] = {.lex_state = 431, .external_lex_state = 11},
  [3850] = {.lex_state = 430, .external_lex_state = 12},
  [3851] = {.lex_state = 431, .external_lex_state = 11},
  [3852] = {.lex_state = 431, .external_lex_state = 10},
  [3853] = {.lex_state = 444, .external_lex_state = 39},
  [3854] = {.lex_state = 430, .external_lex_state = 10},
  [3855] = {.lex_state = 430, .external_lex_state = 12},
  [3856] = {.lex_state = 422, .external_lex_state = 13},
  [3857] = {.lex_state = 422, .external_lex_state = 15},
  [3858] = {.lex_state = 431, .external_lex_state = 11},
  [3859] = {.lex_state = 418, .external_lex_state = 38},
  [3860] = {.lex_state = 417, .external_lex_state = 12},
  [3861] = {.lex_state = 90, .external_lex_state = 9},
  [3862] = {.lex_state = 431, .external_lex_state = 10},
  [3863] = {.lex_state = 417, .external_lex_state = 12},
  [3864] = {.lex_state = 418, .external_lex_state = 10},
  [3865] = {.lex_state = 90, .external_lex_state = 14},
  [3866] = {.lex_state = 90, .external_lex_state = 14},
  [3867] = {.lex_state = 430, .external_lex_state = 12},
  [3868] = {.lex_state = 422, .external_lex_state = 17},
  [3869] = {.lex_state = 418, .external_lex_state = 10},
  [3870] = {.lex_state = 417, .external_lex_state = 11},
  [3871] = {.lex_state = 78, .external_lex_state = 11},
  [3872] = {.lex_state = 418, .external_lex_state = 10},
  [3873] = {.lex_state = 418, .external_lex_state = 10},
  [3874] = {.lex_state = 418, .external_lex_state = 10},
  [3875] = {.lex_state = 418, .external_lex_state = 11},
  [3876] = {.lex_state = 431, .external_lex_state = 12},
  [3877] = {.lex_state = 422, .external_lex_state = 12},
  [3878] = {.lex_state = 422, .external_lex_state = 19},
  [3879] = {.lex_state = 418, .external_lex_state = 11},
  [3880] = {.lex_state = 418, .external_lex_state = 10},
  [3881] = {.lex_state = 418, .external_lex_state = 11},
  [3882] = {.lex_state = 422, .external_lex_state = 12},
  [3883] = {.lex_state = 418, .external_lex_state = 10},
  [3884] = {.lex_state = 418, .external_lex_state = 11},
  [3885] = {.lex_state = 422, .external_lex_state = 12},
  [3886] = {.lex_state = 422, .external_lex_state = 12},
  [3887] = {.lex_state = 418, .external_lex_state = 10},
  [3888] = {.lex_state = 418, .external_lex_state = 11},
  [3889] = {.lex_state = 418, .external_lex_state = 11},
  [3890] = {.lex_state = 78, .external_lex_state = 11},
  [3891] = {.lex_state = 422, .external_lex_state = 11},
  [3892] = {.lex_state = 78, .external_lex_state = 11},
  [3893] = {.lex_state = 422, .external_lex_state = 17},
  [3894] = {.lex_state = 418, .external_lex_state = 10},
  [3895] = {.lex_state = 418, .external_lex_state = 10},
  [3896] = {.lex_state = 418, .external_lex_state = 10},
  [3897] = {.lex_state = 418, .external_lex_state = 11},
  [3898] = {.lex_state = 422, .external_lex_state = 18},
  [3899] = {.lex_state = 444, .external_lex_state = 40},
  [3900] = {.lex_state = 422, .external_lex_state = 19},
  [3901] = {.lex_state = 422, .external_lex_state = 19},
  [3902] = {.lex_state = 418, .external_lex_state = 10},
  [3903] = {.lex_state = 418, .external_lex_state = 10},
  [3904] = {.lex_state = 418, .external_lex_state = 10},
  [3905] = {.lex_state = 418, .external_lex_state = 11},
  [3906] = {.lex_state = 418, .external_lex_state = 11},
  [3907] = {.lex_state = 422, .external_lex_state = 19},
  [3908] = {.lex_state = 418, .external_lex_state = 10},
  [3909] = {.lex_state = 418, .external_lex_state = 10},
  [3910] = {.lex_state = 90, .external_lex_state = 9},
  [3911] = {.lex_state = 418, .external_lex_state = 10},
  [3912] = {.lex_state = 418, .external_lex_state = 11},
  [3913] = {.lex_state = 418, .external_lex_state = 10},
  [3914] = {.lex_state = 90, .external_lex_state = 9},
  [3915] = {.lex_state = 90, .external_lex_state = 9},
  [3916] = {.lex_state = 89, .external_lex_state = 11},
  [3917] = {.lex_state = 418, .external_lex_state = 11},
  [3918] = {.lex_state = 431, .external_lex_state = 12},
  [3919] = {.lex_state = 418, .external_lex_state = 10},
  [3920] = {.lex_state = 444, .external_lex_state = 40},
  [3921] = {.lex_state = 87, .external_lex_state = 13},
  [3922] = {.lex_state = 90, .external_lex_state = 9},
  [3923] = {.lex_state = 418, .external_lex_state = 11},
  [3924] = {.lex_state = 418, .external_lex_state = 11},
  [3925] = {.lex_state = 418, .external_lex_state = 10},
  [3926] = {.lex_state = 418, .external_lex_state = 10},
  [3927] = {.lex_state = 418, .external_lex_state = 11},
  [3928] = {.lex_state = 418, .external_lex_state = 11},
  [3929] = {.lex_state = 418, .external_lex_state = 11},
  [3930] = {.lex_state = 417, .external_lex_state = 11},
  [3931] = {.lex_state = 418, .external_lex_state = 11},
  [3932] = {.lex_state = 79, .external_lex_state = 11},
  [3933] = {.lex_state = 418, .external_lex_state = 10},
  [3934] = {.lex_state = 418, .external_lex_state = 10},
  [3935] = {.lex_state = 87, .external_lex_state = 13},
  [3936] = {.lex_state = 418, .external_lex_state = 11},
  [3937] = {.lex_state = 418, .external_lex_state = 11},
  [3938] = {.lex_state = 418, .external_lex_state = 11},
  [3939] = {.lex_state = 418, .external_lex_state = 10},
  [3940] = {.lex_state = 431, .external_lex_state = 12},
  [3941] = {.lex_state = 418, .external_lex_state = 11},
  [3942] = {.lex_state = 418, .external_lex_state = 10},
  [3943] = {.lex_state = 418, .external_lex_state = 11},
  [3944] = {.lex_state = 78, .external_lex_state = 11},
  [3945] = {.lex_state = 418, .external_lex_state = 11},
  [3946] = {.lex_state = 418, .external_lex_state = 11},
  [3947] = {.lex_state = 418, .external_lex_state = 11},
  [3948] = {.lex_state = 418, .external_lex_state = 10},
  [3949] = {.lex_state = 418, .external_lex_state = 10},
  [3950] = {.lex_state = 418, .external_lex_state = 11},
  [3951] = {.lex_state = 418, .external_lex_state = 11},
  [3952] = {.lex_state = 418, .external_lex_state = 10},
  [3953] = {.lex_state = 418, .external_lex_state = 11},
  [3954] = {.lex_state = 422, .external_lex_state = 11},
  [3955] = {.lex_state = 418, .external_lex_state = 10},
  [3956] = {.lex_state = 418, .external_lex_state = 12},
  [3957] = {.lex_state = 418, .external_lex_state = 11},
  [3958] = {.lex_state = 418, .external_lex_state = 12},
  [3959] = {.lex_state = 87, .external_lex_state = 13},
  [3960] = {.lex_state = 418, .external_lex_state = 11},
  [3961] = {.lex_state = 90, .external_lex_state = 9},
  [3962] = {.lex_state = 418, .external_lex_state = 11},
  [3963] = {.lex_state = 418, .external_lex_state = 10},
  [3964] = {.lex_state = 90, .external_lex_state = 9},
  [3965] = {.lex_state = 418, .external_lex_state = 11},
  [3966] = {.lex_state = 418, .external_lex_state = 10},
  [3967] = {.lex_state = 444, .external_lex_state = 41},
  [3968] = {.lex_state = 79, .external_lex_state = 11},
  [3969] = {.lex_state = 418, .external_lex_state = 11},
  [3970] = {.lex_state = 418, .external_lex_state = 11},
  [3971] = {.lex_state = 418, .external_lex_state = 11},
  [3972] = {.lex_state = 418, .external_lex_state = 10},
  [3973] = {.lex_state = 418, .external_lex_state = 11},
  [3974] = {.lex_state = 418, .external_lex_state = 10},
  [3975] = {.lex_state = 422, .external_lex_state = 19},
  [3976] = {.lex_state = 418, .external_lex_state = 11},
  [3977] = {.lex_state = 418, .external_lex_state = 10},
  [3978] = {.lex_state = 431, .external_lex_state = 12},
  [3979] = {.lex_state = 78, .external_lex_state = 11},
  [3980] = {.lex_state = 87, .external_lex_state = 13},
  [3981] = {.lex_state = 90, .external_lex_state = 9},
  [3982] = {.lex_state = 418, .external_lex_state = 10},
  [3983] = {.lex_state = 418, .external_lex_state = 11},
  [3984] = {.lex_state = 418, .external_lex_state = 11},
  [3985] = {.lex_state = 87, .external_lex_state = 13},
  [3986] = {.lex_state = 431, .external_lex_state = 12},
  [3987] = {.lex_state = 418, .external_lex_state = 11},
  [3988] = {.lex_state = 417, .external_lex_state = 11},
  [3989] = {.lex_state = 418, .external_lex_state = 11},
  [3990] = {.lex_state = 418, .external_lex_state = 10},
  [3991] = {.lex_state = 418, .external_lex_state = 10},
  [3992] = {.lex_state = 418, .external_lex_state = 12},
  [3993] = {.lex_state = 90, .external_lex_state = 9},
  [3994] = {.lex_state = 430, .external_lex_state = 12},
  [3995] = {.lex_state = 418, .external_lex_state = 10},
  [3996] = {.lex_state = 418, .external_lex_state = 10},
  [3997] = {.lex_state = 87, .external_lex_state = 13},
  [3998] = {.lex_state = 431, .external_lex_state = 12},
  [3999] = {.lex_state = 108, .external_lex_state = 34},
  [4000] = {.lex_state = 422, .external_lex_state = 11},
  [4001] = {.lex_state = 418, .external_lex_state = 11},
  [4002] = {.lex_state = 418, .external_lex_state = 10},
  [4003] = {.lex_state = 422, .external_lex_state = 14},
  [4004] = {.lex_state = 422, .external_lex_state = 11},
  [4005] = {.lex_state = 90, .external_lex_state = 9},
  [4006] = {.lex_state = 422, .external_lex_state = 11},
  [4007] = {.lex_state = 418, .external_lex_state = 10},
  [4008] = {.lex_state = 418, .external_lex_state = 10},
  [4009] = {.lex_state = 418, .external_lex_state = 10},
  [4010] = {.lex_state = 418, .external_lex_state = 11},
  [4011] = {.lex_state = 90, .external_lex_state = 9},
  [4012] = {.lex_state = 418, .external_lex_state = 12},
  [4013] = {.lex_state = 418, .external_lex_state = 11},
  [4014] = {.lex_state = 418, .external_lex_state = 10},
  [4015] = {.lex_state = 430, .external_lex_state = 12},
  [4016] = {.lex_state = 418, .external_lex_state = 11},
  [4017] = {.lex_state = 431, .external_lex_state = 12},
  [4018] = {.lex_state = 431, .external_lex_state = 12},
  [4019] = {.lex_state = 89, .external_lex_state = 11},
  [4020] = {.lex_state = 431, .external_lex_state = 12},
  [4021] = {.lex_state = 418, .external_lex_state = 11},
  [4022] = {.lex_state = 418, .external_lex_state = 11},
  [4023] = {.lex_state = 418, .external_lex_state = 10},
  [4024] = {.lex_state = 422, .external_lex_state = 14},
  [4025] = {.lex_state = 418, .external_lex_state = 10},
  [4026] = {.lex_state = 90, .external_lex_state = 9},
  [4027] = {.lex_state = 422, .external_lex_state = 12},
  [4028] = {.lex_state = 90, .external_lex_state = 9},
  [4029] = {.lex_state = 418, .external_lex_state = 11},
  [4030] = {.lex_state = 418, .external_lex_state = 11},
  [4031] = {.lex_state = 418, .external_lex_state = 10},
  [4032] = {.lex_state = 431, .external_lex_state = 12},
  [4033] = {.lex_state = 422, .external_lex_state = 18},
  [4034] = {.lex_state = 418, .external_lex_state = 11},
  [4035] = {.lex_state = 418, .external_lex_state = 10},
  [4036] = {.lex_state = 418, .external_lex_state = 10},
  [4037] = {.lex_state = 418, .external_lex_state = 36},
  [4038] = {.lex_state = 431, .external_lex_state = 12},
  [4039] = {.lex_state = 418, .external_lex_state = 10},
  [4040] = {.lex_state = 418, .external_lex_state = 11},
  [4041] = {.lex_state = 422, .external_lex_state = 14},
  [4042] = {.lex_state = 431, .external_lex_state = 12},
  [4043] = {.lex_state = 422, .external_lex_state = 17},
  [4044] = {.lex_state = 418, .external_lex_state = 10},
  [4045] = {.lex_state = 418, .external_lex_state = 11},
  [4046] = {.lex_state = 444, .external_lex_state = 42},
  [4047] = {.lex_state = 418, .external_lex_state = 10},
  [4048] = {.lex_state = 444, .external_lex_state = 42},
  [4049] = {.lex_state = 418, .external_lex_state = 12},
  [4050] = {.lex_state = 418, .external_lex_state = 11},
  [4051] = {.lex_state = 418, .external_lex_state = 11},
  [4052] = {.lex_state = 418, .external_lex_state = 11},
  [4053] = {.lex_state = 90, .external_lex_state = 9},
  [4054] = {.lex_state = 90, .external_lex_state = 9},
  [4055] = {.lex_state = 418, .external_lex_state = 36},
  [4056] = {.lex_state = 418, .external_lex_state = 11},
  [4057] = {.lex_state = 418, .external_lex_state = 10},
  [4058] = {.lex_state = 431, .external_lex_state = 12},
  [4059] = {.lex_state = 418, .external_lex_state = 11},
  [4060] = {.lex_state = 89, .external_lex_state = 11},
  [4061] = {.lex_state = 418, .external_lex_state = 10},
  [4062] = {.lex_state = 431, .external_lex_state = 12},
  [4063] = {.lex_state = 418, .external_lex_state = 11},
  [4064] = {.lex_state = 418, .external_lex_state = 36},
  [4065] = {.lex_state = 418, .external_lex_state = 11},
  [4066] = {.lex_state = 418, .external_lex_state = 43},
  [4067] = {.lex_state = 422, .external_lex_state = 18},
  [4068] = {.lex_state = 422, .external_lex_state = 18},
  [4069] = {.lex_state = 422, .external_lex_state = 11},
  [4070] = {.lex_state = 422, .external_lex_state = 18},
  [4071] = {.lex_state = 418, .external_lex_state = 10},
  [4072] = {.lex_state = 89, .external_lex_state = 11},
  [4073] = {.lex_state = 418, .external_lex_state = 10},
  [4074] = {.lex_state = 418, .external_lex_state = 11},
  [4075] = {.lex_state = 444, .external_lex_state = 41},
  [4076] = {.lex_state = 418, .external_lex_state = 11},
  [4077] = {.lex_state = 418, .external_lex_state = 43},
  [4078] = {.lex_state = 430, .external_lex_state = 12},
  [4079] = {.lex_state = 430, .external_lex_state = 12},
  [4080] = {.lex_state = 431, .external_lex_state = 12},
  [4081] = {.lex_state = 418, .external_lex_state = 11},
  [4082] = {.lex_state = 418, .external_lex_state = 10},
  [4083] = {.lex_state = 418, .external_lex_state = 11},
  [4084] = {.lex_state = 418, .external_lex_state = 11},
  [4085] = {.lex_state = 90, .external_lex_state = 9},
  [4086] = {.lex_state = 418, .external_lex_state = 10},
  [4087] = {.lex_state = 418, .external_lex_state = 10},
  [4088] = {.lex_state = 418, .external_lex_state = 11},
  [4089] = {.lex_state = 87, .external_lex_state = 13},
  [4090] = {.lex_state = 422, .external_lex_state = 17},
  [4091] = {.lex_state = 418, .external_lex_state = 10},
  [4092] = {.lex_state = 422, .external_lex_state = 14},
  [4093] = {.lex_state = 90, .external_lex_state = 9},
  [4094] = {.lex_state = 418, .external_lex_state = 11},
  [4095] = {.lex_state = 89, .external_lex_state = 11},
  [4096] = {.lex_state = 87, .external_lex_state = 13},
  [4097] = {.lex_state = 418, .external_lex_state = 10},
  [4098] = {.lex_state = 430, .external_lex_state = 12},
  [4099] = {.lex_state = 418, .external_lex_state = 11},
  [4100] = {.lex_state = 418, .external_lex_state = 10},
  [4101] = {.lex_state = 418, .external_lex_state = 11},
  [4102] = {.lex_state = 418, .external_lex_state = 11},
  [4103] = {.lex_state = 422, .external_lex_state = 17},
  [4104] = {.lex_state = 422, .external_lex_state = 14},
  [4105] = {.lex_state = 418, .external_lex_state = 10},
  [4106] = {.lex_state = 417, .external_lex_state = 11},
  [4107] = {.lex_state = 418, .external_lex_state = 10},
  [4108] = {.lex_state = 418, .external_lex_state = 10},
  [4109] = {.lex_state = 418, .external_lex_state = 10},
  [4110] = {.lex_state = 90, .external_lex_state = 9},
  [4111] = {.lex_state = 418, .external_lex_state = 10},
  [4112] = {.lex_state = 418, .external_lex_state = 11},
  [4113] = {.lex_state = 418, .external_lex_state = 11},
  [4114] = {.lex_state = 90, .external_lex_state = 9},
  [4115] = {.lex_state = 418, .external_lex_state = 10},
  [4116] = {.lex_state = 418, .external_lex_state = 10},
  [4117] = {.lex_state = 418, .external_lex_state = 11},
  [4118] = {.lex_state = 418, .external_lex_state = 10},
  [4119] = {.lex_state = 418, .external_lex_state = 10},
  [4120] = {.lex_state = 418, .external_lex_state = 10},
  [4121] = {.lex_state = 418, .external_lex_state = 10},
  [4122] = {.lex_state = 418, .external_lex_state = 10},
  [4123] = {.lex_state = 90, .external_lex_state = 9},
  [4124] = {.lex_state = 418, .external_lex_state = 10},
  [4125] = {.lex_state = 79, .external_lex_state = 11},
  [4126] = {.lex_state = 418, .external_lex_state = 12},
  [4127] = {.lex_state = 78, .external_lex_state = 11},
  [4128] = {.lex_state = 418, .external_lex_state = 12},
  [4129] = {.lex_state = 79, .external_lex_state = 11},
  [4130] = {.lex_state = 79, .external_lex_state = 11},
  [4131] = {.lex_state = 90, .external_lex_state = 14},
  [4132] = {.lex_state = 79, .external_lex_state = 11},
  [4133] = {.lex_state = 422, .external_lex_state = 10},
  [4134] = {.lex_state = 418, .external_lex_state = 12},
  [4135] = {.lex_state = 422, .external_lex_state = 15},
  [4136] = {.lex_state = 422, .external_lex_state = 15},
  [4137] = {.lex_state = 418, .external_lex_state = 12},
  [4138] = {.lex_state = 418, .external_lex_state = 12},
  [4139] = {.lex_state = 418, .external_lex_state = 12},
  [4140] = {.lex_state = 418, .external_lex_state = 12},
  [4141] = {.lex_state = 422, .external_lex_state = 15},
  [4142] = {.lex_state = 418, .external_lex_state = 12},
  [4143] = {.lex_state = 418, .external_lex_state = 12},
  [4144] = {.lex_state = 418, .external_lex_state = 12},
  [4145] = {.lex_state = 90, .external_lex_state = 14},
  [4146] = {.lex_state = 418, .external_lex_state = 12},
  [4147] = {.lex_state = 418, .external_lex_state = 12},
  [4148] = {.lex_state = 90, .external_lex_state = 14},
  [4149] = {.lex_state = 79, .external_lex_state = 11},
  [4150] = {.lex_state = 422, .external_lex_state = 10},
  [4151] = {.lex_state = 422, .external_lex_state = 10},
  [4152] = {.lex_state = 418, .external_lex_state = 12},
  [4153] = {.lex_state = 418, .external_lex_state = 12},
  [4154] = {.lex_state = 418, .external_lex_state = 12},
  [4155] = {.lex_state = 418, .external_lex_state = 12},
  [4156] = {.lex_state = 418, .external_lex_state = 12},
  [4157] = {.lex_state = 418, .external_lex_state = 12},
  [4158] = {.lex_state = 418, .external_lex_state = 12},
  [4159] = {.lex_state = 79, .external_lex_state = 11},
  [4160] = {.lex_state = 418, .external_lex_state = 12},
  [4161] = {.lex_state = 418, .external_lex_state = 12},
  [4162] = {.lex_state = 418, .external_lex_state = 12},
  [4163] = {.lex_state = 418, .external_lex_state = 12},
  [4164] = {.lex_state = 418, .external_lex_state = 12},
  [4165] = {.lex_state = 418, .external_lex_state = 12},
  [4166] = {.lex_state = 418, .external_lex_state = 12},
  [4167] = {.lex_state = 418, .external_lex_state = 12},
  [4168] = {.lex_state = 418, .external_lex_state = 12},
  [4169] = {.lex_state = 422, .external_lex_state = 15},
  [4170] = {.lex_state = 422, .external_lex_state = 15},
  [4171] = {.lex_state = 418, .external_lex_state = 12},
  [4172] = {.lex_state = 418, .external_lex_state = 12},
  [4173] = {.lex_state = 418, .external_lex_state = 12},
  [4174] = {.lex_state = 418, .external_lex_state = 12},
  [4175] = {.lex_state = 418, .external_lex_state = 12},
  [4176] = {.lex_state = 418, .external_lex_state = 12},
  [4177] = {.lex_state = 418, .external_lex_state = 12},
  [4178] = {.lex_state = 418, .external_lex_state = 12},
  [4179] = {.lex_state = 88, .external_lex_state = 13},
  [4180] = {.lex_state = 88, .external_lex_state = 13},
  [4181] = {.lex_state = 418, .external_lex_state = 12},
  [4182] = {.lex_state = 418, .external_lex_state = 12},
  [4183] = {.lex_state = 79, .external_lex_state = 11},
  [4184] = {.lex_state = 79, .external_lex_state = 11},
  [4185] = {.lex_state = 418, .external_lex_state = 12},
  [4186] = {.lex_state = 87, .external_lex_state = 13},
  [4187] = {.lex_state = 418, .external_lex_state = 12},
  [4188] = {.lex_state = 87, .external_lex_state = 13},
  [4189] = {.lex_state = 87, .external_lex_state = 13},
  [4190] = {.lex_state = 422, .external_lex_state = 13},
  [4191] = {.lex_state = 422, .external_lex_state = 13},
  [4192] = {.lex_state = 422, .external_lex_state = 13},
  [4193] = {.lex_state = 444, .external_lex_state = 44},
  [4194] = {.lex_state = 420, .external_lex_state = 11},
  [4195] = {.lex_state = 422, .external_lex_state = 13},
  [4196] = {.lex_state = 422, .external_lex_state = 13},
  [4197] = {.lex_state = 422, .external_lex_state = 13},
  [4198] = {.lex_state = 444, .external_lex_state = 44},
  [4199] = {.lex_state = 418, .external_lex_state = 43},
  [4200] = {.lex_state = 422, .external_lex_state = 13},
  [4201] = {.lex_state = 422, .external_lex_state = 13},
  [4202] = {.lex_state = 418, .external_lex_state = 12},
  [4203] = {.lex_state = 418, .external_lex_state = 12},
  [4204] = {.lex_state = 87, .external_lex_state = 13},
  [4205] = {.lex_state = 87, .external_lex_state = 13},
  [4206] = {.lex_state = 79, .external_lex_state = 11},
  [4207] = {.lex_state = 108, .external_lex_state = 45},
  [4208] = {.lex_state = 87, .external_lex_state = 13},
  [4209] = {.lex_state = 108, .external_lex_state = 45},
  [4210] = {.lex_state = 88, .external_lex_state = 11},
  [4211] = {.lex_state = 79, .external_lex_state = 11},
  [4212] = {.lex_state = 418, .external_lex_state = 37},
  [4213] = {.lex_state = 418, .external_lex_state = 37},
  [4214] = {.lex_state = 444, .external_lex_state = 42},
  [4215] = {.lex_state = 79, .external_lex_state = 11},
  [4216] = {.lex_state = 418, .external_lex_state = 11},
  [4217] = {.lex_state = 79, .external_lex_state = 11},
  [4218] = {.lex_state = 422, .external_lex_state = 10},
  [4219] = {.lex_state = 444, .external_lex_state = 46},
  [4220] = {.lex_state = 420, .external_lex_state = 11},
  [4221] = {.lex_state = 418, .external_lex_state = 11},
  [4222] = {.lex_state = 444, .external_lex_state = 46},
  [4223] = {.lex_state = 418, .external_lex_state = 11},
  [4224] = {.lex_state = 79, .external_lex_state = 11},
  [4225] = {.lex_state = 88, .external_lex_state = 13},
  [4226] = {.lex_state = 102, .external_lex_state = 9},
  [4227] = {.lex_state = 420, .external_lex_state = 43},
  [4228] = {.lex_state = 420, .external_lex_state = 43},
  [4229] = {.lex_state = 418, .external_lex_state = 47},
  [4230] = {.lex_state = 418, .external_lex_state = 47},
  [4231] = {.lex_state = 418, .external_lex_state = 12},
  [4232] = {.lex_state = 418, .external_lex_state = 12},
  [4233] = {.lex_state = 418, .external_lex_state = 12},
  [4234] = {.lex_state = 418, .external_lex_state = 12},
  [4235] = {.lex_state = 418, .external_lex_state = 12},
  [4236] = {.lex_state = 418, .external_lex_state = 12},
  [4237] = {.lex_state = 418, .external_lex_state = 12},
  [4238] = {.lex_state = 418, .external_lex_state = 12},
  [4239] = {.lex_state = 418, .external_lex_state = 12},
  [4240] = {.lex_state = 418, .external_lex_state = 12},
  [4241] = {.lex_state = 418, .external_lex_state = 12},
  [4242] = {.lex_state = 418, .external_lex_state = 12},
  [4243] = {.lex_state = 418, .external_lex_state = 12},
  [4244] = {.lex_state = 418, .external_lex_state = 12},
  [4245] = {.lex_state = 418, .external_lex_state = 12},
  [4246] = {.lex_state = 418, .external_lex_state = 12},
  [4247] = {.lex_state = 418, .external_lex_state = 12},
  [4248] = {.lex_state = 418, .external_lex_state = 12},
  [4249] = {.lex_state = 418, .external_lex_state = 12},
  [4250] = {.lex_state = 418, .external_lex_state = 12},
  [4251] = {.lex_state = 418, .external_lex_state = 12},
  [4252] = {.lex_state = 418, .external_lex_state = 12},
  [4253] = {.lex_state = 418, .external_lex_state = 12},
  [4254] = {.lex_state = 418, .external_lex_state = 12},
  [4255] = {.lex_state = 418, .external_lex_state = 12},
  [4256] = {.lex_state = 418, .external_lex_state = 12},
  [4257] = {.lex_state = 418, .external_lex_state = 12},
  [4258] = {.lex_state = 418, .external_lex_state = 12},
  [4259] = {.lex_state = 90, .external_lex_state = 14},
  [4260] = {.lex_state = 444, .external_lex_state = 42},
  [4261] = {.lex_state = 90, .external_lex_state = 14},
  [4262] = {.lex_state = 422, .external_lex_state = 11},
  [4263] = {.lex_state = 109, .external_lex_state = 36},
  [4264] = {.lex_state = 90, .external_lex_state = 14},
  [4265] = {.lex_state = 418, .external_lex_state = 12},
  [4266] = {.lex_state = 418, .external_lex_state = 12},
  [4267] = {.lex_state = 418, .external_lex_state = 12},
  [4268] = {.lex_state = 109, .external_lex_state = 36},
  [4269] = {.lex_state = 422, .external_lex_state = 10},
  [4270] = {.lex_state = 422, .external_lex_state = 10},
  [4271] = {.lex_state = 422, .external_lex_state = 10},
  [4272] = {.lex_state = 78, .external_lex_state = 11},
  [4273] = {.lex_state = 87, .external_lex_state = 13},
  [4274] = {.lex_state = 87, .external_lex_state = 13},
  [4275] = {.lex_state = 87, .external_lex_state = 13},
  [4276] = {.lex_state = 87, .external_lex_state = 13},
  [4277] = {.lex_state = 88, .external_lex_state = 11},
  [4278] = {.lex_state = 418, .external_lex_state = 12},
  [4279] = {.lex_state = 87, .external_lex_state = 13},
  [4280] = {.lex_state = 87, .external_lex_state = 13},
  [4281] = {.lex_state = 87, .external_lex_state = 13},
  [4282] = {.lex_state = 87, .external_lex_state = 13},
  [4283] = {.lex_state = 88, .external_lex_state = 11},
  [4284] = {.lex_state = 422, .external_lex_state = 15},
  [4285] = {.lex_state = 78, .external_lex_state = 11},
  [4286] = {.lex_state = 90, .external_lex_state = 14},
  [4287] = {.lex_state = 78, .external_lex_state = 11},
  [4288] = {.lex_state = 422, .external_lex_state = 15},
  [4289] = {.lex_state = 90, .external_lex_state = 14},
  [4290] = {.lex_state = 422, .external_lex_state = 15},
  [4291] = {.lex_state = 87, .external_lex_state = 13},
  [4292] = {.lex_state = 87, .external_lex_state = 13},
  [4293] = {.lex_state = 87, .external_lex_state = 13},
  [4294] = {.lex_state = 87, .external_lex_state = 13},
  [4295] = {.lex_state = 78, .external_lex_state = 11},
  [4296] = {.lex_state = 422, .external_lex_state = 10},
  [4297] = {.lex_state = 87, .external_lex_state = 13},
  [4298] = {.lex_state = 79, .external_lex_state = 11},
  [4299] = {.lex_state = 422, .external_lex_state = 15},
  [4300] = {.lex_state = 422, .external_lex_state = 10},
  [4301] = {.lex_state = 422, .external_lex_state = 11},
  [4302] = {.lex_state = 422, .external_lex_state = 11},
  [4303] = {.lex_state = 422, .external_lex_state = 18},
  [4304] = {.lex_state = 422, .external_lex_state = 13},
  [4305] = {.lex_state = 422, .external_lex_state = 15},
  [4306] = {.lex_state = 422, .external_lex_state = 13},
  [4307] = {.lex_state = 422, .external_lex_state = 13},
  [4308] = {.lex_state = 422, .external_lex_state = 13},
  [4309] = {.lex_state = 422, .external_lex_state = 13},
  [4310] = {.lex_state = 422, .external_lex_state = 13},
  [4311] = {.lex_state = 422, .external_lex_state = 17},
  [4312] = {.lex_state = 422, .external_lex_state = 13},
  [4313] = {.lex_state = 422, .external_lex_state = 13},
  [4314] = {.lex_state = 422, .external_lex_state = 13},
  [4315] = {.lex_state = 422, .external_lex_state = 13},
  [4316] = {.lex_state = 422, .external_lex_state = 13},
  [4317] = {.lex_state = 418, .external_lex_state = 11},
  [4318] = {.lex_state = 422, .external_lex_state = 13},
  [4319] = {.lex_state = 422, .external_lex_state = 13},
  [4320] = {.lex_state = 422, .external_lex_state = 13},
  [4321] = {.lex_state = 422, .external_lex_state = 13},
  [4322] = {.lex_state = 113, .external_lex_state = 9},
  [4323] = {.lex_state = 422, .external_lex_state = 18},
  [4324] = {.lex_state = 422, .external_lex_state = 11},
  [4325] = {.lex_state = 422, .external_lex_state = 11},
  [4326] = {.lex_state = 422, .external_lex_state = 13},
  [4327] = {.lex_state = 422, .external_lex_state = 13},
  [4328] = {.lex_state = 422, .external_lex_state = 13},
  [4329] = {.lex_state = 422, .external_lex_state = 11},
  [4330] = {.lex_state = 113, .external_lex_state = 9},
  [4331] = {.lex_state = 422, .external_lex_state = 19},
  [4332] = {.lex_state = 106, .external_lex_state = 9},
  [4333] = {.lex_state = 422, .external_lex_state = 11},
  [4334] = {.lex_state = 422, .external_lex_state = 11},
  [4335] = {.lex_state = 422, .external_lex_state = 13},
  [4336] = {.lex_state = 422, .external_lex_state = 19},
  [4337] = {.lex_state = 422, .external_lex_state = 12},
  [4338] = {.lex_state = 418, .external_lex_state = 11},
  [4339] = {.lex_state = 422, .external_lex_state = 14},
  [4340] = {.lex_state = 422, .external_lex_state = 12},
  [4341] = {.lex_state = 418, .external_lex_state = 48},
  [4342] = {.lex_state = 88, .external_lex_state = 11},
  [4343] = {.lex_state = 422, .external_lex_state = 12},
  [4344] = {.lex_state = 418, .external_lex_state = 49},
  [4345] = {.lex_state = 418, .external_lex_state = 11},
  [4346] = {.lex_state = 422, .external_lex_state = 12},
  [4347] = {.lex_state = 444, .external_lex_state = 50},
  [4348] = {.lex_state = 108, .external_lex_state = 45},
  [4349] = {.lex_state = 444, .external_lex_state = 50},
  [4350] = {.lex_state = 418, .external_lex_state = 11},
  [4351] = {.lex_state = 422, .external_lex_state = 19},
  [4352] = {.lex_state = 418, .external_lex_state = 11},
  [4353] = {.lex_state = 418, .external_lex_state = 11},
  [4354] = {.lex_state = 416, .external_lex_state = 13},
  [4355] = {.lex_state = 113, .external_lex_state = 9},
  [4356] = {.lex_state = 418, .external_lex_state = 11},
  [4357] = {.lex_state = 110, .external_lex_state = 9},
  [4358] = {.lex_state = 418, .external_lex_state = 37},
  [4359] = {.lex_state = 418, .external_lex_state = 11},
  [4360] = {.lex_state = 418, .external_lex_state = 11},
  [4361] = {.lex_state = 418, .external_lex_state = 11},
  [4362] = {.lex_state = 418, .external_lex_state = 11},
  [4363] = {.lex_state = 418, .external_lex_state = 11},
  [4364] = {.lex_state = 418, .external_lex_state = 11},
  [4365] = {.lex_state = 418, .external_lex_state = 11},
  [4366] = {.lex_state = 110, .external_lex_state = 9},
  [4367] = {.lex_state = 418, .external_lex_state = 11},
  [4368] = {.lex_state = 90, .external_lex_state = 14},
  [4369] = {.lex_state = 418, .external_lex_state = 11},
  [4370] = {.lex_state = 418, .external_lex_state = 11},
  [4371] = {.lex_state = 418, .external_lex_state = 11},
  [4372] = {.lex_state = 418, .external_lex_state = 11},
  [4373] = {.lex_state = 418, .external_lex_state = 11},
  [4374] = {.lex_state = 418, .external_lex_state = 11},
  [4375] = {.lex_state = 422, .external_lex_state = 12},
  [4376] = {.lex_state = 418, .external_lex_state = 11},
  [4377] = {.lex_state = 418, .external_lex_state = 11},
  [4378] = {.lex_state = 418, .external_lex_state = 11},
  [4379] = {.lex_state = 418, .external_lex_state = 11},
  [4380] = {.lex_state = 418, .external_lex_state = 11},
  [4381] = {.lex_state = 418, .external_lex_state = 11},
  [4382] = {.lex_state = 418, .external_lex_state = 11},
  [4383] = {.lex_state = 418, .external_lex_state = 11},
  [4384] = {.lex_state = 418, .external_lex_state = 11},
  [4385] = {.lex_state = 418, .external_lex_state = 11},
  [4386] = {.lex_state = 418, .external_lex_state = 11},
  [4387] = {.lex_state = 418, .external_lex_state = 11},
  [4388] = {.lex_state = 422, .external_lex_state = 10},
  [4389] = {.lex_state = 418, .external_lex_state = 11},
  [4390] = {.lex_state = 418, .external_lex_state = 11},
  [4391] = {.lex_state = 418, .external_lex_state = 11},
  [4392] = {.lex_state = 418, .external_lex_state = 11},
  [4393] = {.lex_state = 418, .external_lex_state = 11},
  [4394] = {.lex_state = 418, .external_lex_state = 11},
  [4395] = {.lex_state = 418, .external_lex_state = 11},
  [4396] = {.lex_state = 418, .external_lex_state = 11},
  [4397] = {.lex_state = 418, .external_lex_state = 11},
  [4398] = {.lex_state = 422, .external_lex_state = 17},
  [4399] = {.lex_state = 418, .external_lex_state = 11},
  [4400] = {.lex_state = 418, .external_lex_state = 11},
  [4401] = {.lex_state = 418, .external_lex_state = 11},
  [4402] = {.lex_state = 86, .external_lex_state = 36},
  [4403] = {.lex_state = 422, .external_lex_state = 17},
  [4404] = {.lex_state = 420, .external_lex_state = 36},
  [4405] = {.lex_state = 86, .external_lex_state = 36},
  [4406] = {.lex_state = 420, .external_lex_state = 36},
  [4407] = {.lex_state = 418, .external_lex_state = 11},
  [4408] = {.lex_state = 418, .external_lex_state = 11},
  [4409] = {.lex_state = 418, .external_lex_state = 11},
  [4410] = {.lex_state = 418, .external_lex_state = 11},
  [4411] = {.lex_state = 103, .external_lex_state = 9},
  [4412] = {.lex_state = 103, .external_lex_state = 9},
  [4413] = {.lex_state = 418, .external_lex_state = 11},
  [4414] = {.lex_state = 422, .external_lex_state = 17},
  [4415] = {.lex_state = 103, .external_lex_state = 9},
  [4416] = {.lex_state = 418, .external_lex_state = 11},
  [4417] = {.lex_state = 418, .external_lex_state = 11},
  [4418] = {.lex_state = 418, .external_lex_state = 11},
  [4419] = {.lex_state = 418, .external_lex_state = 11},
  [4420] = {.lex_state = 103, .external_lex_state = 9},
  [4421] = {.lex_state = 420, .external_lex_state = 43},
  [4422] = {.lex_state = 422, .external_lex_state = 14},
  [4423] = {.lex_state = 418, .external_lex_state = 11},
  [4424] = {.lex_state = 418, .external_lex_state = 11},
  [4425] = {.lex_state = 418, .external_lex_state = 11},
  [4426] = {.lex_state = 418, .external_lex_state = 11},
  [4427] = {.lex_state = 418, .external_lex_state = 11},
  [4428] = {.lex_state = 418, .external_lex_state = 11},
  [4429] = {.lex_state = 418, .external_lex_state = 11},
  [4430] = {.lex_state = 422, .external_lex_state = 18},
  [4431] = {.lex_state = 422, .external_lex_state = 14},
  [4432] = {.lex_state = 422, .external_lex_state = 17},
  [4433] = {.lex_state = 422, .external_lex_state = 17},
  [4434] = {.lex_state = 422, .external_lex_state = 18},
  [4435] = {.lex_state = 422, .external_lex_state = 18},
  [4436] = {.lex_state = 422, .external_lex_state = 14},
  [4437] = {.lex_state = 418, .external_lex_state = 11},
  [4438] = {.lex_state = 422, .external_lex_state = 18},
  [4439] = {.lex_state = 444, .external_lex_state = 51},
  [4440] = {.lex_state = 418, .external_lex_state = 48},
  [4441] = {.lex_state = 418, .external_lex_state = 47},
  [4442] = {.lex_state = 444, .external_lex_state = 51},
  [4443] = {.lex_state = 422, .external_lex_state = 17},
  [4444] = {.lex_state = 110, .external_lex_state = 9},
  [4445] = {.lex_state = 422, .external_lex_state = 15},
  [4446] = {.lex_state = 110, .external_lex_state = 9},
  [4447] = {.lex_state = 89, .external_lex_state = 11},
  [4448] = {.lex_state = 89, .external_lex_state = 11},
  [4449] = {.lex_state = 422, .external_lex_state = 15},
  [4450] = {.lex_state = 418, .external_lex_state = 11},
  [4451] = {.lex_state = 422, .external_lex_state = 15},
  [4452] = {.lex_state = 110, .external_lex_state = 9},
  [4453] = {.lex_state = 422, .external_lex_state = 12},
  [4454] = {.lex_state = 422, .external_lex_state = 12},
  [4455] = {.lex_state = 418, .external_lex_state = 49},
  [4456] = {.lex_state = 90, .external_lex_state = 14},
  [4457] = {.lex_state = 89, .external_lex_state = 11},
  [4458] = {.lex_state = 90, .external_lex_state = 14},
  [4459] = {.lex_state = 90, .external_lex_state = 14},
  [4460] = {.lex_state = 418, .external_lex_state = 11},
  [4461] = {.lex_state = 418, .external_lex_state = 11},
  [4462] = {.lex_state = 90, .external_lex_state = 14},
  [4463] = {.lex_state = 90, .external_lex_state = 14},
  [4464] = {.lex_state = 90, .external_lex_state = 14},
  [4465] = {.lex_state = 90, .external_lex_state = 14},
  [4466] = {.lex_state = 418, .external_lex_state = 11},
  [4467] = {.lex_state = 89, .external_lex_state = 11},
  [4468] = {.lex_state = 418, .external_lex_state = 11},
  [4469] = {.lex_state = 90, .external_lex_state = 14},
  [4470] = {.lex_state = 418, .external_lex_state = 52},
  [4471] = {.lex_state = 90, .external_lex_state = 14},
  [4472] = {.lex_state = 90, .external_lex_state = 14},
  [4473] = {.lex_state = 90, .external_lex_state = 14},
  [4474] = {.lex_state = 90, .external_lex_state = 14},
  [4475] = {.lex_state = 88, .external_lex_state = 13},
  [4476] = {.lex_state = 422, .external_lex_state = 10},
  [4477] = {.lex_state = 422, .external_lex_state = 10},
  [4478] = {.lex_state = 422, .external_lex_state = 10},
  [4479] = {.lex_state = 422, .external_lex_state = 14},
  [4480] = {.lex_state = 422, .external_lex_state = 10},
  [4481] = {.lex_state = 422, .external_lex_state = 10},
  [4482] = {.lex_state = 422, .external_lex_state = 10},
  [4483] = {.lex_state = 422, .external_lex_state = 10},
  [4484] = {.lex_state = 422, .external_lex_state = 14},
  [4485] = {.lex_state = 422, .external_lex_state = 10},
  [4486] = {.lex_state = 422, .external_lex_state = 10},
  [4487] = {.lex_state = 422, .external_lex_state = 10},
  [4488] = {.lex_state = 422, .external_lex_state = 10},
  [4489] = {.lex_state = 422, .external_lex_state = 10},
  [4490] = {.lex_state = 109, .external_lex_state = 36},
  [4491] = {.lex_state = 88, .external_lex_state = 13},
  [4492] = {.lex_state = 422, .external_lex_state = 12},
  [4493] = {.lex_state = 90, .external_lex_state = 14},
  [4494] = {.lex_state = 90, .external_lex_state = 14},
  [4495] = {.lex_state = 90, .external_lex_state = 14},
  [4496] = {.lex_state = 112, .external_lex_state = 9},
  [4497] = {.lex_state = 422, .external_lex_state = 17},
  [4498] = {.lex_state = 422, .external_lex_state = 19},
  [4499] = {.lex_state = 418, .external_lex_state = 45},
  [4500] = {.lex_state = 418, .external_lex_state = 11},
  [4501] = {.lex_state = 112, .external_lex_state = 9},
  [4502] = {.lex_state = 422, .external_lex_state = 19},
  [4503] = {.lex_state = 89, .external_lex_state = 11},
  [4504] = {.lex_state = 422, .external_lex_state = 19},
  [4505] = {.lex_state = 418, .external_lex_state = 38},
  [4506] = {.lex_state = 89, .external_lex_state = 11},
  [4507] = {.lex_state = 422, .external_lex_state = 19},
  [4508] = {.lex_state = 422, .external_lex_state = 19},
  [4509] = {.lex_state = 89, .external_lex_state = 11},
  [4510] = {.lex_state = 422, .external_lex_state = 10},
  [4511] = {.lex_state = 422, .external_lex_state = 10},
  [4512] = {.lex_state = 422, .external_lex_state = 10},
  [4513] = {.lex_state = 418, .external_lex_state = 38},
  [4514] = {.lex_state = 422, .external_lex_state = 18},
  [4515] = {.lex_state = 422, .external_lex_state = 14},
  [4516] = {.lex_state = 418, .external_lex_state = 45},
  [4517] = {.lex_state = 89, .external_lex_state = 11},
  [4518] = {.lex_state = 422, .external_lex_state = 14},
  [4519] = {.lex_state = 418, .external_lex_state = 52},
  [4520] = {.lex_state = 112, .external_lex_state = 9},
  [4521] = {.lex_state = 88, .external_lex_state = 11},
  [4522] = {.lex_state = 108, .external_lex_state = 34},
  [4523] = {.lex_state = 422, .external_lex_state = 15},
  [4524] = {.lex_state = 90, .external_lex_state = 14},
  [4525] = {.lex_state = 422, .external_lex_state = 15},
  [4526] = {.lex_state = 422, .external_lex_state = 15},
  [4527] = {.lex_state = 422, .external_lex_state = 15},
  [4528] = {.lex_state = 422, .external_lex_state = 18},
  [4529] = {.lex_state = 90, .external_lex_state = 14},
  [4530] = {.lex_state = 422, .external_lex_state = 15},
  [4531] = {.lex_state = 422, .external_lex_state = 15},
  [4532] = {.lex_state = 422, .external_lex_state = 15},
  [4533] = {.lex_state = 422, .external_lex_state = 15},
  [4534] = {.lex_state = 90, .external_lex_state = 14},
  [4535] = {.lex_state = 422, .external_lex_state = 15},
  [4536] = {.lex_state = 422, .external_lex_state = 15},
  [4537] = {.lex_state = 418, .external_lex_state = 11},
  [4538] = {.lex_state = 422, .external_lex_state = 15},
  [4539] = {.lex_state = 422, .external_lex_state = 15},
  [4540] = {.lex_state = 422, .external_lex_state = 15},
  [4541] = {.lex_state = 422, .external_lex_state = 15},
  [4542] = {.lex_state = 444, .external_lex_state = 53},
  [4543] = {.lex_state = 444, .external_lex_state = 53},
  [4544] = {.lex_state = 108, .external_lex_state = 34},
  [4545] = {.lex_state = 422, .external_lex_state = 10},
  [4546] = {.lex_state = 422, .external_lex_state = 11},
  [4547] = {.lex_state = 422, .external_lex_state = 10},
  [4548] = {.lex_state = 108, .external_lex_state = 34},
  [4549] = {.lex_state = 418, .external_lex_state = 11},
  [4550] = {.lex_state = 422, .external_lex_state = 19},
  [4551] = {.lex_state = 422, .external_lex_state = 11},
  [4552] = {.lex_state = 422, .external_lex_state = 17},
  [4553] = {.lex_state = 422, .external_lex_state = 17},
  [4554] = {.lex_state = 422, .external_lex_state = 14},
  [4555] = {.lex_state = 422, .external_lex_state = 12},
  [4556] = {.lex_state = 418, .external_lex_state = 43},
  [4557] = {.lex_state = 110, .external_lex_state = 9},
  [4558] = {.lex_state = 110, .external_lex_state = 9},
  [4559] = {.lex_state = 93, .external_lex_state = 13},
  [4560] = {.lex_state = 422, .external_lex_state = 19},
  [4561] = {.lex_state = 79, .external_lex_state = 13},
  [4562] = {.lex_state = 102, .external_lex_state = 14},
  [4563] = {.lex_state = 108, .external_lex_state = 9},
  [4564] = {.lex_state = 422, .external_lex_state = 19},
  [4565] = {.lex_state = 89, .external_lex_state = 11},
  [4566] = {.lex_state = 422, .external_lex_state = 11},
  [4567] = {.lex_state = 422, .external_lex_state = 18},
  [4568] = {.lex_state = 422, .external_lex_state = 17},
  [4569] = {.lex_state = 107, .external_lex_state = 9},
  [4570] = {.lex_state = 422, .external_lex_state = 12},
  [4571] = {.lex_state = 422, .external_lex_state = 12},
  [4572] = {.lex_state = 110, .external_lex_state = 9},
  [4573] = {.lex_state = 418, .external_lex_state = 52},
  [4574] = {.lex_state = 78, .external_lex_state = 13},
  [4575] = {.lex_state = 422, .external_lex_state = 12},
  [4576] = {.lex_state = 113, .external_lex_state = 9},
  [4577] = {.lex_state = 89, .external_lex_state = 11},
  [4578] = {.lex_state = 422, .external_lex_state = 14},
  [4579] = {.lex_state = 89, .external_lex_state = 11},
  [4580] = {.lex_state = 422, .external_lex_state = 19},
  [4581] = {.lex_state = 422, .external_lex_state = 12},
  [4582] = {.lex_state = 422, .external_lex_state = 18},
  [4583] = {.lex_state = 422, .external_lex_state = 17},
  [4584] = {.lex_state = 422, .external_lex_state = 18},
  [4585] = {.lex_state = 422, .external_lex_state = 12},
  [4586] = {.lex_state = 106, .external_lex_state = 13},
  [4587] = {.lex_state = 422, .external_lex_state = 17},
  [4588] = {.lex_state = 422, .external_lex_state = 11},
  [4589] = {.lex_state = 422, .external_lex_state = 12},
  [4590] = {.lex_state = 422, .external_lex_state = 11},
  [4591] = {.lex_state = 444, .external_lex_state = 53},
  [4592] = {.lex_state = 89, .external_lex_state = 11},
  [4593] = {.lex_state = 422, .external_lex_state = 18},
  [4594] = {.lex_state = 422, .external_lex_state = 18},
  [4595] = {.lex_state = 422, .external_lex_state = 11},
  [4596] = {.lex_state = 113, .external_lex_state = 9},
  [4597] = {.lex_state = 444, .external_lex_state = 53},
  [4598] = {.lex_state = 422, .external_lex_state = 18},
  [4599] = {.lex_state = 422, .external_lex_state = 19},
  [4600] = {.lex_state = 422, .external_lex_state = 18},
  [4601] = {.lex_state = 78, .external_lex_state = 13},
  [4602] = {.lex_state = 104, .external_lex_state = 11},
  [4603] = {.lex_state = 422, .external_lex_state = 12},
  [4604] = {.lex_state = 78, .external_lex_state = 13},
  [4605] = {.lex_state = 422, .external_lex_state = 12},
  [4606] = {.lex_state = 422, .external_lex_state = 17},
  [4607] = {.lex_state = 422, .external_lex_state = 12},
  [4608] = {.lex_state = 422, .external_lex_state = 12},
  [4609] = {.lex_state = 112, .external_lex_state = 9},
  [4610] = {.lex_state = 422, .external_lex_state = 17},
  [4611] = {.lex_state = 422, .external_lex_state = 11},
  [4612] = {.lex_state = 422, .external_lex_state = 17},
  [4613] = {.lex_state = 78, .external_lex_state = 13},
  [4614] = {.lex_state = 78, .external_lex_state = 13},
  [4615] = {.lex_state = 108, .external_lex_state = 9},
  [4616] = {.lex_state = 422, .external_lex_state = 17},
  [4617] = {.lex_state = 110, .external_lex_state = 9},
  [4618] = {.lex_state = 422, .external_lex_state = 18},
  [4619] = {.lex_state = 422, .external_lex_state = 14},
  [4620] = {.lex_state = 422, .external_lex_state = 12},
  [4621] = {.lex_state = 422, .external_lex_state = 17},
  [4622] = {.lex_state = 422, .external_lex_state = 14},
  [4623] = {.lex_state = 93, .external_lex_state = 13},
  [4624] = {.lex_state = 422, .external_lex_state = 12},
  [4625] = {.lex_state = 113, .external_lex_state = 9},
  [4626] = {.lex_state = 422, .external_lex_state = 11},
  [4627] = {.lex_state = 444, .external_lex_state = 54},
  [4628] = {.lex_state = 418, .external_lex_state = 48},
  [4629] = {.lex_state = 418, .external_lex_state = 43},
  [4630] = {.lex_state = 113, .external_lex_state = 9},
  [4631] = {.lex_state = 420, .external_lex_state = 36},
  [4632] = {.lex_state = 110, .external_lex_state = 9},
  [4633] = {.lex_state = 417, .external_lex_state = 13},
  [4634] = {.lex_state = 86, .external_lex_state = 36},
  [4635] = {.lex_state = 422, .external_lex_state = 19},
  [4636] = {.lex_state = 444, .external_lex_state = 54},
  [4637] = {.lex_state = 422, .external_lex_state = 19},
  [4638] = {.lex_state = 422, .external_lex_state = 19},
  [4639] = {.lex_state = 79, .external_lex_state = 13},
  [4640] = {.lex_state = 113, .external_lex_state = 9},
  [4641] = {.lex_state = 89, .external_lex_state = 11},
  [4642] = {.lex_state = 422, .external_lex_state = 19},
  [4643] = {.lex_state = 113, .external_lex_state = 9},
  [4644] = {.lex_state = 422, .external_lex_state = 19},
  [4645] = {.lex_state = 422, .external_lex_state = 19},
  [4646] = {.lex_state = 112, .external_lex_state = 9},
  [4647] = {.lex_state = 417, .external_lex_state = 13},
  [4648] = {.lex_state = 113, .external_lex_state = 9},
  [4649] = {.lex_state = 422, .external_lex_state = 19},
  [4650] = {.lex_state = 89, .external_lex_state = 11},
  [4651] = {.lex_state = 418, .external_lex_state = 43},
  [4652] = {.lex_state = 89, .external_lex_state = 11},
  [4653] = {.lex_state = 89, .external_lex_state = 11},
  [4654] = {.lex_state = 79, .external_lex_state = 13},
  [4655] = {.lex_state = 113, .external_lex_state = 9},
  [4656] = {.lex_state = 107, .external_lex_state = 9},
  [4657] = {.lex_state = 422, .external_lex_state = 11},
  [4658] = {.lex_state = 422, .external_lex_state = 12},
  [4659] = {.lex_state = 89, .external_lex_state = 11},
  [4660] = {.lex_state = 89, .external_lex_state = 11},
  [4661] = {.lex_state = 422, .external_lex_state = 17},
  [4662] = {.lex_state = 417, .external_lex_state = 13},
  [4663] = {.lex_state = 422, .external_lex_state = 19},
  [4664] = {.lex_state = 422, .external_lex_state = 19},
  [4665] = {.lex_state = 422, .external_lex_state = 19},
  [4666] = {.lex_state = 89, .external_lex_state = 11},
  [4667] = {.lex_state = 108, .external_lex_state = 9},
  [4668] = {.lex_state = 89, .external_lex_state = 11},
  [4669] = {.lex_state = 422, .external_lex_state = 11},
  [4670] = {.lex_state = 89, .external_lex_state = 11},
  [4671] = {.lex_state = 422, .external_lex_state = 12},
  [4672] = {.lex_state = 89, .external_lex_state = 11},
  [4673] = {.lex_state = 422, .external_lex_state = 18},
  [4674] = {.lex_state = 89, .external_lex_state = 11},
  [4675] = {.lex_state = 422, .external_lex_state = 14},
  [4676] = {.lex_state = 422, .external_lex_state = 19},
  [4677] = {.lex_state = 418, .external_lex_state = 38},
  [4678] = {.lex_state = 422, .external_lex_state = 19},
  [4679] = {.lex_state = 109, .external_lex_state = 9},
  [4680] = {.lex_state = 422, .external_lex_state = 14},
  [4681] = {.lex_state = 422, .external_lex_state = 19},
  [4682] = {.lex_state = 113, .external_lex_state = 9},
  [4683] = {.lex_state = 422, .external_lex_state = 17},
  [4684] = {.lex_state = 422, .external_lex_state = 14},
  [4685] = {.lex_state = 418, .external_lex_state = 49},
  [4686] = {.lex_state = 422, .external_lex_state = 17},
  [4687] = {.lex_state = 422, .external_lex_state = 14},
  [4688] = {.lex_state = 417, .external_lex_state = 13},
  [4689] = {.lex_state = 422, .external_lex_state = 17},
  [4690] = {.lex_state = 108, .external_lex_state = 9},
  [4691] = {.lex_state = 422, .external_lex_state = 14},
  [4692] = {.lex_state = 113, .external_lex_state = 9},
  [4693] = {.lex_state = 422, .external_lex_state = 17},
  [4694] = {.lex_state = 422, .external_lex_state = 17},
  [4695] = {.lex_state = 422, .external_lex_state = 18},
  [4696] = {.lex_state = 109, .external_lex_state = 9},
  [4697] = {.lex_state = 107, .external_lex_state = 9},
  [4698] = {.lex_state = 422, .external_lex_state = 12},
  [4699] = {.lex_state = 89, .external_lex_state = 11},
  [4700] = {.lex_state = 89, .external_lex_state = 11},
  [4701] = {.lex_state = 89, .external_lex_state = 11},
  [4702] = {.lex_state = 422, .external_lex_state = 11},
  [4703] = {.lex_state = 422, .external_lex_state = 18},
  [4704] = {.lex_state = 422, .external_lex_state = 18},
  [4705] = {.lex_state = 422, .external_lex_state = 18},
  [4706] = {.lex_state = 422, .external_lex_state = 14},
  [4707] = {.lex_state = 422, .external_lex_state = 14},
  [4708] = {.lex_state = 113, .external_lex_state = 9},
  [4709] = {.lex_state = 422, .external_lex_state = 14},
  [4710] = {.lex_state = 422, .external_lex_state = 18},
  [4711] = {.lex_state = 422, .external_lex_state = 19},
  [4712] = {.lex_state = 422, .external_lex_state = 18},
  [4713] = {.lex_state = 422, .external_lex_state = 11},
  [4714] = {.lex_state = 422, .external_lex_state = 17},
  [4715] = {.lex_state = 422, .external_lex_state = 18},
  [4716] = {.lex_state = 422, .external_lex_state = 14},
  [4717] = {.lex_state = 422, .external_lex_state = 12},
  [4718] = {.lex_state = 422, .external_lex_state = 18},
  [4719] = {.lex_state = 422, .external_lex_state = 11},
  [4720] = {.lex_state = 422, .external_lex_state = 14},
  [4721] = {.lex_state = 107, .external_lex_state = 9},
  [4722] = {.lex_state = 416, .external_lex_state = 15},
  [4723] = {.lex_state = 422, .external_lex_state = 14},
  [4724] = {.lex_state = 422, .external_lex_state = 17},
  [4725] = {.lex_state = 422, .external_lex_state = 11},
  [4726] = {.lex_state = 422, .external_lex_state = 17},
  [4727] = {.lex_state = 416, .external_lex_state = 10},
  [4728] = {.lex_state = 108, .external_lex_state = 9},
  [4729] = {.lex_state = 113, .external_lex_state = 9},
  [4730] = {.lex_state = 113, .external_lex_state = 9},
  [4731] = {.lex_state = 422, .external_lex_state = 18},
  [4732] = {.lex_state = 418, .external_lex_state = 45},
  [4733] = {.lex_state = 422, .external_lex_state = 14},
  [4734] = {.lex_state = 422, .external_lex_state = 11},
  [4735] = {.lex_state = 422, .external_lex_state = 11},
  [4736] = {.lex_state = 422, .external_lex_state = 14},
  [4737] = {.lex_state = 422, .external_lex_state = 12},
  [4738] = {.lex_state = 108, .external_lex_state = 9},
  [4739] = {.lex_state = 422, .external_lex_state = 11},
  [4740] = {.lex_state = 422, .external_lex_state = 14},
  [4741] = {.lex_state = 422, .external_lex_state = 11},
  [4742] = {.lex_state = 422, .external_lex_state = 12},
  [4743] = {.lex_state = 422, .external_lex_state = 14},
  [4744] = {.lex_state = 422, .external_lex_state = 11},
  [4745] = {.lex_state = 422, .external_lex_state = 11},
  [4746] = {.lex_state = 108, .external_lex_state = 9},
  [4747] = {.lex_state = 422, .external_lex_state = 18},
  [4748] = {.lex_state = 89, .external_lex_state = 11},
  [4749] = {.lex_state = 108, .external_lex_state = 9},
  [4750] = {.lex_state = 417, .external_lex_state = 10},
  [4751] = {.lex_state = 430, .external_lex_state = 10},
  [4752] = {.lex_state = 430, .external_lex_state = 10},
  [4753] = {.lex_state = 105, .external_lex_state = 11},
  [4754] = {.lex_state = 418, .external_lex_state = 13},
  [4755] = {.lex_state = 105, .external_lex_state = 11},
  [4756] = {.lex_state = 418, .external_lex_state = 13},
  [4757] = {.lex_state = 105, .external_lex_state = 11},
  [4758] = {.lex_state = 420, .external_lex_state = 15},
  [4759] = {.lex_state = 431, .external_lex_state = 15},
  [4760] = {.lex_state = 79, .external_lex_state = 13},
  [4761] = {.lex_state = 105, .external_lex_state = 11},
  [4762] = {.lex_state = 109, .external_lex_state = 13},
  [4763] = {.lex_state = 114, .external_lex_state = 11},
  [4764] = {.lex_state = 103, .external_lex_state = 14},
  [4765] = {.lex_state = 79, .external_lex_state = 13},
  [4766] = {.lex_state = 416, .external_lex_state = 12},
  [4767] = {.lex_state = 108, .external_lex_state = 45},
  [4768] = {.lex_state = 114, .external_lex_state = 11},
  [4769] = {.lex_state = 420, .external_lex_state = 43},
  [4770] = {.lex_state = 109, .external_lex_state = 13},
  [4771] = {.lex_state = 418, .external_lex_state = 13},
  [4772] = {.lex_state = 108, .external_lex_state = 9},
  [4773] = {.lex_state = 108, .external_lex_state = 9},
  [4774] = {.lex_state = 108, .external_lex_state = 9},
  [4775] = {.lex_state = 108, .external_lex_state = 9},
  [4776] = {.lex_state = 108, .external_lex_state = 14},
  [4777] = {.lex_state = 108, .external_lex_state = 9},
  [4778] = {.lex_state = 108, .external_lex_state = 9},
  [4779] = {.lex_state = 108, .external_lex_state = 9},
  [4780] = {.lex_state = 108, .external_lex_state = 9},
  [4781] = {.lex_state = 108, .external_lex_state = 9},
  [4782] = {.lex_state = 108, .external_lex_state = 9},
  [4783] = {.lex_state = 108, .external_lex_state = 9},
  [4784] = {.lex_state = 108, .external_lex_state = 9},
  [4785] = {.lex_state = 108, .external_lex_state = 9},
  [4786] = {.lex_state = 108, .external_lex_state = 9},
  [4787] = {.lex_state = 108, .external_lex_state = 9},
  [4788] = {.lex_state = 108, .external_lex_state = 9},
  [4789] = {.lex_state = 108, .external_lex_state = 9},
  [4790] = {.lex_state = 108, .external_lex_state = 9},
  [4791] = {.lex_state = 108, .external_lex_state = 9},
  [4792] = {.lex_state = 108, .external_lex_state = 9},
  [4793] = {.lex_state = 108, .external_lex_state = 9},
  [4794] = {.lex_state = 108, .external_lex_state = 9},
  [4795] = {.lex_state = 108, .external_lex_state = 9},
  [4796] = {.lex_state = 108, .external_lex_state = 9},
  [4797] = {.lex_state = 108, .external_lex_state = 9},
  [4798] = {.lex_state = 108, .external_lex_state = 9},
  [4799] = {.lex_state = 108, .external_lex_state = 9},
  [4800] = {.lex_state = 108, .external_lex_state = 9},
  [4801] = {.lex_state = 420, .external_lex_state = 43},
  [4802] = {.lex_state = 420, .external_lex_state = 43},
  [4803] = {.lex_state = 113, .external_lex_state = 14},
  [4804] = {.lex_state = 79, .external_lex_state = 13},
  [4805] = {.lex_state = 430, .external_lex_state = 13},
  [4806] = {.lex_state = 113, .external_lex_state = 14},
  [4807] = {.lex_state = 446, .external_lex_state = 39},
  [4808] = {.lex_state = 418, .external_lex_state = 13},
  [4809] = {.lex_state = 430, .external_lex_state = 13},
  [4810] = {.lex_state = 430, .external_lex_state = 13},
  [4811] = {.lex_state = 430, .external_lex_state = 13},
  [4812] = {.lex_state = 430, .external_lex_state = 13},
  [4813] = {.lex_state = 416, .external_lex_state = 17},
  [4814] = {.lex_state = 108, .external_lex_state = 9},
  [4815] = {.lex_state = 108, .external_lex_state = 9},
  [4816] = {.lex_state = 446, .external_lex_state = 39},
  [4817] = {.lex_state = 108, .external_lex_state = 9},
  [4818] = {.lex_state = 108, .external_lex_state = 9},
  [4819] = {.lex_state = 420, .external_lex_state = 9},
  [4820] = {.lex_state = 107, .external_lex_state = 13},
  [4821] = {.lex_state = 431, .external_lex_state = 13},
  [4822] = {.lex_state = 431, .external_lex_state = 13},
  [4823] = {.lex_state = 418, .external_lex_state = 47},
  [4824] = {.lex_state = 418, .external_lex_state = 13},
  [4825] = {.lex_state = 418, .external_lex_state = 37},
  [4826] = {.lex_state = 418, .external_lex_state = 13},
  [4827] = {.lex_state = 446, .external_lex_state = 39},
  [4828] = {.lex_state = 79, .external_lex_state = 13},
  [4829] = {.lex_state = 446, .external_lex_state = 39},
  [4830] = {.lex_state = 417, .external_lex_state = 10},
  [4831] = {.lex_state = 114, .external_lex_state = 11},
  [4832] = {.lex_state = 446, .external_lex_state = 39},
  [4833] = {.lex_state = 112, .external_lex_state = 14},
  [4834] = {.lex_state = 417, .external_lex_state = 10},
  [4835] = {.lex_state = 416, .external_lex_state = 19},
  [4836] = {.lex_state = 418, .external_lex_state = 47},
  [4837] = {.lex_state = 82, .external_lex_state = 11},
  [4838] = {.lex_state = 430, .external_lex_state = 15},
  [4839] = {.lex_state = 112, .external_lex_state = 14},
  [4840] = {.lex_state = 103, .external_lex_state = 14},
  [4841] = {.lex_state = 417, .external_lex_state = 10},
  [4842] = {.lex_state = 108, .external_lex_state = 45},
  [4843] = {.lex_state = 78, .external_lex_state = 13},
  [4844] = {.lex_state = 430, .external_lex_state = 15},
  [4845] = {.lex_state = 430, .external_lex_state = 15},
  [4846] = {.lex_state = 78, .external_lex_state = 13},
  [4847] = {.lex_state = 107, .external_lex_state = 13},
  [4848] = {.lex_state = 109, .external_lex_state = 36},
  [4849] = {.lex_state = 418, .external_lex_state = 47},
  [4850] = {.lex_state = 430, .external_lex_state = 15},
  [4851] = {.lex_state = 103, .external_lex_state = 14},
  [4852] = {.lex_state = 78, .external_lex_state = 13},
  [4853] = {.lex_state = 430, .external_lex_state = 15},
  [4854] = {.lex_state = 103, .external_lex_state = 14},
  [4855] = {.lex_state = 78, .external_lex_state = 13},
  [4856] = {.lex_state = 108, .external_lex_state = 45},
  [4857] = {.lex_state = 446, .external_lex_state = 39},
  [4858] = {.lex_state = 107, .external_lex_state = 13},
  [4859] = {.lex_state = 79, .external_lex_state = 13},
  [4860] = {.lex_state = 420, .external_lex_state = 10},
  [4861] = {.lex_state = 431, .external_lex_state = 10},
  [4862] = {.lex_state = 431, .external_lex_state = 13},
  [4863] = {.lex_state = 431, .external_lex_state = 10},
  [4864] = {.lex_state = 418, .external_lex_state = 13},
  [4865] = {.lex_state = 431, .external_lex_state = 10},
  [4866] = {.lex_state = 79, .external_lex_state = 13},
  [4867] = {.lex_state = 446, .external_lex_state = 39},
  [4868] = {.lex_state = 111, .external_lex_state = 11},
  [4869] = {.lex_state = 79, .external_lex_state = 13},
  [4870] = {.lex_state = 431, .external_lex_state = 15},
  [4871] = {.lex_state = 110, .external_lex_state = 14},
  [4872] = {.lex_state = 420, .external_lex_state = 9},
  [4873] = {.lex_state = 79, .external_lex_state = 13},
  [4874] = {.lex_state = 107, .external_lex_state = 13},
  [4875] = {.lex_state = 109, .external_lex_state = 36},
  [4876] = {.lex_state = 111, .external_lex_state = 11},
  [4877] = {.lex_state = 79, .external_lex_state = 13},
  [4878] = {.lex_state = 111, .external_lex_state = 11},
  [4879] = {.lex_state = 110, .external_lex_state = 14},
  [4880] = {.lex_state = 111, .external_lex_state = 11},
  [4881] = {.lex_state = 111, .external_lex_state = 11},
  [4882] = {.lex_state = 110, .external_lex_state = 14},
  [4883] = {.lex_state = 79, .external_lex_state = 13},
  [4884] = {.lex_state = 108, .external_lex_state = 9},
  [4885] = {.lex_state = 113, .external_lex_state = 14},
  [4886] = {.lex_state = 418, .external_lex_state = 37},
  [4887] = {.lex_state = 108, .external_lex_state = 9},
  [4888] = {.lex_state = 446, .external_lex_state = 39},
  [4889] = {.lex_state = 110, .external_lex_state = 14},
  [4890] = {.lex_state = 108, .external_lex_state = 9},
  [4891] = {.lex_state = 108, .external_lex_state = 9},
  [4892] = {.lex_state = 108, .external_lex_state = 9},
  [4893] = {.lex_state = 110, .external_lex_state = 14},
  [4894] = {.lex_state = 79, .external_lex_state = 13},
  [4895] = {.lex_state = 108, .external_lex_state = 9},
  [4896] = {.lex_state = 446, .external_lex_state = 39},
  [4897] = {.lex_state = 446, .external_lex_state = 39},
  [4898] = {.lex_state = 108, .external_lex_state = 9},
  [4899] = {.lex_state = 108, .external_lex_state = 9},
  [4900] = {.lex_state = 108, .external_lex_state = 9},
  [4901] = {.lex_state = 108, .external_lex_state = 9},
  [4902] = {.lex_state = 418, .external_lex_state = 11},
  [4903] = {.lex_state = 79, .external_lex_state = 13},
  [4904] = {.lex_state = 108, .external_lex_state = 9},
  [4905] = {.lex_state = 108, .external_lex_state = 9},
  [4906] = {.lex_state = 418, .external_lex_state = 37},
  [4907] = {.lex_state = 112, .external_lex_state = 14},
  [4908] = {.lex_state = 108, .external_lex_state = 14},
  [4909] = {.lex_state = 420, .external_lex_state = 10},
  [4910] = {.lex_state = 431, .external_lex_state = 15},
  [4911] = {.lex_state = 417, .external_lex_state = 15},
  [4912] = {.lex_state = 416, .external_lex_state = 18},
  [4913] = {.lex_state = 417, .external_lex_state = 15},
  [4914] = {.lex_state = 420, .external_lex_state = 9},
  [4915] = {.lex_state = 420, .external_lex_state = 15},
  [4916] = {.lex_state = 78, .external_lex_state = 13},
  [4917] = {.lex_state = 417, .external_lex_state = 15},
  [4918] = {.lex_state = 106, .external_lex_state = 11},
  [4919] = {.lex_state = 430, .external_lex_state = 10},
  [4920] = {.lex_state = 108, .external_lex_state = 9},
  [4921] = {.lex_state = 108, .external_lex_state = 9},
  [4922] = {.lex_state = 108, .external_lex_state = 9},
  [4923] = {.lex_state = 446, .external_lex_state = 39},
  [4924] = {.lex_state = 446, .external_lex_state = 39},
  [4925] = {.lex_state = 108, .external_lex_state = 9},
  [4926] = {.lex_state = 108, .external_lex_state = 9},
  [4927] = {.lex_state = 108, .external_lex_state = 9},
  [4928] = {.lex_state = 108, .external_lex_state = 9},
  [4929] = {.lex_state = 416, .external_lex_state = 14},
  [4930] = {.lex_state = 108, .external_lex_state = 9},
  [4931] = {.lex_state = 108, .external_lex_state = 9},
  [4932] = {.lex_state = 108, .external_lex_state = 9},
  [4933] = {.lex_state = 108, .external_lex_state = 9},
  [4934] = {.lex_state = 108, .external_lex_state = 9},
  [4935] = {.lex_state = 108, .external_lex_state = 9},
  [4936] = {.lex_state = 108, .external_lex_state = 9},
  [4937] = {.lex_state = 108, .external_lex_state = 9},
  [4938] = {.lex_state = 108, .external_lex_state = 9},
  [4939] = {.lex_state = 108, .external_lex_state = 9},
  [4940] = {.lex_state = 108, .external_lex_state = 9},
  [4941] = {.lex_state = 108, .external_lex_state = 9},
  [4942] = {.lex_state = 446, .external_lex_state = 39},
  [4943] = {.lex_state = 108, .external_lex_state = 9},
  [4944] = {.lex_state = 418, .external_lex_state = 13},
  [4945] = {.lex_state = 430, .external_lex_state = 10},
  [4946] = {.lex_state = 430, .external_lex_state = 10},
  [4947] = {.lex_state = 79, .external_lex_state = 13},
  [4948] = {.lex_state = 108, .external_lex_state = 9},
  [4949] = {.lex_state = 108, .external_lex_state = 9},
  [4950] = {.lex_state = 417, .external_lex_state = 15},
  [4951] = {.lex_state = 108, .external_lex_state = 9},
  [4952] = {.lex_state = 109, .external_lex_state = 36},
  [4953] = {.lex_state = 431, .external_lex_state = 12},
  [4954] = {.lex_state = 431, .external_lex_state = 11},
  [4955] = {.lex_state = 446, .external_lex_state = 41},
  [4956] = {.lex_state = 431, .external_lex_state = 13},
  [4957] = {.lex_state = 418, .external_lex_state = 52},
  [4958] = {.lex_state = 418, .external_lex_state = 15},
  [4959] = {.lex_state = 84, .external_lex_state = 11},
  [4960] = {.lex_state = 84, .external_lex_state = 11},
  [4961] = {.lex_state = 86, .external_lex_state = 36},
  [4962] = {.lex_state = 418, .external_lex_state = 48},
  [4963] = {.lex_state = 431, .external_lex_state = 15},
  [4964] = {.lex_state = 431, .external_lex_state = 15},
  [4965] = {.lex_state = 446, .external_lex_state = 42},
  [4966] = {.lex_state = 418, .external_lex_state = 13},
  [4967] = {.lex_state = 418, .external_lex_state = 13},
  [4968] = {.lex_state = 430, .external_lex_state = 17},
  [4969] = {.lex_state = 114, .external_lex_state = 11},
  [4970] = {.lex_state = 418, .external_lex_state = 13},
  [4971] = {.lex_state = 107, .external_lex_state = 11},
  [4972] = {.lex_state = 84, .external_lex_state = 11},
  [4973] = {.lex_state = 84, .external_lex_state = 11},
  [4974] = {.lex_state = 446, .external_lex_state = 41},
  [4975] = {.lex_state = 431, .external_lex_state = 18},
  [4976] = {.lex_state = 431, .external_lex_state = 15},
  [4977] = {.lex_state = 420, .external_lex_state = 36},
  [4978] = {.lex_state = 418, .external_lex_state = 13},
  [4979] = {.lex_state = 446, .external_lex_state = 41},
  [4980] = {.lex_state = 431, .external_lex_state = 15},
  [4981] = {.lex_state = 446, .external_lex_state = 39},
  [4982] = {.lex_state = 418, .external_lex_state = 15},
  [4983] = {.lex_state = 112, .external_lex_state = 14},
  [4984] = {.lex_state = 113, .external_lex_state = 14},
  [4985] = {.lex_state = 114, .external_lex_state = 11},
  [4986] = {.lex_state = 420, .external_lex_state = 18},
  [4987] = {.lex_state = 417, .external_lex_state = 19},
  [4988] = {.lex_state = 446, .external_lex_state = 41},
  [4989] = {.lex_state = 114, .external_lex_state = 11},
  [4990] = {.lex_state = 114, .external_lex_state = 11},
  [4991] = {.lex_state = 420, .external_lex_state = 12},
  [4992] = {.lex_state = 417, .external_lex_state = 19},
  [4993] = {.lex_state = 418, .external_lex_state = 52},
  [4994] = {.lex_state = 114, .external_lex_state = 11},
  [4995] = {.lex_state = 418, .external_lex_state = 13},
  [4996] = {.lex_state = 418, .external_lex_state = 11},
  [4997] = {.lex_state = 417, .external_lex_state = 14},
  [4998] = {.lex_state = 108, .external_lex_state = 14},
  [4999] = {.lex_state = 417, .external_lex_state = 12},
  [5000] = {.lex_state = 446, .external_lex_state = 39},
  [5001] = {.lex_state = 418, .external_lex_state = 13},
  [5002] = {.lex_state = 83, .external_lex_state = 11},
  [5003] = {.lex_state = 114, .external_lex_state = 11},
  [5004] = {.lex_state = 430, .external_lex_state = 13},
  [5005] = {.lex_state = 83, .external_lex_state = 11},
  [5006] = {.lex_state = 418, .external_lex_state = 11},
  [5007] = {.lex_state = 417, .external_lex_state = 19},
  [5008] = {.lex_state = 446, .external_lex_state = 42},
  [5009] = {.lex_state = 431, .external_lex_state = 18},
  [5010] = {.lex_state = 446, .external_lex_state = 42},
  [5011] = {.lex_state = 418, .external_lex_state = 15},
  [5012] = {.lex_state = 431, .external_lex_state = 13},
  [5013] = {.lex_state = 431, .external_lex_state = 13},
  [5014] = {.lex_state = 114, .external_lex_state = 11},
  [5015] = {.lex_state = 420, .external_lex_state = 36},
  [5016] = {.lex_state = 83, .external_lex_state = 11},
  [5017] = {.lex_state = 418, .external_lex_state = 13},
  [5018] = {.lex_state = 112, .external_lex_state = 14},
  [5019] = {.lex_state = 113, .external_lex_state = 14},
  [5020] = {.lex_state = 113, .external_lex_state = 14},
  [5021] = {.lex_state = 418, .external_lex_state = 48},
  [5022] = {.lex_state = 431, .external_lex_state = 19},
  [5023] = {.lex_state = 113, .external_lex_state = 14},
  [5024] = {.lex_state = 109, .external_lex_state = 11},
  [5025] = {.lex_state = 431, .external_lex_state = 19},
  [5026] = {.lex_state = 431, .external_lex_state = 15},
  [5027] = {.lex_state = 431, .external_lex_state = 15},
  [5028] = {.lex_state = 431, .external_lex_state = 17},
  [5029] = {.lex_state = 108, .external_lex_state = 14},
  [5030] = {.lex_state = 418, .external_lex_state = 49},
  [5031] = {.lex_state = 446, .external_lex_state = 41},
  [5032] = {.lex_state = 446, .external_lex_state = 41},
  [5033] = {.lex_state = 446, .external_lex_state = 41},
  [5034] = {.lex_state = 113, .external_lex_state = 14},
  [5035] = {.lex_state = 114, .external_lex_state = 11},
  [5036] = {.lex_state = 113, .external_lex_state = 14},
  [5037] = {.lex_state = 431, .external_lex_state = 13},
  [5038] = {.lex_state = 431, .external_lex_state = 13},
  [5039] = {.lex_state = 431, .external_lex_state = 14},
  [5040] = {.lex_state = 113, .external_lex_state = 14},
  [5041] = {.lex_state = 417, .external_lex_state = 17},
  [5042] = {.lex_state = 417, .external_lex_state = 18},
  [5043] = {.lex_state = 86, .external_lex_state = 36},
  [5044] = {.lex_state = 109, .external_lex_state = 11},
  [5045] = {.lex_state = 431, .external_lex_state = 12},
  [5046] = {.lex_state = 420, .external_lex_state = 14},
  [5047] = {.lex_state = 446, .external_lex_state = 41},
  [5048] = {.lex_state = 418, .external_lex_state = 13},
  [5049] = {.lex_state = 430, .external_lex_state = 18},
  [5050] = {.lex_state = 83, .external_lex_state = 11},
  [5051] = {.lex_state = 114, .external_lex_state = 11},
  [5052] = {.lex_state = 446, .external_lex_state = 42},
  [5053] = {.lex_state = 431, .external_lex_state = 19},
  [5054] = {.lex_state = 417, .external_lex_state = 19},
  [5055] = {.lex_state = 114, .external_lex_state = 11},
  [5056] = {.lex_state = 446, .external_lex_state = 42},
  [5057] = {.lex_state = 446, .external_lex_state = 42},
  [5058] = {.lex_state = 418, .external_lex_state = 45},
  [5059] = {.lex_state = 430, .external_lex_state = 10},
  [5060] = {.lex_state = 446, .external_lex_state = 42},
  [5061] = {.lex_state = 418, .external_lex_state = 13},
  [5062] = {.lex_state = 431, .external_lex_state = 15},
  [5063] = {.lex_state = 431, .external_lex_state = 15},
  [5064] = {.lex_state = 108, .external_lex_state = 14},
  [5065] = {.lex_state = 430, .external_lex_state = 15},
  [5066] = {.lex_state = 430, .external_lex_state = 15},
  [5067] = {.lex_state = 418, .external_lex_state = 13},
  [5068] = {.lex_state = 446, .external_lex_state = 41},
  [5069] = {.lex_state = 430, .external_lex_state = 19},
  [5070] = {.lex_state = 430, .external_lex_state = 17},
  [5071] = {.lex_state = 430, .external_lex_state = 19},
  [5072] = {.lex_state = 430, .external_lex_state = 14},
  [5073] = {.lex_state = 418, .external_lex_state = 13},
  [5074] = {.lex_state = 417, .external_lex_state = 18},
  [5075] = {.lex_state = 108, .external_lex_state = 14},
  [5076] = {.lex_state = 418, .external_lex_state = 13},
  [5077] = {.lex_state = 418, .external_lex_state = 49},
  [5078] = {.lex_state = 418, .external_lex_state = 13},
  [5079] = {.lex_state = 418, .external_lex_state = 13},
  [5080] = {.lex_state = 430, .external_lex_state = 10},
  [5081] = {.lex_state = 430, .external_lex_state = 17},
  [5082] = {.lex_state = 431, .external_lex_state = 11},
  [5083] = {.lex_state = 431, .external_lex_state = 13},
  [5084] = {.lex_state = 431, .external_lex_state = 10},
  [5085] = {.lex_state = 418, .external_lex_state = 13},
  [5086] = {.lex_state = 109, .external_lex_state = 11},
  [5087] = {.lex_state = 430, .external_lex_state = 15},
  [5088] = {.lex_state = 431, .external_lex_state = 13},
  [5089] = {.lex_state = 418, .external_lex_state = 13},
  [5090] = {.lex_state = 430, .external_lex_state = 15},
  [5091] = {.lex_state = 446, .external_lex_state = 42},
  [5092] = {.lex_state = 431, .external_lex_state = 13},
  [5093] = {.lex_state = 446, .external_lex_state = 39},
  [5094] = {.lex_state = 431, .external_lex_state = 13},
  [5095] = {.lex_state = 431, .external_lex_state = 13},
  [5096] = {.lex_state = 109, .external_lex_state = 11},
  [5097] = {.lex_state = 446, .external_lex_state = 42},
  [5098] = {.lex_state = 446, .external_lex_state = 42},
  [5099] = {.lex_state = 431, .external_lex_state = 15},
  [5100] = {.lex_state = 431, .external_lex_state = 10},
  [5101] = {.lex_state = 109, .external_lex_state = 11},
  [5102] = {.lex_state = 418, .external_lex_state = 13},
  [5103] = {.lex_state = 418, .external_lex_state = 13},
  [5104] = {.lex_state = 418, .external_lex_state = 13},
  [5105] = {.lex_state = 418, .external_lex_state = 13},
  [5106] = {.lex_state = 420, .external_lex_state = 13},
  [5107] = {.lex_state = 418, .external_lex_state = 13},
  [5108] = {.lex_state = 113, .external_lex_state = 14},
  [5109] = {.lex_state = 431, .external_lex_state = 10},
  [5110] = {.lex_state = 418, .external_lex_state = 13},
  [5111] = {.lex_state = 431, .external_lex_state = 10},
  [5112] = {.lex_state = 418, .external_lex_state = 13},
  [5113] = {.lex_state = 418, .external_lex_state = 13},
  [5114] = {.lex_state = 418, .external_lex_state = 13},
  [5115] = {.lex_state = 418, .external_lex_state = 13},
  [5116] = {.lex_state = 418, .external_lex_state = 13},
  [5117] = {.lex_state = 418, .external_lex_state = 13},
  [5118] = {.lex_state = 418, .external_lex_state = 13},
  [5119] = {.lex_state = 418, .external_lex_state = 13},
  [5120] = {.lex_state = 418, .external_lex_state = 13},
  [5121] = {.lex_state = 418, .external_lex_state = 13},
  [5122] = {.lex_state = 418, .external_lex_state = 13},
  [5123] = {.lex_state = 418, .external_lex_state = 13},
  [5124] = {.lex_state = 418, .external_lex_state = 13},
  [5125] = {.lex_state = 107, .external_lex_state = 11},
  [5126] = {.lex_state = 418, .external_lex_state = 13},
  [5127] = {.lex_state = 420, .external_lex_state = 13},
  [5128] = {.lex_state = 418, .external_lex_state = 13},
  [5129] = {.lex_state = 420, .external_lex_state = 17},
  [5130] = {.lex_state = 418, .external_lex_state = 13},
  [5131] = {.lex_state = 418, .external_lex_state = 13},
  [5132] = {.lex_state = 418, .external_lex_state = 10},
  [5133] = {.lex_state = 418, .external_lex_state = 13},
  [5134] = {.lex_state = 418, .external_lex_state = 13},
  [5135] = {.lex_state = 113, .external_lex_state = 14},
  [5136] = {.lex_state = 430, .external_lex_state = 11},
  [5137] = {.lex_state = 420, .external_lex_state = 18},
  [5138] = {.lex_state = 418, .external_lex_state = 15},
  [5139] = {.lex_state = 431, .external_lex_state = 14},
  [5140] = {.lex_state = 431, .external_lex_state = 10},
  [5141] = {.lex_state = 418, .external_lex_state = 13},
  [5142] = {.lex_state = 418, .external_lex_state = 13},
  [5143] = {.lex_state = 418, .external_lex_state = 13},
  [5144] = {.lex_state = 446, .external_lex_state = 40},
  [5145] = {.lex_state = 418, .external_lex_state = 13},
  [5146] = {.lex_state = 446, .external_lex_state = 40},
  [5147] = {.lex_state = 114, .external_lex_state = 11},
  [5148] = {.lex_state = 418, .external_lex_state = 45},
  [5149] = {.lex_state = 446, .external_lex_state = 41},
  [5150] = {.lex_state = 418, .external_lex_state = 13},
  [5151] = {.lex_state = 446, .external_lex_state = 40},
  [5152] = {.lex_state = 417, .external_lex_state = 18},
  [5153] = {.lex_state = 85, .external_lex_state = 11},
  [5154] = {.lex_state = 418, .external_lex_state = 52},
  [5155] = {.lex_state = 431, .external_lex_state = 10},
  [5156] = {.lex_state = 108, .external_lex_state = 14},
  [5157] = {.lex_state = 430, .external_lex_state = 15},
  [5158] = {.lex_state = 418, .external_lex_state = 13},
  [5159] = {.lex_state = 430, .external_lex_state = 10},
  [5160] = {.lex_state = 418, .external_lex_state = 13},
  [5161] = {.lex_state = 420, .external_lex_state = 19},
  [5162] = {.lex_state = 446, .external_lex_state = 41},
  [5163] = {.lex_state = 113, .external_lex_state = 14},
  [5164] = {.lex_state = 418, .external_lex_state = 38},
  [5165] = {.lex_state = 418, .external_lex_state = 10},
  [5166] = {.lex_state = 111, .external_lex_state = 11},
  [5167] = {.lex_state = 431, .external_lex_state = 13},
  [5168] = {.lex_state = 111, .external_lex_state = 11},
  [5169] = {.lex_state = 111, .external_lex_state = 11},
  [5170] = {.lex_state = 418, .external_lex_state = 13},
  [5171] = {.lex_state = 113, .external_lex_state = 14},
  [5172] = {.lex_state = 86, .external_lex_state = 36},
  [5173] = {.lex_state = 446, .external_lex_state = 40},
  [5174] = {.lex_state = 418, .external_lex_state = 13},
  [5175] = {.lex_state = 446, .external_lex_state = 42},
  [5176] = {.lex_state = 431, .external_lex_state = 10},
  [5177] = {.lex_state = 111, .external_lex_state = 11},
  [5178] = {.lex_state = 113, .external_lex_state = 14},
  [5179] = {.lex_state = 431, .external_lex_state = 10},
  [5180] = {.lex_state = 111, .external_lex_state = 11},
  [5181] = {.lex_state = 418, .external_lex_state = 38},
  [5182] = {.lex_state = 420, .external_lex_state = 13},
  [5183] = {.lex_state = 418, .external_lex_state = 38},
  [5184] = {.lex_state = 418, .external_lex_state = 13},
  [5185] = {.lex_state = 431, .external_lex_state = 10},
  [5186] = {.lex_state = 420, .external_lex_state = 13},
  [5187] = {.lex_state = 430, .external_lex_state = 11},
  [5188] = {.lex_state = 110, .external_lex_state = 14},
  [5189] = {.lex_state = 430, .external_lex_state = 11},
  [5190] = {.lex_state = 418, .external_lex_state = 45},
  [5191] = {.lex_state = 418, .external_lex_state = 13},
  [5192] = {.lex_state = 110, .external_lex_state = 14},
  [5193] = {.lex_state = 431, .external_lex_state = 13},
  [5194] = {.lex_state = 85, .external_lex_state = 11},
  [5195] = {.lex_state = 431, .external_lex_state = 13},
  [5196] = {.lex_state = 110, .external_lex_state = 14},
  [5197] = {.lex_state = 418, .external_lex_state = 13},
  [5198] = {.lex_state = 418, .external_lex_state = 13},
  [5199] = {.lex_state = 430, .external_lex_state = 11},
  [5200] = {.lex_state = 430, .external_lex_state = 11},
  [5201] = {.lex_state = 431, .external_lex_state = 15},
  [5202] = {.lex_state = 430, .external_lex_state = 19},
  [5203] = {.lex_state = 446, .external_lex_state = 40},
  [5204] = {.lex_state = 418, .external_lex_state = 13},
  [5205] = {.lex_state = 446, .external_lex_state = 39},
  [5206] = {.lex_state = 420, .external_lex_state = 13},
  [5207] = {.lex_state = 446, .external_lex_state = 39},
  [5208] = {.lex_state = 430, .external_lex_state = 19},
  [5209] = {.lex_state = 430, .external_lex_state = 12},
  [5210] = {.lex_state = 430, .external_lex_state = 10},
  [5211] = {.lex_state = 418, .external_lex_state = 10},
  [5212] = {.lex_state = 110, .external_lex_state = 14},
  [5213] = {.lex_state = 85, .external_lex_state = 11},
  [5214] = {.lex_state = 431, .external_lex_state = 10},
  [5215] = {.lex_state = 420, .external_lex_state = 13},
  [5216] = {.lex_state = 417, .external_lex_state = 14},
  [5217] = {.lex_state = 446, .external_lex_state = 40},
  [5218] = {.lex_state = 418, .external_lex_state = 13},
  [5219] = {.lex_state = 418, .external_lex_state = 13},
  [5220] = {.lex_state = 110, .external_lex_state = 14},
  [5221] = {.lex_state = 418, .external_lex_state = 13},
  [5222] = {.lex_state = 418, .external_lex_state = 10},
  [5223] = {.lex_state = 109, .external_lex_state = 11},
  [5224] = {.lex_state = 446, .external_lex_state = 40},
  [5225] = {.lex_state = 418, .external_lex_state = 13},
  [5226] = {.lex_state = 418, .external_lex_state = 13},
  [5227] = {.lex_state = 114, .external_lex_state = 11},
  [5228] = {.lex_state = 446, .external_lex_state = 40},
  [5229] = {.lex_state = 431, .external_lex_state = 14},
  [5230] = {.lex_state = 430, .external_lex_state = 10},
  [5231] = {.lex_state = 430, .external_lex_state = 18},
  [5232] = {.lex_state = 113, .external_lex_state = 14},
  [5233] = {.lex_state = 430, .external_lex_state = 18},
  [5234] = {.lex_state = 418, .external_lex_state = 13},
  [5235] = {.lex_state = 417, .external_lex_state = 18},
  [5236] = {.lex_state = 430, .external_lex_state = 19},
  [5237] = {.lex_state = 446, .external_lex_state = 40},
  [5238] = {.lex_state = 418, .external_lex_state = 13},
  [5239] = {.lex_state = 430, .external_lex_state = 14},
  [5240] = {.lex_state = 430, .external_lex_state = 14},
  [5241] = {.lex_state = 417, .external_lex_state = 14},
  [5242] = {.lex_state = 420, .external_lex_state = 13},
  [5243] = {.lex_state = 431, .external_lex_state = 17},
  [5244] = {.lex_state = 430, .external_lex_state = 14},
  [5245] = {.lex_state = 430, .external_lex_state = 17},
  [5246] = {.lex_state = 420, .external_lex_state = 19},
  [5247] = {.lex_state = 420, .external_lex_state = 13},
  [5248] = {.lex_state = 446, .external_lex_state = 42},
  [5249] = {.lex_state = 420, .external_lex_state = 17},
  [5250] = {.lex_state = 418, .external_lex_state = 49},
  [5251] = {.lex_state = 86, .external_lex_state = 11},
  [5252] = {.lex_state = 418, .external_lex_state = 15},
  [5253] = {.lex_state = 109, .external_lex_state = 11},
  [5254] = {.lex_state = 430, .external_lex_state = 12},
  [5255] = {.lex_state = 430, .external_lex_state = 12},
  [5256] = {.lex_state = 446, .external_lex_state = 40},
  [5257] = {.lex_state = 418, .external_lex_state = 10},
  [5258] = {.lex_state = 418, .external_lex_state = 13},
  [5259] = {.lex_state = 431, .external_lex_state = 10},
  [5260] = {.lex_state = 431, .external_lex_state = 17},
  [5261] = {.lex_state = 446, .external_lex_state = 40},
  [5262] = {.lex_state = 431, .external_lex_state = 15},
  [5263] = {.lex_state = 431, .external_lex_state = 18},
  [5264] = {.lex_state = 430, .external_lex_state = 13},
  [5265] = {.lex_state = 430, .external_lex_state = 13},
  [5266] = {.lex_state = 420, .external_lex_state = 36},
  [5267] = {.lex_state = 430, .external_lex_state = 18},
  [5268] = {.lex_state = 109, .external_lex_state = 11},
  [5269] = {.lex_state = 430, .external_lex_state = 18},
  [5270] = {.lex_state = 417, .external_lex_state = 14},
  [5271] = {.lex_state = 418, .external_lex_state = 13},
  [5272] = {.lex_state = 430, .external_lex_state = 13},
  [5273] = {.lex_state = 418, .external_lex_state = 13},
  [5274] = {.lex_state = 430, .external_lex_state = 13},
  [5275] = {.lex_state = 418, .external_lex_state = 13},
  [5276] = {.lex_state = 430, .external_lex_state = 12},
  [5277] = {.lex_state = 430, .external_lex_state = 12},
  [5278] = {.lex_state = 446, .external_lex_state = 39},
  [5279] = {.lex_state = 446, .external_lex_state = 41},
  [5280] = {.lex_state = 431, .external_lex_state = 10},
  [5281] = {.lex_state = 84, .external_lex_state = 11},
  [5282] = {.lex_state = 446, .external_lex_state = 40},
  [5283] = {.lex_state = 418, .external_lex_state = 48},
  [5284] = {.lex_state = 417, .external_lex_state = 12},
  [5285] = {.lex_state = 420, .external_lex_state = 14},
  [5286] = {.lex_state = 418, .external_lex_state = 13},
  [5287] = {.lex_state = 430, .external_lex_state = 17},
  [5288] = {.lex_state = 86, .external_lex_state = 11},
  [5289] = {.lex_state = 418, .external_lex_state = 13},
  [5290] = {.lex_state = 417, .external_lex_state = 12},
  [5291] = {.lex_state = 446, .external_lex_state = 41},
  [5292] = {.lex_state = 107, .external_lex_state = 11},
  [5293] = {.lex_state = 431, .external_lex_state = 10},
  [5294] = {.lex_state = 418, .external_lex_state = 13},
  [5295] = {.lex_state = 446, .external_lex_state = 42},
  [5296] = {.lex_state = 417, .external_lex_state = 17},
  [5297] = {.lex_state = 114, .external_lex_state = 11},
  [5298] = {.lex_state = 420, .external_lex_state = 12},
  [5299] = {.lex_state = 417, .external_lex_state = 17},
  [5300] = {.lex_state = 431, .external_lex_state = 15},
  [5301] = {.lex_state = 430, .external_lex_state = 14},
  [5302] = {.lex_state = 417, .external_lex_state = 12},
  [5303] = {.lex_state = 417, .external_lex_state = 17},
  [5304] = {.lex_state = 431, .external_lex_state = 12},
  [5305] = {.lex_state = 431, .external_lex_state = 15},
  [5306] = {.lex_state = 446, .external_lex_state = 40},
  [5307] = {.lex_state = 418, .external_lex_state = 13},
  [5308] = {.lex_state = 431, .external_lex_state = 11},
  [5309] = {.lex_state = 107, .external_lex_state = 11},
  [5310] = {.lex_state = 418, .external_lex_state = 10},
  [5311] = {.lex_state = 431, .external_lex_state = 13},
  [5312] = {.lex_state = 431, .external_lex_state = 13},
  [5313] = {.lex_state = 418, .external_lex_state = 10},
  [5314] = {.lex_state = 418, .external_lex_state = 10},
  [5315] = {.lex_state = 108, .external_lex_state = 14},
  [5316] = {.lex_state = 418, .external_lex_state = 10},
  [5317] = {.lex_state = 108, .external_lex_state = 14},
  [5318] = {.lex_state = 418, .external_lex_state = 10},
  [5319] = {.lex_state = 418, .external_lex_state = 10},
  [5320] = {.lex_state = 446, .external_lex_state = 41},
  [5321] = {.lex_state = 418, .external_lex_state = 10},
  [5322] = {.lex_state = 418, .external_lex_state = 10},
  [5323] = {.lex_state = 418, .external_lex_state = 10},
  [5324] = {.lex_state = 418, .external_lex_state = 10},
  [5325] = {.lex_state = 418, .external_lex_state = 10},
  [5326] = {.lex_state = 418, .external_lex_state = 10},
  [5327] = {.lex_state = 418, .external_lex_state = 10},
  [5328] = {.lex_state = 418, .external_lex_state = 10},
  [5329] = {.lex_state = 418, .external_lex_state = 10},
  [5330] = {.lex_state = 418, .external_lex_state = 10},
  [5331] = {.lex_state = 418, .external_lex_state = 10},
  [5332] = {.lex_state = 418, .external_lex_state = 17},
  [5333] = {.lex_state = 418, .external_lex_state = 10},
  [5334] = {.lex_state = 418, .external_lex_state = 10},
  [5335] = {.lex_state = 446, .external_lex_state = 39},
  [5336] = {.lex_state = 418, .external_lex_state = 10},
  [5337] = {.lex_state = 418, .external_lex_state = 10},
  [5338] = {.lex_state = 418, .external_lex_state = 10},
  [5339] = {.lex_state = 418, .external_lex_state = 10},
  [5340] = {.lex_state = 431, .external_lex_state = 12},
  [5341] = {.lex_state = 418, .external_lex_state = 10},
  [5342] = {.lex_state = 418, .external_lex_state = 10},
  [5343] = {.lex_state = 108, .external_lex_state = 14},
  [5344] = {.lex_state = 446, .external_lex_state = 42},
  [5345] = {.lex_state = 418, .external_lex_state = 10},
  [5346] = {.lex_state = 418, .external_lex_state = 10},
  [5347] = {.lex_state = 108, .external_lex_state = 14},
  [5348] = {.lex_state = 418, .external_lex_state = 10},
  [5349] = {.lex_state = 418, .external_lex_state = 10},
  [5350] = {.lex_state = 446, .external_lex_state = 42},
  [5351] = {.lex_state = 108, .external_lex_state = 14},
  [5352] = {.lex_state = 418, .external_lex_state = 10},
  [5353] = {.lex_state = 418, .external_lex_state = 15},
  [5354] = {.lex_state = 85, .external_lex_state = 11},
  [5355] = {.lex_state = 418, .external_lex_state = 15},
  [5356] = {.lex_state = 418, .external_lex_state = 15},
  [5357] = {.lex_state = 418, .external_lex_state = 15},
  [5358] = {.lex_state = 431, .external_lex_state = 18},
  [5359] = {.lex_state = 108, .external_lex_state = 14},
  [5360] = {.lex_state = 418, .external_lex_state = 15},
  [5361] = {.lex_state = 418, .external_lex_state = 15},
  [5362] = {.lex_state = 418, .external_lex_state = 15},
  [5363] = {.lex_state = 418, .external_lex_state = 15},
  [5364] = {.lex_state = 418, .external_lex_state = 15},
  [5365] = {.lex_state = 418, .external_lex_state = 15},
  [5366] = {.lex_state = 418, .external_lex_state = 15},
  [5367] = {.lex_state = 418, .external_lex_state = 15},
  [5368] = {.lex_state = 418, .external_lex_state = 15},
  [5369] = {.lex_state = 418, .external_lex_state = 15},
  [5370] = {.lex_state = 418, .external_lex_state = 15},
  [5371] = {.lex_state = 418, .external_lex_state = 15},
  [5372] = {.lex_state = 418, .external_lex_state = 15},
  [5373] = {.lex_state = 418, .external_lex_state = 15},
  [5374] = {.lex_state = 418, .external_lex_state = 15},
  [5375] = {.lex_state = 418, .external_lex_state = 15},
  [5376] = {.lex_state = 418, .external_lex_state = 15},
  [5377] = {.lex_state = 418, .external_lex_state = 10},
  [5378] = {.lex_state = 108, .external_lex_state = 14},
  [5379] = {.lex_state = 418, .external_lex_state = 15},
  [5380] = {.lex_state = 418, .external_lex_state = 10},
  [5381] = {.lex_state = 108, .external_lex_state = 14},
  [5382] = {.lex_state = 418, .external_lex_state = 10},
  [5383] = {.lex_state = 418, .external_lex_state = 15},
  [5384] = {.lex_state = 418, .external_lex_state = 15},
  [5385] = {.lex_state = 418, .external_lex_state = 15},
  [5386] = {.lex_state = 418, .external_lex_state = 15},
  [5387] = {.lex_state = 418, .external_lex_state = 10},
  [5388] = {.lex_state = 418, .external_lex_state = 10},
  [5389] = {.lex_state = 418, .external_lex_state = 10},
  [5390] = {.lex_state = 431, .external_lex_state = 17},
  [5391] = {.lex_state = 418, .external_lex_state = 10},
  [5392] = {.lex_state = 431, .external_lex_state = 18},
  [5393] = {.lex_state = 418, .external_lex_state = 10},
  [5394] = {.lex_state = 418, .external_lex_state = 10},
  [5395] = {.lex_state = 418, .external_lex_state = 15},
  [5396] = {.lex_state = 418, .external_lex_state = 10},
  [5397] = {.lex_state = 418, .external_lex_state = 10},
  [5398] = {.lex_state = 418, .external_lex_state = 10},
  [5399] = {.lex_state = 418, .external_lex_state = 10},
  [5400] = {.lex_state = 418, .external_lex_state = 10},
  [5401] = {.lex_state = 431, .external_lex_state = 18},
  [5402] = {.lex_state = 431, .external_lex_state = 18},
  [5403] = {.lex_state = 420, .external_lex_state = 11},
  [5404] = {.lex_state = 431, .external_lex_state = 17},
  [5405] = {.lex_state = 418, .external_lex_state = 10},
  [5406] = {.lex_state = 418, .external_lex_state = 10},
  [5407] = {.lex_state = 431, .external_lex_state = 17},
  [5408] = {.lex_state = 431, .external_lex_state = 18},
  [5409] = {.lex_state = 84, .external_lex_state = 11},
  [5410] = {.lex_state = 418, .external_lex_state = 10},
  [5411] = {.lex_state = 431, .external_lex_state = 13},
  [5412] = {.lex_state = 446, .external_lex_state = 44},
  [5413] = {.lex_state = 430, .external_lex_state = 18},
  [5414] = {.lex_state = 430, .external_lex_state = 18},
  [5415] = {.lex_state = 418, .external_lex_state = 10},
  [5416] = {.lex_state = 108, .external_lex_state = 14},
  [5417] = {.lex_state = 418, .external_lex_state = 10},
  [5418] = {.lex_state = 431, .external_lex_state = 18},
  [5419] = {.lex_state = 108, .external_lex_state = 14},
  [5420] = {.lex_state = 418, .external_lex_state = 10},
  [5421] = {.lex_state = 446, .external_lex_state = 44},
  [5422] = {.lex_state = 418, .external_lex_state = 10},
  [5423] = {.lex_state = 418, .external_lex_state = 19},
  [5424] = {.lex_state = 109, .external_lex_state = 11},
  [5425] = {.lex_state = 109, .external_lex_state = 11},
  [5426] = {.lex_state = 420, .external_lex_state = 13},
  [5427] = {.lex_state = 109, .external_lex_state = 11},
  [5428] = {.lex_state = 109, .external_lex_state = 11},
  [5429] = {.lex_state = 418, .external_lex_state = 13},
  [5430] = {.lex_state = 109, .external_lex_state = 11},
  [5431] = {.lex_state = 420, .external_lex_state = 13},
  [5432] = {.lex_state = 108, .external_lex_state = 14},
  [5433] = {.lex_state = 109, .external_lex_state = 11},
  [5434] = {.lex_state = 109, .external_lex_state = 11},
  [5435] = {.lex_state = 109, .external_lex_state = 11},
  [5436] = {.lex_state = 109, .external_lex_state = 11},
  [5437] = {.lex_state = 109, .external_lex_state = 11},
  [5438] = {.lex_state = 109, .external_lex_state = 11},
  [5439] = {.lex_state = 109, .external_lex_state = 11},
  [5440] = {.lex_state = 109, .external_lex_state = 11},
  [5441] = {.lex_state = 109, .external_lex_state = 11},
  [5442] = {.lex_state = 109, .external_lex_state = 11},
  [5443] = {.lex_state = 109, .external_lex_state = 11},
  [5444] = {.lex_state = 109, .external_lex_state = 11},
  [5445] = {.lex_state = 109, .external_lex_state = 11},
  [5446] = {.lex_state = 109, .external_lex_state = 11},
  [5447] = {.lex_state = 109, .external_lex_state = 11},
  [5448] = {.lex_state = 430, .external_lex_state = 12},
  [5449] = {.lex_state = 109, .external_lex_state = 11},
  [5450] = {.lex_state = 109, .external_lex_state = 11},
  [5451] = {.lex_state = 430, .external_lex_state = 18},
  [5452] = {.lex_state = 431, .external_lex_state = 18},
  [5453] = {.lex_state = 420, .external_lex_state = 13},
  [5454] = {.lex_state = 109, .external_lex_state = 11},
  [5455] = {.lex_state = 420, .external_lex_state = 13},
  [5456] = {.lex_state = 420, .external_lex_state = 13},
  [5457] = {.lex_state = 109, .external_lex_state = 11},
  [5458] = {.lex_state = 109, .external_lex_state = 11},
  [5459] = {.lex_state = 109, .external_lex_state = 11},
  [5460] = {.lex_state = 109, .external_lex_state = 11},
  [5461] = {.lex_state = 420, .external_lex_state = 13},
  [5462] = {.lex_state = 418, .external_lex_state = 14},
  [5463] = {.lex_state = 420, .external_lex_state = 13},
  [5464] = {.lex_state = 430, .external_lex_state = 18},
  [5465] = {.lex_state = 108, .external_lex_state = 14},
  [5466] = {.lex_state = 431, .external_lex_state = 17},
  [5467] = {.lex_state = 108, .external_lex_state = 14},
  [5468] = {.lex_state = 109, .external_lex_state = 11},
  [5469] = {.lex_state = 108, .external_lex_state = 14},
  [5470] = {.lex_state = 108, .external_lex_state = 14},
  [5471] = {.lex_state = 85, .external_lex_state = 11},
  [5472] = {.lex_state = 85, .external_lex_state = 11},
  [5473] = {.lex_state = 420, .external_lex_state = 13},
  [5474] = {.lex_state = 431, .external_lex_state = 17},
  [5475] = {.lex_state = 420, .external_lex_state = 13},
  [5476] = {.lex_state = 420, .external_lex_state = 13},
  [5477] = {.lex_state = 420, .external_lex_state = 13},
  [5478] = {.lex_state = 446, .external_lex_state = 40},
  [5479] = {.lex_state = 420, .external_lex_state = 13},
  [5480] = {.lex_state = 86, .external_lex_state = 11},
  [5481] = {.lex_state = 420, .external_lex_state = 13},
  [5482] = {.lex_state = 420, .external_lex_state = 13},
  [5483] = {.lex_state = 420, .external_lex_state = 13},
  [5484] = {.lex_state = 420, .external_lex_state = 13},
  [5485] = {.lex_state = 420, .external_lex_state = 13},
  [5486] = {.lex_state = 431, .external_lex_state = 14},
  [5487] = {.lex_state = 418, .external_lex_state = 18},
  [5488] = {.lex_state = 431, .external_lex_state = 14},
  [5489] = {.lex_state = 420, .external_lex_state = 13},
  [5490] = {.lex_state = 446, .external_lex_state = 40},
  [5491] = {.lex_state = 420, .external_lex_state = 13},
  [5492] = {.lex_state = 85, .external_lex_state = 11},
  [5493] = {.lex_state = 418, .external_lex_state = 19},
  [5494] = {.lex_state = 418, .external_lex_state = 10},
  [5495] = {.lex_state = 420, .external_lex_state = 13},
  [5496] = {.lex_state = 108, .external_lex_state = 14},
  [5497] = {.lex_state = 86, .external_lex_state = 11},
  [5498] = {.lex_state = 431, .external_lex_state = 11},
  [5499] = {.lex_state = 418, .external_lex_state = 15},
  [5500] = {.lex_state = 418, .external_lex_state = 15},
  [5501] = {.lex_state = 420, .external_lex_state = 13},
  [5502] = {.lex_state = 420, .external_lex_state = 11},
  [5503] = {.lex_state = 418, .external_lex_state = 15},
  [5504] = {.lex_state = 418, .external_lex_state = 15},
  [5505] = {.lex_state = 420, .external_lex_state = 13},
  [5506] = {.lex_state = 420, .external_lex_state = 13},
  [5507] = {.lex_state = 420, .external_lex_state = 13},
  [5508] = {.lex_state = 85, .external_lex_state = 11},
  [5509] = {.lex_state = 85, .external_lex_state = 11},
  [5510] = {.lex_state = 431, .external_lex_state = 11},
  [5511] = {.lex_state = 420, .external_lex_state = 11},
  [5512] = {.lex_state = 431, .external_lex_state = 18},
  [5513] = {.lex_state = 431, .external_lex_state = 11},
  [5514] = {.lex_state = 431, .external_lex_state = 11},
  [5515] = {.lex_state = 420, .external_lex_state = 13},
  [5516] = {.lex_state = 431, .external_lex_state = 14},
  [5517] = {.lex_state = 431, .external_lex_state = 19},
  [5518] = {.lex_state = 85, .external_lex_state = 11},
  [5519] = {.lex_state = 420, .external_lex_state = 13},
  [5520] = {.lex_state = 430, .external_lex_state = 17},
  [5521] = {.lex_state = 420, .external_lex_state = 11},
  [5522] = {.lex_state = 420, .external_lex_state = 13},
  [5523] = {.lex_state = 430, .external_lex_state = 17},
  [5524] = {.lex_state = 420, .external_lex_state = 11},
  [5525] = {.lex_state = 420, .external_lex_state = 13},
  [5526] = {.lex_state = 431, .external_lex_state = 17},
  [5527] = {.lex_state = 420, .external_lex_state = 13},
  [5528] = {.lex_state = 431, .external_lex_state = 17},
  [5529] = {.lex_state = 420, .external_lex_state = 13},
  [5530] = {.lex_state = 420, .external_lex_state = 13},
  [5531] = {.lex_state = 420, .external_lex_state = 13},
  [5532] = {.lex_state = 86, .external_lex_state = 11},
  [5533] = {.lex_state = 420, .external_lex_state = 13},
  [5534] = {.lex_state = 446, .external_lex_state = 44},
  [5535] = {.lex_state = 109, .external_lex_state = 11},
  [5536] = {.lex_state = 109, .external_lex_state = 11},
  [5537] = {.lex_state = 446, .external_lex_state = 46},
  [5538] = {.lex_state = 446, .external_lex_state = 46},
  [5539] = {.lex_state = 446, .external_lex_state = 44},
  [5540] = {.lex_state = 418, .external_lex_state = 18},
  [5541] = {.lex_state = 431, .external_lex_state = 14},
  [5542] = {.lex_state = 446, .external_lex_state = 44},
  [5543] = {.lex_state = 418, .external_lex_state = 12},
  [5544] = {.lex_state = 420, .external_lex_state = 13},
  [5545] = {.lex_state = 420, .external_lex_state = 13},
  [5546] = {.lex_state = 420, .external_lex_state = 13},
  [5547] = {.lex_state = 420, .external_lex_state = 13},
  [5548] = {.lex_state = 420, .external_lex_state = 13},
  [5549] = {.lex_state = 446, .external_lex_state = 40},
  [5550] = {.lex_state = 109, .external_lex_state = 11},
  [5551] = {.lex_state = 109, .external_lex_state = 11},
  [5552] = {.lex_state = 431, .external_lex_state = 19},
  [5553] = {.lex_state = 420, .external_lex_state = 13},
  [5554] = {.lex_state = 420, .external_lex_state = 13},
  [5555] = {.lex_state = 431, .external_lex_state = 19},
  [5556] = {.lex_state = 446, .external_lex_state = 41},
  [5557] = {.lex_state = 86, .external_lex_state = 11},
  [5558] = {.lex_state = 431, .external_lex_state = 17},
  [5559] = {.lex_state = 420, .external_lex_state = 13},
  [5560] = {.lex_state = 420, .external_lex_state = 13},
  [5561] = {.lex_state = 418, .external_lex_state = 18},
  [5562] = {.lex_state = 431, .external_lex_state = 14},
  [5563] = {.lex_state = 420, .external_lex_state = 13},
  [5564] = {.lex_state = 420, .external_lex_state = 13},
  [5565] = {.lex_state = 420, .external_lex_state = 13},
  [5566] = {.lex_state = 420, .external_lex_state = 13},
  [5567] = {.lex_state = 420, .external_lex_state = 13},
  [5568] = {.lex_state = 420, .external_lex_state = 13},
  [5569] = {.lex_state = 420, .external_lex_state = 13},
  [5570] = {.lex_state = 420, .external_lex_state = 13},
  [5571] = {.lex_state = 420, .external_lex_state = 13},
  [5572] = {.lex_state = 420, .external_lex_state = 13},
  [5573] = {.lex_state = 420, .external_lex_state = 13},
  [5574] = {.lex_state = 430, .external_lex_state = 11},
  [5575] = {.lex_state = 430, .external_lex_state = 11},
  [5576] = {.lex_state = 418, .external_lex_state = 15},
  [5577] = {.lex_state = 420, .external_lex_state = 13},
  [5578] = {.lex_state = 420, .external_lex_state = 13},
  [5579] = {.lex_state = 420, .external_lex_state = 13},
  [5580] = {.lex_state = 420, .external_lex_state = 13},
  [5581] = {.lex_state = 418, .external_lex_state = 15},
  [5582] = {.lex_state = 420, .external_lex_state = 13},
  [5583] = {.lex_state = 420, .external_lex_state = 13},
  [5584] = {.lex_state = 430, .external_lex_state = 11},
  [5585] = {.lex_state = 420, .external_lex_state = 13},
  [5586] = {.lex_state = 420, .external_lex_state = 13},
  [5587] = {.lex_state = 420, .external_lex_state = 13},
  [5588] = {.lex_state = 420, .external_lex_state = 13},
  [5589] = {.lex_state = 430, .external_lex_state = 11},
  [5590] = {.lex_state = 85, .external_lex_state = 11},
  [5591] = {.lex_state = 420, .external_lex_state = 13},
  [5592] = {.lex_state = 418, .external_lex_state = 15},
  [5593] = {.lex_state = 418, .external_lex_state = 15},
  [5594] = {.lex_state = 418, .external_lex_state = 10},
  [5595] = {.lex_state = 418, .external_lex_state = 15},
  [5596] = {.lex_state = 430, .external_lex_state = 11},
  [5597] = {.lex_state = 431, .external_lex_state = 17},
  [5598] = {.lex_state = 418, .external_lex_state = 15},
  [5599] = {.lex_state = 431, .external_lex_state = 12},
  [5600] = {.lex_state = 448, .external_lex_state = 46},
  [5601] = {.lex_state = 446, .external_lex_state = 41},
  [5602] = {.lex_state = 446, .external_lex_state = 41},
  [5603] = {.lex_state = 420, .external_lex_state = 13},
  [5604] = {.lex_state = 446, .external_lex_state = 39},
  [5605] = {.lex_state = 420, .external_lex_state = 13},
  [5606] = {.lex_state = 420, .external_lex_state = 13},
  [5607] = {.lex_state = 418, .external_lex_state = 15},
  [5608] = {.lex_state = 418, .external_lex_state = 15},
  [5609] = {.lex_state = 418, .external_lex_state = 15},
  [5610] = {.lex_state = 418, .external_lex_state = 15},
  [5611] = {.lex_state = 430, .external_lex_state = 19},
  [5612] = {.lex_state = 446, .external_lex_state = 39},
  [5613] = {.lex_state = 446, .external_lex_state = 39},
  [5614] = {.lex_state = 430, .external_lex_state = 19},
  [5615] = {.lex_state = 430, .external_lex_state = 17},
  [5616] = {.lex_state = 418, .external_lex_state = 15},
  [5617] = {.lex_state = 418, .external_lex_state = 15},
  [5618] = {.lex_state = 446, .external_lex_state = 40},
  [5619] = {.lex_state = 446, .external_lex_state = 40},
  [5620] = {.lex_state = 431, .external_lex_state = 14},
  [5621] = {.lex_state = 420, .external_lex_state = 13},
  [5622] = {.lex_state = 418, .external_lex_state = 14},
  [5623] = {.lex_state = 431, .external_lex_state = 17},
  [5624] = {.lex_state = 446, .external_lex_state = 39},
  [5625] = {.lex_state = 418, .external_lex_state = 15},
  [5626] = {.lex_state = 418, .external_lex_state = 15},
  [5627] = {.lex_state = 418, .external_lex_state = 15},
  [5628] = {.lex_state = 418, .external_lex_state = 15},
  [5629] = {.lex_state = 418, .external_lex_state = 15},
  [5630] = {.lex_state = 418, .external_lex_state = 15},
  [5631] = {.lex_state = 418, .external_lex_state = 15},
  [5632] = {.lex_state = 418, .external_lex_state = 15},
  [5633] = {.lex_state = 418, .external_lex_state = 15},
  [5634] = {.lex_state = 418, .external_lex_state = 15},
  [5635] = {.lex_state = 85, .external_lex_state = 11},
  [5636] = {.lex_state = 418, .external_lex_state = 15},
  [5637] = {.lex_state = 418, .external_lex_state = 15},
  [5638] = {.lex_state = 418, .external_lex_state = 15},
  [5639] = {.lex_state = 430, .external_lex_state = 19},
  [5640] = {.lex_state = 84, .external_lex_state = 11},
  [5641] = {.lex_state = 84, .external_lex_state = 11},
  [5642] = {.lex_state = 418, .external_lex_state = 15},
  [5643] = {.lex_state = 418, .external_lex_state = 15},
  [5644] = {.lex_state = 418, .external_lex_state = 15},
  [5645] = {.lex_state = 431, .external_lex_state = 14},
  [5646] = {.lex_state = 418, .external_lex_state = 15},
  [5647] = {.lex_state = 446, .external_lex_state = 44},
  [5648] = {.lex_state = 446, .external_lex_state = 44},
  [5649] = {.lex_state = 420, .external_lex_state = 11},
  [5650] = {.lex_state = 418, .external_lex_state = 15},
  [5651] = {.lex_state = 431, .external_lex_state = 11},
  [5652] = {.lex_state = 431, .external_lex_state = 11},
  [5653] = {.lex_state = 418, .external_lex_state = 15},
  [5654] = {.lex_state = 430, .external_lex_state = 19},
  [5655] = {.lex_state = 84, .external_lex_state = 11},
  [5656] = {.lex_state = 418, .external_lex_state = 15},
  [5657] = {.lex_state = 431, .external_lex_state = 11},
  [5658] = {.lex_state = 108, .external_lex_state = 14},
  [5659] = {.lex_state = 418, .external_lex_state = 15},
  [5660] = {.lex_state = 108, .external_lex_state = 14},
  [5661] = {.lex_state = 431, .external_lex_state = 17},
  [5662] = {.lex_state = 418, .external_lex_state = 15},
  [5663] = {.lex_state = 418, .external_lex_state = 15},
  [5664] = {.lex_state = 84, .external_lex_state = 11},
  [5665] = {.lex_state = 431, .external_lex_state = 11},
  [5666] = {.lex_state = 431, .external_lex_state = 19},
  [5667] = {.lex_state = 431, .external_lex_state = 12},
  [5668] = {.lex_state = 446, .external_lex_state = 46},
  [5669] = {.lex_state = 446, .external_lex_state = 46},
  [5670] = {.lex_state = 418, .external_lex_state = 15},
  [5671] = {.lex_state = 446, .external_lex_state = 46},
  [5672] = {.lex_state = 446, .external_lex_state = 46},
  [5673] = {.lex_state = 446, .external_lex_state = 46},
  [5674] = {.lex_state = 109, .external_lex_state = 11},
  [5675] = {.lex_state = 108, .external_lex_state = 14},
  [5676] = {.lex_state = 109, .external_lex_state = 11},
  [5677] = {.lex_state = 431, .external_lex_state = 19},
  [5678] = {.lex_state = 109, .external_lex_state = 11},
  [5679] = {.lex_state = 109, .external_lex_state = 11},
  [5680] = {.lex_state = 109, .external_lex_state = 11},
  [5681] = {.lex_state = 431, .external_lex_state = 13},
  [5682] = {.lex_state = 109, .external_lex_state = 11},
  [5683] = {.lex_state = 109, .external_lex_state = 11},
  [5684] = {.lex_state = 109, .external_lex_state = 11},
  [5685] = {.lex_state = 109, .external_lex_state = 11},
  [5686] = {.lex_state = 109, .external_lex_state = 11},
  [5687] = {.lex_state = 109, .external_lex_state = 11},
  [5688] = {.lex_state = 446, .external_lex_state = 44},
  [5689] = {.lex_state = 431, .external_lex_state = 19},
  [5690] = {.lex_state = 446, .external_lex_state = 44},
  [5691] = {.lex_state = 109, .external_lex_state = 11},
  [5692] = {.lex_state = 108, .external_lex_state = 14},
  [5693] = {.lex_state = 108, .external_lex_state = 14},
  [5694] = {.lex_state = 431, .external_lex_state = 13},
  [5695] = {.lex_state = 430, .external_lex_state = 17},
  [5696] = {.lex_state = 108, .external_lex_state = 14},
  [5697] = {.lex_state = 109, .external_lex_state = 11},
  [5698] = {.lex_state = 418, .external_lex_state = 17},
  [5699] = {.lex_state = 86, .external_lex_state = 11},
  [5700] = {.lex_state = 446, .external_lex_state = 44},
  [5701] = {.lex_state = 109, .external_lex_state = 11},
  [5702] = {.lex_state = 109, .external_lex_state = 11},
  [5703] = {.lex_state = 109, .external_lex_state = 11},
  [5704] = {.lex_state = 109, .external_lex_state = 11},
  [5705] = {.lex_state = 109, .external_lex_state = 11},
  [5706] = {.lex_state = 431, .external_lex_state = 13},
  [5707] = {.lex_state = 446, .external_lex_state = 42},
  [5708] = {.lex_state = 109, .external_lex_state = 11},
  [5709] = {.lex_state = 109, .external_lex_state = 11},
  [5710] = {.lex_state = 446, .external_lex_state = 42},
  [5711] = {.lex_state = 418, .external_lex_state = 18},
  [5712] = {.lex_state = 109, .external_lex_state = 11},
  [5713] = {.lex_state = 109, .external_lex_state = 11},
  [5714] = {.lex_state = 109, .external_lex_state = 11},
  [5715] = {.lex_state = 109, .external_lex_state = 11},
  [5716] = {.lex_state = 109, .external_lex_state = 11},
  [5717] = {.lex_state = 109, .external_lex_state = 11},
  [5718] = {.lex_state = 431, .external_lex_state = 11},
  [5719] = {.lex_state = 108, .external_lex_state = 14},
  [5720] = {.lex_state = 418, .external_lex_state = 17},
  [5721] = {.lex_state = 431, .external_lex_state = 12},
  [5722] = {.lex_state = 108, .external_lex_state = 14},
  [5723] = {.lex_state = 108, .external_lex_state = 14},
  [5724] = {.lex_state = 109, .external_lex_state = 11},
  [5725] = {.lex_state = 108, .external_lex_state = 14},
  [5726] = {.lex_state = 108, .external_lex_state = 14},
  [5727] = {.lex_state = 431, .external_lex_state = 11},
  [5728] = {.lex_state = 418, .external_lex_state = 17},
  [5729] = {.lex_state = 430, .external_lex_state = 17},
  [5730] = {.lex_state = 431, .external_lex_state = 11},
  [5731] = {.lex_state = 431, .external_lex_state = 11},
  [5732] = {.lex_state = 109, .external_lex_state = 11},
  [5733] = {.lex_state = 431, .external_lex_state = 11},
  [5734] = {.lex_state = 109, .external_lex_state = 11},
  [5735] = {.lex_state = 109, .external_lex_state = 11},
  [5736] = {.lex_state = 446, .external_lex_state = 44},
  [5737] = {.lex_state = 431, .external_lex_state = 17},
  [5738] = {.lex_state = 109, .external_lex_state = 11},
  [5739] = {.lex_state = 109, .external_lex_state = 11},
  [5740] = {.lex_state = 109, .external_lex_state = 11},
  [5741] = {.lex_state = 109, .external_lex_state = 11},
  [5742] = {.lex_state = 418, .external_lex_state = 10},
  [5743] = {.lex_state = 418, .external_lex_state = 18},
  [5744] = {.lex_state = 109, .external_lex_state = 11},
  [5745] = {.lex_state = 109, .external_lex_state = 11},
  [5746] = {.lex_state = 418, .external_lex_state = 10},
  [5747] = {.lex_state = 430, .external_lex_state = 18},
  [5748] = {.lex_state = 109, .external_lex_state = 11},
  [5749] = {.lex_state = 446, .external_lex_state = 41},
  [5750] = {.lex_state = 431, .external_lex_state = 14},
  [5751] = {.lex_state = 446, .external_lex_state = 42},
  [5752] = {.lex_state = 431, .external_lex_state = 12},
  [5753] = {.lex_state = 431, .external_lex_state = 19},
  [5754] = {.lex_state = 108, .external_lex_state = 14},
  [5755] = {.lex_state = 108, .external_lex_state = 14},
  [5756] = {.lex_state = 418, .external_lex_state = 14},
  [5757] = {.lex_state = 108, .external_lex_state = 14},
  [5758] = {.lex_state = 418, .external_lex_state = 12},
  [5759] = {.lex_state = 431, .external_lex_state = 18},
  [5760] = {.lex_state = 418, .external_lex_state = 10},
  [5761] = {.lex_state = 418, .external_lex_state = 10},
  [5762] = {.lex_state = 431, .external_lex_state = 12},
  [5763] = {.lex_state = 430, .external_lex_state = 14},
  [5764] = {.lex_state = 446, .external_lex_state = 41},
  [5765] = {.lex_state = 85, .external_lex_state = 11},
  [5766] = {.lex_state = 418, .external_lex_state = 13},
  [5767] = {.lex_state = 431, .external_lex_state = 19},
  [5768] = {.lex_state = 418, .external_lex_state = 14},
  [5769] = {.lex_state = 431, .external_lex_state = 12},
  [5770] = {.lex_state = 430, .external_lex_state = 14},
  [5771] = {.lex_state = 430, .external_lex_state = 14},
  [5772] = {.lex_state = 108, .external_lex_state = 14},
  [5773] = {.lex_state = 108, .external_lex_state = 14},
  [5774] = {.lex_state = 431, .external_lex_state = 17},
  [5775] = {.lex_state = 431, .external_lex_state = 18},
  [5776] = {.lex_state = 108, .external_lex_state = 14},
  [5777] = {.lex_state = 418, .external_lex_state = 12},
  [5778] = {.lex_state = 431, .external_lex_state = 19},
  [5779] = {.lex_state = 431, .external_lex_state = 12},
  [5780] = {.lex_state = 108, .external_lex_state = 14},
  [5781] = {.lex_state = 108, .external_lex_state = 14},
  [5782] = {.lex_state = 431, .external_lex_state = 18},
  [5783] = {.lex_state = 108, .external_lex_state = 14},
  [5784] = {.lex_state = 108, .external_lex_state = 14},
  [5785] = {.lex_state = 431, .external_lex_state = 13},
  [5786] = {.lex_state = 431, .external_lex_state = 12},
  [5787] = {.lex_state = 108, .external_lex_state = 14},
  [5788] = {.lex_state = 431, .external_lex_state = 12},
  [5789] = {.lex_state = 431, .external_lex_state = 12},
  [5790] = {.lex_state = 431, .external_lex_state = 13},
  [5791] = {.lex_state = 108, .external_lex_state = 14},
  [5792] = {.lex_state = 108, .external_lex_state = 14},
  [5793] = {.lex_state = 108, .external_lex_state = 14},
  [5794] = {.lex_state = 108, .external_lex_state = 14},
  [5795] = {.lex_state = 108, .external_lex_state = 14},
  [5796] = {.lex_state = 418, .external_lex_state = 19},
  [5797] = {.lex_state = 431, .external_lex_state = 12},
  [5798] = {.lex_state = 418, .external_lex_state = 12},
  [5799] = {.lex_state = 431, .external_lex_state = 19},
  [5800] = {.lex_state = 431, .external_lex_state = 14},
  [5801] = {.lex_state = 431, .external_lex_state = 19},
  [5802] = {.lex_state = 108, .external_lex_state = 14},
  [5803] = {.lex_state = 418, .external_lex_state = 12},
  [5804] = {.lex_state = 431, .external_lex_state = 14},
  [5805] = {.lex_state = 108, .external_lex_state = 14},
  [5806] = {.lex_state = 108, .external_lex_state = 14},
  [5807] = {.lex_state = 446, .external_lex_state = 42},
  [5808] = {.lex_state = 108, .external_lex_state = 14},
  [5809] = {.lex_state = 108, .external_lex_state = 14},
  [5810] = {.lex_state = 431, .external_lex_state = 12},
  [5811] = {.lex_state = 85, .external_lex_state = 11},
  [5812] = {.lex_state = 418, .external_lex_state = 19},
  [5813] = {.lex_state = 108, .external_lex_state = 14},
  [5814] = {.lex_state = 446, .external_lex_state = 44},
  [5815] = {.lex_state = 108, .external_lex_state = 14},
  [5816] = {.lex_state = 108, .external_lex_state = 14},
  [5817] = {.lex_state = 430, .external_lex_state = 14},
  [5818] = {.lex_state = 431, .external_lex_state = 19},
  [5819] = {.lex_state = 108, .external_lex_state = 14},
  [5820] = {.lex_state = 108, .external_lex_state = 14},
  [5821] = {.lex_state = 446, .external_lex_state = 44},
  [5822] = {.lex_state = 108, .external_lex_state = 14},
  [5823] = {.lex_state = 108, .external_lex_state = 14},
  [5824] = {.lex_state = 431, .external_lex_state = 18},
  [5825] = {.lex_state = 85, .external_lex_state = 11},
  [5826] = {.lex_state = 108, .external_lex_state = 14},
  [5827] = {.lex_state = 418, .external_lex_state = 10},
  [5828] = {.lex_state = 85, .external_lex_state = 11},
  [5829] = {.lex_state = 108, .external_lex_state = 14},
  [5830] = {.lex_state = 431, .external_lex_state = 13},
  [5831] = {.lex_state = 431, .external_lex_state = 13},
  [5832] = {.lex_state = 430, .external_lex_state = 14},
  [5833] = {.lex_state = 430, .external_lex_state = 19},
  [5834] = {.lex_state = 418, .external_lex_state = 19},
  [5835] = {.lex_state = 418, .external_lex_state = 10},
  [5836] = {.lex_state = 431, .external_lex_state = 18},
  [5837] = {.lex_state = 418, .external_lex_state = 10},
  [5838] = {.lex_state = 418, .external_lex_state = 10},
  [5839] = {.lex_state = 418, .external_lex_state = 10},
  [5840] = {.lex_state = 109, .external_lex_state = 11},
  [5841] = {.lex_state = 431, .external_lex_state = 14},
  [5842] = {.lex_state = 446, .external_lex_state = 46},
  [5843] = {.lex_state = 431, .external_lex_state = 13},
  [5844] = {.lex_state = 431, .external_lex_state = 13},
  [5845] = {.lex_state = 446, .external_lex_state = 46},
  [5846] = {.lex_state = 108, .external_lex_state = 14},
  [5847] = {.lex_state = 108, .external_lex_state = 14},
  [5848] = {.lex_state = 108, .external_lex_state = 14},
  [5849] = {.lex_state = 108, .external_lex_state = 14},
  [5850] = {.lex_state = 418, .external_lex_state = 10},
  [5851] = {.lex_state = 431, .external_lex_state = 13},
  [5852] = {.lex_state = 431, .external_lex_state = 13},
  [5853] = {.lex_state = 431, .external_lex_state = 14},
  [5854] = {.lex_state = 431, .external_lex_state = 19},
  [5855] = {.lex_state = 418, .external_lex_state = 10},
  [5856] = {.lex_state = 418, .external_lex_state = 10},
  [5857] = {.lex_state = 418, .external_lex_state = 10},
  [5858] = {.lex_state = 418, .external_lex_state = 10},
  [5859] = {.lex_state = 108, .external_lex_state = 14},
  [5860] = {.lex_state = 108, .external_lex_state = 14},
  [5861] = {.lex_state = 430, .external_lex_state = 12},
  [5862] = {.lex_state = 430, .external_lex_state = 12},
  [5863] = {.lex_state = 108, .external_lex_state = 14},
  [5864] = {.lex_state = 108, .external_lex_state = 14},
  [5865] = {.lex_state = 108, .external_lex_state = 14},
  [5866] = {.lex_state = 108, .external_lex_state = 14},
  [5867] = {.lex_state = 431, .external_lex_state = 14},
  [5868] = {.lex_state = 431, .external_lex_state = 13},
  [5869] = {.lex_state = 431, .external_lex_state = 13},
  [5870] = {.lex_state = 418, .external_lex_state = 10},
  [5871] = {.lex_state = 430, .external_lex_state = 12},
  [5872] = {.lex_state = 418, .external_lex_state = 10},
  [5873] = {.lex_state = 418, .external_lex_state = 14},
  [5874] = {.lex_state = 108, .external_lex_state = 14},
  [5875] = {.lex_state = 418, .external_lex_state = 17},
  [5876] = {.lex_state = 431, .external_lex_state = 13},
  [5877] = {.lex_state = 430, .external_lex_state = 12},
  [5878] = {.lex_state = 446, .external_lex_state = 40},
  [5879] = {.lex_state = 446, .external_lex_state = 46},
  [5880] = {.lex_state = 446, .external_lex_state = 46},
  [5881] = {.lex_state = 446, .external_lex_state = 46},
  [5882] = {.lex_state = 418, .external_lex_state = 15},
  [5883] = {.lex_state = 418, .external_lex_state = 12},
  [5884] = {.lex_state = 418, .external_lex_state = 19},
  [5885] = {.lex_state = 86, .external_lex_state = 11},
  [5886] = {.lex_state = 86, .external_lex_state = 11},
  [5887] = {.lex_state = 420, .external_lex_state = 11},
  [5888] = {.lex_state = 418, .external_lex_state = 17},
  [5889] = {.lex_state = 420, .external_lex_state = 11},
  [5890] = {.lex_state = 418, .external_lex_state = 17},
  [5891] = {.lex_state = 420, .external_lex_state = 11},
  [5892] = {.lex_state = 418, .external_lex_state = 14},
  [5893] = {.lex_state = 418, .external_lex_state = 18},
  [5894] = {.lex_state = 418, .external_lex_state = 14},
  [5895] = {.lex_state = 418, .external_lex_state = 12},
  [5896] = {.lex_state = 418, .external_lex_state = 12},
  [5897] = {.lex_state = 420, .external_lex_state = 11},
  [5898] = {.lex_state = 418, .external_lex_state = 19},
  [5899] = {.lex_state = 418, .external_lex_state = 17},
  [5900] = {.lex_state = 418, .external_lex_state = 12},
  [5901] = {.lex_state = 418, .external_lex_state = 12},
  [5902] = {.lex_state = 418, .external_lex_state = 18},
  [5903] = {.lex_state = 418, .external_lex_state = 17},
  [5904] = {.lex_state = 418, .external_lex_state = 17},
  [5905] = {.lex_state = 418, .external_lex_state = 17},
  [5906] = {.lex_state = 420, .external_lex_state = 11},
  [5907] = {.lex_state = 86, .external_lex_state = 11},
  [5908] = {.lex_state = 418, .external_lex_state = 18},
  [5909] = {.lex_state = 418, .external_lex_state = 19},
  [5910] = {.lex_state = 418, .external_lex_state = 17},
  [5911] = {.lex_state = 446, .external_lex_state = 53},
  [5912] = {.lex_state = 446, .external_lex_state = 39},
  [5913] = {.lex_state = 418, .external_lex_state = 17},
  [5914] = {.lex_state = 446, .external_lex_state = 41},
  [5915] = {.lex_state = 418, .external_lex_state = 17},
  [5916] = {.lex_state = 418, .external_lex_state = 18},
  [5917] = {.lex_state = 418, .external_lex_state = 18},
  [5918] = {.lex_state = 446, .external_lex_state = 39},
  [5919] = {.lex_state = 418, .external_lex_state = 17},
  [5920] = {.lex_state = 446, .external_lex_state = 44},
  [5921] = {.lex_state = 446, .external_lex_state = 53},
  [5922] = {.lex_state = 446, .external_lex_state = 39},
  [5923] = {.lex_state = 86, .external_lex_state = 11},
  [5924] = {.lex_state = 446, .external_lex_state = 39},
  [5925] = {.lex_state = 86, .external_lex_state = 11},
  [5926] = {.lex_state = 86, .external_lex_state = 11},
  [5927] = {.lex_state = 86, .external_lex_state = 11},
  [5928] = {.lex_state = 418, .external_lex_state = 17},
  [5929] = {.lex_state = 418, .external_lex_state = 17},
  [5930] = {.lex_state = 446, .external_lex_state = 39},
  [5931] = {.lex_state = 418, .external_lex_state = 14},
  [5932] = {.lex_state = 418, .external_lex_state = 14},
  [5933] = {.lex_state = 418, .external_lex_state = 19},
  [5934] = {.lex_state = 446, .external_lex_state = 39},
  [5935] = {.lex_state = 446, .external_lex_state = 53},
  [5936] = {.lex_state = 418, .external_lex_state = 17},
  [5937] = {.lex_state = 446, .external_lex_state = 44},
  [5938] = {.lex_state = 418, .external_lex_state = 17},
  [5939] = {.lex_state = 418, .external_lex_state = 17},
  [5940] = {.lex_state = 86, .external_lex_state = 11},
  [5941] = {.lex_state = 418, .external_lex_state = 17},
  [5942] = {.lex_state = 418, .external_lex_state = 17},
  [5943] = {.lex_state = 418, .external_lex_state = 17},
  [5944] = {.lex_state = 446, .external_lex_state = 44},
  [5945] = {.lex_state = 418, .external_lex_state = 17},
  [5946] = {.lex_state = 418, .external_lex_state = 14},
  [5947] = {.lex_state = 446, .external_lex_state = 40},
  [5948] = {.lex_state = 446, .external_lex_state = 53},
  [5949] = {.lex_state = 418, .external_lex_state = 14},
  [5950] = {.lex_state = 418, .external_lex_state = 18},
  [5951] = {.lex_state = 418, .external_lex_state = 17},
  [5952] = {.lex_state = 418, .external_lex_state = 17},
  [5953] = {.lex_state = 420, .external_lex_state = 11},
  [5954] = {.lex_state = 446, .external_lex_state = 39},
  [5955] = {.lex_state = 86, .external_lex_state = 11},
  [5956] = {.lex_state = 418, .external_lex_state = 14},
  [5957] = {.lex_state = 446, .external_lex_state = 42},
  [5958] = {.lex_state = 446, .external_lex_state = 39},
  [5959] = {.lex_state = 418, .external_lex_state = 17},
  [5960] = {.lex_state = 418, .external_lex_state = 18},
  [5961] = {.lex_state = 446, .external_lex_state = 39},
  [5962] = {.lex_state = 446, .external_lex_state = 39},
  [5963] = {.lex_state = 446, .external_lex_state = 53},
  [5964] = {.lex_state = 446, .external_lex_state = 39},
  [5965] = {.lex_state = 418, .external_lex_state = 17},
  [5966] = {.lex_state = 446, .external_lex_state = 50},
  [5967] = {.lex_state = 420, .external_lex_state = 11},
  [5968] = {.lex_state = 446, .external_lex_state = 50},
  [5969] = {.lex_state = 418, .external_lex_state = 12},
  [5970] = {.lex_state = 446, .external_lex_state = 44},
  [5971] = {.lex_state = 420, .external_lex_state = 11},
  [5972] = {.lex_state = 418, .external_lex_state = 19},
  [5973] = {.lex_state = 446, .external_lex_state = 46},
  [5974] = {.lex_state = 418, .external_lex_state = 17},
  [5975] = {.lex_state = 446, .external_lex_state = 46},
  [5976] = {.lex_state = 420, .external_lex_state = 11},
  [5977] = {.lex_state = 418, .external_lex_state = 18},
  [5978] = {.lex_state = 418, .external_lex_state = 19},
  [5979] = {.lex_state = 420, .external_lex_state = 11},
  [5980] = {.lex_state = 418, .external_lex_state = 12},
  [5981] = {.lex_state = 418, .external_lex_state = 12},
  [5982] = {.lex_state = 418, .external_lex_state = 18},
  [5983] = {.lex_state = 418, .external_lex_state = 17},
  [5984] = {.lex_state = 418, .external_lex_state = 14},
  [5985] = {.lex_state = 420, .external_lex_state = 11},
  [5986] = {.lex_state = 418, .external_lex_state = 12},
  [5987] = {.lex_state = 418, .external_lex_state = 14},
  [5988] = {.lex_state = 418, .external_lex_state = 12},
  [5989] = {.lex_state = 86, .external_lex_state = 11},
  [5990] = {.lex_state = 86, .external_lex_state = 11},
  [5991] = {.lex_state = 418, .external_lex_state = 19},
  [5992] = {.lex_state = 418, .external_lex_state = 18},
  [5993] = {.lex_state = 420, .external_lex_state = 11},
  [5994] = {.lex_state = 418, .external_lex_state = 17},
  [5995] = {.lex_state = 418, .external_lex_state = 12},
  [5996] = {.lex_state = 420, .external_lex_state = 11},
  [5997] = {.lex_state = 420, .external_lex_state = 11},
  [5998] = {.lex_state = 418, .external_lex_state = 18},
  [5999] = {.lex_state = 418, .external_lex_state = 17},
  [6000] = {.lex_state = 418, .external_lex_state = 14},
  [6001] = {.lex_state = 420, .external_lex_state = 11},
  [6002] = {.lex_state = 418, .external_lex_state = 12},
  [6003] = {.lex_state = 418, .external_lex_state = 17},
  [6004] = {.lex_state = 418, .external_lex_state = 14},
  [6005] = {.lex_state = 418, .external_lex_state = 12},
  [6006] = {.lex_state = 420, .external_lex_state = 11},
  [6007] = {.lex_state = 418, .external_lex_state = 17},
  [6008] = {.lex_state = 418, .external_lex_state = 12},
  [6009] = {.lex_state = 446, .external_lex_state = 50},
  [6010] = {.lex_state = 446, .external_lex_state = 50},
  [6011] = {.lex_state = 446, .external_lex_state = 50},
  [6012] = {.lex_state = 418, .external_lex_state = 12},
  [6013] = {.lex_state = 420, .external_lex_state = 11},
  [6014] = {.lex_state = 446, .external_lex_state = 40},
  [6015] = {.lex_state = 420, .external_lex_state = 11},
  [6016] = {.lex_state = 418, .external_lex_state = 12},
  [6017] = {.lex_state = 418, .external_lex_state = 18},
  [6018] = {.lex_state = 420, .external_lex_state = 11},
  [6019] = {.lex_state = 418, .external_lex_state = 17},
  [6020] = {.lex_state = 418, .external_lex_state = 12},
  [6021] = {.lex_state = 418, .external_lex_state = 14},
  [6022] = {.lex_state = 418, .external_lex_state = 12},
  [6023] = {.lex_state = 418, .external_lex_state = 19},
  [6024] = {.lex_state = 418, .external_lex_state = 19},
  [6025] = {.lex_state = 418, .external_lex_state = 12},
  [6026] = {.lex_state = 418, .external_lex_state = 12},
  [6027] = {.lex_state = 418, .external_lex_state = 12},
  [6028] = {.lex_state = 418, .external_lex_state = 12},
  [6029] = {.lex_state = 418, .external_lex_state = 12},
  [6030] = {.lex_state = 418, .external_lex_state = 12},
  [6031] = {.lex_state = 446, .external_lex_state = 39},
  [6032] = {.lex_state = 418, .external_lex_state = 12},
  [6033] = {.lex_state = 418, .external_lex_state = 13},
  [6034] = {.lex_state = 418, .external_lex_state = 12},
  [6035] = {.lex_state = 418, .external_lex_state = 14},
  [6036] = {.lex_state = 418, .external_lex_state = 19},
  [6037] = {.lex_state = 418, .external_lex_state = 12},
  [6038] = {.lex_state = 418, .external_lex_state = 12},
  [6039] = {.lex_state = 418, .external_lex_state = 14},
  [6040] = {.lex_state = 418, .external_lex_state = 14},
  [6041] = {.lex_state = 446, .external_lex_state = 40},
  [6042] = {.lex_state = 420, .external_lex_state = 11},
  [6043] = {.lex_state = 418, .external_lex_state = 18},
  [6044] = {.lex_state = 446, .external_lex_state = 53},
  [6045] = {.lex_state = 86, .external_lex_state = 11},
  [6046] = {.lex_state = 418, .external_lex_state = 19},
  [6047] = {.lex_state = 418, .external_lex_state = 14},
  [6048] = {.lex_state = 418, .external_lex_state = 19},
  [6049] = {.lex_state = 418, .external_lex_state = 12},
  [6050] = {.lex_state = 420, .external_lex_state = 11},
  [6051] = {.lex_state = 420, .external_lex_state = 11},
  [6052] = {.lex_state = 446, .external_lex_state = 53},
  [6053] = {.lex_state = 420, .external_lex_state = 11},
  [6054] = {.lex_state = 420, .external_lex_state = 11},
  [6055] = {.lex_state = 418, .external_lex_state = 17},
  [6056] = {.lex_state = 418, .external_lex_state = 17},
  [6057] = {.lex_state = 418, .external_lex_state = 12},
  [6058] = {.lex_state = 418, .external_lex_state = 17},
  [6059] = {.lex_state = 418, .external_lex_state = 12},
  [6060] = {.lex_state = 418, .external_lex_state = 12},
  [6061] = {.lex_state = 418, .external_lex_state = 12},
  [6062] = {.lex_state = 418, .external_lex_state = 14},
  [6063] = {.lex_state = 418, .external_lex_state = 19},
  [6064] = {.lex_state = 418, .external_lex_state = 18},
  [6065] = {.lex_state = 418, .external_lex_state = 18},
  [6066] = {.lex_state = 418, .external_lex_state = 17},
  [6067] = {.lex_state = 418, .external_lex_state = 17},
  [6068] = {.lex_state = 420, .external_lex_state = 11},
  [6069] = {.lex_state = 86, .external_lex_state = 11},
  [6070] = {.lex_state = 86, .external_lex_state = 11},
  [6071] = {.lex_state = 418, .external_lex_state = 18},
  [6072] = {.lex_state = 420, .external_lex_state = 11},
  [6073] = {.lex_state = 418, .external_lex_state = 14},
  [6074] = {.lex_state = 418, .external_lex_state = 19},
  [6075] = {.lex_state = 418, .external_lex_state = 19},
  [6076] = {.lex_state = 418, .external_lex_state = 18},
  [6077] = {.lex_state = 418, .external_lex_state = 14},
  [6078] = {.lex_state = 418, .external_lex_state = 17},
  [6079] = {.lex_state = 418, .external_lex_state = 19},
  [6080] = {.lex_state = 418, .external_lex_state = 19},
  [6081] = {.lex_state = 418, .external_lex_state = 18},
  [6082] = {.lex_state = 418, .external_lex_state = 19},
  [6083] = {.lex_state = 418, .external_lex_state = 19},
  [6084] = {.lex_state = 418, .external_lex_state = 19},
  [6085] = {.lex_state = 418, .external_lex_state = 19},
  [6086] = {.lex_state = 418, .external_lex_state = 14},
  [6087] = {.lex_state = 418, .external_lex_state = 19},
  [6088] = {.lex_state = 418, .external_lex_state = 19},
  [6089] = {.lex_state = 446, .external_lex_state = 40},
  [6090] = {.lex_state = 418, .external_lex_state = 14},
  [6091] = {.lex_state = 86, .external_lex_state = 11},
  [6092] = {.lex_state = 86, .external_lex_state = 11},
  [6093] = {.lex_state = 418, .external_lex_state = 19},
  [6094] = {.lex_state = 418, .external_lex_state = 19},
  [6095] = {.lex_state = 418, .external_lex_state = 19},
  [6096] = {.lex_state = 418, .external_lex_state = 19},
  [6097] = {.lex_state = 418, .external_lex_state = 14},
  [6098] = {.lex_state = 418, .external_lex_state = 14},
  [6099] = {.lex_state = 418, .external_lex_state = 17},
  [6100] = {.lex_state = 446, .external_lex_state = 42},
  [6101] = {.lex_state = 418, .external_lex_state = 12},
  [6102] = {.lex_state = 418, .external_lex_state = 19},
  [6103] = {.lex_state = 418, .external_lex_state = 14},
  [6104] = {.lex_state = 446, .external_lex_state = 53},
  [6105] = {.lex_state = 418, .external_lex_state = 19},
  [6106] = {.lex_state = 418, .external_lex_state = 19},
  [6107] = {.lex_state = 446, .external_lex_state = 39},
  [6108] = {.lex_state = 86, .external_lex_state = 11},
  [6109] = {.lex_state = 420, .external_lex_state = 11},
  [6110] = {.lex_state = 420, .external_lex_state = 11},
  [6111] = {.lex_state = 446, .external_lex_state = 39},
  [6112] = {.lex_state = 86, .external_lex_state = 11},
  [6113] = {.lex_state = 418, .external_lex_state = 17},
  [6114] = {.lex_state = 418, .external_lex_state = 18},
  [6115] = {.lex_state = 420, .external_lex_state = 11},
  [6116] = {.lex_state = 86, .external_lex_state = 11},
  [6117] = {.lex_state = 420, .external_lex_state = 11},
  [6118] = {.lex_state = 420, .external_lex_state = 11},
  [6119] = {.lex_state = 86, .external_lex_state = 11},
  [6120] = {.lex_state = 446, .external_lex_state = 53},
  [6121] = {.lex_state = 418, .external_lex_state = 18},
  [6122] = {.lex_state = 418, .external_lex_state = 17},
  [6123] = {.lex_state = 446, .external_lex_state = 42},
  [6124] = {.lex_state = 446, .external_lex_state = 39},
  [6125] = {.lex_state = 420, .external_lex_state = 11},
  [6126] = {.lex_state = 418, .external_lex_state = 19},
  [6127] = {.lex_state = 86, .external_lex_state = 11},
  [6128] = {.lex_state = 86, .external_lex_state = 11},
  [6129] = {.lex_state = 446, .external_lex_state = 41},
  [6130] = {.lex_state = 420, .external_lex_state = 11},
  [6131] = {.lex_state = 418, .external_lex_state = 14},
  [6132] = {.lex_state = 420, .external_lex_state = 11},
  [6133] = {.lex_state = 420, .external_lex_state = 11},
  [6134] = {.lex_state = 420, .external_lex_state = 11},
  [6135] = {.lex_state = 420, .external_lex_state = 11},
  [6136] = {.lex_state = 86, .external_lex_state = 11},
  [6137] = {.lex_state = 86, .external_lex_state = 11},
  [6138] = {.lex_state = 420, .external_lex_state = 11},
  [6139] = {.lex_state = 420, .external_lex_state = 11},
  [6140] = {.lex_state = 86, .external_lex_state = 11},
  [6141] = {.lex_state = 420, .external_lex_state = 11},
  [6142] = {.lex_state = 86, .external_lex_state = 11},
  [6143] = {.lex_state = 418, .external_lex_state = 19},
  [6144] = {.lex_state = 86, .external_lex_state = 11},
  [6145] = {.lex_state = 446, .external_lex_state = 53},
  [6146] = {.lex_state = 86, .external_lex_state = 11},
  [6147] = {.lex_state = 448, .external_lex_state = 51},
  [6148] = {.lex_state = 418, .external_lex_state = 17},
  [6149] = {.lex_state = 418, .external_lex_state = 14},
  [6150] = {.lex_state = 418, .external_lex_state = 19},
  [6151] = {.lex_state = 418, .external_lex_state = 14},
  [6152] = {.lex_state = 420, .external_lex_state = 11},
  [6153] = {.lex_state = 418, .external_lex_state = 19},
  [6154] = {.lex_state = 418, .external_lex_state = 14},
  [6155] = {.lex_state = 420, .external_lex_state = 11},
  [6156] = {.lex_state = 420, .external_lex_state = 11},
  [6157] = {.lex_state = 418, .external_lex_state = 19},
  [6158] = {.lex_state = 420, .external_lex_state = 11},
  [6159] = {.lex_state = 86, .external_lex_state = 11},
  [6160] = {.lex_state = 418, .external_lex_state = 18},
  [6161] = {.lex_state = 418, .external_lex_state = 18},
  [6162] = {.lex_state = 420, .external_lex_state = 11},
  [6163] = {.lex_state = 420, .external_lex_state = 11},
  [6164] = {.lex_state = 86, .external_lex_state = 11},
  [6165] = {.lex_state = 418, .external_lex_state = 18},
  [6166] = {.lex_state = 86, .external_lex_state = 11},
  [6167] = {.lex_state = 86, .external_lex_state = 11},
  [6168] = {.lex_state = 86, .external_lex_state = 11},
  [6169] = {.lex_state = 86, .external_lex_state = 11},
  [6170] = {.lex_state = 86, .external_lex_state = 11},
  [6171] = {.lex_state = 86, .external_lex_state = 11},
  [6172] = {.lex_state = 86, .external_lex_state = 11},
  [6173] = {.lex_state = 418, .external_lex_state = 17},
  [6174] = {.lex_state = 86, .external_lex_state = 11},
  [6175] = {.lex_state = 86, .external_lex_state = 11},
  [6176] = {.lex_state = 86, .external_lex_state = 11},
  [6177] = {.lex_state = 86, .external_lex_state = 11},
  [6178] = {.lex_state = 86, .external_lex_state = 11},
  [6179] = {.lex_state = 420, .external_lex_state = 11},
  [6180] = {.lex_state = 86, .external_lex_state = 11},
  [6181] = {.lex_state = 86, .external_lex_state = 11},
  [6182] = {.lex_state = 420, .external_lex_state = 11},
  [6183] = {.lex_state = 86, .external_lex_state = 11},
  [6184] = {.lex_state = 86, .external_lex_state = 11},
  [6185] = {.lex_state = 418, .external_lex_state = 12},
  [6186] = {.lex_state = 86, .external_lex_state = 11},
  [6187] = {.lex_state = 418, .external_lex_state = 19},
  [6188] = {.lex_state = 418, .external_lex_state = 17},
  [6189] = {.lex_state = 420, .external_lex_state = 11},
  [6190] = {.lex_state = 418, .external_lex_state = 12},
  [6191] = {.lex_state = 420, .external_lex_state = 11},
  [6192] = {.lex_state = 420, .external_lex_state = 11},
  [6193] = {.lex_state = 418, .external_lex_state = 19},
  [6194] = {.lex_state = 86, .external_lex_state = 11},
  [6195] = {.lex_state = 418, .external_lex_state = 17},
  [6196] = {.lex_state = 418, .external_lex_state = 18},
  [6197] = {.lex_state = 418, .external_lex_state = 19},
  [6198] = {.lex_state = 418, .external_lex_state = 18},
  [6199] = {.lex_state = 418, .external_lex_state = 19},
  [6200] = {.lex_state = 86, .external_lex_state = 11},
  [6201] = {.lex_state = 418, .external_lex_state = 17},
  [6202] = {.lex_state = 418, .external_lex_state = 12},
  [6203] = {.lex_state = 418, .external_lex_state = 19},
  [6204] = {.lex_state = 418, .external_lex_state = 19},
  [6205] = {.lex_state = 418, .external_lex_state = 12},
  [6206] = {.lex_state = 420, .external_lex_state = 11},
  [6207] = {.lex_state = 418, .external_lex_state = 14},
  [6208] = {.lex_state = 420, .external_lex_state = 11},
  [6209] = {.lex_state = 86, .external_lex_state = 11},
  [6210] = {.lex_state = 418, .external_lex_state = 17},
  [6211] = {.lex_state = 86, .external_lex_state = 11},
  [6212] = {.lex_state = 418, .external_lex_state = 17},
  [6213] = {.lex_state = 86, .external_lex_state = 11},
  [6214] = {.lex_state = 86, .external_lex_state = 11},
  [6215] = {.lex_state = 418, .external_lex_state = 18},
  [6216] = {.lex_state = 418, .external_lex_state = 12},
  [6217] = {.lex_state = 418, .external_lex_state = 12},
  [6218] = {.lex_state = 448, .external_lex_state = 53},
  [6219] = {.lex_state = 418, .external_lex_state = 12},
  [6220] = {.lex_state = 418, .external_lex_state = 12},
  [6221] = {.lex_state = 418, .external_lex_state = 19},
  [6222] = {.lex_state = 418, .external_lex_state = 19},
  [6223] = {.lex_state = 420, .external_lex_state = 11},
  [6224] = {.lex_state = 86, .external_lex_state = 11},
  [6225] = {.lex_state = 418, .external_lex_state = 19},
  [6226] = {.lex_state = 86, .external_lex_state = 11},
  [6227] = {.lex_state = 446, .external_lex_state = 50},
  [6228] = {.lex_state = 420, .external_lex_state = 11},
  [6229] = {.lex_state = 418, .external_lex_state = 17},
  [6230] = {.lex_state = 418, .external_lex_state = 12},
  [6231] = {.lex_state = 418, .external_lex_state = 17},
  [6232] = {.lex_state = 418, .external_lex_state = 18},
  [6233] = {.lex_state = 418, .external_lex_state = 12},
  [6234] = {.lex_state = 418, .external_lex_state = 17},
  [6235] = {.lex_state = 418, .external_lex_state = 14},
  [6236] = {.lex_state = 418, .external_lex_state = 14},
  [6237] = {.lex_state = 418, .external_lex_state = 17},
  [6238] = {.lex_state = 418, .external_lex_state = 17},
  [6239] = {.lex_state = 418, .external_lex_state = 17},
  [6240] = {.lex_state = 418, .external_lex_state = 19},
  [6241] = {.lex_state = 420, .external_lex_state = 11},
  [6242] = {.lex_state = 418, .external_lex_state = 18},
  [6243] = {.lex_state = 418, .external_lex_state = 18},
  [6244] = {.lex_state = 418, .external_lex_state = 14},
  [6245] = {.lex_state = 418, .external_lex_state = 18},
  [6246] = {.lex_state = 418, .external_lex_state = 17},
  [6247] = {.lex_state = 86, .external_lex_state = 11},
  [6248] = {.lex_state = 420, .external_lex_state = 11},
  [6249] = {.lex_state = 418, .external_lex_state = 19},
  [6250] = {.lex_state = 418, .external_lex_state = 19},
  [6251] = {.lex_state = 418, .external_lex_state = 14},
  [6252] = {.lex_state = 420, .external_lex_state = 11},
  [6253] = {.lex_state = 418, .external_lex_state = 19},
  [6254] = {.lex_state = 418, .external_lex_state = 19},
  [6255] = {.lex_state = 418, .external_lex_state = 18},
  [6256] = {.lex_state = 86, .external_lex_state = 11},
  [6257] = {.lex_state = 418, .external_lex_state = 12},
  [6258] = {.lex_state = 418, .external_lex_state = 12},
  [6259] = {.lex_state = 418, .external_lex_state = 18},
  [6260] = {.lex_state = 418, .external_lex_state = 19},
  [6261] = {.lex_state = 418, .external_lex_state = 19},
  [6262] = {.lex_state = 418, .external_lex_state = 19},
  [6263] = {.lex_state = 418, .external_lex_state = 19},
  [6264] = {.lex_state = 446, .external_lex_state = 46},
  [6265] = {.lex_state = 446, .external_lex_state = 51},
  [6266] = {.lex_state = 446, .external_lex_state = 46},
  [6267] = {.lex_state = 418, .external_lex_state = 17},
  [6268] = {.lex_state = 418, .external_lex_state = 12},
  [6269] = {.lex_state = 418, .external_lex_state = 18},
  [6270] = {.lex_state = 418, .external_lex_state = 12},
  [6271] = {.lex_state = 418, .external_lex_state = 12},
  [6272] = {.lex_state = 418, .external_lex_state = 14},
  [6273] = {.lex_state = 420, .external_lex_state = 11},
  [6274] = {.lex_state = 418, .external_lex_state = 12},
  [6275] = {.lex_state = 418, .external_lex_state = 12},
  [6276] = {.lex_state = 420, .external_lex_state = 11},
  [6277] = {.lex_state = 418, .external_lex_state = 18},
  [6278] = {.lex_state = 418, .external_lex_state = 12},
  [6279] = {.lex_state = 418, .external_lex_state = 19},
  [6280] = {.lex_state = 418, .external_lex_state = 12},
  [6281] = {.lex_state = 418, .external_lex_state = 12},
  [6282] = {.lex_state = 418, .external_lex_state = 12},
  [6283] = {.lex_state = 418, .external_lex_state = 12},
  [6284] = {.lex_state = 418, .external_lex_state = 12},
  [6285] = {.lex_state = 418, .external_lex_state = 18},
  [6286] = {.lex_state = 418, .external_lex_state = 14},
  [6287] = {.lex_state = 446, .external_lex_state = 44},
  [6288] = {.lex_state = 418, .external_lex_state = 14},
  [6289] = {.lex_state = 418, .external_lex_state = 18},
  [6290] = {.lex_state = 418, .external_lex_state = 14},
  [6291] = {.lex_state = 418, .external_lex_state = 12},
  [6292] = {.lex_state = 418, .external_lex_state = 12},
  [6293] = {.lex_state = 418, .external_lex_state = 12},
  [6294] = {.lex_state = 418, .external_lex_state = 12},
  [6295] = {.lex_state = 418, .external_lex_state = 19},
  [6296] = {.lex_state = 418, .external_lex_state = 18},
  [6297] = {.lex_state = 420, .external_lex_state = 11},
  [6298] = {.lex_state = 418, .external_lex_state = 12},
  [6299] = {.lex_state = 418, .external_lex_state = 12},
  [6300] = {.lex_state = 418, .external_lex_state = 19},
  [6301] = {.lex_state = 418, .external_lex_state = 18},
  [6302] = {.lex_state = 418, .external_lex_state = 12},
  [6303] = {.lex_state = 418, .external_lex_state = 12},
  [6304] = {.lex_state = 86, .external_lex_state = 11},
  [6305] = {.lex_state = 418, .external_lex_state = 12},
  [6306] = {.lex_state = 418, .external_lex_state = 12},
  [6307] = {.lex_state = 418, .external_lex_state = 17},
  [6308] = {.lex_state = 446, .external_lex_state = 44},
  [6309] = {.lex_state = 446, .external_lex_state = 51},
  [6310] = {.lex_state = 86, .external_lex_state = 11},
  [6311] = {.lex_state = 446, .external_lex_state = 51},
  [6312] = {.lex_state = 418, .external_lex_state = 12},
  [6313] = {.lex_state = 446, .external_lex_state = 51},
  [6314] = {.lex_state = 446, .external_lex_state = 51},
  [6315] = {.lex_state = 418, .external_lex_state = 14},
  [6316] = {.lex_state = 446, .external_lex_state = 51},
  [6317] = {.lex_state = 420, .external_lex_state = 11},
  [6318] = {.lex_state = 446, .external_lex_state = 42},
  [6319] = {.lex_state = 446, .external_lex_state = 50},
  [6320] = {.lex_state = 418, .external_lex_state = 13},
  [6321] = {.lex_state = 418, .external_lex_state = 14},
  [6322] = {.lex_state = 418, .external_lex_state = 13},
  [6323] = {.lex_state = 446, .external_lex_state = 51},
  [6324] = {.lex_state = 418, .external_lex_state = 19},
  [6325] = {.lex_state = 418, .external_lex_state = 19},
  [6326] = {.lex_state = 418, .external_lex_state = 19},
  [6327] = {.lex_state = 446, .external_lex_state = 51},
  [6328] = {.lex_state = 418, .external_lex_state = 17},
  [6329] = {.lex_state = 420, .external_lex_state = 11},
  [6330] = {.lex_state = 420, .external_lex_state = 11},
  [6331] = {.lex_state = 418, .external_lex_state = 19},
  [6332] = {.lex_state = 418, .external_lex_state = 17},
  [6333] = {.lex_state = 418, .external_lex_state = 18},
  [6334] = {.lex_state = 418, .external_lex_state = 19},
  [6335] = {.lex_state = 418, .external_lex_state = 14},
  [6336] = {.lex_state = 418, .external_lex_state = 14},
  [6337] = {.lex_state = 86, .external_lex_state = 11},
  [6338] = {.lex_state = 418, .external_lex_state = 14},
  [6339] = {.lex_state = 418, .external_lex_state = 17},
  [6340] = {.lex_state = 86, .external_lex_state = 11},
  [6341] = {.lex_state = 418, .external_lex_state = 14},
  [6342] = {.lex_state = 418, .external_lex_state = 19},
  [6343] = {.lex_state = 420, .external_lex_state = 11},
  [6344] = {.lex_state = 418, .external_lex_state = 14},
  [6345] = {.lex_state = 86, .external_lex_state = 11},
  [6346] = {.lex_state = 86, .external_lex_state = 11},
  [6347] = {.lex_state = 418, .external_lex_state = 19},
  [6348] = {.lex_state = 418, .external_lex_state = 19},
  [6349] = {.lex_state = 418, .external_lex_state = 14},
  [6350] = {.lex_state = 418, .external_lex_state = 19},
  [6351] = {.lex_state = 418, .external_lex_state = 14},
  [6352] = {.lex_state = 418, .external_lex_state = 14},
  [6353] = {.lex_state = 418, .external_lex_state = 18},
  [6354] = {.lex_state = 418, .external_lex_state = 17},
  [6355] = {.lex_state = 420, .external_lex_state = 11},
  [6356] = {.lex_state = 418, .external_lex_state = 14},
  [6357] = {.lex_state = 418, .external_lex_state = 18},
  [6358] = {.lex_state = 418, .external_lex_state = 18},
  [6359] = {.lex_state = 418, .external_lex_state = 18},
  [6360] = {.lex_state = 446, .external_lex_state = 41},
  [6361] = {.lex_state = 446, .external_lex_state = 51},
  [6362] = {.lex_state = 418, .external_lex_state = 19},
  [6363] = {.lex_state = 418, .external_lex_state = 18},
  [6364] = {.lex_state = 420, .external_lex_state = 11},
  [6365] = {.lex_state = 418, .external_lex_state = 18},
  [6366] = {.lex_state = 418, .external_lex_state = 14},
  [6367] = {.lex_state = 446, .external_lex_state = 51},
  [6368] = {.lex_state = 418, .external_lex_state = 17},
  [6369] = {.lex_state = 418, .external_lex_state = 18},
  [6370] = {.lex_state = 418, .external_lex_state = 14},
  [6371] = {.lex_state = 418, .external_lex_state = 14},
  [6372] = {.lex_state = 418, .external_lex_state = 18},
  [6373] = {.lex_state = 446, .external_lex_state = 53},
  [6374] = {.lex_state = 418, .external_lex_state = 18},
  [6375] = {.lex_state = 446, .external_lex_state = 51},
  [6376] = {.lex_state = 418, .external_lex_state = 18},
  [6377] = {.lex_state = 418, .external_lex_state = 18},
  [6378] = {.lex_state = 418, .external_lex_state = 18},
  [6379] = {.lex_state = 86, .external_lex_state = 11},
  [6380] = {.lex_state = 446, .external_lex_state = 50},
  [6381] = {.lex_state = 418, .external_lex_state = 17},
  [6382] = {.lex_state = 418, .external_lex_state = 19},
  [6383] = {.lex_state = 418, .external_lex_state = 14},
  [6384] = {.lex_state = 420, .external_lex_state = 11},
  [6385] = {.lex_state = 418, .external_lex_state = 14},
  [6386] = {.lex_state = 418, .external_lex_state = 17},
  [6387] = {.lex_state = 418, .external_lex_state = 14},
  [6388] = {.lex_state = 446, .external_lex_state = 41},
  [6389] = {.lex_state = 418, .external_lex_state = 14},
  [6390] = {.lex_state = 418, .external_lex_state = 14},
  [6391] = {.lex_state = 418, .external_lex_state = 18},
  [6392] = {.lex_state = 418, .external_lex_state = 18},
  [6393] = {.lex_state = 418, .external_lex_state = 18},
  [6394] = {.lex_state = 418, .external_lex_state = 18},
  [6395] = {.lex_state = 418, .external_lex_state = 18},
  [6396] = {.lex_state = 418, .external_lex_state = 14},
  [6397] = {.lex_state = 418, .external_lex_state = 18},
  [6398] = {.lex_state = 418, .external_lex_state = 18},
  [6399] = {.lex_state = 446, .external_lex_state = 50},
  [6400] = {.lex_state = 418, .external_lex_state = 18},
  [6401] = {.lex_state = 418, .external_lex_state = 18},
  [6402] = {.lex_state = 418, .external_lex_state = 14},
  [6403] = {.lex_state = 446, .external_lex_state = 50},
  [6404] = {.lex_state = 446, .external_lex_state = 50},
  [6405] = {.lex_state = 418, .external_lex_state = 14},
  [6406] = {.lex_state = 418, .external_lex_state = 14},
  [6407] = {.lex_state = 448, .external_lex_state = 50},
  [6408] = {.lex_state = 446, .external_lex_state = 50},
  [6409] = {.lex_state = 418, .external_lex_state = 18},
  [6410] = {.lex_state = 86, .external_lex_state = 11},
  [6411] = {.lex_state = 418, .external_lex_state = 14},
  [6412] = {.lex_state = 418, .external_lex_state = 17},
  [6413] = {.lex_state = 446, .external_lex_state = 41},
  [6414] = {.lex_state = 418, .external_lex_state = 17},
  [6415] = {.lex_state = 418, .external_lex_state = 18},
  [6416] = {.lex_state = 446, .external_lex_state = 46},
  [6417] = {.lex_state = 418, .external_lex_state = 14},
  [6418] = {.lex_state = 418, .external_lex_state = 17},
  [6419] = {.lex_state = 418, .external_lex_state = 17},
  [6420] = {.lex_state = 418, .external_lex_state = 17},
  [6421] = {.lex_state = 418, .external_lex_state = 14},
  [6422] = {.lex_state = 418, .external_lex_state = 18},
  [6423] = {.lex_state = 446, .external_lex_state = 42},
  [6424] = {.lex_state = 446, .external_lex_state = 40},
  [6425] = {.lex_state = 418, .external_lex_state = 18},
  [6426] = {.lex_state = 418, .external_lex_state = 18},
  [6427] = {.lex_state = 418, .external_lex_state = 19},
  [6428] = {.lex_state = 418, .external_lex_state = 18},
  [6429] = {.lex_state = 418, .external_lex_state = 18},
  [6430] = {.lex_state = 418, .external_lex_state = 14},
  [6431] = {.lex_state = 446, .external_lex_state = 53},
  [6432] = {.lex_state = 418, .external_lex_state = 18},
  [6433] = {.lex_state = 418, .external_lex_state = 17},
  [6434] = {.lex_state = 418, .external_lex_state = 14},
  [6435] = {.lex_state = 86, .external_lex_state = 11},
  [6436] = {.lex_state = 446, .external_lex_state = 51},
  [6437] = {.lex_state = 418, .external_lex_state = 14},
  [6438] = {.lex_state = 86, .external_lex_state = 11},
  [6439] = {.lex_state = 418, .external_lex_state = 14},
  [6440] = {.lex_state = 86, .external_lex_state = 11},
  [6441] = {.lex_state = 418, .external_lex_state = 14},
  [6442] = {.lex_state = 418, .external_lex_state = 12},
  [6443] = {.lex_state = 446, .external_lex_state = 50},
  [6444] = {.lex_state = 446, .external_lex_state = 42},
  [6445] = {.lex_state = 179, .external_lex_state = 55},
  [6446] = {.lex_state = 446, .external_lex_state = 41},
  [6447] = {.lex_state = 446, .external_lex_state = 44},
  [6448] = {.lex_state = 446, .external_lex_state = 40},
  [6449] = {.lex_state = 446, .external_lex_state = 54},
  [6450] = {.lex_state = 446, .external_lex_state = 39},
  [6451] = {.lex_state = 446, .external_lex_state = 39},
  [6452] = {.lex_state = 179, .external_lex_state = 55},
  [6453] = {.lex_state = 179, .external_lex_state = 55},
  [6454] = {.lex_state = 446, .external_lex_state = 53},
  [6455] = {.lex_state = 446, .external_lex_state = 53},
  [6456] = {.lex_state = 446, .external_lex_state = 40},
  [6457] = {.lex_state = 446, .external_lex_state = 42},
  [6458] = {.lex_state = 165, .external_lex_state = 55},
  [6459] = {.lex_state = 446, .external_lex_state = 46},
  [6460] = {.lex_state = 446, .external_lex_state = 51},
  [6461] = {.lex_state = 446, .external_lex_state = 41},
  [6462] = {.lex_state = 446, .external_lex_state = 46},
  [6463] = {.lex_state = 446, .external_lex_state = 54},
  [6464] = {.lex_state = 446, .external_lex_state = 44},
  [6465] = {.lex_state = 446, .external_lex_state = 40},
  [6466] = {.lex_state = 448, .external_lex_state = 54},
  [6467] = {.lex_state = 446, .external_lex_state = 46},
  [6468] = {.lex_state = 446, .external_lex_state = 51},
  [6469] = {.lex_state = 446, .external_lex_state = 42},
  [6470] = {.lex_state = 446, .external_lex_state = 42},
  [6471] = {.lex_state = 446, .external_lex_state = 42},
  [6472] = {.lex_state = 446, .external_lex_state = 40},
  [6473] = {.lex_state = 446, .external_lex_state = 40},
  [6474] = {.lex_state = 446, .external_lex_state = 40},
  [6475] = {.lex_state = 446, .external_lex_state = 41},
  [6476] = {.lex_state = 446, .external_lex_state = 41},
  [6477] = {.lex_state = 446, .external_lex_state = 42},
  [6478] = {.lex_state = 446, .external_lex_state = 41},
  [6479] = {.lex_state = 446, .external_lex_state = 40},
  [6480] = {.lex_state = 179, .external_lex_state = 55},
  [6481] = {.lex_state = 446, .external_lex_state = 50},
  [6482] = {.lex_state = 446, .external_lex_state = 54},
  [6483] = {.lex_state = 446, .external_lex_state = 51},
  [6484] = {.lex_state = 446, .external_lex_state = 54},
  [6485] = {.lex_state = 446, .external_lex_state = 42},
  [6486] = {.lex_state = 446, .external_lex_state = 46},
  [6487] = {.lex_state = 446, .external_lex_state = 40},
  [6488] = {.lex_state = 446, .external_lex_state = 40},
  [6489] = {.lex_state = 446, .external_lex_state = 51},
  [6490] = {.lex_state = 179, .external_lex_state = 55},
  [6491] = {.lex_state = 446, .external_lex_state = 50},
  [6492] = {.lex_state = 446, .external_lex_state = 41},
  [6493] = {.lex_state = 446, .external_lex_state = 41},
  [6494] = {.lex_state = 179, .external_lex_state = 55},
  [6495] = {.lex_state = 446, .external_lex_state = 44},
  [6496] = {.lex_state = 446, .external_lex_state = 40},
  [6497] = {.lex_state = 446, .external_lex_state = 41},
  [6498] = {.lex_state = 446, .external_lex_state = 50},
  [6499] = {.lex_state = 446, .external_lex_state = 40},
  [6500] = {.lex_state = 446, .external_lex_state = 40},
  [6501] = {.lex_state = 449, .external_lex_state = 46},
  [6502] = {.lex_state = 179, .external_lex_state = 55},
  [6503] = {.lex_state = 446, .external_lex_state = 41},
  [6504] = {.lex_state = 446, .external_lex_state = 44},
  [6505] = {.lex_state = 446, .external_lex_state = 42},
  [6506] = {.lex_state = 446, .external_lex_state = 54},
  [6507] = {.lex_state = 446, .external_lex_state = 40},
  [6508] = {.lex_state = 446, .external_lex_state = 42},
  [6509] = {.lex_state = 446, .external_lex_state = 41},
  [6510] = {.lex_state = 446, .external_lex_state = 54},
  [6511] = {.lex_state = 446, .external_lex_state = 42},
  [6512] = {.lex_state = 446, .external_lex_state = 54},
  [6513] = {.lex_state = 446, .external_lex_state = 53},
  [6514] = {.lex_state = 179, .external_lex_state = 55},
  [6515] = {.lex_state = 446, .external_lex_state = 46},
  [6516] = {.lex_state = 446, .external_lex_state = 42},
  [6517] = {.lex_state = 179, .external_lex_state = 55},
  [6518] = {.lex_state = 446, .external_lex_state = 40},
  [6519] = {.lex_state = 446, .external_lex_state = 42},
  [6520] = {.lex_state = 446, .external_lex_state = 41},
  [6521] = {.lex_state = 446, .external_lex_state = 53},
  [6522] = {.lex_state = 446, .external_lex_state = 42},
  [6523] = {.lex_state = 446, .external_lex_state = 41},
  [6524] = {.lex_state = 446, .external_lex_state = 44},
  [6525] = {.lex_state = 179, .external_lex_state = 55},
  [6526] = {.lex_state = 446, .external_lex_state = 54},
  [6527] = {.lex_state = 179, .external_lex_state = 55},
  [6528] = {.lex_state = 446, .external_lex_state = 54},
  [6529] = {.lex_state = 179, .external_lex_state = 55},
  [6530] = {.lex_state = 446, .external_lex_state = 42},
  [6531] = {.lex_state = 446, .external_lex_state = 41},
  [6532] = {.lex_state = 446, .external_lex_state = 40},
  [6533] = {.lex_state = 179, .external_lex_state = 55},
  [6534] = {.lex_state = 446, .external_lex_state = 51},
  [6535] = {.lex_state = 446, .external_lex_state = 53},
  [6536] = {.lex_state = 446, .external_lex_state = 54},
  [6537] = {.lex_state = 446, .external_lex_state = 54},
  [6538] = {.lex_state = 446, .external_lex_state = 41},
  [6539] = {.lex_state = 444, .external_lex_state = 39},
  [6540] = {.lex_state = 446, .external_lex_state = 54},
  [6541] = {.lex_state = 446, .external_lex_state = 50},
  [6542] = {.lex_state = 448, .external_lex_state = 53},
  [6543] = {.lex_state = 446, .external_lex_state = 41},
  [6544] = {.lex_state = 446, .external_lex_state = 42},
  [6545] = {.lex_state = 165, .external_lex_state = 55},
  [6546] = {.lex_state = 449, .external_lex_state = 50},
  [6547] = {.lex_state = 449, .external_lex_state = 53},
  [6548] = {.lex_state = 446, .external_lex_state = 42},
  [6549] = {.lex_state = 446, .external_lex_state = 42},
  [6550] = {.lex_state = 446, .external_lex_state = 44},
  [6551] = {.lex_state = 446, .external_lex_state = 40},
  [6552] = {.lex_state = 449, .external_lex_state = 51},
  [6553] = {.lex_state = 446, .external_lex_state = 40},
  [6554] = {.lex_state = 446, .external_lex_state = 53},
  [6555] = {.lex_state = 446, .external_lex_state = 53},
  [6556] = {.lex_state = 446, .external_lex_state = 53},
  [6557] = {.lex_state = 444, .external_lex_state = 39},
  [6558] = {.lex_state = 446, .external_lex_state = 53},
  [6559] = {.lex_state = 444, .external_lex_state = 39},
  [6560] = {.lex_state = 444, .external_lex_state = 39},
  [6561] = {.lex_state = 154, .external_lex_state = 56},
  [6562] = {.lex_state = 446, .external_lex_state = 50},
  [6563] = {.lex_state = 446, .external_lex_state = 50},
  [6564] = {.lex_state = 446, .external_lex_state = 50},
  [6565] = {.lex_state = 446, .external_lex_state = 44},
  [6566] = {.lex_state = 446, .external_lex_state = 44},
  [6567] = {.lex_state = 446, .external_lex_state = 44},
  [6568] = {.lex_state = 446, .external_lex_state = 44},
  [6569] = {.lex_state = 446, .external_lex_state = 44},
  [6570] = {.lex_state = 446, .external_lex_state = 44},
  [6571] = {.lex_state = 446, .external_lex_state = 44},
  [6572] = {.lex_state = 446, .external_lex_state = 44},
  [6573] = {.lex_state = 446, .external_lex_state = 44},
  [6574] = {.lex_state = 446, .external_lex_state = 44},
  [6575] = {.lex_state = 446, .external_lex_state = 44},
  [6576] = {.lex_state = 446, .external_lex_state = 53},
  [6577] = {.lex_state = 154, .external_lex_state = 56},
  [6578] = {.lex_state = 446, .external_lex_state = 44},
  [6579] = {.lex_state = 446, .external_lex_state = 44},
  [6580] = {.lex_state = 446, .external_lex_state = 44},
  [6581] = {.lex_state = 446, .external_lex_state = 46},
  [6582] = {.lex_state = 154, .external_lex_state = 57},
  [6583] = {.lex_state = 446, .external_lex_state = 46},
  [6584] = {.lex_state = 446, .external_lex_state = 46},
  [6585] = {.lex_state = 446, .external_lex_state = 46},
  [6586] = {.lex_state = 446, .external_lex_state = 46},
  [6587] = {.lex_state = 446, .external_lex_state = 46},
  [6588] = {.lex_state = 446, .external_lex_state = 46},
  [6589] = {.lex_state = 446, .external_lex_state = 46},
  [6590] = {.lex_state = 446, .external_lex_state = 46},
  [6591] = {.lex_state = 446, .external_lex_state = 46},
  [6592] = {.lex_state = 446, .external_lex_state = 51},
  [6593] = {.lex_state = 446, .external_lex_state = 51},
  [6594] = {.lex_state = 446, .external_lex_state = 51},
  [6595] = {.lex_state = 446, .external_lex_state = 46},
  [6596] = {.lex_state = 446, .external_lex_state = 46},
  [6597] = {.lex_state = 446, .external_lex_state = 46},
  [6598] = {.lex_state = 446, .external_lex_state = 50},
  [6599] = {.lex_state = 444, .external_lex_state = 39},
  [6600] = {.lex_state = 444, .external_lex_state = 39},
  [6601] = {.lex_state = 446, .external_lex_state = 54},
  [6602] = {.lex_state = 446, .external_lex_state = 54},
  [6603] = {.lex_state = 446, .external_lex_state = 51},
  [6604] = {.lex_state = 446, .external_lex_state = 54},
  [6605] = {.lex_state = 446, .external_lex_state = 50},
  [6606] = {.lex_state = 446, .external_lex_state = 54},
  [6607] = {.lex_state = 446, .external_lex_state = 54},
  [6608] = {.lex_state = 446, .external_lex_state = 51},
  [6609] = {.lex_state = 165, .external_lex_state = 55},
  [6610] = {.lex_state = 444, .external_lex_state = 40},
  [6611] = {.lex_state = 444, .external_lex_state = 42},
  [6612] = {.lex_state = 446, .external_lex_state = 41},
  [6613] = {.lex_state = 446, .external_lex_state = 41},
  [6614] = {.lex_state = 154, .external_lex_state = 57},
  [6615] = {.lex_state = 444, .external_lex_state = 41},
  [6616] = {.lex_state = 165, .external_lex_state = 55},
  [6617] = {.lex_state = 446, .external_lex_state = 46},
  [6618] = {.lex_state = 444, .external_lex_state = 40},
  [6619] = {.lex_state = 446, .external_lex_state = 50},
  [6620] = {.lex_state = 446, .external_lex_state = 50},
  [6621] = {.lex_state = 446, .external_lex_state = 53},
  [6622] = {.lex_state = 446, .external_lex_state = 53},
  [6623] = {.lex_state = 154, .external_lex_state = 56},
  [6624] = {.lex_state = 446, .external_lex_state = 44},
  [6625] = {.lex_state = 444, .external_lex_state = 39},
  [6626] = {.lex_state = 446, .external_lex_state = 44},
  [6627] = {.lex_state = 444, .external_lex_state = 42},
  [6628] = {.lex_state = 154, .external_lex_state = 56},
  [6629] = {.lex_state = 444, .external_lex_state = 40},
  [6630] = {.lex_state = 446, .external_lex_state = 50},
  [6631] = {.lex_state = 444, .external_lex_state = 39},
  [6632] = {.lex_state = 446, .external_lex_state = 50},
  [6633] = {.lex_state = 444, .external_lex_state = 40},
  [6634] = {.lex_state = 446, .external_lex_state = 54},
  [6635] = {.lex_state = 446, .external_lex_state = 54},
  [6636] = {.lex_state = 446, .external_lex_state = 54},
  [6637] = {.lex_state = 446, .external_lex_state = 51},
  [6638] = {.lex_state = 446, .external_lex_state = 50},
  [6639] = {.lex_state = 446, .external_lex_state = 51},
  [6640] = {.lex_state = 446, .external_lex_state = 51},
  [6641] = {.lex_state = 446, .external_lex_state = 51},
  [6642] = {.lex_state = 446, .external_lex_state = 51},
  [6643] = {.lex_state = 446, .external_lex_state = 51},
  [6644] = {.lex_state = 446, .external_lex_state = 51},
  [6645] = {.lex_state = 446, .external_lex_state = 50},
  [6646] = {.lex_state = 444, .external_lex_state = 40},
  [6647] = {.lex_state = 446, .external_lex_state = 51},
  [6648] = {.lex_state = 446, .external_lex_state = 51},
  [6649] = {.lex_state = 446, .external_lex_state = 51},
  [6650] = {.lex_state = 446, .external_lex_state = 51},
  [6651] = {.lex_state = 446, .external_lex_state = 50},
  [6652] = {.lex_state = 444, .external_lex_state = 39},
  [6653] = {.lex_state = 446, .external_lex_state = 53},
  [6654] = {.lex_state = 446, .external_lex_state = 50},
  [6655] = {.lex_state = 446, .external_lex_state = 50},
  [6656] = {.lex_state = 154, .external_lex_state = 55},
  [6657] = {.lex_state = 446, .external_lex_state = 50},
  [6658] = {.lex_state = 444, .external_lex_state = 39},
  [6659] = {.lex_state = 446, .external_lex_state = 51},
  [6660] = {.lex_state = 446, .external_lex_state = 51},
  [6661] = {.lex_state = 446, .external_lex_state = 51},
  [6662] = {.lex_state = 446, .external_lex_state = 53},
  [6663] = {.lex_state = 444, .external_lex_state = 39},
  [6664] = {.lex_state = 446, .external_lex_state = 54},
  [6665] = {.lex_state = 446, .external_lex_state = 50},
  [6666] = {.lex_state = 446, .external_lex_state = 50},
  [6667] = {.lex_state = 446, .external_lex_state = 50},
  [6668] = {.lex_state = 446, .external_lex_state = 53},
  [6669] = {.lex_state = 446, .external_lex_state = 53},
  [6670] = {.lex_state = 446, .external_lex_state = 53},
  [6671] = {.lex_state = 446, .external_lex_state = 54},
  [6672] = {.lex_state = 446, .external_lex_state = 53},
  [6673] = {.lex_state = 154, .external_lex_state = 58},
  [6674] = {.lex_state = 446, .external_lex_state = 53},
  [6675] = {.lex_state = 444, .external_lex_state = 44},
  [6676] = {.lex_state = 154, .external_lex_state = 55},
  [6677] = {.lex_state = 446, .external_lex_state = 53},
  [6678] = {.lex_state = 444, .external_lex_state = 40},
  [6679] = {.lex_state = 449, .external_lex_state = 54},
  [6680] = {.lex_state = 154, .external_lex_state = 58},
  [6681] = {.lex_state = 444, .external_lex_state = 42},
  [6682] = {.lex_state = 444, .external_lex_state = 42},
  [6683] = {.lex_state = 444, .external_lex_state = 42},
  [6684] = {.lex_state = 446, .external_lex_state = 53},
  [6685] = {.lex_state = 446, .external_lex_state = 53},
  [6686] = {.lex_state = 154, .external_lex_state = 57},
  [6687] = {.lex_state = 446, .external_lex_state = 53},
  [6688] = {.lex_state = 444, .external_lex_state = 42},
  [6689] = {.lex_state = 444, .external_lex_state = 41},
  [6690] = {.lex_state = 444, .external_lex_state = 41},
  [6691] = {.lex_state = 444, .external_lex_state = 41},
  [6692] = {.lex_state = 449, .external_lex_state = 53},
  [6693] = {.lex_state = 446, .external_lex_state = 53},
  [6694] = {.lex_state = 444, .external_lex_state = 42},
  [6695] = {.lex_state = 154, .external_lex_state = 57},
  [6696] = {.lex_state = 444, .external_lex_state = 41},
  [6697] = {.lex_state = 444, .external_lex_state = 41},
  [6698] = {.lex_state = 446, .external_lex_state = 50},
  [6699] = {.lex_state = 444, .external_lex_state = 39},
  [6700] = {.lex_state = 179, .external_lex_state = 55},
  [6701] = {.lex_state = 179, .external_lex_state = 55},
  [6702] = {.lex_state = 179, .external_lex_state = 59},
  [6703] = {.lex_state = 154, .external_lex_state = 55},
  [6704] = {.lex_state = 444, .external_lex_state = 42},
  [6705] = {.lex_state = 446, .external_lex_state = 54},
  [6706] = {.lex_state = 179, .external_lex_state = 55},
  [6707] = {.lex_state = 179, .external_lex_state = 55},
  [6708] = {.lex_state = 179, .external_lex_state = 59},
  [6709] = {.lex_state = 444, .external_lex_state = 44},
  [6710] = {.lex_state = 444, .external_lex_state = 44},
  [6711] = {.lex_state = 179, .external_lex_state = 55},
  [6712] = {.lex_state = 179, .external_lex_state = 55},
  [6713] = {.lex_state = 179, .external_lex_state = 59},
  [6714] = {.lex_state = 179, .external_lex_state = 55},
  [6715] = {.lex_state = 179, .external_lex_state = 55},
  [6716] = {.lex_state = 179, .external_lex_state = 55},
  [6717] = {.lex_state = 179, .external_lex_state = 59},
  [6718] = {.lex_state = 444, .external_lex_state = 41},
  [6719] = {.lex_state = 179, .external_lex_state = 55},
  [6720] = {.lex_state = 179, .external_lex_state = 59},
  [6721] = {.lex_state = 179, .external_lex_state = 55},
  [6722] = {.lex_state = 179, .external_lex_state = 55},
  [6723] = {.lex_state = 179, .external_lex_state = 59},
  [6724] = {.lex_state = 179, .external_lex_state = 55},
  [6725] = {.lex_state = 179, .external_lex_state = 55},
  [6726] = {.lex_state = 179, .external_lex_state = 59},
  [6727] = {.lex_state = 179, .external_lex_state = 55},
  [6728] = {.lex_state = 179, .external_lex_state = 55},
  [6729] = {.lex_state = 179, .external_lex_state = 59},
  [6730] = {.lex_state = 179, .external_lex_state = 55},
  [6731] = {.lex_state = 179, .external_lex_state = 55},
  [6732] = {.lex_state = 179, .external_lex_state = 59},
  [6733] = {.lex_state = 179, .external_lex_state = 55},
  [6734] = {.lex_state = 179, .external_lex_state = 55},
  [6735] = {.lex_state = 179, .external_lex_state = 59},
  [6736] = {.lex_state = 179, .external_lex_state = 55},
  [6737] = {.lex_state = 179, .external_lex_state = 55},
  [6738] = {.lex_state = 179, .external_lex_state = 59},
  [6739] = {.lex_state = 179, .external_lex_state = 55},
  [6740] = {.lex_state = 179, .external_lex_state = 55},
  [6741] = {.lex_state = 179, .external_lex_state = 59},
  [6742] = {.lex_state = 179, .external_lex_state = 55},
  [6743] = {.lex_state = 179, .external_lex_state = 55},
  [6744] = {.lex_state = 154, .external_lex_state = 55},
  [6745] = {.lex_state = 179, .external_lex_state = 55},
  [6746] = {.lex_state = 179, .external_lex_state = 55},
  [6747] = {.lex_state = 179, .external_lex_state = 55},
  [6748] = {.lex_state = 179, .external_lex_state = 55},
  [6749] = {.lex_state = 179, .external_lex_state = 55},
  [6750] = {.lex_state = 179, .external_lex_state = 55},
  [6751] = {.lex_state = 179, .external_lex_state = 55},
  [6752] = {.lex_state = 179, .external_lex_state = 55},
  [6753] = {.lex_state = 446, .external_lex_state = 54},
  [6754] = {.lex_state = 172, .external_lex_state = 55},
  [6755] = {.lex_state = 179, .external_lex_state = 55},
  [6756] = {.lex_state = 154, .external_lex_state = 58},
  [6757] = {.lex_state = 446, .external_lex_state = 54},
  [6758] = {.lex_state = 179, .external_lex_state = 55},
  [6759] = {.lex_state = 179, .external_lex_state = 55},
  [6760] = {.lex_state = 444, .external_lex_state = 40},
  [6761] = {.lex_state = 179, .external_lex_state = 55},
  [6762] = {.lex_state = 179, .external_lex_state = 55},
  [6763] = {.lex_state = 179, .external_lex_state = 55},
  [6764] = {.lex_state = 179, .external_lex_state = 55},
  [6765] = {.lex_state = 179, .external_lex_state = 55},
  [6766] = {.lex_state = 179, .external_lex_state = 55},
  [6767] = {.lex_state = 179, .external_lex_state = 55},
  [6768] = {.lex_state = 179, .external_lex_state = 55},
  [6769] = {.lex_state = 444, .external_lex_state = 40},
  [6770] = {.lex_state = 444, .external_lex_state = 41},
  [6771] = {.lex_state = 446, .external_lex_state = 54},
  [6772] = {.lex_state = 444, .external_lex_state = 40},
  [6773] = {.lex_state = 444, .external_lex_state = 40},
  [6774] = {.lex_state = 444, .external_lex_state = 40},
  [6775] = {.lex_state = 446, .external_lex_state = 54},
  [6776] = {.lex_state = 154, .external_lex_state = 58},
  [6777] = {.lex_state = 444, .external_lex_state = 44},
  [6778] = {.lex_state = 444, .external_lex_state = 42},
  [6779] = {.lex_state = 446, .external_lex_state = 54},
  [6780] = {.lex_state = 446, .external_lex_state = 54},
  [6781] = {.lex_state = 444, .external_lex_state = 40},
  [6782] = {.lex_state = 444, .external_lex_state = 42},
  [6783] = {.lex_state = 444, .external_lex_state = 39},
  [6784] = {.lex_state = 444, .external_lex_state = 39},
  [6785] = {.lex_state = 444, .external_lex_state = 44},
  [6786] = {.lex_state = 446, .external_lex_state = 54},
  [6787] = {.lex_state = 172, .external_lex_state = 55},
  [6788] = {.lex_state = 179, .external_lex_state = 55},
  [6789] = {.lex_state = 446, .external_lex_state = 54},
  [6790] = {.lex_state = 444, .external_lex_state = 42},
  [6791] = {.lex_state = 444, .external_lex_state = 44},
  [6792] = {.lex_state = 444, .external_lex_state = 41},
  [6793] = {.lex_state = 446, .external_lex_state = 54},
  [6794] = {.lex_state = 444, .external_lex_state = 41},
  [6795] = {.lex_state = 444, .external_lex_state = 39},
  [6796] = {.lex_state = 446, .external_lex_state = 54},
  [6797] = {.lex_state = 446, .external_lex_state = 54},
  [6798] = {.lex_state = 444, .external_lex_state = 42},
  [6799] = {.lex_state = 446, .external_lex_state = 54},
  [6800] = {.lex_state = 444, .external_lex_state = 42},
  [6801] = {.lex_state = 444, .external_lex_state = 41},
  [6802] = {.lex_state = 446, .external_lex_state = 54},
  [6803] = {.lex_state = 179, .external_lex_state = 59},
  [6804] = {.lex_state = 444, .external_lex_state = 41},
  [6805] = {.lex_state = 179, .external_lex_state = 55},
  [6806] = {.lex_state = 179, .external_lex_state = 55},
  [6807] = {.lex_state = 444, .external_lex_state = 46},
  [6808] = {.lex_state = 179, .external_lex_state = 55},
  [6809] = {.lex_state = 179, .external_lex_state = 55},
  [6810] = {.lex_state = 444, .external_lex_state = 46},
  [6811] = {.lex_state = 179, .external_lex_state = 55},
  [6812] = {.lex_state = 179, .external_lex_state = 55},
  [6813] = {.lex_state = 179, .external_lex_state = 55},
  [6814] = {.lex_state = 179, .external_lex_state = 55},
  [6815] = {.lex_state = 444, .external_lex_state = 60},
  [6816] = {.lex_state = 179, .external_lex_state = 55},
  [6817] = {.lex_state = 179, .external_lex_state = 55},
  [6818] = {.lex_state = 179, .external_lex_state = 55},
  [6819] = {.lex_state = 444, .external_lex_state = 46},
  [6820] = {.lex_state = 179, .external_lex_state = 55},
  [6821] = {.lex_state = 179, .external_lex_state = 55},
  [6822] = {.lex_state = 179, .external_lex_state = 55},
  [6823] = {.lex_state = 444, .external_lex_state = 41},
  [6824] = {.lex_state = 179, .external_lex_state = 55},
  [6825] = {.lex_state = 444, .external_lex_state = 41},
  [6826] = {.lex_state = 179, .external_lex_state = 55},
  [6827] = {.lex_state = 179, .external_lex_state = 55},
  [6828] = {.lex_state = 179, .external_lex_state = 55},
  [6829] = {.lex_state = 444, .external_lex_state = 42},
  [6830] = {.lex_state = 444, .external_lex_state = 44},
  [6831] = {.lex_state = 444, .external_lex_state = 42},
  [6832] = {.lex_state = 179, .external_lex_state = 55},
  [6833] = {.lex_state = 444, .external_lex_state = 40},
  [6834] = {.lex_state = 444, .external_lex_state = 40},
  [6835] = {.lex_state = 444, .external_lex_state = 46},
  [6836] = {.lex_state = 444, .external_lex_state = 44},
  [6837] = {.lex_state = 444, .external_lex_state = 44},
  [6838] = {.lex_state = 444, .external_lex_state = 42},
  [6839] = {.lex_state = 179, .external_lex_state = 55},
  [6840] = {.lex_state = 172, .external_lex_state = 55},
  [6841] = {.lex_state = 444, .external_lex_state = 44},
  [6842] = {.lex_state = 172, .external_lex_state = 55},
  [6843] = {.lex_state = 444, .external_lex_state = 44},
  [6844] = {.lex_state = 444, .external_lex_state = 41},
  [6845] = {.lex_state = 179, .external_lex_state = 55},
  [6846] = {.lex_state = 444, .external_lex_state = 44},
  [6847] = {.lex_state = 179, .external_lex_state = 55},
  [6848] = {.lex_state = 444, .external_lex_state = 60},
  [6849] = {.lex_state = 444, .external_lex_state = 39},
  [6850] = {.lex_state = 444, .external_lex_state = 39},
  [6851] = {.lex_state = 444, .external_lex_state = 39},
  [6852] = {.lex_state = 444, .external_lex_state = 39},
  [6853] = {.lex_state = 444, .external_lex_state = 60},
  [6854] = {.lex_state = 444, .external_lex_state = 46},
  [6855] = {.lex_state = 444, .external_lex_state = 46},
  [6856] = {.lex_state = 444, .external_lex_state = 46},
  [6857] = {.lex_state = 444, .external_lex_state = 46},
  [6858] = {.lex_state = 444, .external_lex_state = 46},
  [6859] = {.lex_state = 444, .external_lex_state = 46},
  [6860] = {.lex_state = 444, .external_lex_state = 46},
  [6861] = {.lex_state = 179, .external_lex_state = 55},
  [6862] = {.lex_state = 179, .external_lex_state = 55},
  [6863] = {.lex_state = 444, .external_lex_state = 40},
  [6864] = {.lex_state = 444, .external_lex_state = 46},
  [6865] = {.lex_state = 444, .external_lex_state = 51},
  [6866] = {.lex_state = 444, .external_lex_state = 50},
  [6867] = {.lex_state = 444, .external_lex_state = 53},
  [6868] = {.lex_state = 444, .external_lex_state = 42},
  [6869] = {.lex_state = 444, .external_lex_state = 61},
  [6870] = {.lex_state = 444, .external_lex_state = 61},
  [6871] = {.lex_state = 444, .external_lex_state = 46},
  [6872] = {.lex_state = 444, .external_lex_state = 46},
  [6873] = {.lex_state = 444, .external_lex_state = 53},
  [6874] = {.lex_state = 444, .external_lex_state = 62},
  [6875] = {.lex_state = 444, .external_lex_state = 41},
  [6876] = {.lex_state = 444, .external_lex_state = 41},
  [6877] = {.lex_state = 444, .external_lex_state = 41},
  [6878] = {.lex_state = 444, .external_lex_state = 41},
  [6879] = {.lex_state = 444, .external_lex_state = 62},
  [6880] = {.lex_state = 444, .external_lex_state = 50},
  [6881] = {.lex_state = 444, .external_lex_state = 50},
  [6882] = {.lex_state = 444, .external_lex_state = 50},
  [6883] = {.lex_state = 444, .external_lex_state = 50},
  [6884] = {.lex_state = 444, .external_lex_state = 50},
  [6885] = {.lex_state = 444, .external_lex_state = 50},
  [6886] = {.lex_state = 444, .external_lex_state = 46},
  [6887] = {.lex_state = 444, .external_lex_state = 50},
  [6888] = {.lex_state = 444, .external_lex_state = 50},
  [6889] = {.lex_state = 444, .external_lex_state = 61},
  [6890] = {.lex_state = 444, .external_lex_state = 46},
  [6891] = {.lex_state = 444, .external_lex_state = 46},
  [6892] = {.lex_state = 444, .external_lex_state = 46},
  [6893] = {.lex_state = 444, .external_lex_state = 46},
  [6894] = {.lex_state = 444, .external_lex_state = 51},
  [6895] = {.lex_state = 444, .external_lex_state = 46},
  [6896] = {.lex_state = 444, .external_lex_state = 46},
  [6897] = {.lex_state = 444, .external_lex_state = 46},
  [6898] = {.lex_state = 444, .external_lex_state = 46},
  [6899] = {.lex_state = 444, .external_lex_state = 53},
  [6900] = {.lex_state = 444, .external_lex_state = 51},
  [6901] = {.lex_state = 444, .external_lex_state = 51},
  [6902] = {.lex_state = 444, .external_lex_state = 46},
  [6903] = {.lex_state = 444, .external_lex_state = 46},
  [6904] = {.lex_state = 444, .external_lex_state = 51},
  [6905] = {.lex_state = 444, .external_lex_state = 53},
  [6906] = {.lex_state = 444, .external_lex_state = 51},
  [6907] = {.lex_state = 444, .external_lex_state = 46},
  [6908] = {.lex_state = 444, .external_lex_state = 46},
  [6909] = {.lex_state = 444, .external_lex_state = 46},
  [6910] = {.lex_state = 444, .external_lex_state = 46},
  [6911] = {.lex_state = 444, .external_lex_state = 46},
  [6912] = {.lex_state = 444, .external_lex_state = 63},
  [6913] = {.lex_state = 444, .external_lex_state = 44},
  [6914] = {.lex_state = 444, .external_lex_state = 40},
  [6915] = {.lex_state = 444, .external_lex_state = 53},
  [6916] = {.lex_state = 444, .external_lex_state = 53},
  [6917] = {.lex_state = 444, .external_lex_state = 44},
  [6918] = {.lex_state = 444, .external_lex_state = 46},
  [6919] = {.lex_state = 444, .external_lex_state = 46},
  [6920] = {.lex_state = 444, .external_lex_state = 51},
  [6921] = {.lex_state = 444, .external_lex_state = 46},
  [6922] = {.lex_state = 444, .external_lex_state = 46},
  [6923] = {.lex_state = 444, .external_lex_state = 40},
  [6924] = {.lex_state = 444, .external_lex_state = 42},
  [6925] = {.lex_state = 444, .external_lex_state = 46},
  [6926] = {.lex_state = 444, .external_lex_state = 46},
  [6927] = {.lex_state = 444, .external_lex_state = 46},
  [6928] = {.lex_state = 444, .external_lex_state = 46},
  [6929] = {.lex_state = 444, .external_lex_state = 51},
  [6930] = {.lex_state = 444, .external_lex_state = 40},
  [6931] = {.lex_state = 444, .external_lex_state = 46},
  [6932] = {.lex_state = 444, .external_lex_state = 46},
  [6933] = {.lex_state = 444, .external_lex_state = 53},
  [6934] = {.lex_state = 444, .external_lex_state = 51},
  [6935] = {.lex_state = 444, .external_lex_state = 46},
  [6936] = {.lex_state = 444, .external_lex_state = 46},
  [6937] = {.lex_state = 444, .external_lex_state = 46},
  [6938] = {.lex_state = 444, .external_lex_state = 46},
  [6939] = {.lex_state = 444, .external_lex_state = 53},
  [6940] = {.lex_state = 444, .external_lex_state = 46},
  [6941] = {.lex_state = 444, .external_lex_state = 46},
  [6942] = {.lex_state = 444, .external_lex_state = 46},
  [6943] = {.lex_state = 444, .external_lex_state = 46},
  [6944] = {.lex_state = 444, .external_lex_state = 46},
  [6945] = {.lex_state = 444, .external_lex_state = 46},
  [6946] = {.lex_state = 444, .external_lex_state = 53},
  [6947] = {.lex_state = 444, .external_lex_state = 63},
  [6948] = {.lex_state = 444, .external_lex_state = 46},
  [6949] = {.lex_state = 444, .external_lex_state = 46},
  [6950] = {.lex_state = 444, .external_lex_state = 62},
  [6951] = {.lex_state = 444, .external_lex_state = 42},
  [6952] = {.lex_state = 444, .external_lex_state = 42},
  [6953] = {.lex_state = 444, .external_lex_state = 46},
  [6954] = {.lex_state = 444, .external_lex_state = 46},
  [6955] = {.lex_state = 444, .external_lex_state = 46},
  [6956] = {.lex_state = 444, .external_lex_state = 46},
  [6957] = {.lex_state = 444, .external_lex_state = 46},
  [6958] = {.lex_state = 444, .external_lex_state = 46},
  [6959] = {.lex_state = 444, .external_lex_state = 44},
  [6960] = {.lex_state = 444, .external_lex_state = 50},
  [6961] = {.lex_state = 444, .external_lex_state = 46},
  [6962] = {.lex_state = 444, .external_lex_state = 46},
  [6963] = {.lex_state = 444, .external_lex_state = 40},
  [6964] = {.lex_state = 444, .external_lex_state = 63},
  [6965] = {.lex_state = 444, .external_lex_state = 53},
  [6966] = {.lex_state = 444, .external_lex_state = 51},
  [6967] = {.lex_state = 444, .external_lex_state = 46},
  [6968] = {.lex_state = 444, .external_lex_state = 51},
  [6969] = {.lex_state = 444, .external_lex_state = 53},
  [6970] = {.lex_state = 444, .external_lex_state = 50},
  [6971] = {.lex_state = 444, .external_lex_state = 46},
  [6972] = {.lex_state = 444, .external_lex_state = 53},
  [6973] = {.lex_state = 444, .external_lex_state = 50},
  [6974] = {.lex_state = 444, .external_lex_state = 53},
  [6975] = {.lex_state = 444, .external_lex_state = 50},
  [6976] = {.lex_state = 444, .external_lex_state = 50},
  [6977] = {.lex_state = 444, .external_lex_state = 53},
  [6978] = {.lex_state = 444, .external_lex_state = 53},
  [6979] = {.lex_state = 444, .external_lex_state = 50},
  [6980] = {.lex_state = 444, .external_lex_state = 50},
  [6981] = {.lex_state = 444, .external_lex_state = 50},
  [6982] = {.lex_state = 444, .external_lex_state = 53},
  [6983] = {.lex_state = 444, .external_lex_state = 53},
  [6984] = {.lex_state = 444, .external_lex_state = 53},
  [6985] = {.lex_state = 444, .external_lex_state = 51},
  [6986] = {.lex_state = 444, .external_lex_state = 53},
  [6987] = {.lex_state = 444, .external_lex_state = 53},
  [6988] = {.lex_state = 444, .external_lex_state = 53},
  [6989] = {.lex_state = 444, .external_lex_state = 50},
  [6990] = {.lex_state = 444, .external_lex_state = 50},
  [6991] = {.lex_state = 444, .external_lex_state = 51},
  [6992] = {.lex_state = 444, .external_lex_state = 53},
  [6993] = {.lex_state = 444, .external_lex_state = 50},
  [6994] = {.lex_state = 444, .external_lex_state = 50},
  [6995] = {.lex_state = 444, .external_lex_state = 53},
  [6996] = {.lex_state = 444, .external_lex_state = 51},
  [6997] = {.lex_state = 444, .external_lex_state = 53},
  [6998] = {.lex_state = 444, .external_lex_state = 51},
  [6999] = {.lex_state = 444, .external_lex_state = 51},
  [7000] = {.lex_state = 444, .external_lex_state = 51},
  [7001] = {.lex_state = 444, .external_lex_state = 51},
  [7002] = {.lex_state = 444, .external_lex_state = 50},
  [7003] = {.lex_state = 444, .external_lex_state = 51},
  [7004] = {.lex_state = 444, .external_lex_state = 51},
  [7005] = {.lex_state = 444, .external_lex_state = 51},
  [7006] = {.lex_state = 444, .external_lex_state = 51},
  [7007] = {.lex_state = 444, .external_lex_state = 51},
  [7008] = {.lex_state = 444, .external_lex_state = 54},
  [7009] = {.lex_state = 444, .external_lex_state = 50},
  [7010] = {.lex_state = 444, .external_lex_state = 51},
  [7011] = {.lex_state = 444, .external_lex_state = 51},
  [7012] = {.lex_state = 444, .external_lex_state = 50},
  [7013] = {.lex_state = 444, .external_lex_state = 51},
  [7014] = {.lex_state = 444, .external_lex_state = 53},
  [7015] = {.lex_state = 444, .external_lex_state = 54},
  [7016] = {.lex_state = 444, .external_lex_state = 51},
  [7017] = {.lex_state = 444, .external_lex_state = 51},
  [7018] = {.lex_state = 444, .external_lex_state = 50},
  [7019] = {.lex_state = 444, .external_lex_state = 54},
  [7020] = {.lex_state = 444, .external_lex_state = 50},
  [7021] = {.lex_state = 444, .external_lex_state = 53},
  [7022] = {.lex_state = 444, .external_lex_state = 51},
  [7023] = {.lex_state = 444, .external_lex_state = 51},
  [7024] = {.lex_state = 444, .external_lex_state = 51},
  [7025] = {.lex_state = 444, .external_lex_state = 51},
  [7026] = {.lex_state = 444, .external_lex_state = 51},
  [7027] = {.lex_state = 444, .external_lex_state = 53},
  [7028] = {.lex_state = 444, .external_lex_state = 51},
  [7029] = {.lex_state = 444, .external_lex_state = 51},
  [7030] = {.lex_state = 444, .external_lex_state = 53},
  [7031] = {.lex_state = 444, .external_lex_state = 51},
  [7032] = {.lex_state = 444, .external_lex_state = 53},
  [7033] = {.lex_state = 444, .external_lex_state = 51},
  [7034] = {.lex_state = 444, .external_lex_state = 51},
  [7035] = {.lex_state = 444, .external_lex_state = 51},
  [7036] = {.lex_state = 444, .external_lex_state = 51},
  [7037] = {.lex_state = 444, .external_lex_state = 51},
  [7038] = {.lex_state = 444, .external_lex_state = 53},
  [7039] = {.lex_state = 444, .external_lex_state = 51},
  [7040] = {.lex_state = 444, .external_lex_state = 51},
  [7041] = {.lex_state = 444, .external_lex_state = 50},
  [7042] = {.lex_state = 444, .external_lex_state = 64},
  [7043] = {.lex_state = 444, .external_lex_state = 51},
  [7044] = {.lex_state = 444, .external_lex_state = 51},
  [7045] = {.lex_state = 444, .external_lex_state = 50},
  [7046] = {.lex_state = 444, .external_lex_state = 53},
  [7047] = {.lex_state = 444, .external_lex_state = 53},
  [7048] = {.lex_state = 444, .external_lex_state = 53},
  [7049] = {.lex_state = 444, .external_lex_state = 44},
  [7050] = {.lex_state = 444, .external_lex_state = 51},
  [7051] = {.lex_state = 444, .external_lex_state = 51},
  [7052] = {.lex_state = 444, .external_lex_state = 50},
  [7053] = {.lex_state = 444, .external_lex_state = 53},
  [7054] = {.lex_state = 444, .external_lex_state = 53},
  [7055] = {.lex_state = 444, .external_lex_state = 44},
  [7056] = {.lex_state = 444, .external_lex_state = 51},
  [7057] = {.lex_state = 444, .external_lex_state = 53},
  [7058] = {.lex_state = 444, .external_lex_state = 50},
  [7059] = {.lex_state = 444, .external_lex_state = 50},
  [7060] = {.lex_state = 444, .external_lex_state = 50},
  [7061] = {.lex_state = 444, .external_lex_state = 50},
  [7062] = {.lex_state = 444, .external_lex_state = 51},
  [7063] = {.lex_state = 444, .external_lex_state = 51},
  [7064] = {.lex_state = 444, .external_lex_state = 50},
  [7065] = {.lex_state = 444, .external_lex_state = 50},
  [7066] = {.lex_state = 444, .external_lex_state = 51},
  [7067] = {.lex_state = 444, .external_lex_state = 53},
  [7068] = {.lex_state = 444, .external_lex_state = 50},
  [7069] = {.lex_state = 444, .external_lex_state = 50},
  [7070] = {.lex_state = 444, .external_lex_state = 50},
  [7071] = {.lex_state = 444, .external_lex_state = 53},
  [7072] = {.lex_state = 444, .external_lex_state = 53},
  [7073] = {.lex_state = 444, .external_lex_state = 53},
  [7074] = {.lex_state = 444, .external_lex_state = 50},
  [7075] = {.lex_state = 444, .external_lex_state = 50},
  [7076] = {.lex_state = 444, .external_lex_state = 50},
  [7077] = {.lex_state = 444, .external_lex_state = 50},
  [7078] = {.lex_state = 444, .external_lex_state = 50},
  [7079] = {.lex_state = 444, .external_lex_state = 53},
  [7080] = {.lex_state = 444, .external_lex_state = 51},
  [7081] = {.lex_state = 444, .external_lex_state = 50},
  [7082] = {.lex_state = 444, .external_lex_state = 50},
  [7083] = {.lex_state = 444, .external_lex_state = 51},
  [7084] = {.lex_state = 444, .external_lex_state = 53},
  [7085] = {.lex_state = 444, .external_lex_state = 53},
  [7086] = {.lex_state = 444, .external_lex_state = 50},
  [7087] = {.lex_state = 444, .external_lex_state = 50},
  [7088] = {.lex_state = 444, .external_lex_state = 50},
  [7089] = {.lex_state = 444, .external_lex_state = 50},
  [7090] = {.lex_state = 444, .external_lex_state = 50},
  [7091] = {.lex_state = 444, .external_lex_state = 64},
  [7092] = {.lex_state = 444, .external_lex_state = 50},
  [7093] = {.lex_state = 444, .external_lex_state = 50},
  [7094] = {.lex_state = 444, .external_lex_state = 53},
  [7095] = {.lex_state = 444, .external_lex_state = 53},
  [7096] = {.lex_state = 444, .external_lex_state = 44},
  [7097] = {.lex_state = 444, .external_lex_state = 44},
  [7098] = {.lex_state = 444, .external_lex_state = 51},
  [7099] = {.lex_state = 444, .external_lex_state = 51},
  [7100] = {.lex_state = 444, .external_lex_state = 64},
  [7101] = {.lex_state = 444, .external_lex_state = 53},
  [7102] = {.lex_state = 444, .external_lex_state = 53},
  [7103] = {.lex_state = 444, .external_lex_state = 54},
  [7104] = {.lex_state = 444, .external_lex_state = 54},
  [7105] = {.lex_state = 444, .external_lex_state = 51},
  [7106] = {.lex_state = 444, .external_lex_state = 53},
  [7107] = {.lex_state = 444, .external_lex_state = 53},
  [7108] = {.lex_state = 444, .external_lex_state = 53},
  [7109] = {.lex_state = 444, .external_lex_state = 53},
  [7110] = {.lex_state = 444, .external_lex_state = 53},
  [7111] = {.lex_state = 444, .external_lex_state = 54},
  [7112] = {.lex_state = 444, .external_lex_state = 54},
  [7113] = {.lex_state = 444, .external_lex_state = 54},
  [7114] = {.lex_state = 444, .external_lex_state = 53},
  [7115] = {.lex_state = 444, .external_lex_state = 53},
  [7116] = {.lex_state = 444, .external_lex_state = 53},
  [7117] = {.lex_state = 444, .external_lex_state = 54},
  [7118] = {.lex_state = 444, .external_lex_state = 50},
  [7119] = {.lex_state = 444, .external_lex_state = 50},
  [7120] = {.lex_state = 444, .external_lex_state = 53},
  [7121] = {.lex_state = 444, .external_lex_state = 51},
  [7122] = {.lex_state = 444, .external_lex_state = 51},
  [7123] = {.lex_state = 444, .external_lex_state = 53},
  [7124] = {.lex_state = 444, .external_lex_state = 51},
  [7125] = {.lex_state = 444, .external_lex_state = 50},
  [7126] = {.lex_state = 444, .external_lex_state = 51},
  [7127] = {.lex_state = 444, .external_lex_state = 50},
  [7128] = {.lex_state = 444, .external_lex_state = 51},
  [7129] = {.lex_state = 444, .external_lex_state = 50},
  [7130] = {.lex_state = 444, .external_lex_state = 54},
  [7131] = {.lex_state = 444, .external_lex_state = 54},
  [7132] = {.lex_state = 444, .external_lex_state = 53},
  [7133] = {.lex_state = 444, .external_lex_state = 50},
  [7134] = {.lex_state = 444, .external_lex_state = 51},
  [7135] = {.lex_state = 444, .external_lex_state = 50},
  [7136] = {.lex_state = 444, .external_lex_state = 53},
  [7137] = {.lex_state = 444, .external_lex_state = 53},
  [7138] = {.lex_state = 444, .external_lex_state = 53},
  [7139] = {.lex_state = 444, .external_lex_state = 53},
  [7140] = {.lex_state = 444, .external_lex_state = 50},
  [7141] = {.lex_state = 444, .external_lex_state = 50},
  [7142] = {.lex_state = 444, .external_lex_state = 50},
  [7143] = {.lex_state = 444, .external_lex_state = 54},
  [7144] = {.lex_state = 444, .external_lex_state = 54},
  [7145] = {.lex_state = 444, .external_lex_state = 54},
  [7146] = {.lex_state = 444, .external_lex_state = 54},
  [7147] = {.lex_state = 444, .external_lex_state = 54},
  [7148] = {.lex_state = 146, .external_lex_state = 2},
  [7149] = {.lex_state = 445, .external_lex_state = 16},
  [7150] = {.lex_state = 444, .external_lex_state = 54},
  [7151] = {.lex_state = 445, .external_lex_state = 16},
  [7152] = {.lex_state = 444, .external_lex_state = 54},
  [7153] = {.lex_state = 444, .external_lex_state = 54},
  [7154] = {.lex_state = 444, .external_lex_state = 54},
  [7155] = {.lex_state = 444, .external_lex_state = 54},
  [7156] = {.lex_state = 445, .external_lex_state = 16},
  [7157] = {.lex_state = 445, .external_lex_state = 16},
  [7158] = {.lex_state = 444, .external_lex_state = 54},
  [7159] = {.lex_state = 444, .external_lex_state = 54},
  [7160] = {.lex_state = 444, .external_lex_state = 54},
  [7161] = {.lex_state = 445, .external_lex_state = 16},
  [7162] = {.lex_state = 444, .external_lex_state = 54},
  [7163] = {.lex_state = 444, .external_lex_state = 54},
  [7164] = {.lex_state = 444, .external_lex_state = 54},
  [7165] = {.lex_state = 444, .external_lex_state = 54},
  [7166] = {.lex_state = 444, .external_lex_state = 54},
  [7167] = {.lex_state = 445, .external_lex_state = 16},
  [7168] = {.lex_state = 444, .external_lex_state = 54},
  [7169] = {.lex_state = 444, .external_lex_state = 54},
  [7170] = {.lex_state = 445, .external_lex_state = 16},
  [7171] = {.lex_state = 444, .external_lex_state = 54},
  [7172] = {.lex_state = 444, .external_lex_state = 54},
  [7173] = {.lex_state = 444, .external_lex_state = 54},
  [7174] = {.lex_state = 444, .external_lex_state = 54},
  [7175] = {.lex_state = 444, .external_lex_state = 54},
  [7176] = {.lex_state = 444, .external_lex_state = 54},
  [7177] = {.lex_state = 444, .external_lex_state = 54},
  [7178] = {.lex_state = 444, .external_lex_state = 54},
  [7179] = {.lex_state = 444, .external_lex_state = 54},
  [7180] = {.lex_state = 444, .external_lex_state = 54},
  [7181] = {.lex_state = 444, .external_lex_state = 54},
  [7182] = {.lex_state = 444, .external_lex_state = 54},
  [7183] = {.lex_state = 444, .external_lex_state = 54},
  [7184] = {.lex_state = 444, .external_lex_state = 54},
  [7185] = {.lex_state = 444, .external_lex_state = 54},
  [7186] = {.lex_state = 444, .external_lex_state = 54},
  [7187] = {.lex_state = 444, .external_lex_state = 54},
  [7188] = {.lex_state = 444, .external_lex_state = 54},
  [7189] = {.lex_state = 444, .external_lex_state = 54},
  [7190] = {.lex_state = 444, .external_lex_state = 54},
  [7191] = {.lex_state = 444, .external_lex_state = 54},
  [7192] = {.lex_state = 444, .external_lex_state = 54},
  [7193] = {.lex_state = 444, .external_lex_state = 54},
  [7194] = {.lex_state = 444, .external_lex_state = 54},
  [7195] = {.lex_state = 444, .external_lex_state = 54},
  [7196] = {.lex_state = 444, .external_lex_state = 54},
  [7197] = {.lex_state = 444, .external_lex_state = 54},
  [7198] = {.lex_state = 444, .external_lex_state = 54},
  [7199] = {.lex_state = 444, .external_lex_state = 54},
  [7200] = {.lex_state = 445, .external_lex_state = 16},
  [7201] = {.lex_state = 445, .external_lex_state = 16},
  [7202] = {.lex_state = 445, .external_lex_state = 16},
  [7203] = {.lex_state = 444, .external_lex_state = 54},
  [7204] = {.lex_state = 444, .external_lex_state = 54},
  [7205] = {.lex_state = 179, .external_lex_state = 55},
  [7206] = {.lex_state = 179, .external_lex_state = 55},
  [7207] = {.lex_state = 179, .external_lex_state = 55},
  [7208] = {.lex_state = 179, .external_lex_state = 55},
  [7209] = {.lex_state = 179, .external_lex_state = 55},
  [7210] = {.lex_state = 179, .external_lex_state = 55},
  [7211] = {.lex_state = 445, .external_lex_state = 16},
  [7212] = {.lex_state = 179, .external_lex_state = 55},
  [7213] = {.lex_state = 179, .external_lex_state = 55},
  [7214] = {.lex_state = 179, .external_lex_state = 55},
  [7215] = {.lex_state = 179, .external_lex_state = 55},
  [7216] = {.lex_state = 179, .external_lex_state = 55},
  [7217] = {.lex_state = 179, .external_lex_state = 55},
  [7218] = {.lex_state = 179, .external_lex_state = 55},
  [7219] = {.lex_state = 179, .external_lex_state = 55},
  [7220] = {.lex_state = 179, .external_lex_state = 55},
  [7221] = {.lex_state = 179, .external_lex_state = 55},
  [7222] = {.lex_state = 179, .external_lex_state = 55},
  [7223] = {.lex_state = 179, .external_lex_state = 55},
  [7224] = {.lex_state = 445, .external_lex_state = 20},
  [7225] = {.lex_state = 179, .external_lex_state = 55},
  [7226] = {.lex_state = 179, .external_lex_state = 55},
  [7227] = {.lex_state = 179, .external_lex_state = 55},
  [7228] = {.lex_state = 179, .external_lex_state = 55},
  [7229] = {.lex_state = 179, .external_lex_state = 55},
  [7230] = {.lex_state = 179, .external_lex_state = 55},
  [7231] = {.lex_state = 179, .external_lex_state = 55},
  [7232] = {.lex_state = 179, .external_lex_state = 55},
  [7233] = {.lex_state = 445, .external_lex_state = 20},
  [7234] = {.lex_state = 179, .external_lex_state = 55},
  [7235] = {.lex_state = 179, .external_lex_state = 55},
  [7236] = {.lex_state = 179, .external_lex_state = 55},
  [7237] = {.lex_state = 179, .external_lex_state = 55},
  [7238] = {.lex_state = 179, .external_lex_state = 55},
  [7239] = {.lex_state = 179, .external_lex_state = 55},
  [7240] = {.lex_state = 179, .external_lex_state = 55},
  [7241] = {.lex_state = 179, .external_lex_state = 55},
  [7242] = {.lex_state = 179, .external_lex_state = 55},
  [7243] = {.lex_state = 179, .external_lex_state = 55},
  [7244] = {.lex_state = 179, .external_lex_state = 55},
  [7245] = {.lex_state = 179, .external_lex_state = 55},
  [7246] = {.lex_state = 179, .external_lex_state = 55},
  [7247] = {.lex_state = 179, .external_lex_state = 55},
  [7248] = {.lex_state = 179, .external_lex_state = 55},
  [7249] = {.lex_state = 179, .external_lex_state = 55},
  [7250] = {.lex_state = 179, .external_lex_state = 55},
  [7251] = {.lex_state = 179, .external_lex_state = 55},
  [7252] = {.lex_state = 179, .external_lex_state = 55},
  [7253] = {.lex_state = 179, .external_lex_state = 55},
  [7254] = {.lex_state = 179, .external_lex_state = 55},
  [7255] = {.lex_state = 445, .external_lex_state = 20},
  [7256] = {.lex_state = 179, .external_lex_state = 55},
  [7257] = {.lex_state = 445, .external_lex_state = 20},
  [7258] = {.lex_state = 179, .external_lex_state = 55},
  [7259] = {.lex_state = 179, .external_lex_state = 55},
  [7260] = {.lex_state = 179, .external_lex_state = 55},
  [7261] = {.lex_state = 179, .external_lex_state = 55},
  [7262] = {.lex_state = 179, .external_lex_state = 55},
  [7263] = {.lex_state = 445, .external_lex_state = 21},
  [7264] = {.lex_state = 445, .external_lex_state = 21},
  [7265] = {.lex_state = 179, .external_lex_state = 55},
  [7266] = {.lex_state = 179, .external_lex_state = 55},
  [7267] = {.lex_state = 179, .external_lex_state = 55},
  [7268] = {.lex_state = 179, .external_lex_state = 55},
  [7269] = {.lex_state = 179, .external_lex_state = 55},
  [7270] = {.lex_state = 179, .external_lex_state = 55},
  [7271] = {.lex_state = 179, .external_lex_state = 55},
  [7272] = {.lex_state = 445, .external_lex_state = 20},
  [7273] = {.lex_state = 179, .external_lex_state = 55},
  [7274] = {.lex_state = 179, .external_lex_state = 55},
  [7275] = {.lex_state = 179, .external_lex_state = 55},
  [7276] = {.lex_state = 179, .external_lex_state = 55},
  [7277] = {.lex_state = 179, .external_lex_state = 55},
  [7278] = {.lex_state = 179, .external_lex_state = 55},
  [7279] = {.lex_state = 179, .external_lex_state = 55},
  [7280] = {.lex_state = 179, .external_lex_state = 55},
  [7281] = {.lex_state = 179, .external_lex_state = 55},
  [7282] = {.lex_state = 179, .external_lex_state = 55},
  [7283] = {.lex_state = 179, .external_lex_state = 55},
  [7284] = {.lex_state = 179, .external_lex_state = 55},
  [7285] = {.lex_state = 445, .external_lex_state = 2},
  [7286] = {.lex_state = 179, .external_lex_state = 55},
  [7287] = {.lex_state = 179, .external_lex_state = 55},
  [7288] = {.lex_state = 179, .external_lex_state = 55},
  [7289] = {.lex_state = 179, .external_lex_state = 55},
  [7290] = {.lex_state = 179, .external_lex_state = 55},
  [7291] = {.lex_state = 179, .external_lex_state = 55},
  [7292] = {.lex_state = 179, .external_lex_state = 55},
  [7293] = {.lex_state = 179, .external_lex_state = 55},
  [7294] = {.lex_state = 179, .external_lex_state = 55},
  [7295] = {.lex_state = 179, .external_lex_state = 55},
  [7296] = {.lex_state = 179, .external_lex_state = 55},
  [7297] = {.lex_state = 179, .external_lex_state = 55},
  [7298] = {.lex_state = 179, .external_lex_state = 55},
  [7299] = {.lex_state = 179, .external_lex_state = 55},
  [7300] = {.lex_state = 445, .external_lex_state = 16},
  [7301] = {.lex_state = 179, .external_lex_state = 55},
  [7302] = {.lex_state = 179, .external_lex_state = 55},
  [7303] = {.lex_state = 179, .external_lex_state = 55},
  [7304] = {.lex_state = 445, .external_lex_state = 20},
  [7305] = {.lex_state = 179, .external_lex_state = 55},
  [7306] = {.lex_state = 179, .external_lex_state = 55},
  [7307] = {.lex_state = 445, .external_lex_state = 21},
  [7308] = {.lex_state = 179, .external_lex_state = 55},
  [7309] = {.lex_state = 179, .external_lex_state = 55},
  [7310] = {.lex_state = 445, .external_lex_state = 21},
  [7311] = {.lex_state = 445, .external_lex_state = 20},
  [7312] = {.lex_state = 179, .external_lex_state = 55},
  [7313] = {.lex_state = 179, .external_lex_state = 55},
  [7314] = {.lex_state = 179, .external_lex_state = 55},
  [7315] = {.lex_state = 179, .external_lex_state = 55},
  [7316] = {.lex_state = 179, .external_lex_state = 55},
  [7317] = {.lex_state = 179, .external_lex_state = 55},
  [7318] = {.lex_state = 179, .external_lex_state = 55},
  [7319] = {.lex_state = 179, .external_lex_state = 55},
  [7320] = {.lex_state = 179, .external_lex_state = 55},
  [7321] = {.lex_state = 445, .external_lex_state = 2},
  [7322] = {.lex_state = 179, .external_lex_state = 55},
  [7323] = {.lex_state = 179, .external_lex_state = 55},
  [7324] = {.lex_state = 179, .external_lex_state = 55},
  [7325] = {.lex_state = 179, .external_lex_state = 55},
  [7326] = {.lex_state = 179, .external_lex_state = 55},
  [7327] = {.lex_state = 179, .external_lex_state = 55},
  [7328] = {.lex_state = 179, .external_lex_state = 55},
  [7329] = {.lex_state = 179, .external_lex_state = 55},
  [7330] = {.lex_state = 445, .external_lex_state = 20},
  [7331] = {.lex_state = 179, .external_lex_state = 55},
  [7332] = {.lex_state = 179, .external_lex_state = 55},
  [7333] = {.lex_state = 179, .external_lex_state = 55},
  [7334] = {.lex_state = 179, .external_lex_state = 55},
  [7335] = {.lex_state = 179, .external_lex_state = 55},
  [7336] = {.lex_state = 179, .external_lex_state = 55},
  [7337] = {.lex_state = 179, .external_lex_state = 55},
  [7338] = {.lex_state = 179, .external_lex_state = 55},
  [7339] = {.lex_state = 179, .external_lex_state = 55},
  [7340] = {.lex_state = 179, .external_lex_state = 55},
  [7341] = {.lex_state = 179, .external_lex_state = 55},
  [7342] = {.lex_state = 179, .external_lex_state = 55},
  [7343] = {.lex_state = 179, .external_lex_state = 55},
  [7344] = {.lex_state = 179, .external_lex_state = 55},
  [7345] = {.lex_state = 179, .external_lex_state = 55},
  [7346] = {.lex_state = 179, .external_lex_state = 55},
  [7347] = {.lex_state = 179, .external_lex_state = 55},
  [7348] = {.lex_state = 179, .external_lex_state = 55},
  [7349] = {.lex_state = 179, .external_lex_state = 55},
  [7350] = {.lex_state = 179, .external_lex_state = 55},
  [7351] = {.lex_state = 179, .external_lex_state = 55},
  [7352] = {.lex_state = 179, .external_lex_state = 55},
  [7353] = {.lex_state = 179, .external_lex_state = 55},
  [7354] = {.lex_state = 179, .external_lex_state = 55},
  [7355] = {.lex_state = 179, .external_lex_state = 55},
  [7356] = {.lex_state = 445, .external_lex_state = 20},
  [7357] = {.lex_state = 179, .external_lex_state = 55},
  [7358] = {.lex_state = 179, .external_lex_state = 55},
  [7359] = {.lex_state = 179, .external_lex_state = 55},
  [7360] = {.lex_state = 179, .external_lex_state = 55},
  [7361] = {.lex_state = 179, .external_lex_state = 55},
  [7362] = {.lex_state = 445, .external_lex_state = 2},
  [7363] = {.lex_state = 179, .external_lex_state = 55},
  [7364] = {.lex_state = 179, .external_lex_state = 55},
  [7365] = {.lex_state = 179, .external_lex_state = 55},
  [7366] = {.lex_state = 179, .external_lex_state = 55},
  [7367] = {.lex_state = 179, .external_lex_state = 55},
  [7368] = {.lex_state = 445, .external_lex_state = 16},
  [7369] = {.lex_state = 179, .external_lex_state = 55},
  [7370] = {.lex_state = 445, .external_lex_state = 16},
  [7371] = {.lex_state = 179, .external_lex_state = 55},
  [7372] = {.lex_state = 445, .external_lex_state = 2},
  [7373] = {.lex_state = 179, .external_lex_state = 55},
  [7374] = {.lex_state = 179, .external_lex_state = 55},
  [7375] = {.lex_state = 179, .external_lex_state = 55},
  [7376] = {.lex_state = 179, .external_lex_state = 55},
  [7377] = {.lex_state = 179, .external_lex_state = 55},
  [7378] = {.lex_state = 179, .external_lex_state = 55},
  [7379] = {.lex_state = 179, .external_lex_state = 55},
  [7380] = {.lex_state = 445, .external_lex_state = 21},
  [7381] = {.lex_state = 179, .external_lex_state = 55},
  [7382] = {.lex_state = 445, .external_lex_state = 16},
  [7383] = {.lex_state = 445, .external_lex_state = 2},
  [7384] = {.lex_state = 179, .external_lex_state = 55},
  [7385] = {.lex_state = 445, .external_lex_state = 2},
  [7386] = {.lex_state = 179, .external_lex_state = 55},
  [7387] = {.lex_state = 445, .external_lex_state = 2},
  [7388] = {.lex_state = 179, .external_lex_state = 55},
  [7389] = {.lex_state = 445, .external_lex_state = 21},
  [7390] = {.lex_state = 179, .external_lex_state = 55},
  [7391] = {.lex_state = 179, .external_lex_state = 55},
  [7392] = {.lex_state = 179, .external_lex_state = 55},
  [7393] = {.lex_state = 445, .external_lex_state = 21},
  [7394] = {.lex_state = 179, .external_lex_state = 55},
  [7395] = {.lex_state = 179, .external_lex_state = 55},
  [7396] = {.lex_state = 179, .external_lex_state = 55},
  [7397] = {.lex_state = 179, .external_lex_state = 55},
  [7398] = {.lex_state = 445, .external_lex_state = 21},
  [7399] = {.lex_state = 179, .external_lex_state = 55},
  [7400] = {.lex_state = 179, .external_lex_state = 55},
  [7401] = {.lex_state = 445, .external_lex_state = 21},
  [7402] = {.lex_state = 179, .external_lex_state = 55},
  [7403] = {.lex_state = 445, .external_lex_state = 2},
  [7404] = {.lex_state = 445, .external_lex_state = 21},
  [7405] = {.lex_state = 179, .external_lex_state = 55},
  [7406] = {.lex_state = 179, .external_lex_state = 55},
  [7407] = {.lex_state = 179, .external_lex_state = 55},
  [7408] = {.lex_state = 179, .external_lex_state = 55},
  [7409] = {.lex_state = 179, .external_lex_state = 55},
  [7410] = {.lex_state = 179, .external_lex_state = 55},
  [7411] = {.lex_state = 179, .external_lex_state = 55},
  [7412] = {.lex_state = 445, .external_lex_state = 20},
  [7413] = {.lex_state = 179, .external_lex_state = 55},
  [7414] = {.lex_state = 179, .external_lex_state = 55},
  [7415] = {.lex_state = 179, .external_lex_state = 55},
  [7416] = {.lex_state = 179, .external_lex_state = 55},
  [7417] = {.lex_state = 179, .external_lex_state = 55},
  [7418] = {.lex_state = 179, .external_lex_state = 55},
  [7419] = {.lex_state = 445, .external_lex_state = 2},
  [7420] = {.lex_state = 179, .external_lex_state = 55},
  [7421] = {.lex_state = 179, .external_lex_state = 55},
  [7422] = {.lex_state = 445, .external_lex_state = 2},
  [7423] = {.lex_state = 179, .external_lex_state = 55},
  [7424] = {.lex_state = 179, .external_lex_state = 55},
  [7425] = {.lex_state = 179, .external_lex_state = 55},
  [7426] = {.lex_state = 179, .external_lex_state = 55},
  [7427] = {.lex_state = 179, .external_lex_state = 55},
  [7428] = {.lex_state = 179, .external_lex_state = 55},
  [7429] = {.lex_state = 445, .external_lex_state = 23},
  [7430] = {.lex_state = 444, .external_lex_state = 46},
  [7431] = {.lex_state = 147, .external_lex_state = 2},
  [7432] = {.lex_state = 444, .external_lex_state = 46},
  [7433] = {.lex_state = 450, .external_lex_state = 46},
  [7434] = {.lex_state = 445, .external_lex_state = 16},
  [7435] = {.lex_state = 445, .external_lex_state = 16},
  [7436] = {.lex_state = 445, .external_lex_state = 16},
  [7437] = {.lex_state = 445, .external_lex_state = 21},
  [7438] = {.lex_state = 445, .external_lex_state = 21},
  [7439] = {.lex_state = 445, .external_lex_state = 23},
  [7440] = {.lex_state = 445, .external_lex_state = 23},
  [7441] = {.lex_state = 445, .external_lex_state = 23},
  [7442] = {.lex_state = 445, .external_lex_state = 23},
  [7443] = {.lex_state = 445, .external_lex_state = 23},
  [7444] = {.lex_state = 179, .external_lex_state = 55},
  [7445] = {.lex_state = 445, .external_lex_state = 16},
  [7446] = {.lex_state = 179, .external_lex_state = 55},
  [7447] = {.lex_state = 179, .external_lex_state = 55},
  [7448] = {.lex_state = 179, .external_lex_state = 55},
  [7449] = {.lex_state = 179, .external_lex_state = 55},
  [7450] = {.lex_state = 179, .external_lex_state = 55},
  [7451] = {.lex_state = 179, .external_lex_state = 55},
  [7452] = {.lex_state = 179, .external_lex_state = 55},
  [7453] = {.lex_state = 179, .external_lex_state = 55},
  [7454] = {.lex_state = 179, .external_lex_state = 55},
  [7455] = {.lex_state = 179, .external_lex_state = 55},
  [7456] = {.lex_state = 179, .external_lex_state = 55},
  [7457] = {.lex_state = 179, .external_lex_state = 55},
  [7458] = {.lex_state = 445, .external_lex_state = 21},
  [7459] = {.lex_state = 179, .external_lex_state = 55},
  [7460] = {.lex_state = 179, .external_lex_state = 55},
  [7461] = {.lex_state = 179, .external_lex_state = 55},
  [7462] = {.lex_state = 179, .external_lex_state = 55},
  [7463] = {.lex_state = 179, .external_lex_state = 55},
  [7464] = {.lex_state = 179, .external_lex_state = 55},
  [7465] = {.lex_state = 444, .external_lex_state = 46},
  [7466] = {.lex_state = 179, .external_lex_state = 55},
  [7467] = {.lex_state = 179, .external_lex_state = 55},
  [7468] = {.lex_state = 179, .external_lex_state = 55},
  [7469] = {.lex_state = 445, .external_lex_state = 23},
  [7470] = {.lex_state = 179, .external_lex_state = 55},
  [7471] = {.lex_state = 179, .external_lex_state = 55},
  [7472] = {.lex_state = 445, .external_lex_state = 23},
  [7473] = {.lex_state = 445, .external_lex_state = 23},
  [7474] = {.lex_state = 445, .external_lex_state = 2},
  [7475] = {.lex_state = 445, .external_lex_state = 20},
  [7476] = {.lex_state = 445, .external_lex_state = 20},
  [7477] = {.lex_state = 179, .external_lex_state = 55},
  [7478] = {.lex_state = 179, .external_lex_state = 55},
  [7479] = {.lex_state = 445, .external_lex_state = 2},
  [7480] = {.lex_state = 179, .external_lex_state = 55},
  [7481] = {.lex_state = 445, .external_lex_state = 2},
  [7482] = {.lex_state = 444, .external_lex_state = 46},
  [7483] = {.lex_state = 445, .external_lex_state = 23},
  [7484] = {.lex_state = 445, .external_lex_state = 20},
  [7485] = {.lex_state = 445, .external_lex_state = 21},
  [7486] = {.lex_state = 445, .external_lex_state = 21},
  [7487] = {.lex_state = 445, .external_lex_state = 2},
  [7488] = {.lex_state = 445, .external_lex_state = 16},
  [7489] = {.lex_state = 444, .external_lex_state = 46},
  [7490] = {.lex_state = 445, .external_lex_state = 20},
  [7491] = {.lex_state = 445, .external_lex_state = 20},
  [7492] = {.lex_state = 445, .external_lex_state = 2},
  [7493] = {.lex_state = 179, .external_lex_state = 55},
  [7494] = {.lex_state = 179, .external_lex_state = 55},
  [7495] = {.lex_state = 445, .external_lex_state = 16},
  [7496] = {.lex_state = 445, .external_lex_state = 16},
  [7497] = {.lex_state = 444, .external_lex_state = 50},
  [7498] = {.lex_state = 445, .external_lex_state = 16},
  [7499] = {.lex_state = 445, .external_lex_state = 16},
  [7500] = {.lex_state = 179, .external_lex_state = 55},
  [7501] = {.lex_state = 179, .external_lex_state = 55},
  [7502] = {.lex_state = 445, .external_lex_state = 16},
  [7503] = {.lex_state = 445, .external_lex_state = 16},
  [7504] = {.lex_state = 444, .external_lex_state = 51},
  [7505] = {.lex_state = 179, .external_lex_state = 55},
  [7506] = {.lex_state = 445, .external_lex_state = 16},
  [7507] = {.lex_state = 179, .external_lex_state = 55},
  [7508] = {.lex_state = 445, .external_lex_state = 16},
  [7509] = {.lex_state = 445, .external_lex_state = 16},
  [7510] = {.lex_state = 445, .external_lex_state = 16},
  [7511] = {.lex_state = 179, .external_lex_state = 55},
  [7512] = {.lex_state = 179, .external_lex_state = 55},
  [7513] = {.lex_state = 179, .external_lex_state = 55},
  [7514] = {.lex_state = 179, .external_lex_state = 55},
  [7515] = {.lex_state = 444, .external_lex_state = 50},
  [7516] = {.lex_state = 445, .external_lex_state = 21},
  [7517] = {.lex_state = 179, .external_lex_state = 55},
  [7518] = {.lex_state = 179, .external_lex_state = 55},
  [7519] = {.lex_state = 444, .external_lex_state = 50},
  [7520] = {.lex_state = 444, .external_lex_state = 51},
  [7521] = {.lex_state = 179, .external_lex_state = 55},
  [7522] = {.lex_state = 179, .external_lex_state = 55},
  [7523] = {.lex_state = 179, .external_lex_state = 55},
  [7524] = {.lex_state = 445, .external_lex_state = 16},
  [7525] = {.lex_state = 179, .external_lex_state = 55},
  [7526] = {.lex_state = 445, .external_lex_state = 16},
  [7527] = {.lex_state = 445, .external_lex_state = 16},
  [7528] = {.lex_state = 179, .external_lex_state = 55},
  [7529] = {.lex_state = 179, .external_lex_state = 55},
  [7530] = {.lex_state = 444, .external_lex_state = 51},
  [7531] = {.lex_state = 179, .external_lex_state = 55},
  [7532] = {.lex_state = 444, .external_lex_state = 51},
  [7533] = {.lex_state = 179, .external_lex_state = 55},
  [7534] = {.lex_state = 179, .external_lex_state = 55},
  [7535] = {.lex_state = 179, .external_lex_state = 55},
  [7536] = {.lex_state = 179, .external_lex_state = 55},
  [7537] = {.lex_state = 206, .external_lex_state = 65},
  [7538] = {.lex_state = 179, .external_lex_state = 55},
  [7539] = {.lex_state = 179, .external_lex_state = 55},
  [7540] = {.lex_state = 179, .external_lex_state = 55},
  [7541] = {.lex_state = 179, .external_lex_state = 56},
  [7542] = {.lex_state = 444, .external_lex_state = 51},
  [7543] = {.lex_state = 179, .external_lex_state = 55},
  [7544] = {.lex_state = 179, .external_lex_state = 55},
  [7545] = {.lex_state = 179, .external_lex_state = 55},
  [7546] = {.lex_state = 179, .external_lex_state = 55},
  [7547] = {.lex_state = 179, .external_lex_state = 55},
  [7548] = {.lex_state = 445, .external_lex_state = 20},
  [7549] = {.lex_state = 179, .external_lex_state = 55},
  [7550] = {.lex_state = 179, .external_lex_state = 55},
  [7551] = {.lex_state = 445, .external_lex_state = 20},
  [7552] = {.lex_state = 179, .external_lex_state = 55},
  [7553] = {.lex_state = 445, .external_lex_state = 20},
  [7554] = {.lex_state = 179, .external_lex_state = 55},
  [7555] = {.lex_state = 179, .external_lex_state = 55},
  [7556] = {.lex_state = 179, .external_lex_state = 55},
  [7557] = {.lex_state = 179, .external_lex_state = 55},
  [7558] = {.lex_state = 179, .external_lex_state = 55},
  [7559] = {.lex_state = 179, .external_lex_state = 55},
  [7560] = {.lex_state = 179, .external_lex_state = 55},
  [7561] = {.lex_state = 444, .external_lex_state = 53},
  [7562] = {.lex_state = 179, .external_lex_state = 55},
  [7563] = {.lex_state = 179, .external_lex_state = 55},
  [7564] = {.lex_state = 179, .external_lex_state = 55},
  [7565] = {.lex_state = 179, .external_lex_state = 55},
  [7566] = {.lex_state = 179, .external_lex_state = 55},
  [7567] = {.lex_state = 179, .external_lex_state = 55},
  [7568] = {.lex_state = 179, .external_lex_state = 55},
  [7569] = {.lex_state = 179, .external_lex_state = 55},
  [7570] = {.lex_state = 179, .external_lex_state = 55},
  [7571] = {.lex_state = 206, .external_lex_state = 65},
  [7572] = {.lex_state = 444, .external_lex_state = 50},
  [7573] = {.lex_state = 179, .external_lex_state = 55},
  [7574] = {.lex_state = 179, .external_lex_state = 55},
  [7575] = {.lex_state = 206, .external_lex_state = 65},
  [7576] = {.lex_state = 179, .external_lex_state = 55},
  [7577] = {.lex_state = 179, .external_lex_state = 55},
  [7578] = {.lex_state = 179, .external_lex_state = 55},
  [7579] = {.lex_state = 206, .external_lex_state = 65},
  [7580] = {.lex_state = 179, .external_lex_state = 55},
  [7581] = {.lex_state = 451, .external_lex_state = 46},
  [7582] = {.lex_state = 179, .external_lex_state = 55},
  [7583] = {.lex_state = 206, .external_lex_state = 65},
  [7584] = {.lex_state = 179, .external_lex_state = 55},
  [7585] = {.lex_state = 179, .external_lex_state = 55},
  [7586] = {.lex_state = 179, .external_lex_state = 55},
  [7587] = {.lex_state = 206, .external_lex_state = 65},
  [7588] = {.lex_state = 179, .external_lex_state = 55},
  [7589] = {.lex_state = 179, .external_lex_state = 55},
  [7590] = {.lex_state = 179, .external_lex_state = 55},
  [7591] = {.lex_state = 206, .external_lex_state = 65},
  [7592] = {.lex_state = 450, .external_lex_state = 50},
  [7593] = {.lex_state = 179, .external_lex_state = 55},
  [7594] = {.lex_state = 179, .external_lex_state = 55},
  [7595] = {.lex_state = 206, .external_lex_state = 65},
  [7596] = {.lex_state = 179, .external_lex_state = 55},
  [7597] = {.lex_state = 179, .external_lex_state = 55},
  [7598] = {.lex_state = 179, .external_lex_state = 55},
  [7599] = {.lex_state = 179, .external_lex_state = 55},
  [7600] = {.lex_state = 179, .external_lex_state = 55},
  [7601] = {.lex_state = 179, .external_lex_state = 55},
  [7602] = {.lex_state = 179, .external_lex_state = 55},
  [7603] = {.lex_state = 179, .external_lex_state = 55},
  [7604] = {.lex_state = 444, .external_lex_state = 50},
  [7605] = {.lex_state = 179, .external_lex_state = 55},
  [7606] = {.lex_state = 179, .external_lex_state = 55},
  [7607] = {.lex_state = 179, .external_lex_state = 55},
  [7608] = {.lex_state = 179, .external_lex_state = 55},
  [7609] = {.lex_state = 451, .external_lex_state = 46},
  [7610] = {.lex_state = 445, .external_lex_state = 23},
  [7611] = {.lex_state = 179, .external_lex_state = 55},
  [7612] = {.lex_state = 445, .external_lex_state = 23},
  [7613] = {.lex_state = 179, .external_lex_state = 55},
  [7614] = {.lex_state = 445, .external_lex_state = 21},
  [7615] = {.lex_state = 179, .external_lex_state = 55},
  [7616] = {.lex_state = 445, .external_lex_state = 21},
  [7617] = {.lex_state = 179, .external_lex_state = 55},
  [7618] = {.lex_state = 179, .external_lex_state = 55},
  [7619] = {.lex_state = 445, .external_lex_state = 20},
  [7620] = {.lex_state = 445, .external_lex_state = 21},
  [7621] = {.lex_state = 179, .external_lex_state = 55},
  [7622] = {.lex_state = 179, .external_lex_state = 55},
  [7623] = {.lex_state = 179, .external_lex_state = 55},
  [7624] = {.lex_state = 445, .external_lex_state = 2},
  [7625] = {.lex_state = 196, .external_lex_state = 66},
  [7626] = {.lex_state = 450, .external_lex_state = 51},
  [7627] = {.lex_state = 445, .external_lex_state = 2},
  [7628] = {.lex_state = 179, .external_lex_state = 55},
  [7629] = {.lex_state = 179, .external_lex_state = 55},
  [7630] = {.lex_state = 451, .external_lex_state = 46},
  [7631] = {.lex_state = 179, .external_lex_state = 55},
  [7632] = {.lex_state = 179, .external_lex_state = 55},
  [7633] = {.lex_state = 179, .external_lex_state = 55},
  [7634] = {.lex_state = 450, .external_lex_state = 53},
  [7635] = {.lex_state = 179, .external_lex_state = 55},
  [7636] = {.lex_state = 451, .external_lex_state = 46},
  [7637] = {.lex_state = 179, .external_lex_state = 55},
  [7638] = {.lex_state = 179, .external_lex_state = 55},
  [7639] = {.lex_state = 445, .external_lex_state = 2},
  [7640] = {.lex_state = 179, .external_lex_state = 55},
  [7641] = {.lex_state = 179, .external_lex_state = 55},
  [7642] = {.lex_state = 179, .external_lex_state = 55},
  [7643] = {.lex_state = 179, .external_lex_state = 55},
  [7644] = {.lex_state = 179, .external_lex_state = 55},
  [7645] = {.lex_state = 179, .external_lex_state = 55},
  [7646] = {.lex_state = 179, .external_lex_state = 55},
  [7647] = {.lex_state = 179, .external_lex_state = 55},
  [7648] = {.lex_state = 445, .external_lex_state = 23},
  [7649] = {.lex_state = 179, .external_lex_state = 55},
  [7650] = {.lex_state = 179, .external_lex_state = 55},
  [7651] = {.lex_state = 179, .external_lex_state = 55},
  [7652] = {.lex_state = 445, .external_lex_state = 21},
  [7653] = {.lex_state = 179, .external_lex_state = 55},
  [7654] = {.lex_state = 445, .external_lex_state = 2},
  [7655] = {.lex_state = 445, .external_lex_state = 2},
  [7656] = {.lex_state = 179, .external_lex_state = 55},
  [7657] = {.lex_state = 445, .external_lex_state = 16},
  [7658] = {.lex_state = 179, .external_lex_state = 55},
  [7659] = {.lex_state = 179, .external_lex_state = 55},
  [7660] = {.lex_state = 445, .external_lex_state = 23},
  [7661] = {.lex_state = 444, .external_lex_state = 53},
  [7662] = {.lex_state = 444, .external_lex_state = 53},
  [7663] = {.lex_state = 444, .external_lex_state = 53},
  [7664] = {.lex_state = 445, .external_lex_state = 23},
  [7665] = {.lex_state = 444, .external_lex_state = 53},
  [7666] = {.lex_state = 179, .external_lex_state = 55},
  [7667] = {.lex_state = 445, .external_lex_state = 20},
  [7668] = {.lex_state = 179, .external_lex_state = 55},
  [7669] = {.lex_state = 179, .external_lex_state = 55},
  [7670] = {.lex_state = 179, .external_lex_state = 55},
  [7671] = {.lex_state = 196, .external_lex_state = 55},
  [7672] = {.lex_state = 445, .external_lex_state = 21},
  [7673] = {.lex_state = 179, .external_lex_state = 55},
  [7674] = {.lex_state = 445, .external_lex_state = 21},
  [7675] = {.lex_state = 445, .external_lex_state = 21},
  [7676] = {.lex_state = 179, .external_lex_state = 55},
  [7677] = {.lex_state = 445, .external_lex_state = 21},
  [7678] = {.lex_state = 445, .external_lex_state = 21},
  [7679] = {.lex_state = 445, .external_lex_state = 21},
  [7680] = {.lex_state = 196, .external_lex_state = 55},
  [7681] = {.lex_state = 445, .external_lex_state = 21},
  [7682] = {.lex_state = 445, .external_lex_state = 2},
  [7683] = {.lex_state = 179, .external_lex_state = 55},
  [7684] = {.lex_state = 445, .external_lex_state = 21},
  [7685] = {.lex_state = 445, .external_lex_state = 21},
  [7686] = {.lex_state = 445, .external_lex_state = 21},
  [7687] = {.lex_state = 196, .external_lex_state = 55},
  [7688] = {.lex_state = 196, .external_lex_state = 55},
  [7689] = {.lex_state = 179, .external_lex_state = 55},
  [7690] = {.lex_state = 196, .external_lex_state = 55},
  [7691] = {.lex_state = 444, .external_lex_state = 53},
  [7692] = {.lex_state = 179, .external_lex_state = 55},
  [7693] = {.lex_state = 196, .external_lex_state = 55},
  [7694] = {.lex_state = 179, .external_lex_state = 55},
  [7695] = {.lex_state = 179, .external_lex_state = 55},
  [7696] = {.lex_state = 179, .external_lex_state = 55},
  [7697] = {.lex_state = 196, .external_lex_state = 55},
  [7698] = {.lex_state = 179, .external_lex_state = 55},
  [7699] = {.lex_state = 179, .external_lex_state = 55},
  [7700] = {.lex_state = 179, .external_lex_state = 55},
  [7701] = {.lex_state = 196, .external_lex_state = 55},
  [7702] = {.lex_state = 179, .external_lex_state = 55},
  [7703] = {.lex_state = 179, .external_lex_state = 55},
  [7704] = {.lex_state = 196, .external_lex_state = 55},
  [7705] = {.lex_state = 179, .external_lex_state = 55},
  [7706] = {.lex_state = 179, .external_lex_state = 55},
  [7707] = {.lex_state = 447, .external_lex_state = 16},
  [7708] = {.lex_state = 444, .external_lex_state = 53},
  [7709] = {.lex_state = 196, .external_lex_state = 55},
  [7710] = {.lex_state = 179, .external_lex_state = 55},
  [7711] = {.lex_state = 179, .external_lex_state = 55},
  [7712] = {.lex_state = 179, .external_lex_state = 55},
  [7713] = {.lex_state = 179, .external_lex_state = 55},
  [7714] = {.lex_state = 179, .external_lex_state = 55},
  [7715] = {.lex_state = 179, .external_lex_state = 55},
  [7716] = {.lex_state = 179, .external_lex_state = 55},
  [7717] = {.lex_state = 196, .external_lex_state = 55},
  [7718] = {.lex_state = 179, .external_lex_state = 55},
  [7719] = {.lex_state = 179, .external_lex_state = 55},
  [7720] = {.lex_state = 444, .external_lex_state = 53},
  [7721] = {.lex_state = 179, .external_lex_state = 55},
  [7722] = {.lex_state = 179, .external_lex_state = 55},
  [7723] = {.lex_state = 179, .external_lex_state = 55},
  [7724] = {.lex_state = 196, .external_lex_state = 55},
  [7725] = {.lex_state = 179, .external_lex_state = 55},
  [7726] = {.lex_state = 445, .external_lex_state = 23},
  [7727] = {.lex_state = 196, .external_lex_state = 55},
  [7728] = {.lex_state = 445, .external_lex_state = 23},
  [7729] = {.lex_state = 196, .external_lex_state = 55},
  [7730] = {.lex_state = 444, .external_lex_state = 53},
  [7731] = {.lex_state = 445, .external_lex_state = 23},
  [7732] = {.lex_state = 196, .external_lex_state = 55},
  [7733] = {.lex_state = 196, .external_lex_state = 55},
  [7734] = {.lex_state = 196, .external_lex_state = 55},
  [7735] = {.lex_state = 444, .external_lex_state = 54},
  [7736] = {.lex_state = 445, .external_lex_state = 20},
  [7737] = {.lex_state = 196, .external_lex_state = 55},
  [7738] = {.lex_state = 179, .external_lex_state = 55},
  [7739] = {.lex_state = 445, .external_lex_state = 20},
  [7740] = {.lex_state = 445, .external_lex_state = 20},
  [7741] = {.lex_state = 445, .external_lex_state = 20},
  [7742] = {.lex_state = 196, .external_lex_state = 55},
  [7743] = {.lex_state = 445, .external_lex_state = 20},
  [7744] = {.lex_state = 179, .external_lex_state = 55},
  [7745] = {.lex_state = 445, .external_lex_state = 20},
  [7746] = {.lex_state = 445, .external_lex_state = 20},
  [7747] = {.lex_state = 179, .external_lex_state = 55},
  [7748] = {.lex_state = 445, .external_lex_state = 20},
  [7749] = {.lex_state = 179, .external_lex_state = 55},
  [7750] = {.lex_state = 196, .external_lex_state = 55},
  [7751] = {.lex_state = 179, .external_lex_state = 55},
  [7752] = {.lex_state = 445, .external_lex_state = 20},
  [7753] = {.lex_state = 445, .external_lex_state = 20},
  [7754] = {.lex_state = 445, .external_lex_state = 20},
  [7755] = {.lex_state = 179, .external_lex_state = 55},
  [7756] = {.lex_state = 196, .external_lex_state = 55},
  [7757] = {.lex_state = 179, .external_lex_state = 55},
  [7758] = {.lex_state = 179, .external_lex_state = 55},
  [7759] = {.lex_state = 179, .external_lex_state = 55},
  [7760] = {.lex_state = 196, .external_lex_state = 55},
  [7761] = {.lex_state = 196, .external_lex_state = 55},
  [7762] = {.lex_state = 196, .external_lex_state = 55},
  [7763] = {.lex_state = 196, .external_lex_state = 55},
  [7764] = {.lex_state = 196, .external_lex_state = 55},
  [7765] = {.lex_state = 179, .external_lex_state = 55},
  [7766] = {.lex_state = 196, .external_lex_state = 55},
  [7767] = {.lex_state = 196, .external_lex_state = 55},
  [7768] = {.lex_state = 179, .external_lex_state = 55},
  [7769] = {.lex_state = 196, .external_lex_state = 55},
  [7770] = {.lex_state = 196, .external_lex_state = 55},
  [7771] = {.lex_state = 196, .external_lex_state = 55},
  [7772] = {.lex_state = 196, .external_lex_state = 55},
  [7773] = {.lex_state = 196, .external_lex_state = 55},
  [7774] = {.lex_state = 445, .external_lex_state = 2},
  [7775] = {.lex_state = 196, .external_lex_state = 55},
  [7776] = {.lex_state = 196, .external_lex_state = 55},
  [7777] = {.lex_state = 196, .external_lex_state = 55},
  [7778] = {.lex_state = 196, .external_lex_state = 55},
  [7779] = {.lex_state = 196, .external_lex_state = 55},
  [7780] = {.lex_state = 196, .external_lex_state = 55},
  [7781] = {.lex_state = 196, .external_lex_state = 55},
  [7782] = {.lex_state = 196, .external_lex_state = 55},
  [7783] = {.lex_state = 196, .external_lex_state = 55},
  [7784] = {.lex_state = 196, .external_lex_state = 55},
  [7785] = {.lex_state = 196, .external_lex_state = 55},
  [7786] = {.lex_state = 196, .external_lex_state = 55},
  [7787] = {.lex_state = 196, .external_lex_state = 55},
  [7788] = {.lex_state = 196, .external_lex_state = 55},
  [7789] = {.lex_state = 196, .external_lex_state = 55},
  [7790] = {.lex_state = 196, .external_lex_state = 55},
  [7791] = {.lex_state = 196, .external_lex_state = 55},
  [7792] = {.lex_state = 196, .external_lex_state = 55},
  [7793] = {.lex_state = 196, .external_lex_state = 55},
  [7794] = {.lex_state = 196, .external_lex_state = 55},
  [7795] = {.lex_state = 196, .external_lex_state = 55},
  [7796] = {.lex_state = 196, .external_lex_state = 55},
  [7797] = {.lex_state = 196, .external_lex_state = 55},
  [7798] = {.lex_state = 196, .external_lex_state = 55},
  [7799] = {.lex_state = 196, .external_lex_state = 55},
  [7800] = {.lex_state = 196, .external_lex_state = 55},
  [7801] = {.lex_state = 196, .external_lex_state = 55},
  [7802] = {.lex_state = 196, .external_lex_state = 55},
  [7803] = {.lex_state = 196, .external_lex_state = 55},
  [7804] = {.lex_state = 196, .external_lex_state = 55},
  [7805] = {.lex_state = 196, .external_lex_state = 55},
  [7806] = {.lex_state = 196, .external_lex_state = 55},
  [7807] = {.lex_state = 196, .external_lex_state = 55},
  [7808] = {.lex_state = 196, .external_lex_state = 55},
  [7809] = {.lex_state = 196, .external_lex_state = 55},
  [7810] = {.lex_state = 196, .external_lex_state = 55},
  [7811] = {.lex_state = 196, .external_lex_state = 55},
  [7812] = {.lex_state = 196, .external_lex_state = 55},
  [7813] = {.lex_state = 179, .external_lex_state = 55},
  [7814] = {.lex_state = 196, .external_lex_state = 55},
  [7815] = {.lex_state = 196, .external_lex_state = 55},
  [7816] = {.lex_state = 196, .external_lex_state = 55},
  [7817] = {.lex_state = 196, .external_lex_state = 55},
  [7818] = {.lex_state = 196, .external_lex_state = 55},
  [7819] = {.lex_state = 196, .external_lex_state = 55},
  [7820] = {.lex_state = 196, .external_lex_state = 55},
  [7821] = {.lex_state = 196, .external_lex_state = 55},
  [7822] = {.lex_state = 196, .external_lex_state = 55},
  [7823] = {.lex_state = 196, .external_lex_state = 55},
  [7824] = {.lex_state = 196, .external_lex_state = 55},
  [7825] = {.lex_state = 196, .external_lex_state = 55},
  [7826] = {.lex_state = 196, .external_lex_state = 55},
  [7827] = {.lex_state = 196, .external_lex_state = 55},
  [7828] = {.lex_state = 196, .external_lex_state = 55},
  [7829] = {.lex_state = 196, .external_lex_state = 55},
  [7830] = {.lex_state = 196, .external_lex_state = 55},
  [7831] = {.lex_state = 196, .external_lex_state = 55},
  [7832] = {.lex_state = 196, .external_lex_state = 55},
  [7833] = {.lex_state = 196, .external_lex_state = 55},
  [7834] = {.lex_state = 196, .external_lex_state = 55},
  [7835] = {.lex_state = 196, .external_lex_state = 55},
  [7836] = {.lex_state = 196, .external_lex_state = 55},
  [7837] = {.lex_state = 196, .external_lex_state = 55},
  [7838] = {.lex_state = 451, .external_lex_state = 50},
  [7839] = {.lex_state = 454, .external_lex_state = 46},
  [7840] = {.lex_state = 179, .external_lex_state = 55},
  [7841] = {.lex_state = 196, .external_lex_state = 55},
  [7842] = {.lex_state = 445, .external_lex_state = 2},
  [7843] = {.lex_state = 445, .external_lex_state = 2},
  [7844] = {.lex_state = 179, .external_lex_state = 55},
  [7845] = {.lex_state = 179, .external_lex_state = 55},
  [7846] = {.lex_state = 196, .external_lex_state = 55},
  [7847] = {.lex_state = 451, .external_lex_state = 50},
  [7848] = {.lex_state = 445, .external_lex_state = 20},
  [7849] = {.lex_state = 445, .external_lex_state = 20},
  [7850] = {.lex_state = 445, .external_lex_state = 2},
  [7851] = {.lex_state = 179, .external_lex_state = 55},
  [7852] = {.lex_state = 451, .external_lex_state = 50},
  [7853] = {.lex_state = 445, .external_lex_state = 20},
  [7854] = {.lex_state = 445, .external_lex_state = 2},
  [7855] = {.lex_state = 179, .external_lex_state = 55},
  [7856] = {.lex_state = 444, .external_lex_state = 54},
  [7857] = {.lex_state = 179, .external_lex_state = 55},
  [7858] = {.lex_state = 196, .external_lex_state = 55},
  [7859] = {.lex_state = 179, .external_lex_state = 55},
  [7860] = {.lex_state = 179, .external_lex_state = 55},
  [7861] = {.lex_state = 179, .external_lex_state = 55},
  [7862] = {.lex_state = 179, .external_lex_state = 55},
  [7863] = {.lex_state = 179, .external_lex_state = 55},
  [7864] = {.lex_state = 444, .external_lex_state = 53},
  [7865] = {.lex_state = 179, .external_lex_state = 55},
  [7866] = {.lex_state = 445, .external_lex_state = 23},
  [7867] = {.lex_state = 444, .external_lex_state = 54},
  [7868] = {.lex_state = 179, .external_lex_state = 55},
  [7869] = {.lex_state = 450, .external_lex_state = 54},
  [7870] = {.lex_state = 179, .external_lex_state = 55},
  [7871] = {.lex_state = 179, .external_lex_state = 55},
  [7872] = {.lex_state = 444, .external_lex_state = 54},
  [7873] = {.lex_state = 179, .external_lex_state = 55},
  [7874] = {.lex_state = 444, .external_lex_state = 46},
  [7875] = {.lex_state = 196, .external_lex_state = 55},
  [7876] = {.lex_state = 451, .external_lex_state = 51},
  [7877] = {.lex_state = 451, .external_lex_state = 50},
  [7878] = {.lex_state = 451, .external_lex_state = 51},
  [7879] = {.lex_state = 179, .external_lex_state = 55},
  [7880] = {.lex_state = 179, .external_lex_state = 55},
  [7881] = {.lex_state = 451, .external_lex_state = 51},
  [7882] = {.lex_state = 179, .external_lex_state = 55},
  [7883] = {.lex_state = 454, .external_lex_state = 46},
  [7884] = {.lex_state = 444, .external_lex_state = 54},
  [7885] = {.lex_state = 445, .external_lex_state = 2},
  [7886] = {.lex_state = 451, .external_lex_state = 51},
  [7887] = {.lex_state = 179, .external_lex_state = 55},
  [7888] = {.lex_state = 444, .external_lex_state = 46},
  [7889] = {.lex_state = 445, .external_lex_state = 2},
  [7890] = {.lex_state = 179, .external_lex_state = 55},
  [7891] = {.lex_state = 179, .external_lex_state = 55},
  [7892] = {.lex_state = 447, .external_lex_state = 16},
  [7893] = {.lex_state = 444, .external_lex_state = 46},
  [7894] = {.lex_state = 447, .external_lex_state = 16},
  [7895] = {.lex_state = 451, .external_lex_state = 53},
  [7896] = {.lex_state = 451, .external_lex_state = 53},
  [7897] = {.lex_state = 179, .external_lex_state = 55},
  [7898] = {.lex_state = 179, .external_lex_state = 55},
  [7899] = {.lex_state = 445, .external_lex_state = 2},
  [7900] = {.lex_state = 179, .external_lex_state = 55},
  [7901] = {.lex_state = 196, .external_lex_state = 55},
  [7902] = {.lex_state = 451, .external_lex_state = 53},
  [7903] = {.lex_state = 179, .external_lex_state = 55},
  [7904] = {.lex_state = 445, .external_lex_state = 2},
  [7905] = {.lex_state = 179, .external_lex_state = 55},
  [7906] = {.lex_state = 179, .external_lex_state = 55},
  [7907] = {.lex_state = 179, .external_lex_state = 55},
  [7908] = {.lex_state = 179, .external_lex_state = 55},
  [7909] = {.lex_state = 445, .external_lex_state = 2},
  [7910] = {.lex_state = 179, .external_lex_state = 55},
  [7911] = {.lex_state = 179, .external_lex_state = 55},
  [7912] = {.lex_state = 179, .external_lex_state = 55},
  [7913] = {.lex_state = 451, .external_lex_state = 53},
  [7914] = {.lex_state = 179, .external_lex_state = 55},
  [7915] = {.lex_state = 179, .external_lex_state = 55},
  [7916] = {.lex_state = 179, .external_lex_state = 55},
  [7917] = {.lex_state = 179, .external_lex_state = 55},
  [7918] = {.lex_state = 179, .external_lex_state = 55},
  [7919] = {.lex_state = 179, .external_lex_state = 55},
  [7920] = {.lex_state = 179, .external_lex_state = 55},
  [7921] = {.lex_state = 179, .external_lex_state = 55},
  [7922] = {.lex_state = 179, .external_lex_state = 55},
  [7923] = {.lex_state = 179, .external_lex_state = 55},
  [7924] = {.lex_state = 179, .external_lex_state = 55},
  [7925] = {.lex_state = 179, .external_lex_state = 55},
  [7926] = {.lex_state = 179, .external_lex_state = 55},
  [7927] = {.lex_state = 179, .external_lex_state = 55},
  [7928] = {.lex_state = 445, .external_lex_state = 2},
  [7929] = {.lex_state = 179, .external_lex_state = 55},
  [7930] = {.lex_state = 444, .external_lex_state = 46},
  [7931] = {.lex_state = 179, .external_lex_state = 55},
  [7932] = {.lex_state = 179, .external_lex_state = 55},
  [7933] = {.lex_state = 444, .external_lex_state = 46},
  [7934] = {.lex_state = 179, .external_lex_state = 55},
  [7935] = {.lex_state = 445, .external_lex_state = 2},
  [7936] = {.lex_state = 196, .external_lex_state = 55},
  [7937] = {.lex_state = 445, .external_lex_state = 2},
  [7938] = {.lex_state = 179, .external_lex_state = 55},
  [7939] = {.lex_state = 445, .external_lex_state = 21},
  [7940] = {.lex_state = 445, .external_lex_state = 23},
  [7941] = {.lex_state = 179, .external_lex_state = 55},
  [7942] = {.lex_state = 445, .external_lex_state = 21},
  [7943] = {.lex_state = 445, .external_lex_state = 21},
  [7944] = {.lex_state = 179, .external_lex_state = 55},
  [7945] = {.lex_state = 179, .external_lex_state = 55},
  [7946] = {.lex_state = 179, .external_lex_state = 55},
  [7947] = {.lex_state = 445, .external_lex_state = 21},
  [7948] = {.lex_state = 179, .external_lex_state = 55},
  [7949] = {.lex_state = 179, .external_lex_state = 55},
  [7950] = {.lex_state = 179, .external_lex_state = 55},
  [7951] = {.lex_state = 179, .external_lex_state = 55},
  [7952] = {.lex_state = 179, .external_lex_state = 55},
  [7953] = {.lex_state = 179, .external_lex_state = 55},
  [7954] = {.lex_state = 179, .external_lex_state = 55},
  [7955] = {.lex_state = 179, .external_lex_state = 55},
  [7956] = {.lex_state = 179, .external_lex_state = 55},
  [7957] = {.lex_state = 179, .external_lex_state = 55},
  [7958] = {.lex_state = 179, .external_lex_state = 55},
  [7959] = {.lex_state = 179, .external_lex_state = 55},
  [7960] = {.lex_state = 179, .external_lex_state = 55},
  [7961] = {.lex_state = 179, .external_lex_state = 55},
  [7962] = {.lex_state = 179, .external_lex_state = 55},
  [7963] = {.lex_state = 179, .external_lex_state = 55},
  [7964] = {.lex_state = 179, .external_lex_state = 55},
  [7965] = {.lex_state = 179, .external_lex_state = 55},
  [7966] = {.lex_state = 179, .external_lex_state = 55},
  [7967] = {.lex_state = 179, .external_lex_state = 55},
  [7968] = {.lex_state = 454, .external_lex_state = 46},
  [7969] = {.lex_state = 179, .external_lex_state = 55},
  [7970] = {.lex_state = 179, .external_lex_state = 55},
  [7971] = {.lex_state = 179, .external_lex_state = 55},
  [7972] = {.lex_state = 179, .external_lex_state = 55},
  [7973] = {.lex_state = 179, .external_lex_state = 55},
  [7974] = {.lex_state = 179, .external_lex_state = 55},
  [7975] = {.lex_state = 179, .external_lex_state = 55},
  [7976] = {.lex_state = 179, .external_lex_state = 55},
  [7977] = {.lex_state = 179, .external_lex_state = 55},
  [7978] = {.lex_state = 179, .external_lex_state = 55},
  [7979] = {.lex_state = 179, .external_lex_state = 55},
  [7980] = {.lex_state = 179, .external_lex_state = 55},
  [7981] = {.lex_state = 179, .external_lex_state = 55},
  [7982] = {.lex_state = 196, .external_lex_state = 55},
  [7983] = {.lex_state = 444, .external_lex_state = 46},
  [7984] = {.lex_state = 445, .external_lex_state = 23},
  [7985] = {.lex_state = 451, .external_lex_state = 54},
  [7986] = {.lex_state = 447, .external_lex_state = 16},
  [7987] = {.lex_state = 444, .external_lex_state = 46},
  [7988] = {.lex_state = 444, .external_lex_state = 46},
  [7989] = {.lex_state = 445, .external_lex_state = 23},
  [7990] = {.lex_state = 444, .external_lex_state = 53},
  [7991] = {.lex_state = 445, .external_lex_state = 23},
  [7992] = {.lex_state = 444, .external_lex_state = 53},
  [7993] = {.lex_state = 196, .external_lex_state = 55},
  [7994] = {.lex_state = 444, .external_lex_state = 50},
  [7995] = {.lex_state = 454, .external_lex_state = 50},
  [7996] = {.lex_state = 444, .external_lex_state = 53},
  [7997] = {.lex_state = 444, .external_lex_state = 53},
  [7998] = {.lex_state = 444, .external_lex_state = 53},
  [7999] = {.lex_state = 454, .external_lex_state = 51},
  [8000] = {.lex_state = 444, .external_lex_state = 50},
  [8001] = {.lex_state = 445, .external_lex_state = 23},
  [8002] = {.lex_state = 444, .external_lex_state = 46},
  [8003] = {.lex_state = 196, .external_lex_state = 55},
  [8004] = {.lex_state = 445, .external_lex_state = 23},
  [8005] = {.lex_state = 444, .external_lex_state = 46},
  [8006] = {.lex_state = 444, .external_lex_state = 46},
  [8007] = {.lex_state = 196, .external_lex_state = 55},
  [8008] = {.lex_state = 444, .external_lex_state = 46},
  [8009] = {.lex_state = 444, .external_lex_state = 46},
  [8010] = {.lex_state = 454, .external_lex_state = 51},
  [8011] = {.lex_state = 444, .external_lex_state = 46},
  [8012] = {.lex_state = 444, .external_lex_state = 46},
  [8013] = {.lex_state = 196, .external_lex_state = 55},
  [8014] = {.lex_state = 454, .external_lex_state = 46},
  [8015] = {.lex_state = 444, .external_lex_state = 46},
  [8016] = {.lex_state = 445, .external_lex_state = 23},
  [8017] = {.lex_state = 196, .external_lex_state = 55},
  [8018] = {.lex_state = 451, .external_lex_state = 54},
  [8019] = {.lex_state = 454, .external_lex_state = 46},
  [8020] = {.lex_state = 445, .external_lex_state = 23},
  [8021] = {.lex_state = 447, .external_lex_state = 21},
  [8022] = {.lex_state = 196, .external_lex_state = 55},
  [8023] = {.lex_state = 445, .external_lex_state = 23},
  [8024] = {.lex_state = 454, .external_lex_state = 46},
  [8025] = {.lex_state = 445, .external_lex_state = 23},
  [8026] = {.lex_state = 447, .external_lex_state = 21},
  [8027] = {.lex_state = 454, .external_lex_state = 46},
  [8028] = {.lex_state = 196, .external_lex_state = 55},
  [8029] = {.lex_state = 444, .external_lex_state = 46},
  [8030] = {.lex_state = 447, .external_lex_state = 20},
  [8031] = {.lex_state = 444, .external_lex_state = 46},
  [8032] = {.lex_state = 445, .external_lex_state = 23},
  [8033] = {.lex_state = 443, .external_lex_state = 16},
  [8034] = {.lex_state = 445, .external_lex_state = 23},
  [8035] = {.lex_state = 444, .external_lex_state = 46},
  [8036] = {.lex_state = 447, .external_lex_state = 2},
  [8037] = {.lex_state = 447, .external_lex_state = 21},
  [8038] = {.lex_state = 447, .external_lex_state = 20},
  [8039] = {.lex_state = 447, .external_lex_state = 21},
  [8040] = {.lex_state = 447, .external_lex_state = 20},
  [8041] = {.lex_state = 444, .external_lex_state = 46},
  [8042] = {.lex_state = 447, .external_lex_state = 21},
  [8043] = {.lex_state = 444, .external_lex_state = 46},
  [8044] = {.lex_state = 445, .external_lex_state = 23},
  [8045] = {.lex_state = 196, .external_lex_state = 55},
  [8046] = {.lex_state = 447, .external_lex_state = 21},
  [8047] = {.lex_state = 444, .external_lex_state = 51},
  [8048] = {.lex_state = 444, .external_lex_state = 46},
  [8049] = {.lex_state = 196, .external_lex_state = 55},
  [8050] = {.lex_state = 196, .external_lex_state = 55},
  [8051] = {.lex_state = 444, .external_lex_state = 46},
  [8052] = {.lex_state = 443, .external_lex_state = 16},
  [8053] = {.lex_state = 444, .external_lex_state = 46},
  [8054] = {.lex_state = 444, .external_lex_state = 50},
  [8055] = {.lex_state = 196, .external_lex_state = 55},
  [8056] = {.lex_state = 444, .external_lex_state = 46},
  [8057] = {.lex_state = 444, .external_lex_state = 46},
  [8058] = {.lex_state = 447, .external_lex_state = 2},
  [8059] = {.lex_state = 444, .external_lex_state = 50},
  [8060] = {.lex_state = 454, .external_lex_state = 53},
  [8061] = {.lex_state = 444, .external_lex_state = 46},
  [8062] = {.lex_state = 444, .external_lex_state = 46},
  [8063] = {.lex_state = 451, .external_lex_state = 54},
  [8064] = {.lex_state = 443, .external_lex_state = 16},
  [8065] = {.lex_state = 444, .external_lex_state = 46},
  [8066] = {.lex_state = 454, .external_lex_state = 50},
  [8067] = {.lex_state = 444, .external_lex_state = 50},
  [8068] = {.lex_state = 444, .external_lex_state = 46},
  [8069] = {.lex_state = 444, .external_lex_state = 51},
  [8070] = {.lex_state = 444, .external_lex_state = 46},
  [8071] = {.lex_state = 451, .external_lex_state = 54},
  [8072] = {.lex_state = 454, .external_lex_state = 53},
  [8073] = {.lex_state = 444, .external_lex_state = 51},
  [8074] = {.lex_state = 447, .external_lex_state = 21},
  [8075] = {.lex_state = 454, .external_lex_state = 46},
  [8076] = {.lex_state = 447, .external_lex_state = 2},
  [8077] = {.lex_state = 454, .external_lex_state = 46},
  [8078] = {.lex_state = 444, .external_lex_state = 46},
  [8079] = {.lex_state = 454, .external_lex_state = 51},
  [8080] = {.lex_state = 445, .external_lex_state = 23},
  [8081] = {.lex_state = 454, .external_lex_state = 53},
  [8082] = {.lex_state = 444, .external_lex_state = 46},
  [8083] = {.lex_state = 444, .external_lex_state = 51},
  [8084] = {.lex_state = 444, .external_lex_state = 51},
  [8085] = {.lex_state = 447, .external_lex_state = 21},
  [8086] = {.lex_state = 444, .external_lex_state = 46},
  [8087] = {.lex_state = 445, .external_lex_state = 23},
  [8088] = {.lex_state = 454, .external_lex_state = 50},
  [8089] = {.lex_state = 444, .external_lex_state = 46},
  [8090] = {.lex_state = 444, .external_lex_state = 46},
  [8091] = {.lex_state = 444, .external_lex_state = 46},
  [8092] = {.lex_state = 444, .external_lex_state = 46},
  [8093] = {.lex_state = 444, .external_lex_state = 46},
  [8094] = {.lex_state = 196, .external_lex_state = 55},
  [8095] = {.lex_state = 444, .external_lex_state = 46},
  [8096] = {.lex_state = 444, .external_lex_state = 46},
  [8097] = {.lex_state = 444, .external_lex_state = 54},
  [8098] = {.lex_state = 444, .external_lex_state = 51},
  [8099] = {.lex_state = 447, .external_lex_state = 2},
  [8100] = {.lex_state = 447, .external_lex_state = 23},
  [8101] = {.lex_state = 444, .external_lex_state = 51},
  [8102] = {.lex_state = 454, .external_lex_state = 53},
  [8103] = {.lex_state = 454, .external_lex_state = 54},
  [8104] = {.lex_state = 447, .external_lex_state = 23},
  [8105] = {.lex_state = 444, .external_lex_state = 51},
  [8106] = {.lex_state = 444, .external_lex_state = 51},
  [8107] = {.lex_state = 447, .external_lex_state = 23},
  [8108] = {.lex_state = 447, .external_lex_state = 21},
  [8109] = {.lex_state = 452, .external_lex_state = 16},
  [8110] = {.lex_state = 444, .external_lex_state = 46},
  [8111] = {.lex_state = 444, .external_lex_state = 51},
  [8112] = {.lex_state = 444, .external_lex_state = 46},
  [8113] = {.lex_state = 443, .external_lex_state = 20},
  [8114] = {.lex_state = 444, .external_lex_state = 51},
  [8115] = {.lex_state = 444, .external_lex_state = 51},
  [8116] = {.lex_state = 444, .external_lex_state = 50},
  [8117] = {.lex_state = 444, .external_lex_state = 46},
  [8118] = {.lex_state = 444, .external_lex_state = 50},
  [8119] = {.lex_state = 444, .external_lex_state = 46},
  [8120] = {.lex_state = 444, .external_lex_state = 46},
  [8121] = {.lex_state = 447, .external_lex_state = 23},
  [8122] = {.lex_state = 444, .external_lex_state = 51},
  [8123] = {.lex_state = 444, .external_lex_state = 53},
  [8124] = {.lex_state = 444, .external_lex_state = 51},
  [8125] = {.lex_state = 444, .external_lex_state = 50},
  [8126] = {.lex_state = 444, .external_lex_state = 53},
  [8127] = {.lex_state = 444, .external_lex_state = 46},
  [8128] = {.lex_state = 444, .external_lex_state = 46},
  [8129] = {.lex_state = 444, .external_lex_state = 51},
  [8130] = {.lex_state = 444, .external_lex_state = 46},
  [8131] = {.lex_state = 443, .external_lex_state = 20},
  [8132] = {.lex_state = 447, .external_lex_state = 2},
  [8133] = {.lex_state = 444, .external_lex_state = 51},
  [8134] = {.lex_state = 454, .external_lex_state = 54},
  [8135] = {.lex_state = 447, .external_lex_state = 2},
  [8136] = {.lex_state = 444, .external_lex_state = 51},
  [8137] = {.lex_state = 443, .external_lex_state = 16},
  [8138] = {.lex_state = 444, .external_lex_state = 51},
  [8139] = {.lex_state = 444, .external_lex_state = 50},
  [8140] = {.lex_state = 444, .external_lex_state = 51},
  [8141] = {.lex_state = 444, .external_lex_state = 50},
  [8142] = {.lex_state = 444, .external_lex_state = 50},
  [8143] = {.lex_state = 444, .external_lex_state = 50},
  [8144] = {.lex_state = 443, .external_lex_state = 20},
  [8145] = {.lex_state = 444, .external_lex_state = 53},
  [8146] = {.lex_state = 444, .external_lex_state = 46},
  [8147] = {.lex_state = 444, .external_lex_state = 50},
  [8148] = {.lex_state = 444, .external_lex_state = 46},
  [8149] = {.lex_state = 444, .external_lex_state = 46},
  [8150] = {.lex_state = 444, .external_lex_state = 50},
  [8151] = {.lex_state = 444, .external_lex_state = 50},
  [8152] = {.lex_state = 444, .external_lex_state = 50},
  [8153] = {.lex_state = 444, .external_lex_state = 50},
  [8154] = {.lex_state = 454, .external_lex_state = 51},
  [8155] = {.lex_state = 444, .external_lex_state = 46},
  [8156] = {.lex_state = 444, .external_lex_state = 50},
  [8157] = {.lex_state = 444, .external_lex_state = 50},
  [8158] = {.lex_state = 444, .external_lex_state = 50},
  [8159] = {.lex_state = 444, .external_lex_state = 46},
  [8160] = {.lex_state = 444, .external_lex_state = 50},
  [8161] = {.lex_state = 444, .external_lex_state = 50},
  [8162] = {.lex_state = 444, .external_lex_state = 50},
  [8163] = {.lex_state = 196, .external_lex_state = 55},
  [8164] = {.lex_state = 444, .external_lex_state = 50},
  [8165] = {.lex_state = 444, .external_lex_state = 50},
  [8166] = {.lex_state = 444, .external_lex_state = 46},
  [8167] = {.lex_state = 444, .external_lex_state = 50},
  [8168] = {.lex_state = 444, .external_lex_state = 50},
  [8169] = {.lex_state = 444, .external_lex_state = 50},
  [8170] = {.lex_state = 444, .external_lex_state = 50},
  [8171] = {.lex_state = 444, .external_lex_state = 50},
  [8172] = {.lex_state = 452, .external_lex_state = 16},
  [8173] = {.lex_state = 443, .external_lex_state = 21},
  [8174] = {.lex_state = 444, .external_lex_state = 50},
  [8175] = {.lex_state = 444, .external_lex_state = 50},
  [8176] = {.lex_state = 444, .external_lex_state = 50},
  [8177] = {.lex_state = 444, .external_lex_state = 50},
  [8178] = {.lex_state = 444, .external_lex_state = 46},
  [8179] = {.lex_state = 444, .external_lex_state = 46},
  [8180] = {.lex_state = 454, .external_lex_state = 53},
  [8181] = {.lex_state = 444, .external_lex_state = 50},
  [8182] = {.lex_state = 454, .external_lex_state = 53},
  [8183] = {.lex_state = 447, .external_lex_state = 2},
  [8184] = {.lex_state = 452, .external_lex_state = 16},
  [8185] = {.lex_state = 452, .external_lex_state = 16},
  [8186] = {.lex_state = 444, .external_lex_state = 50},
  [8187] = {.lex_state = 443, .external_lex_state = 21},
  [8188] = {.lex_state = 452, .external_lex_state = 16},
  [8189] = {.lex_state = 444, .external_lex_state = 51},
  [8190] = {.lex_state = 444, .external_lex_state = 51},
  [8191] = {.lex_state = 444, .external_lex_state = 53},
  [8192] = {.lex_state = 444, .external_lex_state = 53},
  [8193] = {.lex_state = 444, .external_lex_state = 54},
  [8194] = {.lex_state = 444, .external_lex_state = 54},
  [8195] = {.lex_state = 443, .external_lex_state = 16},
  [8196] = {.lex_state = 444, .external_lex_state = 53},
  [8197] = {.lex_state = 444, .external_lex_state = 53},
  [8198] = {.lex_state = 444, .external_lex_state = 53},
  [8199] = {.lex_state = 444, .external_lex_state = 46},
  [8200] = {.lex_state = 444, .external_lex_state = 53},
  [8201] = {.lex_state = 444, .external_lex_state = 46},
  [8202] = {.lex_state = 444, .external_lex_state = 53},
  [8203] = {.lex_state = 444, .external_lex_state = 53},
  [8204] = {.lex_state = 444, .external_lex_state = 53},
  [8205] = {.lex_state = 444, .external_lex_state = 53},
  [8206] = {.lex_state = 444, .external_lex_state = 53},
  [8207] = {.lex_state = 444, .external_lex_state = 51},
  [8208] = {.lex_state = 444, .external_lex_state = 53},
  [8209] = {.lex_state = 444, .external_lex_state = 53},
  [8210] = {.lex_state = 444, .external_lex_state = 53},
  [8211] = {.lex_state = 444, .external_lex_state = 53},
  [8212] = {.lex_state = 444, .external_lex_state = 53},
  [8213] = {.lex_state = 444, .external_lex_state = 53},
  [8214] = {.lex_state = 444, .external_lex_state = 53},
  [8215] = {.lex_state = 444, .external_lex_state = 53},
  [8216] = {.lex_state = 444, .external_lex_state = 53},
  [8217] = {.lex_state = 444, .external_lex_state = 53},
  [8218] = {.lex_state = 444, .external_lex_state = 53},
  [8219] = {.lex_state = 447, .external_lex_state = 23},
  [8220] = {.lex_state = 444, .external_lex_state = 51},
  [8221] = {.lex_state = 444, .external_lex_state = 53},
  [8222] = {.lex_state = 444, .external_lex_state = 53},
  [8223] = {.lex_state = 444, .external_lex_state = 53},
  [8224] = {.lex_state = 444, .external_lex_state = 53},
  [8225] = {.lex_state = 444, .external_lex_state = 54},
  [8226] = {.lex_state = 452, .external_lex_state = 16},
  [8227] = {.lex_state = 444, .external_lex_state = 54},
  [8228] = {.lex_state = 444, .external_lex_state = 53},
  [8229] = {.lex_state = 444, .external_lex_state = 53},
  [8230] = {.lex_state = 447, .external_lex_state = 2},
  [8231] = {.lex_state = 447, .external_lex_state = 23},
  [8232] = {.lex_state = 444, .external_lex_state = 53},
  [8233] = {.lex_state = 444, .external_lex_state = 46},
  [8234] = {.lex_state = 444, .external_lex_state = 46},
  [8235] = {.lex_state = 444, .external_lex_state = 51},
  [8236] = {.lex_state = 444, .external_lex_state = 53},
  [8237] = {.lex_state = 444, .external_lex_state = 51},
  [8238] = {.lex_state = 444, .external_lex_state = 46},
  [8239] = {.lex_state = 444, .external_lex_state = 46},
  [8240] = {.lex_state = 444, .external_lex_state = 46},
  [8241] = {.lex_state = 454, .external_lex_state = 50},
  [8242] = {.lex_state = 454, .external_lex_state = 53},
  [8243] = {.lex_state = 454, .external_lex_state = 50},
  [8244] = {.lex_state = 447, .external_lex_state = 23},
  [8245] = {.lex_state = 454, .external_lex_state = 53},
  [8246] = {.lex_state = 444, .external_lex_state = 53},
  [8247] = {.lex_state = 444, .external_lex_state = 53},
  [8248] = {.lex_state = 454, .external_lex_state = 50},
  [8249] = {.lex_state = 452, .external_lex_state = 16},
  [8250] = {.lex_state = 454, .external_lex_state = 54},
  [8251] = {.lex_state = 454, .external_lex_state = 50},
  [8252] = {.lex_state = 444, .external_lex_state = 46},
  [8253] = {.lex_state = 443, .external_lex_state = 2},
  [8254] = {.lex_state = 444, .external_lex_state = 53},
  [8255] = {.lex_state = 444, .external_lex_state = 53},
  [8256] = {.lex_state = 443, .external_lex_state = 16},
  [8257] = {.lex_state = 447, .external_lex_state = 20},
  [8258] = {.lex_state = 454, .external_lex_state = 50},
  [8259] = {.lex_state = 444, .external_lex_state = 51},
  [8260] = {.lex_state = 443, .external_lex_state = 16},
  [8261] = {.lex_state = 444, .external_lex_state = 51},
  [8262] = {.lex_state = 444, .external_lex_state = 46},
  [8263] = {.lex_state = 444, .external_lex_state = 46},
  [8264] = {.lex_state = 454, .external_lex_state = 51},
  [8265] = {.lex_state = 454, .external_lex_state = 51},
  [8266] = {.lex_state = 454, .external_lex_state = 51},
  [8267] = {.lex_state = 454, .external_lex_state = 51},
  [8268] = {.lex_state = 444, .external_lex_state = 50},
  [8269] = {.lex_state = 444, .external_lex_state = 51},
  [8270] = {.lex_state = 444, .external_lex_state = 53},
  [8271] = {.lex_state = 444, .external_lex_state = 51},
  [8272] = {.lex_state = 443, .external_lex_state = 2},
  [8273] = {.lex_state = 444, .external_lex_state = 51},
  [8274] = {.lex_state = 443, .external_lex_state = 2},
  [8275] = {.lex_state = 454, .external_lex_state = 53},
  [8276] = {.lex_state = 443, .external_lex_state = 21},
  [8277] = {.lex_state = 454, .external_lex_state = 51},
  [8278] = {.lex_state = 444, .external_lex_state = 50},
  [8279] = {.lex_state = 444, .external_lex_state = 50},
  [8280] = {.lex_state = 444, .external_lex_state = 51},
  [8281] = {.lex_state = 444, .external_lex_state = 46},
  [8282] = {.lex_state = 444, .external_lex_state = 51},
  [8283] = {.lex_state = 444, .external_lex_state = 51},
  [8284] = {.lex_state = 444, .external_lex_state = 51},
  [8285] = {.lex_state = 444, .external_lex_state = 51},
  [8286] = {.lex_state = 444, .external_lex_state = 51},
  [8287] = {.lex_state = 447, .external_lex_state = 2},
  [8288] = {.lex_state = 444, .external_lex_state = 46},
  [8289] = {.lex_state = 444, .external_lex_state = 51},
  [8290] = {.lex_state = 444, .external_lex_state = 51},
  [8291] = {.lex_state = 444, .external_lex_state = 46},
  [8292] = {.lex_state = 454, .external_lex_state = 50},
  [8293] = {.lex_state = 447, .external_lex_state = 23},
  [8294] = {.lex_state = 444, .external_lex_state = 51},
  [8295] = {.lex_state = 444, .external_lex_state = 53},
  [8296] = {.lex_state = 444, .external_lex_state = 54},
  [8297] = {.lex_state = 443, .external_lex_state = 16},
  [8298] = {.lex_state = 444, .external_lex_state = 51},
  [8299] = {.lex_state = 443, .external_lex_state = 16},
  [8300] = {.lex_state = 443, .external_lex_state = 16},
  [8301] = {.lex_state = 444, .external_lex_state = 53},
  [8302] = {.lex_state = 452, .external_lex_state = 20},
  [8303] = {.lex_state = 443, .external_lex_state = 16},
  [8304] = {.lex_state = 444, .external_lex_state = 50},
  [8305] = {.lex_state = 452, .external_lex_state = 21},
  [8306] = {.lex_state = 443, .external_lex_state = 16},
  [8307] = {.lex_state = 443, .external_lex_state = 16},
  [8308] = {.lex_state = 444, .external_lex_state = 54},
  [8309] = {.lex_state = 443, .external_lex_state = 16},
  [8310] = {.lex_state = 443, .external_lex_state = 2},
  [8311] = {.lex_state = 444, .external_lex_state = 53},
  [8312] = {.lex_state = 443, .external_lex_state = 16},
  [8313] = {.lex_state = 445, .external_lex_state = 24},
  [8314] = {.lex_state = 444, .external_lex_state = 50},
  [8315] = {.lex_state = 443, .external_lex_state = 16},
  [8316] = {.lex_state = 444, .external_lex_state = 51},
  [8317] = {.lex_state = 444, .external_lex_state = 51},
  [8318] = {.lex_state = 444, .external_lex_state = 51},
  [8319] = {.lex_state = 443, .external_lex_state = 16},
  [8320] = {.lex_state = 443, .external_lex_state = 2},
  [8321] = {.lex_state = 443, .external_lex_state = 20},
  [8322] = {.lex_state = 443, .external_lex_state = 20},
  [8323] = {.lex_state = 444, .external_lex_state = 50},
  [8324] = {.lex_state = 444, .external_lex_state = 51},
  [8325] = {.lex_state = 443, .external_lex_state = 16},
  [8326] = {.lex_state = 452, .external_lex_state = 21},
  [8327] = {.lex_state = 444, .external_lex_state = 51},
  [8328] = {.lex_state = 443, .external_lex_state = 16},
  [8329] = {.lex_state = 443, .external_lex_state = 16},
  [8330] = {.lex_state = 443, .external_lex_state = 16},
  [8331] = {.lex_state = 444, .external_lex_state = 54},
  [8332] = {.lex_state = 443, .external_lex_state = 16},
  [8333] = {.lex_state = 443, .external_lex_state = 21},
  [8334] = {.lex_state = 443, .external_lex_state = 2},
  [8335] = {.lex_state = 444, .external_lex_state = 50},
  [8336] = {.lex_state = 444, .external_lex_state = 54},
  [8337] = {.lex_state = 443, .external_lex_state = 16},
  [8338] = {.lex_state = 443, .external_lex_state = 16},
  [8339] = {.lex_state = 444, .external_lex_state = 51},
  [8340] = {.lex_state = 443, .external_lex_state = 16},
  [8341] = {.lex_state = 444, .external_lex_state = 50},
  [8342] = {.lex_state = 452, .external_lex_state = 20},
  [8343] = {.lex_state = 444, .external_lex_state = 53},
  [8344] = {.lex_state = 443, .external_lex_state = 21},
  [8345] = {.lex_state = 444, .external_lex_state = 54},
  [8346] = {.lex_state = 444, .external_lex_state = 50},
  [8347] = {.lex_state = 444, .external_lex_state = 54},
  [8348] = {.lex_state = 443, .external_lex_state = 16},
  [8349] = {.lex_state = 444, .external_lex_state = 54},
  [8350] = {.lex_state = 444, .external_lex_state = 51},
  [8351] = {.lex_state = 443, .external_lex_state = 16},
  [8352] = {.lex_state = 444, .external_lex_state = 50},
  [8353] = {.lex_state = 444, .external_lex_state = 53},
  [8354] = {.lex_state = 443, .external_lex_state = 16},
  [8355] = {.lex_state = 443, .external_lex_state = 16},
  [8356] = {.lex_state = 444, .external_lex_state = 53},
  [8357] = {.lex_state = 444, .external_lex_state = 53},
  [8358] = {.lex_state = 444, .external_lex_state = 53},
  [8359] = {.lex_state = 443, .external_lex_state = 2},
  [8360] = {.lex_state = 444, .external_lex_state = 54},
  [8361] = {.lex_state = 443, .external_lex_state = 16},
  [8362] = {.lex_state = 444, .external_lex_state = 53},
  [8363] = {.lex_state = 444, .external_lex_state = 50},
  [8364] = {.lex_state = 443, .external_lex_state = 16},
  [8365] = {.lex_state = 444, .external_lex_state = 51},
  [8366] = {.lex_state = 444, .external_lex_state = 53},
  [8367] = {.lex_state = 443, .external_lex_state = 16},
  [8368] = {.lex_state = 444, .external_lex_state = 50},
  [8369] = {.lex_state = 444, .external_lex_state = 51},
  [8370] = {.lex_state = 452, .external_lex_state = 21},
  [8371] = {.lex_state = 452, .external_lex_state = 21},
  [8372] = {.lex_state = 443, .external_lex_state = 16},
  [8373] = {.lex_state = 444, .external_lex_state = 53},
  [8374] = {.lex_state = 444, .external_lex_state = 51},
  [8375] = {.lex_state = 452, .external_lex_state = 20},
  [8376] = {.lex_state = 443, .external_lex_state = 21},
  [8377] = {.lex_state = 444, .external_lex_state = 54},
  [8378] = {.lex_state = 443, .external_lex_state = 16},
  [8379] = {.lex_state = 443, .external_lex_state = 16},
  [8380] = {.lex_state = 444, .external_lex_state = 54},
  [8381] = {.lex_state = 444, .external_lex_state = 51},
  [8382] = {.lex_state = 444, .external_lex_state = 54},
  [8383] = {.lex_state = 444, .external_lex_state = 54},
  [8384] = {.lex_state = 444, .external_lex_state = 53},
  [8385] = {.lex_state = 444, .external_lex_state = 54},
  [8386] = {.lex_state = 444, .external_lex_state = 50},
  [8387] = {.lex_state = 444, .external_lex_state = 54},
  [8388] = {.lex_state = 444, .external_lex_state = 53},
  [8389] = {.lex_state = 444, .external_lex_state = 51},
  [8390] = {.lex_state = 444, .external_lex_state = 50},
  [8391] = {.lex_state = 444, .external_lex_state = 51},
  [8392] = {.lex_state = 443, .external_lex_state = 16},
  [8393] = {.lex_state = 444, .external_lex_state = 53},
  [8394] = {.lex_state = 444, .external_lex_state = 50},
  [8395] = {.lex_state = 444, .external_lex_state = 51},
  [8396] = {.lex_state = 444, .external_lex_state = 50},
  [8397] = {.lex_state = 444, .external_lex_state = 54},
  [8398] = {.lex_state = 444, .external_lex_state = 50},
  [8399] = {.lex_state = 444, .external_lex_state = 51},
  [8400] = {.lex_state = 444, .external_lex_state = 54},
  [8401] = {.lex_state = 444, .external_lex_state = 51},
  [8402] = {.lex_state = 443, .external_lex_state = 16},
  [8403] = {.lex_state = 444, .external_lex_state = 51},
  [8404] = {.lex_state = 444, .external_lex_state = 54},
  [8405] = {.lex_state = 452, .external_lex_state = 2},
  [8406] = {.lex_state = 444, .external_lex_state = 50},
  [8407] = {.lex_state = 444, .external_lex_state = 54},
  [8408] = {.lex_state = 443, .external_lex_state = 16},
  [8409] = {.lex_state = 443, .external_lex_state = 16},
  [8410] = {.lex_state = 444, .external_lex_state = 50},
  [8411] = {.lex_state = 444, .external_lex_state = 54},
  [8412] = {.lex_state = 445, .external_lex_state = 24},
  [8413] = {.lex_state = 444, .external_lex_state = 53},
  [8414] = {.lex_state = 454, .external_lex_state = 54},
  [8415] = {.lex_state = 452, .external_lex_state = 2},
  [8416] = {.lex_state = 444, .external_lex_state = 54},
  [8417] = {.lex_state = 444, .external_lex_state = 54},
  [8418] = {.lex_state = 444, .external_lex_state = 53},
  [8419] = {.lex_state = 444, .external_lex_state = 50},
  [8420] = {.lex_state = 444, .external_lex_state = 50},
  [8421] = {.lex_state = 444, .external_lex_state = 53},
  [8422] = {.lex_state = 444, .external_lex_state = 53},
  [8423] = {.lex_state = 444, .external_lex_state = 53},
  [8424] = {.lex_state = 444, .external_lex_state = 53},
  [8425] = {.lex_state = 444, .external_lex_state = 50},
  [8426] = {.lex_state = 444, .external_lex_state = 51},
  [8427] = {.lex_state = 452, .external_lex_state = 21},
  [8428] = {.lex_state = 443, .external_lex_state = 16},
  [8429] = {.lex_state = 452, .external_lex_state = 21},
  [8430] = {.lex_state = 443, .external_lex_state = 20},
  [8431] = {.lex_state = 444, .external_lex_state = 51},
  [8432] = {.lex_state = 444, .external_lex_state = 54},
  [8433] = {.lex_state = 443, .external_lex_state = 23},
  [8434] = {.lex_state = 444, .external_lex_state = 51},
  [8435] = {.lex_state = 452, .external_lex_state = 20},
  [8436] = {.lex_state = 444, .external_lex_state = 51},
  [8437] = {.lex_state = 443, .external_lex_state = 16},
  [8438] = {.lex_state = 452, .external_lex_state = 20},
  [8439] = {.lex_state = 444, .external_lex_state = 53},
  [8440] = {.lex_state = 444, .external_lex_state = 50},
  [8441] = {.lex_state = 444, .external_lex_state = 54},
  [8442] = {.lex_state = 443, .external_lex_state = 16},
  [8443] = {.lex_state = 452, .external_lex_state = 2},
  [8444] = {.lex_state = 444, .external_lex_state = 54},
  [8445] = {.lex_state = 454, .external_lex_state = 54},
  [8446] = {.lex_state = 443, .external_lex_state = 16},
  [8447] = {.lex_state = 454, .external_lex_state = 54},
  [8448] = {.lex_state = 443, .external_lex_state = 16},
  [8449] = {.lex_state = 445, .external_lex_state = 24},
  [8450] = {.lex_state = 454, .external_lex_state = 54},
  [8451] = {.lex_state = 444, .external_lex_state = 54},
  [8452] = {.lex_state = 443, .external_lex_state = 16},
  [8453] = {.lex_state = 454, .external_lex_state = 54},
  [8454] = {.lex_state = 444, .external_lex_state = 50},
  [8455] = {.lex_state = 444, .external_lex_state = 51},
  [8456] = {.lex_state = 443, .external_lex_state = 21},
  [8457] = {.lex_state = 443, .external_lex_state = 23},
  [8458] = {.lex_state = 443, .external_lex_state = 16},
  [8459] = {.lex_state = 444, .external_lex_state = 50},
  [8460] = {.lex_state = 444, .external_lex_state = 51},
  [8461] = {.lex_state = 443, .external_lex_state = 16},
  [8462] = {.lex_state = 454, .external_lex_state = 54},
  [8463] = {.lex_state = 444, .external_lex_state = 54},
  [8464] = {.lex_state = 444, .external_lex_state = 50},
  [8465] = {.lex_state = 444, .external_lex_state = 51},
  [8466] = {.lex_state = 444, .external_lex_state = 51},
  [8467] = {.lex_state = 444, .external_lex_state = 51},
  [8468] = {.lex_state = 444, .external_lex_state = 50},
  [8469] = {.lex_state = 443, .external_lex_state = 16},
  [8470] = {.lex_state = 444, .external_lex_state = 54},
  [8471] = {.lex_state = 444, .external_lex_state = 53},
  [8472] = {.lex_state = 444, .external_lex_state = 51},
  [8473] = {.lex_state = 444, .external_lex_state = 53},
  [8474] = {.lex_state = 444, .external_lex_state = 54},
  [8475] = {.lex_state = 445, .external_lex_state = 24},
  [8476] = {.lex_state = 452, .external_lex_state = 2},
  [8477] = {.lex_state = 444, .external_lex_state = 53},
  [8478] = {.lex_state = 447, .external_lex_state = 23},
  [8479] = {.lex_state = 444, .external_lex_state = 53},
  [8480] = {.lex_state = 444, .external_lex_state = 50},
  [8481] = {.lex_state = 443, .external_lex_state = 16},
  [8482] = {.lex_state = 452, .external_lex_state = 2},
  [8483] = {.lex_state = 444, .external_lex_state = 50},
  [8484] = {.lex_state = 444, .external_lex_state = 54},
  [8485] = {.lex_state = 444, .external_lex_state = 54},
  [8486] = {.lex_state = 443, .external_lex_state = 16},
  [8487] = {.lex_state = 444, .external_lex_state = 53},
  [8488] = {.lex_state = 445, .external_lex_state = 24},
  [8489] = {.lex_state = 444, .external_lex_state = 54},
  [8490] = {.lex_state = 443, .external_lex_state = 16},
  [8491] = {.lex_state = 452, .external_lex_state = 21},
  [8492] = {.lex_state = 444, .external_lex_state = 50},
  [8493] = {.lex_state = 443, .external_lex_state = 20},
  [8494] = {.lex_state = 443, .external_lex_state = 16},
  [8495] = {.lex_state = 445, .external_lex_state = 24},
  [8496] = {.lex_state = 444, .external_lex_state = 51},
  [8497] = {.lex_state = 444, .external_lex_state = 53},
  [8498] = {.lex_state = 444, .external_lex_state = 53},
  [8499] = {.lex_state = 444, .external_lex_state = 54},
  [8500] = {.lex_state = 443, .external_lex_state = 23},
  [8501] = {.lex_state = 444, .external_lex_state = 54},
  [8502] = {.lex_state = 443, .external_lex_state = 16},
  [8503] = {.lex_state = 452, .external_lex_state = 2},
  [8504] = {.lex_state = 444, .external_lex_state = 51},
  [8505] = {.lex_state = 445, .external_lex_state = 24},
  [8506] = {.lex_state = 445, .external_lex_state = 24},
  [8507] = {.lex_state = 444, .external_lex_state = 53},
  [8508] = {.lex_state = 444, .external_lex_state = 53},
  [8509] = {.lex_state = 443, .external_lex_state = 16},
  [8510] = {.lex_state = 444, .external_lex_state = 53},
  [8511] = {.lex_state = 444, .external_lex_state = 50},
  [8512] = {.lex_state = 452, .external_lex_state = 2},
  [8513] = {.lex_state = 444, .external_lex_state = 53},
  [8514] = {.lex_state = 444, .external_lex_state = 50},
  [8515] = {.lex_state = 444, .external_lex_state = 54},
  [8516] = {.lex_state = 452, .external_lex_state = 20},
  [8517] = {.lex_state = 444, .external_lex_state = 50},
  [8518] = {.lex_state = 444, .external_lex_state = 50},
  [8519] = {.lex_state = 444, .external_lex_state = 51},
  [8520] = {.lex_state = 452, .external_lex_state = 20},
  [8521] = {.lex_state = 443, .external_lex_state = 2},
  [8522] = {.lex_state = 444, .external_lex_state = 54},
  [8523] = {.lex_state = 443, .external_lex_state = 21},
  [8524] = {.lex_state = 443, .external_lex_state = 20},
  [8525] = {.lex_state = 443, .external_lex_state = 2},
  [8526] = {.lex_state = 443, .external_lex_state = 20},
  [8527] = {.lex_state = 443, .external_lex_state = 2},
  [8528] = {.lex_state = 443, .external_lex_state = 2},
  [8529] = {.lex_state = 444, .external_lex_state = 54},
  [8530] = {.lex_state = 443, .external_lex_state = 2},
  [8531] = {.lex_state = 452, .external_lex_state = 23},
  [8532] = {.lex_state = 443, .external_lex_state = 23},
  [8533] = {.lex_state = 443, .external_lex_state = 21},
  [8534] = {.lex_state = 444, .external_lex_state = 54},
  [8535] = {.lex_state = 443, .external_lex_state = 21},
  [8536] = {.lex_state = 443, .external_lex_state = 23},
  [8537] = {.lex_state = 444, .external_lex_state = 54},
  [8538] = {.lex_state = 444, .external_lex_state = 54},
  [8539] = {.lex_state = 443, .external_lex_state = 20},
  [8540] = {.lex_state = 443, .external_lex_state = 21},
  [8541] = {.lex_state = 443, .external_lex_state = 21},
  [8542] = {.lex_state = 443, .external_lex_state = 2},
  [8543] = {.lex_state = 444, .external_lex_state = 54},
  [8544] = {.lex_state = 444, .external_lex_state = 54},
  [8545] = {.lex_state = 444, .external_lex_state = 54},
  [8546] = {.lex_state = 444, .external_lex_state = 54},
  [8547] = {.lex_state = 444, .external_lex_state = 54},
  [8548] = {.lex_state = 444, .external_lex_state = 54},
  [8549] = {.lex_state = 444, .external_lex_state = 54},
  [8550] = {.lex_state = 443, .external_lex_state = 21},
  [8551] = {.lex_state = 443, .external_lex_state = 21},
  [8552] = {.lex_state = 444, .external_lex_state = 54},
  [8553] = {.lex_state = 444, .external_lex_state = 54},
  [8554] = {.lex_state = 444, .external_lex_state = 54},
  [8555] = {.lex_state = 443, .external_lex_state = 21},
  [8556] = {.lex_state = 443, .external_lex_state = 20},
  [8557] = {.lex_state = 444, .external_lex_state = 54},
  [8558] = {.lex_state = 443, .external_lex_state = 21},
  [8559] = {.lex_state = 443, .external_lex_state = 2},
  [8560] = {.lex_state = 444, .external_lex_state = 54},
  [8561] = {.lex_state = 444, .external_lex_state = 54},
  [8562] = {.lex_state = 444, .external_lex_state = 54},
  [8563] = {.lex_state = 444, .external_lex_state = 54},
  [8564] = {.lex_state = 443, .external_lex_state = 20},
  [8565] = {.lex_state = 452, .external_lex_state = 23},
  [8566] = {.lex_state = 444, .external_lex_state = 54},
  [8567] = {.lex_state = 444, .external_lex_state = 54},
  [8568] = {.lex_state = 443, .external_lex_state = 2},
  [8569] = {.lex_state = 444, .external_lex_state = 54},
  [8570] = {.lex_state = 444, .external_lex_state = 54},
  [8571] = {.lex_state = 444, .external_lex_state = 54},
  [8572] = {.lex_state = 452, .external_lex_state = 23},
  [8573] = {.lex_state = 444, .external_lex_state = 54},
  [8574] = {.lex_state = 443, .external_lex_state = 2},
  [8575] = {.lex_state = 443, .external_lex_state = 2},
  [8576] = {.lex_state = 443, .external_lex_state = 21},
  [8577] = {.lex_state = 443, .external_lex_state = 21},
  [8578] = {.lex_state = 443, .external_lex_state = 20},
  [8579] = {.lex_state = 443, .external_lex_state = 20},
  [8580] = {.lex_state = 452, .external_lex_state = 23},
  [8581] = {.lex_state = 445, .external_lex_state = 22},
  [8582] = {.lex_state = 452, .external_lex_state = 23},
  [8583] = {.lex_state = 443, .external_lex_state = 2},
  [8584] = {.lex_state = 443, .external_lex_state = 20},
  [8585] = {.lex_state = 443, .external_lex_state = 21},
  [8586] = {.lex_state = 443, .external_lex_state = 2},
  [8587] = {.lex_state = 443, .external_lex_state = 20},
  [8588] = {.lex_state = 445, .external_lex_state = 24},
  [8589] = {.lex_state = 443, .external_lex_state = 21},
  [8590] = {.lex_state = 443, .external_lex_state = 21},
  [8591] = {.lex_state = 443, .external_lex_state = 21},
  [8592] = {.lex_state = 443, .external_lex_state = 21},
  [8593] = {.lex_state = 443, .external_lex_state = 2},
  [8594] = {.lex_state = 445, .external_lex_state = 24},
  [8595] = {.lex_state = 443, .external_lex_state = 21},
  [8596] = {.lex_state = 443, .external_lex_state = 20},
  [8597] = {.lex_state = 445, .external_lex_state = 22},
  [8598] = {.lex_state = 443, .external_lex_state = 2},
  [8599] = {.lex_state = 443, .external_lex_state = 21},
  [8600] = {.lex_state = 443, .external_lex_state = 21},
  [8601] = {.lex_state = 443, .external_lex_state = 21},
  [8602] = {.lex_state = 445, .external_lex_state = 22},
  [8603] = {.lex_state = 443, .external_lex_state = 21},
  [8604] = {.lex_state = 445, .external_lex_state = 22},
  [8605] = {.lex_state = 445, .external_lex_state = 22},
  [8606] = {.lex_state = 443, .external_lex_state = 21},
  [8607] = {.lex_state = 443, .external_lex_state = 2},
  [8608] = {.lex_state = 445, .external_lex_state = 24},
  [8609] = {.lex_state = 443, .external_lex_state = 2},
  [8610] = {.lex_state = 445, .external_lex_state = 22},
  [8611] = {.lex_state = 443, .external_lex_state = 21},
  [8612] = {.lex_state = 443, .external_lex_state = 21},
  [8613] = {.lex_state = 443, .external_lex_state = 21},
  [8614] = {.lex_state = 443, .external_lex_state = 21},
  [8615] = {.lex_state = 443, .external_lex_state = 20},
  [8616] = {.lex_state = 452, .external_lex_state = 23},
  [8617] = {.lex_state = 443, .external_lex_state = 21},
  [8618] = {.lex_state = 443, .external_lex_state = 21},
  [8619] = {.lex_state = 443, .external_lex_state = 2},
  [8620] = {.lex_state = 445, .external_lex_state = 22},
  [8621] = {.lex_state = 443, .external_lex_state = 2},
  [8622] = {.lex_state = 443, .external_lex_state = 21},
  [8623] = {.lex_state = 443, .external_lex_state = 21},
  [8624] = {.lex_state = 443, .external_lex_state = 21},
  [8625] = {.lex_state = 443, .external_lex_state = 2},
  [8626] = {.lex_state = 443, .external_lex_state = 2},
  [8627] = {.lex_state = 443, .external_lex_state = 21},
  [8628] = {.lex_state = 443, .external_lex_state = 2},
  [8629] = {.lex_state = 443, .external_lex_state = 20},
  [8630] = {.lex_state = 443, .external_lex_state = 20},
  [8631] = {.lex_state = 443, .external_lex_state = 21},
  [8632] = {.lex_state = 443, .external_lex_state = 20},
  [8633] = {.lex_state = 443, .external_lex_state = 20},
  [8634] = {.lex_state = 444, .external_lex_state = 54},
  [8635] = {.lex_state = 443, .external_lex_state = 2},
  [8636] = {.lex_state = 443, .external_lex_state = 20},
  [8637] = {.lex_state = 443, .external_lex_state = 20},
  [8638] = {.lex_state = 443, .external_lex_state = 20},
  [8639] = {.lex_state = 443, .external_lex_state = 20},
  [8640] = {.lex_state = 443, .external_lex_state = 20},
  [8641] = {.lex_state = 443, .external_lex_state = 20},
  [8642] = {.lex_state = 443, .external_lex_state = 20},
  [8643] = {.lex_state = 443, .external_lex_state = 20},
  [8644] = {.lex_state = 443, .external_lex_state = 21},
  [8645] = {.lex_state = 443, .external_lex_state = 21},
  [8646] = {.lex_state = 445, .external_lex_state = 24},
  [8647] = {.lex_state = 443, .external_lex_state = 2},
  [8648] = {.lex_state = 443, .external_lex_state = 2},
  [8649] = {.lex_state = 443, .external_lex_state = 20},
  [8650] = {.lex_state = 443, .external_lex_state = 20},
  [8651] = {.lex_state = 443, .external_lex_state = 20},
  [8652] = {.lex_state = 443, .external_lex_state = 20},
  [8653] = {.lex_state = 443, .external_lex_state = 20},
  [8654] = {.lex_state = 443, .external_lex_state = 20},
  [8655] = {.lex_state = 443, .external_lex_state = 23},
  [8656] = {.lex_state = 443, .external_lex_state = 20},
  [8657] = {.lex_state = 443, .external_lex_state = 2},
  [8658] = {.lex_state = 443, .external_lex_state = 20},
  [8659] = {.lex_state = 443, .external_lex_state = 20},
  [8660] = {.lex_state = 443, .external_lex_state = 20},
  [8661] = {.lex_state = 443, .external_lex_state = 20},
  [8662] = {.lex_state = 443, .external_lex_state = 20},
  [8663] = {.lex_state = 443, .external_lex_state = 20},
  [8664] = {.lex_state = 444, .external_lex_state = 54},
  [8665] = {.lex_state = 443, .external_lex_state = 21},
  [8666] = {.lex_state = 443, .external_lex_state = 2},
  [8667] = {.lex_state = 443, .external_lex_state = 20},
  [8668] = {.lex_state = 443, .external_lex_state = 2},
  [8669] = {.lex_state = 444, .external_lex_state = 54},
  [8670] = {.lex_state = 444, .external_lex_state = 54},
  [8671] = {.lex_state = 443, .external_lex_state = 2},
  [8672] = {.lex_state = 443, .external_lex_state = 2},
  [8673] = {.lex_state = 443, .external_lex_state = 20},
  [8674] = {.lex_state = 443, .external_lex_state = 23},
  [8675] = {.lex_state = 452, .external_lex_state = 23},
  [8676] = {.lex_state = 443, .external_lex_state = 2},
  [8677] = {.lex_state = 443, .external_lex_state = 2},
  [8678] = {.lex_state = 443, .external_lex_state = 2},
  [8679] = {.lex_state = 443, .external_lex_state = 2},
  [8680] = {.lex_state = 443, .external_lex_state = 2},
  [8681] = {.lex_state = 443, .external_lex_state = 2},
  [8682] = {.lex_state = 443, .external_lex_state = 20},
  [8683] = {.lex_state = 443, .external_lex_state = 20},
  [8684] = {.lex_state = 443, .external_lex_state = 20},
  [8685] = {.lex_state = 443, .external_lex_state = 20},
  [8686] = {.lex_state = 443, .external_lex_state = 2},
  [8687] = {.lex_state = 443, .external_lex_state = 20},
  [8688] = {.lex_state = 443, .external_lex_state = 2},
  [8689] = {.lex_state = 443, .external_lex_state = 20},
  [8690] = {.lex_state = 443, .external_lex_state = 20},
  [8691] = {.lex_state = 443, .external_lex_state = 20},
  [8692] = {.lex_state = 443, .external_lex_state = 20},
  [8693] = {.lex_state = 445, .external_lex_state = 24},
  [8694] = {.lex_state = 444, .external_lex_state = 54},
  [8695] = {.lex_state = 443, .external_lex_state = 21},
  [8696] = {.lex_state = 443, .external_lex_state = 23},
  [8697] = {.lex_state = 443, .external_lex_state = 23},
  [8698] = {.lex_state = 165, .external_lex_state = 55},
  [8699] = {.lex_state = 445, .external_lex_state = 24},
  [8700] = {.lex_state = 165, .external_lex_state = 55},
  [8701] = {.lex_state = 443, .external_lex_state = 23},
  [8702] = {.lex_state = 443, .external_lex_state = 23},
  [8703] = {.lex_state = 443, .external_lex_state = 23},
  [8704] = {.lex_state = 443, .external_lex_state = 23},
  [8705] = {.lex_state = 443, .external_lex_state = 23},
  [8706] = {.lex_state = 443, .external_lex_state = 23},
  [8707] = {.lex_state = 445, .external_lex_state = 22},
  [8708] = {.lex_state = 443, .external_lex_state = 23},
  [8709] = {.lex_state = 443, .external_lex_state = 23},
  [8710] = {.lex_state = 445, .external_lex_state = 22},
  [8711] = {.lex_state = 445, .external_lex_state = 24},
  [8712] = {.lex_state = 443, .external_lex_state = 23},
  [8713] = {.lex_state = 443, .external_lex_state = 23},
  [8714] = {.lex_state = 443, .external_lex_state = 23},
  [8715] = {.lex_state = 443, .external_lex_state = 23},
  [8716] = {.lex_state = 445, .external_lex_state = 24},
  [8717] = {.lex_state = 445, .external_lex_state = 22},
  [8718] = {.lex_state = 443, .external_lex_state = 23},
  [8719] = {.lex_state = 443, .external_lex_state = 23},
  [8720] = {.lex_state = 443, .external_lex_state = 23},
  [8721] = {.lex_state = 443, .external_lex_state = 23},
  [8722] = {.lex_state = 445, .external_lex_state = 22},
  [8723] = {.lex_state = 445, .external_lex_state = 24},
  [8724] = {.lex_state = 443, .external_lex_state = 23},
  [8725] = {.lex_state = 443, .external_lex_state = 23},
  [8726] = {.lex_state = 443, .external_lex_state = 23},
  [8727] = {.lex_state = 443, .external_lex_state = 23},
  [8728] = {.lex_state = 443, .external_lex_state = 23},
  [8729] = {.lex_state = 443, .external_lex_state = 23},
  [8730] = {.lex_state = 443, .external_lex_state = 23},
  [8731] = {.lex_state = 443, .external_lex_state = 23},
  [8732] = {.lex_state = 443, .external_lex_state = 23},
  [8733] = {.lex_state = 443, .external_lex_state = 2},
  [8734] = {.lex_state = 443, .external_lex_state = 23},
  [8735] = {.lex_state = 443, .external_lex_state = 23},
  [8736] = {.lex_state = 443, .external_lex_state = 23},
  [8737] = {.lex_state = 445, .external_lex_state = 22},
  [8738] = {.lex_state = 443, .external_lex_state = 23},
  [8739] = {.lex_state = 445, .external_lex_state = 24},
  [8740] = {.lex_state = 443, .external_lex_state = 2},
  [8741] = {.lex_state = 443, .external_lex_state = 23},
  [8742] = {.lex_state = 443, .external_lex_state = 23},
  [8743] = {.lex_state = 443, .external_lex_state = 23},
  [8744] = {.lex_state = 443, .external_lex_state = 23},
  [8745] = {.lex_state = 443, .external_lex_state = 23},
  [8746] = {.lex_state = 443, .external_lex_state = 23},
  [8747] = {.lex_state = 445, .external_lex_state = 24},
  [8748] = {.lex_state = 445, .external_lex_state = 24},
  [8749] = {.lex_state = 445, .external_lex_state = 24},
  [8750] = {.lex_state = 445, .external_lex_state = 24},
  [8751] = {.lex_state = 445, .external_lex_state = 24},
  [8752] = {.lex_state = 445, .external_lex_state = 24},
  [8753] = {.lex_state = 445, .external_lex_state = 22},
  [8754] = {.lex_state = 445, .external_lex_state = 24},
  [8755] = {.lex_state = 445, .external_lex_state = 22},
  [8756] = {.lex_state = 445, .external_lex_state = 22},
  [8757] = {.lex_state = 445, .external_lex_state = 24},
  [8758] = {.lex_state = 445, .external_lex_state = 24},
  [8759] = {.lex_state = 445, .external_lex_state = 24},
  [8760] = {.lex_state = 445, .external_lex_state = 24},
  [8761] = {.lex_state = 445, .external_lex_state = 22},
  [8762] = {.lex_state = 445, .external_lex_state = 24},
  [8763] = {.lex_state = 445, .external_lex_state = 22},
  [8764] = {.lex_state = 445, .external_lex_state = 24},
  [8765] = {.lex_state = 445, .external_lex_state = 24},
  [8766] = {.lex_state = 454, .external_lex_state = 67},
  [8767] = {.lex_state = 454, .external_lex_state = 67},
  [8768] = {.lex_state = 445, .external_lex_state = 22},
  [8769] = {.lex_state = 445, .external_lex_state = 22},
  [8770] = {.lex_state = 148, .external_lex_state = 2},
  [8771] = {.lex_state = 445, .external_lex_state = 22},
  [8772] = {.lex_state = 445, .external_lex_state = 22},
  [8773] = {.lex_state = 454, .external_lex_state = 50},
  [8774] = {.lex_state = 454, .external_lex_state = 50},
  [8775] = {.lex_state = 445, .external_lex_state = 22},
  [8776] = {.lex_state = 454, .external_lex_state = 50},
  [8777] = {.lex_state = 445, .external_lex_state = 22},
  [8778] = {.lex_state = 454, .external_lex_state = 67},
  [8779] = {.lex_state = 445, .external_lex_state = 22},
  [8780] = {.lex_state = 445, .external_lex_state = 22},
  [8781] = {.lex_state = 443, .external_lex_state = 2},
  [8782] = {.lex_state = 445, .external_lex_state = 22},
  [8783] = {.lex_state = 445, .external_lex_state = 22},
  [8784] = {.lex_state = 445, .external_lex_state = 22},
  [8785] = {.lex_state = 445, .external_lex_state = 22},
  [8786] = {.lex_state = 443, .external_lex_state = 2},
  [8787] = {.lex_state = 443, .external_lex_state = 2},
  [8788] = {.lex_state = 445, .external_lex_state = 22},
  [8789] = {.lex_state = 445, .external_lex_state = 22},
  [8790] = {.lex_state = 454, .external_lex_state = 67},
  [8791] = {.lex_state = 149, .external_lex_state = 2},
  [8792] = {.lex_state = 454, .external_lex_state = 54},
  [8793] = {.lex_state = 454, .external_lex_state = 67},
  [8794] = {.lex_state = 149, .external_lex_state = 2},
  [8795] = {.lex_state = 454, .external_lex_state = 50},
  [8796] = {.lex_state = 454, .external_lex_state = 50},
  [8797] = {.lex_state = 454, .external_lex_state = 67},
  [8798] = {.lex_state = 454, .external_lex_state = 67},
  [8799] = {.lex_state = 454, .external_lex_state = 54},
  [8800] = {.lex_state = 454, .external_lex_state = 54},
  [8801] = {.lex_state = 454, .external_lex_state = 50},
  [8802] = {.lex_state = 149, .external_lex_state = 2},
  [8803] = {.lex_state = 454, .external_lex_state = 67},
  [8804] = {.lex_state = 454, .external_lex_state = 50},
  [8805] = {.lex_state = 454, .external_lex_state = 67},
  [8806] = {.lex_state = 149, .external_lex_state = 2},
  [8807] = {.lex_state = 454, .external_lex_state = 50},
  [8808] = {.lex_state = 454, .external_lex_state = 50},
  [8809] = {.lex_state = 454, .external_lex_state = 54},
  [8810] = {.lex_state = 452, .external_lex_state = 25},
  [8811] = {.lex_state = 454, .external_lex_state = 54},
  [8812] = {.lex_state = 452, .external_lex_state = 21},
  [8813] = {.lex_state = 452, .external_lex_state = 21},
  [8814] = {.lex_state = 452, .external_lex_state = 21},
  [8815] = {.lex_state = 454, .external_lex_state = 54},
  [8816] = {.lex_state = 454, .external_lex_state = 54},
  [8817] = {.lex_state = 452, .external_lex_state = 2},
  [8818] = {.lex_state = 452, .external_lex_state = 21},
  [8819] = {.lex_state = 452, .external_lex_state = 25},
  [8820] = {.lex_state = 452, .external_lex_state = 21},
  [8821] = {.lex_state = 454, .external_lex_state = 53},
  [8822] = {.lex_state = 452, .external_lex_state = 25},
  [8823] = {.lex_state = 452, .external_lex_state = 25},
  [8824] = {.lex_state = 452, .external_lex_state = 21},
  [8825] = {.lex_state = 452, .external_lex_state = 25},
  [8826] = {.lex_state = 454, .external_lex_state = 53},
  [8827] = {.lex_state = 454, .external_lex_state = 54},
  [8828] = {.lex_state = 452, .external_lex_state = 25},
  [8829] = {.lex_state = 454, .external_lex_state = 54},
  [8830] = {.lex_state = 452, .external_lex_state = 25},
  [8831] = {.lex_state = 452, .external_lex_state = 21},
  [8832] = {.lex_state = 454, .external_lex_state = 53},
  [8833] = {.lex_state = 443, .external_lex_state = 2},
  [8834] = {.lex_state = 443, .external_lex_state = 2},
  [8835] = {.lex_state = 454, .external_lex_state = 53},
  [8836] = {.lex_state = 443, .external_lex_state = 2},
  [8837] = {.lex_state = 443, .external_lex_state = 2},
  [8838] = {.lex_state = 452, .external_lex_state = 23},
  [8839] = {.lex_state = 443, .external_lex_state = 2},
  [8840] = {.lex_state = 443, .external_lex_state = 2},
  [8841] = {.lex_state = 454, .external_lex_state = 53},
  [8842] = {.lex_state = 443, .external_lex_state = 2},
  [8843] = {.lex_state = 443, .external_lex_state = 2},
  [8844] = {.lex_state = 454, .external_lex_state = 53},
  [8845] = {.lex_state = 452, .external_lex_state = 23},
  [8846] = {.lex_state = 443, .external_lex_state = 2},
  [8847] = {.lex_state = 443, .external_lex_state = 2},
  [8848] = {.lex_state = 443, .external_lex_state = 2},
  [8849] = {.lex_state = 443, .external_lex_state = 2},
  [8850] = {.lex_state = 443, .external_lex_state = 2},
  [8851] = {.lex_state = 452, .external_lex_state = 23},
  [8852] = {.lex_state = 454, .external_lex_state = 53},
  [8853] = {.lex_state = 443, .external_lex_state = 2},
  [8854] = {.lex_state = 443, .external_lex_state = 2},
  [8855] = {.lex_state = 443, .external_lex_state = 2},
  [8856] = {.lex_state = 452, .external_lex_state = 23},
  [8857] = {.lex_state = 454, .external_lex_state = 53},
  [8858] = {.lex_state = 443, .external_lex_state = 2},
  [8859] = {.lex_state = 443, .external_lex_state = 2},
  [8860] = {.lex_state = 443, .external_lex_state = 2},
  [8861] = {.lex_state = 443, .external_lex_state = 2},
  [8862] = {.lex_state = 443, .external_lex_state = 2},
  [8863] = {.lex_state = 443, .external_lex_state = 2},
  [8864] = {.lex_state = 443, .external_lex_state = 2},
  [8865] = {.lex_state = 443, .external_lex_state = 2},
  [8866] = {.lex_state = 452, .external_lex_state = 23},
  [8867] = {.lex_state = 443, .external_lex_state = 2},
  [8868] = {.lex_state = 452, .external_lex_state = 23},
  [8869] = {.lex_state = 443, .external_lex_state = 2},
  [8870] = {.lex_state = 443, .external_lex_state = 2},
  [8871] = {.lex_state = 452, .external_lex_state = 23},
  [8872] = {.lex_state = 443, .external_lex_state = 2},
  [8873] = {.lex_state = 443, .external_lex_state = 2},
  [8874] = {.lex_state = 443, .external_lex_state = 2},
  [8875] = {.lex_state = 454, .external_lex_state = 53},
  [8876] = {.lex_state = 443, .external_lex_state = 2},
  [8877] = {.lex_state = 443, .external_lex_state = 2},
  [8878] = {.lex_state = 443, .external_lex_state = 2},
  [8879] = {.lex_state = 452, .external_lex_state = 2},
  [8880] = {.lex_state = 452, .external_lex_state = 2},
  [8881] = {.lex_state = 452, .external_lex_state = 2},
  [8882] = {.lex_state = 452, .external_lex_state = 2},
  [8883] = {.lex_state = 443, .external_lex_state = 2},
  [8884] = {.lex_state = 452, .external_lex_state = 2},
  [8885] = {.lex_state = 452, .external_lex_state = 2},
  [8886] = {.lex_state = 452, .external_lex_state = 2},
  [8887] = {.lex_state = 452, .external_lex_state = 2},
  [8888] = {.lex_state = 452, .external_lex_state = 2},
  [8889] = {.lex_state = 443, .external_lex_state = 2},
  [8890] = {.lex_state = 443, .external_lex_state = 2},
  [8891] = {.lex_state = 443, .external_lex_state = 2},
  [8892] = {.lex_state = 443, .external_lex_state = 2},
  [8893] = {.lex_state = 443, .external_lex_state = 2},
  [8894] = {.lex_state = 443, .external_lex_state = 2},
  [8895] = {.lex_state = 457, .external_lex_state = 55},
  [8896] = {.lex_state = 455, .external_lex_state = 55},
  [8897] = {.lex_state = 457, .external_lex_state = 55},
  [8898] = {.lex_state = 455, .external_lex_state = 55},
  [8899] = {.lex_state = 455, .external_lex_state = 55},
  [8900] = {.lex_state = 455, .external_lex_state = 55},
  [8901] = {.lex_state = 455, .external_lex_state = 55},
  [8902] = {.lex_state = 455, .external_lex_state = 55},
  [8903] = {.lex_state = 455, .external_lex_state = 55},
  [8904] = {.lex_state = 455, .external_lex_state = 55},
  [8905] = {.lex_state = 455, .external_lex_state = 55},
  [8906] = {.lex_state = 455, .external_lex_state = 55},
  [8907] = {.lex_state = 455, .external_lex_state = 55},
  [8908] = {.lex_state = 455, .external_lex_state = 55},
  [8909] = {.lex_state = 455, .external_lex_state = 55},
  [8910] = {.lex_state = 455, .external_lex_state = 55},
  [8911] = {.lex_state = 455, .external_lex_state = 55},
  [8912] = {.lex_state = 455, .external_lex_state = 55},
  [8913] = {.lex_state = 455, .external_lex_state = 55},
  [8914] = {.lex_state = 455, .external_lex_state = 55},
  [8915] = {.lex_state = 455, .external_lex_state = 55},
  [8916] = {.lex_state = 455, .external_lex_state = 55},
  [8917] = {.lex_state = 455, .external_lex_state = 55},
  [8918] = {.lex_state = 455, .external_lex_state = 55},
  [8919] = {.lex_state = 455, .external_lex_state = 55},
  [8920] = {.lex_state = 455, .external_lex_state = 55},
  [8921] = {.lex_state = 455, .external_lex_state = 55},
  [8922] = {.lex_state = 455, .external_lex_state = 55},
  [8923] = {.lex_state = 455, .external_lex_state = 55},
  [8924] = {.lex_state = 455, .external_lex_state = 55},
  [8925] = {.lex_state = 455, .external_lex_state = 55},
  [8926] = {.lex_state = 455, .external_lex_state = 55},
  [8927] = {.lex_state = 455, .external_lex_state = 55},
  [8928] = {.lex_state = 455, .external_lex_state = 55},
  [8929] = {.lex_state = 455, .external_lex_state = 55},
  [8930] = {.lex_state = 455, .external_lex_state = 55},
  [8931] = {.lex_state = 455, .external_lex_state = 55},
  [8932] = {.lex_state = 166, .external_lex_state = 68},
  [8933] = {.lex_state = 155, .external_lex_state = 69},
  [8934] = {.lex_state = 155, .external_lex_state = 70},
  [8935] = {.lex_state = 155, .external_lex_state = 68},
  [8936] = {.lex_state = 155, .external_lex_state = 71},
  [8937] = {.lex_state = 167, .external_lex_state = 68},
  [8938] = {.lex_state = 173, .external_lex_state = 68},
  [8939] = {.lex_state = 156, .external_lex_state = 69},
  [8940] = {.lex_state = 156, .external_lex_state = 70},
  [8941] = {.lex_state = 156, .external_lex_state = 71},
  [8942] = {.lex_state = 156, .external_lex_state = 68},
  [8943] = {.lex_state = 174, .external_lex_state = 68},
  [8944] = {.lex_state = 151, .external_lex_state = 57},
  [8945] = {.lex_state = 151, .external_lex_state = 56},
  [8946] = {.lex_state = 152, .external_lex_state = 72},
  [8947] = {.lex_state = 152, .external_lex_state = 73},
  [8948] = {.lex_state = 152, .external_lex_state = 73},
  [8949] = {.lex_state = 152, .external_lex_state = 72},
  [8950] = {.lex_state = 152, .external_lex_state = 72},
  [8951] = {.lex_state = 152, .external_lex_state = 72},
  [8952] = {.lex_state = 152, .external_lex_state = 73},
  [8953] = {.lex_state = 151, .external_lex_state = 56},
  [8954] = {.lex_state = 152, .external_lex_state = 72},
  [8955] = {.lex_state = 151, .external_lex_state = 57},
  [8956] = {.lex_state = 152, .external_lex_state = 72},
  [8957] = {.lex_state = 152, .external_lex_state = 73},
  [8958] = {.lex_state = 152, .external_lex_state = 72},
  [8959] = {.lex_state = 151, .external_lex_state = 57},
  [8960] = {.lex_state = 151, .external_lex_state = 56},
  [8961] = {.lex_state = 151, .external_lex_state = 55},
  [8962] = {.lex_state = 152, .external_lex_state = 72},
  [8963] = {.lex_state = 151, .external_lex_state = 56},
  [8964] = {.lex_state = 152, .external_lex_state = 73},
  [8965] = {.lex_state = 152, .external_lex_state = 73},
  [8966] = {.lex_state = 151, .external_lex_state = 57},
  [8967] = {.lex_state = 151, .external_lex_state = 56},
  [8968] = {.lex_state = 151, .external_lex_state = 56},
  [8969] = {.lex_state = 151, .external_lex_state = 56},
  [8970] = {.lex_state = 151, .external_lex_state = 57},
  [8971] = {.lex_state = 151, .external_lex_state = 57},
  [8972] = {.lex_state = 152, .external_lex_state = 73},
  [8973] = {.lex_state = 152, .external_lex_state = 73},
  [8974] = {.lex_state = 152, .external_lex_state = 73},
  [8975] = {.lex_state = 151, .external_lex_state = 56},
  [8976] = {.lex_state = 151, .external_lex_state = 57},
  [8977] = {.lex_state = 151, .external_lex_state = 57},
  [8978] = {.lex_state = 151, .external_lex_state = 58},
  [8979] = {.lex_state = 152, .external_lex_state = 72},
  [8980] = {.lex_state = 151, .external_lex_state = 55},
  [8981] = {.lex_state = 152, .external_lex_state = 73},
  [8982] = {.lex_state = 151, .external_lex_state = 55},
  [8983] = {.lex_state = 151, .external_lex_state = 55},
  [8984] = {.lex_state = 161, .external_lex_state = 74},
  [8985] = {.lex_state = 152, .external_lex_state = 75},
  [8986] = {.lex_state = 153, .external_lex_state = 57},
  [8987] = {.lex_state = 152, .external_lex_state = 75},
  [8988] = {.lex_state = 161, .external_lex_state = 74},
  [8989] = {.lex_state = 153, .external_lex_state = 56},
  [8990] = {.lex_state = 161, .external_lex_state = 74},
  [8991] = {.lex_state = 152, .external_lex_state = 74},
  [8992] = {.lex_state = 151, .external_lex_state = 56},
  [8993] = {.lex_state = 161, .external_lex_state = 74},
  [8994] = {.lex_state = 151, .external_lex_state = 55},
  [8995] = {.lex_state = 152, .external_lex_state = 74},
  [8996] = {.lex_state = 152, .external_lex_state = 74},
  [8997] = {.lex_state = 151, .external_lex_state = 55},
  [8998] = {.lex_state = 161, .external_lex_state = 74},
  [8999] = {.lex_state = 151, .external_lex_state = 58},
  [9000] = {.lex_state = 151, .external_lex_state = 58},
  [9001] = {.lex_state = 151, .external_lex_state = 56},
  [9002] = {.lex_state = 153, .external_lex_state = 57},
  [9003] = {.lex_state = 153, .external_lex_state = 56},
  [9004] = {.lex_state = 153, .external_lex_state = 57},
  [9005] = {.lex_state = 153, .external_lex_state = 56},
  [9006] = {.lex_state = 151, .external_lex_state = 56},
  [9007] = {.lex_state = 152, .external_lex_state = 73},
  [9008] = {.lex_state = 153, .external_lex_state = 56},
  [9009] = {.lex_state = 153, .external_lex_state = 56},
  [9010] = {.lex_state = 153, .external_lex_state = 56},
  [9011] = {.lex_state = 152, .external_lex_state = 74},
  [9012] = {.lex_state = 152, .external_lex_state = 74},
  [9013] = {.lex_state = 152, .external_lex_state = 74},
  [9014] = {.lex_state = 162, .external_lex_state = 55},
  [9015] = {.lex_state = 153, .external_lex_state = 56},
  [9016] = {.lex_state = 153, .external_lex_state = 57},
  [9017] = {.lex_state = 153, .external_lex_state = 57},
  [9018] = {.lex_state = 153, .external_lex_state = 57},
  [9019] = {.lex_state = 152, .external_lex_state = 75},
  [9020] = {.lex_state = 151, .external_lex_state = 55},
  [9021] = {.lex_state = 161, .external_lex_state = 74},
  [9022] = {.lex_state = 152, .external_lex_state = 72},
  [9023] = {.lex_state = 152, .external_lex_state = 72},
  [9024] = {.lex_state = 152, .external_lex_state = 75},
  [9025] = {.lex_state = 153, .external_lex_state = 57},
  [9026] = {.lex_state = 152, .external_lex_state = 75},
  [9027] = {.lex_state = 152, .external_lex_state = 74},
  [9028] = {.lex_state = 153, .external_lex_state = 57},
  [9029] = {.lex_state = 152, .external_lex_state = 75},
  [9030] = {.lex_state = 151, .external_lex_state = 58},
  [9031] = {.lex_state = 151, .external_lex_state = 57},
  [9032] = {.lex_state = 151, .external_lex_state = 58},
  [9033] = {.lex_state = 151, .external_lex_state = 58},
  [9034] = {.lex_state = 152, .external_lex_state = 72},
  [9035] = {.lex_state = 151, .external_lex_state = 55},
  [9036] = {.lex_state = 151, .external_lex_state = 56},
  [9037] = {.lex_state = 161, .external_lex_state = 74},
  [9038] = {.lex_state = 152, .external_lex_state = 74},
  [9039] = {.lex_state = 152, .external_lex_state = 75},
  [9040] = {.lex_state = 151, .external_lex_state = 56},
  [9041] = {.lex_state = 152, .external_lex_state = 74},
  [9042] = {.lex_state = 151, .external_lex_state = 57},
  [9043] = {.lex_state = 152, .external_lex_state = 75},
  [9044] = {.lex_state = 152, .external_lex_state = 72},
  [9045] = {.lex_state = 151, .external_lex_state = 57},
  [9046] = {.lex_state = 151, .external_lex_state = 58},
  [9047] = {.lex_state = 152, .external_lex_state = 73},
  [9048] = {.lex_state = 152, .external_lex_state = 73},
  [9049] = {.lex_state = 152, .external_lex_state = 73},
  [9050] = {.lex_state = 151, .external_lex_state = 58},
  [9051] = {.lex_state = 161, .external_lex_state = 74},
  [9052] = {.lex_state = 152, .external_lex_state = 75},
  [9053] = {.lex_state = 151, .external_lex_state = 55},
  [9054] = {.lex_state = 151, .external_lex_state = 57},
  [9055] = {.lex_state = 161, .external_lex_state = 74},
  [9056] = {.lex_state = 153, .external_lex_state = 56},
  [9057] = {.lex_state = 151, .external_lex_state = 57},
  [9058] = {.lex_state = 152, .external_lex_state = 72},
  [9059] = {.lex_state = 152, .external_lex_state = 72},
  [9060] = {.lex_state = 151, .external_lex_state = 55},
  [9061] = {.lex_state = 152, .external_lex_state = 72},
  [9062] = {.lex_state = 162, .external_lex_state = 55},
  [9063] = {.lex_state = 153, .external_lex_state = 55},
  [9064] = {.lex_state = 151, .external_lex_state = 57},
  [9065] = {.lex_state = 153, .external_lex_state = 55},
  [9066] = {.lex_state = 151, .external_lex_state = 58},
  [9067] = {.lex_state = 152, .external_lex_state = 74},
  [9068] = {.lex_state = 151, .external_lex_state = 56},
  [9069] = {.lex_state = 152, .external_lex_state = 75},
  [9070] = {.lex_state = 163, .external_lex_state = 74},
  [9071] = {.lex_state = 152, .external_lex_state = 72},
  [9072] = {.lex_state = 151, .external_lex_state = 55},
  [9073] = {.lex_state = 153, .external_lex_state = 55},
  [9074] = {.lex_state = 161, .external_lex_state = 74},
  [9075] = {.lex_state = 153, .external_lex_state = 55},
  [9076] = {.lex_state = 151, .external_lex_state = 58},
  [9077] = {.lex_state = 151, .external_lex_state = 56},
  [9078] = {.lex_state = 151, .external_lex_state = 58},
  [9079] = {.lex_state = 153, .external_lex_state = 55},
  [9080] = {.lex_state = 163, .external_lex_state = 74},
  [9081] = {.lex_state = 153, .external_lex_state = 57},
  [9082] = {.lex_state = 153, .external_lex_state = 58},
  [9083] = {.lex_state = 152, .external_lex_state = 73},
  [9084] = {.lex_state = 152, .external_lex_state = 74},
  [9085] = {.lex_state = 151, .external_lex_state = 56},
  [9086] = {.lex_state = 153, .external_lex_state = 55},
  [9087] = {.lex_state = 151, .external_lex_state = 58},
  [9088] = {.lex_state = 152, .external_lex_state = 75},
  [9089] = {.lex_state = 152, .external_lex_state = 75},
  [9090] = {.lex_state = 153, .external_lex_state = 57},
  [9091] = {.lex_state = 153, .external_lex_state = 58},
  [9092] = {.lex_state = 161, .external_lex_state = 74},
  [9093] = {.lex_state = 152, .external_lex_state = 75},
  [9094] = {.lex_state = 162, .external_lex_state = 55},
  [9095] = {.lex_state = 151, .external_lex_state = 58},
  [9096] = {.lex_state = 163, .external_lex_state = 74},
  [9097] = {.lex_state = 152, .external_lex_state = 74},
  [9098] = {.lex_state = 163, .external_lex_state = 74},
  [9099] = {.lex_state = 152, .external_lex_state = 73},
  [9100] = {.lex_state = 161, .external_lex_state = 74},
  [9101] = {.lex_state = 161, .external_lex_state = 74},
  [9102] = {.lex_state = 163, .external_lex_state = 74},
  [9103] = {.lex_state = 153, .external_lex_state = 55},
  [9104] = {.lex_state = 162, .external_lex_state = 55},
  [9105] = {.lex_state = 152, .external_lex_state = 72},
  [9106] = {.lex_state = 152, .external_lex_state = 73},
  [9107] = {.lex_state = 151, .external_lex_state = 57},
  [9108] = {.lex_state = 163, .external_lex_state = 74},
  [9109] = {.lex_state = 151, .external_lex_state = 56},
  [9110] = {.lex_state = 151, .external_lex_state = 55},
  [9111] = {.lex_state = 153, .external_lex_state = 55},
  [9112] = {.lex_state = 161, .external_lex_state = 74},
  [9113] = {.lex_state = 162, .external_lex_state = 55},
  [9114] = {.lex_state = 161, .external_lex_state = 74},
  [9115] = {.lex_state = 151, .external_lex_state = 55},
  [9116] = {.lex_state = 162, .external_lex_state = 55},
  [9117] = {.lex_state = 162, .external_lex_state = 55},
  [9118] = {.lex_state = 162, .external_lex_state = 55},
  [9119] = {.lex_state = 153, .external_lex_state = 58},
  [9120] = {.lex_state = 153, .external_lex_state = 58},
  [9121] = {.lex_state = 151, .external_lex_state = 57},
  [9122] = {.lex_state = 153, .external_lex_state = 56},
  [9123] = {.lex_state = 153, .external_lex_state = 58},
  [9124] = {.lex_state = 163, .external_lex_state = 74},
  [9125] = {.lex_state = 153, .external_lex_state = 56},
  [9126] = {.lex_state = 151, .external_lex_state = 57},
  [9127] = {.lex_state = 151, .external_lex_state = 56},
  [9128] = {.lex_state = 152, .external_lex_state = 74},
  [9129] = {.lex_state = 152, .external_lex_state = 74},
  [9130] = {.lex_state = 152, .external_lex_state = 73},
  [9131] = {.lex_state = 163, .external_lex_state = 74},
  [9132] = {.lex_state = 152, .external_lex_state = 73},
  [9133] = {.lex_state = 152, .external_lex_state = 72},
  [9134] = {.lex_state = 153, .external_lex_state = 56},
  [9135] = {.lex_state = 153, .external_lex_state = 57},
  [9136] = {.lex_state = 153, .external_lex_state = 57},
  [9137] = {.lex_state = 153, .external_lex_state = 56},
  [9138] = {.lex_state = 153, .external_lex_state = 57},
  [9139] = {.lex_state = 152, .external_lex_state = 75},
  [9140] = {.lex_state = 161, .external_lex_state = 74},
  [9141] = {.lex_state = 153, .external_lex_state = 58},
  [9142] = {.lex_state = 153, .external_lex_state = 58},
  [9143] = {.lex_state = 153, .external_lex_state = 56},
  [9144] = {.lex_state = 151, .external_lex_state = 55},
  [9145] = {.lex_state = 163, .external_lex_state = 74},
  [9146] = {.lex_state = 151, .external_lex_state = 57},
  [9147] = {.lex_state = 151, .external_lex_state = 55},
  [9148] = {.lex_state = 153, .external_lex_state = 58},
  [9149] = {.lex_state = 151, .external_lex_state = 55},
  [9150] = {.lex_state = 161, .external_lex_state = 74},
  [9151] = {.lex_state = 163, .external_lex_state = 74},
  [9152] = {.lex_state = 151, .external_lex_state = 55},
  [9153] = {.lex_state = 152, .external_lex_state = 73},
  [9154] = {.lex_state = 152, .external_lex_state = 74},
  [9155] = {.lex_state = 153, .external_lex_state = 56},
  [9156] = {.lex_state = 153, .external_lex_state = 57},
  [9157] = {.lex_state = 162, .external_lex_state = 55},
  [9158] = {.lex_state = 153, .external_lex_state = 58},
  [9159] = {.lex_state = 153, .external_lex_state = 57},
  [9160] = {.lex_state = 153, .external_lex_state = 57},
  [9161] = {.lex_state = 152, .external_lex_state = 73},
  [9162] = {.lex_state = 152, .external_lex_state = 73},
  [9163] = {.lex_state = 153, .external_lex_state = 55},
  [9164] = {.lex_state = 164, .external_lex_state = 55},
  [9165] = {.lex_state = 161, .external_lex_state = 74},
  [9166] = {.lex_state = 151, .external_lex_state = 58},
  [9167] = {.lex_state = 152, .external_lex_state = 73},
  [9168] = {.lex_state = 153, .external_lex_state = 56},
  [9169] = {.lex_state = 152, .external_lex_state = 73},
  [9170] = {.lex_state = 163, .external_lex_state = 74},
  [9171] = {.lex_state = 151, .external_lex_state = 57},
  [9172] = {.lex_state = 153, .external_lex_state = 55},
  [9173] = {.lex_state = 190, .external_lex_state = 68},
  [9174] = {.lex_state = 151, .external_lex_state = 57},
  [9175] = {.lex_state = 151, .external_lex_state = 57},
  [9176] = {.lex_state = 162, .external_lex_state = 55},
  [9177] = {.lex_state = 151, .external_lex_state = 57},
  [9178] = {.lex_state = 151, .external_lex_state = 57},
  [9179] = {.lex_state = 153, .external_lex_state = 58},
  [9180] = {.lex_state = 153, .external_lex_state = 56},
  [9181] = {.lex_state = 455, .external_lex_state = 55},
  [9182] = {.lex_state = 151, .external_lex_state = 55},
  [9183] = {.lex_state = 151, .external_lex_state = 57},
  [9184] = {.lex_state = 151, .external_lex_state = 57},
  [9185] = {.lex_state = 151, .external_lex_state = 57},
  [9186] = {.lex_state = 163, .external_lex_state = 74},
  [9187] = {.lex_state = 153, .external_lex_state = 55},
  [9188] = {.lex_state = 151, .external_lex_state = 57},
  [9189] = {.lex_state = 151, .external_lex_state = 57},
  [9190] = {.lex_state = 151, .external_lex_state = 57},
  [9191] = {.lex_state = 153, .external_lex_state = 58},
  [9192] = {.lex_state = 163, .external_lex_state = 74},
  [9193] = {.lex_state = 151, .external_lex_state = 56},
  [9194] = {.lex_state = 152, .external_lex_state = 74},
  [9195] = {.lex_state = 164, .external_lex_state = 55},
  [9196] = {.lex_state = 151, .external_lex_state = 56},
  [9197] = {.lex_state = 152, .external_lex_state = 75},
  [9198] = {.lex_state = 153, .external_lex_state = 57},
  [9199] = {.lex_state = 152, .external_lex_state = 72},
  [9200] = {.lex_state = 164, .external_lex_state = 55},
  [9201] = {.lex_state = 151, .external_lex_state = 56},
  [9202] = {.lex_state = 151, .external_lex_state = 56},
  [9203] = {.lex_state = 183, .external_lex_state = 68},
  [9204] = {.lex_state = 151, .external_lex_state = 56},
  [9205] = {.lex_state = 151, .external_lex_state = 56},
  [9206] = {.lex_state = 161, .external_lex_state = 74},
  [9207] = {.lex_state = 151, .external_lex_state = 56},
  [9208] = {.lex_state = 151, .external_lex_state = 58},
  [9209] = {.lex_state = 152, .external_lex_state = 72},
  [9210] = {.lex_state = 151, .external_lex_state = 56},
  [9211] = {.lex_state = 164, .external_lex_state = 55},
  [9212] = {.lex_state = 151, .external_lex_state = 56},
  [9213] = {.lex_state = 161, .external_lex_state = 74},
  [9214] = {.lex_state = 151, .external_lex_state = 56},
  [9215] = {.lex_state = 152, .external_lex_state = 72},
  [9216] = {.lex_state = 152, .external_lex_state = 75},
  [9217] = {.lex_state = 151, .external_lex_state = 58},
  [9218] = {.lex_state = 151, .external_lex_state = 56},
  [9219] = {.lex_state = 153, .external_lex_state = 55},
  [9220] = {.lex_state = 151, .external_lex_state = 56},
  [9221] = {.lex_state = 152, .external_lex_state = 72},
  [9222] = {.lex_state = 151, .external_lex_state = 56},
  [9223] = {.lex_state = 152, .external_lex_state = 73},
  [9224] = {.lex_state = 152, .external_lex_state = 75},
  [9225] = {.lex_state = 152, .external_lex_state = 72},
  [9226] = {.lex_state = 152, .external_lex_state = 74},
  [9227] = {.lex_state = 152, .external_lex_state = 72},
  [9228] = {.lex_state = 152, .external_lex_state = 72},
  [9229] = {.lex_state = 152, .external_lex_state = 72},
  [9230] = {.lex_state = 152, .external_lex_state = 72},
  [9231] = {.lex_state = 152, .external_lex_state = 73},
  [9232] = {.lex_state = 152, .external_lex_state = 74},
  [9233] = {.lex_state = 162, .external_lex_state = 55},
  [9234] = {.lex_state = 153, .external_lex_state = 55},
  [9235] = {.lex_state = 164, .external_lex_state = 55},
  [9236] = {.lex_state = 164, .external_lex_state = 55},
  [9237] = {.lex_state = 164, .external_lex_state = 55},
  [9238] = {.lex_state = 151, .external_lex_state = 55},
  [9239] = {.lex_state = 162, .external_lex_state = 55},
  [9240] = {.lex_state = 152, .external_lex_state = 73},
  [9241] = {.lex_state = 151, .external_lex_state = 55},
  [9242] = {.lex_state = 153, .external_lex_state = 56},
  [9243] = {.lex_state = 152, .external_lex_state = 72},
  [9244] = {.lex_state = 152, .external_lex_state = 73},
  [9245] = {.lex_state = 152, .external_lex_state = 72},
  [9246] = {.lex_state = 162, .external_lex_state = 55},
  [9247] = {.lex_state = 168, .external_lex_state = 68},
  [9248] = {.lex_state = 152, .external_lex_state = 72},
  [9249] = {.lex_state = 152, .external_lex_state = 73},
  [9250] = {.lex_state = 151, .external_lex_state = 57},
  [9251] = {.lex_state = 152, .external_lex_state = 75},
  [9252] = {.lex_state = 151, .external_lex_state = 57},
  [9253] = {.lex_state = 151, .external_lex_state = 57},
  [9254] = {.lex_state = 152, .external_lex_state = 73},
  [9255] = {.lex_state = 151, .external_lex_state = 58},
  [9256] = {.lex_state = 152, .external_lex_state = 73},
  [9257] = {.lex_state = 152, .external_lex_state = 72},
  [9258] = {.lex_state = 151, .external_lex_state = 58},
  [9259] = {.lex_state = 152, .external_lex_state = 75},
  [9260] = {.lex_state = 152, .external_lex_state = 73},
  [9261] = {.lex_state = 152, .external_lex_state = 72},
  [9262] = {.lex_state = 153, .external_lex_state = 57},
  [9263] = {.lex_state = 152, .external_lex_state = 74},
  [9264] = {.lex_state = 153, .external_lex_state = 58},
  [9265] = {.lex_state = 163, .external_lex_state = 74},
  [9266] = {.lex_state = 153, .external_lex_state = 58},
  [9267] = {.lex_state = 161, .external_lex_state = 74},
  [9268] = {.lex_state = 152, .external_lex_state = 73},
  [9269] = {.lex_state = 164, .external_lex_state = 55},
  [9270] = {.lex_state = 151, .external_lex_state = 56},
  [9271] = {.lex_state = 153, .external_lex_state = 56},
  [9272] = {.lex_state = 151, .external_lex_state = 58},
  [9273] = {.lex_state = 223},
  [9274] = {.lex_state = 203, .external_lex_state = 76},
  [9275] = {.lex_state = 151, .external_lex_state = 58},
  [9276] = {.lex_state = 203, .external_lex_state = 76},
  [9277] = {.lex_state = 223},
  [9278] = {.lex_state = 153, .external_lex_state = 57},
  [9279] = {.lex_state = 153, .external_lex_state = 57},
  [9280] = {.lex_state = 161, .external_lex_state = 74},
  [9281] = {.lex_state = 153, .external_lex_state = 56},
  [9282] = {.lex_state = 153, .external_lex_state = 55},
  [9283] = {.lex_state = 161, .external_lex_state = 74},
  [9284] = {.lex_state = 164, .external_lex_state = 55},
  [9285] = {.lex_state = 153, .external_lex_state = 57},
  [9286] = {.lex_state = 151, .external_lex_state = 58},
  [9287] = {.lex_state = 169, .external_lex_state = 68},
  [9288] = {.lex_state = 157, .external_lex_state = 69},
  [9289] = {.lex_state = 153, .external_lex_state = 58},
  [9290] = {.lex_state = 203, .external_lex_state = 76},
  [9291] = {.lex_state = 203, .external_lex_state = 76},
  [9292] = {.lex_state = 203, .external_lex_state = 76},
  [9293] = {.lex_state = 153, .external_lex_state = 55},
  [9294] = {.lex_state = 153, .external_lex_state = 58},
  [9295] = {.lex_state = 223},
  [9296] = {.lex_state = 161, .external_lex_state = 74},
  [9297] = {.lex_state = 153, .external_lex_state = 57},
  [9298] = {.lex_state = 152, .external_lex_state = 74},
  [9299] = {.lex_state = 151, .external_lex_state = 55},
  [9300] = {.lex_state = 157, .external_lex_state = 70},
  [9301] = {.lex_state = 153, .external_lex_state = 57},
  [9302] = {.lex_state = 164, .external_lex_state = 55},
  [9303] = {.lex_state = 161, .external_lex_state = 74},
  [9304] = {.lex_state = 151, .external_lex_state = 58},
  [9305] = {.lex_state = 153, .external_lex_state = 55},
  [9306] = {.lex_state = 153, .external_lex_state = 55},
  [9307] = {.lex_state = 153, .external_lex_state = 58},
  [9308] = {.lex_state = 203, .external_lex_state = 76},
  [9309] = {.lex_state = 223},
  [9310] = {.lex_state = 153, .external_lex_state = 57},
  [9311] = {.lex_state = 162, .external_lex_state = 55},
  [9312] = {.lex_state = 169, .external_lex_state = 68},
  [9313] = {.lex_state = 153, .external_lex_state = 57},
  [9314] = {.lex_state = 455, .external_lex_state = 55},
  [9315] = {.lex_state = 153, .external_lex_state = 57},
  [9316] = {.lex_state = 169, .external_lex_state = 68},
  [9317] = {.lex_state = 153, .external_lex_state = 58},
  [9318] = {.lex_state = 455, .external_lex_state = 55},
  [9319] = {.lex_state = 151, .external_lex_state = 55},
  [9320] = {.lex_state = 152, .external_lex_state = 75},
  [9321] = {.lex_state = 152, .external_lex_state = 74},
  [9322] = {.lex_state = 223},
  [9323] = {.lex_state = 203, .external_lex_state = 76},
  [9324] = {.lex_state = 153, .external_lex_state = 56},
  [9325] = {.lex_state = 223},
  [9326] = {.lex_state = 203, .external_lex_state = 76},
  [9327] = {.lex_state = 203, .external_lex_state = 76},
  [9328] = {.lex_state = 223},
  [9329] = {.lex_state = 455, .external_lex_state = 55},
  [9330] = {.lex_state = 162, .external_lex_state = 55},
  [9331] = {.lex_state = 164, .external_lex_state = 55},
  [9332] = {.lex_state = 455, .external_lex_state = 55},
  [9333] = {.lex_state = 151, .external_lex_state = 55},
  [9334] = {.lex_state = 152, .external_lex_state = 75},
  [9335] = {.lex_state = 153, .external_lex_state = 57},
  [9336] = {.lex_state = 153, .external_lex_state = 56},
  [9337] = {.lex_state = 223},
  [9338] = {.lex_state = 223},
  [9339] = {.lex_state = 203, .external_lex_state = 76},
  [9340] = {.lex_state = 223},
  [9341] = {.lex_state = 152, .external_lex_state = 74},
  [9342] = {.lex_state = 455, .external_lex_state = 55},
  [9343] = {.lex_state = 161, .external_lex_state = 74},
  [9344] = {.lex_state = 151, .external_lex_state = 55},
  [9345] = {.lex_state = 203, .external_lex_state = 76},
  [9346] = {.lex_state = 152, .external_lex_state = 75},
  [9347] = {.lex_state = 151, .external_lex_state = 55},
  [9348] = {.lex_state = 223},
  [9349] = {.lex_state = 203, .external_lex_state = 76},
  [9350] = {.lex_state = 223},
  [9351] = {.lex_state = 203, .external_lex_state = 76},
  [9352] = {.lex_state = 151, .external_lex_state = 58},
  [9353] = {.lex_state = 152, .external_lex_state = 74},
  [9354] = {.lex_state = 203, .external_lex_state = 76},
  [9355] = {.lex_state = 203, .external_lex_state = 76},
  [9356] = {.lex_state = 151, .external_lex_state = 58},
  [9357] = {.lex_state = 223},
  [9358] = {.lex_state = 151, .external_lex_state = 58},
  [9359] = {.lex_state = 153, .external_lex_state = 56},
  [9360] = {.lex_state = 203, .external_lex_state = 76},
  [9361] = {.lex_state = 223},
  [9362] = {.lex_state = 455, .external_lex_state = 55},
  [9363] = {.lex_state = 455, .external_lex_state = 55},
  [9364] = {.lex_state = 161, .external_lex_state = 74},
  [9365] = {.lex_state = 151, .external_lex_state = 55},
  [9366] = {.lex_state = 223},
  [9367] = {.lex_state = 151, .external_lex_state = 55},
  [9368] = {.lex_state = 153, .external_lex_state = 57},
  [9369] = {.lex_state = 163, .external_lex_state = 74},
  [9370] = {.lex_state = 152, .external_lex_state = 74},
  [9371] = {.lex_state = 151, .external_lex_state = 58},
  [9372] = {.lex_state = 152, .external_lex_state = 74},
  [9373] = {.lex_state = 203, .external_lex_state = 76},
  [9374] = {.lex_state = 153, .external_lex_state = 57},
  [9375] = {.lex_state = 152, .external_lex_state = 75},
  [9376] = {.lex_state = 203, .external_lex_state = 76},
  [9377] = {.lex_state = 223},
  [9378] = {.lex_state = 153, .external_lex_state = 56},
  [9379] = {.lex_state = 152, .external_lex_state = 75},
  [9380] = {.lex_state = 162, .external_lex_state = 55},
  [9381] = {.lex_state = 151, .external_lex_state = 58},
  [9382] = {.lex_state = 152, .external_lex_state = 74},
  [9383] = {.lex_state = 164, .external_lex_state = 55},
  [9384] = {.lex_state = 153, .external_lex_state = 57},
  [9385] = {.lex_state = 455, .external_lex_state = 55},
  [9386] = {.lex_state = 153, .external_lex_state = 57},
  [9387] = {.lex_state = 161, .external_lex_state = 74},
  [9388] = {.lex_state = 151, .external_lex_state = 55},
  [9389] = {.lex_state = 223},
  [9390] = {.lex_state = 169, .external_lex_state = 68},
  [9391] = {.lex_state = 152, .external_lex_state = 75},
  [9392] = {.lex_state = 151, .external_lex_state = 58},
  [9393] = {.lex_state = 151, .external_lex_state = 55},
  [9394] = {.lex_state = 223},
  [9395] = {.lex_state = 223},
  [9396] = {.lex_state = 223},
  [9397] = {.lex_state = 152, .external_lex_state = 75},
  [9398] = {.lex_state = 203, .external_lex_state = 76},
  [9399] = {.lex_state = 152, .external_lex_state = 75},
  [9400] = {.lex_state = 153, .external_lex_state = 56},
  [9401] = {.lex_state = 161, .external_lex_state = 74},
  [9402] = {.lex_state = 153, .external_lex_state = 56},
  [9403] = {.lex_state = 152, .external_lex_state = 75},
  [9404] = {.lex_state = 203, .external_lex_state = 76},
  [9405] = {.lex_state = 223},
  [9406] = {.lex_state = 203, .external_lex_state = 76},
  [9407] = {.lex_state = 203, .external_lex_state = 76},
  [9408] = {.lex_state = 153, .external_lex_state = 56},
  [9409] = {.lex_state = 163, .external_lex_state = 74},
  [9410] = {.lex_state = 152, .external_lex_state = 75},
  [9411] = {.lex_state = 152, .external_lex_state = 74},
  [9412] = {.lex_state = 153, .external_lex_state = 56},
  [9413] = {.lex_state = 203, .external_lex_state = 76},
  [9414] = {.lex_state = 151, .external_lex_state = 58},
  [9415] = {.lex_state = 223},
  [9416] = {.lex_state = 153, .external_lex_state = 56},
  [9417] = {.lex_state = 151, .external_lex_state = 58},
  [9418] = {.lex_state = 223},
  [9419] = {.lex_state = 223},
  [9420] = {.lex_state = 152, .external_lex_state = 75},
  [9421] = {.lex_state = 151, .external_lex_state = 55},
  [9422] = {.lex_state = 151, .external_lex_state = 58},
  [9423] = {.lex_state = 152, .external_lex_state = 74},
  [9424] = {.lex_state = 152, .external_lex_state = 75},
  [9425] = {.lex_state = 152, .external_lex_state = 74},
  [9426] = {.lex_state = 152, .external_lex_state = 75},
  [9427] = {.lex_state = 153, .external_lex_state = 56},
  [9428] = {.lex_state = 152, .external_lex_state = 74},
  [9429] = {.lex_state = 163, .external_lex_state = 74},
  [9430] = {.lex_state = 152, .external_lex_state = 75},
  [9431] = {.lex_state = 163, .external_lex_state = 74},
  [9432] = {.lex_state = 152, .external_lex_state = 74},
  [9433] = {.lex_state = 455, .external_lex_state = 55},
  [9434] = {.lex_state = 203, .external_lex_state = 76},
  [9435] = {.lex_state = 152, .external_lex_state = 74},
  [9436] = {.lex_state = 151, .external_lex_state = 55},
  [9437] = {.lex_state = 223},
  [9438] = {.lex_state = 162, .external_lex_state = 55},
  [9439] = {.lex_state = 163, .external_lex_state = 74},
  [9440] = {.lex_state = 153, .external_lex_state = 55},
  [9441] = {.lex_state = 161, .external_lex_state = 74},
  [9442] = {.lex_state = 223},
  [9443] = {.lex_state = 153, .external_lex_state = 58},
  [9444] = {.lex_state = 164, .external_lex_state = 55},
  [9445] = {.lex_state = 153, .external_lex_state = 56},
  [9446] = {.lex_state = 151, .external_lex_state = 55},
  [9447] = {.lex_state = 151, .external_lex_state = 58},
  [9448] = {.lex_state = 152, .external_lex_state = 74},
  [9449] = {.lex_state = 153, .external_lex_state = 56},
  [9450] = {.lex_state = 153, .external_lex_state = 56},
  [9451] = {.lex_state = 161, .external_lex_state = 74},
  [9452] = {.lex_state = 161, .external_lex_state = 74},
  [9453] = {.lex_state = 223},
  [9454] = {.lex_state = 151, .external_lex_state = 55},
  [9455] = {.lex_state = 455, .external_lex_state = 55},
  [9456] = {.lex_state = 161, .external_lex_state = 74},
  [9457] = {.lex_state = 161, .external_lex_state = 74},
  [9458] = {.lex_state = 161, .external_lex_state = 74},
  [9459] = {.lex_state = 223},
  [9460] = {.lex_state = 162, .external_lex_state = 55},
  [9461] = {.lex_state = 151, .external_lex_state = 55},
  [9462] = {.lex_state = 153, .external_lex_state = 57},
  [9463] = {.lex_state = 153, .external_lex_state = 55},
  [9464] = {.lex_state = 162, .external_lex_state = 55},
  [9465] = {.lex_state = 154, .external_lex_state = 56},
  [9466] = {.lex_state = 223},
  [9467] = {.lex_state = 164, .external_lex_state = 55},
  [9468] = {.lex_state = 223},
  [9469] = {.lex_state = 164, .external_lex_state = 55},
  [9470] = {.lex_state = 158, .external_lex_state = 69},
  [9471] = {.lex_state = 223},
  [9472] = {.lex_state = 164, .external_lex_state = 55},
  [9473] = {.lex_state = 223},
  [9474] = {.lex_state = 223},
  [9475] = {.lex_state = 153, .external_lex_state = 58},
  [9476] = {.lex_state = 223},
  [9477] = {.lex_state = 162, .external_lex_state = 55},
  [9478] = {.lex_state = 223},
  [9479] = {.lex_state = 223},
  [9480] = {.lex_state = 162, .external_lex_state = 55},
  [9481] = {.lex_state = 162, .external_lex_state = 55},
  [9482] = {.lex_state = 162, .external_lex_state = 55},
  [9483] = {.lex_state = 162, .external_lex_state = 55},
  [9484] = {.lex_state = 153, .external_lex_state = 55},
  [9485] = {.lex_state = 157, .external_lex_state = 71},
  [9486] = {.lex_state = 153, .external_lex_state = 58},
  [9487] = {.lex_state = 153, .external_lex_state = 58},
  [9488] = {.lex_state = 162, .external_lex_state = 55},
  [9489] = {.lex_state = 153, .external_lex_state = 58},
  [9490] = {.lex_state = 162, .external_lex_state = 55},
  [9491] = {.lex_state = 158, .external_lex_state = 69},
  [9492] = {.lex_state = 153, .external_lex_state = 58},
  [9493] = {.lex_state = 153, .external_lex_state = 58},
  [9494] = {.lex_state = 153, .external_lex_state = 55},
  [9495] = {.lex_state = 162, .external_lex_state = 55},
  [9496] = {.lex_state = 162, .external_lex_state = 55},
  [9497] = {.lex_state = 162, .external_lex_state = 55},
  [9498] = {.lex_state = 165, .external_lex_state = 55},
  [9499] = {.lex_state = 153, .external_lex_state = 58},
  [9500] = {.lex_state = 153, .external_lex_state = 55},
  [9501] = {.lex_state = 153, .external_lex_state = 58},
  [9502] = {.lex_state = 153, .external_lex_state = 58},
  [9503] = {.lex_state = 165, .external_lex_state = 55},
  [9504] = {.lex_state = 153, .external_lex_state = 58},
  [9505] = {.lex_state = 154, .external_lex_state = 57},
  [9506] = {.lex_state = 153, .external_lex_state = 58},
  [9507] = {.lex_state = 153, .external_lex_state = 58},
  [9508] = {.lex_state = 184, .external_lex_state = 68},
  [9509] = {.lex_state = 153, .external_lex_state = 58},
  [9510] = {.lex_state = 153, .external_lex_state = 58},
  [9511] = {.lex_state = 223},
  [9512] = {.lex_state = 153, .external_lex_state = 55},
  [9513] = {.lex_state = 154, .external_lex_state = 57},
  [9514] = {.lex_state = 163, .external_lex_state = 74},
  [9515] = {.lex_state = 158, .external_lex_state = 69},
  [9516] = {.lex_state = 163, .external_lex_state = 74},
  [9517] = {.lex_state = 163, .external_lex_state = 74},
  [9518] = {.lex_state = 163, .external_lex_state = 74},
  [9519] = {.lex_state = 163, .external_lex_state = 74},
  [9520] = {.lex_state = 163, .external_lex_state = 74},
  [9521] = {.lex_state = 163, .external_lex_state = 74},
  [9522] = {.lex_state = 163, .external_lex_state = 74},
  [9523] = {.lex_state = 163, .external_lex_state = 74},
  [9524] = {.lex_state = 163, .external_lex_state = 74},
  [9525] = {.lex_state = 163, .external_lex_state = 74},
  [9526] = {.lex_state = 153, .external_lex_state = 55},
  [9527] = {.lex_state = 154, .external_lex_state = 57},
  [9528] = {.lex_state = 223},
  [9529] = {.lex_state = 157, .external_lex_state = 68},
  [9530] = {.lex_state = 153, .external_lex_state = 55},
  [9531] = {.lex_state = 158, .external_lex_state = 69},
  [9532] = {.lex_state = 153, .external_lex_state = 55},
  [9533] = {.lex_state = 223},
  [9534] = {.lex_state = 162, .external_lex_state = 55},
  [9535] = {.lex_state = 162, .external_lex_state = 55},
  [9536] = {.lex_state = 162, .external_lex_state = 55},
  [9537] = {.lex_state = 163, .external_lex_state = 74},
  [9538] = {.lex_state = 163, .external_lex_state = 74},
  [9539] = {.lex_state = 163, .external_lex_state = 74},
  [9540] = {.lex_state = 154, .external_lex_state = 56},
  [9541] = {.lex_state = 165, .external_lex_state = 55},
  [9542] = {.lex_state = 165, .external_lex_state = 55},
  [9543] = {.lex_state = 153, .external_lex_state = 55},
  [9544] = {.lex_state = 165, .external_lex_state = 55},
  [9545] = {.lex_state = 223},
  [9546] = {.lex_state = 153, .external_lex_state = 55},
  [9547] = {.lex_state = 164, .external_lex_state = 55},
  [9548] = {.lex_state = 203, .external_lex_state = 76},
  [9549] = {.lex_state = 164, .external_lex_state = 55},
  [9550] = {.lex_state = 223},
  [9551] = {.lex_state = 158, .external_lex_state = 70},
  [9552] = {.lex_state = 158, .external_lex_state = 70},
  [9553] = {.lex_state = 153, .external_lex_state = 55},
  [9554] = {.lex_state = 153, .external_lex_state = 55},
  [9555] = {.lex_state = 223},
  [9556] = {.lex_state = 158, .external_lex_state = 70},
  [9557] = {.lex_state = 153, .external_lex_state = 55},
  [9558] = {.lex_state = 158, .external_lex_state = 70},
  [9559] = {.lex_state = 165, .external_lex_state = 55},
  [9560] = {.lex_state = 154, .external_lex_state = 56},
  [9561] = {.lex_state = 165, .external_lex_state = 55},
  [9562] = {.lex_state = 191, .external_lex_state = 68},
  [9563] = {.lex_state = 153, .external_lex_state = 55},
  [9564] = {.lex_state = 160, .external_lex_state = 73},
  [9565] = {.lex_state = 171, .external_lex_state = 74},
  [9566] = {.lex_state = 164, .external_lex_state = 55},
  [9567] = {.lex_state = 165, .external_lex_state = 55},
  [9568] = {.lex_state = 158, .external_lex_state = 68},
  [9569] = {.lex_state = 154, .external_lex_state = 57},
  [9570] = {.lex_state = 164, .external_lex_state = 55},
  [9571] = {.lex_state = 154, .external_lex_state = 56},
  [9572] = {.lex_state = 158, .external_lex_state = 68},
  [9573] = {.lex_state = 164, .external_lex_state = 55},
  [9574] = {.lex_state = 165, .external_lex_state = 55},
  [9575] = {.lex_state = 203, .external_lex_state = 77},
  [9576] = {.lex_state = 203, .external_lex_state = 77},
  [9577] = {.lex_state = 165, .external_lex_state = 55},
  [9578] = {.lex_state = 165, .external_lex_state = 55},
  [9579] = {.lex_state = 158, .external_lex_state = 68},
  [9580] = {.lex_state = 154, .external_lex_state = 58},
  [9581] = {.lex_state = 203, .external_lex_state = 78},
  [9582] = {.lex_state = 203, .external_lex_state = 77},
  [9583] = {.lex_state = 154, .external_lex_state = 56},
  [9584] = {.lex_state = 203, .external_lex_state = 77},
  [9585] = {.lex_state = 154, .external_lex_state = 58},
  [9586] = {.lex_state = 203, .external_lex_state = 77},
  [9587] = {.lex_state = 203, .external_lex_state = 77},
  [9588] = {.lex_state = 171, .external_lex_state = 74},
  [9589] = {.lex_state = 165, .external_lex_state = 55},
  [9590] = {.lex_state = 154, .external_lex_state = 56},
  [9591] = {.lex_state = 203, .external_lex_state = 77},
  [9592] = {.lex_state = 203, .external_lex_state = 77},
  [9593] = {.lex_state = 164, .external_lex_state = 55},
  [9594] = {.lex_state = 203, .external_lex_state = 77},
  [9595] = {.lex_state = 171, .external_lex_state = 74},
  [9596] = {.lex_state = 164, .external_lex_state = 55},
  [9597] = {.lex_state = 203, .external_lex_state = 77},
  [9598] = {.lex_state = 158, .external_lex_state = 68},
  [9599] = {.lex_state = 203, .external_lex_state = 77},
  [9600] = {.lex_state = 171, .external_lex_state = 74},
  [9601] = {.lex_state = 160, .external_lex_state = 72},
  [9602] = {.lex_state = 164, .external_lex_state = 55},
  [9603] = {.lex_state = 203, .external_lex_state = 77},
  [9604] = {.lex_state = 203, .external_lex_state = 77},
  [9605] = {.lex_state = 165, .external_lex_state = 55},
  [9606] = {.lex_state = 165, .external_lex_state = 55},
  [9607] = {.lex_state = 203, .external_lex_state = 77},
  [9608] = {.lex_state = 154, .external_lex_state = 57},
  [9609] = {.lex_state = 154, .external_lex_state = 56},
  [9610] = {.lex_state = 203, .external_lex_state = 77},
  [9611] = {.lex_state = 165, .external_lex_state = 55},
  [9612] = {.lex_state = 164, .external_lex_state = 55},
  [9613] = {.lex_state = 160, .external_lex_state = 72},
  [9614] = {.lex_state = 203, .external_lex_state = 77},
  [9615] = {.lex_state = 160, .external_lex_state = 72},
  [9616] = {.lex_state = 164, .external_lex_state = 55},
  [9617] = {.lex_state = 158, .external_lex_state = 71},
  [9618] = {.lex_state = 165, .external_lex_state = 68},
  [9619] = {.lex_state = 165, .external_lex_state = 68},
  [9620] = {.lex_state = 165, .external_lex_state = 68},
  [9621] = {.lex_state = 165, .external_lex_state = 68},
  [9622] = {.lex_state = 165, .external_lex_state = 68},
  [9623] = {.lex_state = 165, .external_lex_state = 68},
  [9624] = {.lex_state = 165, .external_lex_state = 68},
  [9625] = {.lex_state = 165, .external_lex_state = 68},
  [9626] = {.lex_state = 165, .external_lex_state = 68},
  [9627] = {.lex_state = 165, .external_lex_state = 68},
  [9628] = {.lex_state = 164, .external_lex_state = 55},
  [9629] = {.lex_state = 165, .external_lex_state = 68},
  [9630] = {.lex_state = 165, .external_lex_state = 68},
  [9631] = {.lex_state = 165, .external_lex_state = 68},
  [9632] = {.lex_state = 165, .external_lex_state = 68},
  [9633] = {.lex_state = 165, .external_lex_state = 68},
  [9634] = {.lex_state = 165, .external_lex_state = 68},
  [9635] = {.lex_state = 165, .external_lex_state = 68},
  [9636] = {.lex_state = 165, .external_lex_state = 68},
  [9637] = {.lex_state = 165, .external_lex_state = 68},
  [9638] = {.lex_state = 165, .external_lex_state = 68},
  [9639] = {.lex_state = 165, .external_lex_state = 68},
  [9640] = {.lex_state = 165, .external_lex_state = 68},
  [9641] = {.lex_state = 171, .external_lex_state = 74},
  [9642] = {.lex_state = 165, .external_lex_state = 68},
  [9643] = {.lex_state = 165, .external_lex_state = 68},
  [9644] = {.lex_state = 165, .external_lex_state = 68},
  [9645] = {.lex_state = 165, .external_lex_state = 68},
  [9646] = {.lex_state = 165, .external_lex_state = 68},
  [9647] = {.lex_state = 164, .external_lex_state = 55},
  [9648] = {.lex_state = 165, .external_lex_state = 55},
  [9649] = {.lex_state = 164, .external_lex_state = 55},
  [9650] = {.lex_state = 165, .external_lex_state = 68},
  [9651] = {.lex_state = 171, .external_lex_state = 74},
  [9652] = {.lex_state = 154, .external_lex_state = 56},
  [9653] = {.lex_state = 165, .external_lex_state = 68},
  [9654] = {.lex_state = 160, .external_lex_state = 73},
  [9655] = {.lex_state = 171, .external_lex_state = 74},
  [9656] = {.lex_state = 203, .external_lex_state = 78},
  [9657] = {.lex_state = 171, .external_lex_state = 74},
  [9658] = {.lex_state = 171, .external_lex_state = 74},
  [9659] = {.lex_state = 223},
  [9660] = {.lex_state = 203, .external_lex_state = 78},
  [9661] = {.lex_state = 165, .external_lex_state = 55},
  [9662] = {.lex_state = 171, .external_lex_state = 74},
  [9663] = {.lex_state = 164, .external_lex_state = 55},
  [9664] = {.lex_state = 164, .external_lex_state = 55},
  [9665] = {.lex_state = 164, .external_lex_state = 55},
  [9666] = {.lex_state = 154, .external_lex_state = 55},
  [9667] = {.lex_state = 165, .external_lex_state = 55},
  [9668] = {.lex_state = 158, .external_lex_state = 71},
  [9669] = {.lex_state = 154, .external_lex_state = 57},
  [9670] = {.lex_state = 154, .external_lex_state = 57},
  [9671] = {.lex_state = 165, .external_lex_state = 68},
  [9672] = {.lex_state = 165, .external_lex_state = 68},
  [9673] = {.lex_state = 165, .external_lex_state = 55},
  [9674] = {.lex_state = 154, .external_lex_state = 55},
  [9675] = {.lex_state = 165, .external_lex_state = 68},
  [9676] = {.lex_state = 154, .external_lex_state = 58},
  [9677] = {.lex_state = 175, .external_lex_state = 68},
  [9678] = {.lex_state = 154, .external_lex_state = 56},
  [9679] = {.lex_state = 158, .external_lex_state = 71},
  [9680] = {.lex_state = 165, .external_lex_state = 55},
  [9681] = {.lex_state = 154, .external_lex_state = 57},
  [9682] = {.lex_state = 160, .external_lex_state = 73},
  [9683] = {.lex_state = 154, .external_lex_state = 55},
  [9684] = {.lex_state = 158, .external_lex_state = 71},
  [9685] = {.lex_state = 165, .external_lex_state = 68},
  [9686] = {.lex_state = 172, .external_lex_state = 55},
  [9687] = {.lex_state = 154, .external_lex_state = 69},
  [9688] = {.lex_state = 160, .external_lex_state = 73},
  [9689] = {.lex_state = 165, .external_lex_state = 55},
  [9690] = {.lex_state = 203, .external_lex_state = 79},
  [9691] = {.lex_state = 160, .external_lex_state = 73},
  [9692] = {.lex_state = 154, .external_lex_state = 69},
  [9693] = {.lex_state = 203, .external_lex_state = 79},
  [9694] = {.lex_state = 165, .external_lex_state = 55},
  [9695] = {.lex_state = 154, .external_lex_state = 58},
  [9696] = {.lex_state = 160, .external_lex_state = 74},
  [9697] = {.lex_state = 160, .external_lex_state = 73},
  [9698] = {.lex_state = 154, .external_lex_state = 69},
  [9699] = {.lex_state = 165, .external_lex_state = 55},
  [9700] = {.lex_state = 172, .external_lex_state = 55},
  [9701] = {.lex_state = 154, .external_lex_state = 57},
  [9702] = {.lex_state = 154, .external_lex_state = 57},
  [9703] = {.lex_state = 203, .external_lex_state = 79},
  [9704] = {.lex_state = 160, .external_lex_state = 73},
  [9705] = {.lex_state = 154, .external_lex_state = 57},
  [9706] = {.lex_state = 154, .external_lex_state = 57},
  [9707] = {.lex_state = 203, .external_lex_state = 79},
  [9708] = {.lex_state = 154, .external_lex_state = 57},
  [9709] = {.lex_state = 154, .external_lex_state = 57},
  [9710] = {.lex_state = 154, .external_lex_state = 69},
  [9711] = {.lex_state = 154, .external_lex_state = 69},
  [9712] = {.lex_state = 154, .external_lex_state = 57},
  [9713] = {.lex_state = 154, .external_lex_state = 69},
  [9714] = {.lex_state = 154, .external_lex_state = 57},
  [9715] = {.lex_state = 203, .external_lex_state = 79},
  [9716] = {.lex_state = 160, .external_lex_state = 73},
  [9717] = {.lex_state = 165, .external_lex_state = 55},
  [9718] = {.lex_state = 203, .external_lex_state = 79},
  [9719] = {.lex_state = 160, .external_lex_state = 75},
  [9720] = {.lex_state = 154, .external_lex_state = 58},
  [9721] = {.lex_state = 160, .external_lex_state = 74},
  [9722] = {.lex_state = 165, .external_lex_state = 55},
  [9723] = {.lex_state = 160, .external_lex_state = 72},
  [9724] = {.lex_state = 165, .external_lex_state = 55},
  [9725] = {.lex_state = 160, .external_lex_state = 72},
  [9726] = {.lex_state = 165, .external_lex_state = 55},
  [9727] = {.lex_state = 154, .external_lex_state = 69},
  [9728] = {.lex_state = 160, .external_lex_state = 72},
  [9729] = {.lex_state = 160, .external_lex_state = 72},
  [9730] = {.lex_state = 154, .external_lex_state = 69},
  [9731] = {.lex_state = 160, .external_lex_state = 72},
  [9732] = {.lex_state = 160, .external_lex_state = 72},
  [9733] = {.lex_state = 160, .external_lex_state = 72},
  [9734] = {.lex_state = 165, .external_lex_state = 55},
  [9735] = {.lex_state = 165, .external_lex_state = 55},
  [9736] = {.lex_state = 154, .external_lex_state = 69},
  [9737] = {.lex_state = 154, .external_lex_state = 69},
  [9738] = {.lex_state = 165, .external_lex_state = 55},
  [9739] = {.lex_state = 154, .external_lex_state = 58},
  [9740] = {.lex_state = 154, .external_lex_state = 69},
  [9741] = {.lex_state = 165, .external_lex_state = 55},
  [9742] = {.lex_state = 165, .external_lex_state = 55},
  [9743] = {.lex_state = 165, .external_lex_state = 55},
  [9744] = {.lex_state = 165, .external_lex_state = 55},
  [9745] = {.lex_state = 165, .external_lex_state = 55},
  [9746] = {.lex_state = 154, .external_lex_state = 69},
  [9747] = {.lex_state = 203, .external_lex_state = 79},
  [9748] = {.lex_state = 176, .external_lex_state = 68},
  [9749] = {.lex_state = 176, .external_lex_state = 68},
  [9750] = {.lex_state = 160, .external_lex_state = 73},
  [9751] = {.lex_state = 165, .external_lex_state = 55},
  [9752] = {.lex_state = 165, .external_lex_state = 55},
  [9753] = {.lex_state = 172, .external_lex_state = 55},
  [9754] = {.lex_state = 165, .external_lex_state = 55},
  [9755] = {.lex_state = 165, .external_lex_state = 55},
  [9756] = {.lex_state = 154, .external_lex_state = 69},
  [9757] = {.lex_state = 154, .external_lex_state = 69},
  [9758] = {.lex_state = 154, .external_lex_state = 69},
  [9759] = {.lex_state = 154, .external_lex_state = 69},
  [9760] = {.lex_state = 154, .external_lex_state = 58},
  [9761] = {.lex_state = 172, .external_lex_state = 55},
  [9762] = {.lex_state = 160, .external_lex_state = 74},
  [9763] = {.lex_state = 154, .external_lex_state = 56},
  [9764] = {.lex_state = 154, .external_lex_state = 56},
  [9765] = {.lex_state = 154, .external_lex_state = 56},
  [9766] = {.lex_state = 165, .external_lex_state = 55},
  [9767] = {.lex_state = 165, .external_lex_state = 55},
  [9768] = {.lex_state = 203, .external_lex_state = 79},
  [9769] = {.lex_state = 154, .external_lex_state = 69},
  [9770] = {.lex_state = 154, .external_lex_state = 69},
  [9771] = {.lex_state = 154, .external_lex_state = 69},
  [9772] = {.lex_state = 154, .external_lex_state = 69},
  [9773] = {.lex_state = 203, .external_lex_state = 79},
  [9774] = {.lex_state = 154, .external_lex_state = 69},
  [9775] = {.lex_state = 154, .external_lex_state = 56},
  [9776] = {.lex_state = 154, .external_lex_state = 55},
  [9777] = {.lex_state = 176, .external_lex_state = 68},
  [9778] = {.lex_state = 154, .external_lex_state = 69},
  [9779] = {.lex_state = 172, .external_lex_state = 55},
  [9780] = {.lex_state = 154, .external_lex_state = 70},
  [9781] = {.lex_state = 154, .external_lex_state = 56},
  [9782] = {.lex_state = 154, .external_lex_state = 56},
  [9783] = {.lex_state = 154, .external_lex_state = 69},
  [9784] = {.lex_state = 203, .external_lex_state = 79},
  [9785] = {.lex_state = 154, .external_lex_state = 69},
  [9786] = {.lex_state = 165, .external_lex_state = 55},
  [9787] = {.lex_state = 154, .external_lex_state = 70},
  [9788] = {.lex_state = 154, .external_lex_state = 70},
  [9789] = {.lex_state = 203, .external_lex_state = 79},
  [9790] = {.lex_state = 154, .external_lex_state = 70},
  [9791] = {.lex_state = 154, .external_lex_state = 70},
  [9792] = {.lex_state = 154, .external_lex_state = 56},
  [9793] = {.lex_state = 154, .external_lex_state = 58},
  [9794] = {.lex_state = 154, .external_lex_state = 70},
  [9795] = {.lex_state = 154, .external_lex_state = 70},
  [9796] = {.lex_state = 165, .external_lex_state = 55},
  [9797] = {.lex_state = 160, .external_lex_state = 73},
  [9798] = {.lex_state = 160, .external_lex_state = 75},
  [9799] = {.lex_state = 154, .external_lex_state = 70},
  [9800] = {.lex_state = 154, .external_lex_state = 56},
  [9801] = {.lex_state = 154, .external_lex_state = 70},
  [9802] = {.lex_state = 154, .external_lex_state = 70},
  [9803] = {.lex_state = 154, .external_lex_state = 70},
  [9804] = {.lex_state = 154, .external_lex_state = 70},
  [9805] = {.lex_state = 154, .external_lex_state = 70},
  [9806] = {.lex_state = 154, .external_lex_state = 70},
  [9807] = {.lex_state = 154, .external_lex_state = 70},
  [9808] = {.lex_state = 154, .external_lex_state = 70},
  [9809] = {.lex_state = 154, .external_lex_state = 70},
  [9810] = {.lex_state = 154, .external_lex_state = 70},
  [9811] = {.lex_state = 154, .external_lex_state = 70},
  [9812] = {.lex_state = 154, .external_lex_state = 70},
  [9813] = {.lex_state = 154, .external_lex_state = 69},
  [9814] = {.lex_state = 154, .external_lex_state = 57},
  [9815] = {.lex_state = 154, .external_lex_state = 70},
  [9816] = {.lex_state = 154, .external_lex_state = 70},
  [9817] = {.lex_state = 154, .external_lex_state = 70},
  [9818] = {.lex_state = 203, .external_lex_state = 79},
  [9819] = {.lex_state = 203, .external_lex_state = 79},
  [9820] = {.lex_state = 165, .external_lex_state = 55},
  [9821] = {.lex_state = 154, .external_lex_state = 69},
  [9822] = {.lex_state = 154, .external_lex_state = 70},
  [9823] = {.lex_state = 154, .external_lex_state = 56},
  [9824] = {.lex_state = 154, .external_lex_state = 70},
  [9825] = {.lex_state = 154, .external_lex_state = 56},
  [9826] = {.lex_state = 154, .external_lex_state = 70},
  [9827] = {.lex_state = 154, .external_lex_state = 70},
  [9828] = {.lex_state = 203, .external_lex_state = 79},
  [9829] = {.lex_state = 154, .external_lex_state = 70},
  [9830] = {.lex_state = 176, .external_lex_state = 68},
  [9831] = {.lex_state = 172, .external_lex_state = 55},
  [9832] = {.lex_state = 154, .external_lex_state = 57},
  [9833] = {.lex_state = 154, .external_lex_state = 55},
  [9834] = {.lex_state = 154, .external_lex_state = 57},
  [9835] = {.lex_state = 165, .external_lex_state = 55},
  [9836] = {.lex_state = 154, .external_lex_state = 56},
  [9837] = {.lex_state = 154, .external_lex_state = 56},
  [9838] = {.lex_state = 154, .external_lex_state = 57},
  [9839] = {.lex_state = 165, .external_lex_state = 55},
  [9840] = {.lex_state = 160, .external_lex_state = 75},
  [9841] = {.lex_state = 154, .external_lex_state = 56},
  [9842] = {.lex_state = 165, .external_lex_state = 55},
  [9843] = {.lex_state = 165, .external_lex_state = 55},
  [9844] = {.lex_state = 154, .external_lex_state = 70},
  [9845] = {.lex_state = 154, .external_lex_state = 70},
  [9846] = {.lex_state = 154, .external_lex_state = 70},
  [9847] = {.lex_state = 154, .external_lex_state = 70},
  [9848] = {.lex_state = 165, .external_lex_state = 55},
  [9849] = {.lex_state = 154, .external_lex_state = 55},
  [9850] = {.lex_state = 171, .external_lex_state = 74},
  [9851] = {.lex_state = 171, .external_lex_state = 74},
  [9852] = {.lex_state = 171, .external_lex_state = 74},
  [9853] = {.lex_state = 165, .external_lex_state = 55},
  [9854] = {.lex_state = 165, .external_lex_state = 55},
  [9855] = {.lex_state = 154, .external_lex_state = 69},
  [9856] = {.lex_state = 154, .external_lex_state = 55},
  [9857] = {.lex_state = 154, .external_lex_state = 69},
  [9858] = {.lex_state = 154, .external_lex_state = 69},
  [9859] = {.lex_state = 154, .external_lex_state = 55},
  [9860] = {.lex_state = 203, .external_lex_state = 79},
  [9861] = {.lex_state = 154, .external_lex_state = 69},
  [9862] = {.lex_state = 154, .external_lex_state = 69},
  [9863] = {.lex_state = 165, .external_lex_state = 55},
  [9864] = {.lex_state = 203, .external_lex_state = 79},
  [9865] = {.lex_state = 154, .external_lex_state = 69},
  [9866] = {.lex_state = 154, .external_lex_state = 57},
  [9867] = {.lex_state = 154, .external_lex_state = 69},
  [9868] = {.lex_state = 165, .external_lex_state = 55},
  [9869] = {.lex_state = 172, .external_lex_state = 55},
  [9870] = {.lex_state = 172, .external_lex_state = 55},
  [9871] = {.lex_state = 172, .external_lex_state = 55},
  [9872] = {.lex_state = 172, .external_lex_state = 55},
  [9873] = {.lex_state = 172, .external_lex_state = 55},
  [9874] = {.lex_state = 172, .external_lex_state = 55},
  [9875] = {.lex_state = 172, .external_lex_state = 55},
  [9876] = {.lex_state = 172, .external_lex_state = 55},
  [9877] = {.lex_state = 172, .external_lex_state = 55},
  [9878] = {.lex_state = 172, .external_lex_state = 55},
  [9879] = {.lex_state = 172, .external_lex_state = 55},
  [9880] = {.lex_state = 172, .external_lex_state = 55},
  [9881] = {.lex_state = 172, .external_lex_state = 55},
  [9882] = {.lex_state = 172, .external_lex_state = 55},
  [9883] = {.lex_state = 172, .external_lex_state = 55},
  [9884] = {.lex_state = 172, .external_lex_state = 55},
  [9885] = {.lex_state = 172, .external_lex_state = 55},
  [9886] = {.lex_state = 172, .external_lex_state = 55},
  [9887] = {.lex_state = 172, .external_lex_state = 55},
  [9888] = {.lex_state = 172, .external_lex_state = 55},
  [9889] = {.lex_state = 172, .external_lex_state = 55},
  [9890] = {.lex_state = 172, .external_lex_state = 55},
  [9891] = {.lex_state = 172, .external_lex_state = 55},
  [9892] = {.lex_state = 172, .external_lex_state = 55},
  [9893] = {.lex_state = 154, .external_lex_state = 70},
  [9894] = {.lex_state = 203, .external_lex_state = 65},
  [9895] = {.lex_state = 223},
  [9896] = {.lex_state = 223},
  [9897] = {.lex_state = 223},
  [9898] = {.lex_state = 154, .external_lex_state = 57},
  [9899] = {.lex_state = 223},
  [9900] = {.lex_state = 154, .external_lex_state = 57},
  [9901] = {.lex_state = 154, .external_lex_state = 57},
  [9902] = {.lex_state = 223},
  [9903] = {.lex_state = 154, .external_lex_state = 56},
  [9904] = {.lex_state = 223},
  [9905] = {.lex_state = 154, .external_lex_state = 71},
  [9906] = {.lex_state = 154, .external_lex_state = 71},
  [9907] = {.lex_state = 223},
  [9908] = {.lex_state = 154, .external_lex_state = 57},
  [9909] = {.lex_state = 154, .external_lex_state = 56},
  [9910] = {.lex_state = 223},
  [9911] = {.lex_state = 154, .external_lex_state = 56},
  [9912] = {.lex_state = 154, .external_lex_state = 56},
  [9913] = {.lex_state = 223},
  [9914] = {.lex_state = 154, .external_lex_state = 55},
  [9915] = {.lex_state = 154, .external_lex_state = 55},
  [9916] = {.lex_state = 154, .external_lex_state = 55},
  [9917] = {.lex_state = 154, .external_lex_state = 55},
  [9918] = {.lex_state = 154, .external_lex_state = 55},
  [9919] = {.lex_state = 154, .external_lex_state = 55},
  [9920] = {.lex_state = 154, .external_lex_state = 56},
  [9921] = {.lex_state = 154, .external_lex_state = 71},
  [9922] = {.lex_state = 154, .external_lex_state = 71},
  [9923] = {.lex_state = 154, .external_lex_state = 57},
  [9924] = {.lex_state = 172, .external_lex_state = 55},
  [9925] = {.lex_state = 223},
  [9926] = {.lex_state = 223},
  [9927] = {.lex_state = 172, .external_lex_state = 55},
  [9928] = {.lex_state = 154, .external_lex_state = 68},
  [9929] = {.lex_state = 178, .external_lex_state = 74},
  [9930] = {.lex_state = 154, .external_lex_state = 68},
  [9931] = {.lex_state = 154, .external_lex_state = 57},
  [9932] = {.lex_state = 154, .external_lex_state = 68},
  [9933] = {.lex_state = 154, .external_lex_state = 68},
  [9934] = {.lex_state = 154, .external_lex_state = 68},
  [9935] = {.lex_state = 154, .external_lex_state = 68},
  [9936] = {.lex_state = 154, .external_lex_state = 68},
  [9937] = {.lex_state = 154, .external_lex_state = 68},
  [9938] = {.lex_state = 154, .external_lex_state = 68},
  [9939] = {.lex_state = 154, .external_lex_state = 68},
  [9940] = {.lex_state = 154, .external_lex_state = 68},
  [9941] = {.lex_state = 154, .external_lex_state = 68},
  [9942] = {.lex_state = 154, .external_lex_state = 68},
  [9943] = {.lex_state = 154, .external_lex_state = 68},
  [9944] = {.lex_state = 154, .external_lex_state = 68},
  [9945] = {.lex_state = 154, .external_lex_state = 68},
  [9946] = {.lex_state = 154, .external_lex_state = 68},
  [9947] = {.lex_state = 154, .external_lex_state = 68},
  [9948] = {.lex_state = 154, .external_lex_state = 68},
  [9949] = {.lex_state = 154, .external_lex_state = 68},
  [9950] = {.lex_state = 154, .external_lex_state = 68},
  [9951] = {.lex_state = 154, .external_lex_state = 68},
  [9952] = {.lex_state = 154, .external_lex_state = 58},
  [9953] = {.lex_state = 154, .external_lex_state = 68},
  [9954] = {.lex_state = 154, .external_lex_state = 56},
  [9955] = {.lex_state = 154, .external_lex_state = 56},
  [9956] = {.lex_state = 154, .external_lex_state = 56},
  [9957] = {.lex_state = 154, .external_lex_state = 56},
  [9958] = {.lex_state = 154, .external_lex_state = 68},
  [9959] = {.lex_state = 203, .external_lex_state = 65},
  [9960] = {.lex_state = 154, .external_lex_state = 68},
  [9961] = {.lex_state = 154, .external_lex_state = 68},
  [9962] = {.lex_state = 154, .external_lex_state = 68},
  [9963] = {.lex_state = 160, .external_lex_state = 75},
  [9964] = {.lex_state = 154, .external_lex_state = 57},
  [9965] = {.lex_state = 154, .external_lex_state = 57},
  [9966] = {.lex_state = 154, .external_lex_state = 56},
  [9967] = {.lex_state = 154, .external_lex_state = 57},
  [9968] = {.lex_state = 154, .external_lex_state = 57},
  [9969] = {.lex_state = 203, .external_lex_state = 65},
  [9970] = {.lex_state = 223},
  [9971] = {.lex_state = 223},
  [9972] = {.lex_state = 154, .external_lex_state = 57},
  [9973] = {.lex_state = 154, .external_lex_state = 71},
  [9974] = {.lex_state = 154, .external_lex_state = 68},
  [9975] = {.lex_state = 154, .external_lex_state = 71},
  [9976] = {.lex_state = 203, .external_lex_state = 65},
  [9977] = {.lex_state = 154, .external_lex_state = 58},
  [9978] = {.lex_state = 160, .external_lex_state = 75},
  [9979] = {.lex_state = 154, .external_lex_state = 57},
  [9980] = {.lex_state = 154, .external_lex_state = 71},
  [9981] = {.lex_state = 154, .external_lex_state = 71},
  [9982] = {.lex_state = 203, .external_lex_state = 65},
  [9983] = {.lex_state = 154, .external_lex_state = 71},
  [9984] = {.lex_state = 154, .external_lex_state = 57},
  [9985] = {.lex_state = 223},
  [9986] = {.lex_state = 154, .external_lex_state = 57},
  [9987] = {.lex_state = 154, .external_lex_state = 58},
  [9988] = {.lex_state = 154, .external_lex_state = 57},
  [9989] = {.lex_state = 223},
  [9990] = {.lex_state = 154, .external_lex_state = 56},
  [9991] = {.lex_state = 203, .external_lex_state = 65},
  [9992] = {.lex_state = 154, .external_lex_state = 71},
  [9993] = {.lex_state = 154, .external_lex_state = 71},
  [9994] = {.lex_state = 154, .external_lex_state = 56},
  [9995] = {.lex_state = 154, .external_lex_state = 56},
  [9996] = {.lex_state = 154, .external_lex_state = 57},
  [9997] = {.lex_state = 160, .external_lex_state = 74},
  [9998] = {.lex_state = 154, .external_lex_state = 57},
  [9999] = {.lex_state = 154, .external_lex_state = 71},
  [10000] = {.lex_state = 160, .external_lex_state = 74},
  [10001] = {.lex_state = 154, .external_lex_state = 71},
  [10002] = {.lex_state = 154, .external_lex_state = 71},
  [10003] = {.lex_state = 154, .external_lex_state = 57},
  [10004] = {.lex_state = 154, .external_lex_state = 57},
  [10005] = {.lex_state = 154, .external_lex_state = 71},
  [10006] = {.lex_state = 154, .external_lex_state = 71},
  [10007] = {.lex_state = 154, .external_lex_state = 71},
  [10008] = {.lex_state = 154, .external_lex_state = 71},
  [10009] = {.lex_state = 154, .external_lex_state = 71},
  [10010] = {.lex_state = 154, .external_lex_state = 71},
  [10011] = {.lex_state = 154, .external_lex_state = 57},
  [10012] = {.lex_state = 154, .external_lex_state = 71},
  [10013] = {.lex_state = 154, .external_lex_state = 71},
  [10014] = {.lex_state = 154, .external_lex_state = 71},
  [10015] = {.lex_state = 154, .external_lex_state = 71},
  [10016] = {.lex_state = 203, .external_lex_state = 65},
  [10017] = {.lex_state = 223},
  [10018] = {.lex_state = 223},
  [10019] = {.lex_state = 223},
  [10020] = {.lex_state = 154, .external_lex_state = 71},
  [10021] = {.lex_state = 154, .external_lex_state = 57},
  [10022] = {.lex_state = 154, .external_lex_state = 57},
  [10023] = {.lex_state = 154, .external_lex_state = 56},
  [10024] = {.lex_state = 160, .external_lex_state = 74},
  [10025] = {.lex_state = 154, .external_lex_state = 68},
  [10026] = {.lex_state = 154, .external_lex_state = 71},
  [10027] = {.lex_state = 154, .external_lex_state = 56},
  [10028] = {.lex_state = 154, .external_lex_state = 56},
  [10029] = {.lex_state = 223},
  [10030] = {.lex_state = 203, .external_lex_state = 65},
  [10031] = {.lex_state = 154, .external_lex_state = 56},
  [10032] = {.lex_state = 160, .external_lex_state = 74},
  [10033] = {.lex_state = 223},
  [10034] = {.lex_state = 223},
  [10035] = {.lex_state = 223},
  [10036] = {.lex_state = 172, .external_lex_state = 55},
  [10037] = {.lex_state = 154, .external_lex_state = 56},
  [10038] = {.lex_state = 154, .external_lex_state = 71},
  [10039] = {.lex_state = 223},
  [10040] = {.lex_state = 154, .external_lex_state = 55},
  [10041] = {.lex_state = 223},
  [10042] = {.lex_state = 154, .external_lex_state = 56},
  [10043] = {.lex_state = 154, .external_lex_state = 71},
  [10044] = {.lex_state = 160, .external_lex_state = 74},
  [10045] = {.lex_state = 154, .external_lex_state = 71},
  [10046] = {.lex_state = 154, .external_lex_state = 71},
  [10047] = {.lex_state = 160, .external_lex_state = 75},
  [10048] = {.lex_state = 223},
  [10049] = {.lex_state = 160, .external_lex_state = 75},
  [10050] = {.lex_state = 223},
  [10051] = {.lex_state = 154, .external_lex_state = 57},
  [10052] = {.lex_state = 160, .external_lex_state = 74},
  [10053] = {.lex_state = 154, .external_lex_state = 71},
  [10054] = {.lex_state = 154, .external_lex_state = 56},
  [10055] = {.lex_state = 160, .external_lex_state = 74},
  [10056] = {.lex_state = 160, .external_lex_state = 75},
  [10057] = {.lex_state = 154, .external_lex_state = 57},
  [10058] = {.lex_state = 223},
  [10059] = {.lex_state = 154, .external_lex_state = 56},
  [10060] = {.lex_state = 154, .external_lex_state = 57},
  [10061] = {.lex_state = 160, .external_lex_state = 75},
  [10062] = {.lex_state = 154, .external_lex_state = 71},
  [10063] = {.lex_state = 223},
  [10064] = {.lex_state = 223},
  [10065] = {.lex_state = 154, .external_lex_state = 56},
  [10066] = {.lex_state = 154, .external_lex_state = 56},
  [10067] = {.lex_state = 223},
  [10068] = {.lex_state = 154, .external_lex_state = 56},
  [10069] = {.lex_state = 223},
  [10070] = {.lex_state = 154, .external_lex_state = 56},
  [10071] = {.lex_state = 154, .external_lex_state = 68},
  [10072] = {.lex_state = 154, .external_lex_state = 68},
  [10073] = {.lex_state = 154, .external_lex_state = 56},
  [10074] = {.lex_state = 178, .external_lex_state = 74},
  [10075] = {.lex_state = 223},
  [10076] = {.lex_state = 223},
  [10077] = {.lex_state = 154, .external_lex_state = 58},
  [10078] = {.lex_state = 223},
  [10079] = {.lex_state = 154, .external_lex_state = 58},
  [10080] = {.lex_state = 154, .external_lex_state = 58},
  [10081] = {.lex_state = 223},
  [10082] = {.lex_state = 154, .external_lex_state = 57},
  [10083] = {.lex_state = 172, .external_lex_state = 55},
  [10084] = {.lex_state = 154, .external_lex_state = 57},
  [10085] = {.lex_state = 172, .external_lex_state = 55},
  [10086] = {.lex_state = 154, .external_lex_state = 68},
  [10087] = {.lex_state = 154, .external_lex_state = 68},
  [10088] = {.lex_state = 154, .external_lex_state = 55},
  [10089] = {.lex_state = 154, .external_lex_state = 57},
  [10090] = {.lex_state = 154, .external_lex_state = 58},
  [10091] = {.lex_state = 154, .external_lex_state = 58},
  [10092] = {.lex_state = 154, .external_lex_state = 58},
  [10093] = {.lex_state = 154, .external_lex_state = 58},
  [10094] = {.lex_state = 154, .external_lex_state = 58},
  [10095] = {.lex_state = 160, .external_lex_state = 75},
  [10096] = {.lex_state = 154, .external_lex_state = 58},
  [10097] = {.lex_state = 154, .external_lex_state = 55},
  [10098] = {.lex_state = 154, .external_lex_state = 56},
  [10099] = {.lex_state = 154, .external_lex_state = 56},
  [10100] = {.lex_state = 223},
  [10101] = {.lex_state = 223},
  [10102] = {.lex_state = 154, .external_lex_state = 56},
  [10103] = {.lex_state = 154, .external_lex_state = 56},
  [10104] = {.lex_state = 154, .external_lex_state = 55},
  [10105] = {.lex_state = 223},
  [10106] = {.lex_state = 154, .external_lex_state = 58},
  [10107] = {.lex_state = 154, .external_lex_state = 55},
  [10108] = {.lex_state = 154, .external_lex_state = 55},
  [10109] = {.lex_state = 154, .external_lex_state = 55},
  [10110] = {.lex_state = 223},
  [10111] = {.lex_state = 154, .external_lex_state = 56},
  [10112] = {.lex_state = 154, .external_lex_state = 57},
  [10113] = {.lex_state = 154, .external_lex_state = 57},
  [10114] = {.lex_state = 154, .external_lex_state = 57},
  [10115] = {.lex_state = 154, .external_lex_state = 57},
  [10116] = {.lex_state = 154, .external_lex_state = 56},
  [10117] = {.lex_state = 178, .external_lex_state = 74},
  [10118] = {.lex_state = 223},
  [10119] = {.lex_state = 154, .external_lex_state = 71},
  [10120] = {.lex_state = 172, .external_lex_state = 68},
  [10121] = {.lex_state = 154, .external_lex_state = 58},
  [10122] = {.lex_state = 154, .external_lex_state = 58},
  [10123] = {.lex_state = 172, .external_lex_state = 68},
  [10124] = {.lex_state = 154, .external_lex_state = 55},
  [10125] = {.lex_state = 154, .external_lex_state = 58},
  [10126] = {.lex_state = 172, .external_lex_state = 68},
  [10127] = {.lex_state = 172, .external_lex_state = 68},
  [10128] = {.lex_state = 172, .external_lex_state = 68},
  [10129] = {.lex_state = 172, .external_lex_state = 68},
  [10130] = {.lex_state = 141, .external_lex_state = 80},
  [10131] = {.lex_state = 154, .external_lex_state = 58},
  [10132] = {.lex_state = 154, .external_lex_state = 55},
  [10133] = {.lex_state = 154, .external_lex_state = 55},
  [10134] = {.lex_state = 172, .external_lex_state = 68},
  [10135] = {.lex_state = 141, .external_lex_state = 81},
  [10136] = {.lex_state = 154, .external_lex_state = 58},
  [10137] = {.lex_state = 154, .external_lex_state = 55},
  [10138] = {.lex_state = 172, .external_lex_state = 68},
  [10139] = {.lex_state = 178, .external_lex_state = 74},
  [10140] = {.lex_state = 178, .external_lex_state = 74},
  [10141] = {.lex_state = 203, .external_lex_state = 76},
  [10142] = {.lex_state = 154, .external_lex_state = 55},
  [10143] = {.lex_state = 154, .external_lex_state = 55},
  [10144] = {.lex_state = 172, .external_lex_state = 68},
  [10145] = {.lex_state = 172, .external_lex_state = 68},
  [10146] = {.lex_state = 172, .external_lex_state = 55},
  [10147] = {.lex_state = 154, .external_lex_state = 55},
  [10148] = {.lex_state = 154, .external_lex_state = 55},
  [10149] = {.lex_state = 172, .external_lex_state = 68},
  [10150] = {.lex_state = 172, .external_lex_state = 55},
  [10151] = {.lex_state = 154, .external_lex_state = 58},
  [10152] = {.lex_state = 154, .external_lex_state = 58},
  [10153] = {.lex_state = 154, .external_lex_state = 55},
  [10154] = {.lex_state = 154, .external_lex_state = 58},
  [10155] = {.lex_state = 154, .external_lex_state = 58},
  [10156] = {.lex_state = 203, .external_lex_state = 65},
  [10157] = {.lex_state = 154, .external_lex_state = 58},
  [10158] = {.lex_state = 172, .external_lex_state = 68},
  [10159] = {.lex_state = 154, .external_lex_state = 58},
  [10160] = {.lex_state = 172, .external_lex_state = 55},
  [10161] = {.lex_state = 172, .external_lex_state = 68},
  [10162] = {.lex_state = 154, .external_lex_state = 55},
  [10163] = {.lex_state = 141, .external_lex_state = 81},
  [10164] = {.lex_state = 172, .external_lex_state = 68},
  [10165] = {.lex_state = 172, .external_lex_state = 68},
  [10166] = {.lex_state = 172, .external_lex_state = 68},
  [10167] = {.lex_state = 203, .external_lex_state = 65},
  [10168] = {.lex_state = 154, .external_lex_state = 58},
  [10169] = {.lex_state = 154, .external_lex_state = 58},
  [10170] = {.lex_state = 203, .external_lex_state = 65},
  [10171] = {.lex_state = 141, .external_lex_state = 80},
  [10172] = {.lex_state = 172, .external_lex_state = 68},
  [10173] = {.lex_state = 154, .external_lex_state = 55},
  [10174] = {.lex_state = 172, .external_lex_state = 68},
  [10175] = {.lex_state = 172, .external_lex_state = 68},
  [10176] = {.lex_state = 172, .external_lex_state = 55},
  [10177] = {.lex_state = 172, .external_lex_state = 55},
  [10178] = {.lex_state = 172, .external_lex_state = 55},
  [10179] = {.lex_state = 172, .external_lex_state = 55},
  [10180] = {.lex_state = 154, .external_lex_state = 55},
  [10181] = {.lex_state = 172, .external_lex_state = 55},
  [10182] = {.lex_state = 172, .external_lex_state = 55},
  [10183] = {.lex_state = 154, .external_lex_state = 58},
  [10184] = {.lex_state = 154, .external_lex_state = 55},
  [10185] = {.lex_state = 172, .external_lex_state = 68},
  [10186] = {.lex_state = 154, .external_lex_state = 55},
  [10187] = {.lex_state = 154, .external_lex_state = 55},
  [10188] = {.lex_state = 154, .external_lex_state = 58},
  [10189] = {.lex_state = 178, .external_lex_state = 74},
  [10190] = {.lex_state = 172, .external_lex_state = 55},
  [10191] = {.lex_state = 172, .external_lex_state = 68},
  [10192] = {.lex_state = 154, .external_lex_state = 55},
  [10193] = {.lex_state = 154, .external_lex_state = 55},
  [10194] = {.lex_state = 154, .external_lex_state = 55},
  [10195] = {.lex_state = 172, .external_lex_state = 68},
  [10196] = {.lex_state = 154, .external_lex_state = 55},
  [10197] = {.lex_state = 172, .external_lex_state = 68},
  [10198] = {.lex_state = 154, .external_lex_state = 55},
  [10199] = {.lex_state = 178, .external_lex_state = 74},
  [10200] = {.lex_state = 154, .external_lex_state = 58},
  [10201] = {.lex_state = 154, .external_lex_state = 55},
  [10202] = {.lex_state = 154, .external_lex_state = 55},
  [10203] = {.lex_state = 154, .external_lex_state = 58},
  [10204] = {.lex_state = 154, .external_lex_state = 58},
  [10205] = {.lex_state = 154, .external_lex_state = 58},
  [10206] = {.lex_state = 154, .external_lex_state = 58},
  [10207] = {.lex_state = 178, .external_lex_state = 74},
  [10208] = {.lex_state = 178, .external_lex_state = 74},
  [10209] = {.lex_state = 172, .external_lex_state = 55},
  [10210] = {.lex_state = 154, .external_lex_state = 58},
  [10211] = {.lex_state = 154, .external_lex_state = 58},
  [10212] = {.lex_state = 154, .external_lex_state = 58},
  [10213] = {.lex_state = 172, .external_lex_state = 68},
  [10214] = {.lex_state = 172, .external_lex_state = 68},
  [10215] = {.lex_state = 154, .external_lex_state = 58},
  [10216] = {.lex_state = 172, .external_lex_state = 68},
  [10217] = {.lex_state = 154, .external_lex_state = 55},
  [10218] = {.lex_state = 172, .external_lex_state = 68},
  [10219] = {.lex_state = 154, .external_lex_state = 55},
  [10220] = {.lex_state = 154, .external_lex_state = 55},
  [10221] = {.lex_state = 172, .external_lex_state = 55},
  [10222] = {.lex_state = 172, .external_lex_state = 68},
  [10223] = {.lex_state = 154, .external_lex_state = 55},
  [10224] = {.lex_state = 154, .external_lex_state = 58},
  [10225] = {.lex_state = 154, .external_lex_state = 58},
  [10226] = {.lex_state = 154, .external_lex_state = 58},
  [10227] = {.lex_state = 154, .external_lex_state = 58},
  [10228] = {.lex_state = 203, .external_lex_state = 76},
  [10229] = {.lex_state = 172, .external_lex_state = 68},
  [10230] = {.lex_state = 154, .external_lex_state = 55},
  [10231] = {.lex_state = 172, .external_lex_state = 68},
  [10232] = {.lex_state = 154, .external_lex_state = 58},
  [10233] = {.lex_state = 172, .external_lex_state = 68},
  [10234] = {.lex_state = 154, .external_lex_state = 58},
  [10235] = {.lex_state = 154, .external_lex_state = 55},
  [10236] = {.lex_state = 172, .external_lex_state = 68},
  [10237] = {.lex_state = 172, .external_lex_state = 55},
  [10238] = {.lex_state = 154, .external_lex_state = 58},
  [10239] = {.lex_state = 154, .external_lex_state = 55},
  [10240] = {.lex_state = 178, .external_lex_state = 74},
  [10241] = {.lex_state = 154, .external_lex_state = 55},
  [10242] = {.lex_state = 154, .external_lex_state = 55},
  [10243] = {.lex_state = 154, .external_lex_state = 55},
  [10244] = {.lex_state = 154, .external_lex_state = 55},
  [10245] = {.lex_state = 154, .external_lex_state = 58},
  [10246] = {.lex_state = 172, .external_lex_state = 68},
  [10247] = {.lex_state = 223},
  [10248] = {.lex_state = 180, .external_lex_state = 55},
  [10249] = {.lex_state = 180, .external_lex_state = 55},
  [10250] = {.lex_state = 172, .external_lex_state = 55},
  [10251] = {.lex_state = 172, .external_lex_state = 55},
  [10252] = {.lex_state = 172, .external_lex_state = 55},
  [10253] = {.lex_state = 172, .external_lex_state = 55},
  [10254] = {.lex_state = 180, .external_lex_state = 55},
  [10255] = {.lex_state = 223},
  [10256] = {.lex_state = 172, .external_lex_state = 55},
  [10257] = {.lex_state = 172, .external_lex_state = 55},
  [10258] = {.lex_state = 172, .external_lex_state = 55},
  [10259] = {.lex_state = 172, .external_lex_state = 55},
  [10260] = {.lex_state = 172, .external_lex_state = 55},
  [10261] = {.lex_state = 172, .external_lex_state = 55},
  [10262] = {.lex_state = 172, .external_lex_state = 55},
  [10263] = {.lex_state = 223},
  [10264] = {.lex_state = 223},
  [10265] = {.lex_state = 172, .external_lex_state = 55},
  [10266] = {.lex_state = 172, .external_lex_state = 55},
  [10267] = {.lex_state = 141, .external_lex_state = 82},
  [10268] = {.lex_state = 172, .external_lex_state = 55},
  [10269] = {.lex_state = 223},
  [10270] = {.lex_state = 172, .external_lex_state = 55},
  [10271] = {.lex_state = 172, .external_lex_state = 55},
  [10272] = {.lex_state = 172, .external_lex_state = 55},
  [10273] = {.lex_state = 172, .external_lex_state = 55},
  [10274] = {.lex_state = 223},
  [10275] = {.lex_state = 180, .external_lex_state = 55},
  [10276] = {.lex_state = 180, .external_lex_state = 55},
  [10277] = {.lex_state = 223},
  [10278] = {.lex_state = 223},
  [10279] = {.lex_state = 172, .external_lex_state = 55},
  [10280] = {.lex_state = 172, .external_lex_state = 55},
  [10281] = {.lex_state = 172, .external_lex_state = 55},
  [10282] = {.lex_state = 180, .external_lex_state = 55},
  [10283] = {.lex_state = 223},
  [10284] = {.lex_state = 223},
  [10285] = {.lex_state = 223},
  [10286] = {.lex_state = 172, .external_lex_state = 55},
  [10287] = {.lex_state = 141, .external_lex_state = 82},
  [10288] = {.lex_state = 223},
  [10289] = {.lex_state = 172, .external_lex_state = 55},
  [10290] = {.lex_state = 223},
  [10291] = {.lex_state = 172, .external_lex_state = 55},
  [10292] = {.lex_state = 223},
  [10293] = {.lex_state = 180, .external_lex_state = 55},
  [10294] = {.lex_state = 172, .external_lex_state = 55},
  [10295] = {.lex_state = 223},
  [10296] = {.lex_state = 180, .external_lex_state = 55},
  [10297] = {.lex_state = 223},
  [10298] = {.lex_state = 172, .external_lex_state = 55},
  [10299] = {.lex_state = 223},
  [10300] = {.lex_state = 172, .external_lex_state = 55},
  [10301] = {.lex_state = 172, .external_lex_state = 55},
  [10302] = {.lex_state = 223},
  [10303] = {.lex_state = 203, .external_lex_state = 78},
  [10304] = {.lex_state = 223},
  [10305] = {.lex_state = 203, .external_lex_state = 78},
  [10306] = {.lex_state = 172, .external_lex_state = 55},
  [10307] = {.lex_state = 172, .external_lex_state = 55},
  [10308] = {.lex_state = 172, .external_lex_state = 55},
  [10309] = {.lex_state = 223},
  [10310] = {.lex_state = 172, .external_lex_state = 55},
  [10311] = {.lex_state = 203, .external_lex_state = 79},
  [10312] = {.lex_state = 203, .external_lex_state = 79},
  [10313] = {.lex_state = 203, .external_lex_state = 79},
  [10314] = {.lex_state = 181, .external_lex_state = 55},
  [10315] = {.lex_state = 223},
  [10316] = {.lex_state = 203, .external_lex_state = 79},
  [10317] = {.lex_state = 181, .external_lex_state = 55},
  [10318] = {.lex_state = 203, .external_lex_state = 79},
  [10319] = {.lex_state = 203, .external_lex_state = 79},
  [10320] = {.lex_state = 181, .external_lex_state = 55},
  [10321] = {.lex_state = 181, .external_lex_state = 55},
  [10322] = {.lex_state = 181, .external_lex_state = 55},
  [10323] = {.lex_state = 203, .external_lex_state = 79},
  [10324] = {.lex_state = 203, .external_lex_state = 79},
  [10325] = {.lex_state = 203, .external_lex_state = 79},
  [10326] = {.lex_state = 180, .external_lex_state = 55},
  [10327] = {.lex_state = 203, .external_lex_state = 79},
  [10328] = {.lex_state = 203, .external_lex_state = 79},
  [10329] = {.lex_state = 180, .external_lex_state = 55},
  [10330] = {.lex_state = 203, .external_lex_state = 79},
  [10331] = {.lex_state = 203, .external_lex_state = 79},
  [10332] = {.lex_state = 203, .external_lex_state = 79},
  [10333] = {.lex_state = 203, .external_lex_state = 79},
  [10334] = {.lex_state = 180, .external_lex_state = 55},
  [10335] = {.lex_state = 203, .external_lex_state = 79},
  [10336] = {.lex_state = 141, .external_lex_state = 83},
  [10337] = {.lex_state = 203, .external_lex_state = 79},
  [10338] = {.lex_state = 181, .external_lex_state = 55},
  [10339] = {.lex_state = 203, .external_lex_state = 79},
  [10340] = {.lex_state = 203, .external_lex_state = 79},
  [10341] = {.lex_state = 181, .external_lex_state = 55},
  [10342] = {.lex_state = 203, .external_lex_state = 79},
  [10343] = {.lex_state = 203, .external_lex_state = 79},
  [10344] = {.lex_state = 203, .external_lex_state = 79},
  [10345] = {.lex_state = 141, .external_lex_state = 83},
  [10346] = {.lex_state = 203, .external_lex_state = 79},
  [10347] = {.lex_state = 203, .external_lex_state = 79},
  [10348] = {.lex_state = 203, .external_lex_state = 79},
  [10349] = {.lex_state = 203, .external_lex_state = 79},
  [10350] = {.lex_state = 203, .external_lex_state = 79},
  [10351] = {.lex_state = 203, .external_lex_state = 79},
  [10352] = {.lex_state = 203, .external_lex_state = 79},
  [10353] = {.lex_state = 180, .external_lex_state = 55},
  [10354] = {.lex_state = 203, .external_lex_state = 79},
  [10355] = {.lex_state = 141, .external_lex_state = 84},
  [10356] = {.lex_state = 141, .external_lex_state = 84},
  [10357] = {.lex_state = 203, .external_lex_state = 79},
  [10358] = {.lex_state = 181, .external_lex_state = 55},
  [10359] = {.lex_state = 203, .external_lex_state = 79},
  [10360] = {.lex_state = 203, .external_lex_state = 79},
  [10361] = {.lex_state = 180, .external_lex_state = 55},
  [10362] = {.lex_state = 179, .external_lex_state = 59},
  [10363] = {.lex_state = 180, .external_lex_state = 55},
  [10364] = {.lex_state = 181, .external_lex_state = 55},
  [10365] = {.lex_state = 141, .external_lex_state = 85},
  [10366] = {.lex_state = 180, .external_lex_state = 55},
  [10367] = {.lex_state = 180, .external_lex_state = 55},
  [10368] = {.lex_state = 203, .external_lex_state = 86},
  [10369] = {.lex_state = 223},
  [10370] = {.lex_state = 180, .external_lex_state = 55},
  [10371] = {.lex_state = 203, .external_lex_state = 76},
  [10372] = {.lex_state = 180, .external_lex_state = 55},
  [10373] = {.lex_state = 181, .external_lex_state = 55},
  [10374] = {.lex_state = 141, .external_lex_state = 85},
  [10375] = {.lex_state = 223},
  [10376] = {.lex_state = 181, .external_lex_state = 55},
  [10377] = {.lex_state = 203, .external_lex_state = 86},
  [10378] = {.lex_state = 181, .external_lex_state = 55},
  [10379] = {.lex_state = 181, .external_lex_state = 55},
  [10380] = {.lex_state = 181, .external_lex_state = 55},
  [10381] = {.lex_state = 203, .external_lex_state = 76},
  [10382] = {.lex_state = 141, .external_lex_state = 80},
  [10383] = {.lex_state = 141, .external_lex_state = 81},
  [10384] = {.lex_state = 141, .external_lex_state = 80},
  [10385] = {.lex_state = 181, .external_lex_state = 55},
  [10386] = {.lex_state = 179, .external_lex_state = 59},
  [10387] = {.lex_state = 181, .external_lex_state = 55},
  [10388] = {.lex_state = 203, .external_lex_state = 87},
  [10389] = {.lex_state = 203, .external_lex_state = 87},
  [10390] = {.lex_state = 141, .external_lex_state = 80},
  [10391] = {.lex_state = 141, .external_lex_state = 80},
  [10392] = {.lex_state = 203, .external_lex_state = 87},
  [10393] = {.lex_state = 179, .external_lex_state = 59},
  [10394] = {.lex_state = 203, .external_lex_state = 76},
  [10395] = {.lex_state = 203, .external_lex_state = 76},
  [10396] = {.lex_state = 223},
  [10397] = {.lex_state = 203, .external_lex_state = 87},
  [10398] = {.lex_state = 179, .external_lex_state = 59},
  [10399] = {.lex_state = 141, .external_lex_state = 81},
  [10400] = {.lex_state = 203, .external_lex_state = 76},
  [10401] = {.lex_state = 180, .external_lex_state = 55},
  [10402] = {.lex_state = 203, .external_lex_state = 87},
  [10403] = {.lex_state = 179, .external_lex_state = 59},
  [10404] = {.lex_state = 180, .external_lex_state = 55},
  [10405] = {.lex_state = 180, .external_lex_state = 55},
  [10406] = {.lex_state = 180, .external_lex_state = 55},
  [10407] = {.lex_state = 141, .external_lex_state = 80},
  [10408] = {.lex_state = 141, .external_lex_state = 81},
  [10409] = {.lex_state = 181, .external_lex_state = 55},
  [10410] = {.lex_state = 141, .external_lex_state = 81},
  [10411] = {.lex_state = 180, .external_lex_state = 55},
  [10412] = {.lex_state = 141, .external_lex_state = 81},
  [10413] = {.lex_state = 203, .external_lex_state = 87},
  [10414] = {.lex_state = 203, .external_lex_state = 87},
  [10415] = {.lex_state = 179, .external_lex_state = 59},
  [10416] = {.lex_state = 180, .external_lex_state = 55},
  [10417] = {.lex_state = 203, .external_lex_state = 76},
  [10418] = {.lex_state = 141, .external_lex_state = 81},
  [10419] = {.lex_state = 179, .external_lex_state = 59},
  [10420] = {.lex_state = 180, .external_lex_state = 55},
  [10421] = {.lex_state = 203, .external_lex_state = 87},
  [10422] = {.lex_state = 203, .external_lex_state = 76},
  [10423] = {.lex_state = 141, .external_lex_state = 80},
  [10424] = {.lex_state = 141, .external_lex_state = 80},
  [10425] = {.lex_state = 141, .external_lex_state = 80},
  [10426] = {.lex_state = 180, .external_lex_state = 55},
  [10427] = {.lex_state = 141, .external_lex_state = 80},
  [10428] = {.lex_state = 203, .external_lex_state = 87},
  [10429] = {.lex_state = 203, .external_lex_state = 87},
  [10430] = {.lex_state = 224, .external_lex_state = 88},
  [10431] = {.lex_state = 179, .external_lex_state = 59},
  [10432] = {.lex_state = 141, .external_lex_state = 81},
  [10433] = {.lex_state = 141, .external_lex_state = 80},
  [10434] = {.lex_state = 180, .external_lex_state = 55},
  [10435] = {.lex_state = 203, .external_lex_state = 76},
  [10436] = {.lex_state = 141, .external_lex_state = 81},
  [10437] = {.lex_state = 203, .external_lex_state = 87},
  [10438] = {.lex_state = 203, .external_lex_state = 87},
  [10439] = {.lex_state = 203, .external_lex_state = 76},
  [10440] = {.lex_state = 141, .external_lex_state = 80},
  [10441] = {.lex_state = 203, .external_lex_state = 87},
  [10442] = {.lex_state = 203, .external_lex_state = 76},
  [10443] = {.lex_state = 180, .external_lex_state = 55},
  [10444] = {.lex_state = 203, .external_lex_state = 76},
  [10445] = {.lex_state = 180, .external_lex_state = 55},
  [10446] = {.lex_state = 141, .external_lex_state = 81},
  [10447] = {.lex_state = 203, .external_lex_state = 87},
  [10448] = {.lex_state = 180, .external_lex_state = 55},
  [10449] = {.lex_state = 203, .external_lex_state = 87},
  [10450] = {.lex_state = 141, .external_lex_state = 80},
  [10451] = {.lex_state = 203, .external_lex_state = 87},
  [10452] = {.lex_state = 181, .external_lex_state = 55},
  [10453] = {.lex_state = 203, .external_lex_state = 87},
  [10454] = {.lex_state = 141, .external_lex_state = 81},
  [10455] = {.lex_state = 203, .external_lex_state = 78},
  [10456] = {.lex_state = 203, .external_lex_state = 87},
  [10457] = {.lex_state = 180, .external_lex_state = 55},
  [10458] = {.lex_state = 203, .external_lex_state = 87},
  [10459] = {.lex_state = 141, .external_lex_state = 81},
  [10460] = {.lex_state = 203, .external_lex_state = 87},
  [10461] = {.lex_state = 141, .external_lex_state = 80},
  [10462] = {.lex_state = 180, .external_lex_state = 55},
  [10463] = {.lex_state = 203, .external_lex_state = 87},
  [10464] = {.lex_state = 203, .external_lex_state = 76},
  [10465] = {.lex_state = 141, .external_lex_state = 81},
  [10466] = {.lex_state = 203, .external_lex_state = 76},
  [10467] = {.lex_state = 203, .external_lex_state = 87},
  [10468] = {.lex_state = 203, .external_lex_state = 76},
  [10469] = {.lex_state = 141, .external_lex_state = 81},
  [10470] = {.lex_state = 141, .external_lex_state = 82},
  [10471] = {.lex_state = 141, .external_lex_state = 80},
  [10472] = {.lex_state = 223},
  [10473] = {.lex_state = 203, .external_lex_state = 78},
  [10474] = {.lex_state = 203, .external_lex_state = 89},
  [10475] = {.lex_state = 141, .external_lex_state = 82},
  [10476] = {.lex_state = 203, .external_lex_state = 89},
  [10477] = {.lex_state = 223},
  [10478] = {.lex_state = 189, .external_lex_state = 55},
  [10479] = {.lex_state = 203, .external_lex_state = 89},
  [10480] = {.lex_state = 141, .external_lex_state = 82},
  [10481] = {.lex_state = 223},
  [10482] = {.lex_state = 203, .external_lex_state = 77},
  [10483] = {.lex_state = 223},
  [10484] = {.lex_state = 141, .external_lex_state = 82},
  [10485] = {.lex_state = 181, .external_lex_state = 55},
  [10486] = {.lex_state = 223},
  [10487] = {.lex_state = 203, .external_lex_state = 78},
  [10488] = {.lex_state = 203, .external_lex_state = 89},
  [10489] = {.lex_state = 223},
  [10490] = {.lex_state = 182, .external_lex_state = 55},
  [10491] = {.lex_state = 203, .external_lex_state = 89},
  [10492] = {.lex_state = 141, .external_lex_state = 81},
  [10493] = {.lex_state = 203, .external_lex_state = 89},
  [10494] = {.lex_state = 223},
  [10495] = {.lex_state = 223},
  [10496] = {.lex_state = 203, .external_lex_state = 76},
  [10497] = {.lex_state = 203, .external_lex_state = 76},
  [10498] = {.lex_state = 223},
  [10499] = {.lex_state = 203, .external_lex_state = 78},
  [10500] = {.lex_state = 203, .external_lex_state = 78},
  [10501] = {.lex_state = 141, .external_lex_state = 82},
  [10502] = {.lex_state = 141, .external_lex_state = 82},
  [10503] = {.lex_state = 223},
  [10504] = {.lex_state = 203, .external_lex_state = 78},
  [10505] = {.lex_state = 203, .external_lex_state = 78},
  [10506] = {.lex_state = 203, .external_lex_state = 89},
  [10507] = {.lex_state = 203, .external_lex_state = 78},
  [10508] = {.lex_state = 181, .external_lex_state = 55},
  [10509] = {.lex_state = 203, .external_lex_state = 78},
  [10510] = {.lex_state = 181, .external_lex_state = 55},
  [10511] = {.lex_state = 223},
  [10512] = {.lex_state = 189, .external_lex_state = 55},
  [10513] = {.lex_state = 203, .external_lex_state = 78},
  [10514] = {.lex_state = 181, .external_lex_state = 55},
  [10515] = {.lex_state = 181, .external_lex_state = 55},
  [10516] = {.lex_state = 203, .external_lex_state = 89},
  [10517] = {.lex_state = 203, .external_lex_state = 76},
  [10518] = {.lex_state = 203, .external_lex_state = 89},
  [10519] = {.lex_state = 223},
  [10520] = {.lex_state = 181, .external_lex_state = 55},
  [10521] = {.lex_state = 223},
  [10522] = {.lex_state = 223},
  [10523] = {.lex_state = 223},
  [10524] = {.lex_state = 223},
  [10525] = {.lex_state = 203, .external_lex_state = 89},
  [10526] = {.lex_state = 223},
  [10527] = {.lex_state = 203, .external_lex_state = 76},
  [10528] = {.lex_state = 223},
  [10529] = {.lex_state = 223},
  [10530] = {.lex_state = 203, .external_lex_state = 89},
  [10531] = {.lex_state = 223},
  [10532] = {.lex_state = 203, .external_lex_state = 78},
  [10533] = {.lex_state = 181, .external_lex_state = 55},
  [10534] = {.lex_state = 223},
  [10535] = {.lex_state = 223},
  [10536] = {.lex_state = 141, .external_lex_state = 80},
  [10537] = {.lex_state = 182, .external_lex_state = 55},
  [10538] = {.lex_state = 181, .external_lex_state = 55},
  [10539] = {.lex_state = 203, .external_lex_state = 89},
  [10540] = {.lex_state = 141, .external_lex_state = 82},
  [10541] = {.lex_state = 181, .external_lex_state = 55},
  [10542] = {.lex_state = 185, .external_lex_state = 68},
  [10543] = {.lex_state = 141, .external_lex_state = 81},
  [10544] = {.lex_state = 223},
  [10545] = {.lex_state = 223},
  [10546] = {.lex_state = 141, .external_lex_state = 81},
  [10547] = {.lex_state = 181, .external_lex_state = 55},
  [10548] = {.lex_state = 223},
  [10549] = {.lex_state = 203, .external_lex_state = 89},
  [10550] = {.lex_state = 223},
  [10551] = {.lex_state = 203, .external_lex_state = 89},
  [10552] = {.lex_state = 223},
  [10553] = {.lex_state = 223},
  [10554] = {.lex_state = 223},
  [10555] = {.lex_state = 141, .external_lex_state = 81},
  [10556] = {.lex_state = 141, .external_lex_state = 81},
  [10557] = {.lex_state = 203, .external_lex_state = 78},
  [10558] = {.lex_state = 223},
  [10559] = {.lex_state = 223},
  [10560] = {.lex_state = 141, .external_lex_state = 81},
  [10561] = {.lex_state = 141, .external_lex_state = 82},
  [10562] = {.lex_state = 193, .external_lex_state = 68},
  [10563] = {.lex_state = 203, .external_lex_state = 89},
  [10564] = {.lex_state = 203, .external_lex_state = 76},
  [10565] = {.lex_state = 189, .external_lex_state = 55},
  [10566] = {.lex_state = 223},
  [10567] = {.lex_state = 223},
  [10568] = {.lex_state = 223},
  [10569] = {.lex_state = 181, .external_lex_state = 55},
  [10570] = {.lex_state = 141, .external_lex_state = 82},
  [10571] = {.lex_state = 223},
  [10572] = {.lex_state = 181, .external_lex_state = 55},
  [10573] = {.lex_state = 223},
  [10574] = {.lex_state = 203, .external_lex_state = 89},
  [10575] = {.lex_state = 141, .external_lex_state = 80},
  [10576] = {.lex_state = 223},
  [10577] = {.lex_state = 141, .external_lex_state = 80},
  [10578] = {.lex_state = 181, .external_lex_state = 55},
  [10579] = {.lex_state = 203, .external_lex_state = 89},
  [10580] = {.lex_state = 223},
  [10581] = {.lex_state = 181, .external_lex_state = 55},
  [10582] = {.lex_state = 141, .external_lex_state = 82},
  [10583] = {.lex_state = 203, .external_lex_state = 89},
  [10584] = {.lex_state = 141, .external_lex_state = 82},
  [10585] = {.lex_state = 223},
  [10586] = {.lex_state = 203, .external_lex_state = 78},
  [10587] = {.lex_state = 223},
  [10588] = {.lex_state = 223},
  [10589] = {.lex_state = 141, .external_lex_state = 82},
  [10590] = {.lex_state = 203, .external_lex_state = 89},
  [10591] = {.lex_state = 141, .external_lex_state = 82},
  [10592] = {.lex_state = 203, .external_lex_state = 78},
  [10593] = {.lex_state = 223},
  [10594] = {.lex_state = 203, .external_lex_state = 89},
  [10595] = {.lex_state = 223},
  [10596] = {.lex_state = 223},
  [10597] = {.lex_state = 141, .external_lex_state = 80},
  [10598] = {.lex_state = 223},
  [10599] = {.lex_state = 223},
  [10600] = {.lex_state = 223},
  [10601] = {.lex_state = 223},
  [10602] = {.lex_state = 223},
  [10603] = {.lex_state = 182, .external_lex_state = 55},
  [10604] = {.lex_state = 223},
  [10605] = {.lex_state = 141, .external_lex_state = 80},
  [10606] = {.lex_state = 141, .external_lex_state = 83},
  [10607] = {.lex_state = 223},
  [10608] = {.lex_state = 141, .external_lex_state = 80},
  [10609] = {.lex_state = 203, .external_lex_state = 79},
  [10610] = {.lex_state = 223},
  [10611] = {.lex_state = 141, .external_lex_state = 82},
  [10612] = {.lex_state = 203, .external_lex_state = 76},
  [10613] = {.lex_state = 223},
  [10614] = {.lex_state = 141, .external_lex_state = 80},
  [10615] = {.lex_state = 141, .external_lex_state = 82},
  [10616] = {.lex_state = 141, .external_lex_state = 82},
  [10617] = {.lex_state = 212, .external_lex_state = 88},
  [10618] = {.lex_state = 141, .external_lex_state = 82},
  [10619] = {.lex_state = 141, .external_lex_state = 84},
  [10620] = {.lex_state = 223},
  [10621] = {.lex_state = 141, .external_lex_state = 83},
  [10622] = {.lex_state = 223},
  [10623] = {.lex_state = 141, .external_lex_state = 84},
  [10624] = {.lex_state = 203, .external_lex_state = 77},
  [10625] = {.lex_state = 203, .external_lex_state = 79},
  [10626] = {.lex_state = 141, .external_lex_state = 81},
  [10627] = {.lex_state = 141, .external_lex_state = 83},
  [10628] = {.lex_state = 141, .external_lex_state = 84},
  [10629] = {.lex_state = 223},
  [10630] = {.lex_state = 182, .external_lex_state = 55},
  [10631] = {.lex_state = 186, .external_lex_state = 68},
  [10632] = {.lex_state = 203, .external_lex_state = 79},
  [10633] = {.lex_state = 141, .external_lex_state = 80},
  [10634] = {.lex_state = 203, .external_lex_state = 76},
  [10635] = {.lex_state = 189, .external_lex_state = 55},
  [10636] = {.lex_state = 141, .external_lex_state = 80},
  [10637] = {.lex_state = 203, .external_lex_state = 77},
  [10638] = {.lex_state = 194, .external_lex_state = 68},
  [10639] = {.lex_state = 194, .external_lex_state = 68},
  [10640] = {.lex_state = 141, .external_lex_state = 82},
  [10641] = {.lex_state = 141, .external_lex_state = 84},
  [10642] = {.lex_state = 223},
  [10643] = {.lex_state = 203, .external_lex_state = 79},
  [10644] = {.lex_state = 223},
  [10645] = {.lex_state = 203, .external_lex_state = 79},
  [10646] = {.lex_state = 223},
  [10647] = {.lex_state = 223},
  [10648] = {.lex_state = 194, .external_lex_state = 68},
  [10649] = {.lex_state = 223},
  [10650] = {.lex_state = 186, .external_lex_state = 68},
  [10651] = {.lex_state = 141, .external_lex_state = 83},
  [10652] = {.lex_state = 223},
  [10653] = {.lex_state = 186, .external_lex_state = 68},
  [10654] = {.lex_state = 141, .external_lex_state = 80},
  [10655] = {.lex_state = 189, .external_lex_state = 55},
  [10656] = {.lex_state = 203, .external_lex_state = 79},
  [10657] = {.lex_state = 182, .external_lex_state = 55},
  [10658] = {.lex_state = 223},
  [10659] = {.lex_state = 141, .external_lex_state = 83},
  [10660] = {.lex_state = 141, .external_lex_state = 81},
  [10661] = {.lex_state = 141, .external_lex_state = 83},
  [10662] = {.lex_state = 141, .external_lex_state = 83},
  [10663] = {.lex_state = 223},
  [10664] = {.lex_state = 223},
  [10665] = {.lex_state = 141, .external_lex_state = 84},
  [10666] = {.lex_state = 203, .external_lex_state = 77},
  [10667] = {.lex_state = 203, .external_lex_state = 79},
  [10668] = {.lex_state = 203, .external_lex_state = 77},
  [10669] = {.lex_state = 203, .external_lex_state = 76},
  [10670] = {.lex_state = 203, .external_lex_state = 78},
  [10671] = {.lex_state = 141, .external_lex_state = 84},
  [10672] = {.lex_state = 182, .external_lex_state = 55},
  [10673] = {.lex_state = 141, .external_lex_state = 81},
  [10674] = {.lex_state = 194, .external_lex_state = 68},
  [10675] = {.lex_state = 141, .external_lex_state = 81},
  [10676] = {.lex_state = 141, .external_lex_state = 84},
  [10677] = {.lex_state = 186, .external_lex_state = 68},
  [10678] = {.lex_state = 203, .external_lex_state = 79},
  [10679] = {.lex_state = 189, .external_lex_state = 55},
  [10680] = {.lex_state = 141, .external_lex_state = 83},
  [10681] = {.lex_state = 141, .external_lex_state = 81},
  [10682] = {.lex_state = 203, .external_lex_state = 78},
  [10683] = {.lex_state = 141, .external_lex_state = 84},
  [10684] = {.lex_state = 179, .external_lex_state = 55},
  [10685] = {.lex_state = 223},
  [10686] = {.lex_state = 203, .external_lex_state = 78},
  [10687] = {.lex_state = 223},
  [10688] = {.lex_state = 141, .external_lex_state = 82},
  [10689] = {.lex_state = 203, .external_lex_state = 77},
  [10690] = {.lex_state = 203, .external_lex_state = 77},
  [10691] = {.lex_state = 203, .external_lex_state = 78},
  [10692] = {.lex_state = 189, .external_lex_state = 55},
  [10693] = {.lex_state = 141, .external_lex_state = 83},
  [10694] = {.lex_state = 223},
  [10695] = {.lex_state = 203, .external_lex_state = 78},
  [10696] = {.lex_state = 223},
  [10697] = {.lex_state = 182, .external_lex_state = 55},
  [10698] = {.lex_state = 203, .external_lex_state = 77},
  [10699] = {.lex_state = 141, .external_lex_state = 84},
  [10700] = {.lex_state = 203, .external_lex_state = 76},
  [10701] = {.lex_state = 203, .external_lex_state = 79},
  [10702] = {.lex_state = 203, .external_lex_state = 79},
  [10703] = {.lex_state = 141, .external_lex_state = 83},
  [10704] = {.lex_state = 203, .external_lex_state = 79},
  [10705] = {.lex_state = 203, .external_lex_state = 79},
  [10706] = {.lex_state = 141, .external_lex_state = 84},
  [10707] = {.lex_state = 203, .external_lex_state = 77},
  [10708] = {.lex_state = 203, .external_lex_state = 76},
  [10709] = {.lex_state = 141, .external_lex_state = 83},
  [10710] = {.lex_state = 141, .external_lex_state = 84},
  [10711] = {.lex_state = 203, .external_lex_state = 76},
  [10712] = {.lex_state = 141, .external_lex_state = 83},
  [10713] = {.lex_state = 223},
  [10714] = {.lex_state = 141, .external_lex_state = 84},
  [10715] = {.lex_state = 203, .external_lex_state = 79},
  [10716] = {.lex_state = 179, .external_lex_state = 55},
  [10717] = {.lex_state = 189, .external_lex_state = 55},
  [10718] = {.lex_state = 223},
  [10719] = {.lex_state = 223},
  [10720] = {.lex_state = 223},
  [10721] = {.lex_state = 223},
  [10722] = {.lex_state = 203, .external_lex_state = 86},
  [10723] = {.lex_state = 223},
  [10724] = {.lex_state = 223},
  [10725] = {.lex_state = 223},
  [10726] = {.lex_state = 223},
  [10727] = {.lex_state = 223},
  [10728] = {.lex_state = 223},
  [10729] = {.lex_state = 223},
  [10730] = {.lex_state = 223},
  [10731] = {.lex_state = 223},
  [10732] = {.lex_state = 223},
  [10733] = {.lex_state = 223},
  [10734] = {.lex_state = 223},
  [10735] = {.lex_state = 223},
  [10736] = {.lex_state = 223},
  [10737] = {.lex_state = 223},
  [10738] = {.lex_state = 223},
  [10739] = {.lex_state = 223},
  [10740] = {.lex_state = 223},
  [10741] = {.lex_state = 223},
  [10742] = {.lex_state = 223},
  [10743] = {.lex_state = 223},
  [10744] = {.lex_state = 223},
  [10745] = {.lex_state = 223},
  [10746] = {.lex_state = 223},
  [10747] = {.lex_state = 223},
  [10748] = {.lex_state = 223},
  [10749] = {.lex_state = 223},
  [10750] = {.lex_state = 223},
  [10751] = {.lex_state = 223},
  [10752] = {.lex_state = 223},
  [10753] = {.lex_state = 223},
  [10754] = {.lex_state = 223},
  [10755] = {.lex_state = 223},
  [10756] = {.lex_state = 223},
  [10757] = {.lex_state = 223},
  [10758] = {.lex_state = 223},
  [10759] = {.lex_state = 223},
  [10760] = {.lex_state = 223},
  [10761] = {.lex_state = 223},
  [10762] = {.lex_state = 223},
  [10763] = {.lex_state = 223},
  [10764] = {.lex_state = 223},
  [10765] = {.lex_state = 203, .external_lex_state = 86},
  [10766] = {.lex_state = 223},
  [10767] = {.lex_state = 223},
  [10768] = {.lex_state = 223},
  [10769] = {.lex_state = 223},
  [10770] = {.lex_state = 223},
  [10771] = {.lex_state = 223},
  [10772] = {.lex_state = 223},
  [10773] = {.lex_state = 223},
  [10774] = {.lex_state = 223},
  [10775] = {.lex_state = 138},
  [10776] = {.lex_state = 223},
  [10777] = {.lex_state = 223},
  [10778] = {.lex_state = 223},
  [10779] = {.lex_state = 223},
  [10780] = {.lex_state = 223},
  [10781] = {.lex_state = 223},
  [10782] = {.lex_state = 223},
  [10783] = {.lex_state = 141, .external_lex_state = 82},
  [10784] = {.lex_state = 223},
  [10785] = {.lex_state = 223},
  [10786] = {.lex_state = 223},
  [10787] = {.lex_state = 223},
  [10788] = {.lex_state = 223},
  [10789] = {.lex_state = 223},
  [10790] = {.lex_state = 223},
  [10791] = {.lex_state = 223},
  [10792] = {.lex_state = 223},
  [10793] = {.lex_state = 223},
  [10794] = {.lex_state = 223},
  [10795] = {.lex_state = 223},
  [10796] = {.lex_state = 223},
  [10797] = {.lex_state = 223},
  [10798] = {.lex_state = 203, .external_lex_state = 76},
  [10799] = {.lex_state = 223},
  [10800] = {.lex_state = 223},
  [10801] = {.lex_state = 223},
  [10802] = {.lex_state = 223},
  [10803] = {.lex_state = 203, .external_lex_state = 76},
  [10804] = {.lex_state = 223},
  [10805] = {.lex_state = 223},
  [10806] = {.lex_state = 223},
  [10807] = {.lex_state = 223},
  [10808] = {.lex_state = 141, .external_lex_state = 84},
  [10809] = {.lex_state = 223},
  [10810] = {.lex_state = 223},
  [10811] = {.lex_state = 223},
  [10812] = {.lex_state = 223},
  [10813] = {.lex_state = 223},
  [10814] = {.lex_state = 223},
  [10815] = {.lex_state = 223},
  [10816] = {.lex_state = 223},
  [10817] = {.lex_state = 223},
  [10818] = {.lex_state = 223},
  [10819] = {.lex_state = 223},
  [10820] = {.lex_state = 223},
  [10821] = {.lex_state = 223},
  [10822] = {.lex_state = 223},
  [10823] = {.lex_state = 223},
  [10824] = {.lex_state = 138},
  [10825] = {.lex_state = 223},
  [10826] = {.lex_state = 223},
  [10827] = {.lex_state = 223},
  [10828] = {.lex_state = 223},
  [10829] = {.lex_state = 223},
  [10830] = {.lex_state = 223},
  [10831] = {.lex_state = 223},
  [10832] = {.lex_state = 223},
  [10833] = {.lex_state = 223},
  [10834] = {.lex_state = 223},
  [10835] = {.lex_state = 203, .external_lex_state = 76},
  [10836] = {.lex_state = 223},
  [10837] = {.lex_state = 223},
  [10838] = {.lex_state = 223},
  [10839] = {.lex_state = 223},
  [10840] = {.lex_state = 223},
  [10841] = {.lex_state = 223},
  [10842] = {.lex_state = 223},
  [10843] = {.lex_state = 223},
  [10844] = {.lex_state = 223},
  [10845] = {.lex_state = 223},
  [10846] = {.lex_state = 223},
  [10847] = {.lex_state = 223},
  [10848] = {.lex_state = 223},
  [10849] = {.lex_state = 223},
  [10850] = {.lex_state = 223},
  [10851] = {.lex_state = 223},
  [10852] = {.lex_state = 223},
  [10853] = {.lex_state = 223},
  [10854] = {.lex_state = 223},
  [10855] = {.lex_state = 223},
  [10856] = {.lex_state = 223},
  [10857] = {.lex_state = 223},
  [10858] = {.lex_state = 138},
  [10859] = {.lex_state = 223},
  [10860] = {.lex_state = 223},
  [10861] = {.lex_state = 223},
  [10862] = {.lex_state = 223},
  [10863] = {.lex_state = 223},
  [10864] = {.lex_state = 223},
  [10865] = {.lex_state = 223},
  [10866] = {.lex_state = 223},
  [10867] = {.lex_state = 223},
  [10868] = {.lex_state = 223},
  [10869] = {.lex_state = 223},
  [10870] = {.lex_state = 223},
  [10871] = {.lex_state = 223},
  [10872] = {.lex_state = 223},
  [10873] = {.lex_state = 223},
  [10874] = {.lex_state = 223},
  [10875] = {.lex_state = 223},
  [10876] = {.lex_state = 223},
  [10877] = {.lex_state = 223},
  [10878] = {.lex_state = 223},
  [10879] = {.lex_state = 223},
  [10880] = {.lex_state = 223},
  [10881] = {.lex_state = 223},
  [10882] = {.lex_state = 223},
  [10883] = {.lex_state = 223},
  [10884] = {.lex_state = 223},
  [10885] = {.lex_state = 223},
  [10886] = {.lex_state = 223},
  [10887] = {.lex_state = 223},
  [10888] = {.lex_state = 223},
  [10889] = {.lex_state = 223},
  [10890] = {.lex_state = 223},
  [10891] = {.lex_state = 223},
  [10892] = {.lex_state = 223},
  [10893] = {.lex_state = 223},
  [10894] = {.lex_state = 223},
  [10895] = {.lex_state = 223},
  [10896] = {.lex_state = 223},
  [10897] = {.lex_state = 223},
  [10898] = {.lex_state = 223},
  [10899] = {.lex_state = 223},
  [10900] = {.lex_state = 223},
  [10901] = {.lex_state = 223},
  [10902] = {.lex_state = 223},
  [10903] = {.lex_state = 223},
  [10904] = {.lex_state = 223},
  [10905] = {.lex_state = 223},
  [10906] = {.lex_state = 223},
  [10907] = {.lex_state = 223},
  [10908] = {.lex_state = 223},
  [10909] = {.lex_state = 223},
  [10910] = {.lex_state = 223},
  [10911] = {.lex_state = 223},
  [10912] = {.lex_state = 223},
  [10913] = {.lex_state = 223},
  [10914] = {.lex_state = 223},
  [10915] = {.lex_state = 223},
  [10916] = {.lex_state = 223},
  [10917] = {.lex_state = 223},
  [10918] = {.lex_state = 223},
  [10919] = {.lex_state = 223},
  [10920] = {.lex_state = 223},
  [10921] = {.lex_state = 223},
  [10922] = {.lex_state = 203, .external_lex_state = 78},
  [10923] = {.lex_state = 223},
  [10924] = {.lex_state = 223},
  [10925] = {.lex_state = 223},
  [10926] = {.lex_state = 223},
  [10927] = {.lex_state = 223},
  [10928] = {.lex_state = 223},
  [10929] = {.lex_state = 223},
  [10930] = {.lex_state = 223},
  [10931] = {.lex_state = 223},
  [10932] = {.lex_state = 223},
  [10933] = {.lex_state = 223},
  [10934] = {.lex_state = 223},
  [10935] = {.lex_state = 223},
  [10936] = {.lex_state = 203, .external_lex_state = 77},
  [10937] = {.lex_state = 223},
  [10938] = {.lex_state = 223},
  [10939] = {.lex_state = 223},
  [10940] = {.lex_state = 223},
  [10941] = {.lex_state = 223},
  [10942] = {.lex_state = 223},
  [10943] = {.lex_state = 223},
  [10944] = {.lex_state = 223},
  [10945] = {.lex_state = 223},
  [10946] = {.lex_state = 223},
  [10947] = {.lex_state = 223},
  [10948] = {.lex_state = 223},
  [10949] = {.lex_state = 223},
  [10950] = {.lex_state = 223},
  [10951] = {.lex_state = 223},
  [10952] = {.lex_state = 457, .external_lex_state = 55},
  [10953] = {.lex_state = 223},
  [10954] = {.lex_state = 223},
  [10955] = {.lex_state = 223},
  [10956] = {.lex_state = 223},
  [10957] = {.lex_state = 223},
  [10958] = {.lex_state = 223},
  [10959] = {.lex_state = 223},
  [10960] = {.lex_state = 223},
  [10961] = {.lex_state = 203, .external_lex_state = 77},
  [10962] = {.lex_state = 223},
  [10963] = {.lex_state = 223},
  [10964] = {.lex_state = 223},
  [10965] = {.lex_state = 223},
  [10966] = {.lex_state = 223},
  [10967] = {.lex_state = 223},
  [10968] = {.lex_state = 223},
  [10969] = {.lex_state = 223},
  [10970] = {.lex_state = 223},
  [10971] = {.lex_state = 223},
  [10972] = {.lex_state = 223},
  [10973] = {.lex_state = 223},
  [10974] = {.lex_state = 223},
  [10975] = {.lex_state = 182, .external_lex_state = 55},
  [10976] = {.lex_state = 223},
  [10977] = {.lex_state = 223},
  [10978] = {.lex_state = 223},
  [10979] = {.lex_state = 141, .external_lex_state = 82},
  [10980] = {.lex_state = 223},
  [10981] = {.lex_state = 223},
  [10982] = {.lex_state = 223},
  [10983] = {.lex_state = 203, .external_lex_state = 79},
  [10984] = {.lex_state = 223},
  [10985] = {.lex_state = 138},
  [10986] = {.lex_state = 141, .external_lex_state = 83},
  [10987] = {.lex_state = 223},
  [10988] = {.lex_state = 223},
  [10989] = {.lex_state = 223},
  [10990] = {.lex_state = 223},
  [10991] = {.lex_state = 223},
  [10992] = {.lex_state = 223},
  [10993] = {.lex_state = 223},
  [10994] = {.lex_state = 223},
  [10995] = {.lex_state = 223},
  [10996] = {.lex_state = 138},
  [10997] = {.lex_state = 138},
  [10998] = {.lex_state = 223},
  [10999] = {.lex_state = 138},
  [11000] = {.lex_state = 223},
  [11001] = {.lex_state = 223},
  [11002] = {.lex_state = 223},
  [11003] = {.lex_state = 223},
  [11004] = {.lex_state = 457, .external_lex_state = 55},
  [11005] = {.lex_state = 223},
  [11006] = {.lex_state = 223},
  [11007] = {.lex_state = 223},
  [11008] = {.lex_state = 223},
  [11009] = {.lex_state = 223},
  [11010] = {.lex_state = 141, .external_lex_state = 82},
  [11011] = {.lex_state = 223},
  [11012] = {.lex_state = 223},
  [11013] = {.lex_state = 223},
  [11014] = {.lex_state = 223},
  [11015] = {.lex_state = 223},
  [11016] = {.lex_state = 223},
  [11017] = {.lex_state = 203, .external_lex_state = 86},
  [11018] = {.lex_state = 223},
  [11019] = {.lex_state = 223},
  [11020] = {.lex_state = 138},
  [11021] = {.lex_state = 223},
  [11022] = {.lex_state = 141, .external_lex_state = 85},
  [11023] = {.lex_state = 138},
  [11024] = {.lex_state = 223},
  [11025] = {.lex_state = 141, .external_lex_state = 85},
  [11026] = {.lex_state = 223},
  [11027] = {.lex_state = 138},
  [11028] = {.lex_state = 455, .external_lex_state = 55},
  [11029] = {.lex_state = 223},
  [11030] = {.lex_state = 223},
  [11031] = {.lex_state = 223},
  [11032] = {.lex_state = 223},
  [11033] = {.lex_state = 223},
  [11034] = {.lex_state = 223},
  [11035] = {.lex_state = 223},
  [11036] = {.lex_state = 223},
  [11037] = {.lex_state = 223},
  [11038] = {.lex_state = 223},
  [11039] = {.lex_state = 223},
  [11040] = {.lex_state = 223},
  [11041] = {.lex_state = 223},
  [11042] = {.lex_state = 223},
  [11043] = {.lex_state = 223},
  [11044] = {.lex_state = 223},
  [11045] = {.lex_state = 223},
  [11046] = {.lex_state = 223},
  [11047] = {.lex_state = 223},
  [11048] = {.lex_state = 223},
  [11049] = {.lex_state = 223},
  [11050] = {.lex_state = 223},
  [11051] = {.lex_state = 223},
  [11052] = {.lex_state = 223},
  [11053] = {.lex_state = 223},
  [11054] = {.lex_state = 223},
  [11055] = {.lex_state = 223},
  [11056] = {.lex_state = 223},
  [11057] = {.lex_state = 141, .external_lex_state = 81},
  [11058] = {.lex_state = 223},
  [11059] = {.lex_state = 141, .external_lex_state = 81},
  [11060] = {.lex_state = 455, .external_lex_state = 55},
  [11061] = {.lex_state = 223},
  [11062] = {.lex_state = 223},
  [11063] = {.lex_state = 223},
  [11064] = {.lex_state = 223},
  [11065] = {.lex_state = 223},
  [11066] = {.lex_state = 223},
  [11067] = {.lex_state = 223},
  [11068] = {.lex_state = 223},
  [11069] = {.lex_state = 223},
  [11070] = {.lex_state = 223},
  [11071] = {.lex_state = 141, .external_lex_state = 81},
  [11072] = {.lex_state = 203, .external_lex_state = 86},
  [11073] = {.lex_state = 141, .external_lex_state = 81},
  [11074] = {.lex_state = 223},
  [11075] = {.lex_state = 182, .external_lex_state = 55},
  [11076] = {.lex_state = 223},
  [11077] = {.lex_state = 141, .external_lex_state = 81},
  [11078] = {.lex_state = 141, .external_lex_state = 81},
  [11079] = {.lex_state = 223},
  [11080] = {.lex_state = 223},
  [11081] = {.lex_state = 223},
  [11082] = {.lex_state = 223},
  [11083] = {.lex_state = 223},
  [11084] = {.lex_state = 223},
  [11085] = {.lex_state = 141, .external_lex_state = 81},
  [11086] = {.lex_state = 189, .external_lex_state = 55},
  [11087] = {.lex_state = 223},
  [11088] = {.lex_state = 223},
  [11089] = {.lex_state = 223},
  [11090] = {.lex_state = 223},
  [11091] = {.lex_state = 223},
  [11092] = {.lex_state = 223},
  [11093] = {.lex_state = 223},
  [11094] = {.lex_state = 223},
  [11095] = {.lex_state = 223},
  [11096] = {.lex_state = 141, .external_lex_state = 81},
  [11097] = {.lex_state = 223},
  [11098] = {.lex_state = 223},
  [11099] = {.lex_state = 223},
  [11100] = {.lex_state = 223},
  [11101] = {.lex_state = 141, .external_lex_state = 85},
  [11102] = {.lex_state = 223},
  [11103] = {.lex_state = 223},
  [11104] = {.lex_state = 141, .external_lex_state = 85},
  [11105] = {.lex_state = 141, .external_lex_state = 81},
  [11106] = {.lex_state = 223},
  [11107] = {.lex_state = 141, .external_lex_state = 81},
  [11108] = {.lex_state = 141, .external_lex_state = 81},
  [11109] = {.lex_state = 223},
  [11110] = {.lex_state = 223},
  [11111] = {.lex_state = 223},
  [11112] = {.lex_state = 223},
  [11113] = {.lex_state = 223},
  [11114] = {.lex_state = 223},
  [11115] = {.lex_state = 223},
  [11116] = {.lex_state = 223},
  [11117] = {.lex_state = 223},
  [11118] = {.lex_state = 223},
  [11119] = {.lex_state = 203, .external_lex_state = 86},
  [11120] = {.lex_state = 223},
  [11121] = {.lex_state = 223},
  [11122] = {.lex_state = 223},
  [11123] = {.lex_state = 223},
  [11124] = {.lex_state = 223},
  [11125] = {.lex_state = 223},
  [11126] = {.lex_state = 223},
  [11127] = {.lex_state = 223},
  [11128] = {.lex_state = 223},
  [11129] = {.lex_state = 203, .external_lex_state = 78},
  [11130] = {.lex_state = 223},
  [11131] = {.lex_state = 223},
  [11132] = {.lex_state = 223},
  [11133] = {.lex_state = 223},
  [11134] = {.lex_state = 141, .external_lex_state = 80},
  [11135] = {.lex_state = 223},
  [11136] = {.lex_state = 223},
  [11137] = {.lex_state = 223},
  [11138] = {.lex_state = 223},
  [11139] = {.lex_state = 141, .external_lex_state = 80},
  [11140] = {.lex_state = 141, .external_lex_state = 80},
  [11141] = {.lex_state = 223},
  [11142] = {.lex_state = 141, .external_lex_state = 80},
  [11143] = {.lex_state = 223},
  [11144] = {.lex_state = 141, .external_lex_state = 80},
  [11145] = {.lex_state = 223},
  [11146] = {.lex_state = 223},
  [11147] = {.lex_state = 223},
  [11148] = {.lex_state = 223},
  [11149] = {.lex_state = 141, .external_lex_state = 80},
  [11150] = {.lex_state = 223},
  [11151] = {.lex_state = 141, .external_lex_state = 80},
  [11152] = {.lex_state = 223},
  [11153] = {.lex_state = 223},
  [11154] = {.lex_state = 223},
  [11155] = {.lex_state = 223},
  [11156] = {.lex_state = 141, .external_lex_state = 80},
  [11157] = {.lex_state = 223},
  [11158] = {.lex_state = 223},
  [11159] = {.lex_state = 223},
  [11160] = {.lex_state = 223},
  [11161] = {.lex_state = 138},
  [11162] = {.lex_state = 223},
  [11163] = {.lex_state = 223},
  [11164] = {.lex_state = 223},
  [11165] = {.lex_state = 223},
  [11166] = {.lex_state = 141, .external_lex_state = 80},
  [11167] = {.lex_state = 223},
  [11168] = {.lex_state = 223},
  [11169] = {.lex_state = 141, .external_lex_state = 80},
  [11170] = {.lex_state = 223},
  [11171] = {.lex_state = 141, .external_lex_state = 80},
  [11172] = {.lex_state = 223},
  [11173] = {.lex_state = 223},
  [11174] = {.lex_state = 223},
  [11175] = {.lex_state = 141, .external_lex_state = 80},
  [11176] = {.lex_state = 223},
  [11177] = {.lex_state = 223},
  [11178] = {.lex_state = 223},
  [11179] = {.lex_state = 223},
  [11180] = {.lex_state = 203, .external_lex_state = 78},
  [11181] = {.lex_state = 223},
  [11182] = {.lex_state = 223},
  [11183] = {.lex_state = 203, .external_lex_state = 78},
  [11184] = {.lex_state = 203, .external_lex_state = 78},
  [11185] = {.lex_state = 223},
  [11186] = {.lex_state = 223},
  [11187] = {.lex_state = 223},
  [11188] = {.lex_state = 223},
  [11189] = {.lex_state = 223},
  [11190] = {.lex_state = 138},
  [11191] = {.lex_state = 141, .external_lex_state = 81},
  [11192] = {.lex_state = 203, .external_lex_state = 78},
  [11193] = {.lex_state = 223},
  [11194] = {.lex_state = 223},
  [11195] = {.lex_state = 223},
  [11196] = {.lex_state = 138},
  [11197] = {.lex_state = 223},
  [11198] = {.lex_state = 223},
  [11199] = {.lex_state = 223},
  [11200] = {.lex_state = 223},
  [11201] = {.lex_state = 457, .external_lex_state = 55},
  [11202] = {.lex_state = 223},
  [11203] = {.lex_state = 223},
  [11204] = {.lex_state = 223},
  [11205] = {.lex_state = 223},
  [11206] = {.lex_state = 223},
  [11207] = {.lex_state = 223},
  [11208] = {.lex_state = 223},
  [11209] = {.lex_state = 138},
  [11210] = {.lex_state = 223},
  [11211] = {.lex_state = 223},
  [11212] = {.lex_state = 223},
  [11213] = {.lex_state = 223},
  [11214] = {.lex_state = 223},
  [11215] = {.lex_state = 223},
  [11216] = {.lex_state = 223},
  [11217] = {.lex_state = 223},
  [11218] = {.lex_state = 223},
  [11219] = {.lex_state = 223},
  [11220] = {.lex_state = 141, .external_lex_state = 84},
  [11221] = {.lex_state = 224, .external_lex_state = 90},
  [11222] = {.lex_state = 141, .external_lex_state = 84},
  [11223] = {.lex_state = 182, .external_lex_state = 55},
  [11224] = {.lex_state = 223},
  [11225] = {.lex_state = 223},
  [11226] = {.lex_state = 223},
  [11227] = {.lex_state = 141, .external_lex_state = 82},
  [11228] = {.lex_state = 224, .external_lex_state = 91},
  [11229] = {.lex_state = 141, .external_lex_state = 85},
  [11230] = {.lex_state = 141, .external_lex_state = 83},
  [11231] = {.lex_state = 141, .external_lex_state = 85},
  [11232] = {.lex_state = 223},
  [11233] = {.lex_state = 141, .external_lex_state = 85},
  [11234] = {.lex_state = 223},
  [11235] = {.lex_state = 223},
  [11236] = {.lex_state = 141, .external_lex_state = 83},
  [11237] = {.lex_state = 141, .external_lex_state = 81},
  [11238] = {.lex_state = 223},
  [11239] = {.lex_state = 141, .external_lex_state = 81},
  [11240] = {.lex_state = 223},
  [11241] = {.lex_state = 223},
  [11242] = {.lex_state = 223},
  [11243] = {.lex_state = 141, .external_lex_state = 81},
  [11244] = {.lex_state = 223},
  [11245] = {.lex_state = 223},
  [11246] = {.lex_state = 223},
  [11247] = {.lex_state = 223},
  [11248] = {.lex_state = 223},
  [11249] = {.lex_state = 223},
  [11250] = {.lex_state = 223},
  [11251] = {.lex_state = 223},
  [11252] = {.lex_state = 223},
  [11253] = {.lex_state = 223},
  [11254] = {.lex_state = 223},
  [11255] = {.lex_state = 223},
  [11256] = {.lex_state = 223},
  [11257] = {.lex_state = 141, .external_lex_state = 84},
  [11258] = {.lex_state = 223},
  [11259] = {.lex_state = 223},
  [11260] = {.lex_state = 223},
  [11261] = {.lex_state = 223},
  [11262] = {.lex_state = 223},
  [11263] = {.lex_state = 223},
  [11264] = {.lex_state = 223},
  [11265] = {.lex_state = 203, .external_lex_state = 76},
  [11266] = {.lex_state = 223},
  [11267] = {.lex_state = 223},
  [11268] = {.lex_state = 223},
  [11269] = {.lex_state = 223},
  [11270] = {.lex_state = 223},
  [11271] = {.lex_state = 223},
  [11272] = {.lex_state = 203, .external_lex_state = 76},
  [11273] = {.lex_state = 141, .external_lex_state = 84},
  [11274] = {.lex_state = 223},
  [11275] = {.lex_state = 223},
  [11276] = {.lex_state = 223},
  [11277] = {.lex_state = 223},
  [11278] = {.lex_state = 223},
  [11279] = {.lex_state = 223},
  [11280] = {.lex_state = 223},
  [11281] = {.lex_state = 189, .external_lex_state = 55},
  [11282] = {.lex_state = 223},
  [11283] = {.lex_state = 223},
  [11284] = {.lex_state = 223},
  [11285] = {.lex_state = 223},
  [11286] = {.lex_state = 223},
  [11287] = {.lex_state = 223},
  [11288] = {.lex_state = 223},
  [11289] = {.lex_state = 203, .external_lex_state = 76},
  [11290] = {.lex_state = 223},
  [11291] = {.lex_state = 223},
  [11292] = {.lex_state = 223},
  [11293] = {.lex_state = 223},
  [11294] = {.lex_state = 223},
  [11295] = {.lex_state = 223},
  [11296] = {.lex_state = 203, .external_lex_state = 79},
  [11297] = {.lex_state = 223},
  [11298] = {.lex_state = 223},
  [11299] = {.lex_state = 223},
  [11300] = {.lex_state = 223},
  [11301] = {.lex_state = 223},
  [11302] = {.lex_state = 223},
  [11303] = {.lex_state = 223},
  [11304] = {.lex_state = 223},
  [11305] = {.lex_state = 138},
  [11306] = {.lex_state = 223},
  [11307] = {.lex_state = 223},
  [11308] = {.lex_state = 223},
  [11309] = {.lex_state = 223},
  [11310] = {.lex_state = 223},
  [11311] = {.lex_state = 223},
  [11312] = {.lex_state = 223},
  [11313] = {.lex_state = 223},
  [11314] = {.lex_state = 203, .external_lex_state = 76},
  [11315] = {.lex_state = 223},
  [11316] = {.lex_state = 141, .external_lex_state = 80},
  [11317] = {.lex_state = 223},
  [11318] = {.lex_state = 223},
  [11319] = {.lex_state = 223},
  [11320] = {.lex_state = 203, .external_lex_state = 76},
  [11321] = {.lex_state = 223},
  [11322] = {.lex_state = 141, .external_lex_state = 80},
  [11323] = {.lex_state = 223},
  [11324] = {.lex_state = 223},
  [11325] = {.lex_state = 223},
  [11326] = {.lex_state = 223},
  [11327] = {.lex_state = 223},
  [11328] = {.lex_state = 223},
  [11329] = {.lex_state = 223},
  [11330] = {.lex_state = 223},
  [11331] = {.lex_state = 223},
  [11332] = {.lex_state = 223},
  [11333] = {.lex_state = 141, .external_lex_state = 80},
  [11334] = {.lex_state = 223},
  [11335] = {.lex_state = 223},
  [11336] = {.lex_state = 223},
  [11337] = {.lex_state = 223},
  [11338] = {.lex_state = 223},
  [11339] = {.lex_state = 223},
  [11340] = {.lex_state = 138},
  [11341] = {.lex_state = 223},
  [11342] = {.lex_state = 203, .external_lex_state = 79},
  [11343] = {.lex_state = 223},
  [11344] = {.lex_state = 138},
  [11345] = {.lex_state = 223},
  [11346] = {.lex_state = 223},
  [11347] = {.lex_state = 223},
  [11348] = {.lex_state = 223},
  [11349] = {.lex_state = 223},
  [11350] = {.lex_state = 223},
  [11351] = {.lex_state = 223},
  [11352] = {.lex_state = 223},
  [11353] = {.lex_state = 223},
  [11354] = {.lex_state = 223},
  [11355] = {.lex_state = 223},
  [11356] = {.lex_state = 223},
  [11357] = {.lex_state = 223},
  [11358] = {.lex_state = 203, .external_lex_state = 76},
  [11359] = {.lex_state = 223},
  [11360] = {.lex_state = 223},
  [11361] = {.lex_state = 138},
  [11362] = {.lex_state = 223},
  [11363] = {.lex_state = 203, .external_lex_state = 76},
  [11364] = {.lex_state = 223},
  [11365] = {.lex_state = 223},
  [11366] = {.lex_state = 138},
  [11367] = {.lex_state = 223},
  [11368] = {.lex_state = 223},
  [11369] = {.lex_state = 223},
  [11370] = {.lex_state = 223},
  [11371] = {.lex_state = 223},
  [11372] = {.lex_state = 223},
  [11373] = {.lex_state = 223},
  [11374] = {.lex_state = 223},
  [11375] = {.lex_state = 223},
  [11376] = {.lex_state = 223},
  [11377] = {.lex_state = 203, .external_lex_state = 76},
  [11378] = {.lex_state = 223},
  [11379] = {.lex_state = 223},
  [11380] = {.lex_state = 138},
  [11381] = {.lex_state = 203, .external_lex_state = 76},
  [11382] = {.lex_state = 223},
  [11383] = {.lex_state = 223},
  [11384] = {.lex_state = 138},
  [11385] = {.lex_state = 223},
  [11386] = {.lex_state = 223},
  [11387] = {.lex_state = 223},
  [11388] = {.lex_state = 203, .external_lex_state = 76},
  [11389] = {.lex_state = 223},
  [11390] = {.lex_state = 223},
  [11391] = {.lex_state = 223},
  [11392] = {.lex_state = 223},
  [11393] = {.lex_state = 223},
  [11394] = {.lex_state = 223},
  [11395] = {.lex_state = 203, .external_lex_state = 76},
  [11396] = {.lex_state = 223},
  [11397] = {.lex_state = 223},
  [11398] = {.lex_state = 223},
  [11399] = {.lex_state = 223},
  [11400] = {.lex_state = 223},
  [11401] = {.lex_state = 138},
  [11402] = {.lex_state = 223},
  [11403] = {.lex_state = 223},
  [11404] = {.lex_state = 223},
  [11405] = {.lex_state = 203, .external_lex_state = 76},
  [11406] = {.lex_state = 223},
  [11407] = {.lex_state = 223},
  [11408] = {.lex_state = 223},
  [11409] = {.lex_state = 203, .external_lex_state = 76},
  [11410] = {.lex_state = 138},
  [11411] = {.lex_state = 203, .external_lex_state = 77},
  [11412] = {.lex_state = 223},
  [11413] = {.lex_state = 223},
  [11414] = {.lex_state = 203, .external_lex_state = 86},
  [11415] = {.lex_state = 223},
  [11416] = {.lex_state = 223},
  [11417] = {.lex_state = 223},
  [11418] = {.lex_state = 223},
  [11419] = {.lex_state = 223},
  [11420] = {.lex_state = 223},
  [11421] = {.lex_state = 223},
  [11422] = {.lex_state = 223},
  [11423] = {.lex_state = 138},
  [11424] = {.lex_state = 223},
  [11425] = {.lex_state = 223},
  [11426] = {.lex_state = 223},
  [11427] = {.lex_state = 223},
  [11428] = {.lex_state = 223},
  [11429] = {.lex_state = 223},
  [11430] = {.lex_state = 223},
  [11431] = {.lex_state = 223},
  [11432] = {.lex_state = 223},
  [11433] = {.lex_state = 223},
  [11434] = {.lex_state = 223},
  [11435] = {.lex_state = 223},
  [11436] = {.lex_state = 223},
  [11437] = {.lex_state = 223},
  [11438] = {.lex_state = 223},
  [11439] = {.lex_state = 223},
  [11440] = {.lex_state = 223},
  [11441] = {.lex_state = 141, .external_lex_state = 82},
  [11442] = {.lex_state = 223},
  [11443] = {.lex_state = 223},
  [11444] = {.lex_state = 223},
  [11445] = {.lex_state = 223},
  [11446] = {.lex_state = 223},
  [11447] = {.lex_state = 223},
  [11448] = {.lex_state = 223},
  [11449] = {.lex_state = 203, .external_lex_state = 77},
  [11450] = {.lex_state = 223},
  [11451] = {.lex_state = 223},
  [11452] = {.lex_state = 223},
  [11453] = {.lex_state = 223},
  [11454] = {.lex_state = 223},
  [11455] = {.lex_state = 223},
  [11456] = {.lex_state = 223},
  [11457] = {.lex_state = 223},
  [11458] = {.lex_state = 223},
  [11459] = {.lex_state = 223},
  [11460] = {.lex_state = 223},
  [11461] = {.lex_state = 223},
  [11462] = {.lex_state = 223},
  [11463] = {.lex_state = 223},
  [11464] = {.lex_state = 223},
  [11465] = {.lex_state = 223},
  [11466] = {.lex_state = 223},
  [11467] = {.lex_state = 223},
  [11468] = {.lex_state = 223},
  [11469] = {.lex_state = 223},
  [11470] = {.lex_state = 223},
  [11471] = {.lex_state = 223},
  [11472] = {.lex_state = 223},
  [11473] = {.lex_state = 223},
  [11474] = {.lex_state = 223},
  [11475] = {.lex_state = 223},
  [11476] = {.lex_state = 223},
  [11477] = {.lex_state = 223},
  [11478] = {.lex_state = 223},
  [11479] = {.lex_state = 223},
  [11480] = {.lex_state = 223},
  [11481] = {.lex_state = 223},
  [11482] = {.lex_state = 223},
  [11483] = {.lex_state = 223},
  [11484] = {.lex_state = 223},
  [11485] = {.lex_state = 223},
  [11486] = {.lex_state = 223},
  [11487] = {.lex_state = 223},
  [11488] = {.lex_state = 223},
  [11489] = {.lex_state = 223},
  [11490] = {.lex_state = 223},
  [11491] = {.lex_state = 223},
  [11492] = {.lex_state = 223},
  [11493] = {.lex_state = 223},
  [11494] = {.lex_state = 223},
  [11495] = {.lex_state = 138},
  [11496] = {.lex_state = 223},
  [11497] = {.lex_state = 223},
  [11498] = {.lex_state = 223},
  [11499] = {.lex_state = 223},
  [11500] = {.lex_state = 223},
  [11501] = {.lex_state = 223},
  [11502] = {.lex_state = 141, .external_lex_state = 83},
  [11503] = {.lex_state = 223},
  [11504] = {.lex_state = 223},
  [11505] = {.lex_state = 223},
  [11506] = {.lex_state = 141, .external_lex_state = 83},
  [11507] = {.lex_state = 223},
  [11508] = {.lex_state = 203, .external_lex_state = 86},
  [11509] = {.lex_state = 223},
  [11510] = {.lex_state = 223},
  [11511] = {.lex_state = 223},
  [11512] = {.lex_state = 223},
  [11513] = {.lex_state = 223},
  [11514] = {.lex_state = 223},
  [11515] = {.lex_state = 223},
  [11516] = {.lex_state = 223},
  [11517] = {.lex_state = 223},
  [11518] = {.lex_state = 223},
  [11519] = {.lex_state = 223},
  [11520] = {.lex_state = 223},
  [11521] = {.lex_state = 223},
  [11522] = {.lex_state = 203, .external_lex_state = 86},
  [11523] = {.lex_state = 223},
  [11524] = {.lex_state = 223},
  [11525] = {.lex_state = 223},
  [11526] = {.lex_state = 203, .external_lex_state = 86},
  [11527] = {.lex_state = 141, .external_lex_state = 85},
  [11528] = {.lex_state = 223},
  [11529] = {.lex_state = 138},
  [11530] = {.lex_state = 141, .external_lex_state = 85},
  [11531] = {.lex_state = 223},
  [11532] = {.lex_state = 223},
  [11533] = {.lex_state = 223},
  [11534] = {.lex_state = 203, .external_lex_state = 79},
  [11535] = {.lex_state = 223},
  [11536] = {.lex_state = 203, .external_lex_state = 79},
  [11537] = {.lex_state = 223},
  [11538] = {.lex_state = 223},
  [11539] = {.lex_state = 223},
  [11540] = {.lex_state = 141, .external_lex_state = 85},
  [11541] = {.lex_state = 223},
  [11542] = {.lex_state = 223},
  [11543] = {.lex_state = 223},
  [11544] = {.lex_state = 223},
  [11545] = {.lex_state = 203, .external_lex_state = 86},
  [11546] = {.lex_state = 141, .external_lex_state = 85},
  [11547] = {.lex_state = 203, .external_lex_state = 86},
  [11548] = {.lex_state = 223},
  [11549] = {.lex_state = 223},
  [11550] = {.lex_state = 223},
  [11551] = {.lex_state = 223},
  [11552] = {.lex_state = 138},
  [11553] = {.lex_state = 223},
  [11554] = {.lex_state = 223},
  [11555] = {.lex_state = 223},
  [11556] = {.lex_state = 141, .external_lex_state = 85},
  [11557] = {.lex_state = 223},
  [11558] = {.lex_state = 138},
  [11559] = {.lex_state = 223},
  [11560] = {.lex_state = 223},
  [11561] = {.lex_state = 223},
  [11562] = {.lex_state = 203, .external_lex_state = 86},
  [11563] = {.lex_state = 223},
  [11564] = {.lex_state = 223},
  [11565] = {.lex_state = 223},
  [11566] = {.lex_state = 138},
  [11567] = {.lex_state = 223},
  [11568] = {.lex_state = 223},
  [11569] = {.lex_state = 223},
  [11570] = {.lex_state = 223},
  [11571] = {.lex_state = 223},
  [11572] = {.lex_state = 223},
  [11573] = {.lex_state = 223},
  [11574] = {.lex_state = 223},
  [11575] = {.lex_state = 223},
  [11576] = {.lex_state = 223},
  [11577] = {.lex_state = 223},
  [11578] = {.lex_state = 223},
  [11579] = {.lex_state = 223},
  [11580] = {.lex_state = 223},
  [11581] = {.lex_state = 223},
  [11582] = {.lex_state = 223},
  [11583] = {.lex_state = 223},
  [11584] = {.lex_state = 223},
  [11585] = {.lex_state = 223},
  [11586] = {.lex_state = 223},
  [11587] = {.lex_state = 138},
  [11588] = {.lex_state = 223},
  [11589] = {.lex_state = 138},
  [11590] = {.lex_state = 138},
  [11591] = {.lex_state = 223},
  [11592] = {.lex_state = 223},
  [11593] = {.lex_state = 223},
  [11594] = {.lex_state = 223},
  [11595] = {.lex_state = 223},
  [11596] = {.lex_state = 203, .external_lex_state = 79},
  [11597] = {.lex_state = 223},
  [11598] = {.lex_state = 223},
  [11599] = {.lex_state = 223},
  [11600] = {.lex_state = 223},
  [11601] = {.lex_state = 223},
  [11602] = {.lex_state = 223},
  [11603] = {.lex_state = 223},
  [11604] = {.lex_state = 223},
  [11605] = {.lex_state = 223},
  [11606] = {.lex_state = 223},
  [11607] = {.lex_state = 223},
  [11608] = {.lex_state = 223},
  [11609] = {.lex_state = 223},
  [11610] = {.lex_state = 223},
  [11611] = {.lex_state = 223},
  [11612] = {.lex_state = 223},
  [11613] = {.lex_state = 223},
  [11614] = {.lex_state = 223},
  [11615] = {.lex_state = 223},
  [11616] = {.lex_state = 223},
  [11617] = {.lex_state = 223},
  [11618] = {.lex_state = 223},
  [11619] = {.lex_state = 223},
  [11620] = {.lex_state = 223},
  [11621] = {.lex_state = 223},
  [11622] = {.lex_state = 138},
  [11623] = {.lex_state = 223},
  [11624] = {.lex_state = 223},
  [11625] = {.lex_state = 223},
  [11626] = {.lex_state = 223},
  [11627] = {.lex_state = 223},
  [11628] = {.lex_state = 223},
  [11629] = {.lex_state = 223},
  [11630] = {.lex_state = 223},
  [11631] = {.lex_state = 455, .external_lex_state = 55},
  [11632] = {.lex_state = 223},
  [11633] = {.lex_state = 223},
  [11634] = {.lex_state = 138},
  [11635] = {.lex_state = 223},
  [11636] = {.lex_state = 223},
  [11637] = {.lex_state = 223},
  [11638] = {.lex_state = 223},
  [11639] = {.lex_state = 223},
  [11640] = {.lex_state = 203, .external_lex_state = 77},
  [11641] = {.lex_state = 223},
  [11642] = {.lex_state = 223},
  [11643] = {.lex_state = 223},
  [11644] = {.lex_state = 223},
  [11645] = {.lex_state = 223},
  [11646] = {.lex_state = 223},
  [11647] = {.lex_state = 223},
  [11648] = {.lex_state = 223},
  [11649] = {.lex_state = 223},
  [11650] = {.lex_state = 223},
  [11651] = {.lex_state = 223},
  [11652] = {.lex_state = 223},
  [11653] = {.lex_state = 223},
  [11654] = {.lex_state = 223},
  [11655] = {.lex_state = 138},
  [11656] = {.lex_state = 223},
  [11657] = {.lex_state = 223},
  [11658] = {.lex_state = 223},
  [11659] = {.lex_state = 223},
  [11660] = {.lex_state = 223},
  [11661] = {.lex_state = 223},
  [11662] = {.lex_state = 223},
  [11663] = {.lex_state = 138},
  [11664] = {.lex_state = 223},
  [11665] = {.lex_state = 223},
  [11666] = {.lex_state = 223},
  [11667] = {.lex_state = 223},
  [11668] = {.lex_state = 223},
  [11669] = {.lex_state = 223},
  [11670] = {.lex_state = 223},
  [11671] = {.lex_state = 223},
  [11672] = {.lex_state = 223},
  [11673] = {.lex_state = 223},
  [11674] = {.lex_state = 223},
  [11675] = {.lex_state = 223},
  [11676] = {.lex_state = 223},
  [11677] = {.lex_state = 223},
  [11678] = {.lex_state = 223},
  [11679] = {.lex_state = 223},
  [11680] = {.lex_state = 223},
  [11681] = {.lex_state = 223},
  [11682] = {.lex_state = 223},
  [11683] = {.lex_state = 223},
  [11684] = {.lex_state = 223},
  [11685] = {.lex_state = 223},
  [11686] = {.lex_state = 223},
  [11687] = {.lex_state = 223},
  [11688] = {.lex_state = 223},
  [11689] = {.lex_state = 223},
  [11690] = {.lex_state = 223},
  [11691] = {.lex_state = 223},
  [11692] = {.lex_state = 223},
  [11693] = {.lex_state = 223},
  [11694] = {.lex_state = 203, .external_lex_state = 78},
  [11695] = {.lex_state = 455, .external_lex_state = 68},
  [11696] = {.lex_state = 455, .external_lex_state = 68},
  [11697] = {.lex_state = 141, .external_lex_state = 85},
  [11698] = {.lex_state = 455, .external_lex_state = 68},
  [11699] = {.lex_state = 455, .external_lex_state = 55},
  [11700] = {.lex_state = 141, .external_lex_state = 85},
  [11701] = {.lex_state = 203, .external_lex_state = 87},
  [11702] = {.lex_state = 179, .external_lex_state = 55},
  [11703] = {.lex_state = 455, .external_lex_state = 68},
  [11704] = {.lex_state = 141, .external_lex_state = 82},
  [11705] = {.lex_state = 457, .external_lex_state = 68},
  [11706] = {.lex_state = 203, .external_lex_state = 86},
  [11707] = {.lex_state = 189, .external_lex_state = 55},
  [11708] = {.lex_state = 457, .external_lex_state = 68},
  [11709] = {.lex_state = 189, .external_lex_state = 55},
  [11710] = {.lex_state = 455, .external_lex_state = 68},
  [11711] = {.lex_state = 141, .external_lex_state = 82},
  [11712] = {.lex_state = 457, .external_lex_state = 68},
  [11713] = {.lex_state = 179, .external_lex_state = 55},
  [11714] = {.lex_state = 455, .external_lex_state = 68},
  [11715] = {.lex_state = 141, .external_lex_state = 84},
  [11716] = {.lex_state = 203, .external_lex_state = 77},
  [11717] = {.lex_state = 457, .external_lex_state = 68},
  [11718] = {.lex_state = 141, .external_lex_state = 83},
  [11719] = {.lex_state = 455, .external_lex_state = 68},
  [11720] = {.lex_state = 179, .external_lex_state = 55},
  [11721] = {.lex_state = 189, .external_lex_state = 55},
  [11722] = {.lex_state = 457, .external_lex_state = 55},
  [11723] = {.lex_state = 179, .external_lex_state = 55},
  [11724] = {.lex_state = 203, .external_lex_state = 87},
  [11725] = {.lex_state = 203, .external_lex_state = 86},
  [11726] = {.lex_state = 455, .external_lex_state = 68},
  [11727] = {.lex_state = 457, .external_lex_state = 68},
  [11728] = {.lex_state = 457, .external_lex_state = 68},
  [11729] = {.lex_state = 203, .external_lex_state = 86},
  [11730] = {.lex_state = 203, .external_lex_state = 86},
  [11731] = {.lex_state = 203, .external_lex_state = 79},
  [11732] = {.lex_state = 455, .external_lex_state = 68},
  [11733] = {.lex_state = 203, .external_lex_state = 87},
  [11734] = {.lex_state = 141, .external_lex_state = 83},
  [11735] = {.lex_state = 457, .external_lex_state = 68},
  [11736] = {.lex_state = 203, .external_lex_state = 78},
  [11737] = {.lex_state = 457, .external_lex_state = 68},
  [11738] = {.lex_state = 203, .external_lex_state = 78},
  [11739] = {.lex_state = 203, .external_lex_state = 78},
  [11740] = {.lex_state = 203, .external_lex_state = 78},
  [11741] = {.lex_state = 455, .external_lex_state = 68},
  [11742] = {.lex_state = 455, .external_lex_state = 68},
  [11743] = {.lex_state = 203, .external_lex_state = 87},
  [11744] = {.lex_state = 203, .external_lex_state = 87},
  [11745] = {.lex_state = 189, .external_lex_state = 55},
  [11746] = {.lex_state = 182, .external_lex_state = 55},
  [11747] = {.lex_state = 141, .external_lex_state = 82},
  [11748] = {.lex_state = 203, .external_lex_state = 87},
  [11749] = {.lex_state = 455, .external_lex_state = 68},
  [11750] = {.lex_state = 224, .external_lex_state = 92},
  [11751] = {.lex_state = 141, .external_lex_state = 84},
  [11752] = {.lex_state = 455, .external_lex_state = 68},
  [11753] = {.lex_state = 203, .external_lex_state = 86},
  [11754] = {.lex_state = 141, .external_lex_state = 85},
  [11755] = {.lex_state = 141, .external_lex_state = 82},
  [11756] = {.lex_state = 203, .external_lex_state = 79},
  [11757] = {.lex_state = 203, .external_lex_state = 87},
  [11758] = {.lex_state = 455, .external_lex_state = 68},
  [11759] = {.lex_state = 203, .external_lex_state = 77},
  [11760] = {.lex_state = 141, .external_lex_state = 85},
  [11761] = {.lex_state = 141, .external_lex_state = 85},
  [11762] = {.lex_state = 203, .external_lex_state = 78},
  [11763] = {.lex_state = 141, .external_lex_state = 82},
  [11764] = {.lex_state = 141, .external_lex_state = 80},
  [11765] = {.lex_state = 141, .external_lex_state = 80},
  [11766] = {.lex_state = 179, .external_lex_state = 55},
  [11767] = {.lex_state = 203, .external_lex_state = 87},
  [11768] = {.lex_state = 455, .external_lex_state = 68},
  [11769] = {.lex_state = 182, .external_lex_state = 55},
  [11770] = {.lex_state = 203, .external_lex_state = 77},
  [11771] = {.lex_state = 141, .external_lex_state = 82},
  [11772] = {.lex_state = 141, .external_lex_state = 82},
  [11773] = {.lex_state = 141, .external_lex_state = 83},
  [11774] = {.lex_state = 203, .external_lex_state = 87},
  [11775] = {.lex_state = 141, .external_lex_state = 82},
  [11776] = {.lex_state = 455, .external_lex_state = 68},
  [11777] = {.lex_state = 203, .external_lex_state = 78},
  [11778] = {.lex_state = 455, .external_lex_state = 68},
  [11779] = {.lex_state = 203, .external_lex_state = 79},
  [11780] = {.lex_state = 203, .external_lex_state = 78},
  [11781] = {.lex_state = 141, .external_lex_state = 84},
  [11782] = {.lex_state = 182, .external_lex_state = 55},
  [11783] = {.lex_state = 141, .external_lex_state = 82},
  [11784] = {.lex_state = 182, .external_lex_state = 55},
  [11785] = {.lex_state = 203, .external_lex_state = 78},
  [11786] = {.lex_state = 182, .external_lex_state = 55},
  [11787] = {.lex_state = 203, .external_lex_state = 78},
  [11788] = {.lex_state = 455, .external_lex_state = 68},
  [11789] = {.lex_state = 141, .external_lex_state = 82},
  [11790] = {.lex_state = 141, .external_lex_state = 81},
  [11791] = {.lex_state = 455, .external_lex_state = 68},
  [11792] = {.lex_state = 203, .external_lex_state = 78},
  [11793] = {.lex_state = 141, .external_lex_state = 81},
  [11794] = {.lex_state = 203, .external_lex_state = 78},
  [11795] = {.lex_state = 455, .external_lex_state = 68},
  [11796] = {.lex_state = 203, .external_lex_state = 78},
  [11797] = {.lex_state = 203, .external_lex_state = 79},
  [11798] = {.lex_state = 203, .external_lex_state = 79},
  [11799] = {.lex_state = 455, .external_lex_state = 68},
  [11800] = {.lex_state = 203, .external_lex_state = 78},
  [11801] = {.lex_state = 203, .external_lex_state = 78},
  [11802] = {.lex_state = 203, .external_lex_state = 78},
  [11803] = {.lex_state = 457, .external_lex_state = 68},
  [11804] = {.lex_state = 457, .external_lex_state = 68},
  [11805] = {.lex_state = 455, .external_lex_state = 68},
  [11806] = {.lex_state = 455, .external_lex_state = 68},
  [11807] = {.lex_state = 455, .external_lex_state = 68},
  [11808] = {.lex_state = 203, .external_lex_state = 77},
  [11809] = {.lex_state = 457, .external_lex_state = 68},
  [11810] = {.lex_state = 457, .external_lex_state = 68},
  [11811] = {.lex_state = 455, .external_lex_state = 68},
  [11812] = {.lex_state = 457, .external_lex_state = 68},
  [11813] = {.lex_state = 457, .external_lex_state = 68},
  [11814] = {.lex_state = 455, .external_lex_state = 68},
  [11815] = {.lex_state = 455, .external_lex_state = 68},
  [11816] = {.lex_state = 455, .external_lex_state = 68},
  [11817] = {.lex_state = 455, .external_lex_state = 68},
  [11818] = {.lex_state = 457, .external_lex_state = 68},
  [11819] = {.lex_state = 141, .external_lex_state = 82},
  [11820] = {.lex_state = 457, .external_lex_state = 68},
  [11821] = {.lex_state = 457, .external_lex_state = 68},
  [11822] = {.lex_state = 457, .external_lex_state = 68},
  [11823] = {.lex_state = 457, .external_lex_state = 68},
  [11824] = {.lex_state = 457, .external_lex_state = 68},
  [11825] = {.lex_state = 457, .external_lex_state = 68},
  [11826] = {.lex_state = 457, .external_lex_state = 68},
  [11827] = {.lex_state = 457, .external_lex_state = 68},
  [11828] = {.lex_state = 457, .external_lex_state = 68},
  [11829] = {.lex_state = 457, .external_lex_state = 68},
  [11830] = {.lex_state = 141, .external_lex_state = 82},
  [11831] = {.lex_state = 457, .external_lex_state = 68},
  [11832] = {.lex_state = 457, .external_lex_state = 68},
  [11833] = {.lex_state = 457, .external_lex_state = 68},
  [11834] = {.lex_state = 457, .external_lex_state = 68},
  [11835] = {.lex_state = 457, .external_lex_state = 68},
  [11836] = {.lex_state = 457, .external_lex_state = 68},
  [11837] = {.lex_state = 457, .external_lex_state = 68},
  [11838] = {.lex_state = 141, .external_lex_state = 82},
  [11839] = {.lex_state = 203, .external_lex_state = 77},
  [11840] = {.lex_state = 141, .external_lex_state = 84},
  [11841] = {.lex_state = 141, .external_lex_state = 82},
  [11842] = {.lex_state = 179, .external_lex_state = 55},
  [11843] = {.lex_state = 179, .external_lex_state = 55},
  [11844] = {.lex_state = 203, .external_lex_state = 87},
  [11845] = {.lex_state = 457, .external_lex_state = 68},
  [11846] = {.lex_state = 455, .external_lex_state = 68},
  [11847] = {.lex_state = 141, .external_lex_state = 83},
  [11848] = {.lex_state = 141, .external_lex_state = 82},
  [11849] = {.lex_state = 189, .external_lex_state = 55},
  [11850] = {.lex_state = 203, .external_lex_state = 87},
  [11851] = {.lex_state = 455, .external_lex_state = 68},
  [11852] = {.lex_state = 203, .external_lex_state = 77},
  [11853] = {.lex_state = 141, .external_lex_state = 83},
  [11854] = {.lex_state = 203, .external_lex_state = 87},
  [11855] = {.lex_state = 141, .external_lex_state = 84},
  [11856] = {.lex_state = 455, .external_lex_state = 68},
  [11857] = {.lex_state = 455, .external_lex_state = 68},
  [11858] = {.lex_state = 455, .external_lex_state = 87},
  [11859] = {.lex_state = 455, .external_lex_state = 87},
  [11860] = {.lex_state = 455, .external_lex_state = 87},
  [11861] = {.lex_state = 455, .external_lex_state = 87},
  [11862] = {.lex_state = 455, .external_lex_state = 87},
  [11863] = {.lex_state = 455, .external_lex_state = 87},
  [11864] = {.lex_state = 141, .external_lex_state = 80},
  [11865] = {.lex_state = 141, .external_lex_state = 85},
  [11866] = {.lex_state = 455, .external_lex_state = 87},
  [11867] = {.lex_state = 455, .external_lex_state = 87},
  [11868] = {.lex_state = 455, .external_lex_state = 87},
  [11869] = {.lex_state = 455, .external_lex_state = 87},
  [11870] = {.lex_state = 455, .external_lex_state = 87},
  [11871] = {.lex_state = 203, .external_lex_state = 86},
  [11872] = {.lex_state = 141, .external_lex_state = 84},
  [11873] = {.lex_state = 455, .external_lex_state = 87},
  [11874] = {.lex_state = 141, .external_lex_state = 85},
  [11875] = {.lex_state = 455, .external_lex_state = 87},
  [11876] = {.lex_state = 141, .external_lex_state = 84},
  [11877] = {.lex_state = 141, .external_lex_state = 81},
  [11878] = {.lex_state = 141, .external_lex_state = 85},
  [11879] = {.lex_state = 203, .external_lex_state = 77},
  [11880] = {.lex_state = 141, .external_lex_state = 84},
  [11881] = {.lex_state = 203, .external_lex_state = 87},
  [11882] = {.lex_state = 455, .external_lex_state = 87},
  [11883] = {.lex_state = 455, .external_lex_state = 87},
  [11884] = {.lex_state = 141, .external_lex_state = 84},
  [11885] = {.lex_state = 141, .external_lex_state = 84},
  [11886] = {.lex_state = 203, .external_lex_state = 87},
  [11887] = {.lex_state = 455, .external_lex_state = 55},
  [11888] = {.lex_state = 455, .external_lex_state = 87},
  [11889] = {.lex_state = 455, .external_lex_state = 87},
  [11890] = {.lex_state = 455, .external_lex_state = 87},
  [11891] = {.lex_state = 203, .external_lex_state = 86},
  [11892] = {.lex_state = 455, .external_lex_state = 55},
  [11893] = {.lex_state = 141, .external_lex_state = 84},
  [11894] = {.lex_state = 455, .external_lex_state = 87},
  [11895] = {.lex_state = 455, .external_lex_state = 55},
  [11896] = {.lex_state = 141, .external_lex_state = 83},
  [11897] = {.lex_state = 457, .external_lex_state = 55},
  [11898] = {.lex_state = 203, .external_lex_state = 77},
  [11899] = {.lex_state = 203, .external_lex_state = 79},
  [11900] = {.lex_state = 457, .external_lex_state = 55},
  [11901] = {.lex_state = 203, .external_lex_state = 79},
  [11902] = {.lex_state = 203, .external_lex_state = 77},
  [11903] = {.lex_state = 214, .external_lex_state = 91},
  [11904] = {.lex_state = 203, .external_lex_state = 77},
  [11905] = {.lex_state = 455, .external_lex_state = 55},
  [11906] = {.lex_state = 457, .external_lex_state = 55},
  [11907] = {.lex_state = 203, .external_lex_state = 77},
  [11908] = {.lex_state = 141, .external_lex_state = 84},
  [11909] = {.lex_state = 455, .external_lex_state = 55},
  [11910] = {.lex_state = 203, .external_lex_state = 77},
  [11911] = {.lex_state = 455, .external_lex_state = 87},
  [11912] = {.lex_state = 455, .external_lex_state = 87},
  [11913] = {.lex_state = 455, .external_lex_state = 87},
  [11914] = {.lex_state = 457, .external_lex_state = 55},
  [11915] = {.lex_state = 457, .external_lex_state = 55},
  [11916] = {.lex_state = 457, .external_lex_state = 55},
  [11917] = {.lex_state = 203, .external_lex_state = 77},
  [11918] = {.lex_state = 455, .external_lex_state = 87},
  [11919] = {.lex_state = 455, .external_lex_state = 87},
  [11920] = {.lex_state = 455, .external_lex_state = 87},
  [11921] = {.lex_state = 455, .external_lex_state = 87},
  [11922] = {.lex_state = 203, .external_lex_state = 77},
  [11923] = {.lex_state = 141, .external_lex_state = 83},
  [11924] = {.lex_state = 203, .external_lex_state = 86},
  [11925] = {.lex_state = 455, .external_lex_state = 87},
  [11926] = {.lex_state = 455, .external_lex_state = 87},
  [11927] = {.lex_state = 455, .external_lex_state = 87},
  [11928] = {.lex_state = 141, .external_lex_state = 84},
  [11929] = {.lex_state = 455, .external_lex_state = 87},
  [11930] = {.lex_state = 203, .external_lex_state = 77},
  [11931] = {.lex_state = 455, .external_lex_state = 87},
  [11932] = {.lex_state = 203, .external_lex_state = 87},
  [11933] = {.lex_state = 457, .external_lex_state = 55},
  [11934] = {.lex_state = 203, .external_lex_state = 79},
  [11935] = {.lex_state = 203, .external_lex_state = 79},
  [11936] = {.lex_state = 141, .external_lex_state = 84},
  [11937] = {.lex_state = 203, .external_lex_state = 79},
  [11938] = {.lex_state = 455, .external_lex_state = 55},
  [11939] = {.lex_state = 141, .external_lex_state = 84},
  [11940] = {.lex_state = 141, .external_lex_state = 83},
  [11941] = {.lex_state = 457, .external_lex_state = 55},
  [11942] = {.lex_state = 203, .external_lex_state = 79},
  [11943] = {.lex_state = 455, .external_lex_state = 87},
  [11944] = {.lex_state = 141, .external_lex_state = 82},
  [11945] = {.lex_state = 141, .external_lex_state = 82},
  [11946] = {.lex_state = 141, .external_lex_state = 83},
  [11947] = {.lex_state = 141, .external_lex_state = 84},
  [11948] = {.lex_state = 141, .external_lex_state = 83},
  [11949] = {.lex_state = 455, .external_lex_state = 55},
  [11950] = {.lex_state = 138, .external_lex_state = 79},
  [11951] = {.lex_state = 203, .external_lex_state = 77},
  [11952] = {.lex_state = 457, .external_lex_state = 55},
  [11953] = {.lex_state = 455, .external_lex_state = 55},
  [11954] = {.lex_state = 141, .external_lex_state = 83},
  [11955] = {.lex_state = 141, .external_lex_state = 83},
  [11956] = {.lex_state = 141, .external_lex_state = 83},
  [11957] = {.lex_state = 457, .external_lex_state = 55},
  [11958] = {.lex_state = 141, .external_lex_state = 83},
  [11959] = {.lex_state = 203, .external_lex_state = 79},
  [11960] = {.lex_state = 203, .external_lex_state = 79},
  [11961] = {.lex_state = 203, .external_lex_state = 79},
  [11962] = {.lex_state = 203, .external_lex_state = 79},
  [11963] = {.lex_state = 203, .external_lex_state = 79},
  [11964] = {.lex_state = 203, .external_lex_state = 77},
  [11965] = {.lex_state = 455, .external_lex_state = 87},
  [11966] = {.lex_state = 455, .external_lex_state = 87},
  [11967] = {.lex_state = 203, .external_lex_state = 87},
  [11968] = {.lex_state = 203, .external_lex_state = 77},
  [11969] = {.lex_state = 455, .external_lex_state = 55},
  [11970] = {.lex_state = 455, .external_lex_state = 55},
  [11971] = {.lex_state = 457, .external_lex_state = 55},
  [11972] = {.lex_state = 141, .external_lex_state = 83},
  [11973] = {.lex_state = 203, .external_lex_state = 79},
  [11974] = {.lex_state = 203, .external_lex_state = 77},
  [11975] = {.lex_state = 455, .external_lex_state = 87},
  [11976] = {.lex_state = 455, .external_lex_state = 87},
  [11977] = {.lex_state = 455, .external_lex_state = 87},
  [11978] = {.lex_state = 455, .external_lex_state = 87},
  [11979] = {.lex_state = 141, .external_lex_state = 83},
  [11980] = {.lex_state = 203, .external_lex_state = 86},
  [11981] = {.lex_state = 203, .external_lex_state = 79},
  [11982] = {.lex_state = 455, .external_lex_state = 87},
  [11983] = {.lex_state = 455, .external_lex_state = 87},
  [11984] = {.lex_state = 455, .external_lex_state = 87},
  [11985] = {.lex_state = 141, .external_lex_state = 84},
  [11986] = {.lex_state = 455, .external_lex_state = 87},
  [11987] = {.lex_state = 141, .external_lex_state = 84},
  [11988] = {.lex_state = 455, .external_lex_state = 87},
  [11989] = {.lex_state = 455, .external_lex_state = 87},
  [11990] = {.lex_state = 141, .external_lex_state = 83},
  [11991] = {.lex_state = 203, .external_lex_state = 77},
  [11992] = {.lex_state = 455, .external_lex_state = 87},
  [11993] = {.lex_state = 203, .external_lex_state = 79},
  [11994] = {.lex_state = 457, .external_lex_state = 55},
  [11995] = {.lex_state = 203, .external_lex_state = 79},
  [11996] = {.lex_state = 214, .external_lex_state = 90},
  [11997] = {.lex_state = 455, .external_lex_state = 87},
  [11998] = {.lex_state = 455, .external_lex_state = 87},
  [11999] = {.lex_state = 455, .external_lex_state = 87},
  [12000] = {.lex_state = 455, .external_lex_state = 55},
  [12001] = {.lex_state = 455, .external_lex_state = 55},
  [12002] = {.lex_state = 455, .external_lex_state = 87},
  [12003] = {.lex_state = 141, .external_lex_state = 85},
  [12004] = {.lex_state = 203, .external_lex_state = 77},
  [12005] = {.lex_state = 141, .external_lex_state = 83},
  [12006] = {.lex_state = 457, .external_lex_state = 55},
  [12007] = {.lex_state = 457, .external_lex_state = 55},
  [12008] = {.lex_state = 141, .external_lex_state = 85},
  [12009] = {.lex_state = 455, .external_lex_state = 55},
  [12010] = {.lex_state = 141, .external_lex_state = 84},
  [12011] = {.lex_state = 455, .external_lex_state = 55},
  [12012] = {.lex_state = 457, .external_lex_state = 55},
  [12013] = {.lex_state = 203, .external_lex_state = 86},
  [12014] = {.lex_state = 141, .external_lex_state = 83},
  [12015] = {.lex_state = 203, .external_lex_state = 77},
  [12016] = {.lex_state = 455, .external_lex_state = 55},
  [12017] = {.lex_state = 457, .external_lex_state = 55},
  [12018] = {.lex_state = 203, .external_lex_state = 87},
  [12019] = {.lex_state = 455, .external_lex_state = 55},
  [12020] = {.lex_state = 203, .external_lex_state = 86},
  [12021] = {.lex_state = 141, .external_lex_state = 81},
  [12022] = {.lex_state = 138, .external_lex_state = 79},
  [12023] = {.lex_state = 138, .external_lex_state = 79},
  [12024] = {.lex_state = 141, .external_lex_state = 81},
  [12025] = {.lex_state = 214, .external_lex_state = 92},
  [12026] = {.lex_state = 141, .external_lex_state = 81},
  [12027] = {.lex_state = 141, .external_lex_state = 85},
  [12028] = {.lex_state = 141, .external_lex_state = 85},
  [12029] = {.lex_state = 141, .external_lex_state = 80},
  [12030] = {.lex_state = 141, .external_lex_state = 81},
  [12031] = {.lex_state = 141, .external_lex_state = 85},
  [12032] = {.lex_state = 141, .external_lex_state = 80},
  [12033] = {.lex_state = 141, .external_lex_state = 85},
  [12034] = {.lex_state = 203, .external_lex_state = 79},
  [12035] = {.lex_state = 203, .external_lex_state = 79},
  [12036] = {.lex_state = 203, .external_lex_state = 86},
  [12037] = {.lex_state = 141, .external_lex_state = 85},
  [12038] = {.lex_state = 203, .external_lex_state = 86},
  [12039] = {.lex_state = 203, .external_lex_state = 87},
  [12040] = {.lex_state = 203, .external_lex_state = 87},
  [12041] = {.lex_state = 203, .external_lex_state = 86},
  [12042] = {.lex_state = 141, .external_lex_state = 85},
  [12043] = {.lex_state = 141, .external_lex_state = 85},
  [12044] = {.lex_state = 141, .external_lex_state = 80},
  [12045] = {.lex_state = 203, .external_lex_state = 86},
  [12046] = {.lex_state = 141, .external_lex_state = 93},
  [12047] = {.lex_state = 141, .external_lex_state = 85},
  [12048] = {.lex_state = 141, .external_lex_state = 93},
  [12049] = {.lex_state = 203, .external_lex_state = 87},
  [12050] = {.lex_state = 138, .external_lex_state = 79},
  [12051] = {.lex_state = 141, .external_lex_state = 85},
  [12052] = {.lex_state = 141, .external_lex_state = 85},
  [12053] = {.lex_state = 141, .external_lex_state = 85},
  [12054] = {.lex_state = 203, .external_lex_state = 86},
  [12055] = {.lex_state = 141, .external_lex_state = 81},
  [12056] = {.lex_state = 203, .external_lex_state = 87},
  [12057] = {.lex_state = 138, .external_lex_state = 79},
  [12058] = {.lex_state = 141, .external_lex_state = 85},
  [12059] = {.lex_state = 203, .external_lex_state = 86},
  [12060] = {.lex_state = 138, .external_lex_state = 79},
  [12061] = {.lex_state = 141, .external_lex_state = 85},
  [12062] = {.lex_state = 203, .external_lex_state = 86},
  [12063] = {.lex_state = 138, .external_lex_state = 79},
  [12064] = {.lex_state = 138, .external_lex_state = 79},
  [12065] = {.lex_state = 141, .external_lex_state = 82},
  [12066] = {.lex_state = 141, .external_lex_state = 93},
  [12067] = {.lex_state = 141, .external_lex_state = 85},
  [12068] = {.lex_state = 203, .external_lex_state = 86},
  [12069] = {.lex_state = 203, .external_lex_state = 86},
  [12070] = {.lex_state = 203, .external_lex_state = 86},
  [12071] = {.lex_state = 203, .external_lex_state = 87},
  [12072] = {.lex_state = 141, .external_lex_state = 80},
  [12073] = {.lex_state = 141, .external_lex_state = 80},
  [12074] = {.lex_state = 203, .external_lex_state = 86},
  [12075] = {.lex_state = 203, .external_lex_state = 86},
  [12076] = {.lex_state = 203, .external_lex_state = 86},
  [12077] = {.lex_state = 138, .external_lex_state = 79},
  [12078] = {.lex_state = 141, .external_lex_state = 82},
  [12079] = {.lex_state = 141, .external_lex_state = 80},
  [12080] = {.lex_state = 141, .external_lex_state = 94},
  [12081] = {.lex_state = 455, .external_lex_state = 55},
  [12082] = {.lex_state = 203, .external_lex_state = 87},
  [12083] = {.lex_state = 455, .external_lex_state = 55},
  [12084] = {.lex_state = 141, .external_lex_state = 81},
  [12085] = {.lex_state = 203, .external_lex_state = 87},
  [12086] = {.lex_state = 138, .external_lex_state = 87},
  [12087] = {.lex_state = 203, .external_lex_state = 87},
  [12088] = {.lex_state = 203, .external_lex_state = 87},
  [12089] = {.lex_state = 455, .external_lex_state = 55},
  [12090] = {.lex_state = 138, .external_lex_state = 87},
  [12091] = {.lex_state = 203, .external_lex_state = 87},
  [12092] = {.lex_state = 138, .external_lex_state = 87},
  [12093] = {.lex_state = 138, .external_lex_state = 95},
  [12094] = {.lex_state = 141, .external_lex_state = 93},
  [12095] = {.lex_state = 141, .external_lex_state = 94},
  [12096] = {.lex_state = 203, .external_lex_state = 87},
  [12097] = {.lex_state = 138, .external_lex_state = 79},
  [12098] = {.lex_state = 138, .external_lex_state = 87},
  [12099] = {.lex_state = 141, .external_lex_state = 80},
  [12100] = {.lex_state = 138, .external_lex_state = 95},
  [12101] = {.lex_state = 203, .external_lex_state = 87},
  [12102] = {.lex_state = 138, .external_lex_state = 87},
  [12103] = {.lex_state = 138, .external_lex_state = 79},
  [12104] = {.lex_state = 203, .external_lex_state = 87},
  [12105] = {.lex_state = 203, .external_lex_state = 87},
  [12106] = {.lex_state = 138, .external_lex_state = 95},
  [12107] = {.lex_state = 203, .external_lex_state = 76},
  [12108] = {.lex_state = 138, .external_lex_state = 95},
  [12109] = {.lex_state = 455, .external_lex_state = 55},
  [12110] = {.lex_state = 141, .external_lex_state = 80},
  [12111] = {.lex_state = 141, .external_lex_state = 93},
  [12112] = {.lex_state = 141, .external_lex_state = 81},
  [12113] = {.lex_state = 141, .external_lex_state = 81},
  [12114] = {.lex_state = 138, .external_lex_state = 79},
  [12115] = {.lex_state = 203, .external_lex_state = 76},
  [12116] = {.lex_state = 141, .external_lex_state = 82},
  [12117] = {.lex_state = 141, .external_lex_state = 82},
  [12118] = {.lex_state = 203, .external_lex_state = 76},
  [12119] = {.lex_state = 141, .external_lex_state = 93},
  [12120] = {.lex_state = 141, .external_lex_state = 94},
  [12121] = {.lex_state = 141, .external_lex_state = 93},
  [12122] = {.lex_state = 141, .external_lex_state = 80},
  [12123] = {.lex_state = 138, .external_lex_state = 79},
  [12124] = {.lex_state = 203, .external_lex_state = 79},
  [12125] = {.lex_state = 138, .external_lex_state = 87},
  [12126] = {.lex_state = 138, .external_lex_state = 87},
  [12127] = {.lex_state = 141, .external_lex_state = 81},
  [12128] = {.lex_state = 138, .external_lex_state = 87},
  [12129] = {.lex_state = 138, .external_lex_state = 87},
  [12130] = {.lex_state = 138, .external_lex_state = 95},
  [12131] = {.lex_state = 138, .external_lex_state = 87},
  [12132] = {.lex_state = 203, .external_lex_state = 87},
  [12133] = {.lex_state = 203, .external_lex_state = 76},
  [12134] = {.lex_state = 138, .external_lex_state = 95},
  [12135] = {.lex_state = 455, .external_lex_state = 55},
  [12136] = {.lex_state = 141, .external_lex_state = 93},
  [12137] = {.lex_state = 141, .external_lex_state = 81},
  [12138] = {.lex_state = 141, .external_lex_state = 82},
  [12139] = {.lex_state = 203, .external_lex_state = 76},
  [12140] = {.lex_state = 138, .external_lex_state = 87},
  [12141] = {.lex_state = 138, .external_lex_state = 87},
  [12142] = {.lex_state = 141, .external_lex_state = 93},
  [12143] = {.lex_state = 138, .external_lex_state = 95},
  [12144] = {.lex_state = 203, .external_lex_state = 76},
  [12145] = {.lex_state = 203, .external_lex_state = 87},
  [12146] = {.lex_state = 203, .external_lex_state = 96},
  [12147] = {.lex_state = 138, .external_lex_state = 95},
  [12148] = {.lex_state = 141, .external_lex_state = 81},
  [12149] = {.lex_state = 203, .external_lex_state = 76},
  [12150] = {.lex_state = 141, .external_lex_state = 80},
  [12151] = {.lex_state = 203, .external_lex_state = 96},
  [12152] = {.lex_state = 138, .external_lex_state = 79},
  [12153] = {.lex_state = 203, .external_lex_state = 87},
  [12154] = {.lex_state = 203, .external_lex_state = 87},
  [12155] = {.lex_state = 203, .external_lex_state = 96},
  [12156] = {.lex_state = 203, .external_lex_state = 76},
  [12157] = {.lex_state = 203, .external_lex_state = 87},
  [12158] = {.lex_state = 141, .external_lex_state = 80},
  [12159] = {.lex_state = 141, .external_lex_state = 82},
  [12160] = {.lex_state = 203, .external_lex_state = 76},
  [12161] = {.lex_state = 224, .external_lex_state = 97},
  [12162] = {.lex_state = 203, .external_lex_state = 76},
  [12163] = {.lex_state = 203, .external_lex_state = 79},
  [12164] = {.lex_state = 203, .external_lex_state = 76},
  [12165] = {.lex_state = 138, .external_lex_state = 87},
  [12166] = {.lex_state = 203, .external_lex_state = 98},
  [12167] = {.lex_state = 455, .external_lex_state = 55},
  [12168] = {.lex_state = 141, .external_lex_state = 80},
  [12169] = {.lex_state = 203, .external_lex_state = 76},
  [12170] = {.lex_state = 203, .external_lex_state = 76},
  [12171] = {.lex_state = 203, .external_lex_state = 76},
  [12172] = {.lex_state = 141, .external_lex_state = 82},
  [12173] = {.lex_state = 138, .external_lex_state = 87},
  [12174] = {.lex_state = 138, .external_lex_state = 87},
  [12175] = {.lex_state = 138, .external_lex_state = 99},
  [12176] = {.lex_state = 141, .external_lex_state = 94},
  [12177] = {.lex_state = 138, .external_lex_state = 99},
  [12178] = {.lex_state = 203, .external_lex_state = 76},
  [12179] = {.lex_state = 203, .external_lex_state = 76},
  [12180] = {.lex_state = 203, .external_lex_state = 76},
  [12181] = {.lex_state = 203, .external_lex_state = 76},
  [12182] = {.lex_state = 141, .external_lex_state = 94},
  [12183] = {.lex_state = 138, .external_lex_state = 99},
  [12184] = {.lex_state = 203, .external_lex_state = 76},
  [12185] = {.lex_state = 203, .external_lex_state = 76},
  [12186] = {.lex_state = 138, .external_lex_state = 99},
  [12187] = {.lex_state = 138, .external_lex_state = 95},
  [12188] = {.lex_state = 138, .external_lex_state = 99},
  [12189] = {.lex_state = 203, .external_lex_state = 76},
  [12190] = {.lex_state = 203, .external_lex_state = 76},
  [12191] = {.lex_state = 203, .external_lex_state = 76},
  [12192] = {.lex_state = 208, .external_lex_state = 100},
  [12193] = {.lex_state = 203, .external_lex_state = 76},
  [12194] = {.lex_state = 141, .external_lex_state = 82},
  [12195] = {.lex_state = 138, .external_lex_state = 79},
  [12196] = {.lex_state = 203, .external_lex_state = 76},
  [12197] = {.lex_state = 455, .external_lex_state = 55},
  [12198] = {.lex_state = 203, .external_lex_state = 78},
  [12199] = {.lex_state = 138, .external_lex_state = 79},
  [12200] = {.lex_state = 141, .external_lex_state = 82},
  [12201] = {.lex_state = 203, .external_lex_state = 76},
  [12202] = {.lex_state = 203, .external_lex_state = 76},
  [12203] = {.lex_state = 203, .external_lex_state = 76},
  [12204] = {.lex_state = 141, .external_lex_state = 82},
  [12205] = {.lex_state = 138, .external_lex_state = 95},
  [12206] = {.lex_state = 203, .external_lex_state = 76},
  [12207] = {.lex_state = 203, .external_lex_state = 78},
  [12208] = {.lex_state = 455, .external_lex_state = 55},
  [12209] = {.lex_state = 138, .external_lex_state = 95},
  [12210] = {.lex_state = 203, .external_lex_state = 98},
  [12211] = {.lex_state = 141, .external_lex_state = 94},
  [12212] = {.lex_state = 138, .external_lex_state = 79},
  [12213] = {.lex_state = 141, .external_lex_state = 94},
  [12214] = {.lex_state = 203, .external_lex_state = 76},
  [12215] = {.lex_state = 203, .external_lex_state = 76},
  [12216] = {.lex_state = 138, .external_lex_state = 87},
  [12217] = {.lex_state = 141, .external_lex_state = 81},
  [12218] = {.lex_state = 203, .external_lex_state = 76},
  [12219] = {.lex_state = 203, .external_lex_state = 76},
  [12220] = {.lex_state = 141, .external_lex_state = 81},
  [12221] = {.lex_state = 203, .external_lex_state = 98},
  [12222] = {.lex_state = 203, .external_lex_state = 79},
  [12223] = {.lex_state = 203, .external_lex_state = 78},
  [12224] = {.lex_state = 138, .external_lex_state = 79},
  [12225] = {.lex_state = 203, .external_lex_state = 76},
  [12226] = {.lex_state = 141, .external_lex_state = 94},
  [12227] = {.lex_state = 208, .external_lex_state = 101},
  [12228] = {.lex_state = 203, .external_lex_state = 76},
  [12229] = {.lex_state = 203, .external_lex_state = 76},
  [12230] = {.lex_state = 138, .external_lex_state = 99},
  [12231] = {.lex_state = 203, .external_lex_state = 76},
  [12232] = {.lex_state = 141, .external_lex_state = 82},
  [12233] = {.lex_state = 138, .external_lex_state = 99},
  [12234] = {.lex_state = 138, .external_lex_state = 95},
  [12235] = {.lex_state = 203, .external_lex_state = 76},
  [12236] = {.lex_state = 203, .external_lex_state = 76},
  [12237] = {.lex_state = 203, .external_lex_state = 76},
  [12238] = {.lex_state = 203, .external_lex_state = 76},
  [12239] = {.lex_state = 203, .external_lex_state = 76},
  [12240] = {.lex_state = 141, .external_lex_state = 82},
  [12241] = {.lex_state = 203, .external_lex_state = 76},
  [12242] = {.lex_state = 203, .external_lex_state = 76},
  [12243] = {.lex_state = 138, .external_lex_state = 99},
  [12244] = {.lex_state = 203, .external_lex_state = 76},
  [12245] = {.lex_state = 203, .external_lex_state = 78},
  [12246] = {.lex_state = 138, .external_lex_state = 95},
  [12247] = {.lex_state = 203, .external_lex_state = 76},
  [12248] = {.lex_state = 203, .external_lex_state = 76},
  [12249] = {.lex_state = 203, .external_lex_state = 76},
  [12250] = {.lex_state = 203, .external_lex_state = 76},
  [12251] = {.lex_state = 203, .external_lex_state = 76},
  [12252] = {.lex_state = 203, .external_lex_state = 76},
  [12253] = {.lex_state = 141, .external_lex_state = 80},
  [12254] = {.lex_state = 203, .external_lex_state = 76},
  [12255] = {.lex_state = 138, .external_lex_state = 87},
  [12256] = {.lex_state = 224, .external_lex_state = 102},
  [12257] = {.lex_state = 203, .external_lex_state = 76},
  [12258] = {.lex_state = 203, .external_lex_state = 76},
  [12259] = {.lex_state = 203, .external_lex_state = 76},
  [12260] = {.lex_state = 138, .external_lex_state = 99},
  [12261] = {.lex_state = 203, .external_lex_state = 78},
  [12262] = {.lex_state = 203, .external_lex_state = 76},
  [12263] = {.lex_state = 203, .external_lex_state = 76},
  [12264] = {.lex_state = 203, .external_lex_state = 78},
  [12265] = {.lex_state = 141, .external_lex_state = 81},
  [12266] = {.lex_state = 203, .external_lex_state = 76},
  [12267] = {.lex_state = 203, .external_lex_state = 79},
  [12268] = {.lex_state = 203, .external_lex_state = 76},
  [12269] = {.lex_state = 138, .external_lex_state = 99},
  [12270] = {.lex_state = 203, .external_lex_state = 79},
  [12271] = {.lex_state = 203, .external_lex_state = 78},
  [12272] = {.lex_state = 203, .external_lex_state = 76},
  [12273] = {.lex_state = 141, .external_lex_state = 80},
  [12274] = {.lex_state = 138, .external_lex_state = 99},
  [12275] = {.lex_state = 203, .external_lex_state = 76},
  [12276] = {.lex_state = 203, .external_lex_state = 76},
  [12277] = {.lex_state = 203, .external_lex_state = 78},
  [12278] = {.lex_state = 203, .external_lex_state = 76},
  [12279] = {.lex_state = 203, .external_lex_state = 76},
  [12280] = {.lex_state = 138, .external_lex_state = 79},
  [12281] = {.lex_state = 141, .external_lex_state = 94},
  [12282] = {.lex_state = 203, .external_lex_state = 76},
  [12283] = {.lex_state = 138, .external_lex_state = 99},
  [12284] = {.lex_state = 138, .external_lex_state = 79},
  [12285] = {.lex_state = 203, .external_lex_state = 79},
  [12286] = {.lex_state = 203, .external_lex_state = 76},
  [12287] = {.lex_state = 203, .external_lex_state = 76},
  [12288] = {.lex_state = 203, .external_lex_state = 76},
  [12289] = {.lex_state = 208, .external_lex_state = 100},
  [12290] = {.lex_state = 209, .external_lex_state = 103},
  [12291] = {.lex_state = 203, .external_lex_state = 78},
  [12292] = {.lex_state = 208, .external_lex_state = 101},
  [12293] = {.lex_state = 141, .external_lex_state = 81},
  [12294] = {.lex_state = 141, .external_lex_state = 84},
  [12295] = {.lex_state = 141, .external_lex_state = 84},
  [12296] = {.lex_state = 138, .external_lex_state = 79},
  [12297] = {.lex_state = 209, .external_lex_state = 103},
  [12298] = {.lex_state = 141, .external_lex_state = 83},
  [12299] = {.lex_state = 209, .external_lex_state = 104},
  [12300] = {.lex_state = 141, .external_lex_state = 105},
  [12301] = {.lex_state = 203, .external_lex_state = 78},
  [12302] = {.lex_state = 138, .external_lex_state = 79},
  [12303] = {.lex_state = 209, .external_lex_state = 103},
  [12304] = {.lex_state = 141, .external_lex_state = 84},
  [12305] = {.lex_state = 203, .external_lex_state = 78},
  [12306] = {.lex_state = 141, .external_lex_state = 84},
  [12307] = {.lex_state = 209, .external_lex_state = 104},
  [12308] = {.lex_state = 222, .external_lex_state = 106},
  [12309] = {.lex_state = 138, .external_lex_state = 79},
  [12310] = {.lex_state = 208},
  [12311] = {.lex_state = 203, .external_lex_state = 78},
  [12312] = {.lex_state = 141, .external_lex_state = 80},
  [12313] = {.lex_state = 138, .external_lex_state = 95},
  [12314] = {.lex_state = 209, .external_lex_state = 103},
  [12315] = {.lex_state = 203},
  [12316] = {.lex_state = 138, .external_lex_state = 79},
  [12317] = {.lex_state = 138, .external_lex_state = 79},
  [12318] = {.lex_state = 203, .external_lex_state = 78},
  [12319] = {.lex_state = 141, .external_lex_state = 84},
  [12320] = {.lex_state = 138, .external_lex_state = 99},
  [12321] = {.lex_state = 141, .external_lex_state = 83},
  [12322] = {.lex_state = 208, .external_lex_state = 99},
  [12323] = {.lex_state = 203, .external_lex_state = 78},
  [12324] = {.lex_state = 455, .external_lex_state = 99},
  [12325] = {.lex_state = 203, .external_lex_state = 78},
  [12326] = {.lex_state = 209, .external_lex_state = 104},
  [12327] = {.lex_state = 203, .external_lex_state = 78},
  [12328] = {.lex_state = 209, .external_lex_state = 104},
  [12329] = {.lex_state = 138, .external_lex_state = 95},
  [12330] = {.lex_state = 208, .external_lex_state = 100},
  [12331] = {.lex_state = 141, .external_lex_state = 80},
  [12332] = {.lex_state = 203, .external_lex_state = 78},
  [12333] = {.lex_state = 203, .external_lex_state = 79},
  [12334] = {.lex_state = 141, .external_lex_state = 80},
  [12335] = {.lex_state = 208, .external_lex_state = 100},
  [12336] = {.lex_state = 209, .external_lex_state = 103},
  [12337] = {.lex_state = 209, .external_lex_state = 103},
  [12338] = {.lex_state = 141, .external_lex_state = 81},
  [12339] = {.lex_state = 138, .external_lex_state = 87},
  [12340] = {.lex_state = 203, .external_lex_state = 78},
  [12341] = {.lex_state = 138, .external_lex_state = 79},
  [12342] = {.lex_state = 141, .external_lex_state = 80},
  [12343] = {.lex_state = 203, .external_lex_state = 78},
  [12344] = {.lex_state = 203, .external_lex_state = 78},
  [12345] = {.lex_state = 203, .external_lex_state = 78},
  [12346] = {.lex_state = 141, .external_lex_state = 107},
  [12347] = {.lex_state = 141, .external_lex_state = 84},
  [12348] = {.lex_state = 455, .external_lex_state = 99},
  [12349] = {.lex_state = 203, .external_lex_state = 78},
  [12350] = {.lex_state = 208, .external_lex_state = 101},
  [12351] = {.lex_state = 203, .external_lex_state = 78},
  [12352] = {.lex_state = 208, .external_lex_state = 101},
  [12353] = {.lex_state = 138, .external_lex_state = 79},
  [12354] = {.lex_state = 222, .external_lex_state = 106},
  [12355] = {.lex_state = 141, .external_lex_state = 81},
  [12356] = {.lex_state = 138, .external_lex_state = 87},
  [12357] = {.lex_state = 203, .external_lex_state = 78},
  [12358] = {.lex_state = 203, .external_lex_state = 78},
  [12359] = {.lex_state = 138, .external_lex_state = 95},
  [12360] = {.lex_state = 141, .external_lex_state = 84},
  [12361] = {.lex_state = 203, .external_lex_state = 78},
  [12362] = {.lex_state = 203, .external_lex_state = 78},
  [12363] = {.lex_state = 138, .external_lex_state = 79},
  [12364] = {.lex_state = 203, .external_lex_state = 78},
  [12365] = {.lex_state = 203, .external_lex_state = 78},
  [12366] = {.lex_state = 222, .external_lex_state = 106},
  [12367] = {.lex_state = 141, .external_lex_state = 82},
  [12368] = {.lex_state = 203, .external_lex_state = 78},
  [12369] = {.lex_state = 203, .external_lex_state = 78},
  [12370] = {.lex_state = 203, .external_lex_state = 78},
  [12371] = {.lex_state = 203, .external_lex_state = 78},
  [12372] = {.lex_state = 203, .external_lex_state = 78},
  [12373] = {.lex_state = 138, .external_lex_state = 87},
  [12374] = {.lex_state = 141, .external_lex_state = 83},
  [12375] = {.lex_state = 138, .external_lex_state = 99},
  [12376] = {.lex_state = 203, .external_lex_state = 79},
  [12377] = {.lex_state = 208, .external_lex_state = 101},
  [12378] = {.lex_state = 138, .external_lex_state = 79},
  [12379] = {.lex_state = 203, .external_lex_state = 78},
  [12380] = {.lex_state = 203, .external_lex_state = 78},
  [12381] = {.lex_state = 138, .external_lex_state = 99},
  [12382] = {.lex_state = 208, .external_lex_state = 101},
  [12383] = {.lex_state = 208, .external_lex_state = 100},
  [12384] = {.lex_state = 203, .external_lex_state = 78},
  [12385] = {.lex_state = 141, .external_lex_state = 81},
  [12386] = {.lex_state = 138, .external_lex_state = 95},
  [12387] = {.lex_state = 222, .external_lex_state = 106},
  [12388] = {.lex_state = 203, .external_lex_state = 78},
  [12389] = {.lex_state = 203},
  [12390] = {.lex_state = 222, .external_lex_state = 106},
  [12391] = {.lex_state = 138, .external_lex_state = 79},
  [12392] = {.lex_state = 209, .external_lex_state = 103},
  [12393] = {.lex_state = 203, .external_lex_state = 78},
  [12394] = {.lex_state = 203, .external_lex_state = 78},
  [12395] = {.lex_state = 203, .external_lex_state = 78},
  [12396] = {.lex_state = 222, .external_lex_state = 106},
  [12397] = {.lex_state = 203, .external_lex_state = 78},
  [12398] = {.lex_state = 141, .external_lex_state = 83},
  [12399] = {.lex_state = 208, .external_lex_state = 101},
  [12400] = {.lex_state = 203, .external_lex_state = 78},
  [12401] = {.lex_state = 203, .external_lex_state = 78},
  [12402] = {.lex_state = 138, .external_lex_state = 79},
  [12403] = {.lex_state = 141, .external_lex_state = 84},
  [12404] = {.lex_state = 203, .external_lex_state = 79},
  [12405] = {.lex_state = 138, .external_lex_state = 79},
  [12406] = {.lex_state = 141, .external_lex_state = 84},
  [12407] = {.lex_state = 203, .external_lex_state = 79},
  [12408] = {.lex_state = 141, .external_lex_state = 83},
  [12409] = {.lex_state = 203, .external_lex_state = 78},
  [12410] = {.lex_state = 203, .external_lex_state = 78},
  [12411] = {.lex_state = 138, .external_lex_state = 95},
  [12412] = {.lex_state = 141, .external_lex_state = 83},
  [12413] = {.lex_state = 141, .external_lex_state = 84},
  [12414] = {.lex_state = 138, .external_lex_state = 79},
  [12415] = {.lex_state = 203, .external_lex_state = 78},
  [12416] = {.lex_state = 203, .external_lex_state = 78},
  [12417] = {.lex_state = 141, .external_lex_state = 82},
  [12418] = {.lex_state = 141, .external_lex_state = 83},
  [12419] = {.lex_state = 209, .external_lex_state = 104},
  [12420] = {.lex_state = 141, .external_lex_state = 83},
  [12421] = {.lex_state = 203, .external_lex_state = 78},
  [12422] = {.lex_state = 208, .external_lex_state = 100},
  [12423] = {.lex_state = 209, .external_lex_state = 104},
  [12424] = {.lex_state = 455, .external_lex_state = 99},
  [12425] = {.lex_state = 203, .external_lex_state = 78},
  [12426] = {.lex_state = 141, .external_lex_state = 83},
  [12427] = {.lex_state = 141, .external_lex_state = 107},
  [12428] = {.lex_state = 222, .external_lex_state = 106},
  [12429] = {.lex_state = 203, .external_lex_state = 78},
  [12430] = {.lex_state = 141, .external_lex_state = 82},
  [12431] = {.lex_state = 203, .external_lex_state = 78},
  [12432] = {.lex_state = 208, .external_lex_state = 100},
  [12433] = {.lex_state = 208, .external_lex_state = 101},
  [12434] = {.lex_state = 203, .external_lex_state = 78},
  [12435] = {.lex_state = 209, .external_lex_state = 104},
  [12436] = {.lex_state = 203, .external_lex_state = 78},
  [12437] = {.lex_state = 222, .external_lex_state = 106},
  [12438] = {.lex_state = 208, .external_lex_state = 100},
  [12439] = {.lex_state = 203, .external_lex_state = 78},
  [12440] = {.lex_state = 203, .external_lex_state = 79},
  [12441] = {.lex_state = 209, .external_lex_state = 103},
  [12442] = {.lex_state = 203, .external_lex_state = 78},
  [12443] = {.lex_state = 203, .external_lex_state = 78},
  [12444] = {.lex_state = 203, .external_lex_state = 78},
  [12445] = {.lex_state = 222, .external_lex_state = 106},
  [12446] = {.lex_state = 138, .external_lex_state = 87},
  [12447] = {.lex_state = 138, .external_lex_state = 79},
  [12448] = {.lex_state = 138, .external_lex_state = 99},
  [12449] = {.lex_state = 141, .external_lex_state = 83},
  [12450] = {.lex_state = 203, .external_lex_state = 78},
  [12451] = {.lex_state = 203, .external_lex_state = 78},
  [12452] = {.lex_state = 203, .external_lex_state = 78},
  [12453] = {.lex_state = 141, .external_lex_state = 105},
  [12454] = {.lex_state = 203, .external_lex_state = 78},
  [12455] = {.lex_state = 141, .external_lex_state = 83},
  [12456] = {.lex_state = 224, .external_lex_state = 108},
  [12457] = {.lex_state = 203, .external_lex_state = 78},
  [12458] = {.lex_state = 203, .external_lex_state = 78},
  [12459] = {.lex_state = 141, .external_lex_state = 107},
  [12460] = {.lex_state = 209, .external_lex_state = 103},
  [12461] = {.lex_state = 141, .external_lex_state = 105},
  [12462] = {.lex_state = 203, .external_lex_state = 79},
  [12463] = {.lex_state = 138, .external_lex_state = 79},
  [12464] = {.lex_state = 203, .external_lex_state = 78},
  [12465] = {.lex_state = 203, .external_lex_state = 78},
  [12466] = {.lex_state = 138, .external_lex_state = 79},
  [12467] = {.lex_state = 203, .external_lex_state = 78},
  [12468] = {.lex_state = 203, .external_lex_state = 78},
  [12469] = {.lex_state = 138, .external_lex_state = 99},
  [12470] = {.lex_state = 209, .external_lex_state = 104},
  [12471] = {.lex_state = 138, .external_lex_state = 95},
  [12472] = {.lex_state = 141, .external_lex_state = 84},
  [12473] = {.lex_state = 209, .external_lex_state = 104},
  [12474] = {.lex_state = 138, .external_lex_state = 87},
  [12475] = {.lex_state = 203, .external_lex_state = 86},
  [12476] = {.lex_state = 141, .external_lex_state = 84},
  [12477] = {.lex_state = 203, .external_lex_state = 79},
  [12478] = {.lex_state = 203, .external_lex_state = 79},
  [12479] = {.lex_state = 217, .external_lex_state = 109},
  [12480] = {.lex_state = 209, .external_lex_state = 102},
  [12481] = {.lex_state = 210, .external_lex_state = 101},
  [12482] = {.lex_state = 209, .external_lex_state = 102},
  [12483] = {.lex_state = 141, .external_lex_state = 110},
  [12484] = {.lex_state = 141, .external_lex_state = 82},
  [12485] = {.lex_state = 210, .external_lex_state = 101},
  [12486] = {.lex_state = 141, .external_lex_state = 82},
  [12487] = {.lex_state = 141, .external_lex_state = 82},
  [12488] = {.lex_state = 141, .external_lex_state = 82},
  [12489] = {.lex_state = 203, .external_lex_state = 79},
  [12490] = {.lex_state = 141, .external_lex_state = 110},
  [12491] = {.lex_state = 141, .external_lex_state = 85},
  [12492] = {.lex_state = 141, .external_lex_state = 85},
  [12493] = {.lex_state = 217, .external_lex_state = 109},
  [12494] = {.lex_state = 141, .external_lex_state = 85},
  [12495] = {.lex_state = 141, .external_lex_state = 85},
  [12496] = {.lex_state = 141, .external_lex_state = 85},
  [12497] = {.lex_state = 141, .external_lex_state = 85},
  [12498] = {.lex_state = 141, .external_lex_state = 85},
  [12499] = {.lex_state = 141, .external_lex_state = 85},
  [12500] = {.lex_state = 210, .external_lex_state = 101},
  [12501] = {.lex_state = 210, .external_lex_state = 101},
  [12502] = {.lex_state = 210, .external_lex_state = 101},
  [12503] = {.lex_state = 208, .external_lex_state = 100},
  [12504] = {.lex_state = 208, .external_lex_state = 100},
  [12505] = {.lex_state = 141, .external_lex_state = 85},
  [12506] = {.lex_state = 141, .external_lex_state = 85},
  [12507] = {.lex_state = 203, .external_lex_state = 111},
  [12508] = {.lex_state = 141, .external_lex_state = 85},
  [12509] = {.lex_state = 141, .external_lex_state = 83},
  [12510] = {.lex_state = 141, .external_lex_state = 83},
  [12511] = {.lex_state = 141, .external_lex_state = 83},
  [12512] = {.lex_state = 141, .external_lex_state = 84},
  [12513] = {.lex_state = 141, .external_lex_state = 83},
  [12514] = {.lex_state = 141, .external_lex_state = 84},
  [12515] = {.lex_state = 141, .external_lex_state = 83},
  [12516] = {.lex_state = 141, .external_lex_state = 84},
  [12517] = {.lex_state = 141, .external_lex_state = 84},
  [12518] = {.lex_state = 141, .external_lex_state = 84},
  [12519] = {.lex_state = 141, .external_lex_state = 83},
  [12520] = {.lex_state = 141, .external_lex_state = 84},
  [12521] = {.lex_state = 141, .external_lex_state = 83},
  [12522] = {.lex_state = 141, .external_lex_state = 83},
  [12523] = {.lex_state = 141, .external_lex_state = 84},
  [12524] = {.lex_state = 141, .external_lex_state = 84},
  [12525] = {.lex_state = 141, .external_lex_state = 83},
  [12526] = {.lex_state = 141, .external_lex_state = 83},
  [12527] = {.lex_state = 141, .external_lex_state = 84},
  [12528] = {.lex_state = 141, .external_lex_state = 84},
  [12529] = {.lex_state = 141, .external_lex_state = 84},
  [12530] = {.lex_state = 141, .external_lex_state = 83},
  [12531] = {.lex_state = 141, .external_lex_state = 83},
  [12532] = {.lex_state = 210, .external_lex_state = 100},
  [12533] = {.lex_state = 141, .external_lex_state = 84},
  [12534] = {.lex_state = 141, .external_lex_state = 84},
  [12535] = {.lex_state = 141, .external_lex_state = 84},
  [12536] = {.lex_state = 141, .external_lex_state = 84},
  [12537] = {.lex_state = 141, .external_lex_state = 84},
  [12538] = {.lex_state = 141, .external_lex_state = 83},
  [12539] = {.lex_state = 141, .external_lex_state = 84},
  [12540] = {.lex_state = 210, .external_lex_state = 100},
  [12541] = {.lex_state = 141, .external_lex_state = 84},
  [12542] = {.lex_state = 141, .external_lex_state = 84},
  [12543] = {.lex_state = 141, .external_lex_state = 84},
  [12544] = {.lex_state = 141, .external_lex_state = 84},
  [12545] = {.lex_state = 141, .external_lex_state = 84},
  [12546] = {.lex_state = 141, .external_lex_state = 84},
  [12547] = {.lex_state = 141, .external_lex_state = 84},
  [12548] = {.lex_state = 141, .external_lex_state = 84},
  [12549] = {.lex_state = 141, .external_lex_state = 84},
  [12550] = {.lex_state = 141, .external_lex_state = 84},
  [12551] = {.lex_state = 141, .external_lex_state = 84},
  [12552] = {.lex_state = 210, .external_lex_state = 101},
  [12553] = {.lex_state = 210, .external_lex_state = 101},
  [12554] = {.lex_state = 141, .external_lex_state = 84},
  [12555] = {.lex_state = 141, .external_lex_state = 84},
  [12556] = {.lex_state = 208, .external_lex_state = 101},
  [12557] = {.lex_state = 208, .external_lex_state = 101},
  [12558] = {.lex_state = 217, .external_lex_state = 109},
  [12559] = {.lex_state = 217, .external_lex_state = 109},
  [12560] = {.lex_state = 208, .external_lex_state = 100},
  [12561] = {.lex_state = 217, .external_lex_state = 109},
  [12562] = {.lex_state = 217, .external_lex_state = 109},
  [12563] = {.lex_state = 217, .external_lex_state = 109},
  [12564] = {.lex_state = 141, .external_lex_state = 83},
  [12565] = {.lex_state = 208},
  [12566] = {.lex_state = 208, .external_lex_state = 101},
  [12567] = {.lex_state = 141, .external_lex_state = 112},
  [12568] = {.lex_state = 208, .external_lex_state = 100},
  [12569] = {.lex_state = 138, .external_lex_state = 87},
  [12570] = {.lex_state = 138, .external_lex_state = 87},
  [12571] = {.lex_state = 138, .external_lex_state = 87},
  [12572] = {.lex_state = 138, .external_lex_state = 87},
  [12573] = {.lex_state = 138, .external_lex_state = 87},
  [12574] = {.lex_state = 138, .external_lex_state = 87},
  [12575] = {.lex_state = 138, .external_lex_state = 87},
  [12576] = {.lex_state = 138, .external_lex_state = 87},
  [12577] = {.lex_state = 138, .external_lex_state = 87},
  [12578] = {.lex_state = 138, .external_lex_state = 87},
  [12579] = {.lex_state = 138, .external_lex_state = 87},
  [12580] = {.lex_state = 208},
  [12581] = {.lex_state = 141, .external_lex_state = 84},
  [12582] = {.lex_state = 141, .external_lex_state = 112},
  [12583] = {.lex_state = 141, .external_lex_state = 84},
  [12584] = {.lex_state = 138, .external_lex_state = 99},
  [12585] = {.lex_state = 141, .external_lex_state = 84},
  [12586] = {.lex_state = 208, .external_lex_state = 100},
  [12587] = {.lex_state = 209, .external_lex_state = 109},
  [12588] = {.lex_state = 209, .external_lex_state = 109},
  [12589] = {.lex_state = 138, .external_lex_state = 87},
  [12590] = {.lex_state = 138, .external_lex_state = 87},
  [12591] = {.lex_state = 138, .external_lex_state = 87},
  [12592] = {.lex_state = 141, .external_lex_state = 84},
  [12593] = {.lex_state = 141, .external_lex_state = 83},
  [12594] = {.lex_state = 141, .external_lex_state = 84},
  [12595] = {.lex_state = 141, .external_lex_state = 112},
  [12596] = {.lex_state = 209, .external_lex_state = 109},
  [12597] = {.lex_state = 209, .external_lex_state = 109},
  [12598] = {.lex_state = 209, .external_lex_state = 109},
  [12599] = {.lex_state = 210, .external_lex_state = 101},
  [12600] = {.lex_state = 141, .external_lex_state = 83},
  [12601] = {.lex_state = 141, .external_lex_state = 83},
  [12602] = {.lex_state = 217, .external_lex_state = 109},
  [12603] = {.lex_state = 217, .external_lex_state = 109},
  [12604] = {.lex_state = 141, .external_lex_state = 83},
  [12605] = {.lex_state = 209, .external_lex_state = 109},
  [12606] = {.lex_state = 141, .external_lex_state = 83},
  [12607] = {.lex_state = 141, .external_lex_state = 83},
  [12608] = {.lex_state = 141, .external_lex_state = 83},
  [12609] = {.lex_state = 141, .external_lex_state = 83},
  [12610] = {.lex_state = 141, .external_lex_state = 83},
  [12611] = {.lex_state = 141, .external_lex_state = 83},
  [12612] = {.lex_state = 209, .external_lex_state = 109},
  [12613] = {.lex_state = 209, .external_lex_state = 102},
  [12614] = {.lex_state = 209, .external_lex_state = 102},
  [12615] = {.lex_state = 141, .external_lex_state = 83},
  [12616] = {.lex_state = 141, .external_lex_state = 83},
  [12617] = {.lex_state = 138, .external_lex_state = 95},
  [12618] = {.lex_state = 138, .external_lex_state = 95},
  [12619] = {.lex_state = 141, .external_lex_state = 84},
  [12620] = {.lex_state = 138, .external_lex_state = 95},
  [12621] = {.lex_state = 138, .external_lex_state = 95},
  [12622] = {.lex_state = 203, .external_lex_state = 111},
  [12623] = {.lex_state = 208},
  [12624] = {.lex_state = 138, .external_lex_state = 95},
  [12625] = {.lex_state = 141, .external_lex_state = 83},
  [12626] = {.lex_state = 138, .external_lex_state = 95},
  [12627] = {.lex_state = 138, .external_lex_state = 95},
  [12628] = {.lex_state = 138, .external_lex_state = 95},
  [12629] = {.lex_state = 138, .external_lex_state = 95},
  [12630] = {.lex_state = 141, .external_lex_state = 83},
  [12631] = {.lex_state = 141, .external_lex_state = 83},
  [12632] = {.lex_state = 209, .external_lex_state = 104},
  [12633] = {.lex_state = 209, .external_lex_state = 104},
  [12634] = {.lex_state = 209, .external_lex_state = 102},
  [12635] = {.lex_state = 209, .external_lex_state = 102},
  [12636] = {.lex_state = 209, .external_lex_state = 102},
  [12637] = {.lex_state = 141, .external_lex_state = 83},
  [12638] = {.lex_state = 141, .external_lex_state = 83},
  [12639] = {.lex_state = 141, .external_lex_state = 83},
  [12640] = {.lex_state = 141, .external_lex_state = 83},
  [12641] = {.lex_state = 141, .external_lex_state = 83},
  [12642] = {.lex_state = 141, .external_lex_state = 83},
  [12643] = {.lex_state = 203, .external_lex_state = 77},
  [12644] = {.lex_state = 141, .external_lex_state = 83},
  [12645] = {.lex_state = 203, .external_lex_state = 77},
  [12646] = {.lex_state = 141, .external_lex_state = 83},
  [12647] = {.lex_state = 141, .external_lex_state = 83},
  [12648] = {.lex_state = 141, .external_lex_state = 83},
  [12649] = {.lex_state = 141, .external_lex_state = 83},
  [12650] = {.lex_state = 141, .external_lex_state = 83},
  [12651] = {.lex_state = 141, .external_lex_state = 83},
  [12652] = {.lex_state = 141, .external_lex_state = 83},
  [12653] = {.lex_state = 141, .external_lex_state = 83},
  [12654] = {.lex_state = 141, .external_lex_state = 83},
  [12655] = {.lex_state = 209, .external_lex_state = 109},
  [12656] = {.lex_state = 208},
  [12657] = {.lex_state = 209, .external_lex_state = 109},
  [12658] = {.lex_state = 138, .external_lex_state = 95},
  [12659] = {.lex_state = 138, .external_lex_state = 95},
  [12660] = {.lex_state = 138, .external_lex_state = 95},
  [12661] = {.lex_state = 208},
  [12662] = {.lex_state = 138, .external_lex_state = 99},
  [12663] = {.lex_state = 138, .external_lex_state = 95},
  [12664] = {.lex_state = 209, .external_lex_state = 104},
  [12665] = {.lex_state = 141, .external_lex_state = 84},
  [12666] = {.lex_state = 138, .external_lex_state = 99},
  [12667] = {.lex_state = 141, .external_lex_state = 83},
  [12668] = {.lex_state = 208},
  [12669] = {.lex_state = 208},
  [12670] = {.lex_state = 208},
  [12671] = {.lex_state = 138, .external_lex_state = 99},
  [12672] = {.lex_state = 210, .external_lex_state = 100},
  [12673] = {.lex_state = 203, .external_lex_state = 77},
  [12674] = {.lex_state = 203, .external_lex_state = 77},
  [12675] = {.lex_state = 208, .external_lex_state = 101},
  [12676] = {.lex_state = 203, .external_lex_state = 111},
  [12677] = {.lex_state = 208, .external_lex_state = 101},
  [12678] = {.lex_state = 141, .external_lex_state = 83},
  [12679] = {.lex_state = 209, .external_lex_state = 102},
  [12680] = {.lex_state = 209, .external_lex_state = 102},
  [12681] = {.lex_state = 203, .external_lex_state = 86},
  [12682] = {.lex_state = 203, .external_lex_state = 86},
  [12683] = {.lex_state = 141, .external_lex_state = 83},
  [12684] = {.lex_state = 203, .external_lex_state = 86},
  [12685] = {.lex_state = 203, .external_lex_state = 86},
  [12686] = {.lex_state = 203, .external_lex_state = 86},
  [12687] = {.lex_state = 141, .external_lex_state = 84},
  [12688] = {.lex_state = 203, .external_lex_state = 86},
  [12689] = {.lex_state = 203, .external_lex_state = 86},
  [12690] = {.lex_state = 141, .external_lex_state = 110},
  [12691] = {.lex_state = 141, .external_lex_state = 84},
  [12692] = {.lex_state = 138},
  [12693] = {.lex_state = 209, .external_lex_state = 103},
  [12694] = {.lex_state = 209, .external_lex_state = 103},
  [12695] = {.lex_state = 208, .external_lex_state = 99},
  [12696] = {.lex_state = 208, .external_lex_state = 99},
  [12697] = {.lex_state = 208, .external_lex_state = 99},
  [12698] = {.lex_state = 208, .external_lex_state = 99},
  [12699] = {.lex_state = 208, .external_lex_state = 99},
  [12700] = {.lex_state = 209, .external_lex_state = 103},
  [12701] = {.lex_state = 210, .external_lex_state = 100},
  [12702] = {.lex_state = 138},
  [12703] = {.lex_state = 210, .external_lex_state = 100},
  [12704] = {.lex_state = 210, .external_lex_state = 100},
  [12705] = {.lex_state = 210, .external_lex_state = 100},
  [12706] = {.lex_state = 210, .external_lex_state = 100},
  [12707] = {.lex_state = 209, .external_lex_state = 104},
  [12708] = {.lex_state = 209, .external_lex_state = 104},
  [12709] = {.lex_state = 141, .external_lex_state = 83},
  [12710] = {.lex_state = 138, .external_lex_state = 99},
  [12711] = {.lex_state = 209, .external_lex_state = 103},
  [12712] = {.lex_state = 209, .external_lex_state = 103},
  [12713] = {.lex_state = 141, .external_lex_state = 83},
  [12714] = {.lex_state = 207},
  [12715] = {.lex_state = 141, .external_lex_state = 84},
  [12716] = {.lex_state = 141, .external_lex_state = 84},
  [12717] = {.lex_state = 141, .external_lex_state = 84},
  [12718] = {.lex_state = 141, .external_lex_state = 84},
  [12719] = {.lex_state = 226, .external_lex_state = 97},
  [12720] = {.lex_state = 141, .external_lex_state = 83},
  [12721] = {.lex_state = 141, .external_lex_state = 84},
  [12722] = {.lex_state = 141, .external_lex_state = 84},
  [12723] = {.lex_state = 226, .external_lex_state = 102},
  [12724] = {.lex_state = 141, .external_lex_state = 84},
  [12725] = {.lex_state = 141, .external_lex_state = 84},
  [12726] = {.lex_state = 141, .external_lex_state = 84},
  [12727] = {.lex_state = 208, .external_lex_state = 99},
  [12728] = {.lex_state = 138, .external_lex_state = 95},
  [12729] = {.lex_state = 138, .external_lex_state = 95},
  [12730] = {.lex_state = 141, .external_lex_state = 84},
  [12731] = {.lex_state = 141, .external_lex_state = 84},
  [12732] = {.lex_state = 208, .external_lex_state = 99},
  [12733] = {.lex_state = 138, .external_lex_state = 95},
  [12734] = {.lex_state = 141, .external_lex_state = 85},
  [12735] = {.lex_state = 209, .external_lex_state = 104},
  [12736] = {.lex_state = 209, .external_lex_state = 104},
  [12737] = {.lex_state = 208},
  [12738] = {.lex_state = 203, .external_lex_state = 86},
  [12739] = {.lex_state = 203, .external_lex_state = 86},
  [12740] = {.lex_state = 218, .external_lex_state = 109},
  [12741] = {.lex_state = 203, .external_lex_state = 86},
  [12742] = {.lex_state = 141, .external_lex_state = 85},
  [12743] = {.lex_state = 203, .external_lex_state = 113},
  [12744] = {.lex_state = 218, .external_lex_state = 109},
  [12745] = {.lex_state = 141, .external_lex_state = 112},
  [12746] = {.lex_state = 141, .external_lex_state = 85},
  [12747] = {.lex_state = 203, .external_lex_state = 113},
  [12748] = {.lex_state = 203, .external_lex_state = 86},
  [12749] = {.lex_state = 203, .external_lex_state = 79},
  [12750] = {.lex_state = 203, .external_lex_state = 79},
  [12751] = {.lex_state = 141, .external_lex_state = 85},
  [12752] = {.lex_state = 141, .external_lex_state = 85},
  [12753] = {.lex_state = 208, .external_lex_state = 99},
  [12754] = {.lex_state = 208, .external_lex_state = 99},
  [12755] = {.lex_state = 141, .external_lex_state = 85},
  [12756] = {.lex_state = 203, .external_lex_state = 79},
  [12757] = {.lex_state = 203, .external_lex_state = 79},
  [12758] = {.lex_state = 217, .external_lex_state = 109},
  [12759] = {.lex_state = 141, .external_lex_state = 85},
  [12760] = {.lex_state = 203, .external_lex_state = 113},
  [12761] = {.lex_state = 217, .external_lex_state = 109},
  [12762] = {.lex_state = 218, .external_lex_state = 109},
  [12763] = {.lex_state = 203, .external_lex_state = 87},
  [12764] = {.lex_state = 203, .external_lex_state = 87},
  [12765] = {.lex_state = 203, .external_lex_state = 86},
  [12766] = {.lex_state = 218, .external_lex_state = 109},
  [12767] = {.lex_state = 203, .external_lex_state = 87},
  [12768] = {.lex_state = 203, .external_lex_state = 87},
  [12769] = {.lex_state = 203, .external_lex_state = 87},
  [12770] = {.lex_state = 203, .external_lex_state = 87},
  [12771] = {.lex_state = 208, .external_lex_state = 101},
  [12772] = {.lex_state = 141, .external_lex_state = 85},
  [12773] = {.lex_state = 203, .external_lex_state = 87},
  [12774] = {.lex_state = 203, .external_lex_state = 87},
  [12775] = {.lex_state = 203, .external_lex_state = 86},
  [12776] = {.lex_state = 141, .external_lex_state = 85},
  [12777] = {.lex_state = 141, .external_lex_state = 85},
  [12778] = {.lex_state = 141, .external_lex_state = 85},
  [12779] = {.lex_state = 203, .external_lex_state = 86},
  [12780] = {.lex_state = 210, .external_lex_state = 100},
  [12781] = {.lex_state = 208, .external_lex_state = 100},
  [12782] = {.lex_state = 210},
  [12783] = {.lex_state = 203, .external_lex_state = 86},
  [12784] = {.lex_state = 141, .external_lex_state = 85},
  [12785] = {.lex_state = 141, .external_lex_state = 85},
  [12786] = {.lex_state = 141, .external_lex_state = 85},
  [12787] = {.lex_state = 210, .external_lex_state = 100},
  [12788] = {.lex_state = 141, .external_lex_state = 85},
  [12789] = {.lex_state = 210},
  [12790] = {.lex_state = 203, .external_lex_state = 86},
  [12791] = {.lex_state = 141, .external_lex_state = 85},
  [12792] = {.lex_state = 141, .external_lex_state = 85},
  [12793] = {.lex_state = 217, .external_lex_state = 109},
  [12794] = {.lex_state = 209, .external_lex_state = 104},
  [12795] = {.lex_state = 141, .external_lex_state = 85},
  [12796] = {.lex_state = 141, .external_lex_state = 85},
  [12797] = {.lex_state = 203, .external_lex_state = 86},
  [12798] = {.lex_state = 226, .external_lex_state = 108},
  [12799] = {.lex_state = 210, .external_lex_state = 100},
  [12800] = {.lex_state = 208},
  [12801] = {.lex_state = 141, .external_lex_state = 85},
  [12802] = {.lex_state = 210},
  [12803] = {.lex_state = 210},
  [12804] = {.lex_state = 210},
  [12805] = {.lex_state = 455},
  [12806] = {.lex_state = 208, .external_lex_state = 101},
  [12807] = {.lex_state = 203, .external_lex_state = 86},
  [12808] = {.lex_state = 203, .external_lex_state = 86},
  [12809] = {.lex_state = 141, .external_lex_state = 85},
  [12810] = {.lex_state = 210, .external_lex_state = 100},
  [12811] = {.lex_state = 203, .external_lex_state = 86},
  [12812] = {.lex_state = 207},
  [12813] = {.lex_state = 141, .external_lex_state = 85},
  [12814] = {.lex_state = 209, .external_lex_state = 109},
  [12815] = {.lex_state = 209, .external_lex_state = 109},
  [12816] = {.lex_state = 203, .external_lex_state = 86},
  [12817] = {.lex_state = 210, .external_lex_state = 101},
  [12818] = {.lex_state = 208, .external_lex_state = 101},
  [12819] = {.lex_state = 210, .external_lex_state = 101},
  [12820] = {.lex_state = 209, .external_lex_state = 103},
  [12821] = {.lex_state = 207},
  [12822] = {.lex_state = 138, .external_lex_state = 99},
  [12823] = {.lex_state = 138, .external_lex_state = 99},
  [12824] = {.lex_state = 141, .external_lex_state = 112},
  [12825] = {.lex_state = 209, .external_lex_state = 103},
  [12826] = {.lex_state = 209, .external_lex_state = 103},
  [12827] = {.lex_state = 208, .external_lex_state = 101},
  [12828] = {.lex_state = 141, .external_lex_state = 112},
  [12829] = {.lex_state = 138, .external_lex_state = 99},
  [12830] = {.lex_state = 138, .external_lex_state = 99},
  [12831] = {.lex_state = 138, .external_lex_state = 99},
  [12832] = {.lex_state = 138, .external_lex_state = 99},
  [12833] = {.lex_state = 203, .external_lex_state = 86},
  [12834] = {.lex_state = 203, .external_lex_state = 86},
  [12835] = {.lex_state = 210, .external_lex_state = 101},
  [12836] = {.lex_state = 141, .external_lex_state = 112},
  [12837] = {.lex_state = 210, .external_lex_state = 101},
  [12838] = {.lex_state = 138, .external_lex_state = 99},
  [12839] = {.lex_state = 138, .external_lex_state = 99},
  [12840] = {.lex_state = 210, .external_lex_state = 100},
  [12841] = {.lex_state = 141, .external_lex_state = 112},
  [12842] = {.lex_state = 209, .external_lex_state = 104},
  [12843] = {.lex_state = 203, .external_lex_state = 86},
  [12844] = {.lex_state = 203, .external_lex_state = 86},
  [12845] = {.lex_state = 208, .external_lex_state = 99},
  [12846] = {.lex_state = 208, .external_lex_state = 101},
  [12847] = {.lex_state = 138, .external_lex_state = 99},
  [12848] = {.lex_state = 209, .external_lex_state = 109},
  [12849] = {.lex_state = 210},
  [12850] = {.lex_state = 138, .external_lex_state = 99},
  [12851] = {.lex_state = 138, .external_lex_state = 99},
  [12852] = {.lex_state = 203, .external_lex_state = 86},
  [12853] = {.lex_state = 203, .external_lex_state = 86},
  [12854] = {.lex_state = 203, .external_lex_state = 86},
  [12855] = {.lex_state = 209, .external_lex_state = 102},
  [12856] = {.lex_state = 203, .external_lex_state = 86},
  [12857] = {.lex_state = 203, .external_lex_state = 86},
  [12858] = {.lex_state = 209, .external_lex_state = 102},
  [12859] = {.lex_state = 208},
  [12860] = {.lex_state = 218, .external_lex_state = 109},
  [12861] = {.lex_state = 218, .external_lex_state = 109},
  [12862] = {.lex_state = 455},
  [12863] = {.lex_state = 203, .external_lex_state = 86},
  [12864] = {.lex_state = 203, .external_lex_state = 86},
  [12865] = {.lex_state = 203, .external_lex_state = 86},
  [12866] = {.lex_state = 203, .external_lex_state = 86},
  [12867] = {.lex_state = 203, .external_lex_state = 86},
  [12868] = {.lex_state = 203, .external_lex_state = 86},
  [12869] = {.lex_state = 455},
  [12870] = {.lex_state = 203, .external_lex_state = 86},
  [12871] = {.lex_state = 203, .external_lex_state = 86},
  [12872] = {.lex_state = 217, .external_lex_state = 109},
  [12873] = {.lex_state = 203, .external_lex_state = 86},
  [12874] = {.lex_state = 141, .external_lex_state = 85},
  [12875] = {.lex_state = 203, .external_lex_state = 86},
  [12876] = {.lex_state = 203, .external_lex_state = 86},
  [12877] = {.lex_state = 141, .external_lex_state = 112},
  [12878] = {.lex_state = 210, .external_lex_state = 101},
  [12879] = {.lex_state = 455},
  [12880] = {.lex_state = 217, .external_lex_state = 109},
  [12881] = {.lex_state = 209, .external_lex_state = 102},
  [12882] = {.lex_state = 455},
  [12883] = {.lex_state = 210, .external_lex_state = 99},
  [12884] = {.lex_state = 209, .external_lex_state = 103},
  [12885] = {.lex_state = 141, .external_lex_state = 85},
  [12886] = {.lex_state = 141, .external_lex_state = 85},
  [12887] = {.lex_state = 455},
  [12888] = {.lex_state = 210, .external_lex_state = 99},
  [12889] = {.lex_state = 210, .external_lex_state = 99},
  [12890] = {.lex_state = 210, .external_lex_state = 99},
  [12891] = {.lex_state = 210, .external_lex_state = 99},
  [12892] = {.lex_state = 455},
  [12893] = {.lex_state = 203, .external_lex_state = 86},
  [12894] = {.lex_state = 210},
  [12895] = {.lex_state = 455},
  [12896] = {.lex_state = 210},
  [12897] = {.lex_state = 208, .external_lex_state = 100},
  [12898] = {.lex_state = 208},
  [12899] = {.lex_state = 138, .external_lex_state = 99},
  [12900] = {.lex_state = 455},
  [12901] = {.lex_state = 138, .external_lex_state = 99},
  [12902] = {.lex_state = 138, .external_lex_state = 99},
  [12903] = {.lex_state = 217, .external_lex_state = 109},
  [12904] = {.lex_state = 224, .external_lex_state = 88},
  [12905] = {.lex_state = 455},
  [12906] = {.lex_state = 217, .external_lex_state = 109},
  [12907] = {.lex_state = 207},
  [12908] = {.lex_state = 455},
  [12909] = {.lex_state = 455},
  [12910] = {.lex_state = 203},
  [12911] = {.lex_state = 141, .external_lex_state = 85},
  [12912] = {.lex_state = 455},
  [12913] = {.lex_state = 203, .external_lex_state = 86},
  [12914] = {.lex_state = 208, .external_lex_state = 100},
  [12915] = {.lex_state = 203, .external_lex_state = 86},
  [12916] = {.lex_state = 207},
  [12917] = {.lex_state = 141, .external_lex_state = 85},
  [12918] = {.lex_state = 208},
  [12919] = {.lex_state = 141, .external_lex_state = 85},
  [12920] = {.lex_state = 207},
  [12921] = {.lex_state = 207},
  [12922] = {.lex_state = 207},
  [12923] = {.lex_state = 203},
  [12924] = {.lex_state = 203, .external_lex_state = 86},
  [12925] = {.lex_state = 224, .external_lex_state = 88},
  [12926] = {.lex_state = 209, .external_lex_state = 104},
  [12927] = {.lex_state = 218, .external_lex_state = 109},
  [12928] = {.lex_state = 141, .external_lex_state = 85},
  [12929] = {.lex_state = 141, .external_lex_state = 85},
  [12930] = {.lex_state = 141, .external_lex_state = 85},
  [12931] = {.lex_state = 209, .external_lex_state = 109},
  [12932] = {.lex_state = 209, .external_lex_state = 109},
  [12933] = {.lex_state = 203, .external_lex_state = 86},
  [12934] = {.lex_state = 218, .external_lex_state = 109},
  [12935] = {.lex_state = 218, .external_lex_state = 109},
  [12936] = {.lex_state = 141, .external_lex_state = 85},
  [12937] = {.lex_state = 141, .external_lex_state = 85},
  [12938] = {.lex_state = 209, .external_lex_state = 102},
  [12939] = {.lex_state = 203},
  [12940] = {.lex_state = 209, .external_lex_state = 102},
  [12941] = {.lex_state = 208, .external_lex_state = 100},
  [12942] = {.lex_state = 203, .external_lex_state = 86},
  [12943] = {.lex_state = 209, .external_lex_state = 103},
  [12944] = {.lex_state = 203, .external_lex_state = 86},
  [12945] = {.lex_state = 203, .external_lex_state = 86},
  [12946] = {.lex_state = 141, .external_lex_state = 85},
  [12947] = {.lex_state = 141, .external_lex_state = 85},
  [12948] = {.lex_state = 141, .external_lex_state = 85},
  [12949] = {.lex_state = 141, .external_lex_state = 85},
  [12950] = {.lex_state = 141, .external_lex_state = 85},
  [12951] = {.lex_state = 141, .external_lex_state = 85},
  [12952] = {.lex_state = 203, .external_lex_state = 86},
  [12953] = {.lex_state = 208, .external_lex_state = 100},
  [12954] = {.lex_state = 203, .external_lex_state = 86},
  [12955] = {.lex_state = 203, .external_lex_state = 86},
  [12956] = {.lex_state = 141, .external_lex_state = 85},
  [12957] = {.lex_state = 141, .external_lex_state = 85},
  [12958] = {.lex_state = 141, .external_lex_state = 85},
  [12959] = {.lex_state = 141, .external_lex_state = 85},
  [12960] = {.lex_state = 141, .external_lex_state = 85},
  [12961] = {.lex_state = 141, .external_lex_state = 85},
  [12962] = {.lex_state = 141, .external_lex_state = 85},
  [12963] = {.lex_state = 224, .external_lex_state = 114},
  [12964] = {.lex_state = 203, .external_lex_state = 86},
  [12965] = {.lex_state = 141, .external_lex_state = 85},
  [12966] = {.lex_state = 203, .external_lex_state = 86},
  [12967] = {.lex_state = 208},
  [12968] = {.lex_state = 141, .external_lex_state = 85},
  [12969] = {.lex_state = 141, .external_lex_state = 85},
  [12970] = {.lex_state = 210, .external_lex_state = 99},
  [12971] = {.lex_state = 210, .external_lex_state = 99},
  [12972] = {.lex_state = 203, .external_lex_state = 86},
  [12973] = {.lex_state = 210, .external_lex_state = 99},
  [12974] = {.lex_state = 208, .external_lex_state = 99},
  [12975] = {.lex_state = 141, .external_lex_state = 85},
  [12976] = {.lex_state = 141, .external_lex_state = 85},
  [12977] = {.lex_state = 203, .external_lex_state = 86},
  [12978] = {.lex_state = 208, .external_lex_state = 99},
  [12979] = {.lex_state = 203, .external_lex_state = 86},
  [12980] = {.lex_state = 208, .external_lex_state = 101},
  [12981] = {.lex_state = 208},
  [12982] = {.lex_state = 208, .external_lex_state = 99},
  [12983] = {.lex_state = 208, .external_lex_state = 101},
  [12984] = {.lex_state = 143, .external_lex_state = 115},
  [12985] = {.lex_state = 208, .external_lex_state = 100},
  [12986] = {.lex_state = 143, .external_lex_state = 115},
  [12987] = {.lex_state = 143, .external_lex_state = 115},
  [12988] = {.lex_state = 143, .external_lex_state = 115},
  [12989] = {.lex_state = 209, .external_lex_state = 103},
  [12990] = {.lex_state = 207},
  [12991] = {.lex_state = 455},
  [12992] = {.lex_state = 207},
  [12993] = {.lex_state = 203, .external_lex_state = 87},
  [12994] = {.lex_state = 203, .external_lex_state = 87},
  [12995] = {.lex_state = 143, .external_lex_state = 115},
  [12996] = {.lex_state = 217, .external_lex_state = 109},
  [12997] = {.lex_state = 208, .external_lex_state = 100},
  [12998] = {.lex_state = 208, .external_lex_state = 101},
  [12999] = {.lex_state = 209, .external_lex_state = 103},
  [13000] = {.lex_state = 209, .external_lex_state = 103},
  [13001] = {.lex_state = 209, .external_lex_state = 103},
  [13002] = {.lex_state = 209, .external_lex_state = 103},
  [13003] = {.lex_state = 143, .external_lex_state = 115},
  [13004] = {.lex_state = 209, .external_lex_state = 103},
  [13005] = {.lex_state = 209, .external_lex_state = 103},
  [13006] = {.lex_state = 143, .external_lex_state = 115},
  [13007] = {.lex_state = 209, .external_lex_state = 103},
  [13008] = {.lex_state = 209, .external_lex_state = 103},
  [13009] = {.lex_state = 209, .external_lex_state = 103},
  [13010] = {.lex_state = 209, .external_lex_state = 103},
  [13011] = {.lex_state = 143, .external_lex_state = 115},
  [13012] = {.lex_state = 209, .external_lex_state = 104},
  [13013] = {.lex_state = 216},
  [13014] = {.lex_state = 143, .external_lex_state = 115},
  [13015] = {.lex_state = 203, .external_lex_state = 87},
  [13016] = {.lex_state = 143, .external_lex_state = 115},
  [13017] = {.lex_state = 203, .external_lex_state = 87},
  [13018] = {.lex_state = 143, .external_lex_state = 115},
  [13019] = {.lex_state = 203, .external_lex_state = 87},
  [13020] = {.lex_state = 143, .external_lex_state = 115},
  [13021] = {.lex_state = 210, .external_lex_state = 101},
  [13022] = {.lex_state = 208, .external_lex_state = 101},
  [13023] = {.lex_state = 210, .external_lex_state = 99},
  [13024] = {.lex_state = 143, .external_lex_state = 115},
  [13025] = {.lex_state = 203, .external_lex_state = 87},
  [13026] = {.lex_state = 208},
  [13027] = {.lex_state = 210, .external_lex_state = 101},
  [13028] = {.lex_state = 143, .external_lex_state = 115},
  [13029] = {.lex_state = 208},
  [13030] = {.lex_state = 208, .external_lex_state = 101},
  [13031] = {.lex_state = 203, .external_lex_state = 87},
  [13032] = {.lex_state = 203, .external_lex_state = 87},
  [13033] = {.lex_state = 203, .external_lex_state = 87},
  [13034] = {.lex_state = 143, .external_lex_state = 115},
  [13035] = {.lex_state = 138},
  [13036] = {.lex_state = 203, .external_lex_state = 87},
  [13037] = {.lex_state = 203, .external_lex_state = 87},
  [13038] = {.lex_state = 143, .external_lex_state = 115},
  [13039] = {.lex_state = 208, .external_lex_state = 101},
  [13040] = {.lex_state = 217, .external_lex_state = 109},
  [13041] = {.lex_state = 203, .external_lex_state = 87},
  [13042] = {.lex_state = 143, .external_lex_state = 115},
  [13043] = {.lex_state = 143, .external_lex_state = 115},
  [13044] = {.lex_state = 218, .external_lex_state = 109},
  [13045] = {.lex_state = 216},
  [13046] = {.lex_state = 216},
  [13047] = {.lex_state = 216},
  [13048] = {.lex_state = 203, .external_lex_state = 87},
  [13049] = {.lex_state = 143, .external_lex_state = 115},
  [13050] = {.lex_state = 203},
  [13051] = {.lex_state = 217, .external_lex_state = 109},
  [13052] = {.lex_state = 143, .external_lex_state = 115},
  [13053] = {.lex_state = 203, .external_lex_state = 87},
  [13054] = {.lex_state = 209, .external_lex_state = 103},
  [13055] = {.lex_state = 203, .external_lex_state = 87},
  [13056] = {.lex_state = 209, .external_lex_state = 103},
  [13057] = {.lex_state = 209, .external_lex_state = 103},
  [13058] = {.lex_state = 209, .external_lex_state = 102},
  [13059] = {.lex_state = 203, .external_lex_state = 87},
  [13060] = {.lex_state = 143, .external_lex_state = 115},
  [13061] = {.lex_state = 143, .external_lex_state = 115},
  [13062] = {.lex_state = 143, .external_lex_state = 115},
  [13063] = {.lex_state = 218, .external_lex_state = 109},
  [13064] = {.lex_state = 203, .external_lex_state = 87},
  [13065] = {.lex_state = 217, .external_lex_state = 109},
  [13066] = {.lex_state = 138},
  [13067] = {.lex_state = 210, .external_lex_state = 101},
  [13068] = {.lex_state = 143, .external_lex_state = 115},
  [13069] = {.lex_state = 143, .external_lex_state = 115},
  [13070] = {.lex_state = 203, .external_lex_state = 87},
  [13071] = {.lex_state = 203, .external_lex_state = 87},
  [13072] = {.lex_state = 143, .external_lex_state = 115},
  [13073] = {.lex_state = 203, .external_lex_state = 87},
  [13074] = {.lex_state = 143, .external_lex_state = 115},
  [13075] = {.lex_state = 203, .external_lex_state = 87},
  [13076] = {.lex_state = 203, .external_lex_state = 87},
  [13077] = {.lex_state = 203, .external_lex_state = 87},
  [13078] = {.lex_state = 143, .external_lex_state = 115},
  [13079] = {.lex_state = 143, .external_lex_state = 115},
  [13080] = {.lex_state = 143, .external_lex_state = 115},
  [13081] = {.lex_state = 203, .external_lex_state = 87},
  [13082] = {.lex_state = 203, .external_lex_state = 87},
  [13083] = {.lex_state = 143, .external_lex_state = 115},
  [13084] = {.lex_state = 203, .external_lex_state = 87},
  [13085] = {.lex_state = 143, .external_lex_state = 115},
  [13086] = {.lex_state = 203, .external_lex_state = 87},
  [13087] = {.lex_state = 203, .external_lex_state = 87},
  [13088] = {.lex_state = 203, .external_lex_state = 87},
  [13089] = {.lex_state = 143, .external_lex_state = 115},
  [13090] = {.lex_state = 203, .external_lex_state = 87},
  [13091] = {.lex_state = 143, .external_lex_state = 115},
  [13092] = {.lex_state = 203, .external_lex_state = 87},
  [13093] = {.lex_state = 208, .external_lex_state = 101},
  [13094] = {.lex_state = 138},
  [13095] = {.lex_state = 143, .external_lex_state = 115},
  [13096] = {.lex_state = 143, .external_lex_state = 115},
  [13097] = {.lex_state = 224, .external_lex_state = 88},
  [13098] = {.lex_state = 203},
  [13099] = {.lex_state = 216},
  [13100] = {.lex_state = 210, .external_lex_state = 100},
  [13101] = {.lex_state = 207},
  [13102] = {.lex_state = 203, .external_lex_state = 87},
  [13103] = {.lex_state = 143, .external_lex_state = 115},
  [13104] = {.lex_state = 143, .external_lex_state = 115},
  [13105] = {.lex_state = 143, .external_lex_state = 115},
  [13106] = {.lex_state = 209, .external_lex_state = 109},
  [13107] = {.lex_state = 203, .external_lex_state = 87},
  [13108] = {.lex_state = 203, .external_lex_state = 87},
  [13109] = {.lex_state = 203, .external_lex_state = 87},
  [13110] = {.lex_state = 203, .external_lex_state = 87},
  [13111] = {.lex_state = 143, .external_lex_state = 115},
  [13112] = {.lex_state = 143, .external_lex_state = 115},
  [13113] = {.lex_state = 203, .external_lex_state = 87},
  [13114] = {.lex_state = 209, .external_lex_state = 109},
  [13115] = {.lex_state = 143, .external_lex_state = 115},
  [13116] = {.lex_state = 208, .external_lex_state = 99},
  [13117] = {.lex_state = 216},
  [13118] = {.lex_state = 210},
  [13119] = {.lex_state = 203, .external_lex_state = 87},
  [13120] = {.lex_state = 143, .external_lex_state = 115},
  [13121] = {.lex_state = 143, .external_lex_state = 115},
  [13122] = {.lex_state = 208, .external_lex_state = 99},
  [13123] = {.lex_state = 216},
  [13124] = {.lex_state = 210},
  [13125] = {.lex_state = 143, .external_lex_state = 115},
  [13126] = {.lex_state = 143, .external_lex_state = 115},
  [13127] = {.lex_state = 143, .external_lex_state = 115},
  [13128] = {.lex_state = 208, .external_lex_state = 99},
  [13129] = {.lex_state = 143, .external_lex_state = 115},
  [13130] = {.lex_state = 209, .external_lex_state = 102},
  [13131] = {.lex_state = 209, .external_lex_state = 109},
  [13132] = {.lex_state = 143, .external_lex_state = 115},
  [13133] = {.lex_state = 143, .external_lex_state = 115},
  [13134] = {.lex_state = 143, .external_lex_state = 115},
  [13135] = {.lex_state = 143, .external_lex_state = 115},
  [13136] = {.lex_state = 209, .external_lex_state = 104},
  [13137] = {.lex_state = 143, .external_lex_state = 115},
  [13138] = {.lex_state = 210},
  [13139] = {.lex_state = 217, .external_lex_state = 109},
  [13140] = {.lex_state = 143, .external_lex_state = 115},
  [13141] = {.lex_state = 143, .external_lex_state = 115},
  [13142] = {.lex_state = 207},
  [13143] = {.lex_state = 143, .external_lex_state = 115},
  [13144] = {.lex_state = 218, .external_lex_state = 109},
  [13145] = {.lex_state = 218, .external_lex_state = 109},
  [13146] = {.lex_state = 209, .external_lex_state = 102},
  [13147] = {.lex_state = 143, .external_lex_state = 115},
  [13148] = {.lex_state = 207},
  [13149] = {.lex_state = 143, .external_lex_state = 115},
  [13150] = {.lex_state = 143, .external_lex_state = 115},
  [13151] = {.lex_state = 207},
  [13152] = {.lex_state = 143, .external_lex_state = 115},
  [13153] = {.lex_state = 143, .external_lex_state = 115},
  [13154] = {.lex_state = 208, .external_lex_state = 100},
  [13155] = {.lex_state = 143, .external_lex_state = 115},
  [13156] = {.lex_state = 143, .external_lex_state = 115},
  [13157] = {.lex_state = 143, .external_lex_state = 115},
  [13158] = {.lex_state = 143, .external_lex_state = 115},
  [13159] = {.lex_state = 138},
  [13160] = {.lex_state = 208},
  [13161] = {.lex_state = 210, .external_lex_state = 100},
  [13162] = {.lex_state = 143, .external_lex_state = 115},
  [13163] = {.lex_state = 210, .external_lex_state = 100},
  [13164] = {.lex_state = 210, .external_lex_state = 100},
  [13165] = {.lex_state = 203, .external_lex_state = 87},
  [13166] = {.lex_state = 143, .external_lex_state = 115},
  [13167] = {.lex_state = 143, .external_lex_state = 115},
  [13168] = {.lex_state = 203, .external_lex_state = 87},
  [13169] = {.lex_state = 143, .external_lex_state = 115},
  [13170] = {.lex_state = 203, .external_lex_state = 87},
  [13171] = {.lex_state = 208, .external_lex_state = 101},
  [13172] = {.lex_state = 209, .external_lex_state = 104},
  [13173] = {.lex_state = 216},
  [13174] = {.lex_state = 224, .external_lex_state = 88},
  [13175] = {.lex_state = 143, .external_lex_state = 115},
  [13176] = {.lex_state = 143, .external_lex_state = 115},
  [13177] = {.lex_state = 143, .external_lex_state = 115},
  [13178] = {.lex_state = 208, .external_lex_state = 101},
  [13179] = {.lex_state = 143, .external_lex_state = 115},
  [13180] = {.lex_state = 208},
  [13181] = {.lex_state = 143, .external_lex_state = 115},
  [13182] = {.lex_state = 209, .external_lex_state = 109},
  [13183] = {.lex_state = 208, .external_lex_state = 101},
  [13184] = {.lex_state = 203, .external_lex_state = 87},
  [13185] = {.lex_state = 208, .external_lex_state = 100},
  [13186] = {.lex_state = 203},
  [13187] = {.lex_state = 143, .external_lex_state = 115},
  [13188] = {.lex_state = 143, .external_lex_state = 115},
  [13189] = {.lex_state = 143, .external_lex_state = 115},
  [13190] = {.lex_state = 143, .external_lex_state = 115},
  [13191] = {.lex_state = 209, .external_lex_state = 104},
  [13192] = {.lex_state = 209, .external_lex_state = 104},
  [13193] = {.lex_state = 209, .external_lex_state = 104},
  [13194] = {.lex_state = 209, .external_lex_state = 102},
  [13195] = {.lex_state = 209, .external_lex_state = 104},
  [13196] = {.lex_state = 138},
  [13197] = {.lex_state = 208, .external_lex_state = 100},
  [13198] = {.lex_state = 208, .external_lex_state = 100},
  [13199] = {.lex_state = 208, .external_lex_state = 100},
  [13200] = {.lex_state = 209, .external_lex_state = 104},
  [13201] = {.lex_state = 203, .external_lex_state = 87},
  [13202] = {.lex_state = 208, .external_lex_state = 100},
  [13203] = {.lex_state = 216},
  [13204] = {.lex_state = 209, .external_lex_state = 104},
  [13205] = {.lex_state = 203, .external_lex_state = 87},
  [13206] = {.lex_state = 203, .external_lex_state = 87},
  [13207] = {.lex_state = 203, .external_lex_state = 87},
  [13208] = {.lex_state = 143, .external_lex_state = 115},
  [13209] = {.lex_state = 143, .external_lex_state = 115},
  [13210] = {.lex_state = 208, .external_lex_state = 101},
  [13211] = {.lex_state = 143, .external_lex_state = 115},
  [13212] = {.lex_state = 210, .external_lex_state = 99},
  [13213] = {.lex_state = 209, .external_lex_state = 104},
  [13214] = {.lex_state = 209, .external_lex_state = 104},
  [13215] = {.lex_state = 143, .external_lex_state = 115},
  [13216] = {.lex_state = 143, .external_lex_state = 115},
  [13217] = {.lex_state = 203, .external_lex_state = 87},
  [13218] = {.lex_state = 208, .external_lex_state = 100},
  [13219] = {.lex_state = 203, .external_lex_state = 87},
  [13220] = {.lex_state = 208, .external_lex_state = 100},
  [13221] = {.lex_state = 210},
  [13222] = {.lex_state = 218, .external_lex_state = 109},
  [13223] = {.lex_state = 208, .external_lex_state = 101},
  [13224] = {.lex_state = 208, .external_lex_state = 101},
  [13225] = {.lex_state = 208, .external_lex_state = 100},
  [13226] = {.lex_state = 143, .external_lex_state = 115},
  [13227] = {.lex_state = 210, .external_lex_state = 99},
  [13228] = {.lex_state = 143, .external_lex_state = 115},
  [13229] = {.lex_state = 210},
  [13230] = {.lex_state = 143, .external_lex_state = 115},
  [13231] = {.lex_state = 209, .external_lex_state = 104},
  [13232] = {.lex_state = 207},
  [13233] = {.lex_state = 143, .external_lex_state = 115},
  [13234] = {.lex_state = 143, .external_lex_state = 115},
  [13235] = {.lex_state = 203, .external_lex_state = 87},
  [13236] = {.lex_state = 143, .external_lex_state = 115},
  [13237] = {.lex_state = 208, .external_lex_state = 100},
  [13238] = {.lex_state = 143, .external_lex_state = 115},
  [13239] = {.lex_state = 208, .external_lex_state = 100},
  [13240] = {.lex_state = 143, .external_lex_state = 115},
  [13241] = {.lex_state = 210, .external_lex_state = 100},
  [13242] = {.lex_state = 203, .external_lex_state = 87},
  [13243] = {.lex_state = 143, .external_lex_state = 115},
  [13244] = {.lex_state = 143, .external_lex_state = 115},
  [13245] = {.lex_state = 210, .external_lex_state = 99},
  [13246] = {.lex_state = 143, .external_lex_state = 115},
  [13247] = {.lex_state = 210, .external_lex_state = 99},
  [13248] = {.lex_state = 208, .external_lex_state = 101},
  [13249] = {.lex_state = 143, .external_lex_state = 115},
  [13250] = {.lex_state = 208, .external_lex_state = 99},
  [13251] = {.lex_state = 207},
  [13252] = {.lex_state = 208, .external_lex_state = 100},
  [13253] = {.lex_state = 224, .external_lex_state = 88},
  [13254] = {.lex_state = 216},
  [13255] = {.lex_state = 143, .external_lex_state = 115},
  [13256] = {.lex_state = 209, .external_lex_state = 109},
  [13257] = {.lex_state = 216},
  [13258] = {.lex_state = 209, .external_lex_state = 104},
  [13259] = {.lex_state = 143, .external_lex_state = 115},
  [13260] = {.lex_state = 209, .external_lex_state = 104},
  [13261] = {.lex_state = 210, .external_lex_state = 101},
  [13262] = {.lex_state = 210, .external_lex_state = 101},
  [13263] = {.lex_state = 143, .external_lex_state = 115},
  [13264] = {.lex_state = 143, .external_lex_state = 115},
  [13265] = {.lex_state = 209, .external_lex_state = 102},
  [13266] = {.lex_state = 224, .external_lex_state = 109},
  [13267] = {.lex_state = 208},
  [13268] = {.lex_state = 203},
  [13269] = {.lex_state = 207},
  [13270] = {.lex_state = 203},
  [13271] = {.lex_state = 217, .external_lex_state = 109},
  [13272] = {.lex_state = 217, .external_lex_state = 109},
  [13273] = {.lex_state = 217, .external_lex_state = 109},
  [13274] = {.lex_state = 203},
  [13275] = {.lex_state = 207},
  [13276] = {.lex_state = 210, .external_lex_state = 101},
  [13277] = {.lex_state = 210, .external_lex_state = 101},
  [13278] = {.lex_state = 203},
  [13279] = {.lex_state = 224, .external_lex_state = 109},
  [13280] = {.lex_state = 209, .external_lex_state = 102},
  [13281] = {.lex_state = 209, .external_lex_state = 102},
  [13282] = {.lex_state = 209, .external_lex_state = 102},
  [13283] = {.lex_state = 208},
  [13284] = {.lex_state = 210, .external_lex_state = 101},
  [13285] = {.lex_state = 208},
  [13286] = {.lex_state = 208},
  [13287] = {.lex_state = 224, .external_lex_state = 109},
  [13288] = {.lex_state = 208},
  [13289] = {.lex_state = 203},
  [13290] = {.lex_state = 210, .external_lex_state = 101},
  [13291] = {.lex_state = 210},
  [13292] = {.lex_state = 209, .external_lex_state = 109},
  [13293] = {.lex_state = 209, .external_lex_state = 109},
  [13294] = {.lex_state = 209, .external_lex_state = 109},
  [13295] = {.lex_state = 209, .external_lex_state = 109},
  [13296] = {.lex_state = 209, .external_lex_state = 109},
  [13297] = {.lex_state = 209, .external_lex_state = 109},
  [13298] = {.lex_state = 208},
  [13299] = {.lex_state = 203},
  [13300] = {.lex_state = 209, .external_lex_state = 109},
  [13301] = {.lex_state = 209, .external_lex_state = 109},
  [13302] = {.lex_state = 208},
  [13303] = {.lex_state = 138},
  [13304] = {.lex_state = 209, .external_lex_state = 109},
  [13305] = {.lex_state = 209, .external_lex_state = 109},
  [13306] = {.lex_state = 209, .external_lex_state = 109},
  [13307] = {.lex_state = 216},
  [13308] = {.lex_state = 203},
  [13309] = {.lex_state = 224, .external_lex_state = 109},
  [13310] = {.lex_state = 208},
  [13311] = {.lex_state = 210, .external_lex_state = 100},
  [13312] = {.lex_state = 210, .external_lex_state = 101},
  [13313] = {.lex_state = 210, .external_lex_state = 101},
  [13314] = {.lex_state = 210},
  [13315] = {.lex_state = 224, .external_lex_state = 109},
  [13316] = {.lex_state = 208},
  [13317] = {.lex_state = 210, .external_lex_state = 101},
  [13318] = {.lex_state = 210, .external_lex_state = 99},
  [13319] = {.lex_state = 210},
  [13320] = {.lex_state = 210, .external_lex_state = 101},
  [13321] = {.lex_state = 216},
  [13322] = {.lex_state = 224, .external_lex_state = 109},
  [13323] = {.lex_state = 207},
  [13324] = {.lex_state = 210, .external_lex_state = 101},
  [13325] = {.lex_state = 224, .external_lex_state = 109},
  [13326] = {.lex_state = 208},
  [13327] = {.lex_state = 203},
  [13328] = {.lex_state = 210, .external_lex_state = 100},
  [13329] = {.lex_state = 208},
  [13330] = {.lex_state = 210, .external_lex_state = 100},
  [13331] = {.lex_state = 210, .external_lex_state = 101},
  [13332] = {.lex_state = 138},
  [13333] = {.lex_state = 210},
  [13334] = {.lex_state = 210, .external_lex_state = 99},
  [13335] = {.lex_state = 210, .external_lex_state = 101},
  [13336] = {.lex_state = 210, .external_lex_state = 101},
  [13337] = {.lex_state = 210, .external_lex_state = 99},
  [13338] = {.lex_state = 210, .external_lex_state = 99},
  [13339] = {.lex_state = 208, .external_lex_state = 99},
  [13340] = {.lex_state = 209, .external_lex_state = 109},
  [13341] = {.lex_state = 208, .external_lex_state = 99},
  [13342] = {.lex_state = 209, .external_lex_state = 109},
  [13343] = {.lex_state = 210, .external_lex_state = 101},
  [13344] = {.lex_state = 208, .external_lex_state = 99},
  [13345] = {.lex_state = 209, .external_lex_state = 109},
  [13346] = {.lex_state = 208},
  [13347] = {.lex_state = 208, .external_lex_state = 99},
  [13348] = {.lex_state = 208, .external_lex_state = 99},
  [13349] = {.lex_state = 216},
  [13350] = {.lex_state = 208, .external_lex_state = 99},
  [13351] = {.lex_state = 218, .external_lex_state = 109},
  [13352] = {.lex_state = 208},
  [13353] = {.lex_state = 208, .external_lex_state = 99},
  [13354] = {.lex_state = 203},
  [13355] = {.lex_state = 208, .external_lex_state = 99},
  [13356] = {.lex_state = 208, .external_lex_state = 99},
  [13357] = {.lex_state = 208, .external_lex_state = 99},
  [13358] = {.lex_state = 138},
  [13359] = {.lex_state = 217, .external_lex_state = 109},
  [13360] = {.lex_state = 210},
  [13361] = {.lex_state = 207},
  [13362] = {.lex_state = 218, .external_lex_state = 109},
  [13363] = {.lex_state = 217, .external_lex_state = 109},
  [13364] = {.lex_state = 217, .external_lex_state = 109},
  [13365] = {.lex_state = 207},
  [13366] = {.lex_state = 217, .external_lex_state = 109},
  [13367] = {.lex_state = 211},
  [13368] = {.lex_state = 210, .external_lex_state = 100},
  [13369] = {.lex_state = 217, .external_lex_state = 109},
  [13370] = {.lex_state = 184, .external_lex_state = 114},
  [13371] = {.lex_state = 211},
  [13372] = {.lex_state = 217, .external_lex_state = 109},
  [13373] = {.lex_state = 207},
  [13374] = {.lex_state = 210, .external_lex_state = 100},
  [13375] = {.lex_state = 210, .external_lex_state = 100},
  [13376] = {.lex_state = 210, .external_lex_state = 100},
  [13377] = {.lex_state = 210, .external_lex_state = 100},
  [13378] = {.lex_state = 211},
  [13379] = {.lex_state = 210, .external_lex_state = 100},
  [13380] = {.lex_state = 217, .external_lex_state = 109},
  [13381] = {.lex_state = 210, .external_lex_state = 100},
  [13382] = {.lex_state = 217, .external_lex_state = 109},
  [13383] = {.lex_state = 217, .external_lex_state = 109},
  [13384] = {.lex_state = 207},
  [13385] = {.lex_state = 217, .external_lex_state = 109},
  [13386] = {.lex_state = 210, .external_lex_state = 100},
  [13387] = {.lex_state = 217, .external_lex_state = 109},
  [13388] = {.lex_state = 210, .external_lex_state = 100},
  [13389] = {.lex_state = 203},
  [13390] = {.lex_state = 218, .external_lex_state = 109},
  [13391] = {.lex_state = 210, .external_lex_state = 100},
  [13392] = {.lex_state = 218, .external_lex_state = 109},
  [13393] = {.lex_state = 208, .external_lex_state = 99},
  [13394] = {.lex_state = 208, .external_lex_state = 99},
  [13395] = {.lex_state = 208, .external_lex_state = 99},
  [13396] = {.lex_state = 218, .external_lex_state = 109},
  [13397] = {.lex_state = 208},
  [13398] = {.lex_state = 209, .external_lex_state = 102},
  [13399] = {.lex_state = 209, .external_lex_state = 102},
  [13400] = {.lex_state = 209, .external_lex_state = 102},
  [13401] = {.lex_state = 209, .external_lex_state = 102},
  [13402] = {.lex_state = 209, .external_lex_state = 102},
  [13403] = {.lex_state = 209, .external_lex_state = 102},
  [13404] = {.lex_state = 209, .external_lex_state = 102},
  [13405] = {.lex_state = 209, .external_lex_state = 102},
  [13406] = {.lex_state = 209, .external_lex_state = 102},
  [13407] = {.lex_state = 203},
  [13408] = {.lex_state = 209, .external_lex_state = 102},
  [13409] = {.lex_state = 209, .external_lex_state = 102},
  [13410] = {.lex_state = 210, .external_lex_state = 101},
  [13411] = {.lex_state = 203},
  [13412] = {.lex_state = 210, .external_lex_state = 100},
  [13413] = {.lex_state = 210, .external_lex_state = 99},
  [13414] = {.lex_state = 203},
  [13415] = {.lex_state = 208, .external_lex_state = 99},
  [13416] = {.lex_state = 198, .external_lex_state = 116},
  [13417] = {.lex_state = 138},
  [13418] = {.lex_state = 224, .external_lex_state = 91},
  [13419] = {.lex_state = 198, .external_lex_state = 116},
  [13420] = {.lex_state = 198, .external_lex_state = 116},
  [13421] = {.lex_state = 198, .external_lex_state = 116},
  [13422] = {.lex_state = 203},
  [13423] = {.lex_state = 216},
  [13424] = {.lex_state = 198, .external_lex_state = 116},
  [13425] = {.lex_state = 219, .external_lex_state = 102},
  [13426] = {.lex_state = 198, .external_lex_state = 116},
  [13427] = {.lex_state = 198, .external_lex_state = 116},
  [13428] = {.lex_state = 210},
  [13429] = {.lex_state = 198, .external_lex_state = 116},
  [13430] = {.lex_state = 198, .external_lex_state = 116},
  [13431] = {.lex_state = 210},
  [13432] = {.lex_state = 198, .external_lex_state = 116},
  [13433] = {.lex_state = 224, .external_lex_state = 88},
  [13434] = {.lex_state = 138},
  [13435] = {.lex_state = 198, .external_lex_state = 116},
  [13436] = {.lex_state = 219, .external_lex_state = 102},
  [13437] = {.lex_state = 198, .external_lex_state = 116},
  [13438] = {.lex_state = 210},
  [13439] = {.lex_state = 198, .external_lex_state = 116},
  [13440] = {.lex_state = 198, .external_lex_state = 116},
  [13441] = {.lex_state = 198, .external_lex_state = 116},
  [13442] = {.lex_state = 198, .external_lex_state = 116},
  [13443] = {.lex_state = 198, .external_lex_state = 116},
  [13444] = {.lex_state = 203},
  [13445] = {.lex_state = 198, .external_lex_state = 116},
  [13446] = {.lex_state = 198, .external_lex_state = 116},
  [13447] = {.lex_state = 198, .external_lex_state = 116},
  [13448] = {.lex_state = 210},
  [13449] = {.lex_state = 224, .external_lex_state = 88},
  [13450] = {.lex_state = 224, .external_lex_state = 88},
  [13451] = {.lex_state = 224, .external_lex_state = 88},
  [13452] = {.lex_state = 224, .external_lex_state = 88},
  [13453] = {.lex_state = 218, .external_lex_state = 109},
  [13454] = {.lex_state = 224, .external_lex_state = 88},
  [13455] = {.lex_state = 198, .external_lex_state = 116},
  [13456] = {.lex_state = 210},
  [13457] = {.lex_state = 198, .external_lex_state = 116},
  [13458] = {.lex_state = 210, .external_lex_state = 99},
  [13459] = {.lex_state = 218, .external_lex_state = 109},
  [13460] = {.lex_state = 210, .external_lex_state = 99},
  [13461] = {.lex_state = 198, .external_lex_state = 116},
  [13462] = {.lex_state = 216},
  [13463] = {.lex_state = 198, .external_lex_state = 116},
  [13464] = {.lex_state = 210, .external_lex_state = 99},
  [13465] = {.lex_state = 210, .external_lex_state = 99},
  [13466] = {.lex_state = 210, .external_lex_state = 99},
  [13467] = {.lex_state = 210, .external_lex_state = 99},
  [13468] = {.lex_state = 210},
  [13469] = {.lex_state = 210, .external_lex_state = 99},
  [13470] = {.lex_state = 218, .external_lex_state = 109},
  [13471] = {.lex_state = 198, .external_lex_state = 116},
  [13472] = {.lex_state = 210, .external_lex_state = 99},
  [13473] = {.lex_state = 210, .external_lex_state = 99},
  [13474] = {.lex_state = 224, .external_lex_state = 88},
  [13475] = {.lex_state = 210, .external_lex_state = 99},
  [13476] = {.lex_state = 224, .external_lex_state = 88},
  [13477] = {.lex_state = 210, .external_lex_state = 99},
  [13478] = {.lex_state = 224, .external_lex_state = 88},
  [13479] = {.lex_state = 224, .external_lex_state = 88},
  [13480] = {.lex_state = 198, .external_lex_state = 116},
  [13481] = {.lex_state = 198, .external_lex_state = 116},
  [13482] = {.lex_state = 224, .external_lex_state = 88},
  [13483] = {.lex_state = 198, .external_lex_state = 116},
  [13484] = {.lex_state = 210},
  [13485] = {.lex_state = 198, .external_lex_state = 116},
  [13486] = {.lex_state = 210},
  [13487] = {.lex_state = 198, .external_lex_state = 116},
  [13488] = {.lex_state = 224, .external_lex_state = 88},
  [13489] = {.lex_state = 198, .external_lex_state = 116},
  [13490] = {.lex_state = 198, .external_lex_state = 116},
  [13491] = {.lex_state = 210},
  [13492] = {.lex_state = 198, .external_lex_state = 116},
  [13493] = {.lex_state = 198, .external_lex_state = 116},
  [13494] = {.lex_state = 198, .external_lex_state = 116},
  [13495] = {.lex_state = 224, .external_lex_state = 88},
  [13496] = {.lex_state = 224, .external_lex_state = 88},
  [13497] = {.lex_state = 224, .external_lex_state = 88},
  [13498] = {.lex_state = 224, .external_lex_state = 88},
  [13499] = {.lex_state = 198, .external_lex_state = 116},
  [13500] = {.lex_state = 224, .external_lex_state = 88},
  [13501] = {.lex_state = 224, .external_lex_state = 88},
  [13502] = {.lex_state = 198, .external_lex_state = 116},
  [13503] = {.lex_state = 224, .external_lex_state = 88},
  [13504] = {.lex_state = 224, .external_lex_state = 88},
  [13505] = {.lex_state = 198, .external_lex_state = 116},
  [13506] = {.lex_state = 224, .external_lex_state = 88},
  [13507] = {.lex_state = 198, .external_lex_state = 116},
  [13508] = {.lex_state = 138},
  [13509] = {.lex_state = 207},
  [13510] = {.lex_state = 198, .external_lex_state = 116},
  [13511] = {.lex_state = 207},
  [13512] = {.lex_state = 203},
  [13513] = {.lex_state = 207},
  [13514] = {.lex_state = 224, .external_lex_state = 88},
  [13515] = {.lex_state = 198, .external_lex_state = 116},
  [13516] = {.lex_state = 224, .external_lex_state = 88},
  [13517] = {.lex_state = 210},
  [13518] = {.lex_state = 198, .external_lex_state = 116},
  [13519] = {.lex_state = 198, .external_lex_state = 116},
  [13520] = {.lex_state = 224, .external_lex_state = 88},
  [13521] = {.lex_state = 224},
  [13522] = {.lex_state = 224, .external_lex_state = 88},
  [13523] = {.lex_state = 198, .external_lex_state = 116},
  [13524] = {.lex_state = 198, .external_lex_state = 116},
  [13525] = {.lex_state = 198, .external_lex_state = 116},
  [13526] = {.lex_state = 224, .external_lex_state = 88},
  [13527] = {.lex_state = 198, .external_lex_state = 116},
  [13528] = {.lex_state = 207},
  [13529] = {.lex_state = 198, .external_lex_state = 116},
  [13530] = {.lex_state = 207},
  [13531] = {.lex_state = 207},
  [13532] = {.lex_state = 198, .external_lex_state = 116},
  [13533] = {.lex_state = 198, .external_lex_state = 116},
  [13534] = {.lex_state = 216},
  [13535] = {.lex_state = 198, .external_lex_state = 116},
  [13536] = {.lex_state = 224, .external_lex_state = 91},
  [13537] = {.lex_state = 207},
  [13538] = {.lex_state = 211},
  [13539] = {.lex_state = 198, .external_lex_state = 116},
  [13540] = {.lex_state = 198, .external_lex_state = 116},
  [13541] = {.lex_state = 210},
  [13542] = {.lex_state = 224, .external_lex_state = 88},
  [13543] = {.lex_state = 224, .external_lex_state = 88},
  [13544] = {.lex_state = 207},
  [13545] = {.lex_state = 198, .external_lex_state = 116},
  [13546] = {.lex_state = 210},
  [13547] = {.lex_state = 207},
  [13548] = {.lex_state = 210},
  [13549] = {.lex_state = 198, .external_lex_state = 116},
  [13550] = {.lex_state = 224, .external_lex_state = 88},
  [13551] = {.lex_state = 224, .external_lex_state = 88},
  [13552] = {.lex_state = 210, .external_lex_state = 99},
  [13553] = {.lex_state = 210, .external_lex_state = 99},
  [13554] = {.lex_state = 198, .external_lex_state = 116},
  [13555] = {.lex_state = 218, .external_lex_state = 109},
  [13556] = {.lex_state = 210, .external_lex_state = 99},
  [13557] = {.lex_state = 198, .external_lex_state = 116},
  [13558] = {.lex_state = 198, .external_lex_state = 116},
  [13559] = {.lex_state = 218, .external_lex_state = 109},
  [13560] = {.lex_state = 218, .external_lex_state = 109},
  [13561] = {.lex_state = 218, .external_lex_state = 109},
  [13562] = {.lex_state = 198, .external_lex_state = 116},
  [13563] = {.lex_state = 218, .external_lex_state = 109},
  [13564] = {.lex_state = 207},
  [13565] = {.lex_state = 198, .external_lex_state = 116},
  [13566] = {.lex_state = 198, .external_lex_state = 116},
  [13567] = {.lex_state = 198, .external_lex_state = 116},
  [13568] = {.lex_state = 218, .external_lex_state = 109},
  [13569] = {.lex_state = 218, .external_lex_state = 109},
  [13570] = {.lex_state = 224, .external_lex_state = 88},
  [13571] = {.lex_state = 198, .external_lex_state = 116},
  [13572] = {.lex_state = 198, .external_lex_state = 116},
  [13573] = {.lex_state = 198, .external_lex_state = 116},
  [13574] = {.lex_state = 216},
  [13575] = {.lex_state = 207},
  [13576] = {.lex_state = 218, .external_lex_state = 109},
  [13577] = {.lex_state = 138},
  [13578] = {.lex_state = 198, .external_lex_state = 116},
  [13579] = {.lex_state = 198, .external_lex_state = 116},
  [13580] = {.lex_state = 138},
  [13581] = {.lex_state = 198, .external_lex_state = 116},
  [13582] = {.lex_state = 138},
  [13583] = {.lex_state = 198, .external_lex_state = 116},
  [13584] = {.lex_state = 218, .external_lex_state = 109},
  [13585] = {.lex_state = 218, .external_lex_state = 109},
  [13586] = {.lex_state = 198, .external_lex_state = 116},
  [13587] = {.lex_state = 219, .external_lex_state = 102},
  [13588] = {.lex_state = 218, .external_lex_state = 109},
  [13589] = {.lex_state = 198, .external_lex_state = 116},
  [13590] = {.lex_state = 198, .external_lex_state = 116},
  [13591] = {.lex_state = 198, .external_lex_state = 116},
  [13592] = {.lex_state = 207},
  [13593] = {.lex_state = 198, .external_lex_state = 116},
  [13594] = {.lex_state = 207},
  [13595] = {.lex_state = 224, .external_lex_state = 88},
  [13596] = {.lex_state = 198, .external_lex_state = 116},
  [13597] = {.lex_state = 198, .external_lex_state = 116},
  [13598] = {.lex_state = 198, .external_lex_state = 116},
  [13599] = {.lex_state = 198, .external_lex_state = 116},
  [13600] = {.lex_state = 224, .external_lex_state = 88},
  [13601] = {.lex_state = 198, .external_lex_state = 116},
  [13602] = {.lex_state = 198, .external_lex_state = 116},
  [13603] = {.lex_state = 210},
  [13604] = {.lex_state = 198, .external_lex_state = 116},
  [13605] = {.lex_state = 198, .external_lex_state = 116},
  [13606] = {.lex_state = 198, .external_lex_state = 116},
  [13607] = {.lex_state = 198, .external_lex_state = 116},
  [13608] = {.lex_state = 198, .external_lex_state = 116},
  [13609] = {.lex_state = 198, .external_lex_state = 116},
  [13610] = {.lex_state = 198, .external_lex_state = 116},
  [13611] = {.lex_state = 224, .external_lex_state = 90},
  [13612] = {.lex_state = 198, .external_lex_state = 116},
  [13613] = {.lex_state = 198, .external_lex_state = 116},
  [13614] = {.lex_state = 198, .external_lex_state = 116},
  [13615] = {.lex_state = 198, .external_lex_state = 116},
  [13616] = {.lex_state = 198, .external_lex_state = 116},
  [13617] = {.lex_state = 224, .external_lex_state = 90},
  [13618] = {.lex_state = 198, .external_lex_state = 116},
  [13619] = {.lex_state = 198, .external_lex_state = 116},
  [13620] = {.lex_state = 198, .external_lex_state = 116},
  [13621] = {.lex_state = 198, .external_lex_state = 116},
  [13622] = {.lex_state = 198, .external_lex_state = 116},
  [13623] = {.lex_state = 207},
  [13624] = {.lex_state = 198, .external_lex_state = 116},
  [13625] = {.lex_state = 198, .external_lex_state = 116},
  [13626] = {.lex_state = 138},
  [13627] = {.lex_state = 210, .external_lex_state = 99},
  [13628] = {.lex_state = 455, .external_lex_state = 99},
  [13629] = {.lex_state = 203},
  [13630] = {.lex_state = 224},
  [13631] = {.lex_state = 216},
  [13632] = {.lex_state = 216},
  [13633] = {.lex_state = 211, .external_lex_state = 100},
  [13634] = {.lex_state = 455, .external_lex_state = 99},
  [13635] = {.lex_state = 455, .external_lex_state = 99},
  [13636] = {.lex_state = 211, .external_lex_state = 101},
  [13637] = {.lex_state = 224},
  [13638] = {.lex_state = 224},
  [13639] = {.lex_state = 203},
  [13640] = {.lex_state = 216},
  [13641] = {.lex_state = 211, .external_lex_state = 100},
  [13642] = {.lex_state = 203},
  [13643] = {.lex_state = 203},
  [13644] = {.lex_state = 224},
  [13645] = {.lex_state = 224, .external_lex_state = 92},
  [13646] = {.lex_state = 216},
  [13647] = {.lex_state = 455, .external_lex_state = 117},
  [13648] = {.lex_state = 224},
  [13649] = {.lex_state = 216},
  [13650] = {.lex_state = 224, .external_lex_state = 91},
  [13651] = {.lex_state = 224},
  [13652] = {.lex_state = 144, .external_lex_state = 116},
  [13653] = {.lex_state = 138, .external_lex_state = 118},
  [13654] = {.lex_state = 138},
  [13655] = {.lex_state = 203},
  [13656] = {.lex_state = 455, .external_lex_state = 95},
  [13657] = {.lex_state = 210, .external_lex_state = 99},
  [13658] = {.lex_state = 219, .external_lex_state = 102},
  [13659] = {.lex_state = 224},
  [13660] = {.lex_state = 455, .external_lex_state = 95},
  [13661] = {.lex_state = 203},
  [13662] = {.lex_state = 224, .external_lex_state = 92},
  [13663] = {.lex_state = 216},
  [13664] = {.lex_state = 144, .external_lex_state = 116},
  [13665] = {.lex_state = 211, .external_lex_state = 101},
  [13666] = {.lex_state = 219, .external_lex_state = 102},
  [13667] = {.lex_state = 138},
  [13668] = {.lex_state = 216},
  [13669] = {.lex_state = 224},
  [13670] = {.lex_state = 203},
  [13671] = {.lex_state = 144, .external_lex_state = 116},
  [13672] = {.lex_state = 219, .external_lex_state = 102},
  [13673] = {.lex_state = 224, .external_lex_state = 91},
  [13674] = {.lex_state = 224},
  [13675] = {.lex_state = 138, .external_lex_state = 118},
  [13676] = {.lex_state = 216},
  [13677] = {.lex_state = 219, .external_lex_state = 102},
  [13678] = {.lex_state = 138, .external_lex_state = 118},
  [13679] = {.lex_state = 216},
  [13680] = {.lex_state = 219, .external_lex_state = 102},
  [13681] = {.lex_state = 216},
  [13682] = {.lex_state = 203},
  [13683] = {.lex_state = 455, .external_lex_state = 95},
  [13684] = {.lex_state = 455, .external_lex_state = 95},
  [13685] = {.lex_state = 211, .external_lex_state = 101},
  [13686] = {.lex_state = 216},
  [13687] = {.lex_state = 455, .external_lex_state = 117},
  [13688] = {.lex_state = 203},
  [13689] = {.lex_state = 455, .external_lex_state = 117},
  [13690] = {.lex_state = 224},
  [13691] = {.lex_state = 203},
  [13692] = {.lex_state = 138},
  [13693] = {.lex_state = 224},
  [13694] = {.lex_state = 216},
  [13695] = {.lex_state = 455, .external_lex_state = 99},
  [13696] = {.lex_state = 224, .external_lex_state = 91},
  [13697] = {.lex_state = 224},
  [13698] = {.lex_state = 203},
  [13699] = {.lex_state = 210, .external_lex_state = 99},
  [13700] = {.lex_state = 224},
  [13701] = {.lex_state = 224},
  [13702] = {.lex_state = 138, .external_lex_state = 118},
  [13703] = {.lex_state = 216},
  [13704] = {.lex_state = 455, .external_lex_state = 99},
  [13705] = {.lex_state = 224},
  [13706] = {.lex_state = 455, .external_lex_state = 99},
  [13707] = {.lex_state = 216},
  [13708] = {.lex_state = 223},
  [13709] = {.lex_state = 203},
  [13710] = {.lex_state = 138, .external_lex_state = 118},
  [13711] = {.lex_state = 224, .external_lex_state = 90},
  [13712] = {.lex_state = 203},
  [13713] = {.lex_state = 455, .external_lex_state = 87},
  [13714] = {.lex_state = 144, .external_lex_state = 116},
  [13715] = {.lex_state = 138, .external_lex_state = 118},
  [13716] = {.lex_state = 223},
  [13717] = {.lex_state = 203},
  [13718] = {.lex_state = 219, .external_lex_state = 102},
  [13719] = {.lex_state = 224},
  [13720] = {.lex_state = 455, .external_lex_state = 87},
  [13721] = {.lex_state = 144, .external_lex_state = 116},
  [13722] = {.lex_state = 203},
  [13723] = {.lex_state = 224},
  [13724] = {.lex_state = 224, .external_lex_state = 90},
  [13725] = {.lex_state = 203},
  [13726] = {.lex_state = 203},
  [13727] = {.lex_state = 203},
  [13728] = {.lex_state = 138, .external_lex_state = 119},
  [13729] = {.lex_state = 203},
  [13730] = {.lex_state = 203},
  [13731] = {.lex_state = 203},
  [13732] = {.lex_state = 211, .external_lex_state = 100},
  [13733] = {.lex_state = 203},
  [13734] = {.lex_state = 203},
  [13735] = {.lex_state = 138, .external_lex_state = 118},
  [13736] = {.lex_state = 203},
  [13737] = {.lex_state = 138, .external_lex_state = 118},
  [13738] = {.lex_state = 203},
  [13739] = {.lex_state = 224},
  [13740] = {.lex_state = 224},
  [13741] = {.lex_state = 224, .external_lex_state = 90},
  [13742] = {.lex_state = 224},
  [13743] = {.lex_state = 224},
  [13744] = {.lex_state = 203},
  [13745] = {.lex_state = 224},
  [13746] = {.lex_state = 203},
  [13747] = {.lex_state = 203},
  [13748] = {.lex_state = 224},
  [13749] = {.lex_state = 455, .external_lex_state = 99},
  [13750] = {.lex_state = 455, .external_lex_state = 99},
  [13751] = {.lex_state = 455, .external_lex_state = 99},
  [13752] = {.lex_state = 224},
  [13753] = {.lex_state = 138, .external_lex_state = 118},
  [13754] = {.lex_state = 224},
  [13755] = {.lex_state = 224},
  [13756] = {.lex_state = 219, .external_lex_state = 109},
  [13757] = {.lex_state = 220},
  [13758] = {.lex_state = 219, .external_lex_state = 109},
  [13759] = {.lex_state = 219, .external_lex_state = 109},
  [13760] = {.lex_state = 455, .external_lex_state = 99},
  [13761] = {.lex_state = 455, .external_lex_state = 99},
  [13762] = {.lex_state = 455, .external_lex_state = 99},
  [13763] = {.lex_state = 220},
  [13764] = {.lex_state = 224},
  [13765] = {.lex_state = 220},
  [13766] = {.lex_state = 220},
  [13767] = {.lex_state = 179},
  [13768] = {.lex_state = 220},
  [13769] = {.lex_state = 455, .external_lex_state = 99},
  [13770] = {.lex_state = 179},
  [13771] = {.lex_state = 220},
  [13772] = {.lex_state = 220},
  [13773] = {.lex_state = 220},
  [13774] = {.lex_state = 455, .external_lex_state = 99},
  [13775] = {.lex_state = 220},
  [13776] = {.lex_state = 220},
  [13777] = {.lex_state = 220},
  [13778] = {.lex_state = 220},
  [13779] = {.lex_state = 220},
  [13780] = {.lex_state = 220},
  [13781] = {.lex_state = 224, .external_lex_state = 103},
  [13782] = {.lex_state = 220},
  [13783] = {.lex_state = 179},
  [13784] = {.lex_state = 455, .external_lex_state = 99},
  [13785] = {.lex_state = 224},
  [13786] = {.lex_state = 455, .external_lex_state = 99},
  [13787] = {.lex_state = 179},
  [13788] = {.lex_state = 220},
  [13789] = {.lex_state = 219, .external_lex_state = 109},
  [13790] = {.lex_state = 224},
  [13791] = {.lex_state = 138},
  [13792] = {.lex_state = 143, .external_lex_state = 115},
  [13793] = {.lex_state = 210, .external_lex_state = 101},
  [13794] = {.lex_state = 138},
  [13795] = {.lex_state = 219, .external_lex_state = 109},
  [13796] = {.lex_state = 138},
  [13797] = {.lex_state = 455, .external_lex_state = 99},
  [13798] = {.lex_state = 210, .external_lex_state = 100},
  [13799] = {.lex_state = 219, .external_lex_state = 109},
  [13800] = {.lex_state = 219, .external_lex_state = 109},
  [13801] = {.lex_state = 179},
  [13802] = {.lex_state = 455, .external_lex_state = 99},
  [13803] = {.lex_state = 455, .external_lex_state = 99},
  [13804] = {.lex_state = 179},
  [13805] = {.lex_state = 455, .external_lex_state = 99},
  [13806] = {.lex_state = 455, .external_lex_state = 99},
  [13807] = {.lex_state = 455, .external_lex_state = 99},
  [13808] = {.lex_state = 455, .external_lex_state = 99},
  [13809] = {.lex_state = 179},
  [13810] = {.lex_state = 220},
  [13811] = {.lex_state = 224, .external_lex_state = 103},
  [13812] = {.lex_state = 138},
  [13813] = {.lex_state = 210, .external_lex_state = 100},
  [13814] = {.lex_state = 455, .external_lex_state = 99},
  [13815] = {.lex_state = 138, .external_lex_state = 114},
  [13816] = {.lex_state = 220},
  [13817] = {.lex_state = 455, .external_lex_state = 99},
  [13818] = {.lex_state = 179},
  [13819] = {.lex_state = 220},
  [13820] = {.lex_state = 455, .external_lex_state = 99},
  [13821] = {.lex_state = 224, .external_lex_state = 104},
  [13822] = {.lex_state = 219, .external_lex_state = 109},
  [13823] = {.lex_state = 220},
  [13824] = {.lex_state = 455, .external_lex_state = 99},
  [13825] = {.lex_state = 138, .external_lex_state = 119},
  [13826] = {.lex_state = 210, .external_lex_state = 99},
  [13827] = {.lex_state = 224},
  [13828] = {.lex_state = 138, .external_lex_state = 119},
  [13829] = {.lex_state = 224, .external_lex_state = 109},
  [13830] = {.lex_state = 224, .external_lex_state = 104},
  [13831] = {.lex_state = 138, .external_lex_state = 119},
  [13832] = {.lex_state = 179},
  [13833] = {.lex_state = 179},
  [13834] = {.lex_state = 220},
  [13835] = {.lex_state = 455, .external_lex_state = 99},
  [13836] = {.lex_state = 220},
  [13837] = {.lex_state = 138, .external_lex_state = 119},
  [13838] = {.lex_state = 224, .external_lex_state = 109},
  [13839] = {.lex_state = 138},
  [13840] = {.lex_state = 179},
  [13841] = {.lex_state = 179},
  [13842] = {.lex_state = 179},
  [13843] = {.lex_state = 179},
  [13844] = {.lex_state = 455, .external_lex_state = 99},
  [13845] = {.lex_state = 210, .external_lex_state = 101},
  [13846] = {.lex_state = 220},
  [13847] = {.lex_state = 179},
  [13848] = {.lex_state = 455, .external_lex_state = 99},
  [13849] = {.lex_state = 455, .external_lex_state = 99},
  [13850] = {.lex_state = 138, .external_lex_state = 119},
  [13851] = {.lex_state = 210, .external_lex_state = 101},
  [13852] = {.lex_state = 220},
  [13853] = {.lex_state = 224},
  [13854] = {.lex_state = 143, .external_lex_state = 115},
  [13855] = {.lex_state = 220},
  [13856] = {.lex_state = 179},
  [13857] = {.lex_state = 455, .external_lex_state = 99},
  [13858] = {.lex_state = 224},
  [13859] = {.lex_state = 138, .external_lex_state = 119},
  [13860] = {.lex_state = 455, .external_lex_state = 99},
  [13861] = {.lex_state = 179},
  [13862] = {.lex_state = 179},
  [13863] = {.lex_state = 220},
  [13864] = {.lex_state = 179},
  [13865] = {.lex_state = 224, .external_lex_state = 92},
  [13866] = {.lex_state = 455, .external_lex_state = 99},
  [13867] = {.lex_state = 455, .external_lex_state = 99},
  [13868] = {.lex_state = 455, .external_lex_state = 99},
  [13869] = {.lex_state = 143, .external_lex_state = 115},
  [13870] = {.lex_state = 220},
  [13871] = {.lex_state = 143, .external_lex_state = 115},
  [13872] = {.lex_state = 224},
  [13873] = {.lex_state = 220},
  [13874] = {.lex_state = 179},
  [13875] = {.lex_state = 224},
  [13876] = {.lex_state = 211, .external_lex_state = 101},
  [13877] = {.lex_state = 179},
  [13878] = {.lex_state = 211, .external_lex_state = 100},
  [13879] = {.lex_state = 455, .external_lex_state = 99},
  [13880] = {.lex_state = 220},
  [13881] = {.lex_state = 224, .external_lex_state = 92},
  [13882] = {.lex_state = 220},
  [13883] = {.lex_state = 179},
  [13884] = {.lex_state = 224, .external_lex_state = 92},
  [13885] = {.lex_state = 219, .external_lex_state = 109},
  [13886] = {.lex_state = 138, .external_lex_state = 114},
  [13887] = {.lex_state = 220},
  [13888] = {.lex_state = 220},
  [13889] = {.lex_state = 138, .external_lex_state = 114},
  [13890] = {.lex_state = 179},
  [13891] = {.lex_state = 220},
  [13892] = {.lex_state = 455, .external_lex_state = 99},
  [13893] = {.lex_state = 224, .external_lex_state = 104},
  [13894] = {.lex_state = 220},
  [13895] = {.lex_state = 224, .external_lex_state = 103},
  [13896] = {.lex_state = 455, .external_lex_state = 99},
  [13897] = {.lex_state = 220},
  [13898] = {.lex_state = 455, .external_lex_state = 99},
  [13899] = {.lex_state = 179},
  [13900] = {.lex_state = 179},
  [13901] = {.lex_state = 220},
  [13902] = {.lex_state = 138, .external_lex_state = 119},
  [13903] = {.lex_state = 224},
  [13904] = {.lex_state = 455, .external_lex_state = 99},
  [13905] = {.lex_state = 138},
  [13906] = {.lex_state = 224, .external_lex_state = 109},
  [13907] = {.lex_state = 224},
  [13908] = {.lex_state = 455, .external_lex_state = 99},
  [13909] = {.lex_state = 455, .external_lex_state = 99},
  [13910] = {.lex_state = 220},
  [13911] = {.lex_state = 179},
  [13912] = {.lex_state = 210, .external_lex_state = 100},
  [13913] = {.lex_state = 179},
  [13914] = {.lex_state = 138},
  [13915] = {.lex_state = 179},
  [13916] = {.lex_state = 220},
  [13917] = {.lex_state = 179},
  [13918] = {.lex_state = 455, .external_lex_state = 99},
  [13919] = {.lex_state = 455, .external_lex_state = 99},
  [13920] = {.lex_state = 455, .external_lex_state = 99},
  [13921] = {.lex_state = 224},
  [13922] = {.lex_state = 179},
  [13923] = {.lex_state = 455},
  [13924] = {.lex_state = 199, .external_lex_state = 116},
  [13925] = {.lex_state = 199, .external_lex_state = 116},
  [13926] = {.lex_state = 224, .external_lex_state = 91},
  [13927] = {.lex_state = 210, .external_lex_state = 101},
  [13928] = {.lex_state = 199, .external_lex_state = 116},
  [13929] = {.lex_state = 224, .external_lex_state = 91},
  [13930] = {.lex_state = 224, .external_lex_state = 91},
  [13931] = {.lex_state = 224, .external_lex_state = 91},
  [13932] = {.lex_state = 224, .external_lex_state = 91},
  [13933] = {.lex_state = 179},
  [13934] = {.lex_state = 199, .external_lex_state = 116},
  [13935] = {.lex_state = 199, .external_lex_state = 116},
  [13936] = {.lex_state = 224, .external_lex_state = 91},
  [13937] = {.lex_state = 224, .external_lex_state = 91},
  [13938] = {.lex_state = 199, .external_lex_state = 116},
  [13939] = {.lex_state = 224, .external_lex_state = 91},
  [13940] = {.lex_state = 224, .external_lex_state = 91},
  [13941] = {.lex_state = 224, .external_lex_state = 91},
  [13942] = {.lex_state = 199, .external_lex_state = 116},
  [13943] = {.lex_state = 179},
  [13944] = {.lex_state = 224, .external_lex_state = 91},
  [13945] = {.lex_state = 199, .external_lex_state = 116},
  [13946] = {.lex_state = 224, .external_lex_state = 91},
  [13947] = {.lex_state = 224, .external_lex_state = 91},
  [13948] = {.lex_state = 199, .external_lex_state = 116},
  [13949] = {.lex_state = 138},
  [13950] = {.lex_state = 224, .external_lex_state = 91},
  [13951] = {.lex_state = 224, .external_lex_state = 91},
  [13952] = {.lex_state = 199, .external_lex_state = 116},
  [13953] = {.lex_state = 199, .external_lex_state = 116},
  [13954] = {.lex_state = 224, .external_lex_state = 91},
  [13955] = {.lex_state = 224, .external_lex_state = 104},
  [13956] = {.lex_state = 199, .external_lex_state = 116},
  [13957] = {.lex_state = 138, .external_lex_state = 114},
  [13958] = {.lex_state = 199, .external_lex_state = 116},
  [13959] = {.lex_state = 199, .external_lex_state = 116},
  [13960] = {.lex_state = 199, .external_lex_state = 116},
  [13961] = {.lex_state = 199, .external_lex_state = 116},
  [13962] = {.lex_state = 223},
  [13963] = {.lex_state = 199, .external_lex_state = 116},
  [13964] = {.lex_state = 138},
  [13965] = {.lex_state = 179},
  [13966] = {.lex_state = 199, .external_lex_state = 116},
  [13967] = {.lex_state = 199, .external_lex_state = 116},
  [13968] = {.lex_state = 199, .external_lex_state = 116},
  [13969] = {.lex_state = 224, .external_lex_state = 91},
  [13970] = {.lex_state = 224, .external_lex_state = 91},
  [13971] = {.lex_state = 199, .external_lex_state = 116},
  [13972] = {.lex_state = 179},
  [13973] = {.lex_state = 224, .external_lex_state = 91},
  [13974] = {.lex_state = 199, .external_lex_state = 116},
  [13975] = {.lex_state = 224, .external_lex_state = 91},
  [13976] = {.lex_state = 199, .external_lex_state = 116},
  [13977] = {.lex_state = 210, .external_lex_state = 100},
  [13978] = {.lex_state = 199, .external_lex_state = 116},
  [13979] = {.lex_state = 199, .external_lex_state = 116},
  [13980] = {.lex_state = 179},
  [13981] = {.lex_state = 138, .external_lex_state = 120},
  [13982] = {.lex_state = 199, .external_lex_state = 116},
  [13983] = {.lex_state = 138, .external_lex_state = 120},
  [13984] = {.lex_state = 138, .external_lex_state = 120},
  [13985] = {.lex_state = 199, .external_lex_state = 116},
  [13986] = {.lex_state = 179},
  [13987] = {.lex_state = 224, .external_lex_state = 104},
  [13988] = {.lex_state = 199, .external_lex_state = 116},
  [13989] = {.lex_state = 224, .external_lex_state = 90},
  [13990] = {.lex_state = 199, .external_lex_state = 116},
  [13991] = {.lex_state = 224, .external_lex_state = 90},
  [13992] = {.lex_state = 179},
  [13993] = {.lex_state = 138, .external_lex_state = 120},
  [13994] = {.lex_state = 224, .external_lex_state = 91},
  [13995] = {.lex_state = 224, .external_lex_state = 104},
  [13996] = {.lex_state = 210, .external_lex_state = 99},
  [13997] = {.lex_state = 179},
  [13998] = {.lex_state = 224, .external_lex_state = 104},
  [13999] = {.lex_state = 199, .external_lex_state = 116},
  [14000] = {.lex_state = 224, .external_lex_state = 99},
  [14001] = {.lex_state = 210, .external_lex_state = 99},
  [14002] = {.lex_state = 199, .external_lex_state = 116},
  [14003] = {.lex_state = 224, .external_lex_state = 104},
  [14004] = {.lex_state = 199, .external_lex_state = 116},
  [14005] = {.lex_state = 179},
  [14006] = {.lex_state = 199, .external_lex_state = 116},
  [14007] = {.lex_state = 179},
  [14008] = {.lex_state = 179},
  [14009] = {.lex_state = 223},
  [14010] = {.lex_state = 223},
  [14011] = {.lex_state = 138, .external_lex_state = 99},
  [14012] = {.lex_state = 199, .external_lex_state = 116},
  [14013] = {.lex_state = 199, .external_lex_state = 116},
  [14014] = {.lex_state = 224, .external_lex_state = 102},
  [14015] = {.lex_state = 199, .external_lex_state = 116},
  [14016] = {.lex_state = 138, .external_lex_state = 121},
  [14017] = {.lex_state = 220},
  [14018] = {.lex_state = 223},
  [14019] = {.lex_state = 138, .external_lex_state = 121},
  [14020] = {.lex_state = 224, .external_lex_state = 104},
  [14021] = {.lex_state = 199, .external_lex_state = 116},
  [14022] = {.lex_state = 455},
  [14023] = {.lex_state = 138, .external_lex_state = 121},
  [14024] = {.lex_state = 219, .external_lex_state = 109},
  [14025] = {.lex_state = 199, .external_lex_state = 116},
  [14026] = {.lex_state = 223},
  [14027] = {.lex_state = 138, .external_lex_state = 121},
  [14028] = {.lex_state = 220},
  [14029] = {.lex_state = 179},
  [14030] = {.lex_state = 138, .external_lex_state = 99},
  [14031] = {.lex_state = 199, .external_lex_state = 116},
  [14032] = {.lex_state = 199, .external_lex_state = 116},
  [14033] = {.lex_state = 179},
  [14034] = {.lex_state = 138, .external_lex_state = 121},
  [14035] = {.lex_state = 138, .external_lex_state = 121},
  [14036] = {.lex_state = 199, .external_lex_state = 116},
  [14037] = {.lex_state = 179},
  [14038] = {.lex_state = 179},
  [14039] = {.lex_state = 223},
  [14040] = {.lex_state = 138},
  [14041] = {.lex_state = 138, .external_lex_state = 119},
  [14042] = {.lex_state = 138, .external_lex_state = 119},
  [14043] = {.lex_state = 455},
  [14044] = {.lex_state = 138, .external_lex_state = 119},
  [14045] = {.lex_state = 138, .external_lex_state = 119},
  [14046] = {.lex_state = 138, .external_lex_state = 120},
  [14047] = {.lex_state = 199, .external_lex_state = 116},
  [14048] = {.lex_state = 223},
  [14049] = {.lex_state = 138},
  [14050] = {.lex_state = 199, .external_lex_state = 116},
  [14051] = {.lex_state = 210, .external_lex_state = 101},
  [14052] = {.lex_state = 455},
  [14053] = {.lex_state = 223},
  [14054] = {.lex_state = 138, .external_lex_state = 121},
  [14055] = {.lex_state = 138, .external_lex_state = 120},
  [14056] = {.lex_state = 199, .external_lex_state = 116},
  [14057] = {.lex_state = 199, .external_lex_state = 116},
  [14058] = {.lex_state = 138},
  [14059] = {.lex_state = 138, .external_lex_state = 121},
  [14060] = {.lex_state = 138, .external_lex_state = 121},
  [14061] = {.lex_state = 223},
  [14062] = {.lex_state = 455},
  [14063] = {.lex_state = 224, .external_lex_state = 109},
  [14064] = {.lex_state = 138},
  [14065] = {.lex_state = 219, .external_lex_state = 109},
  [14066] = {.lex_state = 224, .external_lex_state = 109},
  [14067] = {.lex_state = 219, .external_lex_state = 109},
  [14068] = {.lex_state = 199, .external_lex_state = 116},
  [14069] = {.lex_state = 199, .external_lex_state = 116},
  [14070] = {.lex_state = 138, .external_lex_state = 120},
  [14071] = {.lex_state = 199, .external_lex_state = 116},
  [14072] = {.lex_state = 223},
  [14073] = {.lex_state = 138},
  [14074] = {.lex_state = 199, .external_lex_state = 116},
  [14075] = {.lex_state = 455},
  [14076] = {.lex_state = 138, .external_lex_state = 120},
  [14077] = {.lex_state = 138},
  [14078] = {.lex_state = 199, .external_lex_state = 116},
  [14079] = {.lex_state = 223},
  [14080] = {.lex_state = 138, .external_lex_state = 114},
  [14081] = {.lex_state = 138},
  [14082] = {.lex_state = 455},
  [14083] = {.lex_state = 224, .external_lex_state = 102},
  [14084] = {.lex_state = 179},
  [14085] = {.lex_state = 199, .external_lex_state = 116},
  [14086] = {.lex_state = 224, .external_lex_state = 90},
  [14087] = {.lex_state = 138, .external_lex_state = 120},
  [14088] = {.lex_state = 199, .external_lex_state = 116},
  [14089] = {.lex_state = 138},
  [14090] = {.lex_state = 138, .external_lex_state = 114},
  [14091] = {.lex_state = 138, .external_lex_state = 114},
  [14092] = {.lex_state = 455},
  [14093] = {.lex_state = 138, .external_lex_state = 114},
  [14094] = {.lex_state = 198, .external_lex_state = 116},
  [14095] = {.lex_state = 198, .external_lex_state = 116},
  [14096] = {.lex_state = 179},
  [14097] = {.lex_state = 455},
  [14098] = {.lex_state = 138, .external_lex_state = 114},
  [14099] = {.lex_state = 138, .external_lex_state = 120},
  [14100] = {.lex_state = 138, .external_lex_state = 119},
  [14101] = {.lex_state = 138},
  [14102] = {.lex_state = 199, .external_lex_state = 116},
  [14103] = {.lex_state = 138, .external_lex_state = 120},
  [14104] = {.lex_state = 138},
  [14105] = {.lex_state = 455},
  [14106] = {.lex_state = 138},
  [14107] = {.lex_state = 199, .external_lex_state = 116},
  [14108] = {.lex_state = 179},
  [14109] = {.lex_state = 138, .external_lex_state = 120},
  [14110] = {.lex_state = 138, .external_lex_state = 120},
  [14111] = {.lex_state = 223},
  [14112] = {.lex_state = 455},
  [14113] = {.lex_state = 138},
  [14114] = {.lex_state = 199, .external_lex_state = 116},
  [14115] = {.lex_state = 199, .external_lex_state = 116},
  [14116] = {.lex_state = 138, .external_lex_state = 120},
  [14117] = {.lex_state = 138, .external_lex_state = 99},
  [14118] = {.lex_state = 138, .external_lex_state = 114},
  [14119] = {.lex_state = 199, .external_lex_state = 116},
  [14120] = {.lex_state = 138},
  [14121] = {.lex_state = 210, .external_lex_state = 100},
  [14122] = {.lex_state = 199, .external_lex_state = 116},
  [14123] = {.lex_state = 199, .external_lex_state = 116},
  [14124] = {.lex_state = 199, .external_lex_state = 116},
  [14125] = {.lex_state = 223},
  [14126] = {.lex_state = 138},
  [14127] = {.lex_state = 138, .external_lex_state = 120},
  [14128] = {.lex_state = 224, .external_lex_state = 91},
  [14129] = {.lex_state = 224, .external_lex_state = 91},
  [14130] = {.lex_state = 210, .external_lex_state = 99},
  [14131] = {.lex_state = 199, .external_lex_state = 116},
  [14132] = {.lex_state = 138},
  [14133] = {.lex_state = 224, .external_lex_state = 91},
  [14134] = {.lex_state = 199, .external_lex_state = 116},
  [14135] = {.lex_state = 138, .external_lex_state = 120},
  [14136] = {.lex_state = 138},
  [14137] = {.lex_state = 138, .external_lex_state = 114},
  [14138] = {.lex_state = 224, .external_lex_state = 102},
  [14139] = {.lex_state = 199, .external_lex_state = 116},
  [14140] = {.lex_state = 138, .external_lex_state = 120},
  [14141] = {.lex_state = 138},
  [14142] = {.lex_state = 138, .external_lex_state = 120},
  [14143] = {.lex_state = 138},
  [14144] = {.lex_state = 220},
  [14145] = {.lex_state = 179},
  [14146] = {.lex_state = 138},
  [14147] = {.lex_state = 179},
  [14148] = {.lex_state = 199, .external_lex_state = 116},
  [14149] = {.lex_state = 138, .external_lex_state = 99},
  [14150] = {.lex_state = 138},
  [14151] = {.lex_state = 224, .external_lex_state = 101},
  [14152] = {.lex_state = 199, .external_lex_state = 116},
  [14153] = {.lex_state = 138},
  [14154] = {.lex_state = 224, .external_lex_state = 101},
  [14155] = {.lex_state = 179},
  [14156] = {.lex_state = 138},
  [14157] = {.lex_state = 138},
  [14158] = {.lex_state = 138},
  [14159] = {.lex_state = 224, .external_lex_state = 101},
  [14160] = {.lex_state = 138},
  [14161] = {.lex_state = 224, .external_lex_state = 91},
  [14162] = {.lex_state = 224, .external_lex_state = 91},
  [14163] = {.lex_state = 224, .external_lex_state = 100},
  [14164] = {.lex_state = 138},
  [14165] = {.lex_state = 224, .external_lex_state = 100},
  [14166] = {.lex_state = 224, .external_lex_state = 100},
  [14167] = {.lex_state = 138},
  [14168] = {.lex_state = 199, .external_lex_state = 116},
  [14169] = {.lex_state = 138, .external_lex_state = 121},
  [14170] = {.lex_state = 138, .external_lex_state = 121},
  [14171] = {.lex_state = 138},
  [14172] = {.lex_state = 199, .external_lex_state = 116},
  [14173] = {.lex_state = 455},
  [14174] = {.lex_state = 179},
  [14175] = {.lex_state = 138, .external_lex_state = 121},
  [14176] = {.lex_state = 138},
  [14177] = {.lex_state = 224, .external_lex_state = 104},
  [14178] = {.lex_state = 455, .external_lex_state = 114},
  [14179] = {.lex_state = 138},
  [14180] = {.lex_state = 138},
  [14181] = {.lex_state = 199, .external_lex_state = 116},
  [14182] = {.lex_state = 199, .external_lex_state = 116},
  [14183] = {.lex_state = 220},
  [14184] = {.lex_state = 224, .external_lex_state = 101},
  [14185] = {.lex_state = 224, .external_lex_state = 101},
  [14186] = {.lex_state = 224, .external_lex_state = 101},
  [14187] = {.lex_state = 138},
  [14188] = {.lex_state = 224, .external_lex_state = 101},
  [14189] = {.lex_state = 224, .external_lex_state = 101},
  [14190] = {.lex_state = 223},
  [14191] = {.lex_state = 224, .external_lex_state = 101},
  [14192] = {.lex_state = 199, .external_lex_state = 116},
  [14193] = {.lex_state = 138},
  [14194] = {.lex_state = 138},
  [14195] = {.lex_state = 199, .external_lex_state = 116},
  [14196] = {.lex_state = 138, .external_lex_state = 121},
  [14197] = {.lex_state = 138},
  [14198] = {.lex_state = 199, .external_lex_state = 116},
  [14199] = {.lex_state = 138},
  [14200] = {.lex_state = 138},
  [14201] = {.lex_state = 138},
  [14202] = {.lex_state = 224, .external_lex_state = 100},
  [14203] = {.lex_state = 199, .external_lex_state = 116},
  [14204] = {.lex_state = 138},
  [14205] = {.lex_state = 138},
  [14206] = {.lex_state = 138},
  [14207] = {.lex_state = 138, .external_lex_state = 120},
  [14208] = {.lex_state = 138},
  [14209] = {.lex_state = 138},
  [14210] = {.lex_state = 138},
  [14211] = {.lex_state = 138},
  [14212] = {.lex_state = 138},
  [14213] = {.lex_state = 138},
  [14214] = {.lex_state = 138},
  [14215] = {.lex_state = 138},
  [14216] = {.lex_state = 138},
  [14217] = {.lex_state = 224, .external_lex_state = 90},
  [14218] = {.lex_state = 138},
  [14219] = {.lex_state = 138},
  [14220] = {.lex_state = 224, .external_lex_state = 90},
  [14221] = {.lex_state = 138},
  [14222] = {.lex_state = 199, .external_lex_state = 116},
  [14223] = {.lex_state = 138},
  [14224] = {.lex_state = 455},
  [14225] = {.lex_state = 138},
  [14226] = {.lex_state = 138},
  [14227] = {.lex_state = 138},
  [14228] = {.lex_state = 138},
  [14229] = {.lex_state = 138},
  [14230] = {.lex_state = 138},
  [14231] = {.lex_state = 138},
  [14232] = {.lex_state = 138},
  [14233] = {.lex_state = 138, .external_lex_state = 114},
  [14234] = {.lex_state = 138},
  [14235] = {.lex_state = 138},
  [14236] = {.lex_state = 199, .external_lex_state = 116},
  [14237] = {.lex_state = 138},
  [14238] = {.lex_state = 138},
  [14239] = {.lex_state = 138},
  [14240] = {.lex_state = 138},
  [14241] = {.lex_state = 220},
  [14242] = {.lex_state = 138},
  [14243] = {.lex_state = 138},
  [14244] = {.lex_state = 220},
  [14245] = {.lex_state = 138},
  [14246] = {.lex_state = 220},
  [14247] = {.lex_state = 138},
  [14248] = {.lex_state = 138},
  [14249] = {.lex_state = 138},
  [14250] = {.lex_state = 138},
  [14251] = {.lex_state = 138},
  [14252] = {.lex_state = 138},
  [14253] = {.lex_state = 179},
  [14254] = {.lex_state = 199, .external_lex_state = 116},
  [14255] = {.lex_state = 138},
  [14256] = {.lex_state = 138},
  [14257] = {.lex_state = 224, .external_lex_state = 103},
  [14258] = {.lex_state = 138},
  [14259] = {.lex_state = 224, .external_lex_state = 103},
  [14260] = {.lex_state = 199, .external_lex_state = 116},
  [14261] = {.lex_state = 138},
  [14262] = {.lex_state = 138},
  [14263] = {.lex_state = 138},
  [14264] = {.lex_state = 138},
  [14265] = {.lex_state = 138},
  [14266] = {.lex_state = 138},
  [14267] = {.lex_state = 138},
  [14268] = {.lex_state = 138},
  [14269] = {.lex_state = 138},
  [14270] = {.lex_state = 138},
  [14271] = {.lex_state = 138},
  [14272] = {.lex_state = 138},
  [14273] = {.lex_state = 138},
  [14274] = {.lex_state = 179},
  [14275] = {.lex_state = 138},
  [14276] = {.lex_state = 199, .external_lex_state = 116},
  [14277] = {.lex_state = 138},
  [14278] = {.lex_state = 224, .external_lex_state = 100},
  [14279] = {.lex_state = 138},
  [14280] = {.lex_state = 138},
  [14281] = {.lex_state = 224, .external_lex_state = 100},
  [14282] = {.lex_state = 138},
  [14283] = {.lex_state = 138, .external_lex_state = 120},
  [14284] = {.lex_state = 138},
  [14285] = {.lex_state = 199, .external_lex_state = 116},
  [14286] = {.lex_state = 138},
  [14287] = {.lex_state = 224, .external_lex_state = 100},
  [14288] = {.lex_state = 138},
  [14289] = {.lex_state = 224, .external_lex_state = 100},
  [14290] = {.lex_state = 138},
  [14291] = {.lex_state = 224, .external_lex_state = 100},
  [14292] = {.lex_state = 138},
  [14293] = {.lex_state = 138},
  [14294] = {.lex_state = 138},
  [14295] = {.lex_state = 224, .external_lex_state = 100},
  [14296] = {.lex_state = 138},
  [14297] = {.lex_state = 138, .external_lex_state = 99},
  [14298] = {.lex_state = 138},
  [14299] = {.lex_state = 138},
  [14300] = {.lex_state = 224, .external_lex_state = 103},
  [14301] = {.lex_state = 138},
  [14302] = {.lex_state = 138, .external_lex_state = 99},
  [14303] = {.lex_state = 219, .external_lex_state = 109},
  [14304] = {.lex_state = 138},
  [14305] = {.lex_state = 138},
  [14306] = {.lex_state = 138},
  [14307] = {.lex_state = 223},
  [14308] = {.lex_state = 138},
  [14309] = {.lex_state = 138},
  [14310] = {.lex_state = 138},
  [14311] = {.lex_state = 455},
  [14312] = {.lex_state = 455},
  [14313] = {.lex_state = 199, .external_lex_state = 116},
  [14314] = {.lex_state = 224, .external_lex_state = 109},
  [14315] = {.lex_state = 224, .external_lex_state = 109},
  [14316] = {.lex_state = 199, .external_lex_state = 116},
  [14317] = {.lex_state = 224, .external_lex_state = 103},
  [14318] = {.lex_state = 455},
  [14319] = {.lex_state = 224, .external_lex_state = 91},
  [14320] = {.lex_state = 455},
  [14321] = {.lex_state = 224, .external_lex_state = 91},
  [14322] = {.lex_state = 199, .external_lex_state = 116},
  [14323] = {.lex_state = 224, .external_lex_state = 91},
  [14324] = {.lex_state = 224, .external_lex_state = 103},
  [14325] = {.lex_state = 455},
  [14326] = {.lex_state = 455},
  [14327] = {.lex_state = 224, .external_lex_state = 109},
  [14328] = {.lex_state = 199, .external_lex_state = 116},
  [14329] = {.lex_state = 219, .external_lex_state = 109},
  [14330] = {.lex_state = 199, .external_lex_state = 116},
  [14331] = {.lex_state = 224, .external_lex_state = 91},
  [14332] = {.lex_state = 455},
  [14333] = {.lex_state = 455},
  [14334] = {.lex_state = 224, .external_lex_state = 109},
  [14335] = {.lex_state = 199, .external_lex_state = 116},
  [14336] = {.lex_state = 223},
  [14337] = {.lex_state = 455},
  [14338] = {.lex_state = 455},
  [14339] = {.lex_state = 138},
  [14340] = {.lex_state = 199, .external_lex_state = 116},
  [14341] = {.lex_state = 455},
  [14342] = {.lex_state = 455},
  [14343] = {.lex_state = 199, .external_lex_state = 116},
  [14344] = {.lex_state = 455},
  [14345] = {.lex_state = 224, .external_lex_state = 103},
  [14346] = {.lex_state = 455},
  [14347] = {.lex_state = 199, .external_lex_state = 116},
  [14348] = {.lex_state = 224, .external_lex_state = 103},
  [14349] = {.lex_state = 455},
  [14350] = {.lex_state = 455},
  [14351] = {.lex_state = 199, .external_lex_state = 116},
  [14352] = {.lex_state = 224, .external_lex_state = 90},
  [14353] = {.lex_state = 455},
  [14354] = {.lex_state = 224, .external_lex_state = 90},
  [14355] = {.lex_state = 455},
  [14356] = {.lex_state = 455},
  [14357] = {.lex_state = 455},
  [14358] = {.lex_state = 455},
  [14359] = {.lex_state = 455},
  [14360] = {.lex_state = 224, .external_lex_state = 90},
  [14361] = {.lex_state = 224, .external_lex_state = 90},
  [14362] = {.lex_state = 138},
  [14363] = {.lex_state = 199, .external_lex_state = 116},
  [14364] = {.lex_state = 224, .external_lex_state = 90},
  [14365] = {.lex_state = 455},
  [14366] = {.lex_state = 455},
  [14367] = {.lex_state = 455},
  [14368] = {.lex_state = 224, .external_lex_state = 91},
  [14369] = {.lex_state = 455},
  [14370] = {.lex_state = 455},
  [14371] = {.lex_state = 455},
  [14372] = {.lex_state = 455},
  [14373] = {.lex_state = 455},
  [14374] = {.lex_state = 223},
  [14375] = {.lex_state = 455},
  [14376] = {.lex_state = 455},
  [14377] = {.lex_state = 455},
  [14378] = {.lex_state = 455},
  [14379] = {.lex_state = 179},
  [14380] = {.lex_state = 455},
  [14381] = {.lex_state = 199, .external_lex_state = 116},
  [14382] = {.lex_state = 455},
  [14383] = {.lex_state = 224, .external_lex_state = 90},
  [14384] = {.lex_state = 455},
  [14385] = {.lex_state = 455},
  [14386] = {.lex_state = 224, .external_lex_state = 90},
  [14387] = {.lex_state = 455},
  [14388] = {.lex_state = 224, .external_lex_state = 90},
  [14389] = {.lex_state = 455},
  [14390] = {.lex_state = 224, .external_lex_state = 90},
  [14391] = {.lex_state = 455},
  [14392] = {.lex_state = 199, .external_lex_state = 116},
  [14393] = {.lex_state = 224, .external_lex_state = 90},
  [14394] = {.lex_state = 455},
  [14395] = {.lex_state = 224, .external_lex_state = 90},
  [14396] = {.lex_state = 455},
  [14397] = {.lex_state = 224, .external_lex_state = 90},
  [14398] = {.lex_state = 455},
  [14399] = {.lex_state = 224, .external_lex_state = 90},
  [14400] = {.lex_state = 455},
  [14401] = {.lex_state = 224, .external_lex_state = 90},
  [14402] = {.lex_state = 224, .external_lex_state = 90},
  [14403] = {.lex_state = 179},
  [14404] = {.lex_state = 455},
  [14405] = {.lex_state = 199, .external_lex_state = 116},
  [14406] = {.lex_state = 455},
  [14407] = {.lex_state = 455},
  [14408] = {.lex_state = 224, .external_lex_state = 90},
  [14409] = {.lex_state = 224, .external_lex_state = 90},
  [14410] = {.lex_state = 455},
  [14411] = {.lex_state = 224, .external_lex_state = 90},
  [14412] = {.lex_state = 224, .external_lex_state = 90},
  [14413] = {.lex_state = 224, .external_lex_state = 90},
  [14414] = {.lex_state = 455},
  [14415] = {.lex_state = 455},
  [14416] = {.lex_state = 455},
  [14417] = {.lex_state = 224, .external_lex_state = 90},
  [14418] = {.lex_state = 455},
  [14419] = {.lex_state = 455},
  [14420] = {.lex_state = 455},
  [14421] = {.lex_state = 455},
  [14422] = {.lex_state = 224, .external_lex_state = 90},
  [14423] = {.lex_state = 455},
  [14424] = {.lex_state = 455},
  [14425] = {.lex_state = 455},
  [14426] = {.lex_state = 455},
  [14427] = {.lex_state = 455},
  [14428] = {.lex_state = 455},
  [14429] = {.lex_state = 455},
  [14430] = {.lex_state = 455},
  [14431] = {.lex_state = 455},
  [14432] = {.lex_state = 455},
  [14433] = {.lex_state = 455},
  [14434] = {.lex_state = 455},
  [14435] = {.lex_state = 455, .external_lex_state = 114},
  [14436] = {.lex_state = 455},
  [14437] = {.lex_state = 455},
  [14438] = {.lex_state = 455},
  [14439] = {.lex_state = 455},
  [14440] = {.lex_state = 455},
  [14441] = {.lex_state = 455},
  [14442] = {.lex_state = 455},
  [14443] = {.lex_state = 455},
  [14444] = {.lex_state = 455},
  [14445] = {.lex_state = 455},
  [14446] = {.lex_state = 455},
  [14447] = {.lex_state = 455},
  [14448] = {.lex_state = 455},
  [14449] = {.lex_state = 455},
  [14450] = {.lex_state = 455},
  [14451] = {.lex_state = 455},
  [14452] = {.lex_state = 455},
  [14453] = {.lex_state = 455},
  [14454] = {.lex_state = 455},
  [14455] = {.lex_state = 455},
  [14456] = {.lex_state = 455},
  [14457] = {.lex_state = 455},
  [14458] = {.lex_state = 455},
  [14459] = {.lex_state = 455},
  [14460] = {.lex_state = 455},
  [14461] = {.lex_state = 455},
  [14462] = {.lex_state = 455},
  [14463] = {.lex_state = 455},
  [14464] = {.lex_state = 455},
  [14465] = {.lex_state = 455},
  [14466] = {.lex_state = 455},
  [14467] = {.lex_state = 455},
  [14468] = {.lex_state = 455},
  [14469] = {.lex_state = 455},
  [14470] = {.lex_state = 455},
  [14471] = {.lex_state = 455},
  [14472] = {.lex_state = 455},
  [14473] = {.lex_state = 455},
  [14474] = {.lex_state = 455},
  [14475] = {.lex_state = 455},
  [14476] = {.lex_state = 455},
  [14477] = {.lex_state = 455},
  [14478] = {.lex_state = 455},
  [14479] = {.lex_state = 455},
  [14480] = {.lex_state = 455},
  [14481] = {.lex_state = 455},
  [14482] = {.lex_state = 455},
  [14483] = {.lex_state = 455},
  [14484] = {.lex_state = 455},
  [14485] = {.lex_state = 455},
  [14486] = {.lex_state = 455},
  [14487] = {.lex_state = 455},
  [14488] = {.lex_state = 455},
  [14489] = {.lex_state = 455},
  [14490] = {.lex_state = 455},
  [14491] = {.lex_state = 455},
  [14492] = {.lex_state = 455},
  [14493] = {.lex_state = 455},
  [14494] = {.lex_state = 455},
  [14495] = {.lex_state = 455},
  [14496] = {.lex_state = 455},
  [14497] = {.lex_state = 455},
  [14498] = {.lex_state = 455},
  [14499] = {.lex_state = 455},
  [14500] = {.lex_state = 455},
  [14501] = {.lex_state = 455},
  [14502] = {.lex_state = 455},
  [14503] = {.lex_state = 455},
  [14504] = {.lex_state = 455},
  [14505] = {.lex_state = 455},
  [14506] = {.lex_state = 455},
  [14507] = {.lex_state = 455},
  [14508] = {.lex_state = 455},
  [14509] = {.lex_state = 455},
  [14510] = {.lex_state = 455},
  [14511] = {.lex_state = 455},
  [14512] = {.lex_state = 455},
  [14513] = {.lex_state = 455},
  [14514] = {.lex_state = 455},
  [14515] = {.lex_state = 455},
  [14516] = {.lex_state = 455},
  [14517] = {.lex_state = 455},
  [14518] = {.lex_state = 455},
  [14519] = {.lex_state = 455},
  [14520] = {.lex_state = 455},
  [14521] = {.lex_state = 455},
  [14522] = {.lex_state = 455},
  [14523] = {.lex_state = 455},
  [14524] = {.lex_state = 455},
  [14525] = {.lex_state = 455},
  [14526] = {.lex_state = 455},
  [14527] = {.lex_state = 455},
  [14528] = {.lex_state = 455},
  [14529] = {.lex_state = 455},
  [14530] = {.lex_state = 455},
  [14531] = {.lex_state = 455},
  [14532] = {.lex_state = 455},
  [14533] = {.lex_state = 455},
  [14534] = {.lex_state = 455},
  [14535] = {.lex_state = 455},
  [14536] = {.lex_state = 455},
  [14537] = {.lex_state = 455},
  [14538] = {.lex_state = 455},
  [14539] = {.lex_state = 455},
  [14540] = {.lex_state = 455},
  [14541] = {.lex_state = 455},
  [14542] = {.lex_state = 455},
  [14543] = {.lex_state = 455},
  [14544] = {.lex_state = 455},
  [14545] = {.lex_state = 455},
  [14546] = {.lex_state = 455},
  [14547] = {.lex_state = 455},
  [14548] = {.lex_state = 455},
  [14549] = {.lex_state = 455},
  [14550] = {.lex_state = 455},
  [14551] = {.lex_state = 455},
  [14552] = {.lex_state = 455},
  [14553] = {.lex_state = 455},
  [14554] = {.lex_state = 455},
  [14555] = {.lex_state = 455},
  [14556] = {.lex_state = 455},
  [14557] = {.lex_state = 455},
  [14558] = {.lex_state = 455},
  [14559] = {.lex_state = 455},
  [14560] = {.lex_state = 455},
  [14561] = {.lex_state = 455},
  [14562] = {.lex_state = 455},
  [14563] = {.lex_state = 455},
  [14564] = {.lex_state = 455},
  [14565] = {.lex_state = 455},
  [14566] = {.lex_state = 455},
  [14567] = {.lex_state = 455},
  [14568] = {.lex_state = 455},
  [14569] = {.lex_state = 199, .external_lex_state = 116},
  [14570] = {.lex_state = 455},
  [14571] = {.lex_state = 455},
  [14572] = {.lex_state = 138},
  [14573] = {.lex_state = 455},
  [14574] = {.lex_state = 455},
  [14575] = {.lex_state = 455},
  [14576] = {.lex_state = 455},
  [14577] = {.lex_state = 224, .external_lex_state = 90},
  [14578] = {.lex_state = 138},
  [14579] = {.lex_state = 455},
  [14580] = {.lex_state = 455},
  [14581] = {.lex_state = 198, .external_lex_state = 116},
  [14582] = {.lex_state = 455},
  [14583] = {.lex_state = 455},
  [14584] = {.lex_state = 455},
  [14585] = {.lex_state = 455},
  [14586] = {.lex_state = 455},
  [14587] = {.lex_state = 455},
  [14588] = {.lex_state = 138, .external_lex_state = 99},
  [14589] = {.lex_state = 199, .external_lex_state = 116},
  [14590] = {.lex_state = 224, .external_lex_state = 90},
  [14591] = {.lex_state = 224, .external_lex_state = 90},
  [14592] = {.lex_state = 224, .external_lex_state = 90},
  [14593] = {.lex_state = 224, .external_lex_state = 90},
  [14594] = {.lex_state = 199, .external_lex_state = 116},
  [14595] = {.lex_state = 223},
  [14596] = {.lex_state = 224, .external_lex_state = 99},
  [14597] = {.lex_state = 138},
  [14598] = {.lex_state = 224, .external_lex_state = 90},
  [14599] = {.lex_state = 199, .external_lex_state = 116},
  [14600] = {.lex_state = 224, .external_lex_state = 101},
  [14601] = {.lex_state = 138},
  [14602] = {.lex_state = 224, .external_lex_state = 101},
  [14603] = {.lex_state = 199, .external_lex_state = 116},
  [14604] = {.lex_state = 224, .external_lex_state = 91},
  [14605] = {.lex_state = 199, .external_lex_state = 116},
  [14606] = {.lex_state = 224, .external_lex_state = 91},
  [14607] = {.lex_state = 138},
  [14608] = {.lex_state = 179},
  [14609] = {.lex_state = 138},
  [14610] = {.lex_state = 138, .external_lex_state = 114},
  [14611] = {.lex_state = 138},
  [14612] = {.lex_state = 224, .external_lex_state = 92},
  [14613] = {.lex_state = 224},
  [14614] = {.lex_state = 138},
  [14615] = {.lex_state = 138},
  [14616] = {.lex_state = 210, .external_lex_state = 99},
  [14617] = {.lex_state = 138},
  [14618] = {.lex_state = 138},
  [14619] = {.lex_state = 138},
  [14620] = {.lex_state = 138},
  [14621] = {.lex_state = 179},
  [14622] = {.lex_state = 138},
  [14623] = {.lex_state = 138, .external_lex_state = 119},
  [14624] = {.lex_state = 179},
  [14625] = {.lex_state = 138},
  [14626] = {.lex_state = 179},
  [14627] = {.lex_state = 224, .external_lex_state = 100},
  [14628] = {.lex_state = 144, .external_lex_state = 116},
  [14629] = {.lex_state = 179},
  [14630] = {.lex_state = 224, .external_lex_state = 101},
  [14631] = {.lex_state = 144, .external_lex_state = 116},
  [14632] = {.lex_state = 138},
  [14633] = {.lex_state = 179},
  [14634] = {.lex_state = 138},
  [14635] = {.lex_state = 224},
  [14636] = {.lex_state = 138, .external_lex_state = 120},
  [14637] = {.lex_state = 138, .external_lex_state = 119},
  [14638] = {.lex_state = 138},
  [14639] = {.lex_state = 144, .external_lex_state = 116},
  [14640] = {.lex_state = 220},
  [14641] = {.lex_state = 138},
  [14642] = {.lex_state = 138},
  [14643] = {.lex_state = 138, .external_lex_state = 114},
  [14644] = {.lex_state = 138, .external_lex_state = 120},
  [14645] = {.lex_state = 138},
  [14646] = {.lex_state = 138},
  [14647] = {.lex_state = 224, .external_lex_state = 100},
  [14648] = {.lex_state = 224, .external_lex_state = 100},
  [14649] = {.lex_state = 179},
  [14650] = {.lex_state = 179},
  [14651] = {.lex_state = 224},
  [14652] = {.lex_state = 138},
  [14653] = {.lex_state = 138},
  [14654] = {.lex_state = 138},
  [14655] = {.lex_state = 224},
  [14656] = {.lex_state = 224, .external_lex_state = 101},
  [14657] = {.lex_state = 179},
  [14658] = {.lex_state = 138},
  [14659] = {.lex_state = 138},
  [14660] = {.lex_state = 138, .external_lex_state = 118},
  [14661] = {.lex_state = 138},
  [14662] = {.lex_state = 138},
  [14663] = {.lex_state = 138},
  [14664] = {.lex_state = 220},
  [14665] = {.lex_state = 138, .external_lex_state = 114},
  [14666] = {.lex_state = 224},
  [14667] = {.lex_state = 138, .external_lex_state = 114},
  [14668] = {.lex_state = 224},
  [14669] = {.lex_state = 179},
  [14670] = {.lex_state = 224, .external_lex_state = 101},
  [14671] = {.lex_state = 138},
  [14672] = {.lex_state = 138},
  [14673] = {.lex_state = 138},
  [14674] = {.lex_state = 138},
  [14675] = {.lex_state = 138},
  [14676] = {.lex_state = 138},
  [14677] = {.lex_state = 179},
  [14678] = {.lex_state = 219, .external_lex_state = 109},
  [14679] = {.lex_state = 138},
  [14680] = {.lex_state = 138},
  [14681] = {.lex_state = 224},
  [14682] = {.lex_state = 138},
  [14683] = {.lex_state = 224, .external_lex_state = 100},
  [14684] = {.lex_state = 224, .external_lex_state = 101},
  [14685] = {.lex_state = 224, .external_lex_state = 101},
  [14686] = {.lex_state = 224, .external_lex_state = 101},
  [14687] = {.lex_state = 138},
  [14688] = {.lex_state = 138},
  [14689] = {.lex_state = 210, .external_lex_state = 99},
  [14690] = {.lex_state = 138},
  [14691] = {.lex_state = 224},
  [14692] = {.lex_state = 224},
  [14693] = {.lex_state = 224, .external_lex_state = 102},
  [14694] = {.lex_state = 138},
  [14695] = {.lex_state = 219, .external_lex_state = 109},
  [14696] = {.lex_state = 224, .external_lex_state = 102},
  [14697] = {.lex_state = 138},
  [14698] = {.lex_state = 179},
  [14699] = {.lex_state = 138},
  [14700] = {.lex_state = 201, .external_lex_state = 116},
  [14701] = {.lex_state = 138},
  [14702] = {.lex_state = 224},
  [14703] = {.lex_state = 219, .external_lex_state = 109},
  [14704] = {.lex_state = 224, .external_lex_state = 100},
  [14705] = {.lex_state = 138, .external_lex_state = 122},
  [14706] = {.lex_state = 224, .external_lex_state = 102},
  [14707] = {.lex_state = 138},
  [14708] = {.lex_state = 455},
  [14709] = {.lex_state = 224, .external_lex_state = 100},
  [14710] = {.lex_state = 138},
  [14711] = {.lex_state = 138},
  [14712] = {.lex_state = 224},
  [14713] = {.lex_state = 138},
  [14714] = {.lex_state = 138, .external_lex_state = 119},
  [14715] = {.lex_state = 179},
  [14716] = {.lex_state = 224},
  [14717] = {.lex_state = 138},
  [14718] = {.lex_state = 224},
  [14719] = {.lex_state = 138},
  [14720] = {.lex_state = 179},
  [14721] = {.lex_state = 179},
  [14722] = {.lex_state = 224, .external_lex_state = 102},
  [14723] = {.lex_state = 224},
  [14724] = {.lex_state = 138},
  [14725] = {.lex_state = 138},
  [14726] = {.lex_state = 224, .external_lex_state = 101},
  [14727] = {.lex_state = 179},
  [14728] = {.lex_state = 138},
  [14729] = {.lex_state = 138, .external_lex_state = 114},
  [14730] = {.lex_state = 224, .external_lex_state = 101},
  [14731] = {.lex_state = 179},
  [14732] = {.lex_state = 224},
  [14733] = {.lex_state = 138},
  [14734] = {.lex_state = 138},
  [14735] = {.lex_state = 179},
  [14736] = {.lex_state = 224, .external_lex_state = 101},
  [14737] = {.lex_state = 138},
  [14738] = {.lex_state = 224, .external_lex_state = 99},
  [14739] = {.lex_state = 138},
  [14740] = {.lex_state = 224},
  [14741] = {.lex_state = 179},
  [14742] = {.lex_state = 138},
  [14743] = {.lex_state = 203},
  [14744] = {.lex_state = 224, .external_lex_state = 101},
  [14745] = {.lex_state = 224},
  [14746] = {.lex_state = 224, .external_lex_state = 100},
  [14747] = {.lex_state = 138},
  [14748] = {.lex_state = 138},
  [14749] = {.lex_state = 138},
  [14750] = {.lex_state = 224},
  [14751] = {.lex_state = 179},
  [14752] = {.lex_state = 138},
  [14753] = {.lex_state = 138},
  [14754] = {.lex_state = 179},
  [14755] = {.lex_state = 224},
  [14756] = {.lex_state = 224, .external_lex_state = 99},
  [14757] = {.lex_state = 179},
  [14758] = {.lex_state = 138},
  [14759] = {.lex_state = 224},
  [14760] = {.lex_state = 138},
  [14761] = {.lex_state = 224, .external_lex_state = 101},
  [14762] = {.lex_state = 224, .external_lex_state = 101},
  [14763] = {.lex_state = 224},
  [14764] = {.lex_state = 138},
  [14765] = {.lex_state = 179},
  [14766] = {.lex_state = 224, .external_lex_state = 100},
  [14767] = {.lex_state = 224, .external_lex_state = 101},
  [14768] = {.lex_state = 138},
  [14769] = {.lex_state = 224},
  [14770] = {.lex_state = 138},
  [14771] = {.lex_state = 138, .external_lex_state = 120},
  [14772] = {.lex_state = 219, .external_lex_state = 109},
  [14773] = {.lex_state = 224, .external_lex_state = 102},
  [14774] = {.lex_state = 201, .external_lex_state = 116},
  [14775] = {.lex_state = 138},
  [14776] = {.lex_state = 138},
  [14777] = {.lex_state = 224, .external_lex_state = 101},
  [14778] = {.lex_state = 224},
  [14779] = {.lex_state = 138, .external_lex_state = 122},
  [14780] = {.lex_state = 138, .external_lex_state = 122},
  [14781] = {.lex_state = 224},
  [14782] = {.lex_state = 224},
  [14783] = {.lex_state = 138},
  [14784] = {.lex_state = 224},
  [14785] = {.lex_state = 224, .external_lex_state = 100},
  [14786] = {.lex_state = 224, .external_lex_state = 100},
  [14787] = {.lex_state = 138},
  [14788] = {.lex_state = 138},
  [14789] = {.lex_state = 224},
  [14790] = {.lex_state = 138},
  [14791] = {.lex_state = 179},
  [14792] = {.lex_state = 138},
  [14793] = {.lex_state = 138},
  [14794] = {.lex_state = 224, .external_lex_state = 102},
  [14795] = {.lex_state = 138, .external_lex_state = 114},
  [14796] = {.lex_state = 179},
  [14797] = {.lex_state = 138},
  [14798] = {.lex_state = 138},
  [14799] = {.lex_state = 224, .external_lex_state = 101},
  [14800] = {.lex_state = 138, .external_lex_state = 114},
  [14801] = {.lex_state = 224, .external_lex_state = 99},
  [14802] = {.lex_state = 224, .external_lex_state = 100},
  [14803] = {.lex_state = 201, .external_lex_state = 116},
  [14804] = {.lex_state = 224, .external_lex_state = 100},
  [14805] = {.lex_state = 224, .external_lex_state = 100},
  [14806] = {.lex_state = 224, .external_lex_state = 101},
  [14807] = {.lex_state = 224, .external_lex_state = 100},
  [14808] = {.lex_state = 138},
  [14809] = {.lex_state = 138},
  [14810] = {.lex_state = 138},
  [14811] = {.lex_state = 138},
  [14812] = {.lex_state = 224, .external_lex_state = 99},
  [14813] = {.lex_state = 138},
  [14814] = {.lex_state = 179},
  [14815] = {.lex_state = 224, .external_lex_state = 100},
  [14816] = {.lex_state = 224, .external_lex_state = 100},
  [14817] = {.lex_state = 179},
  [14818] = {.lex_state = 138},
  [14819] = {.lex_state = 138},
  [14820] = {.lex_state = 138},
  [14821] = {.lex_state = 138, .external_lex_state = 114},
  [14822] = {.lex_state = 224, .external_lex_state = 99},
  [14823] = {.lex_state = 138},
  [14824] = {.lex_state = 224, .external_lex_state = 92},
  [14825] = {.lex_state = 224, .external_lex_state = 92},
  [14826] = {.lex_state = 138},
  [14827] = {.lex_state = 138},
  [14828] = {.lex_state = 138},
  [14829] = {.lex_state = 138},
  [14830] = {.lex_state = 138, .external_lex_state = 118},
  [14831] = {.lex_state = 224, .external_lex_state = 100},
  [14832] = {.lex_state = 138},
  [14833] = {.lex_state = 224, .external_lex_state = 100},
  [14834] = {.lex_state = 138},
  [14835] = {.lex_state = 138},
  [14836] = {.lex_state = 138},
  [14837] = {.lex_state = 220},
  [14838] = {.lex_state = 224, .external_lex_state = 99},
  [14839] = {.lex_state = 138},
  [14840] = {.lex_state = 224, .external_lex_state = 100},
  [14841] = {.lex_state = 138},
  [14842] = {.lex_state = 138},
  [14843] = {.lex_state = 224, .external_lex_state = 99},
  [14844] = {.lex_state = 179},
  [14845] = {.lex_state = 224, .external_lex_state = 99},
  [14846] = {.lex_state = 224, .external_lex_state = 99},
  [14847] = {.lex_state = 138},
  [14848] = {.lex_state = 179},
  [14849] = {.lex_state = 138},
  [14850] = {.lex_state = 138, .external_lex_state = 122},
  [14851] = {.lex_state = 224, .external_lex_state = 99},
  [14852] = {.lex_state = 224, .external_lex_state = 92},
  [14853] = {.lex_state = 138},
  [14854] = {.lex_state = 224, .external_lex_state = 92},
  [14855] = {.lex_state = 138},
  [14856] = {.lex_state = 224, .external_lex_state = 100},
  [14857] = {.lex_state = 138},
  [14858] = {.lex_state = 224, .external_lex_state = 100},
  [14859] = {.lex_state = 138},
  [14860] = {.lex_state = 224, .external_lex_state = 92},
  [14861] = {.lex_state = 179},
  [14862] = {.lex_state = 138},
  [14863] = {.lex_state = 138, .external_lex_state = 114},
  [14864] = {.lex_state = 179},
  [14865] = {.lex_state = 179},
  [14866] = {.lex_state = 224, .external_lex_state = 100},
  [14867] = {.lex_state = 224, .external_lex_state = 99},
  [14868] = {.lex_state = 138},
  [14869] = {.lex_state = 179},
  [14870] = {.lex_state = 138, .external_lex_state = 121},
  [14871] = {.lex_state = 138},
  [14872] = {.lex_state = 138},
  [14873] = {.lex_state = 138},
  [14874] = {.lex_state = 138},
  [14875] = {.lex_state = 179},
  [14876] = {.lex_state = 179},
  [14877] = {.lex_state = 138},
  [14878] = {.lex_state = 138},
  [14879] = {.lex_state = 138, .external_lex_state = 114},
  [14880] = {.lex_state = 138},
  [14881] = {.lex_state = 138},
  [14882] = {.lex_state = 138},
  [14883] = {.lex_state = 138},
  [14884] = {.lex_state = 224},
  [14885] = {.lex_state = 138},
  [14886] = {.lex_state = 138},
  [14887] = {.lex_state = 179},
  [14888] = {.lex_state = 138},
  [14889] = {.lex_state = 138},
  [14890] = {.lex_state = 138},
  [14891] = {.lex_state = 138, .external_lex_state = 118},
  [14892] = {.lex_state = 138},
  [14893] = {.lex_state = 138},
  [14894] = {.lex_state = 138, .external_lex_state = 118},
  [14895] = {.lex_state = 138},
  [14896] = {.lex_state = 138},
  [14897] = {.lex_state = 138},
  [14898] = {.lex_state = 224, .external_lex_state = 109},
  [14899] = {.lex_state = 224, .external_lex_state = 102},
  [14900] = {.lex_state = 138},
  [14901] = {.lex_state = 138},
  [14902] = {.lex_state = 138},
  [14903] = {.lex_state = 220},
  [14904] = {.lex_state = 138},
  [14905] = {.lex_state = 138},
  [14906] = {.lex_state = 219, .external_lex_state = 109},
  [14907] = {.lex_state = 138},
  [14908] = {.lex_state = 179},
  [14909] = {.lex_state = 138, .external_lex_state = 118},
  [14910] = {.lex_state = 201, .external_lex_state = 116},
  [14911] = {.lex_state = 138},
  [14912] = {.lex_state = 220},
  [14913] = {.lex_state = 138},
  [14914] = {.lex_state = 138},
  [14915] = {.lex_state = 138},
  [14916] = {.lex_state = 138},
  [14917] = {.lex_state = 138},
  [14918] = {.lex_state = 138, .external_lex_state = 122},
  [14919] = {.lex_state = 179},
  [14920] = {.lex_state = 138, .external_lex_state = 122},
  [14921] = {.lex_state = 201, .external_lex_state = 116},
  [14922] = {.lex_state = 138},
  [14923] = {.lex_state = 138},
  [14924] = {.lex_state = 138},
  [14925] = {.lex_state = 138},
  [14926] = {.lex_state = 224, .external_lex_state = 101},
  [14927] = {.lex_state = 224, .external_lex_state = 100},
  [14928] = {.lex_state = 138},
  [14929] = {.lex_state = 224, .external_lex_state = 101},
  [14930] = {.lex_state = 138},
  [14931] = {.lex_state = 138},
  [14932] = {.lex_state = 224, .external_lex_state = 101},
  [14933] = {.lex_state = 138},
  [14934] = {.lex_state = 138},
  [14935] = {.lex_state = 138},
  [14936] = {.lex_state = 138},
  [14937] = {.lex_state = 224, .external_lex_state = 101},
  [14938] = {.lex_state = 179},
  [14939] = {.lex_state = 138},
  [14940] = {.lex_state = 138, .external_lex_state = 122},
  [14941] = {.lex_state = 203},
  [14942] = {.lex_state = 138},
  [14943] = {.lex_state = 179},
  [14944] = {.lex_state = 138, .external_lex_state = 122},
  [14945] = {.lex_state = 138, .external_lex_state = 122},
  [14946] = {.lex_state = 138, .external_lex_state = 114},
  [14947] = {.lex_state = 224},
  [14948] = {.lex_state = 138},
  [14949] = {.lex_state = 224},
  [14950] = {.lex_state = 138},
  [14951] = {.lex_state = 138, .external_lex_state = 119},
  [14952] = {.lex_state = 138, .external_lex_state = 114},
  [14953] = {.lex_state = 179},
  [14954] = {.lex_state = 138, .external_lex_state = 119},
  [14955] = {.lex_state = 138},
  [14956] = {.lex_state = 179},
  [14957] = {.lex_state = 138},
  [14958] = {.lex_state = 179},
  [14959] = {.lex_state = 138},
  [14960] = {.lex_state = 138},
  [14961] = {.lex_state = 179},
  [14962] = {.lex_state = 138},
  [14963] = {.lex_state = 138},
  [14964] = {.lex_state = 138},
  [14965] = {.lex_state = 138},
  [14966] = {.lex_state = 138},
  [14967] = {.lex_state = 224},
  [14968] = {.lex_state = 224},
  [14969] = {.lex_state = 224},
  [14970] = {.lex_state = 224, .external_lex_state = 100},
  [14971] = {.lex_state = 138, .external_lex_state = 114},
  [14972] = {.lex_state = 138},
  [14973] = {.lex_state = 138},
  [14974] = {.lex_state = 138},
  [14975] = {.lex_state = 138},
  [14976] = {.lex_state = 224, .external_lex_state = 109},
  [14977] = {.lex_state = 224, .external_lex_state = 109},
  [14978] = {.lex_state = 179},
  [14979] = {.lex_state = 224, .external_lex_state = 100},
  [14980] = {.lex_state = 138, .external_lex_state = 122},
  [14981] = {.lex_state = 138, .external_lex_state = 121},
  [14982] = {.lex_state = 138},
  [14983] = {.lex_state = 224, .external_lex_state = 100},
  [14984] = {.lex_state = 224, .external_lex_state = 100},
  [14985] = {.lex_state = 224, .external_lex_state = 100},
  [14986] = {.lex_state = 138},
  [14987] = {.lex_state = 224, .external_lex_state = 99},
  [14988] = {.lex_state = 138},
  [14989] = {.lex_state = 138, .external_lex_state = 122},
  [14990] = {.lex_state = 138},
  [14991] = {.lex_state = 179},
  [14992] = {.lex_state = 138, .external_lex_state = 122},
  [14993] = {.lex_state = 138, .external_lex_state = 121},
  [14994] = {.lex_state = 224},
  [14995] = {.lex_state = 179},
  [14996] = {.lex_state = 224, .external_lex_state = 100},
  [14997] = {.lex_state = 179},
  [14998] = {.lex_state = 138},
  [14999] = {.lex_state = 224, .external_lex_state = 100},
  [15000] = {.lex_state = 224},
  [15001] = {.lex_state = 138},
  [15002] = {.lex_state = 138},
  [15003] = {.lex_state = 224, .external_lex_state = 92},
  [15004] = {.lex_state = 224, .external_lex_state = 92},
  [15005] = {.lex_state = 138},
  [15006] = {.lex_state = 138},
  [15007] = {.lex_state = 224, .external_lex_state = 92},
  [15008] = {.lex_state = 224, .external_lex_state = 92},
  [15009] = {.lex_state = 138, .external_lex_state = 120},
  [15010] = {.lex_state = 179},
  [15011] = {.lex_state = 224, .external_lex_state = 92},
  [15012] = {.lex_state = 224},
  [15013] = {.lex_state = 138, .external_lex_state = 120},
  [15014] = {.lex_state = 179},
  [15015] = {.lex_state = 224, .external_lex_state = 100},
  [15016] = {.lex_state = 138},
  [15017] = {.lex_state = 224, .external_lex_state = 92},
  [15018] = {.lex_state = 138},
  [15019] = {.lex_state = 224, .external_lex_state = 92},
  [15020] = {.lex_state = 224, .external_lex_state = 92},
  [15021] = {.lex_state = 224, .external_lex_state = 92},
  [15022] = {.lex_state = 224, .external_lex_state = 92},
  [15023] = {.lex_state = 224, .external_lex_state = 92},
  [15024] = {.lex_state = 224, .external_lex_state = 92},
  [15025] = {.lex_state = 224, .external_lex_state = 92},
  [15026] = {.lex_state = 138},
  [15027] = {.lex_state = 224, .external_lex_state = 92},
  [15028] = {.lex_state = 224, .external_lex_state = 92},
  [15029] = {.lex_state = 224, .external_lex_state = 92},
  [15030] = {.lex_state = 224, .external_lex_state = 92},
  [15031] = {.lex_state = 224, .external_lex_state = 92},
  [15032] = {.lex_state = 224, .external_lex_state = 92},
  [15033] = {.lex_state = 224, .external_lex_state = 92},
  [15034] = {.lex_state = 224, .external_lex_state = 92},
  [15035] = {.lex_state = 224, .external_lex_state = 92},
  [15036] = {.lex_state = 138},
  [15037] = {.lex_state = 224, .external_lex_state = 101},
  [15038] = {.lex_state = 224, .external_lex_state = 92},
  [15039] = {.lex_state = 138},
  [15040] = {.lex_state = 138},
  [15041] = {.lex_state = 224, .external_lex_state = 101},
  [15042] = {.lex_state = 179},
  [15043] = {.lex_state = 224, .external_lex_state = 100},
  [15044] = {.lex_state = 224},
  [15045] = {.lex_state = 224, .external_lex_state = 100},
  [15046] = {.lex_state = 138},
  [15047] = {.lex_state = 138},
  [15048] = {.lex_state = 224, .external_lex_state = 100},
  [15049] = {.lex_state = 203},
  [15050] = {.lex_state = 224, .external_lex_state = 101},
  [15051] = {.lex_state = 224, .external_lex_state = 100},
  [15052] = {.lex_state = 138},
  [15053] = {.lex_state = 224, .external_lex_state = 100},
  [15054] = {.lex_state = 224, .external_lex_state = 100},
  [15055] = {.lex_state = 224, .external_lex_state = 100},
  [15056] = {.lex_state = 138},
  [15057] = {.lex_state = 224, .external_lex_state = 92},
  [15058] = {.lex_state = 224, .external_lex_state = 92},
  [15059] = {.lex_state = 224, .external_lex_state = 101},
  [15060] = {.lex_state = 224, .external_lex_state = 92},
  [15061] = {.lex_state = 224, .external_lex_state = 92},
  [15062] = {.lex_state = 138},
  [15063] = {.lex_state = 138},
  [15064] = {.lex_state = 179},
  [15065] = {.lex_state = 138},
  [15066] = {.lex_state = 224, .external_lex_state = 101},
  [15067] = {.lex_state = 138},
  [15068] = {.lex_state = 138, .external_lex_state = 122},
  [15069] = {.lex_state = 138},
  [15070] = {.lex_state = 138},
  [15071] = {.lex_state = 138, .external_lex_state = 121},
  [15072] = {.lex_state = 219, .external_lex_state = 109},
  [15073] = {.lex_state = 224, .external_lex_state = 97},
  [15074] = {.lex_state = 224},
  [15075] = {.lex_state = 224},
  [15076] = {.lex_state = 224},
  [15077] = {.lex_state = 138, .external_lex_state = 118},
  [15078] = {.lex_state = 219, .external_lex_state = 109},
  [15079] = {.lex_state = 138},
  [15080] = {.lex_state = 138, .external_lex_state = 114},
  [15081] = {.lex_state = 224},
  [15082] = {.lex_state = 220},
  [15083] = {.lex_state = 219, .external_lex_state = 109},
  [15084] = {.lex_state = 224},
  [15085] = {.lex_state = 455},
  [15086] = {.lex_state = 224, .external_lex_state = 102},
  [15087] = {.lex_state = 138},
  [15088] = {.lex_state = 219, .external_lex_state = 109},
  [15089] = {.lex_state = 455},
  [15090] = {.lex_state = 138, .external_lex_state = 122},
  [15091] = {.lex_state = 219, .external_lex_state = 109},
  [15092] = {.lex_state = 455},
  [15093] = {.lex_state = 220},
  [15094] = {.lex_state = 224, .external_lex_state = 102},
  [15095] = {.lex_state = 455},
  [15096] = {.lex_state = 138, .external_lex_state = 122},
  [15097] = {.lex_state = 138, .external_lex_state = 119},
  [15098] = {.lex_state = 224},
  [15099] = {.lex_state = 224, .external_lex_state = 99},
  [15100] = {.lex_state = 220},
  [15101] = {.lex_state = 224, .external_lex_state = 99},
  [15102] = {.lex_state = 224, .external_lex_state = 99},
  [15103] = {.lex_state = 224, .external_lex_state = 99},
  [15104] = {.lex_state = 219, .external_lex_state = 109},
  [15105] = {.lex_state = 224},
  [15106] = {.lex_state = 224},
  [15107] = {.lex_state = 219, .external_lex_state = 109},
  [15108] = {.lex_state = 223},
  [15109] = {.lex_state = 224, .external_lex_state = 102},
  [15110] = {.lex_state = 224, .external_lex_state = 97},
  [15111] = {.lex_state = 219, .external_lex_state = 109},
  [15112] = {.lex_state = 224},
  [15113] = {.lex_state = 138, .external_lex_state = 119},
  [15114] = {.lex_state = 138, .external_lex_state = 114},
  [15115] = {.lex_state = 138},
  [15116] = {.lex_state = 224, .external_lex_state = 102},
  [15117] = {.lex_state = 138},
  [15118] = {.lex_state = 138, .external_lex_state = 120},
  [15119] = {.lex_state = 455},
  [15120] = {.lex_state = 138},
  [15121] = {.lex_state = 138},
  [15122] = {.lex_state = 138},
  [15123] = {.lex_state = 138},
  [15124] = {.lex_state = 138},
  [15125] = {.lex_state = 138},
  [15126] = {.lex_state = 138},
  [15127] = {.lex_state = 138},
  [15128] = {.lex_state = 138, .external_lex_state = 114},
  [15129] = {.lex_state = 224, .external_lex_state = 102},
  [15130] = {.lex_state = 224, .external_lex_state = 99},
  [15131] = {.lex_state = 138, .external_lex_state = 119},
  [15132] = {.lex_state = 138},
  [15133] = {.lex_state = 138},
  [15134] = {.lex_state = 138},
  [15135] = {.lex_state = 138},
  [15136] = {.lex_state = 138, .external_lex_state = 120},
  [15137] = {.lex_state = 138},
  [15138] = {.lex_state = 224},
  [15139] = {.lex_state = 138},
  [15140] = {.lex_state = 138},
  [15141] = {.lex_state = 138, .external_lex_state = 118},
  [15142] = {.lex_state = 138},
  [15143] = {.lex_state = 138},
  [15144] = {.lex_state = 138},
  [15145] = {.lex_state = 138},
  [15146] = {.lex_state = 138},
  [15147] = {.lex_state = 138},
  [15148] = {.lex_state = 138, .external_lex_state = 121},
  [15149] = {.lex_state = 138},
  [15150] = {.lex_state = 138},
  [15151] = {.lex_state = 138},
  [15152] = {.lex_state = 138},
  [15153] = {.lex_state = 138},
  [15154] = {.lex_state = 138},
  [15155] = {.lex_state = 138},
  [15156] = {.lex_state = 138},
  [15157] = {.lex_state = 138},
  [15158] = {.lex_state = 138},
  [15159] = {.lex_state = 138},
  [15160] = {.lex_state = 138},
  [15161] = {.lex_state = 138},
  [15162] = {.lex_state = 138},
  [15163] = {.lex_state = 220},
  [15164] = {.lex_state = 138},
  [15165] = {.lex_state = 138},
  [15166] = {.lex_state = 138},
  [15167] = {.lex_state = 138, .external_lex_state = 118},
  [15168] = {.lex_state = 138},
  [15169] = {.lex_state = 138},
  [15170] = {.lex_state = 138},
  [15171] = {.lex_state = 224, .external_lex_state = 97},
  [15172] = {.lex_state = 138},
  [15173] = {.lex_state = 138},
  [15174] = {.lex_state = 224},
  [15175] = {.lex_state = 138},
  [15176] = {.lex_state = 138},
  [15177] = {.lex_state = 138},
  [15178] = {.lex_state = 138},
  [15179] = {.lex_state = 224, .external_lex_state = 99},
  [15180] = {.lex_state = 224},
  [15181] = {.lex_state = 224},
  [15182] = {.lex_state = 138},
  [15183] = {.lex_state = 224, .external_lex_state = 99},
  [15184] = {.lex_state = 224},
  [15185] = {.lex_state = 224, .external_lex_state = 99},
  [15186] = {.lex_state = 138},
  [15187] = {.lex_state = 138, .external_lex_state = 119},
  [15188] = {.lex_state = 138},
  [15189] = {.lex_state = 138},
  [15190] = {.lex_state = 455},
  [15191] = {.lex_state = 138, .external_lex_state = 118},
  [15192] = {.lex_state = 138},
  [15193] = {.lex_state = 138, .external_lex_state = 119},
  [15194] = {.lex_state = 224},
  [15195] = {.lex_state = 224},
  [15196] = {.lex_state = 224},
  [15197] = {.lex_state = 138},
  [15198] = {.lex_state = 224},
  [15199] = {.lex_state = 224},
  [15200] = {.lex_state = 224, .external_lex_state = 99},
  [15201] = {.lex_state = 224},
  [15202] = {.lex_state = 138},
  [15203] = {.lex_state = 224},
  [15204] = {.lex_state = 138},
  [15205] = {.lex_state = 138},
  [15206] = {.lex_state = 138},
  [15207] = {.lex_state = 138},
  [15208] = {.lex_state = 138, .external_lex_state = 114},
  [15209] = {.lex_state = 219, .external_lex_state = 109},
  [15210] = {.lex_state = 138},
  [15211] = {.lex_state = 224},
  [15212] = {.lex_state = 224},
  [15213] = {.lex_state = 138, .external_lex_state = 120},
  [15214] = {.lex_state = 138},
  [15215] = {.lex_state = 138},
  [15216] = {.lex_state = 138},
  [15217] = {.lex_state = 138},
  [15218] = {.lex_state = 138},
  [15219] = {.lex_state = 138},
  [15220] = {.lex_state = 138},
  [15221] = {.lex_state = 219, .external_lex_state = 109},
  [15222] = {.lex_state = 138},
  [15223] = {.lex_state = 138},
  [15224] = {.lex_state = 224},
  [15225] = {.lex_state = 223},
  [15226] = {.lex_state = 138},
  [15227] = {.lex_state = 224},
  [15228] = {.lex_state = 138, .external_lex_state = 121},
  [15229] = {.lex_state = 138},
  [15230] = {.lex_state = 138},
  [15231] = {.lex_state = 138},
  [15232] = {.lex_state = 138},
  [15233] = {.lex_state = 224},
  [15234] = {.lex_state = 224, .external_lex_state = 99},
  [15235] = {.lex_state = 138},
  [15236] = {.lex_state = 138},
  [15237] = {.lex_state = 138},
  [15238] = {.lex_state = 138},
  [15239] = {.lex_state = 224, .external_lex_state = 99},
  [15240] = {.lex_state = 224, .external_lex_state = 99},
  [15241] = {.lex_state = 224, .external_lex_state = 99},
  [15242] = {.lex_state = 224, .external_lex_state = 99},
  [15243] = {.lex_state = 224},
  [15244] = {.lex_state = 220},
  [15245] = {.lex_state = 138},
  [15246] = {.lex_state = 138},
  [15247] = {.lex_state = 138},
  [15248] = {.lex_state = 224},
  [15249] = {.lex_state = 138},
  [15250] = {.lex_state = 138, .external_lex_state = 121},
  [15251] = {.lex_state = 224},
  [15252] = {.lex_state = 138, .external_lex_state = 119},
  [15253] = {.lex_state = 224, .external_lex_state = 99},
  [15254] = {.lex_state = 224, .external_lex_state = 99},
  [15255] = {.lex_state = 224},
  [15256] = {.lex_state = 138},
  [15257] = {.lex_state = 138},
  [15258] = {.lex_state = 138},
  [15259] = {.lex_state = 138, .external_lex_state = 122},
  [15260] = {.lex_state = 224, .external_lex_state = 97},
  [15261] = {.lex_state = 138, .external_lex_state = 119},
  [15262] = {.lex_state = 138},
  [15263] = {.lex_state = 138},
  [15264] = {.lex_state = 224},
  [15265] = {.lex_state = 455},
  [15266] = {.lex_state = 224, .external_lex_state = 99},
  [15267] = {.lex_state = 138},
  [15268] = {.lex_state = 224},
  [15269] = {.lex_state = 224},
  [15270] = {.lex_state = 224, .external_lex_state = 99},
  [15271] = {.lex_state = 138},
  [15272] = {.lex_state = 224, .external_lex_state = 99},
  [15273] = {.lex_state = 138},
  [15274] = {.lex_state = 138, .external_lex_state = 120},
  [15275] = {.lex_state = 224},
  [15276] = {.lex_state = 224, .external_lex_state = 97},
  [15277] = {.lex_state = 138},
  [15278] = {.lex_state = 138},
  [15279] = {.lex_state = 224, .external_lex_state = 97},
  [15280] = {.lex_state = 224},
  [15281] = {.lex_state = 138},
  [15282] = {.lex_state = 138},
  [15283] = {.lex_state = 138},
  [15284] = {.lex_state = 138, .external_lex_state = 120},
  [15285] = {.lex_state = 224},
  [15286] = {.lex_state = 138, .external_lex_state = 114},
  [15287] = {.lex_state = 224, .external_lex_state = 99},
  [15288] = {.lex_state = 219, .external_lex_state = 109},
  [15289] = {.lex_state = 138, .external_lex_state = 119},
  [15290] = {.lex_state = 138, .external_lex_state = 119},
  [15291] = {.lex_state = 138, .external_lex_state = 119},
  [15292] = {.lex_state = 224, .external_lex_state = 99},
  [15293] = {.lex_state = 224, .external_lex_state = 97},
  [15294] = {.lex_state = 224},
  [15295] = {.lex_state = 224},
  [15296] = {.lex_state = 224},
  [15297] = {.lex_state = 224},
  [15298] = {.lex_state = 224},
  [15299] = {.lex_state = 138},
  [15300] = {.lex_state = 224},
  [15301] = {.lex_state = 138, .external_lex_state = 114},
  [15302] = {.lex_state = 138, .external_lex_state = 119},
  [15303] = {.lex_state = 138},
  [15304] = {.lex_state = 224},
  [15305] = {.lex_state = 224},
  [15306] = {.lex_state = 224, .external_lex_state = 99},
  [15307] = {.lex_state = 138, .external_lex_state = 119},
  [15308] = {.lex_state = 224},
  [15309] = {.lex_state = 138},
  [15310] = {.lex_state = 455},
  [15311] = {.lex_state = 224},
  [15312] = {.lex_state = 219, .external_lex_state = 109},
  [15313] = {.lex_state = 224, .external_lex_state = 102},
  [15314] = {.lex_state = 224},
  [15315] = {.lex_state = 138},
  [15316] = {.lex_state = 219, .external_lex_state = 109},
  [15317] = {.lex_state = 224},
  [15318] = {.lex_state = 224, .external_lex_state = 99},
  [15319] = {.lex_state = 224},
  [15320] = {.lex_state = 224},
  [15321] = {.lex_state = 224},
  [15322] = {.lex_state = 138},
  [15323] = {.lex_state = 224},
  [15324] = {.lex_state = 224},
  [15325] = {.lex_state = 219, .external_lex_state = 109},
  [15326] = {.lex_state = 224},
  [15327] = {.lex_state = 224},
  [15328] = {.lex_state = 224},
  [15329] = {.lex_state = 224, .external_lex_state = 99},
  [15330] = {.lex_state = 224},
  [15331] = {.lex_state = 224},
  [15332] = {.lex_state = 138, .external_lex_state = 119},
  [15333] = {.lex_state = 224},
  [15334] = {.lex_state = 224},
  [15335] = {.lex_state = 138, .external_lex_state = 119},
  [15336] = {.lex_state = 224},
  [15337] = {.lex_state = 224},
  [15338] = {.lex_state = 224},
  [15339] = {.lex_state = 224},
  [15340] = {.lex_state = 138, .external_lex_state = 114},
  [15341] = {.lex_state = 224},
  [15342] = {.lex_state = 224},
  [15343] = {.lex_state = 224},
  [15344] = {.lex_state = 224},
  [15345] = {.lex_state = 224},
  [15346] = {.lex_state = 224},
  [15347] = {.lex_state = 224},
  [15348] = {.lex_state = 224},
  [15349] = {.lex_state = 224},
  [15350] = {.lex_state = 224},
  [15351] = {.lex_state = 455},
  [15352] = {.lex_state = 138},
  [15353] = {.lex_state = 224},
  [15354] = {.lex_state = 224, .external_lex_state = 99},
  [15355] = {.lex_state = 224},
  [15356] = {.lex_state = 138},
  [15357] = {.lex_state = 138},
  [15358] = {.lex_state = 224, .external_lex_state = 102},
  [15359] = {.lex_state = 138},
  [15360] = {.lex_state = 179},
  [15361] = {.lex_state = 179},
  [15362] = {.lex_state = 138, .external_lex_state = 120},
  [15363] = {.lex_state = 179},
  [15364] = {.lex_state = 179},
  [15365] = {.lex_state = 138},
  [15366] = {.lex_state = 179},
  [15367] = {.lex_state = 179},
  [15368] = {.lex_state = 179},
  [15369] = {.lex_state = 455},
  [15370] = {.lex_state = 179},
  [15371] = {.lex_state = 179},
  [15372] = {.lex_state = 224, .external_lex_state = 108},
  [15373] = {.lex_state = 138, .external_lex_state = 114},
  [15374] = {.lex_state = 179},
  [15375] = {.lex_state = 138},
  [15376] = {.lex_state = 179},
  [15377] = {.lex_state = 179},
  [15378] = {.lex_state = 224, .external_lex_state = 108},
  [15379] = {.lex_state = 179},
  [15380] = {.lex_state = 138},
  [15381] = {.lex_state = 179},
  [15382] = {.lex_state = 199, .external_lex_state = 116},
  [15383] = {.lex_state = 138, .external_lex_state = 122},
  [15384] = {.lex_state = 138},
  [15385] = {.lex_state = 220},
  [15386] = {.lex_state = 224, .external_lex_state = 108},
  [15387] = {.lex_state = 179},
  [15388] = {.lex_state = 179},
  [15389] = {.lex_state = 179},
  [15390] = {.lex_state = 138, .external_lex_state = 121},
  [15391] = {.lex_state = 455},
  [15392] = {.lex_state = 138},
  [15393] = {.lex_state = 179},
  [15394] = {.lex_state = 138, .external_lex_state = 120},
  [15395] = {.lex_state = 138, .external_lex_state = 121},
  [15396] = {.lex_state = 179},
  [15397] = {.lex_state = 179},
  [15398] = {.lex_state = 138},
  [15399] = {.lex_state = 138, .external_lex_state = 114},
  [15400] = {.lex_state = 455},
  [15401] = {.lex_state = 179},
  [15402] = {.lex_state = 179},
  [15403] = {.lex_state = 179},
  [15404] = {.lex_state = 138},
  [15405] = {.lex_state = 179},
  [15406] = {.lex_state = 179},
  [15407] = {.lex_state = 138},
  [15408] = {.lex_state = 138, .external_lex_state = 121},
  [15409] = {.lex_state = 179},
  [15410] = {.lex_state = 138, .external_lex_state = 120},
  [15411] = {.lex_state = 220},
  [15412] = {.lex_state = 179},
  [15413] = {.lex_state = 179},
  [15414] = {.lex_state = 179},
  [15415] = {.lex_state = 179},
  [15416] = {.lex_state = 138, .external_lex_state = 114},
  [15417] = {.lex_state = 179},
  [15418] = {.lex_state = 179},
  [15419] = {.lex_state = 138, .external_lex_state = 121},
  [15420] = {.lex_state = 179},
  [15421] = {.lex_state = 138},
  [15422] = {.lex_state = 138, .external_lex_state = 114},
  [15423] = {.lex_state = 138, .external_lex_state = 121},
  [15424] = {.lex_state = 138},
  [15425] = {.lex_state = 220},
  [15426] = {.lex_state = 179},
  [15427] = {.lex_state = 220},
  [15428] = {.lex_state = 138, .external_lex_state = 121},
  [15429] = {.lex_state = 138, .external_lex_state = 114},
  [15430] = {.lex_state = 138},
  [15431] = {.lex_state = 224, .external_lex_state = 108},
  [15432] = {.lex_state = 138, .external_lex_state = 121},
  [15433] = {.lex_state = 179},
  [15434] = {.lex_state = 138, .external_lex_state = 121},
  [15435] = {.lex_state = 179},
  [15436] = {.lex_state = 138},
  [15437] = {.lex_state = 179},
  [15438] = {.lex_state = 220},
  [15439] = {.lex_state = 138, .external_lex_state = 114},
  [15440] = {.lex_state = 455},
  [15441] = {.lex_state = 138},
  [15442] = {.lex_state = 138, .external_lex_state = 122},
  [15443] = {.lex_state = 138},
  [15444] = {.lex_state = 179},
  [15445] = {.lex_state = 179},
  [15446] = {.lex_state = 179},
  [15447] = {.lex_state = 138, .external_lex_state = 122},
  [15448] = {.lex_state = 179},
  [15449] = {.lex_state = 138},
  [15450] = {.lex_state = 179},
  [15451] = {.lex_state = 179},
  [15452] = {.lex_state = 220},
  [15453] = {.lex_state = 179},
  [15454] = {.lex_state = 138, .external_lex_state = 120},
  [15455] = {.lex_state = 179},
  [15456] = {.lex_state = 224, .external_lex_state = 97},
  [15457] = {.lex_state = 220},
  [15458] = {.lex_state = 138},
  [15459] = {.lex_state = 179},
  [15460] = {.lex_state = 179},
  [15461] = {.lex_state = 220},
  [15462] = {.lex_state = 138},
  [15463] = {.lex_state = 138},
  [15464] = {.lex_state = 220},
  [15465] = {.lex_state = 179},
  [15466] = {.lex_state = 179},
  [15467] = {.lex_state = 138, .external_lex_state = 121},
  [15468] = {.lex_state = 138, .external_lex_state = 120},
  [15469] = {.lex_state = 179},
  [15470] = {.lex_state = 455},
  [15471] = {.lex_state = 179},
  [15472] = {.lex_state = 179},
  [15473] = {.lex_state = 224, .external_lex_state = 102},
  [15474] = {.lex_state = 179},
  [15475] = {.lex_state = 138, .external_lex_state = 122},
  [15476] = {.lex_state = 138},
  [15477] = {.lex_state = 138, .external_lex_state = 114},
  [15478] = {.lex_state = 199, .external_lex_state = 116},
  [15479] = {.lex_state = 138, .external_lex_state = 114},
  [15480] = {.lex_state = 179},
  [15481] = {.lex_state = 138, .external_lex_state = 120},
  [15482] = {.lex_state = 179},
  [15483] = {.lex_state = 138},
  [15484] = {.lex_state = 138, .external_lex_state = 120},
  [15485] = {.lex_state = 220},
  [15486] = {.lex_state = 455},
  [15487] = {.lex_state = 138, .external_lex_state = 114},
  [15488] = {.lex_state = 179},
  [15489] = {.lex_state = 179},
  [15490] = {.lex_state = 138, .external_lex_state = 120},
  [15491] = {.lex_state = 179},
  [15492] = {.lex_state = 179},
  [15493] = {.lex_state = 179},
  [15494] = {.lex_state = 179},
  [15495] = {.lex_state = 179},
  [15496] = {.lex_state = 224, .external_lex_state = 97},
  [15497] = {.lex_state = 224, .external_lex_state = 108},
  [15498] = {.lex_state = 138, .external_lex_state = 114},
  [15499] = {.lex_state = 138, .external_lex_state = 121},
  [15500] = {.lex_state = 179},
  [15501] = {.lex_state = 138, .external_lex_state = 114},
  [15502] = {.lex_state = 179},
  [15503] = {.lex_state = 138, .external_lex_state = 114},
  [15504] = {.lex_state = 224, .external_lex_state = 108},
  [15505] = {.lex_state = 138},
  [15506] = {.lex_state = 138, .external_lex_state = 120},
  [15507] = {.lex_state = 179},
  [15508] = {.lex_state = 220},
  [15509] = {.lex_state = 179},
  [15510] = {.lex_state = 138, .external_lex_state = 120},
  [15511] = {.lex_state = 455},
  [15512] = {.lex_state = 179},
  [15513] = {.lex_state = 138, .external_lex_state = 121},
  [15514] = {.lex_state = 138, .external_lex_state = 120},
  [15515] = {.lex_state = 220},
  [15516] = {.lex_state = 138},
  [15517] = {.lex_state = 179},
  [15518] = {.lex_state = 199, .external_lex_state = 116},
  [15519] = {.lex_state = 179},
  [15520] = {.lex_state = 179},
  [15521] = {.lex_state = 179},
  [15522] = {.lex_state = 179},
  [15523] = {.lex_state = 179},
  [15524] = {.lex_state = 179},
  [15525] = {.lex_state = 138, .external_lex_state = 121},
  [15526] = {.lex_state = 138, .external_lex_state = 120},
  [15527] = {.lex_state = 138},
  [15528] = {.lex_state = 138},
  [15529] = {.lex_state = 224, .external_lex_state = 102},
  [15530] = {.lex_state = 138, .external_lex_state = 120},
  [15531] = {.lex_state = 179},
  [15532] = {.lex_state = 138, .external_lex_state = 114},
  [15533] = {.lex_state = 138},
  [15534] = {.lex_state = 455},
  [15535] = {.lex_state = 138, .external_lex_state = 120},
  [15536] = {.lex_state = 138, .external_lex_state = 121},
  [15537] = {.lex_state = 179},
  [15538] = {.lex_state = 220},
  [15539] = {.lex_state = 179},
  [15540] = {.lex_state = 138},
  [15541] = {.lex_state = 224, .external_lex_state = 108},
  [15542] = {.lex_state = 224, .external_lex_state = 102},
  [15543] = {.lex_state = 179},
  [15544] = {.lex_state = 179},
  [15545] = {.lex_state = 220},
  [15546] = {.lex_state = 138, .external_lex_state = 114},
  [15547] = {.lex_state = 138, .external_lex_state = 114},
  [15548] = {.lex_state = 138, .external_lex_state = 114},
  [15549] = {.lex_state = 224, .external_lex_state = 97},
  [15550] = {.lex_state = 138, .external_lex_state = 121},
  [15551] = {.lex_state = 138},
  [15552] = {.lex_state = 138, .external_lex_state = 114},
  [15553] = {.lex_state = 455},
  [15554] = {.lex_state = 455},
  [15555] = {.lex_state = 138},
  [15556] = {.lex_state = 179},
  [15557] = {.lex_state = 179},
  [15558] = {.lex_state = 203, .external_lex_state = 123},
  [15559] = {.lex_state = 138, .external_lex_state = 118},
  [15560] = {.lex_state = 179},
  [15561] = {.lex_state = 455},
  [15562] = {.lex_state = 203, .external_lex_state = 123},
  [15563] = {.lex_state = 203, .external_lex_state = 123},
  [15564] = {.lex_state = 138, .external_lex_state = 122},
  [15565] = {.lex_state = 203, .external_lex_state = 123},
  [15566] = {.lex_state = 455},
  [15567] = {.lex_state = 203, .external_lex_state = 123},
  [15568] = {.lex_state = 179},
  [15569] = {.lex_state = 203, .external_lex_state = 124},
  [15570] = {.lex_state = 455},
  [15571] = {.lex_state = 203, .external_lex_state = 124},
  [15572] = {.lex_state = 203, .external_lex_state = 123},
  [15573] = {.lex_state = 203, .external_lex_state = 124},
  [15574] = {.lex_state = 138, .external_lex_state = 122},
  [15575] = {.lex_state = 203, .external_lex_state = 124},
  [15576] = {.lex_state = 203, .external_lex_state = 123},
  [15577] = {.lex_state = 138, .external_lex_state = 118},
  [15578] = {.lex_state = 203, .external_lex_state = 123},
  [15579] = {.lex_state = 203, .external_lex_state = 124},
  [15580] = {.lex_state = 179},
  [15581] = {.lex_state = 203, .external_lex_state = 124},
  [15582] = {.lex_state = 203, .external_lex_state = 124},
  [15583] = {.lex_state = 138, .external_lex_state = 114},
  [15584] = {.lex_state = 224, .external_lex_state = 102},
  [15585] = {.lex_state = 203, .external_lex_state = 123},
  [15586] = {.lex_state = 224, .external_lex_state = 97},
  [15587] = {.lex_state = 203, .external_lex_state = 124},
  [15588] = {.lex_state = 224, .external_lex_state = 97},
  [15589] = {.lex_state = 203, .external_lex_state = 123},
  [15590] = {.lex_state = 203, .external_lex_state = 124},
  [15591] = {.lex_state = 203, .external_lex_state = 124},
  [15592] = {.lex_state = 203, .external_lex_state = 123},
  [15593] = {.lex_state = 138, .external_lex_state = 122},
  [15594] = {.lex_state = 179},
  [15595] = {.lex_state = 455},
  [15596] = {.lex_state = 203, .external_lex_state = 124},
  [15597] = {.lex_state = 203, .external_lex_state = 124},
  [15598] = {.lex_state = 455},
  [15599] = {.lex_state = 203, .external_lex_state = 124},
  [15600] = {.lex_state = 179},
  [15601] = {.lex_state = 203, .external_lex_state = 123},
  [15602] = {.lex_state = 224, .external_lex_state = 97},
  [15603] = {.lex_state = 179},
  [15604] = {.lex_state = 455},
  [15605] = {.lex_state = 203, .external_lex_state = 123},
  [15606] = {.lex_state = 203, .external_lex_state = 124},
  [15607] = {.lex_state = 203, .external_lex_state = 124},
  [15608] = {.lex_state = 138, .external_lex_state = 122},
  [15609] = {.lex_state = 224, .external_lex_state = 97},
  [15610] = {.lex_state = 179},
  [15611] = {.lex_state = 203, .external_lex_state = 123},
  [15612] = {.lex_state = 138, .external_lex_state = 118},
  [15613] = {.lex_state = 203, .external_lex_state = 123},
  [15614] = {.lex_state = 138, .external_lex_state = 122},
  [15615] = {.lex_state = 455},
  [15616] = {.lex_state = 203, .external_lex_state = 123},
  [15617] = {.lex_state = 455},
  [15618] = {.lex_state = 201, .external_lex_state = 116},
  [15619] = {.lex_state = 203, .external_lex_state = 124},
  [15620] = {.lex_state = 203, .external_lex_state = 123},
  [15621] = {.lex_state = 138, .external_lex_state = 118},
  [15622] = {.lex_state = 138},
  [15623] = {.lex_state = 179},
  [15624] = {.lex_state = 203, .external_lex_state = 124},
  [15625] = {.lex_state = 455},
  [15626] = {.lex_state = 203, .external_lex_state = 124},
  [15627] = {.lex_state = 224, .external_lex_state = 108},
  [15628] = {.lex_state = 203, .external_lex_state = 124},
  [15629] = {.lex_state = 138, .external_lex_state = 122},
  [15630] = {.lex_state = 224, .external_lex_state = 102},
  [15631] = {.lex_state = 179},
  [15632] = {.lex_state = 138, .external_lex_state = 122},
  [15633] = {.lex_state = 138, .external_lex_state = 118},
  [15634] = {.lex_state = 203, .external_lex_state = 123},
  [15635] = {.lex_state = 138, .external_lex_state = 118},
  [15636] = {.lex_state = 179},
  [15637] = {.lex_state = 203, .external_lex_state = 124},
  [15638] = {.lex_state = 179},
  [15639] = {.lex_state = 224, .external_lex_state = 102},
  [15640] = {.lex_state = 455},
  [15641] = {.lex_state = 203, .external_lex_state = 124},
  [15642] = {.lex_state = 203, .external_lex_state = 123},
  [15643] = {.lex_state = 203, .external_lex_state = 124},
  [15644] = {.lex_state = 203, .external_lex_state = 124},
  [15645] = {.lex_state = 203, .external_lex_state = 123},
  [15646] = {.lex_state = 179},
  [15647] = {.lex_state = 203, .external_lex_state = 124},
  [15648] = {.lex_state = 179},
  [15649] = {.lex_state = 224, .external_lex_state = 102},
  [15650] = {.lex_state = 203, .external_lex_state = 124},
  [15651] = {.lex_state = 203, .external_lex_state = 123},
  [15652] = {.lex_state = 455},
  [15653] = {.lex_state = 203, .external_lex_state = 124},
  [15654] = {.lex_state = 455},
  [15655] = {.lex_state = 179},
  [15656] = {.lex_state = 203, .external_lex_state = 123},
  [15657] = {.lex_state = 203, .external_lex_state = 124},
  [15658] = {.lex_state = 203, .external_lex_state = 123},
  [15659] = {.lex_state = 203, .external_lex_state = 123},
  [15660] = {.lex_state = 203, .external_lex_state = 123},
  [15661] = {.lex_state = 179},
  [15662] = {.lex_state = 455},
  [15663] = {.lex_state = 138, .external_lex_state = 118},
  [15664] = {.lex_state = 203, .external_lex_state = 124},
  [15665] = {.lex_state = 455},
  [15666] = {.lex_state = 138, .external_lex_state = 122},
  [15667] = {.lex_state = 203, .external_lex_state = 124},
  [15668] = {.lex_state = 138, .external_lex_state = 122},
  [15669] = {.lex_state = 203, .external_lex_state = 123},
  [15670] = {.lex_state = 203, .external_lex_state = 123},
  [15671] = {.lex_state = 138, .external_lex_state = 118},
  [15672] = {.lex_state = 179},
  [15673] = {.lex_state = 455},
  [15674] = {.lex_state = 138, .external_lex_state = 122},
  [15675] = {.lex_state = 203, .external_lex_state = 123},
  [15676] = {.lex_state = 138, .external_lex_state = 122},
  [15677] = {.lex_state = 203, .external_lex_state = 123},
  [15678] = {.lex_state = 179},
  [15679] = {.lex_state = 203, .external_lex_state = 124},
  [15680] = {.lex_state = 201, .external_lex_state = 116},
  [15681] = {.lex_state = 203, .external_lex_state = 124},
  [15682] = {.lex_state = 203, .external_lex_state = 124},
  [15683] = {.lex_state = 179},
  [15684] = {.lex_state = 203, .external_lex_state = 123},
  [15685] = {.lex_state = 138, .external_lex_state = 122},
  [15686] = {.lex_state = 203, .external_lex_state = 123},
  [15687] = {.lex_state = 203, .external_lex_state = 123},
  [15688] = {.lex_state = 203, .external_lex_state = 124},
  [15689] = {.lex_state = 455},
  [15690] = {.lex_state = 203, .external_lex_state = 123},
  [15691] = {.lex_state = 201, .external_lex_state = 116},
  [15692] = {.lex_state = 203, .external_lex_state = 124},
  [15693] = {.lex_state = 179},
  [15694] = {.lex_state = 203, .external_lex_state = 123},
  [15695] = {.lex_state = 179},
  [15696] = {.lex_state = 455},
  [15697] = {.lex_state = 224, .external_lex_state = 108},
  [15698] = {.lex_state = 203, .external_lex_state = 124},
  [15699] = {.lex_state = 179},
  [15700] = {.lex_state = 203, .external_lex_state = 123},
  [15701] = {.lex_state = 179},
  [15702] = {.lex_state = 203, .external_lex_state = 124},
  [15703] = {.lex_state = 203, .external_lex_state = 124},
  [15704] = {.lex_state = 203, .external_lex_state = 124},
  [15705] = {.lex_state = 179},
  [15706] = {.lex_state = 138, .external_lex_state = 122},
  [15707] = {.lex_state = 224, .external_lex_state = 108},
  [15708] = {.lex_state = 203, .external_lex_state = 123},
  [15709] = {.lex_state = 203, .external_lex_state = 124},
  [15710] = {.lex_state = 203, .external_lex_state = 123},
  [15711] = {.lex_state = 203, .external_lex_state = 123},
  [15712] = {.lex_state = 455},
  [15713] = {.lex_state = 203, .external_lex_state = 123},
  [15714] = {.lex_state = 138, .external_lex_state = 118},
  [15715] = {.lex_state = 203, .external_lex_state = 123},
  [15716] = {.lex_state = 203, .external_lex_state = 124},
  [15717] = {.lex_state = 203, .external_lex_state = 123},
  [15718] = {.lex_state = 203, .external_lex_state = 123},
  [15719] = {.lex_state = 138, .external_lex_state = 122},
  [15720] = {.lex_state = 179},
  [15721] = {.lex_state = 203, .external_lex_state = 123},
  [15722] = {.lex_state = 203, .external_lex_state = 124},
  [15723] = {.lex_state = 179},
  [15724] = {.lex_state = 203, .external_lex_state = 123},
  [15725] = {.lex_state = 203, .external_lex_state = 124},
  [15726] = {.lex_state = 179},
  [15727] = {.lex_state = 203, .external_lex_state = 123},
  [15728] = {.lex_state = 203, .external_lex_state = 124},
  [15729] = {.lex_state = 203, .external_lex_state = 124},
  [15730] = {.lex_state = 179},
  [15731] = {.lex_state = 203, .external_lex_state = 124},
  [15732] = {.lex_state = 203, .external_lex_state = 123},
  [15733] = {.lex_state = 224, .external_lex_state = 97},
  [15734] = {.lex_state = 224, .external_lex_state = 102},
  [15735] = {.lex_state = 179},
  [15736] = {.lex_state = 203, .external_lex_state = 124},
  [15737] = {.lex_state = 203, .external_lex_state = 124},
  [15738] = {.lex_state = 203, .external_lex_state = 123},
  [15739] = {.lex_state = 179},
  [15740] = {.lex_state = 203, .external_lex_state = 123},
  [15741] = {.lex_state = 203, .external_lex_state = 124},
  [15742] = {.lex_state = 203, .external_lex_state = 123},
  [15743] = {.lex_state = 203, .external_lex_state = 124},
  [15744] = {.lex_state = 224},
  [15745] = {.lex_state = 203, .external_lex_state = 125},
  [15746] = {.lex_state = 203, .external_lex_state = 125},
  [15747] = {.lex_state = 203, .external_lex_state = 126},
  [15748] = {.lex_state = 203, .external_lex_state = 126},
  [15749] = {.lex_state = 203, .external_lex_state = 125},
  [15750] = {.lex_state = 203, .external_lex_state = 125},
  [15751] = {.lex_state = 203, .external_lex_state = 126},
  [15752] = {.lex_state = 224},
  [15753] = {.lex_state = 224},
  [15754] = {.lex_state = 457},
  [15755] = {.lex_state = 203, .external_lex_state = 125},
  [15756] = {.lex_state = 203, .external_lex_state = 125},
  [15757] = {.lex_state = 203, .external_lex_state = 126},
  [15758] = {.lex_state = 203, .external_lex_state = 126},
  [15759] = {.lex_state = 203, .external_lex_state = 125},
  [15760] = {.lex_state = 203, .external_lex_state = 126},
  [15761] = {.lex_state = 224},
  [15762] = {.lex_state = 224},
  [15763] = {.lex_state = 203, .external_lex_state = 126},
  [15764] = {.lex_state = 203, .external_lex_state = 125},
  [15765] = {.lex_state = 203, .external_lex_state = 125},
  [15766] = {.lex_state = 203, .external_lex_state = 126},
  [15767] = {.lex_state = 203, .external_lex_state = 126},
  [15768] = {.lex_state = 203, .external_lex_state = 125},
  [15769] = {.lex_state = 203, .external_lex_state = 126},
  [15770] = {.lex_state = 224},
  [15771] = {.lex_state = 224},
  [15772] = {.lex_state = 203, .external_lex_state = 125},
  [15773] = {.lex_state = 203, .external_lex_state = 125},
  [15774] = {.lex_state = 203, .external_lex_state = 126},
  [15775] = {.lex_state = 203, .external_lex_state = 126},
  [15776] = {.lex_state = 203, .external_lex_state = 125},
  [15777] = {.lex_state = 203, .external_lex_state = 126},
  [15778] = {.lex_state = 224},
  [15779] = {.lex_state = 224},
  [15780] = {.lex_state = 179},
  [15781] = {.lex_state = 203, .external_lex_state = 126},
  [15782] = {.lex_state = 203, .external_lex_state = 125},
  [15783] = {.lex_state = 203, .external_lex_state = 125},
  [15784] = {.lex_state = 203, .external_lex_state = 126},
  [15785] = {.lex_state = 203, .external_lex_state = 126},
  [15786] = {.lex_state = 203, .external_lex_state = 125},
  [15787] = {.lex_state = 203, .external_lex_state = 125},
  [15788] = {.lex_state = 203, .external_lex_state = 126},
  [15789] = {.lex_state = 224},
  [15790] = {.lex_state = 224},
  [15791] = {.lex_state = 203, .external_lex_state = 125},
  [15792] = {.lex_state = 203, .external_lex_state = 125},
  [15793] = {.lex_state = 203, .external_lex_state = 125},
  [15794] = {.lex_state = 203, .external_lex_state = 125},
  [15795] = {.lex_state = 222, .external_lex_state = 127},
  [15796] = {.lex_state = 224},
  [15797] = {.lex_state = 203, .external_lex_state = 126},
  [15798] = {.lex_state = 203, .external_lex_state = 126},
  [15799] = {.lex_state = 203, .external_lex_state = 125},
  [15800] = {.lex_state = 203, .external_lex_state = 126},
  [15801] = {.lex_state = 224},
  [15802] = {.lex_state = 224},
  [15803] = {.lex_state = 203, .external_lex_state = 126},
  [15804] = {.lex_state = 203, .external_lex_state = 126},
  [15805] = {.lex_state = 203, .external_lex_state = 125},
  [15806] = {.lex_state = 203, .external_lex_state = 125},
  [15807] = {.lex_state = 203, .external_lex_state = 125},
  [15808] = {.lex_state = 203, .external_lex_state = 126},
  [15809] = {.lex_state = 203, .external_lex_state = 126},
  [15810] = {.lex_state = 203, .external_lex_state = 125},
  [15811] = {.lex_state = 203, .external_lex_state = 126},
  [15812] = {.lex_state = 224},
  [15813] = {.lex_state = 224},
  [15814] = {.lex_state = 179},
  [15815] = {.lex_state = 203, .external_lex_state = 125},
  [15816] = {.lex_state = 203, .external_lex_state = 125},
  [15817] = {.lex_state = 455},
  [15818] = {.lex_state = 203, .external_lex_state = 126},
  [15819] = {.lex_state = 203, .external_lex_state = 126},
  [15820] = {.lex_state = 203, .external_lex_state = 125},
  [15821] = {.lex_state = 203, .external_lex_state = 126},
  [15822] = {.lex_state = 224},
  [15823] = {.lex_state = 224},
  [15824] = {.lex_state = 203, .external_lex_state = 126},
  [15825] = {.lex_state = 203, .external_lex_state = 125},
  [15826] = {.lex_state = 203, .external_lex_state = 125},
  [15827] = {.lex_state = 224},
  [15828] = {.lex_state = 455},
  [15829] = {.lex_state = 203, .external_lex_state = 126},
  [15830] = {.lex_state = 203, .external_lex_state = 126},
  [15831] = {.lex_state = 203, .external_lex_state = 125},
  [15832] = {.lex_state = 203, .external_lex_state = 126},
  [15833] = {.lex_state = 224},
  [15834] = {.lex_state = 224},
  [15835] = {.lex_state = 203, .external_lex_state = 126},
  [15836] = {.lex_state = 203, .external_lex_state = 125},
  [15837] = {.lex_state = 203, .external_lex_state = 125},
  [15838] = {.lex_state = 203, .external_lex_state = 126},
  [15839] = {.lex_state = 203, .external_lex_state = 126},
  [15840] = {.lex_state = 203, .external_lex_state = 125},
  [15841] = {.lex_state = 455},
  [15842] = {.lex_state = 224},
  [15843] = {.lex_state = 203, .external_lex_state = 126},
  [15844] = {.lex_state = 224},
  [15845] = {.lex_state = 224},
  [15846] = {.lex_state = 224},
  [15847] = {.lex_state = 203, .external_lex_state = 125},
  [15848] = {.lex_state = 203, .external_lex_state = 125},
  [15849] = {.lex_state = 224},
  [15850] = {.lex_state = 203, .external_lex_state = 126},
  [15851] = {.lex_state = 203, .external_lex_state = 126},
  [15852] = {.lex_state = 203, .external_lex_state = 125},
  [15853] = {.lex_state = 138},
  [15854] = {.lex_state = 203, .external_lex_state = 126},
  [15855] = {.lex_state = 224},
  [15856] = {.lex_state = 224},
  [15857] = {.lex_state = 203},
  [15858] = {.lex_state = 203, .external_lex_state = 125},
  [15859] = {.lex_state = 203, .external_lex_state = 125},
  [15860] = {.lex_state = 224, .external_lex_state = 108},
  [15861] = {.lex_state = 455},
  [15862] = {.lex_state = 203, .external_lex_state = 126},
  [15863] = {.lex_state = 203, .external_lex_state = 126},
  [15864] = {.lex_state = 203, .external_lex_state = 125},
  [15865] = {.lex_state = 203, .external_lex_state = 126},
  [15866] = {.lex_state = 224},
  [15867] = {.lex_state = 224},
  [15868] = {.lex_state = 224},
  [15869] = {.lex_state = 224},
  [15870] = {.lex_state = 203},
  [15871] = {.lex_state = 224},
  [15872] = {.lex_state = 224},
  [15873] = {.lex_state = 203, .external_lex_state = 125},
  [15874] = {.lex_state = 224},
  [15875] = {.lex_state = 224},
  [15876] = {.lex_state = 224},
  [15877] = {.lex_state = 224},
  [15878] = {.lex_state = 455},
  [15879] = {.lex_state = 224},
  [15880] = {.lex_state = 224},
  [15881] = {.lex_state = 224},
  [15882] = {.lex_state = 224},
  [15883] = {.lex_state = 224},
  [15884] = {.lex_state = 224},
  [15885] = {.lex_state = 224},
  [15886] = {.lex_state = 224},
  [15887] = {.lex_state = 224},
  [15888] = {.lex_state = 224},
  [15889] = {.lex_state = 179},
  [15890] = {.lex_state = 224},
  [15891] = {.lex_state = 224},
  [15892] = {.lex_state = 224},
  [15893] = {.lex_state = 224},
  [15894] = {.lex_state = 203, .external_lex_state = 125},
  [15895] = {.lex_state = 203, .external_lex_state = 125},
  [15896] = {.lex_state = 224},
  [15897] = {.lex_state = 224},
  [15898] = {.lex_state = 224},
  [15899] = {.lex_state = 224},
  [15900] = {.lex_state = 224},
  [15901] = {.lex_state = 224},
  [15902] = {.lex_state = 179},
  [15903] = {.lex_state = 224},
  [15904] = {.lex_state = 224},
  [15905] = {.lex_state = 138},
  [15906] = {.lex_state = 203},
  [15907] = {.lex_state = 224},
  [15908] = {.lex_state = 224},
  [15909] = {.lex_state = 203, .external_lex_state = 126},
  [15910] = {.lex_state = 203, .external_lex_state = 126},
  [15911] = {.lex_state = 224},
  [15912] = {.lex_state = 224},
  [15913] = {.lex_state = 203, .external_lex_state = 125},
  [15914] = {.lex_state = 224},
  [15915] = {.lex_state = 224},
  [15916] = {.lex_state = 224},
  [15917] = {.lex_state = 224},
  [15918] = {.lex_state = 224},
  [15919] = {.lex_state = 224},
  [15920] = {.lex_state = 203, .external_lex_state = 126},
  [15921] = {.lex_state = 224},
  [15922] = {.lex_state = 224},
  [15923] = {.lex_state = 224},
  [15924] = {.lex_state = 224},
  [15925] = {.lex_state = 224},
  [15926] = {.lex_state = 224},
  [15927] = {.lex_state = 224},
  [15928] = {.lex_state = 455},
  [15929] = {.lex_state = 224},
  [15930] = {.lex_state = 224},
  [15931] = {.lex_state = 455},
  [15932] = {.lex_state = 224},
  [15933] = {.lex_state = 224},
  [15934] = {.lex_state = 224},
  [15935] = {.lex_state = 224},
  [15936] = {.lex_state = 224},
  [15937] = {.lex_state = 203},
  [15938] = {.lex_state = 224},
  [15939] = {.lex_state = 224},
  [15940] = {.lex_state = 224},
  [15941] = {.lex_state = 224},
  [15942] = {.lex_state = 224},
  [15943] = {.lex_state = 224},
  [15944] = {.lex_state = 455},
  [15945] = {.lex_state = 179},
  [15946] = {.lex_state = 224},
  [15947] = {.lex_state = 224},
  [15948] = {.lex_state = 224},
  [15949] = {.lex_state = 224},
  [15950] = {.lex_state = 224},
  [15951] = {.lex_state = 224},
  [15952] = {.lex_state = 224},
  [15953] = {.lex_state = 224},
  [15954] = {.lex_state = 179},
  [15955] = {.lex_state = 224},
  [15956] = {.lex_state = 224},
  [15957] = {.lex_state = 224},
  [15958] = {.lex_state = 224},
  [15959] = {.lex_state = 1},
  [15960] = {.lex_state = 224},
  [15961] = {.lex_state = 224},
  [15962] = {.lex_state = 224},
  [15963] = {.lex_state = 224},
  [15964] = {.lex_state = 224},
  [15965] = {.lex_state = 224},
  [15966] = {.lex_state = 179},
  [15967] = {.lex_state = 224},
  [15968] = {.lex_state = 224},
  [15969] = {.lex_state = 203, .external_lex_state = 125},
  [15970] = {.lex_state = 138},
  [15971] = {.lex_state = 203, .external_lex_state = 125},
  [15972] = {.lex_state = 224},
  [15973] = {.lex_state = 224},
  [15974] = {.lex_state = 224},
  [15975] = {.lex_state = 224},
  [15976] = {.lex_state = 203, .external_lex_state = 125},
  [15977] = {.lex_state = 203, .external_lex_state = 125},
  [15978] = {.lex_state = 224},
  [15979] = {.lex_state = 224},
  [15980] = {.lex_state = 224},
  [15981] = {.lex_state = 224},
  [15982] = {.lex_state = 203, .external_lex_state = 126},
  [15983] = {.lex_state = 203, .external_lex_state = 126},
  [15984] = {.lex_state = 224},
  [15985] = {.lex_state = 224},
  [15986] = {.lex_state = 203, .external_lex_state = 125},
  [15987] = {.lex_state = 224},
  [15988] = {.lex_state = 224},
  [15989] = {.lex_state = 203, .external_lex_state = 126},
  [15990] = {.lex_state = 203, .external_lex_state = 126},
  [15991] = {.lex_state = 224},
  [15992] = {.lex_state = 224},
  [15993] = {.lex_state = 224},
  [15994] = {.lex_state = 203, .external_lex_state = 126},
  [15995] = {.lex_state = 224},
  [15996] = {.lex_state = 224},
  [15997] = {.lex_state = 224},
  [15998] = {.lex_state = 203, .external_lex_state = 125},
  [15999] = {.lex_state = 224},
  [16000] = {.lex_state = 224},
  [16001] = {.lex_state = 224},
  [16002] = {.lex_state = 224},
  [16003] = {.lex_state = 455},
  [16004] = {.lex_state = 224},
  [16005] = {.lex_state = 224},
  [16006] = {.lex_state = 224},
  [16007] = {.lex_state = 224},
  [16008] = {.lex_state = 224},
  [16009] = {.lex_state = 224},
  [16010] = {.lex_state = 224},
  [16011] = {.lex_state = 224},
  [16012] = {.lex_state = 224},
  [16013] = {.lex_state = 224},
  [16014] = {.lex_state = 224},
  [16015] = {.lex_state = 224},
  [16016] = {.lex_state = 224},
  [16017] = {.lex_state = 203},
  [16018] = {.lex_state = 455},
  [16019] = {.lex_state = 203, .external_lex_state = 126},
  [16020] = {.lex_state = 179},
  [16021] = {.lex_state = 222, .external_lex_state = 127},
  [16022] = {.lex_state = 224},
  [16023] = {.lex_state = 455},
  [16024] = {.lex_state = 203, .external_lex_state = 125},
  [16025] = {.lex_state = 203, .external_lex_state = 125},
  [16026] = {.lex_state = 203, .external_lex_state = 126},
  [16027] = {.lex_state = 203, .external_lex_state = 126},
  [16028] = {.lex_state = 203, .external_lex_state = 125},
  [16029] = {.lex_state = 203, .external_lex_state = 125},
  [16030] = {.lex_state = 224, .external_lex_state = 102},
  [16031] = {.lex_state = 224, .external_lex_state = 102},
  [16032] = {.lex_state = 203, .external_lex_state = 126},
  [16033] = {.lex_state = 224},
  [16034] = {.lex_state = 455},
  [16035] = {.lex_state = 224},
  [16036] = {.lex_state = 224},
  [16037] = {.lex_state = 455},
  [16038] = {.lex_state = 179},
  [16039] = {.lex_state = 203, .external_lex_state = 125},
  [16040] = {.lex_state = 203, .external_lex_state = 125},
  [16041] = {.lex_state = 179},
  [16042] = {.lex_state = 203},
  [16043] = {.lex_state = 203, .external_lex_state = 126},
  [16044] = {.lex_state = 203, .external_lex_state = 126},
  [16045] = {.lex_state = 455},
  [16046] = {.lex_state = 203, .external_lex_state = 125},
  [16047] = {.lex_state = 1},
  [16048] = {.lex_state = 203, .external_lex_state = 126},
  [16049] = {.lex_state = 224},
  [16050] = {.lex_state = 224},
  [16051] = {.lex_state = 179},
  [16052] = {.lex_state = 203, .external_lex_state = 125},
  [16053] = {.lex_state = 203, .external_lex_state = 125},
  [16054] = {.lex_state = 455},
  [16055] = {.lex_state = 203},
  [16056] = {.lex_state = 203, .external_lex_state = 126},
  [16057] = {.lex_state = 203, .external_lex_state = 126},
  [16058] = {.lex_state = 203, .external_lex_state = 125},
  [16059] = {.lex_state = 203, .external_lex_state = 125},
  [16060] = {.lex_state = 203, .external_lex_state = 126},
  [16061] = {.lex_state = 224},
  [16062] = {.lex_state = 203, .external_lex_state = 125},
  [16063] = {.lex_state = 224},
  [16064] = {.lex_state = 179},
  [16065] = {.lex_state = 179},
  [16066] = {.lex_state = 179},
  [16067] = {.lex_state = 203, .external_lex_state = 125},
  [16068] = {.lex_state = 203, .external_lex_state = 125},
  [16069] = {.lex_state = 224},
  [16070] = {.lex_state = 203, .external_lex_state = 126},
  [16071] = {.lex_state = 179},
  [16072] = {.lex_state = 203, .external_lex_state = 126},
  [16073] = {.lex_state = 203, .external_lex_state = 126},
  [16074] = {.lex_state = 224, .external_lex_state = 114},
  [16075] = {.lex_state = 203, .external_lex_state = 125},
  [16076] = {.lex_state = 203, .external_lex_state = 126},
  [16077] = {.lex_state = 179},
  [16078] = {.lex_state = 203, .external_lex_state = 126},
  [16079] = {.lex_state = 1},
  [16080] = {.lex_state = 203, .external_lex_state = 126},
  [16081] = {.lex_state = 224, .external_lex_state = 114},
  [16082] = {.lex_state = 224},
  [16083] = {.lex_state = 224},
  [16084] = {.lex_state = 1},
  [16085] = {.lex_state = 203, .external_lex_state = 125},
  [16086] = {.lex_state = 203, .external_lex_state = 125},
  [16087] = {.lex_state = 203, .external_lex_state = 125},
  [16088] = {.lex_state = 179},
  [16089] = {.lex_state = 224, .external_lex_state = 108},
  [16090] = {.lex_state = 203, .external_lex_state = 125},
  [16091] = {.lex_state = 203, .external_lex_state = 125},
  [16092] = {.lex_state = 203, .external_lex_state = 126},
  [16093] = {.lex_state = 203, .external_lex_state = 126},
  [16094] = {.lex_state = 224, .external_lex_state = 108},
  [16095] = {.lex_state = 203, .external_lex_state = 125},
  [16096] = {.lex_state = 203, .external_lex_state = 126},
  [16097] = {.lex_state = 203, .external_lex_state = 126},
  [16098] = {.lex_state = 203, .external_lex_state = 125},
  [16099] = {.lex_state = 203, .external_lex_state = 126},
  [16100] = {.lex_state = 224},
  [16101] = {.lex_state = 224},
  [16102] = {.lex_state = 179},
  [16103] = {.lex_state = 203, .external_lex_state = 126},
  [16104] = {.lex_state = 203, .external_lex_state = 125},
  [16105] = {.lex_state = 203, .external_lex_state = 125},
  [16106] = {.lex_state = 203, .external_lex_state = 126},
  [16107] = {.lex_state = 222, .external_lex_state = 127},
  [16108] = {.lex_state = 224},
  [16109] = {.lex_state = 203, .external_lex_state = 126},
  [16110] = {.lex_state = 203, .external_lex_state = 126},
  [16111] = {.lex_state = 203, .external_lex_state = 125},
  [16112] = {.lex_state = 179},
  [16113] = {.lex_state = 203, .external_lex_state = 126},
  [16114] = {.lex_state = 224},
  [16115] = {.lex_state = 455},
  [16116] = {.lex_state = 455},
  [16117] = {.lex_state = 179},
  [16118] = {.lex_state = 179},
  [16119] = {.lex_state = 224},
  [16120] = {.lex_state = 222, .external_lex_state = 127},
  [16121] = {.lex_state = 224},
  [16122] = {.lex_state = 455},
  [16123] = {.lex_state = 179},
  [16124] = {.lex_state = 179},
  [16125] = {.lex_state = 141, .external_lex_state = 128},
  [16126] = {.lex_state = 224},
  [16127] = {.lex_state = 203, .external_lex_state = 125},
  [16128] = {.lex_state = 203, .external_lex_state = 125},
  [16129] = {.lex_state = 224, .external_lex_state = 97},
  [16130] = {.lex_state = 203, .external_lex_state = 125},
  [16131] = {.lex_state = 203, .external_lex_state = 125},
  [16132] = {.lex_state = 203, .external_lex_state = 126},
  [16133] = {.lex_state = 203, .external_lex_state = 126},
  [16134] = {.lex_state = 203, .external_lex_state = 125},
  [16135] = {.lex_state = 224, .external_lex_state = 97},
  [16136] = {.lex_state = 224, .external_lex_state = 97},
  [16137] = {.lex_state = 224, .external_lex_state = 97},
  [16138] = {.lex_state = 179},
  [16139] = {.lex_state = 224, .external_lex_state = 97},
  [16140] = {.lex_state = 224, .external_lex_state = 97},
  [16141] = {.lex_state = 224, .external_lex_state = 97},
  [16142] = {.lex_state = 203, .external_lex_state = 126},
  [16143] = {.lex_state = 224},
  [16144] = {.lex_state = 224, .external_lex_state = 97},
  [16145] = {.lex_state = 224, .external_lex_state = 97},
  [16146] = {.lex_state = 224, .external_lex_state = 97},
  [16147] = {.lex_state = 224, .external_lex_state = 97},
  [16148] = {.lex_state = 224, .external_lex_state = 97},
  [16149] = {.lex_state = 224, .external_lex_state = 97},
  [16150] = {.lex_state = 224, .external_lex_state = 97},
  [16151] = {.lex_state = 224, .external_lex_state = 97},
  [16152] = {.lex_state = 224, .external_lex_state = 97},
  [16153] = {.lex_state = 224, .external_lex_state = 97},
  [16154] = {.lex_state = 224, .external_lex_state = 97},
  [16155] = {.lex_state = 224, .external_lex_state = 97},
  [16156] = {.lex_state = 224, .external_lex_state = 97},
  [16157] = {.lex_state = 224, .external_lex_state = 97},
  [16158] = {.lex_state = 224, .external_lex_state = 97},
  [16159] = {.lex_state = 224, .external_lex_state = 97},
  [16160] = {.lex_state = 224, .external_lex_state = 97},
  [16161] = {.lex_state = 224, .external_lex_state = 97},
  [16162] = {.lex_state = 224},
  [16163] = {.lex_state = 203},
  [16164] = {.lex_state = 179},
  [16165] = {.lex_state = 224, .external_lex_state = 97},
  [16166] = {.lex_state = 224, .external_lex_state = 97},
  [16167] = {.lex_state = 224, .external_lex_state = 97},
  [16168] = {.lex_state = 224, .external_lex_state = 97},
  [16169] = {.lex_state = 179},
  [16170] = {.lex_state = 224, .external_lex_state = 97},
  [16171] = {.lex_state = 455},
  [16172] = {.lex_state = 203, .external_lex_state = 125},
  [16173] = {.lex_state = 203, .external_lex_state = 125},
  [16174] = {.lex_state = 203, .external_lex_state = 125},
  [16175] = {.lex_state = 222, .external_lex_state = 127},
  [16176] = {.lex_state = 138},
  [16177] = {.lex_state = 203, .external_lex_state = 126},
  [16178] = {.lex_state = 203, .external_lex_state = 126},
  [16179] = {.lex_state = 203, .external_lex_state = 125},
  [16180] = {.lex_state = 224},
  [16181] = {.lex_state = 224, .external_lex_state = 102},
  [16182] = {.lex_state = 224, .external_lex_state = 102},
  [16183] = {.lex_state = 203, .external_lex_state = 126},
  [16184] = {.lex_state = 224, .external_lex_state = 102},
  [16185] = {.lex_state = 224, .external_lex_state = 102},
  [16186] = {.lex_state = 224},
  [16187] = {.lex_state = 224, .external_lex_state = 102},
  [16188] = {.lex_state = 224, .external_lex_state = 102},
  [16189] = {.lex_state = 224, .external_lex_state = 102},
  [16190] = {.lex_state = 224},
  [16191] = {.lex_state = 224, .external_lex_state = 102},
  [16192] = {.lex_state = 224, .external_lex_state = 102},
  [16193] = {.lex_state = 224, .external_lex_state = 102},
  [16194] = {.lex_state = 224, .external_lex_state = 102},
  [16195] = {.lex_state = 224, .external_lex_state = 102},
  [16196] = {.lex_state = 224, .external_lex_state = 102},
  [16197] = {.lex_state = 224, .external_lex_state = 102},
  [16198] = {.lex_state = 224, .external_lex_state = 102},
  [16199] = {.lex_state = 224, .external_lex_state = 102},
  [16200] = {.lex_state = 224, .external_lex_state = 102},
  [16201] = {.lex_state = 455},
  [16202] = {.lex_state = 224, .external_lex_state = 102},
  [16203] = {.lex_state = 224, .external_lex_state = 102},
  [16204] = {.lex_state = 224, .external_lex_state = 102},
  [16205] = {.lex_state = 224, .external_lex_state = 102},
  [16206] = {.lex_state = 224, .external_lex_state = 102},
  [16207] = {.lex_state = 179},
  [16208] = {.lex_state = 224, .external_lex_state = 102},
  [16209] = {.lex_state = 222, .external_lex_state = 127},
  [16210] = {.lex_state = 224, .external_lex_state = 102},
  [16211] = {.lex_state = 203, .external_lex_state = 125},
  [16212] = {.lex_state = 224, .external_lex_state = 102},
  [16213] = {.lex_state = 203, .external_lex_state = 125},
  [16214] = {.lex_state = 224, .external_lex_state = 102},
  [16215] = {.lex_state = 224, .external_lex_state = 102},
  [16216] = {.lex_state = 203, .external_lex_state = 126},
  [16217] = {.lex_state = 203, .external_lex_state = 126},
  [16218] = {.lex_state = 224, .external_lex_state = 102},
  [16219] = {.lex_state = 203, .external_lex_state = 125},
  [16220] = {.lex_state = 224},
  [16221] = {.lex_state = 455},
  [16222] = {.lex_state = 203, .external_lex_state = 126},
  [16223] = {.lex_state = 224},
  [16224] = {.lex_state = 224},
  [16225] = {.lex_state = 203},
  [16226] = {.lex_state = 224},
  [16227] = {.lex_state = 179},
  [16228] = {.lex_state = 224, .external_lex_state = 97},
  [16229] = {.lex_state = 224, .external_lex_state = 97},
  [16230] = {.lex_state = 203, .external_lex_state = 126},
  [16231] = {.lex_state = 203, .external_lex_state = 126},
  [16232] = {.lex_state = 203, .external_lex_state = 125},
  [16233] = {.lex_state = 203, .external_lex_state = 125},
  [16234] = {.lex_state = 457},
  [16235] = {.lex_state = 203},
  [16236] = {.lex_state = 457},
  [16237] = {.lex_state = 203, .external_lex_state = 126},
  [16238] = {.lex_state = 203, .external_lex_state = 126},
  [16239] = {.lex_state = 179},
  [16240] = {.lex_state = 203, .external_lex_state = 125},
  [16241] = {.lex_state = 1},
  [16242] = {.lex_state = 203, .external_lex_state = 126},
  [16243] = {.lex_state = 224},
  [16244] = {.lex_state = 224, .external_lex_state = 97},
  [16245] = {.lex_state = 224, .external_lex_state = 97},
  [16246] = {.lex_state = 224},
  [16247] = {.lex_state = 203, .external_lex_state = 125},
  [16248] = {.lex_state = 203, .external_lex_state = 125},
  [16249] = {.lex_state = 179},
  [16250] = {.lex_state = 224, .external_lex_state = 108},
  [16251] = {.lex_state = 455},
  [16252] = {.lex_state = 203, .external_lex_state = 125},
  [16253] = {.lex_state = 203, .external_lex_state = 125},
  [16254] = {.lex_state = 203, .external_lex_state = 126},
  [16255] = {.lex_state = 203, .external_lex_state = 126},
  [16256] = {.lex_state = 203, .external_lex_state = 126},
  [16257] = {.lex_state = 203, .external_lex_state = 125},
  [16258] = {.lex_state = 203, .external_lex_state = 126},
  [16259] = {.lex_state = 203, .external_lex_state = 126},
  [16260] = {.lex_state = 224},
  [16261] = {.lex_state = 203, .external_lex_state = 125},
  [16262] = {.lex_state = 1},
  [16263] = {.lex_state = 224, .external_lex_state = 108},
  [16264] = {.lex_state = 224},
  [16265] = {.lex_state = 224, .external_lex_state = 102},
  [16266] = {.lex_state = 224, .external_lex_state = 102},
  [16267] = {.lex_state = 203, .external_lex_state = 125},
  [16268] = {.lex_state = 179},
  [16269] = {.lex_state = 138},
  [16270] = {.lex_state = 224, .external_lex_state = 102},
  [16271] = {.lex_state = 224, .external_lex_state = 102},
  [16272] = {.lex_state = 203, .external_lex_state = 125},
  [16273] = {.lex_state = 203, .external_lex_state = 125},
  [16274] = {.lex_state = 455},
  [16275] = {.lex_state = 455},
  [16276] = {.lex_state = 203, .external_lex_state = 126},
  [16277] = {.lex_state = 203, .external_lex_state = 126},
  [16278] = {.lex_state = 203, .external_lex_state = 126},
  [16279] = {.lex_state = 224},
  [16280] = {.lex_state = 203, .external_lex_state = 125},
  [16281] = {.lex_state = 138},
  [16282] = {.lex_state = 203, .external_lex_state = 126},
  [16283] = {.lex_state = 224},
  [16284] = {.lex_state = 1},
  [16285] = {.lex_state = 224},
  [16286] = {.lex_state = 138},
  [16287] = {.lex_state = 138},
  [16288] = {.lex_state = 203, .external_lex_state = 125},
  [16289] = {.lex_state = 203, .external_lex_state = 125},
  [16290] = {.lex_state = 203, .external_lex_state = 125},
  [16291] = {.lex_state = 203, .external_lex_state = 125},
  [16292] = {.lex_state = 455},
  [16293] = {.lex_state = 203, .external_lex_state = 126},
  [16294] = {.lex_state = 203, .external_lex_state = 126},
  [16295] = {.lex_state = 138},
  [16296] = {.lex_state = 203, .external_lex_state = 125},
  [16297] = {.lex_state = 203, .external_lex_state = 126},
  [16298] = {.lex_state = 203, .external_lex_state = 126},
  [16299] = {.lex_state = 224},
  [16300] = {.lex_state = 138},
  [16301] = {.lex_state = 203, .external_lex_state = 126},
  [16302] = {.lex_state = 203, .external_lex_state = 125},
  [16303] = {.lex_state = 224},
  [16304] = {.lex_state = 179},
  [16305] = {.lex_state = 138},
  [16306] = {.lex_state = 224},
  [16307] = {.lex_state = 138},
  [16308] = {.lex_state = 203},
  [16309] = {.lex_state = 455},
  [16310] = {.lex_state = 203, .external_lex_state = 125},
  [16311] = {.lex_state = 203, .external_lex_state = 126},
  [16312] = {.lex_state = 203, .external_lex_state = 125},
  [16313] = {.lex_state = 138},
  [16314] = {.lex_state = 455},
  [16315] = {.lex_state = 222, .external_lex_state = 127},
  [16316] = {.lex_state = 224},
  [16317] = {.lex_state = 203, .external_lex_state = 126},
  [16318] = {.lex_state = 138},
  [16319] = {.lex_state = 203, .external_lex_state = 126},
  [16320] = {.lex_state = 203, .external_lex_state = 125},
  [16321] = {.lex_state = 138},
  [16322] = {.lex_state = 203, .external_lex_state = 126},
  [16323] = {.lex_state = 455},
  [16324] = {.lex_state = 203, .external_lex_state = 126},
  [16325] = {.lex_state = 138},
  [16326] = {.lex_state = 224},
  [16327] = {.lex_state = 224},
  [16328] = {.lex_state = 138},
  [16329] = {.lex_state = 455},
  [16330] = {.lex_state = 203, .external_lex_state = 125},
  [16331] = {.lex_state = 224},
  [16332] = {.lex_state = 138},
  [16333] = {.lex_state = 224},
  [16334] = {.lex_state = 455},
  [16335] = {.lex_state = 203, .external_lex_state = 125},
  [16336] = {.lex_state = 203, .external_lex_state = 125},
  [16337] = {.lex_state = 203},
  [16338] = {.lex_state = 203, .external_lex_state = 126},
  [16339] = {.lex_state = 203, .external_lex_state = 126},
  [16340] = {.lex_state = 455},
  [16341] = {.lex_state = 203, .external_lex_state = 125},
  [16342] = {.lex_state = 203, .external_lex_state = 126},
  [16343] = {.lex_state = 224},
  [16344] = {.lex_state = 224},
  [16345] = {.lex_state = 203, .external_lex_state = 125},
  [16346] = {.lex_state = 203, .external_lex_state = 125},
  [16347] = {.lex_state = 203, .external_lex_state = 126},
  [16348] = {.lex_state = 203, .external_lex_state = 126},
  [16349] = {.lex_state = 203, .external_lex_state = 125},
  [16350] = {.lex_state = 203, .external_lex_state = 126},
  [16351] = {.lex_state = 224},
  [16352] = {.lex_state = 224},
  [16353] = {.lex_state = 203, .external_lex_state = 125},
  [16354] = {.lex_state = 203, .external_lex_state = 125},
  [16355] = {.lex_state = 179},
  [16356] = {.lex_state = 203, .external_lex_state = 126},
  [16357] = {.lex_state = 203, .external_lex_state = 126},
  [16358] = {.lex_state = 203, .external_lex_state = 125},
  [16359] = {.lex_state = 455},
  [16360] = {.lex_state = 203, .external_lex_state = 126},
  [16361] = {.lex_state = 224},
  [16362] = {.lex_state = 224},
  [16363] = {.lex_state = 203, .external_lex_state = 125},
  [16364] = {.lex_state = 203, .external_lex_state = 125},
  [16365] = {.lex_state = 203, .external_lex_state = 125},
  [16366] = {.lex_state = 203, .external_lex_state = 125},
  [16367] = {.lex_state = 203, .external_lex_state = 126},
  [16368] = {.lex_state = 203, .external_lex_state = 126},
  [16369] = {.lex_state = 203, .external_lex_state = 125},
  [16370] = {.lex_state = 179},
  [16371] = {.lex_state = 203, .external_lex_state = 126},
  [16372] = {.lex_state = 224},
  [16373] = {.lex_state = 224},
  [16374] = {.lex_state = 1},
  [16375] = {.lex_state = 203, .external_lex_state = 126},
  [16376] = {.lex_state = 203, .external_lex_state = 126},
  [16377] = {.lex_state = 203, .external_lex_state = 125},
  [16378] = {.lex_state = 203, .external_lex_state = 125},
  [16379] = {.lex_state = 203, .external_lex_state = 126},
  [16380] = {.lex_state = 203, .external_lex_state = 125},
  [16381] = {.lex_state = 203, .external_lex_state = 126},
  [16382] = {.lex_state = 203, .external_lex_state = 126},
  [16383] = {.lex_state = 203, .external_lex_state = 125},
  [16384] = {.lex_state = 203, .external_lex_state = 126},
  [16385] = {.lex_state = 224},
  [16386] = {.lex_state = 224},
  [16387] = {.lex_state = 203, .external_lex_state = 125},
  [16388] = {.lex_state = 203, .external_lex_state = 125},
  [16389] = {.lex_state = 203, .external_lex_state = 126},
  [16390] = {.lex_state = 203, .external_lex_state = 126},
  [16391] = {.lex_state = 203, .external_lex_state = 125},
  [16392] = {.lex_state = 203, .external_lex_state = 126},
  [16393] = {.lex_state = 224},
  [16394] = {.lex_state = 224},
  [16395] = {.lex_state = 203, .external_lex_state = 126},
  [16396] = {.lex_state = 203, .external_lex_state = 126},
  [16397] = {.lex_state = 224},
  [16398] = {.lex_state = 203, .external_lex_state = 125},
  [16399] = {.lex_state = 203, .external_lex_state = 125},
  [16400] = {.lex_state = 203, .external_lex_state = 125},
  [16401] = {.lex_state = 203, .external_lex_state = 126},
  [16402] = {.lex_state = 203, .external_lex_state = 126},
  [16403] = {.lex_state = 203, .external_lex_state = 125},
  [16404] = {.lex_state = 203, .external_lex_state = 125},
  [16405] = {.lex_state = 203, .external_lex_state = 126},
  [16406] = {.lex_state = 224},
  [16407] = {.lex_state = 179},
  [16408] = {.lex_state = 224},
  [16409] = {.lex_state = 222, .external_lex_state = 127},
  [16410] = {.lex_state = 203, .external_lex_state = 125},
  [16411] = {.lex_state = 203, .external_lex_state = 125},
  [16412] = {.lex_state = 224},
  [16413] = {.lex_state = 203, .external_lex_state = 126},
  [16414] = {.lex_state = 203, .external_lex_state = 126},
  [16415] = {.lex_state = 203, .external_lex_state = 125},
  [16416] = {.lex_state = 455},
  [16417] = {.lex_state = 203, .external_lex_state = 126},
  [16418] = {.lex_state = 224},
  [16419] = {.lex_state = 224},
  [16420] = {.lex_state = 224},
  [16421] = {.lex_state = 203, .external_lex_state = 125},
  [16422] = {.lex_state = 203, .external_lex_state = 125},
  [16423] = {.lex_state = 203, .external_lex_state = 126},
  [16424] = {.lex_state = 203, .external_lex_state = 126},
  [16425] = {.lex_state = 203, .external_lex_state = 125},
  [16426] = {.lex_state = 179},
  [16427] = {.lex_state = 203, .external_lex_state = 126},
  [16428] = {.lex_state = 224},
  [16429] = {.lex_state = 224},
  [16430] = {.lex_state = 203},
  [16431] = {.lex_state = 203, .external_lex_state = 125},
  [16432] = {.lex_state = 203, .external_lex_state = 125},
  [16433] = {.lex_state = 203, .external_lex_state = 126},
  [16434] = {.lex_state = 203, .external_lex_state = 126},
  [16435] = {.lex_state = 455},
  [16436] = {.lex_state = 203, .external_lex_state = 126},
  [16437] = {.lex_state = 203, .external_lex_state = 126},
  [16438] = {.lex_state = 203, .external_lex_state = 125},
  [16439] = {.lex_state = 1},
  [16440] = {.lex_state = 203, .external_lex_state = 125},
  [16441] = {.lex_state = 203, .external_lex_state = 125},
  [16442] = {.lex_state = 203, .external_lex_state = 126},
  [16443] = {.lex_state = 224},
  [16444] = {.lex_state = 224},
  [16445] = {.lex_state = 203, .external_lex_state = 126},
  [16446] = {.lex_state = 203, .external_lex_state = 126},
  [16447] = {.lex_state = 455},
  [16448] = {.lex_state = 455},
  [16449] = {.lex_state = 224, .external_lex_state = 108},
  [16450] = {.lex_state = 455},
  [16451] = {.lex_state = 224},
  [16452] = {.lex_state = 224, .external_lex_state = 102},
  [16453] = {.lex_state = 455},
  [16454] = {.lex_state = 455},
  [16455] = {.lex_state = 457, .external_lex_state = 99},
  [16456] = {.lex_state = 224},
  [16457] = {.lex_state = 224, .external_lex_state = 108},
  [16458] = {.lex_state = 224, .external_lex_state = 108},
  [16459] = {.lex_state = 224, .external_lex_state = 108},
  [16460] = {.lex_state = 455, .external_lex_state = 114},
  [16461] = {.lex_state = 224, .external_lex_state = 108},
  [16462] = {.lex_state = 224, .external_lex_state = 102},
  [16463] = {.lex_state = 224, .external_lex_state = 108},
  [16464] = {.lex_state = 224, .external_lex_state = 102},
  [16465] = {.lex_state = 224, .external_lex_state = 108},
  [16466] = {.lex_state = 455, .external_lex_state = 114},
  [16467] = {.lex_state = 455},
  [16468] = {.lex_state = 224, .external_lex_state = 109},
  [16469] = {.lex_state = 224, .external_lex_state = 108},
  [16470] = {.lex_state = 224, .external_lex_state = 102},
  [16471] = {.lex_state = 224, .external_lex_state = 108},
  [16472] = {.lex_state = 224},
  [16473] = {.lex_state = 224, .external_lex_state = 108},
  [16474] = {.lex_state = 457, .external_lex_state = 99},
  [16475] = {.lex_state = 224, .external_lex_state = 108},
  [16476] = {.lex_state = 224, .external_lex_state = 108},
  [16477] = {.lex_state = 224, .external_lex_state = 108},
  [16478] = {.lex_state = 224, .external_lex_state = 108},
  [16479] = {.lex_state = 224, .external_lex_state = 108},
  [16480] = {.lex_state = 224, .external_lex_state = 108},
  [16481] = {.lex_state = 224, .external_lex_state = 108},
  [16482] = {.lex_state = 224, .external_lex_state = 108},
  [16483] = {.lex_state = 224, .external_lex_state = 108},
  [16484] = {.lex_state = 224, .external_lex_state = 108},
  [16485] = {.lex_state = 224, .external_lex_state = 108},
  [16486] = {.lex_state = 224, .external_lex_state = 108},
  [16487] = {.lex_state = 224, .external_lex_state = 108},
  [16488] = {.lex_state = 224, .external_lex_state = 108},
  [16489] = {.lex_state = 224, .external_lex_state = 108},
  [16490] = {.lex_state = 224, .external_lex_state = 97},
  [16491] = {.lex_state = 224, .external_lex_state = 102},
  [16492] = {.lex_state = 457, .external_lex_state = 101},
  [16493] = {.lex_state = 455},
  [16494] = {.lex_state = 455},
  [16495] = {.lex_state = 224, .external_lex_state = 108},
  [16496] = {.lex_state = 455},
  [16497] = {.lex_state = 224, .external_lex_state = 97},
  [16498] = {.lex_state = 224, .external_lex_state = 108},
  [16499] = {.lex_state = 455, .external_lex_state = 114},
  [16500] = {.lex_state = 455},
  [16501] = {.lex_state = 224, .external_lex_state = 108},
  [16502] = {.lex_state = 224, .external_lex_state = 108},
  [16503] = {.lex_state = 457, .external_lex_state = 101},
  [16504] = {.lex_state = 224},
  [16505] = {.lex_state = 455},
  [16506] = {.lex_state = 224},
  [16507] = {.lex_state = 455},
  [16508] = {.lex_state = 224, .external_lex_state = 102},
  [16509] = {.lex_state = 455},
  [16510] = {.lex_state = 455},
  [16511] = {.lex_state = 224, .external_lex_state = 108},
  [16512] = {.lex_state = 224, .external_lex_state = 97},
  [16513] = {.lex_state = 455},
  [16514] = {.lex_state = 455},
  [16515] = {.lex_state = 224, .external_lex_state = 97},
  [16516] = {.lex_state = 455},
  [16517] = {.lex_state = 455},
  [16518] = {.lex_state = 455},
  [16519] = {.lex_state = 455},
  [16520] = {.lex_state = 455},
  [16521] = {.lex_state = 455},
  [16522] = {.lex_state = 224},
  [16523] = {.lex_state = 455},
  [16524] = {.lex_state = 224, .external_lex_state = 109},
  [16525] = {.lex_state = 224, .external_lex_state = 102},
  [16526] = {.lex_state = 455},
  [16527] = {.lex_state = 224, .external_lex_state = 102},
  [16528] = {.lex_state = 455},
  [16529] = {.lex_state = 455},
  [16530] = {.lex_state = 224, .external_lex_state = 102},
  [16531] = {.lex_state = 455},
  [16532] = {.lex_state = 455},
  [16533] = {.lex_state = 224, .external_lex_state = 102},
  [16534] = {.lex_state = 224, .external_lex_state = 102},
  [16535] = {.lex_state = 224, .external_lex_state = 97},
  [16536] = {.lex_state = 224, .external_lex_state = 97},
  [16537] = {.lex_state = 224},
  [16538] = {.lex_state = 455},
  [16539] = {.lex_state = 455},
  [16540] = {.lex_state = 457, .external_lex_state = 101},
  [16541] = {.lex_state = 455},
  [16542] = {.lex_state = 455},
  [16543] = {.lex_state = 455},
  [16544] = {.lex_state = 224, .external_lex_state = 97},
  [16545] = {.lex_state = 224, .external_lex_state = 97},
  [16546] = {.lex_state = 224, .external_lex_state = 102},
  [16547] = {.lex_state = 455},
  [16548] = {.lex_state = 224},
  [16549] = {.lex_state = 455, .external_lex_state = 114},
  [16550] = {.lex_state = 455},
  [16551] = {.lex_state = 455},
  [16552] = {.lex_state = 455, .external_lex_state = 114},
  [16553] = {.lex_state = 455},
  [16554] = {.lex_state = 455, .external_lex_state = 114},
  [16555] = {.lex_state = 455},
  [16556] = {.lex_state = 224, .external_lex_state = 109},
  [16557] = {.lex_state = 224},
  [16558] = {.lex_state = 455},
  [16559] = {.lex_state = 455},
  [16560] = {.lex_state = 224},
  [16561] = {.lex_state = 455},
  [16562] = {.lex_state = 455},
  [16563] = {.lex_state = 455},
  [16564] = {.lex_state = 455},
  [16565] = {.lex_state = 455},
  [16566] = {.lex_state = 224, .external_lex_state = 109},
  [16567] = {.lex_state = 224, .external_lex_state = 102},
  [16568] = {.lex_state = 224, .external_lex_state = 102},
  [16569] = {.lex_state = 224, .external_lex_state = 102},
  [16570] = {.lex_state = 455},
  [16571] = {.lex_state = 455},
  [16572] = {.lex_state = 203, .external_lex_state = 123},
  [16573] = {.lex_state = 455},
  [16574] = {.lex_state = 457, .external_lex_state = 101},
  [16575] = {.lex_state = 224, .external_lex_state = 97},
  [16576] = {.lex_state = 455},
  [16577] = {.lex_state = 455},
  [16578] = {.lex_state = 455},
  [16579] = {.lex_state = 455},
  [16580] = {.lex_state = 455},
  [16581] = {.lex_state = 455},
  [16582] = {.lex_state = 224},
  [16583] = {.lex_state = 455, .external_lex_state = 114},
  [16584] = {.lex_state = 455},
  [16585] = {.lex_state = 224},
  [16586] = {.lex_state = 455},
  [16587] = {.lex_state = 455},
  [16588] = {.lex_state = 455, .external_lex_state = 114},
  [16589] = {.lex_state = 455},
  [16590] = {.lex_state = 224, .external_lex_state = 97},
  [16591] = {.lex_state = 457, .external_lex_state = 101},
  [16592] = {.lex_state = 455, .external_lex_state = 114},
  [16593] = {.lex_state = 224, .external_lex_state = 97},
  [16594] = {.lex_state = 224},
  [16595] = {.lex_state = 455},
  [16596] = {.lex_state = 457, .external_lex_state = 99},
  [16597] = {.lex_state = 455},
  [16598] = {.lex_state = 455, .external_lex_state = 114},
  [16599] = {.lex_state = 455},
  [16600] = {.lex_state = 455},
  [16601] = {.lex_state = 455},
  [16602] = {.lex_state = 224},
  [16603] = {.lex_state = 224, .external_lex_state = 97},
  [16604] = {.lex_state = 224, .external_lex_state = 97},
  [16605] = {.lex_state = 455},
  [16606] = {.lex_state = 455},
  [16607] = {.lex_state = 455},
  [16608] = {.lex_state = 457, .external_lex_state = 121},
  [16609] = {.lex_state = 455},
  [16610] = {.lex_state = 224, .external_lex_state = 102},
  [16611] = {.lex_state = 224},
  [16612] = {.lex_state = 457},
  [16613] = {.lex_state = 455},
  [16614] = {.lex_state = 455},
  [16615] = {.lex_state = 455},
  [16616] = {.lex_state = 224, .external_lex_state = 97},
  [16617] = {.lex_state = 455},
  [16618] = {.lex_state = 455, .external_lex_state = 114},
  [16619] = {.lex_state = 455},
  [16620] = {.lex_state = 455},
  [16621] = {.lex_state = 455},
  [16622] = {.lex_state = 455},
  [16623] = {.lex_state = 455},
  [16624] = {.lex_state = 455, .external_lex_state = 114},
  [16625] = {.lex_state = 224, .external_lex_state = 102},
  [16626] = {.lex_state = 224},
  [16627] = {.lex_state = 455},
  [16628] = {.lex_state = 455},
  [16629] = {.lex_state = 224, .external_lex_state = 97},
  [16630] = {.lex_state = 455, .external_lex_state = 114},
  [16631] = {.lex_state = 455},
  [16632] = {.lex_state = 224, .external_lex_state = 114},
  [16633] = {.lex_state = 224},
  [16634] = {.lex_state = 455},
  [16635] = {.lex_state = 455, .external_lex_state = 114},
  [16636] = {.lex_state = 224},
  [16637] = {.lex_state = 455},
  [16638] = {.lex_state = 224, .external_lex_state = 102},
  [16639] = {.lex_state = 455},
  [16640] = {.lex_state = 224, .external_lex_state = 102},
  [16641] = {.lex_state = 224, .external_lex_state = 97},
  [16642] = {.lex_state = 224},
  [16643] = {.lex_state = 455},
  [16644] = {.lex_state = 455},
  [16645] = {.lex_state = 457, .external_lex_state = 101},
  [16646] = {.lex_state = 455},
  [16647] = {.lex_state = 457, .external_lex_state = 101},
  [16648] = {.lex_state = 455},
  [16649] = {.lex_state = 224},
  [16650] = {.lex_state = 455},
  [16651] = {.lex_state = 224, .external_lex_state = 108},
  [16652] = {.lex_state = 455},
  [16653] = {.lex_state = 224, .external_lex_state = 108},
  [16654] = {.lex_state = 224, .external_lex_state = 102},
  [16655] = {.lex_state = 203, .external_lex_state = 123},
  [16656] = {.lex_state = 455, .external_lex_state = 114},
  [16657] = {.lex_state = 455, .external_lex_state = 114},
  [16658] = {.lex_state = 455},
  [16659] = {.lex_state = 455},
  [16660] = {.lex_state = 455},
  [16661] = {.lex_state = 224, .external_lex_state = 102},
  [16662] = {.lex_state = 224, .external_lex_state = 102},
  [16663] = {.lex_state = 224},
  [16664] = {.lex_state = 224, .external_lex_state = 97},
  [16665] = {.lex_state = 455, .external_lex_state = 114},
  [16666] = {.lex_state = 224, .external_lex_state = 97},
  [16667] = {.lex_state = 455, .external_lex_state = 114},
  [16668] = {.lex_state = 203, .external_lex_state = 124},
  [16669] = {.lex_state = 455, .external_lex_state = 114},
  [16670] = {.lex_state = 457, .external_lex_state = 101},
  [16671] = {.lex_state = 224, .external_lex_state = 114},
  [16672] = {.lex_state = 457, .external_lex_state = 101},
  [16673] = {.lex_state = 224},
  [16674] = {.lex_state = 224},
  [16675] = {.lex_state = 455},
  [16676] = {.lex_state = 224, .external_lex_state = 97},
  [16677] = {.lex_state = 457, .external_lex_state = 121},
  [16678] = {.lex_state = 455},
  [16679] = {.lex_state = 224, .external_lex_state = 97},
  [16680] = {.lex_state = 224},
  [16681] = {.lex_state = 455, .external_lex_state = 114},
  [16682] = {.lex_state = 455},
  [16683] = {.lex_state = 455, .external_lex_state = 114},
  [16684] = {.lex_state = 224, .external_lex_state = 102},
  [16685] = {.lex_state = 224},
  [16686] = {.lex_state = 224},
  [16687] = {.lex_state = 224, .external_lex_state = 114},
  [16688] = {.lex_state = 224},
  [16689] = {.lex_state = 455},
  [16690] = {.lex_state = 455},
  [16691] = {.lex_state = 455},
  [16692] = {.lex_state = 224},
  [16693] = {.lex_state = 224},
  [16694] = {.lex_state = 455},
  [16695] = {.lex_state = 455},
  [16696] = {.lex_state = 224},
  [16697] = {.lex_state = 455},
  [16698] = {.lex_state = 455},
  [16699] = {.lex_state = 203, .external_lex_state = 124},
  [16700] = {.lex_state = 457, .external_lex_state = 101},
  [16701] = {.lex_state = 224},
  [16702] = {.lex_state = 457, .external_lex_state = 99},
  [16703] = {.lex_state = 455},
  [16704] = {.lex_state = 224},
  [16705] = {.lex_state = 455},
  [16706] = {.lex_state = 224, .external_lex_state = 102},
  [16707] = {.lex_state = 455},
  [16708] = {.lex_state = 224},
  [16709] = {.lex_state = 224},
  [16710] = {.lex_state = 455},
  [16711] = {.lex_state = 455, .external_lex_state = 114},
  [16712] = {.lex_state = 455},
  [16713] = {.lex_state = 224},
  [16714] = {.lex_state = 457},
  [16715] = {.lex_state = 455},
  [16716] = {.lex_state = 455},
  [16717] = {.lex_state = 455},
  [16718] = {.lex_state = 455},
  [16719] = {.lex_state = 455},
  [16720] = {.lex_state = 457, .external_lex_state = 99},
  [16721] = {.lex_state = 455},
  [16722] = {.lex_state = 455},
  [16723] = {.lex_state = 455},
  [16724] = {.lex_state = 455},
  [16725] = {.lex_state = 455, .external_lex_state = 114},
  [16726] = {.lex_state = 455, .external_lex_state = 114},
  [16727] = {.lex_state = 455},
  [16728] = {.lex_state = 455},
  [16729] = {.lex_state = 455},
  [16730] = {.lex_state = 224},
  [16731] = {.lex_state = 455},
  [16732] = {.lex_state = 455},
  [16733] = {.lex_state = 455},
  [16734] = {.lex_state = 224, .external_lex_state = 102},
  [16735] = {.lex_state = 455},
  [16736] = {.lex_state = 455},
  [16737] = {.lex_state = 224, .external_lex_state = 97},
  [16738] = {.lex_state = 224, .external_lex_state = 102},
  [16739] = {.lex_state = 224, .external_lex_state = 108},
  [16740] = {.lex_state = 224, .external_lex_state = 102},
  [16741] = {.lex_state = 224, .external_lex_state = 97},
  [16742] = {.lex_state = 457, .external_lex_state = 99},
  [16743] = {.lex_state = 224, .external_lex_state = 108},
  [16744] = {.lex_state = 455, .external_lex_state = 114},
  [16745] = {.lex_state = 224, .external_lex_state = 102},
  [16746] = {.lex_state = 455},
  [16747] = {.lex_state = 224, .external_lex_state = 97},
  [16748] = {.lex_state = 457, .external_lex_state = 99},
  [16749] = {.lex_state = 224, .external_lex_state = 97},
  [16750] = {.lex_state = 455, .external_lex_state = 114},
  [16751] = {.lex_state = 455},
  [16752] = {.lex_state = 224, .external_lex_state = 102},
  [16753] = {.lex_state = 457, .external_lex_state = 101},
  [16754] = {.lex_state = 455},
  [16755] = {.lex_state = 455, .external_lex_state = 114},
  [16756] = {.lex_state = 455},
  [16757] = {.lex_state = 455},
  [16758] = {.lex_state = 224},
  [16759] = {.lex_state = 224, .external_lex_state = 102},
  [16760] = {.lex_state = 224, .external_lex_state = 97},
  [16761] = {.lex_state = 455, .external_lex_state = 114},
  [16762] = {.lex_state = 457, .external_lex_state = 101},
  [16763] = {.lex_state = 224, .external_lex_state = 97},
  [16764] = {.lex_state = 455, .external_lex_state = 114},
  [16765] = {.lex_state = 455, .external_lex_state = 114},
  [16766] = {.lex_state = 224, .external_lex_state = 102},
  [16767] = {.lex_state = 455},
  [16768] = {.lex_state = 455},
  [16769] = {.lex_state = 455},
  [16770] = {.lex_state = 457},
  [16771] = {.lex_state = 224, .external_lex_state = 97},
  [16772] = {.lex_state = 224},
  [16773] = {.lex_state = 224},
  [16774] = {.lex_state = 455},
  [16775] = {.lex_state = 455},
  [16776] = {.lex_state = 455},
  [16777] = {.lex_state = 455},
  [16778] = {.lex_state = 224, .external_lex_state = 97},
  [16779] = {.lex_state = 455},
  [16780] = {.lex_state = 457, .external_lex_state = 121},
  [16781] = {.lex_state = 224, .external_lex_state = 97},
  [16782] = {.lex_state = 457, .external_lex_state = 99},
  [16783] = {.lex_state = 224, .external_lex_state = 97},
  [16784] = {.lex_state = 224, .external_lex_state = 109},
  [16785] = {.lex_state = 455},
  [16786] = {.lex_state = 455, .external_lex_state = 114},
  [16787] = {.lex_state = 224, .external_lex_state = 97},
  [16788] = {.lex_state = 455},
  [16789] = {.lex_state = 455},
  [16790] = {.lex_state = 224, .external_lex_state = 129},
  [16791] = {.lex_state = 457, .external_lex_state = 122},
  [16792] = {.lex_state = 455, .external_lex_state = 130},
  [16793] = {.lex_state = 457, .external_lex_state = 128},
  [16794] = {.lex_state = 457, .external_lex_state = 128},
  [16795] = {.lex_state = 138},
  [16796] = {.lex_state = 455},
  [16797] = {.lex_state = 455},
  [16798] = {.lex_state = 457, .external_lex_state = 121},
  [16799] = {.lex_state = 141},
  [16800] = {.lex_state = 203},
  [16801] = {.lex_state = 203},
  [16802] = {.lex_state = 457, .external_lex_state = 99},
  [16803] = {.lex_state = 203},
  [16804] = {.lex_state = 457, .external_lex_state = 128},
  [16805] = {.lex_state = 457},
  [16806] = {.lex_state = 203},
  [16807] = {.lex_state = 457, .external_lex_state = 128},
  [16808] = {.lex_state = 457},
  [16809] = {.lex_state = 457},
  [16810] = {.lex_state = 457, .external_lex_state = 99},
  [16811] = {.lex_state = 457},
  [16812] = {.lex_state = 203},
  [16813] = {.lex_state = 457, .external_lex_state = 99},
  [16814] = {.lex_state = 457},
  [16815] = {.lex_state = 457, .external_lex_state = 131},
  [16816] = {.lex_state = 190, .external_lex_state = 131},
  [16817] = {.lex_state = 141},
  [16818] = {.lex_state = 455},
  [16819] = {.lex_state = 203},
  [16820] = {.lex_state = 141},
  [16821] = {.lex_state = 455},
  [16822] = {.lex_state = 457},
  [16823] = {.lex_state = 455, .external_lex_state = 130},
  [16824] = {.lex_state = 457, .external_lex_state = 128},
  [16825] = {.lex_state = 203},
  [16826] = {.lex_state = 457, .external_lex_state = 128},
  [16827] = {.lex_state = 203, .external_lex_state = 126},
  [16828] = {.lex_state = 457, .external_lex_state = 99},
  [16829] = {.lex_state = 141},
  [16830] = {.lex_state = 455},
  [16831] = {.lex_state = 457},
  [16832] = {.lex_state = 203},
  [16833] = {.lex_state = 203},
  [16834] = {.lex_state = 203},
  [16835] = {.lex_state = 457},
  [16836] = {.lex_state = 457, .external_lex_state = 99},
  [16837] = {.lex_state = 455, .external_lex_state = 99},
  [16838] = {.lex_state = 457, .external_lex_state = 128},
  [16839] = {.lex_state = 455},
  [16840] = {.lex_state = 203},
  [16841] = {.lex_state = 457, .external_lex_state = 128},
  [16842] = {.lex_state = 457},
  [16843] = {.lex_state = 457, .external_lex_state = 99},
  [16844] = {.lex_state = 224, .external_lex_state = 129},
  [16845] = {.lex_state = 457, .external_lex_state = 99},
  [16846] = {.lex_state = 457, .external_lex_state = 99},
  [16847] = {.lex_state = 457},
  [16848] = {.lex_state = 457},
  [16849] = {.lex_state = 141},
  [16850] = {.lex_state = 457, .external_lex_state = 100},
  [16851] = {.lex_state = 457, .external_lex_state = 128},
  [16852] = {.lex_state = 455},
  [16853] = {.lex_state = 224, .external_lex_state = 129},
  [16854] = {.lex_state = 455},
  [16855] = {.lex_state = 203},
  [16856] = {.lex_state = 455, .external_lex_state = 130},
  [16857] = {.lex_state = 203},
  [16858] = {.lex_state = 455},
  [16859] = {.lex_state = 203},
  [16860] = {.lex_state = 457, .external_lex_state = 128},
  [16861] = {.lex_state = 455},
  [16862] = {.lex_state = 457, .external_lex_state = 131},
  [16863] = {.lex_state = 457, .external_lex_state = 99},
  [16864] = {.lex_state = 190, .external_lex_state = 131},
  [16865] = {.lex_state = 457, .external_lex_state = 131},
  [16866] = {.lex_state = 190, .external_lex_state = 131},
  [16867] = {.lex_state = 457},
  [16868] = {.lex_state = 457, .external_lex_state = 131},
  [16869] = {.lex_state = 190, .external_lex_state = 131},
  [16870] = {.lex_state = 141},
  [16871] = {.lex_state = 457, .external_lex_state = 99},
  [16872] = {.lex_state = 457, .external_lex_state = 99},
  [16873] = {.lex_state = 203},
  [16874] = {.lex_state = 203},
  [16875] = {.lex_state = 457, .external_lex_state = 128},
  [16876] = {.lex_state = 141},
  [16877] = {.lex_state = 457, .external_lex_state = 121},
  [16878] = {.lex_state = 224, .external_lex_state = 129},
  [16879] = {.lex_state = 457, .external_lex_state = 99},
  [16880] = {.lex_state = 203},
  [16881] = {.lex_state = 457},
  [16882] = {.lex_state = 457, .external_lex_state = 99},
  [16883] = {.lex_state = 457, .external_lex_state = 121},
  [16884] = {.lex_state = 455, .external_lex_state = 100},
  [16885] = {.lex_state = 457},
  [16886] = {.lex_state = 203},
  [16887] = {.lex_state = 138},
  [16888] = {.lex_state = 224, .external_lex_state = 129},
  [16889] = {.lex_state = 455},
  [16890] = {.lex_state = 457, .external_lex_state = 128},
  [16891] = {.lex_state = 457, .external_lex_state = 128},
  [16892] = {.lex_state = 455},
  [16893] = {.lex_state = 224, .external_lex_state = 108},
  [16894] = {.lex_state = 455, .external_lex_state = 130},
  [16895] = {.lex_state = 455, .external_lex_state = 121},
  [16896] = {.lex_state = 203},
  [16897] = {.lex_state = 457, .external_lex_state = 128},
  [16898] = {.lex_state = 141},
  [16899] = {.lex_state = 457},
  [16900] = {.lex_state = 455},
  [16901] = {.lex_state = 224, .external_lex_state = 129},
  [16902] = {.lex_state = 141},
  [16903] = {.lex_state = 455},
  [16904] = {.lex_state = 203},
  [16905] = {.lex_state = 141},
  [16906] = {.lex_state = 141},
  [16907] = {.lex_state = 457, .external_lex_state = 99},
  [16908] = {.lex_state = 224, .external_lex_state = 129},
  [16909] = {.lex_state = 203},
  [16910] = {.lex_state = 457, .external_lex_state = 128},
  [16911] = {.lex_state = 455, .external_lex_state = 100},
  [16912] = {.lex_state = 455},
  [16913] = {.lex_state = 224, .external_lex_state = 129},
  [16914] = {.lex_state = 203},
  [16915] = {.lex_state = 457, .external_lex_state = 99},
  [16916] = {.lex_state = 141},
  [16917] = {.lex_state = 203},
  [16918] = {.lex_state = 224, .external_lex_state = 108},
  [16919] = {.lex_state = 457, .external_lex_state = 131},
  [16920] = {.lex_state = 190, .external_lex_state = 131},
  [16921] = {.lex_state = 455},
  [16922] = {.lex_state = 457, .external_lex_state = 131},
  [16923] = {.lex_state = 190, .external_lex_state = 131},
  [16924] = {.lex_state = 455},
  [16925] = {.lex_state = 457, .external_lex_state = 131},
  [16926] = {.lex_state = 455, .external_lex_state = 130},
  [16927] = {.lex_state = 190, .external_lex_state = 131},
  [16928] = {.lex_state = 141},
  [16929] = {.lex_state = 224},
  [16930] = {.lex_state = 457, .external_lex_state = 128},
  [16931] = {.lex_state = 457, .external_lex_state = 128},
  [16932] = {.lex_state = 457, .external_lex_state = 128},
  [16933] = {.lex_state = 455},
  [16934] = {.lex_state = 141},
  [16935] = {.lex_state = 141},
  [16936] = {.lex_state = 203},
  [16937] = {.lex_state = 457, .external_lex_state = 99},
  [16938] = {.lex_state = 457, .external_lex_state = 121},
  [16939] = {.lex_state = 457},
  [16940] = {.lex_state = 203},
  [16941] = {.lex_state = 457, .external_lex_state = 128},
  [16942] = {.lex_state = 203},
  [16943] = {.lex_state = 203},
  [16944] = {.lex_state = 457, .external_lex_state = 99},
  [16945] = {.lex_state = 457, .external_lex_state = 121},
  [16946] = {.lex_state = 203},
  [16947] = {.lex_state = 203},
  [16948] = {.lex_state = 141},
  [16949] = {.lex_state = 457},
  [16950] = {.lex_state = 203},
  [16951] = {.lex_state = 457},
  [16952] = {.lex_state = 455},
  [16953] = {.lex_state = 457},
  [16954] = {.lex_state = 224},
  [16955] = {.lex_state = 455},
  [16956] = {.lex_state = 455},
  [16957] = {.lex_state = 455, .external_lex_state = 130},
  [16958] = {.lex_state = 455},
  [16959] = {.lex_state = 457, .external_lex_state = 128},
  [16960] = {.lex_state = 457, .external_lex_state = 131},
  [16961] = {.lex_state = 455},
  [16962] = {.lex_state = 190, .external_lex_state = 131},
  [16963] = {.lex_state = 141},
  [16964] = {.lex_state = 457, .external_lex_state = 99},
  [16965] = {.lex_state = 141},
  [16966] = {.lex_state = 203},
  [16967] = {.lex_state = 203},
  [16968] = {.lex_state = 457, .external_lex_state = 128},
  [16969] = {.lex_state = 455, .external_lex_state = 100},
  [16970] = {.lex_state = 457, .external_lex_state = 99},
  [16971] = {.lex_state = 457},
  [16972] = {.lex_state = 455},
  [16973] = {.lex_state = 457, .external_lex_state = 99},
  [16974] = {.lex_state = 203},
  [16975] = {.lex_state = 203},
  [16976] = {.lex_state = 141},
  [16977] = {.lex_state = 455},
  [16978] = {.lex_state = 203},
  [16979] = {.lex_state = 141},
  [16980] = {.lex_state = 203},
  [16981] = {.lex_state = 455},
  [16982] = {.lex_state = 455},
  [16983] = {.lex_state = 455, .external_lex_state = 130},
  [16984] = {.lex_state = 457, .external_lex_state = 131},
  [16985] = {.lex_state = 190, .external_lex_state = 131},
  [16986] = {.lex_state = 457, .external_lex_state = 131},
  [16987] = {.lex_state = 190, .external_lex_state = 131},
  [16988] = {.lex_state = 457, .external_lex_state = 128},
  [16989] = {.lex_state = 455, .external_lex_state = 100},
  [16990] = {.lex_state = 455},
  [16991] = {.lex_state = 203},
  [16992] = {.lex_state = 141},
  [16993] = {.lex_state = 457, .external_lex_state = 99},
  [16994] = {.lex_state = 457},
  [16995] = {.lex_state = 457},
  [16996] = {.lex_state = 455},
  [16997] = {.lex_state = 203},
  [16998] = {.lex_state = 457, .external_lex_state = 128},
  [16999] = {.lex_state = 141},
  [17000] = {.lex_state = 457, .external_lex_state = 99},
  [17001] = {.lex_state = 224, .external_lex_state = 129},
  [17002] = {.lex_state = 203},
  [17003] = {.lex_state = 457},
  [17004] = {.lex_state = 457},
  [17005] = {.lex_state = 203},
  [17006] = {.lex_state = 455},
  [17007] = {.lex_state = 203},
  [17008] = {.lex_state = 455},
  [17009] = {.lex_state = 457, .external_lex_state = 128},
  [17010] = {.lex_state = 203},
  [17011] = {.lex_state = 455},
  [17012] = {.lex_state = 455, .external_lex_state = 130},
  [17013] = {.lex_state = 457, .external_lex_state = 131},
  [17014] = {.lex_state = 457, .external_lex_state = 100},
  [17015] = {.lex_state = 457, .external_lex_state = 128},
  [17016] = {.lex_state = 141},
  [17017] = {.lex_state = 457},
  [17018] = {.lex_state = 455},
  [17019] = {.lex_state = 457},
  [17020] = {.lex_state = 179},
  [17021] = {.lex_state = 457, .external_lex_state = 99},
  [17022] = {.lex_state = 203},
  [17023] = {.lex_state = 457, .external_lex_state = 128},
  [17024] = {.lex_state = 141},
  [17025] = {.lex_state = 457, .external_lex_state = 99},
  [17026] = {.lex_state = 203},
  [17027] = {.lex_state = 457},
  [17028] = {.lex_state = 224},
  [17029] = {.lex_state = 455},
  [17030] = {.lex_state = 455},
  [17031] = {.lex_state = 224, .external_lex_state = 129},
  [17032] = {.lex_state = 455, .external_lex_state = 130},
  [17033] = {.lex_state = 457, .external_lex_state = 100},
  [17034] = {.lex_state = 190, .external_lex_state = 131},
  [17035] = {.lex_state = 457, .external_lex_state = 128},
  [17036] = {.lex_state = 455},
  [17037] = {.lex_state = 457, .external_lex_state = 99},
  [17038] = {.lex_state = 203},
  [17039] = {.lex_state = 457, .external_lex_state = 128},
  [17040] = {.lex_state = 457, .external_lex_state = 99},
  [17041] = {.lex_state = 457, .external_lex_state = 131},
  [17042] = {.lex_state = 190, .external_lex_state = 131},
  [17043] = {.lex_state = 138},
  [17044] = {.lex_state = 457},
  [17045] = {.lex_state = 455},
  [17046] = {.lex_state = 457},
  [17047] = {.lex_state = 141},
  [17048] = {.lex_state = 203},
  [17049] = {.lex_state = 457},
  [17050] = {.lex_state = 455, .external_lex_state = 130},
  [17051] = {.lex_state = 141},
  [17052] = {.lex_state = 203},
  [17053] = {.lex_state = 457, .external_lex_state = 128},
  [17054] = {.lex_state = 455},
  [17055] = {.lex_state = 457},
  [17056] = {.lex_state = 457, .external_lex_state = 100},
  [17057] = {.lex_state = 203},
  [17058] = {.lex_state = 203},
  [17059] = {.lex_state = 457},
  [17060] = {.lex_state = 457, .external_lex_state = 99},
  [17061] = {.lex_state = 141},
  [17062] = {.lex_state = 203},
  [17063] = {.lex_state = 203},
  [17064] = {.lex_state = 455},
  [17065] = {.lex_state = 457, .external_lex_state = 100},
  [17066] = {.lex_state = 457, .external_lex_state = 100},
  [17067] = {.lex_state = 457, .external_lex_state = 100},
  [17068] = {.lex_state = 455, .external_lex_state = 130},
  [17069] = {.lex_state = 457, .external_lex_state = 128},
  [17070] = {.lex_state = 455},
  [17071] = {.lex_state = 457},
  [17072] = {.lex_state = 203},
  [17073] = {.lex_state = 224, .external_lex_state = 108},
  [17074] = {.lex_state = 457},
  [17075] = {.lex_state = 457, .external_lex_state = 114},
  [17076] = {.lex_state = 141},
  [17077] = {.lex_state = 455, .external_lex_state = 130},
  [17078] = {.lex_state = 203},
  [17079] = {.lex_state = 457, .external_lex_state = 128},
  [17080] = {.lex_state = 455},
  [17081] = {.lex_state = 203},
  [17082] = {.lex_state = 455, .external_lex_state = 130},
  [17083] = {.lex_state = 457, .external_lex_state = 128},
  [17084] = {.lex_state = 455},
  [17085] = {.lex_state = 457},
  [17086] = {.lex_state = 141},
  [17087] = {.lex_state = 203},
  [17088] = {.lex_state = 457, .external_lex_state = 100},
  [17089] = {.lex_state = 203},
  [17090] = {.lex_state = 190, .external_lex_state = 131},
  [17091] = {.lex_state = 224, .external_lex_state = 129},
  [17092] = {.lex_state = 455, .external_lex_state = 130},
  [17093] = {.lex_state = 457, .external_lex_state = 99},
  [17094] = {.lex_state = 224, .external_lex_state = 129},
  [17095] = {.lex_state = 457, .external_lex_state = 128},
  [17096] = {.lex_state = 457, .external_lex_state = 131},
  [17097] = {.lex_state = 455},
  [17098] = {.lex_state = 190, .external_lex_state = 131},
  [17099] = {.lex_state = 457, .external_lex_state = 100},
  [17100] = {.lex_state = 224, .external_lex_state = 108},
  [17101] = {.lex_state = 141},
  [17102] = {.lex_state = 455, .external_lex_state = 130},
  [17103] = {.lex_state = 203},
  [17104] = {.lex_state = 455, .external_lex_state = 130},
  [17105] = {.lex_state = 457},
  [17106] = {.lex_state = 141},
  [17107] = {.lex_state = 457, .external_lex_state = 128},
  [17108] = {.lex_state = 457, .external_lex_state = 131},
  [17109] = {.lex_state = 457},
  [17110] = {.lex_state = 455},
  [17111] = {.lex_state = 203},
  [17112] = {.lex_state = 457, .external_lex_state = 99},
  [17113] = {.lex_state = 141},
  [17114] = {.lex_state = 455, .external_lex_state = 130},
  [17115] = {.lex_state = 141},
  [17116] = {.lex_state = 457, .external_lex_state = 128},
  [17117] = {.lex_state = 457},
  [17118] = {.lex_state = 455},
  [17119] = {.lex_state = 457},
  [17120] = {.lex_state = 457, .external_lex_state = 100},
  [17121] = {.lex_state = 203},
  [17122] = {.lex_state = 203},
  [17123] = {.lex_state = 455, .external_lex_state = 130},
  [17124] = {.lex_state = 457},
  [17125] = {.lex_state = 457, .external_lex_state = 128},
  [17126] = {.lex_state = 224, .external_lex_state = 129},
  [17127] = {.lex_state = 455, .external_lex_state = 121},
  [17128] = {.lex_state = 455},
  [17129] = {.lex_state = 141},
  [17130] = {.lex_state = 455, .external_lex_state = 130},
  [17131] = {.lex_state = 203},
  [17132] = {.lex_state = 457},
  [17133] = {.lex_state = 457, .external_lex_state = 128},
  [17134] = {.lex_state = 455, .external_lex_state = 99},
  [17135] = {.lex_state = 455},
  [17136] = {.lex_state = 457, .external_lex_state = 99},
  [17137] = {.lex_state = 224},
  [17138] = {.lex_state = 457, .external_lex_state = 114},
  [17139] = {.lex_state = 141},
  [17140] = {.lex_state = 457, .external_lex_state = 131},
  [17141] = {.lex_state = 141},
  [17142] = {.lex_state = 203},
  [17143] = {.lex_state = 455, .external_lex_state = 130},
  [17144] = {.lex_state = 190, .external_lex_state = 131},
  [17145] = {.lex_state = 457, .external_lex_state = 128},
  [17146] = {.lex_state = 455},
  [17147] = {.lex_state = 203},
  [17148] = {.lex_state = 141},
  [17149] = {.lex_state = 203},
  [17150] = {.lex_state = 141},
  [17151] = {.lex_state = 455, .external_lex_state = 130},
  [17152] = {.lex_state = 224, .external_lex_state = 108},
  [17153] = {.lex_state = 224, .external_lex_state = 108},
  [17154] = {.lex_state = 457, .external_lex_state = 128},
  [17155] = {.lex_state = 203},
  [17156] = {.lex_state = 455},
  [17157] = {.lex_state = 457, .external_lex_state = 99},
  [17158] = {.lex_state = 457},
  [17159] = {.lex_state = 455, .external_lex_state = 130},
  [17160] = {.lex_state = 457, .external_lex_state = 121},
  [17161] = {.lex_state = 457, .external_lex_state = 131},
  [17162] = {.lex_state = 190, .external_lex_state = 131},
  [17163] = {.lex_state = 457, .external_lex_state = 128},
  [17164] = {.lex_state = 457, .external_lex_state = 121},
  [17165] = {.lex_state = 455},
  [17166] = {.lex_state = 141},
  [17167] = {.lex_state = 203},
  [17168] = {.lex_state = 141},
  [17169] = {.lex_state = 203},
  [17170] = {.lex_state = 455, .external_lex_state = 130},
  [17171] = {.lex_state = 457},
  [17172] = {.lex_state = 457},
  [17173] = {.lex_state = 457, .external_lex_state = 128},
  [17174] = {.lex_state = 141},
  [17175] = {.lex_state = 455},
  [17176] = {.lex_state = 203},
  [17177] = {.lex_state = 457},
  [17178] = {.lex_state = 224},
  [17179] = {.lex_state = 203},
  [17180] = {.lex_state = 457},
  [17181] = {.lex_state = 224},
  [17182] = {.lex_state = 141},
  [17183] = {.lex_state = 455, .external_lex_state = 130},
  [17184] = {.lex_state = 203},
  [17185] = {.lex_state = 457, .external_lex_state = 128},
  [17186] = {.lex_state = 455},
  [17187] = {.lex_state = 224, .external_lex_state = 129},
  [17188] = {.lex_state = 457},
  [17189] = {.lex_state = 190, .external_lex_state = 131},
  [17190] = {.lex_state = 455},
  [17191] = {.lex_state = 455, .external_lex_state = 130},
  [17192] = {.lex_state = 457, .external_lex_state = 128},
  [17193] = {.lex_state = 455},
  [17194] = {.lex_state = 141},
  [17195] = {.lex_state = 457},
  [17196] = {.lex_state = 203},
  [17197] = {.lex_state = 141},
  [17198] = {.lex_state = 224, .external_lex_state = 129},
  [17199] = {.lex_state = 455},
  [17200] = {.lex_state = 141},
  [17201] = {.lex_state = 203},
  [17202] = {.lex_state = 455},
  [17203] = {.lex_state = 455},
  [17204] = {.lex_state = 457},
  [17205] = {.lex_state = 455, .external_lex_state = 100},
  [17206] = {.lex_state = 203},
  [17207] = {.lex_state = 455},
  [17208] = {.lex_state = 141},
  [17209] = {.lex_state = 203},
  [17210] = {.lex_state = 457, .external_lex_state = 128},
  [17211] = {.lex_state = 455},
  [17212] = {.lex_state = 141},
  [17213] = {.lex_state = 203},
  [17214] = {.lex_state = 457},
  [17215] = {.lex_state = 455},
  [17216] = {.lex_state = 457, .external_lex_state = 131},
  [17217] = {.lex_state = 190, .external_lex_state = 131},
  [17218] = {.lex_state = 224, .external_lex_state = 129},
  [17219] = {.lex_state = 455},
  [17220] = {.lex_state = 455, .external_lex_state = 100},
  [17221] = {.lex_state = 457, .external_lex_state = 131},
  [17222] = {.lex_state = 455},
  [17223] = {.lex_state = 141},
  [17224] = {.lex_state = 457},
  [17225] = {.lex_state = 141},
  [17226] = {.lex_state = 203},
  [17227] = {.lex_state = 224, .external_lex_state = 129},
  [17228] = {.lex_state = 457},
  [17229] = {.lex_state = 190, .external_lex_state = 131},
  [17230] = {.lex_state = 455},
  [17231] = {.lex_state = 457},
  [17232] = {.lex_state = 203},
  [17233] = {.lex_state = 224, .external_lex_state = 129},
  [17234] = {.lex_state = 203},
  [17235] = {.lex_state = 141},
  [17236] = {.lex_state = 203},
  [17237] = {.lex_state = 224, .external_lex_state = 129},
  [17238] = {.lex_state = 224, .external_lex_state = 129},
  [17239] = {.lex_state = 457},
  [17240] = {.lex_state = 457},
  [17241] = {.lex_state = 224, .external_lex_state = 129},
  [17242] = {.lex_state = 224, .external_lex_state = 129},
  [17243] = {.lex_state = 224, .external_lex_state = 108},
  [17244] = {.lex_state = 457, .external_lex_state = 114},
  [17245] = {.lex_state = 203},
  [17246] = {.lex_state = 224, .external_lex_state = 108},
  [17247] = {.lex_state = 141},
  [17248] = {.lex_state = 457, .external_lex_state = 131},
  [17249] = {.lex_state = 203},
  [17250] = {.lex_state = 224, .external_lex_state = 108},
  [17251] = {.lex_state = 224, .external_lex_state = 129},
  [17252] = {.lex_state = 190, .external_lex_state = 131},
  [17253] = {.lex_state = 457},
  [17254] = {.lex_state = 224, .external_lex_state = 108},
  [17255] = {.lex_state = 457, .external_lex_state = 131},
  [17256] = {.lex_state = 224, .external_lex_state = 129},
  [17257] = {.lex_state = 190, .external_lex_state = 131},
  [17258] = {.lex_state = 457, .external_lex_state = 114},
  [17259] = {.lex_state = 203},
  [17260] = {.lex_state = 224, .external_lex_state = 129},
  [17261] = {.lex_state = 141},
  [17262] = {.lex_state = 203},
  [17263] = {.lex_state = 141},
  [17264] = {.lex_state = 457},
  [17265] = {.lex_state = 141},
  [17266] = {.lex_state = 457, .external_lex_state = 99},
  [17267] = {.lex_state = 457},
  [17268] = {.lex_state = 203},
  [17269] = {.lex_state = 203},
  [17270] = {.lex_state = 457},
  [17271] = {.lex_state = 141},
  [17272] = {.lex_state = 457, .external_lex_state = 131},
  [17273] = {.lex_state = 190, .external_lex_state = 131},
  [17274] = {.lex_state = 224},
  [17275] = {.lex_state = 457, .external_lex_state = 131},
  [17276] = {.lex_state = 457},
  [17277] = {.lex_state = 141},
  [17278] = {.lex_state = 141},
  [17279] = {.lex_state = 190, .external_lex_state = 131},
  [17280] = {.lex_state = 203},
  [17281] = {.lex_state = 141},
  [17282] = {.lex_state = 141},
  [17283] = {.lex_state = 224, .external_lex_state = 129},
  [17284] = {.lex_state = 457, .external_lex_state = 131},
  [17285] = {.lex_state = 190, .external_lex_state = 131},
  [17286] = {.lex_state = 457, .external_lex_state = 131},
  [17287] = {.lex_state = 190, .external_lex_state = 131},
  [17288] = {.lex_state = 141},
  [17289] = {.lex_state = 455, .external_lex_state = 100},
  [17290] = {.lex_state = 141},
  [17291] = {.lex_state = 455},
  [17292] = {.lex_state = 203},
  [17293] = {.lex_state = 141},
  [17294] = {.lex_state = 457},
  [17295] = {.lex_state = 141},
  [17296] = {.lex_state = 203},
  [17297] = {.lex_state = 457},
  [17298] = {.lex_state = 457},
  [17299] = {.lex_state = 457},
  [17300] = {.lex_state = 457},
  [17301] = {.lex_state = 141},
  [17302] = {.lex_state = 457},
  [17303] = {.lex_state = 141},
  [17304] = {.lex_state = 457},
  [17305] = {.lex_state = 457, .external_lex_state = 131},
  [17306] = {.lex_state = 203},
  [17307] = {.lex_state = 190, .external_lex_state = 131},
  [17308] = {.lex_state = 141},
  [17309] = {.lex_state = 224, .external_lex_state = 129},
  [17310] = {.lex_state = 141},
  [17311] = {.lex_state = 457},
  [17312] = {.lex_state = 203},
  [17313] = {.lex_state = 457},
  [17314] = {.lex_state = 203},
  [17315] = {.lex_state = 203},
  [17316] = {.lex_state = 457},
  [17317] = {.lex_state = 224},
  [17318] = {.lex_state = 141},
  [17319] = {.lex_state = 457},
  [17320] = {.lex_state = 203},
  [17321] = {.lex_state = 203},
  [17322] = {.lex_state = 457, .external_lex_state = 99},
  [17323] = {.lex_state = 457, .external_lex_state = 100},
  [17324] = {.lex_state = 224, .external_lex_state = 108},
  [17325] = {.lex_state = 457, .external_lex_state = 131},
  [17326] = {.lex_state = 141},
  [17327] = {.lex_state = 457, .external_lex_state = 131},
  [17328] = {.lex_state = 455},
  [17329] = {.lex_state = 203},
  [17330] = {.lex_state = 190, .external_lex_state = 131},
  [17331] = {.lex_state = 190, .external_lex_state = 131},
  [17332] = {.lex_state = 457, .external_lex_state = 100},
  [17333] = {.lex_state = 457},
  [17334] = {.lex_state = 455},
  [17335] = {.lex_state = 141},
  [17336] = {.lex_state = 455},
  [17337] = {.lex_state = 141},
  [17338] = {.lex_state = 455},
  [17339] = {.lex_state = 203},
  [17340] = {.lex_state = 457},
  [17341] = {.lex_state = 203},
  [17342] = {.lex_state = 141},
  [17343] = {.lex_state = 455, .external_lex_state = 100},
  [17344] = {.lex_state = 203, .external_lex_state = 125},
  [17345] = {.lex_state = 141},
  [17346] = {.lex_state = 203},
  [17347] = {.lex_state = 457},
  [17348] = {.lex_state = 457},
  [17349] = {.lex_state = 457},
  [17350] = {.lex_state = 457, .external_lex_state = 99},
  [17351] = {.lex_state = 457, .external_lex_state = 114},
  [17352] = {.lex_state = 203},
  [17353] = {.lex_state = 457, .external_lex_state = 100},
  [17354] = {.lex_state = 455},
  [17355] = {.lex_state = 455},
  [17356] = {.lex_state = 224, .external_lex_state = 108},
  [17357] = {.lex_state = 203},
  [17358] = {.lex_state = 141},
  [17359] = {.lex_state = 203},
  [17360] = {.lex_state = 455},
  [17361] = {.lex_state = 455},
  [17362] = {.lex_state = 455},
  [17363] = {.lex_state = 457},
  [17364] = {.lex_state = 455, .external_lex_state = 100},
  [17365] = {.lex_state = 457, .external_lex_state = 100},
  [17366] = {.lex_state = 457, .external_lex_state = 100},
  [17367] = {.lex_state = 203},
  [17368] = {.lex_state = 455},
  [17369] = {.lex_state = 203},
  [17370] = {.lex_state = 455},
  [17371] = {.lex_state = 141},
  [17372] = {.lex_state = 141},
  [17373] = {.lex_state = 203},
  [17374] = {.lex_state = 457, .external_lex_state = 131},
  [17375] = {.lex_state = 203},
  [17376] = {.lex_state = 455},
  [17377] = {.lex_state = 190, .external_lex_state = 131},
  [17378] = {.lex_state = 455},
  [17379] = {.lex_state = 455},
  [17380] = {.lex_state = 203},
  [17381] = {.lex_state = 203},
  [17382] = {.lex_state = 457, .external_lex_state = 131},
  [17383] = {.lex_state = 190, .external_lex_state = 131},
  [17384] = {.lex_state = 455},
  [17385] = {.lex_state = 141},
  [17386] = {.lex_state = 203},
  [17387] = {.lex_state = 455, .external_lex_state = 100},
  [17388] = {.lex_state = 203},
  [17389] = {.lex_state = 455},
  [17390] = {.lex_state = 141},
  [17391] = {.lex_state = 141},
  [17392] = {.lex_state = 455},
  [17393] = {.lex_state = 224, .external_lex_state = 108},
  [17394] = {.lex_state = 455},
  [17395] = {.lex_state = 455, .external_lex_state = 100},
  [17396] = {.lex_state = 224},
  [17397] = {.lex_state = 457, .external_lex_state = 131},
  [17398] = {.lex_state = 190, .external_lex_state = 131},
  [17399] = {.lex_state = 141},
  [17400] = {.lex_state = 457},
  [17401] = {.lex_state = 203},
  [17402] = {.lex_state = 455},
  [17403] = {.lex_state = 224, .external_lex_state = 108},
  [17404] = {.lex_state = 457},
  [17405] = {.lex_state = 141},
  [17406] = {.lex_state = 455},
  [17407] = {.lex_state = 203},
  [17408] = {.lex_state = 455},
  [17409] = {.lex_state = 203},
  [17410] = {.lex_state = 455},
  [17411] = {.lex_state = 203},
  [17412] = {.lex_state = 203},
  [17413] = {.lex_state = 457},
  [17414] = {.lex_state = 457},
  [17415] = {.lex_state = 455},
  [17416] = {.lex_state = 203},
  [17417] = {.lex_state = 457},
  [17418] = {.lex_state = 203},
  [17419] = {.lex_state = 455},
  [17420] = {.lex_state = 457},
  [17421] = {.lex_state = 224},
  [17422] = {.lex_state = 455},
  [17423] = {.lex_state = 203},
  [17424] = {.lex_state = 455},
  [17425] = {.lex_state = 457},
  [17426] = {.lex_state = 457},
  [17427] = {.lex_state = 203},
  [17428] = {.lex_state = 455, .external_lex_state = 121},
  [17429] = {.lex_state = 457, .external_lex_state = 100},
  [17430] = {.lex_state = 455, .external_lex_state = 100},
  [17431] = {.lex_state = 455},
  [17432] = {.lex_state = 141},
  [17433] = {.lex_state = 203},
  [17434] = {.lex_state = 203},
  [17435] = {.lex_state = 455},
  [17436] = {.lex_state = 457, .external_lex_state = 100},
  [17437] = {.lex_state = 203},
  [17438] = {.lex_state = 455},
  [17439] = {.lex_state = 455},
  [17440] = {.lex_state = 455},
  [17441] = {.lex_state = 203},
  [17442] = {.lex_state = 224},
  [17443] = {.lex_state = 455},
  [17444] = {.lex_state = 455, .external_lex_state = 100},
  [17445] = {.lex_state = 203},
  [17446] = {.lex_state = 224, .external_lex_state = 108},
  [17447] = {.lex_state = 457, .external_lex_state = 131},
  [17448] = {.lex_state = 203},
  [17449] = {.lex_state = 455},
  [17450] = {.lex_state = 224},
  [17451] = {.lex_state = 190, .external_lex_state = 131},
  [17452] = {.lex_state = 455},
  [17453] = {.lex_state = 455},
  [17454] = {.lex_state = 203},
  [17455] = {.lex_state = 457, .external_lex_state = 99},
  [17456] = {.lex_state = 203},
  [17457] = {.lex_state = 457, .external_lex_state = 121},
  [17458] = {.lex_state = 457},
  [17459] = {.lex_state = 455},
  [17460] = {.lex_state = 457},
  [17461] = {.lex_state = 141},
  [17462] = {.lex_state = 203},
  [17463] = {.lex_state = 455},
  [17464] = {.lex_state = 457, .external_lex_state = 100},
  [17465] = {.lex_state = 455},
  [17466] = {.lex_state = 203},
  [17467] = {.lex_state = 455},
  [17468] = {.lex_state = 203},
  [17469] = {.lex_state = 455},
  [17470] = {.lex_state = 203},
  [17471] = {.lex_state = 457, .external_lex_state = 100},
  [17472] = {.lex_state = 455},
  [17473] = {.lex_state = 455, .external_lex_state = 130},
  [17474] = {.lex_state = 457, .external_lex_state = 100},
  [17475] = {.lex_state = 203},
  [17476] = {.lex_state = 457, .external_lex_state = 128},
  [17477] = {.lex_state = 457},
  [17478] = {.lex_state = 455},
  [17479] = {.lex_state = 457, .external_lex_state = 100},
  [17480] = {.lex_state = 455},
  [17481] = {.lex_state = 455},
  [17482] = {.lex_state = 203},
  [17483] = {.lex_state = 224},
  [17484] = {.lex_state = 455},
  [17485] = {.lex_state = 455},
  [17486] = {.lex_state = 457, .external_lex_state = 128},
  [17487] = {.lex_state = 455},
  [17488] = {.lex_state = 455},
  [17489] = {.lex_state = 455},
  [17490] = {.lex_state = 457},
  [17491] = {.lex_state = 457, .external_lex_state = 121},
  [17492] = {.lex_state = 224, .external_lex_state = 129},
  [17493] = {.lex_state = 457, .external_lex_state = 121},
  [17494] = {.lex_state = 203},
  [17495] = {.lex_state = 457, .external_lex_state = 122},
  [17496] = {.lex_state = 455},
  [17497] = {.lex_state = 455},
  [17498] = {.lex_state = 455},
  [17499] = {.lex_state = 455},
  [17500] = {.lex_state = 203},
  [17501] = {.lex_state = 457, .external_lex_state = 114},
  [17502] = {.lex_state = 455},
  [17503] = {.lex_state = 455},
  [17504] = {.lex_state = 455},
  [17505] = {.lex_state = 455},
  [17506] = {.lex_state = 457, .external_lex_state = 114},
  [17507] = {.lex_state = 224, .external_lex_state = 129},
  [17508] = {.lex_state = 457},
  [17509] = {.lex_state = 224, .external_lex_state = 108},
  [17510] = {.lex_state = 455},
  [17511] = {.lex_state = 455},
  [17512] = {.lex_state = 457, .external_lex_state = 121},
  [17513] = {.lex_state = 224, .external_lex_state = 108},
  [17514] = {.lex_state = 224, .external_lex_state = 129},
  [17515] = {.lex_state = 141},
  [17516] = {.lex_state = 224, .external_lex_state = 129},
  [17517] = {.lex_state = 224, .external_lex_state = 108},
  [17518] = {.lex_state = 203},
  [17519] = {.lex_state = 203},
  [17520] = {.lex_state = 457, .external_lex_state = 100},
  [17521] = {.lex_state = 457},
  [17522] = {.lex_state = 457, .external_lex_state = 99},
  [17523] = {.lex_state = 203},
  [17524] = {.lex_state = 224, .external_lex_state = 108},
  [17525] = {.lex_state = 224, .external_lex_state = 129},
  [17526] = {.lex_state = 224, .external_lex_state = 129},
  [17527] = {.lex_state = 455},
  [17528] = {.lex_state = 224, .external_lex_state = 108},
  [17529] = {.lex_state = 457, .external_lex_state = 100},
  [17530] = {.lex_state = 457},
  [17531] = {.lex_state = 455},
  [17532] = {.lex_state = 457},
  [17533] = {.lex_state = 457},
  [17534] = {.lex_state = 138},
  [17535] = {.lex_state = 457, .external_lex_state = 99},
  [17536] = {.lex_state = 203},
  [17537] = {.lex_state = 224, .external_lex_state = 129},
  [17538] = {.lex_state = 457, .external_lex_state = 128},
  [17539] = {.lex_state = 141},
  [17540] = {.lex_state = 457},
  [17541] = {.lex_state = 224, .external_lex_state = 129},
  [17542] = {.lex_state = 224, .external_lex_state = 129},
  [17543] = {.lex_state = 224, .external_lex_state = 129},
  [17544] = {.lex_state = 457},
  [17545] = {.lex_state = 203},
  [17546] = {.lex_state = 203},
  [17547] = {.lex_state = 457, .external_lex_state = 128},
  [17548] = {.lex_state = 203},
  [17549] = {.lex_state = 224, .external_lex_state = 129},
  [17550] = {.lex_state = 224, .external_lex_state = 129},
  [17551] = {.lex_state = 457},
  [17552] = {.lex_state = 224, .external_lex_state = 129},
  [17553] = {.lex_state = 224, .external_lex_state = 129},
  [17554] = {.lex_state = 224, .external_lex_state = 129},
  [17555] = {.lex_state = 457, .external_lex_state = 100},
  [17556] = {.lex_state = 457},
  [17557] = {.lex_state = 141},
  [17558] = {.lex_state = 141},
  [17559] = {.lex_state = 457, .external_lex_state = 99},
  [17560] = {.lex_state = 224, .external_lex_state = 129},
  [17561] = {.lex_state = 141},
  [17562] = {.lex_state = 457},
  [17563] = {.lex_state = 203},
  [17564] = {.lex_state = 457, .external_lex_state = 99},
  [17565] = {.lex_state = 457, .external_lex_state = 131},
  [17566] = {.lex_state = 190, .external_lex_state = 131},
  [17567] = {.lex_state = 203},
  [17568] = {.lex_state = 457},
  [17569] = {.lex_state = 457},
  [17570] = {.lex_state = 224, .external_lex_state = 108},
  [17571] = {.lex_state = 457, .external_lex_state = 131},
  [17572] = {.lex_state = 141},
  [17573] = {.lex_state = 224, .external_lex_state = 129},
  [17574] = {.lex_state = 141},
  [17575] = {.lex_state = 457, .external_lex_state = 99},
  [17576] = {.lex_state = 457},
  [17577] = {.lex_state = 455, .external_lex_state = 130},
  [17578] = {.lex_state = 203},
  [17579] = {.lex_state = 457, .external_lex_state = 131},
  [17580] = {.lex_state = 190, .external_lex_state = 131},
  [17581] = {.lex_state = 457},
  [17582] = {.lex_state = 457},
  [17583] = {.lex_state = 455},
  [17584] = {.lex_state = 224, .external_lex_state = 129},
  [17585] = {.lex_state = 457, .external_lex_state = 131},
  [17586] = {.lex_state = 457},
  [17587] = {.lex_state = 224, .external_lex_state = 129},
  [17588] = {.lex_state = 141},
  [17589] = {.lex_state = 455},
  [17590] = {.lex_state = 141},
  [17591] = {.lex_state = 203},
  [17592] = {.lex_state = 190, .external_lex_state = 131},
  [17593] = {.lex_state = 457},
  [17594] = {.lex_state = 203},
  [17595] = {.lex_state = 203},
  [17596] = {.lex_state = 457},
  [17597] = {.lex_state = 457},
  [17598] = {.lex_state = 457, .external_lex_state = 131},
  [17599] = {.lex_state = 190, .external_lex_state = 131},
  [17600] = {.lex_state = 203},
  [17601] = {.lex_state = 457},
  [17602] = {.lex_state = 457, .external_lex_state = 100},
  [17603] = {.lex_state = 224, .external_lex_state = 129},
  [17604] = {.lex_state = 457, .external_lex_state = 122},
  [17605] = {.lex_state = 141},
  [17606] = {.lex_state = 457},
  [17607] = {.lex_state = 457, .external_lex_state = 100},
  [17608] = {.lex_state = 455},
  [17609] = {.lex_state = 457},
  [17610] = {.lex_state = 141},
  [17611] = {.lex_state = 203},
  [17612] = {.lex_state = 224},
  [17613] = {.lex_state = 455, .external_lex_state = 130},
  [17614] = {.lex_state = 457},
  [17615] = {.lex_state = 457},
  [17616] = {.lex_state = 457, .external_lex_state = 128},
  [17617] = {.lex_state = 457},
  [17618] = {.lex_state = 203},
  [17619] = {.lex_state = 457, .external_lex_state = 128},
  [17620] = {.lex_state = 141},
  [17621] = {.lex_state = 457, .external_lex_state = 100},
  [17622] = {.lex_state = 224, .external_lex_state = 129},
  [17623] = {.lex_state = 457, .external_lex_state = 99},
  [17624] = {.lex_state = 141},
  [17625] = {.lex_state = 457, .external_lex_state = 100},
  [17626] = {.lex_state = 224, .external_lex_state = 129},
  [17627] = {.lex_state = 457},
  [17628] = {.lex_state = 455},
  [17629] = {.lex_state = 224, .external_lex_state = 129},
  [17630] = {.lex_state = 203},
  [17631] = {.lex_state = 457, .external_lex_state = 100},
  [17632] = {.lex_state = 457, .external_lex_state = 114},
  [17633] = {.lex_state = 457, .external_lex_state = 99},
  [17634] = {.lex_state = 224, .external_lex_state = 108},
  [17635] = {.lex_state = 224},
  [17636] = {.lex_state = 457, .external_lex_state = 128},
  [17637] = {.lex_state = 457},
  [17638] = {.lex_state = 224, .external_lex_state = 108},
  [17639] = {.lex_state = 457},
  [17640] = {.lex_state = 457, .external_lex_state = 99},
  [17641] = {.lex_state = 457},
  [17642] = {.lex_state = 141},
  [17643] = {.lex_state = 224, .external_lex_state = 129},
  [17644] = {.lex_state = 457, .external_lex_state = 99},
  [17645] = {.lex_state = 141},
  [17646] = {.lex_state = 224, .external_lex_state = 129},
  [17647] = {.lex_state = 455, .external_lex_state = 100},
  [17648] = {.lex_state = 457, .external_lex_state = 99},
  [17649] = {.lex_state = 203},
  [17650] = {.lex_state = 224, .external_lex_state = 129},
  [17651] = {.lex_state = 457},
  [17652] = {.lex_state = 457, .external_lex_state = 100},
  [17653] = {.lex_state = 203},
  [17654] = {.lex_state = 457, .external_lex_state = 128},
  [17655] = {.lex_state = 224},
  [17656] = {.lex_state = 457, .external_lex_state = 99},
  [17657] = {.lex_state = 457, .external_lex_state = 131},
  [17658] = {.lex_state = 457},
  [17659] = {.lex_state = 224},
  [17660] = {.lex_state = 457},
  [17661] = {.lex_state = 457},
  [17662] = {.lex_state = 203},
  [17663] = {.lex_state = 457, .external_lex_state = 128},
  [17664] = {.lex_state = 141},
  [17665] = {.lex_state = 190, .external_lex_state = 131},
  [17666] = {.lex_state = 457, .external_lex_state = 100},
  [17667] = {.lex_state = 141},
  [17668] = {.lex_state = 457, .external_lex_state = 100},
  [17669] = {.lex_state = 457, .external_lex_state = 99},
  [17670] = {.lex_state = 203},
  [17671] = {.lex_state = 457},
  [17672] = {.lex_state = 203},
  [17673] = {.lex_state = 457, .external_lex_state = 99},
  [17674] = {.lex_state = 224, .external_lex_state = 108},
  [17675] = {.lex_state = 224, .external_lex_state = 108},
  [17676] = {.lex_state = 224, .external_lex_state = 108},
  [17677] = {.lex_state = 457, .external_lex_state = 100},
  [17678] = {.lex_state = 455, .external_lex_state = 99},
  [17679] = {.lex_state = 224, .external_lex_state = 108},
  [17680] = {.lex_state = 457},
  [17681] = {.lex_state = 224, .external_lex_state = 129},
  [17682] = {.lex_state = 457, .external_lex_state = 99},
  [17683] = {.lex_state = 457, .external_lex_state = 99},
  [17684] = {.lex_state = 224, .external_lex_state = 129},
  [17685] = {.lex_state = 457, .external_lex_state = 99},
  [17686] = {.lex_state = 457},
  [17687] = {.lex_state = 457, .external_lex_state = 128},
  [17688] = {.lex_state = 141},
  [17689] = {.lex_state = 224, .external_lex_state = 129},
  [17690] = {.lex_state = 457, .external_lex_state = 100},
  [17691] = {.lex_state = 457},
  [17692] = {.lex_state = 457, .external_lex_state = 114},
  [17693] = {.lex_state = 224, .external_lex_state = 108},
  [17694] = {.lex_state = 203},
  [17695] = {.lex_state = 224, .external_lex_state = 108},
  [17696] = {.lex_state = 224, .external_lex_state = 108},
  [17697] = {.lex_state = 224, .external_lex_state = 108},
  [17698] = {.lex_state = 457},
  [17699] = {.lex_state = 455},
  [17700] = {.lex_state = 457, .external_lex_state = 131},
  [17701] = {.lex_state = 190, .external_lex_state = 131},
  [17702] = {.lex_state = 455},
  [17703] = {.lex_state = 141},
  [17704] = {.lex_state = 455, .external_lex_state = 100},
  [17705] = {.lex_state = 457, .external_lex_state = 131},
  [17706] = {.lex_state = 190, .external_lex_state = 131},
  [17707] = {.lex_state = 203},
  [17708] = {.lex_state = 455},
  [17709] = {.lex_state = 455},
  [17710] = {.lex_state = 457, .external_lex_state = 131},
  [17711] = {.lex_state = 457, .external_lex_state = 131},
  [17712] = {.lex_state = 455},
  [17713] = {.lex_state = 457, .external_lex_state = 132},
  [17714] = {.lex_state = 457, .external_lex_state = 131},
  [17715] = {.lex_state = 455},
  [17716] = {.lex_state = 455},
  [17717] = {.lex_state = 457, .external_lex_state = 131},
  [17718] = {.lex_state = 457, .external_lex_state = 131},
  [17719] = {.lex_state = 457, .external_lex_state = 99},
  [17720] = {.lex_state = 457, .external_lex_state = 132},
  [17721] = {.lex_state = 457, .external_lex_state = 131},
  [17722] = {.lex_state = 457, .external_lex_state = 131},
  [17723] = {.lex_state = 457, .external_lex_state = 132},
  [17724] = {.lex_state = 457, .external_lex_state = 131},
  [17725] = {.lex_state = 457, .external_lex_state = 131},
  [17726] = {.lex_state = 457, .external_lex_state = 131},
  [17727] = {.lex_state = 455, .external_lex_state = 120},
  [17728] = {.lex_state = 457, .external_lex_state = 101},
  [17729] = {.lex_state = 457, .external_lex_state = 131},
  [17730] = {.lex_state = 455},
  [17731] = {.lex_state = 457, .external_lex_state = 131},
  [17732] = {.lex_state = 455},
  [17733] = {.lex_state = 455},
  [17734] = {.lex_state = 457, .external_lex_state = 131},
  [17735] = {.lex_state = 457, .external_lex_state = 131},
  [17736] = {.lex_state = 457, .external_lex_state = 101},
  [17737] = {.lex_state = 457, .external_lex_state = 131},
  [17738] = {.lex_state = 455},
  [17739] = {.lex_state = 457, .external_lex_state = 131},
  [17740] = {.lex_state = 457, .external_lex_state = 131},
  [17741] = {.lex_state = 457, .external_lex_state = 131},
  [17742] = {.lex_state = 457, .external_lex_state = 132},
  [17743] = {.lex_state = 457, .external_lex_state = 131},
  [17744] = {.lex_state = 457, .external_lex_state = 131},
  [17745] = {.lex_state = 457, .external_lex_state = 131},
  [17746] = {.lex_state = 457, .external_lex_state = 131},
  [17747] = {.lex_state = 457, .external_lex_state = 131},
  [17748] = {.lex_state = 457, .external_lex_state = 131},
  [17749] = {.lex_state = 457, .external_lex_state = 131},
  [17750] = {.lex_state = 457, .external_lex_state = 114},
  [17751] = {.lex_state = 457, .external_lex_state = 131},
  [17752] = {.lex_state = 457, .external_lex_state = 132},
  [17753] = {.lex_state = 457, .external_lex_state = 131},
  [17754] = {.lex_state = 457, .external_lex_state = 131},
  [17755] = {.lex_state = 457, .external_lex_state = 131},
  [17756] = {.lex_state = 455, .external_lex_state = 120},
  [17757] = {.lex_state = 457, .external_lex_state = 101},
  [17758] = {.lex_state = 457, .external_lex_state = 114},
  [17759] = {.lex_state = 455},
  [17760] = {.lex_state = 457, .external_lex_state = 131},
  [17761] = {.lex_state = 457, .external_lex_state = 131},
  [17762] = {.lex_state = 457, .external_lex_state = 131},
  [17763] = {.lex_state = 457, .external_lex_state = 114},
  [17764] = {.lex_state = 457, .external_lex_state = 131},
  [17765] = {.lex_state = 457, .external_lex_state = 101},
  [17766] = {.lex_state = 457, .external_lex_state = 114},
  [17767] = {.lex_state = 455},
  [17768] = {.lex_state = 457, .external_lex_state = 121},
  [17769] = {.lex_state = 457, .external_lex_state = 131},
  [17770] = {.lex_state = 457, .external_lex_state = 131},
  [17771] = {.lex_state = 457, .external_lex_state = 132},
  [17772] = {.lex_state = 457, .external_lex_state = 132},
  [17773] = {.lex_state = 457, .external_lex_state = 131},
  [17774] = {.lex_state = 457, .external_lex_state = 131},
  [17775] = {.lex_state = 457, .external_lex_state = 131},
  [17776] = {.lex_state = 457, .external_lex_state = 99},
  [17777] = {.lex_state = 457, .external_lex_state = 131},
  [17778] = {.lex_state = 457, .external_lex_state = 131},
  [17779] = {.lex_state = 457, .external_lex_state = 131},
  [17780] = {.lex_state = 457, .external_lex_state = 131},
  [17781] = {.lex_state = 457, .external_lex_state = 132},
  [17782] = {.lex_state = 455},
  [17783] = {.lex_state = 457, .external_lex_state = 131},
  [17784] = {.lex_state = 457, .external_lex_state = 101},
  [17785] = {.lex_state = 455, .external_lex_state = 120},
  [17786] = {.lex_state = 457, .external_lex_state = 101},
  [17787] = {.lex_state = 457, .external_lex_state = 99},
  [17788] = {.lex_state = 455},
  [17789] = {.lex_state = 457, .external_lex_state = 131},
  [17790] = {.lex_state = 457, .external_lex_state = 132},
  [17791] = {.lex_state = 455},
  [17792] = {.lex_state = 457, .external_lex_state = 121},
  [17793] = {.lex_state = 457, .external_lex_state = 101},
  [17794] = {.lex_state = 203},
  [17795] = {.lex_state = 455},
  [17796] = {.lex_state = 455},
  [17797] = {.lex_state = 457, .external_lex_state = 114},
  [17798] = {.lex_state = 457, .external_lex_state = 131},
  [17799] = {.lex_state = 457, .external_lex_state = 132},
  [17800] = {.lex_state = 455},
  [17801] = {.lex_state = 457, .external_lex_state = 121},
  [17802] = {.lex_state = 455, .external_lex_state = 122},
  [17803] = {.lex_state = 457, .external_lex_state = 131},
  [17804] = {.lex_state = 457, .external_lex_state = 114},
  [17805] = {.lex_state = 457, .external_lex_state = 131},
  [17806] = {.lex_state = 457, .external_lex_state = 99},
  [17807] = {.lex_state = 457, .external_lex_state = 131},
  [17808] = {.lex_state = 457, .external_lex_state = 131},
  [17809] = {.lex_state = 457, .external_lex_state = 132},
  [17810] = {.lex_state = 457, .external_lex_state = 131},
  [17811] = {.lex_state = 457, .external_lex_state = 131},
  [17812] = {.lex_state = 455},
  [17813] = {.lex_state = 455, .external_lex_state = 120},
  [17814] = {.lex_state = 457, .external_lex_state = 101},
  [17815] = {.lex_state = 457, .external_lex_state = 99},
  [17816] = {.lex_state = 455},
  [17817] = {.lex_state = 457, .external_lex_state = 131},
  [17818] = {.lex_state = 457, .external_lex_state = 132},
  [17819] = {.lex_state = 457, .external_lex_state = 131},
  [17820] = {.lex_state = 457, .external_lex_state = 131},
  [17821] = {.lex_state = 457, .external_lex_state = 99},
  [17822] = {.lex_state = 457, .external_lex_state = 101},
  [17823] = {.lex_state = 457, .external_lex_state = 131},
  [17824] = {.lex_state = 455},
  [17825] = {.lex_state = 457, .external_lex_state = 132},
  [17826] = {.lex_state = 224, .external_lex_state = 129},
  [17827] = {.lex_state = 457, .external_lex_state = 131},
  [17828] = {.lex_state = 455, .external_lex_state = 120},
  [17829] = {.lex_state = 457, .external_lex_state = 131},
  [17830] = {.lex_state = 457, .external_lex_state = 131},
  [17831] = {.lex_state = 457, .external_lex_state = 131},
  [17832] = {.lex_state = 455},
  [17833] = {.lex_state = 457, .external_lex_state = 101},
  [17834] = {.lex_state = 457, .external_lex_state = 131},
  [17835] = {.lex_state = 457, .external_lex_state = 131},
  [17836] = {.lex_state = 457, .external_lex_state = 131},
  [17837] = {.lex_state = 455},
  [17838] = {.lex_state = 457, .external_lex_state = 131},
  [17839] = {.lex_state = 457, .external_lex_state = 131},
  [17840] = {.lex_state = 455},
  [17841] = {.lex_state = 457, .external_lex_state = 121},
  [17842] = {.lex_state = 455, .external_lex_state = 120},
  [17843] = {.lex_state = 455},
  [17844] = {.lex_state = 457, .external_lex_state = 101},
  [17845] = {.lex_state = 457, .external_lex_state = 131},
  [17846] = {.lex_state = 457, .external_lex_state = 131},
  [17847] = {.lex_state = 457, .external_lex_state = 131},
  [17848] = {.lex_state = 455},
  [17849] = {.lex_state = 457, .external_lex_state = 131},
  [17850] = {.lex_state = 457, .external_lex_state = 131},
  [17851] = {.lex_state = 455},
  [17852] = {.lex_state = 457, .external_lex_state = 131},
  [17853] = {.lex_state = 455},
  [17854] = {.lex_state = 203},
  [17855] = {.lex_state = 457, .external_lex_state = 131},
  [17856] = {.lex_state = 457, .external_lex_state = 131},
  [17857] = {.lex_state = 457, .external_lex_state = 131},
  [17858] = {.lex_state = 455},
  [17859] = {.lex_state = 457, .external_lex_state = 101},
  [17860] = {.lex_state = 455},
  [17861] = {.lex_state = 457, .external_lex_state = 131},
  [17862] = {.lex_state = 457, .external_lex_state = 121},
  [17863] = {.lex_state = 455},
  [17864] = {.lex_state = 455},
  [17865] = {.lex_state = 455},
  [17866] = {.lex_state = 457, .external_lex_state = 131},
  [17867] = {.lex_state = 457, .external_lex_state = 131},
  [17868] = {.lex_state = 455},
  [17869] = {.lex_state = 457, .external_lex_state = 131},
  [17870] = {.lex_state = 457, .external_lex_state = 131},
  [17871] = {.lex_state = 457, .external_lex_state = 99},
  [17872] = {.lex_state = 455},
  [17873] = {.lex_state = 457, .external_lex_state = 131},
  [17874] = {.lex_state = 457, .external_lex_state = 131},
  [17875] = {.lex_state = 457, .external_lex_state = 131},
  [17876] = {.lex_state = 457, .external_lex_state = 114},
  [17877] = {.lex_state = 455},
  [17878] = {.lex_state = 457, .external_lex_state = 114},
  [17879] = {.lex_state = 457, .external_lex_state = 131},
  [17880] = {.lex_state = 457, .external_lex_state = 131},
  [17881] = {.lex_state = 457, .external_lex_state = 131},
  [17882] = {.lex_state = 455},
  [17883] = {.lex_state = 457, .external_lex_state = 99},
  [17884] = {.lex_state = 457, .external_lex_state = 114},
  [17885] = {.lex_state = 457, .external_lex_state = 131},
  [17886] = {.lex_state = 455},
  [17887] = {.lex_state = 457, .external_lex_state = 131},
  [17888] = {.lex_state = 457, .external_lex_state = 131},
  [17889] = {.lex_state = 457, .external_lex_state = 131},
  [17890] = {.lex_state = 455},
  [17891] = {.lex_state = 457, .external_lex_state = 131},
  [17892] = {.lex_state = 455},
  [17893] = {.lex_state = 457, .external_lex_state = 131},
  [17894] = {.lex_state = 455},
  [17895] = {.lex_state = 457, .external_lex_state = 121},
  [17896] = {.lex_state = 457, .external_lex_state = 131},
  [17897] = {.lex_state = 457, .external_lex_state = 131},
  [17898] = {.lex_state = 457, .external_lex_state = 131},
  [17899] = {.lex_state = 455},
  [17900] = {.lex_state = 457, .external_lex_state = 131},
  [17901] = {.lex_state = 457, .external_lex_state = 131},
  [17902] = {.lex_state = 455},
  [17903] = {.lex_state = 457, .external_lex_state = 131},
  [17904] = {.lex_state = 457, .external_lex_state = 99},
  [17905] = {.lex_state = 457, .external_lex_state = 131},
  [17906] = {.lex_state = 203},
  [17907] = {.lex_state = 455},
  [17908] = {.lex_state = 457, .external_lex_state = 131},
  [17909] = {.lex_state = 457, .external_lex_state = 131},
  [17910] = {.lex_state = 455},
  [17911] = {.lex_state = 455},
  [17912] = {.lex_state = 457, .external_lex_state = 131},
  [17913] = {.lex_state = 455},
  [17914] = {.lex_state = 457, .external_lex_state = 131},
  [17915] = {.lex_state = 457, .external_lex_state = 131},
  [17916] = {.lex_state = 455},
  [17917] = {.lex_state = 457, .external_lex_state = 131},
  [17918] = {.lex_state = 457, .external_lex_state = 101},
  [17919] = {.lex_state = 457, .external_lex_state = 131},
  [17920] = {.lex_state = 455, .external_lex_state = 133},
  [17921] = {.lex_state = 455},
  [17922] = {.lex_state = 457, .external_lex_state = 131},
  [17923] = {.lex_state = 457, .external_lex_state = 131},
  [17924] = {.lex_state = 455},
  [17925] = {.lex_state = 457, .external_lex_state = 131},
  [17926] = {.lex_state = 457, .external_lex_state = 131},
  [17927] = {.lex_state = 457, .external_lex_state = 131},
  [17928] = {.lex_state = 457, .external_lex_state = 131},
  [17929] = {.lex_state = 457, .external_lex_state = 131},
  [17930] = {.lex_state = 457, .external_lex_state = 131},
  [17931] = {.lex_state = 457, .external_lex_state = 132},
  [17932] = {.lex_state = 457, .external_lex_state = 131},
  [17933] = {.lex_state = 457, .external_lex_state = 131},
  [17934] = {.lex_state = 457, .external_lex_state = 131},
  [17935] = {.lex_state = 457, .external_lex_state = 131},
  [17936] = {.lex_state = 457, .external_lex_state = 131},
  [17937] = {.lex_state = 457, .external_lex_state = 131},
  [17938] = {.lex_state = 457, .external_lex_state = 131},
  [17939] = {.lex_state = 457, .external_lex_state = 131},
  [17940] = {.lex_state = 457, .external_lex_state = 114},
  [17941] = {.lex_state = 203},
  [17942] = {.lex_state = 455},
  [17943] = {.lex_state = 455},
  [17944] = {.lex_state = 457, .external_lex_state = 131},
  [17945] = {.lex_state = 457, .external_lex_state = 131},
  [17946] = {.lex_state = 457, .external_lex_state = 131},
  [17947] = {.lex_state = 457, .external_lex_state = 131},
  [17948] = {.lex_state = 457, .external_lex_state = 131},
  [17949] = {.lex_state = 457, .external_lex_state = 131},
  [17950] = {.lex_state = 457, .external_lex_state = 131},
  [17951] = {.lex_state = 457, .external_lex_state = 131},
  [17952] = {.lex_state = 457, .external_lex_state = 131},
  [17953] = {.lex_state = 457, .external_lex_state = 131},
  [17954] = {.lex_state = 457, .external_lex_state = 131},
  [17955] = {.lex_state = 457, .external_lex_state = 131},
  [17956] = {.lex_state = 457, .external_lex_state = 131},
  [17957] = {.lex_state = 457, .external_lex_state = 131},
  [17958] = {.lex_state = 457, .external_lex_state = 131},
  [17959] = {.lex_state = 457, .external_lex_state = 131},
  [17960] = {.lex_state = 457, .external_lex_state = 99},
  [17961] = {.lex_state = 457},
  [17962] = {.lex_state = 457, .external_lex_state = 121},
  [17963] = {.lex_state = 457, .external_lex_state = 132},
  [17964] = {.lex_state = 457, .external_lex_state = 131},
  [17965] = {.lex_state = 457, .external_lex_state = 131},
  [17966] = {.lex_state = 457, .external_lex_state = 131},
  [17967] = {.lex_state = 457, .external_lex_state = 131},
  [17968] = {.lex_state = 457, .external_lex_state = 131},
  [17969] = {.lex_state = 457, .external_lex_state = 131},
  [17970] = {.lex_state = 457, .external_lex_state = 131},
  [17971] = {.lex_state = 457, .external_lex_state = 131},
  [17972] = {.lex_state = 457, .external_lex_state = 131},
  [17973] = {.lex_state = 455, .external_lex_state = 120},
  [17974] = {.lex_state = 457, .external_lex_state = 131},
  [17975] = {.lex_state = 457, .external_lex_state = 101},
  [17976] = {.lex_state = 457, .external_lex_state = 131},
  [17977] = {.lex_state = 455},
  [17978] = {.lex_state = 457, .external_lex_state = 131},
  [17979] = {.lex_state = 457, .external_lex_state = 121},
  [17980] = {.lex_state = 457, .external_lex_state = 131},
  [17981] = {.lex_state = 455},
  [17982] = {.lex_state = 457, .external_lex_state = 131},
  [17983] = {.lex_state = 457, .external_lex_state = 131},
  [17984] = {.lex_state = 457, .external_lex_state = 131},
  [17985] = {.lex_state = 457, .external_lex_state = 131},
  [17986] = {.lex_state = 457, .external_lex_state = 121},
  [17987] = {.lex_state = 457, .external_lex_state = 131},
  [17988] = {.lex_state = 203},
  [17989] = {.lex_state = 457, .external_lex_state = 131},
  [17990] = {.lex_state = 455},
  [17991] = {.lex_state = 457, .external_lex_state = 131},
  [17992] = {.lex_state = 457, .external_lex_state = 131},
  [17993] = {.lex_state = 455, .external_lex_state = 120},
  [17994] = {.lex_state = 457, .external_lex_state = 131},
  [17995] = {.lex_state = 457, .external_lex_state = 131},
  [17996] = {.lex_state = 457, .external_lex_state = 101},
  [17997] = {.lex_state = 457},
  [17998] = {.lex_state = 457, .external_lex_state = 131},
  [17999] = {.lex_state = 455},
  [18000] = {.lex_state = 457, .external_lex_state = 132},
  [18001] = {.lex_state = 457},
  [18002] = {.lex_state = 457, .external_lex_state = 131},
  [18003] = {.lex_state = 455, .external_lex_state = 114},
  [18004] = {.lex_state = 457, .external_lex_state = 121},
  [18005] = {.lex_state = 457, .external_lex_state = 131},
  [18006] = {.lex_state = 455},
  [18007] = {.lex_state = 455},
  [18008] = {.lex_state = 203},
  [18009] = {.lex_state = 455},
  [18010] = {.lex_state = 457, .external_lex_state = 121},
  [18011] = {.lex_state = 457, .external_lex_state = 131},
  [18012] = {.lex_state = 457, .external_lex_state = 131},
  [18013] = {.lex_state = 457, .external_lex_state = 131},
  [18014] = {.lex_state = 457, .external_lex_state = 131},
  [18015] = {.lex_state = 457, .external_lex_state = 131},
  [18016] = {.lex_state = 457, .external_lex_state = 131},
  [18017] = {.lex_state = 457, .external_lex_state = 131},
  [18018] = {.lex_state = 457, .external_lex_state = 131},
  [18019] = {.lex_state = 457, .external_lex_state = 131},
  [18020] = {.lex_state = 455},
  [18021] = {.lex_state = 457, .external_lex_state = 131},
  [18022] = {.lex_state = 457, .external_lex_state = 131},
  [18023] = {.lex_state = 457, .external_lex_state = 131},
  [18024] = {.lex_state = 457, .external_lex_state = 131},
  [18025] = {.lex_state = 455, .external_lex_state = 120},
  [18026] = {.lex_state = 457, .external_lex_state = 131},
  [18027] = {.lex_state = 190},
  [18028] = {.lex_state = 457, .external_lex_state = 101},
  [18029] = {.lex_state = 457, .external_lex_state = 131},
  [18030] = {.lex_state = 457, .external_lex_state = 131},
  [18031] = {.lex_state = 457, .external_lex_state = 131},
  [18032] = {.lex_state = 457, .external_lex_state = 131},
  [18033] = {.lex_state = 457, .external_lex_state = 131},
  [18034] = {.lex_state = 457, .external_lex_state = 131},
  [18035] = {.lex_state = 457, .external_lex_state = 131},
  [18036] = {.lex_state = 457, .external_lex_state = 131},
  [18037] = {.lex_state = 455},
  [18038] = {.lex_state = 457, .external_lex_state = 131},
  [18039] = {.lex_state = 457, .external_lex_state = 131},
  [18040] = {.lex_state = 457, .external_lex_state = 131},
  [18041] = {.lex_state = 457, .external_lex_state = 131},
  [18042] = {.lex_state = 457, .external_lex_state = 131},
  [18043] = {.lex_state = 457, .external_lex_state = 131},
  [18044] = {.lex_state = 224, .external_lex_state = 109},
  [18045] = {.lex_state = 455},
  [18046] = {.lex_state = 457, .external_lex_state = 131},
  [18047] = {.lex_state = 457, .external_lex_state = 131},
  [18048] = {.lex_state = 457, .external_lex_state = 131},
  [18049] = {.lex_state = 457, .external_lex_state = 131},
  [18050] = {.lex_state = 457, .external_lex_state = 131},
  [18051] = {.lex_state = 455},
  [18052] = {.lex_state = 455},
  [18053] = {.lex_state = 455},
  [18054] = {.lex_state = 457, .external_lex_state = 131},
  [18055] = {.lex_state = 457, .external_lex_state = 131},
  [18056] = {.lex_state = 457, .external_lex_state = 131},
  [18057] = {.lex_state = 457, .external_lex_state = 131},
  [18058] = {.lex_state = 457, .external_lex_state = 99},
  [18059] = {.lex_state = 457, .external_lex_state = 131},
  [18060] = {.lex_state = 457, .external_lex_state = 99},
  [18061] = {.lex_state = 457, .external_lex_state = 131},
  [18062] = {.lex_state = 457, .external_lex_state = 131},
  [18063] = {.lex_state = 457, .external_lex_state = 121},
  [18064] = {.lex_state = 457, .external_lex_state = 131},
  [18065] = {.lex_state = 457, .external_lex_state = 132},
  [18066] = {.lex_state = 457, .external_lex_state = 131},
  [18067] = {.lex_state = 457, .external_lex_state = 99},
  [18068] = {.lex_state = 457, .external_lex_state = 131},
  [18069] = {.lex_state = 457, .external_lex_state = 121},
  [18070] = {.lex_state = 457, .external_lex_state = 99},
  [18071] = {.lex_state = 457, .external_lex_state = 99},
  [18072] = {.lex_state = 457, .external_lex_state = 99},
  [18073] = {.lex_state = 457, .external_lex_state = 131},
  [18074] = {.lex_state = 457, .external_lex_state = 99},
  [18075] = {.lex_state = 455},
  [18076] = {.lex_state = 457, .external_lex_state = 131},
  [18077] = {.lex_state = 457, .external_lex_state = 131},
  [18078] = {.lex_state = 457, .external_lex_state = 131},
  [18079] = {.lex_state = 457, .external_lex_state = 131},
  [18080] = {.lex_state = 457, .external_lex_state = 131},
  [18081] = {.lex_state = 457, .external_lex_state = 131},
  [18082] = {.lex_state = 457, .external_lex_state = 131},
  [18083] = {.lex_state = 457, .external_lex_state = 99},
  [18084] = {.lex_state = 457, .external_lex_state = 131},
  [18085] = {.lex_state = 457, .external_lex_state = 131},
  [18086] = {.lex_state = 457, .external_lex_state = 131},
  [18087] = {.lex_state = 457, .external_lex_state = 131},
  [18088] = {.lex_state = 457, .external_lex_state = 131},
  [18089] = {.lex_state = 457, .external_lex_state = 132},
  [18090] = {.lex_state = 457, .external_lex_state = 131},
  [18091] = {.lex_state = 457, .external_lex_state = 131},
  [18092] = {.lex_state = 457, .external_lex_state = 131},
  [18093] = {.lex_state = 455},
  [18094] = {.lex_state = 224, .external_lex_state = 129},
  [18095] = {.lex_state = 457, .external_lex_state = 131},
  [18096] = {.lex_state = 457, .external_lex_state = 131},
  [18097] = {.lex_state = 457, .external_lex_state = 131},
  [18098] = {.lex_state = 455},
  [18099] = {.lex_state = 457, .external_lex_state = 131},
  [18100] = {.lex_state = 457, .external_lex_state = 131},
  [18101] = {.lex_state = 455, .external_lex_state = 120},
  [18102] = {.lex_state = 457, .external_lex_state = 131},
  [18103] = {.lex_state = 457, .external_lex_state = 131},
  [18104] = {.lex_state = 457, .external_lex_state = 101},
  [18105] = {.lex_state = 457, .external_lex_state = 131},
  [18106] = {.lex_state = 457, .external_lex_state = 131},
  [18107] = {.lex_state = 457, .external_lex_state = 131},
  [18108] = {.lex_state = 455},
  [18109] = {.lex_state = 457, .external_lex_state = 131},
  [18110] = {.lex_state = 457, .external_lex_state = 131},
  [18111] = {.lex_state = 457, .external_lex_state = 131},
  [18112] = {.lex_state = 457, .external_lex_state = 121},
  [18113] = {.lex_state = 457, .external_lex_state = 121},
  [18114] = {.lex_state = 457, .external_lex_state = 131},
  [18115] = {.lex_state = 457, .external_lex_state = 131},
  [18116] = {.lex_state = 457, .external_lex_state = 131},
  [18117] = {.lex_state = 457, .external_lex_state = 131},
  [18118] = {.lex_state = 203},
  [18119] = {.lex_state = 457, .external_lex_state = 131},
  [18120] = {.lex_state = 457, .external_lex_state = 101},
  [18121] = {.lex_state = 457, .external_lex_state = 131},
  [18122] = {.lex_state = 457, .external_lex_state = 131},
  [18123] = {.lex_state = 455},
  [18124] = {.lex_state = 457, .external_lex_state = 131},
  [18125] = {.lex_state = 457, .external_lex_state = 131},
  [18126] = {.lex_state = 457, .external_lex_state = 99},
  [18127] = {.lex_state = 457, .external_lex_state = 99},
  [18128] = {.lex_state = 457, .external_lex_state = 131},
  [18129] = {.lex_state = 455},
  [18130] = {.lex_state = 190},
  [18131] = {.lex_state = 457, .external_lex_state = 131},
  [18132] = {.lex_state = 455, .external_lex_state = 133},
  [18133] = {.lex_state = 457, .external_lex_state = 131},
  [18134] = {.lex_state = 457, .external_lex_state = 131},
  [18135] = {.lex_state = 457, .external_lex_state = 131},
  [18136] = {.lex_state = 457, .external_lex_state = 131},
  [18137] = {.lex_state = 457, .external_lex_state = 131},
  [18138] = {.lex_state = 457, .external_lex_state = 99},
  [18139] = {.lex_state = 455, .external_lex_state = 133},
  [18140] = {.lex_state = 457, .external_lex_state = 131},
  [18141] = {.lex_state = 455, .external_lex_state = 122},
  [18142] = {.lex_state = 457, .external_lex_state = 131},
  [18143] = {.lex_state = 457, .external_lex_state = 131},
  [18144] = {.lex_state = 457, .external_lex_state = 132},
  [18145] = {.lex_state = 457, .external_lex_state = 131},
  [18146] = {.lex_state = 457, .external_lex_state = 131},
  [18147] = {.lex_state = 455, .external_lex_state = 133},
  [18148] = {.lex_state = 457, .external_lex_state = 131},
  [18149] = {.lex_state = 457, .external_lex_state = 131},
  [18150] = {.lex_state = 455, .external_lex_state = 133},
  [18151] = {.lex_state = 457, .external_lex_state = 99},
  [18152] = {.lex_state = 457, .external_lex_state = 131},
  [18153] = {.lex_state = 457, .external_lex_state = 131},
  [18154] = {.lex_state = 455},
  [18155] = {.lex_state = 457, .external_lex_state = 131},
  [18156] = {.lex_state = 224, .external_lex_state = 109},
  [18157] = {.lex_state = 203},
  [18158] = {.lex_state = 455},
  [18159] = {.lex_state = 455, .external_lex_state = 133},
  [18160] = {.lex_state = 457, .external_lex_state = 131},
  [18161] = {.lex_state = 457, .external_lex_state = 131},
  [18162] = {.lex_state = 457, .external_lex_state = 131},
  [18163] = {.lex_state = 457, .external_lex_state = 131},
  [18164] = {.lex_state = 457, .external_lex_state = 131},
  [18165] = {.lex_state = 457, .external_lex_state = 131},
  [18166] = {.lex_state = 457, .external_lex_state = 131},
  [18167] = {.lex_state = 457, .external_lex_state = 131},
  [18168] = {.lex_state = 455},
  [18169] = {.lex_state = 457, .external_lex_state = 131},
  [18170] = {.lex_state = 457, .external_lex_state = 131},
  [18171] = {.lex_state = 457, .external_lex_state = 131},
  [18172] = {.lex_state = 457, .external_lex_state = 131},
  [18173] = {.lex_state = 457, .external_lex_state = 131},
  [18174] = {.lex_state = 457, .external_lex_state = 131},
  [18175] = {.lex_state = 457, .external_lex_state = 131},
  [18176] = {.lex_state = 455},
  [18177] = {.lex_state = 457, .external_lex_state = 101},
  [18178] = {.lex_state = 457, .external_lex_state = 132},
  [18179] = {.lex_state = 455},
  [18180] = {.lex_state = 457, .external_lex_state = 131},
  [18181] = {.lex_state = 455},
  [18182] = {.lex_state = 457, .external_lex_state = 131},
  [18183] = {.lex_state = 457, .external_lex_state = 131},
  [18184] = {.lex_state = 457, .external_lex_state = 131},
  [18185] = {.lex_state = 457, .external_lex_state = 131},
  [18186] = {.lex_state = 455},
  [18187] = {.lex_state = 457, .external_lex_state = 131},
  [18188] = {.lex_state = 457, .external_lex_state = 131},
  [18189] = {.lex_state = 457, .external_lex_state = 131},
  [18190] = {.lex_state = 457, .external_lex_state = 121},
  [18191] = {.lex_state = 455, .external_lex_state = 133},
  [18192] = {.lex_state = 457, .external_lex_state = 121},
  [18193] = {.lex_state = 457, .external_lex_state = 131},
  [18194] = {.lex_state = 457, .external_lex_state = 114},
  [18195] = {.lex_state = 455, .external_lex_state = 133},
  [18196] = {.lex_state = 457, .external_lex_state = 131},
  [18197] = {.lex_state = 455},
  [18198] = {.lex_state = 455},
  [18199] = {.lex_state = 457, .external_lex_state = 131},
  [18200] = {.lex_state = 455, .external_lex_state = 100},
  [18201] = {.lex_state = 455, .external_lex_state = 100},
  [18202] = {.lex_state = 455, .external_lex_state = 133},
  [18203] = {.lex_state = 457, .external_lex_state = 132},
  [18204] = {.lex_state = 455},
  [18205] = {.lex_state = 457, .external_lex_state = 132},
  [18206] = {.lex_state = 455, .external_lex_state = 100},
  [18207] = {.lex_state = 457, .external_lex_state = 131},
  [18208] = {.lex_state = 455, .external_lex_state = 133},
  [18209] = {.lex_state = 457, .external_lex_state = 131},
  [18210] = {.lex_state = 457, .external_lex_state = 131},
  [18211] = {.lex_state = 455, .external_lex_state = 133},
  [18212] = {.lex_state = 455},
  [18213] = {.lex_state = 455, .external_lex_state = 120},
  [18214] = {.lex_state = 457, .external_lex_state = 131},
  [18215] = {.lex_state = 457, .external_lex_state = 131},
  [18216] = {.lex_state = 457, .external_lex_state = 101},
  [18217] = {.lex_state = 457, .external_lex_state = 131},
  [18218] = {.lex_state = 457, .external_lex_state = 131},
  [18219] = {.lex_state = 455, .external_lex_state = 133},
  [18220] = {.lex_state = 457, .external_lex_state = 131},
  [18221] = {.lex_state = 457, .external_lex_state = 131},
  [18222] = {.lex_state = 455},
  [18223] = {.lex_state = 457, .external_lex_state = 131},
  [18224] = {.lex_state = 455},
  [18225] = {.lex_state = 457, .external_lex_state = 131},
  [18226] = {.lex_state = 457, .external_lex_state = 131},
  [18227] = {.lex_state = 457, .external_lex_state = 131},
  [18228] = {.lex_state = 457, .external_lex_state = 99},
  [18229] = {.lex_state = 457, .external_lex_state = 131},
  [18230] = {.lex_state = 457, .external_lex_state = 131},
  [18231] = {.lex_state = 457, .external_lex_state = 131},
  [18232] = {.lex_state = 457, .external_lex_state = 131},
  [18233] = {.lex_state = 457, .external_lex_state = 131},
  [18234] = {.lex_state = 457, .external_lex_state = 121},
  [18235] = {.lex_state = 457, .external_lex_state = 131},
  [18236] = {.lex_state = 203},
  [18237] = {.lex_state = 455, .external_lex_state = 133},
  [18238] = {.lex_state = 457, .external_lex_state = 131},
  [18239] = {.lex_state = 455},
  [18240] = {.lex_state = 457, .external_lex_state = 131},
  [18241] = {.lex_state = 457, .external_lex_state = 131},
  [18242] = {.lex_state = 455},
  [18243] = {.lex_state = 457, .external_lex_state = 131},
  [18244] = {.lex_state = 455, .external_lex_state = 133},
  [18245] = {.lex_state = 457, .external_lex_state = 101},
  [18246] = {.lex_state = 203},
  [18247] = {.lex_state = 455},
  [18248] = {.lex_state = 224, .external_lex_state = 129},
  [18249] = {.lex_state = 457, .external_lex_state = 131},
  [18250] = {.lex_state = 455, .external_lex_state = 133},
  [18251] = {.lex_state = 457, .external_lex_state = 131},
  [18252] = {.lex_state = 457, .external_lex_state = 131},
  [18253] = {.lex_state = 455, .external_lex_state = 133},
  [18254] = {.lex_state = 457, .external_lex_state = 131},
  [18255] = {.lex_state = 457, .external_lex_state = 131},
  [18256] = {.lex_state = 457, .external_lex_state = 131},
  [18257] = {.lex_state = 457, .external_lex_state = 131},
  [18258] = {.lex_state = 457, .external_lex_state = 131},
  [18259] = {.lex_state = 457, .external_lex_state = 131},
  [18260] = {.lex_state = 457, .external_lex_state = 131},
  [18261] = {.lex_state = 455, .external_lex_state = 133},
  [18262] = {.lex_state = 455},
  [18263] = {.lex_state = 457, .external_lex_state = 131},
  [18264] = {.lex_state = 455, .external_lex_state = 120},
  [18265] = {.lex_state = 457, .external_lex_state = 131},
  [18266] = {.lex_state = 457, .external_lex_state = 131},
  [18267] = {.lex_state = 457, .external_lex_state = 131},
  [18268] = {.lex_state = 457, .external_lex_state = 131},
  [18269] = {.lex_state = 457, .external_lex_state = 131},
  [18270] = {.lex_state = 457, .external_lex_state = 131},
  [18271] = {.lex_state = 457, .external_lex_state = 131},
  [18272] = {.lex_state = 457, .external_lex_state = 131},
  [18273] = {.lex_state = 457, .external_lex_state = 131},
  [18274] = {.lex_state = 457, .external_lex_state = 101},
  [18275] = {.lex_state = 457, .external_lex_state = 131},
  [18276] = {.lex_state = 457, .external_lex_state = 131},
  [18277] = {.lex_state = 457, .external_lex_state = 131},
  [18278] = {.lex_state = 455, .external_lex_state = 133},
  [18279] = {.lex_state = 457, .external_lex_state = 131},
  [18280] = {.lex_state = 457, .external_lex_state = 131},
  [18281] = {.lex_state = 455},
  [18282] = {.lex_state = 457, .external_lex_state = 114},
  [18283] = {.lex_state = 455, .external_lex_state = 133},
  [18284] = {.lex_state = 457, .external_lex_state = 131},
  [18285] = {.lex_state = 457, .external_lex_state = 131},
  [18286] = {.lex_state = 457, .external_lex_state = 131},
  [18287] = {.lex_state = 457, .external_lex_state = 131},
  [18288] = {.lex_state = 455, .external_lex_state = 133},
  [18289] = {.lex_state = 457, .external_lex_state = 131},
  [18290] = {.lex_state = 457, .external_lex_state = 131},
  [18291] = {.lex_state = 455, .external_lex_state = 133},
  [18292] = {.lex_state = 457, .external_lex_state = 131},
  [18293] = {.lex_state = 455},
  [18294] = {.lex_state = 457, .external_lex_state = 114},
  [18295] = {.lex_state = 457, .external_lex_state = 131},
  [18296] = {.lex_state = 457, .external_lex_state = 121},
  [18297] = {.lex_state = 457, .external_lex_state = 131},
  [18298] = {.lex_state = 455, .external_lex_state = 133},
  [18299] = {.lex_state = 455},
  [18300] = {.lex_state = 457, .external_lex_state = 132},
  [18301] = {.lex_state = 457, .external_lex_state = 121},
  [18302] = {.lex_state = 457, .external_lex_state = 131},
  [18303] = {.lex_state = 457, .external_lex_state = 131},
  [18304] = {.lex_state = 203},
  [18305] = {.lex_state = 457, .external_lex_state = 131},
  [18306] = {.lex_state = 457, .external_lex_state = 131},
  [18307] = {.lex_state = 457, .external_lex_state = 131},
  [18308] = {.lex_state = 457, .external_lex_state = 131},
  [18309] = {.lex_state = 457, .external_lex_state = 131},
  [18310] = {.lex_state = 457, .external_lex_state = 131},
  [18311] = {.lex_state = 457, .external_lex_state = 131},
  [18312] = {.lex_state = 455},
  [18313] = {.lex_state = 457, .external_lex_state = 131},
  [18314] = {.lex_state = 457, .external_lex_state = 131},
  [18315] = {.lex_state = 455, .external_lex_state = 133},
  [18316] = {.lex_state = 457, .external_lex_state = 132},
  [18317] = {.lex_state = 203},
  [18318] = {.lex_state = 457, .external_lex_state = 131},
  [18319] = {.lex_state = 455, .external_lex_state = 133},
  [18320] = {.lex_state = 455},
  [18321] = {.lex_state = 455, .external_lex_state = 120},
  [18322] = {.lex_state = 457, .external_lex_state = 131},
  [18323] = {.lex_state = 455, .external_lex_state = 133},
  [18324] = {.lex_state = 457, .external_lex_state = 131},
  [18325] = {.lex_state = 455, .external_lex_state = 133},
  [18326] = {.lex_state = 457, .external_lex_state = 101},
  [18327] = {.lex_state = 457, .external_lex_state = 131},
  [18328] = {.lex_state = 457, .external_lex_state = 131},
  [18329] = {.lex_state = 457, .external_lex_state = 121},
  [18330] = {.lex_state = 455},
  [18331] = {.lex_state = 457, .external_lex_state = 101},
  [18332] = {.lex_state = 455},
  [18333] = {.lex_state = 455, .external_lex_state = 133},
  [18334] = {.lex_state = 457, .external_lex_state = 131},
  [18335] = {.lex_state = 457, .external_lex_state = 121},
  [18336] = {.lex_state = 455},
  [18337] = {.lex_state = 457, .external_lex_state = 131},
  [18338] = {.lex_state = 203},
  [18339] = {.lex_state = 457, .external_lex_state = 131},
  [18340] = {.lex_state = 457, .external_lex_state = 131},
  [18341] = {.lex_state = 457, .external_lex_state = 131},
  [18342] = {.lex_state = 457, .external_lex_state = 101},
  [18343] = {.lex_state = 457, .external_lex_state = 131},
  [18344] = {.lex_state = 455},
  [18345] = {.lex_state = 455},
  [18346] = {.lex_state = 457, .external_lex_state = 131},
  [18347] = {.lex_state = 455, .external_lex_state = 133},
  [18348] = {.lex_state = 457, .external_lex_state = 131},
  [18349] = {.lex_state = 457, .external_lex_state = 131},
  [18350] = {.lex_state = 457, .external_lex_state = 131},
  [18351] = {.lex_state = 457, .external_lex_state = 131},
  [18352] = {.lex_state = 457, .external_lex_state = 131},
  [18353] = {.lex_state = 455, .external_lex_state = 133},
  [18354] = {.lex_state = 457, .external_lex_state = 131},
  [18355] = {.lex_state = 455},
  [18356] = {.lex_state = 457, .external_lex_state = 131},
  [18357] = {.lex_state = 457, .external_lex_state = 131},
  [18358] = {.lex_state = 455, .external_lex_state = 133},
  [18359] = {.lex_state = 457, .external_lex_state = 131},
  [18360] = {.lex_state = 455, .external_lex_state = 133},
  [18361] = {.lex_state = 457, .external_lex_state = 131},
  [18362] = {.lex_state = 457, .external_lex_state = 131},
  [18363] = {.lex_state = 457, .external_lex_state = 131},
  [18364] = {.lex_state = 455},
  [18365] = {.lex_state = 457, .external_lex_state = 131},
  [18366] = {.lex_state = 457, .external_lex_state = 131},
  [18367] = {.lex_state = 457, .external_lex_state = 131},
  [18368] = {.lex_state = 455, .external_lex_state = 133},
  [18369] = {.lex_state = 457, .external_lex_state = 114},
  [18370] = {.lex_state = 457, .external_lex_state = 131},
  [18371] = {.lex_state = 457, .external_lex_state = 99},
  [18372] = {.lex_state = 455},
  [18373] = {.lex_state = 457, .external_lex_state = 131},
  [18374] = {.lex_state = 457, .external_lex_state = 131},
  [18375] = {.lex_state = 457, .external_lex_state = 131},
  [18376] = {.lex_state = 457, .external_lex_state = 131},
  [18377] = {.lex_state = 457, .external_lex_state = 131},
  [18378] = {.lex_state = 457, .external_lex_state = 131},
  [18379] = {.lex_state = 457, .external_lex_state = 131},
  [18380] = {.lex_state = 457, .external_lex_state = 132},
  [18381] = {.lex_state = 457, .external_lex_state = 131},
  [18382] = {.lex_state = 457, .external_lex_state = 131},
  [18383] = {.lex_state = 457, .external_lex_state = 131},
  [18384] = {.lex_state = 455, .external_lex_state = 133},
  [18385] = {.lex_state = 455},
  [18386] = {.lex_state = 457, .external_lex_state = 131},
  [18387] = {.lex_state = 457, .external_lex_state = 131},
  [18388] = {.lex_state = 455},
  [18389] = {.lex_state = 455},
  [18390] = {.lex_state = 455, .external_lex_state = 133},
  [18391] = {.lex_state = 457, .external_lex_state = 131},
  [18392] = {.lex_state = 457, .external_lex_state = 131},
  [18393] = {.lex_state = 457, .external_lex_state = 131},
  [18394] = {.lex_state = 457, .external_lex_state = 99},
  [18395] = {.lex_state = 455, .external_lex_state = 133},
  [18396] = {.lex_state = 457, .external_lex_state = 131},
  [18397] = {.lex_state = 455, .external_lex_state = 133},
  [18398] = {.lex_state = 457, .external_lex_state = 131},
  [18399] = {.lex_state = 457, .external_lex_state = 131},
  [18400] = {.lex_state = 457, .external_lex_state = 131},
  [18401] = {.lex_state = 457, .external_lex_state = 131},
  [18402] = {.lex_state = 457, .external_lex_state = 131},
  [18403] = {.lex_state = 457, .external_lex_state = 131},
  [18404] = {.lex_state = 457, .external_lex_state = 131},
  [18405] = {.lex_state = 455, .external_lex_state = 133},
  [18406] = {.lex_state = 457, .external_lex_state = 132},
  [18407] = {.lex_state = 457, .external_lex_state = 131},
  [18408] = {.lex_state = 457, .external_lex_state = 131},
  [18409] = {.lex_state = 457, .external_lex_state = 131},
  [18410] = {.lex_state = 457, .external_lex_state = 131},
  [18411] = {.lex_state = 455, .external_lex_state = 120},
  [18412] = {.lex_state = 457, .external_lex_state = 131},
  [18413] = {.lex_state = 457, .external_lex_state = 131},
  [18414] = {.lex_state = 457, .external_lex_state = 99},
  [18415] = {.lex_state = 457, .external_lex_state = 99},
  [18416] = {.lex_state = 457, .external_lex_state = 131},
  [18417] = {.lex_state = 457, .external_lex_state = 131},
  [18418] = {.lex_state = 457, .external_lex_state = 101},
  [18419] = {.lex_state = 457, .external_lex_state = 131},
  [18420] = {.lex_state = 457, .external_lex_state = 131},
  [18421] = {.lex_state = 457, .external_lex_state = 131},
  [18422] = {.lex_state = 455},
  [18423] = {.lex_state = 455},
  [18424] = {.lex_state = 457, .external_lex_state = 131},
  [18425] = {.lex_state = 457, .external_lex_state = 131},
  [18426] = {.lex_state = 457, .external_lex_state = 121},
  [18427] = {.lex_state = 457, .external_lex_state = 131},
  [18428] = {.lex_state = 457, .external_lex_state = 131},
  [18429] = {.lex_state = 203},
  [18430] = {.lex_state = 457, .external_lex_state = 131},
  [18431] = {.lex_state = 457, .external_lex_state = 101},
  [18432] = {.lex_state = 455},
  [18433] = {.lex_state = 457, .external_lex_state = 131},
  [18434] = {.lex_state = 457, .external_lex_state = 131},
  [18435] = {.lex_state = 457, .external_lex_state = 131},
  [18436] = {.lex_state = 457, .external_lex_state = 131},
  [18437] = {.lex_state = 455},
  [18438] = {.lex_state = 455},
  [18439] = {.lex_state = 457, .external_lex_state = 131},
  [18440] = {.lex_state = 457, .external_lex_state = 131},
  [18441] = {.lex_state = 457, .external_lex_state = 131},
  [18442] = {.lex_state = 457, .external_lex_state = 132},
  [18443] = {.lex_state = 457, .external_lex_state = 131},
  [18444] = {.lex_state = 457, .external_lex_state = 131},
  [18445] = {.lex_state = 455},
  [18446] = {.lex_state = 457, .external_lex_state = 131},
  [18447] = {.lex_state = 455, .external_lex_state = 120},
  [18448] = {.lex_state = 457, .external_lex_state = 131},
  [18449] = {.lex_state = 457, .external_lex_state = 131},
  [18450] = {.lex_state = 457, .external_lex_state = 132},
  [18451] = {.lex_state = 457, .external_lex_state = 132},
  [18452] = {.lex_state = 457, .external_lex_state = 131},
  [18453] = {.lex_state = 457, .external_lex_state = 131},
  [18454] = {.lex_state = 457, .external_lex_state = 99},
  [18455] = {.lex_state = 457, .external_lex_state = 131},
  [18456] = {.lex_state = 457, .external_lex_state = 131},
  [18457] = {.lex_state = 457, .external_lex_state = 99},
  [18458] = {.lex_state = 457, .external_lex_state = 131},
  [18459] = {.lex_state = 457, .external_lex_state = 131},
  [18460] = {.lex_state = 457, .external_lex_state = 131},
  [18461] = {.lex_state = 457, .external_lex_state = 131},
  [18462] = {.lex_state = 457, .external_lex_state = 131},
  [18463] = {.lex_state = 457, .external_lex_state = 131},
  [18464] = {.lex_state = 457, .external_lex_state = 131},
  [18465] = {.lex_state = 457, .external_lex_state = 99},
  [18466] = {.lex_state = 457, .external_lex_state = 132},
  [18467] = {.lex_state = 455},
  [18468] = {.lex_state = 457, .external_lex_state = 131},
  [18469] = {.lex_state = 457, .external_lex_state = 99},
  [18470] = {.lex_state = 457, .external_lex_state = 131},
  [18471] = {.lex_state = 455},
  [18472] = {.lex_state = 455, .external_lex_state = 120},
  [18473] = {.lex_state = 457, .external_lex_state = 101},
  [18474] = {.lex_state = 455},
  [18475] = {.lex_state = 457, .external_lex_state = 131},
  [18476] = {.lex_state = 457, .external_lex_state = 131},
  [18477] = {.lex_state = 455},
  [18478] = {.lex_state = 457, .external_lex_state = 131},
  [18479] = {.lex_state = 203},
  [18480] = {.lex_state = 457, .external_lex_state = 131},
  [18481] = {.lex_state = 457, .external_lex_state = 131},
  [18482] = {.lex_state = 457, .external_lex_state = 132},
  [18483] = {.lex_state = 457, .external_lex_state = 131},
  [18484] = {.lex_state = 457, .external_lex_state = 122},
  [18485] = {.lex_state = 457, .external_lex_state = 131},
  [18486] = {.lex_state = 457, .external_lex_state = 121},
  [18487] = {.lex_state = 457, .external_lex_state = 131},
  [18488] = {.lex_state = 457, .external_lex_state = 131},
  [18489] = {.lex_state = 457, .external_lex_state = 131},
  [18490] = {.lex_state = 457, .external_lex_state = 131},
  [18491] = {.lex_state = 457, .external_lex_state = 131},
  [18492] = {.lex_state = 457, .external_lex_state = 131},
  [18493] = {.lex_state = 457, .external_lex_state = 101},
  [18494] = {.lex_state = 457, .external_lex_state = 121},
  [18495] = {.lex_state = 457, .external_lex_state = 131},
  [18496] = {.lex_state = 457, .external_lex_state = 131},
  [18497] = {.lex_state = 455},
  [18498] = {.lex_state = 457, .external_lex_state = 131},
  [18499] = {.lex_state = 457, .external_lex_state = 131},
  [18500] = {.lex_state = 457, .external_lex_state = 131},
  [18501] = {.lex_state = 457, .external_lex_state = 121},
  [18502] = {.lex_state = 457, .external_lex_state = 131},
  [18503] = {.lex_state = 457, .external_lex_state = 131},
  [18504] = {.lex_state = 457, .external_lex_state = 131},
  [18505] = {.lex_state = 457, .external_lex_state = 131},
  [18506] = {.lex_state = 457, .external_lex_state = 131},
  [18507] = {.lex_state = 457, .external_lex_state = 131},
  [18508] = {.lex_state = 457, .external_lex_state = 131},
  [18509] = {.lex_state = 457, .external_lex_state = 131},
  [18510] = {.lex_state = 457, .external_lex_state = 131},
  [18511] = {.lex_state = 455, .external_lex_state = 120},
  [18512] = {.lex_state = 457, .external_lex_state = 131},
  [18513] = {.lex_state = 457, .external_lex_state = 131},
  [18514] = {.lex_state = 457, .external_lex_state = 131},
  [18515] = {.lex_state = 457, .external_lex_state = 131},
  [18516] = {.lex_state = 457, .external_lex_state = 131},
  [18517] = {.lex_state = 457, .external_lex_state = 131},
  [18518] = {.lex_state = 457, .external_lex_state = 131},
  [18519] = {.lex_state = 457, .external_lex_state = 131},
  [18520] = {.lex_state = 457, .external_lex_state = 101},
  [18521] = {.lex_state = 457, .external_lex_state = 132},
  [18522] = {.lex_state = 457, .external_lex_state = 131},
  [18523] = {.lex_state = 455},
  [18524] = {.lex_state = 455},
  [18525] = {.lex_state = 455},
  [18526] = {.lex_state = 457, .external_lex_state = 131},
  [18527] = {.lex_state = 457, .external_lex_state = 131},
  [18528] = {.lex_state = 457, .external_lex_state = 114},
  [18529] = {.lex_state = 457},
  [18530] = {.lex_state = 455, .external_lex_state = 114},
  [18531] = {.lex_state = 457, .external_lex_state = 131},
  [18532] = {.lex_state = 455},
  [18533] = {.lex_state = 457, .external_lex_state = 114},
  [18534] = {.lex_state = 457, .external_lex_state = 131},
  [18535] = {.lex_state = 457, .external_lex_state = 131},
  [18536] = {.lex_state = 457, .external_lex_state = 131},
  [18537] = {.lex_state = 457, .external_lex_state = 131},
  [18538] = {.lex_state = 203},
  [18539] = {.lex_state = 457, .external_lex_state = 132},
  [18540] = {.lex_state = 457, .external_lex_state = 131},
  [18541] = {.lex_state = 455},
  [18542] = {.lex_state = 457, .external_lex_state = 131},
  [18543] = {.lex_state = 224, .external_lex_state = 129},
  [18544] = {.lex_state = 457, .external_lex_state = 131},
  [18545] = {.lex_state = 457, .external_lex_state = 99},
  [18546] = {.lex_state = 455},
  [18547] = {.lex_state = 457, .external_lex_state = 131},
  [18548] = {.lex_state = 457, .external_lex_state = 131},
  [18549] = {.lex_state = 203},
  [18550] = {.lex_state = 457, .external_lex_state = 131},
  [18551] = {.lex_state = 457, .external_lex_state = 131},
  [18552] = {.lex_state = 457, .external_lex_state = 131},
  [18553] = {.lex_state = 455},
  [18554] = {.lex_state = 455},
  [18555] = {.lex_state = 457, .external_lex_state = 131},
  [18556] = {.lex_state = 455, .external_lex_state = 120},
  [18557] = {.lex_state = 457, .external_lex_state = 131},
  [18558] = {.lex_state = 457, .external_lex_state = 131},
  [18559] = {.lex_state = 203},
  [18560] = {.lex_state = 457, .external_lex_state = 131},
  [18561] = {.lex_state = 457, .external_lex_state = 101},
  [18562] = {.lex_state = 457, .external_lex_state = 131},
  [18563] = {.lex_state = 457, .external_lex_state = 131},
  [18564] = {.lex_state = 457, .external_lex_state = 101},
  [18565] = {.lex_state = 455},
  [18566] = {.lex_state = 457, .external_lex_state = 131},
  [18567] = {.lex_state = 457, .external_lex_state = 131},
  [18568] = {.lex_state = 457, .external_lex_state = 131},
  [18569] = {.lex_state = 457, .external_lex_state = 131},
  [18570] = {.lex_state = 457, .external_lex_state = 131},
  [18571] = {.lex_state = 455},
  [18572] = {.lex_state = 457, .external_lex_state = 131},
  [18573] = {.lex_state = 457, .external_lex_state = 131},
  [18574] = {.lex_state = 457, .external_lex_state = 131},
  [18575] = {.lex_state = 457, .external_lex_state = 131},
  [18576] = {.lex_state = 457, .external_lex_state = 131},
  [18577] = {.lex_state = 457, .external_lex_state = 131},
  [18578] = {.lex_state = 457, .external_lex_state = 131},
  [18579] = {.lex_state = 457, .external_lex_state = 131},
  [18580] = {.lex_state = 457, .external_lex_state = 131},
  [18581] = {.lex_state = 457, .external_lex_state = 131},
  [18582] = {.lex_state = 457, .external_lex_state = 131},
  [18583] = {.lex_state = 457, .external_lex_state = 131},
  [18584] = {.lex_state = 457, .external_lex_state = 131},
  [18585] = {.lex_state = 457, .external_lex_state = 131},
  [18586] = {.lex_state = 455},
  [18587] = {.lex_state = 457, .external_lex_state = 131},
  [18588] = {.lex_state = 457, .external_lex_state = 131},
  [18589] = {.lex_state = 457, .external_lex_state = 131},
  [18590] = {.lex_state = 457, .external_lex_state = 131},
  [18591] = {.lex_state = 455},
  [18592] = {.lex_state = 457, .external_lex_state = 131},
  [18593] = {.lex_state = 457, .external_lex_state = 131},
  [18594] = {.lex_state = 457, .external_lex_state = 131},
  [18595] = {.lex_state = 455},
  [18596] = {.lex_state = 455},
  [18597] = {.lex_state = 457, .external_lex_state = 101},
  [18598] = {.lex_state = 455},
  [18599] = {.lex_state = 457, .external_lex_state = 131},
  [18600] = {.lex_state = 455},
  [18601] = {.lex_state = 457, .external_lex_state = 132},
  [18602] = {.lex_state = 457, .external_lex_state = 131},
  [18603] = {.lex_state = 455, .external_lex_state = 122},
  [18604] = {.lex_state = 457, .external_lex_state = 131},
  [18605] = {.lex_state = 457, .external_lex_state = 131},
  [18606] = {.lex_state = 457, .external_lex_state = 131},
  [18607] = {.lex_state = 457, .external_lex_state = 132},
  [18608] = {.lex_state = 455},
  [18609] = {.lex_state = 457, .external_lex_state = 131},
  [18610] = {.lex_state = 455},
  [18611] = {.lex_state = 455},
  [18612] = {.lex_state = 457, .external_lex_state = 131},
  [18613] = {.lex_state = 457, .external_lex_state = 131},
  [18614] = {.lex_state = 455},
  [18615] = {.lex_state = 457, .external_lex_state = 131},
  [18616] = {.lex_state = 457, .external_lex_state = 114},
  [18617] = {.lex_state = 457, .external_lex_state = 131},
  [18618] = {.lex_state = 457, .external_lex_state = 131},
  [18619] = {.lex_state = 457, .external_lex_state = 131},
  [18620] = {.lex_state = 457, .external_lex_state = 131},
  [18621] = {.lex_state = 457, .external_lex_state = 131},
  [18622] = {.lex_state = 455},
  [18623] = {.lex_state = 457, .external_lex_state = 131},
  [18624] = {.lex_state = 457, .external_lex_state = 114},
  [18625] = {.lex_state = 190},
  [18626] = {.lex_state = 457, .external_lex_state = 131},
  [18627] = {.lex_state = 457, .external_lex_state = 131},
  [18628] = {.lex_state = 457, .external_lex_state = 131},
  [18629] = {.lex_state = 457, .external_lex_state = 131},
  [18630] = {.lex_state = 457, .external_lex_state = 131},
  [18631] = {.lex_state = 457, .external_lex_state = 131},
  [18632] = {.lex_state = 457, .external_lex_state = 131},
  [18633] = {.lex_state = 457, .external_lex_state = 131},
  [18634] = {.lex_state = 457, .external_lex_state = 132},
  [18635] = {.lex_state = 457, .external_lex_state = 131},
  [18636] = {.lex_state = 457, .external_lex_state = 131},
  [18637] = {.lex_state = 455},
  [18638] = {.lex_state = 457, .external_lex_state = 131},
  [18639] = {.lex_state = 457, .external_lex_state = 131},
  [18640] = {.lex_state = 457, .external_lex_state = 131},
  [18641] = {.lex_state = 457, .external_lex_state = 131},
  [18642] = {.lex_state = 455},
  [18643] = {.lex_state = 457, .external_lex_state = 131},
  [18644] = {.lex_state = 457, .external_lex_state = 131},
  [18645] = {.lex_state = 457, .external_lex_state = 131},
  [18646] = {.lex_state = 457, .external_lex_state = 99},
  [18647] = {.lex_state = 455, .external_lex_state = 120},
  [18648] = {.lex_state = 457, .external_lex_state = 101},
  [18649] = {.lex_state = 457, .external_lex_state = 131},
  [18650] = {.lex_state = 457, .external_lex_state = 131},
  [18651] = {.lex_state = 457, .external_lex_state = 131},
  [18652] = {.lex_state = 457, .external_lex_state = 131},
  [18653] = {.lex_state = 457, .external_lex_state = 131},
  [18654] = {.lex_state = 457, .external_lex_state = 131},
  [18655] = {.lex_state = 457, .external_lex_state = 131},
  [18656] = {.lex_state = 457, .external_lex_state = 131},
  [18657] = {.lex_state = 457, .external_lex_state = 131},
  [18658] = {.lex_state = 457, .external_lex_state = 131},
  [18659] = {.lex_state = 457, .external_lex_state = 131},
  [18660] = {.lex_state = 457, .external_lex_state = 131},
  [18661] = {.lex_state = 457, .external_lex_state = 99},
  [18662] = {.lex_state = 455},
  [18663] = {.lex_state = 455},
  [18664] = {.lex_state = 457, .external_lex_state = 131},
  [18665] = {.lex_state = 457, .external_lex_state = 131},
  [18666] = {.lex_state = 457, .external_lex_state = 131},
  [18667] = {.lex_state = 457, .external_lex_state = 131},
  [18668] = {.lex_state = 457, .external_lex_state = 131},
  [18669] = {.lex_state = 455},
  [18670] = {.lex_state = 457, .external_lex_state = 131},
  [18671] = {.lex_state = 457, .external_lex_state = 131},
  [18672] = {.lex_state = 455},
  [18673] = {.lex_state = 455},
  [18674] = {.lex_state = 457, .external_lex_state = 114},
  [18675] = {.lex_state = 455},
  [18676] = {.lex_state = 457, .external_lex_state = 131},
  [18677] = {.lex_state = 457, .external_lex_state = 131},
  [18678] = {.lex_state = 457, .external_lex_state = 131},
  [18679] = {.lex_state = 457, .external_lex_state = 131},
  [18680] = {.lex_state = 457, .external_lex_state = 131},
  [18681] = {.lex_state = 455},
  [18682] = {.lex_state = 457},
  [18683] = {.lex_state = 457, .external_lex_state = 131},
  [18684] = {.lex_state = 457, .external_lex_state = 99},
  [18685] = {.lex_state = 457, .external_lex_state = 131},
  [18686] = {.lex_state = 455},
  [18687] = {.lex_state = 457, .external_lex_state = 101},
  [18688] = {.lex_state = 457, .external_lex_state = 131},
  [18689] = {.lex_state = 457, .external_lex_state = 131},
  [18690] = {.lex_state = 457, .external_lex_state = 114},
  [18691] = {.lex_state = 455},
  [18692] = {.lex_state = 457, .external_lex_state = 131},
  [18693] = {.lex_state = 455},
  [18694] = {.lex_state = 457, .external_lex_state = 132},
  [18695] = {.lex_state = 224, .external_lex_state = 129},
  [18696] = {.lex_state = 455},
  [18697] = {.lex_state = 457, .external_lex_state = 131},
  [18698] = {.lex_state = 457, .external_lex_state = 131},
  [18699] = {.lex_state = 455},
  [18700] = {.lex_state = 457, .external_lex_state = 132},
  [18701] = {.lex_state = 455},
  [18702] = {.lex_state = 455},
  [18703] = {.lex_state = 457, .external_lex_state = 131},
  [18704] = {.lex_state = 457, .external_lex_state = 131},
  [18705] = {.lex_state = 457, .external_lex_state = 131},
  [18706] = {.lex_state = 457, .external_lex_state = 114},
  [18707] = {.lex_state = 455},
  [18708] = {.lex_state = 455},
  [18709] = {.lex_state = 457, .external_lex_state = 131},
  [18710] = {.lex_state = 457, .external_lex_state = 131},
  [18711] = {.lex_state = 457},
  [18712] = {.lex_state = 457, .external_lex_state = 131},
  [18713] = {.lex_state = 457, .external_lex_state = 131},
  [18714] = {.lex_state = 457, .external_lex_state = 131},
  [18715] = {.lex_state = 455},
  [18716] = {.lex_state = 455},
  [18717] = {.lex_state = 457, .external_lex_state = 131},
  [18718] = {.lex_state = 457, .external_lex_state = 131},
  [18719] = {.lex_state = 457, .external_lex_state = 131},
  [18720] = {.lex_state = 457, .external_lex_state = 114},
  [18721] = {.lex_state = 455},
  [18722] = {.lex_state = 455},
  [18723] = {.lex_state = 457, .external_lex_state = 131},
  [18724] = {.lex_state = 457, .external_lex_state = 131},
  [18725] = {.lex_state = 457, .external_lex_state = 131},
  [18726] = {.lex_state = 457},
  [18727] = {.lex_state = 457, .external_lex_state = 131},
  [18728] = {.lex_state = 455},
  [18729] = {.lex_state = 457, .external_lex_state = 114},
  [18730] = {.lex_state = 457, .external_lex_state = 121},
  [18731] = {.lex_state = 457, .external_lex_state = 131},
  [18732] = {.lex_state = 457, .external_lex_state = 131},
  [18733] = {.lex_state = 457, .external_lex_state = 132},
  [18734] = {.lex_state = 457, .external_lex_state = 131},
  [18735] = {.lex_state = 455},
  [18736] = {.lex_state = 457, .external_lex_state = 121},
  [18737] = {.lex_state = 457, .external_lex_state = 101},
  [18738] = {.lex_state = 455, .external_lex_state = 120},
  [18739] = {.lex_state = 457, .external_lex_state = 101},
  [18740] = {.lex_state = 457, .external_lex_state = 99},
  [18741] = {.lex_state = 455},
  [18742] = {.lex_state = 138},
  [18743] = {.lex_state = 455},
  [18744] = {.lex_state = 455},
  [18745] = {.lex_state = 455},
  [18746] = {.lex_state = 455},
  [18747] = {.lex_state = 457, .external_lex_state = 131},
  [18748] = {.lex_state = 457, .external_lex_state = 131},
  [18749] = {.lex_state = 455},
  [18750] = {.lex_state = 457, .external_lex_state = 131},
  [18751] = {.lex_state = 457, .external_lex_state = 131},
  [18752] = {.lex_state = 457, .external_lex_state = 131},
  [18753] = {.lex_state = 457, .external_lex_state = 131},
  [18754] = {.lex_state = 457, .external_lex_state = 131},
  [18755] = {.lex_state = 457, .external_lex_state = 131},
  [18756] = {.lex_state = 455},
  [18757] = {.lex_state = 457, .external_lex_state = 101},
  [18758] = {.lex_state = 455},
  [18759] = {.lex_state = 457, .external_lex_state = 131},
  [18760] = {.lex_state = 457, .external_lex_state = 132},
  [18761] = {.lex_state = 457, .external_lex_state = 131},
  [18762] = {.lex_state = 457, .external_lex_state = 131},
  [18763] = {.lex_state = 457, .external_lex_state = 131},
  [18764] = {.lex_state = 457, .external_lex_state = 131},
  [18765] = {.lex_state = 457, .external_lex_state = 131},
  [18766] = {.lex_state = 203},
  [18767] = {.lex_state = 457, .external_lex_state = 131},
  [18768] = {.lex_state = 457, .external_lex_state = 121},
  [18769] = {.lex_state = 457, .external_lex_state = 131},
  [18770] = {.lex_state = 457, .external_lex_state = 131},
  [18771] = {.lex_state = 457, .external_lex_state = 131},
  [18772] = {.lex_state = 457, .external_lex_state = 132},
  [18773] = {.lex_state = 224, .external_lex_state = 129},
  [18774] = {.lex_state = 457, .external_lex_state = 131},
  [18775] = {.lex_state = 457},
  [18776] = {.lex_state = 457},
  [18777] = {.lex_state = 457, .external_lex_state = 131},
  [18778] = {.lex_state = 457, .external_lex_state = 132},
  [18779] = {.lex_state = 457, .external_lex_state = 131},
  [18780] = {.lex_state = 457, .external_lex_state = 99},
  [18781] = {.lex_state = 457, .external_lex_state = 131},
  [18782] = {.lex_state = 455, .external_lex_state = 114},
  [18783] = {.lex_state = 457, .external_lex_state = 131},
  [18784] = {.lex_state = 457, .external_lex_state = 131},
  [18785] = {.lex_state = 457, .external_lex_state = 132},
  [18786] = {.lex_state = 457, .external_lex_state = 131},
  [18787] = {.lex_state = 457, .external_lex_state = 131},
  [18788] = {.lex_state = 457, .external_lex_state = 131},
  [18789] = {.lex_state = 457, .external_lex_state = 131},
  [18790] = {.lex_state = 457, .external_lex_state = 131},
  [18791] = {.lex_state = 455},
  [18792] = {.lex_state = 457, .external_lex_state = 99},
  [18793] = {.lex_state = 455},
  [18794] = {.lex_state = 457, .external_lex_state = 131},
  [18795] = {.lex_state = 455, .external_lex_state = 133},
  [18796] = {.lex_state = 455, .external_lex_state = 120},
  [18797] = {.lex_state = 457, .external_lex_state = 131},
  [18798] = {.lex_state = 457, .external_lex_state = 131},
  [18799] = {.lex_state = 457, .external_lex_state = 131},
  [18800] = {.lex_state = 457, .external_lex_state = 131},
  [18801] = {.lex_state = 457, .external_lex_state = 131},
  [18802] = {.lex_state = 457, .external_lex_state = 131},
  [18803] = {.lex_state = 455, .external_lex_state = 120},
  [18804] = {.lex_state = 457, .external_lex_state = 101},
  [18805] = {.lex_state = 457, .external_lex_state = 131},
  [18806] = {.lex_state = 455, .external_lex_state = 120},
  [18807] = {.lex_state = 455},
  [18808] = {.lex_state = 457, .external_lex_state = 131},
  [18809] = {.lex_state = 457},
  [18810] = {.lex_state = 457, .external_lex_state = 131},
  [18811] = {.lex_state = 457, .external_lex_state = 101},
  [18812] = {.lex_state = 457, .external_lex_state = 99},
  [18813] = {.lex_state = 224, .external_lex_state = 129},
  [18814] = {.lex_state = 457, .external_lex_state = 131},
  [18815] = {.lex_state = 457},
  [18816] = {.lex_state = 457, .external_lex_state = 101},
  [18817] = {.lex_state = 457, .external_lex_state = 131},
  [18818] = {.lex_state = 455},
  [18819] = {.lex_state = 457, .external_lex_state = 131},
  [18820] = {.lex_state = 457, .external_lex_state = 131},
  [18821] = {.lex_state = 457, .external_lex_state = 131},
  [18822] = {.lex_state = 455},
  [18823] = {.lex_state = 457, .external_lex_state = 131},
  [18824] = {.lex_state = 457, .external_lex_state = 131},
  [18825] = {.lex_state = 457, .external_lex_state = 132},
  [18826] = {.lex_state = 457, .external_lex_state = 114},
  [18827] = {.lex_state = 457, .external_lex_state = 131},
  [18828] = {.lex_state = 457, .external_lex_state = 131},
  [18829] = {.lex_state = 457, .external_lex_state = 132},
  [18830] = {.lex_state = 457, .external_lex_state = 131},
  [18831] = {.lex_state = 457, .external_lex_state = 131},
  [18832] = {.lex_state = 457, .external_lex_state = 131},
  [18833] = {.lex_state = 457, .external_lex_state = 131},
  [18834] = {.lex_state = 457, .external_lex_state = 121},
  [18835] = {.lex_state = 457, .external_lex_state = 131},
  [18836] = {.lex_state = 457, .external_lex_state = 131},
  [18837] = {.lex_state = 457, .external_lex_state = 131},
  [18838] = {.lex_state = 457, .external_lex_state = 131},
  [18839] = {.lex_state = 457, .external_lex_state = 131},
  [18840] = {.lex_state = 457, .external_lex_state = 131},
  [18841] = {.lex_state = 457, .external_lex_state = 131},
  [18842] = {.lex_state = 457, .external_lex_state = 99},
  [18843] = {.lex_state = 455},
  [18844] = {.lex_state = 457, .external_lex_state = 131},
  [18845] = {.lex_state = 457, .external_lex_state = 131},
  [18846] = {.lex_state = 457, .external_lex_state = 131},
  [18847] = {.lex_state = 457, .external_lex_state = 131},
  [18848] = {.lex_state = 457, .external_lex_state = 99},
  [18849] = {.lex_state = 457, .external_lex_state = 132},
  [18850] = {.lex_state = 455, .external_lex_state = 133},
  [18851] = {.lex_state = 457, .external_lex_state = 131},
  [18852] = {.lex_state = 457, .external_lex_state = 131},
  [18853] = {.lex_state = 457, .external_lex_state = 114},
  [18854] = {.lex_state = 457, .external_lex_state = 131},
  [18855] = {.lex_state = 455, .external_lex_state = 120},
  [18856] = {.lex_state = 457, .external_lex_state = 131},
  [18857] = {.lex_state = 457, .external_lex_state = 131},
  [18858] = {.lex_state = 457, .external_lex_state = 101},
  [18859] = {.lex_state = 457, .external_lex_state = 101},
  [18860] = {.lex_state = 457, .external_lex_state = 131},
  [18861] = {.lex_state = 457, .external_lex_state = 131},
  [18862] = {.lex_state = 457, .external_lex_state = 132},
  [18863] = {.lex_state = 455},
  [18864] = {.lex_state = 457, .external_lex_state = 131},
  [18865] = {.lex_state = 457, .external_lex_state = 131},
  [18866] = {.lex_state = 457, .external_lex_state = 131},
  [18867] = {.lex_state = 457, .external_lex_state = 131},
  [18868] = {.lex_state = 455},
  [18869] = {.lex_state = 455},
  [18870] = {.lex_state = 190},
  [18871] = {.lex_state = 457, .external_lex_state = 131},
  [18872] = {.lex_state = 457, .external_lex_state = 131},
  [18873] = {.lex_state = 457, .external_lex_state = 121},
  [18874] = {.lex_state = 457, .external_lex_state = 131},
  [18875] = {.lex_state = 224, .external_lex_state = 129},
  [18876] = {.lex_state = 457, .external_lex_state = 131},
  [18877] = {.lex_state = 455, .external_lex_state = 120},
  [18878] = {.lex_state = 457, .external_lex_state = 101},
  [18879] = {.lex_state = 457, .external_lex_state = 101},
  [18880] = {.lex_state = 457, .external_lex_state = 131},
  [18881] = {.lex_state = 455},
  [18882] = {.lex_state = 455},
  [18883] = {.lex_state = 457, .external_lex_state = 114},
  [18884] = {.lex_state = 457, .external_lex_state = 131},
  [18885] = {.lex_state = 455},
  [18886] = {.lex_state = 455},
  [18887] = {.lex_state = 455},
  [18888] = {.lex_state = 203},
  [18889] = {.lex_state = 457, .external_lex_state = 131},
  [18890] = {.lex_state = 457, .external_lex_state = 132},
  [18891] = {.lex_state = 455},
  [18892] = {.lex_state = 455},
  [18893] = {.lex_state = 457, .external_lex_state = 131},
  [18894] = {.lex_state = 457, .external_lex_state = 131},
  [18895] = {.lex_state = 455},
  [18896] = {.lex_state = 457, .external_lex_state = 131},
  [18897] = {.lex_state = 457, .external_lex_state = 131},
  [18898] = {.lex_state = 457, .external_lex_state = 131},
  [18899] = {.lex_state = 455},
  [18900] = {.lex_state = 457, .external_lex_state = 131},
  [18901] = {.lex_state = 455},
  [18902] = {.lex_state = 203},
  [18903] = {.lex_state = 455},
  [18904] = {.lex_state = 457, .external_lex_state = 131},
  [18905] = {.lex_state = 457, .external_lex_state = 131},
  [18906] = {.lex_state = 457, .external_lex_state = 131},
  [18907] = {.lex_state = 457, .external_lex_state = 131},
  [18908] = {.lex_state = 457, .external_lex_state = 131},
  [18909] = {.lex_state = 457, .external_lex_state = 132},
  [18910] = {.lex_state = 203},
  [18911] = {.lex_state = 455},
  [18912] = {.lex_state = 457, .external_lex_state = 131},
  [18913] = {.lex_state = 457, .external_lex_state = 114},
  [18914] = {.lex_state = 455},
  [18915] = {.lex_state = 457, .external_lex_state = 101},
  [18916] = {.lex_state = 457, .external_lex_state = 131},
  [18917] = {.lex_state = 457, .external_lex_state = 101},
  [18918] = {.lex_state = 457, .external_lex_state = 131},
  [18919] = {.lex_state = 455},
  [18920] = {.lex_state = 457, .external_lex_state = 99},
  [18921] = {.lex_state = 457, .external_lex_state = 131},
  [18922] = {.lex_state = 455},
  [18923] = {.lex_state = 455},
  [18924] = {.lex_state = 457, .external_lex_state = 114},
  [18925] = {.lex_state = 457, .external_lex_state = 131},
  [18926] = {.lex_state = 455},
  [18927] = {.lex_state = 457, .external_lex_state = 131},
  [18928] = {.lex_state = 457, .external_lex_state = 131},
  [18929] = {.lex_state = 457, .external_lex_state = 131},
  [18930] = {.lex_state = 455, .external_lex_state = 120},
  [18931] = {.lex_state = 457, .external_lex_state = 101},
  [18932] = {.lex_state = 455},
  [18933] = {.lex_state = 457, .external_lex_state = 131},
  [18934] = {.lex_state = 457, .external_lex_state = 131},
  [18935] = {.lex_state = 455},
  [18936] = {.lex_state = 455},
  [18937] = {.lex_state = 457, .external_lex_state = 99},
  [18938] = {.lex_state = 455},
  [18939] = {.lex_state = 224, .external_lex_state = 129},
  [18940] = {.lex_state = 457, .external_lex_state = 131},
  [18941] = {.lex_state = 224, .external_lex_state = 129},
  [18942] = {.lex_state = 455},
  [18943] = {.lex_state = 457, .external_lex_state = 131},
  [18944] = {.lex_state = 457, .external_lex_state = 114},
  [18945] = {.lex_state = 457, .external_lex_state = 131},
  [18946] = {.lex_state = 457, .external_lex_state = 131},
  [18947] = {.lex_state = 455},
  [18948] = {.lex_state = 457, .external_lex_state = 101},
  [18949] = {.lex_state = 457, .external_lex_state = 131},
  [18950] = {.lex_state = 455},
  [18951] = {.lex_state = 455},
  [18952] = {.lex_state = 457, .external_lex_state = 114},
  [18953] = {.lex_state = 455},
  [18954] = {.lex_state = 457, .external_lex_state = 99},
  [18955] = {.lex_state = 457, .external_lex_state = 114},
  [18956] = {.lex_state = 457, .external_lex_state = 131},
  [18957] = {.lex_state = 457, .external_lex_state = 132},
  [18958] = {.lex_state = 455},
  [18959] = {.lex_state = 455},
  [18960] = {.lex_state = 457, .external_lex_state = 114},
  [18961] = {.lex_state = 457, .external_lex_state = 131},
  [18962] = {.lex_state = 457, .external_lex_state = 131},
  [18963] = {.lex_state = 457, .external_lex_state = 131},
  [18964] = {.lex_state = 457, .external_lex_state = 114},
  [18965] = {.lex_state = 457, .external_lex_state = 131},
  [18966] = {.lex_state = 457, .external_lex_state = 114},
  [18967] = {.lex_state = 457},
  [18968] = {.lex_state = 457, .external_lex_state = 99},
  [18969] = {.lex_state = 457, .external_lex_state = 114},
  [18970] = {.lex_state = 457, .external_lex_state = 131},
  [18971] = {.lex_state = 457, .external_lex_state = 131},
  [18972] = {.lex_state = 457, .external_lex_state = 131},
  [18973] = {.lex_state = 224, .external_lex_state = 109},
  [18974] = {.lex_state = 457, .external_lex_state = 131},
  [18975] = {.lex_state = 457, .external_lex_state = 131},
  [18976] = {.lex_state = 457, .external_lex_state = 131},
  [18977] = {.lex_state = 457, .external_lex_state = 100},
  [18978] = {.lex_state = 457, .external_lex_state = 131},
  [18979] = {.lex_state = 457, .external_lex_state = 114},
  [18980] = {.lex_state = 457, .external_lex_state = 132},
  [18981] = {.lex_state = 457, .external_lex_state = 131},
  [18982] = {.lex_state = 457, .external_lex_state = 99},
  [18983] = {.lex_state = 457, .external_lex_state = 131},
  [18984] = {.lex_state = 457, .external_lex_state = 131},
  [18985] = {.lex_state = 457, .external_lex_state = 131},
  [18986] = {.lex_state = 457, .external_lex_state = 131},
  [18987] = {.lex_state = 457, .external_lex_state = 114},
  [18988] = {.lex_state = 455, .external_lex_state = 120},
  [18989] = {.lex_state = 457, .external_lex_state = 131},
  [18990] = {.lex_state = 457, .external_lex_state = 131},
  [18991] = {.lex_state = 457, .external_lex_state = 131},
  [18992] = {.lex_state = 457, .external_lex_state = 101},
  [18993] = {.lex_state = 457, .external_lex_state = 101},
  [18994] = {.lex_state = 457, .external_lex_state = 131},
  [18995] = {.lex_state = 455},
  [18996] = {.lex_state = 455},
  [18997] = {.lex_state = 455, .external_lex_state = 133},
  [18998] = {.lex_state = 457, .external_lex_state = 114},
  [18999] = {.lex_state = 457, .external_lex_state = 131},
  [19000] = {.lex_state = 457, .external_lex_state = 114},
  [19001] = {.lex_state = 457, .external_lex_state = 131},
  [19002] = {.lex_state = 455},
  [19003] = {.lex_state = 457, .external_lex_state = 101},
  [19004] = {.lex_state = 457, .external_lex_state = 131},
  [19005] = {.lex_state = 457, .external_lex_state = 131},
  [19006] = {.lex_state = 457, .external_lex_state = 131},
  [19007] = {.lex_state = 457, .external_lex_state = 131},
  [19008] = {.lex_state = 455},
  [19009] = {.lex_state = 457, .external_lex_state = 131},
  [19010] = {.lex_state = 457, .external_lex_state = 131},
  [19011] = {.lex_state = 457, .external_lex_state = 131},
  [19012] = {.lex_state = 457, .external_lex_state = 131},
  [19013] = {.lex_state = 457, .external_lex_state = 131},
  [19014] = {.lex_state = 457, .external_lex_state = 131},
  [19015] = {.lex_state = 457, .external_lex_state = 131},
  [19016] = {.lex_state = 190},
  [19017] = {.lex_state = 457, .external_lex_state = 131},
  [19018] = {.lex_state = 457, .external_lex_state = 121},
  [19019] = {.lex_state = 455},
  [19020] = {.lex_state = 457, .external_lex_state = 132},
  [19021] = {.lex_state = 457, .external_lex_state = 131},
  [19022] = {.lex_state = 457, .external_lex_state = 121},
  [19023] = {.lex_state = 457, .external_lex_state = 100},
  [19024] = {.lex_state = 455},
  [19025] = {.lex_state = 457, .external_lex_state = 121},
  [19026] = {.lex_state = 457, .external_lex_state = 131},
  [19027] = {.lex_state = 455},
  [19028] = {.lex_state = 457, .external_lex_state = 131},
  [19029] = {.lex_state = 457, .external_lex_state = 131},
  [19030] = {.lex_state = 457, .external_lex_state = 121},
  [19031] = {.lex_state = 457, .external_lex_state = 114},
  [19032] = {.lex_state = 457, .external_lex_state = 121},
  [19033] = {.lex_state = 457, .external_lex_state = 131},
  [19034] = {.lex_state = 457, .external_lex_state = 131},
  [19035] = {.lex_state = 457, .external_lex_state = 114},
  [19036] = {.lex_state = 457},
  [19037] = {.lex_state = 457, .external_lex_state = 131},
  [19038] = {.lex_state = 457, .external_lex_state = 131},
  [19039] = {.lex_state = 457, .external_lex_state = 131},
  [19040] = {.lex_state = 455},
  [19041] = {.lex_state = 457, .external_lex_state = 131},
  [19042] = {.lex_state = 455},
  [19043] = {.lex_state = 457, .external_lex_state = 132},
  [19044] = {.lex_state = 457, .external_lex_state = 131},
  [19045] = {.lex_state = 457, .external_lex_state = 131},
  [19046] = {.lex_state = 457, .external_lex_state = 131},
  [19047] = {.lex_state = 457, .external_lex_state = 114},
  [19048] = {.lex_state = 455, .external_lex_state = 120},
  [19049] = {.lex_state = 457, .external_lex_state = 114},
  [19050] = {.lex_state = 457, .external_lex_state = 131},
  [19051] = {.lex_state = 457, .external_lex_state = 131},
  [19052] = {.lex_state = 457, .external_lex_state = 101},
  [19053] = {.lex_state = 457, .external_lex_state = 131},
  [19054] = {.lex_state = 457, .external_lex_state = 131},
  [19055] = {.lex_state = 457, .external_lex_state = 131},
  [19056] = {.lex_state = 455},
  [19057] = {.lex_state = 457, .external_lex_state = 114},
  [19058] = {.lex_state = 457, .external_lex_state = 131},
  [19059] = {.lex_state = 457, .external_lex_state = 131},
  [19060] = {.lex_state = 457, .external_lex_state = 131},
  [19061] = {.lex_state = 457, .external_lex_state = 131},
  [19062] = {.lex_state = 457},
  [19063] = {.lex_state = 457, .external_lex_state = 131},
  [19064] = {.lex_state = 455},
  [19065] = {.lex_state = 455},
  [19066] = {.lex_state = 457, .external_lex_state = 131},
  [19067] = {.lex_state = 457, .external_lex_state = 131},
  [19068] = {.lex_state = 457, .external_lex_state = 131},
  [19069] = {.lex_state = 457, .external_lex_state = 132},
  [19070] = {.lex_state = 457, .external_lex_state = 131},
  [19071] = {.lex_state = 457, .external_lex_state = 101},
  [19072] = {.lex_state = 457, .external_lex_state = 131},
  [19073] = {.lex_state = 455},
  [19074] = {.lex_state = 455},
  [19075] = {.lex_state = 457, .external_lex_state = 131},
  [19076] = {.lex_state = 457, .external_lex_state = 131},
  [19077] = {.lex_state = 224, .external_lex_state = 129},
  [19078] = {.lex_state = 457, .external_lex_state = 131},
  [19079] = {.lex_state = 457, .external_lex_state = 99},
  [19080] = {.lex_state = 203},
  [19081] = {.lex_state = 457, .external_lex_state = 131},
  [19082] = {.lex_state = 457},
  [19083] = {.lex_state = 457, .external_lex_state = 114},
  [19084] = {.lex_state = 457, .external_lex_state = 122},
  [19085] = {.lex_state = 457},
  [19086] = {.lex_state = 457},
  [19087] = {.lex_state = 457},
  [19088] = {.lex_state = 190},
  [19089] = {.lex_state = 203},
  [19090] = {.lex_state = 457},
  [19091] = {.lex_state = 457, .external_lex_state = 114},
  [19092] = {.lex_state = 455, .external_lex_state = 134},
  [19093] = {.lex_state = 457},
  [19094] = {.lex_state = 203},
  [19095] = {.lex_state = 203},
  [19096] = {.lex_state = 457},
  [19097] = {.lex_state = 190},
  [19098] = {.lex_state = 455, .external_lex_state = 135},
  [19099] = {.lex_state = 455, .external_lex_state = 134},
  [19100] = {.lex_state = 457},
  [19101] = {.lex_state = 457, .external_lex_state = 114},
  [19102] = {.lex_state = 457},
  [19103] = {.lex_state = 455, .external_lex_state = 134},
  [19104] = {.lex_state = 455},
  [19105] = {.lex_state = 190},
  [19106] = {.lex_state = 457, .external_lex_state = 114},
  [19107] = {.lex_state = 203},
  [19108] = {.lex_state = 457, .external_lex_state = 114},
  [19109] = {.lex_state = 457},
  [19110] = {.lex_state = 457, .external_lex_state = 114},
  [19111] = {.lex_state = 457, .external_lex_state = 114},
  [19112] = {.lex_state = 457, .external_lex_state = 114},
  [19113] = {.lex_state = 457},
  [19114] = {.lex_state = 203},
  [19115] = {.lex_state = 457},
  [19116] = {.lex_state = 455},
  [19117] = {.lex_state = 457, .external_lex_state = 122},
  [19118] = {.lex_state = 203},
  [19119] = {.lex_state = 457, .external_lex_state = 114},
  [19120] = {.lex_state = 457, .external_lex_state = 114},
  [19121] = {.lex_state = 457},
  [19122] = {.lex_state = 455},
  [19123] = {.lex_state = 203},
  [19124] = {.lex_state = 457},
  [19125] = {.lex_state = 457, .external_lex_state = 114},
  [19126] = {.lex_state = 457, .external_lex_state = 114},
  [19127] = {.lex_state = 190},
  [19128] = {.lex_state = 455},
  [19129] = {.lex_state = 457, .external_lex_state = 114},
  [19130] = {.lex_state = 455, .external_lex_state = 134},
  [19131] = {.lex_state = 455},
  [19132] = {.lex_state = 457},
  [19133] = {.lex_state = 457, .external_lex_state = 114},
  [19134] = {.lex_state = 457},
  [19135] = {.lex_state = 457, .external_lex_state = 122},
  [19136] = {.lex_state = 457, .external_lex_state = 114},
  [19137] = {.lex_state = 457},
  [19138] = {.lex_state = 203},
  [19139] = {.lex_state = 457, .external_lex_state = 114},
  [19140] = {.lex_state = 457, .external_lex_state = 114},
  [19141] = {.lex_state = 203},
  [19142] = {.lex_state = 457},
  [19143] = {.lex_state = 457},
  [19144] = {.lex_state = 203},
  [19145] = {.lex_state = 457, .external_lex_state = 114},
  [19146] = {.lex_state = 457},
  [19147] = {.lex_state = 203},
  [19148] = {.lex_state = 457},
  [19149] = {.lex_state = 203},
  [19150] = {.lex_state = 190},
  [19151] = {.lex_state = 457},
  [19152] = {.lex_state = 457, .external_lex_state = 122},
  [19153] = {.lex_state = 457, .external_lex_state = 136},
  [19154] = {.lex_state = 455},
  [19155] = {.lex_state = 455, .external_lex_state = 135},
  [19156] = {.lex_state = 457},
  [19157] = {.lex_state = 457},
  [19158] = {.lex_state = 457},
  [19159] = {.lex_state = 190},
  [19160] = {.lex_state = 457},
  [19161] = {.lex_state = 190},
  [19162] = {.lex_state = 457},
  [19163] = {.lex_state = 455, .external_lex_state = 134},
  [19164] = {.lex_state = 457, .external_lex_state = 114},
  [19165] = {.lex_state = 457},
  [19166] = {.lex_state = 457},
  [19167] = {.lex_state = 1},
  [19168] = {.lex_state = 457},
  [19169] = {.lex_state = 224, .external_lex_state = 109},
  [19170] = {.lex_state = 457},
  [19171] = {.lex_state = 1},
  [19172] = {.lex_state = 455, .external_lex_state = 135},
  [19173] = {.lex_state = 457},
  [19174] = {.lex_state = 457, .external_lex_state = 114},
  [19175] = {.lex_state = 455},
  [19176] = {.lex_state = 457, .external_lex_state = 114},
  [19177] = {.lex_state = 457},
  [19178] = {.lex_state = 455},
  [19179] = {.lex_state = 457, .external_lex_state = 114},
  [19180] = {.lex_state = 455},
  [19181] = {.lex_state = 455},
  [19182] = {.lex_state = 203},
  [19183] = {.lex_state = 190},
  [19184] = {.lex_state = 457},
  [19185] = {.lex_state = 457},
  [19186] = {.lex_state = 455},
  [19187] = {.lex_state = 457, .external_lex_state = 114},
  [19188] = {.lex_state = 457, .external_lex_state = 122},
  [19189] = {.lex_state = 203},
  [19190] = {.lex_state = 455},
  [19191] = {.lex_state = 455},
  [19192] = {.lex_state = 203},
  [19193] = {.lex_state = 455},
  [19194] = {.lex_state = 455},
  [19195] = {.lex_state = 455},
  [19196] = {.lex_state = 457},
  [19197] = {.lex_state = 455},
  [19198] = {.lex_state = 455},
  [19199] = {.lex_state = 457, .external_lex_state = 114},
  [19200] = {.lex_state = 457},
  [19201] = {.lex_state = 457},
  [19202] = {.lex_state = 457},
  [19203] = {.lex_state = 457},
  [19204] = {.lex_state = 455},
  [19205] = {.lex_state = 457},
  [19206] = {.lex_state = 190},
  [19207] = {.lex_state = 457, .external_lex_state = 114},
  [19208] = {.lex_state = 457},
  [19209] = {.lex_state = 457},
  [19210] = {.lex_state = 203},
  [19211] = {.lex_state = 203},
  [19212] = {.lex_state = 457},
  [19213] = {.lex_state = 457},
  [19214] = {.lex_state = 455, .external_lex_state = 135},
  [19215] = {.lex_state = 457},
  [19216] = {.lex_state = 457, .external_lex_state = 114},
  [19217] = {.lex_state = 190},
  [19218] = {.lex_state = 457, .external_lex_state = 114},
  [19219] = {.lex_state = 203},
  [19220] = {.lex_state = 455},
  [19221] = {.lex_state = 457},
  [19222] = {.lex_state = 203},
  [19223] = {.lex_state = 457},
  [19224] = {.lex_state = 190},
  [19225] = {.lex_state = 455},
  [19226] = {.lex_state = 457, .external_lex_state = 114},
  [19227] = {.lex_state = 457, .external_lex_state = 114},
  [19228] = {.lex_state = 455},
  [19229] = {.lex_state = 457, .external_lex_state = 122},
  [19230] = {.lex_state = 457},
  [19231] = {.lex_state = 457},
  [19232] = {.lex_state = 190},
  [19233] = {.lex_state = 457},
  [19234] = {.lex_state = 457, .external_lex_state = 122},
  [19235] = {.lex_state = 457, .external_lex_state = 114},
  [19236] = {.lex_state = 457},
  [19237] = {.lex_state = 455},
  [19238] = {.lex_state = 203},
  [19239] = {.lex_state = 457, .external_lex_state = 114},
  [19240] = {.lex_state = 457, .external_lex_state = 131},
  [19241] = {.lex_state = 457, .external_lex_state = 114},
  [19242] = {.lex_state = 457},
  [19243] = {.lex_state = 455},
  [19244] = {.lex_state = 457, .external_lex_state = 114},
  [19245] = {.lex_state = 457, .external_lex_state = 114},
  [19246] = {.lex_state = 203},
  [19247] = {.lex_state = 457},
  [19248] = {.lex_state = 457, .external_lex_state = 114},
  [19249] = {.lex_state = 457},
  [19250] = {.lex_state = 457},
  [19251] = {.lex_state = 457, .external_lex_state = 122},
  [19252] = {.lex_state = 455},
  [19253] = {.lex_state = 203},
  [19254] = {.lex_state = 457, .external_lex_state = 114},
  [19255] = {.lex_state = 455},
  [19256] = {.lex_state = 203},
  [19257] = {.lex_state = 203},
  [19258] = {.lex_state = 457},
  [19259] = {.lex_state = 455},
  [19260] = {.lex_state = 457, .external_lex_state = 114},
  [19261] = {.lex_state = 203},
  [19262] = {.lex_state = 203},
  [19263] = {.lex_state = 455},
  [19264] = {.lex_state = 455},
  [19265] = {.lex_state = 457},
  [19266] = {.lex_state = 457, .external_lex_state = 122},
  [19267] = {.lex_state = 457, .external_lex_state = 136},
  [19268] = {.lex_state = 455},
  [19269] = {.lex_state = 455, .external_lex_state = 135},
  [19270] = {.lex_state = 457, .external_lex_state = 122},
  [19271] = {.lex_state = 457},
  [19272] = {.lex_state = 457},
  [19273] = {.lex_state = 457, .external_lex_state = 114},
  [19274] = {.lex_state = 457, .external_lex_state = 114},
  [19275] = {.lex_state = 455},
  [19276] = {.lex_state = 457},
  [19277] = {.lex_state = 457},
  [19278] = {.lex_state = 455, .external_lex_state = 134},
  [19279] = {.lex_state = 457},
  [19280] = {.lex_state = 457},
  [19281] = {.lex_state = 203},
  [19282] = {.lex_state = 457},
  [19283] = {.lex_state = 224, .external_lex_state = 109},
  [19284] = {.lex_state = 457},
  [19285] = {.lex_state = 457},
  [19286] = {.lex_state = 455},
  [19287] = {.lex_state = 457},
  [19288] = {.lex_state = 457, .external_lex_state = 114},
  [19289] = {.lex_state = 190},
  [19290] = {.lex_state = 457, .external_lex_state = 114},
  [19291] = {.lex_state = 457, .external_lex_state = 114},
  [19292] = {.lex_state = 457, .external_lex_state = 114},
  [19293] = {.lex_state = 457, .external_lex_state = 114},
  [19294] = {.lex_state = 457},
  [19295] = {.lex_state = 457},
  [19296] = {.lex_state = 190},
  [19297] = {.lex_state = 457, .external_lex_state = 114},
  [19298] = {.lex_state = 457},
  [19299] = {.lex_state = 457, .external_lex_state = 114},
  [19300] = {.lex_state = 457},
  [19301] = {.lex_state = 457, .external_lex_state = 114},
  [19302] = {.lex_state = 190},
  [19303] = {.lex_state = 203},
  [19304] = {.lex_state = 457},
  [19305] = {.lex_state = 455},
  [19306] = {.lex_state = 455},
  [19307] = {.lex_state = 455},
  [19308] = {.lex_state = 457},
  [19309] = {.lex_state = 457},
  [19310] = {.lex_state = 457},
  [19311] = {.lex_state = 455, .external_lex_state = 134},
  [19312] = {.lex_state = 455},
  [19313] = {.lex_state = 457, .external_lex_state = 114},
  [19314] = {.lex_state = 457},
  [19315] = {.lex_state = 457},
  [19316] = {.lex_state = 190},
  [19317] = {.lex_state = 457},
  [19318] = {.lex_state = 455},
  [19319] = {.lex_state = 457},
  [19320] = {.lex_state = 203},
  [19321] = {.lex_state = 457, .external_lex_state = 114},
  [19322] = {.lex_state = 455},
  [19323] = {.lex_state = 457},
  [19324] = {.lex_state = 203},
  [19325] = {.lex_state = 190},
  [19326] = {.lex_state = 457},
  [19327] = {.lex_state = 457},
  [19328] = {.lex_state = 455, .external_lex_state = 135},
  [19329] = {.lex_state = 457},
  [19330] = {.lex_state = 457, .external_lex_state = 114},
  [19331] = {.lex_state = 455},
  [19332] = {.lex_state = 457, .external_lex_state = 114},
  [19333] = {.lex_state = 457},
  [19334] = {.lex_state = 203},
  [19335] = {.lex_state = 457, .external_lex_state = 122},
  [19336] = {.lex_state = 203},
  [19337] = {.lex_state = 203},
  [19338] = {.lex_state = 457, .external_lex_state = 128},
  [19339] = {.lex_state = 457},
  [19340] = {.lex_state = 457},
  [19341] = {.lex_state = 457, .external_lex_state = 114},
  [19342] = {.lex_state = 457},
  [19343] = {.lex_state = 203},
  [19344] = {.lex_state = 457},
  [19345] = {.lex_state = 455},
  [19346] = {.lex_state = 190},
  [19347] = {.lex_state = 457},
  [19348] = {.lex_state = 457, .external_lex_state = 114},
  [19349] = {.lex_state = 457, .external_lex_state = 114},
  [19350] = {.lex_state = 1},
  [19351] = {.lex_state = 190},
  [19352] = {.lex_state = 203},
  [19353] = {.lex_state = 457},
  [19354] = {.lex_state = 457},
  [19355] = {.lex_state = 457, .external_lex_state = 114},
  [19356] = {.lex_state = 190},
  [19357] = {.lex_state = 455, .external_lex_state = 134},
  [19358] = {.lex_state = 457, .external_lex_state = 114},
  [19359] = {.lex_state = 455, .external_lex_state = 134},
  [19360] = {.lex_state = 455, .external_lex_state = 134},
  [19361] = {.lex_state = 457},
  [19362] = {.lex_state = 457, .external_lex_state = 114},
  [19363] = {.lex_state = 457},
  [19364] = {.lex_state = 457, .external_lex_state = 114},
  [19365] = {.lex_state = 457, .external_lex_state = 128},
  [19366] = {.lex_state = 457, .external_lex_state = 114},
  [19367] = {.lex_state = 203},
  [19368] = {.lex_state = 457, .external_lex_state = 114},
  [19369] = {.lex_state = 457},
  [19370] = {.lex_state = 203},
  [19371] = {.lex_state = 455},
  [19372] = {.lex_state = 457},
  [19373] = {.lex_state = 455},
  [19374] = {.lex_state = 457, .external_lex_state = 114},
  [19375] = {.lex_state = 457},
  [19376] = {.lex_state = 203},
  [19377] = {.lex_state = 457, .external_lex_state = 122},
  [19378] = {.lex_state = 457},
  [19379] = {.lex_state = 457},
  [19380] = {.lex_state = 190},
  [19381] = {.lex_state = 457, .external_lex_state = 136},
  [19382] = {.lex_state = 455},
  [19383] = {.lex_state = 455, .external_lex_state = 135},
  [19384] = {.lex_state = 457, .external_lex_state = 114},
  [19385] = {.lex_state = 457, .external_lex_state = 114},
  [19386] = {.lex_state = 457},
  [19387] = {.lex_state = 455},
  [19388] = {.lex_state = 457},
  [19389] = {.lex_state = 203},
  [19390] = {.lex_state = 457},
  [19391] = {.lex_state = 190},
  [19392] = {.lex_state = 457},
  [19393] = {.lex_state = 457},
  [19394] = {.lex_state = 190},
  [19395] = {.lex_state = 457},
  [19396] = {.lex_state = 457},
  [19397] = {.lex_state = 224, .external_lex_state = 109},
  [19398] = {.lex_state = 457},
  [19399] = {.lex_state = 457},
  [19400] = {.lex_state = 457},
  [19401] = {.lex_state = 457},
  [19402] = {.lex_state = 457, .external_lex_state = 114},
  [19403] = {.lex_state = 455, .external_lex_state = 134},
  [19404] = {.lex_state = 457, .external_lex_state = 114},
  [19405] = {.lex_state = 457, .external_lex_state = 131},
  [19406] = {.lex_state = 455, .external_lex_state = 134},
  [19407] = {.lex_state = 190},
  [19408] = {.lex_state = 457},
  [19409] = {.lex_state = 455},
  [19410] = {.lex_state = 457},
  [19411] = {.lex_state = 203},
  [19412] = {.lex_state = 457},
  [19413] = {.lex_state = 455},
  [19414] = {.lex_state = 190},
  [19415] = {.lex_state = 457, .external_lex_state = 114},
  [19416] = {.lex_state = 457, .external_lex_state = 114},
  [19417] = {.lex_state = 457, .external_lex_state = 114},
  [19418] = {.lex_state = 457},
  [19419] = {.lex_state = 457},
  [19420] = {.lex_state = 190},
  [19421] = {.lex_state = 455},
  [19422] = {.lex_state = 457},
  [19423] = {.lex_state = 457},
  [19424] = {.lex_state = 457},
  [19425] = {.lex_state = 457},
  [19426] = {.lex_state = 455, .external_lex_state = 134},
  [19427] = {.lex_state = 457, .external_lex_state = 114},
  [19428] = {.lex_state = 457},
  [19429] = {.lex_state = 457},
  [19430] = {.lex_state = 455, .external_lex_state = 134},
  [19431] = {.lex_state = 455},
  [19432] = {.lex_state = 455},
  [19433] = {.lex_state = 457},
  [19434] = {.lex_state = 457},
  [19435] = {.lex_state = 457, .external_lex_state = 114},
  [19436] = {.lex_state = 457},
  [19437] = {.lex_state = 457},
  [19438] = {.lex_state = 203},
  [19439] = {.lex_state = 455},
  [19440] = {.lex_state = 455},
  [19441] = {.lex_state = 455},
  [19442] = {.lex_state = 455, .external_lex_state = 135},
  [19443] = {.lex_state = 457},
  [19444] = {.lex_state = 457, .external_lex_state = 114},
  [19445] = {.lex_state = 457},
  [19446] = {.lex_state = 455},
  [19447] = {.lex_state = 203},
  [19448] = {.lex_state = 455, .external_lex_state = 134},
  [19449] = {.lex_state = 457},
  [19450] = {.lex_state = 203},
  [19451] = {.lex_state = 457},
  [19452] = {.lex_state = 190},
  [19453] = {.lex_state = 455, .external_lex_state = 134},
  [19454] = {.lex_state = 457},
  [19455] = {.lex_state = 457, .external_lex_state = 114},
  [19456] = {.lex_state = 457},
  [19457] = {.lex_state = 457},
  [19458] = {.lex_state = 190},
  [19459] = {.lex_state = 457},
  [19460] = {.lex_state = 457, .external_lex_state = 114},
  [19461] = {.lex_state = 457},
  [19462] = {.lex_state = 190},
  [19463] = {.lex_state = 457, .external_lex_state = 114},
  [19464] = {.lex_state = 203},
  [19465] = {.lex_state = 457},
  [19466] = {.lex_state = 203},
  [19467] = {.lex_state = 457},
  [19468] = {.lex_state = 190},
  [19469] = {.lex_state = 457, .external_lex_state = 114},
  [19470] = {.lex_state = 457, .external_lex_state = 131},
  [19471] = {.lex_state = 455},
  [19472] = {.lex_state = 457, .external_lex_state = 114},
  [19473] = {.lex_state = 457},
  [19474] = {.lex_state = 190},
  [19475] = {.lex_state = 457, .external_lex_state = 114},
  [19476] = {.lex_state = 457, .external_lex_state = 114},
  [19477] = {.lex_state = 457},
  [19478] = {.lex_state = 203},
  [19479] = {.lex_state = 457},
  [19480] = {.lex_state = 457},
  [19481] = {.lex_state = 203},
  [19482] = {.lex_state = 457, .external_lex_state = 114},
  [19483] = {.lex_state = 203},
  [19484] = {.lex_state = 203},
  [19485] = {.lex_state = 457, .external_lex_state = 122},
  [19486] = {.lex_state = 457},
  [19487] = {.lex_state = 190},
  [19488] = {.lex_state = 457, .external_lex_state = 114},
  [19489] = {.lex_state = 457},
  [19490] = {.lex_state = 203},
  [19491] = {.lex_state = 455},
  [19492] = {.lex_state = 203},
  [19493] = {.lex_state = 457},
  [19494] = {.lex_state = 455},
  [19495] = {.lex_state = 457, .external_lex_state = 136},
  [19496] = {.lex_state = 455},
  [19497] = {.lex_state = 455, .external_lex_state = 135},
  [19498] = {.lex_state = 455},
  [19499] = {.lex_state = 457, .external_lex_state = 114},
  [19500] = {.lex_state = 457, .external_lex_state = 131},
  [19501] = {.lex_state = 457},
  [19502] = {.lex_state = 457},
  [19503] = {.lex_state = 455},
  [19504] = {.lex_state = 457},
  [19505] = {.lex_state = 203},
  [19506] = {.lex_state = 455},
  [19507] = {.lex_state = 457},
  [19508] = {.lex_state = 203},
  [19509] = {.lex_state = 457, .external_lex_state = 122},
  [19510] = {.lex_state = 457},
  [19511] = {.lex_state = 457, .external_lex_state = 114},
  [19512] = {.lex_state = 455, .external_lex_state = 134},
  [19513] = {.lex_state = 457},
  [19514] = {.lex_state = 457, .external_lex_state = 114},
  [19515] = {.lex_state = 190},
  [19516] = {.lex_state = 457, .external_lex_state = 114},
  [19517] = {.lex_state = 457},
  [19518] = {.lex_state = 457},
  [19519] = {.lex_state = 457},
  [19520] = {.lex_state = 457},
  [19521] = {.lex_state = 457},
  [19522] = {.lex_state = 457},
  [19523] = {.lex_state = 457, .external_lex_state = 114},
  [19524] = {.lex_state = 203},
  [19525] = {.lex_state = 457},
  [19526] = {.lex_state = 455, .external_lex_state = 134},
  [19527] = {.lex_state = 190},
  [19528] = {.lex_state = 455, .external_lex_state = 134},
  [19529] = {.lex_state = 455},
  [19530] = {.lex_state = 457},
  [19531] = {.lex_state = 457},
  [19532] = {.lex_state = 457},
  [19533] = {.lex_state = 190},
  [19534] = {.lex_state = 457, .external_lex_state = 114},
  [19535] = {.lex_state = 457, .external_lex_state = 114},
  [19536] = {.lex_state = 457, .external_lex_state = 114},
  [19537] = {.lex_state = 457},
  [19538] = {.lex_state = 203},
  [19539] = {.lex_state = 203},
  [19540] = {.lex_state = 457},
  [19541] = {.lex_state = 455, .external_lex_state = 135},
  [19542] = {.lex_state = 457, .external_lex_state = 114},
  [19543] = {.lex_state = 190},
  [19544] = {.lex_state = 457, .external_lex_state = 114},
  [19545] = {.lex_state = 457},
  [19546] = {.lex_state = 203},
  [19547] = {.lex_state = 457, .external_lex_state = 114},
  [19548] = {.lex_state = 455},
  [19549] = {.lex_state = 457, .external_lex_state = 122},
  [19550] = {.lex_state = 203},
  [19551] = {.lex_state = 457},
  [19552] = {.lex_state = 203},
  [19553] = {.lex_state = 457, .external_lex_state = 122},
  [19554] = {.lex_state = 203},
  [19555] = {.lex_state = 203},
  [19556] = {.lex_state = 457},
  [19557] = {.lex_state = 457},
  [19558] = {.lex_state = 455},
  [19559] = {.lex_state = 203},
  [19560] = {.lex_state = 457, .external_lex_state = 122},
  [19561] = {.lex_state = 457},
  [19562] = {.lex_state = 203},
  [19563] = {.lex_state = 203},
  [19564] = {.lex_state = 457},
  [19565] = {.lex_state = 190},
  [19566] = {.lex_state = 457},
  [19567] = {.lex_state = 203},
  [19568] = {.lex_state = 457, .external_lex_state = 122},
  [19569] = {.lex_state = 457, .external_lex_state = 122},
  [19570] = {.lex_state = 457, .external_lex_state = 122},
  [19571] = {.lex_state = 457},
  [19572] = {.lex_state = 457, .external_lex_state = 136},
  [19573] = {.lex_state = 455},
  [19574] = {.lex_state = 455, .external_lex_state = 135},
  [19575] = {.lex_state = 457},
  [19576] = {.lex_state = 203},
  [19577] = {.lex_state = 457},
  [19578] = {.lex_state = 457, .external_lex_state = 114},
  [19579] = {.lex_state = 190},
  [19580] = {.lex_state = 455, .external_lex_state = 134},
  [19581] = {.lex_state = 457},
  [19582] = {.lex_state = 457},
  [19583] = {.lex_state = 190},
  [19584] = {.lex_state = 457},
  [19585] = {.lex_state = 457, .external_lex_state = 114},
  [19586] = {.lex_state = 457, .external_lex_state = 114},
  [19587] = {.lex_state = 457, .external_lex_state = 114},
  [19588] = {.lex_state = 457},
  [19589] = {.lex_state = 457, .external_lex_state = 122},
  [19590] = {.lex_state = 457},
  [19591] = {.lex_state = 457, .external_lex_state = 114},
  [19592] = {.lex_state = 457},
  [19593] = {.lex_state = 457, .external_lex_state = 114},
  [19594] = {.lex_state = 457},
  [19595] = {.lex_state = 455},
  [19596] = {.lex_state = 190},
  [19597] = {.lex_state = 457, .external_lex_state = 114},
  [19598] = {.lex_state = 457, .external_lex_state = 114},
  [19599] = {.lex_state = 457, .external_lex_state = 114},
  [19600] = {.lex_state = 457, .external_lex_state = 114},
  [19601] = {.lex_state = 455, .external_lex_state = 134},
  [19602] = {.lex_state = 455, .external_lex_state = 134},
  [19603] = {.lex_state = 457},
  [19604] = {.lex_state = 457},
  [19605] = {.lex_state = 455},
  [19606] = {.lex_state = 455},
  [19607] = {.lex_state = 203},
  [19608] = {.lex_state = 457},
  [19609] = {.lex_state = 457},
  [19610] = {.lex_state = 457, .external_lex_state = 114},
  [19611] = {.lex_state = 457, .external_lex_state = 114},
  [19612] = {.lex_state = 457, .external_lex_state = 114},
  [19613] = {.lex_state = 457, .external_lex_state = 122},
  [19614] = {.lex_state = 457},
  [19615] = {.lex_state = 203},
  [19616] = {.lex_state = 457},
  [19617] = {.lex_state = 457},
  [19618] = {.lex_state = 455, .external_lex_state = 135},
  [19619] = {.lex_state = 190},
  [19620] = {.lex_state = 457},
  [19621] = {.lex_state = 457, .external_lex_state = 128},
  [19622] = {.lex_state = 457, .external_lex_state = 131},
  [19623] = {.lex_state = 203},
  [19624] = {.lex_state = 457},
  [19625] = {.lex_state = 457},
  [19626] = {.lex_state = 203},
  [19627] = {.lex_state = 457},
  [19628] = {.lex_state = 457, .external_lex_state = 114},
  [19629] = {.lex_state = 190},
  [19630] = {.lex_state = 457},
  [19631] = {.lex_state = 203},
  [19632] = {.lex_state = 190},
  [19633] = {.lex_state = 455, .external_lex_state = 134},
  [19634] = {.lex_state = 190},
  [19635] = {.lex_state = 457},
  [19636] = {.lex_state = 455, .external_lex_state = 134},
  [19637] = {.lex_state = 457},
  [19638] = {.lex_state = 457, .external_lex_state = 114},
  [19639] = {.lex_state = 203},
  [19640] = {.lex_state = 203},
  [19641] = {.lex_state = 455, .external_lex_state = 134},
  [19642] = {.lex_state = 455},
  [19643] = {.lex_state = 455, .external_lex_state = 134},
  [19644] = {.lex_state = 203},
  [19645] = {.lex_state = 457, .external_lex_state = 122},
  [19646] = {.lex_state = 455},
  [19647] = {.lex_state = 203},
  [19648] = {.lex_state = 457},
  [19649] = {.lex_state = 457, .external_lex_state = 136},
  [19650] = {.lex_state = 455},
  [19651] = {.lex_state = 455, .external_lex_state = 135},
  [19652] = {.lex_state = 457},
  [19653] = {.lex_state = 457, .external_lex_state = 114},
  [19654] = {.lex_state = 457},
  [19655] = {.lex_state = 455},
  [19656] = {.lex_state = 203},
  [19657] = {.lex_state = 457},
  [19658] = {.lex_state = 455},
  [19659] = {.lex_state = 457, .external_lex_state = 114},
  [19660] = {.lex_state = 457},
  [19661] = {.lex_state = 457, .external_lex_state = 122},
  [19662] = {.lex_state = 455, .external_lex_state = 134},
  [19663] = {.lex_state = 457},
  [19664] = {.lex_state = 457},
  [19665] = {.lex_state = 190},
  [19666] = {.lex_state = 457},
  [19667] = {.lex_state = 457, .external_lex_state = 114},
  [19668] = {.lex_state = 457},
  [19669] = {.lex_state = 457, .external_lex_state = 114},
  [19670] = {.lex_state = 457, .external_lex_state = 114},
  [19671] = {.lex_state = 457, .external_lex_state = 114},
  [19672] = {.lex_state = 457},
  [19673] = {.lex_state = 203},
  [19674] = {.lex_state = 457},
  [19675] = {.lex_state = 457},
  [19676] = {.lex_state = 457, .external_lex_state = 114},
  [19677] = {.lex_state = 190},
  [19678] = {.lex_state = 457},
  [19679] = {.lex_state = 190},
  [19680] = {.lex_state = 457, .external_lex_state = 122},
  [19681] = {.lex_state = 455, .external_lex_state = 134},
  [19682] = {.lex_state = 455},
  [19683] = {.lex_state = 457},
  [19684] = {.lex_state = 457, .external_lex_state = 122},
  [19685] = {.lex_state = 457},
  [19686] = {.lex_state = 457},
  [19687] = {.lex_state = 457, .external_lex_state = 114},
  [19688] = {.lex_state = 457, .external_lex_state = 114},
  [19689] = {.lex_state = 455},
  [19690] = {.lex_state = 457},
  [19691] = {.lex_state = 203},
  [19692] = {.lex_state = 457, .external_lex_state = 114},
  [19693] = {.lex_state = 455},
  [19694] = {.lex_state = 455, .external_lex_state = 135},
  [19695] = {.lex_state = 203},
  [19696] = {.lex_state = 455, .external_lex_state = 134},
  [19697] = {.lex_state = 455, .external_lex_state = 134},
  [19698] = {.lex_state = 455, .external_lex_state = 134},
  [19699] = {.lex_state = 203},
  [19700] = {.lex_state = 455, .external_lex_state = 134},
  [19701] = {.lex_state = 457},
  [19702] = {.lex_state = 457},
  [19703] = {.lex_state = 457},
  [19704] = {.lex_state = 455},
  [19705] = {.lex_state = 203},
  [19706] = {.lex_state = 203},
  [19707] = {.lex_state = 190},
  [19708] = {.lex_state = 457, .external_lex_state = 114},
  [19709] = {.lex_state = 457, .external_lex_state = 114},
  [19710] = {.lex_state = 457, .external_lex_state = 128},
  [19711] = {.lex_state = 455, .external_lex_state = 134},
  [19712] = {.lex_state = 457},
  [19713] = {.lex_state = 203},
  [19714] = {.lex_state = 203},
  [19715] = {.lex_state = 457, .external_lex_state = 114},
  [19716] = {.lex_state = 455, .external_lex_state = 134},
  [19717] = {.lex_state = 203},
  [19718] = {.lex_state = 457, .external_lex_state = 122},
  [19719] = {.lex_state = 457, .external_lex_state = 114},
  [19720] = {.lex_state = 457},
  [19721] = {.lex_state = 457, .external_lex_state = 136},
  [19722] = {.lex_state = 455},
  [19723] = {.lex_state = 455, .external_lex_state = 135},
  [19724] = {.lex_state = 203},
  [19725] = {.lex_state = 457},
  [19726] = {.lex_state = 455, .external_lex_state = 134},
  [19727] = {.lex_state = 190},
  [19728] = {.lex_state = 457, .external_lex_state = 128},
  [19729] = {.lex_state = 457},
  [19730] = {.lex_state = 457},
  [19731] = {.lex_state = 190},
  [19732] = {.lex_state = 457},
  [19733] = {.lex_state = 457, .external_lex_state = 128},
  [19734] = {.lex_state = 457},
  [19735] = {.lex_state = 457, .external_lex_state = 114},
  [19736] = {.lex_state = 457, .external_lex_state = 114},
  [19737] = {.lex_state = 457, .external_lex_state = 114},
  [19738] = {.lex_state = 457},
  [19739] = {.lex_state = 457, .external_lex_state = 114},
  [19740] = {.lex_state = 457, .external_lex_state = 114},
  [19741] = {.lex_state = 457, .external_lex_state = 114},
  [19742] = {.lex_state = 455, .external_lex_state = 134},
  [19743] = {.lex_state = 457},
  [19744] = {.lex_state = 455},
  [19745] = {.lex_state = 455, .external_lex_state = 134},
  [19746] = {.lex_state = 457, .external_lex_state = 122},
  [19747] = {.lex_state = 457},
  [19748] = {.lex_state = 457, .external_lex_state = 114},
  [19749] = {.lex_state = 457},
  [19750] = {.lex_state = 455, .external_lex_state = 134},
  [19751] = {.lex_state = 455},
  [19752] = {.lex_state = 457},
  [19753] = {.lex_state = 457},
  [19754] = {.lex_state = 455},
  [19755] = {.lex_state = 455},
  [19756] = {.lex_state = 203},
  [19757] = {.lex_state = 457},
  [19758] = {.lex_state = 455},
  [19759] = {.lex_state = 457, .external_lex_state = 114},
  [19760] = {.lex_state = 457, .external_lex_state = 114},
  [19761] = {.lex_state = 457, .external_lex_state = 122},
  [19762] = {.lex_state = 457},
  [19763] = {.lex_state = 203},
  [19764] = {.lex_state = 457},
  [19765] = {.lex_state = 455, .external_lex_state = 134},
  [19766] = {.lex_state = 455, .external_lex_state = 135},
  [19767] = {.lex_state = 190},
  [19768] = {.lex_state = 457},
  [19769] = {.lex_state = 455},
  [19770] = {.lex_state = 455},
  [19771] = {.lex_state = 203},
  [19772] = {.lex_state = 457},
  [19773] = {.lex_state = 455},
  [19774] = {.lex_state = 455},
  [19775] = {.lex_state = 457},
  [19776] = {.lex_state = 203},
  [19777] = {.lex_state = 457},
  [19778] = {.lex_state = 203},
  [19779] = {.lex_state = 457},
  [19780] = {.lex_state = 190},
  [19781] = {.lex_state = 457},
  [19782] = {.lex_state = 190},
  [19783] = {.lex_state = 203},
  [19784] = {.lex_state = 203},
  [19785] = {.lex_state = 203},
  [19786] = {.lex_state = 203},
  [19787] = {.lex_state = 457, .external_lex_state = 122},
  [19788] = {.lex_state = 455},
  [19789] = {.lex_state = 203},
  [19790] = {.lex_state = 457},
  [19791] = {.lex_state = 455},
  [19792] = {.lex_state = 457},
  [19793] = {.lex_state = 457, .external_lex_state = 136},
  [19794] = {.lex_state = 455},
  [19795] = {.lex_state = 455, .external_lex_state = 135},
  [19796] = {.lex_state = 457, .external_lex_state = 114},
  [19797] = {.lex_state = 455},
  [19798] = {.lex_state = 190},
  [19799] = {.lex_state = 457},
  [19800] = {.lex_state = 457},
  [19801] = {.lex_state = 455, .external_lex_state = 135},
  [19802] = {.lex_state = 457},
  [19803] = {.lex_state = 190},
  [19804] = {.lex_state = 455},
  [19805] = {.lex_state = 455},
  [19806] = {.lex_state = 457},
  [19807] = {.lex_state = 457, .external_lex_state = 114},
  [19808] = {.lex_state = 457},
  [19809] = {.lex_state = 457, .external_lex_state = 114},
  [19810] = {.lex_state = 457, .external_lex_state = 114},
  [19811] = {.lex_state = 457},
  [19812] = {.lex_state = 455},
  [19813] = {.lex_state = 203},
  [19814] = {.lex_state = 457, .external_lex_state = 114},
  [19815] = {.lex_state = 455, .external_lex_state = 134},
  [19816] = {.lex_state = 457, .external_lex_state = 114},
  [19817] = {.lex_state = 457},
  [19818] = {.lex_state = 457, .external_lex_state = 122},
  [19819] = {.lex_state = 457},
  [19820] = {.lex_state = 457},
  [19821] = {.lex_state = 455},
  [19822] = {.lex_state = 190},
  [19823] = {.lex_state = 455, .external_lex_state = 134},
  [19824] = {.lex_state = 457},
  [19825] = {.lex_state = 457, .external_lex_state = 114},
  [19826] = {.lex_state = 457, .external_lex_state = 114},
  [19827] = {.lex_state = 457, .external_lex_state = 114},
  [19828] = {.lex_state = 457},
  [19829] = {.lex_state = 203},
  [19830] = {.lex_state = 455},
  [19831] = {.lex_state = 457},
  [19832] = {.lex_state = 455, .external_lex_state = 135},
  [19833] = {.lex_state = 203},
  [19834] = {.lex_state = 457},
  [19835] = {.lex_state = 457, .external_lex_state = 114},
  [19836] = {.lex_state = 203},
  [19837] = {.lex_state = 190},
  [19838] = {.lex_state = 457},
  [19839] = {.lex_state = 190},
  [19840] = {.lex_state = 203},
  [19841] = {.lex_state = 457, .external_lex_state = 114},
  [19842] = {.lex_state = 457, .external_lex_state = 114},
  [19843] = {.lex_state = 455, .external_lex_state = 134},
  [19844] = {.lex_state = 455, .external_lex_state = 134},
  [19845] = {.lex_state = 457},
  [19846] = {.lex_state = 203},
  [19847] = {.lex_state = 203},
  [19848] = {.lex_state = 457, .external_lex_state = 114},
  [19849] = {.lex_state = 455, .external_lex_state = 134},
  [19850] = {.lex_state = 203},
  [19851] = {.lex_state = 455, .external_lex_state = 134},
  [19852] = {.lex_state = 1},
  [19853] = {.lex_state = 455},
  [19854] = {.lex_state = 457, .external_lex_state = 136},
  [19855] = {.lex_state = 455},
  [19856] = {.lex_state = 455, .external_lex_state = 135},
  [19857] = {.lex_state = 457, .external_lex_state = 114},
  [19858] = {.lex_state = 457, .external_lex_state = 114},
  [19859] = {.lex_state = 457},
  [19860] = {.lex_state = 457, .external_lex_state = 114},
  [19861] = {.lex_state = 457},
  [19862] = {.lex_state = 457},
  [19863] = {.lex_state = 455, .external_lex_state = 134},
  [19864] = {.lex_state = 457},
  [19865] = {.lex_state = 190},
  [19866] = {.lex_state = 457},
  [19867] = {.lex_state = 455},
  [19868] = {.lex_state = 203},
  [19869] = {.lex_state = 457},
  [19870] = {.lex_state = 457, .external_lex_state = 114},
  [19871] = {.lex_state = 455},
  [19872] = {.lex_state = 457, .external_lex_state = 114},
  [19873] = {.lex_state = 457},
  [19874] = {.lex_state = 455},
  [19875] = {.lex_state = 203},
  [19876] = {.lex_state = 457, .external_lex_state = 131},
  [19877] = {.lex_state = 455},
  [19878] = {.lex_state = 457},
  [19879] = {.lex_state = 457, .external_lex_state = 114},
  [19880] = {.lex_state = 457, .external_lex_state = 122},
  [19881] = {.lex_state = 457, .external_lex_state = 114},
  [19882] = {.lex_state = 457},
  [19883] = {.lex_state = 457},
  [19884] = {.lex_state = 455},
  [19885] = {.lex_state = 203},
  [19886] = {.lex_state = 1},
  [19887] = {.lex_state = 457},
  [19888] = {.lex_state = 457},
  [19889] = {.lex_state = 457, .external_lex_state = 114},
  [19890] = {.lex_state = 457, .external_lex_state = 114},
  [19891] = {.lex_state = 457},
  [19892] = {.lex_state = 203},
  [19893] = {.lex_state = 190},
  [19894] = {.lex_state = 190},
  [19895] = {.lex_state = 455, .external_lex_state = 135},
  [19896] = {.lex_state = 457},
  [19897] = {.lex_state = 190},
  [19898] = {.lex_state = 203},
  [19899] = {.lex_state = 457},
  [19900] = {.lex_state = 203},
  [19901] = {.lex_state = 455},
  [19902] = {.lex_state = 203},
  [19903] = {.lex_state = 203},
  [19904] = {.lex_state = 457, .external_lex_state = 114},
  [19905] = {.lex_state = 203},
  [19906] = {.lex_state = 455},
  [19907] = {.lex_state = 457},
  [19908] = {.lex_state = 457},
  [19909] = {.lex_state = 457, .external_lex_state = 136},
  [19910] = {.lex_state = 455},
  [19911] = {.lex_state = 455, .external_lex_state = 135},
  [19912] = {.lex_state = 190},
  [19913] = {.lex_state = 455},
  [19914] = {.lex_state = 457},
  [19915] = {.lex_state = 457},
  [19916] = {.lex_state = 457},
  [19917] = {.lex_state = 190},
  [19918] = {.lex_state = 455},
  [19919] = {.lex_state = 457},
  [19920] = {.lex_state = 455, .external_lex_state = 134},
  [19921] = {.lex_state = 455, .external_lex_state = 134},
  [19922] = {.lex_state = 455, .external_lex_state = 134},
  [19923] = {.lex_state = 457},
  [19924] = {.lex_state = 457},
  [19925] = {.lex_state = 457, .external_lex_state = 114},
  [19926] = {.lex_state = 457},
  [19927] = {.lex_state = 457, .external_lex_state = 114},
  [19928] = {.lex_state = 455},
  [19929] = {.lex_state = 203},
  [19930] = {.lex_state = 457},
  [19931] = {.lex_state = 457, .external_lex_state = 114},
  [19932] = {.lex_state = 455},
  [19933] = {.lex_state = 455},
  [19934] = {.lex_state = 457, .external_lex_state = 114},
  [19935] = {.lex_state = 455, .external_lex_state = 134},
  [19936] = {.lex_state = 457, .external_lex_state = 114},
  [19937] = {.lex_state = 457},
  [19938] = {.lex_state = 457},
  [19939] = {.lex_state = 455},
  [19940] = {.lex_state = 455},
  [19941] = {.lex_state = 457},
  [19942] = {.lex_state = 457},
  [19943] = {.lex_state = 457},
  [19944] = {.lex_state = 457, .external_lex_state = 114},
  [19945] = {.lex_state = 457, .external_lex_state = 114},
  [19946] = {.lex_state = 457},
  [19947] = {.lex_state = 203},
  [19948] = {.lex_state = 457},
  [19949] = {.lex_state = 203},
  [19950] = {.lex_state = 455, .external_lex_state = 135},
  [19951] = {.lex_state = 457},
  [19952] = {.lex_state = 203},
  [19953] = {.lex_state = 190},
  [19954] = {.lex_state = 457},
  [19955] = {.lex_state = 190},
  [19956] = {.lex_state = 455},
  [19957] = {.lex_state = 457, .external_lex_state = 114},
  [19958] = {.lex_state = 457, .external_lex_state = 136},
  [19959] = {.lex_state = 455},
  [19960] = {.lex_state = 455, .external_lex_state = 135},
  [19961] = {.lex_state = 455, .external_lex_state = 134},
  [19962] = {.lex_state = 203},
  [19963] = {.lex_state = 455, .external_lex_state = 134},
  [19964] = {.lex_state = 455},
  [19965] = {.lex_state = 457},
  [19966] = {.lex_state = 455},
  [19967] = {.lex_state = 455},
  [19968] = {.lex_state = 457},
  [19969] = {.lex_state = 1},
  [19970] = {.lex_state = 455},
  [19971] = {.lex_state = 457, .external_lex_state = 122},
  [19972] = {.lex_state = 455},
  [19973] = {.lex_state = 457},
  [19974] = {.lex_state = 457, .external_lex_state = 114},
  [19975] = {.lex_state = 457, .external_lex_state = 131},
  [19976] = {.lex_state = 457, .external_lex_state = 114},
  [19977] = {.lex_state = 457, .external_lex_state = 114},
  [19978] = {.lex_state = 457, .external_lex_state = 114},
  [19979] = {.lex_state = 457},
  [19980] = {.lex_state = 455},
  [19981] = {.lex_state = 203},
  [19982] = {.lex_state = 457},
  [19983] = {.lex_state = 457, .external_lex_state = 114},
  [19984] = {.lex_state = 457},
  [19985] = {.lex_state = 457},
  [19986] = {.lex_state = 457, .external_lex_state = 114},
  [19987] = {.lex_state = 455},
  [19988] = {.lex_state = 190},
  [19989] = {.lex_state = 455, .external_lex_state = 134},
  [19990] = {.lex_state = 457},
  [19991] = {.lex_state = 455, .external_lex_state = 134},
  [19992] = {.lex_state = 457, .external_lex_state = 114},
  [19993] = {.lex_state = 457, .external_lex_state = 114},
  [19994] = {.lex_state = 457},
  [19995] = {.lex_state = 457},
  [19996] = {.lex_state = 457},
  [19997] = {.lex_state = 455, .external_lex_state = 135},
  [19998] = {.lex_state = 457},
  [19999] = {.lex_state = 457},
  [20000] = {.lex_state = 203},
  [20001] = {.lex_state = 457},
  [20002] = {.lex_state = 457, .external_lex_state = 136},
  [20003] = {.lex_state = 455},
  [20004] = {.lex_state = 455, .external_lex_state = 135},
  [20005] = {.lex_state = 457},
  [20006] = {.lex_state = 190},
  [20007] = {.lex_state = 457},
  [20008] = {.lex_state = 190},
  [20009] = {.lex_state = 457},
  [20010] = {.lex_state = 455, .external_lex_state = 134},
  [20011] = {.lex_state = 457, .external_lex_state = 122},
  [20012] = {.lex_state = 457},
  [20013] = {.lex_state = 457},
  [20014] = {.lex_state = 457, .external_lex_state = 114},
  [20015] = {.lex_state = 455, .external_lex_state = 134},
  [20016] = {.lex_state = 457},
  [20017] = {.lex_state = 457},
  [20018] = {.lex_state = 457, .external_lex_state = 114},
  [20019] = {.lex_state = 457},
  [20020] = {.lex_state = 457, .external_lex_state = 114},
  [20021] = {.lex_state = 455},
  [20022] = {.lex_state = 190},
  [20023] = {.lex_state = 455, .external_lex_state = 134},
  [20024] = {.lex_state = 457, .external_lex_state = 114},
  [20025] = {.lex_state = 457, .external_lex_state = 122},
  [20026] = {.lex_state = 457, .external_lex_state = 114},
  [20027] = {.lex_state = 457, .external_lex_state = 114},
  [20028] = {.lex_state = 457},
  [20029] = {.lex_state = 457},
  [20030] = {.lex_state = 455},
  [20031] = {.lex_state = 455},
  [20032] = {.lex_state = 203},
  [20033] = {.lex_state = 457},
  [20034] = {.lex_state = 203},
  [20035] = {.lex_state = 457, .external_lex_state = 114},
  [20036] = {.lex_state = 457, .external_lex_state = 114},
  [20037] = {.lex_state = 457},
  [20038] = {.lex_state = 457, .external_lex_state = 122},
  [20039] = {.lex_state = 455, .external_lex_state = 134},
  [20040] = {.lex_state = 455, .external_lex_state = 135},
  [20041] = {.lex_state = 457},
  [20042] = {.lex_state = 457},
  [20043] = {.lex_state = 457},
  [20044] = {.lex_state = 457},
  [20045] = {.lex_state = 457, .external_lex_state = 136},
  [20046] = {.lex_state = 455},
  [20047] = {.lex_state = 455, .external_lex_state = 135},
  [20048] = {.lex_state = 455},
  [20049] = {.lex_state = 457, .external_lex_state = 114},
  [20050] = {.lex_state = 190},
  [20051] = {.lex_state = 457},
  [20052] = {.lex_state = 457},
  [20053] = {.lex_state = 457},
  [20054] = {.lex_state = 203},
  [20055] = {.lex_state = 457},
  [20056] = {.lex_state = 457},
  [20057] = {.lex_state = 457},
  [20058] = {.lex_state = 457},
  [20059] = {.lex_state = 190},
  [20060] = {.lex_state = 457},
  [20061] = {.lex_state = 457, .external_lex_state = 114},
  [20062] = {.lex_state = 457},
  [20063] = {.lex_state = 457, .external_lex_state = 114},
  [20064] = {.lex_state = 190},
  [20065] = {.lex_state = 455, .external_lex_state = 134},
  [20066] = {.lex_state = 457},
  [20067] = {.lex_state = 203},
  [20068] = {.lex_state = 457, .external_lex_state = 114},
  [20069] = {.lex_state = 457},
  [20070] = {.lex_state = 457, .external_lex_state = 114},
  [20071] = {.lex_state = 457},
  [20072] = {.lex_state = 457, .external_lex_state = 122},
  [20073] = {.lex_state = 455},
  [20074] = {.lex_state = 455},
  [20075] = {.lex_state = 457},
  [20076] = {.lex_state = 457, .external_lex_state = 128},
  [20077] = {.lex_state = 457},
  [20078] = {.lex_state = 457},
  [20079] = {.lex_state = 457, .external_lex_state = 114},
  [20080] = {.lex_state = 457, .external_lex_state = 114},
  [20081] = {.lex_state = 457},
  [20082] = {.lex_state = 190},
  [20083] = {.lex_state = 457, .external_lex_state = 131},
  [20084] = {.lex_state = 455, .external_lex_state = 135},
  [20085] = {.lex_state = 457, .external_lex_state = 122},
  [20086] = {.lex_state = 457, .external_lex_state = 114},
  [20087] = {.lex_state = 455, .external_lex_state = 134},
  [20088] = {.lex_state = 455},
  [20089] = {.lex_state = 457, .external_lex_state = 136},
  [20090] = {.lex_state = 455},
  [20091] = {.lex_state = 455, .external_lex_state = 135},
  [20092] = {.lex_state = 457},
  [20093] = {.lex_state = 455},
  [20094] = {.lex_state = 203},
  [20095] = {.lex_state = 457, .external_lex_state = 114},
  [20096] = {.lex_state = 457},
  [20097] = {.lex_state = 455, .external_lex_state = 134},
  [20098] = {.lex_state = 455, .external_lex_state = 134},
  [20099] = {.lex_state = 457},
  [20100] = {.lex_state = 457, .external_lex_state = 114},
  [20101] = {.lex_state = 457},
  [20102] = {.lex_state = 457},
  [20103] = {.lex_state = 455, .external_lex_state = 134},
  [20104] = {.lex_state = 457},
  [20105] = {.lex_state = 457, .external_lex_state = 114},
  [20106] = {.lex_state = 455, .external_lex_state = 134},
  [20107] = {.lex_state = 457, .external_lex_state = 114},
  [20108] = {.lex_state = 190},
  [20109] = {.lex_state = 457, .external_lex_state = 114},
  [20110] = {.lex_state = 457},
  [20111] = {.lex_state = 203},
  [20112] = {.lex_state = 457},
  [20113] = {.lex_state = 455, .external_lex_state = 134},
  [20114] = {.lex_state = 457, .external_lex_state = 114},
  [20115] = {.lex_state = 190},
  [20116] = {.lex_state = 457},
  [20117] = {.lex_state = 190},
  [20118] = {.lex_state = 455},
  [20119] = {.lex_state = 455},
  [20120] = {.lex_state = 457, .external_lex_state = 122},
  [20121] = {.lex_state = 457},
  [20122] = {.lex_state = 203},
  [20123] = {.lex_state = 457, .external_lex_state = 114},
  [20124] = {.lex_state = 457, .external_lex_state = 114},
  [20125] = {.lex_state = 457},
  [20126] = {.lex_state = 455},
  [20127] = {.lex_state = 203},
  [20128] = {.lex_state = 455, .external_lex_state = 135},
  [20129] = {.lex_state = 455},
  [20130] = {.lex_state = 457},
  [20131] = {.lex_state = 457},
  [20132] = {.lex_state = 203},
  [20133] = {.lex_state = 457, .external_lex_state = 136},
  [20134] = {.lex_state = 455},
  [20135] = {.lex_state = 455, .external_lex_state = 135},
  [20136] = {.lex_state = 457, .external_lex_state = 122},
  [20137] = {.lex_state = 455},
  [20138] = {.lex_state = 455, .external_lex_state = 134},
  [20139] = {.lex_state = 457},
  [20140] = {.lex_state = 457},
  [20141] = {.lex_state = 457},
  [20142] = {.lex_state = 203},
  [20143] = {.lex_state = 457},
  [20144] = {.lex_state = 457},
  [20145] = {.lex_state = 203},
  [20146] = {.lex_state = 190},
  [20147] = {.lex_state = 457},
  [20148] = {.lex_state = 457},
  [20149] = {.lex_state = 457, .external_lex_state = 114},
  [20150] = {.lex_state = 190},
  [20151] = {.lex_state = 457, .external_lex_state = 114},
  [20152] = {.lex_state = 455},
  [20153] = {.lex_state = 457},
  [20154] = {.lex_state = 457},
  [20155] = {.lex_state = 455},
  [20156] = {.lex_state = 190},
  [20157] = {.lex_state = 455},
  [20158] = {.lex_state = 457, .external_lex_state = 114},
  [20159] = {.lex_state = 457},
  [20160] = {.lex_state = 455},
  [20161] = {.lex_state = 455, .external_lex_state = 134},
  [20162] = {.lex_state = 455},
  [20163] = {.lex_state = 457},
  [20164] = {.lex_state = 457, .external_lex_state = 114},
  [20165] = {.lex_state = 457},
  [20166] = {.lex_state = 457, .external_lex_state = 122},
  [20167] = {.lex_state = 457, .external_lex_state = 114},
  [20168] = {.lex_state = 457, .external_lex_state = 114},
  [20169] = {.lex_state = 457},
  [20170] = {.lex_state = 457, .external_lex_state = 122},
  [20171] = {.lex_state = 457},
  [20172] = {.lex_state = 455, .external_lex_state = 135},
  [20173] = {.lex_state = 457},
  [20174] = {.lex_state = 455},
  [20175] = {.lex_state = 190},
  [20176] = {.lex_state = 457},
  [20177] = {.lex_state = 457, .external_lex_state = 136},
  [20178] = {.lex_state = 455},
  [20179] = {.lex_state = 455, .external_lex_state = 135},
  [20180] = {.lex_state = 190},
  [20181] = {.lex_state = 457, .external_lex_state = 128},
  [20182] = {.lex_state = 455},
  [20183] = {.lex_state = 457},
  [20184] = {.lex_state = 457},
  [20185] = {.lex_state = 455},
  [20186] = {.lex_state = 455},
  [20187] = {.lex_state = 457},
  [20188] = {.lex_state = 457, .external_lex_state = 122},
  [20189] = {.lex_state = 457},
  [20190] = {.lex_state = 203},
  [20191] = {.lex_state = 457, .external_lex_state = 114},
  [20192] = {.lex_state = 457},
  [20193] = {.lex_state = 457, .external_lex_state = 114},
  [20194] = {.lex_state = 455},
  [20195] = {.lex_state = 457, .external_lex_state = 114},
  [20196] = {.lex_state = 190},
  [20197] = {.lex_state = 203},
  [20198] = {.lex_state = 455},
  [20199] = {.lex_state = 457},
  [20200] = {.lex_state = 457},
  [20201] = {.lex_state = 457},
  [20202] = {.lex_state = 457, .external_lex_state = 114},
  [20203] = {.lex_state = 455},
  [20204] = {.lex_state = 190},
  [20205] = {.lex_state = 457},
  [20206] = {.lex_state = 455},
  [20207] = {.lex_state = 190},
  [20208] = {.lex_state = 455},
  [20209] = {.lex_state = 457},
  [20210] = {.lex_state = 455},
  [20211] = {.lex_state = 457, .external_lex_state = 114},
  [20212] = {.lex_state = 457, .external_lex_state = 114},
  [20213] = {.lex_state = 457},
  [20214] = {.lex_state = 455},
  [20215] = {.lex_state = 455},
  [20216] = {.lex_state = 455, .external_lex_state = 135},
  [20217] = {.lex_state = 455},
  [20218] = {.lex_state = 457},
  [20219] = {.lex_state = 455},
  [20220] = {.lex_state = 457},
  [20221] = {.lex_state = 457, .external_lex_state = 136},
  [20222] = {.lex_state = 455},
  [20223] = {.lex_state = 455, .external_lex_state = 135},
  [20224] = {.lex_state = 455, .external_lex_state = 134},
  [20225] = {.lex_state = 457, .external_lex_state = 114},
  [20226] = {.lex_state = 457, .external_lex_state = 114},
  [20227] = {.lex_state = 457, .external_lex_state = 131},
  [20228] = {.lex_state = 457},
  [20229] = {.lex_state = 457},
  [20230] = {.lex_state = 457},
  [20231] = {.lex_state = 457},
  [20232] = {.lex_state = 457},
  [20233] = {.lex_state = 190},
  [20234] = {.lex_state = 457},
  [20235] = {.lex_state = 190},
  [20236] = {.lex_state = 457},
  [20237] = {.lex_state = 457, .external_lex_state = 114},
  [20238] = {.lex_state = 455},
  [20239] = {.lex_state = 457, .external_lex_state = 114},
  [20240] = {.lex_state = 457, .external_lex_state = 122},
  [20241] = {.lex_state = 455},
  [20242] = {.lex_state = 455},
  [20243] = {.lex_state = 457, .external_lex_state = 114},
  [20244] = {.lex_state = 457},
  [20245] = {.lex_state = 455, .external_lex_state = 134},
  [20246] = {.lex_state = 457, .external_lex_state = 114},
  [20247] = {.lex_state = 455, .external_lex_state = 134},
  [20248] = {.lex_state = 457},
  [20249] = {.lex_state = 455, .external_lex_state = 134},
  [20250] = {.lex_state = 455},
  [20251] = {.lex_state = 455},
  [20252] = {.lex_state = 457},
  [20253] = {.lex_state = 457},
  [20254] = {.lex_state = 457},
  [20255] = {.lex_state = 457, .external_lex_state = 114},
  [20256] = {.lex_state = 457, .external_lex_state = 114},
  [20257] = {.lex_state = 457},
  [20258] = {.lex_state = 457},
  [20259] = {.lex_state = 457},
  [20260] = {.lex_state = 455, .external_lex_state = 135},
  [20261] = {.lex_state = 1},
  [20262] = {.lex_state = 190},
  [20263] = {.lex_state = 457},
  [20264] = {.lex_state = 190},
  [20265] = {.lex_state = 457, .external_lex_state = 136},
  [20266] = {.lex_state = 455},
  [20267] = {.lex_state = 455, .external_lex_state = 135},
  [20268] = {.lex_state = 455},
  [20269] = {.lex_state = 455, .external_lex_state = 134},
  [20270] = {.lex_state = 455, .external_lex_state = 134},
  [20271] = {.lex_state = 455},
  [20272] = {.lex_state = 457},
  [20273] = {.lex_state = 457},
  [20274] = {.lex_state = 455, .external_lex_state = 134},
  [20275] = {.lex_state = 457},
  [20276] = {.lex_state = 457},
  [20277] = {.lex_state = 457},
  [20278] = {.lex_state = 457},
  [20279] = {.lex_state = 190},
  [20280] = {.lex_state = 457},
  [20281] = {.lex_state = 457, .external_lex_state = 114},
  [20282] = {.lex_state = 457, .external_lex_state = 114},
  [20283] = {.lex_state = 457, .external_lex_state = 114},
  [20284] = {.lex_state = 457, .external_lex_state = 114},
  [20285] = {.lex_state = 457},
  [20286] = {.lex_state = 457},
  [20287] = {.lex_state = 203},
  [20288] = {.lex_state = 190},
  [20289] = {.lex_state = 457},
  [20290] = {.lex_state = 457, .external_lex_state = 114},
  [20291] = {.lex_state = 190},
  [20292] = {.lex_state = 455},
  [20293] = {.lex_state = 214},
  [20294] = {.lex_state = 455},
  [20295] = {.lex_state = 457, .external_lex_state = 114},
  [20296] = {.lex_state = 455},
  [20297] = {.lex_state = 457},
  [20298] = {.lex_state = 457},
  [20299] = {.lex_state = 457, .external_lex_state = 114},
  [20300] = {.lex_state = 457, .external_lex_state = 114},
  [20301] = {.lex_state = 457},
  [20302] = {.lex_state = 203},
  [20303] = {.lex_state = 457, .external_lex_state = 122},
  [20304] = {.lex_state = 455, .external_lex_state = 135},
  [20305] = {.lex_state = 455},
  [20306] = {.lex_state = 457},
  [20307] = {.lex_state = 457},
  [20308] = {.lex_state = 455, .external_lex_state = 135},
  [20309] = {.lex_state = 457, .external_lex_state = 136},
  [20310] = {.lex_state = 455},
  [20311] = {.lex_state = 455, .external_lex_state = 135},
  [20312] = {.lex_state = 457},
  [20313] = {.lex_state = 457},
  [20314] = {.lex_state = 457},
  [20315] = {.lex_state = 455, .external_lex_state = 134},
  [20316] = {.lex_state = 457},
  [20317] = {.lex_state = 190},
  [20318] = {.lex_state = 457},
  [20319] = {.lex_state = 457},
  [20320] = {.lex_state = 190},
  [20321] = {.lex_state = 455},
  [20322] = {.lex_state = 190},
  [20323] = {.lex_state = 455, .external_lex_state = 134},
  [20324] = {.lex_state = 457},
  [20325] = {.lex_state = 457, .external_lex_state = 114},
  [20326] = {.lex_state = 455},
  [20327] = {.lex_state = 457, .external_lex_state = 114},
  [20328] = {.lex_state = 457},
  [20329] = {.lex_state = 190},
  [20330] = {.lex_state = 457},
  [20331] = {.lex_state = 457},
  [20332] = {.lex_state = 457, .external_lex_state = 114},
  [20333] = {.lex_state = 1},
  [20334] = {.lex_state = 457, .external_lex_state = 114},
  [20335] = {.lex_state = 457},
  [20336] = {.lex_state = 457, .external_lex_state = 122},
  [20337] = {.lex_state = 457},
  [20338] = {.lex_state = 455},
  [20339] = {.lex_state = 457},
  [20340] = {.lex_state = 457},
  [20341] = {.lex_state = 457},
  [20342] = {.lex_state = 190},
  [20343] = {.lex_state = 457, .external_lex_state = 114},
  [20344] = {.lex_state = 457, .external_lex_state = 114},
  [20345] = {.lex_state = 457},
  [20346] = {.lex_state = 457},
  [20347] = {.lex_state = 190},
  [20348] = {.lex_state = 455, .external_lex_state = 135},
  [20349] = {.lex_state = 455},
  [20350] = {.lex_state = 457},
  [20351] = {.lex_state = 190},
  [20352] = {.lex_state = 455},
  [20353] = {.lex_state = 457, .external_lex_state = 136},
  [20354] = {.lex_state = 455},
  [20355] = {.lex_state = 455, .external_lex_state = 135},
  [20356] = {.lex_state = 457, .external_lex_state = 114},
  [20357] = {.lex_state = 457, .external_lex_state = 114},
  [20358] = {.lex_state = 457},
  [20359] = {.lex_state = 1},
  [20360] = {.lex_state = 457},
  [20361] = {.lex_state = 455},
  [20362] = {.lex_state = 455},
  [20363] = {.lex_state = 457},
  [20364] = {.lex_state = 457},
  [20365] = {.lex_state = 457},
  [20366] = {.lex_state = 457},
  [20367] = {.lex_state = 455},
  [20368] = {.lex_state = 457},
  [20369] = {.lex_state = 457, .external_lex_state = 114},
  [20370] = {.lex_state = 190},
  [20371] = {.lex_state = 457, .external_lex_state = 114},
  [20372] = {.lex_state = 457},
  [20373] = {.lex_state = 190},
  [20374] = {.lex_state = 455},
  [20375] = {.lex_state = 457, .external_lex_state = 114},
  [20376] = {.lex_state = 457},
  [20377] = {.lex_state = 455},
  [20378] = {.lex_state = 457, .external_lex_state = 114},
  [20379] = {.lex_state = 457, .external_lex_state = 122},
  [20380] = {.lex_state = 457},
  [20381] = {.lex_state = 1},
  [20382] = {.lex_state = 455},
  [20383] = {.lex_state = 203},
  [20384] = {.lex_state = 1},
  [20385] = {.lex_state = 457},
  [20386] = {.lex_state = 455},
  [20387] = {.lex_state = 457, .external_lex_state = 114},
  [20388] = {.lex_state = 457, .external_lex_state = 114},
  [20389] = {.lex_state = 457},
  [20390] = {.lex_state = 455},
  [20391] = {.lex_state = 455},
  [20392] = {.lex_state = 455, .external_lex_state = 135},
  [20393] = {.lex_state = 457},
  [20394] = {.lex_state = 457},
  [20395] = {.lex_state = 455},
  [20396] = {.lex_state = 190},
  [20397] = {.lex_state = 457, .external_lex_state = 136},
  [20398] = {.lex_state = 455},
  [20399] = {.lex_state = 455, .external_lex_state = 135},
  [20400] = {.lex_state = 457},
  [20401] = {.lex_state = 190},
  [20402] = {.lex_state = 455},
  [20403] = {.lex_state = 457},
  [20404] = {.lex_state = 457},
  [20405] = {.lex_state = 457},
  [20406] = {.lex_state = 455},
  [20407] = {.lex_state = 457},
  [20408] = {.lex_state = 457, .external_lex_state = 114},
  [20409] = {.lex_state = 455},
  [20410] = {.lex_state = 457, .external_lex_state = 114},
  [20411] = {.lex_state = 457},
  [20412] = {.lex_state = 457},
  [20413] = {.lex_state = 457, .external_lex_state = 114},
  [20414] = {.lex_state = 455},
  [20415] = {.lex_state = 457, .external_lex_state = 114},
  [20416] = {.lex_state = 457},
  [20417] = {.lex_state = 457},
  [20418] = {.lex_state = 190},
  [20419] = {.lex_state = 457},
  [20420] = {.lex_state = 457},
  [20421] = {.lex_state = 457},
  [20422] = {.lex_state = 457, .external_lex_state = 114},
  [20423] = {.lex_state = 190},
  [20424] = {.lex_state = 457},
  [20425] = {.lex_state = 457, .external_lex_state = 131},
  [20426] = {.lex_state = 455},
  [20427] = {.lex_state = 190},
  [20428] = {.lex_state = 455},
  [20429] = {.lex_state = 457},
  [20430] = {.lex_state = 457},
  [20431] = {.lex_state = 457, .external_lex_state = 114},
  [20432] = {.lex_state = 457, .external_lex_state = 114},
  [20433] = {.lex_state = 457},
  [20434] = {.lex_state = 457, .external_lex_state = 122},
  [20435] = {.lex_state = 455},
  [20436] = {.lex_state = 455, .external_lex_state = 135},
  [20437] = {.lex_state = 457},
  [20438] = {.lex_state = 457},
  [20439] = {.lex_state = 190},
  [20440] = {.lex_state = 190},
  [20441] = {.lex_state = 457, .external_lex_state = 136},
  [20442] = {.lex_state = 455},
  [20443] = {.lex_state = 455, .external_lex_state = 135},
  [20444] = {.lex_state = 457, .external_lex_state = 122},
  [20445] = {.lex_state = 455, .external_lex_state = 134},
  [20446] = {.lex_state = 457},
  [20447] = {.lex_state = 455},
  [20448] = {.lex_state = 457},
  [20449] = {.lex_state = 457},
  [20450] = {.lex_state = 457},
  [20451] = {.lex_state = 457},
  [20452] = {.lex_state = 457},
  [20453] = {.lex_state = 190},
  [20454] = {.lex_state = 457},
  [20455] = {.lex_state = 190},
  [20456] = {.lex_state = 457},
  [20457] = {.lex_state = 457, .external_lex_state = 114},
  [20458] = {.lex_state = 455},
  [20459] = {.lex_state = 457, .external_lex_state = 114},
  [20460] = {.lex_state = 203},
  [20461] = {.lex_state = 455, .external_lex_state = 134},
  [20462] = {.lex_state = 455},
  [20463] = {.lex_state = 455, .external_lex_state = 134},
  [20464] = {.lex_state = 457, .external_lex_state = 114},
  [20465] = {.lex_state = 457},
  [20466] = {.lex_state = 457, .external_lex_state = 114},
  [20467] = {.lex_state = 455},
  [20468] = {.lex_state = 457, .external_lex_state = 114},
  [20469] = {.lex_state = 455, .external_lex_state = 134},
  [20470] = {.lex_state = 455},
  [20471] = {.lex_state = 457, .external_lex_state = 122},
  [20472] = {.lex_state = 457},
  [20473] = {.lex_state = 457},
  [20474] = {.lex_state = 457},
  [20475] = {.lex_state = 457, .external_lex_state = 114},
  [20476] = {.lex_state = 457, .external_lex_state = 114},
  [20477] = {.lex_state = 457},
  [20478] = {.lex_state = 457, .external_lex_state = 131},
  [20479] = {.lex_state = 457},
  [20480] = {.lex_state = 455, .external_lex_state = 135},
  [20481] = {.lex_state = 455},
  [20482] = {.lex_state = 457, .external_lex_state = 136},
  [20483] = {.lex_state = 455},
  [20484] = {.lex_state = 190},
  [20485] = {.lex_state = 457},
  [20486] = {.lex_state = 190},
  [20487] = {.lex_state = 457},
  [20488] = {.lex_state = 455},
  [20489] = {.lex_state = 457},
  [20490] = {.lex_state = 455},
  [20491] = {.lex_state = 457, .external_lex_state = 114},
  [20492] = {.lex_state = 455},
  [20493] = {.lex_state = 457, .external_lex_state = 114},
  [20494] = {.lex_state = 203},
  [20495] = {.lex_state = 457},
  [20496] = {.lex_state = 455},
  [20497] = {.lex_state = 457},
  [20498] = {.lex_state = 455},
  [20499] = {.lex_state = 457, .external_lex_state = 114},
  [20500] = {.lex_state = 457},
  [20501] = {.lex_state = 457, .external_lex_state = 122},
  [20502] = {.lex_state = 457, .external_lex_state = 136},
  [20503] = {.lex_state = 455},
  [20504] = {.lex_state = 203},
  [20505] = {.lex_state = 455},
  [20506] = {.lex_state = 457, .external_lex_state = 114},
  [20507] = {.lex_state = 457},
  [20508] = {.lex_state = 457},
  [20509] = {.lex_state = 457},
  [20510] = {.lex_state = 457},
  [20511] = {.lex_state = 455},
  [20512] = {.lex_state = 457, .external_lex_state = 114},
  [20513] = {.lex_state = 190},
  [20514] = {.lex_state = 457},
  [20515] = {.lex_state = 190},
  [20516] = {.lex_state = 455},
  [20517] = {.lex_state = 457, .external_lex_state = 114},
  [20518] = {.lex_state = 457},
  [20519] = {.lex_state = 457, .external_lex_state = 136},
  [20520] = {.lex_state = 455},
  [20521] = {.lex_state = 457, .external_lex_state = 114},
  [20522] = {.lex_state = 457, .external_lex_state = 114},
  [20523] = {.lex_state = 455},
  [20524] = {.lex_state = 457},
  [20525] = {.lex_state = 457, .external_lex_state = 114},
  [20526] = {.lex_state = 457},
  [20527] = {.lex_state = 457},
  [20528] = {.lex_state = 455},
  [20529] = {.lex_state = 457, .external_lex_state = 114},
  [20530] = {.lex_state = 457},
  [20531] = {.lex_state = 457},
  [20532] = {.lex_state = 455, .external_lex_state = 134},
  [20533] = {.lex_state = 455},
  [20534] = {.lex_state = 457, .external_lex_state = 114},
  [20535] = {.lex_state = 457},
  [20536] = {.lex_state = 457, .external_lex_state = 136},
  [20537] = {.lex_state = 455},
  [20538] = {.lex_state = 457},
  [20539] = {.lex_state = 457},
  [20540] = {.lex_state = 457},
  [20541] = {.lex_state = 457},
  [20542] = {.lex_state = 455, .external_lex_state = 134},
  [20543] = {.lex_state = 457},
  [20544] = {.lex_state = 190},
  [20545] = {.lex_state = 457},
  [20546] = {.lex_state = 457, .external_lex_state = 114},
  [20547] = {.lex_state = 190},
  [20548] = {.lex_state = 455},
  [20549] = {.lex_state = 455},
  [20550] = {.lex_state = 457},
  [20551] = {.lex_state = 457, .external_lex_state = 114},
  [20552] = {.lex_state = 457},
  [20553] = {.lex_state = 457, .external_lex_state = 136},
  [20554] = {.lex_state = 455},
  [20555] = {.lex_state = 455},
  [20556] = {.lex_state = 455},
  [20557] = {.lex_state = 455},
  [20558] = {.lex_state = 457},
  [20559] = {.lex_state = 190},
  [20560] = {.lex_state = 457},
  [20561] = {.lex_state = 455},
  [20562] = {.lex_state = 457},
  [20563] = {.lex_state = 457, .external_lex_state = 114},
  [20564] = {.lex_state = 455},
  [20565] = {.lex_state = 457},
  [20566] = {.lex_state = 457},
  [20567] = {.lex_state = 457, .external_lex_state = 122},
  [20568] = {.lex_state = 457, .external_lex_state = 114},
  [20569] = {.lex_state = 457},
  [20570] = {.lex_state = 457, .external_lex_state = 136},
  [20571] = {.lex_state = 455},
  [20572] = {.lex_state = 457},
  [20573] = {.lex_state = 457},
  [20574] = {.lex_state = 457},
  [20575] = {.lex_state = 457},
  [20576] = {.lex_state = 190},
  [20577] = {.lex_state = 457},
  [20578] = {.lex_state = 457},
  [20579] = {.lex_state = 190},
  [20580] = {.lex_state = 457, .external_lex_state = 114},
  [20581] = {.lex_state = 455},
  [20582] = {.lex_state = 203},
  [20583] = {.lex_state = 457, .external_lex_state = 122},
  [20584] = {.lex_state = 457, .external_lex_state = 114},
  [20585] = {.lex_state = 457},
  [20586] = {.lex_state = 457, .external_lex_state = 136},
  [20587] = {.lex_state = 455},
  [20588] = {.lex_state = 457, .external_lex_state = 114},
  [20589] = {.lex_state = 455},
  [20590] = {.lex_state = 457, .external_lex_state = 122},
  [20591] = {.lex_state = 457},
  [20592] = {.lex_state = 457},
  [20593] = {.lex_state = 457},
  [20594] = {.lex_state = 455, .external_lex_state = 134},
  [20595] = {.lex_state = 455, .external_lex_state = 134},
  [20596] = {.lex_state = 457, .external_lex_state = 114},
  [20597] = {.lex_state = 457},
  [20598] = {.lex_state = 457},
  [20599] = {.lex_state = 455, .external_lex_state = 134},
  [20600] = {.lex_state = 457, .external_lex_state = 114},
  [20601] = {.lex_state = 457},
  [20602] = {.lex_state = 457, .external_lex_state = 136},
  [20603] = {.lex_state = 455},
  [20604] = {.lex_state = 190},
  [20605] = {.lex_state = 457},
  [20606] = {.lex_state = 457},
  [20607] = {.lex_state = 457},
  [20608] = {.lex_state = 457},
  [20609] = {.lex_state = 190},
  [20610] = {.lex_state = 457, .external_lex_state = 114},
  [20611] = {.lex_state = 457},
  [20612] = {.lex_state = 190},
  [20613] = {.lex_state = 455},
  [20614] = {.lex_state = 457, .external_lex_state = 114},
  [20615] = {.lex_state = 457},
  [20616] = {.lex_state = 457, .external_lex_state = 136},
  [20617] = {.lex_state = 455},
  [20618] = {.lex_state = 190},
  [20619] = {.lex_state = 455, .external_lex_state = 134},
  [20620] = {.lex_state = 455},
  [20621] = {.lex_state = 457, .external_lex_state = 114},
  [20622] = {.lex_state = 457},
  [20623] = {.lex_state = 457},
  [20624] = {.lex_state = 457, .external_lex_state = 114},
  [20625] = {.lex_state = 457, .external_lex_state = 114},
  [20626] = {.lex_state = 457},
  [20627] = {.lex_state = 455, .external_lex_state = 134},
  [20628] = {.lex_state = 457, .external_lex_state = 114},
  [20629] = {.lex_state = 457},
  [20630] = {.lex_state = 457, .external_lex_state = 136},
  [20631] = {.lex_state = 457},
  [20632] = {.lex_state = 457, .external_lex_state = 122},
  [20633] = {.lex_state = 457},
  [20634] = {.lex_state = 457},
  [20635] = {.lex_state = 457},
  [20636] = {.lex_state = 457, .external_lex_state = 114},
  [20637] = {.lex_state = 457},
  [20638] = {.lex_state = 457},
  [20639] = {.lex_state = 457, .external_lex_state = 114},
  [20640] = {.lex_state = 457},
  [20641] = {.lex_state = 457, .external_lex_state = 136},
  [20642] = {.lex_state = 190},
  [20643] = {.lex_state = 457},
  [20644] = {.lex_state = 190},
  [20645] = {.lex_state = 455},
  [20646] = {.lex_state = 457},
  [20647] = {.lex_state = 457, .external_lex_state = 114},
  [20648] = {.lex_state = 190},
  [20649] = {.lex_state = 455},
  [20650] = {.lex_state = 457, .external_lex_state = 114},
  [20651] = {.lex_state = 457},
  [20652] = {.lex_state = 457, .external_lex_state = 136},
  [20653] = {.lex_state = 455},
  [20654] = {.lex_state = 457, .external_lex_state = 114},
  [20655] = {.lex_state = 457},
  [20656] = {.lex_state = 457, .external_lex_state = 114},
  [20657] = {.lex_state = 190},
  [20658] = {.lex_state = 457, .external_lex_state = 114},
  [20659] = {.lex_state = 457, .external_lex_state = 136},
  [20660] = {.lex_state = 455, .external_lex_state = 134},
  [20661] = {.lex_state = 457},
  [20662] = {.lex_state = 457},
  [20663] = {.lex_state = 457, .external_lex_state = 114},
  [20664] = {.lex_state = 457, .external_lex_state = 114},
  [20665] = {.lex_state = 457, .external_lex_state = 136},
  [20666] = {.lex_state = 457},
  [20667] = {.lex_state = 457, .external_lex_state = 114},
  [20668] = {.lex_state = 457},
  [20669] = {.lex_state = 457, .external_lex_state = 114},
  [20670] = {.lex_state = 457, .external_lex_state = 114},
  [20671] = {.lex_state = 457, .external_lex_state = 136},
  [20672] = {.lex_state = 457},
  [20673] = {.lex_state = 457, .external_lex_state = 131},
  [20674] = {.lex_state = 455, .external_lex_state = 134},
  [20675] = {.lex_state = 457, .external_lex_state = 114},
  [20676] = {.lex_state = 457, .external_lex_state = 114},
  [20677] = {.lex_state = 457, .external_lex_state = 136},
  [20678] = {.lex_state = 190},
  [20679] = {.lex_state = 457},
  [20680] = {.lex_state = 190},
  [20681] = {.lex_state = 457, .external_lex_state = 114},
  [20682] = {.lex_state = 457, .external_lex_state = 114},
  [20683] = {.lex_state = 457, .external_lex_state = 136},
  [20684] = {.lex_state = 455},
  [20685] = {.lex_state = 457, .external_lex_state = 114},
  [20686] = {.lex_state = 457},
  [20687] = {.lex_state = 457, .external_lex_state = 114},
  [20688] = {.lex_state = 457, .external_lex_state = 114},
  [20689] = {.lex_state = 457, .external_lex_state = 136},
  [20690] = {.lex_state = 455},
  [20691] = {.lex_state = 457},
  [20692] = {.lex_state = 457},
  [20693] = {.lex_state = 457, .external_lex_state = 114},
  [20694] = {.lex_state = 457, .external_lex_state = 114},
  [20695] = {.lex_state = 457, .external_lex_state = 136},
  [20696] = {.lex_state = 457},
  [20697] = {.lex_state = 455},
  [20698] = {.lex_state = 457},
  [20699] = {.lex_state = 457, .external_lex_state = 114},
  [20700] = {.lex_state = 457, .external_lex_state = 114},
  [20701] = {.lex_state = 457, .external_lex_state = 136},
  [20702] = {.lex_state = 457},
  [20703] = {.lex_state = 457},
  [20704] = {.lex_state = 455, .external_lex_state = 134},
  [20705] = {.lex_state = 457, .external_lex_state = 114},
  [20706] = {.lex_state = 457, .external_lex_state = 114},
  [20707] = {.lex_state = 457, .external_lex_state = 136},
  [20708] = {.lex_state = 457},
  [20709] = {.lex_state = 457},
  [20710] = {.lex_state = 455},
  [20711] = {.lex_state = 457, .external_lex_state = 114},
  [20712] = {.lex_state = 457, .external_lex_state = 114},
  [20713] = {.lex_state = 457, .external_lex_state = 136},
  [20714] = {.lex_state = 190},
  [20715] = {.lex_state = 457, .external_lex_state = 114},
  [20716] = {.lex_state = 457},
  [20717] = {.lex_state = 457, .external_lex_state = 114},
  [20718] = {.lex_state = 457, .external_lex_state = 114},
  [20719] = {.lex_state = 190},
  [20720] = {.lex_state = 455},
  [20721] = {.lex_state = 457},
  [20722] = {.lex_state = 457, .external_lex_state = 114},
  [20723] = {.lex_state = 457, .external_lex_state = 114},
  [20724] = {.lex_state = 203},
  [20725] = {.lex_state = 457},
  [20726] = {.lex_state = 455},
  [20727] = {.lex_state = 457, .external_lex_state = 114},
  [20728] = {.lex_state = 457, .external_lex_state = 114},
  [20729] = {.lex_state = 457},
  [20730] = {.lex_state = 457},
  [20731] = {.lex_state = 457},
  [20732] = {.lex_state = 457, .external_lex_state = 114},
  [20733] = {.lex_state = 457, .external_lex_state = 114},
  [20734] = {.lex_state = 457, .external_lex_state = 114},
  [20735] = {.lex_state = 190},
  [20736] = {.lex_state = 190},
  [20737] = {.lex_state = 457, .external_lex_state = 114},
  [20738] = {.lex_state = 457, .external_lex_state = 114},
  [20739] = {.lex_state = 190},
  [20740] = {.lex_state = 455},
  [20741] = {.lex_state = 457},
  [20742] = {.lex_state = 457, .external_lex_state = 114},
  [20743] = {.lex_state = 457, .external_lex_state = 114},
  [20744] = {.lex_state = 457},
  [20745] = {.lex_state = 455},
  [20746] = {.lex_state = 190},
  [20747] = {.lex_state = 457, .external_lex_state = 114},
  [20748] = {.lex_state = 457, .external_lex_state = 114},
  [20749] = {.lex_state = 457},
  [20750] = {.lex_state = 190},
  [20751] = {.lex_state = 455},
  [20752] = {.lex_state = 457, .external_lex_state = 114},
  [20753] = {.lex_state = 457, .external_lex_state = 114},
  [20754] = {.lex_state = 457},
  [20755] = {.lex_state = 457},
  [20756] = {.lex_state = 455},
  [20757] = {.lex_state = 457, .external_lex_state = 114},
  [20758] = {.lex_state = 457, .external_lex_state = 114},
  [20759] = {.lex_state = 455},
  [20760] = {.lex_state = 457, .external_lex_state = 122},
  [20761] = {.lex_state = 455},
  [20762] = {.lex_state = 457, .external_lex_state = 114},
  [20763] = {.lex_state = 457, .external_lex_state = 114},
  [20764] = {.lex_state = 457, .external_lex_state = 114},
  [20765] = {.lex_state = 455},
  [20766] = {.lex_state = 457},
  [20767] = {.lex_state = 457, .external_lex_state = 114},
  [20768] = {.lex_state = 457, .external_lex_state = 114},
  [20769] = {.lex_state = 455},
  [20770] = {.lex_state = 455},
  [20771] = {.lex_state = 203},
  [20772] = {.lex_state = 457, .external_lex_state = 114},
  [20773] = {.lex_state = 457, .external_lex_state = 114},
  [20774] = {.lex_state = 457},
  [20775] = {.lex_state = 457},
  [20776] = {.lex_state = 457},
  [20777] = {.lex_state = 457, .external_lex_state = 114},
  [20778] = {.lex_state = 457, .external_lex_state = 114},
  [20779] = {.lex_state = 190},
  [20780] = {.lex_state = 457},
  [20781] = {.lex_state = 190},
  [20782] = {.lex_state = 457, .external_lex_state = 114},
  [20783] = {.lex_state = 457, .external_lex_state = 114},
  [20784] = {.lex_state = 455},
  [20785] = {.lex_state = 455},
  [20786] = {.lex_state = 457, .external_lex_state = 114},
  [20787] = {.lex_state = 457, .external_lex_state = 114},
  [20788] = {.lex_state = 457, .external_lex_state = 114},
  [20789] = {.lex_state = 455},
  [20790] = {.lex_state = 203},
  [20791] = {.lex_state = 457, .external_lex_state = 114},
  [20792] = {.lex_state = 457, .external_lex_state = 114},
  [20793] = {.lex_state = 457, .external_lex_state = 114},
  [20794] = {.lex_state = 455},
  [20795] = {.lex_state = 457},
  [20796] = {.lex_state = 457},
  [20797] = {.lex_state = 457, .external_lex_state = 114},
  [20798] = {.lex_state = 457, .external_lex_state = 114},
  [20799] = {.lex_state = 455},
  [20800] = {.lex_state = 190},
  [20801] = {.lex_state = 455},
  [20802] = {.lex_state = 457, .external_lex_state = 114},
  [20803] = {.lex_state = 457, .external_lex_state = 114},
  [20804] = {.lex_state = 457},
  [20805] = {.lex_state = 457},
  [20806] = {.lex_state = 455},
  [20807] = {.lex_state = 457, .external_lex_state = 114},
  [20808] = {.lex_state = 457, .external_lex_state = 114},
  [20809] = {.lex_state = 190},
  [20810] = {.lex_state = 457},
  [20811] = {.lex_state = 190},
  [20812] = {.lex_state = 457, .external_lex_state = 114},
  [20813] = {.lex_state = 457, .external_lex_state = 114},
  [20814] = {.lex_state = 455},
  [20815] = {.lex_state = 457, .external_lex_state = 114},
  [20816] = {.lex_state = 455},
  [20817] = {.lex_state = 457, .external_lex_state = 114},
  [20818] = {.lex_state = 457, .external_lex_state = 114},
  [20819] = {.lex_state = 455},
  [20820] = {.lex_state = 455},
  [20821] = {.lex_state = 457},
  [20822] = {.lex_state = 457, .external_lex_state = 114},
  [20823] = {.lex_state = 457, .external_lex_state = 114},
  [20824] = {.lex_state = 203},
  [20825] = {.lex_state = 457, .external_lex_state = 114},
  [20826] = {.lex_state = 457, .external_lex_state = 122},
  [20827] = {.lex_state = 457, .external_lex_state = 114},
  [20828] = {.lex_state = 457, .external_lex_state = 114},
  [20829] = {.lex_state = 455},
  [20830] = {.lex_state = 457},
  [20831] = {.lex_state = 457},
  [20832] = {.lex_state = 457, .external_lex_state = 114},
  [20833] = {.lex_state = 457, .external_lex_state = 114},
  [20834] = {.lex_state = 457},
  [20835] = {.lex_state = 457},
  [20836] = {.lex_state = 457, .external_lex_state = 122},
  [20837] = {.lex_state = 457, .external_lex_state = 114},
  [20838] = {.lex_state = 457, .external_lex_state = 114},
  [20839] = {.lex_state = 190},
  [20840] = {.lex_state = 457},
  [20841] = {.lex_state = 190},
  [20842] = {.lex_state = 457, .external_lex_state = 114},
  [20843] = {.lex_state = 457, .external_lex_state = 114},
  [20844] = {.lex_state = 455},
  [20845] = {.lex_state = 455, .external_lex_state = 135},
  [20846] = {.lex_state = 457},
  [20847] = {.lex_state = 457, .external_lex_state = 114},
  [20848] = {.lex_state = 457, .external_lex_state = 114},
  [20849] = {.lex_state = 457},
  [20850] = {.lex_state = 455},
  [20851] = {.lex_state = 455, .external_lex_state = 134},
  [20852] = {.lex_state = 457, .external_lex_state = 114},
  [20853] = {.lex_state = 457, .external_lex_state = 114},
  [20854] = {.lex_state = 457, .external_lex_state = 114},
  [20855] = {.lex_state = 457, .external_lex_state = 114},
  [20856] = {.lex_state = 457},
  [20857] = {.lex_state = 457, .external_lex_state = 114},
  [20858] = {.lex_state = 457, .external_lex_state = 114},
  [20859] = {.lex_state = 457},
  [20860] = {.lex_state = 457},
  [20861] = {.lex_state = 457, .external_lex_state = 114},
  [20862] = {.lex_state = 457, .external_lex_state = 114},
  [20863] = {.lex_state = 457, .external_lex_state = 114},
  [20864] = {.lex_state = 455, .external_lex_state = 134},
  [20865] = {.lex_state = 457},
  [20866] = {.lex_state = 457},
  [20867] = {.lex_state = 457, .external_lex_state = 114},
  [20868] = {.lex_state = 457, .external_lex_state = 114},
  [20869] = {.lex_state = 457, .external_lex_state = 114},
  [20870] = {.lex_state = 457},
  [20871] = {.lex_state = 190},
  [20872] = {.lex_state = 457, .external_lex_state = 114},
  [20873] = {.lex_state = 457, .external_lex_state = 114},
  [20874] = {.lex_state = 190},
  [20875] = {.lex_state = 457},
  [20876] = {.lex_state = 190},
  [20877] = {.lex_state = 457, .external_lex_state = 114},
  [20878] = {.lex_state = 457, .external_lex_state = 114},
  [20879] = {.lex_state = 455},
  [20880] = {.lex_state = 455},
  [20881] = {.lex_state = 457},
  [20882] = {.lex_state = 457, .external_lex_state = 114},
  [20883] = {.lex_state = 457, .external_lex_state = 114},
  [20884] = {.lex_state = 455},
  [20885] = {.lex_state = 457},
  [20886] = {.lex_state = 457},
  [20887] = {.lex_state = 457, .external_lex_state = 114},
  [20888] = {.lex_state = 457, .external_lex_state = 114},
  [20889] = {.lex_state = 457},
  [20890] = {.lex_state = 457},
  [20891] = {.lex_state = 203},
  [20892] = {.lex_state = 457, .external_lex_state = 114},
  [20893] = {.lex_state = 457, .external_lex_state = 114},
  [20894] = {.lex_state = 457, .external_lex_state = 122},
  [20895] = {.lex_state = 457},
  [20896] = {.lex_state = 457},
  [20897] = {.lex_state = 457, .external_lex_state = 114},
  [20898] = {.lex_state = 457, .external_lex_state = 114},
  [20899] = {.lex_state = 457},
  [20900] = {.lex_state = 457},
  [20901] = {.lex_state = 457},
  [20902] = {.lex_state = 457, .external_lex_state = 114},
  [20903] = {.lex_state = 457, .external_lex_state = 114},
  [20904] = {.lex_state = 190},
  [20905] = {.lex_state = 457},
  [20906] = {.lex_state = 190},
  [20907] = {.lex_state = 457, .external_lex_state = 114},
  [20908] = {.lex_state = 457, .external_lex_state = 114},
  [20909] = {.lex_state = 455},
  [20910] = {.lex_state = 457},
  [20911] = {.lex_state = 455, .external_lex_state = 134},
  [20912] = {.lex_state = 457, .external_lex_state = 114},
  [20913] = {.lex_state = 457, .external_lex_state = 114},
  [20914] = {.lex_state = 455},
  [20915] = {.lex_state = 190},
  [20916] = {.lex_state = 455, .external_lex_state = 134},
  [20917] = {.lex_state = 457, .external_lex_state = 114},
  [20918] = {.lex_state = 457, .external_lex_state = 114},
  [20919] = {.lex_state = 457},
  [20920] = {.lex_state = 457},
  [20921] = {.lex_state = 203},
  [20922] = {.lex_state = 457, .external_lex_state = 114},
  [20923] = {.lex_state = 457, .external_lex_state = 114},
  [20924] = {.lex_state = 457},
  [20925] = {.lex_state = 457},
  [20926] = {.lex_state = 457},
  [20927] = {.lex_state = 457, .external_lex_state = 114},
  [20928] = {.lex_state = 457, .external_lex_state = 114},
  [20929] = {.lex_state = 457},
  [20930] = {.lex_state = 457, .external_lex_state = 114},
  [20931] = {.lex_state = 190},
  [20932] = {.lex_state = 457, .external_lex_state = 114},
  [20933] = {.lex_state = 457, .external_lex_state = 114},
  [20934] = {.lex_state = 457},
  [20935] = {.lex_state = 190},
  [20936] = {.lex_state = 455},
  [20937] = {.lex_state = 457, .external_lex_state = 114},
  [20938] = {.lex_state = 457, .external_lex_state = 114},
  [20939] = {.lex_state = 457, .external_lex_state = 114},
  [20940] = {.lex_state = 457},
  [20941] = {.lex_state = 455},
  [20942] = {.lex_state = 457, .external_lex_state = 114},
  [20943] = {.lex_state = 457, .external_lex_state = 114},
  [20944] = {.lex_state = 457, .external_lex_state = 122},
  [20945] = {.lex_state = 457},
  [20946] = {.lex_state = 457},
  [20947] = {.lex_state = 457, .external_lex_state = 114},
  [20948] = {.lex_state = 457, .external_lex_state = 114},
  [20949] = {.lex_state = 457, .external_lex_state = 122},
  [20950] = {.lex_state = 457},
  [20951] = {.lex_state = 190},
  [20952] = {.lex_state = 457, .external_lex_state = 114},
  [20953] = {.lex_state = 457, .external_lex_state = 114},
  [20954] = {.lex_state = 190},
  [20955] = {.lex_state = 457, .external_lex_state = 114},
  [20956] = {.lex_state = 457},
  [20957] = {.lex_state = 457, .external_lex_state = 114},
  [20958] = {.lex_state = 457, .external_lex_state = 114},
  [20959] = {.lex_state = 457},
  [20960] = {.lex_state = 457},
  [20961] = {.lex_state = 190},
  [20962] = {.lex_state = 457, .external_lex_state = 114},
  [20963] = {.lex_state = 457, .external_lex_state = 114},
  [20964] = {.lex_state = 457},
  [20965] = {.lex_state = 190},
  [20966] = {.lex_state = 455},
  [20967] = {.lex_state = 457, .external_lex_state = 114},
  [20968] = {.lex_state = 457, .external_lex_state = 114},
  [20969] = {.lex_state = 457, .external_lex_state = 122},
  [20970] = {.lex_state = 455},
  [20971] = {.lex_state = 457, .external_lex_state = 122},
  [20972] = {.lex_state = 457, .external_lex_state = 114},
  [20973] = {.lex_state = 457, .external_lex_state = 114},
  [20974] = {.lex_state = 455},
  [20975] = {.lex_state = 457, .external_lex_state = 114},
  [20976] = {.lex_state = 455, .external_lex_state = 134},
  [20977] = {.lex_state = 457, .external_lex_state = 114},
  [20978] = {.lex_state = 457, .external_lex_state = 114},
  [20979] = {.lex_state = 457, .external_lex_state = 114},
  [20980] = {.lex_state = 457, .external_lex_state = 122},
  [20981] = {.lex_state = 457},
  [20982] = {.lex_state = 455, .external_lex_state = 134},
  [20983] = {.lex_state = 203},
  [20984] = {.lex_state = 457, .external_lex_state = 122},
  [20985] = {.lex_state = 457},
  [20986] = {.lex_state = 457},
  [20987] = {.lex_state = 457},
  [20988] = {.lex_state = 190},
  [20989] = {.lex_state = 457},
  [20990] = {.lex_state = 190},
  [20991] = {.lex_state = 455},
  [20992] = {.lex_state = 457},
  [20993] = {.lex_state = 457},
  [20994] = {.lex_state = 455},
  [20995] = {.lex_state = 457, .external_lex_state = 122},
  [20996] = {.lex_state = 190},
  [20997] = {.lex_state = 457},
  [20998] = {.lex_state = 457},
  [20999] = {.lex_state = 203},
  [21000] = {.lex_state = 457, .external_lex_state = 131},
  [21001] = {.lex_state = 455},
  [21002] = {.lex_state = 457},
  [21003] = {.lex_state = 457},
  [21004] = {.lex_state = 457},
  [21005] = {.lex_state = 457},
  [21006] = {.lex_state = 457},
  [21007] = {.lex_state = 455},
  [21008] = {.lex_state = 190},
  [21009] = {.lex_state = 457, .external_lex_state = 131},
  [21010] = {.lex_state = 203},
  [21011] = {.lex_state = 457},
  [21012] = {.lex_state = 190},
  [21013] = {.lex_state = 457, .external_lex_state = 131},
  [21014] = {.lex_state = 455},
  [21015] = {.lex_state = 190},
  [21016] = {.lex_state = 457, .external_lex_state = 114},
  [21017] = {.lex_state = 455},
  [21018] = {.lex_state = 457},
  [21019] = {.lex_state = 457},
  [21020] = {.lex_state = 457},
  [21021] = {.lex_state = 457, .external_lex_state = 131},
  [21022] = {.lex_state = 457, .external_lex_state = 131},
  [21023] = {.lex_state = 457, .external_lex_state = 131},
  [21024] = {.lex_state = 457, .external_lex_state = 131},
  [21025] = {.lex_state = 457},
  [21026] = {.lex_state = 190},
  [21027] = {.lex_state = 455, .external_lex_state = 134},
  [21028] = {.lex_state = 455},
  [21029] = {.lex_state = 457, .external_lex_state = 114},
  [21030] = {.lex_state = 455, .external_lex_state = 134},
  [21031] = {.lex_state = 457},
  [21032] = {.lex_state = 457},
  [21033] = {.lex_state = 457},
  [21034] = {.lex_state = 457, .external_lex_state = 114},
  [21035] = {.lex_state = 457},
  [21036] = {.lex_state = 457},
  [21037] = {.lex_state = 457, .external_lex_state = 131},
  [21038] = {.lex_state = 190},
  [21039] = {.lex_state = 457},
  [21040] = {.lex_state = 190},
  [21041] = {.lex_state = 457, .external_lex_state = 131},
  [21042] = {.lex_state = 455},
  [21043] = {.lex_state = 457, .external_lex_state = 131},
  [21044] = {.lex_state = 457, .external_lex_state = 114},
  [21045] = {.lex_state = 457, .external_lex_state = 131},
  [21046] = {.lex_state = 457, .external_lex_state = 131},
  [21047] = {.lex_state = 457, .external_lex_state = 131},
  [21048] = {.lex_state = 190},
  [21049] = {.lex_state = 457, .external_lex_state = 131},
  [21050] = {.lex_state = 455},
  [21051] = {.lex_state = 457},
  [21052] = {.lex_state = 457, .external_lex_state = 131},
  [21053] = {.lex_state = 457, .external_lex_state = 114},
  [21054] = {.lex_state = 457},
  [21055] = {.lex_state = 455},
  [21056] = {.lex_state = 457},
  [21057] = {.lex_state = 455},
  [21058] = {.lex_state = 455},
  [21059] = {.lex_state = 457},
  [21060] = {.lex_state = 457, .external_lex_state = 114},
  [21061] = {.lex_state = 457},
  [21062] = {.lex_state = 457, .external_lex_state = 131},
  [21063] = {.lex_state = 203},
  [21064] = {.lex_state = 457, .external_lex_state = 122},
  [21065] = {.lex_state = 455},
  [21066] = {.lex_state = 203},
  [21067] = {.lex_state = 457, .external_lex_state = 131},
  [21068] = {.lex_state = 455},
  [21069] = {.lex_state = 457},
  [21070] = {.lex_state = 457},
  [21071] = {.lex_state = 457, .external_lex_state = 114},
  [21072] = {.lex_state = 457},
  [21073] = {.lex_state = 457, .external_lex_state = 131},
  [21074] = {.lex_state = 455},
  [21075] = {.lex_state = 457},
  [21076] = {.lex_state = 190},
  [21077] = {.lex_state = 457},
  [21078] = {.lex_state = 190},
  [21079] = {.lex_state = 457},
  [21080] = {.lex_state = 455},
  [21081] = {.lex_state = 457},
  [21082] = {.lex_state = 457},
  [21083] = {.lex_state = 455},
  [21084] = {.lex_state = 457, .external_lex_state = 122},
  [21085] = {.lex_state = 457},
  [21086] = {.lex_state = 457, .external_lex_state = 122},
  [21087] = {.lex_state = 457, .external_lex_state = 114},
  [21088] = {.lex_state = 457},
  [21089] = {.lex_state = 190},
  [21090] = {.lex_state = 457},
  [21091] = {.lex_state = 457, .external_lex_state = 114},
  [21092] = {.lex_state = 457},
  [21093] = {.lex_state = 457},
  [21094] = {.lex_state = 457, .external_lex_state = 114},
  [21095] = {.lex_state = 190},
  [21096] = {.lex_state = 457},
  [21097] = {.lex_state = 190},
  [21098] = {.lex_state = 457, .external_lex_state = 131},
  [21099] = {.lex_state = 455},
  [21100] = {.lex_state = 455},
  [21101] = {.lex_state = 455},
  [21102] = {.lex_state = 457},
  [21103] = {.lex_state = 455},
  [21104] = {.lex_state = 190},
  [21105] = {.lex_state = 457, .external_lex_state = 131},
  [21106] = {.lex_state = 457, .external_lex_state = 114},
  [21107] = {.lex_state = 457, .external_lex_state = 131},
  [21108] = {.lex_state = 203},
  [21109] = {.lex_state = 457, .external_lex_state = 131},
  [21110] = {.lex_state = 457, .external_lex_state = 114},
  [21111] = {.lex_state = 457, .external_lex_state = 131},
  [21112] = {.lex_state = 457},
  [21113] = {.lex_state = 457, .external_lex_state = 122},
  [21114] = {.lex_state = 457},
  [21115] = {.lex_state = 457},
  [21116] = {.lex_state = 457},
  [21117] = {.lex_state = 457, .external_lex_state = 131},
  [21118] = {.lex_state = 457, .external_lex_state = 131},
  [21119] = {.lex_state = 457, .external_lex_state = 131},
  [21120] = {.lex_state = 457, .external_lex_state = 131},
  [21121] = {.lex_state = 190},
  [21122] = {.lex_state = 457},
  [21123] = {.lex_state = 190},
  [21124] = {.lex_state = 455},
  [21125] = {.lex_state = 457},
  [21126] = {.lex_state = 457},
  [21127] = {.lex_state = 455, .external_lex_state = 134},
  [21128] = {.lex_state = 455},
  [21129] = {.lex_state = 457, .external_lex_state = 131},
  [21130] = {.lex_state = 455},
  [21131] = {.lex_state = 457},
  [21132] = {.lex_state = 457},
  [21133] = {.lex_state = 457, .external_lex_state = 131},
  [21134] = {.lex_state = 457},
  [21135] = {.lex_state = 457, .external_lex_state = 131},
  [21136] = {.lex_state = 203},
  [21137] = {.lex_state = 457, .external_lex_state = 131},
  [21138] = {.lex_state = 457, .external_lex_state = 131},
  [21139] = {.lex_state = 190},
  [21140] = {.lex_state = 457, .external_lex_state = 131},
  [21141] = {.lex_state = 457},
  [21142] = {.lex_state = 457, .external_lex_state = 131},
  [21143] = {.lex_state = 457, .external_lex_state = 122},
  [21144] = {.lex_state = 457, .external_lex_state = 114},
  [21145] = {.lex_state = 455},
  [21146] = {.lex_state = 457},
  [21147] = {.lex_state = 455},
  [21148] = {.lex_state = 457},
  [21149] = {.lex_state = 457, .external_lex_state = 114},
  [21150] = {.lex_state = 457},
  [21151] = {.lex_state = 457, .external_lex_state = 131},
  [21152] = {.lex_state = 190},
  [21153] = {.lex_state = 457},
  [21154] = {.lex_state = 455},
  [21155] = {.lex_state = 190},
  [21156] = {.lex_state = 457, .external_lex_state = 131},
  [21157] = {.lex_state = 455},
  [21158] = {.lex_state = 455},
  [21159] = {.lex_state = 457},
  [21160] = {.lex_state = 457, .external_lex_state = 114},
  [21161] = {.lex_state = 203},
  [21162] = {.lex_state = 457, .external_lex_state = 131},
  [21163] = {.lex_state = 455},
  [21164] = {.lex_state = 457},
  [21165] = {.lex_state = 455},
  [21166] = {.lex_state = 457},
  [21167] = {.lex_state = 203},
  [21168] = {.lex_state = 457, .external_lex_state = 122},
  [21169] = {.lex_state = 457, .external_lex_state = 131},
  [21170] = {.lex_state = 455},
  [21171] = {.lex_state = 457},
  [21172] = {.lex_state = 457},
  [21173] = {.lex_state = 455},
  [21174] = {.lex_state = 190},
  [21175] = {.lex_state = 457},
  [21176] = {.lex_state = 457},
  [21177] = {.lex_state = 457},
  [21178] = {.lex_state = 457, .external_lex_state = 131},
  [21179] = {.lex_state = 203},
  [21180] = {.lex_state = 457},
  [21181] = {.lex_state = 457, .external_lex_state = 131},
  [21182] = {.lex_state = 190},
  [21183] = {.lex_state = 190},
  [21184] = {.lex_state = 457},
  [21185] = {.lex_state = 190},
  [21186] = {.lex_state = 455},
  [21187] = {.lex_state = 457, .external_lex_state = 114},
  [21188] = {.lex_state = 457, .external_lex_state = 114},
  [21189] = {.lex_state = 457, .external_lex_state = 131},
  [21190] = {.lex_state = 457, .external_lex_state = 131},
  [21191] = {.lex_state = 457, .external_lex_state = 131},
  [21192] = {.lex_state = 457, .external_lex_state = 131},
  [21193] = {.lex_state = 455},
  [21194] = {.lex_state = 457},
  [21195] = {.lex_state = 457, .external_lex_state = 114},
  [21196] = {.lex_state = 455},
  [21197] = {.lex_state = 457, .external_lex_state = 114},
  [21198] = {.lex_state = 457},
  [21199] = {.lex_state = 457, .external_lex_state = 131},
  [21200] = {.lex_state = 457, .external_lex_state = 122},
  [21201] = {.lex_state = 457},
  [21202] = {.lex_state = 457},
  [21203] = {.lex_state = 457, .external_lex_state = 131},
  [21204] = {.lex_state = 457},
  [21205] = {.lex_state = 457},
  [21206] = {.lex_state = 457, .external_lex_state = 131},
  [21207] = {.lex_state = 190},
  [21208] = {.lex_state = 457, .external_lex_state = 131},
  [21209] = {.lex_state = 190},
  [21210] = {.lex_state = 457, .external_lex_state = 131},
  [21211] = {.lex_state = 457},
  [21212] = {.lex_state = 190},
  [21213] = {.lex_state = 455},
  [21214] = {.lex_state = 455},
  [21215] = {.lex_state = 455},
  [21216] = {.lex_state = 190},
  [21217] = {.lex_state = 457, .external_lex_state = 114},
  [21218] = {.lex_state = 457, .external_lex_state = 114},
  [21219] = {.lex_state = 457, .external_lex_state = 131},
  [21220] = {.lex_state = 455},
  [21221] = {.lex_state = 457, .external_lex_state = 114},
  [21222] = {.lex_state = 455},
  [21223] = {.lex_state = 457, .external_lex_state = 114},
  [21224] = {.lex_state = 457, .external_lex_state = 131},
  [21225] = {.lex_state = 455},
  [21226] = {.lex_state = 457, .external_lex_state = 114},
  [21227] = {.lex_state = 457, .external_lex_state = 114},
  [21228] = {.lex_state = 457, .external_lex_state = 114},
  [21229] = {.lex_state = 457, .external_lex_state = 114},
  [21230] = {.lex_state = 457, .external_lex_state = 131},
  [21231] = {.lex_state = 455},
  [21232] = {.lex_state = 457},
  [21233] = {.lex_state = 457},
  [21234] = {.lex_state = 1},
  [21235] = {.lex_state = 457, .external_lex_state = 114},
  [21236] = {.lex_state = 457, .external_lex_state = 131},
  [21237] = {.lex_state = 455},
  [21238] = {.lex_state = 457},
  [21239] = {.lex_state = 457},
  [21240] = {.lex_state = 1},
  [21241] = {.lex_state = 190},
  [21242] = {.lex_state = 457},
  [21243] = {.lex_state = 457},
  [21244] = {.lex_state = 190},
  [21245] = {.lex_state = 457, .external_lex_state = 131},
  [21246] = {.lex_state = 203},
  [21247] = {.lex_state = 455},
  [21248] = {.lex_state = 457, .external_lex_state = 131},
  [21249] = {.lex_state = 457, .external_lex_state = 114},
  [21250] = {.lex_state = 203},
  [21251] = {.lex_state = 455},
  [21252] = {.lex_state = 1},
  [21253] = {.lex_state = 457, .external_lex_state = 114},
  [21254] = {.lex_state = 203},
  [21255] = {.lex_state = 457, .external_lex_state = 131},
  [21256] = {.lex_state = 457, .external_lex_state = 131},
  [21257] = {.lex_state = 457, .external_lex_state = 131},
  [21258] = {.lex_state = 457, .external_lex_state = 131},
  [21259] = {.lex_state = 457, .external_lex_state = 114},
  [21260] = {.lex_state = 457, .external_lex_state = 131},
  [21261] = {.lex_state = 203},
  [21262] = {.lex_state = 203},
  [21263] = {.lex_state = 1},
  [21264] = {.lex_state = 457, .external_lex_state = 114},
  [21265] = {.lex_state = 457, .external_lex_state = 114},
  [21266] = {.lex_state = 457, .external_lex_state = 131},
  [21267] = {.lex_state = 457},
  [21268] = {.lex_state = 457},
  [21269] = {.lex_state = 457},
  [21270] = {.lex_state = 457, .external_lex_state = 131},
  [21271] = {.lex_state = 190},
  [21272] = {.lex_state = 457},
  [21273] = {.lex_state = 457, .external_lex_state = 131},
  [21274] = {.lex_state = 457, .external_lex_state = 131},
  [21275] = {.lex_state = 190},
  [21276] = {.lex_state = 457, .external_lex_state = 131},
  [21277] = {.lex_state = 455},
  [21278] = {.lex_state = 457, .external_lex_state = 114},
  [21279] = {.lex_state = 455},
  [21280] = {.lex_state = 455},
  [21281] = {.lex_state = 455},
  [21282] = {.lex_state = 457, .external_lex_state = 114},
  [21283] = {.lex_state = 457, .external_lex_state = 114},
  [21284] = {.lex_state = 455},
  [21285] = {.lex_state = 457, .external_lex_state = 131},
  [21286] = {.lex_state = 455},
  [21287] = {.lex_state = 457, .external_lex_state = 114},
  [21288] = {.lex_state = 455},
  [21289] = {.lex_state = 455},
  [21290] = {.lex_state = 457, .external_lex_state = 131},
  [21291] = {.lex_state = 455},
  [21292] = {.lex_state = 457},
  [21293] = {.lex_state = 457},
  [21294] = {.lex_state = 457, .external_lex_state = 114},
  [21295] = {.lex_state = 190},
  [21296] = {.lex_state = 457, .external_lex_state = 131},
  [21297] = {.lex_state = 455},
  [21298] = {.lex_state = 457},
  [21299] = {.lex_state = 457},
  [21300] = {.lex_state = 455},
  [21301] = {.lex_state = 457},
  [21302] = {.lex_state = 457, .external_lex_state = 131},
  [21303] = {.lex_state = 455},
  [21304] = {.lex_state = 457},
  [21305] = {.lex_state = 455},
  [21306] = {.lex_state = 190},
  [21307] = {.lex_state = 457},
  [21308] = {.lex_state = 190},
  [21309] = {.lex_state = 455},
  [21310] = {.lex_state = 457, .external_lex_state = 131},
  [21311] = {.lex_state = 203},
  [21312] = {.lex_state = 457},
  [21313] = {.lex_state = 457, .external_lex_state = 131},
  [21314] = {.lex_state = 457, .external_lex_state = 114},
  [21315] = {.lex_state = 455},
  [21316] = {.lex_state = 457},
  [21317] = {.lex_state = 457},
  [21318] = {.lex_state = 455},
  [21319] = {.lex_state = 455},
  [21320] = {.lex_state = 457, .external_lex_state = 131},
  [21321] = {.lex_state = 457, .external_lex_state = 131},
  [21322] = {.lex_state = 457, .external_lex_state = 131},
  [21323] = {.lex_state = 457, .external_lex_state = 131},
  [21324] = {.lex_state = 457, .external_lex_state = 122},
  [21325] = {.lex_state = 455},
  [21326] = {.lex_state = 457},
  [21327] = {.lex_state = 190},
  [21328] = {.lex_state = 455},
  [21329] = {.lex_state = 455},
  [21330] = {.lex_state = 457, .external_lex_state = 131},
  [21331] = {.lex_state = 203},
  [21332] = {.lex_state = 457},
  [21333] = {.lex_state = 457, .external_lex_state = 131},
  [21334] = {.lex_state = 457},
  [21335] = {.lex_state = 455},
  [21336] = {.lex_state = 457, .external_lex_state = 131},
  [21337] = {.lex_state = 457, .external_lex_state = 131},
  [21338] = {.lex_state = 457, .external_lex_state = 131},
  [21339] = {.lex_state = 190},
  [21340] = {.lex_state = 457, .external_lex_state = 131},
  [21341] = {.lex_state = 457},
  [21342] = {.lex_state = 190},
  [21343] = {.lex_state = 455},
  [21344] = {.lex_state = 455},
  [21345] = {.lex_state = 455},
  [21346] = {.lex_state = 455},
  [21347] = {.lex_state = 457, .external_lex_state = 114},
  [21348] = {.lex_state = 455},
  [21349] = {.lex_state = 457, .external_lex_state = 131},
  [21350] = {.lex_state = 455},
  [21351] = {.lex_state = 457, .external_lex_state = 114},
  [21352] = {.lex_state = 455},
  [21353] = {.lex_state = 455, .external_lex_state = 134},
  [21354] = {.lex_state = 457, .external_lex_state = 131},
  [21355] = {.lex_state = 455},
  [21356] = {.lex_state = 455, .external_lex_state = 134},
  [21357] = {.lex_state = 457},
  [21358] = {.lex_state = 457, .external_lex_state = 114},
  [21359] = {.lex_state = 457, .external_lex_state = 122},
  [21360] = {.lex_state = 457, .external_lex_state = 131},
  [21361] = {.lex_state = 455},
  [21362] = {.lex_state = 457},
  [21363] = {.lex_state = 457},
  [21364] = {.lex_state = 1},
  [21365] = {.lex_state = 457},
  [21366] = {.lex_state = 457, .external_lex_state = 131},
  [21367] = {.lex_state = 455},
  [21368] = {.lex_state = 457},
  [21369] = {.lex_state = 455},
  [21370] = {.lex_state = 457},
  [21371] = {.lex_state = 457},
  [21372] = {.lex_state = 203},
  [21373] = {.lex_state = 190},
  [21374] = {.lex_state = 457, .external_lex_state = 131},
  [21375] = {.lex_state = 203},
  [21376] = {.lex_state = 457},
  [21377] = {.lex_state = 457, .external_lex_state = 131},
  [21378] = {.lex_state = 190},
  [21379] = {.lex_state = 455},
  [21380] = {.lex_state = 1},
  [21381] = {.lex_state = 190},
  [21382] = {.lex_state = 455},
  [21383] = {.lex_state = 457},
  [21384] = {.lex_state = 457, .external_lex_state = 131},
  [21385] = {.lex_state = 457, .external_lex_state = 131},
  [21386] = {.lex_state = 457, .external_lex_state = 131},
  [21387] = {.lex_state = 457, .external_lex_state = 131},
  [21388] = {.lex_state = 457, .external_lex_state = 114},
  [21389] = {.lex_state = 457},
  [21390] = {.lex_state = 457, .external_lex_state = 114},
  [21391] = {.lex_state = 457, .external_lex_state = 122},
  [21392] = {.lex_state = 457, .external_lex_state = 114},
  [21393] = {.lex_state = 457, .external_lex_state = 131},
  [21394] = {.lex_state = 457},
  [21395] = {.lex_state = 457},
  [21396] = {.lex_state = 457, .external_lex_state = 131},
  [21397] = {.lex_state = 457, .external_lex_state = 114},
  [21398] = {.lex_state = 190},
  [21399] = {.lex_state = 457, .external_lex_state = 131},
  [21400] = {.lex_state = 457, .external_lex_state = 131},
  [21401] = {.lex_state = 457, .external_lex_state = 131},
  [21402] = {.lex_state = 457},
  [21403] = {.lex_state = 457, .external_lex_state = 131},
  [21404] = {.lex_state = 190},
  [21405] = {.lex_state = 455},
  [21406] = {.lex_state = 455},
  [21407] = {.lex_state = 457},
  [21408] = {.lex_state = 455},
  [21409] = {.lex_state = 455},
  [21410] = {.lex_state = 457, .external_lex_state = 114},
  [21411] = {.lex_state = 455},
  [21412] = {.lex_state = 457, .external_lex_state = 131},
  [21413] = {.lex_state = 457},
  [21414] = {.lex_state = 457},
  [21415] = {.lex_state = 455},
  [21416] = {.lex_state = 455, .external_lex_state = 134},
  [21417] = {.lex_state = 457, .external_lex_state = 131},
  [21418] = {.lex_state = 455},
  [21419] = {.lex_state = 457},
  [21420] = {.lex_state = 457},
  [21421] = {.lex_state = 457, .external_lex_state = 114},
  [21422] = {.lex_state = 457},
  [21423] = {.lex_state = 457, .external_lex_state = 131},
  [21424] = {.lex_state = 455},
  [21425] = {.lex_state = 457},
  [21426] = {.lex_state = 457},
  [21427] = {.lex_state = 457},
  [21428] = {.lex_state = 457},
  [21429] = {.lex_state = 457, .external_lex_state = 131},
  [21430] = {.lex_state = 455},
  [21431] = {.lex_state = 457},
  [21432] = {.lex_state = 190},
  [21433] = {.lex_state = 457},
  [21434] = {.lex_state = 455},
  [21435] = {.lex_state = 190},
  [21436] = {.lex_state = 455},
  [21437] = {.lex_state = 457, .external_lex_state = 131},
  [21438] = {.lex_state = 203},
  [21439] = {.lex_state = 190},
  [21440] = {.lex_state = 457, .external_lex_state = 131},
  [21441] = {.lex_state = 457, .external_lex_state = 122},
  [21442] = {.lex_state = 455},
  [21443] = {.lex_state = 190},
  [21444] = {.lex_state = 457, .external_lex_state = 122},
  [21445] = {.lex_state = 190},
  [21446] = {.lex_state = 457},
  [21447] = {.lex_state = 457, .external_lex_state = 131},
  [21448] = {.lex_state = 457, .external_lex_state = 131},
  [21449] = {.lex_state = 457, .external_lex_state = 131},
  [21450] = {.lex_state = 457, .external_lex_state = 131},
  [21451] = {.lex_state = 457},
  [21452] = {.lex_state = 203},
  [21453] = {.lex_state = 457, .external_lex_state = 122},
  [21454] = {.lex_state = 457, .external_lex_state = 114},
  [21455] = {.lex_state = 457},
  [21456] = {.lex_state = 457},
  [21457] = {.lex_state = 457, .external_lex_state = 131},
  [21458] = {.lex_state = 457},
  [21459] = {.lex_state = 190},
  [21460] = {.lex_state = 457, .external_lex_state = 131},
  [21461] = {.lex_state = 457},
  [21462] = {.lex_state = 190},
  [21463] = {.lex_state = 457, .external_lex_state = 131},
  [21464] = {.lex_state = 457, .external_lex_state = 131},
  [21465] = {.lex_state = 455},
  [21466] = {.lex_state = 457, .external_lex_state = 131},
  [21467] = {.lex_state = 457},
  [21468] = {.lex_state = 457},
  [21469] = {.lex_state = 455},
  [21470] = {.lex_state = 455},
  [21471] = {.lex_state = 455},
  [21472] = {.lex_state = 455},
  [21473] = {.lex_state = 457, .external_lex_state = 114},
  [21474] = {.lex_state = 190},
  [21475] = {.lex_state = 457, .external_lex_state = 131},
  [21476] = {.lex_state = 457},
  [21477] = {.lex_state = 455, .external_lex_state = 134},
  [21478] = {.lex_state = 455},
  [21479] = {.lex_state = 457, .external_lex_state = 122},
  [21480] = {.lex_state = 457, .external_lex_state = 131},
  [21481] = {.lex_state = 455},
  [21482] = {.lex_state = 190},
  [21483] = {.lex_state = 457},
  [21484] = {.lex_state = 457, .external_lex_state = 114},
  [21485] = {.lex_state = 457},
  [21486] = {.lex_state = 457, .external_lex_state = 131},
  [21487] = {.lex_state = 455},
  [21488] = {.lex_state = 457},
  [21489] = {.lex_state = 457},
  [21490] = {.lex_state = 457},
  [21491] = {.lex_state = 457},
  [21492] = {.lex_state = 457, .external_lex_state = 131},
  [21493] = {.lex_state = 455},
  [21494] = {.lex_state = 457},
  [21495] = {.lex_state = 190},
  [21496] = {.lex_state = 457},
  [21497] = {.lex_state = 455},
  [21498] = {.lex_state = 455},
  [21499] = {.lex_state = 457, .external_lex_state = 131},
  [21500] = {.lex_state = 203},
  [21501] = {.lex_state = 190},
  [21502] = {.lex_state = 457},
  [21503] = {.lex_state = 455},
  [21504] = {.lex_state = 457, .external_lex_state = 114},
  [21505] = {.lex_state = 190},
  [21506] = {.lex_state = 455},
  [21507] = {.lex_state = 457},
  [21508] = {.lex_state = 457, .external_lex_state = 122},
  [21509] = {.lex_state = 455},
  [21510] = {.lex_state = 457},
  [21511] = {.lex_state = 457, .external_lex_state = 131},
  [21512] = {.lex_state = 457},
  [21513] = {.lex_state = 457},
  [21514] = {.lex_state = 457, .external_lex_state = 131},
  [21515] = {.lex_state = 190},
  [21516] = {.lex_state = 457, .external_lex_state = 137},
  [21517] = {.lex_state = 457},
  [21518] = {.lex_state = 455},
  [21519] = {.lex_state = 457},
  [21520] = {.lex_state = 455},
  [21521] = {.lex_state = 457, .external_lex_state = 114},
  [21522] = {.lex_state = 457, .external_lex_state = 131},
  [21523] = {.lex_state = 190},
  [21524] = {.lex_state = 457},
  [21525] = {.lex_state = 455},
  [21526] = {.lex_state = 190},
  [21527] = {.lex_state = 455},
  [21528] = {.lex_state = 455},
  [21529] = {.lex_state = 457, .external_lex_state = 122},
  [21530] = {.lex_state = 457, .external_lex_state = 114},
  [21531] = {.lex_state = 457, .external_lex_state = 131},
  [21532] = {.lex_state = 455},
  [21533] = {.lex_state = 457},
  [21534] = {.lex_state = 455},
  [21535] = {.lex_state = 457, .external_lex_state = 131},
  [21536] = {.lex_state = 455},
  [21537] = {.lex_state = 457, .external_lex_state = 131},
  [21538] = {.lex_state = 455},
  [21539] = {.lex_state = 455},
  [21540] = {.lex_state = 457},
  [21541] = {.lex_state = 457},
  [21542] = {.lex_state = 457, .external_lex_state = 114},
  [21543] = {.lex_state = 457, .external_lex_state = 114},
  [21544] = {.lex_state = 457, .external_lex_state = 131},
  [21545] = {.lex_state = 203},
  [21546] = {.lex_state = 455},
  [21547] = {.lex_state = 457},
  [21548] = {.lex_state = 455},
  [21549] = {.lex_state = 457},
  [21550] = {.lex_state = 457},
  [21551] = {.lex_state = 455},
  [21552] = {.lex_state = 190},
  [21553] = {.lex_state = 457},
  [21554] = {.lex_state = 190},
  [21555] = {.lex_state = 457, .external_lex_state = 131},
  [21556] = {.lex_state = 457, .external_lex_state = 131},
  [21557] = {.lex_state = 455},
  [21558] = {.lex_state = 203},
  [21559] = {.lex_state = 457, .external_lex_state = 131},
  [21560] = {.lex_state = 457},
  [21561] = {.lex_state = 455},
  [21562] = {.lex_state = 455},
  [21563] = {.lex_state = 455},
  [21564] = {.lex_state = 457},
  [21565] = {.lex_state = 455},
  [21566] = {.lex_state = 190},
  [21567] = {.lex_state = 457, .external_lex_state = 131},
  [21568] = {.lex_state = 457},
  [21569] = {.lex_state = 455},
  [21570] = {.lex_state = 455},
  [21571] = {.lex_state = 455, .external_lex_state = 134},
  [21572] = {.lex_state = 455},
  [21573] = {.lex_state = 457},
  [21574] = {.lex_state = 457},
  [21575] = {.lex_state = 455},
  [21576] = {.lex_state = 457, .external_lex_state = 131},
  [21577] = {.lex_state = 455},
  [21578] = {.lex_state = 457},
  [21579] = {.lex_state = 457, .external_lex_state = 122},
  [21580] = {.lex_state = 455, .external_lex_state = 134},
  [21581] = {.lex_state = 455},
  [21582] = {.lex_state = 457, .external_lex_state = 131},
  [21583] = {.lex_state = 455},
  [21584] = {.lex_state = 455},
  [21585] = {.lex_state = 457},
  [21586] = {.lex_state = 457},
  [21587] = {.lex_state = 455},
  [21588] = {.lex_state = 455},
  [21589] = {.lex_state = 457, .external_lex_state = 131},
  [21590] = {.lex_state = 203},
  [21591] = {.lex_state = 190},
  [21592] = {.lex_state = 457},
  [21593] = {.lex_state = 190},
  [21594] = {.lex_state = 457},
  [21595] = {.lex_state = 455},
  [21596] = {.lex_state = 455, .external_lex_state = 134},
  [21597] = {.lex_state = 457},
  [21598] = {.lex_state = 455},
  [21599] = {.lex_state = 457},
  [21600] = {.lex_state = 190},
  [21601] = {.lex_state = 457, .external_lex_state = 131},
  [21602] = {.lex_state = 455},
  [21603] = {.lex_state = 457},
  [21604] = {.lex_state = 457, .external_lex_state = 131},
  [21605] = {.lex_state = 1},
  [21606] = {.lex_state = 1},
  [21607] = {.lex_state = 457},
  [21608] = {.lex_state = 455},
  [21609] = {.lex_state = 455},
  [21610] = {.lex_state = 457},
  [21611] = {.lex_state = 457, .external_lex_state = 131},
  [21612] = {.lex_state = 457},
  [21613] = {.lex_state = 455},
  [21614] = {.lex_state = 455},
  [21615] = {.lex_state = 190},
  [21616] = {.lex_state = 455},
  [21617] = {.lex_state = 457, .external_lex_state = 122},
  [21618] = {.lex_state = 457},
  [21619] = {.lex_state = 457, .external_lex_state = 131},
  [21620] = {.lex_state = 455},
  [21621] = {.lex_state = 190},
  [21622] = {.lex_state = 455},
  [21623] = {.lex_state = 457},
  [21624] = {.lex_state = 457, .external_lex_state = 131},
  [21625] = {.lex_state = 455},
  [21626] = {.lex_state = 203},
  [21627] = {.lex_state = 455},
  [21628] = {.lex_state = 457},
  [21629] = {.lex_state = 455},
  [21630] = {.lex_state = 457},
  [21631] = {.lex_state = 457, .external_lex_state = 131},
  [21632] = {.lex_state = 203},
  [21633] = {.lex_state = 203},
  [21634] = {.lex_state = 455},
  [21635] = {.lex_state = 457},
  [21636] = {.lex_state = 457, .external_lex_state = 131},
  [21637] = {.lex_state = 457, .external_lex_state = 114},
  [21638] = {.lex_state = 457},
  [21639] = {.lex_state = 457},
  [21640] = {.lex_state = 457},
  [21641] = {.lex_state = 190},
  [21642] = {.lex_state = 457},
  [21643] = {.lex_state = 457, .external_lex_state = 131},
  [21644] = {.lex_state = 457},
  [21645] = {.lex_state = 190},
  [21646] = {.lex_state = 457, .external_lex_state = 131},
  [21647] = {.lex_state = 455},
  [21648] = {.lex_state = 455},
  [21649] = {.lex_state = 455},
  [21650] = {.lex_state = 455},
  [21651] = {.lex_state = 190},
  [21652] = {.lex_state = 203},
  [21653] = {.lex_state = 457, .external_lex_state = 131},
  [21654] = {.lex_state = 457},
  [21655] = {.lex_state = 203},
  [21656] = {.lex_state = 455},
  [21657] = {.lex_state = 457},
  [21658] = {.lex_state = 455},
  [21659] = {.lex_state = 457},
  [21660] = {.lex_state = 457},
  [21661] = {.lex_state = 457, .external_lex_state = 131},
  [21662] = {.lex_state = 455},
  [21663] = {.lex_state = 457},
  [21664] = {.lex_state = 457, .external_lex_state = 114},
  [21665] = {.lex_state = 190},
  [21666] = {.lex_state = 457, .external_lex_state = 131},
  [21667] = {.lex_state = 455},
  [21668] = {.lex_state = 457},
  [21669] = {.lex_state = 190},
  [21670] = {.lex_state = 455},
  [21671] = {.lex_state = 457},
  [21672] = {.lex_state = 203},
  [21673] = {.lex_state = 457, .external_lex_state = 131},
  [21674] = {.lex_state = 203},
  [21675] = {.lex_state = 455},
  [21676] = {.lex_state = 455, .external_lex_state = 134},
  [21677] = {.lex_state = 455},
  [21678] = {.lex_state = 457},
  [21679] = {.lex_state = 457},
  [21680] = {.lex_state = 457},
  [21681] = {.lex_state = 455},
  [21682] = {.lex_state = 455},
  [21683] = {.lex_state = 457, .external_lex_state = 114},
  [21684] = {.lex_state = 457, .external_lex_state = 131},
  [21685] = {.lex_state = 457, .external_lex_state = 122},
  [21686] = {.lex_state = 190},
  [21687] = {.lex_state = 457, .external_lex_state = 114},
  [21688] = {.lex_state = 455},
  [21689] = {.lex_state = 457},
  [21690] = {.lex_state = 457},
  [21691] = {.lex_state = 457, .external_lex_state = 131},
  [21692] = {.lex_state = 203},
  [21693] = {.lex_state = 457},
  [21694] = {.lex_state = 455},
  [21695] = {.lex_state = 457},
  [21696] = {.lex_state = 455},
  [21697] = {.lex_state = 457},
  [21698] = {.lex_state = 457},
  [21699] = {.lex_state = 457, .external_lex_state = 131},
  [21700] = {.lex_state = 455},
  [21701] = {.lex_state = 455},
  [21702] = {.lex_state = 457, .external_lex_state = 114},
  [21703] = {.lex_state = 455},
  [21704] = {.lex_state = 457, .external_lex_state = 131},
  [21705] = {.lex_state = 455},
  [21706] = {.lex_state = 455, .external_lex_state = 134},
  [21707] = {.lex_state = 457},
  [21708] = {.lex_state = 190},
  [21709] = {.lex_state = 203},
  [21710] = {.lex_state = 457},
  [21711] = {.lex_state = 457, .external_lex_state = 131},
  [21712] = {.lex_state = 203},
  [21713] = {.lex_state = 457},
  [21714] = {.lex_state = 457},
  [21715] = {.lex_state = 455},
  [21716] = {.lex_state = 455},
  [21717] = {.lex_state = 455, .external_lex_state = 134},
  [21718] = {.lex_state = 457},
  [21719] = {.lex_state = 457},
  [21720] = {.lex_state = 457},
  [21721] = {.lex_state = 455, .external_lex_state = 134},
  [21722] = {.lex_state = 457},
  [21723] = {.lex_state = 455},
  [21724] = {.lex_state = 455},
  [21725] = {.lex_state = 455},
  [21726] = {.lex_state = 457, .external_lex_state = 114},
  [21727] = {.lex_state = 455},
  [21728] = {.lex_state = 457},
  [21729] = {.lex_state = 455},
  [21730] = {.lex_state = 455},
  [21731] = {.lex_state = 455},
  [21732] = {.lex_state = 455},
  [21733] = {.lex_state = 457, .external_lex_state = 114},
  [21734] = {.lex_state = 457, .external_lex_state = 114},
  [21735] = {.lex_state = 455, .external_lex_state = 134},
  [21736] = {.lex_state = 457, .external_lex_state = 131},
  [21737] = {.lex_state = 455},
  [21738] = {.lex_state = 457, .external_lex_state = 114},
  [21739] = {.lex_state = 457, .external_lex_state = 114},
  [21740] = {.lex_state = 457, .external_lex_state = 131},
  [21741] = {.lex_state = 455},
  [21742] = {.lex_state = 455},
  [21743] = {.lex_state = 457, .external_lex_state = 131},
  [21744] = {.lex_state = 203},
  [21745] = {.lex_state = 455},
  [21746] = {.lex_state = 457},
  [21747] = {.lex_state = 457, .external_lex_state = 122},
  [21748] = {.lex_state = 457, .external_lex_state = 114},
  [21749] = {.lex_state = 455},
  [21750] = {.lex_state = 455},
  [21751] = {.lex_state = 190},
  [21752] = {.lex_state = 457},
  [21753] = {.lex_state = 455},
  [21754] = {.lex_state = 190},
  [21755] = {.lex_state = 457},
  [21756] = {.lex_state = 455},
  [21757] = {.lex_state = 457},
  [21758] = {.lex_state = 457, .external_lex_state = 131},
  [21759] = {.lex_state = 455},
  [21760] = {.lex_state = 190},
  [21761] = {.lex_state = 457},
  [21762] = {.lex_state = 457},
  [21763] = {.lex_state = 455},
  [21764] = {.lex_state = 457, .external_lex_state = 122},
  [21765] = {.lex_state = 457, .external_lex_state = 131},
  [21766] = {.lex_state = 203},
  [21767] = {.lex_state = 455},
  [21768] = {.lex_state = 203},
  [21769] = {.lex_state = 457, .external_lex_state = 122},
  [21770] = {.lex_state = 455},
  [21771] = {.lex_state = 455},
  [21772] = {.lex_state = 457, .external_lex_state = 114},
  [21773] = {.lex_state = 457},
  [21774] = {.lex_state = 455},
  [21775] = {.lex_state = 455},
  [21776] = {.lex_state = 203},
  [21777] = {.lex_state = 457, .external_lex_state = 122},
  [21778] = {.lex_state = 457, .external_lex_state = 131},
  [21779] = {.lex_state = 455},
  [21780] = {.lex_state = 455},
  [21781] = {.lex_state = 457, .external_lex_state = 114},
  [21782] = {.lex_state = 455, .external_lex_state = 134},
  [21783] = {.lex_state = 455},
  [21784] = {.lex_state = 457},
  [21785] = {.lex_state = 457, .external_lex_state = 131},
  [21786] = {.lex_state = 203},
  [21787] = {.lex_state = 455},
  [21788] = {.lex_state = 455},
  [21789] = {.lex_state = 457},
  [21790] = {.lex_state = 455},
  [21791] = {.lex_state = 455},
  [21792] = {.lex_state = 455},
  [21793] = {.lex_state = 455},
  [21794] = {.lex_state = 455},
  [21795] = {.lex_state = 457, .external_lex_state = 122},
  [21796] = {.lex_state = 457, .external_lex_state = 122},
  [21797] = {.lex_state = 455},
  [21798] = {.lex_state = 457, .external_lex_state = 131},
  [21799] = {.lex_state = 455},
  [21800] = {.lex_state = 455},
  [21801] = {.lex_state = 457, .external_lex_state = 122},
  [21802] = {.lex_state = 455},
  [21803] = {.lex_state = 455},
  [21804] = {.lex_state = 457},
  [21805] = {.lex_state = 457, .external_lex_state = 131},
  [21806] = {.lex_state = 203},
  [21807] = {.lex_state = 455},
  [21808] = {.lex_state = 457, .external_lex_state = 114},
  [21809] = {.lex_state = 455},
  [21810] = {.lex_state = 455, .external_lex_state = 134},
  [21811] = {.lex_state = 455},
  [21812] = {.lex_state = 455},
  [21813] = {.lex_state = 455},
  [21814] = {.lex_state = 455},
  [21815] = {.lex_state = 455},
  [21816] = {.lex_state = 455},
  [21817] = {.lex_state = 455},
  [21818] = {.lex_state = 457, .external_lex_state = 131},
  [21819] = {.lex_state = 455},
  [21820] = {.lex_state = 455},
  [21821] = {.lex_state = 455},
  [21822] = {.lex_state = 455},
  [21823] = {.lex_state = 455},
  [21824] = {.lex_state = 455},
  [21825] = {.lex_state = 457, .external_lex_state = 131},
  [21826] = {.lex_state = 203},
  [21827] = {.lex_state = 455},
  [21828] = {.lex_state = 455},
  [21829] = {.lex_state = 455},
  [21830] = {.lex_state = 455},
  [21831] = {.lex_state = 455},
  [21832] = {.lex_state = 455},
  [21833] = {.lex_state = 455},
  [21834] = {.lex_state = 455},
  [21835] = {.lex_state = 457, .external_lex_state = 114},
  [21836] = {.lex_state = 455, .external_lex_state = 134},
  [21837] = {.lex_state = 455},
  [21838] = {.lex_state = 457, .external_lex_state = 131},
  [21839] = {.lex_state = 455},
  [21840] = {.lex_state = 455},
  [21841] = {.lex_state = 455},
  [21842] = {.lex_state = 455},
  [21843] = {.lex_state = 455, .external_lex_state = 134},
  [21844] = {.lex_state = 457},
  [21845] = {.lex_state = 457, .external_lex_state = 131},
  [21846] = {.lex_state = 203},
  [21847] = {.lex_state = 457},
  [21848] = {.lex_state = 457, .external_lex_state = 114},
  [21849] = {.lex_state = 457},
  [21850] = {.lex_state = 457, .external_lex_state = 114},
  [21851] = {.lex_state = 457},
  [21852] = {.lex_state = 457},
  [21853] = {.lex_state = 457, .external_lex_state = 136},
  [21854] = {.lex_state = 190},
  [21855] = {.lex_state = 457},
  [21856] = {.lex_state = 455},
  [21857] = {.lex_state = 455, .external_lex_state = 135},
  [21858] = {.lex_state = 457, .external_lex_state = 131},
  [21859] = {.lex_state = 455},
  [21860] = {.lex_state = 457},
  [21861] = {.lex_state = 455},
  [21862] = {.lex_state = 457, .external_lex_state = 114},
  [21863] = {.lex_state = 455},
  [21864] = {.lex_state = 190},
  [21865] = {.lex_state = 457, .external_lex_state = 131},
  [21866] = {.lex_state = 203},
  [21867] = {.lex_state = 457},
  [21868] = {.lex_state = 190},
  [21869] = {.lex_state = 190},
  [21870] = {.lex_state = 457, .external_lex_state = 114},
  [21871] = {.lex_state = 457, .external_lex_state = 122},
  [21872] = {.lex_state = 457},
  [21873] = {.lex_state = 457, .external_lex_state = 114},
  [21874] = {.lex_state = 457},
  [21875] = {.lex_state = 457, .external_lex_state = 122},
  [21876] = {.lex_state = 457},
  [21877] = {.lex_state = 457},
  [21878] = {.lex_state = 457, .external_lex_state = 131},
  [21879] = {.lex_state = 455},
  [21880] = {.lex_state = 457, .external_lex_state = 122},
  [21881] = {.lex_state = 457},
  [21882] = {.lex_state = 457},
  [21883] = {.lex_state = 457},
  [21884] = {.lex_state = 190},
  [21885] = {.lex_state = 457, .external_lex_state = 131},
  [21886] = {.lex_state = 203},
  [21887] = {.lex_state = 455},
  [21888] = {.lex_state = 224, .external_lex_state = 109},
  [21889] = {.lex_state = 455, .external_lex_state = 134},
  [21890] = {.lex_state = 457},
  [21891] = {.lex_state = 457, .external_lex_state = 114},
  [21892] = {.lex_state = 455, .external_lex_state = 134},
  [21893] = {.lex_state = 457, .external_lex_state = 114},
  [21894] = {.lex_state = 457, .external_lex_state = 114},
  [21895] = {.lex_state = 457, .external_lex_state = 114},
  [21896] = {.lex_state = 457},
  [21897] = {.lex_state = 457, .external_lex_state = 114},
  [21898] = {.lex_state = 457, .external_lex_state = 131},
  [21899] = {.lex_state = 455},
  [21900] = {.lex_state = 455},
  [21901] = {.lex_state = 457},
  [21902] = {.lex_state = 457, .external_lex_state = 114},
  [21903] = {.lex_state = 455},
  [21904] = {.lex_state = 457, .external_lex_state = 114},
  [21905] = {.lex_state = 457, .external_lex_state = 131},
  [21906] = {.lex_state = 203},
  [21907] = {.lex_state = 455},
  [21908] = {.lex_state = 457},
  [21909] = {.lex_state = 457, .external_lex_state = 122},
  [21910] = {.lex_state = 457},
  [21911] = {.lex_state = 457, .external_lex_state = 114},
  [21912] = {.lex_state = 203},
  [21913] = {.lex_state = 457},
  [21914] = {.lex_state = 190},
  [21915] = {.lex_state = 455, .external_lex_state = 134},
  [21916] = {.lex_state = 457},
  [21917] = {.lex_state = 457},
  [21918] = {.lex_state = 457, .external_lex_state = 131},
  [21919] = {.lex_state = 455},
  [21920] = {.lex_state = 457, .external_lex_state = 114},
  [21921] = {.lex_state = 457},
  [21922] = {.lex_state = 455},
  [21923] = {.lex_state = 457, .external_lex_state = 114},
  [21924] = {.lex_state = 457, .external_lex_state = 114},
  [21925] = {.lex_state = 457, .external_lex_state = 131},
  [21926] = {.lex_state = 203},
  [21927] = {.lex_state = 457},
  [21928] = {.lex_state = 190},
  [21929] = {.lex_state = 455, .external_lex_state = 134},
  [21930] = {.lex_state = 455, .external_lex_state = 134},
  [21931] = {.lex_state = 457},
  [21932] = {.lex_state = 455, .external_lex_state = 134},
  [21933] = {.lex_state = 457},
  [21934] = {.lex_state = 457},
  [21935] = {.lex_state = 457},
  [21936] = {.lex_state = 457, .external_lex_state = 114},
  [21937] = {.lex_state = 190},
  [21938] = {.lex_state = 457, .external_lex_state = 131},
  [21939] = {.lex_state = 455},
  [21940] = {.lex_state = 455},
  [21941] = {.lex_state = 457, .external_lex_state = 114},
  [21942] = {.lex_state = 203},
  [21943] = {.lex_state = 457},
  [21944] = {.lex_state = 455, .external_lex_state = 134},
  [21945] = {.lex_state = 457, .external_lex_state = 131},
  [21946] = {.lex_state = 203},
  [21947] = {.lex_state = 457, .external_lex_state = 122},
  [21948] = {.lex_state = 457, .external_lex_state = 122},
  [21949] = {.lex_state = 457, .external_lex_state = 122},
  [21950] = {.lex_state = 457, .external_lex_state = 114},
  [21951] = {.lex_state = 457, .external_lex_state = 122},
  [21952] = {.lex_state = 455, .external_lex_state = 134},
  [21953] = {.lex_state = 455, .external_lex_state = 134},
  [21954] = {.lex_state = 455, .external_lex_state = 134},
  [21955] = {.lex_state = 457, .external_lex_state = 114},
  [21956] = {.lex_state = 457},
  [21957] = {.lex_state = 455},
  [21958] = {.lex_state = 457, .external_lex_state = 131},
  [21959] = {.lex_state = 455},
  [21960] = {.lex_state = 455, .external_lex_state = 134},
  [21961] = {.lex_state = 203},
  [21962] = {.lex_state = 457},
  [21963] = {.lex_state = 457, .external_lex_state = 114},
  [21964] = {.lex_state = 457, .external_lex_state = 122},
  [21965] = {.lex_state = 457, .external_lex_state = 131},
  [21966] = {.lex_state = 203},
  [21967] = {.lex_state = 457},
  [21968] = {.lex_state = 457},
  [21969] = {.lex_state = 203},
  [21970] = {.lex_state = 455},
  [21971] = {.lex_state = 457, .external_lex_state = 114},
  [21972] = {.lex_state = 457},
  [21973] = {.lex_state = 203},
  [21974] = {.lex_state = 457},
  [21975] = {.lex_state = 455, .external_lex_state = 134},
  [21976] = {.lex_state = 455, .external_lex_state = 134},
  [21977] = {.lex_state = 455, .external_lex_state = 134},
  [21978] = {.lex_state = 457, .external_lex_state = 131},
  [21979] = {.lex_state = 455},
  [21980] = {.lex_state = 457},
  [21981] = {.lex_state = 190},
  [21982] = {.lex_state = 455, .external_lex_state = 135},
  [21983] = {.lex_state = 457, .external_lex_state = 114},
  [21984] = {.lex_state = 457, .external_lex_state = 122},
  [21985] = {.lex_state = 457, .external_lex_state = 131},
  [21986] = {.lex_state = 203},
  [21987] = {.lex_state = 457},
  [21988] = {.lex_state = 457, .external_lex_state = 122},
  [21989] = {.lex_state = 455, .external_lex_state = 134},
  [21990] = {.lex_state = 457, .external_lex_state = 114},
  [21991] = {.lex_state = 457},
  [21992] = {.lex_state = 203},
  [21993] = {.lex_state = 190},
  [21994] = {.lex_state = 457},
  [21995] = {.lex_state = 455},
  [21996] = {.lex_state = 203},
  [21997] = {.lex_state = 190},
  [21998] = {.lex_state = 457, .external_lex_state = 114},
  [21999] = {.lex_state = 203},
  [22000] = {.lex_state = 457},
  [22001] = {.lex_state = 455, .external_lex_state = 134},
  [22002] = {.lex_state = 455},
  [22003] = {.lex_state = 203},
  [22004] = {.lex_state = 457, .external_lex_state = 114},
  [22005] = {.lex_state = 457},
  [22006] = {.lex_state = 457, .external_lex_state = 114},
  [22007] = {.lex_state = 457, .external_lex_state = 114},
  [22008] = {.lex_state = 455},
  [22009] = {.lex_state = 203},
  [22010] = {.lex_state = 455, .external_lex_state = 134},
  [22011] = {.lex_state = 457, .external_lex_state = 122},
  [22012] = {.lex_state = 457},
  [22013] = {.lex_state = 455, .external_lex_state = 134},
  [22014] = {.lex_state = 455},
  [22015] = {.lex_state = 203},
  [22016] = {.lex_state = 457},
  [22017] = {.lex_state = 455, .external_lex_state = 134},
  [22018] = {.lex_state = 455, .external_lex_state = 134},
  [22019] = {.lex_state = 457},
  [22020] = {.lex_state = 455},
  [22021] = {.lex_state = 203},
  [22022] = {.lex_state = 455, .external_lex_state = 134},
  [22023] = {.lex_state = 457, .external_lex_state = 114},
  [22024] = {.lex_state = 457},
  [22025] = {.lex_state = 457},
  [22026] = {.lex_state = 455},
  [22027] = {.lex_state = 203},
  [22028] = {.lex_state = 190},
  [22029] = {.lex_state = 457},
  [22030] = {.lex_state = 455, .external_lex_state = 134},
  [22031] = {.lex_state = 455},
  [22032] = {.lex_state = 203},
  [22033] = {.lex_state = 190},
  [22034] = {.lex_state = 457, .external_lex_state = 114},
  [22035] = {.lex_state = 457},
  [22036] = {.lex_state = 203},
  [22037] = {.lex_state = 203},
  [22038] = {.lex_state = 457},
  [22039] = {.lex_state = 457},
  [22040] = {.lex_state = 457, .external_lex_state = 114},
  [22041] = {.lex_state = 457, .external_lex_state = 114},
  [22042] = {.lex_state = 455},
  [22043] = {.lex_state = 457, .external_lex_state = 114},
  [22044] = {.lex_state = 457},
  [22045] = {.lex_state = 203},
  [22046] = {.lex_state = 457, .external_lex_state = 122},
  [22047] = {.lex_state = 457, .external_lex_state = 114},
  [22048] = {.lex_state = 457},
  [22049] = {.lex_state = 455},
  [22050] = {.lex_state = 457},
  [22051] = {.lex_state = 457},
  [22052] = {.lex_state = 203},
  [22053] = {.lex_state = 457, .external_lex_state = 114},
  [22054] = {.lex_state = 457},
  [22055] = {.lex_state = 203},
  [22056] = {.lex_state = 455, .external_lex_state = 134},
  [22057] = {.lex_state = 457},
  [22058] = {.lex_state = 457, .external_lex_state = 114},
  [22059] = {.lex_state = 457, .external_lex_state = 114},
  [22060] = {.lex_state = 190},
  [22061] = {.lex_state = 203},
  [22062] = {.lex_state = 457},
  [22063] = {.lex_state = 457},
  [22064] = {.lex_state = 190},
  [22065] = {.lex_state = 455},
  [22066] = {.lex_state = 457},
  [22067] = {.lex_state = 455, .external_lex_state = 134},
  [22068] = {.lex_state = 190},
  [22069] = {.lex_state = 224, .external_lex_state = 109},
  [22070] = {.lex_state = 455, .external_lex_state = 134},
  [22071] = {.lex_state = 457, .external_lex_state = 114},
  [22072] = {.lex_state = 457},
  [22073] = {.lex_state = 457, .external_lex_state = 122},
  [22074] = {.lex_state = 457},
  [22075] = {.lex_state = 457},
  [22076] = {.lex_state = 190},
  [22077] = {.lex_state = 457, .external_lex_state = 114},
  [22078] = {.lex_state = 457, .external_lex_state = 114},
  [22079] = {.lex_state = 457, .external_lex_state = 128},
  [22080] = {.lex_state = 455, .external_lex_state = 134},
  [22081] = {.lex_state = 457, .external_lex_state = 114},
  [22082] = {.lex_state = 457, .external_lex_state = 136},
  [22083] = {.lex_state = 455},
  [22084] = {.lex_state = 455, .external_lex_state = 135},
  [22085] = {.lex_state = 457},
  [22086] = {.lex_state = 457, .external_lex_state = 122},
  [22087] = {.lex_state = 457},
  [22088] = {.lex_state = 457, .external_lex_state = 122},
  [22089] = {.lex_state = 457, .external_lex_state = 122},
  [22090] = {.lex_state = 455},
  [22091] = {.lex_state = 457},
  [22092] = {.lex_state = 203},
  [22093] = {.lex_state = 457},
  [22094] = {.lex_state = 457},
  [22095] = {.lex_state = 457},
  [22096] = {.lex_state = 457},
  [22097] = {.lex_state = 457},
  [22098] = {.lex_state = 455, .external_lex_state = 134},
  [22099] = {.lex_state = 455},
  [22100] = {.lex_state = 455},
  [22101] = {.lex_state = 455},
  [22102] = {.lex_state = 224, .external_lex_state = 109},
  [22103] = {.lex_state = 457},
  [22104] = {.lex_state = 457},
  [22105] = {.lex_state = 457},
  [22106] = {.lex_state = 457, .external_lex_state = 114},
  [22107] = {.lex_state = 457},
  [22108] = {.lex_state = 457, .external_lex_state = 114},
  [22109] = {.lex_state = 457},
  [22110] = {.lex_state = 457, .external_lex_state = 114},
  [22111] = {.lex_state = 457, .external_lex_state = 114},
  [22112] = {.lex_state = 457, .external_lex_state = 114},
  [22113] = {.lex_state = 203},
  [22114] = {.lex_state = 455},
  [22115] = {.lex_state = 457, .external_lex_state = 114},
  [22116] = {.lex_state = 457},
  [22117] = {.lex_state = 455},
  [22118] = {.lex_state = 455, .external_lex_state = 134},
  [22119] = {.lex_state = 203},
  [22120] = {.lex_state = 455, .external_lex_state = 134},
  [22121] = {.lex_state = 457},
  [22122] = {.lex_state = 457},
  [22123] = {.lex_state = 457, .external_lex_state = 114},
  [22124] = {.lex_state = 457, .external_lex_state = 114},
  [22125] = {.lex_state = 457},
  [22126] = {.lex_state = 190},
  [22127] = {.lex_state = 457},
  [22128] = {.lex_state = 457, .external_lex_state = 114},
  [22129] = {.lex_state = 455},
  [22130] = {.lex_state = 457, .external_lex_state = 114},
  [22131] = {.lex_state = 455, .external_lex_state = 134},
  [22132] = {.lex_state = 457},
  [22133] = {.lex_state = 457},
  [22134] = {.lex_state = 457},
  [22135] = {.lex_state = 457, .external_lex_state = 114},
  [22136] = {.lex_state = 457, .external_lex_state = 114},
  [22137] = {.lex_state = 457, .external_lex_state = 131},
  [22138] = {.lex_state = 457},
  [22139] = {.lex_state = 457},
  [22140] = {.lex_state = 455, .external_lex_state = 134},
  [22141] = {.lex_state = 457},
  [22142] = {.lex_state = 457},
  [22143] = {.lex_state = 457, .external_lex_state = 114},
  [22144] = {.lex_state = 457},
  [22145] = {.lex_state = 457},
  [22146] = {.lex_state = 203},
  [22147] = {.lex_state = 203},
  [22148] = {.lex_state = 457},
  [22149] = {.lex_state = 203},
  [22150] = {.lex_state = 455},
  [22151] = {.lex_state = 457},
  [22152] = {.lex_state = 455},
  [22153] = {.lex_state = 455, .external_lex_state = 135},
  [22154] = {.lex_state = 457},
  [22155] = {.lex_state = 457, .external_lex_state = 122},
  [22156] = {.lex_state = 457, .external_lex_state = 114},
  [22157] = {.lex_state = 190},
  [22158] = {.lex_state = 457},
  [22159] = {.lex_state = 203},
  [22160] = {.lex_state = 190},
  [22161] = {.lex_state = 457},
  [22162] = {.lex_state = 203},
  [22163] = {.lex_state = 455},
  [22164] = {.lex_state = 457},
  [22165] = {.lex_state = 457},
  [22166] = {.lex_state = 190},
  [22167] = {.lex_state = 455},
  [22168] = {.lex_state = 457, .external_lex_state = 114},
  [22169] = {.lex_state = 457},
  [22170] = {.lex_state = 457, .external_lex_state = 114},
  [22171] = {.lex_state = 457},
  [22172] = {.lex_state = 455, .external_lex_state = 134},
  [22173] = {.lex_state = 457, .external_lex_state = 131},
  [22174] = {.lex_state = 457, .external_lex_state = 131},
  [22175] = {.lex_state = 457, .external_lex_state = 131},
  [22176] = {.lex_state = 457},
  [22177] = {.lex_state = 455, .external_lex_state = 134},
  [22178] = {.lex_state = 203},
  [22179] = {.lex_state = 203},
  [22180] = {.lex_state = 457, .external_lex_state = 131},
  [22181] = {.lex_state = 457, .external_lex_state = 122},
  [22182] = {.lex_state = 457},
  [22183] = {.lex_state = 455},
  [22184] = {.lex_state = 455, .external_lex_state = 134},
  [22185] = {.lex_state = 455},
  [22186] = {.lex_state = 457},
  [22187] = {.lex_state = 457, .external_lex_state = 131},
  [22188] = {.lex_state = 455},
  [22189] = {.lex_state = 457, .external_lex_state = 114},
  [22190] = {.lex_state = 455},
  [22191] = {.lex_state = 457, .external_lex_state = 114},
  [22192] = {.lex_state = 455},
  [22193] = {.lex_state = 457},
  [22194] = {.lex_state = 203},
  [22195] = {.lex_state = 455},
  [22196] = {.lex_state = 457},
  [22197] = {.lex_state = 455},
  [22198] = {.lex_state = 457, .external_lex_state = 122},
  [22199] = {.lex_state = 455},
  [22200] = {.lex_state = 457, .external_lex_state = 114},
  [22201] = {.lex_state = 455},
  [22202] = {.lex_state = 190},
  [22203] = {.lex_state = 455},
  [22204] = {.lex_state = 455},
  [22205] = {.lex_state = 457, .external_lex_state = 131},
  [22206] = {.lex_state = 457, .external_lex_state = 114},
  [22207] = {.lex_state = 455},
  [22208] = {.lex_state = 455},
  [22209] = {.lex_state = 455},
  [22210] = {.lex_state = 455},
  [22211] = {.lex_state = 455},
  [22212] = {.lex_state = 455, .external_lex_state = 134},
  [22213] = {.lex_state = 455, .external_lex_state = 134},
  [22214] = {.lex_state = 457},
  [22215] = {.lex_state = 457},
  [22216] = {.lex_state = 457},
  [22217] = {.lex_state = 457},
  [22218] = {.lex_state = 457},
  [22219] = {.lex_state = 457, .external_lex_state = 131},
  [22220] = {.lex_state = 457, .external_lex_state = 131},
  [22221] = {.lex_state = 457, .external_lex_state = 131},
  [22222] = {.lex_state = 457, .external_lex_state = 114},
  [22223] = {.lex_state = 457},
  [22224] = {.lex_state = 203},
  [22225] = {.lex_state = 457, .external_lex_state = 131},
  [22226] = {.lex_state = 457},
  [22227] = {.lex_state = 455},
  [22228] = {.lex_state = 457},
  [22229] = {.lex_state = 457, .external_lex_state = 131},
  [22230] = {.lex_state = 455},
  [22231] = {.lex_state = 455},
  [22232] = {.lex_state = 203},
  [22233] = {.lex_state = 457, .external_lex_state = 114},
  [22234] = {.lex_state = 203},
  [22235] = {.lex_state = 203},
  [22236] = {.lex_state = 455},
  [22237] = {.lex_state = 457, .external_lex_state = 114},
  [22238] = {.lex_state = 455},
  [22239] = {.lex_state = 455},
  [22240] = {.lex_state = 457},
  [22241] = {.lex_state = 457, .external_lex_state = 114},
  [22242] = {.lex_state = 457, .external_lex_state = 114},
  [22243] = {.lex_state = 455},
  [22244] = {.lex_state = 457},
  [22245] = {.lex_state = 457},
  [22246] = {.lex_state = 457},
  [22247] = {.lex_state = 457},
  [22248] = {.lex_state = 457},
  [22249] = {.lex_state = 457, .external_lex_state = 131},
  [22250] = {.lex_state = 457, .external_lex_state = 131},
  [22251] = {.lex_state = 457, .external_lex_state = 131},
  [22252] = {.lex_state = 457, .external_lex_state = 114},
  [22253] = {.lex_state = 190},
  [22254] = {.lex_state = 203},
  [22255] = {.lex_state = 457, .external_lex_state = 131},
  [22256] = {.lex_state = 203},
  [22257] = {.lex_state = 457},
  [22258] = {.lex_state = 457},
  [22259] = {.lex_state = 457, .external_lex_state = 131},
  [22260] = {.lex_state = 455},
  [22261] = {.lex_state = 457, .external_lex_state = 122},
  [22262] = {.lex_state = 457},
  [22263] = {.lex_state = 455, .external_lex_state = 134},
  [22264] = {.lex_state = 455},
  [22265] = {.lex_state = 457},
  [22266] = {.lex_state = 455},
  [22267] = {.lex_state = 455},
  [22268] = {.lex_state = 457, .external_lex_state = 136},
  [22269] = {.lex_state = 455},
  [22270] = {.lex_state = 455, .external_lex_state = 135},
  [22271] = {.lex_state = 455},
  [22272] = {.lex_state = 457},
  [22273] = {.lex_state = 457},
  [22274] = {.lex_state = 457},
  [22275] = {.lex_state = 457},
  [22276] = {.lex_state = 457},
  [22277] = {.lex_state = 457, .external_lex_state = 131},
  [22278] = {.lex_state = 457, .external_lex_state = 131},
  [22279] = {.lex_state = 457, .external_lex_state = 131},
  [22280] = {.lex_state = 190},
  [22281] = {.lex_state = 457},
  [22282] = {.lex_state = 203},
  [22283] = {.lex_state = 457, .external_lex_state = 131},
  [22284] = {.lex_state = 457},
  [22285] = {.lex_state = 457, .external_lex_state = 114},
  [22286] = {.lex_state = 457},
  [22287] = {.lex_state = 457, .external_lex_state = 131},
  [22288] = {.lex_state = 455},
  [22289] = {.lex_state = 457},
  [22290] = {.lex_state = 457},
  [22291] = {.lex_state = 457},
  [22292] = {.lex_state = 455},
  [22293] = {.lex_state = 455},
  [22294] = {.lex_state = 455},
  [22295] = {.lex_state = 455},
  [22296] = {.lex_state = 457},
  [22297] = {.lex_state = 203},
  [22298] = {.lex_state = 455},
  [22299] = {.lex_state = 455},
  [22300] = {.lex_state = 457},
  [22301] = {.lex_state = 457},
  [22302] = {.lex_state = 457},
  [22303] = {.lex_state = 457},
  [22304] = {.lex_state = 457},
  [22305] = {.lex_state = 457, .external_lex_state = 131},
  [22306] = {.lex_state = 457, .external_lex_state = 131},
  [22307] = {.lex_state = 457, .external_lex_state = 131},
  [22308] = {.lex_state = 457},
  [22309] = {.lex_state = 457},
  [22310] = {.lex_state = 203},
  [22311] = {.lex_state = 457, .external_lex_state = 131},
  [22312] = {.lex_state = 224, .external_lex_state = 109},
  [22313] = {.lex_state = 457},
  [22314] = {.lex_state = 455, .external_lex_state = 134},
  [22315] = {.lex_state = 457, .external_lex_state = 131},
  [22316] = {.lex_state = 455},
  [22317] = {.lex_state = 190},
  [22318] = {.lex_state = 457},
  [22319] = {.lex_state = 457, .external_lex_state = 114},
  [22320] = {.lex_state = 455},
  [22321] = {.lex_state = 190},
  [22322] = {.lex_state = 455},
  [22323] = {.lex_state = 455},
  [22324] = {.lex_state = 457, .external_lex_state = 114},
  [22325] = {.lex_state = 457},
  [22326] = {.lex_state = 457},
  [22327] = {.lex_state = 455},
  [22328] = {.lex_state = 457, .external_lex_state = 131},
  [22329] = {.lex_state = 457, .external_lex_state = 131},
  [22330] = {.lex_state = 457, .external_lex_state = 131},
  [22331] = {.lex_state = 457},
  [22332] = {.lex_state = 457},
  [22333] = {.lex_state = 203},
  [22334] = {.lex_state = 457, .external_lex_state = 131},
  [22335] = {.lex_state = 190},
  [22336] = {.lex_state = 457, .external_lex_state = 131},
  [22337] = {.lex_state = 457, .external_lex_state = 131},
  [22338] = {.lex_state = 455},
  [22339] = {.lex_state = 455},
  [22340] = {.lex_state = 457},
  [22341] = {.lex_state = 455, .external_lex_state = 134},
  [22342] = {.lex_state = 455, .external_lex_state = 134},
  [22343] = {.lex_state = 455},
  [22344] = {.lex_state = 457, .external_lex_state = 131},
  [22345] = {.lex_state = 457, .external_lex_state = 131},
  [22346] = {.lex_state = 457, .external_lex_state = 131},
  [22347] = {.lex_state = 457, .external_lex_state = 114},
  [22348] = {.lex_state = 457, .external_lex_state = 114},
  [22349] = {.lex_state = 203},
  [22350] = {.lex_state = 457, .external_lex_state = 131},
  [22351] = {.lex_state = 457},
  [22352] = {.lex_state = 457},
  [22353] = {.lex_state = 457, .external_lex_state = 131},
  [22354] = {.lex_state = 455},
  [22355] = {.lex_state = 455},
  [22356] = {.lex_state = 457},
  [22357] = {.lex_state = 455},
  [22358] = {.lex_state = 455},
  [22359] = {.lex_state = 455},
  [22360] = {.lex_state = 457, .external_lex_state = 131},
  [22361] = {.lex_state = 457, .external_lex_state = 131},
  [22362] = {.lex_state = 457, .external_lex_state = 131},
  [22363] = {.lex_state = 455},
  [22364] = {.lex_state = 203},
  [22365] = {.lex_state = 457, .external_lex_state = 131},
  [22366] = {.lex_state = 457, .external_lex_state = 122},
  [22367] = {.lex_state = 457, .external_lex_state = 131},
  [22368] = {.lex_state = 457},
  [22369] = {.lex_state = 455},
  [22370] = {.lex_state = 457, .external_lex_state = 114},
  [22371] = {.lex_state = 455},
  [22372] = {.lex_state = 457, .external_lex_state = 131},
  [22373] = {.lex_state = 457, .external_lex_state = 131},
  [22374] = {.lex_state = 457, .external_lex_state = 131},
  [22375] = {.lex_state = 457},
  [22376] = {.lex_state = 203},
  [22377] = {.lex_state = 457, .external_lex_state = 131},
  [22378] = {.lex_state = 457, .external_lex_state = 114},
  [22379] = {.lex_state = 457, .external_lex_state = 131},
  [22380] = {.lex_state = 457, .external_lex_state = 131},
  [22381] = {.lex_state = 455},
  [22382] = {.lex_state = 203},
  [22383] = {.lex_state = 455},
  [22384] = {.lex_state = 457, .external_lex_state = 131},
  [22385] = {.lex_state = 457, .external_lex_state = 131},
  [22386] = {.lex_state = 457, .external_lex_state = 131},
  [22387] = {.lex_state = 457, .external_lex_state = 122},
  [22388] = {.lex_state = 203},
  [22389] = {.lex_state = 457, .external_lex_state = 131},
  [22390] = {.lex_state = 455},
  [22391] = {.lex_state = 457, .external_lex_state = 131},
  [22392] = {.lex_state = 457, .external_lex_state = 122},
  [22393] = {.lex_state = 455},
  [22394] = {.lex_state = 457, .external_lex_state = 131},
  [22395] = {.lex_state = 457, .external_lex_state = 131},
  [22396] = {.lex_state = 457, .external_lex_state = 131},
  [22397] = {.lex_state = 457, .external_lex_state = 122},
  [22398] = {.lex_state = 203},
  [22399] = {.lex_state = 457, .external_lex_state = 131},
  [22400] = {.lex_state = 455},
  [22401] = {.lex_state = 457, .external_lex_state = 131},
  [22402] = {.lex_state = 457},
  [22403] = {.lex_state = 455},
  [22404] = {.lex_state = 457, .external_lex_state = 131},
  [22405] = {.lex_state = 457, .external_lex_state = 131},
  [22406] = {.lex_state = 457, .external_lex_state = 131},
  [22407] = {.lex_state = 457, .external_lex_state = 114},
  [22408] = {.lex_state = 203},
  [22409] = {.lex_state = 457, .external_lex_state = 131},
  [22410] = {.lex_state = 457},
  [22411] = {.lex_state = 457, .external_lex_state = 131},
  [22412] = {.lex_state = 457, .external_lex_state = 122},
  [22413] = {.lex_state = 455},
  [22414] = {.lex_state = 457, .external_lex_state = 131},
  [22415] = {.lex_state = 457, .external_lex_state = 131},
  [22416] = {.lex_state = 457, .external_lex_state = 131},
  [22417] = {.lex_state = 457, .external_lex_state = 122},
  [22418] = {.lex_state = 203},
  [22419] = {.lex_state = 457, .external_lex_state = 131},
  [22420] = {.lex_state = 457},
  [22421] = {.lex_state = 457, .external_lex_state = 131},
  [22422] = {.lex_state = 203},
  [22423] = {.lex_state = 457, .external_lex_state = 131},
  [22424] = {.lex_state = 457, .external_lex_state = 131},
  [22425] = {.lex_state = 457, .external_lex_state = 131},
  [22426] = {.lex_state = 457, .external_lex_state = 114},
  [22427] = {.lex_state = 203},
  [22428] = {.lex_state = 457, .external_lex_state = 131},
  [22429] = {.lex_state = 457},
  [22430] = {.lex_state = 457, .external_lex_state = 131},
  [22431] = {.lex_state = 190},
  [22432] = {.lex_state = 457, .external_lex_state = 131},
  [22433] = {.lex_state = 457, .external_lex_state = 131},
  [22434] = {.lex_state = 457, .external_lex_state = 131},
  [22435] = {.lex_state = 455},
  [22436] = {.lex_state = 203},
  [22437] = {.lex_state = 457, .external_lex_state = 131},
  [22438] = {.lex_state = 457, .external_lex_state = 114},
  [22439] = {.lex_state = 457, .external_lex_state = 131},
  [22440] = {.lex_state = 455, .external_lex_state = 135},
  [22441] = {.lex_state = 457, .external_lex_state = 131},
  [22442] = {.lex_state = 457, .external_lex_state = 131},
  [22443] = {.lex_state = 457, .external_lex_state = 131},
  [22444] = {.lex_state = 203},
  [22445] = {.lex_state = 203},
  [22446] = {.lex_state = 457, .external_lex_state = 131},
  [22447] = {.lex_state = 457},
  [22448] = {.lex_state = 457, .external_lex_state = 131},
  [22449] = {.lex_state = 457, .external_lex_state = 122},
  [22450] = {.lex_state = 457, .external_lex_state = 131},
  [22451] = {.lex_state = 457, .external_lex_state = 131},
  [22452] = {.lex_state = 457, .external_lex_state = 131},
  [22453] = {.lex_state = 457, .external_lex_state = 114},
  [22454] = {.lex_state = 203},
  [22455] = {.lex_state = 457, .external_lex_state = 131},
  [22456] = {.lex_state = 455},
  [22457] = {.lex_state = 457, .external_lex_state = 131},
  [22458] = {.lex_state = 457, .external_lex_state = 122},
  [22459] = {.lex_state = 457, .external_lex_state = 131},
  [22460] = {.lex_state = 457, .external_lex_state = 131},
  [22461] = {.lex_state = 457, .external_lex_state = 131},
  [22462] = {.lex_state = 455},
  [22463] = {.lex_state = 203},
  [22464] = {.lex_state = 457, .external_lex_state = 131},
  [22465] = {.lex_state = 457},
  [22466] = {.lex_state = 457, .external_lex_state = 131},
  [22467] = {.lex_state = 455},
  [22468] = {.lex_state = 457, .external_lex_state = 131},
  [22469] = {.lex_state = 457, .external_lex_state = 131},
  [22470] = {.lex_state = 457, .external_lex_state = 131},
  [22471] = {.lex_state = 455},
  [22472] = {.lex_state = 203},
  [22473] = {.lex_state = 457, .external_lex_state = 131},
  [22474] = {.lex_state = 203},
  [22475] = {.lex_state = 457, .external_lex_state = 131},
  [22476] = {.lex_state = 457},
  [22477] = {.lex_state = 457, .external_lex_state = 131},
  [22478] = {.lex_state = 457, .external_lex_state = 131},
  [22479] = {.lex_state = 457, .external_lex_state = 131},
  [22480] = {.lex_state = 203},
  [22481] = {.lex_state = 457, .external_lex_state = 131},
  [22482] = {.lex_state = 455},
  [22483] = {.lex_state = 457, .external_lex_state = 131},
  [22484] = {.lex_state = 455},
  [22485] = {.lex_state = 457, .external_lex_state = 131},
  [22486] = {.lex_state = 457, .external_lex_state = 131},
  [22487] = {.lex_state = 457, .external_lex_state = 131},
  [22488] = {.lex_state = 457},
  [22489] = {.lex_state = 203},
  [22490] = {.lex_state = 457, .external_lex_state = 131},
  [22491] = {.lex_state = 457, .external_lex_state = 114},
  [22492] = {.lex_state = 457, .external_lex_state = 131},
  [22493] = {.lex_state = 455},
  [22494] = {.lex_state = 457, .external_lex_state = 131},
  [22495] = {.lex_state = 457, .external_lex_state = 131},
  [22496] = {.lex_state = 457, .external_lex_state = 131},
  [22497] = {.lex_state = 457},
  [22498] = {.lex_state = 203},
  [22499] = {.lex_state = 457, .external_lex_state = 131},
  [22500] = {.lex_state = 457},
  [22501] = {.lex_state = 457, .external_lex_state = 131},
  [22502] = {.lex_state = 457},
  [22503] = {.lex_state = 457, .external_lex_state = 131},
  [22504] = {.lex_state = 457, .external_lex_state = 131},
  [22505] = {.lex_state = 457, .external_lex_state = 131},
  [22506] = {.lex_state = 190},
  [22507] = {.lex_state = 203},
  [22508] = {.lex_state = 457, .external_lex_state = 131},
  [22509] = {.lex_state = 457},
  [22510] = {.lex_state = 457, .external_lex_state = 131},
  [22511] = {.lex_state = 203},
  [22512] = {.lex_state = 457, .external_lex_state = 131},
  [22513] = {.lex_state = 457, .external_lex_state = 131},
  [22514] = {.lex_state = 457, .external_lex_state = 131},
  [22515] = {.lex_state = 457},
  [22516] = {.lex_state = 203},
  [22517] = {.lex_state = 457, .external_lex_state = 131},
  [22518] = {.lex_state = 457, .external_lex_state = 114},
  [22519] = {.lex_state = 457, .external_lex_state = 131},
  [22520] = {.lex_state = 457},
  [22521] = {.lex_state = 457, .external_lex_state = 131},
  [22522] = {.lex_state = 457, .external_lex_state = 131},
  [22523] = {.lex_state = 457, .external_lex_state = 131},
  [22524] = {.lex_state = 457, .external_lex_state = 136},
  [22525] = {.lex_state = 203},
  [22526] = {.lex_state = 457, .external_lex_state = 131},
  [22527] = {.lex_state = 457},
  [22528] = {.lex_state = 457, .external_lex_state = 131},
  [22529] = {.lex_state = 203},
  [22530] = {.lex_state = 457, .external_lex_state = 131},
  [22531] = {.lex_state = 457, .external_lex_state = 131},
  [22532] = {.lex_state = 457, .external_lex_state = 131},
  [22533] = {.lex_state = 190},
  [22534] = {.lex_state = 203},
  [22535] = {.lex_state = 457, .external_lex_state = 131},
  [22536] = {.lex_state = 457},
  [22537] = {.lex_state = 457, .external_lex_state = 131},
  [22538] = {.lex_state = 457, .external_lex_state = 114},
  [22539] = {.lex_state = 203},
  [22540] = {.lex_state = 457, .external_lex_state = 131},
  [22541] = {.lex_state = 190},
  [22542] = {.lex_state = 457, .external_lex_state = 131},
  [22543] = {.lex_state = 203},
  [22544] = {.lex_state = 457, .external_lex_state = 131},
  [22545] = {.lex_state = 457},
  [22546] = {.lex_state = 457, .external_lex_state = 131},
  [22547] = {.lex_state = 203},
  [22548] = {.lex_state = 457, .external_lex_state = 131},
  [22549] = {.lex_state = 457, .external_lex_state = 114},
  [22550] = {.lex_state = 457, .external_lex_state = 131},
  [22551] = {.lex_state = 203},
  [22552] = {.lex_state = 457, .external_lex_state = 131},
  [22553] = {.lex_state = 457},
  [22554] = {.lex_state = 457, .external_lex_state = 131},
  [22555] = {.lex_state = 203},
  [22556] = {.lex_state = 457, .external_lex_state = 131},
  [22557] = {.lex_state = 457},
  [22558] = {.lex_state = 457, .external_lex_state = 131},
  [22559] = {.lex_state = 203},
  [22560] = {.lex_state = 457, .external_lex_state = 131},
  [22561] = {.lex_state = 457, .external_lex_state = 131},
  [22562] = {.lex_state = 203},
  [22563] = {.lex_state = 457, .external_lex_state = 131},
  [22564] = {.lex_state = 457, .external_lex_state = 131},
  [22565] = {.lex_state = 457, .external_lex_state = 131},
  [22566] = {.lex_state = 457, .external_lex_state = 131},
  [22567] = {.lex_state = 457, .external_lex_state = 131},
  [22568] = {.lex_state = 457, .external_lex_state = 131},
  [22569] = {.lex_state = 457, .external_lex_state = 131},
  [22570] = {.lex_state = 457, .external_lex_state = 131},
  [22571] = {.lex_state = 457, .external_lex_state = 131},
  [22572] = {.lex_state = 457, .external_lex_state = 131},
  [22573] = {.lex_state = 457, .external_lex_state = 131},
  [22574] = {.lex_state = 457, .external_lex_state = 131},
  [22575] = {.lex_state = 457, .external_lex_state = 131},
  [22576] = {.lex_state = 457, .external_lex_state = 131},
  [22577] = {.lex_state = 457, .external_lex_state = 131},
  [22578] = {.lex_state = 457, .external_lex_state = 131},
  [22579] = {.lex_state = 457, .external_lex_state = 131},
  [22580] = {.lex_state = 457, .external_lex_state = 131},
  [22581] = {.lex_state = 457, .external_lex_state = 131},
  [22582] = {.lex_state = 457, .external_lex_state = 131},
  [22583] = {.lex_state = 457, .external_lex_state = 131},
  [22584] = {.lex_state = 457, .external_lex_state = 131},
  [22585] = {.lex_state = 457, .external_lex_state = 131},
  [22586] = {.lex_state = 457, .external_lex_state = 131},
  [22587] = {.lex_state = 457, .external_lex_state = 131},
  [22588] = {.lex_state = 457, .external_lex_state = 131},
  [22589] = {.lex_state = 457, .external_lex_state = 131},
  [22590] = {.lex_state = 457, .external_lex_state = 131},
  [22591] = {.lex_state = 457, .external_lex_state = 131},
  [22592] = {.lex_state = 457, .external_lex_state = 131},
  [22593] = {.lex_state = 457, .external_lex_state = 131},
  [22594] = {.lex_state = 457, .external_lex_state = 131},
  [22595] = {.lex_state = 457, .external_lex_state = 131},
  [22596] = {.lex_state = 457, .external_lex_state = 131},
  [22597] = {.lex_state = 457, .external_lex_state = 131},
  [22598] = {.lex_state = 457, .external_lex_state = 131},
  [22599] = {.lex_state = 457, .external_lex_state = 131},
  [22600] = {.lex_state = 457, .external_lex_state = 131},
  [22601] = {.lex_state = 457, .external_lex_state = 131},
  [22602] = {.lex_state = 457, .external_lex_state = 131},
  [22603] = {.lex_state = 457, .external_lex_state = 131},
  [22604] = {.lex_state = 457, .external_lex_state = 131},
  [22605] = {.lex_state = 457, .external_lex_state = 131},
  [22606] = {.lex_state = 457, .external_lex_state = 131},
  [22607] = {.lex_state = 457, .external_lex_state = 131},
  [22608] = {.lex_state = 457, .external_lex_state = 131},
  [22609] = {.lex_state = 457, .external_lex_state = 131},
  [22610] = {.lex_state = 457, .external_lex_state = 131},
  [22611] = {.lex_state = 457, .external_lex_state = 131},
  [22612] = {.lex_state = 457, .external_lex_state = 131},
  [22613] = {.lex_state = 457, .external_lex_state = 131},
  [22614] = {.lex_state = 457, .external_lex_state = 131},
  [22615] = {.lex_state = 457, .external_lex_state = 131},
  [22616] = {.lex_state = 457, .external_lex_state = 131},
  [22617] = {.lex_state = 457, .external_lex_state = 131},
  [22618] = {.lex_state = 457, .external_lex_state = 131},
  [22619] = {.lex_state = 457, .external_lex_state = 131},
  [22620] = {.lex_state = 457, .external_lex_state = 131},
  [22621] = {.lex_state = 457, .external_lex_state = 131},
  [22622] = {.lex_state = 457, .external_lex_state = 131},
  [22623] = {.lex_state = 457, .external_lex_state = 131},
  [22624] = {.lex_state = 457, .external_lex_state = 131},
  [22625] = {.lex_state = 457, .external_lex_state = 131},
  [22626] = {.lex_state = 457, .external_lex_state = 131},
  [22627] = {.lex_state = 457, .external_lex_state = 131},
  [22628] = {.lex_state = 457, .external_lex_state = 131},
  [22629] = {.lex_state = 457, .external_lex_state = 131},
  [22630] = {.lex_state = 457, .external_lex_state = 131},
  [22631] = {.lex_state = 457, .external_lex_state = 131},
  [22632] = {.lex_state = 457, .external_lex_state = 131},
  [22633] = {.lex_state = 457, .external_lex_state = 131},
  [22634] = {.lex_state = 457, .external_lex_state = 131},
  [22635] = {.lex_state = 457, .external_lex_state = 131},
  [22636] = {.lex_state = 457, .external_lex_state = 131},
  [22637] = {.lex_state = 457, .external_lex_state = 131},
  [22638] = {.lex_state = 457, .external_lex_state = 131},
  [22639] = {.lex_state = 457, .external_lex_state = 131},
  [22640] = {.lex_state = 457, .external_lex_state = 131},
  [22641] = {.lex_state = 457, .external_lex_state = 131},
  [22642] = {.lex_state = 457, .external_lex_state = 131},
  [22643] = {.lex_state = 457, .external_lex_state = 131},
  [22644] = {.lex_state = 457, .external_lex_state = 131},
  [22645] = {.lex_state = 457, .external_lex_state = 131},
  [22646] = {.lex_state = 457, .external_lex_state = 131},
  [22647] = {.lex_state = 457, .external_lex_state = 131},
  [22648] = {.lex_state = 457, .external_lex_state = 131},
  [22649] = {.lex_state = 457, .external_lex_state = 131},
  [22650] = {.lex_state = 457, .external_lex_state = 131},
  [22651] = {.lex_state = 457, .external_lex_state = 131},
  [22652] = {.lex_state = 457, .external_lex_state = 131},
  [22653] = {.lex_state = 457, .external_lex_state = 131},
  [22654] = {.lex_state = 457, .external_lex_state = 131},
  [22655] = {.lex_state = 457, .external_lex_state = 131},
  [22656] = {.lex_state = 457, .external_lex_state = 131},
  [22657] = {.lex_state = 457, .external_lex_state = 131},
  [22658] = {.lex_state = 457, .external_lex_state = 131},
  [22659] = {.lex_state = 457, .external_lex_state = 131},
  [22660] = {.lex_state = 457, .external_lex_state = 131},
  [22661] = {.lex_state = 457, .external_lex_state = 131},
  [22662] = {.lex_state = 457, .external_lex_state = 131},
  [22663] = {.lex_state = 457, .external_lex_state = 131},
  [22664] = {.lex_state = 457, .external_lex_state = 131},
  [22665] = {.lex_state = 457, .external_lex_state = 131},
  [22666] = {.lex_state = 457, .external_lex_state = 131},
  [22667] = {.lex_state = 457, .external_lex_state = 131},
  [22668] = {.lex_state = 457, .external_lex_state = 131},
  [22669] = {.lex_state = 457, .external_lex_state = 131},
  [22670] = {.lex_state = 457, .external_lex_state = 131},
  [22671] = {.lex_state = 457, .external_lex_state = 131},
  [22672] = {.lex_state = 457, .external_lex_state = 131},
  [22673] = {.lex_state = 457, .external_lex_state = 131},
  [22674] = {.lex_state = 457, .external_lex_state = 131},
  [22675] = {.lex_state = 457, .external_lex_state = 131},
  [22676] = {.lex_state = 457, .external_lex_state = 131},
  [22677] = {.lex_state = 457, .external_lex_state = 131},
  [22678] = {.lex_state = 457, .external_lex_state = 131},
  [22679] = {.lex_state = 457, .external_lex_state = 131},
  [22680] = {.lex_state = 457, .external_lex_state = 131},
  [22681] = {.lex_state = 457, .external_lex_state = 131},
  [22682] = {.lex_state = 457, .external_lex_state = 131},
  [22683] = {.lex_state = 457, .external_lex_state = 131},
  [22684] = {.lex_state = 457, .external_lex_state = 131},
  [22685] = {.lex_state = 457, .external_lex_state = 131},
  [22686] = {.lex_state = 457, .external_lex_state = 131},
  [22687] = {.lex_state = 457, .external_lex_state = 131},
  [22688] = {.lex_state = 457, .external_lex_state = 131},
  [22689] = {.lex_state = 457, .external_lex_state = 131},
  [22690] = {.lex_state = 457, .external_lex_state = 131},
  [22691] = {.lex_state = 457, .external_lex_state = 131},
  [22692] = {.lex_state = 457, .external_lex_state = 131},
  [22693] = {.lex_state = 457, .external_lex_state = 131},
  [22694] = {.lex_state = 457, .external_lex_state = 131},
  [22695] = {.lex_state = 457, .external_lex_state = 131},
  [22696] = {.lex_state = 457, .external_lex_state = 131},
  [22697] = {.lex_state = 457, .external_lex_state = 131},
  [22698] = {.lex_state = 457, .external_lex_state = 131},
  [22699] = {.lex_state = 457, .external_lex_state = 131},
  [22700] = {.lex_state = 457, .external_lex_state = 114},
  [22701] = {.lex_state = 457},
  [22702] = {.lex_state = 224, .external_lex_state = 109},
  [22703] = {.lex_state = 190},
  [22704] = {.lex_state = 457},
  [22705] = {.lex_state = 190},
  [22706] = {.lex_state = 457, .external_lex_state = 131},
  [22707] = {.lex_state = 457, .external_lex_state = 131},
  [22708] = {.lex_state = 455},
  [22709] = {.lex_state = 203},
  [22710] = {.lex_state = 203},
  [22711] = {.lex_state = 457, .external_lex_state = 114},
  [22712] = {.lex_state = 455},
  [22713] = {.lex_state = 457, .external_lex_state = 131},
  [22714] = {.lex_state = 457, .external_lex_state = 131},
  [22715] = {.lex_state = 455},
  [22716] = {.lex_state = 203},
  [22717] = {.lex_state = 203},
  [22718] = {.lex_state = 457, .external_lex_state = 114},
  [22719] = {.lex_state = 457, .external_lex_state = 131},
  [22720] = {.lex_state = 457, .external_lex_state = 131},
  [22721] = {.lex_state = 455},
  [22722] = {.lex_state = 203},
  [22723] = {.lex_state = 457, .external_lex_state = 122},
  [22724] = {.lex_state = 457, .external_lex_state = 122},
  [22725] = {.lex_state = 457, .external_lex_state = 131},
  [22726] = {.lex_state = 457, .external_lex_state = 131},
  [22727] = {.lex_state = 455},
  [22728] = {.lex_state = 203},
  [22729] = {.lex_state = 457, .external_lex_state = 122},
  [22730] = {.lex_state = 457},
  [22731] = {.lex_state = 457, .external_lex_state = 131},
  [22732] = {.lex_state = 457, .external_lex_state = 131},
  [22733] = {.lex_state = 455},
  [22734] = {.lex_state = 203},
  [22735] = {.lex_state = 457, .external_lex_state = 122},
  [22736] = {.lex_state = 457},
  [22737] = {.lex_state = 457, .external_lex_state = 131},
  [22738] = {.lex_state = 457, .external_lex_state = 131},
  [22739] = {.lex_state = 455},
  [22740] = {.lex_state = 203},
  [22741] = {.lex_state = 457, .external_lex_state = 114},
  [22742] = {.lex_state = 457},
  [22743] = {.lex_state = 457, .external_lex_state = 131},
  [22744] = {.lex_state = 203},
  [22745] = {.lex_state = 455},
  [22746] = {.lex_state = 203},
  [22747] = {.lex_state = 203},
  [22748] = {.lex_state = 457, .external_lex_state = 122},
  [22749] = {.lex_state = 455},
  [22750] = {.lex_state = 203},
  [22751] = {.lex_state = 457, .external_lex_state = 122},
  [22752] = {.lex_state = 457},
  [22753] = {.lex_state = 455},
  [22754] = {.lex_state = 203},
  [22755] = {.lex_state = 457},
  [22756] = {.lex_state = 190},
  [22757] = {.lex_state = 457},
  [22758] = {.lex_state = 455},
  [22759] = {.lex_state = 457, .external_lex_state = 136},
  [22760] = {.lex_state = 455},
  [22761] = {.lex_state = 455, .external_lex_state = 135},
  [22762] = {.lex_state = 455},
  [22763] = {.lex_state = 457},
  [22764] = {.lex_state = 457},
  [22765] = {.lex_state = 203},
  [22766] = {.lex_state = 455},
  [22767] = {.lex_state = 455},
  [22768] = {.lex_state = 457},
  [22769] = {.lex_state = 457},
  [22770] = {.lex_state = 457},
  [22771] = {.lex_state = 203},
  [22772] = {.lex_state = 457},
  [22773] = {.lex_state = 457},
  [22774] = {.lex_state = 455},
  [22775] = {.lex_state = 1},
  [22776] = {.lex_state = 203},
  [22777] = {.lex_state = 224, .external_lex_state = 109},
  [22778] = {.lex_state = 457},
  [22779] = {.lex_state = 203},
  [22780] = {.lex_state = 457, .external_lex_state = 122},
  [22781] = {.lex_state = 457, .external_lex_state = 114},
  [22782] = {.lex_state = 457},
  [22783] = {.lex_state = 457, .external_lex_state = 114},
  [22784] = {.lex_state = 455},
  [22785] = {.lex_state = 457, .external_lex_state = 114},
  [22786] = {.lex_state = 457},
  [22787] = {.lex_state = 203},
  [22788] = {.lex_state = 457, .external_lex_state = 114},
  [22789] = {.lex_state = 457},
  [22790] = {.lex_state = 457},
  [22791] = {.lex_state = 1},
  [22792] = {.lex_state = 1},
  [22793] = {.lex_state = 457},
  [22794] = {.lex_state = 457, .external_lex_state = 114},
  [22795] = {.lex_state = 455},
  [22796] = {.lex_state = 457},
  [22797] = {.lex_state = 457},
  [22798] = {.lex_state = 457, .external_lex_state = 114},
  [22799] = {.lex_state = 457, .external_lex_state = 114},
  [22800] = {.lex_state = 457, .external_lex_state = 122},
  [22801] = {.lex_state = 457, .external_lex_state = 122},
  [22802] = {.lex_state = 457, .external_lex_state = 114},
  [22803] = {.lex_state = 457, .external_lex_state = 114},
  [22804] = {.lex_state = 457},
  [22805] = {.lex_state = 457, .external_lex_state = 122},
  [22806] = {.lex_state = 457},
  [22807] = {.lex_state = 457},
  [22808] = {.lex_state = 190},
  [22809] = {.lex_state = 203},
  [22810] = {.lex_state = 190},
  [22811] = {.lex_state = 457},
  [22812] = {.lex_state = 203},
  [22813] = {.lex_state = 457},
  [22814] = {.lex_state = 455},
  [22815] = {.lex_state = 457},
  [22816] = {.lex_state = 457, .external_lex_state = 122},
  [22817] = {.lex_state = 457},
  [22818] = {.lex_state = 455},
  [22819] = {.lex_state = 457},
  [22820] = {.lex_state = 457, .external_lex_state = 131},
  [22821] = {(TSStateId)(-1),},
  [22822] = {(TSStateId)(-1),},
  [22823] = {(TSStateId)(-1),},
  [22824] = {(TSStateId)(-1),},
  [22825] = {(TSStateId)(-1),},
  [22826] = {(TSStateId)(-1),},
  [22827] = {(TSStateId)(-1),},
  [22828] = {(TSStateId)(-1),},
  [22829] = {(TSStateId)(-1),},
  [22830] = {(TSStateId)(-1),},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_block_comment] = STATE(0),
    [sym_compiler_directive_decl] = STATE(0),
    [sym_fsi_directive_decl] = STATE(0),
    [sym_preproc_line] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_namespace] = ACTIONS(1),
    [anon_sym_global] = ACTIONS(1),
    [anon_sym_rec] = ACTIONS(1),
    [anon_sym_module] = ACTIONS(1),
    [anon_sym_null] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_begin] = ACTIONS(1),
    [anon_sym_end] = ACTIONS(1),
    [anon_sym_open] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_LBRACK_LT] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_let_BANG] = ACTIONS(1),
    [anon_sym_inline] = ACTIONS(1),
    [anon_sym_mutable] = ACTIONS(1),
    [aux_sym_access_modifier_token1] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_COLON_QMARK] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_LT_AT] = ACTIONS(1),
    [anon_sym_LT_AT_AT] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_RBRACE] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_return_BANG] = ACTIONS(1),
    [anon_sym_yield] = ACTIONS(1),
    [anon_sym_yield_BANG] = ACTIONS(1),
    [anon_sym_lazy] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_upcast] = ACTIONS(1),
    [anon_sym_downcast] = ACTIONS(1),
    [anon_sym_fixed] = ACTIONS(1),
    [anon_sym_COLON_GT] = ACTIONS(1),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_to] = ACTIONS(1),
    [anon_sym_downto] = ACTIONS(1),
    [anon_sym_done] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_while_BANG] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_then] = ACTIONS(1),
    [anon_sym_elif] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_fun] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_finally] = ACTIONS(1),
    [anon_sym_match] = ACTIONS(1),
    [anon_sym_match_BANG] = ACTIONS(1),
    [anon_sym_function] = ACTIONS(1),
    [anon_sym_LT_DASH] = ACTIONS(1),
    [anon_sym_DOT_LBRACK] = ACTIONS(1),
    [anon_sym_use] = ACTIONS(1),
    [anon_sym_use_BANG] = ACTIONS(1),
    [anon_sym_and_BANG] = ACTIONS(1),
    [anon_sym_do_BANG] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_when] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_DOT_DOT2] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_1] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_unit] = ACTIONS(1),
    [anon_sym_not] = ACTIONS(1),
    [anon_sym_enum] = ACTIONS(1),
    [anon_sym_unmanaged] = ACTIONS(1),
    [anon_sym_equality] = ACTIONS(1),
    [anon_sym_comparison] = ACTIONS(1),
    [anon_sym_delegate] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_static] = ACTIONS(1),
    [anon_sym_member] = ACTIONS(1),
    [anon_sym_get] = ACTIONS(1),
    [anon_sym_set] = ACTIONS(1),
    [anon_sym_exception] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_interface] = ACTIONS(1),
    [anon_sym_abstract] = ACTIONS(1),
    [anon_sym_val] = ACTIONS(1),
    [anon_sym_override] = ACTIONS(1),
    [anon_sym_extern] = ACTIONS(1),
    [anon_sym_inherit] = ACTIONS(1),
    [anon_sym_EQ2] = ACTIONS(1),
    [sym__unicodegraph_short] = ACTIONS(1),
    [sym__unicodegraph_long] = ACTIONS(1),
    [sym__trigraph] = ACTIONS(1),
    [sym__hexgraph_short] = ACTIONS(1),
    [aux_sym_char_token1] = ACTIONS(1),
    [anon_sym_LBRACE2] = ACTIONS(1),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_AT_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [anon_sym_DQUOTEB] = ACTIONS(1),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1),
    [sym_bool] = ACTIONS(1),
    [sym_unit] = ACTIONS(1),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_RPAREN] = ACTIONS(1),
    [sym_op_identifier] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_PLUS_DOT] = ACTIONS(1),
    [anon_sym_DASH_DOT] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [aux_sym_infix_op_token1] = ACTIONS(1),
    [aux_sym_infix_op_token2] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(1),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1),
    [sym_int] = ACTIONS(1),
    [sym_xint] = ACTIONS(1),
    [anon_sym_y] = ACTIONS(1),
    [anon_sym_uy] = ACTIONS(1),
    [anon_sym_s] = ACTIONS(1),
    [anon_sym_us] = ACTIONS(1),
    [anon_sym_l] = ACTIONS(1),
    [aux_sym_uint32_token1] = ACTIONS(1),
    [anon_sym_n] = ACTIONS(1),
    [anon_sym_un] = ACTIONS(1),
    [anon_sym_L] = ACTIONS(1),
    [aux_sym_uint64_token1] = ACTIONS(1),
    [anon_sym_f] = ACTIONS(1),
    [anon_sym_lf] = ACTIONS(1),
    [anon_sym_LF] = ACTIONS(1),
    [aux_sym_bignum_token1] = ACTIONS(1),
    [aux_sym_decimal_token1] = ACTIONS(1),
    [anon_sym_DOT2] = ACTIONS(1),
    [aux_sym_float_token1] = ACTIONS(1),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [anon_sym_STAR_RPAREN] = ACTIONS(1),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_POUNDif] = ACTIONS(1),
    [sym__newline] = ACTIONS(1),
    [sym__indent] = ACTIONS(1),
    [sym__dedent] = ACTIONS(1),
    [sym__struct_begin] = ACTIONS(1),
    [sym__interface_begin] = ACTIONS(1),
    [sym__triple_quoted_content] = ACTIONS(1),
    [sym__format_triple_quoted_content] = ACTIONS(1),
    [sym_block_comment_content] = ACTIONS(1),
    [sym__inside_string_marker] = ACTIONS(1),
    [sym__newline_not_aligned] = ACTIONS(1),
    [sym__tuple_marker] = ACTIONS(1),
    [sym__quoted_close] = ACTIONS(1),
    [sym__untyped_quoted_close] = ACTIONS(1),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1),
    [sym__multi_dollar_triple_quoted_content] = ACTIONS(1),
    [sym__multi_dollar_interp_start] = ACTIONS(1),
    [sym__multi_dollar_interp_end] = ACTIONS(1),
    [sym__multi_dollar_triple_quote_end] = ACTIONS(1),
    [sym__tyapp_open] = ACTIONS(1),
    [sym__paren_indent] = ACTIONS(1),
    [sym__type_decl_newline] = ACTIONS(1),
    [sym__in] = ACTIONS(1),
    [sym__error_sentinel] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_file] = STATE(20626),
    [sym_namespace] = STATE(18154),
    [sym_named_module] = STATE(20960),
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1212),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1),
    [sym_compiler_directive_decl] = STATE(1),
    [sym_fsi_directive_decl] = STATE(1),
    [sym_preproc_line] = STATE(1),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat1] = STATE(16694),
    [aux_sym_file_repeat2] = STATE(1057),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(19),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(23),
    [anon_sym_module] = ACTIONS(25),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(2)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(2),
    [sym_compiler_directive_decl] = STATE(2),
    [sym_fsi_directive_decl] = STATE(2),
    [sym_preproc_line] = STATE(2),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(129),
    [anon_sym_module] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(129),
    [anon_sym_open] = ACTIONS(129),
    [anon_sym_type] = ACTIONS(129),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(129),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(133),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_LT_AT] = ACTIONS(129),
    [anon_sym_LT_AT_AT] = ACTIONS(129),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(133),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(145),
    [anon_sym_return] = ACTIONS(129),
    [anon_sym_return_BANG] = ACTIONS(133),
    [anon_sym_yield] = ACTIONS(129),
    [anon_sym_yield_BANG] = ACTIONS(133),
    [anon_sym_lazy] = ACTIONS(129),
    [anon_sym_assert] = ACTIONS(129),
    [anon_sym_upcast] = ACTIONS(129),
    [anon_sym_downcast] = ACTIONS(129),
    [anon_sym_fixed] = ACTIONS(129),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(129),
    [anon_sym_while] = ACTIONS(129),
    [anon_sym_while_BANG] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(129),
    [anon_sym_fun] = ACTIONS(129),
    [anon_sym_try] = ACTIONS(129),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_match_BANG] = ACTIONS(133),
    [anon_sym_function] = ACTIONS(129),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(129),
    [anon_sym_use_BANG] = ACTIONS(133),
    [anon_sym_do_BANG] = ACTIONS(133),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(129),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_exception] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(129),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_AT_DQUOTE] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [sym_bool] = ACTIONS(129),
    [sym_unit] = ACTIONS(133),
    [anon_sym_LPAREN_PIPE] = ACTIONS(129),
    [sym_op_identifier] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_PLUS_DOT] = ACTIONS(129),
    [anon_sym_DASH_DOT] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(133),
    [aux_sym_prefix_op_token1] = ACTIONS(129),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(129),
    [sym_xint] = ACTIONS(133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(133),
    [anon_sym_POUNDload] = ACTIONS(133),
    [anon_sym_POUNDtime] = ACTIONS(133),
    [anon_sym_POUNDI] = ACTIONS(133),
    [anon_sym_POUNDhelp] = ACTIONS(133),
    [anon_sym_POUNDquit] = ACTIONS(133),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [anon_sym_POUNDelse] = ACTIONS(133),
    [sym__newline] = ACTIONS(153),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(133),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(3)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(3),
    [sym_compiler_directive_decl] = STATE(3),
    [sym_fsi_directive_decl] = STATE(3),
    [sym_preproc_line] = STATE(3),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(157),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_namespace] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(4)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(4),
    [sym_compiler_directive_decl] = STATE(4),
    [sym_fsi_directive_decl] = STATE(4),
    [sym_preproc_line] = STATE(4),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_module] = ACTIONS(171),
    [anon_sym_null] = ACTIONS(171),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(171),
    [anon_sym_open] = ACTIONS(171),
    [anon_sym_type] = ACTIONS(171),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(171),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_LBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(173),
    [anon_sym_LBRACE] = ACTIONS(171),
    [anon_sym_LT_AT] = ACTIONS(171),
    [anon_sym_LT_AT_AT] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(173),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(175),
    [anon_sym_return] = ACTIONS(171),
    [anon_sym_return_BANG] = ACTIONS(173),
    [anon_sym_yield] = ACTIONS(171),
    [anon_sym_yield_BANG] = ACTIONS(173),
    [anon_sym_lazy] = ACTIONS(171),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_upcast] = ACTIONS(171),
    [anon_sym_downcast] = ACTIONS(171),
    [anon_sym_fixed] = ACTIONS(171),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(171),
    [anon_sym_while] = ACTIONS(171),
    [anon_sym_while_BANG] = ACTIONS(173),
    [anon_sym_if] = ACTIONS(171),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_try] = ACTIONS(171),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_match_BANG] = ACTIONS(173),
    [anon_sym_function] = ACTIONS(171),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(171),
    [anon_sym_use_BANG] = ACTIONS(173),
    [anon_sym_do_BANG] = ACTIONS(173),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(171),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_exception] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_extern] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(171),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [anon_sym_AT_DQUOTE] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [sym_bool] = ACTIONS(171),
    [sym_unit] = ACTIONS(173),
    [anon_sym_LPAREN_PIPE] = ACTIONS(171),
    [sym_op_identifier] = ACTIONS(171),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_PLUS_DOT] = ACTIONS(171),
    [anon_sym_DASH_DOT] = ACTIONS(171),
    [anon_sym_PERCENT] = ACTIONS(171),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(171),
    [anon_sym_TILDE] = ACTIONS(173),
    [aux_sym_prefix_op_token1] = ACTIONS(171),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(171),
    [sym_xint] = ACTIONS(173),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(173),
    [anon_sym_POUNDload] = ACTIONS(173),
    [anon_sym_POUNDtime] = ACTIONS(173),
    [anon_sym_POUNDI] = ACTIONS(173),
    [anon_sym_POUNDhelp] = ACTIONS(173),
    [anon_sym_POUNDquit] = ACTIONS(173),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [anon_sym_POUNDelse] = ACTIONS(173),
    [sym__newline] = ACTIONS(153),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(173),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(5)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(5),
    [sym_compiler_directive_decl] = STATE(5),
    [sym_fsi_directive_decl] = STATE(5),
    [sym_preproc_line] = STATE(5),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(129),
    [anon_sym_module] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(129),
    [anon_sym_open] = ACTIONS(129),
    [anon_sym_type] = ACTIONS(129),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(129),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(133),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_LT_AT] = ACTIONS(129),
    [anon_sym_LT_AT_AT] = ACTIONS(129),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(133),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(177),
    [anon_sym_return] = ACTIONS(129),
    [anon_sym_return_BANG] = ACTIONS(133),
    [anon_sym_yield] = ACTIONS(129),
    [anon_sym_yield_BANG] = ACTIONS(133),
    [anon_sym_lazy] = ACTIONS(129),
    [anon_sym_assert] = ACTIONS(129),
    [anon_sym_upcast] = ACTIONS(129),
    [anon_sym_downcast] = ACTIONS(129),
    [anon_sym_fixed] = ACTIONS(129),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(129),
    [anon_sym_while] = ACTIONS(129),
    [anon_sym_while_BANG] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(129),
    [anon_sym_fun] = ACTIONS(129),
    [anon_sym_try] = ACTIONS(129),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_match_BANG] = ACTIONS(133),
    [anon_sym_function] = ACTIONS(129),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(129),
    [anon_sym_use_BANG] = ACTIONS(133),
    [anon_sym_do_BANG] = ACTIONS(133),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(129),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_exception] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(129),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_AT_DQUOTE] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [sym_bool] = ACTIONS(129),
    [sym_unit] = ACTIONS(133),
    [anon_sym_LPAREN_PIPE] = ACTIONS(129),
    [sym_op_identifier] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_PLUS_DOT] = ACTIONS(129),
    [anon_sym_DASH_DOT] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(133),
    [aux_sym_prefix_op_token1] = ACTIONS(129),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(129),
    [sym_xint] = ACTIONS(133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(133),
    [anon_sym_POUNDload] = ACTIONS(133),
    [anon_sym_POUNDtime] = ACTIONS(133),
    [anon_sym_POUNDI] = ACTIONS(133),
    [anon_sym_POUNDhelp] = ACTIONS(133),
    [anon_sym_POUNDquit] = ACTIONS(133),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [anon_sym_POUNDelse] = ACTIONS(133),
    [sym__newline] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(133),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(6)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(6),
    [sym_compiler_directive_decl] = STATE(6),
    [sym_fsi_directive_decl] = STATE(6),
    [sym_preproc_line] = STATE(6),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_module] = ACTIONS(171),
    [anon_sym_null] = ACTIONS(171),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(171),
    [anon_sym_open] = ACTIONS(171),
    [anon_sym_type] = ACTIONS(171),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(171),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_LBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(173),
    [anon_sym_LBRACE] = ACTIONS(171),
    [anon_sym_LT_AT] = ACTIONS(171),
    [anon_sym_LT_AT_AT] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(173),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(179),
    [anon_sym_return] = ACTIONS(171),
    [anon_sym_return_BANG] = ACTIONS(173),
    [anon_sym_yield] = ACTIONS(171),
    [anon_sym_yield_BANG] = ACTIONS(173),
    [anon_sym_lazy] = ACTIONS(171),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_upcast] = ACTIONS(171),
    [anon_sym_downcast] = ACTIONS(171),
    [anon_sym_fixed] = ACTIONS(171),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(171),
    [anon_sym_while] = ACTIONS(171),
    [anon_sym_while_BANG] = ACTIONS(173),
    [anon_sym_if] = ACTIONS(171),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_try] = ACTIONS(171),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_match_BANG] = ACTIONS(173),
    [anon_sym_function] = ACTIONS(171),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(171),
    [anon_sym_use_BANG] = ACTIONS(173),
    [anon_sym_do_BANG] = ACTIONS(173),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(171),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_exception] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_extern] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(171),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [anon_sym_AT_DQUOTE] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [sym_bool] = ACTIONS(171),
    [sym_unit] = ACTIONS(173),
    [anon_sym_LPAREN_PIPE] = ACTIONS(171),
    [sym_op_identifier] = ACTIONS(171),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_PLUS_DOT] = ACTIONS(171),
    [anon_sym_DASH_DOT] = ACTIONS(171),
    [anon_sym_PERCENT] = ACTIONS(171),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(171),
    [anon_sym_TILDE] = ACTIONS(173),
    [aux_sym_prefix_op_token1] = ACTIONS(171),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(171),
    [sym_xint] = ACTIONS(173),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(173),
    [anon_sym_POUNDload] = ACTIONS(173),
    [anon_sym_POUNDtime] = ACTIONS(173),
    [anon_sym_POUNDI] = ACTIONS(173),
    [anon_sym_POUNDhelp] = ACTIONS(173),
    [anon_sym_POUNDquit] = ACTIONS(173),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [anon_sym_POUNDelse] = ACTIONS(173),
    [sym__newline] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(173),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(7)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(7),
    [sym_compiler_directive_decl] = STATE(7),
    [sym_fsi_directive_decl] = STATE(7),
    [sym_preproc_line] = STATE(7),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [anon_sym_POUNDelse] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(8)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(8),
    [sym_compiler_directive_decl] = STATE(8),
    [sym_fsi_directive_decl] = STATE(8),
    [sym_preproc_line] = STATE(8),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(185),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_namespace] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(9)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(9),
    [sym_compiler_directive_decl] = STATE(9),
    [sym_fsi_directive_decl] = STATE(9),
    [sym_preproc_line] = STATE(9),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [anon_sym_POUNDelse] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(10)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(10),
    [sym_compiler_directive_decl] = STATE(10),
    [sym_fsi_directive_decl] = STATE(10),
    [sym_preproc_line] = STATE(10),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [anon_sym_POUNDelse] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(11)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(11),
    [sym_compiler_directive_decl] = STATE(11),
    [sym_fsi_directive_decl] = STATE(11),
    [sym_preproc_line] = STATE(11),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(249),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [anon_sym_POUNDelse] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(12)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(12),
    [sym_compiler_directive_decl] = STATE(12),
    [sym_fsi_directive_decl] = STATE(12),
    [sym_preproc_line] = STATE(12),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(284),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [anon_sym_POUNDelse] = ACTIONS(282),
    [sym__newline] = ACTIONS(153),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(13)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(13),
    [sym_compiler_directive_decl] = STATE(13),
    [sym_fsi_directive_decl] = STATE(13),
    [sym_preproc_line] = STATE(13),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(284),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [anon_sym_POUNDelse] = ACTIONS(288),
    [sym__newline] = ACTIONS(153),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(14)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(14),
    [sym_compiler_directive_decl] = STATE(14),
    [sym_fsi_directive_decl] = STATE(14),
    [sym_preproc_line] = STATE(14),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(15)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(15),
    [sym_compiler_directive_decl] = STATE(15),
    [sym_fsi_directive_decl] = STATE(15),
    [sym_preproc_line] = STATE(15),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(284),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [anon_sym_POUNDelse] = ACTIONS(296),
    [sym__newline] = ACTIONS(153),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(16)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(16),
    [sym_compiler_directive_decl] = STATE(16),
    [sym_fsi_directive_decl] = STATE(16),
    [sym_preproc_line] = STATE(16),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [anon_sym_POUNDelse] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(17)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(17),
    [sym_compiler_directive_decl] = STATE(17),
    [sym_fsi_directive_decl] = STATE(17),
    [sym_preproc_line] = STATE(17),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(284),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(18)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(18),
    [sym_compiler_directive_decl] = STATE(18),
    [sym_fsi_directive_decl] = STATE(18),
    [sym_preproc_line] = STATE(18),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(284),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(19)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(19),
    [sym_compiler_directive_decl] = STATE(19),
    [sym_fsi_directive_decl] = STATE(19),
    [sym_preproc_line] = STATE(19),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [anon_sym_POUNDelse] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(20)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(9),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1794),
    [sym_infix_op] = STATE(1780),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(20),
    [sym_compiler_directive_decl] = STATE(20),
    [sym_fsi_directive_decl] = STATE(20),
    [sym_preproc_line] = STATE(20),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_sequential_expression_repeat1] = STATE(2419),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_COMMA] = ACTIONS(137),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(135),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(213),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(217),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(141),
    [anon_sym_DOT] = ACTIONS(143),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_COLON_GT] = ACTIONS(147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(147),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_DOT_LBRACK] = ACTIONS(149),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_LPAREN2] = ACTIONS(151),
    [anon_sym_PIPE_PIPE] = ACTIONS(141),
    [anon_sym_AMP_AMP] = ACTIONS(284),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(189),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [anon_sym_POUNDelse] = ACTIONS(312),
    [sym__newline] = ACTIONS(153),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
    [sym__tyapp_open] = ACTIONS(155),
  },
  [STATE(21)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(21),
    [sym_compiler_directive_decl] = STATE(21),
    [sym_fsi_directive_decl] = STATE(21),
    [sym_preproc_line] = STATE(21),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(199),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(193),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(368),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(22)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(22),
    [sym_compiler_directive_decl] = STATE(22),
    [sym_fsi_directive_decl] = STATE(22),
    [sym_preproc_line] = STATE(22),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(282),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(280),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(397),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(397),
    [anon_sym_AMP_AMP] = ACTIONS(399),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(23)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(23),
    [sym_compiler_directive_decl] = STATE(23),
    [sym_fsi_directive_decl] = STATE(23),
    [sym_preproc_line] = STATE(23),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(288),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(286),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(397),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(397),
    [anon_sym_AMP_AMP] = ACTIONS(399),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(24)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(24),
    [sym_compiler_directive_decl] = STATE(24),
    [sym_fsi_directive_decl] = STATE(24),
    [sym_preproc_line] = STATE(24),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(292),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(25)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(25),
    [sym_compiler_directive_decl] = STATE(25),
    [sym_fsi_directive_decl] = STATE(25),
    [sym_preproc_line] = STATE(25),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(296),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(294),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(397),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(397),
    [anon_sym_AMP_AMP] = ACTIONS(399),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(26)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(26),
    [sym_compiler_directive_decl] = STATE(26),
    [sym_fsi_directive_decl] = STATE(26),
    [sym_preproc_line] = STATE(26),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(300),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(298),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(27)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(27),
    [sym_compiler_directive_decl] = STATE(27),
    [sym_fsi_directive_decl] = STATE(27),
    [sym_preproc_line] = STATE(27),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(302),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(399),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(28)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(28),
    [sym_compiler_directive_decl] = STATE(28),
    [sym_fsi_directive_decl] = STATE(28),
    [sym_preproc_line] = STATE(28),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(302),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(399),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(29)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(29),
    [sym_compiler_directive_decl] = STATE(29),
    [sym_fsi_directive_decl] = STATE(29),
    [sym_preproc_line] = STATE(29),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(308),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(306),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(30)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(30),
    [sym_compiler_directive_decl] = STATE(30),
    [sym_fsi_directive_decl] = STATE(30),
    [sym_preproc_line] = STATE(30),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(312),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_namespace] = ACTIONS(310),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(397),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(397),
    [anon_sym_AMP_AMP] = ACTIONS(399),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(31)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(31),
    [sym_compiler_directive_decl] = STATE(31),
    [sym_fsi_directive_decl] = STATE(31),
    [sym_preproc_line] = STATE(31),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(133),
    [sym_identifier] = ACTIONS(129),
    [anon_sym_namespace] = ACTIONS(129),
    [anon_sym_module] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(129),
    [anon_sym_open] = ACTIONS(129),
    [anon_sym_type] = ACTIONS(129),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(129),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(133),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_LT_AT] = ACTIONS(129),
    [anon_sym_LT_AT_AT] = ACTIONS(129),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym_or] = ACTIONS(397),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(133),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(403),
    [anon_sym_return] = ACTIONS(129),
    [anon_sym_return_BANG] = ACTIONS(133),
    [anon_sym_yield] = ACTIONS(129),
    [anon_sym_yield_BANG] = ACTIONS(133),
    [anon_sym_lazy] = ACTIONS(129),
    [anon_sym_assert] = ACTIONS(129),
    [anon_sym_upcast] = ACTIONS(129),
    [anon_sym_downcast] = ACTIONS(129),
    [anon_sym_fixed] = ACTIONS(129),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(129),
    [anon_sym_while] = ACTIONS(129),
    [anon_sym_while_BANG] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(129),
    [anon_sym_fun] = ACTIONS(129),
    [anon_sym_try] = ACTIONS(129),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_match_BANG] = ACTIONS(133),
    [anon_sym_function] = ACTIONS(129),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(129),
    [anon_sym_use_BANG] = ACTIONS(133),
    [anon_sym_do_BANG] = ACTIONS(133),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(397),
    [anon_sym_AMP_AMP] = ACTIONS(129),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_exception] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(129),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_AT_DQUOTE] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [sym_bool] = ACTIONS(129),
    [sym_unit] = ACTIONS(133),
    [anon_sym_LPAREN_PIPE] = ACTIONS(129),
    [sym_op_identifier] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_PLUS_DOT] = ACTIONS(129),
    [anon_sym_DASH_DOT] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(133),
    [aux_sym_prefix_op_token1] = ACTIONS(129),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(129),
    [sym_xint] = ACTIONS(133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(133),
    [anon_sym_POUNDload] = ACTIONS(133),
    [anon_sym_POUNDtime] = ACTIONS(133),
    [anon_sym_POUNDI] = ACTIONS(133),
    [anon_sym_POUNDhelp] = ACTIONS(133),
    [anon_sym_POUNDquit] = ACTIONS(133),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [sym__newline] = ACTIONS(401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(133),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(32)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(32),
    [sym_compiler_directive_decl] = STATE(32),
    [sym_fsi_directive_decl] = STATE(32),
    [sym_preproc_line] = STATE(32),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(173),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_namespace] = ACTIONS(171),
    [anon_sym_module] = ACTIONS(171),
    [anon_sym_null] = ACTIONS(171),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(171),
    [anon_sym_open] = ACTIONS(171),
    [anon_sym_type] = ACTIONS(171),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(171),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_LBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(173),
    [anon_sym_LBRACE] = ACTIONS(171),
    [anon_sym_LT_AT] = ACTIONS(171),
    [anon_sym_LT_AT_AT] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_or] = ACTIONS(397),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(173),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(405),
    [anon_sym_return] = ACTIONS(171),
    [anon_sym_return_BANG] = ACTIONS(173),
    [anon_sym_yield] = ACTIONS(171),
    [anon_sym_yield_BANG] = ACTIONS(173),
    [anon_sym_lazy] = ACTIONS(171),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_upcast] = ACTIONS(171),
    [anon_sym_downcast] = ACTIONS(171),
    [anon_sym_fixed] = ACTIONS(171),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(171),
    [anon_sym_while] = ACTIONS(171),
    [anon_sym_while_BANG] = ACTIONS(173),
    [anon_sym_if] = ACTIONS(171),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_try] = ACTIONS(171),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_match_BANG] = ACTIONS(173),
    [anon_sym_function] = ACTIONS(171),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(171),
    [anon_sym_use_BANG] = ACTIONS(173),
    [anon_sym_do_BANG] = ACTIONS(173),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(397),
    [anon_sym_AMP_AMP] = ACTIONS(171),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_exception] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_extern] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(171),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [anon_sym_AT_DQUOTE] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [sym_bool] = ACTIONS(171),
    [sym_unit] = ACTIONS(173),
    [anon_sym_LPAREN_PIPE] = ACTIONS(171),
    [sym_op_identifier] = ACTIONS(171),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_PLUS_DOT] = ACTIONS(171),
    [anon_sym_DASH_DOT] = ACTIONS(171),
    [anon_sym_PERCENT] = ACTIONS(171),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(171),
    [anon_sym_TILDE] = ACTIONS(173),
    [aux_sym_prefix_op_token1] = ACTIONS(171),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(171),
    [sym_xint] = ACTIONS(173),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(173),
    [anon_sym_POUNDload] = ACTIONS(173),
    [anon_sym_POUNDtime] = ACTIONS(173),
    [anon_sym_POUNDI] = ACTIONS(173),
    [anon_sym_POUNDhelp] = ACTIONS(173),
    [anon_sym_POUNDquit] = ACTIONS(173),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [sym__newline] = ACTIONS(401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(173),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(33)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(3),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1749),
    [sym_infix_op] = STATE(1730),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(33),
    [sym_compiler_directive_decl] = STATE(33),
    [sym_fsi_directive_decl] = STATE(33),
    [sym_preproc_line] = STATE(33),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2405),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(183),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_namespace] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(34)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(34),
    [sym_compiler_directive_decl] = STATE(34),
    [sym_fsi_directive_decl] = STATE(34),
    [sym_preproc_line] = STATE(34),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(173),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_module] = ACTIONS(171),
    [anon_sym_null] = ACTIONS(171),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(171),
    [anon_sym_open] = ACTIONS(171),
    [anon_sym_type] = ACTIONS(171),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(171),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_LBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(173),
    [anon_sym_LBRACE] = ACTIONS(171),
    [anon_sym_LT_AT] = ACTIONS(171),
    [anon_sym_LT_AT_AT] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_or] = ACTIONS(409),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(173),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(405),
    [anon_sym_return] = ACTIONS(171),
    [anon_sym_return_BANG] = ACTIONS(173),
    [anon_sym_yield] = ACTIONS(171),
    [anon_sym_yield_BANG] = ACTIONS(173),
    [anon_sym_lazy] = ACTIONS(171),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_upcast] = ACTIONS(171),
    [anon_sym_downcast] = ACTIONS(171),
    [anon_sym_fixed] = ACTIONS(171),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(171),
    [anon_sym_while] = ACTIONS(171),
    [anon_sym_while_BANG] = ACTIONS(173),
    [anon_sym_if] = ACTIONS(171),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_try] = ACTIONS(171),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_match_BANG] = ACTIONS(173),
    [anon_sym_function] = ACTIONS(171),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(171),
    [anon_sym_use_BANG] = ACTIONS(173),
    [anon_sym_do_BANG] = ACTIONS(173),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(409),
    [anon_sym_AMP_AMP] = ACTIONS(171),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_exception] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_extern] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(171),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [anon_sym_AT_DQUOTE] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [sym_bool] = ACTIONS(171),
    [sym_unit] = ACTIONS(173),
    [anon_sym_LPAREN_PIPE] = ACTIONS(171),
    [sym_op_identifier] = ACTIONS(171),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_PLUS_DOT] = ACTIONS(171),
    [anon_sym_DASH_DOT] = ACTIONS(171),
    [anon_sym_PERCENT] = ACTIONS(171),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(171),
    [anon_sym_TILDE] = ACTIONS(173),
    [aux_sym_prefix_op_token1] = ACTIONS(171),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(171),
    [sym_xint] = ACTIONS(173),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(173),
    [anon_sym_POUNDload] = ACTIONS(173),
    [anon_sym_POUNDtime] = ACTIONS(173),
    [anon_sym_POUNDI] = ACTIONS(173),
    [anon_sym_POUNDhelp] = ACTIONS(173),
    [anon_sym_POUNDquit] = ACTIONS(173),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [sym__newline] = ACTIONS(411),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(173),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(35)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(35),
    [sym_compiler_directive_decl] = STATE(35),
    [sym_fsi_directive_decl] = STATE(35),
    [sym_preproc_line] = STATE(35),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(312),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(409),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(409),
    [anon_sym_AMP_AMP] = ACTIONS(421),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(411),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(36)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(36),
    [sym_compiler_directive_decl] = STATE(36),
    [sym_fsi_directive_decl] = STATE(36),
    [sym_preproc_line] = STATE(36),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(133),
    [sym_identifier] = ACTIONS(129),
    [anon_sym_module] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(129),
    [anon_sym_open] = ACTIONS(129),
    [anon_sym_type] = ACTIONS(129),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(129),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(133),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_LT_AT] = ACTIONS(129),
    [anon_sym_LT_AT_AT] = ACTIONS(129),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym_or] = ACTIONS(409),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(133),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(403),
    [anon_sym_return] = ACTIONS(129),
    [anon_sym_return_BANG] = ACTIONS(133),
    [anon_sym_yield] = ACTIONS(129),
    [anon_sym_yield_BANG] = ACTIONS(133),
    [anon_sym_lazy] = ACTIONS(129),
    [anon_sym_assert] = ACTIONS(129),
    [anon_sym_upcast] = ACTIONS(129),
    [anon_sym_downcast] = ACTIONS(129),
    [anon_sym_fixed] = ACTIONS(129),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(129),
    [anon_sym_while] = ACTIONS(129),
    [anon_sym_while_BANG] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(129),
    [anon_sym_fun] = ACTIONS(129),
    [anon_sym_try] = ACTIONS(129),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_match_BANG] = ACTIONS(133),
    [anon_sym_function] = ACTIONS(129),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(129),
    [anon_sym_use_BANG] = ACTIONS(133),
    [anon_sym_do_BANG] = ACTIONS(133),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(409),
    [anon_sym_AMP_AMP] = ACTIONS(129),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_exception] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(129),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_AT_DQUOTE] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [sym_bool] = ACTIONS(129),
    [sym_unit] = ACTIONS(133),
    [anon_sym_LPAREN_PIPE] = ACTIONS(129),
    [sym_op_identifier] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_PLUS_DOT] = ACTIONS(129),
    [anon_sym_DASH_DOT] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(133),
    [aux_sym_prefix_op_token1] = ACTIONS(129),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(129),
    [sym_xint] = ACTIONS(133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(133),
    [anon_sym_POUNDload] = ACTIONS(133),
    [anon_sym_POUNDtime] = ACTIONS(133),
    [anon_sym_POUNDI] = ACTIONS(133),
    [anon_sym_POUNDhelp] = ACTIONS(133),
    [anon_sym_POUNDquit] = ACTIONS(133),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [sym__newline] = ACTIONS(411),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(133),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(37)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(37),
    [sym_compiler_directive_decl] = STATE(37),
    [sym_fsi_directive_decl] = STATE(37),
    [sym_preproc_line] = STATE(37),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(129),
    [anon_sym_module] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(129),
    [anon_sym_open] = ACTIONS(129),
    [anon_sym_type] = ACTIONS(129),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(129),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(133),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_LT_AT] = ACTIONS(129),
    [anon_sym_LT_AT_AT] = ACTIONS(129),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(133),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(435),
    [anon_sym_return] = ACTIONS(129),
    [anon_sym_return_BANG] = ACTIONS(133),
    [anon_sym_yield] = ACTIONS(129),
    [anon_sym_yield_BANG] = ACTIONS(133),
    [anon_sym_lazy] = ACTIONS(129),
    [anon_sym_assert] = ACTIONS(129),
    [anon_sym_upcast] = ACTIONS(129),
    [anon_sym_downcast] = ACTIONS(129),
    [anon_sym_fixed] = ACTIONS(129),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(129),
    [anon_sym_while] = ACTIONS(129),
    [anon_sym_while_BANG] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(129),
    [anon_sym_fun] = ACTIONS(129),
    [anon_sym_try] = ACTIONS(129),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_match_BANG] = ACTIONS(133),
    [anon_sym_function] = ACTIONS(129),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(129),
    [anon_sym_use_BANG] = ACTIONS(133),
    [anon_sym_do_BANG] = ACTIONS(133),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(129),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_exception] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(129),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_AT_DQUOTE] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [sym_bool] = ACTIONS(129),
    [sym_unit] = ACTIONS(133),
    [anon_sym_LPAREN_PIPE] = ACTIONS(129),
    [sym_op_identifier] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_PLUS_DOT] = ACTIONS(129),
    [anon_sym_DASH_DOT] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(133),
    [aux_sym_prefix_op_token1] = ACTIONS(129),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(129),
    [sym_xint] = ACTIONS(133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(133),
    [anon_sym_POUNDload] = ACTIONS(133),
    [anon_sym_POUNDtime] = ACTIONS(133),
    [anon_sym_POUNDI] = ACTIONS(133),
    [anon_sym_POUNDhelp] = ACTIONS(133),
    [anon_sym_POUNDquit] = ACTIONS(133),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [sym__newline] = ACTIONS(443),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(133),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(38)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(38),
    [sym_compiler_directive_decl] = STATE(38),
    [sym_fsi_directive_decl] = STATE(38),
    [sym_preproc_line] = STATE(38),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(39)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(39),
    [sym_compiler_directive_decl] = STATE(39),
    [sym_fsi_directive_decl] = STATE(39),
    [sym_preproc_line] = STATE(39),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_module] = ACTIONS(171),
    [anon_sym_null] = ACTIONS(171),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(171),
    [anon_sym_open] = ACTIONS(171),
    [anon_sym_type] = ACTIONS(171),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(171),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_LBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(173),
    [anon_sym_LBRACE] = ACTIONS(171),
    [anon_sym_LT_AT] = ACTIONS(171),
    [anon_sym_LT_AT_AT] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(173),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(447),
    [anon_sym_return] = ACTIONS(171),
    [anon_sym_return_BANG] = ACTIONS(173),
    [anon_sym_yield] = ACTIONS(171),
    [anon_sym_yield_BANG] = ACTIONS(173),
    [anon_sym_lazy] = ACTIONS(171),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_upcast] = ACTIONS(171),
    [anon_sym_downcast] = ACTIONS(171),
    [anon_sym_fixed] = ACTIONS(171),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(171),
    [anon_sym_while] = ACTIONS(171),
    [anon_sym_while_BANG] = ACTIONS(173),
    [anon_sym_if] = ACTIONS(171),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_try] = ACTIONS(171),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_match_BANG] = ACTIONS(173),
    [anon_sym_function] = ACTIONS(171),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(171),
    [anon_sym_use_BANG] = ACTIONS(173),
    [anon_sym_do_BANG] = ACTIONS(173),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(171),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_exception] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_extern] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(171),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [anon_sym_AT_DQUOTE] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [sym_bool] = ACTIONS(171),
    [sym_unit] = ACTIONS(173),
    [anon_sym_LPAREN_PIPE] = ACTIONS(171),
    [sym_op_identifier] = ACTIONS(171),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_PLUS_DOT] = ACTIONS(171),
    [anon_sym_DASH_DOT] = ACTIONS(171),
    [anon_sym_PERCENT] = ACTIONS(171),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(171),
    [anon_sym_TILDE] = ACTIONS(173),
    [aux_sym_prefix_op_token1] = ACTIONS(171),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(171),
    [sym_xint] = ACTIONS(173),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(173),
    [anon_sym_POUNDload] = ACTIONS(173),
    [anon_sym_POUNDtime] = ACTIONS(173),
    [anon_sym_POUNDI] = ACTIONS(173),
    [anon_sym_POUNDhelp] = ACTIONS(173),
    [anon_sym_POUNDquit] = ACTIONS(173),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [sym__newline] = ACTIONS(443),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(173),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(40)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(40),
    [sym_compiler_directive_decl] = STATE(40),
    [sym_fsi_directive_decl] = STATE(40),
    [sym_preproc_line] = STATE(40),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(129),
    [anon_sym_module] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(129),
    [anon_sym_open] = ACTIONS(129),
    [anon_sym_type] = ACTIONS(129),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(129),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(133),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_LT_AT] = ACTIONS(129),
    [anon_sym_LT_AT_AT] = ACTIONS(129),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(133),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(449),
    [anon_sym_return] = ACTIONS(129),
    [anon_sym_return_BANG] = ACTIONS(133),
    [anon_sym_yield] = ACTIONS(129),
    [anon_sym_yield_BANG] = ACTIONS(133),
    [anon_sym_lazy] = ACTIONS(129),
    [anon_sym_assert] = ACTIONS(129),
    [anon_sym_upcast] = ACTIONS(129),
    [anon_sym_downcast] = ACTIONS(129),
    [anon_sym_fixed] = ACTIONS(129),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(129),
    [anon_sym_while] = ACTIONS(129),
    [anon_sym_while_BANG] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(129),
    [anon_sym_fun] = ACTIONS(129),
    [anon_sym_try] = ACTIONS(129),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_match_BANG] = ACTIONS(133),
    [anon_sym_function] = ACTIONS(129),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(129),
    [anon_sym_use_BANG] = ACTIONS(133),
    [anon_sym_do_BANG] = ACTIONS(133),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(129),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_exception] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(129),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_AT_DQUOTE] = ACTIONS(133),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(133),
    [sym_bool] = ACTIONS(129),
    [sym_unit] = ACTIONS(133),
    [anon_sym_LPAREN_PIPE] = ACTIONS(129),
    [sym_op_identifier] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_PLUS_DOT] = ACTIONS(129),
    [anon_sym_DASH_DOT] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(133),
    [aux_sym_prefix_op_token1] = ACTIONS(129),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(129),
    [sym_xint] = ACTIONS(133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(133),
    [anon_sym_POUNDload] = ACTIONS(133),
    [anon_sym_POUNDtime] = ACTIONS(133),
    [anon_sym_POUNDI] = ACTIONS(133),
    [anon_sym_POUNDhelp] = ACTIONS(133),
    [anon_sym_POUNDquit] = ACTIONS(133),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [sym__newline] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(133),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(41)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(41),
    [sym_compiler_directive_decl] = STATE(41),
    [sym_fsi_directive_decl] = STATE(41),
    [sym_preproc_line] = STATE(41),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_module] = ACTIONS(171),
    [anon_sym_null] = ACTIONS(171),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(171),
    [anon_sym_open] = ACTIONS(171),
    [anon_sym_type] = ACTIONS(171),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(171),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_LBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(173),
    [anon_sym_LBRACE] = ACTIONS(171),
    [anon_sym_LT_AT] = ACTIONS(171),
    [anon_sym_LT_AT_AT] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(173),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(451),
    [anon_sym_return] = ACTIONS(171),
    [anon_sym_return_BANG] = ACTIONS(173),
    [anon_sym_yield] = ACTIONS(171),
    [anon_sym_yield_BANG] = ACTIONS(173),
    [anon_sym_lazy] = ACTIONS(171),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_upcast] = ACTIONS(171),
    [anon_sym_downcast] = ACTIONS(171),
    [anon_sym_fixed] = ACTIONS(171),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(171),
    [anon_sym_while] = ACTIONS(171),
    [anon_sym_while_BANG] = ACTIONS(173),
    [anon_sym_if] = ACTIONS(171),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_try] = ACTIONS(171),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_match_BANG] = ACTIONS(173),
    [anon_sym_function] = ACTIONS(171),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(171),
    [anon_sym_use_BANG] = ACTIONS(173),
    [anon_sym_do_BANG] = ACTIONS(173),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(171),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_exception] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_extern] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(171),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [anon_sym_AT_DQUOTE] = ACTIONS(173),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(173),
    [sym_bool] = ACTIONS(171),
    [sym_unit] = ACTIONS(173),
    [anon_sym_LPAREN_PIPE] = ACTIONS(171),
    [sym_op_identifier] = ACTIONS(171),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_PLUS_DOT] = ACTIONS(171),
    [anon_sym_DASH_DOT] = ACTIONS(171),
    [anon_sym_PERCENT] = ACTIONS(171),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(171),
    [anon_sym_TILDE] = ACTIONS(173),
    [aux_sym_prefix_op_token1] = ACTIONS(171),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(171),
    [sym_xint] = ACTIONS(173),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(173),
    [anon_sym_POUNDload] = ACTIONS(173),
    [anon_sym_POUNDtime] = ACTIONS(173),
    [anon_sym_POUNDI] = ACTIONS(173),
    [anon_sym_POUNDhelp] = ACTIONS(173),
    [anon_sym_POUNDquit] = ACTIONS(173),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [sym__newline] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(173),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(42)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(42),
    [sym_compiler_directive_decl] = STATE(42),
    [sym_fsi_directive_decl] = STATE(42),
    [sym_preproc_line] = STATE(42),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(183),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(43)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(43),
    [sym_compiler_directive_decl] = STATE(43),
    [sym_fsi_directive_decl] = STATE(43),
    [sym_preproc_line] = STATE(43),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(157),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(44)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(44),
    [sym_compiler_directive_decl] = STATE(44),
    [sym_fsi_directive_decl] = STATE(44),
    [sym_preproc_line] = STATE(44),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(185),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(45)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(45),
    [sym_compiler_directive_decl] = STATE(45),
    [sym_fsi_directive_decl] = STATE(45),
    [sym_preproc_line] = STATE(45),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(199),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(453),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(46)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(46),
    [sym_compiler_directive_decl] = STATE(46),
    [sym_fsi_directive_decl] = STATE(46),
    [sym_preproc_line] = STATE(46),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(282),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(409),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(409),
    [anon_sym_AMP_AMP] = ACTIONS(421),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(411),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(47)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(47),
    [sym_compiler_directive_decl] = STATE(47),
    [sym_fsi_directive_decl] = STATE(47),
    [sym_preproc_line] = STATE(47),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(48)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(48),
    [sym_compiler_directive_decl] = STATE(48),
    [sym_fsi_directive_decl] = STATE(48),
    [sym_preproc_line] = STATE(48),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(49)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(49),
    [sym_compiler_directive_decl] = STATE(49),
    [sym_fsi_directive_decl] = STATE(49),
    [sym_preproc_line] = STATE(49),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(506),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(50)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(50),
    [sym_compiler_directive_decl] = STATE(50),
    [sym_fsi_directive_decl] = STATE(50),
    [sym_preproc_line] = STATE(50),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(535),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [sym__newline] = ACTIONS(443),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(51)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(51),
    [sym_compiler_directive_decl] = STATE(51),
    [sym_fsi_directive_decl] = STATE(51),
    [sym_preproc_line] = STATE(51),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(535),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [sym__newline] = ACTIONS(443),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(52)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(52),
    [sym_compiler_directive_decl] = STATE(52),
    [sym_fsi_directive_decl] = STATE(52),
    [sym_preproc_line] = STATE(52),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(53)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(53),
    [sym_compiler_directive_decl] = STATE(53),
    [sym_fsi_directive_decl] = STATE(53),
    [sym_preproc_line] = STATE(53),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(535),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [sym__newline] = ACTIONS(443),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(54)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(54),
    [sym_compiler_directive_decl] = STATE(54),
    [sym_fsi_directive_decl] = STATE(54),
    [sym_preproc_line] = STATE(54),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(55)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(55),
    [sym_compiler_directive_decl] = STATE(55),
    [sym_fsi_directive_decl] = STATE(55),
    [sym_preproc_line] = STATE(55),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(535),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(56)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(56),
    [sym_compiler_directive_decl] = STATE(56),
    [sym_fsi_directive_decl] = STATE(56),
    [sym_preproc_line] = STATE(56),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(535),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(57)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(57),
    [sym_compiler_directive_decl] = STATE(57),
    [sym_fsi_directive_decl] = STATE(57),
    [sym_preproc_line] = STATE(57),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(58)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(47),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1805),
    [sym_infix_op] = STATE(1797),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(58),
    [sym_compiler_directive_decl] = STATE(58),
    [sym_fsi_directive_decl] = STATE(58),
    [sym_preproc_line] = STATE(58),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_sequential_expression_repeat1] = STATE(2663),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_COMMA] = ACTIONS(429),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(427),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(476),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(431),
    [anon_sym_DOT] = ACTIONS(433),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_COLON_GT] = ACTIONS(437),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(437),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_DOT_LBRACK] = ACTIONS(439),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_LPAREN2] = ACTIONS(441),
    [anon_sym_PIPE_PIPE] = ACTIONS(431),
    [anon_sym_AMP_AMP] = ACTIONS(535),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(456),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(525),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [sym__newline] = ACTIONS(443),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
    [sym__tyapp_open] = ACTIONS(445),
  },
  [STATE(59)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(59),
    [sym_compiler_directive_decl] = STATE(59),
    [sym_fsi_directive_decl] = STATE(59),
    [sym_preproc_line] = STATE(59),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(288),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(409),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(409),
    [anon_sym_AMP_AMP] = ACTIONS(421),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(411),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(60)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(60),
    [sym_compiler_directive_decl] = STATE(60),
    [sym_fsi_directive_decl] = STATE(60),
    [sym_preproc_line] = STATE(60),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(292),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(61)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(61),
    [sym_compiler_directive_decl] = STATE(61),
    [sym_fsi_directive_decl] = STATE(61),
    [sym_preproc_line] = STATE(61),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(296),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(409),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(409),
    [anon_sym_AMP_AMP] = ACTIONS(421),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(411),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(62)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(62),
    [sym_compiler_directive_decl] = STATE(62),
    [sym_fsi_directive_decl] = STATE(62),
    [sym_preproc_line] = STATE(62),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(300),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(63)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(63),
    [sym_compiler_directive_decl] = STATE(63),
    [sym_fsi_directive_decl] = STATE(63),
    [sym_preproc_line] = STATE(63),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(421),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(64)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(64),
    [sym_compiler_directive_decl] = STATE(64),
    [sym_fsi_directive_decl] = STATE(64),
    [sym_preproc_line] = STATE(64),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(332),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(421),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(65)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(43),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1739),
    [sym_infix_op] = STATE(1720),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(65),
    [sym_compiler_directive_decl] = STATE(65),
    [sym_fsi_directive_decl] = STATE(65),
    [sym_preproc_line] = STATE(65),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_sequential_expression_repeat1] = STATE(2493),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(308),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_COMMA] = ACTIONS(407),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(336),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(161),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_COLON_GT] = ACTIONS(344),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(344),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_DOT_LBRACK] = ACTIONS(163),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_LPAREN2] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(167),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
    [sym__tyapp_open] = ACTIONS(169),
  },
  [STATE(66)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(66),
    [sym_compiler_directive_decl] = STATE(66),
    [sym_fsi_directive_decl] = STATE(66),
    [sym_preproc_line] = STATE(66),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [anon_sym_inherit] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [anon_sym_POUNDelse] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(67)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(67),
    [sym_compiler_directive_decl] = STATE(67),
    [sym_fsi_directive_decl] = STATE(67),
    [sym_preproc_line] = STATE(67),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_inherit] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(68)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(68),
    [sym_compiler_directive_decl] = STATE(68),
    [sym_fsi_directive_decl] = STATE(68),
    [sym_preproc_line] = STATE(68),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(69)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(69),
    [sym_compiler_directive_decl] = STATE(69),
    [sym_fsi_directive_decl] = STATE(69),
    [sym_preproc_line] = STATE(69),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(715),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [anon_sym_POUNDelse] = ACTIONS(296),
    [sym__newline] = ACTIONS(719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(70)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(70),
    [sym_compiler_directive_decl] = STATE(70),
    [sym_fsi_directive_decl] = STATE(70),
    [sym_preproc_line] = STATE(70),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [anon_sym_POUNDelse] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(71)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(71),
    [sym_compiler_directive_decl] = STATE(71),
    [sym_fsi_directive_decl] = STATE(71),
    [sym_preproc_line] = STATE(71),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(72)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(72),
    [sym_compiler_directive_decl] = STATE(72),
    [sym_fsi_directive_decl] = STATE(72),
    [sym_preproc_line] = STATE(72),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(73)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(73),
    [sym_compiler_directive_decl] = STATE(73),
    [sym_fsi_directive_decl] = STATE(73),
    [sym_preproc_line] = STATE(73),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [anon_sym_POUNDelse] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(74)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(74),
    [sym_compiler_directive_decl] = STATE(74),
    [sym_fsi_directive_decl] = STATE(74),
    [sym_preproc_line] = STATE(74),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(715),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [anon_sym_POUNDelse] = ACTIONS(312),
    [sym__newline] = ACTIONS(719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(75)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(75),
    [sym_compiler_directive_decl] = STATE(75),
    [sym_fsi_directive_decl] = STATE(75),
    [sym_preproc_line] = STATE(75),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(721),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(715),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [anon_sym_POUNDelse] = ACTIONS(133),
    [sym__newline] = ACTIONS(719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(76)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(76),
    [sym_compiler_directive_decl] = STATE(76),
    [sym_fsi_directive_decl] = STATE(76),
    [sym_preproc_line] = STATE(76),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(723),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(715),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [anon_sym_POUNDelse] = ACTIONS(173),
    [sym__newline] = ACTIONS(719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(77)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(77),
    [sym_compiler_directive_decl] = STATE(77),
    [sym_fsi_directive_decl] = STATE(77),
    [sym_preproc_line] = STATE(77),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(727),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [anon_sym_inherit] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [anon_sym_POUNDelse] = ACTIONS(296),
    [sym__newline] = ACTIONS(731),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(78)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(78),
    [sym_compiler_directive_decl] = STATE(78),
    [sym_fsi_directive_decl] = STATE(78),
    [sym_preproc_line] = STATE(78),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [anon_sym_inherit] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [anon_sym_POUNDelse] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(79)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(79),
    [sym_compiler_directive_decl] = STATE(79),
    [sym_fsi_directive_decl] = STATE(79),
    [sym_preproc_line] = STATE(79),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_inherit] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(80)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(80),
    [sym_compiler_directive_decl] = STATE(80),
    [sym_fsi_directive_decl] = STATE(80),
    [sym_preproc_line] = STATE(80),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_inherit] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(81)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(81),
    [sym_compiler_directive_decl] = STATE(81),
    [sym_fsi_directive_decl] = STATE(81),
    [sym_preproc_line] = STATE(81),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(129),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(727),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_inherit] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [anon_sym_POUNDelse] = ACTIONS(133),
    [sym__newline] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(82)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(82),
    [sym_compiler_directive_decl] = STATE(82),
    [sym_fsi_directive_decl] = STATE(82),
    [sym_preproc_line] = STATE(82),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [anon_sym_inherit] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [anon_sym_POUNDelse] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(83)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(83),
    [sym_compiler_directive_decl] = STATE(83),
    [sym_fsi_directive_decl] = STATE(83),
    [sym_preproc_line] = STATE(83),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [anon_sym_POUNDelse] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(84)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(84),
    [sym_compiler_directive_decl] = STATE(84),
    [sym_fsi_directive_decl] = STATE(84),
    [sym_preproc_line] = STATE(84),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [anon_sym_inherit] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [anon_sym_POUNDelse] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(85)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(85),
    [sym_compiler_directive_decl] = STATE(85),
    [sym_fsi_directive_decl] = STATE(85),
    [sym_preproc_line] = STATE(85),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(727),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [anon_sym_inherit] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [anon_sym_POUNDelse] = ACTIONS(312),
    [sym__newline] = ACTIONS(731),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(86)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(86),
    [sym_compiler_directive_decl] = STATE(86),
    [sym_fsi_directive_decl] = STATE(86),
    [sym_preproc_line] = STATE(86),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(171),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(727),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_inherit] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [anon_sym_POUNDelse] = ACTIONS(173),
    [sym__newline] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(87)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(87),
    [sym_compiler_directive_decl] = STATE(87),
    [sym_fsi_directive_decl] = STATE(87),
    [sym_preproc_line] = STATE(87),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(733),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [anon_sym_POUNDelse] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(88)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(88),
    [sym_compiler_directive_decl] = STATE(88),
    [sym_fsi_directive_decl] = STATE(88),
    [sym_preproc_line] = STATE(88),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(715),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [anon_sym_POUNDelse] = ACTIONS(282),
    [sym__newline] = ACTIONS(719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(89)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(89),
    [sym_compiler_directive_decl] = STATE(89),
    [sym_fsi_directive_decl] = STATE(89),
    [sym_preproc_line] = STATE(89),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [anon_sym_inherit] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [anon_sym_POUNDelse] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(90)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(90),
    [sym_compiler_directive_decl] = STATE(90),
    [sym_fsi_directive_decl] = STATE(90),
    [sym_preproc_line] = STATE(90),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [anon_sym_POUNDelse] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(91)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(91),
    [sym_compiler_directive_decl] = STATE(91),
    [sym_fsi_directive_decl] = STATE(91),
    [sym_preproc_line] = STATE(91),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(631),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_COMMA] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(631),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(713),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(645),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_COLON_GT] = ACTIONS(655),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(655),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(715),
    [anon_sym_AMP_AMP] = ACTIONS(717),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [anon_sym_POUNDelse] = ACTIONS(288),
    [sym__newline] = ACTIONS(719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(92)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(92),
    [sym_compiler_directive_decl] = STATE(92),
    [sym_fsi_directive_decl] = STATE(92),
    [sym_preproc_line] = STATE(92),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(736),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [anon_sym_inherit] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [anon_sym_POUNDelse] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(93)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(93),
    [sym_compiler_directive_decl] = STATE(93),
    [sym_fsi_directive_decl] = STATE(93),
    [sym_preproc_line] = STATE(93),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(727),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [anon_sym_inherit] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [anon_sym_POUNDelse] = ACTIONS(282),
    [sym__newline] = ACTIONS(731),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(94)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(94),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1768),
    [sym_infix_op] = STATE(1752),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(94),
    [sym_compiler_directive_decl] = STATE(94),
    [sym_fsi_directive_decl] = STATE(94),
    [sym_preproc_line] = STATE(94),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_sequential_expression_repeat1] = STATE(3081),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(673),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(681),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [anon_sym_POUNDelse] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(711),
  },
  [STATE(95)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(82),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1759),
    [sym_infix_op] = STATE(2220),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(95),
    [sym_compiler_directive_decl] = STATE(95),
    [sym_fsi_directive_decl] = STATE(95),
    [sym_preproc_line] = STATE(95),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_sequential_expression_repeat1] = STATE(3033),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(545),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_COMMA] = ACTIONS(551),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(545),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(725),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(559),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(561),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_COLON_GT] = ACTIONS(569),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(569),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_DOT_LBRACK] = ACTIONS(587),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_LPAREN2] = ACTIONS(595),
    [anon_sym_PIPE_PIPE] = ACTIONS(727),
    [anon_sym_AMP_AMP] = ACTIONS(729),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [anon_sym_inherit] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(611),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(615),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [anon_sym_POUNDelse] = ACTIONS(288),
    [sym__newline] = ACTIONS(731),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
    [sym__tyapp_open] = ACTIONS(537),
  },
  [STATE(96)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(96),
    [sym_compiler_directive_decl] = STATE(96),
    [sym_fsi_directive_decl] = STATE(96),
    [sym_preproc_line] = STATE(96),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(97)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(97),
    [sym_compiler_directive_decl] = STATE(97),
    [sym_fsi_directive_decl] = STATE(97),
    [sym_preproc_line] = STATE(97),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(98)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(98),
    [sym_compiler_directive_decl] = STATE(98),
    [sym_fsi_directive_decl] = STATE(98),
    [sym_preproc_line] = STATE(98),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(829),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(99)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(99),
    [sym_compiler_directive_decl] = STATE(99),
    [sym_fsi_directive_decl] = STATE(99),
    [sym_preproc_line] = STATE(99),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(832),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(832),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(836),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(100)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(100),
    [sym_compiler_directive_decl] = STATE(100),
    [sym_fsi_directive_decl] = STATE(100),
    [sym_preproc_line] = STATE(100),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(101)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(101),
    [sym_compiler_directive_decl] = STATE(101),
    [sym_fsi_directive_decl] = STATE(101),
    [sym_preproc_line] = STATE(101),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [anon_sym_inherit] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(102)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(102),
    [sym_compiler_directive_decl] = STATE(102),
    [sym_fsi_directive_decl] = STATE(102),
    [sym_preproc_line] = STATE(102),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(832),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(832),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(836),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(103)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(103),
    [sym_compiler_directive_decl] = STATE(103),
    [sym_fsi_directive_decl] = STATE(103),
    [sym_preproc_line] = STATE(103),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(193),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(912),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(104)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(104),
    [sym_compiler_directive_decl] = STATE(104),
    [sym_fsi_directive_decl] = STATE(104),
    [sym_preproc_line] = STATE(104),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(280),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(941),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(941),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [sym__newline] = ACTIONS(945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(105)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(105),
    [sym_compiler_directive_decl] = STATE(105),
    [sym_fsi_directive_decl] = STATE(105),
    [sym_preproc_line] = STATE(105),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(106)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(106),
    [sym_compiler_directive_decl] = STATE(106),
    [sym_fsi_directive_decl] = STATE(106),
    [sym_preproc_line] = STATE(106),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(286),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(941),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(941),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [sym__newline] = ACTIONS(945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(107)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(107),
    [sym_compiler_directive_decl] = STATE(107),
    [sym_fsi_directive_decl] = STATE(107),
    [sym_preproc_line] = STATE(107),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(108)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(108),
    [sym_compiler_directive_decl] = STATE(108),
    [sym_fsi_directive_decl] = STATE(108),
    [sym_preproc_line] = STATE(108),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(941),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(941),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [sym__newline] = ACTIONS(945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(109)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(109),
    [sym_compiler_directive_decl] = STATE(109),
    [sym_fsi_directive_decl] = STATE(109),
    [sym_preproc_line] = STATE(109),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(110)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(110),
    [sym_compiler_directive_decl] = STATE(110),
    [sym_fsi_directive_decl] = STATE(110),
    [sym_preproc_line] = STATE(110),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(111)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(111),
    [sym_compiler_directive_decl] = STATE(111),
    [sym_fsi_directive_decl] = STATE(111),
    [sym_preproc_line] = STATE(111),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(112)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(112),
    [sym_compiler_directive_decl] = STATE(112),
    [sym_fsi_directive_decl] = STATE(112),
    [sym_preproc_line] = STATE(112),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(294),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(832),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(832),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(836),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(113)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(113),
    [sym_compiler_directive_decl] = STATE(113),
    [sym_fsi_directive_decl] = STATE(113),
    [sym_preproc_line] = STATE(113),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(306),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(114)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(114),
    [sym_compiler_directive_decl] = STATE(114),
    [sym_fsi_directive_decl] = STATE(114),
    [sym_preproc_line] = STATE(114),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(941),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(941),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [sym__newline] = ACTIONS(945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(115)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(115),
    [sym_compiler_directive_decl] = STATE(115),
    [sym_fsi_directive_decl] = STATE(115),
    [sym_preproc_line] = STATE(115),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(116)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(116),
    [sym_compiler_directive_decl] = STATE(116),
    [sym_fsi_directive_decl] = STATE(116),
    [sym_preproc_line] = STATE(116),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(117)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(117),
    [sym_compiler_directive_decl] = STATE(117),
    [sym_fsi_directive_decl] = STATE(117),
    [sym_preproc_line] = STATE(117),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(118)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(118),
    [sym_compiler_directive_decl] = STATE(118),
    [sym_fsi_directive_decl] = STATE(118),
    [sym_preproc_line] = STATE(118),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [anon_sym_inherit] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(119)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(119),
    [sym_compiler_directive_decl] = STATE(119),
    [sym_fsi_directive_decl] = STATE(119),
    [sym_preproc_line] = STATE(119),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(941),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(947),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(941),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [sym__newline] = ACTIONS(945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(120)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(120),
    [sym_compiler_directive_decl] = STATE(120),
    [sym_fsi_directive_decl] = STATE(120),
    [sym_preproc_line] = STATE(120),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(310),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(832),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(832),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(836),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(121)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(121),
    [sym_compiler_directive_decl] = STATE(121),
    [sym_fsi_directive_decl] = STATE(121),
    [sym_preproc_line] = STATE(121),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_and] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(122)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(122),
    [sym_compiler_directive_decl] = STATE(122),
    [sym_fsi_directive_decl] = STATE(122),
    [sym_preproc_line] = STATE(122),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1003),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [anon_sym_inherit] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(123)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(123),
    [sym_compiler_directive_decl] = STATE(123),
    [sym_fsi_directive_decl] = STATE(123),
    [sym_preproc_line] = STATE(123),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1032),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(1032),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [anon_sym_inherit] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [sym__newline] = ACTIONS(1036),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(124)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(124),
    [sym_compiler_directive_decl] = STATE(124),
    [sym_fsi_directive_decl] = STATE(124),
    [sym_preproc_line] = STATE(124),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1032),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(129),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(1032),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [anon_sym_inherit] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [sym__newline] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(125)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(125),
    [sym_compiler_directive_decl] = STATE(125),
    [sym_fsi_directive_decl] = STATE(125),
    [sym_preproc_line] = STATE(125),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1032),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(1032),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [anon_sym_inherit] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [sym__newline] = ACTIONS(1036),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(126)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(126),
    [sym_compiler_directive_decl] = STATE(126),
    [sym_fsi_directive_decl] = STATE(126),
    [sym_preproc_line] = STATE(126),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_inherit] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(127)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(127),
    [sym_compiler_directive_decl] = STATE(127),
    [sym_fsi_directive_decl] = STATE(127),
    [sym_preproc_line] = STATE(127),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1032),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(1032),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [anon_sym_inherit] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [sym__newline] = ACTIONS(1036),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(128)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(128),
    [sym_compiler_directive_decl] = STATE(128),
    [sym_fsi_directive_decl] = STATE(128),
    [sym_preproc_line] = STATE(128),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [anon_sym_inherit] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(129)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(129),
    [sym_compiler_directive_decl] = STATE(129),
    [sym_fsi_directive_decl] = STATE(129),
    [sym_preproc_line] = STATE(129),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_inherit] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(130)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(130),
    [sym_compiler_directive_decl] = STATE(130),
    [sym_fsi_directive_decl] = STATE(130),
    [sym_preproc_line] = STATE(130),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [anon_sym_inherit] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(131)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(131),
    [sym_compiler_directive_decl] = STATE(131),
    [sym_fsi_directive_decl] = STATE(131),
    [sym_preproc_line] = STATE(131),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1032),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(171),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(1032),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [anon_sym_inherit] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [anon_sym_POUNDendif] = ACTIONS(173),
    [sym__newline] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(132)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(132),
    [sym_compiler_directive_decl] = STATE(132),
    [sym_fsi_directive_decl] = STATE(132),
    [sym_preproc_line] = STATE(132),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [anon_sym_inherit] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(133)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(133),
    [sym_compiler_directive_decl] = STATE(133),
    [sym_fsi_directive_decl] = STATE(133),
    [sym_preproc_line] = STATE(133),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(955),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_COMMA] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(955),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(967),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(971),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1032),
    [anon_sym_DOT] = ACTIONS(848),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_COLON_GT] = ACTIONS(979),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(979),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_DOT_LBRACK] = ACTIONS(850),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(852),
    [anon_sym_PIPE_PIPE] = ACTIONS(1032),
    [anon_sym_AMP_AMP] = ACTIONS(1034),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [anon_sym_inherit] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(854),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(1022),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [sym__newline] = ACTIONS(1036),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(134)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(134),
    [sym_compiler_directive_decl] = STATE(134),
    [sym_fsi_directive_decl] = STATE(134),
    [sym_preproc_line] = STATE(134),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(832),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(1038),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(832),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [sym__newline] = ACTIONS(133),
    [sym__dedent] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(135)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(135),
    [sym_compiler_directive_decl] = STATE(135),
    [sym_fsi_directive_decl] = STATE(135),
    [sym_preproc_line] = STATE(135),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_and] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(832),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(1040),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(832),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [sym__newline] = ACTIONS(173),
    [sym__dedent] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(136)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(101),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1785),
    [sym_infix_op] = STATE(1771),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(136),
    [sym_compiler_directive_decl] = STATE(136),
    [sym_fsi_directive_decl] = STATE(136),
    [sym_preproc_line] = STATE(136),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_sequential_expression_repeat1] = STATE(3165),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [anon_sym_inherit] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(856),
  },
  [STATE(137)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(137),
    [sym_compiler_directive_decl] = STATE(137),
    [sym_fsi_directive_decl] = STATE(137),
    [sym_preproc_line] = STATE(137),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_and] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(138)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(138),
    [sym_compiler_directive_decl] = STATE(138),
    [sym_fsi_directive_decl] = STATE(138),
    [sym_preproc_line] = STATE(138),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(139)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(137),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1802),
    [sym_infix_op] = STATE(1788),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(139),
    [sym_compiler_directive_decl] = STATE(139),
    [sym_fsi_directive_decl] = STATE(139),
    [sym_preproc_line] = STATE(139),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_sequential_expression_repeat1] = STATE(3309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(864),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_and] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_COMMA] = ACTIONS(870),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(864),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(876),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(880),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(941),
    [anon_sym_DOT] = ACTIONS(838),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(1042),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_COLON_GT] = ACTIONS(888),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(888),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_DOT_LBRACK] = ACTIONS(840),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_LPAREN2] = ACTIONS(842),
    [anon_sym_PIPE_PIPE] = ACTIONS(941),
    [anon_sym_AMP_AMP] = ACTIONS(943),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(844),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(931),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [anon_sym_POUNDendif] = ACTIONS(133),
    [sym__newline] = ACTIONS(945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
    [sym__tyapp_open] = ACTIONS(846),
  },
  [STATE(140)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(97),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1777),
    [sym_infix_op] = STATE(1763),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(140),
    [sym_compiler_directive_decl] = STATE(140),
    [sym_fsi_directive_decl] = STATE(140),
    [sym_preproc_line] = STATE(140),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_sequential_expression_repeat1] = STATE(3313),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(745),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_and] = ACTIONS(302),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(751),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(745),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(827),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(759),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(761),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_COLON_GT] = ACTIONS(769),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(769),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_DOT_LBRACK] = ACTIONS(787),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(795),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(834),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(815),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
    [sym__tyapp_open] = ACTIONS(825),
  },
  [STATE(141)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(141),
    [sym_compiler_directive_decl] = STATE(141),
    [sym_fsi_directive_decl] = STATE(141),
    [sym_preproc_line] = STATE(141),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(199),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1104),
    [anon_sym_default] = ACTIONS(193),
    [anon_sym_static] = ACTIONS(193),
    [anon_sym_member] = ACTIONS(193),
    [anon_sym_interface] = ACTIONS(193),
    [anon_sym_abstract] = ACTIONS(193),
    [anon_sym_val] = ACTIONS(193),
    [anon_sym_override] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(142)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(142),
    [sym_compiler_directive_decl] = STATE(142),
    [sym_fsi_directive_decl] = STATE(142),
    [sym_preproc_line] = STATE(142),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(1139),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [sym__newline] = ACTIONS(173),
    [sym__dedent] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(143)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(143),
    [sym_compiler_directive_decl] = STATE(143),
    [sym_fsi_directive_decl] = STATE(143),
    [sym_preproc_line] = STATE(143),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(1143),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(144)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(144),
    [sym_compiler_directive_decl] = STATE(144),
    [sym_fsi_directive_decl] = STATE(144),
    [sym_preproc_line] = STATE(144),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(171),
    [anon_sym_let_BANG] = ACTIONS(173),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(1147),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(145)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(145),
    [sym_compiler_directive_decl] = STATE(145),
    [sym_fsi_directive_decl] = STATE(145),
    [sym_preproc_line] = STATE(145),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(1143),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(146)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(146),
    [sym_compiler_directive_decl] = STATE(146),
    [sym_fsi_directive_decl] = STATE(146),
    [sym_preproc_line] = STATE(146),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(1147),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_interface] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(173),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(147)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(147),
    [sym_compiler_directive_decl] = STATE(147),
    [sym_fsi_directive_decl] = STATE(147),
    [sym_preproc_line] = STATE(147),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(129),
    [anon_sym_let] = ACTIONS(129),
    [anon_sym_let_BANG] = ACTIONS(133),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(1149),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_interface] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(133),
    [sym__newline] = ACTIONS(133),
    [sym__dedent] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(148)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(148),
    [sym_compiler_directive_decl] = STATE(148),
    [sym_fsi_directive_decl] = STATE(148),
    [sym_preproc_line] = STATE(148),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [aux_sym_access_modifier_token1] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_default] = ACTIONS(181),
    [anon_sym_static] = ACTIONS(181),
    [anon_sym_member] = ACTIONS(181),
    [anon_sym_interface] = ACTIONS(181),
    [anon_sym_abstract] = ACTIONS(181),
    [anon_sym_val] = ACTIONS(181),
    [anon_sym_override] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(149)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(149),
    [sym_compiler_directive_decl] = STATE(149),
    [sym_fsi_directive_decl] = STATE(149),
    [sym_preproc_line] = STATE(149),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [aux_sym_access_modifier_token1] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_default] = ACTIONS(159),
    [anon_sym_static] = ACTIONS(159),
    [anon_sym_member] = ACTIONS(159),
    [anon_sym_interface] = ACTIONS(159),
    [anon_sym_abstract] = ACTIONS(159),
    [anon_sym_val] = ACTIONS(159),
    [anon_sym_override] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(150)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(150),
    [sym_compiler_directive_decl] = STATE(150),
    [sym_fsi_directive_decl] = STATE(150),
    [sym_preproc_line] = STATE(150),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [aux_sym_access_modifier_token1] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_default] = ACTIONS(187),
    [anon_sym_static] = ACTIONS(187),
    [anon_sym_member] = ACTIONS(187),
    [anon_sym_interface] = ACTIONS(187),
    [anon_sym_abstract] = ACTIONS(187),
    [anon_sym_val] = ACTIONS(187),
    [anon_sym_override] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(151)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(151),
    [sym_compiler_directive_decl] = STATE(151),
    [sym_fsi_directive_decl] = STATE(151),
    [sym_preproc_line] = STATE(151),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(152)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(152),
    [sym_compiler_directive_decl] = STATE(152),
    [sym_fsi_directive_decl] = STATE(152),
    [sym_preproc_line] = STATE(152),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(296),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_static] = ACTIONS(294),
    [anon_sym_member] = ACTIONS(294),
    [anon_sym_interface] = ACTIONS(294),
    [anon_sym_abstract] = ACTIONS(294),
    [anon_sym_val] = ACTIONS(294),
    [anon_sym_override] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(153)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(153),
    [sym_compiler_directive_decl] = STATE(153),
    [sym_fsi_directive_decl] = STATE(153),
    [sym_preproc_line] = STATE(153),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(300),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_default] = ACTIONS(298),
    [anon_sym_static] = ACTIONS(298),
    [anon_sym_member] = ACTIONS(298),
    [anon_sym_interface] = ACTIONS(298),
    [anon_sym_abstract] = ACTIONS(298),
    [anon_sym_val] = ACTIONS(298),
    [anon_sym_override] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(154)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(154),
    [sym_compiler_directive_decl] = STATE(154),
    [sym_fsi_directive_decl] = STATE(154),
    [sym_preproc_line] = STATE(154),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(155)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(155),
    [sym_compiler_directive_decl] = STATE(155),
    [sym_fsi_directive_decl] = STATE(155),
    [sym_preproc_line] = STATE(155),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(304),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(302),
    [anon_sym_static] = ACTIONS(302),
    [anon_sym_member] = ACTIONS(302),
    [anon_sym_interface] = ACTIONS(302),
    [anon_sym_abstract] = ACTIONS(302),
    [anon_sym_val] = ACTIONS(302),
    [anon_sym_override] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(156)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(156),
    [sym_compiler_directive_decl] = STATE(156),
    [sym_fsi_directive_decl] = STATE(156),
    [sym_preproc_line] = STATE(156),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(308),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_default] = ACTIONS(306),
    [anon_sym_static] = ACTIONS(306),
    [anon_sym_member] = ACTIONS(306),
    [anon_sym_interface] = ACTIONS(306),
    [anon_sym_abstract] = ACTIONS(306),
    [anon_sym_val] = ACTIONS(306),
    [anon_sym_override] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(157)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(157),
    [sym_compiler_directive_decl] = STATE(157),
    [sym_fsi_directive_decl] = STATE(157),
    [sym_preproc_line] = STATE(157),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(280),
    [anon_sym_static] = ACTIONS(280),
    [anon_sym_member] = ACTIONS(280),
    [anon_sym_interface] = ACTIONS(280),
    [anon_sym_abstract] = ACTIONS(280),
    [anon_sym_val] = ACTIONS(280),
    [anon_sym_override] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(158)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(158),
    [sym_compiler_directive_decl] = STATE(158),
    [sym_fsi_directive_decl] = STATE(158),
    [sym_preproc_line] = STATE(158),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(288),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(286),
    [anon_sym_static] = ACTIONS(286),
    [anon_sym_member] = ACTIONS(286),
    [anon_sym_interface] = ACTIONS(286),
    [anon_sym_abstract] = ACTIONS(286),
    [anon_sym_val] = ACTIONS(286),
    [anon_sym_override] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(159)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(159),
    [sym_compiler_directive_decl] = STATE(159),
    [sym_fsi_directive_decl] = STATE(159),
    [sym_preproc_line] = STATE(159),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(312),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(310),
    [anon_sym_static] = ACTIONS(310),
    [anon_sym_member] = ACTIONS(310),
    [anon_sym_interface] = ACTIONS(310),
    [anon_sym_abstract] = ACTIONS(310),
    [anon_sym_val] = ACTIONS(310),
    [anon_sym_override] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1145),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(160)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(160),
    [sym_compiler_directive_decl] = STATE(160),
    [sym_fsi_directive_decl] = STATE(160),
    [sym_preproc_line] = STATE(160),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(133),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(133),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(129),
    [anon_sym_with] = ACTIONS(1151),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(129),
    [anon_sym_static] = ACTIONS(129),
    [anon_sym_member] = ACTIONS(129),
    [anon_sym_abstract] = ACTIONS(129),
    [anon_sym_val] = ACTIONS(129),
    [anon_sym_override] = ACTIONS(129),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(133),
    [sym__dedent] = ACTIONS(133),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(161)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(149),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1727),
    [sym_infix_op] = STATE(1594),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(161),
    [sym_compiler_directive_decl] = STATE(161),
    [sym_fsi_directive_decl] = STATE(161),
    [sym_preproc_line] = STATE(161),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(3536),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_LBRACK_LT] = ACTIONS(173),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [aux_sym_access_modifier_token1] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1056),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1137),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(171),
    [anon_sym_with] = ACTIONS(1153),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1137),
    [anon_sym_AMP_AMP] = ACTIONS(1141),
    [anon_sym_default] = ACTIONS(171),
    [anon_sym_static] = ACTIONS(171),
    [anon_sym_member] = ACTIONS(171),
    [anon_sym_abstract] = ACTIONS(171),
    [anon_sym_val] = ACTIONS(171),
    [anon_sym_override] = ACTIONS(171),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(173),
    [sym__dedent] = ACTIONS(173),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(162)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(162),
    [sym_compiler_directive_decl] = STATE(162),
    [sym_fsi_directive_decl] = STATE(162),
    [sym_preproc_line] = STATE(162),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [anon_sym_POUNDelse] = ACTIONS(282),
    [sym__newline] = ACTIONS(1245),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(163)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(163),
    [sym_compiler_directive_decl] = STATE(163),
    [sym_fsi_directive_decl] = STATE(163),
    [sym_preproc_line] = STATE(163),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(1251),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(1251),
    [anon_sym_type] = ACTIONS(1251),
    [anon_sym_LBRACK_LT] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(1251),
    [anon_sym_extern] = ACTIONS(1251),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1253),
    [anon_sym_POUNDload] = ACTIONS(1253),
    [anon_sym_POUNDtime] = ACTIONS(1253),
    [anon_sym_POUNDI] = ACTIONS(1253),
    [anon_sym_POUNDhelp] = ACTIONS(1253),
    [anon_sym_POUNDquit] = ACTIONS(1253),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(1253),
    [anon_sym_POUNDelse] = ACTIONS(1253),
    [sym__newline] = ACTIONS(1245),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(164)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(164),
    [sym_compiler_directive_decl] = STATE(164),
    [sym_fsi_directive_decl] = STATE(164),
    [sym_preproc_line] = STATE(164),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(1255),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(1255),
    [anon_sym_type] = ACTIONS(1255),
    [anon_sym_LBRACK_LT] = ACTIONS(1257),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(1255),
    [anon_sym_extern] = ACTIONS(1255),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1257),
    [anon_sym_POUNDload] = ACTIONS(1257),
    [anon_sym_POUNDtime] = ACTIONS(1257),
    [anon_sym_POUNDI] = ACTIONS(1257),
    [anon_sym_POUNDhelp] = ACTIONS(1257),
    [anon_sym_POUNDquit] = ACTIONS(1257),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(1257),
    [anon_sym_POUNDelse] = ACTIONS(1257),
    [sym__newline] = ACTIONS(1245),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(165)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(165),
    [sym_compiler_directive_decl] = STATE(165),
    [sym_fsi_directive_decl] = STATE(165),
    [sym_preproc_line] = STATE(165),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1259),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [anon_sym_POUNDelse] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(166)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(166),
    [sym_compiler_directive_decl] = STATE(166),
    [sym_fsi_directive_decl] = STATE(166),
    [sym_preproc_line] = STATE(166),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(1257),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(1255),
    [anon_sym_module] = ACTIONS(1255),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(1255),
    [anon_sym_type] = ACTIONS(1255),
    [anon_sym_LBRACK_LT] = ACTIONS(1257),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1272),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(1255),
    [anon_sym_extern] = ACTIONS(1255),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1257),
    [anon_sym_POUNDload] = ACTIONS(1257),
    [anon_sym_POUNDtime] = ACTIONS(1257),
    [anon_sym_POUNDI] = ACTIONS(1257),
    [anon_sym_POUNDhelp] = ACTIONS(1257),
    [anon_sym_POUNDquit] = ACTIONS(1257),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1302),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(167)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(167),
    [sym_compiler_directive_decl] = STATE(167),
    [sym_fsi_directive_decl] = STATE(167),
    [sym_preproc_line] = STATE(167),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [anon_sym_POUNDelse] = ACTIONS(288),
    [sym__newline] = ACTIONS(1245),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(168)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(168),
    [sym_compiler_directive_decl] = STATE(168),
    [sym_fsi_directive_decl] = STATE(168),
    [sym_preproc_line] = STATE(168),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(169)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(169),
    [sym_compiler_directive_decl] = STATE(169),
    [sym_fsi_directive_decl] = STATE(169),
    [sym_preproc_line] = STATE(169),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [anon_sym_POUNDelse] = ACTIONS(296),
    [sym__newline] = ACTIONS(1245),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(170)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(170),
    [sym_compiler_directive_decl] = STATE(170),
    [sym_fsi_directive_decl] = STATE(170),
    [sym_preproc_line] = STATE(170),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [anon_sym_POUNDelse] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(171)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(171),
    [sym_compiler_directive_decl] = STATE(171),
    [sym_fsi_directive_decl] = STATE(171),
    [sym_preproc_line] = STATE(171),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(172)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(172),
    [sym_compiler_directive_decl] = STATE(172),
    [sym_fsi_directive_decl] = STATE(172),
    [sym_preproc_line] = STATE(172),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(173)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(173),
    [sym_compiler_directive_decl] = STATE(173),
    [sym_fsi_directive_decl] = STATE(173),
    [sym_preproc_line] = STATE(173),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [anon_sym_POUNDelse] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(174)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(174),
    [sym_compiler_directive_decl] = STATE(174),
    [sym_fsi_directive_decl] = STATE(174),
    [sym_preproc_line] = STATE(174),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [anon_sym_POUNDelse] = ACTIONS(312),
    [sym__newline] = ACTIONS(1245),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(175)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(175),
    [sym_compiler_directive_decl] = STATE(175),
    [sym_fsi_directive_decl] = STATE(175),
    [sym_preproc_line] = STATE(175),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(157),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_namespace] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(176)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(176),
    [sym_compiler_directive_decl] = STATE(176),
    [sym_fsi_directive_decl] = STATE(176),
    [sym_preproc_line] = STATE(176),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(185),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_namespace] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(177)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(177),
    [sym_compiler_directive_decl] = STATE(177),
    [sym_fsi_directive_decl] = STATE(177),
    [sym_preproc_line] = STATE(177),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(183),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_namespace] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(178)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(178),
    [sym_compiler_directive_decl] = STATE(178),
    [sym_fsi_directive_decl] = STATE(178),
    [sym_preproc_line] = STATE(178),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(199),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(193),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1306),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(179)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(179),
    [sym_compiler_directive_decl] = STATE(179),
    [sym_fsi_directive_decl] = STATE(179),
    [sym_preproc_line] = STATE(179),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(282),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(280),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1272),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1302),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(180)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(180),
    [sym_compiler_directive_decl] = STATE(180),
    [sym_fsi_directive_decl] = STATE(180),
    [sym_preproc_line] = STATE(180),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(288),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(286),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1272),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1302),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(181)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(181),
    [sym_compiler_directive_decl] = STATE(181),
    [sym_fsi_directive_decl] = STATE(181),
    [sym_preproc_line] = STATE(181),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(292),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(182)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(182),
    [sym_compiler_directive_decl] = STATE(182),
    [sym_fsi_directive_decl] = STATE(182),
    [sym_preproc_line] = STATE(182),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(296),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(294),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1272),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1302),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(183)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(183),
    [sym_compiler_directive_decl] = STATE(183),
    [sym_fsi_directive_decl] = STATE(183),
    [sym_preproc_line] = STATE(183),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(300),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(298),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(184)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(184),
    [sym_compiler_directive_decl] = STATE(184),
    [sym_fsi_directive_decl] = STATE(184),
    [sym_preproc_line] = STATE(184),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(302),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(185)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(185),
    [sym_compiler_directive_decl] = STATE(185),
    [sym_fsi_directive_decl] = STATE(185),
    [sym_preproc_line] = STATE(185),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(302),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(186)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(186),
    [sym_compiler_directive_decl] = STATE(186),
    [sym_fsi_directive_decl] = STATE(186),
    [sym_preproc_line] = STATE(186),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(308),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(306),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(187)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(187),
    [sym_compiler_directive_decl] = STATE(187),
    [sym_fsi_directive_decl] = STATE(187),
    [sym_preproc_line] = STATE(187),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(312),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(310),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1272),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1302),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(188)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(188),
    [sym_compiler_directive_decl] = STATE(188),
    [sym_fsi_directive_decl] = STATE(188),
    [sym_preproc_line] = STATE(188),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [anon_sym_POUNDelse] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(189)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(189),
    [sym_compiler_directive_decl] = STATE(189),
    [sym_fsi_directive_decl] = STATE(189),
    [sym_preproc_line] = STATE(189),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [anon_sym_POUNDelse] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(190)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(190),
    [sym_compiler_directive_decl] = STATE(190),
    [sym_fsi_directive_decl] = STATE(190),
    [sym_preproc_line] = STATE(190),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(3864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [anon_sym_POUNDelse] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(191)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(175),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1540),
    [sym_infix_op] = STATE(1554),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(191),
    [sym_compiler_directive_decl] = STATE(191),
    [sym_fsi_directive_decl] = STATE(191),
    [sym_preproc_line] = STATE(191),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3811),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(1253),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(1251),
    [anon_sym_module] = ACTIONS(1251),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(1251),
    [anon_sym_type] = ACTIONS(1251),
    [anon_sym_LBRACK_LT] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1272),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1272),
    [anon_sym_AMP_AMP] = ACTIONS(1290),
    [anon_sym_exception] = ACTIONS(1251),
    [anon_sym_extern] = ACTIONS(1251),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1253),
    [anon_sym_POUNDload] = ACTIONS(1253),
    [anon_sym_POUNDtime] = ACTIONS(1253),
    [anon_sym_POUNDI] = ACTIONS(1253),
    [anon_sym_POUNDhelp] = ACTIONS(1253),
    [anon_sym_POUNDquit] = ACTIONS(1253),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1302),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(192)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(192),
    [sym_compiler_directive_decl] = STATE(192),
    [sym_fsi_directive_decl] = STATE(192),
    [sym_preproc_line] = STATE(192),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(199),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1311),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(193)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(193),
    [sym_compiler_directive_decl] = STATE(193),
    [sym_fsi_directive_decl] = STATE(193),
    [sym_preproc_line] = STATE(193),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(308),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(194)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(194),
    [sym_compiler_directive_decl] = STATE(194),
    [sym_fsi_directive_decl] = STATE(194),
    [sym_preproc_line] = STATE(194),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(312),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1318),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(195)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(195),
    [sym_compiler_directive_decl] = STATE(195),
    [sym_fsi_directive_decl] = STATE(195),
    [sym_preproc_line] = STATE(195),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(183),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(196)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(196),
    [sym_compiler_directive_decl] = STATE(196),
    [sym_fsi_directive_decl] = STATE(196),
    [sym_preproc_line] = STATE(196),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_module] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_open] = ACTIONS(181),
    [anon_sym_type] = ACTIONS(181),
    [anon_sym_LBRACK_LT] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_exception] = ACTIONS(181),
    [anon_sym_extern] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(183),
    [anon_sym_POUNDload] = ACTIONS(183),
    [anon_sym_POUNDtime] = ACTIONS(183),
    [anon_sym_POUNDI] = ACTIONS(183),
    [anon_sym_POUNDhelp] = ACTIONS(183),
    [anon_sym_POUNDquit] = ACTIONS(183),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(197)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(197),
    [sym_compiler_directive_decl] = STATE(197),
    [sym_fsi_directive_decl] = STATE(197),
    [sym_preproc_line] = STATE(197),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(1251),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(1251),
    [anon_sym_type] = ACTIONS(1251),
    [anon_sym_LBRACK_LT] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(1251),
    [anon_sym_extern] = ACTIONS(1251),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1253),
    [anon_sym_POUNDload] = ACTIONS(1253),
    [anon_sym_POUNDtime] = ACTIONS(1253),
    [anon_sym_POUNDI] = ACTIONS(1253),
    [anon_sym_POUNDhelp] = ACTIONS(1253),
    [anon_sym_POUNDquit] = ACTIONS(1253),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(1253),
    [sym__newline] = ACTIONS(1412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(198)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(198),
    [sym_compiler_directive_decl] = STATE(198),
    [sym_fsi_directive_decl] = STATE(198),
    [sym_preproc_line] = STATE(198),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(199)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(199),
    [sym_compiler_directive_decl] = STATE(199),
    [sym_fsi_directive_decl] = STATE(199),
    [sym_preproc_line] = STATE(199),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(282),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1318),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(200)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(200),
    [sym_compiler_directive_decl] = STATE(200),
    [sym_fsi_directive_decl] = STATE(200),
    [sym_preproc_line] = STATE(200),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(1255),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(1255),
    [anon_sym_type] = ACTIONS(1255),
    [anon_sym_LBRACK_LT] = ACTIONS(1257),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(1255),
    [anon_sym_extern] = ACTIONS(1255),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1257),
    [anon_sym_POUNDload] = ACTIONS(1257),
    [anon_sym_POUNDtime] = ACTIONS(1257),
    [anon_sym_POUNDI] = ACTIONS(1257),
    [anon_sym_POUNDhelp] = ACTIONS(1257),
    [anon_sym_POUNDquit] = ACTIONS(1257),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(1257),
    [sym__newline] = ACTIONS(1412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(201)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(201),
    [sym_compiler_directive_decl] = STATE(201),
    [sym_fsi_directive_decl] = STATE(201),
    [sym_preproc_line] = STATE(201),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(288),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1318),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(202)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(202),
    [sym_compiler_directive_decl] = STATE(202),
    [sym_fsi_directive_decl] = STATE(202),
    [sym_preproc_line] = STATE(202),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(157),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_module] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_open] = ACTIONS(159),
    [anon_sym_type] = ACTIONS(159),
    [anon_sym_LBRACK_LT] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_exception] = ACTIONS(159),
    [anon_sym_extern] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(157),
    [anon_sym_POUNDload] = ACTIONS(157),
    [anon_sym_POUNDtime] = ACTIONS(157),
    [anon_sym_POUNDI] = ACTIONS(157),
    [anon_sym_POUNDhelp] = ACTIONS(157),
    [anon_sym_POUNDquit] = ACTIONS(157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(203)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(203),
    [sym_compiler_directive_decl] = STATE(203),
    [sym_fsi_directive_decl] = STATE(203),
    [sym_preproc_line] = STATE(203),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(204)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(204),
    [sym_compiler_directive_decl] = STATE(204),
    [sym_fsi_directive_decl] = STATE(204),
    [sym_preproc_line] = STATE(204),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(1253),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(1251),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(1251),
    [anon_sym_type] = ACTIONS(1251),
    [anon_sym_LBRACK_LT] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(1251),
    [anon_sym_extern] = ACTIONS(1251),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1253),
    [anon_sym_POUNDload] = ACTIONS(1253),
    [anon_sym_POUNDtime] = ACTIONS(1253),
    [anon_sym_POUNDI] = ACTIONS(1253),
    [anon_sym_POUNDhelp] = ACTIONS(1253),
    [anon_sym_POUNDquit] = ACTIONS(1253),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1318),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(205)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(205),
    [sym_compiler_directive_decl] = STATE(205),
    [sym_fsi_directive_decl] = STATE(205),
    [sym_preproc_line] = STATE(205),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(292),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(206)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(206),
    [sym_compiler_directive_decl] = STATE(206),
    [sym_fsi_directive_decl] = STATE(206),
    [sym_preproc_line] = STATE(206),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(296),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1318),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(207)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(207),
    [sym_compiler_directive_decl] = STATE(207),
    [sym_fsi_directive_decl] = STATE(207),
    [sym_preproc_line] = STATE(207),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(300),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(208)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(208),
    [sym_compiler_directive_decl] = STATE(208),
    [sym_fsi_directive_decl] = STATE(208),
    [sym_preproc_line] = STATE(208),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(209)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(209),
    [sym_compiler_directive_decl] = STATE(209),
    [sym_fsi_directive_decl] = STATE(209),
    [sym_preproc_line] = STATE(209),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(304),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(210)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(210),
    [sym_compiler_directive_decl] = STATE(210),
    [sym_fsi_directive_decl] = STATE(210),
    [sym_preproc_line] = STATE(210),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(185),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_module] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_open] = ACTIONS(187),
    [anon_sym_type] = ACTIONS(187),
    [anon_sym_LBRACK_LT] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_exception] = ACTIONS(187),
    [anon_sym_extern] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(185),
    [anon_sym_POUNDload] = ACTIONS(185),
    [anon_sym_POUNDtime] = ACTIONS(185),
    [anon_sym_POUNDI] = ACTIONS(185),
    [anon_sym_POUNDhelp] = ACTIONS(185),
    [anon_sym_POUNDquit] = ACTIONS(185),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(211)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(202),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1815),
    [sym_infix_op] = STATE(1527),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(211),
    [sym_compiler_directive_decl] = STATE(211),
    [sym_fsi_directive_decl] = STATE(211),
    [sym_preproc_line] = STATE(211),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_sequential_expression_repeat1] = STATE(3889),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(1257),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(1255),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(1255),
    [anon_sym_type] = ACTIONS(1255),
    [anon_sym_LBRACK_LT] = ACTIONS(1257),
    [anon_sym_COLON] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1262),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(1270),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_COLON_GT] = ACTIONS(1280),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_DOT_LBRACK] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(1255),
    [anon_sym_extern] = ACTIONS(1255),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(1292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(1257),
    [anon_sym_POUNDload] = ACTIONS(1257),
    [anon_sym_POUNDtime] = ACTIONS(1257),
    [anon_sym_POUNDI] = ACTIONS(1257),
    [anon_sym_POUNDhelp] = ACTIONS(1257),
    [anon_sym_POUNDquit] = ACTIONS(1257),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__newline] = ACTIONS(1318),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
    [sym__tyapp_open] = ACTIONS(1304),
  },
  [STATE(212)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(212),
    [sym_compiler_directive_decl] = STATE(212),
    [sym_fsi_directive_decl] = STATE(212),
    [sym_preproc_line] = STATE(212),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(193),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(193),
    [anon_sym_type] = ACTIONS(193),
    [anon_sym_LBRACK_LT] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1416),
    [anon_sym_exception] = ACTIONS(193),
    [anon_sym_extern] = ACTIONS(193),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(199),
    [anon_sym_POUNDload] = ACTIONS(199),
    [anon_sym_POUNDtime] = ACTIONS(199),
    [anon_sym_POUNDI] = ACTIONS(199),
    [anon_sym_POUNDhelp] = ACTIONS(199),
    [anon_sym_POUNDquit] = ACTIONS(199),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(213)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(213),
    [sym_compiler_directive_decl] = STATE(213),
    [sym_fsi_directive_decl] = STATE(213),
    [sym_preproc_line] = STATE(213),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(280),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(280),
    [anon_sym_type] = ACTIONS(280),
    [anon_sym_LBRACK_LT] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(280),
    [anon_sym_extern] = ACTIONS(280),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(282),
    [anon_sym_POUNDload] = ACTIONS(282),
    [anon_sym_POUNDtime] = ACTIONS(282),
    [anon_sym_POUNDI] = ACTIONS(282),
    [anon_sym_POUNDhelp] = ACTIONS(282),
    [anon_sym_POUNDquit] = ACTIONS(282),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [sym__newline] = ACTIONS(1412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(214)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(214),
    [sym_compiler_directive_decl] = STATE(214),
    [sym_fsi_directive_decl] = STATE(214),
    [sym_preproc_line] = STATE(214),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(286),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(286),
    [anon_sym_type] = ACTIONS(286),
    [anon_sym_LBRACK_LT] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(286),
    [anon_sym_extern] = ACTIONS(286),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(288),
    [anon_sym_POUNDload] = ACTIONS(288),
    [anon_sym_POUNDtime] = ACTIONS(288),
    [anon_sym_POUNDI] = ACTIONS(288),
    [anon_sym_POUNDhelp] = ACTIONS(288),
    [anon_sym_POUNDquit] = ACTIONS(288),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [sym__newline] = ACTIONS(1412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(215)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(215),
    [sym_compiler_directive_decl] = STATE(215),
    [sym_fsi_directive_decl] = STATE(215),
    [sym_preproc_line] = STATE(215),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(216)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(216),
    [sym_compiler_directive_decl] = STATE(216),
    [sym_fsi_directive_decl] = STATE(216),
    [sym_preproc_line] = STATE(216),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(294),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(294),
    [anon_sym_type] = ACTIONS(294),
    [anon_sym_LBRACK_LT] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(294),
    [anon_sym_extern] = ACTIONS(294),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(296),
    [anon_sym_POUNDload] = ACTIONS(296),
    [anon_sym_POUNDtime] = ACTIONS(296),
    [anon_sym_POUNDI] = ACTIONS(296),
    [anon_sym_POUNDhelp] = ACTIONS(296),
    [anon_sym_POUNDquit] = ACTIONS(296),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [sym__newline] = ACTIONS(1412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(217)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(217),
    [sym_compiler_directive_decl] = STATE(217),
    [sym_fsi_directive_decl] = STATE(217),
    [sym_preproc_line] = STATE(217),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(298),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(298),
    [anon_sym_type] = ACTIONS(298),
    [anon_sym_LBRACK_LT] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_exception] = ACTIONS(298),
    [anon_sym_extern] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(300),
    [anon_sym_POUNDload] = ACTIONS(300),
    [anon_sym_POUNDtime] = ACTIONS(300),
    [anon_sym_POUNDI] = ACTIONS(300),
    [anon_sym_POUNDhelp] = ACTIONS(300),
    [anon_sym_POUNDquit] = ACTIONS(300),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(218)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(218),
    [sym_compiler_directive_decl] = STATE(218),
    [sym_fsi_directive_decl] = STATE(218),
    [sym_preproc_line] = STATE(218),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(219)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(219),
    [sym_compiler_directive_decl] = STATE(219),
    [sym_fsi_directive_decl] = STATE(219),
    [sym_preproc_line] = STATE(219),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(302),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(302),
    [anon_sym_type] = ACTIONS(302),
    [anon_sym_LBRACK_LT] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(302),
    [anon_sym_extern] = ACTIONS(302),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(304),
    [anon_sym_POUNDload] = ACTIONS(304),
    [anon_sym_POUNDtime] = ACTIONS(304),
    [anon_sym_POUNDI] = ACTIONS(304),
    [anon_sym_POUNDhelp] = ACTIONS(304),
    [anon_sym_POUNDquit] = ACTIONS(304),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(220)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(220),
    [sym_compiler_directive_decl] = STATE(220),
    [sym_fsi_directive_decl] = STATE(220),
    [sym_preproc_line] = STATE(220),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(306),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(306),
    [anon_sym_type] = ACTIONS(306),
    [anon_sym_LBRACK_LT] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_exception] = ACTIONS(306),
    [anon_sym_extern] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(308),
    [anon_sym_POUNDload] = ACTIONS(308),
    [anon_sym_POUNDtime] = ACTIONS(308),
    [anon_sym_POUNDI] = ACTIONS(308),
    [anon_sym_POUNDhelp] = ACTIONS(308),
    [anon_sym_POUNDquit] = ACTIONS(308),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(221)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(221),
    [sym_compiler_directive_decl] = STATE(221),
    [sym_fsi_directive_decl] = STATE(221),
    [sym_preproc_line] = STATE(221),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(3958),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(310),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(310),
    [anon_sym_type] = ACTIONS(310),
    [anon_sym_LBRACK_LT] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [anon_sym_exception] = ACTIONS(310),
    [anon_sym_extern] = ACTIONS(310),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(312),
    [anon_sym_POUNDload] = ACTIONS(312),
    [anon_sym_POUNDtime] = ACTIONS(312),
    [anon_sym_POUNDI] = ACTIONS(312),
    [anon_sym_POUNDhelp] = ACTIONS(312),
    [anon_sym_POUNDquit] = ACTIONS(312),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [sym__newline] = ACTIONS(1412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(222)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(222),
    [sym_compiler_directive_decl] = STATE(222),
    [sym_fsi_directive_decl] = STATE(222),
    [sym_preproc_line] = STATE(222),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(282),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(282),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(280),
    [anon_sym_downto] = ACTIONS(280),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(223)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(223),
    [sym_compiler_directive_decl] = STATE(223),
    [sym_fsi_directive_decl] = STATE(223),
    [sym_preproc_line] = STATE(223),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_GT_RBRACK] = ACTIONS(157),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_RBRACK] = ACTIONS(157),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_RBRACE] = ACTIONS(157),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_to] = ACTIONS(159),
    [anon_sym_downto] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(224)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(224),
    [sym_compiler_directive_decl] = STATE(224),
    [sym_fsi_directive_decl] = STATE(224),
    [sym_preproc_line] = STATE(224),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(288),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(288),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(288),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(286),
    [anon_sym_downto] = ACTIONS(286),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(225)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym__object_members] = STATE(16598),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(225),
    [sym_compiler_directive_decl] = STATE(225),
    [sym_fsi_directive_decl] = STATE(225),
    [sym_preproc_line] = STATE(225),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(1525),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(1547),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(199),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1579),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(226)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(226),
    [sym_compiler_directive_decl] = STATE(226),
    [sym_fsi_directive_decl] = STATE(226),
    [sym_preproc_line] = STATE(226),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(292),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_RBRACE] = ACTIONS(292),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(290),
    [anon_sym_downto] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(227)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(227),
    [sym_compiler_directive_decl] = STATE(227),
    [sym_fsi_directive_decl] = STATE(227),
    [sym_preproc_line] = STATE(227),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(296),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(296),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(296),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(294),
    [anon_sym_downto] = ACTIONS(294),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(228)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(228),
    [sym_compiler_directive_decl] = STATE(228),
    [sym_fsi_directive_decl] = STATE(228),
    [sym_preproc_line] = STATE(228),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(300),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(300),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_RBRACE] = ACTIONS(300),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(298),
    [anon_sym_downto] = ACTIONS(298),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(229)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(229),
    [sym_compiler_directive_decl] = STATE(229),
    [sym_fsi_directive_decl] = STATE(229),
    [sym_preproc_line] = STATE(229),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(304),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(304),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(302),
    [anon_sym_downto] = ACTIONS(302),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(230)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(230),
    [sym_compiler_directive_decl] = STATE(230),
    [sym_fsi_directive_decl] = STATE(230),
    [sym_preproc_line] = STATE(230),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(304),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(304),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(304),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(302),
    [anon_sym_downto] = ACTIONS(302),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(231)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(231),
    [sym_compiler_directive_decl] = STATE(231),
    [sym_fsi_directive_decl] = STATE(231),
    [sym_preproc_line] = STATE(231),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(308),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_RBRACE] = ACTIONS(308),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(306),
    [anon_sym_downto] = ACTIONS(306),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(232)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(232),
    [sym_compiler_directive_decl] = STATE(232),
    [sym_fsi_directive_decl] = STATE(232),
    [sym_preproc_line] = STATE(232),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_GT_RBRACK] = ACTIONS(185),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_RBRACK] = ACTIONS(185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_RBRACE] = ACTIONS(185),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_to] = ACTIONS(187),
    [anon_sym_downto] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(233)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(233),
    [sym_compiler_directive_decl] = STATE(233),
    [sym_fsi_directive_decl] = STATE(233),
    [sym_preproc_line] = STATE(233),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(312),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(312),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(310),
    [anon_sym_downto] = ACTIONS(310),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(234)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(234),
    [sym_compiler_directive_decl] = STATE(234),
    [sym_fsi_directive_decl] = STATE(234),
    [sym_preproc_line] = STATE(234),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_GT_RBRACK] = ACTIONS(183),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_RBRACK] = ACTIONS(183),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_RBRACE] = ACTIONS(183),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_to] = ACTIONS(181),
    [anon_sym_downto] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(235)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(235),
    [sym_compiler_directive_decl] = STATE(235),
    [sym_fsi_directive_decl] = STATE(235),
    [sym_preproc_line] = STATE(235),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(199),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(199),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(199),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(193),
    [anon_sym_downto] = ACTIONS(193),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1612),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(236)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(236),
    [sym_compiler_directive_decl] = STATE(236),
    [sym_fsi_directive_decl] = STATE(236),
    [sym_preproc_line] = STATE(236),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym__list_elements_repeat1] = STATE(17157),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(1675),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(1681),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1709),
    [sym__dedent] = ACTIONS(1711),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(237)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(237),
    [sym_compiler_directive_decl] = STATE(237),
    [sym_fsi_directive_decl] = STATE(237),
    [sym_preproc_line] = STATE(237),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym__list_elements_repeat1] = STATE(17322),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(1675),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(1681),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1717),
    [sym__dedent] = ACTIONS(1719),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(238)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(238),
    [sym_compiler_directive_decl] = STATE(238),
    [sym_fsi_directive_decl] = STATE(238),
    [sym_preproc_line] = STATE(238),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_as] = ACTIONS(181),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_DOT_DOT] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(239)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(239),
    [sym_compiler_directive_decl] = STATE(239),
    [sym_fsi_directive_decl] = STATE(239),
    [sym_preproc_line] = STATE(239),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(282),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1723),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(1725),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(240)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(240),
    [sym_compiler_directive_decl] = STATE(240),
    [sym_fsi_directive_decl] = STATE(240),
    [sym_preproc_line] = STATE(240),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_as] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_DOT_DOT] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(241)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(241),
    [sym_compiler_directive_decl] = STATE(241),
    [sym_fsi_directive_decl] = STATE(241),
    [sym_preproc_line] = STATE(241),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(286),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1723),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(1725),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(242)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(242),
    [sym_compiler_directive_decl] = STATE(242),
    [sym_fsi_directive_decl] = STATE(242),
    [sym_preproc_line] = STATE(242),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_as] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_DOT_DOT] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(243)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(243),
    [sym_compiler_directive_decl] = STATE(243),
    [sym_fsi_directive_decl] = STATE(243),
    [sym_preproc_line] = STATE(243),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(310),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1723),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(1725),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(244)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(244),
    [sym_compiler_directive_decl] = STATE(244),
    [sym_fsi_directive_decl] = STATE(244),
    [sym_preproc_line] = STATE(244),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(193),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(199),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1579),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(245)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(245),
    [sym_compiler_directive_decl] = STATE(245),
    [sym_fsi_directive_decl] = STATE(245),
    [sym_preproc_line] = STATE(245),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(290),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(246)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(246),
    [sym_compiler_directive_decl] = STATE(246),
    [sym_fsi_directive_decl] = STATE(246),
    [sym_preproc_line] = STATE(246),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(294),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1723),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(1725),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(247)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(247),
    [sym_compiler_directive_decl] = STATE(247),
    [sym_fsi_directive_decl] = STATE(247),
    [sym_preproc_line] = STATE(247),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(298),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(300),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(248)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(248),
    [sym_compiler_directive_decl] = STATE(248),
    [sym_fsi_directive_decl] = STATE(248),
    [sym_preproc_line] = STATE(248),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(302),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1723),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(249)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(249),
    [sym_compiler_directive_decl] = STATE(249),
    [sym_fsi_directive_decl] = STATE(249),
    [sym_preproc_line] = STATE(249),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(302),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1723),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(250)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(240),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1683),
    [sym_infix_op] = STATE(1676),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(250),
    [sym_compiler_directive_decl] = STATE(250),
    [sym_fsi_directive_decl] = STATE(250),
    [sym_preproc_line] = STATE(250),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_sequential_expression_repeat1] = STATE(4690),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_COLON] = ACTIONS(1521),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(306),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_COMMA] = ACTIONS(1529),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1521),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(1539),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_COLON_GT] = ACTIONS(1551),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1551),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_DOT_LBRACK] = ACTIONS(1569),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_DOT_DOT] = ACTIONS(308),
    [anon_sym_LPAREN2] = ACTIONS(1577),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(1596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
    [sym__tyapp_open] = ACTIONS(1610),
  },
  [STATE(251)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(251),
    [sym_compiler_directive_decl] = STATE(251),
    [sym_fsi_directive_decl] = STATE(251),
    [sym_preproc_line] = STATE(251),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(282),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(252)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(252),
    [sym_compiler_directive_decl] = STATE(252),
    [sym_fsi_directive_decl] = STATE(252),
    [sym_preproc_line] = STATE(252),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(298),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_DOT_DOT2] = ACTIONS(300),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(253)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(253),
    [sym_compiler_directive_decl] = STATE(253),
    [sym_fsi_directive_decl] = STATE(253),
    [sym_preproc_line] = STATE(253),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(302),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(304),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(254)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(254),
    [sym_compiler_directive_decl] = STATE(254),
    [sym_fsi_directive_decl] = STATE(254),
    [sym_preproc_line] = STATE(254),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(302),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(304),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(255)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(255),
    [sym_compiler_directive_decl] = STATE(255),
    [sym_fsi_directive_decl] = STATE(255),
    [sym_preproc_line] = STATE(255),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(306),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_DOT_DOT2] = ACTIONS(308),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(256)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(256),
    [sym_compiler_directive_decl] = STATE(256),
    [sym_fsi_directive_decl] = STATE(256),
    [sym_preproc_line] = STATE(256),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(310),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(312),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(257)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(257),
    [sym_compiler_directive_decl] = STATE(257),
    [sym_fsi_directive_decl] = STATE(257),
    [sym_preproc_line] = STATE(257),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(258)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(258),
    [sym_compiler_directive_decl] = STATE(258),
    [sym_fsi_directive_decl] = STATE(258),
    [sym_preproc_line] = STATE(258),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(300),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(259)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(259),
    [sym_compiler_directive_decl] = STATE(259),
    [sym_fsi_directive_decl] = STATE(259),
    [sym_preproc_line] = STATE(259),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_DOT_DOT] = ACTIONS(159),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_DOT_DOT2] = ACTIONS(157),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(260)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym__object_members] = STATE(16598),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(260),
    [sym_compiler_directive_decl] = STATE(260),
    [sym_fsi_directive_decl] = STATE(260),
    [sym_preproc_line] = STATE(260),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(1525),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(1547),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1811),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(261)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(261),
    [sym_compiler_directive_decl] = STATE(261),
    [sym_fsi_directive_decl] = STATE(261),
    [sym_preproc_line] = STATE(261),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(262)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(262),
    [sym_compiler_directive_decl] = STATE(262),
    [sym_fsi_directive_decl] = STATE(262),
    [sym_preproc_line] = STATE(262),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(263)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(263),
    [sym_compiler_directive_decl] = STATE(263),
    [sym_fsi_directive_decl] = STATE(263),
    [sym_preproc_line] = STATE(263),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_DOT_DOT2] = ACTIONS(185),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(264)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(264),
    [sym_compiler_directive_decl] = STATE(264),
    [sym_fsi_directive_decl] = STATE(264),
    [sym_preproc_line] = STATE(264),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_DOT_DOT] = ACTIONS(181),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_DOT_DOT2] = ACTIONS(183),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(265)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(265),
    [sym_compiler_directive_decl] = STATE(265),
    [sym_fsi_directive_decl] = STATE(265),
    [sym_preproc_line] = STATE(265),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(193),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1844),
    [anon_sym_DOT_DOT2] = ACTIONS(199),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(266)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(266),
    [sym_compiler_directive_decl] = STATE(266),
    [sym_fsi_directive_decl] = STATE(266),
    [sym_preproc_line] = STATE(266),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(280),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(282),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(267)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(267),
    [sym_compiler_directive_decl] = STATE(267),
    [sym_fsi_directive_decl] = STATE(267),
    [sym_preproc_line] = STATE(267),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(1675),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(312),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(1847),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(268)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(268),
    [sym_compiler_directive_decl] = STATE(268),
    [sym_fsi_directive_decl] = STATE(268),
    [sym_preproc_line] = STATE(268),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_DOT_DOT] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(269)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(269),
    [sym_compiler_directive_decl] = STATE(269),
    [sym_fsi_directive_decl] = STATE(269),
    [sym_preproc_line] = STATE(269),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(270)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(270),
    [sym_compiler_directive_decl] = STATE(270),
    [sym_fsi_directive_decl] = STATE(270),
    [sym_preproc_line] = STATE(270),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(290),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_DOT_DOT2] = ACTIONS(292),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(271)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(271),
    [sym_compiler_directive_decl] = STATE(271),
    [sym_fsi_directive_decl] = STATE(271),
    [sym_preproc_line] = STATE(271),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(272)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(272),
    [sym_compiler_directive_decl] = STATE(272),
    [sym_fsi_directive_decl] = STATE(272),
    [sym_preproc_line] = STATE(272),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(1852),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(273)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(273),
    [sym_compiler_directive_decl] = STATE(273),
    [sym_fsi_directive_decl] = STATE(273),
    [sym_preproc_line] = STATE(273),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(1675),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(282),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(1855),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(274)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(274),
    [sym_compiler_directive_decl] = STATE(274),
    [sym_fsi_directive_decl] = STATE(274),
    [sym_preproc_line] = STATE(274),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(1847),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(275)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(275),
    [sym_compiler_directive_decl] = STATE(275),
    [sym_fsi_directive_decl] = STATE(275),
    [sym_preproc_line] = STATE(275),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(308),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(276)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(276),
    [sym_compiler_directive_decl] = STATE(276),
    [sym_fsi_directive_decl] = STATE(276),
    [sym_preproc_line] = STATE(276),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(1858),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(1860),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(277)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(277),
    [sym_compiler_directive_decl] = STATE(277),
    [sym_fsi_directive_decl] = STATE(277),
    [sym_preproc_line] = STATE(277),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(1675),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(288),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(1852),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(278)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(278),
    [sym_compiler_directive_decl] = STATE(278),
    [sym_fsi_directive_decl] = STATE(278),
    [sym_preproc_line] = STATE(278),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(294),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(296),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1862),
    [sym__dedent] = ACTIONS(1862),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(279)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(279),
    [sym_compiler_directive_decl] = STATE(279),
    [sym_fsi_directive_decl] = STATE(279),
    [sym_preproc_line] = STATE(279),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(199),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1864),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(280)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(280),
    [sym_compiler_directive_decl] = STATE(280),
    [sym_fsi_directive_decl] = STATE(280),
    [sym_preproc_line] = STATE(280),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_DOT_DOT] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(281)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(281),
    [sym_compiler_directive_decl] = STATE(281),
    [sym_fsi_directive_decl] = STATE(281),
    [sym_preproc_line] = STATE(281),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(294),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(296),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(282)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(282),
    [sym_compiler_directive_decl] = STATE(282),
    [sym_fsi_directive_decl] = STATE(282),
    [sym_preproc_line] = STATE(282),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(283)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(283),
    [sym_compiler_directive_decl] = STATE(283),
    [sym_fsi_directive_decl] = STATE(283),
    [sym_preproc_line] = STATE(283),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_DOT_DOT] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(284)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(280),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1582),
    [sym_infix_op] = STATE(1826),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(284),
    [sym_compiler_directive_decl] = STATE(284),
    [sym_fsi_directive_decl] = STATE(284),
    [sym_preproc_line] = STATE(284),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_sequential_expression_repeat1] = STATE(4914),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_COLON] = ACTIONS(1050),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1050),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1068),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_COLON_GT] = ACTIONS(1076),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1076),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_DOT_LBRACK] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(1121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__newline] = ACTIONS(1747),
    [sym__dedent] = ACTIONS(1855),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
    [sym__tyapp_open] = ACTIONS(1135),
  },
  [STATE(285)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(259),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1570),
    [sym_infix_op] = STATE(1577),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(285),
    [sym_compiler_directive_decl] = STATE(285),
    [sym_fsi_directive_decl] = STATE(285),
    [sym_preproc_line] = STATE(285),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4864),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_DOT_DOT] = ACTIONS(286),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_DOT_DOT2] = ACTIONS(288),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1749),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(286)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(286),
    [sym_compiler_directive_decl] = STATE(286),
    [sym_fsi_directive_decl] = STATE(286),
    [sym_preproc_line] = STATE(286),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(1959),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(1959),
  },
  [STATE(287)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(287),
    [sym_compiler_directive_decl] = STATE(287),
    [sym_fsi_directive_decl] = STATE(287),
    [sym_preproc_line] = STATE(287),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1965),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(199),
  },
  [STATE(288)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(288),
    [sym_compiler_directive_decl] = STATE(288),
    [sym_fsi_directive_decl] = STATE(288),
    [sym_preproc_line] = STATE(288),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(282),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(289)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(289),
    [sym_compiler_directive_decl] = STATE(289),
    [sym_fsi_directive_decl] = STATE(289),
    [sym_preproc_line] = STATE(289),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_DOT_DOT] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(290)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(290),
    [sym_compiler_directive_decl] = STATE(290),
    [sym_fsi_directive_decl] = STATE(290),
    [sym_preproc_line] = STATE(290),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(291)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(291),
    [sym_compiler_directive_decl] = STATE(291),
    [sym_fsi_directive_decl] = STATE(291),
    [sym_preproc_line] = STATE(291),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(292),
  },
  [STATE(292)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(292),
    [sym_compiler_directive_decl] = STATE(292),
    [sym_fsi_directive_decl] = STATE(292),
    [sym_preproc_line] = STATE(292),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(296),
  },
  [STATE(293)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(293),
    [sym_compiler_directive_decl] = STATE(293),
    [sym_fsi_directive_decl] = STATE(293),
    [sym_preproc_line] = STATE(293),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(300),
  },
  [STATE(294)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(294),
    [sym_compiler_directive_decl] = STATE(294),
    [sym_fsi_directive_decl] = STATE(294),
    [sym_preproc_line] = STATE(294),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(304),
  },
  [STATE(295)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(295),
    [sym_compiler_directive_decl] = STATE(295),
    [sym_fsi_directive_decl] = STATE(295),
    [sym_preproc_line] = STATE(295),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(304),
  },
  [STATE(296)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(296),
    [sym_compiler_directive_decl] = STATE(296),
    [sym_fsi_directive_decl] = STATE(296),
    [sym_preproc_line] = STATE(296),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_DOT_DOT] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(297)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(297),
    [sym_compiler_directive_decl] = STATE(297),
    [sym_fsi_directive_decl] = STATE(297),
    [sym_preproc_line] = STATE(297),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(308),
  },
  [STATE(298)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(298),
    [sym_compiler_directive_decl] = STATE(298),
    [sym_fsi_directive_decl] = STATE(298),
    [sym_preproc_line] = STATE(298),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(299)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(299),
    [sym_compiler_directive_decl] = STATE(299),
    [sym_fsi_directive_decl] = STATE(299),
    [sym_preproc_line] = STATE(299),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_as] = ACTIONS(181),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(300)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(300),
    [sym_compiler_directive_decl] = STATE(300),
    [sym_fsi_directive_decl] = STATE(300),
    [sym_preproc_line] = STATE(300),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(1852),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(301)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(301),
    [sym_compiler_directive_decl] = STATE(301),
    [sym_fsi_directive_decl] = STATE(301),
    [sym_preproc_line] = STATE(301),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(199),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2064),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(302)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(302),
    [sym_compiler_directive_decl] = STATE(302),
    [sym_fsi_directive_decl] = STATE(302),
    [sym_preproc_line] = STATE(302),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_DOT_DOT] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(303)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(303),
    [sym_compiler_directive_decl] = STATE(303),
    [sym_fsi_directive_decl] = STATE(303),
    [sym_preproc_line] = STATE(303),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(304)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(304),
    [sym_compiler_directive_decl] = STATE(304),
    [sym_fsi_directive_decl] = STATE(304),
    [sym_preproc_line] = STATE(304),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(305)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(305),
    [sym_compiler_directive_decl] = STATE(305),
    [sym_fsi_directive_decl] = STATE(305),
    [sym_preproc_line] = STATE(305),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(300),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(306)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(306),
    [sym_compiler_directive_decl] = STATE(306),
    [sym_fsi_directive_decl] = STATE(306),
    [sym_preproc_line] = STATE(306),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(307)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(307),
    [sym_compiler_directive_decl] = STATE(307),
    [sym_fsi_directive_decl] = STATE(307),
    [sym_preproc_line] = STATE(307),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(308)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(308),
    [sym_compiler_directive_decl] = STATE(308),
    [sym_fsi_directive_decl] = STATE(308),
    [sym_preproc_line] = STATE(308),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(308),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(309)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(309),
    [sym_compiler_directive_decl] = STATE(309),
    [sym_fsi_directive_decl] = STATE(309),
    [sym_preproc_line] = STATE(309),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(2067),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2069),
    [sym__dedent] = ACTIONS(2069),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(310)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(310),
    [sym_compiler_directive_decl] = STATE(310),
    [sym_fsi_directive_decl] = STATE(310),
    [sym_preproc_line] = STATE(310),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(193),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(1811),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(311)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(311),
    [sym_compiler_directive_decl] = STATE(311),
    [sym_fsi_directive_decl] = STATE(311),
    [sym_preproc_line] = STATE(311),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2073),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(2075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(312)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(312),
    [sym_compiler_directive_decl] = STATE(312),
    [sym_fsi_directive_decl] = STATE(312),
    [sym_preproc_line] = STATE(312),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(286),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2073),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(2075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(313)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(313),
    [sym_compiler_directive_decl] = STATE(313),
    [sym_fsi_directive_decl] = STATE(313),
    [sym_preproc_line] = STATE(313),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(290),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(314)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(314),
    [sym_compiler_directive_decl] = STATE(314),
    [sym_fsi_directive_decl] = STATE(314),
    [sym_preproc_line] = STATE(314),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(294),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2073),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(2075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(315)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(315),
    [sym_compiler_directive_decl] = STATE(315),
    [sym_fsi_directive_decl] = STATE(315),
    [sym_preproc_line] = STATE(315),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2079),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(316)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(316),
    [sym_compiler_directive_decl] = STATE(316),
    [sym_fsi_directive_decl] = STATE(316),
    [sym_preproc_line] = STATE(316),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(298),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(317)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(317),
    [sym_compiler_directive_decl] = STATE(317),
    [sym_fsi_directive_decl] = STATE(317),
    [sym_preproc_line] = STATE(317),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2101),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2101),
  },
  [STATE(318)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(318),
    [sym_compiler_directive_decl] = STATE(318),
    [sym_fsi_directive_decl] = STATE(318),
    [sym_preproc_line] = STATE(318),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(2103),
    [anon_sym_POUNDelse] = ACTIONS(2103),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(319)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(319),
    [sym_compiler_directive_decl] = STATE(319),
    [sym_fsi_directive_decl] = STATE(319),
    [sym_preproc_line] = STATE(319),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [anon_sym_POUNDendif] = ACTIONS(2107),
    [anon_sym_POUNDelse] = ACTIONS(2107),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(320)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(320),
    [sym_compiler_directive_decl] = STATE(320),
    [sym_fsi_directive_decl] = STATE(320),
    [sym_preproc_line] = STATE(320),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(302),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2073),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(321)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1110),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(318),
    [sym_literal_expression] = STATE(4100),
    [sym_srtp_call_expression] = STATE(4100),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3862),
    [sym_tuple_expression] = STATE(4100),
    [sym_brace_expression] = STATE(4100),
    [sym_anon_record_expression] = STATE(4100),
    [sym_prefixed_expression] = STATE(4100),
    [sym_typecast_expression] = STATE(4100),
    [sym_for_expression] = STATE(4100),
    [sym_while_expression] = STATE(4100),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4100),
    [sym_fun_expression] = STATE(4100),
    [sym_try_expression] = STATE(4100),
    [sym_match_expression] = STATE(4100),
    [sym_function_expression] = STATE(4100),
    [sym_mutate_expression] = STATE(4100),
    [sym_index_expression] = STATE(3862),
    [sym_typed_expression] = STATE(4100),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(4100),
    [sym_list_expression] = STATE(4100),
    [sym_array_expression] = STATE(4100),
    [sym_begin_end_expression] = STATE(4100),
    [sym_paren_expression] = STATE(4100),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4100),
    [sym_dot_expression] = STATE(3862),
    [sym_infix_expression] = STATE(4100),
    [sym_ce_expression] = STATE(4100),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4100),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(321),
    [sym_compiler_directive_decl] = STATE(321),
    [sym_fsi_directive_decl] = STATE(321),
    [sym_preproc_line] = STATE(321),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22155),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(21949),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1184),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(942),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(2111),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2143),
    [anon_sym_POUNDendif] = ACTIONS(2145),
    [anon_sym_POUNDelse] = ACTIONS(2147),
    [sym__newline] = ACTIONS(2149),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(322)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(322),
    [sym_compiler_directive_decl] = STATE(322),
    [sym_fsi_directive_decl] = STATE(322),
    [sym_preproc_line] = STATE(322),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(199),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2151),
    [anon_sym_DOT_DOT2] = ACTIONS(199),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(323)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(323),
    [sym_compiler_directive_decl] = STATE(323),
    [sym_fsi_directive_decl] = STATE(323),
    [sym_preproc_line] = STATE(323),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(282),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(282),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(324)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(324),
    [sym_compiler_directive_decl] = STATE(324),
    [sym_fsi_directive_decl] = STATE(324),
    [sym_preproc_line] = STATE(324),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_RBRACK] = ACTIONS(157),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_DOT_DOT2] = ACTIONS(157),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(325)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(325),
    [sym_compiler_directive_decl] = STATE(325),
    [sym_fsi_directive_decl] = STATE(325),
    [sym_preproc_line] = STATE(325),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(288),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(288),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(326)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(326),
    [sym_compiler_directive_decl] = STATE(326),
    [sym_fsi_directive_decl] = STATE(326),
    [sym_preproc_line] = STATE(326),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(292),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_DOT_DOT2] = ACTIONS(292),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(327)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(327),
    [sym_compiler_directive_decl] = STATE(327),
    [sym_fsi_directive_decl] = STATE(327),
    [sym_preproc_line] = STATE(327),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(296),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(296),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(328)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(328),
    [sym_compiler_directive_decl] = STATE(328),
    [sym_fsi_directive_decl] = STATE(328),
    [sym_preproc_line] = STATE(328),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(300),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_DOT_DOT2] = ACTIONS(300),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(329)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(329),
    [sym_compiler_directive_decl] = STATE(329),
    [sym_fsi_directive_decl] = STATE(329),
    [sym_preproc_line] = STATE(329),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(304),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(304),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(330)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(330),
    [sym_compiler_directive_decl] = STATE(330),
    [sym_fsi_directive_decl] = STATE(330),
    [sym_preproc_line] = STATE(330),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(304),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(304),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(331)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(331),
    [sym_compiler_directive_decl] = STATE(331),
    [sym_fsi_directive_decl] = STATE(331),
    [sym_preproc_line] = STATE(331),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_RBRACK] = ACTIONS(185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_DOT_DOT2] = ACTIONS(185),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(332)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(332),
    [sym_compiler_directive_decl] = STATE(332),
    [sym_fsi_directive_decl] = STATE(332),
    [sym_preproc_line] = STATE(332),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_DOT_DOT2] = ACTIONS(308),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(333)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(333),
    [sym_compiler_directive_decl] = STATE(333),
    [sym_fsi_directive_decl] = STATE(333),
    [sym_preproc_line] = STATE(333),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(312),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(334)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(334),
    [sym_compiler_directive_decl] = STATE(334),
    [sym_fsi_directive_decl] = STATE(334),
    [sym_preproc_line] = STATE(334),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_RBRACK] = ACTIONS(183),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_DOT_DOT2] = ACTIONS(183),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(335)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(335),
    [sym_compiler_directive_decl] = STATE(335),
    [sym_fsi_directive_decl] = STATE(335),
    [sym_preproc_line] = STATE(335),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(302),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2073),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(336)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(336),
    [sym_compiler_directive_decl] = STATE(336),
    [sym_fsi_directive_decl] = STATE(336),
    [sym_preproc_line] = STATE(336),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(306),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(337)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(337),
    [sym_compiler_directive_decl] = STATE(337),
    [sym_fsi_directive_decl] = STATE(337),
    [sym_preproc_line] = STATE(337),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_COLON] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_as] = ACTIONS(310),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(1773),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1783),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2073),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__newline] = ACTIONS(2075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(338)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(338),
    [sym_compiler_directive_decl] = STATE(338),
    [sym_fsi_directive_decl] = STATE(338),
    [sym_preproc_line] = STATE(338),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [anon_sym_POUNDelse] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(339)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(339),
    [sym_compiler_directive_decl] = STATE(339),
    [sym_fsi_directive_decl] = STATE(339),
    [sym_preproc_line] = STATE(339),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(1847),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(340)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(340),
    [sym_compiler_directive_decl] = STATE(340),
    [sym_fsi_directive_decl] = STATE(340),
    [sym_preproc_line] = STATE(340),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2156),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(341)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(341),
    [sym_compiler_directive_decl] = STATE(341),
    [sym_fsi_directive_decl] = STATE(341),
    [sym_preproc_line] = STATE(341),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [anon_sym_POUNDelse] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(342)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(342),
    [sym_compiler_directive_decl] = STATE(342),
    [sym_fsi_directive_decl] = STATE(342),
    [sym_preproc_line] = STATE(342),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [anon_sym_POUNDelse] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(343)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(343),
    [sym_compiler_directive_decl] = STATE(343),
    [sym_fsi_directive_decl] = STATE(343),
    [sym_preproc_line] = STATE(343),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(2166),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(2069),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(344)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(344),
    [sym_compiler_directive_decl] = STATE(344),
    [sym_fsi_directive_decl] = STATE(344),
    [sym_preproc_line] = STATE(344),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2168),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2170),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(2170),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(2184),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(2188),
    [sym__dedent] = ACTIONS(2168),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(345)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(345),
    [sym_compiler_directive_decl] = STATE(345),
    [sym_fsi_directive_decl] = STATE(345),
    [sym_preproc_line] = STATE(345),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2170),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(2170),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(1681),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(2188),
    [sym__dedent] = ACTIONS(2192),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(346)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(346),
    [sym_compiler_directive_decl] = STATE(346),
    [sym_fsi_directive_decl] = STATE(346),
    [sym_preproc_line] = STATE(346),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2244),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [anon_sym_POUNDelse] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(347)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(347),
    [sym_compiler_directive_decl] = STATE(347),
    [sym_fsi_directive_decl] = STATE(347),
    [sym_preproc_line] = STATE(347),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [anon_sym_POUNDelse] = ACTIONS(282),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(348)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(348),
    [sym_compiler_directive_decl] = STATE(348),
    [sym_fsi_directive_decl] = STATE(348),
    [sym_preproc_line] = STATE(348),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(2103),
    [anon_sym_POUNDelse] = ACTIONS(2103),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(349)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(349),
    [sym_compiler_directive_decl] = STATE(349),
    [sym_fsi_directive_decl] = STATE(349),
    [sym_preproc_line] = STATE(349),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [anon_sym_POUNDelse] = ACTIONS(288),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(350)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(350),
    [sym_compiler_directive_decl] = STATE(350),
    [sym_fsi_directive_decl] = STATE(350),
    [sym_preproc_line] = STATE(350),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(351)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(351),
    [sym_compiler_directive_decl] = STATE(351),
    [sym_fsi_directive_decl] = STATE(351),
    [sym_preproc_line] = STATE(351),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [anon_sym_POUNDelse] = ACTIONS(296),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(352)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(352),
    [sym_compiler_directive_decl] = STATE(352),
    [sym_fsi_directive_decl] = STATE(352),
    [sym_preproc_line] = STATE(352),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [anon_sym_POUNDelse] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(353)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(353),
    [sym_compiler_directive_decl] = STATE(353),
    [sym_fsi_directive_decl] = STATE(353),
    [sym_preproc_line] = STATE(353),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(354)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(354),
    [sym_compiler_directive_decl] = STATE(354),
    [sym_fsi_directive_decl] = STATE(354),
    [sym_preproc_line] = STATE(354),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [anon_sym_POUNDelse] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(355)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(355),
    [sym_compiler_directive_decl] = STATE(355),
    [sym_fsi_directive_decl] = STATE(355),
    [sym_preproc_line] = STATE(355),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2277),
    [anon_sym_downto] = ACTIONS(2277),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(356)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(356),
    [sym_compiler_directive_decl] = STATE(356),
    [sym_fsi_directive_decl] = STATE(356),
    [sym_preproc_line] = STATE(356),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [anon_sym_POUNDelse] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(357)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(357),
    [sym_compiler_directive_decl] = STATE(357),
    [sym_fsi_directive_decl] = STATE(357),
    [sym_preproc_line] = STATE(357),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [anon_sym_POUNDelse] = ACTIONS(312),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(358)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(358),
    [sym_compiler_directive_decl] = STATE(358),
    [sym_fsi_directive_decl] = STATE(358),
    [sym_preproc_line] = STATE(358),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(280),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(282),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(359)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(359),
    [sym_compiler_directive_decl] = STATE(359),
    [sym_fsi_directive_decl] = STATE(359),
    [sym_preproc_line] = STATE(359),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(286),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(360)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(360),
    [sym_compiler_directive_decl] = STATE(360),
    [sym_fsi_directive_decl] = STATE(360),
    [sym_preproc_line] = STATE(360),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(310),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(361)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(361),
    [sym_compiler_directive_decl] = STATE(361),
    [sym_fsi_directive_decl] = STATE(361),
    [sym_preproc_line] = STATE(361),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(193),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(199),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2375),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(362)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(362),
    [sym_compiler_directive_decl] = STATE(362),
    [sym_fsi_directive_decl] = STATE(362),
    [sym_preproc_line] = STATE(362),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(363)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(363),
    [sym_compiler_directive_decl] = STATE(363),
    [sym_fsi_directive_decl] = STATE(363),
    [sym_preproc_line] = STATE(363),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(294),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(364)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(364),
    [sym_compiler_directive_decl] = STATE(364),
    [sym_fsi_directive_decl] = STATE(364),
    [sym_preproc_line] = STATE(364),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(298),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(300),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(365)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(365),
    [sym_compiler_directive_decl] = STATE(365),
    [sym_fsi_directive_decl] = STATE(365),
    [sym_preproc_line] = STATE(365),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(302),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(366)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(366),
    [sym_compiler_directive_decl] = STATE(366),
    [sym_fsi_directive_decl] = STATE(366),
    [sym_preproc_line] = STATE(366),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(302),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(367)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(367),
    [sym_compiler_directive_decl] = STATE(367),
    [sym_fsi_directive_decl] = STATE(367),
    [sym_preproc_line] = STATE(367),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(306),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(308),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(368)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(368),
    [sym_compiler_directive_decl] = STATE(368),
    [sym_fsi_directive_decl] = STATE(368),
    [sym_preproc_line] = STATE(368),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(2107),
    [anon_sym_POUNDelse] = ACTIONS(2107),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(369)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(369),
    [sym_compiler_directive_decl] = STATE(369),
    [sym_fsi_directive_decl] = STATE(369),
    [sym_preproc_line] = STATE(369),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2378),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(2380),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(2382),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(370)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(370),
    [sym_compiler_directive_decl] = STATE(370),
    [sym_fsi_directive_decl] = STATE(370),
    [sym_preproc_line] = STATE(370),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2168),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2168),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2384),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(371)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(371),
    [sym_compiler_directive_decl] = STATE(371),
    [sym_fsi_directive_decl] = STATE(371),
    [sym_preproc_line] = STATE(371),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2192),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(372)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(372),
    [sym_compiler_directive_decl] = STATE(372),
    [sym_fsi_directive_decl] = STATE(372),
    [sym_preproc_line] = STATE(372),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(1860),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(373)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(373),
    [sym_compiler_directive_decl] = STATE(373),
    [sym_fsi_directive_decl] = STATE(373),
    [sym_preproc_line] = STATE(373),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2386),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2386),
  },
  [STATE(374)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1110),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(318),
    [sym_literal_expression] = STATE(4100),
    [sym_srtp_call_expression] = STATE(4100),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3862),
    [sym_tuple_expression] = STATE(4100),
    [sym_brace_expression] = STATE(4100),
    [sym_anon_record_expression] = STATE(4100),
    [sym_prefixed_expression] = STATE(4100),
    [sym_typecast_expression] = STATE(4100),
    [sym_for_expression] = STATE(4100),
    [sym_while_expression] = STATE(4100),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4100),
    [sym_fun_expression] = STATE(4100),
    [sym_try_expression] = STATE(4100),
    [sym_match_expression] = STATE(4100),
    [sym_function_expression] = STATE(4100),
    [sym_mutate_expression] = STATE(4100),
    [sym_index_expression] = STATE(3862),
    [sym_typed_expression] = STATE(4100),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(4100),
    [sym_list_expression] = STATE(4100),
    [sym_array_expression] = STATE(4100),
    [sym_begin_end_expression] = STATE(4100),
    [sym_paren_expression] = STATE(4100),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4100),
    [sym_dot_expression] = STATE(3862),
    [sym_infix_expression] = STATE(4100),
    [sym_ce_expression] = STATE(4100),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4100),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(374),
    [sym_compiler_directive_decl] = STATE(374),
    [sym_fsi_directive_decl] = STATE(374),
    [sym_preproc_line] = STATE(374),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20995),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(19746),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1171),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(989),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(2111),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2143),
    [anon_sym_POUNDendif] = ACTIONS(2388),
    [anon_sym_POUNDelse] = ACTIONS(2147),
    [sym__newline] = ACTIONS(2149),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(375)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(375),
    [sym_compiler_directive_decl] = STATE(375),
    [sym_fsi_directive_decl] = STATE(375),
    [sym_preproc_line] = STATE(375),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(2390),
    [anon_sym_POUNDelse] = ACTIONS(2390),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(376)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(376),
    [sym_compiler_directive_decl] = STATE(376),
    [sym_fsi_directive_decl] = STATE(376),
    [sym_preproc_line] = STATE(376),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2393),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(2393),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(2395),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(377)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(377),
    [sym_compiler_directive_decl] = STATE(377),
    [sym_fsi_directive_decl] = STATE(377),
    [sym_preproc_line] = STATE(377),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(183),
  },
  [STATE(378)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(341),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1623),
    [sym_infix_op] = STATE(1807),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(378),
    [sym_compiler_directive_decl] = STATE(378),
    [sym_fsi_directive_decl] = STATE(378),
    [sym_preproc_line] = STATE(378),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_sequential_expression_repeat1] = STATE(5132),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_COLON] = ACTIONS(2200),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_COMMA] = ACTIONS(2206),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2200),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(2212),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(2216),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2273),
    [anon_sym_DOT] = ACTIONS(2158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_COLON_GT] = ACTIONS(2224),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2224),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_DOT_LBRACK] = ACTIONS(2160),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_LPAREN2] = ACTIONS(2162),
    [anon_sym_PIPE_PIPE] = ACTIONS(2273),
    [anon_sym_AMP_AMP] = ACTIONS(2275),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(2164),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(2263),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(1257),
    [anon_sym_POUNDelse] = ACTIONS(1257),
    [sym__newline] = ACTIONS(2105),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
    [sym__tyapp_open] = ACTIONS(2154),
  },
  [STATE(379)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(289),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1651),
    [sym_infix_op] = STATE(1642),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(379),
    [sym_compiler_directive_decl] = STATE(379),
    [sym_fsi_directive_decl] = STATE(379),
    [sym_preproc_line] = STATE(379),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_sequential_expression_repeat1] = STATE(5127),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_COLON] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_COMMA] = ACTIONS(1980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(1990),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1992),
    [anon_sym_DOT] = ACTIONS(1994),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_COLON_GT] = ACTIONS(2002),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_DOT_LBRACK] = ACTIONS(2020),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_DOT_DOT] = ACTIONS(1850),
    [anon_sym_LPAREN2] = ACTIONS(2028),
    [anon_sym_PIPE_PIPE] = ACTIONS(1992),
    [anon_sym_AMP_AMP] = ACTIONS(2030),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(2046),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(2050),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__newline] = ACTIONS(2058),
    [sym__dedent] = ACTIONS(1855),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
    [sym__tyapp_open] = ACTIONS(2062),
  },
  [STATE(380)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(380),
    [sym_compiler_directive_decl] = STATE(380),
    [sym_fsi_directive_decl] = STATE(380),
    [sym_preproc_line] = STATE(380),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(282),
  },
  [STATE(381)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(381),
    [sym_compiler_directive_decl] = STATE(381),
    [sym_fsi_directive_decl] = STATE(381),
    [sym_preproc_line] = STATE(381),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2397),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(382)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(382),
    [sym_compiler_directive_decl] = STATE(382),
    [sym_fsi_directive_decl] = STATE(382),
    [sym_preproc_line] = STATE(382),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2399),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2399),
  },
  [STATE(383)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1110),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(318),
    [sym_literal_expression] = STATE(4100),
    [sym_srtp_call_expression] = STATE(4100),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3862),
    [sym_tuple_expression] = STATE(4100),
    [sym_brace_expression] = STATE(4100),
    [sym_anon_record_expression] = STATE(4100),
    [sym_prefixed_expression] = STATE(4100),
    [sym_typecast_expression] = STATE(4100),
    [sym_for_expression] = STATE(4100),
    [sym_while_expression] = STATE(4100),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4100),
    [sym_fun_expression] = STATE(4100),
    [sym_try_expression] = STATE(4100),
    [sym_match_expression] = STATE(4100),
    [sym_function_expression] = STATE(4100),
    [sym_mutate_expression] = STATE(4100),
    [sym_index_expression] = STATE(3862),
    [sym_typed_expression] = STATE(4100),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(4100),
    [sym_list_expression] = STATE(4100),
    [sym_array_expression] = STATE(4100),
    [sym_begin_end_expression] = STATE(4100),
    [sym_paren_expression] = STATE(4100),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4100),
    [sym_dot_expression] = STATE(3862),
    [sym_infix_expression] = STATE(4100),
    [sym_ce_expression] = STATE(4100),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4100),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(383),
    [sym_compiler_directive_decl] = STATE(383),
    [sym_fsi_directive_decl] = STATE(383),
    [sym_preproc_line] = STATE(383),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19117),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(20980),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1197),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(939),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(2111),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2143),
    [anon_sym_POUNDendif] = ACTIONS(2401),
    [anon_sym_POUNDelse] = ACTIONS(2147),
    [sym__newline] = ACTIONS(2149),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(384)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(384),
    [sym_compiler_directive_decl] = STATE(384),
    [sym_fsi_directive_decl] = STATE(384),
    [sym_preproc_line] = STATE(384),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [anon_sym_DOT_DOT2] = ACTIONS(157),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(385)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(385),
    [sym_compiler_directive_decl] = STATE(385),
    [sym_fsi_directive_decl] = STATE(385),
    [sym_preproc_line] = STATE(385),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_DOT_DOT2] = ACTIONS(185),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(386)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(386),
    [sym_compiler_directive_decl] = STATE(386),
    [sym_fsi_directive_decl] = STATE(386),
    [sym_preproc_line] = STATE(386),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [anon_sym_DOT_DOT2] = ACTIONS(183),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(387)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(387),
    [sym_compiler_directive_decl] = STATE(387),
    [sym_fsi_directive_decl] = STATE(387),
    [sym_preproc_line] = STATE(387),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2403),
    [anon_sym_DOT_DOT2] = ACTIONS(199),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(388)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(388),
    [sym_compiler_directive_decl] = STATE(388),
    [sym_fsi_directive_decl] = STATE(388),
    [sym_preproc_line] = STATE(388),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2170),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(2170),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(282),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(2188),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(389)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(389),
    [sym_compiler_directive_decl] = STATE(389),
    [sym_fsi_directive_decl] = STATE(389),
    [sym_preproc_line] = STATE(389),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2170),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(2170),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(288),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(2188),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(390)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(390),
    [sym_compiler_directive_decl] = STATE(390),
    [sym_fsi_directive_decl] = STATE(390),
    [sym_preproc_line] = STATE(390),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_DOT_DOT2] = ACTIONS(292),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(391)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(391),
    [sym_compiler_directive_decl] = STATE(391),
    [sym_fsi_directive_decl] = STATE(391),
    [sym_preproc_line] = STATE(391),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2170),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(2170),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(296),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(2188),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(392)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(392),
    [sym_compiler_directive_decl] = STATE(392),
    [sym_fsi_directive_decl] = STATE(392),
    [sym_preproc_line] = STATE(392),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [anon_sym_DOT_DOT2] = ACTIONS(300),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(393)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(393),
    [sym_compiler_directive_decl] = STATE(393),
    [sym_fsi_directive_decl] = STATE(393),
    [sym_preproc_line] = STATE(393),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(304),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(394)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(394),
    [sym_compiler_directive_decl] = STATE(394),
    [sym_fsi_directive_decl] = STATE(394),
    [sym_preproc_line] = STATE(394),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(304),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(395)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(395),
    [sym_compiler_directive_decl] = STATE(395),
    [sym_fsi_directive_decl] = STATE(395),
    [sym_preproc_line] = STATE(395),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [anon_sym_DOT_DOT2] = ACTIONS(308),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(396)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(384),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1617),
    [sym_infix_op] = STATE(1622),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(396),
    [sym_compiler_directive_decl] = STATE(396),
    [sym_fsi_directive_decl] = STATE(396),
    [sym_preproc_line] = STATE(396),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5107),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2190),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2170),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(2170),
    [anon_sym_AMP_AMP] = ACTIONS(2182),
    [anon_sym_DOT_DOT2] = ACTIONS(312),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(2188),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(397)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(397),
    [sym_compiler_directive_decl] = STATE(397),
    [sym_fsi_directive_decl] = STATE(397),
    [sym_preproc_line] = STATE(397),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2406),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(398)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(398),
    [sym_compiler_directive_decl] = STATE(398),
    [sym_fsi_directive_decl] = STATE(398),
    [sym_preproc_line] = STATE(398),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2408),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2408),
  },
  [STATE(399)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1110),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(318),
    [sym_literal_expression] = STATE(4100),
    [sym_srtp_call_expression] = STATE(4100),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3862),
    [sym_tuple_expression] = STATE(4100),
    [sym_brace_expression] = STATE(4100),
    [sym_anon_record_expression] = STATE(4100),
    [sym_prefixed_expression] = STATE(4100),
    [sym_typecast_expression] = STATE(4100),
    [sym_for_expression] = STATE(4100),
    [sym_while_expression] = STATE(4100),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4100),
    [sym_fun_expression] = STATE(4100),
    [sym_try_expression] = STATE(4100),
    [sym_match_expression] = STATE(4100),
    [sym_function_expression] = STATE(4100),
    [sym_mutate_expression] = STATE(4100),
    [sym_index_expression] = STATE(3862),
    [sym_typed_expression] = STATE(4100),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(4100),
    [sym_list_expression] = STATE(4100),
    [sym_array_expression] = STATE(4100),
    [sym_begin_end_expression] = STATE(4100),
    [sym_paren_expression] = STATE(4100),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4100),
    [sym_dot_expression] = STATE(3862),
    [sym_infix_expression] = STATE(4100),
    [sym_ce_expression] = STATE(4100),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4100),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(399),
    [sym_compiler_directive_decl] = STATE(399),
    [sym_fsi_directive_decl] = STATE(399),
    [sym_preproc_line] = STATE(399),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20995),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(21949),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1171),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(942),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(2111),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2143),
    [anon_sym_POUNDendif] = ACTIONS(2410),
    [anon_sym_POUNDelse] = ACTIONS(2147),
    [sym__newline] = ACTIONS(2149),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(400)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(400),
    [sym_compiler_directive_decl] = STATE(400),
    [sym_fsi_directive_decl] = STATE(400),
    [sym_preproc_line] = STATE(400),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2412),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(401)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(401),
    [sym_compiler_directive_decl] = STATE(401),
    [sym_fsi_directive_decl] = STATE(401),
    [sym_preproc_line] = STATE(401),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2414),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2414),
  },
  [STATE(402)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(402),
    [sym_compiler_directive_decl] = STATE(402),
    [sym_fsi_directive_decl] = STATE(402),
    [sym_preproc_line] = STATE(402),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_as] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(403)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(403),
    [sym_compiler_directive_decl] = STATE(403),
    [sym_fsi_directive_decl] = STATE(403),
    [sym_preproc_line] = STATE(403),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2416),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(404)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(404),
    [sym_compiler_directive_decl] = STATE(404),
    [sym_fsi_directive_decl] = STATE(404),
    [sym_preproc_line] = STATE(404),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2418),
  },
  [STATE(405)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(405),
    [sym_compiler_directive_decl] = STATE(405),
    [sym_fsi_directive_decl] = STATE(405),
    [sym_preproc_line] = STATE(405),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(406)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(406),
    [sym_compiler_directive_decl] = STATE(406),
    [sym_fsi_directive_decl] = STATE(406),
    [sym_preproc_line] = STATE(406),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2422),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2422),
  },
  [STATE(407)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(407),
    [sym_compiler_directive_decl] = STATE(407),
    [sym_fsi_directive_decl] = STATE(407),
    [sym_preproc_line] = STATE(407),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2424),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(408)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(408),
    [sym_compiler_directive_decl] = STATE(408),
    [sym_fsi_directive_decl] = STATE(408),
    [sym_preproc_line] = STATE(408),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2426),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2426),
  },
  [STATE(409)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(409),
    [sym_compiler_directive_decl] = STATE(409),
    [sym_fsi_directive_decl] = STATE(409),
    [sym_preproc_line] = STATE(409),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2428),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(410)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(410),
    [sym_compiler_directive_decl] = STATE(410),
    [sym_fsi_directive_decl] = STATE(410),
    [sym_preproc_line] = STATE(410),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2430),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2430),
  },
  [STATE(411)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(411),
    [sym_compiler_directive_decl] = STATE(411),
    [sym_fsi_directive_decl] = STATE(411),
    [sym_preproc_line] = STATE(411),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2432),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(412)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(412),
    [sym_compiler_directive_decl] = STATE(412),
    [sym_fsi_directive_decl] = STATE(412),
    [sym_preproc_line] = STATE(412),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2434),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2434),
  },
  [STATE(413)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(413),
    [sym_compiler_directive_decl] = STATE(413),
    [sym_fsi_directive_decl] = STATE(413),
    [sym_preproc_line] = STATE(413),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2436),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(414)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(414),
    [sym_compiler_directive_decl] = STATE(414),
    [sym_fsi_directive_decl] = STATE(414),
    [sym_preproc_line] = STATE(414),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2438),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2438),
  },
  [STATE(415)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(415),
    [sym_compiler_directive_decl] = STATE(415),
    [sym_fsi_directive_decl] = STATE(415),
    [sym_preproc_line] = STATE(415),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2440),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(416)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(416),
    [sym_compiler_directive_decl] = STATE(416),
    [sym_fsi_directive_decl] = STATE(416),
    [sym_preproc_line] = STATE(416),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2442),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2442),
  },
  [STATE(417)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(417),
    [sym_compiler_directive_decl] = STATE(417),
    [sym_fsi_directive_decl] = STATE(417),
    [sym_preproc_line] = STATE(417),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2444),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(418)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(418),
    [sym_compiler_directive_decl] = STATE(418),
    [sym_fsi_directive_decl] = STATE(418),
    [sym_preproc_line] = STATE(418),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2446),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2446),
  },
  [STATE(419)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(419),
    [sym_compiler_directive_decl] = STATE(419),
    [sym_fsi_directive_decl] = STATE(419),
    [sym_preproc_line] = STATE(419),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(420)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(420),
    [sym_compiler_directive_decl] = STATE(420),
    [sym_fsi_directive_decl] = STATE(420),
    [sym_preproc_line] = STATE(420),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2450),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2450),
  },
  [STATE(421)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(421),
    [sym_compiler_directive_decl] = STATE(421),
    [sym_fsi_directive_decl] = STATE(421),
    [sym_preproc_line] = STATE(421),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2452),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(422)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(422),
    [sym_compiler_directive_decl] = STATE(422),
    [sym_fsi_directive_decl] = STATE(422),
    [sym_preproc_line] = STATE(422),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2454),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2454),
  },
  [STATE(423)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(423),
    [sym_compiler_directive_decl] = STATE(423),
    [sym_fsi_directive_decl] = STATE(423),
    [sym_preproc_line] = STATE(423),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2456),
    [anon_sym_downto] = ACTIONS(2456),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(424)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(424),
    [sym_compiler_directive_decl] = STATE(424),
    [sym_fsi_directive_decl] = STATE(424),
    [sym_preproc_line] = STATE(424),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2458),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(425)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(425),
    [sym_compiler_directive_decl] = STATE(425),
    [sym_fsi_directive_decl] = STATE(425),
    [sym_preproc_line] = STATE(425),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2460),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2460),
  },
  [STATE(426)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(426),
    [sym_compiler_directive_decl] = STATE(426),
    [sym_fsi_directive_decl] = STATE(426),
    [sym_preproc_line] = STATE(426),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2462),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(427)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(427),
    [sym_compiler_directive_decl] = STATE(427),
    [sym_fsi_directive_decl] = STATE(427),
    [sym_preproc_line] = STATE(427),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2464),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2464),
  },
  [STATE(428)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(428),
    [sym_compiler_directive_decl] = STATE(428),
    [sym_fsi_directive_decl] = STATE(428),
    [sym_preproc_line] = STATE(428),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2466),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(429)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(429),
    [sym_compiler_directive_decl] = STATE(429),
    [sym_fsi_directive_decl] = STATE(429),
    [sym_preproc_line] = STATE(429),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2468),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2468),
  },
  [STATE(430)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(430),
    [sym_compiler_directive_decl] = STATE(430),
    [sym_fsi_directive_decl] = STATE(430),
    [sym_preproc_line] = STATE(430),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2470),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(431)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(431),
    [sym_compiler_directive_decl] = STATE(431),
    [sym_fsi_directive_decl] = STATE(431),
    [sym_preproc_line] = STATE(431),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2472),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2472),
  },
  [STATE(432)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(432),
    [sym_compiler_directive_decl] = STATE(432),
    [sym_fsi_directive_decl] = STATE(432),
    [sym_preproc_line] = STATE(432),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2474),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(433)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(433),
    [sym_compiler_directive_decl] = STATE(433),
    [sym_fsi_directive_decl] = STATE(433),
    [sym_preproc_line] = STATE(433),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2476),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2476),
  },
  [STATE(434)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(434),
    [sym_compiler_directive_decl] = STATE(434),
    [sym_fsi_directive_decl] = STATE(434),
    [sym_preproc_line] = STATE(434),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2478),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(435)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(435),
    [sym_compiler_directive_decl] = STATE(435),
    [sym_fsi_directive_decl] = STATE(435),
    [sym_preproc_line] = STATE(435),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2480),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2480),
  },
  [STATE(436)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(436),
    [sym_compiler_directive_decl] = STATE(436),
    [sym_fsi_directive_decl] = STATE(436),
    [sym_preproc_line] = STATE(436),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2482),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(437)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(437),
    [sym_compiler_directive_decl] = STATE(437),
    [sym_fsi_directive_decl] = STATE(437),
    [sym_preproc_line] = STATE(437),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2484),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2484),
  },
  [STATE(438)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(438),
    [sym_compiler_directive_decl] = STATE(438),
    [sym_fsi_directive_decl] = STATE(438),
    [sym_preproc_line] = STATE(438),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2486),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(439)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(439),
    [sym_compiler_directive_decl] = STATE(439),
    [sym_fsi_directive_decl] = STATE(439),
    [sym_preproc_line] = STATE(439),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2488),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2488),
  },
  [STATE(440)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(440),
    [sym_compiler_directive_decl] = STATE(440),
    [sym_fsi_directive_decl] = STATE(440),
    [sym_preproc_line] = STATE(440),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2490),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(441)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(441),
    [sym_compiler_directive_decl] = STATE(441),
    [sym_fsi_directive_decl] = STATE(441),
    [sym_preproc_line] = STATE(441),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2492),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2492),
  },
  [STATE(442)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(442),
    [sym_compiler_directive_decl] = STATE(442),
    [sym_fsi_directive_decl] = STATE(442),
    [sym_preproc_line] = STATE(442),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2494),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(443)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(443),
    [sym_compiler_directive_decl] = STATE(443),
    [sym_fsi_directive_decl] = STATE(443),
    [sym_preproc_line] = STATE(443),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2496),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2496),
  },
  [STATE(444)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(444),
    [sym_compiler_directive_decl] = STATE(444),
    [sym_fsi_directive_decl] = STATE(444),
    [sym_preproc_line] = STATE(444),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2498),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(445)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(445),
    [sym_compiler_directive_decl] = STATE(445),
    [sym_fsi_directive_decl] = STATE(445),
    [sym_preproc_line] = STATE(445),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2500),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2500),
  },
  [STATE(446)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(324),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1558),
    [sym_infix_op] = STATE(1563),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(446),
    [sym_compiler_directive_decl] = STATE(446),
    [sym_fsi_directive_decl] = STATE(446),
    [sym_preproc_line] = STATE(446),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(5006),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2502),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2081),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(2081),
    [anon_sym_AMP_AMP] = ACTIONS(2093),
    [anon_sym_DOT_DOT2] = ACTIONS(2095),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(2099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(447)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(447),
    [sym_compiler_directive_decl] = STATE(447),
    [sym_fsi_directive_decl] = STATE(447),
    [sym_preproc_line] = STATE(447),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2504),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2504),
  },
  [STATE(448)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(448),
    [sym_compiler_directive_decl] = STATE(448),
    [sym_fsi_directive_decl] = STATE(448),
    [sym_preproc_line] = STATE(448),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2506),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2506),
  },
  [STATE(449)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(449),
    [sym_compiler_directive_decl] = STATE(449),
    [sym_fsi_directive_decl] = STATE(449),
    [sym_preproc_line] = STATE(449),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2508),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2508),
  },
  [STATE(450)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(450),
    [sym_compiler_directive_decl] = STATE(450),
    [sym_fsi_directive_decl] = STATE(450),
    [sym_preproc_line] = STATE(450),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2510),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2510),
  },
  [STATE(451)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(451),
    [sym_compiler_directive_decl] = STATE(451),
    [sym_fsi_directive_decl] = STATE(451),
    [sym_preproc_line] = STATE(451),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2512),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2512),
  },
  [STATE(452)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(452),
    [sym_compiler_directive_decl] = STATE(452),
    [sym_fsi_directive_decl] = STATE(452),
    [sym_preproc_line] = STATE(452),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2514),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2514),
  },
  [STATE(453)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(453),
    [sym_compiler_directive_decl] = STATE(453),
    [sym_fsi_directive_decl] = STATE(453),
    [sym_preproc_line] = STATE(453),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2516),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2516),
  },
  [STATE(454)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(454),
    [sym_compiler_directive_decl] = STATE(454),
    [sym_fsi_directive_decl] = STATE(454),
    [sym_preproc_line] = STATE(454),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2518),
    [anon_sym_downto] = ACTIONS(2518),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(455)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(455),
    [sym_compiler_directive_decl] = STATE(455),
    [sym_fsi_directive_decl] = STATE(455),
    [sym_preproc_line] = STATE(455),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2520),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2520),
  },
  [STATE(456)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(456),
    [sym_compiler_directive_decl] = STATE(456),
    [sym_fsi_directive_decl] = STATE(456),
    [sym_preproc_line] = STATE(456),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2522),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2522),
  },
  [STATE(457)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(457),
    [sym_compiler_directive_decl] = STATE(457),
    [sym_fsi_directive_decl] = STATE(457),
    [sym_preproc_line] = STATE(457),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2524),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2524),
  },
  [STATE(458)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(458),
    [sym_compiler_directive_decl] = STATE(458),
    [sym_fsi_directive_decl] = STATE(458),
    [sym_preproc_line] = STATE(458),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2526),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2526),
  },
  [STATE(459)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(459),
    [sym_compiler_directive_decl] = STATE(459),
    [sym_fsi_directive_decl] = STATE(459),
    [sym_preproc_line] = STATE(459),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(2528),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(2528),
  },
  [STATE(460)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(460),
    [sym_compiler_directive_decl] = STATE(460),
    [sym_fsi_directive_decl] = STATE(460),
    [sym_preproc_line] = STATE(460),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2530),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(2380),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(2382),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(461)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(461),
    [sym_compiler_directive_decl] = STATE(461),
    [sym_fsi_directive_decl] = STATE(461),
    [sym_preproc_line] = STATE(461),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(157),
  },
  [STATE(462)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(462),
    [sym_compiler_directive_decl] = STATE(462),
    [sym_fsi_directive_decl] = STATE(462),
    [sym_preproc_line] = STATE(462),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(288),
  },
  [STATE(463)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(463),
    [sym_compiler_directive_decl] = STATE(463),
    [sym_fsi_directive_decl] = STATE(463),
    [sym_preproc_line] = STATE(463),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_DASH_GT] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_DOT_DOT] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(464)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(464),
    [sym_compiler_directive_decl] = STATE(464),
    [sym_fsi_directive_decl] = STATE(464),
    [sym_preproc_line] = STATE(464),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_DASH_GT] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_DOT_DOT] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(465)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(465),
    [sym_compiler_directive_decl] = STATE(465),
    [sym_fsi_directive_decl] = STATE(465),
    [sym_preproc_line] = STATE(465),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_DOT_DOT] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(466)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(466),
    [sym_compiler_directive_decl] = STATE(466),
    [sym_fsi_directive_decl] = STATE(466),
    [sym_preproc_line] = STATE(466),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(185),
  },
  [STATE(467)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(461),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1639),
    [sym_infix_op] = STATE(1630),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(467),
    [sym_compiler_directive_decl] = STATE(467),
    [sym_fsi_directive_decl] = STATE(467),
    [sym_preproc_line] = STATE(467),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_sequential_expression_repeat1] = STATE(5252),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_COLON] = ACTIONS(1873),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_COMMA] = ACTIONS(1879),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1873),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(1885),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(1889),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1901),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_DOT_LBRACK] = ACTIONS(1919),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_LPAREN2] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1929),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(1945),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(1949),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__newline] = ACTIONS(1957),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
    [sym__tyapp_open] = ACTIONS(1963),
    [sym__in] = ACTIONS(312),
  },
  [STATE(468)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(468),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1574),
    [sym_infix_op] = STATE(1608),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(468),
    [sym_compiler_directive_decl] = STATE(468),
    [sym_fsi_directive_decl] = STATE(468),
    [sym_preproc_line] = STATE(468),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_sequential_expression_repeat1] = STATE(4998),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_as] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1801),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1828),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1842),
  },
  [STATE(469)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(464),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1695),
    [sym_infix_op] = STATE(1686),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(469),
    [sym_compiler_directive_decl] = STATE(469),
    [sym_fsi_directive_decl] = STATE(469),
    [sym_preproc_line] = STATE(469),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5101),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(2285),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(2291),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2285),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2303),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_COLON_GT] = ACTIONS(2313),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2313),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(2380),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_DOT_DOT] = ACTIONS(2382),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(2303),
    [anon_sym_AMP_AMP] = ACTIONS(2341),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(2369),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(470)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(470),
    [sym_compiler_directive_decl] = STATE(470),
    [sym_fsi_directive_decl] = STATE(470),
    [sym_preproc_line] = STATE(470),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2534),
    [anon_sym_downto] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(471)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(471),
    [sym_compiler_directive_decl] = STATE(471),
    [sym_fsi_directive_decl] = STATE(471),
    [sym_preproc_line] = STATE(471),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2536),
    [anon_sym_downto] = ACTIONS(2536),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(472)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(472),
    [sym_compiler_directive_decl] = STATE(472),
    [sym_fsi_directive_decl] = STATE(472),
    [sym_preproc_line] = STATE(472),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2538),
    [anon_sym_downto] = ACTIONS(2538),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(473)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(473),
    [sym_compiler_directive_decl] = STATE(473),
    [sym_fsi_directive_decl] = STATE(473),
    [sym_preproc_line] = STATE(473),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2540),
    [anon_sym_downto] = ACTIONS(2540),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(474)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(474),
    [sym_compiler_directive_decl] = STATE(474),
    [sym_fsi_directive_decl] = STATE(474),
    [sym_preproc_line] = STATE(474),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2542),
    [anon_sym_downto] = ACTIONS(2542),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(475)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(475),
    [sym_compiler_directive_decl] = STATE(475),
    [sym_fsi_directive_decl] = STATE(475),
    [sym_preproc_line] = STATE(475),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2544),
    [anon_sym_downto] = ACTIONS(2544),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(476)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(476),
    [sym_compiler_directive_decl] = STATE(476),
    [sym_fsi_directive_decl] = STATE(476),
    [sym_preproc_line] = STATE(476),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2546),
    [anon_sym_downto] = ACTIONS(2546),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(477)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(477),
    [sym_compiler_directive_decl] = STATE(477),
    [sym_fsi_directive_decl] = STATE(477),
    [sym_preproc_line] = STATE(477),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2548),
    [anon_sym_downto] = ACTIONS(2548),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(478)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(478),
    [sym_compiler_directive_decl] = STATE(478),
    [sym_fsi_directive_decl] = STATE(478),
    [sym_preproc_line] = STATE(478),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2550),
    [anon_sym_downto] = ACTIONS(2550),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(479)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(479),
    [sym_compiler_directive_decl] = STATE(479),
    [sym_fsi_directive_decl] = STATE(479),
    [sym_preproc_line] = STATE(479),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2552),
    [anon_sym_downto] = ACTIONS(2552),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(480)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(480),
    [sym_compiler_directive_decl] = STATE(480),
    [sym_fsi_directive_decl] = STATE(480),
    [sym_preproc_line] = STATE(480),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2554),
    [anon_sym_downto] = ACTIONS(2554),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(481)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(481),
    [sym_compiler_directive_decl] = STATE(481),
    [sym_fsi_directive_decl] = STATE(481),
    [sym_preproc_line] = STATE(481),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2556),
    [anon_sym_downto] = ACTIONS(2556),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(482)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(482),
    [sym_compiler_directive_decl] = STATE(482),
    [sym_fsi_directive_decl] = STATE(482),
    [sym_preproc_line] = STATE(482),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2558),
    [anon_sym_downto] = ACTIONS(2558),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(483)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(483),
    [sym_compiler_directive_decl] = STATE(483),
    [sym_fsi_directive_decl] = STATE(483),
    [sym_preproc_line] = STATE(483),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2560),
    [anon_sym_downto] = ACTIONS(2560),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(484)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(484),
    [sym_compiler_directive_decl] = STATE(484),
    [sym_fsi_directive_decl] = STATE(484),
    [sym_preproc_line] = STATE(484),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2562),
    [anon_sym_downto] = ACTIONS(2562),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(485)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(485),
    [sym_compiler_directive_decl] = STATE(485),
    [sym_fsi_directive_decl] = STATE(485),
    [sym_preproc_line] = STATE(485),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2564),
    [anon_sym_downto] = ACTIONS(2564),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(486)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(486),
    [sym_compiler_directive_decl] = STATE(486),
    [sym_fsi_directive_decl] = STATE(486),
    [sym_preproc_line] = STATE(486),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2566),
    [anon_sym_downto] = ACTIONS(2566),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(487)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(487),
    [sym_compiler_directive_decl] = STATE(487),
    [sym_fsi_directive_decl] = STATE(487),
    [sym_preproc_line] = STATE(487),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2568),
    [anon_sym_downto] = ACTIONS(2568),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(488)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(488),
    [sym_compiler_directive_decl] = STATE(488),
    [sym_fsi_directive_decl] = STATE(488),
    [sym_preproc_line] = STATE(488),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2570),
    [anon_sym_downto] = ACTIONS(2570),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(489)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(489),
    [sym_compiler_directive_decl] = STATE(489),
    [sym_fsi_directive_decl] = STATE(489),
    [sym_preproc_line] = STATE(489),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2572),
    [anon_sym_downto] = ACTIONS(2572),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(490)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(490),
    [sym_compiler_directive_decl] = STATE(490),
    [sym_fsi_directive_decl] = STATE(490),
    [sym_preproc_line] = STATE(490),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2574),
    [anon_sym_downto] = ACTIONS(2574),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(491)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(491),
    [sym_compiler_directive_decl] = STATE(491),
    [sym_fsi_directive_decl] = STATE(491),
    [sym_preproc_line] = STATE(491),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2576),
    [anon_sym_downto] = ACTIONS(2576),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(492)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(492),
    [sym_compiler_directive_decl] = STATE(492),
    [sym_fsi_directive_decl] = STATE(492),
    [sym_preproc_line] = STATE(492),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2578),
    [anon_sym_downto] = ACTIONS(2578),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(493)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(493),
    [sym_compiler_directive_decl] = STATE(493),
    [sym_fsi_directive_decl] = STATE(493),
    [sym_preproc_line] = STATE(493),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_to] = ACTIONS(2580),
    [anon_sym_downto] = ACTIONS(2580),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(494)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1110),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(318),
    [sym_literal_expression] = STATE(4100),
    [sym_srtp_call_expression] = STATE(4100),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3862),
    [sym_tuple_expression] = STATE(4100),
    [sym_brace_expression] = STATE(4100),
    [sym_anon_record_expression] = STATE(4100),
    [sym_prefixed_expression] = STATE(4100),
    [sym_typecast_expression] = STATE(4100),
    [sym_for_expression] = STATE(4100),
    [sym_while_expression] = STATE(4100),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4100),
    [sym_fun_expression] = STATE(4100),
    [sym_try_expression] = STATE(4100),
    [sym_match_expression] = STATE(4100),
    [sym_function_expression] = STATE(4100),
    [sym_mutate_expression] = STATE(4100),
    [sym_index_expression] = STATE(3862),
    [sym_typed_expression] = STATE(4100),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(4100),
    [sym_list_expression] = STATE(4100),
    [sym_array_expression] = STATE(4100),
    [sym_begin_end_expression] = STATE(4100),
    [sym_paren_expression] = STATE(4100),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4100),
    [sym_dot_expression] = STATE(3862),
    [sym_infix_expression] = STATE(4100),
    [sym_ce_expression] = STATE(4100),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4100),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(494),
    [sym_compiler_directive_decl] = STATE(494),
    [sym_fsi_directive_decl] = STATE(494),
    [sym_preproc_line] = STATE(494),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20995),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(20980),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1171),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(939),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(2111),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2143),
    [anon_sym_POUNDendif] = ACTIONS(2582),
    [anon_sym_POUNDelse] = ACTIONS(2147),
    [sym__newline] = ACTIONS(2149),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(495)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(495),
    [sym_compiler_directive_decl] = STATE(495),
    [sym_fsi_directive_decl] = STATE(495),
    [sym_preproc_line] = STATE(495),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(2676),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(496)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(496),
    [sym_compiler_directive_decl] = STATE(496),
    [sym_fsi_directive_decl] = STATE(496),
    [sym_preproc_line] = STATE(496),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(2774),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(497)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(497),
    [sym_compiler_directive_decl] = STATE(497),
    [sym_fsi_directive_decl] = STATE(497),
    [sym_preproc_line] = STATE(497),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(199),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2840),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(498)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(498),
    [sym_compiler_directive_decl] = STATE(498),
    [sym_fsi_directive_decl] = STATE(498),
    [sym_preproc_line] = STATE(498),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(282),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(499)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(499),
    [sym_compiler_directive_decl] = STATE(499),
    [sym_fsi_directive_decl] = STATE(499),
    [sym_preproc_line] = STATE(499),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(500)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(500),
    [sym_compiler_directive_decl] = STATE(500),
    [sym_fsi_directive_decl] = STATE(500),
    [sym_preproc_line] = STATE(500),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(501)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(501),
    [sym_compiler_directive_decl] = STATE(501),
    [sym_fsi_directive_decl] = STATE(501),
    [sym_preproc_line] = STATE(501),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(502)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(502),
    [sym_compiler_directive_decl] = STATE(502),
    [sym_fsi_directive_decl] = STATE(502),
    [sym_preproc_line] = STATE(502),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(300),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(503)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(503),
    [sym_compiler_directive_decl] = STATE(503),
    [sym_fsi_directive_decl] = STATE(503),
    [sym_preproc_line] = STATE(503),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(504)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(504),
    [sym_compiler_directive_decl] = STATE(504),
    [sym_fsi_directive_decl] = STATE(504),
    [sym_preproc_line] = STATE(504),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(304),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(505)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(505),
    [sym_compiler_directive_decl] = STATE(505),
    [sym_fsi_directive_decl] = STATE(505),
    [sym_preproc_line] = STATE(505),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(308),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(506)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(506),
    [sym_compiler_directive_decl] = STATE(506),
    [sym_fsi_directive_decl] = STATE(506),
    [sym_preproc_line] = STATE(506),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(507)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(507),
    [sym_compiler_directive_decl] = STATE(507),
    [sym_fsi_directive_decl] = STATE(507),
    [sym_preproc_line] = STATE(507),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_DOT_DOT] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(508)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(508),
    [sym_compiler_directive_decl] = STATE(508),
    [sym_fsi_directive_decl] = STATE(508),
    [sym_preproc_line] = STATE(508),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(2879),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(199),
    [sym__quoted_close] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(509)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(509),
    [sym_compiler_directive_decl] = STATE(509),
    [sym_fsi_directive_decl] = STATE(509),
    [sym_preproc_line] = STATE(509),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(2914),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(2974),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(510)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(510),
    [sym_compiler_directive_decl] = STATE(510),
    [sym_fsi_directive_decl] = STATE(510),
    [sym_preproc_line] = STATE(510),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(511)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(511),
    [sym_compiler_directive_decl] = STATE(511),
    [sym_fsi_directive_decl] = STATE(511),
    [sym_preproc_line] = STATE(511),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__quoted_close] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(512)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(512),
    [sym_compiler_directive_decl] = STATE(512),
    [sym_fsi_directive_decl] = STATE(512),
    [sym_preproc_line] = STATE(512),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(513)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(513),
    [sym_compiler_directive_decl] = STATE(513),
    [sym_fsi_directive_decl] = STATE(513),
    [sym_preproc_line] = STATE(513),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(292),
    [sym__quoted_close] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(514)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(514),
    [sym_compiler_directive_decl] = STATE(514),
    [sym_fsi_directive_decl] = STATE(514),
    [sym_preproc_line] = STATE(514),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(515)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(515),
    [sym_compiler_directive_decl] = STATE(515),
    [sym_fsi_directive_decl] = STATE(515),
    [sym_preproc_line] = STATE(515),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(300),
    [sym__quoted_close] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(516)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(516),
    [sym_compiler_directive_decl] = STATE(516),
    [sym_fsi_directive_decl] = STATE(516),
    [sym_preproc_line] = STATE(516),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(304),
    [sym__quoted_close] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(517)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(517),
    [sym_compiler_directive_decl] = STATE(517),
    [sym_fsi_directive_decl] = STATE(517),
    [sym_preproc_line] = STATE(517),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(304),
    [sym__quoted_close] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(518)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(518),
    [sym_compiler_directive_decl] = STATE(518),
    [sym_fsi_directive_decl] = STATE(518),
    [sym_preproc_line] = STATE(518),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__quoted_close] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(519)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(519),
    [sym_compiler_directive_decl] = STATE(519),
    [sym_fsi_directive_decl] = STATE(519),
    [sym_preproc_line] = STATE(519),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(2980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(2980),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(520)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(520),
    [sym_compiler_directive_decl] = STATE(520),
    [sym_fsi_directive_decl] = STATE(520),
    [sym_preproc_line] = STATE(520),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_GT] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(521)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(521),
    [sym_compiler_directive_decl] = STATE(521),
    [sym_fsi_directive_decl] = STATE(521),
    [sym_preproc_line] = STATE(521),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(308),
    [sym__quoted_close] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(522)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(522),
    [sym_compiler_directive_decl] = STATE(522),
    [sym_fsi_directive_decl] = STATE(522),
    [sym_preproc_line] = STATE(522),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(280),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3008),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(3008),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(3074),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(523)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(523),
    [sym_compiler_directive_decl] = STATE(523),
    [sym_fsi_directive_decl] = STATE(523),
    [sym_preproc_line] = STATE(523),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_GT] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(524)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(524),
    [sym_compiler_directive_decl] = STATE(524),
    [sym_fsi_directive_decl] = STATE(524),
    [sym_preproc_line] = STATE(524),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(286),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3008),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(3008),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(3074),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(525)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(525),
    [sym_compiler_directive_decl] = STATE(525),
    [sym_fsi_directive_decl] = STATE(525),
    [sym_preproc_line] = STATE(525),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2393),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3078),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(526)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(526),
    [sym_compiler_directive_decl] = STATE(526),
    [sym_fsi_directive_decl] = STATE(526),
    [sym_preproc_line] = STATE(526),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3080),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(527)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(527),
    [sym_compiler_directive_decl] = STATE(527),
    [sym_fsi_directive_decl] = STATE(527),
    [sym_preproc_line] = STATE(527),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(3176),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(528)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(528),
    [sym_compiler_directive_decl] = STATE(528),
    [sym_fsi_directive_decl] = STATE(528),
    [sym_preproc_line] = STATE(528),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(3176),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(529)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(529),
    [sym_compiler_directive_decl] = STATE(529),
    [sym_fsi_directive_decl] = STATE(529),
    [sym_preproc_line] = STATE(529),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3200),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(530)] = {
    [sym__preproc_toplevel_module] = STATE(17491),
    [sym__module_elem] = STATE(17493),
    [sym__attribute_expression] = STATE(17962),
    [sym_module_abbrev] = STATE(18004),
    [sym_module_defn] = STATE(18004),
    [sym_import_decl] = STATE(18004),
    [sym_attributes] = STATE(1228),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(17962),
    [sym_do] = STATE(18063),
    [sym_function_or_value_defn] = STATE(1169),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(375),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(18004),
    [sym_type_definition] = STATE(18004),
    [sym_extern_binding] = STATE(18004),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(530),
    [sym_compiler_directive_decl] = STATE(530),
    [sym_fsi_directive_decl] = STATE(530),
    [sym_preproc_line] = STATE(530),
    [sym_preproc_if] = STATE(18004),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22155),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1184),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_module] = ACTIONS(3202),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_open] = ACTIONS(3204),
    [anon_sym_type] = ACTIONS(3206),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(3208),
    [anon_sym_let] = ACTIONS(3210),
    [anon_sym_let_BANG] = ACTIONS(3212),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(3218),
    [anon_sym_extern] = ACTIONS(3220),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(3226),
    [anon_sym_POUNDload] = ACTIONS(3226),
    [anon_sym_POUNDtime] = ACTIONS(3226),
    [anon_sym_POUNDI] = ACTIONS(3226),
    [anon_sym_POUNDhelp] = ACTIONS(3226),
    [anon_sym_POUNDquit] = ACTIONS(3226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3228),
    [anon_sym_POUNDendif] = ACTIONS(3230),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(531)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(531),
    [sym_compiler_directive_decl] = STATE(531),
    [sym_fsi_directive_decl] = STATE(531),
    [sym_preproc_line] = STATE(531),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(2103),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(532)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(532),
    [sym_compiler_directive_decl] = STATE(532),
    [sym_fsi_directive_decl] = STATE(532),
    [sym_preproc_line] = STATE(532),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(533)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(533),
    [sym_compiler_directive_decl] = STATE(533),
    [sym_fsi_directive_decl] = STATE(533),
    [sym_preproc_line] = STATE(533),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(534)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(534),
    [sym_compiler_directive_decl] = STATE(534),
    [sym_fsi_directive_decl] = STATE(534),
    [sym_preproc_line] = STATE(534),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(310),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3008),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(3008),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(3074),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(535)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(535),
    [sym_compiler_directive_decl] = STATE(535),
    [sym_fsi_directive_decl] = STATE(535),
    [sym_preproc_line] = STATE(535),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(2107),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(536)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(536),
    [sym_compiler_directive_decl] = STATE(536),
    [sym_fsi_directive_decl] = STATE(536),
    [sym_preproc_line] = STATE(536),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(537)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(537),
    [sym_compiler_directive_decl] = STATE(537),
    [sym_fsi_directive_decl] = STATE(537),
    [sym_preproc_line] = STATE(537),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(2103),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(538)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(538),
    [sym_compiler_directive_decl] = STATE(538),
    [sym_fsi_directive_decl] = STATE(538),
    [sym_preproc_line] = STATE(538),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [anon_sym_POUNDendif] = ACTIONS(2107),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(539)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(539),
    [sym_compiler_directive_decl] = STATE(539),
    [sym_fsi_directive_decl] = STATE(539),
    [sym_preproc_line] = STATE(539),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(193),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3334),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(540)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(540),
    [sym_compiler_directive_decl] = STATE(540),
    [sym_fsi_directive_decl] = STATE(540),
    [sym_preproc_line] = STATE(540),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__quoted_close] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(541)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(541),
    [sym_compiler_directive_decl] = STATE(541),
    [sym_fsi_directive_decl] = STATE(541),
    [sym_preproc_line] = STATE(541),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(542)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(542),
    [sym_compiler_directive_decl] = STATE(542),
    [sym_fsi_directive_decl] = STATE(542),
    [sym_preproc_line] = STATE(542),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(294),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3008),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(3008),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(3074),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(543)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(543),
    [sym_compiler_directive_decl] = STATE(543),
    [sym_fsi_directive_decl] = STATE(543),
    [sym_preproc_line] = STATE(543),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(298),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(544)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(544),
    [sym_compiler_directive_decl] = STATE(544),
    [sym_fsi_directive_decl] = STATE(544),
    [sym_preproc_line] = STATE(544),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(545)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(545),
    [sym_compiler_directive_decl] = STATE(545),
    [sym_fsi_directive_decl] = STATE(545),
    [sym_preproc_line] = STATE(545),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(546)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(546),
    [sym_compiler_directive_decl] = STATE(546),
    [sym_fsi_directive_decl] = STATE(546),
    [sym_preproc_line] = STATE(546),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(306),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(547)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(547),
    [sym_compiler_directive_decl] = STATE(547),
    [sym_fsi_directive_decl] = STATE(547),
    [sym_preproc_line] = STATE(547),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(3337),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(548)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(548),
    [sym_compiler_directive_decl] = STATE(548),
    [sym_fsi_directive_decl] = STATE(548),
    [sym_preproc_line] = STATE(548),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_DASH_GT] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(549)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(549),
    [sym_compiler_directive_decl] = STATE(549),
    [sym_fsi_directive_decl] = STATE(549),
    [sym_preproc_line] = STATE(549),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(2980),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(2980),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(550)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(550),
    [sym_compiler_directive_decl] = STATE(550),
    [sym_fsi_directive_decl] = STATE(550),
    [sym_preproc_line] = STATE(550),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3339),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(199),
    [sym__untyped_quoted_close] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(551)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(551),
    [sym_compiler_directive_decl] = STATE(551),
    [sym_fsi_directive_decl] = STATE(551),
    [sym_preproc_line] = STATE(551),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(552)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(552),
    [sym_compiler_directive_decl] = STATE(552),
    [sym_fsi_directive_decl] = STATE(552),
    [sym_preproc_line] = STATE(552),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__untyped_quoted_close] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(553)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(553),
    [sym_compiler_directive_decl] = STATE(553),
    [sym_fsi_directive_decl] = STATE(553),
    [sym_preproc_line] = STATE(553),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [anon_sym_POUNDendif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(554)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(554),
    [sym_compiler_directive_decl] = STATE(554),
    [sym_fsi_directive_decl] = STATE(554),
    [sym_preproc_line] = STATE(554),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(555)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(555),
    [sym_compiler_directive_decl] = STATE(555),
    [sym_fsi_directive_decl] = STATE(555),
    [sym_preproc_line] = STATE(555),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(292),
    [sym__untyped_quoted_close] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(556)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(556),
    [sym_compiler_directive_decl] = STATE(556),
    [sym_fsi_directive_decl] = STATE(556),
    [sym_preproc_line] = STATE(556),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(280),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(557)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(557),
    [sym_compiler_directive_decl] = STATE(557),
    [sym_fsi_directive_decl] = STATE(557),
    [sym_preproc_line] = STATE(557),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_DASH_GT] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(558)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(558),
    [sym_compiler_directive_decl] = STATE(558),
    [sym_fsi_directive_decl] = STATE(558),
    [sym_preproc_line] = STATE(558),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(286),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(559)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(559),
    [sym_compiler_directive_decl] = STATE(559),
    [sym_fsi_directive_decl] = STATE(559),
    [sym_preproc_line] = STATE(559),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(560)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(560),
    [sym_compiler_directive_decl] = STATE(560),
    [sym_fsi_directive_decl] = STATE(560),
    [sym_preproc_line] = STATE(560),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(300),
    [sym__untyped_quoted_close] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(561)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(561),
    [sym_compiler_directive_decl] = STATE(561),
    [sym_fsi_directive_decl] = STATE(561),
    [sym_preproc_line] = STATE(561),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(304),
    [sym__untyped_quoted_close] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(562)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(562),
    [sym_compiler_directive_decl] = STATE(562),
    [sym_fsi_directive_decl] = STATE(562),
    [sym_preproc_line] = STATE(562),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(304),
    [sym__untyped_quoted_close] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(563)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(563),
    [sym_compiler_directive_decl] = STATE(563),
    [sym_fsi_directive_decl] = STATE(563),
    [sym_preproc_line] = STATE(563),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__untyped_quoted_close] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(564)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(564),
    [sym_compiler_directive_decl] = STATE(564),
    [sym_fsi_directive_decl] = STATE(564),
    [sym_preproc_line] = STATE(564),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(565)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(565),
    [sym_compiler_directive_decl] = STATE(565),
    [sym_fsi_directive_decl] = STATE(565),
    [sym_preproc_line] = STATE(565),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(310),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(566)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(566),
    [sym_compiler_directive_decl] = STATE(566),
    [sym_fsi_directive_decl] = STATE(566),
    [sym_preproc_line] = STATE(566),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(308),
    [sym__untyped_quoted_close] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(567)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(567),
    [sym_compiler_directive_decl] = STATE(567),
    [sym_fsi_directive_decl] = STATE(567),
    [sym_preproc_line] = STATE(567),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(193),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3366),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(568)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(568),
    [sym_compiler_directive_decl] = STATE(568),
    [sym_fsi_directive_decl] = STATE(568),
    [sym_preproc_line] = STATE(568),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(282),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(569)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(569),
    [sym_compiler_directive_decl] = STATE(569),
    [sym_fsi_directive_decl] = STATE(569),
    [sym_preproc_line] = STATE(569),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [anon_sym_POUNDendif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(570)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(570),
    [sym_compiler_directive_decl] = STATE(570),
    [sym_fsi_directive_decl] = STATE(570),
    [sym_preproc_line] = STATE(570),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(288),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(571)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(571),
    [sym_compiler_directive_decl] = STATE(571),
    [sym_fsi_directive_decl] = STATE(571),
    [sym_preproc_line] = STATE(571),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(572)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(572),
    [sym_compiler_directive_decl] = STATE(572),
    [sym_fsi_directive_decl] = STATE(572),
    [sym_preproc_line] = STATE(572),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(294),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(573)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(573),
    [sym_compiler_directive_decl] = STATE(573),
    [sym_fsi_directive_decl] = STATE(573),
    [sym_preproc_line] = STATE(573),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(298),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(574)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(574),
    [sym_compiler_directive_decl] = STATE(574),
    [sym_fsi_directive_decl] = STATE(574),
    [sym_preproc_line] = STATE(574),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(302),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(575)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(575),
    [sym_compiler_directive_decl] = STATE(575),
    [sym_fsi_directive_decl] = STATE(575),
    [sym_preproc_line] = STATE(575),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(302),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(576)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(576),
    [sym_compiler_directive_decl] = STATE(576),
    [sym_fsi_directive_decl] = STATE(576),
    [sym_preproc_line] = STATE(576),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [anon_sym_POUNDendif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(577)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(577),
    [sym_compiler_directive_decl] = STATE(577),
    [sym_fsi_directive_decl] = STATE(577),
    [sym_preproc_line] = STATE(577),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(306),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(578)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(578),
    [sym_compiler_directive_decl] = STATE(578),
    [sym_fsi_directive_decl] = STATE(578),
    [sym_preproc_line] = STATE(578),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(312),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(579)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(579),
    [sym_compiler_directive_decl] = STATE(579),
    [sym_fsi_directive_decl] = STATE(579),
    [sym_preproc_line] = STATE(579),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(3369),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(580)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(580),
    [sym_compiler_directive_decl] = STATE(580),
    [sym_fsi_directive_decl] = STATE(580),
    [sym_preproc_line] = STATE(580),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(199),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(581)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(581),
    [sym_compiler_directive_decl] = STATE(581),
    [sym_fsi_directive_decl] = STATE(581),
    [sym_preproc_line] = STATE(581),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3374),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(582)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(582),
    [sym_compiler_directive_decl] = STATE(582),
    [sym_fsi_directive_decl] = STATE(582),
    [sym_preproc_line] = STATE(582),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3378),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(583)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(583),
    [sym_compiler_directive_decl] = STATE(583),
    [sym_fsi_directive_decl] = STATE(583),
    [sym_preproc_line] = STATE(583),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3378),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(584)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(584),
    [sym_compiler_directive_decl] = STATE(584),
    [sym_fsi_directive_decl] = STATE(584),
    [sym_preproc_line] = STATE(584),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(3380),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(585)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(585),
    [sym_compiler_directive_decl] = STATE(585),
    [sym_fsi_directive_decl] = STATE(585),
    [sym_preproc_line] = STATE(585),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(3380),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(586)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(586),
    [sym_compiler_directive_decl] = STATE(586),
    [sym_fsi_directive_decl] = STATE(586),
    [sym_preproc_line] = STATE(586),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3382),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(587)] = {
    [sym__preproc_toplevel_module] = STATE(17457),
    [sym__module_elem] = STATE(17512),
    [sym__attribute_expression] = STATE(17962),
    [sym_module_abbrev] = STATE(18004),
    [sym_module_defn] = STATE(18004),
    [sym_import_decl] = STATE(18004),
    [sym_attributes] = STATE(1228),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(17962),
    [sym_do] = STATE(18063),
    [sym_function_or_value_defn] = STATE(1169),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(375),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(18004),
    [sym_type_definition] = STATE(18004),
    [sym_extern_binding] = STATE(18004),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(587),
    [sym_compiler_directive_decl] = STATE(587),
    [sym_fsi_directive_decl] = STATE(587),
    [sym_preproc_line] = STATE(587),
    [sym_preproc_if] = STATE(18004),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19553),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1201),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_module] = ACTIONS(3202),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_open] = ACTIONS(3204),
    [anon_sym_type] = ACTIONS(3206),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(3208),
    [anon_sym_let] = ACTIONS(3210),
    [anon_sym_let_BANG] = ACTIONS(3212),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(3218),
    [anon_sym_extern] = ACTIONS(3220),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(3226),
    [anon_sym_POUNDload] = ACTIONS(3226),
    [anon_sym_POUNDtime] = ACTIONS(3226),
    [anon_sym_POUNDI] = ACTIONS(3226),
    [anon_sym_POUNDhelp] = ACTIONS(3226),
    [anon_sym_POUNDquit] = ACTIONS(3226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3228),
    [anon_sym_POUNDendif] = ACTIONS(3384),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(588)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(588),
    [sym_compiler_directive_decl] = STATE(588),
    [sym_fsi_directive_decl] = STATE(588),
    [sym_preproc_line] = STATE(588),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(589)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(589),
    [sym_compiler_directive_decl] = STATE(589),
    [sym_fsi_directive_decl] = STATE(589),
    [sym_preproc_line] = STATE(589),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__untyped_quoted_close] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(590)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(590),
    [sym_compiler_directive_decl] = STATE(590),
    [sym_fsi_directive_decl] = STATE(590),
    [sym_preproc_line] = STATE(590),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(591)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(591),
    [sym_compiler_directive_decl] = STATE(591),
    [sym_fsi_directive_decl] = STATE(591),
    [sym_preproc_line] = STATE(591),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3386),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(592)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(592),
    [sym_compiler_directive_decl] = STATE(592),
    [sym_fsi_directive_decl] = STATE(592),
    [sym_preproc_line] = STATE(592),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(296),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(593)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(593),
    [sym_compiler_directive_decl] = STATE(593),
    [sym_fsi_directive_decl] = STATE(593),
    [sym_preproc_line] = STATE(593),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(300),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(594)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(594),
    [sym_compiler_directive_decl] = STATE(594),
    [sym_fsi_directive_decl] = STATE(594),
    [sym_preproc_line] = STATE(594),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(595)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(595),
    [sym_compiler_directive_decl] = STATE(595),
    [sym_fsi_directive_decl] = STATE(595),
    [sym_preproc_line] = STATE(595),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(304),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(596)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(596),
    [sym_compiler_directive_decl] = STATE(596),
    [sym_fsi_directive_decl] = STATE(596),
    [sym_preproc_line] = STATE(596),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_RBRACE] = ACTIONS(3369),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(597)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(597),
    [sym_compiler_directive_decl] = STATE(597),
    [sym_fsi_directive_decl] = STATE(597),
    [sym_preproc_line] = STATE(597),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(308),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(598)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(598),
    [sym_compiler_directive_decl] = STATE(598),
    [sym_fsi_directive_decl] = STATE(598),
    [sym_preproc_line] = STATE(598),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3388),
    [sym__dedent] = ACTIONS(3388),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(599)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(599),
    [sym_compiler_directive_decl] = STATE(599),
    [sym_fsi_directive_decl] = STATE(599),
    [sym_preproc_line] = STATE(599),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(1253),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(600)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(600),
    [sym_compiler_directive_decl] = STATE(600),
    [sym_fsi_directive_decl] = STATE(600),
    [sym_preproc_line] = STATE(600),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3390),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(601)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(601),
    [sym_compiler_directive_decl] = STATE(601),
    [sym_fsi_directive_decl] = STATE(601),
    [sym_preproc_line] = STATE(601),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3390),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(602)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(602),
    [sym_compiler_directive_decl] = STATE(602),
    [sym_fsi_directive_decl] = STATE(602),
    [sym_preproc_line] = STATE(602),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3392),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(603)] = {
    [sym__preproc_toplevel_module] = STATE(16938),
    [sym__module_elem] = STATE(16945),
    [sym__attribute_expression] = STATE(17962),
    [sym_module_abbrev] = STATE(18004),
    [sym_module_defn] = STATE(18004),
    [sym_import_decl] = STATE(18004),
    [sym_attributes] = STATE(1228),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(17962),
    [sym_do] = STATE(18063),
    [sym_function_or_value_defn] = STATE(1169),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(375),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(18004),
    [sym_type_definition] = STATE(18004),
    [sym_extern_binding] = STATE(18004),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(603),
    [sym_compiler_directive_decl] = STATE(603),
    [sym_fsi_directive_decl] = STATE(603),
    [sym_preproc_line] = STATE(603),
    [sym_preproc_if] = STATE(18004),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19117),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1197),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_module] = ACTIONS(3202),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_open] = ACTIONS(3204),
    [anon_sym_type] = ACTIONS(3206),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(3208),
    [anon_sym_let] = ACTIONS(3210),
    [anon_sym_let_BANG] = ACTIONS(3212),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(3218),
    [anon_sym_extern] = ACTIONS(3220),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(3226),
    [anon_sym_POUNDload] = ACTIONS(3226),
    [anon_sym_POUNDtime] = ACTIONS(3226),
    [anon_sym_POUNDI] = ACTIONS(3226),
    [anon_sym_POUNDhelp] = ACTIONS(3226),
    [anon_sym_POUNDquit] = ACTIONS(3226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3228),
    [anon_sym_POUNDendif] = ACTIONS(3394),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(604)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(604),
    [sym_compiler_directive_decl] = STATE(604),
    [sym_fsi_directive_decl] = STATE(604),
    [sym_preproc_line] = STATE(604),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3396),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(605)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(605),
    [sym_compiler_directive_decl] = STATE(605),
    [sym_fsi_directive_decl] = STATE(605),
    [sym_preproc_line] = STATE(605),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3396),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(606)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(606),
    [sym_compiler_directive_decl] = STATE(606),
    [sym_fsi_directive_decl] = STATE(606),
    [sym_preproc_line] = STATE(606),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3398),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(607)] = {
    [sym__preproc_toplevel_module] = STATE(17160),
    [sym__module_elem] = STATE(17164),
    [sym__attribute_expression] = STATE(17962),
    [sym_module_abbrev] = STATE(18004),
    [sym_module_defn] = STATE(18004),
    [sym_import_decl] = STATE(18004),
    [sym_attributes] = STATE(1228),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(17962),
    [sym_do] = STATE(18063),
    [sym_function_or_value_defn] = STATE(1169),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(375),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(18004),
    [sym_type_definition] = STATE(18004),
    [sym_extern_binding] = STATE(18004),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(607),
    [sym_compiler_directive_decl] = STATE(607),
    [sym_fsi_directive_decl] = STATE(607),
    [sym_preproc_line] = STATE(607),
    [sym_preproc_if] = STATE(18004),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19787),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1203),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_module] = ACTIONS(3202),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_open] = ACTIONS(3204),
    [anon_sym_type] = ACTIONS(3206),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(3208),
    [anon_sym_let] = ACTIONS(3210),
    [anon_sym_let_BANG] = ACTIONS(3212),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(3218),
    [anon_sym_extern] = ACTIONS(3220),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(3226),
    [anon_sym_POUNDload] = ACTIONS(3226),
    [anon_sym_POUNDtime] = ACTIONS(3226),
    [anon_sym_POUNDI] = ACTIONS(3226),
    [anon_sym_POUNDhelp] = ACTIONS(3226),
    [anon_sym_POUNDquit] = ACTIONS(3226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3228),
    [anon_sym_POUNDendif] = ACTIONS(3400),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(608)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(523),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1627),
    [sym_infix_op] = STATE(1585),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(608),
    [sym_compiler_directive_decl] = STATE(608),
    [sym_fsi_directive_decl] = STATE(608),
    [sym_preproc_line] = STATE(608),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_sequential_expression_repeat1] = STATE(5480),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_COLON] = ACTIONS(2990),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(2996),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2990),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(3402),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(3006),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3008),
    [anon_sym_DOT] = ACTIONS(3010),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_COLON_GT] = ACTIONS(3018),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_DOT_LBRACK] = ACTIONS(3036),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_LPAREN2] = ACTIONS(3044),
    [anon_sym_PIPE_PIPE] = ACTIONS(3008),
    [anon_sym_AMP_AMP] = ACTIONS(3046),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(3062),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(3066),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__newline] = ACTIONS(3074),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
    [sym__tyapp_open] = ACTIONS(2982),
  },
  [STATE(609)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(609),
    [sym_compiler_directive_decl] = STATE(609),
    [sym_fsi_directive_decl] = STATE(609),
    [sym_preproc_line] = STATE(609),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3404),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(610)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(610),
    [sym_compiler_directive_decl] = STATE(610),
    [sym_fsi_directive_decl] = STATE(610),
    [sym_preproc_line] = STATE(610),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3404),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(611)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(611),
    [sym_compiler_directive_decl] = STATE(611),
    [sym_fsi_directive_decl] = STATE(611),
    [sym_preproc_line] = STATE(611),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3406),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(612)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(612),
    [sym_compiler_directive_decl] = STATE(612),
    [sym_fsi_directive_decl] = STATE(612),
    [sym_preproc_line] = STATE(612),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3408),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(613)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(613),
    [sym_compiler_directive_decl] = STATE(613),
    [sym_fsi_directive_decl] = STATE(613),
    [sym_preproc_line] = STATE(613),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(614)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(614),
    [sym_compiler_directive_decl] = STATE(614),
    [sym_fsi_directive_decl] = STATE(614),
    [sym_preproc_line] = STATE(614),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(615)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(615),
    [sym_compiler_directive_decl] = STATE(615),
    [sym_fsi_directive_decl] = STATE(615),
    [sym_preproc_line] = STATE(615),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3412),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(616)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(616),
    [sym_compiler_directive_decl] = STATE(616),
    [sym_fsi_directive_decl] = STATE(616),
    [sym_preproc_line] = STATE(616),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(3414),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(3414),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(617)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(617),
    [sym_compiler_directive_decl] = STATE(617),
    [sym_fsi_directive_decl] = STATE(617),
    [sym_preproc_line] = STATE(617),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3416),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(618)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(618),
    [sym_compiler_directive_decl] = STATE(618),
    [sym_fsi_directive_decl] = STATE(618),
    [sym_preproc_line] = STATE(618),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(619)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(619),
    [sym_compiler_directive_decl] = STATE(619),
    [sym_fsi_directive_decl] = STATE(619),
    [sym_preproc_line] = STATE(619),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(620)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(620),
    [sym_compiler_directive_decl] = STATE(620),
    [sym_fsi_directive_decl] = STATE(620),
    [sym_preproc_line] = STATE(620),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3420),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(621)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(621),
    [sym_compiler_directive_decl] = STATE(621),
    [sym_fsi_directive_decl] = STATE(621),
    [sym_preproc_line] = STATE(621),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3422),
    [sym__dedent] = ACTIONS(3422),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(622)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(622),
    [sym_compiler_directive_decl] = STATE(622),
    [sym_fsi_directive_decl] = STATE(622),
    [sym_preproc_line] = STATE(622),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3424),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(623)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(623),
    [sym_compiler_directive_decl] = STATE(623),
    [sym_fsi_directive_decl] = STATE(623),
    [sym_preproc_line] = STATE(623),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3426),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(624)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(624),
    [sym_compiler_directive_decl] = STATE(624),
    [sym_fsi_directive_decl] = STATE(624),
    [sym_preproc_line] = STATE(624),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3426),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(625)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(625),
    [sym_compiler_directive_decl] = STATE(625),
    [sym_fsi_directive_decl] = STATE(625),
    [sym_preproc_line] = STATE(625),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3428),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(626)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(626),
    [sym_compiler_directive_decl] = STATE(626),
    [sym_fsi_directive_decl] = STATE(626),
    [sym_preproc_line] = STATE(626),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3430),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(627)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(627),
    [sym_compiler_directive_decl] = STATE(627),
    [sym_fsi_directive_decl] = STATE(627),
    [sym_preproc_line] = STATE(627),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3432),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(628)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(628),
    [sym_compiler_directive_decl] = STATE(628),
    [sym_fsi_directive_decl] = STATE(628),
    [sym_preproc_line] = STATE(628),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3432),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(629)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(629),
    [sym_compiler_directive_decl] = STATE(629),
    [sym_fsi_directive_decl] = STATE(629),
    [sym_preproc_line] = STATE(629),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3434),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(630)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(630),
    [sym_compiler_directive_decl] = STATE(630),
    [sym_fsi_directive_decl] = STATE(630),
    [sym_preproc_line] = STATE(630),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(3436),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3436),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(631)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(631),
    [sym_compiler_directive_decl] = STATE(631),
    [sym_fsi_directive_decl] = STATE(631),
    [sym_preproc_line] = STATE(631),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3438),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(632)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(632),
    [sym_compiler_directive_decl] = STATE(632),
    [sym_fsi_directive_decl] = STATE(632),
    [sym_preproc_line] = STATE(632),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3438),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(633)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(633),
    [sym_compiler_directive_decl] = STATE(633),
    [sym_fsi_directive_decl] = STATE(633),
    [sym_preproc_line] = STATE(633),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3440),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(634)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(634),
    [sym_compiler_directive_decl] = STATE(634),
    [sym_fsi_directive_decl] = STATE(634),
    [sym_preproc_line] = STATE(634),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3442),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(635)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(635),
    [sym_compiler_directive_decl] = STATE(635),
    [sym_fsi_directive_decl] = STATE(635),
    [sym_preproc_line] = STATE(635),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3442),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(636)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(636),
    [sym_compiler_directive_decl] = STATE(636),
    [sym_fsi_directive_decl] = STATE(636),
    [sym_preproc_line] = STATE(636),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3444),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(637)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(637),
    [sym_compiler_directive_decl] = STATE(637),
    [sym_fsi_directive_decl] = STATE(637),
    [sym_preproc_line] = STATE(637),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(193),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3446),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(638)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(638),
    [sym_compiler_directive_decl] = STATE(638),
    [sym_fsi_directive_decl] = STATE(638),
    [sym_preproc_line] = STATE(638),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3449),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(639)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(639),
    [sym_compiler_directive_decl] = STATE(639),
    [sym_fsi_directive_decl] = STATE(639),
    [sym_preproc_line] = STATE(639),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3449),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(640)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(640),
    [sym_compiler_directive_decl] = STATE(640),
    [sym_fsi_directive_decl] = STATE(640),
    [sym_preproc_line] = STATE(640),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3451),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(641)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(569),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1672),
    [sym_infix_op] = STATE(1665),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(641),
    [sym_compiler_directive_decl] = STATE(641),
    [sym_fsi_directive_decl] = STATE(641),
    [sym_preproc_line] = STATE(641),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_sequential_expression_repeat1] = STATE(5758),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_COLON] = ACTIONS(3242),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_COMMA] = ACTIONS(3248),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3242),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(3254),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(3258),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3260),
    [anon_sym_DOT] = ACTIONS(3262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_COLON_GT] = ACTIONS(3270),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3270),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_DOT_LBRACK] = ACTIONS(3288),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_LPAREN2] = ACTIONS(3296),
    [anon_sym_PIPE_PIPE] = ACTIONS(3260),
    [anon_sym_AMP_AMP] = ACTIONS(3298),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(3314),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(3318),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(1257),
    [sym__newline] = ACTIONS(3326),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
    [sym__tyapp_open] = ACTIONS(3330),
  },
  [STATE(642)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(642),
    [sym_compiler_directive_decl] = STATE(642),
    [sym_fsi_directive_decl] = STATE(642),
    [sym_preproc_line] = STATE(642),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(280),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3453),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(643)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(643),
    [sym_compiler_directive_decl] = STATE(643),
    [sym_fsi_directive_decl] = STATE(643),
    [sym_preproc_line] = STATE(643),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3455),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(644)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(644),
    [sym_compiler_directive_decl] = STATE(644),
    [sym_fsi_directive_decl] = STATE(644),
    [sym_preproc_line] = STATE(644),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3455),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(645)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(645),
    [sym_compiler_directive_decl] = STATE(645),
    [sym_fsi_directive_decl] = STATE(645),
    [sym_preproc_line] = STATE(645),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3457),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(646)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(646),
    [sym_compiler_directive_decl] = STATE(646),
    [sym_fsi_directive_decl] = STATE(646),
    [sym_preproc_line] = STATE(646),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_with] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(647)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(647),
    [sym_compiler_directive_decl] = STATE(647),
    [sym_fsi_directive_decl] = STATE(647),
    [sym_preproc_line] = STATE(647),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(286),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3453),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(648)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(648),
    [sym_compiler_directive_decl] = STATE(648),
    [sym_fsi_directive_decl] = STATE(648),
    [sym_preproc_line] = STATE(648),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3459),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(649)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(649),
    [sym_compiler_directive_decl] = STATE(649),
    [sym_fsi_directive_decl] = STATE(649),
    [sym_preproc_line] = STATE(649),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3459),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(650)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(650),
    [sym_compiler_directive_decl] = STATE(650),
    [sym_fsi_directive_decl] = STATE(650),
    [sym_preproc_line] = STATE(650),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3461),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(651)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(651),
    [sym_compiler_directive_decl] = STATE(651),
    [sym_fsi_directive_decl] = STATE(651),
    [sym_preproc_line] = STATE(651),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(652)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(652),
    [sym_compiler_directive_decl] = STATE(652),
    [sym_fsi_directive_decl] = STATE(652),
    [sym_preproc_line] = STATE(652),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(294),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3453),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(653)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(653),
    [sym_compiler_directive_decl] = STATE(653),
    [sym_fsi_directive_decl] = STATE(653),
    [sym_preproc_line] = STATE(653),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(298),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(654)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(654),
    [sym_compiler_directive_decl] = STATE(654),
    [sym_fsi_directive_decl] = STATE(654),
    [sym_preproc_line] = STATE(654),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(2676),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(655)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(655),
    [sym_compiler_directive_decl] = STATE(655),
    [sym_fsi_directive_decl] = STATE(655),
    [sym_preproc_line] = STATE(655),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3463),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(656)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(656),
    [sym_compiler_directive_decl] = STATE(656),
    [sym_fsi_directive_decl] = STATE(656),
    [sym_preproc_line] = STATE(656),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(657)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(657),
    [sym_compiler_directive_decl] = STATE(657),
    [sym_fsi_directive_decl] = STATE(657),
    [sym_preproc_line] = STATE(657),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(658)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(658),
    [sym_compiler_directive_decl] = STATE(658),
    [sym_fsi_directive_decl] = STATE(658),
    [sym_preproc_line] = STATE(658),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3465),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(659)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(659),
    [sym_compiler_directive_decl] = STATE(659),
    [sym_fsi_directive_decl] = STATE(659),
    [sym_preproc_line] = STATE(659),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3465),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(660)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(660),
    [sym_compiler_directive_decl] = STATE(660),
    [sym_fsi_directive_decl] = STATE(660),
    [sym_preproc_line] = STATE(660),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3467),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(661)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(661),
    [sym_compiler_directive_decl] = STATE(661),
    [sym_fsi_directive_decl] = STATE(661),
    [sym_preproc_line] = STATE(661),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_with] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(662)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(662),
    [sym_compiler_directive_decl] = STATE(662),
    [sym_fsi_directive_decl] = STATE(662),
    [sym_preproc_line] = STATE(662),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(306),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(663)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(663),
    [sym_compiler_directive_decl] = STATE(663),
    [sym_fsi_directive_decl] = STATE(663),
    [sym_preproc_line] = STATE(663),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3469),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(664)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(664),
    [sym_compiler_directive_decl] = STATE(664),
    [sym_fsi_directive_decl] = STATE(664),
    [sym_preproc_line] = STATE(664),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3469),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(665)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(665),
    [sym_compiler_directive_decl] = STATE(665),
    [sym_fsi_directive_decl] = STATE(665),
    [sym_preproc_line] = STATE(665),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3471),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(666)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(666),
    [sym_compiler_directive_decl] = STATE(666),
    [sym_fsi_directive_decl] = STATE(666),
    [sym_preproc_line] = STATE(666),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(310),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3453),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(667)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(667),
    [sym_compiler_directive_decl] = STATE(667),
    [sym_fsi_directive_decl] = STATE(667),
    [sym_preproc_line] = STATE(667),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_with] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(668)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(668),
    [sym_compiler_directive_decl] = STATE(668),
    [sym_fsi_directive_decl] = STATE(668),
    [sym_preproc_line] = STATE(668),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3473),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(669)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(669),
    [sym_compiler_directive_decl] = STATE(669),
    [sym_fsi_directive_decl] = STATE(669),
    [sym_preproc_line] = STATE(669),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3473),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(670)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(670),
    [sym_compiler_directive_decl] = STATE(670),
    [sym_fsi_directive_decl] = STATE(670),
    [sym_preproc_line] = STATE(670),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3475),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(671)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(671),
    [sym_compiler_directive_decl] = STATE(671),
    [sym_fsi_directive_decl] = STATE(671),
    [sym_preproc_line] = STATE(671),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3477),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(672)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(672),
    [sym_compiler_directive_decl] = STATE(672),
    [sym_fsi_directive_decl] = STATE(672),
    [sym_preproc_line] = STATE(672),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3477),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(673)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(673),
    [sym_compiler_directive_decl] = STATE(673),
    [sym_fsi_directive_decl] = STATE(673),
    [sym_preproc_line] = STATE(673),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3479),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(674)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(674),
    [sym_compiler_directive_decl] = STATE(674),
    [sym_fsi_directive_decl] = STATE(674),
    [sym_preproc_line] = STATE(674),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(199),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(675)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(675),
    [sym_compiler_directive_decl] = STATE(675),
    [sym_fsi_directive_decl] = STATE(675),
    [sym_preproc_line] = STATE(675),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3484),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(676)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(676),
    [sym_compiler_directive_decl] = STATE(676),
    [sym_fsi_directive_decl] = STATE(676),
    [sym_preproc_line] = STATE(676),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3484),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(677)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(677),
    [sym_compiler_directive_decl] = STATE(677),
    [sym_fsi_directive_decl] = STATE(677),
    [sym_preproc_line] = STATE(677),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3486),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(678)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(678),
    [sym_compiler_directive_decl] = STATE(678),
    [sym_fsi_directive_decl] = STATE(678),
    [sym_preproc_line] = STATE(678),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(282),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(679)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(679),
    [sym_compiler_directive_decl] = STATE(679),
    [sym_fsi_directive_decl] = STATE(679),
    [sym_preproc_line] = STATE(679),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3488),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(680)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(680),
    [sym_compiler_directive_decl] = STATE(680),
    [sym_fsi_directive_decl] = STATE(680),
    [sym_preproc_line] = STATE(680),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3488),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(681)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(681),
    [sym_compiler_directive_decl] = STATE(681),
    [sym_fsi_directive_decl] = STATE(681),
    [sym_preproc_line] = STATE(681),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(682)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(682),
    [sym_compiler_directive_decl] = STATE(682),
    [sym_fsi_directive_decl] = STATE(682),
    [sym_preproc_line] = STATE(682),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__multi_dollar_interp_end] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(683)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(683),
    [sym_compiler_directive_decl] = STATE(683),
    [sym_fsi_directive_decl] = STATE(683),
    [sym_preproc_line] = STATE(683),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(288),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(684)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(684),
    [sym_compiler_directive_decl] = STATE(684),
    [sym_fsi_directive_decl] = STATE(684),
    [sym_preproc_line] = STATE(684),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3492),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(685)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(685),
    [sym_compiler_directive_decl] = STATE(685),
    [sym_fsi_directive_decl] = STATE(685),
    [sym_preproc_line] = STATE(685),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3492),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(686)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(686),
    [sym_compiler_directive_decl] = STATE(686),
    [sym_fsi_directive_decl] = STATE(686),
    [sym_preproc_line] = STATE(686),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(687)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(687),
    [sym_compiler_directive_decl] = STATE(687),
    [sym_fsi_directive_decl] = STATE(687),
    [sym_preproc_line] = STATE(687),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(688)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(688),
    [sym_compiler_directive_decl] = STATE(688),
    [sym_fsi_directive_decl] = STATE(688),
    [sym_preproc_line] = STATE(688),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(296),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(689)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(689),
    [sym_compiler_directive_decl] = STATE(689),
    [sym_fsi_directive_decl] = STATE(689),
    [sym_preproc_line] = STATE(689),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3496),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(690)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(690),
    [sym_compiler_directive_decl] = STATE(690),
    [sym_fsi_directive_decl] = STATE(690),
    [sym_preproc_line] = STATE(690),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3496),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(691)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(691),
    [sym_compiler_directive_decl] = STATE(691),
    [sym_fsi_directive_decl] = STATE(691),
    [sym_preproc_line] = STATE(691),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3498),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(692)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(692),
    [sym_compiler_directive_decl] = STATE(692),
    [sym_fsi_directive_decl] = STATE(692),
    [sym_preproc_line] = STATE(692),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(300),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(693)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(693),
    [sym_compiler_directive_decl] = STATE(693),
    [sym_fsi_directive_decl] = STATE(693),
    [sym_preproc_line] = STATE(693),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(304),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(694)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(694),
    [sym_compiler_directive_decl] = STATE(694),
    [sym_fsi_directive_decl] = STATE(694),
    [sym_preproc_line] = STATE(694),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3500),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(695)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(695),
    [sym_compiler_directive_decl] = STATE(695),
    [sym_fsi_directive_decl] = STATE(695),
    [sym_preproc_line] = STATE(695),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3500),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(696)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(696),
    [sym_compiler_directive_decl] = STATE(696),
    [sym_fsi_directive_decl] = STATE(696),
    [sym_preproc_line] = STATE(696),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3502),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(697)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(697),
    [sym_compiler_directive_decl] = STATE(697),
    [sym_fsi_directive_decl] = STATE(697),
    [sym_preproc_line] = STATE(697),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(304),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(698)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(698),
    [sym_compiler_directive_decl] = STATE(698),
    [sym_fsi_directive_decl] = STATE(698),
    [sym_preproc_line] = STATE(698),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__multi_dollar_interp_end] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(699)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(699),
    [sym_compiler_directive_decl] = STATE(699),
    [sym_fsi_directive_decl] = STATE(699),
    [sym_preproc_line] = STATE(699),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3504),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(700)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(700),
    [sym_compiler_directive_decl] = STATE(700),
    [sym_fsi_directive_decl] = STATE(700),
    [sym_preproc_line] = STATE(700),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3504),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(701)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(701),
    [sym_compiler_directive_decl] = STATE(701),
    [sym_fsi_directive_decl] = STATE(701),
    [sym_preproc_line] = STATE(701),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3506),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(702)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(702),
    [sym_compiler_directive_decl] = STATE(702),
    [sym_fsi_directive_decl] = STATE(702),
    [sym_preproc_line] = STATE(702),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(308),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(703)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(703),
    [sym_compiler_directive_decl] = STATE(703),
    [sym_fsi_directive_decl] = STATE(703),
    [sym_preproc_line] = STATE(703),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_COLON] = ACTIONS(3088),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_COMMA] = ACTIONS(3094),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3088),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(3100),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(3104),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3106),
    [anon_sym_DOT] = ACTIONS(3108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_COLON_GT] = ACTIONS(3116),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3116),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_DOT_LBRACK] = ACTIONS(3134),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_LPAREN2] = ACTIONS(3142),
    [anon_sym_PIPE_PIPE] = ACTIONS(3106),
    [anon_sym_AMP_AMP] = ACTIONS(3144),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(3160),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(3164),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__newline] = ACTIONS(3172),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
    [sym__multi_dollar_interp_end] = ACTIONS(312),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(704)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(704),
    [sym_compiler_directive_decl] = STATE(704),
    [sym_fsi_directive_decl] = STATE(704),
    [sym_preproc_line] = STATE(704),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3508),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(705)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(552),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1834),
    [sym_infix_op] = STATE(1838),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(705),
    [sym_compiler_directive_decl] = STATE(705),
    [sym_fsi_directive_decl] = STATE(705),
    [sym_preproc_line] = STATE(705),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_sequential_expression_repeat1] = STATE(5743),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2688),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(2704),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2706),
    [anon_sym_DOT] = ACTIONS(2708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_COLON_GT] = ACTIONS(2716),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2716),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_DOT_LBRACK] = ACTIONS(2734),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_LPAREN2] = ACTIONS(2742),
    [anon_sym_PIPE_PIPE] = ACTIONS(2706),
    [anon_sym_AMP_AMP] = ACTIONS(2744),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(2760),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(2764),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__newline] = ACTIONS(2772),
    [sym__untyped_quoted_close] = ACTIONS(3508),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
    [sym__tyapp_open] = ACTIONS(2778),
  },
  [STATE(706)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(706),
    [sym_compiler_directive_decl] = STATE(706),
    [sym_fsi_directive_decl] = STATE(706),
    [sym_preproc_line] = STATE(706),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3510),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(707)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(682),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1529),
    [sym_infix_op] = STATE(1533),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(707),
    [sym_compiler_directive_decl] = STATE(707),
    [sym_fsi_directive_decl] = STATE(707),
    [sym_preproc_line] = STATE(707),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_sequential_expression_repeat1] = STATE(5796),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__multi_dollar_interp_end] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(3178),
  },
  [STATE(708)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(708),
    [sym_compiler_directive_decl] = STATE(708),
    [sym_fsi_directive_decl] = STATE(708),
    [sym_preproc_line] = STATE(708),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3512),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(709)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(709),
    [sym_compiler_directive_decl] = STATE(709),
    [sym_fsi_directive_decl] = STATE(709),
    [sym_preproc_line] = STATE(709),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3514),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(710)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(710),
    [sym_compiler_directive_decl] = STATE(710),
    [sym_fsi_directive_decl] = STATE(710),
    [sym_preproc_line] = STATE(710),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3516),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(711)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(711),
    [sym_compiler_directive_decl] = STATE(711),
    [sym_fsi_directive_decl] = STATE(711),
    [sym_preproc_line] = STATE(711),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3518),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(712)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(712),
    [sym_compiler_directive_decl] = STATE(712),
    [sym_fsi_directive_decl] = STATE(712),
    [sym_preproc_line] = STATE(712),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1862),
    [sym__dedent] = ACTIONS(1862),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(713)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(713),
    [sym_compiler_directive_decl] = STATE(713),
    [sym_fsi_directive_decl] = STATE(713),
    [sym_preproc_line] = STATE(713),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3520),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(714)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(714),
    [sym_compiler_directive_decl] = STATE(714),
    [sym_fsi_directive_decl] = STATE(714),
    [sym_preproc_line] = STATE(714),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__dedent] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(715)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(715),
    [sym_compiler_directive_decl] = STATE(715),
    [sym_fsi_directive_decl] = STATE(715),
    [sym_preproc_line] = STATE(715),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3522),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(716)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(716),
    [sym_compiler_directive_decl] = STATE(716),
    [sym_fsi_directive_decl] = STATE(716),
    [sym_preproc_line] = STATE(716),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3524),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(717)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(717),
    [sym_compiler_directive_decl] = STATE(717),
    [sym_fsi_directive_decl] = STATE(717),
    [sym_preproc_line] = STATE(717),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3526),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(718)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(718),
    [sym_compiler_directive_decl] = STATE(718),
    [sym_fsi_directive_decl] = STATE(718),
    [sym_preproc_line] = STATE(718),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3528),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(719)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(719),
    [sym_compiler_directive_decl] = STATE(719),
    [sym_fsi_directive_decl] = STATE(719),
    [sym_preproc_line] = STATE(719),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3530),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(720)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(720),
    [sym_compiler_directive_decl] = STATE(720),
    [sym_fsi_directive_decl] = STATE(720),
    [sym_preproc_line] = STATE(720),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(1858),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3453),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(721)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(721),
    [sym_compiler_directive_decl] = STATE(721),
    [sym_fsi_directive_decl] = STATE(721),
    [sym_preproc_line] = STATE(721),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3532),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(722)] = {
    [sym__preproc_toplevel_module] = STATE(16877),
    [sym__module_elem] = STATE(16883),
    [sym__attribute_expression] = STATE(17962),
    [sym_module_abbrev] = STATE(18004),
    [sym_module_defn] = STATE(18004),
    [sym_import_decl] = STATE(18004),
    [sym_attributes] = STATE(1228),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(17962),
    [sym_do] = STATE(18063),
    [sym_function_or_value_defn] = STATE(1169),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(375),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(18004),
    [sym_type_definition] = STATE(18004),
    [sym_extern_binding] = STATE(18004),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(722),
    [sym_compiler_directive_decl] = STATE(722),
    [sym_fsi_directive_decl] = STATE(722),
    [sym_preproc_line] = STATE(722),
    [sym_preproc_if] = STATE(18004),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19251),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1231),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_module] = ACTIONS(3202),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_open] = ACTIONS(3204),
    [anon_sym_type] = ACTIONS(3206),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(3208),
    [anon_sym_let] = ACTIONS(3210),
    [anon_sym_let_BANG] = ACTIONS(3212),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(3218),
    [anon_sym_extern] = ACTIONS(3220),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(3226),
    [anon_sym_POUNDload] = ACTIONS(3226),
    [anon_sym_POUNDtime] = ACTIONS(3226),
    [anon_sym_POUNDI] = ACTIONS(3226),
    [anon_sym_POUNDhelp] = ACTIONS(3226),
    [anon_sym_POUNDquit] = ACTIONS(3226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3228),
    [anon_sym_POUNDendif] = ACTIONS(3534),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(723)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(723),
    [sym_compiler_directive_decl] = STATE(723),
    [sym_fsi_directive_decl] = STATE(723),
    [sym_preproc_line] = STATE(723),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3536),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(724)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(724),
    [sym_compiler_directive_decl] = STATE(724),
    [sym_fsi_directive_decl] = STATE(724),
    [sym_preproc_line] = STATE(724),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__dedent] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(725)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(725),
    [sym_compiler_directive_decl] = STATE(725),
    [sym_fsi_directive_decl] = STATE(725),
    [sym_preproc_line] = STATE(725),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3538),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(726)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(726),
    [sym_compiler_directive_decl] = STATE(726),
    [sym_fsi_directive_decl] = STATE(726),
    [sym_preproc_line] = STATE(726),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3540),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(727)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(727),
    [sym_compiler_directive_decl] = STATE(727),
    [sym_fsi_directive_decl] = STATE(727),
    [sym_preproc_line] = STATE(727),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3542),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(728)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(728),
    [sym_compiler_directive_decl] = STATE(728),
    [sym_fsi_directive_decl] = STATE(728),
    [sym_preproc_line] = STATE(728),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3544),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(729)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(729),
    [sym_compiler_directive_decl] = STATE(729),
    [sym_fsi_directive_decl] = STATE(729),
    [sym_preproc_line] = STATE(729),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3546),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(730)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(730),
    [sym_compiler_directive_decl] = STATE(730),
    [sym_fsi_directive_decl] = STATE(730),
    [sym_preproc_line] = STATE(730),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3548),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(731)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(731),
    [sym_compiler_directive_decl] = STATE(731),
    [sym_fsi_directive_decl] = STATE(731),
    [sym_preproc_line] = STATE(731),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3550),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(732)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(732),
    [sym_compiler_directive_decl] = STATE(732),
    [sym_fsi_directive_decl] = STATE(732),
    [sym_preproc_line] = STATE(732),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3552),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(733)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(733),
    [sym_compiler_directive_decl] = STATE(733),
    [sym_fsi_directive_decl] = STATE(733),
    [sym_preproc_line] = STATE(733),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3554),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(734)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(734),
    [sym_compiler_directive_decl] = STATE(734),
    [sym_fsi_directive_decl] = STATE(734),
    [sym_preproc_line] = STATE(734),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(735)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(735),
    [sym_compiler_directive_decl] = STATE(735),
    [sym_fsi_directive_decl] = STATE(735),
    [sym_preproc_line] = STATE(735),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(736)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(736),
    [sym_compiler_directive_decl] = STATE(736),
    [sym_fsi_directive_decl] = STATE(736),
    [sym_preproc_line] = STATE(736),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3560),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(737)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(737),
    [sym_compiler_directive_decl] = STATE(737),
    [sym_fsi_directive_decl] = STATE(737),
    [sym_preproc_line] = STATE(737),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(738)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(738),
    [sym_compiler_directive_decl] = STATE(738),
    [sym_fsi_directive_decl] = STATE(738),
    [sym_preproc_line] = STATE(738),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3564),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(739)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(739),
    [sym_compiler_directive_decl] = STATE(739),
    [sym_fsi_directive_decl] = STATE(739),
    [sym_preproc_line] = STATE(739),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3566),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(740)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(740),
    [sym_compiler_directive_decl] = STATE(740),
    [sym_fsi_directive_decl] = STATE(740),
    [sym_preproc_line] = STATE(740),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3568),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(741)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(741),
    [sym_compiler_directive_decl] = STATE(741),
    [sym_fsi_directive_decl] = STATE(741),
    [sym_preproc_line] = STATE(741),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3570),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(742)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(742),
    [sym_compiler_directive_decl] = STATE(742),
    [sym_fsi_directive_decl] = STATE(742),
    [sym_preproc_line] = STATE(742),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3572),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(743)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(743),
    [sym_compiler_directive_decl] = STATE(743),
    [sym_fsi_directive_decl] = STATE(743),
    [sym_preproc_line] = STATE(743),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3574),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(744)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(744),
    [sym_compiler_directive_decl] = STATE(744),
    [sym_fsi_directive_decl] = STATE(744),
    [sym_preproc_line] = STATE(744),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3576),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(745)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(745),
    [sym_compiler_directive_decl] = STATE(745),
    [sym_fsi_directive_decl] = STATE(745),
    [sym_preproc_line] = STATE(745),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3578),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(746)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(746),
    [sym_compiler_directive_decl] = STATE(746),
    [sym_fsi_directive_decl] = STATE(746),
    [sym_preproc_line] = STATE(746),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3580),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(747)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(747),
    [sym_compiler_directive_decl] = STATE(747),
    [sym_fsi_directive_decl] = STATE(747),
    [sym_preproc_line] = STATE(747),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3582),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(748)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(748),
    [sym_compiler_directive_decl] = STATE(748),
    [sym_fsi_directive_decl] = STATE(748),
    [sym_preproc_line] = STATE(748),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3584),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(749)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(749),
    [sym_compiler_directive_decl] = STATE(749),
    [sym_fsi_directive_decl] = STATE(749),
    [sym_preproc_line] = STATE(749),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3586),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(750)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(750),
    [sym_compiler_directive_decl] = STATE(750),
    [sym_fsi_directive_decl] = STATE(750),
    [sym_preproc_line] = STATE(750),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(751)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(751),
    [sym_compiler_directive_decl] = STATE(751),
    [sym_fsi_directive_decl] = STATE(751),
    [sym_preproc_line] = STATE(751),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3590),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(752)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(752),
    [sym_compiler_directive_decl] = STATE(752),
    [sym_fsi_directive_decl] = STATE(752),
    [sym_preproc_line] = STATE(752),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3592),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(753)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(753),
    [sym_compiler_directive_decl] = STATE(753),
    [sym_fsi_directive_decl] = STATE(753),
    [sym_preproc_line] = STATE(753),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3594),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(754)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(754),
    [sym_compiler_directive_decl] = STATE(754),
    [sym_fsi_directive_decl] = STATE(754),
    [sym_preproc_line] = STATE(754),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3596),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(755)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(755),
    [sym_compiler_directive_decl] = STATE(755),
    [sym_fsi_directive_decl] = STATE(755),
    [sym_preproc_line] = STATE(755),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3598),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(756)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(756),
    [sym_compiler_directive_decl] = STATE(756),
    [sym_fsi_directive_decl] = STATE(756),
    [sym_preproc_line] = STATE(756),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3600),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(757)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(757),
    [sym_compiler_directive_decl] = STATE(757),
    [sym_fsi_directive_decl] = STATE(757),
    [sym_preproc_line] = STATE(757),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3602),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(758)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(758),
    [sym_compiler_directive_decl] = STATE(758),
    [sym_fsi_directive_decl] = STATE(758),
    [sym_preproc_line] = STATE(758),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3604),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(759)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(759),
    [sym_compiler_directive_decl] = STATE(759),
    [sym_fsi_directive_decl] = STATE(759),
    [sym_preproc_line] = STATE(759),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(760)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(760),
    [sym_compiler_directive_decl] = STATE(760),
    [sym_fsi_directive_decl] = STATE(760),
    [sym_preproc_line] = STATE(760),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3608),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(761)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(761),
    [sym_compiler_directive_decl] = STATE(761),
    [sym_fsi_directive_decl] = STATE(761),
    [sym_preproc_line] = STATE(761),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3610),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(762)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(762),
    [sym_compiler_directive_decl] = STATE(762),
    [sym_fsi_directive_decl] = STATE(762),
    [sym_preproc_line] = STATE(762),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3612),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(763)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(763),
    [sym_compiler_directive_decl] = STATE(763),
    [sym_fsi_directive_decl] = STATE(763),
    [sym_preproc_line] = STATE(763),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(764)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(764),
    [sym_compiler_directive_decl] = STATE(764),
    [sym_fsi_directive_decl] = STATE(764),
    [sym_preproc_line] = STATE(764),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(765)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(765),
    [sym_compiler_directive_decl] = STATE(765),
    [sym_fsi_directive_decl] = STATE(765),
    [sym_preproc_line] = STATE(765),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3618),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(766)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(766),
    [sym_compiler_directive_decl] = STATE(766),
    [sym_fsi_directive_decl] = STATE(766),
    [sym_preproc_line] = STATE(766),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3620),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(767)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(767),
    [sym_compiler_directive_decl] = STATE(767),
    [sym_fsi_directive_decl] = STATE(767),
    [sym_preproc_line] = STATE(767),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(768)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(768),
    [sym_compiler_directive_decl] = STATE(768),
    [sym_fsi_directive_decl] = STATE(768),
    [sym_preproc_line] = STATE(768),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3624),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(769)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(769),
    [sym_compiler_directive_decl] = STATE(769),
    [sym_fsi_directive_decl] = STATE(769),
    [sym_preproc_line] = STATE(769),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3626),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(770)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(770),
    [sym_compiler_directive_decl] = STATE(770),
    [sym_fsi_directive_decl] = STATE(770),
    [sym_preproc_line] = STATE(770),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(771)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(771),
    [sym_compiler_directive_decl] = STATE(771),
    [sym_fsi_directive_decl] = STATE(771),
    [sym_preproc_line] = STATE(771),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3630),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(772)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(772),
    [sym_compiler_directive_decl] = STATE(772),
    [sym_fsi_directive_decl] = STATE(772),
    [sym_preproc_line] = STATE(772),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3632),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(773)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(773),
    [sym_compiler_directive_decl] = STATE(773),
    [sym_fsi_directive_decl] = STATE(773),
    [sym_preproc_line] = STATE(773),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3634),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(774)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(774),
    [sym_compiler_directive_decl] = STATE(774),
    [sym_fsi_directive_decl] = STATE(774),
    [sym_preproc_line] = STATE(774),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3636),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(775)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(775),
    [sym_compiler_directive_decl] = STATE(775),
    [sym_fsi_directive_decl] = STATE(775),
    [sym_preproc_line] = STATE(775),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3638),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(776)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(776),
    [sym_compiler_directive_decl] = STATE(776),
    [sym_fsi_directive_decl] = STATE(776),
    [sym_preproc_line] = STATE(776),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3640),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(777)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(777),
    [sym_compiler_directive_decl] = STATE(777),
    [sym_fsi_directive_decl] = STATE(777),
    [sym_preproc_line] = STATE(777),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3642),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(778)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(778),
    [sym_compiler_directive_decl] = STATE(778),
    [sym_fsi_directive_decl] = STATE(778),
    [sym_preproc_line] = STATE(778),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3644),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(779)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(779),
    [sym_compiler_directive_decl] = STATE(779),
    [sym_fsi_directive_decl] = STATE(779),
    [sym_preproc_line] = STATE(779),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3646),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(780)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(780),
    [sym_compiler_directive_decl] = STATE(780),
    [sym_fsi_directive_decl] = STATE(780),
    [sym_preproc_line] = STATE(780),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3648),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(781)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(781),
    [sym_compiler_directive_decl] = STATE(781),
    [sym_fsi_directive_decl] = STATE(781),
    [sym_preproc_line] = STATE(781),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3650),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(782)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(782),
    [sym_compiler_directive_decl] = STATE(782),
    [sym_fsi_directive_decl] = STATE(782),
    [sym_preproc_line] = STATE(782),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3652),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(783)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(783),
    [sym_compiler_directive_decl] = STATE(783),
    [sym_fsi_directive_decl] = STATE(783),
    [sym_preproc_line] = STATE(783),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3654),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(784)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(784),
    [sym_compiler_directive_decl] = STATE(784),
    [sym_fsi_directive_decl] = STATE(784),
    [sym_preproc_line] = STATE(784),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3656),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(785)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(785),
    [sym_compiler_directive_decl] = STATE(785),
    [sym_fsi_directive_decl] = STATE(785),
    [sym_preproc_line] = STATE(785),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3658),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(786)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(786),
    [sym_compiler_directive_decl] = STATE(786),
    [sym_fsi_directive_decl] = STATE(786),
    [sym_preproc_line] = STATE(786),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3660),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(787)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(787),
    [sym_compiler_directive_decl] = STATE(787),
    [sym_fsi_directive_decl] = STATE(787),
    [sym_preproc_line] = STATE(787),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3662),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(788)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(788),
    [sym_compiler_directive_decl] = STATE(788),
    [sym_fsi_directive_decl] = STATE(788),
    [sym_preproc_line] = STATE(788),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3664),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(789)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(789),
    [sym_compiler_directive_decl] = STATE(789),
    [sym_fsi_directive_decl] = STATE(789),
    [sym_preproc_line] = STATE(789),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3666),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(790)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(790),
    [sym_compiler_directive_decl] = STATE(790),
    [sym_fsi_directive_decl] = STATE(790),
    [sym_preproc_line] = STATE(790),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3668),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(791)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(791),
    [sym_compiler_directive_decl] = STATE(791),
    [sym_fsi_directive_decl] = STATE(791),
    [sym_preproc_line] = STATE(791),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3670),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(792)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(792),
    [sym_compiler_directive_decl] = STATE(792),
    [sym_fsi_directive_decl] = STATE(792),
    [sym_preproc_line] = STATE(792),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(793)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(793),
    [sym_compiler_directive_decl] = STATE(793),
    [sym_fsi_directive_decl] = STATE(793),
    [sym_preproc_line] = STATE(793),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3674),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(794)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(794),
    [sym_compiler_directive_decl] = STATE(794),
    [sym_fsi_directive_decl] = STATE(794),
    [sym_preproc_line] = STATE(794),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3676),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(795)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(795),
    [sym_compiler_directive_decl] = STATE(795),
    [sym_fsi_directive_decl] = STATE(795),
    [sym_preproc_line] = STATE(795),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3678),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(796)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(796),
    [sym_compiler_directive_decl] = STATE(796),
    [sym_fsi_directive_decl] = STATE(796),
    [sym_preproc_line] = STATE(796),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3680),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(797)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(797),
    [sym_compiler_directive_decl] = STATE(797),
    [sym_fsi_directive_decl] = STATE(797),
    [sym_preproc_line] = STATE(797),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3682),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(798)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(798),
    [sym_compiler_directive_decl] = STATE(798),
    [sym_fsi_directive_decl] = STATE(798),
    [sym_preproc_line] = STATE(798),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3684),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(799)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(799),
    [sym_compiler_directive_decl] = STATE(799),
    [sym_fsi_directive_decl] = STATE(799),
    [sym_preproc_line] = STATE(799),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3686),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(800)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(800),
    [sym_compiler_directive_decl] = STATE(800),
    [sym_fsi_directive_decl] = STATE(800),
    [sym_preproc_line] = STATE(800),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3688),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(801)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(801),
    [sym_compiler_directive_decl] = STATE(801),
    [sym_fsi_directive_decl] = STATE(801),
    [sym_preproc_line] = STATE(801),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3690),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(802)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(802),
    [sym_compiler_directive_decl] = STATE(802),
    [sym_fsi_directive_decl] = STATE(802),
    [sym_preproc_line] = STATE(802),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3692),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(803)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(803),
    [sym_compiler_directive_decl] = STATE(803),
    [sym_fsi_directive_decl] = STATE(803),
    [sym_preproc_line] = STATE(803),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3694),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(804)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(804),
    [sym_compiler_directive_decl] = STATE(804),
    [sym_fsi_directive_decl] = STATE(804),
    [sym_preproc_line] = STATE(804),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(181),
    [anon_sym_null] = ACTIONS(181),
    [anon_sym_EQ] = ACTIONS(181),
    [anon_sym_begin] = ACTIONS(181),
    [anon_sym_COLON] = ACTIONS(181),
    [anon_sym_do] = ACTIONS(181),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_let_BANG] = ACTIONS(183),
    [anon_sym_LPAREN] = ACTIONS(181),
    [anon_sym_COMMA] = ACTIONS(183),
    [anon_sym_QMARK] = ACTIONS(181),
    [anon_sym_COLON_QMARK] = ACTIONS(181),
    [anon_sym_COLON_COLON] = ACTIONS(183),
    [anon_sym_AMP] = ACTIONS(181),
    [anon_sym_LBRACK] = ACTIONS(181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(183),
    [anon_sym_LBRACE] = ACTIONS(181),
    [anon_sym_LT_AT] = ACTIONS(181),
    [anon_sym_LT_AT_AT] = ACTIONS(181),
    [anon_sym_CARET] = ACTIONS(181),
    [anon_sym_or] = ACTIONS(181),
    [anon_sym_DOT] = ACTIONS(181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(183),
    [anon_sym_new] = ACTIONS(181),
    [anon_sym_return] = ACTIONS(181),
    [anon_sym_return_BANG] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(181),
    [anon_sym_yield_BANG] = ACTIONS(183),
    [anon_sym_lazy] = ACTIONS(181),
    [anon_sym_assert] = ACTIONS(181),
    [anon_sym_upcast] = ACTIONS(181),
    [anon_sym_downcast] = ACTIONS(181),
    [anon_sym_fixed] = ACTIONS(181),
    [anon_sym_COLON_GT] = ACTIONS(183),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(183),
    [anon_sym_for] = ACTIONS(181),
    [anon_sym_while] = ACTIONS(181),
    [anon_sym_while_BANG] = ACTIONS(183),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_fun] = ACTIONS(181),
    [anon_sym_try] = ACTIONS(181),
    [anon_sym_match] = ACTIONS(181),
    [anon_sym_match_BANG] = ACTIONS(183),
    [anon_sym_function] = ACTIONS(181),
    [anon_sym_DOT_LBRACK] = ACTIONS(183),
    [anon_sym_use] = ACTIONS(181),
    [anon_sym_use_BANG] = ACTIONS(183),
    [anon_sym_do_BANG] = ACTIONS(183),
    [anon_sym_LPAREN2] = ACTIONS(181),
    [anon_sym_PIPE_PIPE] = ACTIONS(181),
    [anon_sym_AMP_AMP] = ACTIONS(181),
    [aux_sym_char_token1] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(181),
    [anon_sym_AT_DQUOTE] = ACTIONS(183),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(183),
    [sym_bool] = ACTIONS(181),
    [sym_unit] = ACTIONS(183),
    [anon_sym_LPAREN_PIPE] = ACTIONS(181),
    [sym_op_identifier] = ACTIONS(181),
    [anon_sym_PLUS] = ACTIONS(181),
    [anon_sym_DASH] = ACTIONS(181),
    [anon_sym_PLUS_DOT] = ACTIONS(181),
    [anon_sym_DASH_DOT] = ACTIONS(181),
    [anon_sym_PERCENT] = ACTIONS(181),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(183),
    [aux_sym_prefix_op_token1] = ACTIONS(181),
    [aux_sym_infix_op_token1] = ACTIONS(183),
    [aux_sym_infix_op_token2] = ACTIONS(181),
    [aux_sym_infix_op_token3] = ACTIONS(181),
    [anon_sym_BANG_EQ] = ACTIONS(181),
    [anon_sym_COLON_EQ] = ACTIONS(183),
    [anon_sym_DOLLAR] = ACTIONS(181),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(181),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(181),
    [sym_int] = ACTIONS(181),
    [sym_xint] = ACTIONS(183),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(183),
    [sym__newline] = ACTIONS(183),
    [sym__dedent] = ACTIONS(183),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(183),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(805)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(805),
    [sym_compiler_directive_decl] = STATE(805),
    [sym_fsi_directive_decl] = STATE(805),
    [sym_preproc_line] = STATE(805),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(2774),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(806)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(806),
    [sym_compiler_directive_decl] = STATE(806),
    [sym_fsi_directive_decl] = STATE(806),
    [sym_preproc_line] = STATE(806),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(193),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(193),
    [anon_sym_AMP_AMP] = ACTIONS(3696),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(199),
    [sym__dedent] = ACTIONS(199),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(807)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(807),
    [sym_compiler_directive_decl] = STATE(807),
    [sym_fsi_directive_decl] = STATE(807),
    [sym_preproc_line] = STATE(807),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(282),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(808)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(808),
    [sym_compiler_directive_decl] = STATE(808),
    [sym_fsi_directive_decl] = STATE(808),
    [sym_preproc_line] = STATE(808),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3388),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(809)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(809),
    [sym_compiler_directive_decl] = STATE(809),
    [sym_fsi_directive_decl] = STATE(809),
    [sym_preproc_line] = STATE(809),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3699),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(810)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(810),
    [sym_compiler_directive_decl] = STATE(810),
    [sym_fsi_directive_decl] = STATE(810),
    [sym_preproc_line] = STATE(810),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3701),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(811)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(811),
    [sym_compiler_directive_decl] = STATE(811),
    [sym_fsi_directive_decl] = STATE(811),
    [sym_preproc_line] = STATE(811),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(812)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(812),
    [sym_compiler_directive_decl] = STATE(812),
    [sym_fsi_directive_decl] = STATE(812),
    [sym_preproc_line] = STATE(812),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(813)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(813),
    [sym_compiler_directive_decl] = STATE(813),
    [sym_fsi_directive_decl] = STATE(813),
    [sym_preproc_line] = STATE(813),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3705),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(814)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(814),
    [sym_compiler_directive_decl] = STATE(814),
    [sym_fsi_directive_decl] = STATE(814),
    [sym_preproc_line] = STATE(814),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3707),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(815)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(815),
    [sym_compiler_directive_decl] = STATE(815),
    [sym_fsi_directive_decl] = STATE(815),
    [sym_preproc_line] = STATE(815),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(816)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(816),
    [sym_compiler_directive_decl] = STATE(816),
    [sym_fsi_directive_decl] = STATE(816),
    [sym_preproc_line] = STATE(816),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3711),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(817)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(817),
    [sym_compiler_directive_decl] = STATE(817),
    [sym_fsi_directive_decl] = STATE(817),
    [sym_preproc_line] = STATE(817),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3713),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(818)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(818),
    [sym_compiler_directive_decl] = STATE(818),
    [sym_fsi_directive_decl] = STATE(818),
    [sym_preproc_line] = STATE(818),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(819)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(819),
    [sym_compiler_directive_decl] = STATE(819),
    [sym_fsi_directive_decl] = STATE(819),
    [sym_preproc_line] = STATE(819),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3717),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(820)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(820),
    [sym_compiler_directive_decl] = STATE(820),
    [sym_fsi_directive_decl] = STATE(820),
    [sym_preproc_line] = STATE(820),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3719),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(821)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(821),
    [sym_compiler_directive_decl] = STATE(821),
    [sym_fsi_directive_decl] = STATE(821),
    [sym_preproc_line] = STATE(821),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3721),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(822)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(822),
    [sym_compiler_directive_decl] = STATE(822),
    [sym_fsi_directive_decl] = STATE(822),
    [sym_preproc_line] = STATE(822),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3723),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(823)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(823),
    [sym_compiler_directive_decl] = STATE(823),
    [sym_fsi_directive_decl] = STATE(823),
    [sym_preproc_line] = STATE(823),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3725),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(824)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(824),
    [sym_compiler_directive_decl] = STATE(824),
    [sym_fsi_directive_decl] = STATE(824),
    [sym_preproc_line] = STATE(824),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(825)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(825),
    [sym_compiler_directive_decl] = STATE(825),
    [sym_fsi_directive_decl] = STATE(825),
    [sym_preproc_line] = STATE(825),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3729),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(826)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(826),
    [sym_compiler_directive_decl] = STATE(826),
    [sym_fsi_directive_decl] = STATE(826),
    [sym_preproc_line] = STATE(826),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3731),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(827)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(827),
    [sym_compiler_directive_decl] = STATE(827),
    [sym_fsi_directive_decl] = STATE(827),
    [sym_preproc_line] = STATE(827),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3733),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(828)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(828),
    [sym_compiler_directive_decl] = STATE(828),
    [sym_fsi_directive_decl] = STATE(828),
    [sym_preproc_line] = STATE(828),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3735),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(829)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(829),
    [sym_compiler_directive_decl] = STATE(829),
    [sym_fsi_directive_decl] = STATE(829),
    [sym_preproc_line] = STATE(829),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3737),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(830)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(830),
    [sym_compiler_directive_decl] = STATE(830),
    [sym_fsi_directive_decl] = STATE(830),
    [sym_preproc_line] = STATE(830),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3739),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(831)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(831),
    [sym_compiler_directive_decl] = STATE(831),
    [sym_fsi_directive_decl] = STATE(831),
    [sym_preproc_line] = STATE(831),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3741),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(832)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(832),
    [sym_compiler_directive_decl] = STATE(832),
    [sym_fsi_directive_decl] = STATE(832),
    [sym_preproc_line] = STATE(832),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3743),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(833)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(833),
    [sym_compiler_directive_decl] = STATE(833),
    [sym_fsi_directive_decl] = STATE(833),
    [sym_preproc_line] = STATE(833),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3745),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(834)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(834),
    [sym_compiler_directive_decl] = STATE(834),
    [sym_fsi_directive_decl] = STATE(834),
    [sym_preproc_line] = STATE(834),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3747),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(835)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(835),
    [sym_compiler_directive_decl] = STATE(835),
    [sym_fsi_directive_decl] = STATE(835),
    [sym_preproc_line] = STATE(835),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3749),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(836)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(836),
    [sym_compiler_directive_decl] = STATE(836),
    [sym_fsi_directive_decl] = STATE(836),
    [sym_preproc_line] = STATE(836),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(837)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(837),
    [sym_compiler_directive_decl] = STATE(837),
    [sym_fsi_directive_decl] = STATE(837),
    [sym_preproc_line] = STATE(837),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3753),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(838)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(838),
    [sym_compiler_directive_decl] = STATE(838),
    [sym_fsi_directive_decl] = STATE(838),
    [sym_preproc_line] = STATE(838),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3755),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(839)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(839),
    [sym_compiler_directive_decl] = STATE(839),
    [sym_fsi_directive_decl] = STATE(839),
    [sym_preproc_line] = STATE(839),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3757),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(840)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(840),
    [sym_compiler_directive_decl] = STATE(840),
    [sym_fsi_directive_decl] = STATE(840),
    [sym_preproc_line] = STATE(840),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3759),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(841)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(841),
    [sym_compiler_directive_decl] = STATE(841),
    [sym_fsi_directive_decl] = STATE(841),
    [sym_preproc_line] = STATE(841),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3761),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(842)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(842),
    [sym_compiler_directive_decl] = STATE(842),
    [sym_fsi_directive_decl] = STATE(842),
    [sym_preproc_line] = STATE(842),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3763),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(843)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(843),
    [sym_compiler_directive_decl] = STATE(843),
    [sym_fsi_directive_decl] = STATE(843),
    [sym_preproc_line] = STATE(843),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3765),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(844)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(844),
    [sym_compiler_directive_decl] = STATE(844),
    [sym_fsi_directive_decl] = STATE(844),
    [sym_preproc_line] = STATE(844),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3767),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(845)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(845),
    [sym_compiler_directive_decl] = STATE(845),
    [sym_fsi_directive_decl] = STATE(845),
    [sym_preproc_line] = STATE(845),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3769),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(846)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(846),
    [sym_compiler_directive_decl] = STATE(846),
    [sym_fsi_directive_decl] = STATE(846),
    [sym_preproc_line] = STATE(846),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3771),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(847)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(847),
    [sym_compiler_directive_decl] = STATE(847),
    [sym_fsi_directive_decl] = STATE(847),
    [sym_preproc_line] = STATE(847),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3773),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(848)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(848),
    [sym_compiler_directive_decl] = STATE(848),
    [sym_fsi_directive_decl] = STATE(848),
    [sym_preproc_line] = STATE(848),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(849)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(849),
    [sym_compiler_directive_decl] = STATE(849),
    [sym_fsi_directive_decl] = STATE(849),
    [sym_preproc_line] = STATE(849),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3777),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(850)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(850),
    [sym_compiler_directive_decl] = STATE(850),
    [sym_fsi_directive_decl] = STATE(850),
    [sym_preproc_line] = STATE(850),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3779),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(851)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(851),
    [sym_compiler_directive_decl] = STATE(851),
    [sym_fsi_directive_decl] = STATE(851),
    [sym_preproc_line] = STATE(851),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3781),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(852)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(852),
    [sym_compiler_directive_decl] = STATE(852),
    [sym_fsi_directive_decl] = STATE(852),
    [sym_preproc_line] = STATE(852),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3783),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(853)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(853),
    [sym_compiler_directive_decl] = STATE(853),
    [sym_fsi_directive_decl] = STATE(853),
    [sym_preproc_line] = STATE(853),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3785),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(854)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(854),
    [sym_compiler_directive_decl] = STATE(854),
    [sym_fsi_directive_decl] = STATE(854),
    [sym_preproc_line] = STATE(854),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(855)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(855),
    [sym_compiler_directive_decl] = STATE(855),
    [sym_fsi_directive_decl] = STATE(855),
    [sym_preproc_line] = STATE(855),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3789),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(856)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(856),
    [sym_compiler_directive_decl] = STATE(856),
    [sym_fsi_directive_decl] = STATE(856),
    [sym_preproc_line] = STATE(856),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3791),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(857)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(857),
    [sym_compiler_directive_decl] = STATE(857),
    [sym_fsi_directive_decl] = STATE(857),
    [sym_preproc_line] = STATE(857),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3793),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(858)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(858),
    [sym_compiler_directive_decl] = STATE(858),
    [sym_fsi_directive_decl] = STATE(858),
    [sym_preproc_line] = STATE(858),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3795),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(859)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(859),
    [sym_compiler_directive_decl] = STATE(859),
    [sym_fsi_directive_decl] = STATE(859),
    [sym_preproc_line] = STATE(859),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3797),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(860)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(860),
    [sym_compiler_directive_decl] = STATE(860),
    [sym_fsi_directive_decl] = STATE(860),
    [sym_preproc_line] = STATE(860),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(861)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(861),
    [sym_compiler_directive_decl] = STATE(861),
    [sym_fsi_directive_decl] = STATE(861),
    [sym_preproc_line] = STATE(861),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3801),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(862)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(862),
    [sym_compiler_directive_decl] = STATE(862),
    [sym_fsi_directive_decl] = STATE(862),
    [sym_preproc_line] = STATE(862),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3803),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(863)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(863),
    [sym_compiler_directive_decl] = STATE(863),
    [sym_fsi_directive_decl] = STATE(863),
    [sym_preproc_line] = STATE(863),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3805),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(864)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(864),
    [sym_compiler_directive_decl] = STATE(864),
    [sym_fsi_directive_decl] = STATE(864),
    [sym_preproc_line] = STATE(864),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3807),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(865)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(865),
    [sym_compiler_directive_decl] = STATE(865),
    [sym_fsi_directive_decl] = STATE(865),
    [sym_preproc_line] = STATE(865),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3809),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(866)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(866),
    [sym_compiler_directive_decl] = STATE(866),
    [sym_fsi_directive_decl] = STATE(866),
    [sym_preproc_line] = STATE(866),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(867)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(867),
    [sym_compiler_directive_decl] = STATE(867),
    [sym_fsi_directive_decl] = STATE(867),
    [sym_preproc_line] = STATE(867),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3813),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(868)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(868),
    [sym_compiler_directive_decl] = STATE(868),
    [sym_fsi_directive_decl] = STATE(868),
    [sym_preproc_line] = STATE(868),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3815),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(869)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(869),
    [sym_compiler_directive_decl] = STATE(869),
    [sym_fsi_directive_decl] = STATE(869),
    [sym_preproc_line] = STATE(869),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3817),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(870)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(870),
    [sym_compiler_directive_decl] = STATE(870),
    [sym_fsi_directive_decl] = STATE(870),
    [sym_preproc_line] = STATE(870),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3819),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(871)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(871),
    [sym_compiler_directive_decl] = STATE(871),
    [sym_fsi_directive_decl] = STATE(871),
    [sym_preproc_line] = STATE(871),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3821),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(872)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(872),
    [sym_compiler_directive_decl] = STATE(872),
    [sym_fsi_directive_decl] = STATE(872),
    [sym_preproc_line] = STATE(872),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3823),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(873)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(873),
    [sym_compiler_directive_decl] = STATE(873),
    [sym_fsi_directive_decl] = STATE(873),
    [sym_preproc_line] = STATE(873),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3825),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(874)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(874),
    [sym_compiler_directive_decl] = STATE(874),
    [sym_fsi_directive_decl] = STATE(874),
    [sym_preproc_line] = STATE(874),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3827),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(875)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(875),
    [sym_compiler_directive_decl] = STATE(875),
    [sym_fsi_directive_decl] = STATE(875),
    [sym_preproc_line] = STATE(875),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(876)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(876),
    [sym_compiler_directive_decl] = STATE(876),
    [sym_fsi_directive_decl] = STATE(876),
    [sym_preproc_line] = STATE(876),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3831),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(877)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(877),
    [sym_compiler_directive_decl] = STATE(877),
    [sym_fsi_directive_decl] = STATE(877),
    [sym_preproc_line] = STATE(877),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3833),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(878)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(878),
    [sym_compiler_directive_decl] = STATE(878),
    [sym_fsi_directive_decl] = STATE(878),
    [sym_preproc_line] = STATE(878),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3835),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(879)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(879),
    [sym_compiler_directive_decl] = STATE(879),
    [sym_fsi_directive_decl] = STATE(879),
    [sym_preproc_line] = STATE(879),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3837),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(880)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(880),
    [sym_compiler_directive_decl] = STATE(880),
    [sym_fsi_directive_decl] = STATE(880),
    [sym_preproc_line] = STATE(880),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3839),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(881)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(881),
    [sym_compiler_directive_decl] = STATE(881),
    [sym_fsi_directive_decl] = STATE(881),
    [sym_preproc_line] = STATE(881),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(882)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(882),
    [sym_compiler_directive_decl] = STATE(882),
    [sym_fsi_directive_decl] = STATE(882),
    [sym_preproc_line] = STATE(882),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3843),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(883)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(883),
    [sym_compiler_directive_decl] = STATE(883),
    [sym_fsi_directive_decl] = STATE(883),
    [sym_preproc_line] = STATE(883),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3845),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(884)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(884),
    [sym_compiler_directive_decl] = STATE(884),
    [sym_fsi_directive_decl] = STATE(884),
    [sym_preproc_line] = STATE(884),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3847),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(885)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(885),
    [sym_compiler_directive_decl] = STATE(885),
    [sym_fsi_directive_decl] = STATE(885),
    [sym_preproc_line] = STATE(885),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3849),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(886)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(557),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1662),
    [sym_infix_op] = STATE(1654),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(886),
    [sym_compiler_directive_decl] = STATE(886),
    [sym_fsi_directive_decl] = STATE(886),
    [sym_preproc_line] = STATE(886),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_sequential_expression_repeat1] = STATE(5683),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_COLON] = ACTIONS(3342),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_COMMA] = ACTIONS(3344),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(3342),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(2297),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(2301),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3346),
    [anon_sym_DOT] = ACTIONS(2305),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_COLON_GT] = ACTIONS(3352),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(3352),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_DASH_GT] = ACTIONS(3851),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_DOT_LBRACK] = ACTIONS(2331),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_LPAREN2] = ACTIONS(2339),
    [anon_sym_PIPE_PIPE] = ACTIONS(3346),
    [anon_sym_AMP_AMP] = ACTIONS(3360),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(2357),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(2361),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__newline] = ACTIONS(3364),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
    [sym__tyapp_open] = ACTIONS(2373),
  },
  [STATE(887)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(887),
    [sym_compiler_directive_decl] = STATE(887),
    [sym_fsi_directive_decl] = STATE(887),
    [sym_preproc_line] = STATE(887),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3853),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(888)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(888),
    [sym_compiler_directive_decl] = STATE(888),
    [sym_fsi_directive_decl] = STATE(888),
    [sym_preproc_line] = STATE(888),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3855),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(889)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(889),
    [sym_compiler_directive_decl] = STATE(889),
    [sym_fsi_directive_decl] = STATE(889),
    [sym_preproc_line] = STATE(889),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3857),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(890)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(890),
    [sym_compiler_directive_decl] = STATE(890),
    [sym_fsi_directive_decl] = STATE(890),
    [sym_preproc_line] = STATE(890),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_COLON] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(3859),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_COMMA] = ACTIONS(2792),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2786),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(2798),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(2802),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2873),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_COLON_GT] = ACTIONS(2812),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2812),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_DOT_DOT] = ACTIONS(3376),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(2873),
    [anon_sym_AMP_AMP] = ACTIONS(2875),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(2861),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__newline] = ACTIONS(2877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(891)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(891),
    [sym_compiler_directive_decl] = STATE(891),
    [sym_fsi_directive_decl] = STATE(891),
    [sym_preproc_line] = STATE(891),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3861),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3863),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(892)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(892),
    [sym_compiler_directive_decl] = STATE(892),
    [sym_fsi_directive_decl] = STATE(892),
    [sym_preproc_line] = STATE(892),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3865),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(893)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(893),
    [sym_compiler_directive_decl] = STATE(893),
    [sym_fsi_directive_decl] = STATE(893),
    [sym_preproc_line] = STATE(893),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3869),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3871),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(894)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(894),
    [sym_compiler_directive_decl] = STATE(894),
    [sym_fsi_directive_decl] = STATE(894),
    [sym_preproc_line] = STATE(894),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3873),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3875),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(895)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(895),
    [sym_compiler_directive_decl] = STATE(895),
    [sym_fsi_directive_decl] = STATE(895),
    [sym_preproc_line] = STATE(895),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3877),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3879),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(896)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(896),
    [sym_compiler_directive_decl] = STATE(896),
    [sym_fsi_directive_decl] = STATE(896),
    [sym_preproc_line] = STATE(896),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3881),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3883),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(897)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(897),
    [sym_compiler_directive_decl] = STATE(897),
    [sym_fsi_directive_decl] = STATE(897),
    [sym_preproc_line] = STATE(897),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3885),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3887),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(898)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(898),
    [sym_compiler_directive_decl] = STATE(898),
    [sym_fsi_directive_decl] = STATE(898),
    [sym_preproc_line] = STATE(898),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3889),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3891),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(899)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(899),
    [sym_compiler_directive_decl] = STATE(899),
    [sym_fsi_directive_decl] = STATE(899),
    [sym_preproc_line] = STATE(899),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3893),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3895),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(900)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(900),
    [sym_compiler_directive_decl] = STATE(900),
    [sym_fsi_directive_decl] = STATE(900),
    [sym_preproc_line] = STATE(900),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3897),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3899),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(901)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(901),
    [sym_compiler_directive_decl] = STATE(901),
    [sym_fsi_directive_decl] = STATE(901),
    [sym_preproc_line] = STATE(901),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3901),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3903),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(902)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(902),
    [sym_compiler_directive_decl] = STATE(902),
    [sym_fsi_directive_decl] = STATE(902),
    [sym_preproc_line] = STATE(902),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3905),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3907),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(903)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(903),
    [sym_compiler_directive_decl] = STATE(903),
    [sym_fsi_directive_decl] = STATE(903),
    [sym_preproc_line] = STATE(903),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3909),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(904)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(904),
    [sym_compiler_directive_decl] = STATE(904),
    [sym_fsi_directive_decl] = STATE(904),
    [sym_preproc_line] = STATE(904),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3913),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(905)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(905),
    [sym_compiler_directive_decl] = STATE(905),
    [sym_fsi_directive_decl] = STATE(905),
    [sym_preproc_line] = STATE(905),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3917),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(906)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(906),
    [sym_compiler_directive_decl] = STATE(906),
    [sym_fsi_directive_decl] = STATE(906),
    [sym_preproc_line] = STATE(906),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3921),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(907)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(907),
    [sym_compiler_directive_decl] = STATE(907),
    [sym_fsi_directive_decl] = STATE(907),
    [sym_preproc_line] = STATE(907),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3925),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3927),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(908)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(908),
    [sym_compiler_directive_decl] = STATE(908),
    [sym_fsi_directive_decl] = STATE(908),
    [sym_preproc_line] = STATE(908),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3929),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3931),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(909)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(909),
    [sym_compiler_directive_decl] = STATE(909),
    [sym_fsi_directive_decl] = STATE(909),
    [sym_preproc_line] = STATE(909),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3933),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3935),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(910)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(910),
    [sym_compiler_directive_decl] = STATE(910),
    [sym_fsi_directive_decl] = STATE(910),
    [sym_preproc_line] = STATE(910),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3937),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3939),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(911)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(911),
    [sym_compiler_directive_decl] = STATE(911),
    [sym_fsi_directive_decl] = STATE(911),
    [sym_preproc_line] = STATE(911),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3941),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3943),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(912)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(912),
    [sym_compiler_directive_decl] = STATE(912),
    [sym_fsi_directive_decl] = STATE(912),
    [sym_preproc_line] = STATE(912),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3945),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3947),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(913)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(913),
    [sym_compiler_directive_decl] = STATE(913),
    [sym_fsi_directive_decl] = STATE(913),
    [sym_preproc_line] = STATE(913),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3949),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3951),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(914)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(914),
    [sym_compiler_directive_decl] = STATE(914),
    [sym_fsi_directive_decl] = STATE(914),
    [sym_preproc_line] = STATE(914),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3953),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(915)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(915),
    [sym_compiler_directive_decl] = STATE(915),
    [sym_fsi_directive_decl] = STATE(915),
    [sym_preproc_line] = STATE(915),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3957),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3959),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(916)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(646),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1809),
    [sym_infix_op] = STATE(2127),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(916),
    [sym_compiler_directive_decl] = STATE(916),
    [sym_fsi_directive_decl] = STATE(916),
    [sym_preproc_line] = STATE(916),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_sequential_expression_repeat1] = STATE(5622),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_COLON] = ACTIONS(2888),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_COMMA] = ACTIONS(2894),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2888),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2900),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(2904),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2906),
    [anon_sym_DOT] = ACTIONS(2908),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(3961),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_COLON_GT] = ACTIONS(2918),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_DOT_LBRACK] = ACTIONS(2936),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_LPAREN2] = ACTIONS(2944),
    [anon_sym_PIPE_PIPE] = ACTIONS(2906),
    [anon_sym_AMP_AMP] = ACTIONS(2946),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(2962),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__newline] = ACTIONS(3963),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
    [sym__tyapp_open] = ACTIONS(2978),
  },
  [STATE(917)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(917),
    [sym_compiler_directive_decl] = STATE(917),
    [sym_fsi_directive_decl] = STATE(917),
    [sym_preproc_line] = STATE(917),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(918)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(918),
    [sym_compiler_directive_decl] = STATE(918),
    [sym_fsi_directive_decl] = STATE(918),
    [sym_preproc_line] = STATE(918),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(919)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(919),
    [sym_compiler_directive_decl] = STATE(919),
    [sym_fsi_directive_decl] = STATE(919),
    [sym_preproc_line] = STATE(919),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(298),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(298),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(298),
    [anon_sym_AMP_AMP] = ACTIONS(298),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(300),
    [sym__dedent] = ACTIONS(300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(920)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(920),
    [sym_compiler_directive_decl] = STATE(920),
    [sym_fsi_directive_decl] = STATE(920),
    [sym_preproc_line] = STATE(920),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(921)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(921),
    [sym_compiler_directive_decl] = STATE(921),
    [sym_fsi_directive_decl] = STATE(921),
    [sym_preproc_line] = STATE(921),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(302),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(302),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(304),
    [sym__dedent] = ACTIONS(304),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(922)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(922),
    [sym_compiler_directive_decl] = STATE(922),
    [sym_fsi_directive_decl] = STATE(922),
    [sym_preproc_line] = STATE(922),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3337),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(923)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(923),
    [sym_compiler_directive_decl] = STATE(923),
    [sym_fsi_directive_decl] = STATE(923),
    [sym_preproc_line] = STATE(923),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(306),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(306),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(306),
    [anon_sym_AMP_AMP] = ACTIONS(306),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(308),
    [sym__dedent] = ACTIONS(308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(924)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(924),
    [sym_compiler_directive_decl] = STATE(924),
    [sym_fsi_directive_decl] = STATE(924),
    [sym_preproc_line] = STATE(924),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(312),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(925)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(925),
    [sym_compiler_directive_decl] = STATE(925),
    [sym_fsi_directive_decl] = STATE(925),
    [sym_preproc_line] = STATE(925),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(159),
    [anon_sym_null] = ACTIONS(159),
    [anon_sym_EQ] = ACTIONS(159),
    [anon_sym_begin] = ACTIONS(159),
    [anon_sym_COLON] = ACTIONS(159),
    [anon_sym_do] = ACTIONS(159),
    [anon_sym_let] = ACTIONS(159),
    [anon_sym_let_BANG] = ACTIONS(157),
    [anon_sym_LPAREN] = ACTIONS(159),
    [anon_sym_COMMA] = ACTIONS(157),
    [anon_sym_QMARK] = ACTIONS(159),
    [anon_sym_COLON_QMARK] = ACTIONS(159),
    [anon_sym_COLON_COLON] = ACTIONS(157),
    [anon_sym_AMP] = ACTIONS(159),
    [anon_sym_LBRACK] = ACTIONS(159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(157),
    [anon_sym_LBRACE] = ACTIONS(159),
    [anon_sym_LT_AT] = ACTIONS(159),
    [anon_sym_LT_AT_AT] = ACTIONS(159),
    [anon_sym_CARET] = ACTIONS(159),
    [anon_sym_or] = ACTIONS(159),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(157),
    [anon_sym_new] = ACTIONS(159),
    [anon_sym_return] = ACTIONS(159),
    [anon_sym_return_BANG] = ACTIONS(157),
    [anon_sym_yield] = ACTIONS(159),
    [anon_sym_yield_BANG] = ACTIONS(157),
    [anon_sym_lazy] = ACTIONS(159),
    [anon_sym_assert] = ACTIONS(159),
    [anon_sym_upcast] = ACTIONS(159),
    [anon_sym_downcast] = ACTIONS(159),
    [anon_sym_fixed] = ACTIONS(159),
    [anon_sym_COLON_GT] = ACTIONS(157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(157),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_while] = ACTIONS(159),
    [anon_sym_while_BANG] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_fun] = ACTIONS(159),
    [anon_sym_try] = ACTIONS(159),
    [anon_sym_match] = ACTIONS(159),
    [anon_sym_match_BANG] = ACTIONS(157),
    [anon_sym_function] = ACTIONS(159),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_use_BANG] = ACTIONS(157),
    [anon_sym_do_BANG] = ACTIONS(157),
    [anon_sym_DOT_DOT] = ACTIONS(157),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(159),
    [anon_sym_AMP_AMP] = ACTIONS(159),
    [aux_sym_char_token1] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(159),
    [anon_sym_DQUOTE] = ACTIONS(159),
    [anon_sym_AT_DQUOTE] = ACTIONS(157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(157),
    [sym_bool] = ACTIONS(159),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(159),
    [sym_op_identifier] = ACTIONS(159),
    [anon_sym_PLUS] = ACTIONS(159),
    [anon_sym_DASH] = ACTIONS(159),
    [anon_sym_PLUS_DOT] = ACTIONS(159),
    [anon_sym_DASH_DOT] = ACTIONS(159),
    [anon_sym_PERCENT] = ACTIONS(159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(159),
    [anon_sym_TILDE] = ACTIONS(157),
    [aux_sym_prefix_op_token1] = ACTIONS(159),
    [aux_sym_infix_op_token1] = ACTIONS(157),
    [aux_sym_infix_op_token2] = ACTIONS(159),
    [aux_sym_infix_op_token3] = ACTIONS(159),
    [anon_sym_BANG_EQ] = ACTIONS(159),
    [anon_sym_COLON_EQ] = ACTIONS(157),
    [anon_sym_DOLLAR] = ACTIONS(159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(159),
    [sym_int] = ACTIONS(159),
    [sym_xint] = ACTIONS(157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(157),
    [sym__newline] = ACTIONS(157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(157),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(926)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(926),
    [sym_compiler_directive_decl] = STATE(926),
    [sym_fsi_directive_decl] = STATE(926),
    [sym_preproc_line] = STATE(926),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(927)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(927),
    [sym_compiler_directive_decl] = STATE(927),
    [sym_fsi_directive_decl] = STATE(927),
    [sym_preproc_line] = STATE(927),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(5429),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(3198),
    [sym__dedent] = ACTIONS(3967),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(928)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(925),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1605),
    [sym_infix_op] = STATE(1658),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(928),
    [sym_compiler_directive_decl] = STATE(928),
    [sym_fsi_directive_decl] = STATE(928),
    [sym_preproc_line] = STATE(928),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_sequential_expression_repeat1] = STATE(5502),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(187),
    [anon_sym_null] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_begin] = ACTIONS(187),
    [anon_sym_COLON] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_let] = ACTIONS(187),
    [anon_sym_let_BANG] = ACTIONS(185),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(185),
    [anon_sym_QMARK] = ACTIONS(187),
    [anon_sym_COLON_QMARK] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(187),
    [anon_sym_LBRACK_PIPE] = ACTIONS(185),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LT_AT] = ACTIONS(187),
    [anon_sym_LT_AT_AT] = ACTIONS(187),
    [anon_sym_CARET] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(2804),
    [anon_sym_LBRACE_PIPE] = ACTIONS(185),
    [anon_sym_new] = ACTIONS(187),
    [anon_sym_return] = ACTIONS(187),
    [anon_sym_return_BANG] = ACTIONS(185),
    [anon_sym_yield] = ACTIONS(187),
    [anon_sym_yield_BANG] = ACTIONS(185),
    [anon_sym_lazy] = ACTIONS(187),
    [anon_sym_assert] = ACTIONS(187),
    [anon_sym_upcast] = ACTIONS(187),
    [anon_sym_downcast] = ACTIONS(187),
    [anon_sym_fixed] = ACTIONS(187),
    [anon_sym_COLON_GT] = ACTIONS(185),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(185),
    [anon_sym_for] = ACTIONS(187),
    [anon_sym_while] = ACTIONS(187),
    [anon_sym_while_BANG] = ACTIONS(185),
    [anon_sym_if] = ACTIONS(187),
    [anon_sym_fun] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(187),
    [anon_sym_match] = ACTIONS(187),
    [anon_sym_match_BANG] = ACTIONS(185),
    [anon_sym_function] = ACTIONS(187),
    [anon_sym_DOT_LBRACK] = ACTIONS(2830),
    [anon_sym_use] = ACTIONS(187),
    [anon_sym_use_BANG] = ACTIONS(185),
    [anon_sym_do_BANG] = ACTIONS(185),
    [anon_sym_DOT_DOT] = ACTIONS(185),
    [anon_sym_LPAREN2] = ACTIONS(2838),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [aux_sym_char_token1] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(187),
    [anon_sym_DQUOTE] = ACTIONS(187),
    [anon_sym_AT_DQUOTE] = ACTIONS(185),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(185),
    [sym_bool] = ACTIONS(187),
    [sym_unit] = ACTIONS(2857),
    [anon_sym_LPAREN_PIPE] = ACTIONS(187),
    [sym_op_identifier] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_PLUS_DOT] = ACTIONS(187),
    [anon_sym_DASH_DOT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(187),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(185),
    [aux_sym_prefix_op_token1] = ACTIONS(187),
    [aux_sym_infix_op_token1] = ACTIONS(185),
    [aux_sym_infix_op_token2] = ACTIONS(187),
    [aux_sym_infix_op_token3] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_COLON_EQ] = ACTIONS(185),
    [anon_sym_DOLLAR] = ACTIONS(187),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(187),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(187),
    [sym_int] = ACTIONS(187),
    [sym_xint] = ACTIONS(185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(185),
    [sym__newline] = ACTIONS(185),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(185),
    [sym__tyapp_open] = ACTIONS(2871),
  },
  [STATE(929)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(511),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1618),
    [sym_infix_op] = STATE(1735),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(929),
    [sym_compiler_directive_decl] = STATE(929),
    [sym_fsi_directive_decl] = STATE(929),
    [sym_preproc_line] = STATE(929),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_sequential_expression_repeat1] = STATE(5698),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_COLON] = ACTIONS(2590),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_COMMA] = ACTIONS(2596),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(2602),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(2606),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(2608),
    [anon_sym_DOT] = ACTIONS(2610),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_COLON_GT] = ACTIONS(2618),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(2618),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_DOT_LBRACK] = ACTIONS(2636),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_LPAREN2] = ACTIONS(2644),
    [anon_sym_PIPE_PIPE] = ACTIONS(2608),
    [anon_sym_AMP_AMP] = ACTIONS(2646),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(2662),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(2666),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__newline] = ACTIONS(2674),
    [sym__quoted_close] = ACTIONS(3080),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
    [sym__tyapp_open] = ACTIONS(2680),
  },
  [STATE(930)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(930),
    [sym_compiler_directive_decl] = STATE(930),
    [sym_fsi_directive_decl] = STATE(930),
    [sym_preproc_line] = STATE(930),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3969),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(931)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(931),
    [sym_compiler_directive_decl] = STATE(931),
    [sym_fsi_directive_decl] = STATE(931),
    [sym_preproc_line] = STATE(931),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3971),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(932)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(932),
    [sym_compiler_directive_decl] = STATE(932),
    [sym_fsi_directive_decl] = STATE(932),
    [sym_preproc_line] = STATE(932),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3973),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(933)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(933),
    [sym_compiler_directive_decl] = STATE(933),
    [sym_fsi_directive_decl] = STATE(933),
    [sym_preproc_line] = STATE(933),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3975),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(934)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(198),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1717),
    [sym_infix_op] = STATE(1710),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(934),
    [sym_compiler_directive_decl] = STATE(934),
    [sym_fsi_directive_decl] = STATE(934),
    [sym_preproc_line] = STATE(934),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_COLON] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_COLON_GT] = ACTIONS(1356),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_DOT_LBRACK] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1384),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(1400),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
    [sym__tyapp_open] = ACTIONS(1320),
  },
  [STATE(935)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(935),
    [sym_compiler_directive_decl] = STATE(935),
    [sym_fsi_directive_decl] = STATE(935),
    [sym_preproc_line] = STATE(935),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3977),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(936)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(936),
    [sym_compiler_directive_decl] = STATE(936),
    [sym_fsi_directive_decl] = STATE(936),
    [sym_preproc_line] = STATE(936),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3979),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(937)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(937),
    [sym_compiler_directive_decl] = STATE(937),
    [sym_fsi_directive_decl] = STATE(937),
    [sym_preproc_line] = STATE(937),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3981),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(938)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(938),
    [sym_compiler_directive_decl] = STATE(938),
    [sym_fsi_directive_decl] = STATE(938),
    [sym_preproc_line] = STATE(938),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3983),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(939)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1109),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(970),
    [sym_literal_expression] = STATE(4057),
    [sym_srtp_call_expression] = STATE(4057),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3843),
    [sym_tuple_expression] = STATE(4057),
    [sym_brace_expression] = STATE(4057),
    [sym_anon_record_expression] = STATE(4057),
    [sym_prefixed_expression] = STATE(4057),
    [sym_typecast_expression] = STATE(4057),
    [sym_for_expression] = STATE(4057),
    [sym_while_expression] = STATE(4057),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4057),
    [sym_fun_expression] = STATE(4057),
    [sym_try_expression] = STATE(4057),
    [sym_match_expression] = STATE(4057),
    [sym_function_expression] = STATE(4057),
    [sym_mutate_expression] = STATE(4057),
    [sym_index_expression] = STATE(3843),
    [sym_typed_expression] = STATE(4057),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4057),
    [sym_list_expression] = STATE(4057),
    [sym_array_expression] = STATE(4057),
    [sym_begin_end_expression] = STATE(4057),
    [sym_paren_expression] = STATE(4057),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4057),
    [sym_dot_expression] = STATE(3843),
    [sym_infix_expression] = STATE(4057),
    [sym_ce_expression] = STATE(4057),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4057),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(939),
    [sym_compiler_directive_decl] = STATE(939),
    [sym_fsi_directive_decl] = STATE(939),
    [sym_preproc_line] = STATE(939),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(21444),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(992),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(3985),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3987),
    [anon_sym_POUNDendif] = ACTIONS(3989),
    [anon_sym_POUNDelse] = ACTIONS(3991),
    [sym__newline] = ACTIONS(3993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(940)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(940),
    [sym_compiler_directive_decl] = STATE(940),
    [sym_fsi_directive_decl] = STATE(940),
    [sym_preproc_line] = STATE(940),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3995),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(941)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(941),
    [sym_compiler_directive_decl] = STATE(941),
    [sym_fsi_directive_decl] = STATE(941),
    [sym_preproc_line] = STATE(941),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3997),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(942)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1109),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(970),
    [sym_literal_expression] = STATE(4057),
    [sym_srtp_call_expression] = STATE(4057),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3843),
    [sym_tuple_expression] = STATE(4057),
    [sym_brace_expression] = STATE(4057),
    [sym_anon_record_expression] = STATE(4057),
    [sym_prefixed_expression] = STATE(4057),
    [sym_typecast_expression] = STATE(4057),
    [sym_for_expression] = STATE(4057),
    [sym_while_expression] = STATE(4057),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4057),
    [sym_fun_expression] = STATE(4057),
    [sym_try_expression] = STATE(4057),
    [sym_match_expression] = STATE(4057),
    [sym_function_expression] = STATE(4057),
    [sym_mutate_expression] = STATE(4057),
    [sym_index_expression] = STATE(3843),
    [sym_typed_expression] = STATE(4057),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4057),
    [sym_list_expression] = STATE(4057),
    [sym_array_expression] = STATE(4057),
    [sym_begin_end_expression] = STATE(4057),
    [sym_paren_expression] = STATE(4057),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4057),
    [sym_dot_expression] = STATE(3843),
    [sym_infix_expression] = STATE(4057),
    [sym_ce_expression] = STATE(4057),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4057),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(942),
    [sym_compiler_directive_decl] = STATE(942),
    [sym_fsi_directive_decl] = STATE(942),
    [sym_preproc_line] = STATE(942),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(21988),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(992),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(3985),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3987),
    [anon_sym_POUNDendif] = ACTIONS(3999),
    [anon_sym_POUNDelse] = ACTIONS(3991),
    [sym__newline] = ACTIONS(3993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(943)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(943),
    [sym_compiler_directive_decl] = STATE(943),
    [sym_fsi_directive_decl] = STATE(943),
    [sym_preproc_line] = STATE(943),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4001),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(944)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(944),
    [sym_compiler_directive_decl] = STATE(944),
    [sym_fsi_directive_decl] = STATE(944),
    [sym_preproc_line] = STATE(944),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4003),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(945)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(945),
    [sym_compiler_directive_decl] = STATE(945),
    [sym_fsi_directive_decl] = STATE(945),
    [sym_preproc_line] = STATE(945),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4005),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(946)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(946),
    [sym_compiler_directive_decl] = STATE(946),
    [sym_fsi_directive_decl] = STATE(946),
    [sym_preproc_line] = STATE(946),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4007),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(947)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(947),
    [sym_compiler_directive_decl] = STATE(947),
    [sym_fsi_directive_decl] = STATE(947),
    [sym_preproc_line] = STATE(947),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4009),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(948)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(948),
    [sym_compiler_directive_decl] = STATE(948),
    [sym_fsi_directive_decl] = STATE(948),
    [sym_preproc_line] = STATE(948),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4011),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(949)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(949),
    [sym_compiler_directive_decl] = STATE(949),
    [sym_fsi_directive_decl] = STATE(949),
    [sym_preproc_line] = STATE(949),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4013),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(950)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(950),
    [sym_compiler_directive_decl] = STATE(950),
    [sym_fsi_directive_decl] = STATE(950),
    [sym_preproc_line] = STATE(950),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4015),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(951)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(714),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1544),
    [sym_infix_op] = STATE(1595),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(951),
    [sym_compiler_directive_decl] = STATE(951),
    [sym_fsi_directive_decl] = STATE(951),
    [sym_preproc_line] = STATE(951),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_COLON] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(3180),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1621),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1633),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(1637),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(3182),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1649),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_DOT_LBRACK] = ACTIONS(1667),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_LPAREN2] = ACTIONS(1677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3182),
    [anon_sym_AMP_AMP] = ACTIONS(3194),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(1697),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(1701),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
    [sym__tyapp_open] = ACTIONS(1715),
  },
  [STATE(952)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(952),
    [sym_compiler_directive_decl] = STATE(952),
    [sym_fsi_directive_decl] = STATE(952),
    [sym_preproc_line] = STATE(952),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4017),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(953)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(953),
    [sym_compiler_directive_decl] = STATE(953),
    [sym_fsi_directive_decl] = STATE(953),
    [sym_preproc_line] = STATE(953),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4019),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(954)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(954),
    [sym_compiler_directive_decl] = STATE(954),
    [sym_fsi_directive_decl] = STATE(954),
    [sym_preproc_line] = STATE(954),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4021),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(955)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(955),
    [sym_compiler_directive_decl] = STATE(955),
    [sym_fsi_directive_decl] = STATE(955),
    [sym_preproc_line] = STATE(955),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4023),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(956)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(956),
    [sym_compiler_directive_decl] = STATE(956),
    [sym_fsi_directive_decl] = STATE(956),
    [sym_preproc_line] = STATE(956),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4025),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(957)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(957),
    [sym_compiler_directive_decl] = STATE(957),
    [sym_fsi_directive_decl] = STATE(957),
    [sym_preproc_line] = STATE(957),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4027),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(958)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(958),
    [sym_compiler_directive_decl] = STATE(958),
    [sym_fsi_directive_decl] = STATE(958),
    [sym_preproc_line] = STATE(958),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4029),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(959)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(959),
    [sym_compiler_directive_decl] = STATE(959),
    [sym_fsi_directive_decl] = STATE(959),
    [sym_preproc_line] = STATE(959),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4031),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(960)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(960),
    [sym_compiler_directive_decl] = STATE(960),
    [sym_fsi_directive_decl] = STATE(960),
    [sym_preproc_line] = STATE(960),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4033),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(961)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(961),
    [sym_compiler_directive_decl] = STATE(961),
    [sym_fsi_directive_decl] = STATE(961),
    [sym_preproc_line] = STATE(961),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4035),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(962)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(962),
    [sym_compiler_directive_decl] = STATE(962),
    [sym_fsi_directive_decl] = STATE(962),
    [sym_preproc_line] = STATE(962),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4037),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(963)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(963),
    [sym_compiler_directive_decl] = STATE(963),
    [sym_fsi_directive_decl] = STATE(963),
    [sym_preproc_line] = STATE(963),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4039),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(964)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(964),
    [sym_compiler_directive_decl] = STATE(964),
    [sym_fsi_directive_decl] = STATE(964),
    [sym_preproc_line] = STATE(964),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4041),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(965)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(965),
    [sym_compiler_directive_decl] = STATE(965),
    [sym_fsi_directive_decl] = STATE(965),
    [sym_preproc_line] = STATE(965),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4043),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(966)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(966),
    [sym_compiler_directive_decl] = STATE(966),
    [sym_fsi_directive_decl] = STATE(966),
    [sym_preproc_line] = STATE(966),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4045),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(967)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(967),
    [sym_compiler_directive_decl] = STATE(967),
    [sym_fsi_directive_decl] = STATE(967),
    [sym_preproc_line] = STATE(967),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4047),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(968)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(968),
    [sym_compiler_directive_decl] = STATE(968),
    [sym_fsi_directive_decl] = STATE(968),
    [sym_preproc_line] = STATE(968),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4049),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(969)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(969),
    [sym_compiler_directive_decl] = STATE(969),
    [sym_fsi_directive_decl] = STATE(969),
    [sym_preproc_line] = STATE(969),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(3436),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(970)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(189),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1707),
    [sym_infix_op] = STATE(1698),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(970),
    [sym_compiler_directive_decl] = STATE(970),
    [sym_fsi_directive_decl] = STATE(970),
    [sym_preproc_line] = STATE(970),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_COLON] = ACTIONS(1161),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1167),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1161),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(1173),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(1177),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1179),
    [anon_sym_DOT] = ACTIONS(1181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_COLON_GT] = ACTIONS(1189),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_DOT_LBRACK] = ACTIONS(1207),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_LPAREN2] = ACTIONS(1215),
    [anon_sym_PIPE_PIPE] = ACTIONS(1179),
    [anon_sym_AMP_AMP] = ACTIONS(1217),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(1233),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(1237),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
    [sym__tyapp_open] = ACTIONS(1249),
  },
  [STATE(971)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(971),
    [sym_compiler_directive_decl] = STATE(971),
    [sym_fsi_directive_decl] = STATE(971),
    [sym_preproc_line] = STATE(971),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4051),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(972)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(972),
    [sym_compiler_directive_decl] = STATE(972),
    [sym_fsi_directive_decl] = STATE(972),
    [sym_preproc_line] = STATE(972),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4053),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(973)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(973),
    [sym_compiler_directive_decl] = STATE(973),
    [sym_fsi_directive_decl] = STATE(973),
    [sym_preproc_line] = STATE(973),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4055),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(974)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(974),
    [sym_compiler_directive_decl] = STATE(974),
    [sym_fsi_directive_decl] = STATE(974),
    [sym_preproc_line] = STATE(974),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4057),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(975)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(975),
    [sym_compiler_directive_decl] = STATE(975),
    [sym_fsi_directive_decl] = STATE(975),
    [sym_preproc_line] = STATE(975),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4059),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(976)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(976),
    [sym_compiler_directive_decl] = STATE(976),
    [sym_fsi_directive_decl] = STATE(976),
    [sym_preproc_line] = STATE(976),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4061),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(977)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(977),
    [sym_compiler_directive_decl] = STATE(977),
    [sym_fsi_directive_decl] = STATE(977),
    [sym_preproc_line] = STATE(977),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4063),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(978)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(978),
    [sym_compiler_directive_decl] = STATE(978),
    [sym_fsi_directive_decl] = STATE(978),
    [sym_preproc_line] = STATE(978),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4065),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(979)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(979),
    [sym_compiler_directive_decl] = STATE(979),
    [sym_fsi_directive_decl] = STATE(979),
    [sym_preproc_line] = STATE(979),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4067),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(980)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(980),
    [sym_compiler_directive_decl] = STATE(980),
    [sym_fsi_directive_decl] = STATE(980),
    [sym_preproc_line] = STATE(980),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4069),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(981)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(981),
    [sym_compiler_directive_decl] = STATE(981),
    [sym_fsi_directive_decl] = STATE(981),
    [sym_preproc_line] = STATE(981),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4071),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(982)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(982),
    [sym_compiler_directive_decl] = STATE(982),
    [sym_fsi_directive_decl] = STATE(982),
    [sym_preproc_line] = STATE(982),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4073),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(983)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(983),
    [sym_compiler_directive_decl] = STATE(983),
    [sym_fsi_directive_decl] = STATE(983),
    [sym_preproc_line] = STATE(983),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4075),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(984)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(984),
    [sym_compiler_directive_decl] = STATE(984),
    [sym_fsi_directive_decl] = STATE(984),
    [sym_preproc_line] = STATE(984),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4077),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(985)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(985),
    [sym_compiler_directive_decl] = STATE(985),
    [sym_fsi_directive_decl] = STATE(985),
    [sym_preproc_line] = STATE(985),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4079),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(986)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(986),
    [sym_compiler_directive_decl] = STATE(986),
    [sym_fsi_directive_decl] = STATE(986),
    [sym_preproc_line] = STATE(986),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4081),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(987)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(987),
    [sym_compiler_directive_decl] = STATE(987),
    [sym_fsi_directive_decl] = STATE(987),
    [sym_preproc_line] = STATE(987),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4083),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(988)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(988),
    [sym_compiler_directive_decl] = STATE(988),
    [sym_fsi_directive_decl] = STATE(988),
    [sym_preproc_line] = STATE(988),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4085),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(989)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1109),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(970),
    [sym_literal_expression] = STATE(4057),
    [sym_srtp_call_expression] = STATE(4057),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3843),
    [sym_tuple_expression] = STATE(4057),
    [sym_brace_expression] = STATE(4057),
    [sym_anon_record_expression] = STATE(4057),
    [sym_prefixed_expression] = STATE(4057),
    [sym_typecast_expression] = STATE(4057),
    [sym_for_expression] = STATE(4057),
    [sym_while_expression] = STATE(4057),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4057),
    [sym_fun_expression] = STATE(4057),
    [sym_try_expression] = STATE(4057),
    [sym_match_expression] = STATE(4057),
    [sym_function_expression] = STATE(4057),
    [sym_mutate_expression] = STATE(4057),
    [sym_index_expression] = STATE(3843),
    [sym_typed_expression] = STATE(4057),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4057),
    [sym_list_expression] = STATE(4057),
    [sym_array_expression] = STATE(4057),
    [sym_begin_end_expression] = STATE(4057),
    [sym_paren_expression] = STATE(4057),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4057),
    [sym_dot_expression] = STATE(3843),
    [sym_infix_expression] = STATE(4057),
    [sym_ce_expression] = STATE(4057),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4057),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(989),
    [sym_compiler_directive_decl] = STATE(989),
    [sym_fsi_directive_decl] = STATE(989),
    [sym_preproc_line] = STATE(989),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [sym_preproc_else_in_module_body] = STATE(20971),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(992),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(3985),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3987),
    [anon_sym_POUNDendif] = ACTIONS(4087),
    [anon_sym_POUNDelse] = ACTIONS(3991),
    [sym__newline] = ACTIONS(3993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(990)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(223),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8892),
    [sym_prefix_op] = STATE(1916),
    [sym_infix_op] = STATE(2064),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(990),
    [sym_compiler_directive_decl] = STATE(990),
    [sym_fsi_directive_decl] = STATE(990),
    [sym_preproc_line] = STATE(990),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_sequential_expression_repeat1] = STATE(4360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(131),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_COLON] = ACTIONS(1425),
    [anon_sym_do] = ACTIONS(4089),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(1431),
    [anon_sym_QMARK] = ACTIONS(131),
    [anon_sym_COLON_QMARK] = ACTIONS(1425),
    [anon_sym_COLON_COLON] = ACTIONS(139),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(1437),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(1441),
    [anon_sym_CARET] = ACTIONS(219),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1445),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_COLON_GT] = ACTIONS(1453),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(1453),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_DOT_LBRACK] = ACTIONS(1471),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_LPAREN2] = ACTIONS(1479),
    [anon_sym_PIPE_PIPE] = ACTIONS(1443),
    [anon_sym_AMP_AMP] = ACTIONS(1481),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(1497),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(1501),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(45),
    [anon_sym_DASH_DOT] = ACTIONS(45),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(270),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(270),
    [aux_sym_infix_op_token1] = ACTIONS(139),
    [aux_sym_infix_op_token2] = ACTIONS(131),
    [aux_sym_infix_op_token3] = ACTIONS(131),
    [anon_sym_BANG_EQ] = ACTIONS(131),
    [anon_sym_COLON_EQ] = ACTIONS(139),
    [anon_sym_DOLLAR] = ACTIONS(131),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(131),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(131),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(1509),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
    [sym__tyapp_open] = ACTIONS(1513),
  },
  [STATE(991)] = {
    [sym__module_body_elem] = STATE(8691),
    [sym__module_expression] = STATE(8690),
    [sym_module_abbrev] = STATE(8690),
    [sym_module_defn] = STATE(8690),
    [sym_import_decl] = STATE(8690),
    [sym_attributes] = STATE(15049),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8692),
    [sym_do] = STATE(8637),
    [sym_function_or_value_defn] = STATE(1117),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(537),
    [sym_literal_expression] = STATE(4185),
    [sym_srtp_call_expression] = STATE(4185),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(4058),
    [sym_tuple_expression] = STATE(4185),
    [sym_brace_expression] = STATE(4185),
    [sym_anon_record_expression] = STATE(4185),
    [sym_prefixed_expression] = STATE(4185),
    [sym_typecast_expression] = STATE(4185),
    [sym_for_expression] = STATE(4185),
    [sym_while_expression] = STATE(4185),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4185),
    [sym_fun_expression] = STATE(4185),
    [sym_try_expression] = STATE(4185),
    [sym_match_expression] = STATE(4185),
    [sym_function_expression] = STATE(4185),
    [sym_mutate_expression] = STATE(4185),
    [sym_index_expression] = STATE(4058),
    [sym_typed_expression] = STATE(4185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(4185),
    [sym_list_expression] = STATE(4185),
    [sym_array_expression] = STATE(4185),
    [sym_begin_end_expression] = STATE(4185),
    [sym_paren_expression] = STATE(4185),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4185),
    [sym_dot_expression] = STATE(4058),
    [sym_infix_expression] = STATE(4185),
    [sym_ce_expression] = STATE(4185),
    [sym_sequential_expression] = STATE(6185),
    [sym_exception_definition] = STATE(8690),
    [sym_type_definition] = STATE(8690),
    [sym_extern_binding] = STATE(8690),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4185),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(991),
    [sym_compiler_directive_decl] = STATE(991),
    [sym_fsi_directive_decl] = STATE(991),
    [sym_preproc_line] = STATE(991),
    [sym_preproc_if_in_expression] = STATE(6185),
    [sym_preproc_if_in_module_body] = STATE(8690),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1326),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(1011),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4091),
    [anon_sym_null] = ACTIONS(4093),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4095),
    [anon_sym_type] = ACTIONS(4097),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4099),
    [anon_sym_let] = ACTIONS(4101),
    [anon_sym_let_BANG] = ACTIONS(4103),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4109),
    [anon_sym_extern] = ACTIONS(4111),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4117),
    [anon_sym_POUNDload] = ACTIONS(4117),
    [anon_sym_POUNDtime] = ACTIONS(4117),
    [anon_sym_POUNDI] = ACTIONS(4117),
    [anon_sym_POUNDhelp] = ACTIONS(4117),
    [anon_sym_POUNDquit] = ACTIONS(4117),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4119),
    [anon_sym_POUNDendif] = ACTIONS(4121),
    [sym__newline] = ACTIONS(4124),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(992)] = {
    [sym__module_body_elem] = STATE(8328),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1109),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(970),
    [sym_literal_expression] = STATE(4057),
    [sym_srtp_call_expression] = STATE(4057),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3843),
    [sym_tuple_expression] = STATE(4057),
    [sym_brace_expression] = STATE(4057),
    [sym_anon_record_expression] = STATE(4057),
    [sym_prefixed_expression] = STATE(4057),
    [sym_typecast_expression] = STATE(4057),
    [sym_for_expression] = STATE(4057),
    [sym_while_expression] = STATE(4057),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4057),
    [sym_fun_expression] = STATE(4057),
    [sym_try_expression] = STATE(4057),
    [sym_match_expression] = STATE(4057),
    [sym_function_expression] = STATE(4057),
    [sym_mutate_expression] = STATE(4057),
    [sym_index_expression] = STATE(3843),
    [sym_typed_expression] = STATE(4057),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4057),
    [sym_list_expression] = STATE(4057),
    [sym_array_expression] = STATE(4057),
    [sym_begin_end_expression] = STATE(4057),
    [sym_paren_expression] = STATE(4057),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4057),
    [sym_dot_expression] = STATE(3843),
    [sym_infix_expression] = STATE(4057),
    [sym_ce_expression] = STATE(4057),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4057),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(992),
    [sym_compiler_directive_decl] = STATE(992),
    [sym_fsi_directive_decl] = STATE(992),
    [sym_preproc_line] = STATE(992),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(992),
    [sym_identifier] = ACTIONS(4126),
    [anon_sym_module] = ACTIONS(4129),
    [anon_sym_null] = ACTIONS(4132),
    [anon_sym_begin] = ACTIONS(4135),
    [anon_sym_open] = ACTIONS(4138),
    [anon_sym_type] = ACTIONS(4141),
    [anon_sym_LBRACK_LT] = ACTIONS(4144),
    [anon_sym_do] = ACTIONS(4147),
    [anon_sym_let] = ACTIONS(4150),
    [anon_sym_let_BANG] = ACTIONS(4153),
    [anon_sym_LPAREN] = ACTIONS(4156),
    [anon_sym_AMP] = ACTIONS(4159),
    [anon_sym_LBRACK] = ACTIONS(4162),
    [anon_sym_LBRACK_PIPE] = ACTIONS(4165),
    [anon_sym_LBRACE] = ACTIONS(4168),
    [anon_sym_LT_AT] = ACTIONS(4171),
    [anon_sym_LT_AT_AT] = ACTIONS(4174),
    [anon_sym_CARET] = ACTIONS(4177),
    [anon_sym_LBRACE_PIPE] = ACTIONS(4180),
    [anon_sym_new] = ACTIONS(4183),
    [anon_sym_return] = ACTIONS(4183),
    [anon_sym_return_BANG] = ACTIONS(4186),
    [anon_sym_yield] = ACTIONS(4183),
    [anon_sym_yield_BANG] = ACTIONS(4186),
    [anon_sym_lazy] = ACTIONS(4183),
    [anon_sym_assert] = ACTIONS(4183),
    [anon_sym_upcast] = ACTIONS(4183),
    [anon_sym_downcast] = ACTIONS(4183),
    [anon_sym_fixed] = ACTIONS(4183),
    [anon_sym_for] = ACTIONS(4189),
    [anon_sym_while] = ACTIONS(4192),
    [anon_sym_while_BANG] = ACTIONS(4195),
    [anon_sym_if] = ACTIONS(4198),
    [anon_sym_fun] = ACTIONS(4201),
    [anon_sym_try] = ACTIONS(4204),
    [anon_sym_match] = ACTIONS(4207),
    [anon_sym_match_BANG] = ACTIONS(4210),
    [anon_sym_function] = ACTIONS(4213),
    [anon_sym_use] = ACTIONS(4216),
    [anon_sym_use_BANG] = ACTIONS(4219),
    [anon_sym_do_BANG] = ACTIONS(4222),
    [anon_sym_AMP_AMP] = ACTIONS(4225),
    [anon_sym_exception] = ACTIONS(4228),
    [anon_sym_extern] = ACTIONS(4231),
    [aux_sym_char_token1] = ACTIONS(4234),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(4237),
    [anon_sym_DQUOTE] = ACTIONS(4240),
    [anon_sym_AT_DQUOTE] = ACTIONS(4243),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4246),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4249),
    [sym_bool] = ACTIONS(4252),
    [sym_unit] = ACTIONS(4255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(4258),
    [sym_op_identifier] = ACTIONS(4261),
    [anon_sym_PLUS] = ACTIONS(4159),
    [anon_sym_DASH] = ACTIONS(4159),
    [anon_sym_PLUS_DOT] = ACTIONS(4264),
    [anon_sym_DASH_DOT] = ACTIONS(4264),
    [anon_sym_PERCENT] = ACTIONS(4159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(4225),
    [anon_sym_TILDE] = ACTIONS(4267),
    [aux_sym_prefix_op_token1] = ACTIONS(4225),
    [sym_int] = ACTIONS(4270),
    [sym_xint] = ACTIONS(4273),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4276),
    [anon_sym_POUNDload] = ACTIONS(4276),
    [anon_sym_POUNDtime] = ACTIONS(4276),
    [anon_sym_POUNDI] = ACTIONS(4276),
    [anon_sym_POUNDhelp] = ACTIONS(4276),
    [anon_sym_POUNDquit] = ACTIONS(4276),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4279),
    [anon_sym_POUNDendif] = ACTIONS(4282),
    [anon_sym_POUNDelse] = ACTIONS(4282),
    [sym__newline] = ACTIONS(4284),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(4287),
  },
  [STATE(993)] = {
    [sym__module_body_elem] = STATE(8691),
    [sym__module_expression] = STATE(8690),
    [sym_module_abbrev] = STATE(8690),
    [sym_module_defn] = STATE(8690),
    [sym_import_decl] = STATE(8690),
    [sym_attributes] = STATE(15049),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8692),
    [sym_do] = STATE(8637),
    [sym_function_or_value_defn] = STATE(1111),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(934),
    [sym_literal_expression] = STATE(4278),
    [sym_srtp_call_expression] = STATE(4278),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(4038),
    [sym_tuple_expression] = STATE(4278),
    [sym_brace_expression] = STATE(4278),
    [sym_anon_record_expression] = STATE(4278),
    [sym_prefixed_expression] = STATE(4278),
    [sym_typecast_expression] = STATE(4278),
    [sym_for_expression] = STATE(4278),
    [sym_while_expression] = STATE(4278),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4278),
    [sym_fun_expression] = STATE(4278),
    [sym_try_expression] = STATE(4278),
    [sym_match_expression] = STATE(4278),
    [sym_function_expression] = STATE(4278),
    [sym_mutate_expression] = STATE(4278),
    [sym_index_expression] = STATE(4038),
    [sym_typed_expression] = STATE(4278),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4278),
    [sym_list_expression] = STATE(4278),
    [sym_array_expression] = STATE(4278),
    [sym_begin_end_expression] = STATE(4278),
    [sym_paren_expression] = STATE(4278),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4278),
    [sym_dot_expression] = STATE(4038),
    [sym_infix_expression] = STATE(4278),
    [sym_ce_expression] = STATE(4278),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8690),
    [sym_type_definition] = STATE(8690),
    [sym_extern_binding] = STATE(8690),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4278),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(993),
    [sym_compiler_directive_decl] = STATE(993),
    [sym_fsi_directive_decl] = STATE(993),
    [sym_preproc_line] = STATE(993),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8690),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(1011),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4091),
    [anon_sym_null] = ACTIONS(4290),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4095),
    [anon_sym_type] = ACTIONS(4097),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4099),
    [anon_sym_let] = ACTIONS(4101),
    [anon_sym_let_BANG] = ACTIONS(4103),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4109),
    [anon_sym_extern] = ACTIONS(4111),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4117),
    [anon_sym_POUNDload] = ACTIONS(4117),
    [anon_sym_POUNDtime] = ACTIONS(4117),
    [anon_sym_POUNDI] = ACTIONS(4117),
    [anon_sym_POUNDhelp] = ACTIONS(4117),
    [anon_sym_POUNDquit] = ACTIONS(4117),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4292),
    [anon_sym_POUNDendif] = ACTIONS(4294),
    [sym__newline] = ACTIONS(4296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(994)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(994),
    [sym_compiler_directive_decl] = STATE(994),
    [sym_fsi_directive_decl] = STATE(994),
    [sym_preproc_line] = STATE(994),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(997),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4298),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(4300),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(4304),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(995)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(995),
    [sym_compiler_directive_decl] = STATE(995),
    [sym_fsi_directive_decl] = STATE(995),
    [sym_preproc_line] = STATE(995),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1002),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4306),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(4308),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(4304),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(996)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(996),
    [sym_compiler_directive_decl] = STATE(996),
    [sym_fsi_directive_decl] = STATE(996),
    [sym_preproc_line] = STATE(996),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(994),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4310),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(4312),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(4304),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(997)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(997),
    [sym_compiler_directive_decl] = STATE(997),
    [sym_fsi_directive_decl] = STATE(997),
    [sym_preproc_line] = STATE(997),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(997),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4314),
    [sym_identifier] = ACTIONS(4316),
    [anon_sym_namespace] = ACTIONS(4319),
    [anon_sym_module] = ACTIONS(4321),
    [anon_sym_null] = ACTIONS(4324),
    [anon_sym_begin] = ACTIONS(4327),
    [anon_sym_open] = ACTIONS(4330),
    [anon_sym_type] = ACTIONS(4333),
    [anon_sym_LBRACK_LT] = ACTIONS(4336),
    [anon_sym_do] = ACTIONS(4339),
    [anon_sym_let] = ACTIONS(4342),
    [anon_sym_let_BANG] = ACTIONS(4345),
    [anon_sym_LPAREN] = ACTIONS(4348),
    [anon_sym_AMP] = ACTIONS(4351),
    [anon_sym_LBRACK] = ACTIONS(4354),
    [anon_sym_LBRACK_PIPE] = ACTIONS(4357),
    [anon_sym_LBRACE] = ACTIONS(4360),
    [anon_sym_LT_AT] = ACTIONS(4363),
    [anon_sym_LT_AT_AT] = ACTIONS(4366),
    [anon_sym_CARET] = ACTIONS(4369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(4372),
    [anon_sym_new] = ACTIONS(4375),
    [anon_sym_return] = ACTIONS(4375),
    [anon_sym_return_BANG] = ACTIONS(4378),
    [anon_sym_yield] = ACTIONS(4375),
    [anon_sym_yield_BANG] = ACTIONS(4378),
    [anon_sym_lazy] = ACTIONS(4375),
    [anon_sym_assert] = ACTIONS(4375),
    [anon_sym_upcast] = ACTIONS(4375),
    [anon_sym_downcast] = ACTIONS(4375),
    [anon_sym_fixed] = ACTIONS(4375),
    [anon_sym_for] = ACTIONS(4381),
    [anon_sym_while] = ACTIONS(4384),
    [anon_sym_while_BANG] = ACTIONS(4387),
    [anon_sym_if] = ACTIONS(4390),
    [anon_sym_fun] = ACTIONS(4393),
    [anon_sym_try] = ACTIONS(4396),
    [anon_sym_match] = ACTIONS(4399),
    [anon_sym_match_BANG] = ACTIONS(4402),
    [anon_sym_function] = ACTIONS(4405),
    [anon_sym_use] = ACTIONS(4408),
    [anon_sym_use_BANG] = ACTIONS(4411),
    [anon_sym_do_BANG] = ACTIONS(4414),
    [anon_sym_AMP_AMP] = ACTIONS(4417),
    [anon_sym_exception] = ACTIONS(4420),
    [anon_sym_extern] = ACTIONS(4423),
    [aux_sym_char_token1] = ACTIONS(4426),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(4429),
    [anon_sym_DQUOTE] = ACTIONS(4432),
    [anon_sym_AT_DQUOTE] = ACTIONS(4435),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4438),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4441),
    [sym_bool] = ACTIONS(4444),
    [sym_unit] = ACTIONS(4447),
    [anon_sym_LPAREN_PIPE] = ACTIONS(4450),
    [sym_op_identifier] = ACTIONS(4453),
    [anon_sym_PLUS] = ACTIONS(4351),
    [anon_sym_DASH] = ACTIONS(4351),
    [anon_sym_PLUS_DOT] = ACTIONS(4456),
    [anon_sym_DASH_DOT] = ACTIONS(4456),
    [anon_sym_PERCENT] = ACTIONS(4351),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(4417),
    [anon_sym_TILDE] = ACTIONS(4459),
    [aux_sym_prefix_op_token1] = ACTIONS(4417),
    [sym_int] = ACTIONS(4462),
    [sym_xint] = ACTIONS(4465),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4468),
    [anon_sym_POUNDload] = ACTIONS(4468),
    [anon_sym_POUNDtime] = ACTIONS(4468),
    [anon_sym_POUNDI] = ACTIONS(4468),
    [anon_sym_POUNDhelp] = ACTIONS(4468),
    [anon_sym_POUNDquit] = ACTIONS(4468),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4471),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(4474),
  },
  [STATE(998)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(998),
    [sym_compiler_directive_decl] = STATE(998),
    [sym_fsi_directive_decl] = STATE(998),
    [sym_preproc_line] = STATE(998),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1007),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4497),
    [anon_sym_POUNDelse] = ACTIONS(4497),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(999)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(999),
    [sym_compiler_directive_decl] = STATE(999),
    [sym_fsi_directive_decl] = STATE(999),
    [sym_preproc_line] = STATE(999),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1005),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4499),
    [anon_sym_POUNDelse] = ACTIONS(4499),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1000)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1000),
    [sym_compiler_directive_decl] = STATE(1000),
    [sym_fsi_directive_decl] = STATE(1000),
    [sym_preproc_line] = STATE(1000),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1005),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4501),
    [anon_sym_POUNDelse] = ACTIONS(4501),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1001)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1001),
    [sym_compiler_directive_decl] = STATE(1001),
    [sym_fsi_directive_decl] = STATE(1001),
    [sym_preproc_line] = STATE(1001),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(997),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4503),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(4505),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(4304),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1002)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1002),
    [sym_compiler_directive_decl] = STATE(1002),
    [sym_fsi_directive_decl] = STATE(1002),
    [sym_preproc_line] = STATE(1002),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(997),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4507),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(4509),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(4304),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1003)] = {
    [sym__module_body_elem] = STATE(8691),
    [sym__module_expression] = STATE(8690),
    [sym_module_abbrev] = STATE(8690),
    [sym_module_defn] = STATE(8690),
    [sym_import_decl] = STATE(8690),
    [sym_attributes] = STATE(15049),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8692),
    [sym_do] = STATE(8637),
    [sym_function_or_value_defn] = STATE(1111),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(934),
    [sym_literal_expression] = STATE(4278),
    [sym_srtp_call_expression] = STATE(4278),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(4038),
    [sym_tuple_expression] = STATE(4278),
    [sym_brace_expression] = STATE(4278),
    [sym_anon_record_expression] = STATE(4278),
    [sym_prefixed_expression] = STATE(4278),
    [sym_typecast_expression] = STATE(4278),
    [sym_for_expression] = STATE(4278),
    [sym_while_expression] = STATE(4278),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4278),
    [sym_fun_expression] = STATE(4278),
    [sym_try_expression] = STATE(4278),
    [sym_match_expression] = STATE(4278),
    [sym_function_expression] = STATE(4278),
    [sym_mutate_expression] = STATE(4278),
    [sym_index_expression] = STATE(4038),
    [sym_typed_expression] = STATE(4278),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4278),
    [sym_list_expression] = STATE(4278),
    [sym_array_expression] = STATE(4278),
    [sym_begin_end_expression] = STATE(4278),
    [sym_paren_expression] = STATE(4278),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4278),
    [sym_dot_expression] = STATE(4038),
    [sym_infix_expression] = STATE(4278),
    [sym_ce_expression] = STATE(4278),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8690),
    [sym_type_definition] = STATE(8690),
    [sym_extern_binding] = STATE(8690),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4278),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1003),
    [sym_compiler_directive_decl] = STATE(1003),
    [sym_fsi_directive_decl] = STATE(1003),
    [sym_preproc_line] = STATE(1003),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8690),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(1003),
    [sym_identifier] = ACTIONS(4511),
    [anon_sym_module] = ACTIONS(4514),
    [anon_sym_null] = ACTIONS(4517),
    [anon_sym_begin] = ACTIONS(4520),
    [anon_sym_open] = ACTIONS(4523),
    [anon_sym_type] = ACTIONS(4526),
    [anon_sym_LBRACK_LT] = ACTIONS(4144),
    [anon_sym_do] = ACTIONS(4529),
    [anon_sym_let] = ACTIONS(4532),
    [anon_sym_let_BANG] = ACTIONS(4535),
    [anon_sym_LPAREN] = ACTIONS(4538),
    [anon_sym_AMP] = ACTIONS(4159),
    [anon_sym_LBRACK] = ACTIONS(4541),
    [anon_sym_LBRACK_PIPE] = ACTIONS(4544),
    [anon_sym_LBRACE] = ACTIONS(4547),
    [anon_sym_LT_AT] = ACTIONS(4550),
    [anon_sym_LT_AT_AT] = ACTIONS(4553),
    [anon_sym_CARET] = ACTIONS(4177),
    [anon_sym_LBRACE_PIPE] = ACTIONS(4556),
    [anon_sym_new] = ACTIONS(4559),
    [anon_sym_return] = ACTIONS(4559),
    [anon_sym_return_BANG] = ACTIONS(4562),
    [anon_sym_yield] = ACTIONS(4559),
    [anon_sym_yield_BANG] = ACTIONS(4562),
    [anon_sym_lazy] = ACTIONS(4559),
    [anon_sym_assert] = ACTIONS(4559),
    [anon_sym_upcast] = ACTIONS(4559),
    [anon_sym_downcast] = ACTIONS(4559),
    [anon_sym_fixed] = ACTIONS(4559),
    [anon_sym_for] = ACTIONS(4565),
    [anon_sym_while] = ACTIONS(4568),
    [anon_sym_while_BANG] = ACTIONS(4571),
    [anon_sym_if] = ACTIONS(4198),
    [anon_sym_fun] = ACTIONS(4574),
    [anon_sym_try] = ACTIONS(4577),
    [anon_sym_match] = ACTIONS(4580),
    [anon_sym_match_BANG] = ACTIONS(4583),
    [anon_sym_function] = ACTIONS(4586),
    [anon_sym_use] = ACTIONS(4216),
    [anon_sym_use_BANG] = ACTIONS(4219),
    [anon_sym_do_BANG] = ACTIONS(4589),
    [anon_sym_AMP_AMP] = ACTIONS(4225),
    [anon_sym_exception] = ACTIONS(4592),
    [anon_sym_extern] = ACTIONS(4595),
    [aux_sym_char_token1] = ACTIONS(4598),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(4601),
    [anon_sym_DQUOTE] = ACTIONS(4604),
    [anon_sym_AT_DQUOTE] = ACTIONS(4607),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4610),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4613),
    [sym_bool] = ACTIONS(4616),
    [sym_unit] = ACTIONS(4619),
    [anon_sym_LPAREN_PIPE] = ACTIONS(4622),
    [sym_op_identifier] = ACTIONS(4625),
    [anon_sym_PLUS] = ACTIONS(4159),
    [anon_sym_DASH] = ACTIONS(4159),
    [anon_sym_PLUS_DOT] = ACTIONS(4264),
    [anon_sym_DASH_DOT] = ACTIONS(4264),
    [anon_sym_PERCENT] = ACTIONS(4159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(4225),
    [anon_sym_TILDE] = ACTIONS(4267),
    [aux_sym_prefix_op_token1] = ACTIONS(4225),
    [sym_int] = ACTIONS(4628),
    [sym_xint] = ACTIONS(4631),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4634),
    [anon_sym_POUNDload] = ACTIONS(4634),
    [anon_sym_POUNDtime] = ACTIONS(4634),
    [anon_sym_POUNDI] = ACTIONS(4634),
    [anon_sym_POUNDhelp] = ACTIONS(4634),
    [anon_sym_POUNDquit] = ACTIONS(4634),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4637),
    [anon_sym_POUNDendif] = ACTIONS(4282),
    [sym__newline] = ACTIONS(4640),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(4643),
  },
  [STATE(1004)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1004),
    [sym_compiler_directive_decl] = STATE(1004),
    [sym_fsi_directive_decl] = STATE(1004),
    [sym_preproc_line] = STATE(1004),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(999),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4646),
    [anon_sym_POUNDelse] = ACTIONS(4646),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1005)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1005),
    [sym_compiler_directive_decl] = STATE(1005),
    [sym_fsi_directive_decl] = STATE(1005),
    [sym_preproc_line] = STATE(1005),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1005),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(4648),
    [anon_sym_module] = ACTIONS(4651),
    [anon_sym_null] = ACTIONS(4654),
    [anon_sym_begin] = ACTIONS(4657),
    [anon_sym_open] = ACTIONS(4660),
    [anon_sym_type] = ACTIONS(4663),
    [anon_sym_LBRACK_LT] = ACTIONS(4336),
    [anon_sym_do] = ACTIONS(4666),
    [anon_sym_let] = ACTIONS(4669),
    [anon_sym_let_BANG] = ACTIONS(4672),
    [anon_sym_LPAREN] = ACTIONS(4675),
    [anon_sym_AMP] = ACTIONS(4351),
    [anon_sym_LBRACK] = ACTIONS(4678),
    [anon_sym_LBRACK_PIPE] = ACTIONS(4681),
    [anon_sym_LBRACE] = ACTIONS(4684),
    [anon_sym_LT_AT] = ACTIONS(4687),
    [anon_sym_LT_AT_AT] = ACTIONS(4690),
    [anon_sym_CARET] = ACTIONS(4369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(4693),
    [anon_sym_new] = ACTIONS(4696),
    [anon_sym_return] = ACTIONS(4696),
    [anon_sym_return_BANG] = ACTIONS(4699),
    [anon_sym_yield] = ACTIONS(4696),
    [anon_sym_yield_BANG] = ACTIONS(4699),
    [anon_sym_lazy] = ACTIONS(4696),
    [anon_sym_assert] = ACTIONS(4696),
    [anon_sym_upcast] = ACTIONS(4696),
    [anon_sym_downcast] = ACTIONS(4696),
    [anon_sym_fixed] = ACTIONS(4696),
    [anon_sym_for] = ACTIONS(4702),
    [anon_sym_while] = ACTIONS(4705),
    [anon_sym_while_BANG] = ACTIONS(4708),
    [anon_sym_if] = ACTIONS(4390),
    [anon_sym_fun] = ACTIONS(4711),
    [anon_sym_try] = ACTIONS(4714),
    [anon_sym_match] = ACTIONS(4717),
    [anon_sym_match_BANG] = ACTIONS(4720),
    [anon_sym_function] = ACTIONS(4723),
    [anon_sym_use] = ACTIONS(4726),
    [anon_sym_use_BANG] = ACTIONS(4729),
    [anon_sym_do_BANG] = ACTIONS(4732),
    [anon_sym_AMP_AMP] = ACTIONS(4417),
    [anon_sym_exception] = ACTIONS(4735),
    [anon_sym_extern] = ACTIONS(4738),
    [aux_sym_char_token1] = ACTIONS(4741),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(4744),
    [anon_sym_DQUOTE] = ACTIONS(4747),
    [anon_sym_AT_DQUOTE] = ACTIONS(4750),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4753),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4756),
    [sym_bool] = ACTIONS(4759),
    [sym_unit] = ACTIONS(4762),
    [anon_sym_LPAREN_PIPE] = ACTIONS(4765),
    [sym_op_identifier] = ACTIONS(4768),
    [anon_sym_PLUS] = ACTIONS(4351),
    [anon_sym_DASH] = ACTIONS(4351),
    [anon_sym_PLUS_DOT] = ACTIONS(4456),
    [anon_sym_DASH_DOT] = ACTIONS(4456),
    [anon_sym_PERCENT] = ACTIONS(4351),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(4417),
    [anon_sym_TILDE] = ACTIONS(4459),
    [aux_sym_prefix_op_token1] = ACTIONS(4417),
    [sym_int] = ACTIONS(4771),
    [sym_xint] = ACTIONS(4774),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4777),
    [anon_sym_POUNDload] = ACTIONS(4777),
    [anon_sym_POUNDtime] = ACTIONS(4777),
    [anon_sym_POUNDI] = ACTIONS(4777),
    [anon_sym_POUNDhelp] = ACTIONS(4777),
    [anon_sym_POUNDquit] = ACTIONS(4777),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4780),
    [anon_sym_POUNDendif] = ACTIONS(4314),
    [anon_sym_POUNDelse] = ACTIONS(4314),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(4783),
  },
  [STATE(1006)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1195),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1112),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(191),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1006),
    [sym_compiler_directive_decl] = STATE(1006),
    [sym_fsi_directive_decl] = STATE(1006),
    [sym_preproc_line] = STATE(1006),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1001),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4786),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_namespace] = ACTIONS(4788),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(4304),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1007)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1007),
    [sym_compiler_directive_decl] = STATE(1007),
    [sym_fsi_directive_decl] = STATE(1007),
    [sym_preproc_line] = STATE(1007),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1005),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4790),
    [anon_sym_POUNDelse] = ACTIONS(4790),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1008)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1008),
    [sym_compiler_directive_decl] = STATE(1008),
    [sym_fsi_directive_decl] = STATE(1008),
    [sym_preproc_line] = STATE(1008),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1009),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4792),
    [anon_sym_POUNDelse] = ACTIONS(4792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1009)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1009),
    [sym_compiler_directive_decl] = STATE(1009),
    [sym_fsi_directive_decl] = STATE(1009),
    [sym_preproc_line] = STATE(1009),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1005),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4794),
    [anon_sym_POUNDelse] = ACTIONS(4794),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1010)] = {
    [sym__module_elem] = STATE(8614),
    [sym__attribute_expression] = STATE(8576),
    [sym_module_abbrev] = STATE(8577),
    [sym_module_defn] = STATE(8577),
    [sym_import_decl] = STATE(8577),
    [sym_attributes] = STATE(1223),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8576),
    [sym_do] = STATE(8624),
    [sym_function_or_value_defn] = STATE(1114),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(163),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_exception_definition] = STATE(8577),
    [sym_type_definition] = STATE(8577),
    [sym_extern_binding] = STATE(8577),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1010),
    [sym_compiler_directive_decl] = STATE(1010),
    [sym_fsi_directive_decl] = STATE(1010),
    [sym_preproc_line] = STATE(1010),
    [sym_preproc_if] = STATE(8577),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_file_repeat2] = STATE(1000),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(4477),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(4479),
    [anon_sym_type] = ACTIONS(4481),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4489),
    [anon_sym_extern] = ACTIONS(4491),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4493),
    [anon_sym_POUNDload] = ACTIONS(4493),
    [anon_sym_POUNDtime] = ACTIONS(4493),
    [anon_sym_POUNDI] = ACTIONS(4493),
    [anon_sym_POUNDhelp] = ACTIONS(4493),
    [anon_sym_POUNDquit] = ACTIONS(4493),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4495),
    [anon_sym_POUNDendif] = ACTIONS(4796),
    [anon_sym_POUNDelse] = ACTIONS(4796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1011)] = {
    [sym__module_body_elem] = STATE(8691),
    [sym__module_expression] = STATE(8690),
    [sym_module_abbrev] = STATE(8690),
    [sym_module_defn] = STATE(8690),
    [sym_import_decl] = STATE(8690),
    [sym_attributes] = STATE(15049),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8692),
    [sym_do] = STATE(8637),
    [sym_function_or_value_defn] = STATE(1111),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(934),
    [sym_literal_expression] = STATE(4278),
    [sym_srtp_call_expression] = STATE(4278),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(4038),
    [sym_tuple_expression] = STATE(4278),
    [sym_brace_expression] = STATE(4278),
    [sym_anon_record_expression] = STATE(4278),
    [sym_prefixed_expression] = STATE(4278),
    [sym_typecast_expression] = STATE(4278),
    [sym_for_expression] = STATE(4278),
    [sym_while_expression] = STATE(4278),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4278),
    [sym_fun_expression] = STATE(4278),
    [sym_try_expression] = STATE(4278),
    [sym_match_expression] = STATE(4278),
    [sym_function_expression] = STATE(4278),
    [sym_mutate_expression] = STATE(4278),
    [sym_index_expression] = STATE(4038),
    [sym_typed_expression] = STATE(4278),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4278),
    [sym_list_expression] = STATE(4278),
    [sym_array_expression] = STATE(4278),
    [sym_begin_end_expression] = STATE(4278),
    [sym_paren_expression] = STATE(4278),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4278),
    [sym_dot_expression] = STATE(4038),
    [sym_infix_expression] = STATE(4278),
    [sym_ce_expression] = STATE(4278),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8690),
    [sym_type_definition] = STATE(8690),
    [sym_extern_binding] = STATE(8690),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4278),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1011),
    [sym_compiler_directive_decl] = STATE(1011),
    [sym_fsi_directive_decl] = STATE(1011),
    [sym_preproc_line] = STATE(1011),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8690),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_module_body_repeat1] = STATE(1003),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4091),
    [anon_sym_null] = ACTIONS(4290),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4095),
    [anon_sym_type] = ACTIONS(4097),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4099),
    [anon_sym_let] = ACTIONS(4101),
    [anon_sym_let_BANG] = ACTIONS(4103),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4109),
    [anon_sym_extern] = ACTIONS(4111),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4117),
    [anon_sym_POUNDload] = ACTIONS(4117),
    [anon_sym_POUNDtime] = ACTIONS(4117),
    [anon_sym_POUNDI] = ACTIONS(4117),
    [anon_sym_POUNDhelp] = ACTIONS(4117),
    [anon_sym_POUNDquit] = ACTIONS(4117),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4292),
    [anon_sym_POUNDendif] = ACTIONS(4798),
    [sym__newline] = ACTIONS(4296),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1012)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22518),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1012),
    [sym_compiler_directive_decl] = STATE(1012),
    [sym_fsi_directive_decl] = STATE(1012),
    [sym_preproc_line] = STATE(1012),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1013)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19330),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1013),
    [sym_compiler_directive_decl] = STATE(1013),
    [sym_fsi_directive_decl] = STATE(1013),
    [sym_preproc_line] = STATE(1013),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1014)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1014),
    [sym_compiler_directive_decl] = STATE(1014),
    [sym_fsi_directive_decl] = STATE(1014),
    [sym_preproc_line] = STATE(1014),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1018),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4646),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1015)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19235),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1015),
    [sym_compiler_directive_decl] = STATE(1015),
    [sym_fsi_directive_decl] = STATE(1015),
    [sym_preproc_line] = STATE(1015),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4854),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1016)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1016),
    [sym_compiler_directive_decl] = STATE(1016),
    [sym_fsi_directive_decl] = STATE(1016),
    [sym_preproc_line] = STATE(1016),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1051),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4856),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1017)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19349),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1017),
    [sym_compiler_directive_decl] = STATE(1017),
    [sym_fsi_directive_decl] = STATE(1017),
    [sym_preproc_line] = STATE(1017),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4858),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1018)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1018),
    [sym_compiler_directive_decl] = STATE(1018),
    [sym_fsi_directive_decl] = STATE(1018),
    [sym_preproc_line] = STATE(1018),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1056),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4499),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1019)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1019),
    [sym_compiler_directive_decl] = STATE(1019),
    [sym_fsi_directive_decl] = STATE(1019),
    [sym_preproc_line] = STATE(1019),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1061),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4860),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1020)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19362),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1020),
    [sym_compiler_directive_decl] = STATE(1020),
    [sym_fsi_directive_decl] = STATE(1020),
    [sym_preproc_line] = STATE(1020),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4862),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1021)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19374),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1021),
    [sym_compiler_directive_decl] = STATE(1021),
    [sym_fsi_directive_decl] = STATE(1021),
    [sym_preproc_line] = STATE(1021),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4864),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1022)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19120),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1022),
    [sym_compiler_directive_decl] = STATE(1022),
    [sym_fsi_directive_decl] = STATE(1022),
    [sym_preproc_line] = STATE(1022),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4866),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1023)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1023),
    [sym_compiler_directive_decl] = STATE(1023),
    [sym_fsi_directive_decl] = STATE(1023),
    [sym_preproc_line] = STATE(1023),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1051),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4868),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1024)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19248),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1024),
    [sym_compiler_directive_decl] = STATE(1024),
    [sym_fsi_directive_decl] = STATE(1024),
    [sym_preproc_line] = STATE(1024),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4870),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1025)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1025),
    [sym_compiler_directive_decl] = STATE(1025),
    [sym_fsi_directive_decl] = STATE(1025),
    [sym_preproc_line] = STATE(1025),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1056),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4790),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1026)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1026),
    [sym_compiler_directive_decl] = STATE(1026),
    [sym_fsi_directive_decl] = STATE(1026),
    [sym_preproc_line] = STATE(1026),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1062),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1027)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19444),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1027),
    [sym_compiler_directive_decl] = STATE(1027),
    [sym_fsi_directive_decl] = STATE(1027),
    [sym_preproc_line] = STATE(1027),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4872),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1028)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19216),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1028),
    [sym_compiler_directive_decl] = STATE(1028),
    [sym_fsi_directive_decl] = STATE(1028),
    [sym_preproc_line] = STATE(1028),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4874),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1029)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21990),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1029),
    [sym_compiler_directive_decl] = STATE(1029),
    [sym_fsi_directive_decl] = STATE(1029),
    [sym_preproc_line] = STATE(1029),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4876),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1030)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19463),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1030),
    [sym_compiler_directive_decl] = STATE(1030),
    [sym_fsi_directive_decl] = STATE(1030),
    [sym_preproc_line] = STATE(1030),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4878),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1031)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22034),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1031),
    [sym_compiler_directive_decl] = STATE(1031),
    [sym_fsi_directive_decl] = STATE(1031),
    [sym_preproc_line] = STATE(1031),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4880),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1032)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22047),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1032),
    [sym_compiler_directive_decl] = STATE(1032),
    [sym_fsi_directive_decl] = STATE(1032),
    [sym_preproc_line] = STATE(1032),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4882),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1033)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22059),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1033),
    [sym_compiler_directive_decl] = STATE(1033),
    [sym_fsi_directive_decl] = STATE(1033),
    [sym_preproc_line] = STATE(1033),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4884),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1034)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19936),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1034),
    [sym_compiler_directive_decl] = STATE(1034),
    [sym_fsi_directive_decl] = STATE(1034),
    [sym_preproc_line] = STATE(1034),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1035)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1035),
    [sym_compiler_directive_decl] = STATE(1035),
    [sym_fsi_directive_decl] = STATE(1035),
    [sym_preproc_line] = STATE(1035),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1016),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4888),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1036)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19145),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1036),
    [sym_compiler_directive_decl] = STATE(1036),
    [sym_fsi_directive_decl] = STATE(1036),
    [sym_preproc_line] = STATE(1036),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1037)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1037),
    [sym_compiler_directive_decl] = STATE(1037),
    [sym_fsi_directive_decl] = STATE(1037),
    [sym_preproc_line] = STATE(1037),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1042),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1038)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19476),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1038),
    [sym_compiler_directive_decl] = STATE(1038),
    [sym_fsi_directive_decl] = STATE(1038),
    [sym_preproc_line] = STATE(1038),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4892),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1039)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1039),
    [sym_compiler_directive_decl] = STATE(1039),
    [sym_fsi_directive_decl] = STATE(1039),
    [sym_preproc_line] = STATE(1039),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1053),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4894),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1040)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19260),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1040),
    [sym_compiler_directive_decl] = STATE(1040),
    [sym_fsi_directive_decl] = STATE(1040),
    [sym_preproc_line] = STATE(1040),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4896),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1041)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22156),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1041),
    [sym_compiler_directive_decl] = STATE(1041),
    [sym_fsi_directive_decl] = STATE(1041),
    [sym_preproc_line] = STATE(1041),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1042)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1042),
    [sym_compiler_directive_decl] = STATE(1042),
    [sym_fsi_directive_decl] = STATE(1042),
    [sym_preproc_line] = STATE(1042),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1056),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4501),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1043)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22191),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1043),
    [sym_compiler_directive_decl] = STATE(1043),
    [sym_fsi_directive_decl] = STATE(1043),
    [sym_preproc_line] = STATE(1043),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4900),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1044)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19488),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1044),
    [sym_compiler_directive_decl] = STATE(1044),
    [sym_fsi_directive_decl] = STATE(1044),
    [sym_preproc_line] = STATE(1044),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4902),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1045)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22222),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1045),
    [sym_compiler_directive_decl] = STATE(1045),
    [sym_fsi_directive_decl] = STATE(1045),
    [sym_preproc_line] = STATE(1045),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1046)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22252),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1046),
    [sym_compiler_directive_decl] = STATE(1046),
    [sym_fsi_directive_decl] = STATE(1046),
    [sym_preproc_line] = STATE(1046),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4906),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1047)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22370),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1047),
    [sym_compiler_directive_decl] = STATE(1047),
    [sym_fsi_directive_decl] = STATE(1047),
    [sym_preproc_line] = STATE(1047),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4908),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1048)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1048),
    [sym_compiler_directive_decl] = STATE(1048),
    [sym_fsi_directive_decl] = STATE(1048),
    [sym_preproc_line] = STATE(1048),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1023),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4910),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1049)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22453),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1049),
    [sym_compiler_directive_decl] = STATE(1049),
    [sym_fsi_directive_decl] = STATE(1049),
    [sym_preproc_line] = STATE(1049),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4912),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1050)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19101),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1050),
    [sym_compiler_directive_decl] = STATE(1050),
    [sym_fsi_directive_decl] = STATE(1050),
    [sym_preproc_line] = STATE(1050),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4914),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1051)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1051),
    [sym_compiler_directive_decl] = STATE(1051),
    [sym_fsi_directive_decl] = STATE(1051),
    [sym_preproc_line] = STATE(1051),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1051),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4314),
    [sym_identifier] = ACTIONS(4316),
    [anon_sym_module] = ACTIONS(4321),
    [anon_sym_null] = ACTIONS(4324),
    [anon_sym_begin] = ACTIONS(4327),
    [anon_sym_open] = ACTIONS(4330),
    [anon_sym_type] = ACTIONS(4916),
    [anon_sym_LBRACK_LT] = ACTIONS(4336),
    [anon_sym_do] = ACTIONS(4339),
    [anon_sym_let] = ACTIONS(4342),
    [anon_sym_let_BANG] = ACTIONS(4345),
    [anon_sym_LPAREN] = ACTIONS(4348),
    [anon_sym_AMP] = ACTIONS(4351),
    [anon_sym_LBRACK] = ACTIONS(4354),
    [anon_sym_LBRACK_PIPE] = ACTIONS(4357),
    [anon_sym_LBRACE] = ACTIONS(4360),
    [anon_sym_LT_AT] = ACTIONS(4363),
    [anon_sym_LT_AT_AT] = ACTIONS(4366),
    [anon_sym_CARET] = ACTIONS(4369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(4372),
    [anon_sym_new] = ACTIONS(4919),
    [anon_sym_return] = ACTIONS(4919),
    [anon_sym_return_BANG] = ACTIONS(4922),
    [anon_sym_yield] = ACTIONS(4919),
    [anon_sym_yield_BANG] = ACTIONS(4922),
    [anon_sym_lazy] = ACTIONS(4919),
    [anon_sym_assert] = ACTIONS(4919),
    [anon_sym_upcast] = ACTIONS(4919),
    [anon_sym_downcast] = ACTIONS(4919),
    [anon_sym_fixed] = ACTIONS(4919),
    [anon_sym_for] = ACTIONS(4381),
    [anon_sym_while] = ACTIONS(4384),
    [anon_sym_while_BANG] = ACTIONS(4387),
    [anon_sym_if] = ACTIONS(4390),
    [anon_sym_fun] = ACTIONS(4393),
    [anon_sym_try] = ACTIONS(4396),
    [anon_sym_match] = ACTIONS(4399),
    [anon_sym_match_BANG] = ACTIONS(4402),
    [anon_sym_function] = ACTIONS(4405),
    [anon_sym_use] = ACTIONS(4925),
    [anon_sym_use_BANG] = ACTIONS(4928),
    [anon_sym_do_BANG] = ACTIONS(4414),
    [anon_sym_AMP_AMP] = ACTIONS(4417),
    [anon_sym_exception] = ACTIONS(4420),
    [anon_sym_extern] = ACTIONS(4423),
    [aux_sym_char_token1] = ACTIONS(4426),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(4429),
    [anon_sym_DQUOTE] = ACTIONS(4432),
    [anon_sym_AT_DQUOTE] = ACTIONS(4435),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4438),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(4441),
    [sym_bool] = ACTIONS(4444),
    [sym_unit] = ACTIONS(4447),
    [anon_sym_LPAREN_PIPE] = ACTIONS(4450),
    [sym_op_identifier] = ACTIONS(4453),
    [anon_sym_PLUS] = ACTIONS(4351),
    [anon_sym_DASH] = ACTIONS(4351),
    [anon_sym_PLUS_DOT] = ACTIONS(4456),
    [anon_sym_DASH_DOT] = ACTIONS(4456),
    [anon_sym_PERCENT] = ACTIONS(4351),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(4417),
    [anon_sym_TILDE] = ACTIONS(4459),
    [aux_sym_prefix_op_token1] = ACTIONS(4417),
    [sym_int] = ACTIONS(4931),
    [sym_xint] = ACTIONS(4934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4468),
    [anon_sym_POUNDload] = ACTIONS(4468),
    [anon_sym_POUNDtime] = ACTIONS(4468),
    [anon_sym_POUNDI] = ACTIONS(4468),
    [anon_sym_POUNDhelp] = ACTIONS(4468),
    [anon_sym_POUNDquit] = ACTIONS(4468),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4471),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(4474),
  },
  [STATE(1052)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21091),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1052),
    [sym_compiler_directive_decl] = STATE(1052),
    [sym_fsi_directive_decl] = STATE(1052),
    [sym_preproc_line] = STATE(1052),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1053)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1053),
    [sym_compiler_directive_decl] = STATE(1053),
    [sym_fsi_directive_decl] = STATE(1053),
    [sym_preproc_line] = STATE(1053),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1051),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(4939),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1054)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22700),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1054),
    [sym_compiler_directive_decl] = STATE(1054),
    [sym_fsi_directive_decl] = STATE(1054),
    [sym_preproc_line] = STATE(1054),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4941),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1055)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22741),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1055),
    [sym_compiler_directive_decl] = STATE(1055),
    [sym_fsi_directive_decl] = STATE(1055),
    [sym_preproc_line] = STATE(1055),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(4943),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1056)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1056),
    [sym_compiler_directive_decl] = STATE(1056),
    [sym_fsi_directive_decl] = STATE(1056),
    [sym_preproc_line] = STATE(1056),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1056),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(4945),
    [anon_sym_module] = ACTIONS(4948),
    [anon_sym_null] = ACTIONS(4951),
    [anon_sym_begin] = ACTIONS(4954),
    [anon_sym_open] = ACTIONS(4957),
    [anon_sym_type] = ACTIONS(4960),
    [anon_sym_LBRACK_LT] = ACTIONS(4336),
    [anon_sym_do] = ACTIONS(4963),
    [anon_sym_let] = ACTIONS(4966),
    [anon_sym_let_BANG] = ACTIONS(4969),
    [anon_sym_LPAREN] = ACTIONS(4972),
    [anon_sym_AMP] = ACTIONS(4351),
    [anon_sym_LBRACK] = ACTIONS(4975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(4978),
    [anon_sym_LBRACE] = ACTIONS(4981),
    [anon_sym_LT_AT] = ACTIONS(4984),
    [anon_sym_LT_AT_AT] = ACTIONS(4987),
    [anon_sym_CARET] = ACTIONS(4369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(4990),
    [anon_sym_new] = ACTIONS(4993),
    [anon_sym_return] = ACTIONS(4993),
    [anon_sym_return_BANG] = ACTIONS(4996),
    [anon_sym_yield] = ACTIONS(4993),
    [anon_sym_yield_BANG] = ACTIONS(4996),
    [anon_sym_lazy] = ACTIONS(4993),
    [anon_sym_assert] = ACTIONS(4993),
    [anon_sym_upcast] = ACTIONS(4993),
    [anon_sym_downcast] = ACTIONS(4993),
    [anon_sym_fixed] = ACTIONS(4993),
    [anon_sym_for] = ACTIONS(4999),
    [anon_sym_while] = ACTIONS(5002),
    [anon_sym_while_BANG] = ACTIONS(5005),
    [anon_sym_if] = ACTIONS(4390),
    [anon_sym_fun] = ACTIONS(5008),
    [anon_sym_try] = ACTIONS(5011),
    [anon_sym_match] = ACTIONS(5014),
    [anon_sym_match_BANG] = ACTIONS(5017),
    [anon_sym_function] = ACTIONS(5020),
    [anon_sym_use] = ACTIONS(5023),
    [anon_sym_use_BANG] = ACTIONS(5026),
    [anon_sym_do_BANG] = ACTIONS(5029),
    [anon_sym_AMP_AMP] = ACTIONS(4417),
    [anon_sym_exception] = ACTIONS(5032),
    [anon_sym_extern] = ACTIONS(5035),
    [aux_sym_char_token1] = ACTIONS(5038),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5041),
    [anon_sym_DQUOTE] = ACTIONS(5044),
    [anon_sym_AT_DQUOTE] = ACTIONS(5047),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5050),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5053),
    [sym_bool] = ACTIONS(5056),
    [sym_unit] = ACTIONS(5059),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5062),
    [sym_op_identifier] = ACTIONS(5065),
    [anon_sym_PLUS] = ACTIONS(4351),
    [anon_sym_DASH] = ACTIONS(4351),
    [anon_sym_PLUS_DOT] = ACTIONS(4456),
    [anon_sym_DASH_DOT] = ACTIONS(4456),
    [anon_sym_PERCENT] = ACTIONS(4351),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(4417),
    [anon_sym_TILDE] = ACTIONS(4459),
    [aux_sym_prefix_op_token1] = ACTIONS(4417),
    [sym_int] = ACTIONS(5068),
    [sym_xint] = ACTIONS(5071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5074),
    [anon_sym_POUNDload] = ACTIONS(5074),
    [anon_sym_POUNDtime] = ACTIONS(5074),
    [anon_sym_POUNDI] = ACTIONS(5074),
    [anon_sym_POUNDhelp] = ACTIONS(5074),
    [anon_sym_POUNDquit] = ACTIONS(5074),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5077),
    [anon_sym_POUNDendif] = ACTIONS(4314),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5080),
  },
  [STATE(1057)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1057),
    [sym_compiler_directive_decl] = STATE(1057),
    [sym_fsi_directive_decl] = STATE(1057),
    [sym_preproc_line] = STATE(1057),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1051),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(5083),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1058)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21848),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1058),
    [sym_compiler_directive_decl] = STATE(1058),
    [sym_fsi_directive_decl] = STATE(1058),
    [sym_preproc_line] = STATE(1058),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(5085),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1059)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19133),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1059),
    [sym_compiler_directive_decl] = STATE(1059),
    [sym_fsi_directive_decl] = STATE(1059),
    [sym_preproc_line] = STATE(1059),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__dedent] = ACTIONS(5087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1060)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1060),
    [sym_compiler_directive_decl] = STATE(1060),
    [sym_fsi_directive_decl] = STATE(1060),
    [sym_preproc_line] = STATE(1060),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1025),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4497),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1061)] = {
    [sym__module_elem] = STATE(8530),
    [sym__attribute_expression] = STATE(8657),
    [sym_module_abbrev] = STATE(8628),
    [sym_module_defn] = STATE(8628),
    [sym_import_decl] = STATE(8628),
    [sym_attributes] = STATE(1162),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8657),
    [sym_do] = STATE(8666),
    [sym_function_or_value_defn] = STATE(1118),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(204),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_exception_definition] = STATE(8628),
    [sym_type_definition] = STATE(8628),
    [sym_extern_binding] = STATE(8628),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1061),
    [sym_compiler_directive_decl] = STATE(1061),
    [sym_fsi_directive_decl] = STATE(1061),
    [sym_preproc_line] = STATE(1061),
    [sym_preproc_if] = STATE(8628),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_file_repeat2] = STATE(1051),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(5089),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(4302),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_open] = ACTIONS(31),
    [anon_sym_type] = ACTIONS(33),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(91),
    [anon_sym_extern] = ACTIONS(93),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(123),
    [anon_sym_POUNDload] = ACTIONS(123),
    [anon_sym_POUNDtime] = ACTIONS(123),
    [anon_sym_POUNDI] = ACTIONS(123),
    [anon_sym_POUNDhelp] = ACTIONS(123),
    [anon_sym_POUNDquit] = ACTIONS(123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(125),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1062)] = {
    [sym__module_elem] = STATE(8743),
    [sym__attribute_expression] = STATE(8701),
    [sym_module_abbrev] = STATE(8702),
    [sym_module_defn] = STATE(8702),
    [sym_import_decl] = STATE(8702),
    [sym_attributes] = STATE(1166),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(8701),
    [sym_do] = STATE(8715),
    [sym_function_or_value_defn] = STATE(1121),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(197),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_exception_definition] = STATE(8702),
    [sym_type_definition] = STATE(8702),
    [sym_extern_binding] = STATE(8702),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1062),
    [sym_compiler_directive_decl] = STATE(1062),
    [sym_fsi_directive_decl] = STATE(1062),
    [sym_preproc_line] = STATE(1062),
    [sym_preproc_if] = STATE(8702),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_file_repeat2] = STATE(1056),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4834),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4836),
    [anon_sym_type] = ACTIONS(4838),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4846),
    [anon_sym_extern] = ACTIONS(4848),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4850),
    [anon_sym_POUNDload] = ACTIONS(4850),
    [anon_sym_POUNDtime] = ACTIONS(4850),
    [anon_sym_POUNDI] = ACTIONS(4850),
    [anon_sym_POUNDhelp] = ACTIONS(4850),
    [anon_sym_POUNDquit] = ACTIONS(4850),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4852),
    [anon_sym_POUNDendif] = ACTIONS(4794),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1063)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21971),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5831),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1063),
    [sym_compiler_directive_decl] = STATE(1063),
    [sym_fsi_directive_decl] = STATE(1063),
    [sym_preproc_line] = STATE(1063),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1064)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19638),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1064),
    [sym_compiler_directive_decl] = STATE(1064),
    [sym_fsi_directive_decl] = STATE(1064),
    [sym_preproc_line] = STATE(1064),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1065)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21955),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1065),
    [sym_compiler_directive_decl] = STATE(1065),
    [sym_fsi_directive_decl] = STATE(1065),
    [sym_preproc_line] = STATE(1065),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1066)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19638),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5844),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1066),
    [sym_compiler_directive_decl] = STATE(1066),
    [sym_fsi_directive_decl] = STATE(1066),
    [sym_preproc_line] = STATE(1066),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1067)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19111),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1067),
    [sym_compiler_directive_decl] = STATE(1067),
    [sym_fsi_directive_decl] = STATE(1067),
    [sym_preproc_line] = STATE(1067),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1068)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22130),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1068),
    [sym_compiler_directive_decl] = STATE(1068),
    [sym_fsi_directive_decl] = STATE(1068),
    [sym_preproc_line] = STATE(1068),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1069)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21223),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5868),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1069),
    [sym_compiler_directive_decl] = STATE(1069),
    [sym_fsi_directive_decl] = STATE(1069),
    [sym_preproc_line] = STATE(1069),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1070)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19542),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5785),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1070),
    [sym_compiler_directive_decl] = STATE(1070),
    [sym_fsi_directive_decl] = STATE(1070),
    [sym_preproc_line] = STATE(1070),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1071)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20410),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1071),
    [sym_compiler_directive_decl] = STATE(1071),
    [sym_fsi_directive_decl] = STATE(1071),
    [sym_preproc_line] = STATE(1071),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1072)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21543),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1072),
    [sym_compiler_directive_decl] = STATE(1072),
    [sym_fsi_directive_decl] = STATE(1072),
    [sym_preproc_line] = STATE(1072),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1073)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20410),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5790),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1073),
    [sym_compiler_directive_decl] = STATE(1073),
    [sym_fsi_directive_decl] = STATE(1073),
    [sym_preproc_line] = STATE(1073),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1074)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19536),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5311),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1074),
    [sym_compiler_directive_decl] = STATE(1074),
    [sym_fsi_directive_decl] = STATE(1074),
    [sym_preproc_line] = STATE(1074),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1075)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19671),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1075),
    [sym_compiler_directive_decl] = STATE(1075),
    [sym_fsi_directive_decl] = STATE(1075),
    [sym_preproc_line] = STATE(1075),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1076)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20014),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1076),
    [sym_compiler_directive_decl] = STATE(1076),
    [sym_fsi_directive_decl] = STATE(1076),
    [sym_preproc_line] = STATE(1076),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1077)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19708),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1077),
    [sym_compiler_directive_decl] = STATE(1077),
    [sym_fsi_directive_decl] = STATE(1077),
    [sym_preproc_line] = STATE(1077),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1078)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19692),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1078),
    [sym_compiler_directive_decl] = STATE(1078),
    [sym_fsi_directive_decl] = STATE(1078),
    [sym_preproc_line] = STATE(1078),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1079)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21392),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1079),
    [sym_compiler_directive_decl] = STATE(1079),
    [sym_fsi_directive_decl] = STATE(1079),
    [sym_preproc_line] = STATE(1079),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1080)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21702),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5681),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1080),
    [sym_compiler_directive_decl] = STATE(1080),
    [sym_fsi_directive_decl] = STATE(1080),
    [sym_preproc_line] = STATE(1080),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1081)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20468),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1081),
    [sym_compiler_directive_decl] = STATE(1081),
    [sym_fsi_directive_decl] = STATE(1081),
    [sym_preproc_line] = STATE(1081),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1082)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(22788),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5851),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1082),
    [sym_compiler_directive_decl] = STATE(1082),
    [sym_fsi_directive_decl] = STATE(1082),
    [sym_preproc_line] = STATE(1082),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1083)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21029),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5830),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1083),
    [sym_compiler_directive_decl] = STATE(1083),
    [sym_fsi_directive_decl] = STATE(1083),
    [sym_preproc_line] = STATE(1083),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1084)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21808),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1084),
    [sym_compiler_directive_decl] = STATE(1084),
    [sym_fsi_directive_decl] = STATE(1084),
    [sym_preproc_line] = STATE(1084),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1085)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21971),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1085),
    [sym_compiler_directive_decl] = STATE(1085),
    [sym_fsi_directive_decl] = STATE(1085),
    [sym_preproc_line] = STATE(1085),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1086)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19273),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5852),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1086),
    [sym_compiler_directive_decl] = STATE(1086),
    [sym_fsi_directive_decl] = STATE(1086),
    [sym_preproc_line] = STATE(1086),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1087)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19544),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1087),
    [sym_compiler_directive_decl] = STATE(1087),
    [sym_fsi_directive_decl] = STATE(1087),
    [sym_preproc_line] = STATE(1087),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1088)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19179),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1088),
    [sym_compiler_directive_decl] = STATE(1088),
    [sym_fsi_directive_decl] = STATE(1088),
    [sym_preproc_line] = STATE(1088),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1089)] = {
    [sym__preproc_toplevel_module] = STATE(21871),
    [sym__module_elem] = STATE(21875),
    [sym__attribute_expression] = STATE(19568),
    [sym_module_abbrev] = STATE(19569),
    [sym_module_defn] = STATE(19569),
    [sym_import_decl] = STATE(19569),
    [sym_attributes] = STATE(1164),
    [sym__attribute_set] = STATE(8787),
    [sym_value_declaration] = STATE(19568),
    [sym_do] = STATE(20501),
    [sym_function_or_value_defn] = STATE(1255),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(599),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_exception_definition] = STATE(19569),
    [sym_type_definition] = STATE(19569),
    [sym_extern_binding] = STATE(19569),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1089),
    [sym_compiler_directive_decl] = STATE(1089),
    [sym_fsi_directive_decl] = STATE(1089),
    [sym_preproc_line] = STATE(1089),
    [sym_preproc_if] = STATE(19569),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_attributes_repeat1] = STATE(8740),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_module] = ACTIONS(5091),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_open] = ACTIONS(5093),
    [anon_sym_type] = ACTIONS(5095),
    [anon_sym_LBRACK_LT] = ACTIONS(35),
    [anon_sym_do] = ACTIONS(5097),
    [anon_sym_let] = ACTIONS(5099),
    [anon_sym_let_BANG] = ACTIONS(5101),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5107),
    [anon_sym_extern] = ACTIONS(5109),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5115),
    [anon_sym_POUNDload] = ACTIONS(5115),
    [anon_sym_POUNDtime] = ACTIONS(5115),
    [anon_sym_POUNDI] = ACTIONS(5115),
    [anon_sym_POUNDhelp] = ACTIONS(5115),
    [anon_sym_POUNDquit] = ACTIONS(5115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5117),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1090)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19544),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5876),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1090),
    [sym_compiler_directive_decl] = STATE(1090),
    [sym_fsi_directive_decl] = STATE(1090),
    [sym_preproc_line] = STATE(1090),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1091)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21808),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5694),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1091),
    [sym_compiler_directive_decl] = STATE(1091),
    [sym_fsi_directive_decl] = STATE(1091),
    [sym_preproc_line] = STATE(1091),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1092)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20861),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1092),
    [sym_compiler_directive_decl] = STATE(1092),
    [sym_fsi_directive_decl] = STATE(1092),
    [sym_preproc_line] = STATE(1092),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1093)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19692),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5312),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1093),
    [sym_compiler_directive_decl] = STATE(1093),
    [sym_fsi_directive_decl] = STATE(1093),
    [sym_preproc_line] = STATE(1093),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1094)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19737),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1094),
    [sym_compiler_directive_decl] = STATE(1094),
    [sym_fsi_directive_decl] = STATE(1094),
    [sym_preproc_line] = STATE(1094),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1095)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20095),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5411),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1095),
    [sym_compiler_directive_decl] = STATE(1095),
    [sym_fsi_directive_decl] = STATE(1095),
    [sym_preproc_line] = STATE(1095),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1096)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21543),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5869),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1096),
    [sym_compiler_directive_decl] = STATE(1096),
    [sym_fsi_directive_decl] = STATE(1096),
    [sym_preproc_line] = STATE(1096),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1097)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21894),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1097),
    [sym_compiler_directive_decl] = STATE(1097),
    [sym_fsi_directive_decl] = STATE(1097),
    [sym_preproc_line] = STATE(1097),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1098)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19857),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1098),
    [sym_compiler_directive_decl] = STATE(1098),
    [sym_fsi_directive_decl] = STATE(1098),
    [sym_preproc_line] = STATE(1098),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1099)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(20049),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1099),
    [sym_compiler_directive_decl] = STATE(1099),
    [sym_fsi_directive_decl] = STATE(1099),
    [sym_preproc_line] = STATE(1099),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1100)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19273),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1100),
    [sym_compiler_directive_decl] = STATE(1100),
    [sym_fsi_directive_decl] = STATE(1100),
    [sym_preproc_line] = STATE(1100),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1101)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(21983),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1101),
    [sym_compiler_directive_decl] = STATE(1101),
    [sym_fsi_directive_decl] = STATE(1101),
    [sym_preproc_line] = STATE(1101),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1102)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19460),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1102),
    [sym_compiler_directive_decl] = STATE(1102),
    [sym_fsi_directive_decl] = STATE(1102),
    [sym_preproc_line] = STATE(1102),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1103)] = {
    [sym__module_body_elem] = STATE(17112),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym__module_body] = STATE(19226),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(5843),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1103),
    [sym_compiler_directive_decl] = STATE(1103),
    [sym_fsi_directive_decl] = STATE(1103),
    [sym_preproc_line] = STATE(1103),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1104)] = {
    [sym__module_body_elem] = STATE(8636),
    [sym__module_expression] = STATE(8690),
    [sym_module_abbrev] = STATE(8690),
    [sym_module_defn] = STATE(8690),
    [sym_import_decl] = STATE(8690),
    [sym_attributes] = STATE(15049),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8692),
    [sym_do] = STATE(8637),
    [sym_function_or_value_defn] = STATE(1117),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(538),
    [sym_literal_expression] = STATE(4185),
    [sym_srtp_call_expression] = STATE(4185),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(4058),
    [sym_tuple_expression] = STATE(4185),
    [sym_brace_expression] = STATE(4185),
    [sym_anon_record_expression] = STATE(4185),
    [sym_prefixed_expression] = STATE(4185),
    [sym_typecast_expression] = STATE(4185),
    [sym_for_expression] = STATE(4185),
    [sym_while_expression] = STATE(4185),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4185),
    [sym_fun_expression] = STATE(4185),
    [sym_try_expression] = STATE(4185),
    [sym_match_expression] = STATE(4185),
    [sym_function_expression] = STATE(4185),
    [sym_mutate_expression] = STATE(4185),
    [sym_index_expression] = STATE(4058),
    [sym_typed_expression] = STATE(4185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(4185),
    [sym_list_expression] = STATE(4185),
    [sym_array_expression] = STATE(4185),
    [sym_begin_end_expression] = STATE(4185),
    [sym_paren_expression] = STATE(4185),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4185),
    [sym_dot_expression] = STATE(4058),
    [sym_infix_expression] = STATE(4185),
    [sym_ce_expression] = STATE(4185),
    [sym_sequential_expression] = STATE(6185),
    [sym_exception_definition] = STATE(8690),
    [sym_type_definition] = STATE(8690),
    [sym_extern_binding] = STATE(8690),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4185),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1104),
    [sym_compiler_directive_decl] = STATE(1104),
    [sym_fsi_directive_decl] = STATE(1104),
    [sym_preproc_line] = STATE(1104),
    [sym_preproc_if_in_expression] = STATE(6185),
    [sym_preproc_if_in_module_body] = STATE(8690),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4091),
    [anon_sym_null] = ACTIONS(4093),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4095),
    [anon_sym_type] = ACTIONS(4097),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4099),
    [anon_sym_let] = ACTIONS(4101),
    [anon_sym_let_BANG] = ACTIONS(4103),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4109),
    [anon_sym_extern] = ACTIONS(4111),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4117),
    [anon_sym_POUNDload] = ACTIONS(4117),
    [anon_sym_POUNDtime] = ACTIONS(4117),
    [anon_sym_POUNDI] = ACTIONS(4117),
    [anon_sym_POUNDhelp] = ACTIONS(4117),
    [anon_sym_POUNDquit] = ACTIONS(4117),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1105)] = {
    [sym__module_body_elem] = STATE(8494),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1110),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(319),
    [sym_literal_expression] = STATE(4100),
    [sym_srtp_call_expression] = STATE(4100),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3862),
    [sym_tuple_expression] = STATE(4100),
    [sym_brace_expression] = STATE(4100),
    [sym_anon_record_expression] = STATE(4100),
    [sym_prefixed_expression] = STATE(4100),
    [sym_typecast_expression] = STATE(4100),
    [sym_for_expression] = STATE(4100),
    [sym_while_expression] = STATE(4100),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4100),
    [sym_fun_expression] = STATE(4100),
    [sym_try_expression] = STATE(4100),
    [sym_match_expression] = STATE(4100),
    [sym_function_expression] = STATE(4100),
    [sym_mutate_expression] = STATE(4100),
    [sym_index_expression] = STATE(3862),
    [sym_typed_expression] = STATE(4100),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(4100),
    [sym_list_expression] = STATE(4100),
    [sym_array_expression] = STATE(4100),
    [sym_begin_end_expression] = STATE(4100),
    [sym_paren_expression] = STATE(4100),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4100),
    [sym_dot_expression] = STATE(3862),
    [sym_infix_expression] = STATE(4100),
    [sym_ce_expression] = STATE(4100),
    [sym_sequential_expression] = STATE(5827),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4100),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1105),
    [sym_compiler_directive_decl] = STATE(1105),
    [sym_fsi_directive_decl] = STATE(1105),
    [sym_preproc_line] = STATE(1105),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(2111),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1106)] = {
    [sym__module_body_elem] = STATE(8494),
    [sym__module_expression] = STATE(8392),
    [sym_module_abbrev] = STATE(8392),
    [sym_module_defn] = STATE(8392),
    [sym_import_decl] = STATE(8392),
    [sym_attributes] = STATE(14941),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8402),
    [sym_do] = STATE(8306),
    [sym_function_or_value_defn] = STATE(1109),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(970),
    [sym_literal_expression] = STATE(4057),
    [sym_srtp_call_expression] = STATE(4057),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3843),
    [sym_tuple_expression] = STATE(4057),
    [sym_brace_expression] = STATE(4057),
    [sym_anon_record_expression] = STATE(4057),
    [sym_prefixed_expression] = STATE(4057),
    [sym_typecast_expression] = STATE(4057),
    [sym_for_expression] = STATE(4057),
    [sym_while_expression] = STATE(4057),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4057),
    [sym_fun_expression] = STATE(4057),
    [sym_try_expression] = STATE(4057),
    [sym_match_expression] = STATE(4057),
    [sym_function_expression] = STATE(4057),
    [sym_mutate_expression] = STATE(4057),
    [sym_index_expression] = STATE(3843),
    [sym_typed_expression] = STATE(4057),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4057),
    [sym_list_expression] = STATE(4057),
    [sym_array_expression] = STATE(4057),
    [sym_begin_end_expression] = STATE(4057),
    [sym_paren_expression] = STATE(4057),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4057),
    [sym_dot_expression] = STATE(3843),
    [sym_infix_expression] = STATE(4057),
    [sym_ce_expression] = STATE(4057),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8392),
    [sym_type_definition] = STATE(8392),
    [sym_extern_binding] = STATE(8392),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4057),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1106),
    [sym_compiler_directive_decl] = STATE(1106),
    [sym_fsi_directive_decl] = STATE(1106),
    [sym_preproc_line] = STATE(1106),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8392),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(2109),
    [anon_sym_null] = ACTIONS(3985),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2115),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_let_BANG] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_extern] = ACTIONS(2135),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(2141),
    [anon_sym_POUNDload] = ACTIONS(2141),
    [anon_sym_POUNDtime] = ACTIONS(2141),
    [anon_sym_POUNDI] = ACTIONS(2141),
    [anon_sym_POUNDhelp] = ACTIONS(2141),
    [anon_sym_POUNDquit] = ACTIONS(2141),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3987),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1107)] = {
    [sym__module_body_elem] = STATE(17883),
    [sym__module_expression] = STATE(18780),
    [sym_module_abbrev] = STATE(18780),
    [sym_module_defn] = STATE(18780),
    [sym_import_decl] = STATE(18780),
    [sym_attributes] = STATE(14743),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(18792),
    [sym_do] = STATE(13908),
    [sym_function_or_value_defn] = STATE(1221),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(951),
    [sym_literal_expression] = STATE(6320),
    [sym_srtp_call_expression] = STATE(6320),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(5706),
    [sym_tuple_expression] = STATE(6320),
    [sym_brace_expression] = STATE(6320),
    [sym_anon_record_expression] = STATE(6320),
    [sym_prefixed_expression] = STATE(6320),
    [sym_typecast_expression] = STATE(6320),
    [sym_for_expression] = STATE(6320),
    [sym_while_expression] = STATE(6320),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(6320),
    [sym_fun_expression] = STATE(6320),
    [sym_try_expression] = STATE(6320),
    [sym_match_expression] = STATE(6320),
    [sym_function_expression] = STATE(6320),
    [sym_mutate_expression] = STATE(6320),
    [sym_index_expression] = STATE(5706),
    [sym_typed_expression] = STATE(6320),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(6320),
    [sym_list_expression] = STATE(6320),
    [sym_array_expression] = STATE(6320),
    [sym_begin_end_expression] = STATE(6320),
    [sym_paren_expression] = STATE(6320),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(6320),
    [sym_dot_expression] = STATE(5706),
    [sym_infix_expression] = STATE(6320),
    [sym_ce_expression] = STATE(6320),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(18780),
    [sym_type_definition] = STATE(18780),
    [sym_extern_binding] = STATE(18780),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(6320),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1107),
    [sym_compiler_directive_decl] = STATE(1107),
    [sym_fsi_directive_decl] = STATE(1107),
    [sym_preproc_line] = STATE(1107),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(18780),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(4800),
    [anon_sym_null] = ACTIONS(4802),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(4804),
    [anon_sym_type] = ACTIONS(4806),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4808),
    [anon_sym_let] = ACTIONS(4810),
    [anon_sym_let_BANG] = ACTIONS(4812),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4818),
    [anon_sym_extern] = ACTIONS(4820),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4826),
    [anon_sym_POUNDload] = ACTIONS(4826),
    [anon_sym_POUNDtime] = ACTIONS(4826),
    [anon_sym_POUNDI] = ACTIONS(4826),
    [anon_sym_POUNDhelp] = ACTIONS(4826),
    [anon_sym_POUNDquit] = ACTIONS(4826),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1108)] = {
    [sym__module_body_elem] = STATE(8636),
    [sym__module_expression] = STATE(8690),
    [sym_module_abbrev] = STATE(8690),
    [sym_module_defn] = STATE(8690),
    [sym_import_decl] = STATE(8690),
    [sym_attributes] = STATE(15049),
    [sym__attribute_set] = STATE(12910),
    [sym_value_declaration] = STATE(8692),
    [sym_do] = STATE(8637),
    [sym_function_or_value_defn] = STATE(1111),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(934),
    [sym_literal_expression] = STATE(4278),
    [sym_srtp_call_expression] = STATE(4278),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(4038),
    [sym_tuple_expression] = STATE(4278),
    [sym_brace_expression] = STATE(4278),
    [sym_anon_record_expression] = STATE(4278),
    [sym_prefixed_expression] = STATE(4278),
    [sym_typecast_expression] = STATE(4278),
    [sym_for_expression] = STATE(4278),
    [sym_while_expression] = STATE(4278),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4278),
    [sym_fun_expression] = STATE(4278),
    [sym_try_expression] = STATE(4278),
    [sym_match_expression] = STATE(4278),
    [sym_function_expression] = STATE(4278),
    [sym_mutate_expression] = STATE(4278),
    [sym_index_expression] = STATE(4038),
    [sym_typed_expression] = STATE(4278),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4278),
    [sym_list_expression] = STATE(4278),
    [sym_array_expression] = STATE(4278),
    [sym_begin_end_expression] = STATE(4278),
    [sym_paren_expression] = STATE(4278),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4278),
    [sym_dot_expression] = STATE(4038),
    [sym_infix_expression] = STATE(4278),
    [sym_ce_expression] = STATE(4278),
    [sym_sequential_expression] = STATE(4338),
    [sym_exception_definition] = STATE(8690),
    [sym_type_definition] = STATE(8690),
    [sym_extern_binding] = STATE(8690),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4278),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1108),
    [sym_compiler_directive_decl] = STATE(1108),
    [sym_fsi_directive_decl] = STATE(1108),
    [sym_preproc_line] = STATE(1108),
    [sym_preproc_if_in_expression] = STATE(4338),
    [sym_preproc_if_in_module_body] = STATE(8690),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(4091),
    [anon_sym_null] = ACTIONS(4290),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_open] = ACTIONS(4095),
    [anon_sym_type] = ACTIONS(4097),
    [anon_sym_LBRACK_LT] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(4099),
    [anon_sym_let] = ACTIONS(4101),
    [anon_sym_let_BANG] = ACTIONS(4103),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(4109),
    [anon_sym_extern] = ACTIONS(4111),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(4117),
    [anon_sym_POUNDload] = ACTIONS(4117),
    [anon_sym_POUNDtime] = ACTIONS(4117),
    [anon_sym_POUNDI] = ACTIONS(4117),
    [anon_sym_POUNDhelp] = ACTIONS(4117),
    [anon_sym_POUNDquit] = ACTIONS(4117),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(4292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1109)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(222),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1109),
    [sym_compiler_directive_decl] = STATE(1109),
    [sym_fsi_directive_decl] = STATE(1109),
    [sym_preproc_line] = STATE(1109),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(1337),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [anon_sym_POUNDelse] = ACTIONS(5121),
    [sym__newline] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1110)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(347),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1110),
    [sym_compiler_directive_decl] = STATE(1110),
    [sym_fsi_directive_decl] = STATE(1110),
    [sym_preproc_line] = STATE(1110),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_declaration_expression_repeat1] = STATE(1293),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [anon_sym_POUNDelse] = ACTIONS(5121),
    [sym__newline] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1111)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(222),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1111),
    [sym_compiler_directive_decl] = STATE(1111),
    [sym_fsi_directive_decl] = STATE(1111),
    [sym_preproc_line] = STATE(1111),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(1337),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [sym__newline] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1112)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(179),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1112),
    [sym_compiler_directive_decl] = STATE(1112),
    [sym_fsi_directive_decl] = STATE(1112),
    [sym_preproc_line] = STATE(1112),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(1318),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(5121),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_namespace] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1113)] = {
    [sym_function_or_value_defn] = STATE(1151),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(237),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym__list_elements] = STATE(21351),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_range_expression] = STATE(21253),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym__comp_or_range_expression] = STATE(21351),
    [sym_comp_declaration_expression] = STATE(21253),
    [sym_short_comp_expression] = STATE(21253),
    [sym_slice_ranges] = STATE(21351),
    [sym__slice_range_special] = STATE(18369),
    [sym_slice_range] = STATE(17075),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1113),
    [sym_compiler_directive_decl] = STATE(1113),
    [sym_fsi_directive_decl] = STATE(1113),
    [sym_preproc_line] = STATE(1113),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(5125),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(5127),
    [anon_sym_use_BANG] = ACTIONS(5129),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5131),
    [anon_sym_STAR] = ACTIONS(5133),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__newline] = ACTIONS(5135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1114)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(162),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1114),
    [sym_compiler_directive_decl] = STATE(1114),
    [sym_fsi_directive_decl] = STATE(1114),
    [sym_preproc_line] = STATE(1114),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_declaration_expression_repeat1] = STATE(1330),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [anon_sym_POUNDelse] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1115)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(162),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1115),
    [sym_compiler_directive_decl] = STATE(1115),
    [sym_fsi_directive_decl] = STATE(1115),
    [sym_preproc_line] = STATE(1115),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_declaration_expression_repeat1] = STATE(1330),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5137),
    [anon_sym_module] = ACTIONS(5137),
    [anon_sym_null] = ACTIONS(5137),
    [anon_sym_begin] = ACTIONS(5137),
    [anon_sym_open] = ACTIONS(5137),
    [anon_sym_type] = ACTIONS(5137),
    [anon_sym_LBRACK_LT] = ACTIONS(5139),
    [anon_sym_do] = ACTIONS(5137),
    [anon_sym_let] = ACTIONS(5137),
    [anon_sym_let_BANG] = ACTIONS(5139),
    [anon_sym_LPAREN] = ACTIONS(5137),
    [anon_sym_AMP] = ACTIONS(5137),
    [anon_sym_LBRACK] = ACTIONS(5137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5139),
    [anon_sym_LBRACE] = ACTIONS(5137),
    [anon_sym_LT_AT] = ACTIONS(5137),
    [anon_sym_LT_AT_AT] = ACTIONS(5139),
    [anon_sym_CARET] = ACTIONS(5139),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5139),
    [anon_sym_new] = ACTIONS(5137),
    [anon_sym_return] = ACTIONS(5137),
    [anon_sym_return_BANG] = ACTIONS(5139),
    [anon_sym_yield] = ACTIONS(5137),
    [anon_sym_yield_BANG] = ACTIONS(5139),
    [anon_sym_lazy] = ACTIONS(5137),
    [anon_sym_assert] = ACTIONS(5137),
    [anon_sym_upcast] = ACTIONS(5137),
    [anon_sym_downcast] = ACTIONS(5137),
    [anon_sym_fixed] = ACTIONS(5137),
    [anon_sym_for] = ACTIONS(5137),
    [anon_sym_while] = ACTIONS(5137),
    [anon_sym_while_BANG] = ACTIONS(5139),
    [anon_sym_if] = ACTIONS(5137),
    [anon_sym_fun] = ACTIONS(5137),
    [anon_sym_try] = ACTIONS(5137),
    [anon_sym_match] = ACTIONS(5137),
    [anon_sym_match_BANG] = ACTIONS(5139),
    [anon_sym_function] = ACTIONS(5137),
    [anon_sym_use] = ACTIONS(5137),
    [anon_sym_use_BANG] = ACTIONS(5139),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5139),
    [anon_sym_AMP_AMP] = ACTIONS(5139),
    [anon_sym_exception] = ACTIONS(5137),
    [anon_sym_extern] = ACTIONS(5137),
    [aux_sym_char_token1] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5137),
    [anon_sym_DQUOTE] = ACTIONS(5137),
    [anon_sym_AT_DQUOTE] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [sym_bool] = ACTIONS(5137),
    [sym_unit] = ACTIONS(5139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5137),
    [sym_op_identifier] = ACTIONS(5139),
    [anon_sym_PLUS] = ACTIONS(5137),
    [anon_sym_DASH] = ACTIONS(5137),
    [anon_sym_PLUS_DOT] = ACTIONS(5139),
    [anon_sym_DASH_DOT] = ACTIONS(5139),
    [anon_sym_PERCENT] = ACTIONS(5137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5139),
    [anon_sym_TILDE] = ACTIONS(5139),
    [aux_sym_prefix_op_token1] = ACTIONS(5139),
    [sym_int] = ACTIONS(5137),
    [sym_xint] = ACTIONS(5139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5139),
    [anon_sym_POUNDload] = ACTIONS(5139),
    [anon_sym_POUNDtime] = ACTIONS(5139),
    [anon_sym_POUNDI] = ACTIONS(5139),
    [anon_sym_POUNDhelp] = ACTIONS(5139),
    [anon_sym_POUNDquit] = ACTIONS(5139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5139),
    [anon_sym_POUNDendif] = ACTIONS(5139),
    [anon_sym_POUNDelse] = ACTIONS(5139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5139),
  },
  [STATE(1116)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(179),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1116),
    [sym_compiler_directive_decl] = STATE(1116),
    [sym_fsi_directive_decl] = STATE(1116),
    [sym_preproc_line] = STATE(1116),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(1318),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(5139),
    [sym_identifier] = ACTIONS(5137),
    [anon_sym_namespace] = ACTIONS(5137),
    [anon_sym_module] = ACTIONS(5137),
    [anon_sym_null] = ACTIONS(5137),
    [anon_sym_begin] = ACTIONS(5137),
    [anon_sym_open] = ACTIONS(5137),
    [anon_sym_type] = ACTIONS(5137),
    [anon_sym_LBRACK_LT] = ACTIONS(5139),
    [anon_sym_do] = ACTIONS(5137),
    [anon_sym_let] = ACTIONS(5137),
    [anon_sym_let_BANG] = ACTIONS(5139),
    [anon_sym_LPAREN] = ACTIONS(5137),
    [anon_sym_AMP] = ACTIONS(5137),
    [anon_sym_LBRACK] = ACTIONS(5137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5139),
    [anon_sym_LBRACE] = ACTIONS(5137),
    [anon_sym_LT_AT] = ACTIONS(5137),
    [anon_sym_LT_AT_AT] = ACTIONS(5139),
    [anon_sym_CARET] = ACTIONS(5139),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5139),
    [anon_sym_new] = ACTIONS(5137),
    [anon_sym_return] = ACTIONS(5137),
    [anon_sym_return_BANG] = ACTIONS(5139),
    [anon_sym_yield] = ACTIONS(5137),
    [anon_sym_yield_BANG] = ACTIONS(5139),
    [anon_sym_lazy] = ACTIONS(5137),
    [anon_sym_assert] = ACTIONS(5137),
    [anon_sym_upcast] = ACTIONS(5137),
    [anon_sym_downcast] = ACTIONS(5137),
    [anon_sym_fixed] = ACTIONS(5137),
    [anon_sym_for] = ACTIONS(5137),
    [anon_sym_while] = ACTIONS(5137),
    [anon_sym_while_BANG] = ACTIONS(5139),
    [anon_sym_if] = ACTIONS(5137),
    [anon_sym_fun] = ACTIONS(5137),
    [anon_sym_try] = ACTIONS(5137),
    [anon_sym_match] = ACTIONS(5137),
    [anon_sym_match_BANG] = ACTIONS(5139),
    [anon_sym_function] = ACTIONS(5137),
    [anon_sym_use] = ACTIONS(5137),
    [anon_sym_use_BANG] = ACTIONS(5139),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5139),
    [anon_sym_AMP_AMP] = ACTIONS(5139),
    [anon_sym_exception] = ACTIONS(5137),
    [anon_sym_extern] = ACTIONS(5137),
    [aux_sym_char_token1] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5137),
    [anon_sym_DQUOTE] = ACTIONS(5137),
    [anon_sym_AT_DQUOTE] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [sym_bool] = ACTIONS(5137),
    [sym_unit] = ACTIONS(5139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5137),
    [sym_op_identifier] = ACTIONS(5139),
    [anon_sym_PLUS] = ACTIONS(5137),
    [anon_sym_DASH] = ACTIONS(5137),
    [anon_sym_PLUS_DOT] = ACTIONS(5139),
    [anon_sym_DASH_DOT] = ACTIONS(5139),
    [anon_sym_PERCENT] = ACTIONS(5137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5139),
    [anon_sym_TILDE] = ACTIONS(5139),
    [aux_sym_prefix_op_token1] = ACTIONS(5139),
    [sym_int] = ACTIONS(5137),
    [sym_xint] = ACTIONS(5139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5139),
    [anon_sym_POUNDload] = ACTIONS(5139),
    [anon_sym_POUNDtime] = ACTIONS(5139),
    [anon_sym_POUNDI] = ACTIONS(5139),
    [anon_sym_POUNDhelp] = ACTIONS(5139),
    [anon_sym_POUNDquit] = ACTIONS(5139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5139),
  },
  [STATE(1117)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(568),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1117),
    [sym_compiler_directive_decl] = STATE(1117),
    [sym_fsi_directive_decl] = STATE(1117),
    [sym_preproc_line] = STATE(1117),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_declaration_expression_repeat1] = STATE(1309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [sym__newline] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1118)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(199),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1118),
    [sym_compiler_directive_decl] = STATE(1118),
    [sym_fsi_directive_decl] = STATE(1118),
    [sym_preproc_line] = STATE(1118),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(1324),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(5121),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1119)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(199),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1119),
    [sym_compiler_directive_decl] = STATE(1119),
    [sym_fsi_directive_decl] = STATE(1119),
    [sym_preproc_line] = STATE(1119),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(1324),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [ts_builtin_sym_end] = ACTIONS(5139),
    [sym_identifier] = ACTIONS(5137),
    [anon_sym_module] = ACTIONS(5137),
    [anon_sym_null] = ACTIONS(5137),
    [anon_sym_begin] = ACTIONS(5137),
    [anon_sym_open] = ACTIONS(5137),
    [anon_sym_type] = ACTIONS(5137),
    [anon_sym_LBRACK_LT] = ACTIONS(5139),
    [anon_sym_do] = ACTIONS(5137),
    [anon_sym_let] = ACTIONS(5137),
    [anon_sym_let_BANG] = ACTIONS(5139),
    [anon_sym_LPAREN] = ACTIONS(5137),
    [anon_sym_AMP] = ACTIONS(5137),
    [anon_sym_LBRACK] = ACTIONS(5137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5139),
    [anon_sym_LBRACE] = ACTIONS(5137),
    [anon_sym_LT_AT] = ACTIONS(5137),
    [anon_sym_LT_AT_AT] = ACTIONS(5139),
    [anon_sym_CARET] = ACTIONS(5139),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5139),
    [anon_sym_new] = ACTIONS(5137),
    [anon_sym_return] = ACTIONS(5137),
    [anon_sym_return_BANG] = ACTIONS(5139),
    [anon_sym_yield] = ACTIONS(5137),
    [anon_sym_yield_BANG] = ACTIONS(5139),
    [anon_sym_lazy] = ACTIONS(5137),
    [anon_sym_assert] = ACTIONS(5137),
    [anon_sym_upcast] = ACTIONS(5137),
    [anon_sym_downcast] = ACTIONS(5137),
    [anon_sym_fixed] = ACTIONS(5137),
    [anon_sym_for] = ACTIONS(5137),
    [anon_sym_while] = ACTIONS(5137),
    [anon_sym_while_BANG] = ACTIONS(5139),
    [anon_sym_if] = ACTIONS(5137),
    [anon_sym_fun] = ACTIONS(5137),
    [anon_sym_try] = ACTIONS(5137),
    [anon_sym_match] = ACTIONS(5137),
    [anon_sym_match_BANG] = ACTIONS(5139),
    [anon_sym_function] = ACTIONS(5137),
    [anon_sym_use] = ACTIONS(5137),
    [anon_sym_use_BANG] = ACTIONS(5139),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5139),
    [anon_sym_AMP_AMP] = ACTIONS(5139),
    [anon_sym_exception] = ACTIONS(5137),
    [anon_sym_extern] = ACTIONS(5137),
    [aux_sym_char_token1] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5137),
    [anon_sym_DQUOTE] = ACTIONS(5137),
    [anon_sym_AT_DQUOTE] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [sym_bool] = ACTIONS(5137),
    [sym_unit] = ACTIONS(5139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5137),
    [sym_op_identifier] = ACTIONS(5139),
    [anon_sym_PLUS] = ACTIONS(5137),
    [anon_sym_DASH] = ACTIONS(5137),
    [anon_sym_PLUS_DOT] = ACTIONS(5139),
    [anon_sym_DASH_DOT] = ACTIONS(5139),
    [anon_sym_PERCENT] = ACTIONS(5137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5139),
    [anon_sym_TILDE] = ACTIONS(5139),
    [aux_sym_prefix_op_token1] = ACTIONS(5139),
    [sym_int] = ACTIONS(5137),
    [sym_xint] = ACTIONS(5139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5139),
    [anon_sym_POUNDload] = ACTIONS(5139),
    [anon_sym_POUNDtime] = ACTIONS(5139),
    [anon_sym_POUNDI] = ACTIONS(5139),
    [anon_sym_POUNDhelp] = ACTIONS(5139),
    [anon_sym_POUNDquit] = ACTIONS(5139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5139),
  },
  [STATE(1120)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(213),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1120),
    [sym_compiler_directive_decl] = STATE(1120),
    [sym_fsi_directive_decl] = STATE(1120),
    [sym_preproc_line] = STATE(1120),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_declaration_expression_repeat1] = STATE(1334),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5137),
    [anon_sym_module] = ACTIONS(5137),
    [anon_sym_null] = ACTIONS(5137),
    [anon_sym_begin] = ACTIONS(5137),
    [anon_sym_open] = ACTIONS(5137),
    [anon_sym_type] = ACTIONS(5137),
    [anon_sym_LBRACK_LT] = ACTIONS(5139),
    [anon_sym_do] = ACTIONS(5137),
    [anon_sym_let] = ACTIONS(5137),
    [anon_sym_let_BANG] = ACTIONS(5139),
    [anon_sym_LPAREN] = ACTIONS(5137),
    [anon_sym_AMP] = ACTIONS(5137),
    [anon_sym_LBRACK] = ACTIONS(5137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5139),
    [anon_sym_LBRACE] = ACTIONS(5137),
    [anon_sym_LT_AT] = ACTIONS(5137),
    [anon_sym_LT_AT_AT] = ACTIONS(5139),
    [anon_sym_CARET] = ACTIONS(5139),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5139),
    [anon_sym_new] = ACTIONS(5137),
    [anon_sym_return] = ACTIONS(5137),
    [anon_sym_return_BANG] = ACTIONS(5139),
    [anon_sym_yield] = ACTIONS(5137),
    [anon_sym_yield_BANG] = ACTIONS(5139),
    [anon_sym_lazy] = ACTIONS(5137),
    [anon_sym_assert] = ACTIONS(5137),
    [anon_sym_upcast] = ACTIONS(5137),
    [anon_sym_downcast] = ACTIONS(5137),
    [anon_sym_fixed] = ACTIONS(5137),
    [anon_sym_for] = ACTIONS(5137),
    [anon_sym_while] = ACTIONS(5137),
    [anon_sym_while_BANG] = ACTIONS(5139),
    [anon_sym_if] = ACTIONS(5137),
    [anon_sym_fun] = ACTIONS(5137),
    [anon_sym_try] = ACTIONS(5137),
    [anon_sym_match] = ACTIONS(5137),
    [anon_sym_match_BANG] = ACTIONS(5139),
    [anon_sym_function] = ACTIONS(5137),
    [anon_sym_use] = ACTIONS(5137),
    [anon_sym_use_BANG] = ACTIONS(5139),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5139),
    [anon_sym_AMP_AMP] = ACTIONS(5139),
    [anon_sym_exception] = ACTIONS(5137),
    [anon_sym_extern] = ACTIONS(5137),
    [aux_sym_char_token1] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5137),
    [anon_sym_DQUOTE] = ACTIONS(5137),
    [anon_sym_AT_DQUOTE] = ACTIONS(5139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5139),
    [sym_bool] = ACTIONS(5137),
    [sym_unit] = ACTIONS(5139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5137),
    [sym_op_identifier] = ACTIONS(5139),
    [anon_sym_PLUS] = ACTIONS(5137),
    [anon_sym_DASH] = ACTIONS(5137),
    [anon_sym_PLUS_DOT] = ACTIONS(5139),
    [anon_sym_DASH_DOT] = ACTIONS(5139),
    [anon_sym_PERCENT] = ACTIONS(5137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5139),
    [anon_sym_TILDE] = ACTIONS(5139),
    [aux_sym_prefix_op_token1] = ACTIONS(5139),
    [sym_int] = ACTIONS(5137),
    [sym_xint] = ACTIONS(5139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5139),
    [anon_sym_POUNDload] = ACTIONS(5139),
    [anon_sym_POUNDtime] = ACTIONS(5139),
    [anon_sym_POUNDI] = ACTIONS(5139),
    [anon_sym_POUNDhelp] = ACTIONS(5139),
    [anon_sym_POUNDquit] = ACTIONS(5139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5139),
    [anon_sym_POUNDendif] = ACTIONS(5139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5139),
  },
  [STATE(1121)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(213),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1121),
    [sym_compiler_directive_decl] = STATE(1121),
    [sym_fsi_directive_decl] = STATE(1121),
    [sym_preproc_line] = STATE(1121),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_declaration_expression_repeat1] = STATE(1334),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5119),
    [anon_sym_module] = ACTIONS(5119),
    [anon_sym_null] = ACTIONS(5119),
    [anon_sym_begin] = ACTIONS(5119),
    [anon_sym_open] = ACTIONS(5119),
    [anon_sym_type] = ACTIONS(5119),
    [anon_sym_LBRACK_LT] = ACTIONS(5121),
    [anon_sym_do] = ACTIONS(5119),
    [anon_sym_let] = ACTIONS(5119),
    [anon_sym_let_BANG] = ACTIONS(5121),
    [anon_sym_LPAREN] = ACTIONS(5119),
    [anon_sym_AMP] = ACTIONS(5119),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5121),
    [anon_sym_LBRACE] = ACTIONS(5119),
    [anon_sym_LT_AT] = ACTIONS(5119),
    [anon_sym_LT_AT_AT] = ACTIONS(5121),
    [anon_sym_CARET] = ACTIONS(5121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5121),
    [anon_sym_new] = ACTIONS(5119),
    [anon_sym_return] = ACTIONS(5119),
    [anon_sym_return_BANG] = ACTIONS(5121),
    [anon_sym_yield] = ACTIONS(5119),
    [anon_sym_yield_BANG] = ACTIONS(5121),
    [anon_sym_lazy] = ACTIONS(5119),
    [anon_sym_assert] = ACTIONS(5119),
    [anon_sym_upcast] = ACTIONS(5119),
    [anon_sym_downcast] = ACTIONS(5119),
    [anon_sym_fixed] = ACTIONS(5119),
    [anon_sym_for] = ACTIONS(5119),
    [anon_sym_while] = ACTIONS(5119),
    [anon_sym_while_BANG] = ACTIONS(5121),
    [anon_sym_if] = ACTIONS(5119),
    [anon_sym_fun] = ACTIONS(5119),
    [anon_sym_try] = ACTIONS(5119),
    [anon_sym_match] = ACTIONS(5119),
    [anon_sym_match_BANG] = ACTIONS(5121),
    [anon_sym_function] = ACTIONS(5119),
    [anon_sym_use] = ACTIONS(5119),
    [anon_sym_use_BANG] = ACTIONS(5121),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(5121),
    [anon_sym_AMP_AMP] = ACTIONS(5121),
    [anon_sym_exception] = ACTIONS(5119),
    [anon_sym_extern] = ACTIONS(5119),
    [aux_sym_char_token1] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5119),
    [anon_sym_DQUOTE] = ACTIONS(5119),
    [anon_sym_AT_DQUOTE] = ACTIONS(5121),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5121),
    [sym_bool] = ACTIONS(5119),
    [sym_unit] = ACTIONS(5121),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5119),
    [sym_op_identifier] = ACTIONS(5121),
    [anon_sym_PLUS] = ACTIONS(5119),
    [anon_sym_DASH] = ACTIONS(5119),
    [anon_sym_PLUS_DOT] = ACTIONS(5121),
    [anon_sym_DASH_DOT] = ACTIONS(5121),
    [anon_sym_PERCENT] = ACTIONS(5119),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5121),
    [anon_sym_TILDE] = ACTIONS(5121),
    [aux_sym_prefix_op_token1] = ACTIONS(5121),
    [sym_int] = ACTIONS(5119),
    [sym_xint] = ACTIONS(5121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5121),
    [anon_sym_POUNDload] = ACTIONS(5121),
    [anon_sym_POUNDtime] = ACTIONS(5121),
    [anon_sym_POUNDI] = ACTIONS(5121),
    [anon_sym_POUNDhelp] = ACTIONS(5121),
    [anon_sym_POUNDquit] = ACTIONS(5121),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5121),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5121),
  },
  [STATE(1122)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20193),
    [sym_with_field_expression] = STATE(20193),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20211),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20193),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1122),
    [sym_compiler_directive_decl] = STATE(1122),
    [sym_fsi_directive_decl] = STATE(1122),
    [sym_preproc_line] = STATE(1122),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1123)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20018),
    [sym_with_field_expression] = STATE(20018),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20035),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20018),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1123),
    [sym_compiler_directive_decl] = STATE(1123),
    [sym_fsi_directive_decl] = STATE(1123),
    [sym_preproc_line] = STATE(1123),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1124)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19402),
    [sym_with_field_expression] = STATE(19402),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19427),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19402),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1124),
    [sym_compiler_directive_decl] = STATE(1124),
    [sym_fsi_directive_decl] = STATE(1124),
    [sym_preproc_line] = STATE(1124),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1125)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20325),
    [sym_with_field_expression] = STATE(20325),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20343),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20325),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1125),
    [sym_compiler_directive_decl] = STATE(1125),
    [sym_fsi_directive_decl] = STATE(1125),
    [sym_preproc_line] = STATE(1125),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1126)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19174),
    [sym_with_field_expression] = STATE(19174),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19199),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19174),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1126),
    [sym_compiler_directive_decl] = STATE(1126),
    [sym_fsi_directive_decl] = STATE(1126),
    [sym_preproc_line] = STATE(1126),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1127)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(21902),
    [sym_with_field_expression] = STATE(21902),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(21950),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(21902),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1127),
    [sym_compiler_directive_decl] = STATE(1127),
    [sym_fsi_directive_decl] = STATE(1127),
    [sym_preproc_line] = STATE(1127),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1128)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(22319),
    [sym_with_field_expression] = STATE(22319),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(22378),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22319),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1128),
    [sym_compiler_directive_decl] = STATE(1128),
    [sym_fsi_directive_decl] = STATE(1128),
    [sym_preproc_line] = STATE(1128),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1129)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20369),
    [sym_with_field_expression] = STATE(20369),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20387),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20369),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1129),
    [sym_compiler_directive_decl] = STATE(1129),
    [sym_fsi_directive_decl] = STATE(1129),
    [sym_preproc_line] = STATE(1129),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1130)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19925),
    [sym_with_field_expression] = STATE(19925),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19944),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19925),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1130),
    [sym_compiler_directive_decl] = STATE(1130),
    [sym_fsi_directive_decl] = STATE(1130),
    [sym_preproc_line] = STATE(1130),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1131)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20105),
    [sym_with_field_expression] = STATE(20105),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20123),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20105),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1131),
    [sym_compiler_directive_decl] = STATE(1131),
    [sym_fsi_directive_decl] = STATE(1131),
    [sym_preproc_line] = STATE(1131),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1132)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19870),
    [sym_with_field_expression] = STATE(19870),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19889),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19870),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1132),
    [sym_compiler_directive_decl] = STATE(1132),
    [sym_fsi_directive_decl] = STATE(1132),
    [sym_preproc_line] = STATE(1132),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1133)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20237),
    [sym_with_field_expression] = STATE(20237),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20255),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20237),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1133),
    [sym_compiler_directive_decl] = STATE(1133),
    [sym_fsi_directive_decl] = STATE(1133),
    [sym_preproc_line] = STATE(1133),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1134)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(22110),
    [sym_with_field_expression] = STATE(22110),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(22135),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22110),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1134),
    [sym_compiler_directive_decl] = STATE(1134),
    [sym_fsi_directive_decl] = STATE(1134),
    [sym_preproc_line] = STATE(1134),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1135)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20413),
    [sym_with_field_expression] = STATE(20413),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20431),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20413),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1135),
    [sym_compiler_directive_decl] = STATE(1135),
    [sym_fsi_directive_decl] = STATE(1135),
    [sym_preproc_line] = STATE(1135),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1136)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19591),
    [sym_with_field_expression] = STATE(19591),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19611),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19591),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1136),
    [sym_compiler_directive_decl] = STATE(1136),
    [sym_fsi_directive_decl] = STATE(1136),
    [sym_preproc_line] = STATE(1136),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1137)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19807),
    [sym_with_field_expression] = STATE(19807),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19826),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19807),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1137),
    [sym_compiler_directive_decl] = STATE(1137),
    [sym_fsi_directive_decl] = STATE(1137),
    [sym_preproc_line] = STATE(1137),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1138)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20149),
    [sym_with_field_expression] = STATE(20149),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20167),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20149),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1138),
    [sym_compiler_directive_decl] = STATE(1138),
    [sym_fsi_directive_decl] = STATE(1138),
    [sym_preproc_line] = STATE(1138),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1139)] = {
    [sym_function_or_value_defn] = STATE(1151),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(236),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_range_expression] = STATE(21253),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym__comp_or_range_expression] = STATE(20979),
    [sym_comp_declaration_expression] = STATE(21253),
    [sym_short_comp_expression] = STATE(21253),
    [sym_slice_ranges] = STATE(20979),
    [sym__slice_range_special] = STATE(18369),
    [sym_slice_range] = STATE(17075),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1139),
    [sym_compiler_directive_decl] = STATE(1139),
    [sym_fsi_directive_decl] = STATE(1139),
    [sym_preproc_line] = STATE(1139),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(5125),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(5127),
    [anon_sym_use_BANG] = ACTIONS(5129),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5131),
    [anon_sym_STAR] = ACTIONS(5133),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1140)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19974),
    [sym_with_field_expression] = STATE(19974),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19992),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19974),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1140),
    [sym_compiler_directive_decl] = STATE(1140),
    [sym_fsi_directive_decl] = STATE(1140),
    [sym_preproc_line] = STATE(1140),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1141)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19288),
    [sym_with_field_expression] = STATE(19288),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19313),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19288),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1141),
    [sym_compiler_directive_decl] = STATE(1141),
    [sym_fsi_directive_decl] = STATE(1141),
    [sym_preproc_line] = STATE(1141),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1142)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20457),
    [sym_with_field_expression] = STATE(20457),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20475),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20457),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1142),
    [sym_compiler_directive_decl] = STATE(1142),
    [sym_fsi_directive_decl] = STATE(1142),
    [sym_preproc_line] = STATE(1142),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1143)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19514),
    [sym_with_field_expression] = STATE(19514),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19534),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19514),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1143),
    [sym_compiler_directive_decl] = STATE(1143),
    [sym_fsi_directive_decl] = STATE(1143),
    [sym_preproc_line] = STATE(1143),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1144)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19739),
    [sym_with_field_expression] = STATE(19739),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19759),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19739),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1144),
    [sym_compiler_directive_decl] = STATE(1144),
    [sym_fsi_directive_decl] = STATE(1144),
    [sym_preproc_line] = STATE(1144),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1145)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20281),
    [sym_with_field_expression] = STATE(20281),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20299),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20281),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1145),
    [sym_compiler_directive_decl] = STATE(1145),
    [sym_fsi_directive_decl] = STATE(1145),
    [sym_preproc_line] = STATE(1145),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1146)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(22111),
    [sym_with_field_expression] = STATE(22111),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19814),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22111),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1146),
    [sym_compiler_directive_decl] = STATE(1146),
    [sym_fsi_directive_decl] = STATE(1146),
    [sym_preproc_line] = STATE(1146),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1147)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(22783),
    [sym_with_field_expression] = STATE(22783),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19083),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22783),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1147),
    [sym_compiler_directive_decl] = STATE(1147),
    [sym_fsi_directive_decl] = STATE(1147),
    [sym_preproc_line] = STATE(1147),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1148)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(20061),
    [sym_with_field_expression] = STATE(20061),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(20079),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20061),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1148),
    [sym_compiler_directive_decl] = STATE(1148),
    [sym_fsi_directive_decl] = STATE(1148),
    [sym_preproc_line] = STATE(1148),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1149)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(276),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_object_expression] = STATE(19667),
    [sym_with_field_expression] = STATE(19667),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym__comp_expression_block] = STATE(19687),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(17060),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19667),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(4609),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1149),
    [sym_compiler_directive_decl] = STATE(1149),
    [sym_fsi_directive_decl] = STATE(1149),
    [sym_preproc_line] = STATE(1149),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5141),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(5147),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1150)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(616),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_type_arguments] = STATE(8754),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4342),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1150),
    [sym_compiler_directive_decl] = STATE(1150),
    [sym_fsi_directive_decl] = STATE(1150),
    [sym_preproc_line] = STATE(1150),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_compound_type_repeat1] = STATE(8723),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5159),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_GT_RBRACK] = ACTIONS(5161),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_DASH_GT] = ACTIONS(5167),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_STAR] = ACTIONS(5169),
    [anon_sym_LT] = ACTIONS(5171),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(5173),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(5161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1151)] = {
    [sym_function_or_value_defn] = STATE(1151),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(273),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_range_expression] = STATE(21253),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym__comp_or_range_expression] = STATE(21259),
    [sym_comp_declaration_expression] = STATE(21253),
    [sym_short_comp_expression] = STATE(21253),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1151),
    [sym_compiler_directive_decl] = STATE(1151),
    [sym_fsi_directive_decl] = STATE(1151),
    [sym_preproc_line] = STATE(1151),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(1158),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(5125),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(5127),
    [anon_sym_use_BANG] = ACTIONS(5129),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1152)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(536),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_type_arguments] = STATE(8779),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4491),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1152),
    [sym_compiler_directive_decl] = STATE(1152),
    [sym_fsi_directive_decl] = STATE(1152),
    [sym_preproc_line] = STATE(1152),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_compound_type_repeat1] = STATE(8753),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5179),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_DASH_GT] = ACTIONS(5181),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_STAR] = ACTIONS(5183),
    [anon_sym_LT] = ACTIONS(5185),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(5187),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5189),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1153)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(284),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(18151),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1153),
    [sym_compiler_directive_decl] = STATE(1153),
    [sym_fsi_directive_decl] = STATE(1153),
    [sym_preproc_line] = STATE(1153),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_declaration_expression_repeat1] = STATE(1154),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1154)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(272),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(18954),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1154),
    [sym_compiler_directive_decl] = STATE(1154),
    [sym_fsi_directive_decl] = STATE(1154),
    [sym_preproc_line] = STATE(1154),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1155)] = {
    [sym_function_or_value_defn] = STATE(1156),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(300),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym__comp_or_range_expression] = STATE(18954),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1155),
    [sym_compiler_directive_decl] = STATE(1155),
    [sym_fsi_directive_decl] = STATE(1155),
    [sym_preproc_line] = STATE(1155),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(5195),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(5197),
    [anon_sym_use_BANG] = ACTIONS(5199),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1156)] = {
    [sym_function_or_value_defn] = STATE(1156),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(379),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym__comp_or_range_expression] = STATE(18151),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1156),
    [sym_compiler_directive_decl] = STATE(1156),
    [sym_fsi_directive_decl] = STATE(1156),
    [sym_preproc_line] = STATE(1156),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_declaration_expression_repeat1] = STATE(1155),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(5195),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(5197),
    [anon_sym_use_BANG] = ACTIONS(5199),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1157)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(591),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_type_arguments] = STATE(8779),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4491),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1157),
    [sym_compiler_directive_decl] = STATE(1157),
    [sym_fsi_directive_decl] = STATE(1157),
    [sym_preproc_line] = STATE(1157),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_compound_type_repeat1] = STATE(8753),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5179),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_DASH_GT] = ACTIONS(5181),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_STAR] = ACTIONS(5183),
    [anon_sym_LT] = ACTIONS(5185),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(5187),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5205),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1158)] = {
    [sym_function_or_value_defn] = STATE(1151),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(277),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_range_expression] = STATE(21253),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym__comp_or_range_expression] = STATE(19848),
    [sym_comp_declaration_expression] = STATE(21253),
    [sym_short_comp_expression] = STATE(21253),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1158),
    [sym_compiler_directive_decl] = STATE(1158),
    [sym_fsi_directive_decl] = STATE(1158),
    [sym_preproc_line] = STATE(1158),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(5125),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(5127),
    [anon_sym_use_BANG] = ACTIONS(5129),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1159)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(926),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_type_arguments] = STATE(8779),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4491),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1159),
    [sym_compiler_directive_decl] = STATE(1159),
    [sym_fsi_directive_decl] = STATE(1159),
    [sym_preproc_line] = STATE(1159),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_compound_type_repeat1] = STATE(8753),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5179),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_DASH_GT] = ACTIONS(5181),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_STAR] = ACTIONS(5183),
    [anon_sym_LT] = ACTIONS(5185),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(5187),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5207),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1160)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(432),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19998),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1160),
    [sym_compiler_directive_decl] = STATE(1160),
    [sym_fsi_directive_decl] = STATE(1160),
    [sym_preproc_line] = STATE(1160),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1161)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(340),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(21150),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1161),
    [sym_compiler_directive_decl] = STATE(1161),
    [sym_fsi_directive_decl] = STATE(1161),
    [sym_preproc_line] = STATE(1161),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1162)] = {
    [sym_do] = STATE(8680),
    [sym_function_or_value_defn] = STATE(1119),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(211),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1162),
    [sym_compiler_directive_decl] = STATE(1162),
    [sym_fsi_directive_decl] = STATE(1162),
    [sym_preproc_line] = STATE(1162),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(5213),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_type] = ACTIONS(5215),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5217),
    [anon_sym_extern] = ACTIONS(5219),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1163)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1163),
    [sym_compiler_directive_decl] = STATE(1163),
    [sym_fsi_directive_decl] = STATE(1163),
    [sym_preproc_line] = STATE(1163),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19234),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1168),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5221),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1164)] = {
    [sym_do] = STATE(20949),
    [sym_function_or_value_defn] = STATE(1261),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(641),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1164),
    [sym_compiler_directive_decl] = STATE(1164),
    [sym_fsi_directive_decl] = STATE(1164),
    [sym_preproc_line] = STATE(1164),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_module] = ACTIONS(5223),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_type] = ACTIONS(5225),
    [anon_sym_do] = ACTIONS(5097),
    [anon_sym_let] = ACTIONS(5099),
    [anon_sym_let_BANG] = ACTIONS(5101),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5227),
    [anon_sym_extern] = ACTIONS(5229),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1165)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1165),
    [sym_compiler_directive_decl] = STATE(1165),
    [sym_fsi_directive_decl] = STATE(1165),
    [sym_preproc_line] = STATE(1165),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21801),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5231),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1166)] = {
    [sym_do] = STATE(8725),
    [sym_function_or_value_defn] = STATE(1120),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(200),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1166),
    [sym_compiler_directive_decl] = STATE(1166),
    [sym_fsi_directive_decl] = STATE(1166),
    [sym_preproc_line] = STATE(1166),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_module] = ACTIONS(5233),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_type] = ACTIONS(5235),
    [anon_sym_do] = ACTIONS(4840),
    [anon_sym_let] = ACTIONS(4842),
    [anon_sym_let_BANG] = ACTIONS(4844),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5237),
    [anon_sym_extern] = ACTIONS(5239),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1167)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1167),
    [sym_compiler_directive_decl] = STATE(1167),
    [sym_fsi_directive_decl] = STATE(1167),
    [sym_preproc_line] = STATE(1167),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19251),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1231),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(3534),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1168)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1168),
    [sym_compiler_directive_decl] = STATE(1168),
    [sym_fsi_directive_decl] = STATE(1168),
    [sym_preproc_line] = STATE(1168),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20170),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5241),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1169)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(347),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1169),
    [sym_compiler_directive_decl] = STATE(1169),
    [sym_fsi_directive_decl] = STATE(1169),
    [sym_preproc_line] = STATE(1169),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_declaration_expression_repeat1] = STATE(1293),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [anon_sym_POUNDelse] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1170)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1170),
    [sym_compiler_directive_decl] = STATE(1170),
    [sym_fsi_directive_decl] = STATE(1170),
    [sym_preproc_line] = STATE(1170),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20995),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1171),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5243),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1171)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1171),
    [sym_compiler_directive_decl] = STATE(1171),
    [sym_fsi_directive_decl] = STATE(1171),
    [sym_preproc_line] = STATE(1171),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21951),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5245),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1172)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(347),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1172),
    [sym_compiler_directive_decl] = STATE(1172),
    [sym_fsi_directive_decl] = STATE(1172),
    [sym_preproc_line] = STATE(1172),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_declaration_expression_repeat1] = STATE(1293),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5139),
    [anon_sym_POUNDelse] = ACTIONS(5139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1173)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1173),
    [sym_compiler_directive_decl] = STATE(1173),
    [sym_fsi_directive_decl] = STATE(1173),
    [sym_preproc_line] = STATE(1173),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19188),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1174),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5247),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1174)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1174),
    [sym_compiler_directive_decl] = STATE(1174),
    [sym_fsi_directive_decl] = STATE(1174),
    [sym_preproc_line] = STATE(1174),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19613),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5249),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1175)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1175),
    [sym_compiler_directive_decl] = STATE(1175),
    [sym_fsi_directive_decl] = STATE(1175),
    [sym_preproc_line] = STATE(1175),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22780),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1176),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5251),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1176)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1176),
    [sym_compiler_directive_decl] = STATE(1176),
    [sym_fsi_directive_decl] = STATE(1176),
    [sym_preproc_line] = STATE(1176),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19266),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5253),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1177)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1177),
    [sym_compiler_directive_decl] = STATE(1177),
    [sym_fsi_directive_decl] = STATE(1177),
    [sym_preproc_line] = STATE(1177),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21143),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1178),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5255),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1178)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1178),
    [sym_compiler_directive_decl] = STATE(1178),
    [sym_fsi_directive_decl] = STATE(1178),
    [sym_preproc_line] = STATE(1178),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21529),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5257),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1179)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1179),
    [sym_compiler_directive_decl] = STATE(1179),
    [sym_fsi_directive_decl] = STATE(1179),
    [sym_preproc_line] = STATE(1179),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19509),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1180),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5259),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1180)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1180),
    [sym_compiler_directive_decl] = STATE(1180),
    [sym_fsi_directive_decl] = STATE(1180),
    [sym_preproc_line] = STATE(1180),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19680),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5261),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1181)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1181),
    [sym_compiler_directive_decl] = STATE(1181),
    [sym_fsi_directive_decl] = STATE(1181),
    [sym_preproc_line] = STATE(1181),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20826),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1182),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5263),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1182)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1182),
    [sym_compiler_directive_decl] = STATE(1182),
    [sym_fsi_directive_decl] = STATE(1182),
    [sym_preproc_line] = STATE(1182),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21084),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5265),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1183)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1183),
    [sym_compiler_directive_decl] = STATE(1183),
    [sym_fsi_directive_decl] = STATE(1183),
    [sym_preproc_line] = STATE(1183),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19485),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1186),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5267),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1184)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1184),
    [sym_compiler_directive_decl] = STATE(1184),
    [sym_fsi_directive_decl] = STATE(1184),
    [sym_preproc_line] = STATE(1184),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22366),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5269),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1185)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1185),
    [sym_compiler_directive_decl] = STATE(1185),
    [sym_fsi_directive_decl] = STATE(1185),
    [sym_preproc_line] = STATE(1185),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20136),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1188),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5271),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1186)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1186),
    [sym_compiler_directive_decl] = STATE(1186),
    [sym_fsi_directive_decl] = STATE(1186),
    [sym_preproc_line] = STATE(1186),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19570),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5273),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1187)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1187),
    [sym_compiler_directive_decl] = STATE(1187),
    [sym_fsi_directive_decl] = STATE(1187),
    [sym_preproc_line] = STATE(1187),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20984),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1190),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5275),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1188)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1188),
    [sym_compiler_directive_decl] = STATE(1188),
    [sym_fsi_directive_decl] = STATE(1188),
    [sym_preproc_line] = STATE(1188),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20240),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5277),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1189)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1189),
    [sym_compiler_directive_decl] = STATE(1189),
    [sym_fsi_directive_decl] = STATE(1189),
    [sym_preproc_line] = STATE(1189),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21769),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1192),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5279),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1190)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1190),
    [sym_compiler_directive_decl] = STATE(1190),
    [sym_fsi_directive_decl] = STATE(1190),
    [sym_preproc_line] = STATE(1190),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21064),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5281),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1191)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1191),
    [sym_compiler_directive_decl] = STATE(1191),
    [sym_fsi_directive_decl] = STATE(1191),
    [sym_preproc_line] = STATE(1191),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22155),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1184),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(3230),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1192)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1192),
    [sym_compiler_directive_decl] = STATE(1192),
    [sym_fsi_directive_decl] = STATE(1192),
    [sym_preproc_line] = STATE(1192),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21909),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5283),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1193)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1193),
    [sym_compiler_directive_decl] = STATE(1193),
    [sym_fsi_directive_decl] = STATE(1193),
    [sym_preproc_line] = STATE(1193),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22181),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1194),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5285),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1194)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1194),
    [sym_compiler_directive_decl] = STATE(1194),
    [sym_fsi_directive_decl] = STATE(1194),
    [sym_preproc_line] = STATE(1194),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22261),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5287),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1195)] = {
    [sym_do] = STATE(8680),
    [sym_function_or_value_defn] = STATE(1116),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(166),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1195),
    [sym_compiler_directive_decl] = STATE(1195),
    [sym_fsi_directive_decl] = STATE(1195),
    [sym_preproc_line] = STATE(1195),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(5213),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_type] = ACTIONS(5289),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5217),
    [anon_sym_extern] = ACTIONS(5219),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1196)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1196),
    [sym_compiler_directive_decl] = STATE(1196),
    [sym_fsi_directive_decl] = STATE(1196),
    [sym_preproc_line] = STATE(1196),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19117),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1197),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(3394),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1197)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1197),
    [sym_compiler_directive_decl] = STATE(1197),
    [sym_fsi_directive_decl] = STATE(1197),
    [sym_preproc_line] = STATE(1197),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19152),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5291),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1198)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1198),
    [sym_compiler_directive_decl] = STATE(1198),
    [sym_fsi_directive_decl] = STATE(1198),
    [sym_preproc_line] = STATE(1198),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19335),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1199),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5293),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1199)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1199),
    [sym_compiler_directive_decl] = STATE(1199),
    [sym_fsi_directive_decl] = STATE(1199),
    [sym_preproc_line] = STATE(1199),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19377),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5295),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1200)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1200),
    [sym_compiler_directive_decl] = STATE(1200),
    [sym_fsi_directive_decl] = STATE(1200),
    [sym_preproc_line] = STATE(1200),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19553),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1201),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(3384),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1201)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1201),
    [sym_compiler_directive_decl] = STATE(1201),
    [sym_fsi_directive_decl] = STATE(1201),
    [sym_preproc_line] = STATE(1201),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19589),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5297),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1202)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1202),
    [sym_compiler_directive_decl] = STATE(1202),
    [sym_fsi_directive_decl] = STATE(1202),
    [sym_preproc_line] = STATE(1202),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20038),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1205),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5299),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1203)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1203),
    [sym_compiler_directive_decl] = STATE(1203),
    [sym_fsi_directive_decl] = STATE(1203),
    [sym_preproc_line] = STATE(1203),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19818),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5301),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1204)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1204),
    [sym_compiler_directive_decl] = STATE(1204),
    [sym_fsi_directive_decl] = STATE(1204),
    [sym_preproc_line] = STATE(1204),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(19787),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1203),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(3400),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1205)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1205),
    [sym_compiler_directive_decl] = STATE(1205),
    [sym_fsi_directive_decl] = STATE(1205),
    [sym_preproc_line] = STATE(1205),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20072),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5303),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1206)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1206),
    [sym_compiler_directive_decl] = STATE(1206),
    [sym_fsi_directive_decl] = STATE(1206),
    [sym_preproc_line] = STATE(1206),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20303),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1207),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5305),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1207)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1207),
    [sym_compiler_directive_decl] = STATE(1207),
    [sym_fsi_directive_decl] = STATE(1207),
    [sym_preproc_line] = STATE(1207),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20336),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5307),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1208)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(446),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(22016),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1208),
    [sym_compiler_directive_decl] = STATE(1208),
    [sym_fsi_directive_decl] = STATE(1208),
    [sym_preproc_line] = STATE(1208),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1209)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1209),
    [sym_compiler_directive_decl] = STATE(1209),
    [sym_fsi_directive_decl] = STATE(1209),
    [sym_preproc_line] = STATE(1209),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20632),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5309),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1210)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1210),
    [sym_compiler_directive_decl] = STATE(1210),
    [sym_fsi_directive_decl] = STATE(1210),
    [sym_preproc_line] = STATE(1210),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20894),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1211),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5311),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1211)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1211),
    [sym_compiler_directive_decl] = STATE(1211),
    [sym_fsi_directive_decl] = STATE(1211),
    [sym_preproc_line] = STATE(1211),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20944),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5313),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1212)] = {
    [sym_do] = STATE(8680),
    [sym_function_or_value_defn] = STATE(1119),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(211),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1212),
    [sym_compiler_directive_decl] = STATE(1212),
    [sym_fsi_directive_decl] = STATE(1212),
    [sym_preproc_line] = STATE(1212),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_module] = ACTIONS(5315),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_type] = ACTIONS(5215),
    [anon_sym_do] = ACTIONS(37),
    [anon_sym_let] = ACTIONS(39),
    [anon_sym_let_BANG] = ACTIONS(41),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5217),
    [anon_sym_extern] = ACTIONS(5219),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1213)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1213),
    [sym_compiler_directive_decl] = STATE(1213),
    [sym_fsi_directive_decl] = STATE(1213),
    [sym_preproc_line] = STATE(1213),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21168),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1214),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5317),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1214)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1214),
    [sym_compiler_directive_decl] = STATE(1214),
    [sym_fsi_directive_decl] = STATE(1214),
    [sym_preproc_line] = STATE(1214),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21200),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5319),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1215)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1215),
    [sym_compiler_directive_decl] = STATE(1215),
    [sym_fsi_directive_decl] = STATE(1215),
    [sym_preproc_line] = STATE(1215),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21453),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1216),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5321),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1216)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1216),
    [sym_compiler_directive_decl] = STATE(1216),
    [sym_fsi_directive_decl] = STATE(1216),
    [sym_preproc_line] = STATE(1216),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21508),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5323),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1217)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1217),
    [sym_compiler_directive_decl] = STATE(1217),
    [sym_fsi_directive_decl] = STATE(1217),
    [sym_preproc_line] = STATE(1217),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21777),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1218),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5325),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1218)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1218),
    [sym_compiler_directive_decl] = STATE(1218),
    [sym_fsi_directive_decl] = STATE(1218),
    [sym_preproc_line] = STATE(1218),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21880),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5327),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1219)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1219),
    [sym_compiler_directive_decl] = STATE(1219),
    [sym_fsi_directive_decl] = STATE(1219),
    [sym_preproc_line] = STATE(1219),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22046),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1220),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5329),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1220)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1220),
    [sym_compiler_directive_decl] = STATE(1220),
    [sym_fsi_directive_decl] = STATE(1220),
    [sym_preproc_line] = STATE(1220),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(22073),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5331),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1221)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(222),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1221),
    [sym_compiler_directive_decl] = STATE(1221),
    [sym_fsi_directive_decl] = STATE(1221),
    [sym_preproc_line] = STATE(1221),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(1337),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__newline] = ACTIONS(5121),
    [sym__dedent] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1222)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1222),
    [sym_compiler_directive_decl] = STATE(1222),
    [sym_fsi_directive_decl] = STATE(1222),
    [sym_preproc_line] = STATE(1222),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21685),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1165),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5333),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1223)] = {
    [sym_do] = STATE(8665),
    [sym_function_or_value_defn] = STATE(1115),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(164),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1223),
    [sym_compiler_directive_decl] = STATE(1223),
    [sym_fsi_directive_decl] = STATE(1223),
    [sym_preproc_line] = STATE(1223),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_module] = ACTIONS(5335),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_type] = ACTIONS(5337),
    [anon_sym_do] = ACTIONS(4483),
    [anon_sym_let] = ACTIONS(4485),
    [anon_sym_let_BANG] = ACTIONS(4487),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5339),
    [anon_sym_extern] = ACTIONS(5341),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1224)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(315),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(21568),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1224),
    [sym_compiler_directive_decl] = STATE(1224),
    [sym_fsi_directive_decl] = STATE(1224),
    [sym_preproc_line] = STATE(1224),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1225)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(381),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19882),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1225),
    [sym_compiler_directive_decl] = STATE(1225),
    [sym_fsi_directive_decl] = STATE(1225),
    [sym_preproc_line] = STATE(1225),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1226)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(397),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20019),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1226),
    [sym_compiler_directive_decl] = STATE(1226),
    [sym_fsi_directive_decl] = STATE(1226),
    [sym_preproc_line] = STATE(1226),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1227)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(400),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(22281),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1227),
    [sym_compiler_directive_decl] = STATE(1227),
    [sym_fsi_directive_decl] = STATE(1227),
    [sym_preproc_line] = STATE(1227),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1228)] = {
    [sym_do] = STATE(18190),
    [sym_function_or_value_defn] = STATE(1172),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(378),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1228),
    [sym_compiler_directive_decl] = STATE(1228),
    [sym_fsi_directive_decl] = STATE(1228),
    [sym_preproc_line] = STATE(1228),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_module] = ACTIONS(5343),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_type] = ACTIONS(5345),
    [anon_sym_do] = ACTIONS(3208),
    [anon_sym_let] = ACTIONS(3210),
    [anon_sym_let_BANG] = ACTIONS(3212),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_exception] = ACTIONS(5347),
    [anon_sym_extern] = ACTIONS(5349),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1229)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(403),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(22161),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1229),
    [sym_compiler_directive_decl] = STATE(1229),
    [sym_fsi_directive_decl] = STATE(1229),
    [sym_preproc_line] = STATE(1229),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1230)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(405),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20881),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1230),
    [sym_compiler_directive_decl] = STATE(1230),
    [sym_fsi_directive_decl] = STATE(1230),
    [sym_preproc_line] = STATE(1230),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1231)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1231),
    [sym_compiler_directive_decl] = STATE(1231),
    [sym_fsi_directive_decl] = STATE(1231),
    [sym_preproc_line] = STATE(1231),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(21086),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5351),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1232)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(407),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19399),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1232),
    [sym_compiler_directive_decl] = STATE(1232),
    [sym_fsi_directive_decl] = STATE(1232),
    [sym_preproc_line] = STATE(1232),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1233)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(409),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20698),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1233),
    [sym_compiler_directive_decl] = STATE(1233),
    [sym_fsi_directive_decl] = STATE(1233),
    [sym_preproc_line] = STATE(1233),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1234)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(411),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(22091),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1234),
    [sym_compiler_directive_decl] = STATE(1234),
    [sym_fsi_directive_decl] = STATE(1234),
    [sym_preproc_line] = STATE(1234),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1235)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(413),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19428),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1235),
    [sym_compiler_directive_decl] = STATE(1235),
    [sym_fsi_directive_decl] = STATE(1235),
    [sym_preproc_line] = STATE(1235),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1236)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(415),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20051),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1236),
    [sym_compiler_directive_decl] = STATE(1236),
    [sym_fsi_directive_decl] = STATE(1236),
    [sym_preproc_line] = STATE(1236),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1237)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(417),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20889),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1237),
    [sym_compiler_directive_decl] = STATE(1237),
    [sym_fsi_directive_decl] = STATE(1237),
    [sym_preproc_line] = STATE(1237),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1238)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(419),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(21657),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1238),
    [sym_compiler_directive_decl] = STATE(1238),
    [sym_fsi_directive_decl] = STATE(1238),
    [sym_preproc_line] = STATE(1238),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1239)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(421),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(22138),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1239),
    [sym_compiler_directive_decl] = STATE(1239),
    [sym_fsi_directive_decl] = STATE(1239),
    [sym_preproc_line] = STATE(1239),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1240)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(424),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19085),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1240),
    [sym_compiler_directive_decl] = STATE(1240),
    [sym_fsi_directive_decl] = STATE(1240),
    [sym_preproc_line] = STATE(1240),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1241)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(426),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19308),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1241),
    [sym_compiler_directive_decl] = STATE(1241),
    [sym_fsi_directive_decl] = STATE(1241),
    [sym_preproc_line] = STATE(1241),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1242)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(428),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19520),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1242),
    [sym_compiler_directive_decl] = STATE(1242),
    [sym_fsi_directive_decl] = STATE(1242),
    [sym_preproc_line] = STATE(1242),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1243)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(430),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(19747),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1243),
    [sym_compiler_directive_decl] = STATE(1243),
    [sym_fsi_directive_decl] = STATE(1243),
    [sym_preproc_line] = STATE(1243),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1244)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(434),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20252),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1244),
    [sym_compiler_directive_decl] = STATE(1244),
    [sym_fsi_directive_decl] = STATE(1244),
    [sym_preproc_line] = STATE(1244),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1245)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(436),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20530),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1245),
    [sym_compiler_directive_decl] = STATE(1245),
    [sym_fsi_directive_decl] = STATE(1245),
    [sym_preproc_line] = STATE(1245),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1246)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(438),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(20846),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1246),
    [sym_compiler_directive_decl] = STATE(1246),
    [sym_fsi_directive_decl] = STATE(1246),
    [sym_preproc_line] = STATE(1246),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1247)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(440),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(21116),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1247),
    [sym_compiler_directive_decl] = STATE(1247),
    [sym_fsi_directive_decl] = STATE(1247),
    [sym_preproc_line] = STATE(1247),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1248)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(442),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(21413),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1248),
    [sym_compiler_directive_decl] = STATE(1248),
    [sym_fsi_directive_decl] = STATE(1248),
    [sym_preproc_line] = STATE(1248),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1249)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(444),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_slice_ranges] = STATE(21718),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(17119),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1249),
    [sym_compiler_directive_decl] = STATE(1249),
    [sym_fsi_directive_decl] = STATE(1249),
    [sym_preproc_line] = STATE(1249),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1250)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1250),
    [sym_compiler_directive_decl] = STATE(1250),
    [sym_fsi_directive_decl] = STATE(1250),
    [sym_preproc_line] = STATE(1250),
    [sym_preproc_if_in_expression] = STATE(5827),
    [sym_preproc_else_in_expression] = STATE(20583),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1209),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [anon_sym_POUNDendif] = ACTIONS(5353),
    [anon_sym_POUNDelse] = ACTIONS(3232),
    [sym__newline] = ACTIONS(3234),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1251)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20018),
    [sym_with_field_expression] = STATE(20018),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20018),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1251),
    [sym_compiler_directive_decl] = STATE(1251),
    [sym_fsi_directive_decl] = STATE(1251),
    [sym_preproc_line] = STATE(1251),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1252)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(348),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1252),
    [sym_compiler_directive_decl] = STATE(1252),
    [sym_fsi_directive_decl] = STATE(1252),
    [sym_preproc_line] = STATE(1252),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1252),
    [sym_identifier] = ACTIONS(5367),
    [anon_sym_null] = ACTIONS(5370),
    [anon_sym_begin] = ACTIONS(5373),
    [anon_sym_do] = ACTIONS(5376),
    [anon_sym_let] = ACTIONS(5379),
    [anon_sym_let_BANG] = ACTIONS(5382),
    [anon_sym_LPAREN] = ACTIONS(5385),
    [anon_sym_AMP] = ACTIONS(5388),
    [anon_sym_LBRACK] = ACTIONS(5391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5394),
    [anon_sym_LBRACE] = ACTIONS(5397),
    [anon_sym_LT_AT] = ACTIONS(5400),
    [anon_sym_LT_AT_AT] = ACTIONS(5403),
    [anon_sym_CARET] = ACTIONS(5406),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5409),
    [anon_sym_new] = ACTIONS(5412),
    [anon_sym_return] = ACTIONS(5412),
    [anon_sym_return_BANG] = ACTIONS(5415),
    [anon_sym_yield] = ACTIONS(5412),
    [anon_sym_yield_BANG] = ACTIONS(5415),
    [anon_sym_lazy] = ACTIONS(5412),
    [anon_sym_assert] = ACTIONS(5412),
    [anon_sym_upcast] = ACTIONS(5412),
    [anon_sym_downcast] = ACTIONS(5412),
    [anon_sym_fixed] = ACTIONS(5412),
    [anon_sym_for] = ACTIONS(5418),
    [anon_sym_while] = ACTIONS(5421),
    [anon_sym_while_BANG] = ACTIONS(5424),
    [anon_sym_if] = ACTIONS(5427),
    [anon_sym_fun] = ACTIONS(5430),
    [anon_sym_try] = ACTIONS(5433),
    [anon_sym_match] = ACTIONS(5436),
    [anon_sym_match_BANG] = ACTIONS(5439),
    [anon_sym_function] = ACTIONS(5442),
    [anon_sym_use] = ACTIONS(5445),
    [anon_sym_use_BANG] = ACTIONS(5448),
    [anon_sym_do_BANG] = ACTIONS(5451),
    [anon_sym_AMP_AMP] = ACTIONS(5454),
    [aux_sym_char_token1] = ACTIONS(5457),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5460),
    [anon_sym_DQUOTE] = ACTIONS(5463),
    [anon_sym_AT_DQUOTE] = ACTIONS(5466),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5469),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5472),
    [sym_bool] = ACTIONS(5475),
    [sym_unit] = ACTIONS(5478),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5481),
    [sym_op_identifier] = ACTIONS(5484),
    [anon_sym_PLUS] = ACTIONS(5388),
    [anon_sym_DASH] = ACTIONS(5388),
    [anon_sym_PLUS_DOT] = ACTIONS(5487),
    [anon_sym_DASH_DOT] = ACTIONS(5487),
    [anon_sym_PERCENT] = ACTIONS(5388),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5454),
    [anon_sym_TILDE] = ACTIONS(5490),
    [aux_sym_prefix_op_token1] = ACTIONS(5454),
    [sym_int] = ACTIONS(5493),
    [sym_xint] = ACTIONS(5496),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5499),
    [anon_sym_POUNDendif] = ACTIONS(2107),
    [anon_sym_POUNDelse] = ACTIONS(2107),
    [sym__newline] = ACTIONS(5502),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5505),
  },
  [STATE(1253)] = {
    [sym_function_or_value_defn] = STATE(1156),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(372),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym__comp_or_range_expression] = STATE(18469),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1253),
    [sym_compiler_directive_decl] = STATE(1253),
    [sym_fsi_directive_decl] = STATE(1253),
    [sym_preproc_line] = STATE(1253),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(5195),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(5197),
    [anon_sym_use_BANG] = ACTIONS(5199),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1254)] = {
    [sym_function_or_value_defn] = STATE(1156),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(339),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym__comp_or_range_expression] = STATE(18071),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1254),
    [sym_compiler_directive_decl] = STATE(1254),
    [sym_fsi_directive_decl] = STATE(1254),
    [sym_preproc_line] = STATE(1254),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(5195),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(5197),
    [anon_sym_use_BANG] = ACTIONS(5199),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1255)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(568),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1255),
    [sym_compiler_directive_decl] = STATE(1255),
    [sym_fsi_directive_decl] = STATE(1255),
    [sym_preproc_line] = STATE(1255),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_declaration_expression_repeat1] = STATE(1309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(5121),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1256)] = {
    [sym_function_or_value_defn] = STATE(1151),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(267),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_range_expression] = STATE(21253),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym__comp_or_range_expression] = STATE(21388),
    [sym_comp_declaration_expression] = STATE(21253),
    [sym_short_comp_expression] = STATE(21253),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1256),
    [sym_compiler_directive_decl] = STATE(1256),
    [sym_fsi_directive_decl] = STATE(1256),
    [sym_preproc_line] = STATE(1256),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(5125),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(5127),
    [anon_sym_use_BANG] = ACTIONS(5129),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1257)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(371),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym__slice_range_special] = STATE(18815),
    [sym_slice_range] = STATE(18967),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1257),
    [sym_compiler_directive_decl] = STATE(1257),
    [sym_fsi_directive_decl] = STATE(1257),
    [sym_preproc_line] = STATE(1257),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5209),
    [anon_sym_STAR] = ACTIONS(5211),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1258)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(345),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym__slice_range_special] = STATE(18369),
    [sym_slice_range] = STATE(17797),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1258),
    [sym_compiler_directive_decl] = STATE(1258),
    [sym_fsi_directive_decl] = STATE(1258),
    [sym_preproc_line] = STATE(1258),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [anon_sym_DOT_DOT3] = ACTIONS(5131),
    [anon_sym_STAR] = ACTIONS(5133),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1259)] = {
    [sym_function_or_value_defn] = STATE(1153),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(274),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_range_expression] = STATE(18812),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym__comp_or_range_expression] = STATE(18071),
    [sym_comp_declaration_expression] = STATE(18812),
    [sym_short_comp_expression] = STATE(18812),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1259),
    [sym_compiler_directive_decl] = STATE(1259),
    [sym_fsi_directive_decl] = STATE(1259),
    [sym_preproc_line] = STATE(1259),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(5149),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(5151),
    [anon_sym_use_BANG] = ACTIONS(5153),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1260)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(22111),
    [sym_with_field_expression] = STATE(22111),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22111),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1260),
    [sym_compiler_directive_decl] = STATE(1260),
    [sym_fsi_directive_decl] = STATE(1260),
    [sym_preproc_line] = STATE(1260),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1261)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(568),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1261),
    [sym_compiler_directive_decl] = STATE(1261),
    [sym_fsi_directive_decl] = STATE(1261),
    [sym_preproc_line] = STATE(1261),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_declaration_expression_repeat1] = STATE(1309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(5139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1262)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(21902),
    [sym_with_field_expression] = STATE(21902),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(21902),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1262),
    [sym_compiler_directive_decl] = STATE(1262),
    [sym_fsi_directive_decl] = STATE(1262),
    [sym_preproc_line] = STATE(1262),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1263)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(22110),
    [sym_with_field_expression] = STATE(22110),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22110),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1263),
    [sym_compiler_directive_decl] = STATE(1263),
    [sym_fsi_directive_decl] = STATE(1263),
    [sym_preproc_line] = STATE(1263),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1264)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(22319),
    [sym_with_field_expression] = STATE(22319),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22319),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1264),
    [sym_compiler_directive_decl] = STATE(1264),
    [sym_fsi_directive_decl] = STATE(1264),
    [sym_preproc_line] = STATE(1264),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1265)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(22783),
    [sym_with_field_expression] = STATE(22783),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22783),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1265),
    [sym_compiler_directive_decl] = STATE(1265),
    [sym_fsi_directive_decl] = STATE(1265),
    [sym_preproc_line] = STATE(1265),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1266)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19174),
    [sym_with_field_expression] = STATE(19174),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19174),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1266),
    [sym_compiler_directive_decl] = STATE(1266),
    [sym_fsi_directive_decl] = STATE(1266),
    [sym_preproc_line] = STATE(1266),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1267)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20457),
    [sym_with_field_expression] = STATE(20457),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20457),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1267),
    [sym_compiler_directive_decl] = STATE(1267),
    [sym_fsi_directive_decl] = STATE(1267),
    [sym_preproc_line] = STATE(1267),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1268)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19402),
    [sym_with_field_expression] = STATE(19402),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19402),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1268),
    [sym_compiler_directive_decl] = STATE(1268),
    [sym_fsi_directive_decl] = STATE(1268),
    [sym_preproc_line] = STATE(1268),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1269)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19514),
    [sym_with_field_expression] = STATE(19514),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19514),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1269),
    [sym_compiler_directive_decl] = STATE(1269),
    [sym_fsi_directive_decl] = STATE(1269),
    [sym_preproc_line] = STATE(1269),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1270)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19591),
    [sym_with_field_expression] = STATE(19591),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19591),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1270),
    [sym_compiler_directive_decl] = STATE(1270),
    [sym_fsi_directive_decl] = STATE(1270),
    [sym_preproc_line] = STATE(1270),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1271)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19667),
    [sym_with_field_expression] = STATE(19667),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19667),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1271),
    [sym_compiler_directive_decl] = STATE(1271),
    [sym_fsi_directive_decl] = STATE(1271),
    [sym_preproc_line] = STATE(1271),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1272)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19739),
    [sym_with_field_expression] = STATE(19739),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19739),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1272),
    [sym_compiler_directive_decl] = STATE(1272),
    [sym_fsi_directive_decl] = STATE(1272),
    [sym_preproc_line] = STATE(1272),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1273)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19807),
    [sym_with_field_expression] = STATE(19807),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19807),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1273),
    [sym_compiler_directive_decl] = STATE(1273),
    [sym_fsi_directive_decl] = STATE(1273),
    [sym_preproc_line] = STATE(1273),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1274)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19870),
    [sym_with_field_expression] = STATE(19870),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19870),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1274),
    [sym_compiler_directive_decl] = STATE(1274),
    [sym_fsi_directive_decl] = STATE(1274),
    [sym_preproc_line] = STATE(1274),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1275)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19925),
    [sym_with_field_expression] = STATE(19925),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19925),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1275),
    [sym_compiler_directive_decl] = STATE(1275),
    [sym_fsi_directive_decl] = STATE(1275),
    [sym_preproc_line] = STATE(1275),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1276)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19974),
    [sym_with_field_expression] = STATE(19974),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19974),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1276),
    [sym_compiler_directive_decl] = STATE(1276),
    [sym_fsi_directive_decl] = STATE(1276),
    [sym_preproc_line] = STATE(1276),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1277)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20061),
    [sym_with_field_expression] = STATE(20061),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20061),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1277),
    [sym_compiler_directive_decl] = STATE(1277),
    [sym_fsi_directive_decl] = STATE(1277),
    [sym_preproc_line] = STATE(1277),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1278)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20105),
    [sym_with_field_expression] = STATE(20105),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20105),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1278),
    [sym_compiler_directive_decl] = STATE(1278),
    [sym_fsi_directive_decl] = STATE(1278),
    [sym_preproc_line] = STATE(1278),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1279)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20149),
    [sym_with_field_expression] = STATE(20149),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20149),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1279),
    [sym_compiler_directive_decl] = STATE(1279),
    [sym_fsi_directive_decl] = STATE(1279),
    [sym_preproc_line] = STATE(1279),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1280)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20193),
    [sym_with_field_expression] = STATE(20193),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20193),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1280),
    [sym_compiler_directive_decl] = STATE(1280),
    [sym_fsi_directive_decl] = STATE(1280),
    [sym_preproc_line] = STATE(1280),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1281)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20237),
    [sym_with_field_expression] = STATE(20237),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20237),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1281),
    [sym_compiler_directive_decl] = STATE(1281),
    [sym_fsi_directive_decl] = STATE(1281),
    [sym_preproc_line] = STATE(1281),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1282)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20281),
    [sym_with_field_expression] = STATE(20281),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20281),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1282),
    [sym_compiler_directive_decl] = STATE(1282),
    [sym_fsi_directive_decl] = STATE(1282),
    [sym_preproc_line] = STATE(1282),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1283)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20325),
    [sym_with_field_expression] = STATE(20325),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20325),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1283),
    [sym_compiler_directive_decl] = STATE(1283),
    [sym_fsi_directive_decl] = STATE(1283),
    [sym_preproc_line] = STATE(1283),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1284)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20369),
    [sym_with_field_expression] = STATE(20369),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20369),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1284),
    [sym_compiler_directive_decl] = STATE(1284),
    [sym_fsi_directive_decl] = STATE(1284),
    [sym_preproc_line] = STATE(1284),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1285)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(20413),
    [sym_with_field_expression] = STATE(20413),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20413),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1285),
    [sym_compiler_directive_decl] = STATE(1285),
    [sym_fsi_directive_decl] = STATE(1285),
    [sym_preproc_line] = STATE(1285),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1286)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_object_expression] = STATE(19288),
    [sym_with_field_expression] = STATE(19288),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19288),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1286),
    [sym_compiler_directive_decl] = STATE(1286),
    [sym_fsi_directive_decl] = STATE(1286),
    [sym_preproc_line] = STATE(1286),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(5361),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1287)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(531),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1287),
    [sym_compiler_directive_decl] = STATE(1287),
    [sym_fsi_directive_decl] = STATE(1287),
    [sym_preproc_line] = STATE(1287),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1326),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(5508),
    [sym__newline] = ACTIONS(5510),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1288)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(51),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1288),
    [sym_compiler_directive_decl] = STATE(1288),
    [sym_fsi_directive_decl] = STATE(1288),
    [sym_preproc_line] = STATE(1288),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1289)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(358),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1289),
    [sym_compiler_directive_decl] = STATE(1289),
    [sym_fsi_directive_decl] = STATE(1289),
    [sym_preproc_line] = STATE(1289),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_declaration_expression_repeat1] = STATE(1327),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1290)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(510),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1290),
    [sym_compiler_directive_decl] = STATE(1290),
    [sym_fsi_directive_decl] = STATE(1290),
    [sym_preproc_line] = STATE(1290),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_declaration_expression_repeat1] = STATE(1323),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1291)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(158),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1291),
    [sym_compiler_directive_decl] = STATE(1291),
    [sym_fsi_directive_decl] = STATE(1291),
    [sym_preproc_line] = STATE(1291),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1292)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(162),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1292),
    [sym_compiler_directive_decl] = STATE(1292),
    [sym_fsi_directive_decl] = STATE(1292),
    [sym_preproc_line] = STATE(1292),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_declaration_expression_repeat1] = STATE(1330),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1293)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(349),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1293),
    [sym_compiler_directive_decl] = STATE(1293),
    [sym_fsi_directive_decl] = STATE(1293),
    [sym_preproc_line] = STATE(1293),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1294)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(290),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1294),
    [sym_compiler_directive_decl] = STATE(1294),
    [sym_fsi_directive_decl] = STATE(1294),
    [sym_preproc_line] = STATE(1294),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1295)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(157),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1295),
    [sym_compiler_directive_decl] = STATE(1295),
    [sym_fsi_directive_decl] = STATE(1295),
    [sym_preproc_line] = STATE(1295),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_declaration_expression_repeat1] = STATE(1291),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1296)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(678),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1296),
    [sym_compiler_directive_decl] = STATE(1296),
    [sym_fsi_directive_decl] = STATE(1296),
    [sym_preproc_line] = STATE(1296),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_declaration_expression_repeat1] = STATE(1358),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1297)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(46),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1297),
    [sym_compiler_directive_decl] = STATE(1297),
    [sym_fsi_directive_decl] = STATE(1297),
    [sym_preproc_line] = STATE(1297),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_declaration_expression_repeat1] = STATE(1338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1298)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(499),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1298),
    [sym_compiler_directive_decl] = STATE(1298),
    [sym_fsi_directive_decl] = STATE(1298),
    [sym_preproc_line] = STATE(1298),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1299)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(551),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1299),
    [sym_compiler_directive_decl] = STATE(1299),
    [sym_fsi_directive_decl] = STATE(1299),
    [sym_preproc_line] = STATE(1299),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_declaration_expression_repeat1] = STATE(1320),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1300)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(213),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1300),
    [sym_compiler_directive_decl] = STATE(1300),
    [sym_fsi_directive_decl] = STATE(1300),
    [sym_preproc_line] = STATE(1300),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_declaration_expression_repeat1] = STATE(1334),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1301)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(21893),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(21893),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1301),
    [sym_compiler_directive_decl] = STATE(1301),
    [sym_fsi_directive_decl] = STATE(1301),
    [sym_preproc_line] = STATE(1301),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1302)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(647),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1302),
    [sym_compiler_directive_decl] = STATE(1302),
    [sym_fsi_directive_decl] = STATE(1302),
    [sym_preproc_line] = STATE(1302),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1303)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(558),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1303),
    [sym_compiler_directive_decl] = STATE(1303),
    [sym_fsi_directive_decl] = STATE(1303),
    [sym_preproc_line] = STATE(1303),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1304)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(22),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1304),
    [sym_compiler_directive_decl] = STATE(1304),
    [sym_fsi_directive_decl] = STATE(1304),
    [sym_preproc_line] = STATE(1304),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_declaration_expression_repeat1] = STATE(1348),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1305)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(807),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1305),
    [sym_compiler_directive_decl] = STATE(1305),
    [sym_fsi_directive_decl] = STATE(1305),
    [sym_preproc_line] = STATE(1305),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(1336),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1306)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(93),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1306),
    [sym_compiler_directive_decl] = STATE(1306),
    [sym_fsi_directive_decl] = STATE(1306),
    [sym_preproc_line] = STATE(1306),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_declaration_expression_repeat1] = STATE(1307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1307)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(95),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1307),
    [sym_compiler_directive_decl] = STATE(1307),
    [sym_fsi_directive_decl] = STATE(1307),
    [sym_preproc_line] = STATE(1307),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1308)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(88),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1308),
    [sym_compiler_directive_decl] = STATE(1308),
    [sym_fsi_directive_decl] = STATE(1308),
    [sym_preproc_line] = STATE(1308),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_declaration_expression_repeat1] = STATE(1341),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1309)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(570),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1309),
    [sym_compiler_directive_decl] = STATE(1309),
    [sym_fsi_directive_decl] = STATE(1309),
    [sym_preproc_line] = STATE(1309),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1310)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(99),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1310),
    [sym_compiler_directive_decl] = STATE(1310),
    [sym_fsi_directive_decl] = STATE(1310),
    [sym_preproc_line] = STATE(1310),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_declaration_expression_repeat1] = STATE(1350),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1311)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(262),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1311),
    [sym_compiler_directive_decl] = STATE(1311),
    [sym_fsi_directive_decl] = STATE(1311),
    [sym_preproc_line] = STATE(1311),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1312)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(123),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1312),
    [sym_compiler_directive_decl] = STATE(1312),
    [sym_fsi_directive_decl] = STATE(1312),
    [sym_preproc_line] = STATE(1312),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_declaration_expression_repeat1] = STATE(1360),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1313)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(12),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1313),
    [sym_compiler_directive_decl] = STATE(1313),
    [sym_fsi_directive_decl] = STATE(1313),
    [sym_preproc_line] = STATE(1313),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_declaration_expression_repeat1] = STATE(1371),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1314)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(104),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1314),
    [sym_compiler_directive_decl] = STATE(1314),
    [sym_fsi_directive_decl] = STATE(1314),
    [sym_preproc_line] = STATE(1314),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_declaration_expression_repeat1] = STATE(1379),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1315)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(50),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1315),
    [sym_compiler_directive_decl] = STATE(1315),
    [sym_fsi_directive_decl] = STATE(1315),
    [sym_preproc_line] = STATE(1315),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_declaration_expression_repeat1] = STATE(1288),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1316)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(531),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1316),
    [sym_compiler_directive_decl] = STATE(1316),
    [sym_fsi_directive_decl] = STATE(1316),
    [sym_preproc_line] = STATE(1316),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1316),
    [sym_identifier] = ACTIONS(5616),
    [anon_sym_null] = ACTIONS(5619),
    [anon_sym_begin] = ACTIONS(5622),
    [anon_sym_do] = ACTIONS(5625),
    [anon_sym_let] = ACTIONS(5379),
    [anon_sym_let_BANG] = ACTIONS(5382),
    [anon_sym_LPAREN] = ACTIONS(5628),
    [anon_sym_AMP] = ACTIONS(5388),
    [anon_sym_LBRACK] = ACTIONS(5631),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5634),
    [anon_sym_LBRACE] = ACTIONS(5637),
    [anon_sym_LT_AT] = ACTIONS(5640),
    [anon_sym_LT_AT_AT] = ACTIONS(5643),
    [anon_sym_CARET] = ACTIONS(5406),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5646),
    [anon_sym_new] = ACTIONS(5649),
    [anon_sym_return] = ACTIONS(5649),
    [anon_sym_return_BANG] = ACTIONS(5652),
    [anon_sym_yield] = ACTIONS(5649),
    [anon_sym_yield_BANG] = ACTIONS(5652),
    [anon_sym_lazy] = ACTIONS(5649),
    [anon_sym_assert] = ACTIONS(5649),
    [anon_sym_upcast] = ACTIONS(5649),
    [anon_sym_downcast] = ACTIONS(5649),
    [anon_sym_fixed] = ACTIONS(5649),
    [anon_sym_for] = ACTIONS(5655),
    [anon_sym_while] = ACTIONS(5658),
    [anon_sym_while_BANG] = ACTIONS(5661),
    [anon_sym_if] = ACTIONS(5427),
    [anon_sym_fun] = ACTIONS(5664),
    [anon_sym_try] = ACTIONS(5667),
    [anon_sym_match] = ACTIONS(5670),
    [anon_sym_match_BANG] = ACTIONS(5673),
    [anon_sym_function] = ACTIONS(5676),
    [anon_sym_use] = ACTIONS(5679),
    [anon_sym_use_BANG] = ACTIONS(5682),
    [anon_sym_do_BANG] = ACTIONS(5685),
    [anon_sym_AMP_AMP] = ACTIONS(5454),
    [aux_sym_char_token1] = ACTIONS(5688),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5691),
    [anon_sym_DQUOTE] = ACTIONS(5694),
    [anon_sym_AT_DQUOTE] = ACTIONS(5697),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5700),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5703),
    [sym_bool] = ACTIONS(5706),
    [sym_unit] = ACTIONS(5709),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5712),
    [sym_op_identifier] = ACTIONS(5715),
    [anon_sym_PLUS] = ACTIONS(5388),
    [anon_sym_DASH] = ACTIONS(5388),
    [anon_sym_PLUS_DOT] = ACTIONS(5487),
    [anon_sym_DASH_DOT] = ACTIONS(5487),
    [anon_sym_PERCENT] = ACTIONS(5388),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5454),
    [anon_sym_TILDE] = ACTIONS(5490),
    [aux_sym_prefix_op_token1] = ACTIONS(5454),
    [sym_int] = ACTIONS(5718),
    [sym_xint] = ACTIONS(5721),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5724),
    [anon_sym_POUNDendif] = ACTIONS(2107),
    [sym__newline] = ACTIONS(5727),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5730),
  },
  [STATE(1317)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(323),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1317),
    [sym_compiler_directive_decl] = STATE(1317),
    [sym_fsi_directive_decl] = STATE(1317),
    [sym_preproc_line] = STATE(1317),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(1329),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1318)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(180),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1318),
    [sym_compiler_directive_decl] = STATE(1318),
    [sym_fsi_directive_decl] = STATE(1318),
    [sym_preproc_line] = STATE(1318),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1319)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(179),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1319),
    [sym_compiler_directive_decl] = STATE(1319),
    [sym_fsi_directive_decl] = STATE(1319),
    [sym_preproc_line] = STATE(1319),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(1318),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1320)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(554),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1320),
    [sym_compiler_directive_decl] = STATE(1320),
    [sym_fsi_directive_decl] = STATE(1320),
    [sym_preproc_line] = STATE(1320),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1321)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(241),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1321),
    [sym_compiler_directive_decl] = STATE(1321),
    [sym_fsi_directive_decl] = STATE(1321),
    [sym_preproc_line] = STATE(1321),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1322)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(266),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1322),
    [sym_compiler_directive_decl] = STATE(1322),
    [sym_fsi_directive_decl] = STATE(1322),
    [sym_preproc_line] = STATE(1322),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(1333),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1323)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(512),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1323),
    [sym_compiler_directive_decl] = STATE(1323),
    [sym_fsi_directive_decl] = STATE(1323),
    [sym_preproc_line] = STATE(1323),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1324)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(201),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1324),
    [sym_compiler_directive_decl] = STATE(1324),
    [sym_fsi_directive_decl] = STATE(1324),
    [sym_preproc_line] = STATE(1324),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1325)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(388),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1325),
    [sym_compiler_directive_decl] = STATE(1325),
    [sym_fsi_directive_decl] = STATE(1325),
    [sym_preproc_line] = STATE(1325),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(1369),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1326)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(531),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1326),
    [sym_compiler_directive_decl] = STATE(1326),
    [sym_fsi_directive_decl] = STATE(1326),
    [sym_preproc_line] = STATE(1326),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [aux_sym_preproc_if_in_expression_repeat1] = STATE(1316),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [anon_sym_POUNDendif] = ACTIONS(5741),
    [sym__newline] = ACTIONS(5510),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1327)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(359),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1327),
    [sym_compiler_directive_decl] = STATE(1327),
    [sym_fsi_directive_decl] = STATE(1327),
    [sym_preproc_line] = STATE(1327),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1328)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(311),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1328),
    [sym_compiler_directive_decl] = STATE(1328),
    [sym_fsi_directive_decl] = STATE(1328),
    [sym_preproc_line] = STATE(1328),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_declaration_expression_repeat1] = STATE(1332),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1329)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(325),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1329),
    [sym_compiler_directive_decl] = STATE(1329),
    [sym_fsi_directive_decl] = STATE(1329),
    [sym_preproc_line] = STATE(1329),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1330)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(167),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1330),
    [sym_compiler_directive_decl] = STATE(1330),
    [sym_fsi_directive_decl] = STATE(1330),
    [sym_preproc_line] = STATE(1330),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1331)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(498),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1331),
    [sym_compiler_directive_decl] = STATE(1331),
    [sym_fsi_directive_decl] = STATE(1331),
    [sym_preproc_line] = STATE(1331),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_declaration_expression_repeat1] = STATE(1298),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1332)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(312),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1332),
    [sym_compiler_directive_decl] = STATE(1332),
    [sym_fsi_directive_decl] = STATE(1332),
    [sym_preproc_line] = STATE(1332),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1333)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(285),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1333),
    [sym_compiler_directive_decl] = STATE(1333),
    [sym_fsi_directive_decl] = STATE(1333),
    [sym_preproc_line] = STATE(1333),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1334)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(214),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1334),
    [sym_compiler_directive_decl] = STATE(1334),
    [sym_fsi_directive_decl] = STATE(1334),
    [sym_preproc_line] = STATE(1334),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1335)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(522),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1335),
    [sym_compiler_directive_decl] = STATE(1335),
    [sym_fsi_directive_decl] = STATE(1335),
    [sym_preproc_line] = STATE(1335),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_declaration_expression_repeat1] = STATE(1340),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1336)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(812),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1336),
    [sym_compiler_directive_decl] = STATE(1336),
    [sym_fsi_directive_decl] = STATE(1336),
    [sym_preproc_line] = STATE(1336),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1337)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(224),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1337),
    [sym_compiler_directive_decl] = STATE(1337),
    [sym_fsi_directive_decl] = STATE(1337),
    [sym_preproc_line] = STATE(1337),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1338)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(59),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1338),
    [sym_compiler_directive_decl] = STATE(1338),
    [sym_fsi_directive_decl] = STATE(1338),
    [sym_preproc_line] = STATE(1338),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1339)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(642),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1339),
    [sym_compiler_directive_decl] = STATE(1339),
    [sym_fsi_directive_decl] = STATE(1339),
    [sym_preproc_line] = STATE(1339),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_declaration_expression_repeat1] = STATE(1302),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1340)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(524),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1340),
    [sym_compiler_directive_decl] = STATE(1340),
    [sym_fsi_directive_decl] = STATE(1340),
    [sym_preproc_line] = STATE(1340),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1341)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(91),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1341),
    [sym_compiler_directive_decl] = STATE(1341),
    [sym_fsi_directive_decl] = STATE(1341),
    [sym_preproc_line] = STATE(1341),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1342)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(21904),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(21904),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1342),
    [sym_compiler_directive_decl] = STATE(1342),
    [sym_fsi_directive_decl] = STATE(1342),
    [sym_preproc_line] = STATE(1342),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1343)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(22112),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22112),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1343),
    [sym_compiler_directive_decl] = STATE(1343),
    [sym_fsi_directive_decl] = STATE(1343),
    [sym_preproc_line] = STATE(1343),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1344)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(347),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1344),
    [sym_compiler_directive_decl] = STATE(1344),
    [sym_fsi_directive_decl] = STATE(1344),
    [sym_preproc_line] = STATE(1344),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_declaration_expression_repeat1] = STATE(1293),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1345)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(22324),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22324),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1345),
    [sym_compiler_directive_decl] = STATE(1345),
    [sym_fsi_directive_decl] = STATE(1345),
    [sym_preproc_line] = STATE(1345),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1346)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(222),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1346),
    [sym_compiler_directive_decl] = STATE(1346),
    [sym_fsi_directive_decl] = STATE(1346),
    [sym_preproc_line] = STATE(1346),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_declaration_expression_repeat1] = STATE(1337),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1347)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(22785),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(22785),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1347),
    [sym_compiler_directive_decl] = STATE(1347),
    [sym_fsi_directive_decl] = STATE(1347),
    [sym_preproc_line] = STATE(1347),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1348)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(23),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1348),
    [sym_compiler_directive_decl] = STATE(1348),
    [sym_fsi_directive_decl] = STATE(1348),
    [sym_preproc_line] = STATE(1348),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1349)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19176),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19176),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1349),
    [sym_compiler_directive_decl] = STATE(1349),
    [sym_fsi_directive_decl] = STATE(1349),
    [sym_preproc_line] = STATE(1349),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1350)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(102),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1350),
    [sym_compiler_directive_decl] = STATE(1350),
    [sym_fsi_directive_decl] = STATE(1350),
    [sym_preproc_line] = STATE(1350),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1351)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19290),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19290),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1351),
    [sym_compiler_directive_decl] = STATE(1351),
    [sym_fsi_directive_decl] = STATE(1351),
    [sym_preproc_line] = STATE(1351),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1352)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19404),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19404),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1352),
    [sym_compiler_directive_decl] = STATE(1352),
    [sym_fsi_directive_decl] = STATE(1352),
    [sym_preproc_line] = STATE(1352),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1353)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(556),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1353),
    [sym_compiler_directive_decl] = STATE(1353),
    [sym_fsi_directive_decl] = STATE(1353),
    [sym_preproc_line] = STATE(1353),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_declaration_expression_repeat1] = STATE(1303),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1354)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19516),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19516),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1354),
    [sym_compiler_directive_decl] = STATE(1354),
    [sym_fsi_directive_decl] = STATE(1354),
    [sym_preproc_line] = STATE(1354),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1355)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19593),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19593),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1355),
    [sym_compiler_directive_decl] = STATE(1355),
    [sym_fsi_directive_decl] = STATE(1355),
    [sym_preproc_line] = STATE(1355),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1356)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(288),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1356),
    [sym_compiler_directive_decl] = STATE(1356),
    [sym_fsi_directive_decl] = STATE(1356),
    [sym_preproc_line] = STATE(1356),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_declaration_expression_repeat1] = STATE(1294),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1357)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19669),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19669),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1357),
    [sym_compiler_directive_decl] = STATE(1357),
    [sym_fsi_directive_decl] = STATE(1357),
    [sym_preproc_line] = STATE(1357),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1358)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(683),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1358),
    [sym_compiler_directive_decl] = STATE(1358),
    [sym_fsi_directive_decl] = STATE(1358),
    [sym_preproc_line] = STATE(1358),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1359)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19741),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19741),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1359),
    [sym_compiler_directive_decl] = STATE(1359),
    [sym_fsi_directive_decl] = STATE(1359),
    [sym_preproc_line] = STATE(1359),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1360)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(125),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1360),
    [sym_compiler_directive_decl] = STATE(1360),
    [sym_fsi_directive_decl] = STATE(1360),
    [sym_preproc_line] = STATE(1360),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1361)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19809),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19809),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1361),
    [sym_compiler_directive_decl] = STATE(1361),
    [sym_fsi_directive_decl] = STATE(1361),
    [sym_preproc_line] = STATE(1361),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1362)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19872),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19872),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1362),
    [sym_compiler_directive_decl] = STATE(1362),
    [sym_fsi_directive_decl] = STATE(1362),
    [sym_preproc_line] = STATE(1362),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1363)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(568),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1363),
    [sym_compiler_directive_decl] = STATE(1363),
    [sym_fsi_directive_decl] = STATE(1363),
    [sym_preproc_line] = STATE(1363),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_declaration_expression_repeat1] = STATE(1309),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1364)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19927),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19927),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1364),
    [sym_compiler_directive_decl] = STATE(1364),
    [sym_fsi_directive_decl] = STATE(1364),
    [sym_preproc_line] = STATE(1364),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1365)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(199),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1365),
    [sym_compiler_directive_decl] = STATE(1365),
    [sym_fsi_directive_decl] = STATE(1365),
    [sym_preproc_line] = STATE(1365),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_declaration_expression_repeat1] = STATE(1324),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1366)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(19976),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(19976),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1366),
    [sym_compiler_directive_decl] = STATE(1366),
    [sym_fsi_directive_decl] = STATE(1366),
    [sym_preproc_line] = STATE(1366),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1367)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20459),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20459),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1367),
    [sym_compiler_directive_decl] = STATE(1367),
    [sym_fsi_directive_decl] = STATE(1367),
    [sym_preproc_line] = STATE(1367),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1368)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20020),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20020),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1368),
    [sym_compiler_directive_decl] = STATE(1368),
    [sym_fsi_directive_decl] = STATE(1368),
    [sym_preproc_line] = STATE(1368),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1369)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(389),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1369),
    [sym_compiler_directive_decl] = STATE(1369),
    [sym_fsi_directive_decl] = STATE(1369),
    [sym_preproc_line] = STATE(1369),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1370)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20063),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20063),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1370),
    [sym_compiler_directive_decl] = STATE(1370),
    [sym_fsi_directive_decl] = STATE(1370),
    [sym_preproc_line] = STATE(1370),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1371)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(13),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1371),
    [sym_compiler_directive_decl] = STATE(1371),
    [sym_fsi_directive_decl] = STATE(1371),
    [sym_preproc_line] = STATE(1371),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1372)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20107),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20107),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1372),
    [sym_compiler_directive_decl] = STATE(1372),
    [sym_fsi_directive_decl] = STATE(1372),
    [sym_preproc_line] = STATE(1372),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1373)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20151),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20151),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1373),
    [sym_compiler_directive_decl] = STATE(1373),
    [sym_fsi_directive_decl] = STATE(1373),
    [sym_preproc_line] = STATE(1373),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1374)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(239),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1374),
    [sym_compiler_directive_decl] = STATE(1374),
    [sym_fsi_directive_decl] = STATE(1374),
    [sym_preproc_line] = STATE(1374),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_declaration_expression_repeat1] = STATE(1321),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1375)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20195),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20195),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1375),
    [sym_compiler_directive_decl] = STATE(1375),
    [sym_fsi_directive_decl] = STATE(1375),
    [sym_preproc_line] = STATE(1375),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1376)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20239),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20239),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1376),
    [sym_compiler_directive_decl] = STATE(1376),
    [sym_fsi_directive_decl] = STATE(1376),
    [sym_preproc_line] = STATE(1376),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1377)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(462),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1377),
    [sym_compiler_directive_decl] = STATE(1377),
    [sym_fsi_directive_decl] = STATE(1377),
    [sym_preproc_line] = STATE(1377),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1378)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20283),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20283),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1378),
    [sym_compiler_directive_decl] = STATE(1378),
    [sym_fsi_directive_decl] = STATE(1378),
    [sym_preproc_line] = STATE(1378),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1379)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(106),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1379),
    [sym_compiler_directive_decl] = STATE(1379),
    [sym_fsi_directive_decl] = STATE(1379),
    [sym_preproc_line] = STATE(1379),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_declaration_expression_repeat1] = STATE(8817),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1380)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20327),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20327),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1380),
    [sym_compiler_directive_decl] = STATE(1380),
    [sym_fsi_directive_decl] = STATE(1380),
    [sym_preproc_line] = STATE(1380),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1381)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20371),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20371),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1381),
    [sym_compiler_directive_decl] = STATE(1381),
    [sym_fsi_directive_decl] = STATE(1381),
    [sym_preproc_line] = STATE(1381),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1382)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(251),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1382),
    [sym_compiler_directive_decl] = STATE(1382),
    [sym_fsi_directive_decl] = STATE(1382),
    [sym_preproc_line] = STATE(1382),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_declaration_expression_repeat1] = STATE(1311),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1383)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(720),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_with_field_expression] = STATE(20415),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_field_initializer] = STATE(16882),
    [sym_field_initializers] = STATE(20415),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(4983),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1383),
    [sym_compiler_directive_decl] = STATE(1383),
    [sym_fsi_directive_decl] = STATE(1383),
    [sym_preproc_line] = STATE(1383),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(5355),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1384)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(380),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_and_bang] = STATE(8882),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1384),
    [sym_compiler_directive_decl] = STATE(1384),
    [sym_fsi_directive_decl] = STATE(1384),
    [sym_preproc_line] = STATE(1384),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_declaration_expression_repeat1] = STATE(1377),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_and_BANG] = ACTIONS(5123),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1385)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__expression_block] = STATE(4152),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(215),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1385),
    [sym_compiler_directive_decl] = STATE(1385),
    [sym_fsi_directive_decl] = STATE(1385),
    [sym_preproc_line] = STATE(1385),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__indent] = ACTIONS(5767),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1386)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__expression_block] = STATE(2499),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(24),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1386),
    [sym_compiler_directive_decl] = STATE(1386),
    [sym_fsi_directive_decl] = STATE(1386),
    [sym_preproc_line] = STATE(1386),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__indent] = ACTIONS(5769),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1387)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__expression_block] = STATE(3629),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(151),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1387),
    [sym_compiler_directive_decl] = STATE(1387),
    [sym_fsi_directive_decl] = STATE(1387),
    [sym_preproc_line] = STATE(1387),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__indent] = ACTIONS(5771),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1388)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__expression_block] = STATE(5286),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(270),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1388),
    [sym_compiler_directive_decl] = STATE(1388),
    [sym_fsi_directive_decl] = STATE(1388),
    [sym_preproc_line] = STATE(1388),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__indent] = ACTIONS(5773),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1389)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__expression_block] = STATE(5898),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(687),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1389),
    [sym_compiler_directive_decl] = STATE(1389),
    [sym_fsi_directive_decl] = STATE(1389),
    [sym_preproc_line] = STATE(1389),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__indent] = ACTIONS(5775),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1390)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__expression_block] = STATE(5940),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(541),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1390),
    [sym_compiler_directive_decl] = STATE(1390),
    [sym_fsi_directive_decl] = STATE(1390),
    [sym_preproc_line] = STATE(1390),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__indent] = ACTIONS(5777),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1391)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__expression_block] = STATE(6230),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(590),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1391),
    [sym_compiler_directive_decl] = STATE(1391),
    [sym_fsi_directive_decl] = STATE(1391),
    [sym_preproc_line] = STATE(1391),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__indent] = ACTIONS(5779),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1392)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__expression_block] = STATE(6081),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(555),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1392),
    [sym_compiler_directive_decl] = STATE(1392),
    [sym_fsi_directive_decl] = STATE(1392),
    [sym_preproc_line] = STATE(1392),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__indent] = ACTIONS(5781),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1393)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__expression_block] = STATE(5286),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(917),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1393),
    [sym_compiler_directive_decl] = STATE(1393),
    [sym_fsi_directive_decl] = STATE(1393),
    [sym_preproc_line] = STATE(1393),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__indent] = ACTIONS(5773),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1394)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__expression_block] = STATE(3457),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(126),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1394),
    [sym_compiler_directive_decl] = STATE(1394),
    [sym_fsi_directive_decl] = STATE(1394),
    [sym_preproc_line] = STATE(1394),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__indent] = ACTIONS(5783),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1395)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__expression_block] = STATE(5755),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(313),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1395),
    [sym_compiler_directive_decl] = STATE(1395),
    [sym_fsi_directive_decl] = STATE(1395),
    [sym_preproc_line] = STATE(1395),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__indent] = ACTIONS(5785),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1396)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__expression_block] = STATE(4030),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(181),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1396),
    [sym_compiler_directive_decl] = STATE(1396),
    [sym_fsi_directive_decl] = STATE(1396),
    [sym_preproc_line] = STATE(1396),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__indent] = ACTIONS(5787),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1397)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__expression_block] = STATE(6068),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(500),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1397),
    [sym_compiler_directive_decl] = STATE(1397),
    [sym_fsi_directive_decl] = STATE(1397),
    [sym_preproc_line] = STATE(1397),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__indent] = ACTIONS(5789),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1398)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__expression_block] = STATE(2499),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(60),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1398),
    [sym_compiler_directive_decl] = STATE(1398),
    [sym_fsi_directive_decl] = STATE(1398),
    [sym_preproc_line] = STATE(1398),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__indent] = ACTIONS(5769),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1399)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__expression_block] = STATE(4365),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(226),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1399),
    [sym_compiler_directive_decl] = STATE(1399),
    [sym_fsi_directive_decl] = STATE(1399),
    [sym_preproc_line] = STATE(1399),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__indent] = ACTIONS(5791),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1400)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__expression_block] = STATE(4904),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(245),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1400),
    [sym_compiler_directive_decl] = STATE(1400),
    [sym_fsi_directive_decl] = STATE(1400),
    [sym_preproc_line] = STATE(1400),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__indent] = ACTIONS(5793),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1401)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__expression_block] = STATE(4030),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(205),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1401),
    [sym_compiler_directive_decl] = STATE(1401),
    [sym_fsi_directive_decl] = STATE(1401),
    [sym_preproc_line] = STATE(1401),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__indent] = ACTIONS(5787),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1402)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__expression_block] = STATE(3629),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(282),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1402),
    [sym_compiler_directive_decl] = STATE(1402),
    [sym_fsi_directive_decl] = STATE(1402),
    [sym_preproc_line] = STATE(1402),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__indent] = ACTIONS(5771),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1403)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__expression_block] = STATE(5691),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(571),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1403),
    [sym_compiler_directive_decl] = STATE(1403),
    [sym_fsi_directive_decl] = STATE(1403),
    [sym_preproc_line] = STATE(1403),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__indent] = ACTIONS(5795),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1404)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__expression_block] = STATE(3268),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(68),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1404),
    [sym_compiler_directive_decl] = STATE(1404),
    [sym_fsi_directive_decl] = STATE(1404),
    [sym_preproc_line] = STATE(1404),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__indent] = ACTIONS(5797),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1405)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__expression_block] = STATE(2761),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(52),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1405),
    [sym_compiler_directive_decl] = STATE(1405),
    [sym_fsi_directive_decl] = STATE(1405),
    [sym_preproc_line] = STATE(1405),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__indent] = ACTIONS(5799),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1406)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__expression_block] = STATE(2594),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(14),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1406),
    [sym_compiler_directive_decl] = STATE(1406),
    [sym_fsi_directive_decl] = STATE(1406),
    [sym_preproc_line] = STATE(1406),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__indent] = ACTIONS(5801),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1407)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__expression_block] = STATE(5553),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(303),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1407),
    [sym_compiler_directive_decl] = STATE(1407),
    [sym_fsi_directive_decl] = STATE(1407),
    [sym_preproc_line] = STATE(1407),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__indent] = ACTIONS(5803),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1408)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__expression_block] = STATE(5890),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(513),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1408),
    [sym_compiler_directive_decl] = STATE(1408),
    [sym_fsi_directive_decl] = STATE(1408),
    [sym_preproc_line] = STATE(1408),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__indent] = ACTIONS(5805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1409)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__expression_block] = STATE(5870),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(350),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1409),
    [sym_compiler_directive_decl] = STATE(1409),
    [sym_fsi_directive_decl] = STATE(1409),
    [sym_preproc_line] = STATE(1409),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__indent] = ACTIONS(5807),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1410)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__expression_block] = STATE(5286),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(390),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1410),
    [sym_compiler_directive_decl] = STATE(1410),
    [sym_fsi_directive_decl] = STATE(1410),
    [sym_preproc_line] = STATE(1410),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__indent] = ACTIONS(5773),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1411)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__expression_block] = STATE(5691),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(362),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1411),
    [sym_compiler_directive_decl] = STATE(1411),
    [sym_fsi_directive_decl] = STATE(1411),
    [sym_preproc_line] = STATE(1411),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__indent] = ACTIONS(5795),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1412)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__expression_block] = STATE(6086),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(651),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1412),
    [sym_compiler_directive_decl] = STATE(1412),
    [sym_fsi_directive_decl] = STATE(1412),
    [sym_preproc_line] = STATE(1412),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__indent] = ACTIONS(5809),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1413)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(608),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1413),
    [sym_compiler_directive_decl] = STATE(1413),
    [sym_fsi_directive_decl] = STATE(1413),
    [sym_preproc_line] = STATE(1413),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_COMMA] = ACTIONS(5811),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_GT] = ACTIONS(5811),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1414)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__expression_block] = STATE(4365),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(326),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1414),
    [sym_compiler_directive_decl] = STATE(1414),
    [sym_fsi_directive_decl] = STATE(1414),
    [sym_preproc_line] = STATE(1414),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__indent] = ACTIONS(5791),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1415)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(344),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1415),
    [sym_compiler_directive_decl] = STATE(1415),
    [sym_fsi_directive_decl] = STATE(1415),
    [sym_preproc_line] = STATE(1415),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(5813),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5813),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1416)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__expression_block] = STATE(3601),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(107),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1416),
    [sym_compiler_directive_decl] = STATE(1416),
    [sym_fsi_directive_decl] = STATE(1416),
    [sym_preproc_line] = STATE(1416),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__indent] = ACTIONS(5815),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1417)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__expression_block] = STATE(3615),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(105),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1417),
    [sym_compiler_directive_decl] = STATE(1417),
    [sym_fsi_directive_decl] = STATE(1417),
    [sym_preproc_line] = STATE(1417),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__indent] = ACTIONS(5817),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1418)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__expression_block] = STATE(3955),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(168),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1418),
    [sym_compiler_directive_decl] = STATE(1418),
    [sym_fsi_directive_decl] = STATE(1418),
    [sym_preproc_line] = STATE(1418),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__indent] = ACTIONS(5819),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1419)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__expression_block] = STATE(3184),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(67),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1419),
    [sym_compiler_directive_decl] = STATE(1419),
    [sym_fsi_directive_decl] = STATE(1419),
    [sym_preproc_line] = STATE(1419),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__indent] = ACTIONS(5821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1420)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__expression_block] = STATE(5616),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(291),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1420),
    [sym_compiler_directive_decl] = STATE(1420),
    [sym_fsi_directive_decl] = STATE(1420),
    [sym_preproc_line] = STATE(1420),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__indent] = ACTIONS(5823),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1421)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(370),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1421),
    [sym_compiler_directive_decl] = STATE(1421),
    [sym_fsi_directive_decl] = STATE(1421),
    [sym_preproc_line] = STATE(1421),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_COMMA] = ACTIONS(5813),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_RBRACK] = ACTIONS(5813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1422)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(868),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20074),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1422),
    [sym_compiler_directive_decl] = STATE(1422),
    [sym_fsi_directive_decl] = STATE(1422),
    [sym_preproc_line] = STATE(1422),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1423)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1423),
    [sym_compiler_directive_decl] = STATE(1423),
    [sym_fsi_directive_decl] = STATE(1423),
    [sym_preproc_line] = STATE(1423),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5825),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1424)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(560),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1424),
    [sym_compiler_directive_decl] = STATE(1424),
    [sym_fsi_directive_decl] = STATE(1424),
    [sym_preproc_line] = STATE(1424),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__tuple_marker] = ACTIONS(5827),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1425)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(653),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1425),
    [sym_compiler_directive_decl] = STATE(1425),
    [sym_fsi_directive_decl] = STATE(1425),
    [sym_preproc_line] = STATE(1425),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__tuple_marker] = ACTIONS(5829),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1426)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(817),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(22814),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1426),
    [sym_compiler_directive_decl] = STATE(1426),
    [sym_fsi_directive_decl] = STATE(1426),
    [sym_preproc_line] = STATE(1426),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1427)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1427),
    [sym_compiler_directive_decl] = STATE(1427),
    [sym_fsi_directive_decl] = STATE(1427),
    [sym_preproc_line] = STATE(1427),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5831),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1428)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(316),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1428),
    [sym_compiler_directive_decl] = STATE(1428),
    [sym_fsi_directive_decl] = STATE(1428),
    [sym_preproc_line] = STATE(1428),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__tuple_marker] = ACTIONS(5833),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1429)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(252),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1429),
    [sym_compiler_directive_decl] = STATE(1429),
    [sym_fsi_directive_decl] = STATE(1429),
    [sym_preproc_line] = STATE(1429),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__tuple_marker] = ACTIONS(5835),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1430)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(820),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19193),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1430),
    [sym_compiler_directive_decl] = STATE(1430),
    [sym_fsi_directive_decl] = STATE(1430),
    [sym_preproc_line] = STATE(1430),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1431)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1431),
    [sym_compiler_directive_decl] = STATE(1431),
    [sym_fsi_directive_decl] = STATE(1431),
    [sym_preproc_line] = STATE(1431),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5837),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1432)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(247),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1432),
    [sym_compiler_directive_decl] = STATE(1432),
    [sym_fsi_directive_decl] = STATE(1432),
    [sym_preproc_line] = STATE(1432),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__tuple_marker] = ACTIONS(5839),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1433)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(305),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1433),
    [sym_compiler_directive_decl] = STATE(1433),
    [sym_fsi_directive_decl] = STATE(1433),
    [sym_preproc_line] = STATE(1433),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__tuple_marker] = ACTIONS(5841),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1434)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(823),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19307),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1434),
    [sym_compiler_directive_decl] = STATE(1434),
    [sym_fsi_directive_decl] = STATE(1434),
    [sym_preproc_line] = STATE(1434),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1435)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1435),
    [sym_compiler_directive_decl] = STATE(1435),
    [sym_fsi_directive_decl] = STATE(1435),
    [sym_preproc_line] = STATE(1435),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5843),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1436)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(128),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1436),
    [sym_compiler_directive_decl] = STATE(1436),
    [sym_fsi_directive_decl] = STATE(1436),
    [sym_preproc_line] = STATE(1436),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__tuple_marker] = ACTIONS(5845),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1437)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(62),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1437),
    [sym_compiler_directive_decl] = STATE(1437),
    [sym_fsi_directive_decl] = STATE(1437),
    [sym_preproc_line] = STATE(1437),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__tuple_marker] = ACTIONS(5847),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1438)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(826),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(22357),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1438),
    [sym_compiler_directive_decl] = STATE(1438),
    [sym_fsi_directive_decl] = STATE(1438),
    [sym_preproc_line] = STATE(1438),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1439)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1439),
    [sym_compiler_directive_decl] = STATE(1439),
    [sym_fsi_directive_decl] = STATE(1439),
    [sym_preproc_line] = STATE(1439),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5849),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1440)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(207),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1440),
    [sym_compiler_directive_decl] = STATE(1440),
    [sym_fsi_directive_decl] = STATE(1440),
    [sym_preproc_line] = STATE(1440),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__tuple_marker] = ACTIONS(5851),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1441)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(183),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1441),
    [sym_compiler_directive_decl] = STATE(1441),
    [sym_fsi_directive_decl] = STATE(1441),
    [sym_preproc_line] = STATE(1441),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__tuple_marker] = ACTIONS(5853),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1442)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(829),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19307),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1442),
    [sym_compiler_directive_decl] = STATE(1442),
    [sym_fsi_directive_decl] = STATE(1442),
    [sym_preproc_line] = STATE(1442),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1443)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1443),
    [sym_compiler_directive_decl] = STATE(1443),
    [sym_fsi_directive_decl] = STATE(1443),
    [sym_preproc_line] = STATE(1443),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5855),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1444)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(712),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1444),
    [sym_compiler_directive_decl] = STATE(1444),
    [sym_fsi_directive_decl] = STATE(1444),
    [sym_preproc_line] = STATE(1444),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5857),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1445)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(502),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1445),
    [sym_compiler_directive_decl] = STATE(1445),
    [sym_fsi_directive_decl] = STATE(1445),
    [sym_preproc_line] = STATE(1445),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__tuple_marker] = ACTIONS(5859),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1446)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(832),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19421),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1446),
    [sym_compiler_directive_decl] = STATE(1446),
    [sym_fsi_directive_decl] = STATE(1446),
    [sym_preproc_line] = STATE(1446),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1447)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1447),
    [sym_compiler_directive_decl] = STATE(1447),
    [sym_fsi_directive_decl] = STATE(1447),
    [sym_preproc_line] = STATE(1447),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5861),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1448)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(16),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1448),
    [sym_compiler_directive_decl] = STATE(1448),
    [sym_fsi_directive_decl] = STATE(1448),
    [sym_preproc_line] = STATE(1448),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__tuple_marker] = ACTIONS(5863),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1449)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(692),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1449),
    [sym_compiler_directive_decl] = STATE(1449),
    [sym_fsi_directive_decl] = STATE(1449),
    [sym_preproc_line] = STATE(1449),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__tuple_marker] = ACTIONS(5865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1450)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(228),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1450),
    [sym_compiler_directive_decl] = STATE(1450),
    [sym_fsi_directive_decl] = STATE(1450),
    [sym_preproc_line] = STATE(1450),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__tuple_marker] = ACTIONS(5867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1451)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(835),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19529),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1451),
    [sym_compiler_directive_decl] = STATE(1451),
    [sym_fsi_directive_decl] = STATE(1451),
    [sym_preproc_line] = STATE(1451),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1452)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1452),
    [sym_compiler_directive_decl] = STATE(1452),
    [sym_fsi_directive_decl] = STATE(1452),
    [sym_preproc_line] = STATE(1452),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5869),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1453)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(26),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1453),
    [sym_compiler_directive_decl] = STATE(1453),
    [sym_fsi_directive_decl] = STATE(1453),
    [sym_preproc_line] = STATE(1453),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__tuple_marker] = ACTIONS(5871),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1454)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(543),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1454),
    [sym_compiler_directive_decl] = STATE(1454),
    [sym_fsi_directive_decl] = STATE(1454),
    [sym_preproc_line] = STATE(1454),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__tuple_marker] = ACTIONS(5873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1455)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(838),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19606),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1455),
    [sym_compiler_directive_decl] = STATE(1455),
    [sym_fsi_directive_decl] = STATE(1455),
    [sym_preproc_line] = STATE(1455),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1456)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1456),
    [sym_compiler_directive_decl] = STATE(1456),
    [sym_fsi_directive_decl] = STATE(1456),
    [sym_preproc_line] = STATE(1456),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5875),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1457)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(364),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1457),
    [sym_compiler_directive_decl] = STATE(1457),
    [sym_fsi_directive_decl] = STATE(1457),
    [sym_preproc_line] = STATE(1457),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__tuple_marker] = ACTIONS(5877),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1458)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(919),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1458),
    [sym_compiler_directive_decl] = STATE(1458),
    [sym_fsi_directive_decl] = STATE(1458),
    [sym_preproc_line] = STATE(1458),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__tuple_marker] = ACTIONS(5879),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1459)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(841),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19682),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1459),
    [sym_compiler_directive_decl] = STATE(1459),
    [sym_fsi_directive_decl] = STATE(1459),
    [sym_preproc_line] = STATE(1459),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1460)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1460),
    [sym_compiler_directive_decl] = STATE(1460),
    [sym_fsi_directive_decl] = STATE(1460),
    [sym_preproc_line] = STATE(1460),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5881),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1461)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(109),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1461),
    [sym_compiler_directive_decl] = STATE(1461),
    [sym_fsi_directive_decl] = STATE(1461),
    [sym_preproc_line] = STATE(1461),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__tuple_marker] = ACTIONS(5883),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1462)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(369),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_range_expression] = STATE(18596),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1462),
    [sym_compiler_directive_decl] = STATE(1462),
    [sym_fsi_directive_decl] = STATE(1462),
    [sym_preproc_line] = STATE(1462),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1463)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(844),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19754),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1463),
    [sym_compiler_directive_decl] = STATE(1463),
    [sym_fsi_directive_decl] = STATE(1463),
    [sym_preproc_line] = STATE(1463),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1464)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1464),
    [sym_compiler_directive_decl] = STATE(1464),
    [sym_fsi_directive_decl] = STATE(1464),
    [sym_preproc_line] = STATE(1464),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5885),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1465)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(573),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1465),
    [sym_compiler_directive_decl] = STATE(1465),
    [sym_fsi_directive_decl] = STATE(1465),
    [sym_preproc_line] = STATE(1465),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__tuple_marker] = ACTIONS(5887),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1466)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(78),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1466),
    [sym_compiler_directive_decl] = STATE(1466),
    [sym_fsi_directive_decl] = STATE(1466),
    [sym_preproc_line] = STATE(1466),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__tuple_marker] = ACTIONS(5889),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1467)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(847),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19821),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1467),
    [sym_compiler_directive_decl] = STATE(1467),
    [sym_fsi_directive_decl] = STATE(1467),
    [sym_preproc_line] = STATE(1467),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1468)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1468),
    [sym_compiler_directive_decl] = STATE(1468),
    [sym_fsi_directive_decl] = STATE(1468),
    [sym_preproc_line] = STATE(1468),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5891),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1469)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(515),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1469),
    [sym_compiler_directive_decl] = STATE(1469),
    [sym_fsi_directive_decl] = STATE(1469),
    [sym_preproc_line] = STATE(1469),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__tuple_marker] = ACTIONS(5893),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1470)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(392),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1470),
    [sym_compiler_directive_decl] = STATE(1470),
    [sym_fsi_directive_decl] = STATE(1470),
    [sym_preproc_line] = STATE(1470),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__tuple_marker] = ACTIONS(5895),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1471)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(850),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19307),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1471),
    [sym_compiler_directive_decl] = STATE(1471),
    [sym_fsi_directive_decl] = STATE(1471),
    [sym_preproc_line] = STATE(1471),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1472)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1472),
    [sym_compiler_directive_decl] = STATE(1472),
    [sym_fsi_directive_decl] = STATE(1472),
    [sym_preproc_line] = STATE(1472),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5897),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1473)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(54),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1473),
    [sym_compiler_directive_decl] = STATE(1473),
    [sym_fsi_directive_decl] = STATE(1473),
    [sym_preproc_line] = STATE(1473),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__tuple_marker] = ACTIONS(5899),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1474)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(278),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1474),
    [sym_compiler_directive_decl] = STATE(1474),
    [sym_fsi_directive_decl] = STATE(1474),
    [sym_preproc_line] = STATE(1474),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5857),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1475)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(853),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19884),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1475),
    [sym_compiler_directive_decl] = STATE(1475),
    [sym_fsi_directive_decl] = STATE(1475),
    [sym_preproc_line] = STATE(1475),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1476)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1476),
    [sym_compiler_directive_decl] = STATE(1476),
    [sym_fsi_directive_decl] = STATE(1476),
    [sym_preproc_line] = STATE(1476),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5901),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1477)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(328),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1477),
    [sym_compiler_directive_decl] = STATE(1477),
    [sym_fsi_directive_decl] = STATE(1477),
    [sym_preproc_line] = STATE(1477),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__tuple_marker] = ACTIONS(5903),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1478)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(170),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1478),
    [sym_compiler_directive_decl] = STATE(1478),
    [sym_fsi_directive_decl] = STATE(1478),
    [sym_preproc_line] = STATE(1478),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__tuple_marker] = ACTIONS(5905),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1479)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(856),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19939),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1479),
    [sym_compiler_directive_decl] = STATE(1479),
    [sym_fsi_directive_decl] = STATE(1479),
    [sym_preproc_line] = STATE(1479),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1480)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1480),
    [sym_compiler_directive_decl] = STATE(1480),
    [sym_fsi_directive_decl] = STATE(1480),
    [sym_preproc_line] = STATE(1480),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5907),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1481)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(153),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1481),
    [sym_compiler_directive_decl] = STATE(1481),
    [sym_fsi_directive_decl] = STATE(1481),
    [sym_preproc_line] = STATE(1481),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__tuple_marker] = ACTIONS(5909),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1482)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(70),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1482),
    [sym_compiler_directive_decl] = STATE(1482),
    [sym_fsi_directive_decl] = STATE(1482),
    [sym_preproc_line] = STATE(1482),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__tuple_marker] = ACTIONS(5911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1483)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(859),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19987),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1483),
    [sym_compiler_directive_decl] = STATE(1483),
    [sym_fsi_directive_decl] = STATE(1483),
    [sym_preproc_line] = STATE(1483),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1484)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1484),
    [sym_compiler_directive_decl] = STATE(1484),
    [sym_fsi_directive_decl] = STATE(1484),
    [sym_preproc_line] = STATE(1484),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5913),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1485)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(293),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1485),
    [sym_compiler_directive_decl] = STATE(1485),
    [sym_fsi_directive_decl] = STATE(1485),
    [sym_preproc_line] = STATE(1485),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__tuple_marker] = ACTIONS(5915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1486)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(278),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1486),
    [sym_compiler_directive_decl] = STATE(1486),
    [sym_fsi_directive_decl] = STATE(1486),
    [sym_preproc_line] = STATE(1486),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(1711),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1487)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(862),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20030),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1487),
    [sym_compiler_directive_decl] = STATE(1487),
    [sym_fsi_directive_decl] = STATE(1487),
    [sym_preproc_line] = STATE(1487),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1488)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1488),
    [sym_compiler_directive_decl] = STATE(1488),
    [sym_fsi_directive_decl] = STATE(1488),
    [sym_preproc_line] = STATE(1488),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5917),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1489)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(581),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(22065),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1489),
    [sym_compiler_directive_decl] = STATE(1489),
    [sym_fsi_directive_decl] = STATE(1489),
    [sym_preproc_line] = STATE(1489),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1490)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(593),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1490),
    [sym_compiler_directive_decl] = STATE(1490),
    [sym_fsi_directive_decl] = STATE(1490),
    [sym_preproc_line] = STATE(1490),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__tuple_marker] = ACTIONS(5919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1491)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(865),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(19884),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1491),
    [sym_compiler_directive_decl] = STATE(1491),
    [sym_fsi_directive_decl] = STATE(1491),
    [sym_preproc_line] = STATE(1491),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1492)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1492),
    [sym_compiler_directive_decl] = STATE(1492),
    [sym_fsi_directive_decl] = STATE(1492),
    [sym_preproc_line] = STATE(1492),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5921),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1493)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(352),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1493),
    [sym_compiler_directive_decl] = STATE(1493),
    [sym_fsi_directive_decl] = STATE(1493),
    [sym_preproc_line] = STATE(1493),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__tuple_marker] = ACTIONS(5923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1494)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(802),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(21940),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1494),
    [sym_compiler_directive_decl] = STATE(1494),
    [sym_fsi_directive_decl] = STATE(1494),
    [sym_preproc_line] = STATE(1494),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1495)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(890),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20470),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1495),
    [sym_compiler_directive_decl] = STATE(1495),
    [sym_fsi_directive_decl] = STATE(1495),
    [sym_preproc_line] = STATE(1495),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1496)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1496),
    [sym_compiler_directive_decl] = STATE(1496),
    [sym_fsi_directive_decl] = STATE(1496),
    [sym_preproc_line] = STATE(1496),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5925),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1497)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1497),
    [sym_compiler_directive_decl] = STATE(1497),
    [sym_fsi_directive_decl] = STATE(1497),
    [sym_preproc_line] = STATE(1497),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5927),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1498)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(460),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_range_expression] = STATE(18744),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1498),
    [sym_compiler_directive_decl] = STATE(1498),
    [sym_fsi_directive_decl] = STATE(1498),
    [sym_preproc_line] = STATE(1498),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1499)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(871),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20118),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1499),
    [sym_compiler_directive_decl] = STATE(1499),
    [sym_fsi_directive_decl] = STATE(1499),
    [sym_preproc_line] = STATE(1499),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1500)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1500),
    [sym_compiler_directive_decl] = STATE(1500),
    [sym_fsi_directive_decl] = STATE(1500),
    [sym_preproc_line] = STATE(1500),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5929),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1501)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1501),
    [sym_compiler_directive_decl] = STATE(1501),
    [sym_fsi_directive_decl] = STATE(1501),
    [sym_preproc_line] = STATE(1501),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5931),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1502)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(712),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1502),
    [sym_compiler_directive_decl] = STATE(1502),
    [sym_fsi_directive_decl] = STATE(1502),
    [sym_preproc_line] = STATE(1502),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__dedent] = ACTIONS(5933),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1503)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(874),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20030),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1503),
    [sym_compiler_directive_decl] = STATE(1503),
    [sym_fsi_directive_decl] = STATE(1503),
    [sym_preproc_line] = STATE(1503),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1504)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1504),
    [sym_compiler_directive_decl] = STATE(1504),
    [sym_fsi_directive_decl] = STATE(1504),
    [sym_preproc_line] = STATE(1504),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5935),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1505)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(810),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(22129),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1505),
    [sym_compiler_directive_decl] = STATE(1505),
    [sym_fsi_directive_decl] = STATE(1505),
    [sym_preproc_line] = STATE(1505),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1506)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1506),
    [sym_compiler_directive_decl] = STATE(1506),
    [sym_fsi_directive_decl] = STATE(1506),
    [sym_preproc_line] = STATE(1506),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5937),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1507)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(877),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20162),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1507),
    [sym_compiler_directive_decl] = STATE(1507),
    [sym_fsi_directive_decl] = STATE(1507),
    [sym_preproc_line] = STATE(1507),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1508)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1508),
    [sym_compiler_directive_decl] = STATE(1508),
    [sym_fsi_directive_decl] = STATE(1508),
    [sym_preproc_line] = STATE(1508),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5939),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1509)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(469),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_range_expression] = STATE(18996),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1509),
    [sym_compiler_directive_decl] = STATE(1509),
    [sym_fsi_directive_decl] = STATE(1509),
    [sym_preproc_line] = STATE(1509),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1510)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(115),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1510),
    [sym_compiler_directive_decl] = STATE(1510),
    [sym_fsi_directive_decl] = STATE(1510),
    [sym_preproc_line] = STATE(1510),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__tuple_marker] = ACTIONS(5941),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1511)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(880),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20206),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1511),
    [sym_compiler_directive_decl] = STATE(1511),
    [sym_fsi_directive_decl] = STATE(1511),
    [sym_preproc_line] = STATE(1511),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1512)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1512),
    [sym_compiler_directive_decl] = STATE(1512),
    [sym_fsi_directive_decl] = STATE(1512),
    [sym_preproc_line] = STATE(1512),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5943),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1513)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(217),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1513),
    [sym_compiler_directive_decl] = STATE(1513),
    [sym_fsi_directive_decl] = STATE(1513),
    [sym_preproc_line] = STATE(1513),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__tuple_marker] = ACTIONS(5945),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1514)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(814),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(22357),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1514),
    [sym_compiler_directive_decl] = STATE(1514),
    [sym_fsi_directive_decl] = STATE(1514),
    [sym_preproc_line] = STATE(1514),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1515)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(883),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20250),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1515),
    [sym_compiler_directive_decl] = STATE(1515),
    [sym_fsi_directive_decl] = STATE(1515),
    [sym_preproc_line] = STATE(1515),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1516)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1516),
    [sym_compiler_directive_decl] = STATE(1516),
    [sym_fsi_directive_decl] = STATE(1516),
    [sym_preproc_line] = STATE(1516),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_with] = ACTIONS(5947),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1517)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(885),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20294),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1517),
    [sym_compiler_directive_decl] = STATE(1517),
    [sym_fsi_directive_decl] = STATE(1517),
    [sym_preproc_line] = STATE(1517),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1518)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(887),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20338),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1518),
    [sym_compiler_directive_decl] = STATE(1518),
    [sym_fsi_directive_decl] = STATE(1518),
    [sym_preproc_line] = STATE(1518),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1519)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(888),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20382),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1519),
    [sym_compiler_directive_decl] = STATE(1519),
    [sym_fsi_directive_decl] = STATE(1519),
    [sym_preproc_line] = STATE(1519),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1520)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(889),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_range_expression] = STATE(20426),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1520),
    [sym_compiler_directive_decl] = STATE(1520),
    [sym_fsi_directive_decl] = STATE(1520),
    [sym_preproc_line] = STATE(1520),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1521)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(258),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1521),
    [sym_compiler_directive_decl] = STATE(1521),
    [sym_fsi_directive_decl] = STATE(1521),
    [sym_preproc_line] = STATE(1521),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__tuple_marker] = ACTIONS(5949),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1522)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(770),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1522),
    [sym_compiler_directive_decl] = STATE(1522),
    [sym_fsi_directive_decl] = STATE(1522),
    [sym_preproc_line] = STATE(1522),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1523)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(454),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1523),
    [sym_compiler_directive_decl] = STATE(1523),
    [sym_fsi_directive_decl] = STATE(1523),
    [sym_preproc_line] = STATE(1523),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1524)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(662),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1524),
    [sym_compiler_directive_decl] = STATE(1524),
    [sym_fsi_directive_decl] = STATE(1524),
    [sym_preproc_line] = STATE(1524),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1525)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(666),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1525),
    [sym_compiler_directive_decl] = STATE(1525),
    [sym_fsi_directive_decl] = STATE(1525),
    [sym_preproc_line] = STATE(1525),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1526)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(667),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1526),
    [sym_compiler_directive_decl] = STATE(1526),
    [sym_fsi_directive_decl] = STATE(1526),
    [sym_preproc_line] = STATE(1526),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1527)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(210),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1527),
    [sym_compiler_directive_decl] = STATE(1527),
    [sym_fsi_directive_decl] = STATE(1527),
    [sym_preproc_line] = STATE(1527),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1528)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(195),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1528),
    [sym_compiler_directive_decl] = STATE(1528),
    [sym_fsi_directive_decl] = STATE(1528),
    [sym_preproc_line] = STATE(1528),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1529)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(674),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1529),
    [sym_compiler_directive_decl] = STATE(1529),
    [sym_fsi_directive_decl] = STATE(1529),
    [sym_preproc_line] = STATE(1529),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1530)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(182),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1530),
    [sym_compiler_directive_decl] = STATE(1530),
    [sym_fsi_directive_decl] = STATE(1530),
    [sym_preproc_line] = STATE(1530),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1531)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(693),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1531),
    [sym_compiler_directive_decl] = STATE(1531),
    [sym_fsi_directive_decl] = STATE(1531),
    [sym_preproc_line] = STATE(1531),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1532)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(697),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1532),
    [sym_compiler_directive_decl] = STATE(1532),
    [sym_fsi_directive_decl] = STATE(1532),
    [sym_preproc_line] = STATE(1532),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1533)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(698),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1533),
    [sym_compiler_directive_decl] = STATE(1533),
    [sym_fsi_directive_decl] = STATE(1533),
    [sym_preproc_line] = STATE(1533),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1534)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(231),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1534),
    [sym_compiler_directive_decl] = STATE(1534),
    [sym_fsi_directive_decl] = STATE(1534),
    [sym_preproc_line] = STATE(1534),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1535)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(702),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1535),
    [sym_compiler_directive_decl] = STATE(1535),
    [sym_fsi_directive_decl] = STATE(1535),
    [sym_preproc_line] = STATE(1535),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1536)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(808),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1536),
    [sym_compiler_directive_decl] = STATE(1536),
    [sym_fsi_directive_decl] = STATE(1536),
    [sym_preproc_line] = STATE(1536),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1537)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(922),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1537),
    [sym_compiler_directive_decl] = STATE(1537),
    [sym_fsi_directive_decl] = STATE(1537),
    [sym_preproc_line] = STATE(1537),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1538)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(703),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1538),
    [sym_compiler_directive_decl] = STATE(1538),
    [sym_fsi_directive_decl] = STATE(1538),
    [sym_preproc_line] = STATE(1538),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1539)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(712),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1539),
    [sym_compiler_directive_decl] = STATE(1539),
    [sym_fsi_directive_decl] = STATE(1539),
    [sym_preproc_line] = STATE(1539),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1540)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(178),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1540),
    [sym_compiler_directive_decl] = STATE(1540),
    [sym_fsi_directive_decl] = STATE(1540),
    [sym_preproc_line] = STATE(1540),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1541)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(509),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1541),
    [sym_compiler_directive_decl] = STATE(1541),
    [sym_fsi_directive_decl] = STATE(1541),
    [sym_preproc_line] = STATE(1541),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1542)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(707),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1542),
    [sym_compiler_directive_decl] = STATE(1542),
    [sym_fsi_directive_decl] = STATE(1542),
    [sym_preproc_line] = STATE(1542),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1543)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(519),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1543),
    [sym_compiler_directive_decl] = STATE(1543),
    [sym_fsi_directive_decl] = STATE(1543),
    [sym_preproc_line] = STATE(1543),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1544)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(806),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1544),
    [sym_compiler_directive_decl] = STATE(1544),
    [sym_fsi_directive_decl] = STATE(1544),
    [sym_preproc_line] = STATE(1544),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1545)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(225),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1545),
    [sym_compiler_directive_decl] = STATE(1545),
    [sym_fsi_directive_decl] = STATE(1545),
    [sym_preproc_line] = STATE(1545),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1546)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(525),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1546),
    [sym_compiler_directive_decl] = STATE(1546),
    [sym_fsi_directive_decl] = STATE(1546),
    [sym_preproc_line] = STATE(1546),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1547)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(535),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1547),
    [sym_compiler_directive_decl] = STATE(1547),
    [sym_fsi_directive_decl] = STATE(1547),
    [sym_preproc_line] = STATE(1547),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1548)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(579),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1548),
    [sym_compiler_directive_decl] = STATE(1548),
    [sym_fsi_directive_decl] = STATE(1548),
    [sym_preproc_line] = STATE(1548),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1549)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(652),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1549),
    [sym_compiler_directive_decl] = STATE(1549),
    [sym_fsi_directive_decl] = STATE(1549),
    [sym_preproc_line] = STATE(1549),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1550)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(969),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1550),
    [sym_compiler_directive_decl] = STATE(1550),
    [sym_fsi_directive_decl] = STATE(1550),
    [sym_preproc_line] = STATE(1550),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1551)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(547),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1551),
    [sym_compiler_directive_decl] = STATE(1551),
    [sym_fsi_directive_decl] = STATE(1551),
    [sym_preproc_line] = STATE(1551),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1552)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(184),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1552),
    [sym_compiler_directive_decl] = STATE(1552),
    [sym_fsi_directive_decl] = STATE(1552),
    [sym_preproc_line] = STATE(1552),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1553)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(185),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1553),
    [sym_compiler_directive_decl] = STATE(1553),
    [sym_fsi_directive_decl] = STATE(1553),
    [sym_preproc_line] = STATE(1553),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1554)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(176),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1554),
    [sym_compiler_directive_decl] = STATE(1554),
    [sym_fsi_directive_decl] = STATE(1554),
    [sym_preproc_line] = STATE(1554),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1555)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(36),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1555),
    [sym_compiler_directive_decl] = STATE(1555),
    [sym_fsi_directive_decl] = STATE(1555),
    [sym_preproc_line] = STATE(1555),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1556)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(34),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1556),
    [sym_compiler_directive_decl] = STATE(1556),
    [sym_fsi_directive_decl] = STATE(1556),
    [sym_preproc_line] = STATE(1556),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1557)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(596),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1557),
    [sym_compiler_directive_decl] = STATE(1557),
    [sym_fsi_directive_decl] = STATE(1557),
    [sym_preproc_line] = STATE(1557),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1558)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(322),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1558),
    [sym_compiler_directive_decl] = STATE(1558),
    [sym_fsi_directive_decl] = STATE(1558),
    [sym_preproc_line] = STATE(1558),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1559)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(186),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1559),
    [sym_compiler_directive_decl] = STATE(1559),
    [sym_fsi_directive_decl] = STATE(1559),
    [sym_preproc_line] = STATE(1559),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1560)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(327),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1560),
    [sym_compiler_directive_decl] = STATE(1560),
    [sym_fsi_directive_decl] = STATE(1560),
    [sym_preproc_line] = STATE(1560),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1561)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(329),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1561),
    [sym_compiler_directive_decl] = STATE(1561),
    [sym_fsi_directive_decl] = STATE(1561),
    [sym_preproc_line] = STATE(1561),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1562)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(330),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1562),
    [sym_compiler_directive_decl] = STATE(1562),
    [sym_fsi_directive_decl] = STATE(1562),
    [sym_preproc_line] = STATE(1562),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1563)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(331),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1563),
    [sym_compiler_directive_decl] = STATE(1563),
    [sym_fsi_directive_decl] = STATE(1563),
    [sym_preproc_line] = STATE(1563),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1564)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(612),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1564),
    [sym_compiler_directive_decl] = STATE(1564),
    [sym_fsi_directive_decl] = STATE(1564),
    [sym_preproc_line] = STATE(1564),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1565)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(332),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1565),
    [sym_compiler_directive_decl] = STATE(1565),
    [sym_fsi_directive_decl] = STATE(1565),
    [sym_preproc_line] = STATE(1565),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1566)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(333),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1566),
    [sym_compiler_directive_decl] = STATE(1566),
    [sym_fsi_directive_decl] = STATE(1566),
    [sym_preproc_line] = STATE(1566),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1567)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(187),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1567),
    [sym_compiler_directive_decl] = STATE(1567),
    [sym_fsi_directive_decl] = STATE(1567),
    [sym_preproc_line] = STATE(1567),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1568)] = {
    [sym_function_or_value_defn] = STATE(1317),
    [sym__mutable_expression] = STATE(21860),
    [sym__expression] = STATE(334),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22376),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1558),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1568),
    [sym_compiler_directive_decl] = STATE(1568),
    [sym_fsi_directive_decl] = STATE(1568),
    [sym_preproc_line] = STATE(1568),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_return] = ACTIONS(2083),
    [anon_sym_return_BANG] = ACTIONS(2085),
    [anon_sym_yield] = ACTIONS(2083),
    [anon_sym_yield_BANG] = ACTIONS(2085),
    [anon_sym_lazy] = ACTIONS(2083),
    [anon_sym_assert] = ACTIONS(2083),
    [anon_sym_upcast] = ACTIONS(2083),
    [anon_sym_downcast] = ACTIONS(2083),
    [anon_sym_fixed] = ACTIONS(2083),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(2089),
    [anon_sym_use_BANG] = ACTIONS(2091),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2097),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1569)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(918),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1569),
    [sym_compiler_directive_decl] = STATE(1569),
    [sym_fsi_directive_decl] = STATE(1569),
    [sym_preproc_line] = STATE(1569),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1570)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(265),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1570),
    [sym_compiler_directive_decl] = STATE(1570),
    [sym_fsi_directive_decl] = STATE(1570),
    [sym_preproc_line] = STATE(1570),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1571)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(927),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1571),
    [sym_compiler_directive_decl] = STATE(1571),
    [sym_fsi_directive_decl] = STATE(1571),
    [sym_preproc_line] = STATE(1571),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1572)] = {
    [sym_function_or_value_defn] = STATE(1319),
    [sym__mutable_expression] = STATE(20331),
    [sym__expression] = STATE(177),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(22333),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1540),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1572),
    [sym_compiler_directive_decl] = STATE(1572),
    [sym_fsi_directive_decl] = STATE(1572),
    [sym_preproc_line] = STATE(1572),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(1276),
    [anon_sym_return] = ACTIONS(1276),
    [anon_sym_return_BANG] = ACTIONS(1278),
    [anon_sym_yield] = ACTIONS(1276),
    [anon_sym_yield_BANG] = ACTIONS(1278),
    [anon_sym_lazy] = ACTIONS(1276),
    [anon_sym_assert] = ACTIONS(1276),
    [anon_sym_upcast] = ACTIONS(1276),
    [anon_sym_downcast] = ACTIONS(1276),
    [anon_sym_fixed] = ACTIONS(1276),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(1284),
    [anon_sym_use_BANG] = ACTIONS(1286),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1296),
    [sym_xint] = ACTIONS(1298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1573)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(617),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1573),
    [sym_compiler_directive_decl] = STATE(1573),
    [sym_fsi_directive_decl] = STATE(1573),
    [sym_preproc_line] = STATE(1573),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1574)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(310),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1574),
    [sym_compiler_directive_decl] = STATE(1574),
    [sym_fsi_directive_decl] = STATE(1574),
    [sym_preproc_line] = STATE(1574),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1575)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(253),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1575),
    [sym_compiler_directive_decl] = STATE(1575),
    [sym_fsi_directive_decl] = STATE(1575),
    [sym_preproc_line] = STATE(1575),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1576)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(254),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1576),
    [sym_compiler_directive_decl] = STATE(1576),
    [sym_fsi_directive_decl] = STATE(1576),
    [sym_preproc_line] = STATE(1576),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1577)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(263),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1577),
    [sym_compiler_directive_decl] = STATE(1577),
    [sym_fsi_directive_decl] = STATE(1577),
    [sym_preproc_line] = STATE(1577),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1578)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(255),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1578),
    [sym_compiler_directive_decl] = STATE(1578),
    [sym_fsi_directive_decl] = STATE(1578),
    [sym_preproc_line] = STATE(1578),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1579)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(549),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1579),
    [sym_compiler_directive_decl] = STATE(1579),
    [sym_fsi_directive_decl] = STATE(1579),
    [sym_preproc_line] = STATE(1579),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1580)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(256),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1580),
    [sym_compiler_directive_decl] = STATE(1580),
    [sym_fsi_directive_decl] = STATE(1580),
    [sym_preproc_line] = STATE(1580),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1581)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(264),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1581),
    [sym_compiler_directive_decl] = STATE(1581),
    [sym_fsi_directive_decl] = STATE(1581),
    [sym_preproc_line] = STATE(1581),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1582)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(279),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1582),
    [sym_compiler_directive_decl] = STATE(1582),
    [sym_fsi_directive_decl] = STATE(1582),
    [sym_preproc_line] = STATE(1582),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1583)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(544),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1583),
    [sym_compiler_directive_decl] = STATE(1583),
    [sym_fsi_directive_decl] = STATE(1583),
    [sym_preproc_line] = STATE(1583),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1584)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(545),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1584),
    [sym_compiler_directive_decl] = STATE(1584),
    [sym_fsi_directive_decl] = STATE(1584),
    [sym_preproc_line] = STATE(1584),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1585)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(532),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1585),
    [sym_compiler_directive_decl] = STATE(1585),
    [sym_fsi_directive_decl] = STATE(1585),
    [sym_preproc_line] = STATE(1585),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1586)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(546),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1586),
    [sym_compiler_directive_decl] = STATE(1586),
    [sym_fsi_directive_decl] = STATE(1586),
    [sym_preproc_line] = STATE(1586),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1587)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(534),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1587),
    [sym_compiler_directive_decl] = STATE(1587),
    [sym_fsi_directive_decl] = STATE(1587),
    [sym_preproc_line] = STATE(1587),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1588)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(520),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1588),
    [sym_compiler_directive_decl] = STATE(1588),
    [sym_fsi_directive_decl] = STATE(1588),
    [sym_preproc_line] = STATE(1588),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1589)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(920),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1589),
    [sym_compiler_directive_decl] = STATE(1589),
    [sym_fsi_directive_decl] = STATE(1589),
    [sym_preproc_line] = STATE(1589),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1590)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(921),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1590),
    [sym_compiler_directive_decl] = STATE(1590),
    [sym_fsi_directive_decl] = STATE(1590),
    [sym_preproc_line] = STATE(1590),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1591)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(622),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1591),
    [sym_compiler_directive_decl] = STATE(1591),
    [sym_fsi_directive_decl] = STATE(1591),
    [sym_preproc_line] = STATE(1591),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1592)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(154),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1592),
    [sym_compiler_directive_decl] = STATE(1592),
    [sym_fsi_directive_decl] = STATE(1592),
    [sym_preproc_line] = STATE(1592),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1593)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(155),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1593),
    [sym_compiler_directive_decl] = STATE(1593),
    [sym_fsi_directive_decl] = STATE(1593),
    [sym_preproc_line] = STATE(1593),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1594)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(150),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1594),
    [sym_compiler_directive_decl] = STATE(1594),
    [sym_fsi_directive_decl] = STATE(1594),
    [sym_preproc_line] = STATE(1594),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1595)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(724),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1595),
    [sym_compiler_directive_decl] = STATE(1595),
    [sym_fsi_directive_decl] = STATE(1595),
    [sym_preproc_line] = STATE(1595),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1596)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(156),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1596),
    [sym_compiler_directive_decl] = STATE(1596),
    [sym_fsi_directive_decl] = STATE(1596),
    [sym_preproc_line] = STATE(1596),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1597)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(159),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1597),
    [sym_compiler_directive_decl] = STATE(1597),
    [sym_fsi_directive_decl] = STATE(1597),
    [sym_preproc_line] = STATE(1597),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1598)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(923),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1598),
    [sym_compiler_directive_decl] = STATE(1598),
    [sym_fsi_directive_decl] = STATE(1598),
    [sym_preproc_line] = STATE(1598),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1599)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(148),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1599),
    [sym_compiler_directive_decl] = STATE(1599),
    [sym_fsi_directive_decl] = STATE(1599),
    [sym_preproc_line] = STATE(1599),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1600)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(260),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1600),
    [sym_compiler_directive_decl] = STATE(1600),
    [sym_fsi_directive_decl] = STATE(1600),
    [sym_preproc_line] = STATE(1600),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1601)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(147),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1601),
    [sym_compiler_directive_decl] = STATE(1601),
    [sym_fsi_directive_decl] = STATE(1601),
    [sym_preproc_line] = STATE(1601),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1602)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(924),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1602),
    [sym_compiler_directive_decl] = STATE(1602),
    [sym_fsi_directive_decl] = STATE(1602),
    [sym_preproc_line] = STATE(1602),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1603)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(804),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1603),
    [sym_compiler_directive_decl] = STATE(1603),
    [sym_fsi_directive_decl] = STATE(1603),
    [sym_preproc_line] = STATE(1603),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1604)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(355),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1604),
    [sym_compiler_directive_decl] = STATE(1604),
    [sym_fsi_directive_decl] = STATE(1604),
    [sym_preproc_line] = STATE(1604),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1605)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(497),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1605),
    [sym_compiler_directive_decl] = STATE(1605),
    [sym_fsi_directive_decl] = STATE(1605),
    [sym_preproc_line] = STATE(1605),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1606)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(320),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1606),
    [sym_compiler_directive_decl] = STATE(1606),
    [sym_fsi_directive_decl] = STATE(1606),
    [sym_preproc_line] = STATE(1606),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1607)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(335),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1607),
    [sym_compiler_directive_decl] = STATE(1607),
    [sym_fsi_directive_decl] = STATE(1607),
    [sym_preproc_line] = STATE(1607),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1608)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(402),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1608),
    [sym_compiler_directive_decl] = STATE(1608),
    [sym_fsi_directive_decl] = STATE(1608),
    [sym_preproc_line] = STATE(1608),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1609)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(309),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1609),
    [sym_compiler_directive_decl] = STATE(1609),
    [sym_fsi_directive_decl] = STATE(1609),
    [sym_preproc_line] = STATE(1609),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1610)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(336),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1610),
    [sym_compiler_directive_decl] = STATE(1610),
    [sym_fsi_directive_decl] = STATE(1610),
    [sym_preproc_line] = STATE(1610),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1611)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(598),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1611),
    [sym_compiler_directive_decl] = STATE(1611),
    [sym_fsi_directive_decl] = STATE(1611),
    [sym_preproc_line] = STATE(1611),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1612)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(142),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1612),
    [sym_compiler_directive_decl] = STATE(1612),
    [sym_fsi_directive_decl] = STATE(1612),
    [sym_preproc_line] = STATE(1612),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1613)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(143),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1613),
    [sym_compiler_directive_decl] = STATE(1613),
    [sym_fsi_directive_decl] = STATE(1613),
    [sym_preproc_line] = STATE(1613),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1614)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(337),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1614),
    [sym_compiler_directive_decl] = STATE(1614),
    [sym_fsi_directive_decl] = STATE(1614),
    [sym_preproc_line] = STATE(1614),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1615)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(144),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1615),
    [sym_compiler_directive_decl] = STATE(1615),
    [sym_fsi_directive_decl] = STATE(1615),
    [sym_preproc_line] = STATE(1615),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1616)] = {
    [sym_function_or_value_defn] = STATE(1322),
    [sym__mutable_expression] = STATE(20380),
    [sym__expression] = STATE(281),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22364),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1570),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1616),
    [sym_compiler_directive_decl] = STATE(1616),
    [sym_fsi_directive_decl] = STATE(1616),
    [sym_preproc_line] = STATE(1616),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1645),
    [anon_sym_return] = ACTIONS(1645),
    [anon_sym_return_BANG] = ACTIONS(1647),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_yield_BANG] = ACTIONS(1647),
    [anon_sym_lazy] = ACTIONS(1645),
    [anon_sym_assert] = ACTIONS(1645),
    [anon_sym_upcast] = ACTIONS(1645),
    [anon_sym_downcast] = ACTIONS(1645),
    [anon_sym_fixed] = ACTIONS(1645),
    [anon_sym_for] = ACTIONS(1651),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(1669),
    [anon_sym_use_BANG] = ACTIONS(1671),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1703),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1617)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(387),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1617),
    [sym_compiler_directive_decl] = STATE(1617),
    [sym_fsi_directive_decl] = STATE(1617),
    [sym_preproc_line] = STATE(1617),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1618)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(508),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1618),
    [sym_compiler_directive_decl] = STATE(1618),
    [sym_fsi_directive_decl] = STATE(1618),
    [sym_preproc_line] = STATE(1618),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1619)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(299),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1619),
    [sym_compiler_directive_decl] = STATE(1619),
    [sym_fsi_directive_decl] = STATE(1619),
    [sym_preproc_line] = STATE(1619),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1620)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(393),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1620),
    [sym_compiler_directive_decl] = STATE(1620),
    [sym_fsi_directive_decl] = STATE(1620),
    [sym_preproc_line] = STATE(1620),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1621)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(394),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1621),
    [sym_compiler_directive_decl] = STATE(1621),
    [sym_fsi_directive_decl] = STATE(1621),
    [sym_preproc_line] = STATE(1621),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1622)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(385),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1622),
    [sym_compiler_directive_decl] = STATE(1622),
    [sym_fsi_directive_decl] = STATE(1622),
    [sym_preproc_line] = STATE(1622),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1623)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(346),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1623),
    [sym_compiler_directive_decl] = STATE(1623),
    [sym_fsi_directive_decl] = STATE(1623),
    [sym_preproc_line] = STATE(1623),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1624)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(395),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1624),
    [sym_compiler_directive_decl] = STATE(1624),
    [sym_fsi_directive_decl] = STATE(1624),
    [sym_preproc_line] = STATE(1624),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1625)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(396),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1625),
    [sym_compiler_directive_decl] = STATE(1625),
    [sym_fsi_directive_decl] = STATE(1625),
    [sym_preproc_line] = STATE(1625),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1626)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(386),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1626),
    [sym_compiler_directive_decl] = STATE(1626),
    [sym_fsi_directive_decl] = STATE(1626),
    [sym_preproc_line] = STATE(1626),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1627)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(539),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1627),
    [sym_compiler_directive_decl] = STATE(1627),
    [sym_fsi_directive_decl] = STATE(1627),
    [sym_preproc_line] = STATE(1627),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1628)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(294),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1628),
    [sym_compiler_directive_decl] = STATE(1628),
    [sym_fsi_directive_decl] = STATE(1628),
    [sym_preproc_line] = STATE(1628),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1629)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(295),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1629),
    [sym_compiler_directive_decl] = STATE(1629),
    [sym_fsi_directive_decl] = STATE(1629),
    [sym_preproc_line] = STATE(1629),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1630)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(466),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1630),
    [sym_compiler_directive_decl] = STATE(1630),
    [sym_fsi_directive_decl] = STATE(1630),
    [sym_preproc_line] = STATE(1630),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1631)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(297),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1631),
    [sym_compiler_directive_decl] = STATE(1631),
    [sym_fsi_directive_decl] = STATE(1631),
    [sym_preproc_line] = STATE(1631),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1632)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(467),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1632),
    [sym_compiler_directive_decl] = STATE(1632),
    [sym_fsi_directive_decl] = STATE(1632),
    [sym_preproc_line] = STATE(1632),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1633)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(377),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1633),
    [sym_compiler_directive_decl] = STATE(1633),
    [sym_fsi_directive_decl] = STATE(1633),
    [sym_preproc_line] = STATE(1633),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1634)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(376),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1634),
    [sym_compiler_directive_decl] = STATE(1634),
    [sym_fsi_directive_decl] = STATE(1634),
    [sym_preproc_line] = STATE(1634),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1635)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(630),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1635),
    [sym_compiler_directive_decl] = STATE(1635),
    [sym_fsi_directive_decl] = STATE(1635),
    [sym_preproc_line] = STATE(1635),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1636)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(145),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1636),
    [sym_compiler_directive_decl] = STATE(1636),
    [sym_fsi_directive_decl] = STATE(1636),
    [sym_preproc_line] = STATE(1636),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1637)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(146),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1637),
    [sym_compiler_directive_decl] = STATE(1637),
    [sym_fsi_directive_decl] = STATE(1637),
    [sym_preproc_line] = STATE(1637),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1638)] = {
    [sym_function_or_value_defn] = STATE(1328),
    [sym__mutable_expression] = STATE(19272),
    [sym__expression] = STATE(314),
    [sym_literal_expression] = STATE(5660),
    [sym_srtp_call_expression] = STATE(5660),
    [sym__srtp_type_argument] = STATE(22388),
    [sym_long_identifier_or_op] = STATE(4984),
    [sym_tuple_expression] = STATE(5660),
    [sym_brace_expression] = STATE(5660),
    [sym_anon_record_expression] = STATE(5660),
    [sym_prefixed_expression] = STATE(5660),
    [sym_typecast_expression] = STATE(5660),
    [sym_for_expression] = STATE(5660),
    [sym_while_expression] = STATE(5660),
    [sym__if_branch] = STATE(19383),
    [sym_if_expression] = STATE(5660),
    [sym_fun_expression] = STATE(5660),
    [sym_try_expression] = STATE(5660),
    [sym_match_expression] = STATE(5660),
    [sym_function_expression] = STATE(5660),
    [sym_mutate_expression] = STATE(5660),
    [sym_index_expression] = STATE(4984),
    [sym_typed_expression] = STATE(5660),
    [sym_declaration_expression] = STATE(5660),
    [sym_do_expression] = STATE(5660),
    [sym_list_expression] = STATE(5660),
    [sym_array_expression] = STATE(5660),
    [sym_begin_end_expression] = STATE(5660),
    [sym_paren_expression] = STATE(5660),
    [sym__high_prec_app] = STATE(5675),
    [sym__low_prec_app] = STATE(5675),
    [sym_application_expression] = STATE(5660),
    [sym_dot_expression] = STATE(4984),
    [sym_infix_expression] = STATE(5660),
    [sym_ce_expression] = STATE(5660),
    [sym_sequential_expression] = STATE(5660),
    [sym_char] = STATE(5754),
    [sym_format_string] = STATE(5805),
    [sym__string_literal] = STATE(5805),
    [sym_string] = STATE(5754),
    [sym_verbatim_string] = STATE(5754),
    [sym_bytearray] = STATE(5754),
    [sym_verbatim_bytearray] = STATE(5754),
    [sym_format_triple_quoted_string] = STATE(5813),
    [sym_triple_quoted_string] = STATE(5754),
    [sym_const] = STATE(5660),
    [sym_long_identifier] = STATE(5019),
    [sym_active_pattern] = STATE(5108),
    [sym__identifier_or_op] = STATE(5020),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1574),
    [sym_sbyte] = STATE(5754),
    [sym_byte] = STATE(5754),
    [sym_int16] = STATE(5754),
    [sym_uint16] = STATE(5754),
    [sym_int32] = STATE(5754),
    [sym_uint32] = STATE(5754),
    [sym_nativeint] = STATE(5754),
    [sym_unativeint] = STATE(5754),
    [sym_int64] = STATE(5754),
    [sym_uint64] = STATE(5754),
    [sym_ieee32] = STATE(5754),
    [sym_ieee64] = STATE(5754),
    [sym_bignum] = STATE(5754),
    [sym_decimal] = STATE(5754),
    [sym_float] = STATE(4854),
    [sym_block_comment] = STATE(1638),
    [sym_compiler_directive_decl] = STATE(1638),
    [sym_fsi_directive_decl] = STATE(1638),
    [sym_preproc_line] = STATE(1638),
    [sym_preproc_if_in_expression] = STATE(5660),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_null] = ACTIONS(1753),
    [anon_sym_begin] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(5743),
    [anon_sym_LT_AT] = ACTIONS(1771),
    [anon_sym_LT_AT_AT] = ACTIONS(5745),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1779),
    [anon_sym_return_BANG] = ACTIONS(1781),
    [anon_sym_yield] = ACTIONS(1779),
    [anon_sym_yield_BANG] = ACTIONS(1781),
    [anon_sym_lazy] = ACTIONS(1779),
    [anon_sym_assert] = ACTIONS(1779),
    [anon_sym_upcast] = ACTIONS(1779),
    [anon_sym_downcast] = ACTIONS(1779),
    [anon_sym_fixed] = ACTIONS(1779),
    [anon_sym_for] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1787),
    [anon_sym_while_BANG] = ACTIONS(1789),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1791),
    [anon_sym_try] = ACTIONS(1793),
    [anon_sym_match] = ACTIONS(1795),
    [anon_sym_match_BANG] = ACTIONS(1797),
    [anon_sym_function] = ACTIONS(1799),
    [anon_sym_use] = ACTIONS(1803),
    [anon_sym_use_BANG] = ACTIONS(1805),
    [anon_sym_do_BANG] = ACTIONS(1807),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1814),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1816),
    [anon_sym_DQUOTE] = ACTIONS(1818),
    [anon_sym_AT_DQUOTE] = ACTIONS(1820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1824),
    [sym_bool] = ACTIONS(1826),
    [sym_unit] = ACTIONS(5747),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1830),
    [sym_op_identifier] = ACTIONS(5749),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1834),
    [sym_xint] = ACTIONS(1836),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1840),
  },
  [STATE(1639)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(287),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1639),
    [sym_compiler_directive_decl] = STATE(1639),
    [sym_fsi_directive_decl] = STATE(1639),
    [sym_preproc_line] = STATE(1639),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1640)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(306),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1640),
    [sym_compiler_directive_decl] = STATE(1640),
    [sym_fsi_directive_decl] = STATE(1640),
    [sym_preproc_line] = STATE(1640),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1641)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(307),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1641),
    [sym_compiler_directive_decl] = STATE(1641),
    [sym_fsi_directive_decl] = STATE(1641),
    [sym_preproc_line] = STATE(1641),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1642)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(296),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1642),
    [sym_compiler_directive_decl] = STATE(1642),
    [sym_fsi_directive_decl] = STATE(1642),
    [sym_preproc_line] = STATE(1642),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1643)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(308),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1643),
    [sym_compiler_directive_decl] = STATE(1643),
    [sym_fsi_directive_decl] = STATE(1643),
    [sym_preproc_line] = STATE(1643),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1644)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(298),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1644),
    [sym_compiler_directive_decl] = STATE(1644),
    [sym_fsi_directive_decl] = STATE(1644),
    [sym_preproc_line] = STATE(1644),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1645)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(302),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1645),
    [sym_compiler_directive_decl] = STATE(1645),
    [sym_fsi_directive_decl] = STATE(1645),
    [sym_preproc_line] = STATE(1645),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1646)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(503),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1646),
    [sym_compiler_directive_decl] = STATE(1646),
    [sym_fsi_directive_decl] = STATE(1646),
    [sym_preproc_line] = STATE(1646),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1647)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(504),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1647),
    [sym_compiler_directive_decl] = STATE(1647),
    [sym_fsi_directive_decl] = STATE(1647),
    [sym_preproc_line] = STATE(1647),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1648)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(31),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1648),
    [sym_compiler_directive_decl] = STATE(1648),
    [sym_fsi_directive_decl] = STATE(1648),
    [sym_preproc_line] = STATE(1648),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1649)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(32),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1649),
    [sym_compiler_directive_decl] = STATE(1649),
    [sym_fsi_directive_decl] = STATE(1649),
    [sym_preproc_line] = STATE(1649),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1650)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(501),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1650),
    [sym_compiler_directive_decl] = STATE(1650),
    [sym_fsi_directive_decl] = STATE(1650),
    [sym_preproc_line] = STATE(1650),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1651)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(301),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1651),
    [sym_compiler_directive_decl] = STATE(1651),
    [sym_fsi_directive_decl] = STATE(1651),
    [sym_preproc_line] = STATE(1651),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1652)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(574),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1652),
    [sym_compiler_directive_decl] = STATE(1652),
    [sym_fsi_directive_decl] = STATE(1652),
    [sym_preproc_line] = STATE(1652),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1653)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(575),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1653),
    [sym_compiler_directive_decl] = STATE(1653),
    [sym_fsi_directive_decl] = STATE(1653),
    [sym_preproc_line] = STATE(1653),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1654)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(564),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1654),
    [sym_compiler_directive_decl] = STATE(1654),
    [sym_fsi_directive_decl] = STATE(1654),
    [sym_preproc_line] = STATE(1654),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1655)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(577),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1655),
    [sym_compiler_directive_decl] = STATE(1655),
    [sym_fsi_directive_decl] = STATE(1655),
    [sym_preproc_line] = STATE(1655),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1656)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(565),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1656),
    [sym_compiler_directive_decl] = STATE(1656),
    [sym_fsi_directive_decl] = STATE(1656),
    [sym_preproc_line] = STATE(1656),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1657)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(548),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1657),
    [sym_compiler_directive_decl] = STATE(1657),
    [sym_fsi_directive_decl] = STATE(1657),
    [sym_preproc_line] = STATE(1657),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1658)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(928),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1658),
    [sym_compiler_directive_decl] = STATE(1658),
    [sym_fsi_directive_decl] = STATE(1658),
    [sym_preproc_line] = STATE(1658),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1659)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(160),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1659),
    [sym_compiler_directive_decl] = STATE(1659),
    [sym_fsi_directive_decl] = STATE(1659),
    [sym_preproc_line] = STATE(1659),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1660)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(161),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1660),
    [sym_compiler_directive_decl] = STATE(1660),
    [sym_fsi_directive_decl] = STATE(1660),
    [sym_preproc_line] = STATE(1660),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1661)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(351),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1661),
    [sym_compiler_directive_decl] = STATE(1661),
    [sym_fsi_directive_decl] = STATE(1661),
    [sym_preproc_line] = STATE(1661),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1662)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(567),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1662),
    [sym_compiler_directive_decl] = STATE(1662),
    [sym_fsi_directive_decl] = STATE(1662),
    [sym_preproc_line] = STATE(1662),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1663)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(594),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1663),
    [sym_compiler_directive_decl] = STATE(1663),
    [sym_fsi_directive_decl] = STATE(1663),
    [sym_preproc_line] = STATE(1663),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1664)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(595),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1664),
    [sym_compiler_directive_decl] = STATE(1664),
    [sym_fsi_directive_decl] = STATE(1664),
    [sym_preproc_line] = STATE(1664),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1665)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(576),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1665),
    [sym_compiler_directive_decl] = STATE(1665),
    [sym_fsi_directive_decl] = STATE(1665),
    [sym_preproc_line] = STATE(1665),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1666)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(597),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1666),
    [sym_compiler_directive_decl] = STATE(1666),
    [sym_fsi_directive_decl] = STATE(1666),
    [sym_preproc_line] = STATE(1666),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1667)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(578),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1667),
    [sym_compiler_directive_decl] = STATE(1667),
    [sym_fsi_directive_decl] = STATE(1667),
    [sym_preproc_line] = STATE(1667),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1668)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(553),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1668),
    [sym_compiler_directive_decl] = STATE(1668),
    [sym_fsi_directive_decl] = STATE(1668),
    [sym_preproc_line] = STATE(1668),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1669)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(81),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1669),
    [sym_compiler_directive_decl] = STATE(1669),
    [sym_fsi_directive_decl] = STATE(1669),
    [sym_preproc_line] = STATE(1669),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1670)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(86),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1670),
    [sym_compiler_directive_decl] = STATE(1670),
    [sym_fsi_directive_decl] = STATE(1670),
    [sym_preproc_line] = STATE(1670),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1671)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(257),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1671),
    [sym_compiler_directive_decl] = STATE(1671),
    [sym_fsi_directive_decl] = STATE(1671),
    [sym_preproc_line] = STATE(1671),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1672)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(580),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1672),
    [sym_compiler_directive_decl] = STATE(1672),
    [sym_fsi_directive_decl] = STATE(1672),
    [sym_preproc_line] = STATE(1672),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1673)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(233),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1673),
    [sym_compiler_directive_decl] = STATE(1673),
    [sym_fsi_directive_decl] = STATE(1673),
    [sym_preproc_line] = STATE(1673),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1674)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(248),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1674),
    [sym_compiler_directive_decl] = STATE(1674),
    [sym_fsi_directive_decl] = STATE(1674),
    [sym_preproc_line] = STATE(1674),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1675)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(249),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1675),
    [sym_compiler_directive_decl] = STATE(1675),
    [sym_fsi_directive_decl] = STATE(1675),
    [sym_preproc_line] = STATE(1675),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1676)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(242),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1676),
    [sym_compiler_directive_decl] = STATE(1676),
    [sym_fsi_directive_decl] = STATE(1676),
    [sym_preproc_line] = STATE(1676),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1677)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(250),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1677),
    [sym_compiler_directive_decl] = STATE(1677),
    [sym_fsi_directive_decl] = STATE(1677),
    [sym_preproc_line] = STATE(1677),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1678)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(243),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1678),
    [sym_compiler_directive_decl] = STATE(1678),
    [sym_fsi_directive_decl] = STATE(1678),
    [sym_preproc_line] = STATE(1678),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1679)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(238),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1679),
    [sym_compiler_directive_decl] = STATE(1679),
    [sym_fsi_directive_decl] = STATE(1679),
    [sym_preproc_line] = STATE(1679),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1680)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(75),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1680),
    [sym_compiler_directive_decl] = STATE(1680),
    [sym_fsi_directive_decl] = STATE(1680),
    [sym_preproc_line] = STATE(1680),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1681)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(76),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1681),
    [sym_compiler_directive_decl] = STATE(1681),
    [sym_fsi_directive_decl] = STATE(1681),
    [sym_preproc_line] = STATE(1681),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1682)] = {
    [sym_function_or_value_defn] = STATE(1335),
    [sym__mutable_expression] = STATE(22038),
    [sym__expression] = STATE(542),
    [sym_literal_expression] = STATE(6410),
    [sym_srtp_call_expression] = STATE(6410),
    [sym__srtp_type_argument] = STATE(22427),
    [sym_long_identifier_or_op] = STATE(5472),
    [sym_tuple_expression] = STATE(6410),
    [sym_brace_expression] = STATE(6410),
    [sym_anon_record_expression] = STATE(6410),
    [sym_prefixed_expression] = STATE(6410),
    [sym_typecast_expression] = STATE(6410),
    [sym_for_expression] = STATE(6410),
    [sym_while_expression] = STATE(6410),
    [sym__if_branch] = STATE(19723),
    [sym_if_expression] = STATE(6410),
    [sym_fun_expression] = STATE(6410),
    [sym_try_expression] = STATE(6410),
    [sym_match_expression] = STATE(6410),
    [sym_function_expression] = STATE(6410),
    [sym_mutate_expression] = STATE(6410),
    [sym_index_expression] = STATE(5472),
    [sym_typed_expression] = STATE(6410),
    [sym_declaration_expression] = STATE(6410),
    [sym_do_expression] = STATE(6410),
    [sym_list_expression] = STATE(6410),
    [sym_array_expression] = STATE(6410),
    [sym_begin_end_expression] = STATE(6410),
    [sym_paren_expression] = STATE(6410),
    [sym__high_prec_app] = STATE(6435),
    [sym__low_prec_app] = STATE(6435),
    [sym_application_expression] = STATE(6410),
    [sym_dot_expression] = STATE(5472),
    [sym_infix_expression] = STATE(6410),
    [sym_ce_expression] = STATE(6410),
    [sym_sequential_expression] = STATE(6410),
    [sym_char] = STATE(6128),
    [sym_format_string] = STATE(6136),
    [sym__string_literal] = STATE(6136),
    [sym_string] = STATE(6128),
    [sym_verbatim_string] = STATE(6128),
    [sym_bytearray] = STATE(6128),
    [sym_verbatim_bytearray] = STATE(6128),
    [sym_format_triple_quoted_string] = STATE(6137),
    [sym_triple_quoted_string] = STATE(6128),
    [sym_const] = STATE(6410),
    [sym_long_identifier] = STATE(5811),
    [sym_active_pattern] = STATE(5765),
    [sym__identifier_or_op] = STATE(5354),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1627),
    [sym_sbyte] = STATE(6128),
    [sym_byte] = STATE(6128),
    [sym_int16] = STATE(6128),
    [sym_uint16] = STATE(6128),
    [sym_int32] = STATE(6128),
    [sym_uint32] = STATE(6128),
    [sym_nativeint] = STATE(6128),
    [sym_unativeint] = STATE(6128),
    [sym_int64] = STATE(6128),
    [sym_uint64] = STATE(6128),
    [sym_ieee32] = STATE(6128),
    [sym_ieee64] = STATE(6128),
    [sym_bignum] = STATE(6128),
    [sym_decimal] = STATE(6128),
    [sym_float] = STATE(5002),
    [sym_block_comment] = STATE(1682),
    [sym_compiler_directive_decl] = STATE(1682),
    [sym_fsi_directive_decl] = STATE(1682),
    [sym_preproc_line] = STATE(1682),
    [sym_preproc_if_in_expression] = STATE(6410),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2984),
    [anon_sym_null] = ACTIONS(2986),
    [anon_sym_begin] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2992),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3000),
    [anon_sym_LBRACE] = ACTIONS(5751),
    [anon_sym_LT_AT] = ACTIONS(3004),
    [anon_sym_LT_AT_AT] = ACTIONS(5753),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3012),
    [anon_sym_new] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_return_BANG] = ACTIONS(3016),
    [anon_sym_yield] = ACTIONS(3014),
    [anon_sym_yield_BANG] = ACTIONS(3016),
    [anon_sym_lazy] = ACTIONS(3014),
    [anon_sym_assert] = ACTIONS(3014),
    [anon_sym_upcast] = ACTIONS(3014),
    [anon_sym_downcast] = ACTIONS(3014),
    [anon_sym_fixed] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3020),
    [anon_sym_while] = ACTIONS(3022),
    [anon_sym_while_BANG] = ACTIONS(3024),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3026),
    [anon_sym_try] = ACTIONS(3028),
    [anon_sym_match] = ACTIONS(3030),
    [anon_sym_match_BANG] = ACTIONS(3032),
    [anon_sym_function] = ACTIONS(3034),
    [anon_sym_use] = ACTIONS(3038),
    [anon_sym_use_BANG] = ACTIONS(3040),
    [anon_sym_do_BANG] = ACTIONS(3042),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3048),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3050),
    [anon_sym_DQUOTE] = ACTIONS(3052),
    [anon_sym_AT_DQUOTE] = ACTIONS(3054),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3056),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3058),
    [sym_bool] = ACTIONS(3060),
    [sym_unit] = ACTIONS(5755),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3064),
    [sym_op_identifier] = ACTIONS(5757),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3068),
    [sym_xint] = ACTIONS(3070),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3072),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3076),
  },
  [STATE(1683)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(244),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1683),
    [sym_compiler_directive_decl] = STATE(1683),
    [sym_fsi_directive_decl] = STATE(1683),
    [sym_preproc_line] = STATE(1683),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1684)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(365),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1684),
    [sym_compiler_directive_decl] = STATE(1684),
    [sym_fsi_directive_decl] = STATE(1684),
    [sym_preproc_line] = STATE(1684),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1685)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(366),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1685),
    [sym_compiler_directive_decl] = STATE(1685),
    [sym_fsi_directive_decl] = STATE(1685),
    [sym_preproc_line] = STATE(1685),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1686)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(465),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1686),
    [sym_compiler_directive_decl] = STATE(1686),
    [sym_fsi_directive_decl] = STATE(1686),
    [sym_preproc_line] = STATE(1686),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1687)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(367),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1687),
    [sym_compiler_directive_decl] = STATE(1687),
    [sym_fsi_directive_decl] = STATE(1687),
    [sym_preproc_line] = STATE(1687),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1688)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(360),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1688),
    [sym_compiler_directive_decl] = STATE(1688),
    [sym_fsi_directive_decl] = STATE(1688),
    [sym_preproc_line] = STATE(1688),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1689)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(463),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1689),
    [sym_compiler_directive_decl] = STATE(1689),
    [sym_fsi_directive_decl] = STATE(1689),
    [sym_preproc_line] = STATE(1689),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1690)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(514),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1690),
    [sym_compiler_directive_decl] = STATE(1690),
    [sym_fsi_directive_decl] = STATE(1690),
    [sym_preproc_line] = STATE(1690),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1691)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(505),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1691),
    [sym_compiler_directive_decl] = STATE(1691),
    [sym_fsi_directive_decl] = STATE(1691),
    [sym_preproc_line] = STATE(1691),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1692)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(134),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1692),
    [sym_compiler_directive_decl] = STATE(1692),
    [sym_fsi_directive_decl] = STATE(1692),
    [sym_preproc_line] = STATE(1692),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1693)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(135),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1693),
    [sym_compiler_directive_decl] = STATE(1693),
    [sym_fsi_directive_decl] = STATE(1693),
    [sym_preproc_line] = STATE(1693),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1694)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(292),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1694),
    [sym_compiler_directive_decl] = STATE(1694),
    [sym_fsi_directive_decl] = STATE(1694),
    [sym_preproc_line] = STATE(1694),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1695)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(361),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1695),
    [sym_compiler_directive_decl] = STATE(1695),
    [sym_fsi_directive_decl] = STATE(1695),
    [sym_preproc_line] = STATE(1695),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1696)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(171),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1696),
    [sym_compiler_directive_decl] = STATE(1696),
    [sym_fsi_directive_decl] = STATE(1696),
    [sym_preproc_line] = STATE(1696),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1697)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(172),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1697),
    [sym_compiler_directive_decl] = STATE(1697),
    [sym_fsi_directive_decl] = STATE(1697),
    [sym_preproc_line] = STATE(1697),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1698)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(190),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1698),
    [sym_compiler_directive_decl] = STATE(1698),
    [sym_fsi_directive_decl] = STATE(1698),
    [sym_preproc_line] = STATE(1698),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1699)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(173),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1699),
    [sym_compiler_directive_decl] = STATE(1699),
    [sym_fsi_directive_decl] = STATE(1699),
    [sym_preproc_line] = STATE(1699),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1700)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(174),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1700),
    [sym_compiler_directive_decl] = STATE(1700),
    [sym_fsi_directive_decl] = STATE(1700),
    [sym_preproc_line] = STATE(1700),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1701)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(188),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1701),
    [sym_compiler_directive_decl] = STATE(1701),
    [sym_fsi_directive_decl] = STATE(1701),
    [sym_preproc_line] = STATE(1701),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1702)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(516),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1702),
    [sym_compiler_directive_decl] = STATE(1702),
    [sym_fsi_directive_decl] = STATE(1702),
    [sym_preproc_line] = STATE(1702),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1703)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(517),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1703),
    [sym_compiler_directive_decl] = STATE(1703),
    [sym_fsi_directive_decl] = STATE(1703),
    [sym_preproc_line] = STATE(1703),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1704)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(124),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1704),
    [sym_compiler_directive_decl] = STATE(1704),
    [sym_fsi_directive_decl] = STATE(1704),
    [sym_preproc_line] = STATE(1704),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1705)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(131),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1705),
    [sym_compiler_directive_decl] = STATE(1705),
    [sym_fsi_directive_decl] = STATE(1705),
    [sym_preproc_line] = STATE(1705),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1706)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(304),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1706),
    [sym_compiler_directive_decl] = STATE(1706),
    [sym_fsi_directive_decl] = STATE(1706),
    [sym_preproc_line] = STATE(1706),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1707)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(165),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1707),
    [sym_compiler_directive_decl] = STATE(1707),
    [sym_fsi_directive_decl] = STATE(1707),
    [sym_preproc_line] = STATE(1707),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1708)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(218),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1708),
    [sym_compiler_directive_decl] = STATE(1708),
    [sym_fsi_directive_decl] = STATE(1708),
    [sym_preproc_line] = STATE(1708),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1709)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(219),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1709),
    [sym_compiler_directive_decl] = STATE(1709),
    [sym_fsi_directive_decl] = STATE(1709),
    [sym_preproc_line] = STATE(1709),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1710)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(203),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1710),
    [sym_compiler_directive_decl] = STATE(1710),
    [sym_fsi_directive_decl] = STATE(1710),
    [sym_preproc_line] = STATE(1710),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1711)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(220),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1711),
    [sym_compiler_directive_decl] = STATE(1711),
    [sym_fsi_directive_decl] = STATE(1711),
    [sym_preproc_line] = STATE(1711),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1712)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(221),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1712),
    [sym_compiler_directive_decl] = STATE(1712),
    [sym_fsi_directive_decl] = STATE(1712),
    [sym_preproc_line] = STATE(1712),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1713)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(196),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1713),
    [sym_compiler_directive_decl] = STATE(1713),
    [sym_fsi_directive_decl] = STATE(1713),
    [sym_preproc_line] = STATE(1713),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1714)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(2),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1714),
    [sym_compiler_directive_decl] = STATE(1714),
    [sym_fsi_directive_decl] = STATE(1714),
    [sym_preproc_line] = STATE(1714),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1715)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(4),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1715),
    [sym_compiler_directive_decl] = STATE(1715),
    [sym_fsi_directive_decl] = STATE(1715),
    [sym_preproc_line] = STATE(1715),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1716)] = {
    [sym_function_or_value_defn] = STATE(1325),
    [sym__mutable_expression] = STATE(19434),
    [sym__expression] = STATE(391),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22454),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1617),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1716),
    [sym_compiler_directive_decl] = STATE(1716),
    [sym_fsi_directive_decl] = STATE(1716),
    [sym_preproc_line] = STATE(1716),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(2172),
    [anon_sym_return] = ACTIONS(2172),
    [anon_sym_return_BANG] = ACTIONS(2174),
    [anon_sym_yield] = ACTIONS(2172),
    [anon_sym_yield_BANG] = ACTIONS(2174),
    [anon_sym_lazy] = ACTIONS(2172),
    [anon_sym_assert] = ACTIONS(2172),
    [anon_sym_upcast] = ACTIONS(2172),
    [anon_sym_downcast] = ACTIONS(2172),
    [anon_sym_fixed] = ACTIONS(2172),
    [anon_sym_for] = ACTIONS(2176),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(2178),
    [anon_sym_use_BANG] = ACTIONS(2180),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2186),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1717)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(212),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1717),
    [sym_compiler_directive_decl] = STATE(1717),
    [sym_fsi_directive_decl] = STATE(1717),
    [sym_preproc_line] = STATE(1717),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1718)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(63),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1718),
    [sym_compiler_directive_decl] = STATE(1718),
    [sym_fsi_directive_decl] = STATE(1718),
    [sym_preproc_line] = STATE(1718),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1719)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(64),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1719),
    [sym_compiler_directive_decl] = STATE(1719),
    [sym_fsi_directive_decl] = STATE(1719),
    [sym_preproc_line] = STATE(1719),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1720)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(44),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1720),
    [sym_compiler_directive_decl] = STATE(1720),
    [sym_fsi_directive_decl] = STATE(1720),
    [sym_preproc_line] = STATE(1720),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1721)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(65),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1721),
    [sym_compiler_directive_decl] = STATE(1721),
    [sym_fsi_directive_decl] = STATE(1721),
    [sym_preproc_line] = STATE(1721),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1722)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(35),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1722),
    [sym_compiler_directive_decl] = STATE(1722),
    [sym_fsi_directive_decl] = STATE(1722),
    [sym_preproc_line] = STATE(1722),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1723)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(42),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1723),
    [sym_compiler_directive_decl] = STATE(1723),
    [sym_fsi_directive_decl] = STATE(1723),
    [sym_preproc_line] = STATE(1723),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1724)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(139),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1724),
    [sym_compiler_directive_decl] = STATE(1724),
    [sym_fsi_directive_decl] = STATE(1724),
    [sym_preproc_line] = STATE(1724),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1725)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(119),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1725),
    [sym_compiler_directive_decl] = STATE(1725),
    [sym_fsi_directive_decl] = STATE(1725),
    [sym_preproc_line] = STATE(1725),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1726)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(572),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1726),
    [sym_compiler_directive_decl] = STATE(1726),
    [sym_fsi_directive_decl] = STATE(1726),
    [sym_preproc_line] = STATE(1726),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1727)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(141),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1727),
    [sym_compiler_directive_decl] = STATE(1727),
    [sym_fsi_directive_decl] = STATE(1727),
    [sym_preproc_line] = STATE(1727),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1728)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(27),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1728),
    [sym_compiler_directive_decl] = STATE(1728),
    [sym_fsi_directive_decl] = STATE(1728),
    [sym_preproc_line] = STATE(1728),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1729)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(28),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1729),
    [sym_compiler_directive_decl] = STATE(1729),
    [sym_fsi_directive_decl] = STATE(1729),
    [sym_preproc_line] = STATE(1729),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1730)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(8),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1730),
    [sym_compiler_directive_decl] = STATE(1730),
    [sym_fsi_directive_decl] = STATE(1730),
    [sym_preproc_line] = STATE(1730),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1731)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(29),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1731),
    [sym_compiler_directive_decl] = STATE(1731),
    [sym_fsi_directive_decl] = STATE(1731),
    [sym_preproc_line] = STATE(1731),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1732)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(30),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1732),
    [sym_compiler_directive_decl] = STATE(1732),
    [sym_fsi_directive_decl] = STATE(1732),
    [sym_preproc_line] = STATE(1732),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1733)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(33),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1733),
    [sym_compiler_directive_decl] = STATE(1733),
    [sym_fsi_directive_decl] = STATE(1733),
    [sym_preproc_line] = STATE(1733),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1734)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(506),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1734),
    [sym_compiler_directive_decl] = STATE(1734),
    [sym_fsi_directive_decl] = STATE(1734),
    [sym_preproc_line] = STATE(1734),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1735)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(518),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1735),
    [sym_compiler_directive_decl] = STATE(1735),
    [sym_fsi_directive_decl] = STATE(1735),
    [sym_preproc_line] = STATE(1735),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1736)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(5),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1736),
    [sym_compiler_directive_decl] = STATE(1736),
    [sym_fsi_directive_decl] = STATE(1736),
    [sym_preproc_line] = STATE(1736),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1737)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(6),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1737),
    [sym_compiler_directive_decl] = STATE(1737),
    [sym_fsi_directive_decl] = STATE(1737),
    [sym_preproc_line] = STATE(1737),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1738)] = {
    [sym_function_or_value_defn] = STATE(1363),
    [sym__mutable_expression] = STATE(19899),
    [sym__expression] = STATE(592),
    [sym_literal_expression] = STATE(6185),
    [sym_srtp_call_expression] = STATE(6185),
    [sym__srtp_type_argument] = STATE(22472),
    [sym_long_identifier_or_op] = STATE(5752),
    [sym_tuple_expression] = STATE(6185),
    [sym_brace_expression] = STATE(6185),
    [sym_anon_record_expression] = STATE(6185),
    [sym_prefixed_expression] = STATE(6185),
    [sym_typecast_expression] = STATE(6185),
    [sym_for_expression] = STATE(6185),
    [sym_while_expression] = STATE(6185),
    [sym__if_branch] = STATE(19960),
    [sym_if_expression] = STATE(6185),
    [sym_fun_expression] = STATE(6185),
    [sym_try_expression] = STATE(6185),
    [sym_match_expression] = STATE(6185),
    [sym_function_expression] = STATE(6185),
    [sym_mutate_expression] = STATE(6185),
    [sym_index_expression] = STATE(5752),
    [sym_typed_expression] = STATE(6185),
    [sym_declaration_expression] = STATE(6185),
    [sym_do_expression] = STATE(6185),
    [sym_list_expression] = STATE(6185),
    [sym_array_expression] = STATE(6185),
    [sym_begin_end_expression] = STATE(6185),
    [sym_paren_expression] = STATE(6185),
    [sym__high_prec_app] = STATE(6190),
    [sym__low_prec_app] = STATE(6190),
    [sym_application_expression] = STATE(6185),
    [sym_dot_expression] = STATE(5752),
    [sym_infix_expression] = STATE(6185),
    [sym_ce_expression] = STATE(6185),
    [sym_sequential_expression] = STATE(6185),
    [sym_char] = STATE(5981),
    [sym_format_string] = STATE(5986),
    [sym__string_literal] = STATE(5986),
    [sym_string] = STATE(5981),
    [sym_verbatim_string] = STATE(5981),
    [sym_bytearray] = STATE(5981),
    [sym_verbatim_bytearray] = STATE(5981),
    [sym_format_triple_quoted_string] = STATE(5988),
    [sym_triple_quoted_string] = STATE(5981),
    [sym_const] = STATE(6185),
    [sym_long_identifier] = STATE(5340),
    [sym_active_pattern] = STATE(5721),
    [sym__identifier_or_op] = STATE(5599),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1672),
    [sym_sbyte] = STATE(5981),
    [sym_byte] = STATE(5981),
    [sym_int16] = STATE(5981),
    [sym_uint16] = STATE(5981),
    [sym_int32] = STATE(5981),
    [sym_uint32] = STATE(5981),
    [sym_nativeint] = STATE(5981),
    [sym_unativeint] = STATE(5981),
    [sym_int64] = STATE(5981),
    [sym_uint64] = STATE(5981),
    [sym_ieee32] = STATE(5981),
    [sym_ieee64] = STATE(5981),
    [sym_bignum] = STATE(5981),
    [sym_decimal] = STATE(5981),
    [sym_float] = STATE(5284),
    [sym_block_comment] = STATE(1738),
    [sym_compiler_directive_decl] = STATE(1738),
    [sym_fsi_directive_decl] = STATE(1738),
    [sym_preproc_line] = STATE(1738),
    [sym_preproc_if_in_expression] = STATE(6185),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3236),
    [anon_sym_null] = ACTIONS(3238),
    [anon_sym_begin] = ACTIONS(3240),
    [anon_sym_do] = ACTIONS(3244),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3246),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3250),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3252),
    [anon_sym_LBRACE] = ACTIONS(5103),
    [anon_sym_LT_AT] = ACTIONS(3256),
    [anon_sym_LT_AT_AT] = ACTIONS(5105),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3264),
    [anon_sym_new] = ACTIONS(3266),
    [anon_sym_return] = ACTIONS(3266),
    [anon_sym_return_BANG] = ACTIONS(3268),
    [anon_sym_yield] = ACTIONS(3266),
    [anon_sym_yield_BANG] = ACTIONS(3268),
    [anon_sym_lazy] = ACTIONS(3266),
    [anon_sym_assert] = ACTIONS(3266),
    [anon_sym_upcast] = ACTIONS(3266),
    [anon_sym_downcast] = ACTIONS(3266),
    [anon_sym_fixed] = ACTIONS(3266),
    [anon_sym_for] = ACTIONS(3272),
    [anon_sym_while] = ACTIONS(3274),
    [anon_sym_while_BANG] = ACTIONS(3276),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3278),
    [anon_sym_try] = ACTIONS(3280),
    [anon_sym_match] = ACTIONS(3282),
    [anon_sym_match_BANG] = ACTIONS(3284),
    [anon_sym_function] = ACTIONS(3286),
    [anon_sym_use] = ACTIONS(3290),
    [anon_sym_use_BANG] = ACTIONS(3292),
    [anon_sym_do_BANG] = ACTIONS(3294),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3300),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3302),
    [anon_sym_DQUOTE] = ACTIONS(3304),
    [anon_sym_AT_DQUOTE] = ACTIONS(3306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3308),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3310),
    [sym_bool] = ACTIONS(3312),
    [sym_unit] = ACTIONS(5111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3316),
    [sym_op_identifier] = ACTIONS(5113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3320),
    [sym_xint] = ACTIONS(3322),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3324),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3328),
  },
  [STATE(1739)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(45),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1739),
    [sym_compiler_directive_decl] = STATE(1739),
    [sym_fsi_directive_decl] = STATE(1739),
    [sym_preproc_line] = STATE(1739),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1740)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(79),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1740),
    [sym_compiler_directive_decl] = STATE(1740),
    [sym_fsi_directive_decl] = STATE(1740),
    [sym_preproc_line] = STATE(1740),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1741)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(80),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1741),
    [sym_compiler_directive_decl] = STATE(1741),
    [sym_fsi_directive_decl] = STATE(1741),
    [sym_preproc_line] = STATE(1741),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1742)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(423),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1742),
    [sym_compiler_directive_decl] = STATE(1742),
    [sym_fsi_directive_decl] = STATE(1742),
    [sym_preproc_line] = STATE(1742),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1743)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(84),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1743),
    [sym_compiler_directive_decl] = STATE(1743),
    [sym_fsi_directive_decl] = STATE(1743),
    [sym_preproc_line] = STATE(1743),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1744)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(85),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1744),
    [sym_compiler_directive_decl] = STATE(1744),
    [sym_fsi_directive_decl] = STATE(1744),
    [sym_preproc_line] = STATE(1744),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1745)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(66),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1745),
    [sym_compiler_directive_decl] = STATE(1745),
    [sym_fsi_directive_decl] = STATE(1745),
    [sym_preproc_line] = STATE(1745),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1746)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(37),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1746),
    [sym_compiler_directive_decl] = STATE(1746),
    [sym_fsi_directive_decl] = STATE(1746),
    [sym_preproc_line] = STATE(1746),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1747)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(39),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1747),
    [sym_compiler_directive_decl] = STATE(1747),
    [sym_fsi_directive_decl] = STATE(1747),
    [sym_preproc_line] = STATE(1747),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1748)] = {
    [sym_function_or_value_defn] = STATE(1374),
    [sym__mutable_expression] = STATE(20131),
    [sym__expression] = STATE(246),
    [sym_literal_expression] = STATE(4884),
    [sym_srtp_call_expression] = STATE(4884),
    [sym__srtp_type_argument] = STATE(22480),
    [sym_long_identifier_or_op] = STATE(4682),
    [sym_tuple_expression] = STATE(4884),
    [sym_brace_expression] = STATE(4884),
    [sym_anon_record_expression] = STATE(4884),
    [sym_prefixed_expression] = STATE(4884),
    [sym_typecast_expression] = STATE(4884),
    [sym_for_expression] = STATE(4884),
    [sym_while_expression] = STATE(4884),
    [sym__if_branch] = STATE(20004),
    [sym_if_expression] = STATE(4884),
    [sym_fun_expression] = STATE(4884),
    [sym_try_expression] = STATE(4884),
    [sym_match_expression] = STATE(4884),
    [sym_function_expression] = STATE(4884),
    [sym_mutate_expression] = STATE(4884),
    [sym_index_expression] = STATE(4682),
    [sym_typed_expression] = STATE(4884),
    [sym_declaration_expression] = STATE(4884),
    [sym_do_expression] = STATE(4884),
    [sym_list_expression] = STATE(4884),
    [sym_array_expression] = STATE(4884),
    [sym_begin_end_expression] = STATE(4884),
    [sym_paren_expression] = STATE(4884),
    [sym__high_prec_app] = STATE(4887),
    [sym__low_prec_app] = STATE(4887),
    [sym_application_expression] = STATE(4884),
    [sym_dot_expression] = STATE(4682),
    [sym_infix_expression] = STATE(4884),
    [sym_ce_expression] = STATE(4884),
    [sym_sequential_expression] = STATE(4884),
    [sym_char] = STATE(4773),
    [sym_format_string] = STATE(4774),
    [sym__string_literal] = STATE(4774),
    [sym_string] = STATE(4773),
    [sym_verbatim_string] = STATE(4773),
    [sym_bytearray] = STATE(4773),
    [sym_verbatim_bytearray] = STATE(4773),
    [sym_format_triple_quoted_string] = STATE(4775),
    [sym_triple_quoted_string] = STATE(4773),
    [sym_const] = STATE(4884),
    [sym_long_identifier] = STATE(4630),
    [sym_active_pattern] = STATE(4625),
    [sym__identifier_or_op] = STATE(4576),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1683),
    [sym_sbyte] = STATE(4773),
    [sym_byte] = STATE(4773),
    [sym_int16] = STATE(4773),
    [sym_uint16] = STATE(4773),
    [sym_int32] = STATE(4773),
    [sym_uint32] = STATE(4773),
    [sym_nativeint] = STATE(4773),
    [sym_unativeint] = STATE(4773),
    [sym_int64] = STATE(4773),
    [sym_uint64] = STATE(4773),
    [sym_ieee32] = STATE(4773),
    [sym_ieee64] = STATE(4773),
    [sym_bignum] = STATE(4773),
    [sym_decimal] = STATE(4773),
    [sym_float] = STATE(4411),
    [sym_block_comment] = STATE(1748),
    [sym_compiler_directive_decl] = STATE(1748),
    [sym_fsi_directive_decl] = STATE(1748),
    [sym_preproc_line] = STATE(1748),
    [sym_preproc_if_in_expression] = STATE(4884),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_begin] = ACTIONS(1519),
    [anon_sym_do] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1527),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LBRACE] = ACTIONS(5733),
    [anon_sym_LT_AT] = ACTIONS(1537),
    [anon_sym_LT_AT_AT] = ACTIONS(5735),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1545),
    [anon_sym_return] = ACTIONS(1545),
    [anon_sym_return_BANG] = ACTIONS(1549),
    [anon_sym_yield] = ACTIONS(1545),
    [anon_sym_yield_BANG] = ACTIONS(1549),
    [anon_sym_lazy] = ACTIONS(1545),
    [anon_sym_assert] = ACTIONS(1545),
    [anon_sym_upcast] = ACTIONS(1545),
    [anon_sym_downcast] = ACTIONS(1545),
    [anon_sym_fixed] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1553),
    [anon_sym_while] = ACTIONS(1555),
    [anon_sym_while_BANG] = ACTIONS(1557),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1559),
    [anon_sym_try] = ACTIONS(1561),
    [anon_sym_match] = ACTIONS(1563),
    [anon_sym_match_BANG] = ACTIONS(1565),
    [anon_sym_function] = ACTIONS(1567),
    [anon_sym_use] = ACTIONS(1571),
    [anon_sym_use_BANG] = ACTIONS(1573),
    [anon_sym_do_BANG] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1582),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1586),
    [anon_sym_AT_DQUOTE] = ACTIONS(1588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1590),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1592),
    [sym_bool] = ACTIONS(1594),
    [sym_unit] = ACTIONS(5737),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1598),
    [sym_op_identifier] = ACTIONS(5739),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1602),
    [sym_xint] = ACTIONS(1604),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1606),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1608),
  },
  [STATE(1749)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(21),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1749),
    [sym_compiler_directive_decl] = STATE(1749),
    [sym_fsi_directive_decl] = STATE(1749),
    [sym_preproc_line] = STATE(1749),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1750)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(71),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1750),
    [sym_compiler_directive_decl] = STATE(1750),
    [sym_fsi_directive_decl] = STATE(1750),
    [sym_preproc_line] = STATE(1750),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1751)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(72),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1751),
    [sym_compiler_directive_decl] = STATE(1751),
    [sym_fsi_directive_decl] = STATE(1751),
    [sym_preproc_line] = STATE(1751),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1752)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(83),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1752),
    [sym_compiler_directive_decl] = STATE(1752),
    [sym_fsi_directive_decl] = STATE(1752),
    [sym_preproc_line] = STATE(1752),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1753)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(73),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1753),
    [sym_compiler_directive_decl] = STATE(1753),
    [sym_fsi_directive_decl] = STATE(1753),
    [sym_preproc_line] = STATE(1753),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1754)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(74),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1754),
    [sym_compiler_directive_decl] = STATE(1754),
    [sym_fsi_directive_decl] = STATE(1754),
    [sym_preproc_line] = STATE(1754),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1755)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(90),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1755),
    [sym_compiler_directive_decl] = STATE(1755),
    [sym_fsi_directive_decl] = STATE(1755),
    [sym_preproc_line] = STATE(1755),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1756)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(40),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1756),
    [sym_compiler_directive_decl] = STATE(1756),
    [sym_fsi_directive_decl] = STATE(1756),
    [sym_preproc_line] = STATE(1756),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1757)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(41),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1757),
    [sym_compiler_directive_decl] = STATE(1757),
    [sym_fsi_directive_decl] = STATE(1757),
    [sym_preproc_line] = STATE(1757),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1758)] = {
    [sym_function_or_value_defn] = STATE(1289),
    [sym__mutable_expression] = STATE(20403),
    [sym__expression] = STATE(363),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22489),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1695),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1758),
    [sym_compiler_directive_decl] = STATE(1758),
    [sym_fsi_directive_decl] = STATE(1758),
    [sym_preproc_line] = STATE(1758),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(2309),
    [anon_sym_return] = ACTIONS(2309),
    [anon_sym_return_BANG] = ACTIONS(2311),
    [anon_sym_yield] = ACTIONS(2309),
    [anon_sym_yield_BANG] = ACTIONS(2311),
    [anon_sym_lazy] = ACTIONS(2309),
    [anon_sym_assert] = ACTIONS(2309),
    [anon_sym_upcast] = ACTIONS(2309),
    [anon_sym_downcast] = ACTIONS(2309),
    [anon_sym_fixed] = ACTIONS(2309),
    [anon_sym_for] = ACTIONS(2315),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(2333),
    [anon_sym_use_BANG] = ACTIONS(2335),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2363),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1759)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(92),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1759),
    [sym_compiler_directive_decl] = STATE(1759),
    [sym_fsi_directive_decl] = STATE(1759),
    [sym_preproc_line] = STATE(1759),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1760)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(234),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1760),
    [sym_compiler_directive_decl] = STATE(1760),
    [sym_fsi_directive_decl] = STATE(1760),
    [sym_preproc_line] = STATE(1760),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1761)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(140),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1761),
    [sym_compiler_directive_decl] = STATE(1761),
    [sym_fsi_directive_decl] = STATE(1761),
    [sym_preproc_line] = STATE(1761),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1762)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(117),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1762),
    [sym_compiler_directive_decl] = STATE(1762),
    [sym_fsi_directive_decl] = STATE(1762),
    [sym_preproc_line] = STATE(1762),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1763)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(138),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1763),
    [sym_compiler_directive_decl] = STATE(1763),
    [sym_fsi_directive_decl] = STATE(1763),
    [sym_preproc_line] = STATE(1763),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1764)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(96),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1764),
    [sym_compiler_directive_decl] = STATE(1764),
    [sym_fsi_directive_decl] = STATE(1764),
    [sym_preproc_line] = STATE(1764),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1765)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(120),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1765),
    [sym_compiler_directive_decl] = STATE(1765),
    [sym_fsi_directive_decl] = STATE(1765),
    [sym_preproc_line] = STATE(1765),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1766)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(116),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1766),
    [sym_compiler_directive_decl] = STATE(1766),
    [sym_fsi_directive_decl] = STATE(1766),
    [sym_preproc_line] = STATE(1766),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1767)] = {
    [sym_function_or_value_defn] = STATE(1292),
    [sym__mutable_expression] = STATE(20702),
    [sym__expression] = STATE(169),
    [sym_literal_expression] = STATE(4007),
    [sym_srtp_call_expression] = STATE(4007),
    [sym__srtp_type_argument] = STATE(22498),
    [sym_long_identifier_or_op] = STATE(3833),
    [sym_tuple_expression] = STATE(4007),
    [sym_brace_expression] = STATE(4007),
    [sym_anon_record_expression] = STATE(4007),
    [sym_prefixed_expression] = STATE(4007),
    [sym_typecast_expression] = STATE(4007),
    [sym_for_expression] = STATE(4007),
    [sym_while_expression] = STATE(4007),
    [sym__if_branch] = STATE(20047),
    [sym_if_expression] = STATE(4007),
    [sym_fun_expression] = STATE(4007),
    [sym_try_expression] = STATE(4007),
    [sym_match_expression] = STATE(4007),
    [sym_function_expression] = STATE(4007),
    [sym_mutate_expression] = STATE(4007),
    [sym_index_expression] = STATE(3833),
    [sym_typed_expression] = STATE(4007),
    [sym_declaration_expression] = STATE(4007),
    [sym_do_expression] = STATE(4007),
    [sym_list_expression] = STATE(4007),
    [sym_array_expression] = STATE(4007),
    [sym_begin_end_expression] = STATE(4007),
    [sym_paren_expression] = STATE(4007),
    [sym__high_prec_app] = STATE(4035),
    [sym__low_prec_app] = STATE(4035),
    [sym_application_expression] = STATE(4007),
    [sym_dot_expression] = STATE(3833),
    [sym_infix_expression] = STATE(4007),
    [sym_ce_expression] = STATE(4007),
    [sym_sequential_expression] = STATE(4007),
    [sym_char] = STATE(3942),
    [sym_format_string] = STATE(4086),
    [sym__string_literal] = STATE(4086),
    [sym_string] = STATE(3942),
    [sym_verbatim_string] = STATE(3942),
    [sym_bytearray] = STATE(3942),
    [sym_verbatim_bytearray] = STATE(3942),
    [sym_format_triple_quoted_string] = STATE(4091),
    [sym_triple_quoted_string] = STATE(3942),
    [sym_const] = STATE(4007),
    [sym_long_identifier] = STATE(3782),
    [sym_active_pattern] = STATE(3845),
    [sym__identifier_or_op] = STATE(3789),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1707),
    [sym_sbyte] = STATE(3942),
    [sym_byte] = STATE(3942),
    [sym_int16] = STATE(3942),
    [sym_uint16] = STATE(3942),
    [sym_int32] = STATE(3942),
    [sym_uint32] = STATE(3942),
    [sym_nativeint] = STATE(3942),
    [sym_unativeint] = STATE(3942),
    [sym_int64] = STATE(3942),
    [sym_uint64] = STATE(3942),
    [sym_ieee32] = STATE(3942),
    [sym_ieee64] = STATE(3942),
    [sym_bignum] = STATE(3942),
    [sym_decimal] = STATE(3942),
    [sym_float] = STATE(3768),
    [sym_block_comment] = STATE(1767),
    [sym_compiler_directive_decl] = STATE(1767),
    [sym_fsi_directive_decl] = STATE(1767),
    [sym_preproc_line] = STATE(1767),
    [sym_preproc_if_in_expression] = STATE(4007),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1155),
    [anon_sym_null] = ACTIONS(1157),
    [anon_sym_begin] = ACTIONS(1159),
    [anon_sym_do] = ACTIONS(1163),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1169),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1171),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_LT_AT] = ACTIONS(1175),
    [anon_sym_LT_AT_AT] = ACTIONS(2127),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1183),
    [anon_sym_new] = ACTIONS(1185),
    [anon_sym_return] = ACTIONS(1185),
    [anon_sym_return_BANG] = ACTIONS(1187),
    [anon_sym_yield] = ACTIONS(1185),
    [anon_sym_yield_BANG] = ACTIONS(1187),
    [anon_sym_lazy] = ACTIONS(1185),
    [anon_sym_assert] = ACTIONS(1185),
    [anon_sym_upcast] = ACTIONS(1185),
    [anon_sym_downcast] = ACTIONS(1185),
    [anon_sym_fixed] = ACTIONS(1185),
    [anon_sym_for] = ACTIONS(1191),
    [anon_sym_while] = ACTIONS(1193),
    [anon_sym_while_BANG] = ACTIONS(1195),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1197),
    [anon_sym_try] = ACTIONS(1199),
    [anon_sym_match] = ACTIONS(1201),
    [anon_sym_match_BANG] = ACTIONS(1203),
    [anon_sym_function] = ACTIONS(1205),
    [anon_sym_use] = ACTIONS(1209),
    [anon_sym_use_BANG] = ACTIONS(1211),
    [anon_sym_do_BANG] = ACTIONS(1213),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1219),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1223),
    [anon_sym_AT_DQUOTE] = ACTIONS(1225),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1227),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1229),
    [sym_bool] = ACTIONS(1231),
    [sym_unit] = ACTIONS(2137),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1235),
    [sym_op_identifier] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1239),
    [sym_xint] = ACTIONS(1241),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1243),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1247),
  },
  [STATE(1768)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(87),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1768),
    [sym_compiler_directive_decl] = STATE(1768),
    [sym_fsi_directive_decl] = STATE(1768),
    [sym_preproc_line] = STATE(1768),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1769)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(129),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1769),
    [sym_compiler_directive_decl] = STATE(1769),
    [sym_fsi_directive_decl] = STATE(1769),
    [sym_preproc_line] = STATE(1769),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1770)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(130),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1770),
    [sym_compiler_directive_decl] = STATE(1770),
    [sym_fsi_directive_decl] = STATE(1770),
    [sym_preproc_line] = STATE(1770),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1771)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(118),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1771),
    [sym_compiler_directive_decl] = STATE(1771),
    [sym_fsi_directive_decl] = STATE(1771),
    [sym_preproc_line] = STATE(1771),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1772)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(132),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1772),
    [sym_compiler_directive_decl] = STATE(1772),
    [sym_fsi_directive_decl] = STATE(1772),
    [sym_preproc_line] = STATE(1772),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1773)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(133),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1773),
    [sym_compiler_directive_decl] = STATE(1773),
    [sym_fsi_directive_decl] = STATE(1773),
    [sym_preproc_line] = STATE(1773),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1774)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(136),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1774),
    [sym_compiler_directive_decl] = STATE(1774),
    [sym_fsi_directive_decl] = STATE(1774),
    [sym_preproc_line] = STATE(1774),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1775)] = {
    [sym_function_or_value_defn] = STATE(1331),
    [sym__mutable_expression] = STATE(19873),
    [sym__expression] = STATE(507),
    [sym_literal_expression] = STATE(5979),
    [sym_srtp_call_expression] = STATE(5979),
    [sym__srtp_type_argument] = STATE(22398),
    [sym_long_identifier_or_op] = STATE(5498),
    [sym_tuple_expression] = STATE(5979),
    [sym_brace_expression] = STATE(5979),
    [sym_anon_record_expression] = STATE(5979),
    [sym_prefixed_expression] = STATE(5979),
    [sym_typecast_expression] = STATE(5979),
    [sym_for_expression] = STATE(5979),
    [sym_while_expression] = STATE(5979),
    [sym__if_branch] = STATE(19497),
    [sym_if_expression] = STATE(5979),
    [sym_fun_expression] = STATE(5979),
    [sym_try_expression] = STATE(5979),
    [sym_match_expression] = STATE(5979),
    [sym_function_expression] = STATE(5979),
    [sym_mutate_expression] = STATE(5979),
    [sym_index_expression] = STATE(5498),
    [sym_typed_expression] = STATE(5979),
    [sym_declaration_expression] = STATE(5979),
    [sym_do_expression] = STATE(5979),
    [sym_list_expression] = STATE(5979),
    [sym_array_expression] = STATE(5979),
    [sym_begin_end_expression] = STATE(5979),
    [sym_paren_expression] = STATE(5979),
    [sym__high_prec_app] = STATE(5985),
    [sym__low_prec_app] = STATE(5985),
    [sym_application_expression] = STATE(5979),
    [sym_dot_expression] = STATE(5498),
    [sym_infix_expression] = STATE(5979),
    [sym_ce_expression] = STATE(5979),
    [sym_sequential_expression] = STATE(5979),
    [sym_char] = STATE(6252),
    [sym_format_string] = STATE(6273),
    [sym__string_literal] = STATE(6273),
    [sym_string] = STATE(6252),
    [sym_verbatim_string] = STATE(6252),
    [sym_bytearray] = STATE(6252),
    [sym_verbatim_bytearray] = STATE(6252),
    [sym_format_triple_quoted_string] = STATE(6276),
    [sym_triple_quoted_string] = STATE(6252),
    [sym_const] = STATE(5979),
    [sym_long_identifier] = STATE(5651),
    [sym_active_pattern] = STATE(5718),
    [sym__identifier_or_op] = STATE(5652),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1605),
    [sym_sbyte] = STATE(6252),
    [sym_byte] = STATE(6252),
    [sym_int16] = STATE(6252),
    [sym_uint16] = STATE(6252),
    [sym_int32] = STATE(6252),
    [sym_uint32] = STATE(6252),
    [sym_nativeint] = STATE(6252),
    [sym_unativeint] = STATE(6252),
    [sym_int64] = STATE(6252),
    [sym_uint64] = STATE(6252),
    [sym_ieee32] = STATE(6252),
    [sym_ieee64] = STATE(6252),
    [sym_bignum] = STATE(6252),
    [sym_decimal] = STATE(6252),
    [sym_float] = STATE(5125),
    [sym_block_comment] = STATE(1775),
    [sym_compiler_directive_decl] = STATE(1775),
    [sym_fsi_directive_decl] = STATE(1775),
    [sym_preproc_line] = STATE(1775),
    [sym_preproc_if_in_expression] = STATE(5979),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2780),
    [anon_sym_null] = ACTIONS(2782),
    [anon_sym_begin] = ACTIONS(2784),
    [anon_sym_do] = ACTIONS(2788),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2790),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2796),
    [anon_sym_LBRACE] = ACTIONS(5552),
    [anon_sym_LT_AT] = ACTIONS(2800),
    [anon_sym_LT_AT_AT] = ACTIONS(5554),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2806),
    [anon_sym_new] = ACTIONS(2808),
    [anon_sym_return] = ACTIONS(2808),
    [anon_sym_return_BANG] = ACTIONS(2810),
    [anon_sym_yield] = ACTIONS(2808),
    [anon_sym_yield_BANG] = ACTIONS(2810),
    [anon_sym_lazy] = ACTIONS(2808),
    [anon_sym_assert] = ACTIONS(2808),
    [anon_sym_upcast] = ACTIONS(2808),
    [anon_sym_downcast] = ACTIONS(2808),
    [anon_sym_fixed] = ACTIONS(2808),
    [anon_sym_for] = ACTIONS(2814),
    [anon_sym_while] = ACTIONS(2816),
    [anon_sym_while_BANG] = ACTIONS(2818),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2820),
    [anon_sym_try] = ACTIONS(2822),
    [anon_sym_match] = ACTIONS(2824),
    [anon_sym_match_BANG] = ACTIONS(2826),
    [anon_sym_function] = ACTIONS(2828),
    [anon_sym_use] = ACTIONS(2832),
    [anon_sym_use_BANG] = ACTIONS(2834),
    [anon_sym_do_BANG] = ACTIONS(2836),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2843),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2845),
    [anon_sym_DQUOTE] = ACTIONS(2847),
    [anon_sym_AT_DQUOTE] = ACTIONS(2849),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2851),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2853),
    [sym_bool] = ACTIONS(2855),
    [sym_unit] = ACTIONS(5556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2859),
    [sym_op_identifier] = ACTIONS(5558),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2863),
    [sym_xint] = ACTIONS(2865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2867),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2869),
  },
  [STATE(1776)] = {
    [sym_function_or_value_defn] = STATE(1300),
    [sym__mutable_expression] = STATE(21002),
    [sym__expression] = STATE(216),
    [sym_literal_expression] = STATE(4134),
    [sym_srtp_call_expression] = STATE(4134),
    [sym__srtp_type_argument] = STATE(22507),
    [sym_long_identifier_or_op] = STATE(3940),
    [sym_tuple_expression] = STATE(4134),
    [sym_brace_expression] = STATE(4134),
    [sym_anon_record_expression] = STATE(4134),
    [sym_prefixed_expression] = STATE(4134),
    [sym_typecast_expression] = STATE(4134),
    [sym_for_expression] = STATE(4134),
    [sym_while_expression] = STATE(4134),
    [sym__if_branch] = STATE(20091),
    [sym_if_expression] = STATE(4134),
    [sym_fun_expression] = STATE(4134),
    [sym_try_expression] = STATE(4134),
    [sym_match_expression] = STATE(4134),
    [sym_function_expression] = STATE(4134),
    [sym_mutate_expression] = STATE(4134),
    [sym_index_expression] = STATE(3940),
    [sym_typed_expression] = STATE(4134),
    [sym_declaration_expression] = STATE(4134),
    [sym_do_expression] = STATE(4134),
    [sym_list_expression] = STATE(4134),
    [sym_array_expression] = STATE(4134),
    [sym_begin_end_expression] = STATE(4134),
    [sym_paren_expression] = STATE(4134),
    [sym__high_prec_app] = STATE(4137),
    [sym__low_prec_app] = STATE(4137),
    [sym_application_expression] = STATE(4134),
    [sym_dot_expression] = STATE(3940),
    [sym_infix_expression] = STATE(4134),
    [sym_ce_expression] = STATE(4134),
    [sym_sequential_expression] = STATE(4134),
    [sym_char] = STATE(4232),
    [sym_format_string] = STATE(4233),
    [sym__string_literal] = STATE(4233),
    [sym_string] = STATE(4232),
    [sym_verbatim_string] = STATE(4232),
    [sym_bytearray] = STATE(4232),
    [sym_verbatim_bytearray] = STATE(4232),
    [sym_format_triple_quoted_string] = STATE(4234),
    [sym_triple_quoted_string] = STATE(4232),
    [sym_const] = STATE(4134),
    [sym_long_identifier] = STATE(4032),
    [sym_active_pattern] = STATE(4080),
    [sym__identifier_or_op] = STATE(4062),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1717),
    [sym_sbyte] = STATE(4232),
    [sym_byte] = STATE(4232),
    [sym_int16] = STATE(4232),
    [sym_uint16] = STATE(4232),
    [sym_int32] = STATE(4232),
    [sym_uint32] = STATE(4232),
    [sym_nativeint] = STATE(4232),
    [sym_unativeint] = STATE(4232),
    [sym_int64] = STATE(4232),
    [sym_uint64] = STATE(4232),
    [sym_ieee32] = STATE(4232),
    [sym_ieee64] = STATE(4232),
    [sym_bignum] = STATE(4232),
    [sym_decimal] = STATE(4232),
    [sym_float] = STATE(3860),
    [sym_block_comment] = STATE(1776),
    [sym_compiler_directive_decl] = STATE(1776),
    [sym_fsi_directive_decl] = STATE(1776),
    [sym_preproc_line] = STATE(1776),
    [sym_preproc_if_in_expression] = STATE(4134),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_begin] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(4105),
    [anon_sym_LT_AT] = ACTIONS(1342),
    [anon_sym_LT_AT_AT] = ACTIONS(4107),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1350),
    [anon_sym_new] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_return_BANG] = ACTIONS(1354),
    [anon_sym_yield] = ACTIONS(1352),
    [anon_sym_yield_BANG] = ACTIONS(1354),
    [anon_sym_lazy] = ACTIONS(1352),
    [anon_sym_assert] = ACTIONS(1352),
    [anon_sym_upcast] = ACTIONS(1352),
    [anon_sym_downcast] = ACTIONS(1352),
    [anon_sym_fixed] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_while_BANG] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1364),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1368),
    [anon_sym_match_BANG] = ACTIONS(1370),
    [anon_sym_function] = ACTIONS(1372),
    [anon_sym_use] = ACTIONS(1376),
    [anon_sym_use_BANG] = ACTIONS(1378),
    [anon_sym_do_BANG] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1386),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [anon_sym_AT_DQUOTE] = ACTIONS(1392),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1396),
    [sym_bool] = ACTIONS(1398),
    [sym_unit] = ACTIONS(4113),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1402),
    [sym_op_identifier] = ACTIONS(4115),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1406),
    [sym_xint] = ACTIONS(1408),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1410),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1414),
  },
  [STATE(1777)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(98),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1777),
    [sym_compiler_directive_decl] = STATE(1777),
    [sym_fsi_directive_decl] = STATE(1777),
    [sym_preproc_line] = STATE(1777),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1778)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(17),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1778),
    [sym_compiler_directive_decl] = STATE(1778),
    [sym_fsi_directive_decl] = STATE(1778),
    [sym_preproc_line] = STATE(1778),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1779)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(18),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1779),
    [sym_compiler_directive_decl] = STATE(1779),
    [sym_fsi_directive_decl] = STATE(1779),
    [sym_preproc_line] = STATE(1779),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1780)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(10),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1780),
    [sym_compiler_directive_decl] = STATE(1780),
    [sym_fsi_directive_decl] = STATE(1780),
    [sym_preproc_line] = STATE(1780),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1781)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(19),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1781),
    [sym_compiler_directive_decl] = STATE(1781),
    [sym_fsi_directive_decl] = STATE(1781),
    [sym_preproc_line] = STATE(1781),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1782)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(20),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1782),
    [sym_compiler_directive_decl] = STATE(1782),
    [sym_fsi_directive_decl] = STATE(1782),
    [sym_preproc_line] = STATE(1782),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1783)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(7),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1783),
    [sym_compiler_directive_decl] = STATE(1783),
    [sym_fsi_directive_decl] = STATE(1783),
    [sym_preproc_line] = STATE(1783),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1784)] = {
    [sym_function_or_value_defn] = STATE(1295),
    [sym__mutable_expression] = STATE(21269),
    [sym__expression] = STATE(152),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22516),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1727),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(3259),
    [sym_block_comment] = STATE(1784),
    [sym_compiler_directive_decl] = STATE(1784),
    [sym_fsi_directive_decl] = STATE(1784),
    [sym_preproc_line] = STATE(1784),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_return_BANG] = ACTIONS(1074),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_yield_BANG] = ACTIONS(1074),
    [anon_sym_lazy] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_upcast] = ACTIONS(1072),
    [anon_sym_downcast] = ACTIONS(1072),
    [anon_sym_fixed] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1096),
    [anon_sym_use_BANG] = ACTIONS(1098),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1127),
    [sym_xint] = ACTIONS(1129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1785)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(122),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1785),
    [sym_compiler_directive_decl] = STATE(1785),
    [sym_fsi_directive_decl] = STATE(1785),
    [sym_preproc_line] = STATE(1785),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1786)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(110),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1786),
    [sym_compiler_directive_decl] = STATE(1786),
    [sym_fsi_directive_decl] = STATE(1786),
    [sym_preproc_line] = STATE(1786),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1787)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(111),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1787),
    [sym_compiler_directive_decl] = STATE(1787),
    [sym_fsi_directive_decl] = STATE(1787),
    [sym_preproc_line] = STATE(1787),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1788)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(100),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1788),
    [sym_compiler_directive_decl] = STATE(1788),
    [sym_fsi_directive_decl] = STATE(1788),
    [sym_preproc_line] = STATE(1788),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1789)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(113),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1789),
    [sym_compiler_directive_decl] = STATE(1789),
    [sym_fsi_directive_decl] = STATE(1789),
    [sym_preproc_line] = STATE(1789),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1790)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(114),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1790),
    [sym_compiler_directive_decl] = STATE(1790),
    [sym_fsi_directive_decl] = STATE(1790),
    [sym_preproc_line] = STATE(1790),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1791)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(121),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1791),
    [sym_compiler_directive_decl] = STATE(1791),
    [sym_fsi_directive_decl] = STATE(1791),
    [sym_preproc_line] = STATE(1791),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1792)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(206),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1792),
    [sym_compiler_directive_decl] = STATE(1792),
    [sym_fsi_directive_decl] = STATE(1792),
    [sym_preproc_line] = STATE(1792),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1793)] = {
    [sym_function_or_value_defn] = STATE(1297),
    [sym__mutable_expression] = STATE(21574),
    [sym__expression] = STATE(61),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22525),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1739),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1793),
    [sym_compiler_directive_decl] = STATE(1793),
    [sym_fsi_directive_decl] = STATE(1793),
    [sym_preproc_line] = STATE(1793),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(413),
    [anon_sym_return] = ACTIONS(413),
    [anon_sym_return_BANG] = ACTIONS(415),
    [anon_sym_yield] = ACTIONS(413),
    [anon_sym_yield_BANG] = ACTIONS(415),
    [anon_sym_lazy] = ACTIONS(413),
    [anon_sym_assert] = ACTIONS(413),
    [anon_sym_upcast] = ACTIONS(413),
    [anon_sym_downcast] = ACTIONS(413),
    [anon_sym_fixed] = ACTIONS(413),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(417),
    [anon_sym_use_BANG] = ACTIONS(419),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(423),
    [sym_xint] = ACTIONS(425),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1794)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(11),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1794),
    [sym_compiler_directive_decl] = STATE(1794),
    [sym_fsi_directive_decl] = STATE(1794),
    [sym_preproc_line] = STATE(1794),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1795)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(55),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1795),
    [sym_compiler_directive_decl] = STATE(1795),
    [sym_fsi_directive_decl] = STATE(1795),
    [sym_preproc_line] = STATE(1795),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1796)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(56),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1796),
    [sym_compiler_directive_decl] = STATE(1796),
    [sym_fsi_directive_decl] = STATE(1796),
    [sym_preproc_line] = STATE(1796),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1797)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(48),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1797),
    [sym_compiler_directive_decl] = STATE(1797),
    [sym_fsi_directive_decl] = STATE(1797),
    [sym_preproc_line] = STATE(1797),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1798)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(57),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1798),
    [sym_compiler_directive_decl] = STATE(1798),
    [sym_fsi_directive_decl] = STATE(1798),
    [sym_preproc_line] = STATE(1798),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1799)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(58),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1799),
    [sym_compiler_directive_decl] = STATE(1799),
    [sym_fsi_directive_decl] = STATE(1799),
    [sym_preproc_line] = STATE(1799),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1800)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(38),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1800),
    [sym_compiler_directive_decl] = STATE(1800),
    [sym_fsi_directive_decl] = STATE(1800),
    [sym_preproc_line] = STATE(1800),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1801)] = {
    [sym_function_or_value_defn] = STATE(1304),
    [sym__mutable_expression] = STATE(21917),
    [sym__expression] = STATE(25),
    [sym_literal_expression] = STATE(2487),
    [sym_srtp_call_expression] = STATE(2487),
    [sym__srtp_type_argument] = STATE(22534),
    [sym_long_identifier_or_op] = STATE(2424),
    [sym_tuple_expression] = STATE(2487),
    [sym_brace_expression] = STATE(2487),
    [sym_anon_record_expression] = STATE(2487),
    [sym_prefixed_expression] = STATE(2487),
    [sym_typecast_expression] = STATE(2487),
    [sym_for_expression] = STATE(2487),
    [sym_while_expression] = STATE(2487),
    [sym__if_branch] = STATE(20135),
    [sym_if_expression] = STATE(2487),
    [sym_fun_expression] = STATE(2487),
    [sym_try_expression] = STATE(2487),
    [sym_match_expression] = STATE(2487),
    [sym_function_expression] = STATE(2487),
    [sym_mutate_expression] = STATE(2487),
    [sym_index_expression] = STATE(2424),
    [sym_typed_expression] = STATE(2487),
    [sym_declaration_expression] = STATE(2487),
    [sym_do_expression] = STATE(2487),
    [sym_list_expression] = STATE(2487),
    [sym_array_expression] = STATE(2487),
    [sym_begin_end_expression] = STATE(2487),
    [sym_paren_expression] = STATE(2487),
    [sym__high_prec_app] = STATE(2488),
    [sym__low_prec_app] = STATE(2488),
    [sym_application_expression] = STATE(2487),
    [sym_dot_expression] = STATE(2424),
    [sym_infix_expression] = STATE(2487),
    [sym_ce_expression] = STATE(2487),
    [sym_sequential_expression] = STATE(2487),
    [sym_char] = STATE(2596),
    [sym_format_string] = STATE(2485),
    [sym__string_literal] = STATE(2485),
    [sym_string] = STATE(2596),
    [sym_verbatim_string] = STATE(2596),
    [sym_bytearray] = STATE(2596),
    [sym_verbatim_bytearray] = STATE(2596),
    [sym_format_triple_quoted_string] = STATE(2598),
    [sym_triple_quoted_string] = STATE(2596),
    [sym_const] = STATE(2487),
    [sym_long_identifier] = STATE(2445),
    [sym_active_pattern] = STATE(2467),
    [sym__identifier_or_op] = STATE(2446),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1749),
    [sym_sbyte] = STATE(2596),
    [sym_byte] = STATE(2596),
    [sym_int16] = STATE(2596),
    [sym_uint16] = STATE(2596),
    [sym_int32] = STATE(2596),
    [sym_uint32] = STATE(2596),
    [sym_nativeint] = STATE(2596),
    [sym_unativeint] = STATE(2596),
    [sym_int64] = STATE(2596),
    [sym_uint64] = STATE(2596),
    [sym_ieee32] = STATE(2596),
    [sym_ieee64] = STATE(2596),
    [sym_bignum] = STATE(2596),
    [sym_decimal] = STATE(2596),
    [sym_float] = STATE(2369),
    [sym_block_comment] = STATE(1801),
    [sym_compiler_directive_decl] = STATE(1801),
    [sym_fsi_directive_decl] = STATE(1801),
    [sym_preproc_line] = STATE(1801),
    [sym_preproc_if_in_expression] = STATE(2487),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(314),
    [anon_sym_null] = ACTIONS(316),
    [anon_sym_begin] = ACTIONS(318),
    [anon_sym_do] = ACTIONS(322),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(328),
    [anon_sym_LBRACK_PIPE] = ACTIONS(330),
    [anon_sym_LBRACE] = ACTIONS(5544),
    [anon_sym_LT_AT] = ACTIONS(334),
    [anon_sym_LT_AT_AT] = ACTIONS(5546),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(338),
    [anon_sym_new] = ACTIONS(340),
    [anon_sym_return] = ACTIONS(340),
    [anon_sym_return_BANG] = ACTIONS(342),
    [anon_sym_yield] = ACTIONS(340),
    [anon_sym_yield_BANG] = ACTIONS(342),
    [anon_sym_lazy] = ACTIONS(340),
    [anon_sym_assert] = ACTIONS(340),
    [anon_sym_upcast] = ACTIONS(340),
    [anon_sym_downcast] = ACTIONS(340),
    [anon_sym_fixed] = ACTIONS(340),
    [anon_sym_for] = ACTIONS(346),
    [anon_sym_while] = ACTIONS(348),
    [anon_sym_while_BANG] = ACTIONS(350),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(352),
    [anon_sym_try] = ACTIONS(354),
    [anon_sym_match] = ACTIONS(356),
    [anon_sym_match_BANG] = ACTIONS(358),
    [anon_sym_function] = ACTIONS(360),
    [anon_sym_use] = ACTIONS(362),
    [anon_sym_use_BANG] = ACTIONS(364),
    [anon_sym_do_BANG] = ACTIONS(366),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_AT_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(381),
    [sym_bool] = ACTIONS(383),
    [sym_unit] = ACTIONS(5548),
    [anon_sym_LPAREN_PIPE] = ACTIONS(385),
    [sym_op_identifier] = ACTIONS(5550),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(389),
    [sym_xint] = ACTIONS(391),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(395),
  },
  [STATE(1802)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(103),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1802),
    [sym_compiler_directive_decl] = STATE(1802),
    [sym_fsi_directive_decl] = STATE(1802),
    [sym_preproc_line] = STATE(1802),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1803)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(353),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1803),
    [sym_compiler_directive_decl] = STATE(1803),
    [sym_fsi_directive_decl] = STATE(1803),
    [sym_preproc_line] = STATE(1803),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1804)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(77),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(1804),
    [sym_compiler_directive_decl] = STATE(1804),
    [sym_fsi_directive_decl] = STATE(1804),
    [sym_preproc_line] = STATE(1804),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(1805)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(49),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1805),
    [sym_compiler_directive_decl] = STATE(1805),
    [sym_fsi_directive_decl] = STATE(1805),
    [sym_preproc_line] = STATE(1805),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1806)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(354),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1806),
    [sym_compiler_directive_decl] = STATE(1806),
    [sym_fsi_directive_decl] = STATE(1806),
    [sym_preproc_line] = STATE(1806),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1807)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(342),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1807),
    [sym_compiler_directive_decl] = STATE(1807),
    [sym_fsi_directive_decl] = STATE(1807),
    [sym_preproc_line] = STATE(1807),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1808)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(960),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1808),
    [sym_compiler_directive_decl] = STATE(1808),
    [sym_fsi_directive_decl] = STATE(1808),
    [sym_preproc_line] = STATE(1808),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1809)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(637),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(1809),
    [sym_compiler_directive_decl] = STATE(1809),
    [sym_fsi_directive_decl] = STATE(1809),
    [sym_preproc_line] = STATE(1809),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(1810)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(521),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1810),
    [sym_compiler_directive_decl] = STATE(1810),
    [sym_fsi_directive_decl] = STATE(1810),
    [sym_preproc_line] = STATE(1810),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1811)] = {
    [sym_function_or_value_defn] = STATE(1308),
    [sym__mutable_expression] = STATE(22182),
    [sym__expression] = STATE(69),
    [sym_literal_expression] = STATE(3307),
    [sym_srtp_call_expression] = STATE(3307),
    [sym__srtp_type_argument] = STATE(22543),
    [sym_long_identifier_or_op] = STATE(3077),
    [sym_tuple_expression] = STATE(3307),
    [sym_brace_expression] = STATE(3307),
    [sym_anon_record_expression] = STATE(3307),
    [sym_prefixed_expression] = STATE(3307),
    [sym_typecast_expression] = STATE(3307),
    [sym_for_expression] = STATE(3307),
    [sym_while_expression] = STATE(3307),
    [sym__if_branch] = STATE(20223),
    [sym_if_expression] = STATE(3307),
    [sym_fun_expression] = STATE(3307),
    [sym_try_expression] = STATE(3307),
    [sym_match_expression] = STATE(3307),
    [sym_function_expression] = STATE(3307),
    [sym_mutate_expression] = STATE(3307),
    [sym_index_expression] = STATE(3077),
    [sym_typed_expression] = STATE(3307),
    [sym_declaration_expression] = STATE(3307),
    [sym_do_expression] = STATE(3307),
    [sym_list_expression] = STATE(3307),
    [sym_array_expression] = STATE(3307),
    [sym_begin_end_expression] = STATE(3307),
    [sym_paren_expression] = STATE(3307),
    [sym__high_prec_app] = STATE(3322),
    [sym__low_prec_app] = STATE(3322),
    [sym_application_expression] = STATE(3307),
    [sym_dot_expression] = STATE(3077),
    [sym_infix_expression] = STATE(3307),
    [sym_ce_expression] = STATE(3307),
    [sym_sequential_expression] = STATE(3307),
    [sym_char] = STATE(3234),
    [sym_format_string] = STATE(3235),
    [sym__string_literal] = STATE(3235),
    [sym_string] = STATE(3234),
    [sym_verbatim_string] = STATE(3234),
    [sym_bytearray] = STATE(3234),
    [sym_verbatim_bytearray] = STATE(3234),
    [sym_format_triple_quoted_string] = STATE(3239),
    [sym_triple_quoted_string] = STATE(3234),
    [sym_const] = STATE(3307),
    [sym_long_identifier] = STATE(3048),
    [sym_active_pattern] = STATE(3004),
    [sym__identifier_or_op] = STATE(3049),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1768),
    [sym_sbyte] = STATE(3234),
    [sym_byte] = STATE(3234),
    [sym_int16] = STATE(3234),
    [sym_uint16] = STATE(3234),
    [sym_int32] = STATE(3234),
    [sym_uint32] = STATE(3234),
    [sym_nativeint] = STATE(3234),
    [sym_unativeint] = STATE(3234),
    [sym_int64] = STATE(3234),
    [sym_uint64] = STATE(3234),
    [sym_ieee32] = STATE(3234),
    [sym_ieee64] = STATE(3234),
    [sym_bignum] = STATE(3234),
    [sym_decimal] = STATE(3234),
    [sym_float] = STATE(2912),
    [sym_block_comment] = STATE(1811),
    [sym_compiler_directive_decl] = STATE(1811),
    [sym_fsi_directive_decl] = STATE(1811),
    [sym_preproc_line] = STATE(1811),
    [sym_preproc_if_in_expression] = STATE(3307),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(625),
    [anon_sym_null] = ACTIONS(627),
    [anon_sym_begin] = ACTIONS(629),
    [anon_sym_do] = ACTIONS(633),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(639),
    [anon_sym_LBRACK_PIPE] = ACTIONS(641),
    [anon_sym_LBRACE] = ACTIONS(5576),
    [anon_sym_LT_AT] = ACTIONS(643),
    [anon_sym_LT_AT_AT] = ACTIONS(5578),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(649),
    [anon_sym_new] = ACTIONS(651),
    [anon_sym_return] = ACTIONS(651),
    [anon_sym_return_BANG] = ACTIONS(653),
    [anon_sym_yield] = ACTIONS(651),
    [anon_sym_yield_BANG] = ACTIONS(653),
    [anon_sym_lazy] = ACTIONS(651),
    [anon_sym_assert] = ACTIONS(651),
    [anon_sym_upcast] = ACTIONS(651),
    [anon_sym_downcast] = ACTIONS(651),
    [anon_sym_fixed] = ACTIONS(651),
    [anon_sym_for] = ACTIONS(657),
    [anon_sym_while] = ACTIONS(659),
    [anon_sym_while_BANG] = ACTIONS(661),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(663),
    [anon_sym_try] = ACTIONS(665),
    [anon_sym_match] = ACTIONS(667),
    [anon_sym_match_BANG] = ACTIONS(669),
    [anon_sym_function] = ACTIONS(671),
    [anon_sym_use] = ACTIONS(675),
    [anon_sym_use_BANG] = ACTIONS(677),
    [anon_sym_do_BANG] = ACTIONS(679),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(685),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_AT_DQUOTE] = ACTIONS(689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(693),
    [sym_bool] = ACTIONS(695),
    [sym_unit] = ACTIONS(5580),
    [anon_sym_LPAREN_PIPE] = ACTIONS(699),
    [sym_op_identifier] = ACTIONS(5582),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(703),
    [sym_xint] = ACTIONS(705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(709),
  },
  [STATE(1812)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(193),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1812),
    [sym_compiler_directive_decl] = STATE(1812),
    [sym_fsi_directive_decl] = STATE(1812),
    [sym_preproc_line] = STATE(1812),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1813)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(356),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1813),
    [sym_compiler_directive_decl] = STATE(1813),
    [sym_fsi_directive_decl] = STATE(1813),
    [sym_preproc_line] = STATE(1813),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1814)] = {
    [sym_function_or_value_defn] = STATE(1310),
    [sym__mutable_expression] = STATE(22352),
    [sym__expression] = STATE(112),
    [sym_literal_expression] = STATE(3563),
    [sym_srtp_call_expression] = STATE(3563),
    [sym__srtp_type_argument] = STATE(22547),
    [sym_long_identifier_or_op] = STATE(3298),
    [sym_tuple_expression] = STATE(3563),
    [sym_brace_expression] = STATE(3563),
    [sym_anon_record_expression] = STATE(3563),
    [sym_prefixed_expression] = STATE(3563),
    [sym_typecast_expression] = STATE(3563),
    [sym_for_expression] = STATE(3563),
    [sym_while_expression] = STATE(3563),
    [sym__if_branch] = STATE(20267),
    [sym_if_expression] = STATE(3563),
    [sym_fun_expression] = STATE(3563),
    [sym_try_expression] = STATE(3563),
    [sym_match_expression] = STATE(3563),
    [sym_function_expression] = STATE(3563),
    [sym_mutate_expression] = STATE(3563),
    [sym_index_expression] = STATE(3298),
    [sym_typed_expression] = STATE(3563),
    [sym_declaration_expression] = STATE(3563),
    [sym_do_expression] = STATE(3563),
    [sym_list_expression] = STATE(3563),
    [sym_array_expression] = STATE(3563),
    [sym_begin_end_expression] = STATE(3563),
    [sym_paren_expression] = STATE(3563),
    [sym__high_prec_app] = STATE(3566),
    [sym__low_prec_app] = STATE(3566),
    [sym_application_expression] = STATE(3563),
    [sym_dot_expression] = STATE(3298),
    [sym_infix_expression] = STATE(3563),
    [sym_ce_expression] = STATE(3563),
    [sym_sequential_expression] = STATE(3563),
    [sym_char] = STATE(3590),
    [sym_format_string] = STATE(3638),
    [sym__string_literal] = STATE(3638),
    [sym_string] = STATE(3590),
    [sym_verbatim_string] = STATE(3590),
    [sym_bytearray] = STATE(3590),
    [sym_verbatim_bytearray] = STATE(3590),
    [sym_format_triple_quoted_string] = STATE(3639),
    [sym_triple_quoted_string] = STATE(3590),
    [sym_const] = STATE(3563),
    [sym_long_identifier] = STATE(3339),
    [sym_active_pattern] = STATE(3152),
    [sym__identifier_or_op] = STATE(3346),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1777),
    [sym_sbyte] = STATE(3590),
    [sym_byte] = STATE(3590),
    [sym_int16] = STATE(3590),
    [sym_uint16] = STATE(3590),
    [sym_int32] = STATE(3590),
    [sym_uint32] = STATE(3590),
    [sym_nativeint] = STATE(3590),
    [sym_unativeint] = STATE(3590),
    [sym_int64] = STATE(3590),
    [sym_uint64] = STATE(3590),
    [sym_ieee32] = STATE(3590),
    [sym_ieee64] = STATE(3590),
    [sym_bignum] = STATE(3590),
    [sym_decimal] = STATE(3590),
    [sym_float] = STATE(3012),
    [sym_block_comment] = STATE(1814),
    [sym_compiler_directive_decl] = STATE(1814),
    [sym_fsi_directive_decl] = STATE(1814),
    [sym_preproc_line] = STATE(1814),
    [sym_preproc_if_in_expression] = STATE(3563),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(739),
    [anon_sym_null] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(753),
    [anon_sym_LBRACK_PIPE] = ACTIONS(755),
    [anon_sym_LBRACE] = ACTIONS(5584),
    [anon_sym_LT_AT] = ACTIONS(757),
    [anon_sym_LT_AT_AT] = ACTIONS(5586),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(763),
    [anon_sym_new] = ACTIONS(765),
    [anon_sym_return] = ACTIONS(765),
    [anon_sym_return_BANG] = ACTIONS(767),
    [anon_sym_yield] = ACTIONS(765),
    [anon_sym_yield_BANG] = ACTIONS(767),
    [anon_sym_lazy] = ACTIONS(765),
    [anon_sym_assert] = ACTIONS(765),
    [anon_sym_upcast] = ACTIONS(765),
    [anon_sym_downcast] = ACTIONS(765),
    [anon_sym_fixed] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(771),
    [anon_sym_while] = ACTIONS(773),
    [anon_sym_while_BANG] = ACTIONS(775),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(777),
    [anon_sym_try] = ACTIONS(779),
    [anon_sym_match] = ACTIONS(781),
    [anon_sym_match_BANG] = ACTIONS(783),
    [anon_sym_function] = ACTIONS(785),
    [anon_sym_use] = ACTIONS(789),
    [anon_sym_use_BANG] = ACTIONS(791),
    [anon_sym_do_BANG] = ACTIONS(793),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(797),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(799),
    [anon_sym_DQUOTE] = ACTIONS(801),
    [anon_sym_AT_DQUOTE] = ACTIONS(803),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(807),
    [sym_bool] = ACTIONS(809),
    [sym_unit] = ACTIONS(5588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(813),
    [sym_op_identifier] = ACTIONS(5590),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(817),
    [sym_xint] = ACTIONS(819),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(821),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(823),
  },
  [STATE(1815)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(192),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1815),
    [sym_compiler_directive_decl] = STATE(1815),
    [sym_fsi_directive_decl] = STATE(1815),
    [sym_preproc_line] = STATE(1815),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1816)] = {
    [sym_function_or_value_defn] = STATE(1312),
    [sym__mutable_expression] = STATE(22752),
    [sym__expression] = STATE(127),
    [sym_literal_expression] = STATE(3522),
    [sym_srtp_call_expression] = STATE(3522),
    [sym__srtp_type_argument] = STATE(22551),
    [sym_long_identifier_or_op] = STATE(3130),
    [sym_tuple_expression] = STATE(3522),
    [sym_brace_expression] = STATE(3522),
    [sym_anon_record_expression] = STATE(3522),
    [sym_prefixed_expression] = STATE(3522),
    [sym_typecast_expression] = STATE(3522),
    [sym_for_expression] = STATE(3522),
    [sym_while_expression] = STATE(3522),
    [sym__if_branch] = STATE(20311),
    [sym_if_expression] = STATE(3522),
    [sym_fun_expression] = STATE(3522),
    [sym_try_expression] = STATE(3522),
    [sym_match_expression] = STATE(3522),
    [sym_function_expression] = STATE(3522),
    [sym_mutate_expression] = STATE(3522),
    [sym_index_expression] = STATE(3130),
    [sym_typed_expression] = STATE(3522),
    [sym_declaration_expression] = STATE(3522),
    [sym_do_expression] = STATE(3522),
    [sym_list_expression] = STATE(3522),
    [sym_array_expression] = STATE(3522),
    [sym_begin_end_expression] = STATE(3522),
    [sym_paren_expression] = STATE(3522),
    [sym__high_prec_app] = STATE(3534),
    [sym__low_prec_app] = STATE(3534),
    [sym_application_expression] = STATE(3522),
    [sym_dot_expression] = STATE(3130),
    [sym_infix_expression] = STATE(3522),
    [sym_ce_expression] = STATE(3522),
    [sym_sequential_expression] = STATE(3522),
    [sym_char] = STATE(3408),
    [sym_format_string] = STATE(3409),
    [sym__string_literal] = STATE(3409),
    [sym_string] = STATE(3408),
    [sym_verbatim_string] = STATE(3408),
    [sym_bytearray] = STATE(3408),
    [sym_verbatim_bytearray] = STATE(3408),
    [sym_format_triple_quoted_string] = STATE(3411),
    [sym_triple_quoted_string] = STATE(3408),
    [sym_const] = STATE(3522),
    [sym_long_identifier] = STATE(3087),
    [sym_active_pattern] = STATE(3220),
    [sym__identifier_or_op] = STATE(3089),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1785),
    [sym_sbyte] = STATE(3408),
    [sym_byte] = STATE(3408),
    [sym_int16] = STATE(3408),
    [sym_uint16] = STATE(3408),
    [sym_int32] = STATE(3408),
    [sym_uint32] = STATE(3408),
    [sym_nativeint] = STATE(3408),
    [sym_unativeint] = STATE(3408),
    [sym_int64] = STATE(3408),
    [sym_uint64] = STATE(3408),
    [sym_ieee32] = STATE(3408),
    [sym_ieee64] = STATE(3408),
    [sym_bignum] = STATE(3408),
    [sym_decimal] = STATE(3408),
    [sym_float] = STATE(3020),
    [sym_block_comment] = STATE(1816),
    [sym_compiler_directive_decl] = STATE(1816),
    [sym_fsi_directive_decl] = STATE(1816),
    [sym_preproc_line] = STATE(1816),
    [sym_preproc_if_in_expression] = STATE(3522),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(949),
    [anon_sym_null] = ACTIONS(951),
    [anon_sym_begin] = ACTIONS(953),
    [anon_sym_do] = ACTIONS(957),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(959),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(965),
    [anon_sym_LBRACE] = ACTIONS(5592),
    [anon_sym_LT_AT] = ACTIONS(969),
    [anon_sym_LT_AT_AT] = ACTIONS(5594),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(973),
    [anon_sym_new] = ACTIONS(975),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_return_BANG] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(975),
    [anon_sym_yield_BANG] = ACTIONS(977),
    [anon_sym_lazy] = ACTIONS(975),
    [anon_sym_assert] = ACTIONS(975),
    [anon_sym_upcast] = ACTIONS(975),
    [anon_sym_downcast] = ACTIONS(975),
    [anon_sym_fixed] = ACTIONS(975),
    [anon_sym_for] = ACTIONS(981),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_while_BANG] = ACTIONS(985),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(987),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_match] = ACTIONS(991),
    [anon_sym_match_BANG] = ACTIONS(993),
    [anon_sym_function] = ACTIONS(995),
    [anon_sym_use] = ACTIONS(997),
    [anon_sym_use_BANG] = ACTIONS(999),
    [anon_sym_do_BANG] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1006),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1008),
    [anon_sym_DQUOTE] = ACTIONS(1010),
    [anon_sym_AT_DQUOTE] = ACTIONS(1012),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1016),
    [sym_bool] = ACTIONS(1018),
    [sym_unit] = ACTIONS(5596),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1020),
    [sym_op_identifier] = ACTIONS(5598),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1024),
    [sym_xint] = ACTIONS(1026),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1028),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1030),
  },
  [STATE(1817)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(357),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1817),
    [sym_compiler_directive_decl] = STATE(1817),
    [sym_fsi_directive_decl] = STATE(1817),
    [sym_preproc_line] = STATE(1817),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1818)] = {
    [sym_function_or_value_defn] = STATE(1313),
    [sym__mutable_expression] = STATE(19100),
    [sym__expression] = STATE(15),
    [sym_literal_expression] = STATE(2500),
    [sym_srtp_call_expression] = STATE(2500),
    [sym__srtp_type_argument] = STATE(22555),
    [sym_long_identifier_or_op] = STATE(2451),
    [sym_tuple_expression] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_anon_record_expression] = STATE(2500),
    [sym_prefixed_expression] = STATE(2500),
    [sym_typecast_expression] = STATE(2500),
    [sym_for_expression] = STATE(2500),
    [sym_while_expression] = STATE(2500),
    [sym__if_branch] = STATE(20355),
    [sym_if_expression] = STATE(2500),
    [sym_fun_expression] = STATE(2500),
    [sym_try_expression] = STATE(2500),
    [sym_match_expression] = STATE(2500),
    [sym_function_expression] = STATE(2500),
    [sym_mutate_expression] = STATE(2500),
    [sym_index_expression] = STATE(2451),
    [sym_typed_expression] = STATE(2500),
    [sym_declaration_expression] = STATE(2500),
    [sym_do_expression] = STATE(2500),
    [sym_list_expression] = STATE(2500),
    [sym_array_expression] = STATE(2500),
    [sym_begin_end_expression] = STATE(2500),
    [sym_paren_expression] = STATE(2500),
    [sym__high_prec_app] = STATE(2511),
    [sym__low_prec_app] = STATE(2511),
    [sym_application_expression] = STATE(2500),
    [sym_dot_expression] = STATE(2451),
    [sym_infix_expression] = STATE(2500),
    [sym_ce_expression] = STATE(2500),
    [sym_sequential_expression] = STATE(2500),
    [sym_char] = STATE(2539),
    [sym_format_string] = STATE(2543),
    [sym__string_literal] = STATE(2543),
    [sym_string] = STATE(2539),
    [sym_verbatim_string] = STATE(2539),
    [sym_bytearray] = STATE(2539),
    [sym_verbatim_bytearray] = STATE(2539),
    [sym_format_triple_quoted_string] = STATE(2544),
    [sym_triple_quoted_string] = STATE(2539),
    [sym_const] = STATE(2500),
    [sym_long_identifier] = STATE(2477),
    [sym_active_pattern] = STATE(2480),
    [sym__identifier_or_op] = STATE(2478),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1794),
    [sym_sbyte] = STATE(2539),
    [sym_byte] = STATE(2539),
    [sym_int16] = STATE(2539),
    [sym_uint16] = STATE(2539),
    [sym_int32] = STATE(2539),
    [sym_uint32] = STATE(2539),
    [sym_nativeint] = STATE(2539),
    [sym_unativeint] = STATE(2539),
    [sym_int64] = STATE(2539),
    [sym_uint64] = STATE(2539),
    [sym_ieee32] = STATE(2539),
    [sym_ieee64] = STATE(2539),
    [sym_bignum] = STATE(2539),
    [sym_decimal] = STATE(2539),
    [sym_float] = STATE(2377),
    [sym_block_comment] = STATE(1818),
    [sym_compiler_directive_decl] = STATE(1818),
    [sym_fsi_directive_decl] = STATE(1818),
    [sym_preproc_line] = STATE(1818),
    [sym_preproc_if_in_expression] = STATE(2500),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(191),
    [anon_sym_null] = ACTIONS(195),
    [anon_sym_begin] = ACTIONS(197),
    [anon_sym_do] = ACTIONS(201),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(207),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(5600),
    [anon_sym_LT_AT] = ACTIONS(215),
    [anon_sym_LT_AT_AT] = ACTIONS(5602),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(221),
    [anon_sym_new] = ACTIONS(223),
    [anon_sym_return] = ACTIONS(223),
    [anon_sym_return_BANG] = ACTIONS(225),
    [anon_sym_yield] = ACTIONS(223),
    [anon_sym_yield_BANG] = ACTIONS(225),
    [anon_sym_lazy] = ACTIONS(223),
    [anon_sym_assert] = ACTIONS(223),
    [anon_sym_upcast] = ACTIONS(223),
    [anon_sym_downcast] = ACTIONS(223),
    [anon_sym_fixed] = ACTIONS(223),
    [anon_sym_for] = ACTIONS(227),
    [anon_sym_while] = ACTIONS(229),
    [anon_sym_while_BANG] = ACTIONS(231),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(233),
    [anon_sym_try] = ACTIONS(235),
    [anon_sym_match] = ACTIONS(237),
    [anon_sym_match_BANG] = ACTIONS(239),
    [anon_sym_function] = ACTIONS(241),
    [anon_sym_use] = ACTIONS(243),
    [anon_sym_use_BANG] = ACTIONS(245),
    [anon_sym_do_BANG] = ACTIONS(247),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(252),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [anon_sym_AT_DQUOTE] = ACTIONS(258),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(260),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(262),
    [sym_bool] = ACTIONS(264),
    [sym_unit] = ACTIONS(5604),
    [anon_sym_LPAREN_PIPE] = ACTIONS(266),
    [sym_op_identifier] = ACTIONS(5606),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(272),
    [sym_xint] = ACTIONS(274),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(276),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(278),
  },
  [STATE(1819)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(338),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1819),
    [sym_compiler_directive_decl] = STATE(1819),
    [sym_fsi_directive_decl] = STATE(1819),
    [sym_preproc_line] = STATE(1819),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1820)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(985),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1820),
    [sym_compiler_directive_decl] = STATE(1820),
    [sym_fsi_directive_decl] = STATE(1820),
    [sym_preproc_line] = STATE(1820),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1821)] = {
    [sym_function_or_value_defn] = STATE(1314),
    [sym__mutable_expression] = STATE(19168),
    [sym__expression] = STATE(108),
    [sym_literal_expression] = STATE(3568),
    [sym_srtp_call_expression] = STATE(3568),
    [sym__srtp_type_argument] = STATE(22559),
    [sym_long_identifier_or_op] = STATE(3303),
    [sym_tuple_expression] = STATE(3568),
    [sym_brace_expression] = STATE(3568),
    [sym_anon_record_expression] = STATE(3568),
    [sym_prefixed_expression] = STATE(3568),
    [sym_typecast_expression] = STATE(3568),
    [sym_for_expression] = STATE(3568),
    [sym_while_expression] = STATE(3568),
    [sym__if_branch] = STATE(20399),
    [sym_if_expression] = STATE(3568),
    [sym_fun_expression] = STATE(3568),
    [sym_try_expression] = STATE(3568),
    [sym_match_expression] = STATE(3568),
    [sym_function_expression] = STATE(3568),
    [sym_mutate_expression] = STATE(3568),
    [sym_index_expression] = STATE(3303),
    [sym_typed_expression] = STATE(3568),
    [sym_declaration_expression] = STATE(3568),
    [sym_do_expression] = STATE(3568),
    [sym_list_expression] = STATE(3568),
    [sym_array_expression] = STATE(3568),
    [sym_begin_end_expression] = STATE(3568),
    [sym_paren_expression] = STATE(3568),
    [sym__high_prec_app] = STATE(3573),
    [sym__low_prec_app] = STATE(3573),
    [sym_application_expression] = STATE(3568),
    [sym_dot_expression] = STATE(3303),
    [sym_infix_expression] = STATE(3568),
    [sym_ce_expression] = STATE(3568),
    [sym_sequential_expression] = STATE(3568),
    [sym_char] = STATE(3471),
    [sym_format_string] = STATE(3473),
    [sym__string_literal] = STATE(3473),
    [sym_string] = STATE(3471),
    [sym_verbatim_string] = STATE(3471),
    [sym_bytearray] = STATE(3471),
    [sym_verbatim_bytearray] = STATE(3471),
    [sym_format_triple_quoted_string] = STATE(3474),
    [sym_triple_quoted_string] = STATE(3471),
    [sym_const] = STATE(3568),
    [sym_long_identifier] = STATE(3156),
    [sym_active_pattern] = STATE(3252),
    [sym__identifier_or_op] = STATE(3157),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1802),
    [sym_sbyte] = STATE(3471),
    [sym_byte] = STATE(3471),
    [sym_int16] = STATE(3471),
    [sym_uint16] = STATE(3471),
    [sym_int32] = STATE(3471),
    [sym_uint32] = STATE(3471),
    [sym_nativeint] = STATE(3471),
    [sym_unativeint] = STATE(3471),
    [sym_int64] = STATE(3471),
    [sym_uint64] = STATE(3471),
    [sym_ieee32] = STATE(3471),
    [sym_ieee64] = STATE(3471),
    [sym_bignum] = STATE(3471),
    [sym_decimal] = STATE(3471),
    [sym_float] = STATE(3041),
    [sym_block_comment] = STATE(1821),
    [sym_compiler_directive_decl] = STATE(1821),
    [sym_fsi_directive_decl] = STATE(1821),
    [sym_preproc_line] = STATE(1821),
    [sym_preproc_if_in_expression] = STATE(3568),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(858),
    [anon_sym_null] = ACTIONS(860),
    [anon_sym_begin] = ACTIONS(862),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(868),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(874),
    [anon_sym_LBRACE] = ACTIONS(5608),
    [anon_sym_LT_AT] = ACTIONS(878),
    [anon_sym_LT_AT_AT] = ACTIONS(5610),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(882),
    [anon_sym_new] = ACTIONS(884),
    [anon_sym_return] = ACTIONS(884),
    [anon_sym_return_BANG] = ACTIONS(886),
    [anon_sym_yield] = ACTIONS(884),
    [anon_sym_yield_BANG] = ACTIONS(886),
    [anon_sym_lazy] = ACTIONS(884),
    [anon_sym_assert] = ACTIONS(884),
    [anon_sym_upcast] = ACTIONS(884),
    [anon_sym_downcast] = ACTIONS(884),
    [anon_sym_fixed] = ACTIONS(884),
    [anon_sym_for] = ACTIONS(890),
    [anon_sym_while] = ACTIONS(892),
    [anon_sym_while_BANG] = ACTIONS(894),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(896),
    [anon_sym_try] = ACTIONS(898),
    [anon_sym_match] = ACTIONS(900),
    [anon_sym_match_BANG] = ACTIONS(902),
    [anon_sym_function] = ACTIONS(904),
    [anon_sym_use] = ACTIONS(906),
    [anon_sym_use_BANG] = ACTIONS(908),
    [anon_sym_do_BANG] = ACTIONS(910),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE] = ACTIONS(919),
    [anon_sym_AT_DQUOTE] = ACTIONS(921),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(925),
    [sym_bool] = ACTIONS(927),
    [sym_unit] = ACTIONS(5612),
    [anon_sym_LPAREN_PIPE] = ACTIONS(929),
    [sym_op_identifier] = ACTIONS(5614),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(933),
    [sym_xint] = ACTIONS(935),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(937),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(939),
  },
  [STATE(1822)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(533),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1822),
    [sym_compiler_directive_decl] = STATE(1822),
    [sym_fsi_directive_decl] = STATE(1822),
    [sym_preproc_line] = STATE(1822),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1823)] = {
    [sym_function_or_value_defn] = STATE(1315),
    [sym__mutable_expression] = STATE(19233),
    [sym__expression] = STATE(53),
    [sym_literal_expression] = STATE(2804),
    [sym_srtp_call_expression] = STATE(2804),
    [sym__srtp_type_argument] = STATE(22562),
    [sym_long_identifier_or_op] = STATE(2657),
    [sym_tuple_expression] = STATE(2804),
    [sym_brace_expression] = STATE(2804),
    [sym_anon_record_expression] = STATE(2804),
    [sym_prefixed_expression] = STATE(2804),
    [sym_typecast_expression] = STATE(2804),
    [sym_for_expression] = STATE(2804),
    [sym_while_expression] = STATE(2804),
    [sym__if_branch] = STATE(20443),
    [sym_if_expression] = STATE(2804),
    [sym_fun_expression] = STATE(2804),
    [sym_try_expression] = STATE(2804),
    [sym_match_expression] = STATE(2804),
    [sym_function_expression] = STATE(2804),
    [sym_mutate_expression] = STATE(2804),
    [sym_index_expression] = STATE(2657),
    [sym_typed_expression] = STATE(2804),
    [sym_declaration_expression] = STATE(2804),
    [sym_do_expression] = STATE(2804),
    [sym_list_expression] = STATE(2804),
    [sym_array_expression] = STATE(2804),
    [sym_begin_end_expression] = STATE(2804),
    [sym_paren_expression] = STATE(2804),
    [sym__high_prec_app] = STATE(2806),
    [sym__low_prec_app] = STATE(2806),
    [sym_application_expression] = STATE(2804),
    [sym_dot_expression] = STATE(2657),
    [sym_infix_expression] = STATE(2804),
    [sym_ce_expression] = STATE(2804),
    [sym_sequential_expression] = STATE(2804),
    [sym_char] = STATE(2674),
    [sym_format_string] = STATE(2817),
    [sym__string_literal] = STATE(2817),
    [sym_string] = STATE(2674),
    [sym_verbatim_string] = STATE(2674),
    [sym_bytearray] = STATE(2674),
    [sym_verbatim_bytearray] = STATE(2674),
    [sym_format_triple_quoted_string] = STATE(2675),
    [sym_triple_quoted_string] = STATE(2674),
    [sym_const] = STATE(2804),
    [sym_long_identifier] = STATE(2665),
    [sym_active_pattern] = STATE(2667),
    [sym__identifier_or_op] = STATE(2666),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1805),
    [sym_sbyte] = STATE(2674),
    [sym_byte] = STATE(2674),
    [sym_int16] = STATE(2674),
    [sym_uint16] = STATE(2674),
    [sym_int32] = STATE(2674),
    [sym_uint32] = STATE(2674),
    [sym_nativeint] = STATE(2674),
    [sym_unativeint] = STATE(2674),
    [sym_int64] = STATE(2674),
    [sym_uint64] = STATE(2674),
    [sym_ieee32] = STATE(2674),
    [sym_ieee64] = STATE(2674),
    [sym_bignum] = STATE(2674),
    [sym_decimal] = STATE(2674),
    [sym_float] = STATE(2435),
    [sym_block_comment] = STATE(1823),
    [sym_compiler_directive_decl] = STATE(1823),
    [sym_fsi_directive_decl] = STATE(1823),
    [sym_preproc_line] = STATE(1823),
    [sym_preproc_if_in_expression] = STATE(2804),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(458),
    [anon_sym_null] = ACTIONS(460),
    [anon_sym_begin] = ACTIONS(462),
    [anon_sym_do] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(468),
    [anon_sym_LBRACK_PIPE] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(5512),
    [anon_sym_LT_AT] = ACTIONS(474),
    [anon_sym_LT_AT_AT] = ACTIONS(5514),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(478),
    [anon_sym_new] = ACTIONS(480),
    [anon_sym_return] = ACTIONS(480),
    [anon_sym_return_BANG] = ACTIONS(482),
    [anon_sym_yield] = ACTIONS(480),
    [anon_sym_yield_BANG] = ACTIONS(482),
    [anon_sym_lazy] = ACTIONS(480),
    [anon_sym_assert] = ACTIONS(480),
    [anon_sym_upcast] = ACTIONS(480),
    [anon_sym_downcast] = ACTIONS(480),
    [anon_sym_fixed] = ACTIONS(480),
    [anon_sym_for] = ACTIONS(484),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_while_BANG] = ACTIONS(488),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(490),
    [anon_sym_try] = ACTIONS(492),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_match_BANG] = ACTIONS(496),
    [anon_sym_function] = ACTIONS(498),
    [anon_sym_use] = ACTIONS(500),
    [anon_sym_use_BANG] = ACTIONS(502),
    [anon_sym_do_BANG] = ACTIONS(504),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(509),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(511),
    [anon_sym_DQUOTE] = ACTIONS(513),
    [anon_sym_AT_DQUOTE] = ACTIONS(515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(519),
    [sym_bool] = ACTIONS(521),
    [sym_unit] = ACTIONS(5516),
    [anon_sym_LPAREN_PIPE] = ACTIONS(523),
    [sym_op_identifier] = ACTIONS(5518),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(527),
    [sym_xint] = ACTIONS(529),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(531),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(533),
  },
  [STATE(1824)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(261),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1824),
    [sym_compiler_directive_decl] = STATE(1824),
    [sym_fsi_directive_decl] = STATE(1824),
    [sym_preproc_line] = STATE(1824),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1825)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(269),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1825),
    [sym_compiler_directive_decl] = STATE(1825),
    [sym_fsi_directive_decl] = STATE(1825),
    [sym_preproc_line] = STATE(1825),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1826)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(283),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1826),
    [sym_compiler_directive_decl] = STATE(1826),
    [sym_fsi_directive_decl] = STATE(1826),
    [sym_preproc_line] = STATE(1826),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1827)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(275),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1827),
    [sym_compiler_directive_decl] = STATE(1827),
    [sym_fsi_directive_decl] = STATE(1827),
    [sym_preproc_line] = STATE(1827),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1828)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(271),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1828),
    [sym_compiler_directive_decl] = STATE(1828),
    [sym_fsi_directive_decl] = STATE(1828),
    [sym_preproc_line] = STATE(1828),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1829)] = {
    [sym_function_or_value_defn] = STATE(1382),
    [sym__mutable_expression] = STATE(21446),
    [sym__expression] = STATE(268),
    [sym_literal_expression] = STATE(3545),
    [sym_srtp_call_expression] = STATE(3545),
    [sym__srtp_type_argument] = STATE(22418),
    [sym_long_identifier_or_op] = STATE(3277),
    [sym_tuple_expression] = STATE(3545),
    [sym_brace_expression] = STATE(3545),
    [sym_anon_record_expression] = STATE(3545),
    [sym_prefixed_expression] = STATE(3545),
    [sym_typecast_expression] = STATE(3545),
    [sym_for_expression] = STATE(3545),
    [sym_while_expression] = STATE(3545),
    [sym__if_branch] = STATE(19651),
    [sym_if_expression] = STATE(3545),
    [sym_fun_expression] = STATE(3545),
    [sym_try_expression] = STATE(3545),
    [sym_match_expression] = STATE(3545),
    [sym_function_expression] = STATE(3545),
    [sym_mutate_expression] = STATE(3545),
    [sym_index_expression] = STATE(3277),
    [sym_typed_expression] = STATE(3545),
    [sym_declaration_expression] = STATE(3545),
    [sym_do_expression] = STATE(3545),
    [sym_list_expression] = STATE(3545),
    [sym_array_expression] = STATE(3545),
    [sym_begin_end_expression] = STATE(3545),
    [sym_paren_expression] = STATE(3545),
    [sym__high_prec_app] = STATE(3549),
    [sym__low_prec_app] = STATE(3549),
    [sym_application_expression] = STATE(3545),
    [sym_dot_expression] = STATE(3277),
    [sym_infix_expression] = STATE(3545),
    [sym_ce_expression] = STATE(3545),
    [sym_sequential_expression] = STATE(3545),
    [sym_char] = STATE(3597),
    [sym_format_string] = STATE(3616),
    [sym__string_literal] = STATE(3616),
    [sym_string] = STATE(3597),
    [sym_verbatim_string] = STATE(3597),
    [sym_bytearray] = STATE(3597),
    [sym_verbatim_bytearray] = STATE(3597),
    [sym_format_triple_quoted_string] = STATE(3622),
    [sym_triple_quoted_string] = STATE(3597),
    [sym_const] = STATE(3545),
    [sym_long_identifier] = STATE(3299),
    [sym_active_pattern] = STATE(3110),
    [sym__identifier_or_op] = STATE(3106),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1582),
    [sym_sbyte] = STATE(3597),
    [sym_byte] = STATE(3597),
    [sym_int16] = STATE(3597),
    [sym_uint16] = STATE(3597),
    [sym_int32] = STATE(3597),
    [sym_uint32] = STATE(3597),
    [sym_nativeint] = STATE(3597),
    [sym_unativeint] = STATE(3597),
    [sym_int64] = STATE(3597),
    [sym_uint64] = STATE(3597),
    [sym_ieee32] = STATE(3597),
    [sym_ieee64] = STATE(3597),
    [sym_bignum] = STATE(3597),
    [sym_decimal] = STATE(3597),
    [sym_float] = STATE(4656),
    [sym_block_comment] = STATE(1829),
    [sym_compiler_directive_decl] = STATE(1829),
    [sym_fsi_directive_decl] = STATE(1829),
    [sym_preproc_line] = STATE(1829),
    [sym_preproc_if_in_expression] = STATE(3545),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1044),
    [anon_sym_null] = ACTIONS(1046),
    [anon_sym_begin] = ACTIONS(1048),
    [anon_sym_do] = ACTIONS(1052),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1054),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1058),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(5143),
    [anon_sym_LT_AT] = ACTIONS(1064),
    [anon_sym_LT_AT_AT] = ACTIONS(5145),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1731),
    [anon_sym_return] = ACTIONS(1731),
    [anon_sym_return_BANG] = ACTIONS(1733),
    [anon_sym_yield] = ACTIONS(1731),
    [anon_sym_yield_BANG] = ACTIONS(1733),
    [anon_sym_lazy] = ACTIONS(1731),
    [anon_sym_assert] = ACTIONS(1731),
    [anon_sym_upcast] = ACTIONS(1731),
    [anon_sym_downcast] = ACTIONS(1731),
    [anon_sym_fixed] = ACTIONS(1731),
    [anon_sym_for] = ACTIONS(1735),
    [anon_sym_while] = ACTIONS(1080),
    [anon_sym_while_BANG] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1084),
    [anon_sym_try] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1088),
    [anon_sym_match_BANG] = ACTIONS(1090),
    [anon_sym_function] = ACTIONS(1092),
    [anon_sym_use] = ACTIONS(1737),
    [anon_sym_use_BANG] = ACTIONS(1739),
    [anon_sym_do_BANG] = ACTIONS(1100),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1109),
    [anon_sym_DQUOTE] = ACTIONS(1111),
    [anon_sym_AT_DQUOTE] = ACTIONS(1113),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1117),
    [sym_bool] = ACTIONS(1119),
    [sym_unit] = ACTIONS(5155),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1123),
    [sym_op_identifier] = ACTIONS(5157),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1743),
    [sym_xint] = ACTIONS(1745),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1133),
  },
  [STATE(1830)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(540),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1830),
    [sym_compiler_directive_decl] = STATE(1830),
    [sym_fsi_directive_decl] = STATE(1830),
    [sym_preproc_line] = STATE(1830),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1831)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(373),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1831),
    [sym_compiler_directive_decl] = STATE(1831),
    [sym_fsi_directive_decl] = STATE(1831),
    [sym_preproc_line] = STATE(1831),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1832)] = {
    [sym_function_or_value_defn] = STATE(1356),
    [sym__mutable_expression] = STATE(19215),
    [sym__expression] = STATE(343),
    [sym_literal_expression] = STATE(5525),
    [sym_srtp_call_expression] = STATE(5525),
    [sym__srtp_type_argument] = STATE(22445),
    [sym_long_identifier_or_op] = STATE(5038),
    [sym_tuple_expression] = STATE(5525),
    [sym_brace_expression] = STATE(5525),
    [sym_anon_record_expression] = STATE(5525),
    [sym_prefixed_expression] = STATE(5525),
    [sym_typecast_expression] = STATE(5525),
    [sym_for_expression] = STATE(5525),
    [sym_while_expression] = STATE(5525),
    [sym__if_branch] = STATE(19856),
    [sym_if_expression] = STATE(5525),
    [sym_fun_expression] = STATE(5525),
    [sym_try_expression] = STATE(5525),
    [sym_match_expression] = STATE(5525),
    [sym_function_expression] = STATE(5525),
    [sym_mutate_expression] = STATE(5525),
    [sym_index_expression] = STATE(5038),
    [sym_typed_expression] = STATE(5525),
    [sym_declaration_expression] = STATE(5525),
    [sym_do_expression] = STATE(5525),
    [sym_list_expression] = STATE(5525),
    [sym_array_expression] = STATE(5525),
    [sym_begin_end_expression] = STATE(5525),
    [sym_paren_expression] = STATE(5525),
    [sym__high_prec_app] = STATE(5529),
    [sym__low_prec_app] = STATE(5529),
    [sym_application_expression] = STATE(5525),
    [sym_dot_expression] = STATE(5038),
    [sym_infix_expression] = STATE(5525),
    [sym_ce_expression] = STATE(5525),
    [sym_sequential_expression] = STATE(5525),
    [sym_char] = STATE(5431),
    [sym_format_string] = STATE(5453),
    [sym__string_literal] = STATE(5453),
    [sym_string] = STATE(5431),
    [sym_verbatim_string] = STATE(5431),
    [sym_bytearray] = STATE(5431),
    [sym_verbatim_bytearray] = STATE(5431),
    [sym_format_triple_quoted_string] = STATE(5455),
    [sym_triple_quoted_string] = STATE(5431),
    [sym_const] = STATE(5525),
    [sym_long_identifier] = STATE(5012),
    [sym_active_pattern] = STATE(5083),
    [sym__identifier_or_op] = STATE(5013),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1651),
    [sym_sbyte] = STATE(5431),
    [sym_byte] = STATE(5431),
    [sym_int16] = STATE(5431),
    [sym_uint16] = STATE(5431),
    [sym_int32] = STATE(5431),
    [sym_uint32] = STATE(5431),
    [sym_nativeint] = STATE(5431),
    [sym_unativeint] = STATE(5431),
    [sym_int64] = STATE(5431),
    [sym_uint64] = STATE(5431),
    [sym_ieee32] = STATE(5431),
    [sym_ieee64] = STATE(5431),
    [sym_bignum] = STATE(5431),
    [sym_decimal] = STATE(5431),
    [sym_float] = STATE(4820),
    [sym_block_comment] = STATE(1832),
    [sym_compiler_directive_decl] = STATE(1832),
    [sym_fsi_directive_decl] = STATE(1832),
    [sym_preproc_line] = STATE(1832),
    [sym_preproc_if_in_expression] = STATE(5525),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_begin] = ACTIONS(1972),
    [anon_sym_do] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1984),
    [anon_sym_LBRACE] = ACTIONS(5191),
    [anon_sym_LT_AT] = ACTIONS(1988),
    [anon_sym_LT_AT_AT] = ACTIONS(5193),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1996),
    [anon_sym_new] = ACTIONS(1998),
    [anon_sym_return] = ACTIONS(1998),
    [anon_sym_return_BANG] = ACTIONS(2000),
    [anon_sym_yield] = ACTIONS(1998),
    [anon_sym_yield_BANG] = ACTIONS(2000),
    [anon_sym_lazy] = ACTIONS(1998),
    [anon_sym_assert] = ACTIONS(1998),
    [anon_sym_upcast] = ACTIONS(1998),
    [anon_sym_downcast] = ACTIONS(1998),
    [anon_sym_fixed] = ACTIONS(1998),
    [anon_sym_for] = ACTIONS(2004),
    [anon_sym_while] = ACTIONS(2006),
    [anon_sym_while_BANG] = ACTIONS(2008),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2010),
    [anon_sym_try] = ACTIONS(2012),
    [anon_sym_match] = ACTIONS(2014),
    [anon_sym_match_BANG] = ACTIONS(2016),
    [anon_sym_function] = ACTIONS(2018),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_use_BANG] = ACTIONS(2024),
    [anon_sym_do_BANG] = ACTIONS(2026),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2032),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2034),
    [anon_sym_DQUOTE] = ACTIONS(2036),
    [anon_sym_AT_DQUOTE] = ACTIONS(2038),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2040),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2042),
    [sym_bool] = ACTIONS(2044),
    [sym_unit] = ACTIONS(5201),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2048),
    [sym_op_identifier] = ACTIONS(5203),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2052),
    [sym_xint] = ACTIONS(2054),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2056),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2060),
  },
  [STATE(1833)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(621),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1833),
    [sym_compiler_directive_decl] = STATE(1833),
    [sym_fsi_directive_decl] = STATE(1833),
    [sym_preproc_line] = STATE(1833),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1834)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(550),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1834),
    [sym_compiler_directive_decl] = STATE(1834),
    [sym_fsi_directive_decl] = STATE(1834),
    [sym_preproc_line] = STATE(1834),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1835)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(559),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1835),
    [sym_compiler_directive_decl] = STATE(1835),
    [sym_fsi_directive_decl] = STATE(1835),
    [sym_preproc_line] = STATE(1835),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1836)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(561),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1836),
    [sym_compiler_directive_decl] = STATE(1836),
    [sym_fsi_directive_decl] = STATE(1836),
    [sym_preproc_line] = STATE(1836),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1837)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(562),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1837),
    [sym_compiler_directive_decl] = STATE(1837),
    [sym_fsi_directive_decl] = STATE(1837),
    [sym_preproc_line] = STATE(1837),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1838)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(563),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1838),
    [sym_compiler_directive_decl] = STATE(1838),
    [sym_fsi_directive_decl] = STATE(1838),
    [sym_preproc_line] = STATE(1838),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1839)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(929),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1839),
    [sym_compiler_directive_decl] = STATE(1839),
    [sym_fsi_directive_decl] = STATE(1839),
    [sym_preproc_line] = STATE(1839),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1840)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(526),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1840),
    [sym_compiler_directive_decl] = STATE(1840),
    [sym_fsi_directive_decl] = STATE(1840),
    [sym_preproc_line] = STATE(1840),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1841)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(527),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1841),
    [sym_compiler_directive_decl] = STATE(1841),
    [sym_fsi_directive_decl] = STATE(1841),
    [sym_preproc_line] = STATE(1841),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1842)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(528),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1842),
    [sym_compiler_directive_decl] = STATE(1842),
    [sym_fsi_directive_decl] = STATE(1842),
    [sym_preproc_line] = STATE(1842),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1843)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(529),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1843),
    [sym_compiler_directive_decl] = STATE(1843),
    [sym_fsi_directive_decl] = STATE(1843),
    [sym_preproc_line] = STATE(1843),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1844)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(626),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(1844),
    [sym_compiler_directive_decl] = STATE(1844),
    [sym_fsi_directive_decl] = STATE(1844),
    [sym_preproc_line] = STATE(1844),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(1845)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(317),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1845),
    [sym_compiler_directive_decl] = STATE(1845),
    [sym_fsi_directive_decl] = STATE(1845),
    [sym_preproc_line] = STATE(1845),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1846)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(194),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(1846),
    [sym_compiler_directive_decl] = STATE(1846),
    [sym_fsi_directive_decl] = STATE(1846),
    [sym_preproc_line] = STATE(1846),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(1847)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(582),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1847),
    [sym_compiler_directive_decl] = STATE(1847),
    [sym_fsi_directive_decl] = STATE(1847),
    [sym_preproc_line] = STATE(1847),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1848)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(583),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1848),
    [sym_compiler_directive_decl] = STATE(1848),
    [sym_fsi_directive_decl] = STATE(1848),
    [sym_preproc_line] = STATE(1848),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1849)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(584),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(1849),
    [sym_compiler_directive_decl] = STATE(1849),
    [sym_fsi_directive_decl] = STATE(1849),
    [sym_preproc_line] = STATE(1849),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(1850)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(585),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1850),
    [sym_compiler_directive_decl] = STATE(1850),
    [sym_fsi_directive_decl] = STATE(1850),
    [sym_preproc_line] = STATE(1850),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1851)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(586),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1851),
    [sym_compiler_directive_decl] = STATE(1851),
    [sym_fsi_directive_decl] = STATE(1851),
    [sym_preproc_line] = STATE(1851),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1852)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(382),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1852),
    [sym_compiler_directive_decl] = STATE(1852),
    [sym_fsi_directive_decl] = STATE(1852),
    [sym_preproc_line] = STATE(1852),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1853)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(600),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1853),
    [sym_compiler_directive_decl] = STATE(1853),
    [sym_fsi_directive_decl] = STATE(1853),
    [sym_preproc_line] = STATE(1853),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1854)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(601),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1854),
    [sym_compiler_directive_decl] = STATE(1854),
    [sym_fsi_directive_decl] = STATE(1854),
    [sym_preproc_line] = STATE(1854),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1855)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(602),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1855),
    [sym_compiler_directive_decl] = STATE(1855),
    [sym_fsi_directive_decl] = STATE(1855),
    [sym_preproc_line] = STATE(1855),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1856)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(398),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1856),
    [sym_compiler_directive_decl] = STATE(1856),
    [sym_fsi_directive_decl] = STATE(1856),
    [sym_preproc_line] = STATE(1856),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1857)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(566),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1857),
    [sym_compiler_directive_decl] = STATE(1857),
    [sym_fsi_directive_decl] = STATE(1857),
    [sym_preproc_line] = STATE(1857),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1858)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(604),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1858),
    [sym_compiler_directive_decl] = STATE(1858),
    [sym_fsi_directive_decl] = STATE(1858),
    [sym_preproc_line] = STATE(1858),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1859)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(605),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1859),
    [sym_compiler_directive_decl] = STATE(1859),
    [sym_fsi_directive_decl] = STATE(1859),
    [sym_preproc_line] = STATE(1859),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1860)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(606),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1860),
    [sym_compiler_directive_decl] = STATE(1860),
    [sym_fsi_directive_decl] = STATE(1860),
    [sym_preproc_line] = STATE(1860),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1861)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(401),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1861),
    [sym_compiler_directive_decl] = STATE(1861),
    [sym_fsi_directive_decl] = STATE(1861),
    [sym_preproc_line] = STATE(1861),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1862)] = {
    [sym_function_or_value_defn] = STATE(1344),
    [sym__mutable_expression] = STATE(20691),
    [sym__expression] = STATE(368),
    [sym_literal_expression] = STATE(5827),
    [sym_srtp_call_expression] = STATE(5827),
    [sym__srtp_type_argument] = STATE(22408),
    [sym_long_identifier_or_op] = STATE(5084),
    [sym_tuple_expression] = STATE(5827),
    [sym_brace_expression] = STATE(5827),
    [sym_anon_record_expression] = STATE(5827),
    [sym_prefixed_expression] = STATE(5827),
    [sym_typecast_expression] = STATE(5827),
    [sym_for_expression] = STATE(5827),
    [sym_while_expression] = STATE(5827),
    [sym__if_branch] = STATE(19574),
    [sym_if_expression] = STATE(5827),
    [sym_fun_expression] = STATE(5827),
    [sym_try_expression] = STATE(5827),
    [sym_match_expression] = STATE(5827),
    [sym_function_expression] = STATE(5827),
    [sym_mutate_expression] = STATE(5827),
    [sym_index_expression] = STATE(5084),
    [sym_typed_expression] = STATE(5827),
    [sym_declaration_expression] = STATE(5827),
    [sym_do_expression] = STATE(5827),
    [sym_list_expression] = STATE(5827),
    [sym_array_expression] = STATE(5827),
    [sym_begin_end_expression] = STATE(5827),
    [sym_paren_expression] = STATE(5827),
    [sym__high_prec_app] = STATE(5835),
    [sym__low_prec_app] = STATE(5835),
    [sym_application_expression] = STATE(5827),
    [sym_dot_expression] = STATE(5084),
    [sym_infix_expression] = STATE(5827),
    [sym_ce_expression] = STATE(5827),
    [sym_sequential_expression] = STATE(5827),
    [sym_char] = STATE(5329),
    [sym_format_string] = STATE(5336),
    [sym__string_literal] = STATE(5336),
    [sym_string] = STATE(5329),
    [sym_verbatim_string] = STATE(5329),
    [sym_bytearray] = STATE(5329),
    [sym_verbatim_bytearray] = STATE(5329),
    [sym_format_triple_quoted_string] = STATE(5339),
    [sym_triple_quoted_string] = STATE(5329),
    [sym_const] = STATE(5827),
    [sym_long_identifier] = STATE(5109),
    [sym_active_pattern] = STATE(5185),
    [sym__identifier_or_op] = STATE(5111),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1623),
    [sym_sbyte] = STATE(5329),
    [sym_byte] = STATE(5329),
    [sym_int16] = STATE(5329),
    [sym_uint16] = STATE(5329),
    [sym_int32] = STATE(5329),
    [sym_uint32] = STATE(5329),
    [sym_nativeint] = STATE(5329),
    [sym_unativeint] = STATE(5329),
    [sym_int64] = STATE(5329),
    [sym_uint64] = STATE(5329),
    [sym_ieee32] = STATE(5329),
    [sym_ieee64] = STATE(5329),
    [sym_bignum] = STATE(5329),
    [sym_decimal] = STATE(5329),
    [sym_float] = STATE(4830),
    [sym_block_comment] = STATE(1862),
    [sym_compiler_directive_decl] = STATE(1862),
    [sym_fsi_directive_decl] = STATE(1862),
    [sym_preproc_line] = STATE(1862),
    [sym_preproc_if_in_expression] = STATE(5827),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2194),
    [anon_sym_null] = ACTIONS(2196),
    [anon_sym_begin] = ACTIONS(2198),
    [anon_sym_do] = ACTIONS(2202),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2204),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2210),
    [anon_sym_LBRACE] = ACTIONS(3214),
    [anon_sym_LT_AT] = ACTIONS(2214),
    [anon_sym_LT_AT_AT] = ACTIONS(3216),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2218),
    [anon_sym_new] = ACTIONS(2220),
    [anon_sym_return] = ACTIONS(2220),
    [anon_sym_return_BANG] = ACTIONS(2222),
    [anon_sym_yield] = ACTIONS(2220),
    [anon_sym_yield_BANG] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2220),
    [anon_sym_assert] = ACTIONS(2220),
    [anon_sym_upcast] = ACTIONS(2220),
    [anon_sym_downcast] = ACTIONS(2220),
    [anon_sym_fixed] = ACTIONS(2220),
    [anon_sym_for] = ACTIONS(2226),
    [anon_sym_while] = ACTIONS(2228),
    [anon_sym_while_BANG] = ACTIONS(2230),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2232),
    [anon_sym_try] = ACTIONS(2234),
    [anon_sym_match] = ACTIONS(2236),
    [anon_sym_match_BANG] = ACTIONS(2238),
    [anon_sym_function] = ACTIONS(2240),
    [anon_sym_use] = ACTIONS(2129),
    [anon_sym_use_BANG] = ACTIONS(2131),
    [anon_sym_do_BANG] = ACTIONS(2242),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2247),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(2251),
    [anon_sym_AT_DQUOTE] = ACTIONS(2253),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2257),
    [sym_bool] = ACTIONS(2259),
    [sym_unit] = ACTIONS(3222),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2261),
    [sym_op_identifier] = ACTIONS(3224),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2265),
    [sym_xint] = ACTIONS(2267),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2269),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2271),
  },
  [STATE(1863)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(609),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1863),
    [sym_compiler_directive_decl] = STATE(1863),
    [sym_fsi_directive_decl] = STATE(1863),
    [sym_preproc_line] = STATE(1863),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1864)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(610),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1864),
    [sym_compiler_directive_decl] = STATE(1864),
    [sym_fsi_directive_decl] = STATE(1864),
    [sym_preproc_line] = STATE(1864),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1865)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(611),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1865),
    [sym_compiler_directive_decl] = STATE(1865),
    [sym_fsi_directive_decl] = STATE(1865),
    [sym_preproc_line] = STATE(1865),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1866)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(404),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1866),
    [sym_compiler_directive_decl] = STATE(1866),
    [sym_fsi_directive_decl] = STATE(1866),
    [sym_preproc_line] = STATE(1866),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1867)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(613),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1867),
    [sym_compiler_directive_decl] = STATE(1867),
    [sym_fsi_directive_decl] = STATE(1867),
    [sym_preproc_line] = STATE(1867),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1868)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(614),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1868),
    [sym_compiler_directive_decl] = STATE(1868),
    [sym_fsi_directive_decl] = STATE(1868),
    [sym_preproc_line] = STATE(1868),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1869)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(615),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1869),
    [sym_compiler_directive_decl] = STATE(1869),
    [sym_fsi_directive_decl] = STATE(1869),
    [sym_preproc_line] = STATE(1869),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1870)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(588),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1870),
    [sym_compiler_directive_decl] = STATE(1870),
    [sym_fsi_directive_decl] = STATE(1870),
    [sym_preproc_line] = STATE(1870),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1871)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(406),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1871),
    [sym_compiler_directive_decl] = STATE(1871),
    [sym_fsi_directive_decl] = STATE(1871),
    [sym_preproc_line] = STATE(1871),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1872)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(618),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1872),
    [sym_compiler_directive_decl] = STATE(1872),
    [sym_fsi_directive_decl] = STATE(1872),
    [sym_preproc_line] = STATE(1872),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1873)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(619),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1873),
    [sym_compiler_directive_decl] = STATE(1873),
    [sym_fsi_directive_decl] = STATE(1873),
    [sym_preproc_line] = STATE(1873),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1874)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(620),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1874),
    [sym_compiler_directive_decl] = STATE(1874),
    [sym_fsi_directive_decl] = STATE(1874),
    [sym_preproc_line] = STATE(1874),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1875)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(408),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1875),
    [sym_compiler_directive_decl] = STATE(1875),
    [sym_fsi_directive_decl] = STATE(1875),
    [sym_preproc_line] = STATE(1875),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1876)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(623),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1876),
    [sym_compiler_directive_decl] = STATE(1876),
    [sym_fsi_directive_decl] = STATE(1876),
    [sym_preproc_line] = STATE(1876),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1877)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(624),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1877),
    [sym_compiler_directive_decl] = STATE(1877),
    [sym_fsi_directive_decl] = STATE(1877),
    [sym_preproc_line] = STATE(1877),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1878)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(625),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1878),
    [sym_compiler_directive_decl] = STATE(1878),
    [sym_fsi_directive_decl] = STATE(1878),
    [sym_preproc_line] = STATE(1878),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1879)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(410),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1879),
    [sym_compiler_directive_decl] = STATE(1879),
    [sym_fsi_directive_decl] = STATE(1879),
    [sym_preproc_line] = STATE(1879),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1880)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(627),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1880),
    [sym_compiler_directive_decl] = STATE(1880),
    [sym_fsi_directive_decl] = STATE(1880),
    [sym_preproc_line] = STATE(1880),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1881)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(628),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1881),
    [sym_compiler_directive_decl] = STATE(1881),
    [sym_fsi_directive_decl] = STATE(1881),
    [sym_preproc_line] = STATE(1881),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1882)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(629),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1882),
    [sym_compiler_directive_decl] = STATE(1882),
    [sym_fsi_directive_decl] = STATE(1882),
    [sym_preproc_line] = STATE(1882),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1883)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(589),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1883),
    [sym_compiler_directive_decl] = STATE(1883),
    [sym_fsi_directive_decl] = STATE(1883),
    [sym_preproc_line] = STATE(1883),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1884)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(412),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1884),
    [sym_compiler_directive_decl] = STATE(1884),
    [sym_fsi_directive_decl] = STATE(1884),
    [sym_preproc_line] = STATE(1884),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1885)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(631),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1885),
    [sym_compiler_directive_decl] = STATE(1885),
    [sym_fsi_directive_decl] = STATE(1885),
    [sym_preproc_line] = STATE(1885),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1886)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(632),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1886),
    [sym_compiler_directive_decl] = STATE(1886),
    [sym_fsi_directive_decl] = STATE(1886),
    [sym_preproc_line] = STATE(1886),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1887)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(633),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1887),
    [sym_compiler_directive_decl] = STATE(1887),
    [sym_fsi_directive_decl] = STATE(1887),
    [sym_preproc_line] = STATE(1887),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1888)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(414),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1888),
    [sym_compiler_directive_decl] = STATE(1888),
    [sym_fsi_directive_decl] = STATE(1888),
    [sym_preproc_line] = STATE(1888),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1889)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(634),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1889),
    [sym_compiler_directive_decl] = STATE(1889),
    [sym_fsi_directive_decl] = STATE(1889),
    [sym_preproc_line] = STATE(1889),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1890)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(635),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1890),
    [sym_compiler_directive_decl] = STATE(1890),
    [sym_fsi_directive_decl] = STATE(1890),
    [sym_preproc_line] = STATE(1890),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1891)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(636),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1891),
    [sym_compiler_directive_decl] = STATE(1891),
    [sym_fsi_directive_decl] = STATE(1891),
    [sym_preproc_line] = STATE(1891),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1892)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(416),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1892),
    [sym_compiler_directive_decl] = STATE(1892),
    [sym_fsi_directive_decl] = STATE(1892),
    [sym_preproc_line] = STATE(1892),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1893)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(638),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1893),
    [sym_compiler_directive_decl] = STATE(1893),
    [sym_fsi_directive_decl] = STATE(1893),
    [sym_preproc_line] = STATE(1893),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1894)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(639),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1894),
    [sym_compiler_directive_decl] = STATE(1894),
    [sym_fsi_directive_decl] = STATE(1894),
    [sym_preproc_line] = STATE(1894),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1895)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(640),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1895),
    [sym_compiler_directive_decl] = STATE(1895),
    [sym_fsi_directive_decl] = STATE(1895),
    [sym_preproc_line] = STATE(1895),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1896)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(418),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1896),
    [sym_compiler_directive_decl] = STATE(1896),
    [sym_fsi_directive_decl] = STATE(1896),
    [sym_preproc_line] = STATE(1896),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1897)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(643),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1897),
    [sym_compiler_directive_decl] = STATE(1897),
    [sym_fsi_directive_decl] = STATE(1897),
    [sym_preproc_line] = STATE(1897),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1898)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(644),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1898),
    [sym_compiler_directive_decl] = STATE(1898),
    [sym_fsi_directive_decl] = STATE(1898),
    [sym_preproc_line] = STATE(1898),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1899)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(645),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1899),
    [sym_compiler_directive_decl] = STATE(1899),
    [sym_fsi_directive_decl] = STATE(1899),
    [sym_preproc_line] = STATE(1899),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1900)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(420),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1900),
    [sym_compiler_directive_decl] = STATE(1900),
    [sym_fsi_directive_decl] = STATE(1900),
    [sym_preproc_line] = STATE(1900),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1901)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(648),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1901),
    [sym_compiler_directive_decl] = STATE(1901),
    [sym_fsi_directive_decl] = STATE(1901),
    [sym_preproc_line] = STATE(1901),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1902)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(649),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1902),
    [sym_compiler_directive_decl] = STATE(1902),
    [sym_fsi_directive_decl] = STATE(1902),
    [sym_preproc_line] = STATE(1902),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1903)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(650),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1903),
    [sym_compiler_directive_decl] = STATE(1903),
    [sym_fsi_directive_decl] = STATE(1903),
    [sym_preproc_line] = STATE(1903),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1904)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(422),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1904),
    [sym_compiler_directive_decl] = STATE(1904),
    [sym_fsi_directive_decl] = STATE(1904),
    [sym_preproc_line] = STATE(1904),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1905)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(495),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1905),
    [sym_compiler_directive_decl] = STATE(1905),
    [sym_fsi_directive_decl] = STATE(1905),
    [sym_preproc_line] = STATE(1905),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1906)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(654),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1906),
    [sym_compiler_directive_decl] = STATE(1906),
    [sym_fsi_directive_decl] = STATE(1906),
    [sym_preproc_line] = STATE(1906),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1907)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(655),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1907),
    [sym_compiler_directive_decl] = STATE(1907),
    [sym_fsi_directive_decl] = STATE(1907),
    [sym_preproc_line] = STATE(1907),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1908)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(425),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1908),
    [sym_compiler_directive_decl] = STATE(1908),
    [sym_fsi_directive_decl] = STATE(1908),
    [sym_preproc_line] = STATE(1908),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1909)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(658),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1909),
    [sym_compiler_directive_decl] = STATE(1909),
    [sym_fsi_directive_decl] = STATE(1909),
    [sym_preproc_line] = STATE(1909),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1910)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(659),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1910),
    [sym_compiler_directive_decl] = STATE(1910),
    [sym_fsi_directive_decl] = STATE(1910),
    [sym_preproc_line] = STATE(1910),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1911)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(660),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1911),
    [sym_compiler_directive_decl] = STATE(1911),
    [sym_fsi_directive_decl] = STATE(1911),
    [sym_preproc_line] = STATE(1911),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1912)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(427),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1912),
    [sym_compiler_directive_decl] = STATE(1912),
    [sym_fsi_directive_decl] = STATE(1912),
    [sym_preproc_line] = STATE(1912),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1913)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(663),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1913),
    [sym_compiler_directive_decl] = STATE(1913),
    [sym_fsi_directive_decl] = STATE(1913),
    [sym_preproc_line] = STATE(1913),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1914)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(664),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1914),
    [sym_compiler_directive_decl] = STATE(1914),
    [sym_fsi_directive_decl] = STATE(1914),
    [sym_preproc_line] = STATE(1914),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1915)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(665),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1915),
    [sym_compiler_directive_decl] = STATE(1915),
    [sym_fsi_directive_decl] = STATE(1915),
    [sym_preproc_line] = STATE(1915),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1916)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(235),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1916),
    [sym_compiler_directive_decl] = STATE(1916),
    [sym_fsi_directive_decl] = STATE(1916),
    [sym_preproc_line] = STATE(1916),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1917)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(429),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1917),
    [sym_compiler_directive_decl] = STATE(1917),
    [sym_fsi_directive_decl] = STATE(1917),
    [sym_preproc_line] = STATE(1917),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1918)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(668),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1918),
    [sym_compiler_directive_decl] = STATE(1918),
    [sym_fsi_directive_decl] = STATE(1918),
    [sym_preproc_line] = STATE(1918),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1919)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(669),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1919),
    [sym_compiler_directive_decl] = STATE(1919),
    [sym_fsi_directive_decl] = STATE(1919),
    [sym_preproc_line] = STATE(1919),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1920)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(670),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1920),
    [sym_compiler_directive_decl] = STATE(1920),
    [sym_fsi_directive_decl] = STATE(1920),
    [sym_preproc_line] = STATE(1920),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1921)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(431),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1921),
    [sym_compiler_directive_decl] = STATE(1921),
    [sym_fsi_directive_decl] = STATE(1921),
    [sym_preproc_line] = STATE(1921),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1922)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(671),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1922),
    [sym_compiler_directive_decl] = STATE(1922),
    [sym_fsi_directive_decl] = STATE(1922),
    [sym_preproc_line] = STATE(1922),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1923)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(672),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1923),
    [sym_compiler_directive_decl] = STATE(1923),
    [sym_fsi_directive_decl] = STATE(1923),
    [sym_preproc_line] = STATE(1923),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1924)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(673),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1924),
    [sym_compiler_directive_decl] = STATE(1924),
    [sym_fsi_directive_decl] = STATE(1924),
    [sym_preproc_line] = STATE(1924),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1925)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(805),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1925),
    [sym_compiler_directive_decl] = STATE(1925),
    [sym_fsi_directive_decl] = STATE(1925),
    [sym_preproc_line] = STATE(1925),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1926)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(433),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1926),
    [sym_compiler_directive_decl] = STATE(1926),
    [sym_fsi_directive_decl] = STATE(1926),
    [sym_preproc_line] = STATE(1926),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1927)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(675),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1927),
    [sym_compiler_directive_decl] = STATE(1927),
    [sym_fsi_directive_decl] = STATE(1927),
    [sym_preproc_line] = STATE(1927),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1928)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(676),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1928),
    [sym_compiler_directive_decl] = STATE(1928),
    [sym_fsi_directive_decl] = STATE(1928),
    [sym_preproc_line] = STATE(1928),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1929)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(677),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1929),
    [sym_compiler_directive_decl] = STATE(1929),
    [sym_fsi_directive_decl] = STATE(1929),
    [sym_preproc_line] = STATE(1929),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1930)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(435),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1930),
    [sym_compiler_directive_decl] = STATE(1930),
    [sym_fsi_directive_decl] = STATE(1930),
    [sym_preproc_line] = STATE(1930),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1931)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(679),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1931),
    [sym_compiler_directive_decl] = STATE(1931),
    [sym_fsi_directive_decl] = STATE(1931),
    [sym_preproc_line] = STATE(1931),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1932)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(680),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1932),
    [sym_compiler_directive_decl] = STATE(1932),
    [sym_fsi_directive_decl] = STATE(1932),
    [sym_preproc_line] = STATE(1932),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1933)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(681),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1933),
    [sym_compiler_directive_decl] = STATE(1933),
    [sym_fsi_directive_decl] = STATE(1933),
    [sym_preproc_line] = STATE(1933),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1934)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(437),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1934),
    [sym_compiler_directive_decl] = STATE(1934),
    [sym_fsi_directive_decl] = STATE(1934),
    [sym_preproc_line] = STATE(1934),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1935)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(684),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1935),
    [sym_compiler_directive_decl] = STATE(1935),
    [sym_fsi_directive_decl] = STATE(1935),
    [sym_preproc_line] = STATE(1935),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1936)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(685),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1936),
    [sym_compiler_directive_decl] = STATE(1936),
    [sym_fsi_directive_decl] = STATE(1936),
    [sym_preproc_line] = STATE(1936),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1937)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(686),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1937),
    [sym_compiler_directive_decl] = STATE(1937),
    [sym_fsi_directive_decl] = STATE(1937),
    [sym_preproc_line] = STATE(1937),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1938)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(439),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1938),
    [sym_compiler_directive_decl] = STATE(1938),
    [sym_fsi_directive_decl] = STATE(1938),
    [sym_preproc_line] = STATE(1938),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1939)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(689),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1939),
    [sym_compiler_directive_decl] = STATE(1939),
    [sym_fsi_directive_decl] = STATE(1939),
    [sym_preproc_line] = STATE(1939),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1940)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(690),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1940),
    [sym_compiler_directive_decl] = STATE(1940),
    [sym_fsi_directive_decl] = STATE(1940),
    [sym_preproc_line] = STATE(1940),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1941)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(691),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1941),
    [sym_compiler_directive_decl] = STATE(1941),
    [sym_fsi_directive_decl] = STATE(1941),
    [sym_preproc_line] = STATE(1941),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1942)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(441),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1942),
    [sym_compiler_directive_decl] = STATE(1942),
    [sym_fsi_directive_decl] = STATE(1942),
    [sym_preproc_line] = STATE(1942),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1943)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(694),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1943),
    [sym_compiler_directive_decl] = STATE(1943),
    [sym_fsi_directive_decl] = STATE(1943),
    [sym_preproc_line] = STATE(1943),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1944)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(695),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1944),
    [sym_compiler_directive_decl] = STATE(1944),
    [sym_fsi_directive_decl] = STATE(1944),
    [sym_preproc_line] = STATE(1944),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1945)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(696),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1945),
    [sym_compiler_directive_decl] = STATE(1945),
    [sym_fsi_directive_decl] = STATE(1945),
    [sym_preproc_line] = STATE(1945),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1946)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(443),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1946),
    [sym_compiler_directive_decl] = STATE(1946),
    [sym_fsi_directive_decl] = STATE(1946),
    [sym_preproc_line] = STATE(1946),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1947)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(699),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1947),
    [sym_compiler_directive_decl] = STATE(1947),
    [sym_fsi_directive_decl] = STATE(1947),
    [sym_preproc_line] = STATE(1947),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1948)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(700),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1948),
    [sym_compiler_directive_decl] = STATE(1948),
    [sym_fsi_directive_decl] = STATE(1948),
    [sym_preproc_line] = STATE(1948),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1949)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(701),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1949),
    [sym_compiler_directive_decl] = STATE(1949),
    [sym_fsi_directive_decl] = STATE(1949),
    [sym_preproc_line] = STATE(1949),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1950)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(445),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1950),
    [sym_compiler_directive_decl] = STATE(1950),
    [sym_fsi_directive_decl] = STATE(1950),
    [sym_preproc_line] = STATE(1950),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1951)] = {
    [sym_function_or_value_defn] = STATE(1290),
    [sym__mutable_expression] = STATE(21243),
    [sym__expression] = STATE(704),
    [sym_literal_expression] = STATE(6210),
    [sym_srtp_call_expression] = STATE(6210),
    [sym__srtp_type_argument] = STATE(22178),
    [sym_long_identifier_or_op] = STATE(5737),
    [sym_tuple_expression] = STATE(6210),
    [sym_brace_expression] = STATE(6210),
    [sym_anon_record_expression] = STATE(6210),
    [sym_prefixed_expression] = STATE(6210),
    [sym_typecast_expression] = STATE(6210),
    [sym_for_expression] = STATE(6210),
    [sym_while_expression] = STATE(6210),
    [sym__if_branch] = STATE(21857),
    [sym_if_expression] = STATE(6210),
    [sym_fun_expression] = STATE(6210),
    [sym_try_expression] = STATE(6210),
    [sym_match_expression] = STATE(6210),
    [sym_function_expression] = STATE(6210),
    [sym_mutate_expression] = STATE(6210),
    [sym_index_expression] = STATE(5737),
    [sym_typed_expression] = STATE(6210),
    [sym_declaration_expression] = STATE(6210),
    [sym_do_expression] = STATE(6210),
    [sym_list_expression] = STATE(6210),
    [sym_array_expression] = STATE(6210),
    [sym_begin_end_expression] = STATE(6210),
    [sym_paren_expression] = STATE(6210),
    [sym__high_prec_app] = STATE(6332),
    [sym__low_prec_app] = STATE(6332),
    [sym_application_expression] = STATE(6210),
    [sym_dot_expression] = STATE(5737),
    [sym_infix_expression] = STATE(6210),
    [sym_ce_expression] = STATE(6210),
    [sym_sequential_expression] = STATE(6210),
    [sym_char] = STATE(6067),
    [sym_format_string] = STATE(6412),
    [sym__string_literal] = STATE(6412),
    [sym_string] = STATE(6067),
    [sym_verbatim_string] = STATE(6067),
    [sym_bytearray] = STATE(6067),
    [sym_verbatim_bytearray] = STATE(6067),
    [sym_format_triple_quoted_string] = STATE(6419),
    [sym_triple_quoted_string] = STATE(6067),
    [sym_const] = STATE(6210),
    [sym_long_identifier] = STATE(5407),
    [sym_active_pattern] = STATE(5390),
    [sym__identifier_or_op] = STATE(5558),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1618),
    [sym_sbyte] = STATE(6067),
    [sym_byte] = STATE(6067),
    [sym_int16] = STATE(6067),
    [sym_uint16] = STATE(6067),
    [sym_int32] = STATE(6067),
    [sym_uint32] = STATE(6067),
    [sym_nativeint] = STATE(6067),
    [sym_unativeint] = STATE(6067),
    [sym_int64] = STATE(6067),
    [sym_uint64] = STATE(6067),
    [sym_ieee32] = STATE(6067),
    [sym_ieee64] = STATE(6067),
    [sym_bignum] = STATE(6067),
    [sym_decimal] = STATE(6067),
    [sym_float] = STATE(5303),
    [sym_block_comment] = STATE(1951),
    [sym_compiler_directive_decl] = STATE(1951),
    [sym_fsi_directive_decl] = STATE(1951),
    [sym_preproc_line] = STATE(1951),
    [sym_preproc_if_in_expression] = STATE(6210),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2584),
    [anon_sym_null] = ACTIONS(2586),
    [anon_sym_begin] = ACTIONS(2588),
    [anon_sym_do] = ACTIONS(2592),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2594),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2598),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2600),
    [anon_sym_LBRACE] = ACTIONS(5528),
    [anon_sym_LT_AT] = ACTIONS(2604),
    [anon_sym_LT_AT_AT] = ACTIONS(5530),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2612),
    [anon_sym_new] = ACTIONS(2614),
    [anon_sym_return] = ACTIONS(2614),
    [anon_sym_return_BANG] = ACTIONS(2616),
    [anon_sym_yield] = ACTIONS(2614),
    [anon_sym_yield_BANG] = ACTIONS(2616),
    [anon_sym_lazy] = ACTIONS(2614),
    [anon_sym_assert] = ACTIONS(2614),
    [anon_sym_upcast] = ACTIONS(2614),
    [anon_sym_downcast] = ACTIONS(2614),
    [anon_sym_fixed] = ACTIONS(2614),
    [anon_sym_for] = ACTIONS(2620),
    [anon_sym_while] = ACTIONS(2622),
    [anon_sym_while_BANG] = ACTIONS(2624),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2626),
    [anon_sym_try] = ACTIONS(2628),
    [anon_sym_match] = ACTIONS(2630),
    [anon_sym_match_BANG] = ACTIONS(2632),
    [anon_sym_function] = ACTIONS(2634),
    [anon_sym_use] = ACTIONS(2638),
    [anon_sym_use_BANG] = ACTIONS(2640),
    [anon_sym_do_BANG] = ACTIONS(2642),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2648),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2652),
    [anon_sym_AT_DQUOTE] = ACTIONS(2654),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2656),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2658),
    [sym_bool] = ACTIONS(2660),
    [sym_unit] = ACTIONS(5532),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2664),
    [sym_op_identifier] = ACTIONS(5534),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2668),
    [sym_xint] = ACTIONS(2670),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2672),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2678),
  },
  [STATE(1952)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(705),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(1952),
    [sym_compiler_directive_decl] = STATE(1952),
    [sym_fsi_directive_decl] = STATE(1952),
    [sym_preproc_line] = STATE(1952),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(1953)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(706),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1953),
    [sym_compiler_directive_decl] = STATE(1953),
    [sym_fsi_directive_decl] = STATE(1953),
    [sym_preproc_line] = STATE(1953),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1954)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(447),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1954),
    [sym_compiler_directive_decl] = STATE(1954),
    [sym_fsi_directive_decl] = STATE(1954),
    [sym_preproc_line] = STATE(1954),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1955)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(708),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1955),
    [sym_compiler_directive_decl] = STATE(1955),
    [sym_fsi_directive_decl] = STATE(1955),
    [sym_preproc_line] = STATE(1955),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1956)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(448),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1956),
    [sym_compiler_directive_decl] = STATE(1956),
    [sym_fsi_directive_decl] = STATE(1956),
    [sym_preproc_line] = STATE(1956),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1957)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(709),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1957),
    [sym_compiler_directive_decl] = STATE(1957),
    [sym_fsi_directive_decl] = STATE(1957),
    [sym_preproc_line] = STATE(1957),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1958)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(449),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1958),
    [sym_compiler_directive_decl] = STATE(1958),
    [sym_fsi_directive_decl] = STATE(1958),
    [sym_preproc_line] = STATE(1958),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1959)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(710),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1959),
    [sym_compiler_directive_decl] = STATE(1959),
    [sym_fsi_directive_decl] = STATE(1959),
    [sym_preproc_line] = STATE(1959),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1960)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(450),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1960),
    [sym_compiler_directive_decl] = STATE(1960),
    [sym_fsi_directive_decl] = STATE(1960),
    [sym_preproc_line] = STATE(1960),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1961)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(227),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(1961),
    [sym_compiler_directive_decl] = STATE(1961),
    [sym_fsi_directive_decl] = STATE(1961),
    [sym_preproc_line] = STATE(1961),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(1962)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(711),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1962),
    [sym_compiler_directive_decl] = STATE(1962),
    [sym_fsi_directive_decl] = STATE(1962),
    [sym_preproc_line] = STATE(1962),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1963)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(451),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1963),
    [sym_compiler_directive_decl] = STATE(1963),
    [sym_fsi_directive_decl] = STATE(1963),
    [sym_preproc_line] = STATE(1963),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1964)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(713),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1964),
    [sym_compiler_directive_decl] = STATE(1964),
    [sym_fsi_directive_decl] = STATE(1964),
    [sym_preproc_line] = STATE(1964),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1965)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(452),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1965),
    [sym_compiler_directive_decl] = STATE(1965),
    [sym_fsi_directive_decl] = STATE(1965),
    [sym_preproc_line] = STATE(1965),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1966)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(715),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1966),
    [sym_compiler_directive_decl] = STATE(1966),
    [sym_fsi_directive_decl] = STATE(1966),
    [sym_preproc_line] = STATE(1966),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1967)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(453),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1967),
    [sym_compiler_directive_decl] = STATE(1967),
    [sym_fsi_directive_decl] = STATE(1967),
    [sym_preproc_line] = STATE(1967),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1968)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(716),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1968),
    [sym_compiler_directive_decl] = STATE(1968),
    [sym_fsi_directive_decl] = STATE(1968),
    [sym_preproc_line] = STATE(1968),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1969)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(286),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1969),
    [sym_compiler_directive_decl] = STATE(1969),
    [sym_fsi_directive_decl] = STATE(1969),
    [sym_preproc_line] = STATE(1969),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1970)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(717),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1970),
    [sym_compiler_directive_decl] = STATE(1970),
    [sym_fsi_directive_decl] = STATE(1970),
    [sym_preproc_line] = STATE(1970),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1971)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(455),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1971),
    [sym_compiler_directive_decl] = STATE(1971),
    [sym_fsi_directive_decl] = STATE(1971),
    [sym_preproc_line] = STATE(1971),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1972)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(718),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1972),
    [sym_compiler_directive_decl] = STATE(1972),
    [sym_fsi_directive_decl] = STATE(1972),
    [sym_preproc_line] = STATE(1972),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1973)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(456),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1973),
    [sym_compiler_directive_decl] = STATE(1973),
    [sym_fsi_directive_decl] = STATE(1973),
    [sym_preproc_line] = STATE(1973),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1974)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(719),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1974),
    [sym_compiler_directive_decl] = STATE(1974),
    [sym_fsi_directive_decl] = STATE(1974),
    [sym_preproc_line] = STATE(1974),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1975)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(457),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1975),
    [sym_compiler_directive_decl] = STATE(1975),
    [sym_fsi_directive_decl] = STATE(1975),
    [sym_preproc_line] = STATE(1975),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1976)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(721),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1976),
    [sym_compiler_directive_decl] = STATE(1976),
    [sym_fsi_directive_decl] = STATE(1976),
    [sym_preproc_line] = STATE(1976),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1977)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(458),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1977),
    [sym_compiler_directive_decl] = STATE(1977),
    [sym_fsi_directive_decl] = STATE(1977),
    [sym_preproc_line] = STATE(1977),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1978)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(723),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1978),
    [sym_compiler_directive_decl] = STATE(1978),
    [sym_fsi_directive_decl] = STATE(1978),
    [sym_preproc_line] = STATE(1978),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1979)] = {
    [sym_function_or_value_defn] = STATE(1384),
    [sym__mutable_expression] = STATE(22476),
    [sym__expression] = STATE(459),
    [sym_literal_expression] = STATE(5576),
    [sym_srtp_call_expression] = STATE(5576),
    [sym__srtp_type_argument] = STATE(22436),
    [sym_long_identifier_or_op] = STATE(5201),
    [sym_tuple_expression] = STATE(5576),
    [sym_brace_expression] = STATE(5576),
    [sym_anon_record_expression] = STATE(5576),
    [sym_prefixed_expression] = STATE(5576),
    [sym_typecast_expression] = STATE(5576),
    [sym_for_expression] = STATE(5576),
    [sym_while_expression] = STATE(5576),
    [sym__if_branch] = STATE(19795),
    [sym_if_expression] = STATE(5576),
    [sym_fun_expression] = STATE(5576),
    [sym_try_expression] = STATE(5576),
    [sym_match_expression] = STATE(5576),
    [sym_function_expression] = STATE(5576),
    [sym_mutate_expression] = STATE(5576),
    [sym_index_expression] = STATE(5201),
    [sym_typed_expression] = STATE(5576),
    [sym_declaration_expression] = STATE(5576),
    [sym_do_expression] = STATE(5576),
    [sym_list_expression] = STATE(5576),
    [sym_array_expression] = STATE(5576),
    [sym_begin_end_expression] = STATE(5576),
    [sym_paren_expression] = STATE(5576),
    [sym__high_prec_app] = STATE(5581),
    [sym__low_prec_app] = STATE(5581),
    [sym_application_expression] = STATE(5576),
    [sym_dot_expression] = STATE(5201),
    [sym_infix_expression] = STATE(5576),
    [sym_ce_expression] = STATE(5576),
    [sym_sequential_expression] = STATE(5576),
    [sym_char] = STATE(5882),
    [sym_format_string] = STATE(5355),
    [sym__string_literal] = STATE(5355),
    [sym_string] = STATE(5882),
    [sym_verbatim_string] = STATE(5882),
    [sym_bytearray] = STATE(5882),
    [sym_verbatim_bytearray] = STATE(5882),
    [sym_format_triple_quoted_string] = STATE(5356),
    [sym_triple_quoted_string] = STATE(5882),
    [sym_const] = STATE(5576),
    [sym_long_identifier] = STATE(5063),
    [sym_active_pattern] = STATE(4980),
    [sym__identifier_or_op] = STATE(5099),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1639),
    [sym_sbyte] = STATE(5882),
    [sym_byte] = STATE(5882),
    [sym_int16] = STATE(5882),
    [sym_uint16] = STATE(5882),
    [sym_int32] = STATE(5882),
    [sym_uint32] = STATE(5882),
    [sym_nativeint] = STATE(5882),
    [sym_unativeint] = STATE(5882),
    [sym_int64] = STATE(5882),
    [sym_uint64] = STATE(5882),
    [sym_ieee32] = STATE(5882),
    [sym_ieee64] = STATE(5882),
    [sym_bignum] = STATE(5882),
    [sym_decimal] = STATE(5882),
    [sym_float] = STATE(4911),
    [sym_block_comment] = STATE(1979),
    [sym_compiler_directive_decl] = STATE(1979),
    [sym_fsi_directive_decl] = STATE(1979),
    [sym_preproc_line] = STATE(1979),
    [sym_preproc_if_in_expression] = STATE(5576),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1867),
    [anon_sym_null] = ACTIONS(1869),
    [anon_sym_begin] = ACTIONS(1871),
    [anon_sym_do] = ACTIONS(1875),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1877),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1883),
    [anon_sym_LBRACE] = ACTIONS(5759),
    [anon_sym_LT_AT] = ACTIONS(1887),
    [anon_sym_LT_AT_AT] = ACTIONS(5761),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1897),
    [anon_sym_return] = ACTIONS(1897),
    [anon_sym_return_BANG] = ACTIONS(1899),
    [anon_sym_yield] = ACTIONS(1897),
    [anon_sym_yield_BANG] = ACTIONS(1899),
    [anon_sym_lazy] = ACTIONS(1897),
    [anon_sym_assert] = ACTIONS(1897),
    [anon_sym_upcast] = ACTIONS(1897),
    [anon_sym_downcast] = ACTIONS(1897),
    [anon_sym_fixed] = ACTIONS(1897),
    [anon_sym_for] = ACTIONS(1903),
    [anon_sym_while] = ACTIONS(1905),
    [anon_sym_while_BANG] = ACTIONS(1907),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1909),
    [anon_sym_try] = ACTIONS(1911),
    [anon_sym_match] = ACTIONS(1913),
    [anon_sym_match_BANG] = ACTIONS(1915),
    [anon_sym_function] = ACTIONS(1917),
    [anon_sym_use] = ACTIONS(1921),
    [anon_sym_use_BANG] = ACTIONS(1923),
    [anon_sym_do_BANG] = ACTIONS(1925),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1931),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1935),
    [anon_sym_AT_DQUOTE] = ACTIONS(1937),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1939),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1941),
    [sym_bool] = ACTIONS(1943),
    [sym_unit] = ACTIONS(5763),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1947),
    [sym_op_identifier] = ACTIONS(5765),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1951),
    [sym_xint] = ACTIONS(1953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1955),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1961),
  },
  [STATE(1980)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(725),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1980),
    [sym_compiler_directive_decl] = STATE(1980),
    [sym_fsi_directive_decl] = STATE(1980),
    [sym_preproc_line] = STATE(1980),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1981)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(726),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1981),
    [sym_compiler_directive_decl] = STATE(1981),
    [sym_fsi_directive_decl] = STATE(1981),
    [sym_preproc_line] = STATE(1981),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1982)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(727),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1982),
    [sym_compiler_directive_decl] = STATE(1982),
    [sym_fsi_directive_decl] = STATE(1982),
    [sym_preproc_line] = STATE(1982),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1983)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(728),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1983),
    [sym_compiler_directive_decl] = STATE(1983),
    [sym_fsi_directive_decl] = STATE(1983),
    [sym_preproc_line] = STATE(1983),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1984)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(729),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1984),
    [sym_compiler_directive_decl] = STATE(1984),
    [sym_fsi_directive_decl] = STATE(1984),
    [sym_preproc_line] = STATE(1984),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1985)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(730),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1985),
    [sym_compiler_directive_decl] = STATE(1985),
    [sym_fsi_directive_decl] = STATE(1985),
    [sym_preproc_line] = STATE(1985),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1986)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(731),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1986),
    [sym_compiler_directive_decl] = STATE(1986),
    [sym_fsi_directive_decl] = STATE(1986),
    [sym_preproc_line] = STATE(1986),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1987)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(732),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1987),
    [sym_compiler_directive_decl] = STATE(1987),
    [sym_fsi_directive_decl] = STATE(1987),
    [sym_preproc_line] = STATE(1987),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1988)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(733),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1988),
    [sym_compiler_directive_decl] = STATE(1988),
    [sym_fsi_directive_decl] = STATE(1988),
    [sym_preproc_line] = STATE(1988),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1989)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(734),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1989),
    [sym_compiler_directive_decl] = STATE(1989),
    [sym_fsi_directive_decl] = STATE(1989),
    [sym_preproc_line] = STATE(1989),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1990)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(735),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1990),
    [sym_compiler_directive_decl] = STATE(1990),
    [sym_fsi_directive_decl] = STATE(1990),
    [sym_preproc_line] = STATE(1990),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1991)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(736),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1991),
    [sym_compiler_directive_decl] = STATE(1991),
    [sym_fsi_directive_decl] = STATE(1991),
    [sym_preproc_line] = STATE(1991),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1992)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(737),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1992),
    [sym_compiler_directive_decl] = STATE(1992),
    [sym_fsi_directive_decl] = STATE(1992),
    [sym_preproc_line] = STATE(1992),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1993)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(738),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1993),
    [sym_compiler_directive_decl] = STATE(1993),
    [sym_fsi_directive_decl] = STATE(1993),
    [sym_preproc_line] = STATE(1993),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1994)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(739),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1994),
    [sym_compiler_directive_decl] = STATE(1994),
    [sym_fsi_directive_decl] = STATE(1994),
    [sym_preproc_line] = STATE(1994),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1995)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(740),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1995),
    [sym_compiler_directive_decl] = STATE(1995),
    [sym_fsi_directive_decl] = STATE(1995),
    [sym_preproc_line] = STATE(1995),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1996)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(741),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1996),
    [sym_compiler_directive_decl] = STATE(1996),
    [sym_fsi_directive_decl] = STATE(1996),
    [sym_preproc_line] = STATE(1996),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1997)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(742),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1997),
    [sym_compiler_directive_decl] = STATE(1997),
    [sym_fsi_directive_decl] = STATE(1997),
    [sym_preproc_line] = STATE(1997),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1998)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(743),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1998),
    [sym_compiler_directive_decl] = STATE(1998),
    [sym_fsi_directive_decl] = STATE(1998),
    [sym_preproc_line] = STATE(1998),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(1999)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(744),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(1999),
    [sym_compiler_directive_decl] = STATE(1999),
    [sym_fsi_directive_decl] = STATE(1999),
    [sym_preproc_line] = STATE(1999),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2000)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(745),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2000),
    [sym_compiler_directive_decl] = STATE(2000),
    [sym_fsi_directive_decl] = STATE(2000),
    [sym_preproc_line] = STATE(2000),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2001)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(746),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2001),
    [sym_compiler_directive_decl] = STATE(2001),
    [sym_fsi_directive_decl] = STATE(2001),
    [sym_preproc_line] = STATE(2001),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2002)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(747),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2002),
    [sym_compiler_directive_decl] = STATE(2002),
    [sym_fsi_directive_decl] = STATE(2002),
    [sym_preproc_line] = STATE(2002),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2003)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(748),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2003),
    [sym_compiler_directive_decl] = STATE(2003),
    [sym_fsi_directive_decl] = STATE(2003),
    [sym_preproc_line] = STATE(2003),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2004)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(749),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2004),
    [sym_compiler_directive_decl] = STATE(2004),
    [sym_fsi_directive_decl] = STATE(2004),
    [sym_preproc_line] = STATE(2004),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2005)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(750),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2005),
    [sym_compiler_directive_decl] = STATE(2005),
    [sym_fsi_directive_decl] = STATE(2005),
    [sym_preproc_line] = STATE(2005),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2006)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(751),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2006),
    [sym_compiler_directive_decl] = STATE(2006),
    [sym_fsi_directive_decl] = STATE(2006),
    [sym_preproc_line] = STATE(2006),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2007)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(752),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2007),
    [sym_compiler_directive_decl] = STATE(2007),
    [sym_fsi_directive_decl] = STATE(2007),
    [sym_preproc_line] = STATE(2007),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2008)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(753),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2008),
    [sym_compiler_directive_decl] = STATE(2008),
    [sym_fsi_directive_decl] = STATE(2008),
    [sym_preproc_line] = STATE(2008),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2009)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(754),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2009),
    [sym_compiler_directive_decl] = STATE(2009),
    [sym_fsi_directive_decl] = STATE(2009),
    [sym_preproc_line] = STATE(2009),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2010)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(755),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2010),
    [sym_compiler_directive_decl] = STATE(2010),
    [sym_fsi_directive_decl] = STATE(2010),
    [sym_preproc_line] = STATE(2010),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2011)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(756),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2011),
    [sym_compiler_directive_decl] = STATE(2011),
    [sym_fsi_directive_decl] = STATE(2011),
    [sym_preproc_line] = STATE(2011),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2012)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(757),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2012),
    [sym_compiler_directive_decl] = STATE(2012),
    [sym_fsi_directive_decl] = STATE(2012),
    [sym_preproc_line] = STATE(2012),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2013)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(758),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2013),
    [sym_compiler_directive_decl] = STATE(2013),
    [sym_fsi_directive_decl] = STATE(2013),
    [sym_preproc_line] = STATE(2013),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2014)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(759),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2014),
    [sym_compiler_directive_decl] = STATE(2014),
    [sym_fsi_directive_decl] = STATE(2014),
    [sym_preproc_line] = STATE(2014),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2015)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(760),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2015),
    [sym_compiler_directive_decl] = STATE(2015),
    [sym_fsi_directive_decl] = STATE(2015),
    [sym_preproc_line] = STATE(2015),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2016)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(761),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2016),
    [sym_compiler_directive_decl] = STATE(2016),
    [sym_fsi_directive_decl] = STATE(2016),
    [sym_preproc_line] = STATE(2016),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2017)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(762),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2017),
    [sym_compiler_directive_decl] = STATE(2017),
    [sym_fsi_directive_decl] = STATE(2017),
    [sym_preproc_line] = STATE(2017),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2018)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(763),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2018),
    [sym_compiler_directive_decl] = STATE(2018),
    [sym_fsi_directive_decl] = STATE(2018),
    [sym_preproc_line] = STATE(2018),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2019)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(764),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2019),
    [sym_compiler_directive_decl] = STATE(2019),
    [sym_fsi_directive_decl] = STATE(2019),
    [sym_preproc_line] = STATE(2019),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2020)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(765),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2020),
    [sym_compiler_directive_decl] = STATE(2020),
    [sym_fsi_directive_decl] = STATE(2020),
    [sym_preproc_line] = STATE(2020),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2021)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(766),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2021),
    [sym_compiler_directive_decl] = STATE(2021),
    [sym_fsi_directive_decl] = STATE(2021),
    [sym_preproc_line] = STATE(2021),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2022)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(767),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2022),
    [sym_compiler_directive_decl] = STATE(2022),
    [sym_fsi_directive_decl] = STATE(2022),
    [sym_preproc_line] = STATE(2022),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2023)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(768),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2023),
    [sym_compiler_directive_decl] = STATE(2023),
    [sym_fsi_directive_decl] = STATE(2023),
    [sym_preproc_line] = STATE(2023),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2024)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(769),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2024),
    [sym_compiler_directive_decl] = STATE(2024),
    [sym_fsi_directive_decl] = STATE(2024),
    [sym_preproc_line] = STATE(2024),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2025)] = {
    [sym_function_or_value_defn] = STATE(1299),
    [sym__mutable_expression] = STATE(22790),
    [sym__expression] = STATE(496),
    [sym_literal_expression] = STATE(5977),
    [sym_srtp_call_expression] = STATE(5977),
    [sym__srtp_type_argument] = STATE(22224),
    [sym_long_identifier_or_op] = STATE(5401),
    [sym_tuple_expression] = STATE(5977),
    [sym_brace_expression] = STATE(5977),
    [sym_anon_record_expression] = STATE(5977),
    [sym_prefixed_expression] = STATE(5977),
    [sym_typecast_expression] = STATE(5977),
    [sym_for_expression] = STATE(5977),
    [sym_while_expression] = STATE(5977),
    [sym__if_branch] = STATE(22084),
    [sym_if_expression] = STATE(5977),
    [sym_fun_expression] = STATE(5977),
    [sym_try_expression] = STATE(5977),
    [sym_match_expression] = STATE(5977),
    [sym_function_expression] = STATE(5977),
    [sym_mutate_expression] = STATE(5977),
    [sym_index_expression] = STATE(5401),
    [sym_typed_expression] = STATE(5977),
    [sym_declaration_expression] = STATE(5977),
    [sym_do_expression] = STATE(5977),
    [sym_list_expression] = STATE(5977),
    [sym_array_expression] = STATE(5977),
    [sym_begin_end_expression] = STATE(5977),
    [sym_paren_expression] = STATE(5977),
    [sym__high_prec_app] = STATE(6369),
    [sym__low_prec_app] = STATE(6369),
    [sym_application_expression] = STATE(5977),
    [sym_dot_expression] = STATE(5401),
    [sym_infix_expression] = STATE(5977),
    [sym_ce_expression] = STATE(5977),
    [sym_sequential_expression] = STATE(5977),
    [sym_char] = STATE(6198),
    [sym_format_string] = STATE(6242),
    [sym__string_literal] = STATE(6242),
    [sym_string] = STATE(6198),
    [sym_verbatim_string] = STATE(6198),
    [sym_bytearray] = STATE(6198),
    [sym_verbatim_bytearray] = STATE(6198),
    [sym_format_triple_quoted_string] = STATE(6255),
    [sym_triple_quoted_string] = STATE(6198),
    [sym_const] = STATE(5977),
    [sym_long_identifier] = STATE(5782),
    [sym_active_pattern] = STATE(5759),
    [sym__identifier_or_op] = STATE(5408),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1834),
    [sym_sbyte] = STATE(6198),
    [sym_byte] = STATE(6198),
    [sym_int16] = STATE(6198),
    [sym_uint16] = STATE(6198),
    [sym_int32] = STATE(6198),
    [sym_uint32] = STATE(6198),
    [sym_nativeint] = STATE(6198),
    [sym_unativeint] = STATE(6198),
    [sym_int64] = STATE(6198),
    [sym_uint64] = STATE(6198),
    [sym_ieee32] = STATE(6198),
    [sym_ieee64] = STATE(6198),
    [sym_bignum] = STATE(6198),
    [sym_decimal] = STATE(6198),
    [sym_float] = STATE(5042),
    [sym_block_comment] = STATE(2025),
    [sym_compiler_directive_decl] = STATE(2025),
    [sym_fsi_directive_decl] = STATE(2025),
    [sym_preproc_line] = STATE(2025),
    [sym_preproc_if_in_expression] = STATE(5977),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2682),
    [anon_sym_null] = ACTIONS(2684),
    [anon_sym_begin] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2690),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2698),
    [anon_sym_LBRACE] = ACTIONS(5560),
    [anon_sym_LT_AT] = ACTIONS(2702),
    [anon_sym_LT_AT_AT] = ACTIONS(5562),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2710),
    [anon_sym_new] = ACTIONS(2712),
    [anon_sym_return] = ACTIONS(2712),
    [anon_sym_return_BANG] = ACTIONS(2714),
    [anon_sym_yield] = ACTIONS(2712),
    [anon_sym_yield_BANG] = ACTIONS(2714),
    [anon_sym_lazy] = ACTIONS(2712),
    [anon_sym_assert] = ACTIONS(2712),
    [anon_sym_upcast] = ACTIONS(2712),
    [anon_sym_downcast] = ACTIONS(2712),
    [anon_sym_fixed] = ACTIONS(2712),
    [anon_sym_for] = ACTIONS(2718),
    [anon_sym_while] = ACTIONS(2720),
    [anon_sym_while_BANG] = ACTIONS(2722),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2724),
    [anon_sym_try] = ACTIONS(2726),
    [anon_sym_match] = ACTIONS(2728),
    [anon_sym_match_BANG] = ACTIONS(2730),
    [anon_sym_function] = ACTIONS(2732),
    [anon_sym_use] = ACTIONS(2736),
    [anon_sym_use_BANG] = ACTIONS(2738),
    [anon_sym_do_BANG] = ACTIONS(2740),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2746),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2748),
    [anon_sym_DQUOTE] = ACTIONS(2750),
    [anon_sym_AT_DQUOTE] = ACTIONS(2752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2754),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2756),
    [sym_bool] = ACTIONS(2758),
    [sym_unit] = ACTIONS(5564),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2762),
    [sym_op_identifier] = ACTIONS(5566),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2766),
    [sym_xint] = ACTIONS(2768),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2770),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2776),
  },
  [STATE(2026)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(771),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2026),
    [sym_compiler_directive_decl] = STATE(2026),
    [sym_fsi_directive_decl] = STATE(2026),
    [sym_preproc_line] = STATE(2026),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2027)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(772),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2027),
    [sym_compiler_directive_decl] = STATE(2027),
    [sym_fsi_directive_decl] = STATE(2027),
    [sym_preproc_line] = STATE(2027),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2028)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(773),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2028),
    [sym_compiler_directive_decl] = STATE(2028),
    [sym_fsi_directive_decl] = STATE(2028),
    [sym_preproc_line] = STATE(2028),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2029)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(774),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2029),
    [sym_compiler_directive_decl] = STATE(2029),
    [sym_fsi_directive_decl] = STATE(2029),
    [sym_preproc_line] = STATE(2029),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2030)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(775),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2030),
    [sym_compiler_directive_decl] = STATE(2030),
    [sym_fsi_directive_decl] = STATE(2030),
    [sym_preproc_line] = STATE(2030),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2031)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(776),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2031),
    [sym_compiler_directive_decl] = STATE(2031),
    [sym_fsi_directive_decl] = STATE(2031),
    [sym_preproc_line] = STATE(2031),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2032)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(777),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2032),
    [sym_compiler_directive_decl] = STATE(2032),
    [sym_fsi_directive_decl] = STATE(2032),
    [sym_preproc_line] = STATE(2032),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2033)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(778),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2033),
    [sym_compiler_directive_decl] = STATE(2033),
    [sym_fsi_directive_decl] = STATE(2033),
    [sym_preproc_line] = STATE(2033),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2034)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(779),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2034),
    [sym_compiler_directive_decl] = STATE(2034),
    [sym_fsi_directive_decl] = STATE(2034),
    [sym_preproc_line] = STATE(2034),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2035)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(780),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2035),
    [sym_compiler_directive_decl] = STATE(2035),
    [sym_fsi_directive_decl] = STATE(2035),
    [sym_preproc_line] = STATE(2035),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2036)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(781),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2036),
    [sym_compiler_directive_decl] = STATE(2036),
    [sym_fsi_directive_decl] = STATE(2036),
    [sym_preproc_line] = STATE(2036),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2037)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(782),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2037),
    [sym_compiler_directive_decl] = STATE(2037),
    [sym_fsi_directive_decl] = STATE(2037),
    [sym_preproc_line] = STATE(2037),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2038)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(783),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2038),
    [sym_compiler_directive_decl] = STATE(2038),
    [sym_fsi_directive_decl] = STATE(2038),
    [sym_preproc_line] = STATE(2038),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2039)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(784),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2039),
    [sym_compiler_directive_decl] = STATE(2039),
    [sym_fsi_directive_decl] = STATE(2039),
    [sym_preproc_line] = STATE(2039),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2040)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(785),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2040),
    [sym_compiler_directive_decl] = STATE(2040),
    [sym_fsi_directive_decl] = STATE(2040),
    [sym_preproc_line] = STATE(2040),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2041)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(786),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2041),
    [sym_compiler_directive_decl] = STATE(2041),
    [sym_fsi_directive_decl] = STATE(2041),
    [sym_preproc_line] = STATE(2041),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2042)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(787),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2042),
    [sym_compiler_directive_decl] = STATE(2042),
    [sym_fsi_directive_decl] = STATE(2042),
    [sym_preproc_line] = STATE(2042),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2043)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(788),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2043),
    [sym_compiler_directive_decl] = STATE(2043),
    [sym_fsi_directive_decl] = STATE(2043),
    [sym_preproc_line] = STATE(2043),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2044)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(789),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2044),
    [sym_compiler_directive_decl] = STATE(2044),
    [sym_fsi_directive_decl] = STATE(2044),
    [sym_preproc_line] = STATE(2044),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2045)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(790),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2045),
    [sym_compiler_directive_decl] = STATE(2045),
    [sym_fsi_directive_decl] = STATE(2045),
    [sym_preproc_line] = STATE(2045),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2046)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(791),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2046),
    [sym_compiler_directive_decl] = STATE(2046),
    [sym_fsi_directive_decl] = STATE(2046),
    [sym_preproc_line] = STATE(2046),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2047)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(792),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2047),
    [sym_compiler_directive_decl] = STATE(2047),
    [sym_fsi_directive_decl] = STATE(2047),
    [sym_preproc_line] = STATE(2047),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2048)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(793),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2048),
    [sym_compiler_directive_decl] = STATE(2048),
    [sym_fsi_directive_decl] = STATE(2048),
    [sym_preproc_line] = STATE(2048),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2049)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(794),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2049),
    [sym_compiler_directive_decl] = STATE(2049),
    [sym_fsi_directive_decl] = STATE(2049),
    [sym_preproc_line] = STATE(2049),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2050)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(795),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2050),
    [sym_compiler_directive_decl] = STATE(2050),
    [sym_fsi_directive_decl] = STATE(2050),
    [sym_preproc_line] = STATE(2050),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2051)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(796),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2051),
    [sym_compiler_directive_decl] = STATE(2051),
    [sym_fsi_directive_decl] = STATE(2051),
    [sym_preproc_line] = STATE(2051),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2052)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(797),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2052),
    [sym_compiler_directive_decl] = STATE(2052),
    [sym_fsi_directive_decl] = STATE(2052),
    [sym_preproc_line] = STATE(2052),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2053)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(798),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2053),
    [sym_compiler_directive_decl] = STATE(2053),
    [sym_fsi_directive_decl] = STATE(2053),
    [sym_preproc_line] = STATE(2053),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2054)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(799),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2054),
    [sym_compiler_directive_decl] = STATE(2054),
    [sym_fsi_directive_decl] = STATE(2054),
    [sym_preproc_line] = STATE(2054),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2055)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(800),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2055),
    [sym_compiler_directive_decl] = STATE(2055),
    [sym_fsi_directive_decl] = STATE(2055),
    [sym_preproc_line] = STATE(2055),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2056)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(975),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2056),
    [sym_compiler_directive_decl] = STATE(2056),
    [sym_fsi_directive_decl] = STATE(2056),
    [sym_preproc_line] = STATE(2056),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2057)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(801),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2057),
    [sym_compiler_directive_decl] = STATE(2057),
    [sym_fsi_directive_decl] = STATE(2057),
    [sym_preproc_line] = STATE(2057),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2058)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(803),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2058),
    [sym_compiler_directive_decl] = STATE(2058),
    [sym_fsi_directive_decl] = STATE(2058),
    [sym_preproc_line] = STATE(2058),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2059)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(983),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2059),
    [sym_compiler_directive_decl] = STATE(2059),
    [sym_fsi_directive_decl] = STATE(2059),
    [sym_preproc_line] = STATE(2059),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2060)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(984),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2060),
    [sym_compiler_directive_decl] = STATE(2060),
    [sym_fsi_directive_decl] = STATE(2060),
    [sym_preproc_line] = STATE(2060),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2061)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(809),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2061),
    [sym_compiler_directive_decl] = STATE(2061),
    [sym_fsi_directive_decl] = STATE(2061),
    [sym_preproc_line] = STATE(2061),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2062)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(208),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(2062),
    [sym_compiler_directive_decl] = STATE(2062),
    [sym_fsi_directive_decl] = STATE(2062),
    [sym_preproc_line] = STATE(2062),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(2063)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(811),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2063),
    [sym_compiler_directive_decl] = STATE(2063),
    [sym_fsi_directive_decl] = STATE(2063),
    [sym_preproc_line] = STATE(2063),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2064)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(232),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2064),
    [sym_compiler_directive_decl] = STATE(2064),
    [sym_fsi_directive_decl] = STATE(2064),
    [sym_preproc_line] = STATE(2064),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2065)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(986),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2065),
    [sym_compiler_directive_decl] = STATE(2065),
    [sym_fsi_directive_decl] = STATE(2065),
    [sym_preproc_line] = STATE(2065),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2066)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(987),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2066),
    [sym_compiler_directive_decl] = STATE(2066),
    [sym_fsi_directive_decl] = STATE(2066),
    [sym_preproc_line] = STATE(2066),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2067)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(813),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2067),
    [sym_compiler_directive_decl] = STATE(2067),
    [sym_fsi_directive_decl] = STATE(2067),
    [sym_preproc_line] = STATE(2067),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2068)] = {
    [sym_function_or_value_defn] = STATE(1296),
    [sym__mutable_expression] = STATE(21363),
    [sym__expression] = STATE(688),
    [sym_literal_expression] = STATE(6427),
    [sym_srtp_call_expression] = STATE(6427),
    [sym__srtp_type_argument] = STATE(22310),
    [sym_long_identifier_or_op] = STATE(5517),
    [sym_tuple_expression] = STATE(6427),
    [sym_brace_expression] = STATE(6427),
    [sym_anon_record_expression] = STATE(6427),
    [sym_prefixed_expression] = STATE(6427),
    [sym_typecast_expression] = STATE(6427),
    [sym_for_expression] = STATE(6427),
    [sym_while_expression] = STATE(6427),
    [sym__if_branch] = STATE(19155),
    [sym_if_expression] = STATE(6427),
    [sym_fun_expression] = STATE(6427),
    [sym_try_expression] = STATE(6427),
    [sym_match_expression] = STATE(6427),
    [sym_function_expression] = STATE(6427),
    [sym_mutate_expression] = STATE(6427),
    [sym_index_expression] = STATE(5517),
    [sym_typed_expression] = STATE(6427),
    [sym_declaration_expression] = STATE(6427),
    [sym_do_expression] = STATE(6427),
    [sym_list_expression] = STATE(6427),
    [sym_array_expression] = STATE(6427),
    [sym_begin_end_expression] = STATE(6427),
    [sym_paren_expression] = STATE(6427),
    [sym__high_prec_app] = STATE(6300),
    [sym__low_prec_app] = STATE(6300),
    [sym_application_expression] = STATE(6427),
    [sym_dot_expression] = STATE(5517),
    [sym_infix_expression] = STATE(6427),
    [sym_ce_expression] = STATE(6427),
    [sym_sequential_expression] = STATE(6427),
    [sym_char] = STATE(6024),
    [sym_format_string] = STATE(6046),
    [sym__string_literal] = STATE(6046),
    [sym_string] = STATE(6024),
    [sym_verbatim_string] = STATE(6024),
    [sym_bytearray] = STATE(6024),
    [sym_verbatim_bytearray] = STATE(6024),
    [sym_format_triple_quoted_string] = STATE(6048),
    [sym_triple_quoted_string] = STATE(6024),
    [sym_const] = STATE(6427),
    [sym_long_identifier] = STATE(5677),
    [sym_active_pattern] = STATE(5666),
    [sym__identifier_or_op] = STATE(5854),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1529),
    [sym_sbyte] = STATE(6024),
    [sym_byte] = STATE(6024),
    [sym_int16] = STATE(6024),
    [sym_uint16] = STATE(6024),
    [sym_int32] = STATE(6024),
    [sym_uint32] = STATE(6024),
    [sym_nativeint] = STATE(6024),
    [sym_unativeint] = STATE(6024),
    [sym_int64] = STATE(6024),
    [sym_uint64] = STATE(6024),
    [sym_ieee32] = STATE(6024),
    [sym_ieee64] = STATE(6024),
    [sym_bignum] = STATE(6024),
    [sym_decimal] = STATE(6024),
    [sym_float] = STATE(4987),
    [sym_block_comment] = STATE(2068),
    [sym_compiler_directive_decl] = STATE(2068),
    [sym_fsi_directive_decl] = STATE(2068),
    [sym_preproc_line] = STATE(2068),
    [sym_preproc_if_in_expression] = STATE(6427),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(3082),
    [anon_sym_null] = ACTIONS(3084),
    [anon_sym_begin] = ACTIONS(3086),
    [anon_sym_do] = ACTIONS(3090),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(3092),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(3096),
    [anon_sym_LBRACK_PIPE] = ACTIONS(3098),
    [anon_sym_LBRACE] = ACTIONS(5536),
    [anon_sym_LT_AT] = ACTIONS(3102),
    [anon_sym_LT_AT_AT] = ACTIONS(5538),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(3110),
    [anon_sym_new] = ACTIONS(3112),
    [anon_sym_return] = ACTIONS(3112),
    [anon_sym_return_BANG] = ACTIONS(3114),
    [anon_sym_yield] = ACTIONS(3112),
    [anon_sym_yield_BANG] = ACTIONS(3114),
    [anon_sym_lazy] = ACTIONS(3112),
    [anon_sym_assert] = ACTIONS(3112),
    [anon_sym_upcast] = ACTIONS(3112),
    [anon_sym_downcast] = ACTIONS(3112),
    [anon_sym_fixed] = ACTIONS(3112),
    [anon_sym_for] = ACTIONS(3118),
    [anon_sym_while] = ACTIONS(3120),
    [anon_sym_while_BANG] = ACTIONS(3122),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(3124),
    [anon_sym_try] = ACTIONS(3126),
    [anon_sym_match] = ACTIONS(3128),
    [anon_sym_match_BANG] = ACTIONS(3130),
    [anon_sym_function] = ACTIONS(3132),
    [anon_sym_use] = ACTIONS(3136),
    [anon_sym_use_BANG] = ACTIONS(3138),
    [anon_sym_do_BANG] = ACTIONS(3140),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(3146),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(3148),
    [anon_sym_DQUOTE] = ACTIONS(3150),
    [anon_sym_AT_DQUOTE] = ACTIONS(3152),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3154),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(3156),
    [sym_bool] = ACTIONS(3158),
    [sym_unit] = ACTIONS(5540),
    [anon_sym_LPAREN_PIPE] = ACTIONS(3162),
    [sym_op_identifier] = ACTIONS(5542),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3166),
    [sym_xint] = ACTIONS(3168),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(3170),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(3174),
  },
  [STATE(2069)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(815),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2069),
    [sym_compiler_directive_decl] = STATE(2069),
    [sym_fsi_directive_decl] = STATE(2069),
    [sym_preproc_line] = STATE(2069),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2070)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(988),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2070),
    [sym_compiler_directive_decl] = STATE(2070),
    [sym_fsi_directive_decl] = STATE(2070),
    [sym_preproc_line] = STATE(2070),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2071)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(957),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2071),
    [sym_compiler_directive_decl] = STATE(2071),
    [sym_fsi_directive_decl] = STATE(2071),
    [sym_preproc_line] = STATE(2071),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2072)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(816),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2072),
    [sym_compiler_directive_decl] = STATE(2072),
    [sym_fsi_directive_decl] = STATE(2072),
    [sym_preproc_line] = STATE(2072),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2073)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(818),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2073),
    [sym_compiler_directive_decl] = STATE(2073),
    [sym_fsi_directive_decl] = STATE(2073),
    [sym_preproc_line] = STATE(2073),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2074)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(958),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2074),
    [sym_compiler_directive_decl] = STATE(2074),
    [sym_fsi_directive_decl] = STATE(2074),
    [sym_preproc_line] = STATE(2074),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2075)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(931),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2075),
    [sym_compiler_directive_decl] = STATE(2075),
    [sym_fsi_directive_decl] = STATE(2075),
    [sym_preproc_line] = STATE(2075),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2076)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(819),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2076),
    [sym_compiler_directive_decl] = STATE(2076),
    [sym_fsi_directive_decl] = STATE(2076),
    [sym_preproc_line] = STATE(2076),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2077)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(821),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2077),
    [sym_compiler_directive_decl] = STATE(2077),
    [sym_fsi_directive_decl] = STATE(2077),
    [sym_preproc_line] = STATE(2077),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2078)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(932),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2078),
    [sym_compiler_directive_decl] = STATE(2078),
    [sym_fsi_directive_decl] = STATE(2078),
    [sym_preproc_line] = STATE(2078),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2079)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(933),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2079),
    [sym_compiler_directive_decl] = STATE(2079),
    [sym_fsi_directive_decl] = STATE(2079),
    [sym_preproc_line] = STATE(2079),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2080)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(822),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2080),
    [sym_compiler_directive_decl] = STATE(2080),
    [sym_fsi_directive_decl] = STATE(2080),
    [sym_preproc_line] = STATE(2080),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2081)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(824),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2081),
    [sym_compiler_directive_decl] = STATE(2081),
    [sym_fsi_directive_decl] = STATE(2081),
    [sym_preproc_line] = STATE(2081),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2082)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(935),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2082),
    [sym_compiler_directive_decl] = STATE(2082),
    [sym_fsi_directive_decl] = STATE(2082),
    [sym_preproc_line] = STATE(2082),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2083)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(936),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2083),
    [sym_compiler_directive_decl] = STATE(2083),
    [sym_fsi_directive_decl] = STATE(2083),
    [sym_preproc_line] = STATE(2083),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2084)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(825),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2084),
    [sym_compiler_directive_decl] = STATE(2084),
    [sym_fsi_directive_decl] = STATE(2084),
    [sym_preproc_line] = STATE(2084),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2085)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(827),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2085),
    [sym_compiler_directive_decl] = STATE(2085),
    [sym_fsi_directive_decl] = STATE(2085),
    [sym_preproc_line] = STATE(2085),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2086)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(937),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2086),
    [sym_compiler_directive_decl] = STATE(2086),
    [sym_fsi_directive_decl] = STATE(2086),
    [sym_preproc_line] = STATE(2086),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2087)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(938),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2087),
    [sym_compiler_directive_decl] = STATE(2087),
    [sym_fsi_directive_decl] = STATE(2087),
    [sym_preproc_line] = STATE(2087),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2088)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(828),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2088),
    [sym_compiler_directive_decl] = STATE(2088),
    [sym_fsi_directive_decl] = STATE(2088),
    [sym_preproc_line] = STATE(2088),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2089)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(830),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2089),
    [sym_compiler_directive_decl] = STATE(2089),
    [sym_fsi_directive_decl] = STATE(2089),
    [sym_preproc_line] = STATE(2089),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2090)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(940),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2090),
    [sym_compiler_directive_decl] = STATE(2090),
    [sym_fsi_directive_decl] = STATE(2090),
    [sym_preproc_line] = STATE(2090),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2091)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(941),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2091),
    [sym_compiler_directive_decl] = STATE(2091),
    [sym_fsi_directive_decl] = STATE(2091),
    [sym_preproc_line] = STATE(2091),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2092)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(831),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2092),
    [sym_compiler_directive_decl] = STATE(2092),
    [sym_fsi_directive_decl] = STATE(2092),
    [sym_preproc_line] = STATE(2092),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2093)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(833),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2093),
    [sym_compiler_directive_decl] = STATE(2093),
    [sym_fsi_directive_decl] = STATE(2093),
    [sym_preproc_line] = STATE(2093),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2094)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(943),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2094),
    [sym_compiler_directive_decl] = STATE(2094),
    [sym_fsi_directive_decl] = STATE(2094),
    [sym_preproc_line] = STATE(2094),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2095)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(944),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2095),
    [sym_compiler_directive_decl] = STATE(2095),
    [sym_fsi_directive_decl] = STATE(2095),
    [sym_preproc_line] = STATE(2095),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2096)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(834),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2096),
    [sym_compiler_directive_decl] = STATE(2096),
    [sym_fsi_directive_decl] = STATE(2096),
    [sym_preproc_line] = STATE(2096),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2097)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(656),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2097),
    [sym_compiler_directive_decl] = STATE(2097),
    [sym_fsi_directive_decl] = STATE(2097),
    [sym_preproc_line] = STATE(2097),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2098)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(836),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2098),
    [sym_compiler_directive_decl] = STATE(2098),
    [sym_fsi_directive_decl] = STATE(2098),
    [sym_preproc_line] = STATE(2098),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2099)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(945),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2099),
    [sym_compiler_directive_decl] = STATE(2099),
    [sym_fsi_directive_decl] = STATE(2099),
    [sym_preproc_line] = STATE(2099),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2100)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(946),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2100),
    [sym_compiler_directive_decl] = STATE(2100),
    [sym_fsi_directive_decl] = STATE(2100),
    [sym_preproc_line] = STATE(2100),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2101)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(837),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2101),
    [sym_compiler_directive_decl] = STATE(2101),
    [sym_fsi_directive_decl] = STATE(2101),
    [sym_preproc_line] = STATE(2101),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2102)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(839),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2102),
    [sym_compiler_directive_decl] = STATE(2102),
    [sym_fsi_directive_decl] = STATE(2102),
    [sym_preproc_line] = STATE(2102),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2103)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(947),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2103),
    [sym_compiler_directive_decl] = STATE(2103),
    [sym_fsi_directive_decl] = STATE(2103),
    [sym_preproc_line] = STATE(2103),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2104)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(948),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2104),
    [sym_compiler_directive_decl] = STATE(2104),
    [sym_fsi_directive_decl] = STATE(2104),
    [sym_preproc_line] = STATE(2104),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2105)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(840),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2105),
    [sym_compiler_directive_decl] = STATE(2105),
    [sym_fsi_directive_decl] = STATE(2105),
    [sym_preproc_line] = STATE(2105),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2106)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(842),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2106),
    [sym_compiler_directive_decl] = STATE(2106),
    [sym_fsi_directive_decl] = STATE(2106),
    [sym_preproc_line] = STATE(2106),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2107)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(949),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2107),
    [sym_compiler_directive_decl] = STATE(2107),
    [sym_fsi_directive_decl] = STATE(2107),
    [sym_preproc_line] = STATE(2107),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2108)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(950),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2108),
    [sym_compiler_directive_decl] = STATE(2108),
    [sym_fsi_directive_decl] = STATE(2108),
    [sym_preproc_line] = STATE(2108),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2109)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(843),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2109),
    [sym_compiler_directive_decl] = STATE(2109),
    [sym_fsi_directive_decl] = STATE(2109),
    [sym_preproc_line] = STATE(2109),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2110)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(657),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2110),
    [sym_compiler_directive_decl] = STATE(2110),
    [sym_fsi_directive_decl] = STATE(2110),
    [sym_preproc_line] = STATE(2110),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2111)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(845),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2111),
    [sym_compiler_directive_decl] = STATE(2111),
    [sym_fsi_directive_decl] = STATE(2111),
    [sym_preproc_line] = STATE(2111),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2112)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(952),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2112),
    [sym_compiler_directive_decl] = STATE(2112),
    [sym_fsi_directive_decl] = STATE(2112),
    [sym_preproc_line] = STATE(2112),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2113)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(953),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2113),
    [sym_compiler_directive_decl] = STATE(2113),
    [sym_fsi_directive_decl] = STATE(2113),
    [sym_preproc_line] = STATE(2113),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2114)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(846),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2114),
    [sym_compiler_directive_decl] = STATE(2114),
    [sym_fsi_directive_decl] = STATE(2114),
    [sym_preproc_line] = STATE(2114),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2115)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(848),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2115),
    [sym_compiler_directive_decl] = STATE(2115),
    [sym_fsi_directive_decl] = STATE(2115),
    [sym_preproc_line] = STATE(2115),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2116)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(954),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2116),
    [sym_compiler_directive_decl] = STATE(2116),
    [sym_fsi_directive_decl] = STATE(2116),
    [sym_preproc_line] = STATE(2116),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2117)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(955),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2117),
    [sym_compiler_directive_decl] = STATE(2117),
    [sym_fsi_directive_decl] = STATE(2117),
    [sym_preproc_line] = STATE(2117),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2118)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(849),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2118),
    [sym_compiler_directive_decl] = STATE(2118),
    [sym_fsi_directive_decl] = STATE(2118),
    [sym_preproc_line] = STATE(2118),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2119)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(851),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2119),
    [sym_compiler_directive_decl] = STATE(2119),
    [sym_fsi_directive_decl] = STATE(2119),
    [sym_preproc_line] = STATE(2119),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2120)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(956),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2120),
    [sym_compiler_directive_decl] = STATE(2120),
    [sym_fsi_directive_decl] = STATE(2120),
    [sym_preproc_line] = STATE(2120),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2121)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(930),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2121),
    [sym_compiler_directive_decl] = STATE(2121),
    [sym_fsi_directive_decl] = STATE(2121),
    [sym_preproc_line] = STATE(2121),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2122)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(852),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2122),
    [sym_compiler_directive_decl] = STATE(2122),
    [sym_fsi_directive_decl] = STATE(2122),
    [sym_preproc_line] = STATE(2122),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2123)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(854),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2123),
    [sym_compiler_directive_decl] = STATE(2123),
    [sym_fsi_directive_decl] = STATE(2123),
    [sym_preproc_line] = STATE(2123),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2124)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(990),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2124),
    [sym_compiler_directive_decl] = STATE(2124),
    [sym_fsi_directive_decl] = STATE(2124),
    [sym_preproc_line] = STATE(2124),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2125)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(959),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2125),
    [sym_compiler_directive_decl] = STATE(2125),
    [sym_fsi_directive_decl] = STATE(2125),
    [sym_preproc_line] = STATE(2125),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2126)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(855),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2126),
    [sym_compiler_directive_decl] = STATE(2126),
    [sym_fsi_directive_decl] = STATE(2126),
    [sym_preproc_line] = STATE(2126),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2127)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(661),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2127),
    [sym_compiler_directive_decl] = STATE(2127),
    [sym_fsi_directive_decl] = STATE(2127),
    [sym_preproc_line] = STATE(2127),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2128)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(857),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2128),
    [sym_compiler_directive_decl] = STATE(2128),
    [sym_fsi_directive_decl] = STATE(2128),
    [sym_preproc_line] = STATE(2128),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2129)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(961),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2129),
    [sym_compiler_directive_decl] = STATE(2129),
    [sym_fsi_directive_decl] = STATE(2129),
    [sym_preproc_line] = STATE(2129),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2130)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(962),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2130),
    [sym_compiler_directive_decl] = STATE(2130),
    [sym_fsi_directive_decl] = STATE(2130),
    [sym_preproc_line] = STATE(2130),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2131)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(858),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2131),
    [sym_compiler_directive_decl] = STATE(2131),
    [sym_fsi_directive_decl] = STATE(2131),
    [sym_preproc_line] = STATE(2131),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2132)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(860),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2132),
    [sym_compiler_directive_decl] = STATE(2132),
    [sym_fsi_directive_decl] = STATE(2132),
    [sym_preproc_line] = STATE(2132),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2133)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(963),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2133),
    [sym_compiler_directive_decl] = STATE(2133),
    [sym_fsi_directive_decl] = STATE(2133),
    [sym_preproc_line] = STATE(2133),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2134)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(964),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2134),
    [sym_compiler_directive_decl] = STATE(2134),
    [sym_fsi_directive_decl] = STATE(2134),
    [sym_preproc_line] = STATE(2134),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2135)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(861),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2135),
    [sym_compiler_directive_decl] = STATE(2135),
    [sym_fsi_directive_decl] = STATE(2135),
    [sym_preproc_line] = STATE(2135),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2136)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(863),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2136),
    [sym_compiler_directive_decl] = STATE(2136),
    [sym_fsi_directive_decl] = STATE(2136),
    [sym_preproc_line] = STATE(2136),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2137)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(965),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2137),
    [sym_compiler_directive_decl] = STATE(2137),
    [sym_fsi_directive_decl] = STATE(2137),
    [sym_preproc_line] = STATE(2137),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2138)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(966),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2138),
    [sym_compiler_directive_decl] = STATE(2138),
    [sym_fsi_directive_decl] = STATE(2138),
    [sym_preproc_line] = STATE(2138),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2139)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(864),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2139),
    [sym_compiler_directive_decl] = STATE(2139),
    [sym_fsi_directive_decl] = STATE(2139),
    [sym_preproc_line] = STATE(2139),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2140)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(866),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2140),
    [sym_compiler_directive_decl] = STATE(2140),
    [sym_fsi_directive_decl] = STATE(2140),
    [sym_preproc_line] = STATE(2140),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2141)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(967),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2141),
    [sym_compiler_directive_decl] = STATE(2141),
    [sym_fsi_directive_decl] = STATE(2141),
    [sym_preproc_line] = STATE(2141),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2142)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(968),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2142),
    [sym_compiler_directive_decl] = STATE(2142),
    [sym_fsi_directive_decl] = STATE(2142),
    [sym_preproc_line] = STATE(2142),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2143)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(867),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2143),
    [sym_compiler_directive_decl] = STATE(2143),
    [sym_fsi_directive_decl] = STATE(2143),
    [sym_preproc_line] = STATE(2143),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2144)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(869),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2144),
    [sym_compiler_directive_decl] = STATE(2144),
    [sym_fsi_directive_decl] = STATE(2144),
    [sym_preproc_line] = STATE(2144),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2145)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(971),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2145),
    [sym_compiler_directive_decl] = STATE(2145),
    [sym_fsi_directive_decl] = STATE(2145),
    [sym_preproc_line] = STATE(2145),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2146)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(972),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2146),
    [sym_compiler_directive_decl] = STATE(2146),
    [sym_fsi_directive_decl] = STATE(2146),
    [sym_preproc_line] = STATE(2146),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2147)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(870),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2147),
    [sym_compiler_directive_decl] = STATE(2147),
    [sym_fsi_directive_decl] = STATE(2147),
    [sym_preproc_line] = STATE(2147),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2148)] = {
    [sym_function_or_value_defn] = STATE(1365),
    [sym__mutable_expression] = STATE(21972),
    [sym__expression] = STATE(209),
    [sym_literal_expression] = STATE(3962),
    [sym_srtp_call_expression] = STATE(3962),
    [sym__srtp_type_argument] = STATE(19550),
    [sym_long_identifier_or_op] = STATE(3851),
    [sym_tuple_expression] = STATE(3962),
    [sym_brace_expression] = STATE(3962),
    [sym_anon_record_expression] = STATE(3962),
    [sym_prefixed_expression] = STATE(3962),
    [sym_typecast_expression] = STATE(3962),
    [sym_for_expression] = STATE(3962),
    [sym_while_expression] = STATE(3962),
    [sym__if_branch] = STATE(20308),
    [sym_if_expression] = STATE(3962),
    [sym_fun_expression] = STATE(3962),
    [sym_try_expression] = STATE(3962),
    [sym_match_expression] = STATE(3962),
    [sym_function_expression] = STATE(3962),
    [sym_mutate_expression] = STATE(3962),
    [sym_index_expression] = STATE(3851),
    [sym_typed_expression] = STATE(3962),
    [sym_declaration_expression] = STATE(3962),
    [sym_do_expression] = STATE(3962),
    [sym_list_expression] = STATE(3962),
    [sym_array_expression] = STATE(3962),
    [sym_begin_end_expression] = STATE(3962),
    [sym_paren_expression] = STATE(3962),
    [sym__high_prec_app] = STATE(3984),
    [sym__low_prec_app] = STATE(3984),
    [sym_application_expression] = STATE(3962),
    [sym_dot_expression] = STATE(3851),
    [sym_infix_expression] = STATE(3962),
    [sym_ce_expression] = STATE(3962),
    [sym_sequential_expression] = STATE(3962),
    [sym_char] = STATE(4059),
    [sym_format_string] = STATE(3884),
    [sym__string_literal] = STATE(3884),
    [sym_string] = STATE(4059),
    [sym_verbatim_string] = STATE(4059),
    [sym_bytearray] = STATE(4059),
    [sym_verbatim_bytearray] = STATE(4059),
    [sym_format_triple_quoted_string] = STATE(3941),
    [sym_triple_quoted_string] = STATE(4059),
    [sym_const] = STATE(3962),
    [sym_long_identifier] = STATE(3772),
    [sym_active_pattern] = STATE(3849),
    [sym__identifier_or_op] = STATE(3858),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1815),
    [sym_sbyte] = STATE(4059),
    [sym_byte] = STATE(4059),
    [sym_int16] = STATE(4059),
    [sym_uint16] = STATE(4059),
    [sym_int32] = STATE(4059),
    [sym_uint32] = STATE(4059),
    [sym_nativeint] = STATE(4059),
    [sym_unativeint] = STATE(4059),
    [sym_int64] = STATE(4059),
    [sym_uint64] = STATE(4059),
    [sym_ieee32] = STATE(4059),
    [sym_ieee64] = STATE(4059),
    [sym_bignum] = STATE(4059),
    [sym_decimal] = STATE(4059),
    [sym_float] = STATE(3717),
    [sym_block_comment] = STATE(2148),
    [sym_compiler_directive_decl] = STATE(2148),
    [sym_fsi_directive_decl] = STATE(2148),
    [sym_preproc_line] = STATE(2148),
    [sym_preproc_if_in_expression] = STATE(3962),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(21),
    [anon_sym_null] = ACTIONS(27),
    [anon_sym_begin] = ACTIONS(29),
    [anon_sym_do] = ACTIONS(1264),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_LBRACK_PIPE] = ACTIONS(49),
    [anon_sym_LBRACE] = ACTIONS(51),
    [anon_sym_LT_AT] = ACTIONS(53),
    [anon_sym_LT_AT_AT] = ACTIONS(55),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(59),
    [anon_sym_new] = ACTIONS(61),
    [anon_sym_return] = ACTIONS(61),
    [anon_sym_return_BANG] = ACTIONS(63),
    [anon_sym_yield] = ACTIONS(61),
    [anon_sym_yield_BANG] = ACTIONS(63),
    [anon_sym_lazy] = ACTIONS(61),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_upcast] = ACTIONS(61),
    [anon_sym_downcast] = ACTIONS(61),
    [anon_sym_fixed] = ACTIONS(61),
    [anon_sym_for] = ACTIONS(65),
    [anon_sym_while] = ACTIONS(67),
    [anon_sym_while_BANG] = ACTIONS(69),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(75),
    [anon_sym_match] = ACTIONS(77),
    [anon_sym_match_BANG] = ACTIONS(79),
    [anon_sym_function] = ACTIONS(81),
    [anon_sym_use] = ACTIONS(83),
    [anon_sym_use_BANG] = ACTIONS(85),
    [anon_sym_do_BANG] = ACTIONS(87),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(95),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [anon_sym_AT_DQUOTE] = ACTIONS(101),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(105),
    [sym_bool] = ACTIONS(107),
    [sym_unit] = ACTIONS(109),
    [anon_sym_LPAREN_PIPE] = ACTIONS(111),
    [sym_op_identifier] = ACTIONS(113),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(119),
    [sym_xint] = ACTIONS(121),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(127),
  },
  [STATE(2149)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(872),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2149),
    [sym_compiler_directive_decl] = STATE(2149),
    [sym_fsi_directive_decl] = STATE(2149),
    [sym_preproc_line] = STATE(2149),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2150)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(973),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2150),
    [sym_compiler_directive_decl] = STATE(2150),
    [sym_fsi_directive_decl] = STATE(2150),
    [sym_preproc_line] = STATE(2150),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2151)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(974),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2151),
    [sym_compiler_directive_decl] = STATE(2151),
    [sym_fsi_directive_decl] = STATE(2151),
    [sym_preproc_line] = STATE(2151),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2152)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(873),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2152),
    [sym_compiler_directive_decl] = STATE(2152),
    [sym_fsi_directive_decl] = STATE(2152),
    [sym_preproc_line] = STATE(2152),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2153)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(875),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2153),
    [sym_compiler_directive_decl] = STATE(2153),
    [sym_fsi_directive_decl] = STATE(2153),
    [sym_preproc_line] = STATE(2153),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2154)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(976),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2154),
    [sym_compiler_directive_decl] = STATE(2154),
    [sym_fsi_directive_decl] = STATE(2154),
    [sym_preproc_line] = STATE(2154),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2155)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(977),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2155),
    [sym_compiler_directive_decl] = STATE(2155),
    [sym_fsi_directive_decl] = STATE(2155),
    [sym_preproc_line] = STATE(2155),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2156)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(876),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2156),
    [sym_compiler_directive_decl] = STATE(2156),
    [sym_fsi_directive_decl] = STATE(2156),
    [sym_preproc_line] = STATE(2156),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2157)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(878),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2157),
    [sym_compiler_directive_decl] = STATE(2157),
    [sym_fsi_directive_decl] = STATE(2157),
    [sym_preproc_line] = STATE(2157),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2158)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(978),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2158),
    [sym_compiler_directive_decl] = STATE(2158),
    [sym_fsi_directive_decl] = STATE(2158),
    [sym_preproc_line] = STATE(2158),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2159)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(979),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2159),
    [sym_compiler_directive_decl] = STATE(2159),
    [sym_fsi_directive_decl] = STATE(2159),
    [sym_preproc_line] = STATE(2159),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2160)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(879),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2160),
    [sym_compiler_directive_decl] = STATE(2160),
    [sym_fsi_directive_decl] = STATE(2160),
    [sym_preproc_line] = STATE(2160),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2161)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(881),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2161),
    [sym_compiler_directive_decl] = STATE(2161),
    [sym_fsi_directive_decl] = STATE(2161),
    [sym_preproc_line] = STATE(2161),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2162)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(980),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2162),
    [sym_compiler_directive_decl] = STATE(2162),
    [sym_fsi_directive_decl] = STATE(2162),
    [sym_preproc_line] = STATE(2162),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2163)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(981),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2163),
    [sym_compiler_directive_decl] = STATE(2163),
    [sym_fsi_directive_decl] = STATE(2163),
    [sym_preproc_line] = STATE(2163),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2164)] = {
    [sym_function_or_value_defn] = STATE(1305),
    [sym__mutable_expression] = STATE(19166),
    [sym__expression] = STATE(882),
    [sym_literal_expression] = STATE(5048),
    [sym_srtp_call_expression] = STATE(5048),
    [sym__srtp_type_argument] = STATE(22349),
    [sym_long_identifier_or_op] = STATE(4804),
    [sym_tuple_expression] = STATE(5048),
    [sym_brace_expression] = STATE(5048),
    [sym_anon_record_expression] = STATE(5048),
    [sym_prefixed_expression] = STATE(5048),
    [sym_typecast_expression] = STATE(5048),
    [sym_for_expression] = STATE(5048),
    [sym_while_expression] = STATE(5048),
    [sym__if_branch] = STATE(19269),
    [sym_if_expression] = STATE(5048),
    [sym_fun_expression] = STATE(5048),
    [sym_try_expression] = STATE(5048),
    [sym_match_expression] = STATE(5048),
    [sym_function_expression] = STATE(5048),
    [sym_mutate_expression] = STATE(5048),
    [sym_index_expression] = STATE(4804),
    [sym_typed_expression] = STATE(5048),
    [sym_declaration_expression] = STATE(5048),
    [sym_do_expression] = STATE(5048),
    [sym_list_expression] = STATE(5048),
    [sym_array_expression] = STATE(5048),
    [sym_begin_end_expression] = STATE(5048),
    [sym_paren_expression] = STATE(5048),
    [sym__high_prec_app] = STATE(5238),
    [sym__low_prec_app] = STATE(5238),
    [sym_application_expression] = STATE(5048),
    [sym_dot_expression] = STATE(4804),
    [sym_infix_expression] = STATE(5048),
    [sym_ce_expression] = STATE(5048),
    [sym_sequential_expression] = STATE(5048),
    [sym_char] = STATE(5079),
    [sym_format_string] = STATE(5089),
    [sym__string_literal] = STATE(5089),
    [sym_string] = STATE(5079),
    [sym_verbatim_string] = STATE(5079),
    [sym_bytearray] = STATE(5079),
    [sym_verbatim_bytearray] = STATE(5079),
    [sym_format_triple_quoted_string] = STATE(5103),
    [sym_triple_quoted_string] = STATE(5079),
    [sym_const] = STATE(5048),
    [sym_long_identifier] = STATE(4866),
    [sym_active_pattern] = STATE(4859),
    [sym__identifier_or_op] = STATE(4828),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1544),
    [sym_sbyte] = STATE(5079),
    [sym_byte] = STATE(5079),
    [sym_int16] = STATE(5079),
    [sym_uint16] = STATE(5079),
    [sym_int32] = STATE(5079),
    [sym_uint32] = STATE(5079),
    [sym_nativeint] = STATE(5079),
    [sym_unativeint] = STATE(5079),
    [sym_int64] = STATE(5079),
    [sym_uint64] = STATE(5079),
    [sym_ieee32] = STATE(5079),
    [sym_ieee64] = STATE(5079),
    [sym_bignum] = STATE(5079),
    [sym_decimal] = STATE(5079),
    [sym_float] = STATE(4633),
    [sym_block_comment] = STATE(2164),
    [sym_compiler_directive_decl] = STATE(2164),
    [sym_fsi_directive_decl] = STATE(2164),
    [sym_preproc_line] = STATE(2164),
    [sym_preproc_if_in_expression] = STATE(5048),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1615),
    [anon_sym_null] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(4814),
    [anon_sym_LT_AT] = ACTIONS(1635),
    [anon_sym_LT_AT_AT] = ACTIONS(4816),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(3184),
    [anon_sym_return] = ACTIONS(3184),
    [anon_sym_return_BANG] = ACTIONS(3186),
    [anon_sym_yield] = ACTIONS(3184),
    [anon_sym_yield_BANG] = ACTIONS(3186),
    [anon_sym_lazy] = ACTIONS(3184),
    [anon_sym_assert] = ACTIONS(3184),
    [anon_sym_upcast] = ACTIONS(3184),
    [anon_sym_downcast] = ACTIONS(3184),
    [anon_sym_fixed] = ACTIONS(3184),
    [anon_sym_for] = ACTIONS(3188),
    [anon_sym_while] = ACTIONS(1653),
    [anon_sym_while_BANG] = ACTIONS(1655),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1657),
    [anon_sym_try] = ACTIONS(1659),
    [anon_sym_match] = ACTIONS(1661),
    [anon_sym_match_BANG] = ACTIONS(1663),
    [anon_sym_function] = ACTIONS(1665),
    [anon_sym_use] = ACTIONS(3190),
    [anon_sym_use_BANG] = ACTIONS(3192),
    [anon_sym_do_BANG] = ACTIONS(1673),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1683),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1687),
    [anon_sym_AT_DQUOTE] = ACTIONS(1689),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1693),
    [sym_bool] = ACTIONS(1695),
    [sym_unit] = ACTIONS(4822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1699),
    [sym_op_identifier] = ACTIONS(4824),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3196),
    [sym_xint] = ACTIONS(1705),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1707),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1713),
  },
  [STATE(2165)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(884),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2165),
    [sym_compiler_directive_decl] = STATE(2165),
    [sym_fsi_directive_decl] = STATE(2165),
    [sym_preproc_line] = STATE(2165),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2166)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(982),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2166),
    [sym_compiler_directive_decl] = STATE(2166),
    [sym_fsi_directive_decl] = STATE(2166),
    [sym_preproc_line] = STATE(2166),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2167)] = {
    [sym_function_or_value_defn] = STATE(1353),
    [sym__mutable_expression] = STATE(19648),
    [sym__expression] = STATE(886),
    [sym_literal_expression] = STATE(5674),
    [sym_srtp_call_expression] = STATE(5674),
    [sym__srtp_type_argument] = STATE(22463),
    [sym_long_identifier_or_op] = STATE(5014),
    [sym_tuple_expression] = STATE(5674),
    [sym_brace_expression] = STATE(5674),
    [sym_anon_record_expression] = STATE(5674),
    [sym_prefixed_expression] = STATE(5674),
    [sym_typecast_expression] = STATE(5674),
    [sym_for_expression] = STATE(5674),
    [sym_while_expression] = STATE(5674),
    [sym__if_branch] = STATE(19911),
    [sym_if_expression] = STATE(5674),
    [sym_fun_expression] = STATE(5674),
    [sym_try_expression] = STATE(5674),
    [sym_match_expression] = STATE(5674),
    [sym_function_expression] = STATE(5674),
    [sym_mutate_expression] = STATE(5674),
    [sym_index_expression] = STATE(5014),
    [sym_typed_expression] = STATE(5674),
    [sym_declaration_expression] = STATE(5674),
    [sym_do_expression] = STATE(5674),
    [sym_list_expression] = STATE(5674),
    [sym_array_expression] = STATE(5674),
    [sym_begin_end_expression] = STATE(5674),
    [sym_paren_expression] = STATE(5674),
    [sym__high_prec_app] = STATE(5676),
    [sym__low_prec_app] = STATE(5676),
    [sym_application_expression] = STATE(5674),
    [sym_dot_expression] = STATE(5014),
    [sym_infix_expression] = STATE(5674),
    [sym_ce_expression] = STATE(5674),
    [sym_sequential_expression] = STATE(5674),
    [sym_char] = STATE(5425),
    [sym_format_string] = STATE(5427),
    [sym__string_literal] = STATE(5427),
    [sym_string] = STATE(5425),
    [sym_verbatim_string] = STATE(5425),
    [sym_bytearray] = STATE(5425),
    [sym_verbatim_bytearray] = STATE(5425),
    [sym_format_triple_quoted_string] = STATE(5428),
    [sym_triple_quoted_string] = STATE(5425),
    [sym_const] = STATE(5674),
    [sym_long_identifier] = STATE(5003),
    [sym_active_pattern] = STATE(5147),
    [sym__identifier_or_op] = STATE(4990),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1662),
    [sym_sbyte] = STATE(5425),
    [sym_byte] = STATE(5425),
    [sym_int16] = STATE(5425),
    [sym_uint16] = STATE(5425),
    [sym_int32] = STATE(5425),
    [sym_uint32] = STATE(5425),
    [sym_nativeint] = STATE(5425),
    [sym_unativeint] = STATE(5425),
    [sym_int64] = STATE(5425),
    [sym_uint64] = STATE(5425),
    [sym_ieee32] = STATE(5425),
    [sym_ieee64] = STATE(5425),
    [sym_bignum] = STATE(5425),
    [sym_decimal] = STATE(5425),
    [sym_float] = STATE(4753),
    [sym_block_comment] = STATE(2167),
    [sym_compiler_directive_decl] = STATE(2167),
    [sym_fsi_directive_decl] = STATE(2167),
    [sym_preproc_line] = STATE(2167),
    [sym_preproc_if_in_expression] = STATE(5674),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_null] = ACTIONS(2281),
    [anon_sym_begin] = ACTIONS(2283),
    [anon_sym_do] = ACTIONS(2287),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2289),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2293),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2295),
    [anon_sym_LBRACE] = ACTIONS(5520),
    [anon_sym_LT_AT] = ACTIONS(2299),
    [anon_sym_LT_AT_AT] = ACTIONS(5522),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2307),
    [anon_sym_new] = ACTIONS(3348),
    [anon_sym_return] = ACTIONS(3348),
    [anon_sym_return_BANG] = ACTIONS(3350),
    [anon_sym_yield] = ACTIONS(3348),
    [anon_sym_yield_BANG] = ACTIONS(3350),
    [anon_sym_lazy] = ACTIONS(3348),
    [anon_sym_assert] = ACTIONS(3348),
    [anon_sym_upcast] = ACTIONS(3348),
    [anon_sym_downcast] = ACTIONS(3348),
    [anon_sym_fixed] = ACTIONS(3348),
    [anon_sym_for] = ACTIONS(3354),
    [anon_sym_while] = ACTIONS(2317),
    [anon_sym_while_BANG] = ACTIONS(2319),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2321),
    [anon_sym_try] = ACTIONS(2323),
    [anon_sym_match] = ACTIONS(2325),
    [anon_sym_match_BANG] = ACTIONS(2327),
    [anon_sym_function] = ACTIONS(2329),
    [anon_sym_use] = ACTIONS(3356),
    [anon_sym_use_BANG] = ACTIONS(3358),
    [anon_sym_do_BANG] = ACTIONS(2337),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(2347),
    [anon_sym_AT_DQUOTE] = ACTIONS(2349),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2351),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2353),
    [sym_bool] = ACTIONS(2355),
    [sym_unit] = ACTIONS(5524),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2359),
    [sym_op_identifier] = ACTIONS(5526),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(3362),
    [sym_xint] = ACTIONS(2365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2367),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2371),
  },
  [STATE(2168)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(891),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2168),
    [sym_compiler_directive_decl] = STATE(2168),
    [sym_fsi_directive_decl] = STATE(2168),
    [sym_preproc_line] = STATE(2168),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2169)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(892),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2169),
    [sym_compiler_directive_decl] = STATE(2169),
    [sym_fsi_directive_decl] = STATE(2169),
    [sym_preproc_line] = STATE(2169),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2170)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(893),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2170),
    [sym_compiler_directive_decl] = STATE(2170),
    [sym_fsi_directive_decl] = STATE(2170),
    [sym_preproc_line] = STATE(2170),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2171)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(229),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2171),
    [sym_compiler_directive_decl] = STATE(2171),
    [sym_fsi_directive_decl] = STATE(2171),
    [sym_preproc_line] = STATE(2171),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2172)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(894),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2172),
    [sym_compiler_directive_decl] = STATE(2172),
    [sym_fsi_directive_decl] = STATE(2172),
    [sym_preproc_line] = STATE(2172),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2173)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(895),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2173),
    [sym_compiler_directive_decl] = STATE(2173),
    [sym_fsi_directive_decl] = STATE(2173),
    [sym_preproc_line] = STATE(2173),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2174)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(896),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2174),
    [sym_compiler_directive_decl] = STATE(2174),
    [sym_fsi_directive_decl] = STATE(2174),
    [sym_preproc_line] = STATE(2174),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2175)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(897),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2175),
    [sym_compiler_directive_decl] = STATE(2175),
    [sym_fsi_directive_decl] = STATE(2175),
    [sym_preproc_line] = STATE(2175),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2176)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(898),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2176),
    [sym_compiler_directive_decl] = STATE(2176),
    [sym_fsi_directive_decl] = STATE(2176),
    [sym_preproc_line] = STATE(2176),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2177)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(899),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2177),
    [sym_compiler_directive_decl] = STATE(2177),
    [sym_fsi_directive_decl] = STATE(2177),
    [sym_preproc_line] = STATE(2177),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2178)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(900),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2178),
    [sym_compiler_directive_decl] = STATE(2178),
    [sym_fsi_directive_decl] = STATE(2178),
    [sym_preproc_line] = STATE(2178),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2179)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(901),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2179),
    [sym_compiler_directive_decl] = STATE(2179),
    [sym_fsi_directive_decl] = STATE(2179),
    [sym_preproc_line] = STATE(2179),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2180)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(902),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2180),
    [sym_compiler_directive_decl] = STATE(2180),
    [sym_fsi_directive_decl] = STATE(2180),
    [sym_preproc_line] = STATE(2180),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2181)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(903),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2181),
    [sym_compiler_directive_decl] = STATE(2181),
    [sym_fsi_directive_decl] = STATE(2181),
    [sym_preproc_line] = STATE(2181),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2182)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(904),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2182),
    [sym_compiler_directive_decl] = STATE(2182),
    [sym_fsi_directive_decl] = STATE(2182),
    [sym_preproc_line] = STATE(2182),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2183)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(905),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2183),
    [sym_compiler_directive_decl] = STATE(2183),
    [sym_fsi_directive_decl] = STATE(2183),
    [sym_preproc_line] = STATE(2183),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2184)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(906),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2184),
    [sym_compiler_directive_decl] = STATE(2184),
    [sym_fsi_directive_decl] = STATE(2184),
    [sym_preproc_line] = STATE(2184),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2185)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(907),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2185),
    [sym_compiler_directive_decl] = STATE(2185),
    [sym_fsi_directive_decl] = STATE(2185),
    [sym_preproc_line] = STATE(2185),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2186)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(908),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2186),
    [sym_compiler_directive_decl] = STATE(2186),
    [sym_fsi_directive_decl] = STATE(2186),
    [sym_preproc_line] = STATE(2186),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2187)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(909),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2187),
    [sym_compiler_directive_decl] = STATE(2187),
    [sym_fsi_directive_decl] = STATE(2187),
    [sym_preproc_line] = STATE(2187),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2188)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(910),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2188),
    [sym_compiler_directive_decl] = STATE(2188),
    [sym_fsi_directive_decl] = STATE(2188),
    [sym_preproc_line] = STATE(2188),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2189)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(911),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2189),
    [sym_compiler_directive_decl] = STATE(2189),
    [sym_fsi_directive_decl] = STATE(2189),
    [sym_preproc_line] = STATE(2189),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2190)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(912),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2190),
    [sym_compiler_directive_decl] = STATE(2190),
    [sym_fsi_directive_decl] = STATE(2190),
    [sym_preproc_line] = STATE(2190),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2191)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(230),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2191),
    [sym_compiler_directive_decl] = STATE(2191),
    [sym_fsi_directive_decl] = STATE(2191),
    [sym_preproc_line] = STATE(2191),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2192)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(913),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2192),
    [sym_compiler_directive_decl] = STATE(2192),
    [sym_fsi_directive_decl] = STATE(2192),
    [sym_preproc_line] = STATE(2192),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2193)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(914),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2193),
    [sym_compiler_directive_decl] = STATE(2193),
    [sym_fsi_directive_decl] = STATE(2193),
    [sym_preproc_line] = STATE(2193),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2194)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(915),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2194),
    [sym_compiler_directive_decl] = STATE(2194),
    [sym_fsi_directive_decl] = STATE(2194),
    [sym_preproc_line] = STATE(2194),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2195)] = {
    [sym_function_or_value_defn] = STATE(1339),
    [sym__mutable_expression] = STATE(21233),
    [sym__expression] = STATE(916),
    [sym_literal_expression] = STATE(5892),
    [sym_srtp_call_expression] = STATE(5892),
    [sym__srtp_type_argument] = STATE(22282),
    [sym_long_identifier_or_op] = STATE(5488),
    [sym_tuple_expression] = STATE(5892),
    [sym_brace_expression] = STATE(5892),
    [sym_anon_record_expression] = STATE(5892),
    [sym_prefixed_expression] = STATE(5892),
    [sym_typecast_expression] = STATE(5892),
    [sym_for_expression] = STATE(5892),
    [sym_while_expression] = STATE(5892),
    [sym__if_branch] = STATE(22761),
    [sym_if_expression] = STATE(5892),
    [sym_fun_expression] = STATE(5892),
    [sym_try_expression] = STATE(5892),
    [sym_match_expression] = STATE(5892),
    [sym_function_expression] = STATE(5892),
    [sym_mutate_expression] = STATE(5892),
    [sym_index_expression] = STATE(5488),
    [sym_typed_expression] = STATE(5892),
    [sym_declaration_expression] = STATE(5892),
    [sym_do_expression] = STATE(5892),
    [sym_list_expression] = STATE(5892),
    [sym_array_expression] = STATE(5892),
    [sym_begin_end_expression] = STATE(5892),
    [sym_paren_expression] = STATE(5892),
    [sym__high_prec_app] = STATE(5949),
    [sym__low_prec_app] = STATE(5949),
    [sym_application_expression] = STATE(5892),
    [sym_dot_expression] = STATE(5488),
    [sym_infix_expression] = STATE(5892),
    [sym_ce_expression] = STATE(5892),
    [sym_sequential_expression] = STATE(5892),
    [sym_char] = STATE(6370),
    [sym_format_string] = STATE(6389),
    [sym__string_literal] = STATE(6389),
    [sym_string] = STATE(6370),
    [sym_verbatim_string] = STATE(6370),
    [sym_bytearray] = STATE(6370),
    [sym_verbatim_bytearray] = STATE(6370),
    [sym_format_triple_quoted_string] = STATE(6390),
    [sym_triple_quoted_string] = STATE(6370),
    [sym_const] = STATE(5892),
    [sym_long_identifier] = STATE(5853),
    [sym_active_pattern] = STATE(5841),
    [sym__identifier_or_op] = STATE(5867),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1809),
    [sym_sbyte] = STATE(6370),
    [sym_byte] = STATE(6370),
    [sym_int16] = STATE(6370),
    [sym_uint16] = STATE(6370),
    [sym_int32] = STATE(6370),
    [sym_uint32] = STATE(6370),
    [sym_nativeint] = STATE(6370),
    [sym_unativeint] = STATE(6370),
    [sym_int64] = STATE(6370),
    [sym_uint64] = STATE(6370),
    [sym_ieee32] = STATE(6370),
    [sym_ieee64] = STATE(6370),
    [sym_bignum] = STATE(6370),
    [sym_decimal] = STATE(6370),
    [sym_float] = STATE(5241),
    [sym_block_comment] = STATE(2195),
    [sym_compiler_directive_decl] = STATE(2195),
    [sym_fsi_directive_decl] = STATE(2195),
    [sym_preproc_line] = STATE(2195),
    [sym_preproc_if_in_expression] = STATE(5892),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(2882),
    [anon_sym_null] = ACTIONS(2884),
    [anon_sym_begin] = ACTIONS(2886),
    [anon_sym_do] = ACTIONS(2890),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(2896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(5357),
    [anon_sym_LT_AT] = ACTIONS(2902),
    [anon_sym_LT_AT_AT] = ACTIONS(5359),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(2910),
    [anon_sym_new] = ACTIONS(2912),
    [anon_sym_return] = ACTIONS(2912),
    [anon_sym_return_BANG] = ACTIONS(2916),
    [anon_sym_yield] = ACTIONS(2912),
    [anon_sym_yield_BANG] = ACTIONS(2916),
    [anon_sym_lazy] = ACTIONS(2912),
    [anon_sym_assert] = ACTIONS(2912),
    [anon_sym_upcast] = ACTIONS(2912),
    [anon_sym_downcast] = ACTIONS(2912),
    [anon_sym_fixed] = ACTIONS(2912),
    [anon_sym_for] = ACTIONS(2920),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_while_BANG] = ACTIONS(2924),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2928),
    [anon_sym_match] = ACTIONS(2930),
    [anon_sym_match_BANG] = ACTIONS(2932),
    [anon_sym_function] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_use_BANG] = ACTIONS(2940),
    [anon_sym_do_BANG] = ACTIONS(2942),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(2948),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(2950),
    [anon_sym_DQUOTE] = ACTIONS(2952),
    [anon_sym_AT_DQUOTE] = ACTIONS(2954),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2956),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2958),
    [sym_bool] = ACTIONS(2960),
    [sym_unit] = ACTIONS(5363),
    [anon_sym_LPAREN_PIPE] = ACTIONS(2964),
    [sym_op_identifier] = ACTIONS(5365),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(2968),
    [sym_xint] = ACTIONS(2970),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(2972),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(2976),
  },
  [STATE(2196)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(470),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2196),
    [sym_compiler_directive_decl] = STATE(2196),
    [sym_fsi_directive_decl] = STATE(2196),
    [sym_preproc_line] = STATE(2196),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2197)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(471),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2197),
    [sym_compiler_directive_decl] = STATE(2197),
    [sym_fsi_directive_decl] = STATE(2197),
    [sym_preproc_line] = STATE(2197),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2198)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(472),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2198),
    [sym_compiler_directive_decl] = STATE(2198),
    [sym_fsi_directive_decl] = STATE(2198),
    [sym_preproc_line] = STATE(2198),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2199)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(473),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2199),
    [sym_compiler_directive_decl] = STATE(2199),
    [sym_fsi_directive_decl] = STATE(2199),
    [sym_preproc_line] = STATE(2199),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2200)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(474),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2200),
    [sym_compiler_directive_decl] = STATE(2200),
    [sym_fsi_directive_decl] = STATE(2200),
    [sym_preproc_line] = STATE(2200),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2201)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(475),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2201),
    [sym_compiler_directive_decl] = STATE(2201),
    [sym_fsi_directive_decl] = STATE(2201),
    [sym_preproc_line] = STATE(2201),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2202)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(476),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2202),
    [sym_compiler_directive_decl] = STATE(2202),
    [sym_fsi_directive_decl] = STATE(2202),
    [sym_preproc_line] = STATE(2202),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2203)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(477),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2203),
    [sym_compiler_directive_decl] = STATE(2203),
    [sym_fsi_directive_decl] = STATE(2203),
    [sym_preproc_line] = STATE(2203),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2204)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(478),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2204),
    [sym_compiler_directive_decl] = STATE(2204),
    [sym_fsi_directive_decl] = STATE(2204),
    [sym_preproc_line] = STATE(2204),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2205)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(479),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2205),
    [sym_compiler_directive_decl] = STATE(2205),
    [sym_fsi_directive_decl] = STATE(2205),
    [sym_preproc_line] = STATE(2205),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2206)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(480),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2206),
    [sym_compiler_directive_decl] = STATE(2206),
    [sym_fsi_directive_decl] = STATE(2206),
    [sym_preproc_line] = STATE(2206),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2207)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(481),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2207),
    [sym_compiler_directive_decl] = STATE(2207),
    [sym_fsi_directive_decl] = STATE(2207),
    [sym_preproc_line] = STATE(2207),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2208)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(482),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2208),
    [sym_compiler_directive_decl] = STATE(2208),
    [sym_fsi_directive_decl] = STATE(2208),
    [sym_preproc_line] = STATE(2208),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2209)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(483),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2209),
    [sym_compiler_directive_decl] = STATE(2209),
    [sym_fsi_directive_decl] = STATE(2209),
    [sym_preproc_line] = STATE(2209),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2210)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(484),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2210),
    [sym_compiler_directive_decl] = STATE(2210),
    [sym_fsi_directive_decl] = STATE(2210),
    [sym_preproc_line] = STATE(2210),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2211)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(485),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2211),
    [sym_compiler_directive_decl] = STATE(2211),
    [sym_fsi_directive_decl] = STATE(2211),
    [sym_preproc_line] = STATE(2211),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2212)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(486),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2212),
    [sym_compiler_directive_decl] = STATE(2212),
    [sym_fsi_directive_decl] = STATE(2212),
    [sym_preproc_line] = STATE(2212),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2213)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(487),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2213),
    [sym_compiler_directive_decl] = STATE(2213),
    [sym_fsi_directive_decl] = STATE(2213),
    [sym_preproc_line] = STATE(2213),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2214)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(488),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2214),
    [sym_compiler_directive_decl] = STATE(2214),
    [sym_fsi_directive_decl] = STATE(2214),
    [sym_preproc_line] = STATE(2214),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2215)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(489),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2215),
    [sym_compiler_directive_decl] = STATE(2215),
    [sym_fsi_directive_decl] = STATE(2215),
    [sym_preproc_line] = STATE(2215),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2216)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(490),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2216),
    [sym_compiler_directive_decl] = STATE(2216),
    [sym_fsi_directive_decl] = STATE(2216),
    [sym_preproc_line] = STATE(2216),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2217)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(491),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2217),
    [sym_compiler_directive_decl] = STATE(2217),
    [sym_fsi_directive_decl] = STATE(2217),
    [sym_preproc_line] = STATE(2217),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2218)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(492),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2218),
    [sym_compiler_directive_decl] = STATE(2218),
    [sym_fsi_directive_decl] = STATE(2218),
    [sym_preproc_line] = STATE(2218),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2219)] = {
    [sym_function_or_value_defn] = STATE(1346),
    [sym__mutable_expression] = STATE(21603),
    [sym__expression] = STATE(493),
    [sym_literal_expression] = STATE(4338),
    [sym_srtp_call_expression] = STATE(4338),
    [sym__srtp_type_argument] = STATE(22254),
    [sym_long_identifier_or_op] = STATE(4206),
    [sym_tuple_expression] = STATE(4338),
    [sym_brace_expression] = STATE(4338),
    [sym_anon_record_expression] = STATE(4338),
    [sym_prefixed_expression] = STATE(4338),
    [sym_typecast_expression] = STATE(4338),
    [sym_for_expression] = STATE(4338),
    [sym_while_expression] = STATE(4338),
    [sym__if_branch] = STATE(22270),
    [sym_if_expression] = STATE(4338),
    [sym_fun_expression] = STATE(4338),
    [sym_try_expression] = STATE(4338),
    [sym_match_expression] = STATE(4338),
    [sym_function_expression] = STATE(4338),
    [sym_mutate_expression] = STATE(4338),
    [sym_index_expression] = STATE(4206),
    [sym_typed_expression] = STATE(4338),
    [sym_declaration_expression] = STATE(4338),
    [sym_do_expression] = STATE(4338),
    [sym_list_expression] = STATE(4338),
    [sym_array_expression] = STATE(4338),
    [sym_begin_end_expression] = STATE(4338),
    [sym_paren_expression] = STATE(4338),
    [sym__high_prec_app] = STATE(4345),
    [sym__low_prec_app] = STATE(4345),
    [sym_application_expression] = STATE(4338),
    [sym_dot_expression] = STATE(4206),
    [sym_infix_expression] = STATE(4338),
    [sym_ce_expression] = STATE(4338),
    [sym_sequential_expression] = STATE(4338),
    [sym_char] = STATE(4397),
    [sym_format_string] = STATE(4399),
    [sym__string_literal] = STATE(4399),
    [sym_string] = STATE(4397),
    [sym_verbatim_string] = STATE(4397),
    [sym_bytearray] = STATE(4397),
    [sym_verbatim_bytearray] = STATE(4397),
    [sym_format_triple_quoted_string] = STATE(4400),
    [sym_triple_quoted_string] = STATE(4397),
    [sym_const] = STATE(4338),
    [sym_long_identifier] = STATE(4130),
    [sym_active_pattern] = STATE(4129),
    [sym__identifier_or_op] = STATE(4224),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1916),
    [sym_sbyte] = STATE(4397),
    [sym_byte] = STATE(4397),
    [sym_int16] = STATE(4397),
    [sym_uint16] = STATE(4397),
    [sym_int32] = STATE(4397),
    [sym_uint32] = STATE(4397),
    [sym_nativeint] = STATE(4397),
    [sym_unativeint] = STATE(4397),
    [sym_int64] = STATE(4397),
    [sym_uint64] = STATE(4397),
    [sym_ieee32] = STATE(4397),
    [sym_ieee64] = STATE(4397),
    [sym_bignum] = STATE(4397),
    [sym_decimal] = STATE(4397),
    [sym_float] = STATE(3870),
    [sym_block_comment] = STATE(2219),
    [sym_compiler_directive_decl] = STATE(2219),
    [sym_fsi_directive_decl] = STATE(2219),
    [sym_preproc_line] = STATE(2219),
    [sym_preproc_if_in_expression] = STATE(4338),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(1419),
    [anon_sym_null] = ACTIONS(1421),
    [anon_sym_begin] = ACTIONS(1423),
    [anon_sym_do] = ACTIONS(1427),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(1429),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_LBRACE] = ACTIONS(5163),
    [anon_sym_LT_AT] = ACTIONS(1439),
    [anon_sym_LT_AT_AT] = ACTIONS(5165),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_return_BANG] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1449),
    [anon_sym_yield_BANG] = ACTIONS(1451),
    [anon_sym_lazy] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_upcast] = ACTIONS(1449),
    [anon_sym_downcast] = ACTIONS(1449),
    [anon_sym_fixed] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1457),
    [anon_sym_while_BANG] = ACTIONS(1459),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(1461),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_match] = ACTIONS(1465),
    [anon_sym_match_BANG] = ACTIONS(1467),
    [anon_sym_function] = ACTIONS(1469),
    [anon_sym_use] = ACTIONS(1473),
    [anon_sym_use_BANG] = ACTIONS(1475),
    [anon_sym_do_BANG] = ACTIONS(1477),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(1483),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(1487),
    [anon_sym_AT_DQUOTE] = ACTIONS(1489),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1491),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1493),
    [sym_bool] = ACTIONS(1495),
    [sym_unit] = ACTIONS(5175),
    [anon_sym_LPAREN_PIPE] = ACTIONS(1499),
    [sym_op_identifier] = ACTIONS(5177),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(1503),
    [sym_xint] = ACTIONS(1505),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(1507),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(1511),
  },
  [STATE(2220)] = {
    [sym_function_or_value_defn] = STATE(1306),
    [sym__mutable_expression] = STATE(22095),
    [sym__expression] = STATE(89),
    [sym_literal_expression] = STATE(3166),
    [sym_srtp_call_expression] = STATE(3166),
    [sym__srtp_type_argument] = STATE(22539),
    [sym_long_identifier_or_op] = STATE(3030),
    [sym_tuple_expression] = STATE(3166),
    [sym_brace_expression] = STATE(3166),
    [sym_anon_record_expression] = STATE(3166),
    [sym_prefixed_expression] = STATE(3166),
    [sym_typecast_expression] = STATE(3166),
    [sym_for_expression] = STATE(3166),
    [sym_while_expression] = STATE(3166),
    [sym__if_branch] = STATE(20179),
    [sym_if_expression] = STATE(3166),
    [sym_fun_expression] = STATE(3166),
    [sym_try_expression] = STATE(3166),
    [sym_match_expression] = STATE(3166),
    [sym_function_expression] = STATE(3166),
    [sym_mutate_expression] = STATE(3166),
    [sym_index_expression] = STATE(3030),
    [sym_typed_expression] = STATE(3166),
    [sym_declaration_expression] = STATE(3166),
    [sym_do_expression] = STATE(3166),
    [sym_list_expression] = STATE(3166),
    [sym_array_expression] = STATE(3166),
    [sym_begin_end_expression] = STATE(3166),
    [sym_paren_expression] = STATE(3166),
    [sym__high_prec_app] = STATE(3168),
    [sym__low_prec_app] = STATE(3168),
    [sym_application_expression] = STATE(3166),
    [sym_dot_expression] = STATE(3030),
    [sym_infix_expression] = STATE(3166),
    [sym_ce_expression] = STATE(3166),
    [sym_sequential_expression] = STATE(3166),
    [sym_char] = STATE(3135),
    [sym_format_string] = STATE(3136),
    [sym__string_literal] = STATE(3136),
    [sym_string] = STATE(3135),
    [sym_verbatim_string] = STATE(3135),
    [sym_bytearray] = STATE(3135),
    [sym_verbatim_bytearray] = STATE(3135),
    [sym_format_triple_quoted_string] = STATE(3139),
    [sym_triple_quoted_string] = STATE(3135),
    [sym_const] = STATE(3166),
    [sym_long_identifier] = STATE(3057),
    [sym_active_pattern] = STATE(2981),
    [sym__identifier_or_op] = STATE(3067),
    [sym__infix_or_prefix_op] = STATE(8890),
    [sym_prefix_op] = STATE(1759),
    [sym_sbyte] = STATE(3135),
    [sym_byte] = STATE(3135),
    [sym_int16] = STATE(3135),
    [sym_uint16] = STATE(3135),
    [sym_int32] = STATE(3135),
    [sym_uint32] = STATE(3135),
    [sym_nativeint] = STATE(3135),
    [sym_unativeint] = STATE(3135),
    [sym_int64] = STATE(3135),
    [sym_uint64] = STATE(3135),
    [sym_ieee32] = STATE(3135),
    [sym_ieee64] = STATE(3135),
    [sym_bignum] = STATE(3135),
    [sym_decimal] = STATE(3135),
    [sym_float] = STATE(2933),
    [sym_block_comment] = STATE(2220),
    [sym_compiler_directive_decl] = STATE(2220),
    [sym_fsi_directive_decl] = STATE(2220),
    [sym_preproc_line] = STATE(2220),
    [sym_preproc_if_in_expression] = STATE(3166),
    [aux_sym_prefix_op_repeat1] = STATE(8883),
    [sym_identifier] = ACTIONS(539),
    [anon_sym_null] = ACTIONS(541),
    [anon_sym_begin] = ACTIONS(543),
    [anon_sym_do] = ACTIONS(547),
    [anon_sym_let] = ACTIONS(203),
    [anon_sym_let_BANG] = ACTIONS(205),
    [anon_sym_LPAREN] = ACTIONS(549),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(555),
    [anon_sym_LBRACE] = ACTIONS(5568),
    [anon_sym_LT_AT] = ACTIONS(557),
    [anon_sym_LT_AT_AT] = ACTIONS(5570),
    [anon_sym_CARET] = ACTIONS(57),
    [anon_sym_LBRACE_PIPE] = ACTIONS(563),
    [anon_sym_new] = ACTIONS(565),
    [anon_sym_return] = ACTIONS(565),
    [anon_sym_return_BANG] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(565),
    [anon_sym_yield_BANG] = ACTIONS(567),
    [anon_sym_lazy] = ACTIONS(565),
    [anon_sym_assert] = ACTIONS(565),
    [anon_sym_upcast] = ACTIONS(565),
    [anon_sym_downcast] = ACTIONS(565),
    [anon_sym_fixed] = ACTIONS(565),
    [anon_sym_for] = ACTIONS(571),
    [anon_sym_while] = ACTIONS(573),
    [anon_sym_while_BANG] = ACTIONS(575),
    [anon_sym_if] = ACTIONS(71),
    [anon_sym_fun] = ACTIONS(577),
    [anon_sym_try] = ACTIONS(579),
    [anon_sym_match] = ACTIONS(581),
    [anon_sym_match_BANG] = ACTIONS(583),
    [anon_sym_function] = ACTIONS(585),
    [anon_sym_use] = ACTIONS(589),
    [anon_sym_use_BANG] = ACTIONS(591),
    [anon_sym_do_BANG] = ACTIONS(593),
    [anon_sym_AMP_AMP] = ACTIONS(89),
    [aux_sym_char_token1] = ACTIONS(597),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(599),
    [anon_sym_DQUOTE] = ACTIONS(601),
    [anon_sym_AT_DQUOTE] = ACTIONS(603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(607),
    [sym_bool] = ACTIONS(609),
    [sym_unit] = ACTIONS(5572),
    [anon_sym_LPAREN_PIPE] = ACTIONS(613),
    [sym_op_identifier] = ACTIONS(5574),
    [anon_sym_PLUS] = ACTIONS(45),
    [anon_sym_DASH] = ACTIONS(45),
    [anon_sym_PLUS_DOT] = ACTIONS(115),
    [anon_sym_DASH_DOT] = ACTIONS(115),
    [anon_sym_PERCENT] = ACTIONS(45),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(89),
    [anon_sym_TILDE] = ACTIONS(117),
    [aux_sym_prefix_op_token1] = ACTIONS(89),
    [sym_int] = ACTIONS(617),
    [sym_xint] = ACTIONS(619),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(621),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(623),
  },
  [STATE(2221)] = {
    [sym_block_comment] = STATE(2221),
    [sym_compiler_directive_decl] = STATE(2221),
    [sym_fsi_directive_decl] = STATE(2221),
    [sym_preproc_line] = STATE(2221),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(5955),
    [anon_sym_uy] = ACTIONS(5957),
    [anon_sym_s] = ACTIONS(5959),
    [anon_sym_us] = ACTIONS(5961),
    [anon_sym_l] = ACTIONS(5963),
    [aux_sym_uint32_token1] = ACTIONS(5965),
    [anon_sym_n] = ACTIONS(5967),
    [anon_sym_un] = ACTIONS(5969),
    [anon_sym_L] = ACTIONS(5971),
    [aux_sym_uint64_token1] = ACTIONS(5973),
    [aux_sym_bignum_token1] = ACTIONS(5975),
    [aux_sym_decimal_token1] = ACTIONS(5977),
    [anon_sym_DOT2] = ACTIONS(5979),
    [aux_sym_float_token1] = ACTIONS(5981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2222)] = {
    [sym_block_comment] = STATE(2222),
    [sym_compiler_directive_decl] = STATE(2222),
    [sym_fsi_directive_decl] = STATE(2222),
    [sym_preproc_line] = STATE(2222),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(5983),
    [anon_sym_uy] = ACTIONS(5985),
    [anon_sym_s] = ACTIONS(5987),
    [anon_sym_us] = ACTIONS(5989),
    [anon_sym_l] = ACTIONS(5991),
    [aux_sym_uint32_token1] = ACTIONS(5993),
    [anon_sym_n] = ACTIONS(5995),
    [anon_sym_un] = ACTIONS(5997),
    [anon_sym_L] = ACTIONS(5999),
    [aux_sym_uint64_token1] = ACTIONS(6001),
    [aux_sym_bignum_token1] = ACTIONS(6003),
    [aux_sym_decimal_token1] = ACTIONS(6005),
    [anon_sym_DOT2] = ACTIONS(6007),
    [aux_sym_float_token1] = ACTIONS(6009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2223)] = {
    [sym_block_comment] = STATE(2223),
    [sym_compiler_directive_decl] = STATE(2223),
    [sym_fsi_directive_decl] = STATE(2223),
    [sym_preproc_line] = STATE(2223),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6011),
    [anon_sym_uy] = ACTIONS(6013),
    [anon_sym_s] = ACTIONS(6015),
    [anon_sym_us] = ACTIONS(6017),
    [anon_sym_l] = ACTIONS(6019),
    [aux_sym_uint32_token1] = ACTIONS(6021),
    [anon_sym_n] = ACTIONS(6023),
    [anon_sym_un] = ACTIONS(6025),
    [anon_sym_L] = ACTIONS(6027),
    [aux_sym_uint64_token1] = ACTIONS(6029),
    [aux_sym_bignum_token1] = ACTIONS(6031),
    [aux_sym_decimal_token1] = ACTIONS(6033),
    [anon_sym_DOT2] = ACTIONS(6035),
    [aux_sym_float_token1] = ACTIONS(6037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2224)] = {
    [sym_block_comment] = STATE(2224),
    [sym_compiler_directive_decl] = STATE(2224),
    [sym_fsi_directive_decl] = STATE(2224),
    [sym_preproc_line] = STATE(2224),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(5983),
    [anon_sym_uy] = ACTIONS(5985),
    [anon_sym_s] = ACTIONS(5987),
    [anon_sym_us] = ACTIONS(5989),
    [anon_sym_l] = ACTIONS(5991),
    [aux_sym_uint32_token1] = ACTIONS(5993),
    [anon_sym_n] = ACTIONS(5995),
    [anon_sym_un] = ACTIONS(5997),
    [anon_sym_L] = ACTIONS(5999),
    [aux_sym_uint64_token1] = ACTIONS(6001),
    [aux_sym_bignum_token1] = ACTIONS(6003),
    [aux_sym_decimal_token1] = ACTIONS(6005),
    [anon_sym_DOT2] = ACTIONS(6007),
    [aux_sym_float_token1] = ACTIONS(6009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2225)] = {
    [sym_block_comment] = STATE(2225),
    [sym_compiler_directive_decl] = STATE(2225),
    [sym_fsi_directive_decl] = STATE(2225),
    [sym_preproc_line] = STATE(2225),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(5955),
    [anon_sym_uy] = ACTIONS(5957),
    [anon_sym_s] = ACTIONS(5959),
    [anon_sym_us] = ACTIONS(5961),
    [anon_sym_l] = ACTIONS(5963),
    [aux_sym_uint32_token1] = ACTIONS(5965),
    [anon_sym_n] = ACTIONS(5967),
    [anon_sym_un] = ACTIONS(5969),
    [anon_sym_L] = ACTIONS(5971),
    [aux_sym_uint64_token1] = ACTIONS(5973),
    [anon_sym_lf] = ACTIONS(6039),
    [anon_sym_LF] = ACTIONS(6041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2226)] = {
    [sym_block_comment] = STATE(2226),
    [sym_compiler_directive_decl] = STATE(2226),
    [sym_fsi_directive_decl] = STATE(2226),
    [sym_preproc_line] = STATE(2226),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(5983),
    [anon_sym_uy] = ACTIONS(5985),
    [anon_sym_s] = ACTIONS(5987),
    [anon_sym_us] = ACTIONS(5989),
    [anon_sym_l] = ACTIONS(5991),
    [aux_sym_uint32_token1] = ACTIONS(5993),
    [anon_sym_n] = ACTIONS(5995),
    [anon_sym_un] = ACTIONS(5997),
    [anon_sym_L] = ACTIONS(5999),
    [aux_sym_uint64_token1] = ACTIONS(6001),
    [anon_sym_lf] = ACTIONS(6043),
    [anon_sym_LF] = ACTIONS(6045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2227)] = {
    [sym_block_comment] = STATE(2227),
    [sym_compiler_directive_decl] = STATE(2227),
    [sym_fsi_directive_decl] = STATE(2227),
    [sym_preproc_line] = STATE(2227),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(5983),
    [anon_sym_uy] = ACTIONS(5985),
    [anon_sym_s] = ACTIONS(5987),
    [anon_sym_us] = ACTIONS(5989),
    [anon_sym_l] = ACTIONS(5991),
    [aux_sym_uint32_token1] = ACTIONS(5993),
    [anon_sym_n] = ACTIONS(5995),
    [anon_sym_un] = ACTIONS(5997),
    [anon_sym_L] = ACTIONS(5999),
    [aux_sym_uint64_token1] = ACTIONS(6001),
    [anon_sym_lf] = ACTIONS(6043),
    [anon_sym_LF] = ACTIONS(6045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2228)] = {
    [sym_block_comment] = STATE(2228),
    [sym_compiler_directive_decl] = STATE(2228),
    [sym_fsi_directive_decl] = STATE(2228),
    [sym_preproc_line] = STATE(2228),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6011),
    [anon_sym_uy] = ACTIONS(6013),
    [anon_sym_s] = ACTIONS(6015),
    [anon_sym_us] = ACTIONS(6017),
    [anon_sym_l] = ACTIONS(6019),
    [aux_sym_uint32_token1] = ACTIONS(6021),
    [anon_sym_n] = ACTIONS(6023),
    [anon_sym_un] = ACTIONS(6025),
    [anon_sym_L] = ACTIONS(6027),
    [aux_sym_uint64_token1] = ACTIONS(6029),
    [anon_sym_lf] = ACTIONS(6047),
    [anon_sym_LF] = ACTIONS(6049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2229)] = {
    [sym_block_comment] = STATE(2229),
    [sym_compiler_directive_decl] = STATE(2229),
    [sym_fsi_directive_decl] = STATE(2229),
    [sym_preproc_line] = STATE(2229),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6051),
    [anon_sym_uy] = ACTIONS(6053),
    [anon_sym_s] = ACTIONS(6055),
    [anon_sym_us] = ACTIONS(6057),
    [anon_sym_l] = ACTIONS(6059),
    [aux_sym_uint32_token1] = ACTIONS(6061),
    [anon_sym_n] = ACTIONS(6063),
    [anon_sym_un] = ACTIONS(6065),
    [anon_sym_L] = ACTIONS(6067),
    [aux_sym_uint64_token1] = ACTIONS(6069),
    [aux_sym_bignum_token1] = ACTIONS(6071),
    [aux_sym_decimal_token1] = ACTIONS(6073),
    [anon_sym_DOT2] = ACTIONS(6075),
    [aux_sym_float_token1] = ACTIONS(6077),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2230)] = {
    [sym_block_comment] = STATE(2230),
    [sym_compiler_directive_decl] = STATE(2230),
    [sym_fsi_directive_decl] = STATE(2230),
    [sym_preproc_line] = STATE(2230),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6079),
    [anon_sym_uy] = ACTIONS(6081),
    [anon_sym_s] = ACTIONS(6083),
    [anon_sym_us] = ACTIONS(6085),
    [anon_sym_l] = ACTIONS(6087),
    [aux_sym_uint32_token1] = ACTIONS(6089),
    [anon_sym_n] = ACTIONS(6091),
    [anon_sym_un] = ACTIONS(6093),
    [anon_sym_L] = ACTIONS(6095),
    [aux_sym_uint64_token1] = ACTIONS(6097),
    [aux_sym_bignum_token1] = ACTIONS(6099),
    [aux_sym_decimal_token1] = ACTIONS(6101),
    [anon_sym_DOT2] = ACTIONS(6103),
    [aux_sym_float_token1] = ACTIONS(6105),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2231)] = {
    [sym_block_comment] = STATE(2231),
    [sym_compiler_directive_decl] = STATE(2231),
    [sym_fsi_directive_decl] = STATE(2231),
    [sym_preproc_line] = STATE(2231),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6107),
    [anon_sym_uy] = ACTIONS(6109),
    [anon_sym_s] = ACTIONS(6111),
    [anon_sym_us] = ACTIONS(6113),
    [anon_sym_l] = ACTIONS(6115),
    [aux_sym_uint32_token1] = ACTIONS(6117),
    [anon_sym_n] = ACTIONS(6119),
    [anon_sym_un] = ACTIONS(6121),
    [anon_sym_L] = ACTIONS(6123),
    [aux_sym_uint64_token1] = ACTIONS(6125),
    [aux_sym_bignum_token1] = ACTIONS(6127),
    [aux_sym_decimal_token1] = ACTIONS(6129),
    [anon_sym_DOT2] = ACTIONS(6131),
    [aux_sym_float_token1] = ACTIONS(6133),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2232)] = {
    [sym_block_comment] = STATE(2232),
    [sym_compiler_directive_decl] = STATE(2232),
    [sym_fsi_directive_decl] = STATE(2232),
    [sym_preproc_line] = STATE(2232),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6135),
    [anon_sym_uy] = ACTIONS(6137),
    [anon_sym_s] = ACTIONS(6139),
    [anon_sym_us] = ACTIONS(6141),
    [anon_sym_l] = ACTIONS(6143),
    [aux_sym_uint32_token1] = ACTIONS(6145),
    [anon_sym_n] = ACTIONS(6147),
    [anon_sym_un] = ACTIONS(6149),
    [anon_sym_L] = ACTIONS(6151),
    [aux_sym_uint64_token1] = ACTIONS(6153),
    [aux_sym_bignum_token1] = ACTIONS(6155),
    [aux_sym_decimal_token1] = ACTIONS(6157),
    [anon_sym_DOT2] = ACTIONS(6159),
    [aux_sym_float_token1] = ACTIONS(6161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2233)] = {
    [sym_block_comment] = STATE(2233),
    [sym_compiler_directive_decl] = STATE(2233),
    [sym_fsi_directive_decl] = STATE(2233),
    [sym_preproc_line] = STATE(2233),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6163),
    [anon_sym_uy] = ACTIONS(6165),
    [anon_sym_s] = ACTIONS(6167),
    [anon_sym_us] = ACTIONS(6169),
    [anon_sym_l] = ACTIONS(6171),
    [aux_sym_uint32_token1] = ACTIONS(6173),
    [anon_sym_n] = ACTIONS(6175),
    [anon_sym_un] = ACTIONS(6177),
    [anon_sym_L] = ACTIONS(6179),
    [aux_sym_uint64_token1] = ACTIONS(6181),
    [aux_sym_bignum_token1] = ACTIONS(6183),
    [aux_sym_decimal_token1] = ACTIONS(6185),
    [anon_sym_DOT2] = ACTIONS(6187),
    [aux_sym_float_token1] = ACTIONS(6189),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2234)] = {
    [sym_type_arguments] = STATE(2282),
    [sym_long_identifier] = STATE(2284),
    [sym_block_comment] = STATE(2234),
    [sym_compiler_directive_decl] = STATE(2234),
    [sym_fsi_directive_decl] = STATE(2234),
    [sym_preproc_line] = STATE(2234),
    [aux_sym_compound_type_repeat1] = STATE(2264),
    [sym_identifier] = ACTIONS(6191),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6197),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6199),
    [anon_sym_LT] = ACTIONS(6201),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6203),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2235)] = {
    [sym_block_comment] = STATE(2235),
    [sym_compiler_directive_decl] = STATE(2235),
    [sym_fsi_directive_decl] = STATE(2235),
    [sym_preproc_line] = STATE(2235),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6079),
    [anon_sym_uy] = ACTIONS(6081),
    [anon_sym_s] = ACTIONS(6083),
    [anon_sym_us] = ACTIONS(6085),
    [anon_sym_l] = ACTIONS(6087),
    [aux_sym_uint32_token1] = ACTIONS(6089),
    [anon_sym_n] = ACTIONS(6091),
    [anon_sym_un] = ACTIONS(6093),
    [anon_sym_L] = ACTIONS(6095),
    [aux_sym_uint64_token1] = ACTIONS(6097),
    [anon_sym_lf] = ACTIONS(6205),
    [anon_sym_LF] = ACTIONS(6207),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2236)] = {
    [sym_type_arguments] = STATE(2298),
    [sym_long_identifier] = STATE(2325),
    [sym_block_comment] = STATE(2236),
    [sym_compiler_directive_decl] = STATE(2236),
    [sym_fsi_directive_decl] = STATE(2236),
    [sym_preproc_line] = STATE(2236),
    [aux_sym_compound_type_repeat1] = STATE(2268),
    [ts_builtin_sym_end] = ACTIONS(6209),
    [sym_identifier] = ACTIONS(6211),
    [anon_sym_namespace] = ACTIONS(6213),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6215),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6217),
    [anon_sym_LT] = ACTIONS(6219),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6221),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2237)] = {
    [sym_block_comment] = STATE(2237),
    [sym_compiler_directive_decl] = STATE(2237),
    [sym_fsi_directive_decl] = STATE(2237),
    [sym_preproc_line] = STATE(2237),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6051),
    [anon_sym_uy] = ACTIONS(6053),
    [anon_sym_s] = ACTIONS(6055),
    [anon_sym_us] = ACTIONS(6057),
    [anon_sym_l] = ACTIONS(6059),
    [aux_sym_uint32_token1] = ACTIONS(6061),
    [anon_sym_n] = ACTIONS(6063),
    [anon_sym_un] = ACTIONS(6065),
    [anon_sym_L] = ACTIONS(6067),
    [aux_sym_uint64_token1] = ACTIONS(6069),
    [anon_sym_lf] = ACTIONS(6223),
    [anon_sym_LF] = ACTIONS(6225),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2238)] = {
    [sym_type_arguments] = STATE(2298),
    [sym_long_identifier] = STATE(2325),
    [sym_block_comment] = STATE(2238),
    [sym_compiler_directive_decl] = STATE(2238),
    [sym_fsi_directive_decl] = STATE(2238),
    [sym_preproc_line] = STATE(2238),
    [aux_sym_compound_type_repeat1] = STATE(2268),
    [ts_builtin_sym_end] = ACTIONS(6195),
    [sym_identifier] = ACTIONS(6211),
    [anon_sym_namespace] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6215),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6217),
    [anon_sym_LT] = ACTIONS(6219),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6221),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2239)] = {
    [sym_block_comment] = STATE(2239),
    [sym_compiler_directive_decl] = STATE(2239),
    [sym_fsi_directive_decl] = STATE(2239),
    [sym_preproc_line] = STATE(2239),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6227),
    [anon_sym_uy] = ACTIONS(6229),
    [anon_sym_s] = ACTIONS(6231),
    [anon_sym_us] = ACTIONS(6233),
    [anon_sym_l] = ACTIONS(6235),
    [aux_sym_uint32_token1] = ACTIONS(6237),
    [anon_sym_n] = ACTIONS(6239),
    [anon_sym_un] = ACTIONS(6241),
    [anon_sym_L] = ACTIONS(6243),
    [aux_sym_uint64_token1] = ACTIONS(6245),
    [aux_sym_bignum_token1] = ACTIONS(6247),
    [aux_sym_decimal_token1] = ACTIONS(6249),
    [anon_sym_DOT2] = ACTIONS(6251),
    [aux_sym_float_token1] = ACTIONS(6253),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2240)] = {
    [sym_type_arguments] = STATE(2298),
    [sym_long_identifier] = STATE(2325),
    [sym_block_comment] = STATE(2240),
    [sym_compiler_directive_decl] = STATE(2240),
    [sym_fsi_directive_decl] = STATE(2240),
    [sym_preproc_line] = STATE(2240),
    [aux_sym_compound_type_repeat1] = STATE(2268),
    [ts_builtin_sym_end] = ACTIONS(6255),
    [sym_identifier] = ACTIONS(6211),
    [anon_sym_namespace] = ACTIONS(6257),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6215),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6217),
    [anon_sym_LT] = ACTIONS(6219),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6221),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2241)] = {
    [sym_type_arguments] = STATE(2298),
    [sym_long_identifier] = STATE(2325),
    [sym_block_comment] = STATE(2241),
    [sym_compiler_directive_decl] = STATE(2241),
    [sym_fsi_directive_decl] = STATE(2241),
    [sym_preproc_line] = STATE(2241),
    [aux_sym_compound_type_repeat1] = STATE(2268),
    [ts_builtin_sym_end] = ACTIONS(6259),
    [sym_identifier] = ACTIONS(6211),
    [anon_sym_namespace] = ACTIONS(6261),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6215),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6217),
    [anon_sym_LT] = ACTIONS(6219),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6221),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2242)] = {
    [sym_type_arguments] = STATE(2298),
    [sym_long_identifier] = STATE(2325),
    [sym_block_comment] = STATE(2242),
    [sym_compiler_directive_decl] = STATE(2242),
    [sym_fsi_directive_decl] = STATE(2242),
    [sym_preproc_line] = STATE(2242),
    [aux_sym_compound_type_repeat1] = STATE(2268),
    [ts_builtin_sym_end] = ACTIONS(6263),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_namespace] = ACTIONS(6265),
    [anon_sym_module] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_open] = ACTIONS(6265),
    [anon_sym_type] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_and] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6215),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6217),
    [anon_sym_LT] = ACTIONS(6219),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6221),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_exception] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [anon_sym_extern] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6263),
    [anon_sym_POUNDload] = ACTIONS(6263),
    [anon_sym_POUNDtime] = ACTIONS(6263),
    [anon_sym_POUNDI] = ACTIONS(6263),
    [anon_sym_POUNDhelp] = ACTIONS(6263),
    [anon_sym_POUNDquit] = ACTIONS(6263),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2243)] = {
    [sym_type_arguments] = STATE(2282),
    [sym_long_identifier] = STATE(2284),
    [sym_block_comment] = STATE(2243),
    [sym_compiler_directive_decl] = STATE(2243),
    [sym_fsi_directive_decl] = STATE(2243),
    [sym_preproc_line] = STATE(2243),
    [aux_sym_compound_type_repeat1] = STATE(2264),
    [sym_identifier] = ACTIONS(6191),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6197),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6199),
    [anon_sym_LT] = ACTIONS(6201),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6203),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2244)] = {
    [sym_type_arguments] = STATE(2282),
    [sym_long_identifier] = STATE(2284),
    [sym_block_comment] = STATE(2244),
    [sym_compiler_directive_decl] = STATE(2244),
    [sym_fsi_directive_decl] = STATE(2244),
    [sym_preproc_line] = STATE(2244),
    [aux_sym_compound_type_repeat1] = STATE(2264),
    [sym_identifier] = ACTIONS(6191),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6197),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6199),
    [anon_sym_LT] = ACTIONS(6201),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6203),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2245)] = {
    [sym_type_arguments] = STATE(2282),
    [sym_long_identifier] = STATE(2284),
    [sym_block_comment] = STATE(2245),
    [sym_compiler_directive_decl] = STATE(2245),
    [sym_fsi_directive_decl] = STATE(2245),
    [sym_preproc_line] = STATE(2245),
    [aux_sym_compound_type_repeat1] = STATE(2264),
    [sym_identifier] = ACTIONS(6191),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6197),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6199),
    [anon_sym_LT] = ACTIONS(6201),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6203),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2246)] = {
    [sym_type_arguments] = STATE(2282),
    [sym_long_identifier] = STATE(2284),
    [sym_block_comment] = STATE(2246),
    [sym_compiler_directive_decl] = STATE(2246),
    [sym_fsi_directive_decl] = STATE(2246),
    [sym_preproc_line] = STATE(2246),
    [aux_sym_compound_type_repeat1] = STATE(2264),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_module] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_open] = ACTIONS(6265),
    [anon_sym_type] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_and] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6197),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6199),
    [anon_sym_LT] = ACTIONS(6201),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6203),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_exception] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [anon_sym_extern] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6263),
    [anon_sym_POUNDload] = ACTIONS(6263),
    [anon_sym_POUNDtime] = ACTIONS(6263),
    [anon_sym_POUNDI] = ACTIONS(6263),
    [anon_sym_POUNDhelp] = ACTIONS(6263),
    [anon_sym_POUNDquit] = ACTIONS(6263),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [anon_sym_POUNDelse] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2247)] = {
    [sym_block_comment] = STATE(2247),
    [sym_compiler_directive_decl] = STATE(2247),
    [sym_fsi_directive_decl] = STATE(2247),
    [sym_preproc_line] = STATE(2247),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6135),
    [anon_sym_uy] = ACTIONS(6137),
    [anon_sym_s] = ACTIONS(6139),
    [anon_sym_us] = ACTIONS(6141),
    [anon_sym_l] = ACTIONS(6143),
    [aux_sym_uint32_token1] = ACTIONS(6145),
    [anon_sym_n] = ACTIONS(6147),
    [anon_sym_un] = ACTIONS(6149),
    [anon_sym_L] = ACTIONS(6151),
    [aux_sym_uint64_token1] = ACTIONS(6153),
    [anon_sym_lf] = ACTIONS(6267),
    [anon_sym_LF] = ACTIONS(6269),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2248)] = {
    [sym_block_comment] = STATE(2248),
    [sym_compiler_directive_decl] = STATE(2248),
    [sym_fsi_directive_decl] = STATE(2248),
    [sym_preproc_line] = STATE(2248),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6107),
    [anon_sym_uy] = ACTIONS(6109),
    [anon_sym_s] = ACTIONS(6111),
    [anon_sym_us] = ACTIONS(6113),
    [anon_sym_l] = ACTIONS(6115),
    [aux_sym_uint32_token1] = ACTIONS(6117),
    [anon_sym_n] = ACTIONS(6119),
    [anon_sym_un] = ACTIONS(6121),
    [anon_sym_L] = ACTIONS(6123),
    [aux_sym_uint64_token1] = ACTIONS(6125),
    [anon_sym_lf] = ACTIONS(6271),
    [anon_sym_LF] = ACTIONS(6273),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2249)] = {
    [sym_block_comment] = STATE(2249),
    [sym_compiler_directive_decl] = STATE(2249),
    [sym_fsi_directive_decl] = STATE(2249),
    [sym_preproc_line] = STATE(2249),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6163),
    [anon_sym_uy] = ACTIONS(6165),
    [anon_sym_s] = ACTIONS(6167),
    [anon_sym_us] = ACTIONS(6169),
    [anon_sym_l] = ACTIONS(6171),
    [aux_sym_uint32_token1] = ACTIONS(6173),
    [anon_sym_n] = ACTIONS(6175),
    [anon_sym_un] = ACTIONS(6177),
    [anon_sym_L] = ACTIONS(6179),
    [aux_sym_uint64_token1] = ACTIONS(6181),
    [anon_sym_lf] = ACTIONS(6275),
    [anon_sym_LF] = ACTIONS(6277),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2250)] = {
    [sym_type_arguments] = STATE(2335),
    [sym_long_identifier] = STATE(2337),
    [sym_block_comment] = STATE(2250),
    [sym_compiler_directive_decl] = STATE(2250),
    [sym_fsi_directive_decl] = STATE(2250),
    [sym_preproc_line] = STATE(2250),
    [aux_sym_compound_type_repeat1] = STATE(2293),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_module] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_open] = ACTIONS(6265),
    [anon_sym_type] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_and] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6279),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6281),
    [anon_sym_LT] = ACTIONS(6283),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6285),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_exception] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [anon_sym_extern] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6263),
    [anon_sym_POUNDload] = ACTIONS(6263),
    [anon_sym_POUNDtime] = ACTIONS(6263),
    [anon_sym_POUNDI] = ACTIONS(6263),
    [anon_sym_POUNDhelp] = ACTIONS(6263),
    [anon_sym_POUNDquit] = ACTIONS(6263),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2251)] = {
    [sym_type_arguments] = STATE(2335),
    [sym_long_identifier] = STATE(2337),
    [sym_block_comment] = STATE(2251),
    [sym_compiler_directive_decl] = STATE(2251),
    [sym_fsi_directive_decl] = STATE(2251),
    [sym_preproc_line] = STATE(2251),
    [aux_sym_compound_type_repeat1] = STATE(2293),
    [sym_identifier] = ACTIONS(6287),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6279),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6281),
    [anon_sym_LT] = ACTIONS(6283),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6285),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2252)] = {
    [sym_type_arguments] = STATE(2335),
    [sym_long_identifier] = STATE(2337),
    [sym_block_comment] = STATE(2252),
    [sym_compiler_directive_decl] = STATE(2252),
    [sym_fsi_directive_decl] = STATE(2252),
    [sym_preproc_line] = STATE(2252),
    [aux_sym_compound_type_repeat1] = STATE(2293),
    [sym_identifier] = ACTIONS(6287),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6279),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6281),
    [anon_sym_LT] = ACTIONS(6283),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6285),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2253)] = {
    [sym_type_arguments] = STATE(2335),
    [sym_long_identifier] = STATE(2337),
    [sym_block_comment] = STATE(2253),
    [sym_compiler_directive_decl] = STATE(2253),
    [sym_fsi_directive_decl] = STATE(2253),
    [sym_preproc_line] = STATE(2253),
    [aux_sym_compound_type_repeat1] = STATE(2293),
    [sym_identifier] = ACTIONS(6287),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6279),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6281),
    [anon_sym_LT] = ACTIONS(6283),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6285),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2254)] = {
    [sym_type_arguments] = STATE(2335),
    [sym_long_identifier] = STATE(2337),
    [sym_block_comment] = STATE(2254),
    [sym_compiler_directive_decl] = STATE(2254),
    [sym_fsi_directive_decl] = STATE(2254),
    [sym_preproc_line] = STATE(2254),
    [aux_sym_compound_type_repeat1] = STATE(2293),
    [sym_identifier] = ACTIONS(6287),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6279),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6281),
    [anon_sym_LT] = ACTIONS(6283),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6285),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2255)] = {
    [sym_block_comment] = STATE(2255),
    [sym_compiler_directive_decl] = STATE(2255),
    [sym_fsi_directive_decl] = STATE(2255),
    [sym_preproc_line] = STATE(2255),
    [aux_sym_long_identifier_repeat1] = STATE(2275),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(6293),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2256)] = {
    [sym_block_comment] = STATE(2256),
    [sym_compiler_directive_decl] = STATE(2256),
    [sym_fsi_directive_decl] = STATE(2256),
    [sym_preproc_line] = STATE(2256),
    [aux_sym_long_identifier_repeat1] = STATE(2256),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6299),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2257)] = {
    [sym_block_comment] = STATE(2257),
    [sym_compiler_directive_decl] = STATE(2257),
    [sym_fsi_directive_decl] = STATE(2257),
    [sym_preproc_line] = STATE(2257),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6227),
    [anon_sym_uy] = ACTIONS(6229),
    [anon_sym_s] = ACTIONS(6231),
    [anon_sym_us] = ACTIONS(6233),
    [anon_sym_l] = ACTIONS(6235),
    [aux_sym_uint32_token1] = ACTIONS(6237),
    [anon_sym_n] = ACTIONS(6239),
    [anon_sym_un] = ACTIONS(6241),
    [anon_sym_L] = ACTIONS(6243),
    [aux_sym_uint64_token1] = ACTIONS(6245),
    [anon_sym_lf] = ACTIONS(6302),
    [anon_sym_LF] = ACTIONS(6304),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2258)] = {
    [sym_block_comment] = STATE(2258),
    [sym_compiler_directive_decl] = STATE(2258),
    [sym_fsi_directive_decl] = STATE(2258),
    [sym_preproc_line] = STATE(2258),
    [aux_sym_type_argument_repeat1] = STATE(2259),
    [ts_builtin_sym_end] = ACTIONS(6306),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_namespace] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2259)] = {
    [sym_block_comment] = STATE(2259),
    [sym_compiler_directive_decl] = STATE(2259),
    [sym_fsi_directive_decl] = STATE(2259),
    [sym_preproc_line] = STATE(2259),
    [aux_sym_type_argument_repeat1] = STATE(2259),
    [ts_builtin_sym_end] = ACTIONS(6310),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_namespace] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6314),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2260)] = {
    [sym__else_expression] = STATE(2514),
    [sym_elif_expression] = STATE(2354),
    [sym_block_comment] = STATE(2260),
    [sym_compiler_directive_decl] = STATE(2260),
    [sym_fsi_directive_decl] = STATE(2260),
    [sym_preproc_line] = STATE(2260),
    [aux_sym_if_expression_repeat1] = STATE(2262),
    [ts_builtin_sym_end] = ACTIONS(6317),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_namespace] = ACTIONS(6319),
    [anon_sym_module] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_open] = ACTIONS(6319),
    [anon_sym_type] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_and] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(6321),
    [anon_sym_elif] = ACTIONS(6323),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_exception] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [anon_sym_extern] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6317),
    [anon_sym_POUNDload] = ACTIONS(6317),
    [anon_sym_POUNDtime] = ACTIONS(6317),
    [anon_sym_POUNDI] = ACTIONS(6317),
    [anon_sym_POUNDhelp] = ACTIONS(6317),
    [anon_sym_POUNDquit] = ACTIONS(6317),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2261)] = {
    [sym_block_comment] = STATE(2261),
    [sym_compiler_directive_decl] = STATE(2261),
    [sym_fsi_directive_decl] = STATE(2261),
    [sym_preproc_line] = STATE(2261),
    [aux_sym_long_identifier_repeat1] = STATE(2272),
    [ts_builtin_sym_end] = ACTIONS(6291),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_namespace] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(6325),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2262)] = {
    [sym__else_expression] = STATE(2524),
    [sym_elif_expression] = STATE(2354),
    [sym_block_comment] = STATE(2262),
    [sym_compiler_directive_decl] = STATE(2262),
    [sym_fsi_directive_decl] = STATE(2262),
    [sym_preproc_line] = STATE(2262),
    [aux_sym_if_expression_repeat1] = STATE(2314),
    [ts_builtin_sym_end] = ACTIONS(6327),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_namespace] = ACTIONS(6329),
    [anon_sym_module] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_open] = ACTIONS(6329),
    [anon_sym_type] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_and] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(6321),
    [anon_sym_elif] = ACTIONS(6323),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_exception] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [anon_sym_extern] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6327),
    [anon_sym_POUNDload] = ACTIONS(6327),
    [anon_sym_POUNDtime] = ACTIONS(6327),
    [anon_sym_POUNDI] = ACTIONS(6327),
    [anon_sym_POUNDhelp] = ACTIONS(6327),
    [anon_sym_POUNDquit] = ACTIONS(6327),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2263)] = {
    [sym_block_comment] = STATE(2263),
    [sym_compiler_directive_decl] = STATE(2263),
    [sym_fsi_directive_decl] = STATE(2263),
    [sym_preproc_line] = STATE(2263),
    [aux_sym_type_argument_repeat1] = STATE(2269),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6335),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2264)] = {
    [sym_block_comment] = STATE(2264),
    [sym_compiler_directive_decl] = STATE(2264),
    [sym_fsi_directive_decl] = STATE(2264),
    [sym_preproc_line] = STATE(2264),
    [aux_sym_compound_type_repeat1] = STATE(2265),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6199),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [anon_sym_POUNDelse] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2265)] = {
    [sym_block_comment] = STATE(2265),
    [sym_compiler_directive_decl] = STATE(2265),
    [sym_fsi_directive_decl] = STATE(2265),
    [sym_preproc_line] = STATE(2265),
    [aux_sym_compound_type_repeat1] = STATE(2265),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6342),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2266)] = {
    [sym_block_comment] = STATE(2266),
    [sym_compiler_directive_decl] = STATE(2266),
    [sym_fsi_directive_decl] = STATE(2266),
    [sym_preproc_line] = STATE(2266),
    [aux_sym_type_argument_repeat1] = STATE(2258),
    [ts_builtin_sym_end] = ACTIONS(6333),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_namespace] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6345),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2267)] = {
    [sym_block_comment] = STATE(2267),
    [sym_compiler_directive_decl] = STATE(2267),
    [sym_fsi_directive_decl] = STATE(2267),
    [sym_preproc_line] = STATE(2267),
    [aux_sym_compound_type_repeat1] = STATE(2267),
    [ts_builtin_sym_end] = ACTIONS(6195),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_namespace] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6348),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2268)] = {
    [sym_block_comment] = STATE(2268),
    [sym_compiler_directive_decl] = STATE(2268),
    [sym_fsi_directive_decl] = STATE(2268),
    [sym_preproc_line] = STATE(2268),
    [aux_sym_compound_type_repeat1] = STATE(2267),
    [ts_builtin_sym_end] = ACTIONS(6340),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_namespace] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6217),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2269)] = {
    [sym_block_comment] = STATE(2269),
    [sym_compiler_directive_decl] = STATE(2269),
    [sym_fsi_directive_decl] = STATE(2269),
    [sym_preproc_line] = STATE(2269),
    [aux_sym_type_argument_repeat1] = STATE(2270),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2270)] = {
    [sym_block_comment] = STATE(2270),
    [sym_compiler_directive_decl] = STATE(2270),
    [sym_fsi_directive_decl] = STATE(2270),
    [sym_preproc_line] = STATE(2270),
    [aux_sym_type_argument_repeat1] = STATE(2270),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6351),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2271)] = {
    [sym__else_expression] = STATE(2614),
    [sym_elif_expression] = STATE(2398),
    [sym_block_comment] = STATE(2271),
    [sym_compiler_directive_decl] = STATE(2271),
    [sym_fsi_directive_decl] = STATE(2271),
    [sym_preproc_line] = STATE(2271),
    [aux_sym_if_expression_repeat1] = STATE(2274),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_module] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_open] = ACTIONS(6319),
    [anon_sym_type] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_and] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(6354),
    [anon_sym_elif] = ACTIONS(6356),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_exception] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [anon_sym_extern] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6317),
    [anon_sym_POUNDload] = ACTIONS(6317),
    [anon_sym_POUNDtime] = ACTIONS(6317),
    [anon_sym_POUNDI] = ACTIONS(6317),
    [anon_sym_POUNDhelp] = ACTIONS(6317),
    [anon_sym_POUNDquit] = ACTIONS(6317),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [anon_sym_POUNDelse] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2272)] = {
    [sym_block_comment] = STATE(2272),
    [sym_compiler_directive_decl] = STATE(2272),
    [sym_fsi_directive_decl] = STATE(2272),
    [sym_preproc_line] = STATE(2272),
    [aux_sym_long_identifier_repeat1] = STATE(2273),
    [ts_builtin_sym_end] = ACTIONS(6358),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_namespace] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(6325),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2273)] = {
    [sym_block_comment] = STATE(2273),
    [sym_compiler_directive_decl] = STATE(2273),
    [sym_fsi_directive_decl] = STATE(2273),
    [sym_preproc_line] = STATE(2273),
    [aux_sym_long_identifier_repeat1] = STATE(2273),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6362),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2274)] = {
    [sym__else_expression] = STATE(2632),
    [sym_elif_expression] = STATE(2398),
    [sym_block_comment] = STATE(2274),
    [sym_compiler_directive_decl] = STATE(2274),
    [sym_fsi_directive_decl] = STATE(2274),
    [sym_preproc_line] = STATE(2274),
    [aux_sym_if_expression_repeat1] = STATE(2320),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_module] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_open] = ACTIONS(6329),
    [anon_sym_type] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_and] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(6354),
    [anon_sym_elif] = ACTIONS(6356),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_exception] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [anon_sym_extern] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6327),
    [anon_sym_POUNDload] = ACTIONS(6327),
    [anon_sym_POUNDtime] = ACTIONS(6327),
    [anon_sym_POUNDI] = ACTIONS(6327),
    [anon_sym_POUNDhelp] = ACTIONS(6327),
    [anon_sym_POUNDquit] = ACTIONS(6327),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [anon_sym_POUNDelse] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2275)] = {
    [sym_block_comment] = STATE(2275),
    [sym_compiler_directive_decl] = STATE(2275),
    [sym_fsi_directive_decl] = STATE(2275),
    [sym_preproc_line] = STATE(2275),
    [aux_sym_long_identifier_repeat1] = STATE(2256),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(6293),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2276)] = {
    [sym_block_comment] = STATE(2276),
    [sym_compiler_directive_decl] = STATE(2276),
    [sym_fsi_directive_decl] = STATE(2276),
    [sym_preproc_line] = STATE(2276),
    [ts_builtin_sym_end] = ACTIONS(6365),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_namespace] = ACTIONS(6367),
    [anon_sym_module] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_open] = ACTIONS(6367),
    [anon_sym_type] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_and] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_exception] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [anon_sym_extern] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6365),
    [anon_sym_POUNDload] = ACTIONS(6365),
    [anon_sym_POUNDtime] = ACTIONS(6365),
    [anon_sym_POUNDI] = ACTIONS(6365),
    [anon_sym_POUNDhelp] = ACTIONS(6365),
    [anon_sym_POUNDquit] = ACTIONS(6365),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2277)] = {
    [sym_block_comment] = STATE(2277),
    [sym_compiler_directive_decl] = STATE(2277),
    [sym_fsi_directive_decl] = STATE(2277),
    [sym_preproc_line] = STATE(2277),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2278)] = {
    [sym_block_comment] = STATE(2278),
    [sym_compiler_directive_decl] = STATE(2278),
    [sym_fsi_directive_decl] = STATE(2278),
    [sym_preproc_line] = STATE(2278),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_module] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_open] = ACTIONS(6373),
    [anon_sym_type] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_and] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(6377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_exception] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [anon_sym_extern] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6375),
    [anon_sym_POUNDload] = ACTIONS(6375),
    [anon_sym_POUNDtime] = ACTIONS(6375),
    [anon_sym_POUNDI] = ACTIONS(6375),
    [anon_sym_POUNDhelp] = ACTIONS(6375),
    [anon_sym_POUNDquit] = ACTIONS(6375),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [anon_sym_POUNDelse] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2279)] = {
    [sym_block_comment] = STATE(2279),
    [sym_compiler_directive_decl] = STATE(2279),
    [sym_fsi_directive_decl] = STATE(2279),
    [sym_preproc_line] = STATE(2279),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_module] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_open] = ACTIONS(6367),
    [anon_sym_type] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_and] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_exception] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [anon_sym_extern] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6365),
    [anon_sym_POUNDload] = ACTIONS(6365),
    [anon_sym_POUNDtime] = ACTIONS(6365),
    [anon_sym_POUNDI] = ACTIONS(6365),
    [anon_sym_POUNDhelp] = ACTIONS(6365),
    [anon_sym_POUNDquit] = ACTIONS(6365),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [anon_sym_POUNDelse] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2280)] = {
    [sym_block_comment] = STATE(2280),
    [sym_compiler_directive_decl] = STATE(2280),
    [sym_fsi_directive_decl] = STATE(2280),
    [sym_preproc_line] = STATE(2280),
    [ts_builtin_sym_end] = ACTIONS(6371),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_namespace] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2281)] = {
    [sym_block_comment] = STATE(2281),
    [sym_compiler_directive_decl] = STATE(2281),
    [sym_fsi_directive_decl] = STATE(2281),
    [sym_preproc_line] = STATE(2281),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_module] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_open] = ACTIONS(6379),
    [anon_sym_type] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_and] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_exception] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [anon_sym_extern] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6381),
    [anon_sym_POUNDload] = ACTIONS(6381),
    [anon_sym_POUNDtime] = ACTIONS(6381),
    [anon_sym_POUNDI] = ACTIONS(6381),
    [anon_sym_POUNDhelp] = ACTIONS(6381),
    [anon_sym_POUNDquit] = ACTIONS(6381),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [anon_sym_POUNDelse] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2282)] = {
    [sym_block_comment] = STATE(2282),
    [sym_compiler_directive_decl] = STATE(2282),
    [sym_fsi_directive_decl] = STATE(2282),
    [sym_preproc_line] = STATE(2282),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_module] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_open] = ACTIONS(6383),
    [anon_sym_type] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_and] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_exception] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [anon_sym_extern] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6385),
    [anon_sym_POUNDload] = ACTIONS(6385),
    [anon_sym_POUNDtime] = ACTIONS(6385),
    [anon_sym_POUNDI] = ACTIONS(6385),
    [anon_sym_POUNDhelp] = ACTIONS(6385),
    [anon_sym_POUNDquit] = ACTIONS(6385),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [anon_sym_POUNDelse] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2283)] = {
    [sym_block_comment] = STATE(2283),
    [sym_compiler_directive_decl] = STATE(2283),
    [sym_fsi_directive_decl] = STATE(2283),
    [sym_preproc_line] = STATE(2283),
    [ts_builtin_sym_end] = ACTIONS(6387),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_namespace] = ACTIONS(6389),
    [anon_sym_module] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_open] = ACTIONS(6389),
    [anon_sym_type] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_and] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_exception] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [anon_sym_extern] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6387),
    [anon_sym_POUNDload] = ACTIONS(6387),
    [anon_sym_POUNDtime] = ACTIONS(6387),
    [anon_sym_POUNDI] = ACTIONS(6387),
    [anon_sym_POUNDhelp] = ACTIONS(6387),
    [anon_sym_POUNDquit] = ACTIONS(6387),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2284)] = {
    [sym_block_comment] = STATE(2284),
    [sym_compiler_directive_decl] = STATE(2284),
    [sym_fsi_directive_decl] = STATE(2284),
    [sym_preproc_line] = STATE(2284),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_module] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_open] = ACTIONS(6391),
    [anon_sym_type] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_and] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_exception] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [anon_sym_extern] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6393),
    [anon_sym_POUNDload] = ACTIONS(6393),
    [anon_sym_POUNDtime] = ACTIONS(6393),
    [anon_sym_POUNDI] = ACTIONS(6393),
    [anon_sym_POUNDhelp] = ACTIONS(6393),
    [anon_sym_POUNDquit] = ACTIONS(6393),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [anon_sym_POUNDelse] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2285)] = {
    [sym_block_comment] = STATE(2285),
    [sym_compiler_directive_decl] = STATE(2285),
    [sym_fsi_directive_decl] = STATE(2285),
    [sym_preproc_line] = STATE(2285),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_module] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_open] = ACTIONS(6395),
    [anon_sym_type] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_and] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_exception] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [anon_sym_extern] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6397),
    [anon_sym_POUNDload] = ACTIONS(6397),
    [anon_sym_POUNDtime] = ACTIONS(6397),
    [anon_sym_POUNDI] = ACTIONS(6397),
    [anon_sym_POUNDhelp] = ACTIONS(6397),
    [anon_sym_POUNDquit] = ACTIONS(6397),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [anon_sym_POUNDelse] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2286)] = {
    [sym_block_comment] = STATE(2286),
    [sym_compiler_directive_decl] = STATE(2286),
    [sym_fsi_directive_decl] = STATE(2286),
    [sym_preproc_line] = STATE(2286),
    [ts_builtin_sym_end] = ACTIONS(6333),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_namespace] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2287)] = {
    [sym_block_comment] = STATE(2287),
    [sym_compiler_directive_decl] = STATE(2287),
    [sym_fsi_directive_decl] = STATE(2287),
    [sym_preproc_line] = STATE(2287),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_module] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_open] = ACTIONS(6399),
    [anon_sym_type] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_and] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_exception] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [anon_sym_extern] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6401),
    [anon_sym_POUNDload] = ACTIONS(6401),
    [anon_sym_POUNDtime] = ACTIONS(6401),
    [anon_sym_POUNDI] = ACTIONS(6401),
    [anon_sym_POUNDhelp] = ACTIONS(6401),
    [anon_sym_POUNDquit] = ACTIONS(6401),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [anon_sym_POUNDelse] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2288)] = {
    [sym_block_comment] = STATE(2288),
    [sym_compiler_directive_decl] = STATE(2288),
    [sym_fsi_directive_decl] = STATE(2288),
    [sym_preproc_line] = STATE(2288),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_module] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_open] = ACTIONS(6403),
    [anon_sym_type] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_and] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_exception] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [anon_sym_extern] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6405),
    [anon_sym_POUNDload] = ACTIONS(6405),
    [anon_sym_POUNDtime] = ACTIONS(6405),
    [anon_sym_POUNDI] = ACTIONS(6405),
    [anon_sym_POUNDhelp] = ACTIONS(6405),
    [anon_sym_POUNDquit] = ACTIONS(6405),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [anon_sym_POUNDelse] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2289)] = {
    [sym_block_comment] = STATE(2289),
    [sym_compiler_directive_decl] = STATE(2289),
    [sym_fsi_directive_decl] = STATE(2289),
    [sym_preproc_line] = STATE(2289),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_module] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_open] = ACTIONS(6407),
    [anon_sym_type] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_and] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_exception] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [anon_sym_extern] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6409),
    [anon_sym_POUNDload] = ACTIONS(6409),
    [anon_sym_POUNDtime] = ACTIONS(6409),
    [anon_sym_POUNDI] = ACTIONS(6409),
    [anon_sym_POUNDhelp] = ACTIONS(6409),
    [anon_sym_POUNDquit] = ACTIONS(6409),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [anon_sym_POUNDelse] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2290)] = {
    [sym_block_comment] = STATE(2290),
    [sym_compiler_directive_decl] = STATE(2290),
    [sym_fsi_directive_decl] = STATE(2290),
    [sym_preproc_line] = STATE(2290),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_module] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_open] = ACTIONS(6411),
    [anon_sym_type] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_and] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_exception] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [anon_sym_extern] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6413),
    [anon_sym_POUNDload] = ACTIONS(6413),
    [anon_sym_POUNDtime] = ACTIONS(6413),
    [anon_sym_POUNDI] = ACTIONS(6413),
    [anon_sym_POUNDhelp] = ACTIONS(6413),
    [anon_sym_POUNDquit] = ACTIONS(6413),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [anon_sym_POUNDelse] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2291)] = {
    [sym_block_comment] = STATE(2291),
    [sym_compiler_directive_decl] = STATE(2291),
    [sym_fsi_directive_decl] = STATE(2291),
    [sym_preproc_line] = STATE(2291),
    [aux_sym_type_argument_repeat1] = STATE(2308),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6415),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2292)] = {
    [sym_block_comment] = STATE(2292),
    [sym_compiler_directive_decl] = STATE(2292),
    [sym_fsi_directive_decl] = STATE(2292),
    [sym_preproc_line] = STATE(2292),
    [ts_builtin_sym_end] = ACTIONS(6418),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_namespace] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2293)] = {
    [sym_block_comment] = STATE(2293),
    [sym_compiler_directive_decl] = STATE(2293),
    [sym_fsi_directive_decl] = STATE(2293),
    [sym_preproc_line] = STATE(2293),
    [aux_sym_compound_type_repeat1] = STATE(2296),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6281),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2294)] = {
    [sym_block_comment] = STATE(2294),
    [sym_compiler_directive_decl] = STATE(2294),
    [sym_fsi_directive_decl] = STATE(2294),
    [sym_preproc_line] = STATE(2294),
    [ts_builtin_sym_end] = ACTIONS(6375),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_namespace] = ACTIONS(6373),
    [anon_sym_module] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_open] = ACTIONS(6373),
    [anon_sym_type] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_and] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(6422),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_exception] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [anon_sym_extern] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6375),
    [anon_sym_POUNDload] = ACTIONS(6375),
    [anon_sym_POUNDtime] = ACTIONS(6375),
    [anon_sym_POUNDI] = ACTIONS(6375),
    [anon_sym_POUNDhelp] = ACTIONS(6375),
    [anon_sym_POUNDquit] = ACTIONS(6375),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2295)] = {
    [sym_block_comment] = STATE(2295),
    [sym_compiler_directive_decl] = STATE(2295),
    [sym_fsi_directive_decl] = STATE(2295),
    [sym_preproc_line] = STATE(2295),
    [ts_builtin_sym_end] = ACTIONS(6409),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_namespace] = ACTIONS(6407),
    [anon_sym_module] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_open] = ACTIONS(6407),
    [anon_sym_type] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_and] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_exception] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [anon_sym_extern] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6409),
    [anon_sym_POUNDload] = ACTIONS(6409),
    [anon_sym_POUNDtime] = ACTIONS(6409),
    [anon_sym_POUNDI] = ACTIONS(6409),
    [anon_sym_POUNDhelp] = ACTIONS(6409),
    [anon_sym_POUNDquit] = ACTIONS(6409),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2296)] = {
    [sym_block_comment] = STATE(2296),
    [sym_compiler_directive_decl] = STATE(2296),
    [sym_fsi_directive_decl] = STATE(2296),
    [sym_preproc_line] = STATE(2296),
    [aux_sym_compound_type_repeat1] = STATE(2296),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6424),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2297)] = {
    [sym_block_comment] = STATE(2297),
    [sym_compiler_directive_decl] = STATE(2297),
    [sym_fsi_directive_decl] = STATE(2297),
    [sym_preproc_line] = STATE(2297),
    [ts_builtin_sym_end] = ACTIONS(6381),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_namespace] = ACTIONS(6379),
    [anon_sym_module] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_open] = ACTIONS(6379),
    [anon_sym_type] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_and] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_exception] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [anon_sym_extern] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6381),
    [anon_sym_POUNDload] = ACTIONS(6381),
    [anon_sym_POUNDtime] = ACTIONS(6381),
    [anon_sym_POUNDI] = ACTIONS(6381),
    [anon_sym_POUNDhelp] = ACTIONS(6381),
    [anon_sym_POUNDquit] = ACTIONS(6381),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2298)] = {
    [sym_block_comment] = STATE(2298),
    [sym_compiler_directive_decl] = STATE(2298),
    [sym_fsi_directive_decl] = STATE(2298),
    [sym_preproc_line] = STATE(2298),
    [ts_builtin_sym_end] = ACTIONS(6385),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_namespace] = ACTIONS(6383),
    [anon_sym_module] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_open] = ACTIONS(6383),
    [anon_sym_type] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_and] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_exception] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [anon_sym_extern] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6385),
    [anon_sym_POUNDload] = ACTIONS(6385),
    [anon_sym_POUNDtime] = ACTIONS(6385),
    [anon_sym_POUNDI] = ACTIONS(6385),
    [anon_sym_POUNDhelp] = ACTIONS(6385),
    [anon_sym_POUNDquit] = ACTIONS(6385),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2299)] = {
    [sym_block_comment] = STATE(2299),
    [sym_compiler_directive_decl] = STATE(2299),
    [sym_fsi_directive_decl] = STATE(2299),
    [sym_preproc_line] = STATE(2299),
    [ts_builtin_sym_end] = ACTIONS(6310),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_namespace] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2300)] = {
    [sym_block_comment] = STATE(2300),
    [sym_compiler_directive_decl] = STATE(2300),
    [sym_fsi_directive_decl] = STATE(2300),
    [sym_preproc_line] = STATE(2300),
    [ts_builtin_sym_end] = ACTIONS(6401),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_namespace] = ACTIONS(6399),
    [anon_sym_module] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_open] = ACTIONS(6399),
    [anon_sym_type] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_and] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_exception] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [anon_sym_extern] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6401),
    [anon_sym_POUNDload] = ACTIONS(6401),
    [anon_sym_POUNDtime] = ACTIONS(6401),
    [anon_sym_POUNDI] = ACTIONS(6401),
    [anon_sym_POUNDhelp] = ACTIONS(6401),
    [anon_sym_POUNDquit] = ACTIONS(6401),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2301)] = {
    [sym_block_comment] = STATE(2301),
    [sym_compiler_directive_decl] = STATE(2301),
    [sym_fsi_directive_decl] = STATE(2301),
    [sym_preproc_line] = STATE(2301),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_module] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_open] = ACTIONS(6427),
    [anon_sym_type] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_and] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_exception] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [anon_sym_extern] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6429),
    [anon_sym_POUNDload] = ACTIONS(6429),
    [anon_sym_POUNDtime] = ACTIONS(6429),
    [anon_sym_POUNDI] = ACTIONS(6429),
    [anon_sym_POUNDhelp] = ACTIONS(6429),
    [anon_sym_POUNDquit] = ACTIONS(6429),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [anon_sym_POUNDelse] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2302)] = {
    [sym_block_comment] = STATE(2302),
    [sym_compiler_directive_decl] = STATE(2302),
    [sym_fsi_directive_decl] = STATE(2302),
    [sym_preproc_line] = STATE(2302),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_module] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_open] = ACTIONS(6389),
    [anon_sym_type] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_and] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_exception] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [anon_sym_extern] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6387),
    [anon_sym_POUNDload] = ACTIONS(6387),
    [anon_sym_POUNDtime] = ACTIONS(6387),
    [anon_sym_POUNDI] = ACTIONS(6387),
    [anon_sym_POUNDhelp] = ACTIONS(6387),
    [anon_sym_POUNDquit] = ACTIONS(6387),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [anon_sym_POUNDelse] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2303)] = {
    [sym_block_comment] = STATE(2303),
    [sym_compiler_directive_decl] = STATE(2303),
    [sym_fsi_directive_decl] = STATE(2303),
    [sym_preproc_line] = STATE(2303),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2304)] = {
    [sym_block_comment] = STATE(2304),
    [sym_compiler_directive_decl] = STATE(2304),
    [sym_fsi_directive_decl] = STATE(2304),
    [sym_preproc_line] = STATE(2304),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2305)] = {
    [sym_block_comment] = STATE(2305),
    [sym_compiler_directive_decl] = STATE(2305),
    [sym_fsi_directive_decl] = STATE(2305),
    [sym_preproc_line] = STATE(2305),
    [ts_builtin_sym_end] = ACTIONS(6413),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_namespace] = ACTIONS(6411),
    [anon_sym_module] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_open] = ACTIONS(6411),
    [anon_sym_type] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_and] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_exception] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [anon_sym_extern] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6413),
    [anon_sym_POUNDload] = ACTIONS(6413),
    [anon_sym_POUNDtime] = ACTIONS(6413),
    [anon_sym_POUNDI] = ACTIONS(6413),
    [anon_sym_POUNDhelp] = ACTIONS(6413),
    [anon_sym_POUNDquit] = ACTIONS(6413),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2306)] = {
    [sym_block_comment] = STATE(2306),
    [sym_compiler_directive_decl] = STATE(2306),
    [sym_fsi_directive_decl] = STATE(2306),
    [sym_preproc_line] = STATE(2306),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2307)] = {
    [sym_block_comment] = STATE(2307),
    [sym_compiler_directive_decl] = STATE(2307),
    [sym_fsi_directive_decl] = STATE(2307),
    [sym_preproc_line] = STATE(2307),
    [ts_builtin_sym_end] = ACTIONS(6405),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_namespace] = ACTIONS(6403),
    [anon_sym_module] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_open] = ACTIONS(6403),
    [anon_sym_type] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_and] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_exception] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [anon_sym_extern] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6405),
    [anon_sym_POUNDload] = ACTIONS(6405),
    [anon_sym_POUNDtime] = ACTIONS(6405),
    [anon_sym_POUNDI] = ACTIONS(6405),
    [anon_sym_POUNDhelp] = ACTIONS(6405),
    [anon_sym_POUNDquit] = ACTIONS(6405),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2308)] = {
    [sym_block_comment] = STATE(2308),
    [sym_compiler_directive_decl] = STATE(2308),
    [sym_fsi_directive_decl] = STATE(2308),
    [sym_preproc_line] = STATE(2308),
    [aux_sym_type_argument_repeat1] = STATE(2309),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2309)] = {
    [sym_block_comment] = STATE(2309),
    [sym_compiler_directive_decl] = STATE(2309),
    [sym_fsi_directive_decl] = STATE(2309),
    [sym_preproc_line] = STATE(2309),
    [aux_sym_type_argument_repeat1] = STATE(2309),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6431),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2310)] = {
    [sym_block_comment] = STATE(2310),
    [sym_compiler_directive_decl] = STATE(2310),
    [sym_fsi_directive_decl] = STATE(2310),
    [sym_preproc_line] = STATE(2310),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6434),
    [anon_sym_uy] = ACTIONS(6436),
    [anon_sym_s] = ACTIONS(6438),
    [anon_sym_us] = ACTIONS(6440),
    [anon_sym_l] = ACTIONS(6442),
    [aux_sym_uint32_token1] = ACTIONS(6444),
    [anon_sym_n] = ACTIONS(6446),
    [anon_sym_un] = ACTIONS(6448),
    [anon_sym_L] = ACTIONS(6450),
    [aux_sym_uint64_token1] = ACTIONS(6452),
    [aux_sym_bignum_token1] = ACTIONS(6454),
    [aux_sym_decimal_token1] = ACTIONS(6456),
    [anon_sym_DOT2] = ACTIONS(6458),
    [aux_sym_float_token1] = ACTIONS(6460),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2311)] = {
    [sym_block_comment] = STATE(2311),
    [sym_compiler_directive_decl] = STATE(2311),
    [sym_fsi_directive_decl] = STATE(2311),
    [sym_preproc_line] = STATE(2311),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6462),
    [anon_sym_uy] = ACTIONS(6464),
    [anon_sym_s] = ACTIONS(6466),
    [anon_sym_us] = ACTIONS(6468),
    [anon_sym_l] = ACTIONS(6470),
    [aux_sym_uint32_token1] = ACTIONS(6472),
    [anon_sym_n] = ACTIONS(6474),
    [anon_sym_un] = ACTIONS(6476),
    [anon_sym_L] = ACTIONS(6478),
    [aux_sym_uint64_token1] = ACTIONS(6480),
    [aux_sym_bignum_token1] = ACTIONS(6482),
    [aux_sym_decimal_token1] = ACTIONS(6484),
    [anon_sym_DOT2] = ACTIONS(6486),
    [aux_sym_float_token1] = ACTIONS(6488),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2312)] = {
    [sym_block_comment] = STATE(2312),
    [sym_compiler_directive_decl] = STATE(2312),
    [sym_fsi_directive_decl] = STATE(2312),
    [sym_preproc_line] = STATE(2312),
    [ts_builtin_sym_end] = ACTIONS(6490),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_namespace] = ACTIONS(6492),
    [anon_sym_module] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_open] = ACTIONS(6492),
    [anon_sym_type] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_and] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_exception] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [anon_sym_extern] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6490),
    [anon_sym_POUNDload] = ACTIONS(6490),
    [anon_sym_POUNDtime] = ACTIONS(6490),
    [anon_sym_POUNDI] = ACTIONS(6490),
    [anon_sym_POUNDhelp] = ACTIONS(6490),
    [anon_sym_POUNDquit] = ACTIONS(6490),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2313)] = {
    [sym_block_comment] = STATE(2313),
    [sym_compiler_directive_decl] = STATE(2313),
    [sym_fsi_directive_decl] = STATE(2313),
    [sym_preproc_line] = STATE(2313),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2314)] = {
    [sym_elif_expression] = STATE(2354),
    [sym_block_comment] = STATE(2314),
    [sym_compiler_directive_decl] = STATE(2314),
    [sym_fsi_directive_decl] = STATE(2314),
    [sym_preproc_line] = STATE(2314),
    [aux_sym_if_expression_repeat1] = STATE(2314),
    [ts_builtin_sym_end] = ACTIONS(6494),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_namespace] = ACTIONS(6496),
    [anon_sym_module] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_open] = ACTIONS(6496),
    [anon_sym_type] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_and] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(6498),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_exception] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [anon_sym_extern] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6494),
    [anon_sym_POUNDload] = ACTIONS(6494),
    [anon_sym_POUNDtime] = ACTIONS(6494),
    [anon_sym_POUNDI] = ACTIONS(6494),
    [anon_sym_POUNDhelp] = ACTIONS(6494),
    [anon_sym_POUNDquit] = ACTIONS(6494),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2315)] = {
    [sym_block_comment] = STATE(2315),
    [sym_compiler_directive_decl] = STATE(2315),
    [sym_fsi_directive_decl] = STATE(2315),
    [sym_preproc_line] = STATE(2315),
    [aux_sym_long_identifier_repeat1] = STATE(2321),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(6501),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2316)] = {
    [sym_block_comment] = STATE(2316),
    [sym_compiler_directive_decl] = STATE(2316),
    [sym_fsi_directive_decl] = STATE(2316),
    [sym_preproc_line] = STATE(2316),
    [ts_builtin_sym_end] = ACTIONS(6371),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_namespace] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2317)] = {
    [sym_block_comment] = STATE(2317),
    [sym_compiler_directive_decl] = STATE(2317),
    [sym_fsi_directive_decl] = STATE(2317),
    [sym_preproc_line] = STATE(2317),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2318)] = {
    [sym_block_comment] = STATE(2318),
    [sym_compiler_directive_decl] = STATE(2318),
    [sym_fsi_directive_decl] = STATE(2318),
    [sym_preproc_line] = STATE(2318),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2319)] = {
    [sym_block_comment] = STATE(2319),
    [sym_compiler_directive_decl] = STATE(2319),
    [sym_fsi_directive_decl] = STATE(2319),
    [sym_preproc_line] = STATE(2319),
    [ts_builtin_sym_end] = ACTIONS(6429),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_namespace] = ACTIONS(6427),
    [anon_sym_module] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_open] = ACTIONS(6427),
    [anon_sym_type] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_and] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_exception] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [anon_sym_extern] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6429),
    [anon_sym_POUNDload] = ACTIONS(6429),
    [anon_sym_POUNDtime] = ACTIONS(6429),
    [anon_sym_POUNDI] = ACTIONS(6429),
    [anon_sym_POUNDhelp] = ACTIONS(6429),
    [anon_sym_POUNDquit] = ACTIONS(6429),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2320)] = {
    [sym_elif_expression] = STATE(2398),
    [sym_block_comment] = STATE(2320),
    [sym_compiler_directive_decl] = STATE(2320),
    [sym_fsi_directive_decl] = STATE(2320),
    [sym_preproc_line] = STATE(2320),
    [aux_sym_if_expression_repeat1] = STATE(2320),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_module] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_open] = ACTIONS(6496),
    [anon_sym_type] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_and] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(6503),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_exception] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [anon_sym_extern] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6494),
    [anon_sym_POUNDload] = ACTIONS(6494),
    [anon_sym_POUNDtime] = ACTIONS(6494),
    [anon_sym_POUNDI] = ACTIONS(6494),
    [anon_sym_POUNDhelp] = ACTIONS(6494),
    [anon_sym_POUNDquit] = ACTIONS(6494),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [anon_sym_POUNDelse] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2321)] = {
    [sym_block_comment] = STATE(2321),
    [sym_compiler_directive_decl] = STATE(2321),
    [sym_fsi_directive_decl] = STATE(2321),
    [sym_preproc_line] = STATE(2321),
    [aux_sym_long_identifier_repeat1] = STATE(2322),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(6501),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2322)] = {
    [sym_block_comment] = STATE(2322),
    [sym_compiler_directive_decl] = STATE(2322),
    [sym_fsi_directive_decl] = STATE(2322),
    [sym_preproc_line] = STATE(2322),
    [aux_sym_long_identifier_repeat1] = STATE(2322),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6506),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2323)] = {
    [sym__else_expression] = STATE(2777),
    [sym_elif_expression] = STATE(2447),
    [sym_block_comment] = STATE(2323),
    [sym_compiler_directive_decl] = STATE(2323),
    [sym_fsi_directive_decl] = STATE(2323),
    [sym_preproc_line] = STATE(2323),
    [aux_sym_if_expression_repeat1] = STATE(2324),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_module] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_open] = ACTIONS(6319),
    [anon_sym_type] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_and] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(6509),
    [anon_sym_elif] = ACTIONS(6511),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_exception] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [anon_sym_extern] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6317),
    [anon_sym_POUNDload] = ACTIONS(6317),
    [anon_sym_POUNDtime] = ACTIONS(6317),
    [anon_sym_POUNDI] = ACTIONS(6317),
    [anon_sym_POUNDhelp] = ACTIONS(6317),
    [anon_sym_POUNDquit] = ACTIONS(6317),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2324)] = {
    [sym__else_expression] = STATE(2787),
    [sym_elif_expression] = STATE(2447),
    [sym_block_comment] = STATE(2324),
    [sym_compiler_directive_decl] = STATE(2324),
    [sym_fsi_directive_decl] = STATE(2324),
    [sym_preproc_line] = STATE(2324),
    [aux_sym_if_expression_repeat1] = STATE(2349),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_module] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_open] = ACTIONS(6329),
    [anon_sym_type] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_and] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(6509),
    [anon_sym_elif] = ACTIONS(6511),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_exception] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [anon_sym_extern] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6327),
    [anon_sym_POUNDload] = ACTIONS(6327),
    [anon_sym_POUNDtime] = ACTIONS(6327),
    [anon_sym_POUNDI] = ACTIONS(6327),
    [anon_sym_POUNDhelp] = ACTIONS(6327),
    [anon_sym_POUNDquit] = ACTIONS(6327),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2325)] = {
    [sym_block_comment] = STATE(2325),
    [sym_compiler_directive_decl] = STATE(2325),
    [sym_fsi_directive_decl] = STATE(2325),
    [sym_preproc_line] = STATE(2325),
    [ts_builtin_sym_end] = ACTIONS(6393),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_namespace] = ACTIONS(6391),
    [anon_sym_module] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_open] = ACTIONS(6391),
    [anon_sym_type] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_and] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_exception] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [anon_sym_extern] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6393),
    [anon_sym_POUNDload] = ACTIONS(6393),
    [anon_sym_POUNDtime] = ACTIONS(6393),
    [anon_sym_POUNDI] = ACTIONS(6393),
    [anon_sym_POUNDhelp] = ACTIONS(6393),
    [anon_sym_POUNDquit] = ACTIONS(6393),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2326)] = {
    [sym_block_comment] = STATE(2326),
    [sym_compiler_directive_decl] = STATE(2326),
    [sym_fsi_directive_decl] = STATE(2326),
    [sym_preproc_line] = STATE(2326),
    [ts_builtin_sym_end] = ACTIONS(6397),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_namespace] = ACTIONS(6395),
    [anon_sym_module] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_open] = ACTIONS(6395),
    [anon_sym_type] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_and] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_exception] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [anon_sym_extern] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6397),
    [anon_sym_POUNDload] = ACTIONS(6397),
    [anon_sym_POUNDtime] = ACTIONS(6397),
    [anon_sym_POUNDI] = ACTIONS(6397),
    [anon_sym_POUNDhelp] = ACTIONS(6397),
    [anon_sym_POUNDquit] = ACTIONS(6397),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2327)] = {
    [sym_block_comment] = STATE(2327),
    [sym_compiler_directive_decl] = STATE(2327),
    [sym_fsi_directive_decl] = STATE(2327),
    [sym_preproc_line] = STATE(2327),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_module] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_open] = ACTIONS(6492),
    [anon_sym_type] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_and] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_exception] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [anon_sym_extern] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6490),
    [anon_sym_POUNDload] = ACTIONS(6490),
    [anon_sym_POUNDtime] = ACTIONS(6490),
    [anon_sym_POUNDI] = ACTIONS(6490),
    [anon_sym_POUNDhelp] = ACTIONS(6490),
    [anon_sym_POUNDquit] = ACTIONS(6490),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [anon_sym_POUNDelse] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2328)] = {
    [sym_block_comment] = STATE(2328),
    [sym_compiler_directive_decl] = STATE(2328),
    [sym_fsi_directive_decl] = STATE(2328),
    [sym_preproc_line] = STATE(2328),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(6517),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2329)] = {
    [sym_block_comment] = STATE(2329),
    [sym_compiler_directive_decl] = STATE(2329),
    [sym_fsi_directive_decl] = STATE(2329),
    [sym_preproc_line] = STATE(2329),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_module] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_open] = ACTIONS(6403),
    [anon_sym_type] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_and] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_exception] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [anon_sym_extern] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6405),
    [anon_sym_POUNDload] = ACTIONS(6405),
    [anon_sym_POUNDtime] = ACTIONS(6405),
    [anon_sym_POUNDI] = ACTIONS(6405),
    [anon_sym_POUNDhelp] = ACTIONS(6405),
    [anon_sym_POUNDquit] = ACTIONS(6405),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2330)] = {
    [sym_block_comment] = STATE(2330),
    [sym_compiler_directive_decl] = STATE(2330),
    [sym_fsi_directive_decl] = STATE(2330),
    [sym_preproc_line] = STATE(2330),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2331)] = {
    [sym_block_comment] = STATE(2331),
    [sym_compiler_directive_decl] = STATE(2331),
    [sym_fsi_directive_decl] = STATE(2331),
    [sym_preproc_line] = STATE(2331),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_module] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_open] = ACTIONS(6407),
    [anon_sym_type] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_and] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_exception] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [anon_sym_extern] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6409),
    [anon_sym_POUNDload] = ACTIONS(6409),
    [anon_sym_POUNDtime] = ACTIONS(6409),
    [anon_sym_POUNDI] = ACTIONS(6409),
    [anon_sym_POUNDhelp] = ACTIONS(6409),
    [anon_sym_POUNDquit] = ACTIONS(6409),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2332)] = {
    [sym_block_comment] = STATE(2332),
    [sym_compiler_directive_decl] = STATE(2332),
    [sym_fsi_directive_decl] = STATE(2332),
    [sym_preproc_line] = STATE(2332),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_module] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_open] = ACTIONS(6411),
    [anon_sym_type] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_and] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_exception] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [anon_sym_extern] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6413),
    [anon_sym_POUNDload] = ACTIONS(6413),
    [anon_sym_POUNDtime] = ACTIONS(6413),
    [anon_sym_POUNDI] = ACTIONS(6413),
    [anon_sym_POUNDhelp] = ACTIONS(6413),
    [anon_sym_POUNDquit] = ACTIONS(6413),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2333)] = {
    [sym_block_comment] = STATE(2333),
    [sym_compiler_directive_decl] = STATE(2333),
    [sym_fsi_directive_decl] = STATE(2333),
    [sym_preproc_line] = STATE(2333),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2334)] = {
    [sym_block_comment] = STATE(2334),
    [sym_compiler_directive_decl] = STATE(2334),
    [sym_fsi_directive_decl] = STATE(2334),
    [sym_preproc_line] = STATE(2334),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_module] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_open] = ACTIONS(6379),
    [anon_sym_type] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_and] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_exception] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [anon_sym_extern] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6381),
    [anon_sym_POUNDload] = ACTIONS(6381),
    [anon_sym_POUNDtime] = ACTIONS(6381),
    [anon_sym_POUNDI] = ACTIONS(6381),
    [anon_sym_POUNDhelp] = ACTIONS(6381),
    [anon_sym_POUNDquit] = ACTIONS(6381),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2335)] = {
    [sym_block_comment] = STATE(2335),
    [sym_compiler_directive_decl] = STATE(2335),
    [sym_fsi_directive_decl] = STATE(2335),
    [sym_preproc_line] = STATE(2335),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_module] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_open] = ACTIONS(6383),
    [anon_sym_type] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_and] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_exception] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [anon_sym_extern] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6385),
    [anon_sym_POUNDload] = ACTIONS(6385),
    [anon_sym_POUNDtime] = ACTIONS(6385),
    [anon_sym_POUNDI] = ACTIONS(6385),
    [anon_sym_POUNDhelp] = ACTIONS(6385),
    [anon_sym_POUNDquit] = ACTIONS(6385),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2336)] = {
    [sym_block_comment] = STATE(2336),
    [sym_compiler_directive_decl] = STATE(2336),
    [sym_fsi_directive_decl] = STATE(2336),
    [sym_preproc_line] = STATE(2336),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6519),
    [anon_sym_uy] = ACTIONS(6521),
    [anon_sym_s] = ACTIONS(6523),
    [anon_sym_us] = ACTIONS(6525),
    [anon_sym_l] = ACTIONS(6527),
    [aux_sym_uint32_token1] = ACTIONS(6529),
    [anon_sym_n] = ACTIONS(6531),
    [anon_sym_un] = ACTIONS(6533),
    [anon_sym_L] = ACTIONS(6535),
    [aux_sym_uint64_token1] = ACTIONS(6537),
    [aux_sym_bignum_token1] = ACTIONS(6539),
    [aux_sym_decimal_token1] = ACTIONS(6541),
    [anon_sym_DOT2] = ACTIONS(6543),
    [aux_sym_float_token1] = ACTIONS(6545),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2337)] = {
    [sym_block_comment] = STATE(2337),
    [sym_compiler_directive_decl] = STATE(2337),
    [sym_fsi_directive_decl] = STATE(2337),
    [sym_preproc_line] = STATE(2337),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_module] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_open] = ACTIONS(6391),
    [anon_sym_type] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_and] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_exception] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [anon_sym_extern] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6393),
    [anon_sym_POUNDload] = ACTIONS(6393),
    [anon_sym_POUNDtime] = ACTIONS(6393),
    [anon_sym_POUNDI] = ACTIONS(6393),
    [anon_sym_POUNDhelp] = ACTIONS(6393),
    [anon_sym_POUNDquit] = ACTIONS(6393),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2338)] = {
    [sym_block_comment] = STATE(2338),
    [sym_compiler_directive_decl] = STATE(2338),
    [sym_fsi_directive_decl] = STATE(2338),
    [sym_preproc_line] = STATE(2338),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_module] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_open] = ACTIONS(6395),
    [anon_sym_type] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_and] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_exception] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [anon_sym_extern] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6397),
    [anon_sym_POUNDload] = ACTIONS(6397),
    [anon_sym_POUNDtime] = ACTIONS(6397),
    [anon_sym_POUNDI] = ACTIONS(6397),
    [anon_sym_POUNDhelp] = ACTIONS(6397),
    [anon_sym_POUNDquit] = ACTIONS(6397),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2339)] = {
    [sym_block_comment] = STATE(2339),
    [sym_compiler_directive_decl] = STATE(2339),
    [sym_fsi_directive_decl] = STATE(2339),
    [sym_preproc_line] = STATE(2339),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_module] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_open] = ACTIONS(6492),
    [anon_sym_type] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_and] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_exception] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [anon_sym_extern] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6490),
    [anon_sym_POUNDload] = ACTIONS(6490),
    [anon_sym_POUNDtime] = ACTIONS(6490),
    [anon_sym_POUNDI] = ACTIONS(6490),
    [anon_sym_POUNDhelp] = ACTIONS(6490),
    [anon_sym_POUNDquit] = ACTIONS(6490),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2340)] = {
    [sym_block_comment] = STATE(2340),
    [sym_compiler_directive_decl] = STATE(2340),
    [sym_fsi_directive_decl] = STATE(2340),
    [sym_preproc_line] = STATE(2340),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_module] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_open] = ACTIONS(6427),
    [anon_sym_type] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_and] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_exception] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [anon_sym_extern] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6429),
    [anon_sym_POUNDload] = ACTIONS(6429),
    [anon_sym_POUNDtime] = ACTIONS(6429),
    [anon_sym_POUNDI] = ACTIONS(6429),
    [anon_sym_POUNDhelp] = ACTIONS(6429),
    [anon_sym_POUNDquit] = ACTIONS(6429),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2341)] = {
    [sym_block_comment] = STATE(2341),
    [sym_compiler_directive_decl] = STATE(2341),
    [sym_fsi_directive_decl] = STATE(2341),
    [sym_preproc_line] = STATE(2341),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_module] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_open] = ACTIONS(6389),
    [anon_sym_type] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_and] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_exception] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [anon_sym_extern] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6387),
    [anon_sym_POUNDload] = ACTIONS(6387),
    [anon_sym_POUNDtime] = ACTIONS(6387),
    [anon_sym_POUNDI] = ACTIONS(6387),
    [anon_sym_POUNDhelp] = ACTIONS(6387),
    [anon_sym_POUNDquit] = ACTIONS(6387),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2342)] = {
    [sym_block_comment] = STATE(2342),
    [sym_compiler_directive_decl] = STATE(2342),
    [sym_fsi_directive_decl] = STATE(2342),
    [sym_preproc_line] = STATE(2342),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2343)] = {
    [sym_block_comment] = STATE(2343),
    [sym_compiler_directive_decl] = STATE(2343),
    [sym_fsi_directive_decl] = STATE(2343),
    [sym_preproc_line] = STATE(2343),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_module] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_open] = ACTIONS(6373),
    [anon_sym_type] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_and] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(6547),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_exception] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [anon_sym_extern] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6375),
    [anon_sym_POUNDload] = ACTIONS(6375),
    [anon_sym_POUNDtime] = ACTIONS(6375),
    [anon_sym_POUNDI] = ACTIONS(6375),
    [anon_sym_POUNDhelp] = ACTIONS(6375),
    [anon_sym_POUNDquit] = ACTIONS(6375),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2344)] = {
    [sym_block_comment] = STATE(2344),
    [sym_compiler_directive_decl] = STATE(2344),
    [sym_fsi_directive_decl] = STATE(2344),
    [sym_preproc_line] = STATE(2344),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2345)] = {
    [sym_block_comment] = STATE(2345),
    [sym_compiler_directive_decl] = STATE(2345),
    [sym_fsi_directive_decl] = STATE(2345),
    [sym_preproc_line] = STATE(2345),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_module] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_open] = ACTIONS(6367),
    [anon_sym_type] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_and] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_exception] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [anon_sym_extern] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6365),
    [anon_sym_POUNDload] = ACTIONS(6365),
    [anon_sym_POUNDtime] = ACTIONS(6365),
    [anon_sym_POUNDI] = ACTIONS(6365),
    [anon_sym_POUNDhelp] = ACTIONS(6365),
    [anon_sym_POUNDquit] = ACTIONS(6365),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2346)] = {
    [sym_block_comment] = STATE(2346),
    [sym_compiler_directive_decl] = STATE(2346),
    [sym_fsi_directive_decl] = STATE(2346),
    [sym_preproc_line] = STATE(2346),
    [ts_builtin_sym_end] = ACTIONS(6515),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_namespace] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(6549),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2347)] = {
    [sym_block_comment] = STATE(2347),
    [sym_compiler_directive_decl] = STATE(2347),
    [sym_fsi_directive_decl] = STATE(2347),
    [sym_preproc_line] = STATE(2347),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_module] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_open] = ACTIONS(6399),
    [anon_sym_type] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_and] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_exception] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [anon_sym_extern] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6401),
    [anon_sym_POUNDload] = ACTIONS(6401),
    [anon_sym_POUNDtime] = ACTIONS(6401),
    [anon_sym_POUNDI] = ACTIONS(6401),
    [anon_sym_POUNDhelp] = ACTIONS(6401),
    [anon_sym_POUNDquit] = ACTIONS(6401),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2348)] = {
    [sym_block_comment] = STATE(2348),
    [sym_compiler_directive_decl] = STATE(2348),
    [sym_fsi_directive_decl] = STATE(2348),
    [sym_preproc_line] = STATE(2348),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2349)] = {
    [sym_elif_expression] = STATE(2447),
    [sym_block_comment] = STATE(2349),
    [sym_compiler_directive_decl] = STATE(2349),
    [sym_fsi_directive_decl] = STATE(2349),
    [sym_preproc_line] = STATE(2349),
    [aux_sym_if_expression_repeat1] = STATE(2349),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_module] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_open] = ACTIONS(6496),
    [anon_sym_type] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_and] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(6551),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_exception] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [anon_sym_extern] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6494),
    [anon_sym_POUNDload] = ACTIONS(6494),
    [anon_sym_POUNDtime] = ACTIONS(6494),
    [anon_sym_POUNDI] = ACTIONS(6494),
    [anon_sym_POUNDhelp] = ACTIONS(6494),
    [anon_sym_POUNDquit] = ACTIONS(6494),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2350)] = {
    [sym_block_comment] = STATE(2350),
    [sym_compiler_directive_decl] = STATE(2350),
    [sym_fsi_directive_decl] = STATE(2350),
    [sym_preproc_line] = STATE(2350),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2351)] = {
    [sym_block_comment] = STATE(2351),
    [sym_compiler_directive_decl] = STATE(2351),
    [sym_fsi_directive_decl] = STATE(2351),
    [sym_preproc_line] = STATE(2351),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6462),
    [anon_sym_uy] = ACTIONS(6464),
    [anon_sym_s] = ACTIONS(6466),
    [anon_sym_us] = ACTIONS(6468),
    [anon_sym_l] = ACTIONS(6470),
    [aux_sym_uint32_token1] = ACTIONS(6472),
    [anon_sym_n] = ACTIONS(6474),
    [anon_sym_un] = ACTIONS(6476),
    [anon_sym_L] = ACTIONS(6478),
    [aux_sym_uint64_token1] = ACTIONS(6480),
    [aux_sym_bignum_token1] = ACTIONS(6482),
    [aux_sym_decimal_token1] = ACTIONS(6484),
    [anon_sym_DOT2] = ACTIONS(6486),
    [aux_sym_float_token1] = ACTIONS(6488),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2352)] = {
    [sym_block_comment] = STATE(2352),
    [sym_compiler_directive_decl] = STATE(2352),
    [sym_fsi_directive_decl] = STATE(2352),
    [sym_preproc_line] = STATE(2352),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_module] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_open] = ACTIONS(6554),
    [anon_sym_type] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_and] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_exception] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [anon_sym_extern] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6556),
    [anon_sym_POUNDload] = ACTIONS(6556),
    [anon_sym_POUNDtime] = ACTIONS(6556),
    [anon_sym_POUNDI] = ACTIONS(6556),
    [anon_sym_POUNDhelp] = ACTIONS(6556),
    [anon_sym_POUNDquit] = ACTIONS(6556),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [anon_sym_POUNDelse] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2353)] = {
    [sym_block_comment] = STATE(2353),
    [sym_compiler_directive_decl] = STATE(2353),
    [sym_fsi_directive_decl] = STATE(2353),
    [sym_preproc_line] = STATE(2353),
    [ts_builtin_sym_end] = ACTIONS(6558),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_namespace] = ACTIONS(6560),
    [anon_sym_module] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_open] = ACTIONS(6560),
    [anon_sym_type] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_and] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_exception] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [anon_sym_extern] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6558),
    [anon_sym_POUNDload] = ACTIONS(6558),
    [anon_sym_POUNDtime] = ACTIONS(6558),
    [anon_sym_POUNDI] = ACTIONS(6558),
    [anon_sym_POUNDhelp] = ACTIONS(6558),
    [anon_sym_POUNDquit] = ACTIONS(6558),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(2354)] = {
    [sym_block_comment] = STATE(2354),
    [sym_compiler_directive_decl] = STATE(2354),
    [sym_fsi_directive_decl] = STATE(2354),
    [sym_preproc_line] = STATE(2354),
    [ts_builtin_sym_end] = ACTIONS(6562),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_namespace] = ACTIONS(6564),
    [anon_sym_module] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_open] = ACTIONS(6564),
    [anon_sym_type] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_and] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_exception] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [anon_sym_extern] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6562),
    [anon_sym_POUNDload] = ACTIONS(6562),
    [anon_sym_POUNDtime] = ACTIONS(6562),
    [anon_sym_POUNDI] = ACTIONS(6562),
    [anon_sym_POUNDhelp] = ACTIONS(6562),
    [anon_sym_POUNDquit] = ACTIONS(6562),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2355)] = {
    [sym_type_arguments] = STATE(2766),
    [sym_long_identifier] = STATE(2775),
    [sym_block_comment] = STATE(2355),
    [sym_compiler_directive_decl] = STATE(2355),
    [sym_fsi_directive_decl] = STATE(2355),
    [sym_preproc_line] = STATE(2355),
    [aux_sym_compound_type_repeat1] = STATE(2505),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_and] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6566),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6568),
    [anon_sym_LT] = ACTIONS(6570),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6572),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [anon_sym_POUNDelse] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2356)] = {
    [sym_type_arguments] = STATE(2732),
    [sym_long_identifier] = STATE(2698),
    [sym_block_comment] = STATE(2356),
    [sym_compiler_directive_decl] = STATE(2356),
    [sym_fsi_directive_decl] = STATE(2356),
    [sym_preproc_line] = STATE(2356),
    [aux_sym_compound_type_repeat1] = STATE(2486),
    [sym_identifier] = ACTIONS(6574),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6576),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6578),
    [anon_sym_LT] = ACTIONS(6580),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6582),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_inherit] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2357)] = {
    [sym_block_comment] = STATE(2357),
    [sym_compiler_directive_decl] = STATE(2357),
    [sym_fsi_directive_decl] = STATE(2357),
    [sym_preproc_line] = STATE(2357),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(6584),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2358)] = {
    [sym_type_arguments] = STATE(2732),
    [sym_long_identifier] = STATE(2698),
    [sym_block_comment] = STATE(2358),
    [sym_compiler_directive_decl] = STATE(2358),
    [sym_fsi_directive_decl] = STATE(2358),
    [sym_preproc_line] = STATE(2358),
    [aux_sym_compound_type_repeat1] = STATE(2486),
    [sym_identifier] = ACTIONS(6574),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6576),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6578),
    [anon_sym_LT] = ACTIONS(6580),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6582),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_inherit] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2359)] = {
    [sym_type_arguments] = STATE(2732),
    [sym_long_identifier] = STATE(2698),
    [sym_block_comment] = STATE(2359),
    [sym_compiler_directive_decl] = STATE(2359),
    [sym_fsi_directive_decl] = STATE(2359),
    [sym_preproc_line] = STATE(2359),
    [aux_sym_compound_type_repeat1] = STATE(2486),
    [sym_identifier] = ACTIONS(6574),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6576),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6578),
    [anon_sym_LT] = ACTIONS(6580),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6582),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_inherit] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2360)] = {
    [sym_block_comment] = STATE(2360),
    [sym_compiler_directive_decl] = STATE(2360),
    [sym_fsi_directive_decl] = STATE(2360),
    [sym_preproc_line] = STATE(2360),
    [ts_builtin_sym_end] = ACTIONS(6556),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_namespace] = ACTIONS(6554),
    [anon_sym_module] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_open] = ACTIONS(6554),
    [anon_sym_type] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_and] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_exception] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [anon_sym_extern] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6556),
    [anon_sym_POUNDload] = ACTIONS(6556),
    [anon_sym_POUNDtime] = ACTIONS(6556),
    [anon_sym_POUNDI] = ACTIONS(6556),
    [anon_sym_POUNDhelp] = ACTIONS(6556),
    [anon_sym_POUNDquit] = ACTIONS(6556),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2361)] = {
    [sym_type_arguments] = STATE(2766),
    [sym_long_identifier] = STATE(2775),
    [sym_block_comment] = STATE(2361),
    [sym_compiler_directive_decl] = STATE(2361),
    [sym_fsi_directive_decl] = STATE(2361),
    [sym_preproc_line] = STATE(2361),
    [aux_sym_compound_type_repeat1] = STATE(2505),
    [sym_identifier] = ACTIONS(6586),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6566),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6568),
    [anon_sym_LT] = ACTIONS(6570),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6572),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2362)] = {
    [sym_type_arguments] = STATE(2766),
    [sym_long_identifier] = STATE(2775),
    [sym_block_comment] = STATE(2362),
    [sym_compiler_directive_decl] = STATE(2362),
    [sym_fsi_directive_decl] = STATE(2362),
    [sym_preproc_line] = STATE(2362),
    [aux_sym_compound_type_repeat1] = STATE(2505),
    [sym_identifier] = ACTIONS(6586),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6566),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6568),
    [anon_sym_LT] = ACTIONS(6570),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6572),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2363)] = {
    [sym_type_arguments] = STATE(2766),
    [sym_long_identifier] = STATE(2775),
    [sym_block_comment] = STATE(2363),
    [sym_compiler_directive_decl] = STATE(2363),
    [sym_fsi_directive_decl] = STATE(2363),
    [sym_preproc_line] = STATE(2363),
    [aux_sym_compound_type_repeat1] = STATE(2505),
    [sym_identifier] = ACTIONS(6586),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6566),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6568),
    [anon_sym_LT] = ACTIONS(6570),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6572),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2364)] = {
    [sym_block_comment] = STATE(2364),
    [sym_compiler_directive_decl] = STATE(2364),
    [sym_fsi_directive_decl] = STATE(2364),
    [sym_preproc_line] = STATE(2364),
    [aux_sym_rules_repeat1] = STATE(2368),
    [ts_builtin_sym_end] = ACTIONS(6588),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_namespace] = ACTIONS(6590),
    [anon_sym_module] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_open] = ACTIONS(6590),
    [anon_sym_type] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_and] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(6592),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_exception] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [anon_sym_extern] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6588),
    [anon_sym_POUNDload] = ACTIONS(6588),
    [anon_sym_POUNDtime] = ACTIONS(6588),
    [anon_sym_POUNDI] = ACTIONS(6588),
    [anon_sym_POUNDhelp] = ACTIONS(6588),
    [anon_sym_POUNDquit] = ACTIONS(6588),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(6594),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2365)] = {
    [sym_block_comment] = STATE(2365),
    [sym_compiler_directive_decl] = STATE(2365),
    [sym_fsi_directive_decl] = STATE(2365),
    [sym_preproc_line] = STATE(2365),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(6601),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2366)] = {
    [sym_block_comment] = STATE(2366),
    [sym_compiler_directive_decl] = STATE(2366),
    [sym_fsi_directive_decl] = STATE(2366),
    [sym_preproc_line] = STATE(2366),
    [aux_sym_rules_repeat1] = STATE(2370),
    [ts_builtin_sym_end] = ACTIONS(6605),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_namespace] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(6592),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(6609),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2367)] = {
    [sym_block_comment] = STATE(2367),
    [sym_compiler_directive_decl] = STATE(2367),
    [sym_fsi_directive_decl] = STATE(2367),
    [sym_preproc_line] = STATE(2367),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2368)] = {
    [sym_block_comment] = STATE(2368),
    [sym_compiler_directive_decl] = STATE(2368),
    [sym_fsi_directive_decl] = STATE(2368),
    [sym_preproc_line] = STATE(2368),
    [aux_sym_rules_repeat1] = STATE(2379),
    [ts_builtin_sym_end] = ACTIONS(6605),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_namespace] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(6592),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(6609),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2369)] = {
    [sym_block_comment] = STATE(2369),
    [sym_compiler_directive_decl] = STATE(2369),
    [sym_fsi_directive_decl] = STATE(2369),
    [sym_preproc_line] = STATE(2369),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6043),
    [aux_sym_decimal_token1] = ACTIONS(6005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2370)] = {
    [sym_block_comment] = STATE(2370),
    [sym_compiler_directive_decl] = STATE(2370),
    [sym_fsi_directive_decl] = STATE(2370),
    [sym_preproc_line] = STATE(2370),
    [aux_sym_rules_repeat1] = STATE(2379),
    [ts_builtin_sym_end] = ACTIONS(6616),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_namespace] = ACTIONS(6618),
    [anon_sym_module] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_open] = ACTIONS(6618),
    [anon_sym_type] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_and] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(6592),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_exception] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [anon_sym_extern] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6616),
    [anon_sym_POUNDload] = ACTIONS(6616),
    [anon_sym_POUNDtime] = ACTIONS(6616),
    [anon_sym_POUNDI] = ACTIONS(6616),
    [anon_sym_POUNDhelp] = ACTIONS(6616),
    [anon_sym_POUNDquit] = ACTIONS(6616),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(6620),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2371)] = {
    [sym_block_comment] = STATE(2371),
    [sym_compiler_directive_decl] = STATE(2371),
    [sym_fsi_directive_decl] = STATE(2371),
    [sym_preproc_line] = STATE(2371),
    [aux_sym_long_identifier_repeat1] = STATE(2374),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(6623),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2372)] = {
    [sym_block_comment] = STATE(2372),
    [sym_compiler_directive_decl] = STATE(2372),
    [sym_fsi_directive_decl] = STATE(2372),
    [sym_preproc_line] = STATE(2372),
    [aux_sym_long_identifier_repeat1] = STATE(2371),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_module] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_open] = ACTIONS(6625),
    [anon_sym_type] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_and] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(6631),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_exception] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [anon_sym_extern] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6628),
    [anon_sym_POUNDload] = ACTIONS(6628),
    [anon_sym_POUNDtime] = ACTIONS(6628),
    [anon_sym_POUNDI] = ACTIONS(6628),
    [anon_sym_POUNDhelp] = ACTIONS(6628),
    [anon_sym_POUNDquit] = ACTIONS(6628),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [anon_sym_POUNDelse] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(2373)] = {
    [sym_block_comment] = STATE(2373),
    [sym_compiler_directive_decl] = STATE(2373),
    [sym_fsi_directive_decl] = STATE(2373),
    [sym_preproc_line] = STATE(2373),
    [ts_builtin_sym_end] = ACTIONS(6635),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_namespace] = ACTIONS(6637),
    [anon_sym_module] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_open] = ACTIONS(6637),
    [anon_sym_type] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_and] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_exception] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [anon_sym_extern] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(6639),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6635),
    [anon_sym_POUNDload] = ACTIONS(6635),
    [anon_sym_POUNDtime] = ACTIONS(6635),
    [anon_sym_POUNDI] = ACTIONS(6635),
    [anon_sym_POUNDhelp] = ACTIONS(6635),
    [anon_sym_POUNDquit] = ACTIONS(6635),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(2374)] = {
    [sym_block_comment] = STATE(2374),
    [sym_compiler_directive_decl] = STATE(2374),
    [sym_fsi_directive_decl] = STATE(2374),
    [sym_preproc_line] = STATE(2374),
    [aux_sym_long_identifier_repeat1] = STATE(2374),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6641),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2375)] = {
    [sym_block_comment] = STATE(2375),
    [sym_compiler_directive_decl] = STATE(2375),
    [sym_fsi_directive_decl] = STATE(2375),
    [sym_preproc_line] = STATE(2375),
    [aux_sym_rules_repeat1] = STATE(2390),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_module] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_open] = ACTIONS(6618),
    [anon_sym_type] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_and] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(6644),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_exception] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [anon_sym_extern] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6616),
    [anon_sym_POUNDload] = ACTIONS(6616),
    [anon_sym_POUNDtime] = ACTIONS(6616),
    [anon_sym_POUNDI] = ACTIONS(6616),
    [anon_sym_POUNDhelp] = ACTIONS(6616),
    [anon_sym_POUNDquit] = ACTIONS(6616),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [anon_sym_POUNDelse] = ACTIONS(6616),
    [sym__newline] = ACTIONS(6646),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2376)] = {
    [sym_block_comment] = STATE(2376),
    [sym_compiler_directive_decl] = STATE(2376),
    [sym_fsi_directive_decl] = STATE(2376),
    [sym_preproc_line] = STATE(2376),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2377)] = {
    [sym_block_comment] = STATE(2377),
    [sym_compiler_directive_decl] = STATE(2377),
    [sym_fsi_directive_decl] = STATE(2377),
    [sym_preproc_line] = STATE(2377),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6039),
    [aux_sym_decimal_token1] = ACTIONS(5977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2378)] = {
    [sym_block_comment] = STATE(2378),
    [sym_compiler_directive_decl] = STATE(2378),
    [sym_fsi_directive_decl] = STATE(2378),
    [sym_preproc_line] = STATE(2378),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_module] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_open] = ACTIONS(6637),
    [anon_sym_type] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_and] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_exception] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [anon_sym_extern] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(6649),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6635),
    [anon_sym_POUNDload] = ACTIONS(6635),
    [anon_sym_POUNDtime] = ACTIONS(6635),
    [anon_sym_POUNDI] = ACTIONS(6635),
    [anon_sym_POUNDhelp] = ACTIONS(6635),
    [anon_sym_POUNDquit] = ACTIONS(6635),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [anon_sym_POUNDelse] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(2379)] = {
    [sym_block_comment] = STATE(2379),
    [sym_compiler_directive_decl] = STATE(2379),
    [sym_fsi_directive_decl] = STATE(2379),
    [sym_preproc_line] = STATE(2379),
    [aux_sym_rules_repeat1] = STATE(2379),
    [ts_builtin_sym_end] = ACTIONS(6651),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_namespace] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6655),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6658),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2380)] = {
    [sym_block_comment] = STATE(2380),
    [sym_compiler_directive_decl] = STATE(2380),
    [sym_fsi_directive_decl] = STATE(2380),
    [sym_preproc_line] = STATE(2380),
    [aux_sym_long_identifier_repeat1] = STATE(2381),
    [ts_builtin_sym_end] = ACTIONS(6358),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_namespace] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(6661),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2381)] = {
    [sym_block_comment] = STATE(2381),
    [sym_compiler_directive_decl] = STATE(2381),
    [sym_fsi_directive_decl] = STATE(2381),
    [sym_preproc_line] = STATE(2381),
    [aux_sym_long_identifier_repeat1] = STATE(2381),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6663),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2382)] = {
    [sym_block_comment] = STATE(2382),
    [sym_compiler_directive_decl] = STATE(2382),
    [sym_fsi_directive_decl] = STATE(2382),
    [sym_preproc_line] = STATE(2382),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6434),
    [anon_sym_uy] = ACTIONS(6436),
    [anon_sym_s] = ACTIONS(6438),
    [anon_sym_us] = ACTIONS(6440),
    [anon_sym_l] = ACTIONS(6442),
    [aux_sym_uint32_token1] = ACTIONS(6444),
    [anon_sym_n] = ACTIONS(6446),
    [anon_sym_un] = ACTIONS(6448),
    [anon_sym_L] = ACTIONS(6450),
    [aux_sym_uint64_token1] = ACTIONS(6452),
    [anon_sym_lf] = ACTIONS(6666),
    [anon_sym_LF] = ACTIONS(6668),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2383)] = {
    [sym_block_comment] = STATE(2383),
    [sym_compiler_directive_decl] = STATE(2383),
    [sym_fsi_directive_decl] = STATE(2383),
    [sym_preproc_line] = STATE(2383),
    [ts_builtin_sym_end] = ACTIONS(6515),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_namespace] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2384)] = {
    [sym_block_comment] = STATE(2384),
    [sym_compiler_directive_decl] = STATE(2384),
    [sym_fsi_directive_decl] = STATE(2384),
    [sym_preproc_line] = STATE(2384),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2385)] = {
    [sym_block_comment] = STATE(2385),
    [sym_compiler_directive_decl] = STATE(2385),
    [sym_fsi_directive_decl] = STATE(2385),
    [sym_preproc_line] = STATE(2385),
    [aux_sym_rules_repeat1] = STATE(2389),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_module] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_open] = ACTIONS(6590),
    [anon_sym_type] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_and] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(6644),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_exception] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [anon_sym_extern] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6588),
    [anon_sym_POUNDload] = ACTIONS(6588),
    [anon_sym_POUNDtime] = ACTIONS(6588),
    [anon_sym_POUNDI] = ACTIONS(6588),
    [anon_sym_POUNDhelp] = ACTIONS(6588),
    [anon_sym_POUNDquit] = ACTIONS(6588),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [anon_sym_POUNDelse] = ACTIONS(6588),
    [sym__newline] = ACTIONS(6670),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2386)] = {
    [sym_block_comment] = STATE(2386),
    [sym_compiler_directive_decl] = STATE(2386),
    [sym_fsi_directive_decl] = STATE(2386),
    [sym_preproc_line] = STATE(2386),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_module] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_open] = ACTIONS(6560),
    [anon_sym_type] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_and] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_exception] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [anon_sym_extern] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6558),
    [anon_sym_POUNDload] = ACTIONS(6558),
    [anon_sym_POUNDtime] = ACTIONS(6558),
    [anon_sym_POUNDI] = ACTIONS(6558),
    [anon_sym_POUNDhelp] = ACTIONS(6558),
    [anon_sym_POUNDquit] = ACTIONS(6558),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [anon_sym_POUNDelse] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(2387)] = {
    [sym_block_comment] = STATE(2387),
    [sym_compiler_directive_decl] = STATE(2387),
    [sym_fsi_directive_decl] = STATE(2387),
    [sym_preproc_line] = STATE(2387),
    [ts_builtin_sym_end] = ACTIONS(6614),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_namespace] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2388)] = {
    [sym_block_comment] = STATE(2388),
    [sym_compiler_directive_decl] = STATE(2388),
    [sym_fsi_directive_decl] = STATE(2388),
    [sym_preproc_line] = STATE(2388),
    [aux_sym_rules_repeat1] = STATE(2375),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(6644),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(6673),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2389)] = {
    [sym_block_comment] = STATE(2389),
    [sym_compiler_directive_decl] = STATE(2389),
    [sym_fsi_directive_decl] = STATE(2389),
    [sym_preproc_line] = STATE(2389),
    [aux_sym_rules_repeat1] = STATE(2390),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(6644),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(6673),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2390)] = {
    [sym_block_comment] = STATE(2390),
    [sym_compiler_directive_decl] = STATE(2390),
    [sym_fsi_directive_decl] = STATE(2390),
    [sym_preproc_line] = STATE(2390),
    [aux_sym_rules_repeat1] = STATE(2390),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6676),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6679),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2391)] = {
    [sym_type_arguments] = STATE(2732),
    [sym_long_identifier] = STATE(2698),
    [sym_block_comment] = STATE(2391),
    [sym_compiler_directive_decl] = STATE(2391),
    [sym_fsi_directive_decl] = STATE(2391),
    [sym_preproc_line] = STATE(2391),
    [aux_sym_compound_type_repeat1] = STATE(2486),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6576),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6578),
    [anon_sym_LT] = ACTIONS(6580),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6582),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [anon_sym_inherit] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [anon_sym_POUNDelse] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2392)] = {
    [sym_type_arguments] = STATE(2732),
    [sym_long_identifier] = STATE(2698),
    [sym_block_comment] = STATE(2392),
    [sym_compiler_directive_decl] = STATE(2392),
    [sym_fsi_directive_decl] = STATE(2392),
    [sym_preproc_line] = STATE(2392),
    [aux_sym_compound_type_repeat1] = STATE(2486),
    [sym_identifier] = ACTIONS(6574),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6576),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6578),
    [anon_sym_LT] = ACTIONS(6580),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6582),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_inherit] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2393)] = {
    [sym_type_arguments] = STATE(2766),
    [sym_long_identifier] = STATE(2775),
    [sym_block_comment] = STATE(2393),
    [sym_compiler_directive_decl] = STATE(2393),
    [sym_fsi_directive_decl] = STATE(2393),
    [sym_preproc_line] = STATE(2393),
    [aux_sym_compound_type_repeat1] = STATE(2505),
    [sym_identifier] = ACTIONS(6586),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6566),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6568),
    [anon_sym_LT] = ACTIONS(6570),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6572),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2394)] = {
    [sym_block_comment] = STATE(2394),
    [sym_compiler_directive_decl] = STATE(2394),
    [sym_fsi_directive_decl] = STATE(2394),
    [sym_preproc_line] = STATE(2394),
    [ts_builtin_sym_end] = ACTIONS(6614),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_namespace] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2395)] = {
    [sym_block_comment] = STATE(2395),
    [sym_compiler_directive_decl] = STATE(2395),
    [sym_fsi_directive_decl] = STATE(2395),
    [sym_preproc_line] = STATE(2395),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6462),
    [anon_sym_uy] = ACTIONS(6464),
    [anon_sym_s] = ACTIONS(6466),
    [anon_sym_us] = ACTIONS(6468),
    [anon_sym_l] = ACTIONS(6470),
    [aux_sym_uint32_token1] = ACTIONS(6472),
    [anon_sym_n] = ACTIONS(6474),
    [anon_sym_un] = ACTIONS(6476),
    [anon_sym_L] = ACTIONS(6478),
    [aux_sym_uint64_token1] = ACTIONS(6480),
    [anon_sym_lf] = ACTIONS(6682),
    [anon_sym_LF] = ACTIONS(6684),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2396)] = {
    [sym_block_comment] = STATE(2396),
    [sym_compiler_directive_decl] = STATE(2396),
    [sym_fsi_directive_decl] = STATE(2396),
    [sym_preproc_line] = STATE(2396),
    [aux_sym_long_identifier_repeat1] = STATE(2380),
    [ts_builtin_sym_end] = ACTIONS(6628),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_namespace] = ACTIONS(6625),
    [anon_sym_module] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_open] = ACTIONS(6625),
    [anon_sym_type] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_and] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(6686),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_exception] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [anon_sym_extern] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6628),
    [anon_sym_POUNDload] = ACTIONS(6628),
    [anon_sym_POUNDtime] = ACTIONS(6628),
    [anon_sym_POUNDI] = ACTIONS(6628),
    [anon_sym_POUNDhelp] = ACTIONS(6628),
    [anon_sym_POUNDquit] = ACTIONS(6628),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(2397)] = {
    [sym_block_comment] = STATE(2397),
    [sym_compiler_directive_decl] = STATE(2397),
    [sym_fsi_directive_decl] = STATE(2397),
    [sym_preproc_line] = STATE(2397),
    [ts_builtin_sym_end] = ACTIONS(6599),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_namespace] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(6690),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2398)] = {
    [sym_block_comment] = STATE(2398),
    [sym_compiler_directive_decl] = STATE(2398),
    [sym_fsi_directive_decl] = STATE(2398),
    [sym_preproc_line] = STATE(2398),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_module] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_open] = ACTIONS(6564),
    [anon_sym_type] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_and] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_exception] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [anon_sym_extern] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6562),
    [anon_sym_POUNDload] = ACTIONS(6562),
    [anon_sym_POUNDtime] = ACTIONS(6562),
    [anon_sym_POUNDI] = ACTIONS(6562),
    [anon_sym_POUNDhelp] = ACTIONS(6562),
    [anon_sym_POUNDquit] = ACTIONS(6562),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [anon_sym_POUNDelse] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2399)] = {
    [sym_block_comment] = STATE(2399),
    [sym_compiler_directive_decl] = STATE(2399),
    [sym_fsi_directive_decl] = STATE(2399),
    [sym_preproc_line] = STATE(2399),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_module] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_open] = ACTIONS(6692),
    [anon_sym_type] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_and] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_exception] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [anon_sym_extern] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6694),
    [anon_sym_POUNDload] = ACTIONS(6694),
    [anon_sym_POUNDtime] = ACTIONS(6694),
    [anon_sym_POUNDI] = ACTIONS(6694),
    [anon_sym_POUNDhelp] = ACTIONS(6694),
    [anon_sym_POUNDquit] = ACTIONS(6694),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [anon_sym_POUNDelse] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(2400)] = {
    [sym_block_comment] = STATE(2400),
    [sym_compiler_directive_decl] = STATE(2400),
    [sym_fsi_directive_decl] = STATE(2400),
    [sym_preproc_line] = STATE(2400),
    [ts_builtin_sym_end] = ACTIONS(6599),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_namespace] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(6690),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2401)] = {
    [sym_block_comment] = STATE(2401),
    [sym_compiler_directive_decl] = STATE(2401),
    [sym_fsi_directive_decl] = STATE(2401),
    [sym_preproc_line] = STATE(2401),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_GT_RBRACK] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_RBRACK] = ACTIONS(5953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_RBRACE] = ACTIONS(5953),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_to] = ACTIONS(5951),
    [anon_sym_downto] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6696),
    [anon_sym_uy] = ACTIONS(6698),
    [anon_sym_s] = ACTIONS(6700),
    [anon_sym_us] = ACTIONS(6702),
    [anon_sym_l] = ACTIONS(6704),
    [aux_sym_uint32_token1] = ACTIONS(6706),
    [anon_sym_n] = ACTIONS(6708),
    [anon_sym_un] = ACTIONS(6710),
    [anon_sym_L] = ACTIONS(6712),
    [aux_sym_uint64_token1] = ACTIONS(6714),
    [aux_sym_bignum_token1] = ACTIONS(6716),
    [aux_sym_decimal_token1] = ACTIONS(6718),
    [anon_sym_DOT2] = ACTIONS(6720),
    [aux_sym_float_token1] = ACTIONS(6722),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2402)] = {
    [sym_type_arguments] = STATE(2868),
    [sym_long_identifier] = STATE(2873),
    [sym_block_comment] = STATE(2402),
    [sym_compiler_directive_decl] = STATE(2402),
    [sym_fsi_directive_decl] = STATE(2402),
    [sym_preproc_line] = STATE(2402),
    [aux_sym_compound_type_repeat1] = STATE(2796),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6724),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_DOT_DOT] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6726),
    [anon_sym_LT] = ACTIONS(6728),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6730),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__dedent] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2403)] = {
    [sym_type_arguments] = STATE(2868),
    [sym_long_identifier] = STATE(2873),
    [sym_block_comment] = STATE(2403),
    [sym_compiler_directive_decl] = STATE(2403),
    [sym_fsi_directive_decl] = STATE(2403),
    [sym_preproc_line] = STATE(2403),
    [aux_sym_compound_type_repeat1] = STATE(2796),
    [sym_identifier] = ACTIONS(6732),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6724),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_DOT_DOT] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6726),
    [anon_sym_LT] = ACTIONS(6728),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6730),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__dedent] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2404)] = {
    [sym_block_comment] = STATE(2404),
    [sym_compiler_directive_decl] = STATE(2404),
    [sym_fsi_directive_decl] = STATE(2404),
    [sym_preproc_line] = STATE(2404),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6519),
    [anon_sym_uy] = ACTIONS(6521),
    [anon_sym_s] = ACTIONS(6523),
    [anon_sym_us] = ACTIONS(6525),
    [anon_sym_l] = ACTIONS(6527),
    [aux_sym_uint32_token1] = ACTIONS(6529),
    [anon_sym_n] = ACTIONS(6531),
    [anon_sym_un] = ACTIONS(6533),
    [anon_sym_L] = ACTIONS(6535),
    [aux_sym_uint64_token1] = ACTIONS(6537),
    [anon_sym_lf] = ACTIONS(6734),
    [anon_sym_LF] = ACTIONS(6736),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2405)] = {
    [sym_block_comment] = STATE(2405),
    [sym_compiler_directive_decl] = STATE(2405),
    [sym_fsi_directive_decl] = STATE(2405),
    [sym_preproc_line] = STATE(2405),
    [aux_sym_sequential_expression_repeat1] = STATE(2422),
    [ts_builtin_sym_end] = ACTIONS(6738),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_namespace] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(2406)] = {
    [sym_type_arguments] = STATE(2834),
    [sym_long_identifier] = STATE(2844),
    [sym_block_comment] = STATE(2406),
    [sym_compiler_directive_decl] = STATE(2406),
    [sym_fsi_directive_decl] = STATE(2406),
    [sym_preproc_line] = STATE(2406),
    [aux_sym_compound_type_repeat1] = STATE(2726),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6742),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6744),
    [anon_sym_LT] = ACTIONS(6746),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6748),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [anon_sym_inherit] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2407)] = {
    [sym_block_comment] = STATE(2407),
    [sym_compiler_directive_decl] = STATE(2407),
    [sym_fsi_directive_decl] = STATE(2407),
    [sym_preproc_line] = STATE(2407),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_module] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_open] = ACTIONS(6750),
    [anon_sym_type] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_and] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_exception] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6752),
    [anon_sym_POUNDload] = ACTIONS(6752),
    [anon_sym_POUNDtime] = ACTIONS(6752),
    [anon_sym_POUNDI] = ACTIONS(6752),
    [anon_sym_POUNDhelp] = ACTIONS(6752),
    [anon_sym_POUNDquit] = ACTIONS(6752),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [anon_sym_POUNDelse] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(2408)] = {
    [sym_type_arguments] = STATE(2904),
    [sym_long_identifier] = STATE(2901),
    [sym_block_comment] = STATE(2408),
    [sym_compiler_directive_decl] = STATE(2408),
    [sym_fsi_directive_decl] = STATE(2408),
    [sym_preproc_line] = STATE(2408),
    [aux_sym_compound_type_repeat1] = STATE(2815),
    [sym_identifier] = ACTIONS(6754),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6756),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6760),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6762),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2409)] = {
    [sym_block_comment] = STATE(2409),
    [sym_compiler_directive_decl] = STATE(2409),
    [sym_fsi_directive_decl] = STATE(2409),
    [sym_preproc_line] = STATE(2409),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(6764),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2410)] = {
    [sym_block_comment] = STATE(2410),
    [sym_compiler_directive_decl] = STATE(2410),
    [sym_fsi_directive_decl] = STATE(2410),
    [sym_preproc_line] = STATE(2410),
    [ts_builtin_sym_end] = ACTIONS(6766),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_namespace] = ACTIONS(6768),
    [anon_sym_module] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_open] = ACTIONS(6768),
    [anon_sym_type] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_and] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(6770),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_exception] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [anon_sym_extern] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6766),
    [anon_sym_POUNDload] = ACTIONS(6766),
    [anon_sym_POUNDtime] = ACTIONS(6766),
    [anon_sym_POUNDI] = ACTIONS(6766),
    [anon_sym_POUNDhelp] = ACTIONS(6766),
    [anon_sym_POUNDquit] = ACTIONS(6766),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(2411)] = {
    [sym_type_arguments] = STATE(2868),
    [sym_long_identifier] = STATE(2873),
    [sym_block_comment] = STATE(2411),
    [sym_compiler_directive_decl] = STATE(2411),
    [sym_fsi_directive_decl] = STATE(2411),
    [sym_preproc_line] = STATE(2411),
    [aux_sym_compound_type_repeat1] = STATE(2796),
    [sym_identifier] = ACTIONS(6732),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6724),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_DOT_DOT] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6726),
    [anon_sym_LT] = ACTIONS(6728),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6730),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__dedent] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2412)] = {
    [sym_type_arguments] = STATE(2834),
    [sym_long_identifier] = STATE(2844),
    [sym_block_comment] = STATE(2412),
    [sym_compiler_directive_decl] = STATE(2412),
    [sym_fsi_directive_decl] = STATE(2412),
    [sym_preproc_line] = STATE(2412),
    [aux_sym_compound_type_repeat1] = STATE(2726),
    [sym_identifier] = ACTIONS(6772),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6742),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6744),
    [anon_sym_LT] = ACTIONS(6746),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6748),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_inherit] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2413)] = {
    [sym_type_arguments] = STATE(2836),
    [sym_long_identifier] = STATE(2859),
    [sym_block_comment] = STATE(2413),
    [sym_compiler_directive_decl] = STATE(2413),
    [sym_fsi_directive_decl] = STATE(2413),
    [sym_preproc_line] = STATE(2413),
    [aux_sym_compound_type_repeat1] = STATE(2765),
    [sym_identifier] = ACTIONS(6774),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6776),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6778),
    [anon_sym_LT] = ACTIONS(6780),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6782),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__dedent] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2414)] = {
    [sym_block_comment] = STATE(2414),
    [sym_compiler_directive_decl] = STATE(2414),
    [sym_fsi_directive_decl] = STATE(2414),
    [sym_preproc_line] = STATE(2414),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6462),
    [anon_sym_uy] = ACTIONS(6464),
    [anon_sym_s] = ACTIONS(6466),
    [anon_sym_us] = ACTIONS(6468),
    [anon_sym_l] = ACTIONS(6470),
    [aux_sym_uint32_token1] = ACTIONS(6472),
    [anon_sym_n] = ACTIONS(6474),
    [anon_sym_un] = ACTIONS(6476),
    [anon_sym_L] = ACTIONS(6478),
    [aux_sym_uint64_token1] = ACTIONS(6480),
    [anon_sym_lf] = ACTIONS(6682),
    [anon_sym_LF] = ACTIONS(6684),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2415)] = {
    [sym_type_arguments] = STATE(2836),
    [sym_long_identifier] = STATE(2859),
    [sym_block_comment] = STATE(2415),
    [sym_compiler_directive_decl] = STATE(2415),
    [sym_fsi_directive_decl] = STATE(2415),
    [sym_preproc_line] = STATE(2415),
    [aux_sym_compound_type_repeat1] = STATE(2765),
    [sym_identifier] = ACTIONS(6774),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6776),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6778),
    [anon_sym_LT] = ACTIONS(6780),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6782),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2416)] = {
    [sym_block_comment] = STATE(2416),
    [sym_compiler_directive_decl] = STATE(2416),
    [sym_fsi_directive_decl] = STATE(2416),
    [sym_preproc_line] = STATE(2416),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_module] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_open] = ACTIONS(6784),
    [anon_sym_type] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_and] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(6788),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_exception] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6786),
    [anon_sym_POUNDload] = ACTIONS(6786),
    [anon_sym_POUNDtime] = ACTIONS(6786),
    [anon_sym_POUNDI] = ACTIONS(6786),
    [anon_sym_POUNDhelp] = ACTIONS(6786),
    [anon_sym_POUNDquit] = ACTIONS(6786),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [anon_sym_POUNDelse] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(2417)] = {
    [sym_type_arguments] = STATE(2836),
    [sym_long_identifier] = STATE(2859),
    [sym_block_comment] = STATE(2417),
    [sym_compiler_directive_decl] = STATE(2417),
    [sym_fsi_directive_decl] = STATE(2417),
    [sym_preproc_line] = STATE(2417),
    [aux_sym_compound_type_repeat1] = STATE(2765),
    [sym_identifier] = ACTIONS(6774),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6776),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6778),
    [anon_sym_LT] = ACTIONS(6780),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6782),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__dedent] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2418)] = {
    [sym_type_arguments] = STATE(2836),
    [sym_long_identifier] = STATE(2859),
    [sym_block_comment] = STATE(2418),
    [sym_compiler_directive_decl] = STATE(2418),
    [sym_fsi_directive_decl] = STATE(2418),
    [sym_preproc_line] = STATE(2418),
    [aux_sym_compound_type_repeat1] = STATE(2765),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_and] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6776),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6778),
    [anon_sym_LT] = ACTIONS(6780),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6782),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__dedent] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2419)] = {
    [sym_block_comment] = STATE(2419),
    [sym_compiler_directive_decl] = STATE(2419),
    [sym_fsi_directive_decl] = STATE(2419),
    [sym_preproc_line] = STATE(2419),
    [aux_sym_sequential_expression_repeat1] = STATE(2484),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [anon_sym_POUNDelse] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(2420)] = {
    [sym_block_comment] = STATE(2420),
    [sym_compiler_directive_decl] = STATE(2420),
    [sym_fsi_directive_decl] = STATE(2420),
    [sym_preproc_line] = STATE(2420),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_module] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_open] = ACTIONS(6790),
    [anon_sym_type] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_and] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_exception] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6792),
    [anon_sym_POUNDload] = ACTIONS(6792),
    [anon_sym_POUNDtime] = ACTIONS(6792),
    [anon_sym_POUNDI] = ACTIONS(6792),
    [anon_sym_POUNDhelp] = ACTIONS(6792),
    [anon_sym_POUNDquit] = ACTIONS(6792),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [anon_sym_POUNDelse] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(2421)] = {
    [sym_block_comment] = STATE(2421),
    [sym_compiler_directive_decl] = STATE(2421),
    [sym_fsi_directive_decl] = STATE(2421),
    [sym_preproc_line] = STATE(2421),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_module] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_open] = ACTIONS(6794),
    [anon_sym_type] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_and] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_exception] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6796),
    [anon_sym_POUNDload] = ACTIONS(6796),
    [anon_sym_POUNDtime] = ACTIONS(6796),
    [anon_sym_POUNDI] = ACTIONS(6796),
    [anon_sym_POUNDhelp] = ACTIONS(6796),
    [anon_sym_POUNDquit] = ACTIONS(6796),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [anon_sym_POUNDelse] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(2422)] = {
    [sym_block_comment] = STATE(2422),
    [sym_compiler_directive_decl] = STATE(2422),
    [sym_fsi_directive_decl] = STATE(2422),
    [sym_preproc_line] = STATE(2422),
    [aux_sym_sequential_expression_repeat1] = STATE(2422),
    [ts_builtin_sym_end] = ACTIONS(6798),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_namespace] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(6802),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(2423)] = {
    [sym_block_comment] = STATE(2423),
    [sym_compiler_directive_decl] = STATE(2423),
    [sym_fsi_directive_decl] = STATE(2423),
    [sym_preproc_line] = STATE(2423),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2424)] = {
    [sym_block_comment] = STATE(2424),
    [sym_compiler_directive_decl] = STATE(2424),
    [sym_fsi_directive_decl] = STATE(2424),
    [sym_preproc_line] = STATE(2424),
    [ts_builtin_sym_end] = ACTIONS(6805),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_namespace] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(2425)] = {
    [sym_block_comment] = STATE(2425),
    [sym_compiler_directive_decl] = STATE(2425),
    [sym_fsi_directive_decl] = STATE(2425),
    [sym_preproc_line] = STATE(2425),
    [ts_builtin_sym_end] = ACTIONS(6811),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_namespace] = ACTIONS(6813),
    [anon_sym_module] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_open] = ACTIONS(6813),
    [anon_sym_type] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_and] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_exception] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [anon_sym_extern] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6811),
    [anon_sym_POUNDload] = ACTIONS(6811),
    [anon_sym_POUNDtime] = ACTIONS(6811),
    [anon_sym_POUNDI] = ACTIONS(6811),
    [anon_sym_POUNDhelp] = ACTIONS(6811),
    [anon_sym_POUNDquit] = ACTIONS(6811),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(2426)] = {
    [sym_block_comment] = STATE(2426),
    [sym_compiler_directive_decl] = STATE(2426),
    [sym_fsi_directive_decl] = STATE(2426),
    [sym_preproc_line] = STATE(2426),
    [ts_builtin_sym_end] = ACTIONS(6614),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_namespace] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2427)] = {
    [sym_type_arguments] = STATE(2834),
    [sym_long_identifier] = STATE(2844),
    [sym_block_comment] = STATE(2427),
    [sym_compiler_directive_decl] = STATE(2427),
    [sym_fsi_directive_decl] = STATE(2427),
    [sym_preproc_line] = STATE(2427),
    [aux_sym_compound_type_repeat1] = STATE(2726),
    [sym_identifier] = ACTIONS(6772),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6742),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6744),
    [anon_sym_LT] = ACTIONS(6746),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6748),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_inherit] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2428)] = {
    [sym_type_arguments] = STATE(2834),
    [sym_long_identifier] = STATE(2844),
    [sym_block_comment] = STATE(2428),
    [sym_compiler_directive_decl] = STATE(2428),
    [sym_fsi_directive_decl] = STATE(2428),
    [sym_preproc_line] = STATE(2428),
    [aux_sym_compound_type_repeat1] = STATE(2726),
    [sym_identifier] = ACTIONS(6772),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6742),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6744),
    [anon_sym_LT] = ACTIONS(6746),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6748),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_inherit] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2429)] = {
    [sym_type_arguments] = STATE(2834),
    [sym_long_identifier] = STATE(2844),
    [sym_block_comment] = STATE(2429),
    [sym_compiler_directive_decl] = STATE(2429),
    [sym_fsi_directive_decl] = STATE(2429),
    [sym_preproc_line] = STATE(2429),
    [aux_sym_compound_type_repeat1] = STATE(2726),
    [sym_identifier] = ACTIONS(6772),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6742),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6744),
    [anon_sym_LT] = ACTIONS(6746),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6748),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_inherit] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2430)] = {
    [sym_type_arguments] = STATE(2868),
    [sym_long_identifier] = STATE(2873),
    [sym_block_comment] = STATE(2430),
    [sym_compiler_directive_decl] = STATE(2430),
    [sym_fsi_directive_decl] = STATE(2430),
    [sym_preproc_line] = STATE(2430),
    [aux_sym_compound_type_repeat1] = STATE(2796),
    [sym_identifier] = ACTIONS(6732),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6724),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6726),
    [anon_sym_LT] = ACTIONS(6728),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6730),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2431)] = {
    [sym_block_comment] = STATE(2431),
    [sym_compiler_directive_decl] = STATE(2431),
    [sym_fsi_directive_decl] = STATE(2431),
    [sym_preproc_line] = STATE(2431),
    [aux_sym_long_identifier_repeat1] = STATE(2432),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(6815),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2432)] = {
    [sym_block_comment] = STATE(2432),
    [sym_compiler_directive_decl] = STATE(2432),
    [sym_fsi_directive_decl] = STATE(2432),
    [sym_preproc_line] = STATE(2432),
    [aux_sym_long_identifier_repeat1] = STATE(2432),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6817),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2433)] = {
    [sym_block_comment] = STATE(2433),
    [sym_compiler_directive_decl] = STATE(2433),
    [sym_fsi_directive_decl] = STATE(2433),
    [sym_preproc_line] = STATE(2433),
    [ts_builtin_sym_end] = ACTIONS(6752),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_namespace] = ACTIONS(6750),
    [anon_sym_module] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_open] = ACTIONS(6750),
    [anon_sym_type] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_and] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_exception] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6752),
    [anon_sym_POUNDload] = ACTIONS(6752),
    [anon_sym_POUNDtime] = ACTIONS(6752),
    [anon_sym_POUNDI] = ACTIONS(6752),
    [anon_sym_POUNDhelp] = ACTIONS(6752),
    [anon_sym_POUNDquit] = ACTIONS(6752),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(2434)] = {
    [sym_block_comment] = STATE(2434),
    [sym_compiler_directive_decl] = STATE(2434),
    [sym_fsi_directive_decl] = STATE(2434),
    [sym_preproc_line] = STATE(2434),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(6601),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2435)] = {
    [sym_block_comment] = STATE(2435),
    [sym_compiler_directive_decl] = STATE(2435),
    [sym_fsi_directive_decl] = STATE(2435),
    [sym_preproc_line] = STATE(2435),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6047),
    [aux_sym_decimal_token1] = ACTIONS(6033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2436)] = {
    [sym_block_comment] = STATE(2436),
    [sym_compiler_directive_decl] = STATE(2436),
    [sym_fsi_directive_decl] = STATE(2436),
    [sym_preproc_line] = STATE(2436),
    [ts_builtin_sym_end] = ACTIONS(6786),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_namespace] = ACTIONS(6784),
    [anon_sym_module] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_open] = ACTIONS(6784),
    [anon_sym_type] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_and] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(6820),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_exception] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6786),
    [anon_sym_POUNDload] = ACTIONS(6786),
    [anon_sym_POUNDtime] = ACTIONS(6786),
    [anon_sym_POUNDI] = ACTIONS(6786),
    [anon_sym_POUNDhelp] = ACTIONS(6786),
    [anon_sym_POUNDquit] = ACTIONS(6786),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(2437)] = {
    [sym_type_arguments] = STATE(2904),
    [sym_long_identifier] = STATE(2901),
    [sym_block_comment] = STATE(2437),
    [sym_compiler_directive_decl] = STATE(2437),
    [sym_fsi_directive_decl] = STATE(2437),
    [sym_preproc_line] = STATE(2437),
    [aux_sym_compound_type_repeat1] = STATE(2815),
    [sym_identifier] = ACTIONS(6754),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(6756),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6760),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6762),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2438)] = {
    [sym_block_comment] = STATE(2438),
    [sym_compiler_directive_decl] = STATE(2438),
    [sym_fsi_directive_decl] = STATE(2438),
    [sym_preproc_line] = STATE(2438),
    [ts_builtin_sym_end] = ACTIONS(6822),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_namespace] = ACTIONS(6824),
    [anon_sym_module] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_open] = ACTIONS(6824),
    [anon_sym_type] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_and] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_exception] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6822),
    [anon_sym_POUNDload] = ACTIONS(6822),
    [anon_sym_POUNDtime] = ACTIONS(6822),
    [anon_sym_POUNDI] = ACTIONS(6822),
    [anon_sym_POUNDhelp] = ACTIONS(6822),
    [anon_sym_POUNDquit] = ACTIONS(6822),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(2439)] = {
    [sym_type_arguments] = STATE(2904),
    [sym_long_identifier] = STATE(2901),
    [sym_block_comment] = STATE(2439),
    [sym_compiler_directive_decl] = STATE(2439),
    [sym_fsi_directive_decl] = STATE(2439),
    [sym_preproc_line] = STATE(2439),
    [aux_sym_compound_type_repeat1] = STATE(2815),
    [sym_identifier] = ACTIONS(6754),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6756),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6760),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6762),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2440)] = {
    [sym_block_comment] = STATE(2440),
    [sym_compiler_directive_decl] = STATE(2440),
    [sym_fsi_directive_decl] = STATE(2440),
    [sym_preproc_line] = STATE(2440),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_module] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_open] = ACTIONS(6768),
    [anon_sym_type] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_and] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(6826),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_exception] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [anon_sym_extern] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6766),
    [anon_sym_POUNDload] = ACTIONS(6766),
    [anon_sym_POUNDtime] = ACTIONS(6766),
    [anon_sym_POUNDI] = ACTIONS(6766),
    [anon_sym_POUNDhelp] = ACTIONS(6766),
    [anon_sym_POUNDquit] = ACTIONS(6766),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [anon_sym_POUNDelse] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(2441)] = {
    [sym_type_arguments] = STATE(2904),
    [sym_long_identifier] = STATE(2901),
    [sym_block_comment] = STATE(2441),
    [sym_compiler_directive_decl] = STATE(2441),
    [sym_fsi_directive_decl] = STATE(2441),
    [sym_preproc_line] = STATE(2441),
    [aux_sym_compound_type_repeat1] = STATE(2815),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_and] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [aux_sym_access_modifier_token1] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6756),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6760),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6762),
    [anon_sym_default] = ACTIONS(6265),
    [anon_sym_static] = ACTIONS(6265),
    [anon_sym_member] = ACTIONS(6265),
    [anon_sym_interface] = ACTIONS(6265),
    [anon_sym_abstract] = ACTIONS(6265),
    [anon_sym_val] = ACTIONS(6265),
    [anon_sym_override] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2442)] = {
    [sym_block_comment] = STATE(2442),
    [sym_compiler_directive_decl] = STATE(2442),
    [sym_fsi_directive_decl] = STATE(2442),
    [sym_preproc_line] = STATE(2442),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2443)] = {
    [sym_block_comment] = STATE(2443),
    [sym_compiler_directive_decl] = STATE(2443),
    [sym_fsi_directive_decl] = STATE(2443),
    [sym_preproc_line] = STATE(2443),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_module] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_open] = ACTIONS(6828),
    [anon_sym_type] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_and] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_exception] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6830),
    [anon_sym_POUNDload] = ACTIONS(6830),
    [anon_sym_POUNDtime] = ACTIONS(6830),
    [anon_sym_POUNDI] = ACTIONS(6830),
    [anon_sym_POUNDhelp] = ACTIONS(6830),
    [anon_sym_POUNDquit] = ACTIONS(6830),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [anon_sym_POUNDelse] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(2444)] = {
    [sym_type_arguments] = STATE(2904),
    [sym_long_identifier] = STATE(2901),
    [sym_block_comment] = STATE(2444),
    [sym_compiler_directive_decl] = STATE(2444),
    [sym_fsi_directive_decl] = STATE(2444),
    [sym_preproc_line] = STATE(2444),
    [aux_sym_compound_type_repeat1] = STATE(2815),
    [sym_identifier] = ACTIONS(6754),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6756),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6760),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6762),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2445)] = {
    [sym_block_comment] = STATE(2445),
    [sym_compiler_directive_decl] = STATE(2445),
    [sym_fsi_directive_decl] = STATE(2445),
    [sym_preproc_line] = STATE(2445),
    [ts_builtin_sym_end] = ACTIONS(6832),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_namespace] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6836),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(2446)] = {
    [sym_block_comment] = STATE(2446),
    [sym_compiler_directive_decl] = STATE(2446),
    [sym_fsi_directive_decl] = STATE(2446),
    [sym_preproc_line] = STATE(2446),
    [ts_builtin_sym_end] = ACTIONS(6832),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_namespace] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(2447)] = {
    [sym_block_comment] = STATE(2447),
    [sym_compiler_directive_decl] = STATE(2447),
    [sym_fsi_directive_decl] = STATE(2447),
    [sym_preproc_line] = STATE(2447),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_module] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_open] = ACTIONS(6564),
    [anon_sym_type] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_and] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_exception] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [anon_sym_extern] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6562),
    [anon_sym_POUNDload] = ACTIONS(6562),
    [anon_sym_POUNDtime] = ACTIONS(6562),
    [anon_sym_POUNDI] = ACTIONS(6562),
    [anon_sym_POUNDhelp] = ACTIONS(6562),
    [anon_sym_POUNDquit] = ACTIONS(6562),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2448)] = {
    [sym_block_comment] = STATE(2448),
    [sym_compiler_directive_decl] = STATE(2448),
    [sym_fsi_directive_decl] = STATE(2448),
    [sym_preproc_line] = STATE(2448),
    [ts_builtin_sym_end] = ACTIONS(6838),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_namespace] = ACTIONS(6840),
    [anon_sym_module] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_open] = ACTIONS(6840),
    [anon_sym_type] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_and] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_exception] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [anon_sym_extern] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6838),
    [anon_sym_POUNDload] = ACTIONS(6838),
    [anon_sym_POUNDtime] = ACTIONS(6838),
    [anon_sym_POUNDI] = ACTIONS(6838),
    [anon_sym_POUNDhelp] = ACTIONS(6838),
    [anon_sym_POUNDquit] = ACTIONS(6838),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(2449)] = {
    [sym_block_comment] = STATE(2449),
    [sym_compiler_directive_decl] = STATE(2449),
    [sym_fsi_directive_decl] = STATE(2449),
    [sym_preproc_line] = STATE(2449),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2450)] = {
    [sym_block_comment] = STATE(2450),
    [sym_compiler_directive_decl] = STATE(2450),
    [sym_fsi_directive_decl] = STATE(2450),
    [sym_preproc_line] = STATE(2450),
    [ts_builtin_sym_end] = ACTIONS(6651),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_namespace] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2451)] = {
    [sym_block_comment] = STATE(2451),
    [sym_compiler_directive_decl] = STATE(2451),
    [sym_fsi_directive_decl] = STATE(2451),
    [sym_preproc_line] = STATE(2451),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(2452)] = {
    [sym_block_comment] = STATE(2452),
    [sym_compiler_directive_decl] = STATE(2452),
    [sym_fsi_directive_decl] = STATE(2452),
    [sym_preproc_line] = STATE(2452),
    [ts_builtin_sym_end] = ACTIONS(6694),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_namespace] = ACTIONS(6692),
    [anon_sym_module] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_open] = ACTIONS(6692),
    [anon_sym_type] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_and] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_exception] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [anon_sym_extern] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6694),
    [anon_sym_POUNDload] = ACTIONS(6694),
    [anon_sym_POUNDtime] = ACTIONS(6694),
    [anon_sym_POUNDI] = ACTIONS(6694),
    [anon_sym_POUNDhelp] = ACTIONS(6694),
    [anon_sym_POUNDquit] = ACTIONS(6694),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(2453)] = {
    [sym_block_comment] = STATE(2453),
    [sym_compiler_directive_decl] = STATE(2453),
    [sym_fsi_directive_decl] = STATE(2453),
    [sym_preproc_line] = STATE(2453),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_module] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_open] = ACTIONS(6554),
    [anon_sym_type] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_and] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_exception] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [anon_sym_extern] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6556),
    [anon_sym_POUNDload] = ACTIONS(6556),
    [anon_sym_POUNDtime] = ACTIONS(6556),
    [anon_sym_POUNDI] = ACTIONS(6556),
    [anon_sym_POUNDhelp] = ACTIONS(6556),
    [anon_sym_POUNDquit] = ACTIONS(6556),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2454)] = {
    [sym_block_comment] = STATE(2454),
    [sym_compiler_directive_decl] = STATE(2454),
    [sym_fsi_directive_decl] = STATE(2454),
    [sym_preproc_line] = STATE(2454),
    [ts_builtin_sym_end] = ACTIONS(6792),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_namespace] = ACTIONS(6790),
    [anon_sym_module] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_open] = ACTIONS(6790),
    [anon_sym_type] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_and] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_exception] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6792),
    [anon_sym_POUNDload] = ACTIONS(6792),
    [anon_sym_POUNDtime] = ACTIONS(6792),
    [anon_sym_POUNDI] = ACTIONS(6792),
    [anon_sym_POUNDhelp] = ACTIONS(6792),
    [anon_sym_POUNDquit] = ACTIONS(6792),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(2455)] = {
    [sym_block_comment] = STATE(2455),
    [sym_compiler_directive_decl] = STATE(2455),
    [sym_fsi_directive_decl] = STATE(2455),
    [sym_preproc_line] = STATE(2455),
    [ts_builtin_sym_end] = ACTIONS(6796),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_namespace] = ACTIONS(6794),
    [anon_sym_module] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_open] = ACTIONS(6794),
    [anon_sym_type] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_and] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_exception] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6796),
    [anon_sym_POUNDload] = ACTIONS(6796),
    [anon_sym_POUNDtime] = ACTIONS(6796),
    [anon_sym_POUNDI] = ACTIONS(6796),
    [anon_sym_POUNDhelp] = ACTIONS(6796),
    [anon_sym_POUNDquit] = ACTIONS(6796),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(2456)] = {
    [sym_block_comment] = STATE(2456),
    [sym_compiler_directive_decl] = STATE(2456),
    [sym_fsi_directive_decl] = STATE(2456),
    [sym_preproc_line] = STATE(2456),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_module] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_open] = ACTIONS(6637),
    [anon_sym_type] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_and] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_exception] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [anon_sym_extern] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(6842),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6635),
    [anon_sym_POUNDload] = ACTIONS(6635),
    [anon_sym_POUNDtime] = ACTIONS(6635),
    [anon_sym_POUNDI] = ACTIONS(6635),
    [anon_sym_POUNDhelp] = ACTIONS(6635),
    [anon_sym_POUNDquit] = ACTIONS(6635),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(2457)] = {
    [sym_block_comment] = STATE(2457),
    [sym_compiler_directive_decl] = STATE(2457),
    [sym_fsi_directive_decl] = STATE(2457),
    [sym_preproc_line] = STATE(2457),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_module] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_open] = ACTIONS(6840),
    [anon_sym_type] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_and] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_exception] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [anon_sym_extern] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6838),
    [anon_sym_POUNDload] = ACTIONS(6838),
    [anon_sym_POUNDtime] = ACTIONS(6838),
    [anon_sym_POUNDI] = ACTIONS(6838),
    [anon_sym_POUNDhelp] = ACTIONS(6838),
    [anon_sym_POUNDquit] = ACTIONS(6838),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [anon_sym_POUNDelse] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(2458)] = {
    [sym_block_comment] = STATE(2458),
    [sym_compiler_directive_decl] = STATE(2458),
    [sym_fsi_directive_decl] = STATE(2458),
    [sym_preproc_line] = STATE(2458),
    [aux_sym_rules_repeat1] = STATE(2463),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_module] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_open] = ACTIONS(6590),
    [anon_sym_type] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_and] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(6844),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_exception] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [anon_sym_extern] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6588),
    [anon_sym_POUNDload] = ACTIONS(6588),
    [anon_sym_POUNDtime] = ACTIONS(6588),
    [anon_sym_POUNDI] = ACTIONS(6588),
    [anon_sym_POUNDhelp] = ACTIONS(6588),
    [anon_sym_POUNDquit] = ACTIONS(6588),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(6846),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2459)] = {
    [sym_type_arguments] = STATE(2868),
    [sym_long_identifier] = STATE(2873),
    [sym_block_comment] = STATE(2459),
    [sym_compiler_directive_decl] = STATE(2459),
    [sym_fsi_directive_decl] = STATE(2459),
    [sym_preproc_line] = STATE(2459),
    [aux_sym_compound_type_repeat1] = STATE(2796),
    [sym_identifier] = ACTIONS(6732),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(6724),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_DOT_DOT] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(6726),
    [anon_sym_LT] = ACTIONS(6728),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6730),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__dedent] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2460)] = {
    [sym_block_comment] = STATE(2460),
    [sym_compiler_directive_decl] = STATE(2460),
    [sym_fsi_directive_decl] = STATE(2460),
    [sym_preproc_line] = STATE(2460),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_module] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_open] = ACTIONS(6813),
    [anon_sym_type] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_and] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_exception] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [anon_sym_extern] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6811),
    [anon_sym_POUNDload] = ACTIONS(6811),
    [anon_sym_POUNDtime] = ACTIONS(6811),
    [anon_sym_POUNDI] = ACTIONS(6811),
    [anon_sym_POUNDhelp] = ACTIONS(6811),
    [anon_sym_POUNDquit] = ACTIONS(6811),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [anon_sym_POUNDelse] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(2461)] = {
    [sym_block_comment] = STATE(2461),
    [sym_compiler_directive_decl] = STATE(2461),
    [sym_fsi_directive_decl] = STATE(2461),
    [sym_preproc_line] = STATE(2461),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2462)] = {
    [sym_block_comment] = STATE(2462),
    [sym_compiler_directive_decl] = STATE(2462),
    [sym_fsi_directive_decl] = STATE(2462),
    [sym_preproc_line] = STATE(2462),
    [aux_sym_rules_repeat1] = STATE(2464),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(6844),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(6849),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2463)] = {
    [sym_block_comment] = STATE(2463),
    [sym_compiler_directive_decl] = STATE(2463),
    [sym_fsi_directive_decl] = STATE(2463),
    [sym_preproc_line] = STATE(2463),
    [aux_sym_rules_repeat1] = STATE(2466),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(6844),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(6849),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2464)] = {
    [sym_block_comment] = STATE(2464),
    [sym_compiler_directive_decl] = STATE(2464),
    [sym_fsi_directive_decl] = STATE(2464),
    [sym_preproc_line] = STATE(2464),
    [aux_sym_rules_repeat1] = STATE(2466),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_module] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_open] = ACTIONS(6618),
    [anon_sym_type] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_and] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(6844),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_exception] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [anon_sym_extern] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6616),
    [anon_sym_POUNDload] = ACTIONS(6616),
    [anon_sym_POUNDtime] = ACTIONS(6616),
    [anon_sym_POUNDI] = ACTIONS(6616),
    [anon_sym_POUNDhelp] = ACTIONS(6616),
    [anon_sym_POUNDquit] = ACTIONS(6616),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(6852),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2465)] = {
    [sym_block_comment] = STATE(2465),
    [sym_compiler_directive_decl] = STATE(2465),
    [sym_fsi_directive_decl] = STATE(2465),
    [sym_preproc_line] = STATE(2465),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2466)] = {
    [sym_block_comment] = STATE(2466),
    [sym_compiler_directive_decl] = STATE(2466),
    [sym_fsi_directive_decl] = STATE(2466),
    [sym_preproc_line] = STATE(2466),
    [aux_sym_rules_repeat1] = STATE(2466),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6855),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6858),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2467)] = {
    [sym_block_comment] = STATE(2467),
    [sym_compiler_directive_decl] = STATE(2467),
    [sym_fsi_directive_decl] = STATE(2467),
    [sym_preproc_line] = STATE(2467),
    [ts_builtin_sym_end] = ACTIONS(6861),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_namespace] = ACTIONS(6863),
    [anon_sym_module] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_open] = ACTIONS(6863),
    [anon_sym_type] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_and] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_exception] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [anon_sym_extern] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6861),
    [anon_sym_POUNDload] = ACTIONS(6861),
    [anon_sym_POUNDtime] = ACTIONS(6861),
    [anon_sym_POUNDI] = ACTIONS(6861),
    [anon_sym_POUNDhelp] = ACTIONS(6861),
    [anon_sym_POUNDquit] = ACTIONS(6861),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(2468)] = {
    [sym_block_comment] = STATE(2468),
    [sym_compiler_directive_decl] = STATE(2468),
    [sym_fsi_directive_decl] = STATE(2468),
    [sym_preproc_line] = STATE(2468),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_module] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_open] = ACTIONS(6824),
    [anon_sym_type] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_and] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_exception] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6822),
    [anon_sym_POUNDload] = ACTIONS(6822),
    [anon_sym_POUNDtime] = ACTIONS(6822),
    [anon_sym_POUNDI] = ACTIONS(6822),
    [anon_sym_POUNDhelp] = ACTIONS(6822),
    [anon_sym_POUNDquit] = ACTIONS(6822),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [anon_sym_POUNDelse] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(2469)] = {
    [sym_block_comment] = STATE(2469),
    [sym_compiler_directive_decl] = STATE(2469),
    [sym_fsi_directive_decl] = STATE(2469),
    [sym_preproc_line] = STATE(2469),
    [ts_builtin_sym_end] = ACTIONS(6865),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_namespace] = ACTIONS(6867),
    [anon_sym_module] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_open] = ACTIONS(6867),
    [anon_sym_type] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_and] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_exception] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [anon_sym_extern] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6865),
    [anon_sym_POUNDload] = ACTIONS(6865),
    [anon_sym_POUNDtime] = ACTIONS(6865),
    [anon_sym_POUNDI] = ACTIONS(6865),
    [anon_sym_POUNDhelp] = ACTIONS(6865),
    [anon_sym_POUNDquit] = ACTIONS(6865),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(2470)] = {
    [sym_type_arguments] = STATE(2836),
    [sym_long_identifier] = STATE(2859),
    [sym_block_comment] = STATE(2470),
    [sym_compiler_directive_decl] = STATE(2470),
    [sym_fsi_directive_decl] = STATE(2470),
    [sym_preproc_line] = STATE(2470),
    [aux_sym_compound_type_repeat1] = STATE(2765),
    [sym_identifier] = ACTIONS(6774),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(6776),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(6778),
    [anon_sym_LT] = ACTIONS(6780),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6782),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__dedent] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2471)] = {
    [sym_block_comment] = STATE(2471),
    [sym_compiler_directive_decl] = STATE(2471),
    [sym_fsi_directive_decl] = STATE(2471),
    [sym_preproc_line] = STATE(2471),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_module] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_open] = ACTIONS(6560),
    [anon_sym_type] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_and] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_exception] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [anon_sym_extern] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6558),
    [anon_sym_POUNDload] = ACTIONS(6558),
    [anon_sym_POUNDtime] = ACTIONS(6558),
    [anon_sym_POUNDI] = ACTIONS(6558),
    [anon_sym_POUNDhelp] = ACTIONS(6558),
    [anon_sym_POUNDquit] = ACTIONS(6558),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(2472)] = {
    [sym_block_comment] = STATE(2472),
    [sym_compiler_directive_decl] = STATE(2472),
    [sym_fsi_directive_decl] = STATE(2472),
    [sym_preproc_line] = STATE(2472),
    [ts_builtin_sym_end] = ACTIONS(6869),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_namespace] = ACTIONS(6871),
    [anon_sym_module] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_open] = ACTIONS(6871),
    [anon_sym_type] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_and] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_exception] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [anon_sym_extern] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6869),
    [anon_sym_POUNDload] = ACTIONS(6869),
    [anon_sym_POUNDtime] = ACTIONS(6869),
    [anon_sym_POUNDI] = ACTIONS(6869),
    [anon_sym_POUNDhelp] = ACTIONS(6869),
    [anon_sym_POUNDquit] = ACTIONS(6869),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(2473)] = {
    [sym_block_comment] = STATE(2473),
    [sym_compiler_directive_decl] = STATE(2473),
    [sym_fsi_directive_decl] = STATE(2473),
    [sym_preproc_line] = STATE(2473),
    [ts_builtin_sym_end] = ACTIONS(6873),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_namespace] = ACTIONS(6875),
    [anon_sym_module] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_open] = ACTIONS(6875),
    [anon_sym_type] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_and] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_exception] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [anon_sym_extern] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6873),
    [anon_sym_POUNDload] = ACTIONS(6873),
    [anon_sym_POUNDtime] = ACTIONS(6873),
    [anon_sym_POUNDI] = ACTIONS(6873),
    [anon_sym_POUNDhelp] = ACTIONS(6873),
    [anon_sym_POUNDquit] = ACTIONS(6873),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(2474)] = {
    [sym_block_comment] = STATE(2474),
    [sym_compiler_directive_decl] = STATE(2474),
    [sym_fsi_directive_decl] = STATE(2474),
    [sym_preproc_line] = STATE(2474),
    [ts_builtin_sym_end] = ACTIONS(6830),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_namespace] = ACTIONS(6828),
    [anon_sym_module] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_open] = ACTIONS(6828),
    [anon_sym_type] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_and] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_exception] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6830),
    [anon_sym_POUNDload] = ACTIONS(6830),
    [anon_sym_POUNDtime] = ACTIONS(6830),
    [anon_sym_POUNDI] = ACTIONS(6830),
    [anon_sym_POUNDhelp] = ACTIONS(6830),
    [anon_sym_POUNDquit] = ACTIONS(6830),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(2475)] = {
    [sym_block_comment] = STATE(2475),
    [sym_compiler_directive_decl] = STATE(2475),
    [sym_fsi_directive_decl] = STATE(2475),
    [sym_preproc_line] = STATE(2475),
    [aux_sym_long_identifier_repeat1] = STATE(2431),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_module] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_open] = ACTIONS(6625),
    [anon_sym_type] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_and] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(6877),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_exception] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [anon_sym_extern] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6628),
    [anon_sym_POUNDload] = ACTIONS(6628),
    [anon_sym_POUNDtime] = ACTIONS(6628),
    [anon_sym_POUNDI] = ACTIONS(6628),
    [anon_sym_POUNDhelp] = ACTIONS(6628),
    [anon_sym_POUNDquit] = ACTIONS(6628),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(2476)] = {
    [sym_block_comment] = STATE(2476),
    [sym_compiler_directive_decl] = STATE(2476),
    [sym_fsi_directive_decl] = STATE(2476),
    [sym_preproc_line] = STATE(2476),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2477)] = {
    [sym_block_comment] = STATE(2477),
    [sym_compiler_directive_decl] = STATE(2477),
    [sym_fsi_directive_decl] = STATE(2477),
    [sym_preproc_line] = STATE(2477),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6881),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(2478)] = {
    [sym_block_comment] = STATE(2478),
    [sym_compiler_directive_decl] = STATE(2478),
    [sym_fsi_directive_decl] = STATE(2478),
    [sym_preproc_line] = STATE(2478),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(2479)] = {
    [sym_block_comment] = STATE(2479),
    [sym_compiler_directive_decl] = STATE(2479),
    [sym_fsi_directive_decl] = STATE(2479),
    [sym_preproc_line] = STATE(2479),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_module] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_open] = ACTIONS(6875),
    [anon_sym_type] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_and] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_exception] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [anon_sym_extern] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6873),
    [anon_sym_POUNDload] = ACTIONS(6873),
    [anon_sym_POUNDtime] = ACTIONS(6873),
    [anon_sym_POUNDI] = ACTIONS(6873),
    [anon_sym_POUNDhelp] = ACTIONS(6873),
    [anon_sym_POUNDquit] = ACTIONS(6873),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [anon_sym_POUNDelse] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(2480)] = {
    [sym_block_comment] = STATE(2480),
    [sym_compiler_directive_decl] = STATE(2480),
    [sym_fsi_directive_decl] = STATE(2480),
    [sym_preproc_line] = STATE(2480),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_module] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_open] = ACTIONS(6863),
    [anon_sym_type] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_and] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_exception] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [anon_sym_extern] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6861),
    [anon_sym_POUNDload] = ACTIONS(6861),
    [anon_sym_POUNDtime] = ACTIONS(6861),
    [anon_sym_POUNDI] = ACTIONS(6861),
    [anon_sym_POUNDhelp] = ACTIONS(6861),
    [anon_sym_POUNDquit] = ACTIONS(6861),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [anon_sym_POUNDelse] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(2481)] = {
    [sym_block_comment] = STATE(2481),
    [sym_compiler_directive_decl] = STATE(2481),
    [sym_fsi_directive_decl] = STATE(2481),
    [sym_preproc_line] = STATE(2481),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2482)] = {
    [sym_block_comment] = STATE(2482),
    [sym_compiler_directive_decl] = STATE(2482),
    [sym_fsi_directive_decl] = STATE(2482),
    [sym_preproc_line] = STATE(2482),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_module] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_open] = ACTIONS(6867),
    [anon_sym_type] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_and] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_exception] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [anon_sym_extern] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6865),
    [anon_sym_POUNDload] = ACTIONS(6865),
    [anon_sym_POUNDtime] = ACTIONS(6865),
    [anon_sym_POUNDI] = ACTIONS(6865),
    [anon_sym_POUNDhelp] = ACTIONS(6865),
    [anon_sym_POUNDquit] = ACTIONS(6865),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [anon_sym_POUNDelse] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(2483)] = {
    [sym_block_comment] = STATE(2483),
    [sym_compiler_directive_decl] = STATE(2483),
    [sym_fsi_directive_decl] = STATE(2483),
    [sym_preproc_line] = STATE(2483),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_module] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_open] = ACTIONS(6871),
    [anon_sym_type] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_and] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_exception] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [anon_sym_extern] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6869),
    [anon_sym_POUNDload] = ACTIONS(6869),
    [anon_sym_POUNDtime] = ACTIONS(6869),
    [anon_sym_POUNDI] = ACTIONS(6869),
    [anon_sym_POUNDhelp] = ACTIONS(6869),
    [anon_sym_POUNDquit] = ACTIONS(6869),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [anon_sym_POUNDelse] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(2484)] = {
    [sym_block_comment] = STATE(2484),
    [sym_compiler_directive_decl] = STATE(2484),
    [sym_fsi_directive_decl] = STATE(2484),
    [sym_preproc_line] = STATE(2484),
    [aux_sym_sequential_expression_repeat1] = STATE(2484),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [anon_sym_POUNDelse] = ACTIONS(6798),
    [sym__newline] = ACTIONS(6883),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(2485)] = {
    [sym_block_comment] = STATE(2485),
    [sym_compiler_directive_decl] = STATE(2485),
    [sym_fsi_directive_decl] = STATE(2485),
    [sym_preproc_line] = STATE(2485),
    [ts_builtin_sym_end] = ACTIONS(6886),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_namespace] = ACTIONS(6888),
    [anon_sym_module] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_open] = ACTIONS(6888),
    [anon_sym_type] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_and] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_exception] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [anon_sym_extern] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6886),
    [anon_sym_POUNDload] = ACTIONS(6886),
    [anon_sym_POUNDtime] = ACTIONS(6886),
    [anon_sym_POUNDI] = ACTIONS(6886),
    [anon_sym_POUNDhelp] = ACTIONS(6886),
    [anon_sym_POUNDquit] = ACTIONS(6886),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(2486)] = {
    [sym_block_comment] = STATE(2486),
    [sym_compiler_directive_decl] = STATE(2486),
    [sym_fsi_directive_decl] = STATE(2486),
    [sym_preproc_line] = STATE(2486),
    [aux_sym_compound_type_repeat1] = STATE(2494),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6578),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [anon_sym_inherit] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [anon_sym_POUNDelse] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2487)] = {
    [sym_block_comment] = STATE(2487),
    [sym_compiler_directive_decl] = STATE(2487),
    [sym_fsi_directive_decl] = STATE(2487),
    [sym_preproc_line] = STATE(2487),
    [ts_builtin_sym_end] = ACTIONS(6805),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_namespace] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(2488)] = {
    [sym_block_comment] = STATE(2488),
    [sym_compiler_directive_decl] = STATE(2488),
    [sym_fsi_directive_decl] = STATE(2488),
    [sym_preproc_line] = STATE(2488),
    [ts_builtin_sym_end] = ACTIONS(6890),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_namespace] = ACTIONS(6892),
    [anon_sym_module] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_open] = ACTIONS(6892),
    [anon_sym_type] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_and] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_exception] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [anon_sym_extern] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6890),
    [anon_sym_POUNDload] = ACTIONS(6890),
    [anon_sym_POUNDtime] = ACTIONS(6890),
    [anon_sym_POUNDI] = ACTIONS(6890),
    [anon_sym_POUNDhelp] = ACTIONS(6890),
    [anon_sym_POUNDquit] = ACTIONS(6890),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(2489)] = {
    [sym_block_comment] = STATE(2489),
    [sym_compiler_directive_decl] = STATE(2489),
    [sym_fsi_directive_decl] = STATE(2489),
    [sym_preproc_line] = STATE(2489),
    [ts_builtin_sym_end] = ACTIONS(6894),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_namespace] = ACTIONS(6896),
    [anon_sym_module] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_open] = ACTIONS(6896),
    [anon_sym_type] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_and] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_exception] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [anon_sym_extern] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6894),
    [anon_sym_POUNDload] = ACTIONS(6894),
    [anon_sym_POUNDtime] = ACTIONS(6894),
    [anon_sym_POUNDI] = ACTIONS(6894),
    [anon_sym_POUNDhelp] = ACTIONS(6894),
    [anon_sym_POUNDquit] = ACTIONS(6894),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(2490)] = {
    [sym_block_comment] = STATE(2490),
    [sym_compiler_directive_decl] = STATE(2490),
    [sym_fsi_directive_decl] = STATE(2490),
    [sym_preproc_line] = STATE(2490),
    [ts_builtin_sym_end] = ACTIONS(6898),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_namespace] = ACTIONS(6900),
    [anon_sym_module] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_open] = ACTIONS(6900),
    [anon_sym_type] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_and] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_exception] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [anon_sym_extern] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6898),
    [anon_sym_POUNDload] = ACTIONS(6898),
    [anon_sym_POUNDtime] = ACTIONS(6898),
    [anon_sym_POUNDI] = ACTIONS(6898),
    [anon_sym_POUNDhelp] = ACTIONS(6898),
    [anon_sym_POUNDquit] = ACTIONS(6898),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(2491)] = {
    [sym_block_comment] = STATE(2491),
    [sym_compiler_directive_decl] = STATE(2491),
    [sym_fsi_directive_decl] = STATE(2491),
    [sym_preproc_line] = STATE(2491),
    [ts_builtin_sym_end] = ACTIONS(6603),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_namespace] = ACTIONS(6902),
    [anon_sym_module] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_open] = ACTIONS(6902),
    [anon_sym_type] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_and] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_exception] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [anon_sym_extern] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6603),
    [anon_sym_POUNDload] = ACTIONS(6603),
    [anon_sym_POUNDtime] = ACTIONS(6603),
    [anon_sym_POUNDI] = ACTIONS(6603),
    [anon_sym_POUNDhelp] = ACTIONS(6603),
    [anon_sym_POUNDquit] = ACTIONS(6603),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(2492)] = {
    [sym_block_comment] = STATE(2492),
    [sym_compiler_directive_decl] = STATE(2492),
    [sym_fsi_directive_decl] = STATE(2492),
    [sym_preproc_line] = STATE(2492),
    [ts_builtin_sym_end] = ACTIONS(6904),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_namespace] = ACTIONS(6906),
    [anon_sym_module] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_open] = ACTIONS(6906),
    [anon_sym_type] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_and] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_exception] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [anon_sym_extern] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6904),
    [anon_sym_POUNDload] = ACTIONS(6904),
    [anon_sym_POUNDtime] = ACTIONS(6904),
    [anon_sym_POUNDI] = ACTIONS(6904),
    [anon_sym_POUNDhelp] = ACTIONS(6904),
    [anon_sym_POUNDquit] = ACTIONS(6904),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(2493)] = {
    [sym_block_comment] = STATE(2493),
    [sym_compiler_directive_decl] = STATE(2493),
    [sym_fsi_directive_decl] = STATE(2493),
    [sym_preproc_line] = STATE(2493),
    [aux_sym_sequential_expression_repeat1] = STATE(2662),
    [ts_builtin_sym_end] = ACTIONS(6738),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(2494)] = {
    [sym_block_comment] = STATE(2494),
    [sym_compiler_directive_decl] = STATE(2494),
    [sym_fsi_directive_decl] = STATE(2494),
    [sym_preproc_line] = STATE(2494),
    [aux_sym_compound_type_repeat1] = STATE(2494),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6908),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_inherit] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2495)] = {
    [sym_block_comment] = STATE(2495),
    [sym_compiler_directive_decl] = STATE(2495),
    [sym_fsi_directive_decl] = STATE(2495),
    [sym_preproc_line] = STATE(2495),
    [ts_builtin_sym_end] = ACTIONS(6911),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_namespace] = ACTIONS(6913),
    [anon_sym_module] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_open] = ACTIONS(6913),
    [anon_sym_type] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_and] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_exception] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [anon_sym_extern] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6911),
    [anon_sym_POUNDload] = ACTIONS(6911),
    [anon_sym_POUNDtime] = ACTIONS(6911),
    [anon_sym_POUNDI] = ACTIONS(6911),
    [anon_sym_POUNDhelp] = ACTIONS(6911),
    [anon_sym_POUNDquit] = ACTIONS(6911),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(2496)] = {
    [sym_block_comment] = STATE(2496),
    [sym_compiler_directive_decl] = STATE(2496),
    [sym_fsi_directive_decl] = STATE(2496),
    [sym_preproc_line] = STATE(2496),
    [ts_builtin_sym_end] = ACTIONS(6915),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_namespace] = ACTIONS(6917),
    [anon_sym_module] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_open] = ACTIONS(6917),
    [anon_sym_type] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_and] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_exception] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [anon_sym_extern] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6915),
    [anon_sym_POUNDload] = ACTIONS(6915),
    [anon_sym_POUNDtime] = ACTIONS(6915),
    [anon_sym_POUNDI] = ACTIONS(6915),
    [anon_sym_POUNDhelp] = ACTIONS(6915),
    [anon_sym_POUNDquit] = ACTIONS(6915),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(2497)] = {
    [sym_block_comment] = STATE(2497),
    [sym_compiler_directive_decl] = STATE(2497),
    [sym_fsi_directive_decl] = STATE(2497),
    [sym_preproc_line] = STATE(2497),
    [ts_builtin_sym_end] = ACTIONS(6919),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_namespace] = ACTIONS(6921),
    [anon_sym_module] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_open] = ACTIONS(6921),
    [anon_sym_type] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_and] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_exception] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [anon_sym_extern] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6919),
    [anon_sym_POUNDload] = ACTIONS(6919),
    [anon_sym_POUNDtime] = ACTIONS(6919),
    [anon_sym_POUNDI] = ACTIONS(6919),
    [anon_sym_POUNDhelp] = ACTIONS(6919),
    [anon_sym_POUNDquit] = ACTIONS(6919),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(2498)] = {
    [sym_block_comment] = STATE(2498),
    [sym_compiler_directive_decl] = STATE(2498),
    [sym_fsi_directive_decl] = STATE(2498),
    [sym_preproc_line] = STATE(2498),
    [ts_builtin_sym_end] = ACTIONS(6923),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_namespace] = ACTIONS(6925),
    [anon_sym_module] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_open] = ACTIONS(6925),
    [anon_sym_type] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_and] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_exception] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [anon_sym_extern] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6923),
    [anon_sym_POUNDload] = ACTIONS(6923),
    [anon_sym_POUNDtime] = ACTIONS(6923),
    [anon_sym_POUNDI] = ACTIONS(6923),
    [anon_sym_POUNDhelp] = ACTIONS(6923),
    [anon_sym_POUNDquit] = ACTIONS(6923),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(2499)] = {
    [sym_block_comment] = STATE(2499),
    [sym_compiler_directive_decl] = STATE(2499),
    [sym_fsi_directive_decl] = STATE(2499),
    [sym_preproc_line] = STATE(2499),
    [ts_builtin_sym_end] = ACTIONS(292),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_namespace] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(2500)] = {
    [sym_block_comment] = STATE(2500),
    [sym_compiler_directive_decl] = STATE(2500),
    [sym_fsi_directive_decl] = STATE(2500),
    [sym_preproc_line] = STATE(2500),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(2501)] = {
    [sym_block_comment] = STATE(2501),
    [sym_compiler_directive_decl] = STATE(2501),
    [sym_fsi_directive_decl] = STATE(2501),
    [sym_preproc_line] = STATE(2501),
    [aux_sym_type_argument_repeat1] = STATE(2586),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6927),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2502)] = {
    [sym_block_comment] = STATE(2502),
    [sym_compiler_directive_decl] = STATE(2502),
    [sym_fsi_directive_decl] = STATE(2502),
    [sym_preproc_line] = STATE(2502),
    [ts_builtin_sym_end] = ACTIONS(6930),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_namespace] = ACTIONS(6932),
    [anon_sym_module] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_open] = ACTIONS(6932),
    [anon_sym_type] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_and] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_exception] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [anon_sym_extern] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6930),
    [anon_sym_POUNDload] = ACTIONS(6930),
    [anon_sym_POUNDtime] = ACTIONS(6930),
    [anon_sym_POUNDI] = ACTIONS(6930),
    [anon_sym_POUNDhelp] = ACTIONS(6930),
    [anon_sym_POUNDquit] = ACTIONS(6930),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(2503)] = {
    [sym_block_comment] = STATE(2503),
    [sym_compiler_directive_decl] = STATE(2503),
    [sym_fsi_directive_decl] = STATE(2503),
    [sym_preproc_line] = STATE(2503),
    [ts_builtin_sym_end] = ACTIONS(6934),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_namespace] = ACTIONS(6936),
    [anon_sym_module] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_open] = ACTIONS(6936),
    [anon_sym_type] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_and] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_exception] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [anon_sym_extern] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6934),
    [anon_sym_POUNDload] = ACTIONS(6934),
    [anon_sym_POUNDtime] = ACTIONS(6934),
    [anon_sym_POUNDI] = ACTIONS(6934),
    [anon_sym_POUNDhelp] = ACTIONS(6934),
    [anon_sym_POUNDquit] = ACTIONS(6934),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(2504)] = {
    [sym_block_comment] = STATE(2504),
    [sym_compiler_directive_decl] = STATE(2504),
    [sym_fsi_directive_decl] = STATE(2504),
    [sym_preproc_line] = STATE(2504),
    [ts_builtin_sym_end] = ACTIONS(6938),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_namespace] = ACTIONS(6940),
    [anon_sym_module] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_open] = ACTIONS(6940),
    [anon_sym_type] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_and] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_exception] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [anon_sym_extern] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6938),
    [anon_sym_POUNDload] = ACTIONS(6938),
    [anon_sym_POUNDtime] = ACTIONS(6938),
    [anon_sym_POUNDI] = ACTIONS(6938),
    [anon_sym_POUNDhelp] = ACTIONS(6938),
    [anon_sym_POUNDquit] = ACTIONS(6938),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(2505)] = {
    [sym_block_comment] = STATE(2505),
    [sym_compiler_directive_decl] = STATE(2505),
    [sym_fsi_directive_decl] = STATE(2505),
    [sym_preproc_line] = STATE(2505),
    [aux_sym_compound_type_repeat1] = STATE(2509),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6568),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [anon_sym_POUNDelse] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2506)] = {
    [sym_block_comment] = STATE(2506),
    [sym_compiler_directive_decl] = STATE(2506),
    [sym_fsi_directive_decl] = STATE(2506),
    [sym_preproc_line] = STATE(2506),
    [ts_builtin_sym_end] = ACTIONS(6942),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_namespace] = ACTIONS(6944),
    [anon_sym_module] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_open] = ACTIONS(6944),
    [anon_sym_type] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_and] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_exception] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [anon_sym_extern] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6942),
    [anon_sym_POUNDload] = ACTIONS(6942),
    [anon_sym_POUNDtime] = ACTIONS(6942),
    [anon_sym_POUNDI] = ACTIONS(6942),
    [anon_sym_POUNDhelp] = ACTIONS(6942),
    [anon_sym_POUNDquit] = ACTIONS(6942),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(2507)] = {
    [sym_block_comment] = STATE(2507),
    [sym_compiler_directive_decl] = STATE(2507),
    [sym_fsi_directive_decl] = STATE(2507),
    [sym_preproc_line] = STATE(2507),
    [ts_builtin_sym_end] = ACTIONS(6946),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_namespace] = ACTIONS(6948),
    [anon_sym_module] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_open] = ACTIONS(6948),
    [anon_sym_type] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_and] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_exception] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [anon_sym_extern] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6946),
    [anon_sym_POUNDload] = ACTIONS(6946),
    [anon_sym_POUNDtime] = ACTIONS(6946),
    [anon_sym_POUNDI] = ACTIONS(6946),
    [anon_sym_POUNDhelp] = ACTIONS(6946),
    [anon_sym_POUNDquit] = ACTIONS(6946),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(2508)] = {
    [sym_block_comment] = STATE(2508),
    [sym_compiler_directive_decl] = STATE(2508),
    [sym_fsi_directive_decl] = STATE(2508),
    [sym_preproc_line] = STATE(2508),
    [ts_builtin_sym_end] = ACTIONS(6950),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_namespace] = ACTIONS(6952),
    [anon_sym_module] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_open] = ACTIONS(6952),
    [anon_sym_type] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_and] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_exception] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [anon_sym_extern] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6950),
    [anon_sym_POUNDload] = ACTIONS(6950),
    [anon_sym_POUNDtime] = ACTIONS(6950),
    [anon_sym_POUNDI] = ACTIONS(6950),
    [anon_sym_POUNDhelp] = ACTIONS(6950),
    [anon_sym_POUNDquit] = ACTIONS(6950),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(2509)] = {
    [sym_block_comment] = STATE(2509),
    [sym_compiler_directive_decl] = STATE(2509),
    [sym_fsi_directive_decl] = STATE(2509),
    [sym_preproc_line] = STATE(2509),
    [aux_sym_compound_type_repeat1] = STATE(2509),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(6954),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2510)] = {
    [sym_block_comment] = STATE(2510),
    [sym_compiler_directive_decl] = STATE(2510),
    [sym_fsi_directive_decl] = STATE(2510),
    [sym_preproc_line] = STATE(2510),
    [ts_builtin_sym_end] = ACTIONS(6957),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_namespace] = ACTIONS(6959),
    [anon_sym_module] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_open] = ACTIONS(6959),
    [anon_sym_type] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_and] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_exception] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [anon_sym_extern] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6957),
    [anon_sym_POUNDload] = ACTIONS(6957),
    [anon_sym_POUNDtime] = ACTIONS(6957),
    [anon_sym_POUNDI] = ACTIONS(6957),
    [anon_sym_POUNDhelp] = ACTIONS(6957),
    [anon_sym_POUNDquit] = ACTIONS(6957),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(2511)] = {
    [sym_block_comment] = STATE(2511),
    [sym_compiler_directive_decl] = STATE(2511),
    [sym_fsi_directive_decl] = STATE(2511),
    [sym_preproc_line] = STATE(2511),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_module] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_open] = ACTIONS(6892),
    [anon_sym_type] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_and] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_exception] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [anon_sym_extern] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6890),
    [anon_sym_POUNDload] = ACTIONS(6890),
    [anon_sym_POUNDtime] = ACTIONS(6890),
    [anon_sym_POUNDI] = ACTIONS(6890),
    [anon_sym_POUNDhelp] = ACTIONS(6890),
    [anon_sym_POUNDquit] = ACTIONS(6890),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [anon_sym_POUNDelse] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(2512)] = {
    [sym_block_comment] = STATE(2512),
    [sym_compiler_directive_decl] = STATE(2512),
    [sym_fsi_directive_decl] = STATE(2512),
    [sym_preproc_line] = STATE(2512),
    [ts_builtin_sym_end] = ACTIONS(6961),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_namespace] = ACTIONS(6963),
    [anon_sym_module] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_open] = ACTIONS(6963),
    [anon_sym_type] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_and] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_exception] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [anon_sym_extern] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6961),
    [anon_sym_POUNDload] = ACTIONS(6961),
    [anon_sym_POUNDtime] = ACTIONS(6961),
    [anon_sym_POUNDI] = ACTIONS(6961),
    [anon_sym_POUNDhelp] = ACTIONS(6961),
    [anon_sym_POUNDquit] = ACTIONS(6961),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(2513)] = {
    [sym_block_comment] = STATE(2513),
    [sym_compiler_directive_decl] = STATE(2513),
    [sym_fsi_directive_decl] = STATE(2513),
    [sym_preproc_line] = STATE(2513),
    [ts_builtin_sym_end] = ACTIONS(6965),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_namespace] = ACTIONS(6967),
    [anon_sym_module] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_open] = ACTIONS(6967),
    [anon_sym_type] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_and] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_exception] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [anon_sym_extern] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6965),
    [anon_sym_POUNDload] = ACTIONS(6965),
    [anon_sym_POUNDtime] = ACTIONS(6965),
    [anon_sym_POUNDI] = ACTIONS(6965),
    [anon_sym_POUNDhelp] = ACTIONS(6965),
    [anon_sym_POUNDquit] = ACTIONS(6965),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(2514)] = {
    [sym_block_comment] = STATE(2514),
    [sym_compiler_directive_decl] = STATE(2514),
    [sym_fsi_directive_decl] = STATE(2514),
    [sym_preproc_line] = STATE(2514),
    [ts_builtin_sym_end] = ACTIONS(6969),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_namespace] = ACTIONS(6971),
    [anon_sym_module] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_open] = ACTIONS(6971),
    [anon_sym_type] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_and] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_exception] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [anon_sym_extern] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6969),
    [anon_sym_POUNDload] = ACTIONS(6969),
    [anon_sym_POUNDtime] = ACTIONS(6969),
    [anon_sym_POUNDI] = ACTIONS(6969),
    [anon_sym_POUNDhelp] = ACTIONS(6969),
    [anon_sym_POUNDquit] = ACTIONS(6969),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(2515)] = {
    [sym_block_comment] = STATE(2515),
    [sym_compiler_directive_decl] = STATE(2515),
    [sym_fsi_directive_decl] = STATE(2515),
    [sym_preproc_line] = STATE(2515),
    [ts_builtin_sym_end] = ACTIONS(6973),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_namespace] = ACTIONS(6975),
    [anon_sym_module] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_open] = ACTIONS(6975),
    [anon_sym_type] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_and] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_exception] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [anon_sym_extern] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6973),
    [anon_sym_POUNDload] = ACTIONS(6973),
    [anon_sym_POUNDtime] = ACTIONS(6973),
    [anon_sym_POUNDI] = ACTIONS(6973),
    [anon_sym_POUNDhelp] = ACTIONS(6973),
    [anon_sym_POUNDquit] = ACTIONS(6973),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(2516)] = {
    [sym_block_comment] = STATE(2516),
    [sym_compiler_directive_decl] = STATE(2516),
    [sym_fsi_directive_decl] = STATE(2516),
    [sym_preproc_line] = STATE(2516),
    [ts_builtin_sym_end] = ACTIONS(6977),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_namespace] = ACTIONS(6979),
    [anon_sym_module] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_open] = ACTIONS(6979),
    [anon_sym_type] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_and] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_exception] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [anon_sym_extern] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6977),
    [anon_sym_POUNDload] = ACTIONS(6977),
    [anon_sym_POUNDtime] = ACTIONS(6977),
    [anon_sym_POUNDI] = ACTIONS(6977),
    [anon_sym_POUNDhelp] = ACTIONS(6977),
    [anon_sym_POUNDquit] = ACTIONS(6977),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(2517)] = {
    [sym_block_comment] = STATE(2517),
    [sym_compiler_directive_decl] = STATE(2517),
    [sym_fsi_directive_decl] = STATE(2517),
    [sym_preproc_line] = STATE(2517),
    [ts_builtin_sym_end] = ACTIONS(6981),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_namespace] = ACTIONS(6983),
    [anon_sym_module] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_open] = ACTIONS(6983),
    [anon_sym_type] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_and] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_exception] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [anon_sym_extern] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6981),
    [anon_sym_POUNDload] = ACTIONS(6981),
    [anon_sym_POUNDtime] = ACTIONS(6981),
    [anon_sym_POUNDI] = ACTIONS(6981),
    [anon_sym_POUNDhelp] = ACTIONS(6981),
    [anon_sym_POUNDquit] = ACTIONS(6981),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(2518)] = {
    [sym_block_comment] = STATE(2518),
    [sym_compiler_directive_decl] = STATE(2518),
    [sym_fsi_directive_decl] = STATE(2518),
    [sym_preproc_line] = STATE(2518),
    [ts_builtin_sym_end] = ACTIONS(6985),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_namespace] = ACTIONS(6987),
    [anon_sym_module] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_open] = ACTIONS(6987),
    [anon_sym_type] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_and] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_exception] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [anon_sym_extern] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6985),
    [anon_sym_POUNDload] = ACTIONS(6985),
    [anon_sym_POUNDtime] = ACTIONS(6985),
    [anon_sym_POUNDI] = ACTIONS(6985),
    [anon_sym_POUNDhelp] = ACTIONS(6985),
    [anon_sym_POUNDquit] = ACTIONS(6985),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(2519)] = {
    [sym_block_comment] = STATE(2519),
    [sym_compiler_directive_decl] = STATE(2519),
    [sym_fsi_directive_decl] = STATE(2519),
    [sym_preproc_line] = STATE(2519),
    [ts_builtin_sym_end] = ACTIONS(6989),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_namespace] = ACTIONS(6991),
    [anon_sym_module] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_open] = ACTIONS(6991),
    [anon_sym_type] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_and] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_exception] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [anon_sym_extern] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6989),
    [anon_sym_POUNDload] = ACTIONS(6989),
    [anon_sym_POUNDtime] = ACTIONS(6989),
    [anon_sym_POUNDI] = ACTIONS(6989),
    [anon_sym_POUNDhelp] = ACTIONS(6989),
    [anon_sym_POUNDquit] = ACTIONS(6989),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(2520)] = {
    [sym_block_comment] = STATE(2520),
    [sym_compiler_directive_decl] = STATE(2520),
    [sym_fsi_directive_decl] = STATE(2520),
    [sym_preproc_line] = STATE(2520),
    [ts_builtin_sym_end] = ACTIONS(6993),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_namespace] = ACTIONS(6995),
    [anon_sym_module] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_open] = ACTIONS(6995),
    [anon_sym_type] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_and] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_exception] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [anon_sym_extern] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6993),
    [anon_sym_POUNDload] = ACTIONS(6993),
    [anon_sym_POUNDtime] = ACTIONS(6993),
    [anon_sym_POUNDI] = ACTIONS(6993),
    [anon_sym_POUNDhelp] = ACTIONS(6993),
    [anon_sym_POUNDquit] = ACTIONS(6993),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(2521)] = {
    [sym_block_comment] = STATE(2521),
    [sym_compiler_directive_decl] = STATE(2521),
    [sym_fsi_directive_decl] = STATE(2521),
    [sym_preproc_line] = STATE(2521),
    [ts_builtin_sym_end] = ACTIONS(6997),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_namespace] = ACTIONS(6999),
    [anon_sym_module] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_open] = ACTIONS(6999),
    [anon_sym_type] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_and] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_exception] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [anon_sym_extern] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6997),
    [anon_sym_POUNDload] = ACTIONS(6997),
    [anon_sym_POUNDtime] = ACTIONS(6997),
    [anon_sym_POUNDI] = ACTIONS(6997),
    [anon_sym_POUNDhelp] = ACTIONS(6997),
    [anon_sym_POUNDquit] = ACTIONS(6997),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(2522)] = {
    [sym_block_comment] = STATE(2522),
    [sym_compiler_directive_decl] = STATE(2522),
    [sym_fsi_directive_decl] = STATE(2522),
    [sym_preproc_line] = STATE(2522),
    [ts_builtin_sym_end] = ACTIONS(7001),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_namespace] = ACTIONS(7003),
    [anon_sym_module] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_open] = ACTIONS(7003),
    [anon_sym_type] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_and] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_exception] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [anon_sym_extern] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7001),
    [anon_sym_POUNDload] = ACTIONS(7001),
    [anon_sym_POUNDtime] = ACTIONS(7001),
    [anon_sym_POUNDI] = ACTIONS(7001),
    [anon_sym_POUNDhelp] = ACTIONS(7001),
    [anon_sym_POUNDquit] = ACTIONS(7001),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(2523)] = {
    [sym_block_comment] = STATE(2523),
    [sym_compiler_directive_decl] = STATE(2523),
    [sym_fsi_directive_decl] = STATE(2523),
    [sym_preproc_line] = STATE(2523),
    [ts_builtin_sym_end] = ACTIONS(7005),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_namespace] = ACTIONS(7007),
    [anon_sym_module] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_open] = ACTIONS(7007),
    [anon_sym_type] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_and] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_exception] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [anon_sym_extern] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7005),
    [anon_sym_POUNDload] = ACTIONS(7005),
    [anon_sym_POUNDtime] = ACTIONS(7005),
    [anon_sym_POUNDI] = ACTIONS(7005),
    [anon_sym_POUNDhelp] = ACTIONS(7005),
    [anon_sym_POUNDquit] = ACTIONS(7005),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(2524)] = {
    [sym_block_comment] = STATE(2524),
    [sym_compiler_directive_decl] = STATE(2524),
    [sym_fsi_directive_decl] = STATE(2524),
    [sym_preproc_line] = STATE(2524),
    [ts_builtin_sym_end] = ACTIONS(7009),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_namespace] = ACTIONS(7011),
    [anon_sym_module] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_open] = ACTIONS(7011),
    [anon_sym_type] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_and] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_exception] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [anon_sym_extern] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7009),
    [anon_sym_POUNDload] = ACTIONS(7009),
    [anon_sym_POUNDtime] = ACTIONS(7009),
    [anon_sym_POUNDI] = ACTIONS(7009),
    [anon_sym_POUNDhelp] = ACTIONS(7009),
    [anon_sym_POUNDquit] = ACTIONS(7009),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(2525)] = {
    [sym_block_comment] = STATE(2525),
    [sym_compiler_directive_decl] = STATE(2525),
    [sym_fsi_directive_decl] = STATE(2525),
    [sym_preproc_line] = STATE(2525),
    [ts_builtin_sym_end] = ACTIONS(7013),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_namespace] = ACTIONS(7015),
    [anon_sym_module] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_open] = ACTIONS(7015),
    [anon_sym_type] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_and] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_exception] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [anon_sym_extern] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7013),
    [anon_sym_POUNDload] = ACTIONS(7013),
    [anon_sym_POUNDtime] = ACTIONS(7013),
    [anon_sym_POUNDI] = ACTIONS(7013),
    [anon_sym_POUNDhelp] = ACTIONS(7013),
    [anon_sym_POUNDquit] = ACTIONS(7013),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(2526)] = {
    [sym_block_comment] = STATE(2526),
    [sym_compiler_directive_decl] = STATE(2526),
    [sym_fsi_directive_decl] = STATE(2526),
    [sym_preproc_line] = STATE(2526),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_module] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_open] = ACTIONS(6896),
    [anon_sym_type] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_and] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_exception] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [anon_sym_extern] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6894),
    [anon_sym_POUNDload] = ACTIONS(6894),
    [anon_sym_POUNDtime] = ACTIONS(6894),
    [anon_sym_POUNDI] = ACTIONS(6894),
    [anon_sym_POUNDhelp] = ACTIONS(6894),
    [anon_sym_POUNDquit] = ACTIONS(6894),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [anon_sym_POUNDelse] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(2527)] = {
    [sym_block_comment] = STATE(2527),
    [sym_compiler_directive_decl] = STATE(2527),
    [sym_fsi_directive_decl] = STATE(2527),
    [sym_preproc_line] = STATE(2527),
    [ts_builtin_sym_end] = ACTIONS(7017),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_namespace] = ACTIONS(7019),
    [anon_sym_module] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_open] = ACTIONS(7019),
    [anon_sym_type] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_and] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_exception] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [anon_sym_extern] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7017),
    [anon_sym_POUNDload] = ACTIONS(7017),
    [anon_sym_POUNDtime] = ACTIONS(7017),
    [anon_sym_POUNDI] = ACTIONS(7017),
    [anon_sym_POUNDhelp] = ACTIONS(7017),
    [anon_sym_POUNDquit] = ACTIONS(7017),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(2528)] = {
    [sym_block_comment] = STATE(2528),
    [sym_compiler_directive_decl] = STATE(2528),
    [sym_fsi_directive_decl] = STATE(2528),
    [sym_preproc_line] = STATE(2528),
    [ts_builtin_sym_end] = ACTIONS(7021),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_namespace] = ACTIONS(7023),
    [anon_sym_module] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_open] = ACTIONS(7023),
    [anon_sym_type] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_and] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_exception] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [anon_sym_extern] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7021),
    [anon_sym_POUNDload] = ACTIONS(7021),
    [anon_sym_POUNDtime] = ACTIONS(7021),
    [anon_sym_POUNDI] = ACTIONS(7021),
    [anon_sym_POUNDhelp] = ACTIONS(7021),
    [anon_sym_POUNDquit] = ACTIONS(7021),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(2529)] = {
    [sym_block_comment] = STATE(2529),
    [sym_compiler_directive_decl] = STATE(2529),
    [sym_fsi_directive_decl] = STATE(2529),
    [sym_preproc_line] = STATE(2529),
    [ts_builtin_sym_end] = ACTIONS(7025),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_namespace] = ACTIONS(7027),
    [anon_sym_module] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_open] = ACTIONS(7027),
    [anon_sym_type] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_and] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_exception] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [anon_sym_extern] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7025),
    [anon_sym_POUNDload] = ACTIONS(7025),
    [anon_sym_POUNDtime] = ACTIONS(7025),
    [anon_sym_POUNDI] = ACTIONS(7025),
    [anon_sym_POUNDhelp] = ACTIONS(7025),
    [anon_sym_POUNDquit] = ACTIONS(7025),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(2530)] = {
    [sym_block_comment] = STATE(2530),
    [sym_compiler_directive_decl] = STATE(2530),
    [sym_fsi_directive_decl] = STATE(2530),
    [sym_preproc_line] = STATE(2530),
    [ts_builtin_sym_end] = ACTIONS(7029),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_namespace] = ACTIONS(7031),
    [anon_sym_module] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_open] = ACTIONS(7031),
    [anon_sym_type] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_and] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_exception] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [anon_sym_extern] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7029),
    [anon_sym_POUNDload] = ACTIONS(7029),
    [anon_sym_POUNDtime] = ACTIONS(7029),
    [anon_sym_POUNDI] = ACTIONS(7029),
    [anon_sym_POUNDhelp] = ACTIONS(7029),
    [anon_sym_POUNDquit] = ACTIONS(7029),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(2531)] = {
    [sym_block_comment] = STATE(2531),
    [sym_compiler_directive_decl] = STATE(2531),
    [sym_fsi_directive_decl] = STATE(2531),
    [sym_preproc_line] = STATE(2531),
    [ts_builtin_sym_end] = ACTIONS(7033),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_namespace] = ACTIONS(7035),
    [anon_sym_module] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_open] = ACTIONS(7035),
    [anon_sym_type] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_and] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_exception] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [anon_sym_extern] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7033),
    [anon_sym_POUNDload] = ACTIONS(7033),
    [anon_sym_POUNDtime] = ACTIONS(7033),
    [anon_sym_POUNDI] = ACTIONS(7033),
    [anon_sym_POUNDhelp] = ACTIONS(7033),
    [anon_sym_POUNDquit] = ACTIONS(7033),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(2532)] = {
    [sym_block_comment] = STATE(2532),
    [sym_compiler_directive_decl] = STATE(2532),
    [sym_fsi_directive_decl] = STATE(2532),
    [sym_preproc_line] = STATE(2532),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_module] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_open] = ACTIONS(6900),
    [anon_sym_type] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_and] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_exception] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [anon_sym_extern] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6898),
    [anon_sym_POUNDload] = ACTIONS(6898),
    [anon_sym_POUNDtime] = ACTIONS(6898),
    [anon_sym_POUNDI] = ACTIONS(6898),
    [anon_sym_POUNDhelp] = ACTIONS(6898),
    [anon_sym_POUNDquit] = ACTIONS(6898),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [anon_sym_POUNDelse] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(2533)] = {
    [sym_block_comment] = STATE(2533),
    [sym_compiler_directive_decl] = STATE(2533),
    [sym_fsi_directive_decl] = STATE(2533),
    [sym_preproc_line] = STATE(2533),
    [ts_builtin_sym_end] = ACTIONS(7037),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_namespace] = ACTIONS(7039),
    [anon_sym_module] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_open] = ACTIONS(7039),
    [anon_sym_type] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_and] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_exception] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [anon_sym_extern] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7037),
    [anon_sym_POUNDload] = ACTIONS(7037),
    [anon_sym_POUNDtime] = ACTIONS(7037),
    [anon_sym_POUNDI] = ACTIONS(7037),
    [anon_sym_POUNDhelp] = ACTIONS(7037),
    [anon_sym_POUNDquit] = ACTIONS(7037),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(2534)] = {
    [sym_block_comment] = STATE(2534),
    [sym_compiler_directive_decl] = STATE(2534),
    [sym_fsi_directive_decl] = STATE(2534),
    [sym_preproc_line] = STATE(2534),
    [ts_builtin_sym_end] = ACTIONS(7041),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_namespace] = ACTIONS(7043),
    [anon_sym_module] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_open] = ACTIONS(7043),
    [anon_sym_type] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_and] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_exception] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [anon_sym_extern] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7041),
    [anon_sym_POUNDload] = ACTIONS(7041),
    [anon_sym_POUNDtime] = ACTIONS(7041),
    [anon_sym_POUNDI] = ACTIONS(7041),
    [anon_sym_POUNDhelp] = ACTIONS(7041),
    [anon_sym_POUNDquit] = ACTIONS(7041),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(2535)] = {
    [sym_block_comment] = STATE(2535),
    [sym_compiler_directive_decl] = STATE(2535),
    [sym_fsi_directive_decl] = STATE(2535),
    [sym_preproc_line] = STATE(2535),
    [ts_builtin_sym_end] = ACTIONS(7045),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_namespace] = ACTIONS(7047),
    [anon_sym_module] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_open] = ACTIONS(7047),
    [anon_sym_type] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_and] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_exception] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [anon_sym_extern] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7045),
    [anon_sym_POUNDload] = ACTIONS(7045),
    [anon_sym_POUNDtime] = ACTIONS(7045),
    [anon_sym_POUNDI] = ACTIONS(7045),
    [anon_sym_POUNDhelp] = ACTIONS(7045),
    [anon_sym_POUNDquit] = ACTIONS(7045),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(2536)] = {
    [sym_block_comment] = STATE(2536),
    [sym_compiler_directive_decl] = STATE(2536),
    [sym_fsi_directive_decl] = STATE(2536),
    [sym_preproc_line] = STATE(2536),
    [ts_builtin_sym_end] = ACTIONS(7049),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_namespace] = ACTIONS(7051),
    [anon_sym_module] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_open] = ACTIONS(7051),
    [anon_sym_type] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_and] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_exception] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [anon_sym_extern] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7049),
    [anon_sym_POUNDload] = ACTIONS(7049),
    [anon_sym_POUNDtime] = ACTIONS(7049),
    [anon_sym_POUNDI] = ACTIONS(7049),
    [anon_sym_POUNDhelp] = ACTIONS(7049),
    [anon_sym_POUNDquit] = ACTIONS(7049),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(2537)] = {
    [sym_block_comment] = STATE(2537),
    [sym_compiler_directive_decl] = STATE(2537),
    [sym_fsi_directive_decl] = STATE(2537),
    [sym_preproc_line] = STATE(2537),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_module] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_open] = ACTIONS(7053),
    [anon_sym_type] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_and] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_exception] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7055),
    [anon_sym_POUNDload] = ACTIONS(7055),
    [anon_sym_POUNDtime] = ACTIONS(7055),
    [anon_sym_POUNDI] = ACTIONS(7055),
    [anon_sym_POUNDhelp] = ACTIONS(7055),
    [anon_sym_POUNDquit] = ACTIONS(7055),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [anon_sym_POUNDelse] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(2538)] = {
    [sym_block_comment] = STATE(2538),
    [sym_compiler_directive_decl] = STATE(2538),
    [sym_fsi_directive_decl] = STATE(2538),
    [sym_preproc_line] = STATE(2538),
    [ts_builtin_sym_end] = ACTIONS(7057),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_namespace] = ACTIONS(7059),
    [anon_sym_module] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_open] = ACTIONS(7059),
    [anon_sym_type] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_and] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_exception] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [anon_sym_extern] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7057),
    [anon_sym_POUNDload] = ACTIONS(7057),
    [anon_sym_POUNDtime] = ACTIONS(7057),
    [anon_sym_POUNDI] = ACTIONS(7057),
    [anon_sym_POUNDhelp] = ACTIONS(7057),
    [anon_sym_POUNDquit] = ACTIONS(7057),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(2539)] = {
    [sym_block_comment] = STATE(2539),
    [sym_compiler_directive_decl] = STATE(2539),
    [sym_fsi_directive_decl] = STATE(2539),
    [sym_preproc_line] = STATE(2539),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2540)] = {
    [sym_block_comment] = STATE(2540),
    [sym_compiler_directive_decl] = STATE(2540),
    [sym_fsi_directive_decl] = STATE(2540),
    [sym_preproc_line] = STATE(2540),
    [ts_builtin_sym_end] = ACTIONS(7061),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_namespace] = ACTIONS(7063),
    [anon_sym_module] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_open] = ACTIONS(7063),
    [anon_sym_type] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_and] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_exception] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [anon_sym_extern] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7061),
    [anon_sym_POUNDload] = ACTIONS(7061),
    [anon_sym_POUNDtime] = ACTIONS(7061),
    [anon_sym_POUNDI] = ACTIONS(7061),
    [anon_sym_POUNDhelp] = ACTIONS(7061),
    [anon_sym_POUNDquit] = ACTIONS(7061),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(2541)] = {
    [sym_block_comment] = STATE(2541),
    [sym_compiler_directive_decl] = STATE(2541),
    [sym_fsi_directive_decl] = STATE(2541),
    [sym_preproc_line] = STATE(2541),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_module] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_open] = ACTIONS(6902),
    [anon_sym_type] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_and] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_exception] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [anon_sym_extern] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6603),
    [anon_sym_POUNDload] = ACTIONS(6603),
    [anon_sym_POUNDtime] = ACTIONS(6603),
    [anon_sym_POUNDI] = ACTIONS(6603),
    [anon_sym_POUNDhelp] = ACTIONS(6603),
    [anon_sym_POUNDquit] = ACTIONS(6603),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [anon_sym_POUNDelse] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(2542)] = {
    [sym_block_comment] = STATE(2542),
    [sym_compiler_directive_decl] = STATE(2542),
    [sym_fsi_directive_decl] = STATE(2542),
    [sym_preproc_line] = STATE(2542),
    [ts_builtin_sym_end] = ACTIONS(7065),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_namespace] = ACTIONS(7067),
    [anon_sym_module] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_open] = ACTIONS(7067),
    [anon_sym_type] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_and] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_exception] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [anon_sym_extern] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7065),
    [anon_sym_POUNDload] = ACTIONS(7065),
    [anon_sym_POUNDtime] = ACTIONS(7065),
    [anon_sym_POUNDI] = ACTIONS(7065),
    [anon_sym_POUNDhelp] = ACTIONS(7065),
    [anon_sym_POUNDquit] = ACTIONS(7065),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(2543)] = {
    [sym_block_comment] = STATE(2543),
    [sym_compiler_directive_decl] = STATE(2543),
    [sym_fsi_directive_decl] = STATE(2543),
    [sym_preproc_line] = STATE(2543),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_module] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_open] = ACTIONS(6888),
    [anon_sym_type] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_and] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_exception] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [anon_sym_extern] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6886),
    [anon_sym_POUNDload] = ACTIONS(6886),
    [anon_sym_POUNDtime] = ACTIONS(6886),
    [anon_sym_POUNDI] = ACTIONS(6886),
    [anon_sym_POUNDhelp] = ACTIONS(6886),
    [anon_sym_POUNDquit] = ACTIONS(6886),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [anon_sym_POUNDelse] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(2544)] = {
    [sym_block_comment] = STATE(2544),
    [sym_compiler_directive_decl] = STATE(2544),
    [sym_fsi_directive_decl] = STATE(2544),
    [sym_preproc_line] = STATE(2544),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_module] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_open] = ACTIONS(7069),
    [anon_sym_type] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_and] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_exception] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [anon_sym_extern] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7071),
    [anon_sym_POUNDload] = ACTIONS(7071),
    [anon_sym_POUNDtime] = ACTIONS(7071),
    [anon_sym_POUNDI] = ACTIONS(7071),
    [anon_sym_POUNDhelp] = ACTIONS(7071),
    [anon_sym_POUNDquit] = ACTIONS(7071),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [anon_sym_POUNDelse] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(2545)] = {
    [sym_block_comment] = STATE(2545),
    [sym_compiler_directive_decl] = STATE(2545),
    [sym_fsi_directive_decl] = STATE(2545),
    [sym_preproc_line] = STATE(2545),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_module] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_open] = ACTIONS(7073),
    [anon_sym_type] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_and] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_exception] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [anon_sym_extern] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7075),
    [anon_sym_POUNDload] = ACTIONS(7075),
    [anon_sym_POUNDtime] = ACTIONS(7075),
    [anon_sym_POUNDI] = ACTIONS(7075),
    [anon_sym_POUNDhelp] = ACTIONS(7075),
    [anon_sym_POUNDquit] = ACTIONS(7075),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [anon_sym_POUNDelse] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(2546)] = {
    [sym_block_comment] = STATE(2546),
    [sym_compiler_directive_decl] = STATE(2546),
    [sym_fsi_directive_decl] = STATE(2546),
    [sym_preproc_line] = STATE(2546),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_module] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_open] = ACTIONS(6906),
    [anon_sym_type] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_and] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_exception] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [anon_sym_extern] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6904),
    [anon_sym_POUNDload] = ACTIONS(6904),
    [anon_sym_POUNDtime] = ACTIONS(6904),
    [anon_sym_POUNDI] = ACTIONS(6904),
    [anon_sym_POUNDhelp] = ACTIONS(6904),
    [anon_sym_POUNDquit] = ACTIONS(6904),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [anon_sym_POUNDelse] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(2547)] = {
    [sym_block_comment] = STATE(2547),
    [sym_compiler_directive_decl] = STATE(2547),
    [sym_fsi_directive_decl] = STATE(2547),
    [sym_preproc_line] = STATE(2547),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_module] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_open] = ACTIONS(7077),
    [anon_sym_type] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_and] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_exception] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [anon_sym_extern] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7079),
    [anon_sym_POUNDload] = ACTIONS(7079),
    [anon_sym_POUNDtime] = ACTIONS(7079),
    [anon_sym_POUNDI] = ACTIONS(7079),
    [anon_sym_POUNDhelp] = ACTIONS(7079),
    [anon_sym_POUNDquit] = ACTIONS(7079),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [anon_sym_POUNDelse] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(2548)] = {
    [sym_block_comment] = STATE(2548),
    [sym_compiler_directive_decl] = STATE(2548),
    [sym_fsi_directive_decl] = STATE(2548),
    [sym_preproc_line] = STATE(2548),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_module] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_open] = ACTIONS(6913),
    [anon_sym_type] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_and] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_exception] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [anon_sym_extern] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6911),
    [anon_sym_POUNDload] = ACTIONS(6911),
    [anon_sym_POUNDtime] = ACTIONS(6911),
    [anon_sym_POUNDI] = ACTIONS(6911),
    [anon_sym_POUNDhelp] = ACTIONS(6911),
    [anon_sym_POUNDquit] = ACTIONS(6911),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [anon_sym_POUNDelse] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(2549)] = {
    [sym__else_expression] = STATE(3197),
    [sym_elif_expression] = STATE(2940),
    [sym_block_comment] = STATE(2549),
    [sym_compiler_directive_decl] = STATE(2549),
    [sym_fsi_directive_decl] = STATE(2549),
    [sym_preproc_line] = STATE(2549),
    [aux_sym_if_expression_repeat1] = STATE(2559),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7081),
    [anon_sym_elif] = ACTIONS(7083),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [anon_sym_inherit] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [anon_sym_POUNDelse] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2550)] = {
    [sym_block_comment] = STATE(2550),
    [sym_compiler_directive_decl] = STATE(2550),
    [sym_fsi_directive_decl] = STATE(2550),
    [sym_preproc_line] = STATE(2550),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_module] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_open] = ACTIONS(7085),
    [anon_sym_type] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_and] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_exception] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [anon_sym_extern] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7087),
    [anon_sym_POUNDload] = ACTIONS(7087),
    [anon_sym_POUNDtime] = ACTIONS(7087),
    [anon_sym_POUNDI] = ACTIONS(7087),
    [anon_sym_POUNDhelp] = ACTIONS(7087),
    [anon_sym_POUNDquit] = ACTIONS(7087),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [anon_sym_POUNDelse] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(2551)] = {
    [sym_block_comment] = STATE(2551),
    [sym_compiler_directive_decl] = STATE(2551),
    [sym_fsi_directive_decl] = STATE(2551),
    [sym_preproc_line] = STATE(2551),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_module] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_open] = ACTIONS(7089),
    [anon_sym_type] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_and] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_exception] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [anon_sym_extern] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7091),
    [anon_sym_POUNDload] = ACTIONS(7091),
    [anon_sym_POUNDtime] = ACTIONS(7091),
    [anon_sym_POUNDI] = ACTIONS(7091),
    [anon_sym_POUNDhelp] = ACTIONS(7091),
    [anon_sym_POUNDquit] = ACTIONS(7091),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [anon_sym_POUNDelse] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(2552)] = {
    [sym_block_comment] = STATE(2552),
    [sym_compiler_directive_decl] = STATE(2552),
    [sym_fsi_directive_decl] = STATE(2552),
    [sym_preproc_line] = STATE(2552),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_module] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_open] = ACTIONS(7093),
    [anon_sym_type] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_and] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_exception] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [anon_sym_extern] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7095),
    [anon_sym_POUNDload] = ACTIONS(7095),
    [anon_sym_POUNDtime] = ACTIONS(7095),
    [anon_sym_POUNDI] = ACTIONS(7095),
    [anon_sym_POUNDhelp] = ACTIONS(7095),
    [anon_sym_POUNDquit] = ACTIONS(7095),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [anon_sym_POUNDelse] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(2553)] = {
    [sym_block_comment] = STATE(2553),
    [sym_compiler_directive_decl] = STATE(2553),
    [sym_fsi_directive_decl] = STATE(2553),
    [sym_preproc_line] = STATE(2553),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_module] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_open] = ACTIONS(6917),
    [anon_sym_type] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_and] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_exception] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [anon_sym_extern] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6915),
    [anon_sym_POUNDload] = ACTIONS(6915),
    [anon_sym_POUNDtime] = ACTIONS(6915),
    [anon_sym_POUNDI] = ACTIONS(6915),
    [anon_sym_POUNDhelp] = ACTIONS(6915),
    [anon_sym_POUNDquit] = ACTIONS(6915),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [anon_sym_POUNDelse] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(2554)] = {
    [sym_block_comment] = STATE(2554),
    [sym_compiler_directive_decl] = STATE(2554),
    [sym_fsi_directive_decl] = STATE(2554),
    [sym_preproc_line] = STATE(2554),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_module] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_open] = ACTIONS(6921),
    [anon_sym_type] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_and] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_exception] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [anon_sym_extern] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6919),
    [anon_sym_POUNDload] = ACTIONS(6919),
    [anon_sym_POUNDtime] = ACTIONS(6919),
    [anon_sym_POUNDI] = ACTIONS(6919),
    [anon_sym_POUNDhelp] = ACTIONS(6919),
    [anon_sym_POUNDquit] = ACTIONS(6919),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [anon_sym_POUNDelse] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(2555)] = {
    [sym_block_comment] = STATE(2555),
    [sym_compiler_directive_decl] = STATE(2555),
    [sym_fsi_directive_decl] = STATE(2555),
    [sym_preproc_line] = STATE(2555),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_module] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_open] = ACTIONS(7097),
    [anon_sym_type] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_and] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_exception] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [anon_sym_extern] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7099),
    [anon_sym_POUNDload] = ACTIONS(7099),
    [anon_sym_POUNDtime] = ACTIONS(7099),
    [anon_sym_POUNDI] = ACTIONS(7099),
    [anon_sym_POUNDhelp] = ACTIONS(7099),
    [anon_sym_POUNDquit] = ACTIONS(7099),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [anon_sym_POUNDelse] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(2556)] = {
    [sym_block_comment] = STATE(2556),
    [sym_compiler_directive_decl] = STATE(2556),
    [sym_fsi_directive_decl] = STATE(2556),
    [sym_preproc_line] = STATE(2556),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_module] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_open] = ACTIONS(7101),
    [anon_sym_type] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_and] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_exception] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7103),
    [anon_sym_POUNDload] = ACTIONS(7103),
    [anon_sym_POUNDtime] = ACTIONS(7103),
    [anon_sym_POUNDI] = ACTIONS(7103),
    [anon_sym_POUNDhelp] = ACTIONS(7103),
    [anon_sym_POUNDquit] = ACTIONS(7103),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [anon_sym_POUNDelse] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(2557)] = {
    [sym_block_comment] = STATE(2557),
    [sym_compiler_directive_decl] = STATE(2557),
    [sym_fsi_directive_decl] = STATE(2557),
    [sym_preproc_line] = STATE(2557),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_module] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_open] = ACTIONS(6925),
    [anon_sym_type] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_and] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_exception] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [anon_sym_extern] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6923),
    [anon_sym_POUNDload] = ACTIONS(6923),
    [anon_sym_POUNDtime] = ACTIONS(6923),
    [anon_sym_POUNDI] = ACTIONS(6923),
    [anon_sym_POUNDhelp] = ACTIONS(6923),
    [anon_sym_POUNDquit] = ACTIONS(6923),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [anon_sym_POUNDelse] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(2558)] = {
    [sym_block_comment] = STATE(2558),
    [sym_compiler_directive_decl] = STATE(2558),
    [sym_fsi_directive_decl] = STATE(2558),
    [sym_preproc_line] = STATE(2558),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_module] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_open] = ACTIONS(7105),
    [anon_sym_type] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_and] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_exception] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7107),
    [anon_sym_POUNDload] = ACTIONS(7107),
    [anon_sym_POUNDtime] = ACTIONS(7107),
    [anon_sym_POUNDI] = ACTIONS(7107),
    [anon_sym_POUNDhelp] = ACTIONS(7107),
    [anon_sym_POUNDquit] = ACTIONS(7107),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [anon_sym_POUNDelse] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(2559)] = {
    [sym__else_expression] = STATE(3257),
    [sym_elif_expression] = STATE(2940),
    [sym_block_comment] = STATE(2559),
    [sym_compiler_directive_decl] = STATE(2559),
    [sym_fsi_directive_decl] = STATE(2559),
    [sym_preproc_line] = STATE(2559),
    [aux_sym_if_expression_repeat1] = STATE(2720),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7081),
    [anon_sym_elif] = ACTIONS(7083),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [anon_sym_inherit] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [anon_sym_POUNDelse] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2560)] = {
    [sym_block_comment] = STATE(2560),
    [sym_compiler_directive_decl] = STATE(2560),
    [sym_fsi_directive_decl] = STATE(2560),
    [sym_preproc_line] = STATE(2560),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_module] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_open] = ACTIONS(7109),
    [anon_sym_type] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_and] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_exception] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7111),
    [anon_sym_POUNDload] = ACTIONS(7111),
    [anon_sym_POUNDtime] = ACTIONS(7111),
    [anon_sym_POUNDI] = ACTIONS(7111),
    [anon_sym_POUNDhelp] = ACTIONS(7111),
    [anon_sym_POUNDquit] = ACTIONS(7111),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [anon_sym_POUNDelse] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(2561)] = {
    [sym_block_comment] = STATE(2561),
    [sym_compiler_directive_decl] = STATE(2561),
    [sym_fsi_directive_decl] = STATE(2561),
    [sym_preproc_line] = STATE(2561),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_module] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_open] = ACTIONS(7113),
    [anon_sym_type] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_and] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_exception] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7115),
    [anon_sym_POUNDload] = ACTIONS(7115),
    [anon_sym_POUNDtime] = ACTIONS(7115),
    [anon_sym_POUNDI] = ACTIONS(7115),
    [anon_sym_POUNDhelp] = ACTIONS(7115),
    [anon_sym_POUNDquit] = ACTIONS(7115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [anon_sym_POUNDelse] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(2562)] = {
    [sym_block_comment] = STATE(2562),
    [sym_compiler_directive_decl] = STATE(2562),
    [sym_fsi_directive_decl] = STATE(2562),
    [sym_preproc_line] = STATE(2562),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_module] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_open] = ACTIONS(7117),
    [anon_sym_type] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_and] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_exception] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7119),
    [anon_sym_POUNDload] = ACTIONS(7119),
    [anon_sym_POUNDtime] = ACTIONS(7119),
    [anon_sym_POUNDI] = ACTIONS(7119),
    [anon_sym_POUNDhelp] = ACTIONS(7119),
    [anon_sym_POUNDquit] = ACTIONS(7119),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [anon_sym_POUNDelse] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(2563)] = {
    [sym_block_comment] = STATE(2563),
    [sym_compiler_directive_decl] = STATE(2563),
    [sym_fsi_directive_decl] = STATE(2563),
    [sym_preproc_line] = STATE(2563),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_module] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_open] = ACTIONS(7121),
    [anon_sym_type] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_and] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_exception] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7123),
    [anon_sym_POUNDload] = ACTIONS(7123),
    [anon_sym_POUNDtime] = ACTIONS(7123),
    [anon_sym_POUNDI] = ACTIONS(7123),
    [anon_sym_POUNDhelp] = ACTIONS(7123),
    [anon_sym_POUNDquit] = ACTIONS(7123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [anon_sym_POUNDelse] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(2564)] = {
    [sym_block_comment] = STATE(2564),
    [sym_compiler_directive_decl] = STATE(2564),
    [sym_fsi_directive_decl] = STATE(2564),
    [sym_preproc_line] = STATE(2564),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_module] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_open] = ACTIONS(6936),
    [anon_sym_type] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_and] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_exception] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [anon_sym_extern] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6934),
    [anon_sym_POUNDload] = ACTIONS(6934),
    [anon_sym_POUNDtime] = ACTIONS(6934),
    [anon_sym_POUNDI] = ACTIONS(6934),
    [anon_sym_POUNDhelp] = ACTIONS(6934),
    [anon_sym_POUNDquit] = ACTIONS(6934),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [anon_sym_POUNDelse] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(2565)] = {
    [sym_block_comment] = STATE(2565),
    [sym_compiler_directive_decl] = STATE(2565),
    [sym_fsi_directive_decl] = STATE(2565),
    [sym_preproc_line] = STATE(2565),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_module] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_open] = ACTIONS(6940),
    [anon_sym_type] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_and] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_exception] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [anon_sym_extern] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6938),
    [anon_sym_POUNDload] = ACTIONS(6938),
    [anon_sym_POUNDtime] = ACTIONS(6938),
    [anon_sym_POUNDI] = ACTIONS(6938),
    [anon_sym_POUNDhelp] = ACTIONS(6938),
    [anon_sym_POUNDquit] = ACTIONS(6938),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [anon_sym_POUNDelse] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(2566)] = {
    [sym_block_comment] = STATE(2566),
    [sym_compiler_directive_decl] = STATE(2566),
    [sym_fsi_directive_decl] = STATE(2566),
    [sym_preproc_line] = STATE(2566),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_module] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_open] = ACTIONS(7015),
    [anon_sym_type] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_and] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_exception] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [anon_sym_extern] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7013),
    [anon_sym_POUNDload] = ACTIONS(7013),
    [anon_sym_POUNDtime] = ACTIONS(7013),
    [anon_sym_POUNDI] = ACTIONS(7013),
    [anon_sym_POUNDhelp] = ACTIONS(7013),
    [anon_sym_POUNDquit] = ACTIONS(7013),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [anon_sym_POUNDelse] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(2567)] = {
    [sym_block_comment] = STATE(2567),
    [sym_compiler_directive_decl] = STATE(2567),
    [sym_fsi_directive_decl] = STATE(2567),
    [sym_preproc_line] = STATE(2567),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_module] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_open] = ACTIONS(7031),
    [anon_sym_type] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_and] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_exception] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [anon_sym_extern] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7029),
    [anon_sym_POUNDload] = ACTIONS(7029),
    [anon_sym_POUNDtime] = ACTIONS(7029),
    [anon_sym_POUNDI] = ACTIONS(7029),
    [anon_sym_POUNDhelp] = ACTIONS(7029),
    [anon_sym_POUNDquit] = ACTIONS(7029),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [anon_sym_POUNDelse] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(2568)] = {
    [sym_block_comment] = STATE(2568),
    [sym_compiler_directive_decl] = STATE(2568),
    [sym_fsi_directive_decl] = STATE(2568),
    [sym_preproc_line] = STATE(2568),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_module] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_open] = ACTIONS(7047),
    [anon_sym_type] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_and] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_exception] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [anon_sym_extern] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7045),
    [anon_sym_POUNDload] = ACTIONS(7045),
    [anon_sym_POUNDtime] = ACTIONS(7045),
    [anon_sym_POUNDI] = ACTIONS(7045),
    [anon_sym_POUNDhelp] = ACTIONS(7045),
    [anon_sym_POUNDquit] = ACTIONS(7045),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [anon_sym_POUNDelse] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(2569)] = {
    [sym_block_comment] = STATE(2569),
    [sym_compiler_directive_decl] = STATE(2569),
    [sym_fsi_directive_decl] = STATE(2569),
    [sym_preproc_line] = STATE(2569),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_module] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_open] = ACTIONS(7051),
    [anon_sym_type] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_and] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_exception] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [anon_sym_extern] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7049),
    [anon_sym_POUNDload] = ACTIONS(7049),
    [anon_sym_POUNDtime] = ACTIONS(7049),
    [anon_sym_POUNDI] = ACTIONS(7049),
    [anon_sym_POUNDhelp] = ACTIONS(7049),
    [anon_sym_POUNDquit] = ACTIONS(7049),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [anon_sym_POUNDelse] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(2570)] = {
    [sym_block_comment] = STATE(2570),
    [sym_compiler_directive_decl] = STATE(2570),
    [sym_fsi_directive_decl] = STATE(2570),
    [sym_preproc_line] = STATE(2570),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_module] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_open] = ACTIONS(7063),
    [anon_sym_type] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_and] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_exception] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [anon_sym_extern] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7061),
    [anon_sym_POUNDload] = ACTIONS(7061),
    [anon_sym_POUNDtime] = ACTIONS(7061),
    [anon_sym_POUNDI] = ACTIONS(7061),
    [anon_sym_POUNDhelp] = ACTIONS(7061),
    [anon_sym_POUNDquit] = ACTIONS(7061),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [anon_sym_POUNDelse] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(2571)] = {
    [sym_block_comment] = STATE(2571),
    [sym_compiler_directive_decl] = STATE(2571),
    [sym_fsi_directive_decl] = STATE(2571),
    [sym_preproc_line] = STATE(2571),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_module] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_open] = ACTIONS(7067),
    [anon_sym_type] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_and] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_exception] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [anon_sym_extern] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7065),
    [anon_sym_POUNDload] = ACTIONS(7065),
    [anon_sym_POUNDtime] = ACTIONS(7065),
    [anon_sym_POUNDI] = ACTIONS(7065),
    [anon_sym_POUNDhelp] = ACTIONS(7065),
    [anon_sym_POUNDquit] = ACTIONS(7065),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [anon_sym_POUNDelse] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(2572)] = {
    [sym_block_comment] = STATE(2572),
    [sym_compiler_directive_decl] = STATE(2572),
    [sym_fsi_directive_decl] = STATE(2572),
    [sym_preproc_line] = STATE(2572),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_module] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_open] = ACTIONS(7125),
    [anon_sym_type] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_and] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_exception] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [anon_sym_extern] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7127),
    [anon_sym_POUNDload] = ACTIONS(7127),
    [anon_sym_POUNDtime] = ACTIONS(7127),
    [anon_sym_POUNDI] = ACTIONS(7127),
    [anon_sym_POUNDhelp] = ACTIONS(7127),
    [anon_sym_POUNDquit] = ACTIONS(7127),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [anon_sym_POUNDelse] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(2573)] = {
    [sym_block_comment] = STATE(2573),
    [sym_compiler_directive_decl] = STATE(2573),
    [sym_fsi_directive_decl] = STATE(2573),
    [sym_preproc_line] = STATE(2573),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_module] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_open] = ACTIONS(7129),
    [anon_sym_type] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_and] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_exception] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7131),
    [anon_sym_POUNDload] = ACTIONS(7131),
    [anon_sym_POUNDtime] = ACTIONS(7131),
    [anon_sym_POUNDI] = ACTIONS(7131),
    [anon_sym_POUNDhelp] = ACTIONS(7131),
    [anon_sym_POUNDquit] = ACTIONS(7131),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [anon_sym_POUNDelse] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(2574)] = {
    [sym_block_comment] = STATE(2574),
    [sym_compiler_directive_decl] = STATE(2574),
    [sym_fsi_directive_decl] = STATE(2574),
    [sym_preproc_line] = STATE(2574),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_module] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_open] = ACTIONS(7133),
    [anon_sym_type] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_and] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_exception] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7135),
    [anon_sym_POUNDload] = ACTIONS(7135),
    [anon_sym_POUNDtime] = ACTIONS(7135),
    [anon_sym_POUNDI] = ACTIONS(7135),
    [anon_sym_POUNDhelp] = ACTIONS(7135),
    [anon_sym_POUNDquit] = ACTIONS(7135),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [anon_sym_POUNDelse] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(2575)] = {
    [sym_block_comment] = STATE(2575),
    [sym_compiler_directive_decl] = STATE(2575),
    [sym_fsi_directive_decl] = STATE(2575),
    [sym_preproc_line] = STATE(2575),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_module] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_open] = ACTIONS(7137),
    [anon_sym_type] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_and] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_exception] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7139),
    [anon_sym_POUNDload] = ACTIONS(7139),
    [anon_sym_POUNDtime] = ACTIONS(7139),
    [anon_sym_POUNDI] = ACTIONS(7139),
    [anon_sym_POUNDhelp] = ACTIONS(7139),
    [anon_sym_POUNDquit] = ACTIONS(7139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [anon_sym_POUNDelse] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(2576)] = {
    [sym_block_comment] = STATE(2576),
    [sym_compiler_directive_decl] = STATE(2576),
    [sym_fsi_directive_decl] = STATE(2576),
    [sym_preproc_line] = STATE(2576),
    [ts_builtin_sym_end] = ACTIONS(7127),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_namespace] = ACTIONS(7125),
    [anon_sym_module] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_open] = ACTIONS(7125),
    [anon_sym_type] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_and] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_exception] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [anon_sym_extern] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7127),
    [anon_sym_POUNDload] = ACTIONS(7127),
    [anon_sym_POUNDtime] = ACTIONS(7127),
    [anon_sym_POUNDI] = ACTIONS(7127),
    [anon_sym_POUNDhelp] = ACTIONS(7127),
    [anon_sym_POUNDquit] = ACTIONS(7127),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(2577)] = {
    [sym_block_comment] = STATE(2577),
    [sym_compiler_directive_decl] = STATE(2577),
    [sym_fsi_directive_decl] = STATE(2577),
    [sym_preproc_line] = STATE(2577),
    [ts_builtin_sym_end] = ACTIONS(7131),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_namespace] = ACTIONS(7129),
    [anon_sym_module] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_open] = ACTIONS(7129),
    [anon_sym_type] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_and] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_exception] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7131),
    [anon_sym_POUNDload] = ACTIONS(7131),
    [anon_sym_POUNDtime] = ACTIONS(7131),
    [anon_sym_POUNDI] = ACTIONS(7131),
    [anon_sym_POUNDhelp] = ACTIONS(7131),
    [anon_sym_POUNDquit] = ACTIONS(7131),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(2578)] = {
    [sym_block_comment] = STATE(2578),
    [sym_compiler_directive_decl] = STATE(2578),
    [sym_fsi_directive_decl] = STATE(2578),
    [sym_preproc_line] = STATE(2578),
    [ts_builtin_sym_end] = ACTIONS(7135),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_namespace] = ACTIONS(7133),
    [anon_sym_module] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_open] = ACTIONS(7133),
    [anon_sym_type] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_and] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_exception] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7135),
    [anon_sym_POUNDload] = ACTIONS(7135),
    [anon_sym_POUNDtime] = ACTIONS(7135),
    [anon_sym_POUNDI] = ACTIONS(7135),
    [anon_sym_POUNDhelp] = ACTIONS(7135),
    [anon_sym_POUNDquit] = ACTIONS(7135),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(2579)] = {
    [sym_block_comment] = STATE(2579),
    [sym_compiler_directive_decl] = STATE(2579),
    [sym_fsi_directive_decl] = STATE(2579),
    [sym_preproc_line] = STATE(2579),
    [ts_builtin_sym_end] = ACTIONS(7139),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_namespace] = ACTIONS(7137),
    [anon_sym_module] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_open] = ACTIONS(7137),
    [anon_sym_type] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_and] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_exception] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7139),
    [anon_sym_POUNDload] = ACTIONS(7139),
    [anon_sym_POUNDtime] = ACTIONS(7139),
    [anon_sym_POUNDI] = ACTIONS(7139),
    [anon_sym_POUNDhelp] = ACTIONS(7139),
    [anon_sym_POUNDquit] = ACTIONS(7139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(2580)] = {
    [sym_block_comment] = STATE(2580),
    [sym_compiler_directive_decl] = STATE(2580),
    [sym_fsi_directive_decl] = STATE(2580),
    [sym_preproc_line] = STATE(2580),
    [ts_builtin_sym_end] = ACTIONS(7055),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_namespace] = ACTIONS(7053),
    [anon_sym_module] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_open] = ACTIONS(7053),
    [anon_sym_type] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_and] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_exception] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7055),
    [anon_sym_POUNDload] = ACTIONS(7055),
    [anon_sym_POUNDtime] = ACTIONS(7055),
    [anon_sym_POUNDI] = ACTIONS(7055),
    [anon_sym_POUNDhelp] = ACTIONS(7055),
    [anon_sym_POUNDquit] = ACTIONS(7055),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(2581)] = {
    [sym_block_comment] = STATE(2581),
    [sym_compiler_directive_decl] = STATE(2581),
    [sym_fsi_directive_decl] = STATE(2581),
    [sym_preproc_line] = STATE(2581),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_module] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_open] = ACTIONS(6692),
    [anon_sym_type] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_and] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_exception] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [anon_sym_extern] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6694),
    [anon_sym_POUNDload] = ACTIONS(6694),
    [anon_sym_POUNDtime] = ACTIONS(6694),
    [anon_sym_POUNDI] = ACTIONS(6694),
    [anon_sym_POUNDhelp] = ACTIONS(6694),
    [anon_sym_POUNDquit] = ACTIONS(6694),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(2582)] = {
    [sym_block_comment] = STATE(2582),
    [sym_compiler_directive_decl] = STATE(2582),
    [sym_fsi_directive_decl] = STATE(2582),
    [sym_preproc_line] = STATE(2582),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_module] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_open] = ACTIONS(7141),
    [anon_sym_type] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_and] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_exception] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7143),
    [anon_sym_POUNDload] = ACTIONS(7143),
    [anon_sym_POUNDtime] = ACTIONS(7143),
    [anon_sym_POUNDI] = ACTIONS(7143),
    [anon_sym_POUNDhelp] = ACTIONS(7143),
    [anon_sym_POUNDquit] = ACTIONS(7143),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [anon_sym_POUNDelse] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(2583)] = {
    [sym_block_comment] = STATE(2583),
    [sym_compiler_directive_decl] = STATE(2583),
    [sym_fsi_directive_decl] = STATE(2583),
    [sym_preproc_line] = STATE(2583),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_module] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_open] = ACTIONS(7145),
    [anon_sym_type] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_and] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_exception] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7147),
    [anon_sym_POUNDload] = ACTIONS(7147),
    [anon_sym_POUNDtime] = ACTIONS(7147),
    [anon_sym_POUNDI] = ACTIONS(7147),
    [anon_sym_POUNDhelp] = ACTIONS(7147),
    [anon_sym_POUNDquit] = ACTIONS(7147),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [anon_sym_POUNDelse] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(2584)] = {
    [sym_block_comment] = STATE(2584),
    [sym_compiler_directive_decl] = STATE(2584),
    [sym_fsi_directive_decl] = STATE(2584),
    [sym_preproc_line] = STATE(2584),
    [aux_sym_type_argument_repeat1] = STATE(2585),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_inherit] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2585)] = {
    [sym_block_comment] = STATE(2585),
    [sym_compiler_directive_decl] = STATE(2585),
    [sym_fsi_directive_decl] = STATE(2585),
    [sym_preproc_line] = STATE(2585),
    [aux_sym_type_argument_repeat1] = STATE(2585),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7149),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_inherit] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2586)] = {
    [sym_block_comment] = STATE(2586),
    [sym_compiler_directive_decl] = STATE(2586),
    [sym_fsi_directive_decl] = STATE(2586),
    [sym_preproc_line] = STATE(2586),
    [aux_sym_type_argument_repeat1] = STATE(2587),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2587)] = {
    [sym_block_comment] = STATE(2587),
    [sym_compiler_directive_decl] = STATE(2587),
    [sym_fsi_directive_decl] = STATE(2587),
    [sym_preproc_line] = STATE(2587),
    [aux_sym_type_argument_repeat1] = STATE(2587),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7152),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2588)] = {
    [sym_block_comment] = STATE(2588),
    [sym_compiler_directive_decl] = STATE(2588),
    [sym_fsi_directive_decl] = STATE(2588),
    [sym_preproc_line] = STATE(2588),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_module] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_open] = ACTIONS(7155),
    [anon_sym_type] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_and] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_exception] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [anon_sym_extern] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7157),
    [anon_sym_POUNDload] = ACTIONS(7157),
    [anon_sym_POUNDtime] = ACTIONS(7157),
    [anon_sym_POUNDI] = ACTIONS(7157),
    [anon_sym_POUNDhelp] = ACTIONS(7157),
    [anon_sym_POUNDquit] = ACTIONS(7157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [anon_sym_POUNDelse] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(2589)] = {
    [sym_block_comment] = STATE(2589),
    [sym_compiler_directive_decl] = STATE(2589),
    [sym_fsi_directive_decl] = STATE(2589),
    [sym_preproc_line] = STATE(2589),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_module] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_open] = ACTIONS(7159),
    [anon_sym_type] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_and] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_exception] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [anon_sym_extern] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7161),
    [anon_sym_POUNDload] = ACTIONS(7161),
    [anon_sym_POUNDtime] = ACTIONS(7161),
    [anon_sym_POUNDI] = ACTIONS(7161),
    [anon_sym_POUNDhelp] = ACTIONS(7161),
    [anon_sym_POUNDquit] = ACTIONS(7161),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [anon_sym_POUNDelse] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(2590)] = {
    [sym_block_comment] = STATE(2590),
    [sym_compiler_directive_decl] = STATE(2590),
    [sym_fsi_directive_decl] = STATE(2590),
    [sym_preproc_line] = STATE(2590),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7163),
    [anon_sym_uy] = ACTIONS(7165),
    [anon_sym_s] = ACTIONS(7167),
    [anon_sym_us] = ACTIONS(7169),
    [anon_sym_l] = ACTIONS(7171),
    [aux_sym_uint32_token1] = ACTIONS(7173),
    [anon_sym_n] = ACTIONS(7175),
    [anon_sym_un] = ACTIONS(7177),
    [anon_sym_L] = ACTIONS(7179),
    [aux_sym_uint64_token1] = ACTIONS(7181),
    [aux_sym_bignum_token1] = ACTIONS(7183),
    [aux_sym_decimal_token1] = ACTIONS(7185),
    [anon_sym_DOT2] = ACTIONS(7187),
    [aux_sym_float_token1] = ACTIONS(7189),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2591)] = {
    [sym_block_comment] = STATE(2591),
    [sym_compiler_directive_decl] = STATE(2591),
    [sym_fsi_directive_decl] = STATE(2591),
    [sym_preproc_line] = STATE(2591),
    [aux_sym_long_identifier_repeat1] = STATE(2607),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7191),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_inherit] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2592)] = {
    [sym_block_comment] = STATE(2592),
    [sym_compiler_directive_decl] = STATE(2592),
    [sym_fsi_directive_decl] = STATE(2592),
    [sym_preproc_line] = STATE(2592),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2593)] = {
    [sym_block_comment] = STATE(2593),
    [sym_compiler_directive_decl] = STATE(2593),
    [sym_fsi_directive_decl] = STATE(2593),
    [sym_preproc_line] = STATE(2593),
    [aux_sym_long_identifier_repeat1] = STATE(2619),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2594)] = {
    [sym_block_comment] = STATE(2594),
    [sym_compiler_directive_decl] = STATE(2594),
    [sym_fsi_directive_decl] = STATE(2594),
    [sym_preproc_line] = STATE(2594),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(2595)] = {
    [sym_block_comment] = STATE(2595),
    [sym_compiler_directive_decl] = STATE(2595),
    [sym_fsi_directive_decl] = STATE(2595),
    [sym_preproc_line] = STATE(2595),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_module] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_open] = ACTIONS(6932),
    [anon_sym_type] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_and] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_exception] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [anon_sym_extern] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6930),
    [anon_sym_POUNDload] = ACTIONS(6930),
    [anon_sym_POUNDtime] = ACTIONS(6930),
    [anon_sym_POUNDI] = ACTIONS(6930),
    [anon_sym_POUNDhelp] = ACTIONS(6930),
    [anon_sym_POUNDquit] = ACTIONS(6930),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [anon_sym_POUNDelse] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(2596)] = {
    [sym_block_comment] = STATE(2596),
    [sym_compiler_directive_decl] = STATE(2596),
    [sym_fsi_directive_decl] = STATE(2596),
    [sym_preproc_line] = STATE(2596),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2597)] = {
    [sym_block_comment] = STATE(2597),
    [sym_compiler_directive_decl] = STATE(2597),
    [sym_fsi_directive_decl] = STATE(2597),
    [sym_preproc_line] = STATE(2597),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_module] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_open] = ACTIONS(6867),
    [anon_sym_type] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_and] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_exception] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [anon_sym_extern] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6865),
    [anon_sym_POUNDload] = ACTIONS(6865),
    [anon_sym_POUNDtime] = ACTIONS(6865),
    [anon_sym_POUNDI] = ACTIONS(6865),
    [anon_sym_POUNDhelp] = ACTIONS(6865),
    [anon_sym_POUNDquit] = ACTIONS(6865),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(2598)] = {
    [sym_block_comment] = STATE(2598),
    [sym_compiler_directive_decl] = STATE(2598),
    [sym_fsi_directive_decl] = STATE(2598),
    [sym_preproc_line] = STATE(2598),
    [ts_builtin_sym_end] = ACTIONS(7071),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_namespace] = ACTIONS(7069),
    [anon_sym_module] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_open] = ACTIONS(7069),
    [anon_sym_type] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_and] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_exception] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [anon_sym_extern] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7071),
    [anon_sym_POUNDload] = ACTIONS(7071),
    [anon_sym_POUNDtime] = ACTIONS(7071),
    [anon_sym_POUNDI] = ACTIONS(7071),
    [anon_sym_POUNDhelp] = ACTIONS(7071),
    [anon_sym_POUNDquit] = ACTIONS(7071),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(2599)] = {
    [sym_block_comment] = STATE(2599),
    [sym_compiler_directive_decl] = STATE(2599),
    [sym_fsi_directive_decl] = STATE(2599),
    [sym_preproc_line] = STATE(2599),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_module] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_open] = ACTIONS(6871),
    [anon_sym_type] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_and] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_exception] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [anon_sym_extern] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6869),
    [anon_sym_POUNDload] = ACTIONS(6869),
    [anon_sym_POUNDtime] = ACTIONS(6869),
    [anon_sym_POUNDI] = ACTIONS(6869),
    [anon_sym_POUNDhelp] = ACTIONS(6869),
    [anon_sym_POUNDquit] = ACTIONS(6869),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(2600)] = {
    [sym_block_comment] = STATE(2600),
    [sym_compiler_directive_decl] = STATE(2600),
    [sym_fsi_directive_decl] = STATE(2600),
    [sym_preproc_line] = STATE(2600),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_module] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_open] = ACTIONS(6875),
    [anon_sym_type] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_and] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_exception] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [anon_sym_extern] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6873),
    [anon_sym_POUNDload] = ACTIONS(6873),
    [anon_sym_POUNDtime] = ACTIONS(6873),
    [anon_sym_POUNDI] = ACTIONS(6873),
    [anon_sym_POUNDhelp] = ACTIONS(6873),
    [anon_sym_POUNDquit] = ACTIONS(6873),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(2601)] = {
    [sym_block_comment] = STATE(2601),
    [sym_compiler_directive_decl] = STATE(2601),
    [sym_fsi_directive_decl] = STATE(2601),
    [sym_preproc_line] = STATE(2601),
    [ts_builtin_sym_end] = ACTIONS(7075),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_namespace] = ACTIONS(7073),
    [anon_sym_module] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_open] = ACTIONS(7073),
    [anon_sym_type] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_and] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_exception] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [anon_sym_extern] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7075),
    [anon_sym_POUNDload] = ACTIONS(7075),
    [anon_sym_POUNDtime] = ACTIONS(7075),
    [anon_sym_POUNDI] = ACTIONS(7075),
    [anon_sym_POUNDhelp] = ACTIONS(7075),
    [anon_sym_POUNDquit] = ACTIONS(7075),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(2602)] = {
    [sym_block_comment] = STATE(2602),
    [sym_compiler_directive_decl] = STATE(2602),
    [sym_fsi_directive_decl] = STATE(2602),
    [sym_preproc_line] = STATE(2602),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_module] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_open] = ACTIONS(6944),
    [anon_sym_type] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_and] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_exception] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [anon_sym_extern] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6942),
    [anon_sym_POUNDload] = ACTIONS(6942),
    [anon_sym_POUNDtime] = ACTIONS(6942),
    [anon_sym_POUNDI] = ACTIONS(6942),
    [anon_sym_POUNDhelp] = ACTIONS(6942),
    [anon_sym_POUNDquit] = ACTIONS(6942),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [anon_sym_POUNDelse] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(2603)] = {
    [sym_block_comment] = STATE(2603),
    [sym_compiler_directive_decl] = STATE(2603),
    [sym_fsi_directive_decl] = STATE(2603),
    [sym_preproc_line] = STATE(2603),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_module] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_open] = ACTIONS(6948),
    [anon_sym_type] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_and] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_exception] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [anon_sym_extern] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6946),
    [anon_sym_POUNDload] = ACTIONS(6946),
    [anon_sym_POUNDtime] = ACTIONS(6946),
    [anon_sym_POUNDI] = ACTIONS(6946),
    [anon_sym_POUNDhelp] = ACTIONS(6946),
    [anon_sym_POUNDquit] = ACTIONS(6946),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [anon_sym_POUNDelse] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(2604)] = {
    [sym_block_comment] = STATE(2604),
    [sym_compiler_directive_decl] = STATE(2604),
    [sym_fsi_directive_decl] = STATE(2604),
    [sym_preproc_line] = STATE(2604),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_module] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_open] = ACTIONS(6952),
    [anon_sym_type] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_and] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_exception] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [anon_sym_extern] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6950),
    [anon_sym_POUNDload] = ACTIONS(6950),
    [anon_sym_POUNDtime] = ACTIONS(6950),
    [anon_sym_POUNDI] = ACTIONS(6950),
    [anon_sym_POUNDhelp] = ACTIONS(6950),
    [anon_sym_POUNDquit] = ACTIONS(6950),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [anon_sym_POUNDelse] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(2605)] = {
    [sym_block_comment] = STATE(2605),
    [sym_compiler_directive_decl] = STATE(2605),
    [sym_fsi_directive_decl] = STATE(2605),
    [sym_preproc_line] = STATE(2605),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_module] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_open] = ACTIONS(6959),
    [anon_sym_type] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_and] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_exception] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [anon_sym_extern] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6957),
    [anon_sym_POUNDload] = ACTIONS(6957),
    [anon_sym_POUNDtime] = ACTIONS(6957),
    [anon_sym_POUNDI] = ACTIONS(6957),
    [anon_sym_POUNDhelp] = ACTIONS(6957),
    [anon_sym_POUNDquit] = ACTIONS(6957),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [anon_sym_POUNDelse] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(2606)] = {
    [sym_block_comment] = STATE(2606),
    [sym_compiler_directive_decl] = STATE(2606),
    [sym_fsi_directive_decl] = STATE(2606),
    [sym_preproc_line] = STATE(2606),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_module] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_open] = ACTIONS(6963),
    [anon_sym_type] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_and] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_exception] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [anon_sym_extern] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6961),
    [anon_sym_POUNDload] = ACTIONS(6961),
    [anon_sym_POUNDtime] = ACTIONS(6961),
    [anon_sym_POUNDI] = ACTIONS(6961),
    [anon_sym_POUNDhelp] = ACTIONS(6961),
    [anon_sym_POUNDquit] = ACTIONS(6961),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [anon_sym_POUNDelse] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(2607)] = {
    [sym_block_comment] = STATE(2607),
    [sym_compiler_directive_decl] = STATE(2607),
    [sym_fsi_directive_decl] = STATE(2607),
    [sym_preproc_line] = STATE(2607),
    [aux_sym_long_identifier_repeat1] = STATE(2611),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7191),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_inherit] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2608)] = {
    [sym_block_comment] = STATE(2608),
    [sym_compiler_directive_decl] = STATE(2608),
    [sym_fsi_directive_decl] = STATE(2608),
    [sym_preproc_line] = STATE(2608),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_module] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_open] = ACTIONS(6828),
    [anon_sym_type] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_and] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_exception] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6830),
    [anon_sym_POUNDload] = ACTIONS(6830),
    [anon_sym_POUNDtime] = ACTIONS(6830),
    [anon_sym_POUNDI] = ACTIONS(6830),
    [anon_sym_POUNDhelp] = ACTIONS(6830),
    [anon_sym_POUNDquit] = ACTIONS(6830),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(2609)] = {
    [sym_block_comment] = STATE(2609),
    [sym_compiler_directive_decl] = STATE(2609),
    [sym_fsi_directive_decl] = STATE(2609),
    [sym_preproc_line] = STATE(2609),
    [ts_builtin_sym_end] = ACTIONS(7079),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_namespace] = ACTIONS(7077),
    [anon_sym_module] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_open] = ACTIONS(7077),
    [anon_sym_type] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_and] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_exception] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [anon_sym_extern] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7079),
    [anon_sym_POUNDload] = ACTIONS(7079),
    [anon_sym_POUNDtime] = ACTIONS(7079),
    [anon_sym_POUNDI] = ACTIONS(7079),
    [anon_sym_POUNDhelp] = ACTIONS(7079),
    [anon_sym_POUNDquit] = ACTIONS(7079),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(2610)] = {
    [sym_block_comment] = STATE(2610),
    [sym_compiler_directive_decl] = STATE(2610),
    [sym_fsi_directive_decl] = STATE(2610),
    [sym_preproc_line] = STATE(2610),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_module] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_open] = ACTIONS(6967),
    [anon_sym_type] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_and] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_exception] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [anon_sym_extern] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6965),
    [anon_sym_POUNDload] = ACTIONS(6965),
    [anon_sym_POUNDtime] = ACTIONS(6965),
    [anon_sym_POUNDI] = ACTIONS(6965),
    [anon_sym_POUNDhelp] = ACTIONS(6965),
    [anon_sym_POUNDquit] = ACTIONS(6965),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [anon_sym_POUNDelse] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(2611)] = {
    [sym_block_comment] = STATE(2611),
    [sym_compiler_directive_decl] = STATE(2611),
    [sym_fsi_directive_decl] = STATE(2611),
    [sym_preproc_line] = STATE(2611),
    [aux_sym_long_identifier_repeat1] = STATE(2611),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2612)] = {
    [sym_block_comment] = STATE(2612),
    [sym_compiler_directive_decl] = STATE(2612),
    [sym_fsi_directive_decl] = STATE(2612),
    [sym_preproc_line] = STATE(2612),
    [ts_builtin_sym_end] = ACTIONS(7087),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_namespace] = ACTIONS(7085),
    [anon_sym_module] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_open] = ACTIONS(7085),
    [anon_sym_type] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_and] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_exception] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [anon_sym_extern] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7087),
    [anon_sym_POUNDload] = ACTIONS(7087),
    [anon_sym_POUNDtime] = ACTIONS(7087),
    [anon_sym_POUNDI] = ACTIONS(7087),
    [anon_sym_POUNDhelp] = ACTIONS(7087),
    [anon_sym_POUNDquit] = ACTIONS(7087),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(2613)] = {
    [sym_block_comment] = STATE(2613),
    [sym_compiler_directive_decl] = STATE(2613),
    [sym_fsi_directive_decl] = STATE(2613),
    [sym_preproc_line] = STATE(2613),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_GT_RBRACK] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_RBRACK] = ACTIONS(5953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_RBRACE] = ACTIONS(5953),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_to] = ACTIONS(5951),
    [anon_sym_downto] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6696),
    [anon_sym_uy] = ACTIONS(6698),
    [anon_sym_s] = ACTIONS(6700),
    [anon_sym_us] = ACTIONS(6702),
    [anon_sym_l] = ACTIONS(6704),
    [aux_sym_uint32_token1] = ACTIONS(6706),
    [anon_sym_n] = ACTIONS(6708),
    [anon_sym_un] = ACTIONS(6710),
    [anon_sym_L] = ACTIONS(6712),
    [aux_sym_uint64_token1] = ACTIONS(6714),
    [anon_sym_lf] = ACTIONS(7198),
    [anon_sym_LF] = ACTIONS(7200),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2614)] = {
    [sym_block_comment] = STATE(2614),
    [sym_compiler_directive_decl] = STATE(2614),
    [sym_fsi_directive_decl] = STATE(2614),
    [sym_preproc_line] = STATE(2614),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_module] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_open] = ACTIONS(6971),
    [anon_sym_type] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_and] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_exception] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [anon_sym_extern] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6969),
    [anon_sym_POUNDload] = ACTIONS(6969),
    [anon_sym_POUNDtime] = ACTIONS(6969),
    [anon_sym_POUNDI] = ACTIONS(6969),
    [anon_sym_POUNDhelp] = ACTIONS(6969),
    [anon_sym_POUNDquit] = ACTIONS(6969),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [anon_sym_POUNDelse] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(2615)] = {
    [sym_block_comment] = STATE(2615),
    [sym_compiler_directive_decl] = STATE(2615),
    [sym_fsi_directive_decl] = STATE(2615),
    [sym_preproc_line] = STATE(2615),
    [ts_builtin_sym_end] = ACTIONS(7143),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_namespace] = ACTIONS(7141),
    [anon_sym_module] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_open] = ACTIONS(7141),
    [anon_sym_type] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_and] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_exception] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7143),
    [anon_sym_POUNDload] = ACTIONS(7143),
    [anon_sym_POUNDtime] = ACTIONS(7143),
    [anon_sym_POUNDI] = ACTIONS(7143),
    [anon_sym_POUNDhelp] = ACTIONS(7143),
    [anon_sym_POUNDquit] = ACTIONS(7143),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(2616)] = {
    [sym_block_comment] = STATE(2616),
    [sym_compiler_directive_decl] = STATE(2616),
    [sym_fsi_directive_decl] = STATE(2616),
    [sym_preproc_line] = STATE(2616),
    [ts_builtin_sym_end] = ACTIONS(7147),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_namespace] = ACTIONS(7145),
    [anon_sym_module] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_open] = ACTIONS(7145),
    [anon_sym_type] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_and] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_exception] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7147),
    [anon_sym_POUNDload] = ACTIONS(7147),
    [anon_sym_POUNDtime] = ACTIONS(7147),
    [anon_sym_POUNDI] = ACTIONS(7147),
    [anon_sym_POUNDhelp] = ACTIONS(7147),
    [anon_sym_POUNDquit] = ACTIONS(7147),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(2617)] = {
    [sym_block_comment] = STATE(2617),
    [sym_compiler_directive_decl] = STATE(2617),
    [sym_fsi_directive_decl] = STATE(2617),
    [sym_preproc_line] = STATE(2617),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_module] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_open] = ACTIONS(6975),
    [anon_sym_type] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_and] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_exception] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [anon_sym_extern] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6973),
    [anon_sym_POUNDload] = ACTIONS(6973),
    [anon_sym_POUNDtime] = ACTIONS(6973),
    [anon_sym_POUNDI] = ACTIONS(6973),
    [anon_sym_POUNDhelp] = ACTIONS(6973),
    [anon_sym_POUNDquit] = ACTIONS(6973),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [anon_sym_POUNDelse] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(2618)] = {
    [sym_block_comment] = STATE(2618),
    [sym_compiler_directive_decl] = STATE(2618),
    [sym_fsi_directive_decl] = STATE(2618),
    [sym_preproc_line] = STATE(2618),
    [ts_builtin_sym_end] = ACTIONS(7157),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_namespace] = ACTIONS(7155),
    [anon_sym_module] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_open] = ACTIONS(7155),
    [anon_sym_type] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_and] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_exception] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [anon_sym_extern] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7157),
    [anon_sym_POUNDload] = ACTIONS(7157),
    [anon_sym_POUNDtime] = ACTIONS(7157),
    [anon_sym_POUNDI] = ACTIONS(7157),
    [anon_sym_POUNDhelp] = ACTIONS(7157),
    [anon_sym_POUNDquit] = ACTIONS(7157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(2619)] = {
    [sym_block_comment] = STATE(2619),
    [sym_compiler_directive_decl] = STATE(2619),
    [sym_fsi_directive_decl] = STATE(2619),
    [sym_preproc_line] = STATE(2619),
    [aux_sym_long_identifier_repeat1] = STATE(2620),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2620)] = {
    [sym_block_comment] = STATE(2620),
    [sym_compiler_directive_decl] = STATE(2620),
    [sym_fsi_directive_decl] = STATE(2620),
    [sym_preproc_line] = STATE(2620),
    [aux_sym_long_identifier_repeat1] = STATE(2620),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7202),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2621)] = {
    [sym_block_comment] = STATE(2621),
    [sym_compiler_directive_decl] = STATE(2621),
    [sym_fsi_directive_decl] = STATE(2621),
    [sym_preproc_line] = STATE(2621),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_module] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_open] = ACTIONS(6979),
    [anon_sym_type] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_and] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_exception] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [anon_sym_extern] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6977),
    [anon_sym_POUNDload] = ACTIONS(6977),
    [anon_sym_POUNDtime] = ACTIONS(6977),
    [anon_sym_POUNDI] = ACTIONS(6977),
    [anon_sym_POUNDhelp] = ACTIONS(6977),
    [anon_sym_POUNDquit] = ACTIONS(6977),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [anon_sym_POUNDelse] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(2622)] = {
    [sym_block_comment] = STATE(2622),
    [sym_compiler_directive_decl] = STATE(2622),
    [sym_fsi_directive_decl] = STATE(2622),
    [sym_preproc_line] = STATE(2622),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_module] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_open] = ACTIONS(6983),
    [anon_sym_type] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_and] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_exception] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [anon_sym_extern] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6981),
    [anon_sym_POUNDload] = ACTIONS(6981),
    [anon_sym_POUNDtime] = ACTIONS(6981),
    [anon_sym_POUNDI] = ACTIONS(6981),
    [anon_sym_POUNDhelp] = ACTIONS(6981),
    [anon_sym_POUNDquit] = ACTIONS(6981),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [anon_sym_POUNDelse] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(2623)] = {
    [sym_block_comment] = STATE(2623),
    [sym_compiler_directive_decl] = STATE(2623),
    [sym_fsi_directive_decl] = STATE(2623),
    [sym_preproc_line] = STATE(2623),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_module] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_open] = ACTIONS(6987),
    [anon_sym_type] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_and] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_exception] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [anon_sym_extern] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6985),
    [anon_sym_POUNDload] = ACTIONS(6985),
    [anon_sym_POUNDtime] = ACTIONS(6985),
    [anon_sym_POUNDI] = ACTIONS(6985),
    [anon_sym_POUNDhelp] = ACTIONS(6985),
    [anon_sym_POUNDquit] = ACTIONS(6985),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [anon_sym_POUNDelse] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(2624)] = {
    [sym_block_comment] = STATE(2624),
    [sym_compiler_directive_decl] = STATE(2624),
    [sym_fsi_directive_decl] = STATE(2624),
    [sym_preproc_line] = STATE(2624),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_module] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_open] = ACTIONS(6991),
    [anon_sym_type] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_and] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_exception] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [anon_sym_extern] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6989),
    [anon_sym_POUNDload] = ACTIONS(6989),
    [anon_sym_POUNDtime] = ACTIONS(6989),
    [anon_sym_POUNDI] = ACTIONS(6989),
    [anon_sym_POUNDhelp] = ACTIONS(6989),
    [anon_sym_POUNDquit] = ACTIONS(6989),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [anon_sym_POUNDelse] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(2625)] = {
    [sym_block_comment] = STATE(2625),
    [sym_compiler_directive_decl] = STATE(2625),
    [sym_fsi_directive_decl] = STATE(2625),
    [sym_preproc_line] = STATE(2625),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_module] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_open] = ACTIONS(6995),
    [anon_sym_type] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_and] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_exception] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [anon_sym_extern] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6993),
    [anon_sym_POUNDload] = ACTIONS(6993),
    [anon_sym_POUNDtime] = ACTIONS(6993),
    [anon_sym_POUNDI] = ACTIONS(6993),
    [anon_sym_POUNDhelp] = ACTIONS(6993),
    [anon_sym_POUNDquit] = ACTIONS(6993),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [anon_sym_POUNDelse] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(2626)] = {
    [sym_block_comment] = STATE(2626),
    [sym_compiler_directive_decl] = STATE(2626),
    [sym_fsi_directive_decl] = STATE(2626),
    [sym_preproc_line] = STATE(2626),
    [ts_builtin_sym_end] = ACTIONS(7161),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_namespace] = ACTIONS(7159),
    [anon_sym_module] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_open] = ACTIONS(7159),
    [anon_sym_type] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_and] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_exception] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [anon_sym_extern] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7161),
    [anon_sym_POUNDload] = ACTIONS(7161),
    [anon_sym_POUNDtime] = ACTIONS(7161),
    [anon_sym_POUNDI] = ACTIONS(7161),
    [anon_sym_POUNDhelp] = ACTIONS(7161),
    [anon_sym_POUNDquit] = ACTIONS(7161),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(2627)] = {
    [sym_block_comment] = STATE(2627),
    [sym_compiler_directive_decl] = STATE(2627),
    [sym_fsi_directive_decl] = STATE(2627),
    [sym_preproc_line] = STATE(2627),
    [ts_builtin_sym_end] = ACTIONS(7091),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_namespace] = ACTIONS(7089),
    [anon_sym_module] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_open] = ACTIONS(7089),
    [anon_sym_type] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_and] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_exception] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [anon_sym_extern] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7091),
    [anon_sym_POUNDload] = ACTIONS(7091),
    [anon_sym_POUNDtime] = ACTIONS(7091),
    [anon_sym_POUNDI] = ACTIONS(7091),
    [anon_sym_POUNDhelp] = ACTIONS(7091),
    [anon_sym_POUNDquit] = ACTIONS(7091),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(2628)] = {
    [sym_block_comment] = STATE(2628),
    [sym_compiler_directive_decl] = STATE(2628),
    [sym_fsi_directive_decl] = STATE(2628),
    [sym_preproc_line] = STATE(2628),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_module] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_open] = ACTIONS(6999),
    [anon_sym_type] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_and] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_exception] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [anon_sym_extern] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6997),
    [anon_sym_POUNDload] = ACTIONS(6997),
    [anon_sym_POUNDtime] = ACTIONS(6997),
    [anon_sym_POUNDI] = ACTIONS(6997),
    [anon_sym_POUNDhelp] = ACTIONS(6997),
    [anon_sym_POUNDquit] = ACTIONS(6997),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [anon_sym_POUNDelse] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(2629)] = {
    [sym_block_comment] = STATE(2629),
    [sym_compiler_directive_decl] = STATE(2629),
    [sym_fsi_directive_decl] = STATE(2629),
    [sym_preproc_line] = STATE(2629),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_module] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_open] = ACTIONS(7003),
    [anon_sym_type] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_and] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_exception] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [anon_sym_extern] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7001),
    [anon_sym_POUNDload] = ACTIONS(7001),
    [anon_sym_POUNDtime] = ACTIONS(7001),
    [anon_sym_POUNDI] = ACTIONS(7001),
    [anon_sym_POUNDhelp] = ACTIONS(7001),
    [anon_sym_POUNDquit] = ACTIONS(7001),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [anon_sym_POUNDelse] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(2630)] = {
    [sym_block_comment] = STATE(2630),
    [sym_compiler_directive_decl] = STATE(2630),
    [sym_fsi_directive_decl] = STATE(2630),
    [sym_preproc_line] = STATE(2630),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_module] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_open] = ACTIONS(7007),
    [anon_sym_type] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_and] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_exception] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [anon_sym_extern] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7005),
    [anon_sym_POUNDload] = ACTIONS(7005),
    [anon_sym_POUNDtime] = ACTIONS(7005),
    [anon_sym_POUNDI] = ACTIONS(7005),
    [anon_sym_POUNDhelp] = ACTIONS(7005),
    [anon_sym_POUNDquit] = ACTIONS(7005),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [anon_sym_POUNDelse] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(2631)] = {
    [sym_block_comment] = STATE(2631),
    [sym_compiler_directive_decl] = STATE(2631),
    [sym_fsi_directive_decl] = STATE(2631),
    [sym_preproc_line] = STATE(2631),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2632)] = {
    [sym_block_comment] = STATE(2632),
    [sym_compiler_directive_decl] = STATE(2632),
    [sym_fsi_directive_decl] = STATE(2632),
    [sym_preproc_line] = STATE(2632),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_module] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_open] = ACTIONS(7011),
    [anon_sym_type] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_and] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_exception] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [anon_sym_extern] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7009),
    [anon_sym_POUNDload] = ACTIONS(7009),
    [anon_sym_POUNDtime] = ACTIONS(7009),
    [anon_sym_POUNDI] = ACTIONS(7009),
    [anon_sym_POUNDhelp] = ACTIONS(7009),
    [anon_sym_POUNDquit] = ACTIONS(7009),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [anon_sym_POUNDelse] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(2633)] = {
    [sym_block_comment] = STATE(2633),
    [sym_compiler_directive_decl] = STATE(2633),
    [sym_fsi_directive_decl] = STATE(2633),
    [sym_preproc_line] = STATE(2633),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_module] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_open] = ACTIONS(7019),
    [anon_sym_type] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_and] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_exception] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [anon_sym_extern] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7017),
    [anon_sym_POUNDload] = ACTIONS(7017),
    [anon_sym_POUNDtime] = ACTIONS(7017),
    [anon_sym_POUNDI] = ACTIONS(7017),
    [anon_sym_POUNDhelp] = ACTIONS(7017),
    [anon_sym_POUNDquit] = ACTIONS(7017),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [anon_sym_POUNDelse] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(2634)] = {
    [sym_block_comment] = STATE(2634),
    [sym_compiler_directive_decl] = STATE(2634),
    [sym_fsi_directive_decl] = STATE(2634),
    [sym_preproc_line] = STATE(2634),
    [ts_builtin_sym_end] = ACTIONS(7095),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_namespace] = ACTIONS(7093),
    [anon_sym_module] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_open] = ACTIONS(7093),
    [anon_sym_type] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_and] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_exception] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [anon_sym_extern] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7095),
    [anon_sym_POUNDload] = ACTIONS(7095),
    [anon_sym_POUNDtime] = ACTIONS(7095),
    [anon_sym_POUNDI] = ACTIONS(7095),
    [anon_sym_POUNDhelp] = ACTIONS(7095),
    [anon_sym_POUNDquit] = ACTIONS(7095),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(2635)] = {
    [sym_block_comment] = STATE(2635),
    [sym_compiler_directive_decl] = STATE(2635),
    [sym_fsi_directive_decl] = STATE(2635),
    [sym_preproc_line] = STATE(2635),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_module] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_open] = ACTIONS(7023),
    [anon_sym_type] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_and] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_exception] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [anon_sym_extern] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7021),
    [anon_sym_POUNDload] = ACTIONS(7021),
    [anon_sym_POUNDtime] = ACTIONS(7021),
    [anon_sym_POUNDI] = ACTIONS(7021),
    [anon_sym_POUNDhelp] = ACTIONS(7021),
    [anon_sym_POUNDquit] = ACTIONS(7021),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [anon_sym_POUNDelse] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(2636)] = {
    [sym_block_comment] = STATE(2636),
    [sym_compiler_directive_decl] = STATE(2636),
    [sym_fsi_directive_decl] = STATE(2636),
    [sym_preproc_line] = STATE(2636),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_module] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_open] = ACTIONS(7027),
    [anon_sym_type] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_and] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_exception] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [anon_sym_extern] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7025),
    [anon_sym_POUNDload] = ACTIONS(7025),
    [anon_sym_POUNDtime] = ACTIONS(7025),
    [anon_sym_POUNDI] = ACTIONS(7025),
    [anon_sym_POUNDhelp] = ACTIONS(7025),
    [anon_sym_POUNDquit] = ACTIONS(7025),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [anon_sym_POUNDelse] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(2637)] = {
    [sym_block_comment] = STATE(2637),
    [sym_compiler_directive_decl] = STATE(2637),
    [sym_fsi_directive_decl] = STATE(2637),
    [sym_preproc_line] = STATE(2637),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_module] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_open] = ACTIONS(7035),
    [anon_sym_type] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_and] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_exception] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [anon_sym_extern] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7033),
    [anon_sym_POUNDload] = ACTIONS(7033),
    [anon_sym_POUNDtime] = ACTIONS(7033),
    [anon_sym_POUNDI] = ACTIONS(7033),
    [anon_sym_POUNDhelp] = ACTIONS(7033),
    [anon_sym_POUNDquit] = ACTIONS(7033),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [anon_sym_POUNDelse] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(2638)] = {
    [sym__else_expression] = STATE(3236),
    [sym_elif_expression] = STATE(2938),
    [sym_block_comment] = STATE(2638),
    [sym_compiler_directive_decl] = STATE(2638),
    [sym_fsi_directive_decl] = STATE(2638),
    [sym_preproc_line] = STATE(2638),
    [aux_sym_if_expression_repeat1] = STATE(2650),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_and] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7205),
    [anon_sym_elif] = ACTIONS(7207),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [anon_sym_POUNDelse] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2639)] = {
    [sym_block_comment] = STATE(2639),
    [sym_compiler_directive_decl] = STATE(2639),
    [sym_fsi_directive_decl] = STATE(2639),
    [sym_preproc_line] = STATE(2639),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_module] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_open] = ACTIONS(7039),
    [anon_sym_type] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_and] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_exception] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [anon_sym_extern] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7037),
    [anon_sym_POUNDload] = ACTIONS(7037),
    [anon_sym_POUNDtime] = ACTIONS(7037),
    [anon_sym_POUNDI] = ACTIONS(7037),
    [anon_sym_POUNDhelp] = ACTIONS(7037),
    [anon_sym_POUNDquit] = ACTIONS(7037),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [anon_sym_POUNDelse] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(2640)] = {
    [sym_block_comment] = STATE(2640),
    [sym_compiler_directive_decl] = STATE(2640),
    [sym_fsi_directive_decl] = STATE(2640),
    [sym_preproc_line] = STATE(2640),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_module] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_open] = ACTIONS(7043),
    [anon_sym_type] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_and] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_exception] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [anon_sym_extern] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7041),
    [anon_sym_POUNDload] = ACTIONS(7041),
    [anon_sym_POUNDtime] = ACTIONS(7041),
    [anon_sym_POUNDI] = ACTIONS(7041),
    [anon_sym_POUNDhelp] = ACTIONS(7041),
    [anon_sym_POUNDquit] = ACTIONS(7041),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [anon_sym_POUNDelse] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(2641)] = {
    [sym_block_comment] = STATE(2641),
    [sym_compiler_directive_decl] = STATE(2641),
    [sym_fsi_directive_decl] = STATE(2641),
    [sym_preproc_line] = STATE(2641),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_module] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_open] = ACTIONS(7059),
    [anon_sym_type] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_and] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_exception] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [anon_sym_extern] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7057),
    [anon_sym_POUNDload] = ACTIONS(7057),
    [anon_sym_POUNDtime] = ACTIONS(7057),
    [anon_sym_POUNDI] = ACTIONS(7057),
    [anon_sym_POUNDhelp] = ACTIONS(7057),
    [anon_sym_POUNDquit] = ACTIONS(7057),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [anon_sym_POUNDelse] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(2642)] = {
    [sym_block_comment] = STATE(2642),
    [sym_compiler_directive_decl] = STATE(2642),
    [sym_fsi_directive_decl] = STATE(2642),
    [sym_preproc_line] = STATE(2642),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_module] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_open] = ACTIONS(6750),
    [anon_sym_type] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_and] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_exception] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6752),
    [anon_sym_POUNDload] = ACTIONS(6752),
    [anon_sym_POUNDtime] = ACTIONS(6752),
    [anon_sym_POUNDI] = ACTIONS(6752),
    [anon_sym_POUNDhelp] = ACTIONS(6752),
    [anon_sym_POUNDquit] = ACTIONS(6752),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(2643)] = {
    [sym_block_comment] = STATE(2643),
    [sym_compiler_directive_decl] = STATE(2643),
    [sym_fsi_directive_decl] = STATE(2643),
    [sym_preproc_line] = STATE(2643),
    [ts_builtin_sym_end] = ACTIONS(7099),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_namespace] = ACTIONS(7097),
    [anon_sym_module] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_open] = ACTIONS(7097),
    [anon_sym_type] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_and] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_exception] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [anon_sym_extern] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7099),
    [anon_sym_POUNDload] = ACTIONS(7099),
    [anon_sym_POUNDtime] = ACTIONS(7099),
    [anon_sym_POUNDI] = ACTIONS(7099),
    [anon_sym_POUNDhelp] = ACTIONS(7099),
    [anon_sym_POUNDquit] = ACTIONS(7099),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(2644)] = {
    [sym_block_comment] = STATE(2644),
    [sym_compiler_directive_decl] = STATE(2644),
    [sym_fsi_directive_decl] = STATE(2644),
    [sym_preproc_line] = STATE(2644),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_module] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_open] = ACTIONS(6784),
    [anon_sym_type] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_and] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(7209),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_exception] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6786),
    [anon_sym_POUNDload] = ACTIONS(6786),
    [anon_sym_POUNDtime] = ACTIONS(6786),
    [anon_sym_POUNDI] = ACTIONS(6786),
    [anon_sym_POUNDhelp] = ACTIONS(6786),
    [anon_sym_POUNDquit] = ACTIONS(6786),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(2645)] = {
    [sym_block_comment] = STATE(2645),
    [sym_compiler_directive_decl] = STATE(2645),
    [sym_fsi_directive_decl] = STATE(2645),
    [sym_preproc_line] = STATE(2645),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_module] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_open] = ACTIONS(6790),
    [anon_sym_type] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_and] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_exception] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6792),
    [anon_sym_POUNDload] = ACTIONS(6792),
    [anon_sym_POUNDtime] = ACTIONS(6792),
    [anon_sym_POUNDI] = ACTIONS(6792),
    [anon_sym_POUNDhelp] = ACTIONS(6792),
    [anon_sym_POUNDquit] = ACTIONS(6792),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(2646)] = {
    [sym_block_comment] = STATE(2646),
    [sym_compiler_directive_decl] = STATE(2646),
    [sym_fsi_directive_decl] = STATE(2646),
    [sym_preproc_line] = STATE(2646),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_module] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_open] = ACTIONS(6794),
    [anon_sym_type] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_and] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_exception] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6796),
    [anon_sym_POUNDload] = ACTIONS(6796),
    [anon_sym_POUNDtime] = ACTIONS(6796),
    [anon_sym_POUNDI] = ACTIONS(6796),
    [anon_sym_POUNDhelp] = ACTIONS(6796),
    [anon_sym_POUNDquit] = ACTIONS(6796),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(2647)] = {
    [sym_block_comment] = STATE(2647),
    [sym_compiler_directive_decl] = STATE(2647),
    [sym_fsi_directive_decl] = STATE(2647),
    [sym_preproc_line] = STATE(2647),
    [ts_builtin_sym_end] = ACTIONS(7103),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_namespace] = ACTIONS(7101),
    [anon_sym_module] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_open] = ACTIONS(7101),
    [anon_sym_type] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_and] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_exception] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7103),
    [anon_sym_POUNDload] = ACTIONS(7103),
    [anon_sym_POUNDtime] = ACTIONS(7103),
    [anon_sym_POUNDI] = ACTIONS(7103),
    [anon_sym_POUNDhelp] = ACTIONS(7103),
    [anon_sym_POUNDquit] = ACTIONS(7103),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(2648)] = {
    [sym_block_comment] = STATE(2648),
    [sym_compiler_directive_decl] = STATE(2648),
    [sym_fsi_directive_decl] = STATE(2648),
    [sym_preproc_line] = STATE(2648),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(6764),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2649)] = {
    [sym_block_comment] = STATE(2649),
    [sym_compiler_directive_decl] = STATE(2649),
    [sym_fsi_directive_decl] = STATE(2649),
    [sym_preproc_line] = STATE(2649),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_module] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_open] = ACTIONS(6768),
    [anon_sym_type] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_and] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(7211),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_exception] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [anon_sym_extern] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6766),
    [anon_sym_POUNDload] = ACTIONS(6766),
    [anon_sym_POUNDtime] = ACTIONS(6766),
    [anon_sym_POUNDI] = ACTIONS(6766),
    [anon_sym_POUNDhelp] = ACTIONS(6766),
    [anon_sym_POUNDquit] = ACTIONS(6766),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(2650)] = {
    [sym__else_expression] = STATE(3265),
    [sym_elif_expression] = STATE(2938),
    [sym_block_comment] = STATE(2650),
    [sym_compiler_directive_decl] = STATE(2650),
    [sym_fsi_directive_decl] = STATE(2650),
    [sym_preproc_line] = STATE(2650),
    [aux_sym_if_expression_repeat1] = STATE(2814),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_and] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7205),
    [anon_sym_elif] = ACTIONS(7207),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [anon_sym_POUNDelse] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2651)] = {
    [sym_block_comment] = STATE(2651),
    [sym_compiler_directive_decl] = STATE(2651),
    [sym_fsi_directive_decl] = STATE(2651),
    [sym_preproc_line] = STATE(2651),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_module] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_open] = ACTIONS(6840),
    [anon_sym_type] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_and] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_exception] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [anon_sym_extern] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6838),
    [anon_sym_POUNDload] = ACTIONS(6838),
    [anon_sym_POUNDtime] = ACTIONS(6838),
    [anon_sym_POUNDI] = ACTIONS(6838),
    [anon_sym_POUNDhelp] = ACTIONS(6838),
    [anon_sym_POUNDquit] = ACTIONS(6838),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(2652)] = {
    [sym_block_comment] = STATE(2652),
    [sym_compiler_directive_decl] = STATE(2652),
    [sym_fsi_directive_decl] = STATE(2652),
    [sym_preproc_line] = STATE(2652),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2653)] = {
    [sym_block_comment] = STATE(2653),
    [sym_compiler_directive_decl] = STATE(2653),
    [sym_fsi_directive_decl] = STATE(2653),
    [sym_preproc_line] = STATE(2653),
    [ts_builtin_sym_end] = ACTIONS(7107),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_namespace] = ACTIONS(7105),
    [anon_sym_module] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_open] = ACTIONS(7105),
    [anon_sym_type] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_and] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_exception] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7107),
    [anon_sym_POUNDload] = ACTIONS(7107),
    [anon_sym_POUNDtime] = ACTIONS(7107),
    [anon_sym_POUNDI] = ACTIONS(7107),
    [anon_sym_POUNDhelp] = ACTIONS(7107),
    [anon_sym_POUNDquit] = ACTIONS(7107),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(2654)] = {
    [sym_block_comment] = STATE(2654),
    [sym_compiler_directive_decl] = STATE(2654),
    [sym_fsi_directive_decl] = STATE(2654),
    [sym_preproc_line] = STATE(2654),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_module] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_open] = ACTIONS(6813),
    [anon_sym_type] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_and] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_exception] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [anon_sym_extern] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6811),
    [anon_sym_POUNDload] = ACTIONS(6811),
    [anon_sym_POUNDtime] = ACTIONS(6811),
    [anon_sym_POUNDI] = ACTIONS(6811),
    [anon_sym_POUNDhelp] = ACTIONS(6811),
    [anon_sym_POUNDquit] = ACTIONS(6811),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(2655)] = {
    [sym_block_comment] = STATE(2655),
    [sym_compiler_directive_decl] = STATE(2655),
    [sym_fsi_directive_decl] = STATE(2655),
    [sym_preproc_line] = STATE(2655),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_module] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_open] = ACTIONS(6824),
    [anon_sym_type] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_and] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_exception] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6822),
    [anon_sym_POUNDload] = ACTIONS(6822),
    [anon_sym_POUNDtime] = ACTIONS(6822),
    [anon_sym_POUNDI] = ACTIONS(6822),
    [anon_sym_POUNDhelp] = ACTIONS(6822),
    [anon_sym_POUNDquit] = ACTIONS(6822),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(2656)] = {
    [sym_block_comment] = STATE(2656),
    [sym_compiler_directive_decl] = STATE(2656),
    [sym_fsi_directive_decl] = STATE(2656),
    [sym_preproc_line] = STATE(2656),
    [ts_builtin_sym_end] = ACTIONS(7111),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_namespace] = ACTIONS(7109),
    [anon_sym_module] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_open] = ACTIONS(7109),
    [anon_sym_type] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_and] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_exception] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7111),
    [anon_sym_POUNDload] = ACTIONS(7111),
    [anon_sym_POUNDtime] = ACTIONS(7111),
    [anon_sym_POUNDI] = ACTIONS(7111),
    [anon_sym_POUNDhelp] = ACTIONS(7111),
    [anon_sym_POUNDquit] = ACTIONS(7111),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(2657)] = {
    [sym_block_comment] = STATE(2657),
    [sym_compiler_directive_decl] = STATE(2657),
    [sym_fsi_directive_decl] = STATE(2657),
    [sym_preproc_line] = STATE(2657),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(2658)] = {
    [sym_block_comment] = STATE(2658),
    [sym_compiler_directive_decl] = STATE(2658),
    [sym_fsi_directive_decl] = STATE(2658),
    [sym_preproc_line] = STATE(2658),
    [ts_builtin_sym_end] = ACTIONS(7115),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_namespace] = ACTIONS(7113),
    [anon_sym_module] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_open] = ACTIONS(7113),
    [anon_sym_type] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_and] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_exception] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7115),
    [anon_sym_POUNDload] = ACTIONS(7115),
    [anon_sym_POUNDtime] = ACTIONS(7115),
    [anon_sym_POUNDI] = ACTIONS(7115),
    [anon_sym_POUNDhelp] = ACTIONS(7115),
    [anon_sym_POUNDquit] = ACTIONS(7115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(2659)] = {
    [sym_block_comment] = STATE(2659),
    [sym_compiler_directive_decl] = STATE(2659),
    [sym_fsi_directive_decl] = STATE(2659),
    [sym_preproc_line] = STATE(2659),
    [aux_sym_sequential_expression_repeat1] = STATE(2659),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(7213),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(2660)] = {
    [sym_block_comment] = STATE(2660),
    [sym_compiler_directive_decl] = STATE(2660),
    [sym_fsi_directive_decl] = STATE(2660),
    [sym_preproc_line] = STATE(2660),
    [ts_builtin_sym_end] = ACTIONS(7119),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_namespace] = ACTIONS(7117),
    [anon_sym_module] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_open] = ACTIONS(7117),
    [anon_sym_type] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_and] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_exception] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7119),
    [anon_sym_POUNDload] = ACTIONS(7119),
    [anon_sym_POUNDtime] = ACTIONS(7119),
    [anon_sym_POUNDI] = ACTIONS(7119),
    [anon_sym_POUNDhelp] = ACTIONS(7119),
    [anon_sym_POUNDquit] = ACTIONS(7119),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(2661)] = {
    [sym_block_comment] = STATE(2661),
    [sym_compiler_directive_decl] = STATE(2661),
    [sym_fsi_directive_decl] = STATE(2661),
    [sym_preproc_line] = STATE(2661),
    [ts_builtin_sym_end] = ACTIONS(7123),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_namespace] = ACTIONS(7121),
    [anon_sym_module] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_open] = ACTIONS(7121),
    [anon_sym_type] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_and] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_exception] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7123),
    [anon_sym_POUNDload] = ACTIONS(7123),
    [anon_sym_POUNDtime] = ACTIONS(7123),
    [anon_sym_POUNDI] = ACTIONS(7123),
    [anon_sym_POUNDhelp] = ACTIONS(7123),
    [anon_sym_POUNDquit] = ACTIONS(7123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(2662)] = {
    [sym_block_comment] = STATE(2662),
    [sym_compiler_directive_decl] = STATE(2662),
    [sym_fsi_directive_decl] = STATE(2662),
    [sym_preproc_line] = STATE(2662),
    [aux_sym_sequential_expression_repeat1] = STATE(2662),
    [ts_builtin_sym_end] = ACTIONS(6798),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(7216),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(2663)] = {
    [sym_block_comment] = STATE(2663),
    [sym_compiler_directive_decl] = STATE(2663),
    [sym_fsi_directive_decl] = STATE(2663),
    [sym_preproc_line] = STATE(2663),
    [aux_sym_sequential_expression_repeat1] = STATE(2659),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(2664)] = {
    [sym_block_comment] = STATE(2664),
    [sym_compiler_directive_decl] = STATE(2664),
    [sym_fsi_directive_decl] = STATE(2664),
    [sym_preproc_line] = STATE(2664),
    [aux_sym_type_argument_repeat1] = STATE(2584),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(7219),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_inherit] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2665)] = {
    [sym_block_comment] = STATE(2665),
    [sym_compiler_directive_decl] = STATE(2665),
    [sym_fsi_directive_decl] = STATE(2665),
    [sym_preproc_line] = STATE(2665),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(7222),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(2666)] = {
    [sym_block_comment] = STATE(2666),
    [sym_compiler_directive_decl] = STATE(2666),
    [sym_fsi_directive_decl] = STATE(2666),
    [sym_preproc_line] = STATE(2666),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(2667)] = {
    [sym_block_comment] = STATE(2667),
    [sym_compiler_directive_decl] = STATE(2667),
    [sym_fsi_directive_decl] = STATE(2667),
    [sym_preproc_line] = STATE(2667),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_module] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_open] = ACTIONS(6863),
    [anon_sym_type] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_and] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_exception] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [anon_sym_extern] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6861),
    [anon_sym_POUNDload] = ACTIONS(6861),
    [anon_sym_POUNDtime] = ACTIONS(6861),
    [anon_sym_POUNDI] = ACTIONS(6861),
    [anon_sym_POUNDhelp] = ACTIONS(6861),
    [anon_sym_POUNDquit] = ACTIONS(6861),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(2668)] = {
    [sym_block_comment] = STATE(2668),
    [sym_compiler_directive_decl] = STATE(2668),
    [sym_fsi_directive_decl] = STATE(2668),
    [sym_preproc_line] = STATE(2668),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [anon_sym_inherit] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [anon_sym_POUNDelse] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2669)] = {
    [sym_block_comment] = STATE(2669),
    [sym_compiler_directive_decl] = STATE(2669),
    [sym_fsi_directive_decl] = STATE(2669),
    [sym_preproc_line] = STATE(2669),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(7224),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [anon_sym_inherit] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [anon_sym_POUNDelse] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2670)] = {
    [sym_block_comment] = STATE(2670),
    [sym_compiler_directive_decl] = STATE(2670),
    [sym_fsi_directive_decl] = STATE(2670),
    [sym_preproc_line] = STATE(2670),
    [aux_sym_compound_type_repeat1] = STATE(2670),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7226),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2671)] = {
    [sym_block_comment] = STATE(2671),
    [sym_compiler_directive_decl] = STATE(2671),
    [sym_fsi_directive_decl] = STATE(2671),
    [sym_preproc_line] = STATE(2671),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2672)] = {
    [sym_block_comment] = STATE(2672),
    [sym_compiler_directive_decl] = STATE(2672),
    [sym_fsi_directive_decl] = STATE(2672),
    [sym_preproc_line] = STATE(2672),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_and] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [anon_sym_POUNDelse] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2673)] = {
    [sym_block_comment] = STATE(2673),
    [sym_compiler_directive_decl] = STATE(2673),
    [sym_fsi_directive_decl] = STATE(2673),
    [sym_preproc_line] = STATE(2673),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_module] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_open] = ACTIONS(7053),
    [anon_sym_type] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_and] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_exception] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7055),
    [anon_sym_POUNDload] = ACTIONS(7055),
    [anon_sym_POUNDtime] = ACTIONS(7055),
    [anon_sym_POUNDI] = ACTIONS(7055),
    [anon_sym_POUNDhelp] = ACTIONS(7055),
    [anon_sym_POUNDquit] = ACTIONS(7055),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(2674)] = {
    [sym_block_comment] = STATE(2674),
    [sym_compiler_directive_decl] = STATE(2674),
    [sym_fsi_directive_decl] = STATE(2674),
    [sym_preproc_line] = STATE(2674),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2675)] = {
    [sym_block_comment] = STATE(2675),
    [sym_compiler_directive_decl] = STATE(2675),
    [sym_fsi_directive_decl] = STATE(2675),
    [sym_preproc_line] = STATE(2675),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_module] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_open] = ACTIONS(7069),
    [anon_sym_type] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_and] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_exception] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [anon_sym_extern] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7071),
    [anon_sym_POUNDload] = ACTIONS(7071),
    [anon_sym_POUNDtime] = ACTIONS(7071),
    [anon_sym_POUNDI] = ACTIONS(7071),
    [anon_sym_POUNDhelp] = ACTIONS(7071),
    [anon_sym_POUNDquit] = ACTIONS(7071),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(2676)] = {
    [sym_block_comment] = STATE(2676),
    [sym_compiler_directive_decl] = STATE(2676),
    [sym_fsi_directive_decl] = STATE(2676),
    [sym_preproc_line] = STATE(2676),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_module] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_open] = ACTIONS(7073),
    [anon_sym_type] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_and] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_exception] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [anon_sym_extern] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7075),
    [anon_sym_POUNDload] = ACTIONS(7075),
    [anon_sym_POUNDtime] = ACTIONS(7075),
    [anon_sym_POUNDI] = ACTIONS(7075),
    [anon_sym_POUNDhelp] = ACTIONS(7075),
    [anon_sym_POUNDquit] = ACTIONS(7075),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(2677)] = {
    [sym_block_comment] = STATE(2677),
    [sym_compiler_directive_decl] = STATE(2677),
    [sym_fsi_directive_decl] = STATE(2677),
    [sym_preproc_line] = STATE(2677),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_module] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_open] = ACTIONS(7077),
    [anon_sym_type] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_and] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_exception] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [anon_sym_extern] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7079),
    [anon_sym_POUNDload] = ACTIONS(7079),
    [anon_sym_POUNDtime] = ACTIONS(7079),
    [anon_sym_POUNDI] = ACTIONS(7079),
    [anon_sym_POUNDhelp] = ACTIONS(7079),
    [anon_sym_POUNDquit] = ACTIONS(7079),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(2678)] = {
    [sym_block_comment] = STATE(2678),
    [sym_compiler_directive_decl] = STATE(2678),
    [sym_fsi_directive_decl] = STATE(2678),
    [sym_preproc_line] = STATE(2678),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_module] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_open] = ACTIONS(7085),
    [anon_sym_type] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_and] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_exception] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [anon_sym_extern] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7087),
    [anon_sym_POUNDload] = ACTIONS(7087),
    [anon_sym_POUNDtime] = ACTIONS(7087),
    [anon_sym_POUNDI] = ACTIONS(7087),
    [anon_sym_POUNDhelp] = ACTIONS(7087),
    [anon_sym_POUNDquit] = ACTIONS(7087),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(2679)] = {
    [sym_block_comment] = STATE(2679),
    [sym_compiler_directive_decl] = STATE(2679),
    [sym_fsi_directive_decl] = STATE(2679),
    [sym_preproc_line] = STATE(2679),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_module] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_open] = ACTIONS(7089),
    [anon_sym_type] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_and] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_exception] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [anon_sym_extern] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7091),
    [anon_sym_POUNDload] = ACTIONS(7091),
    [anon_sym_POUNDtime] = ACTIONS(7091),
    [anon_sym_POUNDI] = ACTIONS(7091),
    [anon_sym_POUNDhelp] = ACTIONS(7091),
    [anon_sym_POUNDquit] = ACTIONS(7091),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(2680)] = {
    [sym_block_comment] = STATE(2680),
    [sym_compiler_directive_decl] = STATE(2680),
    [sym_fsi_directive_decl] = STATE(2680),
    [sym_preproc_line] = STATE(2680),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_module] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_open] = ACTIONS(7093),
    [anon_sym_type] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_and] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_exception] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [anon_sym_extern] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7095),
    [anon_sym_POUNDload] = ACTIONS(7095),
    [anon_sym_POUNDtime] = ACTIONS(7095),
    [anon_sym_POUNDI] = ACTIONS(7095),
    [anon_sym_POUNDhelp] = ACTIONS(7095),
    [anon_sym_POUNDquit] = ACTIONS(7095),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(2681)] = {
    [sym_block_comment] = STATE(2681),
    [sym_compiler_directive_decl] = STATE(2681),
    [sym_fsi_directive_decl] = STATE(2681),
    [sym_preproc_line] = STATE(2681),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_module] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_open] = ACTIONS(7097),
    [anon_sym_type] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_and] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_exception] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [anon_sym_extern] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7099),
    [anon_sym_POUNDload] = ACTIONS(7099),
    [anon_sym_POUNDtime] = ACTIONS(7099),
    [anon_sym_POUNDI] = ACTIONS(7099),
    [anon_sym_POUNDhelp] = ACTIONS(7099),
    [anon_sym_POUNDquit] = ACTIONS(7099),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(2682)] = {
    [sym_block_comment] = STATE(2682),
    [sym_compiler_directive_decl] = STATE(2682),
    [sym_fsi_directive_decl] = STATE(2682),
    [sym_preproc_line] = STATE(2682),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_module] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_open] = ACTIONS(7101),
    [anon_sym_type] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_and] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_exception] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7103),
    [anon_sym_POUNDload] = ACTIONS(7103),
    [anon_sym_POUNDtime] = ACTIONS(7103),
    [anon_sym_POUNDI] = ACTIONS(7103),
    [anon_sym_POUNDhelp] = ACTIONS(7103),
    [anon_sym_POUNDquit] = ACTIONS(7103),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(2683)] = {
    [sym_block_comment] = STATE(2683),
    [sym_compiler_directive_decl] = STATE(2683),
    [sym_fsi_directive_decl] = STATE(2683),
    [sym_preproc_line] = STATE(2683),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_module] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_open] = ACTIONS(7105),
    [anon_sym_type] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_and] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_exception] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7107),
    [anon_sym_POUNDload] = ACTIONS(7107),
    [anon_sym_POUNDtime] = ACTIONS(7107),
    [anon_sym_POUNDI] = ACTIONS(7107),
    [anon_sym_POUNDhelp] = ACTIONS(7107),
    [anon_sym_POUNDquit] = ACTIONS(7107),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(2684)] = {
    [sym_block_comment] = STATE(2684),
    [sym_compiler_directive_decl] = STATE(2684),
    [sym_fsi_directive_decl] = STATE(2684),
    [sym_preproc_line] = STATE(2684),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_module] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_open] = ACTIONS(7109),
    [anon_sym_type] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_and] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_exception] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7111),
    [anon_sym_POUNDload] = ACTIONS(7111),
    [anon_sym_POUNDtime] = ACTIONS(7111),
    [anon_sym_POUNDI] = ACTIONS(7111),
    [anon_sym_POUNDhelp] = ACTIONS(7111),
    [anon_sym_POUNDquit] = ACTIONS(7111),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(2685)] = {
    [sym_block_comment] = STATE(2685),
    [sym_compiler_directive_decl] = STATE(2685),
    [sym_fsi_directive_decl] = STATE(2685),
    [sym_preproc_line] = STATE(2685),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_module] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_open] = ACTIONS(7113),
    [anon_sym_type] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_and] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_exception] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7115),
    [anon_sym_POUNDload] = ACTIONS(7115),
    [anon_sym_POUNDtime] = ACTIONS(7115),
    [anon_sym_POUNDI] = ACTIONS(7115),
    [anon_sym_POUNDhelp] = ACTIONS(7115),
    [anon_sym_POUNDquit] = ACTIONS(7115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(2686)] = {
    [sym_block_comment] = STATE(2686),
    [sym_compiler_directive_decl] = STATE(2686),
    [sym_fsi_directive_decl] = STATE(2686),
    [sym_preproc_line] = STATE(2686),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_module] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_open] = ACTIONS(7117),
    [anon_sym_type] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_and] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_exception] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7119),
    [anon_sym_POUNDload] = ACTIONS(7119),
    [anon_sym_POUNDtime] = ACTIONS(7119),
    [anon_sym_POUNDI] = ACTIONS(7119),
    [anon_sym_POUNDhelp] = ACTIONS(7119),
    [anon_sym_POUNDquit] = ACTIONS(7119),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(2687)] = {
    [sym_block_comment] = STATE(2687),
    [sym_compiler_directive_decl] = STATE(2687),
    [sym_fsi_directive_decl] = STATE(2687),
    [sym_preproc_line] = STATE(2687),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_module] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_open] = ACTIONS(7121),
    [anon_sym_type] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_and] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_exception] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7123),
    [anon_sym_POUNDload] = ACTIONS(7123),
    [anon_sym_POUNDtime] = ACTIONS(7123),
    [anon_sym_POUNDI] = ACTIONS(7123),
    [anon_sym_POUNDhelp] = ACTIONS(7123),
    [anon_sym_POUNDquit] = ACTIONS(7123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(2688)] = {
    [sym_block_comment] = STATE(2688),
    [sym_compiler_directive_decl] = STATE(2688),
    [sym_fsi_directive_decl] = STATE(2688),
    [sym_preproc_line] = STATE(2688),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_module] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_open] = ACTIONS(6936),
    [anon_sym_type] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_and] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_exception] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [anon_sym_extern] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6934),
    [anon_sym_POUNDload] = ACTIONS(6934),
    [anon_sym_POUNDtime] = ACTIONS(6934),
    [anon_sym_POUNDI] = ACTIONS(6934),
    [anon_sym_POUNDhelp] = ACTIONS(6934),
    [anon_sym_POUNDquit] = ACTIONS(6934),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(2689)] = {
    [sym_block_comment] = STATE(2689),
    [sym_compiler_directive_decl] = STATE(2689),
    [sym_fsi_directive_decl] = STATE(2689),
    [sym_preproc_line] = STATE(2689),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_module] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_open] = ACTIONS(6940),
    [anon_sym_type] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_and] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_exception] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [anon_sym_extern] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6938),
    [anon_sym_POUNDload] = ACTIONS(6938),
    [anon_sym_POUNDtime] = ACTIONS(6938),
    [anon_sym_POUNDI] = ACTIONS(6938),
    [anon_sym_POUNDhelp] = ACTIONS(6938),
    [anon_sym_POUNDquit] = ACTIONS(6938),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(2690)] = {
    [sym_block_comment] = STATE(2690),
    [sym_compiler_directive_decl] = STATE(2690),
    [sym_fsi_directive_decl] = STATE(2690),
    [sym_preproc_line] = STATE(2690),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_module] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_open] = ACTIONS(7015),
    [anon_sym_type] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_and] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_exception] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [anon_sym_extern] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7013),
    [anon_sym_POUNDload] = ACTIONS(7013),
    [anon_sym_POUNDtime] = ACTIONS(7013),
    [anon_sym_POUNDI] = ACTIONS(7013),
    [anon_sym_POUNDhelp] = ACTIONS(7013),
    [anon_sym_POUNDquit] = ACTIONS(7013),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(2691)] = {
    [sym_block_comment] = STATE(2691),
    [sym_compiler_directive_decl] = STATE(2691),
    [sym_fsi_directive_decl] = STATE(2691),
    [sym_preproc_line] = STATE(2691),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_module] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_open] = ACTIONS(7031),
    [anon_sym_type] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_and] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_exception] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [anon_sym_extern] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7029),
    [anon_sym_POUNDload] = ACTIONS(7029),
    [anon_sym_POUNDtime] = ACTIONS(7029),
    [anon_sym_POUNDI] = ACTIONS(7029),
    [anon_sym_POUNDhelp] = ACTIONS(7029),
    [anon_sym_POUNDquit] = ACTIONS(7029),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(2692)] = {
    [sym_block_comment] = STATE(2692),
    [sym_compiler_directive_decl] = STATE(2692),
    [sym_fsi_directive_decl] = STATE(2692),
    [sym_preproc_line] = STATE(2692),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_module] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_open] = ACTIONS(7047),
    [anon_sym_type] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_and] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_exception] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [anon_sym_extern] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7045),
    [anon_sym_POUNDload] = ACTIONS(7045),
    [anon_sym_POUNDtime] = ACTIONS(7045),
    [anon_sym_POUNDI] = ACTIONS(7045),
    [anon_sym_POUNDhelp] = ACTIONS(7045),
    [anon_sym_POUNDquit] = ACTIONS(7045),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(2693)] = {
    [sym_block_comment] = STATE(2693),
    [sym_compiler_directive_decl] = STATE(2693),
    [sym_fsi_directive_decl] = STATE(2693),
    [sym_preproc_line] = STATE(2693),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_module] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_open] = ACTIONS(7051),
    [anon_sym_type] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_and] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_exception] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [anon_sym_extern] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7049),
    [anon_sym_POUNDload] = ACTIONS(7049),
    [anon_sym_POUNDtime] = ACTIONS(7049),
    [anon_sym_POUNDI] = ACTIONS(7049),
    [anon_sym_POUNDhelp] = ACTIONS(7049),
    [anon_sym_POUNDquit] = ACTIONS(7049),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(2694)] = {
    [sym_block_comment] = STATE(2694),
    [sym_compiler_directive_decl] = STATE(2694),
    [sym_fsi_directive_decl] = STATE(2694),
    [sym_preproc_line] = STATE(2694),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_module] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_open] = ACTIONS(7063),
    [anon_sym_type] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_and] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_exception] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [anon_sym_extern] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7061),
    [anon_sym_POUNDload] = ACTIONS(7061),
    [anon_sym_POUNDtime] = ACTIONS(7061),
    [anon_sym_POUNDI] = ACTIONS(7061),
    [anon_sym_POUNDhelp] = ACTIONS(7061),
    [anon_sym_POUNDquit] = ACTIONS(7061),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(2695)] = {
    [sym_block_comment] = STATE(2695),
    [sym_compiler_directive_decl] = STATE(2695),
    [sym_fsi_directive_decl] = STATE(2695),
    [sym_preproc_line] = STATE(2695),
    [aux_sym_type_argument_repeat1] = STATE(2714),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_DOT_DOT] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__dedent] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2696)] = {
    [sym_block_comment] = STATE(2696),
    [sym_compiler_directive_decl] = STATE(2696),
    [sym_fsi_directive_decl] = STATE(2696),
    [sym_preproc_line] = STATE(2696),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_inherit] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2697)] = {
    [sym_block_comment] = STATE(2697),
    [sym_compiler_directive_decl] = STATE(2697),
    [sym_fsi_directive_decl] = STATE(2697),
    [sym_preproc_line] = STATE(2697),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_inherit] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2698)] = {
    [sym_block_comment] = STATE(2698),
    [sym_compiler_directive_decl] = STATE(2698),
    [sym_fsi_directive_decl] = STATE(2698),
    [sym_preproc_line] = STATE(2698),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [anon_sym_inherit] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [anon_sym_POUNDelse] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2699)] = {
    [sym_block_comment] = STATE(2699),
    [sym_compiler_directive_decl] = STATE(2699),
    [sym_fsi_directive_decl] = STATE(2699),
    [sym_preproc_line] = STATE(2699),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_inherit] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2700)] = {
    [sym_block_comment] = STATE(2700),
    [sym_compiler_directive_decl] = STATE(2700),
    [sym_fsi_directive_decl] = STATE(2700),
    [sym_preproc_line] = STATE(2700),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [anon_sym_inherit] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [anon_sym_POUNDelse] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2701)] = {
    [sym_block_comment] = STATE(2701),
    [sym_compiler_directive_decl] = STATE(2701),
    [sym_fsi_directive_decl] = STATE(2701),
    [sym_preproc_line] = STATE(2701),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2702)] = {
    [sym_block_comment] = STATE(2702),
    [sym_compiler_directive_decl] = STATE(2702),
    [sym_fsi_directive_decl] = STATE(2702),
    [sym_preproc_line] = STATE(2702),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2703)] = {
    [sym_block_comment] = STATE(2703),
    [sym_compiler_directive_decl] = STATE(2703),
    [sym_fsi_directive_decl] = STATE(2703),
    [sym_preproc_line] = STATE(2703),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2704)] = {
    [sym_block_comment] = STATE(2704),
    [sym_compiler_directive_decl] = STATE(2704),
    [sym_fsi_directive_decl] = STATE(2704),
    [sym_preproc_line] = STATE(2704),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_module] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_open] = ACTIONS(7067),
    [anon_sym_type] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_and] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_exception] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [anon_sym_extern] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7065),
    [anon_sym_POUNDload] = ACTIONS(7065),
    [anon_sym_POUNDtime] = ACTIONS(7065),
    [anon_sym_POUNDI] = ACTIONS(7065),
    [anon_sym_POUNDhelp] = ACTIONS(7065),
    [anon_sym_POUNDquit] = ACTIONS(7065),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(2705)] = {
    [sym_block_comment] = STATE(2705),
    [sym_compiler_directive_decl] = STATE(2705),
    [sym_fsi_directive_decl] = STATE(2705),
    [sym_preproc_line] = STATE(2705),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_module] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_open] = ACTIONS(7125),
    [anon_sym_type] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_and] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_exception] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [anon_sym_extern] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7127),
    [anon_sym_POUNDload] = ACTIONS(7127),
    [anon_sym_POUNDtime] = ACTIONS(7127),
    [anon_sym_POUNDI] = ACTIONS(7127),
    [anon_sym_POUNDhelp] = ACTIONS(7127),
    [anon_sym_POUNDquit] = ACTIONS(7127),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(2706)] = {
    [sym_block_comment] = STATE(2706),
    [sym_compiler_directive_decl] = STATE(2706),
    [sym_fsi_directive_decl] = STATE(2706),
    [sym_preproc_line] = STATE(2706),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_and] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [anon_sym_POUNDelse] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2707)] = {
    [sym_block_comment] = STATE(2707),
    [sym_compiler_directive_decl] = STATE(2707),
    [sym_fsi_directive_decl] = STATE(2707),
    [sym_preproc_line] = STATE(2707),
    [aux_sym_type_argument_repeat1] = STATE(2708),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__dedent] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2708)] = {
    [sym_block_comment] = STATE(2708),
    [sym_compiler_directive_decl] = STATE(2708),
    [sym_fsi_directive_decl] = STATE(2708),
    [sym_preproc_line] = STATE(2708),
    [aux_sym_type_argument_repeat1] = STATE(2708),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7229),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2709)] = {
    [sym_block_comment] = STATE(2709),
    [sym_compiler_directive_decl] = STATE(2709),
    [sym_fsi_directive_decl] = STATE(2709),
    [sym_preproc_line] = STATE(2709),
    [aux_sym_type_argument_repeat1] = STATE(2710),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_inherit] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2710)] = {
    [sym_block_comment] = STATE(2710),
    [sym_compiler_directive_decl] = STATE(2710),
    [sym_fsi_directive_decl] = STATE(2710),
    [sym_preproc_line] = STATE(2710),
    [aux_sym_type_argument_repeat1] = STATE(2710),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7232),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_inherit] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2711)] = {
    [sym_block_comment] = STATE(2711),
    [sym_compiler_directive_decl] = STATE(2711),
    [sym_fsi_directive_decl] = STATE(2711),
    [sym_preproc_line] = STATE(2711),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_module] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_open] = ACTIONS(7129),
    [anon_sym_type] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_and] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_exception] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7131),
    [anon_sym_POUNDload] = ACTIONS(7131),
    [anon_sym_POUNDtime] = ACTIONS(7131),
    [anon_sym_POUNDI] = ACTIONS(7131),
    [anon_sym_POUNDhelp] = ACTIONS(7131),
    [anon_sym_POUNDquit] = ACTIONS(7131),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(2712)] = {
    [sym_block_comment] = STATE(2712),
    [sym_compiler_directive_decl] = STATE(2712),
    [sym_fsi_directive_decl] = STATE(2712),
    [sym_preproc_line] = STATE(2712),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_module] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_open] = ACTIONS(7133),
    [anon_sym_type] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_and] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_exception] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7135),
    [anon_sym_POUNDload] = ACTIONS(7135),
    [anon_sym_POUNDtime] = ACTIONS(7135),
    [anon_sym_POUNDI] = ACTIONS(7135),
    [anon_sym_POUNDhelp] = ACTIONS(7135),
    [anon_sym_POUNDquit] = ACTIONS(7135),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(2713)] = {
    [sym_block_comment] = STATE(2713),
    [sym_compiler_directive_decl] = STATE(2713),
    [sym_fsi_directive_decl] = STATE(2713),
    [sym_preproc_line] = STATE(2713),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_module] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_open] = ACTIONS(7137),
    [anon_sym_type] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_and] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_exception] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7139),
    [anon_sym_POUNDload] = ACTIONS(7139),
    [anon_sym_POUNDtime] = ACTIONS(7139),
    [anon_sym_POUNDI] = ACTIONS(7139),
    [anon_sym_POUNDhelp] = ACTIONS(7139),
    [anon_sym_POUNDquit] = ACTIONS(7139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(2714)] = {
    [sym_block_comment] = STATE(2714),
    [sym_compiler_directive_decl] = STATE(2714),
    [sym_fsi_directive_decl] = STATE(2714),
    [sym_preproc_line] = STATE(2714),
    [aux_sym_type_argument_repeat1] = STATE(2714),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7235),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2715)] = {
    [sym_block_comment] = STATE(2715),
    [sym_compiler_directive_decl] = STATE(2715),
    [sym_fsi_directive_decl] = STATE(2715),
    [sym_preproc_line] = STATE(2715),
    [aux_sym_type_argument_repeat1] = STATE(2709),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(7238),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_inherit] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2716)] = {
    [sym_block_comment] = STATE(2716),
    [sym_compiler_directive_decl] = STATE(2716),
    [sym_fsi_directive_decl] = STATE(2716),
    [sym_preproc_line] = STATE(2716),
    [aux_sym_type_argument_repeat1] = STATE(2717),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(2717)] = {
    [sym_block_comment] = STATE(2717),
    [sym_compiler_directive_decl] = STATE(2717),
    [sym_fsi_directive_decl] = STATE(2717),
    [sym_preproc_line] = STATE(2717),
    [aux_sym_type_argument_repeat1] = STATE(2717),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7241),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2718)] = {
    [sym_block_comment] = STATE(2718),
    [sym_compiler_directive_decl] = STATE(2718),
    [sym_fsi_directive_decl] = STATE(2718),
    [sym_preproc_line] = STATE(2718),
    [aux_sym_long_identifier_repeat1] = STATE(2743),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__dedent] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2719)] = {
    [sym_block_comment] = STATE(2719),
    [sym_compiler_directive_decl] = STATE(2719),
    [sym_fsi_directive_decl] = STATE(2719),
    [sym_preproc_line] = STATE(2719),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2720)] = {
    [sym_elif_expression] = STATE(2940),
    [sym_block_comment] = STATE(2720),
    [sym_compiler_directive_decl] = STATE(2720),
    [sym_fsi_directive_decl] = STATE(2720),
    [sym_preproc_line] = STATE(2720),
    [aux_sym_if_expression_repeat1] = STATE(2720),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(7246),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [anon_sym_inherit] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [anon_sym_POUNDelse] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2721)] = {
    [sym_block_comment] = STATE(2721),
    [sym_compiler_directive_decl] = STATE(2721),
    [sym_fsi_directive_decl] = STATE(2721),
    [sym_preproc_line] = STATE(2721),
    [aux_sym_long_identifier_repeat1] = STATE(2745),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7249),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_inherit] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2722)] = {
    [sym_block_comment] = STATE(2722),
    [sym_compiler_directive_decl] = STATE(2722),
    [sym_fsi_directive_decl] = STATE(2722),
    [sym_preproc_line] = STATE(2722),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_module] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_open] = ACTIONS(7141),
    [anon_sym_type] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_and] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_exception] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7143),
    [anon_sym_POUNDload] = ACTIONS(7143),
    [anon_sym_POUNDtime] = ACTIONS(7143),
    [anon_sym_POUNDI] = ACTIONS(7143),
    [anon_sym_POUNDhelp] = ACTIONS(7143),
    [anon_sym_POUNDquit] = ACTIONS(7143),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(2723)] = {
    [sym_block_comment] = STATE(2723),
    [sym_compiler_directive_decl] = STATE(2723),
    [sym_fsi_directive_decl] = STATE(2723),
    [sym_preproc_line] = STATE(2723),
    [aux_sym_long_identifier_repeat1] = STATE(2752),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7251),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2724)] = {
    [sym_block_comment] = STATE(2724),
    [sym_compiler_directive_decl] = STATE(2724),
    [sym_fsi_directive_decl] = STATE(2724),
    [sym_preproc_line] = STATE(2724),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_and] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [anon_sym_POUNDelse] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2725)] = {
    [sym_block_comment] = STATE(2725),
    [sym_compiler_directive_decl] = STATE(2725),
    [sym_fsi_directive_decl] = STATE(2725),
    [sym_preproc_line] = STATE(2725),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_module] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_open] = ACTIONS(7145),
    [anon_sym_type] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_and] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_exception] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7147),
    [anon_sym_POUNDload] = ACTIONS(7147),
    [anon_sym_POUNDtime] = ACTIONS(7147),
    [anon_sym_POUNDI] = ACTIONS(7147),
    [anon_sym_POUNDhelp] = ACTIONS(7147),
    [anon_sym_POUNDquit] = ACTIONS(7147),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(2726)] = {
    [sym_block_comment] = STATE(2726),
    [sym_compiler_directive_decl] = STATE(2726),
    [sym_fsi_directive_decl] = STATE(2726),
    [sym_preproc_line] = STATE(2726),
    [aux_sym_compound_type_repeat1] = STATE(2731),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6744),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [anon_sym_inherit] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2727)] = {
    [sym__else_expression] = STATE(3526),
    [sym_elif_expression] = STATE(2999),
    [sym_block_comment] = STATE(2727),
    [sym_compiler_directive_decl] = STATE(2727),
    [sym_fsi_directive_decl] = STATE(2727),
    [sym_preproc_line] = STATE(2727),
    [aux_sym_if_expression_repeat1] = STATE(2747),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7253),
    [anon_sym_elif] = ACTIONS(7255),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [anon_sym_inherit] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2728)] = {
    [sym_block_comment] = STATE(2728),
    [sym_compiler_directive_decl] = STATE(2728),
    [sym_fsi_directive_decl] = STATE(2728),
    [sym_preproc_line] = STATE(2728),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_module] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_open] = ACTIONS(7155),
    [anon_sym_type] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_and] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_exception] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [anon_sym_extern] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7157),
    [anon_sym_POUNDload] = ACTIONS(7157),
    [anon_sym_POUNDtime] = ACTIONS(7157),
    [anon_sym_POUNDI] = ACTIONS(7157),
    [anon_sym_POUNDhelp] = ACTIONS(7157),
    [anon_sym_POUNDquit] = ACTIONS(7157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(2729)] = {
    [sym_block_comment] = STATE(2729),
    [sym_compiler_directive_decl] = STATE(2729),
    [sym_fsi_directive_decl] = STATE(2729),
    [sym_preproc_line] = STATE(2729),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_module] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_open] = ACTIONS(7159),
    [anon_sym_type] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_and] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_exception] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [anon_sym_extern] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7161),
    [anon_sym_POUNDload] = ACTIONS(7161),
    [anon_sym_POUNDtime] = ACTIONS(7161),
    [anon_sym_POUNDI] = ACTIONS(7161),
    [anon_sym_POUNDhelp] = ACTIONS(7161),
    [anon_sym_POUNDquit] = ACTIONS(7161),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(2730)] = {
    [sym_block_comment] = STATE(2730),
    [sym_compiler_directive_decl] = STATE(2730),
    [sym_fsi_directive_decl] = STATE(2730),
    [sym_preproc_line] = STATE(2730),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_and] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [anon_sym_POUNDelse] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2731)] = {
    [sym_block_comment] = STATE(2731),
    [sym_compiler_directive_decl] = STATE(2731),
    [sym_fsi_directive_decl] = STATE(2731),
    [sym_preproc_line] = STATE(2731),
    [aux_sym_compound_type_repeat1] = STATE(2731),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7257),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_inherit] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2732)] = {
    [sym_block_comment] = STATE(2732),
    [sym_compiler_directive_decl] = STATE(2732),
    [sym_fsi_directive_decl] = STATE(2732),
    [sym_preproc_line] = STATE(2732),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [anon_sym_inherit] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [anon_sym_POUNDelse] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2733)] = {
    [sym_block_comment] = STATE(2733),
    [sym_compiler_directive_decl] = STATE(2733),
    [sym_fsi_directive_decl] = STATE(2733),
    [sym_preproc_line] = STATE(2733),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2734)] = {
    [sym_block_comment] = STATE(2734),
    [sym_compiler_directive_decl] = STATE(2734),
    [sym_fsi_directive_decl] = STATE(2734),
    [sym_preproc_line] = STATE(2734),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [anon_sym_inherit] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [anon_sym_POUNDelse] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2735)] = {
    [sym_block_comment] = STATE(2735),
    [sym_compiler_directive_decl] = STATE(2735),
    [sym_fsi_directive_decl] = STATE(2735),
    [sym_preproc_line] = STATE(2735),
    [aux_sym_long_identifier_repeat1] = STATE(2758),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7260),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_DOT_DOT] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__dedent] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(2736)] = {
    [sym_block_comment] = STATE(2736),
    [sym_compiler_directive_decl] = STATE(2736),
    [sym_fsi_directive_decl] = STATE(2736),
    [sym_preproc_line] = STATE(2736),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [anon_sym_inherit] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [anon_sym_POUNDelse] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2737)] = {
    [sym_block_comment] = STATE(2737),
    [sym_compiler_directive_decl] = STATE(2737),
    [sym_fsi_directive_decl] = STATE(2737),
    [sym_preproc_line] = STATE(2737),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_and] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(7262),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [anon_sym_POUNDelse] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2738)] = {
    [sym_block_comment] = STATE(2738),
    [sym_compiler_directive_decl] = STATE(2738),
    [sym_fsi_directive_decl] = STATE(2738),
    [sym_preproc_line] = STATE(2738),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_and] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [anon_sym_POUNDelse] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2739)] = {
    [sym_block_comment] = STATE(2739),
    [sym_compiler_directive_decl] = STATE(2739),
    [sym_fsi_directive_decl] = STATE(2739),
    [sym_preproc_line] = STATE(2739),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2740)] = {
    [sym_block_comment] = STATE(2740),
    [sym_compiler_directive_decl] = STATE(2740),
    [sym_fsi_directive_decl] = STATE(2740),
    [sym_preproc_line] = STATE(2740),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [anon_sym_inherit] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [anon_sym_POUNDelse] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2741)] = {
    [sym_block_comment] = STATE(2741),
    [sym_compiler_directive_decl] = STATE(2741),
    [sym_fsi_directive_decl] = STATE(2741),
    [sym_preproc_line] = STATE(2741),
    [aux_sym_type_argument_repeat1] = STATE(2707),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(7264),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2742)] = {
    [sym_block_comment] = STATE(2742),
    [sym_compiler_directive_decl] = STATE(2742),
    [sym_fsi_directive_decl] = STATE(2742),
    [sym_preproc_line] = STATE(2742),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_and] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [anon_sym_POUNDelse] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2743)] = {
    [sym_block_comment] = STATE(2743),
    [sym_compiler_directive_decl] = STATE(2743),
    [sym_fsi_directive_decl] = STATE(2743),
    [sym_preproc_line] = STATE(2743),
    [aux_sym_long_identifier_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2744)] = {
    [sym_block_comment] = STATE(2744),
    [sym_compiler_directive_decl] = STATE(2744),
    [sym_fsi_directive_decl] = STATE(2744),
    [sym_preproc_line] = STATE(2744),
    [aux_sym_long_identifier_repeat1] = STATE(2744),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7267),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2745)] = {
    [sym_block_comment] = STATE(2745),
    [sym_compiler_directive_decl] = STATE(2745),
    [sym_fsi_directive_decl] = STATE(2745),
    [sym_preproc_line] = STATE(2745),
    [aux_sym_long_identifier_repeat1] = STATE(2746),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7249),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_inherit] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2746)] = {
    [sym_block_comment] = STATE(2746),
    [sym_compiler_directive_decl] = STATE(2746),
    [sym_fsi_directive_decl] = STATE(2746),
    [sym_preproc_line] = STATE(2746),
    [aux_sym_long_identifier_repeat1] = STATE(2746),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7270),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2747)] = {
    [sym__else_expression] = STATE(3575),
    [sym_elif_expression] = STATE(2999),
    [sym_block_comment] = STATE(2747),
    [sym_compiler_directive_decl] = STATE(2747),
    [sym_fsi_directive_decl] = STATE(2747),
    [sym_preproc_line] = STATE(2747),
    [aux_sym_if_expression_repeat1] = STATE(2837),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7253),
    [anon_sym_elif] = ACTIONS(7255),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [anon_sym_inherit] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2748)] = {
    [sym_block_comment] = STATE(2748),
    [sym_compiler_directive_decl] = STATE(2748),
    [sym_fsi_directive_decl] = STATE(2748),
    [sym_preproc_line] = STATE(2748),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_and] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [anon_sym_POUNDelse] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2749)] = {
    [sym_block_comment] = STATE(2749),
    [sym_compiler_directive_decl] = STATE(2749),
    [sym_fsi_directive_decl] = STATE(2749),
    [sym_preproc_line] = STATE(2749),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_and] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [anon_sym_POUNDelse] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2750)] = {
    [sym_block_comment] = STATE(2750),
    [sym_compiler_directive_decl] = STATE(2750),
    [sym_fsi_directive_decl] = STATE(2750),
    [sym_preproc_line] = STATE(2750),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [anon_sym_inherit] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [anon_sym_POUNDelse] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2751)] = {
    [sym_block_comment] = STATE(2751),
    [sym_compiler_directive_decl] = STATE(2751),
    [sym_fsi_directive_decl] = STATE(2751),
    [sym_preproc_line] = STATE(2751),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_and] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [anon_sym_POUNDelse] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2752)] = {
    [sym_block_comment] = STATE(2752),
    [sym_compiler_directive_decl] = STATE(2752),
    [sym_fsi_directive_decl] = STATE(2752),
    [sym_preproc_line] = STATE(2752),
    [aux_sym_long_identifier_repeat1] = STATE(2753),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7251),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2753)] = {
    [sym_block_comment] = STATE(2753),
    [sym_compiler_directive_decl] = STATE(2753),
    [sym_fsi_directive_decl] = STATE(2753),
    [sym_preproc_line] = STATE(2753),
    [aux_sym_long_identifier_repeat1] = STATE(2753),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7273),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2754)] = {
    [sym_block_comment] = STATE(2754),
    [sym_compiler_directive_decl] = STATE(2754),
    [sym_fsi_directive_decl] = STATE(2754),
    [sym_preproc_line] = STATE(2754),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5951),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7276),
    [anon_sym_uy] = ACTIONS(7278),
    [anon_sym_s] = ACTIONS(7280),
    [anon_sym_us] = ACTIONS(7282),
    [anon_sym_l] = ACTIONS(7284),
    [aux_sym_uint32_token1] = ACTIONS(7286),
    [anon_sym_n] = ACTIONS(7288),
    [anon_sym_un] = ACTIONS(7290),
    [anon_sym_L] = ACTIONS(7292),
    [aux_sym_uint64_token1] = ACTIONS(7294),
    [aux_sym_bignum_token1] = ACTIONS(7296),
    [aux_sym_decimal_token1] = ACTIONS(7298),
    [anon_sym_DOT2] = ACTIONS(7300),
    [aux_sym_float_token1] = ACTIONS(7302),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2755)] = {
    [sym_block_comment] = STATE(2755),
    [sym_compiler_directive_decl] = STATE(2755),
    [sym_fsi_directive_decl] = STATE(2755),
    [sym_preproc_line] = STATE(2755),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_module] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_open] = ACTIONS(6913),
    [anon_sym_type] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_and] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_exception] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [anon_sym_extern] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6911),
    [anon_sym_POUNDload] = ACTIONS(6911),
    [anon_sym_POUNDtime] = ACTIONS(6911),
    [anon_sym_POUNDI] = ACTIONS(6911),
    [anon_sym_POUNDhelp] = ACTIONS(6911),
    [anon_sym_POUNDquit] = ACTIONS(6911),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(2756)] = {
    [sym_block_comment] = STATE(2756),
    [sym_compiler_directive_decl] = STATE(2756),
    [sym_fsi_directive_decl] = STATE(2756),
    [sym_preproc_line] = STATE(2756),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_module] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_open] = ACTIONS(6917),
    [anon_sym_type] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_and] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_exception] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [anon_sym_extern] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6915),
    [anon_sym_POUNDload] = ACTIONS(6915),
    [anon_sym_POUNDtime] = ACTIONS(6915),
    [anon_sym_POUNDI] = ACTIONS(6915),
    [anon_sym_POUNDhelp] = ACTIONS(6915),
    [anon_sym_POUNDquit] = ACTIONS(6915),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(2757)] = {
    [sym_block_comment] = STATE(2757),
    [sym_compiler_directive_decl] = STATE(2757),
    [sym_fsi_directive_decl] = STATE(2757),
    [sym_preproc_line] = STATE(2757),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_module] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_open] = ACTIONS(6921),
    [anon_sym_type] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_and] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_exception] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [anon_sym_extern] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6919),
    [anon_sym_POUNDload] = ACTIONS(6919),
    [anon_sym_POUNDtime] = ACTIONS(6919),
    [anon_sym_POUNDI] = ACTIONS(6919),
    [anon_sym_POUNDhelp] = ACTIONS(6919),
    [anon_sym_POUNDquit] = ACTIONS(6919),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(2758)] = {
    [sym_block_comment] = STATE(2758),
    [sym_compiler_directive_decl] = STATE(2758),
    [sym_fsi_directive_decl] = STATE(2758),
    [sym_preproc_line] = STATE(2758),
    [aux_sym_long_identifier_repeat1] = STATE(2763),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7260),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2759)] = {
    [sym__else_expression] = STATE(3672),
    [sym_elif_expression] = STATE(2977),
    [sym_block_comment] = STATE(2759),
    [sym_compiler_directive_decl] = STATE(2759),
    [sym_fsi_directive_decl] = STATE(2759),
    [sym_preproc_line] = STATE(2759),
    [aux_sym_if_expression_repeat1] = STATE(2842),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7304),
    [anon_sym_elif] = ACTIONS(7306),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_DOT_DOT] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__dedent] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2760)] = {
    [sym_block_comment] = STATE(2760),
    [sym_compiler_directive_decl] = STATE(2760),
    [sym_fsi_directive_decl] = STATE(2760),
    [sym_preproc_line] = STATE(2760),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_module] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_open] = ACTIONS(6925),
    [anon_sym_type] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_and] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_exception] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [anon_sym_extern] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6923),
    [anon_sym_POUNDload] = ACTIONS(6923),
    [anon_sym_POUNDtime] = ACTIONS(6923),
    [anon_sym_POUNDI] = ACTIONS(6923),
    [anon_sym_POUNDhelp] = ACTIONS(6923),
    [anon_sym_POUNDquit] = ACTIONS(6923),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(2761)] = {
    [sym_block_comment] = STATE(2761),
    [sym_compiler_directive_decl] = STATE(2761),
    [sym_fsi_directive_decl] = STATE(2761),
    [sym_preproc_line] = STATE(2761),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(2762)] = {
    [sym_block_comment] = STATE(2762),
    [sym_compiler_directive_decl] = STATE(2762),
    [sym_fsi_directive_decl] = STATE(2762),
    [sym_preproc_line] = STATE(2762),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_module] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_open] = ACTIONS(6932),
    [anon_sym_type] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_and] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_exception] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [anon_sym_extern] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6930),
    [anon_sym_POUNDload] = ACTIONS(6930),
    [anon_sym_POUNDtime] = ACTIONS(6930),
    [anon_sym_POUNDI] = ACTIONS(6930),
    [anon_sym_POUNDhelp] = ACTIONS(6930),
    [anon_sym_POUNDquit] = ACTIONS(6930),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(2763)] = {
    [sym_block_comment] = STATE(2763),
    [sym_compiler_directive_decl] = STATE(2763),
    [sym_fsi_directive_decl] = STATE(2763),
    [sym_preproc_line] = STATE(2763),
    [aux_sym_long_identifier_repeat1] = STATE(2763),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2764)] = {
    [sym_block_comment] = STATE(2764),
    [sym_compiler_directive_decl] = STATE(2764),
    [sym_fsi_directive_decl] = STATE(2764),
    [sym_preproc_line] = STATE(2764),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6227),
    [anon_sym_uy] = ACTIONS(6229),
    [anon_sym_s] = ACTIONS(6231),
    [anon_sym_us] = ACTIONS(6233),
    [anon_sym_l] = ACTIONS(6235),
    [aux_sym_uint32_token1] = ACTIONS(6237),
    [anon_sym_n] = ACTIONS(6239),
    [anon_sym_un] = ACTIONS(6241),
    [anon_sym_L] = ACTIONS(6243),
    [aux_sym_uint64_token1] = ACTIONS(6245),
    [aux_sym_bignum_token1] = ACTIONS(6247),
    [aux_sym_decimal_token1] = ACTIONS(6249),
    [anon_sym_DOT2] = ACTIONS(7311),
    [aux_sym_float_token1] = ACTIONS(7313),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2765)] = {
    [sym_block_comment] = STATE(2765),
    [sym_compiler_directive_decl] = STATE(2765),
    [sym_fsi_directive_decl] = STATE(2765),
    [sym_preproc_line] = STATE(2765),
    [aux_sym_compound_type_repeat1] = STATE(2805),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6778),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__dedent] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2766)] = {
    [sym_block_comment] = STATE(2766),
    [sym_compiler_directive_decl] = STATE(2766),
    [sym_fsi_directive_decl] = STATE(2766),
    [sym_preproc_line] = STATE(2766),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_and] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [anon_sym_POUNDelse] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2767)] = {
    [sym_block_comment] = STATE(2767),
    [sym_compiler_directive_decl] = STATE(2767),
    [sym_fsi_directive_decl] = STATE(2767),
    [sym_preproc_line] = STATE(2767),
    [aux_sym_type_argument_repeat1] = STATE(2695),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(7315),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2768)] = {
    [sym__else_expression] = STATE(3487),
    [sym_elif_expression] = STATE(2969),
    [sym_block_comment] = STATE(2768),
    [sym_compiler_directive_decl] = STATE(2768),
    [sym_fsi_directive_decl] = STATE(2768),
    [sym_preproc_line] = STATE(2768),
    [aux_sym_if_expression_repeat1] = STATE(2795),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_and] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7318),
    [anon_sym_elif] = ACTIONS(7320),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__dedent] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2769)] = {
    [sym_block_comment] = STATE(2769),
    [sym_compiler_directive_decl] = STATE(2769),
    [sym_fsi_directive_decl] = STATE(2769),
    [sym_preproc_line] = STATE(2769),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_module] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_open] = ACTIONS(6944),
    [anon_sym_type] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_and] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_exception] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [anon_sym_extern] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6942),
    [anon_sym_POUNDload] = ACTIONS(6942),
    [anon_sym_POUNDtime] = ACTIONS(6942),
    [anon_sym_POUNDI] = ACTIONS(6942),
    [anon_sym_POUNDhelp] = ACTIONS(6942),
    [anon_sym_POUNDquit] = ACTIONS(6942),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(2770)] = {
    [sym_block_comment] = STATE(2770),
    [sym_compiler_directive_decl] = STATE(2770),
    [sym_fsi_directive_decl] = STATE(2770),
    [sym_preproc_line] = STATE(2770),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_module] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_open] = ACTIONS(6948),
    [anon_sym_type] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_and] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_exception] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [anon_sym_extern] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6946),
    [anon_sym_POUNDload] = ACTIONS(6946),
    [anon_sym_POUNDtime] = ACTIONS(6946),
    [anon_sym_POUNDI] = ACTIONS(6946),
    [anon_sym_POUNDhelp] = ACTIONS(6946),
    [anon_sym_POUNDquit] = ACTIONS(6946),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(2771)] = {
    [sym_block_comment] = STATE(2771),
    [sym_compiler_directive_decl] = STATE(2771),
    [sym_fsi_directive_decl] = STATE(2771),
    [sym_preproc_line] = STATE(2771),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_module] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_open] = ACTIONS(6952),
    [anon_sym_type] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_and] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_exception] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [anon_sym_extern] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6950),
    [anon_sym_POUNDload] = ACTIONS(6950),
    [anon_sym_POUNDtime] = ACTIONS(6950),
    [anon_sym_POUNDI] = ACTIONS(6950),
    [anon_sym_POUNDhelp] = ACTIONS(6950),
    [anon_sym_POUNDquit] = ACTIONS(6950),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(2772)] = {
    [sym_block_comment] = STATE(2772),
    [sym_compiler_directive_decl] = STATE(2772),
    [sym_fsi_directive_decl] = STATE(2772),
    [sym_preproc_line] = STATE(2772),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_module] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_open] = ACTIONS(6959),
    [anon_sym_type] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_and] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_exception] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [anon_sym_extern] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6957),
    [anon_sym_POUNDload] = ACTIONS(6957),
    [anon_sym_POUNDtime] = ACTIONS(6957),
    [anon_sym_POUNDI] = ACTIONS(6957),
    [anon_sym_POUNDhelp] = ACTIONS(6957),
    [anon_sym_POUNDquit] = ACTIONS(6957),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(2773)] = {
    [sym_block_comment] = STATE(2773),
    [sym_compiler_directive_decl] = STATE(2773),
    [sym_fsi_directive_decl] = STATE(2773),
    [sym_preproc_line] = STATE(2773),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_module] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_open] = ACTIONS(6963),
    [anon_sym_type] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_and] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_exception] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [anon_sym_extern] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6961),
    [anon_sym_POUNDload] = ACTIONS(6961),
    [anon_sym_POUNDtime] = ACTIONS(6961),
    [anon_sym_POUNDI] = ACTIONS(6961),
    [anon_sym_POUNDhelp] = ACTIONS(6961),
    [anon_sym_POUNDquit] = ACTIONS(6961),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(2774)] = {
    [sym_block_comment] = STATE(2774),
    [sym_compiler_directive_decl] = STATE(2774),
    [sym_fsi_directive_decl] = STATE(2774),
    [sym_preproc_line] = STATE(2774),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [anon_sym_inherit] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [anon_sym_POUNDelse] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2775)] = {
    [sym_block_comment] = STATE(2775),
    [sym_compiler_directive_decl] = STATE(2775),
    [sym_fsi_directive_decl] = STATE(2775),
    [sym_preproc_line] = STATE(2775),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_and] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [anon_sym_POUNDelse] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2776)] = {
    [sym_block_comment] = STATE(2776),
    [sym_compiler_directive_decl] = STATE(2776),
    [sym_fsi_directive_decl] = STATE(2776),
    [sym_preproc_line] = STATE(2776),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_module] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_open] = ACTIONS(6967),
    [anon_sym_type] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_and] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_exception] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [anon_sym_extern] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6965),
    [anon_sym_POUNDload] = ACTIONS(6965),
    [anon_sym_POUNDtime] = ACTIONS(6965),
    [anon_sym_POUNDI] = ACTIONS(6965),
    [anon_sym_POUNDhelp] = ACTIONS(6965),
    [anon_sym_POUNDquit] = ACTIONS(6965),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(2777)] = {
    [sym_block_comment] = STATE(2777),
    [sym_compiler_directive_decl] = STATE(2777),
    [sym_fsi_directive_decl] = STATE(2777),
    [sym_preproc_line] = STATE(2777),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_module] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_open] = ACTIONS(6971),
    [anon_sym_type] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_and] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_exception] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [anon_sym_extern] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6969),
    [anon_sym_POUNDload] = ACTIONS(6969),
    [anon_sym_POUNDtime] = ACTIONS(6969),
    [anon_sym_POUNDI] = ACTIONS(6969),
    [anon_sym_POUNDhelp] = ACTIONS(6969),
    [anon_sym_POUNDquit] = ACTIONS(6969),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(2778)] = {
    [sym_block_comment] = STATE(2778),
    [sym_compiler_directive_decl] = STATE(2778),
    [sym_fsi_directive_decl] = STATE(2778),
    [sym_preproc_line] = STATE(2778),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_module] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_open] = ACTIONS(6975),
    [anon_sym_type] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_and] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_exception] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [anon_sym_extern] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6973),
    [anon_sym_POUNDload] = ACTIONS(6973),
    [anon_sym_POUNDtime] = ACTIONS(6973),
    [anon_sym_POUNDI] = ACTIONS(6973),
    [anon_sym_POUNDhelp] = ACTIONS(6973),
    [anon_sym_POUNDquit] = ACTIONS(6973),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(2779)] = {
    [sym_block_comment] = STATE(2779),
    [sym_compiler_directive_decl] = STATE(2779),
    [sym_fsi_directive_decl] = STATE(2779),
    [sym_preproc_line] = STATE(2779),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_module] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_open] = ACTIONS(6979),
    [anon_sym_type] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_and] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_exception] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [anon_sym_extern] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6977),
    [anon_sym_POUNDload] = ACTIONS(6977),
    [anon_sym_POUNDtime] = ACTIONS(6977),
    [anon_sym_POUNDI] = ACTIONS(6977),
    [anon_sym_POUNDhelp] = ACTIONS(6977),
    [anon_sym_POUNDquit] = ACTIONS(6977),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(2780)] = {
    [sym_block_comment] = STATE(2780),
    [sym_compiler_directive_decl] = STATE(2780),
    [sym_fsi_directive_decl] = STATE(2780),
    [sym_preproc_line] = STATE(2780),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_module] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_open] = ACTIONS(6983),
    [anon_sym_type] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_and] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_exception] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [anon_sym_extern] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6981),
    [anon_sym_POUNDload] = ACTIONS(6981),
    [anon_sym_POUNDtime] = ACTIONS(6981),
    [anon_sym_POUNDI] = ACTIONS(6981),
    [anon_sym_POUNDhelp] = ACTIONS(6981),
    [anon_sym_POUNDquit] = ACTIONS(6981),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(2781)] = {
    [sym_block_comment] = STATE(2781),
    [sym_compiler_directive_decl] = STATE(2781),
    [sym_fsi_directive_decl] = STATE(2781),
    [sym_preproc_line] = STATE(2781),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_module] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_open] = ACTIONS(6987),
    [anon_sym_type] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_and] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_exception] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [anon_sym_extern] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6985),
    [anon_sym_POUNDload] = ACTIONS(6985),
    [anon_sym_POUNDtime] = ACTIONS(6985),
    [anon_sym_POUNDI] = ACTIONS(6985),
    [anon_sym_POUNDhelp] = ACTIONS(6985),
    [anon_sym_POUNDquit] = ACTIONS(6985),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(2782)] = {
    [sym_block_comment] = STATE(2782),
    [sym_compiler_directive_decl] = STATE(2782),
    [sym_fsi_directive_decl] = STATE(2782),
    [sym_preproc_line] = STATE(2782),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_module] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_open] = ACTIONS(6991),
    [anon_sym_type] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_and] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_exception] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [anon_sym_extern] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6989),
    [anon_sym_POUNDload] = ACTIONS(6989),
    [anon_sym_POUNDtime] = ACTIONS(6989),
    [anon_sym_POUNDI] = ACTIONS(6989),
    [anon_sym_POUNDhelp] = ACTIONS(6989),
    [anon_sym_POUNDquit] = ACTIONS(6989),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(2783)] = {
    [sym_block_comment] = STATE(2783),
    [sym_compiler_directive_decl] = STATE(2783),
    [sym_fsi_directive_decl] = STATE(2783),
    [sym_preproc_line] = STATE(2783),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_module] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_open] = ACTIONS(6995),
    [anon_sym_type] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_and] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_exception] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [anon_sym_extern] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6993),
    [anon_sym_POUNDload] = ACTIONS(6993),
    [anon_sym_POUNDtime] = ACTIONS(6993),
    [anon_sym_POUNDI] = ACTIONS(6993),
    [anon_sym_POUNDhelp] = ACTIONS(6993),
    [anon_sym_POUNDquit] = ACTIONS(6993),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(2784)] = {
    [sym_block_comment] = STATE(2784),
    [sym_compiler_directive_decl] = STATE(2784),
    [sym_fsi_directive_decl] = STATE(2784),
    [sym_preproc_line] = STATE(2784),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_module] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_open] = ACTIONS(6999),
    [anon_sym_type] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_and] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_exception] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [anon_sym_extern] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6997),
    [anon_sym_POUNDload] = ACTIONS(6997),
    [anon_sym_POUNDtime] = ACTIONS(6997),
    [anon_sym_POUNDI] = ACTIONS(6997),
    [anon_sym_POUNDhelp] = ACTIONS(6997),
    [anon_sym_POUNDquit] = ACTIONS(6997),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(2785)] = {
    [sym_block_comment] = STATE(2785),
    [sym_compiler_directive_decl] = STATE(2785),
    [sym_fsi_directive_decl] = STATE(2785),
    [sym_preproc_line] = STATE(2785),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_module] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_open] = ACTIONS(7003),
    [anon_sym_type] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_and] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_exception] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [anon_sym_extern] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7001),
    [anon_sym_POUNDload] = ACTIONS(7001),
    [anon_sym_POUNDtime] = ACTIONS(7001),
    [anon_sym_POUNDI] = ACTIONS(7001),
    [anon_sym_POUNDhelp] = ACTIONS(7001),
    [anon_sym_POUNDquit] = ACTIONS(7001),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(2786)] = {
    [sym_block_comment] = STATE(2786),
    [sym_compiler_directive_decl] = STATE(2786),
    [sym_fsi_directive_decl] = STATE(2786),
    [sym_preproc_line] = STATE(2786),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_module] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_open] = ACTIONS(7007),
    [anon_sym_type] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_and] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_exception] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [anon_sym_extern] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7005),
    [anon_sym_POUNDload] = ACTIONS(7005),
    [anon_sym_POUNDtime] = ACTIONS(7005),
    [anon_sym_POUNDI] = ACTIONS(7005),
    [anon_sym_POUNDhelp] = ACTIONS(7005),
    [anon_sym_POUNDquit] = ACTIONS(7005),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(2787)] = {
    [sym_block_comment] = STATE(2787),
    [sym_compiler_directive_decl] = STATE(2787),
    [sym_fsi_directive_decl] = STATE(2787),
    [sym_preproc_line] = STATE(2787),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_module] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_open] = ACTIONS(7011),
    [anon_sym_type] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_and] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_exception] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [anon_sym_extern] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7009),
    [anon_sym_POUNDload] = ACTIONS(7009),
    [anon_sym_POUNDtime] = ACTIONS(7009),
    [anon_sym_POUNDI] = ACTIONS(7009),
    [anon_sym_POUNDhelp] = ACTIONS(7009),
    [anon_sym_POUNDquit] = ACTIONS(7009),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(2788)] = {
    [sym_block_comment] = STATE(2788),
    [sym_compiler_directive_decl] = STATE(2788),
    [sym_fsi_directive_decl] = STATE(2788),
    [sym_preproc_line] = STATE(2788),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_module] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_open] = ACTIONS(7019),
    [anon_sym_type] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_and] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_exception] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [anon_sym_extern] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7017),
    [anon_sym_POUNDload] = ACTIONS(7017),
    [anon_sym_POUNDtime] = ACTIONS(7017),
    [anon_sym_POUNDI] = ACTIONS(7017),
    [anon_sym_POUNDhelp] = ACTIONS(7017),
    [anon_sym_POUNDquit] = ACTIONS(7017),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(2789)] = {
    [sym_block_comment] = STATE(2789),
    [sym_compiler_directive_decl] = STATE(2789),
    [sym_fsi_directive_decl] = STATE(2789),
    [sym_preproc_line] = STATE(2789),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_module] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_open] = ACTIONS(7023),
    [anon_sym_type] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_and] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_exception] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [anon_sym_extern] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7021),
    [anon_sym_POUNDload] = ACTIONS(7021),
    [anon_sym_POUNDtime] = ACTIONS(7021),
    [anon_sym_POUNDI] = ACTIONS(7021),
    [anon_sym_POUNDhelp] = ACTIONS(7021),
    [anon_sym_POUNDquit] = ACTIONS(7021),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(2790)] = {
    [sym_block_comment] = STATE(2790),
    [sym_compiler_directive_decl] = STATE(2790),
    [sym_fsi_directive_decl] = STATE(2790),
    [sym_preproc_line] = STATE(2790),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_module] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_open] = ACTIONS(7027),
    [anon_sym_type] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_and] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_exception] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [anon_sym_extern] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7025),
    [anon_sym_POUNDload] = ACTIONS(7025),
    [anon_sym_POUNDtime] = ACTIONS(7025),
    [anon_sym_POUNDI] = ACTIONS(7025),
    [anon_sym_POUNDhelp] = ACTIONS(7025),
    [anon_sym_POUNDquit] = ACTIONS(7025),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(2791)] = {
    [sym__else_expression] = STATE(3627),
    [sym_elif_expression] = STATE(3078),
    [sym_block_comment] = STATE(2791),
    [sym_compiler_directive_decl] = STATE(2791),
    [sym_fsi_directive_decl] = STATE(2791),
    [sym_preproc_line] = STATE(2791),
    [aux_sym_if_expression_repeat1] = STATE(2799),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_and] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7322),
    [anon_sym_elif] = ACTIONS(7324),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2792)] = {
    [sym_block_comment] = STATE(2792),
    [sym_compiler_directive_decl] = STATE(2792),
    [sym_fsi_directive_decl] = STATE(2792),
    [sym_preproc_line] = STATE(2792),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_module] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_open] = ACTIONS(7035),
    [anon_sym_type] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_and] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_exception] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [anon_sym_extern] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7033),
    [anon_sym_POUNDload] = ACTIONS(7033),
    [anon_sym_POUNDtime] = ACTIONS(7033),
    [anon_sym_POUNDI] = ACTIONS(7033),
    [anon_sym_POUNDhelp] = ACTIONS(7033),
    [anon_sym_POUNDquit] = ACTIONS(7033),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(2793)] = {
    [sym_block_comment] = STATE(2793),
    [sym_compiler_directive_decl] = STATE(2793),
    [sym_fsi_directive_decl] = STATE(2793),
    [sym_preproc_line] = STATE(2793),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_module] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_open] = ACTIONS(7039),
    [anon_sym_type] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_and] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_exception] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [anon_sym_extern] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7037),
    [anon_sym_POUNDload] = ACTIONS(7037),
    [anon_sym_POUNDtime] = ACTIONS(7037),
    [anon_sym_POUNDI] = ACTIONS(7037),
    [anon_sym_POUNDhelp] = ACTIONS(7037),
    [anon_sym_POUNDquit] = ACTIONS(7037),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(2794)] = {
    [sym_block_comment] = STATE(2794),
    [sym_compiler_directive_decl] = STATE(2794),
    [sym_fsi_directive_decl] = STATE(2794),
    [sym_preproc_line] = STATE(2794),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_module] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_open] = ACTIONS(7043),
    [anon_sym_type] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_and] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_exception] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [anon_sym_extern] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7041),
    [anon_sym_POUNDload] = ACTIONS(7041),
    [anon_sym_POUNDtime] = ACTIONS(7041),
    [anon_sym_POUNDI] = ACTIONS(7041),
    [anon_sym_POUNDhelp] = ACTIONS(7041),
    [anon_sym_POUNDquit] = ACTIONS(7041),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(2795)] = {
    [sym__else_expression] = STATE(3432),
    [sym_elif_expression] = STATE(2969),
    [sym_block_comment] = STATE(2795),
    [sym_compiler_directive_decl] = STATE(2795),
    [sym_fsi_directive_decl] = STATE(2795),
    [sym_preproc_line] = STATE(2795),
    [aux_sym_if_expression_repeat1] = STATE(2830),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_and] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7318),
    [anon_sym_elif] = ACTIONS(7320),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__dedent] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2796)] = {
    [sym_block_comment] = STATE(2796),
    [sym_compiler_directive_decl] = STATE(2796),
    [sym_fsi_directive_decl] = STATE(2796),
    [sym_preproc_line] = STATE(2796),
    [aux_sym_compound_type_repeat1] = STATE(2798),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_DOT_DOT] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6726),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__dedent] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2797)] = {
    [sym_block_comment] = STATE(2797),
    [sym_compiler_directive_decl] = STATE(2797),
    [sym_fsi_directive_decl] = STATE(2797),
    [sym_preproc_line] = STATE(2797),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_module] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_open] = ACTIONS(7059),
    [anon_sym_type] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_and] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_exception] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [anon_sym_extern] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7057),
    [anon_sym_POUNDload] = ACTIONS(7057),
    [anon_sym_POUNDtime] = ACTIONS(7057),
    [anon_sym_POUNDI] = ACTIONS(7057),
    [anon_sym_POUNDhelp] = ACTIONS(7057),
    [anon_sym_POUNDquit] = ACTIONS(7057),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(2798)] = {
    [sym_block_comment] = STATE(2798),
    [sym_compiler_directive_decl] = STATE(2798),
    [sym_fsi_directive_decl] = STATE(2798),
    [sym_preproc_line] = STATE(2798),
    [aux_sym_compound_type_repeat1] = STATE(2798),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7326),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2799)] = {
    [sym__else_expression] = STATE(3377),
    [sym_elif_expression] = STATE(3078),
    [sym_block_comment] = STATE(2799),
    [sym_compiler_directive_decl] = STATE(2799),
    [sym_fsi_directive_decl] = STATE(2799),
    [sym_preproc_line] = STATE(2799),
    [aux_sym_if_expression_repeat1] = STATE(2875),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_and] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [aux_sym_access_modifier_token1] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7322),
    [anon_sym_elif] = ACTIONS(7324),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_default] = ACTIONS(6329),
    [anon_sym_static] = ACTIONS(6329),
    [anon_sym_member] = ACTIONS(6329),
    [anon_sym_interface] = ACTIONS(6329),
    [anon_sym_abstract] = ACTIONS(6329),
    [anon_sym_val] = ACTIONS(6329),
    [anon_sym_override] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(2800)] = {
    [sym__else_expression] = STATE(3413),
    [sym_elif_expression] = STATE(2977),
    [sym_block_comment] = STATE(2800),
    [sym_compiler_directive_decl] = STATE(2800),
    [sym_fsi_directive_decl] = STATE(2800),
    [sym_preproc_line] = STATE(2800),
    [aux_sym_if_expression_repeat1] = STATE(2759),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [aux_sym_access_modifier_token1] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7304),
    [anon_sym_elif] = ACTIONS(7306),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_DOT_DOT] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_default] = ACTIONS(6319),
    [anon_sym_static] = ACTIONS(6319),
    [anon_sym_member] = ACTIONS(6319),
    [anon_sym_interface] = ACTIONS(6319),
    [anon_sym_abstract] = ACTIONS(6319),
    [anon_sym_val] = ACTIONS(6319),
    [anon_sym_override] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__dedent] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(2801)] = {
    [sym_block_comment] = STATE(2801),
    [sym_compiler_directive_decl] = STATE(2801),
    [sym_fsi_directive_decl] = STATE(2801),
    [sym_preproc_line] = STATE(2801),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [anon_sym_inherit] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [anon_sym_POUNDelse] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2802)] = {
    [sym_block_comment] = STATE(2802),
    [sym_compiler_directive_decl] = STATE(2802),
    [sym_fsi_directive_decl] = STATE(2802),
    [sym_preproc_line] = STATE(2802),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [anon_sym_inherit] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [anon_sym_POUNDelse] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2803)] = {
    [sym_block_comment] = STATE(2803),
    [sym_compiler_directive_decl] = STATE(2803),
    [sym_fsi_directive_decl] = STATE(2803),
    [sym_preproc_line] = STATE(2803),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_and] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [anon_sym_POUNDelse] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2804)] = {
    [sym_block_comment] = STATE(2804),
    [sym_compiler_directive_decl] = STATE(2804),
    [sym_fsi_directive_decl] = STATE(2804),
    [sym_preproc_line] = STATE(2804),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(2805)] = {
    [sym_block_comment] = STATE(2805),
    [sym_compiler_directive_decl] = STATE(2805),
    [sym_fsi_directive_decl] = STATE(2805),
    [sym_preproc_line] = STATE(2805),
    [aux_sym_compound_type_repeat1] = STATE(2805),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7329),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2806)] = {
    [sym_block_comment] = STATE(2806),
    [sym_compiler_directive_decl] = STATE(2806),
    [sym_fsi_directive_decl] = STATE(2806),
    [sym_preproc_line] = STATE(2806),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_module] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_open] = ACTIONS(6892),
    [anon_sym_type] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_and] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_exception] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [anon_sym_extern] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6890),
    [anon_sym_POUNDload] = ACTIONS(6890),
    [anon_sym_POUNDtime] = ACTIONS(6890),
    [anon_sym_POUNDI] = ACTIONS(6890),
    [anon_sym_POUNDhelp] = ACTIONS(6890),
    [anon_sym_POUNDquit] = ACTIONS(6890),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(2807)] = {
    [sym_block_comment] = STATE(2807),
    [sym_compiler_directive_decl] = STATE(2807),
    [sym_fsi_directive_decl] = STATE(2807),
    [sym_preproc_line] = STATE(2807),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [anon_sym_inherit] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [anon_sym_POUNDelse] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2808)] = {
    [sym_block_comment] = STATE(2808),
    [sym_compiler_directive_decl] = STATE(2808),
    [sym_fsi_directive_decl] = STATE(2808),
    [sym_preproc_line] = STATE(2808),
    [aux_sym_type_argument_repeat1] = STATE(2716),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(7332),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2809)] = {
    [sym_block_comment] = STATE(2809),
    [sym_compiler_directive_decl] = STATE(2809),
    [sym_fsi_directive_decl] = STATE(2809),
    [sym_preproc_line] = STATE(2809),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_module] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_open] = ACTIONS(6896),
    [anon_sym_type] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_and] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_exception] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [anon_sym_extern] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6894),
    [anon_sym_POUNDload] = ACTIONS(6894),
    [anon_sym_POUNDtime] = ACTIONS(6894),
    [anon_sym_POUNDI] = ACTIONS(6894),
    [anon_sym_POUNDhelp] = ACTIONS(6894),
    [anon_sym_POUNDquit] = ACTIONS(6894),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(2810)] = {
    [sym_block_comment] = STATE(2810),
    [sym_compiler_directive_decl] = STATE(2810),
    [sym_fsi_directive_decl] = STATE(2810),
    [sym_preproc_line] = STATE(2810),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_module] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_open] = ACTIONS(6900),
    [anon_sym_type] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_and] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_exception] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [anon_sym_extern] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6898),
    [anon_sym_POUNDload] = ACTIONS(6898),
    [anon_sym_POUNDtime] = ACTIONS(6898),
    [anon_sym_POUNDI] = ACTIONS(6898),
    [anon_sym_POUNDhelp] = ACTIONS(6898),
    [anon_sym_POUNDquit] = ACTIONS(6898),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(2811)] = {
    [sym_block_comment] = STATE(2811),
    [sym_compiler_directive_decl] = STATE(2811),
    [sym_fsi_directive_decl] = STATE(2811),
    [sym_preproc_line] = STATE(2811),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_module] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_open] = ACTIONS(6902),
    [anon_sym_type] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_and] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_exception] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [anon_sym_extern] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6603),
    [anon_sym_POUNDload] = ACTIONS(6603),
    [anon_sym_POUNDtime] = ACTIONS(6603),
    [anon_sym_POUNDI] = ACTIONS(6603),
    [anon_sym_POUNDhelp] = ACTIONS(6603),
    [anon_sym_POUNDquit] = ACTIONS(6603),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(2812)] = {
    [sym_block_comment] = STATE(2812),
    [sym_compiler_directive_decl] = STATE(2812),
    [sym_fsi_directive_decl] = STATE(2812),
    [sym_preproc_line] = STATE(2812),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_module] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_open] = ACTIONS(6906),
    [anon_sym_type] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_and] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_exception] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [anon_sym_extern] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6904),
    [anon_sym_POUNDload] = ACTIONS(6904),
    [anon_sym_POUNDtime] = ACTIONS(6904),
    [anon_sym_POUNDI] = ACTIONS(6904),
    [anon_sym_POUNDhelp] = ACTIONS(6904),
    [anon_sym_POUNDquit] = ACTIONS(6904),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(2813)] = {
    [sym_block_comment] = STATE(2813),
    [sym_compiler_directive_decl] = STATE(2813),
    [sym_fsi_directive_decl] = STATE(2813),
    [sym_preproc_line] = STATE(2813),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_inherit] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2814)] = {
    [sym_elif_expression] = STATE(2938),
    [sym_block_comment] = STATE(2814),
    [sym_compiler_directive_decl] = STATE(2814),
    [sym_fsi_directive_decl] = STATE(2814),
    [sym_preproc_line] = STATE(2814),
    [aux_sym_if_expression_repeat1] = STATE(2814),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_and] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(7335),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [anon_sym_POUNDelse] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2815)] = {
    [sym_block_comment] = STATE(2815),
    [sym_compiler_directive_decl] = STATE(2815),
    [sym_fsi_directive_decl] = STATE(2815),
    [sym_preproc_line] = STATE(2815),
    [aux_sym_compound_type_repeat1] = STATE(2670),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(6758),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(2816)] = {
    [sym_block_comment] = STATE(2816),
    [sym_compiler_directive_decl] = STATE(2816),
    [sym_fsi_directive_decl] = STATE(2816),
    [sym_preproc_line] = STATE(2816),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_inherit] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2817)] = {
    [sym_block_comment] = STATE(2817),
    [sym_compiler_directive_decl] = STATE(2817),
    [sym_fsi_directive_decl] = STATE(2817),
    [sym_preproc_line] = STATE(2817),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_module] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_open] = ACTIONS(6888),
    [anon_sym_type] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_and] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_exception] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [anon_sym_extern] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6886),
    [anon_sym_POUNDload] = ACTIONS(6886),
    [anon_sym_POUNDtime] = ACTIONS(6886),
    [anon_sym_POUNDI] = ACTIONS(6886),
    [anon_sym_POUNDhelp] = ACTIONS(6886),
    [anon_sym_POUNDquit] = ACTIONS(6886),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(2818)] = {
    [sym_block_comment] = STATE(2818),
    [sym_compiler_directive_decl] = STATE(2818),
    [sym_fsi_directive_decl] = STATE(2818),
    [sym_preproc_line] = STATE(2818),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [anon_sym_inherit] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2819)] = {
    [sym_block_comment] = STATE(2819),
    [sym_compiler_directive_decl] = STATE(2819),
    [sym_fsi_directive_decl] = STATE(2819),
    [sym_preproc_line] = STATE(2819),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_DOT_DOT] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__dedent] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2820)] = {
    [sym_block_comment] = STATE(2820),
    [sym_compiler_directive_decl] = STATE(2820),
    [sym_fsi_directive_decl] = STATE(2820),
    [sym_preproc_line] = STATE(2820),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2821)] = {
    [sym_block_comment] = STATE(2821),
    [sym_compiler_directive_decl] = STATE(2821),
    [sym_fsi_directive_decl] = STATE(2821),
    [sym_preproc_line] = STATE(2821),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_and] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2822)] = {
    [sym_block_comment] = STATE(2822),
    [sym_compiler_directive_decl] = STATE(2822),
    [sym_fsi_directive_decl] = STATE(2822),
    [sym_preproc_line] = STATE(2822),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_DASH_GT] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7338),
    [anon_sym_uy] = ACTIONS(7340),
    [anon_sym_s] = ACTIONS(7342),
    [anon_sym_us] = ACTIONS(7344),
    [anon_sym_l] = ACTIONS(7346),
    [aux_sym_uint32_token1] = ACTIONS(7348),
    [anon_sym_n] = ACTIONS(7350),
    [anon_sym_un] = ACTIONS(7352),
    [anon_sym_L] = ACTIONS(7354),
    [aux_sym_uint64_token1] = ACTIONS(7356),
    [aux_sym_bignum_token1] = ACTIONS(7358),
    [aux_sym_decimal_token1] = ACTIONS(7360),
    [anon_sym_DOT2] = ACTIONS(7362),
    [aux_sym_float_token1] = ACTIONS(7364),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2823)] = {
    [sym_block_comment] = STATE(2823),
    [sym_compiler_directive_decl] = STATE(2823),
    [sym_fsi_directive_decl] = STATE(2823),
    [sym_preproc_line] = STATE(2823),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_and] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__dedent] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2824)] = {
    [sym_block_comment] = STATE(2824),
    [sym_compiler_directive_decl] = STATE(2824),
    [sym_fsi_directive_decl] = STATE(2824),
    [sym_preproc_line] = STATE(2824),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_and] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__dedent] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2825)] = {
    [sym_block_comment] = STATE(2825),
    [sym_compiler_directive_decl] = STATE(2825),
    [sym_fsi_directive_decl] = STATE(2825),
    [sym_preproc_line] = STATE(2825),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_and] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__dedent] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2826)] = {
    [sym_block_comment] = STATE(2826),
    [sym_compiler_directive_decl] = STATE(2826),
    [sym_fsi_directive_decl] = STATE(2826),
    [sym_preproc_line] = STATE(2826),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2827)] = {
    [sym_block_comment] = STATE(2827),
    [sym_compiler_directive_decl] = STATE(2827),
    [sym_fsi_directive_decl] = STATE(2827),
    [sym_preproc_line] = STATE(2827),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_and] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__dedent] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2828)] = {
    [sym_block_comment] = STATE(2828),
    [sym_compiler_directive_decl] = STATE(2828),
    [sym_fsi_directive_decl] = STATE(2828),
    [sym_preproc_line] = STATE(2828),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_and] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2829)] = {
    [sym_block_comment] = STATE(2829),
    [sym_compiler_directive_decl] = STATE(2829),
    [sym_fsi_directive_decl] = STATE(2829),
    [sym_preproc_line] = STATE(2829),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [anon_sym_inherit] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2830)] = {
    [sym_elif_expression] = STATE(2969),
    [sym_block_comment] = STATE(2830),
    [sym_compiler_directive_decl] = STATE(2830),
    [sym_fsi_directive_decl] = STATE(2830),
    [sym_preproc_line] = STATE(2830),
    [aux_sym_if_expression_repeat1] = STATE(2830),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_and] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(7366),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__dedent] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2831)] = {
    [sym_block_comment] = STATE(2831),
    [sym_compiler_directive_decl] = STATE(2831),
    [sym_fsi_directive_decl] = STATE(2831),
    [sym_preproc_line] = STATE(2831),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_and] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2832)] = {
    [sym_block_comment] = STATE(2832),
    [sym_compiler_directive_decl] = STATE(2832),
    [sym_fsi_directive_decl] = STATE(2832),
    [sym_preproc_line] = STATE(2832),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2833)] = {
    [sym_block_comment] = STATE(2833),
    [sym_compiler_directive_decl] = STATE(2833),
    [sym_fsi_directive_decl] = STATE(2833),
    [sym_preproc_line] = STATE(2833),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [anon_sym_inherit] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2834)] = {
    [sym_block_comment] = STATE(2834),
    [sym_compiler_directive_decl] = STATE(2834),
    [sym_fsi_directive_decl] = STATE(2834),
    [sym_preproc_line] = STATE(2834),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [anon_sym_inherit] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2835)] = {
    [sym_block_comment] = STATE(2835),
    [sym_compiler_directive_decl] = STATE(2835),
    [sym_fsi_directive_decl] = STATE(2835),
    [sym_preproc_line] = STATE(2835),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2836)] = {
    [sym_block_comment] = STATE(2836),
    [sym_compiler_directive_decl] = STATE(2836),
    [sym_fsi_directive_decl] = STATE(2836),
    [sym_preproc_line] = STATE(2836),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_and] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__dedent] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2837)] = {
    [sym_elif_expression] = STATE(2999),
    [sym_block_comment] = STATE(2837),
    [sym_compiler_directive_decl] = STATE(2837),
    [sym_fsi_directive_decl] = STATE(2837),
    [sym_preproc_line] = STATE(2837),
    [aux_sym_if_expression_repeat1] = STATE(2837),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(7369),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [anon_sym_inherit] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2838)] = {
    [sym_block_comment] = STATE(2838),
    [sym_compiler_directive_decl] = STATE(2838),
    [sym_fsi_directive_decl] = STATE(2838),
    [sym_preproc_line] = STATE(2838),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_inherit] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(7372),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2839)] = {
    [sym_block_comment] = STATE(2839),
    [sym_compiler_directive_decl] = STATE(2839),
    [sym_fsi_directive_decl] = STATE(2839),
    [sym_preproc_line] = STATE(2839),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2840)] = {
    [sym_block_comment] = STATE(2840),
    [sym_compiler_directive_decl] = STATE(2840),
    [sym_fsi_directive_decl] = STATE(2840),
    [sym_preproc_line] = STATE(2840),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [anon_sym_inherit] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2841)] = {
    [sym_block_comment] = STATE(2841),
    [sym_compiler_directive_decl] = STATE(2841),
    [sym_fsi_directive_decl] = STATE(2841),
    [sym_preproc_line] = STATE(2841),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_and] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__dedent] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2842)] = {
    [sym_elif_expression] = STATE(2977),
    [sym_block_comment] = STATE(2842),
    [sym_compiler_directive_decl] = STATE(2842),
    [sym_fsi_directive_decl] = STATE(2842),
    [sym_preproc_line] = STATE(2842),
    [aux_sym_if_expression_repeat1] = STATE(2842),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(7374),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_DOT_DOT] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__dedent] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2843)] = {
    [sym_block_comment] = STATE(2843),
    [sym_compiler_directive_decl] = STATE(2843),
    [sym_fsi_directive_decl] = STATE(2843),
    [sym_preproc_line] = STATE(2843),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_and] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2844)] = {
    [sym_block_comment] = STATE(2844),
    [sym_compiler_directive_decl] = STATE(2844),
    [sym_fsi_directive_decl] = STATE(2844),
    [sym_preproc_line] = STATE(2844),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [anon_sym_inherit] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2845)] = {
    [sym_block_comment] = STATE(2845),
    [sym_compiler_directive_decl] = STATE(2845),
    [sym_fsi_directive_decl] = STATE(2845),
    [sym_preproc_line] = STATE(2845),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_and] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__dedent] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2846)] = {
    [sym_block_comment] = STATE(2846),
    [sym_compiler_directive_decl] = STATE(2846),
    [sym_fsi_directive_decl] = STATE(2846),
    [sym_preproc_line] = STATE(2846),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2847)] = {
    [sym_block_comment] = STATE(2847),
    [sym_compiler_directive_decl] = STATE(2847),
    [sym_fsi_directive_decl] = STATE(2847),
    [sym_preproc_line] = STATE(2847),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__dedent] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2848)] = {
    [sym_block_comment] = STATE(2848),
    [sym_compiler_directive_decl] = STATE(2848),
    [sym_fsi_directive_decl] = STATE(2848),
    [sym_preproc_line] = STATE(2848),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [anon_sym_inherit] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2849)] = {
    [sym_block_comment] = STATE(2849),
    [sym_compiler_directive_decl] = STATE(2849),
    [sym_fsi_directive_decl] = STATE(2849),
    [sym_preproc_line] = STATE(2849),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_RBRACK] = ACTIONS(5953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6696),
    [anon_sym_uy] = ACTIONS(6698),
    [anon_sym_s] = ACTIONS(6700),
    [anon_sym_us] = ACTIONS(6702),
    [anon_sym_l] = ACTIONS(6704),
    [aux_sym_uint32_token1] = ACTIONS(6706),
    [anon_sym_n] = ACTIONS(6708),
    [anon_sym_un] = ACTIONS(6710),
    [anon_sym_L] = ACTIONS(6712),
    [aux_sym_uint64_token1] = ACTIONS(6714),
    [aux_sym_bignum_token1] = ACTIONS(6716),
    [aux_sym_decimal_token1] = ACTIONS(6718),
    [anon_sym_DOT2] = ACTIONS(6720),
    [aux_sym_float_token1] = ACTIONS(6722),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2850)] = {
    [sym_block_comment] = STATE(2850),
    [sym_compiler_directive_decl] = STATE(2850),
    [sym_fsi_directive_decl] = STATE(2850),
    [sym_preproc_line] = STATE(2850),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2851)] = {
    [sym_block_comment] = STATE(2851),
    [sym_compiler_directive_decl] = STATE(2851),
    [sym_fsi_directive_decl] = STATE(2851),
    [sym_preproc_line] = STATE(2851),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_and] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2852)] = {
    [sym_block_comment] = STATE(2852),
    [sym_compiler_directive_decl] = STATE(2852),
    [sym_fsi_directive_decl] = STATE(2852),
    [sym_preproc_line] = STATE(2852),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_DOT_DOT] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(7377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__dedent] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2853)] = {
    [sym_block_comment] = STATE(2853),
    [sym_compiler_directive_decl] = STATE(2853),
    [sym_fsi_directive_decl] = STATE(2853),
    [sym_preproc_line] = STATE(2853),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7379),
    [anon_sym_uy] = ACTIONS(7381),
    [anon_sym_s] = ACTIONS(7383),
    [anon_sym_us] = ACTIONS(7385),
    [anon_sym_l] = ACTIONS(7387),
    [aux_sym_uint32_token1] = ACTIONS(7389),
    [anon_sym_n] = ACTIONS(7391),
    [anon_sym_un] = ACTIONS(7393),
    [anon_sym_L] = ACTIONS(7395),
    [aux_sym_uint64_token1] = ACTIONS(7397),
    [aux_sym_bignum_token1] = ACTIONS(7399),
    [aux_sym_decimal_token1] = ACTIONS(7401),
    [anon_sym_DOT2] = ACTIONS(7403),
    [aux_sym_float_token1] = ACTIONS(7405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2854)] = {
    [sym_block_comment] = STATE(2854),
    [sym_compiler_directive_decl] = STATE(2854),
    [sym_fsi_directive_decl] = STATE(2854),
    [sym_preproc_line] = STATE(2854),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2855)] = {
    [sym_block_comment] = STATE(2855),
    [sym_compiler_directive_decl] = STATE(2855),
    [sym_fsi_directive_decl] = STATE(2855),
    [sym_preproc_line] = STATE(2855),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_and] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2856)] = {
    [sym_block_comment] = STATE(2856),
    [sym_compiler_directive_decl] = STATE(2856),
    [sym_fsi_directive_decl] = STATE(2856),
    [sym_preproc_line] = STATE(2856),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [anon_sym_inherit] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2857)] = {
    [sym_block_comment] = STATE(2857),
    [sym_compiler_directive_decl] = STATE(2857),
    [sym_fsi_directive_decl] = STATE(2857),
    [sym_preproc_line] = STATE(2857),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_inherit] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2858)] = {
    [sym_block_comment] = STATE(2858),
    [sym_compiler_directive_decl] = STATE(2858),
    [sym_fsi_directive_decl] = STATE(2858),
    [sym_preproc_line] = STATE(2858),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_DOT_DOT] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__dedent] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2859)] = {
    [sym_block_comment] = STATE(2859),
    [sym_compiler_directive_decl] = STATE(2859),
    [sym_fsi_directive_decl] = STATE(2859),
    [sym_preproc_line] = STATE(2859),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_and] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2860)] = {
    [sym_block_comment] = STATE(2860),
    [sym_compiler_directive_decl] = STATE(2860),
    [sym_fsi_directive_decl] = STATE(2860),
    [sym_preproc_line] = STATE(2860),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_and] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__dedent] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2861)] = {
    [sym_block_comment] = STATE(2861),
    [sym_compiler_directive_decl] = STATE(2861),
    [sym_fsi_directive_decl] = STATE(2861),
    [sym_preproc_line] = STATE(2861),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_DOT_DOT] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__dedent] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2862)] = {
    [sym_block_comment] = STATE(2862),
    [sym_compiler_directive_decl] = STATE(2862),
    [sym_fsi_directive_decl] = STATE(2862),
    [sym_preproc_line] = STATE(2862),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_inherit] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2863)] = {
    [sym_block_comment] = STATE(2863),
    [sym_compiler_directive_decl] = STATE(2863),
    [sym_fsi_directive_decl] = STATE(2863),
    [sym_preproc_line] = STATE(2863),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2864)] = {
    [sym_block_comment] = STATE(2864),
    [sym_compiler_directive_decl] = STATE(2864),
    [sym_fsi_directive_decl] = STATE(2864),
    [sym_preproc_line] = STATE(2864),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_DOT_DOT] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__dedent] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2865)] = {
    [sym_block_comment] = STATE(2865),
    [sym_compiler_directive_decl] = STATE(2865),
    [sym_fsi_directive_decl] = STATE(2865),
    [sym_preproc_line] = STATE(2865),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(2866)] = {
    [sym_block_comment] = STATE(2866),
    [sym_compiler_directive_decl] = STATE(2866),
    [sym_fsi_directive_decl] = STATE(2866),
    [sym_preproc_line] = STATE(2866),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_and] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__dedent] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2867)] = {
    [sym_block_comment] = STATE(2867),
    [sym_compiler_directive_decl] = STATE(2867),
    [sym_fsi_directive_decl] = STATE(2867),
    [sym_preproc_line] = STATE(2867),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_DOT_DOT] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__dedent] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2868)] = {
    [sym_block_comment] = STATE(2868),
    [sym_compiler_directive_decl] = STATE(2868),
    [sym_fsi_directive_decl] = STATE(2868),
    [sym_preproc_line] = STATE(2868),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_DOT_DOT] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__dedent] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2869)] = {
    [sym_block_comment] = STATE(2869),
    [sym_compiler_directive_decl] = STATE(2869),
    [sym_fsi_directive_decl] = STATE(2869),
    [sym_preproc_line] = STATE(2869),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [aux_sym_access_modifier_token1] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_DOT_DOT] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_default] = ACTIONS(6389),
    [anon_sym_static] = ACTIONS(6389),
    [anon_sym_member] = ACTIONS(6389),
    [anon_sym_interface] = ACTIONS(6389),
    [anon_sym_abstract] = ACTIONS(6389),
    [anon_sym_val] = ACTIONS(6389),
    [anon_sym_override] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__dedent] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(2870)] = {
    [sym_block_comment] = STATE(2870),
    [sym_compiler_directive_decl] = STATE(2870),
    [sym_fsi_directive_decl] = STATE(2870),
    [sym_preproc_line] = STATE(2870),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [anon_sym_inherit] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2871)] = {
    [sym_block_comment] = STATE(2871),
    [sym_compiler_directive_decl] = STATE(2871),
    [sym_fsi_directive_decl] = STATE(2871),
    [sym_preproc_line] = STATE(2871),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [aux_sym_access_modifier_token1] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_default] = ACTIONS(6492),
    [anon_sym_static] = ACTIONS(6492),
    [anon_sym_member] = ACTIONS(6492),
    [anon_sym_interface] = ACTIONS(6492),
    [anon_sym_abstract] = ACTIONS(6492),
    [anon_sym_val] = ACTIONS(6492),
    [anon_sym_override] = ACTIONS(6492),
    [anon_sym_inherit] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(2872)] = {
    [sym_block_comment] = STATE(2872),
    [sym_compiler_directive_decl] = STATE(2872),
    [sym_fsi_directive_decl] = STATE(2872),
    [sym_preproc_line] = STATE(2872),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_and] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2873)] = {
    [sym_block_comment] = STATE(2873),
    [sym_compiler_directive_decl] = STATE(2873),
    [sym_fsi_directive_decl] = STATE(2873),
    [sym_preproc_line] = STATE(2873),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_DOT_DOT] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2874)] = {
    [sym_block_comment] = STATE(2874),
    [sym_compiler_directive_decl] = STATE(2874),
    [sym_fsi_directive_decl] = STATE(2874),
    [sym_preproc_line] = STATE(2874),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [aux_sym_access_modifier_token1] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_DOT_DOT] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_default] = ACTIONS(6395),
    [anon_sym_static] = ACTIONS(6395),
    [anon_sym_member] = ACTIONS(6395),
    [anon_sym_interface] = ACTIONS(6395),
    [anon_sym_abstract] = ACTIONS(6395),
    [anon_sym_val] = ACTIONS(6395),
    [anon_sym_override] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__dedent] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(2875)] = {
    [sym_elif_expression] = STATE(3078),
    [sym_block_comment] = STATE(2875),
    [sym_compiler_directive_decl] = STATE(2875),
    [sym_fsi_directive_decl] = STATE(2875),
    [sym_preproc_line] = STATE(2875),
    [aux_sym_if_expression_repeat1] = STATE(2875),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_and] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [aux_sym_access_modifier_token1] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(7407),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_default] = ACTIONS(6496),
    [anon_sym_static] = ACTIONS(6496),
    [anon_sym_member] = ACTIONS(6496),
    [anon_sym_interface] = ACTIONS(6496),
    [anon_sym_abstract] = ACTIONS(6496),
    [anon_sym_val] = ACTIONS(6496),
    [anon_sym_override] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(2876)] = {
    [sym_block_comment] = STATE(2876),
    [sym_compiler_directive_decl] = STATE(2876),
    [sym_fsi_directive_decl] = STATE(2876),
    [sym_preproc_line] = STATE(2876),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_DOT_DOT] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__dedent] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2877)] = {
    [sym_block_comment] = STATE(2877),
    [sym_compiler_directive_decl] = STATE(2877),
    [sym_fsi_directive_decl] = STATE(2877),
    [sym_preproc_line] = STATE(2877),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_DOT_DOT] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__dedent] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2878)] = {
    [sym_block_comment] = STATE(2878),
    [sym_compiler_directive_decl] = STATE(2878),
    [sym_fsi_directive_decl] = STATE(2878),
    [sym_preproc_line] = STATE(2878),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(7410),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2879)] = {
    [sym_block_comment] = STATE(2879),
    [sym_compiler_directive_decl] = STATE(2879),
    [sym_fsi_directive_decl] = STATE(2879),
    [sym_preproc_line] = STATE(2879),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_DOT_DOT] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__dedent] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2880)] = {
    [sym_block_comment] = STATE(2880),
    [sym_compiler_directive_decl] = STATE(2880),
    [sym_fsi_directive_decl] = STATE(2880),
    [sym_preproc_line] = STATE(2880),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_DOT_DOT] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__dedent] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2881)] = {
    [sym_block_comment] = STATE(2881),
    [sym_compiler_directive_decl] = STATE(2881),
    [sym_fsi_directive_decl] = STATE(2881),
    [sym_preproc_line] = STATE(2881),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [aux_sym_access_modifier_token1] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_default] = ACTIONS(6411),
    [anon_sym_static] = ACTIONS(6411),
    [anon_sym_member] = ACTIONS(6411),
    [anon_sym_interface] = ACTIONS(6411),
    [anon_sym_abstract] = ACTIONS(6411),
    [anon_sym_val] = ACTIONS(6411),
    [anon_sym_override] = ACTIONS(6411),
    [anon_sym_inherit] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(2882)] = {
    [sym_block_comment] = STATE(2882),
    [sym_compiler_directive_decl] = STATE(2882),
    [sym_fsi_directive_decl] = STATE(2882),
    [sym_preproc_line] = STATE(2882),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_and] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [aux_sym_access_modifier_token1] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_default] = ACTIONS(6403),
    [anon_sym_static] = ACTIONS(6403),
    [anon_sym_member] = ACTIONS(6403),
    [anon_sym_interface] = ACTIONS(6403),
    [anon_sym_abstract] = ACTIONS(6403),
    [anon_sym_val] = ACTIONS(6403),
    [anon_sym_override] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__dedent] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(2883)] = {
    [sym_block_comment] = STATE(2883),
    [sym_compiler_directive_decl] = STATE(2883),
    [sym_fsi_directive_decl] = STATE(2883),
    [sym_preproc_line] = STATE(2883),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2884)] = {
    [sym_block_comment] = STATE(2884),
    [sym_compiler_directive_decl] = STATE(2884),
    [sym_fsi_directive_decl] = STATE(2884),
    [sym_preproc_line] = STATE(2884),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2885)] = {
    [sym_block_comment] = STATE(2885),
    [sym_compiler_directive_decl] = STATE(2885),
    [sym_fsi_directive_decl] = STATE(2885),
    [sym_preproc_line] = STATE(2885),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [aux_sym_access_modifier_token1] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_default] = ACTIONS(6427),
    [anon_sym_static] = ACTIONS(6427),
    [anon_sym_member] = ACTIONS(6427),
    [anon_sym_interface] = ACTIONS(6427),
    [anon_sym_abstract] = ACTIONS(6427),
    [anon_sym_val] = ACTIONS(6427),
    [anon_sym_override] = ACTIONS(6427),
    [anon_sym_inherit] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(2886)] = {
    [sym_block_comment] = STATE(2886),
    [sym_compiler_directive_decl] = STATE(2886),
    [sym_fsi_directive_decl] = STATE(2886),
    [sym_preproc_line] = STATE(2886),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2887)] = {
    [sym_block_comment] = STATE(2887),
    [sym_compiler_directive_decl] = STATE(2887),
    [sym_fsi_directive_decl] = STATE(2887),
    [sym_preproc_line] = STATE(2887),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2888)] = {
    [sym_block_comment] = STATE(2888),
    [sym_compiler_directive_decl] = STATE(2888),
    [sym_fsi_directive_decl] = STATE(2888),
    [sym_preproc_line] = STATE(2888),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2889)] = {
    [sym_block_comment] = STATE(2889),
    [sym_compiler_directive_decl] = STATE(2889),
    [sym_fsi_directive_decl] = STATE(2889),
    [sym_preproc_line] = STATE(2889),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2890)] = {
    [sym_block_comment] = STATE(2890),
    [sym_compiler_directive_decl] = STATE(2890),
    [sym_fsi_directive_decl] = STATE(2890),
    [sym_preproc_line] = STATE(2890),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_and] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(7412),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2891)] = {
    [sym_block_comment] = STATE(2891),
    [sym_compiler_directive_decl] = STATE(2891),
    [sym_fsi_directive_decl] = STATE(2891),
    [sym_preproc_line] = STATE(2891),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7414),
    [anon_sym_uy] = ACTIONS(7416),
    [anon_sym_s] = ACTIONS(7418),
    [anon_sym_us] = ACTIONS(7420),
    [anon_sym_l] = ACTIONS(7422),
    [aux_sym_uint32_token1] = ACTIONS(7424),
    [anon_sym_n] = ACTIONS(7426),
    [anon_sym_un] = ACTIONS(7428),
    [anon_sym_L] = ACTIONS(7430),
    [aux_sym_uint64_token1] = ACTIONS(7432),
    [aux_sym_bignum_token1] = ACTIONS(7434),
    [aux_sym_decimal_token1] = ACTIONS(7436),
    [anon_sym_DOT2] = ACTIONS(7438),
    [aux_sym_float_token1] = ACTIONS(7440),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
    [sym__in] = ACTIONS(5953),
  },
  [STATE(2892)] = {
    [sym_block_comment] = STATE(2892),
    [sym_compiler_directive_decl] = STATE(2892),
    [sym_fsi_directive_decl] = STATE(2892),
    [sym_preproc_line] = STATE(2892),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_inherit] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2893)] = {
    [sym_block_comment] = STATE(2893),
    [sym_compiler_directive_decl] = STATE(2893),
    [sym_fsi_directive_decl] = STATE(2893),
    [sym_preproc_line] = STATE(2893),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_and] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [aux_sym_access_modifier_token1] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_default] = ACTIONS(6407),
    [anon_sym_static] = ACTIONS(6407),
    [anon_sym_member] = ACTIONS(6407),
    [anon_sym_interface] = ACTIONS(6407),
    [anon_sym_abstract] = ACTIONS(6407),
    [anon_sym_val] = ACTIONS(6407),
    [anon_sym_override] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(2894)] = {
    [sym_block_comment] = STATE(2894),
    [sym_compiler_directive_decl] = STATE(2894),
    [sym_fsi_directive_decl] = STATE(2894),
    [sym_preproc_line] = STATE(2894),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_and] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(7442),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__dedent] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2895)] = {
    [sym_block_comment] = STATE(2895),
    [sym_compiler_directive_decl] = STATE(2895),
    [sym_fsi_directive_decl] = STATE(2895),
    [sym_preproc_line] = STATE(2895),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_and] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [aux_sym_access_modifier_token1] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_default] = ACTIONS(6367),
    [anon_sym_static] = ACTIONS(6367),
    [anon_sym_member] = ACTIONS(6367),
    [anon_sym_interface] = ACTIONS(6367),
    [anon_sym_abstract] = ACTIONS(6367),
    [anon_sym_val] = ACTIONS(6367),
    [anon_sym_override] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(2896)] = {
    [sym_block_comment] = STATE(2896),
    [sym_compiler_directive_decl] = STATE(2896),
    [sym_fsi_directive_decl] = STATE(2896),
    [sym_preproc_line] = STATE(2896),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_inherit] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(2897)] = {
    [sym_block_comment] = STATE(2897),
    [sym_compiler_directive_decl] = STATE(2897),
    [sym_fsi_directive_decl] = STATE(2897),
    [sym_preproc_line] = STATE(2897),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7444),
    [anon_sym_uy] = ACTIONS(7446),
    [anon_sym_s] = ACTIONS(7448),
    [anon_sym_us] = ACTIONS(7450),
    [anon_sym_l] = ACTIONS(7452),
    [aux_sym_uint32_token1] = ACTIONS(7454),
    [anon_sym_n] = ACTIONS(7456),
    [anon_sym_un] = ACTIONS(7458),
    [anon_sym_L] = ACTIONS(7460),
    [aux_sym_uint64_token1] = ACTIONS(7462),
    [aux_sym_bignum_token1] = ACTIONS(7464),
    [aux_sym_decimal_token1] = ACTIONS(7466),
    [anon_sym_DOT2] = ACTIONS(7468),
    [aux_sym_float_token1] = ACTIONS(7470),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2898)] = {
    [sym_block_comment] = STATE(2898),
    [sym_compiler_directive_decl] = STATE(2898),
    [sym_fsi_directive_decl] = STATE(2898),
    [sym_preproc_line] = STATE(2898),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7472),
    [anon_sym_uy] = ACTIONS(7474),
    [anon_sym_s] = ACTIONS(7476),
    [anon_sym_us] = ACTIONS(7478),
    [anon_sym_l] = ACTIONS(7480),
    [aux_sym_uint32_token1] = ACTIONS(7482),
    [anon_sym_n] = ACTIONS(7484),
    [anon_sym_un] = ACTIONS(7486),
    [anon_sym_L] = ACTIONS(7488),
    [aux_sym_uint64_token1] = ACTIONS(7490),
    [aux_sym_bignum_token1] = ACTIONS(7492),
    [aux_sym_decimal_token1] = ACTIONS(7494),
    [anon_sym_DOT2] = ACTIONS(7496),
    [aux_sym_float_token1] = ACTIONS(7498),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2899)] = {
    [sym_block_comment] = STATE(2899),
    [sym_compiler_directive_decl] = STATE(2899),
    [sym_fsi_directive_decl] = STATE(2899),
    [sym_preproc_line] = STATE(2899),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7163),
    [anon_sym_uy] = ACTIONS(7165),
    [anon_sym_s] = ACTIONS(7167),
    [anon_sym_us] = ACTIONS(7169),
    [anon_sym_l] = ACTIONS(7171),
    [aux_sym_uint32_token1] = ACTIONS(7173),
    [anon_sym_n] = ACTIONS(7175),
    [anon_sym_un] = ACTIONS(7177),
    [anon_sym_L] = ACTIONS(7179),
    [aux_sym_uint64_token1] = ACTIONS(7181),
    [anon_sym_lf] = ACTIONS(7500),
    [anon_sym_LF] = ACTIONS(7502),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2900)] = {
    [sym_block_comment] = STATE(2900),
    [sym_compiler_directive_decl] = STATE(2900),
    [sym_fsi_directive_decl] = STATE(2900),
    [sym_preproc_line] = STATE(2900),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_and] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [aux_sym_access_modifier_token1] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_default] = ACTIONS(6399),
    [anon_sym_static] = ACTIONS(6399),
    [anon_sym_member] = ACTIONS(6399),
    [anon_sym_interface] = ACTIONS(6399),
    [anon_sym_abstract] = ACTIONS(6399),
    [anon_sym_val] = ACTIONS(6399),
    [anon_sym_override] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__dedent] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(2901)] = {
    [sym_block_comment] = STATE(2901),
    [sym_compiler_directive_decl] = STATE(2901),
    [sym_fsi_directive_decl] = STATE(2901),
    [sym_preproc_line] = STATE(2901),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_and] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [aux_sym_access_modifier_token1] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_default] = ACTIONS(6391),
    [anon_sym_static] = ACTIONS(6391),
    [anon_sym_member] = ACTIONS(6391),
    [anon_sym_interface] = ACTIONS(6391),
    [anon_sym_abstract] = ACTIONS(6391),
    [anon_sym_val] = ACTIONS(6391),
    [anon_sym_override] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(2902)] = {
    [sym_block_comment] = STATE(2902),
    [sym_compiler_directive_decl] = STATE(2902),
    [sym_fsi_directive_decl] = STATE(2902),
    [sym_preproc_line] = STATE(2902),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7276),
    [anon_sym_uy] = ACTIONS(7278),
    [anon_sym_s] = ACTIONS(7280),
    [anon_sym_us] = ACTIONS(7282),
    [anon_sym_l] = ACTIONS(7284),
    [aux_sym_uint32_token1] = ACTIONS(7286),
    [anon_sym_n] = ACTIONS(7288),
    [anon_sym_un] = ACTIONS(7290),
    [anon_sym_L] = ACTIONS(7292),
    [aux_sym_uint64_token1] = ACTIONS(7294),
    [aux_sym_bignum_token1] = ACTIONS(7296),
    [aux_sym_decimal_token1] = ACTIONS(7298),
    [anon_sym_DOT2] = ACTIONS(7300),
    [aux_sym_float_token1] = ACTIONS(7302),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2903)] = {
    [sym_block_comment] = STATE(2903),
    [sym_compiler_directive_decl] = STATE(2903),
    [sym_fsi_directive_decl] = STATE(2903),
    [sym_preproc_line] = STATE(2903),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_and] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [aux_sym_access_modifier_token1] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_default] = ACTIONS(6379),
    [anon_sym_static] = ACTIONS(6379),
    [anon_sym_member] = ACTIONS(6379),
    [anon_sym_interface] = ACTIONS(6379),
    [anon_sym_abstract] = ACTIONS(6379),
    [anon_sym_val] = ACTIONS(6379),
    [anon_sym_override] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(2904)] = {
    [sym_block_comment] = STATE(2904),
    [sym_compiler_directive_decl] = STATE(2904),
    [sym_fsi_directive_decl] = STATE(2904),
    [sym_preproc_line] = STATE(2904),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_and] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [aux_sym_access_modifier_token1] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_default] = ACTIONS(6383),
    [anon_sym_static] = ACTIONS(6383),
    [anon_sym_member] = ACTIONS(6383),
    [anon_sym_interface] = ACTIONS(6383),
    [anon_sym_abstract] = ACTIONS(6383),
    [anon_sym_val] = ACTIONS(6383),
    [anon_sym_override] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(2905)] = {
    [sym_block_comment] = STATE(2905),
    [sym_compiler_directive_decl] = STATE(2905),
    [sym_fsi_directive_decl] = STATE(2905),
    [sym_preproc_line] = STATE(2905),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(2906)] = {
    [sym_block_comment] = STATE(2906),
    [sym_compiler_directive_decl] = STATE(2906),
    [sym_fsi_directive_decl] = STATE(2906),
    [sym_preproc_line] = STATE(2906),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [aux_sym_access_modifier_token1] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(7504),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_default] = ACTIONS(6373),
    [anon_sym_static] = ACTIONS(6373),
    [anon_sym_member] = ACTIONS(6373),
    [anon_sym_interface] = ACTIONS(6373),
    [anon_sym_abstract] = ACTIONS(6373),
    [anon_sym_val] = ACTIONS(6373),
    [anon_sym_override] = ACTIONS(6373),
    [anon_sym_inherit] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(2907)] = {
    [sym_block_comment] = STATE(2907),
    [sym_compiler_directive_decl] = STATE(2907),
    [sym_fsi_directive_decl] = STATE(2907),
    [sym_preproc_line] = STATE(2907),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_inherit] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(2908)] = {
    [sym_block_comment] = STATE(2908),
    [sym_compiler_directive_decl] = STATE(2908),
    [sym_fsi_directive_decl] = STATE(2908),
    [sym_preproc_line] = STATE(2908),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7506),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_inherit] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2909)] = {
    [sym_block_comment] = STATE(2909),
    [sym_compiler_directive_decl] = STATE(2909),
    [sym_fsi_directive_decl] = STATE(2909),
    [sym_preproc_line] = STATE(2909),
    [aux_sym_rules_repeat1] = STATE(2941),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7508),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7510),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2910)] = {
    [sym_block_comment] = STATE(2910),
    [sym_compiler_directive_decl] = STATE(2910),
    [sym_fsi_directive_decl] = STATE(2910),
    [sym_preproc_line] = STATE(2910),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [anon_sym_inherit] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [anon_sym_POUNDelse] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(2911)] = {
    [sym_block_comment] = STATE(2911),
    [sym_compiler_directive_decl] = STATE(2911),
    [sym_fsi_directive_decl] = STATE(2911),
    [sym_preproc_line] = STATE(2911),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2912)] = {
    [sym_block_comment] = STATE(2912),
    [sym_compiler_directive_decl] = STATE(2912),
    [sym_fsi_directive_decl] = STATE(2912),
    [sym_preproc_line] = STATE(2912),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6205),
    [aux_sym_decimal_token1] = ACTIONS(6101),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2913)] = {
    [sym_block_comment] = STATE(2913),
    [sym_compiler_directive_decl] = STATE(2913),
    [sym_fsi_directive_decl] = STATE(2913),
    [sym_preproc_line] = STATE(2913),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(7513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2914)] = {
    [sym_block_comment] = STATE(2914),
    [sym_compiler_directive_decl] = STATE(2914),
    [sym_fsi_directive_decl] = STATE(2914),
    [sym_preproc_line] = STATE(2914),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_inherit] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(7515),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2915)] = {
    [sym_block_comment] = STATE(2915),
    [sym_compiler_directive_decl] = STATE(2915),
    [sym_fsi_directive_decl] = STATE(2915),
    [sym_preproc_line] = STATE(2915),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7517),
    [anon_sym_uy] = ACTIONS(7519),
    [anon_sym_s] = ACTIONS(7521),
    [anon_sym_us] = ACTIONS(7523),
    [anon_sym_l] = ACTIONS(7525),
    [aux_sym_uint32_token1] = ACTIONS(7527),
    [anon_sym_n] = ACTIONS(7529),
    [anon_sym_un] = ACTIONS(7531),
    [anon_sym_L] = ACTIONS(7533),
    [aux_sym_uint64_token1] = ACTIONS(7535),
    [aux_sym_bignum_token1] = ACTIONS(7537),
    [aux_sym_decimal_token1] = ACTIONS(7539),
    [anon_sym_DOT2] = ACTIONS(7541),
    [aux_sym_float_token1] = ACTIONS(7543),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2916)] = {
    [sym_block_comment] = STATE(2916),
    [sym_compiler_directive_decl] = STATE(2916),
    [sym_fsi_directive_decl] = STATE(2916),
    [sym_preproc_line] = STATE(2916),
    [aux_sym_rules_repeat1] = STATE(2941),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_and] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(7508),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [anon_sym_POUNDelse] = ACTIONS(6616),
    [sym__newline] = ACTIONS(7545),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2917)] = {
    [sym_block_comment] = STATE(2917),
    [sym_compiler_directive_decl] = STATE(2917),
    [sym_fsi_directive_decl] = STATE(2917),
    [sym_preproc_line] = STATE(2917),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_inherit] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2918)] = {
    [sym_block_comment] = STATE(2918),
    [sym_compiler_directive_decl] = STATE(2918),
    [sym_fsi_directive_decl] = STATE(2918),
    [sym_preproc_line] = STATE(2918),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(7548),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2919)] = {
    [sym_block_comment] = STATE(2919),
    [sym_compiler_directive_decl] = STATE(2919),
    [sym_fsi_directive_decl] = STATE(2919),
    [sym_preproc_line] = STATE(2919),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_inherit] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2920)] = {
    [sym_block_comment] = STATE(2920),
    [sym_compiler_directive_decl] = STATE(2920),
    [sym_fsi_directive_decl] = STATE(2920),
    [sym_preproc_line] = STATE(2920),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [anon_sym_inherit] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(7550),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [anon_sym_POUNDelse] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(2921)] = {
    [sym_block_comment] = STATE(2921),
    [sym_compiler_directive_decl] = STATE(2921),
    [sym_fsi_directive_decl] = STATE(2921),
    [sym_preproc_line] = STATE(2921),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2922)] = {
    [sym_block_comment] = STATE(2922),
    [sym_compiler_directive_decl] = STATE(2922),
    [sym_fsi_directive_decl] = STATE(2922),
    [sym_preproc_line] = STATE(2922),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_GT] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7552),
    [anon_sym_uy] = ACTIONS(7554),
    [anon_sym_s] = ACTIONS(7556),
    [anon_sym_us] = ACTIONS(7558),
    [anon_sym_l] = ACTIONS(7560),
    [aux_sym_uint32_token1] = ACTIONS(7562),
    [anon_sym_n] = ACTIONS(7564),
    [anon_sym_un] = ACTIONS(7566),
    [anon_sym_L] = ACTIONS(7568),
    [aux_sym_uint64_token1] = ACTIONS(7570),
    [aux_sym_bignum_token1] = ACTIONS(7572),
    [aux_sym_decimal_token1] = ACTIONS(7574),
    [anon_sym_DOT2] = ACTIONS(7576),
    [aux_sym_float_token1] = ACTIONS(7578),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2923)] = {
    [sym_type_arguments] = STATE(3581),
    [sym_long_identifier] = STATE(3582),
    [sym_block_comment] = STATE(2923),
    [sym_compiler_directive_decl] = STATE(2923),
    [sym_fsi_directive_decl] = STATE(2923),
    [sym_preproc_line] = STATE(2923),
    [aux_sym_compound_type_repeat1] = STATE(3306),
    [sym_identifier] = ACTIONS(7580),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(7582),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(7584),
    [anon_sym_LT] = ACTIONS(7586),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7588),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2924)] = {
    [sym_block_comment] = STATE(2924),
    [sym_compiler_directive_decl] = STATE(2924),
    [sym_fsi_directive_decl] = STATE(2924),
    [sym_preproc_line] = STATE(2924),
    [aux_sym_rules_repeat1] = STATE(2928),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(7590),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [anon_sym_inherit] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [anon_sym_POUNDelse] = ACTIONS(6616),
    [sym__newline] = ACTIONS(7592),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2925)] = {
    [sym_type_arguments] = STATE(3351),
    [sym_long_identifier] = STATE(3354),
    [sym_block_comment] = STATE(2925),
    [sym_compiler_directive_decl] = STATE(2925),
    [sym_fsi_directive_decl] = STATE(2925),
    [sym_preproc_line] = STATE(2925),
    [aux_sym_compound_type_repeat1] = STATE(3115),
    [ts_builtin_sym_end] = ACTIONS(6263),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_namespace] = ACTIONS(6265),
    [anon_sym_module] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_open] = ACTIONS(6265),
    [anon_sym_type] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(7595),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(7597),
    [anon_sym_LT] = ACTIONS(7599),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7601),
    [anon_sym_exception] = ACTIONS(6265),
    [anon_sym_extern] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6263),
    [anon_sym_POUNDload] = ACTIONS(6263),
    [anon_sym_POUNDtime] = ACTIONS(6263),
    [anon_sym_POUNDI] = ACTIONS(6263),
    [anon_sym_POUNDhelp] = ACTIONS(6263),
    [anon_sym_POUNDquit] = ACTIONS(6263),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2926)] = {
    [sym_block_comment] = STATE(2926),
    [sym_compiler_directive_decl] = STATE(2926),
    [sym_fsi_directive_decl] = STATE(2926),
    [sym_preproc_line] = STATE(2926),
    [aux_sym_rules_repeat1] = STATE(2936),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(7590),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [anon_sym_inherit] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [anon_sym_POUNDelse] = ACTIONS(6588),
    [sym__newline] = ACTIONS(7603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2927)] = {
    [sym_block_comment] = STATE(2927),
    [sym_compiler_directive_decl] = STATE(2927),
    [sym_fsi_directive_decl] = STATE(2927),
    [sym_preproc_line] = STATE(2927),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_and] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [anon_sym_POUNDelse] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(2928)] = {
    [sym_block_comment] = STATE(2928),
    [sym_compiler_directive_decl] = STATE(2928),
    [sym_fsi_directive_decl] = STATE(2928),
    [sym_preproc_line] = STATE(2928),
    [aux_sym_rules_repeat1] = STATE(2928),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(7606),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_inherit] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(7609),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2929)] = {
    [sym_block_comment] = STATE(2929),
    [sym_compiler_directive_decl] = STATE(2929),
    [sym_fsi_directive_decl] = STATE(2929),
    [sym_preproc_line] = STATE(2929),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [anon_sym_inherit] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [anon_sym_POUNDelse] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2930)] = {
    [sym_type_arguments] = STATE(3581),
    [sym_long_identifier] = STATE(3582),
    [sym_block_comment] = STATE(2930),
    [sym_compiler_directive_decl] = STATE(2930),
    [sym_fsi_directive_decl] = STATE(2930),
    [sym_preproc_line] = STATE(2930),
    [aux_sym_compound_type_repeat1] = STATE(3306),
    [sym_identifier] = ACTIONS(7580),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(7582),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7584),
    [anon_sym_LT] = ACTIONS(7586),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7588),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2931)] = {
    [sym_block_comment] = STATE(2931),
    [sym_compiler_directive_decl] = STATE(2931),
    [sym_fsi_directive_decl] = STATE(2931),
    [sym_preproc_line] = STATE(2931),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7612),
    [anon_sym_uy] = ACTIONS(7614),
    [anon_sym_s] = ACTIONS(7616),
    [anon_sym_us] = ACTIONS(7618),
    [anon_sym_l] = ACTIONS(7620),
    [aux_sym_uint32_token1] = ACTIONS(7622),
    [anon_sym_n] = ACTIONS(7624),
    [anon_sym_un] = ACTIONS(7626),
    [anon_sym_L] = ACTIONS(7628),
    [aux_sym_uint64_token1] = ACTIONS(7630),
    [aux_sym_bignum_token1] = ACTIONS(7632),
    [aux_sym_decimal_token1] = ACTIONS(7634),
    [anon_sym_DOT2] = ACTIONS(7636),
    [aux_sym_float_token1] = ACTIONS(7638),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__multi_dollar_interp_end] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2932)] = {
    [sym_block_comment] = STATE(2932),
    [sym_compiler_directive_decl] = STATE(2932),
    [sym_fsi_directive_decl] = STATE(2932),
    [sym_preproc_line] = STATE(2932),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7640),
    [anon_sym_uy] = ACTIONS(7642),
    [anon_sym_s] = ACTIONS(7644),
    [anon_sym_us] = ACTIONS(7646),
    [anon_sym_l] = ACTIONS(7648),
    [aux_sym_uint32_token1] = ACTIONS(7650),
    [anon_sym_n] = ACTIONS(7652),
    [anon_sym_un] = ACTIONS(7654),
    [anon_sym_L] = ACTIONS(7656),
    [aux_sym_uint64_token1] = ACTIONS(7658),
    [aux_sym_bignum_token1] = ACTIONS(7660),
    [aux_sym_decimal_token1] = ACTIONS(7662),
    [anon_sym_DOT2] = ACTIONS(7664),
    [aux_sym_float_token1] = ACTIONS(7666),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__quoted_close] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2933)] = {
    [sym_block_comment] = STATE(2933),
    [sym_compiler_directive_decl] = STATE(2933),
    [sym_fsi_directive_decl] = STATE(2933),
    [sym_preproc_line] = STATE(2933),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6223),
    [aux_sym_decimal_token1] = ACTIONS(6073),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2934)] = {
    [sym_block_comment] = STATE(2934),
    [sym_compiler_directive_decl] = STATE(2934),
    [sym_fsi_directive_decl] = STATE(2934),
    [sym_preproc_line] = STATE(2934),
    [aux_sym_long_identifier_repeat1] = STATE(2945),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7668),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_inherit] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2935)] = {
    [sym_block_comment] = STATE(2935),
    [sym_compiler_directive_decl] = STATE(2935),
    [sym_fsi_directive_decl] = STATE(2935),
    [sym_preproc_line] = STATE(2935),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_inherit] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2936)] = {
    [sym_block_comment] = STATE(2936),
    [sym_compiler_directive_decl] = STATE(2936),
    [sym_fsi_directive_decl] = STATE(2936),
    [sym_preproc_line] = STATE(2936),
    [aux_sym_rules_repeat1] = STATE(2928),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7590),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_inherit] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7670),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2937)] = {
    [sym_block_comment] = STATE(2937),
    [sym_compiler_directive_decl] = STATE(2937),
    [sym_fsi_directive_decl] = STATE(2937),
    [sym_preproc_line] = STATE(2937),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7673),
    [anon_sym_uy] = ACTIONS(7675),
    [anon_sym_s] = ACTIONS(7677),
    [anon_sym_us] = ACTIONS(7679),
    [anon_sym_l] = ACTIONS(7681),
    [aux_sym_uint32_token1] = ACTIONS(7683),
    [anon_sym_n] = ACTIONS(7685),
    [anon_sym_un] = ACTIONS(7687),
    [anon_sym_L] = ACTIONS(7689),
    [aux_sym_uint64_token1] = ACTIONS(7691),
    [aux_sym_bignum_token1] = ACTIONS(7693),
    [aux_sym_decimal_token1] = ACTIONS(7695),
    [anon_sym_DOT2] = ACTIONS(7697),
    [aux_sym_float_token1] = ACTIONS(7699),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__untyped_quoted_close] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2938)] = {
    [sym_block_comment] = STATE(2938),
    [sym_compiler_directive_decl] = STATE(2938),
    [sym_fsi_directive_decl] = STATE(2938),
    [sym_preproc_line] = STATE(2938),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_and] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [anon_sym_POUNDelse] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2939)] = {
    [sym_block_comment] = STATE(2939),
    [sym_compiler_directive_decl] = STATE(2939),
    [sym_fsi_directive_decl] = STATE(2939),
    [sym_preproc_line] = STATE(2939),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7701),
    [anon_sym_uy] = ACTIONS(7703),
    [anon_sym_s] = ACTIONS(7705),
    [anon_sym_us] = ACTIONS(7707),
    [anon_sym_l] = ACTIONS(7709),
    [aux_sym_uint32_token1] = ACTIONS(7711),
    [anon_sym_n] = ACTIONS(7713),
    [anon_sym_un] = ACTIONS(7715),
    [anon_sym_L] = ACTIONS(7717),
    [aux_sym_uint64_token1] = ACTIONS(7719),
    [aux_sym_bignum_token1] = ACTIONS(7721),
    [aux_sym_decimal_token1] = ACTIONS(7723),
    [anon_sym_DOT2] = ACTIONS(7725),
    [aux_sym_float_token1] = ACTIONS(7727),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2940)] = {
    [sym_block_comment] = STATE(2940),
    [sym_compiler_directive_decl] = STATE(2940),
    [sym_fsi_directive_decl] = STATE(2940),
    [sym_preproc_line] = STATE(2940),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [anon_sym_inherit] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [anon_sym_POUNDelse] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2941)] = {
    [sym_block_comment] = STATE(2941),
    [sym_compiler_directive_decl] = STATE(2941),
    [sym_fsi_directive_decl] = STATE(2941),
    [sym_preproc_line] = STATE(2941),
    [aux_sym_rules_repeat1] = STATE(2941),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(7729),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(7732),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2942)] = {
    [sym_block_comment] = STATE(2942),
    [sym_compiler_directive_decl] = STATE(2942),
    [sym_fsi_directive_decl] = STATE(2942),
    [sym_preproc_line] = STATE(2942),
    [aux_sym_rules_repeat1] = STATE(2924),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7590),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_inherit] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7670),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2943)] = {
    [sym_block_comment] = STATE(2943),
    [sym_compiler_directive_decl] = STATE(2943),
    [sym_fsi_directive_decl] = STATE(2943),
    [sym_preproc_line] = STATE(2943),
    [aux_sym_long_identifier_repeat1] = STATE(2934),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(7735),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [anon_sym_inherit] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [anon_sym_POUNDelse] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(2944)] = {
    [sym_type_arguments] = STATE(3351),
    [sym_long_identifier] = STATE(3354),
    [sym_block_comment] = STATE(2944),
    [sym_compiler_directive_decl] = STATE(2944),
    [sym_fsi_directive_decl] = STATE(2944),
    [sym_preproc_line] = STATE(2944),
    [aux_sym_compound_type_repeat1] = STATE(3115),
    [ts_builtin_sym_end] = ACTIONS(6255),
    [sym_identifier] = ACTIONS(7739),
    [anon_sym_namespace] = ACTIONS(6257),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(7595),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(7597),
    [anon_sym_LT] = ACTIONS(7599),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7601),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2945)] = {
    [sym_block_comment] = STATE(2945),
    [sym_compiler_directive_decl] = STATE(2945),
    [sym_fsi_directive_decl] = STATE(2945),
    [sym_preproc_line] = STATE(2945),
    [aux_sym_long_identifier_repeat1] = STATE(2945),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7741),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2946)] = {
    [sym_block_comment] = STATE(2946),
    [sym_compiler_directive_decl] = STATE(2946),
    [sym_fsi_directive_decl] = STATE(2946),
    [sym_preproc_line] = STATE(2946),
    [aux_sym_long_identifier_repeat1] = STATE(2951),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7744),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(2947)] = {
    [sym_type_arguments] = STATE(3351),
    [sym_long_identifier] = STATE(3354),
    [sym_block_comment] = STATE(2947),
    [sym_compiler_directive_decl] = STATE(2947),
    [sym_fsi_directive_decl] = STATE(2947),
    [sym_preproc_line] = STATE(2947),
    [aux_sym_compound_type_repeat1] = STATE(3115),
    [ts_builtin_sym_end] = ACTIONS(6209),
    [sym_identifier] = ACTIONS(7739),
    [anon_sym_namespace] = ACTIONS(6213),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(7595),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(7597),
    [anon_sym_LT] = ACTIONS(7599),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7601),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(2948)] = {
    [sym_type_arguments] = STATE(3351),
    [sym_long_identifier] = STATE(3354),
    [sym_block_comment] = STATE(2948),
    [sym_compiler_directive_decl] = STATE(2948),
    [sym_fsi_directive_decl] = STATE(2948),
    [sym_preproc_line] = STATE(2948),
    [aux_sym_compound_type_repeat1] = STATE(3115),
    [ts_builtin_sym_end] = ACTIONS(6195),
    [sym_identifier] = ACTIONS(7739),
    [anon_sym_namespace] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(7595),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7597),
    [anon_sym_LT] = ACTIONS(7599),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7601),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2949)] = {
    [sym_block_comment] = STATE(2949),
    [sym_compiler_directive_decl] = STATE(2949),
    [sym_fsi_directive_decl] = STATE(2949),
    [sym_preproc_line] = STATE(2949),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7276),
    [anon_sym_uy] = ACTIONS(7278),
    [anon_sym_s] = ACTIONS(7280),
    [anon_sym_us] = ACTIONS(7282),
    [anon_sym_l] = ACTIONS(7284),
    [aux_sym_uint32_token1] = ACTIONS(7286),
    [anon_sym_n] = ACTIONS(7288),
    [anon_sym_un] = ACTIONS(7290),
    [anon_sym_L] = ACTIONS(7292),
    [aux_sym_uint64_token1] = ACTIONS(7294),
    [aux_sym_bignum_token1] = ACTIONS(7296),
    [aux_sym_decimal_token1] = ACTIONS(7298),
    [anon_sym_DOT2] = ACTIONS(7300),
    [aux_sym_float_token1] = ACTIONS(7302),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2950)] = {
    [sym_block_comment] = STATE(2950),
    [sym_compiler_directive_decl] = STATE(2950),
    [sym_fsi_directive_decl] = STATE(2950),
    [sym_preproc_line] = STATE(2950),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5951),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7276),
    [anon_sym_uy] = ACTIONS(7278),
    [anon_sym_s] = ACTIONS(7280),
    [anon_sym_us] = ACTIONS(7282),
    [anon_sym_l] = ACTIONS(7284),
    [aux_sym_uint32_token1] = ACTIONS(7286),
    [anon_sym_n] = ACTIONS(7288),
    [anon_sym_un] = ACTIONS(7290),
    [anon_sym_L] = ACTIONS(7292),
    [aux_sym_uint64_token1] = ACTIONS(7294),
    [anon_sym_lf] = ACTIONS(7746),
    [anon_sym_LF] = ACTIONS(7748),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2951)] = {
    [sym_block_comment] = STATE(2951),
    [sym_compiler_directive_decl] = STATE(2951),
    [sym_fsi_directive_decl] = STATE(2951),
    [sym_preproc_line] = STATE(2951),
    [aux_sym_long_identifier_repeat1] = STATE(2951),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(2952)] = {
    [sym_block_comment] = STATE(2952),
    [sym_compiler_directive_decl] = STATE(2952),
    [sym_fsi_directive_decl] = STATE(2952),
    [sym_preproc_line] = STATE(2952),
    [aux_sym_long_identifier_repeat1] = STATE(2946),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_and] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(7753),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [anon_sym_POUNDelse] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(2953)] = {
    [sym_block_comment] = STATE(2953),
    [sym_compiler_directive_decl] = STATE(2953),
    [sym_fsi_directive_decl] = STATE(2953),
    [sym_preproc_line] = STATE(2953),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_and] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(7757),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [anon_sym_POUNDelse] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(2954)] = {
    [sym_block_comment] = STATE(2954),
    [sym_compiler_directive_decl] = STATE(2954),
    [sym_fsi_directive_decl] = STATE(2954),
    [sym_preproc_line] = STATE(2954),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7759),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2955)] = {
    [sym_block_comment] = STATE(2955),
    [sym_compiler_directive_decl] = STATE(2955),
    [sym_fsi_directive_decl] = STATE(2955),
    [sym_preproc_line] = STATE(2955),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_DASH_GT] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7338),
    [anon_sym_uy] = ACTIONS(7340),
    [anon_sym_s] = ACTIONS(7342),
    [anon_sym_us] = ACTIONS(7344),
    [anon_sym_l] = ACTIONS(7346),
    [aux_sym_uint32_token1] = ACTIONS(7348),
    [anon_sym_n] = ACTIONS(7350),
    [anon_sym_un] = ACTIONS(7352),
    [anon_sym_L] = ACTIONS(7354),
    [aux_sym_uint64_token1] = ACTIONS(7356),
    [aux_sym_bignum_token1] = ACTIONS(7358),
    [aux_sym_decimal_token1] = ACTIONS(7360),
    [anon_sym_DOT2] = ACTIONS(7362),
    [aux_sym_float_token1] = ACTIONS(7364),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2956)] = {
    [sym_block_comment] = STATE(2956),
    [sym_compiler_directive_decl] = STATE(2956),
    [sym_fsi_directive_decl] = STATE(2956),
    [sym_preproc_line] = STATE(2956),
    [aux_sym_rules_repeat1] = STATE(2916),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7508),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7510),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2957)] = {
    [sym_block_comment] = STATE(2957),
    [sym_compiler_directive_decl] = STATE(2957),
    [sym_fsi_directive_decl] = STATE(2957),
    [sym_preproc_line] = STATE(2957),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(2958)] = {
    [sym_block_comment] = STATE(2958),
    [sym_compiler_directive_decl] = STATE(2958),
    [sym_fsi_directive_decl] = STATE(2958),
    [sym_preproc_line] = STATE(2958),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(6227),
    [anon_sym_uy] = ACTIONS(6229),
    [anon_sym_s] = ACTIONS(6231),
    [anon_sym_us] = ACTIONS(6233),
    [anon_sym_l] = ACTIONS(6235),
    [aux_sym_uint32_token1] = ACTIONS(6237),
    [anon_sym_n] = ACTIONS(6239),
    [anon_sym_un] = ACTIONS(6241),
    [anon_sym_L] = ACTIONS(6243),
    [aux_sym_uint64_token1] = ACTIONS(6245),
    [anon_sym_lf] = ACTIONS(6302),
    [anon_sym_LF] = ACTIONS(6304),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2959)] = {
    [sym_type_arguments] = STATE(3581),
    [sym_long_identifier] = STATE(3582),
    [sym_block_comment] = STATE(2959),
    [sym_compiler_directive_decl] = STATE(2959),
    [sym_fsi_directive_decl] = STATE(2959),
    [sym_preproc_line] = STATE(2959),
    [aux_sym_compound_type_repeat1] = STATE(3306),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_module] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_open] = ACTIONS(6265),
    [anon_sym_type] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(7582),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(7584),
    [anon_sym_LT] = ACTIONS(7586),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7588),
    [anon_sym_exception] = ACTIONS(6265),
    [anon_sym_extern] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6263),
    [anon_sym_POUNDload] = ACTIONS(6263),
    [anon_sym_POUNDtime] = ACTIONS(6263),
    [anon_sym_POUNDI] = ACTIONS(6263),
    [anon_sym_POUNDhelp] = ACTIONS(6263),
    [anon_sym_POUNDquit] = ACTIONS(6263),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [anon_sym_POUNDelse] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(2960)] = {
    [sym_block_comment] = STATE(2960),
    [sym_compiler_directive_decl] = STATE(2960),
    [sym_fsi_directive_decl] = STATE(2960),
    [sym_preproc_line] = STATE(2960),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7761),
    [anon_sym_uy] = ACTIONS(7763),
    [anon_sym_s] = ACTIONS(7765),
    [anon_sym_us] = ACTIONS(7767),
    [anon_sym_l] = ACTIONS(7769),
    [aux_sym_uint32_token1] = ACTIONS(7771),
    [anon_sym_n] = ACTIONS(7773),
    [anon_sym_un] = ACTIONS(7775),
    [anon_sym_L] = ACTIONS(7777),
    [aux_sym_uint64_token1] = ACTIONS(7779),
    [aux_sym_bignum_token1] = ACTIONS(7781),
    [aux_sym_decimal_token1] = ACTIONS(7783),
    [anon_sym_DOT2] = ACTIONS(7785),
    [aux_sym_float_token1] = ACTIONS(7787),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(2961)] = {
    [sym_block_comment] = STATE(2961),
    [sym_compiler_directive_decl] = STATE(2961),
    [sym_fsi_directive_decl] = STATE(2961),
    [sym_preproc_line] = STATE(2961),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_and] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [anon_sym_POUNDelse] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2962)] = {
    [sym_type_arguments] = STATE(3351),
    [sym_long_identifier] = STATE(3354),
    [sym_block_comment] = STATE(2962),
    [sym_compiler_directive_decl] = STATE(2962),
    [sym_fsi_directive_decl] = STATE(2962),
    [sym_preproc_line] = STATE(2962),
    [aux_sym_compound_type_repeat1] = STATE(3115),
    [ts_builtin_sym_end] = ACTIONS(6259),
    [sym_identifier] = ACTIONS(7739),
    [anon_sym_namespace] = ACTIONS(6261),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(7595),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(7597),
    [anon_sym_LT] = ACTIONS(7599),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7601),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2963)] = {
    [sym_block_comment] = STATE(2963),
    [sym_compiler_directive_decl] = STATE(2963),
    [sym_fsi_directive_decl] = STATE(2963),
    [sym_preproc_line] = STATE(2963),
    [aux_sym_rules_repeat1] = STATE(2909),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_and] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(7508),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [anon_sym_POUNDelse] = ACTIONS(6588),
    [sym__newline] = ACTIONS(7789),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2964)] = {
    [sym_type_arguments] = STATE(3581),
    [sym_long_identifier] = STATE(3582),
    [sym_block_comment] = STATE(2964),
    [sym_compiler_directive_decl] = STATE(2964),
    [sym_fsi_directive_decl] = STATE(2964),
    [sym_preproc_line] = STATE(2964),
    [aux_sym_compound_type_repeat1] = STATE(3306),
    [sym_identifier] = ACTIONS(7580),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(7582),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(7584),
    [anon_sym_LT] = ACTIONS(7586),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7588),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(2965)] = {
    [sym_type_arguments] = STATE(3581),
    [sym_long_identifier] = STATE(3582),
    [sym_block_comment] = STATE(2965),
    [sym_compiler_directive_decl] = STATE(2965),
    [sym_fsi_directive_decl] = STATE(2965),
    [sym_preproc_line] = STATE(2965),
    [aux_sym_compound_type_repeat1] = STATE(3306),
    [sym_identifier] = ACTIONS(7580),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(7582),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(7584),
    [anon_sym_LT] = ACTIONS(7586),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7588),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2966)] = {
    [sym_block_comment] = STATE(2966),
    [sym_compiler_directive_decl] = STATE(2966),
    [sym_fsi_directive_decl] = STATE(2966),
    [sym_preproc_line] = STATE(2966),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2967)] = {
    [sym_type_arguments] = STATE(3697),
    [sym_long_identifier] = STATE(3698),
    [sym_block_comment] = STATE(2967),
    [sym_compiler_directive_decl] = STATE(2967),
    [sym_fsi_directive_decl] = STATE(2967),
    [sym_preproc_line] = STATE(2967),
    [aux_sym_compound_type_repeat1] = STATE(3661),
    [sym_identifier] = ACTIONS(7792),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(7794),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7796),
    [anon_sym_LT] = ACTIONS(7798),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7800),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(2968)] = {
    [sym_type_arguments] = STATE(3697),
    [sym_long_identifier] = STATE(3698),
    [sym_block_comment] = STATE(2968),
    [sym_compiler_directive_decl] = STATE(2968),
    [sym_fsi_directive_decl] = STATE(2968),
    [sym_preproc_line] = STATE(2968),
    [aux_sym_compound_type_repeat1] = STATE(3661),
    [sym_identifier] = ACTIONS(7792),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(7794),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(7796),
    [anon_sym_LT] = ACTIONS(7798),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7800),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(2969)] = {
    [sym_block_comment] = STATE(2969),
    [sym_compiler_directive_decl] = STATE(2969),
    [sym_fsi_directive_decl] = STATE(2969),
    [sym_preproc_line] = STATE(2969),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_and] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__dedent] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2970)] = {
    [sym_block_comment] = STATE(2970),
    [sym_compiler_directive_decl] = STATE(2970),
    [sym_fsi_directive_decl] = STATE(2970),
    [sym_preproc_line] = STATE(2970),
    [aux_sym_long_identifier_repeat1] = STATE(3031),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_and] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(7802),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(2971)] = {
    [sym_block_comment] = STATE(2971),
    [sym_compiler_directive_decl] = STATE(2971),
    [sym_fsi_directive_decl] = STATE(2971),
    [sym_preproc_line] = STATE(2971),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_and] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [anon_sym_POUNDelse] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(2972)] = {
    [sym_block_comment] = STATE(2972),
    [sym_compiler_directive_decl] = STATE(2972),
    [sym_fsi_directive_decl] = STATE(2972),
    [sym_preproc_line] = STATE(2972),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_and] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [anon_sym_POUNDelse] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(2973)] = {
    [sym_block_comment] = STATE(2973),
    [sym_compiler_directive_decl] = STATE(2973),
    [sym_fsi_directive_decl] = STATE(2973),
    [sym_preproc_line] = STATE(2973),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2974)] = {
    [sym_block_comment] = STATE(2974),
    [sym_compiler_directive_decl] = STATE(2974),
    [sym_fsi_directive_decl] = STATE(2974),
    [sym_preproc_line] = STATE(2974),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7806),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(2975)] = {
    [sym_block_comment] = STATE(2975),
    [sym_compiler_directive_decl] = STATE(2975),
    [sym_fsi_directive_decl] = STATE(2975),
    [sym_preproc_line] = STATE(2975),
    [aux_sym_rules_repeat1] = STATE(3055),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_and] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(7808),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(7810),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2976)] = {
    [sym_block_comment] = STATE(2976),
    [sym_compiler_directive_decl] = STATE(2976),
    [sym_fsi_directive_decl] = STATE(2976),
    [sym_preproc_line] = STATE(2976),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_inherit] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2977)] = {
    [sym_block_comment] = STATE(2977),
    [sym_compiler_directive_decl] = STATE(2977),
    [sym_fsi_directive_decl] = STATE(2977),
    [sym_preproc_line] = STATE(2977),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_DOT_DOT] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__dedent] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(2978)] = {
    [sym_block_comment] = STATE(2978),
    [sym_compiler_directive_decl] = STATE(2978),
    [sym_fsi_directive_decl] = STATE(2978),
    [sym_preproc_line] = STATE(2978),
    [aux_sym_rules_repeat1] = STATE(2979),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_and] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(7808),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(7813),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2979)] = {
    [sym_block_comment] = STATE(2979),
    [sym_compiler_directive_decl] = STATE(2979),
    [sym_fsi_directive_decl] = STATE(2979),
    [sym_preproc_line] = STATE(2979),
    [aux_sym_rules_repeat1] = STATE(2979),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(7816),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(7819),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2980)] = {
    [sym_block_comment] = STATE(2980),
    [sym_compiler_directive_decl] = STATE(2980),
    [sym_fsi_directive_decl] = STATE(2980),
    [sym_preproc_line] = STATE(2980),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_and] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [anon_sym_POUNDelse] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(2981)] = {
    [sym_block_comment] = STATE(2981),
    [sym_compiler_directive_decl] = STATE(2981),
    [sym_fsi_directive_decl] = STATE(2981),
    [sym_preproc_line] = STATE(2981),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [anon_sym_inherit] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [anon_sym_POUNDelse] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(2982)] = {
    [sym_block_comment] = STATE(2982),
    [sym_compiler_directive_decl] = STATE(2982),
    [sym_fsi_directive_decl] = STATE(2982),
    [sym_preproc_line] = STATE(2982),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_and] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2983)] = {
    [sym_block_comment] = STATE(2983),
    [sym_compiler_directive_decl] = STATE(2983),
    [sym_fsi_directive_decl] = STATE(2983),
    [sym_preproc_line] = STATE(2983),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [anon_sym_inherit] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [anon_sym_POUNDelse] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(2984)] = {
    [sym_block_comment] = STATE(2984),
    [sym_compiler_directive_decl] = STATE(2984),
    [sym_fsi_directive_decl] = STATE(2984),
    [sym_preproc_line] = STATE(2984),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_and] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [anon_sym_POUNDelse] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(2985)] = {
    [sym_block_comment] = STATE(2985),
    [sym_compiler_directive_decl] = STATE(2985),
    [sym_fsi_directive_decl] = STATE(2985),
    [sym_preproc_line] = STATE(2985),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_and] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__dedent] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2986)] = {
    [sym_block_comment] = STATE(2986),
    [sym_compiler_directive_decl] = STATE(2986),
    [sym_fsi_directive_decl] = STATE(2986),
    [sym_preproc_line] = STATE(2986),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [anon_sym_inherit] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [anon_sym_POUNDelse] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(2987)] = {
    [sym_block_comment] = STATE(2987),
    [sym_compiler_directive_decl] = STATE(2987),
    [sym_fsi_directive_decl] = STATE(2987),
    [sym_preproc_line] = STATE(2987),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_DOT_DOT] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__dedent] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(2988)] = {
    [sym_block_comment] = STATE(2988),
    [sym_compiler_directive_decl] = STATE(2988),
    [sym_fsi_directive_decl] = STATE(2988),
    [sym_preproc_line] = STATE(2988),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2989)] = {
    [sym_block_comment] = STATE(2989),
    [sym_compiler_directive_decl] = STATE(2989),
    [sym_fsi_directive_decl] = STATE(2989),
    [sym_preproc_line] = STATE(2989),
    [aux_sym_rules_repeat1] = STATE(2997),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_and] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(7822),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(7824),
    [sym__dedent] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2990)] = {
    [sym_block_comment] = STATE(2990),
    [sym_compiler_directive_decl] = STATE(2990),
    [sym_fsi_directive_decl] = STATE(2990),
    [sym_preproc_line] = STATE(2990),
    [aux_sym_rules_repeat1] = STATE(2994),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(7827),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_DOT_DOT] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(7829),
    [sym__dedent] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(2991)] = {
    [sym_block_comment] = STATE(2991),
    [sym_compiler_directive_decl] = STATE(2991),
    [sym_fsi_directive_decl] = STATE(2991),
    [sym_preproc_line] = STATE(2991),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_inherit] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(2992)] = {
    [sym_block_comment] = STATE(2992),
    [sym_compiler_directive_decl] = STATE(2992),
    [sym_fsi_directive_decl] = STATE(2992),
    [sym_preproc_line] = STATE(2992),
    [aux_sym_sequential_expression_repeat1] = STATE(2992),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [anon_sym_POUNDelse] = ACTIONS(6798),
    [sym__newline] = ACTIONS(7832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(2993)] = {
    [sym_block_comment] = STATE(2993),
    [sym_compiler_directive_decl] = STATE(2993),
    [sym_fsi_directive_decl] = STATE(2993),
    [sym_preproc_line] = STATE(2993),
    [aux_sym_rules_repeat1] = STATE(2996),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7827),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7835),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2994)] = {
    [sym_block_comment] = STATE(2994),
    [sym_compiler_directive_decl] = STATE(2994),
    [sym_fsi_directive_decl] = STATE(2994),
    [sym_preproc_line] = STATE(2994),
    [aux_sym_rules_repeat1] = STATE(2998),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7827),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7835),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2995)] = {
    [sym_block_comment] = STATE(2995),
    [sym_compiler_directive_decl] = STATE(2995),
    [sym_fsi_directive_decl] = STATE(2995),
    [sym_preproc_line] = STATE(2995),
    [aux_sym_rules_repeat1] = STATE(3001),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7822),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7838),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2996)] = {
    [sym_block_comment] = STATE(2996),
    [sym_compiler_directive_decl] = STATE(2996),
    [sym_fsi_directive_decl] = STATE(2996),
    [sym_preproc_line] = STATE(2996),
    [aux_sym_rules_repeat1] = STATE(2998),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(7827),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_DOT_DOT] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(7841),
    [sym__dedent] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(2997)] = {
    [sym_block_comment] = STATE(2997),
    [sym_compiler_directive_decl] = STATE(2997),
    [sym_fsi_directive_decl] = STATE(2997),
    [sym_preproc_line] = STATE(2997),
    [aux_sym_rules_repeat1] = STATE(3002),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7822),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7838),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(2998)] = {
    [sym_block_comment] = STATE(2998),
    [sym_compiler_directive_decl] = STATE(2998),
    [sym_fsi_directive_decl] = STATE(2998),
    [sym_preproc_line] = STATE(2998),
    [aux_sym_rules_repeat1] = STATE(2998),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(7844),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(7847),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(2999)] = {
    [sym_block_comment] = STATE(2999),
    [sym_compiler_directive_decl] = STATE(2999),
    [sym_fsi_directive_decl] = STATE(2999),
    [sym_preproc_line] = STATE(2999),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [anon_sym_inherit] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(3000)] = {
    [sym_block_comment] = STATE(3000),
    [sym_compiler_directive_decl] = STATE(3000),
    [sym_fsi_directive_decl] = STATE(3000),
    [sym_preproc_line] = STATE(3000),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [anon_sym_inherit] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [anon_sym_POUNDelse] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3001)] = {
    [sym_block_comment] = STATE(3001),
    [sym_compiler_directive_decl] = STATE(3001),
    [sym_fsi_directive_decl] = STATE(3001),
    [sym_preproc_line] = STATE(3001),
    [aux_sym_rules_repeat1] = STATE(3002),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_and] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(7822),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(7850),
    [sym__dedent] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(3002)] = {
    [sym_block_comment] = STATE(3002),
    [sym_compiler_directive_decl] = STATE(3002),
    [sym_fsi_directive_decl] = STATE(3002),
    [sym_preproc_line] = STATE(3002),
    [aux_sym_rules_repeat1] = STATE(3002),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(7853),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(7856),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3003)] = {
    [sym_block_comment] = STATE(3003),
    [sym_compiler_directive_decl] = STATE(3003),
    [sym_fsi_directive_decl] = STATE(3003),
    [sym_preproc_line] = STATE(3003),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [anon_sym_inherit] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [anon_sym_POUNDelse] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3004)] = {
    [sym_block_comment] = STATE(3004),
    [sym_compiler_directive_decl] = STATE(3004),
    [sym_fsi_directive_decl] = STATE(3004),
    [sym_preproc_line] = STATE(3004),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_and] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [anon_sym_POUNDelse] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3005)] = {
    [sym_block_comment] = STATE(3005),
    [sym_compiler_directive_decl] = STATE(3005),
    [sym_fsi_directive_decl] = STATE(3005),
    [sym_preproc_line] = STATE(3005),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3006)] = {
    [sym_block_comment] = STATE(3006),
    [sym_compiler_directive_decl] = STATE(3006),
    [sym_fsi_directive_decl] = STATE(3006),
    [sym_preproc_line] = STATE(3006),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_and] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [anon_sym_POUNDelse] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3007)] = {
    [sym_block_comment] = STATE(3007),
    [sym_compiler_directive_decl] = STATE(3007),
    [sym_fsi_directive_decl] = STATE(3007),
    [sym_preproc_line] = STATE(3007),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [aux_sym_access_modifier_token1] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_default] = ACTIONS(6554),
    [anon_sym_static] = ACTIONS(6554),
    [anon_sym_member] = ACTIONS(6554),
    [anon_sym_interface] = ACTIONS(6554),
    [anon_sym_abstract] = ACTIONS(6554),
    [anon_sym_val] = ACTIONS(6554),
    [anon_sym_override] = ACTIONS(6554),
    [anon_sym_inherit] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(3008)] = {
    [sym_block_comment] = STATE(3008),
    [sym_compiler_directive_decl] = STATE(3008),
    [sym_fsi_directive_decl] = STATE(3008),
    [sym_preproc_line] = STATE(3008),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_and] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [anon_sym_POUNDelse] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3009)] = {
    [sym_block_comment] = STATE(3009),
    [sym_compiler_directive_decl] = STATE(3009),
    [sym_fsi_directive_decl] = STATE(3009),
    [sym_preproc_line] = STATE(3009),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7379),
    [anon_sym_uy] = ACTIONS(7381),
    [anon_sym_s] = ACTIONS(7383),
    [anon_sym_us] = ACTIONS(7385),
    [anon_sym_l] = ACTIONS(7387),
    [aux_sym_uint32_token1] = ACTIONS(7389),
    [anon_sym_n] = ACTIONS(7391),
    [anon_sym_un] = ACTIONS(7393),
    [anon_sym_L] = ACTIONS(7395),
    [aux_sym_uint64_token1] = ACTIONS(7397),
    [anon_sym_lf] = ACTIONS(7859),
    [anon_sym_LF] = ACTIONS(7861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3010)] = {
    [sym_block_comment] = STATE(3010),
    [sym_compiler_directive_decl] = STATE(3010),
    [sym_fsi_directive_decl] = STATE(3010),
    [sym_preproc_line] = STATE(3010),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_and] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [anon_sym_POUNDelse] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3011)] = {
    [sym_block_comment] = STATE(3011),
    [sym_compiler_directive_decl] = STATE(3011),
    [sym_fsi_directive_decl] = STATE(3011),
    [sym_preproc_line] = STATE(3011),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_and] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [anon_sym_POUNDelse] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3012)] = {
    [sym_block_comment] = STATE(3012),
    [sym_compiler_directive_decl] = STATE(3012),
    [sym_fsi_directive_decl] = STATE(3012),
    [sym_preproc_line] = STATE(3012),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6267),
    [aux_sym_decimal_token1] = ACTIONS(6157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3013)] = {
    [sym_block_comment] = STATE(3013),
    [sym_compiler_directive_decl] = STATE(3013),
    [sym_fsi_directive_decl] = STATE(3013),
    [sym_preproc_line] = STATE(3013),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3014)] = {
    [sym_block_comment] = STATE(3014),
    [sym_compiler_directive_decl] = STATE(3014),
    [sym_fsi_directive_decl] = STATE(3014),
    [sym_preproc_line] = STATE(3014),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_and] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(7863),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3015)] = {
    [sym_block_comment] = STATE(3015),
    [sym_compiler_directive_decl] = STATE(3015),
    [sym_fsi_directive_decl] = STATE(3015),
    [sym_preproc_line] = STATE(3015),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_inherit] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3016)] = {
    [sym_block_comment] = STATE(3016),
    [sym_compiler_directive_decl] = STATE(3016),
    [sym_fsi_directive_decl] = STATE(3016),
    [sym_preproc_line] = STATE(3016),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3017)] = {
    [sym_block_comment] = STATE(3017),
    [sym_compiler_directive_decl] = STATE(3017),
    [sym_fsi_directive_decl] = STATE(3017),
    [sym_preproc_line] = STATE(3017),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_and] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3018)] = {
    [sym_type_arguments] = STATE(3697),
    [sym_long_identifier] = STATE(3698),
    [sym_block_comment] = STATE(3018),
    [sym_compiler_directive_decl] = STATE(3018),
    [sym_fsi_directive_decl] = STATE(3018),
    [sym_preproc_line] = STATE(3018),
    [aux_sym_compound_type_repeat1] = STATE(3661),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_module] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_open] = ACTIONS(6265),
    [anon_sym_type] = ACTIONS(6265),
    [anon_sym_LBRACK_LT] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(7794),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(7796),
    [anon_sym_LT] = ACTIONS(7798),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7800),
    [anon_sym_exception] = ACTIONS(6265),
    [anon_sym_extern] = ACTIONS(6265),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6263),
    [anon_sym_POUNDload] = ACTIONS(6263),
    [anon_sym_POUNDtime] = ACTIONS(6263),
    [anon_sym_POUNDI] = ACTIONS(6263),
    [anon_sym_POUNDhelp] = ACTIONS(6263),
    [anon_sym_POUNDquit] = ACTIONS(6263),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3019)] = {
    [sym_block_comment] = STATE(3019),
    [sym_compiler_directive_decl] = STATE(3019),
    [sym_fsi_directive_decl] = STATE(3019),
    [sym_preproc_line] = STATE(3019),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3020)] = {
    [sym_block_comment] = STATE(3020),
    [sym_compiler_directive_decl] = STATE(3020),
    [sym_fsi_directive_decl] = STATE(3020),
    [sym_preproc_line] = STATE(3020),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6271),
    [aux_sym_decimal_token1] = ACTIONS(6129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3021)] = {
    [sym_block_comment] = STATE(3021),
    [sym_compiler_directive_decl] = STATE(3021),
    [sym_fsi_directive_decl] = STATE(3021),
    [sym_preproc_line] = STATE(3021),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [anon_sym_inherit] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(7865),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3022)] = {
    [sym_block_comment] = STATE(3022),
    [sym_compiler_directive_decl] = STATE(3022),
    [sym_fsi_directive_decl] = STATE(3022),
    [sym_preproc_line] = STATE(3022),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [anon_sym_inherit] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3023)] = {
    [sym_block_comment] = STATE(3023),
    [sym_compiler_directive_decl] = STATE(3023),
    [sym_fsi_directive_decl] = STATE(3023),
    [sym_preproc_line] = STATE(3023),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [anon_sym_inherit] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3024)] = {
    [sym_block_comment] = STATE(3024),
    [sym_compiler_directive_decl] = STATE(3024),
    [sym_fsi_directive_decl] = STATE(3024),
    [sym_preproc_line] = STATE(3024),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3025)] = {
    [sym_block_comment] = STATE(3025),
    [sym_compiler_directive_decl] = STATE(3025),
    [sym_fsi_directive_decl] = STATE(3025),
    [sym_preproc_line] = STATE(3025),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3026)] = {
    [sym_block_comment] = STATE(3026),
    [sym_compiler_directive_decl] = STATE(3026),
    [sym_fsi_directive_decl] = STATE(3026),
    [sym_preproc_line] = STATE(3026),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(7867),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3027)] = {
    [sym_block_comment] = STATE(3027),
    [sym_compiler_directive_decl] = STATE(3027),
    [sym_fsi_directive_decl] = STATE(3027),
    [sym_preproc_line] = STATE(3027),
    [aux_sym_long_identifier_repeat1] = STATE(3027),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7869),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3028)] = {
    [sym_block_comment] = STATE(3028),
    [sym_compiler_directive_decl] = STATE(3028),
    [sym_fsi_directive_decl] = STATE(3028),
    [sym_preproc_line] = STATE(3028),
    [aux_sym_long_identifier_repeat1] = STATE(3029),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7872),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_inherit] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3029)] = {
    [sym_block_comment] = STATE(3029),
    [sym_compiler_directive_decl] = STATE(3029),
    [sym_fsi_directive_decl] = STATE(3029),
    [sym_preproc_line] = STATE(3029),
    [aux_sym_long_identifier_repeat1] = STATE(3029),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7874),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3030)] = {
    [sym_block_comment] = STATE(3030),
    [sym_compiler_directive_decl] = STATE(3030),
    [sym_fsi_directive_decl] = STATE(3030),
    [sym_preproc_line] = STATE(3030),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_inherit] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3031)] = {
    [sym_block_comment] = STATE(3031),
    [sym_compiler_directive_decl] = STATE(3031),
    [sym_fsi_directive_decl] = STATE(3031),
    [sym_preproc_line] = STATE(3031),
    [aux_sym_long_identifier_repeat1] = STATE(3032),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7877),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3032)] = {
    [sym_block_comment] = STATE(3032),
    [sym_compiler_directive_decl] = STATE(3032),
    [sym_fsi_directive_decl] = STATE(3032),
    [sym_preproc_line] = STATE(3032),
    [aux_sym_long_identifier_repeat1] = STATE(3032),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7879),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3033)] = {
    [sym_block_comment] = STATE(3033),
    [sym_compiler_directive_decl] = STATE(3033),
    [sym_fsi_directive_decl] = STATE(3033),
    [sym_preproc_line] = STATE(3033),
    [aux_sym_sequential_expression_repeat1] = STATE(3079),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [anon_sym_inherit] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [anon_sym_POUNDelse] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3034)] = {
    [sym_block_comment] = STATE(3034),
    [sym_compiler_directive_decl] = STATE(3034),
    [sym_fsi_directive_decl] = STATE(3034),
    [sym_preproc_line] = STATE(3034),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [anon_sym_inherit] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [anon_sym_POUNDelse] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3035)] = {
    [sym_block_comment] = STATE(3035),
    [sym_compiler_directive_decl] = STATE(3035),
    [sym_fsi_directive_decl] = STATE(3035),
    [sym_preproc_line] = STATE(3035),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(7882),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [anon_sym_inherit] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [anon_sym_POUNDelse] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3036)] = {
    [sym_block_comment] = STATE(3036),
    [sym_compiler_directive_decl] = STATE(3036),
    [sym_fsi_directive_decl] = STATE(3036),
    [sym_preproc_line] = STATE(3036),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [anon_sym_inherit] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [anon_sym_POUNDelse] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3037)] = {
    [sym_block_comment] = STATE(3037),
    [sym_compiler_directive_decl] = STATE(3037),
    [sym_fsi_directive_decl] = STATE(3037),
    [sym_preproc_line] = STATE(3037),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [anon_sym_inherit] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [anon_sym_POUNDelse] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3038)] = {
    [sym_block_comment] = STATE(3038),
    [sym_compiler_directive_decl] = STATE(3038),
    [sym_fsi_directive_decl] = STATE(3038),
    [sym_preproc_line] = STATE(3038),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_and] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(7884),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [anon_sym_POUNDelse] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3039)] = {
    [sym_block_comment] = STATE(3039),
    [sym_compiler_directive_decl] = STATE(3039),
    [sym_fsi_directive_decl] = STATE(3039),
    [sym_preproc_line] = STATE(3039),
    [aux_sym_rules_repeat1] = STATE(3063),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [aux_sym_access_modifier_token1] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(7886),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_default] = ACTIONS(6590),
    [anon_sym_static] = ACTIONS(6590),
    [anon_sym_member] = ACTIONS(6590),
    [anon_sym_interface] = ACTIONS(6590),
    [anon_sym_abstract] = ACTIONS(6590),
    [anon_sym_val] = ACTIONS(6590),
    [anon_sym_override] = ACTIONS(6590),
    [anon_sym_inherit] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(7888),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(3040)] = {
    [sym_block_comment] = STATE(3040),
    [sym_compiler_directive_decl] = STATE(3040),
    [sym_fsi_directive_decl] = STATE(3040),
    [sym_preproc_line] = STATE(3040),
    [aux_sym_long_identifier_repeat1] = STATE(3084),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_and] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(7891),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3041)] = {
    [sym_block_comment] = STATE(3041),
    [sym_compiler_directive_decl] = STATE(3041),
    [sym_fsi_directive_decl] = STATE(3041),
    [sym_preproc_line] = STATE(3041),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6275),
    [aux_sym_decimal_token1] = ACTIONS(6185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3042)] = {
    [sym_block_comment] = STATE(3042),
    [sym_compiler_directive_decl] = STATE(3042),
    [sym_fsi_directive_decl] = STATE(3042),
    [sym_preproc_line] = STATE(3042),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [anon_sym_inherit] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [anon_sym_POUNDelse] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3043)] = {
    [sym_block_comment] = STATE(3043),
    [sym_compiler_directive_decl] = STATE(3043),
    [sym_fsi_directive_decl] = STATE(3043),
    [sym_preproc_line] = STATE(3043),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7472),
    [anon_sym_uy] = ACTIONS(7474),
    [anon_sym_s] = ACTIONS(7476),
    [anon_sym_us] = ACTIONS(7478),
    [anon_sym_l] = ACTIONS(7480),
    [aux_sym_uint32_token1] = ACTIONS(7482),
    [anon_sym_n] = ACTIONS(7484),
    [anon_sym_un] = ACTIONS(7486),
    [anon_sym_L] = ACTIONS(7488),
    [aux_sym_uint64_token1] = ACTIONS(7490),
    [anon_sym_lf] = ACTIONS(7895),
    [anon_sym_LF] = ACTIONS(7897),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3044)] = {
    [sym_block_comment] = STATE(3044),
    [sym_compiler_directive_decl] = STATE(3044),
    [sym_fsi_directive_decl] = STATE(3044),
    [sym_preproc_line] = STATE(3044),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_and] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [anon_sym_POUNDelse] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3045)] = {
    [sym_block_comment] = STATE(3045),
    [sym_compiler_directive_decl] = STATE(3045),
    [sym_fsi_directive_decl] = STATE(3045),
    [sym_preproc_line] = STATE(3045),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_and] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [anon_sym_POUNDelse] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3046)] = {
    [sym_block_comment] = STATE(3046),
    [sym_compiler_directive_decl] = STATE(3046),
    [sym_fsi_directive_decl] = STATE(3046),
    [sym_preproc_line] = STATE(3046),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7899),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3047)] = {
    [sym_block_comment] = STATE(3047),
    [sym_compiler_directive_decl] = STATE(3047),
    [sym_fsi_directive_decl] = STATE(3047),
    [sym_preproc_line] = STATE(3047),
    [aux_sym_long_identifier_repeat1] = STATE(3028),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(7901),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [anon_sym_inherit] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3048)] = {
    [sym_block_comment] = STATE(3048),
    [sym_compiler_directive_decl] = STATE(3048),
    [sym_fsi_directive_decl] = STATE(3048),
    [sym_preproc_line] = STATE(3048),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(7905),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3049)] = {
    [sym_block_comment] = STATE(3049),
    [sym_compiler_directive_decl] = STATE(3049),
    [sym_fsi_directive_decl] = STATE(3049),
    [sym_preproc_line] = STATE(3049),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3050)] = {
    [sym_block_comment] = STATE(3050),
    [sym_compiler_directive_decl] = STATE(3050),
    [sym_fsi_directive_decl] = STATE(3050),
    [sym_preproc_line] = STATE(3050),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_and] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(7907),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3051)] = {
    [sym_block_comment] = STATE(3051),
    [sym_compiler_directive_decl] = STATE(3051),
    [sym_fsi_directive_decl] = STATE(3051),
    [sym_preproc_line] = STATE(3051),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7414),
    [anon_sym_uy] = ACTIONS(7416),
    [anon_sym_s] = ACTIONS(7418),
    [anon_sym_us] = ACTIONS(7420),
    [anon_sym_l] = ACTIONS(7422),
    [aux_sym_uint32_token1] = ACTIONS(7424),
    [anon_sym_n] = ACTIONS(7426),
    [anon_sym_un] = ACTIONS(7428),
    [anon_sym_L] = ACTIONS(7430),
    [aux_sym_uint64_token1] = ACTIONS(7432),
    [anon_sym_lf] = ACTIONS(7909),
    [anon_sym_LF] = ACTIONS(7911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
    [sym__in] = ACTIONS(5953),
  },
  [STATE(3052)] = {
    [sym_block_comment] = STATE(3052),
    [sym_compiler_directive_decl] = STATE(3052),
    [sym_fsi_directive_decl] = STATE(3052),
    [sym_preproc_line] = STATE(3052),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3053)] = {
    [sym_block_comment] = STATE(3053),
    [sym_compiler_directive_decl] = STATE(3053),
    [sym_fsi_directive_decl] = STATE(3053),
    [sym_preproc_line] = STATE(3053),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7913),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_inherit] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3054)] = {
    [sym_block_comment] = STATE(3054),
    [sym_compiler_directive_decl] = STATE(3054),
    [sym_fsi_directive_decl] = STATE(3054),
    [sym_preproc_line] = STATE(3054),
    [aux_sym_rules_repeat1] = STATE(2978),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7808),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3055)] = {
    [sym_block_comment] = STATE(3055),
    [sym_compiler_directive_decl] = STATE(3055),
    [sym_fsi_directive_decl] = STATE(3055),
    [sym_preproc_line] = STATE(3055),
    [aux_sym_rules_repeat1] = STATE(2979),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_and] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7808),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3056)] = {
    [sym_block_comment] = STATE(3056),
    [sym_compiler_directive_decl] = STATE(3056),
    [sym_fsi_directive_decl] = STATE(3056),
    [sym_preproc_line] = STATE(3056),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_and] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3057)] = {
    [sym_block_comment] = STATE(3057),
    [sym_compiler_directive_decl] = STATE(3057),
    [sym_fsi_directive_decl] = STATE(3057),
    [sym_preproc_line] = STATE(3057),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(7918),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_inherit] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3058)] = {
    [sym_block_comment] = STATE(3058),
    [sym_compiler_directive_decl] = STATE(3058),
    [sym_fsi_directive_decl] = STATE(3058),
    [sym_preproc_line] = STATE(3058),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_and] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3059)] = {
    [sym_block_comment] = STATE(3059),
    [sym_compiler_directive_decl] = STATE(3059),
    [sym_fsi_directive_decl] = STATE(3059),
    [sym_preproc_line] = STATE(3059),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [anon_sym_inherit] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [anon_sym_POUNDelse] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3060)] = {
    [sym_block_comment] = STATE(3060),
    [sym_compiler_directive_decl] = STATE(3060),
    [sym_fsi_directive_decl] = STATE(3060),
    [sym_preproc_line] = STATE(3060),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_inherit] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3061)] = {
    [sym_block_comment] = STATE(3061),
    [sym_compiler_directive_decl] = STATE(3061),
    [sym_fsi_directive_decl] = STATE(3061),
    [sym_preproc_line] = STATE(3061),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7506),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_inherit] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3062)] = {
    [sym_block_comment] = STATE(3062),
    [sym_compiler_directive_decl] = STATE(3062),
    [sym_fsi_directive_decl] = STATE(3062),
    [sym_preproc_line] = STATE(3062),
    [aux_sym_rules_repeat1] = STATE(3068),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7886),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_inherit] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7920),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3063)] = {
    [sym_block_comment] = STATE(3063),
    [sym_compiler_directive_decl] = STATE(3063),
    [sym_fsi_directive_decl] = STATE(3063),
    [sym_preproc_line] = STATE(3063),
    [aux_sym_rules_repeat1] = STATE(3069),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [aux_sym_access_modifier_token1] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(7886),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_default] = ACTIONS(6607),
    [anon_sym_static] = ACTIONS(6607),
    [anon_sym_member] = ACTIONS(6607),
    [anon_sym_interface] = ACTIONS(6607),
    [anon_sym_abstract] = ACTIONS(6607),
    [anon_sym_val] = ACTIONS(6607),
    [anon_sym_override] = ACTIONS(6607),
    [anon_sym_inherit] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(7920),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3064)] = {
    [sym_block_comment] = STATE(3064),
    [sym_compiler_directive_decl] = STATE(3064),
    [sym_fsi_directive_decl] = STATE(3064),
    [sym_preproc_line] = STATE(3064),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7759),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3065)] = {
    [sym_block_comment] = STATE(3065),
    [sym_compiler_directive_decl] = STATE(3065),
    [sym_fsi_directive_decl] = STATE(3065),
    [sym_preproc_line] = STATE(3065),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [anon_sym_inherit] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [anon_sym_POUNDelse] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3066)] = {
    [sym_block_comment] = STATE(3066),
    [sym_compiler_directive_decl] = STATE(3066),
    [sym_fsi_directive_decl] = STATE(3066),
    [sym_preproc_line] = STATE(3066),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_and] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [anon_sym_POUNDelse] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3067)] = {
    [sym_block_comment] = STATE(3067),
    [sym_compiler_directive_decl] = STATE(3067),
    [sym_fsi_directive_decl] = STATE(3067),
    [sym_preproc_line] = STATE(3067),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_inherit] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3068)] = {
    [sym_block_comment] = STATE(3068),
    [sym_compiler_directive_decl] = STATE(3068),
    [sym_fsi_directive_decl] = STATE(3068),
    [sym_preproc_line] = STATE(3068),
    [aux_sym_rules_repeat1] = STATE(3069),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [aux_sym_access_modifier_token1] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(7886),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_default] = ACTIONS(6618),
    [anon_sym_static] = ACTIONS(6618),
    [anon_sym_member] = ACTIONS(6618),
    [anon_sym_interface] = ACTIONS(6618),
    [anon_sym_abstract] = ACTIONS(6618),
    [anon_sym_val] = ACTIONS(6618),
    [anon_sym_override] = ACTIONS(6618),
    [anon_sym_inherit] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(7923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(3069)] = {
    [sym_block_comment] = STATE(3069),
    [sym_compiler_directive_decl] = STATE(3069),
    [sym_fsi_directive_decl] = STATE(3069),
    [sym_preproc_line] = STATE(3069),
    [aux_sym_rules_repeat1] = STATE(3069),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(7926),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_inherit] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(7929),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3070)] = {
    [sym_block_comment] = STATE(3070),
    [sym_compiler_directive_decl] = STATE(3070),
    [sym_fsi_directive_decl] = STATE(3070),
    [sym_preproc_line] = STATE(3070),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [anon_sym_inherit] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [anon_sym_POUNDelse] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3071)] = {
    [sym_block_comment] = STATE(3071),
    [sym_compiler_directive_decl] = STATE(3071),
    [sym_fsi_directive_decl] = STATE(3071),
    [sym_preproc_line] = STATE(3071),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_and] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(7932),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [anon_sym_POUNDelse] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3072)] = {
    [sym_block_comment] = STATE(3072),
    [sym_compiler_directive_decl] = STATE(3072),
    [sym_fsi_directive_decl] = STATE(3072),
    [sym_preproc_line] = STATE(3072),
    [aux_sym_long_identifier_repeat1] = STATE(3074),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [aux_sym_access_modifier_token1] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(7934),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_default] = ACTIONS(6625),
    [anon_sym_static] = ACTIONS(6625),
    [anon_sym_member] = ACTIONS(6625),
    [anon_sym_interface] = ACTIONS(6625),
    [anon_sym_abstract] = ACTIONS(6625),
    [anon_sym_val] = ACTIONS(6625),
    [anon_sym_override] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3073)] = {
    [sym_block_comment] = STATE(3073),
    [sym_compiler_directive_decl] = STATE(3073),
    [sym_fsi_directive_decl] = STATE(3073),
    [sym_preproc_line] = STATE(3073),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(7938),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [anon_sym_inherit] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [anon_sym_POUNDelse] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3074)] = {
    [sym_block_comment] = STATE(3074),
    [sym_compiler_directive_decl] = STATE(3074),
    [sym_fsi_directive_decl] = STATE(3074),
    [sym_preproc_line] = STATE(3074),
    [aux_sym_long_identifier_repeat1] = STATE(3075),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3075)] = {
    [sym_block_comment] = STATE(3075),
    [sym_compiler_directive_decl] = STATE(3075),
    [sym_fsi_directive_decl] = STATE(3075),
    [sym_preproc_line] = STATE(3075),
    [aux_sym_long_identifier_repeat1] = STATE(3075),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7942),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3076)] = {
    [sym_block_comment] = STATE(3076),
    [sym_compiler_directive_decl] = STATE(3076),
    [sym_fsi_directive_decl] = STATE(3076),
    [sym_preproc_line] = STATE(3076),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7444),
    [anon_sym_uy] = ACTIONS(7446),
    [anon_sym_s] = ACTIONS(7448),
    [anon_sym_us] = ACTIONS(7450),
    [anon_sym_l] = ACTIONS(7452),
    [aux_sym_uint32_token1] = ACTIONS(7454),
    [anon_sym_n] = ACTIONS(7456),
    [anon_sym_un] = ACTIONS(7458),
    [anon_sym_L] = ACTIONS(7460),
    [aux_sym_uint64_token1] = ACTIONS(7462),
    [anon_sym_lf] = ACTIONS(7945),
    [anon_sym_LF] = ACTIONS(7947),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3077)] = {
    [sym_block_comment] = STATE(3077),
    [sym_compiler_directive_decl] = STATE(3077),
    [sym_fsi_directive_decl] = STATE(3077),
    [sym_preproc_line] = STATE(3077),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3078)] = {
    [sym_block_comment] = STATE(3078),
    [sym_compiler_directive_decl] = STATE(3078),
    [sym_fsi_directive_decl] = STATE(3078),
    [sym_preproc_line] = STATE(3078),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_and] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [aux_sym_access_modifier_token1] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_default] = ACTIONS(6564),
    [anon_sym_static] = ACTIONS(6564),
    [anon_sym_member] = ACTIONS(6564),
    [anon_sym_interface] = ACTIONS(6564),
    [anon_sym_abstract] = ACTIONS(6564),
    [anon_sym_val] = ACTIONS(6564),
    [anon_sym_override] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(3079)] = {
    [sym_block_comment] = STATE(3079),
    [sym_compiler_directive_decl] = STATE(3079),
    [sym_fsi_directive_decl] = STATE(3079),
    [sym_preproc_line] = STATE(3079),
    [aux_sym_sequential_expression_repeat1] = STATE(3079),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [anon_sym_inherit] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [anon_sym_POUNDelse] = ACTIONS(6798),
    [sym__newline] = ACTIONS(7949),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3080)] = {
    [sym_type_arguments] = STATE(3697),
    [sym_long_identifier] = STATE(3698),
    [sym_block_comment] = STATE(3080),
    [sym_compiler_directive_decl] = STATE(3080),
    [sym_fsi_directive_decl] = STATE(3080),
    [sym_preproc_line] = STATE(3080),
    [aux_sym_compound_type_repeat1] = STATE(3661),
    [sym_identifier] = ACTIONS(7792),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(7794),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(7796),
    [anon_sym_LT] = ACTIONS(7798),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7800),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3081)] = {
    [sym_block_comment] = STATE(3081),
    [sym_compiler_directive_decl] = STATE(3081),
    [sym_fsi_directive_decl] = STATE(3081),
    [sym_preproc_line] = STATE(3081),
    [aux_sym_sequential_expression_repeat1] = STATE(2992),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [anon_sym_POUNDelse] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3082)] = {
    [sym_block_comment] = STATE(3082),
    [sym_compiler_directive_decl] = STATE(3082),
    [sym_fsi_directive_decl] = STATE(3082),
    [sym_preproc_line] = STATE(3082),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_DASH_GT] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7338),
    [anon_sym_uy] = ACTIONS(7340),
    [anon_sym_s] = ACTIONS(7342),
    [anon_sym_us] = ACTIONS(7344),
    [anon_sym_l] = ACTIONS(7346),
    [aux_sym_uint32_token1] = ACTIONS(7348),
    [anon_sym_n] = ACTIONS(7350),
    [anon_sym_un] = ACTIONS(7352),
    [anon_sym_L] = ACTIONS(7354),
    [aux_sym_uint64_token1] = ACTIONS(7356),
    [anon_sym_lf] = ACTIONS(7952),
    [anon_sym_LF] = ACTIONS(7954),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3083)] = {
    [sym_type_arguments] = STATE(3697),
    [sym_long_identifier] = STATE(3698),
    [sym_block_comment] = STATE(3083),
    [sym_compiler_directive_decl] = STATE(3083),
    [sym_fsi_directive_decl] = STATE(3083),
    [sym_preproc_line] = STATE(3083),
    [aux_sym_compound_type_repeat1] = STATE(3661),
    [sym_identifier] = ACTIONS(7792),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(7794),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(7796),
    [anon_sym_LT] = ACTIONS(7798),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7800),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3084)] = {
    [sym_block_comment] = STATE(3084),
    [sym_compiler_directive_decl] = STATE(3084),
    [sym_fsi_directive_decl] = STATE(3084),
    [sym_preproc_line] = STATE(3084),
    [aux_sym_long_identifier_repeat1] = STATE(3027),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7956),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3085)] = {
    [sym_block_comment] = STATE(3085),
    [sym_compiler_directive_decl] = STATE(3085),
    [sym_fsi_directive_decl] = STATE(3085),
    [sym_preproc_line] = STATE(3085),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [anon_sym_inherit] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [anon_sym_POUNDelse] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3086)] = {
    [sym_block_comment] = STATE(3086),
    [sym_compiler_directive_decl] = STATE(3086),
    [sym_fsi_directive_decl] = STATE(3086),
    [sym_preproc_line] = STATE(3086),
    [aux_sym_long_identifier_repeat1] = STATE(3088),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(7958),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3087)] = {
    [sym_block_comment] = STATE(3087),
    [sym_compiler_directive_decl] = STATE(3087),
    [sym_fsi_directive_decl] = STATE(3087),
    [sym_preproc_line] = STATE(3087),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(7960),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_inherit] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3088)] = {
    [sym_block_comment] = STATE(3088),
    [sym_compiler_directive_decl] = STATE(3088),
    [sym_fsi_directive_decl] = STATE(3088),
    [sym_preproc_line] = STATE(3088),
    [aux_sym_long_identifier_repeat1] = STATE(3088),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(7962),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3089)] = {
    [sym_block_comment] = STATE(3089),
    [sym_compiler_directive_decl] = STATE(3089),
    [sym_fsi_directive_decl] = STATE(3089),
    [sym_preproc_line] = STATE(3089),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [anon_sym_inherit] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3090)] = {
    [sym_block_comment] = STATE(3090),
    [sym_compiler_directive_decl] = STATE(3090),
    [sym_fsi_directive_decl] = STATE(3090),
    [sym_preproc_line] = STATE(3090),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_and] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [anon_sym_POUNDelse] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3091)] = {
    [sym_block_comment] = STATE(3091),
    [sym_compiler_directive_decl] = STATE(3091),
    [sym_fsi_directive_decl] = STATE(3091),
    [sym_preproc_line] = STATE(3091),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_and] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [anon_sym_POUNDelse] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3092)] = {
    [sym_block_comment] = STATE(3092),
    [sym_compiler_directive_decl] = STATE(3092),
    [sym_fsi_directive_decl] = STATE(3092),
    [sym_preproc_line] = STATE(3092),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_and] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [anon_sym_POUNDelse] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3093)] = {
    [sym__else_expression] = STATE(4016),
    [sym_elif_expression] = STATE(3752),
    [sym_block_comment] = STATE(3093),
    [sym_compiler_directive_decl] = STATE(3093),
    [sym_fsi_directive_decl] = STATE(3093),
    [sym_preproc_line] = STATE(3093),
    [aux_sym_if_expression_repeat1] = STATE(3177),
    [ts_builtin_sym_end] = ACTIONS(6317),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_namespace] = ACTIONS(6319),
    [anon_sym_module] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_open] = ACTIONS(6319),
    [anon_sym_type] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7965),
    [anon_sym_elif] = ACTIONS(7967),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_exception] = ACTIONS(6319),
    [anon_sym_extern] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6317),
    [anon_sym_POUNDload] = ACTIONS(6317),
    [anon_sym_POUNDtime] = ACTIONS(6317),
    [anon_sym_POUNDI] = ACTIONS(6317),
    [anon_sym_POUNDhelp] = ACTIONS(6317),
    [anon_sym_POUNDquit] = ACTIONS(6317),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(3094)] = {
    [sym_block_comment] = STATE(3094),
    [sym_compiler_directive_decl] = STATE(3094),
    [sym_fsi_directive_decl] = STATE(3094),
    [sym_preproc_line] = STATE(3094),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_and] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [anon_sym_POUNDelse] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3095)] = {
    [sym_block_comment] = STATE(3095),
    [sym_compiler_directive_decl] = STATE(3095),
    [sym_fsi_directive_decl] = STATE(3095),
    [sym_preproc_line] = STATE(3095),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_DOT_DOT] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__dedent] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3096)] = {
    [sym_block_comment] = STATE(3096),
    [sym_compiler_directive_decl] = STATE(3096),
    [sym_fsi_directive_decl] = STATE(3096),
    [sym_preproc_line] = STATE(3096),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_and] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [anon_sym_POUNDelse] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3097)] = {
    [sym_block_comment] = STATE(3097),
    [sym_compiler_directive_decl] = STATE(3097),
    [sym_fsi_directive_decl] = STATE(3097),
    [sym_preproc_line] = STATE(3097),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_and] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__dedent] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3098)] = {
    [sym_block_comment] = STATE(3098),
    [sym_compiler_directive_decl] = STATE(3098),
    [sym_fsi_directive_decl] = STATE(3098),
    [sym_preproc_line] = STATE(3098),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_and] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [anon_sym_POUNDelse] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3099)] = {
    [sym_block_comment] = STATE(3099),
    [sym_compiler_directive_decl] = STATE(3099),
    [sym_fsi_directive_decl] = STATE(3099),
    [sym_preproc_line] = STATE(3099),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3100)] = {
    [sym_block_comment] = STATE(3100),
    [sym_compiler_directive_decl] = STATE(3100),
    [sym_fsi_directive_decl] = STATE(3100),
    [sym_preproc_line] = STATE(3100),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_and] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [anon_sym_POUNDelse] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3101)] = {
    [sym_block_comment] = STATE(3101),
    [sym_compiler_directive_decl] = STATE(3101),
    [sym_fsi_directive_decl] = STATE(3101),
    [sym_preproc_line] = STATE(3101),
    [aux_sym_type_argument_repeat1] = STATE(3101),
    [ts_builtin_sym_end] = ACTIONS(6310),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_namespace] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(7969),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3102)] = {
    [sym_block_comment] = STATE(3102),
    [sym_compiler_directive_decl] = STATE(3102),
    [sym_fsi_directive_decl] = STATE(3102),
    [sym_preproc_line] = STATE(3102),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_and] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__dedent] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3103)] = {
    [sym_block_comment] = STATE(3103),
    [sym_compiler_directive_decl] = STATE(3103),
    [sym_fsi_directive_decl] = STATE(3103),
    [sym_preproc_line] = STATE(3103),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_and] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [anon_sym_POUNDelse] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(3104)] = {
    [sym_block_comment] = STATE(3104),
    [sym_compiler_directive_decl] = STATE(3104),
    [sym_fsi_directive_decl] = STATE(3104),
    [sym_preproc_line] = STATE(3104),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7640),
    [anon_sym_uy] = ACTIONS(7642),
    [anon_sym_s] = ACTIONS(7644),
    [anon_sym_us] = ACTIONS(7646),
    [anon_sym_l] = ACTIONS(7648),
    [aux_sym_uint32_token1] = ACTIONS(7650),
    [anon_sym_n] = ACTIONS(7652),
    [anon_sym_un] = ACTIONS(7654),
    [anon_sym_L] = ACTIONS(7656),
    [aux_sym_uint64_token1] = ACTIONS(7658),
    [anon_sym_lf] = ACTIONS(7972),
    [anon_sym_LF] = ACTIONS(7974),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__quoted_close] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3105)] = {
    [sym_block_comment] = STATE(3105),
    [sym_compiler_directive_decl] = STATE(3105),
    [sym_fsi_directive_decl] = STATE(3105),
    [sym_preproc_line] = STATE(3105),
    [aux_sym_type_argument_repeat1] = STATE(3311),
    [ts_builtin_sym_end] = ACTIONS(6333),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_namespace] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(7976),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3106)] = {
    [sym_block_comment] = STATE(3106),
    [sym_compiler_directive_decl] = STATE(3106),
    [sym_fsi_directive_decl] = STATE(3106),
    [sym_preproc_line] = STATE(3106),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3107)] = {
    [sym_block_comment] = STATE(3107),
    [sym_compiler_directive_decl] = STATE(3107),
    [sym_fsi_directive_decl] = STATE(3107),
    [sym_preproc_line] = STATE(3107),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7913),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [anon_sym_inherit] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3108)] = {
    [sym_block_comment] = STATE(3108),
    [sym_compiler_directive_decl] = STATE(3108),
    [sym_fsi_directive_decl] = STATE(3108),
    [sym_preproc_line] = STATE(3108),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_and] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3109)] = {
    [sym_block_comment] = STATE(3109),
    [sym_compiler_directive_decl] = STATE(3109),
    [sym_fsi_directive_decl] = STATE(3109),
    [sym_preproc_line] = STATE(3109),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_DOT_DOT] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__dedent] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3110)] = {
    [sym_block_comment] = STATE(3110),
    [sym_compiler_directive_decl] = STATE(3110),
    [sym_fsi_directive_decl] = STATE(3110),
    [sym_preproc_line] = STATE(3110),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_DOT_DOT] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__dedent] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3111)] = {
    [sym_block_comment] = STATE(3111),
    [sym_compiler_directive_decl] = STATE(3111),
    [sym_fsi_directive_decl] = STATE(3111),
    [sym_preproc_line] = STATE(3111),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_and] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3112)] = {
    [sym_type_arguments] = STATE(3730),
    [sym_long_identifier] = STATE(3759),
    [sym_block_comment] = STATE(3112),
    [sym_compiler_directive_decl] = STATE(3112),
    [sym_fsi_directive_decl] = STATE(3112),
    [sym_preproc_line] = STATE(3112),
    [aux_sym_compound_type_repeat1] = STATE(3689),
    [sym_identifier] = ACTIONS(7979),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_GT_RBRACK] = ACTIONS(6209),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_RBRACK] = ACTIONS(6209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_RBRACE] = ACTIONS(6209),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_to] = ACTIONS(6213),
    [anon_sym_downto] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(7981),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_DOT_DOT2] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(7985),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7987),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3113)] = {
    [sym_block_comment] = STATE(3113),
    [sym_compiler_directive_decl] = STATE(3113),
    [sym_fsi_directive_decl] = STATE(3113),
    [sym_preproc_line] = STATE(3113),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [anon_sym_inherit] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [anon_sym_POUNDelse] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3114)] = {
    [sym_block_comment] = STATE(3114),
    [sym_compiler_directive_decl] = STATE(3114),
    [sym_fsi_directive_decl] = STATE(3114),
    [sym_preproc_line] = STATE(3114),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_DOT_DOT] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__dedent] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3115)] = {
    [sym_block_comment] = STATE(3115),
    [sym_compiler_directive_decl] = STATE(3115),
    [sym_fsi_directive_decl] = STATE(3115),
    [sym_preproc_line] = STATE(3115),
    [aux_sym_compound_type_repeat1] = STATE(3125),
    [ts_builtin_sym_end] = ACTIONS(6340),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_namespace] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(7597),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(3116)] = {
    [sym_block_comment] = STATE(3116),
    [sym_compiler_directive_decl] = STATE(3116),
    [sym_fsi_directive_decl] = STATE(3116),
    [sym_preproc_line] = STATE(3116),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [anon_sym_inherit] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [anon_sym_POUNDelse] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(3117)] = {
    [sym_block_comment] = STATE(3117),
    [sym_compiler_directive_decl] = STATE(3117),
    [sym_fsi_directive_decl] = STATE(3117),
    [sym_preproc_line] = STATE(3117),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_and] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__dedent] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3118)] = {
    [sym__else_expression] = STATE(3949),
    [sym_elif_expression] = STATE(3754),
    [sym_block_comment] = STATE(3118),
    [sym_compiler_directive_decl] = STATE(3118),
    [sym_fsi_directive_decl] = STATE(3118),
    [sym_preproc_line] = STATE(3118),
    [aux_sym_if_expression_repeat1] = STATE(3120),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_module] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_open] = ACTIONS(6319),
    [anon_sym_type] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(7989),
    [anon_sym_elif] = ACTIONS(7991),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_exception] = ACTIONS(6319),
    [anon_sym_extern] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6317),
    [anon_sym_POUNDload] = ACTIONS(6317),
    [anon_sym_POUNDtime] = ACTIONS(6317),
    [anon_sym_POUNDI] = ACTIONS(6317),
    [anon_sym_POUNDhelp] = ACTIONS(6317),
    [anon_sym_POUNDquit] = ACTIONS(6317),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [anon_sym_POUNDelse] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(3119)] = {
    [sym_block_comment] = STATE(3119),
    [sym_compiler_directive_decl] = STATE(3119),
    [sym_fsi_directive_decl] = STATE(3119),
    [sym_preproc_line] = STATE(3119),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_DOT_DOT] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__dedent] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3120)] = {
    [sym__else_expression] = STATE(4023),
    [sym_elif_expression] = STATE(3754),
    [sym_block_comment] = STATE(3120),
    [sym_compiler_directive_decl] = STATE(3120),
    [sym_fsi_directive_decl] = STATE(3120),
    [sym_preproc_line] = STATE(3120),
    [aux_sym_if_expression_repeat1] = STATE(3372),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_module] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_open] = ACTIONS(6329),
    [anon_sym_type] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7989),
    [anon_sym_elif] = ACTIONS(7991),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_exception] = ACTIONS(6329),
    [anon_sym_extern] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6327),
    [anon_sym_POUNDload] = ACTIONS(6327),
    [anon_sym_POUNDtime] = ACTIONS(6327),
    [anon_sym_POUNDI] = ACTIONS(6327),
    [anon_sym_POUNDhelp] = ACTIONS(6327),
    [anon_sym_POUNDquit] = ACTIONS(6327),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [anon_sym_POUNDelse] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(3121)] = {
    [sym_block_comment] = STATE(3121),
    [sym_compiler_directive_decl] = STATE(3121),
    [sym_fsi_directive_decl] = STATE(3121),
    [sym_preproc_line] = STATE(3121),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_DOT_DOT] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__dedent] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3122)] = {
    [sym_block_comment] = STATE(3122),
    [sym_compiler_directive_decl] = STATE(3122),
    [sym_fsi_directive_decl] = STATE(3122),
    [sym_preproc_line] = STATE(3122),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_DOT_DOT] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__dedent] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3123)] = {
    [sym_block_comment] = STATE(3123),
    [sym_compiler_directive_decl] = STATE(3123),
    [sym_fsi_directive_decl] = STATE(3123),
    [sym_preproc_line] = STATE(3123),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [anon_sym_inherit] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [anon_sym_POUNDelse] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3124)] = {
    [sym_block_comment] = STATE(3124),
    [sym_compiler_directive_decl] = STATE(3124),
    [sym_fsi_directive_decl] = STATE(3124),
    [sym_preproc_line] = STATE(3124),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [anon_sym_inherit] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [anon_sym_POUNDelse] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(3125)] = {
    [sym_block_comment] = STATE(3125),
    [sym_compiler_directive_decl] = STATE(3125),
    [sym_fsi_directive_decl] = STATE(3125),
    [sym_preproc_line] = STATE(3125),
    [aux_sym_compound_type_repeat1] = STATE(3125),
    [ts_builtin_sym_end] = ACTIONS(6195),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_namespace] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(7993),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3126)] = {
    [sym_block_comment] = STATE(3126),
    [sym_compiler_directive_decl] = STATE(3126),
    [sym_fsi_directive_decl] = STATE(3126),
    [sym_preproc_line] = STATE(3126),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_and] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(7996),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__dedent] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3127)] = {
    [sym_block_comment] = STATE(3127),
    [sym_compiler_directive_decl] = STATE(3127),
    [sym_fsi_directive_decl] = STATE(3127),
    [sym_preproc_line] = STATE(3127),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_and] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__dedent] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3128)] = {
    [sym_block_comment] = STATE(3128),
    [sym_compiler_directive_decl] = STATE(3128),
    [sym_fsi_directive_decl] = STATE(3128),
    [sym_preproc_line] = STATE(3128),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_GT] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7552),
    [anon_sym_uy] = ACTIONS(7554),
    [anon_sym_s] = ACTIONS(7556),
    [anon_sym_us] = ACTIONS(7558),
    [anon_sym_l] = ACTIONS(7560),
    [aux_sym_uint32_token1] = ACTIONS(7562),
    [anon_sym_n] = ACTIONS(7564),
    [anon_sym_un] = ACTIONS(7566),
    [anon_sym_L] = ACTIONS(7568),
    [aux_sym_uint64_token1] = ACTIONS(7570),
    [anon_sym_lf] = ACTIONS(7998),
    [anon_sym_LF] = ACTIONS(8000),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3129)] = {
    [sym_block_comment] = STATE(3129),
    [sym_compiler_directive_decl] = STATE(3129),
    [sym_fsi_directive_decl] = STATE(3129),
    [sym_preproc_line] = STATE(3129),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8002),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [anon_sym_inherit] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3130)] = {
    [sym_block_comment] = STATE(3130),
    [sym_compiler_directive_decl] = STATE(3130),
    [sym_fsi_directive_decl] = STATE(3130),
    [sym_preproc_line] = STATE(3130),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_inherit] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3131)] = {
    [sym_block_comment] = STATE(3131),
    [sym_compiler_directive_decl] = STATE(3131),
    [sym_fsi_directive_decl] = STATE(3131),
    [sym_preproc_line] = STATE(3131),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [anon_sym_inherit] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3132)] = {
    [sym_type_arguments] = STATE(3730),
    [sym_long_identifier] = STATE(3759),
    [sym_block_comment] = STATE(3132),
    [sym_compiler_directive_decl] = STATE(3132),
    [sym_fsi_directive_decl] = STATE(3132),
    [sym_preproc_line] = STATE(3132),
    [aux_sym_compound_type_repeat1] = STATE(3689),
    [sym_identifier] = ACTIONS(7979),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_GT_RBRACK] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_RBRACK] = ACTIONS(6195),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_RBRACE] = ACTIONS(6195),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_to] = ACTIONS(6193),
    [anon_sym_downto] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(7981),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_DOT_DOT2] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(7985),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7987),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3133)] = {
    [sym_block_comment] = STATE(3133),
    [sym_compiler_directive_decl] = STATE(3133),
    [sym_fsi_directive_decl] = STATE(3133),
    [sym_preproc_line] = STATE(3133),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_and] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [anon_sym_POUNDelse] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(3134)] = {
    [sym_block_comment] = STATE(3134),
    [sym_compiler_directive_decl] = STATE(3134),
    [sym_fsi_directive_decl] = STATE(3134),
    [sym_preproc_line] = STATE(3134),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [anon_sym_inherit] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [anon_sym_POUNDelse] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3135)] = {
    [sym_block_comment] = STATE(3135),
    [sym_compiler_directive_decl] = STATE(3135),
    [sym_fsi_directive_decl] = STATE(3135),
    [sym_preproc_line] = STATE(3135),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3136)] = {
    [sym_block_comment] = STATE(3136),
    [sym_compiler_directive_decl] = STATE(3136),
    [sym_fsi_directive_decl] = STATE(3136),
    [sym_preproc_line] = STATE(3136),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [anon_sym_inherit] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [anon_sym_POUNDelse] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3137)] = {
    [sym_block_comment] = STATE(3137),
    [sym_compiler_directive_decl] = STATE(3137),
    [sym_fsi_directive_decl] = STATE(3137),
    [sym_preproc_line] = STATE(3137),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_and] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [anon_sym_POUNDelse] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3138)] = {
    [sym_block_comment] = STATE(3138),
    [sym_compiler_directive_decl] = STATE(3138),
    [sym_fsi_directive_decl] = STATE(3138),
    [sym_preproc_line] = STATE(3138),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_and] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [anon_sym_POUNDelse] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(3139)] = {
    [sym_block_comment] = STATE(3139),
    [sym_compiler_directive_decl] = STATE(3139),
    [sym_fsi_directive_decl] = STATE(3139),
    [sym_preproc_line] = STATE(3139),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [anon_sym_inherit] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [anon_sym_POUNDelse] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3140)] = {
    [sym_block_comment] = STATE(3140),
    [sym_compiler_directive_decl] = STATE(3140),
    [sym_fsi_directive_decl] = STATE(3140),
    [sym_preproc_line] = STATE(3140),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_and] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [anon_sym_POUNDelse] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3141)] = {
    [sym_block_comment] = STATE(3141),
    [sym_compiler_directive_decl] = STATE(3141),
    [sym_fsi_directive_decl] = STATE(3141),
    [sym_preproc_line] = STATE(3141),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_and] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [anon_sym_POUNDelse] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(3142)] = {
    [sym_block_comment] = STATE(3142),
    [sym_compiler_directive_decl] = STATE(3142),
    [sym_fsi_directive_decl] = STATE(3142),
    [sym_preproc_line] = STATE(3142),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [anon_sym_inherit] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [anon_sym_POUNDelse] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(3143)] = {
    [sym_block_comment] = STATE(3143),
    [sym_compiler_directive_decl] = STATE(3143),
    [sym_fsi_directive_decl] = STATE(3143),
    [sym_preproc_line] = STATE(3143),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [anon_sym_inherit] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [anon_sym_POUNDelse] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3144)] = {
    [sym_block_comment] = STATE(3144),
    [sym_compiler_directive_decl] = STATE(3144),
    [sym_fsi_directive_decl] = STATE(3144),
    [sym_preproc_line] = STATE(3144),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [anon_sym_inherit] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [anon_sym_POUNDelse] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3145)] = {
    [sym_block_comment] = STATE(3145),
    [sym_compiler_directive_decl] = STATE(3145),
    [sym_fsi_directive_decl] = STATE(3145),
    [sym_preproc_line] = STATE(3145),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8004),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3146)] = {
    [sym_block_comment] = STATE(3146),
    [sym_compiler_directive_decl] = STATE(3146),
    [sym_fsi_directive_decl] = STATE(3146),
    [sym_preproc_line] = STATE(3146),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [anon_sym_inherit] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [anon_sym_POUNDelse] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3147)] = {
    [sym_block_comment] = STATE(3147),
    [sym_compiler_directive_decl] = STATE(3147),
    [sym_fsi_directive_decl] = STATE(3147),
    [sym_preproc_line] = STATE(3147),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3148)] = {
    [sym_block_comment] = STATE(3148),
    [sym_compiler_directive_decl] = STATE(3148),
    [sym_fsi_directive_decl] = STATE(3148),
    [sym_preproc_line] = STATE(3148),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [anon_sym_inherit] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3149)] = {
    [sym_block_comment] = STATE(3149),
    [sym_compiler_directive_decl] = STATE(3149),
    [sym_fsi_directive_decl] = STATE(3149),
    [sym_preproc_line] = STATE(3149),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_and] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3150)] = {
    [sym_block_comment] = STATE(3150),
    [sym_compiler_directive_decl] = STATE(3150),
    [sym_fsi_directive_decl] = STATE(3150),
    [sym_preproc_line] = STATE(3150),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [anon_sym_inherit] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [anon_sym_POUNDelse] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3151)] = {
    [sym_block_comment] = STATE(3151),
    [sym_compiler_directive_decl] = STATE(3151),
    [sym_fsi_directive_decl] = STATE(3151),
    [sym_preproc_line] = STATE(3151),
    [aux_sym_long_identifier_repeat1] = STATE(3224),
    [ts_builtin_sym_end] = ACTIONS(6358),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_namespace] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8006),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3152)] = {
    [sym_block_comment] = STATE(3152),
    [sym_compiler_directive_decl] = STATE(3152),
    [sym_fsi_directive_decl] = STATE(3152),
    [sym_preproc_line] = STATE(3152),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_and] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__dedent] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3153)] = {
    [sym_block_comment] = STATE(3153),
    [sym_compiler_directive_decl] = STATE(3153),
    [sym_fsi_directive_decl] = STATE(3153),
    [sym_preproc_line] = STATE(3153),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [anon_sym_inherit] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [anon_sym_POUNDelse] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3154)] = {
    [sym_block_comment] = STATE(3154),
    [sym_compiler_directive_decl] = STATE(3154),
    [sym_fsi_directive_decl] = STATE(3154),
    [sym_preproc_line] = STATE(3154),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [anon_sym_inherit] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [anon_sym_POUNDelse] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3155)] = {
    [sym_block_comment] = STATE(3155),
    [sym_compiler_directive_decl] = STATE(3155),
    [sym_fsi_directive_decl] = STATE(3155),
    [sym_preproc_line] = STATE(3155),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_and] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3156)] = {
    [sym_block_comment] = STATE(3156),
    [sym_compiler_directive_decl] = STATE(3156),
    [sym_fsi_directive_decl] = STATE(3156),
    [sym_preproc_line] = STATE(3156),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8008),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3157)] = {
    [sym_block_comment] = STATE(3157),
    [sym_compiler_directive_decl] = STATE(3157),
    [sym_fsi_directive_decl] = STATE(3157),
    [sym_preproc_line] = STATE(3157),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3158)] = {
    [sym_block_comment] = STATE(3158),
    [sym_compiler_directive_decl] = STATE(3158),
    [sym_fsi_directive_decl] = STATE(3158),
    [sym_preproc_line] = STATE(3158),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [anon_sym_inherit] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [anon_sym_POUNDelse] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3159)] = {
    [sym_block_comment] = STATE(3159),
    [sym_compiler_directive_decl] = STATE(3159),
    [sym_fsi_directive_decl] = STATE(3159),
    [sym_preproc_line] = STATE(3159),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [anon_sym_inherit] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [anon_sym_POUNDelse] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3160)] = {
    [sym_block_comment] = STATE(3160),
    [sym_compiler_directive_decl] = STATE(3160),
    [sym_fsi_directive_decl] = STATE(3160),
    [sym_preproc_line] = STATE(3160),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [anon_sym_inherit] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [anon_sym_POUNDelse] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3161)] = {
    [sym_block_comment] = STATE(3161),
    [sym_compiler_directive_decl] = STATE(3161),
    [sym_fsi_directive_decl] = STATE(3161),
    [sym_preproc_line] = STATE(3161),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3162)] = {
    [sym_block_comment] = STATE(3162),
    [sym_compiler_directive_decl] = STATE(3162),
    [sym_fsi_directive_decl] = STATE(3162),
    [sym_preproc_line] = STATE(3162),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [anon_sym_inherit] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [anon_sym_POUNDelse] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3163)] = {
    [sym_block_comment] = STATE(3163),
    [sym_compiler_directive_decl] = STATE(3163),
    [sym_fsi_directive_decl] = STATE(3163),
    [sym_preproc_line] = STATE(3163),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [anon_sym_inherit] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [anon_sym_POUNDelse] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3164)] = {
    [sym_block_comment] = STATE(3164),
    [sym_compiler_directive_decl] = STATE(3164),
    [sym_fsi_directive_decl] = STATE(3164),
    [sym_preproc_line] = STATE(3164),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_inherit] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3165)] = {
    [sym_block_comment] = STATE(3165),
    [sym_compiler_directive_decl] = STATE(3165),
    [sym_fsi_directive_decl] = STATE(3165),
    [sym_preproc_line] = STATE(3165),
    [aux_sym_sequential_expression_repeat1] = STATE(3320),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [anon_sym_inherit] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3166)] = {
    [sym_block_comment] = STATE(3166),
    [sym_compiler_directive_decl] = STATE(3166),
    [sym_fsi_directive_decl] = STATE(3166),
    [sym_preproc_line] = STATE(3166),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_inherit] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3167)] = {
    [sym_block_comment] = STATE(3167),
    [sym_compiler_directive_decl] = STATE(3167),
    [sym_fsi_directive_decl] = STATE(3167),
    [sym_preproc_line] = STATE(3167),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_DOT_DOT] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__dedent] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3168)] = {
    [sym_block_comment] = STATE(3168),
    [sym_compiler_directive_decl] = STATE(3168),
    [sym_fsi_directive_decl] = STATE(3168),
    [sym_preproc_line] = STATE(3168),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [anon_sym_inherit] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [anon_sym_POUNDelse] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3169)] = {
    [sym_block_comment] = STATE(3169),
    [sym_compiler_directive_decl] = STATE(3169),
    [sym_fsi_directive_decl] = STATE(3169),
    [sym_preproc_line] = STATE(3169),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [anon_sym_inherit] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [anon_sym_POUNDelse] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3170)] = {
    [sym_block_comment] = STATE(3170),
    [sym_compiler_directive_decl] = STATE(3170),
    [sym_fsi_directive_decl] = STATE(3170),
    [sym_preproc_line] = STATE(3170),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3171)] = {
    [sym_block_comment] = STATE(3171),
    [sym_compiler_directive_decl] = STATE(3171),
    [sym_fsi_directive_decl] = STATE(3171),
    [sym_preproc_line] = STATE(3171),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [anon_sym_inherit] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [anon_sym_POUNDelse] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3172)] = {
    [sym_type_arguments] = STATE(3730),
    [sym_long_identifier] = STATE(3759),
    [sym_block_comment] = STATE(3172),
    [sym_compiler_directive_decl] = STATE(3172),
    [sym_fsi_directive_decl] = STATE(3172),
    [sym_preproc_line] = STATE(3172),
    [aux_sym_compound_type_repeat1] = STATE(3689),
    [sym_identifier] = ACTIONS(7979),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_GT_RBRACK] = ACTIONS(6259),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_RBRACK] = ACTIONS(6259),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_RBRACE] = ACTIONS(6259),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_to] = ACTIONS(6261),
    [anon_sym_downto] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(7981),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_DOT_DOT2] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(7985),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7987),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3173)] = {
    [sym_block_comment] = STATE(3173),
    [sym_compiler_directive_decl] = STATE(3173),
    [sym_fsi_directive_decl] = STATE(3173),
    [sym_preproc_line] = STATE(3173),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [anon_sym_inherit] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [anon_sym_POUNDelse] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3174)] = {
    [sym_block_comment] = STATE(3174),
    [sym_compiler_directive_decl] = STATE(3174),
    [sym_fsi_directive_decl] = STATE(3174),
    [sym_preproc_line] = STATE(3174),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [anon_sym_inherit] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [anon_sym_POUNDelse] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3175)] = {
    [sym_block_comment] = STATE(3175),
    [sym_compiler_directive_decl] = STATE(3175),
    [sym_fsi_directive_decl] = STATE(3175),
    [sym_preproc_line] = STATE(3175),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7673),
    [anon_sym_uy] = ACTIONS(7675),
    [anon_sym_s] = ACTIONS(7677),
    [anon_sym_us] = ACTIONS(7679),
    [anon_sym_l] = ACTIONS(7681),
    [aux_sym_uint32_token1] = ACTIONS(7683),
    [anon_sym_n] = ACTIONS(7685),
    [anon_sym_un] = ACTIONS(7687),
    [anon_sym_L] = ACTIONS(7689),
    [aux_sym_uint64_token1] = ACTIONS(7691),
    [anon_sym_lf] = ACTIONS(8010),
    [anon_sym_LF] = ACTIONS(8012),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__untyped_quoted_close] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3176)] = {
    [sym_block_comment] = STATE(3176),
    [sym_compiler_directive_decl] = STATE(3176),
    [sym_fsi_directive_decl] = STATE(3176),
    [sym_preproc_line] = STATE(3176),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [anon_sym_inherit] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [anon_sym_POUNDelse] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3177)] = {
    [sym__else_expression] = STATE(4074),
    [sym_elif_expression] = STATE(3752),
    [sym_block_comment] = STATE(3177),
    [sym_compiler_directive_decl] = STATE(3177),
    [sym_fsi_directive_decl] = STATE(3177),
    [sym_preproc_line] = STATE(3177),
    [aux_sym_if_expression_repeat1] = STATE(3544),
    [ts_builtin_sym_end] = ACTIONS(6327),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_namespace] = ACTIONS(6329),
    [anon_sym_module] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_open] = ACTIONS(6329),
    [anon_sym_type] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(7965),
    [anon_sym_elif] = ACTIONS(7967),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_exception] = ACTIONS(6329),
    [anon_sym_extern] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6327),
    [anon_sym_POUNDload] = ACTIONS(6327),
    [anon_sym_POUNDtime] = ACTIONS(6327),
    [anon_sym_POUNDI] = ACTIONS(6327),
    [anon_sym_POUNDhelp] = ACTIONS(6327),
    [anon_sym_POUNDquit] = ACTIONS(6327),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(3178)] = {
    [sym_block_comment] = STATE(3178),
    [sym_compiler_directive_decl] = STATE(3178),
    [sym_fsi_directive_decl] = STATE(3178),
    [sym_preproc_line] = STATE(3178),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_and] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__dedent] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3179)] = {
    [sym_block_comment] = STATE(3179),
    [sym_compiler_directive_decl] = STATE(3179),
    [sym_fsi_directive_decl] = STATE(3179),
    [sym_preproc_line] = STATE(3179),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [anon_sym_inherit] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [anon_sym_POUNDelse] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3180)] = {
    [sym_block_comment] = STATE(3180),
    [sym_compiler_directive_decl] = STATE(3180),
    [sym_fsi_directive_decl] = STATE(3180),
    [sym_preproc_line] = STATE(3180),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [anon_sym_inherit] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [anon_sym_POUNDelse] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(3181)] = {
    [sym_block_comment] = STATE(3181),
    [sym_compiler_directive_decl] = STATE(3181),
    [sym_fsi_directive_decl] = STATE(3181),
    [sym_preproc_line] = STATE(3181),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [anon_sym_inherit] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [anon_sym_POUNDelse] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(3182)] = {
    [sym_block_comment] = STATE(3182),
    [sym_compiler_directive_decl] = STATE(3182),
    [sym_fsi_directive_decl] = STATE(3182),
    [sym_preproc_line] = STATE(3182),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [anon_sym_inherit] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [anon_sym_POUNDelse] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3183)] = {
    [sym_block_comment] = STATE(3183),
    [sym_compiler_directive_decl] = STATE(3183),
    [sym_fsi_directive_decl] = STATE(3183),
    [sym_preproc_line] = STATE(3183),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [anon_sym_inherit] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [anon_sym_POUNDelse] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3184)] = {
    [sym_block_comment] = STATE(3184),
    [sym_compiler_directive_decl] = STATE(3184),
    [sym_fsi_directive_decl] = STATE(3184),
    [sym_preproc_line] = STATE(3184),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_inherit] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3185)] = {
    [sym_block_comment] = STATE(3185),
    [sym_compiler_directive_decl] = STATE(3185),
    [sym_fsi_directive_decl] = STATE(3185),
    [sym_preproc_line] = STATE(3185),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [anon_sym_inherit] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [anon_sym_POUNDelse] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(3186)] = {
    [sym_block_comment] = STATE(3186),
    [sym_compiler_directive_decl] = STATE(3186),
    [sym_fsi_directive_decl] = STATE(3186),
    [sym_preproc_line] = STATE(3186),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [anon_sym_inherit] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [anon_sym_POUNDelse] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3187)] = {
    [sym_block_comment] = STATE(3187),
    [sym_compiler_directive_decl] = STATE(3187),
    [sym_fsi_directive_decl] = STATE(3187),
    [sym_preproc_line] = STATE(3187),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [anon_sym_inherit] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [anon_sym_POUNDelse] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3188)] = {
    [sym_block_comment] = STATE(3188),
    [sym_compiler_directive_decl] = STATE(3188),
    [sym_fsi_directive_decl] = STATE(3188),
    [sym_preproc_line] = STATE(3188),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [anon_sym_inherit] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [anon_sym_POUNDelse] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3189)] = {
    [sym_block_comment] = STATE(3189),
    [sym_compiler_directive_decl] = STATE(3189),
    [sym_fsi_directive_decl] = STATE(3189),
    [sym_preproc_line] = STATE(3189),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_and] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3190)] = {
    [sym_block_comment] = STATE(3190),
    [sym_compiler_directive_decl] = STATE(3190),
    [sym_fsi_directive_decl] = STATE(3190),
    [sym_preproc_line] = STATE(3190),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [anon_sym_inherit] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [anon_sym_POUNDelse] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3191)] = {
    [sym_block_comment] = STATE(3191),
    [sym_compiler_directive_decl] = STATE(3191),
    [sym_fsi_directive_decl] = STATE(3191),
    [sym_preproc_line] = STATE(3191),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [anon_sym_inherit] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [anon_sym_POUNDelse] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3192)] = {
    [sym_block_comment] = STATE(3192),
    [sym_compiler_directive_decl] = STATE(3192),
    [sym_fsi_directive_decl] = STATE(3192),
    [sym_preproc_line] = STATE(3192),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [anon_sym_inherit] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [anon_sym_POUNDelse] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3193)] = {
    [sym_block_comment] = STATE(3193),
    [sym_compiler_directive_decl] = STATE(3193),
    [sym_fsi_directive_decl] = STATE(3193),
    [sym_preproc_line] = STATE(3193),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [anon_sym_inherit] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [anon_sym_POUNDelse] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3194)] = {
    [sym_block_comment] = STATE(3194),
    [sym_compiler_directive_decl] = STATE(3194),
    [sym_fsi_directive_decl] = STATE(3194),
    [sym_preproc_line] = STATE(3194),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7517),
    [anon_sym_uy] = ACTIONS(7519),
    [anon_sym_s] = ACTIONS(7521),
    [anon_sym_us] = ACTIONS(7523),
    [anon_sym_l] = ACTIONS(7525),
    [aux_sym_uint32_token1] = ACTIONS(7527),
    [anon_sym_n] = ACTIONS(7529),
    [anon_sym_un] = ACTIONS(7531),
    [anon_sym_L] = ACTIONS(7533),
    [aux_sym_uint64_token1] = ACTIONS(7535),
    [anon_sym_lf] = ACTIONS(8014),
    [anon_sym_LF] = ACTIONS(8016),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3195)] = {
    [sym_block_comment] = STATE(3195),
    [sym_compiler_directive_decl] = STATE(3195),
    [sym_fsi_directive_decl] = STATE(3195),
    [sym_preproc_line] = STATE(3195),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [anon_sym_inherit] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [anon_sym_POUNDelse] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3196)] = {
    [sym_block_comment] = STATE(3196),
    [sym_compiler_directive_decl] = STATE(3196),
    [sym_fsi_directive_decl] = STATE(3196),
    [sym_preproc_line] = STATE(3196),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [anon_sym_inherit] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [anon_sym_POUNDelse] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3197)] = {
    [sym_block_comment] = STATE(3197),
    [sym_compiler_directive_decl] = STATE(3197),
    [sym_fsi_directive_decl] = STATE(3197),
    [sym_preproc_line] = STATE(3197),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [anon_sym_inherit] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [anon_sym_POUNDelse] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3198)] = {
    [sym_block_comment] = STATE(3198),
    [sym_compiler_directive_decl] = STATE(3198),
    [sym_fsi_directive_decl] = STATE(3198),
    [sym_preproc_line] = STATE(3198),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3199)] = {
    [sym_block_comment] = STATE(3199),
    [sym_compiler_directive_decl] = STATE(3199),
    [sym_fsi_directive_decl] = STATE(3199),
    [sym_preproc_line] = STATE(3199),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [anon_sym_inherit] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [anon_sym_POUNDelse] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3200)] = {
    [sym_block_comment] = STATE(3200),
    [sym_compiler_directive_decl] = STATE(3200),
    [sym_fsi_directive_decl] = STATE(3200),
    [sym_preproc_line] = STATE(3200),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [anon_sym_inherit] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [anon_sym_POUNDelse] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3201)] = {
    [sym_block_comment] = STATE(3201),
    [sym_compiler_directive_decl] = STATE(3201),
    [sym_fsi_directive_decl] = STATE(3201),
    [sym_preproc_line] = STATE(3201),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_and] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [anon_sym_POUNDelse] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3202)] = {
    [sym_block_comment] = STATE(3202),
    [sym_compiler_directive_decl] = STATE(3202),
    [sym_fsi_directive_decl] = STATE(3202),
    [sym_preproc_line] = STATE(3202),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [anon_sym_inherit] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [anon_sym_POUNDelse] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3203)] = {
    [sym_block_comment] = STATE(3203),
    [sym_compiler_directive_decl] = STATE(3203),
    [sym_fsi_directive_decl] = STATE(3203),
    [sym_preproc_line] = STATE(3203),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8018),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_DOT_DOT] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__dedent] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3204)] = {
    [sym_block_comment] = STATE(3204),
    [sym_compiler_directive_decl] = STATE(3204),
    [sym_fsi_directive_decl] = STATE(3204),
    [sym_preproc_line] = STATE(3204),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [anon_sym_inherit] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [anon_sym_POUNDelse] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3205)] = {
    [sym_block_comment] = STATE(3205),
    [sym_compiler_directive_decl] = STATE(3205),
    [sym_fsi_directive_decl] = STATE(3205),
    [sym_preproc_line] = STATE(3205),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [anon_sym_inherit] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [anon_sym_POUNDelse] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3206)] = {
    [sym_block_comment] = STATE(3206),
    [sym_compiler_directive_decl] = STATE(3206),
    [sym_fsi_directive_decl] = STATE(3206),
    [sym_preproc_line] = STATE(3206),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_and] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [anon_sym_POUNDelse] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3207)] = {
    [sym_block_comment] = STATE(3207),
    [sym_compiler_directive_decl] = STATE(3207),
    [sym_fsi_directive_decl] = STATE(3207),
    [sym_preproc_line] = STATE(3207),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [anon_sym_inherit] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [anon_sym_POUNDelse] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3208)] = {
    [sym_block_comment] = STATE(3208),
    [sym_compiler_directive_decl] = STATE(3208),
    [sym_fsi_directive_decl] = STATE(3208),
    [sym_preproc_line] = STATE(3208),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [anon_sym_inherit] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [anon_sym_POUNDelse] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3209)] = {
    [sym_block_comment] = STATE(3209),
    [sym_compiler_directive_decl] = STATE(3209),
    [sym_fsi_directive_decl] = STATE(3209),
    [sym_preproc_line] = STATE(3209),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_and] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3210)] = {
    [sym_block_comment] = STATE(3210),
    [sym_compiler_directive_decl] = STATE(3210),
    [sym_fsi_directive_decl] = STATE(3210),
    [sym_preproc_line] = STATE(3210),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [anon_sym_inherit] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [anon_sym_POUNDelse] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3211)] = {
    [sym_block_comment] = STATE(3211),
    [sym_compiler_directive_decl] = STATE(3211),
    [sym_fsi_directive_decl] = STATE(3211),
    [sym_preproc_line] = STATE(3211),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [anon_sym_inherit] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [anon_sym_POUNDelse] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(3212)] = {
    [sym_block_comment] = STATE(3212),
    [sym_compiler_directive_decl] = STATE(3212),
    [sym_fsi_directive_decl] = STATE(3212),
    [sym_preproc_line] = STATE(3212),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_and] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [anon_sym_POUNDelse] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3213)] = {
    [sym_block_comment] = STATE(3213),
    [sym_compiler_directive_decl] = STATE(3213),
    [sym_fsi_directive_decl] = STATE(3213),
    [sym_preproc_line] = STATE(3213),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_and] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [anon_sym_POUNDelse] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3214)] = {
    [sym_block_comment] = STATE(3214),
    [sym_compiler_directive_decl] = STATE(3214),
    [sym_fsi_directive_decl] = STATE(3214),
    [sym_preproc_line] = STATE(3214),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_and] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [anon_sym_POUNDelse] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3215)] = {
    [sym_block_comment] = STATE(3215),
    [sym_compiler_directive_decl] = STATE(3215),
    [sym_fsi_directive_decl] = STATE(3215),
    [sym_preproc_line] = STATE(3215),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_and] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__dedent] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3216)] = {
    [sym_block_comment] = STATE(3216),
    [sym_compiler_directive_decl] = STATE(3216),
    [sym_fsi_directive_decl] = STATE(3216),
    [sym_preproc_line] = STATE(3216),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_and] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__dedent] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3217)] = {
    [sym_block_comment] = STATE(3217),
    [sym_compiler_directive_decl] = STATE(3217),
    [sym_fsi_directive_decl] = STATE(3217),
    [sym_preproc_line] = STATE(3217),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_and] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__dedent] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3218)] = {
    [sym_block_comment] = STATE(3218),
    [sym_compiler_directive_decl] = STATE(3218),
    [sym_fsi_directive_decl] = STATE(3218),
    [sym_preproc_line] = STATE(3218),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8020),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_DOT_DOT] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__dedent] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3219)] = {
    [sym_block_comment] = STATE(3219),
    [sym_compiler_directive_decl] = STATE(3219),
    [sym_fsi_directive_decl] = STATE(3219),
    [sym_preproc_line] = STATE(3219),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [anon_sym_inherit] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [anon_sym_POUNDelse] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3220)] = {
    [sym_block_comment] = STATE(3220),
    [sym_compiler_directive_decl] = STATE(3220),
    [sym_fsi_directive_decl] = STATE(3220),
    [sym_preproc_line] = STATE(3220),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [anon_sym_inherit] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3221)] = {
    [sym_block_comment] = STATE(3221),
    [sym_compiler_directive_decl] = STATE(3221),
    [sym_fsi_directive_decl] = STATE(3221),
    [sym_preproc_line] = STATE(3221),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [anon_sym_inherit] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [anon_sym_POUNDelse] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3222)] = {
    [sym_block_comment] = STATE(3222),
    [sym_compiler_directive_decl] = STATE(3222),
    [sym_fsi_directive_decl] = STATE(3222),
    [sym_preproc_line] = STATE(3222),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [anon_sym_inherit] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3223)] = {
    [sym_block_comment] = STATE(3223),
    [sym_compiler_directive_decl] = STATE(3223),
    [sym_fsi_directive_decl] = STATE(3223),
    [sym_preproc_line] = STATE(3223),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [anon_sym_inherit] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [anon_sym_POUNDelse] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3224)] = {
    [sym_block_comment] = STATE(3224),
    [sym_compiler_directive_decl] = STATE(3224),
    [sym_fsi_directive_decl] = STATE(3224),
    [sym_preproc_line] = STATE(3224),
    [aux_sym_long_identifier_repeat1] = STATE(3224),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8022),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3225)] = {
    [sym_block_comment] = STATE(3225),
    [sym_compiler_directive_decl] = STATE(3225),
    [sym_fsi_directive_decl] = STATE(3225),
    [sym_preproc_line] = STATE(3225),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [anon_sym_inherit] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [anon_sym_POUNDelse] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3226)] = {
    [sym_block_comment] = STATE(3226),
    [sym_compiler_directive_decl] = STATE(3226),
    [sym_fsi_directive_decl] = STATE(3226),
    [sym_preproc_line] = STATE(3226),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_and] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [anon_sym_POUNDelse] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3227)] = {
    [sym_block_comment] = STATE(3227),
    [sym_compiler_directive_decl] = STATE(3227),
    [sym_fsi_directive_decl] = STATE(3227),
    [sym_preproc_line] = STATE(3227),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [aux_sym_access_modifier_token1] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_default] = ACTIONS(6871),
    [anon_sym_static] = ACTIONS(6871),
    [anon_sym_member] = ACTIONS(6871),
    [anon_sym_interface] = ACTIONS(6871),
    [anon_sym_abstract] = ACTIONS(6871),
    [anon_sym_val] = ACTIONS(6871),
    [anon_sym_override] = ACTIONS(6871),
    [anon_sym_inherit] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3228)] = {
    [sym_block_comment] = STATE(3228),
    [sym_compiler_directive_decl] = STATE(3228),
    [sym_fsi_directive_decl] = STATE(3228),
    [sym_preproc_line] = STATE(3228),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [anon_sym_inherit] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3229)] = {
    [sym_block_comment] = STATE(3229),
    [sym_compiler_directive_decl] = STATE(3229),
    [sym_fsi_directive_decl] = STATE(3229),
    [sym_preproc_line] = STATE(3229),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_and] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [aux_sym_access_modifier_token1] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_default] = ACTIONS(6875),
    [anon_sym_static] = ACTIONS(6875),
    [anon_sym_member] = ACTIONS(6875),
    [anon_sym_interface] = ACTIONS(6875),
    [anon_sym_abstract] = ACTIONS(6875),
    [anon_sym_val] = ACTIONS(6875),
    [anon_sym_override] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3230)] = {
    [sym_block_comment] = STATE(3230),
    [sym_compiler_directive_decl] = STATE(3230),
    [sym_fsi_directive_decl] = STATE(3230),
    [sym_preproc_line] = STATE(3230),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [aux_sym_access_modifier_token1] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_default] = ACTIONS(6828),
    [anon_sym_static] = ACTIONS(6828),
    [anon_sym_member] = ACTIONS(6828),
    [anon_sym_interface] = ACTIONS(6828),
    [anon_sym_abstract] = ACTIONS(6828),
    [anon_sym_val] = ACTIONS(6828),
    [anon_sym_override] = ACTIONS(6828),
    [anon_sym_inherit] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3231)] = {
    [sym_block_comment] = STATE(3231),
    [sym_compiler_directive_decl] = STATE(3231),
    [sym_fsi_directive_decl] = STATE(3231),
    [sym_preproc_line] = STATE(3231),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7612),
    [anon_sym_uy] = ACTIONS(7614),
    [anon_sym_s] = ACTIONS(7616),
    [anon_sym_us] = ACTIONS(7618),
    [anon_sym_l] = ACTIONS(7620),
    [aux_sym_uint32_token1] = ACTIONS(7622),
    [anon_sym_n] = ACTIONS(7624),
    [anon_sym_un] = ACTIONS(7626),
    [anon_sym_L] = ACTIONS(7628),
    [aux_sym_uint64_token1] = ACTIONS(7630),
    [anon_sym_lf] = ACTIONS(8025),
    [anon_sym_LF] = ACTIONS(8027),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__multi_dollar_interp_end] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3232)] = {
    [sym_block_comment] = STATE(3232),
    [sym_compiler_directive_decl] = STATE(3232),
    [sym_fsi_directive_decl] = STATE(3232),
    [sym_preproc_line] = STATE(3232),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_and] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [anon_sym_POUNDelse] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3233)] = {
    [sym_block_comment] = STATE(3233),
    [sym_compiler_directive_decl] = STATE(3233),
    [sym_fsi_directive_decl] = STATE(3233),
    [sym_preproc_line] = STATE(3233),
    [aux_sym_long_identifier_repeat1] = STATE(3151),
    [ts_builtin_sym_end] = ACTIONS(6291),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_namespace] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8006),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(3234)] = {
    [sym_block_comment] = STATE(3234),
    [sym_compiler_directive_decl] = STATE(3234),
    [sym_fsi_directive_decl] = STATE(3234),
    [sym_preproc_line] = STATE(3234),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3235)] = {
    [sym_block_comment] = STATE(3235),
    [sym_compiler_directive_decl] = STATE(3235),
    [sym_fsi_directive_decl] = STATE(3235),
    [sym_preproc_line] = STATE(3235),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_and] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [anon_sym_POUNDelse] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3236)] = {
    [sym_block_comment] = STATE(3236),
    [sym_compiler_directive_decl] = STATE(3236),
    [sym_fsi_directive_decl] = STATE(3236),
    [sym_preproc_line] = STATE(3236),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_and] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [anon_sym_POUNDelse] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3237)] = {
    [sym_block_comment] = STATE(3237),
    [sym_compiler_directive_decl] = STATE(3237),
    [sym_fsi_directive_decl] = STATE(3237),
    [sym_preproc_line] = STATE(3237),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_and] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [anon_sym_POUNDelse] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3238)] = {
    [sym_block_comment] = STATE(3238),
    [sym_compiler_directive_decl] = STATE(3238),
    [sym_fsi_directive_decl] = STATE(3238),
    [sym_preproc_line] = STATE(3238),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_and] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [anon_sym_POUNDelse] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3239)] = {
    [sym_block_comment] = STATE(3239),
    [sym_compiler_directive_decl] = STATE(3239),
    [sym_fsi_directive_decl] = STATE(3239),
    [sym_preproc_line] = STATE(3239),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_and] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [anon_sym_POUNDelse] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3240)] = {
    [sym_block_comment] = STATE(3240),
    [sym_compiler_directive_decl] = STATE(3240),
    [sym_fsi_directive_decl] = STATE(3240),
    [sym_preproc_line] = STATE(3240),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_and] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [aux_sym_access_modifier_token1] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_default] = ACTIONS(6692),
    [anon_sym_static] = ACTIONS(6692),
    [anon_sym_member] = ACTIONS(6692),
    [anon_sym_interface] = ACTIONS(6692),
    [anon_sym_abstract] = ACTIONS(6692),
    [anon_sym_val] = ACTIONS(6692),
    [anon_sym_override] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__dedent] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3241)] = {
    [sym_block_comment] = STATE(3241),
    [sym_compiler_directive_decl] = STATE(3241),
    [sym_fsi_directive_decl] = STATE(3241),
    [sym_preproc_line] = STATE(3241),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_and] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [anon_sym_POUNDelse] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3242)] = {
    [sym_block_comment] = STATE(3242),
    [sym_compiler_directive_decl] = STATE(3242),
    [sym_fsi_directive_decl] = STATE(3242),
    [sym_preproc_line] = STATE(3242),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_and] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [aux_sym_access_modifier_token1] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8029),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_default] = ACTIONS(6768),
    [anon_sym_static] = ACTIONS(6768),
    [anon_sym_member] = ACTIONS(6768),
    [anon_sym_interface] = ACTIONS(6768),
    [anon_sym_abstract] = ACTIONS(6768),
    [anon_sym_val] = ACTIONS(6768),
    [anon_sym_override] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3243)] = {
    [sym_block_comment] = STATE(3243),
    [sym_compiler_directive_decl] = STATE(3243),
    [sym_fsi_directive_decl] = STATE(3243),
    [sym_preproc_line] = STATE(3243),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_and] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [anon_sym_POUNDelse] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3244)] = {
    [sym_block_comment] = STATE(3244),
    [sym_compiler_directive_decl] = STATE(3244),
    [sym_fsi_directive_decl] = STATE(3244),
    [sym_preproc_line] = STATE(3244),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_and] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [anon_sym_POUNDelse] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3245)] = {
    [sym_block_comment] = STATE(3245),
    [sym_compiler_directive_decl] = STATE(3245),
    [sym_fsi_directive_decl] = STATE(3245),
    [sym_preproc_line] = STATE(3245),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_and] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [anon_sym_POUNDelse] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3246)] = {
    [sym_block_comment] = STATE(3246),
    [sym_compiler_directive_decl] = STATE(3246),
    [sym_fsi_directive_decl] = STATE(3246),
    [sym_preproc_line] = STATE(3246),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7806),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3247)] = {
    [sym_block_comment] = STATE(3247),
    [sym_compiler_directive_decl] = STATE(3247),
    [sym_fsi_directive_decl] = STATE(3247),
    [sym_preproc_line] = STATE(3247),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_and] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [anon_sym_POUNDelse] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(3248)] = {
    [sym_block_comment] = STATE(3248),
    [sym_compiler_directive_decl] = STATE(3248),
    [sym_fsi_directive_decl] = STATE(3248),
    [sym_preproc_line] = STATE(3248),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_and] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [anon_sym_POUNDelse] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3249)] = {
    [sym_block_comment] = STATE(3249),
    [sym_compiler_directive_decl] = STATE(3249),
    [sym_fsi_directive_decl] = STATE(3249),
    [sym_preproc_line] = STATE(3249),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_and] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [anon_sym_POUNDelse] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3250)] = {
    [sym_block_comment] = STATE(3250),
    [sym_compiler_directive_decl] = STATE(3250),
    [sym_fsi_directive_decl] = STATE(3250),
    [sym_preproc_line] = STATE(3250),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [anon_sym_inherit] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3251)] = {
    [sym_block_comment] = STATE(3251),
    [sym_compiler_directive_decl] = STATE(3251),
    [sym_fsi_directive_decl] = STATE(3251),
    [sym_preproc_line] = STATE(3251),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_and] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [anon_sym_POUNDelse] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3252)] = {
    [sym_block_comment] = STATE(3252),
    [sym_compiler_directive_decl] = STATE(3252),
    [sym_fsi_directive_decl] = STATE(3252),
    [sym_preproc_line] = STATE(3252),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_and] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [aux_sym_access_modifier_token1] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_default] = ACTIONS(6863),
    [anon_sym_static] = ACTIONS(6863),
    [anon_sym_member] = ACTIONS(6863),
    [anon_sym_interface] = ACTIONS(6863),
    [anon_sym_abstract] = ACTIONS(6863),
    [anon_sym_val] = ACTIONS(6863),
    [anon_sym_override] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3253)] = {
    [sym_block_comment] = STATE(3253),
    [sym_compiler_directive_decl] = STATE(3253),
    [sym_fsi_directive_decl] = STATE(3253),
    [sym_preproc_line] = STATE(3253),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_and] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [anon_sym_POUNDelse] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3254)] = {
    [sym_block_comment] = STATE(3254),
    [sym_compiler_directive_decl] = STATE(3254),
    [sym_fsi_directive_decl] = STATE(3254),
    [sym_preproc_line] = STATE(3254),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [anon_sym_inherit] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [anon_sym_POUNDelse] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3255)] = {
    [sym_block_comment] = STATE(3255),
    [sym_compiler_directive_decl] = STATE(3255),
    [sym_fsi_directive_decl] = STATE(3255),
    [sym_preproc_line] = STATE(3255),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_and] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [anon_sym_POUNDelse] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3256)] = {
    [sym_block_comment] = STATE(3256),
    [sym_compiler_directive_decl] = STATE(3256),
    [sym_fsi_directive_decl] = STATE(3256),
    [sym_preproc_line] = STATE(3256),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_and] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3257)] = {
    [sym_block_comment] = STATE(3257),
    [sym_compiler_directive_decl] = STATE(3257),
    [sym_fsi_directive_decl] = STATE(3257),
    [sym_preproc_line] = STATE(3257),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [anon_sym_inherit] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [anon_sym_POUNDelse] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(3258)] = {
    [sym_block_comment] = STATE(3258),
    [sym_compiler_directive_decl] = STATE(3258),
    [sym_fsi_directive_decl] = STATE(3258),
    [sym_preproc_line] = STATE(3258),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_and] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [anon_sym_POUNDelse] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3259)] = {
    [sym_block_comment] = STATE(3259),
    [sym_compiler_directive_decl] = STATE(3259),
    [sym_fsi_directive_decl] = STATE(3259),
    [sym_preproc_line] = STATE(3259),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6302),
    [aux_sym_decimal_token1] = ACTIONS(6249),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3260)] = {
    [sym_block_comment] = STATE(3260),
    [sym_compiler_directive_decl] = STATE(3260),
    [sym_fsi_directive_decl] = STATE(3260),
    [sym_preproc_line] = STATE(3260),
    [aux_sym_sequential_expression_repeat1] = STATE(3260),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8031),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3261)] = {
    [sym_block_comment] = STATE(3261),
    [sym_compiler_directive_decl] = STATE(3261),
    [sym_fsi_directive_decl] = STATE(3261),
    [sym_preproc_line] = STATE(3261),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_and] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [anon_sym_POUNDelse] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(3262)] = {
    [sym_block_comment] = STATE(3262),
    [sym_compiler_directive_decl] = STATE(3262),
    [sym_fsi_directive_decl] = STATE(3262),
    [sym_preproc_line] = STATE(3262),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_and] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [anon_sym_POUNDelse] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3263)] = {
    [sym_block_comment] = STATE(3263),
    [sym_compiler_directive_decl] = STATE(3263),
    [sym_fsi_directive_decl] = STATE(3263),
    [sym_preproc_line] = STATE(3263),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_and] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(7899),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3264)] = {
    [sym_block_comment] = STATE(3264),
    [sym_compiler_directive_decl] = STATE(3264),
    [sym_fsi_directive_decl] = STATE(3264),
    [sym_preproc_line] = STATE(3264),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_and] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [anon_sym_POUNDelse] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3265)] = {
    [sym_block_comment] = STATE(3265),
    [sym_compiler_directive_decl] = STATE(3265),
    [sym_fsi_directive_decl] = STATE(3265),
    [sym_preproc_line] = STATE(3265),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_and] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [anon_sym_POUNDelse] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(3266)] = {
    [sym_block_comment] = STATE(3266),
    [sym_compiler_directive_decl] = STATE(3266),
    [sym_fsi_directive_decl] = STATE(3266),
    [sym_preproc_line] = STATE(3266),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_and] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [anon_sym_POUNDelse] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3267)] = {
    [sym_block_comment] = STATE(3267),
    [sym_compiler_directive_decl] = STATE(3267),
    [sym_fsi_directive_decl] = STATE(3267),
    [sym_preproc_line] = STATE(3267),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [anon_sym_inherit] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [anon_sym_POUNDelse] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3268)] = {
    [sym_block_comment] = STATE(3268),
    [sym_compiler_directive_decl] = STATE(3268),
    [sym_fsi_directive_decl] = STATE(3268),
    [sym_preproc_line] = STATE(3268),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3269)] = {
    [sym_block_comment] = STATE(3269),
    [sym_compiler_directive_decl] = STATE(3269),
    [sym_fsi_directive_decl] = STATE(3269),
    [sym_preproc_line] = STATE(3269),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [anon_sym_inherit] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [anon_sym_POUNDelse] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3270)] = {
    [sym_block_comment] = STATE(3270),
    [sym_compiler_directive_decl] = STATE(3270),
    [sym_fsi_directive_decl] = STATE(3270),
    [sym_preproc_line] = STATE(3270),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_and] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [anon_sym_POUNDelse] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3271)] = {
    [sym_block_comment] = STATE(3271),
    [sym_compiler_directive_decl] = STATE(3271),
    [sym_fsi_directive_decl] = STATE(3271),
    [sym_preproc_line] = STATE(3271),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_and] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [anon_sym_POUNDelse] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3272)] = {
    [sym_block_comment] = STATE(3272),
    [sym_compiler_directive_decl] = STATE(3272),
    [sym_fsi_directive_decl] = STATE(3272),
    [sym_preproc_line] = STATE(3272),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_and] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [anon_sym_POUNDelse] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3273)] = {
    [sym_block_comment] = STATE(3273),
    [sym_compiler_directive_decl] = STATE(3273),
    [sym_fsi_directive_decl] = STATE(3273),
    [sym_preproc_line] = STATE(3273),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_and] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3274)] = {
    [sym_block_comment] = STATE(3274),
    [sym_compiler_directive_decl] = STATE(3274),
    [sym_fsi_directive_decl] = STATE(3274),
    [sym_preproc_line] = STATE(3274),
    [aux_sym_sequential_expression_repeat1] = STATE(3274),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_and] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8034),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3275)] = {
    [sym_block_comment] = STATE(3275),
    [sym_compiler_directive_decl] = STATE(3275),
    [sym_fsi_directive_decl] = STATE(3275),
    [sym_preproc_line] = STATE(3275),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8037),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [anon_sym_inherit] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3276)] = {
    [sym_type_arguments] = STATE(3730),
    [sym_long_identifier] = STATE(3759),
    [sym_block_comment] = STATE(3276),
    [sym_compiler_directive_decl] = STATE(3276),
    [sym_fsi_directive_decl] = STATE(3276),
    [sym_preproc_line] = STATE(3276),
    [aux_sym_compound_type_repeat1] = STATE(3689),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_GT_RBRACK] = ACTIONS(6263),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_RBRACK] = ACTIONS(6263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_RBRACE] = ACTIONS(6263),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_to] = ACTIONS(6265),
    [anon_sym_downto] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(7981),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_DOT_DOT2] = ACTIONS(6263),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(7985),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7987),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3277)] = {
    [sym_block_comment] = STATE(3277),
    [sym_compiler_directive_decl] = STATE(3277),
    [sym_fsi_directive_decl] = STATE(3277),
    [sym_preproc_line] = STATE(3277),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3278)] = {
    [sym_block_comment] = STATE(3278),
    [sym_compiler_directive_decl] = STATE(3278),
    [sym_fsi_directive_decl] = STATE(3278),
    [sym_preproc_line] = STATE(3278),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [anon_sym_inherit] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [anon_sym_POUNDelse] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3279)] = {
    [sym_block_comment] = STATE(3279),
    [sym_compiler_directive_decl] = STATE(3279),
    [sym_fsi_directive_decl] = STATE(3279),
    [sym_preproc_line] = STATE(3279),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_and] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [anon_sym_POUNDelse] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3280)] = {
    [sym_block_comment] = STATE(3280),
    [sym_compiler_directive_decl] = STATE(3280),
    [sym_fsi_directive_decl] = STATE(3280),
    [sym_preproc_line] = STATE(3280),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_and] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [anon_sym_POUNDelse] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3281)] = {
    [sym_block_comment] = STATE(3281),
    [sym_compiler_directive_decl] = STATE(3281),
    [sym_fsi_directive_decl] = STATE(3281),
    [sym_preproc_line] = STATE(3281),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_and] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [anon_sym_POUNDelse] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3282)] = {
    [sym_block_comment] = STATE(3282),
    [sym_compiler_directive_decl] = STATE(3282),
    [sym_fsi_directive_decl] = STATE(3282),
    [sym_preproc_line] = STATE(3282),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [anon_sym_inherit] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [anon_sym_POUNDelse] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3283)] = {
    [sym_block_comment] = STATE(3283),
    [sym_compiler_directive_decl] = STATE(3283),
    [sym_fsi_directive_decl] = STATE(3283),
    [sym_preproc_line] = STATE(3283),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [anon_sym_inherit] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3284)] = {
    [sym_block_comment] = STATE(3284),
    [sym_compiler_directive_decl] = STATE(3284),
    [sym_fsi_directive_decl] = STATE(3284),
    [sym_preproc_line] = STATE(3284),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [anon_sym_inherit] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3285)] = {
    [sym_block_comment] = STATE(3285),
    [sym_compiler_directive_decl] = STATE(3285),
    [sym_fsi_directive_decl] = STATE(3285),
    [sym_preproc_line] = STATE(3285),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_and] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [anon_sym_POUNDelse] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3286)] = {
    [sym_block_comment] = STATE(3286),
    [sym_compiler_directive_decl] = STATE(3286),
    [sym_fsi_directive_decl] = STATE(3286),
    [sym_preproc_line] = STATE(3286),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_and] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [anon_sym_POUNDelse] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3287)] = {
    [sym_block_comment] = STATE(3287),
    [sym_compiler_directive_decl] = STATE(3287),
    [sym_fsi_directive_decl] = STATE(3287),
    [sym_preproc_line] = STATE(3287),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [anon_sym_inherit] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [anon_sym_POUNDelse] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3288)] = {
    [sym_block_comment] = STATE(3288),
    [sym_compiler_directive_decl] = STATE(3288),
    [sym_fsi_directive_decl] = STATE(3288),
    [sym_preproc_line] = STATE(3288),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_and] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [anon_sym_POUNDelse] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3289)] = {
    [sym_block_comment] = STATE(3289),
    [sym_compiler_directive_decl] = STATE(3289),
    [sym_fsi_directive_decl] = STATE(3289),
    [sym_preproc_line] = STATE(3289),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [aux_sym_access_modifier_token1] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8004),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_default] = ACTIONS(6597),
    [anon_sym_static] = ACTIONS(6597),
    [anon_sym_member] = ACTIONS(6597),
    [anon_sym_interface] = ACTIONS(6597),
    [anon_sym_abstract] = ACTIONS(6597),
    [anon_sym_val] = ACTIONS(6597),
    [anon_sym_override] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3290)] = {
    [sym_block_comment] = STATE(3290),
    [sym_compiler_directive_decl] = STATE(3290),
    [sym_fsi_directive_decl] = STATE(3290),
    [sym_preproc_line] = STATE(3290),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [anon_sym_inherit] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3291)] = {
    [sym_block_comment] = STATE(3291),
    [sym_compiler_directive_decl] = STATE(3291),
    [sym_fsi_directive_decl] = STATE(3291),
    [sym_preproc_line] = STATE(3291),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [anon_sym_inherit] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [anon_sym_POUNDelse] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3292)] = {
    [sym_block_comment] = STATE(3292),
    [sym_compiler_directive_decl] = STATE(3292),
    [sym_fsi_directive_decl] = STATE(3292),
    [sym_preproc_line] = STATE(3292),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_and] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [anon_sym_POUNDelse] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(3293)] = {
    [sym_block_comment] = STATE(3293),
    [sym_compiler_directive_decl] = STATE(3293),
    [sym_fsi_directive_decl] = STATE(3293),
    [sym_preproc_line] = STATE(3293),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [aux_sym_access_modifier_token1] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_default] = ACTIONS(6560),
    [anon_sym_static] = ACTIONS(6560),
    [anon_sym_member] = ACTIONS(6560),
    [anon_sym_interface] = ACTIONS(6560),
    [anon_sym_abstract] = ACTIONS(6560),
    [anon_sym_val] = ACTIONS(6560),
    [anon_sym_override] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3294)] = {
    [sym_block_comment] = STATE(3294),
    [sym_compiler_directive_decl] = STATE(3294),
    [sym_fsi_directive_decl] = STATE(3294),
    [sym_preproc_line] = STATE(3294),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_and] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [anon_sym_POUNDelse] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3295)] = {
    [sym_block_comment] = STATE(3295),
    [sym_compiler_directive_decl] = STATE(3295),
    [sym_fsi_directive_decl] = STATE(3295),
    [sym_preproc_line] = STATE(3295),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7701),
    [anon_sym_uy] = ACTIONS(7703),
    [anon_sym_s] = ACTIONS(7705),
    [anon_sym_us] = ACTIONS(7707),
    [anon_sym_l] = ACTIONS(7709),
    [aux_sym_uint32_token1] = ACTIONS(7711),
    [anon_sym_n] = ACTIONS(7713),
    [anon_sym_un] = ACTIONS(7715),
    [anon_sym_L] = ACTIONS(7717),
    [aux_sym_uint64_token1] = ACTIONS(7719),
    [anon_sym_lf] = ACTIONS(8039),
    [anon_sym_LF] = ACTIONS(8041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3296)] = {
    [sym_block_comment] = STATE(3296),
    [sym_compiler_directive_decl] = STATE(3296),
    [sym_fsi_directive_decl] = STATE(3296),
    [sym_preproc_line] = STATE(3296),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [anon_sym_inherit] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3297)] = {
    [sym_block_comment] = STATE(3297),
    [sym_compiler_directive_decl] = STATE(3297),
    [sym_fsi_directive_decl] = STATE(3297),
    [sym_preproc_line] = STATE(3297),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [aux_sym_access_modifier_token1] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_default] = ACTIONS(6612),
    [anon_sym_static] = ACTIONS(6612),
    [anon_sym_member] = ACTIONS(6612),
    [anon_sym_interface] = ACTIONS(6612),
    [anon_sym_abstract] = ACTIONS(6612),
    [anon_sym_val] = ACTIONS(6612),
    [anon_sym_override] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3298)] = {
    [sym_block_comment] = STATE(3298),
    [sym_compiler_directive_decl] = STATE(3298),
    [sym_fsi_directive_decl] = STATE(3298),
    [sym_preproc_line] = STATE(3298),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3299)] = {
    [sym_block_comment] = STATE(3299),
    [sym_compiler_directive_decl] = STATE(3299),
    [sym_fsi_directive_decl] = STATE(3299),
    [sym_preproc_line] = STATE(3299),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3300)] = {
    [sym_block_comment] = STATE(3300),
    [sym_compiler_directive_decl] = STATE(3300),
    [sym_fsi_directive_decl] = STATE(3300),
    [sym_preproc_line] = STATE(3300),
    [aux_sym_type_argument_repeat1] = STATE(3327),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8045),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3301)] = {
    [sym_block_comment] = STATE(3301),
    [sym_compiler_directive_decl] = STATE(3301),
    [sym_fsi_directive_decl] = STATE(3301),
    [sym_preproc_line] = STATE(3301),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [aux_sym_access_modifier_token1] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_DOT_DOT] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_default] = ACTIONS(6840),
    [anon_sym_static] = ACTIONS(6840),
    [anon_sym_member] = ACTIONS(6840),
    [anon_sym_interface] = ACTIONS(6840),
    [anon_sym_abstract] = ACTIONS(6840),
    [anon_sym_val] = ACTIONS(6840),
    [anon_sym_override] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__dedent] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3302)] = {
    [sym_block_comment] = STATE(3302),
    [sym_compiler_directive_decl] = STATE(3302),
    [sym_fsi_directive_decl] = STATE(3302),
    [sym_preproc_line] = STATE(3302),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_and] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [anon_sym_POUNDelse] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(3303)] = {
    [sym_block_comment] = STATE(3303),
    [sym_compiler_directive_decl] = STATE(3303),
    [sym_fsi_directive_decl] = STATE(3303),
    [sym_preproc_line] = STATE(3303),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3304)] = {
    [sym_block_comment] = STATE(3304),
    [sym_compiler_directive_decl] = STATE(3304),
    [sym_fsi_directive_decl] = STATE(3304),
    [sym_preproc_line] = STATE(3304),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [aux_sym_access_modifier_token1] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_default] = ACTIONS(6637),
    [anon_sym_static] = ACTIONS(6637),
    [anon_sym_member] = ACTIONS(6637),
    [anon_sym_interface] = ACTIONS(6637),
    [anon_sym_abstract] = ACTIONS(6637),
    [anon_sym_val] = ACTIONS(6637),
    [anon_sym_override] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(8048),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3305)] = {
    [sym_block_comment] = STATE(3305),
    [sym_compiler_directive_decl] = STATE(3305),
    [sym_fsi_directive_decl] = STATE(3305),
    [sym_preproc_line] = STATE(3305),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [anon_sym_inherit] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [anon_sym_POUNDelse] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3306)] = {
    [sym_block_comment] = STATE(3306),
    [sym_compiler_directive_decl] = STATE(3306),
    [sym_fsi_directive_decl] = STATE(3306),
    [sym_preproc_line] = STATE(3306),
    [aux_sym_compound_type_repeat1] = STATE(3308),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(7584),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [anon_sym_POUNDelse] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(3307)] = {
    [sym_block_comment] = STATE(3307),
    [sym_compiler_directive_decl] = STATE(3307),
    [sym_fsi_directive_decl] = STATE(3307),
    [sym_preproc_line] = STATE(3307),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3308)] = {
    [sym_block_comment] = STATE(3308),
    [sym_compiler_directive_decl] = STATE(3308),
    [sym_fsi_directive_decl] = STATE(3308),
    [sym_preproc_line] = STATE(3308),
    [aux_sym_compound_type_repeat1] = STATE(3308),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8050),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3309)] = {
    [sym_block_comment] = STATE(3309),
    [sym_compiler_directive_decl] = STATE(3309),
    [sym_fsi_directive_decl] = STATE(3309),
    [sym_preproc_line] = STATE(3309),
    [aux_sym_sequential_expression_repeat1] = STATE(3274),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3310)] = {
    [sym_type_arguments] = STATE(3730),
    [sym_long_identifier] = STATE(3759),
    [sym_block_comment] = STATE(3310),
    [sym_compiler_directive_decl] = STATE(3310),
    [sym_fsi_directive_decl] = STATE(3310),
    [sym_preproc_line] = STATE(3310),
    [aux_sym_compound_type_repeat1] = STATE(3689),
    [sym_identifier] = ACTIONS(7979),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_GT_RBRACK] = ACTIONS(6255),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_RBRACK] = ACTIONS(6255),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_RBRACE] = ACTIONS(6255),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_to] = ACTIONS(6257),
    [anon_sym_downto] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(7981),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_DOT_DOT2] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(7985),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7987),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3311)] = {
    [sym_block_comment] = STATE(3311),
    [sym_compiler_directive_decl] = STATE(3311),
    [sym_fsi_directive_decl] = STATE(3311),
    [sym_preproc_line] = STATE(3311),
    [aux_sym_type_argument_repeat1] = STATE(3101),
    [ts_builtin_sym_end] = ACTIONS(6306),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_namespace] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(3312)] = {
    [sym_block_comment] = STATE(3312),
    [sym_compiler_directive_decl] = STATE(3312),
    [sym_fsi_directive_decl] = STATE(3312),
    [sym_preproc_line] = STATE(3312),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [aux_sym_access_modifier_token1] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_default] = ACTIONS(6653),
    [anon_sym_static] = ACTIONS(6653),
    [anon_sym_member] = ACTIONS(6653),
    [anon_sym_interface] = ACTIONS(6653),
    [anon_sym_abstract] = ACTIONS(6653),
    [anon_sym_val] = ACTIONS(6653),
    [anon_sym_override] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3313)] = {
    [sym_block_comment] = STATE(3313),
    [sym_compiler_directive_decl] = STATE(3313),
    [sym_fsi_directive_decl] = STATE(3313),
    [sym_preproc_line] = STATE(3313),
    [aux_sym_sequential_expression_repeat1] = STATE(3260),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_and] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3314)] = {
    [sym_block_comment] = STATE(3314),
    [sym_compiler_directive_decl] = STATE(3314),
    [sym_fsi_directive_decl] = STATE(3314),
    [sym_preproc_line] = STATE(3314),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3315)] = {
    [sym_block_comment] = STATE(3315),
    [sym_compiler_directive_decl] = STATE(3315),
    [sym_fsi_directive_decl] = STATE(3315),
    [sym_preproc_line] = STATE(3315),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_and] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [anon_sym_POUNDelse] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3316)] = {
    [sym_block_comment] = STATE(3316),
    [sym_compiler_directive_decl] = STATE(3316),
    [sym_fsi_directive_decl] = STATE(3316),
    [sym_preproc_line] = STATE(3316),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_and] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [anon_sym_POUNDelse] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3317)] = {
    [sym_block_comment] = STATE(3317),
    [sym_compiler_directive_decl] = STATE(3317),
    [sym_fsi_directive_decl] = STATE(3317),
    [sym_preproc_line] = STATE(3317),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(7761),
    [anon_sym_uy] = ACTIONS(7763),
    [anon_sym_s] = ACTIONS(7765),
    [anon_sym_us] = ACTIONS(7767),
    [anon_sym_l] = ACTIONS(7769),
    [aux_sym_uint32_token1] = ACTIONS(7771),
    [anon_sym_n] = ACTIONS(7773),
    [anon_sym_un] = ACTIONS(7775),
    [anon_sym_L] = ACTIONS(7777),
    [aux_sym_uint64_token1] = ACTIONS(7779),
    [anon_sym_lf] = ACTIONS(8053),
    [anon_sym_LF] = ACTIONS(8055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3318)] = {
    [sym_block_comment] = STATE(3318),
    [sym_compiler_directive_decl] = STATE(3318),
    [sym_fsi_directive_decl] = STATE(3318),
    [sym_preproc_line] = STATE(3318),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_and] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3319)] = {
    [sym_block_comment] = STATE(3319),
    [sym_compiler_directive_decl] = STATE(3319),
    [sym_fsi_directive_decl] = STATE(3319),
    [sym_preproc_line] = STATE(3319),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_and] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [anon_sym_POUNDelse] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3320)] = {
    [sym_block_comment] = STATE(3320),
    [sym_compiler_directive_decl] = STATE(3320),
    [sym_fsi_directive_decl] = STATE(3320),
    [sym_preproc_line] = STATE(3320),
    [aux_sym_sequential_expression_repeat1] = STATE(3320),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [anon_sym_inherit] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3321)] = {
    [sym_block_comment] = STATE(3321),
    [sym_compiler_directive_decl] = STATE(3321),
    [sym_fsi_directive_decl] = STATE(3321),
    [sym_preproc_line] = STATE(3321),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_and] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8060),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3322)] = {
    [sym_block_comment] = STATE(3322),
    [sym_compiler_directive_decl] = STATE(3322),
    [sym_fsi_directive_decl] = STATE(3322),
    [sym_preproc_line] = STATE(3322),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_and] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [anon_sym_POUNDelse] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3323)] = {
    [sym_block_comment] = STATE(3323),
    [sym_compiler_directive_decl] = STATE(3323),
    [sym_fsi_directive_decl] = STATE(3323),
    [sym_preproc_line] = STATE(3323),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_and] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3324)] = {
    [sym_block_comment] = STATE(3324),
    [sym_compiler_directive_decl] = STATE(3324),
    [sym_fsi_directive_decl] = STATE(3324),
    [sym_preproc_line] = STATE(3324),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_and] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3325)] = {
    [sym_block_comment] = STATE(3325),
    [sym_compiler_directive_decl] = STATE(3325),
    [sym_fsi_directive_decl] = STATE(3325),
    [sym_preproc_line] = STATE(3325),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_and] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [anon_sym_POUNDelse] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3326)] = {
    [sym_block_comment] = STATE(3326),
    [sym_compiler_directive_decl] = STATE(3326),
    [sym_fsi_directive_decl] = STATE(3326),
    [sym_preproc_line] = STATE(3326),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [aux_sym_access_modifier_token1] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_DOT_DOT] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_default] = ACTIONS(6813),
    [anon_sym_static] = ACTIONS(6813),
    [anon_sym_member] = ACTIONS(6813),
    [anon_sym_interface] = ACTIONS(6813),
    [anon_sym_abstract] = ACTIONS(6813),
    [anon_sym_val] = ACTIONS(6813),
    [anon_sym_override] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__dedent] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3327)] = {
    [sym_block_comment] = STATE(3327),
    [sym_compiler_directive_decl] = STATE(3327),
    [sym_fsi_directive_decl] = STATE(3327),
    [sym_preproc_line] = STATE(3327),
    [aux_sym_type_argument_repeat1] = STATE(3330),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(3328)] = {
    [sym_block_comment] = STATE(3328),
    [sym_compiler_directive_decl] = STATE(3328),
    [sym_fsi_directive_decl] = STATE(3328),
    [sym_preproc_line] = STATE(3328),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_and] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [anon_sym_POUNDelse] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3329)] = {
    [sym_block_comment] = STATE(3329),
    [sym_compiler_directive_decl] = STATE(3329),
    [sym_fsi_directive_decl] = STATE(3329),
    [sym_preproc_line] = STATE(3329),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_and] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [anon_sym_POUNDelse] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3330)] = {
    [sym_block_comment] = STATE(3330),
    [sym_compiler_directive_decl] = STATE(3330),
    [sym_fsi_directive_decl] = STATE(3330),
    [sym_preproc_line] = STATE(3330),
    [aux_sym_type_argument_repeat1] = STATE(3330),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8062),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3331)] = {
    [sym_block_comment] = STATE(3331),
    [sym_compiler_directive_decl] = STATE(3331),
    [sym_fsi_directive_decl] = STATE(3331),
    [sym_preproc_line] = STATE(3331),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_and] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [anon_sym_POUNDelse] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3332)] = {
    [sym_block_comment] = STATE(3332),
    [sym_compiler_directive_decl] = STATE(3332),
    [sym_fsi_directive_decl] = STATE(3332),
    [sym_preproc_line] = STATE(3332),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_and] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__dedent] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3333)] = {
    [sym_block_comment] = STATE(3333),
    [sym_compiler_directive_decl] = STATE(3333),
    [sym_fsi_directive_decl] = STATE(3333),
    [sym_preproc_line] = STATE(3333),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_and] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [anon_sym_POUNDelse] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3334)] = {
    [sym_block_comment] = STATE(3334),
    [sym_compiler_directive_decl] = STATE(3334),
    [sym_fsi_directive_decl] = STATE(3334),
    [sym_preproc_line] = STATE(3334),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [anon_sym_inherit] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3335)] = {
    [sym_block_comment] = STATE(3335),
    [sym_compiler_directive_decl] = STATE(3335),
    [sym_fsi_directive_decl] = STATE(3335),
    [sym_preproc_line] = STATE(3335),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_and] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [anon_sym_POUNDelse] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3336)] = {
    [sym_block_comment] = STATE(3336),
    [sym_compiler_directive_decl] = STATE(3336),
    [sym_fsi_directive_decl] = STATE(3336),
    [sym_preproc_line] = STATE(3336),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_and] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [anon_sym_POUNDelse] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3337)] = {
    [sym_block_comment] = STATE(3337),
    [sym_compiler_directive_decl] = STATE(3337),
    [sym_fsi_directive_decl] = STATE(3337),
    [sym_preproc_line] = STATE(3337),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_and] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [aux_sym_access_modifier_token1] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8065),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_default] = ACTIONS(6784),
    [anon_sym_static] = ACTIONS(6784),
    [anon_sym_member] = ACTIONS(6784),
    [anon_sym_interface] = ACTIONS(6784),
    [anon_sym_abstract] = ACTIONS(6784),
    [anon_sym_val] = ACTIONS(6784),
    [anon_sym_override] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__dedent] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3338)] = {
    [sym_block_comment] = STATE(3338),
    [sym_compiler_directive_decl] = STATE(3338),
    [sym_fsi_directive_decl] = STATE(3338),
    [sym_preproc_line] = STATE(3338),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_and] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [anon_sym_POUNDelse] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3339)] = {
    [sym_block_comment] = STATE(3339),
    [sym_compiler_directive_decl] = STATE(3339),
    [sym_fsi_directive_decl] = STATE(3339),
    [sym_preproc_line] = STATE(3339),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3340)] = {
    [sym_block_comment] = STATE(3340),
    [sym_compiler_directive_decl] = STATE(3340),
    [sym_fsi_directive_decl] = STATE(3340),
    [sym_preproc_line] = STATE(3340),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [aux_sym_access_modifier_token1] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_DOT_DOT] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_default] = ACTIONS(6750),
    [anon_sym_static] = ACTIONS(6750),
    [anon_sym_member] = ACTIONS(6750),
    [anon_sym_interface] = ACTIONS(6750),
    [anon_sym_abstract] = ACTIONS(6750),
    [anon_sym_val] = ACTIONS(6750),
    [anon_sym_override] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__dedent] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3341)] = {
    [sym_block_comment] = STATE(3341),
    [sym_compiler_directive_decl] = STATE(3341),
    [sym_fsi_directive_decl] = STATE(3341),
    [sym_preproc_line] = STATE(3341),
    [aux_sym_long_identifier_repeat1] = STATE(3086),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(7958),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(3342)] = {
    [sym_block_comment] = STATE(3342),
    [sym_compiler_directive_decl] = STATE(3342),
    [sym_fsi_directive_decl] = STATE(3342),
    [sym_preproc_line] = STATE(3342),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [aux_sym_access_modifier_token1] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_default] = ACTIONS(6794),
    [anon_sym_static] = ACTIONS(6794),
    [anon_sym_member] = ACTIONS(6794),
    [anon_sym_interface] = ACTIONS(6794),
    [anon_sym_abstract] = ACTIONS(6794),
    [anon_sym_val] = ACTIONS(6794),
    [anon_sym_override] = ACTIONS(6794),
    [anon_sym_inherit] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3343)] = {
    [sym_block_comment] = STATE(3343),
    [sym_compiler_directive_decl] = STATE(3343),
    [sym_fsi_directive_decl] = STATE(3343),
    [sym_preproc_line] = STATE(3343),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_and] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [aux_sym_access_modifier_token1] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_default] = ACTIONS(6867),
    [anon_sym_static] = ACTIONS(6867),
    [anon_sym_member] = ACTIONS(6867),
    [anon_sym_interface] = ACTIONS(6867),
    [anon_sym_abstract] = ACTIONS(6867),
    [anon_sym_val] = ACTIONS(6867),
    [anon_sym_override] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3344)] = {
    [sym_block_comment] = STATE(3344),
    [sym_compiler_directive_decl] = STATE(3344),
    [sym_fsi_directive_decl] = STATE(3344),
    [sym_preproc_line] = STATE(3344),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [aux_sym_access_modifier_token1] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_DOT_DOT] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_default] = ACTIONS(6824),
    [anon_sym_static] = ACTIONS(6824),
    [anon_sym_member] = ACTIONS(6824),
    [anon_sym_interface] = ACTIONS(6824),
    [anon_sym_abstract] = ACTIONS(6824),
    [anon_sym_val] = ACTIONS(6824),
    [anon_sym_override] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__dedent] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3345)] = {
    [sym_block_comment] = STATE(3345),
    [sym_compiler_directive_decl] = STATE(3345),
    [sym_fsi_directive_decl] = STATE(3345),
    [sym_preproc_line] = STATE(3345),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [aux_sym_access_modifier_token1] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_default] = ACTIONS(6513),
    [anon_sym_static] = ACTIONS(6513),
    [anon_sym_member] = ACTIONS(6513),
    [anon_sym_interface] = ACTIONS(6513),
    [anon_sym_abstract] = ACTIONS(6513),
    [anon_sym_val] = ACTIONS(6513),
    [anon_sym_override] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3346)] = {
    [sym_block_comment] = STATE(3346),
    [sym_compiler_directive_decl] = STATE(3346),
    [sym_fsi_directive_decl] = STATE(3346),
    [sym_preproc_line] = STATE(3346),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_and] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [aux_sym_access_modifier_token1] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_default] = ACTIONS(6834),
    [anon_sym_static] = ACTIONS(6834),
    [anon_sym_member] = ACTIONS(6834),
    [anon_sym_interface] = ACTIONS(6834),
    [anon_sym_abstract] = ACTIONS(6834),
    [anon_sym_val] = ACTIONS(6834),
    [anon_sym_override] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3347)] = {
    [sym_block_comment] = STATE(3347),
    [sym_compiler_directive_decl] = STATE(3347),
    [sym_fsi_directive_decl] = STATE(3347),
    [sym_preproc_line] = STATE(3347),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_and] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [aux_sym_access_modifier_token1] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_default] = ACTIONS(6790),
    [anon_sym_static] = ACTIONS(6790),
    [anon_sym_member] = ACTIONS(6790),
    [anon_sym_interface] = ACTIONS(6790),
    [anon_sym_abstract] = ACTIONS(6790),
    [anon_sym_val] = ACTIONS(6790),
    [anon_sym_override] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__dedent] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3348)] = {
    [sym_block_comment] = STATE(3348),
    [sym_compiler_directive_decl] = STATE(3348),
    [sym_fsi_directive_decl] = STATE(3348),
    [sym_preproc_line] = STATE(3348),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [anon_sym_inherit] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [anon_sym_POUNDelse] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(3349)] = {
    [sym_block_comment] = STATE(3349),
    [sym_compiler_directive_decl] = STATE(3349),
    [sym_fsi_directive_decl] = STATE(3349),
    [sym_preproc_line] = STATE(3349),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_module] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_open] = ACTIONS(6407),
    [anon_sym_type] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_exception] = ACTIONS(6407),
    [anon_sym_extern] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6409),
    [anon_sym_POUNDload] = ACTIONS(6409),
    [anon_sym_POUNDtime] = ACTIONS(6409),
    [anon_sym_POUNDI] = ACTIONS(6409),
    [anon_sym_POUNDhelp] = ACTIONS(6409),
    [anon_sym_POUNDquit] = ACTIONS(6409),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [anon_sym_POUNDelse] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(3350)] = {
    [sym_block_comment] = STATE(3350),
    [sym_compiler_directive_decl] = STATE(3350),
    [sym_fsi_directive_decl] = STATE(3350),
    [sym_preproc_line] = STATE(3350),
    [ts_builtin_sym_end] = ACTIONS(6381),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_namespace] = ACTIONS(6379),
    [anon_sym_module] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_open] = ACTIONS(6379),
    [anon_sym_type] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_exception] = ACTIONS(6379),
    [anon_sym_extern] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6381),
    [anon_sym_POUNDload] = ACTIONS(6381),
    [anon_sym_POUNDtime] = ACTIONS(6381),
    [anon_sym_POUNDI] = ACTIONS(6381),
    [anon_sym_POUNDhelp] = ACTIONS(6381),
    [anon_sym_POUNDquit] = ACTIONS(6381),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(3351)] = {
    [sym_block_comment] = STATE(3351),
    [sym_compiler_directive_decl] = STATE(3351),
    [sym_fsi_directive_decl] = STATE(3351),
    [sym_preproc_line] = STATE(3351),
    [ts_builtin_sym_end] = ACTIONS(6385),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_namespace] = ACTIONS(6383),
    [anon_sym_module] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_open] = ACTIONS(6383),
    [anon_sym_type] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_exception] = ACTIONS(6383),
    [anon_sym_extern] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6385),
    [anon_sym_POUNDload] = ACTIONS(6385),
    [anon_sym_POUNDtime] = ACTIONS(6385),
    [anon_sym_POUNDI] = ACTIONS(6385),
    [anon_sym_POUNDhelp] = ACTIONS(6385),
    [anon_sym_POUNDquit] = ACTIONS(6385),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(3352)] = {
    [sym_block_comment] = STATE(3352),
    [sym_compiler_directive_decl] = STATE(3352),
    [sym_fsi_directive_decl] = STATE(3352),
    [sym_preproc_line] = STATE(3352),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_DOT_DOT] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__dedent] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3353)] = {
    [sym_block_comment] = STATE(3353),
    [sym_compiler_directive_decl] = STATE(3353),
    [sym_fsi_directive_decl] = STATE(3353),
    [sym_preproc_line] = STATE(3353),
    [aux_sym_sequential_expression_repeat1] = STATE(3353),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [aux_sym_access_modifier_token1] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_default] = ACTIONS(6800),
    [anon_sym_static] = ACTIONS(6800),
    [anon_sym_member] = ACTIONS(6800),
    [anon_sym_interface] = ACTIONS(6800),
    [anon_sym_abstract] = ACTIONS(6800),
    [anon_sym_val] = ACTIONS(6800),
    [anon_sym_override] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8069),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3354)] = {
    [sym_block_comment] = STATE(3354),
    [sym_compiler_directive_decl] = STATE(3354),
    [sym_fsi_directive_decl] = STATE(3354),
    [sym_preproc_line] = STATE(3354),
    [ts_builtin_sym_end] = ACTIONS(6393),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_namespace] = ACTIONS(6391),
    [anon_sym_module] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_open] = ACTIONS(6391),
    [anon_sym_type] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_exception] = ACTIONS(6391),
    [anon_sym_extern] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6393),
    [anon_sym_POUNDload] = ACTIONS(6393),
    [anon_sym_POUNDtime] = ACTIONS(6393),
    [anon_sym_POUNDI] = ACTIONS(6393),
    [anon_sym_POUNDhelp] = ACTIONS(6393),
    [anon_sym_POUNDquit] = ACTIONS(6393),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(3355)] = {
    [sym_block_comment] = STATE(3355),
    [sym_compiler_directive_decl] = STATE(3355),
    [sym_fsi_directive_decl] = STATE(3355),
    [sym_preproc_line] = STATE(3355),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_and] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__dedent] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3356)] = {
    [sym_block_comment] = STATE(3356),
    [sym_compiler_directive_decl] = STATE(3356),
    [sym_fsi_directive_decl] = STATE(3356),
    [sym_preproc_line] = STATE(3356),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_DOT_DOT] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__dedent] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3357)] = {
    [sym_block_comment] = STATE(3357),
    [sym_compiler_directive_decl] = STATE(3357),
    [sym_fsi_directive_decl] = STATE(3357),
    [sym_preproc_line] = STATE(3357),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_and] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__dedent] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3358)] = {
    [sym_block_comment] = STATE(3358),
    [sym_compiler_directive_decl] = STATE(3358),
    [sym_fsi_directive_decl] = STATE(3358),
    [sym_preproc_line] = STATE(3358),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_and] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__dedent] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3359)] = {
    [sym_block_comment] = STATE(3359),
    [sym_compiler_directive_decl] = STATE(3359),
    [sym_fsi_directive_decl] = STATE(3359),
    [sym_preproc_line] = STATE(3359),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_and] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__dedent] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3360)] = {
    [sym_block_comment] = STATE(3360),
    [sym_compiler_directive_decl] = STATE(3360),
    [sym_fsi_directive_decl] = STATE(3360),
    [sym_preproc_line] = STATE(3360),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_and] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__dedent] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3361)] = {
    [sym_block_comment] = STATE(3361),
    [sym_compiler_directive_decl] = STATE(3361),
    [sym_fsi_directive_decl] = STATE(3361),
    [sym_preproc_line] = STATE(3361),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_DOT_DOT] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__dedent] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3362)] = {
    [sym_block_comment] = STATE(3362),
    [sym_compiler_directive_decl] = STATE(3362),
    [sym_fsi_directive_decl] = STATE(3362),
    [sym_preproc_line] = STATE(3362),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_DOT_DOT] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__dedent] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(3363)] = {
    [sym_block_comment] = STATE(3363),
    [sym_compiler_directive_decl] = STATE(3363),
    [sym_fsi_directive_decl] = STATE(3363),
    [sym_preproc_line] = STATE(3363),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_and] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__dedent] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3364)] = {
    [sym_block_comment] = STATE(3364),
    [sym_compiler_directive_decl] = STATE(3364),
    [sym_fsi_directive_decl] = STATE(3364),
    [sym_preproc_line] = STATE(3364),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_DOT_DOT] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__dedent] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3365)] = {
    [sym_block_comment] = STATE(3365),
    [sym_compiler_directive_decl] = STATE(3365),
    [sym_fsi_directive_decl] = STATE(3365),
    [sym_preproc_line] = STATE(3365),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_and] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__dedent] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3366)] = {
    [sym_block_comment] = STATE(3366),
    [sym_compiler_directive_decl] = STATE(3366),
    [sym_fsi_directive_decl] = STATE(3366),
    [sym_preproc_line] = STATE(3366),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_and] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__dedent] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3367)] = {
    [sym_block_comment] = STATE(3367),
    [sym_compiler_directive_decl] = STATE(3367),
    [sym_fsi_directive_decl] = STATE(3367),
    [sym_preproc_line] = STATE(3367),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_and] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__dedent] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3368)] = {
    [sym_block_comment] = STATE(3368),
    [sym_compiler_directive_decl] = STATE(3368),
    [sym_fsi_directive_decl] = STATE(3368),
    [sym_preproc_line] = STATE(3368),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [anon_sym_inherit] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3369)] = {
    [sym_block_comment] = STATE(3369),
    [sym_compiler_directive_decl] = STATE(3369),
    [sym_fsi_directive_decl] = STATE(3369),
    [sym_preproc_line] = STATE(3369),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_and] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__dedent] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3370)] = {
    [sym_block_comment] = STATE(3370),
    [sym_compiler_directive_decl] = STATE(3370),
    [sym_fsi_directive_decl] = STATE(3370),
    [sym_preproc_line] = STATE(3370),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_and] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__dedent] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3371)] = {
    [sym_block_comment] = STATE(3371),
    [sym_compiler_directive_decl] = STATE(3371),
    [sym_fsi_directive_decl] = STATE(3371),
    [sym_preproc_line] = STATE(3371),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_and] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__dedent] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3372)] = {
    [sym_elif_expression] = STATE(3754),
    [sym_block_comment] = STATE(3372),
    [sym_compiler_directive_decl] = STATE(3372),
    [sym_fsi_directive_decl] = STATE(3372),
    [sym_preproc_line] = STATE(3372),
    [aux_sym_if_expression_repeat1] = STATE(3372),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_module] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_open] = ACTIONS(6496),
    [anon_sym_type] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8072),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_exception] = ACTIONS(6496),
    [anon_sym_extern] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6494),
    [anon_sym_POUNDload] = ACTIONS(6494),
    [anon_sym_POUNDtime] = ACTIONS(6494),
    [anon_sym_POUNDI] = ACTIONS(6494),
    [anon_sym_POUNDhelp] = ACTIONS(6494),
    [anon_sym_POUNDquit] = ACTIONS(6494),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [anon_sym_POUNDelse] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(3373)] = {
    [sym_block_comment] = STATE(3373),
    [sym_compiler_directive_decl] = STATE(3373),
    [sym_fsi_directive_decl] = STATE(3373),
    [sym_preproc_line] = STATE(3373),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_and] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__dedent] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3374)] = {
    [sym_block_comment] = STATE(3374),
    [sym_compiler_directive_decl] = STATE(3374),
    [sym_fsi_directive_decl] = STATE(3374),
    [sym_preproc_line] = STATE(3374),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_and] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__dedent] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3375)] = {
    [sym_block_comment] = STATE(3375),
    [sym_compiler_directive_decl] = STATE(3375),
    [sym_fsi_directive_decl] = STATE(3375),
    [sym_preproc_line] = STATE(3375),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_and] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__dedent] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3376)] = {
    [sym_block_comment] = STATE(3376),
    [sym_compiler_directive_decl] = STATE(3376),
    [sym_fsi_directive_decl] = STATE(3376),
    [sym_preproc_line] = STATE(3376),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [anon_sym_inherit] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3377)] = {
    [sym_block_comment] = STATE(3377),
    [sym_compiler_directive_decl] = STATE(3377),
    [sym_fsi_directive_decl] = STATE(3377),
    [sym_preproc_line] = STATE(3377),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_and] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(3378)] = {
    [sym_block_comment] = STATE(3378),
    [sym_compiler_directive_decl] = STATE(3378),
    [sym_fsi_directive_decl] = STATE(3378),
    [sym_preproc_line] = STATE(3378),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_and] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__dedent] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3379)] = {
    [sym_block_comment] = STATE(3379),
    [sym_compiler_directive_decl] = STATE(3379),
    [sym_fsi_directive_decl] = STATE(3379),
    [sym_preproc_line] = STATE(3379),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_and] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__dedent] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3380)] = {
    [sym_block_comment] = STATE(3380),
    [sym_compiler_directive_decl] = STATE(3380),
    [sym_fsi_directive_decl] = STATE(3380),
    [sym_preproc_line] = STATE(3380),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_and] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__dedent] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3381)] = {
    [sym_block_comment] = STATE(3381),
    [sym_compiler_directive_decl] = STATE(3381),
    [sym_fsi_directive_decl] = STATE(3381),
    [sym_preproc_line] = STATE(3381),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_and] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__dedent] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3382)] = {
    [sym_block_comment] = STATE(3382),
    [sym_compiler_directive_decl] = STATE(3382),
    [sym_fsi_directive_decl] = STATE(3382),
    [sym_preproc_line] = STATE(3382),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [anon_sym_inherit] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3383)] = {
    [sym_block_comment] = STATE(3383),
    [sym_compiler_directive_decl] = STATE(3383),
    [sym_fsi_directive_decl] = STATE(3383),
    [sym_preproc_line] = STATE(3383),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [anon_sym_inherit] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3384)] = {
    [sym_block_comment] = STATE(3384),
    [sym_compiler_directive_decl] = STATE(3384),
    [sym_fsi_directive_decl] = STATE(3384),
    [sym_preproc_line] = STATE(3384),
    [ts_builtin_sym_end] = ACTIONS(6397),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_namespace] = ACTIONS(6395),
    [anon_sym_module] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_open] = ACTIONS(6395),
    [anon_sym_type] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_exception] = ACTIONS(6395),
    [anon_sym_extern] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6397),
    [anon_sym_POUNDload] = ACTIONS(6397),
    [anon_sym_POUNDtime] = ACTIONS(6397),
    [anon_sym_POUNDI] = ACTIONS(6397),
    [anon_sym_POUNDhelp] = ACTIONS(6397),
    [anon_sym_POUNDquit] = ACTIONS(6397),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(3385)] = {
    [sym_block_comment] = STATE(3385),
    [sym_compiler_directive_decl] = STATE(3385),
    [sym_fsi_directive_decl] = STATE(3385),
    [sym_preproc_line] = STATE(3385),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_DOT_DOT] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__dedent] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3386)] = {
    [sym_block_comment] = STATE(3386),
    [sym_compiler_directive_decl] = STATE(3386),
    [sym_fsi_directive_decl] = STATE(3386),
    [sym_preproc_line] = STATE(3386),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_and] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__dedent] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3387)] = {
    [sym_block_comment] = STATE(3387),
    [sym_compiler_directive_decl] = STATE(3387),
    [sym_fsi_directive_decl] = STATE(3387),
    [sym_preproc_line] = STATE(3387),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [anon_sym_inherit] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3388)] = {
    [sym_block_comment] = STATE(3388),
    [sym_compiler_directive_decl] = STATE(3388),
    [sym_fsi_directive_decl] = STATE(3388),
    [sym_preproc_line] = STATE(3388),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_and] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__dedent] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3389)] = {
    [sym_block_comment] = STATE(3389),
    [sym_compiler_directive_decl] = STATE(3389),
    [sym_fsi_directive_decl] = STATE(3389),
    [sym_preproc_line] = STATE(3389),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_and] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__dedent] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3390)] = {
    [sym_block_comment] = STATE(3390),
    [sym_compiler_directive_decl] = STATE(3390),
    [sym_fsi_directive_decl] = STATE(3390),
    [sym_preproc_line] = STATE(3390),
    [ts_builtin_sym_end] = ACTIONS(6401),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_namespace] = ACTIONS(6399),
    [anon_sym_module] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_open] = ACTIONS(6399),
    [anon_sym_type] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_exception] = ACTIONS(6399),
    [anon_sym_extern] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6401),
    [anon_sym_POUNDload] = ACTIONS(6401),
    [anon_sym_POUNDtime] = ACTIONS(6401),
    [anon_sym_POUNDI] = ACTIONS(6401),
    [anon_sym_POUNDhelp] = ACTIONS(6401),
    [anon_sym_POUNDquit] = ACTIONS(6401),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(3391)] = {
    [sym__else_expression] = STATE(4161),
    [sym_elif_expression] = STATE(3828),
    [sym_block_comment] = STATE(3391),
    [sym_compiler_directive_decl] = STATE(3391),
    [sym_fsi_directive_decl] = STATE(3391),
    [sym_preproc_line] = STATE(3391),
    [aux_sym_if_expression_repeat1] = STATE(3395),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_module] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_open] = ACTIONS(6319),
    [anon_sym_type] = ACTIONS(6319),
    [anon_sym_LBRACK_LT] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8075),
    [anon_sym_elif] = ACTIONS(8077),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_exception] = ACTIONS(6319),
    [anon_sym_extern] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6317),
    [anon_sym_POUNDload] = ACTIONS(6317),
    [anon_sym_POUNDtime] = ACTIONS(6317),
    [anon_sym_POUNDI] = ACTIONS(6317),
    [anon_sym_POUNDhelp] = ACTIONS(6317),
    [anon_sym_POUNDquit] = ACTIONS(6317),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(3392)] = {
    [sym_block_comment] = STATE(3392),
    [sym_compiler_directive_decl] = STATE(3392),
    [sym_fsi_directive_decl] = STATE(3392),
    [sym_preproc_line] = STATE(3392),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_and] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__dedent] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3393)] = {
    [sym_block_comment] = STATE(3393),
    [sym_compiler_directive_decl] = STATE(3393),
    [sym_fsi_directive_decl] = STATE(3393),
    [sym_preproc_line] = STATE(3393),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_and] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__dedent] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(3394)] = {
    [sym_block_comment] = STATE(3394),
    [sym_compiler_directive_decl] = STATE(3394),
    [sym_fsi_directive_decl] = STATE(3394),
    [sym_preproc_line] = STATE(3394),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_and] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__dedent] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(3395)] = {
    [sym__else_expression] = STATE(4173),
    [sym_elif_expression] = STATE(3828),
    [sym_block_comment] = STATE(3395),
    [sym_compiler_directive_decl] = STATE(3395),
    [sym_fsi_directive_decl] = STATE(3395),
    [sym_preproc_line] = STATE(3395),
    [aux_sym_if_expression_repeat1] = STATE(3695),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_module] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_open] = ACTIONS(6329),
    [anon_sym_type] = ACTIONS(6329),
    [anon_sym_LBRACK_LT] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8075),
    [anon_sym_elif] = ACTIONS(8077),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_exception] = ACTIONS(6329),
    [anon_sym_extern] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6327),
    [anon_sym_POUNDload] = ACTIONS(6327),
    [anon_sym_POUNDtime] = ACTIONS(6327),
    [anon_sym_POUNDI] = ACTIONS(6327),
    [anon_sym_POUNDhelp] = ACTIONS(6327),
    [anon_sym_POUNDquit] = ACTIONS(6327),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(3396)] = {
    [sym_block_comment] = STATE(3396),
    [sym_compiler_directive_decl] = STATE(3396),
    [sym_fsi_directive_decl] = STATE(3396),
    [sym_preproc_line] = STATE(3396),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_DOT_DOT] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__dedent] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3397)] = {
    [sym_block_comment] = STATE(3397),
    [sym_compiler_directive_decl] = STATE(3397),
    [sym_fsi_directive_decl] = STATE(3397),
    [sym_preproc_line] = STATE(3397),
    [ts_builtin_sym_end] = ACTIONS(6405),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_namespace] = ACTIONS(6403),
    [anon_sym_module] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_open] = ACTIONS(6403),
    [anon_sym_type] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_exception] = ACTIONS(6403),
    [anon_sym_extern] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6405),
    [anon_sym_POUNDload] = ACTIONS(6405),
    [anon_sym_POUNDtime] = ACTIONS(6405),
    [anon_sym_POUNDI] = ACTIONS(6405),
    [anon_sym_POUNDhelp] = ACTIONS(6405),
    [anon_sym_POUNDquit] = ACTIONS(6405),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(3398)] = {
    [sym_block_comment] = STATE(3398),
    [sym_compiler_directive_decl] = STATE(3398),
    [sym_fsi_directive_decl] = STATE(3398),
    [sym_preproc_line] = STATE(3398),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_DOT_DOT] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__dedent] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3399)] = {
    [sym_block_comment] = STATE(3399),
    [sym_compiler_directive_decl] = STATE(3399),
    [sym_fsi_directive_decl] = STATE(3399),
    [sym_preproc_line] = STATE(3399),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_and] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3400)] = {
    [sym_block_comment] = STATE(3400),
    [sym_compiler_directive_decl] = STATE(3400),
    [sym_fsi_directive_decl] = STATE(3400),
    [sym_preproc_line] = STATE(3400),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [anon_sym_inherit] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(3401)] = {
    [sym_block_comment] = STATE(3401),
    [sym_compiler_directive_decl] = STATE(3401),
    [sym_fsi_directive_decl] = STATE(3401),
    [sym_preproc_line] = STATE(3401),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_and] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__dedent] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3402)] = {
    [sym_block_comment] = STATE(3402),
    [sym_compiler_directive_decl] = STATE(3402),
    [sym_fsi_directive_decl] = STATE(3402),
    [sym_preproc_line] = STATE(3402),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [anon_sym_inherit] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3403)] = {
    [sym_block_comment] = STATE(3403),
    [sym_compiler_directive_decl] = STATE(3403),
    [sym_fsi_directive_decl] = STATE(3403),
    [sym_preproc_line] = STATE(3403),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [anon_sym_inherit] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(3404)] = {
    [sym_block_comment] = STATE(3404),
    [sym_compiler_directive_decl] = STATE(3404),
    [sym_fsi_directive_decl] = STATE(3404),
    [sym_preproc_line] = STATE(3404),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [anon_sym_inherit] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3405)] = {
    [sym_block_comment] = STATE(3405),
    [sym_compiler_directive_decl] = STATE(3405),
    [sym_fsi_directive_decl] = STATE(3405),
    [sym_preproc_line] = STATE(3405),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [anon_sym_inherit] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3406)] = {
    [sym_block_comment] = STATE(3406),
    [sym_compiler_directive_decl] = STATE(3406),
    [sym_fsi_directive_decl] = STATE(3406),
    [sym_preproc_line] = STATE(3406),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_DOT_DOT] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__dedent] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3407)] = {
    [sym_block_comment] = STATE(3407),
    [sym_compiler_directive_decl] = STATE(3407),
    [sym_fsi_directive_decl] = STATE(3407),
    [sym_preproc_line] = STATE(3407),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [anon_sym_inherit] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3408)] = {
    [sym_block_comment] = STATE(3408),
    [sym_compiler_directive_decl] = STATE(3408),
    [sym_fsi_directive_decl] = STATE(3408),
    [sym_preproc_line] = STATE(3408),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [anon_sym_inherit] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3409)] = {
    [sym_block_comment] = STATE(3409),
    [sym_compiler_directive_decl] = STATE(3409),
    [sym_fsi_directive_decl] = STATE(3409),
    [sym_preproc_line] = STATE(3409),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [anon_sym_inherit] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3410)] = {
    [sym_block_comment] = STATE(3410),
    [sym_compiler_directive_decl] = STATE(3410),
    [sym_fsi_directive_decl] = STATE(3410),
    [sym_preproc_line] = STATE(3410),
    [ts_builtin_sym_end] = ACTIONS(6409),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_namespace] = ACTIONS(6407),
    [anon_sym_module] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_open] = ACTIONS(6407),
    [anon_sym_type] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_exception] = ACTIONS(6407),
    [anon_sym_extern] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6409),
    [anon_sym_POUNDload] = ACTIONS(6409),
    [anon_sym_POUNDtime] = ACTIONS(6409),
    [anon_sym_POUNDI] = ACTIONS(6409),
    [anon_sym_POUNDhelp] = ACTIONS(6409),
    [anon_sym_POUNDquit] = ACTIONS(6409),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(3411)] = {
    [sym_block_comment] = STATE(3411),
    [sym_compiler_directive_decl] = STATE(3411),
    [sym_fsi_directive_decl] = STATE(3411),
    [sym_preproc_line] = STATE(3411),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [anon_sym_inherit] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3412)] = {
    [sym_block_comment] = STATE(3412),
    [sym_compiler_directive_decl] = STATE(3412),
    [sym_fsi_directive_decl] = STATE(3412),
    [sym_preproc_line] = STATE(3412),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [anon_sym_inherit] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(3413)] = {
    [sym_block_comment] = STATE(3413),
    [sym_compiler_directive_decl] = STATE(3413),
    [sym_fsi_directive_decl] = STATE(3413),
    [sym_preproc_line] = STATE(3413),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_DOT_DOT] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__dedent] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3414)] = {
    [sym_block_comment] = STATE(3414),
    [sym_compiler_directive_decl] = STATE(3414),
    [sym_fsi_directive_decl] = STATE(3414),
    [sym_preproc_line] = STATE(3414),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [anon_sym_inherit] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3415)] = {
    [sym_block_comment] = STATE(3415),
    [sym_compiler_directive_decl] = STATE(3415),
    [sym_fsi_directive_decl] = STATE(3415),
    [sym_preproc_line] = STATE(3415),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_and] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__dedent] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3416)] = {
    [sym_block_comment] = STATE(3416),
    [sym_compiler_directive_decl] = STATE(3416),
    [sym_fsi_directive_decl] = STATE(3416),
    [sym_preproc_line] = STATE(3416),
    [ts_builtin_sym_end] = ACTIONS(6413),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_namespace] = ACTIONS(6411),
    [anon_sym_module] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_open] = ACTIONS(6411),
    [anon_sym_type] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_exception] = ACTIONS(6411),
    [anon_sym_extern] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6413),
    [anon_sym_POUNDload] = ACTIONS(6413),
    [anon_sym_POUNDtime] = ACTIONS(6413),
    [anon_sym_POUNDI] = ACTIONS(6413),
    [anon_sym_POUNDhelp] = ACTIONS(6413),
    [anon_sym_POUNDquit] = ACTIONS(6413),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(3417)] = {
    [sym_block_comment] = STATE(3417),
    [sym_compiler_directive_decl] = STATE(3417),
    [sym_fsi_directive_decl] = STATE(3417),
    [sym_preproc_line] = STATE(3417),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [anon_sym_inherit] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3418)] = {
    [sym_block_comment] = STATE(3418),
    [sym_compiler_directive_decl] = STATE(3418),
    [sym_fsi_directive_decl] = STATE(3418),
    [sym_preproc_line] = STATE(3418),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_and] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__dedent] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(3419)] = {
    [sym_block_comment] = STATE(3419),
    [sym_compiler_directive_decl] = STATE(3419),
    [sym_fsi_directive_decl] = STATE(3419),
    [sym_preproc_line] = STATE(3419),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [anon_sym_inherit] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3420)] = {
    [sym_block_comment] = STATE(3420),
    [sym_compiler_directive_decl] = STATE(3420),
    [sym_fsi_directive_decl] = STATE(3420),
    [sym_preproc_line] = STATE(3420),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [anon_sym_inherit] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3421)] = {
    [sym_block_comment] = STATE(3421),
    [sym_compiler_directive_decl] = STATE(3421),
    [sym_fsi_directive_decl] = STATE(3421),
    [sym_preproc_line] = STATE(3421),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [anon_sym_inherit] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3422)] = {
    [sym_block_comment] = STATE(3422),
    [sym_compiler_directive_decl] = STATE(3422),
    [sym_fsi_directive_decl] = STATE(3422),
    [sym_preproc_line] = STATE(3422),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [anon_sym_inherit] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3423)] = {
    [sym_block_comment] = STATE(3423),
    [sym_compiler_directive_decl] = STATE(3423),
    [sym_fsi_directive_decl] = STATE(3423),
    [sym_preproc_line] = STATE(3423),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [anon_sym_inherit] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3424)] = {
    [sym_block_comment] = STATE(3424),
    [sym_compiler_directive_decl] = STATE(3424),
    [sym_fsi_directive_decl] = STATE(3424),
    [sym_preproc_line] = STATE(3424),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [anon_sym_inherit] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3425)] = {
    [sym_block_comment] = STATE(3425),
    [sym_compiler_directive_decl] = STATE(3425),
    [sym_fsi_directive_decl] = STATE(3425),
    [sym_preproc_line] = STATE(3425),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [anon_sym_inherit] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3426)] = {
    [sym_block_comment] = STATE(3426),
    [sym_compiler_directive_decl] = STATE(3426),
    [sym_fsi_directive_decl] = STATE(3426),
    [sym_preproc_line] = STATE(3426),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [anon_sym_inherit] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3427)] = {
    [sym_block_comment] = STATE(3427),
    [sym_compiler_directive_decl] = STATE(3427),
    [sym_fsi_directive_decl] = STATE(3427),
    [sym_preproc_line] = STATE(3427),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [anon_sym_inherit] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3428)] = {
    [sym_block_comment] = STATE(3428),
    [sym_compiler_directive_decl] = STATE(3428),
    [sym_fsi_directive_decl] = STATE(3428),
    [sym_preproc_line] = STATE(3428),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [anon_sym_inherit] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3429)] = {
    [sym_block_comment] = STATE(3429),
    [sym_compiler_directive_decl] = STATE(3429),
    [sym_fsi_directive_decl] = STATE(3429),
    [sym_preproc_line] = STATE(3429),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [anon_sym_inherit] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3430)] = {
    [sym_block_comment] = STATE(3430),
    [sym_compiler_directive_decl] = STATE(3430),
    [sym_fsi_directive_decl] = STATE(3430),
    [sym_preproc_line] = STATE(3430),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [anon_sym_inherit] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3431)] = {
    [sym_block_comment] = STATE(3431),
    [sym_compiler_directive_decl] = STATE(3431),
    [sym_fsi_directive_decl] = STATE(3431),
    [sym_preproc_line] = STATE(3431),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [anon_sym_inherit] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3432)] = {
    [sym_block_comment] = STATE(3432),
    [sym_compiler_directive_decl] = STATE(3432),
    [sym_fsi_directive_decl] = STATE(3432),
    [sym_preproc_line] = STATE(3432),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_and] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__dedent] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(3433)] = {
    [sym_block_comment] = STATE(3433),
    [sym_compiler_directive_decl] = STATE(3433),
    [sym_fsi_directive_decl] = STATE(3433),
    [sym_preproc_line] = STATE(3433),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_and] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__dedent] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3434)] = {
    [sym_block_comment] = STATE(3434),
    [sym_compiler_directive_decl] = STATE(3434),
    [sym_fsi_directive_decl] = STATE(3434),
    [sym_preproc_line] = STATE(3434),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [anon_sym_inherit] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3435)] = {
    [sym_block_comment] = STATE(3435),
    [sym_compiler_directive_decl] = STATE(3435),
    [sym_fsi_directive_decl] = STATE(3435),
    [sym_preproc_line] = STATE(3435),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [anon_sym_inherit] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3436)] = {
    [sym_block_comment] = STATE(3436),
    [sym_compiler_directive_decl] = STATE(3436),
    [sym_fsi_directive_decl] = STATE(3436),
    [sym_preproc_line] = STATE(3436),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [anon_sym_inherit] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3437)] = {
    [sym_block_comment] = STATE(3437),
    [sym_compiler_directive_decl] = STATE(3437),
    [sym_fsi_directive_decl] = STATE(3437),
    [sym_preproc_line] = STATE(3437),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [anon_sym_inherit] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3438)] = {
    [sym_block_comment] = STATE(3438),
    [sym_compiler_directive_decl] = STATE(3438),
    [sym_fsi_directive_decl] = STATE(3438),
    [sym_preproc_line] = STATE(3438),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_and] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__dedent] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3439)] = {
    [sym_block_comment] = STATE(3439),
    [sym_compiler_directive_decl] = STATE(3439),
    [sym_fsi_directive_decl] = STATE(3439),
    [sym_preproc_line] = STATE(3439),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [anon_sym_inherit] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3440)] = {
    [sym_block_comment] = STATE(3440),
    [sym_compiler_directive_decl] = STATE(3440),
    [sym_fsi_directive_decl] = STATE(3440),
    [sym_preproc_line] = STATE(3440),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_and] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__dedent] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(3441)] = {
    [sym_block_comment] = STATE(3441),
    [sym_compiler_directive_decl] = STATE(3441),
    [sym_fsi_directive_decl] = STATE(3441),
    [sym_preproc_line] = STATE(3441),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [anon_sym_inherit] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3442)] = {
    [sym_block_comment] = STATE(3442),
    [sym_compiler_directive_decl] = STATE(3442),
    [sym_fsi_directive_decl] = STATE(3442),
    [sym_preproc_line] = STATE(3442),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [anon_sym_inherit] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3443)] = {
    [sym_block_comment] = STATE(3443),
    [sym_compiler_directive_decl] = STATE(3443),
    [sym_fsi_directive_decl] = STATE(3443),
    [sym_preproc_line] = STATE(3443),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [anon_sym_inherit] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(3444)] = {
    [sym_block_comment] = STATE(3444),
    [sym_compiler_directive_decl] = STATE(3444),
    [sym_fsi_directive_decl] = STATE(3444),
    [sym_preproc_line] = STATE(3444),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_DOT_DOT] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__dedent] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3445)] = {
    [sym_block_comment] = STATE(3445),
    [sym_compiler_directive_decl] = STATE(3445),
    [sym_fsi_directive_decl] = STATE(3445),
    [sym_preproc_line] = STATE(3445),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_DOT_DOT] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__dedent] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3446)] = {
    [sym_block_comment] = STATE(3446),
    [sym_compiler_directive_decl] = STATE(3446),
    [sym_fsi_directive_decl] = STATE(3446),
    [sym_preproc_line] = STATE(3446),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_DOT_DOT] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__dedent] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(3447)] = {
    [sym_block_comment] = STATE(3447),
    [sym_compiler_directive_decl] = STATE(3447),
    [sym_fsi_directive_decl] = STATE(3447),
    [sym_preproc_line] = STATE(3447),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [anon_sym_inherit] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3448)] = {
    [sym_block_comment] = STATE(3448),
    [sym_compiler_directive_decl] = STATE(3448),
    [sym_fsi_directive_decl] = STATE(3448),
    [sym_preproc_line] = STATE(3448),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [anon_sym_inherit] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(3449)] = {
    [sym_block_comment] = STATE(3449),
    [sym_compiler_directive_decl] = STATE(3449),
    [sym_fsi_directive_decl] = STATE(3449),
    [sym_preproc_line] = STATE(3449),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [anon_sym_inherit] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3450)] = {
    [sym_block_comment] = STATE(3450),
    [sym_compiler_directive_decl] = STATE(3450),
    [sym_fsi_directive_decl] = STATE(3450),
    [sym_preproc_line] = STATE(3450),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [anon_sym_inherit] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(3451)] = {
    [sym_block_comment] = STATE(3451),
    [sym_compiler_directive_decl] = STATE(3451),
    [sym_fsi_directive_decl] = STATE(3451),
    [sym_preproc_line] = STATE(3451),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3452)] = {
    [sym_block_comment] = STATE(3452),
    [sym_compiler_directive_decl] = STATE(3452),
    [sym_fsi_directive_decl] = STATE(3452),
    [sym_preproc_line] = STATE(3452),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_and] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3453)] = {
    [sym_block_comment] = STATE(3453),
    [sym_compiler_directive_decl] = STATE(3453),
    [sym_fsi_directive_decl] = STATE(3453),
    [sym_preproc_line] = STATE(3453),
    [aux_sym_long_identifier_repeat1] = STATE(3454),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8079),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3454)] = {
    [sym_block_comment] = STATE(3454),
    [sym_compiler_directive_decl] = STATE(3454),
    [sym_fsi_directive_decl] = STATE(3454),
    [sym_preproc_line] = STATE(3454),
    [aux_sym_long_identifier_repeat1] = STATE(3454),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8081),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3455)] = {
    [sym_block_comment] = STATE(3455),
    [sym_compiler_directive_decl] = STATE(3455),
    [sym_fsi_directive_decl] = STATE(3455),
    [sym_preproc_line] = STATE(3455),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_DOT_DOT] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__dedent] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3456)] = {
    [sym_block_comment] = STATE(3456),
    [sym_compiler_directive_decl] = STATE(3456),
    [sym_fsi_directive_decl] = STATE(3456),
    [sym_preproc_line] = STATE(3456),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_and] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__dedent] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3457)] = {
    [sym_block_comment] = STATE(3457),
    [sym_compiler_directive_decl] = STATE(3457),
    [sym_fsi_directive_decl] = STATE(3457),
    [sym_preproc_line] = STATE(3457),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [anon_sym_inherit] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3458)] = {
    [sym_block_comment] = STATE(3458),
    [sym_compiler_directive_decl] = STATE(3458),
    [sym_fsi_directive_decl] = STATE(3458),
    [sym_preproc_line] = STATE(3458),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [anon_sym_inherit] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(3459)] = {
    [sym_block_comment] = STATE(3459),
    [sym_compiler_directive_decl] = STATE(3459),
    [sym_fsi_directive_decl] = STATE(3459),
    [sym_preproc_line] = STATE(3459),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_and] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__dedent] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3460)] = {
    [sym_block_comment] = STATE(3460),
    [sym_compiler_directive_decl] = STATE(3460),
    [sym_fsi_directive_decl] = STATE(3460),
    [sym_preproc_line] = STATE(3460),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_DOT_DOT] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__dedent] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3461)] = {
    [sym_block_comment] = STATE(3461),
    [sym_compiler_directive_decl] = STATE(3461),
    [sym_fsi_directive_decl] = STATE(3461),
    [sym_preproc_line] = STATE(3461),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_and] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__dedent] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3462)] = {
    [sym_block_comment] = STATE(3462),
    [sym_compiler_directive_decl] = STATE(3462),
    [sym_fsi_directive_decl] = STATE(3462),
    [sym_preproc_line] = STATE(3462),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_and] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__dedent] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3463)] = {
    [sym_block_comment] = STATE(3463),
    [sym_compiler_directive_decl] = STATE(3463),
    [sym_fsi_directive_decl] = STATE(3463),
    [sym_preproc_line] = STATE(3463),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_DOT_DOT] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__dedent] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3464)] = {
    [sym_block_comment] = STATE(3464),
    [sym_compiler_directive_decl] = STATE(3464),
    [sym_fsi_directive_decl] = STATE(3464),
    [sym_preproc_line] = STATE(3464),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_DOT_DOT] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__dedent] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3465)] = {
    [sym_block_comment] = STATE(3465),
    [sym_compiler_directive_decl] = STATE(3465),
    [sym_fsi_directive_decl] = STATE(3465),
    [sym_preproc_line] = STATE(3465),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_DOT_DOT] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__dedent] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3466)] = {
    [sym_block_comment] = STATE(3466),
    [sym_compiler_directive_decl] = STATE(3466),
    [sym_fsi_directive_decl] = STATE(3466),
    [sym_preproc_line] = STATE(3466),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_DOT_DOT] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__dedent] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3467)] = {
    [sym_block_comment] = STATE(3467),
    [sym_compiler_directive_decl] = STATE(3467),
    [sym_fsi_directive_decl] = STATE(3467),
    [sym_preproc_line] = STATE(3467),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_and] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__dedent] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3468)] = {
    [sym_block_comment] = STATE(3468),
    [sym_compiler_directive_decl] = STATE(3468),
    [sym_fsi_directive_decl] = STATE(3468),
    [sym_preproc_line] = STATE(3468),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_and] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [aux_sym_access_modifier_token1] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_default] = ACTIONS(7035),
    [anon_sym_static] = ACTIONS(7035),
    [anon_sym_member] = ACTIONS(7035),
    [anon_sym_interface] = ACTIONS(7035),
    [anon_sym_abstract] = ACTIONS(7035),
    [anon_sym_val] = ACTIONS(7035),
    [anon_sym_override] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__dedent] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3469)] = {
    [sym_block_comment] = STATE(3469),
    [sym_compiler_directive_decl] = STATE(3469),
    [sym_fsi_directive_decl] = STATE(3469),
    [sym_preproc_line] = STATE(3469),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_DOT_DOT] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__dedent] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3470)] = {
    [sym_block_comment] = STATE(3470),
    [sym_compiler_directive_decl] = STATE(3470),
    [sym_fsi_directive_decl] = STATE(3470),
    [sym_preproc_line] = STATE(3470),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_and] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3471)] = {
    [sym_block_comment] = STATE(3471),
    [sym_compiler_directive_decl] = STATE(3471),
    [sym_fsi_directive_decl] = STATE(3471),
    [sym_preproc_line] = STATE(3471),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3472)] = {
    [sym_block_comment] = STATE(3472),
    [sym_compiler_directive_decl] = STATE(3472),
    [sym_fsi_directive_decl] = STATE(3472),
    [sym_preproc_line] = STATE(3472),
    [ts_builtin_sym_end] = ACTIONS(6333),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_namespace] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3473)] = {
    [sym_block_comment] = STATE(3473),
    [sym_compiler_directive_decl] = STATE(3473),
    [sym_fsi_directive_decl] = STATE(3473),
    [sym_preproc_line] = STATE(3473),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_and] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3474)] = {
    [sym_block_comment] = STATE(3474),
    [sym_compiler_directive_decl] = STATE(3474),
    [sym_fsi_directive_decl] = STATE(3474),
    [sym_preproc_line] = STATE(3474),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_and] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3475)] = {
    [sym_block_comment] = STATE(3475),
    [sym_compiler_directive_decl] = STATE(3475),
    [sym_fsi_directive_decl] = STATE(3475),
    [sym_preproc_line] = STATE(3475),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_and] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(3476)] = {
    [sym_block_comment] = STATE(3476),
    [sym_compiler_directive_decl] = STATE(3476),
    [sym_fsi_directive_decl] = STATE(3476),
    [sym_preproc_line] = STATE(3476),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_and] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3477)] = {
    [sym_block_comment] = STATE(3477),
    [sym_compiler_directive_decl] = STATE(3477),
    [sym_fsi_directive_decl] = STATE(3477),
    [sym_preproc_line] = STATE(3477),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_and] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3478)] = {
    [sym_block_comment] = STATE(3478),
    [sym_compiler_directive_decl] = STATE(3478),
    [sym_fsi_directive_decl] = STATE(3478),
    [sym_preproc_line] = STATE(3478),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_and] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3479)] = {
    [sym_block_comment] = STATE(3479),
    [sym_compiler_directive_decl] = STATE(3479),
    [sym_fsi_directive_decl] = STATE(3479),
    [sym_preproc_line] = STATE(3479),
    [ts_builtin_sym_end] = ACTIONS(6310),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_namespace] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3480)] = {
    [sym_block_comment] = STATE(3480),
    [sym_compiler_directive_decl] = STATE(3480),
    [sym_fsi_directive_decl] = STATE(3480),
    [sym_preproc_line] = STATE(3480),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_and] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3481)] = {
    [sym_block_comment] = STATE(3481),
    [sym_compiler_directive_decl] = STATE(3481),
    [sym_fsi_directive_decl] = STATE(3481),
    [sym_preproc_line] = STATE(3481),
    [ts_builtin_sym_end] = ACTIONS(6490),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_namespace] = ACTIONS(6492),
    [anon_sym_module] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_open] = ACTIONS(6492),
    [anon_sym_type] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_exception] = ACTIONS(6492),
    [anon_sym_extern] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6490),
    [anon_sym_POUNDload] = ACTIONS(6490),
    [anon_sym_POUNDtime] = ACTIONS(6490),
    [anon_sym_POUNDI] = ACTIONS(6490),
    [anon_sym_POUNDhelp] = ACTIONS(6490),
    [anon_sym_POUNDquit] = ACTIONS(6490),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(3482)] = {
    [sym_block_comment] = STATE(3482),
    [sym_compiler_directive_decl] = STATE(3482),
    [sym_fsi_directive_decl] = STATE(3482),
    [sym_preproc_line] = STATE(3482),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_and] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3483)] = {
    [sym_block_comment] = STATE(3483),
    [sym_compiler_directive_decl] = STATE(3483),
    [sym_fsi_directive_decl] = STATE(3483),
    [sym_preproc_line] = STATE(3483),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_and] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3484)] = {
    [sym_block_comment] = STATE(3484),
    [sym_compiler_directive_decl] = STATE(3484),
    [sym_fsi_directive_decl] = STATE(3484),
    [sym_preproc_line] = STATE(3484),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_and] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3485)] = {
    [sym_block_comment] = STATE(3485),
    [sym_compiler_directive_decl] = STATE(3485),
    [sym_fsi_directive_decl] = STATE(3485),
    [sym_preproc_line] = STATE(3485),
    [ts_builtin_sym_end] = ACTIONS(6429),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_namespace] = ACTIONS(6427),
    [anon_sym_module] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_open] = ACTIONS(6427),
    [anon_sym_type] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_exception] = ACTIONS(6427),
    [anon_sym_extern] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6429),
    [anon_sym_POUNDload] = ACTIONS(6429),
    [anon_sym_POUNDtime] = ACTIONS(6429),
    [anon_sym_POUNDI] = ACTIONS(6429),
    [anon_sym_POUNDhelp] = ACTIONS(6429),
    [anon_sym_POUNDquit] = ACTIONS(6429),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(3486)] = {
    [sym_block_comment] = STATE(3486),
    [sym_compiler_directive_decl] = STATE(3486),
    [sym_fsi_directive_decl] = STATE(3486),
    [sym_preproc_line] = STATE(3486),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_and] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3487)] = {
    [sym_block_comment] = STATE(3487),
    [sym_compiler_directive_decl] = STATE(3487),
    [sym_fsi_directive_decl] = STATE(3487),
    [sym_preproc_line] = STATE(3487),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_and] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__dedent] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3488)] = {
    [sym_block_comment] = STATE(3488),
    [sym_compiler_directive_decl] = STATE(3488),
    [sym_fsi_directive_decl] = STATE(3488),
    [sym_preproc_line] = STATE(3488),
    [ts_builtin_sym_end] = ACTIONS(6387),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_namespace] = ACTIONS(6389),
    [anon_sym_module] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_open] = ACTIONS(6389),
    [anon_sym_type] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_exception] = ACTIONS(6389),
    [anon_sym_extern] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6387),
    [anon_sym_POUNDload] = ACTIONS(6387),
    [anon_sym_POUNDtime] = ACTIONS(6387),
    [anon_sym_POUNDI] = ACTIONS(6387),
    [anon_sym_POUNDhelp] = ACTIONS(6387),
    [anon_sym_POUNDquit] = ACTIONS(6387),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(3489)] = {
    [sym_block_comment] = STATE(3489),
    [sym_compiler_directive_decl] = STATE(3489),
    [sym_fsi_directive_decl] = STATE(3489),
    [sym_preproc_line] = STATE(3489),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_and] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3490)] = {
    [sym_block_comment] = STATE(3490),
    [sym_compiler_directive_decl] = STATE(3490),
    [sym_fsi_directive_decl] = STATE(3490),
    [sym_preproc_line] = STATE(3490),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_and] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3491)] = {
    [sym_block_comment] = STATE(3491),
    [sym_compiler_directive_decl] = STATE(3491),
    [sym_fsi_directive_decl] = STATE(3491),
    [sym_preproc_line] = STATE(3491),
    [aux_sym_long_identifier_repeat1] = STATE(3453),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8079),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(3492)] = {
    [sym_block_comment] = STATE(3492),
    [sym_compiler_directive_decl] = STATE(3492),
    [sym_fsi_directive_decl] = STATE(3492),
    [sym_preproc_line] = STATE(3492),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_and] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3493)] = {
    [sym_block_comment] = STATE(3493),
    [sym_compiler_directive_decl] = STATE(3493),
    [sym_fsi_directive_decl] = STATE(3493),
    [sym_preproc_line] = STATE(3493),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_and] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [aux_sym_access_modifier_token1] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_default] = ACTIONS(7039),
    [anon_sym_static] = ACTIONS(7039),
    [anon_sym_member] = ACTIONS(7039),
    [anon_sym_interface] = ACTIONS(7039),
    [anon_sym_abstract] = ACTIONS(7039),
    [anon_sym_val] = ACTIONS(7039),
    [anon_sym_override] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__dedent] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3494)] = {
    [sym_block_comment] = STATE(3494),
    [sym_compiler_directive_decl] = STATE(3494),
    [sym_fsi_directive_decl] = STATE(3494),
    [sym_preproc_line] = STATE(3494),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_and] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3495)] = {
    [sym_block_comment] = STATE(3495),
    [sym_compiler_directive_decl] = STATE(3495),
    [sym_fsi_directive_decl] = STATE(3495),
    [sym_preproc_line] = STATE(3495),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_and] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [aux_sym_access_modifier_token1] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_default] = ACTIONS(7043),
    [anon_sym_static] = ACTIONS(7043),
    [anon_sym_member] = ACTIONS(7043),
    [anon_sym_interface] = ACTIONS(7043),
    [anon_sym_abstract] = ACTIONS(7043),
    [anon_sym_val] = ACTIONS(7043),
    [anon_sym_override] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__dedent] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3496)] = {
    [sym_block_comment] = STATE(3496),
    [sym_compiler_directive_decl] = STATE(3496),
    [sym_fsi_directive_decl] = STATE(3496),
    [sym_preproc_line] = STATE(3496),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [anon_sym_inherit] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3497)] = {
    [sym_block_comment] = STATE(3497),
    [sym_compiler_directive_decl] = STATE(3497),
    [sym_fsi_directive_decl] = STATE(3497),
    [sym_preproc_line] = STATE(3497),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_and] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3498)] = {
    [sym_block_comment] = STATE(3498),
    [sym_compiler_directive_decl] = STATE(3498),
    [sym_fsi_directive_decl] = STATE(3498),
    [sym_preproc_line] = STATE(3498),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_and] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3499)] = {
    [sym_block_comment] = STATE(3499),
    [sym_compiler_directive_decl] = STATE(3499),
    [sym_fsi_directive_decl] = STATE(3499),
    [sym_preproc_line] = STATE(3499),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_and] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__dedent] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3500)] = {
    [sym_block_comment] = STATE(3500),
    [sym_compiler_directive_decl] = STATE(3500),
    [sym_fsi_directive_decl] = STATE(3500),
    [sym_preproc_line] = STATE(3500),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [anon_sym_inherit] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3501)] = {
    [sym_block_comment] = STATE(3501),
    [sym_compiler_directive_decl] = STATE(3501),
    [sym_fsi_directive_decl] = STATE(3501),
    [sym_preproc_line] = STATE(3501),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_DOT_DOT] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__dedent] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(3502)] = {
    [sym_block_comment] = STATE(3502),
    [sym_compiler_directive_decl] = STATE(3502),
    [sym_fsi_directive_decl] = STATE(3502),
    [sym_preproc_line] = STATE(3502),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_and] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3503)] = {
    [sym_block_comment] = STATE(3503),
    [sym_compiler_directive_decl] = STATE(3503),
    [sym_fsi_directive_decl] = STATE(3503),
    [sym_preproc_line] = STATE(3503),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_and] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__dedent] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3504)] = {
    [sym_block_comment] = STATE(3504),
    [sym_compiler_directive_decl] = STATE(3504),
    [sym_fsi_directive_decl] = STATE(3504),
    [sym_preproc_line] = STATE(3504),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_and] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3505)] = {
    [sym_block_comment] = STATE(3505),
    [sym_compiler_directive_decl] = STATE(3505),
    [sym_fsi_directive_decl] = STATE(3505),
    [sym_preproc_line] = STATE(3505),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [anon_sym_inherit] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3506)] = {
    [sym_block_comment] = STATE(3506),
    [sym_compiler_directive_decl] = STATE(3506),
    [sym_fsi_directive_decl] = STATE(3506),
    [sym_preproc_line] = STATE(3506),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_DOT_DOT] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__dedent] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3507)] = {
    [sym_block_comment] = STATE(3507),
    [sym_compiler_directive_decl] = STATE(3507),
    [sym_fsi_directive_decl] = STATE(3507),
    [sym_preproc_line] = STATE(3507),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_DOT_DOT] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__dedent] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3508)] = {
    [sym_block_comment] = STATE(3508),
    [sym_compiler_directive_decl] = STATE(3508),
    [sym_fsi_directive_decl] = STATE(3508),
    [sym_preproc_line] = STATE(3508),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [anon_sym_inherit] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3509)] = {
    [sym_block_comment] = STATE(3509),
    [sym_compiler_directive_decl] = STATE(3509),
    [sym_fsi_directive_decl] = STATE(3509),
    [sym_preproc_line] = STATE(3509),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [anon_sym_inherit] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3510)] = {
    [sym_block_comment] = STATE(3510),
    [sym_compiler_directive_decl] = STATE(3510),
    [sym_fsi_directive_decl] = STATE(3510),
    [sym_preproc_line] = STATE(3510),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_and] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3511)] = {
    [sym_block_comment] = STATE(3511),
    [sym_compiler_directive_decl] = STATE(3511),
    [sym_fsi_directive_decl] = STATE(3511),
    [sym_preproc_line] = STATE(3511),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_and] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3512)] = {
    [sym_block_comment] = STATE(3512),
    [sym_compiler_directive_decl] = STATE(3512),
    [sym_fsi_directive_decl] = STATE(3512),
    [sym_preproc_line] = STATE(3512),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_and] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3513)] = {
    [sym_block_comment] = STATE(3513),
    [sym_compiler_directive_decl] = STATE(3513),
    [sym_fsi_directive_decl] = STATE(3513),
    [sym_preproc_line] = STATE(3513),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_and] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3514)] = {
    [sym_block_comment] = STATE(3514),
    [sym_compiler_directive_decl] = STATE(3514),
    [sym_fsi_directive_decl] = STATE(3514),
    [sym_preproc_line] = STATE(3514),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_and] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [aux_sym_access_modifier_token1] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_default] = ACTIONS(7067),
    [anon_sym_static] = ACTIONS(7067),
    [anon_sym_member] = ACTIONS(7067),
    [anon_sym_interface] = ACTIONS(7067),
    [anon_sym_abstract] = ACTIONS(7067),
    [anon_sym_val] = ACTIONS(7067),
    [anon_sym_override] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3515)] = {
    [sym_block_comment] = STATE(3515),
    [sym_compiler_directive_decl] = STATE(3515),
    [sym_fsi_directive_decl] = STATE(3515),
    [sym_preproc_line] = STATE(3515),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_and] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3516)] = {
    [sym_block_comment] = STATE(3516),
    [sym_compiler_directive_decl] = STATE(3516),
    [sym_fsi_directive_decl] = STATE(3516),
    [sym_preproc_line] = STATE(3516),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_and] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3517)] = {
    [sym_block_comment] = STATE(3517),
    [sym_compiler_directive_decl] = STATE(3517),
    [sym_fsi_directive_decl] = STATE(3517),
    [sym_preproc_line] = STATE(3517),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_and] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3518)] = {
    [sym_block_comment] = STATE(3518),
    [sym_compiler_directive_decl] = STATE(3518),
    [sym_fsi_directive_decl] = STATE(3518),
    [sym_preproc_line] = STATE(3518),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_DOT_DOT] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__dedent] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3519)] = {
    [sym_block_comment] = STATE(3519),
    [sym_compiler_directive_decl] = STATE(3519),
    [sym_fsi_directive_decl] = STATE(3519),
    [sym_preproc_line] = STATE(3519),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_and] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__dedent] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3520)] = {
    [sym_block_comment] = STATE(3520),
    [sym_compiler_directive_decl] = STATE(3520),
    [sym_fsi_directive_decl] = STATE(3520),
    [sym_preproc_line] = STATE(3520),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_DOT_DOT] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__dedent] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(3521)] = {
    [sym_block_comment] = STATE(3521),
    [sym_compiler_directive_decl] = STATE(3521),
    [sym_fsi_directive_decl] = STATE(3521),
    [sym_preproc_line] = STATE(3521),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_and] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [aux_sym_access_modifier_token1] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_default] = ACTIONS(7137),
    [anon_sym_static] = ACTIONS(7137),
    [anon_sym_member] = ACTIONS(7137),
    [anon_sym_interface] = ACTIONS(7137),
    [anon_sym_abstract] = ACTIONS(7137),
    [anon_sym_val] = ACTIONS(7137),
    [anon_sym_override] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(3522)] = {
    [sym_block_comment] = STATE(3522),
    [sym_compiler_directive_decl] = STATE(3522),
    [sym_fsi_directive_decl] = STATE(3522),
    [sym_preproc_line] = STATE(3522),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [anon_sym_inherit] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3523)] = {
    [sym_block_comment] = STATE(3523),
    [sym_compiler_directive_decl] = STATE(3523),
    [sym_fsi_directive_decl] = STATE(3523),
    [sym_preproc_line] = STATE(3523),
    [aux_sym_type_argument_repeat1] = STATE(3550),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(3524)] = {
    [sym_block_comment] = STATE(3524),
    [sym_compiler_directive_decl] = STATE(3524),
    [sym_fsi_directive_decl] = STATE(3524),
    [sym_preproc_line] = STATE(3524),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [anon_sym_inherit] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3525)] = {
    [sym_block_comment] = STATE(3525),
    [sym_compiler_directive_decl] = STATE(3525),
    [sym_fsi_directive_decl] = STATE(3525),
    [sym_preproc_line] = STATE(3525),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_and] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3526)] = {
    [sym_block_comment] = STATE(3526),
    [sym_compiler_directive_decl] = STATE(3526),
    [sym_fsi_directive_decl] = STATE(3526),
    [sym_preproc_line] = STATE(3526),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [anon_sym_inherit] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3527)] = {
    [sym_block_comment] = STATE(3527),
    [sym_compiler_directive_decl] = STATE(3527),
    [sym_fsi_directive_decl] = STATE(3527),
    [sym_preproc_line] = STATE(3527),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3528)] = {
    [sym_block_comment] = STATE(3528),
    [sym_compiler_directive_decl] = STATE(3528),
    [sym_fsi_directive_decl] = STATE(3528),
    [sym_preproc_line] = STATE(3528),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [anon_sym_inherit] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3529)] = {
    [sym_block_comment] = STATE(3529),
    [sym_compiler_directive_decl] = STATE(3529),
    [sym_fsi_directive_decl] = STATE(3529),
    [sym_preproc_line] = STATE(3529),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [anon_sym_inherit] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3530)] = {
    [sym_block_comment] = STATE(3530),
    [sym_compiler_directive_decl] = STATE(3530),
    [sym_fsi_directive_decl] = STATE(3530),
    [sym_preproc_line] = STATE(3530),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [anon_sym_inherit] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3531)] = {
    [sym_block_comment] = STATE(3531),
    [sym_compiler_directive_decl] = STATE(3531),
    [sym_fsi_directive_decl] = STATE(3531),
    [sym_preproc_line] = STATE(3531),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [anon_sym_inherit] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3532)] = {
    [sym_block_comment] = STATE(3532),
    [sym_compiler_directive_decl] = STATE(3532),
    [sym_fsi_directive_decl] = STATE(3532),
    [sym_preproc_line] = STATE(3532),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [anon_sym_inherit] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3533)] = {
    [sym_block_comment] = STATE(3533),
    [sym_compiler_directive_decl] = STATE(3533),
    [sym_fsi_directive_decl] = STATE(3533),
    [sym_preproc_line] = STATE(3533),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [anon_sym_inherit] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3534)] = {
    [sym_block_comment] = STATE(3534),
    [sym_compiler_directive_decl] = STATE(3534),
    [sym_fsi_directive_decl] = STATE(3534),
    [sym_preproc_line] = STATE(3534),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [anon_sym_inherit] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3535)] = {
    [sym_block_comment] = STATE(3535),
    [sym_compiler_directive_decl] = STATE(3535),
    [sym_fsi_directive_decl] = STATE(3535),
    [sym_preproc_line] = STATE(3535),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [aux_sym_access_modifier_token1] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_DOT_DOT] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_default] = ACTIONS(7129),
    [anon_sym_static] = ACTIONS(7129),
    [anon_sym_member] = ACTIONS(7129),
    [anon_sym_interface] = ACTIONS(7129),
    [anon_sym_abstract] = ACTIONS(7129),
    [anon_sym_val] = ACTIONS(7129),
    [anon_sym_override] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__dedent] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3536)] = {
    [sym_block_comment] = STATE(3536),
    [sym_compiler_directive_decl] = STATE(3536),
    [sym_fsi_directive_decl] = STATE(3536),
    [sym_preproc_line] = STATE(3536),
    [aux_sym_sequential_expression_repeat1] = STATE(3353),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [aux_sym_access_modifier_token1] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_default] = ACTIONS(6740),
    [anon_sym_static] = ACTIONS(6740),
    [anon_sym_member] = ACTIONS(6740),
    [anon_sym_interface] = ACTIONS(6740),
    [anon_sym_abstract] = ACTIONS(6740),
    [anon_sym_val] = ACTIONS(6740),
    [anon_sym_override] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3537)] = {
    [sym_block_comment] = STATE(3537),
    [sym_compiler_directive_decl] = STATE(3537),
    [sym_fsi_directive_decl] = STATE(3537),
    [sym_preproc_line] = STATE(3537),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_DOT_DOT] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__dedent] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3538)] = {
    [sym_block_comment] = STATE(3538),
    [sym_compiler_directive_decl] = STATE(3538),
    [sym_fsi_directive_decl] = STATE(3538),
    [sym_preproc_line] = STATE(3538),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [aux_sym_access_modifier_token1] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_DOT_DOT] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_default] = ACTIONS(7133),
    [anon_sym_static] = ACTIONS(7133),
    [anon_sym_member] = ACTIONS(7133),
    [anon_sym_interface] = ACTIONS(7133),
    [anon_sym_abstract] = ACTIONS(7133),
    [anon_sym_val] = ACTIONS(7133),
    [anon_sym_override] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__dedent] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3539)] = {
    [sym_block_comment] = STATE(3539),
    [sym_compiler_directive_decl] = STATE(3539),
    [sym_fsi_directive_decl] = STATE(3539),
    [sym_preproc_line] = STATE(3539),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_DOT_DOT] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__dedent] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3540)] = {
    [sym_block_comment] = STATE(3540),
    [sym_compiler_directive_decl] = STATE(3540),
    [sym_fsi_directive_decl] = STATE(3540),
    [sym_preproc_line] = STATE(3540),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_and] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [aux_sym_access_modifier_token1] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_default] = ACTIONS(7141),
    [anon_sym_static] = ACTIONS(7141),
    [anon_sym_member] = ACTIONS(7141),
    [anon_sym_interface] = ACTIONS(7141),
    [anon_sym_abstract] = ACTIONS(7141),
    [anon_sym_val] = ACTIONS(7141),
    [anon_sym_override] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3541)] = {
    [sym_block_comment] = STATE(3541),
    [sym_compiler_directive_decl] = STATE(3541),
    [sym_fsi_directive_decl] = STATE(3541),
    [sym_preproc_line] = STATE(3541),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_and] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [aux_sym_access_modifier_token1] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_default] = ACTIONS(7145),
    [anon_sym_static] = ACTIONS(7145),
    [anon_sym_member] = ACTIONS(7145),
    [anon_sym_interface] = ACTIONS(7145),
    [anon_sym_abstract] = ACTIONS(7145),
    [anon_sym_val] = ACTIONS(7145),
    [anon_sym_override] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(3542)] = {
    [sym_block_comment] = STATE(3542),
    [sym_compiler_directive_decl] = STATE(3542),
    [sym_fsi_directive_decl] = STATE(3542),
    [sym_preproc_line] = STATE(3542),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_and] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [aux_sym_access_modifier_token1] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_default] = ACTIONS(7155),
    [anon_sym_static] = ACTIONS(7155),
    [anon_sym_member] = ACTIONS(7155),
    [anon_sym_interface] = ACTIONS(7155),
    [anon_sym_abstract] = ACTIONS(7155),
    [anon_sym_val] = ACTIONS(7155),
    [anon_sym_override] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3543)] = {
    [sym_block_comment] = STATE(3543),
    [sym_compiler_directive_decl] = STATE(3543),
    [sym_fsi_directive_decl] = STATE(3543),
    [sym_preproc_line] = STATE(3543),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_and] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [aux_sym_access_modifier_token1] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_default] = ACTIONS(7159),
    [anon_sym_static] = ACTIONS(7159),
    [anon_sym_member] = ACTIONS(7159),
    [anon_sym_interface] = ACTIONS(7159),
    [anon_sym_abstract] = ACTIONS(7159),
    [anon_sym_val] = ACTIONS(7159),
    [anon_sym_override] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(3544)] = {
    [sym_elif_expression] = STATE(3752),
    [sym_block_comment] = STATE(3544),
    [sym_compiler_directive_decl] = STATE(3544),
    [sym_fsi_directive_decl] = STATE(3544),
    [sym_preproc_line] = STATE(3544),
    [aux_sym_if_expression_repeat1] = STATE(3544),
    [ts_builtin_sym_end] = ACTIONS(6494),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_namespace] = ACTIONS(6496),
    [anon_sym_module] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_open] = ACTIONS(6496),
    [anon_sym_type] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8084),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_exception] = ACTIONS(6496),
    [anon_sym_extern] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6494),
    [anon_sym_POUNDload] = ACTIONS(6494),
    [anon_sym_POUNDtime] = ACTIONS(6494),
    [anon_sym_POUNDI] = ACTIONS(6494),
    [anon_sym_POUNDhelp] = ACTIONS(6494),
    [anon_sym_POUNDquit] = ACTIONS(6494),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(3545)] = {
    [sym_block_comment] = STATE(3545),
    [sym_compiler_directive_decl] = STATE(3545),
    [sym_fsi_directive_decl] = STATE(3545),
    [sym_preproc_line] = STATE(3545),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3546)] = {
    [sym_block_comment] = STATE(3546),
    [sym_compiler_directive_decl] = STATE(3546),
    [sym_fsi_directive_decl] = STATE(3546),
    [sym_preproc_line] = STATE(3546),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_and] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [aux_sym_access_modifier_token1] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_default] = ACTIONS(7059),
    [anon_sym_static] = ACTIONS(7059),
    [anon_sym_member] = ACTIONS(7059),
    [anon_sym_interface] = ACTIONS(7059),
    [anon_sym_abstract] = ACTIONS(7059),
    [anon_sym_val] = ACTIONS(7059),
    [anon_sym_override] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3547)] = {
    [sym_block_comment] = STATE(3547),
    [sym_compiler_directive_decl] = STATE(3547),
    [sym_fsi_directive_decl] = STATE(3547),
    [sym_preproc_line] = STATE(3547),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [aux_sym_access_modifier_token1] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_DOT_DOT] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_default] = ACTIONS(7125),
    [anon_sym_static] = ACTIONS(7125),
    [anon_sym_member] = ACTIONS(7125),
    [anon_sym_interface] = ACTIONS(7125),
    [anon_sym_abstract] = ACTIONS(7125),
    [anon_sym_val] = ACTIONS(7125),
    [anon_sym_override] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__dedent] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3548)] = {
    [sym_block_comment] = STATE(3548),
    [sym_compiler_directive_decl] = STATE(3548),
    [sym_fsi_directive_decl] = STATE(3548),
    [sym_preproc_line] = STATE(3548),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_and] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3549)] = {
    [sym_block_comment] = STATE(3549),
    [sym_compiler_directive_decl] = STATE(3549),
    [sym_fsi_directive_decl] = STATE(3549),
    [sym_preproc_line] = STATE(3549),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_DOT_DOT] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__dedent] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3550)] = {
    [sym_block_comment] = STATE(3550),
    [sym_compiler_directive_decl] = STATE(3550),
    [sym_fsi_directive_decl] = STATE(3550),
    [sym_preproc_line] = STATE(3550),
    [aux_sym_type_argument_repeat1] = STATE(3550),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8087),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3551)] = {
    [sym_block_comment] = STATE(3551),
    [sym_compiler_directive_decl] = STATE(3551),
    [sym_fsi_directive_decl] = STATE(3551),
    [sym_preproc_line] = STATE(3551),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [anon_sym_inherit] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3552)] = {
    [sym_block_comment] = STATE(3552),
    [sym_compiler_directive_decl] = STATE(3552),
    [sym_fsi_directive_decl] = STATE(3552),
    [sym_preproc_line] = STATE(3552),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [anon_sym_inherit] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3553)] = {
    [sym_block_comment] = STATE(3553),
    [sym_compiler_directive_decl] = STATE(3553),
    [sym_fsi_directive_decl] = STATE(3553),
    [sym_preproc_line] = STATE(3553),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [aux_sym_access_modifier_token1] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_default] = ACTIONS(6999),
    [anon_sym_static] = ACTIONS(6999),
    [anon_sym_member] = ACTIONS(6999),
    [anon_sym_interface] = ACTIONS(6999),
    [anon_sym_abstract] = ACTIONS(6999),
    [anon_sym_val] = ACTIONS(6999),
    [anon_sym_override] = ACTIONS(6999),
    [anon_sym_inherit] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3554)] = {
    [sym_block_comment] = STATE(3554),
    [sym_compiler_directive_decl] = STATE(3554),
    [sym_fsi_directive_decl] = STATE(3554),
    [sym_preproc_line] = STATE(3554),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_DOT_DOT] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__dedent] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3555)] = {
    [sym_block_comment] = STATE(3555),
    [sym_compiler_directive_decl] = STATE(3555),
    [sym_fsi_directive_decl] = STATE(3555),
    [sym_preproc_line] = STATE(3555),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_DOT_DOT] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__dedent] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3556)] = {
    [sym_block_comment] = STATE(3556),
    [sym_compiler_directive_decl] = STATE(3556),
    [sym_fsi_directive_decl] = STATE(3556),
    [sym_preproc_line] = STATE(3556),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [anon_sym_inherit] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(3557)] = {
    [sym_block_comment] = STATE(3557),
    [sym_compiler_directive_decl] = STATE(3557),
    [sym_fsi_directive_decl] = STATE(3557),
    [sym_preproc_line] = STATE(3557),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_and] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3558)] = {
    [sym_block_comment] = STATE(3558),
    [sym_compiler_directive_decl] = STATE(3558),
    [sym_fsi_directive_decl] = STATE(3558),
    [sym_preproc_line] = STATE(3558),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__dedent] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3559)] = {
    [sym_block_comment] = STATE(3559),
    [sym_compiler_directive_decl] = STATE(3559),
    [sym_fsi_directive_decl] = STATE(3559),
    [sym_preproc_line] = STATE(3559),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_and] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [aux_sym_access_modifier_token1] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_default] = ACTIONS(7003),
    [anon_sym_static] = ACTIONS(7003),
    [anon_sym_member] = ACTIONS(7003),
    [anon_sym_interface] = ACTIONS(7003),
    [anon_sym_abstract] = ACTIONS(7003),
    [anon_sym_val] = ACTIONS(7003),
    [anon_sym_override] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(3560)] = {
    [sym_block_comment] = STATE(3560),
    [sym_compiler_directive_decl] = STATE(3560),
    [sym_fsi_directive_decl] = STATE(3560),
    [sym_preproc_line] = STATE(3560),
    [ts_builtin_sym_end] = ACTIONS(6371),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_namespace] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3561)] = {
    [sym_block_comment] = STATE(3561),
    [sym_compiler_directive_decl] = STATE(3561),
    [sym_fsi_directive_decl] = STATE(3561),
    [sym_preproc_line] = STATE(3561),
    [ts_builtin_sym_end] = ACTIONS(6371),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_namespace] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3562)] = {
    [sym_block_comment] = STATE(3562),
    [sym_compiler_directive_decl] = STATE(3562),
    [sym_fsi_directive_decl] = STATE(3562),
    [sym_preproc_line] = STATE(3562),
    [ts_builtin_sym_end] = ACTIONS(6375),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_namespace] = ACTIONS(6373),
    [anon_sym_module] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_open] = ACTIONS(6373),
    [anon_sym_type] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8090),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_exception] = ACTIONS(6373),
    [anon_sym_extern] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6375),
    [anon_sym_POUNDload] = ACTIONS(6375),
    [anon_sym_POUNDtime] = ACTIONS(6375),
    [anon_sym_POUNDI] = ACTIONS(6375),
    [anon_sym_POUNDhelp] = ACTIONS(6375),
    [anon_sym_POUNDquit] = ACTIONS(6375),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(3563)] = {
    [sym_block_comment] = STATE(3563),
    [sym_compiler_directive_decl] = STATE(3563),
    [sym_fsi_directive_decl] = STATE(3563),
    [sym_preproc_line] = STATE(3563),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3564)] = {
    [sym_block_comment] = STATE(3564),
    [sym_compiler_directive_decl] = STATE(3564),
    [sym_fsi_directive_decl] = STATE(3564),
    [sym_preproc_line] = STATE(3564),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [aux_sym_access_modifier_token1] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_default] = ACTIONS(7007),
    [anon_sym_static] = ACTIONS(7007),
    [anon_sym_member] = ACTIONS(7007),
    [anon_sym_interface] = ACTIONS(7007),
    [anon_sym_abstract] = ACTIONS(7007),
    [anon_sym_val] = ACTIONS(7007),
    [anon_sym_override] = ACTIONS(7007),
    [anon_sym_inherit] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3565)] = {
    [sym_block_comment] = STATE(3565),
    [sym_compiler_directive_decl] = STATE(3565),
    [sym_fsi_directive_decl] = STATE(3565),
    [sym_preproc_line] = STATE(3565),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_DOT_DOT] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__dedent] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3566)] = {
    [sym_block_comment] = STATE(3566),
    [sym_compiler_directive_decl] = STATE(3566),
    [sym_fsi_directive_decl] = STATE(3566),
    [sym_preproc_line] = STATE(3566),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_and] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__dedent] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3567)] = {
    [sym_block_comment] = STATE(3567),
    [sym_compiler_directive_decl] = STATE(3567),
    [sym_fsi_directive_decl] = STATE(3567),
    [sym_preproc_line] = STATE(3567),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3568)] = {
    [sym_block_comment] = STATE(3568),
    [sym_compiler_directive_decl] = STATE(3568),
    [sym_fsi_directive_decl] = STATE(3568),
    [sym_preproc_line] = STATE(3568),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_and] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [aux_sym_access_modifier_token1] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_default] = ACTIONS(6807),
    [anon_sym_static] = ACTIONS(6807),
    [anon_sym_member] = ACTIONS(6807),
    [anon_sym_interface] = ACTIONS(6807),
    [anon_sym_abstract] = ACTIONS(6807),
    [anon_sym_val] = ACTIONS(6807),
    [anon_sym_override] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3569)] = {
    [sym_block_comment] = STATE(3569),
    [sym_compiler_directive_decl] = STATE(3569),
    [sym_fsi_directive_decl] = STATE(3569),
    [sym_preproc_line] = STATE(3569),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_and] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__dedent] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3570)] = {
    [sym_block_comment] = STATE(3570),
    [sym_compiler_directive_decl] = STATE(3570),
    [sym_fsi_directive_decl] = STATE(3570),
    [sym_preproc_line] = STATE(3570),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_DOT_DOT] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__dedent] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3571)] = {
    [sym_block_comment] = STATE(3571),
    [sym_compiler_directive_decl] = STATE(3571),
    [sym_fsi_directive_decl] = STATE(3571),
    [sym_preproc_line] = STATE(3571),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3572)] = {
    [sym_block_comment] = STATE(3572),
    [sym_compiler_directive_decl] = STATE(3572),
    [sym_fsi_directive_decl] = STATE(3572),
    [sym_preproc_line] = STATE(3572),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_module] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_open] = ACTIONS(6373),
    [anon_sym_type] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8092),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_exception] = ACTIONS(6373),
    [anon_sym_extern] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6375),
    [anon_sym_POUNDload] = ACTIONS(6375),
    [anon_sym_POUNDtime] = ACTIONS(6375),
    [anon_sym_POUNDI] = ACTIONS(6375),
    [anon_sym_POUNDhelp] = ACTIONS(6375),
    [anon_sym_POUNDquit] = ACTIONS(6375),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [anon_sym_POUNDelse] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(3573)] = {
    [sym_block_comment] = STATE(3573),
    [sym_compiler_directive_decl] = STATE(3573),
    [sym_fsi_directive_decl] = STATE(3573),
    [sym_preproc_line] = STATE(3573),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_and] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [aux_sym_access_modifier_token1] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_default] = ACTIONS(6892),
    [anon_sym_static] = ACTIONS(6892),
    [anon_sym_member] = ACTIONS(6892),
    [anon_sym_interface] = ACTIONS(6892),
    [anon_sym_abstract] = ACTIONS(6892),
    [anon_sym_val] = ACTIONS(6892),
    [anon_sym_override] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3574)] = {
    [sym_block_comment] = STATE(3574),
    [sym_compiler_directive_decl] = STATE(3574),
    [sym_fsi_directive_decl] = STATE(3574),
    [sym_preproc_line] = STATE(3574),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_module] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_open] = ACTIONS(6367),
    [anon_sym_type] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_exception] = ACTIONS(6367),
    [anon_sym_extern] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6365),
    [anon_sym_POUNDload] = ACTIONS(6365),
    [anon_sym_POUNDtime] = ACTIONS(6365),
    [anon_sym_POUNDI] = ACTIONS(6365),
    [anon_sym_POUNDhelp] = ACTIONS(6365),
    [anon_sym_POUNDquit] = ACTIONS(6365),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [anon_sym_POUNDelse] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(3575)] = {
    [sym_block_comment] = STATE(3575),
    [sym_compiler_directive_decl] = STATE(3575),
    [sym_fsi_directive_decl] = STATE(3575),
    [sym_preproc_line] = STATE(3575),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [anon_sym_inherit] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(3576)] = {
    [sym_block_comment] = STATE(3576),
    [sym_compiler_directive_decl] = STATE(3576),
    [sym_fsi_directive_decl] = STATE(3576),
    [sym_preproc_line] = STATE(3576),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_and] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__dedent] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3577)] = {
    [sym_block_comment] = STATE(3577),
    [sym_compiler_directive_decl] = STATE(3577),
    [sym_fsi_directive_decl] = STATE(3577),
    [sym_preproc_line] = STATE(3577),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_and] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__dedent] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3578)] = {
    [sym_block_comment] = STATE(3578),
    [sym_compiler_directive_decl] = STATE(3578),
    [sym_fsi_directive_decl] = STATE(3578),
    [sym_preproc_line] = STATE(3578),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_module] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_open] = ACTIONS(6379),
    [anon_sym_type] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_exception] = ACTIONS(6379),
    [anon_sym_extern] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6381),
    [anon_sym_POUNDload] = ACTIONS(6381),
    [anon_sym_POUNDtime] = ACTIONS(6381),
    [anon_sym_POUNDI] = ACTIONS(6381),
    [anon_sym_POUNDhelp] = ACTIONS(6381),
    [anon_sym_POUNDquit] = ACTIONS(6381),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [anon_sym_POUNDelse] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(3579)] = {
    [sym_block_comment] = STATE(3579),
    [sym_compiler_directive_decl] = STATE(3579),
    [sym_fsi_directive_decl] = STATE(3579),
    [sym_preproc_line] = STATE(3579),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_and] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [aux_sym_access_modifier_token1] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_default] = ACTIONS(6896),
    [anon_sym_static] = ACTIONS(6896),
    [anon_sym_member] = ACTIONS(6896),
    [anon_sym_interface] = ACTIONS(6896),
    [anon_sym_abstract] = ACTIONS(6896),
    [anon_sym_val] = ACTIONS(6896),
    [anon_sym_override] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3580)] = {
    [sym_block_comment] = STATE(3580),
    [sym_compiler_directive_decl] = STATE(3580),
    [sym_fsi_directive_decl] = STATE(3580),
    [sym_preproc_line] = STATE(3580),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_and] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [aux_sym_access_modifier_token1] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_default] = ACTIONS(6900),
    [anon_sym_static] = ACTIONS(6900),
    [anon_sym_member] = ACTIONS(6900),
    [anon_sym_interface] = ACTIONS(6900),
    [anon_sym_abstract] = ACTIONS(6900),
    [anon_sym_val] = ACTIONS(6900),
    [anon_sym_override] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3581)] = {
    [sym_block_comment] = STATE(3581),
    [sym_compiler_directive_decl] = STATE(3581),
    [sym_fsi_directive_decl] = STATE(3581),
    [sym_preproc_line] = STATE(3581),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_module] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_open] = ACTIONS(6383),
    [anon_sym_type] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_exception] = ACTIONS(6383),
    [anon_sym_extern] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6385),
    [anon_sym_POUNDload] = ACTIONS(6385),
    [anon_sym_POUNDtime] = ACTIONS(6385),
    [anon_sym_POUNDI] = ACTIONS(6385),
    [anon_sym_POUNDhelp] = ACTIONS(6385),
    [anon_sym_POUNDquit] = ACTIONS(6385),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [anon_sym_POUNDelse] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(3582)] = {
    [sym_block_comment] = STATE(3582),
    [sym_compiler_directive_decl] = STATE(3582),
    [sym_fsi_directive_decl] = STATE(3582),
    [sym_preproc_line] = STATE(3582),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_module] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_open] = ACTIONS(6391),
    [anon_sym_type] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_exception] = ACTIONS(6391),
    [anon_sym_extern] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6393),
    [anon_sym_POUNDload] = ACTIONS(6393),
    [anon_sym_POUNDtime] = ACTIONS(6393),
    [anon_sym_POUNDI] = ACTIONS(6393),
    [anon_sym_POUNDhelp] = ACTIONS(6393),
    [anon_sym_POUNDquit] = ACTIONS(6393),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [anon_sym_POUNDelse] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(3583)] = {
    [sym_block_comment] = STATE(3583),
    [sym_compiler_directive_decl] = STATE(3583),
    [sym_fsi_directive_decl] = STATE(3583),
    [sym_preproc_line] = STATE(3583),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_module] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_open] = ACTIONS(6395),
    [anon_sym_type] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_exception] = ACTIONS(6395),
    [anon_sym_extern] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6397),
    [anon_sym_POUNDload] = ACTIONS(6397),
    [anon_sym_POUNDtime] = ACTIONS(6397),
    [anon_sym_POUNDI] = ACTIONS(6397),
    [anon_sym_POUNDhelp] = ACTIONS(6397),
    [anon_sym_POUNDquit] = ACTIONS(6397),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [anon_sym_POUNDelse] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(3584)] = {
    [sym_block_comment] = STATE(3584),
    [sym_compiler_directive_decl] = STATE(3584),
    [sym_fsi_directive_decl] = STATE(3584),
    [sym_preproc_line] = STATE(3584),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_and] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3585)] = {
    [sym_block_comment] = STATE(3585),
    [sym_compiler_directive_decl] = STATE(3585),
    [sym_fsi_directive_decl] = STATE(3585),
    [sym_preproc_line] = STATE(3585),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_and] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__dedent] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3586)] = {
    [sym_block_comment] = STATE(3586),
    [sym_compiler_directive_decl] = STATE(3586),
    [sym_fsi_directive_decl] = STATE(3586),
    [sym_preproc_line] = STATE(3586),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_DOT_DOT] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__dedent] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3587)] = {
    [sym_block_comment] = STATE(3587),
    [sym_compiler_directive_decl] = STATE(3587),
    [sym_fsi_directive_decl] = STATE(3587),
    [sym_preproc_line] = STATE(3587),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_and] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3588)] = {
    [sym_block_comment] = STATE(3588),
    [sym_compiler_directive_decl] = STATE(3588),
    [sym_fsi_directive_decl] = STATE(3588),
    [sym_preproc_line] = STATE(3588),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_module] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_open] = ACTIONS(6399),
    [anon_sym_type] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_exception] = ACTIONS(6399),
    [anon_sym_extern] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6401),
    [anon_sym_POUNDload] = ACTIONS(6401),
    [anon_sym_POUNDtime] = ACTIONS(6401),
    [anon_sym_POUNDI] = ACTIONS(6401),
    [anon_sym_POUNDhelp] = ACTIONS(6401),
    [anon_sym_POUNDquit] = ACTIONS(6401),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [anon_sym_POUNDelse] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(3589)] = {
    [sym_block_comment] = STATE(3589),
    [sym_compiler_directive_decl] = STATE(3589),
    [sym_fsi_directive_decl] = STATE(3589),
    [sym_preproc_line] = STATE(3589),
    [ts_builtin_sym_end] = ACTIONS(6418),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_namespace] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(3590)] = {
    [sym_block_comment] = STATE(3590),
    [sym_compiler_directive_decl] = STATE(3590),
    [sym_fsi_directive_decl] = STATE(3590),
    [sym_preproc_line] = STATE(3590),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3591)] = {
    [sym_block_comment] = STATE(3591),
    [sym_compiler_directive_decl] = STATE(3591),
    [sym_fsi_directive_decl] = STATE(3591),
    [sym_preproc_line] = STATE(3591),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [aux_sym_access_modifier_token1] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_DOT_DOT] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_default] = ACTIONS(7053),
    [anon_sym_static] = ACTIONS(7053),
    [anon_sym_member] = ACTIONS(7053),
    [anon_sym_interface] = ACTIONS(7053),
    [anon_sym_abstract] = ACTIONS(7053),
    [anon_sym_val] = ACTIONS(7053),
    [anon_sym_override] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__dedent] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3592)] = {
    [sym_block_comment] = STATE(3592),
    [sym_compiler_directive_decl] = STATE(3592),
    [sym_fsi_directive_decl] = STATE(3592),
    [sym_preproc_line] = STATE(3592),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_and] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(3593)] = {
    [sym_block_comment] = STATE(3593),
    [sym_compiler_directive_decl] = STATE(3593),
    [sym_fsi_directive_decl] = STATE(3593),
    [sym_preproc_line] = STATE(3593),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_and] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(3594)] = {
    [sym_block_comment] = STATE(3594),
    [sym_compiler_directive_decl] = STATE(3594),
    [sym_fsi_directive_decl] = STATE(3594),
    [sym_preproc_line] = STATE(3594),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_and] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [aux_sym_access_modifier_token1] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_default] = ACTIONS(6906),
    [anon_sym_static] = ACTIONS(6906),
    [anon_sym_member] = ACTIONS(6906),
    [anon_sym_interface] = ACTIONS(6906),
    [anon_sym_abstract] = ACTIONS(6906),
    [anon_sym_val] = ACTIONS(6906),
    [anon_sym_override] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__dedent] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3595)] = {
    [sym_block_comment] = STATE(3595),
    [sym_compiler_directive_decl] = STATE(3595),
    [sym_fsi_directive_decl] = STATE(3595),
    [sym_preproc_line] = STATE(3595),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_and] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3596)] = {
    [sym_block_comment] = STATE(3596),
    [sym_compiler_directive_decl] = STATE(3596),
    [sym_fsi_directive_decl] = STATE(3596),
    [sym_preproc_line] = STATE(3596),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_and] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3597)] = {
    [sym_block_comment] = STATE(3597),
    [sym_compiler_directive_decl] = STATE(3597),
    [sym_fsi_directive_decl] = STATE(3597),
    [sym_preproc_line] = STATE(3597),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [aux_sym_access_modifier_token1] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_default] = ACTIONS(5951),
    [anon_sym_static] = ACTIONS(5951),
    [anon_sym_member] = ACTIONS(5951),
    [anon_sym_interface] = ACTIONS(5951),
    [anon_sym_abstract] = ACTIONS(5951),
    [anon_sym_val] = ACTIONS(5951),
    [anon_sym_override] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3598)] = {
    [sym_block_comment] = STATE(3598),
    [sym_compiler_directive_decl] = STATE(3598),
    [sym_fsi_directive_decl] = STATE(3598),
    [sym_preproc_line] = STATE(3598),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_module] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_open] = ACTIONS(6403),
    [anon_sym_type] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_exception] = ACTIONS(6403),
    [anon_sym_extern] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6405),
    [anon_sym_POUNDload] = ACTIONS(6405),
    [anon_sym_POUNDtime] = ACTIONS(6405),
    [anon_sym_POUNDI] = ACTIONS(6405),
    [anon_sym_POUNDhelp] = ACTIONS(6405),
    [anon_sym_POUNDquit] = ACTIONS(6405),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [anon_sym_POUNDelse] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(3599)] = {
    [sym_block_comment] = STATE(3599),
    [sym_compiler_directive_decl] = STATE(3599),
    [sym_fsi_directive_decl] = STATE(3599),
    [sym_preproc_line] = STATE(3599),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_and] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [aux_sym_access_modifier_token1] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_default] = ACTIONS(6995),
    [anon_sym_static] = ACTIONS(6995),
    [anon_sym_member] = ACTIONS(6995),
    [anon_sym_interface] = ACTIONS(6995),
    [anon_sym_abstract] = ACTIONS(6995),
    [anon_sym_val] = ACTIONS(6995),
    [anon_sym_override] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3600)] = {
    [sym_block_comment] = STATE(3600),
    [sym_compiler_directive_decl] = STATE(3600),
    [sym_fsi_directive_decl] = STATE(3600),
    [sym_preproc_line] = STATE(3600),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_module] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_open] = ACTIONS(6411),
    [anon_sym_type] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_exception] = ACTIONS(6411),
    [anon_sym_extern] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6413),
    [anon_sym_POUNDload] = ACTIONS(6413),
    [anon_sym_POUNDtime] = ACTIONS(6413),
    [anon_sym_POUNDI] = ACTIONS(6413),
    [anon_sym_POUNDhelp] = ACTIONS(6413),
    [anon_sym_POUNDquit] = ACTIONS(6413),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [anon_sym_POUNDelse] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(3601)] = {
    [sym_block_comment] = STATE(3601),
    [sym_compiler_directive_decl] = STATE(3601),
    [sym_fsi_directive_decl] = STATE(3601),
    [sym_preproc_line] = STATE(3601),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3602)] = {
    [sym_block_comment] = STATE(3602),
    [sym_compiler_directive_decl] = STATE(3602),
    [sym_fsi_directive_decl] = STATE(3602),
    [sym_preproc_line] = STATE(3602),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_and] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(3603)] = {
    [sym_block_comment] = STATE(3603),
    [sym_compiler_directive_decl] = STATE(3603),
    [sym_fsi_directive_decl] = STATE(3603),
    [sym_preproc_line] = STATE(3603),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_and] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__dedent] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(3604)] = {
    [sym_block_comment] = STATE(3604),
    [sym_compiler_directive_decl] = STATE(3604),
    [sym_fsi_directive_decl] = STATE(3604),
    [sym_preproc_line] = STATE(3604),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_and] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [aux_sym_access_modifier_token1] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_default] = ACTIONS(7027),
    [anon_sym_static] = ACTIONS(7027),
    [anon_sym_member] = ACTIONS(7027),
    [anon_sym_interface] = ACTIONS(7027),
    [anon_sym_abstract] = ACTIONS(7027),
    [anon_sym_val] = ACTIONS(7027),
    [anon_sym_override] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3605)] = {
    [sym_block_comment] = STATE(3605),
    [sym_compiler_directive_decl] = STATE(3605),
    [sym_fsi_directive_decl] = STATE(3605),
    [sym_preproc_line] = STATE(3605),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_and] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__dedent] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(3606)] = {
    [sym_block_comment] = STATE(3606),
    [sym_compiler_directive_decl] = STATE(3606),
    [sym_fsi_directive_decl] = STATE(3606),
    [sym_preproc_line] = STATE(3606),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [aux_sym_access_modifier_token1] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_DOT_DOT] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_default] = ACTIONS(6913),
    [anon_sym_static] = ACTIONS(6913),
    [anon_sym_member] = ACTIONS(6913),
    [anon_sym_interface] = ACTIONS(6913),
    [anon_sym_abstract] = ACTIONS(6913),
    [anon_sym_val] = ACTIONS(6913),
    [anon_sym_override] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__dedent] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(3607)] = {
    [sym_block_comment] = STATE(3607),
    [sym_compiler_directive_decl] = STATE(3607),
    [sym_fsi_directive_decl] = STATE(3607),
    [sym_preproc_line] = STATE(3607),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [aux_sym_access_modifier_token1] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_DOT_DOT] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_default] = ACTIONS(6917),
    [anon_sym_static] = ACTIONS(6917),
    [anon_sym_member] = ACTIONS(6917),
    [anon_sym_interface] = ACTIONS(6917),
    [anon_sym_abstract] = ACTIONS(6917),
    [anon_sym_val] = ACTIONS(6917),
    [anon_sym_override] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__dedent] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(3608)] = {
    [sym_block_comment] = STATE(3608),
    [sym_compiler_directive_decl] = STATE(3608),
    [sym_fsi_directive_decl] = STATE(3608),
    [sym_preproc_line] = STATE(3608),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_and] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__dedent] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3609)] = {
    [sym_block_comment] = STATE(3609),
    [sym_compiler_directive_decl] = STATE(3609),
    [sym_fsi_directive_decl] = STATE(3609),
    [sym_preproc_line] = STATE(3609),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_and] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__dedent] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3610)] = {
    [sym_block_comment] = STATE(3610),
    [sym_compiler_directive_decl] = STATE(3610),
    [sym_fsi_directive_decl] = STATE(3610),
    [sym_preproc_line] = STATE(3610),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [aux_sym_access_modifier_token1] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_DOT_DOT] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_default] = ACTIONS(6921),
    [anon_sym_static] = ACTIONS(6921),
    [anon_sym_member] = ACTIONS(6921),
    [anon_sym_interface] = ACTIONS(6921),
    [anon_sym_abstract] = ACTIONS(6921),
    [anon_sym_val] = ACTIONS(6921),
    [anon_sym_override] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__dedent] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3611)] = {
    [sym_block_comment] = STATE(3611),
    [sym_compiler_directive_decl] = STATE(3611),
    [sym_fsi_directive_decl] = STATE(3611),
    [sym_preproc_line] = STATE(3611),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [aux_sym_access_modifier_token1] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_default] = ACTIONS(7019),
    [anon_sym_static] = ACTIONS(7019),
    [anon_sym_member] = ACTIONS(7019),
    [anon_sym_interface] = ACTIONS(7019),
    [anon_sym_abstract] = ACTIONS(7019),
    [anon_sym_val] = ACTIONS(7019),
    [anon_sym_override] = ACTIONS(7019),
    [anon_sym_inherit] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3612)] = {
    [sym_block_comment] = STATE(3612),
    [sym_compiler_directive_decl] = STATE(3612),
    [sym_fsi_directive_decl] = STATE(3612),
    [sym_preproc_line] = STATE(3612),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_and] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3613)] = {
    [sym_block_comment] = STATE(3613),
    [sym_compiler_directive_decl] = STATE(3613),
    [sym_fsi_directive_decl] = STATE(3613),
    [sym_preproc_line] = STATE(3613),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_and] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3614)] = {
    [sym_block_comment] = STATE(3614),
    [sym_compiler_directive_decl] = STATE(3614),
    [sym_fsi_directive_decl] = STATE(3614),
    [sym_preproc_line] = STATE(3614),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_module] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_open] = ACTIONS(6492),
    [anon_sym_type] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_exception] = ACTIONS(6492),
    [anon_sym_extern] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6490),
    [anon_sym_POUNDload] = ACTIONS(6490),
    [anon_sym_POUNDtime] = ACTIONS(6490),
    [anon_sym_POUNDI] = ACTIONS(6490),
    [anon_sym_POUNDhelp] = ACTIONS(6490),
    [anon_sym_POUNDquit] = ACTIONS(6490),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [anon_sym_POUNDelse] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(3615)] = {
    [sym_block_comment] = STATE(3615),
    [sym_compiler_directive_decl] = STATE(3615),
    [sym_fsi_directive_decl] = STATE(3615),
    [sym_preproc_line] = STATE(3615),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_and] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3616)] = {
    [sym_block_comment] = STATE(3616),
    [sym_compiler_directive_decl] = STATE(3616),
    [sym_fsi_directive_decl] = STATE(3616),
    [sym_preproc_line] = STATE(3616),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_DOT_DOT] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__dedent] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3617)] = {
    [sym_block_comment] = STATE(3617),
    [sym_compiler_directive_decl] = STATE(3617),
    [sym_fsi_directive_decl] = STATE(3617),
    [sym_preproc_line] = STATE(3617),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_and] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3618)] = {
    [sym_block_comment] = STATE(3618),
    [sym_compiler_directive_decl] = STATE(3618),
    [sym_fsi_directive_decl] = STATE(3618),
    [sym_preproc_line] = STATE(3618),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_and] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3619)] = {
    [sym_block_comment] = STATE(3619),
    [sym_compiler_directive_decl] = STATE(3619),
    [sym_fsi_directive_decl] = STATE(3619),
    [sym_preproc_line] = STATE(3619),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_module] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_open] = ACTIONS(6427),
    [anon_sym_type] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_exception] = ACTIONS(6427),
    [anon_sym_extern] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6429),
    [anon_sym_POUNDload] = ACTIONS(6429),
    [anon_sym_POUNDtime] = ACTIONS(6429),
    [anon_sym_POUNDI] = ACTIONS(6429),
    [anon_sym_POUNDhelp] = ACTIONS(6429),
    [anon_sym_POUNDquit] = ACTIONS(6429),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [anon_sym_POUNDelse] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(3620)] = {
    [sym_block_comment] = STATE(3620),
    [sym_compiler_directive_decl] = STATE(3620),
    [sym_fsi_directive_decl] = STATE(3620),
    [sym_preproc_line] = STATE(3620),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_module] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_open] = ACTIONS(6389),
    [anon_sym_type] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_exception] = ACTIONS(6389),
    [anon_sym_extern] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6387),
    [anon_sym_POUNDload] = ACTIONS(6387),
    [anon_sym_POUNDtime] = ACTIONS(6387),
    [anon_sym_POUNDI] = ACTIONS(6387),
    [anon_sym_POUNDhelp] = ACTIONS(6387),
    [anon_sym_POUNDquit] = ACTIONS(6387),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [anon_sym_POUNDelse] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(3621)] = {
    [sym_block_comment] = STATE(3621),
    [sym_compiler_directive_decl] = STATE(3621),
    [sym_fsi_directive_decl] = STATE(3621),
    [sym_preproc_line] = STATE(3621),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_and] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3622)] = {
    [sym_block_comment] = STATE(3622),
    [sym_compiler_directive_decl] = STATE(3622),
    [sym_fsi_directive_decl] = STATE(3622),
    [sym_preproc_line] = STATE(3622),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_DOT_DOT] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__dedent] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3623)] = {
    [sym_block_comment] = STATE(3623),
    [sym_compiler_directive_decl] = STATE(3623),
    [sym_fsi_directive_decl] = STATE(3623),
    [sym_preproc_line] = STATE(3623),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3624)] = {
    [sym_block_comment] = STATE(3624),
    [sym_compiler_directive_decl] = STATE(3624),
    [sym_fsi_directive_decl] = STATE(3624),
    [sym_preproc_line] = STATE(3624),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(3625)] = {
    [sym_block_comment] = STATE(3625),
    [sym_compiler_directive_decl] = STATE(3625),
    [sym_fsi_directive_decl] = STATE(3625),
    [sym_preproc_line] = STATE(3625),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [aux_sym_access_modifier_token1] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_DOT_DOT] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_default] = ACTIONS(6925),
    [anon_sym_static] = ACTIONS(6925),
    [anon_sym_member] = ACTIONS(6925),
    [anon_sym_interface] = ACTIONS(6925),
    [anon_sym_abstract] = ACTIONS(6925),
    [anon_sym_val] = ACTIONS(6925),
    [anon_sym_override] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__dedent] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3626)] = {
    [sym_block_comment] = STATE(3626),
    [sym_compiler_directive_decl] = STATE(3626),
    [sym_fsi_directive_decl] = STATE(3626),
    [sym_preproc_line] = STATE(3626),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_and] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [aux_sym_access_modifier_token1] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_default] = ACTIONS(6967),
    [anon_sym_static] = ACTIONS(6967),
    [anon_sym_member] = ACTIONS(6967),
    [anon_sym_interface] = ACTIONS(6967),
    [anon_sym_abstract] = ACTIONS(6967),
    [anon_sym_val] = ACTIONS(6967),
    [anon_sym_override] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3627)] = {
    [sym_block_comment] = STATE(3627),
    [sym_compiler_directive_decl] = STATE(3627),
    [sym_fsi_directive_decl] = STATE(3627),
    [sym_preproc_line] = STATE(3627),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_and] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [aux_sym_access_modifier_token1] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_default] = ACTIONS(6971),
    [anon_sym_static] = ACTIONS(6971),
    [anon_sym_member] = ACTIONS(6971),
    [anon_sym_interface] = ACTIONS(6971),
    [anon_sym_abstract] = ACTIONS(6971),
    [anon_sym_val] = ACTIONS(6971),
    [anon_sym_override] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3628)] = {
    [sym_block_comment] = STATE(3628),
    [sym_compiler_directive_decl] = STATE(3628),
    [sym_fsi_directive_decl] = STATE(3628),
    [sym_preproc_line] = STATE(3628),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_DOT_DOT] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__dedent] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(3629)] = {
    [sym_block_comment] = STATE(3629),
    [sym_compiler_directive_decl] = STATE(3629),
    [sym_fsi_directive_decl] = STATE(3629),
    [sym_preproc_line] = STATE(3629),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [aux_sym_access_modifier_token1] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_default] = ACTIONS(290),
    [anon_sym_static] = ACTIONS(290),
    [anon_sym_member] = ACTIONS(290),
    [anon_sym_interface] = ACTIONS(290),
    [anon_sym_abstract] = ACTIONS(290),
    [anon_sym_val] = ACTIONS(290),
    [anon_sym_override] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3630)] = {
    [sym_block_comment] = STATE(3630),
    [sym_compiler_directive_decl] = STATE(3630),
    [sym_fsi_directive_decl] = STATE(3630),
    [sym_preproc_line] = STATE(3630),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_and] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3631)] = {
    [sym_block_comment] = STATE(3631),
    [sym_compiler_directive_decl] = STATE(3631),
    [sym_fsi_directive_decl] = STATE(3631),
    [sym_preproc_line] = STATE(3631),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_and] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__dedent] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(3632)] = {
    [sym_block_comment] = STATE(3632),
    [sym_compiler_directive_decl] = STATE(3632),
    [sym_fsi_directive_decl] = STATE(3632),
    [sym_preproc_line] = STATE(3632),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_and] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [aux_sym_access_modifier_token1] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_default] = ACTIONS(6979),
    [anon_sym_static] = ACTIONS(6979),
    [anon_sym_member] = ACTIONS(6979),
    [anon_sym_interface] = ACTIONS(6979),
    [anon_sym_abstract] = ACTIONS(6979),
    [anon_sym_val] = ACTIONS(6979),
    [anon_sym_override] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3633)] = {
    [sym_block_comment] = STATE(3633),
    [sym_compiler_directive_decl] = STATE(3633),
    [sym_fsi_directive_decl] = STATE(3633),
    [sym_preproc_line] = STATE(3633),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_and] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [aux_sym_access_modifier_token1] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_default] = ACTIONS(6983),
    [anon_sym_static] = ACTIONS(6983),
    [anon_sym_member] = ACTIONS(6983),
    [anon_sym_interface] = ACTIONS(6983),
    [anon_sym_abstract] = ACTIONS(6983),
    [anon_sym_val] = ACTIONS(6983),
    [anon_sym_override] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3634)] = {
    [sym_block_comment] = STATE(3634),
    [sym_compiler_directive_decl] = STATE(3634),
    [sym_fsi_directive_decl] = STATE(3634),
    [sym_preproc_line] = STATE(3634),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_and] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [aux_sym_access_modifier_token1] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_default] = ACTIONS(6987),
    [anon_sym_static] = ACTIONS(6987),
    [anon_sym_member] = ACTIONS(6987),
    [anon_sym_interface] = ACTIONS(6987),
    [anon_sym_abstract] = ACTIONS(6987),
    [anon_sym_val] = ACTIONS(6987),
    [anon_sym_override] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3635)] = {
    [sym_block_comment] = STATE(3635),
    [sym_compiler_directive_decl] = STATE(3635),
    [sym_fsi_directive_decl] = STATE(3635),
    [sym_preproc_line] = STATE(3635),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_and] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [aux_sym_access_modifier_token1] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_default] = ACTIONS(6991),
    [anon_sym_static] = ACTIONS(6991),
    [anon_sym_member] = ACTIONS(6991),
    [anon_sym_interface] = ACTIONS(6991),
    [anon_sym_abstract] = ACTIONS(6991),
    [anon_sym_val] = ACTIONS(6991),
    [anon_sym_override] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3636)] = {
    [sym_block_comment] = STATE(3636),
    [sym_compiler_directive_decl] = STATE(3636),
    [sym_fsi_directive_decl] = STATE(3636),
    [sym_preproc_line] = STATE(3636),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_and] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [aux_sym_access_modifier_token1] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_default] = ACTIONS(6975),
    [anon_sym_static] = ACTIONS(6975),
    [anon_sym_member] = ACTIONS(6975),
    [anon_sym_interface] = ACTIONS(6975),
    [anon_sym_abstract] = ACTIONS(6975),
    [anon_sym_val] = ACTIONS(6975),
    [anon_sym_override] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__dedent] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3637)] = {
    [sym_block_comment] = STATE(3637),
    [sym_compiler_directive_decl] = STATE(3637),
    [sym_fsi_directive_decl] = STATE(3637),
    [sym_preproc_line] = STATE(3637),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3638)] = {
    [sym_block_comment] = STATE(3638),
    [sym_compiler_directive_decl] = STATE(3638),
    [sym_fsi_directive_decl] = STATE(3638),
    [sym_preproc_line] = STATE(3638),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_and] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [aux_sym_access_modifier_token1] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_default] = ACTIONS(6888),
    [anon_sym_static] = ACTIONS(6888),
    [anon_sym_member] = ACTIONS(6888),
    [anon_sym_interface] = ACTIONS(6888),
    [anon_sym_abstract] = ACTIONS(6888),
    [anon_sym_val] = ACTIONS(6888),
    [anon_sym_override] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__dedent] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3639)] = {
    [sym_block_comment] = STATE(3639),
    [sym_compiler_directive_decl] = STATE(3639),
    [sym_fsi_directive_decl] = STATE(3639),
    [sym_preproc_line] = STATE(3639),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_and] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [aux_sym_access_modifier_token1] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_default] = ACTIONS(7069),
    [anon_sym_static] = ACTIONS(7069),
    [anon_sym_member] = ACTIONS(7069),
    [anon_sym_interface] = ACTIONS(7069),
    [anon_sym_abstract] = ACTIONS(7069),
    [anon_sym_val] = ACTIONS(7069),
    [anon_sym_override] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__dedent] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3640)] = {
    [sym_block_comment] = STATE(3640),
    [sym_compiler_directive_decl] = STATE(3640),
    [sym_fsi_directive_decl] = STATE(3640),
    [sym_preproc_line] = STATE(3640),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_DOT_DOT] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__dedent] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3641)] = {
    [sym_block_comment] = STATE(3641),
    [sym_compiler_directive_decl] = STATE(3641),
    [sym_fsi_directive_decl] = STATE(3641),
    [sym_preproc_line] = STATE(3641),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [aux_sym_access_modifier_token1] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_DOT_DOT] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_default] = ACTIONS(7085),
    [anon_sym_static] = ACTIONS(7085),
    [anon_sym_member] = ACTIONS(7085),
    [anon_sym_interface] = ACTIONS(7085),
    [anon_sym_abstract] = ACTIONS(7085),
    [anon_sym_val] = ACTIONS(7085),
    [anon_sym_override] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__dedent] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3642)] = {
    [sym_block_comment] = STATE(3642),
    [sym_compiler_directive_decl] = STATE(3642),
    [sym_fsi_directive_decl] = STATE(3642),
    [sym_preproc_line] = STATE(3642),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [aux_sym_access_modifier_token1] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_default] = ACTIONS(6902),
    [anon_sym_static] = ACTIONS(6902),
    [anon_sym_member] = ACTIONS(6902),
    [anon_sym_interface] = ACTIONS(6902),
    [anon_sym_abstract] = ACTIONS(6902),
    [anon_sym_val] = ACTIONS(6902),
    [anon_sym_override] = ACTIONS(6902),
    [anon_sym_inherit] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3643)] = {
    [sym_block_comment] = STATE(3643),
    [sym_compiler_directive_decl] = STATE(3643),
    [sym_fsi_directive_decl] = STATE(3643),
    [sym_preproc_line] = STATE(3643),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [aux_sym_access_modifier_token1] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_DOT_DOT] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_default] = ACTIONS(7089),
    [anon_sym_static] = ACTIONS(7089),
    [anon_sym_member] = ACTIONS(7089),
    [anon_sym_interface] = ACTIONS(7089),
    [anon_sym_abstract] = ACTIONS(7089),
    [anon_sym_val] = ACTIONS(7089),
    [anon_sym_override] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__dedent] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3644)] = {
    [sym_block_comment] = STATE(3644),
    [sym_compiler_directive_decl] = STATE(3644),
    [sym_fsi_directive_decl] = STATE(3644),
    [sym_preproc_line] = STATE(3644),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [aux_sym_access_modifier_token1] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_DOT_DOT] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_default] = ACTIONS(7093),
    [anon_sym_static] = ACTIONS(7093),
    [anon_sym_member] = ACTIONS(7093),
    [anon_sym_interface] = ACTIONS(7093),
    [anon_sym_abstract] = ACTIONS(7093),
    [anon_sym_val] = ACTIONS(7093),
    [anon_sym_override] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__dedent] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3645)] = {
    [sym_block_comment] = STATE(3645),
    [sym_compiler_directive_decl] = STATE(3645),
    [sym_fsi_directive_decl] = STATE(3645),
    [sym_preproc_line] = STATE(3645),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [aux_sym_access_modifier_token1] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_DOT_DOT] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_default] = ACTIONS(7097),
    [anon_sym_static] = ACTIONS(7097),
    [anon_sym_member] = ACTIONS(7097),
    [anon_sym_interface] = ACTIONS(7097),
    [anon_sym_abstract] = ACTIONS(7097),
    [anon_sym_val] = ACTIONS(7097),
    [anon_sym_override] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__dedent] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3646)] = {
    [sym_block_comment] = STATE(3646),
    [sym_compiler_directive_decl] = STATE(3646),
    [sym_fsi_directive_decl] = STATE(3646),
    [sym_preproc_line] = STATE(3646),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [aux_sym_access_modifier_token1] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_DOT_DOT] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_default] = ACTIONS(7101),
    [anon_sym_static] = ACTIONS(7101),
    [anon_sym_member] = ACTIONS(7101),
    [anon_sym_interface] = ACTIONS(7101),
    [anon_sym_abstract] = ACTIONS(7101),
    [anon_sym_val] = ACTIONS(7101),
    [anon_sym_override] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__dedent] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3647)] = {
    [sym_block_comment] = STATE(3647),
    [sym_compiler_directive_decl] = STATE(3647),
    [sym_fsi_directive_decl] = STATE(3647),
    [sym_preproc_line] = STATE(3647),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [aux_sym_access_modifier_token1] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_DOT_DOT] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_default] = ACTIONS(7105),
    [anon_sym_static] = ACTIONS(7105),
    [anon_sym_member] = ACTIONS(7105),
    [anon_sym_interface] = ACTIONS(7105),
    [anon_sym_abstract] = ACTIONS(7105),
    [anon_sym_val] = ACTIONS(7105),
    [anon_sym_override] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__dedent] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3648)] = {
    [sym_block_comment] = STATE(3648),
    [sym_compiler_directive_decl] = STATE(3648),
    [sym_fsi_directive_decl] = STATE(3648),
    [sym_preproc_line] = STATE(3648),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [aux_sym_access_modifier_token1] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_DOT_DOT] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_default] = ACTIONS(7109),
    [anon_sym_static] = ACTIONS(7109),
    [anon_sym_member] = ACTIONS(7109),
    [anon_sym_interface] = ACTIONS(7109),
    [anon_sym_abstract] = ACTIONS(7109),
    [anon_sym_val] = ACTIONS(7109),
    [anon_sym_override] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__dedent] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3649)] = {
    [sym_block_comment] = STATE(3649),
    [sym_compiler_directive_decl] = STATE(3649),
    [sym_fsi_directive_decl] = STATE(3649),
    [sym_preproc_line] = STATE(3649),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [aux_sym_access_modifier_token1] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_DOT_DOT] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_default] = ACTIONS(7113),
    [anon_sym_static] = ACTIONS(7113),
    [anon_sym_member] = ACTIONS(7113),
    [anon_sym_interface] = ACTIONS(7113),
    [anon_sym_abstract] = ACTIONS(7113),
    [anon_sym_val] = ACTIONS(7113),
    [anon_sym_override] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__dedent] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3650)] = {
    [sym_block_comment] = STATE(3650),
    [sym_compiler_directive_decl] = STATE(3650),
    [sym_fsi_directive_decl] = STATE(3650),
    [sym_preproc_line] = STATE(3650),
    [aux_sym_type_argument_repeat1] = STATE(3523),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8094),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3651)] = {
    [sym_block_comment] = STATE(3651),
    [sym_compiler_directive_decl] = STATE(3651),
    [sym_fsi_directive_decl] = STATE(3651),
    [sym_preproc_line] = STATE(3651),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [aux_sym_access_modifier_token1] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_DOT_DOT] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_default] = ACTIONS(7117),
    [anon_sym_static] = ACTIONS(7117),
    [anon_sym_member] = ACTIONS(7117),
    [anon_sym_interface] = ACTIONS(7117),
    [anon_sym_abstract] = ACTIONS(7117),
    [anon_sym_val] = ACTIONS(7117),
    [anon_sym_override] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__dedent] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3652)] = {
    [sym_block_comment] = STATE(3652),
    [sym_compiler_directive_decl] = STATE(3652),
    [sym_fsi_directive_decl] = STATE(3652),
    [sym_preproc_line] = STATE(3652),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [aux_sym_access_modifier_token1] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_DOT_DOT] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_default] = ACTIONS(7121),
    [anon_sym_static] = ACTIONS(7121),
    [anon_sym_member] = ACTIONS(7121),
    [anon_sym_interface] = ACTIONS(7121),
    [anon_sym_abstract] = ACTIONS(7121),
    [anon_sym_val] = ACTIONS(7121),
    [anon_sym_override] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__dedent] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3653)] = {
    [sym_block_comment] = STATE(3653),
    [sym_compiler_directive_decl] = STATE(3653),
    [sym_fsi_directive_decl] = STATE(3653),
    [sym_preproc_line] = STATE(3653),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [aux_sym_access_modifier_token1] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_DOT_DOT] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_default] = ACTIONS(6936),
    [anon_sym_static] = ACTIONS(6936),
    [anon_sym_member] = ACTIONS(6936),
    [anon_sym_interface] = ACTIONS(6936),
    [anon_sym_abstract] = ACTIONS(6936),
    [anon_sym_val] = ACTIONS(6936),
    [anon_sym_override] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__dedent] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3654)] = {
    [sym_block_comment] = STATE(3654),
    [sym_compiler_directive_decl] = STATE(3654),
    [sym_fsi_directive_decl] = STATE(3654),
    [sym_preproc_line] = STATE(3654),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [aux_sym_access_modifier_token1] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_DOT_DOT] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_default] = ACTIONS(6932),
    [anon_sym_static] = ACTIONS(6932),
    [anon_sym_member] = ACTIONS(6932),
    [anon_sym_interface] = ACTIONS(6932),
    [anon_sym_abstract] = ACTIONS(6932),
    [anon_sym_val] = ACTIONS(6932),
    [anon_sym_override] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__dedent] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(3655)] = {
    [sym_block_comment] = STATE(3655),
    [sym_compiler_directive_decl] = STATE(3655),
    [sym_fsi_directive_decl] = STATE(3655),
    [sym_preproc_line] = STATE(3655),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [aux_sym_access_modifier_token1] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_DOT_DOT] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_default] = ACTIONS(6940),
    [anon_sym_static] = ACTIONS(6940),
    [anon_sym_member] = ACTIONS(6940),
    [anon_sym_interface] = ACTIONS(6940),
    [anon_sym_abstract] = ACTIONS(6940),
    [anon_sym_val] = ACTIONS(6940),
    [anon_sym_override] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__dedent] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3656)] = {
    [sym_block_comment] = STATE(3656),
    [sym_compiler_directive_decl] = STATE(3656),
    [sym_fsi_directive_decl] = STATE(3656),
    [sym_preproc_line] = STATE(3656),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [aux_sym_access_modifier_token1] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_DOT_DOT] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_default] = ACTIONS(7015),
    [anon_sym_static] = ACTIONS(7015),
    [anon_sym_member] = ACTIONS(7015),
    [anon_sym_interface] = ACTIONS(7015),
    [anon_sym_abstract] = ACTIONS(7015),
    [anon_sym_val] = ACTIONS(7015),
    [anon_sym_override] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__dedent] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3657)] = {
    [sym_block_comment] = STATE(3657),
    [sym_compiler_directive_decl] = STATE(3657),
    [sym_fsi_directive_decl] = STATE(3657),
    [sym_preproc_line] = STATE(3657),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [aux_sym_access_modifier_token1] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_DOT_DOT] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_default] = ACTIONS(7031),
    [anon_sym_static] = ACTIONS(7031),
    [anon_sym_member] = ACTIONS(7031),
    [anon_sym_interface] = ACTIONS(7031),
    [anon_sym_abstract] = ACTIONS(7031),
    [anon_sym_val] = ACTIONS(7031),
    [anon_sym_override] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__dedent] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3658)] = {
    [sym_block_comment] = STATE(3658),
    [sym_compiler_directive_decl] = STATE(3658),
    [sym_fsi_directive_decl] = STATE(3658),
    [sym_preproc_line] = STATE(3658),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [aux_sym_access_modifier_token1] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_DOT_DOT] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_default] = ACTIONS(7047),
    [anon_sym_static] = ACTIONS(7047),
    [anon_sym_member] = ACTIONS(7047),
    [anon_sym_interface] = ACTIONS(7047),
    [anon_sym_abstract] = ACTIONS(7047),
    [anon_sym_val] = ACTIONS(7047),
    [anon_sym_override] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__dedent] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3659)] = {
    [sym_block_comment] = STATE(3659),
    [sym_compiler_directive_decl] = STATE(3659),
    [sym_fsi_directive_decl] = STATE(3659),
    [sym_preproc_line] = STATE(3659),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [aux_sym_access_modifier_token1] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_DOT_DOT] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_default] = ACTIONS(7051),
    [anon_sym_static] = ACTIONS(7051),
    [anon_sym_member] = ACTIONS(7051),
    [anon_sym_interface] = ACTIONS(7051),
    [anon_sym_abstract] = ACTIONS(7051),
    [anon_sym_val] = ACTIONS(7051),
    [anon_sym_override] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__dedent] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3660)] = {
    [sym_block_comment] = STATE(3660),
    [sym_compiler_directive_decl] = STATE(3660),
    [sym_fsi_directive_decl] = STATE(3660),
    [sym_preproc_line] = STATE(3660),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [aux_sym_access_modifier_token1] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_default] = ACTIONS(7023),
    [anon_sym_static] = ACTIONS(7023),
    [anon_sym_member] = ACTIONS(7023),
    [anon_sym_interface] = ACTIONS(7023),
    [anon_sym_abstract] = ACTIONS(7023),
    [anon_sym_val] = ACTIONS(7023),
    [anon_sym_override] = ACTIONS(7023),
    [anon_sym_inherit] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3661)] = {
    [sym_block_comment] = STATE(3661),
    [sym_compiler_directive_decl] = STATE(3661),
    [sym_fsi_directive_decl] = STATE(3661),
    [sym_preproc_line] = STATE(3661),
    [aux_sym_compound_type_repeat1] = STATE(3668),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(7796),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(3662)] = {
    [sym_block_comment] = STATE(3662),
    [sym_compiler_directive_decl] = STATE(3662),
    [sym_fsi_directive_decl] = STATE(3662),
    [sym_preproc_line] = STATE(3662),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_and] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [aux_sym_access_modifier_token1] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_default] = ACTIONS(7073),
    [anon_sym_static] = ACTIONS(7073),
    [anon_sym_member] = ACTIONS(7073),
    [anon_sym_interface] = ACTIONS(7073),
    [anon_sym_abstract] = ACTIONS(7073),
    [anon_sym_val] = ACTIONS(7073),
    [anon_sym_override] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__dedent] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(3663)] = {
    [sym_block_comment] = STATE(3663),
    [sym_compiler_directive_decl] = STATE(3663),
    [sym_fsi_directive_decl] = STATE(3663),
    [sym_preproc_line] = STATE(3663),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [aux_sym_access_modifier_token1] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_DOT_DOT] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_default] = ACTIONS(7063),
    [anon_sym_static] = ACTIONS(7063),
    [anon_sym_member] = ACTIONS(7063),
    [anon_sym_interface] = ACTIONS(7063),
    [anon_sym_abstract] = ACTIONS(7063),
    [anon_sym_val] = ACTIONS(7063),
    [anon_sym_override] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__dedent] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3664)] = {
    [sym_block_comment] = STATE(3664),
    [sym_compiler_directive_decl] = STATE(3664),
    [sym_fsi_directive_decl] = STATE(3664),
    [sym_preproc_line] = STATE(3664),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_and] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [aux_sym_access_modifier_token1] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_default] = ACTIONS(6944),
    [anon_sym_static] = ACTIONS(6944),
    [anon_sym_member] = ACTIONS(6944),
    [anon_sym_interface] = ACTIONS(6944),
    [anon_sym_abstract] = ACTIONS(6944),
    [anon_sym_val] = ACTIONS(6944),
    [anon_sym_override] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__dedent] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3665)] = {
    [sym_block_comment] = STATE(3665),
    [sym_compiler_directive_decl] = STATE(3665),
    [sym_fsi_directive_decl] = STATE(3665),
    [sym_preproc_line] = STATE(3665),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_and] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [aux_sym_access_modifier_token1] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_default] = ACTIONS(6948),
    [anon_sym_static] = ACTIONS(6948),
    [anon_sym_member] = ACTIONS(6948),
    [anon_sym_interface] = ACTIONS(6948),
    [anon_sym_abstract] = ACTIONS(6948),
    [anon_sym_val] = ACTIONS(6948),
    [anon_sym_override] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__dedent] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3666)] = {
    [sym_block_comment] = STATE(3666),
    [sym_compiler_directive_decl] = STATE(3666),
    [sym_fsi_directive_decl] = STATE(3666),
    [sym_preproc_line] = STATE(3666),
    [ts_builtin_sym_end] = ACTIONS(6365),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_namespace] = ACTIONS(6367),
    [anon_sym_module] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_open] = ACTIONS(6367),
    [anon_sym_type] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_exception] = ACTIONS(6367),
    [anon_sym_extern] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6365),
    [anon_sym_POUNDload] = ACTIONS(6365),
    [anon_sym_POUNDtime] = ACTIONS(6365),
    [anon_sym_POUNDI] = ACTIONS(6365),
    [anon_sym_POUNDhelp] = ACTIONS(6365),
    [anon_sym_POUNDquit] = ACTIONS(6365),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(3667)] = {
    [sym_block_comment] = STATE(3667),
    [sym_compiler_directive_decl] = STATE(3667),
    [sym_fsi_directive_decl] = STATE(3667),
    [sym_preproc_line] = STATE(3667),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_and] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [aux_sym_access_modifier_token1] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_default] = ACTIONS(7077),
    [anon_sym_static] = ACTIONS(7077),
    [anon_sym_member] = ACTIONS(7077),
    [anon_sym_interface] = ACTIONS(7077),
    [anon_sym_abstract] = ACTIONS(7077),
    [anon_sym_val] = ACTIONS(7077),
    [anon_sym_override] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__dedent] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3668)] = {
    [sym_block_comment] = STATE(3668),
    [sym_compiler_directive_decl] = STATE(3668),
    [sym_fsi_directive_decl] = STATE(3668),
    [sym_preproc_line] = STATE(3668),
    [aux_sym_compound_type_repeat1] = STATE(3668),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8097),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3669)] = {
    [sym_block_comment] = STATE(3669),
    [sym_compiler_directive_decl] = STATE(3669),
    [sym_fsi_directive_decl] = STATE(3669),
    [sym_preproc_line] = STATE(3669),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_and] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [aux_sym_access_modifier_token1] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_default] = ACTIONS(6952),
    [anon_sym_static] = ACTIONS(6952),
    [anon_sym_member] = ACTIONS(6952),
    [anon_sym_interface] = ACTIONS(6952),
    [anon_sym_abstract] = ACTIONS(6952),
    [anon_sym_val] = ACTIONS(6952),
    [anon_sym_override] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__dedent] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3670)] = {
    [sym_block_comment] = STATE(3670),
    [sym_compiler_directive_decl] = STATE(3670),
    [sym_fsi_directive_decl] = STATE(3670),
    [sym_preproc_line] = STATE(3670),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_and] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [aux_sym_access_modifier_token1] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_default] = ACTIONS(6959),
    [anon_sym_static] = ACTIONS(6959),
    [anon_sym_member] = ACTIONS(6959),
    [anon_sym_interface] = ACTIONS(6959),
    [anon_sym_abstract] = ACTIONS(6959),
    [anon_sym_val] = ACTIONS(6959),
    [anon_sym_override] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__dedent] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3671)] = {
    [sym_block_comment] = STATE(3671),
    [sym_compiler_directive_decl] = STATE(3671),
    [sym_fsi_directive_decl] = STATE(3671),
    [sym_preproc_line] = STATE(3671),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_and] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [aux_sym_access_modifier_token1] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_default] = ACTIONS(6963),
    [anon_sym_static] = ACTIONS(6963),
    [anon_sym_member] = ACTIONS(6963),
    [anon_sym_interface] = ACTIONS(6963),
    [anon_sym_abstract] = ACTIONS(6963),
    [anon_sym_val] = ACTIONS(6963),
    [anon_sym_override] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__dedent] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3672)] = {
    [sym_block_comment] = STATE(3672),
    [sym_compiler_directive_decl] = STATE(3672),
    [sym_fsi_directive_decl] = STATE(3672),
    [sym_preproc_line] = STATE(3672),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [aux_sym_access_modifier_token1] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_DOT_DOT] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_default] = ACTIONS(7011),
    [anon_sym_static] = ACTIONS(7011),
    [anon_sym_member] = ACTIONS(7011),
    [anon_sym_interface] = ACTIONS(7011),
    [anon_sym_abstract] = ACTIONS(7011),
    [anon_sym_val] = ACTIONS(7011),
    [anon_sym_override] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__dedent] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(3673)] = {
    [sym_type_arguments] = STATE(4054),
    [sym_long_identifier] = STATE(3922),
    [sym_block_comment] = STATE(3673),
    [sym_compiler_directive_decl] = STATE(3673),
    [sym_fsi_directive_decl] = STATE(3673),
    [sym_preproc_line] = STATE(3673),
    [aux_sym_compound_type_repeat1] = STATE(3786),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_as] = ACTIONS(6265),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8100),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_DOT_DOT] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8102),
    [anon_sym_LT] = ACTIONS(8104),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8106),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__dedent] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3674)] = {
    [sym_block_comment] = STATE(3674),
    [sym_compiler_directive_decl] = STATE(3674),
    [sym_fsi_directive_decl] = STATE(3674),
    [sym_preproc_line] = STATE(3674),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3675)] = {
    [sym_type_arguments] = STATE(4054),
    [sym_long_identifier] = STATE(3922),
    [sym_block_comment] = STATE(3675),
    [sym_compiler_directive_decl] = STATE(3675),
    [sym_fsi_directive_decl] = STATE(3675),
    [sym_preproc_line] = STATE(3675),
    [aux_sym_compound_type_repeat1] = STATE(3786),
    [sym_identifier] = ACTIONS(8108),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_as] = ACTIONS(6213),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8100),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_DOT_DOT] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8102),
    [anon_sym_LT] = ACTIONS(8104),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8106),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__dedent] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3676)] = {
    [sym_block_comment] = STATE(3676),
    [sym_compiler_directive_decl] = STATE(3676),
    [sym_fsi_directive_decl] = STATE(3676),
    [sym_preproc_line] = STATE(3676),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(3677)] = {
    [sym_block_comment] = STATE(3677),
    [sym_compiler_directive_decl] = STATE(3677),
    [sym_fsi_directive_decl] = STATE(3677),
    [sym_preproc_line] = STATE(3677),
    [aux_sym_type_argument_repeat1] = STATE(3688),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_GT_RBRACK] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_RBRACK] = ACTIONS(6333),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_RBRACE] = ACTIONS(6333),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8110),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_to] = ACTIONS(6331),
    [anon_sym_downto] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_DOT_DOT2] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3678)] = {
    [sym_type_arguments] = STATE(4054),
    [sym_long_identifier] = STATE(3922),
    [sym_block_comment] = STATE(3678),
    [sym_compiler_directive_decl] = STATE(3678),
    [sym_fsi_directive_decl] = STATE(3678),
    [sym_preproc_line] = STATE(3678),
    [aux_sym_compound_type_repeat1] = STATE(3786),
    [sym_identifier] = ACTIONS(8108),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_as] = ACTIONS(6193),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8100),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8102),
    [anon_sym_LT] = ACTIONS(8104),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8106),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3679)] = {
    [sym_type_arguments] = STATE(4054),
    [sym_long_identifier] = STATE(3922),
    [sym_block_comment] = STATE(3679),
    [sym_compiler_directive_decl] = STATE(3679),
    [sym_fsi_directive_decl] = STATE(3679),
    [sym_preproc_line] = STATE(3679),
    [aux_sym_compound_type_repeat1] = STATE(3786),
    [sym_identifier] = ACTIONS(8108),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_as] = ACTIONS(6261),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8100),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_DOT_DOT] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8102),
    [anon_sym_LT] = ACTIONS(8104),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8106),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__dedent] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3680)] = {
    [sym_block_comment] = STATE(3680),
    [sym_compiler_directive_decl] = STATE(3680),
    [sym_fsi_directive_decl] = STATE(3680),
    [sym_preproc_line] = STATE(3680),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_module] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_open] = ACTIONS(6389),
    [anon_sym_type] = ACTIONS(6389),
    [anon_sym_LBRACK_LT] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [anon_sym_exception] = ACTIONS(6389),
    [anon_sym_extern] = ACTIONS(6389),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6387),
    [anon_sym_POUNDload] = ACTIONS(6387),
    [anon_sym_POUNDtime] = ACTIONS(6387),
    [anon_sym_POUNDI] = ACTIONS(6387),
    [anon_sym_POUNDhelp] = ACTIONS(6387),
    [anon_sym_POUNDquit] = ACTIONS(6387),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(3681)] = {
    [sym_block_comment] = STATE(3681),
    [sym_compiler_directive_decl] = STATE(3681),
    [sym_fsi_directive_decl] = STATE(3681),
    [sym_preproc_line] = STATE(3681),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8113),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3682)] = {
    [sym_block_comment] = STATE(3682),
    [sym_compiler_directive_decl] = STATE(3682),
    [sym_fsi_directive_decl] = STATE(3682),
    [sym_preproc_line] = STATE(3682),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_module] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_open] = ACTIONS(6427),
    [anon_sym_type] = ACTIONS(6427),
    [anon_sym_LBRACK_LT] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [anon_sym_exception] = ACTIONS(6427),
    [anon_sym_extern] = ACTIONS(6427),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6429),
    [anon_sym_POUNDload] = ACTIONS(6429),
    [anon_sym_POUNDtime] = ACTIONS(6429),
    [anon_sym_POUNDI] = ACTIONS(6429),
    [anon_sym_POUNDhelp] = ACTIONS(6429),
    [anon_sym_POUNDquit] = ACTIONS(6429),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(3683)] = {
    [sym_block_comment] = STATE(3683),
    [sym_compiler_directive_decl] = STATE(3683),
    [sym_fsi_directive_decl] = STATE(3683),
    [sym_preproc_line] = STATE(3683),
    [aux_sym_long_identifier_repeat1] = STATE(3700),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_GT_RBRACK] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_RBRACK] = ACTIONS(6291),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_RBRACE] = ACTIONS(6291),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8115),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_to] = ACTIONS(6289),
    [anon_sym_downto] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_DOT_DOT2] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(3684)] = {
    [sym__else_expression] = STATE(4387),
    [sym_elif_expression] = STATE(4037),
    [sym_block_comment] = STATE(3684),
    [sym_compiler_directive_decl] = STATE(3684),
    [sym_fsi_directive_decl] = STATE(3684),
    [sym_preproc_line] = STATE(3684),
    [aux_sym_if_expression_repeat1] = STATE(3746),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_GT_RBRACK] = ACTIONS(6327),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_RBRACK] = ACTIONS(6327),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_RBRACE] = ACTIONS(6327),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_to] = ACTIONS(6329),
    [anon_sym_downto] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8117),
    [anon_sym_elif] = ACTIONS(8119),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_DOT_DOT2] = ACTIONS(6327),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(3685)] = {
    [sym_block_comment] = STATE(3685),
    [sym_compiler_directive_decl] = STATE(3685),
    [sym_fsi_directive_decl] = STATE(3685),
    [sym_preproc_line] = STATE(3685),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3686)] = {
    [sym_block_comment] = STATE(3686),
    [sym_compiler_directive_decl] = STATE(3686),
    [sym_fsi_directive_decl] = STATE(3686),
    [sym_preproc_line] = STATE(3686),
    [aux_sym_type_argument_repeat1] = STATE(3686),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_GT_RBRACK] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_RBRACK] = ACTIONS(6310),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_RBRACE] = ACTIONS(6310),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8121),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_to] = ACTIONS(6312),
    [anon_sym_downto] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_DOT_DOT2] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3687)] = {
    [sym_block_comment] = STATE(3687),
    [sym_compiler_directive_decl] = STATE(3687),
    [sym_fsi_directive_decl] = STATE(3687),
    [sym_preproc_line] = STATE(3687),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_module] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_open] = ACTIONS(6492),
    [anon_sym_type] = ACTIONS(6492),
    [anon_sym_LBRACK_LT] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [anon_sym_exception] = ACTIONS(6492),
    [anon_sym_extern] = ACTIONS(6492),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6490),
    [anon_sym_POUNDload] = ACTIONS(6490),
    [anon_sym_POUNDtime] = ACTIONS(6490),
    [anon_sym_POUNDI] = ACTIONS(6490),
    [anon_sym_POUNDhelp] = ACTIONS(6490),
    [anon_sym_POUNDquit] = ACTIONS(6490),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(3688)] = {
    [sym_block_comment] = STATE(3688),
    [sym_compiler_directive_decl] = STATE(3688),
    [sym_fsi_directive_decl] = STATE(3688),
    [sym_preproc_line] = STATE(3688),
    [aux_sym_type_argument_repeat1] = STATE(3686),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_GT_RBRACK] = ACTIONS(6306),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_RBRACK] = ACTIONS(6306),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_RBRACE] = ACTIONS(6306),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_to] = ACTIONS(6308),
    [anon_sym_downto] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_DOT_DOT2] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(3689)] = {
    [sym_block_comment] = STATE(3689),
    [sym_compiler_directive_decl] = STATE(3689),
    [sym_fsi_directive_decl] = STATE(3689),
    [sym_preproc_line] = STATE(3689),
    [aux_sym_compound_type_repeat1] = STATE(3707),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_GT_RBRACK] = ACTIONS(6340),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_RBRACK] = ACTIONS(6340),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_RBRACE] = ACTIONS(6340),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_to] = ACTIONS(6338),
    [anon_sym_downto] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_DOT_DOT2] = ACTIONS(6340),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(3690)] = {
    [sym_block_comment] = STATE(3690),
    [sym_compiler_directive_decl] = STATE(3690),
    [sym_fsi_directive_decl] = STATE(3690),
    [sym_preproc_line] = STATE(3690),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3691)] = {
    [sym_type_arguments] = STATE(4054),
    [sym_long_identifier] = STATE(3922),
    [sym_block_comment] = STATE(3691),
    [sym_compiler_directive_decl] = STATE(3691),
    [sym_fsi_directive_decl] = STATE(3691),
    [sym_preproc_line] = STATE(3691),
    [aux_sym_compound_type_repeat1] = STATE(3786),
    [sym_identifier] = ACTIONS(8108),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_as] = ACTIONS(6257),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8100),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_DOT_DOT] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8102),
    [anon_sym_LT] = ACTIONS(8104),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8106),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__dedent] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3692)] = {
    [sym_block_comment] = STATE(3692),
    [sym_compiler_directive_decl] = STATE(3692),
    [sym_fsi_directive_decl] = STATE(3692),
    [sym_preproc_line] = STATE(3692),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_module] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_open] = ACTIONS(6373),
    [anon_sym_type] = ACTIONS(6373),
    [anon_sym_LBRACK_LT] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8124),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [anon_sym_exception] = ACTIONS(6373),
    [anon_sym_extern] = ACTIONS(6373),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6375),
    [anon_sym_POUNDload] = ACTIONS(6375),
    [anon_sym_POUNDtime] = ACTIONS(6375),
    [anon_sym_POUNDI] = ACTIONS(6375),
    [anon_sym_POUNDhelp] = ACTIONS(6375),
    [anon_sym_POUNDquit] = ACTIONS(6375),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(3693)] = {
    [sym__else_expression] = STATE(4377),
    [sym_elif_expression] = STATE(4037),
    [sym_block_comment] = STATE(3693),
    [sym_compiler_directive_decl] = STATE(3693),
    [sym_fsi_directive_decl] = STATE(3693),
    [sym_preproc_line] = STATE(3693),
    [aux_sym_if_expression_repeat1] = STATE(3684),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_GT_RBRACK] = ACTIONS(6317),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_RBRACK] = ACTIONS(6317),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_RBRACE] = ACTIONS(6317),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_to] = ACTIONS(6319),
    [anon_sym_downto] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8117),
    [anon_sym_elif] = ACTIONS(8119),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_DOT_DOT2] = ACTIONS(6317),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(3694)] = {
    [sym_block_comment] = STATE(3694),
    [sym_compiler_directive_decl] = STATE(3694),
    [sym_fsi_directive_decl] = STATE(3694),
    [sym_preproc_line] = STATE(3694),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_module] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_open] = ACTIONS(6367),
    [anon_sym_type] = ACTIONS(6367),
    [anon_sym_LBRACK_LT] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [anon_sym_exception] = ACTIONS(6367),
    [anon_sym_extern] = ACTIONS(6367),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6365),
    [anon_sym_POUNDload] = ACTIONS(6365),
    [anon_sym_POUNDtime] = ACTIONS(6365),
    [anon_sym_POUNDI] = ACTIONS(6365),
    [anon_sym_POUNDhelp] = ACTIONS(6365),
    [anon_sym_POUNDquit] = ACTIONS(6365),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(3695)] = {
    [sym_elif_expression] = STATE(3828),
    [sym_block_comment] = STATE(3695),
    [sym_compiler_directive_decl] = STATE(3695),
    [sym_fsi_directive_decl] = STATE(3695),
    [sym_preproc_line] = STATE(3695),
    [aux_sym_if_expression_repeat1] = STATE(3695),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_module] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_open] = ACTIONS(6496),
    [anon_sym_type] = ACTIONS(6496),
    [anon_sym_LBRACK_LT] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8126),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_exception] = ACTIONS(6496),
    [anon_sym_extern] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6494),
    [anon_sym_POUNDload] = ACTIONS(6494),
    [anon_sym_POUNDtime] = ACTIONS(6494),
    [anon_sym_POUNDI] = ACTIONS(6494),
    [anon_sym_POUNDhelp] = ACTIONS(6494),
    [anon_sym_POUNDquit] = ACTIONS(6494),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(3696)] = {
    [sym_block_comment] = STATE(3696),
    [sym_compiler_directive_decl] = STATE(3696),
    [sym_fsi_directive_decl] = STATE(3696),
    [sym_preproc_line] = STATE(3696),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_module] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_open] = ACTIONS(6379),
    [anon_sym_type] = ACTIONS(6379),
    [anon_sym_LBRACK_LT] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [anon_sym_exception] = ACTIONS(6379),
    [anon_sym_extern] = ACTIONS(6379),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6381),
    [anon_sym_POUNDload] = ACTIONS(6381),
    [anon_sym_POUNDtime] = ACTIONS(6381),
    [anon_sym_POUNDI] = ACTIONS(6381),
    [anon_sym_POUNDhelp] = ACTIONS(6381),
    [anon_sym_POUNDquit] = ACTIONS(6381),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(3697)] = {
    [sym_block_comment] = STATE(3697),
    [sym_compiler_directive_decl] = STATE(3697),
    [sym_fsi_directive_decl] = STATE(3697),
    [sym_preproc_line] = STATE(3697),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_module] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_open] = ACTIONS(6383),
    [anon_sym_type] = ACTIONS(6383),
    [anon_sym_LBRACK_LT] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [anon_sym_exception] = ACTIONS(6383),
    [anon_sym_extern] = ACTIONS(6383),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6385),
    [anon_sym_POUNDload] = ACTIONS(6385),
    [anon_sym_POUNDtime] = ACTIONS(6385),
    [anon_sym_POUNDI] = ACTIONS(6385),
    [anon_sym_POUNDhelp] = ACTIONS(6385),
    [anon_sym_POUNDquit] = ACTIONS(6385),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(3698)] = {
    [sym_block_comment] = STATE(3698),
    [sym_compiler_directive_decl] = STATE(3698),
    [sym_fsi_directive_decl] = STATE(3698),
    [sym_preproc_line] = STATE(3698),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_module] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_open] = ACTIONS(6391),
    [anon_sym_type] = ACTIONS(6391),
    [anon_sym_LBRACK_LT] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [anon_sym_exception] = ACTIONS(6391),
    [anon_sym_extern] = ACTIONS(6391),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6393),
    [anon_sym_POUNDload] = ACTIONS(6393),
    [anon_sym_POUNDtime] = ACTIONS(6393),
    [anon_sym_POUNDI] = ACTIONS(6393),
    [anon_sym_POUNDhelp] = ACTIONS(6393),
    [anon_sym_POUNDquit] = ACTIONS(6393),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(3699)] = {
    [sym_block_comment] = STATE(3699),
    [sym_compiler_directive_decl] = STATE(3699),
    [sym_fsi_directive_decl] = STATE(3699),
    [sym_preproc_line] = STATE(3699),
    [ts_builtin_sym_end] = ACTIONS(6515),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_namespace] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8129),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3700)] = {
    [sym_block_comment] = STATE(3700),
    [sym_compiler_directive_decl] = STATE(3700),
    [sym_fsi_directive_decl] = STATE(3700),
    [sym_preproc_line] = STATE(3700),
    [aux_sym_long_identifier_repeat1] = STATE(3708),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_GT_RBRACK] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_RBRACK] = ACTIONS(6358),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_RBRACE] = ACTIONS(6358),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8115),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_to] = ACTIONS(6360),
    [anon_sym_downto] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_DOT_DOT2] = ACTIONS(6358),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3701)] = {
    [sym_block_comment] = STATE(3701),
    [sym_compiler_directive_decl] = STATE(3701),
    [sym_fsi_directive_decl] = STATE(3701),
    [sym_preproc_line] = STATE(3701),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3702)] = {
    [sym_block_comment] = STATE(3702),
    [sym_compiler_directive_decl] = STATE(3702),
    [sym_fsi_directive_decl] = STATE(3702),
    [sym_preproc_line] = STATE(3702),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_module] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_open] = ACTIONS(6395),
    [anon_sym_type] = ACTIONS(6395),
    [anon_sym_LBRACK_LT] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [anon_sym_exception] = ACTIONS(6395),
    [anon_sym_extern] = ACTIONS(6395),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6397),
    [anon_sym_POUNDload] = ACTIONS(6397),
    [anon_sym_POUNDtime] = ACTIONS(6397),
    [anon_sym_POUNDI] = ACTIONS(6397),
    [anon_sym_POUNDhelp] = ACTIONS(6397),
    [anon_sym_POUNDquit] = ACTIONS(6397),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(3703)] = {
    [sym_block_comment] = STATE(3703),
    [sym_compiler_directive_decl] = STATE(3703),
    [sym_fsi_directive_decl] = STATE(3703),
    [sym_preproc_line] = STATE(3703),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_module] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_open] = ACTIONS(6399),
    [anon_sym_type] = ACTIONS(6399),
    [anon_sym_LBRACK_LT] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [anon_sym_exception] = ACTIONS(6399),
    [anon_sym_extern] = ACTIONS(6399),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6401),
    [anon_sym_POUNDload] = ACTIONS(6401),
    [anon_sym_POUNDtime] = ACTIONS(6401),
    [anon_sym_POUNDI] = ACTIONS(6401),
    [anon_sym_POUNDhelp] = ACTIONS(6401),
    [anon_sym_POUNDquit] = ACTIONS(6401),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(3704)] = {
    [sym_block_comment] = STATE(3704),
    [sym_compiler_directive_decl] = STATE(3704),
    [sym_fsi_directive_decl] = STATE(3704),
    [sym_preproc_line] = STATE(3704),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_module] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_open] = ACTIONS(6403),
    [anon_sym_type] = ACTIONS(6403),
    [anon_sym_LBRACK_LT] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [anon_sym_exception] = ACTIONS(6403),
    [anon_sym_extern] = ACTIONS(6403),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6405),
    [anon_sym_POUNDload] = ACTIONS(6405),
    [anon_sym_POUNDtime] = ACTIONS(6405),
    [anon_sym_POUNDI] = ACTIONS(6405),
    [anon_sym_POUNDhelp] = ACTIONS(6405),
    [anon_sym_POUNDquit] = ACTIONS(6405),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(3705)] = {
    [sym_block_comment] = STATE(3705),
    [sym_compiler_directive_decl] = STATE(3705),
    [sym_fsi_directive_decl] = STATE(3705),
    [sym_preproc_line] = STATE(3705),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_module] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_open] = ACTIONS(6407),
    [anon_sym_type] = ACTIONS(6407),
    [anon_sym_LBRACK_LT] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [anon_sym_exception] = ACTIONS(6407),
    [anon_sym_extern] = ACTIONS(6407),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6409),
    [anon_sym_POUNDload] = ACTIONS(6409),
    [anon_sym_POUNDtime] = ACTIONS(6409),
    [anon_sym_POUNDI] = ACTIONS(6409),
    [anon_sym_POUNDhelp] = ACTIONS(6409),
    [anon_sym_POUNDquit] = ACTIONS(6409),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(3706)] = {
    [sym_block_comment] = STATE(3706),
    [sym_compiler_directive_decl] = STATE(3706),
    [sym_fsi_directive_decl] = STATE(3706),
    [sym_preproc_line] = STATE(3706),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_module] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_open] = ACTIONS(6411),
    [anon_sym_type] = ACTIONS(6411),
    [anon_sym_LBRACK_LT] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [anon_sym_exception] = ACTIONS(6411),
    [anon_sym_extern] = ACTIONS(6411),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6413),
    [anon_sym_POUNDload] = ACTIONS(6413),
    [anon_sym_POUNDtime] = ACTIONS(6413),
    [anon_sym_POUNDI] = ACTIONS(6413),
    [anon_sym_POUNDhelp] = ACTIONS(6413),
    [anon_sym_POUNDquit] = ACTIONS(6413),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(3707)] = {
    [sym_block_comment] = STATE(3707),
    [sym_compiler_directive_decl] = STATE(3707),
    [sym_fsi_directive_decl] = STATE(3707),
    [sym_preproc_line] = STATE(3707),
    [aux_sym_compound_type_repeat1] = STATE(3707),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_GT_RBRACK] = ACTIONS(6195),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_RBRACK] = ACTIONS(6195),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_RBRACE] = ACTIONS(6195),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_to] = ACTIONS(6193),
    [anon_sym_downto] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_DOT_DOT2] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(8131),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3708)] = {
    [sym_block_comment] = STATE(3708),
    [sym_compiler_directive_decl] = STATE(3708),
    [sym_fsi_directive_decl] = STATE(3708),
    [sym_preproc_line] = STATE(3708),
    [aux_sym_long_identifier_repeat1] = STATE(3708),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_GT_RBRACK] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_RBRACK] = ACTIONS(6297),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_RBRACE] = ACTIONS(6297),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8134),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_to] = ACTIONS(6295),
    [anon_sym_downto] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3709)] = {
    [sym_block_comment] = STATE(3709),
    [sym_compiler_directive_decl] = STATE(3709),
    [sym_fsi_directive_decl] = STATE(3709),
    [sym_preproc_line] = STATE(3709),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3710)] = {
    [sym_block_comment] = STATE(3710),
    [sym_compiler_directive_decl] = STATE(3710),
    [sym_fsi_directive_decl] = STATE(3710),
    [sym_preproc_line] = STATE(3710),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_GT_RBRACK] = ACTIONS(6310),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_RBRACK] = ACTIONS(6310),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_RBRACE] = ACTIONS(6310),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_to] = ACTIONS(6312),
    [anon_sym_downto] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_DOT_DOT2] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3711)] = {
    [sym_block_comment] = STATE(3711),
    [sym_compiler_directive_decl] = STATE(3711),
    [sym_fsi_directive_decl] = STATE(3711),
    [sym_preproc_line] = STATE(3711),
    [aux_sym_rules_repeat1] = STATE(3720),
    [ts_builtin_sym_end] = ACTIONS(6616),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_namespace] = ACTIONS(6618),
    [anon_sym_module] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_open] = ACTIONS(6618),
    [anon_sym_type] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(8137),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_exception] = ACTIONS(6618),
    [anon_sym_extern] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6616),
    [anon_sym_POUNDload] = ACTIONS(6616),
    [anon_sym_POUNDtime] = ACTIONS(6616),
    [anon_sym_POUNDI] = ACTIONS(6616),
    [anon_sym_POUNDhelp] = ACTIONS(6616),
    [anon_sym_POUNDquit] = ACTIONS(6616),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(8139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(3712)] = {
    [sym_type_arguments] = STATE(4281),
    [sym_long_identifier] = STATE(4282),
    [sym_block_comment] = STATE(3712),
    [sym_compiler_directive_decl] = STATE(3712),
    [sym_fsi_directive_decl] = STATE(3712),
    [sym_preproc_line] = STATE(3712),
    [aux_sym_compound_type_repeat1] = STATE(3935),
    [sym_identifier] = ACTIONS(8142),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8144),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_DOT_DOT] = ACTIONS(6257),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_DOT_DOT2] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(8146),
    [anon_sym_LT] = ACTIONS(8148),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8150),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__dedent] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3713)] = {
    [sym_block_comment] = STATE(3713),
    [sym_compiler_directive_decl] = STATE(3713),
    [sym_fsi_directive_decl] = STATE(3713),
    [sym_preproc_line] = STATE(3713),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_GT_RBRACK] = ACTIONS(6333),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_RBRACK] = ACTIONS(6333),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_RBRACE] = ACTIONS(6333),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_to] = ACTIONS(6331),
    [anon_sym_downto] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_DOT_DOT2] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3714)] = {
    [sym_type_arguments] = STATE(4281),
    [sym_long_identifier] = STATE(4282),
    [sym_block_comment] = STATE(3714),
    [sym_compiler_directive_decl] = STATE(3714),
    [sym_fsi_directive_decl] = STATE(3714),
    [sym_preproc_line] = STATE(3714),
    [aux_sym_compound_type_repeat1] = STATE(3935),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8144),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_DOT_DOT] = ACTIONS(6265),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_DOT_DOT2] = ACTIONS(6263),
    [anon_sym_STAR] = ACTIONS(8146),
    [anon_sym_LT] = ACTIONS(8148),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8150),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__dedent] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3715)] = {
    [sym_block_comment] = STATE(3715),
    [sym_compiler_directive_decl] = STATE(3715),
    [sym_fsi_directive_decl] = STATE(3715),
    [sym_preproc_line] = STATE(3715),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8152),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3716)] = {
    [sym_block_comment] = STATE(3716),
    [sym_compiler_directive_decl] = STATE(3716),
    [sym_fsi_directive_decl] = STATE(3716),
    [sym_preproc_line] = STATE(3716),
    [ts_builtin_sym_end] = ACTIONS(6556),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_namespace] = ACTIONS(6554),
    [anon_sym_module] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_open] = ACTIONS(6554),
    [anon_sym_type] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_exception] = ACTIONS(6554),
    [anon_sym_extern] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6556),
    [anon_sym_POUNDload] = ACTIONS(6556),
    [anon_sym_POUNDtime] = ACTIONS(6556),
    [anon_sym_POUNDI] = ACTIONS(6556),
    [anon_sym_POUNDhelp] = ACTIONS(6556),
    [anon_sym_POUNDquit] = ACTIONS(6556),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(3717)] = {
    [sym_block_comment] = STATE(3717),
    [sym_compiler_directive_decl] = STATE(3717),
    [sym_fsi_directive_decl] = STATE(3717),
    [sym_preproc_line] = STATE(3717),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6682),
    [aux_sym_decimal_token1] = ACTIONS(6484),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3718)] = {
    [sym_block_comment] = STATE(3718),
    [sym_compiler_directive_decl] = STATE(3718),
    [sym_fsi_directive_decl] = STATE(3718),
    [sym_preproc_line] = STATE(3718),
    [ts_builtin_sym_end] = ACTIONS(6614),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_namespace] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3719)] = {
    [sym_block_comment] = STATE(3719),
    [sym_compiler_directive_decl] = STATE(3719),
    [sym_fsi_directive_decl] = STATE(3719),
    [sym_preproc_line] = STATE(3719),
    [aux_sym_long_identifier_repeat1] = STATE(3726),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8154),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3720)] = {
    [sym_block_comment] = STATE(3720),
    [sym_compiler_directive_decl] = STATE(3720),
    [sym_fsi_directive_decl] = STATE(3720),
    [sym_preproc_line] = STATE(3720),
    [aux_sym_rules_repeat1] = STATE(3720),
    [ts_builtin_sym_end] = ACTIONS(6651),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_namespace] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(8156),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(8159),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3721)] = {
    [sym_block_comment] = STATE(3721),
    [sym_compiler_directive_decl] = STATE(3721),
    [sym_fsi_directive_decl] = STATE(3721),
    [sym_preproc_line] = STATE(3721),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_GT_RBRACK] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_RBRACK] = ACTIONS(6371),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_RBRACE] = ACTIONS(6371),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_to] = ACTIONS(6369),
    [anon_sym_downto] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_DOT_DOT2] = ACTIONS(6371),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3722)] = {
    [sym_block_comment] = STATE(3722),
    [sym_compiler_directive_decl] = STATE(3722),
    [sym_fsi_directive_decl] = STATE(3722),
    [sym_preproc_line] = STATE(3722),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_GT_RBRACK] = ACTIONS(6375),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_RBRACK] = ACTIONS(6375),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_RBRACE] = ACTIONS(6375),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_to] = ACTIONS(6373),
    [anon_sym_downto] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_DOT_DOT2] = ACTIONS(6375),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8162),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(3723)] = {
    [sym_block_comment] = STATE(3723),
    [sym_compiler_directive_decl] = STATE(3723),
    [sym_fsi_directive_decl] = STATE(3723),
    [sym_preproc_line] = STATE(3723),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3724)] = {
    [sym_block_comment] = STATE(3724),
    [sym_compiler_directive_decl] = STATE(3724),
    [sym_fsi_directive_decl] = STATE(3724),
    [sym_preproc_line] = STATE(3724),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_module] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_open] = ACTIONS(6554),
    [anon_sym_type] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_exception] = ACTIONS(6554),
    [anon_sym_extern] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6556),
    [anon_sym_POUNDload] = ACTIONS(6556),
    [anon_sym_POUNDtime] = ACTIONS(6556),
    [anon_sym_POUNDI] = ACTIONS(6556),
    [anon_sym_POUNDhelp] = ACTIONS(6556),
    [anon_sym_POUNDquit] = ACTIONS(6556),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [anon_sym_POUNDelse] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(3725)] = {
    [sym_block_comment] = STATE(3725),
    [sym_compiler_directive_decl] = STATE(3725),
    [sym_fsi_directive_decl] = STATE(3725),
    [sym_preproc_line] = STATE(3725),
    [ts_builtin_sym_end] = ACTIONS(6599),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_namespace] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8164),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3726)] = {
    [sym_block_comment] = STATE(3726),
    [sym_compiler_directive_decl] = STATE(3726),
    [sym_fsi_directive_decl] = STATE(3726),
    [sym_preproc_line] = STATE(3726),
    [aux_sym_long_identifier_repeat1] = STATE(3726),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8166),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3727)] = {
    [sym_block_comment] = STATE(3727),
    [sym_compiler_directive_decl] = STATE(3727),
    [sym_fsi_directive_decl] = STATE(3727),
    [sym_preproc_line] = STATE(3727),
    [aux_sym_long_identifier_repeat1] = STATE(3735),
    [ts_builtin_sym_end] = ACTIONS(6628),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_namespace] = ACTIONS(6625),
    [anon_sym_module] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_open] = ACTIONS(6625),
    [anon_sym_type] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8169),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_exception] = ACTIONS(6625),
    [anon_sym_extern] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6628),
    [anon_sym_POUNDload] = ACTIONS(6628),
    [anon_sym_POUNDtime] = ACTIONS(6628),
    [anon_sym_POUNDI] = ACTIONS(6628),
    [anon_sym_POUNDhelp] = ACTIONS(6628),
    [anon_sym_POUNDquit] = ACTIONS(6628),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3728)] = {
    [sym_block_comment] = STATE(3728),
    [sym_compiler_directive_decl] = STATE(3728),
    [sym_fsi_directive_decl] = STATE(3728),
    [sym_preproc_line] = STATE(3728),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_module] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_open] = ACTIONS(6560),
    [anon_sym_type] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_exception] = ACTIONS(6560),
    [anon_sym_extern] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6558),
    [anon_sym_POUNDload] = ACTIONS(6558),
    [anon_sym_POUNDtime] = ACTIONS(6558),
    [anon_sym_POUNDI] = ACTIONS(6558),
    [anon_sym_POUNDhelp] = ACTIONS(6558),
    [anon_sym_POUNDquit] = ACTIONS(6558),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [anon_sym_POUNDelse] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3729)] = {
    [sym_block_comment] = STATE(3729),
    [sym_compiler_directive_decl] = STATE(3729),
    [sym_fsi_directive_decl] = STATE(3729),
    [sym_preproc_line] = STATE(3729),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_GT_RBRACK] = ACTIONS(6405),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_RBRACK] = ACTIONS(6405),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_RBRACE] = ACTIONS(6405),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_to] = ACTIONS(6403),
    [anon_sym_downto] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_DOT_DOT2] = ACTIONS(6405),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(3730)] = {
    [sym_block_comment] = STATE(3730),
    [sym_compiler_directive_decl] = STATE(3730),
    [sym_fsi_directive_decl] = STATE(3730),
    [sym_preproc_line] = STATE(3730),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_GT_RBRACK] = ACTIONS(6385),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_RBRACK] = ACTIONS(6385),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_RBRACE] = ACTIONS(6385),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_to] = ACTIONS(6383),
    [anon_sym_downto] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_DOT_DOT2] = ACTIONS(6385),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(3731)] = {
    [sym_block_comment] = STATE(3731),
    [sym_compiler_directive_decl] = STATE(3731),
    [sym_fsi_directive_decl] = STATE(3731),
    [sym_preproc_line] = STATE(3731),
    [aux_sym_rules_repeat1] = STATE(3740),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_module] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_open] = ACTIONS(6618),
    [anon_sym_type] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(8173),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_exception] = ACTIONS(6618),
    [anon_sym_extern] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6616),
    [anon_sym_POUNDload] = ACTIONS(6616),
    [anon_sym_POUNDtime] = ACTIONS(6616),
    [anon_sym_POUNDI] = ACTIONS(6616),
    [anon_sym_POUNDhelp] = ACTIONS(6616),
    [anon_sym_POUNDquit] = ACTIONS(6616),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [anon_sym_POUNDelse] = ACTIONS(6616),
    [sym__newline] = ACTIONS(8175),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(3732)] = {
    [sym_block_comment] = STATE(3732),
    [sym_compiler_directive_decl] = STATE(3732),
    [sym_fsi_directive_decl] = STATE(3732),
    [sym_preproc_line] = STATE(3732),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_GT_RBRACK] = ACTIONS(6397),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_RBRACK] = ACTIONS(6397),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_RBRACE] = ACTIONS(6397),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_to] = ACTIONS(6395),
    [anon_sym_downto] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_DOT_DOT2] = ACTIONS(6397),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(3733)] = {
    [sym_block_comment] = STATE(3733),
    [sym_compiler_directive_decl] = STATE(3733),
    [sym_fsi_directive_decl] = STATE(3733),
    [sym_preproc_line] = STATE(3733),
    [ts_builtin_sym_end] = ACTIONS(6614),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_namespace] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3734)] = {
    [sym_block_comment] = STATE(3734),
    [sym_compiler_directive_decl] = STATE(3734),
    [sym_fsi_directive_decl] = STATE(3734),
    [sym_preproc_line] = STATE(3734),
    [aux_sym_rules_repeat1] = STATE(3711),
    [ts_builtin_sym_end] = ACTIONS(6605),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_namespace] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8137),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8178),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3735)] = {
    [sym_block_comment] = STATE(3735),
    [sym_compiler_directive_decl] = STATE(3735),
    [sym_fsi_directive_decl] = STATE(3735),
    [sym_preproc_line] = STATE(3735),
    [aux_sym_long_identifier_repeat1] = STATE(3769),
    [ts_builtin_sym_end] = ACTIONS(6358),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_namespace] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8181),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3736)] = {
    [sym_block_comment] = STATE(3736),
    [sym_compiler_directive_decl] = STATE(3736),
    [sym_fsi_directive_decl] = STATE(3736),
    [sym_preproc_line] = STATE(3736),
    [ts_builtin_sym_end] = ACTIONS(6635),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_namespace] = ACTIONS(6637),
    [anon_sym_module] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_open] = ACTIONS(6637),
    [anon_sym_type] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_exception] = ACTIONS(6637),
    [anon_sym_extern] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(8183),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6635),
    [anon_sym_POUNDload] = ACTIONS(6635),
    [anon_sym_POUNDtime] = ACTIONS(6635),
    [anon_sym_POUNDI] = ACTIONS(6635),
    [anon_sym_POUNDhelp] = ACTIONS(6635),
    [anon_sym_POUNDquit] = ACTIONS(6635),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3737)] = {
    [sym_block_comment] = STATE(3737),
    [sym_compiler_directive_decl] = STATE(3737),
    [sym_fsi_directive_decl] = STATE(3737),
    [sym_preproc_line] = STATE(3737),
    [aux_sym_rules_repeat1] = STATE(3720),
    [ts_builtin_sym_end] = ACTIONS(6605),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_namespace] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8137),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8178),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3738)] = {
    [sym_block_comment] = STATE(3738),
    [sym_compiler_directive_decl] = STATE(3738),
    [sym_fsi_directive_decl] = STATE(3738),
    [sym_preproc_line] = STATE(3738),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_GT_RBRACK] = ACTIONS(6371),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_RBRACK] = ACTIONS(6371),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_RBRACE] = ACTIONS(6371),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_to] = ACTIONS(6369),
    [anon_sym_downto] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_DOT_DOT2] = ACTIONS(6371),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3739)] = {
    [sym_block_comment] = STATE(3739),
    [sym_compiler_directive_decl] = STATE(3739),
    [sym_fsi_directive_decl] = STATE(3739),
    [sym_preproc_line] = STATE(3739),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_GT_RBRACK] = ACTIONS(6365),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_RBRACK] = ACTIONS(6365),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_RBRACE] = ACTIONS(6365),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_to] = ACTIONS(6367),
    [anon_sym_downto] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_DOT_DOT2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(3740)] = {
    [sym_block_comment] = STATE(3740),
    [sym_compiler_directive_decl] = STATE(3740),
    [sym_fsi_directive_decl] = STATE(3740),
    [sym_preproc_line] = STATE(3740),
    [aux_sym_rules_repeat1] = STATE(3740),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(8185),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(8188),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3741)] = {
    [sym_type_arguments] = STATE(4281),
    [sym_long_identifier] = STATE(4282),
    [sym_block_comment] = STATE(3741),
    [sym_compiler_directive_decl] = STATE(3741),
    [sym_fsi_directive_decl] = STATE(3741),
    [sym_preproc_line] = STATE(3741),
    [aux_sym_compound_type_repeat1] = STATE(3935),
    [sym_identifier] = ACTIONS(8142),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8144),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6193),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_DOT_DOT2] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(8146),
    [anon_sym_LT] = ACTIONS(8148),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8150),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3742)] = {
    [sym_block_comment] = STATE(3742),
    [sym_compiler_directive_decl] = STATE(3742),
    [sym_fsi_directive_decl] = STATE(3742),
    [sym_preproc_line] = STATE(3742),
    [aux_sym_rules_repeat1] = STATE(3731),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8173),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8191),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3743)] = {
    [sym_block_comment] = STATE(3743),
    [sym_compiler_directive_decl] = STATE(3743),
    [sym_fsi_directive_decl] = STATE(3743),
    [sym_preproc_line] = STATE(3743),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3744)] = {
    [sym_block_comment] = STATE(3744),
    [sym_compiler_directive_decl] = STATE(3744),
    [sym_fsi_directive_decl] = STATE(3744),
    [sym_preproc_line] = STATE(3744),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_GT_RBRACK] = ACTIONS(6490),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_RBRACK] = ACTIONS(6490),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_RBRACE] = ACTIONS(6490),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_to] = ACTIONS(6492),
    [anon_sym_downto] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_DOT_DOT2] = ACTIONS(6490),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(3745)] = {
    [sym_block_comment] = STATE(3745),
    [sym_compiler_directive_decl] = STATE(3745),
    [sym_fsi_directive_decl] = STATE(3745),
    [sym_preproc_line] = STATE(3745),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3746)] = {
    [sym_elif_expression] = STATE(4037),
    [sym_block_comment] = STATE(3746),
    [sym_compiler_directive_decl] = STATE(3746),
    [sym_fsi_directive_decl] = STATE(3746),
    [sym_preproc_line] = STATE(3746),
    [aux_sym_if_expression_repeat1] = STATE(3746),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_GT_RBRACK] = ACTIONS(6494),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_RBRACK] = ACTIONS(6494),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_RBRACE] = ACTIONS(6494),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_to] = ACTIONS(6496),
    [anon_sym_downto] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8194),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_DOT_DOT2] = ACTIONS(6494),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(3747)] = {
    [sym_block_comment] = STATE(3747),
    [sym_compiler_directive_decl] = STATE(3747),
    [sym_fsi_directive_decl] = STATE(3747),
    [sym_preproc_line] = STATE(3747),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_GT_RBRACK] = ACTIONS(6429),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_RBRACK] = ACTIONS(6429),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_RBRACE] = ACTIONS(6429),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_to] = ACTIONS(6427),
    [anon_sym_downto] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_DOT_DOT2] = ACTIONS(6429),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(3748)] = {
    [sym_block_comment] = STATE(3748),
    [sym_compiler_directive_decl] = STATE(3748),
    [sym_fsi_directive_decl] = STATE(3748),
    [sym_preproc_line] = STATE(3748),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_GT_RBRACK] = ACTIONS(6387),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_RBRACK] = ACTIONS(6387),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_RBRACE] = ACTIONS(6387),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_to] = ACTIONS(6389),
    [anon_sym_downto] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_DOT_DOT2] = ACTIONS(6387),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(3749)] = {
    [sym_block_comment] = STATE(3749),
    [sym_compiler_directive_decl] = STATE(3749),
    [sym_fsi_directive_decl] = STATE(3749),
    [sym_preproc_line] = STATE(3749),
    [aux_sym_long_identifier_repeat1] = STATE(3719),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_module] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_open] = ACTIONS(6625),
    [anon_sym_type] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8197),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_exception] = ACTIONS(6625),
    [anon_sym_extern] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6628),
    [anon_sym_POUNDload] = ACTIONS(6628),
    [anon_sym_POUNDtime] = ACTIONS(6628),
    [anon_sym_POUNDI] = ACTIONS(6628),
    [anon_sym_POUNDhelp] = ACTIONS(6628),
    [anon_sym_POUNDquit] = ACTIONS(6628),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [anon_sym_POUNDelse] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3750)] = {
    [sym_block_comment] = STATE(3750),
    [sym_compiler_directive_decl] = STATE(3750),
    [sym_fsi_directive_decl] = STATE(3750),
    [sym_preproc_line] = STATE(3750),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_GT_RBRACK] = ACTIONS(6381),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_RBRACK] = ACTIONS(6381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_RBRACE] = ACTIONS(6381),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_to] = ACTIONS(6379),
    [anon_sym_downto] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_DOT_DOT2] = ACTIONS(6381),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(3751)] = {
    [sym_block_comment] = STATE(3751),
    [sym_compiler_directive_decl] = STATE(3751),
    [sym_fsi_directive_decl] = STATE(3751),
    [sym_preproc_line] = STATE(3751),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_GT_RBRACK] = ACTIONS(6401),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_RBRACK] = ACTIONS(6401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_RBRACE] = ACTIONS(6401),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_to] = ACTIONS(6399),
    [anon_sym_downto] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_DOT_DOT2] = ACTIONS(6401),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(3752)] = {
    [sym_block_comment] = STATE(3752),
    [sym_compiler_directive_decl] = STATE(3752),
    [sym_fsi_directive_decl] = STATE(3752),
    [sym_preproc_line] = STATE(3752),
    [ts_builtin_sym_end] = ACTIONS(6562),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_namespace] = ACTIONS(6564),
    [anon_sym_module] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_open] = ACTIONS(6564),
    [anon_sym_type] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_exception] = ACTIONS(6564),
    [anon_sym_extern] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6562),
    [anon_sym_POUNDload] = ACTIONS(6562),
    [anon_sym_POUNDtime] = ACTIONS(6562),
    [anon_sym_POUNDI] = ACTIONS(6562),
    [anon_sym_POUNDhelp] = ACTIONS(6562),
    [anon_sym_POUNDquit] = ACTIONS(6562),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(3753)] = {
    [sym_block_comment] = STATE(3753),
    [sym_compiler_directive_decl] = STATE(3753),
    [sym_fsi_directive_decl] = STATE(3753),
    [sym_preproc_line] = STATE(3753),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_GT_RBRACK] = ACTIONS(6409),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_RBRACK] = ACTIONS(6409),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_RBRACE] = ACTIONS(6409),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_to] = ACTIONS(6407),
    [anon_sym_downto] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_DOT_DOT2] = ACTIONS(6409),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(3754)] = {
    [sym_block_comment] = STATE(3754),
    [sym_compiler_directive_decl] = STATE(3754),
    [sym_fsi_directive_decl] = STATE(3754),
    [sym_preproc_line] = STATE(3754),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_module] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_open] = ACTIONS(6564),
    [anon_sym_type] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_exception] = ACTIONS(6564),
    [anon_sym_extern] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6562),
    [anon_sym_POUNDload] = ACTIONS(6562),
    [anon_sym_POUNDtime] = ACTIONS(6562),
    [anon_sym_POUNDI] = ACTIONS(6562),
    [anon_sym_POUNDhelp] = ACTIONS(6562),
    [anon_sym_POUNDquit] = ACTIONS(6562),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [anon_sym_POUNDelse] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(3755)] = {
    [sym_block_comment] = STATE(3755),
    [sym_compiler_directive_decl] = STATE(3755),
    [sym_fsi_directive_decl] = STATE(3755),
    [sym_preproc_line] = STATE(3755),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_module] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_open] = ACTIONS(6637),
    [anon_sym_type] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_exception] = ACTIONS(6637),
    [anon_sym_extern] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(8201),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6635),
    [anon_sym_POUNDload] = ACTIONS(6635),
    [anon_sym_POUNDtime] = ACTIONS(6635),
    [anon_sym_POUNDI] = ACTIONS(6635),
    [anon_sym_POUNDhelp] = ACTIONS(6635),
    [anon_sym_POUNDquit] = ACTIONS(6635),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [anon_sym_POUNDelse] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3756)] = {
    [sym_block_comment] = STATE(3756),
    [sym_compiler_directive_decl] = STATE(3756),
    [sym_fsi_directive_decl] = STATE(3756),
    [sym_preproc_line] = STATE(3756),
    [aux_sym_rules_repeat1] = STATE(3740),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8173),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8191),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3757)] = {
    [sym_block_comment] = STATE(3757),
    [sym_compiler_directive_decl] = STATE(3757),
    [sym_fsi_directive_decl] = STATE(3757),
    [sym_preproc_line] = STATE(3757),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_GT_RBRACK] = ACTIONS(6413),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_RBRACK] = ACTIONS(6413),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_RBRACE] = ACTIONS(6413),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_to] = ACTIONS(6411),
    [anon_sym_downto] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_DOT_DOT2] = ACTIONS(6413),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(3758)] = {
    [sym_block_comment] = STATE(3758),
    [sym_compiler_directive_decl] = STATE(3758),
    [sym_fsi_directive_decl] = STATE(3758),
    [sym_preproc_line] = STATE(3758),
    [ts_builtin_sym_end] = ACTIONS(6515),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_namespace] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3759)] = {
    [sym_block_comment] = STATE(3759),
    [sym_compiler_directive_decl] = STATE(3759),
    [sym_fsi_directive_decl] = STATE(3759),
    [sym_preproc_line] = STATE(3759),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_GT_RBRACK] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_RBRACK] = ACTIONS(6393),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_RBRACE] = ACTIONS(6393),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_to] = ACTIONS(6391),
    [anon_sym_downto] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_DOT_DOT2] = ACTIONS(6393),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(3760)] = {
    [sym_type_arguments] = STATE(4281),
    [sym_long_identifier] = STATE(4282),
    [sym_block_comment] = STATE(3760),
    [sym_compiler_directive_decl] = STATE(3760),
    [sym_fsi_directive_decl] = STATE(3760),
    [sym_preproc_line] = STATE(3760),
    [aux_sym_compound_type_repeat1] = STATE(3935),
    [sym_identifier] = ACTIONS(8142),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8144),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_DOT_DOT] = ACTIONS(6261),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_DOT_DOT2] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(8146),
    [anon_sym_LT] = ACTIONS(8148),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8150),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__dedent] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3761)] = {
    [sym_block_comment] = STATE(3761),
    [sym_compiler_directive_decl] = STATE(3761),
    [sym_fsi_directive_decl] = STATE(3761),
    [sym_preproc_line] = STATE(3761),
    [aux_sym_rules_repeat1] = STATE(3737),
    [ts_builtin_sym_end] = ACTIONS(6588),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_namespace] = ACTIONS(6590),
    [anon_sym_module] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_open] = ACTIONS(6590),
    [anon_sym_type] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(8137),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_exception] = ACTIONS(6590),
    [anon_sym_extern] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6588),
    [anon_sym_POUNDload] = ACTIONS(6588),
    [anon_sym_POUNDtime] = ACTIONS(6588),
    [anon_sym_POUNDI] = ACTIONS(6588),
    [anon_sym_POUNDhelp] = ACTIONS(6588),
    [anon_sym_POUNDquit] = ACTIONS(6588),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(8203),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(3762)] = {
    [sym_block_comment] = STATE(3762),
    [sym_compiler_directive_decl] = STATE(3762),
    [sym_fsi_directive_decl] = STATE(3762),
    [sym_preproc_line] = STATE(3762),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_GT_RBRACK] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_RBRACK] = ACTIONS(6297),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_RBRACE] = ACTIONS(6297),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_to] = ACTIONS(6295),
    [anon_sym_downto] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3763)] = {
    [sym_block_comment] = STATE(3763),
    [sym_compiler_directive_decl] = STATE(3763),
    [sym_fsi_directive_decl] = STATE(3763),
    [sym_preproc_line] = STATE(3763),
    [ts_builtin_sym_end] = ACTIONS(6558),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_namespace] = ACTIONS(6560),
    [anon_sym_module] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_open] = ACTIONS(6560),
    [anon_sym_type] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_exception] = ACTIONS(6560),
    [anon_sym_extern] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6558),
    [anon_sym_POUNDload] = ACTIONS(6558),
    [anon_sym_POUNDtime] = ACTIONS(6558),
    [anon_sym_POUNDI] = ACTIONS(6558),
    [anon_sym_POUNDhelp] = ACTIONS(6558),
    [anon_sym_POUNDquit] = ACTIONS(6558),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3764)] = {
    [sym_type_arguments] = STATE(4281),
    [sym_long_identifier] = STATE(4282),
    [sym_block_comment] = STATE(3764),
    [sym_compiler_directive_decl] = STATE(3764),
    [sym_fsi_directive_decl] = STATE(3764),
    [sym_preproc_line] = STATE(3764),
    [aux_sym_compound_type_repeat1] = STATE(3935),
    [sym_identifier] = ACTIONS(8142),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8144),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_DOT_DOT] = ACTIONS(6213),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_DOT_DOT2] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(8146),
    [anon_sym_LT] = ACTIONS(8148),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8150),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__dedent] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3765)] = {
    [sym_block_comment] = STATE(3765),
    [sym_compiler_directive_decl] = STATE(3765),
    [sym_fsi_directive_decl] = STATE(3765),
    [sym_preproc_line] = STATE(3765),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8206),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3766)] = {
    [sym_block_comment] = STATE(3766),
    [sym_compiler_directive_decl] = STATE(3766),
    [sym_fsi_directive_decl] = STATE(3766),
    [sym_preproc_line] = STATE(3766),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_GT_RBRACK] = ACTIONS(6418),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_RBRACK] = ACTIONS(6418),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_RBRACE] = ACTIONS(6418),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_to] = ACTIONS(6420),
    [anon_sym_downto] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_DOT_DOT2] = ACTIONS(6418),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(3767)] = {
    [sym_block_comment] = STATE(3767),
    [sym_compiler_directive_decl] = STATE(3767),
    [sym_fsi_directive_decl] = STATE(3767),
    [sym_preproc_line] = STATE(3767),
    [aux_sym_rules_repeat1] = STATE(3756),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_module] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_open] = ACTIONS(6590),
    [anon_sym_type] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(8173),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_exception] = ACTIONS(6590),
    [anon_sym_extern] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6588),
    [anon_sym_POUNDload] = ACTIONS(6588),
    [anon_sym_POUNDtime] = ACTIONS(6588),
    [anon_sym_POUNDI] = ACTIONS(6588),
    [anon_sym_POUNDhelp] = ACTIONS(6588),
    [anon_sym_POUNDquit] = ACTIONS(6588),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [anon_sym_POUNDelse] = ACTIONS(6588),
    [sym__newline] = ACTIONS(8208),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(3768)] = {
    [sym_block_comment] = STATE(3768),
    [sym_compiler_directive_decl] = STATE(3768),
    [sym_fsi_directive_decl] = STATE(3768),
    [sym_preproc_line] = STATE(3768),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6666),
    [aux_sym_decimal_token1] = ACTIONS(6456),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3769)] = {
    [sym_block_comment] = STATE(3769),
    [sym_compiler_directive_decl] = STATE(3769),
    [sym_fsi_directive_decl] = STATE(3769),
    [sym_preproc_line] = STATE(3769),
    [aux_sym_long_identifier_repeat1] = STATE(3769),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8211),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3770)] = {
    [sym_block_comment] = STATE(3770),
    [sym_compiler_directive_decl] = STATE(3770),
    [sym_fsi_directive_decl] = STATE(3770),
    [sym_preproc_line] = STATE(3770),
    [ts_builtin_sym_end] = ACTIONS(6651),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_namespace] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3771)] = {
    [sym_block_comment] = STATE(3771),
    [sym_compiler_directive_decl] = STATE(3771),
    [sym_fsi_directive_decl] = STATE(3771),
    [sym_preproc_line] = STATE(3771),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_module] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_open] = ACTIONS(6513),
    [anon_sym_type] = ACTIONS(6513),
    [anon_sym_LBRACK_LT] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_exception] = ACTIONS(6513),
    [anon_sym_extern] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6515),
    [anon_sym_POUNDload] = ACTIONS(6515),
    [anon_sym_POUNDtime] = ACTIONS(6515),
    [anon_sym_POUNDI] = ACTIONS(6515),
    [anon_sym_POUNDhelp] = ACTIONS(6515),
    [anon_sym_POUNDquit] = ACTIONS(6515),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3772)] = {
    [sym_block_comment] = STATE(3772),
    [sym_compiler_directive_decl] = STATE(3772),
    [sym_fsi_directive_decl] = STATE(3772),
    [sym_preproc_line] = STATE(3772),
    [ts_builtin_sym_end] = ACTIONS(6832),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_namespace] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8214),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3773)] = {
    [sym_block_comment] = STATE(3773),
    [sym_compiler_directive_decl] = STATE(3773),
    [sym_fsi_directive_decl] = STATE(3773),
    [sym_preproc_line] = STATE(3773),
    [ts_builtin_sym_end] = ACTIONS(6694),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_namespace] = ACTIONS(6692),
    [anon_sym_module] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_open] = ACTIONS(6692),
    [anon_sym_type] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_exception] = ACTIONS(6692),
    [anon_sym_extern] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6694),
    [anon_sym_POUNDload] = ACTIONS(6694),
    [anon_sym_POUNDtime] = ACTIONS(6694),
    [anon_sym_POUNDI] = ACTIONS(6694),
    [anon_sym_POUNDhelp] = ACTIONS(6694),
    [anon_sym_POUNDquit] = ACTIONS(6694),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3774)] = {
    [sym_block_comment] = STATE(3774),
    [sym_compiler_directive_decl] = STATE(3774),
    [sym_fsi_directive_decl] = STATE(3774),
    [sym_preproc_line] = STATE(3774),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_module] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_open] = ACTIONS(6560),
    [anon_sym_type] = ACTIONS(6560),
    [anon_sym_LBRACK_LT] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_exception] = ACTIONS(6560),
    [anon_sym_extern] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6558),
    [anon_sym_POUNDload] = ACTIONS(6558),
    [anon_sym_POUNDtime] = ACTIONS(6558),
    [anon_sym_POUNDI] = ACTIONS(6558),
    [anon_sym_POUNDhelp] = ACTIONS(6558),
    [anon_sym_POUNDquit] = ACTIONS(6558),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(3775)] = {
    [sym_type_arguments] = STATE(4464),
    [sym_long_identifier] = STATE(4465),
    [sym_block_comment] = STATE(3775),
    [sym_compiler_directive_decl] = STATE(3775),
    [sym_fsi_directive_decl] = STATE(3775),
    [sym_preproc_line] = STATE(3775),
    [aux_sym_compound_type_repeat1] = STATE(4261),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_as] = ACTIONS(6265),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8216),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8218),
    [anon_sym_LT] = ACTIONS(8220),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8222),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3776)] = {
    [sym_block_comment] = STATE(3776),
    [sym_compiler_directive_decl] = STATE(3776),
    [sym_fsi_directive_decl] = STATE(3776),
    [sym_preproc_line] = STATE(3776),
    [ts_builtin_sym_end] = ACTIONS(6838),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_namespace] = ACTIONS(6840),
    [anon_sym_module] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_open] = ACTIONS(6840),
    [anon_sym_type] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_exception] = ACTIONS(6840),
    [anon_sym_extern] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6838),
    [anon_sym_POUNDload] = ACTIONS(6838),
    [anon_sym_POUNDtime] = ACTIONS(6838),
    [anon_sym_POUNDI] = ACTIONS(6838),
    [anon_sym_POUNDhelp] = ACTIONS(6838),
    [anon_sym_POUNDquit] = ACTIONS(6838),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3777)] = {
    [sym_block_comment] = STATE(3777),
    [sym_compiler_directive_decl] = STATE(3777),
    [sym_fsi_directive_decl] = STATE(3777),
    [sym_preproc_line] = STATE(3777),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_module] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_open] = ACTIONS(6871),
    [anon_sym_type] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_exception] = ACTIONS(6871),
    [anon_sym_extern] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6869),
    [anon_sym_POUNDload] = ACTIONS(6869),
    [anon_sym_POUNDtime] = ACTIONS(6869),
    [anon_sym_POUNDI] = ACTIONS(6869),
    [anon_sym_POUNDhelp] = ACTIONS(6869),
    [anon_sym_POUNDquit] = ACTIONS(6869),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [anon_sym_POUNDelse] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3778)] = {
    [sym_block_comment] = STATE(3778),
    [sym_compiler_directive_decl] = STATE(3778),
    [sym_fsi_directive_decl] = STATE(3778),
    [sym_preproc_line] = STATE(3778),
    [ts_builtin_sym_end] = ACTIONS(6766),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_namespace] = ACTIONS(6768),
    [anon_sym_module] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_open] = ACTIONS(6768),
    [anon_sym_type] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8224),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_exception] = ACTIONS(6768),
    [anon_sym_extern] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6766),
    [anon_sym_POUNDload] = ACTIONS(6766),
    [anon_sym_POUNDtime] = ACTIONS(6766),
    [anon_sym_POUNDI] = ACTIONS(6766),
    [anon_sym_POUNDhelp] = ACTIONS(6766),
    [anon_sym_POUNDquit] = ACTIONS(6766),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3779)] = {
    [sym_block_comment] = STATE(3779),
    [sym_compiler_directive_decl] = STATE(3779),
    [sym_fsi_directive_decl] = STATE(3779),
    [sym_preproc_line] = STATE(3779),
    [aux_sym_type_argument_repeat1] = STATE(3825),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_as] = ACTIONS(6331),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8226),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3780)] = {
    [sym_block_comment] = STATE(3780),
    [sym_compiler_directive_decl] = STATE(3780),
    [sym_fsi_directive_decl] = STATE(3780),
    [sym_preproc_line] = STATE(3780),
    [ts_builtin_sym_end] = ACTIONS(6811),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_namespace] = ACTIONS(6813),
    [anon_sym_module] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_open] = ACTIONS(6813),
    [anon_sym_type] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_exception] = ACTIONS(6813),
    [anon_sym_extern] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6811),
    [anon_sym_POUNDload] = ACTIONS(6811),
    [anon_sym_POUNDtime] = ACTIONS(6811),
    [anon_sym_POUNDI] = ACTIONS(6811),
    [anon_sym_POUNDhelp] = ACTIONS(6811),
    [anon_sym_POUNDquit] = ACTIONS(6811),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3781)] = {
    [sym_type_arguments] = STATE(4482),
    [sym_long_identifier] = STATE(4483),
    [sym_block_comment] = STATE(3781),
    [sym_compiler_directive_decl] = STATE(3781),
    [sym_fsi_directive_decl] = STATE(3781),
    [sym_preproc_line] = STATE(3781),
    [aux_sym_compound_type_repeat1] = STATE(4270),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8229),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8231),
    [anon_sym_LT] = ACTIONS(8233),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8235),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [anon_sym_POUNDelse] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3782)] = {
    [sym_block_comment] = STATE(3782),
    [sym_compiler_directive_decl] = STATE(3782),
    [sym_fsi_directive_decl] = STATE(3782),
    [sym_preproc_line] = STATE(3782),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8237),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3783)] = {
    [sym_block_comment] = STATE(3783),
    [sym_compiler_directive_decl] = STATE(3783),
    [sym_fsi_directive_decl] = STATE(3783),
    [sym_preproc_line] = STATE(3783),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8152),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3784)] = {
    [sym_block_comment] = STATE(3784),
    [sym_compiler_directive_decl] = STATE(3784),
    [sym_fsi_directive_decl] = STATE(3784),
    [sym_preproc_line] = STATE(3784),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3785)] = {
    [sym_block_comment] = STATE(3785),
    [sym_compiler_directive_decl] = STATE(3785),
    [sym_fsi_directive_decl] = STATE(3785),
    [sym_preproc_line] = STATE(3785),
    [ts_builtin_sym_end] = ACTIONS(6822),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_namespace] = ACTIONS(6824),
    [anon_sym_module] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_open] = ACTIONS(6824),
    [anon_sym_type] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_exception] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6822),
    [anon_sym_POUNDload] = ACTIONS(6822),
    [anon_sym_POUNDtime] = ACTIONS(6822),
    [anon_sym_POUNDI] = ACTIONS(6822),
    [anon_sym_POUNDhelp] = ACTIONS(6822),
    [anon_sym_POUNDquit] = ACTIONS(6822),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3786)] = {
    [sym_block_comment] = STATE(3786),
    [sym_compiler_directive_decl] = STATE(3786),
    [sym_fsi_directive_decl] = STATE(3786),
    [sym_preproc_line] = STATE(3786),
    [aux_sym_compound_type_repeat1] = STATE(3787),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_as] = ACTIONS(6338),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_DOT_DOT] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8102),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__dedent] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(3787)] = {
    [sym_block_comment] = STATE(3787),
    [sym_compiler_directive_decl] = STATE(3787),
    [sym_fsi_directive_decl] = STATE(3787),
    [sym_preproc_line] = STATE(3787),
    [aux_sym_compound_type_repeat1] = STATE(3787),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_as] = ACTIONS(6193),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8239),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3788)] = {
    [sym_block_comment] = STATE(3788),
    [sym_compiler_directive_decl] = STATE(3788),
    [sym_fsi_directive_decl] = STATE(3788),
    [sym_preproc_line] = STATE(3788),
    [ts_builtin_sym_end] = ACTIONS(6865),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_namespace] = ACTIONS(6867),
    [anon_sym_module] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_open] = ACTIONS(6867),
    [anon_sym_type] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_exception] = ACTIONS(6867),
    [anon_sym_extern] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6865),
    [anon_sym_POUNDload] = ACTIONS(6865),
    [anon_sym_POUNDtime] = ACTIONS(6865),
    [anon_sym_POUNDI] = ACTIONS(6865),
    [anon_sym_POUNDhelp] = ACTIONS(6865),
    [anon_sym_POUNDquit] = ACTIONS(6865),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3789)] = {
    [sym_block_comment] = STATE(3789),
    [sym_compiler_directive_decl] = STATE(3789),
    [sym_fsi_directive_decl] = STATE(3789),
    [sym_preproc_line] = STATE(3789),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3790)] = {
    [sym_attributes] = STATE(13274),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6891),
    [sym_access_modifier] = STATE(22393),
    [sym__class_function_or_value_defn] = STATE(6892),
    [aux_sym__type_extension_inner] = STATE(4219),
    [sym_type_extension_elements] = STATE(8128),
    [sym__type_defn_elements] = STATE(6892),
    [sym_interface_implementation] = STATE(6893),
    [sym_member_defn] = STATE(6893),
    [sym_additional_constr_defn] = STATE(6918),
    [sym_block_comment] = STATE(3790),
    [sym_compiler_directive_decl] = STATE(3790),
    [sym_fsi_directive_decl] = STATE(3790),
    [sym_preproc_line] = STATE(3790),
    [sym_preproc_if_in_class_definition] = STATE(6893),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8242),
    [anon_sym_module] = ACTIONS(8242),
    [anon_sym_null] = ACTIONS(8242),
    [anon_sym_begin] = ACTIONS(8242),
    [anon_sym_open] = ACTIONS(8242),
    [anon_sym_type] = ACTIONS(8242),
    [anon_sym_LBRACK_LT] = ACTIONS(8244),
    [anon_sym_do] = ACTIONS(8242),
    [anon_sym_and] = ACTIONS(8242),
    [anon_sym_let] = ACTIONS(8242),
    [anon_sym_let_BANG] = ACTIONS(8244),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8242),
    [anon_sym_AMP] = ACTIONS(8242),
    [anon_sym_LBRACK] = ACTIONS(8242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8244),
    [anon_sym_LBRACE] = ACTIONS(8242),
    [anon_sym_LT_AT] = ACTIONS(8242),
    [anon_sym_LT_AT_AT] = ACTIONS(8244),
    [anon_sym_CARET] = ACTIONS(8244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8244),
    [anon_sym_new] = ACTIONS(8242),
    [anon_sym_with] = ACTIONS(8248),
    [anon_sym_return] = ACTIONS(8242),
    [anon_sym_return_BANG] = ACTIONS(8244),
    [anon_sym_yield] = ACTIONS(8242),
    [anon_sym_yield_BANG] = ACTIONS(8244),
    [anon_sym_lazy] = ACTIONS(8242),
    [anon_sym_assert] = ACTIONS(8242),
    [anon_sym_upcast] = ACTIONS(8242),
    [anon_sym_downcast] = ACTIONS(8242),
    [anon_sym_fixed] = ACTIONS(8242),
    [anon_sym_for] = ACTIONS(8242),
    [anon_sym_while] = ACTIONS(8242),
    [anon_sym_while_BANG] = ACTIONS(8244),
    [anon_sym_if] = ACTIONS(8242),
    [anon_sym_fun] = ACTIONS(8242),
    [anon_sym_try] = ACTIONS(8242),
    [anon_sym_match] = ACTIONS(8242),
    [anon_sym_match_BANG] = ACTIONS(8244),
    [anon_sym_function] = ACTIONS(8242),
    [anon_sym_use] = ACTIONS(8242),
    [anon_sym_use_BANG] = ACTIONS(8244),
    [anon_sym_do_BANG] = ACTIONS(8244),
    [anon_sym_AMP_AMP] = ACTIONS(8244),
    [anon_sym_default] = ACTIONS(8250),
    [anon_sym_static] = ACTIONS(8252),
    [anon_sym_member] = ACTIONS(8254),
    [anon_sym_exception] = ACTIONS(8242),
    [anon_sym_interface] = ACTIONS(8256),
    [anon_sym_abstract] = ACTIONS(8258),
    [anon_sym_val] = ACTIONS(8260),
    [anon_sym_override] = ACTIONS(8250),
    [anon_sym_extern] = ACTIONS(8242),
    [aux_sym_char_token1] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8242),
    [anon_sym_DQUOTE] = ACTIONS(8242),
    [anon_sym_AT_DQUOTE] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [sym_bool] = ACTIONS(8242),
    [sym_unit] = ACTIONS(8244),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8242),
    [sym_op_identifier] = ACTIONS(8244),
    [anon_sym_PLUS] = ACTIONS(8242),
    [anon_sym_DASH] = ACTIONS(8242),
    [anon_sym_PLUS_DOT] = ACTIONS(8244),
    [anon_sym_DASH_DOT] = ACTIONS(8244),
    [anon_sym_PERCENT] = ACTIONS(8242),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8244),
    [anon_sym_TILDE] = ACTIONS(8244),
    [aux_sym_prefix_op_token1] = ACTIONS(8244),
    [sym_int] = ACTIONS(8242),
    [sym_xint] = ACTIONS(8244),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8244),
    [anon_sym_POUNDload] = ACTIONS(8244),
    [anon_sym_POUNDtime] = ACTIONS(8244),
    [anon_sym_POUNDI] = ACTIONS(8244),
    [anon_sym_POUNDhelp] = ACTIONS(8244),
    [anon_sym_POUNDquit] = ACTIONS(8244),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8244),
    [anon_sym_POUNDendif] = ACTIONS(8244),
    [anon_sym_POUNDelse] = ACTIONS(8244),
    [sym__newline] = ACTIONS(8244),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8244),
  },
  [STATE(3791)] = {
    [sym_block_comment] = STATE(3791),
    [sym_compiler_directive_decl] = STATE(3791),
    [sym_fsi_directive_decl] = STATE(3791),
    [sym_preproc_line] = STATE(3791),
    [aux_sym_long_identifier_repeat1] = STATE(3831),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8262),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3792)] = {
    [sym_block_comment] = STATE(3792),
    [sym_compiler_directive_decl] = STATE(3792),
    [sym_fsi_directive_decl] = STATE(3792),
    [sym_preproc_line] = STATE(3792),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_module] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_open] = ACTIONS(6875),
    [anon_sym_type] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_exception] = ACTIONS(6875),
    [anon_sym_extern] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6873),
    [anon_sym_POUNDload] = ACTIONS(6873),
    [anon_sym_POUNDtime] = ACTIONS(6873),
    [anon_sym_POUNDI] = ACTIONS(6873),
    [anon_sym_POUNDhelp] = ACTIONS(6873),
    [anon_sym_POUNDquit] = ACTIONS(6873),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [anon_sym_POUNDelse] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3793)] = {
    [sym_block_comment] = STATE(3793),
    [sym_compiler_directive_decl] = STATE(3793),
    [sym_fsi_directive_decl] = STATE(3793),
    [sym_preproc_line] = STATE(3793),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_module] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_open] = ACTIONS(6867),
    [anon_sym_type] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_exception] = ACTIONS(6867),
    [anon_sym_extern] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6865),
    [anon_sym_POUNDload] = ACTIONS(6865),
    [anon_sym_POUNDtime] = ACTIONS(6865),
    [anon_sym_POUNDI] = ACTIONS(6865),
    [anon_sym_POUNDhelp] = ACTIONS(6865),
    [anon_sym_POUNDquit] = ACTIONS(6865),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [anon_sym_POUNDelse] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(3794)] = {
    [sym_type_arguments] = STATE(4310),
    [sym_long_identifier] = STATE(4312),
    [sym_block_comment] = STATE(3794),
    [sym_compiler_directive_decl] = STATE(3794),
    [sym_fsi_directive_decl] = STATE(3794),
    [sym_preproc_line] = STATE(3794),
    [aux_sym_compound_type_repeat1] = STATE(4191),
    [sym_identifier] = ACTIONS(8264),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8266),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_DOT_DOT] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8268),
    [anon_sym_LT] = ACTIONS(8270),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8272),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__dedent] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3795)] = {
    [sym_block_comment] = STATE(3795),
    [sym_compiler_directive_decl] = STATE(3795),
    [sym_fsi_directive_decl] = STATE(3795),
    [sym_preproc_line] = STATE(3795),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_module] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_open] = ACTIONS(6790),
    [anon_sym_type] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_exception] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6792),
    [anon_sym_POUNDload] = ACTIONS(6792),
    [anon_sym_POUNDtime] = ACTIONS(6792),
    [anon_sym_POUNDI] = ACTIONS(6792),
    [anon_sym_POUNDhelp] = ACTIONS(6792),
    [anon_sym_POUNDquit] = ACTIONS(6792),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [anon_sym_POUNDelse] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3796)] = {
    [sym_type_arguments] = STATE(4532),
    [sym_long_identifier] = STATE(4523),
    [sym_block_comment] = STATE(3796),
    [sym_compiler_directive_decl] = STATE(3796),
    [sym_fsi_directive_decl] = STATE(3796),
    [sym_preproc_line] = STATE(3796),
    [aux_sym_compound_type_repeat1] = STATE(4288),
    [sym_identifier] = ACTIONS(8274),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8276),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8278),
    [anon_sym_LT] = ACTIONS(8280),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8282),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__dedent] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
    [sym__in] = ACTIONS(6209),
  },
  [STATE(3797)] = {
    [sym_block_comment] = STATE(3797),
    [sym_compiler_directive_decl] = STATE(3797),
    [sym_fsi_directive_decl] = STATE(3797),
    [sym_preproc_line] = STATE(3797),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3798)] = {
    [sym_block_comment] = STATE(3798),
    [sym_compiler_directive_decl] = STATE(3798),
    [sym_fsi_directive_decl] = STATE(3798),
    [sym_preproc_line] = STATE(3798),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8284),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3799)] = {
    [sym_type_arguments] = STATE(4532),
    [sym_long_identifier] = STATE(4523),
    [sym_block_comment] = STATE(3799),
    [sym_compiler_directive_decl] = STATE(3799),
    [sym_fsi_directive_decl] = STATE(3799),
    [sym_preproc_line] = STATE(3799),
    [aux_sym_compound_type_repeat1] = STATE(4288),
    [sym_identifier] = ACTIONS(8274),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8276),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8278),
    [anon_sym_LT] = ACTIONS(8280),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8282),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
    [sym__in] = ACTIONS(6195),
  },
  [STATE(3800)] = {
    [sym_type_arguments] = STATE(4532),
    [sym_long_identifier] = STATE(4523),
    [sym_block_comment] = STATE(3800),
    [sym_compiler_directive_decl] = STATE(3800),
    [sym_fsi_directive_decl] = STATE(3800),
    [sym_preproc_line] = STATE(3800),
    [aux_sym_compound_type_repeat1] = STATE(4288),
    [sym_identifier] = ACTIONS(8274),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8276),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8278),
    [anon_sym_LT] = ACTIONS(8280),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8282),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__dedent] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
    [sym__in] = ACTIONS(6259),
  },
  [STATE(3801)] = {
    [sym_block_comment] = STATE(3801),
    [sym_compiler_directive_decl] = STATE(3801),
    [sym_fsi_directive_decl] = STATE(3801),
    [sym_preproc_line] = STATE(3801),
    [aux_sym_rules_repeat1] = STATE(3867),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_module] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_open] = ACTIONS(6590),
    [anon_sym_type] = ACTIONS(6590),
    [anon_sym_LBRACK_LT] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(8286),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_exception] = ACTIONS(6590),
    [anon_sym_extern] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6588),
    [anon_sym_POUNDload] = ACTIONS(6588),
    [anon_sym_POUNDtime] = ACTIONS(6588),
    [anon_sym_POUNDI] = ACTIONS(6588),
    [anon_sym_POUNDhelp] = ACTIONS(6588),
    [anon_sym_POUNDquit] = ACTIONS(6588),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(8288),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(3802)] = {
    [sym_block_comment] = STATE(3802),
    [sym_compiler_directive_decl] = STATE(3802),
    [sym_fsi_directive_decl] = STATE(3802),
    [sym_preproc_line] = STATE(3802),
    [ts_builtin_sym_end] = ACTIONS(6786),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_namespace] = ACTIONS(6784),
    [anon_sym_module] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_open] = ACTIONS(6784),
    [anon_sym_type] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8291),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_exception] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6786),
    [anon_sym_POUNDload] = ACTIONS(6786),
    [anon_sym_POUNDtime] = ACTIONS(6786),
    [anon_sym_POUNDI] = ACTIONS(6786),
    [anon_sym_POUNDhelp] = ACTIONS(6786),
    [anon_sym_POUNDquit] = ACTIONS(6786),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3803)] = {
    [sym_block_comment] = STATE(3803),
    [sym_compiler_directive_decl] = STATE(3803),
    [sym_fsi_directive_decl] = STATE(3803),
    [sym_preproc_line] = STATE(3803),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_module] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_open] = ACTIONS(6794),
    [anon_sym_type] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_exception] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6796),
    [anon_sym_POUNDload] = ACTIONS(6796),
    [anon_sym_POUNDtime] = ACTIONS(6796),
    [anon_sym_POUNDI] = ACTIONS(6796),
    [anon_sym_POUNDhelp] = ACTIONS(6796),
    [anon_sym_POUNDquit] = ACTIONS(6796),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [anon_sym_POUNDelse] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3804)] = {
    [sym_block_comment] = STATE(3804),
    [sym_compiler_directive_decl] = STATE(3804),
    [sym_fsi_directive_decl] = STATE(3804),
    [sym_preproc_line] = STATE(3804),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_module] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_open] = ACTIONS(6768),
    [anon_sym_type] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8293),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_exception] = ACTIONS(6768),
    [anon_sym_extern] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6766),
    [anon_sym_POUNDload] = ACTIONS(6766),
    [anon_sym_POUNDtime] = ACTIONS(6766),
    [anon_sym_POUNDI] = ACTIONS(6766),
    [anon_sym_POUNDhelp] = ACTIONS(6766),
    [anon_sym_POUNDquit] = ACTIONS(6766),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [anon_sym_POUNDelse] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(3805)] = {
    [sym_block_comment] = STATE(3805),
    [sym_compiler_directive_decl] = STATE(3805),
    [sym_fsi_directive_decl] = STATE(3805),
    [sym_preproc_line] = STATE(3805),
    [ts_builtin_sym_end] = ACTIONS(6873),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_namespace] = ACTIONS(6875),
    [anon_sym_module] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_open] = ACTIONS(6875),
    [anon_sym_type] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_exception] = ACTIONS(6875),
    [anon_sym_extern] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6873),
    [anon_sym_POUNDload] = ACTIONS(6873),
    [anon_sym_POUNDtime] = ACTIONS(6873),
    [anon_sym_POUNDI] = ACTIONS(6873),
    [anon_sym_POUNDhelp] = ACTIONS(6873),
    [anon_sym_POUNDquit] = ACTIONS(6873),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3806)] = {
    [sym_block_comment] = STATE(3806),
    [sym_compiler_directive_decl] = STATE(3806),
    [sym_fsi_directive_decl] = STATE(3806),
    [sym_preproc_line] = STATE(3806),
    [ts_builtin_sym_end] = ACTIONS(6869),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_namespace] = ACTIONS(6871),
    [anon_sym_module] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_open] = ACTIONS(6871),
    [anon_sym_type] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_exception] = ACTIONS(6871),
    [anon_sym_extern] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6869),
    [anon_sym_POUNDload] = ACTIONS(6869),
    [anon_sym_POUNDtime] = ACTIONS(6869),
    [anon_sym_POUNDI] = ACTIONS(6869),
    [anon_sym_POUNDhelp] = ACTIONS(6869),
    [anon_sym_POUNDquit] = ACTIONS(6869),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3807)] = {
    [sym_block_comment] = STATE(3807),
    [sym_compiler_directive_decl] = STATE(3807),
    [sym_fsi_directive_decl] = STATE(3807),
    [sym_preproc_line] = STATE(3807),
    [ts_builtin_sym_end] = ACTIONS(6792),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_namespace] = ACTIONS(6790),
    [anon_sym_module] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_open] = ACTIONS(6790),
    [anon_sym_type] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_exception] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6792),
    [anon_sym_POUNDload] = ACTIONS(6792),
    [anon_sym_POUNDtime] = ACTIONS(6792),
    [anon_sym_POUNDI] = ACTIONS(6792),
    [anon_sym_POUNDhelp] = ACTIONS(6792),
    [anon_sym_POUNDquit] = ACTIONS(6792),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(3808)] = {
    [sym_block_comment] = STATE(3808),
    [sym_compiler_directive_decl] = STATE(3808),
    [sym_fsi_directive_decl] = STATE(3808),
    [sym_preproc_line] = STATE(3808),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_module] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_open] = ACTIONS(6750),
    [anon_sym_type] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_exception] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6752),
    [anon_sym_POUNDload] = ACTIONS(6752),
    [anon_sym_POUNDtime] = ACTIONS(6752),
    [anon_sym_POUNDI] = ACTIONS(6752),
    [anon_sym_POUNDhelp] = ACTIONS(6752),
    [anon_sym_POUNDquit] = ACTIONS(6752),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [anon_sym_POUNDelse] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3809)] = {
    [sym_block_comment] = STATE(3809),
    [sym_compiler_directive_decl] = STATE(3809),
    [sym_fsi_directive_decl] = STATE(3809),
    [sym_preproc_line] = STATE(3809),
    [ts_builtin_sym_end] = ACTIONS(6796),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_namespace] = ACTIONS(6794),
    [anon_sym_module] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_open] = ACTIONS(6794),
    [anon_sym_type] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_exception] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6796),
    [anon_sym_POUNDload] = ACTIONS(6796),
    [anon_sym_POUNDtime] = ACTIONS(6796),
    [anon_sym_POUNDI] = ACTIONS(6796),
    [anon_sym_POUNDhelp] = ACTIONS(6796),
    [anon_sym_POUNDquit] = ACTIONS(6796),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(3810)] = {
    [sym_block_comment] = STATE(3810),
    [sym_compiler_directive_decl] = STATE(3810),
    [sym_fsi_directive_decl] = STATE(3810),
    [sym_preproc_line] = STATE(3810),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_module] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_open] = ACTIONS(6692),
    [anon_sym_type] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_exception] = ACTIONS(6692),
    [anon_sym_extern] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6694),
    [anon_sym_POUNDload] = ACTIONS(6694),
    [anon_sym_POUNDtime] = ACTIONS(6694),
    [anon_sym_POUNDI] = ACTIONS(6694),
    [anon_sym_POUNDhelp] = ACTIONS(6694),
    [anon_sym_POUNDquit] = ACTIONS(6694),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [anon_sym_POUNDelse] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(3811)] = {
    [sym_block_comment] = STATE(3811),
    [sym_compiler_directive_decl] = STATE(3811),
    [sym_fsi_directive_decl] = STATE(3811),
    [sym_preproc_line] = STATE(3811),
    [aux_sym_sequential_expression_repeat1] = STATE(3846),
    [ts_builtin_sym_end] = ACTIONS(6738),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_namespace] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3812)] = {
    [sym_block_comment] = STATE(3812),
    [sym_compiler_directive_decl] = STATE(3812),
    [sym_fsi_directive_decl] = STATE(3812),
    [sym_preproc_line] = STATE(3812),
    [aux_sym_long_identifier_repeat1] = STATE(3861),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_as] = ACTIONS(6360),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3813)] = {
    [sym_block_comment] = STATE(3813),
    [sym_compiler_directive_decl] = STATE(3813),
    [sym_fsi_directive_decl] = STATE(3813),
    [sym_preproc_line] = STATE(3813),
    [ts_builtin_sym_end] = ACTIONS(6599),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_namespace] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8164),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(3814)] = {
    [sym_block_comment] = STATE(3814),
    [sym_compiler_directive_decl] = STATE(3814),
    [sym_fsi_directive_decl] = STATE(3814),
    [sym_preproc_line] = STATE(3814),
    [aux_sym_sequential_expression_repeat1] = STATE(3814),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [anon_sym_POUNDelse] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3815)] = {
    [sym__else_expression] = STATE(4928),
    [sym_elif_expression] = STATE(4544),
    [sym_block_comment] = STATE(3815),
    [sym_compiler_directive_decl] = STATE(3815),
    [sym_fsi_directive_decl] = STATE(3815),
    [sym_preproc_line] = STATE(3815),
    [aux_sym_if_expression_repeat1] = STATE(3820),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_as] = ACTIONS(6319),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8300),
    [anon_sym_elif] = ACTIONS(8302),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_DOT_DOT] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__dedent] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(3816)] = {
    [sym_block_comment] = STATE(3816),
    [sym_compiler_directive_decl] = STATE(3816),
    [sym_fsi_directive_decl] = STATE(3816),
    [sym_preproc_line] = STATE(3816),
    [ts_builtin_sym_end] = ACTIONS(6752),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_namespace] = ACTIONS(6750),
    [anon_sym_module] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_open] = ACTIONS(6750),
    [anon_sym_type] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_exception] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6752),
    [anon_sym_POUNDload] = ACTIONS(6752),
    [anon_sym_POUNDtime] = ACTIONS(6752),
    [anon_sym_POUNDI] = ACTIONS(6752),
    [anon_sym_POUNDhelp] = ACTIONS(6752),
    [anon_sym_POUNDquit] = ACTIONS(6752),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3817)] = {
    [sym_block_comment] = STATE(3817),
    [sym_compiler_directive_decl] = STATE(3817),
    [sym_fsi_directive_decl] = STATE(3817),
    [sym_preproc_line] = STATE(3817),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_module] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_open] = ACTIONS(6784),
    [anon_sym_type] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8304),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_exception] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6786),
    [anon_sym_POUNDload] = ACTIONS(6786),
    [anon_sym_POUNDtime] = ACTIONS(6786),
    [anon_sym_POUNDI] = ACTIONS(6786),
    [anon_sym_POUNDhelp] = ACTIONS(6786),
    [anon_sym_POUNDquit] = ACTIONS(6786),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [anon_sym_POUNDelse] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3818)] = {
    [sym_block_comment] = STATE(3818),
    [sym_compiler_directive_decl] = STATE(3818),
    [sym_fsi_directive_decl] = STATE(3818),
    [sym_preproc_line] = STATE(3818),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_GT_RBRACK] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_RBRACK] = ACTIONS(6515),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_RBRACE] = ACTIONS(6515),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_to] = ACTIONS(6513),
    [anon_sym_downto] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_DOT_DOT2] = ACTIONS(6515),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3819)] = {
    [sym_block_comment] = STATE(3819),
    [sym_compiler_directive_decl] = STATE(3819),
    [sym_fsi_directive_decl] = STATE(3819),
    [sym_preproc_line] = STATE(3819),
    [ts_builtin_sym_end] = ACTIONS(6830),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_namespace] = ACTIONS(6828),
    [anon_sym_module] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_open] = ACTIONS(6828),
    [anon_sym_type] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_exception] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6830),
    [anon_sym_POUNDload] = ACTIONS(6830),
    [anon_sym_POUNDtime] = ACTIONS(6830),
    [anon_sym_POUNDI] = ACTIONS(6830),
    [anon_sym_POUNDhelp] = ACTIONS(6830),
    [anon_sym_POUNDquit] = ACTIONS(6830),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3820)] = {
    [sym__else_expression] = STATE(4939),
    [sym_elif_expression] = STATE(4544),
    [sym_block_comment] = STATE(3820),
    [sym_compiler_directive_decl] = STATE(3820),
    [sym_fsi_directive_decl] = STATE(3820),
    [sym_preproc_line] = STATE(3820),
    [aux_sym_if_expression_repeat1] = STATE(3999),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_as] = ACTIONS(6329),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8300),
    [anon_sym_elif] = ACTIONS(8302),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_DOT_DOT] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__dedent] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(3821)] = {
    [sym_block_comment] = STATE(3821),
    [sym_compiler_directive_decl] = STATE(3821),
    [sym_fsi_directive_decl] = STATE(3821),
    [sym_preproc_line] = STATE(3821),
    [aux_sym_rules_repeat1] = STATE(3850),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8286),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3822)] = {
    [sym_type_arguments] = STATE(4464),
    [sym_long_identifier] = STATE(4465),
    [sym_block_comment] = STATE(3822),
    [sym_compiler_directive_decl] = STATE(3822),
    [sym_fsi_directive_decl] = STATE(3822),
    [sym_preproc_line] = STATE(3822),
    [aux_sym_compound_type_repeat1] = STATE(4261),
    [sym_identifier] = ACTIONS(8311),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_as] = ACTIONS(6261),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8216),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8218),
    [anon_sym_LT] = ACTIONS(8220),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8222),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3823)] = {
    [sym_type_arguments] = STATE(4482),
    [sym_long_identifier] = STATE(4483),
    [sym_block_comment] = STATE(3823),
    [sym_compiler_directive_decl] = STATE(3823),
    [sym_fsi_directive_decl] = STATE(3823),
    [sym_preproc_line] = STATE(3823),
    [aux_sym_compound_type_repeat1] = STATE(4270),
    [sym_identifier] = ACTIONS(8313),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8229),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8231),
    [anon_sym_LT] = ACTIONS(8233),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8235),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3824)] = {
    [sym_type_arguments] = STATE(4482),
    [sym_long_identifier] = STATE(4483),
    [sym_block_comment] = STATE(3824),
    [sym_compiler_directive_decl] = STATE(3824),
    [sym_fsi_directive_decl] = STATE(3824),
    [sym_preproc_line] = STATE(3824),
    [aux_sym_compound_type_repeat1] = STATE(4270),
    [sym_identifier] = ACTIONS(8313),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8229),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8231),
    [anon_sym_LT] = ACTIONS(8233),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8235),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3825)] = {
    [sym_block_comment] = STATE(3825),
    [sym_compiler_directive_decl] = STATE(3825),
    [sym_fsi_directive_decl] = STATE(3825),
    [sym_preproc_line] = STATE(3825),
    [aux_sym_type_argument_repeat1] = STATE(3826),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_as] = ACTIONS(6308),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_DOT_DOT] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__dedent] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(3826)] = {
    [sym_block_comment] = STATE(3826),
    [sym_compiler_directive_decl] = STATE(3826),
    [sym_fsi_directive_decl] = STATE(3826),
    [sym_preproc_line] = STATE(3826),
    [aux_sym_type_argument_repeat1] = STATE(3826),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_as] = ACTIONS(6312),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8315),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(3827)] = {
    [sym_type_arguments] = STATE(4464),
    [sym_long_identifier] = STATE(4465),
    [sym_block_comment] = STATE(3827),
    [sym_compiler_directive_decl] = STATE(3827),
    [sym_fsi_directive_decl] = STATE(3827),
    [sym_preproc_line] = STATE(3827),
    [aux_sym_compound_type_repeat1] = STATE(4261),
    [sym_identifier] = ACTIONS(8311),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_as] = ACTIONS(6257),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8216),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8218),
    [anon_sym_LT] = ACTIONS(8220),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8222),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3828)] = {
    [sym_block_comment] = STATE(3828),
    [sym_compiler_directive_decl] = STATE(3828),
    [sym_fsi_directive_decl] = STATE(3828),
    [sym_preproc_line] = STATE(3828),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_module] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_open] = ACTIONS(6564),
    [anon_sym_type] = ACTIONS(6564),
    [anon_sym_LBRACK_LT] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_exception] = ACTIONS(6564),
    [anon_sym_extern] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6562),
    [anon_sym_POUNDload] = ACTIONS(6562),
    [anon_sym_POUNDtime] = ACTIONS(6562),
    [anon_sym_POUNDI] = ACTIONS(6562),
    [anon_sym_POUNDhelp] = ACTIONS(6562),
    [anon_sym_POUNDquit] = ACTIONS(6562),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(3829)] = {
    [sym_block_comment] = STATE(3829),
    [sym_compiler_directive_decl] = STATE(3829),
    [sym_fsi_directive_decl] = STATE(3829),
    [sym_preproc_line] = STATE(3829),
    [aux_sym_long_identifier_repeat1] = STATE(3812),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_as] = ACTIONS(6289),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_DOT_DOT] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__dedent] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(3830)] = {
    [sym_block_comment] = STATE(3830),
    [sym_compiler_directive_decl] = STATE(3830),
    [sym_fsi_directive_decl] = STATE(3830),
    [sym_preproc_line] = STATE(3830),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3831)] = {
    [sym_block_comment] = STATE(3831),
    [sym_compiler_directive_decl] = STATE(3831),
    [sym_fsi_directive_decl] = STATE(3831),
    [sym_preproc_line] = STATE(3831),
    [aux_sym_long_identifier_repeat1] = STATE(3831),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8318),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3832)] = {
    [sym_block_comment] = STATE(3832),
    [sym_compiler_directive_decl] = STATE(3832),
    [sym_fsi_directive_decl] = STATE(3832),
    [sym_preproc_line] = STATE(3832),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_module] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_open] = ACTIONS(6840),
    [anon_sym_type] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_exception] = ACTIONS(6840),
    [anon_sym_extern] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6838),
    [anon_sym_POUNDload] = ACTIONS(6838),
    [anon_sym_POUNDtime] = ACTIONS(6838),
    [anon_sym_POUNDI] = ACTIONS(6838),
    [anon_sym_POUNDhelp] = ACTIONS(6838),
    [anon_sym_POUNDquit] = ACTIONS(6838),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [anon_sym_POUNDelse] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(3833)] = {
    [sym_block_comment] = STATE(3833),
    [sym_compiler_directive_decl] = STATE(3833),
    [sym_fsi_directive_decl] = STATE(3833),
    [sym_preproc_line] = STATE(3833),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3834)] = {
    [sym_block_comment] = STATE(3834),
    [sym_compiler_directive_decl] = STATE(3834),
    [sym_fsi_directive_decl] = STATE(3834),
    [sym_preproc_line] = STATE(3834),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3835)] = {
    [sym_block_comment] = STATE(3835),
    [sym_compiler_directive_decl] = STATE(3835),
    [sym_fsi_directive_decl] = STATE(3835),
    [sym_preproc_line] = STATE(3835),
    [ts_builtin_sym_end] = ACTIONS(6614),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_namespace] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3836)] = {
    [sym_type_arguments] = STATE(4482),
    [sym_long_identifier] = STATE(4483),
    [sym_block_comment] = STATE(3836),
    [sym_compiler_directive_decl] = STATE(3836),
    [sym_fsi_directive_decl] = STATE(3836),
    [sym_preproc_line] = STATE(3836),
    [aux_sym_compound_type_repeat1] = STATE(4270),
    [sym_identifier] = ACTIONS(8313),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8229),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8231),
    [anon_sym_LT] = ACTIONS(8233),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8235),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3837)] = {
    [sym_type_arguments] = STATE(4310),
    [sym_long_identifier] = STATE(4312),
    [sym_block_comment] = STATE(3837),
    [sym_compiler_directive_decl] = STATE(3837),
    [sym_fsi_directive_decl] = STATE(3837),
    [sym_preproc_line] = STATE(3837),
    [aux_sym_compound_type_repeat1] = STATE(4191),
    [sym_identifier] = ACTIONS(8264),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8266),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_DOT_DOT] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8268),
    [anon_sym_LT] = ACTIONS(8270),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8272),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__dedent] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3838)] = {
    [sym_type_arguments] = STATE(4482),
    [sym_long_identifier] = STATE(4483),
    [sym_block_comment] = STATE(3838),
    [sym_compiler_directive_decl] = STATE(3838),
    [sym_fsi_directive_decl] = STATE(3838),
    [sym_preproc_line] = STATE(3838),
    [aux_sym_compound_type_repeat1] = STATE(4270),
    [sym_identifier] = ACTIONS(8313),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8229),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8231),
    [anon_sym_LT] = ACTIONS(8233),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8235),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3839)] = {
    [sym_type_arguments] = STATE(4310),
    [sym_long_identifier] = STATE(4312),
    [sym_block_comment] = STATE(3839),
    [sym_compiler_directive_decl] = STATE(3839),
    [sym_fsi_directive_decl] = STATE(3839),
    [sym_preproc_line] = STATE(3839),
    [aux_sym_compound_type_repeat1] = STATE(4191),
    [sym_identifier] = ACTIONS(8264),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8266),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8268),
    [anon_sym_LT] = ACTIONS(8270),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8272),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3840)] = {
    [sym_block_comment] = STATE(3840),
    [sym_compiler_directive_decl] = STATE(3840),
    [sym_fsi_directive_decl] = STATE(3840),
    [sym_preproc_line] = STATE(3840),
    [aux_sym_long_identifier_repeat1] = STATE(3791),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_module] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_open] = ACTIONS(6625),
    [anon_sym_type] = ACTIONS(6625),
    [anon_sym_LBRACK_LT] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8321),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_exception] = ACTIONS(6625),
    [anon_sym_extern] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6628),
    [anon_sym_POUNDload] = ACTIONS(6628),
    [anon_sym_POUNDtime] = ACTIONS(6628),
    [anon_sym_POUNDI] = ACTIONS(6628),
    [anon_sym_POUNDhelp] = ACTIONS(6628),
    [anon_sym_POUNDquit] = ACTIONS(6628),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3841)] = {
    [sym_type_arguments] = STATE(4310),
    [sym_long_identifier] = STATE(4312),
    [sym_block_comment] = STATE(3841),
    [sym_compiler_directive_decl] = STATE(3841),
    [sym_fsi_directive_decl] = STATE(3841),
    [sym_preproc_line] = STATE(3841),
    [aux_sym_compound_type_repeat1] = STATE(4191),
    [sym_identifier] = ACTIONS(8264),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8266),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_DOT_DOT] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8268),
    [anon_sym_LT] = ACTIONS(8270),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8272),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__dedent] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3842)] = {
    [sym_block_comment] = STATE(3842),
    [sym_compiler_directive_decl] = STATE(3842),
    [sym_fsi_directive_decl] = STATE(3842),
    [sym_preproc_line] = STATE(3842),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_module] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_open] = ACTIONS(6813),
    [anon_sym_type] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_exception] = ACTIONS(6813),
    [anon_sym_extern] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6811),
    [anon_sym_POUNDload] = ACTIONS(6811),
    [anon_sym_POUNDtime] = ACTIONS(6811),
    [anon_sym_POUNDI] = ACTIONS(6811),
    [anon_sym_POUNDhelp] = ACTIONS(6811),
    [anon_sym_POUNDquit] = ACTIONS(6811),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [anon_sym_POUNDelse] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3843)] = {
    [sym_block_comment] = STATE(3843),
    [sym_compiler_directive_decl] = STATE(3843),
    [sym_fsi_directive_decl] = STATE(3843),
    [sym_preproc_line] = STATE(3843),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8330),
    [anon_sym_POUNDelse] = ACTIONS(8330),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3844)] = {
    [sym_block_comment] = STATE(3844),
    [sym_compiler_directive_decl] = STATE(3844),
    [sym_fsi_directive_decl] = STATE(3844),
    [sym_preproc_line] = STATE(3844),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3845)] = {
    [sym_block_comment] = STATE(3845),
    [sym_compiler_directive_decl] = STATE(3845),
    [sym_fsi_directive_decl] = STATE(3845),
    [sym_preproc_line] = STATE(3845),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_module] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_open] = ACTIONS(6863),
    [anon_sym_type] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_exception] = ACTIONS(6863),
    [anon_sym_extern] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6861),
    [anon_sym_POUNDload] = ACTIONS(6861),
    [anon_sym_POUNDtime] = ACTIONS(6861),
    [anon_sym_POUNDI] = ACTIONS(6861),
    [anon_sym_POUNDhelp] = ACTIONS(6861),
    [anon_sym_POUNDquit] = ACTIONS(6861),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [anon_sym_POUNDelse] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3846)] = {
    [sym_block_comment] = STATE(3846),
    [sym_compiler_directive_decl] = STATE(3846),
    [sym_fsi_directive_decl] = STATE(3846),
    [sym_preproc_line] = STATE(3846),
    [aux_sym_sequential_expression_repeat1] = STATE(3846),
    [ts_builtin_sym_end] = ACTIONS(6798),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_namespace] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8335),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3847)] = {
    [sym_type_arguments] = STATE(4532),
    [sym_long_identifier] = STATE(4523),
    [sym_block_comment] = STATE(3847),
    [sym_compiler_directive_decl] = STATE(3847),
    [sym_fsi_directive_decl] = STATE(3847),
    [sym_preproc_line] = STATE(3847),
    [aux_sym_compound_type_repeat1] = STATE(4288),
    [sym_identifier] = ACTIONS(8274),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8276),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8278),
    [anon_sym_LT] = ACTIONS(8280),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8282),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__dedent] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
    [sym__in] = ACTIONS(6255),
  },
  [STATE(3848)] = {
    [sym_block_comment] = STATE(3848),
    [sym_compiler_directive_decl] = STATE(3848),
    [sym_fsi_directive_decl] = STATE(3848),
    [sym_preproc_line] = STATE(3848),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(3849)] = {
    [sym_block_comment] = STATE(3849),
    [sym_compiler_directive_decl] = STATE(3849),
    [sym_fsi_directive_decl] = STATE(3849),
    [sym_preproc_line] = STATE(3849),
    [ts_builtin_sym_end] = ACTIONS(6861),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_namespace] = ACTIONS(6863),
    [anon_sym_module] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_open] = ACTIONS(6863),
    [anon_sym_type] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_exception] = ACTIONS(6863),
    [anon_sym_extern] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6861),
    [anon_sym_POUNDload] = ACTIONS(6861),
    [anon_sym_POUNDtime] = ACTIONS(6861),
    [anon_sym_POUNDI] = ACTIONS(6861),
    [anon_sym_POUNDhelp] = ACTIONS(6861),
    [anon_sym_POUNDquit] = ACTIONS(6861),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(3850)] = {
    [sym_block_comment] = STATE(3850),
    [sym_compiler_directive_decl] = STATE(3850),
    [sym_fsi_directive_decl] = STATE(3850),
    [sym_preproc_line] = STATE(3850),
    [aux_sym_rules_repeat1] = STATE(3855),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_module] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_open] = ACTIONS(6618),
    [anon_sym_type] = ACTIONS(6618),
    [anon_sym_LBRACK_LT] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(8286),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_exception] = ACTIONS(6618),
    [anon_sym_extern] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6616),
    [anon_sym_POUNDload] = ACTIONS(6616),
    [anon_sym_POUNDtime] = ACTIONS(6616),
    [anon_sym_POUNDI] = ACTIONS(6616),
    [anon_sym_POUNDhelp] = ACTIONS(6616),
    [anon_sym_POUNDquit] = ACTIONS(6616),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(8338),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(3851)] = {
    [sym_block_comment] = STATE(3851),
    [sym_compiler_directive_decl] = STATE(3851),
    [sym_fsi_directive_decl] = STATE(3851),
    [sym_preproc_line] = STATE(3851),
    [ts_builtin_sym_end] = ACTIONS(6805),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_namespace] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3852)] = {
    [sym_block_comment] = STATE(3852),
    [sym_compiler_directive_decl] = STATE(3852),
    [sym_fsi_directive_decl] = STATE(3852),
    [sym_preproc_line] = STATE(3852),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_module] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_open] = ACTIONS(6828),
    [anon_sym_type] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_exception] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6830),
    [anon_sym_POUNDload] = ACTIONS(6830),
    [anon_sym_POUNDtime] = ACTIONS(6830),
    [anon_sym_POUNDI] = ACTIONS(6830),
    [anon_sym_POUNDhelp] = ACTIONS(6830),
    [anon_sym_POUNDquit] = ACTIONS(6830),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [anon_sym_POUNDelse] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3853)] = {
    [sym_attributes] = STATE(13274),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6891),
    [sym_access_modifier] = STATE(22393),
    [sym__class_function_or_value_defn] = STATE(6892),
    [aux_sym__type_extension_inner] = STATE(4219),
    [sym_type_extension_elements] = STATE(8166),
    [sym__type_defn_elements] = STATE(6892),
    [sym_interface_implementation] = STATE(6893),
    [sym_member_defn] = STATE(6893),
    [sym_additional_constr_defn] = STATE(6918),
    [sym_block_comment] = STATE(3853),
    [sym_compiler_directive_decl] = STATE(3853),
    [sym_fsi_directive_decl] = STATE(3853),
    [sym_preproc_line] = STATE(3853),
    [sym_preproc_if_in_class_definition] = STATE(6893),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8341),
    [anon_sym_module] = ACTIONS(8341),
    [anon_sym_null] = ACTIONS(8341),
    [anon_sym_begin] = ACTIONS(8341),
    [anon_sym_open] = ACTIONS(8341),
    [anon_sym_type] = ACTIONS(8341),
    [anon_sym_LBRACK_LT] = ACTIONS(8343),
    [anon_sym_do] = ACTIONS(8341),
    [anon_sym_and] = ACTIONS(8341),
    [anon_sym_let] = ACTIONS(8341),
    [anon_sym_let_BANG] = ACTIONS(8343),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8341),
    [anon_sym_AMP] = ACTIONS(8341),
    [anon_sym_LBRACK] = ACTIONS(8341),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8343),
    [anon_sym_LBRACE] = ACTIONS(8341),
    [anon_sym_LT_AT] = ACTIONS(8341),
    [anon_sym_LT_AT_AT] = ACTIONS(8343),
    [anon_sym_CARET] = ACTIONS(8343),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8343),
    [anon_sym_new] = ACTIONS(8341),
    [anon_sym_with] = ACTIONS(8248),
    [anon_sym_return] = ACTIONS(8341),
    [anon_sym_return_BANG] = ACTIONS(8343),
    [anon_sym_yield] = ACTIONS(8341),
    [anon_sym_yield_BANG] = ACTIONS(8343),
    [anon_sym_lazy] = ACTIONS(8341),
    [anon_sym_assert] = ACTIONS(8341),
    [anon_sym_upcast] = ACTIONS(8341),
    [anon_sym_downcast] = ACTIONS(8341),
    [anon_sym_fixed] = ACTIONS(8341),
    [anon_sym_for] = ACTIONS(8341),
    [anon_sym_while] = ACTIONS(8341),
    [anon_sym_while_BANG] = ACTIONS(8343),
    [anon_sym_if] = ACTIONS(8341),
    [anon_sym_fun] = ACTIONS(8341),
    [anon_sym_try] = ACTIONS(8341),
    [anon_sym_match] = ACTIONS(8341),
    [anon_sym_match_BANG] = ACTIONS(8343),
    [anon_sym_function] = ACTIONS(8341),
    [anon_sym_use] = ACTIONS(8341),
    [anon_sym_use_BANG] = ACTIONS(8343),
    [anon_sym_do_BANG] = ACTIONS(8343),
    [anon_sym_AMP_AMP] = ACTIONS(8343),
    [anon_sym_default] = ACTIONS(8250),
    [anon_sym_static] = ACTIONS(8252),
    [anon_sym_member] = ACTIONS(8254),
    [anon_sym_exception] = ACTIONS(8341),
    [anon_sym_interface] = ACTIONS(8256),
    [anon_sym_abstract] = ACTIONS(8258),
    [anon_sym_val] = ACTIONS(8260),
    [anon_sym_override] = ACTIONS(8250),
    [anon_sym_extern] = ACTIONS(8341),
    [aux_sym_char_token1] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8341),
    [anon_sym_DQUOTE] = ACTIONS(8341),
    [anon_sym_AT_DQUOTE] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [sym_bool] = ACTIONS(8341),
    [sym_unit] = ACTIONS(8343),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8341),
    [sym_op_identifier] = ACTIONS(8343),
    [anon_sym_PLUS] = ACTIONS(8341),
    [anon_sym_DASH] = ACTIONS(8341),
    [anon_sym_PLUS_DOT] = ACTIONS(8343),
    [anon_sym_DASH_DOT] = ACTIONS(8343),
    [anon_sym_PERCENT] = ACTIONS(8341),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8343),
    [anon_sym_TILDE] = ACTIONS(8343),
    [aux_sym_prefix_op_token1] = ACTIONS(8343),
    [sym_int] = ACTIONS(8341),
    [sym_xint] = ACTIONS(8343),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8343),
    [anon_sym_POUNDload] = ACTIONS(8343),
    [anon_sym_POUNDtime] = ACTIONS(8343),
    [anon_sym_POUNDI] = ACTIONS(8343),
    [anon_sym_POUNDhelp] = ACTIONS(8343),
    [anon_sym_POUNDquit] = ACTIONS(8343),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8343),
    [anon_sym_POUNDendif] = ACTIONS(8343),
    [anon_sym_POUNDelse] = ACTIONS(8343),
    [sym__newline] = ACTIONS(8343),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8343),
  },
  [STATE(3854)] = {
    [sym_block_comment] = STATE(3854),
    [sym_compiler_directive_decl] = STATE(3854),
    [sym_fsi_directive_decl] = STATE(3854),
    [sym_preproc_line] = STATE(3854),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_module] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_open] = ACTIONS(6824),
    [anon_sym_type] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_exception] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6822),
    [anon_sym_POUNDload] = ACTIONS(6822),
    [anon_sym_POUNDtime] = ACTIONS(6822),
    [anon_sym_POUNDI] = ACTIONS(6822),
    [anon_sym_POUNDhelp] = ACTIONS(6822),
    [anon_sym_POUNDquit] = ACTIONS(6822),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [anon_sym_POUNDelse] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(3855)] = {
    [sym_block_comment] = STATE(3855),
    [sym_compiler_directive_decl] = STATE(3855),
    [sym_fsi_directive_decl] = STATE(3855),
    [sym_preproc_line] = STATE(3855),
    [aux_sym_rules_repeat1] = STATE(3855),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(8345),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(8348),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3856)] = {
    [sym_type_arguments] = STATE(4310),
    [sym_long_identifier] = STATE(4312),
    [sym_block_comment] = STATE(3856),
    [sym_compiler_directive_decl] = STATE(3856),
    [sym_fsi_directive_decl] = STATE(3856),
    [sym_preproc_line] = STATE(3856),
    [aux_sym_compound_type_repeat1] = STATE(4191),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8266),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_DOT_DOT] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8268),
    [anon_sym_LT] = ACTIONS(8270),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8272),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__dedent] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3857)] = {
    [sym_type_arguments] = STATE(4532),
    [sym_long_identifier] = STATE(4523),
    [sym_block_comment] = STATE(3857),
    [sym_compiler_directive_decl] = STATE(3857),
    [sym_fsi_directive_decl] = STATE(3857),
    [sym_preproc_line] = STATE(3857),
    [aux_sym_compound_type_repeat1] = STATE(4288),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8276),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8278),
    [anon_sym_LT] = ACTIONS(8280),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8282),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__dedent] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
    [sym__in] = ACTIONS(6263),
  },
  [STATE(3858)] = {
    [sym_block_comment] = STATE(3858),
    [sym_compiler_directive_decl] = STATE(3858),
    [sym_fsi_directive_decl] = STATE(3858),
    [sym_preproc_line] = STATE(3858),
    [ts_builtin_sym_end] = ACTIONS(6832),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_namespace] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(3859)] = {
    [sym_block_comment] = STATE(3859),
    [sym_compiler_directive_decl] = STATE(3859),
    [sym_fsi_directive_decl] = STATE(3859),
    [sym_preproc_line] = STATE(3859),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_module] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_open] = ACTIONS(6554),
    [anon_sym_type] = ACTIONS(6554),
    [anon_sym_LBRACK_LT] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_exception] = ACTIONS(6554),
    [anon_sym_extern] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6556),
    [anon_sym_POUNDload] = ACTIONS(6556),
    [anon_sym_POUNDtime] = ACTIONS(6556),
    [anon_sym_POUNDI] = ACTIONS(6556),
    [anon_sym_POUNDhelp] = ACTIONS(6556),
    [anon_sym_POUNDquit] = ACTIONS(6556),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(3860)] = {
    [sym_block_comment] = STATE(3860),
    [sym_compiler_directive_decl] = STATE(3860),
    [sym_fsi_directive_decl] = STATE(3860),
    [sym_preproc_line] = STATE(3860),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6734),
    [aux_sym_decimal_token1] = ACTIONS(6541),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3861)] = {
    [sym_block_comment] = STATE(3861),
    [sym_compiler_directive_decl] = STATE(3861),
    [sym_fsi_directive_decl] = STATE(3861),
    [sym_preproc_line] = STATE(3861),
    [aux_sym_long_identifier_repeat1] = STATE(3861),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8351),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3862)] = {
    [sym_block_comment] = STATE(3862),
    [sym_compiler_directive_decl] = STATE(3862),
    [sym_fsi_directive_decl] = STATE(3862),
    [sym_preproc_line] = STATE(3862),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8332),
    [anon_sym_POUNDelse] = ACTIONS(8332),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3863)] = {
    [sym_block_comment] = STATE(3863),
    [sym_compiler_directive_decl] = STATE(3863),
    [sym_fsi_directive_decl] = STATE(3863),
    [sym_preproc_line] = STATE(3863),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_module] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_open] = ACTIONS(6637),
    [anon_sym_type] = ACTIONS(6637),
    [anon_sym_LBRACK_LT] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_exception] = ACTIONS(6637),
    [anon_sym_extern] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(8354),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6635),
    [anon_sym_POUNDload] = ACTIONS(6635),
    [anon_sym_POUNDtime] = ACTIONS(6635),
    [anon_sym_POUNDI] = ACTIONS(6635),
    [anon_sym_POUNDhelp] = ACTIONS(6635),
    [anon_sym_POUNDquit] = ACTIONS(6635),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3864)] = {
    [sym_block_comment] = STATE(3864),
    [sym_compiler_directive_decl] = STATE(3864),
    [sym_fsi_directive_decl] = STATE(3864),
    [sym_preproc_line] = STATE(3864),
    [aux_sym_sequential_expression_repeat1] = STATE(3814),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [anon_sym_POUNDelse] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3865)] = {
    [sym_type_arguments] = STATE(4464),
    [sym_long_identifier] = STATE(4465),
    [sym_block_comment] = STATE(3865),
    [sym_compiler_directive_decl] = STATE(3865),
    [sym_fsi_directive_decl] = STATE(3865),
    [sym_preproc_line] = STATE(3865),
    [aux_sym_compound_type_repeat1] = STATE(4261),
    [sym_identifier] = ACTIONS(8311),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_as] = ACTIONS(6213),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8216),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8218),
    [anon_sym_LT] = ACTIONS(8220),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8222),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3866)] = {
    [sym_type_arguments] = STATE(4464),
    [sym_long_identifier] = STATE(4465),
    [sym_block_comment] = STATE(3866),
    [sym_compiler_directive_decl] = STATE(3866),
    [sym_fsi_directive_decl] = STATE(3866),
    [sym_preproc_line] = STATE(3866),
    [aux_sym_compound_type_repeat1] = STATE(4261),
    [sym_identifier] = ACTIONS(8311),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_as] = ACTIONS(6193),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8216),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8218),
    [anon_sym_LT] = ACTIONS(8220),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8222),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3867)] = {
    [sym_block_comment] = STATE(3867),
    [sym_compiler_directive_decl] = STATE(3867),
    [sym_fsi_directive_decl] = STATE(3867),
    [sym_preproc_line] = STATE(3867),
    [aux_sym_rules_repeat1] = STATE(3855),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_module] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_open] = ACTIONS(6607),
    [anon_sym_type] = ACTIONS(6607),
    [anon_sym_LBRACK_LT] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8286),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_exception] = ACTIONS(6607),
    [anon_sym_extern] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6605),
    [anon_sym_POUNDload] = ACTIONS(6605),
    [anon_sym_POUNDtime] = ACTIONS(6605),
    [anon_sym_POUNDI] = ACTIONS(6605),
    [anon_sym_POUNDhelp] = ACTIONS(6605),
    [anon_sym_POUNDquit] = ACTIONS(6605),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8308),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3868)] = {
    [sym_type_arguments] = STATE(4553),
    [sym_long_identifier] = STATE(4568),
    [sym_block_comment] = STATE(3868),
    [sym_compiler_directive_decl] = STATE(3868),
    [sym_fsi_directive_decl] = STATE(3868),
    [sym_preproc_line] = STATE(3868),
    [aux_sym_compound_type_repeat1] = STATE(4443),
    [sym_identifier] = ACTIONS(8356),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8358),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8360),
    [anon_sym_LT] = ACTIONS(8362),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8364),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__quoted_close] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3869)] = {
    [sym_block_comment] = STATE(3869),
    [sym_compiler_directive_decl] = STATE(3869),
    [sym_fsi_directive_decl] = STATE(3869),
    [sym_preproc_line] = STATE(3869),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_module] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_open] = ACTIONS(7031),
    [anon_sym_type] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_exception] = ACTIONS(7031),
    [anon_sym_extern] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7029),
    [anon_sym_POUNDload] = ACTIONS(7029),
    [anon_sym_POUNDtime] = ACTIONS(7029),
    [anon_sym_POUNDI] = ACTIONS(7029),
    [anon_sym_POUNDhelp] = ACTIONS(7029),
    [anon_sym_POUNDquit] = ACTIONS(7029),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [anon_sym_POUNDelse] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3870)] = {
    [sym_block_comment] = STATE(3870),
    [sym_compiler_directive_decl] = STATE(3870),
    [sym_fsi_directive_decl] = STATE(3870),
    [sym_preproc_line] = STATE(3870),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_GT_RBRACK] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_RBRACK] = ACTIONS(5953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_RBRACE] = ACTIONS(5953),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_to] = ACTIONS(5951),
    [anon_sym_downto] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7198),
    [aux_sym_decimal_token1] = ACTIONS(6718),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3871)] = {
    [sym_block_comment] = STATE(3871),
    [sym_compiler_directive_decl] = STATE(3871),
    [sym_fsi_directive_decl] = STATE(3871),
    [sym_preproc_line] = STATE(3871),
    [aux_sym_rules_repeat1] = STATE(3890),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_GT_RBRACK] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8366),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_RBRACK] = ACTIONS(6605),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_RBRACE] = ACTIONS(6605),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_to] = ACTIONS(6607),
    [anon_sym_downto] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_DOT_DOT2] = ACTIONS(6605),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8368),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3872)] = {
    [sym_block_comment] = STATE(3872),
    [sym_compiler_directive_decl] = STATE(3872),
    [sym_fsi_directive_decl] = STATE(3872),
    [sym_preproc_line] = STATE(3872),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_module] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_open] = ACTIONS(6963),
    [anon_sym_type] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_exception] = ACTIONS(6963),
    [anon_sym_extern] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6961),
    [anon_sym_POUNDload] = ACTIONS(6961),
    [anon_sym_POUNDtime] = ACTIONS(6961),
    [anon_sym_POUNDI] = ACTIONS(6961),
    [anon_sym_POUNDhelp] = ACTIONS(6961),
    [anon_sym_POUNDquit] = ACTIONS(6961),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [anon_sym_POUNDelse] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3873)] = {
    [sym_block_comment] = STATE(3873),
    [sym_compiler_directive_decl] = STATE(3873),
    [sym_fsi_directive_decl] = STATE(3873),
    [sym_preproc_line] = STATE(3873),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_module] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_open] = ACTIONS(7019),
    [anon_sym_type] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_exception] = ACTIONS(7019),
    [anon_sym_extern] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7017),
    [anon_sym_POUNDload] = ACTIONS(7017),
    [anon_sym_POUNDtime] = ACTIONS(7017),
    [anon_sym_POUNDI] = ACTIONS(7017),
    [anon_sym_POUNDhelp] = ACTIONS(7017),
    [anon_sym_POUNDquit] = ACTIONS(7017),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [anon_sym_POUNDelse] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3874)] = {
    [sym_block_comment] = STATE(3874),
    [sym_compiler_directive_decl] = STATE(3874),
    [sym_fsi_directive_decl] = STATE(3874),
    [sym_preproc_line] = STATE(3874),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_module] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_open] = ACTIONS(7047),
    [anon_sym_type] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_exception] = ACTIONS(7047),
    [anon_sym_extern] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7045),
    [anon_sym_POUNDload] = ACTIONS(7045),
    [anon_sym_POUNDtime] = ACTIONS(7045),
    [anon_sym_POUNDI] = ACTIONS(7045),
    [anon_sym_POUNDhelp] = ACTIONS(7045),
    [anon_sym_POUNDquit] = ACTIONS(7045),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [anon_sym_POUNDelse] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3875)] = {
    [sym_block_comment] = STATE(3875),
    [sym_compiler_directive_decl] = STATE(3875),
    [sym_fsi_directive_decl] = STATE(3875),
    [sym_preproc_line] = STATE(3875),
    [ts_builtin_sym_end] = ACTIONS(6919),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_namespace] = ACTIONS(6921),
    [anon_sym_module] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_open] = ACTIONS(6921),
    [anon_sym_type] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_exception] = ACTIONS(6921),
    [anon_sym_extern] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6919),
    [anon_sym_POUNDload] = ACTIONS(6919),
    [anon_sym_POUNDtime] = ACTIONS(6919),
    [anon_sym_POUNDI] = ACTIONS(6919),
    [anon_sym_POUNDhelp] = ACTIONS(6919),
    [anon_sym_POUNDquit] = ACTIONS(6919),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(3876)] = {
    [sym_block_comment] = STATE(3876),
    [sym_compiler_directive_decl] = STATE(3876),
    [sym_fsi_directive_decl] = STATE(3876),
    [sym_preproc_line] = STATE(3876),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_module] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_open] = ACTIONS(6871),
    [anon_sym_type] = ACTIONS(6871),
    [anon_sym_LBRACK_LT] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_exception] = ACTIONS(6871),
    [anon_sym_extern] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6869),
    [anon_sym_POUNDload] = ACTIONS(6869),
    [anon_sym_POUNDtime] = ACTIONS(6869),
    [anon_sym_POUNDI] = ACTIONS(6869),
    [anon_sym_POUNDhelp] = ACTIONS(6869),
    [anon_sym_POUNDquit] = ACTIONS(6869),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(3877)] = {
    [sym_type_arguments] = STATE(4585),
    [sym_long_identifier] = STATE(4742),
    [sym_block_comment] = STATE(3877),
    [sym_compiler_directive_decl] = STATE(3877),
    [sym_fsi_directive_decl] = STATE(3877),
    [sym_preproc_line] = STATE(3877),
    [aux_sym_compound_type_repeat1] = STATE(4343),
    [sym_identifier] = ACTIONS(8371),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8373),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8375),
    [anon_sym_LT] = ACTIONS(8377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8379),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3878)] = {
    [sym_type_arguments] = STATE(4645),
    [sym_long_identifier] = STATE(4649),
    [sym_block_comment] = STATE(3878),
    [sym_compiler_directive_decl] = STATE(3878),
    [sym_fsi_directive_decl] = STATE(3878),
    [sym_preproc_line] = STATE(3878),
    [aux_sym_compound_type_repeat1] = STATE(4502),
    [sym_identifier] = ACTIONS(8381),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8383),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8385),
    [anon_sym_LT] = ACTIONS(8387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8389),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__multi_dollar_interp_end] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(3879)] = {
    [sym_block_comment] = STATE(3879),
    [sym_compiler_directive_decl] = STATE(3879),
    [sym_fsi_directive_decl] = STATE(3879),
    [sym_preproc_line] = STATE(3879),
    [ts_builtin_sym_end] = ACTIONS(7103),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_namespace] = ACTIONS(7101),
    [anon_sym_module] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_open] = ACTIONS(7101),
    [anon_sym_type] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_exception] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7103),
    [anon_sym_POUNDload] = ACTIONS(7103),
    [anon_sym_POUNDtime] = ACTIONS(7103),
    [anon_sym_POUNDI] = ACTIONS(7103),
    [anon_sym_POUNDhelp] = ACTIONS(7103),
    [anon_sym_POUNDquit] = ACTIONS(7103),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3880)] = {
    [sym_block_comment] = STATE(3880),
    [sym_compiler_directive_decl] = STATE(3880),
    [sym_fsi_directive_decl] = STATE(3880),
    [sym_preproc_line] = STATE(3880),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_module] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_open] = ACTIONS(6925),
    [anon_sym_type] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_exception] = ACTIONS(6925),
    [anon_sym_extern] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6923),
    [anon_sym_POUNDload] = ACTIONS(6923),
    [anon_sym_POUNDtime] = ACTIONS(6923),
    [anon_sym_POUNDI] = ACTIONS(6923),
    [anon_sym_POUNDhelp] = ACTIONS(6923),
    [anon_sym_POUNDquit] = ACTIONS(6923),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [anon_sym_POUNDelse] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3881)] = {
    [sym_block_comment] = STATE(3881),
    [sym_compiler_directive_decl] = STATE(3881),
    [sym_fsi_directive_decl] = STATE(3881),
    [sym_preproc_line] = STATE(3881),
    [ts_builtin_sym_end] = ACTIONS(6961),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_namespace] = ACTIONS(6963),
    [anon_sym_module] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_open] = ACTIONS(6963),
    [anon_sym_type] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_exception] = ACTIONS(6963),
    [anon_sym_extern] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6961),
    [anon_sym_POUNDload] = ACTIONS(6961),
    [anon_sym_POUNDtime] = ACTIONS(6961),
    [anon_sym_POUNDI] = ACTIONS(6961),
    [anon_sym_POUNDhelp] = ACTIONS(6961),
    [anon_sym_POUNDquit] = ACTIONS(6961),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(3882)] = {
    [sym_type_arguments] = STATE(4585),
    [sym_long_identifier] = STATE(4742),
    [sym_block_comment] = STATE(3882),
    [sym_compiler_directive_decl] = STATE(3882),
    [sym_fsi_directive_decl] = STATE(3882),
    [sym_preproc_line] = STATE(3882),
    [aux_sym_compound_type_repeat1] = STATE(4343),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8373),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8375),
    [anon_sym_LT] = ACTIONS(8377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8379),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [anon_sym_POUNDendif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3883)] = {
    [sym_block_comment] = STATE(3883),
    [sym_compiler_directive_decl] = STATE(3883),
    [sym_fsi_directive_decl] = STATE(3883),
    [sym_preproc_line] = STATE(3883),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_module] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_open] = ACTIONS(7023),
    [anon_sym_type] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_exception] = ACTIONS(7023),
    [anon_sym_extern] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7021),
    [anon_sym_POUNDload] = ACTIONS(7021),
    [anon_sym_POUNDtime] = ACTIONS(7021),
    [anon_sym_POUNDI] = ACTIONS(7021),
    [anon_sym_POUNDhelp] = ACTIONS(7021),
    [anon_sym_POUNDquit] = ACTIONS(7021),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [anon_sym_POUNDelse] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3884)] = {
    [sym_block_comment] = STATE(3884),
    [sym_compiler_directive_decl] = STATE(3884),
    [sym_fsi_directive_decl] = STATE(3884),
    [sym_preproc_line] = STATE(3884),
    [ts_builtin_sym_end] = ACTIONS(6886),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_namespace] = ACTIONS(6888),
    [anon_sym_module] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_open] = ACTIONS(6888),
    [anon_sym_type] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_exception] = ACTIONS(6888),
    [anon_sym_extern] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6886),
    [anon_sym_POUNDload] = ACTIONS(6886),
    [anon_sym_POUNDtime] = ACTIONS(6886),
    [anon_sym_POUNDI] = ACTIONS(6886),
    [anon_sym_POUNDhelp] = ACTIONS(6886),
    [anon_sym_POUNDquit] = ACTIONS(6886),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(3885)] = {
    [sym_type_arguments] = STATE(4585),
    [sym_long_identifier] = STATE(4742),
    [sym_block_comment] = STATE(3885),
    [sym_compiler_directive_decl] = STATE(3885),
    [sym_fsi_directive_decl] = STATE(3885),
    [sym_preproc_line] = STATE(3885),
    [aux_sym_compound_type_repeat1] = STATE(4343),
    [sym_identifier] = ACTIONS(8371),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8373),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8375),
    [anon_sym_LT] = ACTIONS(8377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8379),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3886)] = {
    [sym_type_arguments] = STATE(4585),
    [sym_long_identifier] = STATE(4742),
    [sym_block_comment] = STATE(3886),
    [sym_compiler_directive_decl] = STATE(3886),
    [sym_fsi_directive_decl] = STATE(3886),
    [sym_preproc_line] = STATE(3886),
    [aux_sym_compound_type_repeat1] = STATE(4343),
    [sym_identifier] = ACTIONS(8371),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8373),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8375),
    [anon_sym_LT] = ACTIONS(8377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8379),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3887)] = {
    [sym_block_comment] = STATE(3887),
    [sym_compiler_directive_decl] = STATE(3887),
    [sym_fsi_directive_decl] = STATE(3887),
    [sym_preproc_line] = STATE(3887),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_module] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_open] = ACTIONS(7051),
    [anon_sym_type] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_exception] = ACTIONS(7051),
    [anon_sym_extern] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7049),
    [anon_sym_POUNDload] = ACTIONS(7049),
    [anon_sym_POUNDtime] = ACTIONS(7049),
    [anon_sym_POUNDI] = ACTIONS(7049),
    [anon_sym_POUNDhelp] = ACTIONS(7049),
    [anon_sym_POUNDquit] = ACTIONS(7049),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [anon_sym_POUNDelse] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(3888)] = {
    [sym_block_comment] = STATE(3888),
    [sym_compiler_directive_decl] = STATE(3888),
    [sym_fsi_directive_decl] = STATE(3888),
    [sym_preproc_line] = STATE(3888),
    [ts_builtin_sym_end] = ACTIONS(7119),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_namespace] = ACTIONS(7117),
    [anon_sym_module] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_open] = ACTIONS(7117),
    [anon_sym_type] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_exception] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7119),
    [anon_sym_POUNDload] = ACTIONS(7119),
    [anon_sym_POUNDtime] = ACTIONS(7119),
    [anon_sym_POUNDI] = ACTIONS(7119),
    [anon_sym_POUNDhelp] = ACTIONS(7119),
    [anon_sym_POUNDquit] = ACTIONS(7119),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(3889)] = {
    [sym_block_comment] = STATE(3889),
    [sym_compiler_directive_decl] = STATE(3889),
    [sym_fsi_directive_decl] = STATE(3889),
    [sym_preproc_line] = STATE(3889),
    [aux_sym_sequential_expression_repeat1] = STATE(4117),
    [ts_builtin_sym_end] = ACTIONS(6738),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3890)] = {
    [sym_block_comment] = STATE(3890),
    [sym_compiler_directive_decl] = STATE(3890),
    [sym_fsi_directive_decl] = STATE(3890),
    [sym_preproc_line] = STATE(3890),
    [aux_sym_rules_repeat1] = STATE(3892),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_GT_RBRACK] = ACTIONS(6616),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(8366),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_RBRACK] = ACTIONS(6616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_RBRACE] = ACTIONS(6616),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_to] = ACTIONS(6618),
    [anon_sym_downto] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_DOT_DOT2] = ACTIONS(6616),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(8391),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(3891)] = {
    [sym_type_arguments] = STATE(4611),
    [sym_long_identifier] = STATE(4669),
    [sym_block_comment] = STATE(3891),
    [sym_compiler_directive_decl] = STATE(3891),
    [sym_fsi_directive_decl] = STATE(3891),
    [sym_preproc_line] = STATE(3891),
    [aux_sym_compound_type_repeat1] = STATE(4301),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8394),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_DOT_DOT] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(8398),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8400),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3892)] = {
    [sym_block_comment] = STATE(3892),
    [sym_compiler_directive_decl] = STATE(3892),
    [sym_fsi_directive_decl] = STATE(3892),
    [sym_preproc_line] = STATE(3892),
    [aux_sym_rules_repeat1] = STATE(3892),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_GT_RBRACK] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(8402),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_RBRACK] = ACTIONS(6651),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_RBRACE] = ACTIONS(6651),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_to] = ACTIONS(6653),
    [anon_sym_downto] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_DOT_DOT2] = ACTIONS(6651),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(8405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(3893)] = {
    [sym_type_arguments] = STATE(4553),
    [sym_long_identifier] = STATE(4568),
    [sym_block_comment] = STATE(3893),
    [sym_compiler_directive_decl] = STATE(3893),
    [sym_fsi_directive_decl] = STATE(3893),
    [sym_preproc_line] = STATE(3893),
    [aux_sym_compound_type_repeat1] = STATE(4443),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8358),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8360),
    [anon_sym_LT] = ACTIONS(8362),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8364),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__quoted_close] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3894)] = {
    [sym_block_comment] = STATE(3894),
    [sym_compiler_directive_decl] = STATE(3894),
    [sym_fsi_directive_decl] = STATE(3894),
    [sym_preproc_line] = STATE(3894),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_module] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_open] = ACTIONS(7077),
    [anon_sym_type] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_exception] = ACTIONS(7077),
    [anon_sym_extern] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7079),
    [anon_sym_POUNDload] = ACTIONS(7079),
    [anon_sym_POUNDtime] = ACTIONS(7079),
    [anon_sym_POUNDI] = ACTIONS(7079),
    [anon_sym_POUNDhelp] = ACTIONS(7079),
    [anon_sym_POUNDquit] = ACTIONS(7079),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [anon_sym_POUNDelse] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3895)] = {
    [sym_block_comment] = STATE(3895),
    [sym_compiler_directive_decl] = STATE(3895),
    [sym_fsi_directive_decl] = STATE(3895),
    [sym_preproc_line] = STATE(3895),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_module] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_open] = ACTIONS(7063),
    [anon_sym_type] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_exception] = ACTIONS(7063),
    [anon_sym_extern] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7061),
    [anon_sym_POUNDload] = ACTIONS(7061),
    [anon_sym_POUNDtime] = ACTIONS(7061),
    [anon_sym_POUNDI] = ACTIONS(7061),
    [anon_sym_POUNDhelp] = ACTIONS(7061),
    [anon_sym_POUNDquit] = ACTIONS(7061),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [anon_sym_POUNDelse] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3896)] = {
    [sym_block_comment] = STATE(3896),
    [sym_compiler_directive_decl] = STATE(3896),
    [sym_fsi_directive_decl] = STATE(3896),
    [sym_preproc_line] = STATE(3896),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_module] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_open] = ACTIONS(7067),
    [anon_sym_type] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_exception] = ACTIONS(7067),
    [anon_sym_extern] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7065),
    [anon_sym_POUNDload] = ACTIONS(7065),
    [anon_sym_POUNDtime] = ACTIONS(7065),
    [anon_sym_POUNDI] = ACTIONS(7065),
    [anon_sym_POUNDhelp] = ACTIONS(7065),
    [anon_sym_POUNDquit] = ACTIONS(7065),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [anon_sym_POUNDelse] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3897)] = {
    [sym_block_comment] = STATE(3897),
    [sym_compiler_directive_decl] = STATE(3897),
    [sym_fsi_directive_decl] = STATE(3897),
    [sym_preproc_line] = STATE(3897),
    [ts_builtin_sym_end] = ACTIONS(6894),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_namespace] = ACTIONS(6896),
    [anon_sym_module] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_open] = ACTIONS(6896),
    [anon_sym_type] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_exception] = ACTIONS(6896),
    [anon_sym_extern] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6894),
    [anon_sym_POUNDload] = ACTIONS(6894),
    [anon_sym_POUNDtime] = ACTIONS(6894),
    [anon_sym_POUNDI] = ACTIONS(6894),
    [anon_sym_POUNDhelp] = ACTIONS(6894),
    [anon_sym_POUNDquit] = ACTIONS(6894),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(3898)] = {
    [sym_type_arguments] = STATE(4618),
    [sym_long_identifier] = STATE(4600),
    [sym_block_comment] = STATE(3898),
    [sym_compiler_directive_decl] = STATE(3898),
    [sym_fsi_directive_decl] = STATE(3898),
    [sym_preproc_line] = STATE(3898),
    [aux_sym_compound_type_repeat1] = STATE(4303),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8408),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8410),
    [anon_sym_LT] = ACTIONS(8412),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8414),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__untyped_quoted_close] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3899)] = {
    [sym_attributes] = STATE(13308),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6991),
    [sym_access_modifier] = STATE(22413),
    [sym__class_function_or_value_defn] = STATE(6996),
    [aux_sym__type_extension_inner] = STATE(4439),
    [sym_type_extension_elements] = STATE(8365),
    [sym__type_defn_elements] = STATE(6996),
    [sym_interface_implementation] = STATE(6999),
    [sym_member_defn] = STATE(6999),
    [sym_additional_constr_defn] = STATE(6985),
    [sym_block_comment] = STATE(3899),
    [sym_compiler_directive_decl] = STATE(3899),
    [sym_fsi_directive_decl] = STATE(3899),
    [sym_preproc_line] = STATE(3899),
    [sym_preproc_if_in_class_definition] = STATE(6999),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8341),
    [anon_sym_module] = ACTIONS(8341),
    [anon_sym_null] = ACTIONS(8341),
    [anon_sym_begin] = ACTIONS(8341),
    [anon_sym_open] = ACTIONS(8341),
    [anon_sym_type] = ACTIONS(8341),
    [anon_sym_LBRACK_LT] = ACTIONS(8343),
    [anon_sym_do] = ACTIONS(8341),
    [anon_sym_and] = ACTIONS(8341),
    [anon_sym_let] = ACTIONS(8341),
    [anon_sym_let_BANG] = ACTIONS(8343),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8341),
    [anon_sym_AMP] = ACTIONS(8341),
    [anon_sym_LBRACK] = ACTIONS(8341),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8343),
    [anon_sym_LBRACE] = ACTIONS(8341),
    [anon_sym_LT_AT] = ACTIONS(8341),
    [anon_sym_LT_AT_AT] = ACTIONS(8343),
    [anon_sym_CARET] = ACTIONS(8343),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8343),
    [anon_sym_new] = ACTIONS(8341),
    [anon_sym_with] = ACTIONS(8416),
    [anon_sym_return] = ACTIONS(8341),
    [anon_sym_return_BANG] = ACTIONS(8343),
    [anon_sym_yield] = ACTIONS(8341),
    [anon_sym_yield_BANG] = ACTIONS(8343),
    [anon_sym_lazy] = ACTIONS(8341),
    [anon_sym_assert] = ACTIONS(8341),
    [anon_sym_upcast] = ACTIONS(8341),
    [anon_sym_downcast] = ACTIONS(8341),
    [anon_sym_fixed] = ACTIONS(8341),
    [anon_sym_for] = ACTIONS(8341),
    [anon_sym_while] = ACTIONS(8341),
    [anon_sym_while_BANG] = ACTIONS(8343),
    [anon_sym_if] = ACTIONS(8341),
    [anon_sym_fun] = ACTIONS(8341),
    [anon_sym_try] = ACTIONS(8341),
    [anon_sym_match] = ACTIONS(8341),
    [anon_sym_match_BANG] = ACTIONS(8343),
    [anon_sym_function] = ACTIONS(8341),
    [anon_sym_use] = ACTIONS(8341),
    [anon_sym_use_BANG] = ACTIONS(8343),
    [anon_sym_do_BANG] = ACTIONS(8343),
    [anon_sym_AMP_AMP] = ACTIONS(8343),
    [anon_sym_default] = ACTIONS(8418),
    [anon_sym_static] = ACTIONS(8420),
    [anon_sym_member] = ACTIONS(8422),
    [anon_sym_exception] = ACTIONS(8341),
    [anon_sym_interface] = ACTIONS(8424),
    [anon_sym_abstract] = ACTIONS(8426),
    [anon_sym_val] = ACTIONS(8428),
    [anon_sym_override] = ACTIONS(8418),
    [anon_sym_extern] = ACTIONS(8341),
    [aux_sym_char_token1] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8341),
    [anon_sym_DQUOTE] = ACTIONS(8341),
    [anon_sym_AT_DQUOTE] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [sym_bool] = ACTIONS(8341),
    [sym_unit] = ACTIONS(8343),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8341),
    [sym_op_identifier] = ACTIONS(8343),
    [anon_sym_PLUS] = ACTIONS(8341),
    [anon_sym_DASH] = ACTIONS(8341),
    [anon_sym_PLUS_DOT] = ACTIONS(8343),
    [anon_sym_DASH_DOT] = ACTIONS(8343),
    [anon_sym_PERCENT] = ACTIONS(8341),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8343),
    [anon_sym_TILDE] = ACTIONS(8343),
    [aux_sym_prefix_op_token1] = ACTIONS(8343),
    [sym_int] = ACTIONS(8341),
    [sym_xint] = ACTIONS(8343),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8343),
    [anon_sym_POUNDload] = ACTIONS(8343),
    [anon_sym_POUNDtime] = ACTIONS(8343),
    [anon_sym_POUNDI] = ACTIONS(8343),
    [anon_sym_POUNDhelp] = ACTIONS(8343),
    [anon_sym_POUNDquit] = ACTIONS(8343),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8343),
    [anon_sym_POUNDendif] = ACTIONS(8343),
    [sym__newline] = ACTIONS(8343),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8343),
  },
  [STATE(3900)] = {
    [sym_type_arguments] = STATE(4645),
    [sym_long_identifier] = STATE(4649),
    [sym_block_comment] = STATE(3900),
    [sym_compiler_directive_decl] = STATE(3900),
    [sym_fsi_directive_decl] = STATE(3900),
    [sym_preproc_line] = STATE(3900),
    [aux_sym_compound_type_repeat1] = STATE(4502),
    [sym_identifier] = ACTIONS(8381),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8383),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8385),
    [anon_sym_LT] = ACTIONS(8387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8389),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__multi_dollar_interp_end] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(3901)] = {
    [sym_type_arguments] = STATE(4645),
    [sym_long_identifier] = STATE(4649),
    [sym_block_comment] = STATE(3901),
    [sym_compiler_directive_decl] = STATE(3901),
    [sym_fsi_directive_decl] = STATE(3901),
    [sym_preproc_line] = STATE(3901),
    [aux_sym_compound_type_repeat1] = STATE(4502),
    [sym_identifier] = ACTIONS(8381),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8383),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8385),
    [anon_sym_LT] = ACTIONS(8387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8389),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__multi_dollar_interp_end] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3902)] = {
    [sym_block_comment] = STATE(3902),
    [sym_compiler_directive_decl] = STATE(3902),
    [sym_fsi_directive_decl] = STATE(3902),
    [sym_preproc_line] = STATE(3902),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_module] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_open] = ACTIONS(7125),
    [anon_sym_type] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_exception] = ACTIONS(7125),
    [anon_sym_extern] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7127),
    [anon_sym_POUNDload] = ACTIONS(7127),
    [anon_sym_POUNDtime] = ACTIONS(7127),
    [anon_sym_POUNDI] = ACTIONS(7127),
    [anon_sym_POUNDhelp] = ACTIONS(7127),
    [anon_sym_POUNDquit] = ACTIONS(7127),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [anon_sym_POUNDelse] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(3903)] = {
    [sym_block_comment] = STATE(3903),
    [sym_compiler_directive_decl] = STATE(3903),
    [sym_fsi_directive_decl] = STATE(3903),
    [sym_preproc_line] = STATE(3903),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_module] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_open] = ACTIONS(7129),
    [anon_sym_type] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_exception] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7131),
    [anon_sym_POUNDload] = ACTIONS(7131),
    [anon_sym_POUNDtime] = ACTIONS(7131),
    [anon_sym_POUNDI] = ACTIONS(7131),
    [anon_sym_POUNDhelp] = ACTIONS(7131),
    [anon_sym_POUNDquit] = ACTIONS(7131),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [anon_sym_POUNDelse] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(3904)] = {
    [sym_block_comment] = STATE(3904),
    [sym_compiler_directive_decl] = STATE(3904),
    [sym_fsi_directive_decl] = STATE(3904),
    [sym_preproc_line] = STATE(3904),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_module] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_open] = ACTIONS(7133),
    [anon_sym_type] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_exception] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7135),
    [anon_sym_POUNDload] = ACTIONS(7135),
    [anon_sym_POUNDtime] = ACTIONS(7135),
    [anon_sym_POUNDI] = ACTIONS(7135),
    [anon_sym_POUNDhelp] = ACTIONS(7135),
    [anon_sym_POUNDquit] = ACTIONS(7135),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [anon_sym_POUNDelse] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(3905)] = {
    [sym_block_comment] = STATE(3905),
    [sym_compiler_directive_decl] = STATE(3905),
    [sym_fsi_directive_decl] = STATE(3905),
    [sym_preproc_line] = STATE(3905),
    [ts_builtin_sym_end] = ACTIONS(6904),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_namespace] = ACTIONS(6906),
    [anon_sym_module] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_open] = ACTIONS(6906),
    [anon_sym_type] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_exception] = ACTIONS(6906),
    [anon_sym_extern] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6904),
    [anon_sym_POUNDload] = ACTIONS(6904),
    [anon_sym_POUNDtime] = ACTIONS(6904),
    [anon_sym_POUNDI] = ACTIONS(6904),
    [anon_sym_POUNDhelp] = ACTIONS(6904),
    [anon_sym_POUNDquit] = ACTIONS(6904),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(3906)] = {
    [sym_block_comment] = STATE(3906),
    [sym_compiler_directive_decl] = STATE(3906),
    [sym_fsi_directive_decl] = STATE(3906),
    [sym_preproc_line] = STATE(3906),
    [ts_builtin_sym_end] = ACTIONS(7061),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_namespace] = ACTIONS(7063),
    [anon_sym_module] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_open] = ACTIONS(7063),
    [anon_sym_type] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_exception] = ACTIONS(7063),
    [anon_sym_extern] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7061),
    [anon_sym_POUNDload] = ACTIONS(7061),
    [anon_sym_POUNDtime] = ACTIONS(7061),
    [anon_sym_POUNDI] = ACTIONS(7061),
    [anon_sym_POUNDhelp] = ACTIONS(7061),
    [anon_sym_POUNDquit] = ACTIONS(7061),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(3907)] = {
    [sym_type_arguments] = STATE(4645),
    [sym_long_identifier] = STATE(4649),
    [sym_block_comment] = STATE(3907),
    [sym_compiler_directive_decl] = STATE(3907),
    [sym_fsi_directive_decl] = STATE(3907),
    [sym_preproc_line] = STATE(3907),
    [aux_sym_compound_type_repeat1] = STATE(4502),
    [sym_identifier] = ACTIONS(8381),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8383),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8385),
    [anon_sym_LT] = ACTIONS(8387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8389),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__multi_dollar_interp_end] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(3908)] = {
    [sym_block_comment] = STATE(3908),
    [sym_compiler_directive_decl] = STATE(3908),
    [sym_fsi_directive_decl] = STATE(3908),
    [sym_preproc_line] = STATE(3908),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_module] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_open] = ACTIONS(7027),
    [anon_sym_type] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_exception] = ACTIONS(7027),
    [anon_sym_extern] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7025),
    [anon_sym_POUNDload] = ACTIONS(7025),
    [anon_sym_POUNDtime] = ACTIONS(7025),
    [anon_sym_POUNDI] = ACTIONS(7025),
    [anon_sym_POUNDhelp] = ACTIONS(7025),
    [anon_sym_POUNDquit] = ACTIONS(7025),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [anon_sym_POUNDelse] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(3909)] = {
    [sym_block_comment] = STATE(3909),
    [sym_compiler_directive_decl] = STATE(3909),
    [sym_fsi_directive_decl] = STATE(3909),
    [sym_preproc_line] = STATE(3909),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_module] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_open] = ACTIONS(7035),
    [anon_sym_type] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_exception] = ACTIONS(7035),
    [anon_sym_extern] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7033),
    [anon_sym_POUNDload] = ACTIONS(7033),
    [anon_sym_POUNDtime] = ACTIONS(7033),
    [anon_sym_POUNDI] = ACTIONS(7033),
    [anon_sym_POUNDhelp] = ACTIONS(7033),
    [anon_sym_POUNDquit] = ACTIONS(7033),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [anon_sym_POUNDelse] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(3910)] = {
    [sym_block_comment] = STATE(3910),
    [sym_compiler_directive_decl] = STATE(3910),
    [sym_fsi_directive_decl] = STATE(3910),
    [sym_preproc_line] = STATE(3910),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_as] = ACTIONS(6492),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_DOT_DOT] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__dedent] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(3911)] = {
    [sym_block_comment] = STATE(3911),
    [sym_compiler_directive_decl] = STATE(3911),
    [sym_fsi_directive_decl] = STATE(3911),
    [sym_preproc_line] = STATE(3911),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_module] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_open] = ACTIONS(7085),
    [anon_sym_type] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_exception] = ACTIONS(7085),
    [anon_sym_extern] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7087),
    [anon_sym_POUNDload] = ACTIONS(7087),
    [anon_sym_POUNDtime] = ACTIONS(7087),
    [anon_sym_POUNDI] = ACTIONS(7087),
    [anon_sym_POUNDhelp] = ACTIONS(7087),
    [anon_sym_POUNDquit] = ACTIONS(7087),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [anon_sym_POUNDelse] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3912)] = {
    [sym_block_comment] = STATE(3912),
    [sym_compiler_directive_decl] = STATE(3912),
    [sym_fsi_directive_decl] = STATE(3912),
    [sym_preproc_line] = STATE(3912),
    [ts_builtin_sym_end] = ACTIONS(7045),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_namespace] = ACTIONS(7047),
    [anon_sym_module] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_open] = ACTIONS(7047),
    [anon_sym_type] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_exception] = ACTIONS(7047),
    [anon_sym_extern] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7045),
    [anon_sym_POUNDload] = ACTIONS(7045),
    [anon_sym_POUNDtime] = ACTIONS(7045),
    [anon_sym_POUNDI] = ACTIONS(7045),
    [anon_sym_POUNDhelp] = ACTIONS(7045),
    [anon_sym_POUNDquit] = ACTIONS(7045),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(3913)] = {
    [sym_block_comment] = STATE(3913),
    [sym_compiler_directive_decl] = STATE(3913),
    [sym_fsi_directive_decl] = STATE(3913),
    [sym_preproc_line] = STATE(3913),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_module] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_open] = ACTIONS(7039),
    [anon_sym_type] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_exception] = ACTIONS(7039),
    [anon_sym_extern] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7037),
    [anon_sym_POUNDload] = ACTIONS(7037),
    [anon_sym_POUNDtime] = ACTIONS(7037),
    [anon_sym_POUNDI] = ACTIONS(7037),
    [anon_sym_POUNDhelp] = ACTIONS(7037),
    [anon_sym_POUNDquit] = ACTIONS(7037),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [anon_sym_POUNDelse] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(3914)] = {
    [sym_block_comment] = STATE(3914),
    [sym_compiler_directive_decl] = STATE(3914),
    [sym_fsi_directive_decl] = STATE(3914),
    [sym_preproc_line] = STATE(3914),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_as] = ACTIONS(6427),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_DOT_DOT] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__dedent] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(3915)] = {
    [sym_block_comment] = STATE(3915),
    [sym_compiler_directive_decl] = STATE(3915),
    [sym_fsi_directive_decl] = STATE(3915),
    [sym_preproc_line] = STATE(3915),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_as] = ACTIONS(6389),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_DOT_DOT] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__dedent] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(3916)] = {
    [sym_type_arguments] = STATE(4660),
    [sym_long_identifier] = STATE(4641),
    [sym_block_comment] = STATE(3916),
    [sym_compiler_directive_decl] = STATE(3916),
    [sym_fsi_directive_decl] = STATE(3916),
    [sym_preproc_line] = STATE(3916),
    [aux_sym_compound_type_repeat1] = STATE(4506),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_GT] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8430),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8432),
    [anon_sym_LT] = ACTIONS(8434),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8436),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3917)] = {
    [sym_block_comment] = STATE(3917),
    [sym_compiler_directive_decl] = STATE(3917),
    [sym_fsi_directive_decl] = STATE(3917),
    [sym_preproc_line] = STATE(3917),
    [ts_builtin_sym_end] = ACTIONS(6923),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_namespace] = ACTIONS(6925),
    [anon_sym_module] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_open] = ACTIONS(6925),
    [anon_sym_type] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_exception] = ACTIONS(6925),
    [anon_sym_extern] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6923),
    [anon_sym_POUNDload] = ACTIONS(6923),
    [anon_sym_POUNDtime] = ACTIONS(6923),
    [anon_sym_POUNDI] = ACTIONS(6923),
    [anon_sym_POUNDhelp] = ACTIONS(6923),
    [anon_sym_POUNDquit] = ACTIONS(6923),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(3918)] = {
    [sym_block_comment] = STATE(3918),
    [sym_compiler_directive_decl] = STATE(3918),
    [sym_fsi_directive_decl] = STATE(3918),
    [sym_preproc_line] = STATE(3918),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_module] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_open] = ACTIONS(6875),
    [anon_sym_type] = ACTIONS(6875),
    [anon_sym_LBRACK_LT] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_exception] = ACTIONS(6875),
    [anon_sym_extern] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6873),
    [anon_sym_POUNDload] = ACTIONS(6873),
    [anon_sym_POUNDtime] = ACTIONS(6873),
    [anon_sym_POUNDI] = ACTIONS(6873),
    [anon_sym_POUNDhelp] = ACTIONS(6873),
    [anon_sym_POUNDquit] = ACTIONS(6873),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(3919)] = {
    [sym_block_comment] = STATE(3919),
    [sym_compiler_directive_decl] = STATE(3919),
    [sym_fsi_directive_decl] = STATE(3919),
    [sym_preproc_line] = STATE(3919),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_module] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_open] = ACTIONS(7053),
    [anon_sym_type] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_exception] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7055),
    [anon_sym_POUNDload] = ACTIONS(7055),
    [anon_sym_POUNDtime] = ACTIONS(7055),
    [anon_sym_POUNDI] = ACTIONS(7055),
    [anon_sym_POUNDhelp] = ACTIONS(7055),
    [anon_sym_POUNDquit] = ACTIONS(7055),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [anon_sym_POUNDelse] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(3920)] = {
    [sym_attributes] = STATE(13308),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6991),
    [sym_access_modifier] = STATE(22413),
    [sym__class_function_or_value_defn] = STATE(6996),
    [aux_sym__type_extension_inner] = STATE(4439),
    [sym_type_extension_elements] = STATE(8504),
    [sym__type_defn_elements] = STATE(6996),
    [sym_interface_implementation] = STATE(6999),
    [sym_member_defn] = STATE(6999),
    [sym_additional_constr_defn] = STATE(6985),
    [sym_block_comment] = STATE(3920),
    [sym_compiler_directive_decl] = STATE(3920),
    [sym_fsi_directive_decl] = STATE(3920),
    [sym_preproc_line] = STATE(3920),
    [sym_preproc_if_in_class_definition] = STATE(6999),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8242),
    [anon_sym_module] = ACTIONS(8242),
    [anon_sym_null] = ACTIONS(8242),
    [anon_sym_begin] = ACTIONS(8242),
    [anon_sym_open] = ACTIONS(8242),
    [anon_sym_type] = ACTIONS(8242),
    [anon_sym_LBRACK_LT] = ACTIONS(8244),
    [anon_sym_do] = ACTIONS(8242),
    [anon_sym_and] = ACTIONS(8242),
    [anon_sym_let] = ACTIONS(8242),
    [anon_sym_let_BANG] = ACTIONS(8244),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8242),
    [anon_sym_AMP] = ACTIONS(8242),
    [anon_sym_LBRACK] = ACTIONS(8242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8244),
    [anon_sym_LBRACE] = ACTIONS(8242),
    [anon_sym_LT_AT] = ACTIONS(8242),
    [anon_sym_LT_AT_AT] = ACTIONS(8244),
    [anon_sym_CARET] = ACTIONS(8244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8244),
    [anon_sym_new] = ACTIONS(8242),
    [anon_sym_with] = ACTIONS(8416),
    [anon_sym_return] = ACTIONS(8242),
    [anon_sym_return_BANG] = ACTIONS(8244),
    [anon_sym_yield] = ACTIONS(8242),
    [anon_sym_yield_BANG] = ACTIONS(8244),
    [anon_sym_lazy] = ACTIONS(8242),
    [anon_sym_assert] = ACTIONS(8242),
    [anon_sym_upcast] = ACTIONS(8242),
    [anon_sym_downcast] = ACTIONS(8242),
    [anon_sym_fixed] = ACTIONS(8242),
    [anon_sym_for] = ACTIONS(8242),
    [anon_sym_while] = ACTIONS(8242),
    [anon_sym_while_BANG] = ACTIONS(8244),
    [anon_sym_if] = ACTIONS(8242),
    [anon_sym_fun] = ACTIONS(8242),
    [anon_sym_try] = ACTIONS(8242),
    [anon_sym_match] = ACTIONS(8242),
    [anon_sym_match_BANG] = ACTIONS(8244),
    [anon_sym_function] = ACTIONS(8242),
    [anon_sym_use] = ACTIONS(8242),
    [anon_sym_use_BANG] = ACTIONS(8244),
    [anon_sym_do_BANG] = ACTIONS(8244),
    [anon_sym_AMP_AMP] = ACTIONS(8244),
    [anon_sym_default] = ACTIONS(8418),
    [anon_sym_static] = ACTIONS(8420),
    [anon_sym_member] = ACTIONS(8422),
    [anon_sym_exception] = ACTIONS(8242),
    [anon_sym_interface] = ACTIONS(8424),
    [anon_sym_abstract] = ACTIONS(8426),
    [anon_sym_val] = ACTIONS(8428),
    [anon_sym_override] = ACTIONS(8418),
    [anon_sym_extern] = ACTIONS(8242),
    [aux_sym_char_token1] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8242),
    [anon_sym_DQUOTE] = ACTIONS(8242),
    [anon_sym_AT_DQUOTE] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [sym_bool] = ACTIONS(8242),
    [sym_unit] = ACTIONS(8244),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8242),
    [sym_op_identifier] = ACTIONS(8244),
    [anon_sym_PLUS] = ACTIONS(8242),
    [anon_sym_DASH] = ACTIONS(8242),
    [anon_sym_PLUS_DOT] = ACTIONS(8244),
    [anon_sym_DASH_DOT] = ACTIONS(8244),
    [anon_sym_PERCENT] = ACTIONS(8242),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8244),
    [anon_sym_TILDE] = ACTIONS(8244),
    [aux_sym_prefix_op_token1] = ACTIONS(8244),
    [sym_int] = ACTIONS(8242),
    [sym_xint] = ACTIONS(8244),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8244),
    [anon_sym_POUNDload] = ACTIONS(8244),
    [anon_sym_POUNDtime] = ACTIONS(8244),
    [anon_sym_POUNDI] = ACTIONS(8244),
    [anon_sym_POUNDhelp] = ACTIONS(8244),
    [anon_sym_POUNDquit] = ACTIONS(8244),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8244),
    [anon_sym_POUNDendif] = ACTIONS(8244),
    [sym__newline] = ACTIONS(8244),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8244),
  },
  [STATE(3921)] = {
    [sym_block_comment] = STATE(3921),
    [sym_compiler_directive_decl] = STATE(3921),
    [sym_fsi_directive_decl] = STATE(3921),
    [sym_preproc_line] = STATE(3921),
    [aux_sym_type_argument_repeat1] = STATE(3980),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8438),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6331),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_DOT_DOT2] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3922)] = {
    [sym_block_comment] = STATE(3922),
    [sym_compiler_directive_decl] = STATE(3922),
    [sym_fsi_directive_decl] = STATE(3922),
    [sym_preproc_line] = STATE(3922),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_as] = ACTIONS(6391),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_DOT_DOT] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(3923)] = {
    [sym_block_comment] = STATE(3923),
    [sym_compiler_directive_decl] = STATE(3923),
    [sym_fsi_directive_decl] = STATE(3923),
    [sym_preproc_line] = STATE(3923),
    [ts_builtin_sym_end] = ACTIONS(6997),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_namespace] = ACTIONS(6999),
    [anon_sym_module] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_open] = ACTIONS(6999),
    [anon_sym_type] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_exception] = ACTIONS(6999),
    [anon_sym_extern] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6997),
    [anon_sym_POUNDload] = ACTIONS(6997),
    [anon_sym_POUNDtime] = ACTIONS(6997),
    [anon_sym_POUNDI] = ACTIONS(6997),
    [anon_sym_POUNDhelp] = ACTIONS(6997),
    [anon_sym_POUNDquit] = ACTIONS(6997),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(3924)] = {
    [sym_block_comment] = STATE(3924),
    [sym_compiler_directive_decl] = STATE(3924),
    [sym_fsi_directive_decl] = STATE(3924),
    [sym_preproc_line] = STATE(3924),
    [ts_builtin_sym_end] = ACTIONS(7107),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_namespace] = ACTIONS(7105),
    [anon_sym_module] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_open] = ACTIONS(7105),
    [anon_sym_type] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_exception] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7107),
    [anon_sym_POUNDload] = ACTIONS(7107),
    [anon_sym_POUNDtime] = ACTIONS(7107),
    [anon_sym_POUNDI] = ACTIONS(7107),
    [anon_sym_POUNDhelp] = ACTIONS(7107),
    [anon_sym_POUNDquit] = ACTIONS(7107),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3925)] = {
    [sym_block_comment] = STATE(3925),
    [sym_compiler_directive_decl] = STATE(3925),
    [sym_fsi_directive_decl] = STATE(3925),
    [sym_preproc_line] = STATE(3925),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_module] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_open] = ACTIONS(7089),
    [anon_sym_type] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_exception] = ACTIONS(7089),
    [anon_sym_extern] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7091),
    [anon_sym_POUNDload] = ACTIONS(7091),
    [anon_sym_POUNDtime] = ACTIONS(7091),
    [anon_sym_POUNDI] = ACTIONS(7091),
    [anon_sym_POUNDhelp] = ACTIONS(7091),
    [anon_sym_POUNDquit] = ACTIONS(7091),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [anon_sym_POUNDelse] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(3926)] = {
    [sym_block_comment] = STATE(3926),
    [sym_compiler_directive_decl] = STATE(3926),
    [sym_fsi_directive_decl] = STATE(3926),
    [sym_preproc_line] = STATE(3926),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_module] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_open] = ACTIONS(7043),
    [anon_sym_type] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_exception] = ACTIONS(7043),
    [anon_sym_extern] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7041),
    [anon_sym_POUNDload] = ACTIONS(7041),
    [anon_sym_POUNDtime] = ACTIONS(7041),
    [anon_sym_POUNDI] = ACTIONS(7041),
    [anon_sym_POUNDhelp] = ACTIONS(7041),
    [anon_sym_POUNDquit] = ACTIONS(7041),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [anon_sym_POUNDelse] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(3927)] = {
    [sym_block_comment] = STATE(3927),
    [sym_compiler_directive_decl] = STATE(3927),
    [sym_fsi_directive_decl] = STATE(3927),
    [sym_preproc_line] = STATE(3927),
    [ts_builtin_sym_end] = ACTIONS(7111),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_namespace] = ACTIONS(7109),
    [anon_sym_module] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_open] = ACTIONS(7109),
    [anon_sym_type] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_exception] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7111),
    [anon_sym_POUNDload] = ACTIONS(7111),
    [anon_sym_POUNDtime] = ACTIONS(7111),
    [anon_sym_POUNDI] = ACTIONS(7111),
    [anon_sym_POUNDhelp] = ACTIONS(7111),
    [anon_sym_POUNDquit] = ACTIONS(7111),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3928)] = {
    [sym_block_comment] = STATE(3928),
    [sym_compiler_directive_decl] = STATE(3928),
    [sym_fsi_directive_decl] = STATE(3928),
    [sym_preproc_line] = STATE(3928),
    [ts_builtin_sym_end] = ACTIONS(7005),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_namespace] = ACTIONS(7007),
    [anon_sym_module] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_open] = ACTIONS(7007),
    [anon_sym_type] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_exception] = ACTIONS(7007),
    [anon_sym_extern] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7005),
    [anon_sym_POUNDload] = ACTIONS(7005),
    [anon_sym_POUNDtime] = ACTIONS(7005),
    [anon_sym_POUNDI] = ACTIONS(7005),
    [anon_sym_POUNDhelp] = ACTIONS(7005),
    [anon_sym_POUNDquit] = ACTIONS(7005),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(3929)] = {
    [sym_block_comment] = STATE(3929),
    [sym_compiler_directive_decl] = STATE(3929),
    [sym_fsi_directive_decl] = STATE(3929),
    [sym_preproc_line] = STATE(3929),
    [ts_builtin_sym_end] = ACTIONS(7123),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_namespace] = ACTIONS(7121),
    [anon_sym_module] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_open] = ACTIONS(7121),
    [anon_sym_type] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_exception] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7123),
    [anon_sym_POUNDload] = ACTIONS(7123),
    [anon_sym_POUNDtime] = ACTIONS(7123),
    [anon_sym_POUNDI] = ACTIONS(7123),
    [anon_sym_POUNDhelp] = ACTIONS(7123),
    [anon_sym_POUNDquit] = ACTIONS(7123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(3930)] = {
    [sym_block_comment] = STATE(3930),
    [sym_compiler_directive_decl] = STATE(3930),
    [sym_fsi_directive_decl] = STATE(3930),
    [sym_preproc_line] = STATE(3930),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_GT_RBRACK] = ACTIONS(6635),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_RBRACK] = ACTIONS(6635),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_RBRACE] = ACTIONS(6635),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_to] = ACTIONS(6637),
    [anon_sym_downto] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_DOT_DOT2] = ACTIONS(6635),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(8441),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(3931)] = {
    [sym_block_comment] = STATE(3931),
    [sym_compiler_directive_decl] = STATE(3931),
    [sym_fsi_directive_decl] = STATE(3931),
    [sym_preproc_line] = STATE(3931),
    [ts_builtin_sym_end] = ACTIONS(6934),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_namespace] = ACTIONS(6936),
    [anon_sym_module] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_open] = ACTIONS(6936),
    [anon_sym_type] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_exception] = ACTIONS(6936),
    [anon_sym_extern] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6934),
    [anon_sym_POUNDload] = ACTIONS(6934),
    [anon_sym_POUNDtime] = ACTIONS(6934),
    [anon_sym_POUNDI] = ACTIONS(6934),
    [anon_sym_POUNDhelp] = ACTIONS(6934),
    [anon_sym_POUNDquit] = ACTIONS(6934),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(3932)] = {
    [sym_block_comment] = STATE(3932),
    [sym_compiler_directive_decl] = STATE(3932),
    [sym_fsi_directive_decl] = STATE(3932),
    [sym_preproc_line] = STATE(3932),
    [aux_sym_long_identifier_repeat1] = STATE(3968),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_GT_RBRACK] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_RBRACK] = ACTIONS(6628),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_RBRACE] = ACTIONS(6628),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8443),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_to] = ACTIONS(6625),
    [anon_sym_downto] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_DOT_DOT2] = ACTIONS(6628),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(3933)] = {
    [sym_block_comment] = STATE(3933),
    [sym_compiler_directive_decl] = STATE(3933),
    [sym_fsi_directive_decl] = STATE(3933),
    [sym_preproc_line] = STATE(3933),
    [sym_identifier] = ACTIONS(8447),
    [anon_sym_module] = ACTIONS(8447),
    [anon_sym_null] = ACTIONS(8447),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(8447),
    [anon_sym_open] = ACTIONS(8447),
    [anon_sym_type] = ACTIONS(8447),
    [anon_sym_LBRACK_LT] = ACTIONS(8449),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(8447),
    [anon_sym_let] = ACTIONS(8447),
    [anon_sym_let_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN] = ACTIONS(8447),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(8447),
    [anon_sym_LBRACK] = ACTIONS(8447),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8449),
    [anon_sym_LBRACE] = ACTIONS(8447),
    [anon_sym_LT_AT] = ACTIONS(8447),
    [anon_sym_LT_AT_AT] = ACTIONS(8447),
    [anon_sym_CARET] = ACTIONS(8447),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8449),
    [anon_sym_new] = ACTIONS(8447),
    [anon_sym_return] = ACTIONS(8447),
    [anon_sym_return_BANG] = ACTIONS(8449),
    [anon_sym_yield] = ACTIONS(8447),
    [anon_sym_yield_BANG] = ACTIONS(8449),
    [anon_sym_lazy] = ACTIONS(8447),
    [anon_sym_assert] = ACTIONS(8447),
    [anon_sym_upcast] = ACTIONS(8447),
    [anon_sym_downcast] = ACTIONS(8447),
    [anon_sym_fixed] = ACTIONS(8447),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(8447),
    [anon_sym_while] = ACTIONS(8447),
    [anon_sym_while_BANG] = ACTIONS(8449),
    [anon_sym_if] = ACTIONS(8447),
    [anon_sym_fun] = ACTIONS(8447),
    [anon_sym_try] = ACTIONS(8447),
    [anon_sym_match] = ACTIONS(8447),
    [anon_sym_match_BANG] = ACTIONS(8449),
    [anon_sym_function] = ACTIONS(8447),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(8447),
    [anon_sym_use_BANG] = ACTIONS(8449),
    [anon_sym_do_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(8447),
    [anon_sym_exception] = ACTIONS(8447),
    [anon_sym_extern] = ACTIONS(8447),
    [aux_sym_char_token1] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8447),
    [anon_sym_DQUOTE] = ACTIONS(8447),
    [anon_sym_AT_DQUOTE] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [sym_bool] = ACTIONS(8447),
    [sym_unit] = ACTIONS(8449),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8447),
    [sym_op_identifier] = ACTIONS(8447),
    [anon_sym_PLUS] = ACTIONS(8447),
    [anon_sym_DASH] = ACTIONS(8447),
    [anon_sym_PLUS_DOT] = ACTIONS(8447),
    [anon_sym_DASH_DOT] = ACTIONS(8447),
    [anon_sym_PERCENT] = ACTIONS(8447),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8447),
    [anon_sym_TILDE] = ACTIONS(8449),
    [aux_sym_prefix_op_token1] = ACTIONS(8447),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(8447),
    [sym_xint] = ACTIONS(8449),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8449),
    [anon_sym_POUNDload] = ACTIONS(8449),
    [anon_sym_POUNDtime] = ACTIONS(8449),
    [anon_sym_POUNDI] = ACTIONS(8449),
    [anon_sym_POUNDhelp] = ACTIONS(8449),
    [anon_sym_POUNDquit] = ACTIONS(8449),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8449),
    [anon_sym_POUNDendif] = ACTIONS(8449),
    [anon_sym_POUNDelse] = ACTIONS(8449),
    [sym__newline] = ACTIONS(8449),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8449),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3934)] = {
    [sym_block_comment] = STATE(3934),
    [sym_compiler_directive_decl] = STATE(3934),
    [sym_fsi_directive_decl] = STATE(3934),
    [sym_preproc_line] = STATE(3934),
    [sym_identifier] = ACTIONS(8447),
    [anon_sym_module] = ACTIONS(8447),
    [anon_sym_null] = ACTIONS(8447),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(8447),
    [anon_sym_open] = ACTIONS(8447),
    [anon_sym_type] = ACTIONS(8447),
    [anon_sym_LBRACK_LT] = ACTIONS(8449),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(8447),
    [anon_sym_let] = ACTIONS(8447),
    [anon_sym_let_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN] = ACTIONS(8447),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(8447),
    [anon_sym_LBRACK] = ACTIONS(8447),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8449),
    [anon_sym_LBRACE] = ACTIONS(8447),
    [anon_sym_LT_AT] = ACTIONS(8447),
    [anon_sym_LT_AT_AT] = ACTIONS(8447),
    [anon_sym_CARET] = ACTIONS(8447),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8449),
    [anon_sym_new] = ACTIONS(8447),
    [anon_sym_return] = ACTIONS(8447),
    [anon_sym_return_BANG] = ACTIONS(8449),
    [anon_sym_yield] = ACTIONS(8447),
    [anon_sym_yield_BANG] = ACTIONS(8449),
    [anon_sym_lazy] = ACTIONS(8447),
    [anon_sym_assert] = ACTIONS(8447),
    [anon_sym_upcast] = ACTIONS(8447),
    [anon_sym_downcast] = ACTIONS(8447),
    [anon_sym_fixed] = ACTIONS(8447),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(8447),
    [anon_sym_while] = ACTIONS(8447),
    [anon_sym_while_BANG] = ACTIONS(8449),
    [anon_sym_if] = ACTIONS(8447),
    [anon_sym_fun] = ACTIONS(8447),
    [anon_sym_try] = ACTIONS(8447),
    [anon_sym_match] = ACTIONS(8447),
    [anon_sym_match_BANG] = ACTIONS(8449),
    [anon_sym_function] = ACTIONS(8447),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(8447),
    [anon_sym_use_BANG] = ACTIONS(8449),
    [anon_sym_do_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(8447),
    [anon_sym_exception] = ACTIONS(8447),
    [anon_sym_extern] = ACTIONS(8447),
    [aux_sym_char_token1] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8447),
    [anon_sym_DQUOTE] = ACTIONS(8447),
    [anon_sym_AT_DQUOTE] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [sym_bool] = ACTIONS(8447),
    [sym_unit] = ACTIONS(8449),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8447),
    [sym_op_identifier] = ACTIONS(8447),
    [anon_sym_PLUS] = ACTIONS(8447),
    [anon_sym_DASH] = ACTIONS(8447),
    [anon_sym_PLUS_DOT] = ACTIONS(8447),
    [anon_sym_DASH_DOT] = ACTIONS(8447),
    [anon_sym_PERCENT] = ACTIONS(8447),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8447),
    [anon_sym_TILDE] = ACTIONS(8449),
    [aux_sym_prefix_op_token1] = ACTIONS(8447),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(8447),
    [sym_xint] = ACTIONS(8449),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8449),
    [anon_sym_POUNDload] = ACTIONS(8449),
    [anon_sym_POUNDtime] = ACTIONS(8449),
    [anon_sym_POUNDI] = ACTIONS(8449),
    [anon_sym_POUNDhelp] = ACTIONS(8449),
    [anon_sym_POUNDquit] = ACTIONS(8449),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8449),
    [anon_sym_POUNDendif] = ACTIONS(8449),
    [anon_sym_POUNDelse] = ACTIONS(8449),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8449),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(3935)] = {
    [sym_block_comment] = STATE(3935),
    [sym_compiler_directive_decl] = STATE(3935),
    [sym_fsi_directive_decl] = STATE(3935),
    [sym_preproc_line] = STATE(3935),
    [aux_sym_compound_type_repeat1] = STATE(3985),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_DOT_DOT] = ACTIONS(6338),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_DOT_DOT2] = ACTIONS(6340),
    [anon_sym_STAR] = ACTIONS(8146),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__dedent] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(3936)] = {
    [sym_block_comment] = STATE(3936),
    [sym_compiler_directive_decl] = STATE(3936),
    [sym_fsi_directive_decl] = STATE(3936),
    [sym_preproc_line] = STATE(3936),
    [ts_builtin_sym_end] = ACTIONS(6942),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_namespace] = ACTIONS(6944),
    [anon_sym_module] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_open] = ACTIONS(6944),
    [anon_sym_type] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_exception] = ACTIONS(6944),
    [anon_sym_extern] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6942),
    [anon_sym_POUNDload] = ACTIONS(6942),
    [anon_sym_POUNDtime] = ACTIONS(6942),
    [anon_sym_POUNDI] = ACTIONS(6942),
    [anon_sym_POUNDhelp] = ACTIONS(6942),
    [anon_sym_POUNDquit] = ACTIONS(6942),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(3937)] = {
    [sym_block_comment] = STATE(3937),
    [sym_compiler_directive_decl] = STATE(3937),
    [sym_fsi_directive_decl] = STATE(3937),
    [sym_preproc_line] = STATE(3937),
    [ts_builtin_sym_end] = ACTIONS(6938),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_namespace] = ACTIONS(6940),
    [anon_sym_module] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_open] = ACTIONS(6940),
    [anon_sym_type] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_exception] = ACTIONS(6940),
    [anon_sym_extern] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6938),
    [anon_sym_POUNDload] = ACTIONS(6938),
    [anon_sym_POUNDtime] = ACTIONS(6938),
    [anon_sym_POUNDI] = ACTIONS(6938),
    [anon_sym_POUNDhelp] = ACTIONS(6938),
    [anon_sym_POUNDquit] = ACTIONS(6938),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(3938)] = {
    [sym_block_comment] = STATE(3938),
    [sym_compiler_directive_decl] = STATE(3938),
    [sym_fsi_directive_decl] = STATE(3938),
    [sym_preproc_line] = STATE(3938),
    [ts_builtin_sym_end] = ACTIONS(7079),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_namespace] = ACTIONS(7077),
    [anon_sym_module] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_open] = ACTIONS(7077),
    [anon_sym_type] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_exception] = ACTIONS(7077),
    [anon_sym_extern] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7079),
    [anon_sym_POUNDload] = ACTIONS(7079),
    [anon_sym_POUNDtime] = ACTIONS(7079),
    [anon_sym_POUNDI] = ACTIONS(7079),
    [anon_sym_POUNDhelp] = ACTIONS(7079),
    [anon_sym_POUNDquit] = ACTIONS(7079),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(3939)] = {
    [sym_block_comment] = STATE(3939),
    [sym_compiler_directive_decl] = STATE(3939),
    [sym_fsi_directive_decl] = STATE(3939),
    [sym_preproc_line] = STATE(3939),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_module] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_open] = ACTIONS(7059),
    [anon_sym_type] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_exception] = ACTIONS(7059),
    [anon_sym_extern] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7057),
    [anon_sym_POUNDload] = ACTIONS(7057),
    [anon_sym_POUNDtime] = ACTIONS(7057),
    [anon_sym_POUNDI] = ACTIONS(7057),
    [anon_sym_POUNDhelp] = ACTIONS(7057),
    [anon_sym_POUNDquit] = ACTIONS(7057),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [anon_sym_POUNDelse] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(3940)] = {
    [sym_block_comment] = STATE(3940),
    [sym_compiler_directive_decl] = STATE(3940),
    [sym_fsi_directive_decl] = STATE(3940),
    [sym_preproc_line] = STATE(3940),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3941)] = {
    [sym_block_comment] = STATE(3941),
    [sym_compiler_directive_decl] = STATE(3941),
    [sym_fsi_directive_decl] = STATE(3941),
    [sym_preproc_line] = STATE(3941),
    [ts_builtin_sym_end] = ACTIONS(7071),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_namespace] = ACTIONS(7069),
    [anon_sym_module] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_open] = ACTIONS(7069),
    [anon_sym_type] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_exception] = ACTIONS(7069),
    [anon_sym_extern] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7071),
    [anon_sym_POUNDload] = ACTIONS(7071),
    [anon_sym_POUNDtime] = ACTIONS(7071),
    [anon_sym_POUNDI] = ACTIONS(7071),
    [anon_sym_POUNDhelp] = ACTIONS(7071),
    [anon_sym_POUNDquit] = ACTIONS(7071),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(3942)] = {
    [sym_block_comment] = STATE(3942),
    [sym_compiler_directive_decl] = STATE(3942),
    [sym_fsi_directive_decl] = STATE(3942),
    [sym_preproc_line] = STATE(3942),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(3943)] = {
    [sym_block_comment] = STATE(3943),
    [sym_compiler_directive_decl] = STATE(3943),
    [sym_fsi_directive_decl] = STATE(3943),
    [sym_preproc_line] = STATE(3943),
    [ts_builtin_sym_end] = ACTIONS(6957),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_namespace] = ACTIONS(6959),
    [anon_sym_module] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_open] = ACTIONS(6959),
    [anon_sym_type] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_exception] = ACTIONS(6959),
    [anon_sym_extern] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6957),
    [anon_sym_POUNDload] = ACTIONS(6957),
    [anon_sym_POUNDtime] = ACTIONS(6957),
    [anon_sym_POUNDI] = ACTIONS(6957),
    [anon_sym_POUNDhelp] = ACTIONS(6957),
    [anon_sym_POUNDquit] = ACTIONS(6957),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(3944)] = {
    [sym_block_comment] = STATE(3944),
    [sym_compiler_directive_decl] = STATE(3944),
    [sym_fsi_directive_decl] = STATE(3944),
    [sym_preproc_line] = STATE(3944),
    [aux_sym_rules_repeat1] = STATE(3892),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_GT_RBRACK] = ACTIONS(6605),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8366),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_RBRACK] = ACTIONS(6605),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_RBRACE] = ACTIONS(6605),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_to] = ACTIONS(6607),
    [anon_sym_downto] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_DOT_DOT2] = ACTIONS(6605),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8368),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(3945)] = {
    [sym_block_comment] = STATE(3945),
    [sym_compiler_directive_decl] = STATE(3945),
    [sym_fsi_directive_decl] = STATE(3945),
    [sym_preproc_line] = STATE(3945),
    [ts_builtin_sym_end] = ACTIONS(6985),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_namespace] = ACTIONS(6987),
    [anon_sym_module] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_open] = ACTIONS(6987),
    [anon_sym_type] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_exception] = ACTIONS(6987),
    [anon_sym_extern] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6985),
    [anon_sym_POUNDload] = ACTIONS(6985),
    [anon_sym_POUNDtime] = ACTIONS(6985),
    [anon_sym_POUNDI] = ACTIONS(6985),
    [anon_sym_POUNDhelp] = ACTIONS(6985),
    [anon_sym_POUNDquit] = ACTIONS(6985),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3946)] = {
    [sym_block_comment] = STATE(3946),
    [sym_compiler_directive_decl] = STATE(3946),
    [sym_fsi_directive_decl] = STATE(3946),
    [sym_preproc_line] = STATE(3946),
    [ts_builtin_sym_end] = ACTIONS(6989),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_namespace] = ACTIONS(6991),
    [anon_sym_module] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_open] = ACTIONS(6991),
    [anon_sym_type] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_exception] = ACTIONS(6991),
    [anon_sym_extern] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6989),
    [anon_sym_POUNDload] = ACTIONS(6989),
    [anon_sym_POUNDtime] = ACTIONS(6989),
    [anon_sym_POUNDI] = ACTIONS(6989),
    [anon_sym_POUNDhelp] = ACTIONS(6989),
    [anon_sym_POUNDquit] = ACTIONS(6989),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3947)] = {
    [sym_block_comment] = STATE(3947),
    [sym_compiler_directive_decl] = STATE(3947),
    [sym_fsi_directive_decl] = STATE(3947),
    [sym_preproc_line] = STATE(3947),
    [ts_builtin_sym_end] = ACTIONS(7013),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_namespace] = ACTIONS(7015),
    [anon_sym_module] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_open] = ACTIONS(7015),
    [anon_sym_type] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_exception] = ACTIONS(7015),
    [anon_sym_extern] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7013),
    [anon_sym_POUNDload] = ACTIONS(7013),
    [anon_sym_POUNDtime] = ACTIONS(7013),
    [anon_sym_POUNDI] = ACTIONS(7013),
    [anon_sym_POUNDhelp] = ACTIONS(7013),
    [anon_sym_POUNDquit] = ACTIONS(7013),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(3948)] = {
    [sym_block_comment] = STATE(3948),
    [sym_compiler_directive_decl] = STATE(3948),
    [sym_fsi_directive_decl] = STATE(3948),
    [sym_preproc_line] = STATE(3948),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_module] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_open] = ACTIONS(6967),
    [anon_sym_type] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_exception] = ACTIONS(6967),
    [anon_sym_extern] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6965),
    [anon_sym_POUNDload] = ACTIONS(6965),
    [anon_sym_POUNDtime] = ACTIONS(6965),
    [anon_sym_POUNDI] = ACTIONS(6965),
    [anon_sym_POUNDhelp] = ACTIONS(6965),
    [anon_sym_POUNDquit] = ACTIONS(6965),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [anon_sym_POUNDelse] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3949)] = {
    [sym_block_comment] = STATE(3949),
    [sym_compiler_directive_decl] = STATE(3949),
    [sym_fsi_directive_decl] = STATE(3949),
    [sym_preproc_line] = STATE(3949),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_module] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_open] = ACTIONS(6971),
    [anon_sym_type] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_exception] = ACTIONS(6971),
    [anon_sym_extern] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6969),
    [anon_sym_POUNDload] = ACTIONS(6969),
    [anon_sym_POUNDtime] = ACTIONS(6969),
    [anon_sym_POUNDI] = ACTIONS(6969),
    [anon_sym_POUNDhelp] = ACTIONS(6969),
    [anon_sym_POUNDquit] = ACTIONS(6969),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [anon_sym_POUNDelse] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(3950)] = {
    [sym_block_comment] = STATE(3950),
    [sym_compiler_directive_decl] = STATE(3950),
    [sym_fsi_directive_decl] = STATE(3950),
    [sym_preproc_line] = STATE(3950),
    [ts_builtin_sym_end] = ACTIONS(7029),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_namespace] = ACTIONS(7031),
    [anon_sym_module] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_open] = ACTIONS(7031),
    [anon_sym_type] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_exception] = ACTIONS(7031),
    [anon_sym_extern] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7029),
    [anon_sym_POUNDload] = ACTIONS(7029),
    [anon_sym_POUNDtime] = ACTIONS(7029),
    [anon_sym_POUNDI] = ACTIONS(7029),
    [anon_sym_POUNDhelp] = ACTIONS(7029),
    [anon_sym_POUNDquit] = ACTIONS(7029),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(3951)] = {
    [sym_block_comment] = STATE(3951),
    [sym_compiler_directive_decl] = STATE(3951),
    [sym_fsi_directive_decl] = STATE(3951),
    [sym_preproc_line] = STATE(3951),
    [ts_builtin_sym_end] = ACTIONS(6993),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_namespace] = ACTIONS(6995),
    [anon_sym_module] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_open] = ACTIONS(6995),
    [anon_sym_type] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_exception] = ACTIONS(6995),
    [anon_sym_extern] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6993),
    [anon_sym_POUNDload] = ACTIONS(6993),
    [anon_sym_POUNDtime] = ACTIONS(6993),
    [anon_sym_POUNDI] = ACTIONS(6993),
    [anon_sym_POUNDhelp] = ACTIONS(6993),
    [anon_sym_POUNDquit] = ACTIONS(6993),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3952)] = {
    [sym_block_comment] = STATE(3952),
    [sym_compiler_directive_decl] = STATE(3952),
    [sym_fsi_directive_decl] = STATE(3952),
    [sym_preproc_line] = STATE(3952),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_module] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_open] = ACTIONS(7093),
    [anon_sym_type] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_exception] = ACTIONS(7093),
    [anon_sym_extern] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7095),
    [anon_sym_POUNDload] = ACTIONS(7095),
    [anon_sym_POUNDtime] = ACTIONS(7095),
    [anon_sym_POUNDI] = ACTIONS(7095),
    [anon_sym_POUNDhelp] = ACTIONS(7095),
    [anon_sym_POUNDquit] = ACTIONS(7095),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [anon_sym_POUNDelse] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(3953)] = {
    [sym_block_comment] = STATE(3953),
    [sym_compiler_directive_decl] = STATE(3953),
    [sym_fsi_directive_decl] = STATE(3953),
    [sym_preproc_line] = STATE(3953),
    [ts_builtin_sym_end] = ACTIONS(7143),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_namespace] = ACTIONS(7141),
    [anon_sym_module] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_open] = ACTIONS(7141),
    [anon_sym_type] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_exception] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7143),
    [anon_sym_POUNDload] = ACTIONS(7143),
    [anon_sym_POUNDtime] = ACTIONS(7143),
    [anon_sym_POUNDI] = ACTIONS(7143),
    [anon_sym_POUNDhelp] = ACTIONS(7143),
    [anon_sym_POUNDquit] = ACTIONS(7143),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(3954)] = {
    [sym_type_arguments] = STATE(4611),
    [sym_long_identifier] = STATE(4669),
    [sym_block_comment] = STATE(3954),
    [sym_compiler_directive_decl] = STATE(3954),
    [sym_fsi_directive_decl] = STATE(3954),
    [sym_preproc_line] = STATE(3954),
    [aux_sym_compound_type_repeat1] = STATE(4301),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6265),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_DOT_DOT] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(8398),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8400),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3955)] = {
    [sym_block_comment] = STATE(3955),
    [sym_compiler_directive_decl] = STATE(3955),
    [sym_fsi_directive_decl] = STATE(3955),
    [sym_preproc_line] = STATE(3955),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [anon_sym_POUNDelse] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(3956)] = {
    [sym_block_comment] = STATE(3956),
    [sym_compiler_directive_decl] = STATE(3956),
    [sym_fsi_directive_decl] = STATE(3956),
    [sym_preproc_line] = STATE(3956),
    [aux_sym_sequential_expression_repeat1] = STATE(3956),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8451),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(3957)] = {
    [sym_block_comment] = STATE(3957),
    [sym_compiler_directive_decl] = STATE(3957),
    [sym_fsi_directive_decl] = STATE(3957),
    [sym_preproc_line] = STATE(3957),
    [ts_builtin_sym_end] = ACTIONS(6898),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_namespace] = ACTIONS(6900),
    [anon_sym_module] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_open] = ACTIONS(6900),
    [anon_sym_type] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_exception] = ACTIONS(6900),
    [anon_sym_extern] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6898),
    [anon_sym_POUNDload] = ACTIONS(6898),
    [anon_sym_POUNDtime] = ACTIONS(6898),
    [anon_sym_POUNDI] = ACTIONS(6898),
    [anon_sym_POUNDhelp] = ACTIONS(6898),
    [anon_sym_POUNDquit] = ACTIONS(6898),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(3958)] = {
    [sym_block_comment] = STATE(3958),
    [sym_compiler_directive_decl] = STATE(3958),
    [sym_fsi_directive_decl] = STATE(3958),
    [sym_preproc_line] = STATE(3958),
    [aux_sym_sequential_expression_repeat1] = STATE(3956),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_module] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_open] = ACTIONS(6740),
    [anon_sym_type] = ACTIONS(6740),
    [anon_sym_LBRACK_LT] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_exception] = ACTIONS(6740),
    [anon_sym_extern] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6738),
    [anon_sym_POUNDload] = ACTIONS(6738),
    [anon_sym_POUNDtime] = ACTIONS(6738),
    [anon_sym_POUNDI] = ACTIONS(6738),
    [anon_sym_POUNDhelp] = ACTIONS(6738),
    [anon_sym_POUNDquit] = ACTIONS(6738),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(3959)] = {
    [sym_block_comment] = STATE(3959),
    [sym_compiler_directive_decl] = STATE(3959),
    [sym_fsi_directive_decl] = STATE(3959),
    [sym_preproc_line] = STATE(3959),
    [aux_sym_long_identifier_repeat1] = STATE(3997),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8454),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6360),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_DOT_DOT2] = ACTIONS(6358),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3960)] = {
    [sym_block_comment] = STATE(3960),
    [sym_compiler_directive_decl] = STATE(3960),
    [sym_fsi_directive_decl] = STATE(3960),
    [sym_preproc_line] = STATE(3960),
    [ts_builtin_sym_end] = ACTIONS(6950),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_namespace] = ACTIONS(6952),
    [anon_sym_module] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_open] = ACTIONS(6952),
    [anon_sym_type] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_exception] = ACTIONS(6952),
    [anon_sym_extern] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6950),
    [anon_sym_POUNDload] = ACTIONS(6950),
    [anon_sym_POUNDtime] = ACTIONS(6950),
    [anon_sym_POUNDI] = ACTIONS(6950),
    [anon_sym_POUNDhelp] = ACTIONS(6950),
    [anon_sym_POUNDquit] = ACTIONS(6950),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(3961)] = {
    [sym_block_comment] = STATE(3961),
    [sym_compiler_directive_decl] = STATE(3961),
    [sym_fsi_directive_decl] = STATE(3961),
    [sym_preproc_line] = STATE(3961),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_as] = ACTIONS(6369),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3962)] = {
    [sym_block_comment] = STATE(3962),
    [sym_compiler_directive_decl] = STATE(3962),
    [sym_fsi_directive_decl] = STATE(3962),
    [sym_preproc_line] = STATE(3962),
    [ts_builtin_sym_end] = ACTIONS(6805),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_namespace] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(3963)] = {
    [sym_block_comment] = STATE(3963),
    [sym_compiler_directive_decl] = STATE(3963),
    [sym_fsi_directive_decl] = STATE(3963),
    [sym_preproc_line] = STATE(3963),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_module] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_open] = ACTIONS(7097),
    [anon_sym_type] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_exception] = ACTIONS(7097),
    [anon_sym_extern] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7099),
    [anon_sym_POUNDload] = ACTIONS(7099),
    [anon_sym_POUNDtime] = ACTIONS(7099),
    [anon_sym_POUNDI] = ACTIONS(7099),
    [anon_sym_POUNDhelp] = ACTIONS(7099),
    [anon_sym_POUNDquit] = ACTIONS(7099),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [anon_sym_POUNDelse] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(3964)] = {
    [sym_block_comment] = STATE(3964),
    [sym_compiler_directive_decl] = STATE(3964),
    [sym_fsi_directive_decl] = STATE(3964),
    [sym_preproc_line] = STATE(3964),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_as] = ACTIONS(6369),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(3965)] = {
    [sym_block_comment] = STATE(3965),
    [sym_compiler_directive_decl] = STATE(3965),
    [sym_fsi_directive_decl] = STATE(3965),
    [sym_preproc_line] = STATE(3965),
    [ts_builtin_sym_end] = ACTIONS(7065),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_namespace] = ACTIONS(7067),
    [anon_sym_module] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_open] = ACTIONS(7067),
    [anon_sym_type] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_exception] = ACTIONS(7067),
    [anon_sym_extern] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7065),
    [anon_sym_POUNDload] = ACTIONS(7065),
    [anon_sym_POUNDtime] = ACTIONS(7065),
    [anon_sym_POUNDI] = ACTIONS(7065),
    [anon_sym_POUNDhelp] = ACTIONS(7065),
    [anon_sym_POUNDquit] = ACTIONS(7065),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(3966)] = {
    [sym_block_comment] = STATE(3966),
    [sym_compiler_directive_decl] = STATE(3966),
    [sym_fsi_directive_decl] = STATE(3966),
    [sym_preproc_line] = STATE(3966),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_module] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_open] = ACTIONS(6975),
    [anon_sym_type] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_exception] = ACTIONS(6975),
    [anon_sym_extern] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6973),
    [anon_sym_POUNDload] = ACTIONS(6973),
    [anon_sym_POUNDtime] = ACTIONS(6973),
    [anon_sym_POUNDI] = ACTIONS(6973),
    [anon_sym_POUNDhelp] = ACTIONS(6973),
    [anon_sym_POUNDquit] = ACTIONS(6973),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [anon_sym_POUNDelse] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(3967)] = {
    [sym_attributes] = STATE(13389),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7140),
    [sym_access_modifier] = STATE(22369),
    [sym__class_function_or_value_defn] = STATE(7141),
    [aux_sym__type_extension_inner] = STATE(4347),
    [sym_type_extension_elements] = STATE(8419),
    [sym__type_defn_elements] = STATE(7141),
    [sym_interface_implementation] = STATE(6973),
    [sym_member_defn] = STATE(6973),
    [sym_additional_constr_defn] = STATE(7045),
    [sym_block_comment] = STATE(3967),
    [sym_compiler_directive_decl] = STATE(3967),
    [sym_fsi_directive_decl] = STATE(3967),
    [sym_preproc_line] = STATE(3967),
    [sym_preproc_if_in_class_definition] = STATE(6973),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8341),
    [anon_sym_module] = ACTIONS(8341),
    [anon_sym_null] = ACTIONS(8341),
    [anon_sym_begin] = ACTIONS(8341),
    [anon_sym_open] = ACTIONS(8341),
    [anon_sym_type] = ACTIONS(8341),
    [anon_sym_LBRACK_LT] = ACTIONS(8343),
    [anon_sym_do] = ACTIONS(8341),
    [anon_sym_and] = ACTIONS(8341),
    [anon_sym_let] = ACTIONS(8341),
    [anon_sym_let_BANG] = ACTIONS(8343),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8341),
    [anon_sym_AMP] = ACTIONS(8341),
    [anon_sym_LBRACK] = ACTIONS(8341),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8343),
    [anon_sym_LBRACE] = ACTIONS(8341),
    [anon_sym_LT_AT] = ACTIONS(8341),
    [anon_sym_LT_AT_AT] = ACTIONS(8343),
    [anon_sym_CARET] = ACTIONS(8343),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8343),
    [anon_sym_new] = ACTIONS(8341),
    [anon_sym_with] = ACTIONS(8456),
    [anon_sym_return] = ACTIONS(8341),
    [anon_sym_return_BANG] = ACTIONS(8343),
    [anon_sym_yield] = ACTIONS(8341),
    [anon_sym_yield_BANG] = ACTIONS(8343),
    [anon_sym_lazy] = ACTIONS(8341),
    [anon_sym_assert] = ACTIONS(8341),
    [anon_sym_upcast] = ACTIONS(8341),
    [anon_sym_downcast] = ACTIONS(8341),
    [anon_sym_fixed] = ACTIONS(8341),
    [anon_sym_for] = ACTIONS(8341),
    [anon_sym_while] = ACTIONS(8341),
    [anon_sym_while_BANG] = ACTIONS(8343),
    [anon_sym_if] = ACTIONS(8341),
    [anon_sym_fun] = ACTIONS(8341),
    [anon_sym_try] = ACTIONS(8341),
    [anon_sym_match] = ACTIONS(8341),
    [anon_sym_match_BANG] = ACTIONS(8343),
    [anon_sym_function] = ACTIONS(8341),
    [anon_sym_use] = ACTIONS(8341),
    [anon_sym_use_BANG] = ACTIONS(8343),
    [anon_sym_do_BANG] = ACTIONS(8343),
    [anon_sym_AMP_AMP] = ACTIONS(8343),
    [anon_sym_default] = ACTIONS(8458),
    [anon_sym_static] = ACTIONS(8460),
    [anon_sym_member] = ACTIONS(8462),
    [anon_sym_exception] = ACTIONS(8341),
    [anon_sym_interface] = ACTIONS(8464),
    [anon_sym_abstract] = ACTIONS(8466),
    [anon_sym_val] = ACTIONS(8468),
    [anon_sym_override] = ACTIONS(8458),
    [anon_sym_extern] = ACTIONS(8341),
    [aux_sym_char_token1] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8341),
    [anon_sym_DQUOTE] = ACTIONS(8341),
    [anon_sym_AT_DQUOTE] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [sym_bool] = ACTIONS(8341),
    [sym_unit] = ACTIONS(8343),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8341),
    [sym_op_identifier] = ACTIONS(8343),
    [anon_sym_PLUS] = ACTIONS(8341),
    [anon_sym_DASH] = ACTIONS(8341),
    [anon_sym_PLUS_DOT] = ACTIONS(8343),
    [anon_sym_DASH_DOT] = ACTIONS(8343),
    [anon_sym_PERCENT] = ACTIONS(8341),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8343),
    [anon_sym_TILDE] = ACTIONS(8343),
    [aux_sym_prefix_op_token1] = ACTIONS(8343),
    [sym_int] = ACTIONS(8341),
    [sym_xint] = ACTIONS(8343),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8343),
    [anon_sym_POUNDload] = ACTIONS(8343),
    [anon_sym_POUNDtime] = ACTIONS(8343),
    [anon_sym_POUNDI] = ACTIONS(8343),
    [anon_sym_POUNDhelp] = ACTIONS(8343),
    [anon_sym_POUNDquit] = ACTIONS(8343),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8343),
    [anon_sym_POUNDendif] = ACTIONS(8343),
    [anon_sym_POUNDelse] = ACTIONS(8343),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8343),
  },
  [STATE(3968)] = {
    [sym_block_comment] = STATE(3968),
    [sym_compiler_directive_decl] = STATE(3968),
    [sym_fsi_directive_decl] = STATE(3968),
    [sym_preproc_line] = STATE(3968),
    [aux_sym_long_identifier_repeat1] = STATE(4125),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_GT_RBRACK] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_RBRACK] = ACTIONS(6358),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_RBRACE] = ACTIONS(6358),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8470),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_to] = ACTIONS(6360),
    [anon_sym_downto] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_DOT_DOT2] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(3969)] = {
    [sym_block_comment] = STATE(3969),
    [sym_compiler_directive_decl] = STATE(3969),
    [sym_fsi_directive_decl] = STATE(3969),
    [sym_preproc_line] = STATE(3969),
    [ts_builtin_sym_end] = ACTIONS(7017),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_namespace] = ACTIONS(7019),
    [anon_sym_module] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_open] = ACTIONS(7019),
    [anon_sym_type] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_exception] = ACTIONS(7019),
    [anon_sym_extern] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7017),
    [anon_sym_POUNDload] = ACTIONS(7017),
    [anon_sym_POUNDtime] = ACTIONS(7017),
    [anon_sym_POUNDI] = ACTIONS(7017),
    [anon_sym_POUNDhelp] = ACTIONS(7017),
    [anon_sym_POUNDquit] = ACTIONS(7017),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(3970)] = {
    [sym_block_comment] = STATE(3970),
    [sym_compiler_directive_decl] = STATE(3970),
    [sym_fsi_directive_decl] = STATE(3970),
    [sym_preproc_line] = STATE(3970),
    [ts_builtin_sym_end] = ACTIONS(6965),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_namespace] = ACTIONS(6967),
    [anon_sym_module] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_open] = ACTIONS(6967),
    [anon_sym_type] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_exception] = ACTIONS(6967),
    [anon_sym_extern] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6965),
    [anon_sym_POUNDload] = ACTIONS(6965),
    [anon_sym_POUNDtime] = ACTIONS(6965),
    [anon_sym_POUNDI] = ACTIONS(6965),
    [anon_sym_POUNDhelp] = ACTIONS(6965),
    [anon_sym_POUNDquit] = ACTIONS(6965),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(3971)] = {
    [sym_block_comment] = STATE(3971),
    [sym_compiler_directive_decl] = STATE(3971),
    [sym_fsi_directive_decl] = STATE(3971),
    [sym_preproc_line] = STATE(3971),
    [ts_builtin_sym_end] = ACTIONS(7157),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_namespace] = ACTIONS(7155),
    [anon_sym_module] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_open] = ACTIONS(7155),
    [anon_sym_type] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_exception] = ACTIONS(7155),
    [anon_sym_extern] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7157),
    [anon_sym_POUNDload] = ACTIONS(7157),
    [anon_sym_POUNDtime] = ACTIONS(7157),
    [anon_sym_POUNDI] = ACTIONS(7157),
    [anon_sym_POUNDhelp] = ACTIONS(7157),
    [anon_sym_POUNDquit] = ACTIONS(7157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(3972)] = {
    [sym_block_comment] = STATE(3972),
    [sym_compiler_directive_decl] = STATE(3972),
    [sym_fsi_directive_decl] = STATE(3972),
    [sym_preproc_line] = STATE(3972),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_module] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_open] = ACTIONS(6979),
    [anon_sym_type] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_exception] = ACTIONS(6979),
    [anon_sym_extern] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6977),
    [anon_sym_POUNDload] = ACTIONS(6977),
    [anon_sym_POUNDtime] = ACTIONS(6977),
    [anon_sym_POUNDI] = ACTIONS(6977),
    [anon_sym_POUNDhelp] = ACTIONS(6977),
    [anon_sym_POUNDquit] = ACTIONS(6977),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [anon_sym_POUNDelse] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3973)] = {
    [sym_block_comment] = STATE(3973),
    [sym_compiler_directive_decl] = STATE(3973),
    [sym_fsi_directive_decl] = STATE(3973),
    [sym_preproc_line] = STATE(3973),
    [ts_builtin_sym_end] = ACTIONS(7087),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_namespace] = ACTIONS(7085),
    [anon_sym_module] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_open] = ACTIONS(7085),
    [anon_sym_type] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_exception] = ACTIONS(7085),
    [anon_sym_extern] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7087),
    [anon_sym_POUNDload] = ACTIONS(7087),
    [anon_sym_POUNDtime] = ACTIONS(7087),
    [anon_sym_POUNDI] = ACTIONS(7087),
    [anon_sym_POUNDhelp] = ACTIONS(7087),
    [anon_sym_POUNDquit] = ACTIONS(7087),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(3974)] = {
    [sym_block_comment] = STATE(3974),
    [sym_compiler_directive_decl] = STATE(3974),
    [sym_fsi_directive_decl] = STATE(3974),
    [sym_preproc_line] = STATE(3974),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_module] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_open] = ACTIONS(7101),
    [anon_sym_type] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_exception] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7103),
    [anon_sym_POUNDload] = ACTIONS(7103),
    [anon_sym_POUNDtime] = ACTIONS(7103),
    [anon_sym_POUNDI] = ACTIONS(7103),
    [anon_sym_POUNDhelp] = ACTIONS(7103),
    [anon_sym_POUNDquit] = ACTIONS(7103),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [anon_sym_POUNDelse] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(3975)] = {
    [sym_type_arguments] = STATE(4645),
    [sym_long_identifier] = STATE(4649),
    [sym_block_comment] = STATE(3975),
    [sym_compiler_directive_decl] = STATE(3975),
    [sym_fsi_directive_decl] = STATE(3975),
    [sym_preproc_line] = STATE(3975),
    [aux_sym_compound_type_repeat1] = STATE(4502),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8383),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8385),
    [anon_sym_LT] = ACTIONS(8387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8389),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__multi_dollar_interp_end] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(3976)] = {
    [sym_block_comment] = STATE(3976),
    [sym_compiler_directive_decl] = STATE(3976),
    [sym_fsi_directive_decl] = STATE(3976),
    [sym_preproc_line] = STATE(3976),
    [ts_builtin_sym_end] = ACTIONS(6977),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_namespace] = ACTIONS(6979),
    [anon_sym_module] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_open] = ACTIONS(6979),
    [anon_sym_type] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_exception] = ACTIONS(6979),
    [anon_sym_extern] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6977),
    [anon_sym_POUNDload] = ACTIONS(6977),
    [anon_sym_POUNDtime] = ACTIONS(6977),
    [anon_sym_POUNDI] = ACTIONS(6977),
    [anon_sym_POUNDhelp] = ACTIONS(6977),
    [anon_sym_POUNDquit] = ACTIONS(6977),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(3977)] = {
    [sym_block_comment] = STATE(3977),
    [sym_compiler_directive_decl] = STATE(3977),
    [sym_fsi_directive_decl] = STATE(3977),
    [sym_preproc_line] = STATE(3977),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_module] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_open] = ACTIONS(6983),
    [anon_sym_type] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_exception] = ACTIONS(6983),
    [anon_sym_extern] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6981),
    [anon_sym_POUNDload] = ACTIONS(6981),
    [anon_sym_POUNDtime] = ACTIONS(6981),
    [anon_sym_POUNDI] = ACTIONS(6981),
    [anon_sym_POUNDhelp] = ACTIONS(6981),
    [anon_sym_POUNDquit] = ACTIONS(6981),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [anon_sym_POUNDelse] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(3978)] = {
    [sym_block_comment] = STATE(3978),
    [sym_compiler_directive_decl] = STATE(3978),
    [sym_fsi_directive_decl] = STATE(3978),
    [sym_preproc_line] = STATE(3978),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_module] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_open] = ACTIONS(6750),
    [anon_sym_type] = ACTIONS(6750),
    [anon_sym_LBRACK_LT] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_exception] = ACTIONS(6750),
    [anon_sym_extern] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6752),
    [anon_sym_POUNDload] = ACTIONS(6752),
    [anon_sym_POUNDtime] = ACTIONS(6752),
    [anon_sym_POUNDI] = ACTIONS(6752),
    [anon_sym_POUNDhelp] = ACTIONS(6752),
    [anon_sym_POUNDquit] = ACTIONS(6752),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(3979)] = {
    [sym_block_comment] = STATE(3979),
    [sym_compiler_directive_decl] = STATE(3979),
    [sym_fsi_directive_decl] = STATE(3979),
    [sym_preproc_line] = STATE(3979),
    [aux_sym_rules_repeat1] = STATE(3944),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_GT_RBRACK] = ACTIONS(6588),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(8366),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_RBRACK] = ACTIONS(6588),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_RBRACE] = ACTIONS(6588),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_to] = ACTIONS(6590),
    [anon_sym_downto] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_DOT_DOT2] = ACTIONS(6588),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(8472),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(3980)] = {
    [sym_block_comment] = STATE(3980),
    [sym_compiler_directive_decl] = STATE(3980),
    [sym_fsi_directive_decl] = STATE(3980),
    [sym_preproc_line] = STATE(3980),
    [aux_sym_type_argument_repeat1] = STATE(4089),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_DOT_DOT] = ACTIONS(6308),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_DOT_DOT2] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__dedent] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(3981)] = {
    [sym_block_comment] = STATE(3981),
    [sym_compiler_directive_decl] = STATE(3981),
    [sym_fsi_directive_decl] = STATE(3981),
    [sym_preproc_line] = STATE(3981),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_as] = ACTIONS(6373),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_DOT_DOT] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8475),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__dedent] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(3982)] = {
    [sym_block_comment] = STATE(3982),
    [sym_compiler_directive_decl] = STATE(3982),
    [sym_fsi_directive_decl] = STATE(3982),
    [sym_preproc_line] = STATE(3982),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_module] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_open] = ACTIONS(6987),
    [anon_sym_type] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_exception] = ACTIONS(6987),
    [anon_sym_extern] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6985),
    [anon_sym_POUNDload] = ACTIONS(6985),
    [anon_sym_POUNDtime] = ACTIONS(6985),
    [anon_sym_POUNDI] = ACTIONS(6985),
    [anon_sym_POUNDhelp] = ACTIONS(6985),
    [anon_sym_POUNDquit] = ACTIONS(6985),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [anon_sym_POUNDelse] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(3983)] = {
    [sym_block_comment] = STATE(3983),
    [sym_compiler_directive_decl] = STATE(3983),
    [sym_fsi_directive_decl] = STATE(3983),
    [sym_preproc_line] = STATE(3983),
    [ts_builtin_sym_end] = ACTIONS(7021),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_namespace] = ACTIONS(7023),
    [anon_sym_module] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_open] = ACTIONS(7023),
    [anon_sym_type] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_exception] = ACTIONS(7023),
    [anon_sym_extern] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7021),
    [anon_sym_POUNDload] = ACTIONS(7021),
    [anon_sym_POUNDtime] = ACTIONS(7021),
    [anon_sym_POUNDI] = ACTIONS(7021),
    [anon_sym_POUNDhelp] = ACTIONS(7021),
    [anon_sym_POUNDquit] = ACTIONS(7021),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(3984)] = {
    [sym_block_comment] = STATE(3984),
    [sym_compiler_directive_decl] = STATE(3984),
    [sym_fsi_directive_decl] = STATE(3984),
    [sym_preproc_line] = STATE(3984),
    [ts_builtin_sym_end] = ACTIONS(6890),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_namespace] = ACTIONS(6892),
    [anon_sym_module] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_open] = ACTIONS(6892),
    [anon_sym_type] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_exception] = ACTIONS(6892),
    [anon_sym_extern] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6890),
    [anon_sym_POUNDload] = ACTIONS(6890),
    [anon_sym_POUNDtime] = ACTIONS(6890),
    [anon_sym_POUNDI] = ACTIONS(6890),
    [anon_sym_POUNDhelp] = ACTIONS(6890),
    [anon_sym_POUNDquit] = ACTIONS(6890),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(3985)] = {
    [sym_block_comment] = STATE(3985),
    [sym_compiler_directive_decl] = STATE(3985),
    [sym_fsi_directive_decl] = STATE(3985),
    [sym_preproc_line] = STATE(3985),
    [aux_sym_compound_type_repeat1] = STATE(3985),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6193),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_DOT_DOT2] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(8477),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(3986)] = {
    [sym_block_comment] = STATE(3986),
    [sym_compiler_directive_decl] = STATE(3986),
    [sym_fsi_directive_decl] = STATE(3986),
    [sym_preproc_line] = STATE(3986),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_module] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_open] = ACTIONS(6828),
    [anon_sym_type] = ACTIONS(6828),
    [anon_sym_LBRACK_LT] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_exception] = ACTIONS(6828),
    [anon_sym_extern] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6830),
    [anon_sym_POUNDload] = ACTIONS(6830),
    [anon_sym_POUNDtime] = ACTIONS(6830),
    [anon_sym_POUNDI] = ACTIONS(6830),
    [anon_sym_POUNDhelp] = ACTIONS(6830),
    [anon_sym_POUNDquit] = ACTIONS(6830),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(3987)] = {
    [sym_block_comment] = STATE(3987),
    [sym_compiler_directive_decl] = STATE(3987),
    [sym_fsi_directive_decl] = STATE(3987),
    [sym_preproc_line] = STATE(3987),
    [ts_builtin_sym_end] = ACTIONS(7115),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_namespace] = ACTIONS(7113),
    [anon_sym_module] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_open] = ACTIONS(7113),
    [anon_sym_type] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_exception] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7115),
    [anon_sym_POUNDload] = ACTIONS(7115),
    [anon_sym_POUNDtime] = ACTIONS(7115),
    [anon_sym_POUNDI] = ACTIONS(7115),
    [anon_sym_POUNDhelp] = ACTIONS(7115),
    [anon_sym_POUNDquit] = ACTIONS(7115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(3988)] = {
    [sym_block_comment] = STATE(3988),
    [sym_compiler_directive_decl] = STATE(3988),
    [sym_fsi_directive_decl] = STATE(3988),
    [sym_preproc_line] = STATE(3988),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_GT_RBRACK] = ACTIONS(6515),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_RBRACK] = ACTIONS(6515),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_RBRACE] = ACTIONS(6515),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_to] = ACTIONS(6513),
    [anon_sym_downto] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_DOT_DOT2] = ACTIONS(6515),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(3989)] = {
    [sym_block_comment] = STATE(3989),
    [sym_compiler_directive_decl] = STATE(3989),
    [sym_fsi_directive_decl] = STATE(3989),
    [sym_preproc_line] = STATE(3989),
    [ts_builtin_sym_end] = ACTIONS(6946),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_namespace] = ACTIONS(6948),
    [anon_sym_module] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_open] = ACTIONS(6948),
    [anon_sym_type] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_exception] = ACTIONS(6948),
    [anon_sym_extern] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6946),
    [anon_sym_POUNDload] = ACTIONS(6946),
    [anon_sym_POUNDtime] = ACTIONS(6946),
    [anon_sym_POUNDI] = ACTIONS(6946),
    [anon_sym_POUNDhelp] = ACTIONS(6946),
    [anon_sym_POUNDquit] = ACTIONS(6946),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(3990)] = {
    [sym_block_comment] = STATE(3990),
    [sym_compiler_directive_decl] = STATE(3990),
    [sym_fsi_directive_decl] = STATE(3990),
    [sym_preproc_line] = STATE(3990),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_module] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_open] = ACTIONS(6991),
    [anon_sym_type] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_exception] = ACTIONS(6991),
    [anon_sym_extern] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6989),
    [anon_sym_POUNDload] = ACTIONS(6989),
    [anon_sym_POUNDtime] = ACTIONS(6989),
    [anon_sym_POUNDI] = ACTIONS(6989),
    [anon_sym_POUNDhelp] = ACTIONS(6989),
    [anon_sym_POUNDquit] = ACTIONS(6989),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [anon_sym_POUNDelse] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(3991)] = {
    [sym_block_comment] = STATE(3991),
    [sym_compiler_directive_decl] = STATE(3991),
    [sym_fsi_directive_decl] = STATE(3991),
    [sym_preproc_line] = STATE(3991),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_module] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_open] = ACTIONS(7105),
    [anon_sym_type] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_exception] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7107),
    [anon_sym_POUNDload] = ACTIONS(7107),
    [anon_sym_POUNDtime] = ACTIONS(7107),
    [anon_sym_POUNDI] = ACTIONS(7107),
    [anon_sym_POUNDhelp] = ACTIONS(7107),
    [anon_sym_POUNDquit] = ACTIONS(7107),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [anon_sym_POUNDelse] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(3992)] = {
    [sym_block_comment] = STATE(3992),
    [sym_compiler_directive_decl] = STATE(3992),
    [sym_fsi_directive_decl] = STATE(3992),
    [sym_preproc_line] = STATE(3992),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_module] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_open] = ACTIONS(6784),
    [anon_sym_type] = ACTIONS(6784),
    [anon_sym_LBRACK_LT] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8480),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_exception] = ACTIONS(6784),
    [anon_sym_extern] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6786),
    [anon_sym_POUNDload] = ACTIONS(6786),
    [anon_sym_POUNDtime] = ACTIONS(6786),
    [anon_sym_POUNDI] = ACTIONS(6786),
    [anon_sym_POUNDhelp] = ACTIONS(6786),
    [anon_sym_POUNDquit] = ACTIONS(6786),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(3993)] = {
    [sym_block_comment] = STATE(3993),
    [sym_compiler_directive_decl] = STATE(3993),
    [sym_fsi_directive_decl] = STATE(3993),
    [sym_preproc_line] = STATE(3993),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_as] = ACTIONS(6331),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(3994)] = {
    [sym_block_comment] = STATE(3994),
    [sym_compiler_directive_decl] = STATE(3994),
    [sym_fsi_directive_decl] = STATE(3994),
    [sym_preproc_line] = STATE(3994),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_module] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_open] = ACTIONS(6813),
    [anon_sym_type] = ACTIONS(6813),
    [anon_sym_LBRACK_LT] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_exception] = ACTIONS(6813),
    [anon_sym_extern] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6811),
    [anon_sym_POUNDload] = ACTIONS(6811),
    [anon_sym_POUNDtime] = ACTIONS(6811),
    [anon_sym_POUNDI] = ACTIONS(6811),
    [anon_sym_POUNDhelp] = ACTIONS(6811),
    [anon_sym_POUNDquit] = ACTIONS(6811),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(3995)] = {
    [sym_block_comment] = STATE(3995),
    [sym_compiler_directive_decl] = STATE(3995),
    [sym_fsi_directive_decl] = STATE(3995),
    [sym_preproc_line] = STATE(3995),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_module] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_open] = ACTIONS(7109),
    [anon_sym_type] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_exception] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7111),
    [anon_sym_POUNDload] = ACTIONS(7111),
    [anon_sym_POUNDtime] = ACTIONS(7111),
    [anon_sym_POUNDI] = ACTIONS(7111),
    [anon_sym_POUNDhelp] = ACTIONS(7111),
    [anon_sym_POUNDquit] = ACTIONS(7111),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [anon_sym_POUNDelse] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(3996)] = {
    [sym_block_comment] = STATE(3996),
    [sym_compiler_directive_decl] = STATE(3996),
    [sym_fsi_directive_decl] = STATE(3996),
    [sym_preproc_line] = STATE(3996),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_module] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_open] = ACTIONS(6995),
    [anon_sym_type] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_exception] = ACTIONS(6995),
    [anon_sym_extern] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6993),
    [anon_sym_POUNDload] = ACTIONS(6993),
    [anon_sym_POUNDtime] = ACTIONS(6993),
    [anon_sym_POUNDI] = ACTIONS(6993),
    [anon_sym_POUNDhelp] = ACTIONS(6993),
    [anon_sym_POUNDquit] = ACTIONS(6993),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [anon_sym_POUNDelse] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(3997)] = {
    [sym_block_comment] = STATE(3997),
    [sym_compiler_directive_decl] = STATE(3997),
    [sym_fsi_directive_decl] = STATE(3997),
    [sym_preproc_line] = STATE(3997),
    [aux_sym_long_identifier_repeat1] = STATE(3997),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8482),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6295),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3998)] = {
    [sym_block_comment] = STATE(3998),
    [sym_compiler_directive_decl] = STATE(3998),
    [sym_fsi_directive_decl] = STATE(3998),
    [sym_preproc_line] = STATE(3998),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(3999)] = {
    [sym_elif_expression] = STATE(4544),
    [sym_block_comment] = STATE(3999),
    [sym_compiler_directive_decl] = STATE(3999),
    [sym_fsi_directive_decl] = STATE(3999),
    [sym_preproc_line] = STATE(3999),
    [aux_sym_if_expression_repeat1] = STATE(3999),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_as] = ACTIONS(6496),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8485),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_DOT_DOT] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__dedent] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4000)] = {
    [sym_type_arguments] = STATE(4611),
    [sym_long_identifier] = STATE(4669),
    [sym_block_comment] = STATE(4000),
    [sym_compiler_directive_decl] = STATE(4000),
    [sym_fsi_directive_decl] = STATE(4000),
    [sym_preproc_line] = STATE(4000),
    [aux_sym_compound_type_repeat1] = STATE(4301),
    [sym_identifier] = ACTIONS(8488),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8394),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_DOT_DOT] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(8398),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8400),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(4001)] = {
    [sym_block_comment] = STATE(4001),
    [sym_compiler_directive_decl] = STATE(4001),
    [sym_fsi_directive_decl] = STATE(4001),
    [sym_preproc_line] = STATE(4001),
    [ts_builtin_sym_end] = ACTIONS(7127),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_namespace] = ACTIONS(7125),
    [anon_sym_module] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_open] = ACTIONS(7125),
    [anon_sym_type] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_exception] = ACTIONS(7125),
    [anon_sym_extern] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7127),
    [anon_sym_POUNDload] = ACTIONS(7127),
    [anon_sym_POUNDtime] = ACTIONS(7127),
    [anon_sym_POUNDI] = ACTIONS(7127),
    [anon_sym_POUNDhelp] = ACTIONS(7127),
    [anon_sym_POUNDquit] = ACTIONS(7127),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(4002)] = {
    [sym_block_comment] = STATE(4002),
    [sym_compiler_directive_decl] = STATE(4002),
    [sym_fsi_directive_decl] = STATE(4002),
    [sym_preproc_line] = STATE(4002),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_module] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_open] = ACTIONS(6944),
    [anon_sym_type] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_exception] = ACTIONS(6944),
    [anon_sym_extern] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6942),
    [anon_sym_POUNDload] = ACTIONS(6942),
    [anon_sym_POUNDtime] = ACTIONS(6942),
    [anon_sym_POUNDI] = ACTIONS(6942),
    [anon_sym_POUNDhelp] = ACTIONS(6942),
    [anon_sym_POUNDquit] = ACTIONS(6942),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [anon_sym_POUNDelse] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(4003)] = {
    [sym_type_arguments] = STATE(4709),
    [sym_long_identifier] = STATE(4716),
    [sym_block_comment] = STATE(4003),
    [sym_compiler_directive_decl] = STATE(4003),
    [sym_fsi_directive_decl] = STATE(4003),
    [sym_preproc_line] = STATE(4003),
    [aux_sym_compound_type_repeat1] = STATE(4431),
    [sym_identifier] = ACTIONS(8490),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8492),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8494),
    [anon_sym_LT] = ACTIONS(8496),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8498),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(4004)] = {
    [sym_type_arguments] = STATE(4611),
    [sym_long_identifier] = STATE(4669),
    [sym_block_comment] = STATE(4004),
    [sym_compiler_directive_decl] = STATE(4004),
    [sym_fsi_directive_decl] = STATE(4004),
    [sym_preproc_line] = STATE(4004),
    [aux_sym_compound_type_repeat1] = STATE(4301),
    [sym_identifier] = ACTIONS(8488),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8394),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(8398),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8400),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4005)] = {
    [sym_block_comment] = STATE(4005),
    [sym_compiler_directive_decl] = STATE(4005),
    [sym_fsi_directive_decl] = STATE(4005),
    [sym_preproc_line] = STATE(4005),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_as] = ACTIONS(6420),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_DOT_DOT] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__dedent] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4006)] = {
    [sym_type_arguments] = STATE(4611),
    [sym_long_identifier] = STATE(4669),
    [sym_block_comment] = STATE(4006),
    [sym_compiler_directive_decl] = STATE(4006),
    [sym_fsi_directive_decl] = STATE(4006),
    [sym_preproc_line] = STATE(4006),
    [aux_sym_compound_type_repeat1] = STATE(4301),
    [sym_identifier] = ACTIONS(8488),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8394),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_DOT_DOT] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(8398),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8400),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(4007)] = {
    [sym_block_comment] = STATE(4007),
    [sym_compiler_directive_decl] = STATE(4007),
    [sym_fsi_directive_decl] = STATE(4007),
    [sym_preproc_line] = STATE(4007),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4008)] = {
    [sym_block_comment] = STATE(4008),
    [sym_compiler_directive_decl] = STATE(4008),
    [sym_fsi_directive_decl] = STATE(4008),
    [sym_preproc_line] = STATE(4008),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_module] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_open] = ACTIONS(6999),
    [anon_sym_type] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_exception] = ACTIONS(6999),
    [anon_sym_extern] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6997),
    [anon_sym_POUNDload] = ACTIONS(6997),
    [anon_sym_POUNDtime] = ACTIONS(6997),
    [anon_sym_POUNDI] = ACTIONS(6997),
    [anon_sym_POUNDhelp] = ACTIONS(6997),
    [anon_sym_POUNDquit] = ACTIONS(6997),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [anon_sym_POUNDelse] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(4009)] = {
    [sym_block_comment] = STATE(4009),
    [sym_compiler_directive_decl] = STATE(4009),
    [sym_fsi_directive_decl] = STATE(4009),
    [sym_preproc_line] = STATE(4009),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_module] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_open] = ACTIONS(7003),
    [anon_sym_type] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_exception] = ACTIONS(7003),
    [anon_sym_extern] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7001),
    [anon_sym_POUNDload] = ACTIONS(7001),
    [anon_sym_POUNDtime] = ACTIONS(7001),
    [anon_sym_POUNDI] = ACTIONS(7001),
    [anon_sym_POUNDhelp] = ACTIONS(7001),
    [anon_sym_POUNDquit] = ACTIONS(7001),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [anon_sym_POUNDelse] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(4010)] = {
    [sym_block_comment] = STATE(4010),
    [sym_compiler_directive_decl] = STATE(4010),
    [sym_fsi_directive_decl] = STATE(4010),
    [sym_preproc_line] = STATE(4010),
    [ts_builtin_sym_end] = ACTIONS(7147),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_namespace] = ACTIONS(7145),
    [anon_sym_module] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_open] = ACTIONS(7145),
    [anon_sym_type] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_exception] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7147),
    [anon_sym_POUNDload] = ACTIONS(7147),
    [anon_sym_POUNDtime] = ACTIONS(7147),
    [anon_sym_POUNDI] = ACTIONS(7147),
    [anon_sym_POUNDhelp] = ACTIONS(7147),
    [anon_sym_POUNDquit] = ACTIONS(7147),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(4011)] = {
    [sym_block_comment] = STATE(4011),
    [sym_compiler_directive_decl] = STATE(4011),
    [sym_fsi_directive_decl] = STATE(4011),
    [sym_preproc_line] = STATE(4011),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_as] = ACTIONS(6312),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4012)] = {
    [sym_block_comment] = STATE(4012),
    [sym_compiler_directive_decl] = STATE(4012),
    [sym_fsi_directive_decl] = STATE(4012),
    [sym_preproc_line] = STATE(4012),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_module] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_open] = ACTIONS(6768),
    [anon_sym_type] = ACTIONS(6768),
    [anon_sym_LBRACK_LT] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8500),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_exception] = ACTIONS(6768),
    [anon_sym_extern] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6766),
    [anon_sym_POUNDload] = ACTIONS(6766),
    [anon_sym_POUNDtime] = ACTIONS(6766),
    [anon_sym_POUNDI] = ACTIONS(6766),
    [anon_sym_POUNDhelp] = ACTIONS(6766),
    [anon_sym_POUNDquit] = ACTIONS(6766),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(4013)] = {
    [sym_block_comment] = STATE(4013),
    [sym_compiler_directive_decl] = STATE(4013),
    [sym_fsi_directive_decl] = STATE(4013),
    [sym_preproc_line] = STATE(4013),
    [ts_builtin_sym_end] = ACTIONS(7161),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_namespace] = ACTIONS(7159),
    [anon_sym_module] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_open] = ACTIONS(7159),
    [anon_sym_type] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_exception] = ACTIONS(7159),
    [anon_sym_extern] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7161),
    [anon_sym_POUNDload] = ACTIONS(7161),
    [anon_sym_POUNDtime] = ACTIONS(7161),
    [anon_sym_POUNDI] = ACTIONS(7161),
    [anon_sym_POUNDhelp] = ACTIONS(7161),
    [anon_sym_POUNDquit] = ACTIONS(7161),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(4014)] = {
    [sym_block_comment] = STATE(4014),
    [sym_compiler_directive_decl] = STATE(4014),
    [sym_fsi_directive_decl] = STATE(4014),
    [sym_preproc_line] = STATE(4014),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_module] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_open] = ACTIONS(7007),
    [anon_sym_type] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_exception] = ACTIONS(7007),
    [anon_sym_extern] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7005),
    [anon_sym_POUNDload] = ACTIONS(7005),
    [anon_sym_POUNDtime] = ACTIONS(7005),
    [anon_sym_POUNDI] = ACTIONS(7005),
    [anon_sym_POUNDhelp] = ACTIONS(7005),
    [anon_sym_POUNDquit] = ACTIONS(7005),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [anon_sym_POUNDelse] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(4015)] = {
    [sym_block_comment] = STATE(4015),
    [sym_compiler_directive_decl] = STATE(4015),
    [sym_fsi_directive_decl] = STATE(4015),
    [sym_preproc_line] = STATE(4015),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_module] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_open] = ACTIONS(6612),
    [anon_sym_type] = ACTIONS(6612),
    [anon_sym_LBRACK_LT] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_exception] = ACTIONS(6612),
    [anon_sym_extern] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6614),
    [anon_sym_POUNDload] = ACTIONS(6614),
    [anon_sym_POUNDtime] = ACTIONS(6614),
    [anon_sym_POUNDI] = ACTIONS(6614),
    [anon_sym_POUNDhelp] = ACTIONS(6614),
    [anon_sym_POUNDquit] = ACTIONS(6614),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4016)] = {
    [sym_block_comment] = STATE(4016),
    [sym_compiler_directive_decl] = STATE(4016),
    [sym_fsi_directive_decl] = STATE(4016),
    [sym_preproc_line] = STATE(4016),
    [ts_builtin_sym_end] = ACTIONS(6969),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_namespace] = ACTIONS(6971),
    [anon_sym_module] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_open] = ACTIONS(6971),
    [anon_sym_type] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_exception] = ACTIONS(6971),
    [anon_sym_extern] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6969),
    [anon_sym_POUNDload] = ACTIONS(6969),
    [anon_sym_POUNDtime] = ACTIONS(6969),
    [anon_sym_POUNDI] = ACTIONS(6969),
    [anon_sym_POUNDhelp] = ACTIONS(6969),
    [anon_sym_POUNDquit] = ACTIONS(6969),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(4017)] = {
    [sym_block_comment] = STATE(4017),
    [sym_compiler_directive_decl] = STATE(4017),
    [sym_fsi_directive_decl] = STATE(4017),
    [sym_preproc_line] = STATE(4017),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_module] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_open] = ACTIONS(6790),
    [anon_sym_type] = ACTIONS(6790),
    [anon_sym_LBRACK_LT] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_exception] = ACTIONS(6790),
    [anon_sym_extern] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6792),
    [anon_sym_POUNDload] = ACTIONS(6792),
    [anon_sym_POUNDtime] = ACTIONS(6792),
    [anon_sym_POUNDI] = ACTIONS(6792),
    [anon_sym_POUNDhelp] = ACTIONS(6792),
    [anon_sym_POUNDquit] = ACTIONS(6792),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(4018)] = {
    [sym_block_comment] = STATE(4018),
    [sym_compiler_directive_decl] = STATE(4018),
    [sym_fsi_directive_decl] = STATE(4018),
    [sym_preproc_line] = STATE(4018),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_module] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_open] = ACTIONS(6794),
    [anon_sym_type] = ACTIONS(6794),
    [anon_sym_LBRACK_LT] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_exception] = ACTIONS(6794),
    [anon_sym_extern] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6796),
    [anon_sym_POUNDload] = ACTIONS(6796),
    [anon_sym_POUNDtime] = ACTIONS(6796),
    [anon_sym_POUNDI] = ACTIONS(6796),
    [anon_sym_POUNDhelp] = ACTIONS(6796),
    [anon_sym_POUNDquit] = ACTIONS(6796),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(4019)] = {
    [sym_type_arguments] = STATE(4660),
    [sym_long_identifier] = STATE(4641),
    [sym_block_comment] = STATE(4019),
    [sym_compiler_directive_decl] = STATE(4019),
    [sym_fsi_directive_decl] = STATE(4019),
    [sym_preproc_line] = STATE(4019),
    [aux_sym_compound_type_repeat1] = STATE(4506),
    [sym_identifier] = ACTIONS(8502),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_GT] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8430),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8432),
    [anon_sym_LT] = ACTIONS(8434),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8436),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(4020)] = {
    [sym_block_comment] = STATE(4020),
    [sym_compiler_directive_decl] = STATE(4020),
    [sym_fsi_directive_decl] = STATE(4020),
    [sym_preproc_line] = STATE(4020),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_module] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_open] = ACTIONS(6692),
    [anon_sym_type] = ACTIONS(6692),
    [anon_sym_LBRACK_LT] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_exception] = ACTIONS(6692),
    [anon_sym_extern] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6694),
    [anon_sym_POUNDload] = ACTIONS(6694),
    [anon_sym_POUNDtime] = ACTIONS(6694),
    [anon_sym_POUNDI] = ACTIONS(6694),
    [anon_sym_POUNDhelp] = ACTIONS(6694),
    [anon_sym_POUNDquit] = ACTIONS(6694),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(4021)] = {
    [sym_block_comment] = STATE(4021),
    [sym_compiler_directive_decl] = STATE(4021),
    [sym_fsi_directive_decl] = STATE(4021),
    [sym_preproc_line] = STATE(4021),
    [ts_builtin_sym_end] = ACTIONS(7131),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_namespace] = ACTIONS(7129),
    [anon_sym_module] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_open] = ACTIONS(7129),
    [anon_sym_type] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_exception] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7131),
    [anon_sym_POUNDload] = ACTIONS(7131),
    [anon_sym_POUNDtime] = ACTIONS(7131),
    [anon_sym_POUNDI] = ACTIONS(7131),
    [anon_sym_POUNDhelp] = ACTIONS(7131),
    [anon_sym_POUNDquit] = ACTIONS(7131),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(4022)] = {
    [sym_block_comment] = STATE(4022),
    [sym_compiler_directive_decl] = STATE(4022),
    [sym_fsi_directive_decl] = STATE(4022),
    [sym_preproc_line] = STATE(4022),
    [ts_builtin_sym_end] = ACTIONS(7057),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_namespace] = ACTIONS(7059),
    [anon_sym_module] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_open] = ACTIONS(7059),
    [anon_sym_type] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_exception] = ACTIONS(7059),
    [anon_sym_extern] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7057),
    [anon_sym_POUNDload] = ACTIONS(7057),
    [anon_sym_POUNDtime] = ACTIONS(7057),
    [anon_sym_POUNDI] = ACTIONS(7057),
    [anon_sym_POUNDhelp] = ACTIONS(7057),
    [anon_sym_POUNDquit] = ACTIONS(7057),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(4023)] = {
    [sym_block_comment] = STATE(4023),
    [sym_compiler_directive_decl] = STATE(4023),
    [sym_fsi_directive_decl] = STATE(4023),
    [sym_preproc_line] = STATE(4023),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_module] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_open] = ACTIONS(7011),
    [anon_sym_type] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_exception] = ACTIONS(7011),
    [anon_sym_extern] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7009),
    [anon_sym_POUNDload] = ACTIONS(7009),
    [anon_sym_POUNDtime] = ACTIONS(7009),
    [anon_sym_POUNDI] = ACTIONS(7009),
    [anon_sym_POUNDhelp] = ACTIONS(7009),
    [anon_sym_POUNDquit] = ACTIONS(7009),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [anon_sym_POUNDelse] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(4024)] = {
    [sym_type_arguments] = STATE(4709),
    [sym_long_identifier] = STATE(4716),
    [sym_block_comment] = STATE(4024),
    [sym_compiler_directive_decl] = STATE(4024),
    [sym_fsi_directive_decl] = STATE(4024),
    [sym_preproc_line] = STATE(4024),
    [aux_sym_compound_type_repeat1] = STATE(4431),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_with] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(8492),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(8494),
    [anon_sym_LT] = ACTIONS(8496),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8498),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(4025)] = {
    [sym_block_comment] = STATE(4025),
    [sym_compiler_directive_decl] = STATE(4025),
    [sym_fsi_directive_decl] = STATE(4025),
    [sym_preproc_line] = STATE(4025),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_module] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_open] = ACTIONS(6948),
    [anon_sym_type] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_exception] = ACTIONS(6948),
    [anon_sym_extern] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6946),
    [anon_sym_POUNDload] = ACTIONS(6946),
    [anon_sym_POUNDtime] = ACTIONS(6946),
    [anon_sym_POUNDI] = ACTIONS(6946),
    [anon_sym_POUNDhelp] = ACTIONS(6946),
    [anon_sym_POUNDquit] = ACTIONS(6946),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [anon_sym_POUNDelse] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(4026)] = {
    [sym_block_comment] = STATE(4026),
    [sym_compiler_directive_decl] = STATE(4026),
    [sym_fsi_directive_decl] = STATE(4026),
    [sym_preproc_line] = STATE(4026),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4027)] = {
    [sym_type_arguments] = STATE(4585),
    [sym_long_identifier] = STATE(4742),
    [sym_block_comment] = STATE(4027),
    [sym_compiler_directive_decl] = STATE(4027),
    [sym_fsi_directive_decl] = STATE(4027),
    [sym_preproc_line] = STATE(4027),
    [aux_sym_compound_type_repeat1] = STATE(4343),
    [sym_identifier] = ACTIONS(8371),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8373),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8375),
    [anon_sym_LT] = ACTIONS(8377),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8379),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(4028)] = {
    [sym_block_comment] = STATE(4028),
    [sym_compiler_directive_decl] = STATE(4028),
    [sym_fsi_directive_decl] = STATE(4028),
    [sym_preproc_line] = STATE(4028),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_as] = ACTIONS(6367),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_DOT_DOT] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__dedent] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4029)] = {
    [sym_block_comment] = STATE(4029),
    [sym_compiler_directive_decl] = STATE(4029),
    [sym_fsi_directive_decl] = STATE(4029),
    [sym_preproc_line] = STATE(4029),
    [ts_builtin_sym_end] = ACTIONS(7075),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_namespace] = ACTIONS(7073),
    [anon_sym_module] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_open] = ACTIONS(7073),
    [anon_sym_type] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_exception] = ACTIONS(7073),
    [anon_sym_extern] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7075),
    [anon_sym_POUNDload] = ACTIONS(7075),
    [anon_sym_POUNDtime] = ACTIONS(7075),
    [anon_sym_POUNDI] = ACTIONS(7075),
    [anon_sym_POUNDhelp] = ACTIONS(7075),
    [anon_sym_POUNDquit] = ACTIONS(7075),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(4030)] = {
    [sym_block_comment] = STATE(4030),
    [sym_compiler_directive_decl] = STATE(4030),
    [sym_fsi_directive_decl] = STATE(4030),
    [sym_preproc_line] = STATE(4030),
    [ts_builtin_sym_end] = ACTIONS(292),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_namespace] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(4031)] = {
    [sym_block_comment] = STATE(4031),
    [sym_compiler_directive_decl] = STATE(4031),
    [sym_fsi_directive_decl] = STATE(4031),
    [sym_preproc_line] = STATE(4031),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_module] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_open] = ACTIONS(7113),
    [anon_sym_type] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_exception] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7115),
    [anon_sym_POUNDload] = ACTIONS(7115),
    [anon_sym_POUNDtime] = ACTIONS(7115),
    [anon_sym_POUNDI] = ACTIONS(7115),
    [anon_sym_POUNDhelp] = ACTIONS(7115),
    [anon_sym_POUNDquit] = ACTIONS(7115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [anon_sym_POUNDelse] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(4032)] = {
    [sym_block_comment] = STATE(4032),
    [sym_compiler_directive_decl] = STATE(4032),
    [sym_fsi_directive_decl] = STATE(4032),
    [sym_preproc_line] = STATE(4032),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8504),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4033)] = {
    [sym_type_arguments] = STATE(4618),
    [sym_long_identifier] = STATE(4600),
    [sym_block_comment] = STATE(4033),
    [sym_compiler_directive_decl] = STATE(4033),
    [sym_fsi_directive_decl] = STATE(4033),
    [sym_preproc_line] = STATE(4033),
    [aux_sym_compound_type_repeat1] = STATE(4303),
    [sym_identifier] = ACTIONS(8506),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8408),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8410),
    [anon_sym_LT] = ACTIONS(8412),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8414),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__untyped_quoted_close] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(4034)] = {
    [sym_block_comment] = STATE(4034),
    [sym_compiler_directive_decl] = STATE(4034),
    [sym_fsi_directive_decl] = STATE(4034),
    [sym_preproc_line] = STATE(4034),
    [ts_builtin_sym_end] = ACTIONS(7025),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_namespace] = ACTIONS(7027),
    [anon_sym_module] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_open] = ACTIONS(7027),
    [anon_sym_type] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_exception] = ACTIONS(7027),
    [anon_sym_extern] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7025),
    [anon_sym_POUNDload] = ACTIONS(7025),
    [anon_sym_POUNDtime] = ACTIONS(7025),
    [anon_sym_POUNDI] = ACTIONS(7025),
    [anon_sym_POUNDhelp] = ACTIONS(7025),
    [anon_sym_POUNDquit] = ACTIONS(7025),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(4035)] = {
    [sym_block_comment] = STATE(4035),
    [sym_compiler_directive_decl] = STATE(4035),
    [sym_fsi_directive_decl] = STATE(4035),
    [sym_preproc_line] = STATE(4035),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_module] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_open] = ACTIONS(6892),
    [anon_sym_type] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_exception] = ACTIONS(6892),
    [anon_sym_extern] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6890),
    [anon_sym_POUNDload] = ACTIONS(6890),
    [anon_sym_POUNDtime] = ACTIONS(6890),
    [anon_sym_POUNDI] = ACTIONS(6890),
    [anon_sym_POUNDhelp] = ACTIONS(6890),
    [anon_sym_POUNDquit] = ACTIONS(6890),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [anon_sym_POUNDelse] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(4036)] = {
    [sym_block_comment] = STATE(4036),
    [sym_compiler_directive_decl] = STATE(4036),
    [sym_fsi_directive_decl] = STATE(4036),
    [sym_preproc_line] = STATE(4036),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_module] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_open] = ACTIONS(6952),
    [anon_sym_type] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_exception] = ACTIONS(6952),
    [anon_sym_extern] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6950),
    [anon_sym_POUNDload] = ACTIONS(6950),
    [anon_sym_POUNDtime] = ACTIONS(6950),
    [anon_sym_POUNDI] = ACTIONS(6950),
    [anon_sym_POUNDhelp] = ACTIONS(6950),
    [anon_sym_POUNDquit] = ACTIONS(6950),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [anon_sym_POUNDelse] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(4037)] = {
    [sym_block_comment] = STATE(4037),
    [sym_compiler_directive_decl] = STATE(4037),
    [sym_fsi_directive_decl] = STATE(4037),
    [sym_preproc_line] = STATE(4037),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_GT_RBRACK] = ACTIONS(6562),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_RBRACK] = ACTIONS(6562),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_RBRACE] = ACTIONS(6562),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_to] = ACTIONS(6564),
    [anon_sym_downto] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_DOT_DOT2] = ACTIONS(6562),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4038)] = {
    [sym_block_comment] = STATE(4038),
    [sym_compiler_directive_decl] = STATE(4038),
    [sym_fsi_directive_decl] = STATE(4038),
    [sym_preproc_line] = STATE(4038),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8330),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4039)] = {
    [sym_block_comment] = STATE(4039),
    [sym_compiler_directive_decl] = STATE(4039),
    [sym_fsi_directive_decl] = STATE(4039),
    [sym_preproc_line] = STATE(4039),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_module] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_open] = ACTIONS(7137),
    [anon_sym_type] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_exception] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7139),
    [anon_sym_POUNDload] = ACTIONS(7139),
    [anon_sym_POUNDtime] = ACTIONS(7139),
    [anon_sym_POUNDI] = ACTIONS(7139),
    [anon_sym_POUNDhelp] = ACTIONS(7139),
    [anon_sym_POUNDquit] = ACTIONS(7139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [anon_sym_POUNDelse] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(4040)] = {
    [sym_block_comment] = STATE(4040),
    [sym_compiler_directive_decl] = STATE(4040),
    [sym_fsi_directive_decl] = STATE(4040),
    [sym_preproc_line] = STATE(4040),
    [ts_builtin_sym_end] = ACTIONS(7033),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_namespace] = ACTIONS(7035),
    [anon_sym_module] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_open] = ACTIONS(7035),
    [anon_sym_type] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_exception] = ACTIONS(7035),
    [anon_sym_extern] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7033),
    [anon_sym_POUNDload] = ACTIONS(7033),
    [anon_sym_POUNDtime] = ACTIONS(7033),
    [anon_sym_POUNDI] = ACTIONS(7033),
    [anon_sym_POUNDhelp] = ACTIONS(7033),
    [anon_sym_POUNDquit] = ACTIONS(7033),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(4041)] = {
    [sym_type_arguments] = STATE(4709),
    [sym_long_identifier] = STATE(4716),
    [sym_block_comment] = STATE(4041),
    [sym_compiler_directive_decl] = STATE(4041),
    [sym_fsi_directive_decl] = STATE(4041),
    [sym_preproc_line] = STATE(4041),
    [aux_sym_compound_type_repeat1] = STATE(4431),
    [sym_identifier] = ACTIONS(8490),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8492),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8494),
    [anon_sym_LT] = ACTIONS(8496),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8498),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4042)] = {
    [sym_block_comment] = STATE(4042),
    [sym_compiler_directive_decl] = STATE(4042),
    [sym_fsi_directive_decl] = STATE(4042),
    [sym_preproc_line] = STATE(4042),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_module] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_open] = ACTIONS(6867),
    [anon_sym_type] = ACTIONS(6867),
    [anon_sym_LBRACK_LT] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_exception] = ACTIONS(6867),
    [anon_sym_extern] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6865),
    [anon_sym_POUNDload] = ACTIONS(6865),
    [anon_sym_POUNDtime] = ACTIONS(6865),
    [anon_sym_POUNDI] = ACTIONS(6865),
    [anon_sym_POUNDhelp] = ACTIONS(6865),
    [anon_sym_POUNDquit] = ACTIONS(6865),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(4043)] = {
    [sym_type_arguments] = STATE(4553),
    [sym_long_identifier] = STATE(4568),
    [sym_block_comment] = STATE(4043),
    [sym_compiler_directive_decl] = STATE(4043),
    [sym_fsi_directive_decl] = STATE(4043),
    [sym_preproc_line] = STATE(4043),
    [aux_sym_compound_type_repeat1] = STATE(4443),
    [sym_identifier] = ACTIONS(8356),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8358),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8360),
    [anon_sym_LT] = ACTIONS(8362),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8364),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__quoted_close] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(4044)] = {
    [sym_block_comment] = STATE(4044),
    [sym_compiler_directive_decl] = STATE(4044),
    [sym_fsi_directive_decl] = STATE(4044),
    [sym_preproc_line] = STATE(4044),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_module] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_open] = ACTIONS(6932),
    [anon_sym_type] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_exception] = ACTIONS(6932),
    [anon_sym_extern] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6930),
    [anon_sym_POUNDload] = ACTIONS(6930),
    [anon_sym_POUNDtime] = ACTIONS(6930),
    [anon_sym_POUNDI] = ACTIONS(6930),
    [anon_sym_POUNDhelp] = ACTIONS(6930),
    [anon_sym_POUNDquit] = ACTIONS(6930),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [anon_sym_POUNDelse] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(4045)] = {
    [sym_block_comment] = STATE(4045),
    [sym_compiler_directive_decl] = STATE(4045),
    [sym_fsi_directive_decl] = STATE(4045),
    [sym_preproc_line] = STATE(4045),
    [ts_builtin_sym_end] = ACTIONS(7091),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_namespace] = ACTIONS(7089),
    [anon_sym_module] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_open] = ACTIONS(7089),
    [anon_sym_type] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_exception] = ACTIONS(7089),
    [anon_sym_extern] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7091),
    [anon_sym_POUNDload] = ACTIONS(7091),
    [anon_sym_POUNDtime] = ACTIONS(7091),
    [anon_sym_POUNDI] = ACTIONS(7091),
    [anon_sym_POUNDhelp] = ACTIONS(7091),
    [anon_sym_POUNDquit] = ACTIONS(7091),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(4046)] = {
    [sym_attributes] = STATE(13414),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4542),
    [sym_type_extension_elements] = STATE(8353),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4046),
    [sym_compiler_directive_decl] = STATE(4046),
    [sym_fsi_directive_decl] = STATE(4046),
    [sym_preproc_line] = STATE(4046),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8343),
    [sym_identifier] = ACTIONS(8341),
    [anon_sym_namespace] = ACTIONS(8341),
    [anon_sym_module] = ACTIONS(8341),
    [anon_sym_null] = ACTIONS(8341),
    [anon_sym_begin] = ACTIONS(8341),
    [anon_sym_open] = ACTIONS(8341),
    [anon_sym_type] = ACTIONS(8341),
    [anon_sym_LBRACK_LT] = ACTIONS(8343),
    [anon_sym_do] = ACTIONS(8341),
    [anon_sym_and] = ACTIONS(8341),
    [anon_sym_let] = ACTIONS(8341),
    [anon_sym_let_BANG] = ACTIONS(8343),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8341),
    [anon_sym_AMP] = ACTIONS(8341),
    [anon_sym_LBRACK] = ACTIONS(8341),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8343),
    [anon_sym_LBRACE] = ACTIONS(8341),
    [anon_sym_LT_AT] = ACTIONS(8341),
    [anon_sym_LT_AT_AT] = ACTIONS(8343),
    [anon_sym_CARET] = ACTIONS(8343),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8343),
    [anon_sym_new] = ACTIONS(8341),
    [anon_sym_with] = ACTIONS(8508),
    [anon_sym_return] = ACTIONS(8341),
    [anon_sym_return_BANG] = ACTIONS(8343),
    [anon_sym_yield] = ACTIONS(8341),
    [anon_sym_yield_BANG] = ACTIONS(8343),
    [anon_sym_lazy] = ACTIONS(8341),
    [anon_sym_assert] = ACTIONS(8341),
    [anon_sym_upcast] = ACTIONS(8341),
    [anon_sym_downcast] = ACTIONS(8341),
    [anon_sym_fixed] = ACTIONS(8341),
    [anon_sym_for] = ACTIONS(8341),
    [anon_sym_while] = ACTIONS(8341),
    [anon_sym_while_BANG] = ACTIONS(8343),
    [anon_sym_if] = ACTIONS(8341),
    [anon_sym_fun] = ACTIONS(8341),
    [anon_sym_try] = ACTIONS(8341),
    [anon_sym_match] = ACTIONS(8341),
    [anon_sym_match_BANG] = ACTIONS(8343),
    [anon_sym_function] = ACTIONS(8341),
    [anon_sym_use] = ACTIONS(8341),
    [anon_sym_use_BANG] = ACTIONS(8343),
    [anon_sym_do_BANG] = ACTIONS(8343),
    [anon_sym_AMP_AMP] = ACTIONS(8343),
    [anon_sym_default] = ACTIONS(8510),
    [anon_sym_static] = ACTIONS(8512),
    [anon_sym_member] = ACTIONS(8514),
    [anon_sym_exception] = ACTIONS(8341),
    [anon_sym_interface] = ACTIONS(8516),
    [anon_sym_abstract] = ACTIONS(8518),
    [anon_sym_val] = ACTIONS(8520),
    [anon_sym_override] = ACTIONS(8510),
    [anon_sym_extern] = ACTIONS(8341),
    [aux_sym_char_token1] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8341),
    [anon_sym_DQUOTE] = ACTIONS(8341),
    [anon_sym_AT_DQUOTE] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [sym_bool] = ACTIONS(8341),
    [sym_unit] = ACTIONS(8343),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8341),
    [sym_op_identifier] = ACTIONS(8343),
    [anon_sym_PLUS] = ACTIONS(8341),
    [anon_sym_DASH] = ACTIONS(8341),
    [anon_sym_PLUS_DOT] = ACTIONS(8343),
    [anon_sym_DASH_DOT] = ACTIONS(8343),
    [anon_sym_PERCENT] = ACTIONS(8341),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8343),
    [anon_sym_TILDE] = ACTIONS(8343),
    [aux_sym_prefix_op_token1] = ACTIONS(8343),
    [sym_int] = ACTIONS(8341),
    [sym_xint] = ACTIONS(8343),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8343),
    [anon_sym_POUNDload] = ACTIONS(8343),
    [anon_sym_POUNDtime] = ACTIONS(8343),
    [anon_sym_POUNDI] = ACTIONS(8343),
    [anon_sym_POUNDhelp] = ACTIONS(8343),
    [anon_sym_POUNDquit] = ACTIONS(8343),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8343),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8343),
  },
  [STATE(4047)] = {
    [sym_block_comment] = STATE(4047),
    [sym_compiler_directive_decl] = STATE(4047),
    [sym_fsi_directive_decl] = STATE(4047),
    [sym_preproc_line] = STATE(4047),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_module] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_open] = ACTIONS(6896),
    [anon_sym_type] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_exception] = ACTIONS(6896),
    [anon_sym_extern] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6894),
    [anon_sym_POUNDload] = ACTIONS(6894),
    [anon_sym_POUNDtime] = ACTIONS(6894),
    [anon_sym_POUNDI] = ACTIONS(6894),
    [anon_sym_POUNDhelp] = ACTIONS(6894),
    [anon_sym_POUNDquit] = ACTIONS(6894),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [anon_sym_POUNDelse] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(4048)] = {
    [sym_attributes] = STATE(13414),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4542),
    [sym_type_extension_elements] = STATE(8362),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4048),
    [sym_compiler_directive_decl] = STATE(4048),
    [sym_fsi_directive_decl] = STATE(4048),
    [sym_preproc_line] = STATE(4048),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8244),
    [sym_identifier] = ACTIONS(8242),
    [anon_sym_namespace] = ACTIONS(8242),
    [anon_sym_module] = ACTIONS(8242),
    [anon_sym_null] = ACTIONS(8242),
    [anon_sym_begin] = ACTIONS(8242),
    [anon_sym_open] = ACTIONS(8242),
    [anon_sym_type] = ACTIONS(8242),
    [anon_sym_LBRACK_LT] = ACTIONS(8244),
    [anon_sym_do] = ACTIONS(8242),
    [anon_sym_and] = ACTIONS(8242),
    [anon_sym_let] = ACTIONS(8242),
    [anon_sym_let_BANG] = ACTIONS(8244),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8242),
    [anon_sym_AMP] = ACTIONS(8242),
    [anon_sym_LBRACK] = ACTIONS(8242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8244),
    [anon_sym_LBRACE] = ACTIONS(8242),
    [anon_sym_LT_AT] = ACTIONS(8242),
    [anon_sym_LT_AT_AT] = ACTIONS(8244),
    [anon_sym_CARET] = ACTIONS(8244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8244),
    [anon_sym_new] = ACTIONS(8242),
    [anon_sym_with] = ACTIONS(8508),
    [anon_sym_return] = ACTIONS(8242),
    [anon_sym_return_BANG] = ACTIONS(8244),
    [anon_sym_yield] = ACTIONS(8242),
    [anon_sym_yield_BANG] = ACTIONS(8244),
    [anon_sym_lazy] = ACTIONS(8242),
    [anon_sym_assert] = ACTIONS(8242),
    [anon_sym_upcast] = ACTIONS(8242),
    [anon_sym_downcast] = ACTIONS(8242),
    [anon_sym_fixed] = ACTIONS(8242),
    [anon_sym_for] = ACTIONS(8242),
    [anon_sym_while] = ACTIONS(8242),
    [anon_sym_while_BANG] = ACTIONS(8244),
    [anon_sym_if] = ACTIONS(8242),
    [anon_sym_fun] = ACTIONS(8242),
    [anon_sym_try] = ACTIONS(8242),
    [anon_sym_match] = ACTIONS(8242),
    [anon_sym_match_BANG] = ACTIONS(8244),
    [anon_sym_function] = ACTIONS(8242),
    [anon_sym_use] = ACTIONS(8242),
    [anon_sym_use_BANG] = ACTIONS(8244),
    [anon_sym_do_BANG] = ACTIONS(8244),
    [anon_sym_AMP_AMP] = ACTIONS(8244),
    [anon_sym_default] = ACTIONS(8510),
    [anon_sym_static] = ACTIONS(8512),
    [anon_sym_member] = ACTIONS(8514),
    [anon_sym_exception] = ACTIONS(8242),
    [anon_sym_interface] = ACTIONS(8516),
    [anon_sym_abstract] = ACTIONS(8518),
    [anon_sym_val] = ACTIONS(8520),
    [anon_sym_override] = ACTIONS(8510),
    [anon_sym_extern] = ACTIONS(8242),
    [aux_sym_char_token1] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8242),
    [anon_sym_DQUOTE] = ACTIONS(8242),
    [anon_sym_AT_DQUOTE] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [sym_bool] = ACTIONS(8242),
    [sym_unit] = ACTIONS(8244),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8242),
    [sym_op_identifier] = ACTIONS(8244),
    [anon_sym_PLUS] = ACTIONS(8242),
    [anon_sym_DASH] = ACTIONS(8242),
    [anon_sym_PLUS_DOT] = ACTIONS(8244),
    [anon_sym_DASH_DOT] = ACTIONS(8244),
    [anon_sym_PERCENT] = ACTIONS(8242),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8244),
    [anon_sym_TILDE] = ACTIONS(8244),
    [aux_sym_prefix_op_token1] = ACTIONS(8244),
    [sym_int] = ACTIONS(8242),
    [sym_xint] = ACTIONS(8244),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8244),
    [anon_sym_POUNDload] = ACTIONS(8244),
    [anon_sym_POUNDtime] = ACTIONS(8244),
    [anon_sym_POUNDI] = ACTIONS(8244),
    [anon_sym_POUNDhelp] = ACTIONS(8244),
    [anon_sym_POUNDquit] = ACTIONS(8244),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8244),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8244),
  },
  [STATE(4049)] = {
    [sym_block_comment] = STATE(4049),
    [sym_compiler_directive_decl] = STATE(4049),
    [sym_fsi_directive_decl] = STATE(4049),
    [sym_preproc_line] = STATE(4049),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_module] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_open] = ACTIONS(6597),
    [anon_sym_type] = ACTIONS(6597),
    [anon_sym_LBRACK_LT] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8284),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_exception] = ACTIONS(6597),
    [anon_sym_extern] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6599),
    [anon_sym_POUNDload] = ACTIONS(6599),
    [anon_sym_POUNDtime] = ACTIONS(6599),
    [anon_sym_POUNDI] = ACTIONS(6599),
    [anon_sym_POUNDhelp] = ACTIONS(6599),
    [anon_sym_POUNDquit] = ACTIONS(6599),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4050)] = {
    [sym_block_comment] = STATE(4050),
    [sym_compiler_directive_decl] = STATE(4050),
    [sym_fsi_directive_decl] = STATE(4050),
    [sym_preproc_line] = STATE(4050),
    [ts_builtin_sym_end] = ACTIONS(7135),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_namespace] = ACTIONS(7133),
    [anon_sym_module] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_open] = ACTIONS(7133),
    [anon_sym_type] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_exception] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7135),
    [anon_sym_POUNDload] = ACTIONS(7135),
    [anon_sym_POUNDtime] = ACTIONS(7135),
    [anon_sym_POUNDI] = ACTIONS(7135),
    [anon_sym_POUNDhelp] = ACTIONS(7135),
    [anon_sym_POUNDquit] = ACTIONS(7135),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(4051)] = {
    [sym_block_comment] = STATE(4051),
    [sym_compiler_directive_decl] = STATE(4051),
    [sym_fsi_directive_decl] = STATE(4051),
    [sym_preproc_line] = STATE(4051),
    [ts_builtin_sym_end] = ACTIONS(6930),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_namespace] = ACTIONS(6932),
    [anon_sym_module] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_open] = ACTIONS(6932),
    [anon_sym_type] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_exception] = ACTIONS(6932),
    [anon_sym_extern] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6930),
    [anon_sym_POUNDload] = ACTIONS(6930),
    [anon_sym_POUNDtime] = ACTIONS(6930),
    [anon_sym_POUNDI] = ACTIONS(6930),
    [anon_sym_POUNDhelp] = ACTIONS(6930),
    [anon_sym_POUNDquit] = ACTIONS(6930),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(4052)] = {
    [sym_block_comment] = STATE(4052),
    [sym_compiler_directive_decl] = STATE(4052),
    [sym_fsi_directive_decl] = STATE(4052),
    [sym_preproc_line] = STATE(4052),
    [ts_builtin_sym_end] = ACTIONS(7049),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_namespace] = ACTIONS(7051),
    [anon_sym_module] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_open] = ACTIONS(7051),
    [anon_sym_type] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_exception] = ACTIONS(7051),
    [anon_sym_extern] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7049),
    [anon_sym_POUNDload] = ACTIONS(7049),
    [anon_sym_POUNDtime] = ACTIONS(7049),
    [anon_sym_POUNDI] = ACTIONS(7049),
    [anon_sym_POUNDhelp] = ACTIONS(7049),
    [anon_sym_POUNDquit] = ACTIONS(7049),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(4053)] = {
    [sym_block_comment] = STATE(4053),
    [sym_compiler_directive_decl] = STATE(4053),
    [sym_fsi_directive_decl] = STATE(4053),
    [sym_preproc_line] = STATE(4053),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_as] = ACTIONS(6379),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_DOT_DOT] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__dedent] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4054)] = {
    [sym_block_comment] = STATE(4054),
    [sym_compiler_directive_decl] = STATE(4054),
    [sym_fsi_directive_decl] = STATE(4054),
    [sym_preproc_line] = STATE(4054),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_as] = ACTIONS(6383),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_DOT_DOT] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__dedent] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4055)] = {
    [sym_block_comment] = STATE(4055),
    [sym_compiler_directive_decl] = STATE(4055),
    [sym_fsi_directive_decl] = STATE(4055),
    [sym_preproc_line] = STATE(4055),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_GT_RBRACK] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_RBRACK] = ACTIONS(6614),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_RBRACE] = ACTIONS(6614),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_to] = ACTIONS(6612),
    [anon_sym_downto] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4056)] = {
    [sym_block_comment] = STATE(4056),
    [sym_compiler_directive_decl] = STATE(4056),
    [sym_fsi_directive_decl] = STATE(4056),
    [sym_preproc_line] = STATE(4056),
    [ts_builtin_sym_end] = ACTIONS(7055),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_namespace] = ACTIONS(7053),
    [anon_sym_module] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_open] = ACTIONS(7053),
    [anon_sym_type] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_exception] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7055),
    [anon_sym_POUNDload] = ACTIONS(7055),
    [anon_sym_POUNDtime] = ACTIONS(7055),
    [anon_sym_POUNDI] = ACTIONS(7055),
    [anon_sym_POUNDhelp] = ACTIONS(7055),
    [anon_sym_POUNDquit] = ACTIONS(7055),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(4057)] = {
    [sym_block_comment] = STATE(4057),
    [sym_compiler_directive_decl] = STATE(4057),
    [sym_fsi_directive_decl] = STATE(4057),
    [sym_preproc_line] = STATE(4057),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8330),
    [anon_sym_POUNDelse] = ACTIONS(8330),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4058)] = {
    [sym_block_comment] = STATE(4058),
    [sym_compiler_directive_decl] = STATE(4058),
    [sym_fsi_directive_decl] = STATE(4058),
    [sym_preproc_line] = STATE(4058),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8332),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4059)] = {
    [sym_block_comment] = STATE(4059),
    [sym_compiler_directive_decl] = STATE(4059),
    [sym_fsi_directive_decl] = STATE(4059),
    [sym_preproc_line] = STATE(4059),
    [ts_builtin_sym_end] = ACTIONS(5953),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_namespace] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4060)] = {
    [sym_type_arguments] = STATE(4660),
    [sym_long_identifier] = STATE(4641),
    [sym_block_comment] = STATE(4060),
    [sym_compiler_directive_decl] = STATE(4060),
    [sym_fsi_directive_decl] = STATE(4060),
    [sym_preproc_line] = STATE(4060),
    [aux_sym_compound_type_repeat1] = STATE(4506),
    [sym_identifier] = ACTIONS(8502),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_GT] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8430),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8432),
    [anon_sym_LT] = ACTIONS(8434),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8436),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4061)] = {
    [sym_block_comment] = STATE(4061),
    [sym_compiler_directive_decl] = STATE(4061),
    [sym_fsi_directive_decl] = STATE(4061),
    [sym_preproc_line] = STATE(4061),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_module] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_open] = ACTIONS(7073),
    [anon_sym_type] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_exception] = ACTIONS(7073),
    [anon_sym_extern] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7075),
    [anon_sym_POUNDload] = ACTIONS(7075),
    [anon_sym_POUNDtime] = ACTIONS(7075),
    [anon_sym_POUNDI] = ACTIONS(7075),
    [anon_sym_POUNDhelp] = ACTIONS(7075),
    [anon_sym_POUNDquit] = ACTIONS(7075),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [anon_sym_POUNDelse] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(4062)] = {
    [sym_block_comment] = STATE(4062),
    [sym_compiler_directive_decl] = STATE(4062),
    [sym_fsi_directive_decl] = STATE(4062),
    [sym_preproc_line] = STATE(4062),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_module] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_open] = ACTIONS(6834),
    [anon_sym_type] = ACTIONS(6834),
    [anon_sym_LBRACK_LT] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_exception] = ACTIONS(6834),
    [anon_sym_extern] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6832),
    [anon_sym_POUNDload] = ACTIONS(6832),
    [anon_sym_POUNDtime] = ACTIONS(6832),
    [anon_sym_POUNDI] = ACTIONS(6832),
    [anon_sym_POUNDhelp] = ACTIONS(6832),
    [anon_sym_POUNDquit] = ACTIONS(6832),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4063)] = {
    [sym_block_comment] = STATE(4063),
    [sym_compiler_directive_decl] = STATE(4063),
    [sym_fsi_directive_decl] = STATE(4063),
    [sym_preproc_line] = STATE(4063),
    [ts_builtin_sym_end] = ACTIONS(6915),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_namespace] = ACTIONS(6917),
    [anon_sym_module] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_open] = ACTIONS(6917),
    [anon_sym_type] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_exception] = ACTIONS(6917),
    [anon_sym_extern] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6915),
    [anon_sym_POUNDload] = ACTIONS(6915),
    [anon_sym_POUNDtime] = ACTIONS(6915),
    [anon_sym_POUNDI] = ACTIONS(6915),
    [anon_sym_POUNDhelp] = ACTIONS(6915),
    [anon_sym_POUNDquit] = ACTIONS(6915),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(4064)] = {
    [sym_block_comment] = STATE(4064),
    [sym_compiler_directive_decl] = STATE(4064),
    [sym_fsi_directive_decl] = STATE(4064),
    [sym_preproc_line] = STATE(4064),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_GT_RBRACK] = ACTIONS(6556),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_RBRACK] = ACTIONS(6556),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_RBRACE] = ACTIONS(6556),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_to] = ACTIONS(6554),
    [anon_sym_downto] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_DOT_DOT2] = ACTIONS(6556),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4065)] = {
    [sym_block_comment] = STATE(4065),
    [sym_compiler_directive_decl] = STATE(4065),
    [sym_fsi_directive_decl] = STATE(4065),
    [sym_preproc_line] = STATE(4065),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_GT_RBRACK] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_RBRACK] = ACTIONS(6614),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_RBRACE] = ACTIONS(6614),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_to] = ACTIONS(6612),
    [anon_sym_downto] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6612),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4066)] = {
    [sym__else_expression] = STATE(5001),
    [sym_elif_expression] = STATE(4629),
    [sym_block_comment] = STATE(4066),
    [sym_compiler_directive_decl] = STATE(4066),
    [sym_fsi_directive_decl] = STATE(4066),
    [sym_preproc_line] = STATE(4066),
    [aux_sym_if_expression_repeat1] = STATE(4077),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8522),
    [anon_sym_elif] = ACTIONS(8524),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_DOT_DOT] = ACTIONS(6319),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [anon_sym_DOT_DOT2] = ACTIONS(6317),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__dedent] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4067)] = {
    [sym_type_arguments] = STATE(4618),
    [sym_long_identifier] = STATE(4600),
    [sym_block_comment] = STATE(4067),
    [sym_compiler_directive_decl] = STATE(4067),
    [sym_fsi_directive_decl] = STATE(4067),
    [sym_preproc_line] = STATE(4067),
    [aux_sym_compound_type_repeat1] = STATE(4303),
    [sym_identifier] = ACTIONS(8506),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8408),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8410),
    [anon_sym_LT] = ACTIONS(8412),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8414),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__untyped_quoted_close] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(4068)] = {
    [sym_type_arguments] = STATE(4618),
    [sym_long_identifier] = STATE(4600),
    [sym_block_comment] = STATE(4068),
    [sym_compiler_directive_decl] = STATE(4068),
    [sym_fsi_directive_decl] = STATE(4068),
    [sym_preproc_line] = STATE(4068),
    [aux_sym_compound_type_repeat1] = STATE(4303),
    [sym_identifier] = ACTIONS(8506),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(8408),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8410),
    [anon_sym_LT] = ACTIONS(8412),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8414),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__untyped_quoted_close] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4069)] = {
    [sym_type_arguments] = STATE(4611),
    [sym_long_identifier] = STATE(4669),
    [sym_block_comment] = STATE(4069),
    [sym_compiler_directive_decl] = STATE(4069),
    [sym_fsi_directive_decl] = STATE(4069),
    [sym_preproc_line] = STATE(4069),
    [aux_sym_compound_type_repeat1] = STATE(4301),
    [sym_identifier] = ACTIONS(8488),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8394),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_DOT_DOT] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(8398),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8400),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(4070)] = {
    [sym_type_arguments] = STATE(4618),
    [sym_long_identifier] = STATE(4600),
    [sym_block_comment] = STATE(4070),
    [sym_compiler_directive_decl] = STATE(4070),
    [sym_fsi_directive_decl] = STATE(4070),
    [sym_preproc_line] = STATE(4070),
    [aux_sym_compound_type_repeat1] = STATE(4303),
    [sym_identifier] = ACTIONS(8506),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8408),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8410),
    [anon_sym_LT] = ACTIONS(8412),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8414),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__untyped_quoted_close] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(4071)] = {
    [sym_block_comment] = STATE(4071),
    [sym_compiler_directive_decl] = STATE(4071),
    [sym_fsi_directive_decl] = STATE(4071),
    [sym_preproc_line] = STATE(4071),
    [sym_identifier] = ACTIONS(8526),
    [anon_sym_module] = ACTIONS(8529),
    [anon_sym_null] = ACTIONS(8526),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(8526),
    [anon_sym_open] = ACTIONS(8529),
    [anon_sym_type] = ACTIONS(8529),
    [anon_sym_LBRACK_LT] = ACTIONS(8531),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(8526),
    [anon_sym_let] = ACTIONS(8526),
    [anon_sym_let_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN] = ACTIONS(8526),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(8526),
    [anon_sym_LBRACK] = ACTIONS(8526),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8533),
    [anon_sym_LBRACE] = ACTIONS(8526),
    [anon_sym_LT_AT] = ACTIONS(8526),
    [anon_sym_LT_AT_AT] = ACTIONS(8526),
    [anon_sym_CARET] = ACTIONS(8526),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8533),
    [anon_sym_new] = ACTIONS(8526),
    [anon_sym_return] = ACTIONS(8526),
    [anon_sym_return_BANG] = ACTIONS(8533),
    [anon_sym_yield] = ACTIONS(8526),
    [anon_sym_yield_BANG] = ACTIONS(8533),
    [anon_sym_lazy] = ACTIONS(8526),
    [anon_sym_assert] = ACTIONS(8526),
    [anon_sym_upcast] = ACTIONS(8526),
    [anon_sym_downcast] = ACTIONS(8526),
    [anon_sym_fixed] = ACTIONS(8526),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(8526),
    [anon_sym_while] = ACTIONS(8526),
    [anon_sym_while_BANG] = ACTIONS(8533),
    [anon_sym_if] = ACTIONS(8526),
    [anon_sym_fun] = ACTIONS(8526),
    [anon_sym_try] = ACTIONS(8526),
    [anon_sym_match] = ACTIONS(8526),
    [anon_sym_match_BANG] = ACTIONS(8533),
    [anon_sym_function] = ACTIONS(8526),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(8526),
    [anon_sym_use_BANG] = ACTIONS(8533),
    [anon_sym_do_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(8526),
    [anon_sym_exception] = ACTIONS(8529),
    [anon_sym_extern] = ACTIONS(8529),
    [aux_sym_char_token1] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8526),
    [anon_sym_DQUOTE] = ACTIONS(8526),
    [anon_sym_AT_DQUOTE] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [sym_bool] = ACTIONS(8526),
    [sym_unit] = ACTIONS(8533),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8526),
    [sym_op_identifier] = ACTIONS(8526),
    [anon_sym_PLUS] = ACTIONS(8526),
    [anon_sym_DASH] = ACTIONS(8526),
    [anon_sym_PLUS_DOT] = ACTIONS(8526),
    [anon_sym_DASH_DOT] = ACTIONS(8526),
    [anon_sym_PERCENT] = ACTIONS(8526),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8526),
    [anon_sym_TILDE] = ACTIONS(8533),
    [aux_sym_prefix_op_token1] = ACTIONS(8526),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(8526),
    [sym_xint] = ACTIONS(8533),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8531),
    [anon_sym_POUNDload] = ACTIONS(8531),
    [anon_sym_POUNDtime] = ACTIONS(8531),
    [anon_sym_POUNDI] = ACTIONS(8531),
    [anon_sym_POUNDhelp] = ACTIONS(8531),
    [anon_sym_POUNDquit] = ACTIONS(8531),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8533),
    [anon_sym_POUNDendif] = ACTIONS(8531),
    [anon_sym_POUNDelse] = ACTIONS(8531),
    [sym__newline] = ACTIONS(8533),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8533),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4072)] = {
    [sym_type_arguments] = STATE(4660),
    [sym_long_identifier] = STATE(4641),
    [sym_block_comment] = STATE(4072),
    [sym_compiler_directive_decl] = STATE(4072),
    [sym_fsi_directive_decl] = STATE(4072),
    [sym_preproc_line] = STATE(4072),
    [aux_sym_compound_type_repeat1] = STATE(4506),
    [sym_identifier] = ACTIONS(8502),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_GT] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8430),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8432),
    [anon_sym_LT] = ACTIONS(8434),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8436),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(4073)] = {
    [sym_block_comment] = STATE(4073),
    [sym_compiler_directive_decl] = STATE(4073),
    [sym_fsi_directive_decl] = STATE(4073),
    [sym_preproc_line] = STATE(4073),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_module] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_open] = ACTIONS(6959),
    [anon_sym_type] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_exception] = ACTIONS(6959),
    [anon_sym_extern] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6957),
    [anon_sym_POUNDload] = ACTIONS(6957),
    [anon_sym_POUNDtime] = ACTIONS(6957),
    [anon_sym_POUNDI] = ACTIONS(6957),
    [anon_sym_POUNDhelp] = ACTIONS(6957),
    [anon_sym_POUNDquit] = ACTIONS(6957),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [anon_sym_POUNDelse] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4074)] = {
    [sym_block_comment] = STATE(4074),
    [sym_compiler_directive_decl] = STATE(4074),
    [sym_fsi_directive_decl] = STATE(4074),
    [sym_preproc_line] = STATE(4074),
    [ts_builtin_sym_end] = ACTIONS(7009),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_namespace] = ACTIONS(7011),
    [anon_sym_module] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_open] = ACTIONS(7011),
    [anon_sym_type] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_exception] = ACTIONS(7011),
    [anon_sym_extern] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7009),
    [anon_sym_POUNDload] = ACTIONS(7009),
    [anon_sym_POUNDtime] = ACTIONS(7009),
    [anon_sym_POUNDI] = ACTIONS(7009),
    [anon_sym_POUNDhelp] = ACTIONS(7009),
    [anon_sym_POUNDquit] = ACTIONS(7009),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(4075)] = {
    [sym_attributes] = STATE(13389),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7140),
    [sym_access_modifier] = STATE(22369),
    [sym__class_function_or_value_defn] = STATE(7141),
    [aux_sym__type_extension_inner] = STATE(4347),
    [sym_type_extension_elements] = STATE(8480),
    [sym__type_defn_elements] = STATE(7141),
    [sym_interface_implementation] = STATE(6973),
    [sym_member_defn] = STATE(6973),
    [sym_additional_constr_defn] = STATE(7045),
    [sym_block_comment] = STATE(4075),
    [sym_compiler_directive_decl] = STATE(4075),
    [sym_fsi_directive_decl] = STATE(4075),
    [sym_preproc_line] = STATE(4075),
    [sym_preproc_if_in_class_definition] = STATE(6973),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8242),
    [anon_sym_module] = ACTIONS(8242),
    [anon_sym_null] = ACTIONS(8242),
    [anon_sym_begin] = ACTIONS(8242),
    [anon_sym_open] = ACTIONS(8242),
    [anon_sym_type] = ACTIONS(8242),
    [anon_sym_LBRACK_LT] = ACTIONS(8244),
    [anon_sym_do] = ACTIONS(8242),
    [anon_sym_and] = ACTIONS(8242),
    [anon_sym_let] = ACTIONS(8242),
    [anon_sym_let_BANG] = ACTIONS(8244),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8242),
    [anon_sym_AMP] = ACTIONS(8242),
    [anon_sym_LBRACK] = ACTIONS(8242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8244),
    [anon_sym_LBRACE] = ACTIONS(8242),
    [anon_sym_LT_AT] = ACTIONS(8242),
    [anon_sym_LT_AT_AT] = ACTIONS(8244),
    [anon_sym_CARET] = ACTIONS(8244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8244),
    [anon_sym_new] = ACTIONS(8242),
    [anon_sym_with] = ACTIONS(8456),
    [anon_sym_return] = ACTIONS(8242),
    [anon_sym_return_BANG] = ACTIONS(8244),
    [anon_sym_yield] = ACTIONS(8242),
    [anon_sym_yield_BANG] = ACTIONS(8244),
    [anon_sym_lazy] = ACTIONS(8242),
    [anon_sym_assert] = ACTIONS(8242),
    [anon_sym_upcast] = ACTIONS(8242),
    [anon_sym_downcast] = ACTIONS(8242),
    [anon_sym_fixed] = ACTIONS(8242),
    [anon_sym_for] = ACTIONS(8242),
    [anon_sym_while] = ACTIONS(8242),
    [anon_sym_while_BANG] = ACTIONS(8244),
    [anon_sym_if] = ACTIONS(8242),
    [anon_sym_fun] = ACTIONS(8242),
    [anon_sym_try] = ACTIONS(8242),
    [anon_sym_match] = ACTIONS(8242),
    [anon_sym_match_BANG] = ACTIONS(8244),
    [anon_sym_function] = ACTIONS(8242),
    [anon_sym_use] = ACTIONS(8242),
    [anon_sym_use_BANG] = ACTIONS(8244),
    [anon_sym_do_BANG] = ACTIONS(8244),
    [anon_sym_AMP_AMP] = ACTIONS(8244),
    [anon_sym_default] = ACTIONS(8458),
    [anon_sym_static] = ACTIONS(8460),
    [anon_sym_member] = ACTIONS(8462),
    [anon_sym_exception] = ACTIONS(8242),
    [anon_sym_interface] = ACTIONS(8464),
    [anon_sym_abstract] = ACTIONS(8466),
    [anon_sym_val] = ACTIONS(8468),
    [anon_sym_override] = ACTIONS(8458),
    [anon_sym_extern] = ACTIONS(8242),
    [aux_sym_char_token1] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8242),
    [anon_sym_DQUOTE] = ACTIONS(8242),
    [anon_sym_AT_DQUOTE] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [sym_bool] = ACTIONS(8242),
    [sym_unit] = ACTIONS(8244),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8242),
    [sym_op_identifier] = ACTIONS(8244),
    [anon_sym_PLUS] = ACTIONS(8242),
    [anon_sym_DASH] = ACTIONS(8242),
    [anon_sym_PLUS_DOT] = ACTIONS(8244),
    [anon_sym_DASH_DOT] = ACTIONS(8244),
    [anon_sym_PERCENT] = ACTIONS(8242),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8244),
    [anon_sym_TILDE] = ACTIONS(8244),
    [aux_sym_prefix_op_token1] = ACTIONS(8244),
    [sym_int] = ACTIONS(8242),
    [sym_xint] = ACTIONS(8244),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8244),
    [anon_sym_POUNDload] = ACTIONS(8244),
    [anon_sym_POUNDtime] = ACTIONS(8244),
    [anon_sym_POUNDI] = ACTIONS(8244),
    [anon_sym_POUNDhelp] = ACTIONS(8244),
    [anon_sym_POUNDquit] = ACTIONS(8244),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8244),
    [anon_sym_POUNDendif] = ACTIONS(8244),
    [anon_sym_POUNDelse] = ACTIONS(8244),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8244),
  },
  [STATE(4076)] = {
    [sym_block_comment] = STATE(4076),
    [sym_compiler_directive_decl] = STATE(4076),
    [sym_fsi_directive_decl] = STATE(4076),
    [sym_preproc_line] = STATE(4076),
    [ts_builtin_sym_end] = ACTIONS(6981),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_namespace] = ACTIONS(6983),
    [anon_sym_module] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_open] = ACTIONS(6983),
    [anon_sym_type] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_exception] = ACTIONS(6983),
    [anon_sym_extern] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6981),
    [anon_sym_POUNDload] = ACTIONS(6981),
    [anon_sym_POUNDtime] = ACTIONS(6981),
    [anon_sym_POUNDI] = ACTIONS(6981),
    [anon_sym_POUNDhelp] = ACTIONS(6981),
    [anon_sym_POUNDquit] = ACTIONS(6981),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(4077)] = {
    [sym__else_expression] = STATE(5234),
    [sym_elif_expression] = STATE(4629),
    [sym_block_comment] = STATE(4077),
    [sym_compiler_directive_decl] = STATE(4077),
    [sym_fsi_directive_decl] = STATE(4077),
    [sym_preproc_line] = STATE(4077),
    [aux_sym_if_expression_repeat1] = STATE(4199),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8522),
    [anon_sym_elif] = ACTIONS(8524),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_DOT_DOT] = ACTIONS(6329),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [anon_sym_DOT_DOT2] = ACTIONS(6327),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__dedent] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4078)] = {
    [sym_block_comment] = STATE(4078),
    [sym_compiler_directive_decl] = STATE(4078),
    [sym_fsi_directive_decl] = STATE(4078),
    [sym_preproc_line] = STATE(4078),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_module] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_open] = ACTIONS(6840),
    [anon_sym_type] = ACTIONS(6840),
    [anon_sym_LBRACK_LT] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_exception] = ACTIONS(6840),
    [anon_sym_extern] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6838),
    [anon_sym_POUNDload] = ACTIONS(6838),
    [anon_sym_POUNDtime] = ACTIONS(6838),
    [anon_sym_POUNDI] = ACTIONS(6838),
    [anon_sym_POUNDhelp] = ACTIONS(6838),
    [anon_sym_POUNDquit] = ACTIONS(6838),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(4079)] = {
    [sym_block_comment] = STATE(4079),
    [sym_compiler_directive_decl] = STATE(4079),
    [sym_fsi_directive_decl] = STATE(4079),
    [sym_preproc_line] = STATE(4079),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_module] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_open] = ACTIONS(6653),
    [anon_sym_type] = ACTIONS(6653),
    [anon_sym_LBRACK_LT] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_exception] = ACTIONS(6653),
    [anon_sym_extern] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6651),
    [anon_sym_POUNDload] = ACTIONS(6651),
    [anon_sym_POUNDtime] = ACTIONS(6651),
    [anon_sym_POUNDI] = ACTIONS(6651),
    [anon_sym_POUNDhelp] = ACTIONS(6651),
    [anon_sym_POUNDquit] = ACTIONS(6651),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4080)] = {
    [sym_block_comment] = STATE(4080),
    [sym_compiler_directive_decl] = STATE(4080),
    [sym_fsi_directive_decl] = STATE(4080),
    [sym_preproc_line] = STATE(4080),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_module] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_open] = ACTIONS(6863),
    [anon_sym_type] = ACTIONS(6863),
    [anon_sym_LBRACK_LT] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_exception] = ACTIONS(6863),
    [anon_sym_extern] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6861),
    [anon_sym_POUNDload] = ACTIONS(6861),
    [anon_sym_POUNDtime] = ACTIONS(6861),
    [anon_sym_POUNDI] = ACTIONS(6861),
    [anon_sym_POUNDhelp] = ACTIONS(6861),
    [anon_sym_POUNDquit] = ACTIONS(6861),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(4081)] = {
    [sym_block_comment] = STATE(4081),
    [sym_compiler_directive_decl] = STATE(4081),
    [sym_fsi_directive_decl] = STATE(4081),
    [sym_preproc_line] = STATE(4081),
    [ts_builtin_sym_end] = ACTIONS(7001),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_namespace] = ACTIONS(7003),
    [anon_sym_module] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_open] = ACTIONS(7003),
    [anon_sym_type] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_exception] = ACTIONS(7003),
    [anon_sym_extern] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7001),
    [anon_sym_POUNDload] = ACTIONS(7001),
    [anon_sym_POUNDtime] = ACTIONS(7001),
    [anon_sym_POUNDI] = ACTIONS(7001),
    [anon_sym_POUNDhelp] = ACTIONS(7001),
    [anon_sym_POUNDquit] = ACTIONS(7001),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(4082)] = {
    [sym_block_comment] = STATE(4082),
    [sym_compiler_directive_decl] = STATE(4082),
    [sym_fsi_directive_decl] = STATE(4082),
    [sym_preproc_line] = STATE(4082),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_module] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_open] = ACTIONS(7117),
    [anon_sym_type] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_exception] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7119),
    [anon_sym_POUNDload] = ACTIONS(7119),
    [anon_sym_POUNDtime] = ACTIONS(7119),
    [anon_sym_POUNDI] = ACTIONS(7119),
    [anon_sym_POUNDhelp] = ACTIONS(7119),
    [anon_sym_POUNDquit] = ACTIONS(7119),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [anon_sym_POUNDelse] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(4083)] = {
    [sym_block_comment] = STATE(4083),
    [sym_compiler_directive_decl] = STATE(4083),
    [sym_fsi_directive_decl] = STATE(4083),
    [sym_preproc_line] = STATE(4083),
    [ts_builtin_sym_end] = ACTIONS(6911),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_namespace] = ACTIONS(6913),
    [anon_sym_module] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_open] = ACTIONS(6913),
    [anon_sym_type] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_exception] = ACTIONS(6913),
    [anon_sym_extern] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6911),
    [anon_sym_POUNDload] = ACTIONS(6911),
    [anon_sym_POUNDtime] = ACTIONS(6911),
    [anon_sym_POUNDI] = ACTIONS(6911),
    [anon_sym_POUNDhelp] = ACTIONS(6911),
    [anon_sym_POUNDquit] = ACTIONS(6911),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(4084)] = {
    [sym_block_comment] = STATE(4084),
    [sym_compiler_directive_decl] = STATE(4084),
    [sym_fsi_directive_decl] = STATE(4084),
    [sym_preproc_line] = STATE(4084),
    [ts_builtin_sym_end] = ACTIONS(8449),
    [sym_identifier] = ACTIONS(8447),
    [anon_sym_namespace] = ACTIONS(8447),
    [anon_sym_module] = ACTIONS(8447),
    [anon_sym_null] = ACTIONS(8447),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(8447),
    [anon_sym_open] = ACTIONS(8447),
    [anon_sym_type] = ACTIONS(8447),
    [anon_sym_LBRACK_LT] = ACTIONS(8449),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(8447),
    [anon_sym_let] = ACTIONS(8447),
    [anon_sym_let_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN] = ACTIONS(8447),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(8447),
    [anon_sym_LBRACK] = ACTIONS(8447),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8449),
    [anon_sym_LBRACE] = ACTIONS(8447),
    [anon_sym_LT_AT] = ACTIONS(8447),
    [anon_sym_LT_AT_AT] = ACTIONS(8447),
    [anon_sym_CARET] = ACTIONS(8447),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8449),
    [anon_sym_new] = ACTIONS(8447),
    [anon_sym_return] = ACTIONS(8447),
    [anon_sym_return_BANG] = ACTIONS(8449),
    [anon_sym_yield] = ACTIONS(8447),
    [anon_sym_yield_BANG] = ACTIONS(8449),
    [anon_sym_lazy] = ACTIONS(8447),
    [anon_sym_assert] = ACTIONS(8447),
    [anon_sym_upcast] = ACTIONS(8447),
    [anon_sym_downcast] = ACTIONS(8447),
    [anon_sym_fixed] = ACTIONS(8447),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(8447),
    [anon_sym_while] = ACTIONS(8447),
    [anon_sym_while_BANG] = ACTIONS(8449),
    [anon_sym_if] = ACTIONS(8447),
    [anon_sym_fun] = ACTIONS(8447),
    [anon_sym_try] = ACTIONS(8447),
    [anon_sym_match] = ACTIONS(8447),
    [anon_sym_match_BANG] = ACTIONS(8449),
    [anon_sym_function] = ACTIONS(8447),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(8447),
    [anon_sym_use_BANG] = ACTIONS(8449),
    [anon_sym_do_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(8447),
    [anon_sym_exception] = ACTIONS(8447),
    [anon_sym_extern] = ACTIONS(8447),
    [aux_sym_char_token1] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8447),
    [anon_sym_DQUOTE] = ACTIONS(8447),
    [anon_sym_AT_DQUOTE] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [sym_bool] = ACTIONS(8447),
    [sym_unit] = ACTIONS(8449),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8447),
    [sym_op_identifier] = ACTIONS(8447),
    [anon_sym_PLUS] = ACTIONS(8447),
    [anon_sym_DASH] = ACTIONS(8447),
    [anon_sym_PLUS_DOT] = ACTIONS(8447),
    [anon_sym_DASH_DOT] = ACTIONS(8447),
    [anon_sym_PERCENT] = ACTIONS(8447),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8447),
    [anon_sym_TILDE] = ACTIONS(8449),
    [aux_sym_prefix_op_token1] = ACTIONS(8447),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(8447),
    [sym_xint] = ACTIONS(8449),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8449),
    [anon_sym_POUNDload] = ACTIONS(8449),
    [anon_sym_POUNDtime] = ACTIONS(8449),
    [anon_sym_POUNDI] = ACTIONS(8449),
    [anon_sym_POUNDhelp] = ACTIONS(8449),
    [anon_sym_POUNDquit] = ACTIONS(8449),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8449),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8449),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4085)] = {
    [sym_block_comment] = STATE(4085),
    [sym_compiler_directive_decl] = STATE(4085),
    [sym_fsi_directive_decl] = STATE(4085),
    [sym_preproc_line] = STATE(4085),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_as] = ACTIONS(6395),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_DOT_DOT] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__dedent] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4086)] = {
    [sym_block_comment] = STATE(4086),
    [sym_compiler_directive_decl] = STATE(4086),
    [sym_fsi_directive_decl] = STATE(4086),
    [sym_preproc_line] = STATE(4086),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_module] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_open] = ACTIONS(6888),
    [anon_sym_type] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_exception] = ACTIONS(6888),
    [anon_sym_extern] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6886),
    [anon_sym_POUNDload] = ACTIONS(6886),
    [anon_sym_POUNDtime] = ACTIONS(6886),
    [anon_sym_POUNDI] = ACTIONS(6886),
    [anon_sym_POUNDhelp] = ACTIONS(6886),
    [anon_sym_POUNDquit] = ACTIONS(6886),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [anon_sym_POUNDelse] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(4087)] = {
    [sym_block_comment] = STATE(4087),
    [sym_compiler_directive_decl] = STATE(4087),
    [sym_fsi_directive_decl] = STATE(4087),
    [sym_preproc_line] = STATE(4087),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_module] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_open] = ACTIONS(6900),
    [anon_sym_type] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_exception] = ACTIONS(6900),
    [anon_sym_extern] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6898),
    [anon_sym_POUNDload] = ACTIONS(6898),
    [anon_sym_POUNDtime] = ACTIONS(6898),
    [anon_sym_POUNDI] = ACTIONS(6898),
    [anon_sym_POUNDhelp] = ACTIONS(6898),
    [anon_sym_POUNDquit] = ACTIONS(6898),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [anon_sym_POUNDelse] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(4088)] = {
    [sym_block_comment] = STATE(4088),
    [sym_compiler_directive_decl] = STATE(4088),
    [sym_fsi_directive_decl] = STATE(4088),
    [sym_preproc_line] = STATE(4088),
    [ts_builtin_sym_end] = ACTIONS(7095),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_namespace] = ACTIONS(7093),
    [anon_sym_module] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_open] = ACTIONS(7093),
    [anon_sym_type] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_exception] = ACTIONS(7093),
    [anon_sym_extern] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7095),
    [anon_sym_POUNDload] = ACTIONS(7095),
    [anon_sym_POUNDtime] = ACTIONS(7095),
    [anon_sym_POUNDI] = ACTIONS(7095),
    [anon_sym_POUNDhelp] = ACTIONS(7095),
    [anon_sym_POUNDquit] = ACTIONS(7095),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(4089)] = {
    [sym_block_comment] = STATE(4089),
    [sym_compiler_directive_decl] = STATE(4089),
    [sym_fsi_directive_decl] = STATE(4089),
    [sym_preproc_line] = STATE(4089),
    [aux_sym_type_argument_repeat1] = STATE(4089),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8536),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6312),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_DOT_DOT2] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4090)] = {
    [sym_type_arguments] = STATE(4553),
    [sym_long_identifier] = STATE(4568),
    [sym_block_comment] = STATE(4090),
    [sym_compiler_directive_decl] = STATE(4090),
    [sym_fsi_directive_decl] = STATE(4090),
    [sym_preproc_line] = STATE(4090),
    [aux_sym_compound_type_repeat1] = STATE(4443),
    [sym_identifier] = ACTIONS(8356),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_EQ] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_COLON] = ACTIONS(6213),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_COMMA] = ACTIONS(6209),
    [anon_sym_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_QMARK] = ACTIONS(6213),
    [anon_sym_COLON_COLON] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6213),
    [anon_sym_CARET] = ACTIONS(6213),
    [anon_sym_or] = ACTIONS(6213),
    [anon_sym_DOT] = ACTIONS(6213),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_COLON_GT] = ACTIONS(6209),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6209),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(8358),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_DOT_LBRACK] = ACTIONS(6209),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_LPAREN2] = ACTIONS(6213),
    [anon_sym_PIPE_PIPE] = ACTIONS(6213),
    [anon_sym_AMP_AMP] = ACTIONS(6213),
    [anon_sym_STAR] = ACTIONS(8360),
    [anon_sym_LT] = ACTIONS(8362),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8364),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6213),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6213),
    [anon_sym_DASH_DOT] = ACTIONS(6213),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6213),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6213),
    [aux_sym_infix_op_token1] = ACTIONS(6209),
    [aux_sym_infix_op_token2] = ACTIONS(6213),
    [aux_sym_infix_op_token3] = ACTIONS(6213),
    [anon_sym_BANG_EQ] = ACTIONS(6213),
    [anon_sym_COLON_EQ] = ACTIONS(6209),
    [anon_sym_DOLLAR] = ACTIONS(6213),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6213),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6213),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__quoted_close] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
    [sym__tyapp_open] = ACTIONS(6209),
  },
  [STATE(4091)] = {
    [sym_block_comment] = STATE(4091),
    [sym_compiler_directive_decl] = STATE(4091),
    [sym_fsi_directive_decl] = STATE(4091),
    [sym_preproc_line] = STATE(4091),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_module] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_open] = ACTIONS(7069),
    [anon_sym_type] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_exception] = ACTIONS(7069),
    [anon_sym_extern] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7071),
    [anon_sym_POUNDload] = ACTIONS(7071),
    [anon_sym_POUNDtime] = ACTIONS(7071),
    [anon_sym_POUNDI] = ACTIONS(7071),
    [anon_sym_POUNDhelp] = ACTIONS(7071),
    [anon_sym_POUNDquit] = ACTIONS(7071),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [anon_sym_POUNDelse] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(4092)] = {
    [sym_type_arguments] = STATE(4709),
    [sym_long_identifier] = STATE(4716),
    [sym_block_comment] = STATE(4092),
    [sym_compiler_directive_decl] = STATE(4092),
    [sym_fsi_directive_decl] = STATE(4092),
    [sym_preproc_line] = STATE(4092),
    [aux_sym_compound_type_repeat1] = STATE(4431),
    [sym_identifier] = ACTIONS(8490),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8492),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8494),
    [anon_sym_LT] = ACTIONS(8496),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8498),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(4093)] = {
    [sym_block_comment] = STATE(4093),
    [sym_compiler_directive_decl] = STATE(4093),
    [sym_fsi_directive_decl] = STATE(4093),
    [sym_preproc_line] = STATE(4093),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_as] = ACTIONS(6399),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_DOT_DOT] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__dedent] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4094)] = {
    [sym_block_comment] = STATE(4094),
    [sym_compiler_directive_decl] = STATE(4094),
    [sym_fsi_directive_decl] = STATE(4094),
    [sym_preproc_line] = STATE(4094),
    [ts_builtin_sym_end] = ACTIONS(6973),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_namespace] = ACTIONS(6975),
    [anon_sym_module] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_open] = ACTIONS(6975),
    [anon_sym_type] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_exception] = ACTIONS(6975),
    [anon_sym_extern] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6973),
    [anon_sym_POUNDload] = ACTIONS(6973),
    [anon_sym_POUNDtime] = ACTIONS(6973),
    [anon_sym_POUNDI] = ACTIONS(6973),
    [anon_sym_POUNDhelp] = ACTIONS(6973),
    [anon_sym_POUNDquit] = ACTIONS(6973),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(4095)] = {
    [sym_type_arguments] = STATE(4660),
    [sym_long_identifier] = STATE(4641),
    [sym_block_comment] = STATE(4095),
    [sym_compiler_directive_decl] = STATE(4095),
    [sym_fsi_directive_decl] = STATE(4095),
    [sym_preproc_line] = STATE(4095),
    [aux_sym_compound_type_repeat1] = STATE(4506),
    [sym_identifier] = ACTIONS(8502),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_GT] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8430),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8432),
    [anon_sym_LT] = ACTIONS(8434),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8436),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(4096)] = {
    [sym_block_comment] = STATE(4096),
    [sym_compiler_directive_decl] = STATE(4096),
    [sym_fsi_directive_decl] = STATE(4096),
    [sym_preproc_line] = STATE(4096),
    [aux_sym_long_identifier_repeat1] = STATE(3959),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8454),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_DOT_DOT] = ACTIONS(6289),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_DOT_DOT2] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__dedent] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4097)] = {
    [sym_block_comment] = STATE(4097),
    [sym_compiler_directive_decl] = STATE(4097),
    [sym_fsi_directive_decl] = STATE(4097),
    [sym_preproc_line] = STATE(4097),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_module] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_open] = ACTIONS(6902),
    [anon_sym_type] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_exception] = ACTIONS(6902),
    [anon_sym_extern] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6603),
    [anon_sym_POUNDload] = ACTIONS(6603),
    [anon_sym_POUNDtime] = ACTIONS(6603),
    [anon_sym_POUNDI] = ACTIONS(6603),
    [anon_sym_POUNDhelp] = ACTIONS(6603),
    [anon_sym_POUNDquit] = ACTIONS(6603),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [anon_sym_POUNDelse] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4098)] = {
    [sym_block_comment] = STATE(4098),
    [sym_compiler_directive_decl] = STATE(4098),
    [sym_fsi_directive_decl] = STATE(4098),
    [sym_preproc_line] = STATE(4098),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_module] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_open] = ACTIONS(6824),
    [anon_sym_type] = ACTIONS(6824),
    [anon_sym_LBRACK_LT] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_exception] = ACTIONS(6824),
    [anon_sym_extern] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6822),
    [anon_sym_POUNDload] = ACTIONS(6822),
    [anon_sym_POUNDtime] = ACTIONS(6822),
    [anon_sym_POUNDI] = ACTIONS(6822),
    [anon_sym_POUNDhelp] = ACTIONS(6822),
    [anon_sym_POUNDquit] = ACTIONS(6822),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(4099)] = {
    [sym_block_comment] = STATE(4099),
    [sym_compiler_directive_decl] = STATE(4099),
    [sym_fsi_directive_decl] = STATE(4099),
    [sym_preproc_line] = STATE(4099),
    [ts_builtin_sym_end] = ACTIONS(7037),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_namespace] = ACTIONS(7039),
    [anon_sym_module] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_open] = ACTIONS(7039),
    [anon_sym_type] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_exception] = ACTIONS(7039),
    [anon_sym_extern] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7037),
    [anon_sym_POUNDload] = ACTIONS(7037),
    [anon_sym_POUNDtime] = ACTIONS(7037),
    [anon_sym_POUNDI] = ACTIONS(7037),
    [anon_sym_POUNDhelp] = ACTIONS(7037),
    [anon_sym_POUNDquit] = ACTIONS(7037),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(4100)] = {
    [sym_block_comment] = STATE(4100),
    [sym_compiler_directive_decl] = STATE(4100),
    [sym_fsi_directive_decl] = STATE(4100),
    [sym_preproc_line] = STATE(4100),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8332),
    [anon_sym_POUNDelse] = ACTIONS(8332),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4101)] = {
    [sym_block_comment] = STATE(4101),
    [sym_compiler_directive_decl] = STATE(4101),
    [sym_fsi_directive_decl] = STATE(4101),
    [sym_preproc_line] = STATE(4101),
    [ts_builtin_sym_end] = ACTIONS(7099),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_namespace] = ACTIONS(7097),
    [anon_sym_module] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_open] = ACTIONS(7097),
    [anon_sym_type] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_exception] = ACTIONS(7097),
    [anon_sym_extern] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7099),
    [anon_sym_POUNDload] = ACTIONS(7099),
    [anon_sym_POUNDtime] = ACTIONS(7099),
    [anon_sym_POUNDI] = ACTIONS(7099),
    [anon_sym_POUNDhelp] = ACTIONS(7099),
    [anon_sym_POUNDquit] = ACTIONS(7099),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(4102)] = {
    [sym_block_comment] = STATE(4102),
    [sym_compiler_directive_decl] = STATE(4102),
    [sym_fsi_directive_decl] = STATE(4102),
    [sym_preproc_line] = STATE(4102),
    [ts_builtin_sym_end] = ACTIONS(7041),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_namespace] = ACTIONS(7043),
    [anon_sym_module] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_open] = ACTIONS(7043),
    [anon_sym_type] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_exception] = ACTIONS(7043),
    [anon_sym_extern] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7041),
    [anon_sym_POUNDload] = ACTIONS(7041),
    [anon_sym_POUNDtime] = ACTIONS(7041),
    [anon_sym_POUNDI] = ACTIONS(7041),
    [anon_sym_POUNDhelp] = ACTIONS(7041),
    [anon_sym_POUNDquit] = ACTIONS(7041),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(4103)] = {
    [sym_type_arguments] = STATE(4553),
    [sym_long_identifier] = STATE(4568),
    [sym_block_comment] = STATE(4103),
    [sym_compiler_directive_decl] = STATE(4103),
    [sym_fsi_directive_decl] = STATE(4103),
    [sym_preproc_line] = STATE(4103),
    [aux_sym_compound_type_repeat1] = STATE(4443),
    [sym_identifier] = ACTIONS(8356),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_EQ] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_COLON] = ACTIONS(6261),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_COMMA] = ACTIONS(6259),
    [anon_sym_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_QMARK] = ACTIONS(6261),
    [anon_sym_COLON_COLON] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6261),
    [anon_sym_CARET] = ACTIONS(6261),
    [anon_sym_or] = ACTIONS(6261),
    [anon_sym_DOT] = ACTIONS(6261),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_COLON_GT] = ACTIONS(6259),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6259),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(8358),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_DOT_LBRACK] = ACTIONS(6259),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_LPAREN2] = ACTIONS(6261),
    [anon_sym_PIPE_PIPE] = ACTIONS(6261),
    [anon_sym_AMP_AMP] = ACTIONS(6261),
    [anon_sym_STAR] = ACTIONS(8360),
    [anon_sym_LT] = ACTIONS(8362),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8364),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6261),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6261),
    [anon_sym_DASH_DOT] = ACTIONS(6261),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6261),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6261),
    [aux_sym_infix_op_token1] = ACTIONS(6259),
    [aux_sym_infix_op_token2] = ACTIONS(6261),
    [aux_sym_infix_op_token3] = ACTIONS(6261),
    [anon_sym_BANG_EQ] = ACTIONS(6261),
    [anon_sym_COLON_EQ] = ACTIONS(6259),
    [anon_sym_DOLLAR] = ACTIONS(6261),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6261),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6261),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__quoted_close] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
    [sym__tyapp_open] = ACTIONS(6259),
  },
  [STATE(4104)] = {
    [sym_type_arguments] = STATE(4709),
    [sym_long_identifier] = STATE(4716),
    [sym_block_comment] = STATE(4104),
    [sym_compiler_directive_decl] = STATE(4104),
    [sym_fsi_directive_decl] = STATE(4104),
    [sym_preproc_line] = STATE(4104),
    [aux_sym_compound_type_repeat1] = STATE(4431),
    [sym_identifier] = ACTIONS(8490),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_EQ] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_COLON] = ACTIONS(6257),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_COMMA] = ACTIONS(6255),
    [anon_sym_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_QMARK] = ACTIONS(6257),
    [anon_sym_COLON_COLON] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6257),
    [anon_sym_CARET] = ACTIONS(6257),
    [anon_sym_or] = ACTIONS(6257),
    [anon_sym_DOT] = ACTIONS(6257),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_COLON_GT] = ACTIONS(6255),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6255),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(8492),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_DOT_LBRACK] = ACTIONS(6255),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_LPAREN2] = ACTIONS(6257),
    [anon_sym_PIPE_PIPE] = ACTIONS(6257),
    [anon_sym_AMP_AMP] = ACTIONS(6257),
    [anon_sym_STAR] = ACTIONS(8494),
    [anon_sym_LT] = ACTIONS(8496),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(8498),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6257),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6257),
    [anon_sym_DASH_DOT] = ACTIONS(6257),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6257),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6257),
    [aux_sym_infix_op_token1] = ACTIONS(6255),
    [aux_sym_infix_op_token2] = ACTIONS(6257),
    [aux_sym_infix_op_token3] = ACTIONS(6257),
    [anon_sym_BANG_EQ] = ACTIONS(6257),
    [anon_sym_COLON_EQ] = ACTIONS(6255),
    [anon_sym_DOLLAR] = ACTIONS(6257),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6257),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6257),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
    [sym__tyapp_open] = ACTIONS(6255),
  },
  [STATE(4105)] = {
    [sym_block_comment] = STATE(4105),
    [sym_compiler_directive_decl] = STATE(4105),
    [sym_fsi_directive_decl] = STATE(4105),
    [sym_preproc_line] = STATE(4105),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_module] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_open] = ACTIONS(7121),
    [anon_sym_type] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_exception] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7123),
    [anon_sym_POUNDload] = ACTIONS(7123),
    [anon_sym_POUNDtime] = ACTIONS(7123),
    [anon_sym_POUNDI] = ACTIONS(7123),
    [anon_sym_POUNDhelp] = ACTIONS(7123),
    [anon_sym_POUNDquit] = ACTIONS(7123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [anon_sym_POUNDelse] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(4106)] = {
    [sym_block_comment] = STATE(4106),
    [sym_compiler_directive_decl] = STATE(4106),
    [sym_fsi_directive_decl] = STATE(4106),
    [sym_preproc_line] = STATE(4106),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_GT_RBRACK] = ACTIONS(6558),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_RBRACK] = ACTIONS(6558),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_RBRACE] = ACTIONS(6558),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_to] = ACTIONS(6560),
    [anon_sym_downto] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_DOT_DOT2] = ACTIONS(6558),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4107)] = {
    [sym_block_comment] = STATE(4107),
    [sym_compiler_directive_decl] = STATE(4107),
    [sym_fsi_directive_decl] = STATE(4107),
    [sym_preproc_line] = STATE(4107),
    [sym_identifier] = ACTIONS(8526),
    [anon_sym_module] = ACTIONS(8529),
    [anon_sym_null] = ACTIONS(8526),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(8526),
    [anon_sym_open] = ACTIONS(8529),
    [anon_sym_type] = ACTIONS(8529),
    [anon_sym_LBRACK_LT] = ACTIONS(8531),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(8526),
    [anon_sym_let] = ACTIONS(8526),
    [anon_sym_let_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN] = ACTIONS(8526),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(8526),
    [anon_sym_LBRACK] = ACTIONS(8526),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8533),
    [anon_sym_LBRACE] = ACTIONS(8526),
    [anon_sym_LT_AT] = ACTIONS(8526),
    [anon_sym_LT_AT_AT] = ACTIONS(8526),
    [anon_sym_CARET] = ACTIONS(8526),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8533),
    [anon_sym_new] = ACTIONS(8526),
    [anon_sym_return] = ACTIONS(8526),
    [anon_sym_return_BANG] = ACTIONS(8533),
    [anon_sym_yield] = ACTIONS(8526),
    [anon_sym_yield_BANG] = ACTIONS(8533),
    [anon_sym_lazy] = ACTIONS(8526),
    [anon_sym_assert] = ACTIONS(8526),
    [anon_sym_upcast] = ACTIONS(8526),
    [anon_sym_downcast] = ACTIONS(8526),
    [anon_sym_fixed] = ACTIONS(8526),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(8526),
    [anon_sym_while] = ACTIONS(8526),
    [anon_sym_while_BANG] = ACTIONS(8533),
    [anon_sym_if] = ACTIONS(8526),
    [anon_sym_fun] = ACTIONS(8526),
    [anon_sym_try] = ACTIONS(8526),
    [anon_sym_match] = ACTIONS(8526),
    [anon_sym_match_BANG] = ACTIONS(8533),
    [anon_sym_function] = ACTIONS(8526),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(8526),
    [anon_sym_use_BANG] = ACTIONS(8533),
    [anon_sym_do_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(8526),
    [anon_sym_exception] = ACTIONS(8529),
    [anon_sym_extern] = ACTIONS(8529),
    [aux_sym_char_token1] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8526),
    [anon_sym_DQUOTE] = ACTIONS(8526),
    [anon_sym_AT_DQUOTE] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [sym_bool] = ACTIONS(8526),
    [sym_unit] = ACTIONS(8533),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8526),
    [sym_op_identifier] = ACTIONS(8526),
    [anon_sym_PLUS] = ACTIONS(8526),
    [anon_sym_DASH] = ACTIONS(8526),
    [anon_sym_PLUS_DOT] = ACTIONS(8526),
    [anon_sym_DASH_DOT] = ACTIONS(8526),
    [anon_sym_PERCENT] = ACTIONS(8526),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8526),
    [anon_sym_TILDE] = ACTIONS(8533),
    [aux_sym_prefix_op_token1] = ACTIONS(8526),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(8526),
    [sym_xint] = ACTIONS(8533),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8531),
    [anon_sym_POUNDload] = ACTIONS(8531),
    [anon_sym_POUNDtime] = ACTIONS(8531),
    [anon_sym_POUNDI] = ACTIONS(8531),
    [anon_sym_POUNDhelp] = ACTIONS(8531),
    [anon_sym_POUNDquit] = ACTIONS(8531),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8533),
    [anon_sym_POUNDendif] = ACTIONS(8533),
    [anon_sym_POUNDelse] = ACTIONS(8533),
    [sym__newline] = ACTIONS(8533),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8533),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4108)] = {
    [sym_block_comment] = STATE(4108),
    [sym_compiler_directive_decl] = STATE(4108),
    [sym_fsi_directive_decl] = STATE(4108),
    [sym_preproc_line] = STATE(4108),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_module] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_open] = ACTIONS(7141),
    [anon_sym_type] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_exception] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7143),
    [anon_sym_POUNDload] = ACTIONS(7143),
    [anon_sym_POUNDtime] = ACTIONS(7143),
    [anon_sym_POUNDI] = ACTIONS(7143),
    [anon_sym_POUNDhelp] = ACTIONS(7143),
    [anon_sym_POUNDquit] = ACTIONS(7143),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [anon_sym_POUNDelse] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(4109)] = {
    [sym_block_comment] = STATE(4109),
    [sym_compiler_directive_decl] = STATE(4109),
    [sym_fsi_directive_decl] = STATE(4109),
    [sym_preproc_line] = STATE(4109),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_module] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_open] = ACTIONS(6906),
    [anon_sym_type] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_exception] = ACTIONS(6906),
    [anon_sym_extern] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6904),
    [anon_sym_POUNDload] = ACTIONS(6904),
    [anon_sym_POUNDtime] = ACTIONS(6904),
    [anon_sym_POUNDI] = ACTIONS(6904),
    [anon_sym_POUNDhelp] = ACTIONS(6904),
    [anon_sym_POUNDquit] = ACTIONS(6904),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [anon_sym_POUNDelse] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(4110)] = {
    [sym_block_comment] = STATE(4110),
    [sym_compiler_directive_decl] = STATE(4110),
    [sym_fsi_directive_decl] = STATE(4110),
    [sym_preproc_line] = STATE(4110),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_as] = ACTIONS(6403),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_DOT_DOT] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__dedent] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4111)] = {
    [sym_block_comment] = STATE(4111),
    [sym_compiler_directive_decl] = STATE(4111),
    [sym_fsi_directive_decl] = STATE(4111),
    [sym_preproc_line] = STATE(4111),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_module] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_open] = ACTIONS(7145),
    [anon_sym_type] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_exception] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7147),
    [anon_sym_POUNDload] = ACTIONS(7147),
    [anon_sym_POUNDtime] = ACTIONS(7147),
    [anon_sym_POUNDI] = ACTIONS(7147),
    [anon_sym_POUNDhelp] = ACTIONS(7147),
    [anon_sym_POUNDquit] = ACTIONS(7147),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [anon_sym_POUNDelse] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(4112)] = {
    [sym_block_comment] = STATE(4112),
    [sym_compiler_directive_decl] = STATE(4112),
    [sym_fsi_directive_decl] = STATE(4112),
    [sym_preproc_line] = STATE(4112),
    [ts_builtin_sym_end] = ACTIONS(7139),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_namespace] = ACTIONS(7137),
    [anon_sym_module] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_open] = ACTIONS(7137),
    [anon_sym_type] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_exception] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7139),
    [anon_sym_POUNDload] = ACTIONS(7139),
    [anon_sym_POUNDtime] = ACTIONS(7139),
    [anon_sym_POUNDI] = ACTIONS(7139),
    [anon_sym_POUNDhelp] = ACTIONS(7139),
    [anon_sym_POUNDquit] = ACTIONS(7139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(4113)] = {
    [sym_block_comment] = STATE(4113),
    [sym_compiler_directive_decl] = STATE(4113),
    [sym_fsi_directive_decl] = STATE(4113),
    [sym_preproc_line] = STATE(4113),
    [ts_builtin_sym_end] = ACTIONS(6603),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_namespace] = ACTIONS(6902),
    [anon_sym_module] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_open] = ACTIONS(6902),
    [anon_sym_type] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_exception] = ACTIONS(6902),
    [anon_sym_extern] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6603),
    [anon_sym_POUNDload] = ACTIONS(6603),
    [anon_sym_POUNDtime] = ACTIONS(6603),
    [anon_sym_POUNDI] = ACTIONS(6603),
    [anon_sym_POUNDhelp] = ACTIONS(6603),
    [anon_sym_POUNDquit] = ACTIONS(6603),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4114)] = {
    [sym_block_comment] = STATE(4114),
    [sym_compiler_directive_decl] = STATE(4114),
    [sym_fsi_directive_decl] = STATE(4114),
    [sym_preproc_line] = STATE(4114),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_as] = ACTIONS(6407),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_DOT_DOT] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__dedent] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4115)] = {
    [sym_block_comment] = STATE(4115),
    [sym_compiler_directive_decl] = STATE(4115),
    [sym_fsi_directive_decl] = STATE(4115),
    [sym_preproc_line] = STATE(4115),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_module] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_open] = ACTIONS(6936),
    [anon_sym_type] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_exception] = ACTIONS(6936),
    [anon_sym_extern] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6934),
    [anon_sym_POUNDload] = ACTIONS(6934),
    [anon_sym_POUNDtime] = ACTIONS(6934),
    [anon_sym_POUNDI] = ACTIONS(6934),
    [anon_sym_POUNDhelp] = ACTIONS(6934),
    [anon_sym_POUNDquit] = ACTIONS(6934),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [anon_sym_POUNDelse] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(4116)] = {
    [sym_block_comment] = STATE(4116),
    [sym_compiler_directive_decl] = STATE(4116),
    [sym_fsi_directive_decl] = STATE(4116),
    [sym_preproc_line] = STATE(4116),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_module] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_open] = ACTIONS(7155),
    [anon_sym_type] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_exception] = ACTIONS(7155),
    [anon_sym_extern] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7157),
    [anon_sym_POUNDload] = ACTIONS(7157),
    [anon_sym_POUNDtime] = ACTIONS(7157),
    [anon_sym_POUNDI] = ACTIONS(7157),
    [anon_sym_POUNDhelp] = ACTIONS(7157),
    [anon_sym_POUNDquit] = ACTIONS(7157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [anon_sym_POUNDelse] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(4117)] = {
    [sym_block_comment] = STATE(4117),
    [sym_compiler_directive_decl] = STATE(4117),
    [sym_fsi_directive_decl] = STATE(4117),
    [sym_preproc_line] = STATE(4117),
    [aux_sym_sequential_expression_repeat1] = STATE(4117),
    [ts_builtin_sym_end] = ACTIONS(6798),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_module] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_open] = ACTIONS(6800),
    [anon_sym_type] = ACTIONS(6800),
    [anon_sym_LBRACK_LT] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_exception] = ACTIONS(6800),
    [anon_sym_extern] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6798),
    [anon_sym_POUNDload] = ACTIONS(6798),
    [anon_sym_POUNDtime] = ACTIONS(6798),
    [anon_sym_POUNDI] = ACTIONS(6798),
    [anon_sym_POUNDhelp] = ACTIONS(6798),
    [anon_sym_POUNDquit] = ACTIONS(6798),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8539),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(4118)] = {
    [sym_block_comment] = STATE(4118),
    [sym_compiler_directive_decl] = STATE(4118),
    [sym_fsi_directive_decl] = STATE(4118),
    [sym_preproc_line] = STATE(4118),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_module] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_open] = ACTIONS(6913),
    [anon_sym_type] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_exception] = ACTIONS(6913),
    [anon_sym_extern] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6911),
    [anon_sym_POUNDload] = ACTIONS(6911),
    [anon_sym_POUNDtime] = ACTIONS(6911),
    [anon_sym_POUNDI] = ACTIONS(6911),
    [anon_sym_POUNDhelp] = ACTIONS(6911),
    [anon_sym_POUNDquit] = ACTIONS(6911),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [anon_sym_POUNDelse] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(4119)] = {
    [sym_block_comment] = STATE(4119),
    [sym_compiler_directive_decl] = STATE(4119),
    [sym_fsi_directive_decl] = STATE(4119),
    [sym_preproc_line] = STATE(4119),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_module] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_open] = ACTIONS(7159),
    [anon_sym_type] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_exception] = ACTIONS(7159),
    [anon_sym_extern] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7161),
    [anon_sym_POUNDload] = ACTIONS(7161),
    [anon_sym_POUNDtime] = ACTIONS(7161),
    [anon_sym_POUNDI] = ACTIONS(7161),
    [anon_sym_POUNDhelp] = ACTIONS(7161),
    [anon_sym_POUNDquit] = ACTIONS(7161),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [anon_sym_POUNDelse] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(4120)] = {
    [sym_block_comment] = STATE(4120),
    [sym_compiler_directive_decl] = STATE(4120),
    [sym_fsi_directive_decl] = STATE(4120),
    [sym_preproc_line] = STATE(4120),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_module] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_open] = ACTIONS(6917),
    [anon_sym_type] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_exception] = ACTIONS(6917),
    [anon_sym_extern] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6915),
    [anon_sym_POUNDload] = ACTIONS(6915),
    [anon_sym_POUNDtime] = ACTIONS(6915),
    [anon_sym_POUNDI] = ACTIONS(6915),
    [anon_sym_POUNDhelp] = ACTIONS(6915),
    [anon_sym_POUNDquit] = ACTIONS(6915),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [anon_sym_POUNDelse] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(4121)] = {
    [sym_block_comment] = STATE(4121),
    [sym_compiler_directive_decl] = STATE(4121),
    [sym_fsi_directive_decl] = STATE(4121),
    [sym_preproc_line] = STATE(4121),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_module] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_open] = ACTIONS(6921),
    [anon_sym_type] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_exception] = ACTIONS(6921),
    [anon_sym_extern] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6919),
    [anon_sym_POUNDload] = ACTIONS(6919),
    [anon_sym_POUNDtime] = ACTIONS(6919),
    [anon_sym_POUNDI] = ACTIONS(6919),
    [anon_sym_POUNDhelp] = ACTIONS(6919),
    [anon_sym_POUNDquit] = ACTIONS(6919),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [anon_sym_POUNDelse] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(4122)] = {
    [sym_block_comment] = STATE(4122),
    [sym_compiler_directive_decl] = STATE(4122),
    [sym_fsi_directive_decl] = STATE(4122),
    [sym_preproc_line] = STATE(4122),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_module] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_open] = ACTIONS(6940),
    [anon_sym_type] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_exception] = ACTIONS(6940),
    [anon_sym_extern] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6938),
    [anon_sym_POUNDload] = ACTIONS(6938),
    [anon_sym_POUNDtime] = ACTIONS(6938),
    [anon_sym_POUNDI] = ACTIONS(6938),
    [anon_sym_POUNDhelp] = ACTIONS(6938),
    [anon_sym_POUNDquit] = ACTIONS(6938),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [anon_sym_POUNDelse] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(4123)] = {
    [sym_block_comment] = STATE(4123),
    [sym_compiler_directive_decl] = STATE(4123),
    [sym_fsi_directive_decl] = STATE(4123),
    [sym_preproc_line] = STATE(4123),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_as] = ACTIONS(6411),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_DOT_DOT] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__dedent] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4124)] = {
    [sym_block_comment] = STATE(4124),
    [sym_compiler_directive_decl] = STATE(4124),
    [sym_fsi_directive_decl] = STATE(4124),
    [sym_preproc_line] = STATE(4124),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_module] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_open] = ACTIONS(7015),
    [anon_sym_type] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_exception] = ACTIONS(7015),
    [anon_sym_extern] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7013),
    [anon_sym_POUNDload] = ACTIONS(7013),
    [anon_sym_POUNDtime] = ACTIONS(7013),
    [anon_sym_POUNDI] = ACTIONS(7013),
    [anon_sym_POUNDhelp] = ACTIONS(7013),
    [anon_sym_POUNDquit] = ACTIONS(7013),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [anon_sym_POUNDelse] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(4125)] = {
    [sym_block_comment] = STATE(4125),
    [sym_compiler_directive_decl] = STATE(4125),
    [sym_fsi_directive_decl] = STATE(4125),
    [sym_preproc_line] = STATE(4125),
    [aux_sym_long_identifier_repeat1] = STATE(4125),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_GT_RBRACK] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_RBRACK] = ACTIONS(6297),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_RBRACE] = ACTIONS(6297),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8542),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_to] = ACTIONS(6295),
    [anon_sym_downto] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4126)] = {
    [sym_block_comment] = STATE(4126),
    [sym_compiler_directive_decl] = STATE(4126),
    [sym_fsi_directive_decl] = STATE(4126),
    [sym_preproc_line] = STATE(4126),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_module] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_open] = ACTIONS(7141),
    [anon_sym_type] = ACTIONS(7141),
    [anon_sym_LBRACK_LT] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_exception] = ACTIONS(7141),
    [anon_sym_extern] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7143),
    [anon_sym_POUNDload] = ACTIONS(7143),
    [anon_sym_POUNDtime] = ACTIONS(7143),
    [anon_sym_POUNDI] = ACTIONS(7143),
    [anon_sym_POUNDhelp] = ACTIONS(7143),
    [anon_sym_POUNDquit] = ACTIONS(7143),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(4127)] = {
    [sym_block_comment] = STATE(4127),
    [sym_compiler_directive_decl] = STATE(4127),
    [sym_fsi_directive_decl] = STATE(4127),
    [sym_preproc_line] = STATE(4127),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_GT_RBRACK] = ACTIONS(6822),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_RBRACK] = ACTIONS(6822),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_RBRACE] = ACTIONS(6822),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_to] = ACTIONS(6824),
    [anon_sym_downto] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_DOT_DOT2] = ACTIONS(6822),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(4128)] = {
    [sym_block_comment] = STATE(4128),
    [sym_compiler_directive_decl] = STATE(4128),
    [sym_fsi_directive_decl] = STATE(4128),
    [sym_preproc_line] = STATE(4128),
    [sym_identifier] = ACTIONS(8526),
    [anon_sym_module] = ACTIONS(8529),
    [anon_sym_null] = ACTIONS(8526),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(8526),
    [anon_sym_open] = ACTIONS(8529),
    [anon_sym_type] = ACTIONS(8529),
    [anon_sym_LBRACK_LT] = ACTIONS(8531),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(8526),
    [anon_sym_let] = ACTIONS(8526),
    [anon_sym_let_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN] = ACTIONS(8526),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(8526),
    [anon_sym_LBRACK] = ACTIONS(8526),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8533),
    [anon_sym_LBRACE] = ACTIONS(8526),
    [anon_sym_LT_AT] = ACTIONS(8526),
    [anon_sym_LT_AT_AT] = ACTIONS(8526),
    [anon_sym_CARET] = ACTIONS(8526),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8533),
    [anon_sym_new] = ACTIONS(8526),
    [anon_sym_return] = ACTIONS(8526),
    [anon_sym_return_BANG] = ACTIONS(8533),
    [anon_sym_yield] = ACTIONS(8526),
    [anon_sym_yield_BANG] = ACTIONS(8533),
    [anon_sym_lazy] = ACTIONS(8526),
    [anon_sym_assert] = ACTIONS(8526),
    [anon_sym_upcast] = ACTIONS(8526),
    [anon_sym_downcast] = ACTIONS(8526),
    [anon_sym_fixed] = ACTIONS(8526),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(8526),
    [anon_sym_while] = ACTIONS(8526),
    [anon_sym_while_BANG] = ACTIONS(8533),
    [anon_sym_if] = ACTIONS(8526),
    [anon_sym_fun] = ACTIONS(8526),
    [anon_sym_try] = ACTIONS(8526),
    [anon_sym_match] = ACTIONS(8526),
    [anon_sym_match_BANG] = ACTIONS(8533),
    [anon_sym_function] = ACTIONS(8526),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(8526),
    [anon_sym_use_BANG] = ACTIONS(8533),
    [anon_sym_do_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(8526),
    [anon_sym_exception] = ACTIONS(8529),
    [anon_sym_extern] = ACTIONS(8529),
    [aux_sym_char_token1] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8526),
    [anon_sym_DQUOTE] = ACTIONS(8526),
    [anon_sym_AT_DQUOTE] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [sym_bool] = ACTIONS(8526),
    [sym_unit] = ACTIONS(8533),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8526),
    [sym_op_identifier] = ACTIONS(8526),
    [anon_sym_PLUS] = ACTIONS(8526),
    [anon_sym_DASH] = ACTIONS(8526),
    [anon_sym_PLUS_DOT] = ACTIONS(8526),
    [anon_sym_DASH_DOT] = ACTIONS(8526),
    [anon_sym_PERCENT] = ACTIONS(8526),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8526),
    [anon_sym_TILDE] = ACTIONS(8533),
    [aux_sym_prefix_op_token1] = ACTIONS(8526),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(8526),
    [sym_xint] = ACTIONS(8533),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8531),
    [anon_sym_POUNDload] = ACTIONS(8531),
    [anon_sym_POUNDtime] = ACTIONS(8531),
    [anon_sym_POUNDI] = ACTIONS(8531),
    [anon_sym_POUNDhelp] = ACTIONS(8531),
    [anon_sym_POUNDquit] = ACTIONS(8531),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8533),
    [anon_sym_POUNDendif] = ACTIONS(8531),
    [sym__newline] = ACTIONS(8533),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8533),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4129)] = {
    [sym_block_comment] = STATE(4129),
    [sym_compiler_directive_decl] = STATE(4129),
    [sym_fsi_directive_decl] = STATE(4129),
    [sym_preproc_line] = STATE(4129),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_GT_RBRACK] = ACTIONS(6861),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_RBRACK] = ACTIONS(6861),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_RBRACE] = ACTIONS(6861),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_to] = ACTIONS(6863),
    [anon_sym_downto] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_DOT_DOT2] = ACTIONS(6861),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(4130)] = {
    [sym_block_comment] = STATE(4130),
    [sym_compiler_directive_decl] = STATE(4130),
    [sym_fsi_directive_decl] = STATE(4130),
    [sym_preproc_line] = STATE(4130),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_GT_RBRACK] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_RBRACK] = ACTIONS(6832),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_RBRACE] = ACTIONS(6832),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8545),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_to] = ACTIONS(6834),
    [anon_sym_downto] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_DOT_DOT2] = ACTIONS(6832),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4131)] = {
    [sym_block_comment] = STATE(4131),
    [sym_compiler_directive_decl] = STATE(4131),
    [sym_fsi_directive_decl] = STATE(4131),
    [sym_preproc_line] = STATE(4131),
    [aux_sym_long_identifier_repeat1] = STATE(4145),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_as] = ACTIONS(6289),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8547),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4132)] = {
    [sym_block_comment] = STATE(4132),
    [sym_compiler_directive_decl] = STATE(4132),
    [sym_fsi_directive_decl] = STATE(4132),
    [sym_preproc_line] = STATE(4132),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_GT_RBRACK] = ACTIONS(6865),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_RBRACK] = ACTIONS(6865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_RBRACE] = ACTIONS(6865),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_to] = ACTIONS(6867),
    [anon_sym_downto] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_DOT_DOT2] = ACTIONS(6865),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(4133)] = {
    [sym_block_comment] = STATE(4133),
    [sym_compiler_directive_decl] = STATE(4133),
    [sym_fsi_directive_decl] = STATE(4133),
    [sym_preproc_line] = STATE(4133),
    [aux_sym_long_identifier_repeat1] = STATE(4150),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8549),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4134)] = {
    [sym_block_comment] = STATE(4134),
    [sym_compiler_directive_decl] = STATE(4134),
    [sym_fsi_directive_decl] = STATE(4134),
    [sym_preproc_line] = STATE(4134),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_module] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_open] = ACTIONS(6807),
    [anon_sym_type] = ACTIONS(6807),
    [anon_sym_LBRACK_LT] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_exception] = ACTIONS(6807),
    [anon_sym_extern] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6805),
    [anon_sym_POUNDload] = ACTIONS(6805),
    [anon_sym_POUNDtime] = ACTIONS(6805),
    [anon_sym_POUNDI] = ACTIONS(6805),
    [anon_sym_POUNDhelp] = ACTIONS(6805),
    [anon_sym_POUNDquit] = ACTIONS(6805),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4135)] = {
    [sym_block_comment] = STATE(4135),
    [sym_compiler_directive_decl] = STATE(4135),
    [sym_fsi_directive_decl] = STATE(4135),
    [sym_preproc_line] = STATE(4135),
    [aux_sym_type_argument_repeat1] = STATE(4136),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__dedent] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
    [sym__in] = ACTIONS(6306),
  },
  [STATE(4136)] = {
    [sym_block_comment] = STATE(4136),
    [sym_compiler_directive_decl] = STATE(4136),
    [sym_fsi_directive_decl] = STATE(4136),
    [sym_preproc_line] = STATE(4136),
    [aux_sym_type_argument_repeat1] = STATE(4136),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8551),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
    [sym__in] = ACTIONS(6310),
  },
  [STATE(4137)] = {
    [sym_block_comment] = STATE(4137),
    [sym_compiler_directive_decl] = STATE(4137),
    [sym_fsi_directive_decl] = STATE(4137),
    [sym_preproc_line] = STATE(4137),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_module] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_open] = ACTIONS(6892),
    [anon_sym_type] = ACTIONS(6892),
    [anon_sym_LBRACK_LT] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_exception] = ACTIONS(6892),
    [anon_sym_extern] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6890),
    [anon_sym_POUNDload] = ACTIONS(6890),
    [anon_sym_POUNDtime] = ACTIONS(6890),
    [anon_sym_POUNDI] = ACTIONS(6890),
    [anon_sym_POUNDhelp] = ACTIONS(6890),
    [anon_sym_POUNDquit] = ACTIONS(6890),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [anon_sym_POUNDendif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(4138)] = {
    [sym_block_comment] = STATE(4138),
    [sym_compiler_directive_decl] = STATE(4138),
    [sym_fsi_directive_decl] = STATE(4138),
    [sym_preproc_line] = STATE(4138),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_module] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_open] = ACTIONS(6896),
    [anon_sym_type] = ACTIONS(6896),
    [anon_sym_LBRACK_LT] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_exception] = ACTIONS(6896),
    [anon_sym_extern] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6894),
    [anon_sym_POUNDload] = ACTIONS(6894),
    [anon_sym_POUNDtime] = ACTIONS(6894),
    [anon_sym_POUNDI] = ACTIONS(6894),
    [anon_sym_POUNDhelp] = ACTIONS(6894),
    [anon_sym_POUNDquit] = ACTIONS(6894),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [anon_sym_POUNDendif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(4139)] = {
    [sym_block_comment] = STATE(4139),
    [sym_compiler_directive_decl] = STATE(4139),
    [sym_fsi_directive_decl] = STATE(4139),
    [sym_preproc_line] = STATE(4139),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_module] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_open] = ACTIONS(6900),
    [anon_sym_type] = ACTIONS(6900),
    [anon_sym_LBRACK_LT] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_exception] = ACTIONS(6900),
    [anon_sym_extern] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6898),
    [anon_sym_POUNDload] = ACTIONS(6898),
    [anon_sym_POUNDtime] = ACTIONS(6898),
    [anon_sym_POUNDI] = ACTIONS(6898),
    [anon_sym_POUNDhelp] = ACTIONS(6898),
    [anon_sym_POUNDquit] = ACTIONS(6898),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [anon_sym_POUNDendif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(4140)] = {
    [sym_block_comment] = STATE(4140),
    [sym_compiler_directive_decl] = STATE(4140),
    [sym_fsi_directive_decl] = STATE(4140),
    [sym_preproc_line] = STATE(4140),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_module] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_open] = ACTIONS(6902),
    [anon_sym_type] = ACTIONS(6902),
    [anon_sym_LBRACK_LT] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_exception] = ACTIONS(6902),
    [anon_sym_extern] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6603),
    [anon_sym_POUNDload] = ACTIONS(6603),
    [anon_sym_POUNDtime] = ACTIONS(6603),
    [anon_sym_POUNDI] = ACTIONS(6603),
    [anon_sym_POUNDhelp] = ACTIONS(6603),
    [anon_sym_POUNDquit] = ACTIONS(6603),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4141)] = {
    [sym_block_comment] = STATE(4141),
    [sym_compiler_directive_decl] = STATE(4141),
    [sym_fsi_directive_decl] = STATE(4141),
    [sym_preproc_line] = STATE(4141),
    [aux_sym_long_identifier_repeat1] = STATE(4169),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__dedent] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
    [sym__in] = ACTIONS(6291),
  },
  [STATE(4142)] = {
    [sym_block_comment] = STATE(4142),
    [sym_compiler_directive_decl] = STATE(4142),
    [sym_fsi_directive_decl] = STATE(4142),
    [sym_preproc_line] = STATE(4142),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_module] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_open] = ACTIONS(6906),
    [anon_sym_type] = ACTIONS(6906),
    [anon_sym_LBRACK_LT] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_exception] = ACTIONS(6906),
    [anon_sym_extern] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6904),
    [anon_sym_POUNDload] = ACTIONS(6904),
    [anon_sym_POUNDtime] = ACTIONS(6904),
    [anon_sym_POUNDI] = ACTIONS(6904),
    [anon_sym_POUNDhelp] = ACTIONS(6904),
    [anon_sym_POUNDquit] = ACTIONS(6904),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [anon_sym_POUNDendif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(4143)] = {
    [sym_block_comment] = STATE(4143),
    [sym_compiler_directive_decl] = STATE(4143),
    [sym_fsi_directive_decl] = STATE(4143),
    [sym_preproc_line] = STATE(4143),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_module] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_open] = ACTIONS(6913),
    [anon_sym_type] = ACTIONS(6913),
    [anon_sym_LBRACK_LT] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_exception] = ACTIONS(6913),
    [anon_sym_extern] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6911),
    [anon_sym_POUNDload] = ACTIONS(6911),
    [anon_sym_POUNDtime] = ACTIONS(6911),
    [anon_sym_POUNDI] = ACTIONS(6911),
    [anon_sym_POUNDhelp] = ACTIONS(6911),
    [anon_sym_POUNDquit] = ACTIONS(6911),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [anon_sym_POUNDendif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(4144)] = {
    [sym_block_comment] = STATE(4144),
    [sym_compiler_directive_decl] = STATE(4144),
    [sym_fsi_directive_decl] = STATE(4144),
    [sym_preproc_line] = STATE(4144),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_module] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_open] = ACTIONS(6917),
    [anon_sym_type] = ACTIONS(6917),
    [anon_sym_LBRACK_LT] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_exception] = ACTIONS(6917),
    [anon_sym_extern] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6915),
    [anon_sym_POUNDload] = ACTIONS(6915),
    [anon_sym_POUNDtime] = ACTIONS(6915),
    [anon_sym_POUNDI] = ACTIONS(6915),
    [anon_sym_POUNDhelp] = ACTIONS(6915),
    [anon_sym_POUNDquit] = ACTIONS(6915),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [anon_sym_POUNDendif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(4145)] = {
    [sym_block_comment] = STATE(4145),
    [sym_compiler_directive_decl] = STATE(4145),
    [sym_fsi_directive_decl] = STATE(4145),
    [sym_preproc_line] = STATE(4145),
    [aux_sym_long_identifier_repeat1] = STATE(4148),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_as] = ACTIONS(6360),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8547),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4146)] = {
    [sym_block_comment] = STATE(4146),
    [sym_compiler_directive_decl] = STATE(4146),
    [sym_fsi_directive_decl] = STATE(4146),
    [sym_preproc_line] = STATE(4146),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_module] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_open] = ACTIONS(6921),
    [anon_sym_type] = ACTIONS(6921),
    [anon_sym_LBRACK_LT] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_exception] = ACTIONS(6921),
    [anon_sym_extern] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6919),
    [anon_sym_POUNDload] = ACTIONS(6919),
    [anon_sym_POUNDtime] = ACTIONS(6919),
    [anon_sym_POUNDI] = ACTIONS(6919),
    [anon_sym_POUNDhelp] = ACTIONS(6919),
    [anon_sym_POUNDquit] = ACTIONS(6919),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [anon_sym_POUNDendif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(4147)] = {
    [sym_block_comment] = STATE(4147),
    [sym_compiler_directive_decl] = STATE(4147),
    [sym_fsi_directive_decl] = STATE(4147),
    [sym_preproc_line] = STATE(4147),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_module] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_open] = ACTIONS(6925),
    [anon_sym_type] = ACTIONS(6925),
    [anon_sym_LBRACK_LT] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_exception] = ACTIONS(6925),
    [anon_sym_extern] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6923),
    [anon_sym_POUNDload] = ACTIONS(6923),
    [anon_sym_POUNDtime] = ACTIONS(6923),
    [anon_sym_POUNDI] = ACTIONS(6923),
    [anon_sym_POUNDhelp] = ACTIONS(6923),
    [anon_sym_POUNDquit] = ACTIONS(6923),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [anon_sym_POUNDendif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(4148)] = {
    [sym_block_comment] = STATE(4148),
    [sym_compiler_directive_decl] = STATE(4148),
    [sym_fsi_directive_decl] = STATE(4148),
    [sym_preproc_line] = STATE(4148),
    [aux_sym_long_identifier_repeat1] = STATE(4148),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8556),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4149)] = {
    [sym_block_comment] = STATE(4149),
    [sym_compiler_directive_decl] = STATE(4149),
    [sym_fsi_directive_decl] = STATE(4149),
    [sym_preproc_line] = STATE(4149),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_GT_RBRACK] = ACTIONS(6694),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_RBRACK] = ACTIONS(6694),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_RBRACE] = ACTIONS(6694),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_to] = ACTIONS(6692),
    [anon_sym_downto] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_DOT_DOT2] = ACTIONS(6694),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(4150)] = {
    [sym_block_comment] = STATE(4150),
    [sym_compiler_directive_decl] = STATE(4150),
    [sym_fsi_directive_decl] = STATE(4150),
    [sym_preproc_line] = STATE(4150),
    [aux_sym_long_identifier_repeat1] = STATE(4151),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8549),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4151)] = {
    [sym_block_comment] = STATE(4151),
    [sym_compiler_directive_decl] = STATE(4151),
    [sym_fsi_directive_decl] = STATE(4151),
    [sym_preproc_line] = STATE(4151),
    [aux_sym_long_identifier_repeat1] = STATE(4151),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8559),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4152)] = {
    [sym_block_comment] = STATE(4152),
    [sym_compiler_directive_decl] = STATE(4152),
    [sym_fsi_directive_decl] = STATE(4152),
    [sym_preproc_line] = STATE(4152),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_module] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_open] = ACTIONS(290),
    [anon_sym_type] = ACTIONS(290),
    [anon_sym_LBRACK_LT] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_exception] = ACTIONS(290),
    [anon_sym_extern] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(292),
    [anon_sym_POUNDload] = ACTIONS(292),
    [anon_sym_POUNDtime] = ACTIONS(292),
    [anon_sym_POUNDI] = ACTIONS(292),
    [anon_sym_POUNDhelp] = ACTIONS(292),
    [anon_sym_POUNDquit] = ACTIONS(292),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [anon_sym_POUNDendif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(4153)] = {
    [sym_block_comment] = STATE(4153),
    [sym_compiler_directive_decl] = STATE(4153),
    [sym_fsi_directive_decl] = STATE(4153),
    [sym_preproc_line] = STATE(4153),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_module] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_open] = ACTIONS(6932),
    [anon_sym_type] = ACTIONS(6932),
    [anon_sym_LBRACK_LT] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_exception] = ACTIONS(6932),
    [anon_sym_extern] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6930),
    [anon_sym_POUNDload] = ACTIONS(6930),
    [anon_sym_POUNDtime] = ACTIONS(6930),
    [anon_sym_POUNDI] = ACTIONS(6930),
    [anon_sym_POUNDhelp] = ACTIONS(6930),
    [anon_sym_POUNDquit] = ACTIONS(6930),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [anon_sym_POUNDendif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(4154)] = {
    [sym_block_comment] = STATE(4154),
    [sym_compiler_directive_decl] = STATE(4154),
    [sym_fsi_directive_decl] = STATE(4154),
    [sym_preproc_line] = STATE(4154),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_module] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_open] = ACTIONS(6944),
    [anon_sym_type] = ACTIONS(6944),
    [anon_sym_LBRACK_LT] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_exception] = ACTIONS(6944),
    [anon_sym_extern] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6942),
    [anon_sym_POUNDload] = ACTIONS(6942),
    [anon_sym_POUNDtime] = ACTIONS(6942),
    [anon_sym_POUNDI] = ACTIONS(6942),
    [anon_sym_POUNDhelp] = ACTIONS(6942),
    [anon_sym_POUNDquit] = ACTIONS(6942),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(4155)] = {
    [sym_block_comment] = STATE(4155),
    [sym_compiler_directive_decl] = STATE(4155),
    [sym_fsi_directive_decl] = STATE(4155),
    [sym_preproc_line] = STATE(4155),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_module] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_open] = ACTIONS(6948),
    [anon_sym_type] = ACTIONS(6948),
    [anon_sym_LBRACK_LT] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_exception] = ACTIONS(6948),
    [anon_sym_extern] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6946),
    [anon_sym_POUNDload] = ACTIONS(6946),
    [anon_sym_POUNDtime] = ACTIONS(6946),
    [anon_sym_POUNDI] = ACTIONS(6946),
    [anon_sym_POUNDhelp] = ACTIONS(6946),
    [anon_sym_POUNDquit] = ACTIONS(6946),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(4156)] = {
    [sym_block_comment] = STATE(4156),
    [sym_compiler_directive_decl] = STATE(4156),
    [sym_fsi_directive_decl] = STATE(4156),
    [sym_preproc_line] = STATE(4156),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_module] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_open] = ACTIONS(6952),
    [anon_sym_type] = ACTIONS(6952),
    [anon_sym_LBRACK_LT] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_exception] = ACTIONS(6952),
    [anon_sym_extern] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6950),
    [anon_sym_POUNDload] = ACTIONS(6950),
    [anon_sym_POUNDtime] = ACTIONS(6950),
    [anon_sym_POUNDI] = ACTIONS(6950),
    [anon_sym_POUNDhelp] = ACTIONS(6950),
    [anon_sym_POUNDquit] = ACTIONS(6950),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(4157)] = {
    [sym_block_comment] = STATE(4157),
    [sym_compiler_directive_decl] = STATE(4157),
    [sym_fsi_directive_decl] = STATE(4157),
    [sym_preproc_line] = STATE(4157),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_module] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_open] = ACTIONS(6959),
    [anon_sym_type] = ACTIONS(6959),
    [anon_sym_LBRACK_LT] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_exception] = ACTIONS(6959),
    [anon_sym_extern] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6957),
    [anon_sym_POUNDload] = ACTIONS(6957),
    [anon_sym_POUNDtime] = ACTIONS(6957),
    [anon_sym_POUNDI] = ACTIONS(6957),
    [anon_sym_POUNDhelp] = ACTIONS(6957),
    [anon_sym_POUNDquit] = ACTIONS(6957),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4158)] = {
    [sym_block_comment] = STATE(4158),
    [sym_compiler_directive_decl] = STATE(4158),
    [sym_fsi_directive_decl] = STATE(4158),
    [sym_preproc_line] = STATE(4158),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_module] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_open] = ACTIONS(6963),
    [anon_sym_type] = ACTIONS(6963),
    [anon_sym_LBRACK_LT] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_exception] = ACTIONS(6963),
    [anon_sym_extern] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6961),
    [anon_sym_POUNDload] = ACTIONS(6961),
    [anon_sym_POUNDtime] = ACTIONS(6961),
    [anon_sym_POUNDI] = ACTIONS(6961),
    [anon_sym_POUNDhelp] = ACTIONS(6961),
    [anon_sym_POUNDquit] = ACTIONS(6961),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(4159)] = {
    [sym_block_comment] = STATE(4159),
    [sym_compiler_directive_decl] = STATE(4159),
    [sym_fsi_directive_decl] = STATE(4159),
    [sym_preproc_line] = STATE(4159),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_GT_RBRACK] = ACTIONS(6869),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_RBRACK] = ACTIONS(6869),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_RBRACE] = ACTIONS(6869),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_to] = ACTIONS(6871),
    [anon_sym_downto] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_DOT_DOT2] = ACTIONS(6869),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(4160)] = {
    [sym_block_comment] = STATE(4160),
    [sym_compiler_directive_decl] = STATE(4160),
    [sym_fsi_directive_decl] = STATE(4160),
    [sym_preproc_line] = STATE(4160),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_module] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_open] = ACTIONS(6967),
    [anon_sym_type] = ACTIONS(6967),
    [anon_sym_LBRACK_LT] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_exception] = ACTIONS(6967),
    [anon_sym_extern] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6965),
    [anon_sym_POUNDload] = ACTIONS(6965),
    [anon_sym_POUNDtime] = ACTIONS(6965),
    [anon_sym_POUNDI] = ACTIONS(6965),
    [anon_sym_POUNDhelp] = ACTIONS(6965),
    [anon_sym_POUNDquit] = ACTIONS(6965),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(4161)] = {
    [sym_block_comment] = STATE(4161),
    [sym_compiler_directive_decl] = STATE(4161),
    [sym_fsi_directive_decl] = STATE(4161),
    [sym_preproc_line] = STATE(4161),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_module] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_open] = ACTIONS(6971),
    [anon_sym_type] = ACTIONS(6971),
    [anon_sym_LBRACK_LT] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_exception] = ACTIONS(6971),
    [anon_sym_extern] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6969),
    [anon_sym_POUNDload] = ACTIONS(6969),
    [anon_sym_POUNDtime] = ACTIONS(6969),
    [anon_sym_POUNDI] = ACTIONS(6969),
    [anon_sym_POUNDhelp] = ACTIONS(6969),
    [anon_sym_POUNDquit] = ACTIONS(6969),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(4162)] = {
    [sym_block_comment] = STATE(4162),
    [sym_compiler_directive_decl] = STATE(4162),
    [sym_fsi_directive_decl] = STATE(4162),
    [sym_preproc_line] = STATE(4162),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_module] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_open] = ACTIONS(6975),
    [anon_sym_type] = ACTIONS(6975),
    [anon_sym_LBRACK_LT] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_exception] = ACTIONS(6975),
    [anon_sym_extern] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6973),
    [anon_sym_POUNDload] = ACTIONS(6973),
    [anon_sym_POUNDtime] = ACTIONS(6973),
    [anon_sym_POUNDI] = ACTIONS(6973),
    [anon_sym_POUNDhelp] = ACTIONS(6973),
    [anon_sym_POUNDquit] = ACTIONS(6973),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(4163)] = {
    [sym_block_comment] = STATE(4163),
    [sym_compiler_directive_decl] = STATE(4163),
    [sym_fsi_directive_decl] = STATE(4163),
    [sym_preproc_line] = STATE(4163),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_module] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_open] = ACTIONS(6979),
    [anon_sym_type] = ACTIONS(6979),
    [anon_sym_LBRACK_LT] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_exception] = ACTIONS(6979),
    [anon_sym_extern] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6977),
    [anon_sym_POUNDload] = ACTIONS(6977),
    [anon_sym_POUNDtime] = ACTIONS(6977),
    [anon_sym_POUNDI] = ACTIONS(6977),
    [anon_sym_POUNDhelp] = ACTIONS(6977),
    [anon_sym_POUNDquit] = ACTIONS(6977),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(4164)] = {
    [sym_block_comment] = STATE(4164),
    [sym_compiler_directive_decl] = STATE(4164),
    [sym_fsi_directive_decl] = STATE(4164),
    [sym_preproc_line] = STATE(4164),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_module] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_open] = ACTIONS(6983),
    [anon_sym_type] = ACTIONS(6983),
    [anon_sym_LBRACK_LT] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_exception] = ACTIONS(6983),
    [anon_sym_extern] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6981),
    [anon_sym_POUNDload] = ACTIONS(6981),
    [anon_sym_POUNDtime] = ACTIONS(6981),
    [anon_sym_POUNDI] = ACTIONS(6981),
    [anon_sym_POUNDhelp] = ACTIONS(6981),
    [anon_sym_POUNDquit] = ACTIONS(6981),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(4165)] = {
    [sym_block_comment] = STATE(4165),
    [sym_compiler_directive_decl] = STATE(4165),
    [sym_fsi_directive_decl] = STATE(4165),
    [sym_preproc_line] = STATE(4165),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_module] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_open] = ACTIONS(6987),
    [anon_sym_type] = ACTIONS(6987),
    [anon_sym_LBRACK_LT] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_exception] = ACTIONS(6987),
    [anon_sym_extern] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6985),
    [anon_sym_POUNDload] = ACTIONS(6985),
    [anon_sym_POUNDtime] = ACTIONS(6985),
    [anon_sym_POUNDI] = ACTIONS(6985),
    [anon_sym_POUNDhelp] = ACTIONS(6985),
    [anon_sym_POUNDquit] = ACTIONS(6985),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(4166)] = {
    [sym_block_comment] = STATE(4166),
    [sym_compiler_directive_decl] = STATE(4166),
    [sym_fsi_directive_decl] = STATE(4166),
    [sym_preproc_line] = STATE(4166),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_module] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_open] = ACTIONS(6991),
    [anon_sym_type] = ACTIONS(6991),
    [anon_sym_LBRACK_LT] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_exception] = ACTIONS(6991),
    [anon_sym_extern] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6989),
    [anon_sym_POUNDload] = ACTIONS(6989),
    [anon_sym_POUNDtime] = ACTIONS(6989),
    [anon_sym_POUNDI] = ACTIONS(6989),
    [anon_sym_POUNDhelp] = ACTIONS(6989),
    [anon_sym_POUNDquit] = ACTIONS(6989),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(4167)] = {
    [sym_block_comment] = STATE(4167),
    [sym_compiler_directive_decl] = STATE(4167),
    [sym_fsi_directive_decl] = STATE(4167),
    [sym_preproc_line] = STATE(4167),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_module] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_open] = ACTIONS(6995),
    [anon_sym_type] = ACTIONS(6995),
    [anon_sym_LBRACK_LT] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_exception] = ACTIONS(6995),
    [anon_sym_extern] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6993),
    [anon_sym_POUNDload] = ACTIONS(6993),
    [anon_sym_POUNDtime] = ACTIONS(6993),
    [anon_sym_POUNDI] = ACTIONS(6993),
    [anon_sym_POUNDhelp] = ACTIONS(6993),
    [anon_sym_POUNDquit] = ACTIONS(6993),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(4168)] = {
    [sym_block_comment] = STATE(4168),
    [sym_compiler_directive_decl] = STATE(4168),
    [sym_fsi_directive_decl] = STATE(4168),
    [sym_preproc_line] = STATE(4168),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_module] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_open] = ACTIONS(6999),
    [anon_sym_type] = ACTIONS(6999),
    [anon_sym_LBRACK_LT] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_exception] = ACTIONS(6999),
    [anon_sym_extern] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6997),
    [anon_sym_POUNDload] = ACTIONS(6997),
    [anon_sym_POUNDtime] = ACTIONS(6997),
    [anon_sym_POUNDI] = ACTIONS(6997),
    [anon_sym_POUNDhelp] = ACTIONS(6997),
    [anon_sym_POUNDquit] = ACTIONS(6997),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(4169)] = {
    [sym_block_comment] = STATE(4169),
    [sym_compiler_directive_decl] = STATE(4169),
    [sym_fsi_directive_decl] = STATE(4169),
    [sym_preproc_line] = STATE(4169),
    [aux_sym_long_identifier_repeat1] = STATE(4170),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
    [sym__in] = ACTIONS(6358),
  },
  [STATE(4170)] = {
    [sym_block_comment] = STATE(4170),
    [sym_compiler_directive_decl] = STATE(4170),
    [sym_fsi_directive_decl] = STATE(4170),
    [sym_preproc_line] = STATE(4170),
    [aux_sym_long_identifier_repeat1] = STATE(4170),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8562),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
    [sym__in] = ACTIONS(6297),
  },
  [STATE(4171)] = {
    [sym_block_comment] = STATE(4171),
    [sym_compiler_directive_decl] = STATE(4171),
    [sym_fsi_directive_decl] = STATE(4171),
    [sym_preproc_line] = STATE(4171),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_module] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_open] = ACTIONS(7003),
    [anon_sym_type] = ACTIONS(7003),
    [anon_sym_LBRACK_LT] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_exception] = ACTIONS(7003),
    [anon_sym_extern] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7001),
    [anon_sym_POUNDload] = ACTIONS(7001),
    [anon_sym_POUNDtime] = ACTIONS(7001),
    [anon_sym_POUNDI] = ACTIONS(7001),
    [anon_sym_POUNDhelp] = ACTIONS(7001),
    [anon_sym_POUNDquit] = ACTIONS(7001),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(4172)] = {
    [sym_block_comment] = STATE(4172),
    [sym_compiler_directive_decl] = STATE(4172),
    [sym_fsi_directive_decl] = STATE(4172),
    [sym_preproc_line] = STATE(4172),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_module] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_open] = ACTIONS(7007),
    [anon_sym_type] = ACTIONS(7007),
    [anon_sym_LBRACK_LT] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_exception] = ACTIONS(7007),
    [anon_sym_extern] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7005),
    [anon_sym_POUNDload] = ACTIONS(7005),
    [anon_sym_POUNDtime] = ACTIONS(7005),
    [anon_sym_POUNDI] = ACTIONS(7005),
    [anon_sym_POUNDhelp] = ACTIONS(7005),
    [anon_sym_POUNDquit] = ACTIONS(7005),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(4173)] = {
    [sym_block_comment] = STATE(4173),
    [sym_compiler_directive_decl] = STATE(4173),
    [sym_fsi_directive_decl] = STATE(4173),
    [sym_preproc_line] = STATE(4173),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_module] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_open] = ACTIONS(7011),
    [anon_sym_type] = ACTIONS(7011),
    [anon_sym_LBRACK_LT] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_exception] = ACTIONS(7011),
    [anon_sym_extern] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7009),
    [anon_sym_POUNDload] = ACTIONS(7009),
    [anon_sym_POUNDtime] = ACTIONS(7009),
    [anon_sym_POUNDI] = ACTIONS(7009),
    [anon_sym_POUNDhelp] = ACTIONS(7009),
    [anon_sym_POUNDquit] = ACTIONS(7009),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(4174)] = {
    [sym_block_comment] = STATE(4174),
    [sym_compiler_directive_decl] = STATE(4174),
    [sym_fsi_directive_decl] = STATE(4174),
    [sym_preproc_line] = STATE(4174),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_module] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_open] = ACTIONS(7019),
    [anon_sym_type] = ACTIONS(7019),
    [anon_sym_LBRACK_LT] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_exception] = ACTIONS(7019),
    [anon_sym_extern] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7017),
    [anon_sym_POUNDload] = ACTIONS(7017),
    [anon_sym_POUNDtime] = ACTIONS(7017),
    [anon_sym_POUNDI] = ACTIONS(7017),
    [anon_sym_POUNDhelp] = ACTIONS(7017),
    [anon_sym_POUNDquit] = ACTIONS(7017),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(4175)] = {
    [sym_block_comment] = STATE(4175),
    [sym_compiler_directive_decl] = STATE(4175),
    [sym_fsi_directive_decl] = STATE(4175),
    [sym_preproc_line] = STATE(4175),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_module] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_open] = ACTIONS(7023),
    [anon_sym_type] = ACTIONS(7023),
    [anon_sym_LBRACK_LT] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_exception] = ACTIONS(7023),
    [anon_sym_extern] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7021),
    [anon_sym_POUNDload] = ACTIONS(7021),
    [anon_sym_POUNDtime] = ACTIONS(7021),
    [anon_sym_POUNDI] = ACTIONS(7021),
    [anon_sym_POUNDhelp] = ACTIONS(7021),
    [anon_sym_POUNDquit] = ACTIONS(7021),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(4176)] = {
    [sym_block_comment] = STATE(4176),
    [sym_compiler_directive_decl] = STATE(4176),
    [sym_fsi_directive_decl] = STATE(4176),
    [sym_preproc_line] = STATE(4176),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_module] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_open] = ACTIONS(7027),
    [anon_sym_type] = ACTIONS(7027),
    [anon_sym_LBRACK_LT] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_exception] = ACTIONS(7027),
    [anon_sym_extern] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7025),
    [anon_sym_POUNDload] = ACTIONS(7025),
    [anon_sym_POUNDtime] = ACTIONS(7025),
    [anon_sym_POUNDI] = ACTIONS(7025),
    [anon_sym_POUNDhelp] = ACTIONS(7025),
    [anon_sym_POUNDquit] = ACTIONS(7025),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(4177)] = {
    [sym_block_comment] = STATE(4177),
    [sym_compiler_directive_decl] = STATE(4177),
    [sym_fsi_directive_decl] = STATE(4177),
    [sym_preproc_line] = STATE(4177),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_module] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_open] = ACTIONS(7035),
    [anon_sym_type] = ACTIONS(7035),
    [anon_sym_LBRACK_LT] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_exception] = ACTIONS(7035),
    [anon_sym_extern] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7033),
    [anon_sym_POUNDload] = ACTIONS(7033),
    [anon_sym_POUNDtime] = ACTIONS(7033),
    [anon_sym_POUNDI] = ACTIONS(7033),
    [anon_sym_POUNDhelp] = ACTIONS(7033),
    [anon_sym_POUNDquit] = ACTIONS(7033),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(4178)] = {
    [sym_block_comment] = STATE(4178),
    [sym_compiler_directive_decl] = STATE(4178),
    [sym_fsi_directive_decl] = STATE(4178),
    [sym_preproc_line] = STATE(4178),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_module] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_open] = ACTIONS(7039),
    [anon_sym_type] = ACTIONS(7039),
    [anon_sym_LBRACK_LT] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_exception] = ACTIONS(7039),
    [anon_sym_extern] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7037),
    [anon_sym_POUNDload] = ACTIONS(7037),
    [anon_sym_POUNDtime] = ACTIONS(7037),
    [anon_sym_POUNDI] = ACTIONS(7037),
    [anon_sym_POUNDhelp] = ACTIONS(7037),
    [anon_sym_POUNDquit] = ACTIONS(7037),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(4179)] = {
    [sym_block_comment] = STATE(4179),
    [sym_compiler_directive_decl] = STATE(4179),
    [sym_fsi_directive_decl] = STATE(4179),
    [sym_preproc_line] = STATE(4179),
    [aux_sym_long_identifier_repeat1] = STATE(4180),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8565),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4180)] = {
    [sym_block_comment] = STATE(4180),
    [sym_compiler_directive_decl] = STATE(4180),
    [sym_fsi_directive_decl] = STATE(4180),
    [sym_preproc_line] = STATE(4180),
    [aux_sym_long_identifier_repeat1] = STATE(4180),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8567),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4181)] = {
    [sym_block_comment] = STATE(4181),
    [sym_compiler_directive_decl] = STATE(4181),
    [sym_fsi_directive_decl] = STATE(4181),
    [sym_preproc_line] = STATE(4181),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_module] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_open] = ACTIONS(7043),
    [anon_sym_type] = ACTIONS(7043),
    [anon_sym_LBRACK_LT] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_exception] = ACTIONS(7043),
    [anon_sym_extern] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7041),
    [anon_sym_POUNDload] = ACTIONS(7041),
    [anon_sym_POUNDtime] = ACTIONS(7041),
    [anon_sym_POUNDI] = ACTIONS(7041),
    [anon_sym_POUNDhelp] = ACTIONS(7041),
    [anon_sym_POUNDquit] = ACTIONS(7041),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(4182)] = {
    [sym_block_comment] = STATE(4182),
    [sym_compiler_directive_decl] = STATE(4182),
    [sym_fsi_directive_decl] = STATE(4182),
    [sym_preproc_line] = STATE(4182),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_module] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_open] = ACTIONS(7059),
    [anon_sym_type] = ACTIONS(7059),
    [anon_sym_LBRACK_LT] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_exception] = ACTIONS(7059),
    [anon_sym_extern] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7057),
    [anon_sym_POUNDload] = ACTIONS(7057),
    [anon_sym_POUNDtime] = ACTIONS(7057),
    [anon_sym_POUNDI] = ACTIONS(7057),
    [anon_sym_POUNDhelp] = ACTIONS(7057),
    [anon_sym_POUNDquit] = ACTIONS(7057),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(4183)] = {
    [sym_block_comment] = STATE(4183),
    [sym_compiler_directive_decl] = STATE(4183),
    [sym_fsi_directive_decl] = STATE(4183),
    [sym_preproc_line] = STATE(4183),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_GT_RBRACK] = ACTIONS(6873),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_RBRACK] = ACTIONS(6873),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_RBRACE] = ACTIONS(6873),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_to] = ACTIONS(6875),
    [anon_sym_downto] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_DOT_DOT2] = ACTIONS(6873),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(4184)] = {
    [sym_block_comment] = STATE(4184),
    [sym_compiler_directive_decl] = STATE(4184),
    [sym_fsi_directive_decl] = STATE(4184),
    [sym_preproc_line] = STATE(4184),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_GT_RBRACK] = ACTIONS(6830),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_RBRACK] = ACTIONS(6830),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_RBRACE] = ACTIONS(6830),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_to] = ACTIONS(6828),
    [anon_sym_downto] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_DOT_DOT2] = ACTIONS(6830),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(4185)] = {
    [sym_block_comment] = STATE(4185),
    [sym_compiler_directive_decl] = STATE(4185),
    [sym_fsi_directive_decl] = STATE(4185),
    [sym_preproc_line] = STATE(4185),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8332),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4186)] = {
    [sym_block_comment] = STATE(4186),
    [sym_compiler_directive_decl] = STATE(4186),
    [sym_fsi_directive_decl] = STATE(4186),
    [sym_preproc_line] = STATE(4186),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_DOT_DOT] = ACTIONS(6492),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_DOT_DOT2] = ACTIONS(6490),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__dedent] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4187)] = {
    [sym_block_comment] = STATE(4187),
    [sym_compiler_directive_decl] = STATE(4187),
    [sym_fsi_directive_decl] = STATE(4187),
    [sym_preproc_line] = STATE(4187),
    [sym_identifier] = ACTIONS(8447),
    [anon_sym_module] = ACTIONS(8447),
    [anon_sym_null] = ACTIONS(8447),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(8447),
    [anon_sym_open] = ACTIONS(8447),
    [anon_sym_type] = ACTIONS(8447),
    [anon_sym_LBRACK_LT] = ACTIONS(8449),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(8447),
    [anon_sym_let] = ACTIONS(8447),
    [anon_sym_let_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN] = ACTIONS(8447),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(8447),
    [anon_sym_LBRACK] = ACTIONS(8447),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8449),
    [anon_sym_LBRACE] = ACTIONS(8447),
    [anon_sym_LT_AT] = ACTIONS(8447),
    [anon_sym_LT_AT_AT] = ACTIONS(8447),
    [anon_sym_CARET] = ACTIONS(8447),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8449),
    [anon_sym_new] = ACTIONS(8447),
    [anon_sym_return] = ACTIONS(8447),
    [anon_sym_return_BANG] = ACTIONS(8449),
    [anon_sym_yield] = ACTIONS(8447),
    [anon_sym_yield_BANG] = ACTIONS(8449),
    [anon_sym_lazy] = ACTIONS(8447),
    [anon_sym_assert] = ACTIONS(8447),
    [anon_sym_upcast] = ACTIONS(8447),
    [anon_sym_downcast] = ACTIONS(8447),
    [anon_sym_fixed] = ACTIONS(8447),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(8447),
    [anon_sym_while] = ACTIONS(8447),
    [anon_sym_while_BANG] = ACTIONS(8449),
    [anon_sym_if] = ACTIONS(8447),
    [anon_sym_fun] = ACTIONS(8447),
    [anon_sym_try] = ACTIONS(8447),
    [anon_sym_match] = ACTIONS(8447),
    [anon_sym_match_BANG] = ACTIONS(8449),
    [anon_sym_function] = ACTIONS(8447),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(8447),
    [anon_sym_use_BANG] = ACTIONS(8449),
    [anon_sym_do_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(8447),
    [anon_sym_exception] = ACTIONS(8447),
    [anon_sym_extern] = ACTIONS(8447),
    [aux_sym_char_token1] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8447),
    [anon_sym_DQUOTE] = ACTIONS(8447),
    [anon_sym_AT_DQUOTE] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [sym_bool] = ACTIONS(8447),
    [sym_unit] = ACTIONS(8449),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8447),
    [sym_op_identifier] = ACTIONS(8447),
    [anon_sym_PLUS] = ACTIONS(8447),
    [anon_sym_DASH] = ACTIONS(8447),
    [anon_sym_PLUS_DOT] = ACTIONS(8447),
    [anon_sym_DASH_DOT] = ACTIONS(8447),
    [anon_sym_PERCENT] = ACTIONS(8447),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8447),
    [anon_sym_TILDE] = ACTIONS(8449),
    [aux_sym_prefix_op_token1] = ACTIONS(8447),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(8447),
    [sym_xint] = ACTIONS(8449),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8449),
    [anon_sym_POUNDload] = ACTIONS(8449),
    [anon_sym_POUNDtime] = ACTIONS(8449),
    [anon_sym_POUNDI] = ACTIONS(8449),
    [anon_sym_POUNDhelp] = ACTIONS(8449),
    [anon_sym_POUNDquit] = ACTIONS(8449),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8449),
    [anon_sym_POUNDendif] = ACTIONS(8449),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8449),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4188)] = {
    [sym_block_comment] = STATE(4188),
    [sym_compiler_directive_decl] = STATE(4188),
    [sym_fsi_directive_decl] = STATE(4188),
    [sym_preproc_line] = STATE(4188),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_DOT_DOT] = ACTIONS(6427),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_DOT_DOT2] = ACTIONS(6429),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__dedent] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4189)] = {
    [sym_block_comment] = STATE(4189),
    [sym_compiler_directive_decl] = STATE(4189),
    [sym_fsi_directive_decl] = STATE(4189),
    [sym_preproc_line] = STATE(4189),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_DOT_DOT] = ACTIONS(6389),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_DOT_DOT2] = ACTIONS(6387),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__dedent] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4190)] = {
    [sym_block_comment] = STATE(4190),
    [sym_compiler_directive_decl] = STATE(4190),
    [sym_fsi_directive_decl] = STATE(4190),
    [sym_preproc_line] = STATE(4190),
    [aux_sym_type_argument_repeat1] = STATE(4195),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8570),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4191)] = {
    [sym_block_comment] = STATE(4191),
    [sym_compiler_directive_decl] = STATE(4191),
    [sym_fsi_directive_decl] = STATE(4191),
    [sym_preproc_line] = STATE(4191),
    [aux_sym_compound_type_repeat1] = STATE(4192),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_DOT_DOT] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8268),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__dedent] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4192)] = {
    [sym_block_comment] = STATE(4192),
    [sym_compiler_directive_decl] = STATE(4192),
    [sym_fsi_directive_decl] = STATE(4192),
    [sym_preproc_line] = STATE(4192),
    [aux_sym_compound_type_repeat1] = STATE(4192),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8573),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4193)] = {
    [sym_attributes] = STATE(13289),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7173),
    [sym_access_modifier] = STATE(22403),
    [sym__class_function_or_value_defn] = STATE(7174),
    [aux_sym__type_extension_inner] = STATE(4627),
    [sym_type_extension_elements] = STATE(8537),
    [sym__type_defn_elements] = STATE(7174),
    [sym_interface_implementation] = STATE(7175),
    [sym_member_defn] = STATE(7175),
    [sym_additional_constr_defn] = STATE(7146),
    [sym_block_comment] = STATE(4193),
    [sym_compiler_directive_decl] = STATE(4193),
    [sym_fsi_directive_decl] = STATE(4193),
    [sym_preproc_line] = STATE(4193),
    [sym_preproc_if_in_class_definition] = STATE(7175),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8341),
    [anon_sym_module] = ACTIONS(8341),
    [anon_sym_null] = ACTIONS(8341),
    [anon_sym_begin] = ACTIONS(8341),
    [anon_sym_open] = ACTIONS(8341),
    [anon_sym_type] = ACTIONS(8341),
    [anon_sym_LBRACK_LT] = ACTIONS(8343),
    [anon_sym_do] = ACTIONS(8341),
    [anon_sym_and] = ACTIONS(8341),
    [anon_sym_let] = ACTIONS(8341),
    [anon_sym_let_BANG] = ACTIONS(8343),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8341),
    [anon_sym_AMP] = ACTIONS(8341),
    [anon_sym_LBRACK] = ACTIONS(8341),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8343),
    [anon_sym_LBRACE] = ACTIONS(8341),
    [anon_sym_LT_AT] = ACTIONS(8341),
    [anon_sym_LT_AT_AT] = ACTIONS(8343),
    [anon_sym_CARET] = ACTIONS(8343),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8343),
    [anon_sym_new] = ACTIONS(8341),
    [anon_sym_with] = ACTIONS(8576),
    [anon_sym_return] = ACTIONS(8341),
    [anon_sym_return_BANG] = ACTIONS(8343),
    [anon_sym_yield] = ACTIONS(8341),
    [anon_sym_yield_BANG] = ACTIONS(8343),
    [anon_sym_lazy] = ACTIONS(8341),
    [anon_sym_assert] = ACTIONS(8341),
    [anon_sym_upcast] = ACTIONS(8341),
    [anon_sym_downcast] = ACTIONS(8341),
    [anon_sym_fixed] = ACTIONS(8341),
    [anon_sym_for] = ACTIONS(8341),
    [anon_sym_while] = ACTIONS(8341),
    [anon_sym_while_BANG] = ACTIONS(8343),
    [anon_sym_if] = ACTIONS(8341),
    [anon_sym_fun] = ACTIONS(8341),
    [anon_sym_try] = ACTIONS(8341),
    [anon_sym_match] = ACTIONS(8341),
    [anon_sym_match_BANG] = ACTIONS(8343),
    [anon_sym_function] = ACTIONS(8341),
    [anon_sym_use] = ACTIONS(8341),
    [anon_sym_use_BANG] = ACTIONS(8343),
    [anon_sym_do_BANG] = ACTIONS(8343),
    [anon_sym_AMP_AMP] = ACTIONS(8343),
    [anon_sym_default] = ACTIONS(8578),
    [anon_sym_static] = ACTIONS(8580),
    [anon_sym_member] = ACTIONS(8582),
    [anon_sym_exception] = ACTIONS(8341),
    [anon_sym_interface] = ACTIONS(8584),
    [anon_sym_abstract] = ACTIONS(8586),
    [anon_sym_val] = ACTIONS(8588),
    [anon_sym_override] = ACTIONS(8578),
    [anon_sym_extern] = ACTIONS(8341),
    [aux_sym_char_token1] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8341),
    [anon_sym_DQUOTE] = ACTIONS(8341),
    [anon_sym_AT_DQUOTE] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [sym_bool] = ACTIONS(8341),
    [sym_unit] = ACTIONS(8343),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8341),
    [sym_op_identifier] = ACTIONS(8343),
    [anon_sym_PLUS] = ACTIONS(8341),
    [anon_sym_DASH] = ACTIONS(8341),
    [anon_sym_PLUS_DOT] = ACTIONS(8343),
    [anon_sym_DASH_DOT] = ACTIONS(8343),
    [anon_sym_PERCENT] = ACTIONS(8341),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8343),
    [anon_sym_TILDE] = ACTIONS(8343),
    [aux_sym_prefix_op_token1] = ACTIONS(8343),
    [sym_int] = ACTIONS(8341),
    [sym_xint] = ACTIONS(8343),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8343),
    [anon_sym_POUNDload] = ACTIONS(8343),
    [anon_sym_POUNDtime] = ACTIONS(8343),
    [anon_sym_POUNDI] = ACTIONS(8343),
    [anon_sym_POUNDhelp] = ACTIONS(8343),
    [anon_sym_POUNDquit] = ACTIONS(8343),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8343),
    [anon_sym_POUNDendif] = ACTIONS(8343),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8343),
  },
  [STATE(4194)] = {
    [sym_block_comment] = STATE(4194),
    [sym_compiler_directive_decl] = STATE(4194),
    [sym_fsi_directive_decl] = STATE(4194),
    [sym_preproc_line] = STATE(4194),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_GT_RBRACK] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_RBRACK] = ACTIONS(6614),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_RBRACE] = ACTIONS(6614),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_to] = ACTIONS(6612),
    [anon_sym_downto] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4195)] = {
    [sym_block_comment] = STATE(4195),
    [sym_compiler_directive_decl] = STATE(4195),
    [sym_fsi_directive_decl] = STATE(4195),
    [sym_preproc_line] = STATE(4195),
    [aux_sym_type_argument_repeat1] = STATE(4196),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_DOT_DOT] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__dedent] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4196)] = {
    [sym_block_comment] = STATE(4196),
    [sym_compiler_directive_decl] = STATE(4196),
    [sym_fsi_directive_decl] = STATE(4196),
    [sym_preproc_line] = STATE(4196),
    [aux_sym_type_argument_repeat1] = STATE(4196),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8590),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4197)] = {
    [sym_block_comment] = STATE(4197),
    [sym_compiler_directive_decl] = STATE(4197),
    [sym_fsi_directive_decl] = STATE(4197),
    [sym_preproc_line] = STATE(4197),
    [aux_sym_long_identifier_repeat1] = STATE(4200),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8593),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_DOT_DOT] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__dedent] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4198)] = {
    [sym_attributes] = STATE(13289),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7173),
    [sym_access_modifier] = STATE(22403),
    [sym__class_function_or_value_defn] = STATE(7174),
    [aux_sym__type_extension_inner] = STATE(4627),
    [sym_type_extension_elements] = STATE(8548),
    [sym__type_defn_elements] = STATE(7174),
    [sym_interface_implementation] = STATE(7175),
    [sym_member_defn] = STATE(7175),
    [sym_additional_constr_defn] = STATE(7146),
    [sym_block_comment] = STATE(4198),
    [sym_compiler_directive_decl] = STATE(4198),
    [sym_fsi_directive_decl] = STATE(4198),
    [sym_preproc_line] = STATE(4198),
    [sym_preproc_if_in_class_definition] = STATE(7175),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8242),
    [anon_sym_module] = ACTIONS(8242),
    [anon_sym_null] = ACTIONS(8242),
    [anon_sym_begin] = ACTIONS(8242),
    [anon_sym_open] = ACTIONS(8242),
    [anon_sym_type] = ACTIONS(8242),
    [anon_sym_LBRACK_LT] = ACTIONS(8244),
    [anon_sym_do] = ACTIONS(8242),
    [anon_sym_and] = ACTIONS(8242),
    [anon_sym_let] = ACTIONS(8242),
    [anon_sym_let_BANG] = ACTIONS(8244),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8242),
    [anon_sym_AMP] = ACTIONS(8242),
    [anon_sym_LBRACK] = ACTIONS(8242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8244),
    [anon_sym_LBRACE] = ACTIONS(8242),
    [anon_sym_LT_AT] = ACTIONS(8242),
    [anon_sym_LT_AT_AT] = ACTIONS(8244),
    [anon_sym_CARET] = ACTIONS(8244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8244),
    [anon_sym_new] = ACTIONS(8242),
    [anon_sym_with] = ACTIONS(8576),
    [anon_sym_return] = ACTIONS(8242),
    [anon_sym_return_BANG] = ACTIONS(8244),
    [anon_sym_yield] = ACTIONS(8242),
    [anon_sym_yield_BANG] = ACTIONS(8244),
    [anon_sym_lazy] = ACTIONS(8242),
    [anon_sym_assert] = ACTIONS(8242),
    [anon_sym_upcast] = ACTIONS(8242),
    [anon_sym_downcast] = ACTIONS(8242),
    [anon_sym_fixed] = ACTIONS(8242),
    [anon_sym_for] = ACTIONS(8242),
    [anon_sym_while] = ACTIONS(8242),
    [anon_sym_while_BANG] = ACTIONS(8244),
    [anon_sym_if] = ACTIONS(8242),
    [anon_sym_fun] = ACTIONS(8242),
    [anon_sym_try] = ACTIONS(8242),
    [anon_sym_match] = ACTIONS(8242),
    [anon_sym_match_BANG] = ACTIONS(8244),
    [anon_sym_function] = ACTIONS(8242),
    [anon_sym_use] = ACTIONS(8242),
    [anon_sym_use_BANG] = ACTIONS(8244),
    [anon_sym_do_BANG] = ACTIONS(8244),
    [anon_sym_AMP_AMP] = ACTIONS(8244),
    [anon_sym_default] = ACTIONS(8578),
    [anon_sym_static] = ACTIONS(8580),
    [anon_sym_member] = ACTIONS(8582),
    [anon_sym_exception] = ACTIONS(8242),
    [anon_sym_interface] = ACTIONS(8584),
    [anon_sym_abstract] = ACTIONS(8586),
    [anon_sym_val] = ACTIONS(8588),
    [anon_sym_override] = ACTIONS(8578),
    [anon_sym_extern] = ACTIONS(8242),
    [aux_sym_char_token1] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8242),
    [anon_sym_DQUOTE] = ACTIONS(8242),
    [anon_sym_AT_DQUOTE] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [sym_bool] = ACTIONS(8242),
    [sym_unit] = ACTIONS(8244),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8242),
    [sym_op_identifier] = ACTIONS(8244),
    [anon_sym_PLUS] = ACTIONS(8242),
    [anon_sym_DASH] = ACTIONS(8242),
    [anon_sym_PLUS_DOT] = ACTIONS(8244),
    [anon_sym_DASH_DOT] = ACTIONS(8244),
    [anon_sym_PERCENT] = ACTIONS(8242),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8244),
    [anon_sym_TILDE] = ACTIONS(8244),
    [aux_sym_prefix_op_token1] = ACTIONS(8244),
    [sym_int] = ACTIONS(8242),
    [sym_xint] = ACTIONS(8244),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8244),
    [anon_sym_POUNDload] = ACTIONS(8244),
    [anon_sym_POUNDtime] = ACTIONS(8244),
    [anon_sym_POUNDI] = ACTIONS(8244),
    [anon_sym_POUNDhelp] = ACTIONS(8244),
    [anon_sym_POUNDquit] = ACTIONS(8244),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8244),
    [anon_sym_POUNDendif] = ACTIONS(8244),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8244),
  },
  [STATE(4199)] = {
    [sym_elif_expression] = STATE(4629),
    [sym_block_comment] = STATE(4199),
    [sym_compiler_directive_decl] = STATE(4199),
    [sym_fsi_directive_decl] = STATE(4199),
    [sym_preproc_line] = STATE(4199),
    [aux_sym_if_expression_repeat1] = STATE(4199),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8595),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_DOT_DOT] = ACTIONS(6496),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [anon_sym_DOT_DOT2] = ACTIONS(6494),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__dedent] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4200)] = {
    [sym_block_comment] = STATE(4200),
    [sym_compiler_directive_decl] = STATE(4200),
    [sym_fsi_directive_decl] = STATE(4200),
    [sym_preproc_line] = STATE(4200),
    [aux_sym_long_identifier_repeat1] = STATE(4201),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8593),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4201)] = {
    [sym_block_comment] = STATE(4201),
    [sym_compiler_directive_decl] = STATE(4201),
    [sym_fsi_directive_decl] = STATE(4201),
    [sym_preproc_line] = STATE(4201),
    [aux_sym_long_identifier_repeat1] = STATE(4201),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8598),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4202)] = {
    [sym_block_comment] = STATE(4202),
    [sym_compiler_directive_decl] = STATE(4202),
    [sym_fsi_directive_decl] = STATE(4202),
    [sym_preproc_line] = STATE(4202),
    [sym_identifier] = ACTIONS(8447),
    [anon_sym_module] = ACTIONS(8447),
    [anon_sym_null] = ACTIONS(8447),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(8447),
    [anon_sym_open] = ACTIONS(8447),
    [anon_sym_type] = ACTIONS(8447),
    [anon_sym_LBRACK_LT] = ACTIONS(8449),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(8447),
    [anon_sym_let] = ACTIONS(8447),
    [anon_sym_let_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN] = ACTIONS(8447),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(8447),
    [anon_sym_LBRACK] = ACTIONS(8447),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8449),
    [anon_sym_LBRACE] = ACTIONS(8447),
    [anon_sym_LT_AT] = ACTIONS(8447),
    [anon_sym_LT_AT_AT] = ACTIONS(8447),
    [anon_sym_CARET] = ACTIONS(8447),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8449),
    [anon_sym_new] = ACTIONS(8447),
    [anon_sym_return] = ACTIONS(8447),
    [anon_sym_return_BANG] = ACTIONS(8449),
    [anon_sym_yield] = ACTIONS(8447),
    [anon_sym_yield_BANG] = ACTIONS(8449),
    [anon_sym_lazy] = ACTIONS(8447),
    [anon_sym_assert] = ACTIONS(8447),
    [anon_sym_upcast] = ACTIONS(8447),
    [anon_sym_downcast] = ACTIONS(8447),
    [anon_sym_fixed] = ACTIONS(8447),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(8447),
    [anon_sym_while] = ACTIONS(8447),
    [anon_sym_while_BANG] = ACTIONS(8449),
    [anon_sym_if] = ACTIONS(8447),
    [anon_sym_fun] = ACTIONS(8447),
    [anon_sym_try] = ACTIONS(8447),
    [anon_sym_match] = ACTIONS(8447),
    [anon_sym_match_BANG] = ACTIONS(8449),
    [anon_sym_function] = ACTIONS(8447),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(8447),
    [anon_sym_use_BANG] = ACTIONS(8449),
    [anon_sym_do_BANG] = ACTIONS(8449),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(8447),
    [anon_sym_exception] = ACTIONS(8447),
    [anon_sym_extern] = ACTIONS(8447),
    [aux_sym_char_token1] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8447),
    [anon_sym_DQUOTE] = ACTIONS(8447),
    [anon_sym_AT_DQUOTE] = ACTIONS(8449),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8449),
    [sym_bool] = ACTIONS(8447),
    [sym_unit] = ACTIONS(8449),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8447),
    [sym_op_identifier] = ACTIONS(8447),
    [anon_sym_PLUS] = ACTIONS(8447),
    [anon_sym_DASH] = ACTIONS(8447),
    [anon_sym_PLUS_DOT] = ACTIONS(8447),
    [anon_sym_DASH_DOT] = ACTIONS(8447),
    [anon_sym_PERCENT] = ACTIONS(8447),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8447),
    [anon_sym_TILDE] = ACTIONS(8449),
    [aux_sym_prefix_op_token1] = ACTIONS(8447),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(8447),
    [sym_xint] = ACTIONS(8449),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8449),
    [anon_sym_POUNDload] = ACTIONS(8449),
    [anon_sym_POUNDtime] = ACTIONS(8449),
    [anon_sym_POUNDI] = ACTIONS(8449),
    [anon_sym_POUNDhelp] = ACTIONS(8449),
    [anon_sym_POUNDquit] = ACTIONS(8449),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8449),
    [anon_sym_POUNDendif] = ACTIONS(8449),
    [sym__newline] = ACTIONS(8449),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8449),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4203)] = {
    [sym_block_comment] = STATE(4203),
    [sym_compiler_directive_decl] = STATE(4203),
    [sym_fsi_directive_decl] = STATE(4203),
    [sym_preproc_line] = STATE(4203),
    [sym_identifier] = ACTIONS(8526),
    [anon_sym_module] = ACTIONS(8529),
    [anon_sym_null] = ACTIONS(8526),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(8526),
    [anon_sym_open] = ACTIONS(8529),
    [anon_sym_type] = ACTIONS(8529),
    [anon_sym_LBRACK_LT] = ACTIONS(8531),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(8526),
    [anon_sym_let] = ACTIONS(8526),
    [anon_sym_let_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN] = ACTIONS(8526),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(8526),
    [anon_sym_LBRACK] = ACTIONS(8526),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8533),
    [anon_sym_LBRACE] = ACTIONS(8526),
    [anon_sym_LT_AT] = ACTIONS(8526),
    [anon_sym_LT_AT_AT] = ACTIONS(8526),
    [anon_sym_CARET] = ACTIONS(8526),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8533),
    [anon_sym_new] = ACTIONS(8526),
    [anon_sym_return] = ACTIONS(8526),
    [anon_sym_return_BANG] = ACTIONS(8533),
    [anon_sym_yield] = ACTIONS(8526),
    [anon_sym_yield_BANG] = ACTIONS(8533),
    [anon_sym_lazy] = ACTIONS(8526),
    [anon_sym_assert] = ACTIONS(8526),
    [anon_sym_upcast] = ACTIONS(8526),
    [anon_sym_downcast] = ACTIONS(8526),
    [anon_sym_fixed] = ACTIONS(8526),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(8526),
    [anon_sym_while] = ACTIONS(8526),
    [anon_sym_while_BANG] = ACTIONS(8533),
    [anon_sym_if] = ACTIONS(8526),
    [anon_sym_fun] = ACTIONS(8526),
    [anon_sym_try] = ACTIONS(8526),
    [anon_sym_match] = ACTIONS(8526),
    [anon_sym_match_BANG] = ACTIONS(8533),
    [anon_sym_function] = ACTIONS(8526),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(8526),
    [anon_sym_use_BANG] = ACTIONS(8533),
    [anon_sym_do_BANG] = ACTIONS(8533),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(8526),
    [anon_sym_exception] = ACTIONS(8529),
    [anon_sym_extern] = ACTIONS(8529),
    [aux_sym_char_token1] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8526),
    [anon_sym_DQUOTE] = ACTIONS(8526),
    [anon_sym_AT_DQUOTE] = ACTIONS(8533),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8533),
    [sym_bool] = ACTIONS(8526),
    [sym_unit] = ACTIONS(8533),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8526),
    [sym_op_identifier] = ACTIONS(8526),
    [anon_sym_PLUS] = ACTIONS(8526),
    [anon_sym_DASH] = ACTIONS(8526),
    [anon_sym_PLUS_DOT] = ACTIONS(8526),
    [anon_sym_DASH_DOT] = ACTIONS(8526),
    [anon_sym_PERCENT] = ACTIONS(8526),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8526),
    [anon_sym_TILDE] = ACTIONS(8533),
    [aux_sym_prefix_op_token1] = ACTIONS(8526),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(8526),
    [sym_xint] = ACTIONS(8533),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8531),
    [anon_sym_POUNDload] = ACTIONS(8531),
    [anon_sym_POUNDtime] = ACTIONS(8531),
    [anon_sym_POUNDI] = ACTIONS(8531),
    [anon_sym_POUNDhelp] = ACTIONS(8531),
    [anon_sym_POUNDquit] = ACTIONS(8531),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8533),
    [anon_sym_POUNDendif] = ACTIONS(8533),
    [sym__newline] = ACTIONS(8533),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8533),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4204)] = {
    [sym_block_comment] = STATE(4204),
    [sym_compiler_directive_decl] = STATE(4204),
    [sym_fsi_directive_decl] = STATE(4204),
    [sym_preproc_line] = STATE(4204),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6331),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_DOT_DOT2] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4205)] = {
    [sym_block_comment] = STATE(4205),
    [sym_compiler_directive_decl] = STATE(4205),
    [sym_fsi_directive_decl] = STATE(4205),
    [sym_preproc_line] = STATE(4205),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_DOT_DOT] = ACTIONS(6420),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_DOT_DOT2] = ACTIONS(6418),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__dedent] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4206)] = {
    [sym_block_comment] = STATE(4206),
    [sym_compiler_directive_decl] = STATE(4206),
    [sym_fsi_directive_decl] = STATE(4206),
    [sym_preproc_line] = STATE(4206),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_GT_RBRACK] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_RBRACK] = ACTIONS(6805),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_RBRACE] = ACTIONS(6805),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_to] = ACTIONS(6807),
    [anon_sym_downto] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_DOT_DOT2] = ACTIONS(6805),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4207)] = {
    [sym__else_expression] = STATE(5784),
    [sym_elif_expression] = STATE(4842),
    [sym_block_comment] = STATE(4207),
    [sym_compiler_directive_decl] = STATE(4207),
    [sym_fsi_directive_decl] = STATE(4207),
    [sym_preproc_line] = STATE(4207),
    [aux_sym_if_expression_repeat1] = STATE(4209),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_as] = ACTIONS(6319),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8601),
    [anon_sym_elif] = ACTIONS(8603),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4208)] = {
    [sym_block_comment] = STATE(4208),
    [sym_compiler_directive_decl] = STATE(4208),
    [sym_fsi_directive_decl] = STATE(4208),
    [sym_preproc_line] = STATE(4208),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6312),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_DOT_DOT2] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4209)] = {
    [sym__else_expression] = STATE(5809),
    [sym_elif_expression] = STATE(4842),
    [sym_block_comment] = STATE(4209),
    [sym_compiler_directive_decl] = STATE(4209),
    [sym_fsi_directive_decl] = STATE(4209),
    [sym_preproc_line] = STATE(4209),
    [aux_sym_if_expression_repeat1] = STATE(4348),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_as] = ACTIONS(6329),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8601),
    [anon_sym_elif] = ACTIONS(8603),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4210)] = {
    [sym_block_comment] = STATE(4210),
    [sym_compiler_directive_decl] = STATE(4210),
    [sym_fsi_directive_decl] = STATE(4210),
    [sym_preproc_line] = STATE(4210),
    [aux_sym_long_identifier_repeat1] = STATE(4277),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_GT_RBRACK] = ACTIONS(6628),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8605),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4211)] = {
    [sym_block_comment] = STATE(4211),
    [sym_compiler_directive_decl] = STATE(4211),
    [sym_fsi_directive_decl] = STATE(4211),
    [sym_preproc_line] = STATE(4211),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_GT_RBRACK] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_RBRACK] = ACTIONS(6297),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_RBRACE] = ACTIONS(6297),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_to] = ACTIONS(6295),
    [anon_sym_downto] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4212)] = {
    [sym__else_expression] = STATE(5322),
    [sym_elif_expression] = STATE(4886),
    [sym_block_comment] = STATE(4212),
    [sym_compiler_directive_decl] = STATE(4212),
    [sym_fsi_directive_decl] = STATE(4212),
    [sym_preproc_line] = STATE(4212),
    [aux_sym_if_expression_repeat1] = STATE(4213),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8609),
    [anon_sym_elif] = ACTIONS(8611),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [anon_sym_POUNDelse] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4213)] = {
    [sym__else_expression] = STATE(5334),
    [sym_elif_expression] = STATE(4886),
    [sym_block_comment] = STATE(4213),
    [sym_compiler_directive_decl] = STATE(4213),
    [sym_fsi_directive_decl] = STATE(4213),
    [sym_preproc_line] = STATE(4213),
    [aux_sym_if_expression_repeat1] = STATE(4358),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8609),
    [anon_sym_elif] = ACTIONS(8611),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [anon_sym_POUNDelse] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4214)] = {
    [sym_attributes] = STATE(13407),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4597),
    [sym_type_extension_elements] = STATE(8362),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4214),
    [sym_compiler_directive_decl] = STATE(4214),
    [sym_fsi_directive_decl] = STATE(4214),
    [sym_preproc_line] = STATE(4214),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8244),
    [sym_identifier] = ACTIONS(8242),
    [anon_sym_module] = ACTIONS(8242),
    [anon_sym_null] = ACTIONS(8242),
    [anon_sym_begin] = ACTIONS(8242),
    [anon_sym_open] = ACTIONS(8242),
    [anon_sym_type] = ACTIONS(8242),
    [anon_sym_LBRACK_LT] = ACTIONS(8244),
    [anon_sym_do] = ACTIONS(8242),
    [anon_sym_and] = ACTIONS(8242),
    [anon_sym_let] = ACTIONS(8242),
    [anon_sym_let_BANG] = ACTIONS(8244),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8242),
    [anon_sym_AMP] = ACTIONS(8242),
    [anon_sym_LBRACK] = ACTIONS(8242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8244),
    [anon_sym_LBRACE] = ACTIONS(8242),
    [anon_sym_LT_AT] = ACTIONS(8242),
    [anon_sym_LT_AT_AT] = ACTIONS(8244),
    [anon_sym_CARET] = ACTIONS(8244),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8244),
    [anon_sym_new] = ACTIONS(8242),
    [anon_sym_with] = ACTIONS(8508),
    [anon_sym_return] = ACTIONS(8242),
    [anon_sym_return_BANG] = ACTIONS(8244),
    [anon_sym_yield] = ACTIONS(8242),
    [anon_sym_yield_BANG] = ACTIONS(8244),
    [anon_sym_lazy] = ACTIONS(8242),
    [anon_sym_assert] = ACTIONS(8242),
    [anon_sym_upcast] = ACTIONS(8242),
    [anon_sym_downcast] = ACTIONS(8242),
    [anon_sym_fixed] = ACTIONS(8242),
    [anon_sym_for] = ACTIONS(8242),
    [anon_sym_while] = ACTIONS(8242),
    [anon_sym_while_BANG] = ACTIONS(8244),
    [anon_sym_if] = ACTIONS(8242),
    [anon_sym_fun] = ACTIONS(8242),
    [anon_sym_try] = ACTIONS(8242),
    [anon_sym_match] = ACTIONS(8242),
    [anon_sym_match_BANG] = ACTIONS(8244),
    [anon_sym_function] = ACTIONS(8242),
    [anon_sym_use] = ACTIONS(8242),
    [anon_sym_use_BANG] = ACTIONS(8244),
    [anon_sym_do_BANG] = ACTIONS(8244),
    [anon_sym_AMP_AMP] = ACTIONS(8244),
    [anon_sym_default] = ACTIONS(8510),
    [anon_sym_static] = ACTIONS(8512),
    [anon_sym_member] = ACTIONS(8613),
    [anon_sym_exception] = ACTIONS(8242),
    [anon_sym_interface] = ACTIONS(8516),
    [anon_sym_abstract] = ACTIONS(8518),
    [anon_sym_val] = ACTIONS(8520),
    [anon_sym_override] = ACTIONS(8510),
    [anon_sym_extern] = ACTIONS(8242),
    [aux_sym_char_token1] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8242),
    [anon_sym_DQUOTE] = ACTIONS(8242),
    [anon_sym_AT_DQUOTE] = ACTIONS(8244),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8244),
    [sym_bool] = ACTIONS(8242),
    [sym_unit] = ACTIONS(8244),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8242),
    [sym_op_identifier] = ACTIONS(8244),
    [anon_sym_PLUS] = ACTIONS(8242),
    [anon_sym_DASH] = ACTIONS(8242),
    [anon_sym_PLUS_DOT] = ACTIONS(8244),
    [anon_sym_DASH_DOT] = ACTIONS(8244),
    [anon_sym_PERCENT] = ACTIONS(8242),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8244),
    [anon_sym_TILDE] = ACTIONS(8244),
    [aux_sym_prefix_op_token1] = ACTIONS(8244),
    [sym_int] = ACTIONS(8242),
    [sym_xint] = ACTIONS(8244),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8244),
    [anon_sym_POUNDload] = ACTIONS(8244),
    [anon_sym_POUNDtime] = ACTIONS(8244),
    [anon_sym_POUNDI] = ACTIONS(8244),
    [anon_sym_POUNDhelp] = ACTIONS(8244),
    [anon_sym_POUNDquit] = ACTIONS(8244),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8244),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8244),
  },
  [STATE(4215)] = {
    [sym_block_comment] = STATE(4215),
    [sym_compiler_directive_decl] = STATE(4215),
    [sym_fsi_directive_decl] = STATE(4215),
    [sym_preproc_line] = STATE(4215),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_GT_RBRACK] = ACTIONS(6752),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_RBRACK] = ACTIONS(6752),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_RBRACE] = ACTIONS(6752),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_to] = ACTIONS(6750),
    [anon_sym_downto] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_DOT_DOT2] = ACTIONS(6752),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(4216)] = {
    [sym_block_comment] = STATE(4216),
    [sym_compiler_directive_decl] = STATE(4216),
    [sym_fsi_directive_decl] = STATE(4216),
    [sym_preproc_line] = STATE(4216),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_GT_RBRACK] = ACTIONS(6786),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_RBRACK] = ACTIONS(6786),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_RBRACE] = ACTIONS(6786),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_to] = ACTIONS(6784),
    [anon_sym_downto] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(8615),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_DOT_DOT2] = ACTIONS(6786),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(4217)] = {
    [sym_block_comment] = STATE(4217),
    [sym_compiler_directive_decl] = STATE(4217),
    [sym_fsi_directive_decl] = STATE(4217),
    [sym_preproc_line] = STATE(4217),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_GT_RBRACK] = ACTIONS(6792),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_RBRACK] = ACTIONS(6792),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_RBRACE] = ACTIONS(6792),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_to] = ACTIONS(6790),
    [anon_sym_downto] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_DOT_DOT2] = ACTIONS(6792),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(4218)] = {
    [sym_block_comment] = STATE(4218),
    [sym_compiler_directive_decl] = STATE(4218),
    [sym_fsi_directive_decl] = STATE(4218),
    [sym_preproc_line] = STATE(4218),
    [aux_sym_type_argument_repeat1] = STATE(4218),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8617),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4219)] = {
    [sym_attributes] = STATE(13274),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6891),
    [sym_access_modifier] = STATE(22393),
    [sym__class_function_or_value_defn] = STATE(6892),
    [aux_sym__type_extension_inner] = STATE(4222),
    [sym__type_defn_elements] = STATE(6892),
    [sym_interface_implementation] = STATE(6893),
    [sym_member_defn] = STATE(6893),
    [sym_additional_constr_defn] = STATE(6918),
    [sym_block_comment] = STATE(4219),
    [sym_compiler_directive_decl] = STATE(4219),
    [sym_fsi_directive_decl] = STATE(4219),
    [sym_preproc_line] = STATE(4219),
    [sym_preproc_if_in_class_definition] = STATE(6893),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8620),
    [anon_sym_module] = ACTIONS(8620),
    [anon_sym_null] = ACTIONS(8620),
    [anon_sym_begin] = ACTIONS(8620),
    [anon_sym_open] = ACTIONS(8620),
    [anon_sym_type] = ACTIONS(8620),
    [anon_sym_LBRACK_LT] = ACTIONS(8622),
    [anon_sym_do] = ACTIONS(8620),
    [anon_sym_and] = ACTIONS(8620),
    [anon_sym_let] = ACTIONS(8620),
    [anon_sym_let_BANG] = ACTIONS(8622),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8620),
    [anon_sym_AMP] = ACTIONS(8620),
    [anon_sym_LBRACK] = ACTIONS(8620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8622),
    [anon_sym_LBRACE] = ACTIONS(8620),
    [anon_sym_LT_AT] = ACTIONS(8620),
    [anon_sym_LT_AT_AT] = ACTIONS(8622),
    [anon_sym_CARET] = ACTIONS(8622),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8622),
    [anon_sym_new] = ACTIONS(8620),
    [anon_sym_return] = ACTIONS(8620),
    [anon_sym_return_BANG] = ACTIONS(8622),
    [anon_sym_yield] = ACTIONS(8620),
    [anon_sym_yield_BANG] = ACTIONS(8622),
    [anon_sym_lazy] = ACTIONS(8620),
    [anon_sym_assert] = ACTIONS(8620),
    [anon_sym_upcast] = ACTIONS(8620),
    [anon_sym_downcast] = ACTIONS(8620),
    [anon_sym_fixed] = ACTIONS(8620),
    [anon_sym_for] = ACTIONS(8620),
    [anon_sym_while] = ACTIONS(8620),
    [anon_sym_while_BANG] = ACTIONS(8622),
    [anon_sym_if] = ACTIONS(8620),
    [anon_sym_fun] = ACTIONS(8620),
    [anon_sym_try] = ACTIONS(8620),
    [anon_sym_match] = ACTIONS(8620),
    [anon_sym_match_BANG] = ACTIONS(8622),
    [anon_sym_function] = ACTIONS(8620),
    [anon_sym_use] = ACTIONS(8620),
    [anon_sym_use_BANG] = ACTIONS(8622),
    [anon_sym_do_BANG] = ACTIONS(8622),
    [anon_sym_AMP_AMP] = ACTIONS(8622),
    [anon_sym_default] = ACTIONS(8250),
    [anon_sym_static] = ACTIONS(8252),
    [anon_sym_member] = ACTIONS(8254),
    [anon_sym_exception] = ACTIONS(8620),
    [anon_sym_interface] = ACTIONS(8256),
    [anon_sym_abstract] = ACTIONS(8258),
    [anon_sym_val] = ACTIONS(8260),
    [anon_sym_override] = ACTIONS(8250),
    [anon_sym_extern] = ACTIONS(8620),
    [aux_sym_char_token1] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8620),
    [anon_sym_DQUOTE] = ACTIONS(8620),
    [anon_sym_AT_DQUOTE] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [sym_bool] = ACTIONS(8620),
    [sym_unit] = ACTIONS(8622),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8620),
    [sym_op_identifier] = ACTIONS(8622),
    [anon_sym_PLUS] = ACTIONS(8620),
    [anon_sym_DASH] = ACTIONS(8620),
    [anon_sym_PLUS_DOT] = ACTIONS(8622),
    [anon_sym_DASH_DOT] = ACTIONS(8622),
    [anon_sym_PERCENT] = ACTIONS(8620),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8622),
    [anon_sym_TILDE] = ACTIONS(8622),
    [aux_sym_prefix_op_token1] = ACTIONS(8622),
    [sym_int] = ACTIONS(8620),
    [sym_xint] = ACTIONS(8622),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8622),
    [anon_sym_POUNDload] = ACTIONS(8622),
    [anon_sym_POUNDtime] = ACTIONS(8622),
    [anon_sym_POUNDI] = ACTIONS(8622),
    [anon_sym_POUNDhelp] = ACTIONS(8622),
    [anon_sym_POUNDquit] = ACTIONS(8622),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8622),
    [anon_sym_POUNDendif] = ACTIONS(8622),
    [anon_sym_POUNDelse] = ACTIONS(8622),
    [sym__newline] = ACTIONS(8622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8622),
  },
  [STATE(4220)] = {
    [sym_block_comment] = STATE(4220),
    [sym_compiler_directive_decl] = STATE(4220),
    [sym_fsi_directive_decl] = STATE(4220),
    [sym_preproc_line] = STATE(4220),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_GT_RBRACK] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_RBRACK] = ACTIONS(6599),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_RBRACE] = ACTIONS(6599),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_to] = ACTIONS(6597),
    [anon_sym_downto] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8624),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4221)] = {
    [sym_block_comment] = STATE(4221),
    [sym_compiler_directive_decl] = STATE(4221),
    [sym_fsi_directive_decl] = STATE(4221),
    [sym_preproc_line] = STATE(4221),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_GT_RBRACK] = ACTIONS(6599),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_RBRACK] = ACTIONS(6599),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_RBRACE] = ACTIONS(6599),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_to] = ACTIONS(6597),
    [anon_sym_downto] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8624),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_DOT_DOT2] = ACTIONS(6599),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4222)] = {
    [sym_attributes] = STATE(13274),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6891),
    [sym_access_modifier] = STATE(22393),
    [sym__class_function_or_value_defn] = STATE(6892),
    [aux_sym__type_extension_inner] = STATE(4222),
    [sym__type_defn_elements] = STATE(6892),
    [sym_interface_implementation] = STATE(6893),
    [sym_member_defn] = STATE(6893),
    [sym_additional_constr_defn] = STATE(6918),
    [sym_block_comment] = STATE(4222),
    [sym_compiler_directive_decl] = STATE(4222),
    [sym_fsi_directive_decl] = STATE(4222),
    [sym_preproc_line] = STATE(4222),
    [sym_preproc_if_in_class_definition] = STATE(6893),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8626),
    [anon_sym_module] = ACTIONS(8626),
    [anon_sym_null] = ACTIONS(8626),
    [anon_sym_begin] = ACTIONS(8626),
    [anon_sym_open] = ACTIONS(8626),
    [anon_sym_type] = ACTIONS(8626),
    [anon_sym_LBRACK_LT] = ACTIONS(8628),
    [anon_sym_do] = ACTIONS(8631),
    [anon_sym_and] = ACTIONS(8626),
    [anon_sym_let] = ACTIONS(8634),
    [anon_sym_let_BANG] = ACTIONS(8637),
    [aux_sym_access_modifier_token1] = ACTIONS(8640),
    [anon_sym_LPAREN] = ACTIONS(8626),
    [anon_sym_AMP] = ACTIONS(8626),
    [anon_sym_LBRACK] = ACTIONS(8626),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8643),
    [anon_sym_LBRACE] = ACTIONS(8626),
    [anon_sym_LT_AT] = ACTIONS(8626),
    [anon_sym_LT_AT_AT] = ACTIONS(8643),
    [anon_sym_CARET] = ACTIONS(8643),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8643),
    [anon_sym_new] = ACTIONS(8645),
    [anon_sym_return] = ACTIONS(8626),
    [anon_sym_return_BANG] = ACTIONS(8643),
    [anon_sym_yield] = ACTIONS(8626),
    [anon_sym_yield_BANG] = ACTIONS(8643),
    [anon_sym_lazy] = ACTIONS(8626),
    [anon_sym_assert] = ACTIONS(8626),
    [anon_sym_upcast] = ACTIONS(8626),
    [anon_sym_downcast] = ACTIONS(8626),
    [anon_sym_fixed] = ACTIONS(8626),
    [anon_sym_for] = ACTIONS(8626),
    [anon_sym_while] = ACTIONS(8626),
    [anon_sym_while_BANG] = ACTIONS(8643),
    [anon_sym_if] = ACTIONS(8626),
    [anon_sym_fun] = ACTIONS(8626),
    [anon_sym_try] = ACTIONS(8626),
    [anon_sym_match] = ACTIONS(8626),
    [anon_sym_match_BANG] = ACTIONS(8643),
    [anon_sym_function] = ACTIONS(8626),
    [anon_sym_use] = ACTIONS(8626),
    [anon_sym_use_BANG] = ACTIONS(8643),
    [anon_sym_do_BANG] = ACTIONS(8643),
    [anon_sym_AMP_AMP] = ACTIONS(8643),
    [anon_sym_default] = ACTIONS(8648),
    [anon_sym_static] = ACTIONS(8651),
    [anon_sym_member] = ACTIONS(8654),
    [anon_sym_exception] = ACTIONS(8626),
    [anon_sym_interface] = ACTIONS(8657),
    [anon_sym_abstract] = ACTIONS(8660),
    [anon_sym_val] = ACTIONS(8663),
    [anon_sym_override] = ACTIONS(8648),
    [anon_sym_extern] = ACTIONS(8626),
    [aux_sym_char_token1] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8626),
    [anon_sym_DQUOTE] = ACTIONS(8626),
    [anon_sym_AT_DQUOTE] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [sym_bool] = ACTIONS(8626),
    [sym_unit] = ACTIONS(8643),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8626),
    [sym_op_identifier] = ACTIONS(8643),
    [anon_sym_PLUS] = ACTIONS(8626),
    [anon_sym_DASH] = ACTIONS(8626),
    [anon_sym_PLUS_DOT] = ACTIONS(8643),
    [anon_sym_DASH_DOT] = ACTIONS(8643),
    [anon_sym_PERCENT] = ACTIONS(8626),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8643),
    [anon_sym_TILDE] = ACTIONS(8643),
    [aux_sym_prefix_op_token1] = ACTIONS(8643),
    [sym_int] = ACTIONS(8626),
    [sym_xint] = ACTIONS(8643),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8643),
    [anon_sym_POUNDload] = ACTIONS(8643),
    [anon_sym_POUNDtime] = ACTIONS(8643),
    [anon_sym_POUNDI] = ACTIONS(8643),
    [anon_sym_POUNDhelp] = ACTIONS(8643),
    [anon_sym_POUNDquit] = ACTIONS(8643),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8666),
    [anon_sym_POUNDendif] = ACTIONS(8643),
    [anon_sym_POUNDelse] = ACTIONS(8643),
    [sym__newline] = ACTIONS(8643),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8643),
  },
  [STATE(4223)] = {
    [sym_block_comment] = STATE(4223),
    [sym_compiler_directive_decl] = STATE(4223),
    [sym_fsi_directive_decl] = STATE(4223),
    [sym_preproc_line] = STATE(4223),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_GT_RBRACK] = ACTIONS(6766),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_RBRACK] = ACTIONS(6766),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_RBRACE] = ACTIONS(6766),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_to] = ACTIONS(6768),
    [anon_sym_downto] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(8669),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_DOT_DOT2] = ACTIONS(6766),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(4224)] = {
    [sym_block_comment] = STATE(4224),
    [sym_compiler_directive_decl] = STATE(4224),
    [sym_fsi_directive_decl] = STATE(4224),
    [sym_preproc_line] = STATE(4224),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_GT_RBRACK] = ACTIONS(6832),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_RBRACK] = ACTIONS(6832),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_RBRACE] = ACTIONS(6832),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_to] = ACTIONS(6834),
    [anon_sym_downto] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_DOT_DOT2] = ACTIONS(6832),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4225)] = {
    [sym_block_comment] = STATE(4225),
    [sym_compiler_directive_decl] = STATE(4225),
    [sym_fsi_directive_decl] = STATE(4225),
    [sym_preproc_line] = STATE(4225),
    [aux_sym_long_identifier_repeat1] = STATE(4179),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8671),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4226)] = {
    [sym_block_comment] = STATE(4226),
    [sym_compiler_directive_decl] = STATE(4226),
    [sym_fsi_directive_decl] = STATE(4226),
    [sym_preproc_line] = STATE(4226),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_as] = ACTIONS(6513),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8675),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4227)] = {
    [sym__else_expression] = STATE(5567),
    [sym_elif_expression] = STATE(4801),
    [sym_block_comment] = STATE(4227),
    [sym_compiler_directive_decl] = STATE(4227),
    [sym_fsi_directive_decl] = STATE(4227),
    [sym_preproc_line] = STATE(4227),
    [aux_sym_if_expression_repeat1] = STATE(4228),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8677),
    [anon_sym_elif] = ACTIONS(8679),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_DOT_DOT] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__dedent] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4228)] = {
    [sym__else_expression] = STATE(5580),
    [sym_elif_expression] = STATE(4801),
    [sym_block_comment] = STATE(4228),
    [sym_compiler_directive_decl] = STATE(4228),
    [sym_fsi_directive_decl] = STATE(4228),
    [sym_preproc_line] = STATE(4228),
    [aux_sym_if_expression_repeat1] = STATE(4421),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8677),
    [anon_sym_elif] = ACTIONS(8679),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_DOT_DOT] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__dedent] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4229)] = {
    [sym__else_expression] = STATE(5631),
    [sym_elif_expression] = STATE(4823),
    [sym_block_comment] = STATE(4229),
    [sym_compiler_directive_decl] = STATE(4229),
    [sym_fsi_directive_decl] = STATE(4229),
    [sym_preproc_line] = STATE(4229),
    [aux_sym_if_expression_repeat1] = STATE(4230),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8681),
    [anon_sym_elif] = ACTIONS(8683),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__dedent] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
    [sym__in] = ACTIONS(6317),
  },
  [STATE(4230)] = {
    [sym__else_expression] = STATE(5646),
    [sym_elif_expression] = STATE(4823),
    [sym_block_comment] = STATE(4230),
    [sym_compiler_directive_decl] = STATE(4230),
    [sym_fsi_directive_decl] = STATE(4230),
    [sym_preproc_line] = STATE(4230),
    [aux_sym_if_expression_repeat1] = STATE(4441),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8681),
    [anon_sym_elif] = ACTIONS(8683),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__dedent] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
    [sym__in] = ACTIONS(6327),
  },
  [STATE(4231)] = {
    [sym_block_comment] = STATE(4231),
    [sym_compiler_directive_decl] = STATE(4231),
    [sym_fsi_directive_decl] = STATE(4231),
    [sym_preproc_line] = STATE(4231),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_module] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_open] = ACTIONS(7053),
    [anon_sym_type] = ACTIONS(7053),
    [anon_sym_LBRACK_LT] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_exception] = ACTIONS(7053),
    [anon_sym_extern] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7055),
    [anon_sym_POUNDload] = ACTIONS(7055),
    [anon_sym_POUNDtime] = ACTIONS(7055),
    [anon_sym_POUNDI] = ACTIONS(7055),
    [anon_sym_POUNDhelp] = ACTIONS(7055),
    [anon_sym_POUNDquit] = ACTIONS(7055),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(4232)] = {
    [sym_block_comment] = STATE(4232),
    [sym_compiler_directive_decl] = STATE(4232),
    [sym_fsi_directive_decl] = STATE(4232),
    [sym_preproc_line] = STATE(4232),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4233)] = {
    [sym_block_comment] = STATE(4233),
    [sym_compiler_directive_decl] = STATE(4233),
    [sym_fsi_directive_decl] = STATE(4233),
    [sym_preproc_line] = STATE(4233),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_module] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_open] = ACTIONS(6888),
    [anon_sym_type] = ACTIONS(6888),
    [anon_sym_LBRACK_LT] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_exception] = ACTIONS(6888),
    [anon_sym_extern] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6886),
    [anon_sym_POUNDload] = ACTIONS(6886),
    [anon_sym_POUNDtime] = ACTIONS(6886),
    [anon_sym_POUNDI] = ACTIONS(6886),
    [anon_sym_POUNDhelp] = ACTIONS(6886),
    [anon_sym_POUNDquit] = ACTIONS(6886),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(4234)] = {
    [sym_block_comment] = STATE(4234),
    [sym_compiler_directive_decl] = STATE(4234),
    [sym_fsi_directive_decl] = STATE(4234),
    [sym_preproc_line] = STATE(4234),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_module] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_open] = ACTIONS(7069),
    [anon_sym_type] = ACTIONS(7069),
    [anon_sym_LBRACK_LT] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_exception] = ACTIONS(7069),
    [anon_sym_extern] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7071),
    [anon_sym_POUNDload] = ACTIONS(7071),
    [anon_sym_POUNDtime] = ACTIONS(7071),
    [anon_sym_POUNDI] = ACTIONS(7071),
    [anon_sym_POUNDhelp] = ACTIONS(7071),
    [anon_sym_POUNDquit] = ACTIONS(7071),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(4235)] = {
    [sym_block_comment] = STATE(4235),
    [sym_compiler_directive_decl] = STATE(4235),
    [sym_fsi_directive_decl] = STATE(4235),
    [sym_preproc_line] = STATE(4235),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_module] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_open] = ACTIONS(7073),
    [anon_sym_type] = ACTIONS(7073),
    [anon_sym_LBRACK_LT] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_exception] = ACTIONS(7073),
    [anon_sym_extern] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7075),
    [anon_sym_POUNDload] = ACTIONS(7075),
    [anon_sym_POUNDtime] = ACTIONS(7075),
    [anon_sym_POUNDI] = ACTIONS(7075),
    [anon_sym_POUNDhelp] = ACTIONS(7075),
    [anon_sym_POUNDquit] = ACTIONS(7075),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(4236)] = {
    [sym_block_comment] = STATE(4236),
    [sym_compiler_directive_decl] = STATE(4236),
    [sym_fsi_directive_decl] = STATE(4236),
    [sym_preproc_line] = STATE(4236),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_module] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_open] = ACTIONS(7077),
    [anon_sym_type] = ACTIONS(7077),
    [anon_sym_LBRACK_LT] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_exception] = ACTIONS(7077),
    [anon_sym_extern] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7079),
    [anon_sym_POUNDload] = ACTIONS(7079),
    [anon_sym_POUNDtime] = ACTIONS(7079),
    [anon_sym_POUNDI] = ACTIONS(7079),
    [anon_sym_POUNDhelp] = ACTIONS(7079),
    [anon_sym_POUNDquit] = ACTIONS(7079),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(4237)] = {
    [sym_block_comment] = STATE(4237),
    [sym_compiler_directive_decl] = STATE(4237),
    [sym_fsi_directive_decl] = STATE(4237),
    [sym_preproc_line] = STATE(4237),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_module] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_open] = ACTIONS(7085),
    [anon_sym_type] = ACTIONS(7085),
    [anon_sym_LBRACK_LT] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_exception] = ACTIONS(7085),
    [anon_sym_extern] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7087),
    [anon_sym_POUNDload] = ACTIONS(7087),
    [anon_sym_POUNDtime] = ACTIONS(7087),
    [anon_sym_POUNDI] = ACTIONS(7087),
    [anon_sym_POUNDhelp] = ACTIONS(7087),
    [anon_sym_POUNDquit] = ACTIONS(7087),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(4238)] = {
    [sym_block_comment] = STATE(4238),
    [sym_compiler_directive_decl] = STATE(4238),
    [sym_fsi_directive_decl] = STATE(4238),
    [sym_preproc_line] = STATE(4238),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_module] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_open] = ACTIONS(7089),
    [anon_sym_type] = ACTIONS(7089),
    [anon_sym_LBRACK_LT] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_exception] = ACTIONS(7089),
    [anon_sym_extern] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7091),
    [anon_sym_POUNDload] = ACTIONS(7091),
    [anon_sym_POUNDtime] = ACTIONS(7091),
    [anon_sym_POUNDI] = ACTIONS(7091),
    [anon_sym_POUNDhelp] = ACTIONS(7091),
    [anon_sym_POUNDquit] = ACTIONS(7091),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(4239)] = {
    [sym_block_comment] = STATE(4239),
    [sym_compiler_directive_decl] = STATE(4239),
    [sym_fsi_directive_decl] = STATE(4239),
    [sym_preproc_line] = STATE(4239),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_module] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_open] = ACTIONS(7093),
    [anon_sym_type] = ACTIONS(7093),
    [anon_sym_LBRACK_LT] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_exception] = ACTIONS(7093),
    [anon_sym_extern] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7095),
    [anon_sym_POUNDload] = ACTIONS(7095),
    [anon_sym_POUNDtime] = ACTIONS(7095),
    [anon_sym_POUNDI] = ACTIONS(7095),
    [anon_sym_POUNDhelp] = ACTIONS(7095),
    [anon_sym_POUNDquit] = ACTIONS(7095),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(4240)] = {
    [sym_block_comment] = STATE(4240),
    [sym_compiler_directive_decl] = STATE(4240),
    [sym_fsi_directive_decl] = STATE(4240),
    [sym_preproc_line] = STATE(4240),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_module] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_open] = ACTIONS(7097),
    [anon_sym_type] = ACTIONS(7097),
    [anon_sym_LBRACK_LT] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_exception] = ACTIONS(7097),
    [anon_sym_extern] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7099),
    [anon_sym_POUNDload] = ACTIONS(7099),
    [anon_sym_POUNDtime] = ACTIONS(7099),
    [anon_sym_POUNDI] = ACTIONS(7099),
    [anon_sym_POUNDhelp] = ACTIONS(7099),
    [anon_sym_POUNDquit] = ACTIONS(7099),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(4241)] = {
    [sym_block_comment] = STATE(4241),
    [sym_compiler_directive_decl] = STATE(4241),
    [sym_fsi_directive_decl] = STATE(4241),
    [sym_preproc_line] = STATE(4241),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_module] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_open] = ACTIONS(7101),
    [anon_sym_type] = ACTIONS(7101),
    [anon_sym_LBRACK_LT] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_exception] = ACTIONS(7101),
    [anon_sym_extern] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7103),
    [anon_sym_POUNDload] = ACTIONS(7103),
    [anon_sym_POUNDtime] = ACTIONS(7103),
    [anon_sym_POUNDI] = ACTIONS(7103),
    [anon_sym_POUNDhelp] = ACTIONS(7103),
    [anon_sym_POUNDquit] = ACTIONS(7103),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(4242)] = {
    [sym_block_comment] = STATE(4242),
    [sym_compiler_directive_decl] = STATE(4242),
    [sym_fsi_directive_decl] = STATE(4242),
    [sym_preproc_line] = STATE(4242),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_module] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_open] = ACTIONS(7105),
    [anon_sym_type] = ACTIONS(7105),
    [anon_sym_LBRACK_LT] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_exception] = ACTIONS(7105),
    [anon_sym_extern] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7107),
    [anon_sym_POUNDload] = ACTIONS(7107),
    [anon_sym_POUNDtime] = ACTIONS(7107),
    [anon_sym_POUNDI] = ACTIONS(7107),
    [anon_sym_POUNDhelp] = ACTIONS(7107),
    [anon_sym_POUNDquit] = ACTIONS(7107),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(4243)] = {
    [sym_block_comment] = STATE(4243),
    [sym_compiler_directive_decl] = STATE(4243),
    [sym_fsi_directive_decl] = STATE(4243),
    [sym_preproc_line] = STATE(4243),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_module] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_open] = ACTIONS(7109),
    [anon_sym_type] = ACTIONS(7109),
    [anon_sym_LBRACK_LT] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_exception] = ACTIONS(7109),
    [anon_sym_extern] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7111),
    [anon_sym_POUNDload] = ACTIONS(7111),
    [anon_sym_POUNDtime] = ACTIONS(7111),
    [anon_sym_POUNDI] = ACTIONS(7111),
    [anon_sym_POUNDhelp] = ACTIONS(7111),
    [anon_sym_POUNDquit] = ACTIONS(7111),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(4244)] = {
    [sym_block_comment] = STATE(4244),
    [sym_compiler_directive_decl] = STATE(4244),
    [sym_fsi_directive_decl] = STATE(4244),
    [sym_preproc_line] = STATE(4244),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_module] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_open] = ACTIONS(7113),
    [anon_sym_type] = ACTIONS(7113),
    [anon_sym_LBRACK_LT] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_exception] = ACTIONS(7113),
    [anon_sym_extern] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7115),
    [anon_sym_POUNDload] = ACTIONS(7115),
    [anon_sym_POUNDtime] = ACTIONS(7115),
    [anon_sym_POUNDI] = ACTIONS(7115),
    [anon_sym_POUNDhelp] = ACTIONS(7115),
    [anon_sym_POUNDquit] = ACTIONS(7115),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(4245)] = {
    [sym_block_comment] = STATE(4245),
    [sym_compiler_directive_decl] = STATE(4245),
    [sym_fsi_directive_decl] = STATE(4245),
    [sym_preproc_line] = STATE(4245),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_module] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_open] = ACTIONS(7117),
    [anon_sym_type] = ACTIONS(7117),
    [anon_sym_LBRACK_LT] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_exception] = ACTIONS(7117),
    [anon_sym_extern] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7119),
    [anon_sym_POUNDload] = ACTIONS(7119),
    [anon_sym_POUNDtime] = ACTIONS(7119),
    [anon_sym_POUNDI] = ACTIONS(7119),
    [anon_sym_POUNDhelp] = ACTIONS(7119),
    [anon_sym_POUNDquit] = ACTIONS(7119),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(4246)] = {
    [sym_block_comment] = STATE(4246),
    [sym_compiler_directive_decl] = STATE(4246),
    [sym_fsi_directive_decl] = STATE(4246),
    [sym_preproc_line] = STATE(4246),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_module] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_open] = ACTIONS(7121),
    [anon_sym_type] = ACTIONS(7121),
    [anon_sym_LBRACK_LT] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_exception] = ACTIONS(7121),
    [anon_sym_extern] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7123),
    [anon_sym_POUNDload] = ACTIONS(7123),
    [anon_sym_POUNDtime] = ACTIONS(7123),
    [anon_sym_POUNDI] = ACTIONS(7123),
    [anon_sym_POUNDhelp] = ACTIONS(7123),
    [anon_sym_POUNDquit] = ACTIONS(7123),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(4247)] = {
    [sym_block_comment] = STATE(4247),
    [sym_compiler_directive_decl] = STATE(4247),
    [sym_fsi_directive_decl] = STATE(4247),
    [sym_preproc_line] = STATE(4247),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_module] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_open] = ACTIONS(6936),
    [anon_sym_type] = ACTIONS(6936),
    [anon_sym_LBRACK_LT] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_exception] = ACTIONS(6936),
    [anon_sym_extern] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6934),
    [anon_sym_POUNDload] = ACTIONS(6934),
    [anon_sym_POUNDtime] = ACTIONS(6934),
    [anon_sym_POUNDI] = ACTIONS(6934),
    [anon_sym_POUNDhelp] = ACTIONS(6934),
    [anon_sym_POUNDquit] = ACTIONS(6934),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(4248)] = {
    [sym_block_comment] = STATE(4248),
    [sym_compiler_directive_decl] = STATE(4248),
    [sym_fsi_directive_decl] = STATE(4248),
    [sym_preproc_line] = STATE(4248),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_module] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_open] = ACTIONS(6940),
    [anon_sym_type] = ACTIONS(6940),
    [anon_sym_LBRACK_LT] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_exception] = ACTIONS(6940),
    [anon_sym_extern] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6938),
    [anon_sym_POUNDload] = ACTIONS(6938),
    [anon_sym_POUNDtime] = ACTIONS(6938),
    [anon_sym_POUNDI] = ACTIONS(6938),
    [anon_sym_POUNDhelp] = ACTIONS(6938),
    [anon_sym_POUNDquit] = ACTIONS(6938),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(4249)] = {
    [sym_block_comment] = STATE(4249),
    [sym_compiler_directive_decl] = STATE(4249),
    [sym_fsi_directive_decl] = STATE(4249),
    [sym_preproc_line] = STATE(4249),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_module] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_open] = ACTIONS(7015),
    [anon_sym_type] = ACTIONS(7015),
    [anon_sym_LBRACK_LT] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_exception] = ACTIONS(7015),
    [anon_sym_extern] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7013),
    [anon_sym_POUNDload] = ACTIONS(7013),
    [anon_sym_POUNDtime] = ACTIONS(7013),
    [anon_sym_POUNDI] = ACTIONS(7013),
    [anon_sym_POUNDhelp] = ACTIONS(7013),
    [anon_sym_POUNDquit] = ACTIONS(7013),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(4250)] = {
    [sym_block_comment] = STATE(4250),
    [sym_compiler_directive_decl] = STATE(4250),
    [sym_fsi_directive_decl] = STATE(4250),
    [sym_preproc_line] = STATE(4250),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_module] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_open] = ACTIONS(7031),
    [anon_sym_type] = ACTIONS(7031),
    [anon_sym_LBRACK_LT] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_exception] = ACTIONS(7031),
    [anon_sym_extern] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7029),
    [anon_sym_POUNDload] = ACTIONS(7029),
    [anon_sym_POUNDtime] = ACTIONS(7029),
    [anon_sym_POUNDI] = ACTIONS(7029),
    [anon_sym_POUNDhelp] = ACTIONS(7029),
    [anon_sym_POUNDquit] = ACTIONS(7029),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(4251)] = {
    [sym_block_comment] = STATE(4251),
    [sym_compiler_directive_decl] = STATE(4251),
    [sym_fsi_directive_decl] = STATE(4251),
    [sym_preproc_line] = STATE(4251),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_module] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_open] = ACTIONS(7047),
    [anon_sym_type] = ACTIONS(7047),
    [anon_sym_LBRACK_LT] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_exception] = ACTIONS(7047),
    [anon_sym_extern] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7045),
    [anon_sym_POUNDload] = ACTIONS(7045),
    [anon_sym_POUNDtime] = ACTIONS(7045),
    [anon_sym_POUNDI] = ACTIONS(7045),
    [anon_sym_POUNDhelp] = ACTIONS(7045),
    [anon_sym_POUNDquit] = ACTIONS(7045),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(4252)] = {
    [sym_block_comment] = STATE(4252),
    [sym_compiler_directive_decl] = STATE(4252),
    [sym_fsi_directive_decl] = STATE(4252),
    [sym_preproc_line] = STATE(4252),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_module] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_open] = ACTIONS(7051),
    [anon_sym_type] = ACTIONS(7051),
    [anon_sym_LBRACK_LT] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_exception] = ACTIONS(7051),
    [anon_sym_extern] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7049),
    [anon_sym_POUNDload] = ACTIONS(7049),
    [anon_sym_POUNDtime] = ACTIONS(7049),
    [anon_sym_POUNDI] = ACTIONS(7049),
    [anon_sym_POUNDhelp] = ACTIONS(7049),
    [anon_sym_POUNDquit] = ACTIONS(7049),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(4253)] = {
    [sym_block_comment] = STATE(4253),
    [sym_compiler_directive_decl] = STATE(4253),
    [sym_fsi_directive_decl] = STATE(4253),
    [sym_preproc_line] = STATE(4253),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_module] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_open] = ACTIONS(7063),
    [anon_sym_type] = ACTIONS(7063),
    [anon_sym_LBRACK_LT] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_exception] = ACTIONS(7063),
    [anon_sym_extern] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7061),
    [anon_sym_POUNDload] = ACTIONS(7061),
    [anon_sym_POUNDtime] = ACTIONS(7061),
    [anon_sym_POUNDI] = ACTIONS(7061),
    [anon_sym_POUNDhelp] = ACTIONS(7061),
    [anon_sym_POUNDquit] = ACTIONS(7061),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(4254)] = {
    [sym_block_comment] = STATE(4254),
    [sym_compiler_directive_decl] = STATE(4254),
    [sym_fsi_directive_decl] = STATE(4254),
    [sym_preproc_line] = STATE(4254),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_module] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_open] = ACTIONS(7067),
    [anon_sym_type] = ACTIONS(7067),
    [anon_sym_LBRACK_LT] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_exception] = ACTIONS(7067),
    [anon_sym_extern] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7065),
    [anon_sym_POUNDload] = ACTIONS(7065),
    [anon_sym_POUNDtime] = ACTIONS(7065),
    [anon_sym_POUNDI] = ACTIONS(7065),
    [anon_sym_POUNDhelp] = ACTIONS(7065),
    [anon_sym_POUNDquit] = ACTIONS(7065),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(4255)] = {
    [sym_block_comment] = STATE(4255),
    [sym_compiler_directive_decl] = STATE(4255),
    [sym_fsi_directive_decl] = STATE(4255),
    [sym_preproc_line] = STATE(4255),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_module] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_open] = ACTIONS(7125),
    [anon_sym_type] = ACTIONS(7125),
    [anon_sym_LBRACK_LT] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_exception] = ACTIONS(7125),
    [anon_sym_extern] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7127),
    [anon_sym_POUNDload] = ACTIONS(7127),
    [anon_sym_POUNDtime] = ACTIONS(7127),
    [anon_sym_POUNDI] = ACTIONS(7127),
    [anon_sym_POUNDhelp] = ACTIONS(7127),
    [anon_sym_POUNDquit] = ACTIONS(7127),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(4256)] = {
    [sym_block_comment] = STATE(4256),
    [sym_compiler_directive_decl] = STATE(4256),
    [sym_fsi_directive_decl] = STATE(4256),
    [sym_preproc_line] = STATE(4256),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_module] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_open] = ACTIONS(7129),
    [anon_sym_type] = ACTIONS(7129),
    [anon_sym_LBRACK_LT] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_exception] = ACTIONS(7129),
    [anon_sym_extern] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7131),
    [anon_sym_POUNDload] = ACTIONS(7131),
    [anon_sym_POUNDtime] = ACTIONS(7131),
    [anon_sym_POUNDI] = ACTIONS(7131),
    [anon_sym_POUNDhelp] = ACTIONS(7131),
    [anon_sym_POUNDquit] = ACTIONS(7131),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(4257)] = {
    [sym_block_comment] = STATE(4257),
    [sym_compiler_directive_decl] = STATE(4257),
    [sym_fsi_directive_decl] = STATE(4257),
    [sym_preproc_line] = STATE(4257),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_module] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_open] = ACTIONS(7133),
    [anon_sym_type] = ACTIONS(7133),
    [anon_sym_LBRACK_LT] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_exception] = ACTIONS(7133),
    [anon_sym_extern] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7135),
    [anon_sym_POUNDload] = ACTIONS(7135),
    [anon_sym_POUNDtime] = ACTIONS(7135),
    [anon_sym_POUNDI] = ACTIONS(7135),
    [anon_sym_POUNDhelp] = ACTIONS(7135),
    [anon_sym_POUNDquit] = ACTIONS(7135),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(4258)] = {
    [sym_block_comment] = STATE(4258),
    [sym_compiler_directive_decl] = STATE(4258),
    [sym_fsi_directive_decl] = STATE(4258),
    [sym_preproc_line] = STATE(4258),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_module] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_open] = ACTIONS(7137),
    [anon_sym_type] = ACTIONS(7137),
    [anon_sym_LBRACK_LT] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_exception] = ACTIONS(7137),
    [anon_sym_extern] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7139),
    [anon_sym_POUNDload] = ACTIONS(7139),
    [anon_sym_POUNDtime] = ACTIONS(7139),
    [anon_sym_POUNDI] = ACTIONS(7139),
    [anon_sym_POUNDhelp] = ACTIONS(7139),
    [anon_sym_POUNDquit] = ACTIONS(7139),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(4259)] = {
    [sym_block_comment] = STATE(4259),
    [sym_compiler_directive_decl] = STATE(4259),
    [sym_fsi_directive_decl] = STATE(4259),
    [sym_preproc_line] = STATE(4259),
    [aux_sym_type_argument_repeat1] = STATE(4286),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_as] = ACTIONS(6331),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8685),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4260)] = {
    [sym_attributes] = STATE(13407),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4597),
    [sym_type_extension_elements] = STATE(8353),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4260),
    [sym_compiler_directive_decl] = STATE(4260),
    [sym_fsi_directive_decl] = STATE(4260),
    [sym_preproc_line] = STATE(4260),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8343),
    [sym_identifier] = ACTIONS(8341),
    [anon_sym_module] = ACTIONS(8341),
    [anon_sym_null] = ACTIONS(8341),
    [anon_sym_begin] = ACTIONS(8341),
    [anon_sym_open] = ACTIONS(8341),
    [anon_sym_type] = ACTIONS(8341),
    [anon_sym_LBRACK_LT] = ACTIONS(8343),
    [anon_sym_do] = ACTIONS(8341),
    [anon_sym_and] = ACTIONS(8341),
    [anon_sym_let] = ACTIONS(8341),
    [anon_sym_let_BANG] = ACTIONS(8343),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8341),
    [anon_sym_AMP] = ACTIONS(8341),
    [anon_sym_LBRACK] = ACTIONS(8341),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8343),
    [anon_sym_LBRACE] = ACTIONS(8341),
    [anon_sym_LT_AT] = ACTIONS(8341),
    [anon_sym_LT_AT_AT] = ACTIONS(8343),
    [anon_sym_CARET] = ACTIONS(8343),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8343),
    [anon_sym_new] = ACTIONS(8341),
    [anon_sym_with] = ACTIONS(8508),
    [anon_sym_return] = ACTIONS(8341),
    [anon_sym_return_BANG] = ACTIONS(8343),
    [anon_sym_yield] = ACTIONS(8341),
    [anon_sym_yield_BANG] = ACTIONS(8343),
    [anon_sym_lazy] = ACTIONS(8341),
    [anon_sym_assert] = ACTIONS(8341),
    [anon_sym_upcast] = ACTIONS(8341),
    [anon_sym_downcast] = ACTIONS(8341),
    [anon_sym_fixed] = ACTIONS(8341),
    [anon_sym_for] = ACTIONS(8341),
    [anon_sym_while] = ACTIONS(8341),
    [anon_sym_while_BANG] = ACTIONS(8343),
    [anon_sym_if] = ACTIONS(8341),
    [anon_sym_fun] = ACTIONS(8341),
    [anon_sym_try] = ACTIONS(8341),
    [anon_sym_match] = ACTIONS(8341),
    [anon_sym_match_BANG] = ACTIONS(8343),
    [anon_sym_function] = ACTIONS(8341),
    [anon_sym_use] = ACTIONS(8341),
    [anon_sym_use_BANG] = ACTIONS(8343),
    [anon_sym_do_BANG] = ACTIONS(8343),
    [anon_sym_AMP_AMP] = ACTIONS(8343),
    [anon_sym_default] = ACTIONS(8510),
    [anon_sym_static] = ACTIONS(8512),
    [anon_sym_member] = ACTIONS(8613),
    [anon_sym_exception] = ACTIONS(8341),
    [anon_sym_interface] = ACTIONS(8516),
    [anon_sym_abstract] = ACTIONS(8518),
    [anon_sym_val] = ACTIONS(8520),
    [anon_sym_override] = ACTIONS(8510),
    [anon_sym_extern] = ACTIONS(8341),
    [aux_sym_char_token1] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8341),
    [anon_sym_DQUOTE] = ACTIONS(8341),
    [anon_sym_AT_DQUOTE] = ACTIONS(8343),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8343),
    [sym_bool] = ACTIONS(8341),
    [sym_unit] = ACTIONS(8343),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8341),
    [sym_op_identifier] = ACTIONS(8343),
    [anon_sym_PLUS] = ACTIONS(8341),
    [anon_sym_DASH] = ACTIONS(8341),
    [anon_sym_PLUS_DOT] = ACTIONS(8343),
    [anon_sym_DASH_DOT] = ACTIONS(8343),
    [anon_sym_PERCENT] = ACTIONS(8341),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8343),
    [anon_sym_TILDE] = ACTIONS(8343),
    [aux_sym_prefix_op_token1] = ACTIONS(8343),
    [sym_int] = ACTIONS(8341),
    [sym_xint] = ACTIONS(8343),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8343),
    [anon_sym_POUNDload] = ACTIONS(8343),
    [anon_sym_POUNDtime] = ACTIONS(8343),
    [anon_sym_POUNDI] = ACTIONS(8343),
    [anon_sym_POUNDhelp] = ACTIONS(8343),
    [anon_sym_POUNDquit] = ACTIONS(8343),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8343),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8343),
  },
  [STATE(4261)] = {
    [sym_block_comment] = STATE(4261),
    [sym_compiler_directive_decl] = STATE(4261),
    [sym_fsi_directive_decl] = STATE(4261),
    [sym_preproc_line] = STATE(4261),
    [aux_sym_compound_type_repeat1] = STATE(4264),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_as] = ACTIONS(6338),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8218),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4262)] = {
    [sym_type_arguments] = STATE(3730),
    [sym_long_identifier] = STATE(3759),
    [sym_block_comment] = STATE(4262),
    [sym_compiler_directive_decl] = STATE(4262),
    [sym_fsi_directive_decl] = STATE(4262),
    [sym_preproc_line] = STATE(4262),
    [aux_sym_compound_type_repeat1] = STATE(3689),
    [sym_identifier] = ACTIONS(6265),
    [anon_sym_null] = ACTIONS(6265),
    [anon_sym_EQ] = ACTIONS(6265),
    [anon_sym_begin] = ACTIONS(6265),
    [anon_sym_COLON] = ACTIONS(6265),
    [anon_sym_do] = ACTIONS(6265),
    [anon_sym_let] = ACTIONS(6265),
    [anon_sym_let_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN] = ACTIONS(6265),
    [anon_sym_COMMA] = ACTIONS(6263),
    [anon_sym_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_QMARK] = ACTIONS(6265),
    [anon_sym_COLON_COLON] = ACTIONS(6263),
    [anon_sym_AMP] = ACTIONS(6265),
    [anon_sym_LBRACK] = ACTIONS(6265),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6263),
    [anon_sym_LBRACE] = ACTIONS(6265),
    [anon_sym_LT_AT] = ACTIONS(6265),
    [anon_sym_LT_AT_AT] = ACTIONS(6265),
    [anon_sym_CARET] = ACTIONS(6265),
    [anon_sym_or] = ACTIONS(6265),
    [anon_sym_DOT] = ACTIONS(6265),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6263),
    [anon_sym_new] = ACTIONS(6265),
    [anon_sym_return] = ACTIONS(6265),
    [anon_sym_return_BANG] = ACTIONS(6263),
    [anon_sym_yield] = ACTIONS(6265),
    [anon_sym_yield_BANG] = ACTIONS(6263),
    [anon_sym_lazy] = ACTIONS(6265),
    [anon_sym_assert] = ACTIONS(6265),
    [anon_sym_upcast] = ACTIONS(6265),
    [anon_sym_downcast] = ACTIONS(6265),
    [anon_sym_fixed] = ACTIONS(6265),
    [anon_sym_COLON_GT] = ACTIONS(6263),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6263),
    [anon_sym_for] = ACTIONS(6265),
    [anon_sym_while] = ACTIONS(6265),
    [anon_sym_while_BANG] = ACTIONS(6263),
    [anon_sym_if] = ACTIONS(6265),
    [anon_sym_fun] = ACTIONS(6265),
    [anon_sym_DASH_GT] = ACTIONS(6265),
    [anon_sym_try] = ACTIONS(6265),
    [anon_sym_match] = ACTIONS(6265),
    [anon_sym_match_BANG] = ACTIONS(6263),
    [anon_sym_function] = ACTIONS(6265),
    [anon_sym_DOT_LBRACK] = ACTIONS(6263),
    [anon_sym_use] = ACTIONS(6265),
    [anon_sym_use_BANG] = ACTIONS(6263),
    [anon_sym_do_BANG] = ACTIONS(6263),
    [anon_sym_LPAREN2] = ACTIONS(6265),
    [anon_sym_PIPE_PIPE] = ACTIONS(6265),
    [anon_sym_AMP_AMP] = ACTIONS(6265),
    [anon_sym_STAR] = ACTIONS(7983),
    [anon_sym_LT] = ACTIONS(7985),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(7987),
    [aux_sym_char_token1] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6265),
    [anon_sym_DQUOTE] = ACTIONS(6265),
    [anon_sym_AT_DQUOTE] = ACTIONS(6263),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6263),
    [sym_bool] = ACTIONS(6265),
    [sym_unit] = ACTIONS(6263),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6265),
    [sym_op_identifier] = ACTIONS(6265),
    [anon_sym_PLUS] = ACTIONS(6265),
    [anon_sym_DASH] = ACTIONS(6265),
    [anon_sym_PLUS_DOT] = ACTIONS(6265),
    [anon_sym_DASH_DOT] = ACTIONS(6265),
    [anon_sym_PERCENT] = ACTIONS(6265),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6265),
    [anon_sym_TILDE] = ACTIONS(6263),
    [aux_sym_prefix_op_token1] = ACTIONS(6265),
    [aux_sym_infix_op_token1] = ACTIONS(6263),
    [aux_sym_infix_op_token2] = ACTIONS(6265),
    [aux_sym_infix_op_token3] = ACTIONS(6265),
    [anon_sym_BANG_EQ] = ACTIONS(6265),
    [anon_sym_COLON_EQ] = ACTIONS(6263),
    [anon_sym_DOLLAR] = ACTIONS(6265),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6265),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6265),
    [sym_int] = ACTIONS(6265),
    [sym_xint] = ACTIONS(6263),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6263),
    [sym__newline] = ACTIONS(6263),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6263),
    [sym__tyapp_open] = ACTIONS(6263),
  },
  [STATE(4263)] = {
    [sym__else_expression] = STATE(5709),
    [sym_elif_expression] = STATE(4848),
    [sym_block_comment] = STATE(4263),
    [sym_compiler_directive_decl] = STATE(4263),
    [sym_fsi_directive_decl] = STATE(4263),
    [sym_preproc_line] = STATE(4263),
    [aux_sym_if_expression_repeat1] = STATE(4268),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8688),
    [anon_sym_elif] = ACTIONS(8690),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_DASH_GT] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_DOT_DOT] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4264)] = {
    [sym_block_comment] = STATE(4264),
    [sym_compiler_directive_decl] = STATE(4264),
    [sym_fsi_directive_decl] = STATE(4264),
    [sym_preproc_line] = STATE(4264),
    [aux_sym_compound_type_repeat1] = STATE(4264),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_as] = ACTIONS(6193),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8692),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4265)] = {
    [sym_block_comment] = STATE(4265),
    [sym_compiler_directive_decl] = STATE(4265),
    [sym_fsi_directive_decl] = STATE(4265),
    [sym_preproc_line] = STATE(4265),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_module] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_open] = ACTIONS(7145),
    [anon_sym_type] = ACTIONS(7145),
    [anon_sym_LBRACK_LT] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_exception] = ACTIONS(7145),
    [anon_sym_extern] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7147),
    [anon_sym_POUNDload] = ACTIONS(7147),
    [anon_sym_POUNDtime] = ACTIONS(7147),
    [anon_sym_POUNDI] = ACTIONS(7147),
    [anon_sym_POUNDhelp] = ACTIONS(7147),
    [anon_sym_POUNDquit] = ACTIONS(7147),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(4266)] = {
    [sym_block_comment] = STATE(4266),
    [sym_compiler_directive_decl] = STATE(4266),
    [sym_fsi_directive_decl] = STATE(4266),
    [sym_preproc_line] = STATE(4266),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_module] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_open] = ACTIONS(7155),
    [anon_sym_type] = ACTIONS(7155),
    [anon_sym_LBRACK_LT] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_exception] = ACTIONS(7155),
    [anon_sym_extern] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7157),
    [anon_sym_POUNDload] = ACTIONS(7157),
    [anon_sym_POUNDtime] = ACTIONS(7157),
    [anon_sym_POUNDI] = ACTIONS(7157),
    [anon_sym_POUNDhelp] = ACTIONS(7157),
    [anon_sym_POUNDquit] = ACTIONS(7157),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(4267)] = {
    [sym_block_comment] = STATE(4267),
    [sym_compiler_directive_decl] = STATE(4267),
    [sym_fsi_directive_decl] = STATE(4267),
    [sym_preproc_line] = STATE(4267),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_module] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_open] = ACTIONS(7159),
    [anon_sym_type] = ACTIONS(7159),
    [anon_sym_LBRACK_LT] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_exception] = ACTIONS(7159),
    [anon_sym_extern] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(7161),
    [anon_sym_POUNDload] = ACTIONS(7161),
    [anon_sym_POUNDtime] = ACTIONS(7161),
    [anon_sym_POUNDI] = ACTIONS(7161),
    [anon_sym_POUNDhelp] = ACTIONS(7161),
    [anon_sym_POUNDquit] = ACTIONS(7161),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(4268)] = {
    [sym__else_expression] = STATE(5735),
    [sym_elif_expression] = STATE(4848),
    [sym_block_comment] = STATE(4268),
    [sym_compiler_directive_decl] = STATE(4268),
    [sym_fsi_directive_decl] = STATE(4268),
    [sym_preproc_line] = STATE(4268),
    [aux_sym_if_expression_repeat1] = STATE(4490),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8688),
    [anon_sym_elif] = ACTIONS(8690),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_DASH_GT] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_DOT_DOT] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4269)] = {
    [sym_block_comment] = STATE(4269),
    [sym_compiler_directive_decl] = STATE(4269),
    [sym_fsi_directive_decl] = STATE(4269),
    [sym_preproc_line] = STATE(4269),
    [aux_sym_type_argument_repeat1] = STATE(4296),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8695),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4270)] = {
    [sym_block_comment] = STATE(4270),
    [sym_compiler_directive_decl] = STATE(4270),
    [sym_fsi_directive_decl] = STATE(4270),
    [sym_preproc_line] = STATE(4270),
    [aux_sym_compound_type_repeat1] = STATE(4271),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8231),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [anon_sym_POUNDelse] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4271)] = {
    [sym_block_comment] = STATE(4271),
    [sym_compiler_directive_decl] = STATE(4271),
    [sym_fsi_directive_decl] = STATE(4271),
    [sym_preproc_line] = STATE(4271),
    [aux_sym_compound_type_repeat1] = STATE(4271),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8698),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4272)] = {
    [sym_block_comment] = STATE(4272),
    [sym_compiler_directive_decl] = STATE(4272),
    [sym_fsi_directive_decl] = STATE(4272),
    [sym_preproc_line] = STATE(4272),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_GT_RBRACK] = ACTIONS(6614),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_RBRACK] = ACTIONS(6614),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_RBRACE] = ACTIONS(6614),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_to] = ACTIONS(6612),
    [anon_sym_downto] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4273)] = {
    [sym_block_comment] = STATE(4273),
    [sym_compiler_directive_decl] = STATE(4273),
    [sym_fsi_directive_decl] = STATE(4273),
    [sym_preproc_line] = STATE(4273),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6295),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4274)] = {
    [sym_block_comment] = STATE(4274),
    [sym_compiler_directive_decl] = STATE(4274),
    [sym_fsi_directive_decl] = STATE(4274),
    [sym_preproc_line] = STATE(4274),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6369),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_DOT_DOT2] = ACTIONS(6371),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4275)] = {
    [sym_block_comment] = STATE(4275),
    [sym_compiler_directive_decl] = STATE(4275),
    [sym_fsi_directive_decl] = STATE(4275),
    [sym_preproc_line] = STATE(4275),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6369),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_DOT_DOT2] = ACTIONS(6371),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4276)] = {
    [sym_block_comment] = STATE(4276),
    [sym_compiler_directive_decl] = STATE(4276),
    [sym_fsi_directive_decl] = STATE(4276),
    [sym_preproc_line] = STATE(4276),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_DOT_DOT] = ACTIONS(6373),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_DOT_DOT2] = ACTIONS(6375),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8701),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__dedent] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4277)] = {
    [sym_block_comment] = STATE(4277),
    [sym_compiler_directive_decl] = STATE(4277),
    [sym_fsi_directive_decl] = STATE(4277),
    [sym_preproc_line] = STATE(4277),
    [aux_sym_long_identifier_repeat1] = STATE(4283),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_GT_RBRACK] = ACTIONS(6358),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8703),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4278)] = {
    [sym_block_comment] = STATE(4278),
    [sym_compiler_directive_decl] = STATE(4278),
    [sym_fsi_directive_decl] = STATE(4278),
    [sym_preproc_line] = STATE(4278),
    [sym_identifier] = ACTIONS(8325),
    [anon_sym_module] = ACTIONS(8328),
    [anon_sym_null] = ACTIONS(8325),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(8325),
    [anon_sym_open] = ACTIONS(8328),
    [anon_sym_type] = ACTIONS(8328),
    [anon_sym_LBRACK_LT] = ACTIONS(8330),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(8325),
    [anon_sym_let] = ACTIONS(8325),
    [anon_sym_let_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN] = ACTIONS(8325),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(8325),
    [anon_sym_LBRACK] = ACTIONS(8325),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8332),
    [anon_sym_LBRACE] = ACTIONS(8325),
    [anon_sym_LT_AT] = ACTIONS(8325),
    [anon_sym_LT_AT_AT] = ACTIONS(8325),
    [anon_sym_CARET] = ACTIONS(8325),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8332),
    [anon_sym_new] = ACTIONS(8325),
    [anon_sym_return] = ACTIONS(8325),
    [anon_sym_return_BANG] = ACTIONS(8332),
    [anon_sym_yield] = ACTIONS(8325),
    [anon_sym_yield_BANG] = ACTIONS(8332),
    [anon_sym_lazy] = ACTIONS(8325),
    [anon_sym_assert] = ACTIONS(8325),
    [anon_sym_upcast] = ACTIONS(8325),
    [anon_sym_downcast] = ACTIONS(8325),
    [anon_sym_fixed] = ACTIONS(8325),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(8325),
    [anon_sym_while] = ACTIONS(8325),
    [anon_sym_while_BANG] = ACTIONS(8332),
    [anon_sym_if] = ACTIONS(8325),
    [anon_sym_fun] = ACTIONS(8325),
    [anon_sym_try] = ACTIONS(8325),
    [anon_sym_match] = ACTIONS(8325),
    [anon_sym_match_BANG] = ACTIONS(8332),
    [anon_sym_function] = ACTIONS(8325),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(8325),
    [anon_sym_use_BANG] = ACTIONS(8332),
    [anon_sym_do_BANG] = ACTIONS(8332),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(8325),
    [anon_sym_exception] = ACTIONS(8328),
    [anon_sym_extern] = ACTIONS(8328),
    [aux_sym_char_token1] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8325),
    [anon_sym_DQUOTE] = ACTIONS(8325),
    [anon_sym_AT_DQUOTE] = ACTIONS(8332),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8332),
    [sym_bool] = ACTIONS(8325),
    [sym_unit] = ACTIONS(8332),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8325),
    [sym_op_identifier] = ACTIONS(8325),
    [anon_sym_PLUS] = ACTIONS(8325),
    [anon_sym_DASH] = ACTIONS(8325),
    [anon_sym_PLUS_DOT] = ACTIONS(8325),
    [anon_sym_DASH_DOT] = ACTIONS(8325),
    [anon_sym_PERCENT] = ACTIONS(8325),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8325),
    [anon_sym_TILDE] = ACTIONS(8332),
    [aux_sym_prefix_op_token1] = ACTIONS(8325),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(8325),
    [sym_xint] = ACTIONS(8332),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8330),
    [anon_sym_POUNDload] = ACTIONS(8330),
    [anon_sym_POUNDtime] = ACTIONS(8330),
    [anon_sym_POUNDI] = ACTIONS(8330),
    [anon_sym_POUNDhelp] = ACTIONS(8330),
    [anon_sym_POUNDquit] = ACTIONS(8330),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8332),
    [anon_sym_POUNDendif] = ACTIONS(8330),
    [sym__newline] = ACTIONS(8332),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8332),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4279)] = {
    [sym_block_comment] = STATE(4279),
    [sym_compiler_directive_decl] = STATE(4279),
    [sym_fsi_directive_decl] = STATE(4279),
    [sym_preproc_line] = STATE(4279),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_DOT_DOT] = ACTIONS(6367),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_DOT_DOT2] = ACTIONS(6365),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__dedent] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4280)] = {
    [sym_block_comment] = STATE(4280),
    [sym_compiler_directive_decl] = STATE(4280),
    [sym_fsi_directive_decl] = STATE(4280),
    [sym_preproc_line] = STATE(4280),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_DOT_DOT] = ACTIONS(6379),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_DOT_DOT2] = ACTIONS(6381),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__dedent] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4281)] = {
    [sym_block_comment] = STATE(4281),
    [sym_compiler_directive_decl] = STATE(4281),
    [sym_fsi_directive_decl] = STATE(4281),
    [sym_preproc_line] = STATE(4281),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_DOT_DOT] = ACTIONS(6383),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_DOT_DOT2] = ACTIONS(6385),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__dedent] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4282)] = {
    [sym_block_comment] = STATE(4282),
    [sym_compiler_directive_decl] = STATE(4282),
    [sym_fsi_directive_decl] = STATE(4282),
    [sym_preproc_line] = STATE(4282),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_DOT_DOT] = ACTIONS(6391),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_DOT_DOT2] = ACTIONS(6393),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4283)] = {
    [sym_block_comment] = STATE(4283),
    [sym_compiler_directive_decl] = STATE(4283),
    [sym_fsi_directive_decl] = STATE(4283),
    [sym_preproc_line] = STATE(4283),
    [aux_sym_long_identifier_repeat1] = STATE(4283),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_GT_RBRACK] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8705),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4284)] = {
    [sym_block_comment] = STATE(4284),
    [sym_compiler_directive_decl] = STATE(4284),
    [sym_fsi_directive_decl] = STATE(4284),
    [sym_preproc_line] = STATE(4284),
    [aux_sym_type_argument_repeat1] = STATE(4135),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8708),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
    [sym__in] = ACTIONS(6333),
  },
  [STATE(4285)] = {
    [sym_block_comment] = STATE(4285),
    [sym_compiler_directive_decl] = STATE(4285),
    [sym_fsi_directive_decl] = STATE(4285),
    [sym_preproc_line] = STATE(4285),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_GT_RBRACK] = ACTIONS(6838),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_RBRACK] = ACTIONS(6838),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_RBRACE] = ACTIONS(6838),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_to] = ACTIONS(6840),
    [anon_sym_downto] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_DOT_DOT2] = ACTIONS(6838),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(4286)] = {
    [sym_block_comment] = STATE(4286),
    [sym_compiler_directive_decl] = STATE(4286),
    [sym_fsi_directive_decl] = STATE(4286),
    [sym_preproc_line] = STATE(4286),
    [aux_sym_type_argument_repeat1] = STATE(4289),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_as] = ACTIONS(6308),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4287)] = {
    [sym_block_comment] = STATE(4287),
    [sym_compiler_directive_decl] = STATE(4287),
    [sym_fsi_directive_decl] = STATE(4287),
    [sym_preproc_line] = STATE(4287),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_GT_RBRACK] = ACTIONS(6651),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_RBRACK] = ACTIONS(6651),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_RBRACE] = ACTIONS(6651),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_to] = ACTIONS(6653),
    [anon_sym_downto] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_DOT_DOT2] = ACTIONS(6651),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4288)] = {
    [sym_block_comment] = STATE(4288),
    [sym_compiler_directive_decl] = STATE(4288),
    [sym_fsi_directive_decl] = STATE(4288),
    [sym_preproc_line] = STATE(4288),
    [aux_sym_compound_type_repeat1] = STATE(4290),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8278),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__dedent] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
    [sym__in] = ACTIONS(6340),
  },
  [STATE(4289)] = {
    [sym_block_comment] = STATE(4289),
    [sym_compiler_directive_decl] = STATE(4289),
    [sym_fsi_directive_decl] = STATE(4289),
    [sym_preproc_line] = STATE(4289),
    [aux_sym_type_argument_repeat1] = STATE(4289),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_as] = ACTIONS(6312),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8711),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4290)] = {
    [sym_block_comment] = STATE(4290),
    [sym_compiler_directive_decl] = STATE(4290),
    [sym_fsi_directive_decl] = STATE(4290),
    [sym_preproc_line] = STATE(4290),
    [aux_sym_compound_type_repeat1] = STATE(4290),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8714),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__dedent] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
    [sym__in] = ACTIONS(6195),
  },
  [STATE(4291)] = {
    [sym_block_comment] = STATE(4291),
    [sym_compiler_directive_decl] = STATE(4291),
    [sym_fsi_directive_decl] = STATE(4291),
    [sym_preproc_line] = STATE(4291),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_DOT_DOT] = ACTIONS(6395),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_DOT_DOT2] = ACTIONS(6397),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__dedent] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4292)] = {
    [sym_block_comment] = STATE(4292),
    [sym_compiler_directive_decl] = STATE(4292),
    [sym_fsi_directive_decl] = STATE(4292),
    [sym_preproc_line] = STATE(4292),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_DOT_DOT] = ACTIONS(6399),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_DOT_DOT2] = ACTIONS(6401),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__dedent] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4293)] = {
    [sym_block_comment] = STATE(4293),
    [sym_compiler_directive_decl] = STATE(4293),
    [sym_fsi_directive_decl] = STATE(4293),
    [sym_preproc_line] = STATE(4293),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_DOT_DOT] = ACTIONS(6403),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_DOT_DOT2] = ACTIONS(6405),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__dedent] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4294)] = {
    [sym_block_comment] = STATE(4294),
    [sym_compiler_directive_decl] = STATE(4294),
    [sym_fsi_directive_decl] = STATE(4294),
    [sym_preproc_line] = STATE(4294),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_DOT_DOT] = ACTIONS(6407),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_DOT_DOT2] = ACTIONS(6409),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__dedent] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4295)] = {
    [sym_block_comment] = STATE(4295),
    [sym_compiler_directive_decl] = STATE(4295),
    [sym_fsi_directive_decl] = STATE(4295),
    [sym_preproc_line] = STATE(4295),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_GT_RBRACK] = ACTIONS(6811),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_RBRACK] = ACTIONS(6811),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_RBRACE] = ACTIONS(6811),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_to] = ACTIONS(6813),
    [anon_sym_downto] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_DOT_DOT2] = ACTIONS(6811),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(4296)] = {
    [sym_block_comment] = STATE(4296),
    [sym_compiler_directive_decl] = STATE(4296),
    [sym_fsi_directive_decl] = STATE(4296),
    [sym_preproc_line] = STATE(4296),
    [aux_sym_type_argument_repeat1] = STATE(4218),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4297)] = {
    [sym_block_comment] = STATE(4297),
    [sym_compiler_directive_decl] = STATE(4297),
    [sym_fsi_directive_decl] = STATE(4297),
    [sym_preproc_line] = STATE(4297),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_DOT_DOT] = ACTIONS(6411),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_DOT_DOT2] = ACTIONS(6413),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__dedent] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4298)] = {
    [sym_block_comment] = STATE(4298),
    [sym_compiler_directive_decl] = STATE(4298),
    [sym_fsi_directive_decl] = STATE(4298),
    [sym_preproc_line] = STATE(4298),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_GT_RBRACK] = ACTIONS(6796),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_RBRACK] = ACTIONS(6796),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_RBRACE] = ACTIONS(6796),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_to] = ACTIONS(6794),
    [anon_sym_downto] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_DOT_DOT2] = ACTIONS(6796),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(4299)] = {
    [sym_block_comment] = STATE(4299),
    [sym_compiler_directive_decl] = STATE(4299),
    [sym_fsi_directive_decl] = STATE(4299),
    [sym_preproc_line] = STATE(4299),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__dedent] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
    [sym__in] = ACTIONS(6429),
  },
  [STATE(4300)] = {
    [sym_block_comment] = STATE(4300),
    [sym_compiler_directive_decl] = STATE(4300),
    [sym_fsi_directive_decl] = STATE(4300),
    [sym_preproc_line] = STATE(4300),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4301)] = {
    [sym_block_comment] = STATE(4301),
    [sym_compiler_directive_decl] = STATE(4301),
    [sym_fsi_directive_decl] = STATE(4301),
    [sym_preproc_line] = STATE(4301),
    [aux_sym_compound_type_repeat1] = STATE(4302),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_DOT_DOT] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8396),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4302)] = {
    [sym_block_comment] = STATE(4302),
    [sym_compiler_directive_decl] = STATE(4302),
    [sym_fsi_directive_decl] = STATE(4302),
    [sym_preproc_line] = STATE(4302),
    [aux_sym_compound_type_repeat1] = STATE(4302),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_DOT_DOT] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8717),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4303)] = {
    [sym_block_comment] = STATE(4303),
    [sym_compiler_directive_decl] = STATE(4303),
    [sym_fsi_directive_decl] = STATE(4303),
    [sym_preproc_line] = STATE(4303),
    [aux_sym_compound_type_repeat1] = STATE(4323),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8410),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__untyped_quoted_close] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4304)] = {
    [sym_block_comment] = STATE(4304),
    [sym_compiler_directive_decl] = STATE(4304),
    [sym_fsi_directive_decl] = STATE(4304),
    [sym_preproc_line] = STATE(4304),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4305)] = {
    [sym_block_comment] = STATE(4305),
    [sym_compiler_directive_decl] = STATE(4305),
    [sym_fsi_directive_decl] = STATE(4305),
    [sym_preproc_line] = STATE(4305),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
    [sym__in] = ACTIONS(6310),
  },
  [STATE(4306)] = {
    [sym_block_comment] = STATE(4306),
    [sym_compiler_directive_decl] = STATE(4306),
    [sym_fsi_directive_decl] = STATE(4306),
    [sym_preproc_line] = STATE(4306),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4307)] = {
    [sym_block_comment] = STATE(4307),
    [sym_compiler_directive_decl] = STATE(4307),
    [sym_fsi_directive_decl] = STATE(4307),
    [sym_preproc_line] = STATE(4307),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_DOT_DOT] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8720),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__dedent] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4308)] = {
    [sym_block_comment] = STATE(4308),
    [sym_compiler_directive_decl] = STATE(4308),
    [sym_fsi_directive_decl] = STATE(4308),
    [sym_preproc_line] = STATE(4308),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_DOT_DOT] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__dedent] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4309)] = {
    [sym_block_comment] = STATE(4309),
    [sym_compiler_directive_decl] = STATE(4309),
    [sym_fsi_directive_decl] = STATE(4309),
    [sym_preproc_line] = STATE(4309),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_DOT_DOT] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__dedent] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4310)] = {
    [sym_block_comment] = STATE(4310),
    [sym_compiler_directive_decl] = STATE(4310),
    [sym_fsi_directive_decl] = STATE(4310),
    [sym_preproc_line] = STATE(4310),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_DOT_DOT] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__dedent] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4311)] = {
    [sym_block_comment] = STATE(4311),
    [sym_compiler_directive_decl] = STATE(4311),
    [sym_fsi_directive_decl] = STATE(4311),
    [sym_preproc_line] = STATE(4311),
    [aux_sym_type_argument_repeat1] = STATE(4398),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8722),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__quoted_close] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4312)] = {
    [sym_block_comment] = STATE(4312),
    [sym_compiler_directive_decl] = STATE(4312),
    [sym_fsi_directive_decl] = STATE(4312),
    [sym_preproc_line] = STATE(4312),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_DOT_DOT] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4313)] = {
    [sym_block_comment] = STATE(4313),
    [sym_compiler_directive_decl] = STATE(4313),
    [sym_fsi_directive_decl] = STATE(4313),
    [sym_preproc_line] = STATE(4313),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_DOT_DOT] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__dedent] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4314)] = {
    [sym_block_comment] = STATE(4314),
    [sym_compiler_directive_decl] = STATE(4314),
    [sym_fsi_directive_decl] = STATE(4314),
    [sym_preproc_line] = STATE(4314),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_DOT_DOT] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__dedent] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4315)] = {
    [sym_block_comment] = STATE(4315),
    [sym_compiler_directive_decl] = STATE(4315),
    [sym_fsi_directive_decl] = STATE(4315),
    [sym_preproc_line] = STATE(4315),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_DOT_DOT] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__dedent] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4316)] = {
    [sym_block_comment] = STATE(4316),
    [sym_compiler_directive_decl] = STATE(4316),
    [sym_fsi_directive_decl] = STATE(4316),
    [sym_preproc_line] = STATE(4316),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_DOT_DOT] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__dedent] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4317)] = {
    [sym_block_comment] = STATE(4317),
    [sym_compiler_directive_decl] = STATE(4317),
    [sym_fsi_directive_decl] = STATE(4317),
    [sym_preproc_line] = STATE(4317),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_GT_RBRACK] = ACTIONS(7157),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_RBRACK] = ACTIONS(7157),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_RBRACE] = ACTIONS(7157),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_to] = ACTIONS(7155),
    [anon_sym_downto] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_DOT_DOT2] = ACTIONS(7157),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(4318)] = {
    [sym_block_comment] = STATE(4318),
    [sym_compiler_directive_decl] = STATE(4318),
    [sym_fsi_directive_decl] = STATE(4318),
    [sym_preproc_line] = STATE(4318),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_DOT_DOT] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__dedent] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4319)] = {
    [sym_block_comment] = STATE(4319),
    [sym_compiler_directive_decl] = STATE(4319),
    [sym_fsi_directive_decl] = STATE(4319),
    [sym_preproc_line] = STATE(4319),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_DOT_DOT] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__dedent] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4320)] = {
    [sym_block_comment] = STATE(4320),
    [sym_compiler_directive_decl] = STATE(4320),
    [sym_fsi_directive_decl] = STATE(4320),
    [sym_preproc_line] = STATE(4320),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_DOT_DOT] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__dedent] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4321)] = {
    [sym_block_comment] = STATE(4321),
    [sym_compiler_directive_decl] = STATE(4321),
    [sym_fsi_directive_decl] = STATE(4321),
    [sym_preproc_line] = STATE(4321),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_DOT_DOT] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__dedent] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4322)] = {
    [sym_block_comment] = STATE(4322),
    [sym_compiler_directive_decl] = STATE(4322),
    [sym_fsi_directive_decl] = STATE(4322),
    [sym_preproc_line] = STATE(4322),
    [aux_sym_long_identifier_repeat1] = STATE(4330),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_as] = ACTIONS(6360),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8725),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4323)] = {
    [sym_block_comment] = STATE(4323),
    [sym_compiler_directive_decl] = STATE(4323),
    [sym_fsi_directive_decl] = STATE(4323),
    [sym_preproc_line] = STATE(4323),
    [aux_sym_compound_type_repeat1] = STATE(4323),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8727),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__untyped_quoted_close] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4324)] = {
    [sym_block_comment] = STATE(4324),
    [sym_compiler_directive_decl] = STATE(4324),
    [sym_fsi_directive_decl] = STATE(4324),
    [sym_preproc_line] = STATE(4324),
    [aux_sym_type_argument_repeat1] = STATE(4325),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_DOT_DOT] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4325)] = {
    [sym_block_comment] = STATE(4325),
    [sym_compiler_directive_decl] = STATE(4325),
    [sym_fsi_directive_decl] = STATE(4325),
    [sym_preproc_line] = STATE(4325),
    [aux_sym_type_argument_repeat1] = STATE(4325),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8730),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4326)] = {
    [sym_block_comment] = STATE(4326),
    [sym_compiler_directive_decl] = STATE(4326),
    [sym_fsi_directive_decl] = STATE(4326),
    [sym_preproc_line] = STATE(4326),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4327)] = {
    [sym_block_comment] = STATE(4327),
    [sym_compiler_directive_decl] = STATE(4327),
    [sym_fsi_directive_decl] = STATE(4327),
    [sym_preproc_line] = STATE(4327),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_DOT_DOT] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__dedent] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4328)] = {
    [sym_block_comment] = STATE(4328),
    [sym_compiler_directive_decl] = STATE(4328),
    [sym_fsi_directive_decl] = STATE(4328),
    [sym_preproc_line] = STATE(4328),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__dedent] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4329)] = {
    [sym_block_comment] = STATE(4329),
    [sym_compiler_directive_decl] = STATE(4329),
    [sym_fsi_directive_decl] = STATE(4329),
    [sym_preproc_line] = STATE(4329),
    [aux_sym_long_identifier_repeat1] = STATE(4333),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8733),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_DOT_DOT] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4330)] = {
    [sym_block_comment] = STATE(4330),
    [sym_compiler_directive_decl] = STATE(4330),
    [sym_fsi_directive_decl] = STATE(4330),
    [sym_preproc_line] = STATE(4330),
    [aux_sym_long_identifier_repeat1] = STATE(4330),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8735),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4331)] = {
    [sym_block_comment] = STATE(4331),
    [sym_compiler_directive_decl] = STATE(4331),
    [sym_fsi_directive_decl] = STATE(4331),
    [sym_preproc_line] = STATE(4331),
    [aux_sym_type_argument_repeat1] = STATE(4336),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__multi_dollar_interp_end] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4332)] = {
    [sym_block_comment] = STATE(4332),
    [sym_compiler_directive_decl] = STATE(4332),
    [sym_fsi_directive_decl] = STATE(4332),
    [sym_preproc_line] = STATE(4332),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4333)] = {
    [sym_block_comment] = STATE(4333),
    [sym_compiler_directive_decl] = STATE(4333),
    [sym_fsi_directive_decl] = STATE(4333),
    [sym_preproc_line] = STATE(4333),
    [aux_sym_long_identifier_repeat1] = STATE(4334),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8733),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4334)] = {
    [sym_block_comment] = STATE(4334),
    [sym_compiler_directive_decl] = STATE(4334),
    [sym_fsi_directive_decl] = STATE(4334),
    [sym_preproc_line] = STATE(4334),
    [aux_sym_long_identifier_repeat1] = STATE(4334),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4335)] = {
    [sym_block_comment] = STATE(4335),
    [sym_compiler_directive_decl] = STATE(4335),
    [sym_fsi_directive_decl] = STATE(4335),
    [sym_preproc_line] = STATE(4335),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4336)] = {
    [sym_block_comment] = STATE(4336),
    [sym_compiler_directive_decl] = STATE(4336),
    [sym_fsi_directive_decl] = STATE(4336),
    [sym_preproc_line] = STATE(4336),
    [aux_sym_type_argument_repeat1] = STATE(4336),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8743),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__multi_dollar_interp_end] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4337)] = {
    [sym_block_comment] = STATE(4337),
    [sym_compiler_directive_decl] = STATE(4337),
    [sym_fsi_directive_decl] = STATE(4337),
    [sym_preproc_line] = STATE(4337),
    [aux_sym_type_argument_repeat1] = STATE(4340),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4338)] = {
    [sym_block_comment] = STATE(4338),
    [sym_compiler_directive_decl] = STATE(4338),
    [sym_fsi_directive_decl] = STATE(4338),
    [sym_preproc_line] = STATE(4338),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_GT_RBRACK] = ACTIONS(6805),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_RBRACK] = ACTIONS(6805),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_RBRACE] = ACTIONS(6805),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_to] = ACTIONS(6807),
    [anon_sym_downto] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_DOT_DOT2] = ACTIONS(6805),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4339)] = {
    [sym_block_comment] = STATE(4339),
    [sym_compiler_directive_decl] = STATE(4339),
    [sym_fsi_directive_decl] = STATE(4339),
    [sym_preproc_line] = STATE(4339),
    [aux_sym_long_identifier_repeat1] = STATE(4479),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8746),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4340)] = {
    [sym_block_comment] = STATE(4340),
    [sym_compiler_directive_decl] = STATE(4340),
    [sym_fsi_directive_decl] = STATE(4340),
    [sym_preproc_line] = STATE(4340),
    [aux_sym_type_argument_repeat1] = STATE(4340),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8748),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4341)] = {
    [sym__else_expression] = STATE(5952),
    [sym_elif_expression] = STATE(4962),
    [sym_block_comment] = STATE(4341),
    [sym_compiler_directive_decl] = STATE(4341),
    [sym_fsi_directive_decl] = STATE(4341),
    [sym_preproc_line] = STATE(4341),
    [aux_sym_if_expression_repeat1] = STATE(4440),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8751),
    [anon_sym_elif] = ACTIONS(8753),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__quoted_close] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4342)] = {
    [sym_block_comment] = STATE(4342),
    [sym_compiler_directive_decl] = STATE(4342),
    [sym_fsi_directive_decl] = STATE(4342),
    [sym_preproc_line] = STATE(4342),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_GT_RBRACK] = ACTIONS(6393),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8545),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4343)] = {
    [sym_block_comment] = STATE(4343),
    [sym_compiler_directive_decl] = STATE(4343),
    [sym_fsi_directive_decl] = STATE(4343),
    [sym_preproc_line] = STATE(4343),
    [aux_sym_compound_type_repeat1] = STATE(4375),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8375),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4344)] = {
    [sym__else_expression] = STATE(6199),
    [sym_elif_expression] = STATE(5077),
    [sym_block_comment] = STATE(4344),
    [sym_compiler_directive_decl] = STATE(4344),
    [sym_fsi_directive_decl] = STATE(4344),
    [sym_preproc_line] = STATE(4344),
    [aux_sym_if_expression_repeat1] = STATE(4455),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8755),
    [anon_sym_elif] = ACTIONS(8757),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__multi_dollar_interp_end] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4345)] = {
    [sym_block_comment] = STATE(4345),
    [sym_compiler_directive_decl] = STATE(4345),
    [sym_fsi_directive_decl] = STATE(4345),
    [sym_preproc_line] = STATE(4345),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_GT_RBRACK] = ACTIONS(6890),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_RBRACK] = ACTIONS(6890),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_RBRACE] = ACTIONS(6890),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_to] = ACTIONS(6892),
    [anon_sym_downto] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_DOT_DOT2] = ACTIONS(6890),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(4346)] = {
    [sym_block_comment] = STATE(4346),
    [sym_compiler_directive_decl] = STATE(4346),
    [sym_fsi_directive_decl] = STATE(4346),
    [sym_preproc_line] = STATE(4346),
    [aux_sym_long_identifier_repeat1] = STATE(4453),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8759),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4347)] = {
    [sym_attributes] = STATE(13389),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7140),
    [sym_access_modifier] = STATE(22369),
    [sym__class_function_or_value_defn] = STATE(7141),
    [aux_sym__type_extension_inner] = STATE(4349),
    [sym__type_defn_elements] = STATE(7141),
    [sym_interface_implementation] = STATE(6973),
    [sym_member_defn] = STATE(6973),
    [sym_additional_constr_defn] = STATE(7045),
    [sym_block_comment] = STATE(4347),
    [sym_compiler_directive_decl] = STATE(4347),
    [sym_fsi_directive_decl] = STATE(4347),
    [sym_preproc_line] = STATE(4347),
    [sym_preproc_if_in_class_definition] = STATE(6973),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8620),
    [anon_sym_module] = ACTIONS(8620),
    [anon_sym_null] = ACTIONS(8620),
    [anon_sym_begin] = ACTIONS(8620),
    [anon_sym_open] = ACTIONS(8620),
    [anon_sym_type] = ACTIONS(8620),
    [anon_sym_LBRACK_LT] = ACTIONS(8622),
    [anon_sym_do] = ACTIONS(8620),
    [anon_sym_and] = ACTIONS(8620),
    [anon_sym_let] = ACTIONS(8620),
    [anon_sym_let_BANG] = ACTIONS(8622),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8620),
    [anon_sym_AMP] = ACTIONS(8620),
    [anon_sym_LBRACK] = ACTIONS(8620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8622),
    [anon_sym_LBRACE] = ACTIONS(8620),
    [anon_sym_LT_AT] = ACTIONS(8620),
    [anon_sym_LT_AT_AT] = ACTIONS(8622),
    [anon_sym_CARET] = ACTIONS(8622),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8622),
    [anon_sym_new] = ACTIONS(8620),
    [anon_sym_return] = ACTIONS(8620),
    [anon_sym_return_BANG] = ACTIONS(8622),
    [anon_sym_yield] = ACTIONS(8620),
    [anon_sym_yield_BANG] = ACTIONS(8622),
    [anon_sym_lazy] = ACTIONS(8620),
    [anon_sym_assert] = ACTIONS(8620),
    [anon_sym_upcast] = ACTIONS(8620),
    [anon_sym_downcast] = ACTIONS(8620),
    [anon_sym_fixed] = ACTIONS(8620),
    [anon_sym_for] = ACTIONS(8620),
    [anon_sym_while] = ACTIONS(8620),
    [anon_sym_while_BANG] = ACTIONS(8622),
    [anon_sym_if] = ACTIONS(8620),
    [anon_sym_fun] = ACTIONS(8620),
    [anon_sym_try] = ACTIONS(8620),
    [anon_sym_match] = ACTIONS(8620),
    [anon_sym_match_BANG] = ACTIONS(8622),
    [anon_sym_function] = ACTIONS(8620),
    [anon_sym_use] = ACTIONS(8620),
    [anon_sym_use_BANG] = ACTIONS(8622),
    [anon_sym_do_BANG] = ACTIONS(8622),
    [anon_sym_AMP_AMP] = ACTIONS(8622),
    [anon_sym_default] = ACTIONS(8458),
    [anon_sym_static] = ACTIONS(8460),
    [anon_sym_member] = ACTIONS(8462),
    [anon_sym_exception] = ACTIONS(8620),
    [anon_sym_interface] = ACTIONS(8464),
    [anon_sym_abstract] = ACTIONS(8466),
    [anon_sym_val] = ACTIONS(8468),
    [anon_sym_override] = ACTIONS(8458),
    [anon_sym_extern] = ACTIONS(8620),
    [aux_sym_char_token1] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8620),
    [anon_sym_DQUOTE] = ACTIONS(8620),
    [anon_sym_AT_DQUOTE] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [sym_bool] = ACTIONS(8620),
    [sym_unit] = ACTIONS(8622),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8620),
    [sym_op_identifier] = ACTIONS(8622),
    [anon_sym_PLUS] = ACTIONS(8620),
    [anon_sym_DASH] = ACTIONS(8620),
    [anon_sym_PLUS_DOT] = ACTIONS(8622),
    [anon_sym_DASH_DOT] = ACTIONS(8622),
    [anon_sym_PERCENT] = ACTIONS(8620),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8622),
    [anon_sym_TILDE] = ACTIONS(8622),
    [aux_sym_prefix_op_token1] = ACTIONS(8622),
    [sym_int] = ACTIONS(8620),
    [sym_xint] = ACTIONS(8622),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8622),
    [anon_sym_POUNDload] = ACTIONS(8622),
    [anon_sym_POUNDtime] = ACTIONS(8622),
    [anon_sym_POUNDI] = ACTIONS(8622),
    [anon_sym_POUNDhelp] = ACTIONS(8622),
    [anon_sym_POUNDquit] = ACTIONS(8622),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8622),
    [anon_sym_POUNDendif] = ACTIONS(8622),
    [anon_sym_POUNDelse] = ACTIONS(8622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8622),
  },
  [STATE(4348)] = {
    [sym_elif_expression] = STATE(4842),
    [sym_block_comment] = STATE(4348),
    [sym_compiler_directive_decl] = STATE(4348),
    [sym_fsi_directive_decl] = STATE(4348),
    [sym_preproc_line] = STATE(4348),
    [aux_sym_if_expression_repeat1] = STATE(4348),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_as] = ACTIONS(6496),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8761),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4349)] = {
    [sym_attributes] = STATE(13389),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7140),
    [sym_access_modifier] = STATE(22369),
    [sym__class_function_or_value_defn] = STATE(7141),
    [aux_sym__type_extension_inner] = STATE(4349),
    [sym__type_defn_elements] = STATE(7141),
    [sym_interface_implementation] = STATE(6973),
    [sym_member_defn] = STATE(6973),
    [sym_additional_constr_defn] = STATE(7045),
    [sym_block_comment] = STATE(4349),
    [sym_compiler_directive_decl] = STATE(4349),
    [sym_fsi_directive_decl] = STATE(4349),
    [sym_preproc_line] = STATE(4349),
    [sym_preproc_if_in_class_definition] = STATE(6973),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8626),
    [anon_sym_module] = ACTIONS(8626),
    [anon_sym_null] = ACTIONS(8626),
    [anon_sym_begin] = ACTIONS(8626),
    [anon_sym_open] = ACTIONS(8626),
    [anon_sym_type] = ACTIONS(8626),
    [anon_sym_LBRACK_LT] = ACTIONS(8628),
    [anon_sym_do] = ACTIONS(8764),
    [anon_sym_and] = ACTIONS(8626),
    [anon_sym_let] = ACTIONS(8767),
    [anon_sym_let_BANG] = ACTIONS(8770),
    [aux_sym_access_modifier_token1] = ACTIONS(8640),
    [anon_sym_LPAREN] = ACTIONS(8626),
    [anon_sym_AMP] = ACTIONS(8626),
    [anon_sym_LBRACK] = ACTIONS(8626),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8643),
    [anon_sym_LBRACE] = ACTIONS(8626),
    [anon_sym_LT_AT] = ACTIONS(8626),
    [anon_sym_LT_AT_AT] = ACTIONS(8643),
    [anon_sym_CARET] = ACTIONS(8643),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8643),
    [anon_sym_new] = ACTIONS(8773),
    [anon_sym_return] = ACTIONS(8626),
    [anon_sym_return_BANG] = ACTIONS(8643),
    [anon_sym_yield] = ACTIONS(8626),
    [anon_sym_yield_BANG] = ACTIONS(8643),
    [anon_sym_lazy] = ACTIONS(8626),
    [anon_sym_assert] = ACTIONS(8626),
    [anon_sym_upcast] = ACTIONS(8626),
    [anon_sym_downcast] = ACTIONS(8626),
    [anon_sym_fixed] = ACTIONS(8626),
    [anon_sym_for] = ACTIONS(8626),
    [anon_sym_while] = ACTIONS(8626),
    [anon_sym_while_BANG] = ACTIONS(8643),
    [anon_sym_if] = ACTIONS(8626),
    [anon_sym_fun] = ACTIONS(8626),
    [anon_sym_try] = ACTIONS(8626),
    [anon_sym_match] = ACTIONS(8626),
    [anon_sym_match_BANG] = ACTIONS(8643),
    [anon_sym_function] = ACTIONS(8626),
    [anon_sym_use] = ACTIONS(8626),
    [anon_sym_use_BANG] = ACTIONS(8643),
    [anon_sym_do_BANG] = ACTIONS(8643),
    [anon_sym_AMP_AMP] = ACTIONS(8643),
    [anon_sym_default] = ACTIONS(8776),
    [anon_sym_static] = ACTIONS(8779),
    [anon_sym_member] = ACTIONS(8782),
    [anon_sym_exception] = ACTIONS(8626),
    [anon_sym_interface] = ACTIONS(8785),
    [anon_sym_abstract] = ACTIONS(8788),
    [anon_sym_val] = ACTIONS(8791),
    [anon_sym_override] = ACTIONS(8776),
    [anon_sym_extern] = ACTIONS(8626),
    [aux_sym_char_token1] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8626),
    [anon_sym_DQUOTE] = ACTIONS(8626),
    [anon_sym_AT_DQUOTE] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [sym_bool] = ACTIONS(8626),
    [sym_unit] = ACTIONS(8643),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8626),
    [sym_op_identifier] = ACTIONS(8643),
    [anon_sym_PLUS] = ACTIONS(8626),
    [anon_sym_DASH] = ACTIONS(8626),
    [anon_sym_PLUS_DOT] = ACTIONS(8643),
    [anon_sym_DASH_DOT] = ACTIONS(8643),
    [anon_sym_PERCENT] = ACTIONS(8626),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8643),
    [anon_sym_TILDE] = ACTIONS(8643),
    [aux_sym_prefix_op_token1] = ACTIONS(8643),
    [sym_int] = ACTIONS(8626),
    [sym_xint] = ACTIONS(8643),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8643),
    [anon_sym_POUNDload] = ACTIONS(8643),
    [anon_sym_POUNDtime] = ACTIONS(8643),
    [anon_sym_POUNDI] = ACTIONS(8643),
    [anon_sym_POUNDhelp] = ACTIONS(8643),
    [anon_sym_POUNDquit] = ACTIONS(8643),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8794),
    [anon_sym_POUNDendif] = ACTIONS(8643),
    [anon_sym_POUNDelse] = ACTIONS(8643),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8643),
  },
  [STATE(4350)] = {
    [sym_block_comment] = STATE(4350),
    [sym_compiler_directive_decl] = STATE(4350),
    [sym_fsi_directive_decl] = STATE(4350),
    [sym_preproc_line] = STATE(4350),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_GT_RBRACK] = ACTIONS(6894),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_RBRACK] = ACTIONS(6894),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_RBRACE] = ACTIONS(6894),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_to] = ACTIONS(6896),
    [anon_sym_downto] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_DOT_DOT2] = ACTIONS(6894),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(4351)] = {
    [sym_block_comment] = STATE(4351),
    [sym_compiler_directive_decl] = STATE(4351),
    [sym_fsi_directive_decl] = STATE(4351),
    [sym_preproc_line] = STATE(4351),
    [aux_sym_long_identifier_repeat1] = STATE(4504),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8797),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__multi_dollar_interp_end] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4352)] = {
    [sym_block_comment] = STATE(4352),
    [sym_compiler_directive_decl] = STATE(4352),
    [sym_fsi_directive_decl] = STATE(4352),
    [sym_preproc_line] = STATE(4352),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_GT_RBRACK] = ACTIONS(6898),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_RBRACK] = ACTIONS(6898),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_RBRACE] = ACTIONS(6898),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_to] = ACTIONS(6900),
    [anon_sym_downto] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_DOT_DOT2] = ACTIONS(6898),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(4353)] = {
    [sym_block_comment] = STATE(4353),
    [sym_compiler_directive_decl] = STATE(4353),
    [sym_fsi_directive_decl] = STATE(4353),
    [sym_preproc_line] = STATE(4353),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_GT_RBRACK] = ACTIONS(6603),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_RBRACK] = ACTIONS(6603),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_RBRACE] = ACTIONS(6603),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_to] = ACTIONS(6902),
    [anon_sym_downto] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_DOT_DOT2] = ACTIONS(6603),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4354)] = {
    [sym_block_comment] = STATE(4354),
    [sym_compiler_directive_decl] = STATE(4354),
    [sym_fsi_directive_decl] = STATE(4354),
    [sym_preproc_line] = STATE(4354),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6513),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_DOT_DOT2] = ACTIONS(6515),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(8799),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4355)] = {
    [sym_block_comment] = STATE(4355),
    [sym_compiler_directive_decl] = STATE(4355),
    [sym_fsi_directive_decl] = STATE(4355),
    [sym_preproc_line] = STATE(4355),
    [aux_sym_long_identifier_repeat1] = STATE(4322),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_as] = ACTIONS(6625),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8801),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4356)] = {
    [sym_block_comment] = STATE(4356),
    [sym_compiler_directive_decl] = STATE(4356),
    [sym_fsi_directive_decl] = STATE(4356),
    [sym_preproc_line] = STATE(4356),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_GT_RBRACK] = ACTIONS(7147),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_RBRACK] = ACTIONS(7147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_RBRACE] = ACTIONS(7147),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_to] = ACTIONS(7145),
    [anon_sym_downto] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_DOT_DOT2] = ACTIONS(7147),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(4357)] = {
    [sym_block_comment] = STATE(4357),
    [sym_compiler_directive_decl] = STATE(4357),
    [sym_fsi_directive_decl] = STATE(4357),
    [sym_preproc_line] = STATE(4357),
    [aux_sym_rules_repeat1] = STATE(4444),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_as] = ACTIONS(6607),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8805),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8807),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4358)] = {
    [sym_elif_expression] = STATE(4886),
    [sym_block_comment] = STATE(4358),
    [sym_compiler_directive_decl] = STATE(4358),
    [sym_fsi_directive_decl] = STATE(4358),
    [sym_preproc_line] = STATE(4358),
    [aux_sym_if_expression_repeat1] = STATE(4358),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8810),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [anon_sym_POUNDelse] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4359)] = {
    [sym_block_comment] = STATE(4359),
    [sym_compiler_directive_decl] = STATE(4359),
    [sym_fsi_directive_decl] = STATE(4359),
    [sym_preproc_line] = STATE(4359),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_GT_RBRACK] = ACTIONS(6904),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_RBRACK] = ACTIONS(6904),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_RBRACE] = ACTIONS(6904),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_to] = ACTIONS(6906),
    [anon_sym_downto] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_DOT_DOT2] = ACTIONS(6904),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(4360)] = {
    [sym_block_comment] = STATE(4360),
    [sym_compiler_directive_decl] = STATE(4360),
    [sym_fsi_directive_decl] = STATE(4360),
    [sym_preproc_line] = STATE(4360),
    [aux_sym_sequential_expression_repeat1] = STATE(4369),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_GT_RBRACK] = ACTIONS(6738),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_RBRACK] = ACTIONS(6738),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_RBRACE] = ACTIONS(6738),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_to] = ACTIONS(6740),
    [anon_sym_downto] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(4361)] = {
    [sym_block_comment] = STATE(4361),
    [sym_compiler_directive_decl] = STATE(4361),
    [sym_fsi_directive_decl] = STATE(4361),
    [sym_preproc_line] = STATE(4361),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_GT_RBRACK] = ACTIONS(6911),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_RBRACK] = ACTIONS(6911),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_RBRACE] = ACTIONS(6911),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_to] = ACTIONS(6913),
    [anon_sym_downto] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_DOT_DOT2] = ACTIONS(6911),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(4362)] = {
    [sym_block_comment] = STATE(4362),
    [sym_compiler_directive_decl] = STATE(4362),
    [sym_fsi_directive_decl] = STATE(4362),
    [sym_preproc_line] = STATE(4362),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_GT_RBRACK] = ACTIONS(6915),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_RBRACK] = ACTIONS(6915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_RBRACE] = ACTIONS(6915),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_to] = ACTIONS(6917),
    [anon_sym_downto] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_DOT_DOT2] = ACTIONS(6915),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(4363)] = {
    [sym_block_comment] = STATE(4363),
    [sym_compiler_directive_decl] = STATE(4363),
    [sym_fsi_directive_decl] = STATE(4363),
    [sym_preproc_line] = STATE(4363),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_GT_RBRACK] = ACTIONS(6919),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_RBRACK] = ACTIONS(6919),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_RBRACE] = ACTIONS(6919),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_to] = ACTIONS(6921),
    [anon_sym_downto] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_DOT_DOT2] = ACTIONS(6919),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(4364)] = {
    [sym_block_comment] = STATE(4364),
    [sym_compiler_directive_decl] = STATE(4364),
    [sym_fsi_directive_decl] = STATE(4364),
    [sym_preproc_line] = STATE(4364),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_GT_RBRACK] = ACTIONS(6923),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_RBRACK] = ACTIONS(6923),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_RBRACE] = ACTIONS(6923),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_to] = ACTIONS(6925),
    [anon_sym_downto] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_DOT_DOT2] = ACTIONS(6923),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(4365)] = {
    [sym_block_comment] = STATE(4365),
    [sym_compiler_directive_decl] = STATE(4365),
    [sym_fsi_directive_decl] = STATE(4365),
    [sym_preproc_line] = STATE(4365),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_GT_RBRACK] = ACTIONS(292),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_RBRACK] = ACTIONS(292),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_RBRACE] = ACTIONS(292),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_to] = ACTIONS(290),
    [anon_sym_downto] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_DOT_DOT2] = ACTIONS(292),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(4366)] = {
    [sym_block_comment] = STATE(4366),
    [sym_compiler_directive_decl] = STATE(4366),
    [sym_fsi_directive_decl] = STATE(4366),
    [sym_preproc_line] = STATE(4366),
    [aux_sym_rules_repeat1] = STATE(4446),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_as] = ACTIONS(6607),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(8805),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(8807),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4367)] = {
    [sym_block_comment] = STATE(4367),
    [sym_compiler_directive_decl] = STATE(4367),
    [sym_fsi_directive_decl] = STATE(4367),
    [sym_preproc_line] = STATE(4367),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_GT_RBRACK] = ACTIONS(6930),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_RBRACK] = ACTIONS(6930),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_RBRACE] = ACTIONS(6930),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_to] = ACTIONS(6932),
    [anon_sym_downto] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_DOT_DOT2] = ACTIONS(6930),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(4368)] = {
    [sym_block_comment] = STATE(4368),
    [sym_compiler_directive_decl] = STATE(4368),
    [sym_fsi_directive_decl] = STATE(4368),
    [sym_preproc_line] = STATE(4368),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4369)] = {
    [sym_block_comment] = STATE(4369),
    [sym_compiler_directive_decl] = STATE(4369),
    [sym_fsi_directive_decl] = STATE(4369),
    [sym_preproc_line] = STATE(4369),
    [aux_sym_sequential_expression_repeat1] = STATE(4369),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_GT_RBRACK] = ACTIONS(6798),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_RBRACK] = ACTIONS(6798),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_RBRACE] = ACTIONS(6798),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_to] = ACTIONS(6800),
    [anon_sym_downto] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(8813),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(4370)] = {
    [sym_block_comment] = STATE(4370),
    [sym_compiler_directive_decl] = STATE(4370),
    [sym_fsi_directive_decl] = STATE(4370),
    [sym_preproc_line] = STATE(4370),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_GT_RBRACK] = ACTIONS(6942),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_RBRACK] = ACTIONS(6942),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_RBRACE] = ACTIONS(6942),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_to] = ACTIONS(6944),
    [anon_sym_downto] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_DOT_DOT2] = ACTIONS(6942),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(4371)] = {
    [sym_block_comment] = STATE(4371),
    [sym_compiler_directive_decl] = STATE(4371),
    [sym_fsi_directive_decl] = STATE(4371),
    [sym_preproc_line] = STATE(4371),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_GT_RBRACK] = ACTIONS(6946),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_RBRACK] = ACTIONS(6946),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_RBRACE] = ACTIONS(6946),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_to] = ACTIONS(6948),
    [anon_sym_downto] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_DOT_DOT2] = ACTIONS(6946),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(4372)] = {
    [sym_block_comment] = STATE(4372),
    [sym_compiler_directive_decl] = STATE(4372),
    [sym_fsi_directive_decl] = STATE(4372),
    [sym_preproc_line] = STATE(4372),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_GT_RBRACK] = ACTIONS(6950),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_RBRACK] = ACTIONS(6950),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_RBRACE] = ACTIONS(6950),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_to] = ACTIONS(6952),
    [anon_sym_downto] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_DOT_DOT2] = ACTIONS(6950),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(4373)] = {
    [sym_block_comment] = STATE(4373),
    [sym_compiler_directive_decl] = STATE(4373),
    [sym_fsi_directive_decl] = STATE(4373),
    [sym_preproc_line] = STATE(4373),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_GT_RBRACK] = ACTIONS(6957),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_RBRACK] = ACTIONS(6957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_RBRACE] = ACTIONS(6957),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_to] = ACTIONS(6959),
    [anon_sym_downto] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_DOT_DOT2] = ACTIONS(6957),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4374)] = {
    [sym_block_comment] = STATE(4374),
    [sym_compiler_directive_decl] = STATE(4374),
    [sym_fsi_directive_decl] = STATE(4374),
    [sym_preproc_line] = STATE(4374),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_GT_RBRACK] = ACTIONS(6961),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_RBRACK] = ACTIONS(6961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_RBRACE] = ACTIONS(6961),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_to] = ACTIONS(6963),
    [anon_sym_downto] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_DOT_DOT2] = ACTIONS(6961),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(4375)] = {
    [sym_block_comment] = STATE(4375),
    [sym_compiler_directive_decl] = STATE(4375),
    [sym_fsi_directive_decl] = STATE(4375),
    [sym_preproc_line] = STATE(4375),
    [aux_sym_compound_type_repeat1] = STATE(4375),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8816),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4376)] = {
    [sym_block_comment] = STATE(4376),
    [sym_compiler_directive_decl] = STATE(4376),
    [sym_fsi_directive_decl] = STATE(4376),
    [sym_preproc_line] = STATE(4376),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_GT_RBRACK] = ACTIONS(6965),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_RBRACK] = ACTIONS(6965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_RBRACE] = ACTIONS(6965),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_to] = ACTIONS(6967),
    [anon_sym_downto] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_DOT_DOT2] = ACTIONS(6965),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(4377)] = {
    [sym_block_comment] = STATE(4377),
    [sym_compiler_directive_decl] = STATE(4377),
    [sym_fsi_directive_decl] = STATE(4377),
    [sym_preproc_line] = STATE(4377),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_GT_RBRACK] = ACTIONS(6969),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_RBRACK] = ACTIONS(6969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_RBRACE] = ACTIONS(6969),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_to] = ACTIONS(6971),
    [anon_sym_downto] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_DOT_DOT2] = ACTIONS(6969),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(4378)] = {
    [sym_block_comment] = STATE(4378),
    [sym_compiler_directive_decl] = STATE(4378),
    [sym_fsi_directive_decl] = STATE(4378),
    [sym_preproc_line] = STATE(4378),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_GT_RBRACK] = ACTIONS(6973),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_RBRACK] = ACTIONS(6973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_RBRACE] = ACTIONS(6973),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_to] = ACTIONS(6975),
    [anon_sym_downto] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_DOT_DOT2] = ACTIONS(6973),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(4379)] = {
    [sym_block_comment] = STATE(4379),
    [sym_compiler_directive_decl] = STATE(4379),
    [sym_fsi_directive_decl] = STATE(4379),
    [sym_preproc_line] = STATE(4379),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_GT_RBRACK] = ACTIONS(6977),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_RBRACK] = ACTIONS(6977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_RBRACE] = ACTIONS(6977),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_to] = ACTIONS(6979),
    [anon_sym_downto] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_DOT_DOT2] = ACTIONS(6977),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(4380)] = {
    [sym_block_comment] = STATE(4380),
    [sym_compiler_directive_decl] = STATE(4380),
    [sym_fsi_directive_decl] = STATE(4380),
    [sym_preproc_line] = STATE(4380),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_GT_RBRACK] = ACTIONS(6981),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_RBRACK] = ACTIONS(6981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_RBRACE] = ACTIONS(6981),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_to] = ACTIONS(6983),
    [anon_sym_downto] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_DOT_DOT2] = ACTIONS(6981),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(4381)] = {
    [sym_block_comment] = STATE(4381),
    [sym_compiler_directive_decl] = STATE(4381),
    [sym_fsi_directive_decl] = STATE(4381),
    [sym_preproc_line] = STATE(4381),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_GT_RBRACK] = ACTIONS(6985),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_RBRACK] = ACTIONS(6985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_RBRACE] = ACTIONS(6985),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_to] = ACTIONS(6987),
    [anon_sym_downto] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_DOT_DOT2] = ACTIONS(6985),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(4382)] = {
    [sym_block_comment] = STATE(4382),
    [sym_compiler_directive_decl] = STATE(4382),
    [sym_fsi_directive_decl] = STATE(4382),
    [sym_preproc_line] = STATE(4382),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_GT_RBRACK] = ACTIONS(6989),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_RBRACK] = ACTIONS(6989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_RBRACE] = ACTIONS(6989),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_to] = ACTIONS(6991),
    [anon_sym_downto] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_DOT_DOT2] = ACTIONS(6989),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(4383)] = {
    [sym_block_comment] = STATE(4383),
    [sym_compiler_directive_decl] = STATE(4383),
    [sym_fsi_directive_decl] = STATE(4383),
    [sym_preproc_line] = STATE(4383),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_GT_RBRACK] = ACTIONS(6993),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_RBRACK] = ACTIONS(6993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_RBRACE] = ACTIONS(6993),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_to] = ACTIONS(6995),
    [anon_sym_downto] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_DOT_DOT2] = ACTIONS(6993),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(4384)] = {
    [sym_block_comment] = STATE(4384),
    [sym_compiler_directive_decl] = STATE(4384),
    [sym_fsi_directive_decl] = STATE(4384),
    [sym_preproc_line] = STATE(4384),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_GT_RBRACK] = ACTIONS(6997),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_RBRACK] = ACTIONS(6997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_RBRACE] = ACTIONS(6997),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_to] = ACTIONS(6999),
    [anon_sym_downto] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_DOT_DOT2] = ACTIONS(6997),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(4385)] = {
    [sym_block_comment] = STATE(4385),
    [sym_compiler_directive_decl] = STATE(4385),
    [sym_fsi_directive_decl] = STATE(4385),
    [sym_preproc_line] = STATE(4385),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_GT_RBRACK] = ACTIONS(7001),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_RBRACK] = ACTIONS(7001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_RBRACE] = ACTIONS(7001),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_to] = ACTIONS(7003),
    [anon_sym_downto] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_DOT_DOT2] = ACTIONS(7001),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(4386)] = {
    [sym_block_comment] = STATE(4386),
    [sym_compiler_directive_decl] = STATE(4386),
    [sym_fsi_directive_decl] = STATE(4386),
    [sym_preproc_line] = STATE(4386),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_GT_RBRACK] = ACTIONS(7005),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_RBRACK] = ACTIONS(7005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_RBRACE] = ACTIONS(7005),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_to] = ACTIONS(7007),
    [anon_sym_downto] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_DOT_DOT2] = ACTIONS(7005),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(4387)] = {
    [sym_block_comment] = STATE(4387),
    [sym_compiler_directive_decl] = STATE(4387),
    [sym_fsi_directive_decl] = STATE(4387),
    [sym_preproc_line] = STATE(4387),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_GT_RBRACK] = ACTIONS(7009),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_RBRACK] = ACTIONS(7009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_RBRACE] = ACTIONS(7009),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_to] = ACTIONS(7011),
    [anon_sym_downto] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_DOT_DOT2] = ACTIONS(7009),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(4388)] = {
    [sym_block_comment] = STATE(4388),
    [sym_compiler_directive_decl] = STATE(4388),
    [sym_fsi_directive_decl] = STATE(4388),
    [sym_preproc_line] = STATE(4388),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4389)] = {
    [sym_block_comment] = STATE(4389),
    [sym_compiler_directive_decl] = STATE(4389),
    [sym_fsi_directive_decl] = STATE(4389),
    [sym_preproc_line] = STATE(4389),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_GT_RBRACK] = ACTIONS(7017),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_RBRACK] = ACTIONS(7017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_RBRACE] = ACTIONS(7017),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_to] = ACTIONS(7019),
    [anon_sym_downto] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_DOT_DOT2] = ACTIONS(7017),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(4390)] = {
    [sym_block_comment] = STATE(4390),
    [sym_compiler_directive_decl] = STATE(4390),
    [sym_fsi_directive_decl] = STATE(4390),
    [sym_preproc_line] = STATE(4390),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_GT_RBRACK] = ACTIONS(7021),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_RBRACK] = ACTIONS(7021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_RBRACE] = ACTIONS(7021),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_to] = ACTIONS(7023),
    [anon_sym_downto] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_DOT_DOT2] = ACTIONS(7021),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(4391)] = {
    [sym_block_comment] = STATE(4391),
    [sym_compiler_directive_decl] = STATE(4391),
    [sym_fsi_directive_decl] = STATE(4391),
    [sym_preproc_line] = STATE(4391),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_GT_RBRACK] = ACTIONS(7025),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_RBRACK] = ACTIONS(7025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_RBRACE] = ACTIONS(7025),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_to] = ACTIONS(7027),
    [anon_sym_downto] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_DOT_DOT2] = ACTIONS(7025),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(4392)] = {
    [sym_block_comment] = STATE(4392),
    [sym_compiler_directive_decl] = STATE(4392),
    [sym_fsi_directive_decl] = STATE(4392),
    [sym_preproc_line] = STATE(4392),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_GT_RBRACK] = ACTIONS(7033),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_RBRACK] = ACTIONS(7033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_RBRACE] = ACTIONS(7033),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_to] = ACTIONS(7035),
    [anon_sym_downto] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_DOT_DOT2] = ACTIONS(7033),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(4393)] = {
    [sym_block_comment] = STATE(4393),
    [sym_compiler_directive_decl] = STATE(4393),
    [sym_fsi_directive_decl] = STATE(4393),
    [sym_preproc_line] = STATE(4393),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_GT_RBRACK] = ACTIONS(7037),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_RBRACK] = ACTIONS(7037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_RBRACE] = ACTIONS(7037),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_to] = ACTIONS(7039),
    [anon_sym_downto] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_DOT_DOT2] = ACTIONS(7037),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(4394)] = {
    [sym_block_comment] = STATE(4394),
    [sym_compiler_directive_decl] = STATE(4394),
    [sym_fsi_directive_decl] = STATE(4394),
    [sym_preproc_line] = STATE(4394),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_GT_RBRACK] = ACTIONS(7041),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_RBRACK] = ACTIONS(7041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_RBRACE] = ACTIONS(7041),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_to] = ACTIONS(7043),
    [anon_sym_downto] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_DOT_DOT2] = ACTIONS(7041),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(4395)] = {
    [sym_block_comment] = STATE(4395),
    [sym_compiler_directive_decl] = STATE(4395),
    [sym_fsi_directive_decl] = STATE(4395),
    [sym_preproc_line] = STATE(4395),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_GT_RBRACK] = ACTIONS(7057),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_RBRACK] = ACTIONS(7057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_RBRACE] = ACTIONS(7057),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_to] = ACTIONS(7059),
    [anon_sym_downto] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_DOT_DOT2] = ACTIONS(7057),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(4396)] = {
    [sym_block_comment] = STATE(4396),
    [sym_compiler_directive_decl] = STATE(4396),
    [sym_fsi_directive_decl] = STATE(4396),
    [sym_preproc_line] = STATE(4396),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_GT_RBRACK] = ACTIONS(7055),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_RBRACK] = ACTIONS(7055),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_RBRACE] = ACTIONS(7055),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_to] = ACTIONS(7053),
    [anon_sym_downto] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_DOT_DOT2] = ACTIONS(7055),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(4397)] = {
    [sym_block_comment] = STATE(4397),
    [sym_compiler_directive_decl] = STATE(4397),
    [sym_fsi_directive_decl] = STATE(4397),
    [sym_preproc_line] = STATE(4397),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_GT_RBRACK] = ACTIONS(5953),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_RBRACK] = ACTIONS(5953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_RBRACE] = ACTIONS(5953),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_to] = ACTIONS(5951),
    [anon_sym_downto] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4398)] = {
    [sym_block_comment] = STATE(4398),
    [sym_compiler_directive_decl] = STATE(4398),
    [sym_fsi_directive_decl] = STATE(4398),
    [sym_preproc_line] = STATE(4398),
    [aux_sym_type_argument_repeat1] = STATE(4403),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__quoted_close] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4399)] = {
    [sym_block_comment] = STATE(4399),
    [sym_compiler_directive_decl] = STATE(4399),
    [sym_fsi_directive_decl] = STATE(4399),
    [sym_preproc_line] = STATE(4399),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_GT_RBRACK] = ACTIONS(6886),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_RBRACK] = ACTIONS(6886),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_RBRACE] = ACTIONS(6886),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_to] = ACTIONS(6888),
    [anon_sym_downto] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_DOT_DOT2] = ACTIONS(6886),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(4400)] = {
    [sym_block_comment] = STATE(4400),
    [sym_compiler_directive_decl] = STATE(4400),
    [sym_fsi_directive_decl] = STATE(4400),
    [sym_preproc_line] = STATE(4400),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_GT_RBRACK] = ACTIONS(7071),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_RBRACK] = ACTIONS(7071),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_RBRACE] = ACTIONS(7071),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_to] = ACTIONS(7069),
    [anon_sym_downto] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_DOT_DOT2] = ACTIONS(7071),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(4401)] = {
    [sym_block_comment] = STATE(4401),
    [sym_compiler_directive_decl] = STATE(4401),
    [sym_fsi_directive_decl] = STATE(4401),
    [sym_preproc_line] = STATE(4401),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_GT_RBRACK] = ACTIONS(7075),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_RBRACK] = ACTIONS(7075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_RBRACE] = ACTIONS(7075),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_to] = ACTIONS(7073),
    [anon_sym_downto] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_DOT_DOT2] = ACTIONS(7075),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(4402)] = {
    [sym__else_expression] = STATE(6092),
    [sym_elif_expression] = STATE(4961),
    [sym_block_comment] = STATE(4402),
    [sym_compiler_directive_decl] = STATE(4402),
    [sym_fsi_directive_decl] = STATE(4402),
    [sym_preproc_line] = STATE(4402),
    [aux_sym_if_expression_repeat1] = STATE(4405),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_GT] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8819),
    [anon_sym_elif] = ACTIONS(8821),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4403)] = {
    [sym_block_comment] = STATE(4403),
    [sym_compiler_directive_decl] = STATE(4403),
    [sym_fsi_directive_decl] = STATE(4403),
    [sym_preproc_line] = STATE(4403),
    [aux_sym_type_argument_repeat1] = STATE(4403),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8823),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__quoted_close] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4404)] = {
    [sym__else_expression] = STATE(6130),
    [sym_elif_expression] = STATE(4977),
    [sym_block_comment] = STATE(4404),
    [sym_compiler_directive_decl] = STATE(4404),
    [sym_fsi_directive_decl] = STATE(4404),
    [sym_preproc_line] = STATE(4404),
    [aux_sym_if_expression_repeat1] = STATE(4406),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8826),
    [anon_sym_elif] = ACTIONS(8828),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_DOT_DOT] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4405)] = {
    [sym__else_expression] = STATE(6226),
    [sym_elif_expression] = STATE(4961),
    [sym_block_comment] = STATE(4405),
    [sym_compiler_directive_decl] = STATE(4405),
    [sym_fsi_directive_decl] = STATE(4405),
    [sym_preproc_line] = STATE(4405),
    [aux_sym_if_expression_repeat1] = STATE(4634),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_GT] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8819),
    [anon_sym_elif] = ACTIONS(8821),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4406)] = {
    [sym__else_expression] = STATE(6158),
    [sym_elif_expression] = STATE(4977),
    [sym_block_comment] = STATE(4406),
    [sym_compiler_directive_decl] = STATE(4406),
    [sym_fsi_directive_decl] = STATE(4406),
    [sym_preproc_line] = STATE(4406),
    [aux_sym_if_expression_repeat1] = STATE(4631),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8826),
    [anon_sym_elif] = ACTIONS(8828),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_DOT_DOT] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4407)] = {
    [sym_block_comment] = STATE(4407),
    [sym_compiler_directive_decl] = STATE(4407),
    [sym_fsi_directive_decl] = STATE(4407),
    [sym_preproc_line] = STATE(4407),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_GT_RBRACK] = ACTIONS(7079),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_RBRACK] = ACTIONS(7079),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_RBRACE] = ACTIONS(7079),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_to] = ACTIONS(7077),
    [anon_sym_downto] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_DOT_DOT2] = ACTIONS(7079),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(4408)] = {
    [sym_block_comment] = STATE(4408),
    [sym_compiler_directive_decl] = STATE(4408),
    [sym_fsi_directive_decl] = STATE(4408),
    [sym_preproc_line] = STATE(4408),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_GT_RBRACK] = ACTIONS(7087),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_RBRACK] = ACTIONS(7087),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_RBRACE] = ACTIONS(7087),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_to] = ACTIONS(7085),
    [anon_sym_downto] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_DOT_DOT2] = ACTIONS(7087),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(4409)] = {
    [sym_block_comment] = STATE(4409),
    [sym_compiler_directive_decl] = STATE(4409),
    [sym_fsi_directive_decl] = STATE(4409),
    [sym_preproc_line] = STATE(4409),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_GT_RBRACK] = ACTIONS(7091),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_RBRACK] = ACTIONS(7091),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_RBRACE] = ACTIONS(7091),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_to] = ACTIONS(7089),
    [anon_sym_downto] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_DOT_DOT2] = ACTIONS(7091),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(4410)] = {
    [sym_block_comment] = STATE(4410),
    [sym_compiler_directive_decl] = STATE(4410),
    [sym_fsi_directive_decl] = STATE(4410),
    [sym_preproc_line] = STATE(4410),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_GT_RBRACK] = ACTIONS(7095),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_RBRACK] = ACTIONS(7095),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_RBRACE] = ACTIONS(7095),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_to] = ACTIONS(7093),
    [anon_sym_downto] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_DOT_DOT2] = ACTIONS(7095),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(4411)] = {
    [sym_block_comment] = STATE(4411),
    [sym_compiler_directive_decl] = STATE(4411),
    [sym_fsi_directive_decl] = STATE(4411),
    [sym_preproc_line] = STATE(4411),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7500),
    [aux_sym_decimal_token1] = ACTIONS(7185),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4412)] = {
    [sym_block_comment] = STATE(4412),
    [sym_compiler_directive_decl] = STATE(4412),
    [sym_fsi_directive_decl] = STATE(4412),
    [sym_preproc_line] = STATE(4412),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_as] = ACTIONS(6637),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_DOT_DOT] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(8830),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4413)] = {
    [sym_block_comment] = STATE(4413),
    [sym_compiler_directive_decl] = STATE(4413),
    [sym_fsi_directive_decl] = STATE(4413),
    [sym_preproc_line] = STATE(4413),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_GT_RBRACK] = ACTIONS(7099),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_RBRACK] = ACTIONS(7099),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_RBRACE] = ACTIONS(7099),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_to] = ACTIONS(7097),
    [anon_sym_downto] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_DOT_DOT2] = ACTIONS(7099),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(4414)] = {
    [sym_block_comment] = STATE(4414),
    [sym_compiler_directive_decl] = STATE(4414),
    [sym_fsi_directive_decl] = STATE(4414),
    [sym_preproc_line] = STATE(4414),
    [aux_sym_compound_type_repeat1] = STATE(4414),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8832),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__quoted_close] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4415)] = {
    [sym_block_comment] = STATE(4415),
    [sym_compiler_directive_decl] = STATE(4415),
    [sym_fsi_directive_decl] = STATE(4415),
    [sym_preproc_line] = STATE(4415),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_as] = ACTIONS(6513),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4416)] = {
    [sym_block_comment] = STATE(4416),
    [sym_compiler_directive_decl] = STATE(4416),
    [sym_fsi_directive_decl] = STATE(4416),
    [sym_preproc_line] = STATE(4416),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_GT_RBRACK] = ACTIONS(7103),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_RBRACK] = ACTIONS(7103),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_RBRACE] = ACTIONS(7103),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_to] = ACTIONS(7101),
    [anon_sym_downto] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_DOT_DOT2] = ACTIONS(7103),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(4417)] = {
    [sym_block_comment] = STATE(4417),
    [sym_compiler_directive_decl] = STATE(4417),
    [sym_fsi_directive_decl] = STATE(4417),
    [sym_preproc_line] = STATE(4417),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_GT_RBRACK] = ACTIONS(7107),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_RBRACK] = ACTIONS(7107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_RBRACE] = ACTIONS(7107),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_to] = ACTIONS(7105),
    [anon_sym_downto] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_DOT_DOT2] = ACTIONS(7107),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(4418)] = {
    [sym_block_comment] = STATE(4418),
    [sym_compiler_directive_decl] = STATE(4418),
    [sym_fsi_directive_decl] = STATE(4418),
    [sym_preproc_line] = STATE(4418),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_GT_RBRACK] = ACTIONS(7111),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_RBRACK] = ACTIONS(7111),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_RBRACE] = ACTIONS(7111),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_to] = ACTIONS(7109),
    [anon_sym_downto] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_DOT_DOT2] = ACTIONS(7111),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(4419)] = {
    [sym_block_comment] = STATE(4419),
    [sym_compiler_directive_decl] = STATE(4419),
    [sym_fsi_directive_decl] = STATE(4419),
    [sym_preproc_line] = STATE(4419),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_GT_RBRACK] = ACTIONS(7115),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_RBRACK] = ACTIONS(7115),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_RBRACE] = ACTIONS(7115),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_to] = ACTIONS(7113),
    [anon_sym_downto] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_DOT_DOT2] = ACTIONS(7115),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(4420)] = {
    [sym_block_comment] = STATE(4420),
    [sym_compiler_directive_decl] = STATE(4420),
    [sym_fsi_directive_decl] = STATE(4420),
    [sym_preproc_line] = STATE(4420),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_as] = ACTIONS(6560),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_DOT_DOT] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4421)] = {
    [sym_elif_expression] = STATE(4801),
    [sym_block_comment] = STATE(4421),
    [sym_compiler_directive_decl] = STATE(4421),
    [sym_fsi_directive_decl] = STATE(4421),
    [sym_preproc_line] = STATE(4421),
    [aux_sym_if_expression_repeat1] = STATE(4421),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8835),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_DOT_DOT] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__dedent] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4422)] = {
    [sym_block_comment] = STATE(4422),
    [sym_compiler_directive_decl] = STATE(4422),
    [sym_fsi_directive_decl] = STATE(4422),
    [sym_preproc_line] = STATE(4422),
    [aux_sym_type_argument_repeat1] = STATE(4515),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8838),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4423)] = {
    [sym_block_comment] = STATE(4423),
    [sym_compiler_directive_decl] = STATE(4423),
    [sym_fsi_directive_decl] = STATE(4423),
    [sym_preproc_line] = STATE(4423),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_GT_RBRACK] = ACTIONS(7119),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_RBRACK] = ACTIONS(7119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_RBRACE] = ACTIONS(7119),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_to] = ACTIONS(7117),
    [anon_sym_downto] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_DOT_DOT2] = ACTIONS(7119),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(4424)] = {
    [sym_block_comment] = STATE(4424),
    [sym_compiler_directive_decl] = STATE(4424),
    [sym_fsi_directive_decl] = STATE(4424),
    [sym_preproc_line] = STATE(4424),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_GT_RBRACK] = ACTIONS(7123),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_RBRACK] = ACTIONS(7123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_RBRACE] = ACTIONS(7123),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_to] = ACTIONS(7121),
    [anon_sym_downto] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_DOT_DOT2] = ACTIONS(7123),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(4425)] = {
    [sym_block_comment] = STATE(4425),
    [sym_compiler_directive_decl] = STATE(4425),
    [sym_fsi_directive_decl] = STATE(4425),
    [sym_preproc_line] = STATE(4425),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_GT_RBRACK] = ACTIONS(6934),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_RBRACK] = ACTIONS(6934),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_RBRACE] = ACTIONS(6934),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_to] = ACTIONS(6936),
    [anon_sym_downto] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_DOT_DOT2] = ACTIONS(6934),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(4426)] = {
    [sym_block_comment] = STATE(4426),
    [sym_compiler_directive_decl] = STATE(4426),
    [sym_fsi_directive_decl] = STATE(4426),
    [sym_preproc_line] = STATE(4426),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_GT_RBRACK] = ACTIONS(6938),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_RBRACK] = ACTIONS(6938),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_RBRACE] = ACTIONS(6938),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_to] = ACTIONS(6940),
    [anon_sym_downto] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_DOT_DOT2] = ACTIONS(6938),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(4427)] = {
    [sym_block_comment] = STATE(4427),
    [sym_compiler_directive_decl] = STATE(4427),
    [sym_fsi_directive_decl] = STATE(4427),
    [sym_preproc_line] = STATE(4427),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_GT_RBRACK] = ACTIONS(7013),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_RBRACK] = ACTIONS(7013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_RBRACE] = ACTIONS(7013),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_to] = ACTIONS(7015),
    [anon_sym_downto] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_DOT_DOT2] = ACTIONS(7013),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(4428)] = {
    [sym_block_comment] = STATE(4428),
    [sym_compiler_directive_decl] = STATE(4428),
    [sym_fsi_directive_decl] = STATE(4428),
    [sym_preproc_line] = STATE(4428),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_GT_RBRACK] = ACTIONS(7029),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_RBRACK] = ACTIONS(7029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_RBRACE] = ACTIONS(7029),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_to] = ACTIONS(7031),
    [anon_sym_downto] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_DOT_DOT2] = ACTIONS(7029),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(4429)] = {
    [sym_block_comment] = STATE(4429),
    [sym_compiler_directive_decl] = STATE(4429),
    [sym_fsi_directive_decl] = STATE(4429),
    [sym_preproc_line] = STATE(4429),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_GT_RBRACK] = ACTIONS(7049),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_RBRACK] = ACTIONS(7049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_RBRACE] = ACTIONS(7049),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_to] = ACTIONS(7051),
    [anon_sym_downto] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_DOT_DOT2] = ACTIONS(7049),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(4430)] = {
    [sym_block_comment] = STATE(4430),
    [sym_compiler_directive_decl] = STATE(4430),
    [sym_fsi_directive_decl] = STATE(4430),
    [sym_preproc_line] = STATE(4430),
    [aux_sym_type_argument_repeat1] = STATE(4434),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__untyped_quoted_close] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4431)] = {
    [sym_block_comment] = STATE(4431),
    [sym_compiler_directive_decl] = STATE(4431),
    [sym_fsi_directive_decl] = STATE(4431),
    [sym_preproc_line] = STATE(4431),
    [aux_sym_compound_type_repeat1] = STATE(4436),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8494),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4432)] = {
    [sym_block_comment] = STATE(4432),
    [sym_compiler_directive_decl] = STATE(4432),
    [sym_fsi_directive_decl] = STATE(4432),
    [sym_preproc_line] = STATE(4432),
    [aux_sym_long_identifier_repeat1] = STATE(4433),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8841),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__quoted_close] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4433)] = {
    [sym_block_comment] = STATE(4433),
    [sym_compiler_directive_decl] = STATE(4433),
    [sym_fsi_directive_decl] = STATE(4433),
    [sym_preproc_line] = STATE(4433),
    [aux_sym_long_identifier_repeat1] = STATE(4433),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8843),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4434)] = {
    [sym_block_comment] = STATE(4434),
    [sym_compiler_directive_decl] = STATE(4434),
    [sym_fsi_directive_decl] = STATE(4434),
    [sym_preproc_line] = STATE(4434),
    [aux_sym_type_argument_repeat1] = STATE(4434),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8846),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__untyped_quoted_close] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4435)] = {
    [sym_block_comment] = STATE(4435),
    [sym_compiler_directive_decl] = STATE(4435),
    [sym_fsi_directive_decl] = STATE(4435),
    [sym_preproc_line] = STATE(4435),
    [aux_sym_long_identifier_repeat1] = STATE(4438),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8849),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__untyped_quoted_close] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4436)] = {
    [sym_block_comment] = STATE(4436),
    [sym_compiler_directive_decl] = STATE(4436),
    [sym_fsi_directive_decl] = STATE(4436),
    [sym_preproc_line] = STATE(4436),
    [aux_sym_compound_type_repeat1] = STATE(4436),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8851),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4437)] = {
    [sym_block_comment] = STATE(4437),
    [sym_compiler_directive_decl] = STATE(4437),
    [sym_fsi_directive_decl] = STATE(4437),
    [sym_preproc_line] = STATE(4437),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_GT_RBRACK] = ACTIONS(7061),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_RBRACK] = ACTIONS(7061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_RBRACE] = ACTIONS(7061),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_to] = ACTIONS(7063),
    [anon_sym_downto] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_DOT_DOT2] = ACTIONS(7061),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(4438)] = {
    [sym_block_comment] = STATE(4438),
    [sym_compiler_directive_decl] = STATE(4438),
    [sym_fsi_directive_decl] = STATE(4438),
    [sym_preproc_line] = STATE(4438),
    [aux_sym_long_identifier_repeat1] = STATE(4438),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8854),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__untyped_quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4439)] = {
    [sym_attributes] = STATE(13308),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6991),
    [sym_access_modifier] = STATE(22413),
    [sym__class_function_or_value_defn] = STATE(6996),
    [aux_sym__type_extension_inner] = STATE(4442),
    [sym__type_defn_elements] = STATE(6996),
    [sym_interface_implementation] = STATE(6999),
    [sym_member_defn] = STATE(6999),
    [sym_additional_constr_defn] = STATE(6985),
    [sym_block_comment] = STATE(4439),
    [sym_compiler_directive_decl] = STATE(4439),
    [sym_fsi_directive_decl] = STATE(4439),
    [sym_preproc_line] = STATE(4439),
    [sym_preproc_if_in_class_definition] = STATE(6999),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8620),
    [anon_sym_module] = ACTIONS(8620),
    [anon_sym_null] = ACTIONS(8620),
    [anon_sym_begin] = ACTIONS(8620),
    [anon_sym_open] = ACTIONS(8620),
    [anon_sym_type] = ACTIONS(8620),
    [anon_sym_LBRACK_LT] = ACTIONS(8622),
    [anon_sym_do] = ACTIONS(8620),
    [anon_sym_and] = ACTIONS(8620),
    [anon_sym_let] = ACTIONS(8620),
    [anon_sym_let_BANG] = ACTIONS(8622),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8620),
    [anon_sym_AMP] = ACTIONS(8620),
    [anon_sym_LBRACK] = ACTIONS(8620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8622),
    [anon_sym_LBRACE] = ACTIONS(8620),
    [anon_sym_LT_AT] = ACTIONS(8620),
    [anon_sym_LT_AT_AT] = ACTIONS(8622),
    [anon_sym_CARET] = ACTIONS(8622),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8622),
    [anon_sym_new] = ACTIONS(8620),
    [anon_sym_return] = ACTIONS(8620),
    [anon_sym_return_BANG] = ACTIONS(8622),
    [anon_sym_yield] = ACTIONS(8620),
    [anon_sym_yield_BANG] = ACTIONS(8622),
    [anon_sym_lazy] = ACTIONS(8620),
    [anon_sym_assert] = ACTIONS(8620),
    [anon_sym_upcast] = ACTIONS(8620),
    [anon_sym_downcast] = ACTIONS(8620),
    [anon_sym_fixed] = ACTIONS(8620),
    [anon_sym_for] = ACTIONS(8620),
    [anon_sym_while] = ACTIONS(8620),
    [anon_sym_while_BANG] = ACTIONS(8622),
    [anon_sym_if] = ACTIONS(8620),
    [anon_sym_fun] = ACTIONS(8620),
    [anon_sym_try] = ACTIONS(8620),
    [anon_sym_match] = ACTIONS(8620),
    [anon_sym_match_BANG] = ACTIONS(8622),
    [anon_sym_function] = ACTIONS(8620),
    [anon_sym_use] = ACTIONS(8620),
    [anon_sym_use_BANG] = ACTIONS(8622),
    [anon_sym_do_BANG] = ACTIONS(8622),
    [anon_sym_AMP_AMP] = ACTIONS(8622),
    [anon_sym_default] = ACTIONS(8418),
    [anon_sym_static] = ACTIONS(8420),
    [anon_sym_member] = ACTIONS(8422),
    [anon_sym_exception] = ACTIONS(8620),
    [anon_sym_interface] = ACTIONS(8424),
    [anon_sym_abstract] = ACTIONS(8426),
    [anon_sym_val] = ACTIONS(8428),
    [anon_sym_override] = ACTIONS(8418),
    [anon_sym_extern] = ACTIONS(8620),
    [aux_sym_char_token1] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8620),
    [anon_sym_DQUOTE] = ACTIONS(8620),
    [anon_sym_AT_DQUOTE] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [sym_bool] = ACTIONS(8620),
    [sym_unit] = ACTIONS(8622),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8620),
    [sym_op_identifier] = ACTIONS(8622),
    [anon_sym_PLUS] = ACTIONS(8620),
    [anon_sym_DASH] = ACTIONS(8620),
    [anon_sym_PLUS_DOT] = ACTIONS(8622),
    [anon_sym_DASH_DOT] = ACTIONS(8622),
    [anon_sym_PERCENT] = ACTIONS(8620),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8622),
    [anon_sym_TILDE] = ACTIONS(8622),
    [aux_sym_prefix_op_token1] = ACTIONS(8622),
    [sym_int] = ACTIONS(8620),
    [sym_xint] = ACTIONS(8622),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8622),
    [anon_sym_POUNDload] = ACTIONS(8622),
    [anon_sym_POUNDtime] = ACTIONS(8622),
    [anon_sym_POUNDI] = ACTIONS(8622),
    [anon_sym_POUNDhelp] = ACTIONS(8622),
    [anon_sym_POUNDquit] = ACTIONS(8622),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8622),
    [anon_sym_POUNDendif] = ACTIONS(8622),
    [sym__newline] = ACTIONS(8622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8622),
  },
  [STATE(4440)] = {
    [sym__else_expression] = STATE(6003),
    [sym_elif_expression] = STATE(4962),
    [sym_block_comment] = STATE(4440),
    [sym_compiler_directive_decl] = STATE(4440),
    [sym_fsi_directive_decl] = STATE(4440),
    [sym_preproc_line] = STATE(4440),
    [aux_sym_if_expression_repeat1] = STATE(4628),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8751),
    [anon_sym_elif] = ACTIONS(8753),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__quoted_close] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4441)] = {
    [sym_elif_expression] = STATE(4823),
    [sym_block_comment] = STATE(4441),
    [sym_compiler_directive_decl] = STATE(4441),
    [sym_fsi_directive_decl] = STATE(4441),
    [sym_preproc_line] = STATE(4441),
    [aux_sym_if_expression_repeat1] = STATE(4441),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8857),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__dedent] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
    [sym__in] = ACTIONS(6494),
  },
  [STATE(4442)] = {
    [sym_attributes] = STATE(13308),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(6991),
    [sym_access_modifier] = STATE(22413),
    [sym__class_function_or_value_defn] = STATE(6996),
    [aux_sym__type_extension_inner] = STATE(4442),
    [sym__type_defn_elements] = STATE(6996),
    [sym_interface_implementation] = STATE(6999),
    [sym_member_defn] = STATE(6999),
    [sym_additional_constr_defn] = STATE(6985),
    [sym_block_comment] = STATE(4442),
    [sym_compiler_directive_decl] = STATE(4442),
    [sym_fsi_directive_decl] = STATE(4442),
    [sym_preproc_line] = STATE(4442),
    [sym_preproc_if_in_class_definition] = STATE(6999),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8626),
    [anon_sym_module] = ACTIONS(8626),
    [anon_sym_null] = ACTIONS(8626),
    [anon_sym_begin] = ACTIONS(8626),
    [anon_sym_open] = ACTIONS(8626),
    [anon_sym_type] = ACTIONS(8626),
    [anon_sym_LBRACK_LT] = ACTIONS(8628),
    [anon_sym_do] = ACTIONS(8860),
    [anon_sym_and] = ACTIONS(8626),
    [anon_sym_let] = ACTIONS(8863),
    [anon_sym_let_BANG] = ACTIONS(8866),
    [aux_sym_access_modifier_token1] = ACTIONS(8640),
    [anon_sym_LPAREN] = ACTIONS(8626),
    [anon_sym_AMP] = ACTIONS(8626),
    [anon_sym_LBRACK] = ACTIONS(8626),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8643),
    [anon_sym_LBRACE] = ACTIONS(8626),
    [anon_sym_LT_AT] = ACTIONS(8626),
    [anon_sym_LT_AT_AT] = ACTIONS(8643),
    [anon_sym_CARET] = ACTIONS(8643),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8643),
    [anon_sym_new] = ACTIONS(8869),
    [anon_sym_return] = ACTIONS(8626),
    [anon_sym_return_BANG] = ACTIONS(8643),
    [anon_sym_yield] = ACTIONS(8626),
    [anon_sym_yield_BANG] = ACTIONS(8643),
    [anon_sym_lazy] = ACTIONS(8626),
    [anon_sym_assert] = ACTIONS(8626),
    [anon_sym_upcast] = ACTIONS(8626),
    [anon_sym_downcast] = ACTIONS(8626),
    [anon_sym_fixed] = ACTIONS(8626),
    [anon_sym_for] = ACTIONS(8626),
    [anon_sym_while] = ACTIONS(8626),
    [anon_sym_while_BANG] = ACTIONS(8643),
    [anon_sym_if] = ACTIONS(8626),
    [anon_sym_fun] = ACTIONS(8626),
    [anon_sym_try] = ACTIONS(8626),
    [anon_sym_match] = ACTIONS(8626),
    [anon_sym_match_BANG] = ACTIONS(8643),
    [anon_sym_function] = ACTIONS(8626),
    [anon_sym_use] = ACTIONS(8626),
    [anon_sym_use_BANG] = ACTIONS(8643),
    [anon_sym_do_BANG] = ACTIONS(8643),
    [anon_sym_AMP_AMP] = ACTIONS(8643),
    [anon_sym_default] = ACTIONS(8872),
    [anon_sym_static] = ACTIONS(8875),
    [anon_sym_member] = ACTIONS(8878),
    [anon_sym_exception] = ACTIONS(8626),
    [anon_sym_interface] = ACTIONS(8881),
    [anon_sym_abstract] = ACTIONS(8884),
    [anon_sym_val] = ACTIONS(8887),
    [anon_sym_override] = ACTIONS(8872),
    [anon_sym_extern] = ACTIONS(8626),
    [aux_sym_char_token1] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8626),
    [anon_sym_DQUOTE] = ACTIONS(8626),
    [anon_sym_AT_DQUOTE] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [sym_bool] = ACTIONS(8626),
    [sym_unit] = ACTIONS(8643),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8626),
    [sym_op_identifier] = ACTIONS(8643),
    [anon_sym_PLUS] = ACTIONS(8626),
    [anon_sym_DASH] = ACTIONS(8626),
    [anon_sym_PLUS_DOT] = ACTIONS(8643),
    [anon_sym_DASH_DOT] = ACTIONS(8643),
    [anon_sym_PERCENT] = ACTIONS(8626),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8643),
    [anon_sym_TILDE] = ACTIONS(8643),
    [aux_sym_prefix_op_token1] = ACTIONS(8643),
    [sym_int] = ACTIONS(8626),
    [sym_xint] = ACTIONS(8643),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8643),
    [anon_sym_POUNDload] = ACTIONS(8643),
    [anon_sym_POUNDtime] = ACTIONS(8643),
    [anon_sym_POUNDI] = ACTIONS(8643),
    [anon_sym_POUNDhelp] = ACTIONS(8643),
    [anon_sym_POUNDquit] = ACTIONS(8643),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8890),
    [anon_sym_POUNDendif] = ACTIONS(8643),
    [sym__newline] = ACTIONS(8643),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8643),
  },
  [STATE(4443)] = {
    [sym_block_comment] = STATE(4443),
    [sym_compiler_directive_decl] = STATE(4443),
    [sym_fsi_directive_decl] = STATE(4443),
    [sym_preproc_line] = STATE(4443),
    [aux_sym_compound_type_repeat1] = STATE(4414),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8360),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__quoted_close] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4444)] = {
    [sym_block_comment] = STATE(4444),
    [sym_compiler_directive_decl] = STATE(4444),
    [sym_fsi_directive_decl] = STATE(4444),
    [sym_preproc_line] = STATE(4444),
    [aux_sym_rules_repeat1] = STATE(4446),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_as] = ACTIONS(6618),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(8805),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_DOT_DOT] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(8893),
    [sym__dedent] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4445)] = {
    [sym_block_comment] = STATE(4445),
    [sym_compiler_directive_decl] = STATE(4445),
    [sym_fsi_directive_decl] = STATE(4445),
    [sym_preproc_line] = STATE(4445),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__dedent] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
    [sym__in] = ACTIONS(6490),
  },
  [STATE(4446)] = {
    [sym_block_comment] = STATE(4446),
    [sym_compiler_directive_decl] = STATE(4446),
    [sym_fsi_directive_decl] = STATE(4446),
    [sym_preproc_line] = STATE(4446),
    [aux_sym_rules_repeat1] = STATE(4446),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_as] = ACTIONS(6653),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(8896),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(8899),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4447)] = {
    [sym_block_comment] = STATE(4447),
    [sym_compiler_directive_decl] = STATE(4447),
    [sym_fsi_directive_decl] = STATE(4447),
    [sym_preproc_line] = STATE(4447),
    [aux_sym_long_identifier_repeat1] = STATE(4448),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_GT] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4448)] = {
    [sym_block_comment] = STATE(4448),
    [sym_compiler_directive_decl] = STATE(4448),
    [sym_fsi_directive_decl] = STATE(4448),
    [sym_preproc_line] = STATE(4448),
    [aux_sym_long_identifier_repeat1] = STATE(4448),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_GT] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8904),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4449)] = {
    [sym_block_comment] = STATE(4449),
    [sym_compiler_directive_decl] = STATE(4449),
    [sym_fsi_directive_decl] = STATE(4449),
    [sym_preproc_line] = STATE(4449),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__dedent] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
    [sym__in] = ACTIONS(6387),
  },
  [STATE(4450)] = {
    [sym_block_comment] = STATE(4450),
    [sym_compiler_directive_decl] = STATE(4450),
    [sym_fsi_directive_decl] = STATE(4450),
    [sym_preproc_line] = STATE(4450),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_GT_RBRACK] = ACTIONS(7161),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_RBRACK] = ACTIONS(7161),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_RBRACE] = ACTIONS(7161),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_to] = ACTIONS(7159),
    [anon_sym_downto] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_DOT_DOT2] = ACTIONS(7161),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(4451)] = {
    [sym_block_comment] = STATE(4451),
    [sym_compiler_directive_decl] = STATE(4451),
    [sym_fsi_directive_decl] = STATE(4451),
    [sym_preproc_line] = STATE(4451),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
    [sym__in] = ACTIONS(6297),
  },
  [STATE(4452)] = {
    [sym_block_comment] = STATE(4452),
    [sym_compiler_directive_decl] = STATE(4452),
    [sym_fsi_directive_decl] = STATE(4452),
    [sym_preproc_line] = STATE(4452),
    [aux_sym_rules_repeat1] = STATE(4366),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_as] = ACTIONS(6590),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(8805),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_DOT_DOT] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(8907),
    [sym__dedent] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(4453)] = {
    [sym_block_comment] = STATE(4453),
    [sym_compiler_directive_decl] = STATE(4453),
    [sym_fsi_directive_decl] = STATE(4453),
    [sym_preproc_line] = STATE(4453),
    [aux_sym_long_identifier_repeat1] = STATE(4454),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8759),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4454)] = {
    [sym_block_comment] = STATE(4454),
    [sym_compiler_directive_decl] = STATE(4454),
    [sym_fsi_directive_decl] = STATE(4454),
    [sym_preproc_line] = STATE(4454),
    [aux_sym_long_identifier_repeat1] = STATE(4454),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8910),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4455)] = {
    [sym__else_expression] = STATE(6295),
    [sym_elif_expression] = STATE(5077),
    [sym_block_comment] = STATE(4455),
    [sym_compiler_directive_decl] = STATE(4455),
    [sym_fsi_directive_decl] = STATE(4455),
    [sym_preproc_line] = STATE(4455),
    [aux_sym_if_expression_repeat1] = STATE(4685),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8755),
    [anon_sym_elif] = ACTIONS(8757),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__multi_dollar_interp_end] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4456)] = {
    [sym_block_comment] = STATE(4456),
    [sym_compiler_directive_decl] = STATE(4456),
    [sym_fsi_directive_decl] = STATE(4456),
    [sym_preproc_line] = STATE(4456),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_as] = ACTIONS(6369),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4457)] = {
    [sym_block_comment] = STATE(4457),
    [sym_compiler_directive_decl] = STATE(4457),
    [sym_fsi_directive_decl] = STATE(4457),
    [sym_preproc_line] = STATE(4457),
    [aux_sym_type_argument_repeat1] = STATE(4467),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_GT] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4458)] = {
    [sym_block_comment] = STATE(4458),
    [sym_compiler_directive_decl] = STATE(4458),
    [sym_fsi_directive_decl] = STATE(4458),
    [sym_preproc_line] = STATE(4458),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_as] = ACTIONS(6369),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4459)] = {
    [sym_block_comment] = STATE(4459),
    [sym_compiler_directive_decl] = STATE(4459),
    [sym_fsi_directive_decl] = STATE(4459),
    [sym_preproc_line] = STATE(4459),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_as] = ACTIONS(6373),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8913),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4460)] = {
    [sym_block_comment] = STATE(4460),
    [sym_compiler_directive_decl] = STATE(4460),
    [sym_fsi_directive_decl] = STATE(4460),
    [sym_preproc_line] = STATE(4460),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_GT_RBRACK] = ACTIONS(7065),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_RBRACK] = ACTIONS(7065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_RBRACE] = ACTIONS(7065),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_to] = ACTIONS(7067),
    [anon_sym_downto] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_DOT_DOT2] = ACTIONS(7065),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(4461)] = {
    [sym_block_comment] = STATE(4461),
    [sym_compiler_directive_decl] = STATE(4461),
    [sym_fsi_directive_decl] = STATE(4461),
    [sym_preproc_line] = STATE(4461),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_GT_RBRACK] = ACTIONS(7127),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_RBRACK] = ACTIONS(7127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_RBRACE] = ACTIONS(7127),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_to] = ACTIONS(7125),
    [anon_sym_downto] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_DOT_DOT2] = ACTIONS(7127),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(4462)] = {
    [sym_block_comment] = STATE(4462),
    [sym_compiler_directive_decl] = STATE(4462),
    [sym_fsi_directive_decl] = STATE(4462),
    [sym_preproc_line] = STATE(4462),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_as] = ACTIONS(6367),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4463)] = {
    [sym_block_comment] = STATE(4463),
    [sym_compiler_directive_decl] = STATE(4463),
    [sym_fsi_directive_decl] = STATE(4463),
    [sym_preproc_line] = STATE(4463),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_as] = ACTIONS(6379),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4464)] = {
    [sym_block_comment] = STATE(4464),
    [sym_compiler_directive_decl] = STATE(4464),
    [sym_fsi_directive_decl] = STATE(4464),
    [sym_preproc_line] = STATE(4464),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_as] = ACTIONS(6383),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4465)] = {
    [sym_block_comment] = STATE(4465),
    [sym_compiler_directive_decl] = STATE(4465),
    [sym_fsi_directive_decl] = STATE(4465),
    [sym_preproc_line] = STATE(4465),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_as] = ACTIONS(6391),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4466)] = {
    [sym_block_comment] = STATE(4466),
    [sym_compiler_directive_decl] = STATE(4466),
    [sym_fsi_directive_decl] = STATE(4466),
    [sym_preproc_line] = STATE(4466),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_GT_RBRACK] = ACTIONS(7131),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_RBRACK] = ACTIONS(7131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_RBRACE] = ACTIONS(7131),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_to] = ACTIONS(7129),
    [anon_sym_downto] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_DOT_DOT2] = ACTIONS(7131),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(4467)] = {
    [sym_block_comment] = STATE(4467),
    [sym_compiler_directive_decl] = STATE(4467),
    [sym_fsi_directive_decl] = STATE(4467),
    [sym_preproc_line] = STATE(4467),
    [aux_sym_type_argument_repeat1] = STATE(4467),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_GT] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8915),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4468)] = {
    [sym_block_comment] = STATE(4468),
    [sym_compiler_directive_decl] = STATE(4468),
    [sym_fsi_directive_decl] = STATE(4468),
    [sym_preproc_line] = STATE(4468),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_GT_RBRACK] = ACTIONS(7135),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_RBRACK] = ACTIONS(7135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_RBRACE] = ACTIONS(7135),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_to] = ACTIONS(7133),
    [anon_sym_downto] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_DOT_DOT2] = ACTIONS(7135),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(4469)] = {
    [sym_block_comment] = STATE(4469),
    [sym_compiler_directive_decl] = STATE(4469),
    [sym_fsi_directive_decl] = STATE(4469),
    [sym_preproc_line] = STATE(4469),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_as] = ACTIONS(6395),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4470)] = {
    [sym__else_expression] = STATE(6365),
    [sym_elif_expression] = STATE(5154),
    [sym_block_comment] = STATE(4470),
    [sym_compiler_directive_decl] = STATE(4470),
    [sym_fsi_directive_decl] = STATE(4470),
    [sym_preproc_line] = STATE(4470),
    [aux_sym_if_expression_repeat1] = STATE(4519),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8918),
    [anon_sym_elif] = ACTIONS(8920),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__untyped_quoted_close] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4471)] = {
    [sym_block_comment] = STATE(4471),
    [sym_compiler_directive_decl] = STATE(4471),
    [sym_fsi_directive_decl] = STATE(4471),
    [sym_preproc_line] = STATE(4471),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_as] = ACTIONS(6399),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4472)] = {
    [sym_block_comment] = STATE(4472),
    [sym_compiler_directive_decl] = STATE(4472),
    [sym_fsi_directive_decl] = STATE(4472),
    [sym_preproc_line] = STATE(4472),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_as] = ACTIONS(6403),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4473)] = {
    [sym_block_comment] = STATE(4473),
    [sym_compiler_directive_decl] = STATE(4473),
    [sym_fsi_directive_decl] = STATE(4473),
    [sym_preproc_line] = STATE(4473),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_as] = ACTIONS(6407),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4474)] = {
    [sym_block_comment] = STATE(4474),
    [sym_compiler_directive_decl] = STATE(4474),
    [sym_fsi_directive_decl] = STATE(4474),
    [sym_preproc_line] = STATE(4474),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_as] = ACTIONS(6411),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4475)] = {
    [sym_block_comment] = STATE(4475),
    [sym_compiler_directive_decl] = STATE(4475),
    [sym_fsi_directive_decl] = STATE(4475),
    [sym_preproc_line] = STATE(4475),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4476)] = {
    [sym_block_comment] = STATE(4476),
    [sym_compiler_directive_decl] = STATE(4476),
    [sym_fsi_directive_decl] = STATE(4476),
    [sym_preproc_line] = STATE(4476),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4477)] = {
    [sym_block_comment] = STATE(4477),
    [sym_compiler_directive_decl] = STATE(4477),
    [sym_fsi_directive_decl] = STATE(4477),
    [sym_preproc_line] = STATE(4477),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4478)] = {
    [sym_block_comment] = STATE(4478),
    [sym_compiler_directive_decl] = STATE(4478),
    [sym_fsi_directive_decl] = STATE(4478),
    [sym_preproc_line] = STATE(4478),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8922),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [anon_sym_POUNDelse] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4479)] = {
    [sym_block_comment] = STATE(4479),
    [sym_compiler_directive_decl] = STATE(4479),
    [sym_fsi_directive_decl] = STATE(4479),
    [sym_preproc_line] = STATE(4479),
    [aux_sym_long_identifier_repeat1] = STATE(4484),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8746),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4480)] = {
    [sym_block_comment] = STATE(4480),
    [sym_compiler_directive_decl] = STATE(4480),
    [sym_fsi_directive_decl] = STATE(4480),
    [sym_preproc_line] = STATE(4480),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [anon_sym_POUNDelse] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4481)] = {
    [sym_block_comment] = STATE(4481),
    [sym_compiler_directive_decl] = STATE(4481),
    [sym_fsi_directive_decl] = STATE(4481),
    [sym_preproc_line] = STATE(4481),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [anon_sym_POUNDelse] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4482)] = {
    [sym_block_comment] = STATE(4482),
    [sym_compiler_directive_decl] = STATE(4482),
    [sym_fsi_directive_decl] = STATE(4482),
    [sym_preproc_line] = STATE(4482),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [anon_sym_POUNDelse] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4483)] = {
    [sym_block_comment] = STATE(4483),
    [sym_compiler_directive_decl] = STATE(4483),
    [sym_fsi_directive_decl] = STATE(4483),
    [sym_preproc_line] = STATE(4483),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [anon_sym_POUNDelse] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4484)] = {
    [sym_block_comment] = STATE(4484),
    [sym_compiler_directive_decl] = STATE(4484),
    [sym_fsi_directive_decl] = STATE(4484),
    [sym_preproc_line] = STATE(4484),
    [aux_sym_long_identifier_repeat1] = STATE(4484),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8924),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4485)] = {
    [sym_block_comment] = STATE(4485),
    [sym_compiler_directive_decl] = STATE(4485),
    [sym_fsi_directive_decl] = STATE(4485),
    [sym_preproc_line] = STATE(4485),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [anon_sym_POUNDelse] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4486)] = {
    [sym_block_comment] = STATE(4486),
    [sym_compiler_directive_decl] = STATE(4486),
    [sym_fsi_directive_decl] = STATE(4486),
    [sym_preproc_line] = STATE(4486),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [anon_sym_POUNDelse] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4487)] = {
    [sym_block_comment] = STATE(4487),
    [sym_compiler_directive_decl] = STATE(4487),
    [sym_fsi_directive_decl] = STATE(4487),
    [sym_preproc_line] = STATE(4487),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [anon_sym_POUNDelse] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4488)] = {
    [sym_block_comment] = STATE(4488),
    [sym_compiler_directive_decl] = STATE(4488),
    [sym_fsi_directive_decl] = STATE(4488),
    [sym_preproc_line] = STATE(4488),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [anon_sym_POUNDelse] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4489)] = {
    [sym_block_comment] = STATE(4489),
    [sym_compiler_directive_decl] = STATE(4489),
    [sym_fsi_directive_decl] = STATE(4489),
    [sym_preproc_line] = STATE(4489),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [anon_sym_POUNDelse] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4490)] = {
    [sym_elif_expression] = STATE(4848),
    [sym_block_comment] = STATE(4490),
    [sym_compiler_directive_decl] = STATE(4490),
    [sym_fsi_directive_decl] = STATE(4490),
    [sym_preproc_line] = STATE(4490),
    [aux_sym_if_expression_repeat1] = STATE(4490),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(8927),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_DASH_GT] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_DOT_DOT] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4491)] = {
    [sym_block_comment] = STATE(4491),
    [sym_compiler_directive_decl] = STATE(4491),
    [sym_fsi_directive_decl] = STATE(4491),
    [sym_preproc_line] = STATE(4491),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4492)] = {
    [sym_block_comment] = STATE(4492),
    [sym_compiler_directive_decl] = STATE(4492),
    [sym_fsi_directive_decl] = STATE(4492),
    [sym_preproc_line] = STATE(4492),
    [aux_sym_type_argument_repeat1] = STATE(4337),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8932),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4493)] = {
    [sym_block_comment] = STATE(4493),
    [sym_compiler_directive_decl] = STATE(4493),
    [sym_fsi_directive_decl] = STATE(4493),
    [sym_preproc_line] = STATE(4493),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_as] = ACTIONS(6492),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4494)] = {
    [sym_block_comment] = STATE(4494),
    [sym_compiler_directive_decl] = STATE(4494),
    [sym_fsi_directive_decl] = STATE(4494),
    [sym_preproc_line] = STATE(4494),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_as] = ACTIONS(6427),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4495)] = {
    [sym_block_comment] = STATE(4495),
    [sym_compiler_directive_decl] = STATE(4495),
    [sym_fsi_directive_decl] = STATE(4495),
    [sym_preproc_line] = STATE(4495),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_as] = ACTIONS(6389),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4496)] = {
    [sym_block_comment] = STATE(4496),
    [sym_compiler_directive_decl] = STATE(4496),
    [sym_fsi_directive_decl] = STATE(4496),
    [sym_preproc_line] = STATE(4496),
    [aux_sym_long_identifier_repeat1] = STATE(4501),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8935),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_EQ2] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4497)] = {
    [sym_block_comment] = STATE(4497),
    [sym_compiler_directive_decl] = STATE(4497),
    [sym_fsi_directive_decl] = STATE(4497),
    [sym_preproc_line] = STATE(4497),
    [aux_sym_long_identifier_repeat1] = STATE(4432),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8841),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__quoted_close] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4498)] = {
    [sym_block_comment] = STATE(4498),
    [sym_compiler_directive_decl] = STATE(4498),
    [sym_fsi_directive_decl] = STATE(4498),
    [sym_preproc_line] = STATE(4498),
    [aux_sym_type_argument_repeat1] = STATE(4331),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8937),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__multi_dollar_interp_end] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4499)] = {
    [sym__else_expression] = STATE(6321),
    [sym_elif_expression] = STATE(5190),
    [sym_block_comment] = STATE(4499),
    [sym_compiler_directive_decl] = STATE(4499),
    [sym_fsi_directive_decl] = STATE(4499),
    [sym_preproc_line] = STATE(4499),
    [aux_sym_if_expression_repeat1] = STATE(4516),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_with] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8940),
    [anon_sym_elif] = ACTIONS(8942),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4500)] = {
    [sym_block_comment] = STATE(4500),
    [sym_compiler_directive_decl] = STATE(4500),
    [sym_fsi_directive_decl] = STATE(4500),
    [sym_preproc_line] = STATE(4500),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_GT_RBRACK] = ACTIONS(7139),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_RBRACK] = ACTIONS(7139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_RBRACE] = ACTIONS(7139),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_to] = ACTIONS(7137),
    [anon_sym_downto] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_DOT_DOT2] = ACTIONS(7139),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(4501)] = {
    [sym_block_comment] = STATE(4501),
    [sym_compiler_directive_decl] = STATE(4501),
    [sym_fsi_directive_decl] = STATE(4501),
    [sym_preproc_line] = STATE(4501),
    [aux_sym_long_identifier_repeat1] = STATE(4501),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_EQ2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4502)] = {
    [sym_block_comment] = STATE(4502),
    [sym_compiler_directive_decl] = STATE(4502),
    [sym_fsi_directive_decl] = STATE(4502),
    [sym_preproc_line] = STATE(4502),
    [aux_sym_compound_type_repeat1] = STATE(4507),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8385),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__multi_dollar_interp_end] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4503)] = {
    [sym_block_comment] = STATE(4503),
    [sym_compiler_directive_decl] = STATE(4503),
    [sym_fsi_directive_decl] = STATE(4503),
    [sym_preproc_line] = STATE(4503),
    [aux_sym_type_argument_repeat1] = STATE(4457),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_GT] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8947),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4504)] = {
    [sym_block_comment] = STATE(4504),
    [sym_compiler_directive_decl] = STATE(4504),
    [sym_fsi_directive_decl] = STATE(4504),
    [sym_preproc_line] = STATE(4504),
    [aux_sym_long_identifier_repeat1] = STATE(4508),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(8797),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_STAR] = ACTIONS(6360),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__multi_dollar_interp_end] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4505)] = {
    [sym__else_expression] = STATE(6275),
    [sym_elif_expression] = STATE(5164),
    [sym_block_comment] = STATE(4505),
    [sym_compiler_directive_decl] = STATE(4505),
    [sym_fsi_directive_decl] = STATE(4505),
    [sym_preproc_line] = STATE(4505),
    [aux_sym_if_expression_repeat1] = STATE(4513),
    [sym_identifier] = ACTIONS(6319),
    [anon_sym_null] = ACTIONS(6319),
    [anon_sym_EQ] = ACTIONS(6319),
    [anon_sym_begin] = ACTIONS(6319),
    [anon_sym_COLON] = ACTIONS(6319),
    [anon_sym_do] = ACTIONS(6319),
    [anon_sym_let] = ACTIONS(6319),
    [anon_sym_let_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN] = ACTIONS(6319),
    [anon_sym_COMMA] = ACTIONS(6317),
    [anon_sym_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_QMARK] = ACTIONS(6319),
    [anon_sym_COLON_COLON] = ACTIONS(6317),
    [anon_sym_AMP] = ACTIONS(6319),
    [anon_sym_LBRACK] = ACTIONS(6319),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6317),
    [anon_sym_LBRACE] = ACTIONS(6319),
    [anon_sym_LT_AT] = ACTIONS(6319),
    [anon_sym_LT_AT_AT] = ACTIONS(6319),
    [anon_sym_CARET] = ACTIONS(6319),
    [anon_sym_or] = ACTIONS(6319),
    [anon_sym_DOT] = ACTIONS(6319),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6317),
    [anon_sym_new] = ACTIONS(6319),
    [anon_sym_return] = ACTIONS(6319),
    [anon_sym_return_BANG] = ACTIONS(6317),
    [anon_sym_yield] = ACTIONS(6319),
    [anon_sym_yield_BANG] = ACTIONS(6317),
    [anon_sym_lazy] = ACTIONS(6319),
    [anon_sym_assert] = ACTIONS(6319),
    [anon_sym_upcast] = ACTIONS(6319),
    [anon_sym_downcast] = ACTIONS(6319),
    [anon_sym_fixed] = ACTIONS(6319),
    [anon_sym_COLON_GT] = ACTIONS(6317),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6317),
    [anon_sym_for] = ACTIONS(6319),
    [anon_sym_while] = ACTIONS(6319),
    [anon_sym_while_BANG] = ACTIONS(6317),
    [anon_sym_else] = ACTIONS(8950),
    [anon_sym_elif] = ACTIONS(8952),
    [anon_sym_if] = ACTIONS(6319),
    [anon_sym_fun] = ACTIONS(6319),
    [anon_sym_try] = ACTIONS(6319),
    [anon_sym_match] = ACTIONS(6319),
    [anon_sym_match_BANG] = ACTIONS(6317),
    [anon_sym_function] = ACTIONS(6319),
    [anon_sym_DOT_LBRACK] = ACTIONS(6317),
    [anon_sym_use] = ACTIONS(6319),
    [anon_sym_use_BANG] = ACTIONS(6317),
    [anon_sym_do_BANG] = ACTIONS(6317),
    [anon_sym_LPAREN2] = ACTIONS(6319),
    [anon_sym_PIPE_PIPE] = ACTIONS(6319),
    [anon_sym_AMP_AMP] = ACTIONS(6319),
    [aux_sym_char_token1] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6319),
    [anon_sym_DQUOTE] = ACTIONS(6319),
    [anon_sym_AT_DQUOTE] = ACTIONS(6317),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6317),
    [sym_bool] = ACTIONS(6319),
    [sym_unit] = ACTIONS(6317),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6319),
    [sym_op_identifier] = ACTIONS(6319),
    [anon_sym_PLUS] = ACTIONS(6319),
    [anon_sym_DASH] = ACTIONS(6319),
    [anon_sym_PLUS_DOT] = ACTIONS(6319),
    [anon_sym_DASH_DOT] = ACTIONS(6319),
    [anon_sym_PERCENT] = ACTIONS(6319),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6319),
    [anon_sym_TILDE] = ACTIONS(6317),
    [aux_sym_prefix_op_token1] = ACTIONS(6319),
    [aux_sym_infix_op_token1] = ACTIONS(6317),
    [aux_sym_infix_op_token2] = ACTIONS(6319),
    [aux_sym_infix_op_token3] = ACTIONS(6319),
    [anon_sym_BANG_EQ] = ACTIONS(6319),
    [anon_sym_COLON_EQ] = ACTIONS(6317),
    [anon_sym_DOLLAR] = ACTIONS(6319),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6319),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6319),
    [sym_int] = ACTIONS(6319),
    [sym_xint] = ACTIONS(6317),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6317),
    [anon_sym_POUNDendif] = ACTIONS(6317),
    [sym__newline] = ACTIONS(6317),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6317),
    [sym__tyapp_open] = ACTIONS(6317),
  },
  [STATE(4506)] = {
    [sym_block_comment] = STATE(4506),
    [sym_compiler_directive_decl] = STATE(4506),
    [sym_fsi_directive_decl] = STATE(4506),
    [sym_preproc_line] = STATE(4506),
    [aux_sym_compound_type_repeat1] = STATE(4509),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_EQ] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_COLON] = ACTIONS(6338),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_COMMA] = ACTIONS(6340),
    [anon_sym_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_QMARK] = ACTIONS(6338),
    [anon_sym_COLON_COLON] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_GT] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6338),
    [anon_sym_CARET] = ACTIONS(6338),
    [anon_sym_or] = ACTIONS(6338),
    [anon_sym_DOT] = ACTIONS(6338),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_COLON_GT] = ACTIONS(6340),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6340),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6338),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_DOT_LBRACK] = ACTIONS(6340),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_LPAREN2] = ACTIONS(6338),
    [anon_sym_PIPE_PIPE] = ACTIONS(6338),
    [anon_sym_AMP_AMP] = ACTIONS(6338),
    [anon_sym_STAR] = ACTIONS(8432),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6338),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6338),
    [anon_sym_DASH_DOT] = ACTIONS(6338),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6338),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6338),
    [aux_sym_infix_op_token1] = ACTIONS(6340),
    [aux_sym_infix_op_token2] = ACTIONS(6338),
    [aux_sym_infix_op_token3] = ACTIONS(6338),
    [anon_sym_BANG_EQ] = ACTIONS(6338),
    [anon_sym_COLON_EQ] = ACTIONS(6340),
    [anon_sym_DOLLAR] = ACTIONS(6338),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6338),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6338),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
    [sym__tyapp_open] = ACTIONS(6340),
  },
  [STATE(4507)] = {
    [sym_block_comment] = STATE(4507),
    [sym_compiler_directive_decl] = STATE(4507),
    [sym_fsi_directive_decl] = STATE(4507),
    [sym_preproc_line] = STATE(4507),
    [aux_sym_compound_type_repeat1] = STATE(4507),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8954),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__multi_dollar_interp_end] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4508)] = {
    [sym_block_comment] = STATE(4508),
    [sym_compiler_directive_decl] = STATE(4508),
    [sym_fsi_directive_decl] = STATE(4508),
    [sym_preproc_line] = STATE(4508),
    [aux_sym_long_identifier_repeat1] = STATE(4508),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(8957),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__multi_dollar_interp_end] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4509)] = {
    [sym_block_comment] = STATE(4509),
    [sym_compiler_directive_decl] = STATE(4509),
    [sym_fsi_directive_decl] = STATE(4509),
    [sym_preproc_line] = STATE(4509),
    [aux_sym_compound_type_repeat1] = STATE(4509),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_EQ] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_COLON] = ACTIONS(6193),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_COMMA] = ACTIONS(6195),
    [anon_sym_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_QMARK] = ACTIONS(6193),
    [anon_sym_COLON_COLON] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_GT] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6193),
    [anon_sym_CARET] = ACTIONS(6193),
    [anon_sym_or] = ACTIONS(6193),
    [anon_sym_DOT] = ACTIONS(6193),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_COLON_GT] = ACTIONS(6195),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6195),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6193),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_DOT_LBRACK] = ACTIONS(6195),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_LPAREN2] = ACTIONS(6193),
    [anon_sym_PIPE_PIPE] = ACTIONS(6193),
    [anon_sym_AMP_AMP] = ACTIONS(6193),
    [anon_sym_STAR] = ACTIONS(8960),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6193),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6193),
    [anon_sym_DASH_DOT] = ACTIONS(6193),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6193),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6193),
    [aux_sym_infix_op_token1] = ACTIONS(6195),
    [aux_sym_infix_op_token2] = ACTIONS(6193),
    [aux_sym_infix_op_token3] = ACTIONS(6193),
    [anon_sym_BANG_EQ] = ACTIONS(6193),
    [anon_sym_COLON_EQ] = ACTIONS(6195),
    [anon_sym_DOLLAR] = ACTIONS(6193),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6193),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6193),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
    [sym__tyapp_open] = ACTIONS(6195),
  },
  [STATE(4510)] = {
    [sym_block_comment] = STATE(4510),
    [sym_compiler_directive_decl] = STATE(4510),
    [sym_fsi_directive_decl] = STATE(4510),
    [sym_preproc_line] = STATE(4510),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [anon_sym_POUNDelse] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4511)] = {
    [sym_block_comment] = STATE(4511),
    [sym_compiler_directive_decl] = STATE(4511),
    [sym_fsi_directive_decl] = STATE(4511),
    [sym_preproc_line] = STATE(4511),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [anon_sym_POUNDelse] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4512)] = {
    [sym_block_comment] = STATE(4512),
    [sym_compiler_directive_decl] = STATE(4512),
    [sym_fsi_directive_decl] = STATE(4512),
    [sym_preproc_line] = STATE(4512),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [anon_sym_POUNDelse] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4513)] = {
    [sym__else_expression] = STATE(6294),
    [sym_elif_expression] = STATE(5164),
    [sym_block_comment] = STATE(4513),
    [sym_compiler_directive_decl] = STATE(4513),
    [sym_fsi_directive_decl] = STATE(4513),
    [sym_preproc_line] = STATE(4513),
    [aux_sym_if_expression_repeat1] = STATE(4677),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8950),
    [anon_sym_elif] = ACTIONS(8952),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [anon_sym_POUNDendif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4514)] = {
    [sym_block_comment] = STATE(4514),
    [sym_compiler_directive_decl] = STATE(4514),
    [sym_fsi_directive_decl] = STATE(4514),
    [sym_preproc_line] = STATE(4514),
    [aux_sym_long_identifier_repeat1] = STATE(4435),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8849),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__untyped_quoted_close] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4515)] = {
    [sym_block_comment] = STATE(4515),
    [sym_compiler_directive_decl] = STATE(4515),
    [sym_fsi_directive_decl] = STATE(4515),
    [sym_preproc_line] = STATE(4515),
    [aux_sym_type_argument_repeat1] = STATE(4518),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_EQ] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_COLON] = ACTIONS(6308),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_COMMA] = ACTIONS(6306),
    [anon_sym_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_QMARK] = ACTIONS(6308),
    [anon_sym_COLON_COLON] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6308),
    [anon_sym_CARET] = ACTIONS(6308),
    [anon_sym_or] = ACTIONS(6308),
    [anon_sym_DOT] = ACTIONS(6308),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6308),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_DOT_LBRACK] = ACTIONS(6306),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_LPAREN2] = ACTIONS(6308),
    [anon_sym_PIPE_PIPE] = ACTIONS(6308),
    [anon_sym_AMP_AMP] = ACTIONS(6308),
    [anon_sym_STAR] = ACTIONS(6308),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6308),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6308),
    [anon_sym_DASH_DOT] = ACTIONS(6308),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6308),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6308),
    [aux_sym_infix_op_token1] = ACTIONS(6306),
    [aux_sym_infix_op_token2] = ACTIONS(6308),
    [aux_sym_infix_op_token3] = ACTIONS(6308),
    [anon_sym_BANG_EQ] = ACTIONS(6308),
    [anon_sym_COLON_EQ] = ACTIONS(6306),
    [anon_sym_DOLLAR] = ACTIONS(6308),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6308),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6308),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
    [sym__tyapp_open] = ACTIONS(6306),
  },
  [STATE(4516)] = {
    [sym__else_expression] = STATE(6396),
    [sym_elif_expression] = STATE(5190),
    [sym_block_comment] = STATE(4516),
    [sym_compiler_directive_decl] = STATE(4516),
    [sym_fsi_directive_decl] = STATE(4516),
    [sym_preproc_line] = STATE(4516),
    [aux_sym_if_expression_repeat1] = STATE(4732),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_with] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8940),
    [anon_sym_elif] = ACTIONS(8942),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4517)] = {
    [sym_block_comment] = STATE(4517),
    [sym_compiler_directive_decl] = STATE(4517),
    [sym_fsi_directive_decl] = STATE(4517),
    [sym_preproc_line] = STATE(4517),
    [aux_sym_long_identifier_repeat1] = STATE(4447),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_EQ] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_COLON] = ACTIONS(6289),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_COMMA] = ACTIONS(6291),
    [anon_sym_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_QMARK] = ACTIONS(6289),
    [anon_sym_COLON_COLON] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_GT] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6289),
    [anon_sym_CARET] = ACTIONS(6289),
    [anon_sym_or] = ACTIONS(6289),
    [anon_sym_DOT] = ACTIONS(8902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_COLON_GT] = ACTIONS(6291),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6291),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6289),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_DOT_LBRACK] = ACTIONS(6291),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_LPAREN2] = ACTIONS(6289),
    [anon_sym_PIPE_PIPE] = ACTIONS(6289),
    [anon_sym_AMP_AMP] = ACTIONS(6289),
    [anon_sym_STAR] = ACTIONS(6289),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6289),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6289),
    [anon_sym_DASH_DOT] = ACTIONS(6289),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6289),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6289),
    [aux_sym_infix_op_token1] = ACTIONS(6291),
    [aux_sym_infix_op_token2] = ACTIONS(6289),
    [aux_sym_infix_op_token3] = ACTIONS(6289),
    [anon_sym_BANG_EQ] = ACTIONS(6289),
    [anon_sym_COLON_EQ] = ACTIONS(6291),
    [anon_sym_DOLLAR] = ACTIONS(6289),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6289),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6289),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
    [sym__tyapp_open] = ACTIONS(6291),
  },
  [STATE(4518)] = {
    [sym_block_comment] = STATE(4518),
    [sym_compiler_directive_decl] = STATE(4518),
    [sym_fsi_directive_decl] = STATE(4518),
    [sym_preproc_line] = STATE(4518),
    [aux_sym_type_argument_repeat1] = STATE(4518),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(8963),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4519)] = {
    [sym__else_expression] = STATE(6245),
    [sym_elif_expression] = STATE(5154),
    [sym_block_comment] = STATE(4519),
    [sym_compiler_directive_decl] = STATE(4519),
    [sym_fsi_directive_decl] = STATE(4519),
    [sym_preproc_line] = STATE(4519),
    [aux_sym_if_expression_repeat1] = STATE(4573),
    [sym_identifier] = ACTIONS(6329),
    [anon_sym_null] = ACTIONS(6329),
    [anon_sym_EQ] = ACTIONS(6329),
    [anon_sym_begin] = ACTIONS(6329),
    [anon_sym_COLON] = ACTIONS(6329),
    [anon_sym_do] = ACTIONS(6329),
    [anon_sym_let] = ACTIONS(6329),
    [anon_sym_let_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN] = ACTIONS(6329),
    [anon_sym_COMMA] = ACTIONS(6327),
    [anon_sym_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_QMARK] = ACTIONS(6329),
    [anon_sym_COLON_COLON] = ACTIONS(6327),
    [anon_sym_AMP] = ACTIONS(6329),
    [anon_sym_LBRACK] = ACTIONS(6329),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6327),
    [anon_sym_LBRACE] = ACTIONS(6329),
    [anon_sym_LT_AT] = ACTIONS(6329),
    [anon_sym_LT_AT_AT] = ACTIONS(6329),
    [anon_sym_CARET] = ACTIONS(6329),
    [anon_sym_or] = ACTIONS(6329),
    [anon_sym_DOT] = ACTIONS(6329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6327),
    [anon_sym_new] = ACTIONS(6329),
    [anon_sym_return] = ACTIONS(6329),
    [anon_sym_return_BANG] = ACTIONS(6327),
    [anon_sym_yield] = ACTIONS(6329),
    [anon_sym_yield_BANG] = ACTIONS(6327),
    [anon_sym_lazy] = ACTIONS(6329),
    [anon_sym_assert] = ACTIONS(6329),
    [anon_sym_upcast] = ACTIONS(6329),
    [anon_sym_downcast] = ACTIONS(6329),
    [anon_sym_fixed] = ACTIONS(6329),
    [anon_sym_COLON_GT] = ACTIONS(6327),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6327),
    [anon_sym_for] = ACTIONS(6329),
    [anon_sym_while] = ACTIONS(6329),
    [anon_sym_while_BANG] = ACTIONS(6327),
    [anon_sym_else] = ACTIONS(8918),
    [anon_sym_elif] = ACTIONS(8920),
    [anon_sym_if] = ACTIONS(6329),
    [anon_sym_fun] = ACTIONS(6329),
    [anon_sym_try] = ACTIONS(6329),
    [anon_sym_match] = ACTIONS(6329),
    [anon_sym_match_BANG] = ACTIONS(6327),
    [anon_sym_function] = ACTIONS(6329),
    [anon_sym_DOT_LBRACK] = ACTIONS(6327),
    [anon_sym_use] = ACTIONS(6329),
    [anon_sym_use_BANG] = ACTIONS(6327),
    [anon_sym_do_BANG] = ACTIONS(6327),
    [anon_sym_LPAREN2] = ACTIONS(6329),
    [anon_sym_PIPE_PIPE] = ACTIONS(6329),
    [anon_sym_AMP_AMP] = ACTIONS(6329),
    [aux_sym_char_token1] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6329),
    [anon_sym_DQUOTE] = ACTIONS(6329),
    [anon_sym_AT_DQUOTE] = ACTIONS(6327),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6327),
    [sym_bool] = ACTIONS(6329),
    [sym_unit] = ACTIONS(6327),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6329),
    [sym_op_identifier] = ACTIONS(6329),
    [anon_sym_PLUS] = ACTIONS(6329),
    [anon_sym_DASH] = ACTIONS(6329),
    [anon_sym_PLUS_DOT] = ACTIONS(6329),
    [anon_sym_DASH_DOT] = ACTIONS(6329),
    [anon_sym_PERCENT] = ACTIONS(6329),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6329),
    [anon_sym_TILDE] = ACTIONS(6327),
    [aux_sym_prefix_op_token1] = ACTIONS(6329),
    [aux_sym_infix_op_token1] = ACTIONS(6327),
    [aux_sym_infix_op_token2] = ACTIONS(6329),
    [aux_sym_infix_op_token3] = ACTIONS(6329),
    [anon_sym_BANG_EQ] = ACTIONS(6329),
    [anon_sym_COLON_EQ] = ACTIONS(6327),
    [anon_sym_DOLLAR] = ACTIONS(6329),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6329),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6329),
    [sym_int] = ACTIONS(6329),
    [sym_xint] = ACTIONS(6327),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6327),
    [sym__newline] = ACTIONS(6327),
    [sym__untyped_quoted_close] = ACTIONS(6327),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6327),
    [sym__tyapp_open] = ACTIONS(6327),
  },
  [STATE(4520)] = {
    [sym_block_comment] = STATE(4520),
    [sym_compiler_directive_decl] = STATE(4520),
    [sym_fsi_directive_decl] = STATE(4520),
    [sym_preproc_line] = STATE(4520),
    [aux_sym_long_identifier_repeat1] = STATE(4496),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(8966),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_EQ2] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4521)] = {
    [sym_block_comment] = STATE(4521),
    [sym_compiler_directive_decl] = STATE(4521),
    [sym_fsi_directive_decl] = STATE(4521),
    [sym_preproc_line] = STATE(4521),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_GT_RBRACK] = ACTIONS(6297),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4522)] = {
    [sym_block_comment] = STATE(4522),
    [sym_compiler_directive_decl] = STATE(4522),
    [sym_fsi_directive_decl] = STATE(4522),
    [sym_preproc_line] = STATE(4522),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_as] = ACTIONS(6612),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4523)] = {
    [sym_block_comment] = STATE(4523),
    [sym_compiler_directive_decl] = STATE(4523),
    [sym_fsi_directive_decl] = STATE(4523),
    [sym_preproc_line] = STATE(4523),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__dedent] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
    [sym__in] = ACTIONS(6393),
  },
  [STATE(4524)] = {
    [sym_block_comment] = STATE(4524),
    [sym_compiler_directive_decl] = STATE(4524),
    [sym_fsi_directive_decl] = STATE(4524),
    [sym_preproc_line] = STATE(4524),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_as] = ACTIONS(6331),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4525)] = {
    [sym_block_comment] = STATE(4525),
    [sym_compiler_directive_decl] = STATE(4525),
    [sym_fsi_directive_decl] = STATE(4525),
    [sym_preproc_line] = STATE(4525),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
    [sym__in] = ACTIONS(6371),
  },
  [STATE(4526)] = {
    [sym_block_comment] = STATE(4526),
    [sym_compiler_directive_decl] = STATE(4526),
    [sym_fsi_directive_decl] = STATE(4526),
    [sym_preproc_line] = STATE(4526),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__dedent] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
    [sym__in] = ACTIONS(6371),
  },
  [STATE(4527)] = {
    [sym_block_comment] = STATE(4527),
    [sym_compiler_directive_decl] = STATE(4527),
    [sym_fsi_directive_decl] = STATE(4527),
    [sym_preproc_line] = STATE(4527),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(8970),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__dedent] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
    [sym__in] = ACTIONS(6375),
  },
  [STATE(4528)] = {
    [sym_block_comment] = STATE(4528),
    [sym_compiler_directive_decl] = STATE(4528),
    [sym_fsi_directive_decl] = STATE(4528),
    [sym_preproc_line] = STATE(4528),
    [aux_sym_type_argument_repeat1] = STATE(4430),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(8972),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__untyped_quoted_close] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4529)] = {
    [sym_block_comment] = STATE(4529),
    [sym_compiler_directive_decl] = STATE(4529),
    [sym_fsi_directive_decl] = STATE(4529),
    [sym_preproc_line] = STATE(4529),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_as] = ACTIONS(6420),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4530)] = {
    [sym_block_comment] = STATE(4530),
    [sym_compiler_directive_decl] = STATE(4530),
    [sym_fsi_directive_decl] = STATE(4530),
    [sym_preproc_line] = STATE(4530),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__dedent] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
    [sym__in] = ACTIONS(6365),
  },
  [STATE(4531)] = {
    [sym_block_comment] = STATE(4531),
    [sym_compiler_directive_decl] = STATE(4531),
    [sym_fsi_directive_decl] = STATE(4531),
    [sym_preproc_line] = STATE(4531),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__dedent] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
    [sym__in] = ACTIONS(6381),
  },
  [STATE(4532)] = {
    [sym_block_comment] = STATE(4532),
    [sym_compiler_directive_decl] = STATE(4532),
    [sym_fsi_directive_decl] = STATE(4532),
    [sym_preproc_line] = STATE(4532),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__dedent] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
    [sym__in] = ACTIONS(6385),
  },
  [STATE(4533)] = {
    [sym_block_comment] = STATE(4533),
    [sym_compiler_directive_decl] = STATE(4533),
    [sym_fsi_directive_decl] = STATE(4533),
    [sym_preproc_line] = STATE(4533),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__dedent] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
    [sym__in] = ACTIONS(6333),
  },
  [STATE(4534)] = {
    [sym_block_comment] = STATE(4534),
    [sym_compiler_directive_decl] = STATE(4534),
    [sym_fsi_directive_decl] = STATE(4534),
    [sym_preproc_line] = STATE(4534),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_as] = ACTIONS(6312),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4535)] = {
    [sym_block_comment] = STATE(4535),
    [sym_compiler_directive_decl] = STATE(4535),
    [sym_fsi_directive_decl] = STATE(4535),
    [sym_preproc_line] = STATE(4535),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__dedent] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
    [sym__in] = ACTIONS(6418),
  },
  [STATE(4536)] = {
    [sym_block_comment] = STATE(4536),
    [sym_compiler_directive_decl] = STATE(4536),
    [sym_fsi_directive_decl] = STATE(4536),
    [sym_preproc_line] = STATE(4536),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__dedent] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
    [sym__in] = ACTIONS(6397),
  },
  [STATE(4537)] = {
    [sym_block_comment] = STATE(4537),
    [sym_compiler_directive_decl] = STATE(4537),
    [sym_fsi_directive_decl] = STATE(4537),
    [sym_preproc_line] = STATE(4537),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_GT_RBRACK] = ACTIONS(7143),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_RBRACK] = ACTIONS(7143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_RBRACE] = ACTIONS(7143),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_to] = ACTIONS(7141),
    [anon_sym_downto] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_DOT_DOT2] = ACTIONS(7143),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(4538)] = {
    [sym_block_comment] = STATE(4538),
    [sym_compiler_directive_decl] = STATE(4538),
    [sym_fsi_directive_decl] = STATE(4538),
    [sym_preproc_line] = STATE(4538),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__dedent] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
    [sym__in] = ACTIONS(6401),
  },
  [STATE(4539)] = {
    [sym_block_comment] = STATE(4539),
    [sym_compiler_directive_decl] = STATE(4539),
    [sym_fsi_directive_decl] = STATE(4539),
    [sym_preproc_line] = STATE(4539),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__dedent] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
    [sym__in] = ACTIONS(6405),
  },
  [STATE(4540)] = {
    [sym_block_comment] = STATE(4540),
    [sym_compiler_directive_decl] = STATE(4540),
    [sym_fsi_directive_decl] = STATE(4540),
    [sym_preproc_line] = STATE(4540),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__dedent] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
    [sym__in] = ACTIONS(6409),
  },
  [STATE(4541)] = {
    [sym_block_comment] = STATE(4541),
    [sym_compiler_directive_decl] = STATE(4541),
    [sym_fsi_directive_decl] = STATE(4541),
    [sym_preproc_line] = STATE(4541),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__dedent] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
    [sym__in] = ACTIONS(6413),
  },
  [STATE(4542)] = {
    [sym_attributes] = STATE(13414),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4543),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4542),
    [sym_compiler_directive_decl] = STATE(4542),
    [sym_fsi_directive_decl] = STATE(4542),
    [sym_preproc_line] = STATE(4542),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8622),
    [sym_identifier] = ACTIONS(8620),
    [anon_sym_namespace] = ACTIONS(8620),
    [anon_sym_module] = ACTIONS(8620),
    [anon_sym_null] = ACTIONS(8620),
    [anon_sym_begin] = ACTIONS(8620),
    [anon_sym_open] = ACTIONS(8620),
    [anon_sym_type] = ACTIONS(8620),
    [anon_sym_LBRACK_LT] = ACTIONS(8622),
    [anon_sym_do] = ACTIONS(8620),
    [anon_sym_and] = ACTIONS(8620),
    [anon_sym_let] = ACTIONS(8620),
    [anon_sym_let_BANG] = ACTIONS(8622),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8620),
    [anon_sym_AMP] = ACTIONS(8620),
    [anon_sym_LBRACK] = ACTIONS(8620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8622),
    [anon_sym_LBRACE] = ACTIONS(8620),
    [anon_sym_LT_AT] = ACTIONS(8620),
    [anon_sym_LT_AT_AT] = ACTIONS(8622),
    [anon_sym_CARET] = ACTIONS(8622),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8622),
    [anon_sym_new] = ACTIONS(8620),
    [anon_sym_return] = ACTIONS(8620),
    [anon_sym_return_BANG] = ACTIONS(8622),
    [anon_sym_yield] = ACTIONS(8620),
    [anon_sym_yield_BANG] = ACTIONS(8622),
    [anon_sym_lazy] = ACTIONS(8620),
    [anon_sym_assert] = ACTIONS(8620),
    [anon_sym_upcast] = ACTIONS(8620),
    [anon_sym_downcast] = ACTIONS(8620),
    [anon_sym_fixed] = ACTIONS(8620),
    [anon_sym_for] = ACTIONS(8620),
    [anon_sym_while] = ACTIONS(8620),
    [anon_sym_while_BANG] = ACTIONS(8622),
    [anon_sym_if] = ACTIONS(8620),
    [anon_sym_fun] = ACTIONS(8620),
    [anon_sym_try] = ACTIONS(8620),
    [anon_sym_match] = ACTIONS(8620),
    [anon_sym_match_BANG] = ACTIONS(8622),
    [anon_sym_function] = ACTIONS(8620),
    [anon_sym_use] = ACTIONS(8620),
    [anon_sym_use_BANG] = ACTIONS(8622),
    [anon_sym_do_BANG] = ACTIONS(8622),
    [anon_sym_AMP_AMP] = ACTIONS(8622),
    [anon_sym_default] = ACTIONS(8510),
    [anon_sym_static] = ACTIONS(8512),
    [anon_sym_member] = ACTIONS(8514),
    [anon_sym_exception] = ACTIONS(8620),
    [anon_sym_interface] = ACTIONS(8516),
    [anon_sym_abstract] = ACTIONS(8518),
    [anon_sym_val] = ACTIONS(8520),
    [anon_sym_override] = ACTIONS(8510),
    [anon_sym_extern] = ACTIONS(8620),
    [aux_sym_char_token1] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8620),
    [anon_sym_DQUOTE] = ACTIONS(8620),
    [anon_sym_AT_DQUOTE] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [sym_bool] = ACTIONS(8620),
    [sym_unit] = ACTIONS(8622),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8620),
    [sym_op_identifier] = ACTIONS(8622),
    [anon_sym_PLUS] = ACTIONS(8620),
    [anon_sym_DASH] = ACTIONS(8620),
    [anon_sym_PLUS_DOT] = ACTIONS(8622),
    [anon_sym_DASH_DOT] = ACTIONS(8622),
    [anon_sym_PERCENT] = ACTIONS(8620),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8622),
    [anon_sym_TILDE] = ACTIONS(8622),
    [aux_sym_prefix_op_token1] = ACTIONS(8622),
    [sym_int] = ACTIONS(8620),
    [sym_xint] = ACTIONS(8622),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8622),
    [anon_sym_POUNDload] = ACTIONS(8622),
    [anon_sym_POUNDtime] = ACTIONS(8622),
    [anon_sym_POUNDI] = ACTIONS(8622),
    [anon_sym_POUNDhelp] = ACTIONS(8622),
    [anon_sym_POUNDquit] = ACTIONS(8622),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8622),
  },
  [STATE(4543)] = {
    [sym_attributes] = STATE(13414),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4543),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4543),
    [sym_compiler_directive_decl] = STATE(4543),
    [sym_fsi_directive_decl] = STATE(4543),
    [sym_preproc_line] = STATE(4543),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8643),
    [sym_identifier] = ACTIONS(8626),
    [anon_sym_namespace] = ACTIONS(8626),
    [anon_sym_module] = ACTIONS(8626),
    [anon_sym_null] = ACTIONS(8626),
    [anon_sym_begin] = ACTIONS(8626),
    [anon_sym_open] = ACTIONS(8626),
    [anon_sym_type] = ACTIONS(8626),
    [anon_sym_LBRACK_LT] = ACTIONS(8628),
    [anon_sym_do] = ACTIONS(8975),
    [anon_sym_and] = ACTIONS(8626),
    [anon_sym_let] = ACTIONS(8978),
    [anon_sym_let_BANG] = ACTIONS(8981),
    [aux_sym_access_modifier_token1] = ACTIONS(8640),
    [anon_sym_LPAREN] = ACTIONS(8626),
    [anon_sym_AMP] = ACTIONS(8626),
    [anon_sym_LBRACK] = ACTIONS(8626),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8643),
    [anon_sym_LBRACE] = ACTIONS(8626),
    [anon_sym_LT_AT] = ACTIONS(8626),
    [anon_sym_LT_AT_AT] = ACTIONS(8643),
    [anon_sym_CARET] = ACTIONS(8643),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8643),
    [anon_sym_new] = ACTIONS(8984),
    [anon_sym_return] = ACTIONS(8626),
    [anon_sym_return_BANG] = ACTIONS(8643),
    [anon_sym_yield] = ACTIONS(8626),
    [anon_sym_yield_BANG] = ACTIONS(8643),
    [anon_sym_lazy] = ACTIONS(8626),
    [anon_sym_assert] = ACTIONS(8626),
    [anon_sym_upcast] = ACTIONS(8626),
    [anon_sym_downcast] = ACTIONS(8626),
    [anon_sym_fixed] = ACTIONS(8626),
    [anon_sym_for] = ACTIONS(8626),
    [anon_sym_while] = ACTIONS(8626),
    [anon_sym_while_BANG] = ACTIONS(8643),
    [anon_sym_if] = ACTIONS(8626),
    [anon_sym_fun] = ACTIONS(8626),
    [anon_sym_try] = ACTIONS(8626),
    [anon_sym_match] = ACTIONS(8626),
    [anon_sym_match_BANG] = ACTIONS(8643),
    [anon_sym_function] = ACTIONS(8626),
    [anon_sym_use] = ACTIONS(8626),
    [anon_sym_use_BANG] = ACTIONS(8643),
    [anon_sym_do_BANG] = ACTIONS(8643),
    [anon_sym_AMP_AMP] = ACTIONS(8643),
    [anon_sym_default] = ACTIONS(8987),
    [anon_sym_static] = ACTIONS(8990),
    [anon_sym_member] = ACTIONS(8993),
    [anon_sym_exception] = ACTIONS(8626),
    [anon_sym_interface] = ACTIONS(8996),
    [anon_sym_abstract] = ACTIONS(8999),
    [anon_sym_val] = ACTIONS(9002),
    [anon_sym_override] = ACTIONS(8987),
    [anon_sym_extern] = ACTIONS(8626),
    [aux_sym_char_token1] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8626),
    [anon_sym_DQUOTE] = ACTIONS(8626),
    [anon_sym_AT_DQUOTE] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [sym_bool] = ACTIONS(8626),
    [sym_unit] = ACTIONS(8643),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8626),
    [sym_op_identifier] = ACTIONS(8643),
    [anon_sym_PLUS] = ACTIONS(8626),
    [anon_sym_DASH] = ACTIONS(8626),
    [anon_sym_PLUS_DOT] = ACTIONS(8643),
    [anon_sym_DASH_DOT] = ACTIONS(8643),
    [anon_sym_PERCENT] = ACTIONS(8626),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8643),
    [anon_sym_TILDE] = ACTIONS(8643),
    [aux_sym_prefix_op_token1] = ACTIONS(8643),
    [sym_int] = ACTIONS(8626),
    [sym_xint] = ACTIONS(8643),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8643),
    [anon_sym_POUNDload] = ACTIONS(8643),
    [anon_sym_POUNDtime] = ACTIONS(8643),
    [anon_sym_POUNDI] = ACTIONS(8643),
    [anon_sym_POUNDhelp] = ACTIONS(8643),
    [anon_sym_POUNDquit] = ACTIONS(8643),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8643),
  },
  [STATE(4544)] = {
    [sym_block_comment] = STATE(4544),
    [sym_compiler_directive_decl] = STATE(4544),
    [sym_fsi_directive_decl] = STATE(4544),
    [sym_preproc_line] = STATE(4544),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_as] = ACTIONS(6564),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_DOT_DOT] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__dedent] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4545)] = {
    [sym_block_comment] = STATE(4545),
    [sym_compiler_directive_decl] = STATE(4545),
    [sym_fsi_directive_decl] = STATE(4545),
    [sym_preproc_line] = STATE(4545),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4546)] = {
    [sym_block_comment] = STATE(4546),
    [sym_compiler_directive_decl] = STATE(4546),
    [sym_fsi_directive_decl] = STATE(4546),
    [sym_preproc_line] = STATE(4546),
    [aux_sym_type_argument_repeat1] = STATE(4324),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(9008),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4547)] = {
    [sym_block_comment] = STATE(4547),
    [sym_compiler_directive_decl] = STATE(4547),
    [sym_fsi_directive_decl] = STATE(4547),
    [sym_preproc_line] = STATE(4547),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4548)] = {
    [sym_block_comment] = STATE(4548),
    [sym_compiler_directive_decl] = STATE(4548),
    [sym_fsi_directive_decl] = STATE(4548),
    [sym_preproc_line] = STATE(4548),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_as] = ACTIONS(6554),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_DOT_DOT] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__dedent] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4549)] = {
    [sym_block_comment] = STATE(4549),
    [sym_compiler_directive_decl] = STATE(4549),
    [sym_fsi_directive_decl] = STATE(4549),
    [sym_preproc_line] = STATE(4549),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_GT_RBRACK] = ACTIONS(7045),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_RBRACK] = ACTIONS(7045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_RBRACE] = ACTIONS(7045),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_to] = ACTIONS(7047),
    [anon_sym_downto] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_DOT_DOT2] = ACTIONS(7045),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(4550)] = {
    [sym_block_comment] = STATE(4550),
    [sym_compiler_directive_decl] = STATE(4550),
    [sym_fsi_directive_decl] = STATE(4550),
    [sym_preproc_line] = STATE(4550),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__multi_dollar_interp_end] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4551)] = {
    [sym_block_comment] = STATE(4551),
    [sym_compiler_directive_decl] = STATE(4551),
    [sym_fsi_directive_decl] = STATE(4551),
    [sym_preproc_line] = STATE(4551),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_DOT_DOT] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4552)] = {
    [sym_block_comment] = STATE(4552),
    [sym_compiler_directive_decl] = STATE(4552),
    [sym_fsi_directive_decl] = STATE(4552),
    [sym_preproc_line] = STATE(4552),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__quoted_close] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4553)] = {
    [sym_block_comment] = STATE(4553),
    [sym_compiler_directive_decl] = STATE(4553),
    [sym_fsi_directive_decl] = STATE(4553),
    [sym_preproc_line] = STATE(4553),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__quoted_close] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4554)] = {
    [sym_block_comment] = STATE(4554),
    [sym_compiler_directive_decl] = STATE(4554),
    [sym_fsi_directive_decl] = STATE(4554),
    [sym_preproc_line] = STATE(4554),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_with] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4555)] = {
    [sym_block_comment] = STATE(4555),
    [sym_compiler_directive_decl] = STATE(4555),
    [sym_fsi_directive_decl] = STATE(4555),
    [sym_preproc_line] = STATE(4555),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4556)] = {
    [sym_block_comment] = STATE(4556),
    [sym_compiler_directive_decl] = STATE(4556),
    [sym_fsi_directive_decl] = STATE(4556),
    [sym_preproc_line] = STATE(4556),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_DOT_DOT] = ACTIONS(6554),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [anon_sym_DOT_DOT2] = ACTIONS(6556),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__dedent] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4557)] = {
    [sym_block_comment] = STATE(4557),
    [sym_compiler_directive_decl] = STATE(4557),
    [sym_fsi_directive_decl] = STATE(4557),
    [sym_preproc_line] = STATE(4557),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_as] = ACTIONS(6840),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_DOT_DOT] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__dedent] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(4558)] = {
    [sym_block_comment] = STATE(4558),
    [sym_compiler_directive_decl] = STATE(4558),
    [sym_fsi_directive_decl] = STATE(4558),
    [sym_preproc_line] = STATE(4558),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_as] = ACTIONS(6653),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4559)] = {
    [sym_block_comment] = STATE(4559),
    [sym_compiler_directive_decl] = STATE(4559),
    [sym_fsi_directive_decl] = STATE(4559),
    [sym_preproc_line] = STATE(4559),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_DASH_GT] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6612),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4560)] = {
    [sym_block_comment] = STATE(4560),
    [sym_compiler_directive_decl] = STATE(4560),
    [sym_fsi_directive_decl] = STATE(4560),
    [sym_preproc_line] = STATE(4560),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__multi_dollar_interp_end] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4561)] = {
    [sym_block_comment] = STATE(4561),
    [sym_compiler_directive_decl] = STATE(4561),
    [sym_fsi_directive_decl] = STATE(4561),
    [sym_preproc_line] = STATE(4561),
    [aux_sym_long_identifier_repeat1] = STATE(4639),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6625),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_DOT_DOT2] = ACTIONS(6628),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4562)] = {
    [sym_block_comment] = STATE(4562),
    [sym_compiler_directive_decl] = STATE(4562),
    [sym_fsi_directive_decl] = STATE(4562),
    [sym_preproc_line] = STATE(4562),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_as] = ACTIONS(6513),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9015),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4563)] = {
    [sym_block_comment] = STATE(4563),
    [sym_compiler_directive_decl] = STATE(4563),
    [sym_fsi_directive_decl] = STATE(4563),
    [sym_preproc_line] = STATE(4563),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_as] = ACTIONS(6768),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9017),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_DOT_DOT] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__dedent] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(4564)] = {
    [sym_block_comment] = STATE(4564),
    [sym_compiler_directive_decl] = STATE(4564),
    [sym_fsi_directive_decl] = STATE(4564),
    [sym_preproc_line] = STATE(4564),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__multi_dollar_interp_end] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4565)] = {
    [sym_block_comment] = STATE(4565),
    [sym_compiler_directive_decl] = STATE(4565),
    [sym_fsi_directive_decl] = STATE(4565),
    [sym_preproc_line] = STATE(4565),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_GT] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4566)] = {
    [sym_block_comment] = STATE(4566),
    [sym_compiler_directive_decl] = STATE(4566),
    [sym_fsi_directive_decl] = STATE(4566),
    [sym_preproc_line] = STATE(4566),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_DOT_DOT] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4567)] = {
    [sym_block_comment] = STATE(4567),
    [sym_compiler_directive_decl] = STATE(4567),
    [sym_fsi_directive_decl] = STATE(4567),
    [sym_preproc_line] = STATE(4567),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__untyped_quoted_close] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4568)] = {
    [sym_block_comment] = STATE(4568),
    [sym_compiler_directive_decl] = STATE(4568),
    [sym_fsi_directive_decl] = STATE(4568),
    [sym_preproc_line] = STATE(4568),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__quoted_close] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4569)] = {
    [sym_block_comment] = STATE(4569),
    [sym_compiler_directive_decl] = STATE(4569),
    [sym_fsi_directive_decl] = STATE(4569),
    [sym_preproc_line] = STATE(4569),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_DOT_DOT] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4570)] = {
    [sym_block_comment] = STATE(4570),
    [sym_compiler_directive_decl] = STATE(4570),
    [sym_fsi_directive_decl] = STATE(4570),
    [sym_preproc_line] = STATE(4570),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9019),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [anon_sym_POUNDendif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4571)] = {
    [sym_block_comment] = STATE(4571),
    [sym_compiler_directive_decl] = STATE(4571),
    [sym_fsi_directive_decl] = STATE(4571),
    [sym_preproc_line] = STATE(4571),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4572)] = {
    [sym_block_comment] = STATE(4572),
    [sym_compiler_directive_decl] = STATE(4572),
    [sym_fsi_directive_decl] = STATE(4572),
    [sym_preproc_line] = STATE(4572),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_as] = ACTIONS(6813),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_DOT_DOT] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__dedent] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(4573)] = {
    [sym_elif_expression] = STATE(5154),
    [sym_block_comment] = STATE(4573),
    [sym_compiler_directive_decl] = STATE(4573),
    [sym_fsi_directive_decl] = STATE(4573),
    [sym_preproc_line] = STATE(4573),
    [aux_sym_if_expression_repeat1] = STATE(4573),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9021),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__untyped_quoted_close] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4574)] = {
    [sym_block_comment] = STATE(4574),
    [sym_compiler_directive_decl] = STATE(4574),
    [sym_fsi_directive_decl] = STATE(4574),
    [sym_preproc_line] = STATE(4574),
    [aux_sym_rules_repeat1] = STATE(4604),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9024),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_DOT_DOT] = ACTIONS(6590),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [anon_sym_DOT_DOT2] = ACTIONS(6588),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9026),
    [sym__dedent] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(4575)] = {
    [sym_block_comment] = STATE(4575),
    [sym_compiler_directive_decl] = STATE(4575),
    [sym_fsi_directive_decl] = STATE(4575),
    [sym_preproc_line] = STATE(4575),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4576)] = {
    [sym_block_comment] = STATE(4576),
    [sym_compiler_directive_decl] = STATE(4576),
    [sym_fsi_directive_decl] = STATE(4576),
    [sym_preproc_line] = STATE(4576),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_as] = ACTIONS(6834),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4577)] = {
    [sym_block_comment] = STATE(4577),
    [sym_compiler_directive_decl] = STATE(4577),
    [sym_fsi_directive_decl] = STATE(4577),
    [sym_preproc_line] = STATE(4577),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_GT] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4578)] = {
    [sym_block_comment] = STATE(4578),
    [sym_compiler_directive_decl] = STATE(4578),
    [sym_fsi_directive_decl] = STATE(4578),
    [sym_preproc_line] = STATE(4578),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_with] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4579)] = {
    [sym_block_comment] = STATE(4579),
    [sym_compiler_directive_decl] = STATE(4579),
    [sym_fsi_directive_decl] = STATE(4579),
    [sym_preproc_line] = STATE(4579),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_GT] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4580)] = {
    [sym_block_comment] = STATE(4580),
    [sym_compiler_directive_decl] = STATE(4580),
    [sym_fsi_directive_decl] = STATE(4580),
    [sym_preproc_line] = STATE(4580),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__multi_dollar_interp_end] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4581)] = {
    [sym_block_comment] = STATE(4581),
    [sym_compiler_directive_decl] = STATE(4581),
    [sym_fsi_directive_decl] = STATE(4581),
    [sym_preproc_line] = STATE(4581),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [anon_sym_POUNDendif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4582)] = {
    [sym_block_comment] = STATE(4582),
    [sym_compiler_directive_decl] = STATE(4582),
    [sym_fsi_directive_decl] = STATE(4582),
    [sym_preproc_line] = STATE(4582),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__untyped_quoted_close] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4583)] = {
    [sym_block_comment] = STATE(4583),
    [sym_compiler_directive_decl] = STATE(4583),
    [sym_fsi_directive_decl] = STATE(4583),
    [sym_preproc_line] = STATE(4583),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__quoted_close] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4584)] = {
    [sym_block_comment] = STATE(4584),
    [sym_compiler_directive_decl] = STATE(4584),
    [sym_fsi_directive_decl] = STATE(4584),
    [sym_preproc_line] = STATE(4584),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__untyped_quoted_close] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4585)] = {
    [sym_block_comment] = STATE(4585),
    [sym_compiler_directive_decl] = STATE(4585),
    [sym_fsi_directive_decl] = STATE(4585),
    [sym_preproc_line] = STATE(4585),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [anon_sym_POUNDendif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4586)] = {
    [sym_block_comment] = STATE(4586),
    [sym_compiler_directive_decl] = STATE(4586),
    [sym_fsi_directive_decl] = STATE(4586),
    [sym_preproc_line] = STATE(4586),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4587)] = {
    [sym_block_comment] = STATE(4587),
    [sym_compiler_directive_decl] = STATE(4587),
    [sym_fsi_directive_decl] = STATE(4587),
    [sym_preproc_line] = STATE(4587),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9031),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__quoted_close] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4588)] = {
    [sym_block_comment] = STATE(4588),
    [sym_compiler_directive_decl] = STATE(4588),
    [sym_fsi_directive_decl] = STATE(4588),
    [sym_preproc_line] = STATE(4588),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_DOT_DOT] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4589)] = {
    [sym_block_comment] = STATE(4589),
    [sym_compiler_directive_decl] = STATE(4589),
    [sym_fsi_directive_decl] = STATE(4589),
    [sym_preproc_line] = STATE(4589),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [anon_sym_POUNDendif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4590)] = {
    [sym_block_comment] = STATE(4590),
    [sym_compiler_directive_decl] = STATE(4590),
    [sym_fsi_directive_decl] = STATE(4590),
    [sym_preproc_line] = STATE(4590),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_DOT_DOT] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4591)] = {
    [sym_attributes] = STATE(13407),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4591),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4591),
    [sym_compiler_directive_decl] = STATE(4591),
    [sym_fsi_directive_decl] = STATE(4591),
    [sym_preproc_line] = STATE(4591),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8643),
    [sym_identifier] = ACTIONS(8626),
    [anon_sym_module] = ACTIONS(8626),
    [anon_sym_null] = ACTIONS(8626),
    [anon_sym_begin] = ACTIONS(8626),
    [anon_sym_open] = ACTIONS(8626),
    [anon_sym_type] = ACTIONS(8626),
    [anon_sym_LBRACK_LT] = ACTIONS(8628),
    [anon_sym_do] = ACTIONS(8975),
    [anon_sym_and] = ACTIONS(8626),
    [anon_sym_let] = ACTIONS(8978),
    [anon_sym_let_BANG] = ACTIONS(8981),
    [aux_sym_access_modifier_token1] = ACTIONS(8640),
    [anon_sym_LPAREN] = ACTIONS(8626),
    [anon_sym_AMP] = ACTIONS(8626),
    [anon_sym_LBRACK] = ACTIONS(8626),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8643),
    [anon_sym_LBRACE] = ACTIONS(8626),
    [anon_sym_LT_AT] = ACTIONS(8626),
    [anon_sym_LT_AT_AT] = ACTIONS(8643),
    [anon_sym_CARET] = ACTIONS(8643),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8643),
    [anon_sym_new] = ACTIONS(8984),
    [anon_sym_return] = ACTIONS(8626),
    [anon_sym_return_BANG] = ACTIONS(8643),
    [anon_sym_yield] = ACTIONS(8626),
    [anon_sym_yield_BANG] = ACTIONS(8643),
    [anon_sym_lazy] = ACTIONS(8626),
    [anon_sym_assert] = ACTIONS(8626),
    [anon_sym_upcast] = ACTIONS(8626),
    [anon_sym_downcast] = ACTIONS(8626),
    [anon_sym_fixed] = ACTIONS(8626),
    [anon_sym_for] = ACTIONS(8626),
    [anon_sym_while] = ACTIONS(8626),
    [anon_sym_while_BANG] = ACTIONS(8643),
    [anon_sym_if] = ACTIONS(8626),
    [anon_sym_fun] = ACTIONS(8626),
    [anon_sym_try] = ACTIONS(8626),
    [anon_sym_match] = ACTIONS(8626),
    [anon_sym_match_BANG] = ACTIONS(8643),
    [anon_sym_function] = ACTIONS(8626),
    [anon_sym_use] = ACTIONS(8626),
    [anon_sym_use_BANG] = ACTIONS(8643),
    [anon_sym_do_BANG] = ACTIONS(8643),
    [anon_sym_AMP_AMP] = ACTIONS(8643),
    [anon_sym_default] = ACTIONS(8987),
    [anon_sym_static] = ACTIONS(8990),
    [anon_sym_member] = ACTIONS(9033),
    [anon_sym_exception] = ACTIONS(8626),
    [anon_sym_interface] = ACTIONS(8996),
    [anon_sym_abstract] = ACTIONS(8999),
    [anon_sym_val] = ACTIONS(9002),
    [anon_sym_override] = ACTIONS(8987),
    [anon_sym_extern] = ACTIONS(8626),
    [aux_sym_char_token1] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8626),
    [anon_sym_DQUOTE] = ACTIONS(8626),
    [anon_sym_AT_DQUOTE] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [sym_bool] = ACTIONS(8626),
    [sym_unit] = ACTIONS(8643),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8626),
    [sym_op_identifier] = ACTIONS(8643),
    [anon_sym_PLUS] = ACTIONS(8626),
    [anon_sym_DASH] = ACTIONS(8626),
    [anon_sym_PLUS_DOT] = ACTIONS(8643),
    [anon_sym_DASH_DOT] = ACTIONS(8643),
    [anon_sym_PERCENT] = ACTIONS(8626),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8643),
    [anon_sym_TILDE] = ACTIONS(8643),
    [aux_sym_prefix_op_token1] = ACTIONS(8643),
    [sym_int] = ACTIONS(8626),
    [sym_xint] = ACTIONS(8643),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8643),
    [anon_sym_POUNDload] = ACTIONS(8643),
    [anon_sym_POUNDtime] = ACTIONS(8643),
    [anon_sym_POUNDI] = ACTIONS(8643),
    [anon_sym_POUNDhelp] = ACTIONS(8643),
    [anon_sym_POUNDquit] = ACTIONS(8643),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8643),
  },
  [STATE(4592)] = {
    [sym_block_comment] = STATE(4592),
    [sym_compiler_directive_decl] = STATE(4592),
    [sym_fsi_directive_decl] = STATE(4592),
    [sym_preproc_line] = STATE(4592),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_GT] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4593)] = {
    [sym_block_comment] = STATE(4593),
    [sym_compiler_directive_decl] = STATE(4593),
    [sym_fsi_directive_decl] = STATE(4593),
    [sym_preproc_line] = STATE(4593),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__untyped_quoted_close] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4594)] = {
    [sym_block_comment] = STATE(4594),
    [sym_compiler_directive_decl] = STATE(4594),
    [sym_fsi_directive_decl] = STATE(4594),
    [sym_preproc_line] = STATE(4594),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__untyped_quoted_close] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4595)] = {
    [sym_block_comment] = STATE(4595),
    [sym_compiler_directive_decl] = STATE(4595),
    [sym_fsi_directive_decl] = STATE(4595),
    [sym_preproc_line] = STATE(4595),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_DOT_DOT] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4596)] = {
    [sym_block_comment] = STATE(4596),
    [sym_compiler_directive_decl] = STATE(4596),
    [sym_fsi_directive_decl] = STATE(4596),
    [sym_preproc_line] = STATE(4596),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_as] = ACTIONS(6692),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_DOT_DOT] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__dedent] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(4597)] = {
    [sym_attributes] = STATE(13407),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7106),
    [sym_access_modifier] = STATE(21434),
    [sym__class_function_or_value_defn] = STATE(7107),
    [aux_sym__type_extension_inner] = STATE(4591),
    [sym__type_defn_elements] = STATE(7107),
    [sym_interface_implementation] = STATE(7108),
    [sym_member_defn] = STATE(7108),
    [sym_additional_constr_defn] = STATE(7109),
    [sym_block_comment] = STATE(4597),
    [sym_compiler_directive_decl] = STATE(4597),
    [sym_fsi_directive_decl] = STATE(4597),
    [sym_preproc_line] = STATE(4597),
    [sym_preproc_if_in_class_definition] = STATE(7108),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [ts_builtin_sym_end] = ACTIONS(8622),
    [sym_identifier] = ACTIONS(8620),
    [anon_sym_module] = ACTIONS(8620),
    [anon_sym_null] = ACTIONS(8620),
    [anon_sym_begin] = ACTIONS(8620),
    [anon_sym_open] = ACTIONS(8620),
    [anon_sym_type] = ACTIONS(8620),
    [anon_sym_LBRACK_LT] = ACTIONS(8622),
    [anon_sym_do] = ACTIONS(8620),
    [anon_sym_and] = ACTIONS(8620),
    [anon_sym_let] = ACTIONS(8620),
    [anon_sym_let_BANG] = ACTIONS(8622),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8620),
    [anon_sym_AMP] = ACTIONS(8620),
    [anon_sym_LBRACK] = ACTIONS(8620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8622),
    [anon_sym_LBRACE] = ACTIONS(8620),
    [anon_sym_LT_AT] = ACTIONS(8620),
    [anon_sym_LT_AT_AT] = ACTIONS(8622),
    [anon_sym_CARET] = ACTIONS(8622),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8622),
    [anon_sym_new] = ACTIONS(8620),
    [anon_sym_return] = ACTIONS(8620),
    [anon_sym_return_BANG] = ACTIONS(8622),
    [anon_sym_yield] = ACTIONS(8620),
    [anon_sym_yield_BANG] = ACTIONS(8622),
    [anon_sym_lazy] = ACTIONS(8620),
    [anon_sym_assert] = ACTIONS(8620),
    [anon_sym_upcast] = ACTIONS(8620),
    [anon_sym_downcast] = ACTIONS(8620),
    [anon_sym_fixed] = ACTIONS(8620),
    [anon_sym_for] = ACTIONS(8620),
    [anon_sym_while] = ACTIONS(8620),
    [anon_sym_while_BANG] = ACTIONS(8622),
    [anon_sym_if] = ACTIONS(8620),
    [anon_sym_fun] = ACTIONS(8620),
    [anon_sym_try] = ACTIONS(8620),
    [anon_sym_match] = ACTIONS(8620),
    [anon_sym_match_BANG] = ACTIONS(8622),
    [anon_sym_function] = ACTIONS(8620),
    [anon_sym_use] = ACTIONS(8620),
    [anon_sym_use_BANG] = ACTIONS(8622),
    [anon_sym_do_BANG] = ACTIONS(8622),
    [anon_sym_AMP_AMP] = ACTIONS(8622),
    [anon_sym_default] = ACTIONS(8510),
    [anon_sym_static] = ACTIONS(8512),
    [anon_sym_member] = ACTIONS(8613),
    [anon_sym_exception] = ACTIONS(8620),
    [anon_sym_interface] = ACTIONS(8516),
    [anon_sym_abstract] = ACTIONS(8518),
    [anon_sym_val] = ACTIONS(8520),
    [anon_sym_override] = ACTIONS(8510),
    [anon_sym_extern] = ACTIONS(8620),
    [aux_sym_char_token1] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8620),
    [anon_sym_DQUOTE] = ACTIONS(8620),
    [anon_sym_AT_DQUOTE] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [sym_bool] = ACTIONS(8620),
    [sym_unit] = ACTIONS(8622),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8620),
    [sym_op_identifier] = ACTIONS(8622),
    [anon_sym_PLUS] = ACTIONS(8620),
    [anon_sym_DASH] = ACTIONS(8620),
    [anon_sym_PLUS_DOT] = ACTIONS(8622),
    [anon_sym_DASH_DOT] = ACTIONS(8622),
    [anon_sym_PERCENT] = ACTIONS(8620),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8622),
    [anon_sym_TILDE] = ACTIONS(8622),
    [aux_sym_prefix_op_token1] = ACTIONS(8622),
    [sym_int] = ACTIONS(8620),
    [sym_xint] = ACTIONS(8622),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8622),
    [anon_sym_POUNDload] = ACTIONS(8622),
    [anon_sym_POUNDtime] = ACTIONS(8622),
    [anon_sym_POUNDI] = ACTIONS(8622),
    [anon_sym_POUNDhelp] = ACTIONS(8622),
    [anon_sym_POUNDquit] = ACTIONS(8622),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8622),
  },
  [STATE(4598)] = {
    [sym_block_comment] = STATE(4598),
    [sym_compiler_directive_decl] = STATE(4598),
    [sym_fsi_directive_decl] = STATE(4598),
    [sym_preproc_line] = STATE(4598),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__untyped_quoted_close] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4599)] = {
    [sym_block_comment] = STATE(4599),
    [sym_compiler_directive_decl] = STATE(4599),
    [sym_fsi_directive_decl] = STATE(4599),
    [sym_preproc_line] = STATE(4599),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__multi_dollar_interp_end] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4600)] = {
    [sym_block_comment] = STATE(4600),
    [sym_compiler_directive_decl] = STATE(4600),
    [sym_fsi_directive_decl] = STATE(4600),
    [sym_preproc_line] = STATE(4600),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__untyped_quoted_close] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4601)] = {
    [sym_block_comment] = STATE(4601),
    [sym_compiler_directive_decl] = STATE(4601),
    [sym_fsi_directive_decl] = STATE(4601),
    [sym_preproc_line] = STATE(4601),
    [aux_sym_rules_repeat1] = STATE(4613),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9024),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6607),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_DOT_DOT2] = ACTIONS(6605),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9036),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4602)] = {
    [sym_block_comment] = STATE(4602),
    [sym_compiler_directive_decl] = STATE(4602),
    [sym_fsi_directive_decl] = STATE(4602),
    [sym_preproc_line] = STATE(4602),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_DASH_GT] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9039),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4603)] = {
    [sym_block_comment] = STATE(4603),
    [sym_compiler_directive_decl] = STATE(4603),
    [sym_fsi_directive_decl] = STATE(4603),
    [sym_preproc_line] = STATE(4603),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [anon_sym_POUNDendif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4604)] = {
    [sym_block_comment] = STATE(4604),
    [sym_compiler_directive_decl] = STATE(4604),
    [sym_fsi_directive_decl] = STATE(4604),
    [sym_preproc_line] = STATE(4604),
    [aux_sym_rules_repeat1] = STATE(4614),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9024),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6607),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [anon_sym_DOT_DOT2] = ACTIONS(6605),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9036),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4605)] = {
    [sym_block_comment] = STATE(4605),
    [sym_compiler_directive_decl] = STATE(4605),
    [sym_fsi_directive_decl] = STATE(4605),
    [sym_preproc_line] = STATE(4605),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [anon_sym_POUNDendif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4606)] = {
    [sym_block_comment] = STATE(4606),
    [sym_compiler_directive_decl] = STATE(4606),
    [sym_fsi_directive_decl] = STATE(4606),
    [sym_preproc_line] = STATE(4606),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__quoted_close] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4607)] = {
    [sym_block_comment] = STATE(4607),
    [sym_compiler_directive_decl] = STATE(4607),
    [sym_fsi_directive_decl] = STATE(4607),
    [sym_preproc_line] = STATE(4607),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [anon_sym_POUNDendif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4608)] = {
    [sym_block_comment] = STATE(4608),
    [sym_compiler_directive_decl] = STATE(4608),
    [sym_fsi_directive_decl] = STATE(4608),
    [sym_preproc_line] = STATE(4608),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [anon_sym_POUNDendif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4609)] = {
    [sym_block_comment] = STATE(4609),
    [sym_compiler_directive_decl] = STATE(4609),
    [sym_fsi_directive_decl] = STATE(4609),
    [sym_preproc_line] = STATE(4609),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_EQ2] = ACTIONS(9041),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4610)] = {
    [sym_block_comment] = STATE(4610),
    [sym_compiler_directive_decl] = STATE(4610),
    [sym_fsi_directive_decl] = STATE(4610),
    [sym_preproc_line] = STATE(4610),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__quoted_close] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4611)] = {
    [sym_block_comment] = STATE(4611),
    [sym_compiler_directive_decl] = STATE(4611),
    [sym_fsi_directive_decl] = STATE(4611),
    [sym_preproc_line] = STATE(4611),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_DOT_DOT] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4612)] = {
    [sym_block_comment] = STATE(4612),
    [sym_compiler_directive_decl] = STATE(4612),
    [sym_fsi_directive_decl] = STATE(4612),
    [sym_preproc_line] = STATE(4612),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__quoted_close] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4613)] = {
    [sym_block_comment] = STATE(4613),
    [sym_compiler_directive_decl] = STATE(4613),
    [sym_fsi_directive_decl] = STATE(4613),
    [sym_preproc_line] = STATE(4613),
    [aux_sym_rules_repeat1] = STATE(4614),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9024),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_DOT_DOT] = ACTIONS(6618),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [anon_sym_DOT_DOT2] = ACTIONS(6616),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9043),
    [sym__dedent] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4614)] = {
    [sym_block_comment] = STATE(4614),
    [sym_compiler_directive_decl] = STATE(4614),
    [sym_fsi_directive_decl] = STATE(4614),
    [sym_preproc_line] = STATE(4614),
    [aux_sym_rules_repeat1] = STATE(4614),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9046),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6653),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_DOT_DOT2] = ACTIONS(6651),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9049),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4615)] = {
    [sym_block_comment] = STATE(4615),
    [sym_compiler_directive_decl] = STATE(4615),
    [sym_fsi_directive_decl] = STATE(4615),
    [sym_preproc_line] = STATE(4615),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_as] = ACTIONS(6597),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9052),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4616)] = {
    [sym_block_comment] = STATE(4616),
    [sym_compiler_directive_decl] = STATE(4616),
    [sym_fsi_directive_decl] = STATE(4616),
    [sym_preproc_line] = STATE(4616),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__quoted_close] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4617)] = {
    [sym_block_comment] = STATE(4617),
    [sym_compiler_directive_decl] = STATE(4617),
    [sym_fsi_directive_decl] = STATE(4617),
    [sym_preproc_line] = STATE(4617),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_as] = ACTIONS(6824),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_DOT_DOT] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__dedent] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(4618)] = {
    [sym_block_comment] = STATE(4618),
    [sym_compiler_directive_decl] = STATE(4618),
    [sym_fsi_directive_decl] = STATE(4618),
    [sym_preproc_line] = STATE(4618),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__untyped_quoted_close] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4619)] = {
    [sym_block_comment] = STATE(4619),
    [sym_compiler_directive_decl] = STATE(4619),
    [sym_fsi_directive_decl] = STATE(4619),
    [sym_preproc_line] = STATE(4619),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4620)] = {
    [sym_block_comment] = STATE(4620),
    [sym_compiler_directive_decl] = STATE(4620),
    [sym_fsi_directive_decl] = STATE(4620),
    [sym_preproc_line] = STATE(4620),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [anon_sym_POUNDendif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4621)] = {
    [sym_block_comment] = STATE(4621),
    [sym_compiler_directive_decl] = STATE(4621),
    [sym_fsi_directive_decl] = STATE(4621),
    [sym_preproc_line] = STATE(4621),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__quoted_close] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4622)] = {
    [sym_block_comment] = STATE(4622),
    [sym_compiler_directive_decl] = STATE(4622),
    [sym_fsi_directive_decl] = STATE(4622),
    [sym_preproc_line] = STATE(4622),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4623)] = {
    [sym_block_comment] = STATE(4623),
    [sym_compiler_directive_decl] = STATE(4623),
    [sym_fsi_directive_decl] = STATE(4623),
    [sym_preproc_line] = STATE(4623),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9054),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_DASH_GT] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6597),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_DOT_DOT2] = ACTIONS(6599),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4624)] = {
    [sym_block_comment] = STATE(4624),
    [sym_compiler_directive_decl] = STATE(4624),
    [sym_fsi_directive_decl] = STATE(4624),
    [sym_preproc_line] = STATE(4624),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [anon_sym_POUNDendif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4625)] = {
    [sym_block_comment] = STATE(4625),
    [sym_compiler_directive_decl] = STATE(4625),
    [sym_fsi_directive_decl] = STATE(4625),
    [sym_preproc_line] = STATE(4625),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_as] = ACTIONS(6863),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_DOT_DOT] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__dedent] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(4626)] = {
    [sym_block_comment] = STATE(4626),
    [sym_compiler_directive_decl] = STATE(4626),
    [sym_fsi_directive_decl] = STATE(4626),
    [sym_preproc_line] = STATE(4626),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_DOT_DOT] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4627)] = {
    [sym_attributes] = STATE(13289),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7173),
    [sym_access_modifier] = STATE(22403),
    [sym__class_function_or_value_defn] = STATE(7174),
    [aux_sym__type_extension_inner] = STATE(4636),
    [sym__type_defn_elements] = STATE(7174),
    [sym_interface_implementation] = STATE(7175),
    [sym_member_defn] = STATE(7175),
    [sym_additional_constr_defn] = STATE(7146),
    [sym_block_comment] = STATE(4627),
    [sym_compiler_directive_decl] = STATE(4627),
    [sym_fsi_directive_decl] = STATE(4627),
    [sym_preproc_line] = STATE(4627),
    [sym_preproc_if_in_class_definition] = STATE(7175),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8620),
    [anon_sym_module] = ACTIONS(8620),
    [anon_sym_null] = ACTIONS(8620),
    [anon_sym_begin] = ACTIONS(8620),
    [anon_sym_open] = ACTIONS(8620),
    [anon_sym_type] = ACTIONS(8620),
    [anon_sym_LBRACK_LT] = ACTIONS(8622),
    [anon_sym_do] = ACTIONS(8620),
    [anon_sym_and] = ACTIONS(8620),
    [anon_sym_let] = ACTIONS(8620),
    [anon_sym_let_BANG] = ACTIONS(8622),
    [aux_sym_access_modifier_token1] = ACTIONS(8246),
    [anon_sym_LPAREN] = ACTIONS(8620),
    [anon_sym_AMP] = ACTIONS(8620),
    [anon_sym_LBRACK] = ACTIONS(8620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8622),
    [anon_sym_LBRACE] = ACTIONS(8620),
    [anon_sym_LT_AT] = ACTIONS(8620),
    [anon_sym_LT_AT_AT] = ACTIONS(8622),
    [anon_sym_CARET] = ACTIONS(8622),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8622),
    [anon_sym_new] = ACTIONS(8620),
    [anon_sym_return] = ACTIONS(8620),
    [anon_sym_return_BANG] = ACTIONS(8622),
    [anon_sym_yield] = ACTIONS(8620),
    [anon_sym_yield_BANG] = ACTIONS(8622),
    [anon_sym_lazy] = ACTIONS(8620),
    [anon_sym_assert] = ACTIONS(8620),
    [anon_sym_upcast] = ACTIONS(8620),
    [anon_sym_downcast] = ACTIONS(8620),
    [anon_sym_fixed] = ACTIONS(8620),
    [anon_sym_for] = ACTIONS(8620),
    [anon_sym_while] = ACTIONS(8620),
    [anon_sym_while_BANG] = ACTIONS(8622),
    [anon_sym_if] = ACTIONS(8620),
    [anon_sym_fun] = ACTIONS(8620),
    [anon_sym_try] = ACTIONS(8620),
    [anon_sym_match] = ACTIONS(8620),
    [anon_sym_match_BANG] = ACTIONS(8622),
    [anon_sym_function] = ACTIONS(8620),
    [anon_sym_use] = ACTIONS(8620),
    [anon_sym_use_BANG] = ACTIONS(8622),
    [anon_sym_do_BANG] = ACTIONS(8622),
    [anon_sym_AMP_AMP] = ACTIONS(8622),
    [anon_sym_default] = ACTIONS(8578),
    [anon_sym_static] = ACTIONS(8580),
    [anon_sym_member] = ACTIONS(8582),
    [anon_sym_exception] = ACTIONS(8620),
    [anon_sym_interface] = ACTIONS(8584),
    [anon_sym_abstract] = ACTIONS(8586),
    [anon_sym_val] = ACTIONS(8588),
    [anon_sym_override] = ACTIONS(8578),
    [anon_sym_extern] = ACTIONS(8620),
    [aux_sym_char_token1] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8620),
    [anon_sym_DQUOTE] = ACTIONS(8620),
    [anon_sym_AT_DQUOTE] = ACTIONS(8622),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8622),
    [sym_bool] = ACTIONS(8620),
    [sym_unit] = ACTIONS(8622),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8620),
    [sym_op_identifier] = ACTIONS(8622),
    [anon_sym_PLUS] = ACTIONS(8620),
    [anon_sym_DASH] = ACTIONS(8620),
    [anon_sym_PLUS_DOT] = ACTIONS(8622),
    [anon_sym_DASH_DOT] = ACTIONS(8622),
    [anon_sym_PERCENT] = ACTIONS(8620),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8622),
    [anon_sym_TILDE] = ACTIONS(8622),
    [aux_sym_prefix_op_token1] = ACTIONS(8622),
    [sym_int] = ACTIONS(8620),
    [sym_xint] = ACTIONS(8622),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8622),
    [anon_sym_POUNDload] = ACTIONS(8622),
    [anon_sym_POUNDtime] = ACTIONS(8622),
    [anon_sym_POUNDI] = ACTIONS(8622),
    [anon_sym_POUNDhelp] = ACTIONS(8622),
    [anon_sym_POUNDquit] = ACTIONS(8622),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(8622),
    [anon_sym_POUNDendif] = ACTIONS(8622),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8622),
  },
  [STATE(4628)] = {
    [sym_elif_expression] = STATE(4962),
    [sym_block_comment] = STATE(4628),
    [sym_compiler_directive_decl] = STATE(4628),
    [sym_fsi_directive_decl] = STATE(4628),
    [sym_preproc_line] = STATE(4628),
    [aux_sym_if_expression_repeat1] = STATE(4628),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9056),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__quoted_close] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4629)] = {
    [sym_block_comment] = STATE(4629),
    [sym_compiler_directive_decl] = STATE(4629),
    [sym_fsi_directive_decl] = STATE(4629),
    [sym_preproc_line] = STATE(4629),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_DOT_DOT] = ACTIONS(6564),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [anon_sym_DOT_DOT2] = ACTIONS(6562),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__dedent] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4630)] = {
    [sym_block_comment] = STATE(4630),
    [sym_compiler_directive_decl] = STATE(4630),
    [sym_fsi_directive_decl] = STATE(4630),
    [sym_preproc_line] = STATE(4630),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_as] = ACTIONS(6834),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4631)] = {
    [sym_elif_expression] = STATE(4977),
    [sym_block_comment] = STATE(4631),
    [sym_compiler_directive_decl] = STATE(4631),
    [sym_fsi_directive_decl] = STATE(4631),
    [sym_preproc_line] = STATE(4631),
    [aux_sym_if_expression_repeat1] = STATE(4631),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9061),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_DOT_DOT] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4632)] = {
    [sym_block_comment] = STATE(4632),
    [sym_compiler_directive_decl] = STATE(4632),
    [sym_fsi_directive_decl] = STATE(4632),
    [sym_preproc_line] = STATE(4632),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_as] = ACTIONS(6612),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4633)] = {
    [sym_block_comment] = STATE(4633),
    [sym_compiler_directive_decl] = STATE(4633),
    [sym_fsi_directive_decl] = STATE(4633),
    [sym_preproc_line] = STATE(4633),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5951),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7746),
    [aux_sym_decimal_token1] = ACTIONS(7298),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4634)] = {
    [sym_elif_expression] = STATE(4961),
    [sym_block_comment] = STATE(4634),
    [sym_compiler_directive_decl] = STATE(4634),
    [sym_fsi_directive_decl] = STATE(4634),
    [sym_preproc_line] = STATE(4634),
    [aux_sym_if_expression_repeat1] = STATE(4634),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_GT] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9064),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4635)] = {
    [sym_block_comment] = STATE(4635),
    [sym_compiler_directive_decl] = STATE(4635),
    [sym_fsi_directive_decl] = STATE(4635),
    [sym_preproc_line] = STATE(4635),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__multi_dollar_interp_end] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4636)] = {
    [sym_attributes] = STATE(13289),
    [sym__attribute_set] = STATE(12910),
    [sym_function_or_value_defn] = STATE(7173),
    [sym_access_modifier] = STATE(22403),
    [sym__class_function_or_value_defn] = STATE(7174),
    [aux_sym__type_extension_inner] = STATE(4636),
    [sym__type_defn_elements] = STATE(7174),
    [sym_interface_implementation] = STATE(7175),
    [sym_member_defn] = STATE(7175),
    [sym_additional_constr_defn] = STATE(7146),
    [sym_block_comment] = STATE(4636),
    [sym_compiler_directive_decl] = STATE(4636),
    [sym_fsi_directive_decl] = STATE(4636),
    [sym_preproc_line] = STATE(4636),
    [sym_preproc_if_in_class_definition] = STATE(7175),
    [aux_sym_attributes_repeat1] = STATE(12315),
    [sym_identifier] = ACTIONS(8626),
    [anon_sym_module] = ACTIONS(8626),
    [anon_sym_null] = ACTIONS(8626),
    [anon_sym_begin] = ACTIONS(8626),
    [anon_sym_open] = ACTIONS(8626),
    [anon_sym_type] = ACTIONS(8626),
    [anon_sym_LBRACK_LT] = ACTIONS(8628),
    [anon_sym_do] = ACTIONS(9067),
    [anon_sym_and] = ACTIONS(8626),
    [anon_sym_let] = ACTIONS(9070),
    [anon_sym_let_BANG] = ACTIONS(9073),
    [aux_sym_access_modifier_token1] = ACTIONS(8640),
    [anon_sym_LPAREN] = ACTIONS(8626),
    [anon_sym_AMP] = ACTIONS(8626),
    [anon_sym_LBRACK] = ACTIONS(8626),
    [anon_sym_LBRACK_PIPE] = ACTIONS(8643),
    [anon_sym_LBRACE] = ACTIONS(8626),
    [anon_sym_LT_AT] = ACTIONS(8626),
    [anon_sym_LT_AT_AT] = ACTIONS(8643),
    [anon_sym_CARET] = ACTIONS(8643),
    [anon_sym_LBRACE_PIPE] = ACTIONS(8643),
    [anon_sym_new] = ACTIONS(9076),
    [anon_sym_return] = ACTIONS(8626),
    [anon_sym_return_BANG] = ACTIONS(8643),
    [anon_sym_yield] = ACTIONS(8626),
    [anon_sym_yield_BANG] = ACTIONS(8643),
    [anon_sym_lazy] = ACTIONS(8626),
    [anon_sym_assert] = ACTIONS(8626),
    [anon_sym_upcast] = ACTIONS(8626),
    [anon_sym_downcast] = ACTIONS(8626),
    [anon_sym_fixed] = ACTIONS(8626),
    [anon_sym_for] = ACTIONS(8626),
    [anon_sym_while] = ACTIONS(8626),
    [anon_sym_while_BANG] = ACTIONS(8643),
    [anon_sym_if] = ACTIONS(8626),
    [anon_sym_fun] = ACTIONS(8626),
    [anon_sym_try] = ACTIONS(8626),
    [anon_sym_match] = ACTIONS(8626),
    [anon_sym_match_BANG] = ACTIONS(8643),
    [anon_sym_function] = ACTIONS(8626),
    [anon_sym_use] = ACTIONS(8626),
    [anon_sym_use_BANG] = ACTIONS(8643),
    [anon_sym_do_BANG] = ACTIONS(8643),
    [anon_sym_AMP_AMP] = ACTIONS(8643),
    [anon_sym_default] = ACTIONS(9079),
    [anon_sym_static] = ACTIONS(9082),
    [anon_sym_member] = ACTIONS(9085),
    [anon_sym_exception] = ACTIONS(8626),
    [anon_sym_interface] = ACTIONS(9088),
    [anon_sym_abstract] = ACTIONS(9091),
    [anon_sym_val] = ACTIONS(9094),
    [anon_sym_override] = ACTIONS(9079),
    [anon_sym_extern] = ACTIONS(8626),
    [aux_sym_char_token1] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(8626),
    [anon_sym_DQUOTE] = ACTIONS(8626),
    [anon_sym_AT_DQUOTE] = ACTIONS(8643),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(8643),
    [sym_bool] = ACTIONS(8626),
    [sym_unit] = ACTIONS(8643),
    [anon_sym_LPAREN_PIPE] = ACTIONS(8626),
    [sym_op_identifier] = ACTIONS(8643),
    [anon_sym_PLUS] = ACTIONS(8626),
    [anon_sym_DASH] = ACTIONS(8626),
    [anon_sym_PLUS_DOT] = ACTIONS(8643),
    [anon_sym_DASH_DOT] = ACTIONS(8643),
    [anon_sym_PERCENT] = ACTIONS(8626),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(8643),
    [anon_sym_TILDE] = ACTIONS(8643),
    [aux_sym_prefix_op_token1] = ACTIONS(8643),
    [sym_int] = ACTIONS(8626),
    [sym_xint] = ACTIONS(8643),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(8643),
    [anon_sym_POUNDload] = ACTIONS(8643),
    [anon_sym_POUNDtime] = ACTIONS(8643),
    [anon_sym_POUNDI] = ACTIONS(8643),
    [anon_sym_POUNDhelp] = ACTIONS(8643),
    [anon_sym_POUNDquit] = ACTIONS(8643),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9097),
    [anon_sym_POUNDendif] = ACTIONS(8643),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(8643),
  },
  [STATE(4637)] = {
    [sym_block_comment] = STATE(4637),
    [sym_compiler_directive_decl] = STATE(4637),
    [sym_fsi_directive_decl] = STATE(4637),
    [sym_preproc_line] = STATE(4637),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__multi_dollar_interp_end] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4638)] = {
    [sym_block_comment] = STATE(4638),
    [sym_compiler_directive_decl] = STATE(4638),
    [sym_fsi_directive_decl] = STATE(4638),
    [sym_preproc_line] = STATE(4638),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9100),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__multi_dollar_interp_end] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4639)] = {
    [sym_block_comment] = STATE(4639),
    [sym_compiler_directive_decl] = STATE(4639),
    [sym_fsi_directive_decl] = STATE(4639),
    [sym_preproc_line] = STATE(4639),
    [aux_sym_long_identifier_repeat1] = STATE(4654),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9102),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6360),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_DOT_DOT2] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4640)] = {
    [sym_block_comment] = STATE(4640),
    [sym_compiler_directive_decl] = STATE(4640),
    [sym_fsi_directive_decl] = STATE(4640),
    [sym_preproc_line] = STATE(4640),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_as] = ACTIONS(6867),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_DOT_DOT] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__dedent] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(4641)] = {
    [sym_block_comment] = STATE(4641),
    [sym_compiler_directive_decl] = STATE(4641),
    [sym_fsi_directive_decl] = STATE(4641),
    [sym_preproc_line] = STATE(4641),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_GT] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4642)] = {
    [sym_block_comment] = STATE(4642),
    [sym_compiler_directive_decl] = STATE(4642),
    [sym_fsi_directive_decl] = STATE(4642),
    [sym_preproc_line] = STATE(4642),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__multi_dollar_interp_end] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4643)] = {
    [sym_block_comment] = STATE(4643),
    [sym_compiler_directive_decl] = STATE(4643),
    [sym_fsi_directive_decl] = STATE(4643),
    [sym_preproc_line] = STATE(4643),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_as] = ACTIONS(6871),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_DOT_DOT] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__dedent] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(4644)] = {
    [sym_block_comment] = STATE(4644),
    [sym_compiler_directive_decl] = STATE(4644),
    [sym_fsi_directive_decl] = STATE(4644),
    [sym_preproc_line] = STATE(4644),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__multi_dollar_interp_end] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4645)] = {
    [sym_block_comment] = STATE(4645),
    [sym_compiler_directive_decl] = STATE(4645),
    [sym_fsi_directive_decl] = STATE(4645),
    [sym_preproc_line] = STATE(4645),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__multi_dollar_interp_end] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4646)] = {
    [sym_block_comment] = STATE(4646),
    [sym_compiler_directive_decl] = STATE(4646),
    [sym_fsi_directive_decl] = STATE(4646),
    [sym_preproc_line] = STATE(4646),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_EQ2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4647)] = {
    [sym_block_comment] = STATE(4647),
    [sym_compiler_directive_decl] = STATE(4647),
    [sym_fsi_directive_decl] = STATE(4647),
    [sym_preproc_line] = STATE(4647),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_DOT_DOT] = ACTIONS(6637),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [anon_sym_DOT_DOT2] = ACTIONS(6635),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9104),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4648)] = {
    [sym_block_comment] = STATE(4648),
    [sym_compiler_directive_decl] = STATE(4648),
    [sym_fsi_directive_decl] = STATE(4648),
    [sym_preproc_line] = STATE(4648),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_as] = ACTIONS(6875),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_DOT_DOT] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__dedent] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(4649)] = {
    [sym_block_comment] = STATE(4649),
    [sym_compiler_directive_decl] = STATE(4649),
    [sym_fsi_directive_decl] = STATE(4649),
    [sym_preproc_line] = STATE(4649),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__multi_dollar_interp_end] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4650)] = {
    [sym_block_comment] = STATE(4650),
    [sym_compiler_directive_decl] = STATE(4650),
    [sym_fsi_directive_decl] = STATE(4650),
    [sym_preproc_line] = STATE(4650),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_GT] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4651)] = {
    [sym_block_comment] = STATE(4651),
    [sym_compiler_directive_decl] = STATE(4651),
    [sym_fsi_directive_decl] = STATE(4651),
    [sym_preproc_line] = STATE(4651),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6612),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4652)] = {
    [sym_block_comment] = STATE(4652),
    [sym_compiler_directive_decl] = STATE(4652),
    [sym_fsi_directive_decl] = STATE(4652),
    [sym_preproc_line] = STATE(4652),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_GT] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4653)] = {
    [sym_block_comment] = STATE(4653),
    [sym_compiler_directive_decl] = STATE(4653),
    [sym_fsi_directive_decl] = STATE(4653),
    [sym_preproc_line] = STATE(4653),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_GT] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9106),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4654)] = {
    [sym_block_comment] = STATE(4654),
    [sym_compiler_directive_decl] = STATE(4654),
    [sym_fsi_directive_decl] = STATE(4654),
    [sym_preproc_line] = STATE(4654),
    [aux_sym_long_identifier_repeat1] = STATE(4654),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9108),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6295),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4655)] = {
    [sym_block_comment] = STATE(4655),
    [sym_compiler_directive_decl] = STATE(4655),
    [sym_fsi_directive_decl] = STATE(4655),
    [sym_preproc_line] = STATE(4655),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_as] = ACTIONS(6828),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_DOT_DOT] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__dedent] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(4656)] = {
    [sym_block_comment] = STATE(4656),
    [sym_compiler_directive_decl] = STATE(4656),
    [sym_fsi_directive_decl] = STATE(4656),
    [sym_preproc_line] = STATE(4656),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(6302),
    [aux_sym_decimal_token1] = ACTIONS(6249),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4657)] = {
    [sym_block_comment] = STATE(4657),
    [sym_compiler_directive_decl] = STATE(4657),
    [sym_fsi_directive_decl] = STATE(4657),
    [sym_preproc_line] = STATE(4657),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_DOT_DOT] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4658)] = {
    [sym_block_comment] = STATE(4658),
    [sym_compiler_directive_decl] = STATE(4658),
    [sym_fsi_directive_decl] = STATE(4658),
    [sym_preproc_line] = STATE(4658),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4659)] = {
    [sym_block_comment] = STATE(4659),
    [sym_compiler_directive_decl] = STATE(4659),
    [sym_fsi_directive_decl] = STATE(4659),
    [sym_preproc_line] = STATE(4659),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_GT] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4660)] = {
    [sym_block_comment] = STATE(4660),
    [sym_compiler_directive_decl] = STATE(4660),
    [sym_fsi_directive_decl] = STATE(4660),
    [sym_preproc_line] = STATE(4660),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_GT] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4661)] = {
    [sym_block_comment] = STATE(4661),
    [sym_compiler_directive_decl] = STATE(4661),
    [sym_fsi_directive_decl] = STATE(4661),
    [sym_preproc_line] = STATE(4661),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__quoted_close] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4662)] = {
    [sym_block_comment] = STATE(4662),
    [sym_compiler_directive_decl] = STATE(4662),
    [sym_fsi_directive_decl] = STATE(4662),
    [sym_preproc_line] = STATE(4662),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6513),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [anon_sym_DOT_DOT2] = ACTIONS(6515),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4663)] = {
    [sym_block_comment] = STATE(4663),
    [sym_compiler_directive_decl] = STATE(4663),
    [sym_fsi_directive_decl] = STATE(4663),
    [sym_preproc_line] = STATE(4663),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__multi_dollar_interp_end] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4664)] = {
    [sym_block_comment] = STATE(4664),
    [sym_compiler_directive_decl] = STATE(4664),
    [sym_fsi_directive_decl] = STATE(4664),
    [sym_preproc_line] = STATE(4664),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__multi_dollar_interp_end] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4665)] = {
    [sym_block_comment] = STATE(4665),
    [sym_compiler_directive_decl] = STATE(4665),
    [sym_fsi_directive_decl] = STATE(4665),
    [sym_preproc_line] = STATE(4665),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__multi_dollar_interp_end] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4666)] = {
    [sym_block_comment] = STATE(4666),
    [sym_compiler_directive_decl] = STATE(4666),
    [sym_fsi_directive_decl] = STATE(4666),
    [sym_preproc_line] = STATE(4666),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_GT] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4667)] = {
    [sym_block_comment] = STATE(4667),
    [sym_compiler_directive_decl] = STATE(4667),
    [sym_fsi_directive_decl] = STATE(4667),
    [sym_preproc_line] = STATE(4667),
    [aux_sym_sequential_expression_repeat1] = STATE(4667),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_as] = ACTIONS(6800),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_DOT_DOT] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9111),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(4668)] = {
    [sym_block_comment] = STATE(4668),
    [sym_compiler_directive_decl] = STATE(4668),
    [sym_fsi_directive_decl] = STATE(4668),
    [sym_preproc_line] = STATE(4668),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_GT] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4669)] = {
    [sym_block_comment] = STATE(4669),
    [sym_compiler_directive_decl] = STATE(4669),
    [sym_fsi_directive_decl] = STATE(4669),
    [sym_preproc_line] = STATE(4669),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_DOT_DOT] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4670)] = {
    [sym_block_comment] = STATE(4670),
    [sym_compiler_directive_decl] = STATE(4670),
    [sym_fsi_directive_decl] = STATE(4670),
    [sym_preproc_line] = STATE(4670),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_GT] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4671)] = {
    [sym_block_comment] = STATE(4671),
    [sym_compiler_directive_decl] = STATE(4671),
    [sym_fsi_directive_decl] = STATE(4671),
    [sym_preproc_line] = STATE(4671),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [anon_sym_POUNDendif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4672)] = {
    [sym_block_comment] = STATE(4672),
    [sym_compiler_directive_decl] = STATE(4672),
    [sym_fsi_directive_decl] = STATE(4672),
    [sym_preproc_line] = STATE(4672),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_GT] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4673)] = {
    [sym_block_comment] = STATE(4673),
    [sym_compiler_directive_decl] = STATE(4673),
    [sym_fsi_directive_decl] = STATE(4673),
    [sym_preproc_line] = STATE(4673),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_EQ] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_COLON] = ACTIONS(6331),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_COMMA] = ACTIONS(6333),
    [anon_sym_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_QMARK] = ACTIONS(6331),
    [anon_sym_COLON_COLON] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6331),
    [anon_sym_CARET] = ACTIONS(6331),
    [anon_sym_or] = ACTIONS(6331),
    [anon_sym_DOT] = ACTIONS(6331),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6331),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_DOT_LBRACK] = ACTIONS(6333),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_LPAREN2] = ACTIONS(6331),
    [anon_sym_PIPE_PIPE] = ACTIONS(6331),
    [anon_sym_AMP_AMP] = ACTIONS(6331),
    [anon_sym_STAR] = ACTIONS(6331),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6331),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6331),
    [anon_sym_DASH_DOT] = ACTIONS(6331),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6331),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6331),
    [aux_sym_infix_op_token1] = ACTIONS(6333),
    [aux_sym_infix_op_token2] = ACTIONS(6331),
    [aux_sym_infix_op_token3] = ACTIONS(6331),
    [anon_sym_BANG_EQ] = ACTIONS(6331),
    [anon_sym_COLON_EQ] = ACTIONS(6333),
    [anon_sym_DOLLAR] = ACTIONS(6331),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6331),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6331),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__untyped_quoted_close] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
    [sym__tyapp_open] = ACTIONS(6333),
  },
  [STATE(4674)] = {
    [sym_block_comment] = STATE(4674),
    [sym_compiler_directive_decl] = STATE(4674),
    [sym_fsi_directive_decl] = STATE(4674),
    [sym_preproc_line] = STATE(4674),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_GT] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4675)] = {
    [sym_block_comment] = STATE(4675),
    [sym_compiler_directive_decl] = STATE(4675),
    [sym_fsi_directive_decl] = STATE(4675),
    [sym_preproc_line] = STATE(4675),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4676)] = {
    [sym_block_comment] = STATE(4676),
    [sym_compiler_directive_decl] = STATE(4676),
    [sym_fsi_directive_decl] = STATE(4676),
    [sym_preproc_line] = STATE(4676),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__multi_dollar_interp_end] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4677)] = {
    [sym_elif_expression] = STATE(5164),
    [sym_block_comment] = STATE(4677),
    [sym_compiler_directive_decl] = STATE(4677),
    [sym_fsi_directive_decl] = STATE(4677),
    [sym_preproc_line] = STATE(4677),
    [aux_sym_if_expression_repeat1] = STATE(4677),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9114),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [anon_sym_POUNDendif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4678)] = {
    [sym_block_comment] = STATE(4678),
    [sym_compiler_directive_decl] = STATE(4678),
    [sym_fsi_directive_decl] = STATE(4678),
    [sym_preproc_line] = STATE(4678),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__multi_dollar_interp_end] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4679)] = {
    [sym_block_comment] = STATE(4679),
    [sym_compiler_directive_decl] = STATE(4679),
    [sym_fsi_directive_decl] = STATE(4679),
    [sym_preproc_line] = STATE(4679),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8004),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_DASH_GT] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4680)] = {
    [sym_block_comment] = STATE(4680),
    [sym_compiler_directive_decl] = STATE(4680),
    [sym_fsi_directive_decl] = STATE(4680),
    [sym_preproc_line] = STATE(4680),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4681)] = {
    [sym_block_comment] = STATE(4681),
    [sym_compiler_directive_decl] = STATE(4681),
    [sym_fsi_directive_decl] = STATE(4681),
    [sym_preproc_line] = STATE(4681),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__multi_dollar_interp_end] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4682)] = {
    [sym_block_comment] = STATE(4682),
    [sym_compiler_directive_decl] = STATE(4682),
    [sym_fsi_directive_decl] = STATE(4682),
    [sym_preproc_line] = STATE(4682),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_as] = ACTIONS(6807),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4683)] = {
    [sym_block_comment] = STATE(4683),
    [sym_compiler_directive_decl] = STATE(4683),
    [sym_fsi_directive_decl] = STATE(4683),
    [sym_preproc_line] = STATE(4683),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__quoted_close] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4684)] = {
    [sym_block_comment] = STATE(4684),
    [sym_compiler_directive_decl] = STATE(4684),
    [sym_fsi_directive_decl] = STATE(4684),
    [sym_preproc_line] = STATE(4684),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4685)] = {
    [sym_elif_expression] = STATE(5077),
    [sym_block_comment] = STATE(4685),
    [sym_compiler_directive_decl] = STATE(4685),
    [sym_fsi_directive_decl] = STATE(4685),
    [sym_preproc_line] = STATE(4685),
    [aux_sym_if_expression_repeat1] = STATE(4685),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9117),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__multi_dollar_interp_end] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4686)] = {
    [sym_block_comment] = STATE(4686),
    [sym_compiler_directive_decl] = STATE(4686),
    [sym_fsi_directive_decl] = STATE(4686),
    [sym_preproc_line] = STATE(4686),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__quoted_close] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4687)] = {
    [sym_block_comment] = STATE(4687),
    [sym_compiler_directive_decl] = STATE(4687),
    [sym_fsi_directive_decl] = STATE(4687),
    [sym_preproc_line] = STATE(4687),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_with] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9120),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4688)] = {
    [sym_block_comment] = STATE(4688),
    [sym_compiler_directive_decl] = STATE(4688),
    [sym_fsi_directive_decl] = STATE(4688),
    [sym_preproc_line] = STATE(4688),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_DOT_DOT] = ACTIONS(6560),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [anon_sym_DOT_DOT2] = ACTIONS(6558),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4689)] = {
    [sym_block_comment] = STATE(4689),
    [sym_compiler_directive_decl] = STATE(4689),
    [sym_fsi_directive_decl] = STATE(4689),
    [sym_preproc_line] = STATE(4689),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__quoted_close] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4690)] = {
    [sym_block_comment] = STATE(4690),
    [sym_compiler_directive_decl] = STATE(4690),
    [sym_fsi_directive_decl] = STATE(4690),
    [sym_preproc_line] = STATE(4690),
    [aux_sym_sequential_expression_repeat1] = STATE(4667),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_as] = ACTIONS(6740),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_DOT_DOT] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(4691)] = {
    [sym_block_comment] = STATE(4691),
    [sym_compiler_directive_decl] = STATE(4691),
    [sym_fsi_directive_decl] = STATE(4691),
    [sym_preproc_line] = STATE(4691),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4692)] = {
    [sym_block_comment] = STATE(4692),
    [sym_compiler_directive_decl] = STATE(4692),
    [sym_fsi_directive_decl] = STATE(4692),
    [sym_preproc_line] = STATE(4692),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4693)] = {
    [sym_block_comment] = STATE(4693),
    [sym_compiler_directive_decl] = STATE(4693),
    [sym_fsi_directive_decl] = STATE(4693),
    [sym_preproc_line] = STATE(4693),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__quoted_close] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4694)] = {
    [sym_block_comment] = STATE(4694),
    [sym_compiler_directive_decl] = STATE(4694),
    [sym_fsi_directive_decl] = STATE(4694),
    [sym_preproc_line] = STATE(4694),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__quoted_close] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4695)] = {
    [sym_block_comment] = STATE(4695),
    [sym_compiler_directive_decl] = STATE(4695),
    [sym_fsi_directive_decl] = STATE(4695),
    [sym_preproc_line] = STATE(4695),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__untyped_quoted_close] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4696)] = {
    [sym_block_comment] = STATE(4696),
    [sym_compiler_directive_decl] = STATE(4696),
    [sym_fsi_directive_decl] = STATE(4696),
    [sym_preproc_line] = STATE(4696),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_DASH_GT] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4697)] = {
    [sym_block_comment] = STATE(4697),
    [sym_compiler_directive_decl] = STATE(4697),
    [sym_fsi_directive_decl] = STATE(4697),
    [sym_preproc_line] = STATE(4697),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_DOT_DOT] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9122),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4698)] = {
    [sym_block_comment] = STATE(4698),
    [sym_compiler_directive_decl] = STATE(4698),
    [sym_fsi_directive_decl] = STATE(4698),
    [sym_preproc_line] = STATE(4698),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4699)] = {
    [sym_block_comment] = STATE(4699),
    [sym_compiler_directive_decl] = STATE(4699),
    [sym_fsi_directive_decl] = STATE(4699),
    [sym_preproc_line] = STATE(4699),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_GT] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4700)] = {
    [sym_block_comment] = STATE(4700),
    [sym_compiler_directive_decl] = STATE(4700),
    [sym_fsi_directive_decl] = STATE(4700),
    [sym_preproc_line] = STATE(4700),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_GT] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4701)] = {
    [sym_block_comment] = STATE(4701),
    [sym_compiler_directive_decl] = STATE(4701),
    [sym_fsi_directive_decl] = STATE(4701),
    [sym_preproc_line] = STATE(4701),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_GT] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4702)] = {
    [sym_block_comment] = STATE(4702),
    [sym_compiler_directive_decl] = STATE(4702),
    [sym_fsi_directive_decl] = STATE(4702),
    [sym_preproc_line] = STATE(4702),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_DOT_DOT] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4703)] = {
    [sym_block_comment] = STATE(4703),
    [sym_compiler_directive_decl] = STATE(4703),
    [sym_fsi_directive_decl] = STATE(4703),
    [sym_preproc_line] = STATE(4703),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__untyped_quoted_close] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4704)] = {
    [sym_block_comment] = STATE(4704),
    [sym_compiler_directive_decl] = STATE(4704),
    [sym_fsi_directive_decl] = STATE(4704),
    [sym_preproc_line] = STATE(4704),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__untyped_quoted_close] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4705)] = {
    [sym_block_comment] = STATE(4705),
    [sym_compiler_directive_decl] = STATE(4705),
    [sym_fsi_directive_decl] = STATE(4705),
    [sym_preproc_line] = STATE(4705),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__untyped_quoted_close] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4706)] = {
    [sym_block_comment] = STATE(4706),
    [sym_compiler_directive_decl] = STATE(4706),
    [sym_fsi_directive_decl] = STATE(4706),
    [sym_preproc_line] = STATE(4706),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_with] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4707)] = {
    [sym_block_comment] = STATE(4707),
    [sym_compiler_directive_decl] = STATE(4707),
    [sym_fsi_directive_decl] = STATE(4707),
    [sym_preproc_line] = STATE(4707),
    [sym_identifier] = ACTIONS(6379),
    [anon_sym_null] = ACTIONS(6379),
    [anon_sym_EQ] = ACTIONS(6379),
    [anon_sym_begin] = ACTIONS(6379),
    [anon_sym_COLON] = ACTIONS(6379),
    [anon_sym_do] = ACTIONS(6379),
    [anon_sym_let] = ACTIONS(6379),
    [anon_sym_let_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN] = ACTIONS(6379),
    [anon_sym_COMMA] = ACTIONS(6381),
    [anon_sym_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_QMARK] = ACTIONS(6379),
    [anon_sym_COLON_COLON] = ACTIONS(6381),
    [anon_sym_AMP] = ACTIONS(6379),
    [anon_sym_LBRACK] = ACTIONS(6379),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6381),
    [anon_sym_LBRACE] = ACTIONS(6379),
    [anon_sym_LT_AT] = ACTIONS(6379),
    [anon_sym_LT_AT_AT] = ACTIONS(6379),
    [anon_sym_CARET] = ACTIONS(6379),
    [anon_sym_or] = ACTIONS(6379),
    [anon_sym_DOT] = ACTIONS(6379),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6381),
    [anon_sym_new] = ACTIONS(6379),
    [anon_sym_with] = ACTIONS(6379),
    [anon_sym_return] = ACTIONS(6379),
    [anon_sym_return_BANG] = ACTIONS(6381),
    [anon_sym_yield] = ACTIONS(6379),
    [anon_sym_yield_BANG] = ACTIONS(6381),
    [anon_sym_lazy] = ACTIONS(6379),
    [anon_sym_assert] = ACTIONS(6379),
    [anon_sym_upcast] = ACTIONS(6379),
    [anon_sym_downcast] = ACTIONS(6379),
    [anon_sym_fixed] = ACTIONS(6379),
    [anon_sym_COLON_GT] = ACTIONS(6381),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6381),
    [anon_sym_for] = ACTIONS(6379),
    [anon_sym_while] = ACTIONS(6379),
    [anon_sym_while_BANG] = ACTIONS(6381),
    [anon_sym_if] = ACTIONS(6379),
    [anon_sym_fun] = ACTIONS(6379),
    [anon_sym_DASH_GT] = ACTIONS(6379),
    [anon_sym_try] = ACTIONS(6379),
    [anon_sym_match] = ACTIONS(6379),
    [anon_sym_match_BANG] = ACTIONS(6381),
    [anon_sym_function] = ACTIONS(6379),
    [anon_sym_DOT_LBRACK] = ACTIONS(6381),
    [anon_sym_use] = ACTIONS(6379),
    [anon_sym_use_BANG] = ACTIONS(6381),
    [anon_sym_do_BANG] = ACTIONS(6381),
    [anon_sym_LPAREN2] = ACTIONS(6379),
    [anon_sym_PIPE_PIPE] = ACTIONS(6379),
    [anon_sym_AMP_AMP] = ACTIONS(6379),
    [anon_sym_STAR] = ACTIONS(6379),
    [anon_sym_LT] = ACTIONS(6379),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6381),
    [aux_sym_char_token1] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6379),
    [anon_sym_DQUOTE] = ACTIONS(6379),
    [anon_sym_AT_DQUOTE] = ACTIONS(6381),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6381),
    [sym_bool] = ACTIONS(6379),
    [sym_unit] = ACTIONS(6381),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6379),
    [sym_op_identifier] = ACTIONS(6379),
    [anon_sym_PLUS] = ACTIONS(6379),
    [anon_sym_DASH] = ACTIONS(6379),
    [anon_sym_PLUS_DOT] = ACTIONS(6379),
    [anon_sym_DASH_DOT] = ACTIONS(6379),
    [anon_sym_PERCENT] = ACTIONS(6379),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6379),
    [anon_sym_TILDE] = ACTIONS(6381),
    [aux_sym_prefix_op_token1] = ACTIONS(6379),
    [aux_sym_infix_op_token1] = ACTIONS(6381),
    [aux_sym_infix_op_token2] = ACTIONS(6379),
    [aux_sym_infix_op_token3] = ACTIONS(6379),
    [anon_sym_BANG_EQ] = ACTIONS(6379),
    [anon_sym_COLON_EQ] = ACTIONS(6381),
    [anon_sym_DOLLAR] = ACTIONS(6379),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6379),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6379),
    [sym_int] = ACTIONS(6379),
    [sym_xint] = ACTIONS(6381),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6381),
    [sym__newline] = ACTIONS(6381),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6381),
    [sym__tyapp_open] = ACTIONS(6381),
  },
  [STATE(4708)] = {
    [sym_block_comment] = STATE(4708),
    [sym_compiler_directive_decl] = STATE(4708),
    [sym_fsi_directive_decl] = STATE(4708),
    [sym_preproc_line] = STATE(4708),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_as] = ACTIONS(6750),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_DOT_DOT] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__dedent] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(4709)] = {
    [sym_block_comment] = STATE(4709),
    [sym_compiler_directive_decl] = STATE(4709),
    [sym_fsi_directive_decl] = STATE(4709),
    [sym_preproc_line] = STATE(4709),
    [sym_identifier] = ACTIONS(6383),
    [anon_sym_null] = ACTIONS(6383),
    [anon_sym_EQ] = ACTIONS(6383),
    [anon_sym_begin] = ACTIONS(6383),
    [anon_sym_COLON] = ACTIONS(6383),
    [anon_sym_do] = ACTIONS(6383),
    [anon_sym_let] = ACTIONS(6383),
    [anon_sym_let_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN] = ACTIONS(6383),
    [anon_sym_COMMA] = ACTIONS(6385),
    [anon_sym_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_QMARK] = ACTIONS(6383),
    [anon_sym_COLON_COLON] = ACTIONS(6385),
    [anon_sym_AMP] = ACTIONS(6383),
    [anon_sym_LBRACK] = ACTIONS(6383),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6385),
    [anon_sym_LBRACE] = ACTIONS(6383),
    [anon_sym_LT_AT] = ACTIONS(6383),
    [anon_sym_LT_AT_AT] = ACTIONS(6383),
    [anon_sym_CARET] = ACTIONS(6383),
    [anon_sym_or] = ACTIONS(6383),
    [anon_sym_DOT] = ACTIONS(6383),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6385),
    [anon_sym_new] = ACTIONS(6383),
    [anon_sym_with] = ACTIONS(6383),
    [anon_sym_return] = ACTIONS(6383),
    [anon_sym_return_BANG] = ACTIONS(6385),
    [anon_sym_yield] = ACTIONS(6383),
    [anon_sym_yield_BANG] = ACTIONS(6385),
    [anon_sym_lazy] = ACTIONS(6383),
    [anon_sym_assert] = ACTIONS(6383),
    [anon_sym_upcast] = ACTIONS(6383),
    [anon_sym_downcast] = ACTIONS(6383),
    [anon_sym_fixed] = ACTIONS(6383),
    [anon_sym_COLON_GT] = ACTIONS(6385),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6385),
    [anon_sym_for] = ACTIONS(6383),
    [anon_sym_while] = ACTIONS(6383),
    [anon_sym_while_BANG] = ACTIONS(6385),
    [anon_sym_if] = ACTIONS(6383),
    [anon_sym_fun] = ACTIONS(6383),
    [anon_sym_DASH_GT] = ACTIONS(6383),
    [anon_sym_try] = ACTIONS(6383),
    [anon_sym_match] = ACTIONS(6383),
    [anon_sym_match_BANG] = ACTIONS(6385),
    [anon_sym_function] = ACTIONS(6383),
    [anon_sym_DOT_LBRACK] = ACTIONS(6385),
    [anon_sym_use] = ACTIONS(6383),
    [anon_sym_use_BANG] = ACTIONS(6385),
    [anon_sym_do_BANG] = ACTIONS(6385),
    [anon_sym_LPAREN2] = ACTIONS(6383),
    [anon_sym_PIPE_PIPE] = ACTIONS(6383),
    [anon_sym_AMP_AMP] = ACTIONS(6383),
    [anon_sym_STAR] = ACTIONS(6383),
    [anon_sym_LT] = ACTIONS(6383),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6385),
    [aux_sym_char_token1] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6383),
    [anon_sym_DQUOTE] = ACTIONS(6383),
    [anon_sym_AT_DQUOTE] = ACTIONS(6385),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6385),
    [sym_bool] = ACTIONS(6383),
    [sym_unit] = ACTIONS(6385),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6383),
    [sym_op_identifier] = ACTIONS(6383),
    [anon_sym_PLUS] = ACTIONS(6383),
    [anon_sym_DASH] = ACTIONS(6383),
    [anon_sym_PLUS_DOT] = ACTIONS(6383),
    [anon_sym_DASH_DOT] = ACTIONS(6383),
    [anon_sym_PERCENT] = ACTIONS(6383),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6383),
    [anon_sym_TILDE] = ACTIONS(6385),
    [aux_sym_prefix_op_token1] = ACTIONS(6383),
    [aux_sym_infix_op_token1] = ACTIONS(6385),
    [aux_sym_infix_op_token2] = ACTIONS(6383),
    [aux_sym_infix_op_token3] = ACTIONS(6383),
    [anon_sym_BANG_EQ] = ACTIONS(6383),
    [anon_sym_COLON_EQ] = ACTIONS(6385),
    [anon_sym_DOLLAR] = ACTIONS(6383),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6383),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6383),
    [sym_int] = ACTIONS(6383),
    [sym_xint] = ACTIONS(6385),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6385),
    [sym__newline] = ACTIONS(6385),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6385),
    [sym__tyapp_open] = ACTIONS(6385),
  },
  [STATE(4710)] = {
    [sym_block_comment] = STATE(4710),
    [sym_compiler_directive_decl] = STATE(4710),
    [sym_fsi_directive_decl] = STATE(4710),
    [sym_preproc_line] = STATE(4710),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__untyped_quoted_close] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4711)] = {
    [sym_block_comment] = STATE(4711),
    [sym_compiler_directive_decl] = STATE(4711),
    [sym_fsi_directive_decl] = STATE(4711),
    [sym_preproc_line] = STATE(4711),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_EQ] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_COLON] = ACTIONS(6420),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_COMMA] = ACTIONS(6418),
    [anon_sym_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_QMARK] = ACTIONS(6420),
    [anon_sym_COLON_COLON] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6420),
    [anon_sym_CARET] = ACTIONS(6420),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_DOT] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6420),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_DOT_LBRACK] = ACTIONS(6418),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_LPAREN2] = ACTIONS(6420),
    [anon_sym_PIPE_PIPE] = ACTIONS(6420),
    [anon_sym_AMP_AMP] = ACTIONS(6420),
    [anon_sym_STAR] = ACTIONS(6420),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6420),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6420),
    [anon_sym_DASH_DOT] = ACTIONS(6420),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6420),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6420),
    [aux_sym_infix_op_token1] = ACTIONS(6418),
    [aux_sym_infix_op_token2] = ACTIONS(6420),
    [aux_sym_infix_op_token3] = ACTIONS(6420),
    [anon_sym_BANG_EQ] = ACTIONS(6420),
    [anon_sym_COLON_EQ] = ACTIONS(6418),
    [anon_sym_DOLLAR] = ACTIONS(6420),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6420),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6420),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
    [sym__multi_dollar_interp_end] = ACTIONS(6418),
    [sym__tyapp_open] = ACTIONS(6418),
  },
  [STATE(4712)] = {
    [sym_block_comment] = STATE(4712),
    [sym_compiler_directive_decl] = STATE(4712),
    [sym_fsi_directive_decl] = STATE(4712),
    [sym_preproc_line] = STATE(4712),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9124),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__untyped_quoted_close] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4713)] = {
    [sym_block_comment] = STATE(4713),
    [sym_compiler_directive_decl] = STATE(4713),
    [sym_fsi_directive_decl] = STATE(4713),
    [sym_preproc_line] = STATE(4713),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_DOT_DOT] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4714)] = {
    [sym_block_comment] = STATE(4714),
    [sym_compiler_directive_decl] = STATE(4714),
    [sym_fsi_directive_decl] = STATE(4714),
    [sym_preproc_line] = STATE(4714),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4715)] = {
    [sym_block_comment] = STATE(4715),
    [sym_compiler_directive_decl] = STATE(4715),
    [sym_fsi_directive_decl] = STATE(4715),
    [sym_preproc_line] = STATE(4715),
    [sym_identifier] = ACTIONS(6411),
    [anon_sym_null] = ACTIONS(6411),
    [anon_sym_EQ] = ACTIONS(6411),
    [anon_sym_begin] = ACTIONS(6411),
    [anon_sym_COLON] = ACTIONS(6411),
    [anon_sym_do] = ACTIONS(6411),
    [anon_sym_let] = ACTIONS(6411),
    [anon_sym_let_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN] = ACTIONS(6411),
    [anon_sym_COMMA] = ACTIONS(6413),
    [anon_sym_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_QMARK] = ACTIONS(6411),
    [anon_sym_COLON_COLON] = ACTIONS(6413),
    [anon_sym_AMP] = ACTIONS(6411),
    [anon_sym_LBRACK] = ACTIONS(6411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6413),
    [anon_sym_LBRACE] = ACTIONS(6411),
    [anon_sym_LT_AT] = ACTIONS(6411),
    [anon_sym_LT_AT_AT] = ACTIONS(6411),
    [anon_sym_CARET] = ACTIONS(6411),
    [anon_sym_or] = ACTIONS(6411),
    [anon_sym_DOT] = ACTIONS(6411),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6413),
    [anon_sym_new] = ACTIONS(6411),
    [anon_sym_return] = ACTIONS(6411),
    [anon_sym_return_BANG] = ACTIONS(6413),
    [anon_sym_yield] = ACTIONS(6411),
    [anon_sym_yield_BANG] = ACTIONS(6413),
    [anon_sym_lazy] = ACTIONS(6411),
    [anon_sym_assert] = ACTIONS(6411),
    [anon_sym_upcast] = ACTIONS(6411),
    [anon_sym_downcast] = ACTIONS(6411),
    [anon_sym_fixed] = ACTIONS(6411),
    [anon_sym_COLON_GT] = ACTIONS(6413),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6413),
    [anon_sym_for] = ACTIONS(6411),
    [anon_sym_while] = ACTIONS(6411),
    [anon_sym_while_BANG] = ACTIONS(6413),
    [anon_sym_if] = ACTIONS(6411),
    [anon_sym_fun] = ACTIONS(6411),
    [anon_sym_DASH_GT] = ACTIONS(6411),
    [anon_sym_try] = ACTIONS(6411),
    [anon_sym_match] = ACTIONS(6411),
    [anon_sym_match_BANG] = ACTIONS(6413),
    [anon_sym_function] = ACTIONS(6411),
    [anon_sym_DOT_LBRACK] = ACTIONS(6413),
    [anon_sym_use] = ACTIONS(6411),
    [anon_sym_use_BANG] = ACTIONS(6413),
    [anon_sym_do_BANG] = ACTIONS(6413),
    [anon_sym_LPAREN2] = ACTIONS(6411),
    [anon_sym_PIPE_PIPE] = ACTIONS(6411),
    [anon_sym_AMP_AMP] = ACTIONS(6411),
    [anon_sym_STAR] = ACTIONS(6411),
    [anon_sym_LT] = ACTIONS(6411),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6413),
    [aux_sym_char_token1] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6411),
    [anon_sym_DQUOTE] = ACTIONS(6411),
    [anon_sym_AT_DQUOTE] = ACTIONS(6413),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6413),
    [sym_bool] = ACTIONS(6411),
    [sym_unit] = ACTIONS(6413),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6411),
    [sym_op_identifier] = ACTIONS(6411),
    [anon_sym_PLUS] = ACTIONS(6411),
    [anon_sym_DASH] = ACTIONS(6411),
    [anon_sym_PLUS_DOT] = ACTIONS(6411),
    [anon_sym_DASH_DOT] = ACTIONS(6411),
    [anon_sym_PERCENT] = ACTIONS(6411),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6411),
    [anon_sym_TILDE] = ACTIONS(6413),
    [aux_sym_prefix_op_token1] = ACTIONS(6411),
    [aux_sym_infix_op_token1] = ACTIONS(6413),
    [aux_sym_infix_op_token2] = ACTIONS(6411),
    [aux_sym_infix_op_token3] = ACTIONS(6411),
    [anon_sym_BANG_EQ] = ACTIONS(6411),
    [anon_sym_COLON_EQ] = ACTIONS(6413),
    [anon_sym_DOLLAR] = ACTIONS(6411),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6411),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6411),
    [sym_int] = ACTIONS(6411),
    [sym_xint] = ACTIONS(6413),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6413),
    [sym__newline] = ACTIONS(6413),
    [sym__untyped_quoted_close] = ACTIONS(6413),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6413),
    [sym__tyapp_open] = ACTIONS(6413),
  },
  [STATE(4716)] = {
    [sym_block_comment] = STATE(4716),
    [sym_compiler_directive_decl] = STATE(4716),
    [sym_fsi_directive_decl] = STATE(4716),
    [sym_preproc_line] = STATE(4716),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_with] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4717)] = {
    [sym_block_comment] = STATE(4717),
    [sym_compiler_directive_decl] = STATE(4717),
    [sym_fsi_directive_decl] = STATE(4717),
    [sym_preproc_line] = STATE(4717),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4718)] = {
    [sym_block_comment] = STATE(4718),
    [sym_compiler_directive_decl] = STATE(4718),
    [sym_fsi_directive_decl] = STATE(4718),
    [sym_preproc_line] = STATE(4718),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_EQ] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_COLON] = ACTIONS(6312),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_COMMA] = ACTIONS(6310),
    [anon_sym_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_QMARK] = ACTIONS(6312),
    [anon_sym_COLON_COLON] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6312),
    [anon_sym_CARET] = ACTIONS(6312),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_DOT] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6312),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_DOT_LBRACK] = ACTIONS(6310),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_LPAREN2] = ACTIONS(6312),
    [anon_sym_PIPE_PIPE] = ACTIONS(6312),
    [anon_sym_AMP_AMP] = ACTIONS(6312),
    [anon_sym_STAR] = ACTIONS(6312),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6312),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6312),
    [anon_sym_DASH_DOT] = ACTIONS(6312),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6312),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6312),
    [aux_sym_infix_op_token1] = ACTIONS(6310),
    [aux_sym_infix_op_token2] = ACTIONS(6312),
    [aux_sym_infix_op_token3] = ACTIONS(6312),
    [anon_sym_BANG_EQ] = ACTIONS(6312),
    [anon_sym_COLON_EQ] = ACTIONS(6310),
    [anon_sym_DOLLAR] = ACTIONS(6312),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6312),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6312),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__untyped_quoted_close] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
    [sym__tyapp_open] = ACTIONS(6310),
  },
  [STATE(4719)] = {
    [sym_block_comment] = STATE(4719),
    [sym_compiler_directive_decl] = STATE(4719),
    [sym_fsi_directive_decl] = STATE(4719),
    [sym_preproc_line] = STATE(4719),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_DOT_DOT] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4720)] = {
    [sym_block_comment] = STATE(4720),
    [sym_compiler_directive_decl] = STATE(4720),
    [sym_fsi_directive_decl] = STATE(4720),
    [sym_preproc_line] = STATE(4720),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_with] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4721)] = {
    [sym_block_comment] = STATE(4721),
    [sym_compiler_directive_decl] = STATE(4721),
    [sym_fsi_directive_decl] = STATE(4721),
    [sym_preproc_line] = STATE(4721),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4722)] = {
    [sym_block_comment] = STATE(4722),
    [sym_compiler_directive_decl] = STATE(4722),
    [sym_fsi_directive_decl] = STATE(4722),
    [sym_preproc_line] = STATE(4722),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9126),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
    [sym__in] = ACTIONS(6515),
  },
  [STATE(4723)] = {
    [sym_block_comment] = STATE(4723),
    [sym_compiler_directive_decl] = STATE(4723),
    [sym_fsi_directive_decl] = STATE(4723),
    [sym_preproc_line] = STATE(4723),
    [sym_identifier] = ACTIONS(6399),
    [anon_sym_null] = ACTIONS(6399),
    [anon_sym_EQ] = ACTIONS(6399),
    [anon_sym_begin] = ACTIONS(6399),
    [anon_sym_COLON] = ACTIONS(6399),
    [anon_sym_do] = ACTIONS(6399),
    [anon_sym_let] = ACTIONS(6399),
    [anon_sym_let_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN] = ACTIONS(6399),
    [anon_sym_COMMA] = ACTIONS(6401),
    [anon_sym_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_QMARK] = ACTIONS(6399),
    [anon_sym_COLON_COLON] = ACTIONS(6401),
    [anon_sym_AMP] = ACTIONS(6399),
    [anon_sym_LBRACK] = ACTIONS(6399),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6401),
    [anon_sym_LBRACE] = ACTIONS(6399),
    [anon_sym_LT_AT] = ACTIONS(6399),
    [anon_sym_LT_AT_AT] = ACTIONS(6399),
    [anon_sym_CARET] = ACTIONS(6399),
    [anon_sym_or] = ACTIONS(6399),
    [anon_sym_DOT] = ACTIONS(6399),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6401),
    [anon_sym_new] = ACTIONS(6399),
    [anon_sym_with] = ACTIONS(6399),
    [anon_sym_return] = ACTIONS(6399),
    [anon_sym_return_BANG] = ACTIONS(6401),
    [anon_sym_yield] = ACTIONS(6399),
    [anon_sym_yield_BANG] = ACTIONS(6401),
    [anon_sym_lazy] = ACTIONS(6399),
    [anon_sym_assert] = ACTIONS(6399),
    [anon_sym_upcast] = ACTIONS(6399),
    [anon_sym_downcast] = ACTIONS(6399),
    [anon_sym_fixed] = ACTIONS(6399),
    [anon_sym_COLON_GT] = ACTIONS(6401),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6401),
    [anon_sym_for] = ACTIONS(6399),
    [anon_sym_while] = ACTIONS(6399),
    [anon_sym_while_BANG] = ACTIONS(6401),
    [anon_sym_if] = ACTIONS(6399),
    [anon_sym_fun] = ACTIONS(6399),
    [anon_sym_DASH_GT] = ACTIONS(6399),
    [anon_sym_try] = ACTIONS(6399),
    [anon_sym_match] = ACTIONS(6399),
    [anon_sym_match_BANG] = ACTIONS(6401),
    [anon_sym_function] = ACTIONS(6399),
    [anon_sym_DOT_LBRACK] = ACTIONS(6401),
    [anon_sym_use] = ACTIONS(6399),
    [anon_sym_use_BANG] = ACTIONS(6401),
    [anon_sym_do_BANG] = ACTIONS(6401),
    [anon_sym_LPAREN2] = ACTIONS(6399),
    [anon_sym_PIPE_PIPE] = ACTIONS(6399),
    [anon_sym_AMP_AMP] = ACTIONS(6399),
    [anon_sym_STAR] = ACTIONS(6399),
    [anon_sym_LT] = ACTIONS(6399),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6401),
    [aux_sym_char_token1] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6399),
    [anon_sym_DQUOTE] = ACTIONS(6399),
    [anon_sym_AT_DQUOTE] = ACTIONS(6401),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6401),
    [sym_bool] = ACTIONS(6399),
    [sym_unit] = ACTIONS(6401),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6399),
    [sym_op_identifier] = ACTIONS(6399),
    [anon_sym_PLUS] = ACTIONS(6399),
    [anon_sym_DASH] = ACTIONS(6399),
    [anon_sym_PLUS_DOT] = ACTIONS(6399),
    [anon_sym_DASH_DOT] = ACTIONS(6399),
    [anon_sym_PERCENT] = ACTIONS(6399),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6399),
    [anon_sym_TILDE] = ACTIONS(6401),
    [aux_sym_prefix_op_token1] = ACTIONS(6399),
    [aux_sym_infix_op_token1] = ACTIONS(6401),
    [aux_sym_infix_op_token2] = ACTIONS(6399),
    [aux_sym_infix_op_token3] = ACTIONS(6399),
    [anon_sym_BANG_EQ] = ACTIONS(6399),
    [anon_sym_COLON_EQ] = ACTIONS(6401),
    [anon_sym_DOLLAR] = ACTIONS(6399),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6399),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6399),
    [sym_int] = ACTIONS(6399),
    [sym_xint] = ACTIONS(6401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6401),
    [sym__newline] = ACTIONS(6401),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6401),
    [sym__tyapp_open] = ACTIONS(6401),
  },
  [STATE(4724)] = {
    [sym_block_comment] = STATE(4724),
    [sym_compiler_directive_decl] = STATE(4724),
    [sym_fsi_directive_decl] = STATE(4724),
    [sym_preproc_line] = STATE(4724),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__quoted_close] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4725)] = {
    [sym_block_comment] = STATE(4725),
    [sym_compiler_directive_decl] = STATE(4725),
    [sym_fsi_directive_decl] = STATE(4725),
    [sym_preproc_line] = STATE(4725),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_DOT_DOT] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4726)] = {
    [sym_block_comment] = STATE(4726),
    [sym_compiler_directive_decl] = STATE(4726),
    [sym_fsi_directive_decl] = STATE(4726),
    [sym_preproc_line] = STATE(4726),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__quoted_close] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4727)] = {
    [sym_block_comment] = STATE(4727),
    [sym_compiler_directive_decl] = STATE(4727),
    [sym_fsi_directive_decl] = STATE(4727),
    [sym_preproc_line] = STATE(4727),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9128),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4728)] = {
    [sym_block_comment] = STATE(4728),
    [sym_compiler_directive_decl] = STATE(4728),
    [sym_fsi_directive_decl] = STATE(4728),
    [sym_preproc_line] = STATE(4728),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_as] = ACTIONS(6784),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9130),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_DOT_DOT] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__dedent] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(4729)] = {
    [sym_block_comment] = STATE(4729),
    [sym_compiler_directive_decl] = STATE(4729),
    [sym_fsi_directive_decl] = STATE(4729),
    [sym_preproc_line] = STATE(4729),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_as] = ACTIONS(6790),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_DOT_DOT] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__dedent] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(4730)] = {
    [sym_block_comment] = STATE(4730),
    [sym_compiler_directive_decl] = STATE(4730),
    [sym_fsi_directive_decl] = STATE(4730),
    [sym_preproc_line] = STATE(4730),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_as] = ACTIONS(6794),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_DOT_DOT] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__dedent] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(4731)] = {
    [sym_block_comment] = STATE(4731),
    [sym_compiler_directive_decl] = STATE(4731),
    [sym_fsi_directive_decl] = STATE(4731),
    [sym_preproc_line] = STATE(4731),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__untyped_quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4732)] = {
    [sym_elif_expression] = STATE(5190),
    [sym_block_comment] = STATE(4732),
    [sym_compiler_directive_decl] = STATE(4732),
    [sym_fsi_directive_decl] = STATE(4732),
    [sym_preproc_line] = STATE(4732),
    [aux_sym_if_expression_repeat1] = STATE(4732),
    [sym_identifier] = ACTIONS(6496),
    [anon_sym_null] = ACTIONS(6496),
    [anon_sym_EQ] = ACTIONS(6496),
    [anon_sym_begin] = ACTIONS(6496),
    [anon_sym_COLON] = ACTIONS(6496),
    [anon_sym_do] = ACTIONS(6496),
    [anon_sym_let] = ACTIONS(6496),
    [anon_sym_let_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN] = ACTIONS(6496),
    [anon_sym_COMMA] = ACTIONS(6494),
    [anon_sym_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_QMARK] = ACTIONS(6496),
    [anon_sym_COLON_COLON] = ACTIONS(6494),
    [anon_sym_AMP] = ACTIONS(6496),
    [anon_sym_LBRACK] = ACTIONS(6496),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6494),
    [anon_sym_LBRACE] = ACTIONS(6496),
    [anon_sym_LT_AT] = ACTIONS(6496),
    [anon_sym_LT_AT_AT] = ACTIONS(6496),
    [anon_sym_CARET] = ACTIONS(6496),
    [anon_sym_or] = ACTIONS(6496),
    [anon_sym_DOT] = ACTIONS(6496),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6494),
    [anon_sym_new] = ACTIONS(6496),
    [anon_sym_with] = ACTIONS(6496),
    [anon_sym_return] = ACTIONS(6496),
    [anon_sym_return_BANG] = ACTIONS(6494),
    [anon_sym_yield] = ACTIONS(6496),
    [anon_sym_yield_BANG] = ACTIONS(6494),
    [anon_sym_lazy] = ACTIONS(6496),
    [anon_sym_assert] = ACTIONS(6496),
    [anon_sym_upcast] = ACTIONS(6496),
    [anon_sym_downcast] = ACTIONS(6496),
    [anon_sym_fixed] = ACTIONS(6496),
    [anon_sym_COLON_GT] = ACTIONS(6494),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6494),
    [anon_sym_for] = ACTIONS(6496),
    [anon_sym_while] = ACTIONS(6496),
    [anon_sym_while_BANG] = ACTIONS(6494),
    [anon_sym_else] = ACTIONS(6496),
    [anon_sym_elif] = ACTIONS(9132),
    [anon_sym_if] = ACTIONS(6496),
    [anon_sym_fun] = ACTIONS(6496),
    [anon_sym_try] = ACTIONS(6496),
    [anon_sym_match] = ACTIONS(6496),
    [anon_sym_match_BANG] = ACTIONS(6494),
    [anon_sym_function] = ACTIONS(6496),
    [anon_sym_DOT_LBRACK] = ACTIONS(6494),
    [anon_sym_use] = ACTIONS(6496),
    [anon_sym_use_BANG] = ACTIONS(6494),
    [anon_sym_do_BANG] = ACTIONS(6494),
    [anon_sym_LPAREN2] = ACTIONS(6496),
    [anon_sym_PIPE_PIPE] = ACTIONS(6496),
    [anon_sym_AMP_AMP] = ACTIONS(6496),
    [aux_sym_char_token1] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6496),
    [anon_sym_DQUOTE] = ACTIONS(6496),
    [anon_sym_AT_DQUOTE] = ACTIONS(6494),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6494),
    [sym_bool] = ACTIONS(6496),
    [sym_unit] = ACTIONS(6494),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6496),
    [sym_op_identifier] = ACTIONS(6496),
    [anon_sym_PLUS] = ACTIONS(6496),
    [anon_sym_DASH] = ACTIONS(6496),
    [anon_sym_PLUS_DOT] = ACTIONS(6496),
    [anon_sym_DASH_DOT] = ACTIONS(6496),
    [anon_sym_PERCENT] = ACTIONS(6496),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6496),
    [anon_sym_TILDE] = ACTIONS(6494),
    [aux_sym_prefix_op_token1] = ACTIONS(6496),
    [aux_sym_infix_op_token1] = ACTIONS(6494),
    [aux_sym_infix_op_token2] = ACTIONS(6496),
    [aux_sym_infix_op_token3] = ACTIONS(6496),
    [anon_sym_BANG_EQ] = ACTIONS(6496),
    [anon_sym_COLON_EQ] = ACTIONS(6494),
    [anon_sym_DOLLAR] = ACTIONS(6496),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6496),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6496),
    [sym_int] = ACTIONS(6496),
    [sym_xint] = ACTIONS(6494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6494),
    [sym__newline] = ACTIONS(6494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6494),
    [sym__tyapp_open] = ACTIONS(6494),
  },
  [STATE(4733)] = {
    [sym_block_comment] = STATE(4733),
    [sym_compiler_directive_decl] = STATE(4733),
    [sym_fsi_directive_decl] = STATE(4733),
    [sym_preproc_line] = STATE(4733),
    [sym_identifier] = ACTIONS(6492),
    [anon_sym_null] = ACTIONS(6492),
    [anon_sym_EQ] = ACTIONS(6492),
    [anon_sym_begin] = ACTIONS(6492),
    [anon_sym_COLON] = ACTIONS(6492),
    [anon_sym_do] = ACTIONS(6492),
    [anon_sym_let] = ACTIONS(6492),
    [anon_sym_let_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN] = ACTIONS(6492),
    [anon_sym_COMMA] = ACTIONS(6490),
    [anon_sym_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_QMARK] = ACTIONS(6492),
    [anon_sym_COLON_COLON] = ACTIONS(6490),
    [anon_sym_AMP] = ACTIONS(6492),
    [anon_sym_LBRACK] = ACTIONS(6492),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6490),
    [anon_sym_LBRACE] = ACTIONS(6492),
    [anon_sym_LT_AT] = ACTIONS(6492),
    [anon_sym_LT_AT_AT] = ACTIONS(6492),
    [anon_sym_CARET] = ACTIONS(6492),
    [anon_sym_or] = ACTIONS(6492),
    [anon_sym_DOT] = ACTIONS(6492),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6490),
    [anon_sym_new] = ACTIONS(6492),
    [anon_sym_with] = ACTIONS(6492),
    [anon_sym_return] = ACTIONS(6492),
    [anon_sym_return_BANG] = ACTIONS(6490),
    [anon_sym_yield] = ACTIONS(6492),
    [anon_sym_yield_BANG] = ACTIONS(6490),
    [anon_sym_lazy] = ACTIONS(6492),
    [anon_sym_assert] = ACTIONS(6492),
    [anon_sym_upcast] = ACTIONS(6492),
    [anon_sym_downcast] = ACTIONS(6492),
    [anon_sym_fixed] = ACTIONS(6492),
    [anon_sym_COLON_GT] = ACTIONS(6490),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6490),
    [anon_sym_for] = ACTIONS(6492),
    [anon_sym_while] = ACTIONS(6492),
    [anon_sym_while_BANG] = ACTIONS(6490),
    [anon_sym_if] = ACTIONS(6492),
    [anon_sym_fun] = ACTIONS(6492),
    [anon_sym_DASH_GT] = ACTIONS(6492),
    [anon_sym_try] = ACTIONS(6492),
    [anon_sym_match] = ACTIONS(6492),
    [anon_sym_match_BANG] = ACTIONS(6490),
    [anon_sym_function] = ACTIONS(6492),
    [anon_sym_DOT_LBRACK] = ACTIONS(6490),
    [anon_sym_use] = ACTIONS(6492),
    [anon_sym_use_BANG] = ACTIONS(6490),
    [anon_sym_do_BANG] = ACTIONS(6490),
    [anon_sym_LPAREN2] = ACTIONS(6492),
    [anon_sym_PIPE_PIPE] = ACTIONS(6492),
    [anon_sym_AMP_AMP] = ACTIONS(6492),
    [anon_sym_STAR] = ACTIONS(6492),
    [anon_sym_LT] = ACTIONS(6492),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6490),
    [aux_sym_char_token1] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6492),
    [anon_sym_DQUOTE] = ACTIONS(6492),
    [anon_sym_AT_DQUOTE] = ACTIONS(6490),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6490),
    [sym_bool] = ACTIONS(6492),
    [sym_unit] = ACTIONS(6490),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6492),
    [sym_op_identifier] = ACTIONS(6492),
    [anon_sym_PLUS] = ACTIONS(6492),
    [anon_sym_DASH] = ACTIONS(6492),
    [anon_sym_PLUS_DOT] = ACTIONS(6492),
    [anon_sym_DASH_DOT] = ACTIONS(6492),
    [anon_sym_PERCENT] = ACTIONS(6492),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6492),
    [anon_sym_TILDE] = ACTIONS(6490),
    [aux_sym_prefix_op_token1] = ACTIONS(6492),
    [aux_sym_infix_op_token1] = ACTIONS(6490),
    [aux_sym_infix_op_token2] = ACTIONS(6492),
    [aux_sym_infix_op_token3] = ACTIONS(6492),
    [anon_sym_BANG_EQ] = ACTIONS(6492),
    [anon_sym_COLON_EQ] = ACTIONS(6490),
    [anon_sym_DOLLAR] = ACTIONS(6492),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6492),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6492),
    [sym_int] = ACTIONS(6492),
    [sym_xint] = ACTIONS(6490),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6490),
    [sym__newline] = ACTIONS(6490),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6490),
    [sym__tyapp_open] = ACTIONS(6490),
  },
  [STATE(4734)] = {
    [sym_block_comment] = STATE(4734),
    [sym_compiler_directive_decl] = STATE(4734),
    [sym_fsi_directive_decl] = STATE(4734),
    [sym_preproc_line] = STATE(4734),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4735)] = {
    [sym_block_comment] = STATE(4735),
    [sym_compiler_directive_decl] = STATE(4735),
    [sym_fsi_directive_decl] = STATE(4735),
    [sym_preproc_line] = STATE(4735),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_STAR] = ACTIONS(6295),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4736)] = {
    [sym_block_comment] = STATE(4736),
    [sym_compiler_directive_decl] = STATE(4736),
    [sym_fsi_directive_decl] = STATE(4736),
    [sym_preproc_line] = STATE(4736),
    [sym_identifier] = ACTIONS(6427),
    [anon_sym_null] = ACTIONS(6427),
    [anon_sym_EQ] = ACTIONS(6427),
    [anon_sym_begin] = ACTIONS(6427),
    [anon_sym_COLON] = ACTIONS(6427),
    [anon_sym_do] = ACTIONS(6427),
    [anon_sym_let] = ACTIONS(6427),
    [anon_sym_let_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN] = ACTIONS(6427),
    [anon_sym_COMMA] = ACTIONS(6429),
    [anon_sym_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_QMARK] = ACTIONS(6427),
    [anon_sym_COLON_COLON] = ACTIONS(6429),
    [anon_sym_AMP] = ACTIONS(6427),
    [anon_sym_LBRACK] = ACTIONS(6427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6429),
    [anon_sym_LBRACE] = ACTIONS(6427),
    [anon_sym_LT_AT] = ACTIONS(6427),
    [anon_sym_LT_AT_AT] = ACTIONS(6427),
    [anon_sym_CARET] = ACTIONS(6427),
    [anon_sym_or] = ACTIONS(6427),
    [anon_sym_DOT] = ACTIONS(6427),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6429),
    [anon_sym_new] = ACTIONS(6427),
    [anon_sym_with] = ACTIONS(6427),
    [anon_sym_return] = ACTIONS(6427),
    [anon_sym_return_BANG] = ACTIONS(6429),
    [anon_sym_yield] = ACTIONS(6427),
    [anon_sym_yield_BANG] = ACTIONS(6429),
    [anon_sym_lazy] = ACTIONS(6427),
    [anon_sym_assert] = ACTIONS(6427),
    [anon_sym_upcast] = ACTIONS(6427),
    [anon_sym_downcast] = ACTIONS(6427),
    [anon_sym_fixed] = ACTIONS(6427),
    [anon_sym_COLON_GT] = ACTIONS(6429),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6429),
    [anon_sym_for] = ACTIONS(6427),
    [anon_sym_while] = ACTIONS(6427),
    [anon_sym_while_BANG] = ACTIONS(6429),
    [anon_sym_if] = ACTIONS(6427),
    [anon_sym_fun] = ACTIONS(6427),
    [anon_sym_DASH_GT] = ACTIONS(6427),
    [anon_sym_try] = ACTIONS(6427),
    [anon_sym_match] = ACTIONS(6427),
    [anon_sym_match_BANG] = ACTIONS(6429),
    [anon_sym_function] = ACTIONS(6427),
    [anon_sym_DOT_LBRACK] = ACTIONS(6429),
    [anon_sym_use] = ACTIONS(6427),
    [anon_sym_use_BANG] = ACTIONS(6429),
    [anon_sym_do_BANG] = ACTIONS(6429),
    [anon_sym_LPAREN2] = ACTIONS(6427),
    [anon_sym_PIPE_PIPE] = ACTIONS(6427),
    [anon_sym_AMP_AMP] = ACTIONS(6427),
    [anon_sym_STAR] = ACTIONS(6427),
    [anon_sym_LT] = ACTIONS(6427),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6429),
    [aux_sym_char_token1] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6427),
    [anon_sym_DQUOTE] = ACTIONS(6427),
    [anon_sym_AT_DQUOTE] = ACTIONS(6429),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6429),
    [sym_bool] = ACTIONS(6427),
    [sym_unit] = ACTIONS(6429),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6427),
    [sym_op_identifier] = ACTIONS(6427),
    [anon_sym_PLUS] = ACTIONS(6427),
    [anon_sym_DASH] = ACTIONS(6427),
    [anon_sym_PLUS_DOT] = ACTIONS(6427),
    [anon_sym_DASH_DOT] = ACTIONS(6427),
    [anon_sym_PERCENT] = ACTIONS(6427),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6427),
    [anon_sym_TILDE] = ACTIONS(6429),
    [aux_sym_prefix_op_token1] = ACTIONS(6427),
    [aux_sym_infix_op_token1] = ACTIONS(6429),
    [aux_sym_infix_op_token2] = ACTIONS(6427),
    [aux_sym_infix_op_token3] = ACTIONS(6427),
    [anon_sym_BANG_EQ] = ACTIONS(6427),
    [anon_sym_COLON_EQ] = ACTIONS(6429),
    [anon_sym_DOLLAR] = ACTIONS(6427),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6427),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6427),
    [sym_int] = ACTIONS(6427),
    [sym_xint] = ACTIONS(6429),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6429),
    [sym__newline] = ACTIONS(6429),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6429),
    [sym__tyapp_open] = ACTIONS(6429),
  },
  [STATE(4737)] = {
    [sym_block_comment] = STATE(4737),
    [sym_compiler_directive_decl] = STATE(4737),
    [sym_fsi_directive_decl] = STATE(4737),
    [sym_preproc_line] = STATE(4737),
    [sym_identifier] = ACTIONS(6395),
    [anon_sym_null] = ACTIONS(6395),
    [anon_sym_EQ] = ACTIONS(6395),
    [anon_sym_begin] = ACTIONS(6395),
    [anon_sym_COLON] = ACTIONS(6395),
    [anon_sym_do] = ACTIONS(6395),
    [anon_sym_let] = ACTIONS(6395),
    [anon_sym_let_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN] = ACTIONS(6395),
    [anon_sym_COMMA] = ACTIONS(6397),
    [anon_sym_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_QMARK] = ACTIONS(6395),
    [anon_sym_COLON_COLON] = ACTIONS(6397),
    [anon_sym_AMP] = ACTIONS(6395),
    [anon_sym_LBRACK] = ACTIONS(6395),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6397),
    [anon_sym_LBRACE] = ACTIONS(6395),
    [anon_sym_LT_AT] = ACTIONS(6395),
    [anon_sym_LT_AT_AT] = ACTIONS(6395),
    [anon_sym_CARET] = ACTIONS(6395),
    [anon_sym_or] = ACTIONS(6395),
    [anon_sym_DOT] = ACTIONS(6395),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6397),
    [anon_sym_new] = ACTIONS(6395),
    [anon_sym_return] = ACTIONS(6395),
    [anon_sym_return_BANG] = ACTIONS(6397),
    [anon_sym_yield] = ACTIONS(6395),
    [anon_sym_yield_BANG] = ACTIONS(6397),
    [anon_sym_lazy] = ACTIONS(6395),
    [anon_sym_assert] = ACTIONS(6395),
    [anon_sym_upcast] = ACTIONS(6395),
    [anon_sym_downcast] = ACTIONS(6395),
    [anon_sym_fixed] = ACTIONS(6395),
    [anon_sym_COLON_GT] = ACTIONS(6397),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6397),
    [anon_sym_for] = ACTIONS(6395),
    [anon_sym_while] = ACTIONS(6395),
    [anon_sym_while_BANG] = ACTIONS(6397),
    [anon_sym_if] = ACTIONS(6395),
    [anon_sym_fun] = ACTIONS(6395),
    [anon_sym_DASH_GT] = ACTIONS(6395),
    [anon_sym_try] = ACTIONS(6395),
    [anon_sym_match] = ACTIONS(6395),
    [anon_sym_match_BANG] = ACTIONS(6397),
    [anon_sym_function] = ACTIONS(6395),
    [anon_sym_DOT_LBRACK] = ACTIONS(6397),
    [anon_sym_use] = ACTIONS(6395),
    [anon_sym_use_BANG] = ACTIONS(6397),
    [anon_sym_do_BANG] = ACTIONS(6397),
    [anon_sym_LPAREN2] = ACTIONS(6395),
    [anon_sym_PIPE_PIPE] = ACTIONS(6395),
    [anon_sym_AMP_AMP] = ACTIONS(6395),
    [anon_sym_STAR] = ACTIONS(6395),
    [anon_sym_LT] = ACTIONS(6395),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6397),
    [aux_sym_char_token1] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6395),
    [anon_sym_DQUOTE] = ACTIONS(6395),
    [anon_sym_AT_DQUOTE] = ACTIONS(6397),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6397),
    [sym_bool] = ACTIONS(6395),
    [sym_unit] = ACTIONS(6397),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6395),
    [sym_op_identifier] = ACTIONS(6395),
    [anon_sym_PLUS] = ACTIONS(6395),
    [anon_sym_DASH] = ACTIONS(6395),
    [anon_sym_PLUS_DOT] = ACTIONS(6395),
    [anon_sym_DASH_DOT] = ACTIONS(6395),
    [anon_sym_PERCENT] = ACTIONS(6395),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6395),
    [anon_sym_TILDE] = ACTIONS(6397),
    [aux_sym_prefix_op_token1] = ACTIONS(6395),
    [aux_sym_infix_op_token1] = ACTIONS(6397),
    [aux_sym_infix_op_token2] = ACTIONS(6395),
    [aux_sym_infix_op_token3] = ACTIONS(6395),
    [anon_sym_BANG_EQ] = ACTIONS(6395),
    [anon_sym_COLON_EQ] = ACTIONS(6397),
    [anon_sym_DOLLAR] = ACTIONS(6395),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6395),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6395),
    [sym_int] = ACTIONS(6395),
    [sym_xint] = ACTIONS(6397),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6397),
    [anon_sym_POUNDendif] = ACTIONS(6397),
    [sym__newline] = ACTIONS(6397),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6397),
    [sym__tyapp_open] = ACTIONS(6397),
  },
  [STATE(4738)] = {
    [sym_block_comment] = STATE(4738),
    [sym_compiler_directive_decl] = STATE(4738),
    [sym_fsi_directive_decl] = STATE(4738),
    [sym_preproc_line] = STATE(4738),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_as] = ACTIONS(6597),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9052),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4739)] = {
    [sym_block_comment] = STATE(4739),
    [sym_compiler_directive_decl] = STATE(4739),
    [sym_fsi_directive_decl] = STATE(4739),
    [sym_preproc_line] = STATE(4739),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_EQ] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_COLON] = ACTIONS(6369),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_COMMA] = ACTIONS(6371),
    [anon_sym_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_QMARK] = ACTIONS(6369),
    [anon_sym_COLON_COLON] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6369),
    [anon_sym_CARET] = ACTIONS(6369),
    [anon_sym_or] = ACTIONS(6369),
    [anon_sym_DOT] = ACTIONS(6369),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(6371),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6371),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6369),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_DOT_LBRACK] = ACTIONS(6371),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_DOT_DOT] = ACTIONS(6371),
    [anon_sym_LPAREN2] = ACTIONS(6369),
    [anon_sym_PIPE_PIPE] = ACTIONS(6369),
    [anon_sym_AMP_AMP] = ACTIONS(6369),
    [anon_sym_STAR] = ACTIONS(6369),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6369),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6369),
    [anon_sym_DASH_DOT] = ACTIONS(6369),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6369),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6369),
    [aux_sym_infix_op_token1] = ACTIONS(6371),
    [aux_sym_infix_op_token2] = ACTIONS(6369),
    [aux_sym_infix_op_token3] = ACTIONS(6369),
    [anon_sym_BANG_EQ] = ACTIONS(6369),
    [anon_sym_COLON_EQ] = ACTIONS(6371),
    [anon_sym_DOLLAR] = ACTIONS(6369),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6369),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6369),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
    [sym__tyapp_open] = ACTIONS(6371),
  },
  [STATE(4740)] = {
    [sym_block_comment] = STATE(4740),
    [sym_compiler_directive_decl] = STATE(4740),
    [sym_fsi_directive_decl] = STATE(4740),
    [sym_preproc_line] = STATE(4740),
    [sym_identifier] = ACTIONS(6403),
    [anon_sym_null] = ACTIONS(6403),
    [anon_sym_EQ] = ACTIONS(6403),
    [anon_sym_begin] = ACTIONS(6403),
    [anon_sym_COLON] = ACTIONS(6403),
    [anon_sym_do] = ACTIONS(6403),
    [anon_sym_let] = ACTIONS(6403),
    [anon_sym_let_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN] = ACTIONS(6403),
    [anon_sym_COMMA] = ACTIONS(6405),
    [anon_sym_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_QMARK] = ACTIONS(6403),
    [anon_sym_COLON_COLON] = ACTIONS(6405),
    [anon_sym_AMP] = ACTIONS(6403),
    [anon_sym_LBRACK] = ACTIONS(6403),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6405),
    [anon_sym_LBRACE] = ACTIONS(6403),
    [anon_sym_LT_AT] = ACTIONS(6403),
    [anon_sym_LT_AT_AT] = ACTIONS(6403),
    [anon_sym_CARET] = ACTIONS(6403),
    [anon_sym_or] = ACTIONS(6403),
    [anon_sym_DOT] = ACTIONS(6403),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6405),
    [anon_sym_new] = ACTIONS(6403),
    [anon_sym_with] = ACTIONS(6403),
    [anon_sym_return] = ACTIONS(6403),
    [anon_sym_return_BANG] = ACTIONS(6405),
    [anon_sym_yield] = ACTIONS(6403),
    [anon_sym_yield_BANG] = ACTIONS(6405),
    [anon_sym_lazy] = ACTIONS(6403),
    [anon_sym_assert] = ACTIONS(6403),
    [anon_sym_upcast] = ACTIONS(6403),
    [anon_sym_downcast] = ACTIONS(6403),
    [anon_sym_fixed] = ACTIONS(6403),
    [anon_sym_COLON_GT] = ACTIONS(6405),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6405),
    [anon_sym_for] = ACTIONS(6403),
    [anon_sym_while] = ACTIONS(6403),
    [anon_sym_while_BANG] = ACTIONS(6405),
    [anon_sym_if] = ACTIONS(6403),
    [anon_sym_fun] = ACTIONS(6403),
    [anon_sym_DASH_GT] = ACTIONS(6403),
    [anon_sym_try] = ACTIONS(6403),
    [anon_sym_match] = ACTIONS(6403),
    [anon_sym_match_BANG] = ACTIONS(6405),
    [anon_sym_function] = ACTIONS(6403),
    [anon_sym_DOT_LBRACK] = ACTIONS(6405),
    [anon_sym_use] = ACTIONS(6403),
    [anon_sym_use_BANG] = ACTIONS(6405),
    [anon_sym_do_BANG] = ACTIONS(6405),
    [anon_sym_LPAREN2] = ACTIONS(6403),
    [anon_sym_PIPE_PIPE] = ACTIONS(6403),
    [anon_sym_AMP_AMP] = ACTIONS(6403),
    [anon_sym_STAR] = ACTIONS(6403),
    [anon_sym_LT] = ACTIONS(6403),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6405),
    [aux_sym_char_token1] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6403),
    [anon_sym_DQUOTE] = ACTIONS(6403),
    [anon_sym_AT_DQUOTE] = ACTIONS(6405),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6405),
    [sym_bool] = ACTIONS(6403),
    [sym_unit] = ACTIONS(6405),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6403),
    [sym_op_identifier] = ACTIONS(6403),
    [anon_sym_PLUS] = ACTIONS(6403),
    [anon_sym_DASH] = ACTIONS(6403),
    [anon_sym_PLUS_DOT] = ACTIONS(6403),
    [anon_sym_DASH_DOT] = ACTIONS(6403),
    [anon_sym_PERCENT] = ACTIONS(6403),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6403),
    [anon_sym_TILDE] = ACTIONS(6405),
    [aux_sym_prefix_op_token1] = ACTIONS(6403),
    [aux_sym_infix_op_token1] = ACTIONS(6405),
    [aux_sym_infix_op_token2] = ACTIONS(6403),
    [aux_sym_infix_op_token3] = ACTIONS(6403),
    [anon_sym_BANG_EQ] = ACTIONS(6403),
    [anon_sym_COLON_EQ] = ACTIONS(6405),
    [anon_sym_DOLLAR] = ACTIONS(6403),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6403),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6403),
    [sym_int] = ACTIONS(6403),
    [sym_xint] = ACTIONS(6405),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6405),
    [sym__newline] = ACTIONS(6405),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6405),
    [sym__tyapp_open] = ACTIONS(6405),
  },
  [STATE(4741)] = {
    [sym_block_comment] = STATE(4741),
    [sym_compiler_directive_decl] = STATE(4741),
    [sym_fsi_directive_decl] = STATE(4741),
    [sym_preproc_line] = STATE(4741),
    [sym_identifier] = ACTIONS(6373),
    [anon_sym_null] = ACTIONS(6373),
    [anon_sym_EQ] = ACTIONS(6373),
    [anon_sym_begin] = ACTIONS(6373),
    [anon_sym_COLON] = ACTIONS(6373),
    [anon_sym_do] = ACTIONS(6373),
    [anon_sym_let] = ACTIONS(6373),
    [anon_sym_let_BANG] = ACTIONS(6375),
    [anon_sym_LPAREN] = ACTIONS(6373),
    [anon_sym_COMMA] = ACTIONS(6375),
    [anon_sym_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_QMARK] = ACTIONS(6373),
    [anon_sym_COLON_COLON] = ACTIONS(6375),
    [anon_sym_AMP] = ACTIONS(6373),
    [anon_sym_LBRACK] = ACTIONS(6373),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6375),
    [anon_sym_LBRACE] = ACTIONS(6373),
    [anon_sym_LT_AT] = ACTIONS(6373),
    [anon_sym_LT_AT_AT] = ACTIONS(6373),
    [anon_sym_CARET] = ACTIONS(6373),
    [anon_sym_or] = ACTIONS(6373),
    [anon_sym_DOT] = ACTIONS(6373),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6375),
    [anon_sym_new] = ACTIONS(6373),
    [anon_sym_return] = ACTIONS(6373),
    [anon_sym_return_BANG] = ACTIONS(6375),
    [anon_sym_yield] = ACTIONS(6373),
    [anon_sym_yield_BANG] = ACTIONS(6375),
    [anon_sym_lazy] = ACTIONS(6373),
    [anon_sym_assert] = ACTIONS(6373),
    [anon_sym_upcast] = ACTIONS(6373),
    [anon_sym_downcast] = ACTIONS(6373),
    [anon_sym_fixed] = ACTIONS(6373),
    [anon_sym_COLON_GT] = ACTIONS(6375),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6375),
    [anon_sym_for] = ACTIONS(6373),
    [anon_sym_while] = ACTIONS(6373),
    [anon_sym_while_BANG] = ACTIONS(6375),
    [anon_sym_if] = ACTIONS(6373),
    [anon_sym_fun] = ACTIONS(6373),
    [anon_sym_DASH_GT] = ACTIONS(6373),
    [anon_sym_try] = ACTIONS(6373),
    [anon_sym_match] = ACTIONS(6373),
    [anon_sym_match_BANG] = ACTIONS(6375),
    [anon_sym_function] = ACTIONS(6373),
    [anon_sym_DOT_LBRACK] = ACTIONS(6375),
    [anon_sym_use] = ACTIONS(6373),
    [anon_sym_use_BANG] = ACTIONS(6375),
    [anon_sym_do_BANG] = ACTIONS(6375),
    [anon_sym_DOT_DOT] = ACTIONS(6375),
    [anon_sym_LPAREN2] = ACTIONS(6373),
    [anon_sym_PIPE_PIPE] = ACTIONS(6373),
    [anon_sym_AMP_AMP] = ACTIONS(6373),
    [anon_sym_STAR] = ACTIONS(6373),
    [anon_sym_LT] = ACTIONS(9135),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6375),
    [aux_sym_char_token1] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6373),
    [anon_sym_DQUOTE] = ACTIONS(6373),
    [anon_sym_AT_DQUOTE] = ACTIONS(6375),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6375),
    [sym_bool] = ACTIONS(6373),
    [sym_unit] = ACTIONS(6375),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6373),
    [sym_op_identifier] = ACTIONS(6373),
    [anon_sym_PLUS] = ACTIONS(6373),
    [anon_sym_DASH] = ACTIONS(6373),
    [anon_sym_PLUS_DOT] = ACTIONS(6373),
    [anon_sym_DASH_DOT] = ACTIONS(6373),
    [anon_sym_PERCENT] = ACTIONS(6373),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6373),
    [anon_sym_TILDE] = ACTIONS(6375),
    [aux_sym_prefix_op_token1] = ACTIONS(6373),
    [aux_sym_infix_op_token1] = ACTIONS(6375),
    [aux_sym_infix_op_token2] = ACTIONS(6373),
    [aux_sym_infix_op_token3] = ACTIONS(6373),
    [anon_sym_BANG_EQ] = ACTIONS(6373),
    [anon_sym_COLON_EQ] = ACTIONS(6375),
    [anon_sym_DOLLAR] = ACTIONS(6373),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6373),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6373),
    [sym_int] = ACTIONS(6373),
    [sym_xint] = ACTIONS(6375),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6375),
    [sym__newline] = ACTIONS(6375),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6375),
    [sym__tyapp_open] = ACTIONS(6375),
  },
  [STATE(4742)] = {
    [sym_block_comment] = STATE(4742),
    [sym_compiler_directive_decl] = STATE(4742),
    [sym_fsi_directive_decl] = STATE(4742),
    [sym_preproc_line] = STATE(4742),
    [sym_identifier] = ACTIONS(6391),
    [anon_sym_null] = ACTIONS(6391),
    [anon_sym_EQ] = ACTIONS(6391),
    [anon_sym_begin] = ACTIONS(6391),
    [anon_sym_COLON] = ACTIONS(6391),
    [anon_sym_do] = ACTIONS(6391),
    [anon_sym_let] = ACTIONS(6391),
    [anon_sym_let_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN] = ACTIONS(6391),
    [anon_sym_COMMA] = ACTIONS(6393),
    [anon_sym_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_QMARK] = ACTIONS(6391),
    [anon_sym_COLON_COLON] = ACTIONS(6393),
    [anon_sym_AMP] = ACTIONS(6391),
    [anon_sym_LBRACK] = ACTIONS(6391),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6393),
    [anon_sym_LBRACE] = ACTIONS(6391),
    [anon_sym_LT_AT] = ACTIONS(6391),
    [anon_sym_LT_AT_AT] = ACTIONS(6391),
    [anon_sym_CARET] = ACTIONS(6391),
    [anon_sym_or] = ACTIONS(6391),
    [anon_sym_DOT] = ACTIONS(6391),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6393),
    [anon_sym_new] = ACTIONS(6391),
    [anon_sym_return] = ACTIONS(6391),
    [anon_sym_return_BANG] = ACTIONS(6393),
    [anon_sym_yield] = ACTIONS(6391),
    [anon_sym_yield_BANG] = ACTIONS(6393),
    [anon_sym_lazy] = ACTIONS(6391),
    [anon_sym_assert] = ACTIONS(6391),
    [anon_sym_upcast] = ACTIONS(6391),
    [anon_sym_downcast] = ACTIONS(6391),
    [anon_sym_fixed] = ACTIONS(6391),
    [anon_sym_COLON_GT] = ACTIONS(6393),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6393),
    [anon_sym_for] = ACTIONS(6391),
    [anon_sym_while] = ACTIONS(6391),
    [anon_sym_while_BANG] = ACTIONS(6393),
    [anon_sym_if] = ACTIONS(6391),
    [anon_sym_fun] = ACTIONS(6391),
    [anon_sym_DASH_GT] = ACTIONS(6391),
    [anon_sym_try] = ACTIONS(6391),
    [anon_sym_match] = ACTIONS(6391),
    [anon_sym_match_BANG] = ACTIONS(6393),
    [anon_sym_function] = ACTIONS(6391),
    [anon_sym_DOT_LBRACK] = ACTIONS(6393),
    [anon_sym_use] = ACTIONS(6391),
    [anon_sym_use_BANG] = ACTIONS(6393),
    [anon_sym_do_BANG] = ACTIONS(6393),
    [anon_sym_LPAREN2] = ACTIONS(6391),
    [anon_sym_PIPE_PIPE] = ACTIONS(6391),
    [anon_sym_AMP_AMP] = ACTIONS(6391),
    [anon_sym_STAR] = ACTIONS(6391),
    [anon_sym_LT] = ACTIONS(6391),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6393),
    [aux_sym_char_token1] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6391),
    [anon_sym_DQUOTE] = ACTIONS(6391),
    [anon_sym_AT_DQUOTE] = ACTIONS(6393),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6393),
    [sym_bool] = ACTIONS(6391),
    [sym_unit] = ACTIONS(6393),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6391),
    [sym_op_identifier] = ACTIONS(6391),
    [anon_sym_PLUS] = ACTIONS(6391),
    [anon_sym_DASH] = ACTIONS(6391),
    [anon_sym_PLUS_DOT] = ACTIONS(6391),
    [anon_sym_DASH_DOT] = ACTIONS(6391),
    [anon_sym_PERCENT] = ACTIONS(6391),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6391),
    [anon_sym_TILDE] = ACTIONS(6393),
    [aux_sym_prefix_op_token1] = ACTIONS(6391),
    [aux_sym_infix_op_token1] = ACTIONS(6393),
    [aux_sym_infix_op_token2] = ACTIONS(6391),
    [aux_sym_infix_op_token3] = ACTIONS(6391),
    [anon_sym_BANG_EQ] = ACTIONS(6391),
    [anon_sym_COLON_EQ] = ACTIONS(6393),
    [anon_sym_DOLLAR] = ACTIONS(6391),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6391),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6391),
    [sym_int] = ACTIONS(6391),
    [sym_xint] = ACTIONS(6393),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6393),
    [anon_sym_POUNDendif] = ACTIONS(6393),
    [sym__newline] = ACTIONS(6393),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6393),
    [sym__tyapp_open] = ACTIONS(6393),
  },
  [STATE(4743)] = {
    [sym_block_comment] = STATE(4743),
    [sym_compiler_directive_decl] = STATE(4743),
    [sym_fsi_directive_decl] = STATE(4743),
    [sym_preproc_line] = STATE(4743),
    [sym_identifier] = ACTIONS(6389),
    [anon_sym_null] = ACTIONS(6389),
    [anon_sym_EQ] = ACTIONS(6389),
    [anon_sym_begin] = ACTIONS(6389),
    [anon_sym_COLON] = ACTIONS(6389),
    [anon_sym_do] = ACTIONS(6389),
    [anon_sym_let] = ACTIONS(6389),
    [anon_sym_let_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN] = ACTIONS(6389),
    [anon_sym_COMMA] = ACTIONS(6387),
    [anon_sym_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_QMARK] = ACTIONS(6389),
    [anon_sym_COLON_COLON] = ACTIONS(6387),
    [anon_sym_AMP] = ACTIONS(6389),
    [anon_sym_LBRACK] = ACTIONS(6389),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6387),
    [anon_sym_LBRACE] = ACTIONS(6389),
    [anon_sym_LT_AT] = ACTIONS(6389),
    [anon_sym_LT_AT_AT] = ACTIONS(6389),
    [anon_sym_CARET] = ACTIONS(6389),
    [anon_sym_or] = ACTIONS(6389),
    [anon_sym_DOT] = ACTIONS(6389),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6387),
    [anon_sym_new] = ACTIONS(6389),
    [anon_sym_with] = ACTIONS(6389),
    [anon_sym_return] = ACTIONS(6389),
    [anon_sym_return_BANG] = ACTIONS(6387),
    [anon_sym_yield] = ACTIONS(6389),
    [anon_sym_yield_BANG] = ACTIONS(6387),
    [anon_sym_lazy] = ACTIONS(6389),
    [anon_sym_assert] = ACTIONS(6389),
    [anon_sym_upcast] = ACTIONS(6389),
    [anon_sym_downcast] = ACTIONS(6389),
    [anon_sym_fixed] = ACTIONS(6389),
    [anon_sym_COLON_GT] = ACTIONS(6387),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6387),
    [anon_sym_for] = ACTIONS(6389),
    [anon_sym_while] = ACTIONS(6389),
    [anon_sym_while_BANG] = ACTIONS(6387),
    [anon_sym_if] = ACTIONS(6389),
    [anon_sym_fun] = ACTIONS(6389),
    [anon_sym_DASH_GT] = ACTIONS(6389),
    [anon_sym_try] = ACTIONS(6389),
    [anon_sym_match] = ACTIONS(6389),
    [anon_sym_match_BANG] = ACTIONS(6387),
    [anon_sym_function] = ACTIONS(6389),
    [anon_sym_DOT_LBRACK] = ACTIONS(6387),
    [anon_sym_use] = ACTIONS(6389),
    [anon_sym_use_BANG] = ACTIONS(6387),
    [anon_sym_do_BANG] = ACTIONS(6387),
    [anon_sym_LPAREN2] = ACTIONS(6389),
    [anon_sym_PIPE_PIPE] = ACTIONS(6389),
    [anon_sym_AMP_AMP] = ACTIONS(6389),
    [anon_sym_STAR] = ACTIONS(6389),
    [anon_sym_LT] = ACTIONS(6389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6387),
    [aux_sym_char_token1] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6389),
    [anon_sym_DQUOTE] = ACTIONS(6389),
    [anon_sym_AT_DQUOTE] = ACTIONS(6387),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6387),
    [sym_bool] = ACTIONS(6389),
    [sym_unit] = ACTIONS(6387),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6389),
    [sym_op_identifier] = ACTIONS(6389),
    [anon_sym_PLUS] = ACTIONS(6389),
    [anon_sym_DASH] = ACTIONS(6389),
    [anon_sym_PLUS_DOT] = ACTIONS(6389),
    [anon_sym_DASH_DOT] = ACTIONS(6389),
    [anon_sym_PERCENT] = ACTIONS(6389),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6389),
    [anon_sym_TILDE] = ACTIONS(6387),
    [aux_sym_prefix_op_token1] = ACTIONS(6389),
    [aux_sym_infix_op_token1] = ACTIONS(6387),
    [aux_sym_infix_op_token2] = ACTIONS(6389),
    [aux_sym_infix_op_token3] = ACTIONS(6389),
    [anon_sym_BANG_EQ] = ACTIONS(6389),
    [anon_sym_COLON_EQ] = ACTIONS(6387),
    [anon_sym_DOLLAR] = ACTIONS(6389),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6389),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6389),
    [sym_int] = ACTIONS(6389),
    [sym_xint] = ACTIONS(6387),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6387),
    [sym__newline] = ACTIONS(6387),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6387),
    [sym__tyapp_open] = ACTIONS(6387),
  },
  [STATE(4744)] = {
    [sym_block_comment] = STATE(4744),
    [sym_compiler_directive_decl] = STATE(4744),
    [sym_fsi_directive_decl] = STATE(4744),
    [sym_preproc_line] = STATE(4744),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_DOT_DOT] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4745)] = {
    [sym_block_comment] = STATE(4745),
    [sym_compiler_directive_decl] = STATE(4745),
    [sym_fsi_directive_decl] = STATE(4745),
    [sym_preproc_line] = STATE(4745),
    [sym_identifier] = ACTIONS(6407),
    [anon_sym_null] = ACTIONS(6407),
    [anon_sym_EQ] = ACTIONS(6407),
    [anon_sym_begin] = ACTIONS(6407),
    [anon_sym_COLON] = ACTIONS(6407),
    [anon_sym_do] = ACTIONS(6407),
    [anon_sym_let] = ACTIONS(6407),
    [anon_sym_let_BANG] = ACTIONS(6409),
    [anon_sym_LPAREN] = ACTIONS(6407),
    [anon_sym_COMMA] = ACTIONS(6409),
    [anon_sym_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_QMARK] = ACTIONS(6407),
    [anon_sym_COLON_COLON] = ACTIONS(6409),
    [anon_sym_AMP] = ACTIONS(6407),
    [anon_sym_LBRACK] = ACTIONS(6407),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6409),
    [anon_sym_LBRACE] = ACTIONS(6407),
    [anon_sym_LT_AT] = ACTIONS(6407),
    [anon_sym_LT_AT_AT] = ACTIONS(6407),
    [anon_sym_CARET] = ACTIONS(6407),
    [anon_sym_or] = ACTIONS(6407),
    [anon_sym_DOT] = ACTIONS(6407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6409),
    [anon_sym_new] = ACTIONS(6407),
    [anon_sym_return] = ACTIONS(6407),
    [anon_sym_return_BANG] = ACTIONS(6409),
    [anon_sym_yield] = ACTIONS(6407),
    [anon_sym_yield_BANG] = ACTIONS(6409),
    [anon_sym_lazy] = ACTIONS(6407),
    [anon_sym_assert] = ACTIONS(6407),
    [anon_sym_upcast] = ACTIONS(6407),
    [anon_sym_downcast] = ACTIONS(6407),
    [anon_sym_fixed] = ACTIONS(6407),
    [anon_sym_COLON_GT] = ACTIONS(6409),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6409),
    [anon_sym_for] = ACTIONS(6407),
    [anon_sym_while] = ACTIONS(6407),
    [anon_sym_while_BANG] = ACTIONS(6409),
    [anon_sym_if] = ACTIONS(6407),
    [anon_sym_fun] = ACTIONS(6407),
    [anon_sym_DASH_GT] = ACTIONS(6407),
    [anon_sym_try] = ACTIONS(6407),
    [anon_sym_match] = ACTIONS(6407),
    [anon_sym_match_BANG] = ACTIONS(6409),
    [anon_sym_function] = ACTIONS(6407),
    [anon_sym_DOT_LBRACK] = ACTIONS(6409),
    [anon_sym_use] = ACTIONS(6407),
    [anon_sym_use_BANG] = ACTIONS(6409),
    [anon_sym_do_BANG] = ACTIONS(6409),
    [anon_sym_DOT_DOT] = ACTIONS(6409),
    [anon_sym_LPAREN2] = ACTIONS(6407),
    [anon_sym_PIPE_PIPE] = ACTIONS(6407),
    [anon_sym_AMP_AMP] = ACTIONS(6407),
    [anon_sym_STAR] = ACTIONS(6407),
    [anon_sym_LT] = ACTIONS(6407),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6409),
    [aux_sym_char_token1] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6407),
    [anon_sym_DQUOTE] = ACTIONS(6407),
    [anon_sym_AT_DQUOTE] = ACTIONS(6409),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6409),
    [sym_bool] = ACTIONS(6407),
    [sym_unit] = ACTIONS(6409),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6407),
    [sym_op_identifier] = ACTIONS(6407),
    [anon_sym_PLUS] = ACTIONS(6407),
    [anon_sym_DASH] = ACTIONS(6407),
    [anon_sym_PLUS_DOT] = ACTIONS(6407),
    [anon_sym_DASH_DOT] = ACTIONS(6407),
    [anon_sym_PERCENT] = ACTIONS(6407),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6407),
    [anon_sym_TILDE] = ACTIONS(6409),
    [aux_sym_prefix_op_token1] = ACTIONS(6407),
    [aux_sym_infix_op_token1] = ACTIONS(6409),
    [aux_sym_infix_op_token2] = ACTIONS(6407),
    [aux_sym_infix_op_token3] = ACTIONS(6407),
    [anon_sym_BANG_EQ] = ACTIONS(6407),
    [anon_sym_COLON_EQ] = ACTIONS(6409),
    [anon_sym_DOLLAR] = ACTIONS(6407),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6407),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6407),
    [sym_int] = ACTIONS(6407),
    [sym_xint] = ACTIONS(6409),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6409),
    [sym__newline] = ACTIONS(6409),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6409),
    [sym__tyapp_open] = ACTIONS(6409),
  },
  [STATE(4746)] = {
    [sym_block_comment] = STATE(4746),
    [sym_compiler_directive_decl] = STATE(4746),
    [sym_fsi_directive_decl] = STATE(4746),
    [sym_preproc_line] = STATE(4746),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_as] = ACTIONS(6612),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4747)] = {
    [sym_block_comment] = STATE(4747),
    [sym_compiler_directive_decl] = STATE(4747),
    [sym_fsi_directive_decl] = STATE(4747),
    [sym_preproc_line] = STATE(4747),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__untyped_quoted_close] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4748)] = {
    [sym_block_comment] = STATE(4748),
    [sym_compiler_directive_decl] = STATE(4748),
    [sym_fsi_directive_decl] = STATE(4748),
    [sym_preproc_line] = STATE(4748),
    [sym_identifier] = ACTIONS(6367),
    [anon_sym_null] = ACTIONS(6367),
    [anon_sym_EQ] = ACTIONS(6367),
    [anon_sym_begin] = ACTIONS(6367),
    [anon_sym_COLON] = ACTIONS(6367),
    [anon_sym_do] = ACTIONS(6367),
    [anon_sym_let] = ACTIONS(6367),
    [anon_sym_let_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN] = ACTIONS(6367),
    [anon_sym_COMMA] = ACTIONS(6365),
    [anon_sym_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_QMARK] = ACTIONS(6367),
    [anon_sym_COLON_COLON] = ACTIONS(6365),
    [anon_sym_AMP] = ACTIONS(6367),
    [anon_sym_GT] = ACTIONS(6367),
    [anon_sym_LBRACK] = ACTIONS(6367),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6365),
    [anon_sym_LBRACE] = ACTIONS(6367),
    [anon_sym_LT_AT] = ACTIONS(6367),
    [anon_sym_LT_AT_AT] = ACTIONS(6367),
    [anon_sym_CARET] = ACTIONS(6367),
    [anon_sym_or] = ACTIONS(6367),
    [anon_sym_DOT] = ACTIONS(6367),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6365),
    [anon_sym_new] = ACTIONS(6367),
    [anon_sym_return] = ACTIONS(6367),
    [anon_sym_return_BANG] = ACTIONS(6365),
    [anon_sym_yield] = ACTIONS(6367),
    [anon_sym_yield_BANG] = ACTIONS(6365),
    [anon_sym_lazy] = ACTIONS(6367),
    [anon_sym_assert] = ACTIONS(6367),
    [anon_sym_upcast] = ACTIONS(6367),
    [anon_sym_downcast] = ACTIONS(6367),
    [anon_sym_fixed] = ACTIONS(6367),
    [anon_sym_COLON_GT] = ACTIONS(6365),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6365),
    [anon_sym_for] = ACTIONS(6367),
    [anon_sym_while] = ACTIONS(6367),
    [anon_sym_while_BANG] = ACTIONS(6365),
    [anon_sym_if] = ACTIONS(6367),
    [anon_sym_fun] = ACTIONS(6367),
    [anon_sym_DASH_GT] = ACTIONS(6367),
    [anon_sym_try] = ACTIONS(6367),
    [anon_sym_match] = ACTIONS(6367),
    [anon_sym_match_BANG] = ACTIONS(6365),
    [anon_sym_function] = ACTIONS(6367),
    [anon_sym_DOT_LBRACK] = ACTIONS(6365),
    [anon_sym_use] = ACTIONS(6367),
    [anon_sym_use_BANG] = ACTIONS(6365),
    [anon_sym_do_BANG] = ACTIONS(6365),
    [anon_sym_LPAREN2] = ACTIONS(6367),
    [anon_sym_PIPE_PIPE] = ACTIONS(6367),
    [anon_sym_AMP_AMP] = ACTIONS(6367),
    [anon_sym_STAR] = ACTIONS(6367),
    [anon_sym_LT] = ACTIONS(6367),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6365),
    [aux_sym_char_token1] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6367),
    [anon_sym_DQUOTE] = ACTIONS(6367),
    [anon_sym_AT_DQUOTE] = ACTIONS(6365),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6365),
    [sym_bool] = ACTIONS(6367),
    [sym_unit] = ACTIONS(6365),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6367),
    [sym_op_identifier] = ACTIONS(6367),
    [anon_sym_PLUS] = ACTIONS(6367),
    [anon_sym_DASH] = ACTIONS(6367),
    [anon_sym_PLUS_DOT] = ACTIONS(6367),
    [anon_sym_DASH_DOT] = ACTIONS(6367),
    [anon_sym_PERCENT] = ACTIONS(6367),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6367),
    [anon_sym_TILDE] = ACTIONS(6365),
    [aux_sym_prefix_op_token1] = ACTIONS(6367),
    [aux_sym_infix_op_token1] = ACTIONS(6365),
    [aux_sym_infix_op_token2] = ACTIONS(6367),
    [aux_sym_infix_op_token3] = ACTIONS(6367),
    [anon_sym_BANG_EQ] = ACTIONS(6367),
    [anon_sym_COLON_EQ] = ACTIONS(6365),
    [anon_sym_DOLLAR] = ACTIONS(6367),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6367),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6367),
    [sym_int] = ACTIONS(6367),
    [sym_xint] = ACTIONS(6365),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6365),
    [sym__newline] = ACTIONS(6365),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6365),
    [sym__tyapp_open] = ACTIONS(6365),
  },
  [STATE(4749)] = {
    [sym_block_comment] = STATE(4749),
    [sym_compiler_directive_decl] = STATE(4749),
    [sym_fsi_directive_decl] = STATE(4749),
    [sym_preproc_line] = STATE(4749),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_as] = ACTIONS(7035),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_with] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_DOT_DOT] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__dedent] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(4750)] = {
    [sym_block_comment] = STATE(4750),
    [sym_compiler_directive_decl] = STATE(4750),
    [sym_fsi_directive_decl] = STATE(4750),
    [sym_preproc_line] = STATE(4750),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [anon_sym_POUNDelse] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4751)] = {
    [sym_block_comment] = STATE(4751),
    [sym_compiler_directive_decl] = STATE(4751),
    [sym_fsi_directive_decl] = STATE(4751),
    [sym_preproc_line] = STATE(4751),
    [aux_sym_rules_repeat1] = STATE(4752),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9137),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [anon_sym_POUNDelse] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4752)] = {
    [sym_block_comment] = STATE(4752),
    [sym_compiler_directive_decl] = STATE(4752),
    [sym_fsi_directive_decl] = STATE(4752),
    [sym_preproc_line] = STATE(4752),
    [aux_sym_rules_repeat1] = STATE(4752),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9142),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9145),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4753)] = {
    [sym_block_comment] = STATE(4753),
    [sym_compiler_directive_decl] = STATE(4753),
    [sym_fsi_directive_decl] = STATE(4753),
    [sym_preproc_line] = STATE(4753),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_DASH_GT] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7952),
    [aux_sym_decimal_token1] = ACTIONS(7360),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4754)] = {
    [sym_block_comment] = STATE(4754),
    [sym_compiler_directive_decl] = STATE(4754),
    [sym_fsi_directive_decl] = STATE(4754),
    [sym_preproc_line] = STATE(4754),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9054),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6597),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_DOT_DOT2] = ACTIONS(6599),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4755)] = {
    [sym_block_comment] = STATE(4755),
    [sym_compiler_directive_decl] = STATE(4755),
    [sym_fsi_directive_decl] = STATE(4755),
    [sym_preproc_line] = STATE(4755),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_DASH_GT] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_DOT_DOT] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9148),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4756)] = {
    [sym_block_comment] = STATE(4756),
    [sym_compiler_directive_decl] = STATE(4756),
    [sym_fsi_directive_decl] = STATE(4756),
    [sym_preproc_line] = STATE(4756),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9150),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_DOT_DOT] = ACTIONS(6784),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [anon_sym_DOT_DOT2] = ACTIONS(6786),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__dedent] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(4757)] = {
    [sym_block_comment] = STATE(4757),
    [sym_compiler_directive_decl] = STATE(4757),
    [sym_fsi_directive_decl] = STATE(4757),
    [sym_preproc_line] = STATE(4757),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_DASH_GT] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4758)] = {
    [sym_block_comment] = STATE(4758),
    [sym_compiler_directive_decl] = STATE(4758),
    [sym_fsi_directive_decl] = STATE(4758),
    [sym_preproc_line] = STATE(4758),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
    [sym__in] = ACTIONS(6614),
  },
  [STATE(4759)] = {
    [sym_block_comment] = STATE(4759),
    [sym_compiler_directive_decl] = STATE(4759),
    [sym_fsi_directive_decl] = STATE(4759),
    [sym_preproc_line] = STATE(4759),
    [aux_sym_long_identifier_repeat1] = STATE(4910),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9152),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
    [sym__in] = ACTIONS(6628),
  },
  [STATE(4760)] = {
    [sym_block_comment] = STATE(4760),
    [sym_compiler_directive_decl] = STATE(4760),
    [sym_fsi_directive_decl] = STATE(4760),
    [sym_preproc_line] = STATE(4760),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_DOT_DOT] = ACTIONS(6790),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [anon_sym_DOT_DOT2] = ACTIONS(6792),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__dedent] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(4761)] = {
    [sym_block_comment] = STATE(4761),
    [sym_compiler_directive_decl] = STATE(4761),
    [sym_fsi_directive_decl] = STATE(4761),
    [sym_preproc_line] = STATE(4761),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_DASH_GT] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_DOT_DOT] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4762)] = {
    [sym_block_comment] = STATE(4762),
    [sym_compiler_directive_decl] = STATE(4762),
    [sym_fsi_directive_decl] = STATE(4762),
    [sym_preproc_line] = STATE(4762),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9156),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_DASH_GT] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4763)] = {
    [sym_block_comment] = STATE(4763),
    [sym_compiler_directive_decl] = STATE(4763),
    [sym_fsi_directive_decl] = STATE(4763),
    [sym_preproc_line] = STATE(4763),
    [aux_sym_long_identifier_repeat1] = STATE(4768),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9158),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4764)] = {
    [sym_block_comment] = STATE(4764),
    [sym_compiler_directive_decl] = STATE(4764),
    [sym_fsi_directive_decl] = STATE(4764),
    [sym_preproc_line] = STATE(4764),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_as] = ACTIONS(6637),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9160),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4765)] = {
    [sym_block_comment] = STATE(4765),
    [sym_compiler_directive_decl] = STATE(4765),
    [sym_fsi_directive_decl] = STATE(4765),
    [sym_preproc_line] = STATE(4765),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_DOT_DOT] = ACTIONS(6794),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [anon_sym_DOT_DOT2] = ACTIONS(6796),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__dedent] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(4766)] = {
    [sym_block_comment] = STATE(4766),
    [sym_compiler_directive_decl] = STATE(4766),
    [sym_fsi_directive_decl] = STATE(4766),
    [sym_preproc_line] = STATE(4766),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9162),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4767)] = {
    [sym_block_comment] = STATE(4767),
    [sym_compiler_directive_decl] = STATE(4767),
    [sym_fsi_directive_decl] = STATE(4767),
    [sym_preproc_line] = STATE(4767),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_as] = ACTIONS(6612),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4768)] = {
    [sym_block_comment] = STATE(4768),
    [sym_compiler_directive_decl] = STATE(4768),
    [sym_fsi_directive_decl] = STATE(4768),
    [sym_preproc_line] = STATE(4768),
    [aux_sym_long_identifier_repeat1] = STATE(4768),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9164),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4769)] = {
    [sym_block_comment] = STATE(4769),
    [sym_compiler_directive_decl] = STATE(4769),
    [sym_fsi_directive_decl] = STATE(4769),
    [sym_preproc_line] = STATE(4769),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4770)] = {
    [sym_block_comment] = STATE(4770),
    [sym_compiler_directive_decl] = STATE(4770),
    [sym_fsi_directive_decl] = STATE(4770),
    [sym_preproc_line] = STATE(4770),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_DASH_GT] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4771)] = {
    [sym_block_comment] = STATE(4771),
    [sym_compiler_directive_decl] = STATE(4771),
    [sym_fsi_directive_decl] = STATE(4771),
    [sym_preproc_line] = STATE(4771),
    [aux_sym_sequential_expression_repeat1] = STATE(4771),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_DOT_DOT] = ACTIONS(6800),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_DOT_DOT2] = ACTIONS(6798),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9167),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(4772)] = {
    [sym_block_comment] = STATE(4772),
    [sym_compiler_directive_decl] = STATE(4772),
    [sym_fsi_directive_decl] = STATE(4772),
    [sym_preproc_line] = STATE(4772),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_as] = ACTIONS(7053),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_DOT_DOT] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__dedent] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(4773)] = {
    [sym_block_comment] = STATE(4773),
    [sym_compiler_directive_decl] = STATE(4773),
    [sym_fsi_directive_decl] = STATE(4773),
    [sym_preproc_line] = STATE(4773),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4774)] = {
    [sym_block_comment] = STATE(4774),
    [sym_compiler_directive_decl] = STATE(4774),
    [sym_fsi_directive_decl] = STATE(4774),
    [sym_preproc_line] = STATE(4774),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_as] = ACTIONS(6888),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_with] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_DOT_DOT] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__dedent] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(4775)] = {
    [sym_block_comment] = STATE(4775),
    [sym_compiler_directive_decl] = STATE(4775),
    [sym_fsi_directive_decl] = STATE(4775),
    [sym_preproc_line] = STATE(4775),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_as] = ACTIONS(7069),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_with] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_DOT_DOT] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__dedent] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(4776)] = {
    [sym_block_comment] = STATE(4776),
    [sym_compiler_directive_decl] = STATE(4776),
    [sym_fsi_directive_decl] = STATE(4776),
    [sym_preproc_line] = STATE(4776),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_as] = ACTIONS(6597),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9170),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4777)] = {
    [sym_block_comment] = STATE(4777),
    [sym_compiler_directive_decl] = STATE(4777),
    [sym_fsi_directive_decl] = STATE(4777),
    [sym_preproc_line] = STATE(4777),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_as] = ACTIONS(7073),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_with] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_DOT_DOT] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__dedent] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(4778)] = {
    [sym_block_comment] = STATE(4778),
    [sym_compiler_directive_decl] = STATE(4778),
    [sym_fsi_directive_decl] = STATE(4778),
    [sym_preproc_line] = STATE(4778),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_as] = ACTIONS(7077),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_with] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_DOT_DOT] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__dedent] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(4779)] = {
    [sym_block_comment] = STATE(4779),
    [sym_compiler_directive_decl] = STATE(4779),
    [sym_fsi_directive_decl] = STATE(4779),
    [sym_preproc_line] = STATE(4779),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_as] = ACTIONS(7085),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_with] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_DOT_DOT] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__dedent] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(4780)] = {
    [sym_block_comment] = STATE(4780),
    [sym_compiler_directive_decl] = STATE(4780),
    [sym_fsi_directive_decl] = STATE(4780),
    [sym_preproc_line] = STATE(4780),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_as] = ACTIONS(7089),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_with] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_DOT_DOT] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__dedent] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(4781)] = {
    [sym_block_comment] = STATE(4781),
    [sym_compiler_directive_decl] = STATE(4781),
    [sym_fsi_directive_decl] = STATE(4781),
    [sym_preproc_line] = STATE(4781),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_as] = ACTIONS(7093),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_with] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_DOT_DOT] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__dedent] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(4782)] = {
    [sym_block_comment] = STATE(4782),
    [sym_compiler_directive_decl] = STATE(4782),
    [sym_fsi_directive_decl] = STATE(4782),
    [sym_preproc_line] = STATE(4782),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_as] = ACTIONS(7097),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_with] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_DOT_DOT] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__dedent] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(4783)] = {
    [sym_block_comment] = STATE(4783),
    [sym_compiler_directive_decl] = STATE(4783),
    [sym_fsi_directive_decl] = STATE(4783),
    [sym_preproc_line] = STATE(4783),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_as] = ACTIONS(7101),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_with] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_DOT_DOT] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__dedent] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(4784)] = {
    [sym_block_comment] = STATE(4784),
    [sym_compiler_directive_decl] = STATE(4784),
    [sym_fsi_directive_decl] = STATE(4784),
    [sym_preproc_line] = STATE(4784),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_as] = ACTIONS(7105),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_with] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_DOT_DOT] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__dedent] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(4785)] = {
    [sym_block_comment] = STATE(4785),
    [sym_compiler_directive_decl] = STATE(4785),
    [sym_fsi_directive_decl] = STATE(4785),
    [sym_preproc_line] = STATE(4785),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_as] = ACTIONS(7109),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_with] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_DOT_DOT] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__dedent] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(4786)] = {
    [sym_block_comment] = STATE(4786),
    [sym_compiler_directive_decl] = STATE(4786),
    [sym_fsi_directive_decl] = STATE(4786),
    [sym_preproc_line] = STATE(4786),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_as] = ACTIONS(7113),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_with] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_DOT_DOT] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__dedent] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(4787)] = {
    [sym_block_comment] = STATE(4787),
    [sym_compiler_directive_decl] = STATE(4787),
    [sym_fsi_directive_decl] = STATE(4787),
    [sym_preproc_line] = STATE(4787),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_as] = ACTIONS(7117),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_with] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_DOT_DOT] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__dedent] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(4788)] = {
    [sym_block_comment] = STATE(4788),
    [sym_compiler_directive_decl] = STATE(4788),
    [sym_fsi_directive_decl] = STATE(4788),
    [sym_preproc_line] = STATE(4788),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_as] = ACTIONS(7121),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_with] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_DOT_DOT] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__dedent] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(4789)] = {
    [sym_block_comment] = STATE(4789),
    [sym_compiler_directive_decl] = STATE(4789),
    [sym_fsi_directive_decl] = STATE(4789),
    [sym_preproc_line] = STATE(4789),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_as] = ACTIONS(6936),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_DOT_DOT] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__dedent] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(4790)] = {
    [sym_block_comment] = STATE(4790),
    [sym_compiler_directive_decl] = STATE(4790),
    [sym_fsi_directive_decl] = STATE(4790),
    [sym_preproc_line] = STATE(4790),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_as] = ACTIONS(6940),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_DOT_DOT] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__dedent] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(4791)] = {
    [sym_block_comment] = STATE(4791),
    [sym_compiler_directive_decl] = STATE(4791),
    [sym_fsi_directive_decl] = STATE(4791),
    [sym_preproc_line] = STATE(4791),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_as] = ACTIONS(7015),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_DOT_DOT] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__dedent] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(4792)] = {
    [sym_block_comment] = STATE(4792),
    [sym_compiler_directive_decl] = STATE(4792),
    [sym_fsi_directive_decl] = STATE(4792),
    [sym_preproc_line] = STATE(4792),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_as] = ACTIONS(7031),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_DOT_DOT] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__dedent] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(4793)] = {
    [sym_block_comment] = STATE(4793),
    [sym_compiler_directive_decl] = STATE(4793),
    [sym_fsi_directive_decl] = STATE(4793),
    [sym_preproc_line] = STATE(4793),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_as] = ACTIONS(7047),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_DOT_DOT] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__dedent] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(4794)] = {
    [sym_block_comment] = STATE(4794),
    [sym_compiler_directive_decl] = STATE(4794),
    [sym_fsi_directive_decl] = STATE(4794),
    [sym_preproc_line] = STATE(4794),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_as] = ACTIONS(7051),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_DOT_DOT] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__dedent] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(4795)] = {
    [sym_block_comment] = STATE(4795),
    [sym_compiler_directive_decl] = STATE(4795),
    [sym_fsi_directive_decl] = STATE(4795),
    [sym_preproc_line] = STATE(4795),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_as] = ACTIONS(7063),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_DOT_DOT] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__dedent] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(4796)] = {
    [sym_block_comment] = STATE(4796),
    [sym_compiler_directive_decl] = STATE(4796),
    [sym_fsi_directive_decl] = STATE(4796),
    [sym_preproc_line] = STATE(4796),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_as] = ACTIONS(7067),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_DOT_DOT] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__dedent] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(4797)] = {
    [sym_block_comment] = STATE(4797),
    [sym_compiler_directive_decl] = STATE(4797),
    [sym_fsi_directive_decl] = STATE(4797),
    [sym_preproc_line] = STATE(4797),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_as] = ACTIONS(7125),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_DOT_DOT] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__dedent] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(4798)] = {
    [sym_block_comment] = STATE(4798),
    [sym_compiler_directive_decl] = STATE(4798),
    [sym_fsi_directive_decl] = STATE(4798),
    [sym_preproc_line] = STATE(4798),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_as] = ACTIONS(7129),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_DOT_DOT] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__dedent] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(4799)] = {
    [sym_block_comment] = STATE(4799),
    [sym_compiler_directive_decl] = STATE(4799),
    [sym_fsi_directive_decl] = STATE(4799),
    [sym_preproc_line] = STATE(4799),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_as] = ACTIONS(7133),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_DOT_DOT] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__dedent] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(4800)] = {
    [sym_block_comment] = STATE(4800),
    [sym_compiler_directive_decl] = STATE(4800),
    [sym_fsi_directive_decl] = STATE(4800),
    [sym_preproc_line] = STATE(4800),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_as] = ACTIONS(7137),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_DOT_DOT] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__dedent] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(4801)] = {
    [sym_block_comment] = STATE(4801),
    [sym_compiler_directive_decl] = STATE(4801),
    [sym_fsi_directive_decl] = STATE(4801),
    [sym_preproc_line] = STATE(4801),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_DOT_DOT] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__dedent] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4802)] = {
    [sym_block_comment] = STATE(4802),
    [sym_compiler_directive_decl] = STATE(4802),
    [sym_fsi_directive_decl] = STATE(4802),
    [sym_preproc_line] = STATE(4802),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_DOT_DOT] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__dedent] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4803)] = {
    [sym_block_comment] = STATE(4803),
    [sym_compiler_directive_decl] = STATE(4803),
    [sym_fsi_directive_decl] = STATE(4803),
    [sym_preproc_line] = STATE(4803),
    [aux_sym_long_identifier_repeat1] = STATE(4806),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_as] = ACTIONS(6360),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9172),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4804)] = {
    [sym_block_comment] = STATE(4804),
    [sym_compiler_directive_decl] = STATE(4804),
    [sym_fsi_directive_decl] = STATE(4804),
    [sym_preproc_line] = STATE(4804),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6807),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_DOT_DOT2] = ACTIONS(6805),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4805)] = {
    [sym_block_comment] = STATE(4805),
    [sym_compiler_directive_decl] = STATE(4805),
    [sym_fsi_directive_decl] = STATE(4805),
    [sym_preproc_line] = STATE(4805),
    [aux_sym_rules_repeat1] = STATE(4810),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9174),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_DOT_DOT] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9176),
    [sym__dedent] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(4806)] = {
    [sym_block_comment] = STATE(4806),
    [sym_compiler_directive_decl] = STATE(4806),
    [sym_fsi_directive_decl] = STATE(4806),
    [sym_preproc_line] = STATE(4806),
    [aux_sym_long_identifier_repeat1] = STATE(4806),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9179),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4807)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4807),
    [sym_compiler_directive_decl] = STATE(4807),
    [sym_fsi_directive_decl] = STATE(4807),
    [sym_preproc_line] = STATE(4807),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9182),
    [anon_sym_module] = ACTIONS(9182),
    [anon_sym_null] = ACTIONS(9182),
    [anon_sym_begin] = ACTIONS(9182),
    [anon_sym_open] = ACTIONS(9182),
    [anon_sym_type] = ACTIONS(9182),
    [anon_sym_LBRACK_LT] = ACTIONS(9184),
    [anon_sym_do] = ACTIONS(9182),
    [anon_sym_and] = ACTIONS(9182),
    [anon_sym_let] = ACTIONS(9182),
    [anon_sym_let_BANG] = ACTIONS(9184),
    [aux_sym_access_modifier_token1] = ACTIONS(9184),
    [anon_sym_LPAREN] = ACTIONS(9182),
    [anon_sym_PIPE] = ACTIONS(9184),
    [anon_sym_AMP] = ACTIONS(9182),
    [anon_sym_LBRACK] = ACTIONS(9182),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9184),
    [anon_sym_LBRACE] = ACTIONS(9182),
    [anon_sym_LT_AT] = ACTIONS(9182),
    [anon_sym_LT_AT_AT] = ACTIONS(9184),
    [anon_sym_CARET] = ACTIONS(9184),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9184),
    [anon_sym_new] = ACTIONS(9182),
    [anon_sym_with] = ACTIONS(9182),
    [anon_sym_return] = ACTIONS(9182),
    [anon_sym_return_BANG] = ACTIONS(9184),
    [anon_sym_yield] = ACTIONS(9182),
    [anon_sym_yield_BANG] = ACTIONS(9184),
    [anon_sym_lazy] = ACTIONS(9182),
    [anon_sym_assert] = ACTIONS(9182),
    [anon_sym_upcast] = ACTIONS(9182),
    [anon_sym_downcast] = ACTIONS(9182),
    [anon_sym_fixed] = ACTIONS(9182),
    [anon_sym_for] = ACTIONS(9182),
    [anon_sym_while] = ACTIONS(9182),
    [anon_sym_while_BANG] = ACTIONS(9184),
    [anon_sym_if] = ACTIONS(9182),
    [anon_sym_fun] = ACTIONS(9182),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(9182),
    [anon_sym_match] = ACTIONS(9182),
    [anon_sym_match_BANG] = ACTIONS(9184),
    [anon_sym_function] = ACTIONS(9182),
    [anon_sym_use] = ACTIONS(9182),
    [anon_sym_use_BANG] = ACTIONS(9184),
    [anon_sym_do_BANG] = ACTIONS(9184),
    [anon_sym_AMP_AMP] = ACTIONS(9184),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(9182),
    [anon_sym_static] = ACTIONS(9182),
    [anon_sym_member] = ACTIONS(9182),
    [anon_sym_exception] = ACTIONS(9182),
    [anon_sym_interface] = ACTIONS(9182),
    [anon_sym_abstract] = ACTIONS(9182),
    [anon_sym_val] = ACTIONS(9182),
    [anon_sym_override] = ACTIONS(9182),
    [anon_sym_extern] = ACTIONS(9182),
    [aux_sym_char_token1] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9182),
    [anon_sym_DQUOTE] = ACTIONS(9182),
    [anon_sym_AT_DQUOTE] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [sym_bool] = ACTIONS(9182),
    [sym_unit] = ACTIONS(9184),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9182),
    [sym_op_identifier] = ACTIONS(9184),
    [anon_sym_PLUS] = ACTIONS(9182),
    [anon_sym_DASH] = ACTIONS(9182),
    [anon_sym_PLUS_DOT] = ACTIONS(9184),
    [anon_sym_DASH_DOT] = ACTIONS(9184),
    [anon_sym_PERCENT] = ACTIONS(9182),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9184),
    [anon_sym_TILDE] = ACTIONS(9184),
    [aux_sym_prefix_op_token1] = ACTIONS(9184),
    [sym_int] = ACTIONS(9182),
    [sym_xint] = ACTIONS(9184),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9184),
    [anon_sym_POUNDload] = ACTIONS(9184),
    [anon_sym_POUNDtime] = ACTIONS(9184),
    [anon_sym_POUNDI] = ACTIONS(9184),
    [anon_sym_POUNDhelp] = ACTIONS(9184),
    [anon_sym_POUNDquit] = ACTIONS(9184),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9184),
    [anon_sym_POUNDendif] = ACTIONS(9184),
    [anon_sym_POUNDelse] = ACTIONS(9184),
    [sym__newline] = ACTIONS(9184),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9184),
  },
  [STATE(4808)] = {
    [sym_block_comment] = STATE(4808),
    [sym_compiler_directive_decl] = STATE(4808),
    [sym_fsi_directive_decl] = STATE(4808),
    [sym_preproc_line] = STATE(4808),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9054),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6597),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_DOT_DOT2] = ACTIONS(6599),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4809)] = {
    [sym_block_comment] = STATE(4809),
    [sym_compiler_directive_decl] = STATE(4809),
    [sym_fsi_directive_decl] = STATE(4809),
    [sym_preproc_line] = STATE(4809),
    [aux_sym_rules_repeat1] = STATE(4811),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9174),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9194),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4810)] = {
    [sym_block_comment] = STATE(4810),
    [sym_compiler_directive_decl] = STATE(4810),
    [sym_fsi_directive_decl] = STATE(4810),
    [sym_preproc_line] = STATE(4810),
    [aux_sym_rules_repeat1] = STATE(4812),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9174),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9194),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4811)] = {
    [sym_block_comment] = STATE(4811),
    [sym_compiler_directive_decl] = STATE(4811),
    [sym_fsi_directive_decl] = STATE(4811),
    [sym_preproc_line] = STATE(4811),
    [aux_sym_rules_repeat1] = STATE(4812),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9174),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_DOT_DOT] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9197),
    [sym__dedent] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4812)] = {
    [sym_block_comment] = STATE(4812),
    [sym_compiler_directive_decl] = STATE(4812),
    [sym_fsi_directive_decl] = STATE(4812),
    [sym_preproc_line] = STATE(4812),
    [aux_sym_rules_repeat1] = STATE(4812),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9200),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9203),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4813)] = {
    [sym_block_comment] = STATE(4813),
    [sym_compiler_directive_decl] = STATE(4813),
    [sym_fsi_directive_decl] = STATE(4813),
    [sym_preproc_line] = STATE(4813),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9206),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__quoted_close] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4814)] = {
    [sym_block_comment] = STATE(4814),
    [sym_compiler_directive_decl] = STATE(4814),
    [sym_fsi_directive_decl] = STATE(4814),
    [sym_preproc_line] = STATE(4814),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_as] = ACTIONS(7141),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_DOT_DOT] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__dedent] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(4815)] = {
    [sym_block_comment] = STATE(4815),
    [sym_compiler_directive_decl] = STATE(4815),
    [sym_fsi_directive_decl] = STATE(4815),
    [sym_preproc_line] = STATE(4815),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_as] = ACTIONS(7145),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_DOT_DOT] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__dedent] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(4816)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4816),
    [sym_compiler_directive_decl] = STATE(4816),
    [sym_fsi_directive_decl] = STATE(4816),
    [sym_preproc_line] = STATE(4816),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9208),
    [anon_sym_module] = ACTIONS(9208),
    [anon_sym_null] = ACTIONS(9208),
    [anon_sym_begin] = ACTIONS(9208),
    [anon_sym_open] = ACTIONS(9208),
    [anon_sym_type] = ACTIONS(9208),
    [anon_sym_LBRACK_LT] = ACTIONS(9210),
    [anon_sym_do] = ACTIONS(9208),
    [anon_sym_and] = ACTIONS(9208),
    [anon_sym_let] = ACTIONS(9208),
    [anon_sym_let_BANG] = ACTIONS(9210),
    [aux_sym_access_modifier_token1] = ACTIONS(9210),
    [anon_sym_LPAREN] = ACTIONS(9208),
    [anon_sym_PIPE] = ACTIONS(9210),
    [anon_sym_AMP] = ACTIONS(9208),
    [anon_sym_LBRACK] = ACTIONS(9208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9210),
    [anon_sym_LBRACE] = ACTIONS(9208),
    [anon_sym_LT_AT] = ACTIONS(9208),
    [anon_sym_LT_AT_AT] = ACTIONS(9210),
    [anon_sym_CARET] = ACTIONS(9210),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9210),
    [anon_sym_new] = ACTIONS(9208),
    [anon_sym_with] = ACTIONS(9208),
    [anon_sym_return] = ACTIONS(9208),
    [anon_sym_return_BANG] = ACTIONS(9210),
    [anon_sym_yield] = ACTIONS(9208),
    [anon_sym_yield_BANG] = ACTIONS(9210),
    [anon_sym_lazy] = ACTIONS(9208),
    [anon_sym_assert] = ACTIONS(9208),
    [anon_sym_upcast] = ACTIONS(9208),
    [anon_sym_downcast] = ACTIONS(9208),
    [anon_sym_fixed] = ACTIONS(9208),
    [anon_sym_for] = ACTIONS(9208),
    [anon_sym_while] = ACTIONS(9208),
    [anon_sym_while_BANG] = ACTIONS(9210),
    [anon_sym_if] = ACTIONS(9208),
    [anon_sym_fun] = ACTIONS(9208),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(9208),
    [anon_sym_match] = ACTIONS(9208),
    [anon_sym_match_BANG] = ACTIONS(9210),
    [anon_sym_function] = ACTIONS(9208),
    [anon_sym_use] = ACTIONS(9208),
    [anon_sym_use_BANG] = ACTIONS(9210),
    [anon_sym_do_BANG] = ACTIONS(9210),
    [anon_sym_AMP_AMP] = ACTIONS(9210),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(9208),
    [anon_sym_static] = ACTIONS(9208),
    [anon_sym_member] = ACTIONS(9208),
    [anon_sym_exception] = ACTIONS(9208),
    [anon_sym_interface] = ACTIONS(9208),
    [anon_sym_abstract] = ACTIONS(9208),
    [anon_sym_val] = ACTIONS(9208),
    [anon_sym_override] = ACTIONS(9208),
    [anon_sym_extern] = ACTIONS(9208),
    [aux_sym_char_token1] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9208),
    [anon_sym_DQUOTE] = ACTIONS(9208),
    [anon_sym_AT_DQUOTE] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [sym_bool] = ACTIONS(9208),
    [sym_unit] = ACTIONS(9210),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9208),
    [sym_op_identifier] = ACTIONS(9210),
    [anon_sym_PLUS] = ACTIONS(9208),
    [anon_sym_DASH] = ACTIONS(9208),
    [anon_sym_PLUS_DOT] = ACTIONS(9210),
    [anon_sym_DASH_DOT] = ACTIONS(9210),
    [anon_sym_PERCENT] = ACTIONS(9208),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9210),
    [anon_sym_TILDE] = ACTIONS(9210),
    [aux_sym_prefix_op_token1] = ACTIONS(9210),
    [sym_int] = ACTIONS(9208),
    [sym_xint] = ACTIONS(9210),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9210),
    [anon_sym_POUNDload] = ACTIONS(9210),
    [anon_sym_POUNDtime] = ACTIONS(9210),
    [anon_sym_POUNDI] = ACTIONS(9210),
    [anon_sym_POUNDhelp] = ACTIONS(9210),
    [anon_sym_POUNDquit] = ACTIONS(9210),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9210),
    [anon_sym_POUNDendif] = ACTIONS(9210),
    [anon_sym_POUNDelse] = ACTIONS(9210),
    [sym__newline] = ACTIONS(9210),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9210),
  },
  [STATE(4817)] = {
    [sym_block_comment] = STATE(4817),
    [sym_compiler_directive_decl] = STATE(4817),
    [sym_fsi_directive_decl] = STATE(4817),
    [sym_preproc_line] = STATE(4817),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_as] = ACTIONS(7155),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_DOT_DOT] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__dedent] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(4818)] = {
    [sym_block_comment] = STATE(4818),
    [sym_compiler_directive_decl] = STATE(4818),
    [sym_fsi_directive_decl] = STATE(4818),
    [sym_preproc_line] = STATE(4818),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_as] = ACTIONS(7159),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_DOT_DOT] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__dedent] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(4819)] = {
    [sym_block_comment] = STATE(4819),
    [sym_compiler_directive_decl] = STATE(4819),
    [sym_fsi_directive_decl] = STATE(4819),
    [sym_preproc_line] = STATE(4819),
    [aux_sym_sequential_expression_repeat1] = STATE(4819),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_DOT_DOT] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9212),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(4820)] = {
    [sym_block_comment] = STATE(4820),
    [sym_compiler_directive_decl] = STATE(4820),
    [sym_fsi_directive_decl] = STATE(4820),
    [sym_preproc_line] = STATE(4820),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7895),
    [aux_sym_decimal_token1] = ACTIONS(7494),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4821)] = {
    [sym_block_comment] = STATE(4821),
    [sym_compiler_directive_decl] = STATE(4821),
    [sym_fsi_directive_decl] = STATE(4821),
    [sym_preproc_line] = STATE(4821),
    [aux_sym_long_identifier_repeat1] = STATE(4822),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9215),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4822)] = {
    [sym_block_comment] = STATE(4822),
    [sym_compiler_directive_decl] = STATE(4822),
    [sym_fsi_directive_decl] = STATE(4822),
    [sym_preproc_line] = STATE(4822),
    [aux_sym_long_identifier_repeat1] = STATE(4822),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9217),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4823)] = {
    [sym_block_comment] = STATE(4823),
    [sym_compiler_directive_decl] = STATE(4823),
    [sym_fsi_directive_decl] = STATE(4823),
    [sym_preproc_line] = STATE(4823),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__dedent] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
    [sym__in] = ACTIONS(6562),
  },
  [STATE(4824)] = {
    [sym_block_comment] = STATE(4824),
    [sym_compiler_directive_decl] = STATE(4824),
    [sym_fsi_directive_decl] = STATE(4824),
    [sym_preproc_line] = STATE(4824),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9054),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6902),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_DOT_DOT2] = ACTIONS(6599),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4825)] = {
    [sym_block_comment] = STATE(4825),
    [sym_compiler_directive_decl] = STATE(4825),
    [sym_fsi_directive_decl] = STATE(4825),
    [sym_preproc_line] = STATE(4825),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4826)] = {
    [sym_block_comment] = STATE(4826),
    [sym_compiler_directive_decl] = STATE(4826),
    [sym_fsi_directive_decl] = STATE(4826),
    [sym_preproc_line] = STATE(4826),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9220),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_DOT_DOT] = ACTIONS(6768),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [anon_sym_DOT_DOT2] = ACTIONS(6766),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__dedent] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(4827)] = {
    [sym_type_arguments] = STATE(5930),
    [sym__object_members] = STATE(6928),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4827),
    [sym_compiler_directive_decl] = STATE(4827),
    [sym_fsi_directive_decl] = STATE(4827),
    [sym_preproc_line] = STATE(4827),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(9224),
    [anon_sym_null] = ACTIONS(9224),
    [anon_sym_begin] = ACTIONS(9224),
    [anon_sym_open] = ACTIONS(9224),
    [anon_sym_type] = ACTIONS(9224),
    [anon_sym_LBRACK_LT] = ACTIONS(9226),
    [anon_sym_do] = ACTIONS(9224),
    [anon_sym_and] = ACTIONS(9224),
    [anon_sym_let] = ACTIONS(9224),
    [anon_sym_let_BANG] = ACTIONS(9226),
    [aux_sym_access_modifier_token1] = ACTIONS(9226),
    [anon_sym_LPAREN] = ACTIONS(9224),
    [anon_sym_AMP] = ACTIONS(9224),
    [anon_sym_LBRACK] = ACTIONS(9224),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9226),
    [anon_sym_LBRACE] = ACTIONS(9224),
    [anon_sym_LT_AT] = ACTIONS(9224),
    [anon_sym_LT_AT_AT] = ACTIONS(9226),
    [anon_sym_CARET] = ACTIONS(9226),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9226),
    [anon_sym_new] = ACTIONS(9224),
    [anon_sym_with] = ACTIONS(9228),
    [anon_sym_return] = ACTIONS(9224),
    [anon_sym_return_BANG] = ACTIONS(9226),
    [anon_sym_yield] = ACTIONS(9224),
    [anon_sym_yield_BANG] = ACTIONS(9226),
    [anon_sym_lazy] = ACTIONS(9224),
    [anon_sym_assert] = ACTIONS(9224),
    [anon_sym_upcast] = ACTIONS(9224),
    [anon_sym_downcast] = ACTIONS(9224),
    [anon_sym_fixed] = ACTIONS(9224),
    [anon_sym_for] = ACTIONS(9224),
    [anon_sym_while] = ACTIONS(9224),
    [anon_sym_while_BANG] = ACTIONS(9226),
    [anon_sym_if] = ACTIONS(9224),
    [anon_sym_fun] = ACTIONS(9224),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(9224),
    [anon_sym_match] = ACTIONS(9224),
    [anon_sym_match_BANG] = ACTIONS(9226),
    [anon_sym_function] = ACTIONS(9224),
    [anon_sym_use] = ACTIONS(9224),
    [anon_sym_use_BANG] = ACTIONS(9226),
    [anon_sym_do_BANG] = ACTIONS(9226),
    [anon_sym_AMP_AMP] = ACTIONS(9226),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(9224),
    [anon_sym_static] = ACTIONS(9224),
    [anon_sym_member] = ACTIONS(9224),
    [anon_sym_exception] = ACTIONS(9224),
    [anon_sym_interface] = ACTIONS(9224),
    [anon_sym_abstract] = ACTIONS(9224),
    [anon_sym_val] = ACTIONS(9224),
    [anon_sym_override] = ACTIONS(9224),
    [anon_sym_extern] = ACTIONS(9224),
    [aux_sym_char_token1] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9224),
    [anon_sym_DQUOTE] = ACTIONS(9224),
    [anon_sym_AT_DQUOTE] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [sym_bool] = ACTIONS(9224),
    [sym_unit] = ACTIONS(9226),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9224),
    [sym_op_identifier] = ACTIONS(9226),
    [anon_sym_PLUS] = ACTIONS(9224),
    [anon_sym_DASH] = ACTIONS(9224),
    [anon_sym_PLUS_DOT] = ACTIONS(9226),
    [anon_sym_DASH_DOT] = ACTIONS(9226),
    [anon_sym_PERCENT] = ACTIONS(9224),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9226),
    [anon_sym_TILDE] = ACTIONS(9226),
    [aux_sym_prefix_op_token1] = ACTIONS(9226),
    [sym_int] = ACTIONS(9224),
    [sym_xint] = ACTIONS(9226),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9226),
    [anon_sym_POUNDload] = ACTIONS(9226),
    [anon_sym_POUNDtime] = ACTIONS(9226),
    [anon_sym_POUNDI] = ACTIONS(9226),
    [anon_sym_POUNDhelp] = ACTIONS(9226),
    [anon_sym_POUNDquit] = ACTIONS(9226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9226),
    [anon_sym_POUNDendif] = ACTIONS(9226),
    [anon_sym_POUNDelse] = ACTIONS(9226),
    [sym__newline] = ACTIONS(9226),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9226),
  },
  [STATE(4828)] = {
    [sym_block_comment] = STATE(4828),
    [sym_compiler_directive_decl] = STATE(4828),
    [sym_fsi_directive_decl] = STATE(4828),
    [sym_preproc_line] = STATE(4828),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6834),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_DOT_DOT2] = ACTIONS(6832),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4829)] = {
    [sym_block_comment] = STATE(4829),
    [sym_compiler_directive_decl] = STATE(4829),
    [sym_fsi_directive_decl] = STATE(4829),
    [sym_preproc_line] = STATE(4829),
    [aux_sym_type_argument_repeat1] = STATE(4832),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_PIPE] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6306),
    [anon_sym_CARET] = ACTIONS(6306),
    [anon_sym_or] = ACTIONS(9230),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6306),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_AMP_AMP] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6306),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6306),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6306),
    [anon_sym_DASH_DOT] = ACTIONS(6306),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6306),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6306),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
  },
  [STATE(4830)] = {
    [sym_block_comment] = STATE(4830),
    [sym_compiler_directive_decl] = STATE(4830),
    [sym_fsi_directive_decl] = STATE(4830),
    [sym_preproc_line] = STATE(4830),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7945),
    [aux_sym_decimal_token1] = ACTIONS(7466),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4831)] = {
    [sym_block_comment] = STATE(4831),
    [sym_compiler_directive_decl] = STATE(4831),
    [sym_fsi_directive_decl] = STATE(4831),
    [sym_preproc_line] = STATE(4831),
    [aux_sym_long_identifier_repeat1] = STATE(4763),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9232),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_DASH_GT] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4832)] = {
    [sym_block_comment] = STATE(4832),
    [sym_compiler_directive_decl] = STATE(4832),
    [sym_fsi_directive_decl] = STATE(4832),
    [sym_preproc_line] = STATE(4832),
    [aux_sym_type_argument_repeat1] = STATE(4832),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(9236),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(4833)] = {
    [sym_block_comment] = STATE(4833),
    [sym_compiler_directive_decl] = STATE(4833),
    [sym_fsi_directive_decl] = STATE(4833),
    [sym_preproc_line] = STATE(4833),
    [aux_sym_long_identifier_repeat1] = STATE(4839),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9239),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [anon_sym_EQ2] = ACTIONS(6358),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4834)] = {
    [sym_block_comment] = STATE(4834),
    [sym_compiler_directive_decl] = STATE(4834),
    [sym_fsi_directive_decl] = STATE(4834),
    [sym_preproc_line] = STATE(4834),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9241),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [anon_sym_POUNDelse] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4835)] = {
    [sym_block_comment] = STATE(4835),
    [sym_compiler_directive_decl] = STATE(4835),
    [sym_fsi_directive_decl] = STATE(4835),
    [sym_preproc_line] = STATE(4835),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9243),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__multi_dollar_interp_end] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4836)] = {
    [sym_block_comment] = STATE(4836),
    [sym_compiler_directive_decl] = STATE(4836),
    [sym_fsi_directive_decl] = STATE(4836),
    [sym_preproc_line] = STATE(4836),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__dedent] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
    [sym__in] = ACTIONS(6556),
  },
  [STATE(4837)] = {
    [sym_block_comment] = STATE(4837),
    [sym_compiler_directive_decl] = STATE(4837),
    [sym_fsi_directive_decl] = STATE(4837),
    [sym_preproc_line] = STATE(4837),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_GT] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9245),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4838)] = {
    [sym_block_comment] = STATE(4838),
    [sym_compiler_directive_decl] = STATE(4838),
    [sym_fsi_directive_decl] = STATE(4838),
    [sym_preproc_line] = STATE(4838),
    [aux_sym_rules_repeat1] = STATE(4845),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9247),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9249),
    [sym__dedent] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
    [sym__in] = ACTIONS(6588),
  },
  [STATE(4839)] = {
    [sym_block_comment] = STATE(4839),
    [sym_compiler_directive_decl] = STATE(4839),
    [sym_fsi_directive_decl] = STATE(4839),
    [sym_preproc_line] = STATE(4839),
    [aux_sym_long_identifier_repeat1] = STATE(4839),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9252),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_EQ2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4840)] = {
    [sym_block_comment] = STATE(4840),
    [sym_compiler_directive_decl] = STATE(4840),
    [sym_fsi_directive_decl] = STATE(4840),
    [sym_preproc_line] = STATE(4840),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_as] = ACTIONS(6560),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4841)] = {
    [sym_block_comment] = STATE(4841),
    [sym_compiler_directive_decl] = STATE(4841),
    [sym_fsi_directive_decl] = STATE(4841),
    [sym_preproc_line] = STATE(4841),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [anon_sym_POUNDelse] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4842)] = {
    [sym_block_comment] = STATE(4842),
    [sym_compiler_directive_decl] = STATE(4842),
    [sym_fsi_directive_decl] = STATE(4842),
    [sym_preproc_line] = STATE(4842),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_as] = ACTIONS(6564),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4843)] = {
    [sym_block_comment] = STATE(4843),
    [sym_compiler_directive_decl] = STATE(4843),
    [sym_fsi_directive_decl] = STATE(4843),
    [sym_preproc_line] = STATE(4843),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_DOT_DOT] = ACTIONS(6840),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [anon_sym_DOT_DOT2] = ACTIONS(6838),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__dedent] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(4844)] = {
    [sym_block_comment] = STATE(4844),
    [sym_compiler_directive_decl] = STATE(4844),
    [sym_fsi_directive_decl] = STATE(4844),
    [sym_preproc_line] = STATE(4844),
    [aux_sym_rules_repeat1] = STATE(4850),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9247),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9255),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
    [sym__in] = ACTIONS(6605),
  },
  [STATE(4845)] = {
    [sym_block_comment] = STATE(4845),
    [sym_compiler_directive_decl] = STATE(4845),
    [sym_fsi_directive_decl] = STATE(4845),
    [sym_preproc_line] = STATE(4845),
    [aux_sym_rules_repeat1] = STATE(4853),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9247),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9255),
    [sym__dedent] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
    [sym__in] = ACTIONS(6605),
  },
  [STATE(4846)] = {
    [sym_block_comment] = STATE(4846),
    [sym_compiler_directive_decl] = STATE(4846),
    [sym_fsi_directive_decl] = STATE(4846),
    [sym_preproc_line] = STATE(4846),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6653),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [anon_sym_DOT_DOT2] = ACTIONS(6651),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4847)] = {
    [sym_block_comment] = STATE(4847),
    [sym_compiler_directive_decl] = STATE(4847),
    [sym_fsi_directive_decl] = STATE(4847),
    [sym_preproc_line] = STATE(4847),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_DOT_DOT] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9258),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4848)] = {
    [sym_block_comment] = STATE(4848),
    [sym_compiler_directive_decl] = STATE(4848),
    [sym_fsi_directive_decl] = STATE(4848),
    [sym_preproc_line] = STATE(4848),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_DASH_GT] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_DOT_DOT] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4849)] = {
    [sym_block_comment] = STATE(4849),
    [sym_compiler_directive_decl] = STATE(4849),
    [sym_fsi_directive_decl] = STATE(4849),
    [sym_preproc_line] = STATE(4849),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
    [sym__in] = ACTIONS(6614),
  },
  [STATE(4850)] = {
    [sym_block_comment] = STATE(4850),
    [sym_compiler_directive_decl] = STATE(4850),
    [sym_fsi_directive_decl] = STATE(4850),
    [sym_preproc_line] = STATE(4850),
    [aux_sym_rules_repeat1] = STATE(4853),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9247),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9260),
    [sym__dedent] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
    [sym__in] = ACTIONS(6616),
  },
  [STATE(4851)] = {
    [sym_block_comment] = STATE(4851),
    [sym_compiler_directive_decl] = STATE(4851),
    [sym_fsi_directive_decl] = STATE(4851),
    [sym_preproc_line] = STATE(4851),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_as] = ACTIONS(6513),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4852)] = {
    [sym_block_comment] = STATE(4852),
    [sym_compiler_directive_decl] = STATE(4852),
    [sym_fsi_directive_decl] = STATE(4852),
    [sym_preproc_line] = STATE(4852),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_DOT_DOT] = ACTIONS(6813),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [anon_sym_DOT_DOT2] = ACTIONS(6811),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__dedent] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(4853)] = {
    [sym_block_comment] = STATE(4853),
    [sym_compiler_directive_decl] = STATE(4853),
    [sym_fsi_directive_decl] = STATE(4853),
    [sym_preproc_line] = STATE(4853),
    [aux_sym_rules_repeat1] = STATE(4853),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9263),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9266),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
    [sym__in] = ACTIONS(6651),
  },
  [STATE(4854)] = {
    [sym_block_comment] = STATE(4854),
    [sym_compiler_directive_decl] = STATE(4854),
    [sym_fsi_directive_decl] = STATE(4854),
    [sym_preproc_line] = STATE(4854),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7859),
    [aux_sym_decimal_token1] = ACTIONS(7401),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4855)] = {
    [sym_block_comment] = STATE(4855),
    [sym_compiler_directive_decl] = STATE(4855),
    [sym_fsi_directive_decl] = STATE(4855),
    [sym_preproc_line] = STATE(4855),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_DOT_DOT] = ACTIONS(6824),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [anon_sym_DOT_DOT2] = ACTIONS(6822),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__dedent] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(4856)] = {
    [sym_block_comment] = STATE(4856),
    [sym_compiler_directive_decl] = STATE(4856),
    [sym_fsi_directive_decl] = STATE(4856),
    [sym_preproc_line] = STATE(4856),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_as] = ACTIONS(6554),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4857)] = {
    [sym_block_comment] = STATE(4857),
    [sym_compiler_directive_decl] = STATE(4857),
    [sym_fsi_directive_decl] = STATE(4857),
    [sym_preproc_line] = STATE(4857),
    [aux_sym_type_argument_repeat1] = STATE(4829),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_PIPE] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6333),
    [anon_sym_CARET] = ACTIONS(6333),
    [anon_sym_or] = ACTIONS(9230),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6333),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_AMP_AMP] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6333),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6333),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6333),
    [anon_sym_DASH_DOT] = ACTIONS(6333),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6333),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6333),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
  },
  [STATE(4858)] = {
    [sym_block_comment] = STATE(4858),
    [sym_compiler_directive_decl] = STATE(4858),
    [sym_fsi_directive_decl] = STATE(4858),
    [sym_preproc_line] = STATE(4858),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4859)] = {
    [sym_block_comment] = STATE(4859),
    [sym_compiler_directive_decl] = STATE(4859),
    [sym_fsi_directive_decl] = STATE(4859),
    [sym_preproc_line] = STATE(4859),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_DOT_DOT] = ACTIONS(6863),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [anon_sym_DOT_DOT2] = ACTIONS(6861),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__dedent] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(4860)] = {
    [sym_block_comment] = STATE(4860),
    [sym_compiler_directive_decl] = STATE(4860),
    [sym_fsi_directive_decl] = STATE(4860),
    [sym_preproc_line] = STATE(4860),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9269),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4861)] = {
    [sym_block_comment] = STATE(4861),
    [sym_compiler_directive_decl] = STATE(4861),
    [sym_fsi_directive_decl] = STATE(4861),
    [sym_preproc_line] = STATE(4861),
    [aux_sym_long_identifier_repeat1] = STATE(4863),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9271),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [anon_sym_POUNDelse] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4862)] = {
    [sym_block_comment] = STATE(4862),
    [sym_compiler_directive_decl] = STATE(4862),
    [sym_fsi_directive_decl] = STATE(4862),
    [sym_preproc_line] = STATE(4862),
    [aux_sym_long_identifier_repeat1] = STATE(4821),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9275),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__dedent] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4863)] = {
    [sym_block_comment] = STATE(4863),
    [sym_compiler_directive_decl] = STATE(4863),
    [sym_fsi_directive_decl] = STATE(4863),
    [sym_preproc_line] = STATE(4863),
    [aux_sym_long_identifier_repeat1] = STATE(4865),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9279),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4864)] = {
    [sym_block_comment] = STATE(4864),
    [sym_compiler_directive_decl] = STATE(4864),
    [sym_fsi_directive_decl] = STATE(4864),
    [sym_preproc_line] = STATE(4864),
    [aux_sym_sequential_expression_repeat1] = STATE(4771),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_DOT_DOT] = ACTIONS(6740),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_DOT_DOT2] = ACTIONS(6738),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(4865)] = {
    [sym_block_comment] = STATE(4865),
    [sym_compiler_directive_decl] = STATE(4865),
    [sym_fsi_directive_decl] = STATE(4865),
    [sym_preproc_line] = STATE(4865),
    [aux_sym_long_identifier_repeat1] = STATE(4865),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9281),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4866)] = {
    [sym_block_comment] = STATE(4866),
    [sym_compiler_directive_decl] = STATE(4866),
    [sym_fsi_directive_decl] = STATE(4866),
    [sym_preproc_line] = STATE(4866),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6834),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_DOT_DOT2] = ACTIONS(6832),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4867)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4867),
    [sym_compiler_directive_decl] = STATE(4867),
    [sym_fsi_directive_decl] = STATE(4867),
    [sym_preproc_line] = STATE(4867),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_PIPE] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6255),
    [anon_sym_CARET] = ACTIONS(6255),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_AMP_AMP] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6255),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6255),
    [anon_sym_DASH_DOT] = ACTIONS(6255),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6255),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6255),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
  },
  [STATE(4868)] = {
    [sym_block_comment] = STATE(4868),
    [sym_compiler_directive_decl] = STATE(4868),
    [sym_fsi_directive_decl] = STATE(4868),
    [sym_preproc_line] = STATE(4868),
    [aux_sym_rules_repeat1] = STATE(4878),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9284),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_DASH_GT] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_DOT_DOT] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9286),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(4869)] = {
    [sym_block_comment] = STATE(4869),
    [sym_compiler_directive_decl] = STATE(4869),
    [sym_fsi_directive_decl] = STATE(4869),
    [sym_preproc_line] = STATE(4869),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6295),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_DOT_DOT2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4870)] = {
    [sym_block_comment] = STATE(4870),
    [sym_compiler_directive_decl] = STATE(4870),
    [sym_fsi_directive_decl] = STATE(4870),
    [sym_preproc_line] = STATE(4870),
    [aux_sym_long_identifier_repeat1] = STATE(4870),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9289),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
    [sym__in] = ACTIONS(6297),
  },
  [STATE(4871)] = {
    [sym_block_comment] = STATE(4871),
    [sym_compiler_directive_decl] = STATE(4871),
    [sym_fsi_directive_decl] = STATE(4871),
    [sym_preproc_line] = STATE(4871),
    [aux_sym_rules_repeat1] = STATE(4882),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_as] = ACTIONS(6590),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9292),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9294),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(4872)] = {
    [sym_block_comment] = STATE(4872),
    [sym_compiler_directive_decl] = STATE(4872),
    [sym_fsi_directive_decl] = STATE(4872),
    [sym_preproc_line] = STATE(4872),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8004),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4873)] = {
    [sym_block_comment] = STATE(4873),
    [sym_compiler_directive_decl] = STATE(4873),
    [sym_fsi_directive_decl] = STATE(4873),
    [sym_preproc_line] = STATE(4873),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_DOT_DOT] = ACTIONS(6867),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [anon_sym_DOT_DOT2] = ACTIONS(6865),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__dedent] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(4874)] = {
    [sym_block_comment] = STATE(4874),
    [sym_compiler_directive_decl] = STATE(4874),
    [sym_fsi_directive_decl] = STATE(4874),
    [sym_preproc_line] = STATE(4874),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_DOT_DOT] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(4875)] = {
    [sym_block_comment] = STATE(4875),
    [sym_compiler_directive_decl] = STATE(4875),
    [sym_fsi_directive_decl] = STATE(4875),
    [sym_preproc_line] = STATE(4875),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_DASH_GT] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4876)] = {
    [sym_block_comment] = STATE(4876),
    [sym_compiler_directive_decl] = STATE(4876),
    [sym_fsi_directive_decl] = STATE(4876),
    [sym_preproc_line] = STATE(4876),
    [aux_sym_rules_repeat1] = STATE(4880),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9284),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_DASH_GT] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4877)] = {
    [sym_block_comment] = STATE(4877),
    [sym_compiler_directive_decl] = STATE(4877),
    [sym_fsi_directive_decl] = STATE(4877),
    [sym_preproc_line] = STATE(4877),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_DOT_DOT] = ACTIONS(6692),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [anon_sym_DOT_DOT2] = ACTIONS(6694),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__dedent] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(4878)] = {
    [sym_block_comment] = STATE(4878),
    [sym_compiler_directive_decl] = STATE(4878),
    [sym_fsi_directive_decl] = STATE(4878),
    [sym_preproc_line] = STATE(4878),
    [aux_sym_rules_repeat1] = STATE(4881),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9284),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_DASH_GT] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4879)] = {
    [sym_block_comment] = STATE(4879),
    [sym_compiler_directive_decl] = STATE(4879),
    [sym_fsi_directive_decl] = STATE(4879),
    [sym_preproc_line] = STATE(4879),
    [aux_sym_rules_repeat1] = STATE(4889),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_as] = ACTIONS(6607),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9292),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4880)] = {
    [sym_block_comment] = STATE(4880),
    [sym_compiler_directive_decl] = STATE(4880),
    [sym_fsi_directive_decl] = STATE(4880),
    [sym_preproc_line] = STATE(4880),
    [aux_sym_rules_repeat1] = STATE(4881),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9284),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_DASH_GT] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_DOT_DOT] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9303),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4881)] = {
    [sym_block_comment] = STATE(4881),
    [sym_compiler_directive_decl] = STATE(4881),
    [sym_fsi_directive_decl] = STATE(4881),
    [sym_preproc_line] = STATE(4881),
    [aux_sym_rules_repeat1] = STATE(4881),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9306),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_DASH_GT] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9309),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4882)] = {
    [sym_block_comment] = STATE(4882),
    [sym_compiler_directive_decl] = STATE(4882),
    [sym_fsi_directive_decl] = STATE(4882),
    [sym_preproc_line] = STATE(4882),
    [aux_sym_rules_repeat1] = STATE(4893),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_as] = ACTIONS(6607),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9292),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9300),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4883)] = {
    [sym_block_comment] = STATE(4883),
    [sym_compiler_directive_decl] = STATE(4883),
    [sym_fsi_directive_decl] = STATE(4883),
    [sym_preproc_line] = STATE(4883),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_DOT_DOT] = ACTIONS(6871),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [anon_sym_DOT_DOT2] = ACTIONS(6869),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__dedent] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(4884)] = {
    [sym_block_comment] = STATE(4884),
    [sym_compiler_directive_decl] = STATE(4884),
    [sym_fsi_directive_decl] = STATE(4884),
    [sym_preproc_line] = STATE(4884),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_as] = ACTIONS(6807),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4885)] = {
    [sym_block_comment] = STATE(4885),
    [sym_compiler_directive_decl] = STATE(4885),
    [sym_fsi_directive_decl] = STATE(4885),
    [sym_preproc_line] = STATE(4885),
    [aux_sym_long_identifier_repeat1] = STATE(4803),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_as] = ACTIONS(6625),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4886)] = {
    [sym_block_comment] = STATE(4886),
    [sym_compiler_directive_decl] = STATE(4886),
    [sym_fsi_directive_decl] = STATE(4886),
    [sym_preproc_line] = STATE(4886),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [anon_sym_POUNDelse] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4887)] = {
    [sym_block_comment] = STATE(4887),
    [sym_compiler_directive_decl] = STATE(4887),
    [sym_fsi_directive_decl] = STATE(4887),
    [sym_preproc_line] = STATE(4887),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_as] = ACTIONS(6892),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_DOT_DOT] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__dedent] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(4888)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4888),
    [sym_compiler_directive_decl] = STATE(4888),
    [sym_fsi_directive_decl] = STATE(4888),
    [sym_preproc_line] = STATE(4888),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_PIPE] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6209),
    [anon_sym_CARET] = ACTIONS(6209),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_AMP_AMP] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6209),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6209),
    [anon_sym_DASH_DOT] = ACTIONS(6209),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6209),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6209),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
  },
  [STATE(4889)] = {
    [sym_block_comment] = STATE(4889),
    [sym_compiler_directive_decl] = STATE(4889),
    [sym_fsi_directive_decl] = STATE(4889),
    [sym_preproc_line] = STATE(4889),
    [aux_sym_rules_repeat1] = STATE(4893),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_as] = ACTIONS(6618),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9292),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9316),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4890)] = {
    [sym_block_comment] = STATE(4890),
    [sym_compiler_directive_decl] = STATE(4890),
    [sym_fsi_directive_decl] = STATE(4890),
    [sym_preproc_line] = STATE(4890),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_as] = ACTIONS(6896),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_DOT_DOT] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__dedent] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(4891)] = {
    [sym_block_comment] = STATE(4891),
    [sym_compiler_directive_decl] = STATE(4891),
    [sym_fsi_directive_decl] = STATE(4891),
    [sym_preproc_line] = STATE(4891),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_as] = ACTIONS(6900),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_DOT_DOT] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__dedent] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(4892)] = {
    [sym_block_comment] = STATE(4892),
    [sym_compiler_directive_decl] = STATE(4892),
    [sym_fsi_directive_decl] = STATE(4892),
    [sym_preproc_line] = STATE(4892),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_as] = ACTIONS(6902),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__dedent] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4893)] = {
    [sym_block_comment] = STATE(4893),
    [sym_compiler_directive_decl] = STATE(4893),
    [sym_fsi_directive_decl] = STATE(4893),
    [sym_preproc_line] = STATE(4893),
    [aux_sym_rules_repeat1] = STATE(4893),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_as] = ACTIONS(6653),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9319),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9322),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4894)] = {
    [sym_block_comment] = STATE(4894),
    [sym_compiler_directive_decl] = STATE(4894),
    [sym_fsi_directive_decl] = STATE(4894),
    [sym_preproc_line] = STATE(4894),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_DOT_DOT] = ACTIONS(6875),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [anon_sym_DOT_DOT2] = ACTIONS(6873),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__dedent] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(4895)] = {
    [sym_block_comment] = STATE(4895),
    [sym_compiler_directive_decl] = STATE(4895),
    [sym_fsi_directive_decl] = STATE(4895),
    [sym_preproc_line] = STATE(4895),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_as] = ACTIONS(6906),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_DOT_DOT] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__dedent] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(4896)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4896),
    [sym_compiler_directive_decl] = STATE(4896),
    [sym_fsi_directive_decl] = STATE(4896),
    [sym_preproc_line] = STATE(4896),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_PIPE] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6195),
    [anon_sym_CARET] = ACTIONS(6195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_AMP_AMP] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6195),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6195),
    [anon_sym_DASH_DOT] = ACTIONS(6195),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6195),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6195),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
  },
  [STATE(4897)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4897),
    [sym_compiler_directive_decl] = STATE(4897),
    [sym_fsi_directive_decl] = STATE(4897),
    [sym_preproc_line] = STATE(4897),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_PIPE] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6259),
    [anon_sym_CARET] = ACTIONS(6259),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_AMP_AMP] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6259),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6259),
    [anon_sym_DASH_DOT] = ACTIONS(6259),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6259),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6259),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
  },
  [STATE(4898)] = {
    [sym_block_comment] = STATE(4898),
    [sym_compiler_directive_decl] = STATE(4898),
    [sym_fsi_directive_decl] = STATE(4898),
    [sym_preproc_line] = STATE(4898),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_as] = ACTIONS(6913),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_DOT_DOT] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__dedent] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(4899)] = {
    [sym_block_comment] = STATE(4899),
    [sym_compiler_directive_decl] = STATE(4899),
    [sym_fsi_directive_decl] = STATE(4899),
    [sym_preproc_line] = STATE(4899),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_as] = ACTIONS(6917),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_DOT_DOT] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__dedent] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(4900)] = {
    [sym_block_comment] = STATE(4900),
    [sym_compiler_directive_decl] = STATE(4900),
    [sym_fsi_directive_decl] = STATE(4900),
    [sym_preproc_line] = STATE(4900),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_as] = ACTIONS(6921),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_DOT_DOT] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__dedent] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(4901)] = {
    [sym_block_comment] = STATE(4901),
    [sym_compiler_directive_decl] = STATE(4901),
    [sym_fsi_directive_decl] = STATE(4901),
    [sym_preproc_line] = STATE(4901),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_as] = ACTIONS(6925),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_DOT_DOT] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__dedent] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(4902)] = {
    [sym_block_comment] = STATE(4902),
    [sym_compiler_directive_decl] = STATE(4902),
    [sym_fsi_directive_decl] = STATE(4902),
    [sym_preproc_line] = STATE(4902),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_RBRACK] = ACTIONS(6599),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(8624),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6902),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [anon_sym_DOT_DOT2] = ACTIONS(6599),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4903)] = {
    [sym_block_comment] = STATE(4903),
    [sym_compiler_directive_decl] = STATE(4903),
    [sym_fsi_directive_decl] = STATE(4903),
    [sym_preproc_line] = STATE(4903),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_DOT_DOT] = ACTIONS(6828),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [anon_sym_DOT_DOT2] = ACTIONS(6830),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__dedent] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(4904)] = {
    [sym_block_comment] = STATE(4904),
    [sym_compiler_directive_decl] = STATE(4904),
    [sym_fsi_directive_decl] = STATE(4904),
    [sym_preproc_line] = STATE(4904),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_as] = ACTIONS(290),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(4905)] = {
    [sym_block_comment] = STATE(4905),
    [sym_compiler_directive_decl] = STATE(4905),
    [sym_fsi_directive_decl] = STATE(4905),
    [sym_preproc_line] = STATE(4905),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_as] = ACTIONS(6932),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_DOT_DOT] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__dedent] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(4906)] = {
    [sym_block_comment] = STATE(4906),
    [sym_compiler_directive_decl] = STATE(4906),
    [sym_fsi_directive_decl] = STATE(4906),
    [sym_preproc_line] = STATE(4906),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [anon_sym_POUNDelse] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4907)] = {
    [sym_block_comment] = STATE(4907),
    [sym_compiler_directive_decl] = STATE(4907),
    [sym_fsi_directive_decl] = STATE(4907),
    [sym_preproc_line] = STATE(4907),
    [aux_sym_long_identifier_repeat1] = STATE(4833),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9325),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [anon_sym_EQ2] = ACTIONS(6291),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4908)] = {
    [sym_block_comment] = STATE(4908),
    [sym_compiler_directive_decl] = STATE(4908),
    [sym_fsi_directive_decl] = STATE(4908),
    [sym_preproc_line] = STATE(4908),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_as] = ACTIONS(6612),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4909)] = {
    [sym_block_comment] = STATE(4909),
    [sym_compiler_directive_decl] = STATE(4909),
    [sym_fsi_directive_decl] = STATE(4909),
    [sym_preproc_line] = STATE(4909),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4910)] = {
    [sym_block_comment] = STATE(4910),
    [sym_compiler_directive_decl] = STATE(4910),
    [sym_fsi_directive_decl] = STATE(4910),
    [sym_preproc_line] = STATE(4910),
    [aux_sym_long_identifier_repeat1] = STATE(4870),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9329),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__dedent] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
    [sym__in] = ACTIONS(6358),
  },
  [STATE(4911)] = {
    [sym_block_comment] = STATE(4911),
    [sym_compiler_directive_decl] = STATE(4911),
    [sym_fsi_directive_decl] = STATE(4911),
    [sym_preproc_line] = STATE(4911),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7909),
    [aux_sym_decimal_token1] = ACTIONS(7436),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
    [sym__in] = ACTIONS(5953),
  },
  [STATE(4912)] = {
    [sym_block_comment] = STATE(4912),
    [sym_compiler_directive_decl] = STATE(4912),
    [sym_fsi_directive_decl] = STATE(4912),
    [sym_preproc_line] = STATE(4912),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9331),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__untyped_quoted_close] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4913)] = {
    [sym_block_comment] = STATE(4913),
    [sym_compiler_directive_decl] = STATE(4913),
    [sym_fsi_directive_decl] = STATE(4913),
    [sym_preproc_line] = STATE(4913),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9333),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__dedent] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
    [sym__in] = ACTIONS(6635),
  },
  [STATE(4914)] = {
    [sym_block_comment] = STATE(4914),
    [sym_compiler_directive_decl] = STATE(4914),
    [sym_fsi_directive_decl] = STATE(4914),
    [sym_preproc_line] = STATE(4914),
    [aux_sym_sequential_expression_repeat1] = STATE(4819),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_DOT_DOT] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(4915)] = {
    [sym_block_comment] = STATE(4915),
    [sym_compiler_directive_decl] = STATE(4915),
    [sym_fsi_directive_decl] = STATE(4915),
    [sym_preproc_line] = STATE(4915),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9335),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
    [sym__in] = ACTIONS(6599),
  },
  [STATE(4916)] = {
    [sym_block_comment] = STATE(4916),
    [sym_compiler_directive_decl] = STATE(4916),
    [sym_fsi_directive_decl] = STATE(4916),
    [sym_preproc_line] = STATE(4916),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6612),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4917)] = {
    [sym_block_comment] = STATE(4917),
    [sym_compiler_directive_decl] = STATE(4917),
    [sym_fsi_directive_decl] = STATE(4917),
    [sym_preproc_line] = STATE(4917),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__dedent] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
    [sym__in] = ACTIONS(6515),
  },
  [STATE(4918)] = {
    [sym_block_comment] = STATE(4918),
    [sym_compiler_directive_decl] = STATE(4918),
    [sym_fsi_directive_decl] = STATE(4918),
    [sym_preproc_line] = STATE(4918),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9337),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4919)] = {
    [sym_block_comment] = STATE(4919),
    [sym_compiler_directive_decl] = STATE(4919),
    [sym_fsi_directive_decl] = STATE(4919),
    [sym_preproc_line] = STATE(4919),
    [aux_sym_rules_repeat1] = STATE(4946),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9137),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [anon_sym_POUNDelse] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9339),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(4920)] = {
    [sym_block_comment] = STATE(4920),
    [sym_compiler_directive_decl] = STATE(4920),
    [sym_fsi_directive_decl] = STATE(4920),
    [sym_preproc_line] = STATE(4920),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_as] = ACTIONS(6944),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_with] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_DOT_DOT] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__dedent] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(4921)] = {
    [sym_block_comment] = STATE(4921),
    [sym_compiler_directive_decl] = STATE(4921),
    [sym_fsi_directive_decl] = STATE(4921),
    [sym_preproc_line] = STATE(4921),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_as] = ACTIONS(6948),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_with] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_DOT_DOT] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__dedent] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(4922)] = {
    [sym_block_comment] = STATE(4922),
    [sym_compiler_directive_decl] = STATE(4922),
    [sym_fsi_directive_decl] = STATE(4922),
    [sym_preproc_line] = STATE(4922),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_as] = ACTIONS(6952),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_with] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_DOT_DOT] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__dedent] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(4923)] = {
    [sym_block_comment] = STATE(4923),
    [sym_compiler_directive_decl] = STATE(4923),
    [sym_fsi_directive_decl] = STATE(4923),
    [sym_preproc_line] = STATE(4923),
    [aux_sym_long_identifier_repeat1] = STATE(5205),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9342),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9344),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(4924)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4924),
    [sym_compiler_directive_decl] = STATE(4924),
    [sym_fsi_directive_decl] = STATE(4924),
    [sym_preproc_line] = STATE(4924),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(9346),
    [anon_sym_null] = ACTIONS(9346),
    [anon_sym_begin] = ACTIONS(9346),
    [anon_sym_open] = ACTIONS(9346),
    [anon_sym_type] = ACTIONS(9346),
    [anon_sym_LBRACK_LT] = ACTIONS(9348),
    [anon_sym_do] = ACTIONS(9346),
    [anon_sym_and] = ACTIONS(9346),
    [anon_sym_let] = ACTIONS(9346),
    [anon_sym_let_BANG] = ACTIONS(9348),
    [aux_sym_access_modifier_token1] = ACTIONS(9348),
    [anon_sym_LPAREN] = ACTIONS(9346),
    [anon_sym_PIPE] = ACTIONS(9348),
    [anon_sym_AMP] = ACTIONS(9346),
    [anon_sym_LBRACK] = ACTIONS(9346),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9348),
    [anon_sym_LBRACE] = ACTIONS(9346),
    [anon_sym_LT_AT] = ACTIONS(9346),
    [anon_sym_LT_AT_AT] = ACTIONS(9348),
    [anon_sym_CARET] = ACTIONS(9348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9348),
    [anon_sym_new] = ACTIONS(9346),
    [anon_sym_with] = ACTIONS(9346),
    [anon_sym_return] = ACTIONS(9346),
    [anon_sym_return_BANG] = ACTIONS(9348),
    [anon_sym_yield] = ACTIONS(9346),
    [anon_sym_yield_BANG] = ACTIONS(9348),
    [anon_sym_lazy] = ACTIONS(9346),
    [anon_sym_assert] = ACTIONS(9346),
    [anon_sym_upcast] = ACTIONS(9346),
    [anon_sym_downcast] = ACTIONS(9346),
    [anon_sym_fixed] = ACTIONS(9346),
    [anon_sym_for] = ACTIONS(9346),
    [anon_sym_while] = ACTIONS(9346),
    [anon_sym_while_BANG] = ACTIONS(9348),
    [anon_sym_if] = ACTIONS(9346),
    [anon_sym_fun] = ACTIONS(9346),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(9346),
    [anon_sym_match] = ACTIONS(9346),
    [anon_sym_match_BANG] = ACTIONS(9348),
    [anon_sym_function] = ACTIONS(9346),
    [anon_sym_use] = ACTIONS(9346),
    [anon_sym_use_BANG] = ACTIONS(9348),
    [anon_sym_do_BANG] = ACTIONS(9348),
    [anon_sym_AMP_AMP] = ACTIONS(9348),
    [anon_sym_STAR] = ACTIONS(9348),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(9346),
    [anon_sym_static] = ACTIONS(9346),
    [anon_sym_member] = ACTIONS(9346),
    [anon_sym_exception] = ACTIONS(9346),
    [anon_sym_interface] = ACTIONS(9346),
    [anon_sym_abstract] = ACTIONS(9346),
    [anon_sym_val] = ACTIONS(9346),
    [anon_sym_override] = ACTIONS(9346),
    [anon_sym_extern] = ACTIONS(9346),
    [aux_sym_char_token1] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9346),
    [anon_sym_DQUOTE] = ACTIONS(9346),
    [anon_sym_AT_DQUOTE] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [sym_bool] = ACTIONS(9346),
    [sym_unit] = ACTIONS(9348),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9346),
    [sym_op_identifier] = ACTIONS(9348),
    [anon_sym_PLUS] = ACTIONS(9346),
    [anon_sym_DASH] = ACTIONS(9346),
    [anon_sym_PLUS_DOT] = ACTIONS(9348),
    [anon_sym_DASH_DOT] = ACTIONS(9348),
    [anon_sym_PERCENT] = ACTIONS(9346),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9348),
    [anon_sym_TILDE] = ACTIONS(9348),
    [aux_sym_prefix_op_token1] = ACTIONS(9348),
    [sym_int] = ACTIONS(9346),
    [sym_xint] = ACTIONS(9348),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9348),
    [anon_sym_POUNDload] = ACTIONS(9348),
    [anon_sym_POUNDtime] = ACTIONS(9348),
    [anon_sym_POUNDI] = ACTIONS(9348),
    [anon_sym_POUNDhelp] = ACTIONS(9348),
    [anon_sym_POUNDquit] = ACTIONS(9348),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9348),
    [anon_sym_POUNDendif] = ACTIONS(9348),
    [anon_sym_POUNDelse] = ACTIONS(9348),
    [sym__newline] = ACTIONS(9348),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9348),
  },
  [STATE(4925)] = {
    [sym_block_comment] = STATE(4925),
    [sym_compiler_directive_decl] = STATE(4925),
    [sym_fsi_directive_decl] = STATE(4925),
    [sym_preproc_line] = STATE(4925),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_as] = ACTIONS(6959),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_with] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_DOT_DOT] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__dedent] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(4926)] = {
    [sym_block_comment] = STATE(4926),
    [sym_compiler_directive_decl] = STATE(4926),
    [sym_fsi_directive_decl] = STATE(4926),
    [sym_preproc_line] = STATE(4926),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_as] = ACTIONS(6963),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_with] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_DOT_DOT] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__dedent] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(4927)] = {
    [sym_block_comment] = STATE(4927),
    [sym_compiler_directive_decl] = STATE(4927),
    [sym_fsi_directive_decl] = STATE(4927),
    [sym_preproc_line] = STATE(4927),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_as] = ACTIONS(6967),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_with] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_DOT_DOT] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__dedent] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(4928)] = {
    [sym_block_comment] = STATE(4928),
    [sym_compiler_directive_decl] = STATE(4928),
    [sym_fsi_directive_decl] = STATE(4928),
    [sym_preproc_line] = STATE(4928),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_as] = ACTIONS(6971),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_with] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_DOT_DOT] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__dedent] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(4929)] = {
    [sym_block_comment] = STATE(4929),
    [sym_compiler_directive_decl] = STATE(4929),
    [sym_fsi_directive_decl] = STATE(4929),
    [sym_preproc_line] = STATE(4929),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [aux_sym_float_token1] = ACTIONS(9350),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4930)] = {
    [sym_block_comment] = STATE(4930),
    [sym_compiler_directive_decl] = STATE(4930),
    [sym_fsi_directive_decl] = STATE(4930),
    [sym_preproc_line] = STATE(4930),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_as] = ACTIONS(6975),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_with] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_DOT_DOT] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__dedent] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(4931)] = {
    [sym_block_comment] = STATE(4931),
    [sym_compiler_directive_decl] = STATE(4931),
    [sym_fsi_directive_decl] = STATE(4931),
    [sym_preproc_line] = STATE(4931),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_as] = ACTIONS(6979),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_with] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_DOT_DOT] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__dedent] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(4932)] = {
    [sym_block_comment] = STATE(4932),
    [sym_compiler_directive_decl] = STATE(4932),
    [sym_fsi_directive_decl] = STATE(4932),
    [sym_preproc_line] = STATE(4932),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_as] = ACTIONS(6983),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_with] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_DOT_DOT] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__dedent] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(4933)] = {
    [sym_block_comment] = STATE(4933),
    [sym_compiler_directive_decl] = STATE(4933),
    [sym_fsi_directive_decl] = STATE(4933),
    [sym_preproc_line] = STATE(4933),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_as] = ACTIONS(6987),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_with] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_DOT_DOT] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__dedent] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(4934)] = {
    [sym_block_comment] = STATE(4934),
    [sym_compiler_directive_decl] = STATE(4934),
    [sym_fsi_directive_decl] = STATE(4934),
    [sym_preproc_line] = STATE(4934),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_as] = ACTIONS(6991),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_with] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_DOT_DOT] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__dedent] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(4935)] = {
    [sym_block_comment] = STATE(4935),
    [sym_compiler_directive_decl] = STATE(4935),
    [sym_fsi_directive_decl] = STATE(4935),
    [sym_preproc_line] = STATE(4935),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_as] = ACTIONS(6995),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_with] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_DOT_DOT] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__dedent] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(4936)] = {
    [sym_block_comment] = STATE(4936),
    [sym_compiler_directive_decl] = STATE(4936),
    [sym_fsi_directive_decl] = STATE(4936),
    [sym_preproc_line] = STATE(4936),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_as] = ACTIONS(6999),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_with] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_DOT_DOT] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__dedent] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(4937)] = {
    [sym_block_comment] = STATE(4937),
    [sym_compiler_directive_decl] = STATE(4937),
    [sym_fsi_directive_decl] = STATE(4937),
    [sym_preproc_line] = STATE(4937),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_as] = ACTIONS(7003),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_with] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_DOT_DOT] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__dedent] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(4938)] = {
    [sym_block_comment] = STATE(4938),
    [sym_compiler_directive_decl] = STATE(4938),
    [sym_fsi_directive_decl] = STATE(4938),
    [sym_preproc_line] = STATE(4938),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_as] = ACTIONS(7007),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_with] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_DOT_DOT] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__dedent] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(4939)] = {
    [sym_block_comment] = STATE(4939),
    [sym_compiler_directive_decl] = STATE(4939),
    [sym_fsi_directive_decl] = STATE(4939),
    [sym_preproc_line] = STATE(4939),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_as] = ACTIONS(7011),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_with] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_DOT_DOT] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__dedent] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(4940)] = {
    [sym_block_comment] = STATE(4940),
    [sym_compiler_directive_decl] = STATE(4940),
    [sym_fsi_directive_decl] = STATE(4940),
    [sym_preproc_line] = STATE(4940),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_as] = ACTIONS(7019),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_with] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_DOT_DOT] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__dedent] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(4941)] = {
    [sym_block_comment] = STATE(4941),
    [sym_compiler_directive_decl] = STATE(4941),
    [sym_fsi_directive_decl] = STATE(4941),
    [sym_preproc_line] = STATE(4941),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_as] = ACTIONS(7023),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_with] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_DOT_DOT] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__dedent] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(4942)] = {
    [sym_type_arguments] = STATE(5930),
    [sym_long_identifier] = STATE(5934),
    [sym_block_comment] = STATE(4942),
    [sym_compiler_directive_decl] = STATE(4942),
    [sym_fsi_directive_decl] = STATE(4942),
    [sym_preproc_line] = STATE(4942),
    [aux_sym_compound_type_repeat1] = STATE(5613),
    [sym_identifier] = ACTIONS(9222),
    [anon_sym_module] = ACTIONS(9352),
    [anon_sym_null] = ACTIONS(9352),
    [anon_sym_begin] = ACTIONS(9352),
    [anon_sym_open] = ACTIONS(9352),
    [anon_sym_type] = ACTIONS(9352),
    [anon_sym_LBRACK_LT] = ACTIONS(9354),
    [anon_sym_do] = ACTIONS(9352),
    [anon_sym_and] = ACTIONS(9352),
    [anon_sym_let] = ACTIONS(9352),
    [anon_sym_let_BANG] = ACTIONS(9354),
    [aux_sym_access_modifier_token1] = ACTIONS(9354),
    [anon_sym_LPAREN] = ACTIONS(9352),
    [anon_sym_PIPE] = ACTIONS(9354),
    [anon_sym_AMP] = ACTIONS(9352),
    [anon_sym_LBRACK] = ACTIONS(9352),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9354),
    [anon_sym_LBRACE] = ACTIONS(9352),
    [anon_sym_LT_AT] = ACTIONS(9352),
    [anon_sym_LT_AT_AT] = ACTIONS(9354),
    [anon_sym_CARET] = ACTIONS(9354),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9354),
    [anon_sym_new] = ACTIONS(9352),
    [anon_sym_with] = ACTIONS(9352),
    [anon_sym_return] = ACTIONS(9352),
    [anon_sym_return_BANG] = ACTIONS(9354),
    [anon_sym_yield] = ACTIONS(9352),
    [anon_sym_yield_BANG] = ACTIONS(9354),
    [anon_sym_lazy] = ACTIONS(9352),
    [anon_sym_assert] = ACTIONS(9352),
    [anon_sym_upcast] = ACTIONS(9352),
    [anon_sym_downcast] = ACTIONS(9352),
    [anon_sym_fixed] = ACTIONS(9352),
    [anon_sym_for] = ACTIONS(9352),
    [anon_sym_while] = ACTIONS(9352),
    [anon_sym_while_BANG] = ACTIONS(9354),
    [anon_sym_if] = ACTIONS(9352),
    [anon_sym_fun] = ACTIONS(9352),
    [anon_sym_DASH_GT] = ACTIONS(9186),
    [anon_sym_try] = ACTIONS(9352),
    [anon_sym_match] = ACTIONS(9352),
    [anon_sym_match_BANG] = ACTIONS(9354),
    [anon_sym_function] = ACTIONS(9352),
    [anon_sym_use] = ACTIONS(9352),
    [anon_sym_use_BANG] = ACTIONS(9354),
    [anon_sym_do_BANG] = ACTIONS(9354),
    [anon_sym_AMP_AMP] = ACTIONS(9354),
    [anon_sym_STAR] = ACTIONS(9354),
    [anon_sym_LT] = ACTIONS(9190),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9192),
    [anon_sym_default] = ACTIONS(9352),
    [anon_sym_static] = ACTIONS(9352),
    [anon_sym_member] = ACTIONS(9352),
    [anon_sym_exception] = ACTIONS(9352),
    [anon_sym_interface] = ACTIONS(9352),
    [anon_sym_abstract] = ACTIONS(9352),
    [anon_sym_val] = ACTIONS(9352),
    [anon_sym_override] = ACTIONS(9352),
    [anon_sym_extern] = ACTIONS(9352),
    [aux_sym_char_token1] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9352),
    [anon_sym_DQUOTE] = ACTIONS(9352),
    [anon_sym_AT_DQUOTE] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [sym_bool] = ACTIONS(9352),
    [sym_unit] = ACTIONS(9354),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9352),
    [sym_op_identifier] = ACTIONS(9354),
    [anon_sym_PLUS] = ACTIONS(9352),
    [anon_sym_DASH] = ACTIONS(9352),
    [anon_sym_PLUS_DOT] = ACTIONS(9354),
    [anon_sym_DASH_DOT] = ACTIONS(9354),
    [anon_sym_PERCENT] = ACTIONS(9352),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9354),
    [anon_sym_TILDE] = ACTIONS(9354),
    [aux_sym_prefix_op_token1] = ACTIONS(9354),
    [sym_int] = ACTIONS(9352),
    [sym_xint] = ACTIONS(9354),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9354),
    [anon_sym_POUNDload] = ACTIONS(9354),
    [anon_sym_POUNDtime] = ACTIONS(9354),
    [anon_sym_POUNDI] = ACTIONS(9354),
    [anon_sym_POUNDhelp] = ACTIONS(9354),
    [anon_sym_POUNDquit] = ACTIONS(9354),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9354),
    [anon_sym_POUNDendif] = ACTIONS(9354),
    [anon_sym_POUNDelse] = ACTIONS(9354),
    [sym__newline] = ACTIONS(9354),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9354),
  },
  [STATE(4943)] = {
    [sym_block_comment] = STATE(4943),
    [sym_compiler_directive_decl] = STATE(4943),
    [sym_fsi_directive_decl] = STATE(4943),
    [sym_preproc_line] = STATE(4943),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_as] = ACTIONS(7027),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_with] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_DOT_DOT] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__dedent] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(4944)] = {
    [sym_block_comment] = STATE(4944),
    [sym_compiler_directive_decl] = STATE(4944),
    [sym_fsi_directive_decl] = STATE(4944),
    [sym_preproc_line] = STATE(4944),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6612),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [anon_sym_DOT_DOT2] = ACTIONS(6614),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4945)] = {
    [sym_block_comment] = STATE(4945),
    [sym_compiler_directive_decl] = STATE(4945),
    [sym_fsi_directive_decl] = STATE(4945),
    [sym_preproc_line] = STATE(4945),
    [aux_sym_rules_repeat1] = STATE(4751),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9137),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9356),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4946)] = {
    [sym_block_comment] = STATE(4946),
    [sym_compiler_directive_decl] = STATE(4946),
    [sym_fsi_directive_decl] = STATE(4946),
    [sym_preproc_line] = STATE(4946),
    [aux_sym_rules_repeat1] = STATE(4752),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9137),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [anon_sym_POUNDelse] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9356),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4947)] = {
    [sym_block_comment] = STATE(4947),
    [sym_compiler_directive_decl] = STATE(4947),
    [sym_fsi_directive_decl] = STATE(4947),
    [sym_preproc_line] = STATE(4947),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_DOT_DOT] = ACTIONS(6750),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [anon_sym_DOT_DOT2] = ACTIONS(6752),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__dedent] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(4948)] = {
    [sym_block_comment] = STATE(4948),
    [sym_compiler_directive_decl] = STATE(4948),
    [sym_fsi_directive_decl] = STATE(4948),
    [sym_preproc_line] = STATE(4948),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_as] = ACTIONS(7039),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_with] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_DOT_DOT] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__dedent] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(4949)] = {
    [sym_block_comment] = STATE(4949),
    [sym_compiler_directive_decl] = STATE(4949),
    [sym_fsi_directive_decl] = STATE(4949),
    [sym_preproc_line] = STATE(4949),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_as] = ACTIONS(7043),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_with] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_DOT_DOT] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__dedent] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(4950)] = {
    [sym_block_comment] = STATE(4950),
    [sym_compiler_directive_decl] = STATE(4950),
    [sym_fsi_directive_decl] = STATE(4950),
    [sym_preproc_line] = STATE(4950),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__dedent] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
    [sym__in] = ACTIONS(6558),
  },
  [STATE(4951)] = {
    [sym_block_comment] = STATE(4951),
    [sym_compiler_directive_decl] = STATE(4951),
    [sym_fsi_directive_decl] = STATE(4951),
    [sym_preproc_line] = STATE(4951),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_as] = ACTIONS(7059),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_with] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_DOT_DOT] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__dedent] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(4952)] = {
    [sym_block_comment] = STATE(4952),
    [sym_compiler_directive_decl] = STATE(4952),
    [sym_fsi_directive_decl] = STATE(4952),
    [sym_preproc_line] = STATE(4952),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_DASH_GT] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_DOT_DOT] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4953)] = {
    [sym_block_comment] = STATE(4953),
    [sym_compiler_directive_decl] = STATE(4953),
    [sym_fsi_directive_decl] = STATE(4953),
    [sym_preproc_line] = STATE(4953),
    [aux_sym_long_identifier_repeat1] = STATE(5304),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9359),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [anon_sym_POUNDendif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(4954)] = {
    [sym_block_comment] = STATE(4954),
    [sym_compiler_directive_decl] = STATE(4954),
    [sym_fsi_directive_decl] = STATE(4954),
    [sym_preproc_line] = STATE(4954),
    [aux_sym_long_identifier_repeat1] = STATE(4954),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9363),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4955)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(4955),
    [sym_compiler_directive_decl] = STATE(4955),
    [sym_fsi_directive_decl] = STATE(4955),
    [sym_preproc_line] = STATE(4955),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(9346),
    [anon_sym_null] = ACTIONS(9346),
    [anon_sym_begin] = ACTIONS(9346),
    [anon_sym_open] = ACTIONS(9346),
    [anon_sym_type] = ACTIONS(9346),
    [anon_sym_LBRACK_LT] = ACTIONS(9348),
    [anon_sym_do] = ACTIONS(9346),
    [anon_sym_and] = ACTIONS(9346),
    [anon_sym_let] = ACTIONS(9346),
    [anon_sym_let_BANG] = ACTIONS(9348),
    [aux_sym_access_modifier_token1] = ACTIONS(9348),
    [anon_sym_LPAREN] = ACTIONS(9346),
    [anon_sym_PIPE] = ACTIONS(9348),
    [anon_sym_AMP] = ACTIONS(9346),
    [anon_sym_LBRACK] = ACTIONS(9346),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9348),
    [anon_sym_LBRACE] = ACTIONS(9346),
    [anon_sym_LT_AT] = ACTIONS(9346),
    [anon_sym_LT_AT_AT] = ACTIONS(9348),
    [anon_sym_CARET] = ACTIONS(9348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9348),
    [anon_sym_new] = ACTIONS(9346),
    [anon_sym_with] = ACTIONS(9346),
    [anon_sym_return] = ACTIONS(9346),
    [anon_sym_return_BANG] = ACTIONS(9348),
    [anon_sym_yield] = ACTIONS(9346),
    [anon_sym_yield_BANG] = ACTIONS(9348),
    [anon_sym_lazy] = ACTIONS(9346),
    [anon_sym_assert] = ACTIONS(9346),
    [anon_sym_upcast] = ACTIONS(9346),
    [anon_sym_downcast] = ACTIONS(9346),
    [anon_sym_fixed] = ACTIONS(9346),
    [anon_sym_for] = ACTIONS(9346),
    [anon_sym_while] = ACTIONS(9346),
    [anon_sym_while_BANG] = ACTIONS(9348),
    [anon_sym_if] = ACTIONS(9346),
    [anon_sym_fun] = ACTIONS(9346),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(9346),
    [anon_sym_match] = ACTIONS(9346),
    [anon_sym_match_BANG] = ACTIONS(9348),
    [anon_sym_function] = ACTIONS(9346),
    [anon_sym_use] = ACTIONS(9346),
    [anon_sym_use_BANG] = ACTIONS(9348),
    [anon_sym_do_BANG] = ACTIONS(9348),
    [anon_sym_AMP_AMP] = ACTIONS(9348),
    [anon_sym_STAR] = ACTIONS(9348),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(9346),
    [anon_sym_static] = ACTIONS(9346),
    [anon_sym_member] = ACTIONS(9346),
    [anon_sym_exception] = ACTIONS(9346),
    [anon_sym_interface] = ACTIONS(9346),
    [anon_sym_abstract] = ACTIONS(9346),
    [anon_sym_val] = ACTIONS(9346),
    [anon_sym_override] = ACTIONS(9346),
    [anon_sym_extern] = ACTIONS(9346),
    [aux_sym_char_token1] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9346),
    [anon_sym_DQUOTE] = ACTIONS(9346),
    [anon_sym_AT_DQUOTE] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [sym_bool] = ACTIONS(9346),
    [sym_unit] = ACTIONS(9348),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9346),
    [sym_op_identifier] = ACTIONS(9348),
    [anon_sym_PLUS] = ACTIONS(9346),
    [anon_sym_DASH] = ACTIONS(9346),
    [anon_sym_PLUS_DOT] = ACTIONS(9348),
    [anon_sym_DASH_DOT] = ACTIONS(9348),
    [anon_sym_PERCENT] = ACTIONS(9346),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9348),
    [anon_sym_TILDE] = ACTIONS(9348),
    [aux_sym_prefix_op_token1] = ACTIONS(9348),
    [sym_int] = ACTIONS(9346),
    [sym_xint] = ACTIONS(9348),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9348),
    [anon_sym_POUNDload] = ACTIONS(9348),
    [anon_sym_POUNDtime] = ACTIONS(9348),
    [anon_sym_POUNDI] = ACTIONS(9348),
    [anon_sym_POUNDhelp] = ACTIONS(9348),
    [anon_sym_POUNDquit] = ACTIONS(9348),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9348),
    [anon_sym_POUNDendif] = ACTIONS(9348),
    [anon_sym_POUNDelse] = ACTIONS(9348),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9348),
  },
  [STATE(4956)] = {
    [sym_block_comment] = STATE(4956),
    [sym_compiler_directive_decl] = STATE(4956),
    [sym_fsi_directive_decl] = STATE(4956),
    [sym_preproc_line] = STATE(4956),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4957)] = {
    [sym_block_comment] = STATE(4957),
    [sym_compiler_directive_decl] = STATE(4957),
    [sym_fsi_directive_decl] = STATE(4957),
    [sym_preproc_line] = STATE(4957),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__untyped_quoted_close] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(4958)] = {
    [sym_block_comment] = STATE(4958),
    [sym_compiler_directive_decl] = STATE(4958),
    [sym_fsi_directive_decl] = STATE(4958),
    [sym_preproc_line] = STATE(4958),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9374),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__dedent] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
    [sym__in] = ACTIONS(6786),
  },
  [STATE(4959)] = {
    [sym_block_comment] = STATE(4959),
    [sym_compiler_directive_decl] = STATE(4959),
    [sym_fsi_directive_decl] = STATE(4959),
    [sym_preproc_line] = STATE(4959),
    [aux_sym_rules_repeat1] = STATE(4972),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9376),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_GT] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9378),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4960)] = {
    [sym_block_comment] = STATE(4960),
    [sym_compiler_directive_decl] = STATE(4960),
    [sym_fsi_directive_decl] = STATE(4960),
    [sym_preproc_line] = STATE(4960),
    [aux_sym_rules_repeat1] = STATE(4973),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9376),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_GT] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9378),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4961)] = {
    [sym_block_comment] = STATE(4961),
    [sym_compiler_directive_decl] = STATE(4961),
    [sym_fsi_directive_decl] = STATE(4961),
    [sym_preproc_line] = STATE(4961),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_GT] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4962)] = {
    [sym_block_comment] = STATE(4962),
    [sym_compiler_directive_decl] = STATE(4962),
    [sym_fsi_directive_decl] = STATE(4962),
    [sym_preproc_line] = STATE(4962),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__quoted_close] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4963)] = {
    [sym_block_comment] = STATE(4963),
    [sym_compiler_directive_decl] = STATE(4963),
    [sym_fsi_directive_decl] = STATE(4963),
    [sym_preproc_line] = STATE(4963),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__dedent] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
    [sym__in] = ACTIONS(6792),
  },
  [STATE(4964)] = {
    [sym_block_comment] = STATE(4964),
    [sym_compiler_directive_decl] = STATE(4964),
    [sym_fsi_directive_decl] = STATE(4964),
    [sym_preproc_line] = STATE(4964),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__dedent] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
    [sym__in] = ACTIONS(6796),
  },
  [STATE(4965)] = {
    [sym_type_arguments] = STATE(6477),
    [sym__object_members] = STATE(7120),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(4965),
    [sym_compiler_directive_decl] = STATE(4965),
    [sym_fsi_directive_decl] = STATE(4965),
    [sym_preproc_line] = STATE(4965),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(9226),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(9224),
    [anon_sym_module] = ACTIONS(9224),
    [anon_sym_null] = ACTIONS(9224),
    [anon_sym_begin] = ACTIONS(9224),
    [anon_sym_open] = ACTIONS(9224),
    [anon_sym_type] = ACTIONS(9224),
    [anon_sym_LBRACK_LT] = ACTIONS(9226),
    [anon_sym_do] = ACTIONS(9224),
    [anon_sym_and] = ACTIONS(9224),
    [anon_sym_let] = ACTIONS(9224),
    [anon_sym_let_BANG] = ACTIONS(9226),
    [aux_sym_access_modifier_token1] = ACTIONS(9226),
    [anon_sym_LPAREN] = ACTIONS(9224),
    [anon_sym_AMP] = ACTIONS(9224),
    [anon_sym_LBRACK] = ACTIONS(9224),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9226),
    [anon_sym_LBRACE] = ACTIONS(9224),
    [anon_sym_LT_AT] = ACTIONS(9224),
    [anon_sym_LT_AT_AT] = ACTIONS(9226),
    [anon_sym_CARET] = ACTIONS(9226),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9226),
    [anon_sym_new] = ACTIONS(9224),
    [anon_sym_with] = ACTIONS(9383),
    [anon_sym_return] = ACTIONS(9224),
    [anon_sym_return_BANG] = ACTIONS(9226),
    [anon_sym_yield] = ACTIONS(9224),
    [anon_sym_yield_BANG] = ACTIONS(9226),
    [anon_sym_lazy] = ACTIONS(9224),
    [anon_sym_assert] = ACTIONS(9224),
    [anon_sym_upcast] = ACTIONS(9224),
    [anon_sym_downcast] = ACTIONS(9224),
    [anon_sym_fixed] = ACTIONS(9224),
    [anon_sym_for] = ACTIONS(9224),
    [anon_sym_while] = ACTIONS(9224),
    [anon_sym_while_BANG] = ACTIONS(9226),
    [anon_sym_if] = ACTIONS(9224),
    [anon_sym_fun] = ACTIONS(9224),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(9224),
    [anon_sym_match] = ACTIONS(9224),
    [anon_sym_match_BANG] = ACTIONS(9226),
    [anon_sym_function] = ACTIONS(9224),
    [anon_sym_use] = ACTIONS(9224),
    [anon_sym_use_BANG] = ACTIONS(9226),
    [anon_sym_do_BANG] = ACTIONS(9226),
    [anon_sym_AMP_AMP] = ACTIONS(9226),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(9224),
    [anon_sym_static] = ACTIONS(9224),
    [anon_sym_member] = ACTIONS(9224),
    [anon_sym_exception] = ACTIONS(9224),
    [anon_sym_interface] = ACTIONS(9224),
    [anon_sym_abstract] = ACTIONS(9224),
    [anon_sym_val] = ACTIONS(9224),
    [anon_sym_override] = ACTIONS(9224),
    [anon_sym_extern] = ACTIONS(9224),
    [aux_sym_char_token1] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9224),
    [anon_sym_DQUOTE] = ACTIONS(9224),
    [anon_sym_AT_DQUOTE] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [sym_bool] = ACTIONS(9224),
    [sym_unit] = ACTIONS(9226),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9224),
    [sym_op_identifier] = ACTIONS(9226),
    [anon_sym_PLUS] = ACTIONS(9224),
    [anon_sym_DASH] = ACTIONS(9224),
    [anon_sym_PLUS_DOT] = ACTIONS(9226),
    [anon_sym_DASH_DOT] = ACTIONS(9226),
    [anon_sym_PERCENT] = ACTIONS(9224),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9226),
    [anon_sym_TILDE] = ACTIONS(9226),
    [aux_sym_prefix_op_token1] = ACTIONS(9226),
    [sym_int] = ACTIONS(9224),
    [sym_xint] = ACTIONS(9226),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9226),
    [anon_sym_POUNDload] = ACTIONS(9226),
    [anon_sym_POUNDtime] = ACTIONS(9226),
    [anon_sym_POUNDI] = ACTIONS(9226),
    [anon_sym_POUNDhelp] = ACTIONS(9226),
    [anon_sym_POUNDquit] = ACTIONS(9226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9226),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9226),
  },
  [STATE(4966)] = {
    [sym_block_comment] = STATE(4966),
    [sym_compiler_directive_decl] = STATE(4966),
    [sym_fsi_directive_decl] = STATE(4966),
    [sym_preproc_line] = STATE(4966),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_DOT_DOT] = ACTIONS(6896),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [anon_sym_DOT_DOT2] = ACTIONS(6894),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__dedent] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(4967)] = {
    [sym_block_comment] = STATE(4967),
    [sym_compiler_directive_decl] = STATE(4967),
    [sym_fsi_directive_decl] = STATE(4967),
    [sym_preproc_line] = STATE(4967),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_DOT_DOT] = ACTIONS(6900),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [anon_sym_DOT_DOT2] = ACTIONS(6898),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__dedent] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(4968)] = {
    [sym_block_comment] = STATE(4968),
    [sym_compiler_directive_decl] = STATE(4968),
    [sym_fsi_directive_decl] = STATE(4968),
    [sym_preproc_line] = STATE(4968),
    [aux_sym_rules_repeat1] = STATE(5287),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9393),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9395),
    [sym__quoted_close] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(4969)] = {
    [sym_block_comment] = STATE(4969),
    [sym_compiler_directive_decl] = STATE(4969),
    [sym_fsi_directive_decl] = STATE(4969),
    [sym_preproc_line] = STATE(4969),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(4970)] = {
    [sym_block_comment] = STATE(4970),
    [sym_compiler_directive_decl] = STATE(4970),
    [sym_fsi_directive_decl] = STATE(4970),
    [sym_preproc_line] = STATE(4970),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_DOT_DOT] = ACTIONS(6902),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [anon_sym_DOT_DOT2] = ACTIONS(6603),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__dedent] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(4971)] = {
    [sym_block_comment] = STATE(4971),
    [sym_compiler_directive_decl] = STATE(4971),
    [sym_fsi_directive_decl] = STATE(4971),
    [sym_preproc_line] = STATE(4971),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_DOT_DOT] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(4972)] = {
    [sym_block_comment] = STATE(4972),
    [sym_compiler_directive_decl] = STATE(4972),
    [sym_fsi_directive_decl] = STATE(4972),
    [sym_preproc_line] = STATE(4972),
    [aux_sym_rules_repeat1] = STATE(4973),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9376),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_GT] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9398),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(4973)] = {
    [sym_block_comment] = STATE(4973),
    [sym_compiler_directive_decl] = STATE(4973),
    [sym_fsi_directive_decl] = STATE(4973),
    [sym_preproc_line] = STATE(4973),
    [aux_sym_rules_repeat1] = STATE(4973),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9401),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_GT] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9404),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(4974)] = {
    [sym_block_comment] = STATE(4974),
    [sym_compiler_directive_decl] = STATE(4974),
    [sym_fsi_directive_decl] = STATE(4974),
    [sym_preproc_line] = STATE(4974),
    [aux_sym_type_argument_repeat1] = STATE(5149),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_PIPE] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6333),
    [anon_sym_CARET] = ACTIONS(6333),
    [anon_sym_or] = ACTIONS(9407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6333),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_AMP_AMP] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6333),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6333),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6333),
    [anon_sym_DASH_DOT] = ACTIONS(6333),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6333),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6333),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
  },
  [STATE(4975)] = {
    [sym_block_comment] = STATE(4975),
    [sym_compiler_directive_decl] = STATE(4975),
    [sym_fsi_directive_decl] = STATE(4975),
    [sym_preproc_line] = STATE(4975),
    [aux_sym_long_identifier_repeat1] = STATE(5263),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9409),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__untyped_quoted_close] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(4976)] = {
    [sym_block_comment] = STATE(4976),
    [sym_compiler_directive_decl] = STATE(4976),
    [sym_fsi_directive_decl] = STATE(4976),
    [sym_preproc_line] = STATE(4976),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__dedent] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
    [sym__in] = ACTIONS(6694),
  },
  [STATE(4977)] = {
    [sym_block_comment] = STATE(4977),
    [sym_compiler_directive_decl] = STATE(4977),
    [sym_fsi_directive_decl] = STATE(4977),
    [sym_preproc_line] = STATE(4977),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_DOT_DOT] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(4978)] = {
    [sym_block_comment] = STATE(4978),
    [sym_compiler_directive_decl] = STATE(4978),
    [sym_fsi_directive_decl] = STATE(4978),
    [sym_preproc_line] = STATE(4978),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_DOT_DOT] = ACTIONS(7059),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [anon_sym_DOT_DOT2] = ACTIONS(7057),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__dedent] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(4979)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(4979),
    [sym_compiler_directive_decl] = STATE(4979),
    [sym_fsi_directive_decl] = STATE(4979),
    [sym_preproc_line] = STATE(4979),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_PIPE] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6255),
    [anon_sym_CARET] = ACTIONS(6255),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_AMP_AMP] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6255),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6255),
    [anon_sym_DASH_DOT] = ACTIONS(6255),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6255),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6255),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [anon_sym_POUNDelse] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
  },
  [STATE(4980)] = {
    [sym_block_comment] = STATE(4980),
    [sym_compiler_directive_decl] = STATE(4980),
    [sym_fsi_directive_decl] = STATE(4980),
    [sym_preproc_line] = STATE(4980),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__dedent] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
    [sym__in] = ACTIONS(6861),
  },
  [STATE(4981)] = {
    [sym_block_comment] = STATE(4981),
    [sym_compiler_directive_decl] = STATE(4981),
    [sym_fsi_directive_decl] = STATE(4981),
    [sym_preproc_line] = STATE(4981),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_PIPE] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6418),
    [anon_sym_CARET] = ACTIONS(6418),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6418),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_AMP_AMP] = ACTIONS(6418),
    [anon_sym_STAR] = ACTIONS(6418),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6418),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6418),
    [anon_sym_DASH_DOT] = ACTIONS(6418),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6418),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6418),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
  },
  [STATE(4982)] = {
    [sym_block_comment] = STATE(4982),
    [sym_compiler_directive_decl] = STATE(4982),
    [sym_fsi_directive_decl] = STATE(4982),
    [sym_preproc_line] = STATE(4982),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9335),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
    [sym__in] = ACTIONS(6599),
  },
  [STATE(4983)] = {
    [sym_block_comment] = STATE(4983),
    [sym_compiler_directive_decl] = STATE(4983),
    [sym_fsi_directive_decl] = STATE(4983),
    [sym_preproc_line] = STATE(4983),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9413),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [anon_sym_EQ2] = ACTIONS(9041),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4984)] = {
    [sym_block_comment] = STATE(4984),
    [sym_compiler_directive_decl] = STATE(4984),
    [sym_fsi_directive_decl] = STATE(4984),
    [sym_preproc_line] = STATE(4984),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_as] = ACTIONS(6807),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(4985)] = {
    [sym_block_comment] = STATE(4985),
    [sym_compiler_directive_decl] = STATE(4985),
    [sym_fsi_directive_decl] = STATE(4985),
    [sym_preproc_line] = STATE(4985),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_DASH_GT] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_DOT_DOT] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(4986)] = {
    [sym_block_comment] = STATE(4986),
    [sym_compiler_directive_decl] = STATE(4986),
    [sym_fsi_directive_decl] = STATE(4986),
    [sym_preproc_line] = STATE(4986),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9415),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__untyped_quoted_close] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4987)] = {
    [sym_block_comment] = STATE(4987),
    [sym_compiler_directive_decl] = STATE(4987),
    [sym_fsi_directive_decl] = STATE(4987),
    [sym_preproc_line] = STATE(4987),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(8025),
    [aux_sym_decimal_token1] = ACTIONS(7634),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__multi_dollar_interp_end] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(4988)] = {
    [sym_type_arguments] = STATE(6476),
    [sym__object_members] = STATE(7061),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(4988),
    [sym_compiler_directive_decl] = STATE(4988),
    [sym_fsi_directive_decl] = STATE(4988),
    [sym_preproc_line] = STATE(4988),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(9224),
    [anon_sym_null] = ACTIONS(9224),
    [anon_sym_begin] = ACTIONS(9224),
    [anon_sym_open] = ACTIONS(9224),
    [anon_sym_type] = ACTIONS(9224),
    [anon_sym_LBRACK_LT] = ACTIONS(9226),
    [anon_sym_do] = ACTIONS(9224),
    [anon_sym_and] = ACTIONS(9224),
    [anon_sym_let] = ACTIONS(9224),
    [anon_sym_let_BANG] = ACTIONS(9226),
    [aux_sym_access_modifier_token1] = ACTIONS(9226),
    [anon_sym_LPAREN] = ACTIONS(9224),
    [anon_sym_AMP] = ACTIONS(9224),
    [anon_sym_LBRACK] = ACTIONS(9224),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9226),
    [anon_sym_LBRACE] = ACTIONS(9224),
    [anon_sym_LT_AT] = ACTIONS(9224),
    [anon_sym_LT_AT_AT] = ACTIONS(9226),
    [anon_sym_CARET] = ACTIONS(9226),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9226),
    [anon_sym_new] = ACTIONS(9224),
    [anon_sym_with] = ACTIONS(9417),
    [anon_sym_return] = ACTIONS(9224),
    [anon_sym_return_BANG] = ACTIONS(9226),
    [anon_sym_yield] = ACTIONS(9224),
    [anon_sym_yield_BANG] = ACTIONS(9226),
    [anon_sym_lazy] = ACTIONS(9224),
    [anon_sym_assert] = ACTIONS(9224),
    [anon_sym_upcast] = ACTIONS(9224),
    [anon_sym_downcast] = ACTIONS(9224),
    [anon_sym_fixed] = ACTIONS(9224),
    [anon_sym_for] = ACTIONS(9224),
    [anon_sym_while] = ACTIONS(9224),
    [anon_sym_while_BANG] = ACTIONS(9226),
    [anon_sym_if] = ACTIONS(9224),
    [anon_sym_fun] = ACTIONS(9224),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(9224),
    [anon_sym_match] = ACTIONS(9224),
    [anon_sym_match_BANG] = ACTIONS(9226),
    [anon_sym_function] = ACTIONS(9224),
    [anon_sym_use] = ACTIONS(9224),
    [anon_sym_use_BANG] = ACTIONS(9226),
    [anon_sym_do_BANG] = ACTIONS(9226),
    [anon_sym_AMP_AMP] = ACTIONS(9226),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(9224),
    [anon_sym_static] = ACTIONS(9224),
    [anon_sym_member] = ACTIONS(9224),
    [anon_sym_exception] = ACTIONS(9224),
    [anon_sym_interface] = ACTIONS(9224),
    [anon_sym_abstract] = ACTIONS(9224),
    [anon_sym_val] = ACTIONS(9224),
    [anon_sym_override] = ACTIONS(9224),
    [anon_sym_extern] = ACTIONS(9224),
    [aux_sym_char_token1] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9224),
    [anon_sym_DQUOTE] = ACTIONS(9224),
    [anon_sym_AT_DQUOTE] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [sym_bool] = ACTIONS(9224),
    [sym_unit] = ACTIONS(9226),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9224),
    [sym_op_identifier] = ACTIONS(9226),
    [anon_sym_PLUS] = ACTIONS(9224),
    [anon_sym_DASH] = ACTIONS(9224),
    [anon_sym_PLUS_DOT] = ACTIONS(9226),
    [anon_sym_DASH_DOT] = ACTIONS(9226),
    [anon_sym_PERCENT] = ACTIONS(9224),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9226),
    [anon_sym_TILDE] = ACTIONS(9226),
    [aux_sym_prefix_op_token1] = ACTIONS(9226),
    [sym_int] = ACTIONS(9224),
    [sym_xint] = ACTIONS(9226),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9226),
    [anon_sym_POUNDload] = ACTIONS(9226),
    [anon_sym_POUNDtime] = ACTIONS(9226),
    [anon_sym_POUNDI] = ACTIONS(9226),
    [anon_sym_POUNDhelp] = ACTIONS(9226),
    [anon_sym_POUNDquit] = ACTIONS(9226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9226),
    [anon_sym_POUNDendif] = ACTIONS(9226),
    [anon_sym_POUNDelse] = ACTIONS(9226),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9226),
  },
  [STATE(4989)] = {
    [sym_block_comment] = STATE(4989),
    [sym_compiler_directive_decl] = STATE(4989),
    [sym_fsi_directive_decl] = STATE(4989),
    [sym_preproc_line] = STATE(4989),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_DASH_GT] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_DOT_DOT] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(4990)] = {
    [sym_block_comment] = STATE(4990),
    [sym_compiler_directive_decl] = STATE(4990),
    [sym_fsi_directive_decl] = STATE(4990),
    [sym_preproc_line] = STATE(4990),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_DASH_GT] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(4991)] = {
    [sym_block_comment] = STATE(4991),
    [sym_compiler_directive_decl] = STATE(4991),
    [sym_fsi_directive_decl] = STATE(4991),
    [sym_preproc_line] = STATE(4991),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9419),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(4992)] = {
    [sym_block_comment] = STATE(4992),
    [sym_compiler_directive_decl] = STATE(4992),
    [sym_fsi_directive_decl] = STATE(4992),
    [sym_preproc_line] = STATE(4992),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9421),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__multi_dollar_interp_end] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4993)] = {
    [sym_block_comment] = STATE(4993),
    [sym_compiler_directive_decl] = STATE(4993),
    [sym_fsi_directive_decl] = STATE(4993),
    [sym_preproc_line] = STATE(4993),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__untyped_quoted_close] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(4994)] = {
    [sym_block_comment] = STATE(4994),
    [sym_compiler_directive_decl] = STATE(4994),
    [sym_fsi_directive_decl] = STATE(4994),
    [sym_preproc_line] = STATE(4994),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_DASH_GT] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_DOT_DOT] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(4995)] = {
    [sym_block_comment] = STATE(4995),
    [sym_compiler_directive_decl] = STATE(4995),
    [sym_fsi_directive_decl] = STATE(4995),
    [sym_preproc_line] = STATE(4995),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_DOT_DOT] = ACTIONS(6967),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [anon_sym_DOT_DOT2] = ACTIONS(6965),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__dedent] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(4996)] = {
    [sym_block_comment] = STATE(4996),
    [sym_compiler_directive_decl] = STATE(4996),
    [sym_fsi_directive_decl] = STATE(4996),
    [sym_preproc_line] = STATE(4996),
    [aux_sym_sequential_expression_repeat1] = STATE(4996),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_RBRACK] = ACTIONS(6798),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_DOT_DOT2] = ACTIONS(6798),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9423),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(4997)] = {
    [sym_block_comment] = STATE(4997),
    [sym_compiler_directive_decl] = STATE(4997),
    [sym_fsi_directive_decl] = STATE(4997),
    [sym_preproc_line] = STATE(4997),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_with] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9426),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(4998)] = {
    [sym_block_comment] = STATE(4998),
    [sym_compiler_directive_decl] = STATE(4998),
    [sym_fsi_directive_decl] = STATE(4998),
    [sym_preproc_line] = STATE(4998),
    [aux_sym_sequential_expression_repeat1] = STATE(5156),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_as] = ACTIONS(6740),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(4999)] = {
    [sym_block_comment] = STATE(4999),
    [sym_compiler_directive_decl] = STATE(4999),
    [sym_fsi_directive_decl] = STATE(4999),
    [sym_preproc_line] = STATE(4999),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [anon_sym_POUNDendif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5000)] = {
    [sym_block_comment] = STATE(5000),
    [sym_compiler_directive_decl] = STATE(5000),
    [sym_fsi_directive_decl] = STATE(5000),
    [sym_preproc_line] = STATE(5000),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5001)] = {
    [sym_block_comment] = STATE(5001),
    [sym_compiler_directive_decl] = STATE(5001),
    [sym_fsi_directive_decl] = STATE(5001),
    [sym_preproc_line] = STATE(5001),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_DOT_DOT] = ACTIONS(6971),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [anon_sym_DOT_DOT2] = ACTIONS(6969),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__dedent] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(5002)] = {
    [sym_block_comment] = STATE(5002),
    [sym_compiler_directive_decl] = STATE(5002),
    [sym_fsi_directive_decl] = STATE(5002),
    [sym_preproc_line] = STATE(5002),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_GT] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7998),
    [aux_sym_decimal_token1] = ACTIONS(7574),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5003)] = {
    [sym_block_comment] = STATE(5003),
    [sym_compiler_directive_decl] = STATE(5003),
    [sym_fsi_directive_decl] = STATE(5003),
    [sym_preproc_line] = STATE(5003),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9428),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_DASH_GT] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5004)] = {
    [sym_block_comment] = STATE(5004),
    [sym_compiler_directive_decl] = STATE(5004),
    [sym_fsi_directive_decl] = STATE(5004),
    [sym_preproc_line] = STATE(5004),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5005)] = {
    [sym_block_comment] = STATE(5005),
    [sym_compiler_directive_decl] = STATE(5005),
    [sym_fsi_directive_decl] = STATE(5005),
    [sym_preproc_line] = STATE(5005),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_GT] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9430),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(5006)] = {
    [sym_block_comment] = STATE(5006),
    [sym_compiler_directive_decl] = STATE(5006),
    [sym_fsi_directive_decl] = STATE(5006),
    [sym_preproc_line] = STATE(5006),
    [aux_sym_sequential_expression_repeat1] = STATE(4996),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_RBRACK] = ACTIONS(6738),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_DOT_DOT2] = ACTIONS(6738),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5007)] = {
    [sym_block_comment] = STATE(5007),
    [sym_compiler_directive_decl] = STATE(5007),
    [sym_fsi_directive_decl] = STATE(5007),
    [sym_preproc_line] = STATE(5007),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__multi_dollar_interp_end] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(5008)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5008),
    [sym_compiler_directive_decl] = STATE(5008),
    [sym_fsi_directive_decl] = STATE(5008),
    [sym_preproc_line] = STATE(5008),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(9354),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(9352),
    [anon_sym_module] = ACTIONS(9352),
    [anon_sym_null] = ACTIONS(9352),
    [anon_sym_begin] = ACTIONS(9352),
    [anon_sym_open] = ACTIONS(9352),
    [anon_sym_type] = ACTIONS(9352),
    [anon_sym_LBRACK_LT] = ACTIONS(9354),
    [anon_sym_do] = ACTIONS(9352),
    [anon_sym_and] = ACTIONS(9352),
    [anon_sym_let] = ACTIONS(9352),
    [anon_sym_let_BANG] = ACTIONS(9354),
    [aux_sym_access_modifier_token1] = ACTIONS(9354),
    [anon_sym_LPAREN] = ACTIONS(9352),
    [anon_sym_PIPE] = ACTIONS(9354),
    [anon_sym_AMP] = ACTIONS(9352),
    [anon_sym_LBRACK] = ACTIONS(9352),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9354),
    [anon_sym_LBRACE] = ACTIONS(9352),
    [anon_sym_LT_AT] = ACTIONS(9352),
    [anon_sym_LT_AT_AT] = ACTIONS(9354),
    [anon_sym_CARET] = ACTIONS(9354),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9354),
    [anon_sym_new] = ACTIONS(9352),
    [anon_sym_with] = ACTIONS(9352),
    [anon_sym_return] = ACTIONS(9352),
    [anon_sym_return_BANG] = ACTIONS(9354),
    [anon_sym_yield] = ACTIONS(9352),
    [anon_sym_yield_BANG] = ACTIONS(9354),
    [anon_sym_lazy] = ACTIONS(9352),
    [anon_sym_assert] = ACTIONS(9352),
    [anon_sym_upcast] = ACTIONS(9352),
    [anon_sym_downcast] = ACTIONS(9352),
    [anon_sym_fixed] = ACTIONS(9352),
    [anon_sym_for] = ACTIONS(9352),
    [anon_sym_while] = ACTIONS(9352),
    [anon_sym_while_BANG] = ACTIONS(9354),
    [anon_sym_if] = ACTIONS(9352),
    [anon_sym_fun] = ACTIONS(9352),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(9352),
    [anon_sym_match] = ACTIONS(9352),
    [anon_sym_match_BANG] = ACTIONS(9354),
    [anon_sym_function] = ACTIONS(9352),
    [anon_sym_use] = ACTIONS(9352),
    [anon_sym_use_BANG] = ACTIONS(9354),
    [anon_sym_do_BANG] = ACTIONS(9354),
    [anon_sym_AMP_AMP] = ACTIONS(9354),
    [anon_sym_STAR] = ACTIONS(9354),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(9352),
    [anon_sym_static] = ACTIONS(9352),
    [anon_sym_member] = ACTIONS(9352),
    [anon_sym_exception] = ACTIONS(9352),
    [anon_sym_interface] = ACTIONS(9352),
    [anon_sym_abstract] = ACTIONS(9352),
    [anon_sym_val] = ACTIONS(9352),
    [anon_sym_override] = ACTIONS(9352),
    [anon_sym_extern] = ACTIONS(9352),
    [aux_sym_char_token1] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9352),
    [anon_sym_DQUOTE] = ACTIONS(9352),
    [anon_sym_AT_DQUOTE] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [sym_bool] = ACTIONS(9352),
    [sym_unit] = ACTIONS(9354),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9352),
    [sym_op_identifier] = ACTIONS(9354),
    [anon_sym_PLUS] = ACTIONS(9352),
    [anon_sym_DASH] = ACTIONS(9352),
    [anon_sym_PLUS_DOT] = ACTIONS(9354),
    [anon_sym_DASH_DOT] = ACTIONS(9354),
    [anon_sym_PERCENT] = ACTIONS(9352),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9354),
    [anon_sym_TILDE] = ACTIONS(9354),
    [aux_sym_prefix_op_token1] = ACTIONS(9354),
    [sym_int] = ACTIONS(9352),
    [sym_xint] = ACTIONS(9354),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9354),
    [anon_sym_POUNDload] = ACTIONS(9354),
    [anon_sym_POUNDtime] = ACTIONS(9354),
    [anon_sym_POUNDI] = ACTIONS(9354),
    [anon_sym_POUNDhelp] = ACTIONS(9354),
    [anon_sym_POUNDquit] = ACTIONS(9354),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9354),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9354),
  },
  [STATE(5009)] = {
    [sym_block_comment] = STATE(5009),
    [sym_compiler_directive_decl] = STATE(5009),
    [sym_fsi_directive_decl] = STATE(5009),
    [sym_preproc_line] = STATE(5009),
    [aux_sym_long_identifier_repeat1] = STATE(4975),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9432),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__untyped_quoted_close] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(5010)] = {
    [sym_block_comment] = STATE(5010),
    [sym_compiler_directive_decl] = STATE(5010),
    [sym_fsi_directive_decl] = STATE(5010),
    [sym_preproc_line] = STATE(5010),
    [aux_sym_type_argument_repeat1] = STATE(5052),
    [ts_builtin_sym_end] = ACTIONS(6333),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_namespace] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_PIPE] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6333),
    [anon_sym_CARET] = ACTIONS(6333),
    [anon_sym_or] = ACTIONS(9436),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6333),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_AMP_AMP] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6333),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6333),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6333),
    [anon_sym_DASH_DOT] = ACTIONS(6333),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6333),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6333),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
  },
  [STATE(5011)] = {
    [sym_block_comment] = STATE(5011),
    [sym_compiler_directive_decl] = STATE(5011),
    [sym_fsi_directive_decl] = STATE(5011),
    [sym_preproc_line] = STATE(5011),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9438),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__dedent] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
    [sym__in] = ACTIONS(6766),
  },
  [STATE(5012)] = {
    [sym_block_comment] = STATE(5012),
    [sym_compiler_directive_decl] = STATE(5012),
    [sym_fsi_directive_decl] = STATE(5012),
    [sym_preproc_line] = STATE(5012),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9440),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5013)] = {
    [sym_block_comment] = STATE(5013),
    [sym_compiler_directive_decl] = STATE(5013),
    [sym_fsi_directive_decl] = STATE(5013),
    [sym_preproc_line] = STATE(5013),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5014)] = {
    [sym_block_comment] = STATE(5014),
    [sym_compiler_directive_decl] = STATE(5014),
    [sym_fsi_directive_decl] = STATE(5014),
    [sym_preproc_line] = STATE(5014),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_DASH_GT] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5015)] = {
    [sym_block_comment] = STATE(5015),
    [sym_compiler_directive_decl] = STATE(5015),
    [sym_fsi_directive_decl] = STATE(5015),
    [sym_preproc_line] = STATE(5015),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_DOT_DOT] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(5016)] = {
    [sym_block_comment] = STATE(5016),
    [sym_compiler_directive_decl] = STATE(5016),
    [sym_fsi_directive_decl] = STATE(5016),
    [sym_preproc_line] = STATE(5016),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_GT] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(5017)] = {
    [sym_block_comment] = STATE(5017),
    [sym_compiler_directive_decl] = STATE(5017),
    [sym_fsi_directive_decl] = STATE(5017),
    [sym_preproc_line] = STATE(5017),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_DOT_DOT] = ACTIONS(6906),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [anon_sym_DOT_DOT2] = ACTIONS(6904),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__dedent] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(5018)] = {
    [sym_block_comment] = STATE(5018),
    [sym_compiler_directive_decl] = STATE(5018),
    [sym_fsi_directive_decl] = STATE(5018),
    [sym_preproc_line] = STATE(5018),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [anon_sym_EQ2] = ACTIONS(6297),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5019)] = {
    [sym_block_comment] = STATE(5019),
    [sym_compiler_directive_decl] = STATE(5019),
    [sym_fsi_directive_decl] = STATE(5019),
    [sym_preproc_line] = STATE(5019),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_as] = ACTIONS(6834),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9442),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5020)] = {
    [sym_block_comment] = STATE(5020),
    [sym_compiler_directive_decl] = STATE(5020),
    [sym_fsi_directive_decl] = STATE(5020),
    [sym_preproc_line] = STATE(5020),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_as] = ACTIONS(6834),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_with] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5021)] = {
    [sym_block_comment] = STATE(5021),
    [sym_compiler_directive_decl] = STATE(5021),
    [sym_fsi_directive_decl] = STATE(5021),
    [sym_preproc_line] = STATE(5021),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__quoted_close] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(5022)] = {
    [sym_block_comment] = STATE(5022),
    [sym_compiler_directive_decl] = STATE(5022),
    [sym_fsi_directive_decl] = STATE(5022),
    [sym_preproc_line] = STATE(5022),
    [aux_sym_long_identifier_repeat1] = STATE(5025),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9444),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__multi_dollar_interp_end] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(5023)] = {
    [sym_block_comment] = STATE(5023),
    [sym_compiler_directive_decl] = STATE(5023),
    [sym_fsi_directive_decl] = STATE(5023),
    [sym_preproc_line] = STATE(5023),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_as] = ACTIONS(6750),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5024)] = {
    [sym_block_comment] = STATE(5024),
    [sym_compiler_directive_decl] = STATE(5024),
    [sym_fsi_directive_decl] = STATE(5024),
    [sym_preproc_line] = STATE(5024),
    [aux_sym_sequential_expression_repeat1] = STATE(5024),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_DASH_GT] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_DOT_DOT] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9446),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5025)] = {
    [sym_block_comment] = STATE(5025),
    [sym_compiler_directive_decl] = STATE(5025),
    [sym_fsi_directive_decl] = STATE(5025),
    [sym_preproc_line] = STATE(5025),
    [aux_sym_long_identifier_repeat1] = STATE(5025),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9449),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__multi_dollar_interp_end] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5026)] = {
    [sym_block_comment] = STATE(5026),
    [sym_compiler_directive_decl] = STATE(5026),
    [sym_fsi_directive_decl] = STATE(5026),
    [sym_preproc_line] = STATE(5026),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__dedent] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
    [sym__in] = ACTIONS(6830),
  },
  [STATE(5027)] = {
    [sym_block_comment] = STATE(5027),
    [sym_compiler_directive_decl] = STATE(5027),
    [sym_fsi_directive_decl] = STATE(5027),
    [sym_preproc_line] = STATE(5027),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__dedent] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
    [sym__in] = ACTIONS(6865),
  },
  [STATE(5028)] = {
    [sym_block_comment] = STATE(5028),
    [sym_compiler_directive_decl] = STATE(5028),
    [sym_fsi_directive_decl] = STATE(5028),
    [sym_preproc_line] = STATE(5028),
    [aux_sym_long_identifier_repeat1] = STATE(5028),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9452),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5029)] = {
    [sym_block_comment] = STATE(5029),
    [sym_compiler_directive_decl] = STATE(5029),
    [sym_fsi_directive_decl] = STATE(5029),
    [sym_preproc_line] = STATE(5029),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_as] = ACTIONS(6784),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9455),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5030)] = {
    [sym_block_comment] = STATE(5030),
    [sym_compiler_directive_decl] = STATE(5030),
    [sym_fsi_directive_decl] = STATE(5030),
    [sym_preproc_line] = STATE(5030),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__multi_dollar_interp_end] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(5031)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(5031),
    [sym_compiler_directive_decl] = STATE(5031),
    [sym_fsi_directive_decl] = STATE(5031),
    [sym_preproc_line] = STATE(5031),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_PIPE] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6209),
    [anon_sym_CARET] = ACTIONS(6209),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_AMP_AMP] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6209),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6209),
    [anon_sym_DASH_DOT] = ACTIONS(6209),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6209),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6209),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [anon_sym_POUNDelse] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
  },
  [STATE(5032)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(5032),
    [sym_compiler_directive_decl] = STATE(5032),
    [sym_fsi_directive_decl] = STATE(5032),
    [sym_preproc_line] = STATE(5032),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_PIPE] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6195),
    [anon_sym_CARET] = ACTIONS(6195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_AMP_AMP] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6195),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6195),
    [anon_sym_DASH_DOT] = ACTIONS(6195),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6195),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6195),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
  },
  [STATE(5033)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(5033),
    [sym_compiler_directive_decl] = STATE(5033),
    [sym_fsi_directive_decl] = STATE(5033),
    [sym_preproc_line] = STATE(5033),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_PIPE] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6259),
    [anon_sym_CARET] = ACTIONS(6259),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_AMP_AMP] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6259),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6259),
    [anon_sym_DASH_DOT] = ACTIONS(6259),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6259),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6259),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [anon_sym_POUNDelse] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
  },
  [STATE(5034)] = {
    [sym_block_comment] = STATE(5034),
    [sym_compiler_directive_decl] = STATE(5034),
    [sym_fsi_directive_decl] = STATE(5034),
    [sym_preproc_line] = STATE(5034),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_as] = ACTIONS(6790),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_with] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5035)] = {
    [sym_block_comment] = STATE(5035),
    [sym_compiler_directive_decl] = STATE(5035),
    [sym_fsi_directive_decl] = STATE(5035),
    [sym_preproc_line] = STATE(5035),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_DASH_GT] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_DOT_DOT] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5036)] = {
    [sym_block_comment] = STATE(5036),
    [sym_compiler_directive_decl] = STATE(5036),
    [sym_fsi_directive_decl] = STATE(5036),
    [sym_preproc_line] = STATE(5036),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_as] = ACTIONS(6794),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_with] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5037)] = {
    [sym_block_comment] = STATE(5037),
    [sym_compiler_directive_decl] = STATE(5037),
    [sym_fsi_directive_decl] = STATE(5037),
    [sym_preproc_line] = STATE(5037),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_DOT_DOT] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__dedent] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5038)] = {
    [sym_block_comment] = STATE(5038),
    [sym_compiler_directive_decl] = STATE(5038),
    [sym_fsi_directive_decl] = STATE(5038),
    [sym_preproc_line] = STATE(5038),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5039)] = {
    [sym_block_comment] = STATE(5039),
    [sym_compiler_directive_decl] = STATE(5039),
    [sym_fsi_directive_decl] = STATE(5039),
    [sym_preproc_line] = STATE(5039),
    [aux_sym_long_identifier_repeat1] = STATE(5139),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9457),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_with] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(5040)] = {
    [sym_block_comment] = STATE(5040),
    [sym_compiler_directive_decl] = STATE(5040),
    [sym_fsi_directive_decl] = STATE(5040),
    [sym_preproc_line] = STATE(5040),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_as] = ACTIONS(6692),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5041)] = {
    [sym_block_comment] = STATE(5041),
    [sym_compiler_directive_decl] = STATE(5041),
    [sym_fsi_directive_decl] = STATE(5041),
    [sym_preproc_line] = STATE(5041),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__quoted_close] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(5042)] = {
    [sym_block_comment] = STATE(5042),
    [sym_compiler_directive_decl] = STATE(5042),
    [sym_fsi_directive_decl] = STATE(5042),
    [sym_preproc_line] = STATE(5042),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(8010),
    [aux_sym_decimal_token1] = ACTIONS(7695),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__untyped_quoted_close] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5043)] = {
    [sym_block_comment] = STATE(5043),
    [sym_compiler_directive_decl] = STATE(5043),
    [sym_fsi_directive_decl] = STATE(5043),
    [sym_preproc_line] = STATE(5043),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_GT] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5044)] = {
    [sym_block_comment] = STATE(5044),
    [sym_compiler_directive_decl] = STATE(5044),
    [sym_fsi_directive_decl] = STATE(5044),
    [sym_preproc_line] = STATE(5044),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9461),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_DASH_GT] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_DOT_DOT] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5045)] = {
    [sym_block_comment] = STATE(5045),
    [sym_compiler_directive_decl] = STATE(5045),
    [sym_fsi_directive_decl] = STATE(5045),
    [sym_preproc_line] = STATE(5045),
    [aux_sym_long_identifier_repeat1] = STATE(5045),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9463),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5046)] = {
    [sym_block_comment] = STATE(5046),
    [sym_compiler_directive_decl] = STATE(5046),
    [sym_fsi_directive_decl] = STATE(5046),
    [sym_preproc_line] = STATE(5046),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5047)] = {
    [sym_block_comment] = STATE(5047),
    [sym_compiler_directive_decl] = STATE(5047),
    [sym_fsi_directive_decl] = STATE(5047),
    [sym_preproc_line] = STATE(5047),
    [aux_sym_type_argument_repeat1] = STATE(5047),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(9466),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5048)] = {
    [sym_block_comment] = STATE(5048),
    [sym_compiler_directive_decl] = STATE(5048),
    [sym_fsi_directive_decl] = STATE(5048),
    [sym_preproc_line] = STATE(5048),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6807),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [anon_sym_DOT_DOT2] = ACTIONS(6805),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5049)] = {
    [sym_block_comment] = STATE(5049),
    [sym_compiler_directive_decl] = STATE(5049),
    [sym_fsi_directive_decl] = STATE(5049),
    [sym_preproc_line] = STATE(5049),
    [aux_sym_rules_repeat1] = STATE(5049),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9469),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9472),
    [sym__untyped_quoted_close] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5050)] = {
    [sym_block_comment] = STATE(5050),
    [sym_compiler_directive_decl] = STATE(5050),
    [sym_fsi_directive_decl] = STATE(5050),
    [sym_preproc_line] = STATE(5050),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_GT] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5051)] = {
    [sym_block_comment] = STATE(5051),
    [sym_compiler_directive_decl] = STATE(5051),
    [sym_fsi_directive_decl] = STATE(5051),
    [sym_preproc_line] = STATE(5051),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_DASH_GT] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_DOT_DOT] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5052)] = {
    [sym_block_comment] = STATE(5052),
    [sym_compiler_directive_decl] = STATE(5052),
    [sym_fsi_directive_decl] = STATE(5052),
    [sym_preproc_line] = STATE(5052),
    [aux_sym_type_argument_repeat1] = STATE(5056),
    [ts_builtin_sym_end] = ACTIONS(6306),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_namespace] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_PIPE] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6306),
    [anon_sym_CARET] = ACTIONS(6306),
    [anon_sym_or] = ACTIONS(9436),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6306),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_AMP_AMP] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6306),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6306),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6306),
    [anon_sym_DASH_DOT] = ACTIONS(6306),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6306),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6306),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
  },
  [STATE(5053)] = {
    [sym_block_comment] = STATE(5053),
    [sym_compiler_directive_decl] = STATE(5053),
    [sym_fsi_directive_decl] = STATE(5053),
    [sym_preproc_line] = STATE(5053),
    [aux_sym_long_identifier_repeat1] = STATE(5022),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9475),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__multi_dollar_interp_end] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(5054)] = {
    [sym_block_comment] = STATE(5054),
    [sym_compiler_directive_decl] = STATE(5054),
    [sym_fsi_directive_decl] = STATE(5054),
    [sym_preproc_line] = STATE(5054),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__multi_dollar_interp_end] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5055)] = {
    [sym_block_comment] = STATE(5055),
    [sym_compiler_directive_decl] = STATE(5055),
    [sym_fsi_directive_decl] = STATE(5055),
    [sym_preproc_line] = STATE(5055),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_DASH_GT] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_DOT_DOT] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5056)] = {
    [sym_block_comment] = STATE(5056),
    [sym_compiler_directive_decl] = STATE(5056),
    [sym_fsi_directive_decl] = STATE(5056),
    [sym_preproc_line] = STATE(5056),
    [aux_sym_type_argument_repeat1] = STATE(5056),
    [ts_builtin_sym_end] = ACTIONS(6310),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_namespace] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(9479),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5057)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5057),
    [sym_compiler_directive_decl] = STATE(5057),
    [sym_fsi_directive_decl] = STATE(5057),
    [sym_preproc_line] = STATE(5057),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(9348),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(9346),
    [anon_sym_module] = ACTIONS(9346),
    [anon_sym_null] = ACTIONS(9346),
    [anon_sym_begin] = ACTIONS(9346),
    [anon_sym_open] = ACTIONS(9346),
    [anon_sym_type] = ACTIONS(9346),
    [anon_sym_LBRACK_LT] = ACTIONS(9348),
    [anon_sym_do] = ACTIONS(9346),
    [anon_sym_and] = ACTIONS(9346),
    [anon_sym_let] = ACTIONS(9346),
    [anon_sym_let_BANG] = ACTIONS(9348),
    [aux_sym_access_modifier_token1] = ACTIONS(9348),
    [anon_sym_LPAREN] = ACTIONS(9346),
    [anon_sym_PIPE] = ACTIONS(9348),
    [anon_sym_AMP] = ACTIONS(9346),
    [anon_sym_LBRACK] = ACTIONS(9346),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9348),
    [anon_sym_LBRACE] = ACTIONS(9346),
    [anon_sym_LT_AT] = ACTIONS(9346),
    [anon_sym_LT_AT_AT] = ACTIONS(9348),
    [anon_sym_CARET] = ACTIONS(9348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9348),
    [anon_sym_new] = ACTIONS(9346),
    [anon_sym_with] = ACTIONS(9346),
    [anon_sym_return] = ACTIONS(9346),
    [anon_sym_return_BANG] = ACTIONS(9348),
    [anon_sym_yield] = ACTIONS(9346),
    [anon_sym_yield_BANG] = ACTIONS(9348),
    [anon_sym_lazy] = ACTIONS(9346),
    [anon_sym_assert] = ACTIONS(9346),
    [anon_sym_upcast] = ACTIONS(9346),
    [anon_sym_downcast] = ACTIONS(9346),
    [anon_sym_fixed] = ACTIONS(9346),
    [anon_sym_for] = ACTIONS(9346),
    [anon_sym_while] = ACTIONS(9346),
    [anon_sym_while_BANG] = ACTIONS(9348),
    [anon_sym_if] = ACTIONS(9346),
    [anon_sym_fun] = ACTIONS(9346),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(9346),
    [anon_sym_match] = ACTIONS(9346),
    [anon_sym_match_BANG] = ACTIONS(9348),
    [anon_sym_function] = ACTIONS(9346),
    [anon_sym_use] = ACTIONS(9346),
    [anon_sym_use_BANG] = ACTIONS(9348),
    [anon_sym_do_BANG] = ACTIONS(9348),
    [anon_sym_AMP_AMP] = ACTIONS(9348),
    [anon_sym_STAR] = ACTIONS(9348),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(9346),
    [anon_sym_static] = ACTIONS(9346),
    [anon_sym_member] = ACTIONS(9346),
    [anon_sym_exception] = ACTIONS(9346),
    [anon_sym_interface] = ACTIONS(9346),
    [anon_sym_abstract] = ACTIONS(9346),
    [anon_sym_val] = ACTIONS(9346),
    [anon_sym_override] = ACTIONS(9346),
    [anon_sym_extern] = ACTIONS(9346),
    [aux_sym_char_token1] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9346),
    [anon_sym_DQUOTE] = ACTIONS(9346),
    [anon_sym_AT_DQUOTE] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [sym_bool] = ACTIONS(9346),
    [sym_unit] = ACTIONS(9348),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9346),
    [sym_op_identifier] = ACTIONS(9348),
    [anon_sym_PLUS] = ACTIONS(9346),
    [anon_sym_DASH] = ACTIONS(9346),
    [anon_sym_PLUS_DOT] = ACTIONS(9348),
    [anon_sym_DASH_DOT] = ACTIONS(9348),
    [anon_sym_PERCENT] = ACTIONS(9346),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9348),
    [anon_sym_TILDE] = ACTIONS(9348),
    [aux_sym_prefix_op_token1] = ACTIONS(9348),
    [sym_int] = ACTIONS(9346),
    [sym_xint] = ACTIONS(9348),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9348),
    [anon_sym_POUNDload] = ACTIONS(9348),
    [anon_sym_POUNDtime] = ACTIONS(9348),
    [anon_sym_POUNDI] = ACTIONS(9348),
    [anon_sym_POUNDhelp] = ACTIONS(9348),
    [anon_sym_POUNDquit] = ACTIONS(9348),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9348),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9348),
  },
  [STATE(5058)] = {
    [sym_block_comment] = STATE(5058),
    [sym_compiler_directive_decl] = STATE(5058),
    [sym_fsi_directive_decl] = STATE(5058),
    [sym_preproc_line] = STATE(5058),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_with] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(5059)] = {
    [sym_block_comment] = STATE(5059),
    [sym_compiler_directive_decl] = STATE(5059),
    [sym_fsi_directive_decl] = STATE(5059),
    [sym_preproc_line] = STATE(5059),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [anon_sym_POUNDendif] = ACTIONS(6811),
    [anon_sym_POUNDelse] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5060)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5060),
    [sym_compiler_directive_decl] = STATE(5060),
    [sym_fsi_directive_decl] = STATE(5060),
    [sym_preproc_line] = STATE(5060),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(6255),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(6257),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_PIPE] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6255),
    [anon_sym_CARET] = ACTIONS(6255),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_AMP_AMP] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6255),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6255),
    [anon_sym_DASH_DOT] = ACTIONS(6255),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6255),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6255),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
  },
  [STATE(5061)] = {
    [sym_block_comment] = STATE(5061),
    [sym_compiler_directive_decl] = STATE(5061),
    [sym_fsi_directive_decl] = STATE(5061),
    [sym_preproc_line] = STATE(5061),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_DOT_DOT] = ACTIONS(6944),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [anon_sym_DOT_DOT2] = ACTIONS(6942),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__dedent] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(5062)] = {
    [sym_block_comment] = STATE(5062),
    [sym_compiler_directive_decl] = STATE(5062),
    [sym_fsi_directive_decl] = STATE(5062),
    [sym_preproc_line] = STATE(5062),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__dedent] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
    [sym__in] = ACTIONS(6297),
  },
  [STATE(5063)] = {
    [sym_block_comment] = STATE(5063),
    [sym_compiler_directive_decl] = STATE(5063),
    [sym_fsi_directive_decl] = STATE(5063),
    [sym_preproc_line] = STATE(5063),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9482),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
    [sym__in] = ACTIONS(6832),
  },
  [STATE(5064)] = {
    [sym_block_comment] = STATE(5064),
    [sym_compiler_directive_decl] = STATE(5064),
    [sym_fsi_directive_decl] = STATE(5064),
    [sym_preproc_line] = STATE(5064),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_as] = ACTIONS(6597),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9170),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5065)] = {
    [sym_block_comment] = STATE(5065),
    [sym_compiler_directive_decl] = STATE(5065),
    [sym_fsi_directive_decl] = STATE(5065),
    [sym_preproc_line] = STATE(5065),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__dedent] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
    [sym__in] = ACTIONS(6838),
  },
  [STATE(5066)] = {
    [sym_block_comment] = STATE(5066),
    [sym_compiler_directive_decl] = STATE(5066),
    [sym_fsi_directive_decl] = STATE(5066),
    [sym_preproc_line] = STATE(5066),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
    [sym__in] = ACTIONS(6651),
  },
  [STATE(5067)] = {
    [sym_block_comment] = STATE(5067),
    [sym_compiler_directive_decl] = STATE(5067),
    [sym_fsi_directive_decl] = STATE(5067),
    [sym_preproc_line] = STATE(5067),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_DOT_DOT] = ACTIONS(6948),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [anon_sym_DOT_DOT2] = ACTIONS(6946),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__dedent] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(5068)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(5068),
    [sym_compiler_directive_decl] = STATE(5068),
    [sym_fsi_directive_decl] = STATE(5068),
    [sym_preproc_line] = STATE(5068),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9208),
    [anon_sym_module] = ACTIONS(9208),
    [anon_sym_null] = ACTIONS(9208),
    [anon_sym_begin] = ACTIONS(9208),
    [anon_sym_open] = ACTIONS(9208),
    [anon_sym_type] = ACTIONS(9208),
    [anon_sym_LBRACK_LT] = ACTIONS(9210),
    [anon_sym_do] = ACTIONS(9208),
    [anon_sym_and] = ACTIONS(9208),
    [anon_sym_let] = ACTIONS(9208),
    [anon_sym_let_BANG] = ACTIONS(9210),
    [aux_sym_access_modifier_token1] = ACTIONS(9210),
    [anon_sym_LPAREN] = ACTIONS(9208),
    [anon_sym_PIPE] = ACTIONS(9210),
    [anon_sym_AMP] = ACTIONS(9208),
    [anon_sym_LBRACK] = ACTIONS(9208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9210),
    [anon_sym_LBRACE] = ACTIONS(9208),
    [anon_sym_LT_AT] = ACTIONS(9208),
    [anon_sym_LT_AT_AT] = ACTIONS(9210),
    [anon_sym_CARET] = ACTIONS(9210),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9210),
    [anon_sym_new] = ACTIONS(9208),
    [anon_sym_with] = ACTIONS(9208),
    [anon_sym_return] = ACTIONS(9208),
    [anon_sym_return_BANG] = ACTIONS(9210),
    [anon_sym_yield] = ACTIONS(9208),
    [anon_sym_yield_BANG] = ACTIONS(9210),
    [anon_sym_lazy] = ACTIONS(9208),
    [anon_sym_assert] = ACTIONS(9208),
    [anon_sym_upcast] = ACTIONS(9208),
    [anon_sym_downcast] = ACTIONS(9208),
    [anon_sym_fixed] = ACTIONS(9208),
    [anon_sym_for] = ACTIONS(9208),
    [anon_sym_while] = ACTIONS(9208),
    [anon_sym_while_BANG] = ACTIONS(9210),
    [anon_sym_if] = ACTIONS(9208),
    [anon_sym_fun] = ACTIONS(9208),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(9208),
    [anon_sym_match] = ACTIONS(9208),
    [anon_sym_match_BANG] = ACTIONS(9210),
    [anon_sym_function] = ACTIONS(9208),
    [anon_sym_use] = ACTIONS(9208),
    [anon_sym_use_BANG] = ACTIONS(9210),
    [anon_sym_do_BANG] = ACTIONS(9210),
    [anon_sym_AMP_AMP] = ACTIONS(9210),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(9208),
    [anon_sym_static] = ACTIONS(9208),
    [anon_sym_member] = ACTIONS(9208),
    [anon_sym_exception] = ACTIONS(9208),
    [anon_sym_interface] = ACTIONS(9208),
    [anon_sym_abstract] = ACTIONS(9208),
    [anon_sym_val] = ACTIONS(9208),
    [anon_sym_override] = ACTIONS(9208),
    [anon_sym_extern] = ACTIONS(9208),
    [aux_sym_char_token1] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9208),
    [anon_sym_DQUOTE] = ACTIONS(9208),
    [anon_sym_AT_DQUOTE] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [sym_bool] = ACTIONS(9208),
    [sym_unit] = ACTIONS(9210),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9208),
    [sym_op_identifier] = ACTIONS(9210),
    [anon_sym_PLUS] = ACTIONS(9208),
    [anon_sym_DASH] = ACTIONS(9208),
    [anon_sym_PLUS_DOT] = ACTIONS(9210),
    [anon_sym_DASH_DOT] = ACTIONS(9210),
    [anon_sym_PERCENT] = ACTIONS(9208),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9210),
    [anon_sym_TILDE] = ACTIONS(9210),
    [aux_sym_prefix_op_token1] = ACTIONS(9210),
    [sym_int] = ACTIONS(9208),
    [sym_xint] = ACTIONS(9210),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9210),
    [anon_sym_POUNDload] = ACTIONS(9210),
    [anon_sym_POUNDtime] = ACTIONS(9210),
    [anon_sym_POUNDI] = ACTIONS(9210),
    [anon_sym_POUNDhelp] = ACTIONS(9210),
    [anon_sym_POUNDquit] = ACTIONS(9210),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9210),
    [anon_sym_POUNDendif] = ACTIONS(9210),
    [anon_sym_POUNDelse] = ACTIONS(9210),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9210),
  },
  [STATE(5069)] = {
    [sym_block_comment] = STATE(5069),
    [sym_compiler_directive_decl] = STATE(5069),
    [sym_fsi_directive_decl] = STATE(5069),
    [sym_preproc_line] = STATE(5069),
    [aux_sym_rules_repeat1] = STATE(5202),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9484),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9486),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__multi_dollar_interp_end] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5070)] = {
    [sym_block_comment] = STATE(5070),
    [sym_compiler_directive_decl] = STATE(5070),
    [sym_fsi_directive_decl] = STATE(5070),
    [sym_preproc_line] = STATE(5070),
    [aux_sym_rules_repeat1] = STATE(5287),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9393),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9489),
    [sym__quoted_close] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(5071)] = {
    [sym_block_comment] = STATE(5071),
    [sym_compiler_directive_decl] = STATE(5071),
    [sym_fsi_directive_decl] = STATE(5071),
    [sym_preproc_line] = STATE(5071),
    [aux_sym_rules_repeat1] = STATE(5208),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9484),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9486),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__multi_dollar_interp_end] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5072)] = {
    [sym_block_comment] = STATE(5072),
    [sym_compiler_directive_decl] = STATE(5072),
    [sym_fsi_directive_decl] = STATE(5072),
    [sym_preproc_line] = STATE(5072),
    [aux_sym_rules_repeat1] = STATE(5244),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9492),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_with] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9494),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5073)] = {
    [sym_block_comment] = STATE(5073),
    [sym_compiler_directive_decl] = STATE(5073),
    [sym_fsi_directive_decl] = STATE(5073),
    [sym_preproc_line] = STATE(5073),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_DOT_DOT] = ACTIONS(6952),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [anon_sym_DOT_DOT2] = ACTIONS(6950),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__dedent] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(5074)] = {
    [sym_block_comment] = STATE(5074),
    [sym_compiler_directive_decl] = STATE(5074),
    [sym_fsi_directive_decl] = STATE(5074),
    [sym_preproc_line] = STATE(5074),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__untyped_quoted_close] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5075)] = {
    [sym_block_comment] = STATE(5075),
    [sym_compiler_directive_decl] = STATE(5075),
    [sym_fsi_directive_decl] = STATE(5075),
    [sym_preproc_line] = STATE(5075),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_as] = ACTIONS(6768),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9497),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5076)] = {
    [sym_block_comment] = STATE(5076),
    [sym_compiler_directive_decl] = STATE(5076),
    [sym_fsi_directive_decl] = STATE(5076),
    [sym_preproc_line] = STATE(5076),
    [aux_sym_sequential_expression_repeat1] = STATE(5076),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [anon_sym_DOT_DOT2] = ACTIONS(6798),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9499),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5077)] = {
    [sym_block_comment] = STATE(5077),
    [sym_compiler_directive_decl] = STATE(5077),
    [sym_fsi_directive_decl] = STATE(5077),
    [sym_preproc_line] = STATE(5077),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__multi_dollar_interp_end] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(5078)] = {
    [sym_block_comment] = STATE(5078),
    [sym_compiler_directive_decl] = STATE(5078),
    [sym_fsi_directive_decl] = STATE(5078),
    [sym_preproc_line] = STATE(5078),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_DOT_DOT] = ACTIONS(7053),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [anon_sym_DOT_DOT2] = ACTIONS(7055),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__dedent] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(5079)] = {
    [sym_block_comment] = STATE(5079),
    [sym_compiler_directive_decl] = STATE(5079),
    [sym_fsi_directive_decl] = STATE(5079),
    [sym_preproc_line] = STATE(5079),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5951),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [anon_sym_DOT_DOT2] = ACTIONS(5953),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5080)] = {
    [sym_block_comment] = STATE(5080),
    [sym_compiler_directive_decl] = STATE(5080),
    [sym_fsi_directive_decl] = STATE(5080),
    [sym_preproc_line] = STATE(5080),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [anon_sym_POUNDendif] = ACTIONS(6822),
    [anon_sym_POUNDelse] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5081)] = {
    [sym_block_comment] = STATE(5081),
    [sym_compiler_directive_decl] = STATE(5081),
    [sym_fsi_directive_decl] = STATE(5081),
    [sym_preproc_line] = STATE(5081),
    [aux_sym_rules_repeat1] = STATE(5070),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9393),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9395),
    [sym__quoted_close] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5082)] = {
    [sym_block_comment] = STATE(5082),
    [sym_compiler_directive_decl] = STATE(5082),
    [sym_fsi_directive_decl] = STATE(5082),
    [sym_preproc_line] = STATE(5082),
    [aux_sym_long_identifier_repeat1] = STATE(5308),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9502),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_DOT_DOT] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(5083)] = {
    [sym_block_comment] = STATE(5083),
    [sym_compiler_directive_decl] = STATE(5083),
    [sym_fsi_directive_decl] = STATE(5083),
    [sym_preproc_line] = STATE(5083),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_DOT_DOT] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__dedent] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5084)] = {
    [sym_block_comment] = STATE(5084),
    [sym_compiler_directive_decl] = STATE(5084),
    [sym_fsi_directive_decl] = STATE(5084),
    [sym_preproc_line] = STATE(5084),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [anon_sym_POUNDelse] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5085)] = {
    [sym_block_comment] = STATE(5085),
    [sym_compiler_directive_decl] = STATE(5085),
    [sym_fsi_directive_decl] = STATE(5085),
    [sym_preproc_line] = STATE(5085),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_DOT_DOT] = ACTIONS(6959),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [anon_sym_DOT_DOT2] = ACTIONS(6957),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__dedent] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(5086)] = {
    [sym_block_comment] = STATE(5086),
    [sym_compiler_directive_decl] = STATE(5086),
    [sym_fsi_directive_decl] = STATE(5086),
    [sym_preproc_line] = STATE(5086),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9506),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_DASH_GT] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5087)] = {
    [sym_block_comment] = STATE(5087),
    [sym_compiler_directive_decl] = STATE(5087),
    [sym_fsi_directive_decl] = STATE(5087),
    [sym_preproc_line] = STATE(5087),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__dedent] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
    [sym__in] = ACTIONS(6811),
  },
  [STATE(5088)] = {
    [sym_block_comment] = STATE(5088),
    [sym_compiler_directive_decl] = STATE(5088),
    [sym_fsi_directive_decl] = STATE(5088),
    [sym_preproc_line] = STATE(5088),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_DOT_DOT] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__dedent] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5089)] = {
    [sym_block_comment] = STATE(5089),
    [sym_compiler_directive_decl] = STATE(5089),
    [sym_fsi_directive_decl] = STATE(5089),
    [sym_preproc_line] = STATE(5089),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_DOT_DOT] = ACTIONS(6888),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [anon_sym_DOT_DOT2] = ACTIONS(6886),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__dedent] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(5090)] = {
    [sym_block_comment] = STATE(5090),
    [sym_compiler_directive_decl] = STATE(5090),
    [sym_fsi_directive_decl] = STATE(5090),
    [sym_preproc_line] = STATE(5090),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__dedent] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
    [sym__in] = ACTIONS(6822),
  },
  [STATE(5091)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5091),
    [sym_compiler_directive_decl] = STATE(5091),
    [sym_fsi_directive_decl] = STATE(5091),
    [sym_preproc_line] = STATE(5091),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(6209),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(6213),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_PIPE] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6209),
    [anon_sym_CARET] = ACTIONS(6209),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_AMP_AMP] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6209),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6209),
    [anon_sym_DASH_DOT] = ACTIONS(6209),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6209),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6209),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
  },
  [STATE(5092)] = {
    [sym_block_comment] = STATE(5092),
    [sym_compiler_directive_decl] = STATE(5092),
    [sym_fsi_directive_decl] = STATE(5092),
    [sym_preproc_line] = STATE(5092),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_DOT_DOT] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__dedent] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5093)] = {
    [sym_block_comment] = STATE(5093),
    [sym_compiler_directive_decl] = STATE(5093),
    [sym_fsi_directive_decl] = STATE(5093),
    [sym_preproc_line] = STATE(5093),
    [aux_sym_long_identifier_repeat1] = STATE(5205),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9508),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9344),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5094)] = {
    [sym_block_comment] = STATE(5094),
    [sym_compiler_directive_decl] = STATE(5094),
    [sym_fsi_directive_decl] = STATE(5094),
    [sym_preproc_line] = STATE(5094),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_DOT_DOT] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__dedent] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5095)] = {
    [sym_block_comment] = STATE(5095),
    [sym_compiler_directive_decl] = STATE(5095),
    [sym_fsi_directive_decl] = STATE(5095),
    [sym_preproc_line] = STATE(5095),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_DOT_DOT] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__dedent] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5096)] = {
    [sym_block_comment] = STATE(5096),
    [sym_compiler_directive_decl] = STATE(5096),
    [sym_fsi_directive_decl] = STATE(5096),
    [sym_preproc_line] = STATE(5096),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9510),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_DASH_GT] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_DOT_DOT] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5097)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5097),
    [sym_compiler_directive_decl] = STATE(5097),
    [sym_fsi_directive_decl] = STATE(5097),
    [sym_preproc_line] = STATE(5097),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(6195),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_PIPE] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6195),
    [anon_sym_CARET] = ACTIONS(6195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_AMP_AMP] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6195),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6195),
    [anon_sym_DASH_DOT] = ACTIONS(6195),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6195),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6195),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
  },
  [STATE(5098)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5098),
    [sym_compiler_directive_decl] = STATE(5098),
    [sym_fsi_directive_decl] = STATE(5098),
    [sym_preproc_line] = STATE(5098),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(6259),
    [sym_identifier] = ACTIONS(9381),
    [anon_sym_namespace] = ACTIONS(6261),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_PIPE] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6259),
    [anon_sym_CARET] = ACTIONS(6259),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_AMP_AMP] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6259),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6259),
    [anon_sym_DASH_DOT] = ACTIONS(6259),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6259),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6259),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
  },
  [STATE(5099)] = {
    [sym_block_comment] = STATE(5099),
    [sym_compiler_directive_decl] = STATE(5099),
    [sym_fsi_directive_decl] = STATE(5099),
    [sym_preproc_line] = STATE(5099),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
    [sym__in] = ACTIONS(6832),
  },
  [STATE(5100)] = {
    [sym_block_comment] = STATE(5100),
    [sym_compiler_directive_decl] = STATE(5100),
    [sym_fsi_directive_decl] = STATE(5100),
    [sym_preproc_line] = STATE(5100),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5101)] = {
    [sym_block_comment] = STATE(5101),
    [sym_compiler_directive_decl] = STATE(5101),
    [sym_fsi_directive_decl] = STATE(5101),
    [sym_preproc_line] = STATE(5101),
    [aux_sym_sequential_expression_repeat1] = STATE(5024),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_DASH_GT] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_DOT_DOT] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5102)] = {
    [sym_block_comment] = STATE(5102),
    [sym_compiler_directive_decl] = STATE(5102),
    [sym_fsi_directive_decl] = STATE(5102),
    [sym_preproc_line] = STATE(5102),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_DOT_DOT] = ACTIONS(6975),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [anon_sym_DOT_DOT2] = ACTIONS(6973),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__dedent] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(5103)] = {
    [sym_block_comment] = STATE(5103),
    [sym_compiler_directive_decl] = STATE(5103),
    [sym_fsi_directive_decl] = STATE(5103),
    [sym_preproc_line] = STATE(5103),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_DOT_DOT] = ACTIONS(7069),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [anon_sym_DOT_DOT2] = ACTIONS(7071),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__dedent] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(5104)] = {
    [sym_block_comment] = STATE(5104),
    [sym_compiler_directive_decl] = STATE(5104),
    [sym_fsi_directive_decl] = STATE(5104),
    [sym_preproc_line] = STATE(5104),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_DOT_DOT] = ACTIONS(6963),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [anon_sym_DOT_DOT2] = ACTIONS(6961),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__dedent] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(5105)] = {
    [sym_block_comment] = STATE(5105),
    [sym_compiler_directive_decl] = STATE(5105),
    [sym_fsi_directive_decl] = STATE(5105),
    [sym_preproc_line] = STATE(5105),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_DOT_DOT] = ACTIONS(7141),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [anon_sym_DOT_DOT2] = ACTIONS(7143),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__dedent] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(5106)] = {
    [sym_block_comment] = STATE(5106),
    [sym_compiler_directive_decl] = STATE(5106),
    [sym_fsi_directive_decl] = STATE(5106),
    [sym_preproc_line] = STATE(5106),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9156),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5107)] = {
    [sym_block_comment] = STATE(5107),
    [sym_compiler_directive_decl] = STATE(5107),
    [sym_fsi_directive_decl] = STATE(5107),
    [sym_preproc_line] = STATE(5107),
    [aux_sym_sequential_expression_repeat1] = STATE(5076),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [anon_sym_DOT_DOT2] = ACTIONS(6738),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5108)] = {
    [sym_block_comment] = STATE(5108),
    [sym_compiler_directive_decl] = STATE(5108),
    [sym_fsi_directive_decl] = STATE(5108),
    [sym_preproc_line] = STATE(5108),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_as] = ACTIONS(6863),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_with] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5109)] = {
    [sym_block_comment] = STATE(5109),
    [sym_compiler_directive_decl] = STATE(5109),
    [sym_fsi_directive_decl] = STATE(5109),
    [sym_preproc_line] = STATE(5109),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9512),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5110)] = {
    [sym_block_comment] = STATE(5110),
    [sym_compiler_directive_decl] = STATE(5110),
    [sym_fsi_directive_decl] = STATE(5110),
    [sym_preproc_line] = STATE(5110),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_DOT_DOT] = ACTIONS(7073),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [anon_sym_DOT_DOT2] = ACTIONS(7075),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__dedent] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(5111)] = {
    [sym_block_comment] = STATE(5111),
    [sym_compiler_directive_decl] = STATE(5111),
    [sym_fsi_directive_decl] = STATE(5111),
    [sym_preproc_line] = STATE(5111),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [anon_sym_POUNDelse] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5112)] = {
    [sym_block_comment] = STATE(5112),
    [sym_compiler_directive_decl] = STATE(5112),
    [sym_fsi_directive_decl] = STATE(5112),
    [sym_preproc_line] = STATE(5112),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_DOT_DOT] = ACTIONS(7077),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [anon_sym_DOT_DOT2] = ACTIONS(7079),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__dedent] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(5113)] = {
    [sym_block_comment] = STATE(5113),
    [sym_compiler_directive_decl] = STATE(5113),
    [sym_fsi_directive_decl] = STATE(5113),
    [sym_preproc_line] = STATE(5113),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_DOT_DOT] = ACTIONS(7085),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [anon_sym_DOT_DOT2] = ACTIONS(7087),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__dedent] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(5114)] = {
    [sym_block_comment] = STATE(5114),
    [sym_compiler_directive_decl] = STATE(5114),
    [sym_fsi_directive_decl] = STATE(5114),
    [sym_preproc_line] = STATE(5114),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_DOT_DOT] = ACTIONS(7089),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [anon_sym_DOT_DOT2] = ACTIONS(7091),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__dedent] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(5115)] = {
    [sym_block_comment] = STATE(5115),
    [sym_compiler_directive_decl] = STATE(5115),
    [sym_fsi_directive_decl] = STATE(5115),
    [sym_preproc_line] = STATE(5115),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_DOT_DOT] = ACTIONS(7093),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [anon_sym_DOT_DOT2] = ACTIONS(7095),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__dedent] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(5116)] = {
    [sym_block_comment] = STATE(5116),
    [sym_compiler_directive_decl] = STATE(5116),
    [sym_fsi_directive_decl] = STATE(5116),
    [sym_preproc_line] = STATE(5116),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_DOT_DOT] = ACTIONS(7097),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [anon_sym_DOT_DOT2] = ACTIONS(7099),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__dedent] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(5117)] = {
    [sym_block_comment] = STATE(5117),
    [sym_compiler_directive_decl] = STATE(5117),
    [sym_fsi_directive_decl] = STATE(5117),
    [sym_preproc_line] = STATE(5117),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_DOT_DOT] = ACTIONS(7101),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [anon_sym_DOT_DOT2] = ACTIONS(7103),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__dedent] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(5118)] = {
    [sym_block_comment] = STATE(5118),
    [sym_compiler_directive_decl] = STATE(5118),
    [sym_fsi_directive_decl] = STATE(5118),
    [sym_preproc_line] = STATE(5118),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_DOT_DOT] = ACTIONS(7105),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [anon_sym_DOT_DOT2] = ACTIONS(7107),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__dedent] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(5119)] = {
    [sym_block_comment] = STATE(5119),
    [sym_compiler_directive_decl] = STATE(5119),
    [sym_fsi_directive_decl] = STATE(5119),
    [sym_preproc_line] = STATE(5119),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_DOT_DOT] = ACTIONS(7109),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [anon_sym_DOT_DOT2] = ACTIONS(7111),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__dedent] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(5120)] = {
    [sym_block_comment] = STATE(5120),
    [sym_compiler_directive_decl] = STATE(5120),
    [sym_fsi_directive_decl] = STATE(5120),
    [sym_preproc_line] = STATE(5120),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_DOT_DOT] = ACTIONS(7113),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [anon_sym_DOT_DOT2] = ACTIONS(7115),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__dedent] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(5121)] = {
    [sym_block_comment] = STATE(5121),
    [sym_compiler_directive_decl] = STATE(5121),
    [sym_fsi_directive_decl] = STATE(5121),
    [sym_preproc_line] = STATE(5121),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_DOT_DOT] = ACTIONS(7117),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [anon_sym_DOT_DOT2] = ACTIONS(7119),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__dedent] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(5122)] = {
    [sym_block_comment] = STATE(5122),
    [sym_compiler_directive_decl] = STATE(5122),
    [sym_fsi_directive_decl] = STATE(5122),
    [sym_preproc_line] = STATE(5122),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_DOT_DOT] = ACTIONS(7121),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [anon_sym_DOT_DOT2] = ACTIONS(7123),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__dedent] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(5123)] = {
    [sym_block_comment] = STATE(5123),
    [sym_compiler_directive_decl] = STATE(5123),
    [sym_fsi_directive_decl] = STATE(5123),
    [sym_preproc_line] = STATE(5123),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_DOT_DOT] = ACTIONS(6936),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [anon_sym_DOT_DOT2] = ACTIONS(6934),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__dedent] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(5124)] = {
    [sym_block_comment] = STATE(5124),
    [sym_compiler_directive_decl] = STATE(5124),
    [sym_fsi_directive_decl] = STATE(5124),
    [sym_preproc_line] = STATE(5124),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_DOT_DOT] = ACTIONS(6940),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [anon_sym_DOT_DOT2] = ACTIONS(6938),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__dedent] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(5125)] = {
    [sym_block_comment] = STATE(5125),
    [sym_compiler_directive_decl] = STATE(5125),
    [sym_fsi_directive_decl] = STATE(5125),
    [sym_preproc_line] = STATE(5125),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(8039),
    [aux_sym_decimal_token1] = ACTIONS(7723),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5126)] = {
    [sym_block_comment] = STATE(5126),
    [sym_compiler_directive_decl] = STATE(5126),
    [sym_fsi_directive_decl] = STATE(5126),
    [sym_preproc_line] = STATE(5126),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_DOT_DOT] = ACTIONS(7015),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [anon_sym_DOT_DOT2] = ACTIONS(7013),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__dedent] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(5127)] = {
    [sym_block_comment] = STATE(5127),
    [sym_compiler_directive_decl] = STATE(5127),
    [sym_fsi_directive_decl] = STATE(5127),
    [sym_preproc_line] = STATE(5127),
    [aux_sym_sequential_expression_repeat1] = STATE(5247),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_DOT_DOT] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5128)] = {
    [sym_block_comment] = STATE(5128),
    [sym_compiler_directive_decl] = STATE(5128),
    [sym_fsi_directive_decl] = STATE(5128),
    [sym_preproc_line] = STATE(5128),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_DOT_DOT] = ACTIONS(7031),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [anon_sym_DOT_DOT2] = ACTIONS(7029),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__dedent] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(5129)] = {
    [sym_block_comment] = STATE(5129),
    [sym_compiler_directive_decl] = STATE(5129),
    [sym_fsi_directive_decl] = STATE(5129),
    [sym_preproc_line] = STATE(5129),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9514),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__quoted_close] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5130)] = {
    [sym_block_comment] = STATE(5130),
    [sym_compiler_directive_decl] = STATE(5130),
    [sym_fsi_directive_decl] = STATE(5130),
    [sym_preproc_line] = STATE(5130),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_DOT_DOT] = ACTIONS(7047),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [anon_sym_DOT_DOT2] = ACTIONS(7045),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__dedent] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(5131)] = {
    [sym_block_comment] = STATE(5131),
    [sym_compiler_directive_decl] = STATE(5131),
    [sym_fsi_directive_decl] = STATE(5131),
    [sym_preproc_line] = STATE(5131),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_DOT_DOT] = ACTIONS(7051),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [anon_sym_DOT_DOT2] = ACTIONS(7049),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__dedent] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(5132)] = {
    [sym_block_comment] = STATE(5132),
    [sym_compiler_directive_decl] = STATE(5132),
    [sym_fsi_directive_decl] = STATE(5132),
    [sym_preproc_line] = STATE(5132),
    [aux_sym_sequential_expression_repeat1] = STATE(5257),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [anon_sym_POUNDelse] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5133)] = {
    [sym_block_comment] = STATE(5133),
    [sym_compiler_directive_decl] = STATE(5133),
    [sym_fsi_directive_decl] = STATE(5133),
    [sym_preproc_line] = STATE(5133),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_DOT_DOT] = ACTIONS(7145),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [anon_sym_DOT_DOT2] = ACTIONS(7147),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__dedent] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(5134)] = {
    [sym_block_comment] = STATE(5134),
    [sym_compiler_directive_decl] = STATE(5134),
    [sym_fsi_directive_decl] = STATE(5134),
    [sym_preproc_line] = STATE(5134),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_DOT_DOT] = ACTIONS(7063),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [anon_sym_DOT_DOT2] = ACTIONS(7061),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__dedent] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(5135)] = {
    [sym_block_comment] = STATE(5135),
    [sym_compiler_directive_decl] = STATE(5135),
    [sym_fsi_directive_decl] = STATE(5135),
    [sym_preproc_line] = STATE(5135),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_as] = ACTIONS(6867),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5136)] = {
    [sym_block_comment] = STATE(5136),
    [sym_compiler_directive_decl] = STATE(5136),
    [sym_fsi_directive_decl] = STATE(5136),
    [sym_preproc_line] = STATE(5136),
    [aux_sym_rules_repeat1] = STATE(5189),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9516),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_DOT_DOT] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9518),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5137)] = {
    [sym_block_comment] = STATE(5137),
    [sym_compiler_directive_decl] = STATE(5137),
    [sym_fsi_directive_decl] = STATE(5137),
    [sym_preproc_line] = STATE(5137),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__untyped_quoted_close] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5138)] = {
    [sym_block_comment] = STATE(5138),
    [sym_compiler_directive_decl] = STATE(5138),
    [sym_fsi_directive_decl] = STATE(5138),
    [sym_preproc_line] = STATE(5138),
    [aux_sym_sequential_expression_repeat1] = STATE(5138),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9521),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
    [sym__in] = ACTIONS(6798),
  },
  [STATE(5139)] = {
    [sym_block_comment] = STATE(5139),
    [sym_compiler_directive_decl] = STATE(5139),
    [sym_fsi_directive_decl] = STATE(5139),
    [sym_preproc_line] = STATE(5139),
    [aux_sym_long_identifier_repeat1] = STATE(5229),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9524),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(5140)] = {
    [sym_block_comment] = STATE(5140),
    [sym_compiler_directive_decl] = STATE(5140),
    [sym_fsi_directive_decl] = STATE(5140),
    [sym_preproc_line] = STATE(5140),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [anon_sym_POUNDelse] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5141)] = {
    [sym_block_comment] = STATE(5141),
    [sym_compiler_directive_decl] = STATE(5141),
    [sym_fsi_directive_decl] = STATE(5141),
    [sym_preproc_line] = STATE(5141),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_DOT_DOT] = ACTIONS(7067),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [anon_sym_DOT_DOT2] = ACTIONS(7065),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__dedent] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(5142)] = {
    [sym_block_comment] = STATE(5142),
    [sym_compiler_directive_decl] = STATE(5142),
    [sym_fsi_directive_decl] = STATE(5142),
    [sym_preproc_line] = STATE(5142),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_DOT_DOT] = ACTIONS(7125),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [anon_sym_DOT_DOT2] = ACTIONS(7127),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__dedent] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(5143)] = {
    [sym_block_comment] = STATE(5143),
    [sym_compiler_directive_decl] = STATE(5143),
    [sym_fsi_directive_decl] = STATE(5143),
    [sym_preproc_line] = STATE(5143),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_DOT_DOT] = ACTIONS(7129),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [anon_sym_DOT_DOT2] = ACTIONS(7131),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__dedent] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(5144)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5144),
    [sym_compiler_directive_decl] = STATE(5144),
    [sym_fsi_directive_decl] = STATE(5144),
    [sym_preproc_line] = STATE(5144),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(9346),
    [anon_sym_null] = ACTIONS(9346),
    [anon_sym_begin] = ACTIONS(9346),
    [anon_sym_open] = ACTIONS(9346),
    [anon_sym_type] = ACTIONS(9346),
    [anon_sym_LBRACK_LT] = ACTIONS(9348),
    [anon_sym_do] = ACTIONS(9346),
    [anon_sym_and] = ACTIONS(9346),
    [anon_sym_let] = ACTIONS(9346),
    [anon_sym_let_BANG] = ACTIONS(9348),
    [aux_sym_access_modifier_token1] = ACTIONS(9348),
    [anon_sym_LPAREN] = ACTIONS(9346),
    [anon_sym_PIPE] = ACTIONS(9348),
    [anon_sym_AMP] = ACTIONS(9346),
    [anon_sym_LBRACK] = ACTIONS(9346),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9348),
    [anon_sym_LBRACE] = ACTIONS(9346),
    [anon_sym_LT_AT] = ACTIONS(9346),
    [anon_sym_LT_AT_AT] = ACTIONS(9348),
    [anon_sym_CARET] = ACTIONS(9348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9348),
    [anon_sym_new] = ACTIONS(9346),
    [anon_sym_with] = ACTIONS(9346),
    [anon_sym_return] = ACTIONS(9346),
    [anon_sym_return_BANG] = ACTIONS(9348),
    [anon_sym_yield] = ACTIONS(9346),
    [anon_sym_yield_BANG] = ACTIONS(9348),
    [anon_sym_lazy] = ACTIONS(9346),
    [anon_sym_assert] = ACTIONS(9346),
    [anon_sym_upcast] = ACTIONS(9346),
    [anon_sym_downcast] = ACTIONS(9346),
    [anon_sym_fixed] = ACTIONS(9346),
    [anon_sym_for] = ACTIONS(9346),
    [anon_sym_while] = ACTIONS(9346),
    [anon_sym_while_BANG] = ACTIONS(9348),
    [anon_sym_if] = ACTIONS(9346),
    [anon_sym_fun] = ACTIONS(9346),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(9346),
    [anon_sym_match] = ACTIONS(9346),
    [anon_sym_match_BANG] = ACTIONS(9348),
    [anon_sym_function] = ACTIONS(9346),
    [anon_sym_use] = ACTIONS(9346),
    [anon_sym_use_BANG] = ACTIONS(9348),
    [anon_sym_do_BANG] = ACTIONS(9348),
    [anon_sym_AMP_AMP] = ACTIONS(9348),
    [anon_sym_STAR] = ACTIONS(9348),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(9346),
    [anon_sym_static] = ACTIONS(9346),
    [anon_sym_member] = ACTIONS(9346),
    [anon_sym_exception] = ACTIONS(9346),
    [anon_sym_interface] = ACTIONS(9346),
    [anon_sym_abstract] = ACTIONS(9346),
    [anon_sym_val] = ACTIONS(9346),
    [anon_sym_override] = ACTIONS(9346),
    [anon_sym_extern] = ACTIONS(9346),
    [aux_sym_char_token1] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9346),
    [anon_sym_DQUOTE] = ACTIONS(9346),
    [anon_sym_AT_DQUOTE] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [sym_bool] = ACTIONS(9346),
    [sym_unit] = ACTIONS(9348),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9346),
    [sym_op_identifier] = ACTIONS(9348),
    [anon_sym_PLUS] = ACTIONS(9346),
    [anon_sym_DASH] = ACTIONS(9346),
    [anon_sym_PLUS_DOT] = ACTIONS(9348),
    [anon_sym_DASH_DOT] = ACTIONS(9348),
    [anon_sym_PERCENT] = ACTIONS(9346),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9348),
    [anon_sym_TILDE] = ACTIONS(9348),
    [aux_sym_prefix_op_token1] = ACTIONS(9348),
    [sym_int] = ACTIONS(9346),
    [sym_xint] = ACTIONS(9348),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9348),
    [anon_sym_POUNDload] = ACTIONS(9348),
    [anon_sym_POUNDtime] = ACTIONS(9348),
    [anon_sym_POUNDI] = ACTIONS(9348),
    [anon_sym_POUNDhelp] = ACTIONS(9348),
    [anon_sym_POUNDquit] = ACTIONS(9348),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9348),
    [anon_sym_POUNDendif] = ACTIONS(9348),
    [sym__newline] = ACTIONS(9348),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9348),
  },
  [STATE(5145)] = {
    [sym_block_comment] = STATE(5145),
    [sym_compiler_directive_decl] = STATE(5145),
    [sym_fsi_directive_decl] = STATE(5145),
    [sym_preproc_line] = STATE(5145),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_DOT_DOT] = ACTIONS(7133),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [anon_sym_DOT_DOT2] = ACTIONS(7135),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__dedent] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(5146)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5146),
    [sym_compiler_directive_decl] = STATE(5146),
    [sym_fsi_directive_decl] = STATE(5146),
    [sym_preproc_line] = STATE(5146),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(9352),
    [anon_sym_null] = ACTIONS(9352),
    [anon_sym_begin] = ACTIONS(9352),
    [anon_sym_open] = ACTIONS(9352),
    [anon_sym_type] = ACTIONS(9352),
    [anon_sym_LBRACK_LT] = ACTIONS(9354),
    [anon_sym_do] = ACTIONS(9352),
    [anon_sym_and] = ACTIONS(9352),
    [anon_sym_let] = ACTIONS(9352),
    [anon_sym_let_BANG] = ACTIONS(9354),
    [aux_sym_access_modifier_token1] = ACTIONS(9354),
    [anon_sym_LPAREN] = ACTIONS(9352),
    [anon_sym_PIPE] = ACTIONS(9354),
    [anon_sym_AMP] = ACTIONS(9352),
    [anon_sym_LBRACK] = ACTIONS(9352),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9354),
    [anon_sym_LBRACE] = ACTIONS(9352),
    [anon_sym_LT_AT] = ACTIONS(9352),
    [anon_sym_LT_AT_AT] = ACTIONS(9354),
    [anon_sym_CARET] = ACTIONS(9354),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9354),
    [anon_sym_new] = ACTIONS(9352),
    [anon_sym_with] = ACTIONS(9352),
    [anon_sym_return] = ACTIONS(9352),
    [anon_sym_return_BANG] = ACTIONS(9354),
    [anon_sym_yield] = ACTIONS(9352),
    [anon_sym_yield_BANG] = ACTIONS(9354),
    [anon_sym_lazy] = ACTIONS(9352),
    [anon_sym_assert] = ACTIONS(9352),
    [anon_sym_upcast] = ACTIONS(9352),
    [anon_sym_downcast] = ACTIONS(9352),
    [anon_sym_fixed] = ACTIONS(9352),
    [anon_sym_for] = ACTIONS(9352),
    [anon_sym_while] = ACTIONS(9352),
    [anon_sym_while_BANG] = ACTIONS(9354),
    [anon_sym_if] = ACTIONS(9352),
    [anon_sym_fun] = ACTIONS(9352),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(9352),
    [anon_sym_match] = ACTIONS(9352),
    [anon_sym_match_BANG] = ACTIONS(9354),
    [anon_sym_function] = ACTIONS(9352),
    [anon_sym_use] = ACTIONS(9352),
    [anon_sym_use_BANG] = ACTIONS(9354),
    [anon_sym_do_BANG] = ACTIONS(9354),
    [anon_sym_AMP_AMP] = ACTIONS(9354),
    [anon_sym_STAR] = ACTIONS(9354),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(9352),
    [anon_sym_static] = ACTIONS(9352),
    [anon_sym_member] = ACTIONS(9352),
    [anon_sym_exception] = ACTIONS(9352),
    [anon_sym_interface] = ACTIONS(9352),
    [anon_sym_abstract] = ACTIONS(9352),
    [anon_sym_val] = ACTIONS(9352),
    [anon_sym_override] = ACTIONS(9352),
    [anon_sym_extern] = ACTIONS(9352),
    [aux_sym_char_token1] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9352),
    [anon_sym_DQUOTE] = ACTIONS(9352),
    [anon_sym_AT_DQUOTE] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [sym_bool] = ACTIONS(9352),
    [sym_unit] = ACTIONS(9354),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9352),
    [sym_op_identifier] = ACTIONS(9354),
    [anon_sym_PLUS] = ACTIONS(9352),
    [anon_sym_DASH] = ACTIONS(9352),
    [anon_sym_PLUS_DOT] = ACTIONS(9354),
    [anon_sym_DASH_DOT] = ACTIONS(9354),
    [anon_sym_PERCENT] = ACTIONS(9352),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9354),
    [anon_sym_TILDE] = ACTIONS(9354),
    [aux_sym_prefix_op_token1] = ACTIONS(9354),
    [sym_int] = ACTIONS(9352),
    [sym_xint] = ACTIONS(9354),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9354),
    [anon_sym_POUNDload] = ACTIONS(9354),
    [anon_sym_POUNDtime] = ACTIONS(9354),
    [anon_sym_POUNDI] = ACTIONS(9354),
    [anon_sym_POUNDhelp] = ACTIONS(9354),
    [anon_sym_POUNDquit] = ACTIONS(9354),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9354),
    [anon_sym_POUNDendif] = ACTIONS(9354),
    [sym__newline] = ACTIONS(9354),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9354),
  },
  [STATE(5147)] = {
    [sym_block_comment] = STATE(5147),
    [sym_compiler_directive_decl] = STATE(5147),
    [sym_fsi_directive_decl] = STATE(5147),
    [sym_preproc_line] = STATE(5147),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_DASH_GT] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_DOT_DOT] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5148)] = {
    [sym_block_comment] = STATE(5148),
    [sym_compiler_directive_decl] = STATE(5148),
    [sym_fsi_directive_decl] = STATE(5148),
    [sym_preproc_line] = STATE(5148),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5149)] = {
    [sym_block_comment] = STATE(5149),
    [sym_compiler_directive_decl] = STATE(5149),
    [sym_fsi_directive_decl] = STATE(5149),
    [sym_preproc_line] = STATE(5149),
    [aux_sym_type_argument_repeat1] = STATE(5047),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_PIPE] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6306),
    [anon_sym_CARET] = ACTIONS(6306),
    [anon_sym_or] = ACTIONS(9407),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6306),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_AMP_AMP] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6306),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6306),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6306),
    [anon_sym_DASH_DOT] = ACTIONS(6306),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6306),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6306),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
  },
  [STATE(5150)] = {
    [sym_block_comment] = STATE(5150),
    [sym_compiler_directive_decl] = STATE(5150),
    [sym_fsi_directive_decl] = STATE(5150),
    [sym_preproc_line] = STATE(5150),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_DOT_DOT] = ACTIONS(6913),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [anon_sym_DOT_DOT2] = ACTIONS(6911),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__dedent] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(5151)] = {
    [sym_block_comment] = STATE(5151),
    [sym_compiler_directive_decl] = STATE(5151),
    [sym_fsi_directive_decl] = STATE(5151),
    [sym_preproc_line] = STATE(5151),
    [aux_sym_type_argument_repeat1] = STATE(5224),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_PIPE] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6306),
    [anon_sym_CARET] = ACTIONS(6306),
    [anon_sym_or] = ACTIONS(9534),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6306),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_AMP_AMP] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6306),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6306),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6306),
    [anon_sym_DASH_DOT] = ACTIONS(6306),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6306),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6306),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
  },
  [STATE(5152)] = {
    [sym_block_comment] = STATE(5152),
    [sym_compiler_directive_decl] = STATE(5152),
    [sym_fsi_directive_decl] = STATE(5152),
    [sym_preproc_line] = STATE(5152),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9536),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__untyped_quoted_close] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(5153)] = {
    [sym_block_comment] = STATE(5153),
    [sym_compiler_directive_decl] = STATE(5153),
    [sym_fsi_directive_decl] = STATE(5153),
    [sym_preproc_line] = STATE(5153),
    [aux_sym_long_identifier_repeat1] = STATE(5194),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_GT] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9538),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(5154)] = {
    [sym_block_comment] = STATE(5154),
    [sym_compiler_directive_decl] = STATE(5154),
    [sym_fsi_directive_decl] = STATE(5154),
    [sym_preproc_line] = STATE(5154),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__untyped_quoted_close] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(5155)] = {
    [sym_block_comment] = STATE(5155),
    [sym_compiler_directive_decl] = STATE(5155),
    [sym_fsi_directive_decl] = STATE(5155),
    [sym_preproc_line] = STATE(5155),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [anon_sym_POUNDelse] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5156)] = {
    [sym_block_comment] = STATE(5156),
    [sym_compiler_directive_decl] = STATE(5156),
    [sym_fsi_directive_decl] = STATE(5156),
    [sym_preproc_line] = STATE(5156),
    [aux_sym_sequential_expression_repeat1] = STATE(5156),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_as] = ACTIONS(6800),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9542),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5157)] = {
    [sym_block_comment] = STATE(5157),
    [sym_compiler_directive_decl] = STATE(5157),
    [sym_fsi_directive_decl] = STATE(5157),
    [sym_preproc_line] = STATE(5157),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
    [sym__in] = ACTIONS(6614),
  },
  [STATE(5158)] = {
    [sym_block_comment] = STATE(5158),
    [sym_compiler_directive_decl] = STATE(5158),
    [sym_fsi_directive_decl] = STATE(5158),
    [sym_preproc_line] = STATE(5158),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_DOT_DOT] = ACTIONS(6917),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [anon_sym_DOT_DOT2] = ACTIONS(6915),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__dedent] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(5159)] = {
    [sym_block_comment] = STATE(5159),
    [sym_compiler_directive_decl] = STATE(5159),
    [sym_fsi_directive_decl] = STATE(5159),
    [sym_preproc_line] = STATE(5159),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [anon_sym_POUNDelse] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5160)] = {
    [sym_block_comment] = STATE(5160),
    [sym_compiler_directive_decl] = STATE(5160),
    [sym_fsi_directive_decl] = STATE(5160),
    [sym_preproc_line] = STATE(5160),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_DOT_DOT] = ACTIONS(7137),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [anon_sym_DOT_DOT2] = ACTIONS(7139),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__dedent] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(5161)] = {
    [sym_block_comment] = STATE(5161),
    [sym_compiler_directive_decl] = STATE(5161),
    [sym_fsi_directive_decl] = STATE(5161),
    [sym_preproc_line] = STATE(5161),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__multi_dollar_interp_end] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5162)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(5162),
    [sym_compiler_directive_decl] = STATE(5162),
    [sym_fsi_directive_decl] = STATE(5162),
    [sym_preproc_line] = STATE(5162),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9366),
    [anon_sym_module] = ACTIONS(9352),
    [anon_sym_null] = ACTIONS(9352),
    [anon_sym_begin] = ACTIONS(9352),
    [anon_sym_open] = ACTIONS(9352),
    [anon_sym_type] = ACTIONS(9352),
    [anon_sym_LBRACK_LT] = ACTIONS(9354),
    [anon_sym_do] = ACTIONS(9352),
    [anon_sym_and] = ACTIONS(9352),
    [anon_sym_let] = ACTIONS(9352),
    [anon_sym_let_BANG] = ACTIONS(9354),
    [aux_sym_access_modifier_token1] = ACTIONS(9354),
    [anon_sym_LPAREN] = ACTIONS(9352),
    [anon_sym_PIPE] = ACTIONS(9354),
    [anon_sym_AMP] = ACTIONS(9352),
    [anon_sym_LBRACK] = ACTIONS(9352),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9354),
    [anon_sym_LBRACE] = ACTIONS(9352),
    [anon_sym_LT_AT] = ACTIONS(9352),
    [anon_sym_LT_AT_AT] = ACTIONS(9354),
    [anon_sym_CARET] = ACTIONS(9354),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9354),
    [anon_sym_new] = ACTIONS(9352),
    [anon_sym_with] = ACTIONS(9352),
    [anon_sym_return] = ACTIONS(9352),
    [anon_sym_return_BANG] = ACTIONS(9354),
    [anon_sym_yield] = ACTIONS(9352),
    [anon_sym_yield_BANG] = ACTIONS(9354),
    [anon_sym_lazy] = ACTIONS(9352),
    [anon_sym_assert] = ACTIONS(9352),
    [anon_sym_upcast] = ACTIONS(9352),
    [anon_sym_downcast] = ACTIONS(9352),
    [anon_sym_fixed] = ACTIONS(9352),
    [anon_sym_for] = ACTIONS(9352),
    [anon_sym_while] = ACTIONS(9352),
    [anon_sym_while_BANG] = ACTIONS(9354),
    [anon_sym_if] = ACTIONS(9352),
    [anon_sym_fun] = ACTIONS(9352),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(9352),
    [anon_sym_match] = ACTIONS(9352),
    [anon_sym_match_BANG] = ACTIONS(9354),
    [anon_sym_function] = ACTIONS(9352),
    [anon_sym_use] = ACTIONS(9352),
    [anon_sym_use_BANG] = ACTIONS(9354),
    [anon_sym_do_BANG] = ACTIONS(9354),
    [anon_sym_AMP_AMP] = ACTIONS(9354),
    [anon_sym_STAR] = ACTIONS(9354),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(9352),
    [anon_sym_static] = ACTIONS(9352),
    [anon_sym_member] = ACTIONS(9352),
    [anon_sym_exception] = ACTIONS(9352),
    [anon_sym_interface] = ACTIONS(9352),
    [anon_sym_abstract] = ACTIONS(9352),
    [anon_sym_val] = ACTIONS(9352),
    [anon_sym_override] = ACTIONS(9352),
    [anon_sym_extern] = ACTIONS(9352),
    [aux_sym_char_token1] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9352),
    [anon_sym_DQUOTE] = ACTIONS(9352),
    [anon_sym_AT_DQUOTE] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [sym_bool] = ACTIONS(9352),
    [sym_unit] = ACTIONS(9354),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9352),
    [sym_op_identifier] = ACTIONS(9354),
    [anon_sym_PLUS] = ACTIONS(9352),
    [anon_sym_DASH] = ACTIONS(9352),
    [anon_sym_PLUS_DOT] = ACTIONS(9354),
    [anon_sym_DASH_DOT] = ACTIONS(9354),
    [anon_sym_PERCENT] = ACTIONS(9352),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9354),
    [anon_sym_TILDE] = ACTIONS(9354),
    [aux_sym_prefix_op_token1] = ACTIONS(9354),
    [sym_int] = ACTIONS(9352),
    [sym_xint] = ACTIONS(9354),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9354),
    [anon_sym_POUNDload] = ACTIONS(9354),
    [anon_sym_POUNDtime] = ACTIONS(9354),
    [anon_sym_POUNDI] = ACTIONS(9354),
    [anon_sym_POUNDhelp] = ACTIONS(9354),
    [anon_sym_POUNDquit] = ACTIONS(9354),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9354),
    [anon_sym_POUNDendif] = ACTIONS(9354),
    [anon_sym_POUNDelse] = ACTIONS(9354),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9354),
  },
  [STATE(5163)] = {
    [sym_block_comment] = STATE(5163),
    [sym_compiler_directive_decl] = STATE(5163),
    [sym_fsi_directive_decl] = STATE(5163),
    [sym_preproc_line] = STATE(5163),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_as] = ACTIONS(6871),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5164)] = {
    [sym_block_comment] = STATE(5164),
    [sym_compiler_directive_decl] = STATE(5164),
    [sym_fsi_directive_decl] = STATE(5164),
    [sym_preproc_line] = STATE(5164),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [anon_sym_POUNDendif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(5165)] = {
    [sym_block_comment] = STATE(5165),
    [sym_compiler_directive_decl] = STATE(5165),
    [sym_fsi_directive_decl] = STATE(5165),
    [sym_preproc_line] = STATE(5165),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9545),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [anon_sym_POUNDendif] = ACTIONS(6786),
    [anon_sym_POUNDelse] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5166)] = {
    [sym_block_comment] = STATE(5166),
    [sym_compiler_directive_decl] = STATE(5166),
    [sym_fsi_directive_decl] = STATE(5166),
    [sym_preproc_line] = STATE(5166),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_DASH_GT] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5167)] = {
    [sym_block_comment] = STATE(5167),
    [sym_compiler_directive_decl] = STATE(5167),
    [sym_fsi_directive_decl] = STATE(5167),
    [sym_preproc_line] = STATE(5167),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_DOT_DOT] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__dedent] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5168)] = {
    [sym_block_comment] = STATE(5168),
    [sym_compiler_directive_decl] = STATE(5168),
    [sym_fsi_directive_decl] = STATE(5168),
    [sym_preproc_line] = STATE(5168),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_DASH_GT] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_DOT_DOT] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5169)] = {
    [sym_block_comment] = STATE(5169),
    [sym_compiler_directive_decl] = STATE(5169),
    [sym_fsi_directive_decl] = STATE(5169),
    [sym_preproc_line] = STATE(5169),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_DASH_GT] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5170)] = {
    [sym_block_comment] = STATE(5170),
    [sym_compiler_directive_decl] = STATE(5170),
    [sym_fsi_directive_decl] = STATE(5170),
    [sym_preproc_line] = STATE(5170),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_DOT_DOT] = ACTIONS(6979),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [anon_sym_DOT_DOT2] = ACTIONS(6977),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__dedent] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(5171)] = {
    [sym_block_comment] = STATE(5171),
    [sym_compiler_directive_decl] = STATE(5171),
    [sym_fsi_directive_decl] = STATE(5171),
    [sym_preproc_line] = STATE(5171),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_as] = ACTIONS(6875),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5172)] = {
    [sym_block_comment] = STATE(5172),
    [sym_compiler_directive_decl] = STATE(5172),
    [sym_fsi_directive_decl] = STATE(5172),
    [sym_preproc_line] = STATE(5172),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_GT] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(5173)] = {
    [sym_type_arguments] = STATE(6479),
    [sym__object_members] = STATE(7004),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5173),
    [sym_compiler_directive_decl] = STATE(5173),
    [sym_fsi_directive_decl] = STATE(5173),
    [sym_preproc_line] = STATE(5173),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(9224),
    [anon_sym_null] = ACTIONS(9224),
    [anon_sym_begin] = ACTIONS(9224),
    [anon_sym_open] = ACTIONS(9224),
    [anon_sym_type] = ACTIONS(9224),
    [anon_sym_LBRACK_LT] = ACTIONS(9226),
    [anon_sym_do] = ACTIONS(9224),
    [anon_sym_and] = ACTIONS(9224),
    [anon_sym_let] = ACTIONS(9224),
    [anon_sym_let_BANG] = ACTIONS(9226),
    [aux_sym_access_modifier_token1] = ACTIONS(9226),
    [anon_sym_LPAREN] = ACTIONS(9224),
    [anon_sym_AMP] = ACTIONS(9224),
    [anon_sym_LBRACK] = ACTIONS(9224),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9226),
    [anon_sym_LBRACE] = ACTIONS(9224),
    [anon_sym_LT_AT] = ACTIONS(9224),
    [anon_sym_LT_AT_AT] = ACTIONS(9226),
    [anon_sym_CARET] = ACTIONS(9226),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9226),
    [anon_sym_new] = ACTIONS(9224),
    [anon_sym_with] = ACTIONS(9547),
    [anon_sym_return] = ACTIONS(9224),
    [anon_sym_return_BANG] = ACTIONS(9226),
    [anon_sym_yield] = ACTIONS(9224),
    [anon_sym_yield_BANG] = ACTIONS(9226),
    [anon_sym_lazy] = ACTIONS(9224),
    [anon_sym_assert] = ACTIONS(9224),
    [anon_sym_upcast] = ACTIONS(9224),
    [anon_sym_downcast] = ACTIONS(9224),
    [anon_sym_fixed] = ACTIONS(9224),
    [anon_sym_for] = ACTIONS(9224),
    [anon_sym_while] = ACTIONS(9224),
    [anon_sym_while_BANG] = ACTIONS(9226),
    [anon_sym_if] = ACTIONS(9224),
    [anon_sym_fun] = ACTIONS(9224),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(9224),
    [anon_sym_match] = ACTIONS(9224),
    [anon_sym_match_BANG] = ACTIONS(9226),
    [anon_sym_function] = ACTIONS(9224),
    [anon_sym_use] = ACTIONS(9224),
    [anon_sym_use_BANG] = ACTIONS(9226),
    [anon_sym_do_BANG] = ACTIONS(9226),
    [anon_sym_AMP_AMP] = ACTIONS(9226),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(9224),
    [anon_sym_static] = ACTIONS(9224),
    [anon_sym_member] = ACTIONS(9224),
    [anon_sym_exception] = ACTIONS(9224),
    [anon_sym_interface] = ACTIONS(9224),
    [anon_sym_abstract] = ACTIONS(9224),
    [anon_sym_val] = ACTIONS(9224),
    [anon_sym_override] = ACTIONS(9224),
    [anon_sym_extern] = ACTIONS(9224),
    [aux_sym_char_token1] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9224),
    [anon_sym_DQUOTE] = ACTIONS(9224),
    [anon_sym_AT_DQUOTE] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [sym_bool] = ACTIONS(9224),
    [sym_unit] = ACTIONS(9226),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9224),
    [sym_op_identifier] = ACTIONS(9226),
    [anon_sym_PLUS] = ACTIONS(9224),
    [anon_sym_DASH] = ACTIONS(9224),
    [anon_sym_PLUS_DOT] = ACTIONS(9226),
    [anon_sym_DASH_DOT] = ACTIONS(9226),
    [anon_sym_PERCENT] = ACTIONS(9224),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9226),
    [anon_sym_TILDE] = ACTIONS(9226),
    [aux_sym_prefix_op_token1] = ACTIONS(9226),
    [sym_int] = ACTIONS(9224),
    [sym_xint] = ACTIONS(9226),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9226),
    [anon_sym_POUNDload] = ACTIONS(9226),
    [anon_sym_POUNDtime] = ACTIONS(9226),
    [anon_sym_POUNDI] = ACTIONS(9226),
    [anon_sym_POUNDhelp] = ACTIONS(9226),
    [anon_sym_POUNDquit] = ACTIONS(9226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9226),
    [anon_sym_POUNDendif] = ACTIONS(9226),
    [sym__newline] = ACTIONS(9226),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9226),
  },
  [STATE(5174)] = {
    [sym_block_comment] = STATE(5174),
    [sym_compiler_directive_decl] = STATE(5174),
    [sym_fsi_directive_decl] = STATE(5174),
    [sym_preproc_line] = STATE(5174),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_DOT_DOT] = ACTIONS(6983),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [anon_sym_DOT_DOT2] = ACTIONS(6981),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__dedent] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(5175)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5175),
    [sym_compiler_directive_decl] = STATE(5175),
    [sym_fsi_directive_decl] = STATE(5175),
    [sym_preproc_line] = STATE(5175),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(9184),
    [sym_identifier] = ACTIONS(9182),
    [anon_sym_namespace] = ACTIONS(9182),
    [anon_sym_module] = ACTIONS(9182),
    [anon_sym_null] = ACTIONS(9182),
    [anon_sym_begin] = ACTIONS(9182),
    [anon_sym_open] = ACTIONS(9182),
    [anon_sym_type] = ACTIONS(9182),
    [anon_sym_LBRACK_LT] = ACTIONS(9184),
    [anon_sym_do] = ACTIONS(9182),
    [anon_sym_and] = ACTIONS(9182),
    [anon_sym_let] = ACTIONS(9182),
    [anon_sym_let_BANG] = ACTIONS(9184),
    [aux_sym_access_modifier_token1] = ACTIONS(9184),
    [anon_sym_LPAREN] = ACTIONS(9182),
    [anon_sym_PIPE] = ACTIONS(9184),
    [anon_sym_AMP] = ACTIONS(9182),
    [anon_sym_LBRACK] = ACTIONS(9182),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9184),
    [anon_sym_LBRACE] = ACTIONS(9182),
    [anon_sym_LT_AT] = ACTIONS(9182),
    [anon_sym_LT_AT_AT] = ACTIONS(9184),
    [anon_sym_CARET] = ACTIONS(9184),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9184),
    [anon_sym_new] = ACTIONS(9182),
    [anon_sym_with] = ACTIONS(9182),
    [anon_sym_return] = ACTIONS(9182),
    [anon_sym_return_BANG] = ACTIONS(9184),
    [anon_sym_yield] = ACTIONS(9182),
    [anon_sym_yield_BANG] = ACTIONS(9184),
    [anon_sym_lazy] = ACTIONS(9182),
    [anon_sym_assert] = ACTIONS(9182),
    [anon_sym_upcast] = ACTIONS(9182),
    [anon_sym_downcast] = ACTIONS(9182),
    [anon_sym_fixed] = ACTIONS(9182),
    [anon_sym_for] = ACTIONS(9182),
    [anon_sym_while] = ACTIONS(9182),
    [anon_sym_while_BANG] = ACTIONS(9184),
    [anon_sym_if] = ACTIONS(9182),
    [anon_sym_fun] = ACTIONS(9182),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(9182),
    [anon_sym_match] = ACTIONS(9182),
    [anon_sym_match_BANG] = ACTIONS(9184),
    [anon_sym_function] = ACTIONS(9182),
    [anon_sym_use] = ACTIONS(9182),
    [anon_sym_use_BANG] = ACTIONS(9184),
    [anon_sym_do_BANG] = ACTIONS(9184),
    [anon_sym_AMP_AMP] = ACTIONS(9184),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(9182),
    [anon_sym_static] = ACTIONS(9182),
    [anon_sym_member] = ACTIONS(9182),
    [anon_sym_exception] = ACTIONS(9182),
    [anon_sym_interface] = ACTIONS(9182),
    [anon_sym_abstract] = ACTIONS(9182),
    [anon_sym_val] = ACTIONS(9182),
    [anon_sym_override] = ACTIONS(9182),
    [anon_sym_extern] = ACTIONS(9182),
    [aux_sym_char_token1] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9182),
    [anon_sym_DQUOTE] = ACTIONS(9182),
    [anon_sym_AT_DQUOTE] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [sym_bool] = ACTIONS(9182),
    [sym_unit] = ACTIONS(9184),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9182),
    [sym_op_identifier] = ACTIONS(9184),
    [anon_sym_PLUS] = ACTIONS(9182),
    [anon_sym_DASH] = ACTIONS(9182),
    [anon_sym_PLUS_DOT] = ACTIONS(9184),
    [anon_sym_DASH_DOT] = ACTIONS(9184),
    [anon_sym_PERCENT] = ACTIONS(9182),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9184),
    [anon_sym_TILDE] = ACTIONS(9184),
    [aux_sym_prefix_op_token1] = ACTIONS(9184),
    [sym_int] = ACTIONS(9182),
    [sym_xint] = ACTIONS(9184),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9184),
    [anon_sym_POUNDload] = ACTIONS(9184),
    [anon_sym_POUNDtime] = ACTIONS(9184),
    [anon_sym_POUNDI] = ACTIONS(9184),
    [anon_sym_POUNDhelp] = ACTIONS(9184),
    [anon_sym_POUNDquit] = ACTIONS(9184),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9184),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9184),
  },
  [STATE(5176)] = {
    [sym_block_comment] = STATE(5176),
    [sym_compiler_directive_decl] = STATE(5176),
    [sym_fsi_directive_decl] = STATE(5176),
    [sym_preproc_line] = STATE(5176),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [anon_sym_POUNDendif] = ACTIONS(6792),
    [anon_sym_POUNDelse] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5177)] = {
    [sym_block_comment] = STATE(5177),
    [sym_compiler_directive_decl] = STATE(5177),
    [sym_fsi_directive_decl] = STATE(5177),
    [sym_preproc_line] = STATE(5177),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_DASH_GT] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_DOT_DOT] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5178)] = {
    [sym_block_comment] = STATE(5178),
    [sym_compiler_directive_decl] = STATE(5178),
    [sym_fsi_directive_decl] = STATE(5178),
    [sym_preproc_line] = STATE(5178),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_as] = ACTIONS(6828),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5179)] = {
    [sym_block_comment] = STATE(5179),
    [sym_compiler_directive_decl] = STATE(5179),
    [sym_fsi_directive_decl] = STATE(5179),
    [sym_preproc_line] = STATE(5179),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [anon_sym_POUNDendif] = ACTIONS(6796),
    [anon_sym_POUNDelse] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5180)] = {
    [sym_block_comment] = STATE(5180),
    [sym_compiler_directive_decl] = STATE(5180),
    [sym_fsi_directive_decl] = STATE(5180),
    [sym_preproc_line] = STATE(5180),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_DASH_GT] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_DOT_DOT] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5181)] = {
    [sym_block_comment] = STATE(5181),
    [sym_compiler_directive_decl] = STATE(5181),
    [sym_fsi_directive_decl] = STATE(5181),
    [sym_preproc_line] = STATE(5181),
    [sym_identifier] = ACTIONS(6554),
    [anon_sym_null] = ACTIONS(6554),
    [anon_sym_EQ] = ACTIONS(6554),
    [anon_sym_begin] = ACTIONS(6554),
    [anon_sym_COLON] = ACTIONS(6554),
    [anon_sym_do] = ACTIONS(6554),
    [anon_sym_let] = ACTIONS(6554),
    [anon_sym_let_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN] = ACTIONS(6554),
    [anon_sym_COMMA] = ACTIONS(6556),
    [anon_sym_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_QMARK] = ACTIONS(6554),
    [anon_sym_COLON_COLON] = ACTIONS(6556),
    [anon_sym_AMP] = ACTIONS(6554),
    [anon_sym_LBRACK] = ACTIONS(6554),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6556),
    [anon_sym_LBRACE] = ACTIONS(6554),
    [anon_sym_LT_AT] = ACTIONS(6554),
    [anon_sym_LT_AT_AT] = ACTIONS(6554),
    [anon_sym_CARET] = ACTIONS(6554),
    [anon_sym_or] = ACTIONS(6554),
    [anon_sym_DOT] = ACTIONS(6554),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6556),
    [anon_sym_new] = ACTIONS(6554),
    [anon_sym_return] = ACTIONS(6554),
    [anon_sym_return_BANG] = ACTIONS(6556),
    [anon_sym_yield] = ACTIONS(6554),
    [anon_sym_yield_BANG] = ACTIONS(6556),
    [anon_sym_lazy] = ACTIONS(6554),
    [anon_sym_assert] = ACTIONS(6554),
    [anon_sym_upcast] = ACTIONS(6554),
    [anon_sym_downcast] = ACTIONS(6554),
    [anon_sym_fixed] = ACTIONS(6554),
    [anon_sym_COLON_GT] = ACTIONS(6556),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6556),
    [anon_sym_for] = ACTIONS(6554),
    [anon_sym_while] = ACTIONS(6554),
    [anon_sym_while_BANG] = ACTIONS(6556),
    [anon_sym_else] = ACTIONS(6554),
    [anon_sym_elif] = ACTIONS(6554),
    [anon_sym_if] = ACTIONS(6554),
    [anon_sym_fun] = ACTIONS(6554),
    [anon_sym_try] = ACTIONS(6554),
    [anon_sym_match] = ACTIONS(6554),
    [anon_sym_match_BANG] = ACTIONS(6556),
    [anon_sym_function] = ACTIONS(6554),
    [anon_sym_DOT_LBRACK] = ACTIONS(6556),
    [anon_sym_use] = ACTIONS(6554),
    [anon_sym_use_BANG] = ACTIONS(6556),
    [anon_sym_do_BANG] = ACTIONS(6556),
    [anon_sym_LPAREN2] = ACTIONS(6554),
    [anon_sym_PIPE_PIPE] = ACTIONS(6554),
    [anon_sym_AMP_AMP] = ACTIONS(6554),
    [aux_sym_char_token1] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6554),
    [anon_sym_DQUOTE] = ACTIONS(6554),
    [anon_sym_AT_DQUOTE] = ACTIONS(6556),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6556),
    [sym_bool] = ACTIONS(6554),
    [sym_unit] = ACTIONS(6556),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6554),
    [sym_op_identifier] = ACTIONS(6554),
    [anon_sym_PLUS] = ACTIONS(6554),
    [anon_sym_DASH] = ACTIONS(6554),
    [anon_sym_PLUS_DOT] = ACTIONS(6554),
    [anon_sym_DASH_DOT] = ACTIONS(6554),
    [anon_sym_PERCENT] = ACTIONS(6554),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6554),
    [anon_sym_TILDE] = ACTIONS(6556),
    [aux_sym_prefix_op_token1] = ACTIONS(6554),
    [aux_sym_infix_op_token1] = ACTIONS(6556),
    [aux_sym_infix_op_token2] = ACTIONS(6554),
    [aux_sym_infix_op_token3] = ACTIONS(6554),
    [anon_sym_BANG_EQ] = ACTIONS(6554),
    [anon_sym_COLON_EQ] = ACTIONS(6556),
    [anon_sym_DOLLAR] = ACTIONS(6554),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6554),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6554),
    [sym_int] = ACTIONS(6554),
    [sym_xint] = ACTIONS(6556),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6556),
    [anon_sym_POUNDendif] = ACTIONS(6556),
    [sym__newline] = ACTIONS(6556),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6556),
    [sym__tyapp_open] = ACTIONS(6556),
  },
  [STATE(5182)] = {
    [sym_block_comment] = STATE(5182),
    [sym_compiler_directive_decl] = STATE(5182),
    [sym_fsi_directive_decl] = STATE(5182),
    [sym_preproc_line] = STATE(5182),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__dedent] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5183)] = {
    [sym_block_comment] = STATE(5183),
    [sym_compiler_directive_decl] = STATE(5183),
    [sym_fsi_directive_decl] = STATE(5183),
    [sym_preproc_line] = STATE(5183),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5184)] = {
    [sym_block_comment] = STATE(5184),
    [sym_compiler_directive_decl] = STATE(5184),
    [sym_fsi_directive_decl] = STATE(5184),
    [sym_preproc_line] = STATE(5184),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_DOT_DOT] = ACTIONS(6987),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [anon_sym_DOT_DOT2] = ACTIONS(6985),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__dedent] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(5185)] = {
    [sym_block_comment] = STATE(5185),
    [sym_compiler_directive_decl] = STATE(5185),
    [sym_fsi_directive_decl] = STATE(5185),
    [sym_preproc_line] = STATE(5185),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [anon_sym_POUNDelse] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5186)] = {
    [sym_block_comment] = STATE(5186),
    [sym_compiler_directive_decl] = STATE(5186),
    [sym_fsi_directive_decl] = STATE(5186),
    [sym_preproc_line] = STATE(5186),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9551),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_DOT_DOT] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__dedent] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5187)] = {
    [sym_block_comment] = STATE(5187),
    [sym_compiler_directive_decl] = STATE(5187),
    [sym_fsi_directive_decl] = STATE(5187),
    [sym_preproc_line] = STATE(5187),
    [aux_sym_rules_repeat1] = STATE(5199),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9516),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9553),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5188)] = {
    [sym_block_comment] = STATE(5188),
    [sym_compiler_directive_decl] = STATE(5188),
    [sym_fsi_directive_decl] = STATE(5188),
    [sym_preproc_line] = STATE(5188),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_as] = ACTIONS(6612),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5189)] = {
    [sym_block_comment] = STATE(5189),
    [sym_compiler_directive_decl] = STATE(5189),
    [sym_fsi_directive_decl] = STATE(5189),
    [sym_preproc_line] = STATE(5189),
    [aux_sym_rules_repeat1] = STATE(5200),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9516),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_DOT_DOT] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9553),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5190)] = {
    [sym_block_comment] = STATE(5190),
    [sym_compiler_directive_decl] = STATE(5190),
    [sym_fsi_directive_decl] = STATE(5190),
    [sym_preproc_line] = STATE(5190),
    [sym_identifier] = ACTIONS(6564),
    [anon_sym_null] = ACTIONS(6564),
    [anon_sym_EQ] = ACTIONS(6564),
    [anon_sym_begin] = ACTIONS(6564),
    [anon_sym_COLON] = ACTIONS(6564),
    [anon_sym_do] = ACTIONS(6564),
    [anon_sym_let] = ACTIONS(6564),
    [anon_sym_let_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN] = ACTIONS(6564),
    [anon_sym_COMMA] = ACTIONS(6562),
    [anon_sym_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_QMARK] = ACTIONS(6564),
    [anon_sym_COLON_COLON] = ACTIONS(6562),
    [anon_sym_AMP] = ACTIONS(6564),
    [anon_sym_LBRACK] = ACTIONS(6564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6562),
    [anon_sym_LBRACE] = ACTIONS(6564),
    [anon_sym_LT_AT] = ACTIONS(6564),
    [anon_sym_LT_AT_AT] = ACTIONS(6564),
    [anon_sym_CARET] = ACTIONS(6564),
    [anon_sym_or] = ACTIONS(6564),
    [anon_sym_DOT] = ACTIONS(6564),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6562),
    [anon_sym_new] = ACTIONS(6564),
    [anon_sym_with] = ACTIONS(6564),
    [anon_sym_return] = ACTIONS(6564),
    [anon_sym_return_BANG] = ACTIONS(6562),
    [anon_sym_yield] = ACTIONS(6564),
    [anon_sym_yield_BANG] = ACTIONS(6562),
    [anon_sym_lazy] = ACTIONS(6564),
    [anon_sym_assert] = ACTIONS(6564),
    [anon_sym_upcast] = ACTIONS(6564),
    [anon_sym_downcast] = ACTIONS(6564),
    [anon_sym_fixed] = ACTIONS(6564),
    [anon_sym_COLON_GT] = ACTIONS(6562),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6562),
    [anon_sym_for] = ACTIONS(6564),
    [anon_sym_while] = ACTIONS(6564),
    [anon_sym_while_BANG] = ACTIONS(6562),
    [anon_sym_else] = ACTIONS(6564),
    [anon_sym_elif] = ACTIONS(6564),
    [anon_sym_if] = ACTIONS(6564),
    [anon_sym_fun] = ACTIONS(6564),
    [anon_sym_try] = ACTIONS(6564),
    [anon_sym_match] = ACTIONS(6564),
    [anon_sym_match_BANG] = ACTIONS(6562),
    [anon_sym_function] = ACTIONS(6564),
    [anon_sym_DOT_LBRACK] = ACTIONS(6562),
    [anon_sym_use] = ACTIONS(6564),
    [anon_sym_use_BANG] = ACTIONS(6562),
    [anon_sym_do_BANG] = ACTIONS(6562),
    [anon_sym_LPAREN2] = ACTIONS(6564),
    [anon_sym_PIPE_PIPE] = ACTIONS(6564),
    [anon_sym_AMP_AMP] = ACTIONS(6564),
    [aux_sym_char_token1] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6564),
    [anon_sym_DQUOTE] = ACTIONS(6564),
    [anon_sym_AT_DQUOTE] = ACTIONS(6562),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6562),
    [sym_bool] = ACTIONS(6564),
    [sym_unit] = ACTIONS(6562),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6564),
    [sym_op_identifier] = ACTIONS(6564),
    [anon_sym_PLUS] = ACTIONS(6564),
    [anon_sym_DASH] = ACTIONS(6564),
    [anon_sym_PLUS_DOT] = ACTIONS(6564),
    [anon_sym_DASH_DOT] = ACTIONS(6564),
    [anon_sym_PERCENT] = ACTIONS(6564),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6564),
    [anon_sym_TILDE] = ACTIONS(6562),
    [aux_sym_prefix_op_token1] = ACTIONS(6564),
    [aux_sym_infix_op_token1] = ACTIONS(6562),
    [aux_sym_infix_op_token2] = ACTIONS(6564),
    [aux_sym_infix_op_token3] = ACTIONS(6564),
    [anon_sym_BANG_EQ] = ACTIONS(6564),
    [anon_sym_COLON_EQ] = ACTIONS(6562),
    [anon_sym_DOLLAR] = ACTIONS(6564),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6564),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6564),
    [sym_int] = ACTIONS(6564),
    [sym_xint] = ACTIONS(6562),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6562),
    [sym__newline] = ACTIONS(6562),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6562),
    [sym__tyapp_open] = ACTIONS(6562),
  },
  [STATE(5191)] = {
    [sym_block_comment] = STATE(5191),
    [sym_compiler_directive_decl] = STATE(5191),
    [sym_fsi_directive_decl] = STATE(5191),
    [sym_preproc_line] = STATE(5191),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_DOT_DOT] = ACTIONS(6991),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [anon_sym_DOT_DOT2] = ACTIONS(6989),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__dedent] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(5192)] = {
    [sym_block_comment] = STATE(5192),
    [sym_compiler_directive_decl] = STATE(5192),
    [sym_fsi_directive_decl] = STATE(5192),
    [sym_preproc_line] = STATE(5192),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_as] = ACTIONS(6840),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5193)] = {
    [sym_block_comment] = STATE(5193),
    [sym_compiler_directive_decl] = STATE(5193),
    [sym_fsi_directive_decl] = STATE(5193),
    [sym_preproc_line] = STATE(5193),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_DOT_DOT] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__dedent] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5194)] = {
    [sym_block_comment] = STATE(5194),
    [sym_compiler_directive_decl] = STATE(5194),
    [sym_fsi_directive_decl] = STATE(5194),
    [sym_preproc_line] = STATE(5194),
    [aux_sym_long_identifier_repeat1] = STATE(5213),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_GT] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9556),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(5195)] = {
    [sym_block_comment] = STATE(5195),
    [sym_compiler_directive_decl] = STATE(5195),
    [sym_fsi_directive_decl] = STATE(5195),
    [sym_preproc_line] = STATE(5195),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_DOT_DOT] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__dedent] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5196)] = {
    [sym_block_comment] = STATE(5196),
    [sym_compiler_directive_decl] = STATE(5196),
    [sym_fsi_directive_decl] = STATE(5196),
    [sym_preproc_line] = STATE(5196),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_as] = ACTIONS(6653),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5197)] = {
    [sym_block_comment] = STATE(5197),
    [sym_compiler_directive_decl] = STATE(5197),
    [sym_fsi_directive_decl] = STATE(5197),
    [sym_preproc_line] = STATE(5197),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_DOT_DOT] = ACTIONS(7155),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [anon_sym_DOT_DOT2] = ACTIONS(7157),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__dedent] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(5198)] = {
    [sym_block_comment] = STATE(5198),
    [sym_compiler_directive_decl] = STATE(5198),
    [sym_fsi_directive_decl] = STATE(5198),
    [sym_preproc_line] = STATE(5198),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_DOT_DOT] = ACTIONS(6995),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [anon_sym_DOT_DOT2] = ACTIONS(6993),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__dedent] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(5199)] = {
    [sym_block_comment] = STATE(5199),
    [sym_compiler_directive_decl] = STATE(5199),
    [sym_fsi_directive_decl] = STATE(5199),
    [sym_preproc_line] = STATE(5199),
    [aux_sym_rules_repeat1] = STATE(5200),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9516),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_DOT_DOT] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(5200)] = {
    [sym_block_comment] = STATE(5200),
    [sym_compiler_directive_decl] = STATE(5200),
    [sym_fsi_directive_decl] = STATE(5200),
    [sym_preproc_line] = STATE(5200),
    [aux_sym_rules_repeat1] = STATE(5200),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9561),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9564),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5201)] = {
    [sym_block_comment] = STATE(5201),
    [sym_compiler_directive_decl] = STATE(5201),
    [sym_fsi_directive_decl] = STATE(5201),
    [sym_preproc_line] = STATE(5201),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
    [sym__in] = ACTIONS(6805),
  },
  [STATE(5202)] = {
    [sym_block_comment] = STATE(5202),
    [sym_compiler_directive_decl] = STATE(5202),
    [sym_fsi_directive_decl] = STATE(5202),
    [sym_preproc_line] = STATE(5202),
    [aux_sym_rules_repeat1] = STATE(5208),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9484),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9567),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__multi_dollar_interp_end] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(5203)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5203),
    [sym_compiler_directive_decl] = STATE(5203),
    [sym_fsi_directive_decl] = STATE(5203),
    [sym_preproc_line] = STATE(5203),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9182),
    [anon_sym_module] = ACTIONS(9182),
    [anon_sym_null] = ACTIONS(9182),
    [anon_sym_begin] = ACTIONS(9182),
    [anon_sym_open] = ACTIONS(9182),
    [anon_sym_type] = ACTIONS(9182),
    [anon_sym_LBRACK_LT] = ACTIONS(9184),
    [anon_sym_do] = ACTIONS(9182),
    [anon_sym_and] = ACTIONS(9182),
    [anon_sym_let] = ACTIONS(9182),
    [anon_sym_let_BANG] = ACTIONS(9184),
    [aux_sym_access_modifier_token1] = ACTIONS(9184),
    [anon_sym_LPAREN] = ACTIONS(9182),
    [anon_sym_PIPE] = ACTIONS(9184),
    [anon_sym_AMP] = ACTIONS(9182),
    [anon_sym_LBRACK] = ACTIONS(9182),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9184),
    [anon_sym_LBRACE] = ACTIONS(9182),
    [anon_sym_LT_AT] = ACTIONS(9182),
    [anon_sym_LT_AT_AT] = ACTIONS(9184),
    [anon_sym_CARET] = ACTIONS(9184),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9184),
    [anon_sym_new] = ACTIONS(9182),
    [anon_sym_with] = ACTIONS(9182),
    [anon_sym_return] = ACTIONS(9182),
    [anon_sym_return_BANG] = ACTIONS(9184),
    [anon_sym_yield] = ACTIONS(9182),
    [anon_sym_yield_BANG] = ACTIONS(9184),
    [anon_sym_lazy] = ACTIONS(9182),
    [anon_sym_assert] = ACTIONS(9182),
    [anon_sym_upcast] = ACTIONS(9182),
    [anon_sym_downcast] = ACTIONS(9182),
    [anon_sym_fixed] = ACTIONS(9182),
    [anon_sym_for] = ACTIONS(9182),
    [anon_sym_while] = ACTIONS(9182),
    [anon_sym_while_BANG] = ACTIONS(9184),
    [anon_sym_if] = ACTIONS(9182),
    [anon_sym_fun] = ACTIONS(9182),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(9182),
    [anon_sym_match] = ACTIONS(9182),
    [anon_sym_match_BANG] = ACTIONS(9184),
    [anon_sym_function] = ACTIONS(9182),
    [anon_sym_use] = ACTIONS(9182),
    [anon_sym_use_BANG] = ACTIONS(9184),
    [anon_sym_do_BANG] = ACTIONS(9184),
    [anon_sym_AMP_AMP] = ACTIONS(9184),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(9182),
    [anon_sym_static] = ACTIONS(9182),
    [anon_sym_member] = ACTIONS(9182),
    [anon_sym_exception] = ACTIONS(9182),
    [anon_sym_interface] = ACTIONS(9182),
    [anon_sym_abstract] = ACTIONS(9182),
    [anon_sym_val] = ACTIONS(9182),
    [anon_sym_override] = ACTIONS(9182),
    [anon_sym_extern] = ACTIONS(9182),
    [aux_sym_char_token1] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9182),
    [anon_sym_DQUOTE] = ACTIONS(9182),
    [anon_sym_AT_DQUOTE] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [sym_bool] = ACTIONS(9182),
    [sym_unit] = ACTIONS(9184),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9182),
    [sym_op_identifier] = ACTIONS(9184),
    [anon_sym_PLUS] = ACTIONS(9182),
    [anon_sym_DASH] = ACTIONS(9182),
    [anon_sym_PLUS_DOT] = ACTIONS(9184),
    [anon_sym_DASH_DOT] = ACTIONS(9184),
    [anon_sym_PERCENT] = ACTIONS(9182),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9184),
    [anon_sym_TILDE] = ACTIONS(9184),
    [aux_sym_prefix_op_token1] = ACTIONS(9184),
    [sym_int] = ACTIONS(9182),
    [sym_xint] = ACTIONS(9184),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9184),
    [anon_sym_POUNDload] = ACTIONS(9184),
    [anon_sym_POUNDtime] = ACTIONS(9184),
    [anon_sym_POUNDI] = ACTIONS(9184),
    [anon_sym_POUNDhelp] = ACTIONS(9184),
    [anon_sym_POUNDquit] = ACTIONS(9184),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9184),
    [anon_sym_POUNDendif] = ACTIONS(9184),
    [sym__newline] = ACTIONS(9184),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9184),
  },
  [STATE(5204)] = {
    [sym_block_comment] = STATE(5204),
    [sym_compiler_directive_decl] = STATE(5204),
    [sym_fsi_directive_decl] = STATE(5204),
    [sym_preproc_line] = STATE(5204),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_DOT_DOT] = ACTIONS(6921),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [anon_sym_DOT_DOT2] = ACTIONS(6919),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__dedent] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(5205)] = {
    [sym_block_comment] = STATE(5205),
    [sym_compiler_directive_decl] = STATE(5205),
    [sym_fsi_directive_decl] = STATE(5205),
    [sym_preproc_line] = STATE(5205),
    [aux_sym_long_identifier_repeat1] = STATE(5207),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_PIPE] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6358),
    [anon_sym_CARET] = ACTIONS(6358),
    [anon_sym_DOT] = ACTIONS(9344),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6358),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_AMP_AMP] = ACTIONS(6358),
    [anon_sym_STAR] = ACTIONS(6358),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6358),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6358),
    [anon_sym_DASH_DOT] = ACTIONS(6358),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6358),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6358),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
  },
  [STATE(5206)] = {
    [sym_block_comment] = STATE(5206),
    [sym_compiler_directive_decl] = STATE(5206),
    [sym_fsi_directive_decl] = STATE(5206),
    [sym_preproc_line] = STATE(5206),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9156),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5207)] = {
    [sym_block_comment] = STATE(5207),
    [sym_compiler_directive_decl] = STATE(5207),
    [sym_fsi_directive_decl] = STATE(5207),
    [sym_preproc_line] = STATE(5207),
    [aux_sym_long_identifier_repeat1] = STATE(5207),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_PIPE] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6297),
    [anon_sym_CARET] = ACTIONS(6297),
    [anon_sym_DOT] = ACTIONS(9570),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6297),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_AMP_AMP] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6297),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6297),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6297),
    [anon_sym_DASH_DOT] = ACTIONS(6297),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6297),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6297),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
  },
  [STATE(5208)] = {
    [sym_block_comment] = STATE(5208),
    [sym_compiler_directive_decl] = STATE(5208),
    [sym_fsi_directive_decl] = STATE(5208),
    [sym_preproc_line] = STATE(5208),
    [aux_sym_rules_repeat1] = STATE(5208),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9573),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9576),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__multi_dollar_interp_end] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5209)] = {
    [sym_block_comment] = STATE(5209),
    [sym_compiler_directive_decl] = STATE(5209),
    [sym_fsi_directive_decl] = STATE(5209),
    [sym_preproc_line] = STATE(5209),
    [aux_sym_rules_repeat1] = STATE(5255),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9579),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [anon_sym_POUNDendif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9581),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5210)] = {
    [sym_block_comment] = STATE(5210),
    [sym_compiler_directive_decl] = STATE(5210),
    [sym_fsi_directive_decl] = STATE(5210),
    [sym_preproc_line] = STATE(5210),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [anon_sym_POUNDendif] = ACTIONS(6838),
    [anon_sym_POUNDelse] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5211)] = {
    [sym_block_comment] = STATE(5211),
    [sym_compiler_directive_decl] = STATE(5211),
    [sym_fsi_directive_decl] = STATE(5211),
    [sym_preproc_line] = STATE(5211),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9269),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [anon_sym_POUNDendif] = ACTIONS(6599),
    [anon_sym_POUNDelse] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5212)] = {
    [sym_block_comment] = STATE(5212),
    [sym_compiler_directive_decl] = STATE(5212),
    [sym_fsi_directive_decl] = STATE(5212),
    [sym_preproc_line] = STATE(5212),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_as] = ACTIONS(6813),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_with] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5213)] = {
    [sym_block_comment] = STATE(5213),
    [sym_compiler_directive_decl] = STATE(5213),
    [sym_fsi_directive_decl] = STATE(5213),
    [sym_preproc_line] = STATE(5213),
    [aux_sym_long_identifier_repeat1] = STATE(5213),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_GT] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9584),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5214)] = {
    [sym_block_comment] = STATE(5214),
    [sym_compiler_directive_decl] = STATE(5214),
    [sym_fsi_directive_decl] = STATE(5214),
    [sym_preproc_line] = STATE(5214),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [anon_sym_POUNDelse] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5215)] = {
    [sym_block_comment] = STATE(5215),
    [sym_compiler_directive_decl] = STATE(5215),
    [sym_fsi_directive_decl] = STATE(5215),
    [sym_preproc_line] = STATE(5215),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9587),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_DOT_DOT] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__dedent] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5216)] = {
    [sym_block_comment] = STATE(5216),
    [sym_compiler_directive_decl] = STATE(5216),
    [sym_fsi_directive_decl] = STATE(5216),
    [sym_preproc_line] = STATE(5216),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_with] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5217)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5217),
    [sym_compiler_directive_decl] = STATE(5217),
    [sym_fsi_directive_decl] = STATE(5217),
    [sym_preproc_line] = STATE(5217),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9208),
    [anon_sym_module] = ACTIONS(9208),
    [anon_sym_null] = ACTIONS(9208),
    [anon_sym_begin] = ACTIONS(9208),
    [anon_sym_open] = ACTIONS(9208),
    [anon_sym_type] = ACTIONS(9208),
    [anon_sym_LBRACK_LT] = ACTIONS(9210),
    [anon_sym_do] = ACTIONS(9208),
    [anon_sym_and] = ACTIONS(9208),
    [anon_sym_let] = ACTIONS(9208),
    [anon_sym_let_BANG] = ACTIONS(9210),
    [aux_sym_access_modifier_token1] = ACTIONS(9210),
    [anon_sym_LPAREN] = ACTIONS(9208),
    [anon_sym_PIPE] = ACTIONS(9210),
    [anon_sym_AMP] = ACTIONS(9208),
    [anon_sym_LBRACK] = ACTIONS(9208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9210),
    [anon_sym_LBRACE] = ACTIONS(9208),
    [anon_sym_LT_AT] = ACTIONS(9208),
    [anon_sym_LT_AT_AT] = ACTIONS(9210),
    [anon_sym_CARET] = ACTIONS(9210),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9210),
    [anon_sym_new] = ACTIONS(9208),
    [anon_sym_with] = ACTIONS(9208),
    [anon_sym_return] = ACTIONS(9208),
    [anon_sym_return_BANG] = ACTIONS(9210),
    [anon_sym_yield] = ACTIONS(9208),
    [anon_sym_yield_BANG] = ACTIONS(9210),
    [anon_sym_lazy] = ACTIONS(9208),
    [anon_sym_assert] = ACTIONS(9208),
    [anon_sym_upcast] = ACTIONS(9208),
    [anon_sym_downcast] = ACTIONS(9208),
    [anon_sym_fixed] = ACTIONS(9208),
    [anon_sym_for] = ACTIONS(9208),
    [anon_sym_while] = ACTIONS(9208),
    [anon_sym_while_BANG] = ACTIONS(9210),
    [anon_sym_if] = ACTIONS(9208),
    [anon_sym_fun] = ACTIONS(9208),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(9208),
    [anon_sym_match] = ACTIONS(9208),
    [anon_sym_match_BANG] = ACTIONS(9210),
    [anon_sym_function] = ACTIONS(9208),
    [anon_sym_use] = ACTIONS(9208),
    [anon_sym_use_BANG] = ACTIONS(9210),
    [anon_sym_do_BANG] = ACTIONS(9210),
    [anon_sym_AMP_AMP] = ACTIONS(9210),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(9208),
    [anon_sym_static] = ACTIONS(9208),
    [anon_sym_member] = ACTIONS(9208),
    [anon_sym_exception] = ACTIONS(9208),
    [anon_sym_interface] = ACTIONS(9208),
    [anon_sym_abstract] = ACTIONS(9208),
    [anon_sym_val] = ACTIONS(9208),
    [anon_sym_override] = ACTIONS(9208),
    [anon_sym_extern] = ACTIONS(9208),
    [aux_sym_char_token1] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9208),
    [anon_sym_DQUOTE] = ACTIONS(9208),
    [anon_sym_AT_DQUOTE] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [sym_bool] = ACTIONS(9208),
    [sym_unit] = ACTIONS(9210),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9208),
    [sym_op_identifier] = ACTIONS(9210),
    [anon_sym_PLUS] = ACTIONS(9208),
    [anon_sym_DASH] = ACTIONS(9208),
    [anon_sym_PLUS_DOT] = ACTIONS(9210),
    [anon_sym_DASH_DOT] = ACTIONS(9210),
    [anon_sym_PERCENT] = ACTIONS(9208),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9210),
    [anon_sym_TILDE] = ACTIONS(9210),
    [aux_sym_prefix_op_token1] = ACTIONS(9210),
    [sym_int] = ACTIONS(9208),
    [sym_xint] = ACTIONS(9210),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9210),
    [anon_sym_POUNDload] = ACTIONS(9210),
    [anon_sym_POUNDtime] = ACTIONS(9210),
    [anon_sym_POUNDI] = ACTIONS(9210),
    [anon_sym_POUNDhelp] = ACTIONS(9210),
    [anon_sym_POUNDquit] = ACTIONS(9210),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9210),
    [anon_sym_POUNDendif] = ACTIONS(9210),
    [sym__newline] = ACTIONS(9210),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9210),
  },
  [STATE(5218)] = {
    [sym_block_comment] = STATE(5218),
    [sym_compiler_directive_decl] = STATE(5218),
    [sym_fsi_directive_decl] = STATE(5218),
    [sym_preproc_line] = STATE(5218),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_DOT_DOT] = ACTIONS(6999),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [anon_sym_DOT_DOT2] = ACTIONS(6997),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__dedent] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(5219)] = {
    [sym_block_comment] = STATE(5219),
    [sym_compiler_directive_decl] = STATE(5219),
    [sym_fsi_directive_decl] = STATE(5219),
    [sym_preproc_line] = STATE(5219),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_DOT_DOT] = ACTIONS(6925),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [anon_sym_DOT_DOT2] = ACTIONS(6923),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__dedent] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(5220)] = {
    [sym_block_comment] = STATE(5220),
    [sym_compiler_directive_decl] = STATE(5220),
    [sym_fsi_directive_decl] = STATE(5220),
    [sym_preproc_line] = STATE(5220),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_as] = ACTIONS(6824),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_with] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5221)] = {
    [sym_block_comment] = STATE(5221),
    [sym_compiler_directive_decl] = STATE(5221),
    [sym_fsi_directive_decl] = STATE(5221),
    [sym_preproc_line] = STATE(5221),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_DOT_DOT] = ACTIONS(7003),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [anon_sym_DOT_DOT2] = ACTIONS(7001),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__dedent] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(5222)] = {
    [sym_block_comment] = STATE(5222),
    [sym_compiler_directive_decl] = STATE(5222),
    [sym_fsi_directive_decl] = STATE(5222),
    [sym_preproc_line] = STATE(5222),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9589),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [anon_sym_POUNDendif] = ACTIONS(6766),
    [anon_sym_POUNDelse] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5223)] = {
    [sym_block_comment] = STATE(5223),
    [sym_compiler_directive_decl] = STATE(5223),
    [sym_fsi_directive_decl] = STATE(5223),
    [sym_preproc_line] = STATE(5223),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9506),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_DASH_GT] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5224)] = {
    [sym_block_comment] = STATE(5224),
    [sym_compiler_directive_decl] = STATE(5224),
    [sym_fsi_directive_decl] = STATE(5224),
    [sym_preproc_line] = STATE(5224),
    [aux_sym_type_argument_repeat1] = STATE(5224),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(9591),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5225)] = {
    [sym_block_comment] = STATE(5225),
    [sym_compiler_directive_decl] = STATE(5225),
    [sym_fsi_directive_decl] = STATE(5225),
    [sym_preproc_line] = STATE(5225),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_DOT_DOT] = ACTIONS(7007),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [anon_sym_DOT_DOT2] = ACTIONS(7005),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__dedent] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(5226)] = {
    [sym_block_comment] = STATE(5226),
    [sym_compiler_directive_decl] = STATE(5226),
    [sym_fsi_directive_decl] = STATE(5226),
    [sym_preproc_line] = STATE(5226),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_DOT_DOT] = ACTIONS(7159),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [anon_sym_DOT_DOT2] = ACTIONS(7161),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__dedent] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(5227)] = {
    [sym_block_comment] = STATE(5227),
    [sym_compiler_directive_decl] = STATE(5227),
    [sym_fsi_directive_decl] = STATE(5227),
    [sym_preproc_line] = STATE(5227),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_DASH_GT] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_DOT_DOT] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5228)] = {
    [sym_block_comment] = STATE(5228),
    [sym_compiler_directive_decl] = STATE(5228),
    [sym_fsi_directive_decl] = STATE(5228),
    [sym_preproc_line] = STATE(5228),
    [aux_sym_type_argument_repeat1] = STATE(5151),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_PIPE] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6333),
    [anon_sym_CARET] = ACTIONS(6333),
    [anon_sym_or] = ACTIONS(9534),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6333),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_AMP_AMP] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6333),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6333),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6333),
    [anon_sym_DASH_DOT] = ACTIONS(6333),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6333),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6333),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
  },
  [STATE(5229)] = {
    [sym_block_comment] = STATE(5229),
    [sym_compiler_directive_decl] = STATE(5229),
    [sym_fsi_directive_decl] = STATE(5229),
    [sym_preproc_line] = STATE(5229),
    [aux_sym_long_identifier_repeat1] = STATE(5229),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9594),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5230)] = {
    [sym_block_comment] = STATE(5230),
    [sym_compiler_directive_decl] = STATE(5230),
    [sym_fsi_directive_decl] = STATE(5230),
    [sym_preproc_line] = STATE(5230),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [anon_sym_POUNDelse] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5231)] = {
    [sym_block_comment] = STATE(5231),
    [sym_compiler_directive_decl] = STATE(5231),
    [sym_fsi_directive_decl] = STATE(5231),
    [sym_preproc_line] = STATE(5231),
    [aux_sym_rules_repeat1] = STATE(5269),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9597),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9599),
    [sym__untyped_quoted_close] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5232)] = {
    [sym_block_comment] = STATE(5232),
    [sym_compiler_directive_decl] = STATE(5232),
    [sym_fsi_directive_decl] = STATE(5232),
    [sym_preproc_line] = STATE(5232),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_as] = ACTIONS(6295),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5233)] = {
    [sym_block_comment] = STATE(5233),
    [sym_compiler_directive_decl] = STATE(5233),
    [sym_fsi_directive_decl] = STATE(5233),
    [sym_preproc_line] = STATE(5233),
    [aux_sym_rules_repeat1] = STATE(5049),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9597),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9602),
    [sym__untyped_quoted_close] = ACTIONS(6616),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(5234)] = {
    [sym_block_comment] = STATE(5234),
    [sym_compiler_directive_decl] = STATE(5234),
    [sym_fsi_directive_decl] = STATE(5234),
    [sym_preproc_line] = STATE(5234),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_DOT_DOT] = ACTIONS(7011),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [anon_sym_DOT_DOT2] = ACTIONS(7009),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__dedent] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(5235)] = {
    [sym_block_comment] = STATE(5235),
    [sym_compiler_directive_decl] = STATE(5235),
    [sym_fsi_directive_decl] = STATE(5235),
    [sym_preproc_line] = STATE(5235),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__untyped_quoted_close] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(5236)] = {
    [sym_block_comment] = STATE(5236),
    [sym_compiler_directive_decl] = STATE(5236),
    [sym_fsi_directive_decl] = STATE(5236),
    [sym_preproc_line] = STATE(5236),
    [aux_sym_rules_repeat1] = STATE(5071),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9484),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__multi_dollar_interp_end] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5237)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5237),
    [sym_compiler_directive_decl] = STATE(5237),
    [sym_fsi_directive_decl] = STATE(5237),
    [sym_preproc_line] = STATE(5237),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_PIPE] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6255),
    [anon_sym_CARET] = ACTIONS(6255),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_AMP_AMP] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6255),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6255),
    [anon_sym_DASH_DOT] = ACTIONS(6255),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6255),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6255),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__newline] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
  },
  [STATE(5238)] = {
    [sym_block_comment] = STATE(5238),
    [sym_compiler_directive_decl] = STATE(5238),
    [sym_fsi_directive_decl] = STATE(5238),
    [sym_preproc_line] = STATE(5238),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_DOT_DOT] = ACTIONS(6892),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [anon_sym_DOT_DOT2] = ACTIONS(6890),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__dedent] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(5239)] = {
    [sym_block_comment] = STATE(5239),
    [sym_compiler_directive_decl] = STATE(5239),
    [sym_fsi_directive_decl] = STATE(5239),
    [sym_preproc_line] = STATE(5239),
    [aux_sym_rules_repeat1] = STATE(5301),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9492),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_with] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9608),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(5240)] = {
    [sym_block_comment] = STATE(5240),
    [sym_compiler_directive_decl] = STATE(5240),
    [sym_fsi_directive_decl] = STATE(5240),
    [sym_preproc_line] = STATE(5240),
    [aux_sym_rules_repeat1] = STATE(5239),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9492),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9611),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5241)] = {
    [sym_block_comment] = STATE(5241),
    [sym_compiler_directive_decl] = STATE(5241),
    [sym_fsi_directive_decl] = STATE(5241),
    [sym_preproc_line] = STATE(5241),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(8014),
    [aux_sym_decimal_token1] = ACTIONS(7539),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5242)] = {
    [sym_block_comment] = STATE(5242),
    [sym_compiler_directive_decl] = STATE(5242),
    [sym_fsi_directive_decl] = STATE(5242),
    [sym_preproc_line] = STATE(5242),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9054),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__dedent] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5243)] = {
    [sym_block_comment] = STATE(5243),
    [sym_compiler_directive_decl] = STATE(5243),
    [sym_fsi_directive_decl] = STATE(5243),
    [sym_preproc_line] = STATE(5243),
    [aux_sym_long_identifier_repeat1] = STATE(5260),
    [sym_identifier] = ACTIONS(6625),
    [anon_sym_null] = ACTIONS(6625),
    [anon_sym_EQ] = ACTIONS(6625),
    [anon_sym_begin] = ACTIONS(6625),
    [anon_sym_COLON] = ACTIONS(6625),
    [anon_sym_do] = ACTIONS(6625),
    [anon_sym_let] = ACTIONS(6625),
    [anon_sym_let_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN] = ACTIONS(6625),
    [anon_sym_COMMA] = ACTIONS(6628),
    [anon_sym_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_QMARK] = ACTIONS(6625),
    [anon_sym_COLON_COLON] = ACTIONS(6628),
    [anon_sym_AMP] = ACTIONS(6625),
    [anon_sym_LBRACK] = ACTIONS(6625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6628),
    [anon_sym_LBRACE] = ACTIONS(6625),
    [anon_sym_LT_AT] = ACTIONS(6625),
    [anon_sym_LT_AT_AT] = ACTIONS(6625),
    [anon_sym_CARET] = ACTIONS(6625),
    [anon_sym_or] = ACTIONS(6625),
    [anon_sym_DOT] = ACTIONS(9614),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6628),
    [anon_sym_new] = ACTIONS(6625),
    [anon_sym_return] = ACTIONS(6625),
    [anon_sym_return_BANG] = ACTIONS(6628),
    [anon_sym_yield] = ACTIONS(6625),
    [anon_sym_yield_BANG] = ACTIONS(6628),
    [anon_sym_lazy] = ACTIONS(6625),
    [anon_sym_assert] = ACTIONS(6625),
    [anon_sym_upcast] = ACTIONS(6625),
    [anon_sym_downcast] = ACTIONS(6625),
    [anon_sym_fixed] = ACTIONS(6625),
    [anon_sym_COLON_GT] = ACTIONS(6628),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6628),
    [anon_sym_for] = ACTIONS(6625),
    [anon_sym_while] = ACTIONS(6625),
    [anon_sym_while_BANG] = ACTIONS(6628),
    [anon_sym_if] = ACTIONS(6625),
    [anon_sym_fun] = ACTIONS(6625),
    [anon_sym_try] = ACTIONS(6625),
    [anon_sym_match] = ACTIONS(6625),
    [anon_sym_match_BANG] = ACTIONS(6628),
    [anon_sym_function] = ACTIONS(6625),
    [anon_sym_LT_DASH] = ACTIONS(6625),
    [anon_sym_DOT_LBRACK] = ACTIONS(6628),
    [anon_sym_use] = ACTIONS(6625),
    [anon_sym_use_BANG] = ACTIONS(6628),
    [anon_sym_do_BANG] = ACTIONS(6628),
    [anon_sym_LPAREN2] = ACTIONS(6625),
    [anon_sym_PIPE_PIPE] = ACTIONS(6625),
    [anon_sym_AMP_AMP] = ACTIONS(6625),
    [aux_sym_char_token1] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6625),
    [anon_sym_DQUOTE] = ACTIONS(6625),
    [anon_sym_AT_DQUOTE] = ACTIONS(6628),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6628),
    [sym_bool] = ACTIONS(6625),
    [sym_unit] = ACTIONS(6628),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6625),
    [sym_op_identifier] = ACTIONS(6625),
    [anon_sym_PLUS] = ACTIONS(6625),
    [anon_sym_DASH] = ACTIONS(6625),
    [anon_sym_PLUS_DOT] = ACTIONS(6625),
    [anon_sym_DASH_DOT] = ACTIONS(6625),
    [anon_sym_PERCENT] = ACTIONS(6625),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6625),
    [anon_sym_TILDE] = ACTIONS(6628),
    [aux_sym_prefix_op_token1] = ACTIONS(6625),
    [aux_sym_infix_op_token1] = ACTIONS(6628),
    [aux_sym_infix_op_token2] = ACTIONS(6625),
    [aux_sym_infix_op_token3] = ACTIONS(6625),
    [anon_sym_BANG_EQ] = ACTIONS(6625),
    [anon_sym_COLON_EQ] = ACTIONS(6628),
    [anon_sym_DOLLAR] = ACTIONS(6625),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6625),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6625),
    [sym_int] = ACTIONS(6625),
    [sym_xint] = ACTIONS(6628),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6628),
    [sym__newline] = ACTIONS(6628),
    [sym__quoted_close] = ACTIONS(6628),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6628),
    [sym__tyapp_open] = ACTIONS(6628),
  },
  [STATE(5244)] = {
    [sym_block_comment] = STATE(5244),
    [sym_compiler_directive_decl] = STATE(5244),
    [sym_fsi_directive_decl] = STATE(5244),
    [sym_preproc_line] = STATE(5244),
    [aux_sym_rules_repeat1] = STATE(5301),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9492),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_with] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9611),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5245)] = {
    [sym_block_comment] = STATE(5245),
    [sym_compiler_directive_decl] = STATE(5245),
    [sym_fsi_directive_decl] = STATE(5245),
    [sym_preproc_line] = STATE(5245),
    [aux_sym_rules_repeat1] = STATE(4968),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9393),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9618),
    [sym__quoted_close] = ACTIONS(6588),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5246)] = {
    [sym_block_comment] = STATE(5246),
    [sym_compiler_directive_decl] = STATE(5246),
    [sym_fsi_directive_decl] = STATE(5246),
    [sym_preproc_line] = STATE(5246),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9621),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__multi_dollar_interp_end] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5247)] = {
    [sym_block_comment] = STATE(5247),
    [sym_compiler_directive_decl] = STATE(5247),
    [sym_fsi_directive_decl] = STATE(5247),
    [sym_preproc_line] = STATE(5247),
    [aux_sym_sequential_expression_repeat1] = STATE(5247),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_DOT_DOT] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9623),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5248)] = {
    [sym_block_comment] = STATE(5248),
    [sym_compiler_directive_decl] = STATE(5248),
    [sym_fsi_directive_decl] = STATE(5248),
    [sym_preproc_line] = STATE(5248),
    [aux_sym_long_identifier_repeat1] = STATE(5344),
    [ts_builtin_sym_end] = ACTIONS(6291),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_namespace] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9626),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9628),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5249)] = {
    [sym_block_comment] = STATE(5249),
    [sym_compiler_directive_decl] = STATE(5249),
    [sym_fsi_directive_decl] = STATE(5249),
    [sym_preproc_line] = STATE(5249),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__quoted_close] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5250)] = {
    [sym_block_comment] = STATE(5250),
    [sym_compiler_directive_decl] = STATE(5250),
    [sym_fsi_directive_decl] = STATE(5250),
    [sym_preproc_line] = STATE(5250),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__multi_dollar_interp_end] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5251)] = {
    [sym_block_comment] = STATE(5251),
    [sym_compiler_directive_decl] = STATE(5251),
    [sym_fsi_directive_decl] = STATE(5251),
    [sym_preproc_line] = STATE(5251),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_GT] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5252)] = {
    [sym_block_comment] = STATE(5252),
    [sym_compiler_directive_decl] = STATE(5252),
    [sym_fsi_directive_decl] = STATE(5252),
    [sym_preproc_line] = STATE(5252),
    [aux_sym_sequential_expression_repeat1] = STATE(5138),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
    [sym__in] = ACTIONS(6738),
  },
  [STATE(5253)] = {
    [sym_block_comment] = STATE(5253),
    [sym_compiler_directive_decl] = STATE(5253),
    [sym_fsi_directive_decl] = STATE(5253),
    [sym_preproc_line] = STATE(5253),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_DASH_GT] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5254)] = {
    [sym_block_comment] = STATE(5254),
    [sym_compiler_directive_decl] = STATE(5254),
    [sym_fsi_directive_decl] = STATE(5254),
    [sym_preproc_line] = STATE(5254),
    [aux_sym_rules_repeat1] = STATE(5276),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9579),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9630),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5255)] = {
    [sym_block_comment] = STATE(5255),
    [sym_compiler_directive_decl] = STATE(5255),
    [sym_fsi_directive_decl] = STATE(5255),
    [sym_preproc_line] = STATE(5255),
    [aux_sym_rules_repeat1] = STATE(5277),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9579),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [anon_sym_POUNDendif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9630),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5256)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5256),
    [sym_compiler_directive_decl] = STATE(5256),
    [sym_fsi_directive_decl] = STATE(5256),
    [sym_preproc_line] = STATE(5256),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_PIPE] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6209),
    [anon_sym_CARET] = ACTIONS(6209),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_AMP_AMP] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6209),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6209),
    [anon_sym_DASH_DOT] = ACTIONS(6209),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6209),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6209),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__newline] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
  },
  [STATE(5257)] = {
    [sym_block_comment] = STATE(5257),
    [sym_compiler_directive_decl] = STATE(5257),
    [sym_fsi_directive_decl] = STATE(5257),
    [sym_preproc_line] = STATE(5257),
    [aux_sym_sequential_expression_repeat1] = STATE(5257),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [anon_sym_POUNDelse] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9633),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5258)] = {
    [sym_block_comment] = STATE(5258),
    [sym_compiler_directive_decl] = STATE(5258),
    [sym_fsi_directive_decl] = STATE(5258),
    [sym_preproc_line] = STATE(5258),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_DOT_DOT] = ACTIONS(7019),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [anon_sym_DOT_DOT2] = ACTIONS(7017),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__dedent] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(5259)] = {
    [sym_block_comment] = STATE(5259),
    [sym_compiler_directive_decl] = STATE(5259),
    [sym_fsi_directive_decl] = STATE(5259),
    [sym_preproc_line] = STATE(5259),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [anon_sym_POUNDendif] = ACTIONS(6869),
    [anon_sym_POUNDelse] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5260)] = {
    [sym_block_comment] = STATE(5260),
    [sym_compiler_directive_decl] = STATE(5260),
    [sym_fsi_directive_decl] = STATE(5260),
    [sym_preproc_line] = STATE(5260),
    [aux_sym_long_identifier_repeat1] = STATE(5028),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9636),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__quoted_close] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(5261)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5261),
    [sym_compiler_directive_decl] = STATE(5261),
    [sym_fsi_directive_decl] = STATE(5261),
    [sym_preproc_line] = STATE(5261),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_PIPE] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6195),
    [anon_sym_CARET] = ACTIONS(6195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_AMP_AMP] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6195),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6195),
    [anon_sym_DASH_DOT] = ACTIONS(6195),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6195),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6195),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
  },
  [STATE(5262)] = {
    [sym_block_comment] = STATE(5262),
    [sym_compiler_directive_decl] = STATE(5262),
    [sym_fsi_directive_decl] = STATE(5262),
    [sym_preproc_line] = STATE(5262),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__dedent] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
    [sym__in] = ACTIONS(6869),
  },
  [STATE(5263)] = {
    [sym_block_comment] = STATE(5263),
    [sym_compiler_directive_decl] = STATE(5263),
    [sym_fsi_directive_decl] = STATE(5263),
    [sym_preproc_line] = STATE(5263),
    [aux_sym_long_identifier_repeat1] = STATE(5263),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(9638),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__untyped_quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5264)] = {
    [sym_block_comment] = STATE(5264),
    [sym_compiler_directive_decl] = STATE(5264),
    [sym_fsi_directive_decl] = STATE(5264),
    [sym_preproc_line] = STATE(5264),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_DOT_DOT] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__dedent] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5265)] = {
    [sym_block_comment] = STATE(5265),
    [sym_compiler_directive_decl] = STATE(5265),
    [sym_fsi_directive_decl] = STATE(5265),
    [sym_preproc_line] = STATE(5265),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__dedent] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5266)] = {
    [sym_block_comment] = STATE(5266),
    [sym_compiler_directive_decl] = STATE(5266),
    [sym_fsi_directive_decl] = STATE(5266),
    [sym_preproc_line] = STATE(5266),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5267)] = {
    [sym_block_comment] = STATE(5267),
    [sym_compiler_directive_decl] = STATE(5267),
    [sym_fsi_directive_decl] = STATE(5267),
    [sym_preproc_line] = STATE(5267),
    [aux_sym_rules_repeat1] = STATE(5233),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9597),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9641),
    [sym__untyped_quoted_close] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5268)] = {
    [sym_block_comment] = STATE(5268),
    [sym_compiler_directive_decl] = STATE(5268),
    [sym_fsi_directive_decl] = STATE(5268),
    [sym_preproc_line] = STATE(5268),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9506),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_DASH_GT] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5269)] = {
    [sym_block_comment] = STATE(5269),
    [sym_compiler_directive_decl] = STATE(5269),
    [sym_fsi_directive_decl] = STATE(5269),
    [sym_preproc_line] = STATE(5269),
    [aux_sym_rules_repeat1] = STATE(5049),
    [sym_identifier] = ACTIONS(6607),
    [anon_sym_null] = ACTIONS(6607),
    [anon_sym_EQ] = ACTIONS(6607),
    [anon_sym_begin] = ACTIONS(6607),
    [anon_sym_COLON] = ACTIONS(6607),
    [anon_sym_do] = ACTIONS(6607),
    [anon_sym_let] = ACTIONS(6607),
    [anon_sym_let_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN] = ACTIONS(6607),
    [anon_sym_COMMA] = ACTIONS(6605),
    [anon_sym_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_QMARK] = ACTIONS(6607),
    [anon_sym_COLON_COLON] = ACTIONS(6605),
    [anon_sym_PIPE] = ACTIONS(9597),
    [anon_sym_AMP] = ACTIONS(6607),
    [anon_sym_LBRACK] = ACTIONS(6607),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6605),
    [anon_sym_LBRACE] = ACTIONS(6607),
    [anon_sym_LT_AT] = ACTIONS(6607),
    [anon_sym_LT_AT_AT] = ACTIONS(6607),
    [anon_sym_CARET] = ACTIONS(6607),
    [anon_sym_or] = ACTIONS(6607),
    [anon_sym_DOT] = ACTIONS(6607),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6605),
    [anon_sym_new] = ACTIONS(6607),
    [anon_sym_return] = ACTIONS(6607),
    [anon_sym_return_BANG] = ACTIONS(6605),
    [anon_sym_yield] = ACTIONS(6607),
    [anon_sym_yield_BANG] = ACTIONS(6605),
    [anon_sym_lazy] = ACTIONS(6607),
    [anon_sym_assert] = ACTIONS(6607),
    [anon_sym_upcast] = ACTIONS(6607),
    [anon_sym_downcast] = ACTIONS(6607),
    [anon_sym_fixed] = ACTIONS(6607),
    [anon_sym_COLON_GT] = ACTIONS(6605),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6605),
    [anon_sym_for] = ACTIONS(6607),
    [anon_sym_while] = ACTIONS(6607),
    [anon_sym_while_BANG] = ACTIONS(6605),
    [anon_sym_if] = ACTIONS(6607),
    [anon_sym_fun] = ACTIONS(6607),
    [anon_sym_try] = ACTIONS(6607),
    [anon_sym_match] = ACTIONS(6607),
    [anon_sym_match_BANG] = ACTIONS(6605),
    [anon_sym_function] = ACTIONS(6607),
    [anon_sym_DOT_LBRACK] = ACTIONS(6605),
    [anon_sym_use] = ACTIONS(6607),
    [anon_sym_use_BANG] = ACTIONS(6605),
    [anon_sym_do_BANG] = ACTIONS(6605),
    [anon_sym_LPAREN2] = ACTIONS(6607),
    [anon_sym_PIPE_PIPE] = ACTIONS(6607),
    [anon_sym_AMP_AMP] = ACTIONS(6607),
    [aux_sym_char_token1] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6607),
    [anon_sym_DQUOTE] = ACTIONS(6607),
    [anon_sym_AT_DQUOTE] = ACTIONS(6605),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6605),
    [sym_bool] = ACTIONS(6607),
    [sym_unit] = ACTIONS(6605),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6607),
    [sym_op_identifier] = ACTIONS(6607),
    [anon_sym_PLUS] = ACTIONS(6607),
    [anon_sym_DASH] = ACTIONS(6607),
    [anon_sym_PLUS_DOT] = ACTIONS(6607),
    [anon_sym_DASH_DOT] = ACTIONS(6607),
    [anon_sym_PERCENT] = ACTIONS(6607),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6607),
    [anon_sym_TILDE] = ACTIONS(6605),
    [aux_sym_prefix_op_token1] = ACTIONS(6607),
    [aux_sym_infix_op_token1] = ACTIONS(6605),
    [aux_sym_infix_op_token2] = ACTIONS(6607),
    [aux_sym_infix_op_token3] = ACTIONS(6607),
    [anon_sym_BANG_EQ] = ACTIONS(6607),
    [anon_sym_COLON_EQ] = ACTIONS(6605),
    [anon_sym_DOLLAR] = ACTIONS(6607),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6607),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6607),
    [sym_int] = ACTIONS(6607),
    [sym_xint] = ACTIONS(6605),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6605),
    [sym__newline] = ACTIONS(9641),
    [sym__untyped_quoted_close] = ACTIONS(6605),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6605),
    [sym__tyapp_open] = ACTIONS(6605),
  },
  [STATE(5270)] = {
    [sym_block_comment] = STATE(5270),
    [sym_compiler_directive_decl] = STATE(5270),
    [sym_fsi_directive_decl] = STATE(5270),
    [sym_preproc_line] = STATE(5270),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_with] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(5271)] = {
    [sym_block_comment] = STATE(5271),
    [sym_compiler_directive_decl] = STATE(5271),
    [sym_fsi_directive_decl] = STATE(5271),
    [sym_preproc_line] = STATE(5271),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_DOT_DOT] = ACTIONS(7023),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [anon_sym_DOT_DOT2] = ACTIONS(7021),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__dedent] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(5272)] = {
    [sym_block_comment] = STATE(5272),
    [sym_compiler_directive_decl] = STATE(5272),
    [sym_fsi_directive_decl] = STATE(5272),
    [sym_preproc_line] = STATE(5272),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_DOT_DOT] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__dedent] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5273)] = {
    [sym_block_comment] = STATE(5273),
    [sym_compiler_directive_decl] = STATE(5273),
    [sym_fsi_directive_decl] = STATE(5273),
    [sym_preproc_line] = STATE(5273),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_DOT_DOT] = ACTIONS(7027),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [anon_sym_DOT_DOT2] = ACTIONS(7025),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__dedent] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(5274)] = {
    [sym_block_comment] = STATE(5274),
    [sym_compiler_directive_decl] = STATE(5274),
    [sym_fsi_directive_decl] = STATE(5274),
    [sym_preproc_line] = STATE(5274),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_DOT_DOT] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__dedent] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5275)] = {
    [sym_block_comment] = STATE(5275),
    [sym_compiler_directive_decl] = STATE(5275),
    [sym_fsi_directive_decl] = STATE(5275),
    [sym_preproc_line] = STATE(5275),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_DOT_DOT] = ACTIONS(7035),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [anon_sym_DOT_DOT2] = ACTIONS(7033),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__dedent] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(5276)] = {
    [sym_block_comment] = STATE(5276),
    [sym_compiler_directive_decl] = STATE(5276),
    [sym_fsi_directive_decl] = STATE(5276),
    [sym_preproc_line] = STATE(5276),
    [aux_sym_rules_repeat1] = STATE(5277),
    [sym_identifier] = ACTIONS(6618),
    [anon_sym_null] = ACTIONS(6618),
    [anon_sym_EQ] = ACTIONS(6618),
    [anon_sym_begin] = ACTIONS(6618),
    [anon_sym_COLON] = ACTIONS(6618),
    [anon_sym_do] = ACTIONS(6618),
    [anon_sym_let] = ACTIONS(6618),
    [anon_sym_let_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN] = ACTIONS(6618),
    [anon_sym_COMMA] = ACTIONS(6616),
    [anon_sym_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_QMARK] = ACTIONS(6618),
    [anon_sym_COLON_COLON] = ACTIONS(6616),
    [anon_sym_PIPE] = ACTIONS(9579),
    [anon_sym_AMP] = ACTIONS(6618),
    [anon_sym_LBRACK] = ACTIONS(6618),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6616),
    [anon_sym_LBRACE] = ACTIONS(6618),
    [anon_sym_LT_AT] = ACTIONS(6618),
    [anon_sym_LT_AT_AT] = ACTIONS(6618),
    [anon_sym_CARET] = ACTIONS(6618),
    [anon_sym_or] = ACTIONS(6618),
    [anon_sym_DOT] = ACTIONS(6618),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6616),
    [anon_sym_new] = ACTIONS(6618),
    [anon_sym_return] = ACTIONS(6618),
    [anon_sym_return_BANG] = ACTIONS(6616),
    [anon_sym_yield] = ACTIONS(6618),
    [anon_sym_yield_BANG] = ACTIONS(6616),
    [anon_sym_lazy] = ACTIONS(6618),
    [anon_sym_assert] = ACTIONS(6618),
    [anon_sym_upcast] = ACTIONS(6618),
    [anon_sym_downcast] = ACTIONS(6618),
    [anon_sym_fixed] = ACTIONS(6618),
    [anon_sym_COLON_GT] = ACTIONS(6616),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6616),
    [anon_sym_for] = ACTIONS(6618),
    [anon_sym_while] = ACTIONS(6618),
    [anon_sym_while_BANG] = ACTIONS(6616),
    [anon_sym_if] = ACTIONS(6618),
    [anon_sym_fun] = ACTIONS(6618),
    [anon_sym_try] = ACTIONS(6618),
    [anon_sym_match] = ACTIONS(6618),
    [anon_sym_match_BANG] = ACTIONS(6616),
    [anon_sym_function] = ACTIONS(6618),
    [anon_sym_DOT_LBRACK] = ACTIONS(6616),
    [anon_sym_use] = ACTIONS(6618),
    [anon_sym_use_BANG] = ACTIONS(6616),
    [anon_sym_do_BANG] = ACTIONS(6616),
    [anon_sym_LPAREN2] = ACTIONS(6618),
    [anon_sym_PIPE_PIPE] = ACTIONS(6618),
    [anon_sym_AMP_AMP] = ACTIONS(6618),
    [aux_sym_char_token1] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6618),
    [anon_sym_DQUOTE] = ACTIONS(6618),
    [anon_sym_AT_DQUOTE] = ACTIONS(6616),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6616),
    [sym_bool] = ACTIONS(6618),
    [sym_unit] = ACTIONS(6616),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6618),
    [sym_op_identifier] = ACTIONS(6618),
    [anon_sym_PLUS] = ACTIONS(6618),
    [anon_sym_DASH] = ACTIONS(6618),
    [anon_sym_PLUS_DOT] = ACTIONS(6618),
    [anon_sym_DASH_DOT] = ACTIONS(6618),
    [anon_sym_PERCENT] = ACTIONS(6618),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6618),
    [anon_sym_TILDE] = ACTIONS(6616),
    [aux_sym_prefix_op_token1] = ACTIONS(6618),
    [aux_sym_infix_op_token1] = ACTIONS(6616),
    [aux_sym_infix_op_token2] = ACTIONS(6618),
    [aux_sym_infix_op_token3] = ACTIONS(6618),
    [anon_sym_BANG_EQ] = ACTIONS(6618),
    [anon_sym_COLON_EQ] = ACTIONS(6616),
    [anon_sym_DOLLAR] = ACTIONS(6618),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6618),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6618),
    [sym_int] = ACTIONS(6618),
    [sym_xint] = ACTIONS(6616),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6616),
    [anon_sym_POUNDendif] = ACTIONS(6616),
    [sym__newline] = ACTIONS(9644),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6616),
    [sym__tyapp_open] = ACTIONS(6616),
  },
  [STATE(5277)] = {
    [sym_block_comment] = STATE(5277),
    [sym_compiler_directive_decl] = STATE(5277),
    [sym_fsi_directive_decl] = STATE(5277),
    [sym_preproc_line] = STATE(5277),
    [aux_sym_rules_repeat1] = STATE(5277),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9647),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [anon_sym_POUNDendif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9650),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5278)] = {
    [sym_block_comment] = STATE(5278),
    [sym_compiler_directive_decl] = STATE(5278),
    [sym_fsi_directive_decl] = STATE(5278),
    [sym_preproc_line] = STATE(5278),
    [aux_sym_long_identifier_repeat1] = STATE(5205),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9344),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5279)] = {
    [sym_type_arguments] = STATE(6476),
    [sym_long_identifier] = STATE(6492),
    [sym_block_comment] = STATE(5279),
    [sym_compiler_directive_decl] = STATE(5279),
    [sym_fsi_directive_decl] = STATE(5279),
    [sym_preproc_line] = STATE(5279),
    [aux_sym_compound_type_repeat1] = STATE(6388),
    [sym_identifier] = ACTIONS(9182),
    [anon_sym_module] = ACTIONS(9182),
    [anon_sym_null] = ACTIONS(9182),
    [anon_sym_begin] = ACTIONS(9182),
    [anon_sym_open] = ACTIONS(9182),
    [anon_sym_type] = ACTIONS(9182),
    [anon_sym_LBRACK_LT] = ACTIONS(9184),
    [anon_sym_do] = ACTIONS(9182),
    [anon_sym_and] = ACTIONS(9182),
    [anon_sym_let] = ACTIONS(9182),
    [anon_sym_let_BANG] = ACTIONS(9184),
    [aux_sym_access_modifier_token1] = ACTIONS(9184),
    [anon_sym_LPAREN] = ACTIONS(9182),
    [anon_sym_PIPE] = ACTIONS(9184),
    [anon_sym_AMP] = ACTIONS(9182),
    [anon_sym_LBRACK] = ACTIONS(9182),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9184),
    [anon_sym_LBRACE] = ACTIONS(9182),
    [anon_sym_LT_AT] = ACTIONS(9182),
    [anon_sym_LT_AT_AT] = ACTIONS(9184),
    [anon_sym_CARET] = ACTIONS(9184),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9184),
    [anon_sym_new] = ACTIONS(9182),
    [anon_sym_with] = ACTIONS(9182),
    [anon_sym_return] = ACTIONS(9182),
    [anon_sym_return_BANG] = ACTIONS(9184),
    [anon_sym_yield] = ACTIONS(9182),
    [anon_sym_yield_BANG] = ACTIONS(9184),
    [anon_sym_lazy] = ACTIONS(9182),
    [anon_sym_assert] = ACTIONS(9182),
    [anon_sym_upcast] = ACTIONS(9182),
    [anon_sym_downcast] = ACTIONS(9182),
    [anon_sym_fixed] = ACTIONS(9182),
    [anon_sym_for] = ACTIONS(9182),
    [anon_sym_while] = ACTIONS(9182),
    [anon_sym_while_BANG] = ACTIONS(9184),
    [anon_sym_if] = ACTIONS(9182),
    [anon_sym_fun] = ACTIONS(9182),
    [anon_sym_DASH_GT] = ACTIONS(9368),
    [anon_sym_try] = ACTIONS(9182),
    [anon_sym_match] = ACTIONS(9182),
    [anon_sym_match_BANG] = ACTIONS(9184),
    [anon_sym_function] = ACTIONS(9182),
    [anon_sym_use] = ACTIONS(9182),
    [anon_sym_use_BANG] = ACTIONS(9184),
    [anon_sym_do_BANG] = ACTIONS(9184),
    [anon_sym_AMP_AMP] = ACTIONS(9184),
    [anon_sym_STAR] = ACTIONS(9411),
    [anon_sym_LT] = ACTIONS(9370),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9372),
    [anon_sym_default] = ACTIONS(9182),
    [anon_sym_static] = ACTIONS(9182),
    [anon_sym_member] = ACTIONS(9182),
    [anon_sym_exception] = ACTIONS(9182),
    [anon_sym_interface] = ACTIONS(9182),
    [anon_sym_abstract] = ACTIONS(9182),
    [anon_sym_val] = ACTIONS(9182),
    [anon_sym_override] = ACTIONS(9182),
    [anon_sym_extern] = ACTIONS(9182),
    [aux_sym_char_token1] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9182),
    [anon_sym_DQUOTE] = ACTIONS(9182),
    [anon_sym_AT_DQUOTE] = ACTIONS(9184),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9184),
    [sym_bool] = ACTIONS(9182),
    [sym_unit] = ACTIONS(9184),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9182),
    [sym_op_identifier] = ACTIONS(9184),
    [anon_sym_PLUS] = ACTIONS(9182),
    [anon_sym_DASH] = ACTIONS(9182),
    [anon_sym_PLUS_DOT] = ACTIONS(9184),
    [anon_sym_DASH_DOT] = ACTIONS(9184),
    [anon_sym_PERCENT] = ACTIONS(9182),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9184),
    [anon_sym_TILDE] = ACTIONS(9184),
    [aux_sym_prefix_op_token1] = ACTIONS(9184),
    [sym_int] = ACTIONS(9182),
    [sym_xint] = ACTIONS(9184),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9184),
    [anon_sym_POUNDload] = ACTIONS(9184),
    [anon_sym_POUNDtime] = ACTIONS(9184),
    [anon_sym_POUNDI] = ACTIONS(9184),
    [anon_sym_POUNDhelp] = ACTIONS(9184),
    [anon_sym_POUNDquit] = ACTIONS(9184),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9184),
    [anon_sym_POUNDendif] = ACTIONS(9184),
    [anon_sym_POUNDelse] = ACTIONS(9184),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9184),
  },
  [STATE(5280)] = {
    [sym_block_comment] = STATE(5280),
    [sym_compiler_directive_decl] = STATE(5280),
    [sym_fsi_directive_decl] = STATE(5280),
    [sym_preproc_line] = STATE(5280),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [anon_sym_POUNDendif] = ACTIONS(6873),
    [anon_sym_POUNDelse] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5281)] = {
    [sym_block_comment] = STATE(5281),
    [sym_compiler_directive_decl] = STATE(5281),
    [sym_fsi_directive_decl] = STATE(5281),
    [sym_preproc_line] = STATE(5281),
    [aux_sym_rules_repeat1] = STATE(4960),
    [sym_identifier] = ACTIONS(6590),
    [anon_sym_null] = ACTIONS(6590),
    [anon_sym_EQ] = ACTIONS(6590),
    [anon_sym_begin] = ACTIONS(6590),
    [anon_sym_COLON] = ACTIONS(6590),
    [anon_sym_do] = ACTIONS(6590),
    [anon_sym_let] = ACTIONS(6590),
    [anon_sym_let_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN] = ACTIONS(6590),
    [anon_sym_COMMA] = ACTIONS(6588),
    [anon_sym_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_QMARK] = ACTIONS(6590),
    [anon_sym_COLON_COLON] = ACTIONS(6588),
    [anon_sym_PIPE] = ACTIONS(9376),
    [anon_sym_AMP] = ACTIONS(6590),
    [anon_sym_GT] = ACTIONS(6590),
    [anon_sym_LBRACK] = ACTIONS(6590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6588),
    [anon_sym_LBRACE] = ACTIONS(6590),
    [anon_sym_LT_AT] = ACTIONS(6590),
    [anon_sym_LT_AT_AT] = ACTIONS(6590),
    [anon_sym_CARET] = ACTIONS(6590),
    [anon_sym_or] = ACTIONS(6590),
    [anon_sym_DOT] = ACTIONS(6590),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6588),
    [anon_sym_new] = ACTIONS(6590),
    [anon_sym_return] = ACTIONS(6590),
    [anon_sym_return_BANG] = ACTIONS(6588),
    [anon_sym_yield] = ACTIONS(6590),
    [anon_sym_yield_BANG] = ACTIONS(6588),
    [anon_sym_lazy] = ACTIONS(6590),
    [anon_sym_assert] = ACTIONS(6590),
    [anon_sym_upcast] = ACTIONS(6590),
    [anon_sym_downcast] = ACTIONS(6590),
    [anon_sym_fixed] = ACTIONS(6590),
    [anon_sym_COLON_GT] = ACTIONS(6588),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6588),
    [anon_sym_for] = ACTIONS(6590),
    [anon_sym_while] = ACTIONS(6590),
    [anon_sym_while_BANG] = ACTIONS(6588),
    [anon_sym_if] = ACTIONS(6590),
    [anon_sym_fun] = ACTIONS(6590),
    [anon_sym_try] = ACTIONS(6590),
    [anon_sym_match] = ACTIONS(6590),
    [anon_sym_match_BANG] = ACTIONS(6588),
    [anon_sym_function] = ACTIONS(6590),
    [anon_sym_DOT_LBRACK] = ACTIONS(6588),
    [anon_sym_use] = ACTIONS(6590),
    [anon_sym_use_BANG] = ACTIONS(6588),
    [anon_sym_do_BANG] = ACTIONS(6588),
    [anon_sym_LPAREN2] = ACTIONS(6590),
    [anon_sym_PIPE_PIPE] = ACTIONS(6590),
    [anon_sym_AMP_AMP] = ACTIONS(6590),
    [aux_sym_char_token1] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6590),
    [anon_sym_DQUOTE] = ACTIONS(6590),
    [anon_sym_AT_DQUOTE] = ACTIONS(6588),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6588),
    [sym_bool] = ACTIONS(6590),
    [sym_unit] = ACTIONS(6588),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6590),
    [sym_op_identifier] = ACTIONS(6590),
    [anon_sym_PLUS] = ACTIONS(6590),
    [anon_sym_DASH] = ACTIONS(6590),
    [anon_sym_PLUS_DOT] = ACTIONS(6590),
    [anon_sym_DASH_DOT] = ACTIONS(6590),
    [anon_sym_PERCENT] = ACTIONS(6590),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6590),
    [anon_sym_TILDE] = ACTIONS(6588),
    [aux_sym_prefix_op_token1] = ACTIONS(6590),
    [aux_sym_infix_op_token1] = ACTIONS(6588),
    [aux_sym_infix_op_token2] = ACTIONS(6590),
    [aux_sym_infix_op_token3] = ACTIONS(6590),
    [anon_sym_BANG_EQ] = ACTIONS(6590),
    [anon_sym_COLON_EQ] = ACTIONS(6588),
    [anon_sym_DOLLAR] = ACTIONS(6590),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6590),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6590),
    [sym_int] = ACTIONS(6590),
    [sym_xint] = ACTIONS(6588),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6588),
    [sym__newline] = ACTIONS(9653),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6588),
    [sym__tyapp_open] = ACTIONS(6588),
  },
  [STATE(5282)] = {
    [sym_type_arguments] = STATE(6479),
    [sym_long_identifier] = STATE(6488),
    [sym_block_comment] = STATE(5282),
    [sym_compiler_directive_decl] = STATE(5282),
    [sym_fsi_directive_decl] = STATE(5282),
    [sym_preproc_line] = STATE(5282),
    [aux_sym_compound_type_repeat1] = STATE(6041),
    [sym_identifier] = ACTIONS(9526),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_PIPE] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6259),
    [anon_sym_CARET] = ACTIONS(6259),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(9528),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_AMP_AMP] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(9549),
    [anon_sym_LT] = ACTIONS(9530),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9532),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6259),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6259),
    [anon_sym_DASH_DOT] = ACTIONS(6259),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6259),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6259),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__newline] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
  },
  [STATE(5283)] = {
    [sym_block_comment] = STATE(5283),
    [sym_compiler_directive_decl] = STATE(5283),
    [sym_fsi_directive_decl] = STATE(5283),
    [sym_preproc_line] = STATE(5283),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_else] = ACTIONS(6612),
    [anon_sym_elif] = ACTIONS(6612),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__quoted_close] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5284)] = {
    [sym_block_comment] = STATE(5284),
    [sym_compiler_directive_decl] = STATE(5284),
    [sym_fsi_directive_decl] = STATE(5284),
    [sym_preproc_line] = STATE(5284),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(8053),
    [aux_sym_decimal_token1] = ACTIONS(7783),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5285)] = {
    [sym_block_comment] = STATE(5285),
    [sym_compiler_directive_decl] = STATE(5285),
    [sym_fsi_directive_decl] = STATE(5285),
    [sym_preproc_line] = STATE(5285),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_with] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9656),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5286)] = {
    [sym_block_comment] = STATE(5286),
    [sym_compiler_directive_decl] = STATE(5286),
    [sym_fsi_directive_decl] = STATE(5286),
    [sym_preproc_line] = STATE(5286),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_DOT_DOT] = ACTIONS(290),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [anon_sym_DOT_DOT2] = ACTIONS(292),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(5287)] = {
    [sym_block_comment] = STATE(5287),
    [sym_compiler_directive_decl] = STATE(5287),
    [sym_fsi_directive_decl] = STATE(5287),
    [sym_preproc_line] = STATE(5287),
    [aux_sym_rules_repeat1] = STATE(5287),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9658),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9661),
    [sym__quoted_close] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5288)] = {
    [sym_block_comment] = STATE(5288),
    [sym_compiler_directive_decl] = STATE(5288),
    [sym_fsi_directive_decl] = STATE(5288),
    [sym_preproc_line] = STATE(5288),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_GT] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9664),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5289)] = {
    [sym_block_comment] = STATE(5289),
    [sym_compiler_directive_decl] = STATE(5289),
    [sym_fsi_directive_decl] = STATE(5289),
    [sym_preproc_line] = STATE(5289),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_DOT_DOT] = ACTIONS(6932),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [anon_sym_DOT_DOT2] = ACTIONS(6930),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__dedent] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(5290)] = {
    [sym_block_comment] = STATE(5290),
    [sym_compiler_directive_decl] = STATE(5290),
    [sym_fsi_directive_decl] = STATE(5290),
    [sym_preproc_line] = STATE(5290),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9666),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [anon_sym_POUNDendif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(5291)] = {
    [sym_block_comment] = STATE(5291),
    [sym_compiler_directive_decl] = STATE(5291),
    [sym_fsi_directive_decl] = STATE(5291),
    [sym_preproc_line] = STATE(5291),
    [aux_sym_long_identifier_repeat1] = STATE(5601),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9668),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9670),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5292)] = {
    [sym_block_comment] = STATE(5292),
    [sym_compiler_directive_decl] = STATE(5292),
    [sym_fsi_directive_decl] = STATE(5292),
    [sym_preproc_line] = STATE(5292),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_DOT_DOT] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5293)] = {
    [sym_block_comment] = STATE(5293),
    [sym_compiler_directive_decl] = STATE(5293),
    [sym_fsi_directive_decl] = STATE(5293),
    [sym_preproc_line] = STATE(5293),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [anon_sym_POUNDendif] = ACTIONS(6830),
    [anon_sym_POUNDelse] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5294)] = {
    [sym_block_comment] = STATE(5294),
    [sym_compiler_directive_decl] = STATE(5294),
    [sym_fsi_directive_decl] = STATE(5294),
    [sym_preproc_line] = STATE(5294),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_DOT_DOT] = ACTIONS(7039),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [anon_sym_DOT_DOT2] = ACTIONS(7037),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__dedent] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(5295)] = {
    [sym_type_arguments] = STATE(6477),
    [sym_long_identifier] = STATE(6505),
    [sym_block_comment] = STATE(5295),
    [sym_compiler_directive_decl] = STATE(5295),
    [sym_fsi_directive_decl] = STATE(5295),
    [sym_preproc_line] = STATE(5295),
    [aux_sym_compound_type_repeat1] = STATE(6318),
    [ts_builtin_sym_end] = ACTIONS(9210),
    [sym_identifier] = ACTIONS(9208),
    [anon_sym_namespace] = ACTIONS(9208),
    [anon_sym_module] = ACTIONS(9208),
    [anon_sym_null] = ACTIONS(9208),
    [anon_sym_begin] = ACTIONS(9208),
    [anon_sym_open] = ACTIONS(9208),
    [anon_sym_type] = ACTIONS(9208),
    [anon_sym_LBRACK_LT] = ACTIONS(9210),
    [anon_sym_do] = ACTIONS(9208),
    [anon_sym_and] = ACTIONS(9208),
    [anon_sym_let] = ACTIONS(9208),
    [anon_sym_let_BANG] = ACTIONS(9210),
    [aux_sym_access_modifier_token1] = ACTIONS(9210),
    [anon_sym_LPAREN] = ACTIONS(9208),
    [anon_sym_PIPE] = ACTIONS(9210),
    [anon_sym_AMP] = ACTIONS(9208),
    [anon_sym_LBRACK] = ACTIONS(9208),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9210),
    [anon_sym_LBRACE] = ACTIONS(9208),
    [anon_sym_LT_AT] = ACTIONS(9208),
    [anon_sym_LT_AT_AT] = ACTIONS(9210),
    [anon_sym_CARET] = ACTIONS(9210),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9210),
    [anon_sym_new] = ACTIONS(9208),
    [anon_sym_with] = ACTIONS(9208),
    [anon_sym_return] = ACTIONS(9208),
    [anon_sym_return_BANG] = ACTIONS(9210),
    [anon_sym_yield] = ACTIONS(9208),
    [anon_sym_yield_BANG] = ACTIONS(9210),
    [anon_sym_lazy] = ACTIONS(9208),
    [anon_sym_assert] = ACTIONS(9208),
    [anon_sym_upcast] = ACTIONS(9208),
    [anon_sym_downcast] = ACTIONS(9208),
    [anon_sym_fixed] = ACTIONS(9208),
    [anon_sym_for] = ACTIONS(9208),
    [anon_sym_while] = ACTIONS(9208),
    [anon_sym_while_BANG] = ACTIONS(9210),
    [anon_sym_if] = ACTIONS(9208),
    [anon_sym_fun] = ACTIONS(9208),
    [anon_sym_DASH_GT] = ACTIONS(9385),
    [anon_sym_try] = ACTIONS(9208),
    [anon_sym_match] = ACTIONS(9208),
    [anon_sym_match_BANG] = ACTIONS(9210),
    [anon_sym_function] = ACTIONS(9208),
    [anon_sym_use] = ACTIONS(9208),
    [anon_sym_use_BANG] = ACTIONS(9210),
    [anon_sym_do_BANG] = ACTIONS(9210),
    [anon_sym_AMP_AMP] = ACTIONS(9210),
    [anon_sym_STAR] = ACTIONS(9387),
    [anon_sym_LT] = ACTIONS(9389),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9391),
    [anon_sym_default] = ACTIONS(9208),
    [anon_sym_static] = ACTIONS(9208),
    [anon_sym_member] = ACTIONS(9208),
    [anon_sym_exception] = ACTIONS(9208),
    [anon_sym_interface] = ACTIONS(9208),
    [anon_sym_abstract] = ACTIONS(9208),
    [anon_sym_val] = ACTIONS(9208),
    [anon_sym_override] = ACTIONS(9208),
    [anon_sym_extern] = ACTIONS(9208),
    [aux_sym_char_token1] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9208),
    [anon_sym_DQUOTE] = ACTIONS(9208),
    [anon_sym_AT_DQUOTE] = ACTIONS(9210),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9210),
    [sym_bool] = ACTIONS(9208),
    [sym_unit] = ACTIONS(9210),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9208),
    [sym_op_identifier] = ACTIONS(9210),
    [anon_sym_PLUS] = ACTIONS(9208),
    [anon_sym_DASH] = ACTIONS(9208),
    [anon_sym_PLUS_DOT] = ACTIONS(9210),
    [anon_sym_DASH_DOT] = ACTIONS(9210),
    [anon_sym_PERCENT] = ACTIONS(9208),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9210),
    [anon_sym_TILDE] = ACTIONS(9210),
    [aux_sym_prefix_op_token1] = ACTIONS(9210),
    [sym_int] = ACTIONS(9208),
    [sym_xint] = ACTIONS(9210),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9210),
    [anon_sym_POUNDload] = ACTIONS(9210),
    [anon_sym_POUNDtime] = ACTIONS(9210),
    [anon_sym_POUNDI] = ACTIONS(9210),
    [anon_sym_POUNDhelp] = ACTIONS(9210),
    [anon_sym_POUNDquit] = ACTIONS(9210),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9210),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9210),
  },
  [STATE(5296)] = {
    [sym_block_comment] = STATE(5296),
    [sym_compiler_directive_decl] = STATE(5296),
    [sym_fsi_directive_decl] = STATE(5296),
    [sym_preproc_line] = STATE(5296),
    [sym_identifier] = ACTIONS(6560),
    [anon_sym_null] = ACTIONS(6560),
    [anon_sym_EQ] = ACTIONS(6560),
    [anon_sym_begin] = ACTIONS(6560),
    [anon_sym_COLON] = ACTIONS(6560),
    [anon_sym_do] = ACTIONS(6560),
    [anon_sym_let] = ACTIONS(6560),
    [anon_sym_let_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN] = ACTIONS(6560),
    [anon_sym_COMMA] = ACTIONS(6558),
    [anon_sym_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_QMARK] = ACTIONS(6560),
    [anon_sym_COLON_COLON] = ACTIONS(6558),
    [anon_sym_AMP] = ACTIONS(6560),
    [anon_sym_LBRACK] = ACTIONS(6560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6558),
    [anon_sym_LBRACE] = ACTIONS(6560),
    [anon_sym_LT_AT] = ACTIONS(6560),
    [anon_sym_LT_AT_AT] = ACTIONS(6560),
    [anon_sym_CARET] = ACTIONS(6560),
    [anon_sym_or] = ACTIONS(6560),
    [anon_sym_DOT] = ACTIONS(6560),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6558),
    [anon_sym_new] = ACTIONS(6560),
    [anon_sym_return] = ACTIONS(6560),
    [anon_sym_return_BANG] = ACTIONS(6558),
    [anon_sym_yield] = ACTIONS(6560),
    [anon_sym_yield_BANG] = ACTIONS(6558),
    [anon_sym_lazy] = ACTIONS(6560),
    [anon_sym_assert] = ACTIONS(6560),
    [anon_sym_upcast] = ACTIONS(6560),
    [anon_sym_downcast] = ACTIONS(6560),
    [anon_sym_fixed] = ACTIONS(6560),
    [anon_sym_COLON_GT] = ACTIONS(6558),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6558),
    [anon_sym_for] = ACTIONS(6560),
    [anon_sym_while] = ACTIONS(6560),
    [anon_sym_while_BANG] = ACTIONS(6558),
    [anon_sym_if] = ACTIONS(6560),
    [anon_sym_fun] = ACTIONS(6560),
    [anon_sym_try] = ACTIONS(6560),
    [anon_sym_match] = ACTIONS(6560),
    [anon_sym_match_BANG] = ACTIONS(6558),
    [anon_sym_function] = ACTIONS(6560),
    [anon_sym_DOT_LBRACK] = ACTIONS(6558),
    [anon_sym_use] = ACTIONS(6560),
    [anon_sym_use_BANG] = ACTIONS(6558),
    [anon_sym_do_BANG] = ACTIONS(6558),
    [anon_sym_LPAREN2] = ACTIONS(6560),
    [anon_sym_PIPE_PIPE] = ACTIONS(6560),
    [anon_sym_AMP_AMP] = ACTIONS(6560),
    [aux_sym_char_token1] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6560),
    [anon_sym_DQUOTE] = ACTIONS(6560),
    [anon_sym_AT_DQUOTE] = ACTIONS(6558),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6558),
    [sym_bool] = ACTIONS(6560),
    [sym_unit] = ACTIONS(6558),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6560),
    [sym_op_identifier] = ACTIONS(6560),
    [anon_sym_PLUS] = ACTIONS(6560),
    [anon_sym_DASH] = ACTIONS(6560),
    [anon_sym_PLUS_DOT] = ACTIONS(6560),
    [anon_sym_DASH_DOT] = ACTIONS(6560),
    [anon_sym_PERCENT] = ACTIONS(6560),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6560),
    [anon_sym_TILDE] = ACTIONS(6558),
    [aux_sym_prefix_op_token1] = ACTIONS(6560),
    [aux_sym_infix_op_token1] = ACTIONS(6558),
    [aux_sym_infix_op_token2] = ACTIONS(6560),
    [aux_sym_infix_op_token3] = ACTIONS(6560),
    [anon_sym_BANG_EQ] = ACTIONS(6560),
    [anon_sym_COLON_EQ] = ACTIONS(6558),
    [anon_sym_DOLLAR] = ACTIONS(6560),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6560),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6560),
    [sym_int] = ACTIONS(6560),
    [sym_xint] = ACTIONS(6558),
    [anon_sym_f] = ACTIONS(6560),
    [aux_sym_decimal_token1] = ACTIONS(6560),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6558),
    [sym__newline] = ACTIONS(6558),
    [sym__quoted_close] = ACTIONS(6558),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6558),
    [sym__tyapp_open] = ACTIONS(6558),
  },
  [STATE(5297)] = {
    [sym_block_comment] = STATE(5297),
    [sym_compiler_directive_decl] = STATE(5297),
    [sym_fsi_directive_decl] = STATE(5297),
    [sym_preproc_line] = STATE(5297),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_DASH_GT] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_DOT_DOT] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5298)] = {
    [sym_block_comment] = STATE(5298),
    [sym_compiler_directive_decl] = STATE(5298),
    [sym_fsi_directive_decl] = STATE(5298),
    [sym_preproc_line] = STATE(5298),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_done] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5299)] = {
    [sym_block_comment] = STATE(5299),
    [sym_compiler_directive_decl] = STATE(5299),
    [sym_fsi_directive_decl] = STATE(5299),
    [sym_preproc_line] = STATE(5299),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9672),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__quoted_close] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(5300)] = {
    [sym_block_comment] = STATE(5300),
    [sym_compiler_directive_decl] = STATE(5300),
    [sym_fsi_directive_decl] = STATE(5300),
    [sym_preproc_line] = STATE(5300),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__dedent] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
    [sym__in] = ACTIONS(6752),
  },
  [STATE(5301)] = {
    [sym_block_comment] = STATE(5301),
    [sym_compiler_directive_decl] = STATE(5301),
    [sym_fsi_directive_decl] = STATE(5301),
    [sym_preproc_line] = STATE(5301),
    [aux_sym_rules_repeat1] = STATE(5301),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(9674),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_with] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(9677),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5302)] = {
    [sym_block_comment] = STATE(5302),
    [sym_compiler_directive_decl] = STATE(5302),
    [sym_fsi_directive_decl] = STATE(5302),
    [sym_preproc_line] = STATE(5302),
    [sym_identifier] = ACTIONS(6513),
    [anon_sym_null] = ACTIONS(6513),
    [anon_sym_EQ] = ACTIONS(6513),
    [anon_sym_begin] = ACTIONS(6513),
    [anon_sym_COLON] = ACTIONS(6513),
    [anon_sym_do] = ACTIONS(6513),
    [anon_sym_let] = ACTIONS(6513),
    [anon_sym_let_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN] = ACTIONS(6513),
    [anon_sym_COMMA] = ACTIONS(6515),
    [anon_sym_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_QMARK] = ACTIONS(6513),
    [anon_sym_COLON_COLON] = ACTIONS(6515),
    [anon_sym_AMP] = ACTIONS(6513),
    [anon_sym_LBRACK] = ACTIONS(6513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6515),
    [anon_sym_LBRACE] = ACTIONS(6513),
    [anon_sym_LT_AT] = ACTIONS(6513),
    [anon_sym_LT_AT_AT] = ACTIONS(6513),
    [anon_sym_CARET] = ACTIONS(6513),
    [anon_sym_or] = ACTIONS(6513),
    [anon_sym_DOT] = ACTIONS(6513),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6515),
    [anon_sym_new] = ACTIONS(6513),
    [anon_sym_return] = ACTIONS(6513),
    [anon_sym_return_BANG] = ACTIONS(6515),
    [anon_sym_yield] = ACTIONS(6513),
    [anon_sym_yield_BANG] = ACTIONS(6515),
    [anon_sym_lazy] = ACTIONS(6513),
    [anon_sym_assert] = ACTIONS(6513),
    [anon_sym_upcast] = ACTIONS(6513),
    [anon_sym_downcast] = ACTIONS(6513),
    [anon_sym_fixed] = ACTIONS(6513),
    [anon_sym_COLON_GT] = ACTIONS(6515),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6515),
    [anon_sym_for] = ACTIONS(6513),
    [anon_sym_while] = ACTIONS(6513),
    [anon_sym_while_BANG] = ACTIONS(6515),
    [anon_sym_if] = ACTIONS(6513),
    [anon_sym_fun] = ACTIONS(6513),
    [anon_sym_try] = ACTIONS(6513),
    [anon_sym_match] = ACTIONS(6513),
    [anon_sym_match_BANG] = ACTIONS(6515),
    [anon_sym_function] = ACTIONS(6513),
    [anon_sym_DOT_LBRACK] = ACTIONS(6515),
    [anon_sym_use] = ACTIONS(6513),
    [anon_sym_use_BANG] = ACTIONS(6515),
    [anon_sym_do_BANG] = ACTIONS(6515),
    [anon_sym_LPAREN2] = ACTIONS(6513),
    [anon_sym_PIPE_PIPE] = ACTIONS(6513),
    [anon_sym_AMP_AMP] = ACTIONS(6513),
    [aux_sym_char_token1] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6513),
    [anon_sym_DQUOTE] = ACTIONS(6513),
    [anon_sym_AT_DQUOTE] = ACTIONS(6515),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6515),
    [sym_bool] = ACTIONS(6513),
    [sym_unit] = ACTIONS(6515),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6513),
    [sym_op_identifier] = ACTIONS(6513),
    [anon_sym_PLUS] = ACTIONS(6513),
    [anon_sym_DASH] = ACTIONS(6513),
    [anon_sym_PLUS_DOT] = ACTIONS(6513),
    [anon_sym_DASH_DOT] = ACTIONS(6513),
    [anon_sym_PERCENT] = ACTIONS(6513),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6513),
    [anon_sym_TILDE] = ACTIONS(6515),
    [aux_sym_prefix_op_token1] = ACTIONS(6513),
    [aux_sym_infix_op_token1] = ACTIONS(6515),
    [aux_sym_infix_op_token2] = ACTIONS(6513),
    [aux_sym_infix_op_token3] = ACTIONS(6513),
    [anon_sym_BANG_EQ] = ACTIONS(6513),
    [anon_sym_COLON_EQ] = ACTIONS(6515),
    [anon_sym_DOLLAR] = ACTIONS(6513),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6513),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6513),
    [sym_int] = ACTIONS(6513),
    [sym_xint] = ACTIONS(6515),
    [anon_sym_f] = ACTIONS(6513),
    [aux_sym_decimal_token1] = ACTIONS(6513),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6515),
    [anon_sym_POUNDendif] = ACTIONS(6515),
    [sym__newline] = ACTIONS(6515),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6515),
    [sym__tyapp_open] = ACTIONS(6515),
  },
  [STATE(5303)] = {
    [sym_block_comment] = STATE(5303),
    [sym_compiler_directive_decl] = STATE(5303),
    [sym_fsi_directive_decl] = STATE(5303),
    [sym_preproc_line] = STATE(5303),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_f] = ACTIONS(7972),
    [aux_sym_decimal_token1] = ACTIONS(7662),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__quoted_close] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5304)] = {
    [sym_block_comment] = STATE(5304),
    [sym_compiler_directive_decl] = STATE(5304),
    [sym_fsi_directive_decl] = STATE(5304),
    [sym_preproc_line] = STATE(5304),
    [aux_sym_long_identifier_repeat1] = STATE(5045),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9680),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(5305)] = {
    [sym_block_comment] = STATE(5305),
    [sym_compiler_directive_decl] = STATE(5305),
    [sym_fsi_directive_decl] = STATE(5305),
    [sym_preproc_line] = STATE(5305),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__dedent] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
    [sym__in] = ACTIONS(6873),
  },
  [STATE(5306)] = {
    [sym_block_comment] = STATE(5306),
    [sym_compiler_directive_decl] = STATE(5306),
    [sym_fsi_directive_decl] = STATE(5306),
    [sym_preproc_line] = STATE(5306),
    [aux_sym_long_identifier_repeat1] = STATE(5618),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9682),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9684),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5307)] = {
    [sym_block_comment] = STATE(5307),
    [sym_compiler_directive_decl] = STATE(5307),
    [sym_fsi_directive_decl] = STATE(5307),
    [sym_preproc_line] = STATE(5307),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_DOT_DOT] = ACTIONS(7043),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [anon_sym_DOT_DOT2] = ACTIONS(7041),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__dedent] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(5308)] = {
    [sym_block_comment] = STATE(5308),
    [sym_compiler_directive_decl] = STATE(5308),
    [sym_fsi_directive_decl] = STATE(5308),
    [sym_preproc_line] = STATE(5308),
    [aux_sym_long_identifier_repeat1] = STATE(4954),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_EQ] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_COLON] = ACTIONS(6360),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_COMMA] = ACTIONS(6358),
    [anon_sym_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_QMARK] = ACTIONS(6360),
    [anon_sym_COLON_COLON] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6360),
    [anon_sym_CARET] = ACTIONS(6360),
    [anon_sym_or] = ACTIONS(6360),
    [anon_sym_DOT] = ACTIONS(9686),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_COLON_GT] = ACTIONS(6358),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6358),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_LT_DASH] = ACTIONS(6360),
    [anon_sym_DOT_LBRACK] = ACTIONS(6358),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_DOT_DOT] = ACTIONS(6358),
    [anon_sym_LPAREN2] = ACTIONS(6360),
    [anon_sym_PIPE_PIPE] = ACTIONS(6360),
    [anon_sym_AMP_AMP] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6360),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6360),
    [anon_sym_DASH_DOT] = ACTIONS(6360),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6360),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6360),
    [aux_sym_infix_op_token1] = ACTIONS(6358),
    [aux_sym_infix_op_token2] = ACTIONS(6360),
    [aux_sym_infix_op_token3] = ACTIONS(6360),
    [anon_sym_BANG_EQ] = ACTIONS(6360),
    [anon_sym_COLON_EQ] = ACTIONS(6358),
    [anon_sym_DOLLAR] = ACTIONS(6360),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6360),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6360),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
    [sym__tyapp_open] = ACTIONS(6358),
  },
  [STATE(5309)] = {
    [sym_block_comment] = STATE(5309),
    [sym_compiler_directive_decl] = STATE(5309),
    [sym_fsi_directive_decl] = STATE(5309),
    [sym_preproc_line] = STATE(5309),
    [sym_identifier] = ACTIONS(6637),
    [anon_sym_null] = ACTIONS(6637),
    [anon_sym_EQ] = ACTIONS(6637),
    [anon_sym_begin] = ACTIONS(6637),
    [anon_sym_COLON] = ACTIONS(6637),
    [anon_sym_do] = ACTIONS(6637),
    [anon_sym_let] = ACTIONS(6637),
    [anon_sym_let_BANG] = ACTIONS(6635),
    [anon_sym_LPAREN] = ACTIONS(6637),
    [anon_sym_COMMA] = ACTIONS(6635),
    [anon_sym_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_QMARK] = ACTIONS(6637),
    [anon_sym_COLON_COLON] = ACTIONS(6635),
    [anon_sym_AMP] = ACTIONS(6637),
    [anon_sym_LBRACK] = ACTIONS(6637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6635),
    [anon_sym_LBRACE] = ACTIONS(6637),
    [anon_sym_LT_AT] = ACTIONS(6637),
    [anon_sym_LT_AT_AT] = ACTIONS(6637),
    [anon_sym_CARET] = ACTIONS(6637),
    [anon_sym_or] = ACTIONS(6637),
    [anon_sym_DOT] = ACTIONS(6637),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6635),
    [anon_sym_new] = ACTIONS(6637),
    [anon_sym_return] = ACTIONS(6637),
    [anon_sym_return_BANG] = ACTIONS(6635),
    [anon_sym_yield] = ACTIONS(6637),
    [anon_sym_yield_BANG] = ACTIONS(6635),
    [anon_sym_lazy] = ACTIONS(6637),
    [anon_sym_assert] = ACTIONS(6637),
    [anon_sym_upcast] = ACTIONS(6637),
    [anon_sym_downcast] = ACTIONS(6637),
    [anon_sym_fixed] = ACTIONS(6637),
    [anon_sym_COLON_GT] = ACTIONS(6635),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6635),
    [anon_sym_for] = ACTIONS(6637),
    [anon_sym_while] = ACTIONS(6637),
    [anon_sym_while_BANG] = ACTIONS(6635),
    [anon_sym_if] = ACTIONS(6637),
    [anon_sym_fun] = ACTIONS(6637),
    [anon_sym_try] = ACTIONS(6637),
    [anon_sym_match] = ACTIONS(6637),
    [anon_sym_match_BANG] = ACTIONS(6635),
    [anon_sym_function] = ACTIONS(6637),
    [anon_sym_DOT_LBRACK] = ACTIONS(6635),
    [anon_sym_use] = ACTIONS(6637),
    [anon_sym_use_BANG] = ACTIONS(6635),
    [anon_sym_do_BANG] = ACTIONS(6635),
    [anon_sym_DOT_DOT] = ACTIONS(6635),
    [anon_sym_LPAREN2] = ACTIONS(6637),
    [anon_sym_PIPE_PIPE] = ACTIONS(6637),
    [anon_sym_AMP_AMP] = ACTIONS(6637),
    [aux_sym_char_token1] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6637),
    [anon_sym_DQUOTE] = ACTIONS(6637),
    [anon_sym_AT_DQUOTE] = ACTIONS(6635),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6635),
    [sym_bool] = ACTIONS(6637),
    [sym_unit] = ACTIONS(6635),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6637),
    [sym_op_identifier] = ACTIONS(6637),
    [anon_sym_PLUS] = ACTIONS(6637),
    [anon_sym_DASH] = ACTIONS(6637),
    [anon_sym_PLUS_DOT] = ACTIONS(6637),
    [anon_sym_DASH_DOT] = ACTIONS(6637),
    [anon_sym_PERCENT] = ACTIONS(6637),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6637),
    [anon_sym_TILDE] = ACTIONS(6635),
    [aux_sym_prefix_op_token1] = ACTIONS(6637),
    [aux_sym_infix_op_token1] = ACTIONS(6635),
    [aux_sym_infix_op_token2] = ACTIONS(6637),
    [aux_sym_infix_op_token3] = ACTIONS(6637),
    [anon_sym_BANG_EQ] = ACTIONS(6637),
    [anon_sym_COLON_EQ] = ACTIONS(6635),
    [anon_sym_DOLLAR] = ACTIONS(6637),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6637),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6637),
    [sym_int] = ACTIONS(9688),
    [sym_xint] = ACTIONS(6635),
    [anon_sym_f] = ACTIONS(6637),
    [aux_sym_decimal_token1] = ACTIONS(6637),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6635),
    [sym__newline] = ACTIONS(6635),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6635),
    [sym__tyapp_open] = ACTIONS(6635),
  },
  [STATE(5310)] = {
    [sym_block_comment] = STATE(5310),
    [sym_compiler_directive_decl] = STATE(5310),
    [sym_fsi_directive_decl] = STATE(5310),
    [sym_preproc_line] = STATE(5310),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [anon_sym_POUNDendif] = ACTIONS(6977),
    [anon_sym_POUNDelse] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(5311)] = {
    [sym_block_comment] = STATE(5311),
    [sym_compiler_directive_decl] = STATE(5311),
    [sym_fsi_directive_decl] = STATE(5311),
    [sym_preproc_line] = STATE(5311),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(9690),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5312)] = {
    [sym_block_comment] = STATE(5312),
    [sym_compiler_directive_decl] = STATE(5312),
    [sym_fsi_directive_decl] = STATE(5312),
    [sym_preproc_line] = STATE(5312),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(9692),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5313)] = {
    [sym_block_comment] = STATE(5313),
    [sym_compiler_directive_decl] = STATE(5313),
    [sym_fsi_directive_decl] = STATE(5313),
    [sym_preproc_line] = STATE(5313),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [anon_sym_POUNDendif] = ACTIONS(6942),
    [anon_sym_POUNDelse] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(5314)] = {
    [sym_block_comment] = STATE(5314),
    [sym_compiler_directive_decl] = STATE(5314),
    [sym_fsi_directive_decl] = STATE(5314),
    [sym_preproc_line] = STATE(5314),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [anon_sym_POUNDendif] = ACTIONS(6946),
    [anon_sym_POUNDelse] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(5315)] = {
    [sym_block_comment] = STATE(5315),
    [sym_compiler_directive_decl] = STATE(5315),
    [sym_fsi_directive_decl] = STATE(5315),
    [sym_preproc_line] = STATE(5315),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_as] = ACTIONS(6936),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_with] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(5316)] = {
    [sym_block_comment] = STATE(5316),
    [sym_compiler_directive_decl] = STATE(5316),
    [sym_fsi_directive_decl] = STATE(5316),
    [sym_preproc_line] = STATE(5316),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [anon_sym_POUNDendif] = ACTIONS(6950),
    [anon_sym_POUNDelse] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(5317)] = {
    [sym_block_comment] = STATE(5317),
    [sym_compiler_directive_decl] = STATE(5317),
    [sym_fsi_directive_decl] = STATE(5317),
    [sym_preproc_line] = STATE(5317),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_as] = ACTIONS(6940),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_with] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(5318)] = {
    [sym_block_comment] = STATE(5318),
    [sym_compiler_directive_decl] = STATE(5318),
    [sym_fsi_directive_decl] = STATE(5318),
    [sym_preproc_line] = STATE(5318),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [anon_sym_POUNDendif] = ACTIONS(6957),
    [anon_sym_POUNDelse] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(5319)] = {
    [sym_block_comment] = STATE(5319),
    [sym_compiler_directive_decl] = STATE(5319),
    [sym_fsi_directive_decl] = STATE(5319),
    [sym_preproc_line] = STATE(5319),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [anon_sym_POUNDendif] = ACTIONS(6961),
    [anon_sym_POUNDelse] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(5320)] = {
    [sym_block_comment] = STATE(5320),
    [sym_compiler_directive_decl] = STATE(5320),
    [sym_fsi_directive_decl] = STATE(5320),
    [sym_preproc_line] = STATE(5320),
    [aux_sym_long_identifier_repeat1] = STATE(5601),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9670),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5321)] = {
    [sym_block_comment] = STATE(5321),
    [sym_compiler_directive_decl] = STATE(5321),
    [sym_fsi_directive_decl] = STATE(5321),
    [sym_preproc_line] = STATE(5321),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [anon_sym_POUNDendif] = ACTIONS(6965),
    [anon_sym_POUNDelse] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(5322)] = {
    [sym_block_comment] = STATE(5322),
    [sym_compiler_directive_decl] = STATE(5322),
    [sym_fsi_directive_decl] = STATE(5322),
    [sym_preproc_line] = STATE(5322),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [anon_sym_POUNDendif] = ACTIONS(6969),
    [anon_sym_POUNDelse] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(5323)] = {
    [sym_block_comment] = STATE(5323),
    [sym_compiler_directive_decl] = STATE(5323),
    [sym_fsi_directive_decl] = STATE(5323),
    [sym_preproc_line] = STATE(5323),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [anon_sym_POUNDendif] = ACTIONS(6973),
    [anon_sym_POUNDelse] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(5324)] = {
    [sym_block_comment] = STATE(5324),
    [sym_compiler_directive_decl] = STATE(5324),
    [sym_fsi_directive_decl] = STATE(5324),
    [sym_preproc_line] = STATE(5324),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [anon_sym_POUNDendif] = ACTIONS(6981),
    [anon_sym_POUNDelse] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(5325)] = {
    [sym_block_comment] = STATE(5325),
    [sym_compiler_directive_decl] = STATE(5325),
    [sym_fsi_directive_decl] = STATE(5325),
    [sym_preproc_line] = STATE(5325),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [anon_sym_POUNDendif] = ACTIONS(6985),
    [anon_sym_POUNDelse] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(5326)] = {
    [sym_block_comment] = STATE(5326),
    [sym_compiler_directive_decl] = STATE(5326),
    [sym_fsi_directive_decl] = STATE(5326),
    [sym_preproc_line] = STATE(5326),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [anon_sym_POUNDendif] = ACTIONS(6989),
    [anon_sym_POUNDelse] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(5327)] = {
    [sym_block_comment] = STATE(5327),
    [sym_compiler_directive_decl] = STATE(5327),
    [sym_fsi_directive_decl] = STATE(5327),
    [sym_preproc_line] = STATE(5327),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [anon_sym_POUNDendif] = ACTIONS(6993),
    [anon_sym_POUNDelse] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(5328)] = {
    [sym_block_comment] = STATE(5328),
    [sym_compiler_directive_decl] = STATE(5328),
    [sym_fsi_directive_decl] = STATE(5328),
    [sym_preproc_line] = STATE(5328),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [anon_sym_POUNDendif] = ACTIONS(7055),
    [anon_sym_POUNDelse] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(5329)] = {
    [sym_block_comment] = STATE(5329),
    [sym_compiler_directive_decl] = STATE(5329),
    [sym_fsi_directive_decl] = STATE(5329),
    [sym_preproc_line] = STATE(5329),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5330)] = {
    [sym_block_comment] = STATE(5330),
    [sym_compiler_directive_decl] = STATE(5330),
    [sym_fsi_directive_decl] = STATE(5330),
    [sym_preproc_line] = STATE(5330),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [anon_sym_POUNDendif] = ACTIONS(6997),
    [anon_sym_POUNDelse] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(5331)] = {
    [sym_block_comment] = STATE(5331),
    [sym_compiler_directive_decl] = STATE(5331),
    [sym_fsi_directive_decl] = STATE(5331),
    [sym_preproc_line] = STATE(5331),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [anon_sym_POUNDendif] = ACTIONS(7001),
    [anon_sym_POUNDelse] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(5332)] = {
    [sym_block_comment] = STATE(5332),
    [sym_compiler_directive_decl] = STATE(5332),
    [sym_fsi_directive_decl] = STATE(5332),
    [sym_preproc_line] = STATE(5332),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9694),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__quoted_close] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5333)] = {
    [sym_block_comment] = STATE(5333),
    [sym_compiler_directive_decl] = STATE(5333),
    [sym_fsi_directive_decl] = STATE(5333),
    [sym_preproc_line] = STATE(5333),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [anon_sym_POUNDendif] = ACTIONS(7005),
    [anon_sym_POUNDelse] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(5334)] = {
    [sym_block_comment] = STATE(5334),
    [sym_compiler_directive_decl] = STATE(5334),
    [sym_fsi_directive_decl] = STATE(5334),
    [sym_preproc_line] = STATE(5334),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [anon_sym_POUNDendif] = ACTIONS(7009),
    [anon_sym_POUNDelse] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(5335)] = {
    [sym_block_comment] = STATE(5335),
    [sym_compiler_directive_decl] = STATE(5335),
    [sym_fsi_directive_decl] = STATE(5335),
    [sym_preproc_line] = STATE(5335),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_PIPE] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6333),
    [anon_sym_CARET] = ACTIONS(6333),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6333),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_AMP_AMP] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6333),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6333),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6333),
    [anon_sym_DASH_DOT] = ACTIONS(6333),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6333),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6333),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [anon_sym_POUNDelse] = ACTIONS(6333),
    [sym__newline] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
  },
  [STATE(5336)] = {
    [sym_block_comment] = STATE(5336),
    [sym_compiler_directive_decl] = STATE(5336),
    [sym_fsi_directive_decl] = STATE(5336),
    [sym_preproc_line] = STATE(5336),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [anon_sym_POUNDendif] = ACTIONS(6886),
    [anon_sym_POUNDelse] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(5337)] = {
    [sym_block_comment] = STATE(5337),
    [sym_compiler_directive_decl] = STATE(5337),
    [sym_fsi_directive_decl] = STATE(5337),
    [sym_preproc_line] = STATE(5337),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [anon_sym_POUNDendif] = ACTIONS(7017),
    [anon_sym_POUNDelse] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(5338)] = {
    [sym_block_comment] = STATE(5338),
    [sym_compiler_directive_decl] = STATE(5338),
    [sym_fsi_directive_decl] = STATE(5338),
    [sym_preproc_line] = STATE(5338),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [anon_sym_POUNDendif] = ACTIONS(7021),
    [anon_sym_POUNDelse] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(5339)] = {
    [sym_block_comment] = STATE(5339),
    [sym_compiler_directive_decl] = STATE(5339),
    [sym_fsi_directive_decl] = STATE(5339),
    [sym_preproc_line] = STATE(5339),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [anon_sym_POUNDendif] = ACTIONS(7071),
    [anon_sym_POUNDelse] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(5340)] = {
    [sym_block_comment] = STATE(5340),
    [sym_compiler_directive_decl] = STATE(5340),
    [sym_fsi_directive_decl] = STATE(5340),
    [sym_preproc_line] = STATE(5340),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9696),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5341)] = {
    [sym_block_comment] = STATE(5341),
    [sym_compiler_directive_decl] = STATE(5341),
    [sym_fsi_directive_decl] = STATE(5341),
    [sym_preproc_line] = STATE(5341),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [anon_sym_POUNDendif] = ACTIONS(7025),
    [anon_sym_POUNDelse] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(5342)] = {
    [sym_block_comment] = STATE(5342),
    [sym_compiler_directive_decl] = STATE(5342),
    [sym_fsi_directive_decl] = STATE(5342),
    [sym_preproc_line] = STATE(5342),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [anon_sym_POUNDendif] = ACTIONS(7033),
    [anon_sym_POUNDelse] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(5343)] = {
    [sym_block_comment] = STATE(5343),
    [sym_compiler_directive_decl] = STATE(5343),
    [sym_fsi_directive_decl] = STATE(5343),
    [sym_preproc_line] = STATE(5343),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_as] = ACTIONS(7015),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_with] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(5344)] = {
    [sym_block_comment] = STATE(5344),
    [sym_compiler_directive_decl] = STATE(5344),
    [sym_fsi_directive_decl] = STATE(5344),
    [sym_preproc_line] = STATE(5344),
    [aux_sym_long_identifier_repeat1] = STATE(5350),
    [ts_builtin_sym_end] = ACTIONS(6358),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_namespace] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_PIPE] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6358),
    [anon_sym_CARET] = ACTIONS(6358),
    [anon_sym_DOT] = ACTIONS(9628),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6358),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_AMP_AMP] = ACTIONS(6358),
    [anon_sym_STAR] = ACTIONS(6358),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6358),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6358),
    [anon_sym_DASH_DOT] = ACTIONS(6358),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6358),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6358),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
  },
  [STATE(5345)] = {
    [sym_block_comment] = STATE(5345),
    [sym_compiler_directive_decl] = STATE(5345),
    [sym_fsi_directive_decl] = STATE(5345),
    [sym_preproc_line] = STATE(5345),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [anon_sym_POUNDendif] = ACTIONS(7037),
    [anon_sym_POUNDelse] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(5346)] = {
    [sym_block_comment] = STATE(5346),
    [sym_compiler_directive_decl] = STATE(5346),
    [sym_fsi_directive_decl] = STATE(5346),
    [sym_preproc_line] = STATE(5346),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [anon_sym_POUNDendif] = ACTIONS(7041),
    [anon_sym_POUNDelse] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(5347)] = {
    [sym_block_comment] = STATE(5347),
    [sym_compiler_directive_decl] = STATE(5347),
    [sym_fsi_directive_decl] = STATE(5347),
    [sym_preproc_line] = STATE(5347),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_as] = ACTIONS(7031),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_with] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(5348)] = {
    [sym_block_comment] = STATE(5348),
    [sym_compiler_directive_decl] = STATE(5348),
    [sym_fsi_directive_decl] = STATE(5348),
    [sym_preproc_line] = STATE(5348),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [anon_sym_POUNDendif] = ACTIONS(7075),
    [anon_sym_POUNDelse] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(5349)] = {
    [sym_block_comment] = STATE(5349),
    [sym_compiler_directive_decl] = STATE(5349),
    [sym_fsi_directive_decl] = STATE(5349),
    [sym_preproc_line] = STATE(5349),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [anon_sym_POUNDendif] = ACTIONS(7057),
    [anon_sym_POUNDelse] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(5350)] = {
    [sym_block_comment] = STATE(5350),
    [sym_compiler_directive_decl] = STATE(5350),
    [sym_fsi_directive_decl] = STATE(5350),
    [sym_preproc_line] = STATE(5350),
    [aux_sym_long_identifier_repeat1] = STATE(5350),
    [ts_builtin_sym_end] = ACTIONS(6297),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_namespace] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_PIPE] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6297),
    [anon_sym_CARET] = ACTIONS(6297),
    [anon_sym_DOT] = ACTIONS(9698),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6297),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_AMP_AMP] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6297),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6297),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6297),
    [anon_sym_DASH_DOT] = ACTIONS(6297),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6297),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6297),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
  },
  [STATE(5351)] = {
    [sym_block_comment] = STATE(5351),
    [sym_compiler_directive_decl] = STATE(5351),
    [sym_fsi_directive_decl] = STATE(5351),
    [sym_preproc_line] = STATE(5351),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_as] = ACTIONS(7047),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_with] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(5352)] = {
    [sym_block_comment] = STATE(5352),
    [sym_compiler_directive_decl] = STATE(5352),
    [sym_fsi_directive_decl] = STATE(5352),
    [sym_preproc_line] = STATE(5352),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [anon_sym_POUNDendif] = ACTIONS(7079),
    [anon_sym_POUNDelse] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(5353)] = {
    [sym_block_comment] = STATE(5353),
    [sym_compiler_directive_decl] = STATE(5353),
    [sym_fsi_directive_decl] = STATE(5353),
    [sym_preproc_line] = STATE(5353),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__dedent] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
    [sym__in] = ACTIONS(7055),
  },
  [STATE(5354)] = {
    [sym_block_comment] = STATE(5354),
    [sym_compiler_directive_decl] = STATE(5354),
    [sym_fsi_directive_decl] = STATE(5354),
    [sym_preproc_line] = STATE(5354),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_GT] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5355)] = {
    [sym_block_comment] = STATE(5355),
    [sym_compiler_directive_decl] = STATE(5355),
    [sym_fsi_directive_decl] = STATE(5355),
    [sym_preproc_line] = STATE(5355),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__dedent] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
    [sym__in] = ACTIONS(6886),
  },
  [STATE(5356)] = {
    [sym_block_comment] = STATE(5356),
    [sym_compiler_directive_decl] = STATE(5356),
    [sym_fsi_directive_decl] = STATE(5356),
    [sym_preproc_line] = STATE(5356),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__dedent] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
    [sym__in] = ACTIONS(7071),
  },
  [STATE(5357)] = {
    [sym_block_comment] = STATE(5357),
    [sym_compiler_directive_decl] = STATE(5357),
    [sym_fsi_directive_decl] = STATE(5357),
    [sym_preproc_line] = STATE(5357),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__dedent] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
    [sym__in] = ACTIONS(7075),
  },
  [STATE(5358)] = {
    [sym_block_comment] = STATE(5358),
    [sym_compiler_directive_decl] = STATE(5358),
    [sym_fsi_directive_decl] = STATE(5358),
    [sym_preproc_line] = STATE(5358),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__untyped_quoted_close] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5359)] = {
    [sym_block_comment] = STATE(5359),
    [sym_compiler_directive_decl] = STATE(5359),
    [sym_fsi_directive_decl] = STATE(5359),
    [sym_preproc_line] = STATE(5359),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_as] = ACTIONS(7141),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_with] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(5360)] = {
    [sym_block_comment] = STATE(5360),
    [sym_compiler_directive_decl] = STATE(5360),
    [sym_fsi_directive_decl] = STATE(5360),
    [sym_preproc_line] = STATE(5360),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__dedent] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
    [sym__in] = ACTIONS(7079),
  },
  [STATE(5361)] = {
    [sym_block_comment] = STATE(5361),
    [sym_compiler_directive_decl] = STATE(5361),
    [sym_fsi_directive_decl] = STATE(5361),
    [sym_preproc_line] = STATE(5361),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__dedent] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
    [sym__in] = ACTIONS(7087),
  },
  [STATE(5362)] = {
    [sym_block_comment] = STATE(5362),
    [sym_compiler_directive_decl] = STATE(5362),
    [sym_fsi_directive_decl] = STATE(5362),
    [sym_preproc_line] = STATE(5362),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__dedent] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
    [sym__in] = ACTIONS(7091),
  },
  [STATE(5363)] = {
    [sym_block_comment] = STATE(5363),
    [sym_compiler_directive_decl] = STATE(5363),
    [sym_fsi_directive_decl] = STATE(5363),
    [sym_preproc_line] = STATE(5363),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__dedent] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
    [sym__in] = ACTIONS(7095),
  },
  [STATE(5364)] = {
    [sym_block_comment] = STATE(5364),
    [sym_compiler_directive_decl] = STATE(5364),
    [sym_fsi_directive_decl] = STATE(5364),
    [sym_preproc_line] = STATE(5364),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__dedent] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
    [sym__in] = ACTIONS(7099),
  },
  [STATE(5365)] = {
    [sym_block_comment] = STATE(5365),
    [sym_compiler_directive_decl] = STATE(5365),
    [sym_fsi_directive_decl] = STATE(5365),
    [sym_preproc_line] = STATE(5365),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__dedent] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
    [sym__in] = ACTIONS(7103),
  },
  [STATE(5366)] = {
    [sym_block_comment] = STATE(5366),
    [sym_compiler_directive_decl] = STATE(5366),
    [sym_fsi_directive_decl] = STATE(5366),
    [sym_preproc_line] = STATE(5366),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__dedent] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
    [sym__in] = ACTIONS(7107),
  },
  [STATE(5367)] = {
    [sym_block_comment] = STATE(5367),
    [sym_compiler_directive_decl] = STATE(5367),
    [sym_fsi_directive_decl] = STATE(5367),
    [sym_preproc_line] = STATE(5367),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__dedent] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
    [sym__in] = ACTIONS(7111),
  },
  [STATE(5368)] = {
    [sym_block_comment] = STATE(5368),
    [sym_compiler_directive_decl] = STATE(5368),
    [sym_fsi_directive_decl] = STATE(5368),
    [sym_preproc_line] = STATE(5368),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__dedent] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
    [sym__in] = ACTIONS(7115),
  },
  [STATE(5369)] = {
    [sym_block_comment] = STATE(5369),
    [sym_compiler_directive_decl] = STATE(5369),
    [sym_fsi_directive_decl] = STATE(5369),
    [sym_preproc_line] = STATE(5369),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__dedent] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
    [sym__in] = ACTIONS(7119),
  },
  [STATE(5370)] = {
    [sym_block_comment] = STATE(5370),
    [sym_compiler_directive_decl] = STATE(5370),
    [sym_fsi_directive_decl] = STATE(5370),
    [sym_preproc_line] = STATE(5370),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__dedent] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
    [sym__in] = ACTIONS(7123),
  },
  [STATE(5371)] = {
    [sym_block_comment] = STATE(5371),
    [sym_compiler_directive_decl] = STATE(5371),
    [sym_fsi_directive_decl] = STATE(5371),
    [sym_preproc_line] = STATE(5371),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__dedent] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
    [sym__in] = ACTIONS(6934),
  },
  [STATE(5372)] = {
    [sym_block_comment] = STATE(5372),
    [sym_compiler_directive_decl] = STATE(5372),
    [sym_fsi_directive_decl] = STATE(5372),
    [sym_preproc_line] = STATE(5372),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__dedent] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
    [sym__in] = ACTIONS(6938),
  },
  [STATE(5373)] = {
    [sym_block_comment] = STATE(5373),
    [sym_compiler_directive_decl] = STATE(5373),
    [sym_fsi_directive_decl] = STATE(5373),
    [sym_preproc_line] = STATE(5373),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__dedent] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
    [sym__in] = ACTIONS(7013),
  },
  [STATE(5374)] = {
    [sym_block_comment] = STATE(5374),
    [sym_compiler_directive_decl] = STATE(5374),
    [sym_fsi_directive_decl] = STATE(5374),
    [sym_preproc_line] = STATE(5374),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__dedent] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
    [sym__in] = ACTIONS(7029),
  },
  [STATE(5375)] = {
    [sym_block_comment] = STATE(5375),
    [sym_compiler_directive_decl] = STATE(5375),
    [sym_fsi_directive_decl] = STATE(5375),
    [sym_preproc_line] = STATE(5375),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__dedent] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
    [sym__in] = ACTIONS(7045),
  },
  [STATE(5376)] = {
    [sym_block_comment] = STATE(5376),
    [sym_compiler_directive_decl] = STATE(5376),
    [sym_fsi_directive_decl] = STATE(5376),
    [sym_preproc_line] = STATE(5376),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__dedent] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
    [sym__in] = ACTIONS(7049),
  },
  [STATE(5377)] = {
    [sym_block_comment] = STATE(5377),
    [sym_compiler_directive_decl] = STATE(5377),
    [sym_fsi_directive_decl] = STATE(5377),
    [sym_preproc_line] = STATE(5377),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [anon_sym_POUNDendif] = ACTIONS(7087),
    [anon_sym_POUNDelse] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(5378)] = {
    [sym_block_comment] = STATE(5378),
    [sym_compiler_directive_decl] = STATE(5378),
    [sym_fsi_directive_decl] = STATE(5378),
    [sym_preproc_line] = STATE(5378),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_as] = ACTIONS(7145),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_with] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(5379)] = {
    [sym_block_comment] = STATE(5379),
    [sym_compiler_directive_decl] = STATE(5379),
    [sym_fsi_directive_decl] = STATE(5379),
    [sym_preproc_line] = STATE(5379),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__dedent] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
    [sym__in] = ACTIONS(7061),
  },
  [STATE(5380)] = {
    [sym_block_comment] = STATE(5380),
    [sym_compiler_directive_decl] = STATE(5380),
    [sym_fsi_directive_decl] = STATE(5380),
    [sym_preproc_line] = STATE(5380),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [anon_sym_POUNDendif] = ACTIONS(7091),
    [anon_sym_POUNDelse] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(5381)] = {
    [sym_block_comment] = STATE(5381),
    [sym_compiler_directive_decl] = STATE(5381),
    [sym_fsi_directive_decl] = STATE(5381),
    [sym_preproc_line] = STATE(5381),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_as] = ACTIONS(7051),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_with] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(5382)] = {
    [sym_block_comment] = STATE(5382),
    [sym_compiler_directive_decl] = STATE(5382),
    [sym_fsi_directive_decl] = STATE(5382),
    [sym_preproc_line] = STATE(5382),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [anon_sym_POUNDendif] = ACTIONS(7095),
    [anon_sym_POUNDelse] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(5383)] = {
    [sym_block_comment] = STATE(5383),
    [sym_compiler_directive_decl] = STATE(5383),
    [sym_fsi_directive_decl] = STATE(5383),
    [sym_preproc_line] = STATE(5383),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__dedent] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
    [sym__in] = ACTIONS(7065),
  },
  [STATE(5384)] = {
    [sym_block_comment] = STATE(5384),
    [sym_compiler_directive_decl] = STATE(5384),
    [sym_fsi_directive_decl] = STATE(5384),
    [sym_preproc_line] = STATE(5384),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__dedent] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
    [sym__in] = ACTIONS(7127),
  },
  [STATE(5385)] = {
    [sym_block_comment] = STATE(5385),
    [sym_compiler_directive_decl] = STATE(5385),
    [sym_fsi_directive_decl] = STATE(5385),
    [sym_preproc_line] = STATE(5385),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__dedent] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
    [sym__in] = ACTIONS(7131),
  },
  [STATE(5386)] = {
    [sym_block_comment] = STATE(5386),
    [sym_compiler_directive_decl] = STATE(5386),
    [sym_fsi_directive_decl] = STATE(5386),
    [sym_preproc_line] = STATE(5386),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__dedent] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
    [sym__in] = ACTIONS(7135),
  },
  [STATE(5387)] = {
    [sym_block_comment] = STATE(5387),
    [sym_compiler_directive_decl] = STATE(5387),
    [sym_fsi_directive_decl] = STATE(5387),
    [sym_preproc_line] = STATE(5387),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [anon_sym_POUNDendif] = ACTIONS(7099),
    [anon_sym_POUNDelse] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(5388)] = {
    [sym_block_comment] = STATE(5388),
    [sym_compiler_directive_decl] = STATE(5388),
    [sym_fsi_directive_decl] = STATE(5388),
    [sym_preproc_line] = STATE(5388),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [anon_sym_POUNDendif] = ACTIONS(7103),
    [anon_sym_POUNDelse] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(5389)] = {
    [sym_block_comment] = STATE(5389),
    [sym_compiler_directive_decl] = STATE(5389),
    [sym_fsi_directive_decl] = STATE(5389),
    [sym_preproc_line] = STATE(5389),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [anon_sym_POUNDendif] = ACTIONS(7107),
    [anon_sym_POUNDelse] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(5390)] = {
    [sym_block_comment] = STATE(5390),
    [sym_compiler_directive_decl] = STATE(5390),
    [sym_fsi_directive_decl] = STATE(5390),
    [sym_preproc_line] = STATE(5390),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__quoted_close] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5391)] = {
    [sym_block_comment] = STATE(5391),
    [sym_compiler_directive_decl] = STATE(5391),
    [sym_fsi_directive_decl] = STATE(5391),
    [sym_preproc_line] = STATE(5391),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [anon_sym_POUNDendif] = ACTIONS(7111),
    [anon_sym_POUNDelse] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(5392)] = {
    [sym_block_comment] = STATE(5392),
    [sym_compiler_directive_decl] = STATE(5392),
    [sym_fsi_directive_decl] = STATE(5392),
    [sym_preproc_line] = STATE(5392),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__untyped_quoted_close] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5393)] = {
    [sym_block_comment] = STATE(5393),
    [sym_compiler_directive_decl] = STATE(5393),
    [sym_fsi_directive_decl] = STATE(5393),
    [sym_preproc_line] = STATE(5393),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [anon_sym_POUNDendif] = ACTIONS(7115),
    [anon_sym_POUNDelse] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(5394)] = {
    [sym_block_comment] = STATE(5394),
    [sym_compiler_directive_decl] = STATE(5394),
    [sym_fsi_directive_decl] = STATE(5394),
    [sym_preproc_line] = STATE(5394),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [anon_sym_POUNDendif] = ACTIONS(7119),
    [anon_sym_POUNDelse] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(5395)] = {
    [sym_block_comment] = STATE(5395),
    [sym_compiler_directive_decl] = STATE(5395),
    [sym_fsi_directive_decl] = STATE(5395),
    [sym_preproc_line] = STATE(5395),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__dedent] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
    [sym__in] = ACTIONS(7139),
  },
  [STATE(5396)] = {
    [sym_block_comment] = STATE(5396),
    [sym_compiler_directive_decl] = STATE(5396),
    [sym_fsi_directive_decl] = STATE(5396),
    [sym_preproc_line] = STATE(5396),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [anon_sym_POUNDendif] = ACTIONS(7123),
    [anon_sym_POUNDelse] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(5397)] = {
    [sym_block_comment] = STATE(5397),
    [sym_compiler_directive_decl] = STATE(5397),
    [sym_fsi_directive_decl] = STATE(5397),
    [sym_preproc_line] = STATE(5397),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [anon_sym_POUNDendif] = ACTIONS(6934),
    [anon_sym_POUNDelse] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(5398)] = {
    [sym_block_comment] = STATE(5398),
    [sym_compiler_directive_decl] = STATE(5398),
    [sym_fsi_directive_decl] = STATE(5398),
    [sym_preproc_line] = STATE(5398),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [anon_sym_POUNDendif] = ACTIONS(6938),
    [anon_sym_POUNDelse] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(5399)] = {
    [sym_block_comment] = STATE(5399),
    [sym_compiler_directive_decl] = STATE(5399),
    [sym_fsi_directive_decl] = STATE(5399),
    [sym_preproc_line] = STATE(5399),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [anon_sym_POUNDendif] = ACTIONS(7013),
    [anon_sym_POUNDelse] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(5400)] = {
    [sym_block_comment] = STATE(5400),
    [sym_compiler_directive_decl] = STATE(5400),
    [sym_fsi_directive_decl] = STATE(5400),
    [sym_preproc_line] = STATE(5400),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [anon_sym_POUNDendif] = ACTIONS(7029),
    [anon_sym_POUNDelse] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(5401)] = {
    [sym_block_comment] = STATE(5401),
    [sym_compiler_directive_decl] = STATE(5401),
    [sym_fsi_directive_decl] = STATE(5401),
    [sym_preproc_line] = STATE(5401),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__untyped_quoted_close] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5402)] = {
    [sym_block_comment] = STATE(5402),
    [sym_compiler_directive_decl] = STATE(5402),
    [sym_fsi_directive_decl] = STATE(5402),
    [sym_preproc_line] = STATE(5402),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__untyped_quoted_close] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5403)] = {
    [sym_block_comment] = STATE(5403),
    [sym_compiler_directive_decl] = STATE(5403),
    [sym_fsi_directive_decl] = STATE(5403),
    [sym_preproc_line] = STATE(5403),
    [aux_sym_sequential_expression_repeat1] = STATE(5403),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_DOT_DOT] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9701),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5404)] = {
    [sym_block_comment] = STATE(5404),
    [sym_compiler_directive_decl] = STATE(5404),
    [sym_fsi_directive_decl] = STATE(5404),
    [sym_preproc_line] = STATE(5404),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__quoted_close] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5405)] = {
    [sym_block_comment] = STATE(5405),
    [sym_compiler_directive_decl] = STATE(5405),
    [sym_fsi_directive_decl] = STATE(5405),
    [sym_preproc_line] = STATE(5405),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [anon_sym_POUNDendif] = ACTIONS(7045),
    [anon_sym_POUNDelse] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(5406)] = {
    [sym_block_comment] = STATE(5406),
    [sym_compiler_directive_decl] = STATE(5406),
    [sym_fsi_directive_decl] = STATE(5406),
    [sym_preproc_line] = STATE(5406),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [anon_sym_POUNDendif] = ACTIONS(7049),
    [anon_sym_POUNDelse] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(5407)] = {
    [sym_block_comment] = STATE(5407),
    [sym_compiler_directive_decl] = STATE(5407),
    [sym_fsi_directive_decl] = STATE(5407),
    [sym_preproc_line] = STATE(5407),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9704),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__quoted_close] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5408)] = {
    [sym_block_comment] = STATE(5408),
    [sym_compiler_directive_decl] = STATE(5408),
    [sym_fsi_directive_decl] = STATE(5408),
    [sym_preproc_line] = STATE(5408),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__untyped_quoted_close] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5409)] = {
    [sym_block_comment] = STATE(5409),
    [sym_compiler_directive_decl] = STATE(5409),
    [sym_fsi_directive_decl] = STATE(5409),
    [sym_preproc_line] = STATE(5409),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_GT] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5410)] = {
    [sym_block_comment] = STATE(5410),
    [sym_compiler_directive_decl] = STATE(5410),
    [sym_fsi_directive_decl] = STATE(5410),
    [sym_preproc_line] = STATE(5410),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [anon_sym_POUNDendif] = ACTIONS(7061),
    [anon_sym_POUNDelse] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(5411)] = {
    [sym_block_comment] = STATE(5411),
    [sym_compiler_directive_decl] = STATE(5411),
    [sym_fsi_directive_decl] = STATE(5411),
    [sym_preproc_line] = STATE(5411),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(9706),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5412)] = {
    [sym_block_comment] = STATE(5412),
    [sym_compiler_directive_decl] = STATE(5412),
    [sym_fsi_directive_decl] = STATE(5412),
    [sym_preproc_line] = STATE(5412),
    [aux_sym_type_argument_repeat1] = STATE(5421),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_PIPE] = ACTIONS(6306),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6306),
    [anon_sym_CARET] = ACTIONS(6306),
    [anon_sym_or] = ACTIONS(9708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_with] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6306),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_AMP_AMP] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6306),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6306),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6306),
    [anon_sym_DASH_DOT] = ACTIONS(6306),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6306),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6306),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
  },
  [STATE(5413)] = {
    [sym_block_comment] = STATE(5413),
    [sym_compiler_directive_decl] = STATE(5413),
    [sym_fsi_directive_decl] = STATE(5413),
    [sym_preproc_line] = STATE(5413),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__untyped_quoted_close] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5414)] = {
    [sym_block_comment] = STATE(5414),
    [sym_compiler_directive_decl] = STATE(5414),
    [sym_fsi_directive_decl] = STATE(5414),
    [sym_preproc_line] = STATE(5414),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__untyped_quoted_close] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5415)] = {
    [sym_block_comment] = STATE(5415),
    [sym_compiler_directive_decl] = STATE(5415),
    [sym_fsi_directive_decl] = STATE(5415),
    [sym_preproc_line] = STATE(5415),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [anon_sym_POUNDendif] = ACTIONS(7065),
    [anon_sym_POUNDelse] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(5416)] = {
    [sym_block_comment] = STATE(5416),
    [sym_compiler_directive_decl] = STATE(5416),
    [sym_fsi_directive_decl] = STATE(5416),
    [sym_preproc_line] = STATE(5416),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_as] = ACTIONS(7155),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_with] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(5417)] = {
    [sym_block_comment] = STATE(5417),
    [sym_compiler_directive_decl] = STATE(5417),
    [sym_fsi_directive_decl] = STATE(5417),
    [sym_preproc_line] = STATE(5417),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [anon_sym_POUNDendif] = ACTIONS(7127),
    [anon_sym_POUNDelse] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(5418)] = {
    [sym_block_comment] = STATE(5418),
    [sym_compiler_directive_decl] = STATE(5418),
    [sym_fsi_directive_decl] = STATE(5418),
    [sym_preproc_line] = STATE(5418),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__untyped_quoted_close] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5419)] = {
    [sym_block_comment] = STATE(5419),
    [sym_compiler_directive_decl] = STATE(5419),
    [sym_fsi_directive_decl] = STATE(5419),
    [sym_preproc_line] = STATE(5419),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_as] = ACTIONS(7159),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_with] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(5420)] = {
    [sym_block_comment] = STATE(5420),
    [sym_compiler_directive_decl] = STATE(5420),
    [sym_fsi_directive_decl] = STATE(5420),
    [sym_preproc_line] = STATE(5420),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [anon_sym_POUNDendif] = ACTIONS(7131),
    [anon_sym_POUNDelse] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(5421)] = {
    [sym_block_comment] = STATE(5421),
    [sym_compiler_directive_decl] = STATE(5421),
    [sym_fsi_directive_decl] = STATE(5421),
    [sym_preproc_line] = STATE(5421),
    [aux_sym_type_argument_repeat1] = STATE(5421),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(9710),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5422)] = {
    [sym_block_comment] = STATE(5422),
    [sym_compiler_directive_decl] = STATE(5422),
    [sym_fsi_directive_decl] = STATE(5422),
    [sym_preproc_line] = STATE(5422),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [anon_sym_POUNDendif] = ACTIONS(7135),
    [anon_sym_POUNDelse] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(5423)] = {
    [sym_block_comment] = STATE(5423),
    [sym_compiler_directive_decl] = STATE(5423),
    [sym_fsi_directive_decl] = STATE(5423),
    [sym_preproc_line] = STATE(5423),
    [aux_sym_sequential_expression_repeat1] = STATE(5423),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9713),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__multi_dollar_interp_end] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5424)] = {
    [sym_block_comment] = STATE(5424),
    [sym_compiler_directive_decl] = STATE(5424),
    [sym_fsi_directive_decl] = STATE(5424),
    [sym_preproc_line] = STATE(5424),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_DASH_GT] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_DOT_DOT] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(5425)] = {
    [sym_block_comment] = STATE(5425),
    [sym_compiler_directive_decl] = STATE(5425),
    [sym_fsi_directive_decl] = STATE(5425),
    [sym_preproc_line] = STATE(5425),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_DASH_GT] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5426)] = {
    [sym_block_comment] = STATE(5426),
    [sym_compiler_directive_decl] = STATE(5426),
    [sym_fsi_directive_decl] = STATE(5426),
    [sym_preproc_line] = STATE(5426),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_DOT_DOT] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__dedent] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(5427)] = {
    [sym_block_comment] = STATE(5427),
    [sym_compiler_directive_decl] = STATE(5427),
    [sym_fsi_directive_decl] = STATE(5427),
    [sym_preproc_line] = STATE(5427),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_DASH_GT] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_DOT_DOT] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(5428)] = {
    [sym_block_comment] = STATE(5428),
    [sym_compiler_directive_decl] = STATE(5428),
    [sym_fsi_directive_decl] = STATE(5428),
    [sym_preproc_line] = STATE(5428),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_DASH_GT] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_DOT_DOT] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(5429)] = {
    [sym_block_comment] = STATE(5429),
    [sym_compiler_directive_decl] = STATE(5429),
    [sym_fsi_directive_decl] = STATE(5429),
    [sym_preproc_line] = STATE(5429),
    [aux_sym_sequential_expression_repeat1] = STATE(5766),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__dedent] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5430)] = {
    [sym_block_comment] = STATE(5430),
    [sym_compiler_directive_decl] = STATE(5430),
    [sym_fsi_directive_decl] = STATE(5430),
    [sym_preproc_line] = STATE(5430),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_DASH_GT] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_DOT_DOT] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(5431)] = {
    [sym_block_comment] = STATE(5431),
    [sym_compiler_directive_decl] = STATE(5431),
    [sym_fsi_directive_decl] = STATE(5431),
    [sym_preproc_line] = STATE(5431),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_DOT_DOT] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__dedent] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5432)] = {
    [sym_block_comment] = STATE(5432),
    [sym_compiler_directive_decl] = STATE(5432),
    [sym_fsi_directive_decl] = STATE(5432),
    [sym_preproc_line] = STATE(5432),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_as] = ACTIONS(7063),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_with] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(5433)] = {
    [sym_block_comment] = STATE(5433),
    [sym_compiler_directive_decl] = STATE(5433),
    [sym_fsi_directive_decl] = STATE(5433),
    [sym_preproc_line] = STATE(5433),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_DASH_GT] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_DOT_DOT] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(5434)] = {
    [sym_block_comment] = STATE(5434),
    [sym_compiler_directive_decl] = STATE(5434),
    [sym_fsi_directive_decl] = STATE(5434),
    [sym_preproc_line] = STATE(5434),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_DASH_GT] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_DOT_DOT] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(5435)] = {
    [sym_block_comment] = STATE(5435),
    [sym_compiler_directive_decl] = STATE(5435),
    [sym_fsi_directive_decl] = STATE(5435),
    [sym_preproc_line] = STATE(5435),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_DASH_GT] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_DOT_DOT] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(5436)] = {
    [sym_block_comment] = STATE(5436),
    [sym_compiler_directive_decl] = STATE(5436),
    [sym_fsi_directive_decl] = STATE(5436),
    [sym_preproc_line] = STATE(5436),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_DASH_GT] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_DOT_DOT] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(5437)] = {
    [sym_block_comment] = STATE(5437),
    [sym_compiler_directive_decl] = STATE(5437),
    [sym_fsi_directive_decl] = STATE(5437),
    [sym_preproc_line] = STATE(5437),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_DASH_GT] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_DOT_DOT] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(5438)] = {
    [sym_block_comment] = STATE(5438),
    [sym_compiler_directive_decl] = STATE(5438),
    [sym_fsi_directive_decl] = STATE(5438),
    [sym_preproc_line] = STATE(5438),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_DASH_GT] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_DOT_DOT] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(5439)] = {
    [sym_block_comment] = STATE(5439),
    [sym_compiler_directive_decl] = STATE(5439),
    [sym_fsi_directive_decl] = STATE(5439),
    [sym_preproc_line] = STATE(5439),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_DASH_GT] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_DOT_DOT] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(5440)] = {
    [sym_block_comment] = STATE(5440),
    [sym_compiler_directive_decl] = STATE(5440),
    [sym_fsi_directive_decl] = STATE(5440),
    [sym_preproc_line] = STATE(5440),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_DASH_GT] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_DOT_DOT] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(5441)] = {
    [sym_block_comment] = STATE(5441),
    [sym_compiler_directive_decl] = STATE(5441),
    [sym_fsi_directive_decl] = STATE(5441),
    [sym_preproc_line] = STATE(5441),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_DASH_GT] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_DOT_DOT] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(5442)] = {
    [sym_block_comment] = STATE(5442),
    [sym_compiler_directive_decl] = STATE(5442),
    [sym_fsi_directive_decl] = STATE(5442),
    [sym_preproc_line] = STATE(5442),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_DASH_GT] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_DOT_DOT] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(5443)] = {
    [sym_block_comment] = STATE(5443),
    [sym_compiler_directive_decl] = STATE(5443),
    [sym_fsi_directive_decl] = STATE(5443),
    [sym_preproc_line] = STATE(5443),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_DASH_GT] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_DOT_DOT] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(5444)] = {
    [sym_block_comment] = STATE(5444),
    [sym_compiler_directive_decl] = STATE(5444),
    [sym_fsi_directive_decl] = STATE(5444),
    [sym_preproc_line] = STATE(5444),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_DASH_GT] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_DOT_DOT] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(5445)] = {
    [sym_block_comment] = STATE(5445),
    [sym_compiler_directive_decl] = STATE(5445),
    [sym_fsi_directive_decl] = STATE(5445),
    [sym_preproc_line] = STATE(5445),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_DASH_GT] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_DOT_DOT] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(5446)] = {
    [sym_block_comment] = STATE(5446),
    [sym_compiler_directive_decl] = STATE(5446),
    [sym_fsi_directive_decl] = STATE(5446),
    [sym_preproc_line] = STATE(5446),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_DASH_GT] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_DOT_DOT] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(5447)] = {
    [sym_block_comment] = STATE(5447),
    [sym_compiler_directive_decl] = STATE(5447),
    [sym_fsi_directive_decl] = STATE(5447),
    [sym_preproc_line] = STATE(5447),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_DASH_GT] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_DOT_DOT] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(5448)] = {
    [sym_block_comment] = STATE(5448),
    [sym_compiler_directive_decl] = STATE(5448),
    [sym_fsi_directive_decl] = STATE(5448),
    [sym_preproc_line] = STATE(5448),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [anon_sym_POUNDendif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5449)] = {
    [sym_block_comment] = STATE(5449),
    [sym_compiler_directive_decl] = STATE(5449),
    [sym_fsi_directive_decl] = STATE(5449),
    [sym_preproc_line] = STATE(5449),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_DASH_GT] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_DOT_DOT] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(5450)] = {
    [sym_block_comment] = STATE(5450),
    [sym_compiler_directive_decl] = STATE(5450),
    [sym_fsi_directive_decl] = STATE(5450),
    [sym_preproc_line] = STATE(5450),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_DASH_GT] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_DOT_DOT] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(5451)] = {
    [sym_block_comment] = STATE(5451),
    [sym_compiler_directive_decl] = STATE(5451),
    [sym_fsi_directive_decl] = STATE(5451),
    [sym_preproc_line] = STATE(5451),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__untyped_quoted_close] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5452)] = {
    [sym_block_comment] = STATE(5452),
    [sym_compiler_directive_decl] = STATE(5452),
    [sym_fsi_directive_decl] = STATE(5452),
    [sym_preproc_line] = STATE(5452),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__untyped_quoted_close] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5453)] = {
    [sym_block_comment] = STATE(5453),
    [sym_compiler_directive_decl] = STATE(5453),
    [sym_fsi_directive_decl] = STATE(5453),
    [sym_preproc_line] = STATE(5453),
    [sym_identifier] = ACTIONS(6888),
    [anon_sym_null] = ACTIONS(6888),
    [anon_sym_EQ] = ACTIONS(6888),
    [anon_sym_begin] = ACTIONS(6888),
    [anon_sym_COLON] = ACTIONS(6888),
    [anon_sym_do] = ACTIONS(6888),
    [anon_sym_let] = ACTIONS(6888),
    [anon_sym_let_BANG] = ACTIONS(6886),
    [anon_sym_LPAREN] = ACTIONS(6888),
    [anon_sym_COMMA] = ACTIONS(6886),
    [anon_sym_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_QMARK] = ACTIONS(6888),
    [anon_sym_COLON_COLON] = ACTIONS(6886),
    [anon_sym_AMP] = ACTIONS(6888),
    [anon_sym_LBRACK] = ACTIONS(6888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6886),
    [anon_sym_LBRACE] = ACTIONS(6888),
    [anon_sym_LT_AT] = ACTIONS(6888),
    [anon_sym_LT_AT_AT] = ACTIONS(6888),
    [anon_sym_CARET] = ACTIONS(6888),
    [anon_sym_or] = ACTIONS(6888),
    [anon_sym_DOT] = ACTIONS(6888),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6886),
    [anon_sym_new] = ACTIONS(6888),
    [anon_sym_return] = ACTIONS(6888),
    [anon_sym_return_BANG] = ACTIONS(6886),
    [anon_sym_yield] = ACTIONS(6888),
    [anon_sym_yield_BANG] = ACTIONS(6886),
    [anon_sym_lazy] = ACTIONS(6888),
    [anon_sym_assert] = ACTIONS(6888),
    [anon_sym_upcast] = ACTIONS(6888),
    [anon_sym_downcast] = ACTIONS(6888),
    [anon_sym_fixed] = ACTIONS(6888),
    [anon_sym_COLON_GT] = ACTIONS(6886),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6886),
    [anon_sym_for] = ACTIONS(6888),
    [anon_sym_while] = ACTIONS(6888),
    [anon_sym_while_BANG] = ACTIONS(6886),
    [anon_sym_if] = ACTIONS(6888),
    [anon_sym_fun] = ACTIONS(6888),
    [anon_sym_try] = ACTIONS(6888),
    [anon_sym_match] = ACTIONS(6888),
    [anon_sym_match_BANG] = ACTIONS(6886),
    [anon_sym_function] = ACTIONS(6888),
    [anon_sym_DOT_LBRACK] = ACTIONS(6886),
    [anon_sym_use] = ACTIONS(6888),
    [anon_sym_use_BANG] = ACTIONS(6886),
    [anon_sym_do_BANG] = ACTIONS(6886),
    [anon_sym_DOT_DOT] = ACTIONS(6886),
    [anon_sym_LPAREN2] = ACTIONS(6888),
    [anon_sym_PIPE_PIPE] = ACTIONS(6888),
    [anon_sym_AMP_AMP] = ACTIONS(6888),
    [aux_sym_char_token1] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6888),
    [anon_sym_DQUOTE] = ACTIONS(6888),
    [anon_sym_AT_DQUOTE] = ACTIONS(6886),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6886),
    [sym_bool] = ACTIONS(6888),
    [sym_unit] = ACTIONS(6886),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6888),
    [sym_op_identifier] = ACTIONS(6888),
    [anon_sym_PLUS] = ACTIONS(6888),
    [anon_sym_DASH] = ACTIONS(6888),
    [anon_sym_PLUS_DOT] = ACTIONS(6888),
    [anon_sym_DASH_DOT] = ACTIONS(6888),
    [anon_sym_PERCENT] = ACTIONS(6888),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6888),
    [anon_sym_TILDE] = ACTIONS(6886),
    [aux_sym_prefix_op_token1] = ACTIONS(6888),
    [aux_sym_infix_op_token1] = ACTIONS(6886),
    [aux_sym_infix_op_token2] = ACTIONS(6888),
    [aux_sym_infix_op_token3] = ACTIONS(6888),
    [anon_sym_BANG_EQ] = ACTIONS(6888),
    [anon_sym_COLON_EQ] = ACTIONS(6886),
    [anon_sym_DOLLAR] = ACTIONS(6888),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6888),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6888),
    [sym_int] = ACTIONS(6888),
    [sym_xint] = ACTIONS(6886),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6886),
    [sym__newline] = ACTIONS(6886),
    [sym__dedent] = ACTIONS(6886),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6886),
    [sym__tyapp_open] = ACTIONS(6886),
  },
  [STATE(5454)] = {
    [sym_block_comment] = STATE(5454),
    [sym_compiler_directive_decl] = STATE(5454),
    [sym_fsi_directive_decl] = STATE(5454),
    [sym_preproc_line] = STATE(5454),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_DASH_GT] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_DOT_DOT] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(5455)] = {
    [sym_block_comment] = STATE(5455),
    [sym_compiler_directive_decl] = STATE(5455),
    [sym_fsi_directive_decl] = STATE(5455),
    [sym_preproc_line] = STATE(5455),
    [sym_identifier] = ACTIONS(7069),
    [anon_sym_null] = ACTIONS(7069),
    [anon_sym_EQ] = ACTIONS(7069),
    [anon_sym_begin] = ACTIONS(7069),
    [anon_sym_COLON] = ACTIONS(7069),
    [anon_sym_do] = ACTIONS(7069),
    [anon_sym_let] = ACTIONS(7069),
    [anon_sym_let_BANG] = ACTIONS(7071),
    [anon_sym_LPAREN] = ACTIONS(7069),
    [anon_sym_COMMA] = ACTIONS(7071),
    [anon_sym_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_QMARK] = ACTIONS(7069),
    [anon_sym_COLON_COLON] = ACTIONS(7071),
    [anon_sym_AMP] = ACTIONS(7069),
    [anon_sym_LBRACK] = ACTIONS(7069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7071),
    [anon_sym_LBRACE] = ACTIONS(7069),
    [anon_sym_LT_AT] = ACTIONS(7069),
    [anon_sym_LT_AT_AT] = ACTIONS(7069),
    [anon_sym_CARET] = ACTIONS(7069),
    [anon_sym_or] = ACTIONS(7069),
    [anon_sym_DOT] = ACTIONS(7069),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7071),
    [anon_sym_new] = ACTIONS(7069),
    [anon_sym_return] = ACTIONS(7069),
    [anon_sym_return_BANG] = ACTIONS(7071),
    [anon_sym_yield] = ACTIONS(7069),
    [anon_sym_yield_BANG] = ACTIONS(7071),
    [anon_sym_lazy] = ACTIONS(7069),
    [anon_sym_assert] = ACTIONS(7069),
    [anon_sym_upcast] = ACTIONS(7069),
    [anon_sym_downcast] = ACTIONS(7069),
    [anon_sym_fixed] = ACTIONS(7069),
    [anon_sym_COLON_GT] = ACTIONS(7071),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7071),
    [anon_sym_for] = ACTIONS(7069),
    [anon_sym_while] = ACTIONS(7069),
    [anon_sym_while_BANG] = ACTIONS(7071),
    [anon_sym_if] = ACTIONS(7069),
    [anon_sym_fun] = ACTIONS(7069),
    [anon_sym_try] = ACTIONS(7069),
    [anon_sym_match] = ACTIONS(7069),
    [anon_sym_match_BANG] = ACTIONS(7071),
    [anon_sym_function] = ACTIONS(7069),
    [anon_sym_DOT_LBRACK] = ACTIONS(7071),
    [anon_sym_use] = ACTIONS(7069),
    [anon_sym_use_BANG] = ACTIONS(7071),
    [anon_sym_do_BANG] = ACTIONS(7071),
    [anon_sym_DOT_DOT] = ACTIONS(7071),
    [anon_sym_LPAREN2] = ACTIONS(7069),
    [anon_sym_PIPE_PIPE] = ACTIONS(7069),
    [anon_sym_AMP_AMP] = ACTIONS(7069),
    [aux_sym_char_token1] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7069),
    [anon_sym_DQUOTE] = ACTIONS(7069),
    [anon_sym_AT_DQUOTE] = ACTIONS(7071),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7071),
    [sym_bool] = ACTIONS(7069),
    [sym_unit] = ACTIONS(7071),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7069),
    [sym_op_identifier] = ACTIONS(7069),
    [anon_sym_PLUS] = ACTIONS(7069),
    [anon_sym_DASH] = ACTIONS(7069),
    [anon_sym_PLUS_DOT] = ACTIONS(7069),
    [anon_sym_DASH_DOT] = ACTIONS(7069),
    [anon_sym_PERCENT] = ACTIONS(7069),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7069),
    [anon_sym_TILDE] = ACTIONS(7071),
    [aux_sym_prefix_op_token1] = ACTIONS(7069),
    [aux_sym_infix_op_token1] = ACTIONS(7071),
    [aux_sym_infix_op_token2] = ACTIONS(7069),
    [aux_sym_infix_op_token3] = ACTIONS(7069),
    [anon_sym_BANG_EQ] = ACTIONS(7069),
    [anon_sym_COLON_EQ] = ACTIONS(7071),
    [anon_sym_DOLLAR] = ACTIONS(7069),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7069),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7069),
    [sym_int] = ACTIONS(7069),
    [sym_xint] = ACTIONS(7071),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7071),
    [sym__newline] = ACTIONS(7071),
    [sym__dedent] = ACTIONS(7071),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7071),
    [sym__tyapp_open] = ACTIONS(7071),
  },
  [STATE(5456)] = {
    [sym_block_comment] = STATE(5456),
    [sym_compiler_directive_decl] = STATE(5456),
    [sym_fsi_directive_decl] = STATE(5456),
    [sym_preproc_line] = STATE(5456),
    [sym_identifier] = ACTIONS(7073),
    [anon_sym_null] = ACTIONS(7073),
    [anon_sym_EQ] = ACTIONS(7073),
    [anon_sym_begin] = ACTIONS(7073),
    [anon_sym_COLON] = ACTIONS(7073),
    [anon_sym_do] = ACTIONS(7073),
    [anon_sym_let] = ACTIONS(7073),
    [anon_sym_let_BANG] = ACTIONS(7075),
    [anon_sym_LPAREN] = ACTIONS(7073),
    [anon_sym_COMMA] = ACTIONS(7075),
    [anon_sym_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_QMARK] = ACTIONS(7073),
    [anon_sym_COLON_COLON] = ACTIONS(7075),
    [anon_sym_AMP] = ACTIONS(7073),
    [anon_sym_LBRACK] = ACTIONS(7073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7075),
    [anon_sym_LBRACE] = ACTIONS(7073),
    [anon_sym_LT_AT] = ACTIONS(7073),
    [anon_sym_LT_AT_AT] = ACTIONS(7073),
    [anon_sym_CARET] = ACTIONS(7073),
    [anon_sym_or] = ACTIONS(7073),
    [anon_sym_DOT] = ACTIONS(7073),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7075),
    [anon_sym_new] = ACTIONS(7073),
    [anon_sym_return] = ACTIONS(7073),
    [anon_sym_return_BANG] = ACTIONS(7075),
    [anon_sym_yield] = ACTIONS(7073),
    [anon_sym_yield_BANG] = ACTIONS(7075),
    [anon_sym_lazy] = ACTIONS(7073),
    [anon_sym_assert] = ACTIONS(7073),
    [anon_sym_upcast] = ACTIONS(7073),
    [anon_sym_downcast] = ACTIONS(7073),
    [anon_sym_fixed] = ACTIONS(7073),
    [anon_sym_COLON_GT] = ACTIONS(7075),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7075),
    [anon_sym_for] = ACTIONS(7073),
    [anon_sym_while] = ACTIONS(7073),
    [anon_sym_while_BANG] = ACTIONS(7075),
    [anon_sym_if] = ACTIONS(7073),
    [anon_sym_fun] = ACTIONS(7073),
    [anon_sym_try] = ACTIONS(7073),
    [anon_sym_match] = ACTIONS(7073),
    [anon_sym_match_BANG] = ACTIONS(7075),
    [anon_sym_function] = ACTIONS(7073),
    [anon_sym_DOT_LBRACK] = ACTIONS(7075),
    [anon_sym_use] = ACTIONS(7073),
    [anon_sym_use_BANG] = ACTIONS(7075),
    [anon_sym_do_BANG] = ACTIONS(7075),
    [anon_sym_DOT_DOT] = ACTIONS(7075),
    [anon_sym_LPAREN2] = ACTIONS(7073),
    [anon_sym_PIPE_PIPE] = ACTIONS(7073),
    [anon_sym_AMP_AMP] = ACTIONS(7073),
    [aux_sym_char_token1] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7073),
    [anon_sym_DQUOTE] = ACTIONS(7073),
    [anon_sym_AT_DQUOTE] = ACTIONS(7075),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7075),
    [sym_bool] = ACTIONS(7073),
    [sym_unit] = ACTIONS(7075),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7073),
    [sym_op_identifier] = ACTIONS(7073),
    [anon_sym_PLUS] = ACTIONS(7073),
    [anon_sym_DASH] = ACTIONS(7073),
    [anon_sym_PLUS_DOT] = ACTIONS(7073),
    [anon_sym_DASH_DOT] = ACTIONS(7073),
    [anon_sym_PERCENT] = ACTIONS(7073),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7073),
    [anon_sym_TILDE] = ACTIONS(7075),
    [aux_sym_prefix_op_token1] = ACTIONS(7073),
    [aux_sym_infix_op_token1] = ACTIONS(7075),
    [aux_sym_infix_op_token2] = ACTIONS(7073),
    [aux_sym_infix_op_token3] = ACTIONS(7073),
    [anon_sym_BANG_EQ] = ACTIONS(7073),
    [anon_sym_COLON_EQ] = ACTIONS(7075),
    [anon_sym_DOLLAR] = ACTIONS(7073),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7073),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7073),
    [sym_int] = ACTIONS(7073),
    [sym_xint] = ACTIONS(7075),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7075),
    [sym__newline] = ACTIONS(7075),
    [sym__dedent] = ACTIONS(7075),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7075),
    [sym__tyapp_open] = ACTIONS(7075),
  },
  [STATE(5457)] = {
    [sym_block_comment] = STATE(5457),
    [sym_compiler_directive_decl] = STATE(5457),
    [sym_fsi_directive_decl] = STATE(5457),
    [sym_preproc_line] = STATE(5457),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_DASH_GT] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_DOT_DOT] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(5458)] = {
    [sym_block_comment] = STATE(5458),
    [sym_compiler_directive_decl] = STATE(5458),
    [sym_fsi_directive_decl] = STATE(5458),
    [sym_preproc_line] = STATE(5458),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_DASH_GT] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_DOT_DOT] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(5459)] = {
    [sym_block_comment] = STATE(5459),
    [sym_compiler_directive_decl] = STATE(5459),
    [sym_fsi_directive_decl] = STATE(5459),
    [sym_preproc_line] = STATE(5459),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_DASH_GT] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_DOT_DOT] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(5460)] = {
    [sym_block_comment] = STATE(5460),
    [sym_compiler_directive_decl] = STATE(5460),
    [sym_fsi_directive_decl] = STATE(5460),
    [sym_preproc_line] = STATE(5460),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_DASH_GT] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_DOT_DOT] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(5461)] = {
    [sym_block_comment] = STATE(5461),
    [sym_compiler_directive_decl] = STATE(5461),
    [sym_fsi_directive_decl] = STATE(5461),
    [sym_preproc_line] = STATE(5461),
    [sym_identifier] = ACTIONS(7077),
    [anon_sym_null] = ACTIONS(7077),
    [anon_sym_EQ] = ACTIONS(7077),
    [anon_sym_begin] = ACTIONS(7077),
    [anon_sym_COLON] = ACTIONS(7077),
    [anon_sym_do] = ACTIONS(7077),
    [anon_sym_let] = ACTIONS(7077),
    [anon_sym_let_BANG] = ACTIONS(7079),
    [anon_sym_LPAREN] = ACTIONS(7077),
    [anon_sym_COMMA] = ACTIONS(7079),
    [anon_sym_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_QMARK] = ACTIONS(7077),
    [anon_sym_COLON_COLON] = ACTIONS(7079),
    [anon_sym_AMP] = ACTIONS(7077),
    [anon_sym_LBRACK] = ACTIONS(7077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7079),
    [anon_sym_LBRACE] = ACTIONS(7077),
    [anon_sym_LT_AT] = ACTIONS(7077),
    [anon_sym_LT_AT_AT] = ACTIONS(7077),
    [anon_sym_CARET] = ACTIONS(7077),
    [anon_sym_or] = ACTIONS(7077),
    [anon_sym_DOT] = ACTIONS(7077),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7079),
    [anon_sym_new] = ACTIONS(7077),
    [anon_sym_return] = ACTIONS(7077),
    [anon_sym_return_BANG] = ACTIONS(7079),
    [anon_sym_yield] = ACTIONS(7077),
    [anon_sym_yield_BANG] = ACTIONS(7079),
    [anon_sym_lazy] = ACTIONS(7077),
    [anon_sym_assert] = ACTIONS(7077),
    [anon_sym_upcast] = ACTIONS(7077),
    [anon_sym_downcast] = ACTIONS(7077),
    [anon_sym_fixed] = ACTIONS(7077),
    [anon_sym_COLON_GT] = ACTIONS(7079),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7079),
    [anon_sym_for] = ACTIONS(7077),
    [anon_sym_while] = ACTIONS(7077),
    [anon_sym_while_BANG] = ACTIONS(7079),
    [anon_sym_if] = ACTIONS(7077),
    [anon_sym_fun] = ACTIONS(7077),
    [anon_sym_try] = ACTIONS(7077),
    [anon_sym_match] = ACTIONS(7077),
    [anon_sym_match_BANG] = ACTIONS(7079),
    [anon_sym_function] = ACTIONS(7077),
    [anon_sym_DOT_LBRACK] = ACTIONS(7079),
    [anon_sym_use] = ACTIONS(7077),
    [anon_sym_use_BANG] = ACTIONS(7079),
    [anon_sym_do_BANG] = ACTIONS(7079),
    [anon_sym_DOT_DOT] = ACTIONS(7079),
    [anon_sym_LPAREN2] = ACTIONS(7077),
    [anon_sym_PIPE_PIPE] = ACTIONS(7077),
    [anon_sym_AMP_AMP] = ACTIONS(7077),
    [aux_sym_char_token1] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7077),
    [anon_sym_DQUOTE] = ACTIONS(7077),
    [anon_sym_AT_DQUOTE] = ACTIONS(7079),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7079),
    [sym_bool] = ACTIONS(7077),
    [sym_unit] = ACTIONS(7079),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7077),
    [sym_op_identifier] = ACTIONS(7077),
    [anon_sym_PLUS] = ACTIONS(7077),
    [anon_sym_DASH] = ACTIONS(7077),
    [anon_sym_PLUS_DOT] = ACTIONS(7077),
    [anon_sym_DASH_DOT] = ACTIONS(7077),
    [anon_sym_PERCENT] = ACTIONS(7077),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7077),
    [anon_sym_TILDE] = ACTIONS(7079),
    [aux_sym_prefix_op_token1] = ACTIONS(7077),
    [aux_sym_infix_op_token1] = ACTIONS(7079),
    [aux_sym_infix_op_token2] = ACTIONS(7077),
    [aux_sym_infix_op_token3] = ACTIONS(7077),
    [anon_sym_BANG_EQ] = ACTIONS(7077),
    [anon_sym_COLON_EQ] = ACTIONS(7079),
    [anon_sym_DOLLAR] = ACTIONS(7077),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7077),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7077),
    [sym_int] = ACTIONS(7077),
    [sym_xint] = ACTIONS(7079),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7079),
    [sym__newline] = ACTIONS(7079),
    [sym__dedent] = ACTIONS(7079),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7079),
    [sym__tyapp_open] = ACTIONS(7079),
  },
  [STATE(5462)] = {
    [sym_block_comment] = STATE(5462),
    [sym_compiler_directive_decl] = STATE(5462),
    [sym_fsi_directive_decl] = STATE(5462),
    [sym_preproc_line] = STATE(5462),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_with] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9716),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5463)] = {
    [sym_block_comment] = STATE(5463),
    [sym_compiler_directive_decl] = STATE(5463),
    [sym_fsi_directive_decl] = STATE(5463),
    [sym_preproc_line] = STATE(5463),
    [sym_identifier] = ACTIONS(7085),
    [anon_sym_null] = ACTIONS(7085),
    [anon_sym_EQ] = ACTIONS(7085),
    [anon_sym_begin] = ACTIONS(7085),
    [anon_sym_COLON] = ACTIONS(7085),
    [anon_sym_do] = ACTIONS(7085),
    [anon_sym_let] = ACTIONS(7085),
    [anon_sym_let_BANG] = ACTIONS(7087),
    [anon_sym_LPAREN] = ACTIONS(7085),
    [anon_sym_COMMA] = ACTIONS(7087),
    [anon_sym_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_QMARK] = ACTIONS(7085),
    [anon_sym_COLON_COLON] = ACTIONS(7087),
    [anon_sym_AMP] = ACTIONS(7085),
    [anon_sym_LBRACK] = ACTIONS(7085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7087),
    [anon_sym_LBRACE] = ACTIONS(7085),
    [anon_sym_LT_AT] = ACTIONS(7085),
    [anon_sym_LT_AT_AT] = ACTIONS(7085),
    [anon_sym_CARET] = ACTIONS(7085),
    [anon_sym_or] = ACTIONS(7085),
    [anon_sym_DOT] = ACTIONS(7085),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7087),
    [anon_sym_new] = ACTIONS(7085),
    [anon_sym_return] = ACTIONS(7085),
    [anon_sym_return_BANG] = ACTIONS(7087),
    [anon_sym_yield] = ACTIONS(7085),
    [anon_sym_yield_BANG] = ACTIONS(7087),
    [anon_sym_lazy] = ACTIONS(7085),
    [anon_sym_assert] = ACTIONS(7085),
    [anon_sym_upcast] = ACTIONS(7085),
    [anon_sym_downcast] = ACTIONS(7085),
    [anon_sym_fixed] = ACTIONS(7085),
    [anon_sym_COLON_GT] = ACTIONS(7087),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7087),
    [anon_sym_for] = ACTIONS(7085),
    [anon_sym_while] = ACTIONS(7085),
    [anon_sym_while_BANG] = ACTIONS(7087),
    [anon_sym_if] = ACTIONS(7085),
    [anon_sym_fun] = ACTIONS(7085),
    [anon_sym_try] = ACTIONS(7085),
    [anon_sym_match] = ACTIONS(7085),
    [anon_sym_match_BANG] = ACTIONS(7087),
    [anon_sym_function] = ACTIONS(7085),
    [anon_sym_DOT_LBRACK] = ACTIONS(7087),
    [anon_sym_use] = ACTIONS(7085),
    [anon_sym_use_BANG] = ACTIONS(7087),
    [anon_sym_do_BANG] = ACTIONS(7087),
    [anon_sym_DOT_DOT] = ACTIONS(7087),
    [anon_sym_LPAREN2] = ACTIONS(7085),
    [anon_sym_PIPE_PIPE] = ACTIONS(7085),
    [anon_sym_AMP_AMP] = ACTIONS(7085),
    [aux_sym_char_token1] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7085),
    [anon_sym_DQUOTE] = ACTIONS(7085),
    [anon_sym_AT_DQUOTE] = ACTIONS(7087),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7087),
    [sym_bool] = ACTIONS(7085),
    [sym_unit] = ACTIONS(7087),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7085),
    [sym_op_identifier] = ACTIONS(7085),
    [anon_sym_PLUS] = ACTIONS(7085),
    [anon_sym_DASH] = ACTIONS(7085),
    [anon_sym_PLUS_DOT] = ACTIONS(7085),
    [anon_sym_DASH_DOT] = ACTIONS(7085),
    [anon_sym_PERCENT] = ACTIONS(7085),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7085),
    [anon_sym_TILDE] = ACTIONS(7087),
    [aux_sym_prefix_op_token1] = ACTIONS(7085),
    [aux_sym_infix_op_token1] = ACTIONS(7087),
    [aux_sym_infix_op_token2] = ACTIONS(7085),
    [aux_sym_infix_op_token3] = ACTIONS(7085),
    [anon_sym_BANG_EQ] = ACTIONS(7085),
    [anon_sym_COLON_EQ] = ACTIONS(7087),
    [anon_sym_DOLLAR] = ACTIONS(7085),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7085),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7085),
    [sym_int] = ACTIONS(7085),
    [sym_xint] = ACTIONS(7087),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7087),
    [sym__newline] = ACTIONS(7087),
    [sym__dedent] = ACTIONS(7087),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7087),
    [sym__tyapp_open] = ACTIONS(7087),
  },
  [STATE(5464)] = {
    [sym_block_comment] = STATE(5464),
    [sym_compiler_directive_decl] = STATE(5464),
    [sym_fsi_directive_decl] = STATE(5464),
    [sym_preproc_line] = STATE(5464),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__untyped_quoted_close] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5465)] = {
    [sym_block_comment] = STATE(5465),
    [sym_compiler_directive_decl] = STATE(5465),
    [sym_fsi_directive_decl] = STATE(5465),
    [sym_preproc_line] = STATE(5465),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_as] = ACTIONS(7067),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_with] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(5466)] = {
    [sym_block_comment] = STATE(5466),
    [sym_compiler_directive_decl] = STATE(5466),
    [sym_fsi_directive_decl] = STATE(5466),
    [sym_preproc_line] = STATE(5466),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__quoted_close] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5467)] = {
    [sym_block_comment] = STATE(5467),
    [sym_compiler_directive_decl] = STATE(5467),
    [sym_fsi_directive_decl] = STATE(5467),
    [sym_preproc_line] = STATE(5467),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_as] = ACTIONS(7125),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_with] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(5468)] = {
    [sym_block_comment] = STATE(5468),
    [sym_compiler_directive_decl] = STATE(5468),
    [sym_fsi_directive_decl] = STATE(5468),
    [sym_preproc_line] = STATE(5468),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_DASH_GT] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_DOT_DOT] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(5469)] = {
    [sym_block_comment] = STATE(5469),
    [sym_compiler_directive_decl] = STATE(5469),
    [sym_fsi_directive_decl] = STATE(5469),
    [sym_preproc_line] = STATE(5469),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_as] = ACTIONS(7129),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_with] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(5470)] = {
    [sym_block_comment] = STATE(5470),
    [sym_compiler_directive_decl] = STATE(5470),
    [sym_fsi_directive_decl] = STATE(5470),
    [sym_preproc_line] = STATE(5470),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_as] = ACTIONS(7133),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_with] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(5471)] = {
    [sym_block_comment] = STATE(5471),
    [sym_compiler_directive_decl] = STATE(5471),
    [sym_fsi_directive_decl] = STATE(5471),
    [sym_preproc_line] = STATE(5471),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_GT] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5472)] = {
    [sym_block_comment] = STATE(5472),
    [sym_compiler_directive_decl] = STATE(5472),
    [sym_fsi_directive_decl] = STATE(5472),
    [sym_preproc_line] = STATE(5472),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_GT] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5473)] = {
    [sym_block_comment] = STATE(5473),
    [sym_compiler_directive_decl] = STATE(5473),
    [sym_fsi_directive_decl] = STATE(5473),
    [sym_preproc_line] = STATE(5473),
    [sym_identifier] = ACTIONS(7089),
    [anon_sym_null] = ACTIONS(7089),
    [anon_sym_EQ] = ACTIONS(7089),
    [anon_sym_begin] = ACTIONS(7089),
    [anon_sym_COLON] = ACTIONS(7089),
    [anon_sym_do] = ACTIONS(7089),
    [anon_sym_let] = ACTIONS(7089),
    [anon_sym_let_BANG] = ACTIONS(7091),
    [anon_sym_LPAREN] = ACTIONS(7089),
    [anon_sym_COMMA] = ACTIONS(7091),
    [anon_sym_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_QMARK] = ACTIONS(7089),
    [anon_sym_COLON_COLON] = ACTIONS(7091),
    [anon_sym_AMP] = ACTIONS(7089),
    [anon_sym_LBRACK] = ACTIONS(7089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7091),
    [anon_sym_LBRACE] = ACTIONS(7089),
    [anon_sym_LT_AT] = ACTIONS(7089),
    [anon_sym_LT_AT_AT] = ACTIONS(7089),
    [anon_sym_CARET] = ACTIONS(7089),
    [anon_sym_or] = ACTIONS(7089),
    [anon_sym_DOT] = ACTIONS(7089),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7091),
    [anon_sym_new] = ACTIONS(7089),
    [anon_sym_return] = ACTIONS(7089),
    [anon_sym_return_BANG] = ACTIONS(7091),
    [anon_sym_yield] = ACTIONS(7089),
    [anon_sym_yield_BANG] = ACTIONS(7091),
    [anon_sym_lazy] = ACTIONS(7089),
    [anon_sym_assert] = ACTIONS(7089),
    [anon_sym_upcast] = ACTIONS(7089),
    [anon_sym_downcast] = ACTIONS(7089),
    [anon_sym_fixed] = ACTIONS(7089),
    [anon_sym_COLON_GT] = ACTIONS(7091),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7091),
    [anon_sym_for] = ACTIONS(7089),
    [anon_sym_while] = ACTIONS(7089),
    [anon_sym_while_BANG] = ACTIONS(7091),
    [anon_sym_if] = ACTIONS(7089),
    [anon_sym_fun] = ACTIONS(7089),
    [anon_sym_try] = ACTIONS(7089),
    [anon_sym_match] = ACTIONS(7089),
    [anon_sym_match_BANG] = ACTIONS(7091),
    [anon_sym_function] = ACTIONS(7089),
    [anon_sym_DOT_LBRACK] = ACTIONS(7091),
    [anon_sym_use] = ACTIONS(7089),
    [anon_sym_use_BANG] = ACTIONS(7091),
    [anon_sym_do_BANG] = ACTIONS(7091),
    [anon_sym_DOT_DOT] = ACTIONS(7091),
    [anon_sym_LPAREN2] = ACTIONS(7089),
    [anon_sym_PIPE_PIPE] = ACTIONS(7089),
    [anon_sym_AMP_AMP] = ACTIONS(7089),
    [aux_sym_char_token1] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7089),
    [anon_sym_DQUOTE] = ACTIONS(7089),
    [anon_sym_AT_DQUOTE] = ACTIONS(7091),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7091),
    [sym_bool] = ACTIONS(7089),
    [sym_unit] = ACTIONS(7091),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7089),
    [sym_op_identifier] = ACTIONS(7089),
    [anon_sym_PLUS] = ACTIONS(7089),
    [anon_sym_DASH] = ACTIONS(7089),
    [anon_sym_PLUS_DOT] = ACTIONS(7089),
    [anon_sym_DASH_DOT] = ACTIONS(7089),
    [anon_sym_PERCENT] = ACTIONS(7089),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7089),
    [anon_sym_TILDE] = ACTIONS(7091),
    [aux_sym_prefix_op_token1] = ACTIONS(7089),
    [aux_sym_infix_op_token1] = ACTIONS(7091),
    [aux_sym_infix_op_token2] = ACTIONS(7089),
    [aux_sym_infix_op_token3] = ACTIONS(7089),
    [anon_sym_BANG_EQ] = ACTIONS(7089),
    [anon_sym_COLON_EQ] = ACTIONS(7091),
    [anon_sym_DOLLAR] = ACTIONS(7089),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7089),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7089),
    [sym_int] = ACTIONS(7089),
    [sym_xint] = ACTIONS(7091),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7091),
    [sym__newline] = ACTIONS(7091),
    [sym__dedent] = ACTIONS(7091),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7091),
    [sym__tyapp_open] = ACTIONS(7091),
  },
  [STATE(5474)] = {
    [sym_block_comment] = STATE(5474),
    [sym_compiler_directive_decl] = STATE(5474),
    [sym_fsi_directive_decl] = STATE(5474),
    [sym_preproc_line] = STATE(5474),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__quoted_close] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5475)] = {
    [sym_block_comment] = STATE(5475),
    [sym_compiler_directive_decl] = STATE(5475),
    [sym_fsi_directive_decl] = STATE(5475),
    [sym_preproc_line] = STATE(5475),
    [sym_identifier] = ACTIONS(7093),
    [anon_sym_null] = ACTIONS(7093),
    [anon_sym_EQ] = ACTIONS(7093),
    [anon_sym_begin] = ACTIONS(7093),
    [anon_sym_COLON] = ACTIONS(7093),
    [anon_sym_do] = ACTIONS(7093),
    [anon_sym_let] = ACTIONS(7093),
    [anon_sym_let_BANG] = ACTIONS(7095),
    [anon_sym_LPAREN] = ACTIONS(7093),
    [anon_sym_COMMA] = ACTIONS(7095),
    [anon_sym_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_QMARK] = ACTIONS(7093),
    [anon_sym_COLON_COLON] = ACTIONS(7095),
    [anon_sym_AMP] = ACTIONS(7093),
    [anon_sym_LBRACK] = ACTIONS(7093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7095),
    [anon_sym_LBRACE] = ACTIONS(7093),
    [anon_sym_LT_AT] = ACTIONS(7093),
    [anon_sym_LT_AT_AT] = ACTIONS(7093),
    [anon_sym_CARET] = ACTIONS(7093),
    [anon_sym_or] = ACTIONS(7093),
    [anon_sym_DOT] = ACTIONS(7093),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7095),
    [anon_sym_new] = ACTIONS(7093),
    [anon_sym_return] = ACTIONS(7093),
    [anon_sym_return_BANG] = ACTIONS(7095),
    [anon_sym_yield] = ACTIONS(7093),
    [anon_sym_yield_BANG] = ACTIONS(7095),
    [anon_sym_lazy] = ACTIONS(7093),
    [anon_sym_assert] = ACTIONS(7093),
    [anon_sym_upcast] = ACTIONS(7093),
    [anon_sym_downcast] = ACTIONS(7093),
    [anon_sym_fixed] = ACTIONS(7093),
    [anon_sym_COLON_GT] = ACTIONS(7095),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7095),
    [anon_sym_for] = ACTIONS(7093),
    [anon_sym_while] = ACTIONS(7093),
    [anon_sym_while_BANG] = ACTIONS(7095),
    [anon_sym_if] = ACTIONS(7093),
    [anon_sym_fun] = ACTIONS(7093),
    [anon_sym_try] = ACTIONS(7093),
    [anon_sym_match] = ACTIONS(7093),
    [anon_sym_match_BANG] = ACTIONS(7095),
    [anon_sym_function] = ACTIONS(7093),
    [anon_sym_DOT_LBRACK] = ACTIONS(7095),
    [anon_sym_use] = ACTIONS(7093),
    [anon_sym_use_BANG] = ACTIONS(7095),
    [anon_sym_do_BANG] = ACTIONS(7095),
    [anon_sym_DOT_DOT] = ACTIONS(7095),
    [anon_sym_LPAREN2] = ACTIONS(7093),
    [anon_sym_PIPE_PIPE] = ACTIONS(7093),
    [anon_sym_AMP_AMP] = ACTIONS(7093),
    [aux_sym_char_token1] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7093),
    [anon_sym_DQUOTE] = ACTIONS(7093),
    [anon_sym_AT_DQUOTE] = ACTIONS(7095),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7095),
    [sym_bool] = ACTIONS(7093),
    [sym_unit] = ACTIONS(7095),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7093),
    [sym_op_identifier] = ACTIONS(7093),
    [anon_sym_PLUS] = ACTIONS(7093),
    [anon_sym_DASH] = ACTIONS(7093),
    [anon_sym_PLUS_DOT] = ACTIONS(7093),
    [anon_sym_DASH_DOT] = ACTIONS(7093),
    [anon_sym_PERCENT] = ACTIONS(7093),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7093),
    [anon_sym_TILDE] = ACTIONS(7095),
    [aux_sym_prefix_op_token1] = ACTIONS(7093),
    [aux_sym_infix_op_token1] = ACTIONS(7095),
    [aux_sym_infix_op_token2] = ACTIONS(7093),
    [aux_sym_infix_op_token3] = ACTIONS(7093),
    [anon_sym_BANG_EQ] = ACTIONS(7093),
    [anon_sym_COLON_EQ] = ACTIONS(7095),
    [anon_sym_DOLLAR] = ACTIONS(7093),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7093),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7093),
    [sym_int] = ACTIONS(7093),
    [sym_xint] = ACTIONS(7095),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7095),
    [sym__newline] = ACTIONS(7095),
    [sym__dedent] = ACTIONS(7095),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7095),
    [sym__tyapp_open] = ACTIONS(7095),
  },
  [STATE(5476)] = {
    [sym_block_comment] = STATE(5476),
    [sym_compiler_directive_decl] = STATE(5476),
    [sym_fsi_directive_decl] = STATE(5476),
    [sym_preproc_line] = STATE(5476),
    [sym_identifier] = ACTIONS(7097),
    [anon_sym_null] = ACTIONS(7097),
    [anon_sym_EQ] = ACTIONS(7097),
    [anon_sym_begin] = ACTIONS(7097),
    [anon_sym_COLON] = ACTIONS(7097),
    [anon_sym_do] = ACTIONS(7097),
    [anon_sym_let] = ACTIONS(7097),
    [anon_sym_let_BANG] = ACTIONS(7099),
    [anon_sym_LPAREN] = ACTIONS(7097),
    [anon_sym_COMMA] = ACTIONS(7099),
    [anon_sym_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_QMARK] = ACTIONS(7097),
    [anon_sym_COLON_COLON] = ACTIONS(7099),
    [anon_sym_AMP] = ACTIONS(7097),
    [anon_sym_LBRACK] = ACTIONS(7097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7099),
    [anon_sym_LBRACE] = ACTIONS(7097),
    [anon_sym_LT_AT] = ACTIONS(7097),
    [anon_sym_LT_AT_AT] = ACTIONS(7097),
    [anon_sym_CARET] = ACTIONS(7097),
    [anon_sym_or] = ACTIONS(7097),
    [anon_sym_DOT] = ACTIONS(7097),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7099),
    [anon_sym_new] = ACTIONS(7097),
    [anon_sym_return] = ACTIONS(7097),
    [anon_sym_return_BANG] = ACTIONS(7099),
    [anon_sym_yield] = ACTIONS(7097),
    [anon_sym_yield_BANG] = ACTIONS(7099),
    [anon_sym_lazy] = ACTIONS(7097),
    [anon_sym_assert] = ACTIONS(7097),
    [anon_sym_upcast] = ACTIONS(7097),
    [anon_sym_downcast] = ACTIONS(7097),
    [anon_sym_fixed] = ACTIONS(7097),
    [anon_sym_COLON_GT] = ACTIONS(7099),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7099),
    [anon_sym_for] = ACTIONS(7097),
    [anon_sym_while] = ACTIONS(7097),
    [anon_sym_while_BANG] = ACTIONS(7099),
    [anon_sym_if] = ACTIONS(7097),
    [anon_sym_fun] = ACTIONS(7097),
    [anon_sym_try] = ACTIONS(7097),
    [anon_sym_match] = ACTIONS(7097),
    [anon_sym_match_BANG] = ACTIONS(7099),
    [anon_sym_function] = ACTIONS(7097),
    [anon_sym_DOT_LBRACK] = ACTIONS(7099),
    [anon_sym_use] = ACTIONS(7097),
    [anon_sym_use_BANG] = ACTIONS(7099),
    [anon_sym_do_BANG] = ACTIONS(7099),
    [anon_sym_DOT_DOT] = ACTIONS(7099),
    [anon_sym_LPAREN2] = ACTIONS(7097),
    [anon_sym_PIPE_PIPE] = ACTIONS(7097),
    [anon_sym_AMP_AMP] = ACTIONS(7097),
    [aux_sym_char_token1] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7097),
    [anon_sym_DQUOTE] = ACTIONS(7097),
    [anon_sym_AT_DQUOTE] = ACTIONS(7099),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7099),
    [sym_bool] = ACTIONS(7097),
    [sym_unit] = ACTIONS(7099),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7097),
    [sym_op_identifier] = ACTIONS(7097),
    [anon_sym_PLUS] = ACTIONS(7097),
    [anon_sym_DASH] = ACTIONS(7097),
    [anon_sym_PLUS_DOT] = ACTIONS(7097),
    [anon_sym_DASH_DOT] = ACTIONS(7097),
    [anon_sym_PERCENT] = ACTIONS(7097),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7097),
    [anon_sym_TILDE] = ACTIONS(7099),
    [aux_sym_prefix_op_token1] = ACTIONS(7097),
    [aux_sym_infix_op_token1] = ACTIONS(7099),
    [aux_sym_infix_op_token2] = ACTIONS(7097),
    [aux_sym_infix_op_token3] = ACTIONS(7097),
    [anon_sym_BANG_EQ] = ACTIONS(7097),
    [anon_sym_COLON_EQ] = ACTIONS(7099),
    [anon_sym_DOLLAR] = ACTIONS(7097),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7097),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7097),
    [sym_int] = ACTIONS(7097),
    [sym_xint] = ACTIONS(7099),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7099),
    [sym__newline] = ACTIONS(7099),
    [sym__dedent] = ACTIONS(7099),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7099),
    [sym__tyapp_open] = ACTIONS(7099),
  },
  [STATE(5477)] = {
    [sym_block_comment] = STATE(5477),
    [sym_compiler_directive_decl] = STATE(5477),
    [sym_fsi_directive_decl] = STATE(5477),
    [sym_preproc_line] = STATE(5477),
    [sym_identifier] = ACTIONS(7101),
    [anon_sym_null] = ACTIONS(7101),
    [anon_sym_EQ] = ACTIONS(7101),
    [anon_sym_begin] = ACTIONS(7101),
    [anon_sym_COLON] = ACTIONS(7101),
    [anon_sym_do] = ACTIONS(7101),
    [anon_sym_let] = ACTIONS(7101),
    [anon_sym_let_BANG] = ACTIONS(7103),
    [anon_sym_LPAREN] = ACTIONS(7101),
    [anon_sym_COMMA] = ACTIONS(7103),
    [anon_sym_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_QMARK] = ACTIONS(7101),
    [anon_sym_COLON_COLON] = ACTIONS(7103),
    [anon_sym_AMP] = ACTIONS(7101),
    [anon_sym_LBRACK] = ACTIONS(7101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7103),
    [anon_sym_LBRACE] = ACTIONS(7101),
    [anon_sym_LT_AT] = ACTIONS(7101),
    [anon_sym_LT_AT_AT] = ACTIONS(7101),
    [anon_sym_CARET] = ACTIONS(7101),
    [anon_sym_or] = ACTIONS(7101),
    [anon_sym_DOT] = ACTIONS(7101),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7103),
    [anon_sym_new] = ACTIONS(7101),
    [anon_sym_return] = ACTIONS(7101),
    [anon_sym_return_BANG] = ACTIONS(7103),
    [anon_sym_yield] = ACTIONS(7101),
    [anon_sym_yield_BANG] = ACTIONS(7103),
    [anon_sym_lazy] = ACTIONS(7101),
    [anon_sym_assert] = ACTIONS(7101),
    [anon_sym_upcast] = ACTIONS(7101),
    [anon_sym_downcast] = ACTIONS(7101),
    [anon_sym_fixed] = ACTIONS(7101),
    [anon_sym_COLON_GT] = ACTIONS(7103),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7103),
    [anon_sym_for] = ACTIONS(7101),
    [anon_sym_while] = ACTIONS(7101),
    [anon_sym_while_BANG] = ACTIONS(7103),
    [anon_sym_if] = ACTIONS(7101),
    [anon_sym_fun] = ACTIONS(7101),
    [anon_sym_try] = ACTIONS(7101),
    [anon_sym_match] = ACTIONS(7101),
    [anon_sym_match_BANG] = ACTIONS(7103),
    [anon_sym_function] = ACTIONS(7101),
    [anon_sym_DOT_LBRACK] = ACTIONS(7103),
    [anon_sym_use] = ACTIONS(7101),
    [anon_sym_use_BANG] = ACTIONS(7103),
    [anon_sym_do_BANG] = ACTIONS(7103),
    [anon_sym_DOT_DOT] = ACTIONS(7103),
    [anon_sym_LPAREN2] = ACTIONS(7101),
    [anon_sym_PIPE_PIPE] = ACTIONS(7101),
    [anon_sym_AMP_AMP] = ACTIONS(7101),
    [aux_sym_char_token1] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7101),
    [anon_sym_DQUOTE] = ACTIONS(7101),
    [anon_sym_AT_DQUOTE] = ACTIONS(7103),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7103),
    [sym_bool] = ACTIONS(7101),
    [sym_unit] = ACTIONS(7103),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7101),
    [sym_op_identifier] = ACTIONS(7101),
    [anon_sym_PLUS] = ACTIONS(7101),
    [anon_sym_DASH] = ACTIONS(7101),
    [anon_sym_PLUS_DOT] = ACTIONS(7101),
    [anon_sym_DASH_DOT] = ACTIONS(7101),
    [anon_sym_PERCENT] = ACTIONS(7101),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7101),
    [anon_sym_TILDE] = ACTIONS(7103),
    [aux_sym_prefix_op_token1] = ACTIONS(7101),
    [aux_sym_infix_op_token1] = ACTIONS(7103),
    [aux_sym_infix_op_token2] = ACTIONS(7101),
    [aux_sym_infix_op_token3] = ACTIONS(7101),
    [anon_sym_BANG_EQ] = ACTIONS(7101),
    [anon_sym_COLON_EQ] = ACTIONS(7103),
    [anon_sym_DOLLAR] = ACTIONS(7101),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7101),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7101),
    [sym_int] = ACTIONS(7101),
    [sym_xint] = ACTIONS(7103),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7103),
    [sym__newline] = ACTIONS(7103),
    [sym__dedent] = ACTIONS(7103),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7103),
    [sym__tyapp_open] = ACTIONS(7103),
  },
  [STATE(5478)] = {
    [sym_block_comment] = STATE(5478),
    [sym_compiler_directive_decl] = STATE(5478),
    [sym_fsi_directive_decl] = STATE(5478),
    [sym_preproc_line] = STATE(5478),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_PIPE] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6418),
    [anon_sym_CARET] = ACTIONS(6418),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6418),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_AMP_AMP] = ACTIONS(6418),
    [anon_sym_STAR] = ACTIONS(6418),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6418),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6418),
    [anon_sym_DASH_DOT] = ACTIONS(6418),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6418),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6418),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [sym__newline] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
  },
  [STATE(5479)] = {
    [sym_block_comment] = STATE(5479),
    [sym_compiler_directive_decl] = STATE(5479),
    [sym_fsi_directive_decl] = STATE(5479),
    [sym_preproc_line] = STATE(5479),
    [sym_identifier] = ACTIONS(7105),
    [anon_sym_null] = ACTIONS(7105),
    [anon_sym_EQ] = ACTIONS(7105),
    [anon_sym_begin] = ACTIONS(7105),
    [anon_sym_COLON] = ACTIONS(7105),
    [anon_sym_do] = ACTIONS(7105),
    [anon_sym_let] = ACTIONS(7105),
    [anon_sym_let_BANG] = ACTIONS(7107),
    [anon_sym_LPAREN] = ACTIONS(7105),
    [anon_sym_COMMA] = ACTIONS(7107),
    [anon_sym_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_QMARK] = ACTIONS(7105),
    [anon_sym_COLON_COLON] = ACTIONS(7107),
    [anon_sym_AMP] = ACTIONS(7105),
    [anon_sym_LBRACK] = ACTIONS(7105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7107),
    [anon_sym_LBRACE] = ACTIONS(7105),
    [anon_sym_LT_AT] = ACTIONS(7105),
    [anon_sym_LT_AT_AT] = ACTIONS(7105),
    [anon_sym_CARET] = ACTIONS(7105),
    [anon_sym_or] = ACTIONS(7105),
    [anon_sym_DOT] = ACTIONS(7105),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7107),
    [anon_sym_new] = ACTIONS(7105),
    [anon_sym_return] = ACTIONS(7105),
    [anon_sym_return_BANG] = ACTIONS(7107),
    [anon_sym_yield] = ACTIONS(7105),
    [anon_sym_yield_BANG] = ACTIONS(7107),
    [anon_sym_lazy] = ACTIONS(7105),
    [anon_sym_assert] = ACTIONS(7105),
    [anon_sym_upcast] = ACTIONS(7105),
    [anon_sym_downcast] = ACTIONS(7105),
    [anon_sym_fixed] = ACTIONS(7105),
    [anon_sym_COLON_GT] = ACTIONS(7107),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7107),
    [anon_sym_for] = ACTIONS(7105),
    [anon_sym_while] = ACTIONS(7105),
    [anon_sym_while_BANG] = ACTIONS(7107),
    [anon_sym_if] = ACTIONS(7105),
    [anon_sym_fun] = ACTIONS(7105),
    [anon_sym_try] = ACTIONS(7105),
    [anon_sym_match] = ACTIONS(7105),
    [anon_sym_match_BANG] = ACTIONS(7107),
    [anon_sym_function] = ACTIONS(7105),
    [anon_sym_DOT_LBRACK] = ACTIONS(7107),
    [anon_sym_use] = ACTIONS(7105),
    [anon_sym_use_BANG] = ACTIONS(7107),
    [anon_sym_do_BANG] = ACTIONS(7107),
    [anon_sym_DOT_DOT] = ACTIONS(7107),
    [anon_sym_LPAREN2] = ACTIONS(7105),
    [anon_sym_PIPE_PIPE] = ACTIONS(7105),
    [anon_sym_AMP_AMP] = ACTIONS(7105),
    [aux_sym_char_token1] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7105),
    [anon_sym_DQUOTE] = ACTIONS(7105),
    [anon_sym_AT_DQUOTE] = ACTIONS(7107),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7107),
    [sym_bool] = ACTIONS(7105),
    [sym_unit] = ACTIONS(7107),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7105),
    [sym_op_identifier] = ACTIONS(7105),
    [anon_sym_PLUS] = ACTIONS(7105),
    [anon_sym_DASH] = ACTIONS(7105),
    [anon_sym_PLUS_DOT] = ACTIONS(7105),
    [anon_sym_DASH_DOT] = ACTIONS(7105),
    [anon_sym_PERCENT] = ACTIONS(7105),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7105),
    [anon_sym_TILDE] = ACTIONS(7107),
    [aux_sym_prefix_op_token1] = ACTIONS(7105),
    [aux_sym_infix_op_token1] = ACTIONS(7107),
    [aux_sym_infix_op_token2] = ACTIONS(7105),
    [aux_sym_infix_op_token3] = ACTIONS(7105),
    [anon_sym_BANG_EQ] = ACTIONS(7105),
    [anon_sym_COLON_EQ] = ACTIONS(7107),
    [anon_sym_DOLLAR] = ACTIONS(7105),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7105),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7105),
    [sym_int] = ACTIONS(7105),
    [sym_xint] = ACTIONS(7107),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7107),
    [sym__newline] = ACTIONS(7107),
    [sym__dedent] = ACTIONS(7107),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7107),
    [sym__tyapp_open] = ACTIONS(7107),
  },
  [STATE(5480)] = {
    [sym_block_comment] = STATE(5480),
    [sym_compiler_directive_decl] = STATE(5480),
    [sym_fsi_directive_decl] = STATE(5480),
    [sym_preproc_line] = STATE(5480),
    [aux_sym_sequential_expression_repeat1] = STATE(5699),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_GT] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5481)] = {
    [sym_block_comment] = STATE(5481),
    [sym_compiler_directive_decl] = STATE(5481),
    [sym_fsi_directive_decl] = STATE(5481),
    [sym_preproc_line] = STATE(5481),
    [sym_identifier] = ACTIONS(7109),
    [anon_sym_null] = ACTIONS(7109),
    [anon_sym_EQ] = ACTIONS(7109),
    [anon_sym_begin] = ACTIONS(7109),
    [anon_sym_COLON] = ACTIONS(7109),
    [anon_sym_do] = ACTIONS(7109),
    [anon_sym_let] = ACTIONS(7109),
    [anon_sym_let_BANG] = ACTIONS(7111),
    [anon_sym_LPAREN] = ACTIONS(7109),
    [anon_sym_COMMA] = ACTIONS(7111),
    [anon_sym_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_QMARK] = ACTIONS(7109),
    [anon_sym_COLON_COLON] = ACTIONS(7111),
    [anon_sym_AMP] = ACTIONS(7109),
    [anon_sym_LBRACK] = ACTIONS(7109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7111),
    [anon_sym_LBRACE] = ACTIONS(7109),
    [anon_sym_LT_AT] = ACTIONS(7109),
    [anon_sym_LT_AT_AT] = ACTIONS(7109),
    [anon_sym_CARET] = ACTIONS(7109),
    [anon_sym_or] = ACTIONS(7109),
    [anon_sym_DOT] = ACTIONS(7109),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7111),
    [anon_sym_new] = ACTIONS(7109),
    [anon_sym_return] = ACTIONS(7109),
    [anon_sym_return_BANG] = ACTIONS(7111),
    [anon_sym_yield] = ACTIONS(7109),
    [anon_sym_yield_BANG] = ACTIONS(7111),
    [anon_sym_lazy] = ACTIONS(7109),
    [anon_sym_assert] = ACTIONS(7109),
    [anon_sym_upcast] = ACTIONS(7109),
    [anon_sym_downcast] = ACTIONS(7109),
    [anon_sym_fixed] = ACTIONS(7109),
    [anon_sym_COLON_GT] = ACTIONS(7111),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7111),
    [anon_sym_for] = ACTIONS(7109),
    [anon_sym_while] = ACTIONS(7109),
    [anon_sym_while_BANG] = ACTIONS(7111),
    [anon_sym_if] = ACTIONS(7109),
    [anon_sym_fun] = ACTIONS(7109),
    [anon_sym_try] = ACTIONS(7109),
    [anon_sym_match] = ACTIONS(7109),
    [anon_sym_match_BANG] = ACTIONS(7111),
    [anon_sym_function] = ACTIONS(7109),
    [anon_sym_DOT_LBRACK] = ACTIONS(7111),
    [anon_sym_use] = ACTIONS(7109),
    [anon_sym_use_BANG] = ACTIONS(7111),
    [anon_sym_do_BANG] = ACTIONS(7111),
    [anon_sym_DOT_DOT] = ACTIONS(7111),
    [anon_sym_LPAREN2] = ACTIONS(7109),
    [anon_sym_PIPE_PIPE] = ACTIONS(7109),
    [anon_sym_AMP_AMP] = ACTIONS(7109),
    [aux_sym_char_token1] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7109),
    [anon_sym_DQUOTE] = ACTIONS(7109),
    [anon_sym_AT_DQUOTE] = ACTIONS(7111),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7111),
    [sym_bool] = ACTIONS(7109),
    [sym_unit] = ACTIONS(7111),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7109),
    [sym_op_identifier] = ACTIONS(7109),
    [anon_sym_PLUS] = ACTIONS(7109),
    [anon_sym_DASH] = ACTIONS(7109),
    [anon_sym_PLUS_DOT] = ACTIONS(7109),
    [anon_sym_DASH_DOT] = ACTIONS(7109),
    [anon_sym_PERCENT] = ACTIONS(7109),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7109),
    [anon_sym_TILDE] = ACTIONS(7111),
    [aux_sym_prefix_op_token1] = ACTIONS(7109),
    [aux_sym_infix_op_token1] = ACTIONS(7111),
    [aux_sym_infix_op_token2] = ACTIONS(7109),
    [aux_sym_infix_op_token3] = ACTIONS(7109),
    [anon_sym_BANG_EQ] = ACTIONS(7109),
    [anon_sym_COLON_EQ] = ACTIONS(7111),
    [anon_sym_DOLLAR] = ACTIONS(7109),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7109),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7109),
    [sym_int] = ACTIONS(7109),
    [sym_xint] = ACTIONS(7111),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7111),
    [sym__newline] = ACTIONS(7111),
    [sym__dedent] = ACTIONS(7111),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7111),
    [sym__tyapp_open] = ACTIONS(7111),
  },
  [STATE(5482)] = {
    [sym_block_comment] = STATE(5482),
    [sym_compiler_directive_decl] = STATE(5482),
    [sym_fsi_directive_decl] = STATE(5482),
    [sym_preproc_line] = STATE(5482),
    [sym_identifier] = ACTIONS(7113),
    [anon_sym_null] = ACTIONS(7113),
    [anon_sym_EQ] = ACTIONS(7113),
    [anon_sym_begin] = ACTIONS(7113),
    [anon_sym_COLON] = ACTIONS(7113),
    [anon_sym_do] = ACTIONS(7113),
    [anon_sym_let] = ACTIONS(7113),
    [anon_sym_let_BANG] = ACTIONS(7115),
    [anon_sym_LPAREN] = ACTIONS(7113),
    [anon_sym_COMMA] = ACTIONS(7115),
    [anon_sym_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_QMARK] = ACTIONS(7113),
    [anon_sym_COLON_COLON] = ACTIONS(7115),
    [anon_sym_AMP] = ACTIONS(7113),
    [anon_sym_LBRACK] = ACTIONS(7113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7115),
    [anon_sym_LBRACE] = ACTIONS(7113),
    [anon_sym_LT_AT] = ACTIONS(7113),
    [anon_sym_LT_AT_AT] = ACTIONS(7113),
    [anon_sym_CARET] = ACTIONS(7113),
    [anon_sym_or] = ACTIONS(7113),
    [anon_sym_DOT] = ACTIONS(7113),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7115),
    [anon_sym_new] = ACTIONS(7113),
    [anon_sym_return] = ACTIONS(7113),
    [anon_sym_return_BANG] = ACTIONS(7115),
    [anon_sym_yield] = ACTIONS(7113),
    [anon_sym_yield_BANG] = ACTIONS(7115),
    [anon_sym_lazy] = ACTIONS(7113),
    [anon_sym_assert] = ACTIONS(7113),
    [anon_sym_upcast] = ACTIONS(7113),
    [anon_sym_downcast] = ACTIONS(7113),
    [anon_sym_fixed] = ACTIONS(7113),
    [anon_sym_COLON_GT] = ACTIONS(7115),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7115),
    [anon_sym_for] = ACTIONS(7113),
    [anon_sym_while] = ACTIONS(7113),
    [anon_sym_while_BANG] = ACTIONS(7115),
    [anon_sym_if] = ACTIONS(7113),
    [anon_sym_fun] = ACTIONS(7113),
    [anon_sym_try] = ACTIONS(7113),
    [anon_sym_match] = ACTIONS(7113),
    [anon_sym_match_BANG] = ACTIONS(7115),
    [anon_sym_function] = ACTIONS(7113),
    [anon_sym_DOT_LBRACK] = ACTIONS(7115),
    [anon_sym_use] = ACTIONS(7113),
    [anon_sym_use_BANG] = ACTIONS(7115),
    [anon_sym_do_BANG] = ACTIONS(7115),
    [anon_sym_DOT_DOT] = ACTIONS(7115),
    [anon_sym_LPAREN2] = ACTIONS(7113),
    [anon_sym_PIPE_PIPE] = ACTIONS(7113),
    [anon_sym_AMP_AMP] = ACTIONS(7113),
    [aux_sym_char_token1] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7113),
    [anon_sym_DQUOTE] = ACTIONS(7113),
    [anon_sym_AT_DQUOTE] = ACTIONS(7115),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7115),
    [sym_bool] = ACTIONS(7113),
    [sym_unit] = ACTIONS(7115),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7113),
    [sym_op_identifier] = ACTIONS(7113),
    [anon_sym_PLUS] = ACTIONS(7113),
    [anon_sym_DASH] = ACTIONS(7113),
    [anon_sym_PLUS_DOT] = ACTIONS(7113),
    [anon_sym_DASH_DOT] = ACTIONS(7113),
    [anon_sym_PERCENT] = ACTIONS(7113),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7113),
    [anon_sym_TILDE] = ACTIONS(7115),
    [aux_sym_prefix_op_token1] = ACTIONS(7113),
    [aux_sym_infix_op_token1] = ACTIONS(7115),
    [aux_sym_infix_op_token2] = ACTIONS(7113),
    [aux_sym_infix_op_token3] = ACTIONS(7113),
    [anon_sym_BANG_EQ] = ACTIONS(7113),
    [anon_sym_COLON_EQ] = ACTIONS(7115),
    [anon_sym_DOLLAR] = ACTIONS(7113),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7113),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7113),
    [sym_int] = ACTIONS(7113),
    [sym_xint] = ACTIONS(7115),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7115),
    [sym__newline] = ACTIONS(7115),
    [sym__dedent] = ACTIONS(7115),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7115),
    [sym__tyapp_open] = ACTIONS(7115),
  },
  [STATE(5483)] = {
    [sym_block_comment] = STATE(5483),
    [sym_compiler_directive_decl] = STATE(5483),
    [sym_fsi_directive_decl] = STATE(5483),
    [sym_preproc_line] = STATE(5483),
    [sym_identifier] = ACTIONS(7117),
    [anon_sym_null] = ACTIONS(7117),
    [anon_sym_EQ] = ACTIONS(7117),
    [anon_sym_begin] = ACTIONS(7117),
    [anon_sym_COLON] = ACTIONS(7117),
    [anon_sym_do] = ACTIONS(7117),
    [anon_sym_let] = ACTIONS(7117),
    [anon_sym_let_BANG] = ACTIONS(7119),
    [anon_sym_LPAREN] = ACTIONS(7117),
    [anon_sym_COMMA] = ACTIONS(7119),
    [anon_sym_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_QMARK] = ACTIONS(7117),
    [anon_sym_COLON_COLON] = ACTIONS(7119),
    [anon_sym_AMP] = ACTIONS(7117),
    [anon_sym_LBRACK] = ACTIONS(7117),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7119),
    [anon_sym_LBRACE] = ACTIONS(7117),
    [anon_sym_LT_AT] = ACTIONS(7117),
    [anon_sym_LT_AT_AT] = ACTIONS(7117),
    [anon_sym_CARET] = ACTIONS(7117),
    [anon_sym_or] = ACTIONS(7117),
    [anon_sym_DOT] = ACTIONS(7117),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7119),
    [anon_sym_new] = ACTIONS(7117),
    [anon_sym_return] = ACTIONS(7117),
    [anon_sym_return_BANG] = ACTIONS(7119),
    [anon_sym_yield] = ACTIONS(7117),
    [anon_sym_yield_BANG] = ACTIONS(7119),
    [anon_sym_lazy] = ACTIONS(7117),
    [anon_sym_assert] = ACTIONS(7117),
    [anon_sym_upcast] = ACTIONS(7117),
    [anon_sym_downcast] = ACTIONS(7117),
    [anon_sym_fixed] = ACTIONS(7117),
    [anon_sym_COLON_GT] = ACTIONS(7119),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7119),
    [anon_sym_for] = ACTIONS(7117),
    [anon_sym_while] = ACTIONS(7117),
    [anon_sym_while_BANG] = ACTIONS(7119),
    [anon_sym_if] = ACTIONS(7117),
    [anon_sym_fun] = ACTIONS(7117),
    [anon_sym_try] = ACTIONS(7117),
    [anon_sym_match] = ACTIONS(7117),
    [anon_sym_match_BANG] = ACTIONS(7119),
    [anon_sym_function] = ACTIONS(7117),
    [anon_sym_DOT_LBRACK] = ACTIONS(7119),
    [anon_sym_use] = ACTIONS(7117),
    [anon_sym_use_BANG] = ACTIONS(7119),
    [anon_sym_do_BANG] = ACTIONS(7119),
    [anon_sym_DOT_DOT] = ACTIONS(7119),
    [anon_sym_LPAREN2] = ACTIONS(7117),
    [anon_sym_PIPE_PIPE] = ACTIONS(7117),
    [anon_sym_AMP_AMP] = ACTIONS(7117),
    [aux_sym_char_token1] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7117),
    [anon_sym_DQUOTE] = ACTIONS(7117),
    [anon_sym_AT_DQUOTE] = ACTIONS(7119),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7119),
    [sym_bool] = ACTIONS(7117),
    [sym_unit] = ACTIONS(7119),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7117),
    [sym_op_identifier] = ACTIONS(7117),
    [anon_sym_PLUS] = ACTIONS(7117),
    [anon_sym_DASH] = ACTIONS(7117),
    [anon_sym_PLUS_DOT] = ACTIONS(7117),
    [anon_sym_DASH_DOT] = ACTIONS(7117),
    [anon_sym_PERCENT] = ACTIONS(7117),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7117),
    [anon_sym_TILDE] = ACTIONS(7119),
    [aux_sym_prefix_op_token1] = ACTIONS(7117),
    [aux_sym_infix_op_token1] = ACTIONS(7119),
    [aux_sym_infix_op_token2] = ACTIONS(7117),
    [aux_sym_infix_op_token3] = ACTIONS(7117),
    [anon_sym_BANG_EQ] = ACTIONS(7117),
    [anon_sym_COLON_EQ] = ACTIONS(7119),
    [anon_sym_DOLLAR] = ACTIONS(7117),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7117),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7117),
    [sym_int] = ACTIONS(7117),
    [sym_xint] = ACTIONS(7119),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7119),
    [sym__newline] = ACTIONS(7119),
    [sym__dedent] = ACTIONS(7119),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7119),
    [sym__tyapp_open] = ACTIONS(7119),
  },
  [STATE(5484)] = {
    [sym_block_comment] = STATE(5484),
    [sym_compiler_directive_decl] = STATE(5484),
    [sym_fsi_directive_decl] = STATE(5484),
    [sym_preproc_line] = STATE(5484),
    [sym_identifier] = ACTIONS(7121),
    [anon_sym_null] = ACTIONS(7121),
    [anon_sym_EQ] = ACTIONS(7121),
    [anon_sym_begin] = ACTIONS(7121),
    [anon_sym_COLON] = ACTIONS(7121),
    [anon_sym_do] = ACTIONS(7121),
    [anon_sym_let] = ACTIONS(7121),
    [anon_sym_let_BANG] = ACTIONS(7123),
    [anon_sym_LPAREN] = ACTIONS(7121),
    [anon_sym_COMMA] = ACTIONS(7123),
    [anon_sym_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_QMARK] = ACTIONS(7121),
    [anon_sym_COLON_COLON] = ACTIONS(7123),
    [anon_sym_AMP] = ACTIONS(7121),
    [anon_sym_LBRACK] = ACTIONS(7121),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7123),
    [anon_sym_LBRACE] = ACTIONS(7121),
    [anon_sym_LT_AT] = ACTIONS(7121),
    [anon_sym_LT_AT_AT] = ACTIONS(7121),
    [anon_sym_CARET] = ACTIONS(7121),
    [anon_sym_or] = ACTIONS(7121),
    [anon_sym_DOT] = ACTIONS(7121),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7123),
    [anon_sym_new] = ACTIONS(7121),
    [anon_sym_return] = ACTIONS(7121),
    [anon_sym_return_BANG] = ACTIONS(7123),
    [anon_sym_yield] = ACTIONS(7121),
    [anon_sym_yield_BANG] = ACTIONS(7123),
    [anon_sym_lazy] = ACTIONS(7121),
    [anon_sym_assert] = ACTIONS(7121),
    [anon_sym_upcast] = ACTIONS(7121),
    [anon_sym_downcast] = ACTIONS(7121),
    [anon_sym_fixed] = ACTIONS(7121),
    [anon_sym_COLON_GT] = ACTIONS(7123),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7123),
    [anon_sym_for] = ACTIONS(7121),
    [anon_sym_while] = ACTIONS(7121),
    [anon_sym_while_BANG] = ACTIONS(7123),
    [anon_sym_if] = ACTIONS(7121),
    [anon_sym_fun] = ACTIONS(7121),
    [anon_sym_try] = ACTIONS(7121),
    [anon_sym_match] = ACTIONS(7121),
    [anon_sym_match_BANG] = ACTIONS(7123),
    [anon_sym_function] = ACTIONS(7121),
    [anon_sym_DOT_LBRACK] = ACTIONS(7123),
    [anon_sym_use] = ACTIONS(7121),
    [anon_sym_use_BANG] = ACTIONS(7123),
    [anon_sym_do_BANG] = ACTIONS(7123),
    [anon_sym_DOT_DOT] = ACTIONS(7123),
    [anon_sym_LPAREN2] = ACTIONS(7121),
    [anon_sym_PIPE_PIPE] = ACTIONS(7121),
    [anon_sym_AMP_AMP] = ACTIONS(7121),
    [aux_sym_char_token1] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7121),
    [anon_sym_DQUOTE] = ACTIONS(7121),
    [anon_sym_AT_DQUOTE] = ACTIONS(7123),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7123),
    [sym_bool] = ACTIONS(7121),
    [sym_unit] = ACTIONS(7123),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7121),
    [sym_op_identifier] = ACTIONS(7121),
    [anon_sym_PLUS] = ACTIONS(7121),
    [anon_sym_DASH] = ACTIONS(7121),
    [anon_sym_PLUS_DOT] = ACTIONS(7121),
    [anon_sym_DASH_DOT] = ACTIONS(7121),
    [anon_sym_PERCENT] = ACTIONS(7121),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7121),
    [anon_sym_TILDE] = ACTIONS(7123),
    [aux_sym_prefix_op_token1] = ACTIONS(7121),
    [aux_sym_infix_op_token1] = ACTIONS(7123),
    [aux_sym_infix_op_token2] = ACTIONS(7121),
    [aux_sym_infix_op_token3] = ACTIONS(7121),
    [anon_sym_BANG_EQ] = ACTIONS(7121),
    [anon_sym_COLON_EQ] = ACTIONS(7123),
    [anon_sym_DOLLAR] = ACTIONS(7121),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7121),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7121),
    [sym_int] = ACTIONS(7121),
    [sym_xint] = ACTIONS(7123),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7123),
    [sym__newline] = ACTIONS(7123),
    [sym__dedent] = ACTIONS(7123),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7123),
    [sym__tyapp_open] = ACTIONS(7123),
  },
  [STATE(5485)] = {
    [sym_block_comment] = STATE(5485),
    [sym_compiler_directive_decl] = STATE(5485),
    [sym_fsi_directive_decl] = STATE(5485),
    [sym_preproc_line] = STATE(5485),
    [sym_identifier] = ACTIONS(6936),
    [anon_sym_null] = ACTIONS(6936),
    [anon_sym_EQ] = ACTIONS(6936),
    [anon_sym_begin] = ACTIONS(6936),
    [anon_sym_COLON] = ACTIONS(6936),
    [anon_sym_do] = ACTIONS(6936),
    [anon_sym_let] = ACTIONS(6936),
    [anon_sym_let_BANG] = ACTIONS(6934),
    [anon_sym_LPAREN] = ACTIONS(6936),
    [anon_sym_COMMA] = ACTIONS(6934),
    [anon_sym_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_QMARK] = ACTIONS(6936),
    [anon_sym_COLON_COLON] = ACTIONS(6934),
    [anon_sym_AMP] = ACTIONS(6936),
    [anon_sym_LBRACK] = ACTIONS(6936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6934),
    [anon_sym_LBRACE] = ACTIONS(6936),
    [anon_sym_LT_AT] = ACTIONS(6936),
    [anon_sym_LT_AT_AT] = ACTIONS(6936),
    [anon_sym_CARET] = ACTIONS(6936),
    [anon_sym_or] = ACTIONS(6936),
    [anon_sym_DOT] = ACTIONS(6936),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6934),
    [anon_sym_new] = ACTIONS(6936),
    [anon_sym_return] = ACTIONS(6936),
    [anon_sym_return_BANG] = ACTIONS(6934),
    [anon_sym_yield] = ACTIONS(6936),
    [anon_sym_yield_BANG] = ACTIONS(6934),
    [anon_sym_lazy] = ACTIONS(6936),
    [anon_sym_assert] = ACTIONS(6936),
    [anon_sym_upcast] = ACTIONS(6936),
    [anon_sym_downcast] = ACTIONS(6936),
    [anon_sym_fixed] = ACTIONS(6936),
    [anon_sym_COLON_GT] = ACTIONS(6934),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6934),
    [anon_sym_for] = ACTIONS(6936),
    [anon_sym_while] = ACTIONS(6936),
    [anon_sym_while_BANG] = ACTIONS(6934),
    [anon_sym_if] = ACTIONS(6936),
    [anon_sym_fun] = ACTIONS(6936),
    [anon_sym_try] = ACTIONS(6936),
    [anon_sym_match] = ACTIONS(6936),
    [anon_sym_match_BANG] = ACTIONS(6934),
    [anon_sym_function] = ACTIONS(6936),
    [anon_sym_DOT_LBRACK] = ACTIONS(6934),
    [anon_sym_use] = ACTIONS(6936),
    [anon_sym_use_BANG] = ACTIONS(6934),
    [anon_sym_do_BANG] = ACTIONS(6934),
    [anon_sym_DOT_DOT] = ACTIONS(6934),
    [anon_sym_LPAREN2] = ACTIONS(6936),
    [anon_sym_PIPE_PIPE] = ACTIONS(6936),
    [anon_sym_AMP_AMP] = ACTIONS(6936),
    [aux_sym_char_token1] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6936),
    [anon_sym_DQUOTE] = ACTIONS(6936),
    [anon_sym_AT_DQUOTE] = ACTIONS(6934),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6934),
    [sym_bool] = ACTIONS(6936),
    [sym_unit] = ACTIONS(6934),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6936),
    [sym_op_identifier] = ACTIONS(6936),
    [anon_sym_PLUS] = ACTIONS(6936),
    [anon_sym_DASH] = ACTIONS(6936),
    [anon_sym_PLUS_DOT] = ACTIONS(6936),
    [anon_sym_DASH_DOT] = ACTIONS(6936),
    [anon_sym_PERCENT] = ACTIONS(6936),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6936),
    [anon_sym_TILDE] = ACTIONS(6934),
    [aux_sym_prefix_op_token1] = ACTIONS(6936),
    [aux_sym_infix_op_token1] = ACTIONS(6934),
    [aux_sym_infix_op_token2] = ACTIONS(6936),
    [aux_sym_infix_op_token3] = ACTIONS(6936),
    [anon_sym_BANG_EQ] = ACTIONS(6936),
    [anon_sym_COLON_EQ] = ACTIONS(6934),
    [anon_sym_DOLLAR] = ACTIONS(6936),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6936),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6936),
    [sym_int] = ACTIONS(6936),
    [sym_xint] = ACTIONS(6934),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6934),
    [sym__newline] = ACTIONS(6934),
    [sym__dedent] = ACTIONS(6934),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6934),
    [sym__tyapp_open] = ACTIONS(6934),
  },
  [STATE(5486)] = {
    [sym_block_comment] = STATE(5486),
    [sym_compiler_directive_decl] = STATE(5486),
    [sym_fsi_directive_decl] = STATE(5486),
    [sym_preproc_line] = STATE(5486),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5487)] = {
    [sym_block_comment] = STATE(5487),
    [sym_compiler_directive_decl] = STATE(5487),
    [sym_fsi_directive_decl] = STATE(5487),
    [sym_preproc_line] = STATE(5487),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9718),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__untyped_quoted_close] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5488)] = {
    [sym_block_comment] = STATE(5488),
    [sym_compiler_directive_decl] = STATE(5488),
    [sym_fsi_directive_decl] = STATE(5488),
    [sym_preproc_line] = STATE(5488),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5489)] = {
    [sym_block_comment] = STATE(5489),
    [sym_compiler_directive_decl] = STATE(5489),
    [sym_fsi_directive_decl] = STATE(5489),
    [sym_preproc_line] = STATE(5489),
    [sym_identifier] = ACTIONS(6940),
    [anon_sym_null] = ACTIONS(6940),
    [anon_sym_EQ] = ACTIONS(6940),
    [anon_sym_begin] = ACTIONS(6940),
    [anon_sym_COLON] = ACTIONS(6940),
    [anon_sym_do] = ACTIONS(6940),
    [anon_sym_let] = ACTIONS(6940),
    [anon_sym_let_BANG] = ACTIONS(6938),
    [anon_sym_LPAREN] = ACTIONS(6940),
    [anon_sym_COMMA] = ACTIONS(6938),
    [anon_sym_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_QMARK] = ACTIONS(6940),
    [anon_sym_COLON_COLON] = ACTIONS(6938),
    [anon_sym_AMP] = ACTIONS(6940),
    [anon_sym_LBRACK] = ACTIONS(6940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6938),
    [anon_sym_LBRACE] = ACTIONS(6940),
    [anon_sym_LT_AT] = ACTIONS(6940),
    [anon_sym_LT_AT_AT] = ACTIONS(6940),
    [anon_sym_CARET] = ACTIONS(6940),
    [anon_sym_or] = ACTIONS(6940),
    [anon_sym_DOT] = ACTIONS(6940),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6938),
    [anon_sym_new] = ACTIONS(6940),
    [anon_sym_return] = ACTIONS(6940),
    [anon_sym_return_BANG] = ACTIONS(6938),
    [anon_sym_yield] = ACTIONS(6940),
    [anon_sym_yield_BANG] = ACTIONS(6938),
    [anon_sym_lazy] = ACTIONS(6940),
    [anon_sym_assert] = ACTIONS(6940),
    [anon_sym_upcast] = ACTIONS(6940),
    [anon_sym_downcast] = ACTIONS(6940),
    [anon_sym_fixed] = ACTIONS(6940),
    [anon_sym_COLON_GT] = ACTIONS(6938),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6938),
    [anon_sym_for] = ACTIONS(6940),
    [anon_sym_while] = ACTIONS(6940),
    [anon_sym_while_BANG] = ACTIONS(6938),
    [anon_sym_if] = ACTIONS(6940),
    [anon_sym_fun] = ACTIONS(6940),
    [anon_sym_try] = ACTIONS(6940),
    [anon_sym_match] = ACTIONS(6940),
    [anon_sym_match_BANG] = ACTIONS(6938),
    [anon_sym_function] = ACTIONS(6940),
    [anon_sym_DOT_LBRACK] = ACTIONS(6938),
    [anon_sym_use] = ACTIONS(6940),
    [anon_sym_use_BANG] = ACTIONS(6938),
    [anon_sym_do_BANG] = ACTIONS(6938),
    [anon_sym_DOT_DOT] = ACTIONS(6938),
    [anon_sym_LPAREN2] = ACTIONS(6940),
    [anon_sym_PIPE_PIPE] = ACTIONS(6940),
    [anon_sym_AMP_AMP] = ACTIONS(6940),
    [aux_sym_char_token1] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6940),
    [anon_sym_DQUOTE] = ACTIONS(6940),
    [anon_sym_AT_DQUOTE] = ACTIONS(6938),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6938),
    [sym_bool] = ACTIONS(6940),
    [sym_unit] = ACTIONS(6938),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6940),
    [sym_op_identifier] = ACTIONS(6940),
    [anon_sym_PLUS] = ACTIONS(6940),
    [anon_sym_DASH] = ACTIONS(6940),
    [anon_sym_PLUS_DOT] = ACTIONS(6940),
    [anon_sym_DASH_DOT] = ACTIONS(6940),
    [anon_sym_PERCENT] = ACTIONS(6940),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6940),
    [anon_sym_TILDE] = ACTIONS(6938),
    [aux_sym_prefix_op_token1] = ACTIONS(6940),
    [aux_sym_infix_op_token1] = ACTIONS(6938),
    [aux_sym_infix_op_token2] = ACTIONS(6940),
    [aux_sym_infix_op_token3] = ACTIONS(6940),
    [anon_sym_BANG_EQ] = ACTIONS(6940),
    [anon_sym_COLON_EQ] = ACTIONS(6938),
    [anon_sym_DOLLAR] = ACTIONS(6940),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6940),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6940),
    [sym_int] = ACTIONS(6940),
    [sym_xint] = ACTIONS(6938),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6938),
    [sym__newline] = ACTIONS(6938),
    [sym__dedent] = ACTIONS(6938),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6938),
    [sym__tyapp_open] = ACTIONS(6938),
  },
  [STATE(5490)] = {
    [sym_block_comment] = STATE(5490),
    [sym_compiler_directive_decl] = STATE(5490),
    [sym_fsi_directive_decl] = STATE(5490),
    [sym_preproc_line] = STATE(5490),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5491)] = {
    [sym_block_comment] = STATE(5491),
    [sym_compiler_directive_decl] = STATE(5491),
    [sym_fsi_directive_decl] = STATE(5491),
    [sym_preproc_line] = STATE(5491),
    [sym_identifier] = ACTIONS(7015),
    [anon_sym_null] = ACTIONS(7015),
    [anon_sym_EQ] = ACTIONS(7015),
    [anon_sym_begin] = ACTIONS(7015),
    [anon_sym_COLON] = ACTIONS(7015),
    [anon_sym_do] = ACTIONS(7015),
    [anon_sym_let] = ACTIONS(7015),
    [anon_sym_let_BANG] = ACTIONS(7013),
    [anon_sym_LPAREN] = ACTIONS(7015),
    [anon_sym_COMMA] = ACTIONS(7013),
    [anon_sym_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_QMARK] = ACTIONS(7015),
    [anon_sym_COLON_COLON] = ACTIONS(7013),
    [anon_sym_AMP] = ACTIONS(7015),
    [anon_sym_LBRACK] = ACTIONS(7015),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7013),
    [anon_sym_LBRACE] = ACTIONS(7015),
    [anon_sym_LT_AT] = ACTIONS(7015),
    [anon_sym_LT_AT_AT] = ACTIONS(7015),
    [anon_sym_CARET] = ACTIONS(7015),
    [anon_sym_or] = ACTIONS(7015),
    [anon_sym_DOT] = ACTIONS(7015),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7013),
    [anon_sym_new] = ACTIONS(7015),
    [anon_sym_return] = ACTIONS(7015),
    [anon_sym_return_BANG] = ACTIONS(7013),
    [anon_sym_yield] = ACTIONS(7015),
    [anon_sym_yield_BANG] = ACTIONS(7013),
    [anon_sym_lazy] = ACTIONS(7015),
    [anon_sym_assert] = ACTIONS(7015),
    [anon_sym_upcast] = ACTIONS(7015),
    [anon_sym_downcast] = ACTIONS(7015),
    [anon_sym_fixed] = ACTIONS(7015),
    [anon_sym_COLON_GT] = ACTIONS(7013),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7013),
    [anon_sym_for] = ACTIONS(7015),
    [anon_sym_while] = ACTIONS(7015),
    [anon_sym_while_BANG] = ACTIONS(7013),
    [anon_sym_if] = ACTIONS(7015),
    [anon_sym_fun] = ACTIONS(7015),
    [anon_sym_try] = ACTIONS(7015),
    [anon_sym_match] = ACTIONS(7015),
    [anon_sym_match_BANG] = ACTIONS(7013),
    [anon_sym_function] = ACTIONS(7015),
    [anon_sym_DOT_LBRACK] = ACTIONS(7013),
    [anon_sym_use] = ACTIONS(7015),
    [anon_sym_use_BANG] = ACTIONS(7013),
    [anon_sym_do_BANG] = ACTIONS(7013),
    [anon_sym_DOT_DOT] = ACTIONS(7013),
    [anon_sym_LPAREN2] = ACTIONS(7015),
    [anon_sym_PIPE_PIPE] = ACTIONS(7015),
    [anon_sym_AMP_AMP] = ACTIONS(7015),
    [aux_sym_char_token1] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7015),
    [anon_sym_DQUOTE] = ACTIONS(7015),
    [anon_sym_AT_DQUOTE] = ACTIONS(7013),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7013),
    [sym_bool] = ACTIONS(7015),
    [sym_unit] = ACTIONS(7013),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7015),
    [sym_op_identifier] = ACTIONS(7015),
    [anon_sym_PLUS] = ACTIONS(7015),
    [anon_sym_DASH] = ACTIONS(7015),
    [anon_sym_PLUS_DOT] = ACTIONS(7015),
    [anon_sym_DASH_DOT] = ACTIONS(7015),
    [anon_sym_PERCENT] = ACTIONS(7015),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7015),
    [anon_sym_TILDE] = ACTIONS(7013),
    [aux_sym_prefix_op_token1] = ACTIONS(7015),
    [aux_sym_infix_op_token1] = ACTIONS(7013),
    [aux_sym_infix_op_token2] = ACTIONS(7015),
    [aux_sym_infix_op_token3] = ACTIONS(7015),
    [anon_sym_BANG_EQ] = ACTIONS(7015),
    [anon_sym_COLON_EQ] = ACTIONS(7013),
    [anon_sym_DOLLAR] = ACTIONS(7015),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7015),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7015),
    [sym_int] = ACTIONS(7015),
    [sym_xint] = ACTIONS(7013),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7013),
    [sym__newline] = ACTIONS(7013),
    [sym__dedent] = ACTIONS(7013),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7013),
    [sym__tyapp_open] = ACTIONS(7013),
  },
  [STATE(5492)] = {
    [sym_block_comment] = STATE(5492),
    [sym_compiler_directive_decl] = STATE(5492),
    [sym_fsi_directive_decl] = STATE(5492),
    [sym_preproc_line] = STATE(5492),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_GT] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5493)] = {
    [sym_block_comment] = STATE(5493),
    [sym_compiler_directive_decl] = STATE(5493),
    [sym_fsi_directive_decl] = STATE(5493),
    [sym_preproc_line] = STATE(5493),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9720),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__multi_dollar_interp_end] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5494)] = {
    [sym_block_comment] = STATE(5494),
    [sym_compiler_directive_decl] = STATE(5494),
    [sym_fsi_directive_decl] = STATE(5494),
    [sym_preproc_line] = STATE(5494),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [anon_sym_POUNDendif] = ACTIONS(7139),
    [anon_sym_POUNDelse] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(5495)] = {
    [sym_block_comment] = STATE(5495),
    [sym_compiler_directive_decl] = STATE(5495),
    [sym_fsi_directive_decl] = STATE(5495),
    [sym_preproc_line] = STATE(5495),
    [sym_identifier] = ACTIONS(7031),
    [anon_sym_null] = ACTIONS(7031),
    [anon_sym_EQ] = ACTIONS(7031),
    [anon_sym_begin] = ACTIONS(7031),
    [anon_sym_COLON] = ACTIONS(7031),
    [anon_sym_do] = ACTIONS(7031),
    [anon_sym_let] = ACTIONS(7031),
    [anon_sym_let_BANG] = ACTIONS(7029),
    [anon_sym_LPAREN] = ACTIONS(7031),
    [anon_sym_COMMA] = ACTIONS(7029),
    [anon_sym_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_QMARK] = ACTIONS(7031),
    [anon_sym_COLON_COLON] = ACTIONS(7029),
    [anon_sym_AMP] = ACTIONS(7031),
    [anon_sym_LBRACK] = ACTIONS(7031),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7029),
    [anon_sym_LBRACE] = ACTIONS(7031),
    [anon_sym_LT_AT] = ACTIONS(7031),
    [anon_sym_LT_AT_AT] = ACTIONS(7031),
    [anon_sym_CARET] = ACTIONS(7031),
    [anon_sym_or] = ACTIONS(7031),
    [anon_sym_DOT] = ACTIONS(7031),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7029),
    [anon_sym_new] = ACTIONS(7031),
    [anon_sym_return] = ACTIONS(7031),
    [anon_sym_return_BANG] = ACTIONS(7029),
    [anon_sym_yield] = ACTIONS(7031),
    [anon_sym_yield_BANG] = ACTIONS(7029),
    [anon_sym_lazy] = ACTIONS(7031),
    [anon_sym_assert] = ACTIONS(7031),
    [anon_sym_upcast] = ACTIONS(7031),
    [anon_sym_downcast] = ACTIONS(7031),
    [anon_sym_fixed] = ACTIONS(7031),
    [anon_sym_COLON_GT] = ACTIONS(7029),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7029),
    [anon_sym_for] = ACTIONS(7031),
    [anon_sym_while] = ACTIONS(7031),
    [anon_sym_while_BANG] = ACTIONS(7029),
    [anon_sym_if] = ACTIONS(7031),
    [anon_sym_fun] = ACTIONS(7031),
    [anon_sym_try] = ACTIONS(7031),
    [anon_sym_match] = ACTIONS(7031),
    [anon_sym_match_BANG] = ACTIONS(7029),
    [anon_sym_function] = ACTIONS(7031),
    [anon_sym_DOT_LBRACK] = ACTIONS(7029),
    [anon_sym_use] = ACTIONS(7031),
    [anon_sym_use_BANG] = ACTIONS(7029),
    [anon_sym_do_BANG] = ACTIONS(7029),
    [anon_sym_DOT_DOT] = ACTIONS(7029),
    [anon_sym_LPAREN2] = ACTIONS(7031),
    [anon_sym_PIPE_PIPE] = ACTIONS(7031),
    [anon_sym_AMP_AMP] = ACTIONS(7031),
    [aux_sym_char_token1] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7031),
    [anon_sym_DQUOTE] = ACTIONS(7031),
    [anon_sym_AT_DQUOTE] = ACTIONS(7029),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7029),
    [sym_bool] = ACTIONS(7031),
    [sym_unit] = ACTIONS(7029),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7031),
    [sym_op_identifier] = ACTIONS(7031),
    [anon_sym_PLUS] = ACTIONS(7031),
    [anon_sym_DASH] = ACTIONS(7031),
    [anon_sym_PLUS_DOT] = ACTIONS(7031),
    [anon_sym_DASH_DOT] = ACTIONS(7031),
    [anon_sym_PERCENT] = ACTIONS(7031),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7031),
    [anon_sym_TILDE] = ACTIONS(7029),
    [aux_sym_prefix_op_token1] = ACTIONS(7031),
    [aux_sym_infix_op_token1] = ACTIONS(7029),
    [aux_sym_infix_op_token2] = ACTIONS(7031),
    [aux_sym_infix_op_token3] = ACTIONS(7031),
    [anon_sym_BANG_EQ] = ACTIONS(7031),
    [anon_sym_COLON_EQ] = ACTIONS(7029),
    [anon_sym_DOLLAR] = ACTIONS(7031),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7031),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7031),
    [sym_int] = ACTIONS(7031),
    [sym_xint] = ACTIONS(7029),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7029),
    [sym__newline] = ACTIONS(7029),
    [sym__dedent] = ACTIONS(7029),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7029),
    [sym__tyapp_open] = ACTIONS(7029),
  },
  [STATE(5496)] = {
    [sym_block_comment] = STATE(5496),
    [sym_compiler_directive_decl] = STATE(5496),
    [sym_fsi_directive_decl] = STATE(5496),
    [sym_preproc_line] = STATE(5496),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_as] = ACTIONS(7137),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_with] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(5497)] = {
    [sym_block_comment] = STATE(5497),
    [sym_compiler_directive_decl] = STATE(5497),
    [sym_fsi_directive_decl] = STATE(5497),
    [sym_preproc_line] = STATE(5497),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_GT] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9722),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5498)] = {
    [sym_block_comment] = STATE(5498),
    [sym_compiler_directive_decl] = STATE(5498),
    [sym_fsi_directive_decl] = STATE(5498),
    [sym_preproc_line] = STATE(5498),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5499)] = {
    [sym_block_comment] = STATE(5499),
    [sym_compiler_directive_decl] = STATE(5499),
    [sym_fsi_directive_decl] = STATE(5499),
    [sym_preproc_line] = STATE(5499),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__dedent] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
    [sym__in] = ACTIONS(7143),
  },
  [STATE(5500)] = {
    [sym_block_comment] = STATE(5500),
    [sym_compiler_directive_decl] = STATE(5500),
    [sym_fsi_directive_decl] = STATE(5500),
    [sym_preproc_line] = STATE(5500),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__dedent] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
    [sym__in] = ACTIONS(7147),
  },
  [STATE(5501)] = {
    [sym_block_comment] = STATE(5501),
    [sym_compiler_directive_decl] = STATE(5501),
    [sym_fsi_directive_decl] = STATE(5501),
    [sym_preproc_line] = STATE(5501),
    [sym_identifier] = ACTIONS(7047),
    [anon_sym_null] = ACTIONS(7047),
    [anon_sym_EQ] = ACTIONS(7047),
    [anon_sym_begin] = ACTIONS(7047),
    [anon_sym_COLON] = ACTIONS(7047),
    [anon_sym_do] = ACTIONS(7047),
    [anon_sym_let] = ACTIONS(7047),
    [anon_sym_let_BANG] = ACTIONS(7045),
    [anon_sym_LPAREN] = ACTIONS(7047),
    [anon_sym_COMMA] = ACTIONS(7045),
    [anon_sym_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_QMARK] = ACTIONS(7047),
    [anon_sym_COLON_COLON] = ACTIONS(7045),
    [anon_sym_AMP] = ACTIONS(7047),
    [anon_sym_LBRACK] = ACTIONS(7047),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7045),
    [anon_sym_LBRACE] = ACTIONS(7047),
    [anon_sym_LT_AT] = ACTIONS(7047),
    [anon_sym_LT_AT_AT] = ACTIONS(7047),
    [anon_sym_CARET] = ACTIONS(7047),
    [anon_sym_or] = ACTIONS(7047),
    [anon_sym_DOT] = ACTIONS(7047),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7045),
    [anon_sym_new] = ACTIONS(7047),
    [anon_sym_return] = ACTIONS(7047),
    [anon_sym_return_BANG] = ACTIONS(7045),
    [anon_sym_yield] = ACTIONS(7047),
    [anon_sym_yield_BANG] = ACTIONS(7045),
    [anon_sym_lazy] = ACTIONS(7047),
    [anon_sym_assert] = ACTIONS(7047),
    [anon_sym_upcast] = ACTIONS(7047),
    [anon_sym_downcast] = ACTIONS(7047),
    [anon_sym_fixed] = ACTIONS(7047),
    [anon_sym_COLON_GT] = ACTIONS(7045),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7045),
    [anon_sym_for] = ACTIONS(7047),
    [anon_sym_while] = ACTIONS(7047),
    [anon_sym_while_BANG] = ACTIONS(7045),
    [anon_sym_if] = ACTIONS(7047),
    [anon_sym_fun] = ACTIONS(7047),
    [anon_sym_try] = ACTIONS(7047),
    [anon_sym_match] = ACTIONS(7047),
    [anon_sym_match_BANG] = ACTIONS(7045),
    [anon_sym_function] = ACTIONS(7047),
    [anon_sym_DOT_LBRACK] = ACTIONS(7045),
    [anon_sym_use] = ACTIONS(7047),
    [anon_sym_use_BANG] = ACTIONS(7045),
    [anon_sym_do_BANG] = ACTIONS(7045),
    [anon_sym_DOT_DOT] = ACTIONS(7045),
    [anon_sym_LPAREN2] = ACTIONS(7047),
    [anon_sym_PIPE_PIPE] = ACTIONS(7047),
    [anon_sym_AMP_AMP] = ACTIONS(7047),
    [aux_sym_char_token1] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7047),
    [anon_sym_DQUOTE] = ACTIONS(7047),
    [anon_sym_AT_DQUOTE] = ACTIONS(7045),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7045),
    [sym_bool] = ACTIONS(7047),
    [sym_unit] = ACTIONS(7045),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7047),
    [sym_op_identifier] = ACTIONS(7047),
    [anon_sym_PLUS] = ACTIONS(7047),
    [anon_sym_DASH] = ACTIONS(7047),
    [anon_sym_PLUS_DOT] = ACTIONS(7047),
    [anon_sym_DASH_DOT] = ACTIONS(7047),
    [anon_sym_PERCENT] = ACTIONS(7047),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7047),
    [anon_sym_TILDE] = ACTIONS(7045),
    [aux_sym_prefix_op_token1] = ACTIONS(7047),
    [aux_sym_infix_op_token1] = ACTIONS(7045),
    [aux_sym_infix_op_token2] = ACTIONS(7047),
    [aux_sym_infix_op_token3] = ACTIONS(7047),
    [anon_sym_BANG_EQ] = ACTIONS(7047),
    [anon_sym_COLON_EQ] = ACTIONS(7045),
    [anon_sym_DOLLAR] = ACTIONS(7047),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7047),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7047),
    [sym_int] = ACTIONS(7047),
    [sym_xint] = ACTIONS(7045),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7045),
    [sym__newline] = ACTIONS(7045),
    [sym__dedent] = ACTIONS(7045),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7045),
    [sym__tyapp_open] = ACTIONS(7045),
  },
  [STATE(5502)] = {
    [sym_block_comment] = STATE(5502),
    [sym_compiler_directive_decl] = STATE(5502),
    [sym_fsi_directive_decl] = STATE(5502),
    [sym_preproc_line] = STATE(5502),
    [aux_sym_sequential_expression_repeat1] = STATE(5403),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_DOT_DOT] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5503)] = {
    [sym_block_comment] = STATE(5503),
    [sym_compiler_directive_decl] = STATE(5503),
    [sym_fsi_directive_decl] = STATE(5503),
    [sym_preproc_line] = STATE(5503),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__dedent] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
    [sym__in] = ACTIONS(7157),
  },
  [STATE(5504)] = {
    [sym_block_comment] = STATE(5504),
    [sym_compiler_directive_decl] = STATE(5504),
    [sym_fsi_directive_decl] = STATE(5504),
    [sym_preproc_line] = STATE(5504),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__dedent] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
    [sym__in] = ACTIONS(7161),
  },
  [STATE(5505)] = {
    [sym_block_comment] = STATE(5505),
    [sym_compiler_directive_decl] = STATE(5505),
    [sym_fsi_directive_decl] = STATE(5505),
    [sym_preproc_line] = STATE(5505),
    [sym_identifier] = ACTIONS(7051),
    [anon_sym_null] = ACTIONS(7051),
    [anon_sym_EQ] = ACTIONS(7051),
    [anon_sym_begin] = ACTIONS(7051),
    [anon_sym_COLON] = ACTIONS(7051),
    [anon_sym_do] = ACTIONS(7051),
    [anon_sym_let] = ACTIONS(7051),
    [anon_sym_let_BANG] = ACTIONS(7049),
    [anon_sym_LPAREN] = ACTIONS(7051),
    [anon_sym_COMMA] = ACTIONS(7049),
    [anon_sym_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_QMARK] = ACTIONS(7051),
    [anon_sym_COLON_COLON] = ACTIONS(7049),
    [anon_sym_AMP] = ACTIONS(7051),
    [anon_sym_LBRACK] = ACTIONS(7051),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7049),
    [anon_sym_LBRACE] = ACTIONS(7051),
    [anon_sym_LT_AT] = ACTIONS(7051),
    [anon_sym_LT_AT_AT] = ACTIONS(7051),
    [anon_sym_CARET] = ACTIONS(7051),
    [anon_sym_or] = ACTIONS(7051),
    [anon_sym_DOT] = ACTIONS(7051),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7049),
    [anon_sym_new] = ACTIONS(7051),
    [anon_sym_return] = ACTIONS(7051),
    [anon_sym_return_BANG] = ACTIONS(7049),
    [anon_sym_yield] = ACTIONS(7051),
    [anon_sym_yield_BANG] = ACTIONS(7049),
    [anon_sym_lazy] = ACTIONS(7051),
    [anon_sym_assert] = ACTIONS(7051),
    [anon_sym_upcast] = ACTIONS(7051),
    [anon_sym_downcast] = ACTIONS(7051),
    [anon_sym_fixed] = ACTIONS(7051),
    [anon_sym_COLON_GT] = ACTIONS(7049),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7049),
    [anon_sym_for] = ACTIONS(7051),
    [anon_sym_while] = ACTIONS(7051),
    [anon_sym_while_BANG] = ACTIONS(7049),
    [anon_sym_if] = ACTIONS(7051),
    [anon_sym_fun] = ACTIONS(7051),
    [anon_sym_try] = ACTIONS(7051),
    [anon_sym_match] = ACTIONS(7051),
    [anon_sym_match_BANG] = ACTIONS(7049),
    [anon_sym_function] = ACTIONS(7051),
    [anon_sym_DOT_LBRACK] = ACTIONS(7049),
    [anon_sym_use] = ACTIONS(7051),
    [anon_sym_use_BANG] = ACTIONS(7049),
    [anon_sym_do_BANG] = ACTIONS(7049),
    [anon_sym_DOT_DOT] = ACTIONS(7049),
    [anon_sym_LPAREN2] = ACTIONS(7051),
    [anon_sym_PIPE_PIPE] = ACTIONS(7051),
    [anon_sym_AMP_AMP] = ACTIONS(7051),
    [aux_sym_char_token1] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7051),
    [anon_sym_DQUOTE] = ACTIONS(7051),
    [anon_sym_AT_DQUOTE] = ACTIONS(7049),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7049),
    [sym_bool] = ACTIONS(7051),
    [sym_unit] = ACTIONS(7049),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7051),
    [sym_op_identifier] = ACTIONS(7051),
    [anon_sym_PLUS] = ACTIONS(7051),
    [anon_sym_DASH] = ACTIONS(7051),
    [anon_sym_PLUS_DOT] = ACTIONS(7051),
    [anon_sym_DASH_DOT] = ACTIONS(7051),
    [anon_sym_PERCENT] = ACTIONS(7051),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7051),
    [anon_sym_TILDE] = ACTIONS(7049),
    [aux_sym_prefix_op_token1] = ACTIONS(7051),
    [aux_sym_infix_op_token1] = ACTIONS(7049),
    [aux_sym_infix_op_token2] = ACTIONS(7051),
    [aux_sym_infix_op_token3] = ACTIONS(7051),
    [anon_sym_BANG_EQ] = ACTIONS(7051),
    [anon_sym_COLON_EQ] = ACTIONS(7049),
    [anon_sym_DOLLAR] = ACTIONS(7051),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7051),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7051),
    [sym_int] = ACTIONS(7051),
    [sym_xint] = ACTIONS(7049),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7049),
    [sym__newline] = ACTIONS(7049),
    [sym__dedent] = ACTIONS(7049),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7049),
    [sym__tyapp_open] = ACTIONS(7049),
  },
  [STATE(5506)] = {
    [sym_block_comment] = STATE(5506),
    [sym_compiler_directive_decl] = STATE(5506),
    [sym_fsi_directive_decl] = STATE(5506),
    [sym_preproc_line] = STATE(5506),
    [sym_identifier] = ACTIONS(7063),
    [anon_sym_null] = ACTIONS(7063),
    [anon_sym_EQ] = ACTIONS(7063),
    [anon_sym_begin] = ACTIONS(7063),
    [anon_sym_COLON] = ACTIONS(7063),
    [anon_sym_do] = ACTIONS(7063),
    [anon_sym_let] = ACTIONS(7063),
    [anon_sym_let_BANG] = ACTIONS(7061),
    [anon_sym_LPAREN] = ACTIONS(7063),
    [anon_sym_COMMA] = ACTIONS(7061),
    [anon_sym_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_QMARK] = ACTIONS(7063),
    [anon_sym_COLON_COLON] = ACTIONS(7061),
    [anon_sym_AMP] = ACTIONS(7063),
    [anon_sym_LBRACK] = ACTIONS(7063),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7061),
    [anon_sym_LBRACE] = ACTIONS(7063),
    [anon_sym_LT_AT] = ACTIONS(7063),
    [anon_sym_LT_AT_AT] = ACTIONS(7063),
    [anon_sym_CARET] = ACTIONS(7063),
    [anon_sym_or] = ACTIONS(7063),
    [anon_sym_DOT] = ACTIONS(7063),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7061),
    [anon_sym_new] = ACTIONS(7063),
    [anon_sym_return] = ACTIONS(7063),
    [anon_sym_return_BANG] = ACTIONS(7061),
    [anon_sym_yield] = ACTIONS(7063),
    [anon_sym_yield_BANG] = ACTIONS(7061),
    [anon_sym_lazy] = ACTIONS(7063),
    [anon_sym_assert] = ACTIONS(7063),
    [anon_sym_upcast] = ACTIONS(7063),
    [anon_sym_downcast] = ACTIONS(7063),
    [anon_sym_fixed] = ACTIONS(7063),
    [anon_sym_COLON_GT] = ACTIONS(7061),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7061),
    [anon_sym_for] = ACTIONS(7063),
    [anon_sym_while] = ACTIONS(7063),
    [anon_sym_while_BANG] = ACTIONS(7061),
    [anon_sym_if] = ACTIONS(7063),
    [anon_sym_fun] = ACTIONS(7063),
    [anon_sym_try] = ACTIONS(7063),
    [anon_sym_match] = ACTIONS(7063),
    [anon_sym_match_BANG] = ACTIONS(7061),
    [anon_sym_function] = ACTIONS(7063),
    [anon_sym_DOT_LBRACK] = ACTIONS(7061),
    [anon_sym_use] = ACTIONS(7063),
    [anon_sym_use_BANG] = ACTIONS(7061),
    [anon_sym_do_BANG] = ACTIONS(7061),
    [anon_sym_DOT_DOT] = ACTIONS(7061),
    [anon_sym_LPAREN2] = ACTIONS(7063),
    [anon_sym_PIPE_PIPE] = ACTIONS(7063),
    [anon_sym_AMP_AMP] = ACTIONS(7063),
    [aux_sym_char_token1] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7063),
    [anon_sym_DQUOTE] = ACTIONS(7063),
    [anon_sym_AT_DQUOTE] = ACTIONS(7061),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7061),
    [sym_bool] = ACTIONS(7063),
    [sym_unit] = ACTIONS(7061),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7063),
    [sym_op_identifier] = ACTIONS(7063),
    [anon_sym_PLUS] = ACTIONS(7063),
    [anon_sym_DASH] = ACTIONS(7063),
    [anon_sym_PLUS_DOT] = ACTIONS(7063),
    [anon_sym_DASH_DOT] = ACTIONS(7063),
    [anon_sym_PERCENT] = ACTIONS(7063),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7063),
    [anon_sym_TILDE] = ACTIONS(7061),
    [aux_sym_prefix_op_token1] = ACTIONS(7063),
    [aux_sym_infix_op_token1] = ACTIONS(7061),
    [aux_sym_infix_op_token2] = ACTIONS(7063),
    [aux_sym_infix_op_token3] = ACTIONS(7063),
    [anon_sym_BANG_EQ] = ACTIONS(7063),
    [anon_sym_COLON_EQ] = ACTIONS(7061),
    [anon_sym_DOLLAR] = ACTIONS(7063),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7063),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7063),
    [sym_int] = ACTIONS(7063),
    [sym_xint] = ACTIONS(7061),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7061),
    [sym__newline] = ACTIONS(7061),
    [sym__dedent] = ACTIONS(7061),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7061),
    [sym__tyapp_open] = ACTIONS(7061),
  },
  [STATE(5507)] = {
    [sym_block_comment] = STATE(5507),
    [sym_compiler_directive_decl] = STATE(5507),
    [sym_fsi_directive_decl] = STATE(5507),
    [sym_preproc_line] = STATE(5507),
    [sym_identifier] = ACTIONS(7067),
    [anon_sym_null] = ACTIONS(7067),
    [anon_sym_EQ] = ACTIONS(7067),
    [anon_sym_begin] = ACTIONS(7067),
    [anon_sym_COLON] = ACTIONS(7067),
    [anon_sym_do] = ACTIONS(7067),
    [anon_sym_let] = ACTIONS(7067),
    [anon_sym_let_BANG] = ACTIONS(7065),
    [anon_sym_LPAREN] = ACTIONS(7067),
    [anon_sym_COMMA] = ACTIONS(7065),
    [anon_sym_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_QMARK] = ACTIONS(7067),
    [anon_sym_COLON_COLON] = ACTIONS(7065),
    [anon_sym_AMP] = ACTIONS(7067),
    [anon_sym_LBRACK] = ACTIONS(7067),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7065),
    [anon_sym_LBRACE] = ACTIONS(7067),
    [anon_sym_LT_AT] = ACTIONS(7067),
    [anon_sym_LT_AT_AT] = ACTIONS(7067),
    [anon_sym_CARET] = ACTIONS(7067),
    [anon_sym_or] = ACTIONS(7067),
    [anon_sym_DOT] = ACTIONS(7067),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7065),
    [anon_sym_new] = ACTIONS(7067),
    [anon_sym_return] = ACTIONS(7067),
    [anon_sym_return_BANG] = ACTIONS(7065),
    [anon_sym_yield] = ACTIONS(7067),
    [anon_sym_yield_BANG] = ACTIONS(7065),
    [anon_sym_lazy] = ACTIONS(7067),
    [anon_sym_assert] = ACTIONS(7067),
    [anon_sym_upcast] = ACTIONS(7067),
    [anon_sym_downcast] = ACTIONS(7067),
    [anon_sym_fixed] = ACTIONS(7067),
    [anon_sym_COLON_GT] = ACTIONS(7065),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7065),
    [anon_sym_for] = ACTIONS(7067),
    [anon_sym_while] = ACTIONS(7067),
    [anon_sym_while_BANG] = ACTIONS(7065),
    [anon_sym_if] = ACTIONS(7067),
    [anon_sym_fun] = ACTIONS(7067),
    [anon_sym_try] = ACTIONS(7067),
    [anon_sym_match] = ACTIONS(7067),
    [anon_sym_match_BANG] = ACTIONS(7065),
    [anon_sym_function] = ACTIONS(7067),
    [anon_sym_DOT_LBRACK] = ACTIONS(7065),
    [anon_sym_use] = ACTIONS(7067),
    [anon_sym_use_BANG] = ACTIONS(7065),
    [anon_sym_do_BANG] = ACTIONS(7065),
    [anon_sym_DOT_DOT] = ACTIONS(7065),
    [anon_sym_LPAREN2] = ACTIONS(7067),
    [anon_sym_PIPE_PIPE] = ACTIONS(7067),
    [anon_sym_AMP_AMP] = ACTIONS(7067),
    [aux_sym_char_token1] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7067),
    [anon_sym_DQUOTE] = ACTIONS(7067),
    [anon_sym_AT_DQUOTE] = ACTIONS(7065),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7065),
    [sym_bool] = ACTIONS(7067),
    [sym_unit] = ACTIONS(7065),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7067),
    [sym_op_identifier] = ACTIONS(7067),
    [anon_sym_PLUS] = ACTIONS(7067),
    [anon_sym_DASH] = ACTIONS(7067),
    [anon_sym_PLUS_DOT] = ACTIONS(7067),
    [anon_sym_DASH_DOT] = ACTIONS(7067),
    [anon_sym_PERCENT] = ACTIONS(7067),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7067),
    [anon_sym_TILDE] = ACTIONS(7065),
    [aux_sym_prefix_op_token1] = ACTIONS(7067),
    [aux_sym_infix_op_token1] = ACTIONS(7065),
    [aux_sym_infix_op_token2] = ACTIONS(7067),
    [aux_sym_infix_op_token3] = ACTIONS(7067),
    [anon_sym_BANG_EQ] = ACTIONS(7067),
    [anon_sym_COLON_EQ] = ACTIONS(7065),
    [anon_sym_DOLLAR] = ACTIONS(7067),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7067),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7067),
    [sym_int] = ACTIONS(7067),
    [sym_xint] = ACTIONS(7065),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7065),
    [sym__newline] = ACTIONS(7065),
    [sym__dedent] = ACTIONS(7065),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7065),
    [sym__tyapp_open] = ACTIONS(7065),
  },
  [STATE(5508)] = {
    [sym_block_comment] = STATE(5508),
    [sym_compiler_directive_decl] = STATE(5508),
    [sym_fsi_directive_decl] = STATE(5508),
    [sym_preproc_line] = STATE(5508),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_GT] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5509)] = {
    [sym_block_comment] = STATE(5509),
    [sym_compiler_directive_decl] = STATE(5509),
    [sym_fsi_directive_decl] = STATE(5509),
    [sym_preproc_line] = STATE(5509),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_GT] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5510)] = {
    [sym_block_comment] = STATE(5510),
    [sym_compiler_directive_decl] = STATE(5510),
    [sym_fsi_directive_decl] = STATE(5510),
    [sym_preproc_line] = STATE(5510),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_DOT_DOT] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5511)] = {
    [sym_block_comment] = STATE(5511),
    [sym_compiler_directive_decl] = STATE(5511),
    [sym_fsi_directive_decl] = STATE(5511),
    [sym_preproc_line] = STATE(5511),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9724),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_DOT_DOT] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5512)] = {
    [sym_block_comment] = STATE(5512),
    [sym_compiler_directive_decl] = STATE(5512),
    [sym_fsi_directive_decl] = STATE(5512),
    [sym_preproc_line] = STATE(5512),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__untyped_quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5513)] = {
    [sym_block_comment] = STATE(5513),
    [sym_compiler_directive_decl] = STATE(5513),
    [sym_fsi_directive_decl] = STATE(5513),
    [sym_preproc_line] = STATE(5513),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_DOT_DOT] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5514)] = {
    [sym_block_comment] = STATE(5514),
    [sym_compiler_directive_decl] = STATE(5514),
    [sym_fsi_directive_decl] = STATE(5514),
    [sym_preproc_line] = STATE(5514),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_DOT_DOT] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5515)] = {
    [sym_block_comment] = STATE(5515),
    [sym_compiler_directive_decl] = STATE(5515),
    [sym_fsi_directive_decl] = STATE(5515),
    [sym_preproc_line] = STATE(5515),
    [sym_identifier] = ACTIONS(7125),
    [anon_sym_null] = ACTIONS(7125),
    [anon_sym_EQ] = ACTIONS(7125),
    [anon_sym_begin] = ACTIONS(7125),
    [anon_sym_COLON] = ACTIONS(7125),
    [anon_sym_do] = ACTIONS(7125),
    [anon_sym_let] = ACTIONS(7125),
    [anon_sym_let_BANG] = ACTIONS(7127),
    [anon_sym_LPAREN] = ACTIONS(7125),
    [anon_sym_COMMA] = ACTIONS(7127),
    [anon_sym_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_QMARK] = ACTIONS(7125),
    [anon_sym_COLON_COLON] = ACTIONS(7127),
    [anon_sym_AMP] = ACTIONS(7125),
    [anon_sym_LBRACK] = ACTIONS(7125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7127),
    [anon_sym_LBRACE] = ACTIONS(7125),
    [anon_sym_LT_AT] = ACTIONS(7125),
    [anon_sym_LT_AT_AT] = ACTIONS(7125),
    [anon_sym_CARET] = ACTIONS(7125),
    [anon_sym_or] = ACTIONS(7125),
    [anon_sym_DOT] = ACTIONS(7125),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7127),
    [anon_sym_new] = ACTIONS(7125),
    [anon_sym_return] = ACTIONS(7125),
    [anon_sym_return_BANG] = ACTIONS(7127),
    [anon_sym_yield] = ACTIONS(7125),
    [anon_sym_yield_BANG] = ACTIONS(7127),
    [anon_sym_lazy] = ACTIONS(7125),
    [anon_sym_assert] = ACTIONS(7125),
    [anon_sym_upcast] = ACTIONS(7125),
    [anon_sym_downcast] = ACTIONS(7125),
    [anon_sym_fixed] = ACTIONS(7125),
    [anon_sym_COLON_GT] = ACTIONS(7127),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7127),
    [anon_sym_for] = ACTIONS(7125),
    [anon_sym_while] = ACTIONS(7125),
    [anon_sym_while_BANG] = ACTIONS(7127),
    [anon_sym_if] = ACTIONS(7125),
    [anon_sym_fun] = ACTIONS(7125),
    [anon_sym_try] = ACTIONS(7125),
    [anon_sym_match] = ACTIONS(7125),
    [anon_sym_match_BANG] = ACTIONS(7127),
    [anon_sym_function] = ACTIONS(7125),
    [anon_sym_DOT_LBRACK] = ACTIONS(7127),
    [anon_sym_use] = ACTIONS(7125),
    [anon_sym_use_BANG] = ACTIONS(7127),
    [anon_sym_do_BANG] = ACTIONS(7127),
    [anon_sym_DOT_DOT] = ACTIONS(7127),
    [anon_sym_LPAREN2] = ACTIONS(7125),
    [anon_sym_PIPE_PIPE] = ACTIONS(7125),
    [anon_sym_AMP_AMP] = ACTIONS(7125),
    [aux_sym_char_token1] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7125),
    [anon_sym_DQUOTE] = ACTIONS(7125),
    [anon_sym_AT_DQUOTE] = ACTIONS(7127),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7127),
    [sym_bool] = ACTIONS(7125),
    [sym_unit] = ACTIONS(7127),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7125),
    [sym_op_identifier] = ACTIONS(7125),
    [anon_sym_PLUS] = ACTIONS(7125),
    [anon_sym_DASH] = ACTIONS(7125),
    [anon_sym_PLUS_DOT] = ACTIONS(7125),
    [anon_sym_DASH_DOT] = ACTIONS(7125),
    [anon_sym_PERCENT] = ACTIONS(7125),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7125),
    [anon_sym_TILDE] = ACTIONS(7127),
    [aux_sym_prefix_op_token1] = ACTIONS(7125),
    [aux_sym_infix_op_token1] = ACTIONS(7127),
    [aux_sym_infix_op_token2] = ACTIONS(7125),
    [aux_sym_infix_op_token3] = ACTIONS(7125),
    [anon_sym_BANG_EQ] = ACTIONS(7125),
    [anon_sym_COLON_EQ] = ACTIONS(7127),
    [anon_sym_DOLLAR] = ACTIONS(7125),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7125),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7125),
    [sym_int] = ACTIONS(7125),
    [sym_xint] = ACTIONS(7127),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7127),
    [sym__newline] = ACTIONS(7127),
    [sym__dedent] = ACTIONS(7127),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7127),
    [sym__tyapp_open] = ACTIONS(7127),
  },
  [STATE(5516)] = {
    [sym_block_comment] = STATE(5516),
    [sym_compiler_directive_decl] = STATE(5516),
    [sym_fsi_directive_decl] = STATE(5516),
    [sym_preproc_line] = STATE(5516),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_with] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5517)] = {
    [sym_block_comment] = STATE(5517),
    [sym_compiler_directive_decl] = STATE(5517),
    [sym_fsi_directive_decl] = STATE(5517),
    [sym_preproc_line] = STATE(5517),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__multi_dollar_interp_end] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5518)] = {
    [sym_block_comment] = STATE(5518),
    [sym_compiler_directive_decl] = STATE(5518),
    [sym_fsi_directive_decl] = STATE(5518),
    [sym_preproc_line] = STATE(5518),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_GT] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5519)] = {
    [sym_block_comment] = STATE(5519),
    [sym_compiler_directive_decl] = STATE(5519),
    [sym_fsi_directive_decl] = STATE(5519),
    [sym_preproc_line] = STATE(5519),
    [sym_identifier] = ACTIONS(7129),
    [anon_sym_null] = ACTIONS(7129),
    [anon_sym_EQ] = ACTIONS(7129),
    [anon_sym_begin] = ACTIONS(7129),
    [anon_sym_COLON] = ACTIONS(7129),
    [anon_sym_do] = ACTIONS(7129),
    [anon_sym_let] = ACTIONS(7129),
    [anon_sym_let_BANG] = ACTIONS(7131),
    [anon_sym_LPAREN] = ACTIONS(7129),
    [anon_sym_COMMA] = ACTIONS(7131),
    [anon_sym_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_QMARK] = ACTIONS(7129),
    [anon_sym_COLON_COLON] = ACTIONS(7131),
    [anon_sym_AMP] = ACTIONS(7129),
    [anon_sym_LBRACK] = ACTIONS(7129),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7131),
    [anon_sym_LBRACE] = ACTIONS(7129),
    [anon_sym_LT_AT] = ACTIONS(7129),
    [anon_sym_LT_AT_AT] = ACTIONS(7129),
    [anon_sym_CARET] = ACTIONS(7129),
    [anon_sym_or] = ACTIONS(7129),
    [anon_sym_DOT] = ACTIONS(7129),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7131),
    [anon_sym_new] = ACTIONS(7129),
    [anon_sym_return] = ACTIONS(7129),
    [anon_sym_return_BANG] = ACTIONS(7131),
    [anon_sym_yield] = ACTIONS(7129),
    [anon_sym_yield_BANG] = ACTIONS(7131),
    [anon_sym_lazy] = ACTIONS(7129),
    [anon_sym_assert] = ACTIONS(7129),
    [anon_sym_upcast] = ACTIONS(7129),
    [anon_sym_downcast] = ACTIONS(7129),
    [anon_sym_fixed] = ACTIONS(7129),
    [anon_sym_COLON_GT] = ACTIONS(7131),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7131),
    [anon_sym_for] = ACTIONS(7129),
    [anon_sym_while] = ACTIONS(7129),
    [anon_sym_while_BANG] = ACTIONS(7131),
    [anon_sym_if] = ACTIONS(7129),
    [anon_sym_fun] = ACTIONS(7129),
    [anon_sym_try] = ACTIONS(7129),
    [anon_sym_match] = ACTIONS(7129),
    [anon_sym_match_BANG] = ACTIONS(7131),
    [anon_sym_function] = ACTIONS(7129),
    [anon_sym_DOT_LBRACK] = ACTIONS(7131),
    [anon_sym_use] = ACTIONS(7129),
    [anon_sym_use_BANG] = ACTIONS(7131),
    [anon_sym_do_BANG] = ACTIONS(7131),
    [anon_sym_DOT_DOT] = ACTIONS(7131),
    [anon_sym_LPAREN2] = ACTIONS(7129),
    [anon_sym_PIPE_PIPE] = ACTIONS(7129),
    [anon_sym_AMP_AMP] = ACTIONS(7129),
    [aux_sym_char_token1] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7129),
    [anon_sym_DQUOTE] = ACTIONS(7129),
    [anon_sym_AT_DQUOTE] = ACTIONS(7131),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7131),
    [sym_bool] = ACTIONS(7129),
    [sym_unit] = ACTIONS(7131),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7129),
    [sym_op_identifier] = ACTIONS(7129),
    [anon_sym_PLUS] = ACTIONS(7129),
    [anon_sym_DASH] = ACTIONS(7129),
    [anon_sym_PLUS_DOT] = ACTIONS(7129),
    [anon_sym_DASH_DOT] = ACTIONS(7129),
    [anon_sym_PERCENT] = ACTIONS(7129),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7129),
    [anon_sym_TILDE] = ACTIONS(7131),
    [aux_sym_prefix_op_token1] = ACTIONS(7129),
    [aux_sym_infix_op_token1] = ACTIONS(7131),
    [aux_sym_infix_op_token2] = ACTIONS(7129),
    [aux_sym_infix_op_token3] = ACTIONS(7129),
    [anon_sym_BANG_EQ] = ACTIONS(7129),
    [anon_sym_COLON_EQ] = ACTIONS(7131),
    [anon_sym_DOLLAR] = ACTIONS(7129),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7129),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7129),
    [sym_int] = ACTIONS(7129),
    [sym_xint] = ACTIONS(7131),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7131),
    [sym__newline] = ACTIONS(7131),
    [sym__dedent] = ACTIONS(7131),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7131),
    [sym__tyapp_open] = ACTIONS(7131),
  },
  [STATE(5520)] = {
    [sym_block_comment] = STATE(5520),
    [sym_compiler_directive_decl] = STATE(5520),
    [sym_fsi_directive_decl] = STATE(5520),
    [sym_preproc_line] = STATE(5520),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__quoted_close] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5521)] = {
    [sym_block_comment] = STATE(5521),
    [sym_compiler_directive_decl] = STATE(5521),
    [sym_fsi_directive_decl] = STATE(5521),
    [sym_preproc_line] = STATE(5521),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9726),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5522)] = {
    [sym_block_comment] = STATE(5522),
    [sym_compiler_directive_decl] = STATE(5522),
    [sym_fsi_directive_decl] = STATE(5522),
    [sym_preproc_line] = STATE(5522),
    [sym_identifier] = ACTIONS(7133),
    [anon_sym_null] = ACTIONS(7133),
    [anon_sym_EQ] = ACTIONS(7133),
    [anon_sym_begin] = ACTIONS(7133),
    [anon_sym_COLON] = ACTIONS(7133),
    [anon_sym_do] = ACTIONS(7133),
    [anon_sym_let] = ACTIONS(7133),
    [anon_sym_let_BANG] = ACTIONS(7135),
    [anon_sym_LPAREN] = ACTIONS(7133),
    [anon_sym_COMMA] = ACTIONS(7135),
    [anon_sym_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_QMARK] = ACTIONS(7133),
    [anon_sym_COLON_COLON] = ACTIONS(7135),
    [anon_sym_AMP] = ACTIONS(7133),
    [anon_sym_LBRACK] = ACTIONS(7133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7135),
    [anon_sym_LBRACE] = ACTIONS(7133),
    [anon_sym_LT_AT] = ACTIONS(7133),
    [anon_sym_LT_AT_AT] = ACTIONS(7133),
    [anon_sym_CARET] = ACTIONS(7133),
    [anon_sym_or] = ACTIONS(7133),
    [anon_sym_DOT] = ACTIONS(7133),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7135),
    [anon_sym_new] = ACTIONS(7133),
    [anon_sym_return] = ACTIONS(7133),
    [anon_sym_return_BANG] = ACTIONS(7135),
    [anon_sym_yield] = ACTIONS(7133),
    [anon_sym_yield_BANG] = ACTIONS(7135),
    [anon_sym_lazy] = ACTIONS(7133),
    [anon_sym_assert] = ACTIONS(7133),
    [anon_sym_upcast] = ACTIONS(7133),
    [anon_sym_downcast] = ACTIONS(7133),
    [anon_sym_fixed] = ACTIONS(7133),
    [anon_sym_COLON_GT] = ACTIONS(7135),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7135),
    [anon_sym_for] = ACTIONS(7133),
    [anon_sym_while] = ACTIONS(7133),
    [anon_sym_while_BANG] = ACTIONS(7135),
    [anon_sym_if] = ACTIONS(7133),
    [anon_sym_fun] = ACTIONS(7133),
    [anon_sym_try] = ACTIONS(7133),
    [anon_sym_match] = ACTIONS(7133),
    [anon_sym_match_BANG] = ACTIONS(7135),
    [anon_sym_function] = ACTIONS(7133),
    [anon_sym_DOT_LBRACK] = ACTIONS(7135),
    [anon_sym_use] = ACTIONS(7133),
    [anon_sym_use_BANG] = ACTIONS(7135),
    [anon_sym_do_BANG] = ACTIONS(7135),
    [anon_sym_DOT_DOT] = ACTIONS(7135),
    [anon_sym_LPAREN2] = ACTIONS(7133),
    [anon_sym_PIPE_PIPE] = ACTIONS(7133),
    [anon_sym_AMP_AMP] = ACTIONS(7133),
    [aux_sym_char_token1] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7133),
    [anon_sym_DQUOTE] = ACTIONS(7133),
    [anon_sym_AT_DQUOTE] = ACTIONS(7135),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7135),
    [sym_bool] = ACTIONS(7133),
    [sym_unit] = ACTIONS(7135),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7133),
    [sym_op_identifier] = ACTIONS(7133),
    [anon_sym_PLUS] = ACTIONS(7133),
    [anon_sym_DASH] = ACTIONS(7133),
    [anon_sym_PLUS_DOT] = ACTIONS(7133),
    [anon_sym_DASH_DOT] = ACTIONS(7133),
    [anon_sym_PERCENT] = ACTIONS(7133),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7133),
    [anon_sym_TILDE] = ACTIONS(7135),
    [aux_sym_prefix_op_token1] = ACTIONS(7133),
    [aux_sym_infix_op_token1] = ACTIONS(7135),
    [aux_sym_infix_op_token2] = ACTIONS(7133),
    [aux_sym_infix_op_token3] = ACTIONS(7133),
    [anon_sym_BANG_EQ] = ACTIONS(7133),
    [anon_sym_COLON_EQ] = ACTIONS(7135),
    [anon_sym_DOLLAR] = ACTIONS(7133),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7133),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7133),
    [sym_int] = ACTIONS(7133),
    [sym_xint] = ACTIONS(7135),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7135),
    [sym__newline] = ACTIONS(7135),
    [sym__dedent] = ACTIONS(7135),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7135),
    [sym__tyapp_open] = ACTIONS(7135),
  },
  [STATE(5523)] = {
    [sym_block_comment] = STATE(5523),
    [sym_compiler_directive_decl] = STATE(5523),
    [sym_fsi_directive_decl] = STATE(5523),
    [sym_preproc_line] = STATE(5523),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__quoted_close] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5524)] = {
    [sym_block_comment] = STATE(5524),
    [sym_compiler_directive_decl] = STATE(5524),
    [sym_fsi_directive_decl] = STATE(5524),
    [sym_preproc_line] = STATE(5524),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9728),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_DOT_DOT] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5525)] = {
    [sym_block_comment] = STATE(5525),
    [sym_compiler_directive_decl] = STATE(5525),
    [sym_fsi_directive_decl] = STATE(5525),
    [sym_preproc_line] = STATE(5525),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5526)] = {
    [sym_block_comment] = STATE(5526),
    [sym_compiler_directive_decl] = STATE(5526),
    [sym_fsi_directive_decl] = STATE(5526),
    [sym_preproc_line] = STATE(5526),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__quoted_close] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5527)] = {
    [sym_block_comment] = STATE(5527),
    [sym_compiler_directive_decl] = STATE(5527),
    [sym_fsi_directive_decl] = STATE(5527),
    [sym_preproc_line] = STATE(5527),
    [sym_identifier] = ACTIONS(7137),
    [anon_sym_null] = ACTIONS(7137),
    [anon_sym_EQ] = ACTIONS(7137),
    [anon_sym_begin] = ACTIONS(7137),
    [anon_sym_COLON] = ACTIONS(7137),
    [anon_sym_do] = ACTIONS(7137),
    [anon_sym_let] = ACTIONS(7137),
    [anon_sym_let_BANG] = ACTIONS(7139),
    [anon_sym_LPAREN] = ACTIONS(7137),
    [anon_sym_COMMA] = ACTIONS(7139),
    [anon_sym_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_QMARK] = ACTIONS(7137),
    [anon_sym_COLON_COLON] = ACTIONS(7139),
    [anon_sym_AMP] = ACTIONS(7137),
    [anon_sym_LBRACK] = ACTIONS(7137),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7139),
    [anon_sym_LBRACE] = ACTIONS(7137),
    [anon_sym_LT_AT] = ACTIONS(7137),
    [anon_sym_LT_AT_AT] = ACTIONS(7137),
    [anon_sym_CARET] = ACTIONS(7137),
    [anon_sym_or] = ACTIONS(7137),
    [anon_sym_DOT] = ACTIONS(7137),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7139),
    [anon_sym_new] = ACTIONS(7137),
    [anon_sym_return] = ACTIONS(7137),
    [anon_sym_return_BANG] = ACTIONS(7139),
    [anon_sym_yield] = ACTIONS(7137),
    [anon_sym_yield_BANG] = ACTIONS(7139),
    [anon_sym_lazy] = ACTIONS(7137),
    [anon_sym_assert] = ACTIONS(7137),
    [anon_sym_upcast] = ACTIONS(7137),
    [anon_sym_downcast] = ACTIONS(7137),
    [anon_sym_fixed] = ACTIONS(7137),
    [anon_sym_COLON_GT] = ACTIONS(7139),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7139),
    [anon_sym_for] = ACTIONS(7137),
    [anon_sym_while] = ACTIONS(7137),
    [anon_sym_while_BANG] = ACTIONS(7139),
    [anon_sym_if] = ACTIONS(7137),
    [anon_sym_fun] = ACTIONS(7137),
    [anon_sym_try] = ACTIONS(7137),
    [anon_sym_match] = ACTIONS(7137),
    [anon_sym_match_BANG] = ACTIONS(7139),
    [anon_sym_function] = ACTIONS(7137),
    [anon_sym_DOT_LBRACK] = ACTIONS(7139),
    [anon_sym_use] = ACTIONS(7137),
    [anon_sym_use_BANG] = ACTIONS(7139),
    [anon_sym_do_BANG] = ACTIONS(7139),
    [anon_sym_DOT_DOT] = ACTIONS(7139),
    [anon_sym_LPAREN2] = ACTIONS(7137),
    [anon_sym_PIPE_PIPE] = ACTIONS(7137),
    [anon_sym_AMP_AMP] = ACTIONS(7137),
    [aux_sym_char_token1] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7137),
    [anon_sym_DQUOTE] = ACTIONS(7137),
    [anon_sym_AT_DQUOTE] = ACTIONS(7139),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7139),
    [sym_bool] = ACTIONS(7137),
    [sym_unit] = ACTIONS(7139),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7137),
    [sym_op_identifier] = ACTIONS(7137),
    [anon_sym_PLUS] = ACTIONS(7137),
    [anon_sym_DASH] = ACTIONS(7137),
    [anon_sym_PLUS_DOT] = ACTIONS(7137),
    [anon_sym_DASH_DOT] = ACTIONS(7137),
    [anon_sym_PERCENT] = ACTIONS(7137),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7137),
    [anon_sym_TILDE] = ACTIONS(7139),
    [aux_sym_prefix_op_token1] = ACTIONS(7137),
    [aux_sym_infix_op_token1] = ACTIONS(7139),
    [aux_sym_infix_op_token2] = ACTIONS(7137),
    [aux_sym_infix_op_token3] = ACTIONS(7137),
    [anon_sym_BANG_EQ] = ACTIONS(7137),
    [anon_sym_COLON_EQ] = ACTIONS(7139),
    [anon_sym_DOLLAR] = ACTIONS(7137),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7137),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7137),
    [sym_int] = ACTIONS(7137),
    [sym_xint] = ACTIONS(7139),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7139),
    [sym__newline] = ACTIONS(7139),
    [sym__dedent] = ACTIONS(7139),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7139),
    [sym__tyapp_open] = ACTIONS(7139),
  },
  [STATE(5528)] = {
    [sym_block_comment] = STATE(5528),
    [sym_compiler_directive_decl] = STATE(5528),
    [sym_fsi_directive_decl] = STATE(5528),
    [sym_preproc_line] = STATE(5528),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__quoted_close] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5529)] = {
    [sym_block_comment] = STATE(5529),
    [sym_compiler_directive_decl] = STATE(5529),
    [sym_fsi_directive_decl] = STATE(5529),
    [sym_preproc_line] = STATE(5529),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_DOT_DOT] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__dedent] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(5530)] = {
    [sym_block_comment] = STATE(5530),
    [sym_compiler_directive_decl] = STATE(5530),
    [sym_fsi_directive_decl] = STATE(5530),
    [sym_preproc_line] = STATE(5530),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_DOT_DOT] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__dedent] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(5531)] = {
    [sym_block_comment] = STATE(5531),
    [sym_compiler_directive_decl] = STATE(5531),
    [sym_fsi_directive_decl] = STATE(5531),
    [sym_preproc_line] = STATE(5531),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_DOT_DOT] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__dedent] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(5532)] = {
    [sym_block_comment] = STATE(5532),
    [sym_compiler_directive_decl] = STATE(5532),
    [sym_fsi_directive_decl] = STATE(5532),
    [sym_preproc_line] = STATE(5532),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_GT] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9664),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5533)] = {
    [sym_block_comment] = STATE(5533),
    [sym_compiler_directive_decl] = STATE(5533),
    [sym_fsi_directive_decl] = STATE(5533),
    [sym_preproc_line] = STATE(5533),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__dedent] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(5534)] = {
    [sym_block_comment] = STATE(5534),
    [sym_compiler_directive_decl] = STATE(5534),
    [sym_fsi_directive_decl] = STATE(5534),
    [sym_preproc_line] = STATE(5534),
    [aux_sym_long_identifier_repeat1] = STATE(5937),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9730),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9732),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5535)] = {
    [sym_block_comment] = STATE(5535),
    [sym_compiler_directive_decl] = STATE(5535),
    [sym_fsi_directive_decl] = STATE(5535),
    [sym_preproc_line] = STATE(5535),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_DASH_GT] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_DOT_DOT] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(5536)] = {
    [sym_block_comment] = STATE(5536),
    [sym_compiler_directive_decl] = STATE(5536),
    [sym_fsi_directive_decl] = STATE(5536),
    [sym_preproc_line] = STATE(5536),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_DASH_GT] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_DOT_DOT] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(5537)] = {
    [sym_block_comment] = STATE(5537),
    [sym_compiler_directive_decl] = STATE(5537),
    [sym_fsi_directive_decl] = STATE(5537),
    [sym_preproc_line] = STATE(5537),
    [aux_sym_type_argument_repeat1] = STATE(5538),
    [sym_identifier] = ACTIONS(6308),
    [anon_sym_module] = ACTIONS(6308),
    [anon_sym_null] = ACTIONS(6308),
    [anon_sym_begin] = ACTIONS(6308),
    [anon_sym_open] = ACTIONS(6308),
    [anon_sym_type] = ACTIONS(6308),
    [anon_sym_LBRACK_LT] = ACTIONS(6306),
    [anon_sym_do] = ACTIONS(6308),
    [anon_sym_and] = ACTIONS(6308),
    [anon_sym_let] = ACTIONS(6308),
    [anon_sym_let_BANG] = ACTIONS(6306),
    [aux_sym_access_modifier_token1] = ACTIONS(6306),
    [anon_sym_LPAREN] = ACTIONS(6308),
    [anon_sym_AMP] = ACTIONS(6308),
    [anon_sym_LBRACK] = ACTIONS(6308),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6306),
    [anon_sym_LBRACE] = ACTIONS(6308),
    [anon_sym_LT_AT] = ACTIONS(6308),
    [anon_sym_LT_AT_AT] = ACTIONS(6306),
    [anon_sym_CARET] = ACTIONS(6306),
    [anon_sym_or] = ACTIONS(9734),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6306),
    [anon_sym_new] = ACTIONS(6308),
    [anon_sym_return] = ACTIONS(6308),
    [anon_sym_return_BANG] = ACTIONS(6306),
    [anon_sym_yield] = ACTIONS(6308),
    [anon_sym_yield_BANG] = ACTIONS(6306),
    [anon_sym_lazy] = ACTIONS(6308),
    [anon_sym_assert] = ACTIONS(6308),
    [anon_sym_upcast] = ACTIONS(6308),
    [anon_sym_downcast] = ACTIONS(6308),
    [anon_sym_fixed] = ACTIONS(6308),
    [anon_sym_COLON_GT] = ACTIONS(6306),
    [anon_sym_for] = ACTIONS(6308),
    [anon_sym_while] = ACTIONS(6308),
    [anon_sym_while_BANG] = ACTIONS(6306),
    [anon_sym_if] = ACTIONS(6308),
    [anon_sym_fun] = ACTIONS(6308),
    [anon_sym_DASH_GT] = ACTIONS(6306),
    [anon_sym_try] = ACTIONS(6308),
    [anon_sym_match] = ACTIONS(6308),
    [anon_sym_match_BANG] = ACTIONS(6306),
    [anon_sym_function] = ACTIONS(6308),
    [anon_sym_use] = ACTIONS(6308),
    [anon_sym_use_BANG] = ACTIONS(6306),
    [anon_sym_do_BANG] = ACTIONS(6306),
    [anon_sym_AMP_AMP] = ACTIONS(6306),
    [anon_sym_STAR] = ACTIONS(6306),
    [anon_sym_LT] = ACTIONS(6308),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6306),
    [anon_sym_default] = ACTIONS(6308),
    [anon_sym_static] = ACTIONS(6308),
    [anon_sym_member] = ACTIONS(6308),
    [anon_sym_exception] = ACTIONS(6308),
    [anon_sym_interface] = ACTIONS(6308),
    [anon_sym_abstract] = ACTIONS(6308),
    [anon_sym_val] = ACTIONS(6308),
    [anon_sym_override] = ACTIONS(6308),
    [anon_sym_extern] = ACTIONS(6308),
    [aux_sym_char_token1] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6308),
    [anon_sym_DQUOTE] = ACTIONS(6308),
    [anon_sym_AT_DQUOTE] = ACTIONS(6306),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6306),
    [sym_bool] = ACTIONS(6308),
    [sym_unit] = ACTIONS(6306),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6308),
    [sym_op_identifier] = ACTIONS(6306),
    [anon_sym_PLUS] = ACTIONS(6308),
    [anon_sym_DASH] = ACTIONS(6308),
    [anon_sym_PLUS_DOT] = ACTIONS(6306),
    [anon_sym_DASH_DOT] = ACTIONS(6306),
    [anon_sym_PERCENT] = ACTIONS(6308),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6306),
    [anon_sym_TILDE] = ACTIONS(6306),
    [aux_sym_prefix_op_token1] = ACTIONS(6306),
    [sym_int] = ACTIONS(6308),
    [sym_xint] = ACTIONS(6306),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6306),
    [anon_sym_POUNDload] = ACTIONS(6306),
    [anon_sym_POUNDtime] = ACTIONS(6306),
    [anon_sym_POUNDI] = ACTIONS(6306),
    [anon_sym_POUNDhelp] = ACTIONS(6306),
    [anon_sym_POUNDquit] = ACTIONS(6306),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6306),
    [anon_sym_POUNDendif] = ACTIONS(6306),
    [anon_sym_POUNDelse] = ACTIONS(6306),
    [sym__newline] = ACTIONS(6306),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6306),
  },
  [STATE(5538)] = {
    [sym_block_comment] = STATE(5538),
    [sym_compiler_directive_decl] = STATE(5538),
    [sym_fsi_directive_decl] = STATE(5538),
    [sym_preproc_line] = STATE(5538),
    [aux_sym_type_argument_repeat1] = STATE(5538),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(9736),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__newline] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5539)] = {
    [sym_type_arguments] = STATE(6569),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5539),
    [sym_compiler_directive_decl] = STATE(5539),
    [sym_fsi_directive_decl] = STATE(5539),
    [sym_preproc_line] = STATE(5539),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(9352),
    [anon_sym_null] = ACTIONS(9352),
    [anon_sym_begin] = ACTIONS(9352),
    [anon_sym_open] = ACTIONS(9352),
    [anon_sym_type] = ACTIONS(9352),
    [anon_sym_LBRACK_LT] = ACTIONS(9354),
    [anon_sym_do] = ACTIONS(9352),
    [anon_sym_and] = ACTIONS(9352),
    [anon_sym_let] = ACTIONS(9352),
    [anon_sym_let_BANG] = ACTIONS(9354),
    [aux_sym_access_modifier_token1] = ACTIONS(9354),
    [anon_sym_LPAREN] = ACTIONS(9352),
    [anon_sym_PIPE] = ACTIONS(9354),
    [anon_sym_AMP] = ACTIONS(9352),
    [anon_sym_LBRACK] = ACTIONS(9352),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9354),
    [anon_sym_LBRACE] = ACTIONS(9352),
    [anon_sym_LT_AT] = ACTIONS(9352),
    [anon_sym_LT_AT_AT] = ACTIONS(9354),
    [anon_sym_CARET] = ACTIONS(9354),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9354),
    [anon_sym_new] = ACTIONS(9352),
    [anon_sym_with] = ACTIONS(9352),
    [anon_sym_return] = ACTIONS(9352),
    [anon_sym_return_BANG] = ACTIONS(9354),
    [anon_sym_yield] = ACTIONS(9352),
    [anon_sym_yield_BANG] = ACTIONS(9354),
    [anon_sym_lazy] = ACTIONS(9352),
    [anon_sym_assert] = ACTIONS(9352),
    [anon_sym_upcast] = ACTIONS(9352),
    [anon_sym_downcast] = ACTIONS(9352),
    [anon_sym_fixed] = ACTIONS(9352),
    [anon_sym_for] = ACTIONS(9352),
    [anon_sym_while] = ACTIONS(9352),
    [anon_sym_while_BANG] = ACTIONS(9354),
    [anon_sym_if] = ACTIONS(9352),
    [anon_sym_fun] = ACTIONS(9352),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(9352),
    [anon_sym_match] = ACTIONS(9352),
    [anon_sym_match_BANG] = ACTIONS(9354),
    [anon_sym_function] = ACTIONS(9352),
    [anon_sym_use] = ACTIONS(9352),
    [anon_sym_use_BANG] = ACTIONS(9354),
    [anon_sym_do_BANG] = ACTIONS(9354),
    [anon_sym_AMP_AMP] = ACTIONS(9354),
    [anon_sym_STAR] = ACTIONS(9354),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(9352),
    [anon_sym_static] = ACTIONS(9352),
    [anon_sym_member] = ACTIONS(9352),
    [anon_sym_exception] = ACTIONS(9352),
    [anon_sym_interface] = ACTIONS(9352),
    [anon_sym_abstract] = ACTIONS(9352),
    [anon_sym_val] = ACTIONS(9352),
    [anon_sym_override] = ACTIONS(9352),
    [anon_sym_extern] = ACTIONS(9352),
    [aux_sym_char_token1] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9352),
    [anon_sym_DQUOTE] = ACTIONS(9352),
    [anon_sym_AT_DQUOTE] = ACTIONS(9354),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9354),
    [sym_bool] = ACTIONS(9352),
    [sym_unit] = ACTIONS(9354),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9352),
    [sym_op_identifier] = ACTIONS(9354),
    [anon_sym_PLUS] = ACTIONS(9352),
    [anon_sym_DASH] = ACTIONS(9352),
    [anon_sym_PLUS_DOT] = ACTIONS(9354),
    [anon_sym_DASH_DOT] = ACTIONS(9354),
    [anon_sym_PERCENT] = ACTIONS(9352),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9354),
    [anon_sym_TILDE] = ACTIONS(9354),
    [aux_sym_prefix_op_token1] = ACTIONS(9354),
    [sym_int] = ACTIONS(9352),
    [sym_xint] = ACTIONS(9354),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9354),
    [anon_sym_POUNDload] = ACTIONS(9354),
    [anon_sym_POUNDtime] = ACTIONS(9354),
    [anon_sym_POUNDI] = ACTIONS(9354),
    [anon_sym_POUNDhelp] = ACTIONS(9354),
    [anon_sym_POUNDquit] = ACTIONS(9354),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9354),
    [anon_sym_POUNDendif] = ACTIONS(9354),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9354),
  },
  [STATE(5540)] = {
    [sym_block_comment] = STATE(5540),
    [sym_compiler_directive_decl] = STATE(5540),
    [sym_fsi_directive_decl] = STATE(5540),
    [sym_preproc_line] = STATE(5540),
    [aux_sym_sequential_expression_repeat1] = STATE(5540),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9747),
    [sym__untyped_quoted_close] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5541)] = {
    [sym_block_comment] = STATE(5541),
    [sym_compiler_directive_decl] = STATE(5541),
    [sym_fsi_directive_decl] = STATE(5541),
    [sym_preproc_line] = STATE(5541),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_with] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5542)] = {
    [sym_type_arguments] = STATE(6569),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5542),
    [sym_compiler_directive_decl] = STATE(5542),
    [sym_fsi_directive_decl] = STATE(5542),
    [sym_preproc_line] = STATE(5542),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(9346),
    [anon_sym_null] = ACTIONS(9346),
    [anon_sym_begin] = ACTIONS(9346),
    [anon_sym_open] = ACTIONS(9346),
    [anon_sym_type] = ACTIONS(9346),
    [anon_sym_LBRACK_LT] = ACTIONS(9348),
    [anon_sym_do] = ACTIONS(9346),
    [anon_sym_and] = ACTIONS(9346),
    [anon_sym_let] = ACTIONS(9346),
    [anon_sym_let_BANG] = ACTIONS(9348),
    [aux_sym_access_modifier_token1] = ACTIONS(9348),
    [anon_sym_LPAREN] = ACTIONS(9346),
    [anon_sym_PIPE] = ACTIONS(9348),
    [anon_sym_AMP] = ACTIONS(9346),
    [anon_sym_LBRACK] = ACTIONS(9346),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9348),
    [anon_sym_LBRACE] = ACTIONS(9346),
    [anon_sym_LT_AT] = ACTIONS(9346),
    [anon_sym_LT_AT_AT] = ACTIONS(9348),
    [anon_sym_CARET] = ACTIONS(9348),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9348),
    [anon_sym_new] = ACTIONS(9346),
    [anon_sym_with] = ACTIONS(9346),
    [anon_sym_return] = ACTIONS(9346),
    [anon_sym_return_BANG] = ACTIONS(9348),
    [anon_sym_yield] = ACTIONS(9346),
    [anon_sym_yield_BANG] = ACTIONS(9348),
    [anon_sym_lazy] = ACTIONS(9346),
    [anon_sym_assert] = ACTIONS(9346),
    [anon_sym_upcast] = ACTIONS(9346),
    [anon_sym_downcast] = ACTIONS(9346),
    [anon_sym_fixed] = ACTIONS(9346),
    [anon_sym_for] = ACTIONS(9346),
    [anon_sym_while] = ACTIONS(9346),
    [anon_sym_while_BANG] = ACTIONS(9348),
    [anon_sym_if] = ACTIONS(9346),
    [anon_sym_fun] = ACTIONS(9346),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(9346),
    [anon_sym_match] = ACTIONS(9346),
    [anon_sym_match_BANG] = ACTIONS(9348),
    [anon_sym_function] = ACTIONS(9346),
    [anon_sym_use] = ACTIONS(9346),
    [anon_sym_use_BANG] = ACTIONS(9348),
    [anon_sym_do_BANG] = ACTIONS(9348),
    [anon_sym_AMP_AMP] = ACTIONS(9348),
    [anon_sym_STAR] = ACTIONS(9348),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(9346),
    [anon_sym_static] = ACTIONS(9346),
    [anon_sym_member] = ACTIONS(9346),
    [anon_sym_exception] = ACTIONS(9346),
    [anon_sym_interface] = ACTIONS(9346),
    [anon_sym_abstract] = ACTIONS(9346),
    [anon_sym_val] = ACTIONS(9346),
    [anon_sym_override] = ACTIONS(9346),
    [anon_sym_extern] = ACTIONS(9346),
    [aux_sym_char_token1] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9346),
    [anon_sym_DQUOTE] = ACTIONS(9346),
    [anon_sym_AT_DQUOTE] = ACTIONS(9348),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9348),
    [sym_bool] = ACTIONS(9346),
    [sym_unit] = ACTIONS(9348),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9346),
    [sym_op_identifier] = ACTIONS(9348),
    [anon_sym_PLUS] = ACTIONS(9346),
    [anon_sym_DASH] = ACTIONS(9346),
    [anon_sym_PLUS_DOT] = ACTIONS(9348),
    [anon_sym_DASH_DOT] = ACTIONS(9348),
    [anon_sym_PERCENT] = ACTIONS(9346),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9348),
    [anon_sym_TILDE] = ACTIONS(9348),
    [aux_sym_prefix_op_token1] = ACTIONS(9348),
    [sym_int] = ACTIONS(9346),
    [sym_xint] = ACTIONS(9348),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9348),
    [anon_sym_POUNDload] = ACTIONS(9348),
    [anon_sym_POUNDtime] = ACTIONS(9348),
    [anon_sym_POUNDI] = ACTIONS(9348),
    [anon_sym_POUNDhelp] = ACTIONS(9348),
    [anon_sym_POUNDquit] = ACTIONS(9348),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9348),
    [anon_sym_POUNDendif] = ACTIONS(9348),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9348),
  },
  [STATE(5543)] = {
    [sym_block_comment] = STATE(5543),
    [sym_compiler_directive_decl] = STATE(5543),
    [sym_fsi_directive_decl] = STATE(5543),
    [sym_preproc_line] = STATE(5543),
    [aux_sym_sequential_expression_repeat1] = STATE(5543),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [anon_sym_POUNDendif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9750),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5544)] = {
    [sym_block_comment] = STATE(5544),
    [sym_compiler_directive_decl] = STATE(5544),
    [sym_fsi_directive_decl] = STATE(5544),
    [sym_preproc_line] = STATE(5544),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_DOT_DOT] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__dedent] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(5545)] = {
    [sym_block_comment] = STATE(5545),
    [sym_compiler_directive_decl] = STATE(5545),
    [sym_fsi_directive_decl] = STATE(5545),
    [sym_preproc_line] = STATE(5545),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_DOT_DOT] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__dedent] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(5546)] = {
    [sym_block_comment] = STATE(5546),
    [sym_compiler_directive_decl] = STATE(5546),
    [sym_fsi_directive_decl] = STATE(5546),
    [sym_preproc_line] = STATE(5546),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_DOT_DOT] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__dedent] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(5547)] = {
    [sym_block_comment] = STATE(5547),
    [sym_compiler_directive_decl] = STATE(5547),
    [sym_fsi_directive_decl] = STATE(5547),
    [sym_preproc_line] = STATE(5547),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_DOT_DOT] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__dedent] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(5548)] = {
    [sym_block_comment] = STATE(5548),
    [sym_compiler_directive_decl] = STATE(5548),
    [sym_fsi_directive_decl] = STATE(5548),
    [sym_preproc_line] = STATE(5548),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_DOT_DOT] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__dedent] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(5549)] = {
    [sym_block_comment] = STATE(5549),
    [sym_compiler_directive_decl] = STATE(5549),
    [sym_fsi_directive_decl] = STATE(5549),
    [sym_preproc_line] = STATE(5549),
    [aux_sym_long_identifier_repeat1] = STATE(5618),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9684),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [sym__newline] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5550)] = {
    [sym_block_comment] = STATE(5550),
    [sym_compiler_directive_decl] = STATE(5550),
    [sym_fsi_directive_decl] = STATE(5550),
    [sym_preproc_line] = STATE(5550),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_DASH_GT] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_DOT_DOT] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(5551)] = {
    [sym_block_comment] = STATE(5551),
    [sym_compiler_directive_decl] = STATE(5551),
    [sym_fsi_directive_decl] = STATE(5551),
    [sym_preproc_line] = STATE(5551),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_DASH_GT] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_DOT_DOT] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(5552)] = {
    [sym_block_comment] = STATE(5552),
    [sym_compiler_directive_decl] = STATE(5552),
    [sym_fsi_directive_decl] = STATE(5552),
    [sym_preproc_line] = STATE(5552),
    [sym_identifier] = ACTIONS(6790),
    [anon_sym_null] = ACTIONS(6790),
    [anon_sym_EQ] = ACTIONS(6790),
    [anon_sym_begin] = ACTIONS(6790),
    [anon_sym_COLON] = ACTIONS(6790),
    [anon_sym_do] = ACTIONS(6790),
    [anon_sym_let] = ACTIONS(6790),
    [anon_sym_let_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN] = ACTIONS(6790),
    [anon_sym_COMMA] = ACTIONS(6792),
    [anon_sym_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_QMARK] = ACTIONS(6790),
    [anon_sym_COLON_COLON] = ACTIONS(6792),
    [anon_sym_AMP] = ACTIONS(6790),
    [anon_sym_LBRACK] = ACTIONS(6790),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6792),
    [anon_sym_LBRACE] = ACTIONS(6790),
    [anon_sym_LT_AT] = ACTIONS(6790),
    [anon_sym_LT_AT_AT] = ACTIONS(6790),
    [anon_sym_CARET] = ACTIONS(6790),
    [anon_sym_or] = ACTIONS(6790),
    [anon_sym_DOT] = ACTIONS(6790),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6792),
    [anon_sym_new] = ACTIONS(6790),
    [anon_sym_return] = ACTIONS(6790),
    [anon_sym_return_BANG] = ACTIONS(6792),
    [anon_sym_yield] = ACTIONS(6790),
    [anon_sym_yield_BANG] = ACTIONS(6792),
    [anon_sym_lazy] = ACTIONS(6790),
    [anon_sym_assert] = ACTIONS(6790),
    [anon_sym_upcast] = ACTIONS(6790),
    [anon_sym_downcast] = ACTIONS(6790),
    [anon_sym_fixed] = ACTIONS(6790),
    [anon_sym_COLON_GT] = ACTIONS(6792),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6792),
    [anon_sym_for] = ACTIONS(6790),
    [anon_sym_while] = ACTIONS(6790),
    [anon_sym_while_BANG] = ACTIONS(6792),
    [anon_sym_if] = ACTIONS(6790),
    [anon_sym_fun] = ACTIONS(6790),
    [anon_sym_try] = ACTIONS(6790),
    [anon_sym_match] = ACTIONS(6790),
    [anon_sym_match_BANG] = ACTIONS(6792),
    [anon_sym_function] = ACTIONS(6790),
    [anon_sym_LT_DASH] = ACTIONS(6790),
    [anon_sym_DOT_LBRACK] = ACTIONS(6792),
    [anon_sym_use] = ACTIONS(6790),
    [anon_sym_use_BANG] = ACTIONS(6792),
    [anon_sym_do_BANG] = ACTIONS(6792),
    [anon_sym_LPAREN2] = ACTIONS(6790),
    [anon_sym_PIPE_PIPE] = ACTIONS(6790),
    [anon_sym_AMP_AMP] = ACTIONS(6790),
    [aux_sym_char_token1] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6790),
    [anon_sym_DQUOTE] = ACTIONS(6790),
    [anon_sym_AT_DQUOTE] = ACTIONS(6792),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6792),
    [sym_bool] = ACTIONS(6790),
    [sym_unit] = ACTIONS(6792),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6790),
    [sym_op_identifier] = ACTIONS(6790),
    [anon_sym_PLUS] = ACTIONS(6790),
    [anon_sym_DASH] = ACTIONS(6790),
    [anon_sym_PLUS_DOT] = ACTIONS(6790),
    [anon_sym_DASH_DOT] = ACTIONS(6790),
    [anon_sym_PERCENT] = ACTIONS(6790),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6790),
    [anon_sym_TILDE] = ACTIONS(6792),
    [aux_sym_prefix_op_token1] = ACTIONS(6790),
    [aux_sym_infix_op_token1] = ACTIONS(6792),
    [aux_sym_infix_op_token2] = ACTIONS(6790),
    [aux_sym_infix_op_token3] = ACTIONS(6790),
    [anon_sym_BANG_EQ] = ACTIONS(6790),
    [anon_sym_COLON_EQ] = ACTIONS(6792),
    [anon_sym_DOLLAR] = ACTIONS(6790),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6790),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6790),
    [sym_int] = ACTIONS(6790),
    [sym_xint] = ACTIONS(6792),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6792),
    [sym__newline] = ACTIONS(6792),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6792),
    [sym__multi_dollar_interp_end] = ACTIONS(6792),
    [sym__tyapp_open] = ACTIONS(6792),
  },
  [STATE(5553)] = {
    [sym_block_comment] = STATE(5553),
    [sym_compiler_directive_decl] = STATE(5553),
    [sym_fsi_directive_decl] = STATE(5553),
    [sym_preproc_line] = STATE(5553),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(5554)] = {
    [sym_block_comment] = STATE(5554),
    [sym_compiler_directive_decl] = STATE(5554),
    [sym_fsi_directive_decl] = STATE(5554),
    [sym_preproc_line] = STATE(5554),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_DOT_DOT] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__dedent] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(5555)] = {
    [sym_block_comment] = STATE(5555),
    [sym_compiler_directive_decl] = STATE(5555),
    [sym_fsi_directive_decl] = STATE(5555),
    [sym_preproc_line] = STATE(5555),
    [sym_identifier] = ACTIONS(6794),
    [anon_sym_null] = ACTIONS(6794),
    [anon_sym_EQ] = ACTIONS(6794),
    [anon_sym_begin] = ACTIONS(6794),
    [anon_sym_COLON] = ACTIONS(6794),
    [anon_sym_do] = ACTIONS(6794),
    [anon_sym_let] = ACTIONS(6794),
    [anon_sym_let_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN] = ACTIONS(6794),
    [anon_sym_COMMA] = ACTIONS(6796),
    [anon_sym_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_QMARK] = ACTIONS(6794),
    [anon_sym_COLON_COLON] = ACTIONS(6796),
    [anon_sym_AMP] = ACTIONS(6794),
    [anon_sym_LBRACK] = ACTIONS(6794),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6796),
    [anon_sym_LBRACE] = ACTIONS(6794),
    [anon_sym_LT_AT] = ACTIONS(6794),
    [anon_sym_LT_AT_AT] = ACTIONS(6794),
    [anon_sym_CARET] = ACTIONS(6794),
    [anon_sym_or] = ACTIONS(6794),
    [anon_sym_DOT] = ACTIONS(6794),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6796),
    [anon_sym_new] = ACTIONS(6794),
    [anon_sym_return] = ACTIONS(6794),
    [anon_sym_return_BANG] = ACTIONS(6796),
    [anon_sym_yield] = ACTIONS(6794),
    [anon_sym_yield_BANG] = ACTIONS(6796),
    [anon_sym_lazy] = ACTIONS(6794),
    [anon_sym_assert] = ACTIONS(6794),
    [anon_sym_upcast] = ACTIONS(6794),
    [anon_sym_downcast] = ACTIONS(6794),
    [anon_sym_fixed] = ACTIONS(6794),
    [anon_sym_COLON_GT] = ACTIONS(6796),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6796),
    [anon_sym_for] = ACTIONS(6794),
    [anon_sym_while] = ACTIONS(6794),
    [anon_sym_while_BANG] = ACTIONS(6796),
    [anon_sym_if] = ACTIONS(6794),
    [anon_sym_fun] = ACTIONS(6794),
    [anon_sym_try] = ACTIONS(6794),
    [anon_sym_match] = ACTIONS(6794),
    [anon_sym_match_BANG] = ACTIONS(6796),
    [anon_sym_function] = ACTIONS(6794),
    [anon_sym_LT_DASH] = ACTIONS(6794),
    [anon_sym_DOT_LBRACK] = ACTIONS(6796),
    [anon_sym_use] = ACTIONS(6794),
    [anon_sym_use_BANG] = ACTIONS(6796),
    [anon_sym_do_BANG] = ACTIONS(6796),
    [anon_sym_LPAREN2] = ACTIONS(6794),
    [anon_sym_PIPE_PIPE] = ACTIONS(6794),
    [anon_sym_AMP_AMP] = ACTIONS(6794),
    [aux_sym_char_token1] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6794),
    [anon_sym_DQUOTE] = ACTIONS(6794),
    [anon_sym_AT_DQUOTE] = ACTIONS(6796),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6796),
    [sym_bool] = ACTIONS(6794),
    [sym_unit] = ACTIONS(6796),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6794),
    [sym_op_identifier] = ACTIONS(6794),
    [anon_sym_PLUS] = ACTIONS(6794),
    [anon_sym_DASH] = ACTIONS(6794),
    [anon_sym_PLUS_DOT] = ACTIONS(6794),
    [anon_sym_DASH_DOT] = ACTIONS(6794),
    [anon_sym_PERCENT] = ACTIONS(6794),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6794),
    [anon_sym_TILDE] = ACTIONS(6796),
    [aux_sym_prefix_op_token1] = ACTIONS(6794),
    [aux_sym_infix_op_token1] = ACTIONS(6796),
    [aux_sym_infix_op_token2] = ACTIONS(6794),
    [aux_sym_infix_op_token3] = ACTIONS(6794),
    [anon_sym_BANG_EQ] = ACTIONS(6794),
    [anon_sym_COLON_EQ] = ACTIONS(6796),
    [anon_sym_DOLLAR] = ACTIONS(6794),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6794),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6794),
    [sym_int] = ACTIONS(6794),
    [sym_xint] = ACTIONS(6796),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6796),
    [sym__newline] = ACTIONS(6796),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6796),
    [sym__multi_dollar_interp_end] = ACTIONS(6796),
    [sym__tyapp_open] = ACTIONS(6796),
  },
  [STATE(5556)] = {
    [sym_block_comment] = STATE(5556),
    [sym_compiler_directive_decl] = STATE(5556),
    [sym_fsi_directive_decl] = STATE(5556),
    [sym_preproc_line] = STATE(5556),
    [aux_sym_long_identifier_repeat1] = STATE(5601),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_COLON] = ACTIONS(9508),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9670),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [anon_sym_POUNDendif] = ACTIONS(6291),
    [anon_sym_POUNDelse] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5557)] = {
    [sym_block_comment] = STATE(5557),
    [sym_compiler_directive_decl] = STATE(5557),
    [sym_fsi_directive_decl] = STATE(5557),
    [sym_preproc_line] = STATE(5557),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_GT] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9753),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5558)] = {
    [sym_block_comment] = STATE(5558),
    [sym_compiler_directive_decl] = STATE(5558),
    [sym_fsi_directive_decl] = STATE(5558),
    [sym_preproc_line] = STATE(5558),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__quoted_close] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5559)] = {
    [sym_block_comment] = STATE(5559),
    [sym_compiler_directive_decl] = STATE(5559),
    [sym_fsi_directive_decl] = STATE(5559),
    [sym_preproc_line] = STATE(5559),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_DOT_DOT] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__dedent] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(5560)] = {
    [sym_block_comment] = STATE(5560),
    [sym_compiler_directive_decl] = STATE(5560),
    [sym_fsi_directive_decl] = STATE(5560),
    [sym_preproc_line] = STATE(5560),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_DOT_DOT] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__dedent] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(5561)] = {
    [sym_block_comment] = STATE(5561),
    [sym_compiler_directive_decl] = STATE(5561),
    [sym_fsi_directive_decl] = STATE(5561),
    [sym_preproc_line] = STATE(5561),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9415),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__untyped_quoted_close] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5562)] = {
    [sym_block_comment] = STATE(5562),
    [sym_compiler_directive_decl] = STATE(5562),
    [sym_fsi_directive_decl] = STATE(5562),
    [sym_preproc_line] = STATE(5562),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_with] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5563)] = {
    [sym_block_comment] = STATE(5563),
    [sym_compiler_directive_decl] = STATE(5563),
    [sym_fsi_directive_decl] = STATE(5563),
    [sym_preproc_line] = STATE(5563),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_DOT_DOT] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__dedent] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(5564)] = {
    [sym_block_comment] = STATE(5564),
    [sym_compiler_directive_decl] = STATE(5564),
    [sym_fsi_directive_decl] = STATE(5564),
    [sym_preproc_line] = STATE(5564),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_DOT_DOT] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__dedent] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(5565)] = {
    [sym_block_comment] = STATE(5565),
    [sym_compiler_directive_decl] = STATE(5565),
    [sym_fsi_directive_decl] = STATE(5565),
    [sym_preproc_line] = STATE(5565),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_DOT_DOT] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__dedent] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(5566)] = {
    [sym_block_comment] = STATE(5566),
    [sym_compiler_directive_decl] = STATE(5566),
    [sym_fsi_directive_decl] = STATE(5566),
    [sym_preproc_line] = STATE(5566),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_DOT_DOT] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__dedent] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(5567)] = {
    [sym_block_comment] = STATE(5567),
    [sym_compiler_directive_decl] = STATE(5567),
    [sym_fsi_directive_decl] = STATE(5567),
    [sym_preproc_line] = STATE(5567),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_DOT_DOT] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__dedent] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(5568)] = {
    [sym_block_comment] = STATE(5568),
    [sym_compiler_directive_decl] = STATE(5568),
    [sym_fsi_directive_decl] = STATE(5568),
    [sym_preproc_line] = STATE(5568),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_DOT_DOT] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__dedent] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(5569)] = {
    [sym_block_comment] = STATE(5569),
    [sym_compiler_directive_decl] = STATE(5569),
    [sym_fsi_directive_decl] = STATE(5569),
    [sym_preproc_line] = STATE(5569),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_DOT_DOT] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__dedent] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(5570)] = {
    [sym_block_comment] = STATE(5570),
    [sym_compiler_directive_decl] = STATE(5570),
    [sym_fsi_directive_decl] = STATE(5570),
    [sym_preproc_line] = STATE(5570),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_DOT_DOT] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__dedent] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(5571)] = {
    [sym_block_comment] = STATE(5571),
    [sym_compiler_directive_decl] = STATE(5571),
    [sym_fsi_directive_decl] = STATE(5571),
    [sym_preproc_line] = STATE(5571),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_DOT_DOT] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__dedent] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(5572)] = {
    [sym_block_comment] = STATE(5572),
    [sym_compiler_directive_decl] = STATE(5572),
    [sym_fsi_directive_decl] = STATE(5572),
    [sym_preproc_line] = STATE(5572),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_DOT_DOT] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__dedent] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(5573)] = {
    [sym_block_comment] = STATE(5573),
    [sym_compiler_directive_decl] = STATE(5573),
    [sym_fsi_directive_decl] = STATE(5573),
    [sym_preproc_line] = STATE(5573),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_DOT_DOT] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__dedent] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(5574)] = {
    [sym_block_comment] = STATE(5574),
    [sym_compiler_directive_decl] = STATE(5574),
    [sym_fsi_directive_decl] = STATE(5574),
    [sym_preproc_line] = STATE(5574),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_DOT_DOT] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5575)] = {
    [sym_block_comment] = STATE(5575),
    [sym_compiler_directive_decl] = STATE(5575),
    [sym_fsi_directive_decl] = STATE(5575),
    [sym_preproc_line] = STATE(5575),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_DOT_DOT] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5576)] = {
    [sym_block_comment] = STATE(5576),
    [sym_compiler_directive_decl] = STATE(5576),
    [sym_fsi_directive_decl] = STATE(5576),
    [sym_preproc_line] = STATE(5576),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__dedent] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
    [sym__in] = ACTIONS(6805),
  },
  [STATE(5577)] = {
    [sym_block_comment] = STATE(5577),
    [sym_compiler_directive_decl] = STATE(5577),
    [sym_fsi_directive_decl] = STATE(5577),
    [sym_preproc_line] = STATE(5577),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_DOT_DOT] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__dedent] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(5578)] = {
    [sym_block_comment] = STATE(5578),
    [sym_compiler_directive_decl] = STATE(5578),
    [sym_fsi_directive_decl] = STATE(5578),
    [sym_preproc_line] = STATE(5578),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_DOT_DOT] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__dedent] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(5579)] = {
    [sym_block_comment] = STATE(5579),
    [sym_compiler_directive_decl] = STATE(5579),
    [sym_fsi_directive_decl] = STATE(5579),
    [sym_preproc_line] = STATE(5579),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_DOT_DOT] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__dedent] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(5580)] = {
    [sym_block_comment] = STATE(5580),
    [sym_compiler_directive_decl] = STATE(5580),
    [sym_fsi_directive_decl] = STATE(5580),
    [sym_preproc_line] = STATE(5580),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_DOT_DOT] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__dedent] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(5581)] = {
    [sym_block_comment] = STATE(5581),
    [sym_compiler_directive_decl] = STATE(5581),
    [sym_fsi_directive_decl] = STATE(5581),
    [sym_preproc_line] = STATE(5581),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__dedent] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
    [sym__in] = ACTIONS(6890),
  },
  [STATE(5582)] = {
    [sym_block_comment] = STATE(5582),
    [sym_compiler_directive_decl] = STATE(5582),
    [sym_fsi_directive_decl] = STATE(5582),
    [sym_preproc_line] = STATE(5582),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_DOT_DOT] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__dedent] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(5583)] = {
    [sym_block_comment] = STATE(5583),
    [sym_compiler_directive_decl] = STATE(5583),
    [sym_fsi_directive_decl] = STATE(5583),
    [sym_preproc_line] = STATE(5583),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_DOT_DOT] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__dedent] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(5584)] = {
    [sym_block_comment] = STATE(5584),
    [sym_compiler_directive_decl] = STATE(5584),
    [sym_fsi_directive_decl] = STATE(5584),
    [sym_preproc_line] = STATE(5584),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_DOT_DOT] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5585)] = {
    [sym_block_comment] = STATE(5585),
    [sym_compiler_directive_decl] = STATE(5585),
    [sym_fsi_directive_decl] = STATE(5585),
    [sym_preproc_line] = STATE(5585),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_DOT_DOT] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__dedent] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(5586)] = {
    [sym_block_comment] = STATE(5586),
    [sym_compiler_directive_decl] = STATE(5586),
    [sym_fsi_directive_decl] = STATE(5586),
    [sym_preproc_line] = STATE(5586),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_DOT_DOT] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__dedent] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(5587)] = {
    [sym_block_comment] = STATE(5587),
    [sym_compiler_directive_decl] = STATE(5587),
    [sym_fsi_directive_decl] = STATE(5587),
    [sym_preproc_line] = STATE(5587),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_DOT_DOT] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__dedent] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(5588)] = {
    [sym_block_comment] = STATE(5588),
    [sym_compiler_directive_decl] = STATE(5588),
    [sym_fsi_directive_decl] = STATE(5588),
    [sym_preproc_line] = STATE(5588),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_DOT_DOT] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__dedent] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(5589)] = {
    [sym_block_comment] = STATE(5589),
    [sym_compiler_directive_decl] = STATE(5589),
    [sym_fsi_directive_decl] = STATE(5589),
    [sym_preproc_line] = STATE(5589),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_DOT_DOT] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5590)] = {
    [sym_block_comment] = STATE(5590),
    [sym_compiler_directive_decl] = STATE(5590),
    [sym_fsi_directive_decl] = STATE(5590),
    [sym_preproc_line] = STATE(5590),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_GT] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5591)] = {
    [sym_block_comment] = STATE(5591),
    [sym_compiler_directive_decl] = STATE(5591),
    [sym_fsi_directive_decl] = STATE(5591),
    [sym_preproc_line] = STATE(5591),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_DOT_DOT] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__dedent] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(5592)] = {
    [sym_block_comment] = STATE(5592),
    [sym_compiler_directive_decl] = STATE(5592),
    [sym_fsi_directive_decl] = STATE(5592),
    [sym_preproc_line] = STATE(5592),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__dedent] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
    [sym__in] = ACTIONS(6894),
  },
  [STATE(5593)] = {
    [sym_block_comment] = STATE(5593),
    [sym_compiler_directive_decl] = STATE(5593),
    [sym_fsi_directive_decl] = STATE(5593),
    [sym_preproc_line] = STATE(5593),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__dedent] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
    [sym__in] = ACTIONS(6898),
  },
  [STATE(5594)] = {
    [sym_block_comment] = STATE(5594),
    [sym_compiler_directive_decl] = STATE(5594),
    [sym_fsi_directive_decl] = STATE(5594),
    [sym_preproc_line] = STATE(5594),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [anon_sym_POUNDendif] = ACTIONS(8449),
    [anon_sym_POUNDelse] = ACTIONS(8449),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(5595)] = {
    [sym_block_comment] = STATE(5595),
    [sym_compiler_directive_decl] = STATE(5595),
    [sym_fsi_directive_decl] = STATE(5595),
    [sym_preproc_line] = STATE(5595),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__dedent] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
    [sym__in] = ACTIONS(6603),
  },
  [STATE(5596)] = {
    [sym_block_comment] = STATE(5596),
    [sym_compiler_directive_decl] = STATE(5596),
    [sym_fsi_directive_decl] = STATE(5596),
    [sym_preproc_line] = STATE(5596),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_DOT_DOT] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5597)] = {
    [sym_block_comment] = STATE(5597),
    [sym_compiler_directive_decl] = STATE(5597),
    [sym_fsi_directive_decl] = STATE(5597),
    [sym_preproc_line] = STATE(5597),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__quoted_close] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5598)] = {
    [sym_block_comment] = STATE(5598),
    [sym_compiler_directive_decl] = STATE(5598),
    [sym_fsi_directive_decl] = STATE(5598),
    [sym_preproc_line] = STATE(5598),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__dedent] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
    [sym__in] = ACTIONS(6904),
  },
  [STATE(5599)] = {
    [sym_block_comment] = STATE(5599),
    [sym_compiler_directive_decl] = STATE(5599),
    [sym_fsi_directive_decl] = STATE(5599),
    [sym_preproc_line] = STATE(5599),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [anon_sym_POUNDendif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5600)] = {
    [sym_block_comment] = STATE(5600),
    [sym_compiler_directive_decl] = STATE(5600),
    [sym_fsi_directive_decl] = STATE(5600),
    [sym_preproc_line] = STATE(5600),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_module] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_open] = ACTIONS(5951),
    [anon_sym_type] = ACTIONS(5951),
    [anon_sym_LBRACK_LT] = ACTIONS(5953),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_and] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_PIPE] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5953),
    [anon_sym_CARET] = ACTIONS(5953),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_AMP_AMP] = ACTIONS(5953),
    [anon_sym_exception] = ACTIONS(5951),
    [anon_sym_extern] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5953),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5953),
    [anon_sym_DASH_DOT] = ACTIONS(5953),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5953),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5953),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_y] = ACTIONS(9755),
    [anon_sym_uy] = ACTIONS(9757),
    [anon_sym_s] = ACTIONS(9759),
    [anon_sym_us] = ACTIONS(9761),
    [anon_sym_l] = ACTIONS(9763),
    [aux_sym_uint32_token1] = ACTIONS(9765),
    [anon_sym_n] = ACTIONS(9767),
    [anon_sym_un] = ACTIONS(9769),
    [anon_sym_L] = ACTIONS(9771),
    [aux_sym_uint64_token1] = ACTIONS(9773),
    [aux_sym_bignum_token1] = ACTIONS(9775),
    [aux_sym_decimal_token1] = ACTIONS(9777),
    [anon_sym_DOT2] = ACTIONS(9779),
    [aux_sym_float_token1] = ACTIONS(9781),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(5953),
    [anon_sym_POUNDload] = ACTIONS(5953),
    [anon_sym_POUNDtime] = ACTIONS(5953),
    [anon_sym_POUNDI] = ACTIONS(5953),
    [anon_sym_POUNDhelp] = ACTIONS(5953),
    [anon_sym_POUNDquit] = ACTIONS(5953),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [anon_sym_POUNDendif] = ACTIONS(5953),
    [anon_sym_POUNDelse] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
  },
  [STATE(5601)] = {
    [sym_block_comment] = STATE(5601),
    [sym_compiler_directive_decl] = STATE(5601),
    [sym_fsi_directive_decl] = STATE(5601),
    [sym_preproc_line] = STATE(5601),
    [aux_sym_long_identifier_repeat1] = STATE(5602),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_PIPE] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6358),
    [anon_sym_CARET] = ACTIONS(6358),
    [anon_sym_DOT] = ACTIONS(9670),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6358),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_AMP_AMP] = ACTIONS(6358),
    [anon_sym_STAR] = ACTIONS(6358),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6358),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6358),
    [anon_sym_DASH_DOT] = ACTIONS(6358),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6358),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6358),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [anon_sym_POUNDelse] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
  },
  [STATE(5602)] = {
    [sym_block_comment] = STATE(5602),
    [sym_compiler_directive_decl] = STATE(5602),
    [sym_fsi_directive_decl] = STATE(5602),
    [sym_preproc_line] = STATE(5602),
    [aux_sym_long_identifier_repeat1] = STATE(5602),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_PIPE] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6297),
    [anon_sym_CARET] = ACTIONS(6297),
    [anon_sym_DOT] = ACTIONS(9783),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6297),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_AMP_AMP] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6297),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6297),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6297),
    [anon_sym_DASH_DOT] = ACTIONS(6297),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6297),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6297),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
  },
  [STATE(5603)] = {
    [sym_block_comment] = STATE(5603),
    [sym_compiler_directive_decl] = STATE(5603),
    [sym_fsi_directive_decl] = STATE(5603),
    [sym_preproc_line] = STATE(5603),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_DOT_DOT] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__dedent] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(5604)] = {
    [sym_block_comment] = STATE(5604),
    [sym_compiler_directive_decl] = STATE(5604),
    [sym_fsi_directive_decl] = STATE(5604),
    [sym_preproc_line] = STATE(5604),
    [sym_identifier] = ACTIONS(6369),
    [anon_sym_module] = ACTIONS(6369),
    [anon_sym_null] = ACTIONS(6369),
    [anon_sym_begin] = ACTIONS(6369),
    [anon_sym_open] = ACTIONS(6369),
    [anon_sym_type] = ACTIONS(6369),
    [anon_sym_LBRACK_LT] = ACTIONS(6371),
    [anon_sym_do] = ACTIONS(6369),
    [anon_sym_and] = ACTIONS(6369),
    [anon_sym_let] = ACTIONS(6369),
    [anon_sym_let_BANG] = ACTIONS(6371),
    [aux_sym_access_modifier_token1] = ACTIONS(6371),
    [anon_sym_LPAREN] = ACTIONS(6369),
    [anon_sym_PIPE] = ACTIONS(6371),
    [anon_sym_AMP] = ACTIONS(6369),
    [anon_sym_LBRACK] = ACTIONS(6369),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6371),
    [anon_sym_LBRACE] = ACTIONS(6369),
    [anon_sym_LT_AT] = ACTIONS(6369),
    [anon_sym_LT_AT_AT] = ACTIONS(6371),
    [anon_sym_CARET] = ACTIONS(6371),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6371),
    [anon_sym_new] = ACTIONS(6369),
    [anon_sym_with] = ACTIONS(6369),
    [anon_sym_return] = ACTIONS(6369),
    [anon_sym_return_BANG] = ACTIONS(6371),
    [anon_sym_yield] = ACTIONS(6369),
    [anon_sym_yield_BANG] = ACTIONS(6371),
    [anon_sym_lazy] = ACTIONS(6369),
    [anon_sym_assert] = ACTIONS(6369),
    [anon_sym_upcast] = ACTIONS(6369),
    [anon_sym_downcast] = ACTIONS(6369),
    [anon_sym_fixed] = ACTIONS(6369),
    [anon_sym_COLON_GT] = ACTIONS(9786),
    [anon_sym_for] = ACTIONS(6369),
    [anon_sym_while] = ACTIONS(6369),
    [anon_sym_while_BANG] = ACTIONS(6371),
    [anon_sym_if] = ACTIONS(6369),
    [anon_sym_fun] = ACTIONS(6369),
    [anon_sym_DASH_GT] = ACTIONS(6371),
    [anon_sym_try] = ACTIONS(6369),
    [anon_sym_match] = ACTIONS(6369),
    [anon_sym_match_BANG] = ACTIONS(6371),
    [anon_sym_function] = ACTIONS(6369),
    [anon_sym_use] = ACTIONS(6369),
    [anon_sym_use_BANG] = ACTIONS(6371),
    [anon_sym_do_BANG] = ACTIONS(6371),
    [anon_sym_AMP_AMP] = ACTIONS(6371),
    [anon_sym_STAR] = ACTIONS(6371),
    [anon_sym_LT] = ACTIONS(6369),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6371),
    [anon_sym_default] = ACTIONS(6369),
    [anon_sym_static] = ACTIONS(6369),
    [anon_sym_member] = ACTIONS(6369),
    [anon_sym_exception] = ACTIONS(6369),
    [anon_sym_interface] = ACTIONS(6369),
    [anon_sym_abstract] = ACTIONS(6369),
    [anon_sym_val] = ACTIONS(6369),
    [anon_sym_override] = ACTIONS(6369),
    [anon_sym_extern] = ACTIONS(6369),
    [aux_sym_char_token1] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6369),
    [anon_sym_DQUOTE] = ACTIONS(6369),
    [anon_sym_AT_DQUOTE] = ACTIONS(6371),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6371),
    [sym_bool] = ACTIONS(6369),
    [sym_unit] = ACTIONS(6371),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6369),
    [sym_op_identifier] = ACTIONS(6371),
    [anon_sym_PLUS] = ACTIONS(6369),
    [anon_sym_DASH] = ACTIONS(6369),
    [anon_sym_PLUS_DOT] = ACTIONS(6371),
    [anon_sym_DASH_DOT] = ACTIONS(6371),
    [anon_sym_PERCENT] = ACTIONS(6369),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6371),
    [anon_sym_TILDE] = ACTIONS(6371),
    [aux_sym_prefix_op_token1] = ACTIONS(6371),
    [sym_int] = ACTIONS(6369),
    [sym_xint] = ACTIONS(6371),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6371),
    [anon_sym_POUNDload] = ACTIONS(6371),
    [anon_sym_POUNDtime] = ACTIONS(6371),
    [anon_sym_POUNDI] = ACTIONS(6371),
    [anon_sym_POUNDhelp] = ACTIONS(6371),
    [anon_sym_POUNDquit] = ACTIONS(6371),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6371),
    [anon_sym_POUNDendif] = ACTIONS(6371),
    [anon_sym_POUNDelse] = ACTIONS(6371),
    [sym__newline] = ACTIONS(6371),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6371),
  },
  [STATE(5605)] = {
    [sym_block_comment] = STATE(5605),
    [sym_compiler_directive_decl] = STATE(5605),
    [sym_fsi_directive_decl] = STATE(5605),
    [sym_preproc_line] = STATE(5605),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_DOT_DOT] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__dedent] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(5606)] = {
    [sym_block_comment] = STATE(5606),
    [sym_compiler_directive_decl] = STATE(5606),
    [sym_fsi_directive_decl] = STATE(5606),
    [sym_preproc_line] = STATE(5606),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_DOT_DOT] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__dedent] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(5607)] = {
    [sym_block_comment] = STATE(5607),
    [sym_compiler_directive_decl] = STATE(5607),
    [sym_fsi_directive_decl] = STATE(5607),
    [sym_preproc_line] = STATE(5607),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__dedent] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
    [sym__in] = ACTIONS(6911),
  },
  [STATE(5608)] = {
    [sym_block_comment] = STATE(5608),
    [sym_compiler_directive_decl] = STATE(5608),
    [sym_fsi_directive_decl] = STATE(5608),
    [sym_preproc_line] = STATE(5608),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__dedent] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
    [sym__in] = ACTIONS(6915),
  },
  [STATE(5609)] = {
    [sym_block_comment] = STATE(5609),
    [sym_compiler_directive_decl] = STATE(5609),
    [sym_fsi_directive_decl] = STATE(5609),
    [sym_preproc_line] = STATE(5609),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__dedent] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
    [sym__in] = ACTIONS(6919),
  },
  [STATE(5610)] = {
    [sym_block_comment] = STATE(5610),
    [sym_compiler_directive_decl] = STATE(5610),
    [sym_fsi_directive_decl] = STATE(5610),
    [sym_preproc_line] = STATE(5610),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__dedent] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
    [sym__in] = ACTIONS(6923),
  },
  [STATE(5611)] = {
    [sym_block_comment] = STATE(5611),
    [sym_compiler_directive_decl] = STATE(5611),
    [sym_fsi_directive_decl] = STATE(5611),
    [sym_preproc_line] = STATE(5611),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__multi_dollar_interp_end] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5612)] = {
    [sym_block_comment] = STATE(5612),
    [sym_compiler_directive_decl] = STATE(5612),
    [sym_fsi_directive_decl] = STATE(5612),
    [sym_preproc_line] = STATE(5612),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_PIPE] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6297),
    [anon_sym_CARET] = ACTIONS(6297),
    [anon_sym_DOT] = ACTIONS(6297),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6297),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_AMP_AMP] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6297),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6297),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6297),
    [anon_sym_DASH_DOT] = ACTIONS(6297),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6297),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6297),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [anon_sym_POUNDelse] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
  },
  [STATE(5613)] = {
    [sym_block_comment] = STATE(5613),
    [sym_compiler_directive_decl] = STATE(5613),
    [sym_fsi_directive_decl] = STATE(5613),
    [sym_preproc_line] = STATE(5613),
    [aux_sym_compound_type_repeat1] = STATE(5624),
    [sym_identifier] = ACTIONS(6338),
    [anon_sym_module] = ACTIONS(6338),
    [anon_sym_null] = ACTIONS(6338),
    [anon_sym_begin] = ACTIONS(6338),
    [anon_sym_open] = ACTIONS(6338),
    [anon_sym_type] = ACTIONS(6338),
    [anon_sym_LBRACK_LT] = ACTIONS(6340),
    [anon_sym_do] = ACTIONS(6338),
    [anon_sym_and] = ACTIONS(6338),
    [anon_sym_let] = ACTIONS(6338),
    [anon_sym_let_BANG] = ACTIONS(6340),
    [aux_sym_access_modifier_token1] = ACTIONS(6340),
    [anon_sym_LPAREN] = ACTIONS(6338),
    [anon_sym_PIPE] = ACTIONS(6340),
    [anon_sym_AMP] = ACTIONS(6338),
    [anon_sym_LBRACK] = ACTIONS(6338),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6340),
    [anon_sym_LBRACE] = ACTIONS(6338),
    [anon_sym_LT_AT] = ACTIONS(6338),
    [anon_sym_LT_AT_AT] = ACTIONS(6340),
    [anon_sym_CARET] = ACTIONS(6340),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6340),
    [anon_sym_new] = ACTIONS(6338),
    [anon_sym_with] = ACTIONS(6338),
    [anon_sym_return] = ACTIONS(6338),
    [anon_sym_return_BANG] = ACTIONS(6340),
    [anon_sym_yield] = ACTIONS(6338),
    [anon_sym_yield_BANG] = ACTIONS(6340),
    [anon_sym_lazy] = ACTIONS(6338),
    [anon_sym_assert] = ACTIONS(6338),
    [anon_sym_upcast] = ACTIONS(6338),
    [anon_sym_downcast] = ACTIONS(6338),
    [anon_sym_fixed] = ACTIONS(6338),
    [anon_sym_for] = ACTIONS(6338),
    [anon_sym_while] = ACTIONS(6338),
    [anon_sym_while_BANG] = ACTIONS(6340),
    [anon_sym_if] = ACTIONS(6338),
    [anon_sym_fun] = ACTIONS(6338),
    [anon_sym_DASH_GT] = ACTIONS(6340),
    [anon_sym_try] = ACTIONS(6338),
    [anon_sym_match] = ACTIONS(6338),
    [anon_sym_match_BANG] = ACTIONS(6340),
    [anon_sym_function] = ACTIONS(6338),
    [anon_sym_use] = ACTIONS(6338),
    [anon_sym_use_BANG] = ACTIONS(6340),
    [anon_sym_do_BANG] = ACTIONS(6340),
    [anon_sym_AMP_AMP] = ACTIONS(6340),
    [anon_sym_STAR] = ACTIONS(9188),
    [anon_sym_LT] = ACTIONS(6338),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6340),
    [anon_sym_default] = ACTIONS(6338),
    [anon_sym_static] = ACTIONS(6338),
    [anon_sym_member] = ACTIONS(6338),
    [anon_sym_exception] = ACTIONS(6338),
    [anon_sym_interface] = ACTIONS(6338),
    [anon_sym_abstract] = ACTIONS(6338),
    [anon_sym_val] = ACTIONS(6338),
    [anon_sym_override] = ACTIONS(6338),
    [anon_sym_extern] = ACTIONS(6338),
    [aux_sym_char_token1] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6338),
    [anon_sym_DQUOTE] = ACTIONS(6338),
    [anon_sym_AT_DQUOTE] = ACTIONS(6340),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6340),
    [sym_bool] = ACTIONS(6338),
    [sym_unit] = ACTIONS(6340),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6338),
    [sym_op_identifier] = ACTIONS(6340),
    [anon_sym_PLUS] = ACTIONS(6338),
    [anon_sym_DASH] = ACTIONS(6338),
    [anon_sym_PLUS_DOT] = ACTIONS(6340),
    [anon_sym_DASH_DOT] = ACTIONS(6340),
    [anon_sym_PERCENT] = ACTIONS(6338),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6340),
    [anon_sym_TILDE] = ACTIONS(6340),
    [aux_sym_prefix_op_token1] = ACTIONS(6340),
    [sym_int] = ACTIONS(6338),
    [sym_xint] = ACTIONS(6340),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6340),
    [anon_sym_POUNDload] = ACTIONS(6340),
    [anon_sym_POUNDtime] = ACTIONS(6340),
    [anon_sym_POUNDI] = ACTIONS(6340),
    [anon_sym_POUNDhelp] = ACTIONS(6340),
    [anon_sym_POUNDquit] = ACTIONS(6340),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6340),
    [anon_sym_POUNDendif] = ACTIONS(6340),
    [anon_sym_POUNDelse] = ACTIONS(6340),
    [sym__newline] = ACTIONS(6340),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6340),
  },
  [STATE(5614)] = {
    [sym_block_comment] = STATE(5614),
    [sym_compiler_directive_decl] = STATE(5614),
    [sym_fsi_directive_decl] = STATE(5614),
    [sym_preproc_line] = STATE(5614),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__multi_dollar_interp_end] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5615)] = {
    [sym_block_comment] = STATE(5615),
    [sym_compiler_directive_decl] = STATE(5615),
    [sym_fsi_directive_decl] = STATE(5615),
    [sym_preproc_line] = STATE(5615),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__quoted_close] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5616)] = {
    [sym_block_comment] = STATE(5616),
    [sym_compiler_directive_decl] = STATE(5616),
    [sym_fsi_directive_decl] = STATE(5616),
    [sym_preproc_line] = STATE(5616),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__dedent] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
    [sym__in] = ACTIONS(292),
  },
  [STATE(5617)] = {
    [sym_block_comment] = STATE(5617),
    [sym_compiler_directive_decl] = STATE(5617),
    [sym_fsi_directive_decl] = STATE(5617),
    [sym_preproc_line] = STATE(5617),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__dedent] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
    [sym__in] = ACTIONS(6930),
  },
  [STATE(5618)] = {
    [sym_block_comment] = STATE(5618),
    [sym_compiler_directive_decl] = STATE(5618),
    [sym_fsi_directive_decl] = STATE(5618),
    [sym_preproc_line] = STATE(5618),
    [aux_sym_long_identifier_repeat1] = STATE(5619),
    [sym_identifier] = ACTIONS(6360),
    [anon_sym_module] = ACTIONS(6360),
    [anon_sym_null] = ACTIONS(6360),
    [anon_sym_begin] = ACTIONS(6360),
    [anon_sym_open] = ACTIONS(6360),
    [anon_sym_type] = ACTIONS(6360),
    [anon_sym_LBRACK_LT] = ACTIONS(6358),
    [anon_sym_do] = ACTIONS(6360),
    [anon_sym_and] = ACTIONS(6360),
    [anon_sym_let] = ACTIONS(6360),
    [anon_sym_let_BANG] = ACTIONS(6358),
    [aux_sym_access_modifier_token1] = ACTIONS(6358),
    [anon_sym_LPAREN] = ACTIONS(6360),
    [anon_sym_PIPE] = ACTIONS(6358),
    [anon_sym_AMP] = ACTIONS(6360),
    [anon_sym_LBRACK] = ACTIONS(6360),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6358),
    [anon_sym_LBRACE] = ACTIONS(6360),
    [anon_sym_LT_AT] = ACTIONS(6360),
    [anon_sym_LT_AT_AT] = ACTIONS(6358),
    [anon_sym_CARET] = ACTIONS(6358),
    [anon_sym_DOT] = ACTIONS(9684),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6358),
    [anon_sym_new] = ACTIONS(6360),
    [anon_sym_with] = ACTIONS(6360),
    [anon_sym_return] = ACTIONS(6360),
    [anon_sym_return_BANG] = ACTIONS(6358),
    [anon_sym_yield] = ACTIONS(6360),
    [anon_sym_yield_BANG] = ACTIONS(6358),
    [anon_sym_lazy] = ACTIONS(6360),
    [anon_sym_assert] = ACTIONS(6360),
    [anon_sym_upcast] = ACTIONS(6360),
    [anon_sym_downcast] = ACTIONS(6360),
    [anon_sym_fixed] = ACTIONS(6360),
    [anon_sym_for] = ACTIONS(6360),
    [anon_sym_while] = ACTIONS(6360),
    [anon_sym_while_BANG] = ACTIONS(6358),
    [anon_sym_if] = ACTIONS(6360),
    [anon_sym_fun] = ACTIONS(6360),
    [anon_sym_DASH_GT] = ACTIONS(6358),
    [anon_sym_try] = ACTIONS(6360),
    [anon_sym_match] = ACTIONS(6360),
    [anon_sym_match_BANG] = ACTIONS(6358),
    [anon_sym_function] = ACTIONS(6360),
    [anon_sym_use] = ACTIONS(6360),
    [anon_sym_use_BANG] = ACTIONS(6358),
    [anon_sym_do_BANG] = ACTIONS(6358),
    [anon_sym_AMP_AMP] = ACTIONS(6358),
    [anon_sym_STAR] = ACTIONS(6358),
    [anon_sym_LT] = ACTIONS(6360),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6358),
    [anon_sym_default] = ACTIONS(6360),
    [anon_sym_static] = ACTIONS(6360),
    [anon_sym_member] = ACTIONS(6360),
    [anon_sym_exception] = ACTIONS(6360),
    [anon_sym_interface] = ACTIONS(6360),
    [anon_sym_abstract] = ACTIONS(6360),
    [anon_sym_val] = ACTIONS(6360),
    [anon_sym_override] = ACTIONS(6360),
    [anon_sym_extern] = ACTIONS(6360),
    [aux_sym_char_token1] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6360),
    [anon_sym_DQUOTE] = ACTIONS(6360),
    [anon_sym_AT_DQUOTE] = ACTIONS(6358),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6358),
    [sym_bool] = ACTIONS(6360),
    [sym_unit] = ACTIONS(6358),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6360),
    [sym_op_identifier] = ACTIONS(6358),
    [anon_sym_PLUS] = ACTIONS(6360),
    [anon_sym_DASH] = ACTIONS(6360),
    [anon_sym_PLUS_DOT] = ACTIONS(6358),
    [anon_sym_DASH_DOT] = ACTIONS(6358),
    [anon_sym_PERCENT] = ACTIONS(6360),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6358),
    [anon_sym_TILDE] = ACTIONS(6358),
    [aux_sym_prefix_op_token1] = ACTIONS(6358),
    [sym_int] = ACTIONS(6360),
    [sym_xint] = ACTIONS(6358),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6358),
    [anon_sym_POUNDload] = ACTIONS(6358),
    [anon_sym_POUNDtime] = ACTIONS(6358),
    [anon_sym_POUNDI] = ACTIONS(6358),
    [anon_sym_POUNDhelp] = ACTIONS(6358),
    [anon_sym_POUNDquit] = ACTIONS(6358),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6358),
    [anon_sym_POUNDendif] = ACTIONS(6358),
    [sym__newline] = ACTIONS(6358),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6358),
  },
  [STATE(5619)] = {
    [sym_block_comment] = STATE(5619),
    [sym_compiler_directive_decl] = STATE(5619),
    [sym_fsi_directive_decl] = STATE(5619),
    [sym_preproc_line] = STATE(5619),
    [aux_sym_long_identifier_repeat1] = STATE(5619),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_module] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_open] = ACTIONS(6295),
    [anon_sym_type] = ACTIONS(6295),
    [anon_sym_LBRACK_LT] = ACTIONS(6297),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_and] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [aux_sym_access_modifier_token1] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_PIPE] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6297),
    [anon_sym_CARET] = ACTIONS(6297),
    [anon_sym_DOT] = ACTIONS(9788),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_with] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_DASH_GT] = ACTIONS(6297),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_AMP_AMP] = ACTIONS(6297),
    [anon_sym_STAR] = ACTIONS(6297),
    [anon_sym_LT] = ACTIONS(6295),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6297),
    [anon_sym_default] = ACTIONS(6295),
    [anon_sym_static] = ACTIONS(6295),
    [anon_sym_member] = ACTIONS(6295),
    [anon_sym_exception] = ACTIONS(6295),
    [anon_sym_interface] = ACTIONS(6295),
    [anon_sym_abstract] = ACTIONS(6295),
    [anon_sym_val] = ACTIONS(6295),
    [anon_sym_override] = ACTIONS(6295),
    [anon_sym_extern] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6297),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6297),
    [anon_sym_DASH_DOT] = ACTIONS(6297),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6297),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6297),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6297),
    [anon_sym_POUNDload] = ACTIONS(6297),
    [anon_sym_POUNDtime] = ACTIONS(6297),
    [anon_sym_POUNDI] = ACTIONS(6297),
    [anon_sym_POUNDhelp] = ACTIONS(6297),
    [anon_sym_POUNDquit] = ACTIONS(6297),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [anon_sym_POUNDendif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
  },
  [STATE(5620)] = {
    [sym_block_comment] = STATE(5620),
    [sym_compiler_directive_decl] = STATE(5620),
    [sym_fsi_directive_decl] = STATE(5620),
    [sym_preproc_line] = STATE(5620),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_with] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5621)] = {
    [sym_block_comment] = STATE(5621),
    [sym_compiler_directive_decl] = STATE(5621),
    [sym_fsi_directive_decl] = STATE(5621),
    [sym_preproc_line] = STATE(5621),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_DOT_DOT] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__dedent] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(5622)] = {
    [sym_block_comment] = STATE(5622),
    [sym_compiler_directive_decl] = STATE(5622),
    [sym_fsi_directive_decl] = STATE(5622),
    [sym_preproc_line] = STATE(5622),
    [aux_sym_sequential_expression_repeat1] = STATE(5756),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_with] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5623)] = {
    [sym_block_comment] = STATE(5623),
    [sym_compiler_directive_decl] = STATE(5623),
    [sym_fsi_directive_decl] = STATE(5623),
    [sym_preproc_line] = STATE(5623),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__quoted_close] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5624)] = {
    [sym_block_comment] = STATE(5624),
    [sym_compiler_directive_decl] = STATE(5624),
    [sym_fsi_directive_decl] = STATE(5624),
    [sym_preproc_line] = STATE(5624),
    [aux_sym_compound_type_repeat1] = STATE(5624),
    [sym_identifier] = ACTIONS(6193),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_PIPE] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6195),
    [anon_sym_CARET] = ACTIONS(6195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(6195),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_AMP_AMP] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(9791),
    [anon_sym_LT] = ACTIONS(6193),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6195),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6195),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6195),
    [anon_sym_DASH_DOT] = ACTIONS(6195),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6195),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6195),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [anon_sym_POUNDelse] = ACTIONS(6195),
    [sym__newline] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
  },
  [STATE(5625)] = {
    [sym_block_comment] = STATE(5625),
    [sym_compiler_directive_decl] = STATE(5625),
    [sym_fsi_directive_decl] = STATE(5625),
    [sym_preproc_line] = STATE(5625),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__dedent] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
    [sym__in] = ACTIONS(6942),
  },
  [STATE(5626)] = {
    [sym_block_comment] = STATE(5626),
    [sym_compiler_directive_decl] = STATE(5626),
    [sym_fsi_directive_decl] = STATE(5626),
    [sym_preproc_line] = STATE(5626),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__dedent] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
    [sym__in] = ACTIONS(6946),
  },
  [STATE(5627)] = {
    [sym_block_comment] = STATE(5627),
    [sym_compiler_directive_decl] = STATE(5627),
    [sym_fsi_directive_decl] = STATE(5627),
    [sym_preproc_line] = STATE(5627),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__dedent] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
    [sym__in] = ACTIONS(6950),
  },
  [STATE(5628)] = {
    [sym_block_comment] = STATE(5628),
    [sym_compiler_directive_decl] = STATE(5628),
    [sym_fsi_directive_decl] = STATE(5628),
    [sym_preproc_line] = STATE(5628),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__dedent] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
    [sym__in] = ACTIONS(6957),
  },
  [STATE(5629)] = {
    [sym_block_comment] = STATE(5629),
    [sym_compiler_directive_decl] = STATE(5629),
    [sym_fsi_directive_decl] = STATE(5629),
    [sym_preproc_line] = STATE(5629),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__dedent] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
    [sym__in] = ACTIONS(6961),
  },
  [STATE(5630)] = {
    [sym_block_comment] = STATE(5630),
    [sym_compiler_directive_decl] = STATE(5630),
    [sym_fsi_directive_decl] = STATE(5630),
    [sym_preproc_line] = STATE(5630),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__dedent] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
    [sym__in] = ACTIONS(6965),
  },
  [STATE(5631)] = {
    [sym_block_comment] = STATE(5631),
    [sym_compiler_directive_decl] = STATE(5631),
    [sym_fsi_directive_decl] = STATE(5631),
    [sym_preproc_line] = STATE(5631),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__dedent] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
    [sym__in] = ACTIONS(6969),
  },
  [STATE(5632)] = {
    [sym_block_comment] = STATE(5632),
    [sym_compiler_directive_decl] = STATE(5632),
    [sym_fsi_directive_decl] = STATE(5632),
    [sym_preproc_line] = STATE(5632),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__dedent] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
    [sym__in] = ACTIONS(6973),
  },
  [STATE(5633)] = {
    [sym_block_comment] = STATE(5633),
    [sym_compiler_directive_decl] = STATE(5633),
    [sym_fsi_directive_decl] = STATE(5633),
    [sym_preproc_line] = STATE(5633),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__dedent] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
    [sym__in] = ACTIONS(6977),
  },
  [STATE(5634)] = {
    [sym_block_comment] = STATE(5634),
    [sym_compiler_directive_decl] = STATE(5634),
    [sym_fsi_directive_decl] = STATE(5634),
    [sym_preproc_line] = STATE(5634),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__dedent] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
    [sym__in] = ACTIONS(6981),
  },
  [STATE(5635)] = {
    [sym_block_comment] = STATE(5635),
    [sym_compiler_directive_decl] = STATE(5635),
    [sym_fsi_directive_decl] = STATE(5635),
    [sym_preproc_line] = STATE(5635),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_GT] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5636)] = {
    [sym_block_comment] = STATE(5636),
    [sym_compiler_directive_decl] = STATE(5636),
    [sym_fsi_directive_decl] = STATE(5636),
    [sym_preproc_line] = STATE(5636),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__dedent] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
    [sym__in] = ACTIONS(6985),
  },
  [STATE(5637)] = {
    [sym_block_comment] = STATE(5637),
    [sym_compiler_directive_decl] = STATE(5637),
    [sym_fsi_directive_decl] = STATE(5637),
    [sym_preproc_line] = STATE(5637),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__dedent] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
    [sym__in] = ACTIONS(6989),
  },
  [STATE(5638)] = {
    [sym_block_comment] = STATE(5638),
    [sym_compiler_directive_decl] = STATE(5638),
    [sym_fsi_directive_decl] = STATE(5638),
    [sym_preproc_line] = STATE(5638),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__dedent] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
    [sym__in] = ACTIONS(6993),
  },
  [STATE(5639)] = {
    [sym_block_comment] = STATE(5639),
    [sym_compiler_directive_decl] = STATE(5639),
    [sym_fsi_directive_decl] = STATE(5639),
    [sym_preproc_line] = STATE(5639),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__multi_dollar_interp_end] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5640)] = {
    [sym_block_comment] = STATE(5640),
    [sym_compiler_directive_decl] = STATE(5640),
    [sym_fsi_directive_decl] = STATE(5640),
    [sym_preproc_line] = STATE(5640),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_GT] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5641)] = {
    [sym_block_comment] = STATE(5641),
    [sym_compiler_directive_decl] = STATE(5641),
    [sym_fsi_directive_decl] = STATE(5641),
    [sym_preproc_line] = STATE(5641),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN] = ACTIONS(6653),
    [anon_sym_COMMA] = ACTIONS(6651),
    [anon_sym_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_QMARK] = ACTIONS(6653),
    [anon_sym_COLON_COLON] = ACTIONS(6651),
    [anon_sym_PIPE] = ACTIONS(6653),
    [anon_sym_AMP] = ACTIONS(6653),
    [anon_sym_GT] = ACTIONS(6653),
    [anon_sym_LBRACK] = ACTIONS(6653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6651),
    [anon_sym_LBRACE] = ACTIONS(6653),
    [anon_sym_LT_AT] = ACTIONS(6653),
    [anon_sym_LT_AT_AT] = ACTIONS(6653),
    [anon_sym_CARET] = ACTIONS(6653),
    [anon_sym_or] = ACTIONS(6653),
    [anon_sym_DOT] = ACTIONS(6653),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6651),
    [anon_sym_new] = ACTIONS(6653),
    [anon_sym_return] = ACTIONS(6653),
    [anon_sym_return_BANG] = ACTIONS(6651),
    [anon_sym_yield] = ACTIONS(6653),
    [anon_sym_yield_BANG] = ACTIONS(6651),
    [anon_sym_lazy] = ACTIONS(6653),
    [anon_sym_assert] = ACTIONS(6653),
    [anon_sym_upcast] = ACTIONS(6653),
    [anon_sym_downcast] = ACTIONS(6653),
    [anon_sym_fixed] = ACTIONS(6653),
    [anon_sym_COLON_GT] = ACTIONS(6651),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6651),
    [anon_sym_for] = ACTIONS(6653),
    [anon_sym_while] = ACTIONS(6653),
    [anon_sym_while_BANG] = ACTIONS(6651),
    [anon_sym_if] = ACTIONS(6653),
    [anon_sym_fun] = ACTIONS(6653),
    [anon_sym_try] = ACTIONS(6653),
    [anon_sym_match] = ACTIONS(6653),
    [anon_sym_match_BANG] = ACTIONS(6651),
    [anon_sym_function] = ACTIONS(6653),
    [anon_sym_DOT_LBRACK] = ACTIONS(6651),
    [anon_sym_use] = ACTIONS(6653),
    [anon_sym_use_BANG] = ACTIONS(6651),
    [anon_sym_do_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN2] = ACTIONS(6653),
    [anon_sym_PIPE_PIPE] = ACTIONS(6653),
    [anon_sym_AMP_AMP] = ACTIONS(6653),
    [aux_sym_char_token1] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6653),
    [anon_sym_DQUOTE] = ACTIONS(6653),
    [anon_sym_AT_DQUOTE] = ACTIONS(6651),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6651),
    [sym_bool] = ACTIONS(6653),
    [sym_unit] = ACTIONS(6651),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6653),
    [sym_op_identifier] = ACTIONS(6653),
    [anon_sym_PLUS] = ACTIONS(6653),
    [anon_sym_DASH] = ACTIONS(6653),
    [anon_sym_PLUS_DOT] = ACTIONS(6653),
    [anon_sym_DASH_DOT] = ACTIONS(6653),
    [anon_sym_PERCENT] = ACTIONS(6653),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6653),
    [anon_sym_TILDE] = ACTIONS(6651),
    [aux_sym_prefix_op_token1] = ACTIONS(6653),
    [aux_sym_infix_op_token1] = ACTIONS(6651),
    [aux_sym_infix_op_token2] = ACTIONS(6653),
    [aux_sym_infix_op_token3] = ACTIONS(6653),
    [anon_sym_BANG_EQ] = ACTIONS(6653),
    [anon_sym_COLON_EQ] = ACTIONS(6651),
    [anon_sym_DOLLAR] = ACTIONS(6653),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6653),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6653),
    [sym_int] = ACTIONS(6653),
    [sym_xint] = ACTIONS(6651),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6651),
    [sym__newline] = ACTIONS(6651),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6651),
    [sym__tyapp_open] = ACTIONS(6651),
  },
  [STATE(5642)] = {
    [sym_block_comment] = STATE(5642),
    [sym_compiler_directive_decl] = STATE(5642),
    [sym_fsi_directive_decl] = STATE(5642),
    [sym_preproc_line] = STATE(5642),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__dedent] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
    [sym__in] = ACTIONS(6997),
  },
  [STATE(5643)] = {
    [sym_block_comment] = STATE(5643),
    [sym_compiler_directive_decl] = STATE(5643),
    [sym_fsi_directive_decl] = STATE(5643),
    [sym_preproc_line] = STATE(5643),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__dedent] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
    [sym__in] = ACTIONS(7001),
  },
  [STATE(5644)] = {
    [sym_block_comment] = STATE(5644),
    [sym_compiler_directive_decl] = STATE(5644),
    [sym_fsi_directive_decl] = STATE(5644),
    [sym_preproc_line] = STATE(5644),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__dedent] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
    [sym__in] = ACTIONS(7005),
  },
  [STATE(5645)] = {
    [sym_block_comment] = STATE(5645),
    [sym_compiler_directive_decl] = STATE(5645),
    [sym_fsi_directive_decl] = STATE(5645),
    [sym_preproc_line] = STATE(5645),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_with] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5646)] = {
    [sym_block_comment] = STATE(5646),
    [sym_compiler_directive_decl] = STATE(5646),
    [sym_fsi_directive_decl] = STATE(5646),
    [sym_preproc_line] = STATE(5646),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__dedent] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
    [sym__in] = ACTIONS(7009),
  },
  [STATE(5647)] = {
    [sym_block_comment] = STATE(5647),
    [sym_compiler_directive_decl] = STATE(5647),
    [sym_fsi_directive_decl] = STATE(5647),
    [sym_preproc_line] = STATE(5647),
    [aux_sym_type_argument_repeat1] = STATE(5412),
    [sym_identifier] = ACTIONS(6331),
    [anon_sym_module] = ACTIONS(6331),
    [anon_sym_null] = ACTIONS(6331),
    [anon_sym_begin] = ACTIONS(6331),
    [anon_sym_open] = ACTIONS(6331),
    [anon_sym_type] = ACTIONS(6331),
    [anon_sym_LBRACK_LT] = ACTIONS(6333),
    [anon_sym_do] = ACTIONS(6331),
    [anon_sym_and] = ACTIONS(6331),
    [anon_sym_let] = ACTIONS(6331),
    [anon_sym_let_BANG] = ACTIONS(6333),
    [aux_sym_access_modifier_token1] = ACTIONS(6333),
    [anon_sym_LPAREN] = ACTIONS(6331),
    [anon_sym_PIPE] = ACTIONS(6333),
    [anon_sym_AMP] = ACTIONS(6331),
    [anon_sym_LBRACK] = ACTIONS(6331),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6333),
    [anon_sym_LBRACE] = ACTIONS(6331),
    [anon_sym_LT_AT] = ACTIONS(6331),
    [anon_sym_LT_AT_AT] = ACTIONS(6333),
    [anon_sym_CARET] = ACTIONS(6333),
    [anon_sym_or] = ACTIONS(9708),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6333),
    [anon_sym_new] = ACTIONS(6331),
    [anon_sym_with] = ACTIONS(6331),
    [anon_sym_return] = ACTIONS(6331),
    [anon_sym_return_BANG] = ACTIONS(6333),
    [anon_sym_yield] = ACTIONS(6331),
    [anon_sym_yield_BANG] = ACTIONS(6333),
    [anon_sym_lazy] = ACTIONS(6331),
    [anon_sym_assert] = ACTIONS(6331),
    [anon_sym_upcast] = ACTIONS(6331),
    [anon_sym_downcast] = ACTIONS(6331),
    [anon_sym_fixed] = ACTIONS(6331),
    [anon_sym_COLON_GT] = ACTIONS(6333),
    [anon_sym_for] = ACTIONS(6331),
    [anon_sym_while] = ACTIONS(6331),
    [anon_sym_while_BANG] = ACTIONS(6333),
    [anon_sym_if] = ACTIONS(6331),
    [anon_sym_fun] = ACTIONS(6331),
    [anon_sym_DASH_GT] = ACTIONS(6333),
    [anon_sym_try] = ACTIONS(6331),
    [anon_sym_match] = ACTIONS(6331),
    [anon_sym_match_BANG] = ACTIONS(6333),
    [anon_sym_function] = ACTIONS(6331),
    [anon_sym_use] = ACTIONS(6331),
    [anon_sym_use_BANG] = ACTIONS(6333),
    [anon_sym_do_BANG] = ACTIONS(6333),
    [anon_sym_AMP_AMP] = ACTIONS(6333),
    [anon_sym_STAR] = ACTIONS(6333),
    [anon_sym_LT] = ACTIONS(6331),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6333),
    [anon_sym_default] = ACTIONS(6331),
    [anon_sym_static] = ACTIONS(6331),
    [anon_sym_member] = ACTIONS(6331),
    [anon_sym_exception] = ACTIONS(6331),
    [anon_sym_interface] = ACTIONS(6331),
    [anon_sym_abstract] = ACTIONS(6331),
    [anon_sym_val] = ACTIONS(6331),
    [anon_sym_override] = ACTIONS(6331),
    [anon_sym_extern] = ACTIONS(6331),
    [aux_sym_char_token1] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6331),
    [anon_sym_DQUOTE] = ACTIONS(6331),
    [anon_sym_AT_DQUOTE] = ACTIONS(6333),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6333),
    [sym_bool] = ACTIONS(6331),
    [sym_unit] = ACTIONS(6333),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6331),
    [sym_op_identifier] = ACTIONS(6333),
    [anon_sym_PLUS] = ACTIONS(6331),
    [anon_sym_DASH] = ACTIONS(6331),
    [anon_sym_PLUS_DOT] = ACTIONS(6333),
    [anon_sym_DASH_DOT] = ACTIONS(6333),
    [anon_sym_PERCENT] = ACTIONS(6331),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6333),
    [anon_sym_TILDE] = ACTIONS(6333),
    [aux_sym_prefix_op_token1] = ACTIONS(6333),
    [sym_int] = ACTIONS(6331),
    [sym_xint] = ACTIONS(6333),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6333),
    [anon_sym_POUNDload] = ACTIONS(6333),
    [anon_sym_POUNDtime] = ACTIONS(6333),
    [anon_sym_POUNDI] = ACTIONS(6333),
    [anon_sym_POUNDhelp] = ACTIONS(6333),
    [anon_sym_POUNDquit] = ACTIONS(6333),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6333),
    [anon_sym_POUNDendif] = ACTIONS(6333),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6333),
  },
  [STATE(5648)] = {
    [sym_type_arguments] = STATE(6569),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5648),
    [sym_compiler_directive_decl] = STATE(5648),
    [sym_fsi_directive_decl] = STATE(5648),
    [sym_preproc_line] = STATE(5648),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(6257),
    [anon_sym_null] = ACTIONS(6257),
    [anon_sym_begin] = ACTIONS(6257),
    [anon_sym_open] = ACTIONS(6257),
    [anon_sym_type] = ACTIONS(6257),
    [anon_sym_LBRACK_LT] = ACTIONS(6255),
    [anon_sym_do] = ACTIONS(6257),
    [anon_sym_and] = ACTIONS(6257),
    [anon_sym_let] = ACTIONS(6257),
    [anon_sym_let_BANG] = ACTIONS(6255),
    [aux_sym_access_modifier_token1] = ACTIONS(6255),
    [anon_sym_LPAREN] = ACTIONS(6257),
    [anon_sym_PIPE] = ACTIONS(6255),
    [anon_sym_AMP] = ACTIONS(6257),
    [anon_sym_LBRACK] = ACTIONS(6257),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6255),
    [anon_sym_LBRACE] = ACTIONS(6257),
    [anon_sym_LT_AT] = ACTIONS(6257),
    [anon_sym_LT_AT_AT] = ACTIONS(6255),
    [anon_sym_CARET] = ACTIONS(6255),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6255),
    [anon_sym_new] = ACTIONS(6257),
    [anon_sym_with] = ACTIONS(6257),
    [anon_sym_return] = ACTIONS(6257),
    [anon_sym_return_BANG] = ACTIONS(6255),
    [anon_sym_yield] = ACTIONS(6257),
    [anon_sym_yield_BANG] = ACTIONS(6255),
    [anon_sym_lazy] = ACTIONS(6257),
    [anon_sym_assert] = ACTIONS(6257),
    [anon_sym_upcast] = ACTIONS(6257),
    [anon_sym_downcast] = ACTIONS(6257),
    [anon_sym_fixed] = ACTIONS(6257),
    [anon_sym_for] = ACTIONS(6257),
    [anon_sym_while] = ACTIONS(6257),
    [anon_sym_while_BANG] = ACTIONS(6255),
    [anon_sym_if] = ACTIONS(6257),
    [anon_sym_fun] = ACTIONS(6257),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(6257),
    [anon_sym_match] = ACTIONS(6257),
    [anon_sym_match_BANG] = ACTIONS(6255),
    [anon_sym_function] = ACTIONS(6257),
    [anon_sym_use] = ACTIONS(6257),
    [anon_sym_use_BANG] = ACTIONS(6255),
    [anon_sym_do_BANG] = ACTIONS(6255),
    [anon_sym_AMP_AMP] = ACTIONS(6255),
    [anon_sym_STAR] = ACTIONS(9794),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(6257),
    [anon_sym_static] = ACTIONS(6257),
    [anon_sym_member] = ACTIONS(6257),
    [anon_sym_exception] = ACTIONS(6257),
    [anon_sym_interface] = ACTIONS(6257),
    [anon_sym_abstract] = ACTIONS(6257),
    [anon_sym_val] = ACTIONS(6257),
    [anon_sym_override] = ACTIONS(6257),
    [anon_sym_extern] = ACTIONS(6257),
    [aux_sym_char_token1] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6257),
    [anon_sym_DQUOTE] = ACTIONS(6257),
    [anon_sym_AT_DQUOTE] = ACTIONS(6255),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6255),
    [sym_bool] = ACTIONS(6257),
    [sym_unit] = ACTIONS(6255),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6257),
    [sym_op_identifier] = ACTIONS(6255),
    [anon_sym_PLUS] = ACTIONS(6257),
    [anon_sym_DASH] = ACTIONS(6257),
    [anon_sym_PLUS_DOT] = ACTIONS(6255),
    [anon_sym_DASH_DOT] = ACTIONS(6255),
    [anon_sym_PERCENT] = ACTIONS(6257),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6255),
    [anon_sym_TILDE] = ACTIONS(6255),
    [aux_sym_prefix_op_token1] = ACTIONS(6255),
    [sym_int] = ACTIONS(6257),
    [sym_xint] = ACTIONS(6255),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6255),
    [anon_sym_POUNDload] = ACTIONS(6255),
    [anon_sym_POUNDtime] = ACTIONS(6255),
    [anon_sym_POUNDI] = ACTIONS(6255),
    [anon_sym_POUNDhelp] = ACTIONS(6255),
    [anon_sym_POUNDquit] = ACTIONS(6255),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6255),
    [anon_sym_POUNDendif] = ACTIONS(6255),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6255),
  },
  [STATE(5649)] = {
    [sym_block_comment] = STATE(5649),
    [sym_compiler_directive_decl] = STATE(5649),
    [sym_fsi_directive_decl] = STATE(5649),
    [sym_preproc_line] = STATE(5649),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9726),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_DOT_DOT] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5650)] = {
    [sym_block_comment] = STATE(5650),
    [sym_compiler_directive_decl] = STATE(5650),
    [sym_fsi_directive_decl] = STATE(5650),
    [sym_preproc_line] = STATE(5650),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__dedent] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
    [sym__in] = ACTIONS(7017),
  },
  [STATE(5651)] = {
    [sym_block_comment] = STATE(5651),
    [sym_compiler_directive_decl] = STATE(5651),
    [sym_fsi_directive_decl] = STATE(5651),
    [sym_preproc_line] = STATE(5651),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9796),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5652)] = {
    [sym_block_comment] = STATE(5652),
    [sym_compiler_directive_decl] = STATE(5652),
    [sym_fsi_directive_decl] = STATE(5652),
    [sym_preproc_line] = STATE(5652),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(6834),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_DOT_DOT] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5653)] = {
    [sym_block_comment] = STATE(5653),
    [sym_compiler_directive_decl] = STATE(5653),
    [sym_fsi_directive_decl] = STATE(5653),
    [sym_preproc_line] = STATE(5653),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__dedent] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
    [sym__in] = ACTIONS(7021),
  },
  [STATE(5654)] = {
    [sym_block_comment] = STATE(5654),
    [sym_compiler_directive_decl] = STATE(5654),
    [sym_fsi_directive_decl] = STATE(5654),
    [sym_preproc_line] = STATE(5654),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__multi_dollar_interp_end] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5655)] = {
    [sym_block_comment] = STATE(5655),
    [sym_compiler_directive_decl] = STATE(5655),
    [sym_fsi_directive_decl] = STATE(5655),
    [sym_preproc_line] = STATE(5655),
    [sym_identifier] = ACTIONS(6813),
    [anon_sym_null] = ACTIONS(6813),
    [anon_sym_EQ] = ACTIONS(6813),
    [anon_sym_begin] = ACTIONS(6813),
    [anon_sym_COLON] = ACTIONS(6813),
    [anon_sym_do] = ACTIONS(6813),
    [anon_sym_let] = ACTIONS(6813),
    [anon_sym_let_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN] = ACTIONS(6813),
    [anon_sym_COMMA] = ACTIONS(6811),
    [anon_sym_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_QMARK] = ACTIONS(6813),
    [anon_sym_COLON_COLON] = ACTIONS(6811),
    [anon_sym_PIPE] = ACTIONS(6813),
    [anon_sym_AMP] = ACTIONS(6813),
    [anon_sym_GT] = ACTIONS(6813),
    [anon_sym_LBRACK] = ACTIONS(6813),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6811),
    [anon_sym_LBRACE] = ACTIONS(6813),
    [anon_sym_LT_AT] = ACTIONS(6813),
    [anon_sym_LT_AT_AT] = ACTIONS(6813),
    [anon_sym_CARET] = ACTIONS(6813),
    [anon_sym_or] = ACTIONS(6813),
    [anon_sym_DOT] = ACTIONS(6813),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6811),
    [anon_sym_new] = ACTIONS(6813),
    [anon_sym_return] = ACTIONS(6813),
    [anon_sym_return_BANG] = ACTIONS(6811),
    [anon_sym_yield] = ACTIONS(6813),
    [anon_sym_yield_BANG] = ACTIONS(6811),
    [anon_sym_lazy] = ACTIONS(6813),
    [anon_sym_assert] = ACTIONS(6813),
    [anon_sym_upcast] = ACTIONS(6813),
    [anon_sym_downcast] = ACTIONS(6813),
    [anon_sym_fixed] = ACTIONS(6813),
    [anon_sym_COLON_GT] = ACTIONS(6811),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6811),
    [anon_sym_for] = ACTIONS(6813),
    [anon_sym_while] = ACTIONS(6813),
    [anon_sym_while_BANG] = ACTIONS(6811),
    [anon_sym_if] = ACTIONS(6813),
    [anon_sym_fun] = ACTIONS(6813),
    [anon_sym_try] = ACTIONS(6813),
    [anon_sym_match] = ACTIONS(6813),
    [anon_sym_match_BANG] = ACTIONS(6811),
    [anon_sym_function] = ACTIONS(6813),
    [anon_sym_DOT_LBRACK] = ACTIONS(6811),
    [anon_sym_use] = ACTIONS(6813),
    [anon_sym_use_BANG] = ACTIONS(6811),
    [anon_sym_do_BANG] = ACTIONS(6811),
    [anon_sym_LPAREN2] = ACTIONS(6813),
    [anon_sym_PIPE_PIPE] = ACTIONS(6813),
    [anon_sym_AMP_AMP] = ACTIONS(6813),
    [aux_sym_char_token1] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6813),
    [anon_sym_DQUOTE] = ACTIONS(6813),
    [anon_sym_AT_DQUOTE] = ACTIONS(6811),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6811),
    [sym_bool] = ACTIONS(6813),
    [sym_unit] = ACTIONS(6811),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6813),
    [sym_op_identifier] = ACTIONS(6813),
    [anon_sym_PLUS] = ACTIONS(6813),
    [anon_sym_DASH] = ACTIONS(6813),
    [anon_sym_PLUS_DOT] = ACTIONS(6813),
    [anon_sym_DASH_DOT] = ACTIONS(6813),
    [anon_sym_PERCENT] = ACTIONS(6813),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6813),
    [anon_sym_TILDE] = ACTIONS(6811),
    [aux_sym_prefix_op_token1] = ACTIONS(6813),
    [aux_sym_infix_op_token1] = ACTIONS(6811),
    [aux_sym_infix_op_token2] = ACTIONS(6813),
    [aux_sym_infix_op_token3] = ACTIONS(6813),
    [anon_sym_BANG_EQ] = ACTIONS(6813),
    [anon_sym_COLON_EQ] = ACTIONS(6811),
    [anon_sym_DOLLAR] = ACTIONS(6813),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6813),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6813),
    [sym_int] = ACTIONS(6813),
    [sym_xint] = ACTIONS(6811),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6811),
    [sym__newline] = ACTIONS(6811),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6811),
    [sym__tyapp_open] = ACTIONS(6811),
  },
  [STATE(5656)] = {
    [sym_block_comment] = STATE(5656),
    [sym_compiler_directive_decl] = STATE(5656),
    [sym_fsi_directive_decl] = STATE(5656),
    [sym_preproc_line] = STATE(5656),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__dedent] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
    [sym__in] = ACTIONS(7025),
  },
  [STATE(5657)] = {
    [sym_block_comment] = STATE(5657),
    [sym_compiler_directive_decl] = STATE(5657),
    [sym_fsi_directive_decl] = STATE(5657),
    [sym_preproc_line] = STATE(5657),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_DOT_DOT] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5658)] = {
    [sym_block_comment] = STATE(5658),
    [sym_compiler_directive_decl] = STATE(5658),
    [sym_fsi_directive_decl] = STATE(5658),
    [sym_preproc_line] = STATE(5658),
    [sym_identifier] = ACTIONS(7053),
    [anon_sym_null] = ACTIONS(7053),
    [anon_sym_EQ] = ACTIONS(7053),
    [anon_sym_begin] = ACTIONS(7053),
    [anon_sym_COLON] = ACTIONS(7053),
    [anon_sym_do] = ACTIONS(7053),
    [anon_sym_let] = ACTIONS(7053),
    [anon_sym_let_BANG] = ACTIONS(7055),
    [anon_sym_as] = ACTIONS(7053),
    [anon_sym_LPAREN] = ACTIONS(7053),
    [anon_sym_COMMA] = ACTIONS(7055),
    [anon_sym_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_QMARK] = ACTIONS(7053),
    [anon_sym_COLON_COLON] = ACTIONS(7055),
    [anon_sym_AMP] = ACTIONS(7053),
    [anon_sym_LBRACK] = ACTIONS(7053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7055),
    [anon_sym_LBRACE] = ACTIONS(7053),
    [anon_sym_LT_AT] = ACTIONS(7053),
    [anon_sym_LT_AT_AT] = ACTIONS(7053),
    [anon_sym_CARET] = ACTIONS(7053),
    [anon_sym_or] = ACTIONS(7053),
    [anon_sym_DOT] = ACTIONS(7053),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7055),
    [anon_sym_new] = ACTIONS(7053),
    [anon_sym_with] = ACTIONS(7053),
    [anon_sym_return] = ACTIONS(7053),
    [anon_sym_return_BANG] = ACTIONS(7055),
    [anon_sym_yield] = ACTIONS(7053),
    [anon_sym_yield_BANG] = ACTIONS(7055),
    [anon_sym_lazy] = ACTIONS(7053),
    [anon_sym_assert] = ACTIONS(7053),
    [anon_sym_upcast] = ACTIONS(7053),
    [anon_sym_downcast] = ACTIONS(7053),
    [anon_sym_fixed] = ACTIONS(7053),
    [anon_sym_COLON_GT] = ACTIONS(7055),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7055),
    [anon_sym_for] = ACTIONS(7053),
    [anon_sym_while] = ACTIONS(7053),
    [anon_sym_while_BANG] = ACTIONS(7055),
    [anon_sym_if] = ACTIONS(7053),
    [anon_sym_fun] = ACTIONS(7053),
    [anon_sym_try] = ACTIONS(7053),
    [anon_sym_match] = ACTIONS(7053),
    [anon_sym_match_BANG] = ACTIONS(7055),
    [anon_sym_function] = ACTIONS(7053),
    [anon_sym_DOT_LBRACK] = ACTIONS(7055),
    [anon_sym_use] = ACTIONS(7053),
    [anon_sym_use_BANG] = ACTIONS(7055),
    [anon_sym_do_BANG] = ACTIONS(7055),
    [anon_sym_LPAREN2] = ACTIONS(7053),
    [anon_sym_PIPE_PIPE] = ACTIONS(7053),
    [anon_sym_AMP_AMP] = ACTIONS(7053),
    [aux_sym_char_token1] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7053),
    [anon_sym_DQUOTE] = ACTIONS(7053),
    [anon_sym_AT_DQUOTE] = ACTIONS(7055),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7055),
    [sym_bool] = ACTIONS(7053),
    [sym_unit] = ACTIONS(7055),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7053),
    [sym_op_identifier] = ACTIONS(7053),
    [anon_sym_PLUS] = ACTIONS(7053),
    [anon_sym_DASH] = ACTIONS(7053),
    [anon_sym_PLUS_DOT] = ACTIONS(7053),
    [anon_sym_DASH_DOT] = ACTIONS(7053),
    [anon_sym_PERCENT] = ACTIONS(7053),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7053),
    [anon_sym_TILDE] = ACTIONS(7055),
    [aux_sym_prefix_op_token1] = ACTIONS(7053),
    [aux_sym_infix_op_token1] = ACTIONS(7055),
    [aux_sym_infix_op_token2] = ACTIONS(7053),
    [aux_sym_infix_op_token3] = ACTIONS(7053),
    [anon_sym_BANG_EQ] = ACTIONS(7053),
    [anon_sym_COLON_EQ] = ACTIONS(7055),
    [anon_sym_DOLLAR] = ACTIONS(7053),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7053),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7053),
    [sym_int] = ACTIONS(7053),
    [sym_xint] = ACTIONS(7055),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7055),
    [sym__newline] = ACTIONS(7055),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7055),
    [sym__tyapp_open] = ACTIONS(7055),
  },
  [STATE(5659)] = {
    [sym_block_comment] = STATE(5659),
    [sym_compiler_directive_decl] = STATE(5659),
    [sym_fsi_directive_decl] = STATE(5659),
    [sym_preproc_line] = STATE(5659),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__dedent] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
    [sym__in] = ACTIONS(7033),
  },
  [STATE(5660)] = {
    [sym_block_comment] = STATE(5660),
    [sym_compiler_directive_decl] = STATE(5660),
    [sym_fsi_directive_decl] = STATE(5660),
    [sym_preproc_line] = STATE(5660),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_as] = ACTIONS(6807),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_with] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5661)] = {
    [sym_block_comment] = STATE(5661),
    [sym_compiler_directive_decl] = STATE(5661),
    [sym_fsi_directive_decl] = STATE(5661),
    [sym_preproc_line] = STATE(5661),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__quoted_close] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5662)] = {
    [sym_block_comment] = STATE(5662),
    [sym_compiler_directive_decl] = STATE(5662),
    [sym_fsi_directive_decl] = STATE(5662),
    [sym_preproc_line] = STATE(5662),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__dedent] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
    [sym__in] = ACTIONS(7037),
  },
  [STATE(5663)] = {
    [sym_block_comment] = STATE(5663),
    [sym_compiler_directive_decl] = STATE(5663),
    [sym_fsi_directive_decl] = STATE(5663),
    [sym_preproc_line] = STATE(5663),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__dedent] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
    [sym__in] = ACTIONS(7041),
  },
  [STATE(5664)] = {
    [sym_block_comment] = STATE(5664),
    [sym_compiler_directive_decl] = STATE(5664),
    [sym_fsi_directive_decl] = STATE(5664),
    [sym_preproc_line] = STATE(5664),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_GT] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5665)] = {
    [sym_block_comment] = STATE(5665),
    [sym_compiler_directive_decl] = STATE(5665),
    [sym_fsi_directive_decl] = STATE(5665),
    [sym_preproc_line] = STATE(5665),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_DOT_DOT] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5666)] = {
    [sym_block_comment] = STATE(5666),
    [sym_compiler_directive_decl] = STATE(5666),
    [sym_fsi_directive_decl] = STATE(5666),
    [sym_preproc_line] = STATE(5666),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__multi_dollar_interp_end] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5667)] = {
    [sym_block_comment] = STATE(5667),
    [sym_compiler_directive_decl] = STATE(5667),
    [sym_fsi_directive_decl] = STATE(5667),
    [sym_preproc_line] = STATE(5667),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [anon_sym_POUNDendif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5668)] = {
    [sym_type_arguments] = STATE(6585),
    [sym_long_identifier] = STATE(6586),
    [sym_block_comment] = STATE(5668),
    [sym_compiler_directive_decl] = STATE(5668),
    [sym_fsi_directive_decl] = STATE(5668),
    [sym_preproc_line] = STATE(5668),
    [aux_sym_compound_type_repeat1] = STATE(6462),
    [sym_identifier] = ACTIONS(9798),
    [anon_sym_module] = ACTIONS(9798),
    [anon_sym_null] = ACTIONS(9798),
    [anon_sym_begin] = ACTIONS(9798),
    [anon_sym_open] = ACTIONS(9798),
    [anon_sym_type] = ACTIONS(9798),
    [anon_sym_LBRACK_LT] = ACTIONS(9800),
    [anon_sym_do] = ACTIONS(9798),
    [anon_sym_and] = ACTIONS(9798),
    [anon_sym_let] = ACTIONS(9798),
    [anon_sym_let_BANG] = ACTIONS(9800),
    [aux_sym_access_modifier_token1] = ACTIONS(9800),
    [anon_sym_LPAREN] = ACTIONS(9798),
    [anon_sym_AMP] = ACTIONS(9798),
    [anon_sym_LBRACK] = ACTIONS(9798),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9800),
    [anon_sym_LBRACE] = ACTIONS(9798),
    [anon_sym_LT_AT] = ACTIONS(9798),
    [anon_sym_LT_AT_AT] = ACTIONS(9800),
    [anon_sym_CARET] = ACTIONS(9800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9800),
    [anon_sym_new] = ACTIONS(9798),
    [anon_sym_return] = ACTIONS(9798),
    [anon_sym_return_BANG] = ACTIONS(9800),
    [anon_sym_yield] = ACTIONS(9798),
    [anon_sym_yield_BANG] = ACTIONS(9800),
    [anon_sym_lazy] = ACTIONS(9798),
    [anon_sym_assert] = ACTIONS(9798),
    [anon_sym_upcast] = ACTIONS(9798),
    [anon_sym_downcast] = ACTIONS(9798),
    [anon_sym_fixed] = ACTIONS(9798),
    [anon_sym_for] = ACTIONS(9798),
    [anon_sym_while] = ACTIONS(9798),
    [anon_sym_while_BANG] = ACTIONS(9800),
    [anon_sym_if] = ACTIONS(9798),
    [anon_sym_fun] = ACTIONS(9798),
    [anon_sym_DASH_GT] = ACTIONS(9802),
    [anon_sym_try] = ACTIONS(9798),
    [anon_sym_match] = ACTIONS(9798),
    [anon_sym_match_BANG] = ACTIONS(9800),
    [anon_sym_function] = ACTIONS(9798),
    [anon_sym_use] = ACTIONS(9798),
    [anon_sym_use_BANG] = ACTIONS(9800),
    [anon_sym_do_BANG] = ACTIONS(9800),
    [anon_sym_AMP_AMP] = ACTIONS(9800),
    [anon_sym_STAR] = ACTIONS(9804),
    [anon_sym_LT] = ACTIONS(9806),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9808),
    [anon_sym_default] = ACTIONS(9798),
    [anon_sym_static] = ACTIONS(9798),
    [anon_sym_member] = ACTIONS(9798),
    [anon_sym_exception] = ACTIONS(9798),
    [anon_sym_interface] = ACTIONS(9798),
    [anon_sym_abstract] = ACTIONS(9798),
    [anon_sym_val] = ACTIONS(9798),
    [anon_sym_override] = ACTIONS(9798),
    [anon_sym_extern] = ACTIONS(9798),
    [aux_sym_char_token1] = ACTIONS(9800),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9798),
    [anon_sym_DQUOTE] = ACTIONS(9798),
    [anon_sym_AT_DQUOTE] = ACTIONS(9800),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9800),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9800),
    [sym_bool] = ACTIONS(9798),
    [sym_unit] = ACTIONS(9800),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9798),
    [sym_op_identifier] = ACTIONS(9800),
    [anon_sym_PLUS] = ACTIONS(9798),
    [anon_sym_DASH] = ACTIONS(9798),
    [anon_sym_PLUS_DOT] = ACTIONS(9800),
    [anon_sym_DASH_DOT] = ACTIONS(9800),
    [anon_sym_PERCENT] = ACTIONS(9798),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9800),
    [anon_sym_TILDE] = ACTIONS(9800),
    [aux_sym_prefix_op_token1] = ACTIONS(9800),
    [sym_int] = ACTIONS(9798),
    [sym_xint] = ACTIONS(9800),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9800),
    [anon_sym_POUNDload] = ACTIONS(9800),
    [anon_sym_POUNDtime] = ACTIONS(9800),
    [anon_sym_POUNDI] = ACTIONS(9800),
    [anon_sym_POUNDhelp] = ACTIONS(9800),
    [anon_sym_POUNDquit] = ACTIONS(9800),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9800),
    [anon_sym_POUNDendif] = ACTIONS(9800),
    [anon_sym_POUNDelse] = ACTIONS(9800),
    [sym__newline] = ACTIONS(9800),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9800),
  },
  [STATE(5669)] = {
    [sym_type_arguments] = STATE(6585),
    [sym_long_identifier] = STATE(6586),
    [sym_block_comment] = STATE(5669),
    [sym_compiler_directive_decl] = STATE(5669),
    [sym_fsi_directive_decl] = STATE(5669),
    [sym_preproc_line] = STATE(5669),
    [aux_sym_compound_type_repeat1] = STATE(6462),
    [sym_identifier] = ACTIONS(9810),
    [anon_sym_module] = ACTIONS(9810),
    [anon_sym_null] = ACTIONS(9810),
    [anon_sym_begin] = ACTIONS(9810),
    [anon_sym_open] = ACTIONS(9810),
    [anon_sym_type] = ACTIONS(9810),
    [anon_sym_LBRACK_LT] = ACTIONS(9812),
    [anon_sym_do] = ACTIONS(9810),
    [anon_sym_and] = ACTIONS(9810),
    [anon_sym_let] = ACTIONS(9810),
    [anon_sym_let_BANG] = ACTIONS(9812),
    [aux_sym_access_modifier_token1] = ACTIONS(9812),
    [anon_sym_LPAREN] = ACTIONS(9810),
    [anon_sym_AMP] = ACTIONS(9810),
    [anon_sym_LBRACK] = ACTIONS(9810),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9812),
    [anon_sym_LBRACE] = ACTIONS(9810),
    [anon_sym_LT_AT] = ACTIONS(9810),
    [anon_sym_LT_AT_AT] = ACTIONS(9812),
    [anon_sym_CARET] = ACTIONS(9812),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9812),
    [anon_sym_new] = ACTIONS(9810),
    [anon_sym_return] = ACTIONS(9810),
    [anon_sym_return_BANG] = ACTIONS(9812),
    [anon_sym_yield] = ACTIONS(9810),
    [anon_sym_yield_BANG] = ACTIONS(9812),
    [anon_sym_lazy] = ACTIONS(9810),
    [anon_sym_assert] = ACTIONS(9810),
    [anon_sym_upcast] = ACTIONS(9810),
    [anon_sym_downcast] = ACTIONS(9810),
    [anon_sym_fixed] = ACTIONS(9810),
    [anon_sym_for] = ACTIONS(9810),
    [anon_sym_while] = ACTIONS(9810),
    [anon_sym_while_BANG] = ACTIONS(9812),
    [anon_sym_if] = ACTIONS(9810),
    [anon_sym_fun] = ACTIONS(9810),
    [anon_sym_DASH_GT] = ACTIONS(9802),
    [anon_sym_try] = ACTIONS(9810),
    [anon_sym_match] = ACTIONS(9810),
    [anon_sym_match_BANG] = ACTIONS(9812),
    [anon_sym_function] = ACTIONS(9810),
    [anon_sym_use] = ACTIONS(9810),
    [anon_sym_use_BANG] = ACTIONS(9812),
    [anon_sym_do_BANG] = ACTIONS(9812),
    [anon_sym_AMP_AMP] = ACTIONS(9812),
    [anon_sym_STAR] = ACTIONS(9804),
    [anon_sym_LT] = ACTIONS(9806),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9808),
    [anon_sym_default] = ACTIONS(9810),
    [anon_sym_static] = ACTIONS(9810),
    [anon_sym_member] = ACTIONS(9810),
    [anon_sym_exception] = ACTIONS(9810),
    [anon_sym_interface] = ACTIONS(9810),
    [anon_sym_abstract] = ACTIONS(9810),
    [anon_sym_val] = ACTIONS(9810),
    [anon_sym_override] = ACTIONS(9810),
    [anon_sym_extern] = ACTIONS(9810),
    [aux_sym_char_token1] = ACTIONS(9812),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9810),
    [anon_sym_DQUOTE] = ACTIONS(9810),
    [anon_sym_AT_DQUOTE] = ACTIONS(9812),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9812),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9812),
    [sym_bool] = ACTIONS(9810),
    [sym_unit] = ACTIONS(9812),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9810),
    [sym_op_identifier] = ACTIONS(9812),
    [anon_sym_PLUS] = ACTIONS(9810),
    [anon_sym_DASH] = ACTIONS(9810),
    [anon_sym_PLUS_DOT] = ACTIONS(9812),
    [anon_sym_DASH_DOT] = ACTIONS(9812),
    [anon_sym_PERCENT] = ACTIONS(9810),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9812),
    [anon_sym_TILDE] = ACTIONS(9812),
    [aux_sym_prefix_op_token1] = ACTIONS(9812),
    [sym_int] = ACTIONS(9810),
    [sym_xint] = ACTIONS(9812),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9812),
    [anon_sym_POUNDload] = ACTIONS(9812),
    [anon_sym_POUNDtime] = ACTIONS(9812),
    [anon_sym_POUNDI] = ACTIONS(9812),
    [anon_sym_POUNDhelp] = ACTIONS(9812),
    [anon_sym_POUNDquit] = ACTIONS(9812),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9812),
    [anon_sym_POUNDendif] = ACTIONS(9812),
    [anon_sym_POUNDelse] = ACTIONS(9812),
    [sym__newline] = ACTIONS(9812),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9812),
  },
  [STATE(5670)] = {
    [sym_block_comment] = STATE(5670),
    [sym_compiler_directive_decl] = STATE(5670),
    [sym_fsi_directive_decl] = STATE(5670),
    [sym_preproc_line] = STATE(5670),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__dedent] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
    [sym__in] = ACTIONS(7057),
  },
  [STATE(5671)] = {
    [sym_type_arguments] = STATE(6585),
    [sym_long_identifier] = STATE(6586),
    [sym_block_comment] = STATE(5671),
    [sym_compiler_directive_decl] = STATE(5671),
    [sym_fsi_directive_decl] = STATE(5671),
    [sym_preproc_line] = STATE(5671),
    [aux_sym_compound_type_repeat1] = STATE(6462),
    [sym_identifier] = ACTIONS(9814),
    [anon_sym_module] = ACTIONS(9814),
    [anon_sym_null] = ACTIONS(9814),
    [anon_sym_begin] = ACTIONS(9814),
    [anon_sym_open] = ACTIONS(9814),
    [anon_sym_type] = ACTIONS(9814),
    [anon_sym_LBRACK_LT] = ACTIONS(9816),
    [anon_sym_do] = ACTIONS(9814),
    [anon_sym_and] = ACTIONS(9814),
    [anon_sym_let] = ACTIONS(9814),
    [anon_sym_let_BANG] = ACTIONS(9816),
    [aux_sym_access_modifier_token1] = ACTIONS(9816),
    [anon_sym_LPAREN] = ACTIONS(9814),
    [anon_sym_AMP] = ACTIONS(9814),
    [anon_sym_LBRACK] = ACTIONS(9814),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9816),
    [anon_sym_LBRACE] = ACTIONS(9814),
    [anon_sym_LT_AT] = ACTIONS(9814),
    [anon_sym_LT_AT_AT] = ACTIONS(9816),
    [anon_sym_CARET] = ACTIONS(9816),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9816),
    [anon_sym_new] = ACTIONS(9814),
    [anon_sym_return] = ACTIONS(9814),
    [anon_sym_return_BANG] = ACTIONS(9816),
    [anon_sym_yield] = ACTIONS(9814),
    [anon_sym_yield_BANG] = ACTIONS(9816),
    [anon_sym_lazy] = ACTIONS(9814),
    [anon_sym_assert] = ACTIONS(9814),
    [anon_sym_upcast] = ACTIONS(9814),
    [anon_sym_downcast] = ACTIONS(9814),
    [anon_sym_fixed] = ACTIONS(9814),
    [anon_sym_for] = ACTIONS(9814),
    [anon_sym_while] = ACTIONS(9814),
    [anon_sym_while_BANG] = ACTIONS(9816),
    [anon_sym_if] = ACTIONS(9814),
    [anon_sym_fun] = ACTIONS(9814),
    [anon_sym_DASH_GT] = ACTIONS(9802),
    [anon_sym_try] = ACTIONS(9814),
    [anon_sym_match] = ACTIONS(9814),
    [anon_sym_match_BANG] = ACTIONS(9816),
    [anon_sym_function] = ACTIONS(9814),
    [anon_sym_use] = ACTIONS(9814),
    [anon_sym_use_BANG] = ACTIONS(9816),
    [anon_sym_do_BANG] = ACTIONS(9816),
    [anon_sym_AMP_AMP] = ACTIONS(9816),
    [anon_sym_STAR] = ACTIONS(9804),
    [anon_sym_LT] = ACTIONS(9806),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9808),
    [anon_sym_default] = ACTIONS(9814),
    [anon_sym_static] = ACTIONS(9814),
    [anon_sym_member] = ACTIONS(9814),
    [anon_sym_exception] = ACTIONS(9814),
    [anon_sym_interface] = ACTIONS(9814),
    [anon_sym_abstract] = ACTIONS(9814),
    [anon_sym_val] = ACTIONS(9814),
    [anon_sym_override] = ACTIONS(9814),
    [anon_sym_extern] = ACTIONS(9814),
    [aux_sym_char_token1] = ACTIONS(9816),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9814),
    [anon_sym_DQUOTE] = ACTIONS(9814),
    [anon_sym_AT_DQUOTE] = ACTIONS(9816),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9816),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9816),
    [sym_bool] = ACTIONS(9814),
    [sym_unit] = ACTIONS(9816),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9814),
    [sym_op_identifier] = ACTIONS(9816),
    [anon_sym_PLUS] = ACTIONS(9814),
    [anon_sym_DASH] = ACTIONS(9814),
    [anon_sym_PLUS_DOT] = ACTIONS(9816),
    [anon_sym_DASH_DOT] = ACTIONS(9816),
    [anon_sym_PERCENT] = ACTIONS(9814),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9816),
    [anon_sym_TILDE] = ACTIONS(9816),
    [aux_sym_prefix_op_token1] = ACTIONS(9816),
    [sym_int] = ACTIONS(9814),
    [sym_xint] = ACTIONS(9816),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9816),
    [anon_sym_POUNDload] = ACTIONS(9816),
    [anon_sym_POUNDtime] = ACTIONS(9816),
    [anon_sym_POUNDI] = ACTIONS(9816),
    [anon_sym_POUNDhelp] = ACTIONS(9816),
    [anon_sym_POUNDquit] = ACTIONS(9816),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9816),
    [anon_sym_POUNDendif] = ACTIONS(9816),
    [anon_sym_POUNDelse] = ACTIONS(9816),
    [sym__newline] = ACTIONS(9816),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9816),
  },
  [STATE(5672)] = {
    [sym_type_arguments] = STATE(6585),
    [sym_long_identifier] = STATE(6586),
    [sym_block_comment] = STATE(5672),
    [sym_compiler_directive_decl] = STATE(5672),
    [sym_fsi_directive_decl] = STATE(5672),
    [sym_preproc_line] = STATE(5672),
    [aux_sym_compound_type_repeat1] = STATE(6462),
    [sym_identifier] = ACTIONS(9818),
    [anon_sym_module] = ACTIONS(9818),
    [anon_sym_null] = ACTIONS(9818),
    [anon_sym_begin] = ACTIONS(9818),
    [anon_sym_open] = ACTIONS(9818),
    [anon_sym_type] = ACTIONS(9818),
    [anon_sym_LBRACK_LT] = ACTIONS(9820),
    [anon_sym_do] = ACTIONS(9818),
    [anon_sym_and] = ACTIONS(9818),
    [anon_sym_let] = ACTIONS(9818),
    [anon_sym_let_BANG] = ACTIONS(9820),
    [aux_sym_access_modifier_token1] = ACTIONS(9820),
    [anon_sym_LPAREN] = ACTIONS(9818),
    [anon_sym_AMP] = ACTIONS(9818),
    [anon_sym_LBRACK] = ACTIONS(9818),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9820),
    [anon_sym_LBRACE] = ACTIONS(9818),
    [anon_sym_LT_AT] = ACTIONS(9818),
    [anon_sym_LT_AT_AT] = ACTIONS(9820),
    [anon_sym_CARET] = ACTIONS(9820),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9820),
    [anon_sym_new] = ACTIONS(9818),
    [anon_sym_return] = ACTIONS(9818),
    [anon_sym_return_BANG] = ACTIONS(9820),
    [anon_sym_yield] = ACTIONS(9818),
    [anon_sym_yield_BANG] = ACTIONS(9820),
    [anon_sym_lazy] = ACTIONS(9818),
    [anon_sym_assert] = ACTIONS(9818),
    [anon_sym_upcast] = ACTIONS(9818),
    [anon_sym_downcast] = ACTIONS(9818),
    [anon_sym_fixed] = ACTIONS(9818),
    [anon_sym_for] = ACTIONS(9818),
    [anon_sym_while] = ACTIONS(9818),
    [anon_sym_while_BANG] = ACTIONS(9820),
    [anon_sym_if] = ACTIONS(9818),
    [anon_sym_fun] = ACTIONS(9818),
    [anon_sym_DASH_GT] = ACTIONS(9802),
    [anon_sym_try] = ACTIONS(9818),
    [anon_sym_match] = ACTIONS(9818),
    [anon_sym_match_BANG] = ACTIONS(9820),
    [anon_sym_function] = ACTIONS(9818),
    [anon_sym_use] = ACTIONS(9818),
    [anon_sym_use_BANG] = ACTIONS(9820),
    [anon_sym_do_BANG] = ACTIONS(9820),
    [anon_sym_AMP_AMP] = ACTIONS(9820),
    [anon_sym_STAR] = ACTIONS(9804),
    [anon_sym_LT] = ACTIONS(9806),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9808),
    [anon_sym_default] = ACTIONS(9818),
    [anon_sym_static] = ACTIONS(9818),
    [anon_sym_member] = ACTIONS(9818),
    [anon_sym_exception] = ACTIONS(9818),
    [anon_sym_interface] = ACTIONS(9818),
    [anon_sym_abstract] = ACTIONS(9818),
    [anon_sym_val] = ACTIONS(9818),
    [anon_sym_override] = ACTIONS(9818),
    [anon_sym_extern] = ACTIONS(9818),
    [aux_sym_char_token1] = ACTIONS(9820),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9818),
    [anon_sym_DQUOTE] = ACTIONS(9818),
    [anon_sym_AT_DQUOTE] = ACTIONS(9820),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9820),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9820),
    [sym_bool] = ACTIONS(9818),
    [sym_unit] = ACTIONS(9820),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9818),
    [sym_op_identifier] = ACTIONS(9820),
    [anon_sym_PLUS] = ACTIONS(9818),
    [anon_sym_DASH] = ACTIONS(9818),
    [anon_sym_PLUS_DOT] = ACTIONS(9820),
    [anon_sym_DASH_DOT] = ACTIONS(9820),
    [anon_sym_PERCENT] = ACTIONS(9818),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9820),
    [anon_sym_TILDE] = ACTIONS(9820),
    [aux_sym_prefix_op_token1] = ACTIONS(9820),
    [sym_int] = ACTIONS(9818),
    [sym_xint] = ACTIONS(9820),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9820),
    [anon_sym_POUNDload] = ACTIONS(9820),
    [anon_sym_POUNDtime] = ACTIONS(9820),
    [anon_sym_POUNDI] = ACTIONS(9820),
    [anon_sym_POUNDhelp] = ACTIONS(9820),
    [anon_sym_POUNDquit] = ACTIONS(9820),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9820),
    [anon_sym_POUNDendif] = ACTIONS(9820),
    [anon_sym_POUNDelse] = ACTIONS(9820),
    [sym__newline] = ACTIONS(9820),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9820),
  },
  [STATE(5673)] = {
    [sym_type_arguments] = STATE(6585),
    [sym_long_identifier] = STATE(6586),
    [sym_block_comment] = STATE(5673),
    [sym_compiler_directive_decl] = STATE(5673),
    [sym_fsi_directive_decl] = STATE(5673),
    [sym_preproc_line] = STATE(5673),
    [aux_sym_compound_type_repeat1] = STATE(6462),
    [sym_identifier] = ACTIONS(9822),
    [anon_sym_module] = ACTIONS(9822),
    [anon_sym_null] = ACTIONS(9822),
    [anon_sym_begin] = ACTIONS(9822),
    [anon_sym_open] = ACTIONS(9822),
    [anon_sym_type] = ACTIONS(9822),
    [anon_sym_LBRACK_LT] = ACTIONS(9824),
    [anon_sym_do] = ACTIONS(9822),
    [anon_sym_and] = ACTIONS(9822),
    [anon_sym_let] = ACTIONS(9822),
    [anon_sym_let_BANG] = ACTIONS(9824),
    [aux_sym_access_modifier_token1] = ACTIONS(9824),
    [anon_sym_LPAREN] = ACTIONS(9822),
    [anon_sym_AMP] = ACTIONS(9822),
    [anon_sym_LBRACK] = ACTIONS(9822),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9824),
    [anon_sym_LBRACE] = ACTIONS(9822),
    [anon_sym_LT_AT] = ACTIONS(9822),
    [anon_sym_LT_AT_AT] = ACTIONS(9824),
    [anon_sym_CARET] = ACTIONS(9824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9824),
    [anon_sym_new] = ACTIONS(9822),
    [anon_sym_return] = ACTIONS(9822),
    [anon_sym_return_BANG] = ACTIONS(9824),
    [anon_sym_yield] = ACTIONS(9822),
    [anon_sym_yield_BANG] = ACTIONS(9824),
    [anon_sym_lazy] = ACTIONS(9822),
    [anon_sym_assert] = ACTIONS(9822),
    [anon_sym_upcast] = ACTIONS(9822),
    [anon_sym_downcast] = ACTIONS(9822),
    [anon_sym_fixed] = ACTIONS(9822),
    [anon_sym_for] = ACTIONS(9822),
    [anon_sym_while] = ACTIONS(9822),
    [anon_sym_while_BANG] = ACTIONS(9824),
    [anon_sym_if] = ACTIONS(9822),
    [anon_sym_fun] = ACTIONS(9822),
    [anon_sym_DASH_GT] = ACTIONS(9802),
    [anon_sym_try] = ACTIONS(9822),
    [anon_sym_match] = ACTIONS(9822),
    [anon_sym_match_BANG] = ACTIONS(9824),
    [anon_sym_function] = ACTIONS(9822),
    [anon_sym_use] = ACTIONS(9822),
    [anon_sym_use_BANG] = ACTIONS(9824),
    [anon_sym_do_BANG] = ACTIONS(9824),
    [anon_sym_AMP_AMP] = ACTIONS(9824),
    [anon_sym_STAR] = ACTIONS(9804),
    [anon_sym_LT] = ACTIONS(9806),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9808),
    [anon_sym_default] = ACTIONS(9822),
    [anon_sym_static] = ACTIONS(9822),
    [anon_sym_member] = ACTIONS(9822),
    [anon_sym_exception] = ACTIONS(9822),
    [anon_sym_interface] = ACTIONS(9822),
    [anon_sym_abstract] = ACTIONS(9822),
    [anon_sym_val] = ACTIONS(9822),
    [anon_sym_override] = ACTIONS(9822),
    [anon_sym_extern] = ACTIONS(9822),
    [aux_sym_char_token1] = ACTIONS(9824),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9822),
    [anon_sym_DQUOTE] = ACTIONS(9822),
    [anon_sym_AT_DQUOTE] = ACTIONS(9824),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9824),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9824),
    [sym_bool] = ACTIONS(9822),
    [sym_unit] = ACTIONS(9824),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9822),
    [sym_op_identifier] = ACTIONS(9824),
    [anon_sym_PLUS] = ACTIONS(9822),
    [anon_sym_DASH] = ACTIONS(9822),
    [anon_sym_PLUS_DOT] = ACTIONS(9824),
    [anon_sym_DASH_DOT] = ACTIONS(9824),
    [anon_sym_PERCENT] = ACTIONS(9822),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9824),
    [anon_sym_TILDE] = ACTIONS(9824),
    [aux_sym_prefix_op_token1] = ACTIONS(9824),
    [sym_int] = ACTIONS(9822),
    [sym_xint] = ACTIONS(9824),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9824),
    [anon_sym_POUNDload] = ACTIONS(9824),
    [anon_sym_POUNDtime] = ACTIONS(9824),
    [anon_sym_POUNDI] = ACTIONS(9824),
    [anon_sym_POUNDhelp] = ACTIONS(9824),
    [anon_sym_POUNDquit] = ACTIONS(9824),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9824),
    [anon_sym_POUNDendif] = ACTIONS(9824),
    [anon_sym_POUNDelse] = ACTIONS(9824),
    [sym__newline] = ACTIONS(9824),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9824),
  },
  [STATE(5674)] = {
    [sym_block_comment] = STATE(5674),
    [sym_compiler_directive_decl] = STATE(5674),
    [sym_fsi_directive_decl] = STATE(5674),
    [sym_preproc_line] = STATE(5674),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_DASH_GT] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_DOT_DOT] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5675)] = {
    [sym_block_comment] = STATE(5675),
    [sym_compiler_directive_decl] = STATE(5675),
    [sym_fsi_directive_decl] = STATE(5675),
    [sym_preproc_line] = STATE(5675),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_as] = ACTIONS(6892),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_with] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(5676)] = {
    [sym_block_comment] = STATE(5676),
    [sym_compiler_directive_decl] = STATE(5676),
    [sym_fsi_directive_decl] = STATE(5676),
    [sym_preproc_line] = STATE(5676),
    [sym_identifier] = ACTIONS(6892),
    [anon_sym_null] = ACTIONS(6892),
    [anon_sym_EQ] = ACTIONS(6892),
    [anon_sym_begin] = ACTIONS(6892),
    [anon_sym_COLON] = ACTIONS(6892),
    [anon_sym_do] = ACTIONS(6892),
    [anon_sym_let] = ACTIONS(6892),
    [anon_sym_let_BANG] = ACTIONS(6890),
    [anon_sym_LPAREN] = ACTIONS(6892),
    [anon_sym_COMMA] = ACTIONS(6890),
    [anon_sym_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_QMARK] = ACTIONS(6892),
    [anon_sym_COLON_COLON] = ACTIONS(6890),
    [anon_sym_AMP] = ACTIONS(6892),
    [anon_sym_LBRACK] = ACTIONS(6892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6890),
    [anon_sym_LBRACE] = ACTIONS(6892),
    [anon_sym_LT_AT] = ACTIONS(6892),
    [anon_sym_LT_AT_AT] = ACTIONS(6892),
    [anon_sym_CARET] = ACTIONS(6892),
    [anon_sym_or] = ACTIONS(6892),
    [anon_sym_DOT] = ACTIONS(6892),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6890),
    [anon_sym_new] = ACTIONS(6892),
    [anon_sym_return] = ACTIONS(6892),
    [anon_sym_return_BANG] = ACTIONS(6890),
    [anon_sym_yield] = ACTIONS(6892),
    [anon_sym_yield_BANG] = ACTIONS(6890),
    [anon_sym_lazy] = ACTIONS(6892),
    [anon_sym_assert] = ACTIONS(6892),
    [anon_sym_upcast] = ACTIONS(6892),
    [anon_sym_downcast] = ACTIONS(6892),
    [anon_sym_fixed] = ACTIONS(6892),
    [anon_sym_COLON_GT] = ACTIONS(6890),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6890),
    [anon_sym_for] = ACTIONS(6892),
    [anon_sym_while] = ACTIONS(6892),
    [anon_sym_while_BANG] = ACTIONS(6890),
    [anon_sym_if] = ACTIONS(6892),
    [anon_sym_fun] = ACTIONS(6892),
    [anon_sym_DASH_GT] = ACTIONS(6892),
    [anon_sym_try] = ACTIONS(6892),
    [anon_sym_match] = ACTIONS(6892),
    [anon_sym_match_BANG] = ACTIONS(6890),
    [anon_sym_function] = ACTIONS(6892),
    [anon_sym_DOT_LBRACK] = ACTIONS(6890),
    [anon_sym_use] = ACTIONS(6892),
    [anon_sym_use_BANG] = ACTIONS(6890),
    [anon_sym_do_BANG] = ACTIONS(6890),
    [anon_sym_DOT_DOT] = ACTIONS(6890),
    [anon_sym_LPAREN2] = ACTIONS(6892),
    [anon_sym_PIPE_PIPE] = ACTIONS(6892),
    [anon_sym_AMP_AMP] = ACTIONS(6892),
    [aux_sym_char_token1] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6892),
    [anon_sym_DQUOTE] = ACTIONS(6892),
    [anon_sym_AT_DQUOTE] = ACTIONS(6890),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6890),
    [sym_bool] = ACTIONS(6892),
    [sym_unit] = ACTIONS(6890),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6892),
    [sym_op_identifier] = ACTIONS(6892),
    [anon_sym_PLUS] = ACTIONS(6892),
    [anon_sym_DASH] = ACTIONS(6892),
    [anon_sym_PLUS_DOT] = ACTIONS(6892),
    [anon_sym_DASH_DOT] = ACTIONS(6892),
    [anon_sym_PERCENT] = ACTIONS(6892),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6892),
    [anon_sym_TILDE] = ACTIONS(6890),
    [aux_sym_prefix_op_token1] = ACTIONS(6892),
    [aux_sym_infix_op_token1] = ACTIONS(6890),
    [aux_sym_infix_op_token2] = ACTIONS(6892),
    [aux_sym_infix_op_token3] = ACTIONS(6892),
    [anon_sym_BANG_EQ] = ACTIONS(6892),
    [anon_sym_COLON_EQ] = ACTIONS(6890),
    [anon_sym_DOLLAR] = ACTIONS(6892),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6892),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6892),
    [sym_int] = ACTIONS(6892),
    [sym_xint] = ACTIONS(6890),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6890),
    [sym__newline] = ACTIONS(6890),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6890),
    [sym__tyapp_open] = ACTIONS(6890),
  },
  [STATE(5677)] = {
    [sym_block_comment] = STATE(5677),
    [sym_compiler_directive_decl] = STATE(5677),
    [sym_fsi_directive_decl] = STATE(5677),
    [sym_preproc_line] = STATE(5677),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(9826),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__multi_dollar_interp_end] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5678)] = {
    [sym_block_comment] = STATE(5678),
    [sym_compiler_directive_decl] = STATE(5678),
    [sym_fsi_directive_decl] = STATE(5678),
    [sym_preproc_line] = STATE(5678),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_DASH_GT] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_DOT_DOT] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(5679)] = {
    [sym_block_comment] = STATE(5679),
    [sym_compiler_directive_decl] = STATE(5679),
    [sym_fsi_directive_decl] = STATE(5679),
    [sym_preproc_line] = STATE(5679),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_DASH_GT] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_DOT_DOT] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(5680)] = {
    [sym_block_comment] = STATE(5680),
    [sym_compiler_directive_decl] = STATE(5680),
    [sym_fsi_directive_decl] = STATE(5680),
    [sym_preproc_line] = STATE(5680),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_DASH_GT] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_DOT_DOT] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(5681)] = {
    [sym_block_comment] = STATE(5681),
    [sym_compiler_directive_decl] = STATE(5681),
    [sym_fsi_directive_decl] = STATE(5681),
    [sym_preproc_line] = STATE(5681),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(9828),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5682)] = {
    [sym_block_comment] = STATE(5682),
    [sym_compiler_directive_decl] = STATE(5682),
    [sym_fsi_directive_decl] = STATE(5682),
    [sym_preproc_line] = STATE(5682),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_DASH_GT] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_DOT_DOT] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(5683)] = {
    [sym_block_comment] = STATE(5683),
    [sym_compiler_directive_decl] = STATE(5683),
    [sym_fsi_directive_decl] = STATE(5683),
    [sym_preproc_line] = STATE(5683),
    [aux_sym_sequential_expression_repeat1] = STATE(5840),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_DASH_GT] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5684)] = {
    [sym_block_comment] = STATE(5684),
    [sym_compiler_directive_decl] = STATE(5684),
    [sym_fsi_directive_decl] = STATE(5684),
    [sym_preproc_line] = STATE(5684),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_DASH_GT] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_DOT_DOT] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(5685)] = {
    [sym_block_comment] = STATE(5685),
    [sym_compiler_directive_decl] = STATE(5685),
    [sym_fsi_directive_decl] = STATE(5685),
    [sym_preproc_line] = STATE(5685),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_DASH_GT] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_DOT_DOT] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(5686)] = {
    [sym_block_comment] = STATE(5686),
    [sym_compiler_directive_decl] = STATE(5686),
    [sym_fsi_directive_decl] = STATE(5686),
    [sym_preproc_line] = STATE(5686),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_DASH_GT] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_DOT_DOT] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(5687)] = {
    [sym_block_comment] = STATE(5687),
    [sym_compiler_directive_decl] = STATE(5687),
    [sym_fsi_directive_decl] = STATE(5687),
    [sym_preproc_line] = STATE(5687),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_DASH_GT] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_DOT_DOT] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(5688)] = {
    [sym_type_arguments] = STATE(6569),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5688),
    [sym_compiler_directive_decl] = STATE(5688),
    [sym_fsi_directive_decl] = STATE(5688),
    [sym_preproc_line] = STATE(5688),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(6213),
    [anon_sym_null] = ACTIONS(6213),
    [anon_sym_begin] = ACTIONS(6213),
    [anon_sym_open] = ACTIONS(6213),
    [anon_sym_type] = ACTIONS(6213),
    [anon_sym_LBRACK_LT] = ACTIONS(6209),
    [anon_sym_do] = ACTIONS(6213),
    [anon_sym_and] = ACTIONS(6213),
    [anon_sym_let] = ACTIONS(6213),
    [anon_sym_let_BANG] = ACTIONS(6209),
    [aux_sym_access_modifier_token1] = ACTIONS(6209),
    [anon_sym_LPAREN] = ACTIONS(6213),
    [anon_sym_PIPE] = ACTIONS(6209),
    [anon_sym_AMP] = ACTIONS(6213),
    [anon_sym_LBRACK] = ACTIONS(6213),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6209),
    [anon_sym_LBRACE] = ACTIONS(6213),
    [anon_sym_LT_AT] = ACTIONS(6213),
    [anon_sym_LT_AT_AT] = ACTIONS(6209),
    [anon_sym_CARET] = ACTIONS(6209),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6209),
    [anon_sym_new] = ACTIONS(6213),
    [anon_sym_with] = ACTIONS(6213),
    [anon_sym_return] = ACTIONS(6213),
    [anon_sym_return_BANG] = ACTIONS(6209),
    [anon_sym_yield] = ACTIONS(6213),
    [anon_sym_yield_BANG] = ACTIONS(6209),
    [anon_sym_lazy] = ACTIONS(6213),
    [anon_sym_assert] = ACTIONS(6213),
    [anon_sym_upcast] = ACTIONS(6213),
    [anon_sym_downcast] = ACTIONS(6213),
    [anon_sym_fixed] = ACTIONS(6213),
    [anon_sym_for] = ACTIONS(6213),
    [anon_sym_while] = ACTIONS(6213),
    [anon_sym_while_BANG] = ACTIONS(6209),
    [anon_sym_if] = ACTIONS(6213),
    [anon_sym_fun] = ACTIONS(6213),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(6213),
    [anon_sym_match] = ACTIONS(6213),
    [anon_sym_match_BANG] = ACTIONS(6209),
    [anon_sym_function] = ACTIONS(6213),
    [anon_sym_use] = ACTIONS(6213),
    [anon_sym_use_BANG] = ACTIONS(6209),
    [anon_sym_do_BANG] = ACTIONS(6209),
    [anon_sym_AMP_AMP] = ACTIONS(6209),
    [anon_sym_STAR] = ACTIONS(9794),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(6213),
    [anon_sym_static] = ACTIONS(6213),
    [anon_sym_member] = ACTIONS(6213),
    [anon_sym_exception] = ACTIONS(6213),
    [anon_sym_interface] = ACTIONS(6213),
    [anon_sym_abstract] = ACTIONS(6213),
    [anon_sym_val] = ACTIONS(6213),
    [anon_sym_override] = ACTIONS(6213),
    [anon_sym_extern] = ACTIONS(6213),
    [aux_sym_char_token1] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6213),
    [anon_sym_DQUOTE] = ACTIONS(6213),
    [anon_sym_AT_DQUOTE] = ACTIONS(6209),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6209),
    [sym_bool] = ACTIONS(6213),
    [sym_unit] = ACTIONS(6209),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6213),
    [sym_op_identifier] = ACTIONS(6209),
    [anon_sym_PLUS] = ACTIONS(6213),
    [anon_sym_DASH] = ACTIONS(6213),
    [anon_sym_PLUS_DOT] = ACTIONS(6209),
    [anon_sym_DASH_DOT] = ACTIONS(6209),
    [anon_sym_PERCENT] = ACTIONS(6213),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6209),
    [anon_sym_TILDE] = ACTIONS(6209),
    [aux_sym_prefix_op_token1] = ACTIONS(6209),
    [sym_int] = ACTIONS(6213),
    [sym_xint] = ACTIONS(6209),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6209),
    [anon_sym_POUNDload] = ACTIONS(6209),
    [anon_sym_POUNDtime] = ACTIONS(6209),
    [anon_sym_POUNDI] = ACTIONS(6209),
    [anon_sym_POUNDhelp] = ACTIONS(6209),
    [anon_sym_POUNDquit] = ACTIONS(6209),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6209),
    [anon_sym_POUNDendif] = ACTIONS(6209),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6209),
  },
  [STATE(5689)] = {
    [sym_block_comment] = STATE(5689),
    [sym_compiler_directive_decl] = STATE(5689),
    [sym_fsi_directive_decl] = STATE(5689),
    [sym_preproc_line] = STATE(5689),
    [sym_identifier] = ACTIONS(6295),
    [anon_sym_null] = ACTIONS(6295),
    [anon_sym_EQ] = ACTIONS(6295),
    [anon_sym_begin] = ACTIONS(6295),
    [anon_sym_COLON] = ACTIONS(6295),
    [anon_sym_do] = ACTIONS(6295),
    [anon_sym_let] = ACTIONS(6295),
    [anon_sym_let_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN] = ACTIONS(6295),
    [anon_sym_COMMA] = ACTIONS(6297),
    [anon_sym_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_QMARK] = ACTIONS(6295),
    [anon_sym_COLON_COLON] = ACTIONS(6297),
    [anon_sym_AMP] = ACTIONS(6295),
    [anon_sym_LBRACK] = ACTIONS(6295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6297),
    [anon_sym_LBRACE] = ACTIONS(6295),
    [anon_sym_LT_AT] = ACTIONS(6295),
    [anon_sym_LT_AT_AT] = ACTIONS(6295),
    [anon_sym_CARET] = ACTIONS(6295),
    [anon_sym_or] = ACTIONS(6295),
    [anon_sym_DOT] = ACTIONS(6295),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6297),
    [anon_sym_new] = ACTIONS(6295),
    [anon_sym_return] = ACTIONS(6295),
    [anon_sym_return_BANG] = ACTIONS(6297),
    [anon_sym_yield] = ACTIONS(6295),
    [anon_sym_yield_BANG] = ACTIONS(6297),
    [anon_sym_lazy] = ACTIONS(6295),
    [anon_sym_assert] = ACTIONS(6295),
    [anon_sym_upcast] = ACTIONS(6295),
    [anon_sym_downcast] = ACTIONS(6295),
    [anon_sym_fixed] = ACTIONS(6295),
    [anon_sym_COLON_GT] = ACTIONS(6297),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6297),
    [anon_sym_for] = ACTIONS(6295),
    [anon_sym_while] = ACTIONS(6295),
    [anon_sym_while_BANG] = ACTIONS(6297),
    [anon_sym_if] = ACTIONS(6295),
    [anon_sym_fun] = ACTIONS(6295),
    [anon_sym_try] = ACTIONS(6295),
    [anon_sym_match] = ACTIONS(6295),
    [anon_sym_match_BANG] = ACTIONS(6297),
    [anon_sym_function] = ACTIONS(6295),
    [anon_sym_LT_DASH] = ACTIONS(6295),
    [anon_sym_DOT_LBRACK] = ACTIONS(6297),
    [anon_sym_use] = ACTIONS(6295),
    [anon_sym_use_BANG] = ACTIONS(6297),
    [anon_sym_do_BANG] = ACTIONS(6297),
    [anon_sym_LPAREN2] = ACTIONS(6295),
    [anon_sym_PIPE_PIPE] = ACTIONS(6295),
    [anon_sym_AMP_AMP] = ACTIONS(6295),
    [aux_sym_char_token1] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6295),
    [anon_sym_DQUOTE] = ACTIONS(6295),
    [anon_sym_AT_DQUOTE] = ACTIONS(6297),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6297),
    [sym_bool] = ACTIONS(6295),
    [sym_unit] = ACTIONS(6297),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6295),
    [sym_op_identifier] = ACTIONS(6295),
    [anon_sym_PLUS] = ACTIONS(6295),
    [anon_sym_DASH] = ACTIONS(6295),
    [anon_sym_PLUS_DOT] = ACTIONS(6295),
    [anon_sym_DASH_DOT] = ACTIONS(6295),
    [anon_sym_PERCENT] = ACTIONS(6295),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6295),
    [anon_sym_TILDE] = ACTIONS(6297),
    [aux_sym_prefix_op_token1] = ACTIONS(6295),
    [aux_sym_infix_op_token1] = ACTIONS(6297),
    [aux_sym_infix_op_token2] = ACTIONS(6295),
    [aux_sym_infix_op_token3] = ACTIONS(6295),
    [anon_sym_BANG_EQ] = ACTIONS(6295),
    [anon_sym_COLON_EQ] = ACTIONS(6297),
    [anon_sym_DOLLAR] = ACTIONS(6295),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6295),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6295),
    [sym_int] = ACTIONS(6295),
    [sym_xint] = ACTIONS(6297),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6297),
    [sym__newline] = ACTIONS(6297),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6297),
    [sym__multi_dollar_interp_end] = ACTIONS(6297),
    [sym__tyapp_open] = ACTIONS(6297),
  },
  [STATE(5690)] = {
    [sym_type_arguments] = STATE(6569),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5690),
    [sym_compiler_directive_decl] = STATE(5690),
    [sym_fsi_directive_decl] = STATE(5690),
    [sym_preproc_line] = STATE(5690),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(6193),
    [anon_sym_null] = ACTIONS(6193),
    [anon_sym_begin] = ACTIONS(6193),
    [anon_sym_open] = ACTIONS(6193),
    [anon_sym_type] = ACTIONS(6193),
    [anon_sym_LBRACK_LT] = ACTIONS(6195),
    [anon_sym_do] = ACTIONS(6193),
    [anon_sym_and] = ACTIONS(6193),
    [anon_sym_let] = ACTIONS(6193),
    [anon_sym_let_BANG] = ACTIONS(6195),
    [aux_sym_access_modifier_token1] = ACTIONS(6195),
    [anon_sym_LPAREN] = ACTIONS(6193),
    [anon_sym_PIPE] = ACTIONS(6195),
    [anon_sym_AMP] = ACTIONS(6193),
    [anon_sym_LBRACK] = ACTIONS(6193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6195),
    [anon_sym_LBRACE] = ACTIONS(6193),
    [anon_sym_LT_AT] = ACTIONS(6193),
    [anon_sym_LT_AT_AT] = ACTIONS(6195),
    [anon_sym_CARET] = ACTIONS(6195),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6195),
    [anon_sym_new] = ACTIONS(6193),
    [anon_sym_with] = ACTIONS(6193),
    [anon_sym_return] = ACTIONS(6193),
    [anon_sym_return_BANG] = ACTIONS(6195),
    [anon_sym_yield] = ACTIONS(6193),
    [anon_sym_yield_BANG] = ACTIONS(6195),
    [anon_sym_lazy] = ACTIONS(6193),
    [anon_sym_assert] = ACTIONS(6193),
    [anon_sym_upcast] = ACTIONS(6193),
    [anon_sym_downcast] = ACTIONS(6193),
    [anon_sym_fixed] = ACTIONS(6193),
    [anon_sym_for] = ACTIONS(6193),
    [anon_sym_while] = ACTIONS(6193),
    [anon_sym_while_BANG] = ACTIONS(6195),
    [anon_sym_if] = ACTIONS(6193),
    [anon_sym_fun] = ACTIONS(6193),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(6193),
    [anon_sym_match] = ACTIONS(6193),
    [anon_sym_match_BANG] = ACTIONS(6195),
    [anon_sym_function] = ACTIONS(6193),
    [anon_sym_use] = ACTIONS(6193),
    [anon_sym_use_BANG] = ACTIONS(6195),
    [anon_sym_do_BANG] = ACTIONS(6195),
    [anon_sym_AMP_AMP] = ACTIONS(6195),
    [anon_sym_STAR] = ACTIONS(9794),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(6193),
    [anon_sym_static] = ACTIONS(6193),
    [anon_sym_member] = ACTIONS(6193),
    [anon_sym_exception] = ACTIONS(6193),
    [anon_sym_interface] = ACTIONS(6193),
    [anon_sym_abstract] = ACTIONS(6193),
    [anon_sym_val] = ACTIONS(6193),
    [anon_sym_override] = ACTIONS(6193),
    [anon_sym_extern] = ACTIONS(6193),
    [aux_sym_char_token1] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6193),
    [anon_sym_DQUOTE] = ACTIONS(6193),
    [anon_sym_AT_DQUOTE] = ACTIONS(6195),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6195),
    [sym_bool] = ACTIONS(6193),
    [sym_unit] = ACTIONS(6195),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6193),
    [sym_op_identifier] = ACTIONS(6195),
    [anon_sym_PLUS] = ACTIONS(6193),
    [anon_sym_DASH] = ACTIONS(6193),
    [anon_sym_PLUS_DOT] = ACTIONS(6195),
    [anon_sym_DASH_DOT] = ACTIONS(6195),
    [anon_sym_PERCENT] = ACTIONS(6193),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6195),
    [anon_sym_TILDE] = ACTIONS(6195),
    [aux_sym_prefix_op_token1] = ACTIONS(6195),
    [sym_int] = ACTIONS(6193),
    [sym_xint] = ACTIONS(6195),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6195),
    [anon_sym_POUNDload] = ACTIONS(6195),
    [anon_sym_POUNDtime] = ACTIONS(6195),
    [anon_sym_POUNDI] = ACTIONS(6195),
    [anon_sym_POUNDhelp] = ACTIONS(6195),
    [anon_sym_POUNDquit] = ACTIONS(6195),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6195),
    [anon_sym_POUNDendif] = ACTIONS(6195),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6195),
  },
  [STATE(5691)] = {
    [sym_block_comment] = STATE(5691),
    [sym_compiler_directive_decl] = STATE(5691),
    [sym_fsi_directive_decl] = STATE(5691),
    [sym_preproc_line] = STATE(5691),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_DASH_GT] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_DOT_DOT] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(5692)] = {
    [sym_block_comment] = STATE(5692),
    [sym_compiler_directive_decl] = STATE(5692),
    [sym_fsi_directive_decl] = STATE(5692),
    [sym_preproc_line] = STATE(5692),
    [sym_identifier] = ACTIONS(6896),
    [anon_sym_null] = ACTIONS(6896),
    [anon_sym_EQ] = ACTIONS(6896),
    [anon_sym_begin] = ACTIONS(6896),
    [anon_sym_COLON] = ACTIONS(6896),
    [anon_sym_do] = ACTIONS(6896),
    [anon_sym_let] = ACTIONS(6896),
    [anon_sym_let_BANG] = ACTIONS(6894),
    [anon_sym_as] = ACTIONS(6896),
    [anon_sym_LPAREN] = ACTIONS(6896),
    [anon_sym_COMMA] = ACTIONS(6894),
    [anon_sym_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_QMARK] = ACTIONS(6896),
    [anon_sym_COLON_COLON] = ACTIONS(6894),
    [anon_sym_AMP] = ACTIONS(6896),
    [anon_sym_LBRACK] = ACTIONS(6896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6894),
    [anon_sym_LBRACE] = ACTIONS(6896),
    [anon_sym_LT_AT] = ACTIONS(6896),
    [anon_sym_LT_AT_AT] = ACTIONS(6896),
    [anon_sym_CARET] = ACTIONS(6896),
    [anon_sym_or] = ACTIONS(6896),
    [anon_sym_DOT] = ACTIONS(6896),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6894),
    [anon_sym_new] = ACTIONS(6896),
    [anon_sym_with] = ACTIONS(6896),
    [anon_sym_return] = ACTIONS(6896),
    [anon_sym_return_BANG] = ACTIONS(6894),
    [anon_sym_yield] = ACTIONS(6896),
    [anon_sym_yield_BANG] = ACTIONS(6894),
    [anon_sym_lazy] = ACTIONS(6896),
    [anon_sym_assert] = ACTIONS(6896),
    [anon_sym_upcast] = ACTIONS(6896),
    [anon_sym_downcast] = ACTIONS(6896),
    [anon_sym_fixed] = ACTIONS(6896),
    [anon_sym_COLON_GT] = ACTIONS(6894),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6894),
    [anon_sym_for] = ACTIONS(6896),
    [anon_sym_while] = ACTIONS(6896),
    [anon_sym_while_BANG] = ACTIONS(6894),
    [anon_sym_if] = ACTIONS(6896),
    [anon_sym_fun] = ACTIONS(6896),
    [anon_sym_try] = ACTIONS(6896),
    [anon_sym_match] = ACTIONS(6896),
    [anon_sym_match_BANG] = ACTIONS(6894),
    [anon_sym_function] = ACTIONS(6896),
    [anon_sym_DOT_LBRACK] = ACTIONS(6894),
    [anon_sym_use] = ACTIONS(6896),
    [anon_sym_use_BANG] = ACTIONS(6894),
    [anon_sym_do_BANG] = ACTIONS(6894),
    [anon_sym_LPAREN2] = ACTIONS(6896),
    [anon_sym_PIPE_PIPE] = ACTIONS(6896),
    [anon_sym_AMP_AMP] = ACTIONS(6896),
    [aux_sym_char_token1] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6896),
    [anon_sym_DQUOTE] = ACTIONS(6896),
    [anon_sym_AT_DQUOTE] = ACTIONS(6894),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6894),
    [sym_bool] = ACTIONS(6896),
    [sym_unit] = ACTIONS(6894),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6896),
    [sym_op_identifier] = ACTIONS(6896),
    [anon_sym_PLUS] = ACTIONS(6896),
    [anon_sym_DASH] = ACTIONS(6896),
    [anon_sym_PLUS_DOT] = ACTIONS(6896),
    [anon_sym_DASH_DOT] = ACTIONS(6896),
    [anon_sym_PERCENT] = ACTIONS(6896),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6896),
    [anon_sym_TILDE] = ACTIONS(6894),
    [aux_sym_prefix_op_token1] = ACTIONS(6896),
    [aux_sym_infix_op_token1] = ACTIONS(6894),
    [aux_sym_infix_op_token2] = ACTIONS(6896),
    [aux_sym_infix_op_token3] = ACTIONS(6896),
    [anon_sym_BANG_EQ] = ACTIONS(6896),
    [anon_sym_COLON_EQ] = ACTIONS(6894),
    [anon_sym_DOLLAR] = ACTIONS(6896),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6896),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6896),
    [sym_int] = ACTIONS(6896),
    [sym_xint] = ACTIONS(6894),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6894),
    [sym__newline] = ACTIONS(6894),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6894),
    [sym__tyapp_open] = ACTIONS(6894),
  },
  [STATE(5693)] = {
    [sym_block_comment] = STATE(5693),
    [sym_compiler_directive_decl] = STATE(5693),
    [sym_fsi_directive_decl] = STATE(5693),
    [sym_preproc_line] = STATE(5693),
    [sym_identifier] = ACTIONS(6900),
    [anon_sym_null] = ACTIONS(6900),
    [anon_sym_EQ] = ACTIONS(6900),
    [anon_sym_begin] = ACTIONS(6900),
    [anon_sym_COLON] = ACTIONS(6900),
    [anon_sym_do] = ACTIONS(6900),
    [anon_sym_let] = ACTIONS(6900),
    [anon_sym_let_BANG] = ACTIONS(6898),
    [anon_sym_as] = ACTIONS(6900),
    [anon_sym_LPAREN] = ACTIONS(6900),
    [anon_sym_COMMA] = ACTIONS(6898),
    [anon_sym_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_QMARK] = ACTIONS(6900),
    [anon_sym_COLON_COLON] = ACTIONS(6898),
    [anon_sym_AMP] = ACTIONS(6900),
    [anon_sym_LBRACK] = ACTIONS(6900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6898),
    [anon_sym_LBRACE] = ACTIONS(6900),
    [anon_sym_LT_AT] = ACTIONS(6900),
    [anon_sym_LT_AT_AT] = ACTIONS(6900),
    [anon_sym_CARET] = ACTIONS(6900),
    [anon_sym_or] = ACTIONS(6900),
    [anon_sym_DOT] = ACTIONS(6900),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6898),
    [anon_sym_new] = ACTIONS(6900),
    [anon_sym_with] = ACTIONS(6900),
    [anon_sym_return] = ACTIONS(6900),
    [anon_sym_return_BANG] = ACTIONS(6898),
    [anon_sym_yield] = ACTIONS(6900),
    [anon_sym_yield_BANG] = ACTIONS(6898),
    [anon_sym_lazy] = ACTIONS(6900),
    [anon_sym_assert] = ACTIONS(6900),
    [anon_sym_upcast] = ACTIONS(6900),
    [anon_sym_downcast] = ACTIONS(6900),
    [anon_sym_fixed] = ACTIONS(6900),
    [anon_sym_COLON_GT] = ACTIONS(6898),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6898),
    [anon_sym_for] = ACTIONS(6900),
    [anon_sym_while] = ACTIONS(6900),
    [anon_sym_while_BANG] = ACTIONS(6898),
    [anon_sym_if] = ACTIONS(6900),
    [anon_sym_fun] = ACTIONS(6900),
    [anon_sym_try] = ACTIONS(6900),
    [anon_sym_match] = ACTIONS(6900),
    [anon_sym_match_BANG] = ACTIONS(6898),
    [anon_sym_function] = ACTIONS(6900),
    [anon_sym_DOT_LBRACK] = ACTIONS(6898),
    [anon_sym_use] = ACTIONS(6900),
    [anon_sym_use_BANG] = ACTIONS(6898),
    [anon_sym_do_BANG] = ACTIONS(6898),
    [anon_sym_LPAREN2] = ACTIONS(6900),
    [anon_sym_PIPE_PIPE] = ACTIONS(6900),
    [anon_sym_AMP_AMP] = ACTIONS(6900),
    [aux_sym_char_token1] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6900),
    [anon_sym_DQUOTE] = ACTIONS(6900),
    [anon_sym_AT_DQUOTE] = ACTIONS(6898),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6898),
    [sym_bool] = ACTIONS(6900),
    [sym_unit] = ACTIONS(6898),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6900),
    [sym_op_identifier] = ACTIONS(6900),
    [anon_sym_PLUS] = ACTIONS(6900),
    [anon_sym_DASH] = ACTIONS(6900),
    [anon_sym_PLUS_DOT] = ACTIONS(6900),
    [anon_sym_DASH_DOT] = ACTIONS(6900),
    [anon_sym_PERCENT] = ACTIONS(6900),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6900),
    [anon_sym_TILDE] = ACTIONS(6898),
    [aux_sym_prefix_op_token1] = ACTIONS(6900),
    [aux_sym_infix_op_token1] = ACTIONS(6898),
    [aux_sym_infix_op_token2] = ACTIONS(6900),
    [aux_sym_infix_op_token3] = ACTIONS(6900),
    [anon_sym_BANG_EQ] = ACTIONS(6900),
    [anon_sym_COLON_EQ] = ACTIONS(6898),
    [anon_sym_DOLLAR] = ACTIONS(6900),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6900),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6900),
    [sym_int] = ACTIONS(6900),
    [sym_xint] = ACTIONS(6898),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6898),
    [sym__newline] = ACTIONS(6898),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6898),
    [sym__tyapp_open] = ACTIONS(6898),
  },
  [STATE(5694)] = {
    [sym_block_comment] = STATE(5694),
    [sym_compiler_directive_decl] = STATE(5694),
    [sym_fsi_directive_decl] = STATE(5694),
    [sym_preproc_line] = STATE(5694),
    [sym_identifier] = ACTIONS(6834),
    [anon_sym_null] = ACTIONS(6834),
    [anon_sym_EQ] = ACTIONS(6834),
    [anon_sym_begin] = ACTIONS(6834),
    [anon_sym_COLON] = ACTIONS(6834),
    [anon_sym_do] = ACTIONS(6834),
    [anon_sym_let] = ACTIONS(6834),
    [anon_sym_let_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN] = ACTIONS(6834),
    [anon_sym_COMMA] = ACTIONS(6832),
    [anon_sym_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_QMARK] = ACTIONS(6834),
    [anon_sym_COLON_COLON] = ACTIONS(6832),
    [anon_sym_AMP] = ACTIONS(6834),
    [anon_sym_LBRACK] = ACTIONS(6834),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6832),
    [anon_sym_LBRACE] = ACTIONS(6834),
    [anon_sym_LT_AT] = ACTIONS(6834),
    [anon_sym_LT_AT_AT] = ACTIONS(6834),
    [anon_sym_CARET] = ACTIONS(6834),
    [anon_sym_or] = ACTIONS(6834),
    [anon_sym_DOT] = ACTIONS(8930),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6832),
    [anon_sym_new] = ACTIONS(6834),
    [anon_sym_return] = ACTIONS(6834),
    [anon_sym_return_BANG] = ACTIONS(6832),
    [anon_sym_yield] = ACTIONS(6834),
    [anon_sym_yield_BANG] = ACTIONS(6832),
    [anon_sym_lazy] = ACTIONS(6834),
    [anon_sym_assert] = ACTIONS(6834),
    [anon_sym_upcast] = ACTIONS(6834),
    [anon_sym_downcast] = ACTIONS(6834),
    [anon_sym_fixed] = ACTIONS(6834),
    [anon_sym_COLON_GT] = ACTIONS(6832),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6832),
    [anon_sym_for] = ACTIONS(6834),
    [anon_sym_while] = ACTIONS(6834),
    [anon_sym_while_BANG] = ACTIONS(6832),
    [anon_sym_if] = ACTIONS(6834),
    [anon_sym_fun] = ACTIONS(6834),
    [anon_sym_try] = ACTIONS(6834),
    [anon_sym_match] = ACTIONS(6834),
    [anon_sym_match_BANG] = ACTIONS(6832),
    [anon_sym_function] = ACTIONS(6834),
    [anon_sym_LT_DASH] = ACTIONS(6834),
    [anon_sym_DOT_LBRACK] = ACTIONS(6832),
    [anon_sym_use] = ACTIONS(6834),
    [anon_sym_use_BANG] = ACTIONS(6832),
    [anon_sym_do_BANG] = ACTIONS(6832),
    [anon_sym_LPAREN2] = ACTIONS(6834),
    [anon_sym_PIPE_PIPE] = ACTIONS(6834),
    [anon_sym_AMP_AMP] = ACTIONS(6834),
    [aux_sym_char_token1] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6834),
    [anon_sym_DQUOTE] = ACTIONS(6834),
    [anon_sym_AT_DQUOTE] = ACTIONS(6832),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6832),
    [sym_bool] = ACTIONS(6834),
    [sym_unit] = ACTIONS(6832),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6834),
    [sym_op_identifier] = ACTIONS(6834),
    [anon_sym_PLUS] = ACTIONS(6834),
    [anon_sym_DASH] = ACTIONS(6834),
    [anon_sym_PLUS_DOT] = ACTIONS(6834),
    [anon_sym_DASH_DOT] = ACTIONS(6834),
    [anon_sym_PERCENT] = ACTIONS(6834),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6834),
    [anon_sym_TILDE] = ACTIONS(6832),
    [aux_sym_prefix_op_token1] = ACTIONS(6834),
    [aux_sym_infix_op_token1] = ACTIONS(6832),
    [aux_sym_infix_op_token2] = ACTIONS(6834),
    [aux_sym_infix_op_token3] = ACTIONS(6834),
    [anon_sym_BANG_EQ] = ACTIONS(6834),
    [anon_sym_COLON_EQ] = ACTIONS(6832),
    [anon_sym_DOLLAR] = ACTIONS(6834),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6834),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6834),
    [sym_int] = ACTIONS(6834),
    [sym_xint] = ACTIONS(6832),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6832),
    [sym__newline] = ACTIONS(6832),
    [sym__dedent] = ACTIONS(9830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6832),
    [sym__tyapp_open] = ACTIONS(6832),
  },
  [STATE(5695)] = {
    [sym_block_comment] = STATE(5695),
    [sym_compiler_directive_decl] = STATE(5695),
    [sym_fsi_directive_decl] = STATE(5695),
    [sym_preproc_line] = STATE(5695),
    [sym_identifier] = ACTIONS(6824),
    [anon_sym_null] = ACTIONS(6824),
    [anon_sym_EQ] = ACTIONS(6824),
    [anon_sym_begin] = ACTIONS(6824),
    [anon_sym_COLON] = ACTIONS(6824),
    [anon_sym_do] = ACTIONS(6824),
    [anon_sym_let] = ACTIONS(6824),
    [anon_sym_let_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN] = ACTIONS(6824),
    [anon_sym_COMMA] = ACTIONS(6822),
    [anon_sym_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_QMARK] = ACTIONS(6824),
    [anon_sym_COLON_COLON] = ACTIONS(6822),
    [anon_sym_PIPE] = ACTIONS(6824),
    [anon_sym_AMP] = ACTIONS(6824),
    [anon_sym_LBRACK] = ACTIONS(6824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6822),
    [anon_sym_LBRACE] = ACTIONS(6824),
    [anon_sym_LT_AT] = ACTIONS(6824),
    [anon_sym_LT_AT_AT] = ACTIONS(6824),
    [anon_sym_CARET] = ACTIONS(6824),
    [anon_sym_or] = ACTIONS(6824),
    [anon_sym_DOT] = ACTIONS(6824),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6822),
    [anon_sym_new] = ACTIONS(6824),
    [anon_sym_return] = ACTIONS(6824),
    [anon_sym_return_BANG] = ACTIONS(6822),
    [anon_sym_yield] = ACTIONS(6824),
    [anon_sym_yield_BANG] = ACTIONS(6822),
    [anon_sym_lazy] = ACTIONS(6824),
    [anon_sym_assert] = ACTIONS(6824),
    [anon_sym_upcast] = ACTIONS(6824),
    [anon_sym_downcast] = ACTIONS(6824),
    [anon_sym_fixed] = ACTIONS(6824),
    [anon_sym_COLON_GT] = ACTIONS(6822),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6822),
    [anon_sym_for] = ACTIONS(6824),
    [anon_sym_while] = ACTIONS(6824),
    [anon_sym_while_BANG] = ACTIONS(6822),
    [anon_sym_if] = ACTIONS(6824),
    [anon_sym_fun] = ACTIONS(6824),
    [anon_sym_try] = ACTIONS(6824),
    [anon_sym_match] = ACTIONS(6824),
    [anon_sym_match_BANG] = ACTIONS(6822),
    [anon_sym_function] = ACTIONS(6824),
    [anon_sym_DOT_LBRACK] = ACTIONS(6822),
    [anon_sym_use] = ACTIONS(6824),
    [anon_sym_use_BANG] = ACTIONS(6822),
    [anon_sym_do_BANG] = ACTIONS(6822),
    [anon_sym_LPAREN2] = ACTIONS(6824),
    [anon_sym_PIPE_PIPE] = ACTIONS(6824),
    [anon_sym_AMP_AMP] = ACTIONS(6824),
    [aux_sym_char_token1] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6824),
    [anon_sym_DQUOTE] = ACTIONS(6824),
    [anon_sym_AT_DQUOTE] = ACTIONS(6822),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6822),
    [sym_bool] = ACTIONS(6824),
    [sym_unit] = ACTIONS(6822),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6824),
    [sym_op_identifier] = ACTIONS(6824),
    [anon_sym_PLUS] = ACTIONS(6824),
    [anon_sym_DASH] = ACTIONS(6824),
    [anon_sym_PLUS_DOT] = ACTIONS(6824),
    [anon_sym_DASH_DOT] = ACTIONS(6824),
    [anon_sym_PERCENT] = ACTIONS(6824),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6824),
    [anon_sym_TILDE] = ACTIONS(6822),
    [aux_sym_prefix_op_token1] = ACTIONS(6824),
    [aux_sym_infix_op_token1] = ACTIONS(6822),
    [aux_sym_infix_op_token2] = ACTIONS(6824),
    [aux_sym_infix_op_token3] = ACTIONS(6824),
    [anon_sym_BANG_EQ] = ACTIONS(6824),
    [anon_sym_COLON_EQ] = ACTIONS(6822),
    [anon_sym_DOLLAR] = ACTIONS(6824),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6824),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6824),
    [sym_int] = ACTIONS(6824),
    [sym_xint] = ACTIONS(6822),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6822),
    [sym__newline] = ACTIONS(6822),
    [sym__quoted_close] = ACTIONS(6822),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6822),
    [sym__tyapp_open] = ACTIONS(6822),
  },
  [STATE(5696)] = {
    [sym_block_comment] = STATE(5696),
    [sym_compiler_directive_decl] = STATE(5696),
    [sym_fsi_directive_decl] = STATE(5696),
    [sym_preproc_line] = STATE(5696),
    [sym_identifier] = ACTIONS(6902),
    [anon_sym_null] = ACTIONS(6902),
    [anon_sym_EQ] = ACTIONS(6902),
    [anon_sym_begin] = ACTIONS(6902),
    [anon_sym_COLON] = ACTIONS(6902),
    [anon_sym_do] = ACTIONS(6902),
    [anon_sym_let] = ACTIONS(6902),
    [anon_sym_let_BANG] = ACTIONS(6603),
    [anon_sym_as] = ACTIONS(6902),
    [anon_sym_LPAREN] = ACTIONS(6902),
    [anon_sym_COMMA] = ACTIONS(6603),
    [anon_sym_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_QMARK] = ACTIONS(6902),
    [anon_sym_COLON_COLON] = ACTIONS(6603),
    [anon_sym_AMP] = ACTIONS(6902),
    [anon_sym_LBRACK] = ACTIONS(6902),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6603),
    [anon_sym_LBRACE] = ACTIONS(6902),
    [anon_sym_LT_AT] = ACTIONS(6902),
    [anon_sym_LT_AT_AT] = ACTIONS(6902),
    [anon_sym_CARET] = ACTIONS(6902),
    [anon_sym_or] = ACTIONS(6902),
    [anon_sym_DOT] = ACTIONS(6902),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6603),
    [anon_sym_new] = ACTIONS(6902),
    [anon_sym_with] = ACTIONS(6902),
    [anon_sym_return] = ACTIONS(6902),
    [anon_sym_return_BANG] = ACTIONS(6603),
    [anon_sym_yield] = ACTIONS(6902),
    [anon_sym_yield_BANG] = ACTIONS(6603),
    [anon_sym_lazy] = ACTIONS(6902),
    [anon_sym_assert] = ACTIONS(6902),
    [anon_sym_upcast] = ACTIONS(6902),
    [anon_sym_downcast] = ACTIONS(6902),
    [anon_sym_fixed] = ACTIONS(6902),
    [anon_sym_COLON_GT] = ACTIONS(6603),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6603),
    [anon_sym_for] = ACTIONS(6902),
    [anon_sym_while] = ACTIONS(6902),
    [anon_sym_while_BANG] = ACTIONS(6603),
    [anon_sym_if] = ACTIONS(6902),
    [anon_sym_fun] = ACTIONS(6902),
    [anon_sym_try] = ACTIONS(6902),
    [anon_sym_match] = ACTIONS(6902),
    [anon_sym_match_BANG] = ACTIONS(6603),
    [anon_sym_function] = ACTIONS(6902),
    [anon_sym_DOT_LBRACK] = ACTIONS(6603),
    [anon_sym_use] = ACTIONS(6902),
    [anon_sym_use_BANG] = ACTIONS(6603),
    [anon_sym_do_BANG] = ACTIONS(6603),
    [anon_sym_LPAREN2] = ACTIONS(6902),
    [anon_sym_PIPE_PIPE] = ACTIONS(6902),
    [anon_sym_AMP_AMP] = ACTIONS(6902),
    [aux_sym_char_token1] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6902),
    [anon_sym_DQUOTE] = ACTIONS(6902),
    [anon_sym_AT_DQUOTE] = ACTIONS(6603),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6603),
    [sym_bool] = ACTIONS(6902),
    [sym_unit] = ACTIONS(6603),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6902),
    [sym_op_identifier] = ACTIONS(6902),
    [anon_sym_PLUS] = ACTIONS(6902),
    [anon_sym_DASH] = ACTIONS(6902),
    [anon_sym_PLUS_DOT] = ACTIONS(6902),
    [anon_sym_DASH_DOT] = ACTIONS(6902),
    [anon_sym_PERCENT] = ACTIONS(6902),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6902),
    [anon_sym_TILDE] = ACTIONS(6603),
    [aux_sym_prefix_op_token1] = ACTIONS(6902),
    [aux_sym_infix_op_token1] = ACTIONS(6603),
    [aux_sym_infix_op_token2] = ACTIONS(6902),
    [aux_sym_infix_op_token3] = ACTIONS(6902),
    [anon_sym_BANG_EQ] = ACTIONS(6902),
    [anon_sym_COLON_EQ] = ACTIONS(6603),
    [anon_sym_DOLLAR] = ACTIONS(6902),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6902),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6902),
    [sym_int] = ACTIONS(6902),
    [sym_xint] = ACTIONS(6603),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6603),
    [sym__newline] = ACTIONS(6603),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6603),
    [sym__tyapp_open] = ACTIONS(6603),
  },
  [STATE(5697)] = {
    [sym_block_comment] = STATE(5697),
    [sym_compiler_directive_decl] = STATE(5697),
    [sym_fsi_directive_decl] = STATE(5697),
    [sym_preproc_line] = STATE(5697),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_DASH_GT] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_DOT_DOT] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(5698)] = {
    [sym_block_comment] = STATE(5698),
    [sym_compiler_directive_decl] = STATE(5698),
    [sym_fsi_directive_decl] = STATE(5698),
    [sym_preproc_line] = STATE(5698),
    [aux_sym_sequential_expression_repeat1] = STATE(5720),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__quoted_close] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5699)] = {
    [sym_block_comment] = STATE(5699),
    [sym_compiler_directive_decl] = STATE(5699),
    [sym_fsi_directive_decl] = STATE(5699),
    [sym_preproc_line] = STATE(5699),
    [aux_sym_sequential_expression_repeat1] = STATE(5699),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_GT] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9832),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5700)] = {
    [sym_type_arguments] = STATE(6569),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5700),
    [sym_compiler_directive_decl] = STATE(5700),
    [sym_fsi_directive_decl] = STATE(5700),
    [sym_preproc_line] = STATE(5700),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(6261),
    [anon_sym_null] = ACTIONS(6261),
    [anon_sym_begin] = ACTIONS(6261),
    [anon_sym_open] = ACTIONS(6261),
    [anon_sym_type] = ACTIONS(6261),
    [anon_sym_LBRACK_LT] = ACTIONS(6259),
    [anon_sym_do] = ACTIONS(6261),
    [anon_sym_and] = ACTIONS(6261),
    [anon_sym_let] = ACTIONS(6261),
    [anon_sym_let_BANG] = ACTIONS(6259),
    [aux_sym_access_modifier_token1] = ACTIONS(6259),
    [anon_sym_LPAREN] = ACTIONS(6261),
    [anon_sym_PIPE] = ACTIONS(6259),
    [anon_sym_AMP] = ACTIONS(6261),
    [anon_sym_LBRACK] = ACTIONS(6261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6259),
    [anon_sym_LBRACE] = ACTIONS(6261),
    [anon_sym_LT_AT] = ACTIONS(6261),
    [anon_sym_LT_AT_AT] = ACTIONS(6259),
    [anon_sym_CARET] = ACTIONS(6259),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6259),
    [anon_sym_new] = ACTIONS(6261),
    [anon_sym_with] = ACTIONS(6261),
    [anon_sym_return] = ACTIONS(6261),
    [anon_sym_return_BANG] = ACTIONS(6259),
    [anon_sym_yield] = ACTIONS(6261),
    [anon_sym_yield_BANG] = ACTIONS(6259),
    [anon_sym_lazy] = ACTIONS(6261),
    [anon_sym_assert] = ACTIONS(6261),
    [anon_sym_upcast] = ACTIONS(6261),
    [anon_sym_downcast] = ACTIONS(6261),
    [anon_sym_fixed] = ACTIONS(6261),
    [anon_sym_for] = ACTIONS(6261),
    [anon_sym_while] = ACTIONS(6261),
    [anon_sym_while_BANG] = ACTIONS(6259),
    [anon_sym_if] = ACTIONS(6261),
    [anon_sym_fun] = ACTIONS(6261),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(6261),
    [anon_sym_match] = ACTIONS(6261),
    [anon_sym_match_BANG] = ACTIONS(6259),
    [anon_sym_function] = ACTIONS(6261),
    [anon_sym_use] = ACTIONS(6261),
    [anon_sym_use_BANG] = ACTIONS(6259),
    [anon_sym_do_BANG] = ACTIONS(6259),
    [anon_sym_AMP_AMP] = ACTIONS(6259),
    [anon_sym_STAR] = ACTIONS(9794),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(6261),
    [anon_sym_static] = ACTIONS(6261),
    [anon_sym_member] = ACTIONS(6261),
    [anon_sym_exception] = ACTIONS(6261),
    [anon_sym_interface] = ACTIONS(6261),
    [anon_sym_abstract] = ACTIONS(6261),
    [anon_sym_val] = ACTIONS(6261),
    [anon_sym_override] = ACTIONS(6261),
    [anon_sym_extern] = ACTIONS(6261),
    [aux_sym_char_token1] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6261),
    [anon_sym_DQUOTE] = ACTIONS(6261),
    [anon_sym_AT_DQUOTE] = ACTIONS(6259),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6259),
    [sym_bool] = ACTIONS(6261),
    [sym_unit] = ACTIONS(6259),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6261),
    [sym_op_identifier] = ACTIONS(6259),
    [anon_sym_PLUS] = ACTIONS(6261),
    [anon_sym_DASH] = ACTIONS(6261),
    [anon_sym_PLUS_DOT] = ACTIONS(6259),
    [anon_sym_DASH_DOT] = ACTIONS(6259),
    [anon_sym_PERCENT] = ACTIONS(6261),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6259),
    [anon_sym_TILDE] = ACTIONS(6259),
    [aux_sym_prefix_op_token1] = ACTIONS(6259),
    [sym_int] = ACTIONS(6261),
    [sym_xint] = ACTIONS(6259),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6259),
    [anon_sym_POUNDload] = ACTIONS(6259),
    [anon_sym_POUNDtime] = ACTIONS(6259),
    [anon_sym_POUNDI] = ACTIONS(6259),
    [anon_sym_POUNDhelp] = ACTIONS(6259),
    [anon_sym_POUNDquit] = ACTIONS(6259),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6259),
    [anon_sym_POUNDendif] = ACTIONS(6259),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6259),
  },
  [STATE(5701)] = {
    [sym_block_comment] = STATE(5701),
    [sym_compiler_directive_decl] = STATE(5701),
    [sym_fsi_directive_decl] = STATE(5701),
    [sym_preproc_line] = STATE(5701),
    [sym_identifier] = ACTIONS(6944),
    [anon_sym_null] = ACTIONS(6944),
    [anon_sym_EQ] = ACTIONS(6944),
    [anon_sym_begin] = ACTIONS(6944),
    [anon_sym_COLON] = ACTIONS(6944),
    [anon_sym_do] = ACTIONS(6944),
    [anon_sym_let] = ACTIONS(6944),
    [anon_sym_let_BANG] = ACTIONS(6942),
    [anon_sym_LPAREN] = ACTIONS(6944),
    [anon_sym_COMMA] = ACTIONS(6942),
    [anon_sym_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_QMARK] = ACTIONS(6944),
    [anon_sym_COLON_COLON] = ACTIONS(6942),
    [anon_sym_AMP] = ACTIONS(6944),
    [anon_sym_LBRACK] = ACTIONS(6944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6942),
    [anon_sym_LBRACE] = ACTIONS(6944),
    [anon_sym_LT_AT] = ACTIONS(6944),
    [anon_sym_LT_AT_AT] = ACTIONS(6944),
    [anon_sym_CARET] = ACTIONS(6944),
    [anon_sym_or] = ACTIONS(6944),
    [anon_sym_DOT] = ACTIONS(6944),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6942),
    [anon_sym_new] = ACTIONS(6944),
    [anon_sym_return] = ACTIONS(6944),
    [anon_sym_return_BANG] = ACTIONS(6942),
    [anon_sym_yield] = ACTIONS(6944),
    [anon_sym_yield_BANG] = ACTIONS(6942),
    [anon_sym_lazy] = ACTIONS(6944),
    [anon_sym_assert] = ACTIONS(6944),
    [anon_sym_upcast] = ACTIONS(6944),
    [anon_sym_downcast] = ACTIONS(6944),
    [anon_sym_fixed] = ACTIONS(6944),
    [anon_sym_COLON_GT] = ACTIONS(6942),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6942),
    [anon_sym_for] = ACTIONS(6944),
    [anon_sym_while] = ACTIONS(6944),
    [anon_sym_while_BANG] = ACTIONS(6942),
    [anon_sym_if] = ACTIONS(6944),
    [anon_sym_fun] = ACTIONS(6944),
    [anon_sym_DASH_GT] = ACTIONS(6944),
    [anon_sym_try] = ACTIONS(6944),
    [anon_sym_match] = ACTIONS(6944),
    [anon_sym_match_BANG] = ACTIONS(6942),
    [anon_sym_function] = ACTIONS(6944),
    [anon_sym_DOT_LBRACK] = ACTIONS(6942),
    [anon_sym_use] = ACTIONS(6944),
    [anon_sym_use_BANG] = ACTIONS(6942),
    [anon_sym_do_BANG] = ACTIONS(6942),
    [anon_sym_DOT_DOT] = ACTIONS(6942),
    [anon_sym_LPAREN2] = ACTIONS(6944),
    [anon_sym_PIPE_PIPE] = ACTIONS(6944),
    [anon_sym_AMP_AMP] = ACTIONS(6944),
    [aux_sym_char_token1] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6944),
    [anon_sym_DQUOTE] = ACTIONS(6944),
    [anon_sym_AT_DQUOTE] = ACTIONS(6942),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6942),
    [sym_bool] = ACTIONS(6944),
    [sym_unit] = ACTIONS(6942),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6944),
    [sym_op_identifier] = ACTIONS(6944),
    [anon_sym_PLUS] = ACTIONS(6944),
    [anon_sym_DASH] = ACTIONS(6944),
    [anon_sym_PLUS_DOT] = ACTIONS(6944),
    [anon_sym_DASH_DOT] = ACTIONS(6944),
    [anon_sym_PERCENT] = ACTIONS(6944),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6944),
    [anon_sym_TILDE] = ACTIONS(6942),
    [aux_sym_prefix_op_token1] = ACTIONS(6944),
    [aux_sym_infix_op_token1] = ACTIONS(6942),
    [aux_sym_infix_op_token2] = ACTIONS(6944),
    [aux_sym_infix_op_token3] = ACTIONS(6944),
    [anon_sym_BANG_EQ] = ACTIONS(6944),
    [anon_sym_COLON_EQ] = ACTIONS(6942),
    [anon_sym_DOLLAR] = ACTIONS(6944),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6944),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6944),
    [sym_int] = ACTIONS(6944),
    [sym_xint] = ACTIONS(6942),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6942),
    [sym__newline] = ACTIONS(6942),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6942),
    [sym__tyapp_open] = ACTIONS(6942),
  },
  [STATE(5702)] = {
    [sym_block_comment] = STATE(5702),
    [sym_compiler_directive_decl] = STATE(5702),
    [sym_fsi_directive_decl] = STATE(5702),
    [sym_preproc_line] = STATE(5702),
    [sym_identifier] = ACTIONS(6948),
    [anon_sym_null] = ACTIONS(6948),
    [anon_sym_EQ] = ACTIONS(6948),
    [anon_sym_begin] = ACTIONS(6948),
    [anon_sym_COLON] = ACTIONS(6948),
    [anon_sym_do] = ACTIONS(6948),
    [anon_sym_let] = ACTIONS(6948),
    [anon_sym_let_BANG] = ACTIONS(6946),
    [anon_sym_LPAREN] = ACTIONS(6948),
    [anon_sym_COMMA] = ACTIONS(6946),
    [anon_sym_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_QMARK] = ACTIONS(6948),
    [anon_sym_COLON_COLON] = ACTIONS(6946),
    [anon_sym_AMP] = ACTIONS(6948),
    [anon_sym_LBRACK] = ACTIONS(6948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6946),
    [anon_sym_LBRACE] = ACTIONS(6948),
    [anon_sym_LT_AT] = ACTIONS(6948),
    [anon_sym_LT_AT_AT] = ACTIONS(6948),
    [anon_sym_CARET] = ACTIONS(6948),
    [anon_sym_or] = ACTIONS(6948),
    [anon_sym_DOT] = ACTIONS(6948),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6946),
    [anon_sym_new] = ACTIONS(6948),
    [anon_sym_return] = ACTIONS(6948),
    [anon_sym_return_BANG] = ACTIONS(6946),
    [anon_sym_yield] = ACTIONS(6948),
    [anon_sym_yield_BANG] = ACTIONS(6946),
    [anon_sym_lazy] = ACTIONS(6948),
    [anon_sym_assert] = ACTIONS(6948),
    [anon_sym_upcast] = ACTIONS(6948),
    [anon_sym_downcast] = ACTIONS(6948),
    [anon_sym_fixed] = ACTIONS(6948),
    [anon_sym_COLON_GT] = ACTIONS(6946),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6946),
    [anon_sym_for] = ACTIONS(6948),
    [anon_sym_while] = ACTIONS(6948),
    [anon_sym_while_BANG] = ACTIONS(6946),
    [anon_sym_if] = ACTIONS(6948),
    [anon_sym_fun] = ACTIONS(6948),
    [anon_sym_DASH_GT] = ACTIONS(6948),
    [anon_sym_try] = ACTIONS(6948),
    [anon_sym_match] = ACTIONS(6948),
    [anon_sym_match_BANG] = ACTIONS(6946),
    [anon_sym_function] = ACTIONS(6948),
    [anon_sym_DOT_LBRACK] = ACTIONS(6946),
    [anon_sym_use] = ACTIONS(6948),
    [anon_sym_use_BANG] = ACTIONS(6946),
    [anon_sym_do_BANG] = ACTIONS(6946),
    [anon_sym_DOT_DOT] = ACTIONS(6946),
    [anon_sym_LPAREN2] = ACTIONS(6948),
    [anon_sym_PIPE_PIPE] = ACTIONS(6948),
    [anon_sym_AMP_AMP] = ACTIONS(6948),
    [aux_sym_char_token1] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6948),
    [anon_sym_DQUOTE] = ACTIONS(6948),
    [anon_sym_AT_DQUOTE] = ACTIONS(6946),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6946),
    [sym_bool] = ACTIONS(6948),
    [sym_unit] = ACTIONS(6946),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6948),
    [sym_op_identifier] = ACTIONS(6948),
    [anon_sym_PLUS] = ACTIONS(6948),
    [anon_sym_DASH] = ACTIONS(6948),
    [anon_sym_PLUS_DOT] = ACTIONS(6948),
    [anon_sym_DASH_DOT] = ACTIONS(6948),
    [anon_sym_PERCENT] = ACTIONS(6948),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6948),
    [anon_sym_TILDE] = ACTIONS(6946),
    [aux_sym_prefix_op_token1] = ACTIONS(6948),
    [aux_sym_infix_op_token1] = ACTIONS(6946),
    [aux_sym_infix_op_token2] = ACTIONS(6948),
    [aux_sym_infix_op_token3] = ACTIONS(6948),
    [anon_sym_BANG_EQ] = ACTIONS(6948),
    [anon_sym_COLON_EQ] = ACTIONS(6946),
    [anon_sym_DOLLAR] = ACTIONS(6948),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6948),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6948),
    [sym_int] = ACTIONS(6948),
    [sym_xint] = ACTIONS(6946),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6946),
    [sym__newline] = ACTIONS(6946),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6946),
    [sym__tyapp_open] = ACTIONS(6946),
  },
  [STATE(5703)] = {
    [sym_block_comment] = STATE(5703),
    [sym_compiler_directive_decl] = STATE(5703),
    [sym_fsi_directive_decl] = STATE(5703),
    [sym_preproc_line] = STATE(5703),
    [sym_identifier] = ACTIONS(6952),
    [anon_sym_null] = ACTIONS(6952),
    [anon_sym_EQ] = ACTIONS(6952),
    [anon_sym_begin] = ACTIONS(6952),
    [anon_sym_COLON] = ACTIONS(6952),
    [anon_sym_do] = ACTIONS(6952),
    [anon_sym_let] = ACTIONS(6952),
    [anon_sym_let_BANG] = ACTIONS(6950),
    [anon_sym_LPAREN] = ACTIONS(6952),
    [anon_sym_COMMA] = ACTIONS(6950),
    [anon_sym_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_QMARK] = ACTIONS(6952),
    [anon_sym_COLON_COLON] = ACTIONS(6950),
    [anon_sym_AMP] = ACTIONS(6952),
    [anon_sym_LBRACK] = ACTIONS(6952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6950),
    [anon_sym_LBRACE] = ACTIONS(6952),
    [anon_sym_LT_AT] = ACTIONS(6952),
    [anon_sym_LT_AT_AT] = ACTIONS(6952),
    [anon_sym_CARET] = ACTIONS(6952),
    [anon_sym_or] = ACTIONS(6952),
    [anon_sym_DOT] = ACTIONS(6952),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6950),
    [anon_sym_new] = ACTIONS(6952),
    [anon_sym_return] = ACTIONS(6952),
    [anon_sym_return_BANG] = ACTIONS(6950),
    [anon_sym_yield] = ACTIONS(6952),
    [anon_sym_yield_BANG] = ACTIONS(6950),
    [anon_sym_lazy] = ACTIONS(6952),
    [anon_sym_assert] = ACTIONS(6952),
    [anon_sym_upcast] = ACTIONS(6952),
    [anon_sym_downcast] = ACTIONS(6952),
    [anon_sym_fixed] = ACTIONS(6952),
    [anon_sym_COLON_GT] = ACTIONS(6950),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6950),
    [anon_sym_for] = ACTIONS(6952),
    [anon_sym_while] = ACTIONS(6952),
    [anon_sym_while_BANG] = ACTIONS(6950),
    [anon_sym_if] = ACTIONS(6952),
    [anon_sym_fun] = ACTIONS(6952),
    [anon_sym_DASH_GT] = ACTIONS(6952),
    [anon_sym_try] = ACTIONS(6952),
    [anon_sym_match] = ACTIONS(6952),
    [anon_sym_match_BANG] = ACTIONS(6950),
    [anon_sym_function] = ACTIONS(6952),
    [anon_sym_DOT_LBRACK] = ACTIONS(6950),
    [anon_sym_use] = ACTIONS(6952),
    [anon_sym_use_BANG] = ACTIONS(6950),
    [anon_sym_do_BANG] = ACTIONS(6950),
    [anon_sym_DOT_DOT] = ACTIONS(6950),
    [anon_sym_LPAREN2] = ACTIONS(6952),
    [anon_sym_PIPE_PIPE] = ACTIONS(6952),
    [anon_sym_AMP_AMP] = ACTIONS(6952),
    [aux_sym_char_token1] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6952),
    [anon_sym_DQUOTE] = ACTIONS(6952),
    [anon_sym_AT_DQUOTE] = ACTIONS(6950),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6950),
    [sym_bool] = ACTIONS(6952),
    [sym_unit] = ACTIONS(6950),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6952),
    [sym_op_identifier] = ACTIONS(6952),
    [anon_sym_PLUS] = ACTIONS(6952),
    [anon_sym_DASH] = ACTIONS(6952),
    [anon_sym_PLUS_DOT] = ACTIONS(6952),
    [anon_sym_DASH_DOT] = ACTIONS(6952),
    [anon_sym_PERCENT] = ACTIONS(6952),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6952),
    [anon_sym_TILDE] = ACTIONS(6950),
    [aux_sym_prefix_op_token1] = ACTIONS(6952),
    [aux_sym_infix_op_token1] = ACTIONS(6950),
    [aux_sym_infix_op_token2] = ACTIONS(6952),
    [aux_sym_infix_op_token3] = ACTIONS(6952),
    [anon_sym_BANG_EQ] = ACTIONS(6952),
    [anon_sym_COLON_EQ] = ACTIONS(6950),
    [anon_sym_DOLLAR] = ACTIONS(6952),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6952),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6952),
    [sym_int] = ACTIONS(6952),
    [sym_xint] = ACTIONS(6950),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6950),
    [sym__newline] = ACTIONS(6950),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6950),
    [sym__tyapp_open] = ACTIONS(6950),
  },
  [STATE(5704)] = {
    [sym_block_comment] = STATE(5704),
    [sym_compiler_directive_decl] = STATE(5704),
    [sym_fsi_directive_decl] = STATE(5704),
    [sym_preproc_line] = STATE(5704),
    [sym_identifier] = ACTIONS(6959),
    [anon_sym_null] = ACTIONS(6959),
    [anon_sym_EQ] = ACTIONS(6959),
    [anon_sym_begin] = ACTIONS(6959),
    [anon_sym_COLON] = ACTIONS(6959),
    [anon_sym_do] = ACTIONS(6959),
    [anon_sym_let] = ACTIONS(6959),
    [anon_sym_let_BANG] = ACTIONS(6957),
    [anon_sym_LPAREN] = ACTIONS(6959),
    [anon_sym_COMMA] = ACTIONS(6957),
    [anon_sym_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_QMARK] = ACTIONS(6959),
    [anon_sym_COLON_COLON] = ACTIONS(6957),
    [anon_sym_AMP] = ACTIONS(6959),
    [anon_sym_LBRACK] = ACTIONS(6959),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6957),
    [anon_sym_LBRACE] = ACTIONS(6959),
    [anon_sym_LT_AT] = ACTIONS(6959),
    [anon_sym_LT_AT_AT] = ACTIONS(6959),
    [anon_sym_CARET] = ACTIONS(6959),
    [anon_sym_or] = ACTIONS(6959),
    [anon_sym_DOT] = ACTIONS(6959),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6957),
    [anon_sym_new] = ACTIONS(6959),
    [anon_sym_return] = ACTIONS(6959),
    [anon_sym_return_BANG] = ACTIONS(6957),
    [anon_sym_yield] = ACTIONS(6959),
    [anon_sym_yield_BANG] = ACTIONS(6957),
    [anon_sym_lazy] = ACTIONS(6959),
    [anon_sym_assert] = ACTIONS(6959),
    [anon_sym_upcast] = ACTIONS(6959),
    [anon_sym_downcast] = ACTIONS(6959),
    [anon_sym_fixed] = ACTIONS(6959),
    [anon_sym_COLON_GT] = ACTIONS(6957),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6957),
    [anon_sym_for] = ACTIONS(6959),
    [anon_sym_while] = ACTIONS(6959),
    [anon_sym_while_BANG] = ACTIONS(6957),
    [anon_sym_if] = ACTIONS(6959),
    [anon_sym_fun] = ACTIONS(6959),
    [anon_sym_DASH_GT] = ACTIONS(6959),
    [anon_sym_try] = ACTIONS(6959),
    [anon_sym_match] = ACTIONS(6959),
    [anon_sym_match_BANG] = ACTIONS(6957),
    [anon_sym_function] = ACTIONS(6959),
    [anon_sym_DOT_LBRACK] = ACTIONS(6957),
    [anon_sym_use] = ACTIONS(6959),
    [anon_sym_use_BANG] = ACTIONS(6957),
    [anon_sym_do_BANG] = ACTIONS(6957),
    [anon_sym_DOT_DOT] = ACTIONS(6957),
    [anon_sym_LPAREN2] = ACTIONS(6959),
    [anon_sym_PIPE_PIPE] = ACTIONS(6959),
    [anon_sym_AMP_AMP] = ACTIONS(6959),
    [aux_sym_char_token1] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6959),
    [anon_sym_DQUOTE] = ACTIONS(6959),
    [anon_sym_AT_DQUOTE] = ACTIONS(6957),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6957),
    [sym_bool] = ACTIONS(6959),
    [sym_unit] = ACTIONS(6957),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6959),
    [sym_op_identifier] = ACTIONS(6959),
    [anon_sym_PLUS] = ACTIONS(6959),
    [anon_sym_DASH] = ACTIONS(6959),
    [anon_sym_PLUS_DOT] = ACTIONS(6959),
    [anon_sym_DASH_DOT] = ACTIONS(6959),
    [anon_sym_PERCENT] = ACTIONS(6959),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6959),
    [anon_sym_TILDE] = ACTIONS(6957),
    [aux_sym_prefix_op_token1] = ACTIONS(6959),
    [aux_sym_infix_op_token1] = ACTIONS(6957),
    [aux_sym_infix_op_token2] = ACTIONS(6959),
    [aux_sym_infix_op_token3] = ACTIONS(6959),
    [anon_sym_BANG_EQ] = ACTIONS(6959),
    [anon_sym_COLON_EQ] = ACTIONS(6957),
    [anon_sym_DOLLAR] = ACTIONS(6959),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6959),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6959),
    [sym_int] = ACTIONS(6959),
    [sym_xint] = ACTIONS(6957),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6957),
    [sym__newline] = ACTIONS(6957),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6957),
    [sym__tyapp_open] = ACTIONS(6957),
  },
  [STATE(5705)] = {
    [sym_block_comment] = STATE(5705),
    [sym_compiler_directive_decl] = STATE(5705),
    [sym_fsi_directive_decl] = STATE(5705),
    [sym_preproc_line] = STATE(5705),
    [sym_identifier] = ACTIONS(6963),
    [anon_sym_null] = ACTIONS(6963),
    [anon_sym_EQ] = ACTIONS(6963),
    [anon_sym_begin] = ACTIONS(6963),
    [anon_sym_COLON] = ACTIONS(6963),
    [anon_sym_do] = ACTIONS(6963),
    [anon_sym_let] = ACTIONS(6963),
    [anon_sym_let_BANG] = ACTIONS(6961),
    [anon_sym_LPAREN] = ACTIONS(6963),
    [anon_sym_COMMA] = ACTIONS(6961),
    [anon_sym_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_QMARK] = ACTIONS(6963),
    [anon_sym_COLON_COLON] = ACTIONS(6961),
    [anon_sym_AMP] = ACTIONS(6963),
    [anon_sym_LBRACK] = ACTIONS(6963),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6961),
    [anon_sym_LBRACE] = ACTIONS(6963),
    [anon_sym_LT_AT] = ACTIONS(6963),
    [anon_sym_LT_AT_AT] = ACTIONS(6963),
    [anon_sym_CARET] = ACTIONS(6963),
    [anon_sym_or] = ACTIONS(6963),
    [anon_sym_DOT] = ACTIONS(6963),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6961),
    [anon_sym_new] = ACTIONS(6963),
    [anon_sym_return] = ACTIONS(6963),
    [anon_sym_return_BANG] = ACTIONS(6961),
    [anon_sym_yield] = ACTIONS(6963),
    [anon_sym_yield_BANG] = ACTIONS(6961),
    [anon_sym_lazy] = ACTIONS(6963),
    [anon_sym_assert] = ACTIONS(6963),
    [anon_sym_upcast] = ACTIONS(6963),
    [anon_sym_downcast] = ACTIONS(6963),
    [anon_sym_fixed] = ACTIONS(6963),
    [anon_sym_COLON_GT] = ACTIONS(6961),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6961),
    [anon_sym_for] = ACTIONS(6963),
    [anon_sym_while] = ACTIONS(6963),
    [anon_sym_while_BANG] = ACTIONS(6961),
    [anon_sym_if] = ACTIONS(6963),
    [anon_sym_fun] = ACTIONS(6963),
    [anon_sym_DASH_GT] = ACTIONS(6963),
    [anon_sym_try] = ACTIONS(6963),
    [anon_sym_match] = ACTIONS(6963),
    [anon_sym_match_BANG] = ACTIONS(6961),
    [anon_sym_function] = ACTIONS(6963),
    [anon_sym_DOT_LBRACK] = ACTIONS(6961),
    [anon_sym_use] = ACTIONS(6963),
    [anon_sym_use_BANG] = ACTIONS(6961),
    [anon_sym_do_BANG] = ACTIONS(6961),
    [anon_sym_DOT_DOT] = ACTIONS(6961),
    [anon_sym_LPAREN2] = ACTIONS(6963),
    [anon_sym_PIPE_PIPE] = ACTIONS(6963),
    [anon_sym_AMP_AMP] = ACTIONS(6963),
    [aux_sym_char_token1] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6963),
    [anon_sym_DQUOTE] = ACTIONS(6963),
    [anon_sym_AT_DQUOTE] = ACTIONS(6961),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6961),
    [sym_bool] = ACTIONS(6963),
    [sym_unit] = ACTIONS(6961),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6963),
    [sym_op_identifier] = ACTIONS(6963),
    [anon_sym_PLUS] = ACTIONS(6963),
    [anon_sym_DASH] = ACTIONS(6963),
    [anon_sym_PLUS_DOT] = ACTIONS(6963),
    [anon_sym_DASH_DOT] = ACTIONS(6963),
    [anon_sym_PERCENT] = ACTIONS(6963),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6963),
    [anon_sym_TILDE] = ACTIONS(6961),
    [aux_sym_prefix_op_token1] = ACTIONS(6963),
    [aux_sym_infix_op_token1] = ACTIONS(6961),
    [aux_sym_infix_op_token2] = ACTIONS(6963),
    [aux_sym_infix_op_token3] = ACTIONS(6963),
    [anon_sym_BANG_EQ] = ACTIONS(6963),
    [anon_sym_COLON_EQ] = ACTIONS(6961),
    [anon_sym_DOLLAR] = ACTIONS(6963),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6963),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6963),
    [sym_int] = ACTIONS(6963),
    [sym_xint] = ACTIONS(6961),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6961),
    [sym__newline] = ACTIONS(6961),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6961),
    [sym__tyapp_open] = ACTIONS(6961),
  },
  [STATE(5706)] = {
    [sym_block_comment] = STATE(5706),
    [sym_compiler_directive_decl] = STATE(5706),
    [sym_fsi_directive_decl] = STATE(5706),
    [sym_preproc_line] = STATE(5706),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(8332),
    [sym__dedent] = ACTIONS(8330),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5707)] = {
    [sym_block_comment] = STATE(5707),
    [sym_compiler_directive_decl] = STATE(5707),
    [sym_fsi_directive_decl] = STATE(5707),
    [sym_preproc_line] = STATE(5707),
    [ts_builtin_sym_end] = ACTIONS(6418),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_namespace] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_PIPE] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6418),
    [anon_sym_CARET] = ACTIONS(6418),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6418),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_AMP_AMP] = ACTIONS(6418),
    [anon_sym_STAR] = ACTIONS(6418),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6418),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6418),
    [anon_sym_DASH_DOT] = ACTIONS(6418),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6418),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6418),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
  },
  [STATE(5708)] = {
    [sym_block_comment] = STATE(5708),
    [sym_compiler_directive_decl] = STATE(5708),
    [sym_fsi_directive_decl] = STATE(5708),
    [sym_preproc_line] = STATE(5708),
    [sym_identifier] = ACTIONS(6967),
    [anon_sym_null] = ACTIONS(6967),
    [anon_sym_EQ] = ACTIONS(6967),
    [anon_sym_begin] = ACTIONS(6967),
    [anon_sym_COLON] = ACTIONS(6967),
    [anon_sym_do] = ACTIONS(6967),
    [anon_sym_let] = ACTIONS(6967),
    [anon_sym_let_BANG] = ACTIONS(6965),
    [anon_sym_LPAREN] = ACTIONS(6967),
    [anon_sym_COMMA] = ACTIONS(6965),
    [anon_sym_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_QMARK] = ACTIONS(6967),
    [anon_sym_COLON_COLON] = ACTIONS(6965),
    [anon_sym_AMP] = ACTIONS(6967),
    [anon_sym_LBRACK] = ACTIONS(6967),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6965),
    [anon_sym_LBRACE] = ACTIONS(6967),
    [anon_sym_LT_AT] = ACTIONS(6967),
    [anon_sym_LT_AT_AT] = ACTIONS(6967),
    [anon_sym_CARET] = ACTIONS(6967),
    [anon_sym_or] = ACTIONS(6967),
    [anon_sym_DOT] = ACTIONS(6967),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6965),
    [anon_sym_new] = ACTIONS(6967),
    [anon_sym_return] = ACTIONS(6967),
    [anon_sym_return_BANG] = ACTIONS(6965),
    [anon_sym_yield] = ACTIONS(6967),
    [anon_sym_yield_BANG] = ACTIONS(6965),
    [anon_sym_lazy] = ACTIONS(6967),
    [anon_sym_assert] = ACTIONS(6967),
    [anon_sym_upcast] = ACTIONS(6967),
    [anon_sym_downcast] = ACTIONS(6967),
    [anon_sym_fixed] = ACTIONS(6967),
    [anon_sym_COLON_GT] = ACTIONS(6965),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6965),
    [anon_sym_for] = ACTIONS(6967),
    [anon_sym_while] = ACTIONS(6967),
    [anon_sym_while_BANG] = ACTIONS(6965),
    [anon_sym_if] = ACTIONS(6967),
    [anon_sym_fun] = ACTIONS(6967),
    [anon_sym_DASH_GT] = ACTIONS(6967),
    [anon_sym_try] = ACTIONS(6967),
    [anon_sym_match] = ACTIONS(6967),
    [anon_sym_match_BANG] = ACTIONS(6965),
    [anon_sym_function] = ACTIONS(6967),
    [anon_sym_DOT_LBRACK] = ACTIONS(6965),
    [anon_sym_use] = ACTIONS(6967),
    [anon_sym_use_BANG] = ACTIONS(6965),
    [anon_sym_do_BANG] = ACTIONS(6965),
    [anon_sym_DOT_DOT] = ACTIONS(6965),
    [anon_sym_LPAREN2] = ACTIONS(6967),
    [anon_sym_PIPE_PIPE] = ACTIONS(6967),
    [anon_sym_AMP_AMP] = ACTIONS(6967),
    [aux_sym_char_token1] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6967),
    [anon_sym_DQUOTE] = ACTIONS(6967),
    [anon_sym_AT_DQUOTE] = ACTIONS(6965),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6965),
    [sym_bool] = ACTIONS(6967),
    [sym_unit] = ACTIONS(6965),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6967),
    [sym_op_identifier] = ACTIONS(6967),
    [anon_sym_PLUS] = ACTIONS(6967),
    [anon_sym_DASH] = ACTIONS(6967),
    [anon_sym_PLUS_DOT] = ACTIONS(6967),
    [anon_sym_DASH_DOT] = ACTIONS(6967),
    [anon_sym_PERCENT] = ACTIONS(6967),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6967),
    [anon_sym_TILDE] = ACTIONS(6965),
    [aux_sym_prefix_op_token1] = ACTIONS(6967),
    [aux_sym_infix_op_token1] = ACTIONS(6965),
    [aux_sym_infix_op_token2] = ACTIONS(6967),
    [aux_sym_infix_op_token3] = ACTIONS(6967),
    [anon_sym_BANG_EQ] = ACTIONS(6967),
    [anon_sym_COLON_EQ] = ACTIONS(6965),
    [anon_sym_DOLLAR] = ACTIONS(6967),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6967),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6967),
    [sym_int] = ACTIONS(6967),
    [sym_xint] = ACTIONS(6965),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6965),
    [sym__newline] = ACTIONS(6965),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6965),
    [sym__tyapp_open] = ACTIONS(6965),
  },
  [STATE(5709)] = {
    [sym_block_comment] = STATE(5709),
    [sym_compiler_directive_decl] = STATE(5709),
    [sym_fsi_directive_decl] = STATE(5709),
    [sym_preproc_line] = STATE(5709),
    [sym_identifier] = ACTIONS(6971),
    [anon_sym_null] = ACTIONS(6971),
    [anon_sym_EQ] = ACTIONS(6971),
    [anon_sym_begin] = ACTIONS(6971),
    [anon_sym_COLON] = ACTIONS(6971),
    [anon_sym_do] = ACTIONS(6971),
    [anon_sym_let] = ACTIONS(6971),
    [anon_sym_let_BANG] = ACTIONS(6969),
    [anon_sym_LPAREN] = ACTIONS(6971),
    [anon_sym_COMMA] = ACTIONS(6969),
    [anon_sym_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_QMARK] = ACTIONS(6971),
    [anon_sym_COLON_COLON] = ACTIONS(6969),
    [anon_sym_AMP] = ACTIONS(6971),
    [anon_sym_LBRACK] = ACTIONS(6971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6969),
    [anon_sym_LBRACE] = ACTIONS(6971),
    [anon_sym_LT_AT] = ACTIONS(6971),
    [anon_sym_LT_AT_AT] = ACTIONS(6971),
    [anon_sym_CARET] = ACTIONS(6971),
    [anon_sym_or] = ACTIONS(6971),
    [anon_sym_DOT] = ACTIONS(6971),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6969),
    [anon_sym_new] = ACTIONS(6971),
    [anon_sym_return] = ACTIONS(6971),
    [anon_sym_return_BANG] = ACTIONS(6969),
    [anon_sym_yield] = ACTIONS(6971),
    [anon_sym_yield_BANG] = ACTIONS(6969),
    [anon_sym_lazy] = ACTIONS(6971),
    [anon_sym_assert] = ACTIONS(6971),
    [anon_sym_upcast] = ACTIONS(6971),
    [anon_sym_downcast] = ACTIONS(6971),
    [anon_sym_fixed] = ACTIONS(6971),
    [anon_sym_COLON_GT] = ACTIONS(6969),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6969),
    [anon_sym_for] = ACTIONS(6971),
    [anon_sym_while] = ACTIONS(6971),
    [anon_sym_while_BANG] = ACTIONS(6969),
    [anon_sym_if] = ACTIONS(6971),
    [anon_sym_fun] = ACTIONS(6971),
    [anon_sym_DASH_GT] = ACTIONS(6971),
    [anon_sym_try] = ACTIONS(6971),
    [anon_sym_match] = ACTIONS(6971),
    [anon_sym_match_BANG] = ACTIONS(6969),
    [anon_sym_function] = ACTIONS(6971),
    [anon_sym_DOT_LBRACK] = ACTIONS(6969),
    [anon_sym_use] = ACTIONS(6971),
    [anon_sym_use_BANG] = ACTIONS(6969),
    [anon_sym_do_BANG] = ACTIONS(6969),
    [anon_sym_DOT_DOT] = ACTIONS(6969),
    [anon_sym_LPAREN2] = ACTIONS(6971),
    [anon_sym_PIPE_PIPE] = ACTIONS(6971),
    [anon_sym_AMP_AMP] = ACTIONS(6971),
    [aux_sym_char_token1] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6971),
    [anon_sym_DQUOTE] = ACTIONS(6971),
    [anon_sym_AT_DQUOTE] = ACTIONS(6969),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6969),
    [sym_bool] = ACTIONS(6971),
    [sym_unit] = ACTIONS(6969),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6971),
    [sym_op_identifier] = ACTIONS(6971),
    [anon_sym_PLUS] = ACTIONS(6971),
    [anon_sym_DASH] = ACTIONS(6971),
    [anon_sym_PLUS_DOT] = ACTIONS(6971),
    [anon_sym_DASH_DOT] = ACTIONS(6971),
    [anon_sym_PERCENT] = ACTIONS(6971),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6971),
    [anon_sym_TILDE] = ACTIONS(6969),
    [aux_sym_prefix_op_token1] = ACTIONS(6971),
    [aux_sym_infix_op_token1] = ACTIONS(6969),
    [aux_sym_infix_op_token2] = ACTIONS(6971),
    [aux_sym_infix_op_token3] = ACTIONS(6971),
    [anon_sym_BANG_EQ] = ACTIONS(6971),
    [anon_sym_COLON_EQ] = ACTIONS(6969),
    [anon_sym_DOLLAR] = ACTIONS(6971),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6971),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6971),
    [sym_int] = ACTIONS(6971),
    [sym_xint] = ACTIONS(6969),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6969),
    [sym__newline] = ACTIONS(6969),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6969),
    [sym__tyapp_open] = ACTIONS(6969),
  },
  [STATE(5710)] = {
    [sym_block_comment] = STATE(5710),
    [sym_compiler_directive_decl] = STATE(5710),
    [sym_fsi_directive_decl] = STATE(5710),
    [sym_preproc_line] = STATE(5710),
    [ts_builtin_sym_end] = ACTIONS(6310),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_namespace] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5711)] = {
    [sym_block_comment] = STATE(5711),
    [sym_compiler_directive_decl] = STATE(5711),
    [sym_fsi_directive_decl] = STATE(5711),
    [sym_preproc_line] = STATE(5711),
    [sym_identifier] = ACTIONS(6768),
    [anon_sym_null] = ACTIONS(6768),
    [anon_sym_EQ] = ACTIONS(6768),
    [anon_sym_begin] = ACTIONS(6768),
    [anon_sym_COLON] = ACTIONS(6768),
    [anon_sym_do] = ACTIONS(6768),
    [anon_sym_let] = ACTIONS(6768),
    [anon_sym_let_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN] = ACTIONS(6768),
    [anon_sym_COMMA] = ACTIONS(6766),
    [anon_sym_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_QMARK] = ACTIONS(6768),
    [anon_sym_COLON_COLON] = ACTIONS(6766),
    [anon_sym_AMP] = ACTIONS(6768),
    [anon_sym_LBRACK] = ACTIONS(6768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6766),
    [anon_sym_LBRACE] = ACTIONS(6768),
    [anon_sym_LT_AT] = ACTIONS(6768),
    [anon_sym_LT_AT_AT] = ACTIONS(6768),
    [anon_sym_CARET] = ACTIONS(6768),
    [anon_sym_or] = ACTIONS(6768),
    [anon_sym_DOT] = ACTIONS(6768),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6766),
    [anon_sym_new] = ACTIONS(6768),
    [anon_sym_return] = ACTIONS(6768),
    [anon_sym_return_BANG] = ACTIONS(6766),
    [anon_sym_yield] = ACTIONS(6768),
    [anon_sym_yield_BANG] = ACTIONS(6766),
    [anon_sym_lazy] = ACTIONS(6768),
    [anon_sym_assert] = ACTIONS(6768),
    [anon_sym_upcast] = ACTIONS(6768),
    [anon_sym_downcast] = ACTIONS(6768),
    [anon_sym_fixed] = ACTIONS(6768),
    [anon_sym_COLON_GT] = ACTIONS(6766),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6766),
    [anon_sym_for] = ACTIONS(6768),
    [anon_sym_done] = ACTIONS(9835),
    [anon_sym_while] = ACTIONS(6768),
    [anon_sym_while_BANG] = ACTIONS(6766),
    [anon_sym_if] = ACTIONS(6768),
    [anon_sym_fun] = ACTIONS(6768),
    [anon_sym_try] = ACTIONS(6768),
    [anon_sym_match] = ACTIONS(6768),
    [anon_sym_match_BANG] = ACTIONS(6766),
    [anon_sym_function] = ACTIONS(6768),
    [anon_sym_DOT_LBRACK] = ACTIONS(6766),
    [anon_sym_use] = ACTIONS(6768),
    [anon_sym_use_BANG] = ACTIONS(6766),
    [anon_sym_do_BANG] = ACTIONS(6766),
    [anon_sym_LPAREN2] = ACTIONS(6768),
    [anon_sym_PIPE_PIPE] = ACTIONS(6768),
    [anon_sym_AMP_AMP] = ACTIONS(6768),
    [aux_sym_char_token1] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6768),
    [anon_sym_DQUOTE] = ACTIONS(6768),
    [anon_sym_AT_DQUOTE] = ACTIONS(6766),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6766),
    [sym_bool] = ACTIONS(6768),
    [sym_unit] = ACTIONS(6766),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6768),
    [sym_op_identifier] = ACTIONS(6768),
    [anon_sym_PLUS] = ACTIONS(6768),
    [anon_sym_DASH] = ACTIONS(6768),
    [anon_sym_PLUS_DOT] = ACTIONS(6768),
    [anon_sym_DASH_DOT] = ACTIONS(6768),
    [anon_sym_PERCENT] = ACTIONS(6768),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6768),
    [anon_sym_TILDE] = ACTIONS(6766),
    [aux_sym_prefix_op_token1] = ACTIONS(6768),
    [aux_sym_infix_op_token1] = ACTIONS(6766),
    [aux_sym_infix_op_token2] = ACTIONS(6768),
    [aux_sym_infix_op_token3] = ACTIONS(6768),
    [anon_sym_BANG_EQ] = ACTIONS(6768),
    [anon_sym_COLON_EQ] = ACTIONS(6766),
    [anon_sym_DOLLAR] = ACTIONS(6768),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6768),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6768),
    [sym_int] = ACTIONS(6768),
    [sym_xint] = ACTIONS(6766),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6766),
    [sym__newline] = ACTIONS(6766),
    [sym__untyped_quoted_close] = ACTIONS(6766),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6766),
    [sym__tyapp_open] = ACTIONS(6766),
  },
  [STATE(5712)] = {
    [sym_block_comment] = STATE(5712),
    [sym_compiler_directive_decl] = STATE(5712),
    [sym_fsi_directive_decl] = STATE(5712),
    [sym_preproc_line] = STATE(5712),
    [sym_identifier] = ACTIONS(6975),
    [anon_sym_null] = ACTIONS(6975),
    [anon_sym_EQ] = ACTIONS(6975),
    [anon_sym_begin] = ACTIONS(6975),
    [anon_sym_COLON] = ACTIONS(6975),
    [anon_sym_do] = ACTIONS(6975),
    [anon_sym_let] = ACTIONS(6975),
    [anon_sym_let_BANG] = ACTIONS(6973),
    [anon_sym_LPAREN] = ACTIONS(6975),
    [anon_sym_COMMA] = ACTIONS(6973),
    [anon_sym_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_QMARK] = ACTIONS(6975),
    [anon_sym_COLON_COLON] = ACTIONS(6973),
    [anon_sym_AMP] = ACTIONS(6975),
    [anon_sym_LBRACK] = ACTIONS(6975),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6973),
    [anon_sym_LBRACE] = ACTIONS(6975),
    [anon_sym_LT_AT] = ACTIONS(6975),
    [anon_sym_LT_AT_AT] = ACTIONS(6975),
    [anon_sym_CARET] = ACTIONS(6975),
    [anon_sym_or] = ACTIONS(6975),
    [anon_sym_DOT] = ACTIONS(6975),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6973),
    [anon_sym_new] = ACTIONS(6975),
    [anon_sym_return] = ACTIONS(6975),
    [anon_sym_return_BANG] = ACTIONS(6973),
    [anon_sym_yield] = ACTIONS(6975),
    [anon_sym_yield_BANG] = ACTIONS(6973),
    [anon_sym_lazy] = ACTIONS(6975),
    [anon_sym_assert] = ACTIONS(6975),
    [anon_sym_upcast] = ACTIONS(6975),
    [anon_sym_downcast] = ACTIONS(6975),
    [anon_sym_fixed] = ACTIONS(6975),
    [anon_sym_COLON_GT] = ACTIONS(6973),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6973),
    [anon_sym_for] = ACTIONS(6975),
    [anon_sym_while] = ACTIONS(6975),
    [anon_sym_while_BANG] = ACTIONS(6973),
    [anon_sym_if] = ACTIONS(6975),
    [anon_sym_fun] = ACTIONS(6975),
    [anon_sym_DASH_GT] = ACTIONS(6975),
    [anon_sym_try] = ACTIONS(6975),
    [anon_sym_match] = ACTIONS(6975),
    [anon_sym_match_BANG] = ACTIONS(6973),
    [anon_sym_function] = ACTIONS(6975),
    [anon_sym_DOT_LBRACK] = ACTIONS(6973),
    [anon_sym_use] = ACTIONS(6975),
    [anon_sym_use_BANG] = ACTIONS(6973),
    [anon_sym_do_BANG] = ACTIONS(6973),
    [anon_sym_DOT_DOT] = ACTIONS(6973),
    [anon_sym_LPAREN2] = ACTIONS(6975),
    [anon_sym_PIPE_PIPE] = ACTIONS(6975),
    [anon_sym_AMP_AMP] = ACTIONS(6975),
    [aux_sym_char_token1] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6975),
    [anon_sym_DQUOTE] = ACTIONS(6975),
    [anon_sym_AT_DQUOTE] = ACTIONS(6973),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6973),
    [sym_bool] = ACTIONS(6975),
    [sym_unit] = ACTIONS(6973),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6975),
    [sym_op_identifier] = ACTIONS(6975),
    [anon_sym_PLUS] = ACTIONS(6975),
    [anon_sym_DASH] = ACTIONS(6975),
    [anon_sym_PLUS_DOT] = ACTIONS(6975),
    [anon_sym_DASH_DOT] = ACTIONS(6975),
    [anon_sym_PERCENT] = ACTIONS(6975),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6975),
    [anon_sym_TILDE] = ACTIONS(6973),
    [aux_sym_prefix_op_token1] = ACTIONS(6975),
    [aux_sym_infix_op_token1] = ACTIONS(6973),
    [aux_sym_infix_op_token2] = ACTIONS(6975),
    [aux_sym_infix_op_token3] = ACTIONS(6975),
    [anon_sym_BANG_EQ] = ACTIONS(6975),
    [anon_sym_COLON_EQ] = ACTIONS(6973),
    [anon_sym_DOLLAR] = ACTIONS(6975),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6975),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6975),
    [sym_int] = ACTIONS(6975),
    [sym_xint] = ACTIONS(6973),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6973),
    [sym__newline] = ACTIONS(6973),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6973),
    [sym__tyapp_open] = ACTIONS(6973),
  },
  [STATE(5713)] = {
    [sym_block_comment] = STATE(5713),
    [sym_compiler_directive_decl] = STATE(5713),
    [sym_fsi_directive_decl] = STATE(5713),
    [sym_preproc_line] = STATE(5713),
    [sym_identifier] = ACTIONS(6979),
    [anon_sym_null] = ACTIONS(6979),
    [anon_sym_EQ] = ACTIONS(6979),
    [anon_sym_begin] = ACTIONS(6979),
    [anon_sym_COLON] = ACTIONS(6979),
    [anon_sym_do] = ACTIONS(6979),
    [anon_sym_let] = ACTIONS(6979),
    [anon_sym_let_BANG] = ACTIONS(6977),
    [anon_sym_LPAREN] = ACTIONS(6979),
    [anon_sym_COMMA] = ACTIONS(6977),
    [anon_sym_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_QMARK] = ACTIONS(6979),
    [anon_sym_COLON_COLON] = ACTIONS(6977),
    [anon_sym_AMP] = ACTIONS(6979),
    [anon_sym_LBRACK] = ACTIONS(6979),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6977),
    [anon_sym_LBRACE] = ACTIONS(6979),
    [anon_sym_LT_AT] = ACTIONS(6979),
    [anon_sym_LT_AT_AT] = ACTIONS(6979),
    [anon_sym_CARET] = ACTIONS(6979),
    [anon_sym_or] = ACTIONS(6979),
    [anon_sym_DOT] = ACTIONS(6979),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6977),
    [anon_sym_new] = ACTIONS(6979),
    [anon_sym_return] = ACTIONS(6979),
    [anon_sym_return_BANG] = ACTIONS(6977),
    [anon_sym_yield] = ACTIONS(6979),
    [anon_sym_yield_BANG] = ACTIONS(6977),
    [anon_sym_lazy] = ACTIONS(6979),
    [anon_sym_assert] = ACTIONS(6979),
    [anon_sym_upcast] = ACTIONS(6979),
    [anon_sym_downcast] = ACTIONS(6979),
    [anon_sym_fixed] = ACTIONS(6979),
    [anon_sym_COLON_GT] = ACTIONS(6977),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6977),
    [anon_sym_for] = ACTIONS(6979),
    [anon_sym_while] = ACTIONS(6979),
    [anon_sym_while_BANG] = ACTIONS(6977),
    [anon_sym_if] = ACTIONS(6979),
    [anon_sym_fun] = ACTIONS(6979),
    [anon_sym_DASH_GT] = ACTIONS(6979),
    [anon_sym_try] = ACTIONS(6979),
    [anon_sym_match] = ACTIONS(6979),
    [anon_sym_match_BANG] = ACTIONS(6977),
    [anon_sym_function] = ACTIONS(6979),
    [anon_sym_DOT_LBRACK] = ACTIONS(6977),
    [anon_sym_use] = ACTIONS(6979),
    [anon_sym_use_BANG] = ACTIONS(6977),
    [anon_sym_do_BANG] = ACTIONS(6977),
    [anon_sym_DOT_DOT] = ACTIONS(6977),
    [anon_sym_LPAREN2] = ACTIONS(6979),
    [anon_sym_PIPE_PIPE] = ACTIONS(6979),
    [anon_sym_AMP_AMP] = ACTIONS(6979),
    [aux_sym_char_token1] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6979),
    [anon_sym_DQUOTE] = ACTIONS(6979),
    [anon_sym_AT_DQUOTE] = ACTIONS(6977),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6977),
    [sym_bool] = ACTIONS(6979),
    [sym_unit] = ACTIONS(6977),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6979),
    [sym_op_identifier] = ACTIONS(6979),
    [anon_sym_PLUS] = ACTIONS(6979),
    [anon_sym_DASH] = ACTIONS(6979),
    [anon_sym_PLUS_DOT] = ACTIONS(6979),
    [anon_sym_DASH_DOT] = ACTIONS(6979),
    [anon_sym_PERCENT] = ACTIONS(6979),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6979),
    [anon_sym_TILDE] = ACTIONS(6977),
    [aux_sym_prefix_op_token1] = ACTIONS(6979),
    [aux_sym_infix_op_token1] = ACTIONS(6977),
    [aux_sym_infix_op_token2] = ACTIONS(6979),
    [aux_sym_infix_op_token3] = ACTIONS(6979),
    [anon_sym_BANG_EQ] = ACTIONS(6979),
    [anon_sym_COLON_EQ] = ACTIONS(6977),
    [anon_sym_DOLLAR] = ACTIONS(6979),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6979),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6979),
    [sym_int] = ACTIONS(6979),
    [sym_xint] = ACTIONS(6977),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6977),
    [sym__newline] = ACTIONS(6977),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6977),
    [sym__tyapp_open] = ACTIONS(6977),
  },
  [STATE(5714)] = {
    [sym_block_comment] = STATE(5714),
    [sym_compiler_directive_decl] = STATE(5714),
    [sym_fsi_directive_decl] = STATE(5714),
    [sym_preproc_line] = STATE(5714),
    [sym_identifier] = ACTIONS(6983),
    [anon_sym_null] = ACTIONS(6983),
    [anon_sym_EQ] = ACTIONS(6983),
    [anon_sym_begin] = ACTIONS(6983),
    [anon_sym_COLON] = ACTIONS(6983),
    [anon_sym_do] = ACTIONS(6983),
    [anon_sym_let] = ACTIONS(6983),
    [anon_sym_let_BANG] = ACTIONS(6981),
    [anon_sym_LPAREN] = ACTIONS(6983),
    [anon_sym_COMMA] = ACTIONS(6981),
    [anon_sym_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_QMARK] = ACTIONS(6983),
    [anon_sym_COLON_COLON] = ACTIONS(6981),
    [anon_sym_AMP] = ACTIONS(6983),
    [anon_sym_LBRACK] = ACTIONS(6983),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6981),
    [anon_sym_LBRACE] = ACTIONS(6983),
    [anon_sym_LT_AT] = ACTIONS(6983),
    [anon_sym_LT_AT_AT] = ACTIONS(6983),
    [anon_sym_CARET] = ACTIONS(6983),
    [anon_sym_or] = ACTIONS(6983),
    [anon_sym_DOT] = ACTIONS(6983),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6981),
    [anon_sym_new] = ACTIONS(6983),
    [anon_sym_return] = ACTIONS(6983),
    [anon_sym_return_BANG] = ACTIONS(6981),
    [anon_sym_yield] = ACTIONS(6983),
    [anon_sym_yield_BANG] = ACTIONS(6981),
    [anon_sym_lazy] = ACTIONS(6983),
    [anon_sym_assert] = ACTIONS(6983),
    [anon_sym_upcast] = ACTIONS(6983),
    [anon_sym_downcast] = ACTIONS(6983),
    [anon_sym_fixed] = ACTIONS(6983),
    [anon_sym_COLON_GT] = ACTIONS(6981),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6981),
    [anon_sym_for] = ACTIONS(6983),
    [anon_sym_while] = ACTIONS(6983),
    [anon_sym_while_BANG] = ACTIONS(6981),
    [anon_sym_if] = ACTIONS(6983),
    [anon_sym_fun] = ACTIONS(6983),
    [anon_sym_DASH_GT] = ACTIONS(6983),
    [anon_sym_try] = ACTIONS(6983),
    [anon_sym_match] = ACTIONS(6983),
    [anon_sym_match_BANG] = ACTIONS(6981),
    [anon_sym_function] = ACTIONS(6983),
    [anon_sym_DOT_LBRACK] = ACTIONS(6981),
    [anon_sym_use] = ACTIONS(6983),
    [anon_sym_use_BANG] = ACTIONS(6981),
    [anon_sym_do_BANG] = ACTIONS(6981),
    [anon_sym_DOT_DOT] = ACTIONS(6981),
    [anon_sym_LPAREN2] = ACTIONS(6983),
    [anon_sym_PIPE_PIPE] = ACTIONS(6983),
    [anon_sym_AMP_AMP] = ACTIONS(6983),
    [aux_sym_char_token1] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6983),
    [anon_sym_DQUOTE] = ACTIONS(6983),
    [anon_sym_AT_DQUOTE] = ACTIONS(6981),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6981),
    [sym_bool] = ACTIONS(6983),
    [sym_unit] = ACTIONS(6981),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6983),
    [sym_op_identifier] = ACTIONS(6983),
    [anon_sym_PLUS] = ACTIONS(6983),
    [anon_sym_DASH] = ACTIONS(6983),
    [anon_sym_PLUS_DOT] = ACTIONS(6983),
    [anon_sym_DASH_DOT] = ACTIONS(6983),
    [anon_sym_PERCENT] = ACTIONS(6983),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6983),
    [anon_sym_TILDE] = ACTIONS(6981),
    [aux_sym_prefix_op_token1] = ACTIONS(6983),
    [aux_sym_infix_op_token1] = ACTIONS(6981),
    [aux_sym_infix_op_token2] = ACTIONS(6983),
    [aux_sym_infix_op_token3] = ACTIONS(6983),
    [anon_sym_BANG_EQ] = ACTIONS(6983),
    [anon_sym_COLON_EQ] = ACTIONS(6981),
    [anon_sym_DOLLAR] = ACTIONS(6983),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6983),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6983),
    [sym_int] = ACTIONS(6983),
    [sym_xint] = ACTIONS(6981),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6981),
    [sym__newline] = ACTIONS(6981),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6981),
    [sym__tyapp_open] = ACTIONS(6981),
  },
  [STATE(5715)] = {
    [sym_block_comment] = STATE(5715),
    [sym_compiler_directive_decl] = STATE(5715),
    [sym_fsi_directive_decl] = STATE(5715),
    [sym_preproc_line] = STATE(5715),
    [sym_identifier] = ACTIONS(6987),
    [anon_sym_null] = ACTIONS(6987),
    [anon_sym_EQ] = ACTIONS(6987),
    [anon_sym_begin] = ACTIONS(6987),
    [anon_sym_COLON] = ACTIONS(6987),
    [anon_sym_do] = ACTIONS(6987),
    [anon_sym_let] = ACTIONS(6987),
    [anon_sym_let_BANG] = ACTIONS(6985),
    [anon_sym_LPAREN] = ACTIONS(6987),
    [anon_sym_COMMA] = ACTIONS(6985),
    [anon_sym_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_QMARK] = ACTIONS(6987),
    [anon_sym_COLON_COLON] = ACTIONS(6985),
    [anon_sym_AMP] = ACTIONS(6987),
    [anon_sym_LBRACK] = ACTIONS(6987),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6985),
    [anon_sym_LBRACE] = ACTIONS(6987),
    [anon_sym_LT_AT] = ACTIONS(6987),
    [anon_sym_LT_AT_AT] = ACTIONS(6987),
    [anon_sym_CARET] = ACTIONS(6987),
    [anon_sym_or] = ACTIONS(6987),
    [anon_sym_DOT] = ACTIONS(6987),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6985),
    [anon_sym_new] = ACTIONS(6987),
    [anon_sym_return] = ACTIONS(6987),
    [anon_sym_return_BANG] = ACTIONS(6985),
    [anon_sym_yield] = ACTIONS(6987),
    [anon_sym_yield_BANG] = ACTIONS(6985),
    [anon_sym_lazy] = ACTIONS(6987),
    [anon_sym_assert] = ACTIONS(6987),
    [anon_sym_upcast] = ACTIONS(6987),
    [anon_sym_downcast] = ACTIONS(6987),
    [anon_sym_fixed] = ACTIONS(6987),
    [anon_sym_COLON_GT] = ACTIONS(6985),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6985),
    [anon_sym_for] = ACTIONS(6987),
    [anon_sym_while] = ACTIONS(6987),
    [anon_sym_while_BANG] = ACTIONS(6985),
    [anon_sym_if] = ACTIONS(6987),
    [anon_sym_fun] = ACTIONS(6987),
    [anon_sym_DASH_GT] = ACTIONS(6987),
    [anon_sym_try] = ACTIONS(6987),
    [anon_sym_match] = ACTIONS(6987),
    [anon_sym_match_BANG] = ACTIONS(6985),
    [anon_sym_function] = ACTIONS(6987),
    [anon_sym_DOT_LBRACK] = ACTIONS(6985),
    [anon_sym_use] = ACTIONS(6987),
    [anon_sym_use_BANG] = ACTIONS(6985),
    [anon_sym_do_BANG] = ACTIONS(6985),
    [anon_sym_DOT_DOT] = ACTIONS(6985),
    [anon_sym_LPAREN2] = ACTIONS(6987),
    [anon_sym_PIPE_PIPE] = ACTIONS(6987),
    [anon_sym_AMP_AMP] = ACTIONS(6987),
    [aux_sym_char_token1] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6987),
    [anon_sym_DQUOTE] = ACTIONS(6987),
    [anon_sym_AT_DQUOTE] = ACTIONS(6985),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6985),
    [sym_bool] = ACTIONS(6987),
    [sym_unit] = ACTIONS(6985),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6987),
    [sym_op_identifier] = ACTIONS(6987),
    [anon_sym_PLUS] = ACTIONS(6987),
    [anon_sym_DASH] = ACTIONS(6987),
    [anon_sym_PLUS_DOT] = ACTIONS(6987),
    [anon_sym_DASH_DOT] = ACTIONS(6987),
    [anon_sym_PERCENT] = ACTIONS(6987),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6987),
    [anon_sym_TILDE] = ACTIONS(6985),
    [aux_sym_prefix_op_token1] = ACTIONS(6987),
    [aux_sym_infix_op_token1] = ACTIONS(6985),
    [aux_sym_infix_op_token2] = ACTIONS(6987),
    [aux_sym_infix_op_token3] = ACTIONS(6987),
    [anon_sym_BANG_EQ] = ACTIONS(6987),
    [anon_sym_COLON_EQ] = ACTIONS(6985),
    [anon_sym_DOLLAR] = ACTIONS(6987),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6987),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6987),
    [sym_int] = ACTIONS(6987),
    [sym_xint] = ACTIONS(6985),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6985),
    [sym__newline] = ACTIONS(6985),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6985),
    [sym__tyapp_open] = ACTIONS(6985),
  },
  [STATE(5716)] = {
    [sym_block_comment] = STATE(5716),
    [sym_compiler_directive_decl] = STATE(5716),
    [sym_fsi_directive_decl] = STATE(5716),
    [sym_preproc_line] = STATE(5716),
    [sym_identifier] = ACTIONS(6991),
    [anon_sym_null] = ACTIONS(6991),
    [anon_sym_EQ] = ACTIONS(6991),
    [anon_sym_begin] = ACTIONS(6991),
    [anon_sym_COLON] = ACTIONS(6991),
    [anon_sym_do] = ACTIONS(6991),
    [anon_sym_let] = ACTIONS(6991),
    [anon_sym_let_BANG] = ACTIONS(6989),
    [anon_sym_LPAREN] = ACTIONS(6991),
    [anon_sym_COMMA] = ACTIONS(6989),
    [anon_sym_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_QMARK] = ACTIONS(6991),
    [anon_sym_COLON_COLON] = ACTIONS(6989),
    [anon_sym_AMP] = ACTIONS(6991),
    [anon_sym_LBRACK] = ACTIONS(6991),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6989),
    [anon_sym_LBRACE] = ACTIONS(6991),
    [anon_sym_LT_AT] = ACTIONS(6991),
    [anon_sym_LT_AT_AT] = ACTIONS(6991),
    [anon_sym_CARET] = ACTIONS(6991),
    [anon_sym_or] = ACTIONS(6991),
    [anon_sym_DOT] = ACTIONS(6991),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6989),
    [anon_sym_new] = ACTIONS(6991),
    [anon_sym_return] = ACTIONS(6991),
    [anon_sym_return_BANG] = ACTIONS(6989),
    [anon_sym_yield] = ACTIONS(6991),
    [anon_sym_yield_BANG] = ACTIONS(6989),
    [anon_sym_lazy] = ACTIONS(6991),
    [anon_sym_assert] = ACTIONS(6991),
    [anon_sym_upcast] = ACTIONS(6991),
    [anon_sym_downcast] = ACTIONS(6991),
    [anon_sym_fixed] = ACTIONS(6991),
    [anon_sym_COLON_GT] = ACTIONS(6989),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6989),
    [anon_sym_for] = ACTIONS(6991),
    [anon_sym_while] = ACTIONS(6991),
    [anon_sym_while_BANG] = ACTIONS(6989),
    [anon_sym_if] = ACTIONS(6991),
    [anon_sym_fun] = ACTIONS(6991),
    [anon_sym_DASH_GT] = ACTIONS(6991),
    [anon_sym_try] = ACTIONS(6991),
    [anon_sym_match] = ACTIONS(6991),
    [anon_sym_match_BANG] = ACTIONS(6989),
    [anon_sym_function] = ACTIONS(6991),
    [anon_sym_DOT_LBRACK] = ACTIONS(6989),
    [anon_sym_use] = ACTIONS(6991),
    [anon_sym_use_BANG] = ACTIONS(6989),
    [anon_sym_do_BANG] = ACTIONS(6989),
    [anon_sym_DOT_DOT] = ACTIONS(6989),
    [anon_sym_LPAREN2] = ACTIONS(6991),
    [anon_sym_PIPE_PIPE] = ACTIONS(6991),
    [anon_sym_AMP_AMP] = ACTIONS(6991),
    [aux_sym_char_token1] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6991),
    [anon_sym_DQUOTE] = ACTIONS(6991),
    [anon_sym_AT_DQUOTE] = ACTIONS(6989),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6989),
    [sym_bool] = ACTIONS(6991),
    [sym_unit] = ACTIONS(6989),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6991),
    [sym_op_identifier] = ACTIONS(6991),
    [anon_sym_PLUS] = ACTIONS(6991),
    [anon_sym_DASH] = ACTIONS(6991),
    [anon_sym_PLUS_DOT] = ACTIONS(6991),
    [anon_sym_DASH_DOT] = ACTIONS(6991),
    [anon_sym_PERCENT] = ACTIONS(6991),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6991),
    [anon_sym_TILDE] = ACTIONS(6989),
    [aux_sym_prefix_op_token1] = ACTIONS(6991),
    [aux_sym_infix_op_token1] = ACTIONS(6989),
    [aux_sym_infix_op_token2] = ACTIONS(6991),
    [aux_sym_infix_op_token3] = ACTIONS(6991),
    [anon_sym_BANG_EQ] = ACTIONS(6991),
    [anon_sym_COLON_EQ] = ACTIONS(6989),
    [anon_sym_DOLLAR] = ACTIONS(6991),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6991),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6991),
    [sym_int] = ACTIONS(6991),
    [sym_xint] = ACTIONS(6989),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6989),
    [sym__newline] = ACTIONS(6989),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6989),
    [sym__tyapp_open] = ACTIONS(6989),
  },
  [STATE(5717)] = {
    [sym_block_comment] = STATE(5717),
    [sym_compiler_directive_decl] = STATE(5717),
    [sym_fsi_directive_decl] = STATE(5717),
    [sym_preproc_line] = STATE(5717),
    [sym_identifier] = ACTIONS(6995),
    [anon_sym_null] = ACTIONS(6995),
    [anon_sym_EQ] = ACTIONS(6995),
    [anon_sym_begin] = ACTIONS(6995),
    [anon_sym_COLON] = ACTIONS(6995),
    [anon_sym_do] = ACTIONS(6995),
    [anon_sym_let] = ACTIONS(6995),
    [anon_sym_let_BANG] = ACTIONS(6993),
    [anon_sym_LPAREN] = ACTIONS(6995),
    [anon_sym_COMMA] = ACTIONS(6993),
    [anon_sym_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_QMARK] = ACTIONS(6995),
    [anon_sym_COLON_COLON] = ACTIONS(6993),
    [anon_sym_AMP] = ACTIONS(6995),
    [anon_sym_LBRACK] = ACTIONS(6995),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6993),
    [anon_sym_LBRACE] = ACTIONS(6995),
    [anon_sym_LT_AT] = ACTIONS(6995),
    [anon_sym_LT_AT_AT] = ACTIONS(6995),
    [anon_sym_CARET] = ACTIONS(6995),
    [anon_sym_or] = ACTIONS(6995),
    [anon_sym_DOT] = ACTIONS(6995),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6993),
    [anon_sym_new] = ACTIONS(6995),
    [anon_sym_return] = ACTIONS(6995),
    [anon_sym_return_BANG] = ACTIONS(6993),
    [anon_sym_yield] = ACTIONS(6995),
    [anon_sym_yield_BANG] = ACTIONS(6993),
    [anon_sym_lazy] = ACTIONS(6995),
    [anon_sym_assert] = ACTIONS(6995),
    [anon_sym_upcast] = ACTIONS(6995),
    [anon_sym_downcast] = ACTIONS(6995),
    [anon_sym_fixed] = ACTIONS(6995),
    [anon_sym_COLON_GT] = ACTIONS(6993),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6993),
    [anon_sym_for] = ACTIONS(6995),
    [anon_sym_while] = ACTIONS(6995),
    [anon_sym_while_BANG] = ACTIONS(6993),
    [anon_sym_if] = ACTIONS(6995),
    [anon_sym_fun] = ACTIONS(6995),
    [anon_sym_DASH_GT] = ACTIONS(6995),
    [anon_sym_try] = ACTIONS(6995),
    [anon_sym_match] = ACTIONS(6995),
    [anon_sym_match_BANG] = ACTIONS(6993),
    [anon_sym_function] = ACTIONS(6995),
    [anon_sym_DOT_LBRACK] = ACTIONS(6993),
    [anon_sym_use] = ACTIONS(6995),
    [anon_sym_use_BANG] = ACTIONS(6993),
    [anon_sym_do_BANG] = ACTIONS(6993),
    [anon_sym_DOT_DOT] = ACTIONS(6993),
    [anon_sym_LPAREN2] = ACTIONS(6995),
    [anon_sym_PIPE_PIPE] = ACTIONS(6995),
    [anon_sym_AMP_AMP] = ACTIONS(6995),
    [aux_sym_char_token1] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6995),
    [anon_sym_DQUOTE] = ACTIONS(6995),
    [anon_sym_AT_DQUOTE] = ACTIONS(6993),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6993),
    [sym_bool] = ACTIONS(6995),
    [sym_unit] = ACTIONS(6993),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6995),
    [sym_op_identifier] = ACTIONS(6995),
    [anon_sym_PLUS] = ACTIONS(6995),
    [anon_sym_DASH] = ACTIONS(6995),
    [anon_sym_PLUS_DOT] = ACTIONS(6995),
    [anon_sym_DASH_DOT] = ACTIONS(6995),
    [anon_sym_PERCENT] = ACTIONS(6995),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6995),
    [anon_sym_TILDE] = ACTIONS(6993),
    [aux_sym_prefix_op_token1] = ACTIONS(6995),
    [aux_sym_infix_op_token1] = ACTIONS(6993),
    [aux_sym_infix_op_token2] = ACTIONS(6995),
    [aux_sym_infix_op_token3] = ACTIONS(6995),
    [anon_sym_BANG_EQ] = ACTIONS(6995),
    [anon_sym_COLON_EQ] = ACTIONS(6993),
    [anon_sym_DOLLAR] = ACTIONS(6995),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6995),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6995),
    [sym_int] = ACTIONS(6995),
    [sym_xint] = ACTIONS(6993),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6993),
    [sym__newline] = ACTIONS(6993),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6993),
    [sym__tyapp_open] = ACTIONS(6993),
  },
  [STATE(5718)] = {
    [sym_block_comment] = STATE(5718),
    [sym_compiler_directive_decl] = STATE(5718),
    [sym_fsi_directive_decl] = STATE(5718),
    [sym_preproc_line] = STATE(5718),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_DOT_DOT] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5719)] = {
    [sym_block_comment] = STATE(5719),
    [sym_compiler_directive_decl] = STATE(5719),
    [sym_fsi_directive_decl] = STATE(5719),
    [sym_preproc_line] = STATE(5719),
    [sym_identifier] = ACTIONS(6906),
    [anon_sym_null] = ACTIONS(6906),
    [anon_sym_EQ] = ACTIONS(6906),
    [anon_sym_begin] = ACTIONS(6906),
    [anon_sym_COLON] = ACTIONS(6906),
    [anon_sym_do] = ACTIONS(6906),
    [anon_sym_let] = ACTIONS(6906),
    [anon_sym_let_BANG] = ACTIONS(6904),
    [anon_sym_as] = ACTIONS(6906),
    [anon_sym_LPAREN] = ACTIONS(6906),
    [anon_sym_COMMA] = ACTIONS(6904),
    [anon_sym_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_QMARK] = ACTIONS(6906),
    [anon_sym_COLON_COLON] = ACTIONS(6904),
    [anon_sym_AMP] = ACTIONS(6906),
    [anon_sym_LBRACK] = ACTIONS(6906),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6904),
    [anon_sym_LBRACE] = ACTIONS(6906),
    [anon_sym_LT_AT] = ACTIONS(6906),
    [anon_sym_LT_AT_AT] = ACTIONS(6906),
    [anon_sym_CARET] = ACTIONS(6906),
    [anon_sym_or] = ACTIONS(6906),
    [anon_sym_DOT] = ACTIONS(6906),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6904),
    [anon_sym_new] = ACTIONS(6906),
    [anon_sym_with] = ACTIONS(6906),
    [anon_sym_return] = ACTIONS(6906),
    [anon_sym_return_BANG] = ACTIONS(6904),
    [anon_sym_yield] = ACTIONS(6906),
    [anon_sym_yield_BANG] = ACTIONS(6904),
    [anon_sym_lazy] = ACTIONS(6906),
    [anon_sym_assert] = ACTIONS(6906),
    [anon_sym_upcast] = ACTIONS(6906),
    [anon_sym_downcast] = ACTIONS(6906),
    [anon_sym_fixed] = ACTIONS(6906),
    [anon_sym_COLON_GT] = ACTIONS(6904),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6904),
    [anon_sym_for] = ACTIONS(6906),
    [anon_sym_while] = ACTIONS(6906),
    [anon_sym_while_BANG] = ACTIONS(6904),
    [anon_sym_if] = ACTIONS(6906),
    [anon_sym_fun] = ACTIONS(6906),
    [anon_sym_try] = ACTIONS(6906),
    [anon_sym_match] = ACTIONS(6906),
    [anon_sym_match_BANG] = ACTIONS(6904),
    [anon_sym_function] = ACTIONS(6906),
    [anon_sym_DOT_LBRACK] = ACTIONS(6904),
    [anon_sym_use] = ACTIONS(6906),
    [anon_sym_use_BANG] = ACTIONS(6904),
    [anon_sym_do_BANG] = ACTIONS(6904),
    [anon_sym_LPAREN2] = ACTIONS(6906),
    [anon_sym_PIPE_PIPE] = ACTIONS(6906),
    [anon_sym_AMP_AMP] = ACTIONS(6906),
    [aux_sym_char_token1] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6906),
    [anon_sym_DQUOTE] = ACTIONS(6906),
    [anon_sym_AT_DQUOTE] = ACTIONS(6904),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6904),
    [sym_bool] = ACTIONS(6906),
    [sym_unit] = ACTIONS(6904),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6906),
    [sym_op_identifier] = ACTIONS(6906),
    [anon_sym_PLUS] = ACTIONS(6906),
    [anon_sym_DASH] = ACTIONS(6906),
    [anon_sym_PLUS_DOT] = ACTIONS(6906),
    [anon_sym_DASH_DOT] = ACTIONS(6906),
    [anon_sym_PERCENT] = ACTIONS(6906),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6906),
    [anon_sym_TILDE] = ACTIONS(6904),
    [aux_sym_prefix_op_token1] = ACTIONS(6906),
    [aux_sym_infix_op_token1] = ACTIONS(6904),
    [aux_sym_infix_op_token2] = ACTIONS(6906),
    [aux_sym_infix_op_token3] = ACTIONS(6906),
    [anon_sym_BANG_EQ] = ACTIONS(6906),
    [anon_sym_COLON_EQ] = ACTIONS(6904),
    [anon_sym_DOLLAR] = ACTIONS(6906),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6906),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6906),
    [sym_int] = ACTIONS(6906),
    [sym_xint] = ACTIONS(6904),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6904),
    [sym__newline] = ACTIONS(6904),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6904),
    [sym__tyapp_open] = ACTIONS(6904),
  },
  [STATE(5720)] = {
    [sym_block_comment] = STATE(5720),
    [sym_compiler_directive_decl] = STATE(5720),
    [sym_fsi_directive_decl] = STATE(5720),
    [sym_preproc_line] = STATE(5720),
    [aux_sym_sequential_expression_repeat1] = STATE(5720),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9837),
    [sym__quoted_close] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5721)] = {
    [sym_block_comment] = STATE(5721),
    [sym_compiler_directive_decl] = STATE(5721),
    [sym_fsi_directive_decl] = STATE(5721),
    [sym_preproc_line] = STATE(5721),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [anon_sym_POUNDendif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5722)] = {
    [sym_block_comment] = STATE(5722),
    [sym_compiler_directive_decl] = STATE(5722),
    [sym_fsi_directive_decl] = STATE(5722),
    [sym_preproc_line] = STATE(5722),
    [sym_identifier] = ACTIONS(6913),
    [anon_sym_null] = ACTIONS(6913),
    [anon_sym_EQ] = ACTIONS(6913),
    [anon_sym_begin] = ACTIONS(6913),
    [anon_sym_COLON] = ACTIONS(6913),
    [anon_sym_do] = ACTIONS(6913),
    [anon_sym_let] = ACTIONS(6913),
    [anon_sym_let_BANG] = ACTIONS(6911),
    [anon_sym_as] = ACTIONS(6913),
    [anon_sym_LPAREN] = ACTIONS(6913),
    [anon_sym_COMMA] = ACTIONS(6911),
    [anon_sym_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_QMARK] = ACTIONS(6913),
    [anon_sym_COLON_COLON] = ACTIONS(6911),
    [anon_sym_AMP] = ACTIONS(6913),
    [anon_sym_LBRACK] = ACTIONS(6913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6911),
    [anon_sym_LBRACE] = ACTIONS(6913),
    [anon_sym_LT_AT] = ACTIONS(6913),
    [anon_sym_LT_AT_AT] = ACTIONS(6913),
    [anon_sym_CARET] = ACTIONS(6913),
    [anon_sym_or] = ACTIONS(6913),
    [anon_sym_DOT] = ACTIONS(6913),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6911),
    [anon_sym_new] = ACTIONS(6913),
    [anon_sym_with] = ACTIONS(6913),
    [anon_sym_return] = ACTIONS(6913),
    [anon_sym_return_BANG] = ACTIONS(6911),
    [anon_sym_yield] = ACTIONS(6913),
    [anon_sym_yield_BANG] = ACTIONS(6911),
    [anon_sym_lazy] = ACTIONS(6913),
    [anon_sym_assert] = ACTIONS(6913),
    [anon_sym_upcast] = ACTIONS(6913),
    [anon_sym_downcast] = ACTIONS(6913),
    [anon_sym_fixed] = ACTIONS(6913),
    [anon_sym_COLON_GT] = ACTIONS(6911),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6911),
    [anon_sym_for] = ACTIONS(6913),
    [anon_sym_while] = ACTIONS(6913),
    [anon_sym_while_BANG] = ACTIONS(6911),
    [anon_sym_if] = ACTIONS(6913),
    [anon_sym_fun] = ACTIONS(6913),
    [anon_sym_try] = ACTIONS(6913),
    [anon_sym_match] = ACTIONS(6913),
    [anon_sym_match_BANG] = ACTIONS(6911),
    [anon_sym_function] = ACTIONS(6913),
    [anon_sym_DOT_LBRACK] = ACTIONS(6911),
    [anon_sym_use] = ACTIONS(6913),
    [anon_sym_use_BANG] = ACTIONS(6911),
    [anon_sym_do_BANG] = ACTIONS(6911),
    [anon_sym_LPAREN2] = ACTIONS(6913),
    [anon_sym_PIPE_PIPE] = ACTIONS(6913),
    [anon_sym_AMP_AMP] = ACTIONS(6913),
    [aux_sym_char_token1] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6913),
    [anon_sym_DQUOTE] = ACTIONS(6913),
    [anon_sym_AT_DQUOTE] = ACTIONS(6911),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6911),
    [sym_bool] = ACTIONS(6913),
    [sym_unit] = ACTIONS(6911),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6913),
    [sym_op_identifier] = ACTIONS(6913),
    [anon_sym_PLUS] = ACTIONS(6913),
    [anon_sym_DASH] = ACTIONS(6913),
    [anon_sym_PLUS_DOT] = ACTIONS(6913),
    [anon_sym_DASH_DOT] = ACTIONS(6913),
    [anon_sym_PERCENT] = ACTIONS(6913),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6913),
    [anon_sym_TILDE] = ACTIONS(6911),
    [aux_sym_prefix_op_token1] = ACTIONS(6913),
    [aux_sym_infix_op_token1] = ACTIONS(6911),
    [aux_sym_infix_op_token2] = ACTIONS(6913),
    [aux_sym_infix_op_token3] = ACTIONS(6913),
    [anon_sym_BANG_EQ] = ACTIONS(6913),
    [anon_sym_COLON_EQ] = ACTIONS(6911),
    [anon_sym_DOLLAR] = ACTIONS(6913),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6913),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6913),
    [sym_int] = ACTIONS(6913),
    [sym_xint] = ACTIONS(6911),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6911),
    [sym__newline] = ACTIONS(6911),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6911),
    [sym__tyapp_open] = ACTIONS(6911),
  },
  [STATE(5723)] = {
    [sym_block_comment] = STATE(5723),
    [sym_compiler_directive_decl] = STATE(5723),
    [sym_fsi_directive_decl] = STATE(5723),
    [sym_preproc_line] = STATE(5723),
    [sym_identifier] = ACTIONS(6917),
    [anon_sym_null] = ACTIONS(6917),
    [anon_sym_EQ] = ACTIONS(6917),
    [anon_sym_begin] = ACTIONS(6917),
    [anon_sym_COLON] = ACTIONS(6917),
    [anon_sym_do] = ACTIONS(6917),
    [anon_sym_let] = ACTIONS(6917),
    [anon_sym_let_BANG] = ACTIONS(6915),
    [anon_sym_as] = ACTIONS(6917),
    [anon_sym_LPAREN] = ACTIONS(6917),
    [anon_sym_COMMA] = ACTIONS(6915),
    [anon_sym_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_QMARK] = ACTIONS(6917),
    [anon_sym_COLON_COLON] = ACTIONS(6915),
    [anon_sym_AMP] = ACTIONS(6917),
    [anon_sym_LBRACK] = ACTIONS(6917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6915),
    [anon_sym_LBRACE] = ACTIONS(6917),
    [anon_sym_LT_AT] = ACTIONS(6917),
    [anon_sym_LT_AT_AT] = ACTIONS(6917),
    [anon_sym_CARET] = ACTIONS(6917),
    [anon_sym_or] = ACTIONS(6917),
    [anon_sym_DOT] = ACTIONS(6917),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6915),
    [anon_sym_new] = ACTIONS(6917),
    [anon_sym_with] = ACTIONS(6917),
    [anon_sym_return] = ACTIONS(6917),
    [anon_sym_return_BANG] = ACTIONS(6915),
    [anon_sym_yield] = ACTIONS(6917),
    [anon_sym_yield_BANG] = ACTIONS(6915),
    [anon_sym_lazy] = ACTIONS(6917),
    [anon_sym_assert] = ACTIONS(6917),
    [anon_sym_upcast] = ACTIONS(6917),
    [anon_sym_downcast] = ACTIONS(6917),
    [anon_sym_fixed] = ACTIONS(6917),
    [anon_sym_COLON_GT] = ACTIONS(6915),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6915),
    [anon_sym_for] = ACTIONS(6917),
    [anon_sym_while] = ACTIONS(6917),
    [anon_sym_while_BANG] = ACTIONS(6915),
    [anon_sym_if] = ACTIONS(6917),
    [anon_sym_fun] = ACTIONS(6917),
    [anon_sym_try] = ACTIONS(6917),
    [anon_sym_match] = ACTIONS(6917),
    [anon_sym_match_BANG] = ACTIONS(6915),
    [anon_sym_function] = ACTIONS(6917),
    [anon_sym_DOT_LBRACK] = ACTIONS(6915),
    [anon_sym_use] = ACTIONS(6917),
    [anon_sym_use_BANG] = ACTIONS(6915),
    [anon_sym_do_BANG] = ACTIONS(6915),
    [anon_sym_LPAREN2] = ACTIONS(6917),
    [anon_sym_PIPE_PIPE] = ACTIONS(6917),
    [anon_sym_AMP_AMP] = ACTIONS(6917),
    [aux_sym_char_token1] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6917),
    [anon_sym_DQUOTE] = ACTIONS(6917),
    [anon_sym_AT_DQUOTE] = ACTIONS(6915),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6915),
    [sym_bool] = ACTIONS(6917),
    [sym_unit] = ACTIONS(6915),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6917),
    [sym_op_identifier] = ACTIONS(6917),
    [anon_sym_PLUS] = ACTIONS(6917),
    [anon_sym_DASH] = ACTIONS(6917),
    [anon_sym_PLUS_DOT] = ACTIONS(6917),
    [anon_sym_DASH_DOT] = ACTIONS(6917),
    [anon_sym_PERCENT] = ACTIONS(6917),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6917),
    [anon_sym_TILDE] = ACTIONS(6915),
    [aux_sym_prefix_op_token1] = ACTIONS(6917),
    [aux_sym_infix_op_token1] = ACTIONS(6915),
    [aux_sym_infix_op_token2] = ACTIONS(6917),
    [aux_sym_infix_op_token3] = ACTIONS(6917),
    [anon_sym_BANG_EQ] = ACTIONS(6917),
    [anon_sym_COLON_EQ] = ACTIONS(6915),
    [anon_sym_DOLLAR] = ACTIONS(6917),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6917),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6917),
    [sym_int] = ACTIONS(6917),
    [sym_xint] = ACTIONS(6915),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6915),
    [sym__newline] = ACTIONS(6915),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6915),
    [sym__tyapp_open] = ACTIONS(6915),
  },
  [STATE(5724)] = {
    [sym_block_comment] = STATE(5724),
    [sym_compiler_directive_decl] = STATE(5724),
    [sym_fsi_directive_decl] = STATE(5724),
    [sym_preproc_line] = STATE(5724),
    [sym_identifier] = ACTIONS(6999),
    [anon_sym_null] = ACTIONS(6999),
    [anon_sym_EQ] = ACTIONS(6999),
    [anon_sym_begin] = ACTIONS(6999),
    [anon_sym_COLON] = ACTIONS(6999),
    [anon_sym_do] = ACTIONS(6999),
    [anon_sym_let] = ACTIONS(6999),
    [anon_sym_let_BANG] = ACTIONS(6997),
    [anon_sym_LPAREN] = ACTIONS(6999),
    [anon_sym_COMMA] = ACTIONS(6997),
    [anon_sym_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_QMARK] = ACTIONS(6999),
    [anon_sym_COLON_COLON] = ACTIONS(6997),
    [anon_sym_AMP] = ACTIONS(6999),
    [anon_sym_LBRACK] = ACTIONS(6999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6997),
    [anon_sym_LBRACE] = ACTIONS(6999),
    [anon_sym_LT_AT] = ACTIONS(6999),
    [anon_sym_LT_AT_AT] = ACTIONS(6999),
    [anon_sym_CARET] = ACTIONS(6999),
    [anon_sym_or] = ACTIONS(6999),
    [anon_sym_DOT] = ACTIONS(6999),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6997),
    [anon_sym_new] = ACTIONS(6999),
    [anon_sym_return] = ACTIONS(6999),
    [anon_sym_return_BANG] = ACTIONS(6997),
    [anon_sym_yield] = ACTIONS(6999),
    [anon_sym_yield_BANG] = ACTIONS(6997),
    [anon_sym_lazy] = ACTIONS(6999),
    [anon_sym_assert] = ACTIONS(6999),
    [anon_sym_upcast] = ACTIONS(6999),
    [anon_sym_downcast] = ACTIONS(6999),
    [anon_sym_fixed] = ACTIONS(6999),
    [anon_sym_COLON_GT] = ACTIONS(6997),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6997),
    [anon_sym_for] = ACTIONS(6999),
    [anon_sym_while] = ACTIONS(6999),
    [anon_sym_while_BANG] = ACTIONS(6997),
    [anon_sym_if] = ACTIONS(6999),
    [anon_sym_fun] = ACTIONS(6999),
    [anon_sym_DASH_GT] = ACTIONS(6999),
    [anon_sym_try] = ACTIONS(6999),
    [anon_sym_match] = ACTIONS(6999),
    [anon_sym_match_BANG] = ACTIONS(6997),
    [anon_sym_function] = ACTIONS(6999),
    [anon_sym_DOT_LBRACK] = ACTIONS(6997),
    [anon_sym_use] = ACTIONS(6999),
    [anon_sym_use_BANG] = ACTIONS(6997),
    [anon_sym_do_BANG] = ACTIONS(6997),
    [anon_sym_DOT_DOT] = ACTIONS(6997),
    [anon_sym_LPAREN2] = ACTIONS(6999),
    [anon_sym_PIPE_PIPE] = ACTIONS(6999),
    [anon_sym_AMP_AMP] = ACTIONS(6999),
    [aux_sym_char_token1] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6999),
    [anon_sym_DQUOTE] = ACTIONS(6999),
    [anon_sym_AT_DQUOTE] = ACTIONS(6997),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6997),
    [sym_bool] = ACTIONS(6999),
    [sym_unit] = ACTIONS(6997),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6999),
    [sym_op_identifier] = ACTIONS(6999),
    [anon_sym_PLUS] = ACTIONS(6999),
    [anon_sym_DASH] = ACTIONS(6999),
    [anon_sym_PLUS_DOT] = ACTIONS(6999),
    [anon_sym_DASH_DOT] = ACTIONS(6999),
    [anon_sym_PERCENT] = ACTIONS(6999),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6999),
    [anon_sym_TILDE] = ACTIONS(6997),
    [aux_sym_prefix_op_token1] = ACTIONS(6999),
    [aux_sym_infix_op_token1] = ACTIONS(6997),
    [aux_sym_infix_op_token2] = ACTIONS(6999),
    [aux_sym_infix_op_token3] = ACTIONS(6999),
    [anon_sym_BANG_EQ] = ACTIONS(6999),
    [anon_sym_COLON_EQ] = ACTIONS(6997),
    [anon_sym_DOLLAR] = ACTIONS(6999),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6999),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6999),
    [sym_int] = ACTIONS(6999),
    [sym_xint] = ACTIONS(6997),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6997),
    [sym__newline] = ACTIONS(6997),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6997),
    [sym__tyapp_open] = ACTIONS(6997),
  },
  [STATE(5725)] = {
    [sym_block_comment] = STATE(5725),
    [sym_compiler_directive_decl] = STATE(5725),
    [sym_fsi_directive_decl] = STATE(5725),
    [sym_preproc_line] = STATE(5725),
    [sym_identifier] = ACTIONS(6921),
    [anon_sym_null] = ACTIONS(6921),
    [anon_sym_EQ] = ACTIONS(6921),
    [anon_sym_begin] = ACTIONS(6921),
    [anon_sym_COLON] = ACTIONS(6921),
    [anon_sym_do] = ACTIONS(6921),
    [anon_sym_let] = ACTIONS(6921),
    [anon_sym_let_BANG] = ACTIONS(6919),
    [anon_sym_as] = ACTIONS(6921),
    [anon_sym_LPAREN] = ACTIONS(6921),
    [anon_sym_COMMA] = ACTIONS(6919),
    [anon_sym_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_QMARK] = ACTIONS(6921),
    [anon_sym_COLON_COLON] = ACTIONS(6919),
    [anon_sym_AMP] = ACTIONS(6921),
    [anon_sym_LBRACK] = ACTIONS(6921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6919),
    [anon_sym_LBRACE] = ACTIONS(6921),
    [anon_sym_LT_AT] = ACTIONS(6921),
    [anon_sym_LT_AT_AT] = ACTIONS(6921),
    [anon_sym_CARET] = ACTIONS(6921),
    [anon_sym_or] = ACTIONS(6921),
    [anon_sym_DOT] = ACTIONS(6921),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6919),
    [anon_sym_new] = ACTIONS(6921),
    [anon_sym_with] = ACTIONS(6921),
    [anon_sym_return] = ACTIONS(6921),
    [anon_sym_return_BANG] = ACTIONS(6919),
    [anon_sym_yield] = ACTIONS(6921),
    [anon_sym_yield_BANG] = ACTIONS(6919),
    [anon_sym_lazy] = ACTIONS(6921),
    [anon_sym_assert] = ACTIONS(6921),
    [anon_sym_upcast] = ACTIONS(6921),
    [anon_sym_downcast] = ACTIONS(6921),
    [anon_sym_fixed] = ACTIONS(6921),
    [anon_sym_COLON_GT] = ACTIONS(6919),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6919),
    [anon_sym_for] = ACTIONS(6921),
    [anon_sym_while] = ACTIONS(6921),
    [anon_sym_while_BANG] = ACTIONS(6919),
    [anon_sym_if] = ACTIONS(6921),
    [anon_sym_fun] = ACTIONS(6921),
    [anon_sym_try] = ACTIONS(6921),
    [anon_sym_match] = ACTIONS(6921),
    [anon_sym_match_BANG] = ACTIONS(6919),
    [anon_sym_function] = ACTIONS(6921),
    [anon_sym_DOT_LBRACK] = ACTIONS(6919),
    [anon_sym_use] = ACTIONS(6921),
    [anon_sym_use_BANG] = ACTIONS(6919),
    [anon_sym_do_BANG] = ACTIONS(6919),
    [anon_sym_LPAREN2] = ACTIONS(6921),
    [anon_sym_PIPE_PIPE] = ACTIONS(6921),
    [anon_sym_AMP_AMP] = ACTIONS(6921),
    [aux_sym_char_token1] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6921),
    [anon_sym_DQUOTE] = ACTIONS(6921),
    [anon_sym_AT_DQUOTE] = ACTIONS(6919),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6919),
    [sym_bool] = ACTIONS(6921),
    [sym_unit] = ACTIONS(6919),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6921),
    [sym_op_identifier] = ACTIONS(6921),
    [anon_sym_PLUS] = ACTIONS(6921),
    [anon_sym_DASH] = ACTIONS(6921),
    [anon_sym_PLUS_DOT] = ACTIONS(6921),
    [anon_sym_DASH_DOT] = ACTIONS(6921),
    [anon_sym_PERCENT] = ACTIONS(6921),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6921),
    [anon_sym_TILDE] = ACTIONS(6919),
    [aux_sym_prefix_op_token1] = ACTIONS(6921),
    [aux_sym_infix_op_token1] = ACTIONS(6919),
    [aux_sym_infix_op_token2] = ACTIONS(6921),
    [aux_sym_infix_op_token3] = ACTIONS(6921),
    [anon_sym_BANG_EQ] = ACTIONS(6921),
    [anon_sym_COLON_EQ] = ACTIONS(6919),
    [anon_sym_DOLLAR] = ACTIONS(6921),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6921),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6921),
    [sym_int] = ACTIONS(6921),
    [sym_xint] = ACTIONS(6919),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6919),
    [sym__newline] = ACTIONS(6919),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6919),
    [sym__tyapp_open] = ACTIONS(6919),
  },
  [STATE(5726)] = {
    [sym_block_comment] = STATE(5726),
    [sym_compiler_directive_decl] = STATE(5726),
    [sym_fsi_directive_decl] = STATE(5726),
    [sym_preproc_line] = STATE(5726),
    [sym_identifier] = ACTIONS(6925),
    [anon_sym_null] = ACTIONS(6925),
    [anon_sym_EQ] = ACTIONS(6925),
    [anon_sym_begin] = ACTIONS(6925),
    [anon_sym_COLON] = ACTIONS(6925),
    [anon_sym_do] = ACTIONS(6925),
    [anon_sym_let] = ACTIONS(6925),
    [anon_sym_let_BANG] = ACTIONS(6923),
    [anon_sym_as] = ACTIONS(6925),
    [anon_sym_LPAREN] = ACTIONS(6925),
    [anon_sym_COMMA] = ACTIONS(6923),
    [anon_sym_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_QMARK] = ACTIONS(6925),
    [anon_sym_COLON_COLON] = ACTIONS(6923),
    [anon_sym_AMP] = ACTIONS(6925),
    [anon_sym_LBRACK] = ACTIONS(6925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6923),
    [anon_sym_LBRACE] = ACTIONS(6925),
    [anon_sym_LT_AT] = ACTIONS(6925),
    [anon_sym_LT_AT_AT] = ACTIONS(6925),
    [anon_sym_CARET] = ACTIONS(6925),
    [anon_sym_or] = ACTIONS(6925),
    [anon_sym_DOT] = ACTIONS(6925),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6923),
    [anon_sym_new] = ACTIONS(6925),
    [anon_sym_with] = ACTIONS(6925),
    [anon_sym_return] = ACTIONS(6925),
    [anon_sym_return_BANG] = ACTIONS(6923),
    [anon_sym_yield] = ACTIONS(6925),
    [anon_sym_yield_BANG] = ACTIONS(6923),
    [anon_sym_lazy] = ACTIONS(6925),
    [anon_sym_assert] = ACTIONS(6925),
    [anon_sym_upcast] = ACTIONS(6925),
    [anon_sym_downcast] = ACTIONS(6925),
    [anon_sym_fixed] = ACTIONS(6925),
    [anon_sym_COLON_GT] = ACTIONS(6923),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6923),
    [anon_sym_for] = ACTIONS(6925),
    [anon_sym_while] = ACTIONS(6925),
    [anon_sym_while_BANG] = ACTIONS(6923),
    [anon_sym_if] = ACTIONS(6925),
    [anon_sym_fun] = ACTIONS(6925),
    [anon_sym_try] = ACTIONS(6925),
    [anon_sym_match] = ACTIONS(6925),
    [anon_sym_match_BANG] = ACTIONS(6923),
    [anon_sym_function] = ACTIONS(6925),
    [anon_sym_DOT_LBRACK] = ACTIONS(6923),
    [anon_sym_use] = ACTIONS(6925),
    [anon_sym_use_BANG] = ACTIONS(6923),
    [anon_sym_do_BANG] = ACTIONS(6923),
    [anon_sym_LPAREN2] = ACTIONS(6925),
    [anon_sym_PIPE_PIPE] = ACTIONS(6925),
    [anon_sym_AMP_AMP] = ACTIONS(6925),
    [aux_sym_char_token1] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6925),
    [anon_sym_DQUOTE] = ACTIONS(6925),
    [anon_sym_AT_DQUOTE] = ACTIONS(6923),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6923),
    [sym_bool] = ACTIONS(6925),
    [sym_unit] = ACTIONS(6923),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6925),
    [sym_op_identifier] = ACTIONS(6925),
    [anon_sym_PLUS] = ACTIONS(6925),
    [anon_sym_DASH] = ACTIONS(6925),
    [anon_sym_PLUS_DOT] = ACTIONS(6925),
    [anon_sym_DASH_DOT] = ACTIONS(6925),
    [anon_sym_PERCENT] = ACTIONS(6925),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6925),
    [anon_sym_TILDE] = ACTIONS(6923),
    [aux_sym_prefix_op_token1] = ACTIONS(6925),
    [aux_sym_infix_op_token1] = ACTIONS(6923),
    [aux_sym_infix_op_token2] = ACTIONS(6925),
    [aux_sym_infix_op_token3] = ACTIONS(6925),
    [anon_sym_BANG_EQ] = ACTIONS(6925),
    [anon_sym_COLON_EQ] = ACTIONS(6923),
    [anon_sym_DOLLAR] = ACTIONS(6925),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6925),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6925),
    [sym_int] = ACTIONS(6925),
    [sym_xint] = ACTIONS(6923),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6923),
    [sym__newline] = ACTIONS(6923),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6923),
    [sym__tyapp_open] = ACTIONS(6923),
  },
  [STATE(5727)] = {
    [sym_block_comment] = STATE(5727),
    [sym_compiler_directive_decl] = STATE(5727),
    [sym_fsi_directive_decl] = STATE(5727),
    [sym_preproc_line] = STATE(5727),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_DOT_DOT] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5728)] = {
    [sym_block_comment] = STATE(5728),
    [sym_compiler_directive_decl] = STATE(5728),
    [sym_fsi_directive_decl] = STATE(5728),
    [sym_preproc_line] = STATE(5728),
    [sym_identifier] = ACTIONS(6597),
    [anon_sym_null] = ACTIONS(6597),
    [anon_sym_EQ] = ACTIONS(6597),
    [anon_sym_begin] = ACTIONS(6597),
    [anon_sym_COLON] = ACTIONS(6597),
    [anon_sym_do] = ACTIONS(6597),
    [anon_sym_let] = ACTIONS(6597),
    [anon_sym_let_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN] = ACTIONS(6597),
    [anon_sym_COMMA] = ACTIONS(6599),
    [anon_sym_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_QMARK] = ACTIONS(6597),
    [anon_sym_COLON_COLON] = ACTIONS(6599),
    [anon_sym_AMP] = ACTIONS(6597),
    [anon_sym_LBRACK] = ACTIONS(6597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6599),
    [anon_sym_LBRACE] = ACTIONS(6597),
    [anon_sym_LT_AT] = ACTIONS(6597),
    [anon_sym_LT_AT_AT] = ACTIONS(6597),
    [anon_sym_CARET] = ACTIONS(6597),
    [anon_sym_or] = ACTIONS(6597),
    [anon_sym_DOT] = ACTIONS(6597),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6599),
    [anon_sym_new] = ACTIONS(6597),
    [anon_sym_return] = ACTIONS(6597),
    [anon_sym_return_BANG] = ACTIONS(6599),
    [anon_sym_yield] = ACTIONS(6597),
    [anon_sym_yield_BANG] = ACTIONS(6599),
    [anon_sym_lazy] = ACTIONS(6597),
    [anon_sym_assert] = ACTIONS(6597),
    [anon_sym_upcast] = ACTIONS(6597),
    [anon_sym_downcast] = ACTIONS(6597),
    [anon_sym_fixed] = ACTIONS(6597),
    [anon_sym_COLON_GT] = ACTIONS(6599),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6599),
    [anon_sym_for] = ACTIONS(6597),
    [anon_sym_done] = ACTIONS(9514),
    [anon_sym_while] = ACTIONS(6597),
    [anon_sym_while_BANG] = ACTIONS(6599),
    [anon_sym_if] = ACTIONS(6597),
    [anon_sym_fun] = ACTIONS(6597),
    [anon_sym_try] = ACTIONS(6597),
    [anon_sym_match] = ACTIONS(6597),
    [anon_sym_match_BANG] = ACTIONS(6599),
    [anon_sym_function] = ACTIONS(6597),
    [anon_sym_DOT_LBRACK] = ACTIONS(6599),
    [anon_sym_use] = ACTIONS(6597),
    [anon_sym_use_BANG] = ACTIONS(6599),
    [anon_sym_do_BANG] = ACTIONS(6599),
    [anon_sym_LPAREN2] = ACTIONS(6597),
    [anon_sym_PIPE_PIPE] = ACTIONS(6597),
    [anon_sym_AMP_AMP] = ACTIONS(6597),
    [aux_sym_char_token1] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6597),
    [anon_sym_DQUOTE] = ACTIONS(6597),
    [anon_sym_AT_DQUOTE] = ACTIONS(6599),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6599),
    [sym_bool] = ACTIONS(6597),
    [sym_unit] = ACTIONS(6599),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6597),
    [sym_op_identifier] = ACTIONS(6597),
    [anon_sym_PLUS] = ACTIONS(6597),
    [anon_sym_DASH] = ACTIONS(6597),
    [anon_sym_PLUS_DOT] = ACTIONS(6597),
    [anon_sym_DASH_DOT] = ACTIONS(6597),
    [anon_sym_PERCENT] = ACTIONS(6597),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6597),
    [anon_sym_TILDE] = ACTIONS(6599),
    [aux_sym_prefix_op_token1] = ACTIONS(6597),
    [aux_sym_infix_op_token1] = ACTIONS(6599),
    [aux_sym_infix_op_token2] = ACTIONS(6597),
    [aux_sym_infix_op_token3] = ACTIONS(6597),
    [anon_sym_BANG_EQ] = ACTIONS(6597),
    [anon_sym_COLON_EQ] = ACTIONS(6599),
    [anon_sym_DOLLAR] = ACTIONS(6597),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6597),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6597),
    [sym_int] = ACTIONS(6597),
    [sym_xint] = ACTIONS(6599),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6599),
    [sym__newline] = ACTIONS(6599),
    [sym__quoted_close] = ACTIONS(6599),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6599),
    [sym__tyapp_open] = ACTIONS(6599),
  },
  [STATE(5729)] = {
    [sym_block_comment] = STATE(5729),
    [sym_compiler_directive_decl] = STATE(5729),
    [sym_fsi_directive_decl] = STATE(5729),
    [sym_preproc_line] = STATE(5729),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__quoted_close] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5730)] = {
    [sym_block_comment] = STATE(5730),
    [sym_compiler_directive_decl] = STATE(5730),
    [sym_fsi_directive_decl] = STATE(5730),
    [sym_preproc_line] = STATE(5730),
    [sym_identifier] = ACTIONS(6871),
    [anon_sym_null] = ACTIONS(6871),
    [anon_sym_EQ] = ACTIONS(6871),
    [anon_sym_begin] = ACTIONS(6871),
    [anon_sym_COLON] = ACTIONS(6871),
    [anon_sym_do] = ACTIONS(6871),
    [anon_sym_let] = ACTIONS(6871),
    [anon_sym_let_BANG] = ACTIONS(6869),
    [anon_sym_LPAREN] = ACTIONS(6871),
    [anon_sym_COMMA] = ACTIONS(6869),
    [anon_sym_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_QMARK] = ACTIONS(6871),
    [anon_sym_COLON_COLON] = ACTIONS(6869),
    [anon_sym_AMP] = ACTIONS(6871),
    [anon_sym_LBRACK] = ACTIONS(6871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6869),
    [anon_sym_LBRACE] = ACTIONS(6871),
    [anon_sym_LT_AT] = ACTIONS(6871),
    [anon_sym_LT_AT_AT] = ACTIONS(6871),
    [anon_sym_CARET] = ACTIONS(6871),
    [anon_sym_or] = ACTIONS(6871),
    [anon_sym_DOT] = ACTIONS(6871),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6869),
    [anon_sym_new] = ACTIONS(6871),
    [anon_sym_return] = ACTIONS(6871),
    [anon_sym_return_BANG] = ACTIONS(6869),
    [anon_sym_yield] = ACTIONS(6871),
    [anon_sym_yield_BANG] = ACTIONS(6869),
    [anon_sym_lazy] = ACTIONS(6871),
    [anon_sym_assert] = ACTIONS(6871),
    [anon_sym_upcast] = ACTIONS(6871),
    [anon_sym_downcast] = ACTIONS(6871),
    [anon_sym_fixed] = ACTIONS(6871),
    [anon_sym_COLON_GT] = ACTIONS(6869),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6869),
    [anon_sym_for] = ACTIONS(6871),
    [anon_sym_while] = ACTIONS(6871),
    [anon_sym_while_BANG] = ACTIONS(6869),
    [anon_sym_if] = ACTIONS(6871),
    [anon_sym_fun] = ACTIONS(6871),
    [anon_sym_try] = ACTIONS(6871),
    [anon_sym_match] = ACTIONS(6871),
    [anon_sym_match_BANG] = ACTIONS(6869),
    [anon_sym_function] = ACTIONS(6871),
    [anon_sym_LT_DASH] = ACTIONS(6871),
    [anon_sym_DOT_LBRACK] = ACTIONS(6869),
    [anon_sym_use] = ACTIONS(6871),
    [anon_sym_use_BANG] = ACTIONS(6869),
    [anon_sym_do_BANG] = ACTIONS(6869),
    [anon_sym_DOT_DOT] = ACTIONS(6869),
    [anon_sym_LPAREN2] = ACTIONS(6871),
    [anon_sym_PIPE_PIPE] = ACTIONS(6871),
    [anon_sym_AMP_AMP] = ACTIONS(6871),
    [aux_sym_char_token1] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6871),
    [anon_sym_DQUOTE] = ACTIONS(6871),
    [anon_sym_AT_DQUOTE] = ACTIONS(6869),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6869),
    [sym_bool] = ACTIONS(6871),
    [sym_unit] = ACTIONS(6869),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6871),
    [sym_op_identifier] = ACTIONS(6871),
    [anon_sym_PLUS] = ACTIONS(6871),
    [anon_sym_DASH] = ACTIONS(6871),
    [anon_sym_PLUS_DOT] = ACTIONS(6871),
    [anon_sym_DASH_DOT] = ACTIONS(6871),
    [anon_sym_PERCENT] = ACTIONS(6871),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6871),
    [anon_sym_TILDE] = ACTIONS(6869),
    [aux_sym_prefix_op_token1] = ACTIONS(6871),
    [aux_sym_infix_op_token1] = ACTIONS(6869),
    [aux_sym_infix_op_token2] = ACTIONS(6871),
    [aux_sym_infix_op_token3] = ACTIONS(6871),
    [anon_sym_BANG_EQ] = ACTIONS(6871),
    [anon_sym_COLON_EQ] = ACTIONS(6869),
    [anon_sym_DOLLAR] = ACTIONS(6871),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6871),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6871),
    [sym_int] = ACTIONS(6871),
    [sym_xint] = ACTIONS(6869),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6869),
    [sym__newline] = ACTIONS(6869),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6869),
    [sym__tyapp_open] = ACTIONS(6869),
  },
  [STATE(5731)] = {
    [sym_block_comment] = STATE(5731),
    [sym_compiler_directive_decl] = STATE(5731),
    [sym_fsi_directive_decl] = STATE(5731),
    [sym_preproc_line] = STATE(5731),
    [sym_identifier] = ACTIONS(6875),
    [anon_sym_null] = ACTIONS(6875),
    [anon_sym_EQ] = ACTIONS(6875),
    [anon_sym_begin] = ACTIONS(6875),
    [anon_sym_COLON] = ACTIONS(6875),
    [anon_sym_do] = ACTIONS(6875),
    [anon_sym_let] = ACTIONS(6875),
    [anon_sym_let_BANG] = ACTIONS(6873),
    [anon_sym_LPAREN] = ACTIONS(6875),
    [anon_sym_COMMA] = ACTIONS(6873),
    [anon_sym_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_QMARK] = ACTIONS(6875),
    [anon_sym_COLON_COLON] = ACTIONS(6873),
    [anon_sym_AMP] = ACTIONS(6875),
    [anon_sym_LBRACK] = ACTIONS(6875),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6873),
    [anon_sym_LBRACE] = ACTIONS(6875),
    [anon_sym_LT_AT] = ACTIONS(6875),
    [anon_sym_LT_AT_AT] = ACTIONS(6875),
    [anon_sym_CARET] = ACTIONS(6875),
    [anon_sym_or] = ACTIONS(6875),
    [anon_sym_DOT] = ACTIONS(6875),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6873),
    [anon_sym_new] = ACTIONS(6875),
    [anon_sym_return] = ACTIONS(6875),
    [anon_sym_return_BANG] = ACTIONS(6873),
    [anon_sym_yield] = ACTIONS(6875),
    [anon_sym_yield_BANG] = ACTIONS(6873),
    [anon_sym_lazy] = ACTIONS(6875),
    [anon_sym_assert] = ACTIONS(6875),
    [anon_sym_upcast] = ACTIONS(6875),
    [anon_sym_downcast] = ACTIONS(6875),
    [anon_sym_fixed] = ACTIONS(6875),
    [anon_sym_COLON_GT] = ACTIONS(6873),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6873),
    [anon_sym_for] = ACTIONS(6875),
    [anon_sym_while] = ACTIONS(6875),
    [anon_sym_while_BANG] = ACTIONS(6873),
    [anon_sym_if] = ACTIONS(6875),
    [anon_sym_fun] = ACTIONS(6875),
    [anon_sym_try] = ACTIONS(6875),
    [anon_sym_match] = ACTIONS(6875),
    [anon_sym_match_BANG] = ACTIONS(6873),
    [anon_sym_function] = ACTIONS(6875),
    [anon_sym_LT_DASH] = ACTIONS(6875),
    [anon_sym_DOT_LBRACK] = ACTIONS(6873),
    [anon_sym_use] = ACTIONS(6875),
    [anon_sym_use_BANG] = ACTIONS(6873),
    [anon_sym_do_BANG] = ACTIONS(6873),
    [anon_sym_DOT_DOT] = ACTIONS(6873),
    [anon_sym_LPAREN2] = ACTIONS(6875),
    [anon_sym_PIPE_PIPE] = ACTIONS(6875),
    [anon_sym_AMP_AMP] = ACTIONS(6875),
    [aux_sym_char_token1] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6875),
    [anon_sym_DQUOTE] = ACTIONS(6875),
    [anon_sym_AT_DQUOTE] = ACTIONS(6873),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6873),
    [sym_bool] = ACTIONS(6875),
    [sym_unit] = ACTIONS(6873),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6875),
    [sym_op_identifier] = ACTIONS(6875),
    [anon_sym_PLUS] = ACTIONS(6875),
    [anon_sym_DASH] = ACTIONS(6875),
    [anon_sym_PLUS_DOT] = ACTIONS(6875),
    [anon_sym_DASH_DOT] = ACTIONS(6875),
    [anon_sym_PERCENT] = ACTIONS(6875),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6875),
    [anon_sym_TILDE] = ACTIONS(6873),
    [aux_sym_prefix_op_token1] = ACTIONS(6875),
    [aux_sym_infix_op_token1] = ACTIONS(6873),
    [aux_sym_infix_op_token2] = ACTIONS(6875),
    [aux_sym_infix_op_token3] = ACTIONS(6875),
    [anon_sym_BANG_EQ] = ACTIONS(6875),
    [anon_sym_COLON_EQ] = ACTIONS(6873),
    [anon_sym_DOLLAR] = ACTIONS(6875),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6875),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6875),
    [sym_int] = ACTIONS(6875),
    [sym_xint] = ACTIONS(6873),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6873),
    [sym__newline] = ACTIONS(6873),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6873),
    [sym__tyapp_open] = ACTIONS(6873),
  },
  [STATE(5732)] = {
    [sym_block_comment] = STATE(5732),
    [sym_compiler_directive_decl] = STATE(5732),
    [sym_fsi_directive_decl] = STATE(5732),
    [sym_preproc_line] = STATE(5732),
    [sym_identifier] = ACTIONS(7003),
    [anon_sym_null] = ACTIONS(7003),
    [anon_sym_EQ] = ACTIONS(7003),
    [anon_sym_begin] = ACTIONS(7003),
    [anon_sym_COLON] = ACTIONS(7003),
    [anon_sym_do] = ACTIONS(7003),
    [anon_sym_let] = ACTIONS(7003),
    [anon_sym_let_BANG] = ACTIONS(7001),
    [anon_sym_LPAREN] = ACTIONS(7003),
    [anon_sym_COMMA] = ACTIONS(7001),
    [anon_sym_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_QMARK] = ACTIONS(7003),
    [anon_sym_COLON_COLON] = ACTIONS(7001),
    [anon_sym_AMP] = ACTIONS(7003),
    [anon_sym_LBRACK] = ACTIONS(7003),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7001),
    [anon_sym_LBRACE] = ACTIONS(7003),
    [anon_sym_LT_AT] = ACTIONS(7003),
    [anon_sym_LT_AT_AT] = ACTIONS(7003),
    [anon_sym_CARET] = ACTIONS(7003),
    [anon_sym_or] = ACTIONS(7003),
    [anon_sym_DOT] = ACTIONS(7003),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7001),
    [anon_sym_new] = ACTIONS(7003),
    [anon_sym_return] = ACTIONS(7003),
    [anon_sym_return_BANG] = ACTIONS(7001),
    [anon_sym_yield] = ACTIONS(7003),
    [anon_sym_yield_BANG] = ACTIONS(7001),
    [anon_sym_lazy] = ACTIONS(7003),
    [anon_sym_assert] = ACTIONS(7003),
    [anon_sym_upcast] = ACTIONS(7003),
    [anon_sym_downcast] = ACTIONS(7003),
    [anon_sym_fixed] = ACTIONS(7003),
    [anon_sym_COLON_GT] = ACTIONS(7001),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7001),
    [anon_sym_for] = ACTIONS(7003),
    [anon_sym_while] = ACTIONS(7003),
    [anon_sym_while_BANG] = ACTIONS(7001),
    [anon_sym_if] = ACTIONS(7003),
    [anon_sym_fun] = ACTIONS(7003),
    [anon_sym_DASH_GT] = ACTIONS(7003),
    [anon_sym_try] = ACTIONS(7003),
    [anon_sym_match] = ACTIONS(7003),
    [anon_sym_match_BANG] = ACTIONS(7001),
    [anon_sym_function] = ACTIONS(7003),
    [anon_sym_DOT_LBRACK] = ACTIONS(7001),
    [anon_sym_use] = ACTIONS(7003),
    [anon_sym_use_BANG] = ACTIONS(7001),
    [anon_sym_do_BANG] = ACTIONS(7001),
    [anon_sym_DOT_DOT] = ACTIONS(7001),
    [anon_sym_LPAREN2] = ACTIONS(7003),
    [anon_sym_PIPE_PIPE] = ACTIONS(7003),
    [anon_sym_AMP_AMP] = ACTIONS(7003),
    [aux_sym_char_token1] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7003),
    [anon_sym_DQUOTE] = ACTIONS(7003),
    [anon_sym_AT_DQUOTE] = ACTIONS(7001),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7001),
    [sym_bool] = ACTIONS(7003),
    [sym_unit] = ACTIONS(7001),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7003),
    [sym_op_identifier] = ACTIONS(7003),
    [anon_sym_PLUS] = ACTIONS(7003),
    [anon_sym_DASH] = ACTIONS(7003),
    [anon_sym_PLUS_DOT] = ACTIONS(7003),
    [anon_sym_DASH_DOT] = ACTIONS(7003),
    [anon_sym_PERCENT] = ACTIONS(7003),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7003),
    [anon_sym_TILDE] = ACTIONS(7001),
    [aux_sym_prefix_op_token1] = ACTIONS(7003),
    [aux_sym_infix_op_token1] = ACTIONS(7001),
    [aux_sym_infix_op_token2] = ACTIONS(7003),
    [aux_sym_infix_op_token3] = ACTIONS(7003),
    [anon_sym_BANG_EQ] = ACTIONS(7003),
    [anon_sym_COLON_EQ] = ACTIONS(7001),
    [anon_sym_DOLLAR] = ACTIONS(7003),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7003),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7003),
    [sym_int] = ACTIONS(7003),
    [sym_xint] = ACTIONS(7001),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7001),
    [sym__newline] = ACTIONS(7001),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7001),
    [sym__tyapp_open] = ACTIONS(7001),
  },
  [STATE(5733)] = {
    [sym_block_comment] = STATE(5733),
    [sym_compiler_directive_decl] = STATE(5733),
    [sym_fsi_directive_decl] = STATE(5733),
    [sym_preproc_line] = STATE(5733),
    [sym_identifier] = ACTIONS(6828),
    [anon_sym_null] = ACTIONS(6828),
    [anon_sym_EQ] = ACTIONS(6828),
    [anon_sym_begin] = ACTIONS(6828),
    [anon_sym_COLON] = ACTIONS(6828),
    [anon_sym_do] = ACTIONS(6828),
    [anon_sym_let] = ACTIONS(6828),
    [anon_sym_let_BANG] = ACTIONS(6830),
    [anon_sym_LPAREN] = ACTIONS(6828),
    [anon_sym_COMMA] = ACTIONS(6830),
    [anon_sym_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_QMARK] = ACTIONS(6828),
    [anon_sym_COLON_COLON] = ACTIONS(6830),
    [anon_sym_AMP] = ACTIONS(6828),
    [anon_sym_LBRACK] = ACTIONS(6828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6830),
    [anon_sym_LBRACE] = ACTIONS(6828),
    [anon_sym_LT_AT] = ACTIONS(6828),
    [anon_sym_LT_AT_AT] = ACTIONS(6828),
    [anon_sym_CARET] = ACTIONS(6828),
    [anon_sym_or] = ACTIONS(6828),
    [anon_sym_DOT] = ACTIONS(6828),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6830),
    [anon_sym_new] = ACTIONS(6828),
    [anon_sym_return] = ACTIONS(6828),
    [anon_sym_return_BANG] = ACTIONS(6830),
    [anon_sym_yield] = ACTIONS(6828),
    [anon_sym_yield_BANG] = ACTIONS(6830),
    [anon_sym_lazy] = ACTIONS(6828),
    [anon_sym_assert] = ACTIONS(6828),
    [anon_sym_upcast] = ACTIONS(6828),
    [anon_sym_downcast] = ACTIONS(6828),
    [anon_sym_fixed] = ACTIONS(6828),
    [anon_sym_COLON_GT] = ACTIONS(6830),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6830),
    [anon_sym_for] = ACTIONS(6828),
    [anon_sym_while] = ACTIONS(6828),
    [anon_sym_while_BANG] = ACTIONS(6830),
    [anon_sym_if] = ACTIONS(6828),
    [anon_sym_fun] = ACTIONS(6828),
    [anon_sym_try] = ACTIONS(6828),
    [anon_sym_match] = ACTIONS(6828),
    [anon_sym_match_BANG] = ACTIONS(6830),
    [anon_sym_function] = ACTIONS(6828),
    [anon_sym_LT_DASH] = ACTIONS(6828),
    [anon_sym_DOT_LBRACK] = ACTIONS(6830),
    [anon_sym_use] = ACTIONS(6828),
    [anon_sym_use_BANG] = ACTIONS(6830),
    [anon_sym_do_BANG] = ACTIONS(6830),
    [anon_sym_DOT_DOT] = ACTIONS(6830),
    [anon_sym_LPAREN2] = ACTIONS(6828),
    [anon_sym_PIPE_PIPE] = ACTIONS(6828),
    [anon_sym_AMP_AMP] = ACTIONS(6828),
    [aux_sym_char_token1] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6828),
    [anon_sym_DQUOTE] = ACTIONS(6828),
    [anon_sym_AT_DQUOTE] = ACTIONS(6830),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6830),
    [sym_bool] = ACTIONS(6828),
    [sym_unit] = ACTIONS(6830),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6828),
    [sym_op_identifier] = ACTIONS(6828),
    [anon_sym_PLUS] = ACTIONS(6828),
    [anon_sym_DASH] = ACTIONS(6828),
    [anon_sym_PLUS_DOT] = ACTIONS(6828),
    [anon_sym_DASH_DOT] = ACTIONS(6828),
    [anon_sym_PERCENT] = ACTIONS(6828),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6828),
    [anon_sym_TILDE] = ACTIONS(6830),
    [aux_sym_prefix_op_token1] = ACTIONS(6828),
    [aux_sym_infix_op_token1] = ACTIONS(6830),
    [aux_sym_infix_op_token2] = ACTIONS(6828),
    [aux_sym_infix_op_token3] = ACTIONS(6828),
    [anon_sym_BANG_EQ] = ACTIONS(6828),
    [anon_sym_COLON_EQ] = ACTIONS(6830),
    [anon_sym_DOLLAR] = ACTIONS(6828),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6828),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6828),
    [sym_int] = ACTIONS(6828),
    [sym_xint] = ACTIONS(6830),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6830),
    [sym__newline] = ACTIONS(6830),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6830),
    [sym__tyapp_open] = ACTIONS(6830),
  },
  [STATE(5734)] = {
    [sym_block_comment] = STATE(5734),
    [sym_compiler_directive_decl] = STATE(5734),
    [sym_fsi_directive_decl] = STATE(5734),
    [sym_preproc_line] = STATE(5734),
    [sym_identifier] = ACTIONS(7007),
    [anon_sym_null] = ACTIONS(7007),
    [anon_sym_EQ] = ACTIONS(7007),
    [anon_sym_begin] = ACTIONS(7007),
    [anon_sym_COLON] = ACTIONS(7007),
    [anon_sym_do] = ACTIONS(7007),
    [anon_sym_let] = ACTIONS(7007),
    [anon_sym_let_BANG] = ACTIONS(7005),
    [anon_sym_LPAREN] = ACTIONS(7007),
    [anon_sym_COMMA] = ACTIONS(7005),
    [anon_sym_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_QMARK] = ACTIONS(7007),
    [anon_sym_COLON_COLON] = ACTIONS(7005),
    [anon_sym_AMP] = ACTIONS(7007),
    [anon_sym_LBRACK] = ACTIONS(7007),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7005),
    [anon_sym_LBRACE] = ACTIONS(7007),
    [anon_sym_LT_AT] = ACTIONS(7007),
    [anon_sym_LT_AT_AT] = ACTIONS(7007),
    [anon_sym_CARET] = ACTIONS(7007),
    [anon_sym_or] = ACTIONS(7007),
    [anon_sym_DOT] = ACTIONS(7007),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7005),
    [anon_sym_new] = ACTIONS(7007),
    [anon_sym_return] = ACTIONS(7007),
    [anon_sym_return_BANG] = ACTIONS(7005),
    [anon_sym_yield] = ACTIONS(7007),
    [anon_sym_yield_BANG] = ACTIONS(7005),
    [anon_sym_lazy] = ACTIONS(7007),
    [anon_sym_assert] = ACTIONS(7007),
    [anon_sym_upcast] = ACTIONS(7007),
    [anon_sym_downcast] = ACTIONS(7007),
    [anon_sym_fixed] = ACTIONS(7007),
    [anon_sym_COLON_GT] = ACTIONS(7005),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7005),
    [anon_sym_for] = ACTIONS(7007),
    [anon_sym_while] = ACTIONS(7007),
    [anon_sym_while_BANG] = ACTIONS(7005),
    [anon_sym_if] = ACTIONS(7007),
    [anon_sym_fun] = ACTIONS(7007),
    [anon_sym_DASH_GT] = ACTIONS(7007),
    [anon_sym_try] = ACTIONS(7007),
    [anon_sym_match] = ACTIONS(7007),
    [anon_sym_match_BANG] = ACTIONS(7005),
    [anon_sym_function] = ACTIONS(7007),
    [anon_sym_DOT_LBRACK] = ACTIONS(7005),
    [anon_sym_use] = ACTIONS(7007),
    [anon_sym_use_BANG] = ACTIONS(7005),
    [anon_sym_do_BANG] = ACTIONS(7005),
    [anon_sym_DOT_DOT] = ACTIONS(7005),
    [anon_sym_LPAREN2] = ACTIONS(7007),
    [anon_sym_PIPE_PIPE] = ACTIONS(7007),
    [anon_sym_AMP_AMP] = ACTIONS(7007),
    [aux_sym_char_token1] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7007),
    [anon_sym_DQUOTE] = ACTIONS(7007),
    [anon_sym_AT_DQUOTE] = ACTIONS(7005),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7005),
    [sym_bool] = ACTIONS(7007),
    [sym_unit] = ACTIONS(7005),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7007),
    [sym_op_identifier] = ACTIONS(7007),
    [anon_sym_PLUS] = ACTIONS(7007),
    [anon_sym_DASH] = ACTIONS(7007),
    [anon_sym_PLUS_DOT] = ACTIONS(7007),
    [anon_sym_DASH_DOT] = ACTIONS(7007),
    [anon_sym_PERCENT] = ACTIONS(7007),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7007),
    [anon_sym_TILDE] = ACTIONS(7005),
    [aux_sym_prefix_op_token1] = ACTIONS(7007),
    [aux_sym_infix_op_token1] = ACTIONS(7005),
    [aux_sym_infix_op_token2] = ACTIONS(7007),
    [aux_sym_infix_op_token3] = ACTIONS(7007),
    [anon_sym_BANG_EQ] = ACTIONS(7007),
    [anon_sym_COLON_EQ] = ACTIONS(7005),
    [anon_sym_DOLLAR] = ACTIONS(7007),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7007),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7007),
    [sym_int] = ACTIONS(7007),
    [sym_xint] = ACTIONS(7005),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7005),
    [sym__newline] = ACTIONS(7005),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7005),
    [sym__tyapp_open] = ACTIONS(7005),
  },
  [STATE(5735)] = {
    [sym_block_comment] = STATE(5735),
    [sym_compiler_directive_decl] = STATE(5735),
    [sym_fsi_directive_decl] = STATE(5735),
    [sym_preproc_line] = STATE(5735),
    [sym_identifier] = ACTIONS(7011),
    [anon_sym_null] = ACTIONS(7011),
    [anon_sym_EQ] = ACTIONS(7011),
    [anon_sym_begin] = ACTIONS(7011),
    [anon_sym_COLON] = ACTIONS(7011),
    [anon_sym_do] = ACTIONS(7011),
    [anon_sym_let] = ACTIONS(7011),
    [anon_sym_let_BANG] = ACTIONS(7009),
    [anon_sym_LPAREN] = ACTIONS(7011),
    [anon_sym_COMMA] = ACTIONS(7009),
    [anon_sym_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_QMARK] = ACTIONS(7011),
    [anon_sym_COLON_COLON] = ACTIONS(7009),
    [anon_sym_AMP] = ACTIONS(7011),
    [anon_sym_LBRACK] = ACTIONS(7011),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7009),
    [anon_sym_LBRACE] = ACTIONS(7011),
    [anon_sym_LT_AT] = ACTIONS(7011),
    [anon_sym_LT_AT_AT] = ACTIONS(7011),
    [anon_sym_CARET] = ACTIONS(7011),
    [anon_sym_or] = ACTIONS(7011),
    [anon_sym_DOT] = ACTIONS(7011),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7009),
    [anon_sym_new] = ACTIONS(7011),
    [anon_sym_return] = ACTIONS(7011),
    [anon_sym_return_BANG] = ACTIONS(7009),
    [anon_sym_yield] = ACTIONS(7011),
    [anon_sym_yield_BANG] = ACTIONS(7009),
    [anon_sym_lazy] = ACTIONS(7011),
    [anon_sym_assert] = ACTIONS(7011),
    [anon_sym_upcast] = ACTIONS(7011),
    [anon_sym_downcast] = ACTIONS(7011),
    [anon_sym_fixed] = ACTIONS(7011),
    [anon_sym_COLON_GT] = ACTIONS(7009),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7009),
    [anon_sym_for] = ACTIONS(7011),
    [anon_sym_while] = ACTIONS(7011),
    [anon_sym_while_BANG] = ACTIONS(7009),
    [anon_sym_if] = ACTIONS(7011),
    [anon_sym_fun] = ACTIONS(7011),
    [anon_sym_DASH_GT] = ACTIONS(7011),
    [anon_sym_try] = ACTIONS(7011),
    [anon_sym_match] = ACTIONS(7011),
    [anon_sym_match_BANG] = ACTIONS(7009),
    [anon_sym_function] = ACTIONS(7011),
    [anon_sym_DOT_LBRACK] = ACTIONS(7009),
    [anon_sym_use] = ACTIONS(7011),
    [anon_sym_use_BANG] = ACTIONS(7009),
    [anon_sym_do_BANG] = ACTIONS(7009),
    [anon_sym_DOT_DOT] = ACTIONS(7009),
    [anon_sym_LPAREN2] = ACTIONS(7011),
    [anon_sym_PIPE_PIPE] = ACTIONS(7011),
    [anon_sym_AMP_AMP] = ACTIONS(7011),
    [aux_sym_char_token1] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7011),
    [anon_sym_DQUOTE] = ACTIONS(7011),
    [anon_sym_AT_DQUOTE] = ACTIONS(7009),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7009),
    [sym_bool] = ACTIONS(7011),
    [sym_unit] = ACTIONS(7009),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7011),
    [sym_op_identifier] = ACTIONS(7011),
    [anon_sym_PLUS] = ACTIONS(7011),
    [anon_sym_DASH] = ACTIONS(7011),
    [anon_sym_PLUS_DOT] = ACTIONS(7011),
    [anon_sym_DASH_DOT] = ACTIONS(7011),
    [anon_sym_PERCENT] = ACTIONS(7011),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7011),
    [anon_sym_TILDE] = ACTIONS(7009),
    [aux_sym_prefix_op_token1] = ACTIONS(7011),
    [aux_sym_infix_op_token1] = ACTIONS(7009),
    [aux_sym_infix_op_token2] = ACTIONS(7011),
    [aux_sym_infix_op_token3] = ACTIONS(7011),
    [anon_sym_BANG_EQ] = ACTIONS(7011),
    [anon_sym_COLON_EQ] = ACTIONS(7009),
    [anon_sym_DOLLAR] = ACTIONS(7011),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7011),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7011),
    [sym_int] = ACTIONS(7011),
    [sym_xint] = ACTIONS(7009),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7009),
    [sym__newline] = ACTIONS(7009),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7009),
    [sym__tyapp_open] = ACTIONS(7009),
  },
  [STATE(5736)] = {
    [sym_type_arguments] = STATE(6569),
    [sym__object_members] = STATE(7199),
    [sym_long_identifier] = STATE(6570),
    [sym_block_comment] = STATE(5736),
    [sym_compiler_directive_decl] = STATE(5736),
    [sym_fsi_directive_decl] = STATE(5736),
    [sym_preproc_line] = STATE(5736),
    [aux_sym_compound_type_repeat1] = STATE(6524),
    [sym_identifier] = ACTIONS(9739),
    [anon_sym_module] = ACTIONS(9224),
    [anon_sym_null] = ACTIONS(9224),
    [anon_sym_begin] = ACTIONS(9224),
    [anon_sym_open] = ACTIONS(9224),
    [anon_sym_type] = ACTIONS(9224),
    [anon_sym_LBRACK_LT] = ACTIONS(9226),
    [anon_sym_do] = ACTIONS(9224),
    [anon_sym_and] = ACTIONS(9224),
    [anon_sym_let] = ACTIONS(9224),
    [anon_sym_let_BANG] = ACTIONS(9226),
    [aux_sym_access_modifier_token1] = ACTIONS(9226),
    [anon_sym_LPAREN] = ACTIONS(9224),
    [anon_sym_AMP] = ACTIONS(9224),
    [anon_sym_LBRACK] = ACTIONS(9224),
    [anon_sym_LBRACK_PIPE] = ACTIONS(9226),
    [anon_sym_LBRACE] = ACTIONS(9224),
    [anon_sym_LT_AT] = ACTIONS(9224),
    [anon_sym_LT_AT_AT] = ACTIONS(9226),
    [anon_sym_CARET] = ACTIONS(9226),
    [anon_sym_LBRACE_PIPE] = ACTIONS(9226),
    [anon_sym_new] = ACTIONS(9224),
    [anon_sym_with] = ACTIONS(9840),
    [anon_sym_return] = ACTIONS(9224),
    [anon_sym_return_BANG] = ACTIONS(9226),
    [anon_sym_yield] = ACTIONS(9224),
    [anon_sym_yield_BANG] = ACTIONS(9226),
    [anon_sym_lazy] = ACTIONS(9224),
    [anon_sym_assert] = ACTIONS(9224),
    [anon_sym_upcast] = ACTIONS(9224),
    [anon_sym_downcast] = ACTIONS(9224),
    [anon_sym_fixed] = ACTIONS(9224),
    [anon_sym_for] = ACTIONS(9224),
    [anon_sym_while] = ACTIONS(9224),
    [anon_sym_while_BANG] = ACTIONS(9226),
    [anon_sym_if] = ACTIONS(9224),
    [anon_sym_fun] = ACTIONS(9224),
    [anon_sym_DASH_GT] = ACTIONS(9741),
    [anon_sym_try] = ACTIONS(9224),
    [anon_sym_match] = ACTIONS(9224),
    [anon_sym_match_BANG] = ACTIONS(9226),
    [anon_sym_function] = ACTIONS(9224),
    [anon_sym_use] = ACTIONS(9224),
    [anon_sym_use_BANG] = ACTIONS(9226),
    [anon_sym_do_BANG] = ACTIONS(9226),
    [anon_sym_AMP_AMP] = ACTIONS(9226),
    [anon_sym_STAR] = ACTIONS(9794),
    [anon_sym_LT] = ACTIONS(9743),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(9745),
    [anon_sym_default] = ACTIONS(9224),
    [anon_sym_static] = ACTIONS(9224),
    [anon_sym_member] = ACTIONS(9224),
    [anon_sym_exception] = ACTIONS(9224),
    [anon_sym_interface] = ACTIONS(9224),
    [anon_sym_abstract] = ACTIONS(9224),
    [anon_sym_val] = ACTIONS(9224),
    [anon_sym_override] = ACTIONS(9224),
    [anon_sym_extern] = ACTIONS(9224),
    [aux_sym_char_token1] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(9224),
    [anon_sym_DQUOTE] = ACTIONS(9224),
    [anon_sym_AT_DQUOTE] = ACTIONS(9226),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(9226),
    [sym_bool] = ACTIONS(9224),
    [sym_unit] = ACTIONS(9226),
    [anon_sym_LPAREN_PIPE] = ACTIONS(9224),
    [sym_op_identifier] = ACTIONS(9226),
    [anon_sym_PLUS] = ACTIONS(9224),
    [anon_sym_DASH] = ACTIONS(9224),
    [anon_sym_PLUS_DOT] = ACTIONS(9226),
    [anon_sym_DASH_DOT] = ACTIONS(9226),
    [anon_sym_PERCENT] = ACTIONS(9224),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(9226),
    [anon_sym_TILDE] = ACTIONS(9226),
    [aux_sym_prefix_op_token1] = ACTIONS(9226),
    [sym_int] = ACTIONS(9224),
    [sym_xint] = ACTIONS(9226),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(9226),
    [anon_sym_POUNDload] = ACTIONS(9226),
    [anon_sym_POUNDtime] = ACTIONS(9226),
    [anon_sym_POUNDI] = ACTIONS(9226),
    [anon_sym_POUNDhelp] = ACTIONS(9226),
    [anon_sym_POUNDquit] = ACTIONS(9226),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(9226),
    [anon_sym_POUNDendif] = ACTIONS(9226),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(9226),
  },
  [STATE(5737)] = {
    [sym_block_comment] = STATE(5737),
    [sym_compiler_directive_decl] = STATE(5737),
    [sym_fsi_directive_decl] = STATE(5737),
    [sym_preproc_line] = STATE(5737),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__quoted_close] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5738)] = {
    [sym_block_comment] = STATE(5738),
    [sym_compiler_directive_decl] = STATE(5738),
    [sym_fsi_directive_decl] = STATE(5738),
    [sym_preproc_line] = STATE(5738),
    [sym_identifier] = ACTIONS(7019),
    [anon_sym_null] = ACTIONS(7019),
    [anon_sym_EQ] = ACTIONS(7019),
    [anon_sym_begin] = ACTIONS(7019),
    [anon_sym_COLON] = ACTIONS(7019),
    [anon_sym_do] = ACTIONS(7019),
    [anon_sym_let] = ACTIONS(7019),
    [anon_sym_let_BANG] = ACTIONS(7017),
    [anon_sym_LPAREN] = ACTIONS(7019),
    [anon_sym_COMMA] = ACTIONS(7017),
    [anon_sym_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_QMARK] = ACTIONS(7019),
    [anon_sym_COLON_COLON] = ACTIONS(7017),
    [anon_sym_AMP] = ACTIONS(7019),
    [anon_sym_LBRACK] = ACTIONS(7019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7017),
    [anon_sym_LBRACE] = ACTIONS(7019),
    [anon_sym_LT_AT] = ACTIONS(7019),
    [anon_sym_LT_AT_AT] = ACTIONS(7019),
    [anon_sym_CARET] = ACTIONS(7019),
    [anon_sym_or] = ACTIONS(7019),
    [anon_sym_DOT] = ACTIONS(7019),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7017),
    [anon_sym_new] = ACTIONS(7019),
    [anon_sym_return] = ACTIONS(7019),
    [anon_sym_return_BANG] = ACTIONS(7017),
    [anon_sym_yield] = ACTIONS(7019),
    [anon_sym_yield_BANG] = ACTIONS(7017),
    [anon_sym_lazy] = ACTIONS(7019),
    [anon_sym_assert] = ACTIONS(7019),
    [anon_sym_upcast] = ACTIONS(7019),
    [anon_sym_downcast] = ACTIONS(7019),
    [anon_sym_fixed] = ACTIONS(7019),
    [anon_sym_COLON_GT] = ACTIONS(7017),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7017),
    [anon_sym_for] = ACTIONS(7019),
    [anon_sym_while] = ACTIONS(7019),
    [anon_sym_while_BANG] = ACTIONS(7017),
    [anon_sym_if] = ACTIONS(7019),
    [anon_sym_fun] = ACTIONS(7019),
    [anon_sym_DASH_GT] = ACTIONS(7019),
    [anon_sym_try] = ACTIONS(7019),
    [anon_sym_match] = ACTIONS(7019),
    [anon_sym_match_BANG] = ACTIONS(7017),
    [anon_sym_function] = ACTIONS(7019),
    [anon_sym_DOT_LBRACK] = ACTIONS(7017),
    [anon_sym_use] = ACTIONS(7019),
    [anon_sym_use_BANG] = ACTIONS(7017),
    [anon_sym_do_BANG] = ACTIONS(7017),
    [anon_sym_DOT_DOT] = ACTIONS(7017),
    [anon_sym_LPAREN2] = ACTIONS(7019),
    [anon_sym_PIPE_PIPE] = ACTIONS(7019),
    [anon_sym_AMP_AMP] = ACTIONS(7019),
    [aux_sym_char_token1] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7019),
    [anon_sym_DQUOTE] = ACTIONS(7019),
    [anon_sym_AT_DQUOTE] = ACTIONS(7017),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7017),
    [sym_bool] = ACTIONS(7019),
    [sym_unit] = ACTIONS(7017),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7019),
    [sym_op_identifier] = ACTIONS(7019),
    [anon_sym_PLUS] = ACTIONS(7019),
    [anon_sym_DASH] = ACTIONS(7019),
    [anon_sym_PLUS_DOT] = ACTIONS(7019),
    [anon_sym_DASH_DOT] = ACTIONS(7019),
    [anon_sym_PERCENT] = ACTIONS(7019),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7019),
    [anon_sym_TILDE] = ACTIONS(7017),
    [aux_sym_prefix_op_token1] = ACTIONS(7019),
    [aux_sym_infix_op_token1] = ACTIONS(7017),
    [aux_sym_infix_op_token2] = ACTIONS(7019),
    [aux_sym_infix_op_token3] = ACTIONS(7019),
    [anon_sym_BANG_EQ] = ACTIONS(7019),
    [anon_sym_COLON_EQ] = ACTIONS(7017),
    [anon_sym_DOLLAR] = ACTIONS(7019),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7019),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7019),
    [sym_int] = ACTIONS(7019),
    [sym_xint] = ACTIONS(7017),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7017),
    [sym__newline] = ACTIONS(7017),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7017),
    [sym__tyapp_open] = ACTIONS(7017),
  },
  [STATE(5739)] = {
    [sym_block_comment] = STATE(5739),
    [sym_compiler_directive_decl] = STATE(5739),
    [sym_fsi_directive_decl] = STATE(5739),
    [sym_preproc_line] = STATE(5739),
    [sym_identifier] = ACTIONS(7023),
    [anon_sym_null] = ACTIONS(7023),
    [anon_sym_EQ] = ACTIONS(7023),
    [anon_sym_begin] = ACTIONS(7023),
    [anon_sym_COLON] = ACTIONS(7023),
    [anon_sym_do] = ACTIONS(7023),
    [anon_sym_let] = ACTIONS(7023),
    [anon_sym_let_BANG] = ACTIONS(7021),
    [anon_sym_LPAREN] = ACTIONS(7023),
    [anon_sym_COMMA] = ACTIONS(7021),
    [anon_sym_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_QMARK] = ACTIONS(7023),
    [anon_sym_COLON_COLON] = ACTIONS(7021),
    [anon_sym_AMP] = ACTIONS(7023),
    [anon_sym_LBRACK] = ACTIONS(7023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7021),
    [anon_sym_LBRACE] = ACTIONS(7023),
    [anon_sym_LT_AT] = ACTIONS(7023),
    [anon_sym_LT_AT_AT] = ACTIONS(7023),
    [anon_sym_CARET] = ACTIONS(7023),
    [anon_sym_or] = ACTIONS(7023),
    [anon_sym_DOT] = ACTIONS(7023),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7021),
    [anon_sym_new] = ACTIONS(7023),
    [anon_sym_return] = ACTIONS(7023),
    [anon_sym_return_BANG] = ACTIONS(7021),
    [anon_sym_yield] = ACTIONS(7023),
    [anon_sym_yield_BANG] = ACTIONS(7021),
    [anon_sym_lazy] = ACTIONS(7023),
    [anon_sym_assert] = ACTIONS(7023),
    [anon_sym_upcast] = ACTIONS(7023),
    [anon_sym_downcast] = ACTIONS(7023),
    [anon_sym_fixed] = ACTIONS(7023),
    [anon_sym_COLON_GT] = ACTIONS(7021),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7021),
    [anon_sym_for] = ACTIONS(7023),
    [anon_sym_while] = ACTIONS(7023),
    [anon_sym_while_BANG] = ACTIONS(7021),
    [anon_sym_if] = ACTIONS(7023),
    [anon_sym_fun] = ACTIONS(7023),
    [anon_sym_DASH_GT] = ACTIONS(7023),
    [anon_sym_try] = ACTIONS(7023),
    [anon_sym_match] = ACTIONS(7023),
    [anon_sym_match_BANG] = ACTIONS(7021),
    [anon_sym_function] = ACTIONS(7023),
    [anon_sym_DOT_LBRACK] = ACTIONS(7021),
    [anon_sym_use] = ACTIONS(7023),
    [anon_sym_use_BANG] = ACTIONS(7021),
    [anon_sym_do_BANG] = ACTIONS(7021),
    [anon_sym_DOT_DOT] = ACTIONS(7021),
    [anon_sym_LPAREN2] = ACTIONS(7023),
    [anon_sym_PIPE_PIPE] = ACTIONS(7023),
    [anon_sym_AMP_AMP] = ACTIONS(7023),
    [aux_sym_char_token1] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7023),
    [anon_sym_DQUOTE] = ACTIONS(7023),
    [anon_sym_AT_DQUOTE] = ACTIONS(7021),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7021),
    [sym_bool] = ACTIONS(7023),
    [sym_unit] = ACTIONS(7021),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7023),
    [sym_op_identifier] = ACTIONS(7023),
    [anon_sym_PLUS] = ACTIONS(7023),
    [anon_sym_DASH] = ACTIONS(7023),
    [anon_sym_PLUS_DOT] = ACTIONS(7023),
    [anon_sym_DASH_DOT] = ACTIONS(7023),
    [anon_sym_PERCENT] = ACTIONS(7023),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7023),
    [anon_sym_TILDE] = ACTIONS(7021),
    [aux_sym_prefix_op_token1] = ACTIONS(7023),
    [aux_sym_infix_op_token1] = ACTIONS(7021),
    [aux_sym_infix_op_token2] = ACTIONS(7023),
    [aux_sym_infix_op_token3] = ACTIONS(7023),
    [anon_sym_BANG_EQ] = ACTIONS(7023),
    [anon_sym_COLON_EQ] = ACTIONS(7021),
    [anon_sym_DOLLAR] = ACTIONS(7023),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7023),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7023),
    [sym_int] = ACTIONS(7023),
    [sym_xint] = ACTIONS(7021),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7021),
    [sym__newline] = ACTIONS(7021),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7021),
    [sym__tyapp_open] = ACTIONS(7021),
  },
  [STATE(5740)] = {
    [sym_block_comment] = STATE(5740),
    [sym_compiler_directive_decl] = STATE(5740),
    [sym_fsi_directive_decl] = STATE(5740),
    [sym_preproc_line] = STATE(5740),
    [sym_identifier] = ACTIONS(7027),
    [anon_sym_null] = ACTIONS(7027),
    [anon_sym_EQ] = ACTIONS(7027),
    [anon_sym_begin] = ACTIONS(7027),
    [anon_sym_COLON] = ACTIONS(7027),
    [anon_sym_do] = ACTIONS(7027),
    [anon_sym_let] = ACTIONS(7027),
    [anon_sym_let_BANG] = ACTIONS(7025),
    [anon_sym_LPAREN] = ACTIONS(7027),
    [anon_sym_COMMA] = ACTIONS(7025),
    [anon_sym_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_QMARK] = ACTIONS(7027),
    [anon_sym_COLON_COLON] = ACTIONS(7025),
    [anon_sym_AMP] = ACTIONS(7027),
    [anon_sym_LBRACK] = ACTIONS(7027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7025),
    [anon_sym_LBRACE] = ACTIONS(7027),
    [anon_sym_LT_AT] = ACTIONS(7027),
    [anon_sym_LT_AT_AT] = ACTIONS(7027),
    [anon_sym_CARET] = ACTIONS(7027),
    [anon_sym_or] = ACTIONS(7027),
    [anon_sym_DOT] = ACTIONS(7027),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7025),
    [anon_sym_new] = ACTIONS(7027),
    [anon_sym_return] = ACTIONS(7027),
    [anon_sym_return_BANG] = ACTIONS(7025),
    [anon_sym_yield] = ACTIONS(7027),
    [anon_sym_yield_BANG] = ACTIONS(7025),
    [anon_sym_lazy] = ACTIONS(7027),
    [anon_sym_assert] = ACTIONS(7027),
    [anon_sym_upcast] = ACTIONS(7027),
    [anon_sym_downcast] = ACTIONS(7027),
    [anon_sym_fixed] = ACTIONS(7027),
    [anon_sym_COLON_GT] = ACTIONS(7025),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7025),
    [anon_sym_for] = ACTIONS(7027),
    [anon_sym_while] = ACTIONS(7027),
    [anon_sym_while_BANG] = ACTIONS(7025),
    [anon_sym_if] = ACTIONS(7027),
    [anon_sym_fun] = ACTIONS(7027),
    [anon_sym_DASH_GT] = ACTIONS(7027),
    [anon_sym_try] = ACTIONS(7027),
    [anon_sym_match] = ACTIONS(7027),
    [anon_sym_match_BANG] = ACTIONS(7025),
    [anon_sym_function] = ACTIONS(7027),
    [anon_sym_DOT_LBRACK] = ACTIONS(7025),
    [anon_sym_use] = ACTIONS(7027),
    [anon_sym_use_BANG] = ACTIONS(7025),
    [anon_sym_do_BANG] = ACTIONS(7025),
    [anon_sym_DOT_DOT] = ACTIONS(7025),
    [anon_sym_LPAREN2] = ACTIONS(7027),
    [anon_sym_PIPE_PIPE] = ACTIONS(7027),
    [anon_sym_AMP_AMP] = ACTIONS(7027),
    [aux_sym_char_token1] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7027),
    [anon_sym_DQUOTE] = ACTIONS(7027),
    [anon_sym_AT_DQUOTE] = ACTIONS(7025),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7025),
    [sym_bool] = ACTIONS(7027),
    [sym_unit] = ACTIONS(7025),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7027),
    [sym_op_identifier] = ACTIONS(7027),
    [anon_sym_PLUS] = ACTIONS(7027),
    [anon_sym_DASH] = ACTIONS(7027),
    [anon_sym_PLUS_DOT] = ACTIONS(7027),
    [anon_sym_DASH_DOT] = ACTIONS(7027),
    [anon_sym_PERCENT] = ACTIONS(7027),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7027),
    [anon_sym_TILDE] = ACTIONS(7025),
    [aux_sym_prefix_op_token1] = ACTIONS(7027),
    [aux_sym_infix_op_token1] = ACTIONS(7025),
    [aux_sym_infix_op_token2] = ACTIONS(7027),
    [aux_sym_infix_op_token3] = ACTIONS(7027),
    [anon_sym_BANG_EQ] = ACTIONS(7027),
    [anon_sym_COLON_EQ] = ACTIONS(7025),
    [anon_sym_DOLLAR] = ACTIONS(7027),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7027),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7027),
    [sym_int] = ACTIONS(7027),
    [sym_xint] = ACTIONS(7025),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7025),
    [sym__newline] = ACTIONS(7025),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7025),
    [sym__tyapp_open] = ACTIONS(7025),
  },
  [STATE(5741)] = {
    [sym_block_comment] = STATE(5741),
    [sym_compiler_directive_decl] = STATE(5741),
    [sym_fsi_directive_decl] = STATE(5741),
    [sym_preproc_line] = STATE(5741),
    [sym_identifier] = ACTIONS(7035),
    [anon_sym_null] = ACTIONS(7035),
    [anon_sym_EQ] = ACTIONS(7035),
    [anon_sym_begin] = ACTIONS(7035),
    [anon_sym_COLON] = ACTIONS(7035),
    [anon_sym_do] = ACTIONS(7035),
    [anon_sym_let] = ACTIONS(7035),
    [anon_sym_let_BANG] = ACTIONS(7033),
    [anon_sym_LPAREN] = ACTIONS(7035),
    [anon_sym_COMMA] = ACTIONS(7033),
    [anon_sym_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_QMARK] = ACTIONS(7035),
    [anon_sym_COLON_COLON] = ACTIONS(7033),
    [anon_sym_AMP] = ACTIONS(7035),
    [anon_sym_LBRACK] = ACTIONS(7035),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7033),
    [anon_sym_LBRACE] = ACTIONS(7035),
    [anon_sym_LT_AT] = ACTIONS(7035),
    [anon_sym_LT_AT_AT] = ACTIONS(7035),
    [anon_sym_CARET] = ACTIONS(7035),
    [anon_sym_or] = ACTIONS(7035),
    [anon_sym_DOT] = ACTIONS(7035),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7033),
    [anon_sym_new] = ACTIONS(7035),
    [anon_sym_return] = ACTIONS(7035),
    [anon_sym_return_BANG] = ACTIONS(7033),
    [anon_sym_yield] = ACTIONS(7035),
    [anon_sym_yield_BANG] = ACTIONS(7033),
    [anon_sym_lazy] = ACTIONS(7035),
    [anon_sym_assert] = ACTIONS(7035),
    [anon_sym_upcast] = ACTIONS(7035),
    [anon_sym_downcast] = ACTIONS(7035),
    [anon_sym_fixed] = ACTIONS(7035),
    [anon_sym_COLON_GT] = ACTIONS(7033),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7033),
    [anon_sym_for] = ACTIONS(7035),
    [anon_sym_while] = ACTIONS(7035),
    [anon_sym_while_BANG] = ACTIONS(7033),
    [anon_sym_if] = ACTIONS(7035),
    [anon_sym_fun] = ACTIONS(7035),
    [anon_sym_DASH_GT] = ACTIONS(7035),
    [anon_sym_try] = ACTIONS(7035),
    [anon_sym_match] = ACTIONS(7035),
    [anon_sym_match_BANG] = ACTIONS(7033),
    [anon_sym_function] = ACTIONS(7035),
    [anon_sym_DOT_LBRACK] = ACTIONS(7033),
    [anon_sym_use] = ACTIONS(7035),
    [anon_sym_use_BANG] = ACTIONS(7033),
    [anon_sym_do_BANG] = ACTIONS(7033),
    [anon_sym_DOT_DOT] = ACTIONS(7033),
    [anon_sym_LPAREN2] = ACTIONS(7035),
    [anon_sym_PIPE_PIPE] = ACTIONS(7035),
    [anon_sym_AMP_AMP] = ACTIONS(7035),
    [aux_sym_char_token1] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7035),
    [anon_sym_DQUOTE] = ACTIONS(7035),
    [anon_sym_AT_DQUOTE] = ACTIONS(7033),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7033),
    [sym_bool] = ACTIONS(7035),
    [sym_unit] = ACTIONS(7033),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7035),
    [sym_op_identifier] = ACTIONS(7035),
    [anon_sym_PLUS] = ACTIONS(7035),
    [anon_sym_DASH] = ACTIONS(7035),
    [anon_sym_PLUS_DOT] = ACTIONS(7035),
    [anon_sym_DASH_DOT] = ACTIONS(7035),
    [anon_sym_PERCENT] = ACTIONS(7035),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7035),
    [anon_sym_TILDE] = ACTIONS(7033),
    [aux_sym_prefix_op_token1] = ACTIONS(7035),
    [aux_sym_infix_op_token1] = ACTIONS(7033),
    [aux_sym_infix_op_token2] = ACTIONS(7035),
    [aux_sym_infix_op_token3] = ACTIONS(7035),
    [anon_sym_BANG_EQ] = ACTIONS(7035),
    [anon_sym_COLON_EQ] = ACTIONS(7033),
    [anon_sym_DOLLAR] = ACTIONS(7035),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7035),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7035),
    [sym_int] = ACTIONS(7035),
    [sym_xint] = ACTIONS(7033),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7033),
    [sym__newline] = ACTIONS(7033),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7033),
    [sym__tyapp_open] = ACTIONS(7033),
  },
  [STATE(5742)] = {
    [sym_block_comment] = STATE(5742),
    [sym_compiler_directive_decl] = STATE(5742),
    [sym_fsi_directive_decl] = STATE(5742),
    [sym_preproc_line] = STATE(5742),
    [sym_identifier] = ACTIONS(7141),
    [anon_sym_null] = ACTIONS(7141),
    [anon_sym_EQ] = ACTIONS(7141),
    [anon_sym_begin] = ACTIONS(7141),
    [anon_sym_COLON] = ACTIONS(7141),
    [anon_sym_do] = ACTIONS(7141),
    [anon_sym_let] = ACTIONS(7141),
    [anon_sym_let_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN] = ACTIONS(7141),
    [anon_sym_COMMA] = ACTIONS(7143),
    [anon_sym_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_QMARK] = ACTIONS(7141),
    [anon_sym_COLON_COLON] = ACTIONS(7143),
    [anon_sym_AMP] = ACTIONS(7141),
    [anon_sym_LBRACK] = ACTIONS(7141),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7143),
    [anon_sym_LBRACE] = ACTIONS(7141),
    [anon_sym_LT_AT] = ACTIONS(7141),
    [anon_sym_LT_AT_AT] = ACTIONS(7141),
    [anon_sym_CARET] = ACTIONS(7141),
    [anon_sym_or] = ACTIONS(7141),
    [anon_sym_DOT] = ACTIONS(7141),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7143),
    [anon_sym_new] = ACTIONS(7141),
    [anon_sym_return] = ACTIONS(7141),
    [anon_sym_return_BANG] = ACTIONS(7143),
    [anon_sym_yield] = ACTIONS(7141),
    [anon_sym_yield_BANG] = ACTIONS(7143),
    [anon_sym_lazy] = ACTIONS(7141),
    [anon_sym_assert] = ACTIONS(7141),
    [anon_sym_upcast] = ACTIONS(7141),
    [anon_sym_downcast] = ACTIONS(7141),
    [anon_sym_fixed] = ACTIONS(7141),
    [anon_sym_COLON_GT] = ACTIONS(7143),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7143),
    [anon_sym_for] = ACTIONS(7141),
    [anon_sym_while] = ACTIONS(7141),
    [anon_sym_while_BANG] = ACTIONS(7143),
    [anon_sym_if] = ACTIONS(7141),
    [anon_sym_fun] = ACTIONS(7141),
    [anon_sym_try] = ACTIONS(7141),
    [anon_sym_match] = ACTIONS(7141),
    [anon_sym_match_BANG] = ACTIONS(7143),
    [anon_sym_function] = ACTIONS(7141),
    [anon_sym_DOT_LBRACK] = ACTIONS(7143),
    [anon_sym_use] = ACTIONS(7141),
    [anon_sym_use_BANG] = ACTIONS(7143),
    [anon_sym_do_BANG] = ACTIONS(7143),
    [anon_sym_LPAREN2] = ACTIONS(7141),
    [anon_sym_PIPE_PIPE] = ACTIONS(7141),
    [anon_sym_AMP_AMP] = ACTIONS(7141),
    [aux_sym_char_token1] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7141),
    [anon_sym_DQUOTE] = ACTIONS(7141),
    [anon_sym_AT_DQUOTE] = ACTIONS(7143),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7143),
    [sym_bool] = ACTIONS(7141),
    [sym_unit] = ACTIONS(7143),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7141),
    [sym_op_identifier] = ACTIONS(7141),
    [anon_sym_PLUS] = ACTIONS(7141),
    [anon_sym_DASH] = ACTIONS(7141),
    [anon_sym_PLUS_DOT] = ACTIONS(7141),
    [anon_sym_DASH_DOT] = ACTIONS(7141),
    [anon_sym_PERCENT] = ACTIONS(7141),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7141),
    [anon_sym_TILDE] = ACTIONS(7143),
    [aux_sym_prefix_op_token1] = ACTIONS(7141),
    [aux_sym_infix_op_token1] = ACTIONS(7143),
    [aux_sym_infix_op_token2] = ACTIONS(7141),
    [aux_sym_infix_op_token3] = ACTIONS(7141),
    [anon_sym_BANG_EQ] = ACTIONS(7141),
    [anon_sym_COLON_EQ] = ACTIONS(7143),
    [anon_sym_DOLLAR] = ACTIONS(7141),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7141),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7141),
    [sym_int] = ACTIONS(7141),
    [sym_xint] = ACTIONS(7143),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7143),
    [anon_sym_POUNDendif] = ACTIONS(7143),
    [anon_sym_POUNDelse] = ACTIONS(7143),
    [sym__newline] = ACTIONS(7143),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7143),
    [sym__tyapp_open] = ACTIONS(7143),
  },
  [STATE(5743)] = {
    [sym_block_comment] = STATE(5743),
    [sym_compiler_directive_decl] = STATE(5743),
    [sym_fsi_directive_decl] = STATE(5743),
    [sym_preproc_line] = STATE(5743),
    [aux_sym_sequential_expression_repeat1] = STATE(5540),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__untyped_quoted_close] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5744)] = {
    [sym_block_comment] = STATE(5744),
    [sym_compiler_directive_decl] = STATE(5744),
    [sym_fsi_directive_decl] = STATE(5744),
    [sym_preproc_line] = STATE(5744),
    [sym_identifier] = ACTIONS(7039),
    [anon_sym_null] = ACTIONS(7039),
    [anon_sym_EQ] = ACTIONS(7039),
    [anon_sym_begin] = ACTIONS(7039),
    [anon_sym_COLON] = ACTIONS(7039),
    [anon_sym_do] = ACTIONS(7039),
    [anon_sym_let] = ACTIONS(7039),
    [anon_sym_let_BANG] = ACTIONS(7037),
    [anon_sym_LPAREN] = ACTIONS(7039),
    [anon_sym_COMMA] = ACTIONS(7037),
    [anon_sym_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_QMARK] = ACTIONS(7039),
    [anon_sym_COLON_COLON] = ACTIONS(7037),
    [anon_sym_AMP] = ACTIONS(7039),
    [anon_sym_LBRACK] = ACTIONS(7039),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7037),
    [anon_sym_LBRACE] = ACTIONS(7039),
    [anon_sym_LT_AT] = ACTIONS(7039),
    [anon_sym_LT_AT_AT] = ACTIONS(7039),
    [anon_sym_CARET] = ACTIONS(7039),
    [anon_sym_or] = ACTIONS(7039),
    [anon_sym_DOT] = ACTIONS(7039),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7037),
    [anon_sym_new] = ACTIONS(7039),
    [anon_sym_return] = ACTIONS(7039),
    [anon_sym_return_BANG] = ACTIONS(7037),
    [anon_sym_yield] = ACTIONS(7039),
    [anon_sym_yield_BANG] = ACTIONS(7037),
    [anon_sym_lazy] = ACTIONS(7039),
    [anon_sym_assert] = ACTIONS(7039),
    [anon_sym_upcast] = ACTIONS(7039),
    [anon_sym_downcast] = ACTIONS(7039),
    [anon_sym_fixed] = ACTIONS(7039),
    [anon_sym_COLON_GT] = ACTIONS(7037),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7037),
    [anon_sym_for] = ACTIONS(7039),
    [anon_sym_while] = ACTIONS(7039),
    [anon_sym_while_BANG] = ACTIONS(7037),
    [anon_sym_if] = ACTIONS(7039),
    [anon_sym_fun] = ACTIONS(7039),
    [anon_sym_DASH_GT] = ACTIONS(7039),
    [anon_sym_try] = ACTIONS(7039),
    [anon_sym_match] = ACTIONS(7039),
    [anon_sym_match_BANG] = ACTIONS(7037),
    [anon_sym_function] = ACTIONS(7039),
    [anon_sym_DOT_LBRACK] = ACTIONS(7037),
    [anon_sym_use] = ACTIONS(7039),
    [anon_sym_use_BANG] = ACTIONS(7037),
    [anon_sym_do_BANG] = ACTIONS(7037),
    [anon_sym_DOT_DOT] = ACTIONS(7037),
    [anon_sym_LPAREN2] = ACTIONS(7039),
    [anon_sym_PIPE_PIPE] = ACTIONS(7039),
    [anon_sym_AMP_AMP] = ACTIONS(7039),
    [aux_sym_char_token1] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7039),
    [anon_sym_DQUOTE] = ACTIONS(7039),
    [anon_sym_AT_DQUOTE] = ACTIONS(7037),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7037),
    [sym_bool] = ACTIONS(7039),
    [sym_unit] = ACTIONS(7037),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7039),
    [sym_op_identifier] = ACTIONS(7039),
    [anon_sym_PLUS] = ACTIONS(7039),
    [anon_sym_DASH] = ACTIONS(7039),
    [anon_sym_PLUS_DOT] = ACTIONS(7039),
    [anon_sym_DASH_DOT] = ACTIONS(7039),
    [anon_sym_PERCENT] = ACTIONS(7039),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7039),
    [anon_sym_TILDE] = ACTIONS(7037),
    [aux_sym_prefix_op_token1] = ACTIONS(7039),
    [aux_sym_infix_op_token1] = ACTIONS(7037),
    [aux_sym_infix_op_token2] = ACTIONS(7039),
    [aux_sym_infix_op_token3] = ACTIONS(7039),
    [anon_sym_BANG_EQ] = ACTIONS(7039),
    [anon_sym_COLON_EQ] = ACTIONS(7037),
    [anon_sym_DOLLAR] = ACTIONS(7039),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7039),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7039),
    [sym_int] = ACTIONS(7039),
    [sym_xint] = ACTIONS(7037),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7037),
    [sym__newline] = ACTIONS(7037),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7037),
    [sym__tyapp_open] = ACTIONS(7037),
  },
  [STATE(5745)] = {
    [sym_block_comment] = STATE(5745),
    [sym_compiler_directive_decl] = STATE(5745),
    [sym_fsi_directive_decl] = STATE(5745),
    [sym_preproc_line] = STATE(5745),
    [sym_identifier] = ACTIONS(7043),
    [anon_sym_null] = ACTIONS(7043),
    [anon_sym_EQ] = ACTIONS(7043),
    [anon_sym_begin] = ACTIONS(7043),
    [anon_sym_COLON] = ACTIONS(7043),
    [anon_sym_do] = ACTIONS(7043),
    [anon_sym_let] = ACTIONS(7043),
    [anon_sym_let_BANG] = ACTIONS(7041),
    [anon_sym_LPAREN] = ACTIONS(7043),
    [anon_sym_COMMA] = ACTIONS(7041),
    [anon_sym_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_QMARK] = ACTIONS(7043),
    [anon_sym_COLON_COLON] = ACTIONS(7041),
    [anon_sym_AMP] = ACTIONS(7043),
    [anon_sym_LBRACK] = ACTIONS(7043),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7041),
    [anon_sym_LBRACE] = ACTIONS(7043),
    [anon_sym_LT_AT] = ACTIONS(7043),
    [anon_sym_LT_AT_AT] = ACTIONS(7043),
    [anon_sym_CARET] = ACTIONS(7043),
    [anon_sym_or] = ACTIONS(7043),
    [anon_sym_DOT] = ACTIONS(7043),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7041),
    [anon_sym_new] = ACTIONS(7043),
    [anon_sym_return] = ACTIONS(7043),
    [anon_sym_return_BANG] = ACTIONS(7041),
    [anon_sym_yield] = ACTIONS(7043),
    [anon_sym_yield_BANG] = ACTIONS(7041),
    [anon_sym_lazy] = ACTIONS(7043),
    [anon_sym_assert] = ACTIONS(7043),
    [anon_sym_upcast] = ACTIONS(7043),
    [anon_sym_downcast] = ACTIONS(7043),
    [anon_sym_fixed] = ACTIONS(7043),
    [anon_sym_COLON_GT] = ACTIONS(7041),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7041),
    [anon_sym_for] = ACTIONS(7043),
    [anon_sym_while] = ACTIONS(7043),
    [anon_sym_while_BANG] = ACTIONS(7041),
    [anon_sym_if] = ACTIONS(7043),
    [anon_sym_fun] = ACTIONS(7043),
    [anon_sym_DASH_GT] = ACTIONS(7043),
    [anon_sym_try] = ACTIONS(7043),
    [anon_sym_match] = ACTIONS(7043),
    [anon_sym_match_BANG] = ACTIONS(7041),
    [anon_sym_function] = ACTIONS(7043),
    [anon_sym_DOT_LBRACK] = ACTIONS(7041),
    [anon_sym_use] = ACTIONS(7043),
    [anon_sym_use_BANG] = ACTIONS(7041),
    [anon_sym_do_BANG] = ACTIONS(7041),
    [anon_sym_DOT_DOT] = ACTIONS(7041),
    [anon_sym_LPAREN2] = ACTIONS(7043),
    [anon_sym_PIPE_PIPE] = ACTIONS(7043),
    [anon_sym_AMP_AMP] = ACTIONS(7043),
    [aux_sym_char_token1] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7043),
    [anon_sym_DQUOTE] = ACTIONS(7043),
    [anon_sym_AT_DQUOTE] = ACTIONS(7041),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7041),
    [sym_bool] = ACTIONS(7043),
    [sym_unit] = ACTIONS(7041),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7043),
    [sym_op_identifier] = ACTIONS(7043),
    [anon_sym_PLUS] = ACTIONS(7043),
    [anon_sym_DASH] = ACTIONS(7043),
    [anon_sym_PLUS_DOT] = ACTIONS(7043),
    [anon_sym_DASH_DOT] = ACTIONS(7043),
    [anon_sym_PERCENT] = ACTIONS(7043),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7043),
    [anon_sym_TILDE] = ACTIONS(7041),
    [aux_sym_prefix_op_token1] = ACTIONS(7043),
    [aux_sym_infix_op_token1] = ACTIONS(7041),
    [aux_sym_infix_op_token2] = ACTIONS(7043),
    [aux_sym_infix_op_token3] = ACTIONS(7043),
    [anon_sym_BANG_EQ] = ACTIONS(7043),
    [anon_sym_COLON_EQ] = ACTIONS(7041),
    [anon_sym_DOLLAR] = ACTIONS(7043),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7043),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7043),
    [sym_int] = ACTIONS(7043),
    [sym_xint] = ACTIONS(7041),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7041),
    [sym__newline] = ACTIONS(7041),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7041),
    [sym__tyapp_open] = ACTIONS(7041),
  },
  [STATE(5746)] = {
    [sym_block_comment] = STATE(5746),
    [sym_compiler_directive_decl] = STATE(5746),
    [sym_fsi_directive_decl] = STATE(5746),
    [sym_preproc_line] = STATE(5746),
    [sym_identifier] = ACTIONS(7145),
    [anon_sym_null] = ACTIONS(7145),
    [anon_sym_EQ] = ACTIONS(7145),
    [anon_sym_begin] = ACTIONS(7145),
    [anon_sym_COLON] = ACTIONS(7145),
    [anon_sym_do] = ACTIONS(7145),
    [anon_sym_let] = ACTIONS(7145),
    [anon_sym_let_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN] = ACTIONS(7145),
    [anon_sym_COMMA] = ACTIONS(7147),
    [anon_sym_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_QMARK] = ACTIONS(7145),
    [anon_sym_COLON_COLON] = ACTIONS(7147),
    [anon_sym_AMP] = ACTIONS(7145),
    [anon_sym_LBRACK] = ACTIONS(7145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7147),
    [anon_sym_LBRACE] = ACTIONS(7145),
    [anon_sym_LT_AT] = ACTIONS(7145),
    [anon_sym_LT_AT_AT] = ACTIONS(7145),
    [anon_sym_CARET] = ACTIONS(7145),
    [anon_sym_or] = ACTIONS(7145),
    [anon_sym_DOT] = ACTIONS(7145),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7147),
    [anon_sym_new] = ACTIONS(7145),
    [anon_sym_return] = ACTIONS(7145),
    [anon_sym_return_BANG] = ACTIONS(7147),
    [anon_sym_yield] = ACTIONS(7145),
    [anon_sym_yield_BANG] = ACTIONS(7147),
    [anon_sym_lazy] = ACTIONS(7145),
    [anon_sym_assert] = ACTIONS(7145),
    [anon_sym_upcast] = ACTIONS(7145),
    [anon_sym_downcast] = ACTIONS(7145),
    [anon_sym_fixed] = ACTIONS(7145),
    [anon_sym_COLON_GT] = ACTIONS(7147),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7147),
    [anon_sym_for] = ACTIONS(7145),
    [anon_sym_while] = ACTIONS(7145),
    [anon_sym_while_BANG] = ACTIONS(7147),
    [anon_sym_if] = ACTIONS(7145),
    [anon_sym_fun] = ACTIONS(7145),
    [anon_sym_try] = ACTIONS(7145),
    [anon_sym_match] = ACTIONS(7145),
    [anon_sym_match_BANG] = ACTIONS(7147),
    [anon_sym_function] = ACTIONS(7145),
    [anon_sym_DOT_LBRACK] = ACTIONS(7147),
    [anon_sym_use] = ACTIONS(7145),
    [anon_sym_use_BANG] = ACTIONS(7147),
    [anon_sym_do_BANG] = ACTIONS(7147),
    [anon_sym_LPAREN2] = ACTIONS(7145),
    [anon_sym_PIPE_PIPE] = ACTIONS(7145),
    [anon_sym_AMP_AMP] = ACTIONS(7145),
    [aux_sym_char_token1] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7145),
    [anon_sym_DQUOTE] = ACTIONS(7145),
    [anon_sym_AT_DQUOTE] = ACTIONS(7147),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7147),
    [sym_bool] = ACTIONS(7145),
    [sym_unit] = ACTIONS(7147),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7145),
    [sym_op_identifier] = ACTIONS(7145),
    [anon_sym_PLUS] = ACTIONS(7145),
    [anon_sym_DASH] = ACTIONS(7145),
    [anon_sym_PLUS_DOT] = ACTIONS(7145),
    [anon_sym_DASH_DOT] = ACTIONS(7145),
    [anon_sym_PERCENT] = ACTIONS(7145),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7145),
    [anon_sym_TILDE] = ACTIONS(7147),
    [aux_sym_prefix_op_token1] = ACTIONS(7145),
    [aux_sym_infix_op_token1] = ACTIONS(7147),
    [aux_sym_infix_op_token2] = ACTIONS(7145),
    [aux_sym_infix_op_token3] = ACTIONS(7145),
    [anon_sym_BANG_EQ] = ACTIONS(7145),
    [anon_sym_COLON_EQ] = ACTIONS(7147),
    [anon_sym_DOLLAR] = ACTIONS(7145),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7145),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7145),
    [sym_int] = ACTIONS(7145),
    [sym_xint] = ACTIONS(7147),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7147),
    [anon_sym_POUNDendif] = ACTIONS(7147),
    [anon_sym_POUNDelse] = ACTIONS(7147),
    [sym__newline] = ACTIONS(7147),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7147),
    [sym__tyapp_open] = ACTIONS(7147),
  },
  [STATE(5747)] = {
    [sym_block_comment] = STATE(5747),
    [sym_compiler_directive_decl] = STATE(5747),
    [sym_fsi_directive_decl] = STATE(5747),
    [sym_preproc_line] = STATE(5747),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__untyped_quoted_close] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5748)] = {
    [sym_block_comment] = STATE(5748),
    [sym_compiler_directive_decl] = STATE(5748),
    [sym_fsi_directive_decl] = STATE(5748),
    [sym_preproc_line] = STATE(5748),
    [sym_identifier] = ACTIONS(7059),
    [anon_sym_null] = ACTIONS(7059),
    [anon_sym_EQ] = ACTIONS(7059),
    [anon_sym_begin] = ACTIONS(7059),
    [anon_sym_COLON] = ACTIONS(7059),
    [anon_sym_do] = ACTIONS(7059),
    [anon_sym_let] = ACTIONS(7059),
    [anon_sym_let_BANG] = ACTIONS(7057),
    [anon_sym_LPAREN] = ACTIONS(7059),
    [anon_sym_COMMA] = ACTIONS(7057),
    [anon_sym_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_QMARK] = ACTIONS(7059),
    [anon_sym_COLON_COLON] = ACTIONS(7057),
    [anon_sym_AMP] = ACTIONS(7059),
    [anon_sym_LBRACK] = ACTIONS(7059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7057),
    [anon_sym_LBRACE] = ACTIONS(7059),
    [anon_sym_LT_AT] = ACTIONS(7059),
    [anon_sym_LT_AT_AT] = ACTIONS(7059),
    [anon_sym_CARET] = ACTIONS(7059),
    [anon_sym_or] = ACTIONS(7059),
    [anon_sym_DOT] = ACTIONS(7059),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7057),
    [anon_sym_new] = ACTIONS(7059),
    [anon_sym_return] = ACTIONS(7059),
    [anon_sym_return_BANG] = ACTIONS(7057),
    [anon_sym_yield] = ACTIONS(7059),
    [anon_sym_yield_BANG] = ACTIONS(7057),
    [anon_sym_lazy] = ACTIONS(7059),
    [anon_sym_assert] = ACTIONS(7059),
    [anon_sym_upcast] = ACTIONS(7059),
    [anon_sym_downcast] = ACTIONS(7059),
    [anon_sym_fixed] = ACTIONS(7059),
    [anon_sym_COLON_GT] = ACTIONS(7057),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7057),
    [anon_sym_for] = ACTIONS(7059),
    [anon_sym_while] = ACTIONS(7059),
    [anon_sym_while_BANG] = ACTIONS(7057),
    [anon_sym_if] = ACTIONS(7059),
    [anon_sym_fun] = ACTIONS(7059),
    [anon_sym_DASH_GT] = ACTIONS(7059),
    [anon_sym_try] = ACTIONS(7059),
    [anon_sym_match] = ACTIONS(7059),
    [anon_sym_match_BANG] = ACTIONS(7057),
    [anon_sym_function] = ACTIONS(7059),
    [anon_sym_DOT_LBRACK] = ACTIONS(7057),
    [anon_sym_use] = ACTIONS(7059),
    [anon_sym_use_BANG] = ACTIONS(7057),
    [anon_sym_do_BANG] = ACTIONS(7057),
    [anon_sym_DOT_DOT] = ACTIONS(7057),
    [anon_sym_LPAREN2] = ACTIONS(7059),
    [anon_sym_PIPE_PIPE] = ACTIONS(7059),
    [anon_sym_AMP_AMP] = ACTIONS(7059),
    [aux_sym_char_token1] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7059),
    [anon_sym_DQUOTE] = ACTIONS(7059),
    [anon_sym_AT_DQUOTE] = ACTIONS(7057),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7057),
    [sym_bool] = ACTIONS(7059),
    [sym_unit] = ACTIONS(7057),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7059),
    [sym_op_identifier] = ACTIONS(7059),
    [anon_sym_PLUS] = ACTIONS(7059),
    [anon_sym_DASH] = ACTIONS(7059),
    [anon_sym_PLUS_DOT] = ACTIONS(7059),
    [anon_sym_DASH_DOT] = ACTIONS(7059),
    [anon_sym_PERCENT] = ACTIONS(7059),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7059),
    [anon_sym_TILDE] = ACTIONS(7057),
    [aux_sym_prefix_op_token1] = ACTIONS(7059),
    [aux_sym_infix_op_token1] = ACTIONS(7057),
    [aux_sym_infix_op_token2] = ACTIONS(7059),
    [aux_sym_infix_op_token3] = ACTIONS(7059),
    [anon_sym_BANG_EQ] = ACTIONS(7059),
    [anon_sym_COLON_EQ] = ACTIONS(7057),
    [anon_sym_DOLLAR] = ACTIONS(7059),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7059),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7059),
    [sym_int] = ACTIONS(7059),
    [sym_xint] = ACTIONS(7057),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7057),
    [sym__newline] = ACTIONS(7057),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7057),
    [sym__tyapp_open] = ACTIONS(7057),
  },
  [STATE(5749)] = {
    [sym_block_comment] = STATE(5749),
    [sym_compiler_directive_decl] = STATE(5749),
    [sym_fsi_directive_decl] = STATE(5749),
    [sym_preproc_line] = STATE(5749),
    [sym_identifier] = ACTIONS(6420),
    [anon_sym_module] = ACTIONS(6420),
    [anon_sym_null] = ACTIONS(6420),
    [anon_sym_begin] = ACTIONS(6420),
    [anon_sym_open] = ACTIONS(6420),
    [anon_sym_type] = ACTIONS(6420),
    [anon_sym_LBRACK_LT] = ACTIONS(6418),
    [anon_sym_do] = ACTIONS(6420),
    [anon_sym_and] = ACTIONS(6420),
    [anon_sym_let] = ACTIONS(6420),
    [anon_sym_let_BANG] = ACTIONS(6418),
    [aux_sym_access_modifier_token1] = ACTIONS(6418),
    [anon_sym_LPAREN] = ACTIONS(6420),
    [anon_sym_PIPE] = ACTIONS(6418),
    [anon_sym_AMP] = ACTIONS(6420),
    [anon_sym_LBRACK] = ACTIONS(6420),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6418),
    [anon_sym_LBRACE] = ACTIONS(6420),
    [anon_sym_LT_AT] = ACTIONS(6420),
    [anon_sym_LT_AT_AT] = ACTIONS(6418),
    [anon_sym_CARET] = ACTIONS(6418),
    [anon_sym_or] = ACTIONS(6420),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6418),
    [anon_sym_new] = ACTIONS(6420),
    [anon_sym_with] = ACTIONS(6420),
    [anon_sym_return] = ACTIONS(6420),
    [anon_sym_return_BANG] = ACTIONS(6418),
    [anon_sym_yield] = ACTIONS(6420),
    [anon_sym_yield_BANG] = ACTIONS(6418),
    [anon_sym_lazy] = ACTIONS(6420),
    [anon_sym_assert] = ACTIONS(6420),
    [anon_sym_upcast] = ACTIONS(6420),
    [anon_sym_downcast] = ACTIONS(6420),
    [anon_sym_fixed] = ACTIONS(6420),
    [anon_sym_COLON_GT] = ACTIONS(6418),
    [anon_sym_for] = ACTIONS(6420),
    [anon_sym_while] = ACTIONS(6420),
    [anon_sym_while_BANG] = ACTIONS(6418),
    [anon_sym_if] = ACTIONS(6420),
    [anon_sym_fun] = ACTIONS(6420),
    [anon_sym_DASH_GT] = ACTIONS(6418),
    [anon_sym_try] = ACTIONS(6420),
    [anon_sym_match] = ACTIONS(6420),
    [anon_sym_match_BANG] = ACTIONS(6418),
    [anon_sym_function] = ACTIONS(6420),
    [anon_sym_use] = ACTIONS(6420),
    [anon_sym_use_BANG] = ACTIONS(6418),
    [anon_sym_do_BANG] = ACTIONS(6418),
    [anon_sym_AMP_AMP] = ACTIONS(6418),
    [anon_sym_STAR] = ACTIONS(6418),
    [anon_sym_LT] = ACTIONS(6420),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6418),
    [anon_sym_default] = ACTIONS(6420),
    [anon_sym_static] = ACTIONS(6420),
    [anon_sym_member] = ACTIONS(6420),
    [anon_sym_exception] = ACTIONS(6420),
    [anon_sym_interface] = ACTIONS(6420),
    [anon_sym_abstract] = ACTIONS(6420),
    [anon_sym_val] = ACTIONS(6420),
    [anon_sym_override] = ACTIONS(6420),
    [anon_sym_extern] = ACTIONS(6420),
    [aux_sym_char_token1] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6420),
    [anon_sym_DQUOTE] = ACTIONS(6420),
    [anon_sym_AT_DQUOTE] = ACTIONS(6418),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6418),
    [sym_bool] = ACTIONS(6420),
    [sym_unit] = ACTIONS(6418),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6420),
    [sym_op_identifier] = ACTIONS(6418),
    [anon_sym_PLUS] = ACTIONS(6420),
    [anon_sym_DASH] = ACTIONS(6420),
    [anon_sym_PLUS_DOT] = ACTIONS(6418),
    [anon_sym_DASH_DOT] = ACTIONS(6418),
    [anon_sym_PERCENT] = ACTIONS(6420),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6418),
    [anon_sym_TILDE] = ACTIONS(6418),
    [aux_sym_prefix_op_token1] = ACTIONS(6418),
    [sym_int] = ACTIONS(6420),
    [sym_xint] = ACTIONS(6418),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6418),
    [anon_sym_POUNDload] = ACTIONS(6418),
    [anon_sym_POUNDtime] = ACTIONS(6418),
    [anon_sym_POUNDI] = ACTIONS(6418),
    [anon_sym_POUNDhelp] = ACTIONS(6418),
    [anon_sym_POUNDquit] = ACTIONS(6418),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6418),
    [anon_sym_POUNDendif] = ACTIONS(6418),
    [anon_sym_POUNDelse] = ACTIONS(6418),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6418),
  },
  [STATE(5750)] = {
    [sym_block_comment] = STATE(5750),
    [sym_compiler_directive_decl] = STATE(5750),
    [sym_fsi_directive_decl] = STATE(5750),
    [sym_preproc_line] = STATE(5750),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_with] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5751)] = {
    [sym_block_comment] = STATE(5751),
    [sym_compiler_directive_decl] = STATE(5751),
    [sym_fsi_directive_decl] = STATE(5751),
    [sym_preproc_line] = STATE(5751),
    [aux_sym_long_identifier_repeat1] = STATE(5344),
    [ts_builtin_sym_end] = ACTIONS(6291),
    [sym_identifier] = ACTIONS(6289),
    [anon_sym_namespace] = ACTIONS(6289),
    [anon_sym_module] = ACTIONS(6289),
    [anon_sym_null] = ACTIONS(6289),
    [anon_sym_begin] = ACTIONS(6289),
    [anon_sym_open] = ACTIONS(6289),
    [anon_sym_type] = ACTIONS(6289),
    [anon_sym_LBRACK_LT] = ACTIONS(6291),
    [anon_sym_do] = ACTIONS(6289),
    [anon_sym_and] = ACTIONS(6289),
    [anon_sym_let] = ACTIONS(6289),
    [anon_sym_let_BANG] = ACTIONS(6291),
    [aux_sym_access_modifier_token1] = ACTIONS(6291),
    [anon_sym_LPAREN] = ACTIONS(6289),
    [anon_sym_PIPE] = ACTIONS(6291),
    [anon_sym_AMP] = ACTIONS(6289),
    [anon_sym_LBRACK] = ACTIONS(6289),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6291),
    [anon_sym_LBRACE] = ACTIONS(6289),
    [anon_sym_LT_AT] = ACTIONS(6289),
    [anon_sym_LT_AT_AT] = ACTIONS(6291),
    [anon_sym_CARET] = ACTIONS(6291),
    [anon_sym_DOT] = ACTIONS(9628),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6291),
    [anon_sym_new] = ACTIONS(6289),
    [anon_sym_with] = ACTIONS(6289),
    [anon_sym_return] = ACTIONS(6289),
    [anon_sym_return_BANG] = ACTIONS(6291),
    [anon_sym_yield] = ACTIONS(6289),
    [anon_sym_yield_BANG] = ACTIONS(6291),
    [anon_sym_lazy] = ACTIONS(6289),
    [anon_sym_assert] = ACTIONS(6289),
    [anon_sym_upcast] = ACTIONS(6289),
    [anon_sym_downcast] = ACTIONS(6289),
    [anon_sym_fixed] = ACTIONS(6289),
    [anon_sym_for] = ACTIONS(6289),
    [anon_sym_while] = ACTIONS(6289),
    [anon_sym_while_BANG] = ACTIONS(6291),
    [anon_sym_if] = ACTIONS(6289),
    [anon_sym_fun] = ACTIONS(6289),
    [anon_sym_DASH_GT] = ACTIONS(6291),
    [anon_sym_try] = ACTIONS(6289),
    [anon_sym_match] = ACTIONS(6289),
    [anon_sym_match_BANG] = ACTIONS(6291),
    [anon_sym_function] = ACTIONS(6289),
    [anon_sym_use] = ACTIONS(6289),
    [anon_sym_use_BANG] = ACTIONS(6291),
    [anon_sym_do_BANG] = ACTIONS(6291),
    [anon_sym_AMP_AMP] = ACTIONS(6291),
    [anon_sym_STAR] = ACTIONS(6291),
    [anon_sym_LT] = ACTIONS(6289),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6291),
    [anon_sym_default] = ACTIONS(6289),
    [anon_sym_static] = ACTIONS(6289),
    [anon_sym_member] = ACTIONS(6289),
    [anon_sym_exception] = ACTIONS(6289),
    [anon_sym_interface] = ACTIONS(6289),
    [anon_sym_abstract] = ACTIONS(6289),
    [anon_sym_val] = ACTIONS(6289),
    [anon_sym_override] = ACTIONS(6289),
    [anon_sym_extern] = ACTIONS(6289),
    [aux_sym_char_token1] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6289),
    [anon_sym_DQUOTE] = ACTIONS(6289),
    [anon_sym_AT_DQUOTE] = ACTIONS(6291),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6291),
    [sym_bool] = ACTIONS(6289),
    [sym_unit] = ACTIONS(6291),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6289),
    [sym_op_identifier] = ACTIONS(6291),
    [anon_sym_PLUS] = ACTIONS(6289),
    [anon_sym_DASH] = ACTIONS(6289),
    [anon_sym_PLUS_DOT] = ACTIONS(6291),
    [anon_sym_DASH_DOT] = ACTIONS(6291),
    [anon_sym_PERCENT] = ACTIONS(6289),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6291),
    [anon_sym_TILDE] = ACTIONS(6291),
    [aux_sym_prefix_op_token1] = ACTIONS(6291),
    [sym_int] = ACTIONS(6289),
    [sym_xint] = ACTIONS(6291),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6291),
    [anon_sym_POUNDload] = ACTIONS(6291),
    [anon_sym_POUNDtime] = ACTIONS(6291),
    [anon_sym_POUNDI] = ACTIONS(6291),
    [anon_sym_POUNDhelp] = ACTIONS(6291),
    [anon_sym_POUNDquit] = ACTIONS(6291),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6291),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6291),
  },
  [STATE(5752)] = {
    [sym_block_comment] = STATE(5752),
    [sym_compiler_directive_decl] = STATE(5752),
    [sym_fsi_directive_decl] = STATE(5752),
    [sym_preproc_line] = STATE(5752),
    [sym_identifier] = ACTIONS(6807),
    [anon_sym_null] = ACTIONS(6807),
    [anon_sym_EQ] = ACTIONS(6807),
    [anon_sym_begin] = ACTIONS(6807),
    [anon_sym_COLON] = ACTIONS(6807),
    [anon_sym_do] = ACTIONS(6807),
    [anon_sym_let] = ACTIONS(6807),
    [anon_sym_let_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN] = ACTIONS(6807),
    [anon_sym_COMMA] = ACTIONS(6805),
    [anon_sym_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_QMARK] = ACTIONS(6807),
    [anon_sym_COLON_COLON] = ACTIONS(6805),
    [anon_sym_AMP] = ACTIONS(6807),
    [anon_sym_LBRACK] = ACTIONS(6807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6805),
    [anon_sym_LBRACE] = ACTIONS(6807),
    [anon_sym_LT_AT] = ACTIONS(6807),
    [anon_sym_LT_AT_AT] = ACTIONS(6807),
    [anon_sym_CARET] = ACTIONS(6807),
    [anon_sym_or] = ACTIONS(6807),
    [anon_sym_DOT] = ACTIONS(6807),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6805),
    [anon_sym_new] = ACTIONS(6807),
    [anon_sym_return] = ACTIONS(6807),
    [anon_sym_return_BANG] = ACTIONS(6805),
    [anon_sym_yield] = ACTIONS(6807),
    [anon_sym_yield_BANG] = ACTIONS(6805),
    [anon_sym_lazy] = ACTIONS(6807),
    [anon_sym_assert] = ACTIONS(6807),
    [anon_sym_upcast] = ACTIONS(6807),
    [anon_sym_downcast] = ACTIONS(6807),
    [anon_sym_fixed] = ACTIONS(6807),
    [anon_sym_COLON_GT] = ACTIONS(6805),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6805),
    [anon_sym_for] = ACTIONS(6807),
    [anon_sym_while] = ACTIONS(6807),
    [anon_sym_while_BANG] = ACTIONS(6805),
    [anon_sym_if] = ACTIONS(6807),
    [anon_sym_fun] = ACTIONS(6807),
    [anon_sym_try] = ACTIONS(6807),
    [anon_sym_match] = ACTIONS(6807),
    [anon_sym_match_BANG] = ACTIONS(6805),
    [anon_sym_function] = ACTIONS(6807),
    [anon_sym_LT_DASH] = ACTIONS(6809),
    [anon_sym_DOT_LBRACK] = ACTIONS(6805),
    [anon_sym_use] = ACTIONS(6807),
    [anon_sym_use_BANG] = ACTIONS(6805),
    [anon_sym_do_BANG] = ACTIONS(6805),
    [anon_sym_LPAREN2] = ACTIONS(6807),
    [anon_sym_PIPE_PIPE] = ACTIONS(6807),
    [anon_sym_AMP_AMP] = ACTIONS(6807),
    [aux_sym_char_token1] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6807),
    [anon_sym_DQUOTE] = ACTIONS(6807),
    [anon_sym_AT_DQUOTE] = ACTIONS(6805),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6805),
    [sym_bool] = ACTIONS(6807),
    [sym_unit] = ACTIONS(6805),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6807),
    [sym_op_identifier] = ACTIONS(6807),
    [anon_sym_PLUS] = ACTIONS(6807),
    [anon_sym_DASH] = ACTIONS(6807),
    [anon_sym_PLUS_DOT] = ACTIONS(6807),
    [anon_sym_DASH_DOT] = ACTIONS(6807),
    [anon_sym_PERCENT] = ACTIONS(6807),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6807),
    [anon_sym_TILDE] = ACTIONS(6805),
    [aux_sym_prefix_op_token1] = ACTIONS(6807),
    [aux_sym_infix_op_token1] = ACTIONS(6805),
    [aux_sym_infix_op_token2] = ACTIONS(6807),
    [aux_sym_infix_op_token3] = ACTIONS(6807),
    [anon_sym_BANG_EQ] = ACTIONS(6807),
    [anon_sym_COLON_EQ] = ACTIONS(6805),
    [anon_sym_DOLLAR] = ACTIONS(6807),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6807),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6807),
    [sym_int] = ACTIONS(6807),
    [sym_xint] = ACTIONS(6805),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6805),
    [anon_sym_POUNDendif] = ACTIONS(6805),
    [sym__newline] = ACTIONS(6805),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6805),
    [sym__tyapp_open] = ACTIONS(6805),
  },
  [STATE(5753)] = {
    [sym_block_comment] = STATE(5753),
    [sym_compiler_directive_decl] = STATE(5753),
    [sym_fsi_directive_decl] = STATE(5753),
    [sym_preproc_line] = STATE(5753),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__multi_dollar_interp_end] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5754)] = {
    [sym_block_comment] = STATE(5754),
    [sym_compiler_directive_decl] = STATE(5754),
    [sym_fsi_directive_decl] = STATE(5754),
    [sym_preproc_line] = STATE(5754),
    [sym_identifier] = ACTIONS(5951),
    [anon_sym_null] = ACTIONS(5951),
    [anon_sym_EQ] = ACTIONS(5951),
    [anon_sym_begin] = ACTIONS(5951),
    [anon_sym_COLON] = ACTIONS(5951),
    [anon_sym_do] = ACTIONS(5951),
    [anon_sym_let] = ACTIONS(5951),
    [anon_sym_let_BANG] = ACTIONS(5953),
    [anon_sym_as] = ACTIONS(5951),
    [anon_sym_LPAREN] = ACTIONS(5951),
    [anon_sym_COMMA] = ACTIONS(5953),
    [anon_sym_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_QMARK] = ACTIONS(5951),
    [anon_sym_COLON_COLON] = ACTIONS(5953),
    [anon_sym_AMP] = ACTIONS(5951),
    [anon_sym_LBRACK] = ACTIONS(5951),
    [anon_sym_LBRACK_PIPE] = ACTIONS(5953),
    [anon_sym_LBRACE] = ACTIONS(5951),
    [anon_sym_LT_AT] = ACTIONS(5951),
    [anon_sym_LT_AT_AT] = ACTIONS(5951),
    [anon_sym_CARET] = ACTIONS(5951),
    [anon_sym_or] = ACTIONS(5951),
    [anon_sym_DOT] = ACTIONS(5951),
    [anon_sym_LBRACE_PIPE] = ACTIONS(5953),
    [anon_sym_new] = ACTIONS(5951),
    [anon_sym_with] = ACTIONS(5951),
    [anon_sym_return] = ACTIONS(5951),
    [anon_sym_return_BANG] = ACTIONS(5953),
    [anon_sym_yield] = ACTIONS(5951),
    [anon_sym_yield_BANG] = ACTIONS(5953),
    [anon_sym_lazy] = ACTIONS(5951),
    [anon_sym_assert] = ACTIONS(5951),
    [anon_sym_upcast] = ACTIONS(5951),
    [anon_sym_downcast] = ACTIONS(5951),
    [anon_sym_fixed] = ACTIONS(5951),
    [anon_sym_COLON_GT] = ACTIONS(5953),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(5953),
    [anon_sym_for] = ACTIONS(5951),
    [anon_sym_while] = ACTIONS(5951),
    [anon_sym_while_BANG] = ACTIONS(5953),
    [anon_sym_if] = ACTIONS(5951),
    [anon_sym_fun] = ACTIONS(5951),
    [anon_sym_try] = ACTIONS(5951),
    [anon_sym_match] = ACTIONS(5951),
    [anon_sym_match_BANG] = ACTIONS(5953),
    [anon_sym_function] = ACTIONS(5951),
    [anon_sym_DOT_LBRACK] = ACTIONS(5953),
    [anon_sym_use] = ACTIONS(5951),
    [anon_sym_use_BANG] = ACTIONS(5953),
    [anon_sym_do_BANG] = ACTIONS(5953),
    [anon_sym_LPAREN2] = ACTIONS(5951),
    [anon_sym_PIPE_PIPE] = ACTIONS(5951),
    [anon_sym_AMP_AMP] = ACTIONS(5951),
    [aux_sym_char_token1] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(5951),
    [anon_sym_DQUOTE] = ACTIONS(5951),
    [anon_sym_AT_DQUOTE] = ACTIONS(5953),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(5953),
    [sym_bool] = ACTIONS(5951),
    [sym_unit] = ACTIONS(5953),
    [anon_sym_LPAREN_PIPE] = ACTIONS(5951),
    [sym_op_identifier] = ACTIONS(5951),
    [anon_sym_PLUS] = ACTIONS(5951),
    [anon_sym_DASH] = ACTIONS(5951),
    [anon_sym_PLUS_DOT] = ACTIONS(5951),
    [anon_sym_DASH_DOT] = ACTIONS(5951),
    [anon_sym_PERCENT] = ACTIONS(5951),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(5951),
    [anon_sym_TILDE] = ACTIONS(5953),
    [aux_sym_prefix_op_token1] = ACTIONS(5951),
    [aux_sym_infix_op_token1] = ACTIONS(5953),
    [aux_sym_infix_op_token2] = ACTIONS(5951),
    [aux_sym_infix_op_token3] = ACTIONS(5951),
    [anon_sym_BANG_EQ] = ACTIONS(5951),
    [anon_sym_COLON_EQ] = ACTIONS(5953),
    [anon_sym_DOLLAR] = ACTIONS(5951),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(5951),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(5951),
    [sym_int] = ACTIONS(5951),
    [sym_xint] = ACTIONS(5953),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(5953),
    [sym__newline] = ACTIONS(5953),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(5953),
    [sym__tyapp_open] = ACTIONS(5953),
  },
  [STATE(5755)] = {
    [sym_block_comment] = STATE(5755),
    [sym_compiler_directive_decl] = STATE(5755),
    [sym_fsi_directive_decl] = STATE(5755),
    [sym_preproc_line] = STATE(5755),
    [sym_identifier] = ACTIONS(290),
    [anon_sym_null] = ACTIONS(290),
    [anon_sym_EQ] = ACTIONS(290),
    [anon_sym_begin] = ACTIONS(290),
    [anon_sym_COLON] = ACTIONS(290),
    [anon_sym_do] = ACTIONS(290),
    [anon_sym_let] = ACTIONS(290),
    [anon_sym_let_BANG] = ACTIONS(292),
    [anon_sym_as] = ACTIONS(290),
    [anon_sym_LPAREN] = ACTIONS(290),
    [anon_sym_COMMA] = ACTIONS(292),
    [anon_sym_QMARK] = ACTIONS(290),
    [anon_sym_COLON_QMARK] = ACTIONS(290),
    [anon_sym_COLON_COLON] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(290),
    [anon_sym_LBRACK] = ACTIONS(290),
    [anon_sym_LBRACK_PIPE] = ACTIONS(292),
    [anon_sym_LBRACE] = ACTIONS(290),
    [anon_sym_LT_AT] = ACTIONS(290),
    [anon_sym_LT_AT_AT] = ACTIONS(290),
    [anon_sym_CARET] = ACTIONS(290),
    [anon_sym_or] = ACTIONS(290),
    [anon_sym_DOT] = ACTIONS(290),
    [anon_sym_LBRACE_PIPE] = ACTIONS(292),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_with] = ACTIONS(290),
    [anon_sym_return] = ACTIONS(290),
    [anon_sym_return_BANG] = ACTIONS(292),
    [anon_sym_yield] = ACTIONS(290),
    [anon_sym_yield_BANG] = ACTIONS(292),
    [anon_sym_lazy] = ACTIONS(290),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_upcast] = ACTIONS(290),
    [anon_sym_downcast] = ACTIONS(290),
    [anon_sym_fixed] = ACTIONS(290),
    [anon_sym_COLON_GT] = ACTIONS(292),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(292),
    [anon_sym_for] = ACTIONS(290),
    [anon_sym_while] = ACTIONS(290),
    [anon_sym_while_BANG] = ACTIONS(292),
    [anon_sym_if] = ACTIONS(290),
    [anon_sym_fun] = ACTIONS(290),
    [anon_sym_try] = ACTIONS(290),
    [anon_sym_match] = ACTIONS(290),
    [anon_sym_match_BANG] = ACTIONS(292),
    [anon_sym_function] = ACTIONS(290),
    [anon_sym_DOT_LBRACK] = ACTIONS(292),
    [anon_sym_use] = ACTIONS(290),
    [anon_sym_use_BANG] = ACTIONS(292),
    [anon_sym_do_BANG] = ACTIONS(292),
    [anon_sym_LPAREN2] = ACTIONS(290),
    [anon_sym_PIPE_PIPE] = ACTIONS(290),
    [anon_sym_AMP_AMP] = ACTIONS(290),
    [aux_sym_char_token1] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(290),
    [anon_sym_DQUOTE] = ACTIONS(290),
    [anon_sym_AT_DQUOTE] = ACTIONS(292),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(292),
    [sym_bool] = ACTIONS(290),
    [sym_unit] = ACTIONS(292),
    [anon_sym_LPAREN_PIPE] = ACTIONS(290),
    [sym_op_identifier] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(290),
    [anon_sym_DASH] = ACTIONS(290),
    [anon_sym_PLUS_DOT] = ACTIONS(290),
    [anon_sym_DASH_DOT] = ACTIONS(290),
    [anon_sym_PERCENT] = ACTIONS(290),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(290),
    [anon_sym_TILDE] = ACTIONS(292),
    [aux_sym_prefix_op_token1] = ACTIONS(290),
    [aux_sym_infix_op_token1] = ACTIONS(292),
    [aux_sym_infix_op_token2] = ACTIONS(290),
    [aux_sym_infix_op_token3] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_COLON_EQ] = ACTIONS(292),
    [anon_sym_DOLLAR] = ACTIONS(290),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(290),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(290),
    [sym_int] = ACTIONS(290),
    [sym_xint] = ACTIONS(292),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(292),
    [sym__newline] = ACTIONS(292),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(292),
    [sym__tyapp_open] = ACTIONS(292),
  },
  [STATE(5756)] = {
    [sym_block_comment] = STATE(5756),
    [sym_compiler_directive_decl] = STATE(5756),
    [sym_fsi_directive_decl] = STATE(5756),
    [sym_preproc_line] = STATE(5756),
    [aux_sym_sequential_expression_repeat1] = STATE(5756),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_with] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9842),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5757)] = {
    [sym_block_comment] = STATE(5757),
    [sym_compiler_directive_decl] = STATE(5757),
    [sym_fsi_directive_decl] = STATE(5757),
    [sym_preproc_line] = STATE(5757),
    [sym_identifier] = ACTIONS(6932),
    [anon_sym_null] = ACTIONS(6932),
    [anon_sym_EQ] = ACTIONS(6932),
    [anon_sym_begin] = ACTIONS(6932),
    [anon_sym_COLON] = ACTIONS(6932),
    [anon_sym_do] = ACTIONS(6932),
    [anon_sym_let] = ACTIONS(6932),
    [anon_sym_let_BANG] = ACTIONS(6930),
    [anon_sym_as] = ACTIONS(6932),
    [anon_sym_LPAREN] = ACTIONS(6932),
    [anon_sym_COMMA] = ACTIONS(6930),
    [anon_sym_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_QMARK] = ACTIONS(6932),
    [anon_sym_COLON_COLON] = ACTIONS(6930),
    [anon_sym_AMP] = ACTIONS(6932),
    [anon_sym_LBRACK] = ACTIONS(6932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6930),
    [anon_sym_LBRACE] = ACTIONS(6932),
    [anon_sym_LT_AT] = ACTIONS(6932),
    [anon_sym_LT_AT_AT] = ACTIONS(6932),
    [anon_sym_CARET] = ACTIONS(6932),
    [anon_sym_or] = ACTIONS(6932),
    [anon_sym_DOT] = ACTIONS(6932),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6930),
    [anon_sym_new] = ACTIONS(6932),
    [anon_sym_with] = ACTIONS(6932),
    [anon_sym_return] = ACTIONS(6932),
    [anon_sym_return_BANG] = ACTIONS(6930),
    [anon_sym_yield] = ACTIONS(6932),
    [anon_sym_yield_BANG] = ACTIONS(6930),
    [anon_sym_lazy] = ACTIONS(6932),
    [anon_sym_assert] = ACTIONS(6932),
    [anon_sym_upcast] = ACTIONS(6932),
    [anon_sym_downcast] = ACTIONS(6932),
    [anon_sym_fixed] = ACTIONS(6932),
    [anon_sym_COLON_GT] = ACTIONS(6930),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6930),
    [anon_sym_for] = ACTIONS(6932),
    [anon_sym_while] = ACTIONS(6932),
    [anon_sym_while_BANG] = ACTIONS(6930),
    [anon_sym_if] = ACTIONS(6932),
    [anon_sym_fun] = ACTIONS(6932),
    [anon_sym_try] = ACTIONS(6932),
    [anon_sym_match] = ACTIONS(6932),
    [anon_sym_match_BANG] = ACTIONS(6930),
    [anon_sym_function] = ACTIONS(6932),
    [anon_sym_DOT_LBRACK] = ACTIONS(6930),
    [anon_sym_use] = ACTIONS(6932),
    [anon_sym_use_BANG] = ACTIONS(6930),
    [anon_sym_do_BANG] = ACTIONS(6930),
    [anon_sym_LPAREN2] = ACTIONS(6932),
    [anon_sym_PIPE_PIPE] = ACTIONS(6932),
    [anon_sym_AMP_AMP] = ACTIONS(6932),
    [aux_sym_char_token1] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6932),
    [anon_sym_DQUOTE] = ACTIONS(6932),
    [anon_sym_AT_DQUOTE] = ACTIONS(6930),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6930),
    [sym_bool] = ACTIONS(6932),
    [sym_unit] = ACTIONS(6930),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6932),
    [sym_op_identifier] = ACTIONS(6932),
    [anon_sym_PLUS] = ACTIONS(6932),
    [anon_sym_DASH] = ACTIONS(6932),
    [anon_sym_PLUS_DOT] = ACTIONS(6932),
    [anon_sym_DASH_DOT] = ACTIONS(6932),
    [anon_sym_PERCENT] = ACTIONS(6932),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6932),
    [anon_sym_TILDE] = ACTIONS(6930),
    [aux_sym_prefix_op_token1] = ACTIONS(6932),
    [aux_sym_infix_op_token1] = ACTIONS(6930),
    [aux_sym_infix_op_token2] = ACTIONS(6932),
    [aux_sym_infix_op_token3] = ACTIONS(6932),
    [anon_sym_BANG_EQ] = ACTIONS(6932),
    [anon_sym_COLON_EQ] = ACTIONS(6930),
    [anon_sym_DOLLAR] = ACTIONS(6932),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6932),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6932),
    [sym_int] = ACTIONS(6932),
    [sym_xint] = ACTIONS(6930),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6930),
    [sym__newline] = ACTIONS(6930),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6930),
    [sym__tyapp_open] = ACTIONS(6930),
  },
  [STATE(5758)] = {
    [sym_block_comment] = STATE(5758),
    [sym_compiler_directive_decl] = STATE(5758),
    [sym_fsi_directive_decl] = STATE(5758),
    [sym_preproc_line] = STATE(5758),
    [aux_sym_sequential_expression_repeat1] = STATE(5543),
    [sym_identifier] = ACTIONS(6740),
    [anon_sym_null] = ACTIONS(6740),
    [anon_sym_EQ] = ACTIONS(6740),
    [anon_sym_begin] = ACTIONS(6740),
    [anon_sym_COLON] = ACTIONS(6740),
    [anon_sym_do] = ACTIONS(6740),
    [anon_sym_let] = ACTIONS(6740),
    [anon_sym_let_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN] = ACTIONS(6740),
    [anon_sym_COMMA] = ACTIONS(6738),
    [anon_sym_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_QMARK] = ACTIONS(6740),
    [anon_sym_COLON_COLON] = ACTIONS(6738),
    [anon_sym_AMP] = ACTIONS(6740),
    [anon_sym_LBRACK] = ACTIONS(6740),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6738),
    [anon_sym_LBRACE] = ACTIONS(6740),
    [anon_sym_LT_AT] = ACTIONS(6740),
    [anon_sym_LT_AT_AT] = ACTIONS(6740),
    [anon_sym_CARET] = ACTIONS(6740),
    [anon_sym_or] = ACTIONS(6740),
    [anon_sym_DOT] = ACTIONS(6740),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6738),
    [anon_sym_new] = ACTIONS(6740),
    [anon_sym_return] = ACTIONS(6740),
    [anon_sym_return_BANG] = ACTIONS(6738),
    [anon_sym_yield] = ACTIONS(6740),
    [anon_sym_yield_BANG] = ACTIONS(6738),
    [anon_sym_lazy] = ACTIONS(6740),
    [anon_sym_assert] = ACTIONS(6740),
    [anon_sym_upcast] = ACTIONS(6740),
    [anon_sym_downcast] = ACTIONS(6740),
    [anon_sym_fixed] = ACTIONS(6740),
    [anon_sym_COLON_GT] = ACTIONS(6738),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6738),
    [anon_sym_for] = ACTIONS(6740),
    [anon_sym_while] = ACTIONS(6740),
    [anon_sym_while_BANG] = ACTIONS(6738),
    [anon_sym_if] = ACTIONS(6740),
    [anon_sym_fun] = ACTIONS(6740),
    [anon_sym_try] = ACTIONS(6740),
    [anon_sym_match] = ACTIONS(6740),
    [anon_sym_match_BANG] = ACTIONS(6738),
    [anon_sym_function] = ACTIONS(6740),
    [anon_sym_DOT_LBRACK] = ACTIONS(6738),
    [anon_sym_use] = ACTIONS(6740),
    [anon_sym_use_BANG] = ACTIONS(6738),
    [anon_sym_do_BANG] = ACTIONS(6738),
    [anon_sym_LPAREN2] = ACTIONS(6740),
    [anon_sym_PIPE_PIPE] = ACTIONS(6740),
    [anon_sym_AMP_AMP] = ACTIONS(6740),
    [aux_sym_char_token1] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6740),
    [anon_sym_DQUOTE] = ACTIONS(6740),
    [anon_sym_AT_DQUOTE] = ACTIONS(6738),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6738),
    [sym_bool] = ACTIONS(6740),
    [sym_unit] = ACTIONS(6738),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6740),
    [sym_op_identifier] = ACTIONS(6740),
    [anon_sym_PLUS] = ACTIONS(6740),
    [anon_sym_DASH] = ACTIONS(6740),
    [anon_sym_PLUS_DOT] = ACTIONS(6740),
    [anon_sym_DASH_DOT] = ACTIONS(6740),
    [anon_sym_PERCENT] = ACTIONS(6740),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6740),
    [anon_sym_TILDE] = ACTIONS(6738),
    [aux_sym_prefix_op_token1] = ACTIONS(6740),
    [aux_sym_infix_op_token1] = ACTIONS(6738),
    [aux_sym_infix_op_token2] = ACTIONS(6740),
    [aux_sym_infix_op_token3] = ACTIONS(6740),
    [anon_sym_BANG_EQ] = ACTIONS(6740),
    [anon_sym_COLON_EQ] = ACTIONS(6738),
    [anon_sym_DOLLAR] = ACTIONS(6740),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6740),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6740),
    [sym_int] = ACTIONS(6740),
    [sym_xint] = ACTIONS(6738),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6738),
    [anon_sym_POUNDendif] = ACTIONS(6738),
    [sym__newline] = ACTIONS(6738),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6738),
    [sym__tyapp_open] = ACTIONS(6738),
  },
  [STATE(5759)] = {
    [sym_block_comment] = STATE(5759),
    [sym_compiler_directive_decl] = STATE(5759),
    [sym_fsi_directive_decl] = STATE(5759),
    [sym_preproc_line] = STATE(5759),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__untyped_quoted_close] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5760)] = {
    [sym_block_comment] = STATE(5760),
    [sym_compiler_directive_decl] = STATE(5760),
    [sym_fsi_directive_decl] = STATE(5760),
    [sym_preproc_line] = STATE(5760),
    [sym_identifier] = ACTIONS(7155),
    [anon_sym_null] = ACTIONS(7155),
    [anon_sym_EQ] = ACTIONS(7155),
    [anon_sym_begin] = ACTIONS(7155),
    [anon_sym_COLON] = ACTIONS(7155),
    [anon_sym_do] = ACTIONS(7155),
    [anon_sym_let] = ACTIONS(7155),
    [anon_sym_let_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN] = ACTIONS(7155),
    [anon_sym_COMMA] = ACTIONS(7157),
    [anon_sym_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_QMARK] = ACTIONS(7155),
    [anon_sym_COLON_COLON] = ACTIONS(7157),
    [anon_sym_AMP] = ACTIONS(7155),
    [anon_sym_LBRACK] = ACTIONS(7155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7157),
    [anon_sym_LBRACE] = ACTIONS(7155),
    [anon_sym_LT_AT] = ACTIONS(7155),
    [anon_sym_LT_AT_AT] = ACTIONS(7155),
    [anon_sym_CARET] = ACTIONS(7155),
    [anon_sym_or] = ACTIONS(7155),
    [anon_sym_DOT] = ACTIONS(7155),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7157),
    [anon_sym_new] = ACTIONS(7155),
    [anon_sym_return] = ACTIONS(7155),
    [anon_sym_return_BANG] = ACTIONS(7157),
    [anon_sym_yield] = ACTIONS(7155),
    [anon_sym_yield_BANG] = ACTIONS(7157),
    [anon_sym_lazy] = ACTIONS(7155),
    [anon_sym_assert] = ACTIONS(7155),
    [anon_sym_upcast] = ACTIONS(7155),
    [anon_sym_downcast] = ACTIONS(7155),
    [anon_sym_fixed] = ACTIONS(7155),
    [anon_sym_COLON_GT] = ACTIONS(7157),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7157),
    [anon_sym_for] = ACTIONS(7155),
    [anon_sym_while] = ACTIONS(7155),
    [anon_sym_while_BANG] = ACTIONS(7157),
    [anon_sym_if] = ACTIONS(7155),
    [anon_sym_fun] = ACTIONS(7155),
    [anon_sym_try] = ACTIONS(7155),
    [anon_sym_match] = ACTIONS(7155),
    [anon_sym_match_BANG] = ACTIONS(7157),
    [anon_sym_function] = ACTIONS(7155),
    [anon_sym_DOT_LBRACK] = ACTIONS(7157),
    [anon_sym_use] = ACTIONS(7155),
    [anon_sym_use_BANG] = ACTIONS(7157),
    [anon_sym_do_BANG] = ACTIONS(7157),
    [anon_sym_LPAREN2] = ACTIONS(7155),
    [anon_sym_PIPE_PIPE] = ACTIONS(7155),
    [anon_sym_AMP_AMP] = ACTIONS(7155),
    [aux_sym_char_token1] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7155),
    [anon_sym_DQUOTE] = ACTIONS(7155),
    [anon_sym_AT_DQUOTE] = ACTIONS(7157),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7157),
    [sym_bool] = ACTIONS(7155),
    [sym_unit] = ACTIONS(7157),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7155),
    [sym_op_identifier] = ACTIONS(7155),
    [anon_sym_PLUS] = ACTIONS(7155),
    [anon_sym_DASH] = ACTIONS(7155),
    [anon_sym_PLUS_DOT] = ACTIONS(7155),
    [anon_sym_DASH_DOT] = ACTIONS(7155),
    [anon_sym_PERCENT] = ACTIONS(7155),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7155),
    [anon_sym_TILDE] = ACTIONS(7157),
    [aux_sym_prefix_op_token1] = ACTIONS(7155),
    [aux_sym_infix_op_token1] = ACTIONS(7157),
    [aux_sym_infix_op_token2] = ACTIONS(7155),
    [aux_sym_infix_op_token3] = ACTIONS(7155),
    [anon_sym_BANG_EQ] = ACTIONS(7155),
    [anon_sym_COLON_EQ] = ACTIONS(7157),
    [anon_sym_DOLLAR] = ACTIONS(7155),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7155),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7155),
    [sym_int] = ACTIONS(7155),
    [sym_xint] = ACTIONS(7157),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7157),
    [anon_sym_POUNDendif] = ACTIONS(7157),
    [anon_sym_POUNDelse] = ACTIONS(7157),
    [sym__newline] = ACTIONS(7157),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7157),
    [sym__tyapp_open] = ACTIONS(7157),
  },
  [STATE(5761)] = {
    [sym_block_comment] = STATE(5761),
    [sym_compiler_directive_decl] = STATE(5761),
    [sym_fsi_directive_decl] = STATE(5761),
    [sym_preproc_line] = STATE(5761),
    [sym_identifier] = ACTIONS(7159),
    [anon_sym_null] = ACTIONS(7159),
    [anon_sym_EQ] = ACTIONS(7159),
    [anon_sym_begin] = ACTIONS(7159),
    [anon_sym_COLON] = ACTIONS(7159),
    [anon_sym_do] = ACTIONS(7159),
    [anon_sym_let] = ACTIONS(7159),
    [anon_sym_let_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN] = ACTIONS(7159),
    [anon_sym_COMMA] = ACTIONS(7161),
    [anon_sym_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_QMARK] = ACTIONS(7159),
    [anon_sym_COLON_COLON] = ACTIONS(7161),
    [anon_sym_AMP] = ACTIONS(7159),
    [anon_sym_LBRACK] = ACTIONS(7159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(7161),
    [anon_sym_LBRACE] = ACTIONS(7159),
    [anon_sym_LT_AT] = ACTIONS(7159),
    [anon_sym_LT_AT_AT] = ACTIONS(7159),
    [anon_sym_CARET] = ACTIONS(7159),
    [anon_sym_or] = ACTIONS(7159),
    [anon_sym_DOT] = ACTIONS(7159),
    [anon_sym_LBRACE_PIPE] = ACTIONS(7161),
    [anon_sym_new] = ACTIONS(7159),
    [anon_sym_return] = ACTIONS(7159),
    [anon_sym_return_BANG] = ACTIONS(7161),
    [anon_sym_yield] = ACTIONS(7159),
    [anon_sym_yield_BANG] = ACTIONS(7161),
    [anon_sym_lazy] = ACTIONS(7159),
    [anon_sym_assert] = ACTIONS(7159),
    [anon_sym_upcast] = ACTIONS(7159),
    [anon_sym_downcast] = ACTIONS(7159),
    [anon_sym_fixed] = ACTIONS(7159),
    [anon_sym_COLON_GT] = ACTIONS(7161),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(7161),
    [anon_sym_for] = ACTIONS(7159),
    [anon_sym_while] = ACTIONS(7159),
    [anon_sym_while_BANG] = ACTIONS(7161),
    [anon_sym_if] = ACTIONS(7159),
    [anon_sym_fun] = ACTIONS(7159),
    [anon_sym_try] = ACTIONS(7159),
    [anon_sym_match] = ACTIONS(7159),
    [anon_sym_match_BANG] = ACTIONS(7161),
    [anon_sym_function] = ACTIONS(7159),
    [anon_sym_DOT_LBRACK] = ACTIONS(7161),
    [anon_sym_use] = ACTIONS(7159),
    [anon_sym_use_BANG] = ACTIONS(7161),
    [anon_sym_do_BANG] = ACTIONS(7161),
    [anon_sym_LPAREN2] = ACTIONS(7159),
    [anon_sym_PIPE_PIPE] = ACTIONS(7159),
    [anon_sym_AMP_AMP] = ACTIONS(7159),
    [aux_sym_char_token1] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(7159),
    [anon_sym_DQUOTE] = ACTIONS(7159),
    [anon_sym_AT_DQUOTE] = ACTIONS(7161),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(7161),
    [sym_bool] = ACTIONS(7159),
    [sym_unit] = ACTIONS(7161),
    [anon_sym_LPAREN_PIPE] = ACTIONS(7159),
    [sym_op_identifier] = ACTIONS(7159),
    [anon_sym_PLUS] = ACTIONS(7159),
    [anon_sym_DASH] = ACTIONS(7159),
    [anon_sym_PLUS_DOT] = ACTIONS(7159),
    [anon_sym_DASH_DOT] = ACTIONS(7159),
    [anon_sym_PERCENT] = ACTIONS(7159),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(7159),
    [anon_sym_TILDE] = ACTIONS(7161),
    [aux_sym_prefix_op_token1] = ACTIONS(7159),
    [aux_sym_infix_op_token1] = ACTIONS(7161),
    [aux_sym_infix_op_token2] = ACTIONS(7159),
    [aux_sym_infix_op_token3] = ACTIONS(7159),
    [anon_sym_BANG_EQ] = ACTIONS(7159),
    [anon_sym_COLON_EQ] = ACTIONS(7161),
    [anon_sym_DOLLAR] = ACTIONS(7159),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(7159),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(7159),
    [sym_int] = ACTIONS(7159),
    [sym_xint] = ACTIONS(7161),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(7161),
    [anon_sym_POUNDendif] = ACTIONS(7161),
    [anon_sym_POUNDelse] = ACTIONS(7161),
    [sym__newline] = ACTIONS(7161),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(7161),
    [sym__tyapp_open] = ACTIONS(7161),
  },
  [STATE(5762)] = {
    [sym_block_comment] = STATE(5762),
    [sym_compiler_directive_decl] = STATE(5762),
    [sym_fsi_directive_decl] = STATE(5762),
    [sym_preproc_line] = STATE(5762),
    [sym_identifier] = ACTIONS(6867),
    [anon_sym_null] = ACTIONS(6867),
    [anon_sym_EQ] = ACTIONS(6867),
    [anon_sym_begin] = ACTIONS(6867),
    [anon_sym_COLON] = ACTIONS(6867),
    [anon_sym_do] = ACTIONS(6867),
    [anon_sym_let] = ACTIONS(6867),
    [anon_sym_let_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN] = ACTIONS(6867),
    [anon_sym_COMMA] = ACTIONS(6865),
    [anon_sym_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_QMARK] = ACTIONS(6867),
    [anon_sym_COLON_COLON] = ACTIONS(6865),
    [anon_sym_AMP] = ACTIONS(6867),
    [anon_sym_LBRACK] = ACTIONS(6867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6865),
    [anon_sym_LBRACE] = ACTIONS(6867),
    [anon_sym_LT_AT] = ACTIONS(6867),
    [anon_sym_LT_AT_AT] = ACTIONS(6867),
    [anon_sym_CARET] = ACTIONS(6867),
    [anon_sym_or] = ACTIONS(6867),
    [anon_sym_DOT] = ACTIONS(6867),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6865),
    [anon_sym_new] = ACTIONS(6867),
    [anon_sym_return] = ACTIONS(6867),
    [anon_sym_return_BANG] = ACTIONS(6865),
    [anon_sym_yield] = ACTIONS(6867),
    [anon_sym_yield_BANG] = ACTIONS(6865),
    [anon_sym_lazy] = ACTIONS(6867),
    [anon_sym_assert] = ACTIONS(6867),
    [anon_sym_upcast] = ACTIONS(6867),
    [anon_sym_downcast] = ACTIONS(6867),
    [anon_sym_fixed] = ACTIONS(6867),
    [anon_sym_COLON_GT] = ACTIONS(6865),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6865),
    [anon_sym_for] = ACTIONS(6867),
    [anon_sym_while] = ACTIONS(6867),
    [anon_sym_while_BANG] = ACTIONS(6865),
    [anon_sym_if] = ACTIONS(6867),
    [anon_sym_fun] = ACTIONS(6867),
    [anon_sym_try] = ACTIONS(6867),
    [anon_sym_match] = ACTIONS(6867),
    [anon_sym_match_BANG] = ACTIONS(6865),
    [anon_sym_function] = ACTIONS(6867),
    [anon_sym_LT_DASH] = ACTIONS(6867),
    [anon_sym_DOT_LBRACK] = ACTIONS(6865),
    [anon_sym_use] = ACTIONS(6867),
    [anon_sym_use_BANG] = ACTIONS(6865),
    [anon_sym_do_BANG] = ACTIONS(6865),
    [anon_sym_LPAREN2] = ACTIONS(6867),
    [anon_sym_PIPE_PIPE] = ACTIONS(6867),
    [anon_sym_AMP_AMP] = ACTIONS(6867),
    [aux_sym_char_token1] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6867),
    [anon_sym_DQUOTE] = ACTIONS(6867),
    [anon_sym_AT_DQUOTE] = ACTIONS(6865),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6865),
    [sym_bool] = ACTIONS(6867),
    [sym_unit] = ACTIONS(6865),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6867),
    [sym_op_identifier] = ACTIONS(6867),
    [anon_sym_PLUS] = ACTIONS(6867),
    [anon_sym_DASH] = ACTIONS(6867),
    [anon_sym_PLUS_DOT] = ACTIONS(6867),
    [anon_sym_DASH_DOT] = ACTIONS(6867),
    [anon_sym_PERCENT] = ACTIONS(6867),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6867),
    [anon_sym_TILDE] = ACTIONS(6865),
    [aux_sym_prefix_op_token1] = ACTIONS(6867),
    [aux_sym_infix_op_token1] = ACTIONS(6865),
    [aux_sym_infix_op_token2] = ACTIONS(6867),
    [aux_sym_infix_op_token3] = ACTIONS(6867),
    [anon_sym_BANG_EQ] = ACTIONS(6867),
    [anon_sym_COLON_EQ] = ACTIONS(6865),
    [anon_sym_DOLLAR] = ACTIONS(6867),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6867),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6867),
    [sym_int] = ACTIONS(6867),
    [sym_xint] = ACTIONS(6865),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6865),
    [anon_sym_POUNDendif] = ACTIONS(6865),
    [sym__newline] = ACTIONS(6865),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6865),
    [sym__tyapp_open] = ACTIONS(6865),
  },
  [STATE(5763)] = {
    [sym_block_comment] = STATE(5763),
    [sym_compiler_directive_decl] = STATE(5763),
    [sym_fsi_directive_decl] = STATE(5763),
    [sym_preproc_line] = STATE(5763),
    [sym_identifier] = ACTIONS(6612),
    [anon_sym_null] = ACTIONS(6612),
    [anon_sym_EQ] = ACTIONS(6612),
    [anon_sym_begin] = ACTIONS(6612),
    [anon_sym_COLON] = ACTIONS(6612),
    [anon_sym_do] = ACTIONS(6612),
    [anon_sym_let] = ACTIONS(6612),
    [anon_sym_let_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN] = ACTIONS(6612),
    [anon_sym_COMMA] = ACTIONS(6614),
    [anon_sym_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_QMARK] = ACTIONS(6612),
    [anon_sym_COLON_COLON] = ACTIONS(6614),
    [anon_sym_PIPE] = ACTIONS(6612),
    [anon_sym_AMP] = ACTIONS(6612),
    [anon_sym_LBRACK] = ACTIONS(6612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6614),
    [anon_sym_LBRACE] = ACTIONS(6612),
    [anon_sym_LT_AT] = ACTIONS(6612),
    [anon_sym_LT_AT_AT] = ACTIONS(6612),
    [anon_sym_CARET] = ACTIONS(6612),
    [anon_sym_or] = ACTIONS(6612),
    [anon_sym_DOT] = ACTIONS(6612),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6614),
    [anon_sym_new] = ACTIONS(6612),
    [anon_sym_with] = ACTIONS(6612),
    [anon_sym_return] = ACTIONS(6612),
    [anon_sym_return_BANG] = ACTIONS(6614),
    [anon_sym_yield] = ACTIONS(6612),
    [anon_sym_yield_BANG] = ACTIONS(6614),
    [anon_sym_lazy] = ACTIONS(6612),
    [anon_sym_assert] = ACTIONS(6612),
    [anon_sym_upcast] = ACTIONS(6612),
    [anon_sym_downcast] = ACTIONS(6612),
    [anon_sym_fixed] = ACTIONS(6612),
    [anon_sym_COLON_GT] = ACTIONS(6614),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6614),
    [anon_sym_for] = ACTIONS(6612),
    [anon_sym_while] = ACTIONS(6612),
    [anon_sym_while_BANG] = ACTIONS(6614),
    [anon_sym_if] = ACTIONS(6612),
    [anon_sym_fun] = ACTIONS(6612),
    [anon_sym_try] = ACTIONS(6612),
    [anon_sym_match] = ACTIONS(6612),
    [anon_sym_match_BANG] = ACTIONS(6614),
    [anon_sym_function] = ACTIONS(6612),
    [anon_sym_DOT_LBRACK] = ACTIONS(6614),
    [anon_sym_use] = ACTIONS(6612),
    [anon_sym_use_BANG] = ACTIONS(6614),
    [anon_sym_do_BANG] = ACTIONS(6614),
    [anon_sym_LPAREN2] = ACTIONS(6612),
    [anon_sym_PIPE_PIPE] = ACTIONS(6612),
    [anon_sym_AMP_AMP] = ACTIONS(6612),
    [aux_sym_char_token1] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6612),
    [anon_sym_DQUOTE] = ACTIONS(6612),
    [anon_sym_AT_DQUOTE] = ACTIONS(6614),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6614),
    [sym_bool] = ACTIONS(6612),
    [sym_unit] = ACTIONS(6614),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6612),
    [sym_op_identifier] = ACTIONS(6612),
    [anon_sym_PLUS] = ACTIONS(6612),
    [anon_sym_DASH] = ACTIONS(6612),
    [anon_sym_PLUS_DOT] = ACTIONS(6612),
    [anon_sym_DASH_DOT] = ACTIONS(6612),
    [anon_sym_PERCENT] = ACTIONS(6612),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6612),
    [anon_sym_TILDE] = ACTIONS(6614),
    [aux_sym_prefix_op_token1] = ACTIONS(6612),
    [aux_sym_infix_op_token1] = ACTIONS(6614),
    [aux_sym_infix_op_token2] = ACTIONS(6612),
    [aux_sym_infix_op_token3] = ACTIONS(6612),
    [anon_sym_BANG_EQ] = ACTIONS(6612),
    [anon_sym_COLON_EQ] = ACTIONS(6614),
    [anon_sym_DOLLAR] = ACTIONS(6612),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6612),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6612),
    [sym_int] = ACTIONS(6612),
    [sym_xint] = ACTIONS(6614),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6614),
    [sym__newline] = ACTIONS(6614),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6614),
    [sym__tyapp_open] = ACTIONS(6614),
  },
  [STATE(5764)] = {
    [sym_block_comment] = STATE(5764),
    [sym_compiler_directive_decl] = STATE(5764),
    [sym_fsi_directive_decl] = STATE(5764),
    [sym_preproc_line] = STATE(5764),
    [sym_identifier] = ACTIONS(6312),
    [anon_sym_module] = ACTIONS(6312),
    [anon_sym_null] = ACTIONS(6312),
    [anon_sym_begin] = ACTIONS(6312),
    [anon_sym_open] = ACTIONS(6312),
    [anon_sym_type] = ACTIONS(6312),
    [anon_sym_LBRACK_LT] = ACTIONS(6310),
    [anon_sym_do] = ACTIONS(6312),
    [anon_sym_and] = ACTIONS(6312),
    [anon_sym_let] = ACTIONS(6312),
    [anon_sym_let_BANG] = ACTIONS(6310),
    [aux_sym_access_modifier_token1] = ACTIONS(6310),
    [anon_sym_LPAREN] = ACTIONS(6312),
    [anon_sym_PIPE] = ACTIONS(6310),
    [anon_sym_AMP] = ACTIONS(6312),
    [anon_sym_LBRACK] = ACTIONS(6312),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6310),
    [anon_sym_LBRACE] = ACTIONS(6312),
    [anon_sym_LT_AT] = ACTIONS(6312),
    [anon_sym_LT_AT_AT] = ACTIONS(6310),
    [anon_sym_CARET] = ACTIONS(6310),
    [anon_sym_or] = ACTIONS(6312),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6310),
    [anon_sym_new] = ACTIONS(6312),
    [anon_sym_with] = ACTIONS(6312),
    [anon_sym_return] = ACTIONS(6312),
    [anon_sym_return_BANG] = ACTIONS(6310),
    [anon_sym_yield] = ACTIONS(6312),
    [anon_sym_yield_BANG] = ACTIONS(6310),
    [anon_sym_lazy] = ACTIONS(6312),
    [anon_sym_assert] = ACTIONS(6312),
    [anon_sym_upcast] = ACTIONS(6312),
    [anon_sym_downcast] = ACTIONS(6312),
    [anon_sym_fixed] = ACTIONS(6312),
    [anon_sym_COLON_GT] = ACTIONS(6310),
    [anon_sym_for] = ACTIONS(6312),
    [anon_sym_while] = ACTIONS(6312),
    [anon_sym_while_BANG] = ACTIONS(6310),
    [anon_sym_if] = ACTIONS(6312),
    [anon_sym_fun] = ACTIONS(6312),
    [anon_sym_DASH_GT] = ACTIONS(6310),
    [anon_sym_try] = ACTIONS(6312),
    [anon_sym_match] = ACTIONS(6312),
    [anon_sym_match_BANG] = ACTIONS(6310),
    [anon_sym_function] = ACTIONS(6312),
    [anon_sym_use] = ACTIONS(6312),
    [anon_sym_use_BANG] = ACTIONS(6310),
    [anon_sym_do_BANG] = ACTIONS(6310),
    [anon_sym_AMP_AMP] = ACTIONS(6310),
    [anon_sym_STAR] = ACTIONS(6310),
    [anon_sym_LT] = ACTIONS(6312),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(6310),
    [anon_sym_default] = ACTIONS(6312),
    [anon_sym_static] = ACTIONS(6312),
    [anon_sym_member] = ACTIONS(6312),
    [anon_sym_exception] = ACTIONS(6312),
    [anon_sym_interface] = ACTIONS(6312),
    [anon_sym_abstract] = ACTIONS(6312),
    [anon_sym_val] = ACTIONS(6312),
    [anon_sym_override] = ACTIONS(6312),
    [anon_sym_extern] = ACTIONS(6312),
    [aux_sym_char_token1] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6312),
    [anon_sym_DQUOTE] = ACTIONS(6312),
    [anon_sym_AT_DQUOTE] = ACTIONS(6310),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6310),
    [sym_bool] = ACTIONS(6312),
    [sym_unit] = ACTIONS(6310),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6312),
    [sym_op_identifier] = ACTIONS(6310),
    [anon_sym_PLUS] = ACTIONS(6312),
    [anon_sym_DASH] = ACTIONS(6312),
    [anon_sym_PLUS_DOT] = ACTIONS(6310),
    [anon_sym_DASH_DOT] = ACTIONS(6310),
    [anon_sym_PERCENT] = ACTIONS(6312),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6310),
    [anon_sym_TILDE] = ACTIONS(6310),
    [aux_sym_prefix_op_token1] = ACTIONS(6310),
    [sym_int] = ACTIONS(6312),
    [sym_xint] = ACTIONS(6310),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(6310),
    [anon_sym_POUNDload] = ACTIONS(6310),
    [anon_sym_POUNDtime] = ACTIONS(6310),
    [anon_sym_POUNDI] = ACTIONS(6310),
    [anon_sym_POUNDhelp] = ACTIONS(6310),
    [anon_sym_POUNDquit] = ACTIONS(6310),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6310),
    [anon_sym_POUNDendif] = ACTIONS(6310),
    [anon_sym_POUNDelse] = ACTIONS(6310),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6310),
  },
  [STATE(5765)] = {
    [sym_block_comment] = STATE(5765),
    [sym_compiler_directive_decl] = STATE(5765),
    [sym_fsi_directive_decl] = STATE(5765),
    [sym_preproc_line] = STATE(5765),
    [sym_identifier] = ACTIONS(6863),
    [anon_sym_null] = ACTIONS(6863),
    [anon_sym_EQ] = ACTIONS(6863),
    [anon_sym_begin] = ACTIONS(6863),
    [anon_sym_COLON] = ACTIONS(6863),
    [anon_sym_do] = ACTIONS(6863),
    [anon_sym_let] = ACTIONS(6863),
    [anon_sym_let_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN] = ACTIONS(6863),
    [anon_sym_COMMA] = ACTIONS(6861),
    [anon_sym_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_QMARK] = ACTIONS(6863),
    [anon_sym_COLON_COLON] = ACTIONS(6861),
    [anon_sym_AMP] = ACTIONS(6863),
    [anon_sym_GT] = ACTIONS(6863),
    [anon_sym_LBRACK] = ACTIONS(6863),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6861),
    [anon_sym_LBRACE] = ACTIONS(6863),
    [anon_sym_LT_AT] = ACTIONS(6863),
    [anon_sym_LT_AT_AT] = ACTIONS(6863),
    [anon_sym_CARET] = ACTIONS(6863),
    [anon_sym_or] = ACTIONS(6863),
    [anon_sym_DOT] = ACTIONS(6863),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6861),
    [anon_sym_new] = ACTIONS(6863),
    [anon_sym_return] = ACTIONS(6863),
    [anon_sym_return_BANG] = ACTIONS(6861),
    [anon_sym_yield] = ACTIONS(6863),
    [anon_sym_yield_BANG] = ACTIONS(6861),
    [anon_sym_lazy] = ACTIONS(6863),
    [anon_sym_assert] = ACTIONS(6863),
    [anon_sym_upcast] = ACTIONS(6863),
    [anon_sym_downcast] = ACTIONS(6863),
    [anon_sym_fixed] = ACTIONS(6863),
    [anon_sym_COLON_GT] = ACTIONS(6861),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6861),
    [anon_sym_for] = ACTIONS(6863),
    [anon_sym_while] = ACTIONS(6863),
    [anon_sym_while_BANG] = ACTIONS(6861),
    [anon_sym_if] = ACTIONS(6863),
    [anon_sym_fun] = ACTIONS(6863),
    [anon_sym_try] = ACTIONS(6863),
    [anon_sym_match] = ACTIONS(6863),
    [anon_sym_match_BANG] = ACTIONS(6861),
    [anon_sym_function] = ACTIONS(6863),
    [anon_sym_LT_DASH] = ACTIONS(6863),
    [anon_sym_DOT_LBRACK] = ACTIONS(6861),
    [anon_sym_use] = ACTIONS(6863),
    [anon_sym_use_BANG] = ACTIONS(6861),
    [anon_sym_do_BANG] = ACTIONS(6861),
    [anon_sym_LPAREN2] = ACTIONS(6863),
    [anon_sym_PIPE_PIPE] = ACTIONS(6863),
    [anon_sym_AMP_AMP] = ACTIONS(6863),
    [aux_sym_char_token1] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6863),
    [anon_sym_DQUOTE] = ACTIONS(6863),
    [anon_sym_AT_DQUOTE] = ACTIONS(6861),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6861),
    [sym_bool] = ACTIONS(6863),
    [sym_unit] = ACTIONS(6861),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6863),
    [sym_op_identifier] = ACTIONS(6863),
    [anon_sym_PLUS] = ACTIONS(6863),
    [anon_sym_DASH] = ACTIONS(6863),
    [anon_sym_PLUS_DOT] = ACTIONS(6863),
    [anon_sym_DASH_DOT] = ACTIONS(6863),
    [anon_sym_PERCENT] = ACTIONS(6863),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6863),
    [anon_sym_TILDE] = ACTIONS(6861),
    [aux_sym_prefix_op_token1] = ACTIONS(6863),
    [aux_sym_infix_op_token1] = ACTIONS(6861),
    [aux_sym_infix_op_token2] = ACTIONS(6863),
    [aux_sym_infix_op_token3] = ACTIONS(6863),
    [anon_sym_BANG_EQ] = ACTIONS(6863),
    [anon_sym_COLON_EQ] = ACTIONS(6861),
    [anon_sym_DOLLAR] = ACTIONS(6863),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6863),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6863),
    [sym_int] = ACTIONS(6863),
    [sym_xint] = ACTIONS(6861),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6861),
    [sym__newline] = ACTIONS(6861),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6861),
    [sym__tyapp_open] = ACTIONS(6861),
  },
  [STATE(5766)] = {
    [sym_block_comment] = STATE(5766),
    [sym_compiler_directive_decl] = STATE(5766),
    [sym_fsi_directive_decl] = STATE(5766),
    [sym_preproc_line] = STATE(5766),
    [aux_sym_sequential_expression_repeat1] = STATE(5766),
    [sym_identifier] = ACTIONS(6800),
    [anon_sym_null] = ACTIONS(6800),
    [anon_sym_EQ] = ACTIONS(6800),
    [anon_sym_begin] = ACTIONS(6800),
    [anon_sym_COLON] = ACTIONS(6800),
    [anon_sym_do] = ACTIONS(6800),
    [anon_sym_let] = ACTIONS(6800),
    [anon_sym_let_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN] = ACTIONS(6800),
    [anon_sym_COMMA] = ACTIONS(6798),
    [anon_sym_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_QMARK] = ACTIONS(6800),
    [anon_sym_COLON_COLON] = ACTIONS(6798),
    [anon_sym_AMP] = ACTIONS(6800),
    [anon_sym_LBRACK] = ACTIONS(6800),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6798),
    [anon_sym_LBRACE] = ACTIONS(6800),
    [anon_sym_LT_AT] = ACTIONS(6800),
    [anon_sym_LT_AT_AT] = ACTIONS(6800),
    [anon_sym_CARET] = ACTIONS(6800),
    [anon_sym_or] = ACTIONS(6800),
    [anon_sym_DOT] = ACTIONS(6800),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6798),
    [anon_sym_new] = ACTIONS(6800),
    [anon_sym_return] = ACTIONS(6800),
    [anon_sym_return_BANG] = ACTIONS(6798),
    [anon_sym_yield] = ACTIONS(6800),
    [anon_sym_yield_BANG] = ACTIONS(6798),
    [anon_sym_lazy] = ACTIONS(6800),
    [anon_sym_assert] = ACTIONS(6800),
    [anon_sym_upcast] = ACTIONS(6800),
    [anon_sym_downcast] = ACTIONS(6800),
    [anon_sym_fixed] = ACTIONS(6800),
    [anon_sym_COLON_GT] = ACTIONS(6798),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6798),
    [anon_sym_for] = ACTIONS(6800),
    [anon_sym_while] = ACTIONS(6800),
    [anon_sym_while_BANG] = ACTIONS(6798),
    [anon_sym_if] = ACTIONS(6800),
    [anon_sym_fun] = ACTIONS(6800),
    [anon_sym_try] = ACTIONS(6800),
    [anon_sym_match] = ACTIONS(6800),
    [anon_sym_match_BANG] = ACTIONS(6798),
    [anon_sym_function] = ACTIONS(6800),
    [anon_sym_DOT_LBRACK] = ACTIONS(6798),
    [anon_sym_use] = ACTIONS(6800),
    [anon_sym_use_BANG] = ACTIONS(6798),
    [anon_sym_do_BANG] = ACTIONS(6798),
    [anon_sym_LPAREN2] = ACTIONS(6800),
    [anon_sym_PIPE_PIPE] = ACTIONS(6800),
    [anon_sym_AMP_AMP] = ACTIONS(6800),
    [aux_sym_char_token1] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6800),
    [anon_sym_DQUOTE] = ACTIONS(6800),
    [anon_sym_AT_DQUOTE] = ACTIONS(6798),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6798),
    [sym_bool] = ACTIONS(6800),
    [sym_unit] = ACTIONS(6798),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6800),
    [sym_op_identifier] = ACTIONS(6800),
    [anon_sym_PLUS] = ACTIONS(6800),
    [anon_sym_DASH] = ACTIONS(6800),
    [anon_sym_PLUS_DOT] = ACTIONS(6800),
    [anon_sym_DASH_DOT] = ACTIONS(6800),
    [anon_sym_PERCENT] = ACTIONS(6800),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6800),
    [anon_sym_TILDE] = ACTIONS(6798),
    [aux_sym_prefix_op_token1] = ACTIONS(6800),
    [aux_sym_infix_op_token1] = ACTIONS(6798),
    [aux_sym_infix_op_token2] = ACTIONS(6800),
    [aux_sym_infix_op_token3] = ACTIONS(6800),
    [anon_sym_BANG_EQ] = ACTIONS(6800),
    [anon_sym_COLON_EQ] = ACTIONS(6798),
    [anon_sym_DOLLAR] = ACTIONS(6800),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6800),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6800),
    [sym_int] = ACTIONS(6800),
    [sym_xint] = ACTIONS(6798),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6798),
    [sym__newline] = ACTIONS(9845),
    [sym__dedent] = ACTIONS(6798),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6798),
    [sym__tyapp_open] = ACTIONS(6798),
  },
  [STATE(5767)] = {
    [sym_block_comment] = STATE(5767),
    [sym_compiler_directive_decl] = STATE(5767),
    [sym_fsi_directive_decl] = STATE(5767),
    [sym_preproc_line] = STATE(5767),
    [sym_identifier] = ACTIONS(6692),
    [anon_sym_null] = ACTIONS(6692),
    [anon_sym_EQ] = ACTIONS(6692),
    [anon_sym_begin] = ACTIONS(6692),
    [anon_sym_COLON] = ACTIONS(6692),
    [anon_sym_do] = ACTIONS(6692),
    [anon_sym_let] = ACTIONS(6692),
    [anon_sym_let_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN] = ACTIONS(6692),
    [anon_sym_COMMA] = ACTIONS(6694),
    [anon_sym_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_QMARK] = ACTIONS(6692),
    [anon_sym_COLON_COLON] = ACTIONS(6694),
    [anon_sym_AMP] = ACTIONS(6692),
    [anon_sym_LBRACK] = ACTIONS(6692),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6694),
    [anon_sym_LBRACE] = ACTIONS(6692),
    [anon_sym_LT_AT] = ACTIONS(6692),
    [anon_sym_LT_AT_AT] = ACTIONS(6692),
    [anon_sym_CARET] = ACTIONS(6692),
    [anon_sym_or] = ACTIONS(6692),
    [anon_sym_DOT] = ACTIONS(6692),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6694),
    [anon_sym_new] = ACTIONS(6692),
    [anon_sym_return] = ACTIONS(6692),
    [anon_sym_return_BANG] = ACTIONS(6694),
    [anon_sym_yield] = ACTIONS(6692),
    [anon_sym_yield_BANG] = ACTIONS(6694),
    [anon_sym_lazy] = ACTIONS(6692),
    [anon_sym_assert] = ACTIONS(6692),
    [anon_sym_upcast] = ACTIONS(6692),
    [anon_sym_downcast] = ACTIONS(6692),
    [anon_sym_fixed] = ACTIONS(6692),
    [anon_sym_COLON_GT] = ACTIONS(6694),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6694),
    [anon_sym_for] = ACTIONS(6692),
    [anon_sym_while] = ACTIONS(6692),
    [anon_sym_while_BANG] = ACTIONS(6694),
    [anon_sym_if] = ACTIONS(6692),
    [anon_sym_fun] = ACTIONS(6692),
    [anon_sym_try] = ACTIONS(6692),
    [anon_sym_match] = ACTIONS(6692),
    [anon_sym_match_BANG] = ACTIONS(6694),
    [anon_sym_function] = ACTIONS(6692),
    [anon_sym_LT_DASH] = ACTIONS(6692),
    [anon_sym_DOT_LBRACK] = ACTIONS(6694),
    [anon_sym_use] = ACTIONS(6692),
    [anon_sym_use_BANG] = ACTIONS(6694),
    [anon_sym_do_BANG] = ACTIONS(6694),
    [anon_sym_LPAREN2] = ACTIONS(6692),
    [anon_sym_PIPE_PIPE] = ACTIONS(6692),
    [anon_sym_AMP_AMP] = ACTIONS(6692),
    [aux_sym_char_token1] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6692),
    [anon_sym_DQUOTE] = ACTIONS(6692),
    [anon_sym_AT_DQUOTE] = ACTIONS(6694),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6694),
    [sym_bool] = ACTIONS(6692),
    [sym_unit] = ACTIONS(6694),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6692),
    [sym_op_identifier] = ACTIONS(6692),
    [anon_sym_PLUS] = ACTIONS(6692),
    [anon_sym_DASH] = ACTIONS(6692),
    [anon_sym_PLUS_DOT] = ACTIONS(6692),
    [anon_sym_DASH_DOT] = ACTIONS(6692),
    [anon_sym_PERCENT] = ACTIONS(6692),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6692),
    [anon_sym_TILDE] = ACTIONS(6694),
    [aux_sym_prefix_op_token1] = ACTIONS(6692),
    [aux_sym_infix_op_token1] = ACTIONS(6694),
    [aux_sym_infix_op_token2] = ACTIONS(6692),
    [aux_sym_infix_op_token3] = ACTIONS(6692),
    [anon_sym_BANG_EQ] = ACTIONS(6692),
    [anon_sym_COLON_EQ] = ACTIONS(6694),
    [anon_sym_DOLLAR] = ACTIONS(6692),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6692),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6692),
    [sym_int] = ACTIONS(6692),
    [sym_xint] = ACTIONS(6694),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6694),
    [sym__newline] = ACTIONS(6694),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6694),
    [sym__multi_dollar_interp_end] = ACTIONS(6694),
    [sym__tyapp_open] = ACTIONS(6694),
  },
  [STATE(5768)] = {
    [sym_block_comment] = STATE(5768),
    [sym_compiler_directive_decl] = STATE(5768),
    [sym_fsi_directive_decl] = STATE(5768),
    [sym_preproc_line] = STATE(5768),
    [sym_identifier] = ACTIONS(6784),
    [anon_sym_null] = ACTIONS(6784),
    [anon_sym_EQ] = ACTIONS(6784),
    [anon_sym_begin] = ACTIONS(6784),
    [anon_sym_COLON] = ACTIONS(6784),
    [anon_sym_do] = ACTIONS(6784),
    [anon_sym_let] = ACTIONS(6784),
    [anon_sym_let_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN] = ACTIONS(6784),
    [anon_sym_COMMA] = ACTIONS(6786),
    [anon_sym_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_QMARK] = ACTIONS(6784),
    [anon_sym_COLON_COLON] = ACTIONS(6786),
    [anon_sym_AMP] = ACTIONS(6784),
    [anon_sym_LBRACK] = ACTIONS(6784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6786),
    [anon_sym_LBRACE] = ACTIONS(6784),
    [anon_sym_LT_AT] = ACTIONS(6784),
    [anon_sym_LT_AT_AT] = ACTIONS(6784),
    [anon_sym_CARET] = ACTIONS(6784),
    [anon_sym_or] = ACTIONS(6784),
    [anon_sym_DOT] = ACTIONS(6784),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6786),
    [anon_sym_new] = ACTIONS(6784),
    [anon_sym_with] = ACTIONS(6784),
    [anon_sym_return] = ACTIONS(6784),
    [anon_sym_return_BANG] = ACTIONS(6786),
    [anon_sym_yield] = ACTIONS(6784),
    [anon_sym_yield_BANG] = ACTIONS(6786),
    [anon_sym_lazy] = ACTIONS(6784),
    [anon_sym_assert] = ACTIONS(6784),
    [anon_sym_upcast] = ACTIONS(6784),
    [anon_sym_downcast] = ACTIONS(6784),
    [anon_sym_fixed] = ACTIONS(6784),
    [anon_sym_COLON_GT] = ACTIONS(6786),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6786),
    [anon_sym_for] = ACTIONS(6784),
    [anon_sym_done] = ACTIONS(9848),
    [anon_sym_while] = ACTIONS(6784),
    [anon_sym_while_BANG] = ACTIONS(6786),
    [anon_sym_if] = ACTIONS(6784),
    [anon_sym_fun] = ACTIONS(6784),
    [anon_sym_try] = ACTIONS(6784),
    [anon_sym_match] = ACTIONS(6784),
    [anon_sym_match_BANG] = ACTIONS(6786),
    [anon_sym_function] = ACTIONS(6784),
    [anon_sym_DOT_LBRACK] = ACTIONS(6786),
    [anon_sym_use] = ACTIONS(6784),
    [anon_sym_use_BANG] = ACTIONS(6786),
    [anon_sym_do_BANG] = ACTIONS(6786),
    [anon_sym_LPAREN2] = ACTIONS(6784),
    [anon_sym_PIPE_PIPE] = ACTIONS(6784),
    [anon_sym_AMP_AMP] = ACTIONS(6784),
    [aux_sym_char_token1] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6784),
    [anon_sym_DQUOTE] = ACTIONS(6784),
    [anon_sym_AT_DQUOTE] = ACTIONS(6786),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6786),
    [sym_bool] = ACTIONS(6784),
    [sym_unit] = ACTIONS(6786),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6784),
    [sym_op_identifier] = ACTIONS(6784),
    [anon_sym_PLUS] = ACTIONS(6784),
    [anon_sym_DASH] = ACTIONS(6784),
    [anon_sym_PLUS_DOT] = ACTIONS(6784),
    [anon_sym_DASH_DOT] = ACTIONS(6784),
    [anon_sym_PERCENT] = ACTIONS(6784),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6784),
    [anon_sym_TILDE] = ACTIONS(6786),
    [aux_sym_prefix_op_token1] = ACTIONS(6784),
    [aux_sym_infix_op_token1] = ACTIONS(6786),
    [aux_sym_infix_op_token2] = ACTIONS(6784),
    [aux_sym_infix_op_token3] = ACTIONS(6784),
    [anon_sym_BANG_EQ] = ACTIONS(6784),
    [anon_sym_COLON_EQ] = ACTIONS(6786),
    [anon_sym_DOLLAR] = ACTIONS(6784),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6784),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6784),
    [sym_int] = ACTIONS(6784),
    [sym_xint] = ACTIONS(6786),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6786),
    [sym__newline] = ACTIONS(6786),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6786),
    [sym__tyapp_open] = ACTIONS(6786),
  },
  [STATE(5769)] = {
    [sym_block_comment] = STATE(5769),
    [sym_compiler_directive_decl] = STATE(5769),
    [sym_fsi_directive_decl] = STATE(5769),
    [sym_preproc_line] = STATE(5769),
    [sym_identifier] = ACTIONS(6750),
    [anon_sym_null] = ACTIONS(6750),
    [anon_sym_EQ] = ACTIONS(6750),
    [anon_sym_begin] = ACTIONS(6750),
    [anon_sym_COLON] = ACTIONS(6750),
    [anon_sym_do] = ACTIONS(6750),
    [anon_sym_let] = ACTIONS(6750),
    [anon_sym_let_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN] = ACTIONS(6750),
    [anon_sym_COMMA] = ACTIONS(6752),
    [anon_sym_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_QMARK] = ACTIONS(6750),
    [anon_sym_COLON_COLON] = ACTIONS(6752),
    [anon_sym_AMP] = ACTIONS(6750),
    [anon_sym_LBRACK] = ACTIONS(6750),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6752),
    [anon_sym_LBRACE] = ACTIONS(6750),
    [anon_sym_LT_AT] = ACTIONS(6750),
    [anon_sym_LT_AT_AT] = ACTIONS(6750),
    [anon_sym_CARET] = ACTIONS(6750),
    [anon_sym_or] = ACTIONS(6750),
    [anon_sym_DOT] = ACTIONS(6750),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6752),
    [anon_sym_new] = ACTIONS(6750),
    [anon_sym_return] = ACTIONS(6750),
    [anon_sym_return_BANG] = ACTIONS(6752),
    [anon_sym_yield] = ACTIONS(6750),
    [anon_sym_yield_BANG] = ACTIONS(6752),
    [anon_sym_lazy] = ACTIONS(6750),
    [anon_sym_assert] = ACTIONS(6750),
    [anon_sym_upcast] = ACTIONS(6750),
    [anon_sym_downcast] = ACTIONS(6750),
    [anon_sym_fixed] = ACTIONS(6750),
    [anon_sym_COLON_GT] = ACTIONS(6752),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6752),
    [anon_sym_for] = ACTIONS(6750),
    [anon_sym_while] = ACTIONS(6750),
    [anon_sym_while_BANG] = ACTIONS(6752),
    [anon_sym_if] = ACTIONS(6750),
    [anon_sym_fun] = ACTIONS(6750),
    [anon_sym_try] = ACTIONS(6750),
    [anon_sym_match] = ACTIONS(6750),
    [anon_sym_match_BANG] = ACTIONS(6752),
    [anon_sym_function] = ACTIONS(6750),
    [anon_sym_LT_DASH] = ACTIONS(6750),
    [anon_sym_DOT_LBRACK] = ACTIONS(6752),
    [anon_sym_use] = ACTIONS(6750),
    [anon_sym_use_BANG] = ACTIONS(6752),
    [anon_sym_do_BANG] = ACTIONS(6752),
    [anon_sym_LPAREN2] = ACTIONS(6750),
    [anon_sym_PIPE_PIPE] = ACTIONS(6750),
    [anon_sym_AMP_AMP] = ACTIONS(6750),
    [aux_sym_char_token1] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6750),
    [anon_sym_DQUOTE] = ACTIONS(6750),
    [anon_sym_AT_DQUOTE] = ACTIONS(6752),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6752),
    [sym_bool] = ACTIONS(6750),
    [sym_unit] = ACTIONS(6752),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6750),
    [sym_op_identifier] = ACTIONS(6750),
    [anon_sym_PLUS] = ACTIONS(6750),
    [anon_sym_DASH] = ACTIONS(6750),
    [anon_sym_PLUS_DOT] = ACTIONS(6750),
    [anon_sym_DASH_DOT] = ACTIONS(6750),
    [anon_sym_PERCENT] = ACTIONS(6750),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6750),
    [anon_sym_TILDE] = ACTIONS(6752),
    [aux_sym_prefix_op_token1] = ACTIONS(6750),
    [aux_sym_infix_op_token1] = ACTIONS(6752),
    [aux_sym_infix_op_token2] = ACTIONS(6750),
    [aux_sym_infix_op_token3] = ACTIONS(6750),
    [anon_sym_BANG_EQ] = ACTIONS(6750),
    [anon_sym_COLON_EQ] = ACTIONS(6752),
    [anon_sym_DOLLAR] = ACTIONS(6750),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6750),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6750),
    [sym_int] = ACTIONS(6750),
    [sym_xint] = ACTIONS(6752),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6752),
    [anon_sym_POUNDendif] = ACTIONS(6752),
    [sym__newline] = ACTIONS(6752),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6752),
    [sym__tyapp_open] = ACTIONS(6752),
  },
  [STATE(5770)] = {
    [sym_block_comment] = STATE(5770),
    [sym_compiler_directive_decl] = STATE(5770),
    [sym_fsi_directive_decl] = STATE(5770),
    [sym_preproc_line] = STATE(5770),
    [sym_identifier] = ACTIONS(6840),
    [anon_sym_null] = ACTIONS(6840),
    [anon_sym_EQ] = ACTIONS(6840),
    [anon_sym_begin] = ACTIONS(6840),
    [anon_sym_COLON] = ACTIONS(6840),
    [anon_sym_do] = ACTIONS(6840),
    [anon_sym_let] = ACTIONS(6840),
    [anon_sym_let_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN] = ACTIONS(6840),
    [anon_sym_COMMA] = ACTIONS(6838),
    [anon_sym_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_QMARK] = ACTIONS(6840),
    [anon_sym_COLON_COLON] = ACTIONS(6838),
    [anon_sym_PIPE] = ACTIONS(6840),
    [anon_sym_AMP] = ACTIONS(6840),
    [anon_sym_LBRACK] = ACTIONS(6840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(6838),
    [anon_sym_LBRACE] = ACTIONS(6840),
    [anon_sym_LT_AT] = ACTIONS(6840),
    [anon_sym_LT_AT_AT] = ACTIONS(6840),
    [anon_sym_CARET] = ACTIONS(6840),
    [anon_sym_or] = ACTIONS(6840),
    [anon_sym_DOT] = ACTIONS(6840),
    [anon_sym_LBRACE_PIPE] = ACTIONS(6838),
    [anon_sym_new] = ACTIONS(6840),
    [anon_sym_with] = ACTIONS(6840),
    [anon_sym_return] = ACTIONS(6840),
    [anon_sym_return_BANG] = ACTIONS(6838),
    [anon_sym_yield] = ACTIONS(6840),
    [anon_sym_yield_BANG] = ACTIONS(6838),
    [anon_sym_lazy] = ACTIONS(6840),
    [anon_sym_assert] = ACTIONS(6840),
    [anon_sym_upcast] = ACTIONS(6840),
    [anon_sym_downcast] = ACTIONS(6840),
    [anon_sym_fixed] = ACTIONS(6840),
    [anon_sym_COLON_GT] = ACTIONS(6838),
    [anon_sym_COLON_QMARK_GT] = ACTIONS(6838),
    [anon_sym_for] = ACTIONS(6840),
    [anon_sym_while] = ACTIONS(6840),
    [anon_sym_while_BANG] = ACTIONS(6838),
    [anon_sym_if] = ACTIONS(6840),
    [anon_sym_fun] = ACTIONS(6840),
    [anon_sym_try] = ACTIONS(6840),
    [anon_sym_match] = ACTIONS(6840),
    [anon_sym_match_BANG] = ACTIONS(6838),
    [anon_sym_function] = ACTIONS(6840),
    [anon_sym_DOT_LBRACK] = ACTIONS(6838),
    [anon_sym_use] = ACTIONS(6840),
    [anon_sym_use_BANG] = ACTIONS(6838),
    [anon_sym_do_BANG] = ACTIONS(6838),
    [anon_sym_LPAREN2] = ACTIONS(6840),
    [anon_sym_PIPE_PIPE] = ACTIONS(6840),
    [anon_sym_AMP_AMP] = ACTIONS(6840),
    [aux_sym_char_token1] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE] = ACTIONS(6840),
    [anon_sym_DQUOTE] = ACTIONS(6840),
    [anon_sym_AT_DQUOTE] = ACTIONS(6838),
    [anon_sym_DOLLAR_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(6838),
    [sym_bool] = ACTIONS(6840),
    [sym_unit] = ACTIONS(6838),
    [anon_sym_LPAREN_PIPE] = ACTIONS(6840),
    [sym_op_identifier] = ACTIONS(6840),
    [anon_sym_PLUS] = ACTIONS(6840),
    [anon_sym_DASH] = ACTIONS(6840),
    [anon_sym_PLUS_DOT] = ACTIONS(6840),
    [anon_sym_DASH_DOT] = ACTIONS(6840),
    [anon_sym_PERCENT] = ACTIONS(6840),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(6840),
    [anon_sym_TILDE] = ACTIONS(6838),
    [aux_sym_prefix_op_token1] = ACTIONS(6840),
    [aux_sym_infix_op_token1] = ACTIONS(6838),
    [aux_sym_infix_op_token2] = ACTIONS(6840),
    [aux_sym_infix_op_token3] = ACTIONS(6840),
    [anon_sym_BANG_EQ] = ACTIONS(6840),
    [anon_sym_COLON_EQ] = ACTIONS(6838),
    [anon_sym_DOLLAR] = ACTIONS(6840),
    [anon_sym_QMARK_LT_DASH] = ACTIONS(6840),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(6840),
    [sym_int] = ACTIONS(6840),
    [sym_xint] = ACTIONS(6838),
    [anon_sym_LPAREN_STAR] = ACTIONS(3),
    [sym_line_comment] = ACTIONS(5),
    [sym_xml_doc] = ACTIONS(7),
    [anon_sym_POUNDnowarn] = ACTIONS(9),
    [anon_sym_POUNDwarnon] = ACTIONS(11),
    [anon_sym_POUNDlight] = ACTIONS(13),
    [anon_sym_POUNDr] = ACTIONS(15),
    [anon_sym_POUNDload] = ACTIONS(15),
    [anon_sym_POUNDtime] = ACTIONS(15),
    [anon_sym_POUNDI] = ACTIONS(15),
    [anon_sym_POUNDhelp] = ACTIONS(15),
    [anon_sym_POUNDquit] = ACTIONS(15),
    [aux_sym_preproc_line_token1] = ACTIONS(17),
    [anon_sym_POUNDif] = ACTIONS(6838),
    [sym__newline] = ACTIONS(6838),
    [sym__multi_dollar_triple_quote_start] = ACTIONS(6838),
    [sym__tyapp_open] = ACTIONS(6838),
  },
  [STATE(5771)] = {
    [sym_block_comment] = STATE(5771),
    [sym_compiler_directive_decl] = STATE(5771),
    [sym_fsi_directive_decl] = STATE(5771),
    [sym_preproc_line] = STATE(5771),
    [sym_identifier] = ACTIONS(6653),
    [anon_sym_null] = ACTIONS(6653),
    [anon_sym_EQ] = ACTIONS(6653),
    [anon_sym_begin] = ACTIONS(6653),
    [anon_sym_COLON] = ACTIONS(6653),
    [anon_sym_do] = ACTIONS(6653),
    [anon_sym_let] = ACTIONS(6653),
    [anon_sym_let_BANG] = ACTIONS(6651),
    [anon_sym_LPAREN]